# $Revision$

# states: -1: Output file header (see comment in BEGIN section)
#         0: Searching for start of machine
#         1: Specifying name of task
#         2: Defining events
#         3: Expecting start function
#         4: Defining start function
#         5: Creating new state
#         6: Defining transitions
#         7: Expecting end function
#         8: Defining end function
#         9: Expecting suspend function (which is optional)
#        10: Defining suspend function
#        11: Copying non-machine lines

BEGIN {
    # Awk allows you to define variables on the command line, but they
    # aren't available in the BEGIN section.  Hence, we set our state
    # variable to -1, and do our BEGIN stuff when we get the first
    # line of input.
    state = -1;
}

state == -1 {
    # Set suppress_line_directives = "1" to include #line directives
    # in the output code (Mips 2.20 compiler breaks when using #line
    # directives).
    if (suppress_line_directives+0 == 0)
    {
	lstart = "#line ";
	lend = "";
    }
    else
    {
	lstart = "/* ==== #line ";
	lend = " ==== */";
    }

    state=0;

    # The number of states in the machine in the input file
    nstates=0;

    # The number of subtasks
    nsubtasks=0;

    # The number of subframes
    nsubframes=0;

    # The number of criteria
    ncriteria=0;

    # Flag indicating whether we are defining a subframe
    inframe=0;

    taskname = "";

    if (FILENAME == "-")
    {
	print "ERROR: Cannot process input from stdin.";
	exit;
    }

    for (i=length(FILENAME);i>0;i--)
      if (substr(FILENAME, i, 1) == ".")
	break;
    if (i == 0)
    {
	cfile = FILENAME ".c";
	hfile = FILENAME ".h";
    }
    else
    {
	cfile = substr(FILENAME, 1, i) "c";
	hfile = substr(FILENAME, 1, i) "h";
    }

    print "Creating output files " cfile " and " hfile;

    print lstart "1 \"" FILENAME "\"" lend	> cfile;
}

/^`/ {
    if (state == 0)
      state = 1;
    else if (state == 1)
      state = 2;
    else if (state == 2)
    {
	# Finish up this event declaration

	functions[1000*which+flines[which]] = "{";

	functions[1000*which+flines[which]+1]= \
	  "    " statevars " *state = (" statevars "*)state_object->data;";
	functions[1000*which+flines[which]+2]= \
	  "    " paramvars " *parameters = (" paramvars "*)(PO_TASK_DATA(task_entry).data);";
	functions[1000*which+flines[which]+3] = \
	  "    " private " *private = (" private "*)";
	functions[1000*which+flines[which]+4] = \
	  "      class_get_user_data(vtab_get_vehicle(vehicle_id),";
	functions[1000*which+flines[which]+5] = \
	  "                        " taskname "_user_data_handle);";
	functions[1000*which+flines[which]+6] = \
	  "    PO_DB_ENTRY *unit_entry = task_get_unit(task_entry);";
	flines[which] += 7;

	for (i=0;i<ncriteria;i++)
	{
	    functions[1000*which+flines[which]] = \
	      "    int32 " criteria[i] \
	        "_enabled = taskpri_enabled(vehicle_id, " \
		  criteria[i] "_symbol, task_entry);";
	    flines[which] += 1;
        }

	functions[1000*which+flines[which]] = \
	  "    enum " taskname "_states _initial_state = state->state;";
	functions[1000*which+flines[which]+1] = \
	  "    if (!private || !unit_entry)";
	functions[1000*which+flines[which]+2] = \
	  "      return;";
	flines[which] += 3;

	# Add debugging for everything but 'tick'
	if (which != 1)
	{
	    functions[1000*which+flines[which]] = \
	      "    CLASS_DEBUG((CLASS_USER_DATA_TYPE)" \
	      "vtab_get_vehicle(vehicle_id), " taskname \
	      "_user_data_handle, (\"%d (" taskname " " events[which] \
	      "): %s\\n\", vehicle_id, _statename(state->state)));";
	    flines[which] += 1;
	}

	# If a machine re-spawns a task, that task may already be running,
	# but have re-initialized state.  In those cases, the machine will
	# get a 'params' event, so we can test for that here, and run the
	# start function if necessary.

	if (events[which] == "params")
	{
	    functions[1000*which+flines[which]] = \
	      "    if (state->state == not_started)";
	    functions[1000*which+flines[which]+1] = \
	      "      {" taskname "_start(vehicle_id, task_entry, state_object);return;}";
	    flines[which] += 2;
	}

	functions[1000*which+flines[which]] = \
	  "    switch (state->state)";
	functions[1000*which+flines[which]+1] = "    {";
	flines[which] += 2;

	# Now start a new event
	which++;
	nevents++;
    }
    else if (state == 3)
    {
	state = 4;
	functions[flines[0]] = lstart NR+1 " \"" FILENAME "\"" lend;
	flines[0]++;
    }
    else if (state == 4)
    {
	# Finish up the definition of the start function
	functions[flines[0]] = "}";
	flines[0]++;

	# The next thing we should find is the first state
	state = 5;
    }
    else if (state == 5)
    {
	state = 6;
	nstates++;
	which = 0;
    }
    else if ((state == 7) || (state == 9))
    {
	print lstart NR+1 " \"" FILENAME "\"" lend		> cfile;
	state += 1;
    }
    else if ((state == 8) || (state == 10))
    {
	# Finish up the end or suspend function
        if ((state == 8) && (nsubtasks != 0))
          print "    _delete_subtasks(task_entry, state_object);" > cfile;

	# Finish up the suspend function
	if (state == 10)
	  print "state->state = suspended;"			> cfile;

	print "}"						> cfile;

	print lstart NR+1 " \"" FILENAME "\"" lend		> cfile;

	state += 1;
    }

    # Note that this is not an else-if, because I need to look at the "`"
    # line that just put me into this state.
    if (state == 6)
    {
	which++;
	if (which == nevents)
	{
	    state = 5;
	    next;
	}

	if (events[which] != $2)
	{
	    print "ERROR on line " NR ": Expected transition for '" \
	      events[which] "' but found '" $2 "'.";
	    exit;
	}

        # Put the break after the last block of code (except the first time)

	if (nstates > 1)
	{
	    functions[1000*which+flines[which]] = "\t    break;";
	    flines[which]++;
	}

        # Put the 'case' into the cfile

	functions[1000*which+flines[which]] = "\tcase " thisstate ":";
	functions[1000*which+flines[which]+1] = \
	  lstart NR+1 " \"" FILENAME "\"" lend;
	flines[which]+=2;
    }
    next;
}

/^START$/ {
    if (state == 2)
    {
	state = 3;
	functions[0]="/*ARGSUSED*/"
	functions[1]=lstart NR+1 " \"" FILENAME "\"" lend;
	functions[2]="static void " taskname "_start(";
	functions[3]="    int32 vehicle_id,";
	functions[4]="    PO_DB_ENTRY *task_entry, ";
	functions[5]="    TaskStateClass *state_object) ";
	functions[6]="{";
	functions[7]="    " statevars " *state = (" statevars "*)state_object->data;";
	functions[8]="    " paramvars " *parameters = (" paramvars "*)(PO_TASK_DATA(task_entry).data);";
	functions[9]="    " private " *private = (" private "*)";
	functions[10]= \
	  "      class_get_user_data(vtab_get_vehicle(vehicle_id),";
	functions[11]= \
	  "                        " taskname "_user_data_handle);";
	functions[12]= \
	  "   PO_DB_ENTRY *unit_entry = task_get_unit(task_entry);";
	functions[13]="    if (!private || !unit_entry) /* Passive error detection */";
	functions[14]="      return;";
	functions[15] = \
	      "    CLASS_DEBUG((CLASS_USER_DATA_TYPE)" \
	      "vtab_get_vehicle(vehicle_id), " taskname \
	      "_user_data_handle, (\"%d (" taskname " START" \
	      "): %s\\n\", vehicle_id, _statename(state->state)));";
	functions[16] = "    if (state->state == ended) return;";
	functions[17] = "";

	flines[0] = 18;
	which = 0;
	next;
    }
    else
    {
	print "ERROR on line " NR ": Not expecting START here";
	exit;
    }
}

/^END$/ {
    if (state != 5)
    {
	print "ERROR on line " NR ": Not expecting END here";
	exit;
    }

    print "#include <libtask.h>"				> hfile;

    # Print the enumerated type of states

    print "enum " taskname "_states {"				> hfile;
    print "    not_started,"					> hfile;
    print "    ended,"						> hfile;
    print "    suspended,"					> hfile;
    for (i=0;i<nstates-1;i++)
      print "    " states[i] ","				> hfile;
    print "    " states[nstates-1]                              > hfile;
    print "};"							> hfile;
    print ""							> hfile;

    # Prototype the functions

    print "extern void " taskname "_init_fsm("			> hfile;
    print "    uint32             safmodel,"			> hfile;
    print "    TASK_PREDICATE_FCN pred,"			> hfile;
    print "    int32              n_b,"				> hfile;
    print "    uint32             b[],"				> hfile;
    print "    int32              n_a,"				> hfile;
    print "    uint32             a[]);"			> hfile;
    print "extern int32 " taskname "_ended("			> hfile;
    print "    int32           vehicle_id,"			> hfile;
    print "    PO_DB_ENTRY    *task_entry,"			> hfile;
    print "    TaskStateClass *state_object);"			> hfile;

    print "static void " taskname "_start("		        > cfile;
    print "    int32           vehicle_id,"			> cfile;
    print "    PO_DB_ENTRY    *task_entry,"			> cfile;
    print "    TaskStateClass *state_object);"			> cfile;
    for (i=1;i<nevents;i++)
    {
        for (j=1;j<declines[i]-1;j++)
          print functions[1000*i+j]				> cfile;
        print functions[1000*i+j] ";"				> cfile;
    }
    print "static void " taskname "_end("			> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object);"			> cfile;
    print "static void " taskname "_suspend("			> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object);"			> cfile;
    if (nsubtasks != 0)
    {
        print "static void _suspend_subtasks("			> cfile;
	print "    PO_DB_ENTRY *task_entry,"			> cfile;
	print "    TaskStateClass *state_object);"		> cfile;
	print "static void _delete_subtasks("			> cfile;
	print "    PO_DB_ENTRY *task_entry,"			> cfile;
	print "    TaskStateClass *state_object);"		> cfile;
	print "static void _update_subtasks("			> cfile;
	print "    int32 vehicle_id,"				> cfile;
	print "    PO_DB_ENTRY *task_entry,"			> cfile;
	print "    TaskStateClass *state_object, "		> cfile;
	print "    " statevars " *state,"			> cfile;
	print "    " paramvars " *parameters,"			> cfile;
	print "    " private " *private);"			> cfile;
    }
    print ""                                                    > cfile;

    print "void " taskname "_init_fsm("				> cfile;
    print "    uint32             safmodel,"			> cfile;
    print "    TASK_PREDICATE_FCN pred,"			> cfile;
    print "    int32              n_b,"				> cfile;
    print "    uint32             b[],"				> cfile;
    print "    int32              n_a,"				> cfile;
    print "    uint32             a[])"				> cfile;
    print "{"							> cfile;
    print "    if (!pred) pred = (TASK_PREDICATE_FCN) " \
      taskname "_ended;"					> cfile;
    print "    task_register(safmodel, params, pred, " \
      taskname "_ended, " taskname "_start, tick, "		> cfile;
    print "                  "  taskname "_end, " \
      taskname "_suspend,"					> cfile;
    print "                  sizeof(" statevars \
      "), n_b, b, n_a, a, \"" classifications "\");"		> cfile;
    for (i=0;i<ncriteria;i++)
    {
        print "    " criteria[i] "_symbol = reader_get_symbol" \
          "(\"" criteria[i] "\");"				> cfile;
        print "    taskpri_register_for_group(safmodel,"	> cfile;
        print "      TASKPRI_OFFSET(" statevars ")," \
	  criteria[i] "_symbol," criteria[i] "_stack);"		> cfile;
    }

    print "}"							> cfile;

    # Declare a function to translate state numbers to strings

    print "static char *_statename("				> cfile;
    print "    enum " taskname "_states state)"			> cfile;
    print "{"							> cfile;
    print "    switch(state)"					> cfile;
    print "    {"						> cfile;
    print "        case not_started: return \"not_started\";"	> cfile;
    print "        case ended: return \"ended\";"	> cfile;
    print "        case suspended: return \"suspended\";"	> cfile;
    for (i=0;i<nstates;i++)
      print "        case " states[i] ": return \"" states[i] "\";" > cfile;
    print "    }"						> cfile;
    print "    return \"Unknown\";"				> cfile;
    print "}"							> cfile;
    print ""							> cfile;


    # Declare the ended predicate

    print "/*ARGSUSED*/"                                        > cfile;
    print "int32 " taskname "_ended("				> cfile;
    print "    int32           vehicle_id,"			> cfile;
    print "    PO_DB_ENTRY    *task_entry,"			> cfile;
    print "    TaskStateClass *state_object)"			> cfile;
    print "{"							> cfile;
    print "    " statevars " *state = (" statevars \
      "*)state_object->data;"					> cfile;
    print "    return state->state == ended;"			> cfile;
    print "}"							> cfile;

    # Print the functions for each of the events

    for (i=0;i<nevents;i++)
    {
	for (j=0;j<flines[i];j++)
	  print functions[1000*i+j]				> cfile;

	if (i > 0)
	{
	    print "\t    break;"				> cfile;
	    print "\t}"						> cfile;

	    # Generate debugging for state transitions
	    print "\tif (_initial_state != state->state)"	> cfile;
	    print "\t  CLASS_DEBUG((CLASS_USER_DATA_TYPE)" \
	      "vtab_get_vehicle(vehicle_id), " taskname \
	      "_user_data_handle, (\"%d (" taskname " " events[i] \
	      "): %s -> %s\\n\", vehicle_id, _statename(_initial_state), " \
	      "_statename(state->state)));"			> cfile;

            # Don't bother updating subtasks after a tick
            if ((i > 1) && (nsubtasks != 0))
	      print "_update_subtasks(vehicle_id, task_entry, " \
	        "state_object, state, parameters, private);"	> cfile;

	    print "}"						> cfile;
	}

	print ""						> cfile;
    }

# Declare the end function

    state = 7;
    print "/*ARGSUSED*/"					> cfile;
    print lstart NR+1 " \"" FILENAME "\"" lend			> cfile;
    print "static void " taskname "_end("			> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry, "			> cfile;
    print "    TaskStateClass *state_object) "			> cfile;
    print "{"							> cfile;
    print "    " statevars " *state = (" statevars \
      "*)state_object->data;"					> cfile;
    print "    " paramvars " *parameters = (" paramvars \
      "*)(PO_TASK_DATA(task_entry).data);"			> cfile;
    print "    " private " *private = (" private "*)"		> cfile;
    print "      class_get_user_data(vtab_get_vehicle(vehicle_id)," > cfile;
    print "                        " taskname "_user_data_handle);" > cfile;
    print "    if (!private) /* Passive error detection */" > cfile;
    print "      return;"					> cfile;
    print "state->state = ended;"				> cfile;
    print "    CLASS_DEBUG((CLASS_USER_DATA_TYPE)" \
      "vtab_get_vehicle(vehicle_id), " taskname \
      "_user_data_handle, (\"%d (" taskname " END" \
      ")\\n\", vehicle_id));"					> cfile;

    next;
}    

/^SUSPEND$/ {
    if (state != 9)
    {
	print "ERROR on line " NR ": Not expecting SUSPEND here";
	exit;
    }

# Declare the suspend function

    print "/*ARGSUSED*/"					> cfile;
    print "static void " taskname "_suspend("			> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry, "			> cfile;
    print "    TaskStateClass *state_object) "			> cfile;
    print "{"							> cfile;
    print "    " statevars " *state = (" statevars \
      "*)state_object->data;"					> cfile;
    print "    " paramvars " *parameters = (" paramvars \
      "*)(PO_TASK_DATA(task_entry).data);"			> cfile;
    print "    " private " *private = (" private "*)"		> cfile;
    print "      class_get_user_data(vtab_get_vehicle(vehicle_id)," > cfile;
    print "                        " taskname "_user_data_handle);" > cfile;
    print "    PO_DB_ENTRY *unit_entry = task_get_unit(task_entry);" > cfile;
    if (nsubtasks != 0)
      print "    _suspend_subtasks(task_entry, state_object);"	> cfile;
    print "    if (!private || !unit_entry) /* Passive error detection */" > cfile;
    print "      return;"					> cfile;
    print "    CLASS_DEBUG((CLASS_USER_DATA_TYPE)" \
      "vtab_get_vehicle(vehicle_id), " taskname \
      "_user_data_handle, (\"%d (" taskname " SUSPEND" \
      ")\\n\", vehicle_id));"					> cfile;

    next;
}    

# This rule catches the beginning of a subframe declaration
#	SUBFRAME: name member instruction [opaque] [destroy] [CLEANUP]
(state == 1) && /^SUBFRAME:/ {
    if ((NF < 4) || (NF > 7))
    {
	print "ERROR on line " NR ": Expect SUBFRAME: name member instruction [opaque] [destroy] [CLEANUP]";
	exit;
    }
    subframes[nsubframes] = $2;
    submember[$2] = $3;
    nframetasks[$2] = 0;
    isframe[$2] = 1;
	if ($0 ~ /CLEANUP/)
	  docleanup[$2] = 1;
	else
	  docleanup[$2] = 0;
    save_frame = $2;
    inframe = 1;
    print "Subframe: " $2;

    nsubframes += 1;

    if (nsubframes == 1)
    {
	print "static void get_current_taskframe( " 		> cfile;
	print "    ObjectID *sub_unit_id, " 			> cfile;
	print "    PO_DATABASE *db," 				> cfile;
	print "    ObjectID *next_task_frame)" 			> cfile;
	print "{" 						> cfile;
	print "    PO_DB_ENTRY *unit_entry;" 			> cfile;
	print "" 						> cfile;
	print "    if ((unit_entry = " 				> cfile;
	print "         po_get_object(db, sub_unit_id)) &&" 	> cfile;
	print "        (PO_OBJECT_CLASS(unit_entry) == objectClassUnit))" > cfile;
	print "      *next_task_frame = "			> cfile;
	print "         PO_UNIT_DATA(unit_entry).taskFrameStack;"> cfile;
	print "" 						> cfile;
	print "    /* Next task frame should already bzero before being passed in */" > cfile;
	print "}"						> cfile;
	print "" 						> cfile;
    }
    # Declare the subframe spawn function 
    print "/*ARGSUSED*/"					> cfile;
    print "static void spawn_" $2 "("				> cfile;
    print "    int32 ignored,"					> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object, "			> cfile;
    print "    " statevars " *state,"				> cfile;
    print "    " paramvars " *parameters,"			> cfile;
    print "    " private " *private,"				> cfile;
    print "    ObjectID *sub_unit_id,"				> cfile;
    print "    int32 which,"					> cfile;
    print "    ObjectID *sup_unit_id,"				> cfile;
    print "    ObjectID *ignored2)"				> cfile;
    print "{"							> cfile;
    print "    PO_DB_ENTRY *thisframe, *subframe;"		> cfile;
    print "    PO_BUFFER(TaskFrameClass, subframe_object, buf);" > cfile;
    print "    PO_DATABASE *db = task_get_db(task_entry);"	> cfile;
    print "    ObjectID *spawn_id = &state->" $3 "[which];"	> cfile;
    print ""							> cfile;
    print "    if (which >= ARRAYLENGTH(state->" $3 "))"	> cfile;
    print "    { printf(\"Instantiation %d of " $2 \
      " exceeds maximum\\n\", which+1); return; }"		> cfile;
    print ""							> cfile;
    print "    strncpy(subframe_object->name, \"" $2 \
      "\", maxTaskFrameNameLength);"				> cfile;
    print "    subframe_object->name[maxTaskFrameNameLength-1] = 0;" > cfile;
    if ($0 ~ /opaque/)
      print "    subframe_object->opaque = 1;"			> cfile;
    else
      print "    subframe_object->opaque = 0;"			> cfile;
    if ($0 ~ /destroy/)
      print "    subframe_object->destroyWhenDone = 1;"		> cfile;
    else
      print "    subframe_object->destroyWhenDone = 0;"		> cfile;
    print "    subframe_object->instruction = " $4 ";"		> cfile;
    print "    subframe_object->commandingUnit = *sup_unit_id;"	> cfile;
    print "    subframe_object->name[maxTaskFrameNameLength-1] = 0;" > cfile
    print "    get_current_taskframe(sub_unit_id, db, " \
      "&subframe_object->nextStackFrame);" 			> cfile;
    print "    bzero(&subframe_object->previousMissionFrame, " \
      "sizeof(ObjectID));"					> cfile;
    print "    bzero(&subframe_object->unit, sizeof(ObjectID));" > cfile;
    # Set the sponsoring task onlyh if we are pushing the frame
    if ($4 == "TIIPopNone")
    {
	print "    if (NS_OBJECT_IDS_EQUAL(*sub_unit_id, *sup_unit_id))" > cfile;
	print "      subframe_object->sponsoringTask = " \
	  "PO_OBJECT_ID(task_entry);"				> cfile;
	print "    else"						> cfile;
    }
    print "      bzero(&subframe_object->sponsoringTask, " \
      "sizeof(ObjectID));"					> cfile;
    print "    subframe_object->logicStack[0] = " \
      "taskFrameLogicStackSTOP;"				> cfile;
    print "    subframe_object->etaskCount = 0;"		> cfile;
    print ""							> cfile;
    print "    subframe ="					> cfile;
    print "      po_create_object(db, 0, " \
      "objectClassTaskFrame, FALSE, subframe_object, " \
	"PRO_PO_TASK_FRAME_CLASS_SIZE(0), 0);"			> cfile;
    print "    if (!subframe) return;"				> cfile;
    print ""							> cfile;
    print "    *spawn_id = PO_OBJECT_ID(subframe);"		> cfile;
    print ""							> cfile;
    print lstart NR+1 " \"" FILENAME "\"" lend			> cfile;
    next;
}

# This rule catches subtasks included in subframes
#	INCLUDE: subtask
(state == 1) && /^INCLUDE:/ {
    if (NF != 2)
    {
	print "ERROR on line " NR ": Expect INCLUDE: subtask";
	exit;
    }

    if (!inframe)
    {
	print "ERROR on line " NR ": Not in SUBFRAME definition";
	exit;
    }

    subincluded[$2] = 1;
    frametasks[nframetasks[save_frame]] = $2;
    nframetasks[save_frame] += 1;
    print "\tIncludes " $2;

    print "    spawn_" $2 "((int32)0, vehicle_id, task_entry, " \
	"state_object, state, parameters, private, " \
	  "spawn_id, which, (ObjectID *)0, sub_unit_id);"	> cfile;
    next;
}

# This rule catches the end of a subframe declaration
#	END_SUBFRAME
(state == 1) && /^END_SUBFRAME$/ {
    print ""							> cfile;
    print "    subframe_object->unit = *sub_unit_id;"		> cfile;
    print "    subframe_object->assigned = 1;"			> cfile;
    print "    subframe ="					> cfile;
    print "      po_change_object(db, subframe, subframe_object, " \
      "PRO_PO_TASK_FRAME_CLASS_SIZE(0));"			> cfile;
    print "    if (subframe)"					> cfile;
    print "      *spawn_id = PO_OBJECT_ID(subframe);"		> cfile;
    print "}"							> cfile;
    print ""							> cfile;
    inframe = 0;
    next;
}

# This rule catches the beginning of a subtask declaration
#	SUBTASK: name member model params refcount init [background]
(state == 1) && /^SUBTASK:/ {
    if ((NF < 7) || (NF > 9))
    {
	print "ERROR on line " NR ": Expect SUBTASK: name member model params refcount init [BCKGRND] [OPTIONAL]";
	exit;
    }

    if (!($0 ~ /BCKGRND/) && ($0 ~ /OPTIONAL/))
    {
	print "ERROR on line " NR ": Can't have optional without background [BCKGRND] [OPTIONAL]";
	exit;
    }

    subtasks[nsubtasks] = $2;
    submember[$2] = $3;
    if ($0 ~ /BCKGRND/) 
    {
	save_background = 1;
	subincluded[$2] = 1;
	if ($0 ~ /OPTIONAL/)
	  print "Optional Background task: " $2;
	else
	  print "Background task: " $2;
    }
    else
    {
	save_background = 0;
	subincluded[$2] = 0;
	print "Subtask: " $2;
    }



    nsubtasks += 1;

    # Save these for the END_SUBTASK
    save_init = $7;
    save_member = $3;



    # Declare the subtask spawn function
    print "/*ARGSUSED*/"					> cfile;
    print "static void spawn_" $2 "("				> cfile;
    print "    int32 is_update,"				> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object, "			> cfile;
    print "    " statevars " *state,"				> cfile;
    print "    " paramvars " *parameters,"			> cfile;
    print "    " private " *private,"				> cfile;
    print "    ObjectID *unit_or_frame_id,"			> cfile;
    print "    int32 which,"					> cfile;
    print "    ObjectID *ignored,"				> cfile;
    print "    ObjectID *sub_unit_id) /*Only for included frames*/"> cfile;
    print "{"							> cfile;
    print "    PO_BUFFER(TaskClass, subtask_object, buf);"	> cfile;
    print "    PO_DB_ENTRY *bgt_entry;"				> cfile;
    print "    " $5 " *subtask = (" $5 " *)subtask_object->data;" > cfile;
    print "    PO_DATABASE *db = task_get_db(task_entry);"	> cfile;
    print "    PO_DB_ENTRY *unit_or_frame_entry=0;"		> cfile;
    print ""							> cfile;
    print "    bzero(buf, sizeof(buf));"			> cfile;
    print "    if (is_update)"					> cfile;
    print "    {"						> cfile;
    print "        PO_DB_ENTRY *sub_task_entry;"		> cfile;
    print "        if (sub_task_entry = po_get_object(db, &state->" \
      save_member "[which]))"					> cfile;
    print "          unit_or_frame_entry = task_get_unit(sub_task_entry);" > cfile;
    print "    }"						> cfile;
    print "    else"						> cfile;
    print "      unit_or_frame_entry = po_get_object(db, unit_or_frame_id);" > cfile;
    print "    if (!unit_or_frame_entry) return;"		> cfile;
    print "    if (which >= ARRAYLENGTH(state->" save_member "))" > cfile;
    print "    { printf(\"Instantiation %d of " $2 \
      " exceeds maximum\\n\", which+1); return; }"		> cfile;
    print ""							> cfile;

    print "    subtask_object->model = " $4 ";"			> cfile;
    print "    subtask_object->refcount = " $6 ";"		> cfile;
    print "    subtask_object->size = sizeof(" $5 ");"		> cfile;
    print ""							> cfile;
	
    if (save_background)
    {
	print "    bgt_entry = taskfr_get_background_task(db,"	> cfile;
	print "        unit_or_frame_entry, subtask_object->model);" > cfile;
	print "    if (bgt_entry)"				> cfile;
	print "      state->" save_member "[which] = PO_OBJECT_ID(bgt_entry);" > cfile;
	print "    else"					> cfile;

        if ($0 ~ /OPTIONAL/)
	print "      return;"	         			> cfile
        else 
	print "      printf(\"" taskname " unable to find background task " $4 "\\n\");" > cfile;
    }
    else
      print "    if (is_update)"					> cfile;
# yes, the preceeding else body is exactly what it looks like.  nasty.
    
    print "    {"						> cfile;
    print "        if (!task_subtask_parameters(task_entry,"	> cfile;
    print "               &state->" save_member "[which],"	> cfile;
    print "               (ADDRESS)subtask))"			> cfile;
    print "          return;"					> cfile;
    print "    }"						> cfile;
    print ""							> cfile;
    print lstart NR+1 " \"" FILENAME "\"" lend			> cfile;
    next;
}

# This rule catches the end of a subtask declaration
#	END_SUBTASK
(state == 1) && /^END_SUBTASK$/ {

	if (save_background)
	{
	    print "      task_update_subtask(task_entry,"	> cfile;
	    print "                          subtask_object, "	> cfile;
	    print "        &state->" save_member "[which]);"	> cfile;
	}
	else 
	{
	    print "    if (is_update)"				> cfile;
	    print "      task_update_subtask(task_entry,"	> cfile;
	    print "                          subtask_object, "	> cfile;
	    print "        &state->" save_member "[which]);"	> cfile;
	    print "    else"					> cfile;
	    print "      task_spawn_subtask(task_entry, unit_or_frame_id," > cfile;
	    print "                         subtask_object, " save_init "," > cfile;
	    print "        &state->" save_member "[which]);"	> cfile;
	}
	print "}"						> cfile;
	print ""						> cfile;

    next;
}

# This rule catches the beginning of a CRITERIA declaration
#	CRITERIA: group
(state == 1) && /^CRITERIA:/ {
    if (NF != 2)
    {
	print "ERROR on line " NR ": Expect CRITERIA: group";
	exit;
    }

    criteria[ncriteria] = $2;
    print "Criteria: " $2;

    ncriteria += 1;

    # Declare the variable which will hold the libreader symbol for
    # this group name
    print "static char *" $2 "_symbol;"				> cfile;

    # Declare the beginning of the logic stack for this criteria
    print "static TASKPRI_LOGIC " $2 "_stack[] = {"		> cfile;
    next;
}

# This rule catches the end of a CRITERIA declaration
#	END_CRITERIA
(state == 1) && /^END_CRITERIA$/ {
    print "TASKPRI_END };"					> cfile;
    next;
}

# This rule catches the body of the subtask and criteria declarations
(state == 1) && (taskname != "") {
    print							> cfile;
    next;
}

# This rule catches the command POP_OWN_TASKFRAME in the END body
(state == 8) && /^[ \t]*POP_OWN_TASKFRAME$/ {
    print "task_pop_ownframe(task_entry);"			> cfile;
    next;
}

# This rule catches POP_OWN_TASKFRAME in illegal places
/POP_OWN_TASKFRAME/ {
    print "ERROR on line " NR ": POP_OWN_TASKFRAME only allowed in END body";
    exit;
}

# This rule matches the CLEAR command:
#	CLEAR subframe {which}

/^[ \t]*CLEAR / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect CLEAR subframe {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    # Clear the tasks

    if (isframe[$2])
    {
	for (i=0;i<nframetasks[$2];i++)
	{
	    if (state == 8)
	    {
		print "bzero((char *)&state->" \
		  submember[frametasks[i]] "[" w "], sizeof(ObjectID));" > cfile;
	    }
	    else
	    {
		functions[1000*which+flines[which]] = \
		  "bzero((char *)&state->" \
		    submember[frametasks[i]] "[" w "], sizeof(ObjectID));";
		flines[which]++;
	    }
	}
    }

    # Clear the frame
    if (state == 8)
    {
	print "bzero((char *)&state->" \
	  submember[$2] "[" w "], sizeof(ObjectID));" > cfile;
    }
    else
    {
	functions[1000*which+flines[which]] = \
	  "bzero((char *)&state->" \
	    submember[$2] "[" w "], sizeof(ObjectID));";
	flines[which]++;
    }

    next;
}

# This rule matches UPDATE commands:

/^[ \t]*UPDATE / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect UPDATE subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    line = "spawn_" $2 "((int32)1, vehicle_id, task_entry, " \
      "state_object, state, parameters, private, (ObjectID *) 0, " \
	"(int32)(" w "), (ObjectID *)0, (ObjectID *)0);";

    if ((state == 8) || (state == 10))
      print line						> cfile;
    else
    {
	functions[1000*which+flines[which]] = line;
	flines[which]++;
    }
    next;
}

(state == 0) || (state >= 8) {
    print							> cfile;
    next;
}

state == 1 {

    # Grab the task name

    if ((NF < 4) || (NF > 5))
    {
	print "ERROR on line " NR ": Expect taskname PARAMETERS_DATA_TYPE STATE_DATA_TYPE PRIVATE_DATA_TYPE classifications";
	exit;
    }
    taskname = $1;
    paramvars = $2;
    statevars = $3;
    private = $4;
    if (NF == 5)
      classifications = $5;
    else
      classifications = $1;
    print "Defining task " taskname "...";
    print "Classifications: " classifications;
    which = 1;
    nevents = 1;

    print "#define SPAWN_UNIT_ENTRY \\"				> cfile;
    print "  (sub_unit_id ? po_get_object(db, sub_unit_id) : " \
      "unit_or_frame_entry)"					> cfile;
    print ""							> cfile;

    next;
}

state == 2 {
# Is this first line of the definition of the event?
    if (events[which] == "")
    {

# Find the event name.  It is the set of characters in the first field
# just prior to the '('.

	for (i=1;i<=length($1);i++)
	{
	    if (substr($1, i, 1) == "(")
	    {
		events[which] = substr($1, 1, i-1);
		break;
	    }
	}
	if (events[which] == "")
	{
	    print "ERROR parsing line " NR ": Expected '(' in " $1;
	    print "Note: no space is allowed between event name and '('";
	    exit;
	}
	print "Event: " events[which];

# Find the '(' again, but this time determine its location in the context
# of the whole line.

	for (i=1;i<=length($0);i++)
	{
	    if (substr($0, i, 1) == "(")
	      break;
	}

# Make the function declaration

	functions[1000*which] = lstart NR+1 " \"" FILENAME "\"" lend;
        functions[1000*which+1] = "static void " events[which] "(";
	functions[1000*which+2] = "    int32 vehicle_id,";
	functions[1000*which+3] = "    PO_DB_ENTRY *task_entry,";
	if (substr($0, i+1, 1) == ")")
        {
	    functions[1000*which+4] = "    TaskStateClass *state_object)";
	    flines[which] = 5;
            declines[which] = 5;
        }
	else
        {
	    functions[1000*which+4] = "    TaskStateClass *state_object,";
            functions[1000*which+5] = "    " substr($0, i+1, length($0)-i);
            flines[which] = 6;
            declines[which] = 6;
        }
    }
    else
# This is after the first line of the declaration of the event.  Just
# copy the lines verbatim.
    {
	functions[1000*which+flines[which]] = \
	  lstart NR+1 " \"" FILENAME "\"" lend;
	functions[1000*which+flines[which]+1] = $0;
	flines[which] += 2;
    }
    next;
}

state == 5 {
# This line should name a state.  Just take the first token, to allow
# comments to appear thereafter.
    if (states[nstates] == "")
    {
	states[nstates] = $1;
	thisstate = $1;
	print "State: " $1;
    }
    next;
}

# At this point we should be in state 4, 6, or 7

# See if there is a transition to END
/^[ \t]*\^END;/ {
    functions[1000*which+flines[which]] = \
      "{ " taskname "_end(vehicle_id, task_entry, state_object); return; }";
    flines[which]++;
    next;
}

# The next rule matches transitions which look like this:
#	^new_state;

/^[ \t]*\^[^ \t;]*;/ {

    # In the start, end, and suspend clauses, transitions make us return.
    # In events, we just break (so that _update_subtasks can be called).
    if (state == 6)
      action = "break";
    else
      action = "return";

    functions[1000*which+flines[which]] = \
      "{ state->state = " substr($1, 2, length($1)-1) " " action "; }";
    flines[which]++;
    next;
}

# This rule matches SPAWN commands:

/^[ \t]*SPAWN / {
    if ((NF < 3) || (NF > 4))
    {
	print "ERROR on line " NR ": Expect SPAWN who|SELF subtask {which}";
	exit;
    }

    if (NF == 3)
      w = 0;
    else
      w = $4;

    if (($2 == "SELF") || ($2 == "BACKGROUND"))
    {
	if (isframe[$3] || ($2 == "BACKGROUND"))
	  line = "spawn_" $3 "( (int32)0, vehicle_id, task_entry, " \
	    "state_object, state, parameters, private, " \
	      "&PO_OBJECT_ID(unit_entry), (int32)(" w "), " \
		"&PO_OBJECT_ID(unit_entry), (ObjectID *)0);";
	else
	  line = "spawn_" $3 "( (int32)0, vehicle_id, task_entry, " \
	    "state_object, state, parameters, private, " \
	      "&PO_TASK_DATA(task_entry).frame, " \
		"(int32)(" w "), &PO_OBJECT_ID(unit_entry), " \
		  "&PO_OBJECT_ID(unit_entry));";
    }
    else
      line = "spawn_" $3 "( (int32)0, vehicle_id, task_entry, " \
	"state_object, state, parameters, private, " $2 ", " \
	  "(int32)(" w "), &PO_OBJECT_ID(unit_entry), (ObjectID *)0);";

    functions[1000*which+flines[which]] = line;
    flines[which]++;
    next;
}

# This rule matches STOP commands:
#	STOP subtask {which}

/^[ \t]*STOP / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect STOP subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    if (isframe[$2])
      functions[1000*which+flines[which]] = \
	"task_stop_subframe(task_entry, &state->" submember[$2] "[" w "]);";
    else
      functions[1000*which+flines[which]] = \
	"task_stop_subtask(task_entry, &state->" submember[$2] "[" w "]);";

    flines[which]++;
    next;
}

# This rule matches SUSPEND commands:
#	SUSPEND subtask {which}

/^[ \t]*SUSPEND / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect SUSPEND subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    functions[1000*which+flines[which]] = \
      "task_suspend_subtask(task_entry, &state->" submember[$2] "[" w "]);";

    flines[which]++;
    next;
}

# This rule matches RESUME commands:
#	RESUME subtask {which}

/^[ \t]*RESUME / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect RESUME subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    functions[1000*which+flines[which]] = \
      "task_resume_subtask(task_entry, &state->" submember[$2] "[" w "]);";

    flines[which]++;
    next;
}

# This rule matches DELETE commands:
#	DELETE subtask {which}

/^[ \t]*DELETE / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect DELETE subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    functions[1000*which+flines[which]] = \
      "task_delete_subtask(task_entry, &state->" submember[$2] "[" w "]);";

    flines[which]++;
    next;
}

# This rule matches FINISHED predicates:
#	FINISHED subtask {which}

/^[ \t]*FINISHED / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect FINISHED subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    functions[1000*which+flines[which]] = \
      "task_complete(task_entry, &state->" submember[$2] "[" w "])";

    flines[which]++;
    next;
}

# This rule matches ENDED predicates:
#	ENDED subtask {which}

/^[ \t]*ENDED / {
    if ((NF < 2) || (NF > 3))
    {
	print "ERROR on line " NR ": Expect ENDED subtask {which}";
	exit;
    }

    if (NF == 2)
      w = 0;
    else
      w = $3;

    functions[1000*which+flines[which]] = \
      "task_ended(task_entry, &state->" submember[$2] "[" w "])";

    flines[which]++;
    next;
}

# This rule matches PERIOD commands:
#	PERIOD <msecs>

/^[ \t]*PERIOD / {
    if (NF != 2)
    {
	print "ERROR on line " NR ": Expect PERIOD <milliseconds>";
	exit;
    }

    functions[1000*which+flines[which]] = "task_set_period(task_entry, "$2");";

    flines[which]++;
    next;
}

# Just copy out those lines which are not transitions or special commands.
{
    functions[1000*which+flines[which]] = $0;
    flines[which]++;
    next;
}    

END {
    # If the task definition did not include a suspend function, put
    # one in now.

    if (state == 9)
    {
    print "/*ARGSUSED*/"					> cfile;
    print "static void " taskname "_suspend("			> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object) "			> cfile;
    print "{"							> cfile;
    print "    " statevars " *state = (" statevars \
      "*)state_object->data;"					> cfile;
    print "    " paramvars " *parameters = (" paramvars \
      "*)(PO_TASK_DATA(task_entry).data);"			> cfile;
    print "    " private " *private = (" private "*)"		> cfile;
    print "      class_get_user_data(vtab_get_vehicle(vehicle_id)," > cfile;
    print "                        " taskname "_user_data_handle);" > cfile;
    print "    PO_DB_ENTRY *unit_entry = task_get_unit(task_entry);" > cfile;
    print "    if (!private || !unit_entry) /* Passive error detection */"	> cfile;
    print "      return;"					> cfile;
    if (nsubtasks != 0)
      print "    _suspend_subtasks(task_entry, state_object);"	> cfile;
    print "    CLASS_DEBUG((CLASS_USER_DATA_TYPE)" \
      "vtab_get_vehicle(vehicle_id), " taskname \
      "_user_data_handle, (\"%d (" taskname " SUSPEND" \
      ")\\n\", vehicle_id));"					> cfile;
    print "state->state = suspended;"				> cfile;
    print "}"							> cfile;
    }

    # If there aren't any subtasks, we're done
    if (nsubtasks == 0)
      exit;

    # Declare the subtask utility functions

    print "static void _suspend_subtasks("			> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object)"			> cfile;
    print "{"							> cfile;
    print "    int32 i;"					> cfile;
    print "    " statevars " *state = (" statevars \
      "*)state_object->data;"					> cfile;
    for (i=0;i<nsubtasks;i++)
    {
	# Skip subtasks which are included in spawned frames
	if (subincluded[subtasks[i]])
	  continue;

	print "    for (i=0;i<ARRAYLENGTH(state->" \
	  submember[subtasks[i]] ");i++)"			> cfile;
	print "      task_suspend_subtask(task_entry, " \
	  "&state->" submember[subtasks[i]] "[i]);"		> cfile;
    }
    print ";}"							> cfile;
    print "static void _delete_subtasks("			> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object)"			> cfile;
    print "{"							> cfile;
    print "    int32 i;"					> cfile;
    print "    " statevars " *state = (" statevars \
      "*)state_object->data;"					> cfile;
    for (i=0;i<nsubtasks;i++)
    {
	print "    for (i=0;i<ARRAYLENGTH(state->" \
	  submember[subtasks[i]] ");i++)"			> cfile;

	# Just forget about subtasks which are included in spawned frames
	if (subincluded[subtasks[i]])
	print "      bzero(&state->" submember[subtasks[i]] "[i], sizeof(ObjectID));"		> cfile;
    else
	print "      task_delete_subtask(task_entry, " \
	  "&state->" submember[subtasks[i]] "[i]);"		> cfile;
    }
    for (i=0;i<nsubframes;i++)
    {
        if (!docleanup[subframes[i]])
          continue;

	print "    for (i=0;i<ARRAYLENGTH(state->" \
	  submember[subframes[i]] ");i++)"			> cfile;
	print "      task_stop_subframe(task_entry, &state->" \
	  submember[subframes[i]] "[i]),"		> cfile;
	print "      bzero(&state->" \
	  submember[subframes[i]] "[i], sizeof(ObjectID));"		> cfile;
	}
    print ";}"							> cfile;

    print "/*ARGSUSED*/"					> cfile;
    print "static void _update_subtasks("			> cfile;
    print "    int32 vehicle_id,"				> cfile;
    print "    PO_DB_ENTRY *task_entry,"			> cfile;
    print "    TaskStateClass *state_object, "			> cfile;
    print "    " statevars " *state,"				> cfile;
    print "    " paramvars " *parameters,"			> cfile;
    print "    " private " *private)"				> cfile;
    print "{"							> cfile;
    print "    int32 i;"					> cfile;
    for (i=0;i<nsubtasks;i++)
    {
	print "    for (i=0;i<ARRAYLENGTH(state->" \
	  submember[subtasks[i]] ");i++)"			> cfile;
	print "      spawn_" subtasks[i] "((int32)1, vehicle_id, " \
	  "task_entry, state_object, state, parameters, private, " \
	    "(ObjectID *)0, i, (ObjectID *)0, (ObjectID *)0);"	> cfile;
    }
    print ";}"							> cfile;
}

