# $Revision$

# states: 0: Searching for START
#         1: Reading start body
#         2: Expecting state name
#         3: Reading state body

BEGIN {
    state = 0;
    marks = 0;
    nstates = 0;
    ntrans = 0;

# List of lozenge locations.  Note that START and END count as states.
# To accommodate larger machines, add more clever locations to this list
# and increase 'maxstates' appropriately.

    xloc[0] = 340; yloc[0] = 340;
    xloc[1] = 340; yloc[1] = 500;
    xloc[2] = 340; yloc[2] = 180;
    xloc[3] = 500; yloc[3] = 340;
    xloc[4] = 180; yloc[4] = 340;
    xloc[5] = 500; yloc[5] = 500;
    xloc[6] = 180; yloc[6] = 180;
    xloc[7] = 180; yloc[7] = 500;
    xloc[8] = 500; yloc[8] = 180;
    maxstates = 9;

    if (FILENAME == "-")
    {
	print "ERROR: Cannot process input from stdin.";
	exit;
    }

    for (i=length(FILENAME);i>0;i--)
      if (substr(FILENAME, i, 1) == ".")
	break;
    if (i == 0)
      file = FILENAME ".fig";
    else
      file = substr(FILENAME, 1, i) "fig";

    print "Creating output file " file;

# Write the FIG header
    
    print "#FIG 2.0" > file;
    print "80 2" > file;
}

/^START$/ {
    if (state != 0)
    {
	print "ERROR on line " NR ": Not expecting START here";
	exit;
    }
    state = 1;
    nevents = marks-2;
    marks = 0;

# Make a state for start
    states[nstates] = "START";
    curstate = "START";
    curevent = "";
    nstates += 1;

    next;
}

/^END$/ {
    if (state != 2)
    {
	print "ERROR on line " NR ": Not expecting START here";
	exit;
    }

# Make a state for end
    states[nstates] = "END";
    nstates += 1;

    for (i=0;i<nstates;i++)
    {
	x = xloc[i];
	y = yloc[i];
	xs[states[i]] = x;
	ys[states[i]] = y;

# Make a lozenge
	print "4 1 0 11 0 0 0 0.000 1 15 100 " x " " y " " \
	  states[i] "" > file;
	print "1 1 0 1 0 0 0 0 1 1 0.000 " x " " y-5 " 60 15 " \
	  x " " y " " (x + 60) " " (y + 10) > file;
    }

    for (i=0;i<ntrans;i++)
    {
	xfrom = xs[gofrom[i]];
	yfrom = ys[gofrom[i]];
	xto = xs[goto[i]];
	yto = ys[goto[i]];

	if (yfrom > yto)
	{
	    yfrom -= 20;
	    yto += 10;
	    xmid = 100 + int((xfrom + xto) / 2);
	}
	else
	{
	    yfrom += 10;
	    yto -= 20;
	    xmid = -100 + int((xfrom + xto) / 2);
	}

	if (xfrom > xto)
	  ymid = -20 + int((yfrom + yto) / 2);
	else
	  ymid = 20 + int((yfrom + yto) / 2);

# See if we have tread this way before
	if (path[gofrom[i] goto[i]] == 1)
	{
	    ymid += 40 * path[gofrom[i] goto[i]];
	    path[gofrom[i] goto[i]]++;
	}

# Remember the path, to avoid stomping on ourselves later
	path[gofrom[i] goto[i]] = 1;

# Put the label halfway between start and midpoint
	xtext = int(xfrom + (xmid - xfrom) / 4);
	ytext = int(yfrom + (ymid - yfrom) / 4);

# Make an arc
	print "3 0 0 1 0 0 0 0 0.000 1 0" > file;
	print "\t0 0 1.000 4.000 8.000" > file;
	print "\t " xfrom " " yfrom " " xmid " " ymid " " xto " " yto \
	  " 9999 9999" > file;
	if (goevent[i] != "")
	  print "4 0 0 11 0 0 0 0.000 1 15 100 " xtext " " ytext " " \
	    goevent[i] "" > file;
    }

    exit;
}

/^`$/ {
    marks++;

    if (state == 1)
    {
	if (marks == 2)
	  state = 2;
    }
    else if (state == 3)
    {
	state = 2;
    }
    next;
}

state == 2 {
    print "State: " $1;

    states[nstates] = $1;
    curstate = $1;
    nstates += 1;

    if (nstates > maxstates)
    {
	print "ERROR: Can only handle " maxstates " states!\n";
	exit;
    }

    state = 3;
    next;
}

(state == 3) && ($1 == "`") {
    curevent = $2;
    next;
}

((state == 3) || (state == 1)) && /^[ \t]*\^[^ \t;]*;$/ {
    transition = substr($1, 2, length($1)-2)

    print curevent " goes to " transition;

    gofrom[ntrans] = curstate;
    goto[ntrans] = transition;
    goevent[ntrans] = curevent;
    ntrans += 1;

    next;
}
