/*
 * $RCSfile: enparg.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */

/*
 * File:	enparg.h
 * Author:	Daniel J. Van Hook
 *
 * Copyright (c) 1985 by BBN Laboratories.
 * Cambridge, MA  02238.
 * All Rights Reserved.
 *
 * This include file contains low level constants and declarations
 * relating to the enp device.
 *
 */

#ifndef ENPARG_H_INCLUDED

#define ENPARG_H_INCLUDED

#ifdef mips
#include </usr/include/sys/sysmacros.h>
#include <sys/types.h>
#include </usr/include/sys/immu.h>

#define PAGE_SIZE		NBPP	/* number of bytes per page */

#define CMC_BUFFER_SIZE                     (1024 * 1024)
#define CMC_DEVICE_SIZE                     (512 * 1024)
#define CMC_BUFFER_PAGES                    (CMC_BUFFER_SIZE/PAGE_SIZE)
#define PAGE_ALIGN(x) ((char *)((((unsigned) x/PAGE_SIZE) + 1) * PAGE_SIZE))
#endif

/*
 * Hardware supported.
 */
#define ENP30	0
#define ENP10	1
#define ENP66	2
#define ENP30S	3				/* ENP30 byte swapped */
#define ENP100	4
#define ENPSIMULATION 88             /* mips lance simulation of enp100 */
#define ENPNOT	99				/* not an enp */

#ifdef MSC
#define TIMER_REG_OFS	0x00f0
#define ENP_REG_OFS	0x00e0
#define REG_REG_OFS	0x00b0
#else
#define TIMER_REG_OFS	0x10f0
#define ENP_REG_OFS	0x10e0
#define REG_REG_OFS	0x10b0
#endif

#define VME_VECTOR_OFF		0x10f8	/* location to put vector to use */
#define ENP_REG_OFF		0x10fa	/* location to put ENP card type */

/*
 * Lowest portion of ram.
 */
struct lowstuff {
    short com_address;			/* command location */
    short control_status;		/* control status */
    long add_address;			/* starting address */
    long kernel_address;		/* kernel starting address */
    char *top_of_useable_ram;		/* above this reserved for kernel */
    short powerup_diagnostics[12];	/* power up diagnostic stuff */
};

/*
 * Enp commands.
 */
#define NORMMODE		1	/* normal operating mode */
#define EXTLMODE		2	/* external loopback mode */
#define INTLMODE		3	/* internal loopback mode */
#define GETSTAT			4	/* return statistics */
#define ZEROSTAT		5	/* initialize statistics to zero */
#define STOPENP			6	/* stop the enp */
#define AREYOUALIVE		7	/* probe to find out if there's life*/
#define FLUSHRCV		8	/* flush received packets */
#define FLUSHXMT		9	/* flush transmitted packets */
#define SETTIME			10	/* set long timer value */
#define GETTIME			11	/* get long timer value */
#define DISABLE_TIMESTAMP 	12	/* disable timestamping */
#define ENABLE_TIMESTAMP	13	/* enable timestamping */
#define STORE_DATA		18	/* store data table */
#define GET_DATA		19	/* get data table */
#define ADDMCA			20	/* add multicast address */
#define DELMCA			21	/* delete multicast address */
#define INITMCA			22	/* init multicast address list */
#define PROMMODE		23	/* promiscuous mode */
#define STOMPTIME		24	/* set timer to waiting xmt packet */
#define ADDTYPE                 25      /* add type for type code checking */ 
#define INITTYPE                26      /* init type code checking */ 
#define SETDATAPROTOCOL		27	/* set data link protocol number */
#define DISABLE_LOOPBACK	28	/* disable packet loopback */
#define ENABLE_LOOPBACK		29	/* enable packet loopback */
#define IOCONTROL		30	/* pass bytes to/from filter code */
#define VERSION_STRING		31	/* get version string */
#define SENDNOW			32	/* send the packets in the outring */

/*
 * Enp responses.
 */
#define ENPREADY 0			/* enp has completed the command */
#define ENPFAULT 99			/* enp has failed the command */


/*
 * Enp argument structures.
 */
struct startenp_arg {			/* normal and ext/int loopback */
    struct lowstuff ls;
    short command;			/* operational command */
    char enetaddr[6];			/* ethernet address */
    char *inr;				/* input ring address */
    long ins;				/* size of each ring element */
    long inn;				/* number of ring elements */
    char *outr;				/* output ring address */
    long outs;				/* size of each ring element */
    long outn;				/* number of output rings */
    char *dlrdr;			/* data link read ring */
    long dlrdn;				/* number of elements */
    char *dlwrr;			/* data link write ring */
    long dlwrn;				/* number of elements */
    long local_buffers;			/* -1 if local, 0 if in host */
    char *share_address;		/* shared memory (in host) */
    long share_size;			/* size of shared memory */
    long xmtbufs;			/* transmit buffers (returned) */
    long rcvbufs;			/* receive buffers (returned) */
    long hostbuf_ofs;			/* host buffer space offset */ 
    long hostbuf_size;			/* size of host buffer space */ 
    long mca_num;			/* number of multicast addresses */
};


struct getstat_arg {
    struct lowstuff ls;
    short command;
    short pad;
#define NUMBER_OF_STATS 24		/* number of statistics */
    long stats[NUMBER_OF_STATS];	/* statistics stuff returned here */
};

struct gettime_arg {
    struct lowstuff ls;
    short command;
    short pad;
    long time;
};

struct settime_arg {
    struct lowstuff ls;
    short command;
    short pad;
    long time;
};

struct data_arg {
    struct lowstuff ls;
    short command;
    short pad;
#define NUMBER_OF_DATA_ITEMS 32
    long data[NUMBER_OF_DATA_ITEMS];
};

struct mca_arg {
    struct lowstuff ls;
    short command;
    short pad;
    char eaddr[6];
};

struct type_arg {
    struct lowstuff ls;
    short command;
    short pad;
    int type;				/* type code to add to checking list */
};

struct version_arg {
    struct lowstuff ls;
    short command;
    short pad;
#define VERSION_SIZE 20
    char version[VERSION_SIZE];		/* version string placed here */
};

struct templet_arg {
    struct lowstuff ls;
    short command;			/* command code */
    short pad;
    char args_rets;			/* arguments and returns */
};

struct enp_registers {
#define SYSERR_BUSERROR		0
#define SYSERR_ADDRESSERROR	1
#define SYSERR_ILLEGALINST	2
#define SYSERR_PRIVILEGE	3
#define SYSERR_SPURINT		4
#define NUM_SYSERRORS		8
    char syserrors[NUM_SYSERRORS];		/* system statistics */	
    char pad1[8];
    short timer;
    short heartbeat;
    short rint;
    short tint;
    char vme_vector;			/* interrupt vector number */
    char modifier;			/* vmebus address modifier */
    char enp;				/* enp type */
    char pad3;
    short pad4;
    short pad5;
};

#endif

