/*
 * $RCSfile: enpioctl.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */

/*
 * File:	enpioctl.h
 *
 * Copyright (c) 1985, 1986, 1987, 1988 by BBN Laboratories.
 * Cambridge, MA  02238.
 * All Rights Reserved.
 *
 * This include file contains declarations and definitions pertinent to
 * ioctl access of the enp device.
 *
 */

#ifndef ENPIOCTL_H_INCLUDED

#define ENPIOCTL_H_INCLUDED

/* added SUN system ioctl.h file needed for SUN cmc driver -lms 3/7/90 */
#ifdef SUN
#include <sys/ioctl.h>
#endif /* SUN */

#ifdef mips
#ifdef sgi
#include <sys/termio.h>
#else
#include <sysv/sys/termio.h>
#endif

#if !defined(sgi5)
/* These macros are missing on RiscOS and Irix 4.0.x */
#if defined(__STDC__) || defined(__GNUC__) || defined(sgi)
#define _IORN(x,y,t)    (IOC_OUT|(((t)&IOCPARM_MASK)<<16)|(x<<8)|y)
#define _IOWN(x,y,t)    (IOC_IN|(((t)&IOCPARM_MASK)<<16)|(x<<8)|y)
#define _IOWRN(x,y,t)   (IOC_INOUT|(((t)&IOCPARM_MASK)<<16)|(x<<8)|y)
#else  /*  STDC or GNUC or sgi */
#define _IORN(x,y,t)    (IOC_OUT|(((t)&IOCPARM_MASK)<<16)|('x'<<8)|y)
#define _IOWN(x,y,t)    (IOC_IN|(((t)&IOCPARM_MASK)<<16)|('x'<<8)|y)
#define _IOWRN(x,y,t)   (IOC_INOUT|(((t)&IOCPARM_MASK)<<16)|('x'<<8)|y)
#endif  /*  STDC or GNUC or sgi */
#endif /* !sgi or !SVR4 */
#endif /* mips */

/* If this system doesn't support SIMNET, blank out the ioctls */
#if !defined(SUN) && !defined(mips)
#undef _IO
#undef _IOR
#undef _IOW
#undef _IOWR
#undef _IOWN
#undef _IORN
#define _IO(a,b)        (b)
#define _IOR(a,b,c)     (b)
#define _IOW(a,b,c)     (b)
#define _IOWR(a,b,c)    (b)
#define _IOWN(a,b,c)    (b)
#define _IORN(a,b,c)    (b)
#endif



#define N_STATS 32			/* number of statistics returned */

/* Enp command constants. */


/* ANSI C doesn't do macro replacement in strings and characters.  
 * SGI and Solaris both have ioctl definitions without quotes.
 * Mips and SunOS do have quotes, but when gcc is installed, fixincludes
 * removes them.
 */
#if defined(__STDC__) || defined(__GNUC__) || defined(sun5) || defined(sgi)


/* start the enp running */
#define ENPRUN		_IO  ('E', 1)
/* get prommed ethernet address */
#define ENPGETEADDR	_IOR ('E', 2, NetworkAddress)
/* get current statistics */
#define ENPGETSTATS	_IORN('E', 3, sizeof (int) * N_STATS)
/* zero the statistics */
#define ENPZEROSTATS	_IO  ('E', 4)
/* normal execution */
#define ENPNORM		_IOWR('E', 5, struct enprun_arg)
/* external loopback */
#define ENPEXTLOOP	_IOWR('E', 6, struct enprun_arg)
/* internal loopback */
#define ENPINTLOOP	_IOWR('E', 7, struct enprun_arg)
/* stop the enp */
#define ENPSTOP		_IO  ('E', 8)
/* reset the enp */
#define ENPRESET	_IO  ('E', 9)
/* check if the enp is running */
#define ENPALIVE	_IO  ('E', 10)
/* flush received packets */
#define ENPFLUSHRCV	_IO  ('E', 11)
/* flush transmit packets */
#define ENPFLUSHXMT	_IO  ('E', 12)

#define ENPINTSON	_IO  ('E', 13)

#define ENPINTSOFF	_IO  ('E', 14)
/* set long timer value */
#define ENPSETTIME	_IOW ('E', 15, int)
/* get the long timer value */
#define ENPGETTIME	_IOR ('E', 16, int)
/* enable timestamping */
#define ENPENBTIMESTAMP	_IO  ('E', 17)
/* disable timestamping */
#define ENPDISTIMESTAMP	_IO  ('E', 18)
#ifdef MASSCOMP
/* get parameter array */
#define ENPGETPARAMS    _IORN('E', 20, sizeof (int) * ENPNUMPARAMS)
/* set parameter array */
#define ENPSETPARAMS    _IOWN('E', 21, sizeof (int) * ENPNUMPARAMS)
#else
#define ENPGETPARAMS   _IOW ('E', 20, int)
#define ENPSETPARAMS   _IOW ('E', 21, int)
#endif
/* support for blocking */
#define ENPREADSUPPORT	_IO  ('E', 22)
/* support for blocking */
#define ENPWRITESUPPORT	_IO  ('E', 23)
/* get the address of the enp */
#define ENPADDRESS	_IORN('E', 25, sizeof (char *))
/* as ENPTABLECONTROL, but pass pointer to value */
#define ENPTBLCNTRL	_IOW ('E', 26, int)
/* add multicast address */
#define ENPADDMCA	_IOWN ('E', 28, ENETADDRSZ)
/* delete multicast address */
#define ENPDELMCA	_IOWN ('E', 29, ENETADDRSZ)
/* add multicast address */
#define ENPINITMCA	_IO  ('E', 30)
/* promiscuous execution */
#define ENPPROM		_IOWR('E', 31, struct enprun_arg)
/* get number of receive and transmit buffers */
#define ENPBUFS		_IORN('E', 32, 2 * sizeof(long))
#ifndef mips
#ifndef SUN
#ifndef HP
/* get parameter array */
#define ENPGETPARAMS2	_IORN('E', 33, sizeof (int) * ENPNUMPARAMS2)
/* set parameter array */
#define ENPSETPARAMS2	_IOWN('E', 34, sizeof (int) * ENPNUMPARAMS2)
#endif
#endif
#endif
/* set timer to timestamp of next pkt */
#define ENPSTOMPTIME	_IO ('E', 35)
#define ENPGETWRITE	_IOR ('E', 36, int)
#define ENPGETREAD	_IOR ('E', 37, int)
/* get host buffer offset (on the card; from card start) and size */
#define ENPGETHOSTBUF	_IORN('E', 38, 2 * sizeof(long))
/* get parameter array */
#define ENPACCESS	_IOWR('E', 41, struct enpaccess_arg)
/* get type of interface */
#define ENPTYPE		_IOR ('E', 42, int)
/* get IFX ccb pointer */
#define ENPGETCCB       43  
#define ENPGETIOCTL     44
#define ENPBLOCKING     45
#define ENPPOLLED       46
/* send all the packets in the outring (MIPS MAGNUM ONLY) */
#define ENPSENDNOW	_IO('E', 47)
 
/* add UDP port */
#define ENPADDUDP	_IOWN ('E', 48, sizeof(short))
/* init UDP port */
#define ENPINITUDP	_IO  ('E', 49)
 
#define DL_TIMEOUT	1		/* set timeout for reads and writes */
#define DL_BLOCKING	2		/* enable/disable blocking */
#define NOBLOCK_ON_RD	0x1		/* bit mapped arguments */
#define NOBLOCK_ON_WR	0x2	


#ifdef MASSCOMP
#define ENPNUMPARAMS	12		/* number of params driver can hold */
#define ENPNUMPARAMS2	36		/* number of params driver can hold */
#else
#define ENPNUMPARAMS	50		/* number of params driver can hold */
#endif





#else     /* __STDC__, __GNUC__, sun5, or sgi */





/* start the enp running */
#define ENPRUN		_IO  (E, 1)
/* get prommed ethernet address */
#define ENPGETEADDR	_IOR (E, 2, NetworkAddress)
/* get current statistics */
#define ENPGETSTATS	_IORN(E, 3, sizeof (int) * N_STATS)
/* zero the statistics */
#define ENPZEROSTATS	_IO  (E, 4)
/* normal execution */
#define ENPNORM		_IOWR(E, 5, struct enprun_arg)
/* external loopback */
#define ENPEXTLOOP	_IOWR(E, 6, struct enprun_arg)
/* internal loopback */
#define ENPINTLOOP	_IOWR(E, 7, struct enprun_arg)
/* stop the enp */
#define ENPSTOP		_IO  (E, 8)
/* reset the enp */
#define ENPRESET	_IO  (E, 9)
/* check if the enp is running */
#define ENPALIVE	_IO  (E, 10)
/* flush received packets */
#define ENPFLUSHRCV	_IO  (E, 11)
/* flush transmit packets */
#define ENPFLUSHXMT	_IO  (E, 12)

#define ENPINTSON	_IO  (E, 13)

#define ENPINTSOFF	_IO  (E, 14)
/* set long timer value */
#define ENPSETTIME	_IOW (E, 15, int)
/* get the long timer value */
#define ENPGETTIME	_IOR (E, 16, int)
/* enable timestamping */
#define ENPENBTIMESTAMP	_IO  (E, 17)
/* disable timestamping */
#define ENPDISTIMESTAMP	_IO  (E, 18)
#ifdef MASSCOMP
/* get parameter array */
#define ENPGETPARAMS    _IORN(E, 20, sizeof (int) * ENPNUMPARAMS)
/* set parameter array */
#define ENPSETPARAMS    _IOWN(E, 21, sizeof (int) * ENPNUMPARAMS)
#else
#define ENPGETPARAMS   _IOW (E, 20, int)
#define ENPSETPARAMS   _IOW (E, 21, int)
#endif
/* support for blocking */
#define ENPREADSUPPORT	_IO  (E, 22)
/* support for blocking */
#define ENPWRITESUPPORT	_IO  (E, 23)
/* get the address of the enp */
#define ENPADDRESS	_IORN(E, 25, sizeof (char *))
/* as ENPTABLECONTROL, but pass pointer to value */
#define ENPTBLCNTRL	_IOW (E, 26, int)
/* add multicast address */
#define ENPADDMCA	_IOWN (E, 28, ENETADDRSZ)
/* delete multicast address */
#define ENPDELMCA	_IOWN (E, 29, ENETADDRSZ)
/* add multicast address */
#define ENPINITMCA	_IO  (E, 30)
/* promiscuous execution */
#define ENPPROM		_IOWR(E, 31, struct enprun_arg)
/* get number of receive and transmit buffers */
#define ENPBUFS		_IORN(E, 32, 2 * sizeof(long))
#ifndef mips
#ifndef SUN
#ifndef HP
/* get parameter array */
#define ENPGETPARAMS2	_IORN(E, 33, sizeof (int) * ENPNUMPARAMS2)
/* set parameter array */
#define ENPSETPARAMS2	_IOWN(E, 34, sizeof (int) * ENPNUMPARAMS2)
#endif
#endif
#endif
/* set timer to timestamp of next pkt */
#define ENPSTOMPTIME	_IO (E, 35)
#define ENPGETWRITE	_IOR (E, 36, int)
#define ENPGETREAD	_IOR (E, 37, int)
/* get host buffer offset (on the card; from card start) and size */
#define ENPGETHOSTBUF	_IORN(E, 38, 2 * sizeof(long))
/* get parameter array */
#define ENPACCESS	_IOWR(E, 41, struct enpaccess_arg)
/* get type of interface */
#define ENPTYPE		_IOR (E, 42, int)
/* get IFX ccb pointer */
#define ENPGETCCB       43  
#define ENPGETIOCTL     44
#define ENPBLOCKING     45
#define ENPPOLLED       46
/* send all the packets in the outring (MIPS MAGNUM ONLY) */
#define ENPSENDNOW	_IO(E, 47)
 
/* add UDP port */
#define ENPADDUDP	_IOWN (E, 48, sizeof(short))
/* init UDP port */
#define ENPINITUDP	_IO  (E, 49)
 
#define DL_TIMEOUT	1		/* set timeout for reads and writes */
#define DL_BLOCKING	2		/* enable/disable blocking */
#define NOBLOCK_ON_RD	0x1		/* bit mapped arguments */
#define NOBLOCK_ON_WR	0x2	


#ifdef MASSCOMP
#define ENPNUMPARAMS	12		/* number of params driver can hold */
#define ENPNUMPARAMS2	36		/* number of params driver can hold */
#else
#define ENPNUMPARAMS	50		/* number of params driver can hold */
#endif


#endif     /* __STDC__, __GNUC__, sun5, or sgi */






/* Argument structure for normal, external loopback, internal loopback
   and promiscuous running modes.  */

struct enprun_arg {
    char enetaddr[6];			/* ethernet address */
    char *inr;				/* pointer to input ring */
    int  ins;				/* size of each intput ring entry */
    int  inn;				/* number of input ring entries */
    char *outr;				/* pointer to output ring */
    int  outs;				/* size of each output ring entry */
    int  outn;				/* number of output ring entries */
    char *share_address;		/* shared memory address */
    int  share_size;			/* shared memory size */
    int  local_buffers;
    int  hostbufsize;			/* size host buffer */
    int  mcanum;			/* number of multicast entries */
};

struct enprwsupport_arg {
    short *pfull;
    int tout_val;
};

/* Generic access argument structure.  Bytes_to_enp bytes are copied
   to the enp from buffer_to_enp as arguments to the enp for the
   command.  After completion, bytes_to_host bytes are copied to
   buffer_to_host from the enp. */

struct enpaccess_arg {
    int command;
    char *buffer_to_enp;
    int bytes_to_enp;
    char *buffer_to_host;
    int bytes_to_host;
};

#endif
