/*----------------------------------------------------------------------*
 | File: dis204_types.h                                                 |
 |                                                                      |
 | Description: Contains common types for the DIS protocol as defined   |
 |              in the "Military Standard (2.0.4) Protocol Data Units   |
 |		for Entity Information and Entity Interaction in a      |
 |		Distributed Interactive Simulation" document.           |
 |                                                                      |
 |              This code was modified from existing code written by:   |
 |                                                                      |
 |          Loral Advanced Distributed Simulation                       |
 |          50 Moulton Street                                           |
 |          Cambridge, Ma. 02138                                        |
 |          617-873-3600                                                |
 |                                                                      |
 | Author: Julienne P. Chappell                                         |
 |                                                                      |
 | Company: E-OIR Measurements, Inc.                                    |
 |          9241 Courthouse Road                                        |
 |          P.O. Box 1240                                               |
 |          Spotsylvania, VA 22553                                      |
 |                                                                      |
 |                                                                      |
 | Copyright:                                                           |
 |   + This software is Copyright (C) E-OIR Measurements, Inc., 1998    |
 |   + E-OIR Measurements, Inc. has all commercial interest and rights. |
 |   + E-OIR Measurements, Inc. grants a royalty-free license and       |
 |     right to use, free of charge, with the following terms and       | 
 |     conditions:                                                      |
 |                                                                      |
 |       (1) You can redistribute this software in source form.  If you |
 |           redistribute this software in compiled form you will       |
 |           include the source code.                                   |
 |       (2) You own full rights to any output files you generate with  |
 |           this software.                                             |
 |       (3) You can make modifications to this software and use it for |
 |           in-house use only.  Under no circumstances can modified    |
 |           software be redistributed.                                 |
 |       (4) E-OIR Measurements, Inc. shall be credited should this     |
 |           software be used in a product or written about in any      |
 |           publication.                                               |
 |       (5) This software is provided "as is", without warranty by     |
 |           E-OIR Measurements, Inc. In no event shall E-OIR           |
 |           Measurements, Inc. be liable for any loss or for any       |
 |           indirect, special, punitive, exemplary, incidental, or     |
 |           consequential damages arising from the use, possession or  |
 |           performance of this software.                              |
 |                                                                      |
 *----------------------------------------------------------------------*/

/*--------------------------- RCS Information --------------------------*
 * $RCSfile: dis204_types.h,v $
 * $Revision: 1.1.1.1 $
 * $Author: ksnively $
 * $Date: 1999/04/13 17:02:15 $
 *----------------------------- Change Log -----------------------------*
 * $Log: dis204_types.h,v $
 * Revision 1.1.1.1  1999/04/13 17:02:15  ksnively
 * Import include directory
 *
 *
//
 *----------------------------------------------------------------------*/


#ifndef _dis204_types_included
#define _dis204_types_included



#include <stdtypes.h>
#include <dis203_types.h>
#include <dis203_values.h> 
#include <dis204_values.h> 

/* 204
 * Angle Representation
 *
 * Version 2.0.4 Angle Representation (section 5.3.1, page 78)
 * Angles shall be specified as 32-bit float32s expressed in radians
 */

/*typedef float32            DIS204_ANGLE_TYPE;   */ /* same as 203 */


/* 204
 * The Angular Velocity Representation
 *
 */

/*typedef float32            DIS204_ANGULAR_VELOCITY_TYPE;    */ /* same as 203 */

/* 204
 * Change Indicator
 *
 */

typedef uint8   DIS204_CHANGE_INDICATOR_TYPE; /* this combined with next was uint16 */

/* 204
 * Parameter Type Designator
 * 5.3.5.1
 */

typedef uint8   DIS204_PARAMETER_TYPE_DESIGNATOR_TYPE;

/* 204
 * ID - Part Attached to
 *
 */

/*typedef uint16   DIS204_ID_PART_ATTACHED_TO_TYPE;    */ /* same as 203 */


/* 204
 * Parameter Type
 *
 */

/*typedef uint32    DIS204_PARAMETER_TYPE_TYPE;    */ /* same as 203 */


/* 204
 * Parameter Value
 *
 */

typedef float64            DIS204_PARAMETER_VALUE_TYPE; 

 /* 204 Boolean */

 /* Shall be represented as a single bit representing a true-false */
 /* value.  This bit shall represent an enumeration type of one bit, */
 /* where the value 0 is interpreted as false, and the value 1 as true. */

/*typedef struct {
    unsigned       value  : 1;           
    unsigned       _unused_1 : 7;       */ /* padding */
/*} DIS204_BOOLEAN_RECORD;                     */        /* same as 203 */
 /*  204 */
 /* Quantity */

/*typedef uint16 DIS204_QUANTITY_TYPE;  */   /* same as 203 */


/* 204
 * Rate
 *
 * In units of rounds per second
 *
 */

/*typedef uint16 DIS204_RATE_TYPE;  */   /* same as 203 */


/* 204
 * Entity Appearance
 *
 */

/*typedef uint32 DIS204_ENTITY_APPEARANCE_TYPE;  */   /* same as 203 */


/* 204
 * Entity Capabilities
 *
 */

/*typedef struct {
    unsigned       _unused_2                     : 28; */ /* padding */
/*    unsigned       repair_capability             : 1; 
    unsigned       recovery_capability           : 1;  
    unsigned       fuel_supply_capability        : 1;  
    unsigned       ammunition_supply_capability  : 1;  
} DIS204_ENTITY_CAPABILITIES_RECORD;   */  /* same as 203 */


/* 204
 * Site ID
 * No site shall be assigned an ID containing all zeroes, (2^16 - 1),
 * or (2^16 - 2).
 */

/*typedef uint16 DIS204_SITE_ID_TYPE;   */  /* same as 203 */

/*  204 
 * Application ID 
 * No simulation application shall be assigned an ID containing all zeroes, 
 * (2^16 - 1), or (2^16 - 2).
 */

/*typedef uint16 DIS204_APPLICATION_ID_TYPE;   */  /* same as 203 */


/* 204
 * Entity ID
 * No Entity ID shall be assigned an ID containing all zeroes, 
 * (2^16 - 1), or (2^16 - 2).
 */

/*typedef uint16 DIS204_ENTITY_ID_TYPE;    */ /* same as 203 */


/* 204
 * Group ID
 *
 */

/*typedef uint16 DIS204_GROUP_ID_TYPE;    */ /* same as 203 */

 /* 204  5.3.15 */
 /* Character Set */

/*typedef uint8 DIS204_CHARACTER_SET_TYPE;  */   /* same as 203 */
 /*  204 */
 /* Character */

/*typedef uint8 DIS204_CHARACTER_TYPE;    */ /* same as 203 */


/* 204
 * Entity Kind
 * 5.3.16 
 */

/*typedef uint8 DIS204_KIND_TYPE;   */  /* same as 203 */


/* 204
 * Domain
 *
 */

/*typedef uint8 DIS204_DOMAIN_TYPE;   */  /* same as 203 */


/* 204
 * Country
 *
 */

/*typedef uint16 DIS204_COUNTRY_TYPE;    */ /* same as 203 */


/* 204
 * Category
 *
 */

/*typedef uint8 DIS204_CATEGORY_TYPE;   */  /* same as 203 */


/* 204
 * Subcategory
 *
 */

/*typedef uint8 DIS204_SUBCATEGORY_TYPE;    */ /* same as 203 */


/* 204
 * Specific
 *
 */

/*typedef uint8 DIS204_SPECIFIC_TYPE;   */  /* same as 203 */


/* 204
 * Extra
 * 5.3.16.7
 */

/*typedef uint8 DIS204_EXTRA_TYPE;    */ /* same as 203 */


/* 204
 * Event ID
 * 5.3.18
 */

/*typedef uint16 DIS204_EVENT_ID_TYPE;    */ /* same as 203 */


/* 204
 * Exercise ID
 * 5.3.19
 */

/*typedef uint8 DIS204_EXERCISE_ID_TYPE;    */ /* same as 203 */


/* 204
 * Force ID
 * 5.3.21
 */

/*typedef uint8 DIS204_FORCE_ID_TYPE;   */  /* same as 203 */


/* 204
 * Protocol Version
 * 5.3.24
 */

/*typedef uint8 DIS204_PROTOCOL_VERSION_TYPE;   */  /* same as 203 */


/* 204
 * Protocol Family
 *
 */

typedef uint8 DIS204_PROTOCOL_FAMILY_TYPE;
 
/* 204
 * PDU Type
 *
 */

/*typedef uint8 DIS204_PDU_TYPE_TYPE;    */ /* same as 203 */


/* 204
 * Time Stamp
 * 5.3.24 (look at 5.3.30)
 */

/*typedef struct {
    unsigned       value      : 31;      
    unsigned       reference  : 1;       
} DIS204_TIMESTAMP_RECORD;    */ /* same as 203 */


/* 204
 * Time
 *
 */

/*typedef uint32 DIS204_TIME_TYPE;  */   /* same as 203 */


/* 204
 * Length
 *
 */

/*typedef uint16 DIS204_LENGTH_TYPE;    */ /* same as 203 */


/* 204
 * Repair
 * 5.3.25
 */

/*typedef uint16 DIS204_REPAIR_TYPE_TYPE;   */  /* same as 203 */


/* 204
 * Service
 * 5.3.27
 */

/*typedef uint8 DIS204_SERVICE_TYPE_TYPE;  */   /* same as 203 */


/* 204
 * Supply Quantity
 * 5.3.29
 */

/*typedef float32 DIS204_SUPPLY_QUANTITY_TYPE;  */   /* same as 203 */


/* 204
 * Vector Component
 * 5.3.32
 */

/*typedef float32 DIS204_VECTOR_COMPONENT_TYPE;   */  /* same as 203 */

 /* 204  */
 /* World Coordinates Component */
 /* 5.3.33 */

/*typedef float64 DIS204_WORLD_COORDINATES_COMPONENT_TYPE;    */ /* same as 203 */


/* 204
 * Dead Reckoning Algorithm
 * 5.4.3
 */

/*typedef uint8 DIS204_DEAD_RECKONING_ALGORITHM_TYPE;   */  /* same as 203 */


/* 204
 * Number of Articulation Parameters
 *
 */

/*typedef uint8 DIS204_NUMBER_OF_ARTICULATION_PARAMETERS_TYPE;    */ /* same as 203 */


/* 204
 * Range
 * 5.4.4
 */

/*typedef float32 DIS204_RANGE_TYPE;   */  /* same as 203 */


/* 204
 * Detonation Result
 * 5.4.4.2
 */

/*typedef uint8 DIS204_DETONATION_RESULT_TYPE;    */ /* same as 203 */


/* 204
 * Number of Supply Types
 * 5.4.5.1
 */

/*typedef uint8 DIS204_NUMBER_OF_SUPPLY_TYPES_TYPE; */    /* same as 203 */


/* 204
 * Repair Result
 * 5.4.5.6
 */

/*typedef uint8 DIS204_REPAIR_RESULT_TYPE;    */ /* same as 203 */


/* 204
 * Mass
 * 5.4.3.2
 */

/*typedef float32 DIS204_MASS_TYPE;    */ /* same as 203 */


/* 204
 * Reasons to Stop (4.3.1.10)
 * 5.4.6.4
 */

/*typedef uint8 DIS204_REASONS_TO_STOP_TYPE;   */  /* same as 203 */


/* 204
 * Acknowledgement Flags 
 * 5.4.6.5
 */

/*typedef uint16 DIS204_ACKNOWLEDGEMENT_FLAGS_TYPE;  */   /* same as 203 */


/* 204
 * Response Flags (None defined)
 *
 */

/*typedef uint16 DIS204_RESPONSE_FLAGS_TYPE;   */  /* same as 203 */

 /* 204  */
 /* Request ID  */


/*typedef uint32 DIS204_REQUEST_ID_TYPE;  */   /* same as 203 */


/* 204
 * Action ID
 * 5.4.6.6
 */

/*typedef uint32 DIS204_ACTION_ID_TYPE;  */   /* same as 203 */


/* 204
 * Request Status
 * 5.4.6.7
 */

/*typedef uint32 DIS204_REQUEST_STATUS_TYPE;  */   /* same as 203 */


/* 204
 * Time Interval
 * 5.4.6.8
 */

/*typedef uint32 DIS204_TIME_INTERVAL_TYPE;    */ /* same as 203 */

 /*  204  */
 /* Number of Fixed Datum Fields   */

/*typedef uint32 DIS204_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE;  */   /* same as 203 */
 /*  204  */
 /* Number of Variable Datum Fields   */

/*typedef uint32 DIS204_NUMBER_OF_VARIABLE_DATUM_FIELDS_TYPE;   */  /* same as 203 */


/* 204
 * Datum ID
 *
 */

/*typedef uint32 DIS204_DATUM_ID_TYPE;  */   /* same as 203 */


/* 204
 * Event Type
 * 5.4.6.11
 */

/*typedef uint32 DIS204_EVENT_TYPE_TYPE;    */ /* same as 203 */


/* 204
 * Radio ID
 * 5.4.8.1
 */

/*typedef uint16 DIS204_RADIO_ID_TYPE;  */   /* same as 203 */


/* 204
 * Sample rate (for Signal PDU)
 * 5.4.8.2
 */

/*typedef uint32 DIS204_SAMPLE_RATE_TYPE;   */  /* same as 203 */


/* 204
 * Signal length (for Signal PDU)
 *
 */

/*typedef uint16 DIS204_SIGNAL_LENGTH_TYPE;  */   /* same as 203 */


/* 204
 * Number of Samples (for Signal PDU)
 *
 */

/*typedef uint16 DIS204_SAMPLES_TYPE;  */   /* same as 203 */


/* 204
 * Radio Transmit State
 * 5.4.8.1
 */

/*typedef uint8 DIS204_RADIO_TRANSMIT_STATE_TYPE;  */    /* same as 203 */


/* 204
 * Radio Input Source
 *
 */

/*typedef uint8 DIS204_RADIO_INPUT_SOURCE_TYPE;    */  /* same as 203 */


/* 204
 * Antenna Pattern Type
 *
 */

/*typedef uint16 DIS204_ANTENNA_PATTERN_TYPE_TYPE;    */  /* same as 203 */


/* 204
 * Antenna Pattern Length
 *
 */

/*typedef uint16 DIS204_ANTENNA_PATTERN_LENGTH_TYPE;     */ /* same as 203 */


/* 204
 * Frequency
 *
 */

/*typedef struct { 
    uint32 high_order;            
    uint32 low_order;             
} DIS204_TRANSMITTER_FREQUENCY_TYPE;   */   /* same as 203 */


/* 204
 * Frequency Bandwidth
 *
 */

/*typedef float32 DIS204_FREQUENCY_BANDWIDTH_TYPE;    */  /* same as 203 */


/* 204
 * Power
 *
 */

/*typedef float32 DIS204_POWER_TYPE;  */    /* same as 203 */


/* 204
 * Crypto System
 *
 */

/*typedef uint16 DIS204_CRYPTO_SYSTEM_TYPE;  */    /* same as 203 */


/* 204
 * Crypto Key ID
 *
 */

/*typedef uint16 DIS204_CRYPTO_KEY_ID_TYPE;   */   /* same as 203 */


/* 204
 * Length of Modulation Parameters (for Transmitter PDU)
 *
 */

/*typedef uint8 DIS204_MODULATION_LENGTH_TYPE; */     /* same as 203 */


/* 204
 * Receiver State (for Receiver PDU)
 * 5.4.8.3
 */

/*typedef uint16 DIS204_RECEIVER_STATE_TYPE;    */  /* same as 203 */


/* 204
 * Transmitter Radio ID (for Receiver PDU)
 *
 */

/*typedef uint16 TRANSMITTER_RADIO_ID_TYPE;   */   /* same as 203 */


/* 204
 * State Update Indicator (for  Electromagnetic Emission PDU)
 * 5.4.7.1
 */

/*typedef uint8 DIS204_STATE_UPDATE_INDICATOR_TYPE;    */  /* same as 203 */


/* 204
 * Number of Systems (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_NUMBER_OF_SYSTEMS_TYPE;   */   /* same as 203 */


/* 204
 * System Data Length (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_SYSTEM_DATA_LENGTH_TYPE;   */   /* same as 203 */


/* 204
 * Number of Beams (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_NUMBER_OF_BEAMS_TYPE;    */  /* same as 203 */


/* 204
 * Beam Data Length (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_BEAM_DATA_LENGTH_TYPE;    */  /* same as 203 */


/* 204
 * Beam ID Number (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_BEAM_ID_NUMBER_TYPE;     */ /* same as 203 */


/* 204
 * Beam Parameter Index (for Electromagnetic  Emission PDU)
 *
 */

/*typedef uint16 DIS204_BEAM_PARAMETER_INDEX_TYPE;  */    /* same as 203 */


/* 204
 * Beam Function (for Electromagnetic  Emission PDU)
 *
 */

/*typedef uint8 DIS204_BEAM_FUNCTION_TYPE;   */   /* same as 203 */


/* 204
 * Number of Targets (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_NUMBER_OF_TARGETS_TYPE;  */    /* same as 203 */


/* 204
 * High Density Track Jam (for Electromagnetic  Emission PDU)
 *
 */

/*typedef uint8 DIS204_HIGH_DENSITY_TRACK_JAM_TYPE;   */   /* same as 203 */


/* 204
 * Jamming Mode Sequence (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint32 DIS204_JAMMING_MODE_SEQUENCE_TYPE;     */  /* same as 203 */


/* 204
 * Emitter Name (for Electromagnetic  Emission PDU)
 *
 */

/*typedef uint16 DIS204_EMITTER_NAME_TYPE;   */    /* same as 203 */


/* 204
 * Emitter Function (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_EMITTER_FUNCTION_TYPE;    */   /* same as 203 */


/* 204
 * Emitter ID Number (for  Electromagnetic Emission PDU)
 *
 */

/*typedef uint8 DIS204_EMITTER_ID_NUMBER_TYPE;   */    /* same as 203 */


/* 204
 * Emission Frequency (for  Electromagnetic Emission PDU)
 *
 */

/*typedef float32 DIS204_EMISSION_FREQUENCY_TYPE;    */   /* same as 203 */


/* 204
 * Frequency Range (for  Electromagnetic Emission PDU)
 *
 */

/*typedef float32 DIS204_FREQUENCY_RANGE_TYPE;    */   /* same as 203 */


/* 204
 * Effective Radiated Power (for  Electromagnetic Emission PDU)
 *
 */

/*typedef float32 DIS204_EFFECTIVE_RADIATED_POWER_TYPE;   */    /* same as 203 */


/* 204
 * Pulse Width (for  Electromagnetic Emission PDU)
 *
 */

/*typedef float32 DIS204_PULSE_WIDTH_TYPE;   */    /* same as 203 */


/* 204
 * Beam Sweep Sync (for  Electromagnetic Emission PDU)
 *
 */

/*typedef float32 DIS204_BEAM_SWEEP_SYNC_TYPE;   */    /* same as 203 */


/* 204
 * Code Name (for Designator PDU)
 * 5.4.7.1
 */

/*typedef uint16 DIS204_CODE_NAME_TYPE;   */    /* same as 203 */


/* 204
 *  Designator Code (for Designator  PDU)
 *
 */

/*typedef uint8 DIS204_LASER_CODE_TYPE;     */  /* same as 203 */


/* 204
 *  Designator Wave Length (for Designator  PDU)
 *
 */

/*typedef float32 DIS204_LASER_WAVE_LENGTH_TYPE;   */    /* same as 203 */





 /* 204 END LOWEST-LEVEL COMPONENTS */
 /* 1st-LEVEL COMPONENTS */

 /* Angular Velocity Vector Record (section 5.3.2, Page 78) */

 /* These rates shall be measured in radians per second about each of the */
 /* entity's own coordinate axes.  The first field shall consist of the */
 /* velocity about the x-axis, the second about the y-axis and the third */
 /* about the z-axis. */

/*typedef struct {
    DIS204_ANGULAR_VELOCITY_TYPE rate_about_x_axis;  
    DIS204_ANGULAR_VELOCITY_TYPE rate_about_y_axis;  
    DIS204_ANGULAR_VELOCITY_TYPE rate_about_z_axis;  
} DIS204_ANGULAR_VELOCITY_VECTOR_RECORD;  */     /* same as 203 */


/* 204
 * Simulation Address Record
 * 5.3.14.1
 */

/*typedef struct {
    DIS204_SITE_ID_TYPE        site_id;  
    DIS204_APPLICATION_ID_TYPE application_id;  
} DIS204_SIMULATION_ADDRESS_RECORD;   */    /* same as 203 */


/* 204
 * Entity Marking Record
 * 5.3.15
 */

/*typedef struct {
    DIS204_CHARACTER_SET_TYPE Character_set;  
    DIS204_CHARACTER_TYPE     Character_string[11];  
} DIS204_ENTITY_MARKING_RECORD;  */     /* same as 203 */


/* 204
 * Entity Type Record
 * 5.3.16
 */

/*typedef struct {
    DIS204_KIND_TYPE        kind;    
    DIS204_DOMAIN_TYPE      domain;   
    DIS204_COUNTRY_TYPE     country;  
    DIS204_CATEGORY_TYPE    category;  
    DIS204_SUBCATEGORY_TYPE subcategory;  
    DIS204_SPECIFIC_TYPE    specific;  
    DIS204_EXTRA_TYPE       extra;    
} DIS204_ENTITY_TYPE_RECORD;   */    /* same as 203 */  


/* 204
 * Euler Angles Record
 * 5.3.17
 */

/*typedef struct {
    DIS204_ANGLE_TYPE psi;            
    DIS204_ANGLE_TYPE theta;          
    DIS204_ANGLE_TYPE phi;            
} DIS204_EULER_ANGLES_RECORD;    */   /* same as 203 */


 /*  204 */
 /* Protocol Data Unit (PDU) Header Record */
 /* 5.3.24 */

typedef struct {
    DIS203_PROTOCOL_VERSION_TYPE protocol_version; 
    DIS203_EXERCISE_ID_TYPE      exercise_id; 
    DIS203_PDU_TYPE_TYPE         pdu_type; 
    DIS204_PROTOCOL_FAMILY_TYPE  protocol_family;   /*  different from 203 */
    DIS203_TIMESTAMP_RECORD      timestamp; 
    DIS203_LENGTH_TYPE           length; 
    uint16                  _unused_4;  /* padding */
} DIS204_PDU_HEADER_RECORD; 


/* 204
 * Vector Record
 * 5.3.32
 */

/*typedef struct { 
    DIS204_VECTOR_COMPONENT_TYPE x;   
    DIS204_VECTOR_COMPONENT_TYPE y;   
    DIS204_VECTOR_COMPONENT_TYPE z;   
} DIS204_VECTOR_RECORD;      */ /* same as 203 */


/* 204
 * Entity Coordinate Vector Record
 * 5.3.32.1
 */

/*typedef DIS204_VECTOR_RECORD DIS204_ENTITY_COORDINATE_VECTOR_RECORD;   */ /* same as 203 */


/* 204
 * Linear Acceleration Vector Record
 * 5.3.32.2
 */

/*typedef DIS204_VECTOR_RECORD DIS204_LINEAR_ACCELERATION_VECTOR_RECORD; */ /* same as 203 */


/* 204
 * Linear Velocity Vector Record
 * 5.3.32.3
 */

/*typedef DIS204_VECTOR_RECORD DIS204_LINEAR_VELOCITY_VECTOR_RECORD; */    /* same as 203 */


/* 204
 * World Coordinates Record
 * 5.3.33
 */

/*typedef struct {
    DIS204_WORLD_COORDINATES_COMPONENT_TYPE x;  
    DIS204_WORLD_COORDINATES_COMPONENT_TYPE y;  
    DIS204_WORLD_COORDINATES_COMPONENT_TYPE z;  
} DIS204_WORLD_COORDINATES_RECORD;   */    /* same as 203 */


/* 204
 * Clock Time 
 * 5.3.8
 */

/*typedef struct {
    uint32              hour;     
    DIS204_TIMESTAMP_RECORD time_past_the_hour;  
} DIS204_CLOCK_TIME_RECORD; */      /* same as 203 */


/* 204
 * Fixed Datum Record
 * 5.3.10
 */

/*typedef struct {
    DIS204_DATUM_ID_TYPE datum_id;    
    uint32           datum_value;  
} DIS204_FIXED_DATUM_RECORD;     */  /* same as 203 */


/* 204
 * Datum Value Record
 *
 */

/*typedef struct {
    uint32 datum_value_1;         
    uint32 datum_value_2;         
} DIS204_DATUM_VALUE_RECORD;  */     /* same as 203 */


/* 204
 * Variable Datum Record
 * 5.3.10
 */

/*typedef struct {
    DIS204_DATUM_ID_TYPE      datum_id;  
    uint32                datum_length;  
    DIS204_DATUM_VALUE_RECORD datum_value[1];  
} DIS204_VARIABLE_DATUM_RECORD;      */  /* same as 203 */


/* 204
 * Modulation Type 
 * 5.4.8.1 [13]
 */

/*typedef struct {
    uint16 spread_spectrum;      
    uint16 mod_major;            
    uint16 mod_detail;           
    uint16 mod_system;           
} DIS204_MODULATION_TYPE_RECORD;      */  /* same as 203 */


/* 204
 * Encoding Scheme 
 * 5.4.8.2
 */

/*typedef uint16 DIS204_ENCODING_SCHEME;     */   /* same as 203 */


/* 204
 * Emitter System (for Electromagnetic Emission PDU)
 * 5.4.7
 */

/*typedef struct {
    DIS204_EMITTER_NAME_TYPE      emitter_name;  
    DIS204_EMITTER_FUNCTION_TYPE  function;  
    DIS204_EMITTER_ID_NUMBER_TYPE emitter_id_number;  
} DIS204_EMITTER_SYSTEM_RECORD;       */   /* same as 203 */


/* 204
 * Fundamental Parameter Data Record (for Electromagnetic Emission PDU)
 * 5.4.7
 */

/*typedef struct {
    DIS204_EMISSION_FREQUENCY_TYPE       frequency;  
    DIS204_FREQUENCY_RANGE_TYPE          frequency_range;  
    DIS204_EFFECTIVE_RADIATED_POWER_TYPE effective_radiated_power;  
    DIS204_EMISSION_FREQUENCY_TYPE       pulse_repetition_frequency;  
    DIS204_PULSE_WIDTH_TYPE              pulse_width;  
    DIS204_ANGLE_TYPE                    beam_azimuth_center;  
    DIS204_ANGLE_TYPE                    beam_azimuth_sweep;  
    DIS204_ANGLE_TYPE                    beam_elevation_center;  
    DIS204_ANGLE_TYPE                    beam_elevation_sweep;  
    DIS204_BEAM_SWEEP_SYNC_TYPE          beam_sweep_sync;  
} DIS204_FUNDAMENTAL_PARAMETER_DATA_RECORD;       */    /* same as 203 */

 /* 204 END 1st-LEVEL COMPONENTS */





 /*  204 2nd-LEVEL COMPONENTS */

 /* Articulation Parameter Record 
  * 5.3.5.5
  */

 /* The change indicator shall be set to zero for each exercise and */
 /* sequentially incremented by one for each change in articulation */
 /* parameters.  In the case where all possible values are exhausted, */
 /* the numbers shall be reused beginning at zero. */

 /* The id - part attached to field shall contain the value zero if */
 /* the articulated part is attached directly to the entity. */

typedef struct {
    DIS204_PARAMETER_TYPE_DESIGNATOR_TYPE   parameter_type_designator; /* diff from 203 */
    DIS204_CHANGE_INDICATOR_TYPE    change_indicator;  /* diff from 203 */
    DIS203_ID_PART_ATTACHED_TO_TYPE id_part_attached_to; 

    union {
         /* Parameter Type Record */
         /* if whole, want to access all 32 bits at once for comparisons */
        DIS203_PARAMETER_TYPE_TYPE          value;
         /* if articulated part */
        struct {
          /*  unsigned       part_type         : 1; */ /* this no longer exists in 2.04 */
            unsigned       part_type_class   : 27; 
            unsigned       part_type_metric  : 5; 
        } articulation_parameter;
         /* if attached part */
        struct {
          /*  unsigned       part_type          : 1; */ /* this no longer exists in 2.04 */ 
            unsigned       part_type_station  : 32; 
        } attached_part;
    } parameter_type;          /*  way different from 203 */           

    union {
         /* Parameter Value Record */
         /* if attached part */
        DIS203_ENTITY_TYPE_RECORD           entity_type;
         /* if articulated part */
        struct {
            DIS203_PARAMETER_VALUE_TYPE value;   /*this is different from 203 */
            uint32                  _unused_5;   /* padding is gone from 204 */
        } numeric;
    } parameter_value;                  
} DIS204_ARTICULATION_PARAMETER_RECORD;        /*  way different from 203 */ 


/* 204
 * Supply Quantity Record
 *
 */

/*typedef struct {
    DIS204_ENTITY_TYPE_RECORD   supply_type;  
    DIS204_SUPPLY_QUANTITY_TYPE supply_quantity;  
} DIS204_SUPPLY_QUANTITY_RECORD;  */                /* same as 203 */


/* 204
 * Burst Descriptor Record
 *
 */

/*typedef struct {
    DIS204_ENTITY_TYPE_RECORD munition;  
    DIS204_WARHEAD_TYPE       warhead;  */ /* this is not defined in this file */
/*   DIS204_FUZE_TYPE          fuze;     */ /* this is not defined in this file */
/*    DIS204_QUANTITY_TYPE      quantity; 
    DIS204_RATE_TYPE          rate;   
} DIS204_BURST_DESCRIPTOR_RECORD;   */      /* same as 203 */


/* 204
 * Entity Identifier Record
 * 5.3.14.2
 */

/*typedef struct {
    DIS204_SIMULATION_ADDRESS_RECORD simulation_address;  
    DIS204_ENTITY_ID_TYPE            entity_id;  
} DIS204_ENTITY_ID_RECORD;   */      /* same as 203 */


/* 204
 * Event Identifier Record
 * 5.3.18
 */

/*typedef struct {
    DIS204_SIMULATION_ADDRESS_RECORD simulation_address;  
    DIS204_EVENT_ID_TYPE             event_id;  
} DIS204_EVENT_ID_RECORD;     */   /* same as 203 */ 


/* 204
 * Dead Reckoning Parameters
 * 5.4.3.1
 */

/*typedef struct {
    DIS204_DEAD_RECKONING_ALGORITHM_TYPE     dead_reckoning_algorithm;  
    unsigned                                    _unused_6_7               : 24; */
     /* padding */
/*    uint32                               _unused_6_8;*/  /* padding */
/*    uint32                               _unused_6_9;*/  /* padding */
/*    uint32                               _unused_6;*/  /* padding */
/*    DIS204_LINEAR_ACCELERATION_VECTOR_RECORD entity_linear_accel; 
    DIS204_ANGULAR_VELOCITY_VECTOR_RECORD    entity_angular_velocity;  
} DIS204_DEAD_RECKONING_PARAMETERS_RECORD;    */     /* same as 203 */


/* 204
 * Track Jam Record (for Emission PDU)
 * 
 */

/*typedef struct {
    DIS204_ENTITY_ID_RECORD       entity_id;  
    DIS204_EMITTER_ID_NUMBER_TYPE emitter_id;  
    DIS204_BEAM_ID_NUMBER_TYPE    beam_id;  
} DIS204_TRACK_JAM_RECORD;      */  /* same as 203 */
 /* END 2nd-LEVEL COMPONENTS */


#endif
