#
# File:     basic.cdf
# Contents: basic data elements
# System:   simnet development.
# Created:  10-Oct-1989
# Author:   kanarick
# Remarks:
#  History:
#    12/01/89  Kanarick	Vectors are now arrays
#    12/07/89  Kanarick	Added objectIrrelevant = 0
#    08/01/90	Pope		added BattleScheme type definition
#    12/01/90	Kanarick	Added timeUnknown, changed BattleScheme
#    15/01/90  Kanarick	Added new force IDs
#    22/01/90	Kanarick	Added simulator_SIMNET_STEALTH
#
# $RCSfile: basic.cdf,v $
# $Revision: 1.1 $
# $State: Exp $
# COPYRIGHT 1990 BBN Systems and Technologies Corporation
# 10 Moulton Street
# Cambridge, Ma. 02138
# 617-873-3000

hfile:
#include <stdtypes.h>
endhfile:

needs: address.cdf

# An angle, represented as a fixed-point fraction of a circle:
defs: Angle uint32

# A Battle Scheme describes how force IDs and vehicle guises are applied:
defs: BattleScheme uint8
  value: battleSchemeOther       0     # none of those listed below
  value: battleSchemeAbsolute    1    
  value: battleSchemeRelative    2    

# A Boolean value:
defs: SP_Boolean unsigned
hfile:
#ifndef CMS
endhfile:
  value: false                   0    
  value: true                    1
hfile:    
#endif
endhfile:

# #define to keep X header files happy
hfile:
#define Boolean SP_Boolean
endhfile:

# The type of some object, such as a vehicle or munition:
defs: ObjectType uint32
  value: objectIrrelevant        0

# A Burst Descriptor describes the firing of one or more rounds:
hfile:
typedef struct {
    ObjectType  projectile;           /* the type of projectile fired */
    ObjectType  detonator;            /* the type of fuze */
    uint16      quantity;             /* number of rounds in burst */
    uint16      rate;                 /* rounds per second */
} BurstDescriptor; 
endhfile:

defs: BurstDescriptor
  value: burstRateIrrelevant     0

# An event identifier refers to a particular event in the simulated world:
defs: EventID uint16
  value: eventIDIrrelevant       0

# Concurrent simulation exercises are distinguished by unique identifiers:
defs: ExerciseID uint8
  value: exerciseIDIrrelevant    0

# Each vehicle belongs to a force identified by a value of Force ID:
defs: ForceID uint8
  value: forceIDIrrelevant       0
  value: distinguishedForceID    1
  value: otherForceID            2
  value: observerForceID         3
  value: targetForceID           4

# An amount of some munition is represented by a Munition Quantity:
hfile:
typedef struct {
    ObjectType munition;              /* type of munition */
    float32    quantity;              /* quantity of it */
} MunitionQuantity; 
endhfile:

# An amphibious craft can carry things:
hfile:
typedef struct {
    ObjectType   type;              /* type of vehicle */
    int32        number;            /* number on board */
} AmpCrftInventory;
endhfile:

# Types of damages are identified by codes:
defs: DamageType uint16
  value: damageEverything        0

# Types of repairs are identified by codes:
defs: RepairType uint16
  value: repairEverything        0

# Types of simulators are identified by codes:
defs: SimulatorType uint16
  value: simulatorUnknown        0    
  value: simulator_SIMNET_MCC    1     # SIMNET MCC system
  value: simulator_SIMNET_SAF    2     # SIMNET SAF system
  value: simulator_SIMNET_M1     3     # SIMNET M1 (Abrams) simulator
  value: simulator_SIMNET_M2     4     # SIMNET M2/3 (Bradley) simulator
  value: simulator_SIMNET_FRED   5     # SIMNET rotary-wing simulator
  value: simulator_SIMNET_FWA    6     # SIMNET fixed-wing simulator
  value: simulator_SIMNET_FAAD_LOS_H 7 # SIMNET FAAD-LOS-H simulator
  value: simulator_SIMNET_STEALTH 8    # SIMNET Stealth simulator
  value: simulator_SIMNET_DI     9     # SIMNET Dismounted Infantry
  value: simulator_AGPT_LEO2     10    # AGPT Leo2 simulator
  value: simulator_AGPT_ULF      11    # AGPT Stealth simulator
  value: simulator_AGPT_DATA_LOGGER 12 # AGPT Data Logger
  value: simulator_SIMNET_LOSAT  13    # SIMNET Losat vehicle simualator
  value: simulator_SIMNET_SAFSTATION 14 # SIMNET SAF frontend interface
  value: simulator_SIMNET_SAFSIM 15    # SIMNET SAF backend simulator
  value: simulator_LL_SAFSIM     16    # Load-Leveling SAF backend
  value: simulator_LL_SAFSTATION 17    # Load-Leveling SAF frontend
  value: simulator_LL_ENVSIM     18    # Load-Leveling Environmental backend

# A terrain database is identified by a name and version number:
defs: TerrainDatabaseID
  value: maxTerrainNameLength    14
  value: latestTerrainVersion    0

hfile:
typedef struct {
    char    terrainName[maxTerrainNameLength]; 
    uint16  terrainVersion;      
} TerrainDatabaseID; 
endhfile:

# The condition of the sky is described by:
defs: SkyColor uint8
  value: skyColorClear           0     # No cloud cover
  value: skyColorPartlyCloudy    1     # 0-50 percent cloud cover
  value: skyColorPartlySunny     2     # 50-100 percent cloud cover
  value: skyColorOvercast        3     # total, light colored cloud cover
  value: skyColorRainy           4     # total, dark colored cloud cover

# The general weather conditions are described by the following, which is
# an enhancement of the skycolor, but is more encompassing:
defs: WeatherConditions uint8
  value: weatherClear            0    
  value: weatherRaining          1    
  value: weatherSnowing          2    
  value: weatherMisting          3    
  value: weatherSleeting         4    
  value: weatherHailing          5    
  value: weatherSandStorm        6    

# Absolute time is measured in seconds since 0 GMT, 1 January 1970:
defs: SP_Time uint32
  value: timeUnknown             0

# #define to keep X headers happy
hfile:
#define Time SP_Time
endhfile:

# A unit is identified by its type and number, e.g., 3rd Battalion:
defs: UnitType uint8
  value: unitTypeIrrelevant      0    
  value: unitTypeArmy            1    
  value: unitTypeBattalion       2    
  value: unitTypeBattery         3    
  value: unitTypeBrigade         4    
  value: unitTypeCompany         5    
  value: unitTypeCorps           6    
  value: unitTypeDivision        7    
  value: unitTypeFlight          8    
  value: unitTypeGroup           9    
  value: unitTypePlatoon         10   
  value: unitTypeRegiment        11   
  value: unitTypeSection         12   
  value: unitTypeSquad           13   
  value: unitTypeSquadron        14   
  value: unitTypeTaskForce       15   
  value: unitTypeTeam            16   
  value: unitTypeTroop           17   
  value: unitTypeWing            18   

defs: UnitNumber uint8
  value: unitNumberIrrelevant    0

hfile:
typedef struct {
    UnitNumber  unitNumber;          
    UnitType    unitType;            
} UnitIdentifier; 
endhfile:


# A vehicle's capabilities for interacting with other vehicles:
hfile:
typedef struct {
    unsigned  ammunitionSupply  : 1;     /* Boolean */
    unsigned  fuelSupply        : 1;     /* Boolean */
    unsigned  recovery          : 1;     /* Boolean */
    unsigned  repair            : 1;     /* Boolean */
    unsigned  _unused_1         : 28;    /* padding */
} VehicleCapabilities; 
endhfile:

# Vehicle classification by dead reckoning algorithm and degrees of freedom:
defs: VehicleClass uint8
  value: vehicleClassIrrelevant  0     # class irrelevant
  value: vehicleClassStatic      1     # static class
  value: vehicleClassSimple      2     # simple class
  value: vehicleClassTank        3     # tank class
  value: vehicleClassShip        4     # ship class



# A location wrt a vehicle is described by a Vehicle Coordinates:

# The origin of the coordinate system is taken to be the center
# of the vehicle, unless there is a VehicleComponent associated
# with the VehicleCoordinates.
hfile:
typedef float32 VehicleCoordinates[3]; 
endhfile:


# A vehicle's appearance to other vehicles is described by:
hfile:
typedef struct {
    ObjectType distinguished;        
    ObjectType other;                
} VehicleGuises; 
endhfile:

# Every vehicle in an exercise has a unique vehicle identifier:
hfile:
typedef struct {
    SimulationAddress simulator;      /* address of vehicle's simulator */
    uint16            vehicle;        /* unique for that simulator */
} VehicleID; 
endhfile:

defs: VehicleID
  value: vehicleIrrelevant   0

# Every object in a database has a unique identifier
hfile:
typedef struct {
    SimulationAddress simulator;      /* address of object's simulator */
    uint16            object;         /* unique for that simulator */
} ObjectID; 
endhfile:

# A data structure that is no longer really used
hfile:
typedef struct {
    uint32 time;                  /* HHMMSS */
    uint32 date;                  /* YYMMDD */
    uint32 secondsSince1970;     
} TimeDateGroup; 
endhfile:

# A vehicle's nameplate or bumper number:
defs: CharacterSetType uint8
  value: asciiCharacterSet       0    

defs: VehicleMarking
  value: maxVehicleMarkingLength 11

hfile:
typedef struct {
    CharacterSetType characterSet;   
    uint8            text[maxVehicleMarkingLength]; 
} VehicleMarking; 
endhfile:

# Velocity is in units of world coordinates per second:
hfile:
typedef float32 VelocityVector[3]; 
endhfile:

# A location in the simulated world is described by a World Coordinates:
# where WorldCoordinates are GCS coordinates (X,Y,Z,CELL3D)
hfile:
typedef float64 WorldCoordinates[4]; 
typedef float64 SIMNET_WorldCoordinates[3]; 
endhfile:

# Sometimes, one only wants the XY coordinates
# where WorldCoordinates are GCS coordinates (X,Y,CELL2D)
hfile:
typedef float64 XYCoordinates[3]; 
typedef float64 SIMNET_XYCoordinates[2]; 
endhfile:

# Sometime, one wants to save space at the expense of accuracy:
hfile:
typedef struct {
    int32 x;                             
    int32 y;                             
    int32 cell;                          
} PointLocation; 

typedef struct {
    float32 x;                            
    float32 y;                            
    float32 z;                            
    float32 cell;                         
} PointLocation3D; 
endhfile:

# OrganizationType describes the type of organizational heirarchy:
defs: OrganizationType uint8
  value: organizationIrrelevant  0    
  value: organizationArmy        1    

defs: OrganizationalUnit
  value: organizationalLevels    9

# An Oranizational Unit describes a branch in an organizational hierarchy:
hfile:
typedef struct {
    ForceID           force;            
    OrganizationType  organizationType; 
    UnitIdentifier    hierarchy[organizationalLevels]; 
} OrganizationalUnit; 
endhfile:


# What is known about a target being fired upon:
defs: TargetType uint8
  value: targetUnknown           0     # the target vehicle is not known
  value: targetIsNotVehicle      1     # target known, but not a vehicle
  value: targetIsVehicle         2     # target known, and a vehicle

# Description of target being fired upon:
hfile:
typedef struct {
    TargetType targetType;           
    uint8      _unused_2;             /* padding */
    VehicleID  vehicleID;            
} TargetDescriptor; 
endhfile:

# A component is a "major" part of a vehicle:
defs: VehicleComponent uint16
  value: hullComponent           1    
  value: turretComponent         2    
  value: airframeComponent       3    

# A vehicle can be directed to take certain types of damage:
defs: DirectedDamage uint8
  value: directedDamageNone      0    
  value: directedDamageMobility  1    
  value: directedDamageFirepower 2    
  value: directedDamageMobilityFirepower 3
  value: directedDamageCatastrophic 4 
  value: directedDamageCatastrophicNotBurning 5
  value: directedDamageCatastrophicShredded 6


# Descriptor for radar targetting:
hfile:
typedef struct {
    VehicleID   vehicleID;         /* Identity of the painted target */
     /* ARG!!!  the next field is a 32 bit field, but */
     /* this packet is not 32 bit alligned.  set up so can define */
     /* packet, and will have to do copy into field. */
     /* this field is labeled with what it really should  */
     /* represent.  for more on this, see RadiateVariant in p_sim.drn */
    uint16      radarData[2];      /* int32 */
} RadarTargetData; 
endhfile:

# Type of sight for a visual sensor:
defs: SightID uint16
  value: MAGNIFIED_SIGHT_GUNNER  0    
  value: MAGNIFIED_SIGHT_CMDR    1    
  value: MAGNIFIED_SIGHT_OTHER   2    

# Type of precipitation that is occurring:
defs: SimpleWeatherPrecipitationType uint32
  value: weatherPrecipitationNone 0   
  value: weatherPrecipitationOther 1  
  value: weatherPrecipitationRain 2   
  value: weatherPrecipitationSleet 3  
  value: weatherPrecipitationHail 4   
  value: weatherPrecipitationSnow 5   

# What is visibility like:
defs: SimpleWeatherExtinctionType uint32
  value: weatherExtinctionNone   0    
  value: weatherExtinctionOther  1    
  value: weatherExtinctionRural  2    
  value: weatherExtinctionMaritime 3  
  value: weatherExtinctionUrban  4    
  value: weatherExtinctionTropospheric 5
  value: weatherExtinctionAdvectiveFog 6
  value: weatherExtinctionRadiativeFog 7
  value: weatherExtinctionDesert 8    
  value: weatherExtinctionDesertSummer 9
  value: weatherExtinctionDesertWinter 10
  value: weatherExtinctionTemperateSummerDay 11
  value: weatherExtinctionTemperateSummerNight 12
  value: weatherExtinctionTemperateWinter 13
  value: weatherExtinctionDustStorm 14
  value: weatherExtinctionSnow   15   
  value: weatherExtinctionBlowingSnow 16
  value: weatherExtinctionFogSnow 17  

# DVO is BAND_47MM (0.4 - 0.7 um)
# NVO is BAND_49MM (0.4 - 0.9 um)
# IR is BAND_812MM (8.0 - 12.0 um)
# What wavelengths of light are we trying to see:
defs: WavebandType uint32
  value: wavebandUnknown         0    
  value: waveband47MM            1    
  value: waveband49MM            2    
  value: waveband712             3    
  value: waveband106             4    
  value: waveband35              5    
  value: waveband812             6    
  value: waveband1060            7    

defs: EnvCloud
  value: EnvCloudDomainMask      0x1ff
  value: EnvCloudDomainShift     0
  value: EnvCloudDefault         ( 0 << EnvCloudDomainShift )
  value: EnvCloudCirrus          ( 1 << EnvCloudDomainShift )
  value: EnvCloudCirrocumulus    ( 2 << EnvCloudDomainShift )
  value: EnvCloudCirrostratus    ( 4 << EnvCloudDomainShift )
  value: EnvCloudStratus         ( 8 << EnvCloudDomainShift )
  value: EnvCloudAltostratus     ( 16 << EnvCloudDomainShift )
  value: EnvCloudNimbostratus    ( 32 << EnvCloudDomainShift )
  value: EnvCloudStratocumulus   ( 64 << EnvCloudDomainShift )
  value: EnvCloudCumulus         ( 128 << EnvCloudDomainShift )
  value: EnvCloudAltocumulus     ( 256 << EnvCloudDomainShift )
  value: EnvCloudCompCumulusCirrus        ( EnvCloudCumulus |
	EnvCloudCirrus )
  value: EnvCloudCompStratocumulusCirrus  ( EnvCloudStratocumulus | 
	EnvCloudCirrus )


# Why is this in basic?

namespace_def: SHIPGUNS_MESSAGE
  value: FIRE_AT_TRACK    1
  value: FIRE_AT_LOCATION 2
  value: CIWS             3
  value: CEASE_FIRE       4

namespace_def: SHIPGUNS_MESSAGE_CHOICE
  value: FIRE_AT_TRACK    ( SHIPGUNS_MESSAGE_FIRE_AT_TRACK - 1 )
  value: FIRE_AT_LOCATION ( SHIPGUNS_MESSAGE_FIRE_AT_LOCATION - 1 )
  value: CIWS             ( SHIPGUNS_MESSAGE_CIWS - 1 )
  value: CEASE_FIRE       ( SHIPGUNS_MESSAGE_CEASE_FIRE - 1 )

namespace_def: SHIPGUNS_GUN
  value: FWD_5_54 0
  value: AFT_5_54 1
  value: FWD      2
  value: AFT      3

namespace_def: SHIPGUNS_CIWS
  value: STBD     0
  value: PORT     1
  value: FWD      2
  value: AFT      3
  value: FWD_STBD 4
  value: FWD_PORT 5
  value: AFT_STBD 6
  value: AFT_PORT 7
  value: ALL      8

namespace_def: SHIPGUNS_MUNITION
  value: VTIR  0
  value: HEPD  1
  value: HICAP 2
  value: WP    3
  value: MTPD  4
  value: 30HE  5
  value: 23AP  6
