/*  This test program illustrates how libbgr_ library calls and macros work.
**  It has a simple parser that allows symbols to be assembled and displayed
**  from text file descriptions.  It does NOT illustrate how to write such
**  a parser.
*/

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/bgrex.c,v 1.1 1999/04/12 18:23:24 ksnively Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "libbgr.h"
#include <stdalloc.h>
#include <stdlib.h>
#include <stdstring.h>

BgrDC	Dc;

Display	*Disp;
Window	Win;

static int	Xsize, Ysize;
static int	Gridx = 100, Gridy = 100;
static int	Xpos = 50, Ypos = 50;
static int	Row = 0, Col = 0;
static int	Nrows, Ncols;
static Bool	Linepips = FALSE;


unsigned long	GetNamedColor();
unsigned long	Red, Black, White, Green, Blue, Yellow, Gray;
GC		Friendgc, Foegc, Greengc, Yellowgc, Redgc, Whitegc, Blackgc;

int		Resized(), Readline(), Input();
void		UserHandler();

#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define MAX(a,b)	((a) > (b) ? (a) : (b))

#ifdef lint
#define exit (void)
#endif

main (argc, argv)
int	argc;
char**	argv;
{
    Widget shell;
    static Arg size[] = { XtNwidth, 700, XtNheight, 600 };
    extern time_t time();

    *argv = "symbols";
    srandom((int) time((time_t*) 0));

    shell = XtInitialize("symbols", "Symbols",
			 (XrmOptionDescRec*) NULL, 0,
			 &argc, argv);
    XtSetValues(shell, size, 2);

    XtAddEventHandler(shell, StructureNotifyMask, False,
		      Resized, (caddr_t) NULL);
    XtAddEventHandler(shell, ButtonPressMask, False,
		      Input, (caddr_t) NULL);
    XtAddInput(0, XtInputReadMask, Readline, (caddr_t) NULL);

    XtRealizeWidget(shell);
    Disp = XtDisplay(shell);
    Win = XtWindow(shell);

    Red = GetNamedColor(Disp, "red");
    Black = GetNamedColor(Disp, "black");
    White = GetNamedColor(Disp, "white");
    Green = GetNamedColor(Disp, "green");
    Blue = GetNamedColor(Disp, "skyblue");
    Yellow = GetNamedColor(Disp, "yellow");
    Gray = GetNamedColor(Disp, "dimgray");

    {	Arg color;
	XtSetArg(color, XtNbackground, Black);
	XtSetValues(shell, &color, 1);
    }

    /*  Create gcs  */

    Friendgc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Friendgc, Blue);
    XSetBackground(Disp, Friendgc, Gray);
    XSetLineAttributes(Disp, Friendgc, 0, LineSolid, CapProjecting, JoinBevel);

    Foegc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Foegc, Red);
    XSetBackground(Disp, Foegc, Gray);
    XSetLineAttributes(Disp, Foegc, 0, LineSolid, CapProjecting, JoinBevel);

    Redgc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Redgc, Red);
    XSetLineAttributes(Disp, Redgc, 0, LineSolid, CapButt, JoinBevel);

    Whitegc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Whitegc, White);
    XSetLineAttributes(Disp, Whitegc, 0, LineSolid, CapButt, JoinBevel);

    Blackgc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Blackgc, Black);
    XSetLineAttributes(Disp, Blackgc, 0, LineSolid, CapButt, JoinBevel);

    Greengc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Greengc, Green);
    XSetLineAttributes(Disp, Greengc, 0, LineOnOffDash, CapButt, JoinBevel);

    Yellowgc = XCreateGC(Disp, Win, 0, 0);
    XSetForeground(Disp, Yellowgc, Yellow);
    XSetLineAttributes(Disp, Yellowgc, 0, LineOnOffDash, CapButt, JoinBevel);

    /*  Initalize libbgr.
     *  Set up draw context.
     */

    if (!BgrInit(Disp)) 
        exit(1);

    bgrexInstallBgrSymbols();

    Dc = BgrCreateDC();
    BgrSetOutput(Dc, bgrXOutput);
    BgrSetXOutputInfo(Dc, Disp, Win);
    BgrSetXOutputUnits(Dc, 1);
    BgrSetXGC(Dc, Friendgc);
    BgrSetSize(Dc, 50.0);
    BgrSetTagSize(Dc, 0);
    BgrSetTagBorder(Dc, 3);
    BgrSetUserHandler(Dc, UserHandler);

    Resized(shell);
    XtMainLoop();

    /*NOTREACHED*/
}

unsigned long GetNamedColor (display, name)
Display	*display;
char	*name;
{
    XColor	requested, exact;
    Status	status;
    long	val;

    if (*name == '#')
    {
	sscanf(name+1, "%lx", &val);
	requested.red = (val & 0xff0000) >> 16;
	requested.green = (val & 0xff00) >> 8;
	requested.blue = (val & 0xff);
	status = XAllocColor(display,
			     DefaultColormap(display, DefaultScreen(display)),
			     &requested);
    }
    else
	status = XAllocNamedColor(display,
				  DefaultColormap(display,
						  DefaultScreen(display)),
				  name, &requested, &exact);

    if (!status)
    {
	fprintf(stderr, "can't alloc color '%s'\n", name);
	return(-1);
    }

    return(requested.pixel);
}

Resized (w)
Widget w;
{
    XWindowAttributes attribs;
    XGetWindowAttributes(Disp, XtWindow(w), &attribs);
    Xsize = attribs.width;
    Ysize = attribs.height;

    BgrSetXMappingAnchor(Dc, 0, Ysize * Dc->x11.unitsPerPixel);

    /*  reissue size command with old value; recalculates row and column
     *  counts based on new window size.
     */

    {	char buf[20];
	sprintf(buf, "size %d", (int) Dc->size);
	Command(buf);
    }
}

Input (w, cdata, event)
Widget		w;
caddr_t		cdata;
XButtonEvent	*event;
{
    static int npoints = 0;

    if (event->button == Button3)
    {
	Command("clear");
	Command("lzero");
	npoints = 0;
    }
    else
    {
	char buf[1000];
	sprintf(buf, "lpoint %d %d %d", npoints++,
		(int) (event->x / Dc->x11.unitsPerPixel),
		(int) ((Ysize - event->y) / Dc->x11.unitsPerPixel));
	Command(buf);
	if (event->button == Button2)
	{
	    Command("ldraw");
	    Command("lzero");
	    npoints = 0;
	}
    }
}

Readline (w)
Widget w;
{
    char buf[1000];
    if (!gets(buf))
	exit(0);
    Command(buf);
}

/*=====  command processing  ===============================================*/

/*  Parse a command from stdin or a data file; assembles symbols, displays
**  them, changes display parameters, etc.
*/

Command (s)
char *s;
{
    static BgrSymbolElt	elts[20];
    static int		nelts = 0;
    static BgrPoint	lpoints[50];
    static int		nlpoints = 0;
    static BgrPolySpec	pspec = { bgrGenericLine, 1, FALSE, FALSE, 20, 20 };
    static struct BgrCompPoly *cpoly = NULL;

    char		cmd[50], arg1[50], arg2[50], arg3[50], arg4[50];

    *cmd = *arg1 = *arg2 = *arg3 = *arg4 = '\0';
    sscanf(s, " %s %s %s %s %s", cmd, arg1, arg2, arg3, arg4);

#define ISCMD(c) (strcmp(cmd,c) == 0)

    if (!*cmd || *cmd == '#')
	;

    /*  alignment -- set friend or foe
     */

    else if (ISCMD("alignment"))
    {
	if (strcmp(arg1, "friend") == 0)
	    BgrSetXGC(Dc, Friendgc);
	else if (strcmp(arg1, "foe") == 0)
	    BgrSetXGC(Dc, Foegc);
	else
	    fputs("specify friend or foe\n", stderr);
    }

    /*  bounds -- indicate that the internal and external extents of a
     *  symbol are to be calculated; redisplay code below shows these bounds
     *  as dashed boxes.
     */

    else if (ISCMD("bounds"))
    {
	if (strcmp(arg1, "on") == 0)
	    BgrSetOutput(Dc, bgrXExtOutput);
	else if (strcmp(arg1, "off") == 0)
	    BgrClearOutput(Dc, bgrXExtOutput);
	else
	    fputs("specify on or off\n", stderr);
    }

    else if (ISCMD("clear"))
	XClearWindow(Disp, Win);

    else if (ISCMD("dashed"))
    {
	if (strcmp(arg1, "on") == 0)
	    BgrSetDashed(Dc, TRUE);
	else if (strcmp(arg1, "off") == 0)
	    BgrSetDashed(Dc, FALSE);
	else
	    fputs("specify on or off\n", stderr);
    }

    /*  draw -- draw currently assembled symbol.
     */

    else if (ISCMD("draw"))
    {
	if (nelts == 0) {
	    fputs("?no elements specified\n", stderr);
	    return;
	}

	BgrSetPosition(Dc, Xpos, Ypos);
	if (!BgrDrawSymbol(Dc, elts, nelts))
	    return;

	/*  draw bounding rectangles if specified by "bounds" command.
	 */

	if (Dc->flags & bgrXExtOutput)
	{
	    BgrExtents	*iext = &Dc->x11.iExtents;
	    BgrExtents 	*eext = &Dc->x11.eExtents;
	    int		x, y;

	    BgrMapX11Position(Dc, (float) Xpos, (float) Ypos, &x, &y);
	    XDrawRectangle(Disp, Win, Greengc,
			   x - iext->left + iext->xoff,
			   y - iext->up + iext->yoff,
			   iext->left + iext->right,
			   iext->up + iext->down);

	    if (iext->up != eext->up ||
		iext->down != eext->down ||
		iext->left != eext->left ||
		iext->right != eext->right)
		XDrawRectangle(Disp, Win, Yellowgc,
			       x - eext->left + eext->xoff,
			       y - eext->up + eext->yoff,
			       eext->left + eext->right,
			       eext->up + eext->down);
	}
    }

    else if (ISCMD("echo"))
	puts(s+5);

    else if (ISCMD("elt"))
    {
	extern BgrSymbolElt *BgrParseElement();
	char *start = strchr(strchr(s, 'e'), ' ');
	if (BgrParseElement(start ? start : "", &elts[nelts]))
	    ++nelts;
    }

    else if (ISCMD("file"))
    {
	FILE *fp = fopen(arg1, "r");
	if (!fp)
	    fprintf(stderr, "?can't open %s\n", arg1);
	else
	{
	    char inbuf[1000];
	    while (fgets(inbuf, sizeof(inbuf)-1, fp))
	    {
		inbuf[strlen(inbuf)-1] = '\0';
		Command(inbuf);
	    }
	    fclose(fp);
	    XFlush(Disp);
	}
    }

    else if (ISCMD("linemask"))
	BgrSetLineMask(Dc, atof(arg1));

    else if (ISCMD("polymask"))
	BgrSetPolyMask(Dc, atof(arg1));

    else if (ISCMD("new"))
	nelts = 0;

    else if (ISCMD("quit"))
	exit(0);

    /*  number of available rows and columns for displaying symbols is
     *  dictated by "size" command and dimensions of display window.  rc m n
     *  sets current position to be in row m, column n.  rc next advances to
     *  the next column
     */

    else if (ISCMD("rc"))
    {
	if (strcmp(arg1, "next") == 0)
	{
	    if (++Col == Ncols)
		++Row, Col = 0;
	    if (Row == Nrows)
		Row = 0;
	}
	else
	{
	    Row = atoi(arg1);
	    Col = atoi(arg2);
	}

	Xpos = Col * Gridx + Gridx/2;
	Ypos = Ysize - (Row * Gridy + Gridy/2);
	BgrSetPosition(Dc, Xpos, Ypos);
    }

    else if ISCMD("rot")
	BgrSetRotation(Dc, atof(arg1));

    /*  size -- set size of displayed symbols; alters number of rows and
     *  columns that can fit in the window.
     */

    else if (ISCMD("size"))
    {
	int size = atoi(arg1);
	BgrSetSize(Dc, size);

	Gridx = size * 1.5;
	Gridy = size * 1.5;
	Ncols = Xsize / Gridx;
	Nrows = Ysize / Gridy;
	Xpos = Gridx / 2;
	Ypos = Gridy / 2;
	Row = Col = 0;
    }

    else if (ISCMD("sleep"))
    {
	float sec = atof(arg1);
	struct timeval tm;
	tm.tv_sec = (int) sec;
	tm.tv_usec = (int) (fmod(sec, 1.0) * 1E6);
	select(0, (fd_set*) NULL, (fd_set*) NULL, (fd_set*) NULL, &tm);
    }

    else if (ISCMD("spin"))
    {
	int theta;
	for ( theta = 0; theta < 360; theta += 5 )
	{
	    BgrSetRotation(Dc, (float) theta);
	    Command("clear");
	    Command("draw");
	    XSync(Disp, True);
	    Command("sleep .1");
	}

	BgrSetRotation(Dc, 0.0);
	Command("clear");
	Command("draw");
    }

    else if (ISCMD("tborder"))
	BgrSetTagBorder(Dc, (float) atoi(arg1));

    else if (ISCMD("thickness"))
	BgrSetThickness(Dc, atof(arg1));

    else if (ISCMD("tsize"))
	BgrSetTagSize(Dc, atof(arg1));

    else if (ISCMD("uratio"))
	BgrSetUnitSizeRatio(Dc, atof(arg1));

    else if (ISCMD("units"))
    {
	BgrSetXOutputUnits(Dc, atof(arg1));
	BgrSetXMappingAnchor(Dc, 0, Ysize * Dc->x11.unitsPerPixel);
    }

    else if (ISCMD("usefonts"))
    {
	FILE *fp = fopen(arg1, "r");
	if (!fp)
	    perror("can't open file");
	else
	{
	    char name[100], *names[50];
	    int count = 0;
	    while (fgets(name, sizeof(name)-1, fp))
	    {
		if (*name == '#')
		    continue;
		name[strlen(name)-1] = '\0';
		strcpy(names[count++] = STDALLOC(strlen(name)+1), name);
	    }
	    fclose(fp);
	    BgrUseXFonts(Disp, names, count);
	    while (--count >= 0)
		STDDEALLOC(names[count]);
	}
    }

    else if (ISCMD("wait"))
    {
	fputs(s+5, stdout);
	while (getchar() != '\n')
	    ;
    }

    /*  xy -- explicitly set symbol output position; ignores current row and
     *  column position.
     */

    else if (ISCMD("xy"))
    {
	if (arg1[0] == '+') Xpos += atoi(&arg1[1]);
	else if (arg1[0] == '-') Xpos -= atoi(&arg1[1]);
	else Xpos = atoi(arg1);

	if (arg2[0] == '+') Ypos += atoi(&arg2[1]);
	else if (arg2[0] == '-') Ypos -= atoi(&arg2[1]);
	else Ypos = atoi(arg2);

	BgrSetPosition(Dc, Xpos, Ypos);
    }

    /*  Polyline commands.
     */

    else if (ISCMD("ldraw"))
    {
	if (cpoly)
	    BgrFreePoly(cpoly);
	BgrSetOutput(Dc, bgrXOutput);
	if (cpoly = BgrMakePoly(lpoints, nlpoints, &pspec))
	    Command("lredraw");
    }

    else if (ISCMD("lendelt"))
    {
	if (!*arg1)
	    pspec.end_elt = 0;
	else
	{
	    int elt = BgrDataName2Value(arg1);
	    if (elt)
		pspec.end_elt = elt;
	}
    }

    else if (ISCMD("lredraw"))
    {
	if (cpoly)
	{
	    BgrDrawPoly(Dc, cpoly);
	    if (Dc->flags & bgrXExtOutput)
	    {
		int p1x, p1y, p2x, p2y;
		BgrPolyBounds pb;
		BgrBoundPoly(cpoly, &pb, Dc);
		BgrMapX11Position(Dc, pb.xmin, pb.ymin, &p1x, &p1y);
		BgrMapX11Position(Dc, pb.xmax, pb.ymax, &p2x, &p2y);
		XDrawRectangle(Disp, Win, Greengc,
			       p1x, p1y, p2x-p1x, p2y-p1y);
	    }
	    if (Linepips)
	    {
		int i, x, y;
		for ( i = 0; i < nlpoints; i++ )
		{
		    BgrMapX11Position(Dc, lpoints[i].x, lpoints[i].y, &x, &y);
		    XDrawLine(Disp, Win, Whitegc, x-1, y-1, x+1, y+1);
		    XDrawLine(Disp, Win, Whitegc, x-1, y+1, x+1, y-1);
		}
	    }
	}
    }

    else if (ISCMD("lclosed"))
	pspec.closed = atoi(arg1);

    else if (ISCMD("lpoint"))
    {
	int idx = atoi(arg1);
	get_float(arg2, &lpoints[idx].x);
	get_float(arg3, &lpoints[idx].y);
	if (idx + 1 > nlpoints)
	    nlpoints = idx + 1;
    }

    else if (ISCMD("lpoints"))
    {
	char *p = strtok(s, " ");
	nlpoints = 0;
	while (p = strtok((char*) NULL, " "))
	{
	    get_float(p, &lpoints[nlpoints].x);
	    if (p = strtok((char*) NULL, " "))
		get_float(p, &lpoints[nlpoints++].y);
	}
    }

    else if (ISCMD("lfixsize"))
	pspec.fixed_size = atoi(arg1);

    else if (ISCMD("lpips"))
    {
	if (strcmp(arg1, "on") == 0)
	    Linepips = TRUE;
	else if (strcmp(arg1, "off") == 0)
	    Linepips = FALSE;
	else
	    fputs("specify on or off\n", stderr);
    }

    else if (ISCMD("lshift"))
    {
	int i;
	float xshift = atof(arg1), yshift = atof(arg2);
	for ( i = 0; i < nlpoints; i++ )
	{
	    lpoints[i].x += xshift;
	    lpoints[i].y += yshift;
	}
    }

    else if (ISCMD("lsplining"))
	pspec.splining = atoi(arg1);

    else if (ISCMD("ltype"))
    {
	int ptype = BgrDataName2Value(arg1);
	if (ptype)
	    pspec.type = ptype;
    }

    else if (ISCMD("lsizes"))
    {
	pspec.sizes[0] = atof(arg1);
	pspec.sizes[1] = atof(arg2);
	pspec.sizes[2] = atof(arg3);
    }

    else if (ISCMD("lzero"))
	nlpoints = 0;

    else
	fprintf(stderr, "?unknown command: '%s'\n", cmd);
}

static get_float (str, valp)
char	*str;
float	*valp;
{
    if (*str == '+')
	*valp += atof(str+1);
    else if (*str == '-')
	*valp -= atof(str+1);
    else
	*valp = atof(str);
}

void UserHandler (dc, elt, x, y, width, height)
BgrDC		dc;
BgrSymbolElt	*elt;
float		x, y;
float		width, height;
{
    int	bx, by, d = MIN(width,height) / 2;
    float sin30 = sin(30*3.14/180);
    float sin60 = sin(60*3.14/180);

    BgrMapX11Position(dc, x, y, &bx, &by);

    switch (elt->data)
    {
    case 1:
	XDrawArc(Disp, Win, Whitegc, bx - d, by - d,
		 d*2+1, d*2+1, 0*64, 360*64);
	XDrawLine(Disp, Win, Whitegc, bx, by, bx, by - d);
	XDrawLine(Disp, Win, Whitegc, bx, by,
		  (int) (bx + d*sin60), (int) (by + d*sin30));
	XDrawLine(Disp, Win, Whitegc, bx, by,
		  (int) (bx - d*sin60), (int) (by + d*sin30));
	break;
    case 2:
	XFillArc(Disp, Win, Greengc, bx - d, by - d,
		 d*2, d*2, 0*64, 90*64);
	XFillArc(Disp, Win, Yellowgc, bx - d, by - d,
		 d*2, d*2, 90*64, 90*64);
	XFillArc(Disp, Win, Redgc, bx - d, by - d,
		 d*2, d*2, 180*64, 90*64);
	XDrawLine(Disp, Win, Blackgc, bx, by - d, bx, by + d);
	XDrawLine(Disp, Win, Blackgc, bx - d, by, bx + d, by);
	XDrawArc(Disp, Win, Whitegc, bx - d, by - d, d*2, d*2, 0*64, 360*64);
	break;
    default:
	printf("whoa.  data == %ld\n", elt->data);
    }
}
