/*
**  composite.c
**
**  Author:		Christopher Ross
**  Creation date:	13-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	03-Jul-1991
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/composite.c,v 1.1 1999/04/12 18:23:27 ksnively Exp $";
#endif

#include <stdio.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

/*  This routine is hard to follow.  Read carefully, and have your
**  FM-101-5-1 open to pp. 2-23 and 2-24.
*/

INTERNAL void BgrDrawCompositeAviation (DRAWFNARGS)
DRAWFNDECL;
{
    static BgrSegment	*segs = NULL;
    register BgrSegment	*segp;
    float		u, u2, uhw;	/*  grid unit size of symbol	*/
    float		endy;		/*   and end of vertical seg	*/
    Bool		maskedp;
    long		what = celt->elt->data;

    /*  This describes how the various elements fit together.  up and down
     *  are how many vertical units each element adds TO THE END of the
     *  centerline, if any.  If own_mask is TRUE, the element is unusually
     *  wide, and will handle its own masking.  n_segs is how may segments
     *  are required to draw the element, unless it has a separate draw
     *  handler.
     */

    typedef struct {
	int	flag;
	float	up, down, out;
	int	n_segs;
	Bool	own_mask;
    } AvDimen;

    AvDimen *dimen;

#define HW	6		/* height of line, width of bowtie  */

    static AvDimen dimens[] =
    {
	{ bgrAvArmy,		0,	0,	HW/2,	0,	TRUE	},
	{ bgrAvAirForce,	0,	0,	HW/2,	0,	TRUE	},
	{ bgrAvAttack,		0,	0,	1,	2,	FALSE	},
	{ bgrAvAssault,		1,	0,	1,	2,	FALSE	},
	{ bgrAvTransport,	4,	0,	2,	0,	TRUE	},
	{ bgrAvFuel,		3,	0,	1.5,	0,	FALSE	},
	{ bgrAvRPV,		0,	2,	2,	0,	TRUE	},
	{ bgrAvBomber,		0,	2,	1,	0,	FALSE	},
	{ bgrAvRecon,		0,	1,	1,	1,	FALSE	},
	{ bgrAvHelo,		0,	0,	1,	1,	FALSE	},
	{ bgrAvMedium,		0,	0,	1,	1,	FALSE	},
	{ bgrAvHeavy,		0,	0,	1,	2,	FALSE	},
	{ bgrAvSovFWA,	0,	0,	HW/2,	2,	TRUE	},
	{ bgrAvSovRWA,	1,	0,	HW/2,	2,	TRUE	},
	{ bgrAvSovAttack,	0,	1,	1,	2,	FALSE	},
	{ 0 }
    };

    maskedp = (flags & P_MASK);

    /*  One-time allocation of maximum possible number of segments
     *  required.
     */

    if (!segs)
    {
	int n_segs = 0;
	for ( dimen = dimens; dimen->flag; ++dimen )
	    n_segs += dimen->n_segs;
	segs = (BgrSegment*) MALLOC(n_segs * sizeof(*segs));
    }

    segp = segs;

    /*  Determine unit-bounds of symbol.  u_width and u_height are the
     *  overall width and height of the symbol in dimensional units.
     *  u_up, _down, and _out are the number of units that the symbol
     *  extends from the draw point.  m_up, _down, and _out are the same
     *  quantities, but only for those elements that don't do their own
     *  masking.
     */

    {	float	u_width, u_height;
	float	u_up = 0, u_down = 0, u_out = 0;
	float	m_up = 0, m_down = 0, m_out = 0;

	for ( dimen = dimens; dimen->flag; ++dimen )
	{
	    if (!(what & dimen->flag))
		continue;

	    SETMAX(u_up, dimen->up);
	    SETMAX(u_down, dimen->down);
	    SETMAX(u_out, dimen->out);

	    if (maskedp && !dimen->own_mask)
	    {
		SETMAX(m_up, dimen->up);
		SETMAX(m_down, dimen->down);
		SETMAX(m_out, dimen->out);
	    }
	}

	/*  Determine unit dimension.  Overall width and height in units are
	 *  taken into account.  If we're not held by another element, we
	 *  can spill over the end, so leave out the unit height beyond the
	 *  centerline. 
	 */

	u_width = u_out * 2;
	u_up += HW;
	m_up += HW;

	if (flags & P_HELD)
	    u_height = u_up + u_down;
	else
	    u_height = HW;

	if (u_width == 0 && u_height == 0)
	    return;

	/*  u_height and u_width determine the aviation symbol's aspect
	 *  ratio; constrain the passed size.  The larger ovderall dimension
	 *  then determines the unit dimension.
	 */

	{   SFLOAT nwidth = width, nheight = height;
	    BgrConstrainDimensions((float) u_height / (float) u_width,
				   &nwidth, &nheight);
	    if (u_width > u_height)
		u = nwidth / u_width;
	    else
		u = nheight / u_height;
	}

	/*  Move y so that the symbol is actually entirely above the base
	 *  point.
	 */

	y += u * u_down;
	endy = y + u * HW;

	if (maskedp)
	    BgrDoBox(dc,
		     x - u * m_out, y - u * m_down,
		     x + u * m_out, y + u * m_up,
		     flags | P_MASK);
    }

    u2 = u * 2;
    uhw = u * HW/2;

    if (!maskedp)
    {
	/*  Centerline.
	 */

	segp->x1 = x;
	segp->y1 = y;
	segp->x2 = x;
	segp->y2 = endy;
	++segp;

	if (what & bgrAvAttack)
	{
	    segp->x1 = x;		/*  left barb	*/
	    segp->y1 = endy;
	    segp->x2 = x - u;
	    segp->y2 = endy - u;
	    ++segp;
	    *segp = segp[-1];		/*  right barb	*/
	    segp->x2 = x + u;
	    ++segp;
	}

	if (what & bgrAvAssault)
	{
	    segp->x1 = x;		/*  left barb	*/
	    segp->y1 = endy;
	    segp->x2 = x - u;
	    segp->y2 = endy + u;
	    ++segp;
	    *segp = segp[-1];		/*  right barb	*/
	    segp->x2 = x + u;
	    ++segp;
	}

	if (what & bgrAvSovAttack)
	{
	    segp->x1 = x;		/*  left barb	*/
	    segp->y1 = y;
	    segp->x2 = x - u;
	    segp->y2 = y - u;
	    ++segp;
	    *segp = segp[-1];		/*  right barb	*/
	    segp->x2 = x + u;
	    ++segp;
	}

	if (what & bgrAvHelo)
	{
	    segp->x1 = x - u;
	    segp->x2 = x + u;
	    segp->y1 = segp->y2 = y;
	    ++segp;
	}

	if (what & bgrAvRecon)
	{
	    segp->x1 = x - u;
	    segp->y1 = y - u;
	    segp->x2 = x + u;
	    segp->y2 = y + u;
	    ++segp;
	}

	if (what & bgrAvSovFWA)
	{
	    segp->x1 = x - uhw;
	    segp->x2 = x + uhw;
	    segp->y1 = segp->y2 = endy;
	    ++segp;
	    segp->x1 = x - uhw;
	    segp->x2 = x + uhw;
	    segp->y1 = segp->y2 = endy - u;
	    ++segp;
	}

	if (what & bgrAvSovRWA)
	{
	    segp->x1 = x - uhw;
	    segp->y1 = endy + u;
	    segp->x2 = x + uhw;
	    segp->y2 = endy - u;
	    ++segp;
	    segp->x1 = x - uhw;
	    segp->y1 = endy - u;
	    segp->x2 = x + uhw;
	    segp->y2 = endy + u;
	    ++segp;
	}

	/*  Medium and Heavy aircraft are denoted by the usual horizontal
	 *  segments, at 2 and 3 units above the base.
	 */

	if (what & (bgrAvMedium | bgrAvHeavy))
	{
	    segp->x1 = x - u;
	    segp->x2 = x + u;
	    segp->y1 = segp->y2 = y + u2;
	    ++segp;

	    if (what & bgrAvHeavy)
	    {
		*segp = segp[-1];
		segp->y1 += u;
		segp->y2 += u;
		++segp;
	    }
	}

	BgrDoSegments(dc, segs, segp - segs, flags);

	if (what & bgrAvBomber)
	    BgrDoCircle(dc, x, y - u, u,
			0.0, 360.0, flags | P_OUTLINE | P_FILL);

	if (what & bgrAvFuel)
	    BgrDoTriangle(dc, x, endy,
			  x - u*1.5, endy + u*3, x + u*1.5, endy + u*3,
			  flags | P_OUTLINE);
    }
    else
    {
	if (what & bgrAvSovFWA)
	    BgrDoBox(dc, x - uhw, endy, x + uhw, endy - u, flags | P_MASK);
	if (what & bgrAvSovRWA)
	    BgrDoBox(dc, x - uhw, endy + u, x + uhw, endy - u, flags | P_MASK);
    }

    /*  The following elements do their own masking.
     */

    if (what & bgrAvArmy)
	BgrQuickDrawElement(dc, bgrRArmyServiceBranch, x, y + u*4,
			    u*HW, u*HW, flags);
    else if (what & bgrAvAirForce)
	BgrQuickDrawElement(dc, bgrRAirForceServiceBranch, x, y + u*4,
			    u*HW, u*HW, flags);

    if (what & bgrAvTransport)
	BgrQuickDrawElement(dc, bgrRTransportation, x, endy + u2,
			    u2*2, u2*2, flags);

    if (what & bgrAvRPV)
	BgrQuickDrawElement(dc, bgrRUnmannedAirRecon, x, y,
			    u2*2, u2*2, flags);
} 

/*
*/

INTERNAL void BgrDrawCompositeWeaponry (DRAWFNARGS)
DRAWFNDECL;
{
    BgrSegment		segs[20];
    register BgrSegment	*segp = segs;

    float	u, u2, leftx, rightx;
    long	what = celt->elt->data;

    /*  Uncapped weapons run to 8 units above the base.  WpWeapon extends
     *  this 1 unit into the arrow head; WpRocket adds 2 onto that.
     *  WpMissile is 1 unit above normal.  WpHighTraj and FlatTraj add an
     *  extra unit at the bottom.
     *
     *  Basic width is 4 units; missiles add 2 units at the base, unless
     *  NoBase has been specified.
     */

    {	int unitsUp = 8, unitsDown = 0;
	int unitsOut = 2;

	if (what & (bgrWpWeapon | bgrWpMissile))
	    ++unitsUp;
	else if (what & bgrWpRocket)
	    unitsUp += 3;
	if (what & (bgrWpHighTraj | bgrWpFlatTraj))
	    ++unitsDown;
	if ((what & bgrWpMissile) && !(what & bgrWpNoBase))
	    ++unitsOut;

	if (flags & P_HELD)
	    u = height / (unitsUp + unitsDown);
	else u = height / 8;

	u2 = u * 2;
	leftx = x - u2;
	rightx = x + u2;
	y += u * unitsDown;

	RFASTMASK(dc, x - u * unitsOut, y - u * unitsDown,
		  x + u * unitsOut, y + u * unitsUp, flags);
	RFASTBOUNDS(dc, x - u * unitsOut, y - u * unitsDown,
		    x + u * unitsOut, y + u * unitsUp, flags);
    }

    /*  Center line segment of weapon starts at draw point; trajectory and
     *  air defense elements push this up by one unit.  Top of this segment
     *  is normally 8 units higher, pushed down by a missile, pushed up into
     *  a weapon/rocket arrowhead.
     */

    segp->x1 = segp->x2 = x;
    segp->y1 = (what & (bgrWpHighTraj | bgrWpFlatTraj | bgrWpAirDefense))
		? y + u : y;
    if (what & (bgrWpWeapon | bgrWpRocket))
	segp->y2 = y + u*9;
    else if (what & bgrWpMissile)
	segp->y2 = y + u*7;
    else
	segp->y2 = y + u*8;
    ++segp;

    /*  Medium-weapon bar is drawn 4 units above base point.  Heavy-weapon
     *  bars are 3 and 5 units above.  Bars are 4 units wide, unless a
     *  missile cap is present, in which case they're 2 wide.
     */

    if (what & bgrWpMedium)
    {
	float d = (what & bgrWpMissile ? u : u2);
	segp->x1 = x - d;
	segp->x2 = x + d;
	segp->y1 = segp->y2 = y + u*4;
	++segp;
    }
    else if (what & bgrWpHeavy)
    {
	float d = (what & bgrWpMissile ? u : u2);
	segp->x1 = x - d;
	segp->x2 = x + d;
	segp->y1 = segp->y2 = y + u*3;
	++segp;
	segp->x1 = x - d;
	segp->x2 = x + d;
	segp->y1 = segp->y2 = y + u*5;
	++segp;
    }

    /*  A missile cap is 9 units high; two 7-unit segments and a top
     *  hemisphere 2 units in radius.  Its bottom part is one unit below the
     *  base if it has to contain either trajectory part.
     */

    if (what & bgrWpMissile)
    {
	float ybase = (what & (bgrWpHighTraj | bgrWpFlatTraj))
			? y - u : y;
	float ytop = y + u*7;

	segp->x1 = segp->x2 = x - u2;
	segp->y1 = ybase;
	segp->y2 = ytop;
	++segp;
	segp->x1 = segp->x2 = x + u2;
	segp->y1 = ybase;
	segp->y2 = ytop;
	++segp;

	if (!(what & bgrWpNoBase))
	{
	    segp->x1 = x - u*3;
	    segp->x2 = x + u*3;
	    segp->y1 = segp->y2 = ybase;
	    ++segp;
	}

	BgrDoCircle(dc, x, ytop, u2, 0.0, 180.0,
		    flags | P_OUTLINE);
    }

    /*  Non-missiles may have a gun part, and a weapon single- or rocket
     *  double-arrow part.
     */

    else
    {
	if (what & (bgrWpWeapon | bgrWpRocket))
	{
	    segp->x1 = x - u2;
	    segp->y1 = y + u*7;
	    segp->x2 = x;
	    segp->y2 = y + u*9;
	    ++segp;
	    *segp = segp[-1];
	    segp->x1 = x + u2;
	    ++segp;

	    if (what & bgrWpRocket)
	    {
		segp->x1 = x - u2;
		segp->y1 = y + u*9;
		segp->x2 = x;
		segp->y2 = y + u*11;
		++segp;
		*segp = segp[-1];
		segp->x1 = x + u2;
		++segp;
	    }
	}

	if (what & bgrWpGun)
	{
	    segp[0].x1 = segp[0].x2 = x - u2;
	    segp[1].x1 = segp[1].x2 = x + u2;
	    segp[0].y1 = segp[1].y1 = y + u2;
	    segp[0].y2 = segp[1].y2 = y + u*6;
	    segp += 2;
	}
    }

    /*  Do one of high trajectory circle, flat trajectory arrow, or air
     *  defense semicircle.
     */

    if (what & bgrWpFlatTraj)
    {
	segp->x1 = x - u2;
	segp->y1 = y - u;
	segp->x2 = x;
	segp->y2 = y + u;
	++segp;
	*segp = segp[-1];
	segp->x1 = x + u2;
	++segp;
    }
    else if (what & bgrWpHighTraj)
	BgrDoCircle(dc, x, y, u, 0.0, 360.0, flags | P_OUTLINE);
    else if (what & bgrWpAirDefense)
    {
	BgrDoArc3P(dc, leftx, y, x, y+u, rightx, y,
		   TRUE, flags | P_OUTLINE);

	/*  If the weapon does not have a bottom horizontal segment from a
	 *  missile part, supply one unless told not to.
	 */

	if (!(what & (bgrWpMissile | bgrWpNoBase)))
	{
	    segp->x1 = leftx;
	    segp->x2 = rightx;
	    segp->y1 = segp->y2 = y;
	    ++segp;
	}
    }

    BgrDoSegments(dc, segs, segp - segs, flags);
} 

#endif /* USE_X */
