/*
**  libbgr.h
**
**  Author:		Christopher Ross
**  Creation date:	17-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	09-Sep-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
**
**  Much of the documentation for symbol elements and flags is in the libbgr
**  man page.  The comments herein deal primarily with libbgr structs.
*/

#ifndef _libbgr_h_
#define _libbgr_h_

#ifdef __cplusplus
extern "C" {
#endif

#ifdef USE_X

/*
    $Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/libbgr.h,v 1.1 1999/04/12 18:23:33 ksnively Exp $
*/


#include <X11/Intrinsic.h>
#include <stdalloc.h>
#include <stdtypes.h>

#define BgrVERSION	2
#define BgrREVISION	3

#define BgrInit(disp) BgrInitWithDisplay(disp, BgrVERSION, BgrREVISION)

/*  Limits.
*/

#define bgrMaxSymbolElts	50

/*  BgrSymbolElt specifies part of a symbol; all symbols in all classes are
**  made up of lists of this structure.  .id specifies the unique integer
**  code for the element; possible values are listed below.  .data contains
**  arbitrary information specific to the element code; some use it, some
**  ignore it.  .flags controls certain per-element attributes (mainly
**  related to layout) which may be configured at runtime.
**
**  This is not implemented as a union because a union will only get in our
**  way, especially in static aggregate initialization (which most compilers
**  cannot handle.)
*/

typedef struct {
    short	id;
    short	flags;
    long	data;
    char	tag[24];
} BgrSymbolElt;

/*  Possible values for BgrSymbolElt.id.  These are in more or less logical
**  groups based on their appearance in FM-101-5-1.  Unless otherwise
**  specified, new identifiers must be added at the _end_ of one of these
**  groups; existing values must _not_ change, since applications using
**  libbgr may be databasing them.  Don't attempt to preserve alphabetical
**  order within groups.
*/

#define bgrNoElement		0

/*  Control elements.
*/

#define bgrPopElement		1

/*  Miscellaneous elements that rarely appear by themselves, and are the
**  most often used to /Hold other elements.
*/

#define bgrUnitRect		10
#define bgrRoundPoint		11
#define bgrSpikePoint		12
#define bgrDiamond		13
#define bgrUnitRectB		14
#define bgrTriangleUp		15

/*  Blank holders and spacers.
*/

#define bgrCenterHolder		30
#define bgrTopHolder		31
#define bgrBottomHolder		32
#define bgrBlankSpace		33
#define bgrThinSpace		34
#define bgrUserSymbol		35
    
/*  These elements are themselves complex composites; their appearance is
**  determined by ORable values in the element .data field; these values are
**  listed below.  Aviation symbols are on pp 2-23 and 2-24 of FM-101-5-1.
**  Weaponry symbols are on pp 2-19 to 2-21.
*/

#define bgrCompositeAviation	50
#define bgrCompositeWeaponry	51

/*  Role elements handle many of the symbols shown on pp 2-7 to 2-19.  These
**  are most often held within other elements.
*/

#define bgrRAerialObservation		100
#define bgrRAirDefense			101
#define bgrRAirForceServiceBranch	102
#define bgrRAmmunition			103
#define bgrRAmphibious			104
#define bgrRAntiArmor			105
#define bgrRArmor			106
#define bgrRArmyServiceBranch		107
#define bgrRArtillery			108
#define bgrRBIFV			109
#define bgrRBridging			110
#define bgrRCavalryRecon		111
#define bgrRChemical			112
#define bgrREngineer			113
#define bgrRFinance			114
#define bgrRGraveService		115
#define bgrRHospital			116
#define bgrRInfantry			117
#define bgrRLinkup			118
#define bgrRMaintenance			119
#define bgrRMajorEndItems		120
#define bgrRMedical			121
#define bgrRMissile			122
#define bgrRMotorized			123
#define bgrRMountain			124
#define bgrRNATOAirborne		125
#define bgrRNATOAirAssault		126
#define bgrRNATOAirAssaultInsuff	127
#define bgrROrdnance			128
#define bgrRParking			129
#define bgrRPersonal			130
#define bgrRPetroleum			131
#define bgrRPsychological		132
#define bgrRQuartermaster		133
#define bgrRRepairParts			134
#define bgrRRocketArtillery		135
#define bgrRSignalComm			136
#define bgrRSmallArms			137
#define bgrRSolidFuel			138
#define bgrRSoundRanging		139
#define bgrRSubsistence			140
#define bgrRSupply			141
#define bgrRSurvey			142
#define bgrRTextual			143
#define bgrRTopographic			144
#define bgrRTrafficControl		145
#define bgrRTransportation		146
#define bgrRUnmannedAirRecon		147
#define bgrRUSAirAssault		148
#define bgrRUSAirAssaultInsuff		149
#define bgrRUSAirborne			150
#define bgrRWater			151
#define bgrRMP                          152
#define bgrRMechEngineer                 153

/*  Mobility elements, p 2-23.
*/

#define bgrMbAmphibious			200
#define bgrMbBargeOrBoat		201
#define bgrMbOverSnow			202
#define bgrMbTowedTrailer		203
#define bgrMbSelfPropelled		204
#define bgrMbWheeled			205
#define bgrMbCrossCountry		206
#define bgrMbWheeledTracked		207

/*  Vehicle symbols, p 2-22 et al.
*/

#define bgrVehAPC			250
#define bgrVehArmoredEngineer		251
#define bgrVehAVLB			252
#define bgrVehBIFV			253
#define bgrVehCFV			254
#define bgrVehHovercraft		255
#define bgrVehTractor			256
#define bgrVehTank			257
#define bgrVehCargo			258
#define bgrVehACE                       259
#define bgrVehAVLM                      260
#define bgrVehHAB                       261
#define bgrVehCEV                       262
#define bgrVehTrailerMICLIC             263

/*  Tag elements.  These deal with most of the textual positions shown on
**  page 2-73.
*/

#define bgrLeftTag1		300
#define bgrLeftTag2		301
#define bgrLeftTag3		302
#define bgrLeftTag4		303
#define bgrRightTag1		304
#define bgrRightTag2		305
#define bgrRightTag3		306
#define bgrRightTag4		307
#define bgrTopTag1		308
#define bgrBottomTag1		309

/*  Unit sizes, pp 2-5, 2-6.  See below for element data values applying to
**  unit sizes.
*/

#define bgrUSquad		350
#define bgrUSection		351
#define bgrUPlatoon		352
#define bgrUCompany		353
#define bgrUBattalion		354
#define bgrURegiment		355
#define bgrUBrigade		356
#define bgrUDivision		357
#define bgrUCorps		358
#define bgrUArmy		359
#define bgrUArmyGroup		360
#define bgrUTheaterArmy		361

/*  Miscellaneous point symbols, p. 2-28 et. al.
*/

#define bgrCoordinatingPoint	400
#define bgrContactPoint		401
#define bgrEarthwork		402
#define bgrSurfaceShelter	403
#define bgrUndergroundShelter	404
#define bgrNuclearEvent		405
#define bgrBioChemEvent		406
				/* 407 */
				/* 408 */
#define bgrAbatis		409
#define bgrReferencePoint	410
#define bgrFiredDemolition	411
#define bgrAntipersonnelMine	412
#define bgrAntitankMine		413
#define bgrUnspecifiedMine	414
#define bgrTankDitchPoint	415
#define bgrPlannedDemolition	416
#define bgrPreparedDemolition1	417
#define bgrPreparedDemolition2	418
				/* 419 */
#define bgrTargetRefPoint	420
#define bgrFLOTPoint		421

/*  Headquarters staff, p 2-26.  When HqAnchor (see .data values below)
**  is specified, the anchor is changed based on the dimensions of the
**  _first_ inner element.
*/

#define bgrHqStaff		550
    
/*  ORable .data values for use when BgrSymbolElt.id is
**  bgrCompositeAviation.  Like the element ids, these must not change.
*/

#define bgrAvArmy		(1<<31)
#define bgrAvAirForce		(1<<30)

#define bgrAvAttack		(1<<29)
#define bgrAvTransport		(1<<28)
#define bgrAvFuel		(1<<27)
#define bgrAvAssault		(1<<26)

#define bgrAvRPV		(1<<25)
#define bgrAvHelo		(1<<24)
#define bgrAvBomber		(1<<23)
#define bgrAvRecon		(1<<22)

#define bgrAvMedium		(1<<21)
#define bgrAvHeavy		(1<<20)

#define bgrAvSovFWA		(1<<19)
#define bgrAvSovRWA		(1<<18)
#define bgrAvSovAttack		(1<<17)

/*  ORable .data values for use when BgrSymbolElt.id is
**  bgrCompositeWeaponry.
*/

#define bgrWpGun		(1<<31)
#define bgrWpWeapon		(1<<30)
#define bgrWpMissile		(1<<29)
#define bgrWpRocket		(1<<28)

#define bgrWpHighTraj		(1<<27)
#define bgrWpFlatTraj		(1<<26)
#define bgrWpAirDefense		(1<<25)

#define bgrWpMedium		(1<<24)
#define bgrWpHeavy		(1<<23)

#define bgrWpNoBase		(1<<22)

/*  ORable .data values for use when the symbol element is a unit size.
*/

#define bgrUGrouping		(1<<31)
#define bgrUBigMask		(1<<30)
#define bgrUUnattached		(1<<29)
#define bgrUAtopHolder		(1<<28)

/*  ORable .data values for vehicle elements.
*/

#define bgrVehDirectionPip	(1<<31)
#define bgrVehMedium		(1<<30)
#define bgrVehHeavy		(1<<29)

/*  .data values for use when symbol element is bgrHqStaff.  Use one of
**  Left, Center, and Right, optionally ORed with Anchor, which will make
**  the symbol anchor at the base of the staff.
*/

#define bgrHqLeft		1
#define bgrHqCenter		2
#define bgrHqRight		3
#define bgrHqLocMask		03

#define bgrHqAnchor		(1<<31)

/*  ORable values for SymbolElt.flags.
**  These are documented in the man page.
*/

#define bgrHold		(1)
#define bgrColumn	(1<<1)
#define bgrStack	bgrColumn
#define bgrFill		(1<<2)
#define bgrRow		(1<<3)
#define bgrExplicit	(1<<4)
#define bgrFilled	(1<<5)

/*=====  polyline stuff  ====================================================*/

/*  Attribute struct for compiling polylines.  .type is one of the constants
**  listed below.  .splining is the number of output points for each input
**  point; 1 means no splining.  .closed means a cyclic spline; libbgr will
**  automatically connect the last point to the first point.  .sizes[]
**  specifies things like the head and tail widths of a wide polyline; its
**  use is specific to each poly type.  .fixed_size TRUE means that .sizes[]
**  will be in device output units (e.g., pixels); FALSE means that the
**  sizes are in world coordinates and will scale.
*/

typedef struct {
    int		type;
    short	splining;
    char	closed;
    char	fixed_size;
    float	sizes[3];
    int		end_elt;
} BgrPolySpec;

struct BgrCompPoly;

/*  Possible values for BgrPolySpec.type.
*/

#define bgrGenericLine			1
#define bgrDirectionOfAdvance		2
#define bgrMainDirectionOfAdvance	3
#define bgrAxisOfAdvance		4
#define bgrMainAxisOfAdvance		5
#define bgrRectangularTarget		6
#define bgrUnspecifiedMineField		7
#define bgrAntitankMineField		8
#define bgrAntipersonnelMineField	9
#define bgrTankDitch			10
#define bgrLinearTarget			11
#define bgrLinearAbatis			12
#define bgrFLOT				13

/*=====  public types and macros  ==========================================*/

typedef unsigned long BgrFlags;

/*  BgrPoint is libbgr's version of XPoint, used for line drawing.
*/

typedef struct BgrPoint {
    float x, y;
} BgrPoint;

/*  BgrExtents is used to return X11 symbol extent information to the
**  application.  Distances given are in pixels from the center specified in
**  the draw context.
*/

typedef struct {
    short up, down;
    short left, right;
    short xoff, yoff;
} BgrExtents;

/*  BgrPolyBounds is used to return information from BgrGetPolyBounds.
*/

typedef struct {
    float	xmin, ymin;
    float	xmax, ymax;
} BgrPolyBounds;

/*  This is used internally for bounds calculation, but must be defined
**  here because it's used in BgrDC.
*/

typedef struct {
    float	up, down;
    float	left, right;
} BgrAxisBounds;

/*  The draw context tells libbgr how to output a symbol to the screen
**  device.  Many of the fields are of interest only to libbgr; they are
**  shown here because both the application and the internals of libbgr need
**  to see them.  Draw contexts must NOT be accessed except via the macros
**  defined below.
*/

typedef struct BgrDC {

    /*  flags is an OR of the following values:
     *
     *  bgrXOutput causes graphical output using the X display, window, and
     *  GC given in the .x11 substruct below.  bgrXExtOutput causes X
     *  extents information to be written.
     */

    BgrFlags	flags;

#define bgrXOutput	(1)
#define bgrXExtOutput	(1 << 1)

#define bgrAllOutput	(bgrXOutput | bgrXExtOutput)

    /*  .x,.y is the user-specified output position for a drawn symbol, in
     *  terms of a Cartesian world-coordinate system defined on top of the
     *  device output coordinates.
     */

    float	x, y;

    /*  .size is the larger dimension (X vs Y) of the symbol to be output;
     *  width of unit symbols and vehicles, diameter of round points, height
     *  of aviation elements, etc.
     */

    float	size;

    /*  .tagSize is used to determine the point size of fonts drawn for
     *  symbol tags.  This defaults to 0, which results in a point size
     *  roughly 1/3 the height of the symbol.  If non-zero, libbgr will use
     *  the largest font it knows of equal to or below the specified point
     *  size.  .tagBorder is the distance in device units between the edge
     *  of the symbol and tag text; it defaults to zero.
     */

    float	tagSize;
    float	tagBorder;

    /*  .thickness (default 1) is the thickness of output lines.  .poly_fill
     *  means that closed polylines drawn with BgrDrawPoly are to be filled.
     *  .poly_mask, if nonzero, means that symbols are to be output with a
     *  filled background which extends that number of world units beyond
     *  the outermost edges.  Alternately, .line_mask, if nonzero, means
     *  that symbol lines are to be outlined with a background color line
     *  which is that number of units thicker than .thickness.
     */

    float	thickness;
    float	poly_mask;
    float	line_mask;
    Bool	poly_fill;

    /*  .usize_ratio is the ratio of the height of attached unit symbols to
     *  the size of the element atop which they are drawn; this is most
     *  often the outermost element.
     */

    float	usize_ratio;

    /*  .theta is the user-specified symbol rotation in clockwise degrees;
     *  .sin_theta and .cos_theta are set by libbgr when .theta is nonzero;
     *  they specify the canonical values for the required 2D transform.
     */

    float	theta;
    float	sin_theta;
    float	cos_theta;

    /*  These are used internally to determine extents as a symbol is drawn.
     *  .rb is the raw bounds; it tracks the edges of the symbol geometry
     *  independent of rotation and masking; .eb is the effective bounds,
     *  with rotation and masking taken into account.
     *
     *  Tags, for example, determine their position using the raw bounds,
     *  since use of the effective bounds in a rotated or masked symbol
     *  would put them too far away.
     */

    BgrAxisBounds	rb;
    BgrAxisBounds	eb;

    /*  dashed indicates that conditional dashed lines are enabled; see the
     *  manual page.
     */

    Bool		dashed;

    /*  Information for X output.
     */

    struct
    {
	/*  Display, window, and GC must be specified to a draw context for
	 *  X output.
	 */

	Display		*disp;
	Window		win;
	GC		gc;

	/*  These parameters map libbgr world coordinates to the output
	 *  drawable.  .xanchor and .yanchor are the world coordinates which
	 *  correspond to 0,0 in the drawable.  .unitsPerPixel specifies
	 *  output scaling.
	 */

	float		xanchor, yanchor;
	float		unitsPerPixel;

	/*  More internals.  During symbol output, all coordinates are
	 *  specified relative to a world 0,0; this corresponds to .xbase,
	 *  .ybase, which is the symbol's anchor in device coordinates.
	 *  .symbolRelative is set to TRUE.  If it is FALSE, coordinate
	 *  output mapping is done normally, relative to .xanchor and
	 *  .yanchor.
	 */

	int		xbase, ybase;
	Bool		symbolRelative;

#ifdef BGR_INTERNALS

#define DCX2SCREENX(dc, x) \
    (dc->x11.symbolRelative ? dc->x11.xbase + ROUND(x) \
     : (int) (((x) - (dc)->x11.xanchor) / dc->x11.unitsPerPixel))
#define DCY2SCREENY(dc, y) \
    (dc->x11.symbolRelative ? dc->x11.ybase - ROUND(y) \
     : (int) (((dc)->x11.yanchor - (y)) / dc->x11.unitsPerPixel))
					     
#endif

	/*  .iExtents and .eExtents will be written if bgrXExtOutput is
	 *  specified in .flags above.  They give the symbol extents in
	 *  pixels from the device output location corresponding to
	 *  BgrDC.x,.y.
	 */

	BgrExtents	iExtents;
	BgrExtents	eExtents;

	/*  The various output routines may desire to change line dashing
	 *  and thickness more times than actually necessary.  gc_dashing
	 *  and gc_thickness are the values currently set in the GC, or -1
	 *  if we don't know what values are set.  These prevent us from
	 *  making unnecessary XSetFoo calls.
	 */

	short		gc_dashing;
	short		gc_thickness;
    }
    x11;

    /*  These are set if the symbol bgrUserSymbol is to be used.
     */

    void	(*user_hdlr)();
    ADDRESS	user_data;

} _BgrDC, *BgrDC;

/*  Macros for modifying a draw context.
*/

#define BgrSetOutput(dc,out)		((dc)->flags |= (out))
#define BgrClearOutput(dc,out)		((dc)->flags &= ~(out))
#define BgrSetSize(dc,s)		((dc)->size = (s))
/* #define BgrSetRotation(dc,ang)	procedure  */
#define BgrSetTagSize(dc,s)		((dc)->tagSize = (s))
#define BgrSetTagBorder(dc,b)		((dc)->tagBorder = (b))
#define BgrSetUnitSizeRatio(dc,r)	((dc)->usize_ratio = (r))
#define BgrSetPolyMask(dc,pm)		((dc)->poly_mask = (pm))
#define BgrSetLineMask(dc,lm)		((dc)->line_mask = (lm))
#define BgrSetPolyFill(dc,pf)		((dc)->poly_fill = (pf))
#define BgrSetDashed(dc,d)		((dc)->dashed = (d), \
					 (dc)->x11.gc_dashing = -1)
#define BgrSetThickness(dc,t)		((dc)->thickness = (t), \
					 (dc->x11.gc_thickness = -1))
#define BgrSetUserHandler(dc,h)		((dc)->user_hdlr = (h))
#define BgrSetUserData(dc,d)		((dc)->user_data = (d))

#define BgrSetPosition(dc,xp,yp) ((dc)->x = (xp), (dc)->y = (yp))

#define BgrSetXOutputUnits(dc,u) ((dc)->x11.unitsPerPixel = (u))
#define BgrSetXMappingAnchor(dc,x,y) \
    ((dc)->x11.xanchor = (x), (dc)->x11.yanchor = (y))
#define BgrSetXOutputInfo(dc,d,w) \
    ((dc)->x11.disp = (d), (dc)->x11.win = (w))
#define BgrSetXGC(dc,g) ((dc)->x11.gc = (g), (dc)->x11.gc_dashing = -1, \
			 (dc)->x11.gc_thickness = -1)

/*=====  public routines  ==================================================*/

/* BEGIN_PUBLIC_DECLS */

/*  PUBLIC declarations extracted by excdecls
**  at Tue Nov 26 18:13:01 EST 1991 with command:
**
**  excdecls -p PUBLIC -a libbgr.h composite.c data.c misc.c poly.c prims.c
**   symbol.c tags.c utils.c
**  Don't modify this section by hand if you normally use excdecls.
*/

extern void BgrSetRotation ();
extern void BgrInstallEltInfo ();
extern int BgrGenerateCurve ();
extern void BgrGetNearbyPoints ();
extern struct BgrCompPoly *BgrMakePoly ();
extern void BgrBoundPoly ();
extern void BgrFreePoly ();
extern void BgrDrawPoly ();
extern int BgrDrawSymbol ();
extern int BgrInitWithDisplay ();
extern int BgrUseXFonts ();
extern BgrDC BgrCreateDC ();
extern void BgrMapX11Position ();
extern BgrSymbolElt *BgrParseElement ();
extern int BgrDataName2Value ();

/* END_PUBLIC_DECLS */

/*=====  INTERNALS  ========================================================*/

#ifdef BGR_INTERNALS

/*  Internal table information for each element code.
*/

typedef struct {

    /*  .name is the text of the C identifier for the element id, stripped of
     *  the 'bgr' prefix.  .id is the actual element code.  .id is the first
     *  field so it matches the BgrDataInfo type.
     */

    int		id;
    char	*name;

    /*  .compiler is invoked by BgrCompileSymbol to compile each element in
     *  the user-specified list.  .drawer is invoked during output to run
     *  the graphics primitives for drawing this element.  .layout is used
     *  to arrange the subelements of an element for drawing; see the
     *  BgrCompSym type, below.
     */

    Bool	(*compiler)();
    void	(*drawer)();
    void	(*layout)();

    /*  .gflags specifies general element attribute flags, common to a wide
     *  variety of elements.  Possible values:
     *  
     *  V_HOLDER means that the element may hold one or more other elements.
     *  V_CIRCULAR means the element is circular; this may be significant to
     *  its subelements.  V_TAGGABLE means that a tag may be specified in
     *  the elements .tag field, usually in addition to held elements.
     *  V_FILLER means the element is not to be given a border when it is
     *  held in a piled list.  V_NO_CONSTRAIN and V_FLEX_GRIDS are used by
     *  BgrDrawOneElement.  V_SEP_ROT means that the .data field for this
     *  element specifies a rotation for it separate from the overall
     *  rotation for the symbol.
     *
     *  V_NO_IMAGE means an element has no image, therefore no bounds or
     *  mask.  V_OWN_BOUNDS and V_OWN_MASK means the element does its own
     *  bounding and/or masking; it is otherwise done automatically by
     *  BgrDrawOneElement.
     */

    BgrFlags	gflags;

#define V_HOLDER	(1)
#define V_NO_IMAGE	(1<<1)
#define V_NO_CONSTRAIN	(1<<2)
#define V_CIRCULAR	(1<<3)
#define V_TAGGABLE	(1<<4)
#define V_FILLER	(1<<5)
#define V_OWN_MASK	(1<<7)
#define V_OWN_BOUNDS	(1<<8)
#define V_FLEX_GRIDS	(1<<9)
#define V_SEP_ROT	(1<<10)

    /*  .sflags specifies specific attributes common to some logical
     *  grouping of elements, such as unit sizes.  Possible values:
     *  
     *  Unit Sizes.
     *  UV_TOUCHES means the unit size touches the attached symbol, as
     *  opposed to being slightly offset.
     */

    BgrFlags	sflags;

#define UV_TOUCHES	(1)

    /*  .ratio is the aspect ratio for an element (height / width); it is
     *  used to determine how much of a given space can be filled by an
     *  element (see BgrConstrainDimensions and callers thereof.)
     *
     *  .xdim and .ydim are also used for drawing one or more held piled
     *  elements, provided they are not V_FILLERs; the element is drawn on a
     *  grid of xdim+2, ydim+2, with one unit serving as a border on all
     *  sides.  Thus, a 3x3 piled element will be drawn on a 5x5 grid, and
     *  will take up 60% of the available space.  A given MxN symbol may be
     *  made to take up more space by making it 2Mx2N, 3Mx3N, etc.  .ratio
     *  is always == .ydim/.xdim.
     *
     *  .rprop and .cprop specify the relative sizes of multiple elements
     *  held in a row or column.  They default to 1; increasing or
     *  decreasing them makes one element larger or smaller than the others.
     *  For example, the .cprop field of CompositeWeaponry is 1.5, since it
     *  is taller than most other elements.  This information is empirical,
     *  and cannot be determined from the aspect ratio.
     *
     *  NOTE: a given element .drawer routine does NOT need to take any of
     *  these values into account.  The width and height passed in to the
     *  are already modified based on the available space and aspect ratio.
     */

    float	ratio;
    int		xdim, ydim;
    float	rprop, cprop;

    /*  .e_hold and .t_hold specify the hold space and tag space location
     *  and size if the element is V_HOLDER or V_TAGGABLE, respectively.
     *  Note that held elements are run through BgrLayoutElements, and will
     *  thus be constrained and given borders.  Held text is given directly
     *  to BgrDrawBoxedText, so if a border is desired, it should be
     *  expressed in .t_hold.
     */

    struct BgrHoldInfo	*e_hold;
    struct BgrHoldInfo	*t_hold;

    /*  Most elements center themselves at the requested draw position.  For
     *  those that don't (e.g., CompositeAviation, MarkerFlag), their anchor
     *  must be explicitly specified.  .x_anchor and .y_anchor express this,
     *  in terms of the element's width and height.  Default is 0,0.
     */

    float	x_anchor;
    float	y_anchor;

    /*  To make it easier to draw elements with natural size differences
     *  (e.g., bgrUnspecifiedMine versus just about anything else),
     *  .rel_size specifies the ratio of this element's size to the norm.
     *  This only applies if the element is not held; if held, elements
     *  expand to fit the hold space.  Default is 1.
     */

    float	rel_size;

    /*  Most elements draw themselves using a list of primitives specified
     *  textually in the file data.sym.  Those that do store the primitive
     *  list and count in .prims and .pcount.  The primitives are in terms
     *  of Cartesian grid units from a 0,0 corresponding to the element's
     *  draw anchor; the actual size of a grid unit is given by
     *
     *  	HW / .grids
     *
     *  where HW is the larger of the width or height of the space in which
     *  the element is being drawn.  (Exception: elements for which
     *  V_FLEX_GRIDS is set may have non-square grid boxes.)
     */

    struct BgrTPrim	*prims;
    int			pcount;
    int			grids;

    /*  Element appearance may be specified in terms of additions to another
     *  element's primitives; for those that use this feature, .based_on
     *  contains the code of the other element.
     */

    int		based_on;

    /*  If any of the primitives for this element are conditionally drawn
     *  (see the "draw if" directive in data.sym), .eval_conds holds a
     *  pointer to the routine that determines these conditions.
     */

    void	(*eval_conds)();

} BgrEltInfo;

/*  BgrHoldInfo describes how one symbol element may hold another.  For
**  V_HOLDER elements, these fields describe how the size of held elements
**  is constrained.  .xpos and .ypos are the position of the held elements
**  relative to the center of the holding element, in units of the width and
**  height used to draw the outer part.  .width and .height are the
**  dimensions of the inner part, using the same units.
**
**  Example: an RMedical element may hold another element in the center of
**  its lower right corner; the four parameters are thus .25, -.25, .5, .5.
**  
**  Similarly, the default values for an element which does not constrain
**  the size of held elements are 0,0,1,1.
*/

typedef struct BgrHoldInfo
{
    float	xpos, ypos;
    float	width, height;
} BgrHoldInfo;

/*  BgrTPrim holds a reference to single typed primitive in the array
**  BgrEltInfo.prims.  .type is the type of the primitive, one of the PT_
**  constants listed below.  .flags is OR'ed with the existing draw flags;
**  it specifies things like P_CLOSED and P_FILLED (see below.)  .count is
**  the number of data elements in the primitive; it's used if .data
**  references an array of points.  .cond, if nonzero, is the draw
**  conditional index which specifies whether or not to draw this primitive.
**  See BgrDrawPrimList for further details.
*/

typedef struct BgrTPrim {
    short	type, cond, count;
    BgrFlags	flags;
    ADDRESS	data;
} BgrTPrim;

#define PT_LINE		1
#define PT_SEGMENT	2
#define PT_CIRCLE	3
#define PT_ARC3P	4
#define PT_BOX		5

/*  BgrCompElt describes a compiled symbol element, an intermediate step in
**  the drawing process.  .elt points to the user-specified element.  .info
**  is the BgrEltInfo pointer for .elt->id.  .next is the element that
**  follows this in a held list, or NULL if none.  .child is this element's
**  first held child, or NULL if none are being held.  .parent is this
**  element's holder, or NULL if the element is not being held.
*/

typedef struct BgrCompElt {
    BgrSymbolElt	*elt;
    BgrEltInfo		*info;
    struct BgrCompElt	*next;
    struct BgrCompElt	*child;
    struct BgrCompElt	*parent;
} BgrCompElt;

/*  BgrCompSym completely specifies an error-checked set of compiled
**  elements, arranged in linked child lists for drawing.  .celts contains
**  the compiled elements in the same order given by the user.  .first is a
**  pointer to the outermost holder element.  .cur_holder and .last are used
**  during compilation.  .mobility, .unit_size, and .hq_staff point to the
**  single specified elements of those types, or NULL if none were
**  specified.  .tags is an array referencing the compiled tag elements.
*/

typedef struct
{
    BgrCompElt	celts[bgrMaxSymbolElts+1];
    BgrCompElt	*first, *cur_holder, *last;
    BgrCompElt	*mobility;
    BgrCompElt	*unit_size;
    BgrCompElt	*hq_staff;
    BgrCompElt	*tags;
} BgrCompSym;

/*  Internal information describing each type of polyline libbgr can draw.
*/

typedef struct BgrPolyInfo
{
    /*  .ptype is the same as BgrPolySpec.type.  .name is for debugging and
     *  tabling purposes.
     */

    short	ptype;
    char	*name;

    /*  .flags is an OR of the following attribute flags:
     *
     *  LF_INNER_HEAD and LF_OUTER_HEAD indicate the arrowhead types to be
     *  drawn for axes and directions of advance.
     *
     *  LF_PIPES means the output poly has two (more or less) parallel lines
     *  which appear in place of the supplied points.  LF_END_CAPS means the
     *  endpoints of the line are capped with line segments; if LF_PIPES is
     *  set, these caps will close the pipe ends.  LF_INVIS indicates a pipe
     *  polyline whose lines are invisible, such as a LineOfContact, but for
     *  which the pipe information must be compiled anyway.  LF_SLOTS
     *  indicates a pipe polyline which is patterned with other symbols,
     *  such as a minefield.
     *
     *  LF_END_ELTS_OK means that BgrPolySpec.end_elt may be used to specify
     *  an endpoint element for this poly type.
     *
     *  LF_EXTRAS means that BgrCompPoly.ex must be allocated for this line.
     *
     *  The LF_SHRINK flags are for use by BgrShrinkPoly.
     *  The LF_FAKE flags are for use by BgrMaybeFakePoly.
     */

    int		flags;

#define LF_PIPES		1
#define LF_INNER_HEAD		(1<<1)
#define LF_OUTER_HEAD		(1<<2)
#define LF_BOTH_HEADS		(LF_INNER_HEAD | LF_OUTER_HEAD)
#define LF_END_CAPS		(1<<3)
#define LF_INVIS		(1<<4)
#define LF_SLOTS		(1<<5)
#define LF_SHRINK_START		(1<<6)
#define LF_SHRINK_END		(1<<7)
#define LF_SHRINK_BOTH		(LF_SHRINK_START | LF_SHRINK_END)
#define LF_END_ELTS_OK		(1<<8)
#define LF_EXTRAS		(1<<9)
#define LF_FAKE_START		(1<<10)
#define LF_FAKE_END		(1<<11)
#define LF_FAKE_LINE		(1<<12)
#define LF_FAKE_MASK		(LF_FAKE_START | LF_FAKE_END | LF_FAKE_LINE)

    /*  For pipe polylines which use LF_SLOTS, .slot_elt gives the symbol
     *  element id used to pattern the polyline, e.g., bgrUnspecifiedMine.
     *  .slot_ratio is the size (between 0 and 1) of the pattern symbol
     *  relative to the width of the pipe.
     */

    int		slot_elt;
    float	slot_ratio;

    /*  .shrink_ratio is the percentage (between 0 and 1) of spec.sizes[0]
     *  or .sizes[1] by which to shrink the end of a line for which an
     *  LF_SHRINK flag is set.  Example: the .shrink_ratio for LinearAbatis
     *  is 1, because the abatis notch is spec.sizes[0] units wide.  For
     *  DirectionOfAdvance, it is .5, because the length of the arrowhead
     *  (and thus the amount by which the line must be shrunk) is half the
     *  width of the pipe, which itself is spec.sizes[1].
     */

    float	shrink_ratio;

    /*  For each .sizes[i] in BgrPolySpec that this type of polyline uses,
     *  bit i is set in .size_flags.
     */

    int		size_flags;

} BgrPolyInfo;

/*  Struct returned by BgrMakePoly.  Like other libbgr internals, it is
**  subject to change without notice.
*/

typedef struct BgrCompPoly
{
    /*  .info and .spec are useful private copies of supplied data.
     */

    BgrPolyInfo	*info;
    BgrPolySpec	spec;

    /*  unitsPerPixel is the draw context scaling it was last compiled with.
     *  Polylines need to be recompiled after changes to the scale, because
     *  some of the frill data may be in terms of bitmap coordinates.
     */

    float	unitsPerPixel;

    /*  .p contains the generated spline for a set of input points.  If
     *  info.flags has LF_PIPES, .p1 and .p2 are the pipe sides; they are
     *  otherwise NULL.  .nout is the number of points in .p, .p1, and .p2.
     */

    BgrPoint	*p;
    BgrPoint	*p1, *p2;
    int		nout;

    /*  After compilation of a patterned polyline, .slots contains one
     *  FrillSlot struct for each point on the line where the pattern symbol
     *  is to be drawn.  .n_slots is the count; .na_slots is the number of
     *  structs allocated in .slots.
     */

    struct BgrFrillSlot	*slots;
    int			n_slots, na_slots;

    /*  .ex is is used if info.flags matches LF_EXTRAS; it contains special
     *  data specific to each type of polyline that uses it.
     */

    union BgrPolyExtras *ex;

    /*  If a line has either LF_SHRINK flag, then .p, .p1, .p2, and .nout
     *  above reflect the line after shrinking; .orig_p and .orig_nout
     *  contain the information returned by BgrGenerateCurve.  If a line was
     *  not shrunk, .orig_p == .p and .orig_nout == .nout.
     */

    BgrPoint	*orig_p;
    int		orig_nout;

    /*  .poof is set to TRUE by BgrShrinkPoly, is otherwise FALSE.
     *  .faked is set to TRUE by BgrMaybeFakePoly, is otherwise FALSE.
     */

    Bool	poof;
    Bool	faked;

} BgrCompPoly;

typedef struct BgrFrillSlot
{
    float	x, y;
    float	angle;
} BgrFrillSlot;

/*  Extra compiled-poly data.  .arrowhead is used if info.flags has
**  LF_INNER_HEAD or LF_OUTER_HEAD set.  .caps is used for LF_END_CAPS.
**  .abatis is used for the LinearAbatis polyline.
*/

typedef union BgrPolyExtras
{
    struct {
	BgrPoint	points[7];
	int		count;
	BgrPoint	*notch;
    } arrowhead;

    struct {
	BgrPoint	cap1[2];
	BgrPoint	cap2[2];
    } caps;

    struct {
	BgrPoint	points[3];
    } abatis;
} BgrPolyExtras;

/*  Geometry elements used inside libbgr.  BgrSegment is an analog of
**  XSegment.  BgrCircle is like XArc, but center-based for ease of use.
**  BgrArc3P is used to specify an arc given 3 points.
*/

typedef struct {
    float x1, y1;
    float x2, y2;
} BgrSegment;

typedef BgrSegment BgrBox;

typedef struct {
    float	x, y;
    float	radius;
    float	a1, a2;
} BgrCircle;

typedef struct {
    float x1, y1;
    float x2, y2;
    float x3, y3;
    Bool clockwise;
} BgrArc3P;

/*  A BgrDict is used to map strings to arbitrary pointers using a
**  lexicographically sorted array.
*/

typedef struct {
    char	*key;
    ADDRESS	value;
} BgrDictEntry;

typedef struct {
    char		sorted;
    int			nalloc;
    int			nentries;
    BgrDictEntry	*entries;
} BgrDict;

/*  BgrDataInfo allows mapping of names to simple integer values (e.g.,
**  AvTransport, WpHeavy, etc.) via a BgrDict.  .value is the first field so
**  it matches the BgrEltInfo type.
*/

typedef struct {
    int		value;
    char	*name;
} BgrDataInfo;

/*  As libbgr wanders up and down its call tree during symbol drawing, it
**  passes around a flags parameter which controls certain output
**  attributes.  This field is initialized from BgrDC.flags, and may take on
**  the following additional values.
**
**  P_OUTLINE causes stroking of the outline of a primitive.  P_CLOSED
**  connects the first and last points of a line-based primitive.  P_FILL
**  causes a primitive to be filled with the foreground, if filling is
**  possible.  P_MASK is used to draw masks of primitives; it implies P_FILL
**  and P_OUTLINE.  Not all primitive routines support P_MASK; see prims.c.
**  
**  P_BOUNDS indicates to various drawing routines to extend BgrDC.up, down,
**  left, right by the bounds of their output; P_BOUNDS_ONLY does the same,
**  but inhibits actual device output.  P_ROTATE specifies to the draw
**  primitives that rotation is being done using BgrDC.theta; symbols use
**  this, polylines do not.
**
**  P_HELD means an element is being drawn inside another one.
**  P_CIRCLE_HOLD (in addition to P_HELD) means that the holding element is
**  circular; this matters to some elements, like RAirDefense.
**
**  P_DASHED indicates a primitive which is always drawn with dashed lines.
**  P_DASHABLE indicates one which is drawn dashed only if the user
**  specified TRUE to BgrSetDashed *and* the element is not held by another
**  element with an image.  See the man page for more on this.
*/

#define P_OUTLINE	(1 << 31)
#define P_CLOSED	(1 << 30)
#define P_FILL		(1 << 29)
#define P_BOUNDS	(1 << 28)
#define P_ROTATE	(1 << 27)
#define P_MASK		(1 << 26)

#define P_HELD		(1 << 25)
#define P_CIRCLE_HOLD	(1 << 24)

#define P_BOUNDS_ONLY	(1 << 23)

#define P_DASHABLE	(1 << 22)
#define P_DASHED	(1 << 21)

/*=====  code macros  ======================================================*/

/*  DRAWFNARGS and DRAWFNDECLS specify the parameter list of a
**  BgrEltInfo.drawer; nearly all draw handlers use this format.
**  Subordinate handlers, such as those for aviation, do weird stuff.
*/

#define DRAWFNARGS dc, celt, x, y, width, height, flags
#define DRAWFNDECL \
BgrDC		dc; \
BgrCompElt	*celt; \
float		x, y; \
float		width; \
float		height; \
BgrFlags	flags

/*  COMPFNARGS and COMPFNDECLS specify the parameter list of a
**  BgrEltInfo.compiler.
*/

#define COMPFNARGS cs, celt
#define COMPFNDECL \
    BgrCompSym	*cs; \
    BgrCompElt	*celt

/*  RFASTMASK and RFASTBOUNDS may be used to specify the masking and bounds
**  for an element, if they are simple rectangles.  This saves the effort of
**  doing the bounds hacking in the primitives layer.  Note that RFASTMASK
**  returns if it does a mask, on the assumption that any subsequent drawing
**  or masking is within the image's bounds, and is thus irrelevant.
**
**  CFASTBOUNDS and CFASTMASK do circular bounds and masks.
*/

#ifndef lint

#define RFASTBOUNDS(dc, x1, y1, x2, y2, flags)			\
    STATEMENTS(if (flags & P_BOUNDS)				\
	       {						\
		   BgrAddBoxToBounds(dc, x1, y1, x2, y2, 0.0);	\
		   flags &= ~P_BOUNDS;				\
	       })

#define RFASTMASK(dc, x1, y1, x2, y2, flags)			\
    STATEMENTS(if (flags & P_MASK)				\
	       {						\
		   BgrDoBox(dc, x1, y1, x2, y2, flags);		\
		   return;					\
	       })

#define CFASTBOUNDS(dc, x, y, r, flags)				\
    STATEMENTS(if (flags & P_BOUNDS)				\
	       {						\
		   BgrAddCircleToBounds(dc, x, y, r, 0.0);	\
		   flags &= ~P_BOUNDS;				\
	       })

#define CFASTMASK(dc, x, y, r, flags)				\
    STATEMENTS(if (flags & P_MASK)				\
	       {						\
		   BgrDoCircle(dc, x, y, r, 0.0, 360.0, flags);	\
		   return;					\
	       })

#else
extern void RFASTBOUNDS();
extern void RFASTMASK();
extern void CFASTBOUNDS();
extern void CFASTMASK();
#endif

/*  Non-ANSI C compilers cannot pass pointers to floating-point formal
**  parameters; they end up as (double*) on the other end.  As a workaround,
**  libbgr makes local copies of such parameters when needed.  The SFLOAT
**  macro indicates when this is being done.
*/

#define SFLOAT float

/*  Other useful stuff.
*/

#undef TRUE
#undef FALSE
#define TRUE 1
#define FALSE 0

#define PUBLIC
#define INTERNAL
#define PRIVATE static
#define REG register

#define ROUND(n)	((int) ((n) < 0 ? (n) - .5 : (n) + .5))
#define ARRAYSIZE(a)	(sizeof(a) / sizeof((a)[0]))
#define STATEMENTS(a)	do { a } while (0)

#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#define ABS(a)		((a) < 0 ? (-(a)) : (a))
#define SGN(a)		((a) < 0 ? (-1) : (a) > 0 ? (1) : (0))

#ifndef lint
#define SETMAX(a,b)	((b) > (a) && ((a) = (b), 1))
#define SETMIN(a,b)	((b) < (a) && ((a) = (b), 1))
#else
#define SETMAX(a,b)	((a) = (b))	/*  avoid "null effect" idiocy  */
#define SETMIN(a,b)	((a) = (b))
#endif

#undef min
#undef max
#undef abs

#define SINE_30		.5
#define SINE_40		.643
#define SINE_45		.7071
#define SINE_50		.766
#define SINE_60		.866

#define COSINE_30	.866
#define COSINE_45	.7071
#define COSINE_60	.5

#define SEG(s,px1,py1,px2,py2) \
    ((s).x1 = px1, (s).x2 = px2, (s).y1 = py1, (s).y2 = py2)

/*  Memory allocation.  MALLOC, REALLOC, FREE, and FREE_RTN are used in
**  place of the usual.  If these macros are not already defined (as, say,
**  alternate allocation routines), they will be defined as the default
**  system routines.  Under lint, they are defined to prevent warnings.
**  Always cast the return from MALLOC and REALLOC, even for char*.
*/

#define MALLOC		STDALLOC
#define REALLOC		STDREALLOC
#define FREE		STDDEALLOC

#define SREALLOC(p,n)	((p) ? REALLOC(p, n) : MALLOC(n))
#define DO_REALLOC(p,n) \
    *((char**) &(p)) = (char*) SREALLOC((char*) (p), (n) * sizeof(*(p)))


/*  Work around some lint lossage.
*/

#ifdef lint
#define sprintf (void)
#define qsort (void)
#define exit (void)
#define sleep (void)
#undef isdigit
#define isdigit(c) (c)		/*  needed for MIPS <ctype.h>  */
#undef va_arg
#define va_arg(args,type) (args, (type) NULL)
#endif

#ifdef lint
#define LPTR(expr) (expr, NULL)
#else
#define LPTR(expr) expr
#endif

/* BEGIN_INTERNAL_DECLS */

/*  INTERNAL declarations extracted by excdecls
**  at Tue Nov 26 18:13:06 EST 1991 with command:
**
**  excdecls -p INTERNAL -a libbgr.h composite.c data.c misc.c poly.c
**   prims.c symbol.c tags.c utils.c
**  Don't modify this section by hand if you normally use excdecls.
*/

extern void BgrDrawCompositeAviation ();
extern void BgrDrawCompositeWeaponry ();
extern BgrEltInfo	**BgrPerEltInfo;
extern int		BgrMaxEltId;
extern BgrDict	*BgrEltDict;
extern BgrDict	*BgrDataDict;
extern void BgrInitSymbolData ();
extern void BgrDrawRTextual ();
extern void BgrDrawUserSymbol ();
extern void BgrDrawUnitSize();
extern void BgrDrawHqStaff ();
extern void BgrGetHqAnchor ();
extern void BgrInitPolyData ();
extern void BgrDoLine ();
extern void BgrDoLines ();
extern void BgrDoSegments ();
extern void BgrDoTriangle ();
extern void BgrDoBox ();
extern void BgrDoCircle ();
extern void BgrDoCircles ();
extern void BgrDoArc3P ();
extern void BgrReverseFgBg ();
extern void BgrChangeThickness ();
extern void BgrChangeDashing ();
extern Bool BgrCompileGeneric ();
extern Bool BgrCompileControl ();
extern Bool BgrCompileTag ();
extern Bool BgrCompileMobility ();
extern Bool BgrCompileUnitSize ();
extern Bool BgrCompileHqStaff ();
extern void BgrLayoutElements ();
extern void BgrDrawOneElement ();
extern void BgrQuickDrawElement ();
extern void BgrDrawPrimList ();
extern void BgrDrawTags ();
extern void BgrDrawBoxedText ();
extern BgrDict *BgrMakeDict ();
extern ADDRESS BgrAddDictEntry ();
extern ADDRESS BgrLookup ();
extern void BgrSortDict ();
extern void BgrRotatePoint ();
extern void BgrRotatePoints ();
extern void BgrError ();
extern void BgrAddBoxToBounds ();
extern void BgrAddCircleToBounds ();
extern void BgrAddPointToBounds ();
extern void BgrConstrainDimensions ();
extern void BgrConstrainToHolder ();
extern void BgrSleazeElement ();
extern int *BgrInstallBgrSymbols();

/* END_INTERNAL_DECLS */

#endif  /* ifdef BGR_INTERNALS */

#endif /* USE_X */

#ifdef __cplusplus
}
#endif

#endif  /* _libbgr_h_ */
