\"  $Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/libbgr.man,v 1.1 1999/04/12 18:23:34 ksnively Exp $
.fp 4 R
\"  .Ss "text"
\"  subsection with a blank line after heading
.de Ss
.SS "\\$1"
.rs
.sp
..
\"  .Ks, .Ke
\"  Simple delimited keep.
.de Ks
.br
.di K=
.nr Kk 1
..
.de Ke
.di
.ne \\n(dnu
.nr Ki \\n(.i
.if !\\n(.u .in 0
.K=
.in \\n(Kiu
.nr Kk 0
..
\"  .Cs, .Ce
\"  start and end code block; indented, nofill, fixed-width font
\"  (if available.)  .Cs K means do a keep as well.
.de Cs
.RS
.nf
.sp
.ft 4
.if "\\$1"K" .Ks
..
.de Ce
.if \\n(Kk .Ke
.RE
.fi
.sp
.ns
.ft 1
..
\"  --------------------  begin man page  --------------------
.TH libbgr 3X
.SH NAME
libbgr \- battlefield graphics library
.SH SYNOPSIS
  The CVCC Battlefield Graphics Library is a facility for runtime or offline
generation of a variety of point-based symbols and line graphics from the
Army's ``\fIOperational Terms and Symbols\fP'' manual (FM 101-5-1, October
1985.)  At present, it can output only to X11 Drawables, but since most of
it is layered on a set of draw primitives, support for other devices (e.g.,
SunView screens, PostScript plotters, text files) is not difficult to
implement.
.PP
Libbgr can produce
.IP -
 All text fields, p. 2-2, except direction indicator.
.IP -
 Unit sizes, pp. 2-5 to 2-6, excluding mission command.
.IP -
 Unit roles, pp. 2-7 to 2-13.
.IP -
 Installations, pp. 2-13 to 2-14, 2-16 to 2-19.
.IP -
 Weapons, pp. 2-19 to 2-21, excluding flamethrower.
.IP -
 Vehicles, p. 2-22.
.IP -
 Mobility indicators, p. 2-23.
.IP -
 Aviation symbols, pp. 2-23 to 2-24.
.IP -
 Points of interest, pp. 2-28 to 2-30, excluding airfield.
.IP -
 Generic open and closed polylines, optionally splined.
.IP -
 Axes and directions of attack, p. 2-42.
.IP -
Some point and linear minefields, p. 2-47 to 2-49.
.IP -
 NBC point symbols, pp. 2-58, 2-61.
.SH "GENERATING SYMBOLS"
.Ss "Format"
A symbol is specified to libbgr using an array of
.IR elements, " where an " element
represents some small, non-decomposable part of a symbol, such as a truck
shape, a piece of text, or a battalion indicator.  The C struct representing
an element is
.Cs
.Ks
typedef struct {
    short  id;
    short  flags;
    long   data;
    char   text[24];
} BgrSymbolElt;
.Ke
.Ce
\f4\&.id\fP is the unique identifying code of the element; there are currently
just over a hundred of these.  \f4\&.flags\fP contains limited information on
how to draw this element in this instance.  \f4\&.data\fP contains information
which is specific to a given \f4\&.id\fP; it is usually zero. \f4\&.tag\fP
contains textual information also specific to a given \f4\&.id\fP; it is
usually set to the empty string.
.PP
This array-based format is a trifle clumsy when constructing a deep element
hierarchy (which rarely occurs in practice), but is useful for allowing
symbols to be completely specified at compile-time, and to be transmitted
in protocol packets (a CVCC project requirement.)
.Ss "Element parsing"
Before we look at how elements fit together, the following brief digression
is useful.  As an alternative to filling element structs by hand, the
routine \f4BgrParseElement\fP may be used.  It allows specification of
element fields in textual format, which is an invaluable aid in writing test
programs.  It takes a string of the form
.Cs K
"element_name [ +data ... ] [ /flag ... ] [ 'tag text' ]"
.Ce
\f4element_name\fP is the same as the C identifier for an \f4id\fP.  An
arbitrary number of optional data and flag tokens are summed and placed in
the \f4\&.data\fP and \f4\&.flags\fP fields of the element; these tokens may be
numeric, or any flag/data identifier from libbgr.h.  The tag is optional,
must be the last token, and is delimited by either single or double quotes.
All non-numeric tokens are given without the ``bgr'' prefix.  Example:
.Cs
.Ks
BgrSymbolElt elts[3];

/*  this ...  */

BgrParseElement("UnitRect /Hold", &elts[0]);
BgrParseElement("RAirDefense /Hold 'SP'", &elts[1]);
BgrParseElement("CompositeAviation +AvArmy +AvHelo", &elts[2]);
.Ke

.Ks
/*  ... is the same as this  */

elts[0].id = bgrUnitRect;
elts[0].data = 0;
elts[0].flags = bgrHold;
*elts[0].tag = '\e0';

elts[1].id = bgrRAirDefense;
elts[1].data = 0;
elts[1].flags = bgrHold;
strcpy(elts[1].tag, "SP");

elts[2].id = bgrCompositeAviation;
elts[2].data = bgrAvArmy | bgrAvHelo;
elts[2].flags = 0;
*elts[2].tag = '\e0';
.Ke
.Ce
For the sake of brevity, example element lists in the manual page are
described in the textual format.
.Ss "Element types"
Libbgr elements are divided into several types:
.IP -
\fIInternal elements\fP form a (usually) constant-sized core of a symbol,
such as the boxed portion of a unit and the encircled portion of a logistics
installation.  The symbols shown on pp. 2-7 to 2-19 are all combinations of
internal elements.  Some of these elements may hold other elements, as in
the Artillery Ammunition symbol on page 2-17.
.IP -
\fIExternal elements\fP are attached to the edges of the combined
\fIinternal\fP elements of a symbol.  These are unit sizes (pp. 2-5, 2-6),
mobility indicators (p. 2-23), and text tags (p. 2-73.)
.IP -
\fIControl elements\fP do not produce graphical output, but change the way
the hierarchy of elements is parsed.  There is currently only one control
element, and it is rarely used.
.Ss "Basic symbol construction"
Symbols are constructed by specifying one or more internal elements,
followed by zero or more external elements, interspersed with zero or more
control elements, in an array of\f4 BgrSymbolElt\fP structs.  The internal
elements are drawn first, and the external elements are then attached to the
edges of the resulting bounding box.  A simple element list example, which
produces a generic wheeled unit with a textual tag above it, is
.Cs K
UnitRect
MbWheeled
TopTag1 "Unit A"
.Ce
The positions of external elements never conflict with each other.  The
different tag fields do not overlap, and only one mobility or unit size
indicator will be drawn for a given symbol.
.PP
The internal elements, however, are a different story.  If more than one is
specified, they will, by default, appear piled atop one another.  This is
fairly useless, and almost never occurs in practice.  Generally, the first
internal element specified acts as a \fIholder\fP; this is done by setting
the \f4bgrHold\fP flag in its \f4\&.flags\fP field.  All the following
internal elements will then be constrained to fit inside the holding
element, usually with a border between the holder and the held elements.
Most any internal element with some enclosed empty space in its graphical
image can be used as a holder.  For example, to produce the Division Armored
Artillery Battalion on page 2-76, specify
.Cs K
UnitRect /Hold
RArmor /Hold
RArtillery
UBattalion
.Ce
The specifics of how each element holds others are listed in the element
index below.
.PP
By default, a list of more than one held element will produce a pile in the
same manner as unheld elements.  To make held elements appear in a row or
column with space between them, use \f4bgrRow\fP or \f4bgrColumn\fP instead.
For example, the Airborne Armored Battalion on page 2-76 is constructed with
.Cs K
UnitRect /Column
RArmor
RUSAirborne
UBattalion
.Ce
Note that the \f4bgrHold\fP, \f4bgrRow\fP, and \f4bgrColumn\fP flags have no
effect on external elements.  For hysterical raisins, \f4bgrStack\fP may be
used as a synonym for \f4bgrColumn\fP.
.PP
.Ss "Tricks, kludges, and other features"
The above information suffices to specify the majority of drawable symbols.
This section discusses some of the more arcane tricks libbgr can perform to
satisfy nitpickers.
.PP
Holding an element usually produces a border between the holder and the held
element.  Specifying the \f4bgrFill\fP flag in the held element's
\f4\&.flags\fP field makes it fill as much of the hold space as possible.
The element list
.Cs K
UnitRect /Hold
RTransportation /Fill
.Ce
makes an transport unit whose ``spoked wheel'' element completely fills the
rectangle.  Some elements, like \f4RInfantry\fP, automatically fill their
holder, because it makes no sense for them to have a border.
.PP
Construction of arbitrary element hierarchies is a bit unusual, given the
flat structure of the element list and the effect of the \f4bgrHold\fP,
\f4bgrRow\fP, and \f4bgrColumn\fP flags, which only allow one element in any
held list to contain other elements.  To get around this, use an element
with the control element \f4\&.id bgrPopElement\fP.  This causes the element
scanner to pop out one level of held elements, resuming the current held
list as if the previous element with a \f4/Hold\fP-type flag was not
present.  For example, you could construct a unit with two armor elements,
each containing an artillery element, by specifying
.Cs K
UnitRect /Column
RArmor /Hold
RArtillery
PopElement
RArmor /Hold
RArtillery
.Ce
.Ss "Drawing symbols"
Once you've decided how to construct a symbol using an array of elements,
you give it to the following routine:
.Cs K
void BgrDrawSymbol (dc, elts, nelts)
BgrDC         dc;
BgrSymbolElt  elts[];
int           nelts;
.Ce
which draws it on the screen, or the plotter, or whatever.  The first
parameter, the draw context, specified how to output the symbol.  It is
similar to the Xlib graphics context, and is required because of the number
of possible output configurations.  Your can have as many draw contexts in
your program as you like, although you only need one.  They are created with
the following routine:
.Cs K
BgrDC BgrCreateDC ()
.Ce
and are modified using a variety of macros and routines.  Here are the basic
ones which you can't do without.
.IP "\f4BgrSetSize (BgrDC dc, float size)\fP"
Sets the symbol output size in device units, e.g., screen pixels.  The
largest dimension of the grouping of internal symbol elements will not
exceed this value (but see the \f4bgrComposite*\fP symbols in the index.)
It does not apply to the external elements.  The default is 50.
.IP "\f4BgrSetOutput (BgrDC dc, int output_flag)\fP"
.IP "\f4BgrClearOutput (BgrDC dc, int output_flag)\fP"
These macros set and clear the flags which determine the types of device and
informational output libbgr will produce.  Currently supported output flags
are:
.RS
.IP \f4bgrXOutput\fP
Output to the Drawable specified by \f4BgrSetXOutputInfo\fP.
.RE
.IP "\f4BgrSetXOutputInfo (dc, Display *disp, Window win)\fP"
.IP "\f4BgrSetXGC (dc, GC gc)\fP"
These set the display and window or GC for X11 output.
.IP "\f4BgrSetXMappingAnchor (dc, float x, float y)\fP"
.IP "\f4BgrSetXOutputUnits (dc, float units_pp)\fP"
These define a simple cartesian world coordinate system on top of the X11
Drawable, which is required by the libbgr polyline routines, and also used
for symbol positioning for the sake of consistency.  \f4x\fP and \f4y\fP
specify the world coordinate corresponding to (0,0) in the bitmap;
\f4units_pp\fP specifies how many world coordinates map to a single pixel.
The units-per-pixel mapping does not affect symbol size.
.IP "\f4BgrSetPosition (dc, float xp, float yp)\fP"
Sets the symbol output position, in world coordinates.
.PP
The following output macros and flags are for more esoteric output control,
and may not be necessary to the casual user.  They exist primarily because
of the individual requirements of the various projects that have funded
libbgr.
.IP "\f4BgrSetRotation (BgrDC dc, float angle)\fP"
Sets the symbol's rotation about its anchor, in clockwise degrees.  Unless
otherwise specified, a symbol's anchor is the center of its outermost holder
element.  The default is zero.
.IP "\f4BgrSetThickness (BgrDC dc, float thickness)\fP"
Sets the thickness of lines drawn.  For pixel-based device, this value will
be truncated to an integer.  For efficient X11 output, you may wish to set
this to zero.  This may be reset as many times as desired without adversely
affect X11 requests; it will be changed in the draw context's GC only if the
GC's value doesn't match.  The default is 1.
.IP "\f4BgrSetTagSize (BgrDC dc, float size)\fP"
Sets the size of external text tag elements (does not apply to the
\f4RTextual\fP element, which is internal.)  The height (point size) of tags
will be made as close as possible to this value without exceeding it.  On
devices where tags are printed using raster fonts, a close approximation
will be made based on loaded fonts.  If set to zero, libbgr will choose a
text size which looks best based on the symbol size.  The default is zero.
.IP "\f4BgrSetTagBorder (BgrDC dc, float border)\fP"
Sets the amount of space (in output device units) between a symbol's inner
elements and any external tags.  The default is 5.
.IP "\f4BgrSetPolyMask (BgrDC dc, float border)\fP"
Normally, only the device-specific foreground color specified in the draw
context is used.  If masking is on, a masking polygon is drawn and filled in
the backround color; it is guaranteed to cover at least the surface area of
all the symbol elements, and also produces a border of the specified size
around the edges.  The default is zero (no poly masking.)
.IP "\f4BgrSetLineMask (BgrDC dc, float border)\fP"
Like poly masking except gives each drawn line a border of the specified
size by temporarily increasing the device-specific line thickness.  This is
_very_ slow under X11.  The default is zero (no line masking.)
.IP "\f4BgrSetUnitSizeRatio\fP"
Sets the height of unit size elements as a ratio of the overall symbol size.
The default is 1/6; anything larger than 1/4 will probably be too much.
.IP "\f4BgrSetDashed\fP"
Causes certain elements to be output with some of their lines dashed,
indicating an unconfirmed battlefield graphic.  This applies primarily to
holder elements like \f4UnitRect\fP and \f4RoundPoint\fP, and does not
affect elements which always use dashed lines (e.g.,
\f4PlannedDemolition\fP.)  In a /Held element hierarchy, only the outermost
element will be affected.
.IP "\f4bgrXExtOutput\fP"
This flag, set with \f4BgrSetOutput\fP, causes libbgr to write extents
information to the draw context during the draw operation, which the caller
can then retrieve.  This information specifies the bounding boxes of the
internal and external elements, in the following format:
.Cs K
typedef struct {
    short up, down;
    short left, right;
    short xoff, yoff;
} BgrExtents;
.Ce
\f4up\fP, \f4down\fP, \f4left\fP, and \f4right\fP give the symbol's extents
in output units from the draw position.  \f4xoff\fP and \f4yoff\fP are zero
unless a specially anchored symbol (e.g., \f4HqStaff\fP, below) was drawn
which caused the symbol anchor to not be based on the position of the
internal elements.
.SH "SYMBOL ELEMENT INDEX"
\"  .Eh
\"  Start a heading for a group of symbol element information.
.de Eh
.PP
..
\"  .El name
\"  Begin describing the named element.
.de El
.IP "\\$1"
..
The following lists libbgr's element names, in textual format.  Most of them
have no associated \f4\&.data\fP or \f4\&.tag\fP; those that do generally
have them to satisfy odd user requirements.
\" SYMBOL INDEX GOES HERE
.SH SEE ALSO
bgrex(1)
.SH AUTHOR
Chris Ross <cjross@bbn.com>
.SH BUGS
All filler elements and some other elements will almost certainly not look
right when held by elements whose hold space is not rougly square or
circular.  Try holding \f4RInfantry\fP with \f4Diamond \fP to get an idea of
the problem.
.PP
Most irregularly shaped non-masked symbols rotated to non-right angles will
generate bounds that are a trifle larger than the actual symbol geometry.
.PP
Wide polylines can't make sharp turns without buckling.  Use several
closely spaced points to make a sharp turn.  A better algorithm (say, from
Xlib) should be used.
.PP
Any polyline decorated with symbols (UnspecifiedMineField, LineOfContact,
etc.) is drawn by determining how many symbols of the width of the polyline
pipe will fit in the length of each segment.  This tends to produce
unsatisfactory output if a splining factor greater than 1 is used.
.PP
Drawing a mobility element and a headquarters staff on the same symbol will
not work; the staff will start at the base of the mobility element, not the
base of the inner elements.
.PP
An apparently nondeterministic bug in Sun X11/NeWS causes arc endpoints to
sometimes be drawn one pixel away from the specified location, distorting
the appearance of symbols where arc endpoints and line endpoints must meet.
.PP
Because of their highly limited uses, most of the draw context output
controls are incomprehensible without examples.  Fortunately, \fIbgrex\fP(1)
and its associated test files provide a wealth of them.
