/*
**  poly.c
**
**  Author:		Christopher Ross
**  Creation date:	10-Mar-1991
**  Version author:	Christopher Ross
**  Version date:	02-Feb-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/poly.c,v 1.1 1999/04/12 18:23:39 ksnively Exp $";
#endif

#include <stdio.h>
#include <math.h>
#include <stdext.h> /*common/include/global*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

PRIVATE void
    BgrAddLineBounds(),
    BgrCompileAbatis(),
    BgrCompileArrowheads(),
    BgrCompileCaps(),
    BgrCompilePipe(),
    BgrCompilePoly(),
    BgrCompileSlots(),
    BgrDrawCaps(),
    BgrDrawArrowheads(),
    BgrDrawEndSymbols(),
    BgrDrawSlots(),
    BgrMaybeFakePoly(),
    BgrShrinkPoly();

/*  Determine polyline sizes relative to cpoly.  WORLD_SIZE indicates a size
**  in world coordinates is needed, so convert from pixels if the polyline
**  frills are fixed-size.  PIXEL_SIZE indicates a size in pixel, so do the
**  opposite conversion for an unfixed polyline.
*/

#define WORLD_SIZE(cpoly, size) \
    ((cpoly)->spec.fixed_size ? (size) * (cpoly)->unitsPerPixel : (size))
#define PIXEL_SIZE(cpoly, size) \
    ((cpoly)->spec.fixed_size ? (size) : (size) / (cpoly)->unitsPerPixel)

/*  Polyline information by type.  Fortunately, this table isn't
**  sufficiently complex to merit another script like bgrsym.
*/

static int		MaxPolyType;
static BgrPolyInfo	**PolyInfo;
static BgrPolyInfo	RawPolyInfo[] =
{
    {
	bgrGenericLine,				/*  .ptype		*/
	"GenericLine",				/*  .name		*/
	LF_END_ELTS_OK,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	0					/*  .size_flags		*/
	},
    {
	bgrDirectionOfAdvance,			/*  .ptype		*/
	"DirectionOfAdvance",			/*  .name		*/
	LF_EXTRAS | LF_INNER_HEAD | LF_SHRINK_END | LF_FAKE_END,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0 | 1<<1)				/*  .size_flags		*/
	},
    {
	bgrMainDirectionOfAdvance,		/*  .ptype		*/
	"MainDirectionOfAdvance",		/*  .name		*/
	LF_EXTRAS | LF_BOTH_HEADS | LF_SHRINK_END | LF_FAKE_END,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.75,					/*  .shrink_ratio	*/
	(1<<0 | 1<<1)				/*  .size_flags		*/
	},
    {
	bgrAxisOfAdvance,			/*  .ptype		*/
	"AxisOfAdvance",			/*  .name		*/
	LF_EXTRAS | LF_OUTER_HEAD | LF_PIPES | LF_SHRINK_END | LF_FAKE_END,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.75,					/*  .shrink_ratio	*/
	(1<<0) | (1<<1)				/*  .size_flags		*/
	},
    {
	bgrMainAxisOfAdvance,			/*  .ptype		*/
	"MainAxisOfAdvance",			/*  .name		*/
	LF_EXTRAS | LF_BOTH_HEADS | LF_PIPES | LF_SHRINK_END | LF_FAKE_END,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.75,					/*  .shrink_ratio	*/
	(1<<0) | (1<<1)				/*  .size_flags		*/
	},
    {
	bgrLinearTarget,			/*  .ptype		*/
	"LinearTarget",				/*  .name		*/
	LF_EXTRAS | LF_END_CAPS | LF_FAKE_LINE,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrRectangularTarget,			/*  .ptype		*/
	"RectangularTarget",			/*  .name		*/
	LF_EXTRAS | LF_PIPES | LF_END_CAPS | LF_FAKE_LINE,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrUnspecifiedMineField,		/*  .ptype		*/
	"UnspecifiedMineField",			/*  .name		*/
	LF_EXTRAS | LF_SLOTS | LF_PIPES | LF_END_CAPS | LF_FAKE_LINE,
	bgrUnspecifiedMine,			/*  .slot_elt		*/
	.5,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrAntitankMineField,			/*  .ptype		*/
	"AntitankMineField",			/*  .name		*/
	LF_EXTRAS | LF_PIPES | LF_SLOTS | LF_END_CAPS | LF_FAKE_LINE,
	bgrAntitankMine,			/*  .slot_elt		*/
	.5,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrAntipersonnelMineField,		/*  .ptype		*/
	"AntipersonnelMineField",		/*  .name		*/
	LF_EXTRAS | LF_PIPES | LF_SLOTS | LF_END_CAPS | LF_FAKE_LINE,
	bgrAntipersonnelMine,			/*  .slot_elt		*/
	.8,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrTankDitch,				/*  .ptype		*/
	"TankDitch",				/*  .name		*/
	LF_EXTRAS | LF_SLOTS | LF_PIPES | LF_END_CAPS | LF_FAKE_LINE,
	bgrTankDitchPoint,			/*  .slot_elt		*/
	.5,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrFLOT,				/*  .ptype		*/
	"FLOT",					/*  .name		*/
	LF_EXTRAS | LF_END_ELTS_OK | LF_SLOTS | LF_PIPES | LF_INVIS | LF_FAKE_LINE,
	bgrFLOTPoint,				/*  .slot_elt		*/
	.8,					/*  .slot_ratio		*/
	.5,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	},
    {
	bgrLinearAbatis,			/*  .ptype		*/
	"LinearAbatis",				/*  .name		*/
	LF_EXTRAS | LF_SHRINK_START | LF_FAKE_START,
	0,					/*  .slot_elt		*/
	0,					/*  .slot_ratio		*/
	1,					/*  .shrink_ratio	*/
	(1<<0)					/*  .size_flags		*/
	}
};

/*  BgrInitPolyData
**
**  Yet another init routine.
*/

INTERNAL void BgrInitPolyData ()
{
    int		i, max = 0;
    unsigned	size;

    /*  Arrange for PolyInfo to be indexable by polyline type constants,
     *  based on the information given in RawPolyInfo.
     */

    for ( i = 0; i < ARRAYSIZE(RawPolyInfo); i++ )
	max = MAX(max, RawPolyInfo[i].ptype);

    size = (max + 1) * sizeof(*PolyInfo);
    PolyInfo = (BgrPolyInfo**) MALLOC(size);
			   
    for ( i = 0; i < max; i++ )
	PolyInfo[i] = NULL;
    for ( i = 0; i < ARRAYSIZE(RawPolyInfo); i++ )
	PolyInfo[RawPolyInfo[i].ptype] = &RawPolyInfo[i];

    MaxPolyType = max;
}

/*  BgrMakePoly
**
**  Compile a polyline with the specified points and attributes.  Returns
**  compiled polyline data on success, NULL on failure.  The generated
**  polyline may be passed to BgrDrawPoly, or freed using BgrFreePoly.
*/

PUBLIC BgrCompPoly *BgrMakePoly (pts, npts, spec)
BgrPoint	*pts;
int		npts;
BgrPolySpec	*spec;
{
    BgrPolySpec	vspec;		/*  passed spec, after validation  */
    BgrCompPoly	*cpoly;
    BgrPolyInfo	*info;
    int		i, nalloc;
    char	*me = "BgrMakePoly";

    /*  Error-check points and attributes.
     */

    if (npts <= 0)
    {
	BgrError("%s: no points specified in polyline", me);
	return(NULL);
    }

    if (spec->type < 0 || spec->type > MaxPolyType
	|| !(info = PolyInfo[spec->type]))
    {
	BgrError("%s: unknown poly type: %d", me, spec->type);
	return(NULL);
    }

    vspec = *spec;

    if (spec->splining < 1)
    {
	BgrError("%s: bad poly output splining (%d); using 1",
		 me, spec->splining);
	vspec.splining = 1;
    }

    if (spec->end_elt != 0)
    {
	if (!(info->flags & LF_END_ELTS_OK))
	{
	    BgrError("%s: poly type %s does not support end elements",
		     me, info->name);
	    vspec.end_elt = 0;
	}
	else if (spec->end_elt < 1 || spec->end_elt > BgrMaxEltId
		 || !BgrPerEltInfo[spec->end_elt])
	{
	    BgrError("%s: unknown end element id (%d)", me, spec->end_elt);
	    vspec.end_elt = 0;
	}
    }

    /*  If this poly type uses any elements from the .sizes array, the sizes
     *  must be nonzero.  Line end size defaults to line start size if not
     *  explicitly specified.
     */

    for ( i = 0; i < ARRAYSIZE(spec->sizes); i++ )
	if ((info->size_flags & (1<<i)) && spec->sizes[i] == 0)
	{
	    BgrError("%s: spec->sizes[%d] is <= 0", me, i);
	    return(NULL);
	}

    if (!(info->size_flags & (1<<1)))
	vspec.sizes[1] = vspec.sizes[0];

    /*  Allocate compiled-poly struct and determine required space.  Always
     *  malloc one more point than necessary so BgrMaybeFakePoly is assured
     *  of having space for two points.
     */

    cpoly = (BgrCompPoly*) MALLOC(sizeof(*cpoly));
    cpoly->info = info;
    cpoly->spec = vspec;
    cpoly->ex = NULL;
    cpoly->poof = FALSE;
    cpoly->unitsPerPixel = 0;

    nalloc = 1 + ((spec->closed ? npts : npts-1) * spec->splining + 1);
    cpoly->p = (BgrPoint*) MALLOC(nalloc * sizeof(BgrPoint));

    if (cpoly->info->flags & LF_PIPES)
    {
	cpoly->p1 = (BgrPoint*) MALLOC(nalloc * sizeof(BgrPoint));
	cpoly->p2 = (BgrPoint*) MALLOC(nalloc * sizeof(BgrPoint));
    }
    else
	cpoly->p1 = cpoly->p2 = NULL;

    if (cpoly->info->flags & LF_EXTRAS)
	cpoly->ex = (BgrPolyExtras*) MALLOC(sizeof(*cpoly->ex));
    else
	cpoly->ex = NULL;

    /*  We don't know how many of these we'll need; BgrCompileSlots will
     *  MALLOC this field if appropriate.
     */

    cpoly->slots = NULL;
    cpoly->n_slots = cpoly->na_slots = 0;

    /*  Generate the curve and accept the return value of the new output
     *  point count, because the curve generation may have eliminated
     *  adjacent points.
     */

    if (npts > 1)
	cpoly->nout = BgrGenerateCurve(pts, npts, cpoly->p,
				       vspec.splining, vspec.closed);
    else
    {
	*cpoly->p = pts[0];
	cpoly->nout = 1;
    }

    /*  If we'll be shrinking the line ends during compilation, we need to
     *  make a copy of the original line.
     */

    if ((cpoly->info->flags & LF_SHRINK_BOTH) ||
	cpoly->spec.end_elt != 0)
    {
	int nbytes = cpoly->nout * sizeof(BgrPoint);
	cpoly->orig_p = (BgrPoint*) MALLOC((unsigned) nbytes);
	bcopy((char*) cpoly->p, (char*) cpoly->orig_p, nbytes);
    }
    else
	cpoly->orig_p = cpoly->p;
    cpoly->orig_nout = cpoly->nout;

    return(cpoly);
}

/*  BgrBoundPoly
**
**  Writes to *bounds the bounds all the geometry in the specified compiled
**  polyline.  A draw context is required to specify the scaling, because
**  some poly bounds information is based on pixel coordinates.
*/

PUBLIC void BgrBoundPoly (cpoly, bounds, dc)
BgrCompPoly	*cpoly;
BgrPolyBounds	*bounds;
BgrDC		dc;
{
    BgrCompilePoly(dc, cpoly);
    bounds->xmin = bounds->ymin = 1.0E+36;
    bounds->xmax = bounds->ymax = -1.0E+36;

    if (!cpoly->faked || (cpoly->info->flags & LF_FAKE_MASK) == LF_FAKE_LINE)
    {
	if (cpoly->info->flags & LF_PIPES)
	{
	    BgrAddLineBounds(cpoly->p1, cpoly->nout, bounds);
	    BgrAddLineBounds(cpoly->p2, cpoly->nout, bounds);
	}
	else
	    BgrAddLineBounds(cpoly->p, cpoly->nout, bounds);
    }

    if (cpoly->info->ptype == bgrLinearAbatis)
	BgrAddLineBounds(cpoly->ex->abatis.points, 3, bounds);

    else if (cpoly->info->flags & LF_END_CAPS)
    {
	BgrAddLineBounds(cpoly->ex->caps.cap1, 2, bounds);
	BgrAddLineBounds(cpoly->ex->caps.cap2, 2, bounds);
    }

    else if (cpoly->info->flags & LF_BOTH_HEADS)
	BgrAddLineBounds(cpoly->ex->arrowhead.points,
			 cpoly->ex->arrowhead.count, bounds);

    /*  If there are end elements, add the bounds of the boxes which will
     *  enclose them.
     */

    if (cpoly->spec.end_elt != 0)
    {
	BgrPoint	*p, box[4];
	double		hwidth = WORLD_SIZE(cpoly, cpoly->spec.sizes[2] / 2);

	p = cpoly->orig_p;
	box[0].x = box[3].x = p->x - hwidth;
	box[1].x = box[2].x = p->x + hwidth;
	box[0].y = box[1].y = p->y + hwidth;
	box[2].y = box[3].y = p->y - hwidth;
	BgrAddLineBounds(box, 4, bounds);

	p = &cpoly->orig_p[cpoly->orig_nout-1];
	box[0].x = box[3].x = p->x - hwidth;
	box[1].x = box[2].x = p->x + hwidth;
	box[0].y = box[1].y = p->y + hwidth;
	box[2].y = box[3].y = p->y - hwidth;
	BgrAddLineBounds(box, 4, bounds);
    }
}

PRIVATE void BgrAddLineBounds (p, n, bounds)
REG BgrPoint		*p;
int			n;
REG BgrPolyBounds	*bounds;
{
    for ( ; n-- > 0; ++p )
    {
	SETMIN(bounds->xmin, p->x);
	SETMAX(bounds->xmax, p->x);
	SETMIN(bounds->ymin, p->y);
	SETMAX(bounds->ymax, p->y);
    }
}

/*  BgrCompilePoly
**
**  Internal routine for invoking the compiler for a given poly type; runs
**  the compilation if the poly has not been compiled for the current DC
**  transformation.
*/

PRIVATE void BgrCompilePoly (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    if (!dc)
	BgrError("libbgr: can't draw/bound poly; no draw context specified");

    else if (cpoly->unitsPerPixel != dc->x11.unitsPerPixel)
    {
	cpoly->unitsPerPixel = dc->x11.unitsPerPixel;
	cpoly->faked = FALSE;
	cpoly->poof = FALSE;

	if ((cpoly->info->flags & LF_SHRINK_BOTH) ||
	    cpoly->spec.end_elt != 0)
	{
	    BgrShrinkPoly(dc, cpoly);
	}
	else
	{
	    cpoly->orig_p = cpoly->p;
	    cpoly->orig_nout = cpoly->nout;
	}

	if (cpoly->info->flags & LF_FAKE_MASK)
	    BgrMaybeFakePoly(cpoly);

	if (cpoly->info->flags & LF_PIPES)
	    BgrCompilePipe(dc, cpoly);

	if (cpoly->info->flags & LF_SLOTS)
	    BgrCompileSlots(dc, cpoly);

	if (cpoly->info->ptype == bgrLinearAbatis)
	    BgrCompileAbatis(dc, cpoly);
	else if (cpoly->info->flags & LF_BOTH_HEADS)
	    BgrCompileArrowheads(dc, cpoly);
	else if (cpoly->info->flags & LF_END_CAPS)
	    BgrCompileCaps(dc, cpoly);
    }
}

/*  BgrDrawPoly
**
**  Draws a compiled polyline generated with BgrMakePoly; uses the mapping
**  transforms defined in the draw context.  Returns 1 on success, 0 on
**  failure (bad draw context contents.)
*/

PUBLIC void BgrDrawPoly (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    BgrFlags	flags = dc->flags | P_OUTLINE;

    if (dc->flags & bgrXOutput &&
	!(dc->x11.disp && dc->x11.win && dc->x11.gc))
    {
	BgrError("draw context is missing X output information");
	return;
    }
    if (dc->dashed)
	flags |= P_DASHED;

    BgrCompilePoly(dc, cpoly);
    BgrChangeThickness(dc, dc->thickness);

    if (!(cpoly->info->flags & LF_INVIS) &&
	(!cpoly->faked || (cpoly->info->flags & LF_FAKE_MASK) == LF_FAKE_LINE))
    {
	if (cpoly->info->flags & LF_PIPES)
	{
	    BgrDoLines(dc, cpoly->p1, cpoly->nout, flags);
	    BgrDoLines(dc, cpoly->p2, cpoly->nout, flags);
	}
	else
	    BgrDoLines(dc, cpoly->p, cpoly->nout,
		       flags | (dc->poly_fill ? P_FILL : 0));
    }

    if (cpoly->info->flags & LF_SLOTS)
	BgrDrawSlots(dc, cpoly, flags);

    if (cpoly->info->ptype == bgrLinearAbatis)
	BgrDoLines(dc, cpoly->ex->abatis.points, 3, flags);

    else if (cpoly->info->flags & LF_END_CAPS)
	BgrDrawCaps(dc, cpoly, flags);

    else if (cpoly->info->flags & LF_BOTH_HEADS)
	BgrDrawArrowheads(dc, cpoly, flags);

    if (cpoly->spec.end_elt != 0)
	BgrDrawEndSymbols(dc, cpoly, flags);
}

/*  BgrFreePoly
**
**  Frees polyline information returned by BgrMakePoly.
*/

PUBLIC void BgrFreePoly (cpoly)
BgrCompPoly *cpoly;
{
    FREE((char*) cpoly->p);
    if (cpoly->p1)
	FREE((char*) cpoly->p1);
    if (cpoly->p2)
	FREE((char*) cpoly->p2);
    if (cpoly->ex)
	FREE((char*) cpoly->ex);
    if (cpoly->slots)
	FREE((char*) cpoly->slots);
    if (cpoly->orig_p && cpoly->orig_p != cpoly->p)
	FREE((char*) cpoly->orig_p);
    FREE((char*) cpoly);
}

PRIVATE void BgrCompileArrowheads (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    /*  Tip of arrowhead will be at end of line, oriented to the slope of
     *  the last segment.  If the line was faked, the end hasn't been shrunk
     *  for us, so do it ourselves.
     */

    BgrPoint	*end = &cpoly->p[cpoly->nout-1], alt_end;
    double	dx = end->x - end[-1].x;
    double	dy = end->y - end[-1].y;
    double	dist = WORLD_SIZE(cpoly, cpoly->spec.sizes[1] / 2);

    int		*countp;
    BgrPoint	*headp;
    BgrPoint	inner_left_barb, outer_left_barb;
    BgrPoint	inner_right_barb, outer_right_barb;
    BgrPoint	inner_tip, outer_tip;

    if (cpoly->faked)
    {
	BgrGetNearbyPoints(end, dx, dy, &alt_end, (BgrPoint*) NULL,
			   (WORLD_SIZE(cpoly, cpoly->spec.sizes[1])
			    * cpoly->info->shrink_ratio), FALSE);
	end = &alt_end;
    }

    /*  The arrowhead is one contiguous point list for all poly types;
     *  *countp keeps count as we build it.  The inner barbs are required
     *  for outer arrowheads as well, because they are where the outer barbs
     *  extend outward from.
     */

    headp = cpoly->ex->arrowhead.points;
    countp = &cpoly->ex->arrowhead.count;
    *countp = 0;

    BgrGetNearbyPoints(end, dx, dy, &inner_left_barb,
		       &inner_right_barb, dist, TRUE);

    if (cpoly->info->flags & LF_INNER_HEAD)
	BgrGetNearbyPoints(end, dx, dy, (BgrPoint*) NULL,
			   &inner_tip, dist, FALSE);

    if (cpoly->info->flags & LF_OUTER_HEAD)
    {
	BgrGetNearbyPoints(end, dx, dy, &outer_left_barb,
			   &outer_right_barb, dist * 1.5, TRUE);
	BgrGetNearbyPoints(end, dx, dy, (BgrPoint*) NULL,
			   &outer_tip, dist * 1.5, FALSE);
    }

    /*  This logic may seem confusing.  Try thinking about each case
     *  separately -- inner head only, outer head only, both.
     */

    if (cpoly->info->flags & LF_INNER_HEAD)
    {
	headp[0] = inner_left_barb;
	headp[1] = inner_tip;
	headp[2] = inner_right_barb;
	cpoly->ex->arrowhead.notch = headp+1;
	*countp += 3;
	headp += 3;
    }

    else if (cpoly->info->flags & LF_OUTER_HEAD)
    {
	headp[0] = inner_right_barb;
	++*countp;
	++headp;
    }

    if (cpoly->info->flags & LF_OUTER_HEAD)
    {
	headp[0] = outer_right_barb;
	headp[1] = outer_tip;
	headp[2] = outer_left_barb;
	headp[3] = inner_left_barb;
	*countp += 4;
	headp += 4;
    }
}

/*  BgrDrawArrowheads
**
**  Draw compiled arrowhead frills.  If the arrow is not a pipe, extend the
**  main line to touch the notch of the arrow, which is the endpoint of the
**  original unshrunk line.
*/

PRIVATE void BgrDrawArrowheads (dc, cpoly, flags)
BgrDC		dc;
BgrCompPoly	*cpoly;
BgrFlags	flags;
{
    BgrDoLines(dc, cpoly->ex->arrowhead.points,
	       cpoly->ex->arrowhead.count, flags);

    if (!cpoly->faked
	&& !(cpoly->info->flags & LF_PIPES)
	&& (cpoly->info->flags & LF_INNER_HEAD))
    {
	BgrPoint *end = &cpoly->p[cpoly->nout-1];
	BgrPoint *notch = cpoly->ex->arrowhead.notch;
	BgrDoLine(dc, end->x, end->y, notch->x, notch->y, flags);
    }
}

PRIVATE void BgrCompileCaps (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    BgrPoint	*end;
    double	dx, dy;
    double	dist;

    dist = WORLD_SIZE(cpoly, cpoly->spec.sizes[0] / 2);
    end = cpoly->p;
    dx = end[1].x - end->x;
    dy = end[1].y - end->y;

    BgrGetNearbyPoints(end, dx, dy, &cpoly->ex->caps.cap1[0],
		       &cpoly->ex->caps.cap1[1], dist, TRUE);

    dist = WORLD_SIZE(cpoly, cpoly->spec.sizes[1] / 2);
    end = &cpoly->p[cpoly->nout-1];
    dx = end->x - end[-1].x;
    dy = end->y - end[-1].y;

    BgrGetNearbyPoints(end, dx, dy, &cpoly->ex->caps.cap2[0],
		       &cpoly->ex->caps.cap2[1], dist, TRUE);
}

PRIVATE void BgrDrawCaps (dc, cpoly, flags)
BgrDC		dc;
BgrCompPoly	*cpoly;
BgrFlags	flags;
{
    BgrPoint *cap;
    cap = cpoly->ex->caps.cap1;
    BgrDoLine(dc, cap[0].x, cap[0].y, cap[1].x, cap[1].y, flags);
    cap = cpoly->ex->caps.cap2;
    BgrDoLine(dc, cap[0].x, cap[0].y, cap[1].x, cap[1].y, flags);
}

PRIVATE void BgrCompilePipe (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    /*  We will make p1 the left side of the pipe and p2 the right side, as
     *  if the pipe were drawn upward from its tail.  curwidth is the
     *  current width of the pipe at each point, as it changes in increments
     *  of wdiff from spec.sizes[0] (tail) to spec.sizes[1] (head.)  To
     *  determine the slope at each point on p, take the average of the
     *  surrounding segments; for the first and last points, use the
     *  adjacent segment.
     */

    BgrPoint	*p = cpoly->p, *p1 = cpoly->p1, *p2 = cpoly->p2, slope;
    float	*sizes = cpoly->spec.sizes;
    double	wdiff, curwidth;
    int		i;

    wdiff = WORLD_SIZE(cpoly, (sizes[1] - sizes[0]) / (cpoly->nout - 1));
    curwidth = WORLD_SIZE(cpoly, sizes[0]);

    for ( i = 0; i < cpoly->nout; ++i, ++p, ++p1, ++p2, curwidth += wdiff )
    {
	if (i == 0)
	{
	    slope.x = p[1].x - p->x;
	    slope.y = p[1].y - p->y;
	}
	else if (i == cpoly->nout - 1)
	{
	    slope.x = p->x - p[-1].x;
	    slope.y = p->y - p[-1].y;
	}
	else
	{
	    slope.x = ((p->x - p[-1].x) + (p[1].x - p->x)) / 2;
	    slope.y = ((p->y - p[-1].y) + (p[1].y - p->y)) / 2;
	}

	BgrGetNearbyPoints(p, slope.x, slope.y,
			   p1, p2, curwidth / 2, TRUE);
    }
}

PRIVATE void BgrCompileSlots (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    /*  Iterate over each section of pipe.  The pipe width is mapped to
     *  device output units, which is what we draw symbols with.
     */

    BgrPoint	*p = cpoly->p;
    double	seg_width = PIXEL_SIZE(cpoly, cpoly->spec.sizes[0]);
    double	seg_length, seg_angle;
    int		i;

    cpoly->n_slots = 0;
    if (cpoly->poof)
	return;

    for ( i = 0; i < cpoly->nout-1; ++i, ++p )
    {
	BgrFrillSlot	*slot;
	int		n_seg_slots;
	double		slot_x, slot_y;
	double		dx, dy;

	/*  Each pipe segment will hold as many slots as its length divided
	 *  by its width.  Start at the first point, advance half the space
	 *  available to each slot, place the slot, and advance down the
	 *  pipe by increments of space available.
	 */

	dx = p[1].x - p->x;
	dy = p[1].y - p->y;
	seg_length = sqrt(dx*dx + dy*dy) / cpoly->unitsPerPixel;
	if ((n_seg_slots = seg_length / seg_width) == 0)
	    continue;

	seg_angle = NS_ATAN2(dy, dx) / PI * 180;
	if (seg_angle > 0)
	    seg_angle = 360 - seg_angle;
	else
	    seg_angle = -seg_angle;

	dx = (p[1].x - p->x) / n_seg_slots;
	dy = (p[1].y - p->y) / n_seg_slots;
	slot_x = p->x + dx / 2;
	slot_y = p->y + dy / 2;
	
	while (n_seg_slots-- > 0)
	{
	    if (cpoly->n_slots+1 > cpoly->na_slots)
	    {
		cpoly->na_slots += 10;
		DO_REALLOC(cpoly->slots, cpoly->na_slots);
	    }

	    slot = &cpoly->slots[cpoly->n_slots++];
	    slot->x = slot_x;
	    slot->y = slot_y;
	    slot->angle = seg_angle;

	    slot_x += dx;
	    slot_y += dy;
	}
    }
}

PRIVATE void BgrDrawSlots (dc, cpoly, flags)
BgrDC		dc;
BgrCompPoly	*cpoly;
BgrFlags	flags;
{
    _BgrDC	sym_dc;
    int		i;

    if (cpoly->n_slots == 0)
	return;

    sym_dc = *dc;
    BgrSetSize(&sym_dc, (PIXEL_SIZE(cpoly, cpoly->spec.sizes[0])
			 * cpoly->info->slot_ratio));

    for ( i = 0; i < cpoly->n_slots; i++ )
    {
	BgrSetPosition(&sym_dc, cpoly->slots[i].x, cpoly->slots[i].y);
	BgrSetRotation(&sym_dc, cpoly->slots[i].angle);
	BgrSleazeElement(&sym_dc, cpoly->info->slot_elt);
    }
}

PRIVATE void BgrCompileAbatis (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    BgrPoint	*ap;
    double	width = WORLD_SIZE(cpoly, cpoly->spec.sizes[0]);

    /*  The abatis points indexed 0..2 give the end point, notch point, and
     *  point at which the notch rejoins the shrunk end of the main line.
     *  The end point is the original end of the main line.
     *
     *  If the line is faked, put point 0 at the faked start point, and
     *  determine point 2 from the specified notch width and slope of the
     *  faked line.
     */

    ap = cpoly->ex->abatis.points;

    if (cpoly->faked)
    {
	ap[0] = *cpoly->p;
	BgrGetNearbyPoints(ap, cpoly->p[1].x - ap->x, cpoly->p[1].y - ap->y,
			   (BgrPoint*) NULL, ap+2, width, FALSE);
    }
    else
    {
	ap[0] = *cpoly->orig_p;
	ap[2] = cpoly->p[0];
    }

    /*  Make the depth of the notch equal its width.
     */

    ap[1].x = (ap[0].x + ap[2].x) / 2;
    ap[1].y = (ap[0].y + ap[2].y) / 2;
    
    BgrGetNearbyPoints(ap+1, ap[2].x - ap[0].x, ap[2].y - ap[0].y,
		       ap+1, (BgrPoint*) NULL, width, TRUE);
}

/*  BgrShrinkPoly
**
**  Starting from polyline values in cpoly->orig_p and cpoly->orig_nout,
**  shrink the line at one or both ends.  If cpoly->info for this polyline
**  specifies LF_SHRINK_START, shrink the start of the line away from its
**  original position by cpoly->spec.sizes[0]; if LF_SHRINK_END is
**  specified, shrink the end of the line by cpoly->spec.sizes[1].  If
**  neither flag is specified, shrink both, on the assumption that we were
**  invoked to make room for end elements.  Points that lie within the
**  shrink distances from the original endpoints are discarded.  cpoly->p
**  is written with cpoly->nout points which reflect the line with one or
**  both ends shrunk.
**
**  If the original line has insufficient length to permit the shrinkage
**  without disappearing, cpoly->poof is set to TRUE, and cpoly->p,nout are
**  given their original values.
**
**  You will likely need to draw a diagram in order to understand the
**  internals of this routine.
*/

PRIVATE void BgrShrinkPoly (dc, cpoly)
BgrDC		dc;
BgrCompPoly	*cpoly;
{
    double	dx, dy, length;
    BgrPoint	*p, *base;
    int		i;

    /*  Information about shrinkage of the start or end of the line.
     *  .amount is the amount by which we shrink that end.  .do_shrink
     *  indicates whether we shrink it at all.  .index is the point index on
     *  the original line to which the shrunk start or end index must move
     *  to be at least .amount units away from that end.
     */

    struct {
	int	index;
	Bool	do_shrink;
	double	amount;
    } start, end;

    if (cpoly->orig_nout == 1)
    {
	cpoly->poof = TRUE;
	return;
    }

    if (!(cpoly->info->flags & LF_SHRINK_BOTH))
	start.do_shrink = end.do_shrink = TRUE;
    else
    {
	start.do_shrink = (cpoly->info->flags & LF_SHRINK_START);
	end.do_shrink = (cpoly->info->flags & LF_SHRINK_END);
    }

    /*  To shrink the start of the line, put <base> at the start, and move
     *  <p> down the line away from <base> until the distance between them
     *  is greater than <start.amount>; <start.index> will be the index of
     *  that point.
     */

    if (!start.do_shrink)
    {
	start.amount = 0;
	start.index = 0;
    }
    else
    {
	start.amount = (WORLD_SIZE(cpoly, cpoly->spec.sizes[0])
			* cpoly->info->shrink_ratio);

	base = cpoly->orig_p;
	p = base + 1;

	for ( start.index = 1;
	      start.index < cpoly->orig_nout;
	      ++start.index, ++p )
	{
	    dx = p->x - base->x;
	    dy = p->y - base->y;
	    length = sqrt(dx*dx + dy*dy);
	    if (length >= start.amount)
		break;
	}
    }

    /*  To shrink the end of the line, put <base> at the end, and move <p>
     *  back up the line away from <base> until the distance between them is
     *  greater than <end.amount>; <end.index> will be the index of that
     *  point.
     */

    if (!end.do_shrink)
    {
	end.amount = 0;
	end.index = cpoly->orig_nout-1;
    }
    else
    {
	end.amount = (WORLD_SIZE(cpoly, cpoly->spec.sizes[1])
		      * cpoly->info->shrink_ratio);

	base = &cpoly->orig_p[cpoly->orig_nout-1];
	p = base - 1;

	for ( end.index = cpoly->orig_nout-2;
	      end.index >= 0;
	      --end.index, --p )
	{
	    dx = base->x - p->x;
	    dy = base->y - p->y;
	    length = sqrt(dx*dx + dy*dy);
	    if (length >= end.amount)
		break;
	}
    }

    /*  If <start.index> is greater than <end.index>, we may have shrunk the
     *  line out of existence.  If they're not adjacent, we certainly have.
     *  If they are adjacent, the segment they define is the only segment on
     *  the line which has not been discarded by the shrinkage.  We won't
     *  even bother drawing it; rather, we'll just draw one segment from the
     *  start point to the end point.  Make sure that this single segment is
     *  greater than the combined amount of shrinkage.
     */

    if (start.index > end.index)
    {
	if (start.index - 1 > end.index)
	    cpoly->poof = TRUE;
	else
	{
	    dx = cpoly->orig_p[cpoly->orig_nout-1].x - cpoly->orig_p->x;
	    dy = cpoly->orig_p[cpoly->orig_nout-1].y - cpoly->orig_p->y;
	    length = sqrt(dx*dx + dy*dy);
	    if (start.amount + end.amount > length)
		cpoly->poof = TRUE;
	    else
	    {
		cpoly->p[0] = cpoly->orig_p[0];
		cpoly->p[1] = cpoly->orig_p[cpoly->orig_nout-1];
		cpoly->nout = 2;
		return;
	    }
	}
    }

    if (cpoly->poof)
    {
	cpoly->nout = cpoly->orig_nout;
	bcopy((char*) cpoly->orig_p, (char*) cpoly->p,
	      cpoly->orig_nout * sizeof(BgrPoint));
	return;
    }

    /*  Otherwise, the new cpoly->p will be the points from <start.index> to
     *  <end.index>, inclusive, plus one extra on either end which are the
     *  points to be shrunk from their original positions.
     */

    if (start.do_shrink)
	--start.index;
    if (end.do_shrink)
	++end.index;
    cpoly->nout = end.index - start.index + 1;

#ifdef HPUX_CCOPT_BUGS
    for ( i = start.index, p = cpoly->p; i <= end.index; ++i, ++p )
	    memcpy(p, &cpoly->orig_p[i], sizeof(BgrPoint));
#else
    for ( i = start.index, p = cpoly->p; i <= end.index; ++i, ++p )
	*p = cpoly->orig_p[i];
#endif

    /*  Finally, determine the shrunk endpoints.
     */

    p = cpoly->p;
    base = cpoly->orig_p;

    if (start.do_shrink)
	BgrGetNearbyPoints(base, p[1].x - base->x, p[1].y - base->y,
			   (BgrPoint*) NULL, p, start.amount, FALSE);
    else
	*p = *base;

    p = &cpoly->p[cpoly->nout-1];
    base = &cpoly->orig_p[cpoly->orig_nout-1];

    if (end.do_shrink)
	BgrGetNearbyPoints(base, base->x - p[-1].x, base->y - p[-1].y,
			   p, (BgrPoint*) NULL, end.amount, FALSE);
    else
	*p = *base;
}

/*  BgrMaybeFakePoly
**
**  Another weapon in the war against the pathological special cases
**  involved in generating polylines.  The compilation process for certain
**  lines requires at least two output points, even if the original input
**  line contained only one point, or if a call to BgrShrinkPoly resulted in
**  the line going poof.
**
**  For example, the LinearAbatis polyline needs a start point at which to
**  place one end of its notch, and another point to determine in what
**  direction to point the notch.  Similarly, any drawn minefield polyline
**  is meaningless unless it has room for at least one of its slot symbols.
**
**  If the current points in cpoly->p can't meet these needs, either because
**  cpoly->poof is TRUE or only one output point was generated, rewrite
**  cpoly->p with a two-point line which will satisfy the needs of the
**  polyline compilation process.  If the polyline's LF_FAKE_MASK matches
**  LF_FAKE_START, the polyline compilation will be using cpoly->p[0] as a
**  base point for frill generation, with cpoly->p[1] used to determine the
**  slope (example: LinearAbatis.)  If the FAKE_MASK matches LF_FAKE_END,
**  the reverse is true (example: AxisOfAdvance.)  If it matches
**  LF_FAKE_LINE, we must generate real line data which will be drawn in
**  place of the bogus output points (example: any minefield.)
**
**  Read the code.
*/

PRIVATE void BgrMaybeFakePoly (cpoly)
BgrCompPoly *cpoly;
{
    BgrPoint	*p = cpoly->p;
    double	dx, dy;
    double	dist, length;

    /*  Determine if we can get away with not constructing a fake line.  We
     *  certainly can't if the line doesn't exist (poof == TRUE).  If it
     *  does, check if the polyline type requires that we have real data
     *  (FAKE_MASK matches FAKE_LINE.)  If the length of the current output
     *  line is less than the requested width, there's no room for symbol
     *  patterning, so we have to construct fake data.
     */

    if (!cpoly->poof)
    {
	if ((cpoly->info->flags & LF_FAKE_MASK) != LF_FAKE_LINE)
	    return;
	dx = cpoly->p[cpoly->nout-1].x - cpoly->p->x;
	dy = cpoly->p[cpoly->nout-1].y - cpoly->p->y;
	length = sqrt(dx*dx + dy*dy);
	if (length > WORLD_SIZE(cpoly, cpoly->spec.sizes[0]))
	    return;
    }

    /*  Case 1: there is only one output point, or the first and last output
     *  points are adjacent.  For the FAKE_START or FAKE_END flags, use the
     *  existing p[0] as the start or end point, and generate the other
     *  point so that the line is horizontal.  For the FAKE_LINE case,
     *  generate a horizontal line centered at p[0].
     */

    if (cpoly->nout == 1 || length == 0)
    {
	switch (cpoly->info->flags & LF_FAKE_MASK)
	{
	case LF_FAKE_START:
	    p[1].x = p->x + 1;
	    p[1].y = p->y;
	    break;
	case LF_FAKE_END:
	    p[1] = *p;
	    p->x = p[1].x - 1;
	    break;
	case LF_FAKE_LINE:
	    dist = WORLD_SIZE(cpoly, cpoly->spec.sizes[0]) / 2;
	    p[1].x = p->x + dist;
	    p[1].y = p->y;
	    p[0].x -= dist;
	    break;
	}
    }

    /*  Case 2: there are at least two separate output points, and they
     *  aren't far enough apart.  For the FAKE_START or FAKE_END flags,
     *  we're again only concerned with the slope, so just write p[1] with
     *  the last point already in the output point list.  For the FAKE_LINE
     *  case, create a line of the necessary length, centered between the
     *  first and last output points, with slope the same as an imaginary
     *  segment joining the first and last output points.
     *
     *  As if this weren't absurd enough, we have to increase the desired
     *  line length by the merest fraction to ensure that, if the line will
     *  be used by BgrCompileSlots, the calculated segment length will not
     *  be less than what we assigned here due to float/double precision
     *  problems.  (During testing, a precision flub of 1e-8 caused this
     *  case to manifest.)
     *
     *  CHRIST ON DRUGS, I'M SICK OF THIS SHIT.
     */

    else
    {
	BgrPoint tmp, *end = &p[cpoly->nout-1];
	switch (cpoly->info->flags & LF_FAKE_MASK)
	{
	case LF_FAKE_START:
	case LF_FAKE_END:
	    p[1] = *end;
	    break;
	case LF_FAKE_LINE:
	    dx = end->x - p->x;
	    dy = end->y - p->y;
	    tmp.x = (p->x + end->x) / 2;
	    tmp.y = (p->y + end->y) / 2;
	    length = WORLD_SIZE(cpoly, cpoly->spec.sizes[0]) * 1.01;
	    BgrGetNearbyPoints(&tmp, dx, dy, p, p+1, length/2, FALSE);
	    break;
	}
    }

    cpoly->nout = 2;
    cpoly->faked = TRUE;
}

PRIVATE void BgrDrawEndSymbols (dc, cpoly, flags)
BgrDC		dc;
BgrCompPoly	*cpoly;
BgrFlags	flags;
{
    _BgrDC sym_dc;

    sym_dc = *dc;
    BgrSetSize(&sym_dc, PIXEL_SIZE(cpoly, cpoly->spec.sizes[2]));
    BgrSetRotation(&sym_dc, 0.0);

    BgrSetPosition(&sym_dc, cpoly->orig_p->x, cpoly->orig_p->y);
    BgrSleazeElement(&sym_dc, cpoly->spec.end_elt);

    BgrSetPosition(&sym_dc, cpoly->orig_p[cpoly->orig_nout-1].x,
		   cpoly->orig_p[cpoly->orig_nout-1].y);
    BgrSleazeElement(&sym_dc, cpoly->spec.end_elt);
}

#endif /* USE_X */
