#ifndef lint
static char rcsid [] = "$RCSfile: cb_handler.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cb_handler.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to register/unregister event handlers                *
*       Created: Wed Feb 26 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include <stdio.h>
#include "libcb_local.h"
#include <stdalloc.h> /*common/include/global*/

/* We allocate room for handlers in groups of 8.  Therefore, the relation
 * between the number of handlers stored to those allocated is:
 * 0 : 0;	1-8 : 8;	9-16 : 16;	etc.
 */
#define ALLOC_QUANTUM 8
#define ALLOCED_HANDLERS(stored) \
  (ALLOC_QUANTUM * (((stored)+ALLOC_QUANTUM-1)/ALLOC_QUANTUM))

void callback_register_handler(CALLBACK_EVENT_PTR ev,
			       CALLBACK_HANDLER function,
			       ADDRESS user_data)
{
    CB_HANDLER *h;
    int32 nold, nnew;
    int32 i;

    /* Idiot checks */
    if (!ev)
    {
	printf("Warning: callback_register_handler passed a NULL event handle\n");
	return;
    }
    if (!function)
    {
	printf("Warning: callback_register_handler passed a NULL function\n");
	return;
    }
    if (ev->nhandlers == -1)
    {
	printf("Warning: callback_register_handler passed a free'd event handle\n");
	return;
    }

    /* Check whether there is an opening (a handler was removed) */
    for(i=0;i<ev->nhandlers;i++)
    {
	if (!ev->handlers[i].func)
	{
	    ev->handlers[i].func = function;
	    ev->handlers[i].user_data = user_data;
            ev->nacthandlers += 1;
	    return;
	}
    }
 
    /* See if there is room for another handler */
    if (ev->nhandlers == ALLOCED_HANDLERS(ev->nhandlers))
    {
	/* Make room for more */
	nold = ev->nhandlers;
	nnew = ALLOCED_HANDLERS(nold+1);
	h = (CB_HANDLER *)STDALLOC(nnew * sizeof(CB_HANDLER));

	if (nold)	/* if there were zero, don't copy or deallocate */
	{
	    bcopy((char *)ev->handlers,
		  (char *)h,
		  (int)(nold * sizeof(CB_HANDLER)));
	    STDDEALLOC (ev->handlers);
	}
	
	ev->handlers = h;
    }

    /* Store the new one */
    ev->handlers[ev->nhandlers].func = function;
    ev->handlers[ev->nhandlers].user_data = user_data;
    ev->nhandlers += 1;
    ev->nacthandlers += 1;
}

void callback_unregister_handler(CALLBACK_EVENT_PTR ev,
				 CALLBACK_HANDLER function,
				 ADDRESS user_data)
{
    int32 i;
    int32 found;

    found = 0;
    /* Idiot checks */
    if (!ev)
    {
	printf("Warning: callback_unregister_handler passed a NULL event handle\n");
	return;
    }
    if (!function)
    {
	printf("Warning: callback_unregister_handler passed a NULL function\n");
	return;
    }
    if (ev->nhandlers == -1)
    {
	printf("Warning: callback_unregister_handler passed a free'd event handle\n");
	return;
    }

    /* Nullify all references that match */
    for(i=0;i<ev->nhandlers;i++)
    {
	if ((ev->handlers[i].func == function) &&
	    (ev->handlers[i].user_data == user_data))
	{
	    ev->handlers[i].func = (CALLBACK_HANDLER)0;
            ev->nacthandlers--;
	    found = 1;
	}
    }
    if (!found)
      printf("Warning:  callback_unregister_handler received a handler that was not found\n");
}

    
/* 
 *  This function returns the number of registrations for the event
 *  that are currently active
 */

int32 callback_event_registrations(CALLBACK_EVENT_PTR ev)
{

    /* Idiot checks */
    if (!ev)
    {
	printf("Warning: callback_event_registrations passed a NULL event handle\n");
	return 0;
    }
    return ev->nacthandlers;
}    
    
