#ifndef lint
static char rcsid [] = "$RCSfile: cb_invoke.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cb_invoke.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to invoke a bunch of event handlers                  *
*       Created: Wed Feb 26 1990                                            *
*       Author: jesmith                                                     *
*       Remarks: This is based on the libsched dispatch function            *
*                                                                           *
****************************************************************************/

#include "libcb_local.h"
#include <stdext.h> /*common/include/global*/
#include <stdio.h>
#include <stdarg.h>

/*VARARGS*/
void callback_fire_event(CALLBACK_EVENT_PTR ev, ...)
{
    struct
    {
	uint16 tag;
	union
	{
	    int32     int_arg;
	    float64  double_arg;
	} arg;
    } a[CALLBACK_MAX_ARGS];
    int32 i;
    va_list ap;

    /* Idiot checks */
    if (!ev)
    {
	printf("Warning: callback_fire_event passed a NULL event handle\n");
	return;
    }
    if (ev->nhandlers == -1)
    {
	printf("Warning: callback_fire_event passed a free'd event handle\n");
	return;
    }

    /* Get the arguments */
    va_start(ap, ev);
    for (i=0;i<ev->nargs;i++)
    {
	a[i].tag = ev->arg_types[i];
	switch (a[i].tag)
	{
	  case A_INT:
	    a[i].arg.int_arg = va_arg(ap, int);
	    break;
	  case A_DOUBLE:
	    a[i].arg.double_arg = va_arg(ap, double);
	    break;
	}
    }
    va_end(ap);

    /* This is a little gross, but I'm hoping the compiler can untangle
       it into some decent code... */

#define F for (i=0;i<ev->nhandlers;i++) \
	    if (ev->handlers[i].func) \
	      ev->handlers[i].func(ev->handlers[i].user_data)
#define F1(w) for (i=0;i<ev->nhandlers;i++) \
		if (ev->handlers[i].func) \
		  ev->handlers[i].func(w, ev->handlers[i].user_data)
#define F2(w,x) for (i=0;i<ev->nhandlers;i++) \
		  if (ev->handlers[i].func) \
		    ev->handlers[i].func(w,x, ev->handlers[i].user_data)
#define F3(w,x,y) for (i=0;i<ev->nhandlers;i++) \
		    if (ev->handlers[i].func) \
		      ev->handlers[i].func(w,x,y, ev->handlers[i].user_data)
#define F4(w,x,y,z) for (i=0;i<ev->nhandlers;i++) \
		      if (ev->handlers[i].func) \
		        ev->handlers[i].func(w,x,y,z, \
					     ev->handlers[i].user_data)

#define C1(t0) F1(a[0].arg.t0);
#define C2(t0,t1) F2(a[0].arg.t0,a[1].arg.t1);
#define C3(t0,t1,t2) F3(a[0].arg.t0,a[1].arg.t1,a[2].arg.t2);
#define C4(t0,t1,t2,t3) F4(a[0].arg.t0,a[1].arg.t1,a[2].arg.t2,a[3].arg.t3);
#define S(n,x,y) \
	switch (a[n].tag) \
	{ \
	  case A_INT: x ; break ; \
	  case A_DOUBLE: y ; break ; \
	}


  /* First, set the executing flag to TRUE, to prevent the event from
   * being deallocated in the middle of the for statement.  The
   * deallocate will be deferred until the end of the switch.
   */
  ev->executing = TRUE;
    
    switch(ev->nargs)
    {
      case 0:
	F;
	break;
      case 1:
	S(0,
	  { C1(int_arg) },
	  { C1(double_arg) } );
	break;
      case 2:
	S(0,
	  S(1,
	    { C2(int_arg,int_arg) },
	    { C2(int_arg,double_arg) } ),
	  S(1,
	    { C2(double_arg,int_arg) },
	    { C2(double_arg,double_arg) } )
	  );
	break;
      case 3:
	S(0,
	  S(1,
	    S(2,
	      { C3(int_arg,int_arg,int_arg) },
	      { C3(int_arg,int_arg,double_arg) }),
	    S(2,
	      { C3(int_arg,double_arg,int_arg) },
	      { C3(int_arg,double_arg,double_arg) })),
	  S(1,
	    S(2,
	      { C3(double_arg,int_arg,int_arg) },
	      { C3(double_arg,int_arg,double_arg) }),
	    S(2,
	      { C3(double_arg,double_arg,int_arg) },
	      { C3(double_arg,double_arg,double_arg) }))
	  );
	break;
      case 4:
	S(0,
	  S(1,
	    S(2,
	      S(3,
		{ C4(int_arg,int_arg,int_arg,int_arg) },
		{ C4(int_arg,int_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(int_arg,int_arg,double_arg,int_arg) },
		{ C4(int_arg,int_arg,double_arg,double_arg) })),
	    S(2,
	      S(3,
		{ C4(int_arg,double_arg,int_arg,int_arg) },
		{ C4(int_arg,double_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(int_arg,double_arg,double_arg,int_arg) },
		{ C4(int_arg,double_arg,double_arg,double_arg) }))),
	  S(1,
	    S(2,
	      S(3,
		{ C4(double_arg,int_arg,int_arg,int_arg) },
		{ C4(double_arg,int_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(double_arg,int_arg,double_arg,int_arg) },
		{ C4(double_arg,int_arg,double_arg,double_arg) })),
	    S(2,
	      S(3,
		{ C4(double_arg,double_arg,int_arg,int_arg) },
		{ C4(double_arg,double_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(double_arg,double_arg,double_arg,int_arg) },
		{ C4(double_arg,double_arg,double_arg,double_arg) })))
	  );
	break;
      default:
	printf("Cannot invoke functions with %d args.\n", ev->nargs);
	return;
    }
    
    /* Reset the executing flag */
    ev->executing = FALSE;
    
    /* Now, if there are no more handlers, call destroy event callback,
     * so that the event can be deallocated.
     */
    if (ev->nhandlers == -1)
      callback_destroy_event(ev);
    
}


