/*
 * $RCSfile: libcallback.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libcallback.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libcallback                        *
*       Created: Wed Feb 26 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBCALLBACK_H_INCLUDED
#define _LIBCALLBACK_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> /*common/include/global*/

/* Generic function type of event handlers */
typedef void (*CALLBACK_HANDLER) (/* [arg, [arg, [arg, [arg,]]]]
				     ADDRESS user_data
				     */);

/* callback_define_event:
 *
 * Creates an event and returns a handle which can be used to register
 * handlers, fire the event, or destroy the event.  The arguments specify
 * the types of up to four arguments which will be passed when the event
 * fires.  (e.g., callback_define_event(A_INT, A_INT, A_DOUBLE, A_END)).
 */
typedef struct cb_event *CALLBACK_EVENT_PTR;
extern CALLBACK_EVENT_PTR callback_define_event(int32 arg, ...);

/* callback_register_handler:
 *
 * Notes that a function is to be called when an event occurs.  The
 * event_handle identifies the event (created with callback_define_event),
 * the handler_function specifies the address of the function which is
 * to be called.  The handler function should expect the argument types
 * specified when the event was created (via callback_define_event), as
 * well as a trailing argument which is the constant value passed here
 * as user data.
 */
extern void callback_register_handler(CALLBACK_EVENT_PTR ev,
				      CALLBACK_HANDLER function,
				      ADDRESS user_data);

/* callback_fire_event:
 *
 * Invokes all the handlers defined for the event in no particular
 * order (if order is important, do the dispatching yourself).
 */
extern void callback_fire_event(CALLBACK_EVENT_PTR ev, ...);

/* callback_unregister_handler:
 *
 * If a handler for the event_handle passed can be found which matches
 * the handler_function/user_data passed, it will be removed from
 * the list of functions called when the event is fired.
 */
extern void callback_unregister_handler(CALLBACK_EVENT_PTR ev,
				        CALLBACK_HANDLER function,
				        ADDRESS user_data);

/* callback_destroy_event:
 *
 * The memory associated with the passed event_handle is freed.
 * Referencing this handle after after a call to this function will
 * probably make the program crash (don't do it).
 */
extern void callback_destroy_event(CALLBACK_EVENT_PTR ev);


/* callback_event_registrations:
 *
 *  This function returns the number of registrations currently active
 *  for the event.  If 0 is returned, there are no handlers active and
 *  code processing required for the callback need not be done.
 *  
 */
extern int32 callback_event_registrations(CALLBACK_EVENT_PTR ev);

#ifdef __cplusplus
}
#endif

#endif /*_LIBCALLBACK_H_INCLUDED*/
