#ifndef lint
static char rcsid [] = "$RCSfile: cl_access.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cl_access.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Software which handles allocation/access to classes       *
*       Created: Tue Mar 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libclass_local.h"
#include <stdalloc.h> /*common/include/global*/

CLASS_USER_DATA_TYPE class_alloc_user_data(CLASS_PTR class)
{
    uint32 size;
    CLASS_SPECIAL_INFO *special;
    CLASS_USER_DATA_TYPE *data;

    /* Protect the user from his own stupidity */
    if (!class)
      return (CLASS_USER_DATA_TYPE)0;

    /* If the class does not yet have global debug flags, allocate
     * those now.
     */
    if (!class->debug_flags)
      class->debug_flags = class_alloc_debug_flags(class->num_slots);

    /* Make a block of memory big enough to hold all the user
     * data's for this class.  Note that slot 0 is reserved by
     * us for the debugging routines.
     */
    size = sizeof(CLASS_USER_DATA_TYPE) * (class->num_slots + 1);
    data = (CLASS_USER_DATA_TYPE *)STDALLOC(size);
    if (!data)
      return (CLASS_USER_DATA_TYPE)(0);
    bzero(data, size);

    /* Fill in the magic 0 entry */
    special = (CLASS_SPECIAL_INFO *)STDALLOC(sizeof(CLASS_SPECIAL_INFO));
    if (!special)
      return (CLASS_USER_DATA_TYPE)(0);
    bzero(special, sizeof(CLASS_SPECIAL_INFO));
    special->class = class;
    special->debug_flags = class_alloc_debug_flags(class->num_slots);

    data[0] = (CLASS_USER_DATA_TYPE)(special);

    /* Give the user a pointer to the block of memory */
    return (CLASS_USER_DATA_TYPE)(data);
}

void class_free_user_data(CLASS_USER_DATA_TYPE block)
{
    CLASS_SPECIAL_INFO *special;

    if (!block)
      return;

    special = ((CLASS_SPECIAL_INFO **)block)[0];

    if (!special)
      return;

    STDDEALLOC(special->debug_flags);
    STDDEALLOC(special);
    /* In case the user accidently references the block */
    ((CLASS_USER_DATA_TYPE *)block)[0] = (CLASS_USER_DATA_TYPE)0;
    STDDEALLOC(block);
}

void class_set_user_data(CLASS_USER_DATA_TYPE block,
			 int32 handle,
			 CLASS_USER_DATA_TYPE user_data)
{
    /* Protect the user from his own stupidity */
    if (!block || !handle)
      return;

    ((CLASS_USER_DATA_TYPE *)block)[handle] = user_data;
}

void class_set_global_data(CLASS_USER_DATA_TYPE block,
			   CLASS_USER_DATA_TYPE user_data)
{
    CLASS_SPECIAL_INFO *special;

    if (!block)
      return;

    special = ((CLASS_SPECIAL_INFO **)block)[0];

    if (!special)
      return;

    special->global_data = user_data;
}

CLASS_USER_DATA_TYPE class_get_global_data(CLASS_USER_DATA_TYPE block)
{
    CLASS_SPECIAL_INFO *special;

    if (!block)
      return 0;

    special = ((CLASS_SPECIAL_INFO **)block)[0];

    if (!special)
      return 0;

    return special->global_data;
}
