#ifndef lint
static char rcsid [] = "$RCSfile: cl_debug.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cl_debug.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle debug flags                                *
*       Created: Thu Mar 26 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libclass_local.h"
#include <stdalloc.h> /*common/include/global*/

#include <stdstring.h> /*common/include/global*/

static int32 slot_listed(char *slot, char *list);
static int32 debug_check = 0;
#define TOKENS " \t,;:|&."

/* h=1 sets bit 1, h=2 sets bit 2, and so on */
#define SET_FLAG(d,s)  ((char *)(d))[(s)>>3] |=  (1 << ((s) & 0x7))
#define CLR_FLAG(d,s)  ((char *)(d))[(s)>>3] &= ~(1 << ((s) & 0x7))
#define TST_FLAG(d,s) (d ? (((char *)(d))[(s)>>3] &   (1 << ((s) & 0x7))) : 0)

char *class_alloc_debug_flags(int32 num_slots)
{
    char *f;

    if (f = (char *)STDALLOC(1 + num_slots/8))
      bzero(f, 1 + num_slots/8);

    return f;
}

static void class_set_debug(CLASS_USER_DATA_TYPE block, int32 slot, int32 bit)
{
    CLASS_SPECIAL_INFO *special;
    int32 i;

    if (!block)
      return;

    special = ((CLASS_SPECIAL_INFO **)block)[0];

    /* Note that somebody might enable debugging before any of these are
     * created.  Deal with that possibility.
     */
    if (!special->debug_flags)
      special->debug_flags =
	class_alloc_debug_flags(special->class->num_slots);

    if (slot)
    {
	if (bit)
	  SET_FLAG(special->debug_flags, slot);
	else
	  CLR_FLAG(special->debug_flags, slot);
    }
    else
    {
	for (i=0;i<=special->class->num_slots;i++)
	{
	    if (bit)
	      SET_FLAG(special->debug_flags, i);
	    else
	      CLR_FLAG(special->debug_flags, i);
	}
    }
}

static void class_set_debug_global(CLASS_PTR class, int32 slot, int32 bit)
{
    int32 i;

    /* Note that somebody might enable debugging before any of these are
     * created.  Deal with that possibility.
     */
    if (!class->debug_flags)
      class->debug_flags = class_alloc_debug_flags(class->num_slots);

    if (slot)
    {
	if (bit)
	  SET_FLAG(class->debug_flags, slot);
	else
	  CLR_FLAG(class->debug_flags, slot);
    }
    else
    {
	for (i=0;i<=class->num_slots;i++)
	{
	    if (bit)
	      SET_FLAG(class->debug_flags, i);
	    else
	      CLR_FLAG(class->debug_flags, i);
	}
    }
}

int32 class_get_debug(CLASS_USER_DATA_TYPE block, int32 slot)
{
    CLASS_SPECIAL_INFO *special;

    /* See if anyone ever set a debug flag in the system */
    if (!debug_check)
      return 0;

    if (!block)
      return 0;

    special = ((CLASS_SPECIAL_INFO **)block)[0];

    if (!slot)
      return 0;

    return (TST_FLAG(special->class->debug_flags, slot) ||
	    TST_FLAG(special->debug_flags, slot));
}

void class_debug_byname(CLASS_USER_DATA_TYPE block, char *fields, int32 bit)
{
    CLASS_SPECIAL_INFO *special;
    CLASS_INFO *c;
    CLASS_SLOT_INFO *s;

    /* Turn on the flag that says something somewhere 
     * might be being debugged
     */
    debug_check = 1;

    /* If they provided a block, apply it to just that one.  Otherwise,
     * apply it to every class.
     */
    if (block)
      special = ((CLASS_SPECIAL_INFO **)block)[0];
    else
      special = (CLASS_SPECIAL_INFO *)0;

    if (!strcmp(fields, "all"))
    {
	if (special)
	  class_set_debug(block, 0, bit);
	else
	  for (c=class_classes();c;c=c->next)
	    class_set_debug_global(c, 0, bit);
    }
    else if (special)
    {
	for (s=special->class->slots;s;s=s->next)
	  if (slot_listed(s->name, fields))
	    class_set_debug(block, s->number, bit);
    }
    else
    {
	for (c=class_classes();c;c=c->next)
	  for (s=c->slots;s;s=s->next)
	    if (slot_listed(s->name, fields))
	      class_set_debug_global(c, s->number, bit);
    }
}

void class_show_names (void)
{
    CLASS_INFO *c;
    CLASS_SLOT_INFO *s;

    for (c=class_classes();c;c=c->next)
	for (s=c->slots;s;s=s->next)
		printf ("class %d-%d:  %s\n", c->number,
			s->number, s->name);
}

void class_show(CLASS_USER_DATA_TYPE user_data, char *fields)
{
    CLASS_USER_DATA_TYPE *block = (CLASS_USER_DATA_TYPE *)user_data;
    CLASS_INFO *class;
    CLASS_SLOT_INFO *slot;
    int32 name_len, i;
    int32 just_what, print_all;

    if (!block)
      return;

    class = ((CLASS_SPECIAL_INFO **)(block))[0]->class;

    /* Special: If fields is "what", just print the names of non-empty
     *          slots.
     */
    just_what = !strcmp(fields, "what");

    /* Special: If fields is "all", print every non-empty slot.
     */
    print_all = !strcmp(fields, "all");

    /* Now loop through the slots defined for this class and print
     * any who's names appear in the list (except in the two cases
     * just tested).
     */
    for (slot=class->slots;slot;slot=slot->next)
    {
	/* Skip 0 or not-filled-in slots */
	if (!block[slot->number])
	  continue;

	if (just_what)
	{
	    printf("%s\n", slot->name);
	    continue;
	}

	if (print_all || slot_listed(slot->name, fields))
	{
	    /* Pretty-print the name of the slot */
	    name_len = strlen(slot->name)/2;
	    for (i=0;i<35-name_len;i++)
	      printf("\\");
	    printf("%s", slot->name);
	    for (i=0;i<35-name_len;i++)
	      printf("/");
	    printf("\n");

	    /* Now call the print routine for the slot */
	    slot->printer(block[slot->number]);
	}
    }
}

static int32 slot_listed(char *slot, char *list)
{
    char toked[1024]; /* Better safe than segv! */
    char *t;

    /* Copy the list, because strtok is destructive */
    strncpy(toked, list, 1024);

    /* Get the first token */
    t = strtok(toked,TOKENS);

    while(t)
    {
	/* If the current token matches, return success */
	if (!strcmp(t, slot))
	  return 1;

	/* Get the next token */
	t = strtok((char *)0,TOKENS);
    }

    /* Ran out of tokens.  I guess it's not here */
    return 0;
}
