#ifndef lint
static char rcsid [] = "$RCSfile: cl_declare.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cl_declare.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Software to handle declaration of a class                 *
*       Created: Tue Mar 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/
#include "libclass_local.h"
#include <stdalloc.h> /*common/include/global*/

#include <stdstring.h> /*common/include/global*/

/* The structures maintained in this file just keep track of how many
 * slots are in each class and stuff like that.  Good old singly linked
 * lists are used to maintain these structures, since they are only
 * searched at init time, or for debugging.
 */

static int32 n_classes = 0;
static CLASS_INFO *classes = (CLASS_INFO *)0;
static void generic_printer(CLASS_USER_DATA_TYPE data);

/* This a very private function which is shared within the library.
 * Don't even THINK of calling it from an application!
 */
CLASS_INFO *class_classes(void) {return classes;}

CLASS_PTR class_declare_class(void)
{
    CLASS_INFO *new;

    /* Allocate a class_info structure to remember stuff about
     * this class.
     */
    new = (CLASS_INFO *)STDALLOC(sizeof(CLASS_INFO));
    bzero(new, sizeof(CLASS_INFO));
    new->next = classes;
    classes = new;

    /* Give it a number */
    new->number = n_classes;
    n_classes += 1;

    return new;
}

int32 class_reserve_user_data(CLASS_PTR class,
			      char *name,
			      CLASS_PRINTER printer)
{
    CLASS_SLOT_INFO *slot;

    /* Make sure no instances have already been created of this class */
    if (class->debug_flags)
    {
	printf("Fatal Error: Attempted to reserve user data for a class which has already\n             been instantiated.  Dumping core.\n");
	abort();
    }

    /* Initialize the description of this slot */
    slot = (CLASS_SLOT_INFO *)STDALLOC(sizeof(CLASS_SLOT_INFO));
    bzero(slot, sizeof(CLASS_SLOT_INFO));

    /* Store the slot with its class */
    class->num_slots += 1;
    slot->next = class->slots;
    class->slots = slot;

    /* Note the slot's number.  We do this after incrementing
     * num_slots, so the numbers start with 1 (the 0 entry of each
     * instance of a class is used to store type information for
     * debugging).
     */
    slot->number = class->num_slots;

    /* Fill in miscellaneous data */
    if (name && *name)
      strcpy(slot->name, name);
    else
      sprintf(slot->name, "<Unnamed slot %d of class %d>",
	      slot->number, class->number);
    if (printer)
      slot->printer = printer;
    else
      slot->printer = generic_printer;

    return slot->number;
}

static void generic_printer(CLASS_USER_DATA_TYPE data)
{
    /* This gets called when the application neglected to provide
     * a printer for class data.
     */
    printf("No print function available.\nUser data = %d = 0x%x\n",
	   (int)data, (int)data);
}
