#ifndef lint
static char rcsid [] = "$RCSfile: cmd_aggregate.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cmd_aggregate.c                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government         *
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause at               *
*       DFARS 252.227-7013 (OCT 1988).                                      *
*                                                                           *
*       Contents: Aggregate command line option arrays.                     *
*       Created: August 25, 1993                                            *
*       Author: bcollard                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcmdline.h"
#include <stdio.h>
#include <stdext.h>
#include <stdalloc.h>

/*
 *  This function aggregates command line option arrays.
 */

int32 cmd_aggregate(
    CMD_OPTION **aggregate,        /* Pointer to aggregate array. */
    uint32      *sizeof_aggregate, /* Size of aggregate array. */
    CMD_OPTION  *piece,            /* Array to be added to aggregate. */
    uint32       sizeof_piece,     /* Size of array to be added. */
    CMD_TYPES ***map)              /* Pointer to array containing a mapping
				      between aggregate array and its pieces.*/
{
    CMD_OPTION *new_aggregate; /* Pointer to array containing original
                                  aggregate and the new piece. */
    CMD_TYPES **new_map;       /* Pointer to array containing original
                                  map and new map entries. */
    int32 num_aggregate;       /* Number of elements in original aggregate
                                  array. */
    int32 num_piece;           /* Number of elements in piece array. */
    int32 total;               /* Number of elements in new aggregate array. */
    int32 i;                   /* A counter. */

    /*
     *  Find the sizes of the arrays.
     */

    num_aggregate = *sizeof_aggregate / sizeof (CMD_OPTION);
    num_piece = sizeof_piece / sizeof (CMD_OPTION);
    total = num_aggregate + num_piece;

    /*
     *  Allocate memory for new option array and map.
     */

    if ((new_aggregate = (CMD_OPTION *) STDALLOC (*sizeof_aggregate +
            sizeof_piece)) == NULL)
      return (FALSE);
    if ((new_map = (CMD_TYPES **) STDCALLOC (num_aggregate + num_piece,
            sizeof (CMD_TYPES *))) == NULL)
    {
        STDDEALLOC (new_aggregate);
        return (FALSE);
    }

    /*
     *  Copy original aggregate array (it there is one) and new piece
     *  into the new aggregate array.
     */

    if (*sizeof_aggregate > 0)
      bcopy ((char *) *aggregate, (char *) new_aggregate, *sizeof_aggregate);
    bcopy ((char *) piece, (char *) &(new_aggregate[num_aggregate]),
            sizeof_piece);
    if (*sizeof_aggregate > 0)
      STDDEALLOC (*aggregate);
    *aggregate = new_aggregate;
    *sizeof_aggregate += sizeof_piece;

    /*
     *  Copy old map (if there is one) into the new map.
     */

    if (num_aggregate > 0)
    {
        bcopy ((char *) *map, (char *) new_map,
                num_aggregate * sizeof (CMD_TYPES *));
        STDDEALLOC (*map);
    }
    *map = new_map;

    /*
     *  Save pointers to the value entry in the static piece.
     */

    for (i = num_aggregate; i < total; i++)
      new_map[i] = &(piece[i - num_aggregate].value);

    return (TRUE);
}

/*
 *  This routine process the command line arguments in an aggregated
 *  command line option array.
 */

void cmd_process_aggregate_options(
    int         argc,
    argv_t      argv,
    int32      *leftover_argc,
    argv_t     *leftover_argv,
    CMD_OPTION *options,
    uint32      sizeof_options,
    CMD_TYPES **map,
    int32       verbose)
{
    int32 i;
    int32 num_options;

    /*
     *  Process the command line options.
     */

    cmd_process_options(argc, argv, /* args to main() */
			leftover_argc, leftover_argv, /* unrecognized */
			options, /* options array */
			sizeof_options, /* size of options array */
			verbose /* Verbose mode */
			);

    /*
     *  Find the number of options.
     */

    num_options = sizeof_options / sizeof (CMD_OPTION);

    /*
     *  Copy the values from the aggregate array to the static pieces.
     */

    for (i = 0; i < num_options; i++)
        bcopy ((char *) &(options[i].value), (char *) map[i],
                sizeof (CMD_TYPES));

    /*
     *  Deallocate memory.
     */

    STDDEALLOC (options);
    STDDEALLOC (map);
}
