#ifndef lint
static char rcsid [] = "$RCSfile: cmd_proc.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: cmd_proc.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Command line processing                                   *
*       Created: Sun Jun 21 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcmdline.h"
#include <stdio.h>
#include <stdext.h> /*common/include/global*/
#include <stdalloc.h>
#include <stdstring.h> /*common/include/global*/
#include <stdlib.h>

#if (mips & lint)
extern char *strtok (/* char *s1, *s2 */);
#endif

#define MAX_ENV_ARGS 128 /* Should be plenty.  Feel free to increase */

static int32 tokenize_environment(char *env, char *argv[]);
static int32 check_options(CMD_OPTION options[], uint32 n_options);
static char *value_string(
    CMD_OPTION options[],
    uint32     n_options,
    int32      index,
    int32      default_p);
static char *type_string(
    CMD_OPTION options[],
    uint32     n_options,
    int32      index);
static void print_help(
    char       *progname,
    CMD_OPTION *options,
    uint32      n_options,
    char       *env_name);
static int32 process_arglist(
    char       *progname,
    int32       argc,
    argv_t      argv,
    CMD_OPTION *options,
    uint32      n_options,
    argv_t      leftover_argv);
static void check_depend(CMD_OPTION options[], uint32 n_options);
static int32 depends_match(
    char      *dep,
    CMD_OPTION options[],
    uint32     n_options);
static char *describe_depend(
    char        startwith[],
    CMD_OPTION *option,
    char        endwith[]);

void cmd_process_options(
    int         argc,
    argv_t      argv,
    int32      *leftover_argc,
    argv_t     *leftover_argv,
    CMD_OPTION *options,
    uint32      sizeof_options,
    int32       verbose)
{
    int32 i;
    uint32 n_options = sizeof_options/sizeof(CMD_OPTION);
    int32 env_argc;
    char *env_argv[MAX_ENV_ARGS];
    char *progname;
    char env_name[128], *e, *p;
    int32 leftover_env_argc;
    char **leftover_env_argv;

    progname = strrchr(argv[0], '/');
    if (progname)
      progname++;
    else
      progname = argv[0];

    /* Derive the name of the environment variable */
    *env_name = 0;
    
    for (e=env_name,p=progname;*p;p++)
    {
	if ((*p >= 'A') && (*p <= 'Z') || (*p >= '0') && (*p <= '9'))
	  *e++ = *p;
	else if ((*p >= 'a') && (*p <= 'z'))
	  *e++ = *p - 'a' + 'A';
	else
	  break;		/* bail out whenever nonalpha */
    }
    *e = 0;

    if (*env_name == 0)
      strcat(env_name, "DEFAULT");

    strcat(env_name, "ARGS");

    /* Check legality of passed options.
     */
    if (!check_options(options, n_options))
      exit(0);	

    /* If any characteristic help signals are passed, just print the
     * help and exit.
     */
    for (i=0;i<argc;i++)
    {
	if (!strcmp(argv[i], "-h") ||
	    !strcmp(argv[i], "-?") ||
	    !strcmp(argv[i], "-help"))
	{
	    print_help(progname, options, n_options, env_name);
	    exit(0);
	    /*NOTREACHED*/
	}
    }

    /* First, run through and copy all the defaults to be the result values.
     */
    for (i=0;i<n_options;i++)
      options[i].value = options[i].defalt;

    /* Now go get the environmental defaults.
     */
    env_argc = tokenize_environment(env_name, env_argv);

    /* Allocate space to store the pointers to the leftovers.
     */
    leftover_env_argv = (char **)STDALLOC((env_argc + 1) * sizeof(char *));
    *leftover_argv = (argv_t)STDALLOC((argc + env_argc) * sizeof(char *));


    /* Process the environmental defaults */
    leftover_env_argc = process_arglist(progname,
					(int32)env_argc, (argv_t)env_argv,
                                        options, n_options,
                                        (argv_t)leftover_env_argv);

    /* Process the command line */
    *leftover_argc = process_arglist(progname,
				     (int32)argc, argv,
				     options, n_options,
				     (*leftover_argv));

    /* Add the leftovers (we want to process the environmentals first,
       but we want to parse the command line first for griping */
    for (i = 0; i < leftover_env_argc; i++)
      (*leftover_argv)[*leftover_argc + i] = leftover_env_argv[i];

    *leftover_argc += leftover_env_argc;

    STDDEALLOC(leftover_env_argv);

    /* Check the dependencies */
    check_depend(options, n_options);

    /* If the verbose switch is TRUE, run through and print all
     * the results.
     */
    if (verbose)
      for (i=0;i<n_options;i++)
	if (*options[i].option &&
	    depends_match(options[i].dependent_option,
			  options, n_options))
	  printf("%40s: %s\n", options[i].name,
		 value_string(options, n_options, i, FALSE));
}

static int32 tokenize_environment(
    char *env,
    char *argv[])
{
    char *env_value = getenv(env);
    char *env_copy;
    int32 argc;

    /* This function takes a string and turns it into argc,argv like
     * unix gives to main.
     */

    if (!env_value || !*env_value)
      return 0;

    /* Copy the environment variable value into some allocated space
     * (I doubt it's safe to go changing the string getenv gave me).
     */
    env_copy = (char *)STDALLOC(strlen(env_value)+1);
    strcpy(env_copy, env_value);

    argv[0] = strtok(env_copy, " \t");
    for (argc=1;;argc++)
      if (!(argv[argc] = strtok(NULL, " \t")))
	break;

    return argc;
}

static void print_help(
    char       *progname,
    CMD_OPTION *options,
    uint32      n_options,
    char       *env_name)
{
    int32 i, lines;
    char ignored[128];

    printf("Usage: %s {optionally followed by...}\n", progname);

    lines = 1;
    for (i=0;i<n_options;i++)
    {
	/* Skip continuation options */
	if (!*options[i].name)
	  continue;

	/* Stop the output every 10 items (every 20th line) */
	if (!(lines%10))
	{
	    printf("---hit <Return> for more---");
	    fflush(stdout);
	    gets(ignored);
	}
	lines += 1;

	switch (options[i].type)
	{
	  case CMD_BOOLEAN:
	  case CMD_TOGGLE:
	    printf("  -%s | -%s\t%s\n\t%s (defaults to %s)\n",
		   options[i].option, options[i].anti_option,
		   describe_depend("(allowed if", options+i,
				   " also specified)"),
		   options[i].help,
		   options[i].defalt.boolean ?
		   options[i].option : options[i].anti_option);
	    break;
	  case CMD_INTEGER:
	    if (options[i].anti_option)
	      printf("  -%s %s| -%s\t%s\n\t%s (defaults to %s)\n",
		     options[i].option,
		     type_string(options, n_options, i),
		     options[i].anti_option,
		     describe_depend("(allowed if", options+i,
				     " also specified)"),
		     options[i].help,
		     options[i].defalt.integer ?
		     value_string(options, n_options, i, TRUE) :
		     options[i].anti_option);
	    else
	      printf("  -%s %s\t%s\n\t%s (defaults to %s)\n",
		     options[i].option,
		     type_string(options, n_options, i),
		     describe_depend("(allowed if", options+i,
				     " also specified)"),
		     options[i].help,
		     value_string(options, n_options, i, TRUE));
	    break;
	  case CMD_STRING:
	    if (options[i].anti_option)
	      printf("  -%s %s| -%s\t%s\n\t%s (defaults to %s)\n",
		     options[i].option,
		     type_string(options, n_options, i),
		     options[i].anti_option,
		     describe_depend("(allowed if", options+i,
				     " also specified)"),
		     options[i].help,
		     options[i].defalt.string ?
		     value_string(options, n_options, i, TRUE) :
		     options[i].anti_option);
	    else
	      printf("  -%s %s\t%s\n\t%s (defaults to %s)\n",
		     options[i].option,
		     type_string(options, n_options, i),
		     describe_depend("(allowed if", options+i,
				     " also specified)"),
		     options[i].help,
		     value_string(options, n_options, i, TRUE));
	    break;
	  case CMD_FLOAT:
	    if (options[i].anti_option)
	      printf("  -%s %s| -%s\t%s\n\t%s (defaults to %s)\n",
		     options[i].option,
		     type_string(options, n_options, i),
		     options[i].anti_option,
		     describe_depend("(allowed if", options+i,
				     " also specified)"),
		     options[i].help,
		     options[i].defalt.floating ?
		     value_string(options, n_options, i, TRUE) :
		     options[i].anti_option);
	    else
	      printf("  -%s %s\t%s\n\t%s (defaults to %s)\n",
		     options[i].option,
		     type_string(options, n_options, i),
		     describe_depend("(allowed if", options+i,
				     " also specified)"),
		     options[i].help,
		     value_string(options, n_options, i, TRUE));
	    break;
	}
    }

    if (env_name)
	{
	char *env_val = getenv(env_name);
	printf("These defaults may be changed with the environment variable: %s\nwhich currently has the value: \"%s\"\n", env_name, env_val ? env_val : "");
	printf("You can change this value with the command:\n\tsetenv %s \"{default command line options}\"\n", env_name);
	}
}

static char *value_string(
    CMD_OPTION options[],
    uint32     n_options,
    int32      index,
    int32      default_p)
{
    static char buf[128];
    char hold[128];
    union cmd_types *value;
    int i;

    /* This function cobbles together a string showing the all values
     * associated with an option (or the defaults), including all
     * continuations.
     */

    *buf = 0;
    for (i=index;i<n_options;i++)
    {
	/* Do this one, and any continuations */
	if ((i > index) && *options[i].name)
	  break;

	if (default_p)
	  value = &options[i].defalt;
	else
	  value = &options[i].value;

	switch(options[i].type)
	{
	    /* Note that BOOLEANs and TOGGLEs cannot be involved with
	     * continuations, so we can just return those values.
	     */
	  case CMD_BOOLEAN:
	    if (value->boolean)
	      return "True";
	    else
	      return "False";
	  case CMD_TOGGLE:
	    if (value->toggle)
	      return "On";
	    else
	      return "Off";
	  case CMD_INTEGER:
	    sprintf(hold, "%d ", value->integer);
	    strcat(buf, hold);
	    break;
	  case CMD_STRING:
	    if (value->string)
	    {
		strcat(buf, value->string);
		strcat(buf, " ");
	    }
	    else
	      strcat(buf, "(null) ");
	    break;
	  case CMD_FLOAT:
	    sprintf(hold, "%f ", value->floating);
	    strcat(buf, hold);
	    break;
	  default:
	    return "?Unknown Type?";
	}
    }
    return buf;
}

static char *type_string(
    CMD_OPTION options[],
    uint32     n_options,
    int32      index)
{
    static char buf[128];
    int i;

    /* This function cobbles together a string showing the all types
     * expected for an option, including all continuations.
     */

    *buf = 0;
    for (i=index;i<n_options;i++)
    {
	/* Do this one, and any continuations */
	if ((i > index) && *options[i].name)
	  break;

	switch(options[i].type)
	{
	  case CMD_INTEGER:
	    strcat(buf, "<integer> ");
	    break;
	  case CMD_STRING:
	    strcat(buf, "<string> ");
	    break;
	  case CMD_FLOAT:
	    strcat(buf, "<float> ");
	    break;
	  default:
	    return "?Unexpected Type?";
	}
    }
    return buf;
}

static int32 process_arglist(
    char       *progname,
    int32       argc,
    argv_t      argv,
    CMD_OPTION *options,
    uint32      n_options,
    argv_t      leftover_argv)
{
    int32 i, j;
    char *arg;
#define MATCHED_OPT  1
#define MATCHED_ANTI 2
    int32 match, which_match;
    int32 left = 0;
    int32 len, mcount;

    /* This is the workhorse function.  It runs down a list of arguments
     * (argv) and figures out which options they correspond to.  Unmatched
     * arguments are retuned in leftover_argv.  The return value is the
     * number of things in leftover_argv.
     */

    for (i=0;i<argc;i++)
    {
	match = 0;

	/* Grab this arg.  Don't even look at arguments which don't
	 * start with '-'.
	 */
	arg = argv[i];
	if (arg[0] == '-')
	{
	    /* See if we can find a match (look at arg+1 to skip the -)
	     */
	    for (j=0;j<n_options;j++)
	    {
		/* Skip continuations */
		if (!*options[j].name)
		  continue;

		if (!strcmp(arg+1, options[j].option))
		{
		    match = MATCHED_OPT;
		    which_match = j;
		    break;
		}
		if (options[j].anti_option &&
		    !strcmp(arg+1, options[j].anti_option))
		{
		    match = MATCHED_ANTI;
		    which_match = j;
		    break;
		}
	    }

	    /* If that didn't find anything, and the option ends with
	     * a period (such as -e.) assume it is an abbreviation,
	     * and try to find a unique match.
	     */
	    if (!match && (arg[strlen(arg)-1] == '.'))
	    {
		len = strlen(arg)-2; /* skip the - and the . */
		mcount = 0;
		for (j=0;j<n_options;j++)
		{
		    /* Skip continuations */
		    if (!*options[j].name)
		      continue;

		    if (!strncmp(arg+1, options[j].option, (int)len))
		    {
			match = MATCHED_OPT;
			which_match = j;
			mcount++;
			printf("Note: %s is a recognized abbreviation for %s\n",
			       arg, options[j].option);
		    }
		    if (options[j].anti_option &&
			!strncmp(arg+1, options[j].anti_option, (int)len))
		    {
			match = MATCHED_ANTI;
			which_match = j;
			mcount++;
			printf("Note: %s is a recognized abbreviation for %s\n",
			       arg, options[j].anti_option);
		    }
		}
		/* If we matched more than one, forget about it */
		if (mcount > 1)
		{
		    printf("Warning: %s is not a unique abbreviation\n", arg);
		    match = 0;
		}
	    }
	}

	/* If there was not a match, relegate this to the leftover pile,
	 * and move on.
	 */
	if (!match)
	{
	    leftover_argv[left++] = arg;
	    continue;
	}

	/* If we matched the anti-option, set the value of the option and
	 * any continuations to zero, and move on.
	 */
	if (match == MATCHED_ANTI)
	{
	    options[which_match].value.integer = 0;
	    for (j=which_match+1;j<n_options;j++)
	    {
		if (*options[j].name)
		  break;

		options[j].value.integer = 0;
	    }
	    continue;
	}

	/* If we matched the option, get the value for this guy and
	 * any continuations.
	 */
	for (j=which_match;j<n_options;j++)
	{
	    /* Only do this one and any continuations */
	    if ((j > which_match) && *options[j].name)
	      break;

	    /* The scheme used to get the values associated with the
	     * option is to advance the for loop variable 'i' as each
	     * option is fetched.  That way, arguments which are matched
	     * are skipped in the next loop of processing.  This allows
	     * strings arguments to start with '-' without causing any
	     * problems.
	     *
	     * If we expect an argument and there aren't any, or if
	     * sscanf doesn't recognize the format of the argument, then
	     * print the help message and bail.
	     */

	    switch (options[j].type)
	    {
	      case CMD_BOOLEAN:
		options[j].value.boolean = 1;
		continue;
	      case CMD_TOGGLE:
		options[j].value.toggle = 1;
		continue;
	      case CMD_INTEGER:
		if ((i < argc-1) &&
		    sscanf(argv[++i],
			   "%d", &options[j].value.integer) == 1)
		  continue;
		else
		{
		    print_help(progname, options, n_options, NULL);
		    exit(0);
		    /*NOTREACHED*/
		}
	      case CMD_STRING:
		if (i < argc-1)
		{
		    options[j].value.string =
		      (char *)STDALLOC(strlen(argv[++i])+1);
		    strcpy(options[j].value.string, argv[i]);
		    continue;
		}
		else
		{
		    print_help(progname, options, n_options, NULL);
		    exit(0);
		    /*NOTREACHED*/
		}
	      case CMD_FLOAT:
		if ((i < argc-1) &&
		    sscanf(argv[++i],
			   "%f", &options[j].value.floating) == 1)
		  continue;
		else
		{
		    print_help(progname, options, n_options, NULL);
		    exit(0);
		    /*NOTREACHED*/
		}
	    }
	}
    }

    return left;
}

void cmd_gripe(int32 argc, argv_t argv)
{
    int32 i;

    if (argc > 1) /* Expect 1 (the program name) */
    {
	printf("\007Warning: these options not recognized:\n");
	for (i=1;i<argc;i++)
	  printf("\t%s", argv[i]);
	printf("\nRestart with option -help for usage.\n");
    }
}

static int32 check_options(
    CMD_OPTION options[],
    uint32     n_options)
{
    int32 i;
    int32 j;
    int32 can_continue;

    /* Check all the fields of all the options for proper values.
     * In some cases, correct the value, in others, barf.  The point of
     * this function is to find common mistakes during testing, instead
     * of having bizarre behavior show up when we give the code to
     * end users.  If you make a mistake in setting up the argument list
     * which this function does not detect, by all means, update this
     * function to find that problem in the future.
     */
    can_continue = 0;
    for (i=0;i<n_options;i++)
    {
	if (!options[i].name || !*options[i].name)
	{
	    if (can_continue)
	      options[i].name = "";
	    else
	    {
		printf("Cannot specify continuation of option #%d\n", i);
		return 0;
	    }

	    if ((options[i].type == CMD_BOOLEAN) ||
		(options[i].type == CMD_TOGGLE))
	    {
		printf("Continuation option #%d must be integer, string, or float\n",
		       i+1);
		return 0;
	    }

	    if ((options[i].help && *options[i].help) ||
		(options[i].option && *options[i].option) ||
		(options[i].dependent_option &&
		 *options[i].dependent_option) ||
		(options[i].anti_option && *options[i].anti_option))
	    {
		printf("Continuation option #%d must have help, option, dependent_option,\nand anti_option all = NULL.\n",
		       i+1);
		return 0;
	    }
		
	}

	if (!options[i].help)
	  options[i].help = "";

	/* This can be NULL, but not "" */
	if (options[i].dependent_option && !*options[i].dependent_option)
	  options[i].dependent_option = NULL;

	/* Make sure the user didn't try to use both conjunctive and
	 * disjunctive constructs in the dependencies.
	 */
	if (options[i].dependent_option &&
	    strchr(options[i].dependent_option, '&') &&
	    strchr(options[i].dependent_option, '|'))
	{
	    printf("You can use either & or | in the dependecies for %s,\nbut not both.\n",
		   options[i].name);
	    return 0;
	}

	if ((options[i].type < CMD_BOOLEAN) || (options[i].type > CMD_FLOAT))
	{
	    printf("%s has unrecognized type %d\n",
		   options[i].name, options[i].type);
	    return 0;
	}

	if (!options[i].option || !*options[i].option)
	{
	    /* It's only OK to have "" as the option if the name is
	     * also "" (i.e., it is a continuation of the previous option.
	     */
	    if (*options[i].name)
	    {
		printf("%s has no -{option} specified\n",
		       options[i].name);
		return 0;
	    }
	    options[i].option = "";
	}

	/* This can be NULL, but not "" */
	if (options[i].anti_option && !*options[i].anti_option)
	  options[i].anti_option = NULL;

	/* but it can't be NULL if boolean or toggle */
	if (!options[i].anti_option &&
	    ((options[i].type == CMD_BOOLEAN) ||
	     (options[i].type == CMD_TOGGLE)))
	{
	    printf("%s must have anti_option specified\n", options[i].name);
	    return 0;
	}

        /* Options or anti-options cannot be duplicated. */
        /* Check to make sure this is not a continued option. */
        if (strcmp (options[i].name, ""))
        {
            for (j = 0; j < i; j++)
            {
                if (options[j].option && options[i].option &&
                        !strcmp (options[j].option, options[i].option))
                {
                    printf ("Option %s duplicated.\n", options[j].option);
                    printf ("\tLeading option name is: %s\n", options[j].name);
                    printf ("\tLeading option help is: %s\n", options[j].help);
                    printf ("\tTrailing option name is: %s\n",options[i].name);
                    printf ("\tTrailing option help is: %s\n",options[i].help);
                    return 0;
                }
                if (options[j].option && options[i].anti_option &&
                        !strcmp (options[j].option, options[i].anti_option))
                {
                    printf ("Option %s duplicated by anti-option.\n",
                            options[j].option);
                    printf ("\tLeading option name is: %s\n", options[j].name);
                    printf ("\tLeading option help is: %s\n", options[j].help);
                    printf ("\tTrailing anti-option name is: %s\n",
                            options[i].name);
                    printf ("\tTrailing anti-option help is: %s\n",
                            options[i].help);
                    return 0;
                }
                if (options[j].anti_option && options[i].option &&
                        !strcmp (options[j].anti_option, options[i].option))
                {
                    printf ("Anti-option %s duplicated by option.\n",
                            options[j].anti_option);
                    printf ("\tLeading anti-option name is: %s\n",
                            options[j].name);
                    printf ("\tLeading anti-option help is: %s\n",
                            options[j].help);
                    printf ("\tTrailing option name is: %s\n",options[i].name);
                    printf ("\tTrailing option help is: %s\n",options[i].help);
                    return 0;
                }
                if (options[j].anti_option && options[i].anti_option &&
                        !strcmp (options[j].anti_option,
                        options[i].anti_option))
                {
                    printf ("Anti-option %s duplicated.\n",
                            options[j].anti_option);
                    printf ("\tLeading anti-option name is: %s\n",
                            options[j].name);
                    printf ("\tLeading anti-option help is: %s\n",
                            options[j].help);
                    printf ("\tTrailing anti-option name is: %s\n",
                            options[i].name);
                    printf ("\tTrailing anti-option help is: %s\n",
                            options[i].help);
                    return 0;
                }
            }
        }

	can_continue = ((options[i].type != CMD_BOOLEAN) &&
			(options[i].type != CMD_TOGGLE));
    }

    return 1;
}

static void check_depend(
    CMD_OPTION options[],
    uint32     n_options)
{
    int32 i;
    char *dep;
    int32 should_affirm;

    for (i=0;i<n_options;i++)
    {
	/* Is this option dependent on any other options? */
	if (!(dep = options[i].dependent_option))
	  continue;

	/* Does this option have a value different from the system default?
	 * If not, don't complain to the user about something over which
	 * he has no control.
	 */
	if (options[i].defalt.integer == options[i].value.integer)
	  continue;

	if ((options[i].type == CMD_BOOLEAN) ||
	    (options[i].type == CMD_TOGGLE))
	  should_affirm = options[i].value.integer;
	else
	  should_affirm = 1;

	if (depends_match(dep, options, n_options) != should_affirm)
	{
	    printf("\007Warning: -%s cannot be specified without%s\n",
		   options[i].option,
		   describe_depend("", options+i, ""));
	    options[i].value = options[i].defalt;
	}
    }
}

static char *describe_depend(
    char        startwith[],
    CMD_OPTION *option,
    char        endwith[])
{
    char dep[128];
    char *d;
    char hold[128];
    char *junctive, *token;
    static char buffer[256];

    /* This function cobbles together a string describing the
     * dependencies attached to an option.
     */

    if (!option->dependent_option)
      return "";

    strcpy(dep, option->dependent_option);

    if (strchr(dep,'&'))
      junctive = "and", token = "&";
    else
      junctive = "or", token = "|";

    d = strtok(dep, token);
    sprintf(buffer, "%s -%s", startwith, d);
    while (d = strtok(NULL, token))
    {
	sprintf(hold, " %s -%s", junctive, d);
	strcat(buffer, hold);
    }
    strcat(buffer, endwith);
    return buffer;
}

static int32 depends_match(
    char      *dep,
    CMD_OPTION options[],
    uint32     n_options)
{
    char depcopy[128];
    char *d;
    char *token;
    int32 is_and;
    int32 i;
    int32 unexpected;

    /* This function determines if the user followed the rules in
     * specifying options.  Depending on a BOOLEAN or a TOGGLE means
     * requiring that the value is TRUE.  Depending on any other type
     * means requiring that the value be different from the default.
     * Depending on an anti-option means that the value must be 0.
     */

    if (!dep)
      return 1;

    if (strchr(dep,'&'))
      token = "&", is_and = 1;
    else
      token = "|", is_and = 0;

    strcpy(depcopy, dep);
    d = strtok(depcopy, token);
    do
    {
	for (i=0;i<n_options;i++)
	{
	    /* What value would we not expect to see if the dependecies
	     * were kosher?
	     */
	    if (!strcmp(options[i].option, d))
	    {
		if ((options[i].type == CMD_BOOLEAN) ||
		    (options[i].type == CMD_TOGGLE))
		  unexpected = 0;
		else
		  unexpected = options[i].defalt.integer;
	    }
	    else if (options[i].anti_option &&
		     !strcmp(options[i].anti_option, d))
	      unexpected = 1;
	    else
	      continue;

	    /* If the dependencies have been violated, then whether
	     * TRUE or FALSE stops evaluation depends upon whether
	     * this is a conjunctive or disjunctive expression.
	     */
	    if ((options[i].value.integer == unexpected) && is_and)
	      return 0;
	    else if ((options[i].value.integer != unexpected) && !is_and)
	      return 1;
	}
    } while (d = strtok(NULL, token));

    if (is_and)
      return 1;
    else
      return 0;
}

