/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libcmdline                               *
*       Created: Sun Jun 21 1992                                            *
*       Author: jesmith                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcmdline.h"
#include <stdext.h> /*common/include/global*/

struct
{
    CMD_TOGGLE_OPTION safsim;
    CMD_TOGGLE_OPTION safstation;
    CMD_STRING_OPTION terrain;
    CMD_INTEGER_OPTION exercise;
    CMD_TOGGLE_OPTION network;
    CMD_BOOLEAN_OPTION synch;
    CMD_INTEGER_OPTION mcache;
    CMD_STRING_OPTION startmap;
    CMD_INTEGER_OPTION scalenum;
    CMD_INTEGER_OPTION scaledenom;
    CMD_FLOAT_OPTION rndseed;
    CMD_INTEGER_OPTION xloc;
    CMD_INTEGER_OPTION yloc;
} options = {
    {
	"SAFSIM" , "Selects whether to run SAF simulation processes" ,
	NULL ,
	CMD_TOGGLE ,	"sim" ,		"nosim" ,	TRUE ,
    } ,
    {
	"SAFSTATION" , "Selects whether to run SAF workstation processes" ,
	NULL ,
	CMD_TOGGLE ,	"station" ,	"nostation" ,	TRUE ,
    } ,
    {
	"Terrain Database Name" , "Specifies terrain database" ,
	"station|sim" ,
	CMD_STRING ,	"terrain" ,	"nowhere" ,	"knox-0311" ,
    } ,
    {
	"Network" , "Selects whether to use network" ,
	NULL ,
	CMD_TOGGLE ,	"network" ,	"nonet" ,	TRUE ,
    } ,
    {
	"Exercise ID" , "Specifies simulation exercise id" ,
	"network" ,
	CMD_INTEGER ,	"exercise" ,	NULL ,		1 ,
    } ,
    {
	"Synchronous" , "Run X windows in sychronous mode for debugging" ,
	"station" ,
	CMD_BOOLEAN ,	"synch" ,	"asynch" ,	FALSE ,
    } ,
    {
	"Map Cache" , "Specifies number of map screens to cache" ,
	"station" ,
	CMD_INTEGER ,	"mcache" ,	"nomcache" ,	2 ,
    } ,
    { 
	"Map Starting Grid" , "Specifies initial location of map" ,
	"station" ,
	CMD_STRING ,	"startmap" ,	"center" ,	NULL ,
    } ,
    {
	"Map Starting Scale" , "Specifies initial map scale # : #" ,
	"station" ,
	CMD_INTEGER ,	"scale" ,	NULL ,		1 ,
    } ,
    {
	/* Map Starting Scale takes two arguments */
	NULL , NULL , NULL , CMD_INTEGER , NULL , NULL , 200000 ,
    } ,
    { 
	"Random Number Seed" , "Specifies seed for random number generator" ,
	"sim" ,
	CMD_FLOAT ,	"rand" ,	NULL ,		0.0 ,
    } ,
    {
	"X" , "Starting X location" ,
	"y&station" ,
	CMD_INTEGER ,	"x" ,		NULL ,		0 ,
    } ,
    {
	"Y" , "Starting Y location" ,
	"x&station" ,
	CMD_INTEGER ,	"y" ,		NULL ,		0 ,
    } ,
};

main(argc, argv)
    int   argc;
    char *argv[];
{
    int32 leftover_argc;
    char **leftover_argv;

    cmd_process_options(argc, argv, /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized */
			(CMD_OPTION *)&options, /* options array */
			sizeof(options), /* size of options array */
			TRUE /* Verbose mode */
			);

    /* Pass remaining args to X windows */

    cmd_gripe(leftover_argc, leftover_argv);

    /* Example usage... */
    printf("Reading /saf/terrain/%s\n", options.terrain.value);
}
