/****************************************************************************
*   File: test_aggregate.c                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government         *
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause at               *
*       DFARS 252.227-7013 (OCT 1988).                                      *
*                                                                           *
*       Contents: Test program for aggregate                                *
*       Created: Tues Aug 24, 1993                                          *
*       Author: bcollard                                                    *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcmdline.h"
#include <stdext.h> /*common/include/global*/

/* Process module. */

struct
{
    CMD_TOGGLE_OPTION safsim;
    CMD_TOGGLE_OPTION safstation;
} process_options = {
    {
        "SAFSIM" , "Selects whether to run SAF simulation processes" ,
        NULL ,
        CMD_TOGGLE ,    "sim" ,         "nosim" ,       TRUE ,
    } ,
    {
        "SAFSTATION" , "Selects whether to run SAF workstation processes" ,
        NULL ,
        CMD_TOGGLE ,    "station" ,     "nostation" ,   TRUE ,
    } ,
};

void process_init (options, sizeof_options, map)
    CMD_OPTION **options;
    uint32      *sizeof_options;
    CMD_TYPES ***map;
{
    cmd_aggregate (options, sizeof_options,
                   (CMD_OPTION *)&process_options,
                   sizeof (process_options),
                   map);
}

void process_simul ()
{
    printf ("Simulation process is %s\n",
            process_options.safsim.value ? "On" : "Off");
    printf ("Workstation process is %s\n",
            process_options.safstation.value ? "On" : "Off");
}

/* Network module */

struct
{
    CMD_TOGGLE_OPTION network;
    CMD_INTEGER_OPTION exercise;
} network_options = {
    {
        "Network" , "Selects whether to use network" ,
        NULL ,
        CMD_TOGGLE ,    "network" ,     "nonet" ,       TRUE ,
    } ,
    {
        "Exercise ID" , "Specifies simulation exercise id" ,
        "network" ,
        CMD_INTEGER ,   "exercise" ,    NULL ,          1 ,
    } ,
};

void network_init (options, sizeof_options, map)
    CMD_OPTION **options;
    uint32      *sizeof_options;
    CMD_TYPES ***map;
{
    cmd_aggregate (options, sizeof_options,
                   (CMD_OPTION *)&network_options,
                   sizeof (network_options),
                   map);
}

void network_simul ()
{
    printf ("Network is %s\n",
            network_options.network.value ? "On" : "Off");
    if (network_options.network.value)
      printf ("Exercise ID is %d\n", network_options.exercise.value);
}

/* Main module */

main(argc, argv)
    int   argc;
    char *argv[];
{
    int32 leftover_argc;
    char **leftover_argv;
    CMD_OPTION *options;
    uint32 sizeof_options;
    CMD_TYPES **map;

    sizeof_options = 0;

    process_init (&options, &sizeof_options, &map);
    network_init (&options, &sizeof_options, &map);

    cmd_process_aggregate_options(argc, argv, /* args to main() */
                        &leftover_argc, &leftover_argv, /* unrecognized */
                        options, /* options array */
                        sizeof_options, /* size of options array */
                        map, /* option value map */
                        TRUE /* Verbose mode */
                        );

   process_simul ();
   network_simul ();
}
