#ifndef lint
static char rcsid [] = "$RCSfile: const_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: const_init.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Libconstants (the whole thing)                            *
*       Created: Wed Mar 20 1996                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libconstants.h"
#include <libreader.h>
#include <stdstring.h>
#include <stdalloc.h>
#include <stdext.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>

static READER_UNION *namespaces = NULL;
char *const_error;

typedef struct fileorder
{
    char *name;
    char *depname;
    READER_UNION ru;
    int32 level;
} FILEORD;

/* Qsort comparater for finding load order */
static int sort_order(const void *a, const void *b)
{
    if (((FILEORD *)a)->level > ((FILEORD *)b)->level)
      return 1;
    else if (((FILEORD *)a)->level < ((FILEORD *)b)->level)
      return -1;
    else
      return (strcmp(((FILEORD *)a)->name, ((FILEORD *)b)->name));
}


/* Read all the constant files in the passed directory.
 */
void const_init(char *const_dir)
{
    DIR *dirp;
    struct dirent *dp;
    READER_UNION ru;
    int32 start;
    int32 i,j,l;
    FILEORD *fname;
    int32 numnames = 0;
    int32 numalloced = 0;
    char tmp[255];
    char *sym;
    int32 max_level, done, loops, depok, maxl;

    const_error = reader_get_symbol("Constant Unknown");

    if (!(dirp = opendir(const_dir)))
      return;


    /* Make list of all files in directory with .rdr in name */
    while (dp = readdir(dirp))
    {
	if (strcmp(dp->d_name + strlen(dp->d_name) - 4, ".rdr"))
	  continue;

	if (numnames >= numalloced)
	{
	    if (numalloced)
	    {
		numalloced += 10;
		fname = (FILEORD *)STDREALLOC(fname, 
					      numalloced * sizeof(FILEORD));
	    }
	    else
	    {
		numalloced = 10;
		fname = (FILEORD *)STDALLOC(numalloced * sizeof(FILEORD));
	    }
	}
		
	fname[numnames].name = reader_get_symbol(dp->d_name);
	strcpy(tmp, dp->d_name);
	tmp[strlen(tmp) - 3] = 'd';
	tmp[strlen(tmp) - 2] = 'e';
	tmp[strlen(tmp) - 1] = 'p';
	fname[numnames].depname = reader_get_symbol(tmp);
	fname[numnames].level = -1;
	fname[numnames].ru.array = NULL;
	numnames++;
    }

    /* Find all files in directory with .dep in name */
    rewinddir(dirp);
    while (dp = readdir(dirp))
    {
	if (strcmp(dp->d_name + strlen(dp->d_name) - 4, ".dep"))
	  continue;

	sym = reader_get_symbol(dp->d_name);

	for (i=0;i<numnames;i++)
	{
	    if (sym == fname[i].depname)
	    {
		if (reader_read(dp->d_name, const_dir, &fname[i].ru, 0))
		{
		    printf("Error reading %s/%s\n", const_dir, dp->d_name);
		    fname[i].ru.array = NULL;
		    break;
		}
		break;
	    }
	}
	if (i == numnames)
	  printf("Found %s with no matching reader file\n", dp->d_name);
    }
    closedir(dirp);

    /* Find dependency order to read files in (from src/depends/depends.c) */
    max_level = -1;
    
    done = FALSE;
    for (loops=0; (!done) && (loops < numnames + 4); loops++)
    {
	/* Assume we succeed */
	done = TRUE;
	/* Go through all the files, and see if they are in correct order */
	for (j=0;j<numnames;j++)
	  
	  /* If it has no level, check it. */
          if (fname[j].level == -1)
          {
              depok = TRUE;
              maxl = -1;

	      /* If it has no dependencies, it goes in level 0 */
	      if (fname[j].ru.array)

		/* Go through all the dependencies.  If they all have levels,
		 * our level is one higher than the highest level of our
		 * dependencies.
		 */
		for (i=1;i<fname[j].ru.array[0].integer;i++)
		{
		    /* Find the dependency */
		    for (l=0;l<numnames;l++)
		      if (fname[l].name == fname[j].ru.array[i].charptr)
			break;

		    if (l == numnames)
		    {
			printf("Could not find needed reader file `%s'",
			       fname[j].ru.array[i].charptr);
			continue;
		    }
			
		    /* If the level of this dependency is not known yet, 
		     * we can't determine this file's level yet either. 
		     */
		    if (fname[l].level == -1)
		    {
			depok = FALSE;
			break;
		    }

		    if (fname[l].level > maxl)
		      maxl = fname[l].level;
		}
	      
	      /* We found the level of this file. */
              if (depok)
              {
                  fname[j].level = maxl+1;
                  if ((maxl + 1) > max_level)
                    max_level = maxl + 1;
              }
              else
                done = FALSE;
          }
    }
    
    /* Did we run into a loop? */
    if (loops == numnames + 4) 
    {
	printf("Circular dependency found in constants files.  Check:\n");
	for (i=0;i<numnames;i++)
	  if (fname[i].level == -1)
	    printf("%s ", fname[i].name);
	printf("\n");
    }

	    
    /* Sort the file entries */
    qsort(fname, numnames, sizeof(FILEORD), sort_order);
    
    /* Read the reader files. */
    for (i=0; i<numnames; i++)
    {
	if (reader_read(fname[i].name, const_dir, &ru, READER_TYPING))
	{
	    printf("Error reading %s/%s\n", const_dir, fname[i].name);
	    continue;
	}

	if (!namespaces)
	{
	    namespaces = (READER_UNION *)
	      STDALLOC(ru.array[0].integer * sizeof(READER_UNION));
	    start = 1;
	    namespaces[0].integer = ru.array[0].integer;
	}
	else
	{
	    start = namespaces[0].integer;
	    namespaces[0].integer += ru.array[0].integer - 1;
	    namespaces = (READER_UNION *)
	      STDREALLOC(namespaces,
			 namespaces[0].integer * sizeof(READER_UNION));
	}
	bcopy((char *)(ru.array+1), (char *)(namespaces+start),
	      (ru.array[0].integer-1) * sizeof(READER_UNION));
    }


    /* Sort the namespaces */
    reader_sort_tag_table(namespaces, READER_UNTAGGED);

    /* Sort the constants in the namespaces */
    for (i=1;i<namespaces[0].integer;i++)
      reader_sort_tag_table(namespaces[i].array, READER_TAGGED);

    /* Deallocate filename list */
    if (numalloced)
      STDDEALLOC(fname);
}

/* const_name_to_value:
 *
 * Translate a fully qualified constant (NAMESPACE_NAME) to its
 * numeric value.  name should be a libreader symbol.
 */
int32 const_name_to_value(char *name)
{
    /* If you crash on this next line, it's a good bet you forgot
     * to call reader_get_symbol on the argument to this function!
     */
    READER_UNION *ru = reader_get_symbol_value(name);

    if (!ru)
      return 0;

    if (ru[0].integer == 2)
      return ru[1].integer;

    /* Something went wrong. */
    if (name != reader_get_symbol(name))
      fprintf(stderr, "const_name_to_value passed raw string \"%s\" instead of libreader symbol\n",
	      name);
    else
      fprintf(stderr, "const_name_to_value passed unknown symbol \"%s\"\n",
	      name);
    return 0;
}

/* const_value_to_name:
 *
 * Translate a constant to its name (returned as a libreader symbol).
 * namespace should be a libreader symbol.
 */
char *const_value_to_name(char *namespace, int32 value)
{
    READER_UNION *ru;

    /* Find the namespace */
    if (!(ru = reader_find_tag_sorted(namespace, namespaces,
				      READER_UNTAGGED, READER_NO_ERRORS)))
    {
	/* Something went wrong */
	if (namespace != reader_get_symbol(namespace))
	  fprintf(stderr, "const_value_to_name passed raw string \"%s\" instead of libreader symbol\n",
		  namespace);
	else
	  fprintf(stderr, "const_value_to_name passed unknown namespace \"%s\"\n",
		  namespace);
	return const_error;
    }

    /* Find the value */
    if (!(ru = reader_find_tag_sorted((char *)value, ru,
				      READER_TAGGED, READER_NO_ERRORS)))
    {
	/* Something went wrong */
	return const_error;
    }

    return ru[2].charptr;
}

int32 const_convert_namespace(int32 value,
			      char *from_namespace,
			      char *to_namespace)
{
    int32 fc, tc;
    char buf[256];
    char *name = const_value_to_name(from_namespace, value);

    if (name == const_error)
      return 0;

    /* Build the name using the other namespace */
    fc = strlen(from_namespace);
    tc = strlen(to_namespace);
    bcopy(to_namespace, buf, tc);
    strcpy(buf+tc, name+fc);
    return const_name_to_value(reader_get_symbol(buf));
}

/* Dump the constants files to stdout.
 */
void const_dump()
{
    int32 i, j;
    READER_UNION *ru;

    for (i=1;i<namespaces[0].integer;i++)
    {
	ru = namespaces[i].array;
	printf("\nnamespace: %s\n", ru[1].charptr);
	for (j=2;j<ru[0].integer;j++)
	  printf("  value: %-40s %d\n",
		 ru[j].array[2].charptr, ru[j].array[1].integer);
    }
}
