/*
 * $RCSfile: libconstants.h,v $ $Revision: 1.2 $ $State: Exp $
 */
/****************************************************************************
*   File: libconstants.h                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libconstants                       *
*       Created: Wed Mar 20 1996                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBCONSTANTS_H_INCLUDED
#define _LIBCONSTANTS_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h>

/* const_init:
 *
 * Read all the constant files in the passed directory.
 */
extern void const_init(char *const_dir);

/* const_name_to_value:
 *
 * Translate a fully qualified constant (NAMESPACE_NAME) to its
 * numeric value.  name should be a libreader symbol.
 */
extern int32 const_name_to_value(char *name);

/* const_value_to_name:
 *
 * Translate a constant to its name (returned as a libreader symbol).
 * namespace should be a libreader symbol.  If this fails, it returns
 * the value of const_error.
 */
extern char *const_value_to_name(char *n_space, int32 value);

/* String returned by const_value_to_name when the value cannot be found.
 */
extern char *const_error;

/* const_convert_namespace:
 *
 * Translate a numerical value between two namespaces.  For example,
 * namespace: DIS203
 *     value: FOO 1
 * namespace: DIS204
 *     value: FOO 2
 * const_convert_namespace(1, DIS203_symbol, DIS204_symbol) == 2
 */
extern int32 const_convert_namespace(int32 value,
				     char *from_namespace,
				     char *to_namespace);

/* const_dump:
 *
 * Dump the constants files to stdout.
 */
extern void const_dump();

#ifdef __cplusplus
}
#endif

#endif /*_LIBCONSTANTS_H_INCLUDED*/
