#ifndef lint
static char rcsid [] = "$RCSfile: coord_convert.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: coord_convert.c                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to do simple conversions                             *
*       Created: Fri Aug 14 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Updated Mon Sep 11 1995
*                                                                           *
****************************************************************************/

#include "libcoord_local.h"
#include <libgcs.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdstring.h> /*common/include/global*/
#include <stdcoord.h>  /* common/include/global */
#include <math.h>

static int32 coord_tcc_aware = 1;

void
coord_set_tcc_awareness(int32 aware)
{
    coord_tcc_aware = aware;
}

#define ROUND(x) ((int32)((x)+0.5))

/*VARARGS*/
int32 coord_convert(COORD_SYSTEM from_system, ...)
{
    va_list       ap;
    /* Possible from args */
    float64       from_x;
    float64       from_y;
    float64       from_z;
    COORD_TCC_PTR from_tcc;
    int32         from_zone;
    float64       from_northing;
    float64       from_easting;
    int32         from_resolution;
    char         *from_string;
    float64       from_latitude;
    float64       from_longitude;
    int32         from_local_datum;
    int32         from_cell;
    /* Possible to args */
    COORD_SYSTEM  to_system;
    float64      *to_x;
    float64      *to_y;
    float64      *to_z;
    COORD_TCC_PTR to_tcc;
    int32        *to_zone;
    float64      *to_northing;
    float64      *to_easting;
    int32         to_resolution;
    char         *to_string;
    float64      *to_latitude;
    float64      *to_longitude;
    int32         to_local_datum;
    int32        *to_cell;
    int32         dest_cell;
    /* Intermediate step variables */
    int32         ret;
    float64       northing, easting;
    float64       latitude, longitude;
    float64       x, y, z;
    float64       u, v, w;
    int32         datum;
    int32         zone;
    int32         cell;
    int32         dummy_zone = 0;

    va_start(ap, from_system);

    switch(from_system)
    {
      case COORD_GCC:
	from_x = va_arg(ap,float64);
	from_y = va_arg(ap,float64);
	from_z = va_arg(ap,float64);
	break;

      case COORD_GCSCS:
	/* This is the same as COORD_GCS when used as the from system */
	from_system = COORD_GCS;

      case COORD_GCS:
#ifndef COORD_FORCE_SIMNET
	from_cell = va_arg(ap, int32);
#else
	from_tcc = coord_tcc_list[0];
#endif
	from_x = va_arg(ap, float64);
	from_y = va_arg(ap, float64);
	from_z = va_arg(ap, float64);
	break;

      case COORD_TCC:
	from_tcc = va_arg(ap,COORD_TCC_PTR);
	if ( coord_tcc_aware )
	{
	    if (!from_tcc)
	      return COORD_BAD_TCC;
	}
	else
	{
	    from_tcc    = NULL;
	    from_system = COORD_GCS;
	    from_cell   = GCS_ILLEGAL_CELL;
	}
	
	from_x = va_arg(ap,float64);
	from_y = va_arg(ap,float64);
	from_z = va_arg(ap,float64);
	break;

      case COORD_UTM_NE:
	from_zone     = va_arg(ap,int32);
	from_northing = va_arg(ap,float64);
	from_easting  = va_arg(ap,float64);
	from_z        = va_arg(ap,float64);
#if 0
	printf("LIBCOORDINATES: %d %f %f %f\n",
	       from_zone,from_northing,from_easting,from_z);
#endif
	break;

      case COORD_UTM_GRID:
	from_zone       = va_arg(ap,int32);
	from_resolution = va_arg(ap,int32);
	from_string     = va_arg(ap,char *);
	from_z          = va_arg(ap,float64);
	break;

      case COORD_LATLON:
	from_latitude    = va_arg(ap,float64);
	from_longitude   = va_arg(ap,float64);
	from_z           = va_arg(ap,float64);
	from_local_datum = va_arg(ap,int32);
	break;

      default:
	return COORD_SYSTEM_UNKNOWN;
    }

    /* Get the to_system args */

    to_system = (COORD_SYSTEM)va_arg(ap, int32);
    switch(to_system)
    {
      case COORD_GCC:
	to_x = va_arg(ap,float64 *);
	to_y = va_arg(ap,float64 *);
	to_z = va_arg(ap,float64 *);
	break;

      case COORD_GCS:
	to_cell = va_arg(ap, int32 *);
	to_x = va_arg(ap, float64 *);
	to_y = va_arg(ap, float64 *);
	to_z = va_arg(ap, float64 *);
	break;

      case COORD_GCSCS:
	dest_cell = va_arg(ap, int32);
	to_x = va_arg(ap, float64 *);
	to_y = va_arg(ap, float64 *);
	to_z = va_arg(ap, float64 *);
	break;

      case COORD_TCC:
	to_tcc = va_arg(ap,COORD_TCC_PTR);
	if ( coord_tcc_aware )
	{
	    if (!to_tcc)
	      return COORD_BAD_TCC;
	}
	else
	{
	    to_tcc    = NULL;
	    to_system = COORD_GCSCS;
	    dest_cell = GCS_ILLEGAL_CELL;
	}
	
	to_x = va_arg(ap,float64 *);
	to_y = va_arg(ap,float64 *);
	to_z = va_arg(ap,float64 *);
	break;

      case COORD_UTM_NE:
	to_zone     = va_arg(ap,int32 *);
	to_northing = va_arg(ap,float64 *);
	to_easting  = va_arg(ap,float64 *);
	to_z        = va_arg(ap,float64 *);
	break;

      case COORD_UTM_GRID:
	to_zone       = va_arg(ap,int32 *);
	if (to_zone == COORD_DEFAULT_ZONE)
	  to_zone     = &dummy_zone;
	to_resolution = va_arg(ap,int32);
	to_string     = va_arg(ap,char *);
	to_z          = va_arg(ap,float64 *);
	break;

      case COORD_LATLON:
	to_latitude    = va_arg(ap,float64 *);
	to_longitude   = va_arg(ap,float64 *);
	to_z           = va_arg(ap,float64 *);
	to_local_datum = va_arg(ap,int32);
	break;

      case COORD_CHARACTER_STRING:
	to_string = va_arg(ap,char *);
	break;

      default:
	return COORD_SYSTEM_UNKNOWN;
    }


    /* ---------------------------------------------------------------------
     * -------------  COORDINATE SYSTEM TRANSFORMATIONS  -------------------
     * ---------------------------------------------------------------------
     */
    switch(from_system)
    {
      /* -------------------------------------------------------------------
       * -----------   GCC      TO ANY OTHER COORDINATE SYSTEM  ------------
       * -------------------------------------------------------------------
       */
      case COORD_GCC:
	switch(to_system)
	{
	  case COORD_GCC:
	    *to_x = from_x;
	    *to_y = from_y;
	    *to_z = from_z;
	    return 0;

	  case COORD_GCS:
	    return !gcs_gcc3e_to_gcs(from_x, from_y, from_z,
				   to_cell, to_x, to_y, to_z);

	  case COORD_GCSCS:
	    return !gcs_gcc3e_to_gcscs(from_x, from_y, from_z,
				     dest_cell, to_x, to_y, to_z);

	  case COORD_TCC:
	    if (to_tcc->source == COORD_GCC)
	      coord_gcc_to_tcc(from_x, from_y, from_z, 
			       to_tcc, to_x, to_y, to_z);
	    else /* (to_tcc->source == COORD_UTM_NE) */
	      coord_gcc_to_tccf(from_x, from_y, from_z, 
				to_tcc, to_x, to_y, to_z);
	    
	    return 0;
	    
	  case COORD_UTM_NE:
	    /* First we translate to llz using the GCC datum, then
	     * we will have a lat/long with which we can find out what
	     * datum we should have used.
	     */
            gcs_gcc_to_utm(from_x, from_y, from_z,
                           to_zone, to_northing, to_easting, to_z);
 
	    return 0;

	  case COORD_UTM_GRID:
	    zone = *to_zone;

	    return (coord_convert(COORD_GCC,from_x, from_y, from_z,
				  COORD_UTM_NE,&zone, &northing, &easting, &z)
		    ||
		    coord_convert(COORD_UTM_NE, zone, northing, 
				  easting, z,
				  COORD_UTM_GRID, &zone, to_resolution, 
				  to_string, to_z));

	  case COORD_LATLON:
	    if ( gcs_gcc_to_llz(from_x, from_y, from_z,
				&latitude, &longitude, to_z, 
				(to_local_datum?DATUM_UNKNOWN:DATUM_WGS84
				 )) )
	    {
		*to_latitude = latitude;
		*to_longitude = longitude;
		return 0;
	    }
	    return 1;

	  case COORD_CHARACTER_STRING:
	    sprintf(to_string,"%.1f, %.1f, %.1f", from_x,from_y,from_z);
	    return 0;

	}
	break;

      /* -------------------------------------------------------------------
       * -----------   GCS      TO ANY OTHER COORDINATE SYSTEM  ------------
       * -------------------------------------------------------------------
       */
      case COORD_GCS:
	switch(to_system)
	{
	  case COORD_GCC:
	    return !gcs_gcs3e_to_gcc(from_cell, from_x, from_y, from_z, 
				   to_x, to_y, to_z);

	  case COORD_GCS:
	    return (!gcs_gcs3e_to_gcs(from_cell,from_x,from_y,from_z,
				    to_cell,to_x,to_y,to_z));

	  case COORD_GCSCS:
	    return !gcs_gcs3e_to_gcscs(from_cell, from_x, from_y, from_z,
				     dest_cell, to_x, to_y, to_z);

	  case COORD_TCC:
	    return (!gcs_gcs3e_to_gcc(from_cell, from_x, from_y, from_z,
				    &u, &v, &w) ||
		    coord_convert(COORD_GCC, u, v, w,
				  COORD_TCC, to_tcc,to_x, to_y, to_z));
	    
	  case COORD_UTM_NE:
	    return (!gcs_gcs3e_to_gcc(from_cell, from_x, from_y, from_z, 
				    &u, &v, &w) ||
		    coord_convert(COORD_GCC, u, v, w,
				  COORD_UTM_NE, to_zone, to_northing,
				  to_easting, to_z));
	    
	  case COORD_UTM_GRID:
	    zone = *to_zone;

	    return (coord_convert(COORD_GCS,from_cell, from_x, from_y, from_z,
				  COORD_UTM_NE,&zone, &northing, &easting, &z)
		    ||
		    coord_convert(COORD_UTM_NE, zone, floor(northing + 0.5),
				  floor(easting + 0.5), z,
				  COORD_UTM_GRID, &zone, to_resolution, 
				  to_string, to_z));
	  case COORD_LATLON:
	    return ( !gcs_gcs_to_llz(from_cell, from_x, from_y, from_z,
				     to_latitude, to_longitude, to_z, 
				     (to_local_datum?DATUM_UNKNOWN:
				      DATUM_WGS84)));

	  case COORD_CHARACTER_STRING:
	    sprintf(to_string,"cell:%05d  %d, %d, %d",
		    from_cell, ROUND(from_x),ROUND(from_y),ROUND(from_z));
	    return 0;
	}
	break;

      /* -------------------------------------------------------------------
       * -----------   TCC      TO ANY OTHER COORDINATE SYSTEM  ------------
       * -------------------------------------------------------------------
       */
      case COORD_TCC:
	switch(to_system)
	{
	  case COORD_GCC:
	    if (from_tcc->source == COORD_GCC)
	      coord_tcc_to_gcc(from_tcc, from_x, from_y, from_z,
			       to_x, to_y, to_z);
	    else /* (from_tcc->source == COORD_UTM_NE) */
	      coord_tccf_to_gcc(from_tcc, from_x, from_y, from_z,
				to_x, to_y, to_z);
	    return 0;

	  case COORD_GCS:
	    return (coord_convert(COORD_TCC, from_tcc, from_x, from_y, from_z,
				  COORD_GCC, &u, &v, &w) ||
		    !gcs_gcc3e_to_gcs(u, v, w, to_cell, to_x, to_y, to_z));

	  case COORD_GCSCS:
	    return (coord_convert(COORD_TCC, from_tcc, from_x, from_y, from_z,
				  COORD_GCC, &u, &v, &w) ||
		    !gcs_gcc3e_to_gcscs(u, v, w, dest_cell, to_x, to_y, to_z));

	  case COORD_TCC:
	    if (to_tcc == from_tcc)
	    {
		*to_x = from_x;
		*to_y = from_y;
		*to_z = from_z;
		return 0;
	    }
	    else if ((from_tcc->source == COORD_GCC) ||
		     (to_tcc->source == COORD_GCC) ||
		     (from_tcc->datum != to_tcc->datum) ||
		     (from_tcc->zone != to_tcc->zone))
	    {
		return (coord_convert(COORD_TCC,from_tcc,from_x,from_y,from_z,
				      COORD_GCC,&x, &y, &z) || 
			coord_convert(COORD_GCC,x, y, z,
				      COORD_TCC,to_tcc, to_x, to_y, to_z));
		
	    }
	    else /* TCC's differ only in origin */
	    {
		*to_x = from_x + from_tcc->easting - to_tcc->easting;
		*to_y = from_y + from_tcc->northing - to_tcc->northing;
		*to_z = from_z;
		return 0;
	    }

	  case COORD_UTM_NE:
	    if (!*to_zone)
	      *to_zone = from_tcc->zone;

	    if ((from_tcc->source == COORD_UTM_NE) &&
		(from_tcc->zone == *to_zone))
	    {
		/* Compute the northing and easting, assuming the datum
		 * is right.
		 */
		*to_easting = from_tcc->easting + from_x;
		*to_northing = from_tcc->northing + from_y;

		/* See if that is the preferred datum */
		if (from_tcc->datum ==
		    coord_lookup_datum_by_ne(from_tcc->zone,
					     *to_northing, *to_easting))
		{
		    *to_z = from_z;
		    return 0;
		}
	    }

	    /* There are differences between the TCC and the preferred
	     * UTM datum here, or the TCC is not UTM-based.  Go through GCC.
	     */
	    return(coord_convert(COORD_TCC,from_tcc, from_x, from_y, from_z,
				    COORD_GCC,&x, &y, &z) ||
		   coord_convert(COORD_GCC, x, y, z,
				 COORD_UTM_NE, to_zone, to_northing, 
				 to_easting, to_z));

	  case COORD_UTM_GRID:
	    zone = *to_zone;

	    return (coord_convert(COORD_TCC,from_tcc, from_x, from_y, from_z,
				  COORD_UTM_NE,&zone, &northing, &easting, &z)
		    ||
		    coord_convert(COORD_UTM_NE, zone, northing, easting, z,
				  COORD_UTM_GRID, &zone, to_resolution, 
				  to_string, to_z));

	  case COORD_LATLON:
	    return (coord_convert(COORD_TCC,from_tcc, from_x, from_y, from_z,
				  COORD_GCC,&x, &y, &z) ||
		    coord_convert(COORD_GCC,x, y, z,
				  COORD_LATLON,to_latitude, to_longitude, to_z,
				  to_local_datum));

	  case COORD_CHARACTER_STRING:
	    sprintf(to_string,"%d, %d, %d",
		    ROUND(from_x),ROUND(from_y),ROUND(from_z));
	    return 0;

	}
	break;

      /* -------------------------------------------------------------------
       * -----------   UTM NE   TO ANY OTHER COORDINATE SYSTEM  ------------
       * -------------------------------------------------------------------
       */
      case COORD_UTM_NE:
	switch(to_system)
	{

	  case COORD_GCC:
	    datum = coord_lookup_datum_by_ne(from_zone,
					     from_northing, from_easting);
	    coord_utm_to_ll(datum, from_northing, from_easting, from_zone,
			    &latitude, &longitude);
	    gcs_llz_to_gcc(latitude, longitude, from_z, datum, 
			   to_x, to_y, to_z);
	    return 0;


	  case COORD_GCS:
	    return (coord_convert(COORD_UTM_NE, from_zone, from_northing, 
				  from_easting, from_z,
				  COORD_GCC, &u, &v, &w) ||
		    !gcs_gcc3e_to_gcs(u, v, w, to_cell, to_x, to_y, to_z));
	    

	  case COORD_GCSCS:
	    return (coord_convert(COORD_UTM_NE, from_zone, from_northing, 
				  from_easting, from_z,
				  COORD_GCC, &u, &v, &w) ||
		    !gcs_gcc3e_to_gcscs(u, v, w, dest_cell, to_x, to_y, to_z));

	  case COORD_TCC:
	    if (!from_zone)
	      from_zone = to_tcc->zone;

	    if ((to_tcc->source == COORD_UTM_NE) &&
		(to_tcc->zone == from_zone) &&
		(to_tcc->datum ==
		 coord_lookup_datum_by_ne(from_zone,
					  from_northing, from_easting)))
	    {
		*to_x = from_easting - to_tcc->easting;
		*to_y = from_northing - to_tcc->northing;
		*to_z = from_z;
		return 0;
	    }
	    else /* (to_system->source == COORD_GCC) or there are
		  * zone/datum discrepancies.
		  */
	    {
		return (coord_convert(COORD_UTM_NE,from_zone,from_northing, 
				      from_easting, from_z,
				      COORD_GCC,&x, &y, &z) ||
			coord_convert(COORD_GCC, x, y, z,
				      COORD_TCC, to_tcc, to_x, to_y, to_z));
	    }

	  case COORD_UTM_NE:
	    *to_z = from_z;
	    coord_change_utm_zone(from_zone, from_northing, from_easting,
				  to_zone, to_northing, to_easting);
	    return 0;

	  case COORD_UTM_GRID:
	    *to_z = from_z;
	    coord_utm_to_grid(from_zone, from_northing, from_easting,
			      to_zone, to_resolution, to_string);
	    return 0;

	  case COORD_LATLON:
	    datum = coord_lookup_datum_by_ne(from_zone,
					     from_northing, from_easting);
	    coord_utm_to_ll(datum, from_northing, from_easting, from_zone,
			    &latitude, &longitude);
	    if ((datum != DATUM_WGS84) && !to_local_datum)
	    {
		/* Convert to local_datum is requested */
		gcs_llz_to_gcc(latitude, longitude, from_z, datum,
			       &x, &y, &z);
		/* Convert to COORD_GCC/WGS-84 datum */
		gcs_gcc_to_llz(x, y, z,
			       to_latitude, to_longitude, to_z, DATUM_WGS84);
	    }
	    else
	    {
		*to_latitude = latitude;
		*to_longitude = longitude;
		*to_z = from_z;
	    }
	    return 0;

	  case COORD_CHARACTER_STRING:
	    sprintf(to_string,"N%d E%d [%d] ^%d",
		    ROUND(from_northing),ROUND(from_easting),from_zone,
		    ROUND(from_z));
	    return 0;
	}
	break;

      /* -------------------------------------------------------------------
       * -----------   UTM GRID TO ANY OTHER COORDINATE SYSTEM  ------------
       * -------------------------------------------------------------------
       */
      case COORD_UTM_GRID:
	switch(to_system)
	{
	  case COORD_GCC:
	    zone = from_zone;
	    return (coord_convert(COORD_UTM_GRID,from_zone, from_resolution, 
				  from_string,from_z,
				  COORD_UTM_NE,&zone, &northing, &easting, &z)
		    || 
		    coord_convert(COORD_UTM_NE, zone, northing, easting, z,
				  COORD_GCC, to_x, to_y, to_z));

	  case COORD_GCS:
	    return (coord_convert(COORD_UTM_GRID, from_zone, from_resolution, 
				  from_string, from_z, 
				  COORD_UTM_NE, &zone, &northing, &easting, 
				  &z) ||
		    coord_convert(COORD_UTM_NE, zone, floor(northing + 0.5),
				  floor(easting + 0.5), z, COORD_GCS, to_cell,
				  to_x, to_y, to_z));

	  case COORD_GCSCS:
	    return (coord_convert(COORD_UTM_GRID, from_zone, from_resolution,
				  from_string, from_z, 
				  COORD_GCC, &u, &v, &w) ||
		    !gcs_gcc3e_to_gcscs(u, v, w, dest_cell, to_x, to_y, to_z));

	  case COORD_TCC:
	    zone = to_tcc->zone;
	    return (coord_convert(COORD_UTM_GRID,from_zone, from_resolution, 
				  from_string,from_z,
				  COORD_UTM_NE,&zone, &northing, &easting, &z)
		    ||
		    coord_convert(COORD_UTM_NE,zone, northing, easting, z,
				  COORD_TCC,to_tcc, to_x, to_y, to_z));

	  case COORD_UTM_NE:
	    *to_z = from_z;
	    return coord_grid_to_utm(from_zone, from_string,
				     to_zone, to_northing, to_easting);

	  case COORD_UTM_GRID:
	    *to_z = from_z;
	    strcpy(to_string, from_string);
	    return 0;

	  case COORD_LATLON:
	    zone = from_zone;
	    return (coord_convert(COORD_UTM_GRID,from_zone, from_resolution, 
				  from_string, from_z,
				  COORD_UTM_NE,&zone, &northing, &easting, &z)
		    ||
		    coord_convert(COORD_UTM_NE, zone, northing, easting, z,
				  COORD_LATLON,to_latitude, to_longitude, to_z,
				  to_local_datum));

	  case COORD_CHARACTER_STRING:
	    strcpy(to_string, from_string);
	    return 0;

	}
	break;

      /* -------------------------------------------------------------------
       * -----------   LAT/LONG TO ANY OTHER COORDINATE SYSTEM  ------------
       * -------------------------------------------------------------------
       */
      case COORD_LATLON:
	switch(to_system)
	{
	  case COORD_GCC:
	    gcs_llz_to_gcc(from_latitude, from_longitude, from_z, 
			   (int32)(from_local_datum?
				   coord_lookup_datum_by_ll(from_latitude,
							    from_longitude):
			    DATUM_WGS84), to_x, to_y, to_z);
	    return 0;

	  case COORD_GCS:
	    return (coord_convert(COORD_LATLON,from_latitude, from_longitude,
				  from_z,from_local_datum,COORD_GCC,&x,&y,&z)
		    || !gcs_gcc3e_to_gcs(x,y,z,to_cell,to_x,to_y,to_z));

	  case COORD_GCSCS:
	    return (coord_convert(COORD_LATLON,from_latitude, from_longitude,
				  from_z,from_local_datum,COORD_GCC,&x,&y,&z)
		    || !gcs_gcc3e_to_gcscs(x,y,z,dest_cell,to_x,to_y,to_z));

	  case COORD_TCC:
	    return (coord_convert(COORD_LATLON,from_latitude, from_longitude, 
				  from_z, from_local_datum,
				  COORD_GCC,&x, &y, &z) ||
		    coord_convert(COORD_GCC, x, y, z,
				  COORD_TCC,to_tcc, to_x, to_y, to_z));

	  case COORD_UTM_NE:
	    *to_z = from_z;

	    datum = coord_lookup_datum_by_ll(from_latitude, from_longitude);

	    if ((datum != DATUM_WGS84) && !from_local_datum)
	    {
		/* Translate the lat/long from the GCC datum to the
		 * local datum.
		 */
		gcs_llz_to_gcc(from_latitude, from_longitude, from_z, datum,
			       &x, &y, &z);
		gcs_gcc_to_llz(x, y, z,
			       &latitude, &longitude, to_z, DATUM_WGS84);
	    }
	    else
	    {
		latitude = from_latitude;
		longitude = from_longitude;
	    }
	    coord_ll_to_utm(datum, latitude, longitude,
			    to_zone, to_northing, to_easting);
	    return 0;

	  case COORD_UTM_GRID:
	    return (coord_convert(COORD_LATLON,from_latitude, from_longitude, 
				  from_z,from_local_datum, COORD_UTM_NE, 
				  to_zone, &northing, &easting, &z) ||
		    coord_convert(COORD_UTM_NE, *to_zone, northing, easting, z,
				  COORD_UTM_GRID, to_zone, to_resolution, 
				  to_string, to_z));

	  case COORD_LATLON:
	    if (from_local_datum == to_local_datum)
	    {
		*to_latitude = from_latitude;
		*to_longitude = from_longitude;
		*to_z = from_z;
		return 0;
	    }
	    else
	    {
		gcs_llz_to_gcc(from_latitude, from_longitude, from_z,
			       (int32)(from_local_datum?
				       coord_lookup_datum_by_ll(from_latitude, 
							       from_longitude):
			       DATUM_WGS84), &x, &y, &z);
		
		gcs_gcc_to_llz(x, y, z,
			       to_latitude, to_longitude, to_z, 
			       (to_local_datum?DATUM_UNKNOWN:
				DATUM_WGS84));
		return 0;
	    }
	    
	    
	  case COORD_CHARACTER_STRING:
	    sprintf(to_string, "%s ^%d",
		    coord_format_latlon(from_latitude, from_longitude),
		    ROUND(from_z));
	    return 0;
	}
	break;
    }

    /* If we got here, one of the switches must have fallen through */
    return COORD_SYSTEM_UNKNOWN;
}

char *coord_error(int32 diag)
{
    switch (diag)
    {
      case 0:				return "No Error";
      case COORD_SYNTAX:		return "UTM Syntax Error";
      case COORD_TOO_LONG:		return "UTM Too Long";
      case COORD_ZONE_UNKNOWN:		return "UTM in Unrecognized Zone";
      case COORD_GRID_UNKNOWN:		return "UTM Grid not Found in Zone";
      case COORD_BAD_TCC:		return "Bad TCC pointer";
      case COORD_SYSTEM_UNKNOWN:	return "Unknown Coordinate System";
      default:				return "Unknown Error";
    }
}
