#ifndef lint
static char rcsid [] = "$RCSfile: coord_trans.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: coord_trans.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to do messy coordinate system transforms             *
*       Created: Mon Aug 17 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Updated Mon Sep 11 1995
*                                                                           *
****************************************************************************/

#include "libcoord_local.h"
#include <libvecmat.h>
#include <math.h>
#include <libgcs.h>

/* DEFINES */
#define k0            0.9996                 /* scale on central meridian */
static void 
calc_ellipse_params(
		    int32    datum,
		    float64 *a,
		    float64 *e_sq,
		    float64 *ep_sq,
		    float64 *ef
		    );

void 
coord_ll_to_utm(int32 datum, float64 lat, float64 lon, int32 *zone,
		float64 *northing, float64 *easting)
                   
                 
                 
                   /* Forced, if non-zero */
                      
                     
{
    /* Equations are taken from "Map Projections - A Working Manual" by
     * John Snyder, USGS Professional Paper 1395, Page 61.
     */

    float64 phai;		/* input latitude in degrees               */
    float64 lamda;		/* input longitude in degrees              */
    float64 lamda0;		/* longitude of central meridian           */
    float64 M,N,T,C,A;		/* basic parameters used for calcualtaions */
    float64 A2, A3, A4, A5, A6;	/* 'D' raised to the numbered power        */
    float64 T2;			/* 'T' raised to the numbered power        */
    float64 C2;			/* 'C' raised to the numbered power        */
    float64 P1, P2, P3, P4, P5;	/* miscellaneous terms                     */
    float64 cos_phai;		/* cosine of latitude                      */
    float64 cos_2phai;		/* cosine of 2*latitude                    */
    float64 sin_phai;		/* sine of latitude                        */
    float64 tan_phai;		/* tangent of latitude                     */
    float64 radius;		/* Equatorial radius of ellipsoid          */
    float64 e_sq;		/* eccentricity squared, see equations below*/
    float64 ep_sq;		/* e-prime squared, see equations below    */
    float64 ef;			/* e factor, see equations below           */
    int32  izone;		/* utm zone                   */
    float64 e_sq_2, e_sq_3, e_sq_4;

    /* Calculate utm zone */
    if (*zone)
      izone = *zone;
    else
      izone = COORD_LATLONG_TO_GRID_ZONE(lat, lon);

    calc_ellipse_params(datum, &radius, &e_sq, &ep_sq, &ef);

    e_sq_2 = e_sq*e_sq;
    e_sq_3 = e_sq_2*e_sq;
    e_sq_4 = e_sq_3*e_sq;

    P1 = 1.0 - (e_sq/4.0) - (3.0*e_sq_2/64.0) - (5.0*e_sq_3/256.0)
      - (175*e_sq_4/16384.0);
    P2 = (3.0*e_sq/8.0) + (3.0*e_sq_2/32.0) + (45.0*e_sq_3/1024.0)
      - (1365*e_sq_4/32768); 
    P3 = (15.0*e_sq_2/256.0) + (45.0*e_sq_3/1024.0) 
      - (1155*e_sq_4/32768.0);
    P4 = 35.0*e_sq_3/3072.0 - (1435.0*e_sq_4/98304);

    P5 = -(315*e_sq_4/131072);

    /* Calculate longitude of the central meridian */
    if ( izone < 31 )
      lamda0 = (float64) ((izone * 6) - 183);
    else
      lamda0 = (float64) (((izone - 31) * 6) + 3);

    /* convert latitudes and longitude to radians */
    phai = DEG_TO_RAD(lat);
    lamda = DEG_TO_RAD(lon);
    lamda0 = DEG_TO_RAD(lamda0);

    /* calculate the basic parameters used in calculations */
    cos_phai = cos(phai);
    sin_phai = sin(phai);
    tan_phai = tan(phai);
    C = ep_sq * cos_phai * cos_phai;
    T = tan_phai * tan_phai;
    N = radius / sqrt( 1.0 - (e_sq * sin_phai * sin_phai));
    A = (lamda - lamda0) * cos_phai;
#ifdef OLD_VERSION
    M = radius * (P1*phai - P2*sin(2.0*phai) + P3*sin(4.0*phai) -
		  P4*sin(6.0*phai) + P5*sin(8.0*phai));
#else
    cos_2phai = cos(2.0*phai);
    M = radius *
      (P1*phai + (sin(2.0*phai) *
		  ((P4-P2) +
		   cos_2phai * ((2.0*P3-4.0*P5) +
				cos_2phai * (-4.0*P4 +
					     cos_2phai * 8.0*P5)))));
#endif

    /* calculate parameters derived from above */
    T2 = T * T;
    C2 = C * C;
    A2 = A * A;
    A3 = A2 * A;
    A4 = A3 * A;
    A5 = A4 * A;
    A6 = A5 * A;

    /* calculate the easting */
    P1 = (1 - T + C);
    P2 = (5.0 - 18.0*T + T2 + 72.0*C - 58.0*ep_sq);
    *easting = k0 * N * (A + (P1*A3/6.0) + (P2*A5/120.0));

    /* calculate the northing */
    P1 = (5.0 - T + 9.0*C + 4.0*C2);
    P2 = (61.0 - 58.0*T + T2 + 600.0*C - 330.0*ep_sq);
    *northing = k0 * (M + N * tan_phai * 
		      (A2/2.0 + (P1*A4/24.0) + (P2*A6/720.0)));

    /* add false northing and easting */
    if (lat < 0)
      *northing = -( (*northing) + 10000000.0);
    *easting += 500000.0;
    *zone = izone;
}

void 
coord_utm_to_ll(int32 datum, float64 northing, float64 easting, int32 zone,
		float64 *lat, float64 *lon)
{
    float64 phai;               /* resulting latitude in radians           */
    float64 lamda;              /* resutling longitude in radians          */
    float64 phai1;              /* footprint latitude                      */
    float64 lamda0;             /* longitude of central meridian           */
    float64 mu,M,N,R,T,C,D;     /* basic parameters used for calcualtaions */
    float64 D2, D3, D4, D5, D6; /* 'D' raised to the numbered power        */
    float64 T2;                 /* 'T' raised to the numbered power        */
    float64 C2;                 /* 'C' raised to the numbered power        */
    float64 P1, P2, P3, P4, P5; /* miscellaneous terms                     */
    float64 cos_2mu;
    float64 cos_phai1;          /* cosine of latitude                      */
    float64 sin_phai1;          /* sine of latitude                        */
    float64 tan_phai1;          /* tangent of latitude                     */
    float64 radius;             /* Equatorial radius of ellipsoid          */
    float64 e_sq;               /* eccentricity squared                    */
    float64 ep_sq;              /* e-prime squared                         */
    float64 ef;                 /* e factor                                */
    float64 adj_easting;        /* easting minus false easting             */
    float64 adj_northing;       /* northing minus false northing           */
    float64 ef_4, ef_3, e_sq_2, e_sq_3;
   
    /* Equations are taken from "Map Projections - A Working Manual" by
     * John Snyder, USGS Professional Paper 1395, Pages 63 and 64.
     */

    calc_ellipse_params(datum, &radius, &e_sq, &ep_sq, &ef);

    ef_3 = ef*ef*ef;
    ef_4 = ef_3*ef;
    e_sq_2 = e_sq*e_sq;
    e_sq_3 = e_sq_2*e_sq;

    P1 =  1.0 - (e_sq/4.0) - (3.0*e_sq_2/64.0) - (5.0*e_sq_3/256.0);
    P2 = (3.0 * ef/2.0) - (27.0 * ef_3/32.0);
    P3 = (21.0 * ef*ef / 16.0 ) - (55.0 * ef_4 / 32.0); 
    P4 = 151.0 * ef_3 / 96.0;
    P5 = 1097.0 * ef_4 / 512.0;

    adj_easting = easting - 500000.0;
    if (northing < 0)
      adj_northing = -northing - 10000000.0;
    else
      adj_northing = northing;

    /* Calculate longitude of the central meridian */
    if ( zone < 31 )
      lamda0 = ( float64) ( ( zone * 6 ) - 183);
    else
      lamda0 =( float64)( ( ( zone - 31 ) * 6 ) + 3 );

    /* convert to radians */
    lamda0 = DEG_TO_RAD(lamda0);

    /* calculate true distance along central meridian from equator to phi */
    M = adj_northing / k0;

    /* calculate the footprint latitude, phai1 */
    mu = M / (radius * P1);
#ifdef OLD_VERSION
    phai1 = mu + P2*sin(2.0*mu) + P3*sin(4.0*mu) + P4*sin(6.0*mu) +
      P5*sin(8.0*mu);
#else
    cos_2mu = cos(2.0*mu);
    phai1 = mu + (sin(2.0*mu) *
		  ((P2-P4) +
		   cos_2mu * ((2.0*P3-4.0*P5) +
			      cos_2mu * (4.0*P4 +
					 cos_2mu * 8.0*P5))));
#endif

    /* calculate the basic parameters used calculations */
    cos_phai1 = cos(phai1);
    sin_phai1 = sin(phai1);
    tan_phai1 = tan(phai1);
    C = ep_sq * cos_phai1 * cos_phai1;
    T = tan_phai1 * tan_phai1;
    N = radius / sqrt( 1.0 - (e_sq * sin_phai1 * sin_phai1));
    P1 = sin_phai1 * sin_phai1;
    R = ( radius * ( 1 - e_sq ) ) /
      ( (1 - e_sq*P1)*sqrt(1 - e_sq*P1) );
    D =  adj_easting / ( N * k0);

    /* calculate various derived parameters from squaring, cubing, etc. */
    D2 = D * D ;
    D3 = D2 * D;
    D4 = D3 * D;
    D5 = D4 * D;
    D6 = D5 * D;
    C2 =  C * C;
    T2 =  T * T;

    /* calculate the latitude, phai */
    P1 =   D2/2.0;
    P2 = ( D4/24.0  ) * ( 5.0 + 3.0*T + 10.0*C - 4.0*C2 - 9.0*ep_sq) ;
    P3 = ( D6/720.0 ) * (61.0 + 90.0*T + 298.0*C + 45.0*T2 - 
			 252.0*ep_sq - 3.0*C2 );
    phai = phai1 - ( N * tan_phai1 / R ) * ( P1 - P2 + P3 );

    /* calculate the longitude, lambda */
    P1 = ( 1.0 + 2.0*T + C ) * D3 / 6.0;
    P2 = ( 5.0 - 2.0*C + 28.0*T - 3.0*C2 + 8.0*ep_sq + 24.0*T2 ) * D5 / 120.0;
    lamda = lamda0 + ((D - P1 + P2) / cos_phai1);

    /* convert to degrees and fraction thereof */
    *lat = RAD_TO_DEG(phai);
    *lon = RAD_TO_DEG(lamda);
}

void 
coord_tccf_to_gcc(COORD_TCC_PTR tcc, float64 x, float64 y, float64 z, 
		  float64 *u, float64 *v, float64 *w)
{
    float64 lat,lng;
 
    /* Convert to UTM in TCC's datum */
    x += tcc->easting;
    y += tcc->northing;
    
    /* Take it to LL, using the TCC's datum */
    coord_utm_to_ll(tcc->datum, y, x, tcc->zone, &lat, &lng);
    
    /* Now take it to GCC, which will switch datums */
    gcs_llz_to_gcc(lat, lng, z, tcc->datum, u, v, w);    
}

void 
coord_gcc_to_tccf(float64 x, float64 y, float64 z, COORD_TCC_PTR tcc,
		  float64 *x_out, float64 *y_out, float64 *z_out)
{
    float64 lat,lng;
    
    /* First go from GCC to LLZ, and take it into the
     * desired datum.
     */
    gcs_gcc_to_llz(x, y, z, &lat, &lng, z_out, tcc->datum );
    
    /* Now take it to UTM, same datum, required zone */
    coord_ll_to_utm(tcc->datum, lat, lng, &tcc->zone, y_out, x_out);
    
    /* Now shift to TCC origin */
    *x_out -= tcc->easting;
    *y_out -= tcc->northing;
}

void 
coord_gcc_to_tcc(float64 x_in, float64 y_in, float64 z_in, COORD_TCC_PTR tcc,
		 float64 *x_out, float64 *y_out, float64 *z_out)
{
    float64 result[XYZ];

    /* This routine *assumes* the TCC is GCC derived */

    vmat3e_mat_vec_mul64(tcc->t_to_g_rot,
			 x_in - tcc->center_in_gcc[X],
			 y_in - tcc->center_in_gcc[Y],
			 z_in - tcc->center_in_gcc[Z],
			 result);

    *x_out = result[X] + tcc->half_width; 
    *y_out = result[Y] + tcc->half_height;
    *z_out = result[Z];
}

void 
coord_tcc_to_gcc(COORD_TCC_PTR tcc, float64 x_in, float64 y_in, float64 z_in,
		 float64 *x_out, float64 *y_out, float64 *z_out)
{
    float64 result[3];

    /* This routine *assumes* the TCC is GCC derived */

    result[X] = x_in - tcc->half_width;
    result[Y] = y_in - tcc->half_height;
    result[Z] = z_in;

    vmat3_vec_mat_mul64(result, tcc->t_to_g_rot, result);
    VMAT3_VEC_ADD(result, tcc->center_in_gcc, result);

    *x_out = result[X];
    *y_out = result[Y];
    *z_out = result[Z];
}

void 
coord_utm_to_grid(int32 from_zone, float64 from_northing, float64 from_easting,
		  int32 *to_zone, int32 to_resolution, char *to_string)
{
    float64 to_northing;
    float64 to_easting;

    coord_change_utm_zone(from_zone, from_northing, from_easting,
			  to_zone, &to_northing, &to_easting);

    coord_lookup_utm(*to_zone, to_northing, to_easting,
		     to_resolution, to_string);
}

int32 
coord_grid_to_utm(int32 from_zone, char *from_string, int32 *to_zone,
		  float64 *to_northing, float64 *to_easting)
{
    int32 zone;
    float64 northing;
    float64 easting;
    int32 ret;
    COORD_TCC_PTR tcc;

    if (!coord_tcc_list)
    {
	if (ret = coord_parse_utm(0, 0, from_string,
				  &zone, &northing, &easting))
	  return ret;
    }
    else
    {
	/* Try the zone data from each tcc */
	for (tcc=coord_tcc_list;tcc;tcc=tcc->next)
	{
	    if (from_zone == COORD_DEFAULT_ZONE)
	      zone = tcc->zone;
	    else
	      zone = from_zone;

	    if (!(ret = coord_parse_utm(zone, tcc->zone_letter, from_string,
					&zone, &northing, &easting)))
	      break;
	}

	/* If 'ret' is set, we had errors with every TCC we tried */
	if (ret)
	  return ret;
    }

    *to_zone = zone;
    coord_change_utm_zone(zone, northing, easting,
			  to_zone, to_northing, to_easting);
    return 0;
}

void 
coord_change_utm_zone(int32 from_zone, float64 from_northing,
		      float64 from_easting, int32 *to_zone, 
		      float64 *to_northing, float64 *to_easting)
{
    float64 latitude;
    float64 longitude;
    int32 from_datum, to_datum;

    /* Check the simplest case */
    if ((*to_zone == COORD_DEFAULT_ZONE) ||
	(from_zone == *to_zone))
    {
	*to_zone = from_zone;
	*to_northing = from_northing;
	*to_easting = from_easting;
	return;
    }

    from_datum = coord_lookup_datum_by_ne(from_zone,
					  from_northing, from_easting);
    coord_utm_to_ll(from_datum, from_northing, from_easting, from_zone,
		    &latitude, &longitude);

    to_datum = coord_lookup_datum_by_ll(latitude, longitude);
    coord_ll_to_utm(to_datum, latitude, longitude,
		    to_zone, to_northing, to_easting);
}


static void 
calc_ellipse_params(int32 datum, float64 *a, float64 *e_sq, float64 *ep_sq,
		    float64 *ef)
{
    float64 b;
    float64 f;

    *a = coord_get_major_axis(datum);
    b = coord_get_minor_axis(datum);

    f = 1 - (b/(*a));

    *e_sq   = (2.0 - f) * f;
    *ep_sq  = *e_sq / (1.0 - *e_sq);
    *ef     = f / (2.0 - f);
}
