#ifndef lint
static char rcsid [] = "$RCSfile: gridtest.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: gridtest.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Program for testing grid-line generation                  *
*       Created: Thu Sep 10 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcoordinates.h"
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#define MIN_X   0
#define MIN_Y   0
#define MAX_X   360000
#define MAX_Y   290000
#define SPACING 10000

/*SAKI*/
#define SOURCE    COORD_GCC
#define NORTHING  3105000.0
#define EASTING   508000.0
#define ZNUM      38
#define ZLET      'R'
#define DATUM     3

#define MAX_SEGS 10000
#define MAX_LBLS 1000

#define BORDER 10

COORD_TCC_PTR tcc;

draw_grid(w, ignored, event)
    Widget w;
    int32 ignored;
    XExposeEvent *event;
{
    int32               n_thin_segments;
    float32             thin_segments[MAX_SEGS*4];
    int32               n_thick_segments;
    float32             thick_segments[MAX_SEGS*4];
    int32               n_labels;
    COORD_GRID_LABEL    labels[MAX_LBLS];
    int16               pts[MAX_SEGS*4];
    int32               i;
    float64             sx,sy,scale_ppm;
    XWindowAttributes   att;
    static GC           thin_gc=0, thick_gc;

    if (event->count)
      return;

    XClearWindow(XtDisplay(w), XtWindow(w));

    if (!thin_gc)
    {
	thin_gc = XCreateGC(XtDisplay(w), XtWindow(w), 0, NULL);
	thick_gc = XCreateGC(XtDisplay(w), XtWindow(w), 0, NULL);
	XSetLineAttributes(XtDisplay(w), thick_gc, 2, LineSolid,
			   CapButt, JoinMiter);
    }

    XGetWindowAttributes(XtDisplay(w), XtWindow(w), &att);
    att.width -= 2*BORDER;
    att.height -= 2*BORDER;

    coord_generate_grid(tcc,
			MIN_X, MAX_X, MIN_Y, MAX_Y, SPACING, 1,
			COORD_LABEL_ALL,
			MAX_SEGS, MAX_SEGS, MAX_LBLS,
			&n_thin_segments, thin_segments,
			&n_thick_segments, thick_segments,
			&n_labels, labels);

    sx = (float64)att.width / (double)(MAX_X - MIN_X);
    sy = (float64)att.height / (double)(MAX_Y - MIN_Y);
    if (sx > sy) scale_ppm = sy;
    else scale_ppm = sx;

    for (i=0;i<n_thin_segments*4;i+=2)
    {
	pts[i] = (thin_segments[i] - MIN_X) * scale_ppm + BORDER;
	pts[i+1] =
	  att.height - (thin_segments[i+1] - MIN_Y) * scale_ppm + BORDER;
    }

    XDrawSegments(XtDisplay(w), XtWindow(w), thin_gc, pts, n_thin_segments);

    for (i=0;i<n_thick_segments*4;i+=2)
    {
	pts[i] = (thick_segments[i] - MIN_X) * scale_ppm + BORDER;
	pts[i+1] =
	  att.height - (thick_segments[i+1] - MIN_Y) * scale_ppm + BORDER;
    }

    XDrawSegments(XtDisplay(w), XtWindow(w), thick_gc,
		  pts, n_thick_segments);

    for (i=0;i<n_labels;i++)
    {
	int16 x = (labels[i].x - MIN_X) * scale_ppm + BORDER;
	int16 y = att.height - (labels[i].y - MIN_Y) * scale_ppm + BORDER;

	XDrawString(XtDisplay(w), XtWindow(w), thin_gc,
		    x, y, labels[i].txt, strlen(labels[i].txt));
    }

    /* Label the center */
    {
	int16 x = ((MAX_X + MIN_X)/2) * scale_ppm + BORDER;
	int16 y = att.height - ((MAX_Y + MIN_Y)/2) * scale_ppm + BORDER;

	XDrawString(XtDisplay(w), XtWindow(w), thin_gc,
		    x, y, "C", 1);
    }
}

main(argc,argv)
    int32 argc;
    char *argv[];
{
    int32             ret;
    Widget            toplevel, area;
    int32             hold_argc = argc;
    XWindowAttributes att;
    extern float64 atof();

    if (coord_init(".", 0) != 0)
      exit(0);

    tcc = coord_define_tcc(SOURCE,
			   NORTHING, EASTING, ZNUM, ZLET, DATUM,
			   MAX_X, MAX_Y);

    toplevel = XtInitialize(argv[0], "Grids",
			    NULL, 0, &hold_argc, argv);

    XtRealizeWidget(toplevel);

    att.bit_gravity = ForgetGravity;
    XChangeWindowAttributes(XtDisplay(toplevel), XtWindow(toplevel),
			    CWBitGravity, &att);

    XtAddEventHandler(toplevel, ExposureMask, FALSE, draw_grid, 0);

    XtMainLoop();
}
