/*
 * $RCSfile: libcoord_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libcoord_local.h                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libcoordinates                    *
*       Created: Fri Aug 14 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Updated Mon Sep 11 1995
*                                                                           *
****************************************************************************/

#include "libcoordinates.h"

/*
 * Private data structures.
 */

struct coord_tcc
{
    COORD_SYSTEM      source; /* GCC or UTM_NE */
    int32             datum;
    int32             zone;
    char              zone_letter;
    float64           northing; /* of center, NOT SW corner */
    float64           easting;  /* of center, NOT SW corner */
    float64           half_width;
    float64           half_height;
    float64           t_to_g_rot[3][3];
    float64           center_in_gcc[3];
    struct coord_tcc *next;
};

extern COORD_TCC_PTR coord_tcc_list;

/*
 * Data structures for interpreting libreader data file.
 */

typedef struct coord_letter_list
{
    int32   length;
    char   *letters[1];
} COORD_LETTER_LIST;
#define NUM_LETTERS(list) \
  ((list).length-sizeof(COORD_LETTER_LIST)/sizeof(int32)+1)

typedef struct coord_letter_range
{
    int32   min_northing;
    int32   max_northing;
    char   *letter;
} COORD_LETTER_RANGE;

typedef struct coord_letter_table
{
    int32                length;
    COORD_LETTER_RANGE   ranges[1];
} COORD_LETTER_TABLE;
#define NUM_RANGES(list) \
  (((list).length-sizeof(COORD_LETTER_LIST)/sizeof(int32))/3+1)

typedef struct coord_grid_box
{
    int32                always_12;
    int32                min_northing;
    int32                max_northing;
    int32                min_easting;
    int32                max_easting;
    float32              min_latitude;
    float32              max_latitude;
    float32              min_longitude;
    float32              max_longitude;
    int32                datum;
    COORD_LETTER_LIST   *n_letters;
    COORD_LETTER_LIST   *e_letters;
} COORD_GRID_BOX;

typedef struct coord_zone_table
{
    int32                 length;
    int32                 zone_number;
    COORD_LETTER_TABLE   *z_letters;
    COORD_GRID_BOX       *boxes[1];
} COORD_ZONE_TABLE;
#define NUM_BOXES(table) \
  ((table).length-sizeof(COORD_ZONE_TABLE)/sizeof(int32)+1)

typedef struct coord_utm_table
{
    int32             length;
    char             *tag;
    COORD_ZONE_TABLE *zones[1];
} COORD_UTM_TABLE;
#define NUM_ZONES(table) \
  ((table).length-sizeof(COORD_UTM_TABLE)/sizeof(int32)+1)

typedef struct coord_molodenskiy_xform
{
    int32   length;
    int32   datum;
    char   *description;
    float32 dx;
    float32 dy;
    float32 dz;
    char   *major_axis;
    char   *minor_axis;
} COORD_MOLODENSKIY_XFORM;

typedef struct coord_datum_table
{
    int32                      length;
    char                      *tag;
    COORD_MOLODENSKIY_XFORM   *datums[1];
} COORD_DATUM_TABLE;
#define NUM_DATUMS(table) \
  ((table).length-sizeof(COORD_DATUM_TABLE)/sizeof(int32)+1)

/* Private prototypes */

extern void 
coord_utm_to_ll (     
		    int32    datum,
		    float64  northing,
		    float64  easting,
		    int32    zone,
		    float64 *lat,
		    float64 *lon 
		    );

		    
extern void 
coord_ll_to_utm (     
		    int32    datum,
		    float64  lat,
		    float64  lon,
		    int32   *zone, /* Forced, if non-zero */
		    float64 *northing,
		    float64 *easting 
		    );
extern void 
coord_gcc_to_tcc (    
		     float64       x_in,
		     float64       y_in,
		     float64       z_in,
		     COORD_TCC_PTR tcc,
		     float64      *x_out, 
		     float64      *y_out,
		     float64      *z_out 
		     );
extern void 
coord_tcc_to_gcc (     
		     COORD_TCC_PTR tcc,
		     float64       x_in,
		     float64       y_in,
		     float64       z_in,
		     float64      *x_out,
		     float64      *y_out,
		     float64      *z_out
		     );
extern void 
coord_utm_to_grid (    
		      int32     from_zone,
		      float64   from_northing, 
		      float64   from_easting,
		      int32    *to_zone,
		      int32     to_resolution,
		      char     *to_string 
		      );
extern int32 
coord_grid_to_utm (     
		      int32     from_zone,
		      char     *from_string,
		      int32    *to_zone,
		      float64  *to_northing,
		      float64  *to_easting 
		      );
extern void 
coord_change_utm_zone (     
			  int32     from_zone,
			  float64   from_northing,
			  float64   from_easting,
			  int32    *to_zone,
			  float64  *to_northing,
			  float64  *to_easting 
			  );

extern float64 
coord_get_major_axis ( int32 datum );

extern float64 
coord_get_minor_axis ( int32 datum );

extern void 
coord_get_datum_shifts ( 
			   int32 datum,
			   float64 *dx, 
			   float64 *dy, 
			   float64 *dz
			   );

extern int32 
coord_lookup_datum_by_ne (     
			     int32 zone,
			     float64 northing,
			     float64 easting 
			     );
extern int32 
coord_lookup_datum_by_ll (     
			     float64 latitude,
			     float64 longitude 
			     );
extern char 
coord_lookup_zone_letter (     
			     int32    zone,
			     float64 northing 
			     );
extern void 
coord_lookup_utm (     
		     int32    zone,
		     float64 northing,
		     float64 easting,
		     int32   resolution,
		     char   *string 
		     );
extern int32 
coord_parse_utm (     
		    int32     default_zone,
		    char      default_zl,
		    char     *string,
		    int32    *zone,
		    float64  *northing,
		    float64  *easting 
		    );
