#ifndef lint
static char rcsid [] = "$RCSfile: rvatest.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: rvatest.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to find error introduced by RVAing in a bad way      *
*       Created: Tue Oct 20 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcoordinates.h"
#include "libcoord_local.h"
#include <libgcs.h>
#include <libvecmat.h>
#include <math.h>
#include <stdext.h> /*common/include/global*/

#define TIME  5.0   /* seconds */

static void register_tcc(tcc)
    COORD_TCC_PTR tcc;
{
    float64 offset[XY];
    float64 rotation[XYZ][XYZ];
    float64 gcc[XYZ];
    
    GCS_FLAT_EARTH  flat_earth;
    flat_earth.tccf_to_gcc = 
      (GCS_TCC_TO_GCC_FUNCTION)coord_tccf_to_gcc;
    flat_earth.gcc_to_tccf = 
      (GCS_GCC_TO_TCC_FUNCTION)coord_gcc_to_tccf;
    flat_earth.tcc_ptr = tcc;
    gcs_init_cell( GCS_ILLEGAL_CELL, &flat_earth);
    
    gcs_offset_origin(GCS_ILLEGAL_CELL, -tcc->half_width,
		      -tcc->half_height);
    
    coord_get_gcc_origin(tcc,gcc); 
    coord_get_rotation(tcc,rotation); 
    vmat3_transpose64(rotation, rotation); 
    gcs_set_coord_data(GCS_ILLEGAL_CELL,rotation,gcc); 
}

main(argc,argv)
    int32 argc;
    char *argv[];
{
    COORD_TCC_PTR utm_tcc;
    float64 x, y, z;
    float64 gx, gy, gz, gx1, gy1, gz1;
    float64 dx, dy, dz;
    float64 rot[9];
    float64 v[3], gv[3];
    float64 speed;

    reader_init(0);
    if (coord_init(".", 0) != 0)
      exit(0);

    gcs_init("../../data",0,coord_get_datum_info);

    /* Define a flat-earth TCC for Hunter-Liggett */
    utm_tcc = coord_define_tcc(COORD_UTM_NE,
			       3953200.0, 635000.0, 10, 'S', 3,
			       100000, 100000);
#define USE_GCS
#ifdef USE_GCS
    gcs_set_cell_awareness(GCS_SINGLE_CELL);
    register_tcc(utm_tcc);
    coord_set_tcc_awareness(FALSE);
#else
    coord_set_tcc_awareness(TRUE);
#endif
    for (speed=10.0;speed<=330.0;speed+=20.0)
    {

	/* Start our vehicle at the center of the database, at
	 * 500 meters ASL
	 */
	x = 50000.0;
	y = 50000.0;
	z = 500.0;

	/* Translate its location to GCC */
	coord_convert(COORD_TCC, utm_tcc, x, y, z,
		      COORD_GCC, &gx, &gy, &gz);

	/* Make a velocity vector in flat-world space */
	v[X] = v[Y] = v[Z] = speed * sqrt(3.0);

	/* Find the equivalent vector in curved-world space */
	coord_tcc_gcc_rotation(utm_tcc, rot);
	vmat3_vec_mat_mul64(v, rot, gv);

	/* Move the vehicle forward in flat-world space */
	x += v[X] * TIME;
	y += v[Y] * TIME;
	z += v[Z] * TIME;

	/* Move the vehicle forward in curved-world space */
	gx += gv[X] * TIME;
	gy += gv[Y] * TIME;
	gz += gv[Z] * TIME;

	/* Translate the new flat-world location to GCC */
	coord_convert(COORD_TCC, utm_tcc, x, y, z,
		      COORD_GCC, &gx1, &gy1, &gz1);

	/* Compute the error */
	dx = gx1-gx;
	dy = gy1-gy;
	dz = gz1-gz;
	printf("Error in %f seconds at %f mps: %f meters\n",
	       TIME, speed, sqrt(dx*dx+dy*dy+dz*dz));
    }
}

    
