#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program which converts typed input                   *
*       Created: Wed Sep  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include <stdio.h>
#include "libcoordinates.h"
#include "libcoord_local.h"
#include <libgcs.h>
#include <math.h>
#include <stdext.h> /*common/include/global*/

char *help_string = "Input format is as follows:  <from> <data> <data>... <to>\n<from>, <to> are single letter codes indicating coordinate system:\nCode\tData\t\t\t\tDescription\n----\t----\t\t\t\t-----------\n U  \t<zone> <northing> <easting> <Z>\tUTM Northing/Easting\n M  \t<milgrid designation> <Z>\tUTM MilGrid designation\n G  \t<x> <y> <z>\t\t\tGeocentric coordinates\n #  \t<x> <y> <z>\t\t\tTopocentric Coordinates (specify number)\n L  \t<latitude> <longitude> <Z>\tLat/Long assuming local datum\n S  \t<latitude> <longitude> <Z>\tLat/Long assuming WGS 84\n\nTo define a TCC system:\n D <northing> <easting> <zone num> <zone let> <datum> <width> <height> <G/U>\nTo print all TCC rotation matrices: R\n";
static int32 reg_tcc=0;

float64 angle_between(a, b)
    float64 *a, *b;
{
    float64 dot;

    dot = a[X]*b[X] + a[Y]*b[Y] + a[Z]*b[Z];

    if ((dot > 1.0) || (dot < -1.0))
      return 0.0;
    else
      return RAD_TO_DEG(acos(dot));
}

static void 
register_tcc(COORD_TCC_PTR tcc)
{
    float64 offset[XY];
    float64 rotation[XYZ][XYZ];
    float64 gcc[XYZ];
    
    if ( reg_tcc )  
    { 
	GCS_FLAT_EARTH  flat_earth;
	flat_earth.tccf_to_gcc = 
	  (GCS_TCC_TO_GCC_FUNCTION)coord_tccf_to_gcc;
	flat_earth.gcc_to_tccf = 
	  (GCS_GCC_TO_TCC_FUNCTION)coord_gcc_to_tccf;
	flat_earth.tcc_ptr = tcc;
	gcs_init_cell( GCS_ILLEGAL_CELL, &flat_earth);

	gcs_offset_origin(GCS_ILLEGAL_CELL, -tcc->half_width, 
		      -tcc->half_height);

	coord_get_gcc_origin(tcc,gcc); 
	coord_get_rotation(tcc,rotation); 
	vmat3_transpose64(rotation, rotation); 
	gcs_set_coord_data(GCS_ILLEGAL_CELL,rotation,gcc); 
    }	    
}

main(int32 argc, char *argv[])
{
    int32         ret, i, j;
    char          inbuf[100];
    char          outbuf0[100], outbuf1[100];
    int32         Izone, Inorthing, Ieasting;
    float64       Ilatitude, Ilongitude;
    int32         It;
    int32         Ix, Iy, Iz;
    int32         Ozone;
    float64       Onorthing, Oeasting;
    float64       Olatitude, Olongitude;
    int32         Ot;
    float64       Ox, Oy, Oz;
    char          to_system;
    COORD_TCC_PTR tcc[10];
    float64       rots[10][3][3], *rot;
    int32         tcc_count = 0;
    int32 mode,tcc_aware;

    reader_init(0);

    if ( coord_init("../../data", 0) != 0 )
      exit(0);

    gcs_init(coord_get_datum_info);
    
    printf("Are you running simnet (0) or single cell(1) or multicell(2) mode? ");
    scanf("%d",&mode);
    gcs_set_cell_awareness((mode == 2));

    printf("Are you allowing TCC and GCS conversion? (0/1)");
    scanf("%d",&tcc_aware);
    coord_set_tcc_awareness(tcc_aware);

    /* running a simnet db and not TCC processing */
    if ( mode == 0 )
      reg_tcc = 1;
    
    bzero(tcc, sizeof(tcc));
    
    printf("%s\n", help_string);

    while (1)
    {
	if (!gets(inbuf))
	  exit(0);

	if ((inbuf[0] >= 'a') && (inbuf[0] <= 'z'))
	  inbuf[0] -= ('a' - 'A');

	if ((inbuf[0] >= '0') && (inbuf[0] <= '9'))
	{
	    It = inbuf[0] - '0' ;
	    register_tcc(tcc[It]);
	    inbuf[0] = 'T';
	}
	
	switch (inbuf[0])
	{
	  case '?':
	  case 'H':
	    printf("%s", help_string);
	    continue;
	  case '#':
	    printf("%s\n", inbuf+1);
	    continue;
	  case 'R':
	    printf("Rotations:\n");
	    for (i=0;i<tcc_count;i++)
	    {
		rot = &rots[i][X][X]; 
		register_tcc(tcc[i]);
		coord_tcc_gcc_rotation(tcc[i], rots[i]);
		printf("#%d:\t%f\t%f\t%f\n\t%f\t%f\t%f\n\t%f\t%f\t%f\n",
		       i, rot[0], rot[1], rot[2], rot[3], rot[4], rot[5],
		       rot[6], rot[7], rot[8]);
	    }
	    printf("Angles between rotations (degrees):\n");
	    for (i=0;i<tcc_count;i++)
	      for (j=0;j<tcc_count;j++)
	      {
		  printf("%d vs %d:\t%2.2f [X], %2.2f [Y], %2.2f [Z]\n",
			 i, j,
			 angle_between(rots[i][X], rots[j][X]),
			 angle_between(rots[i][Y], rots[j][Y]),
			 angle_between(rots[i][Z], rots[j][Z]));
	      }
	    continue;
	  case 'D':
	    {
		int32 n, e, z, d, w, h;
		char l, s;

		if (sscanf(inbuf+1, " %d %d %d %c %d %d %d %c",
			   &n, &e, &z, &l, &d, &w, &h, &s) != 8)
		{
		    printf("Expect 3 integers, a letter, 3 integers, and a letter\n");
		    continue;
		}
		if ((s == 'g') || (s == 'G'))
		  tcc[tcc_count] = coord_define_tcc(COORD_GCC,
						    (float64)n, (float64)e,
						    z, l, d, w, h);
		else if ((s == 'u') || (s == 'U'))
		  tcc[tcc_count] = coord_define_tcc(COORD_UTM_NE,
						    (float64)n, (float64)e,
						    z, l, d, w, h);
		else
		{
		    printf("Expect either G or U as the source system, not `%c'\n",
			   s);
		    continue;
		}
		printf("Allocated TCC #%d\n", tcc_count);
		tcc_count += 1;
	    }
	    continue;
	  case 'Q':
	    exit(0);
	    
#define CONVERT(args)							\
  if (ret = coord_convert args )					\
  {									\
      printf("ERROR: %s\n", coord_error(ret));				\
      continue;								\
  }

#define BODY								\
  {									\
      if (sscanf(inbuf+1, PARSE, &to_system) != COUNT+1)		\
      {									\
	  printf("Expect %s and an output format (U,M,G,0-9,L,S)\n",	\
		 MESSAGE);						\
	  continue;							\
      }									\
      if ((to_system >= 'a') && (to_system <= 'z'))			\
	to_system -= ('a' - 'A');					\
      if ((to_system >= '0') && (to_system <= '9'))			\
      {                                        \
	   Ot = to_system - '0';               \
	   to_system = 'T';                    \
           register_tcc( tcc[Ot] );	       \
      }			                       \
      CONVERT((FROM, COORD_CHARACTER_STRING, outbuf0));			\
      switch(to_system)							\
      {									\
	case 'U':							\
	  Ozone = COORD_DEFAULT_ZONE;					\
	  CONVERT((FROM,						\
		   COORD_UTM_NE, &Ozone, &Onorthing, &Oeasting, &Oz));	\
	  CONVERT((COORD_UTM_NE,  Ozone,  Onorthing,  Oeasting,  Oz,	\
		   COORD_CHARACTER_STRING, outbuf1));			\
	  break;							\
	case 'M':							\
	  Ozone = COORD_DEFAULT_ZONE;					\
	  CONVERT((FROM,						\
		   COORD_UTM_GRID, &Ozone, 10, outbuf1, &Oz));		\
	  break;							\
	case 'G':							\
	  CONVERT((FROM,						\
		   COORD_GCC, &Ox, &Oy, &Oz));				\
	  CONVERT((COORD_GCC,  Ox,  Oy,  Oz,				\
		   COORD_CHARACTER_STRING, outbuf1));			\
	  break;							\
	case 'T':							\
	  CONVERT((FROM,						\
		   COORD_TCC, tcc[Ot], &Ox, &Oy, &Oz));			\
	  CONVERT((COORD_TCC, tcc[Ot],  Ox,  Oy,  Oz,			\
		   COORD_CHARACTER_STRING, outbuf1));			\
	  break;							\
	case 'L':							\
	  CONVERT((FROM,						\
		   COORD_LATLON, &Olatitude, &Olongitude, &Oz, TRUE));	\
	  CONVERT((COORD_LATLON,  Olatitude,  Olongitude,  Oz, TRUE,	\
		   COORD_CHARACTER_STRING, outbuf1));			\
	  break;							\
	case 'S':							\
	  CONVERT((FROM,						\
		   COORD_LATLON, &Olatitude, &Olongitude, &Oz, FALSE));	\
          CONVERT((COORD_LATLON, Olatitude, Olongitude, Oz, FALSE,	\
		   COORD_CHARACTER_STRING, outbuf1));   		\
	  break;							\
	default:							\
	  printf("Expect %s and an output format (U,M,G,0-9,L,S)\n",	\
		 MESSAGE);						\
	  continue;							\
      }									\
  }									\
  break

	  case 'U':
    
#define PARSE   " %d %d %d %d %c", &Izone, &Inorthing, &Ieasting, &Iz
#define COUNT   4
#define MESSAGE "4 integers"
#define FROM    \
  COORD_UTM_NE, Izone, (float64)Inorthing, (float64)Ieasting, (float64)Iz

	    BODY;

#undef FROM
#undef PARSE
#undef COUNT
#undef MESSAGE

	  case 'M':
    
#define PARSE   " %s %d %c", outbuf0, &Iz
#define COUNT   2
#define MESSAGE "a string, an integer,"
#define FROM    COORD_UTM_GRID, COORD_DEFAULT_ZONE, 0, outbuf0, (float64)Iz

	    BODY;

#undef FROM
#undef PARSE
#undef COUNT
#undef MESSAGE

	  case 'G':

#define PARSE   " %d %d %d %c", &Ix, &Iy, &Iz
#define COUNT   3
#define MESSAGE "three integers"
#define FROM    COORD_GCC, (float64)Ix, (float64)Iy, (float64)Iz

	    BODY;

#undef FROM
#undef PARSE
#undef COUNT
#undef MESSAGE

	  case 'T':

#define PARSE   " %d %d %d %c", &Ix, &Iy, &Iz
#define COUNT   3
#define MESSAGE "three integers"
#define FROM    COORD_TCC, tcc[It], (float64)Ix, (float64)Iy, (float64)Iz

	    BODY;

#undef FROM
#undef PARSE
#undef COUNT
#undef MESSAGE

	  case 'L':

#define PARSE   " %F %F %d %c", &Ilatitude, &Ilongitude, &Iz
#define COUNT   3
#define MESSAGE "two floats, an integer,"
#define FROM    COORD_LATLON, Ilatitude, Ilongitude, (float64)Iz, TRUE

	    BODY;

#undef FROM
#undef PARSE
#undef COUNT
#undef MESSAGE

	  case 'S':

#define PARSE   " %F %F %d %c", &Ilatitude, &Ilongitude, &Iz
#define COUNT   3
#define MESSAGE "two floats, an integer,"
#define FROM    COORD_LATLON, Ilatitude, Ilongitude, (float64)Iz, FALSE

	    BODY;

#undef FROM
#undef PARSE
#undef COUNT
#undef MESSAGE

	}

	printf("%40s = %s\n", outbuf0, outbuf1);
    }
}
