#ifndef lint
static char rcsid [] = "$RCSfile: timetest.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: timetest.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program to time performance                          *
*       Created: Tue Oct 20 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libcoordinates.h"
#include <libtime.h>
#include <stdext.h> /*common/include/global*/

#define NITS 10000

COORD_TCC_PTR utm_tcc;
COORD_TCC_PTR gcc_tcc;

float64 utm_x, utm_y, utm_z;
float64 gcc_x, gcc_y, gcc_z;

static void results(name, t0, t1, t2)
    char *name;
    uint32 t0, t1, t2;
{
    float64 dt = (t2 - t1) - (t1 - t0);

    printf("%s: %d usec/op; %f op/sec\n",
	   name, (int)(dt * 1000.0 / NITS), NITS / (dt / 1000.0));
}

static void dummy_fcn()
{
    return;
}

static void flat_to_curved()
{
    float64 x, y, z;

    coord_convert(COORD_TCC,
		  utm_tcc, utm_x, utm_y, utm_z,
		  COORD_TCC,
		  gcc_tcc, &x, &y, &z);
}

static void curved_to_flat()
{
    float64 x, y, z;

    coord_convert(COORD_TCC,
		  gcc_tcc, gcc_x, gcc_y, gcc_z,
		  COORD_TCC,
		  utm_tcc, &x, &y, &z);
}

main(argc,argv)
    int32 argc;
    char *argv[];
{
    uint32 t0, t1, t2;
    int32 i;

    reader_init(0);
    if (coord_init(".", 0) != 0)
      exit(0);

    utm_tcc = coord_define_tcc(COORD_UTM_NE,
			       4155000.0, 545000.0, 16, 'S', 3,
			       75000, 50000);

    gcc_tcc = coord_define_tcc(COORD_GCC,
			       4155000.0, 545000.0, 16, 'S', 3,
			       75000, 50000);

    utm_x = utm_y = 30000.0;
    utm_z = 300.0;
    coord_convert(COORD_TCC,
		  utm_tcc, utm_x, utm_y, utm_z,
		  COORD_GCC,
		  &gcc_x, &gcc_y, &gcc_z);

    time_init(0);

    t0 = time_realtime_clock();

    for (i=0;i<NITS;i++)
      dummy_fcn();

    t1 = time_realtime_clock();

    for (i=0;i<NITS;i++)
      flat_to_curved();

    t2 = time_realtime_clock();

    results("Flat to curved", t0, t1, t2);

    t0 = time_realtime_clock();

    for (i=0;i<NITS;i++)
      dummy_fcn();

    t1 = time_realtime_clock();

    for (i=0;i<NITS;i++)
      curved_to_flat();

    t2 = time_realtime_clock();

    results("Curved to flat", t0, t1, t2);
}

    
