/*
 * $RCSfile: ct_cmp_fmt2.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt2.c                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 2 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/

#include "ct_cmplr.h"

/***************************************************************************
 * READ ROUTINES                                                           *
 ***************************************************************************/

static void 
ctdb_unpack_header_2(ctdb, hdr)
    CTDB_2             *ctdb;
    CTDB_FILE_HEADER_2 *hdr;
{
    ctdb->input_format = hdr->format;

    ctdb->is_curved  = hdr->is_curved;
    ctdb->is_sw_ne   = hdr->is_sw_ne;

    ctdb->num_features = hdr->num_features;

    ctdb->fixed_point_basis = hdr->fixed_point_basis;
    ctdb->inv_fixed_point_basis = 1.0/hdr->fixed_point_basis;

    bcopy(hdr->soil_map, ctdb->soil_map, sizeof(ctdb->soil_map));

    ctdb->incr = hdr->post_increment;
    ctdb->incrD = ctdb->incr;
    ctdb->inv_incr = 1.0/ctdb->incrD;

    ctdb->patch_incr = hdr->posts_per_patch_side;
    ctdb->patch_incrD = ctdb->patch_incr;
    ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide = hdr->pages_wide;
    ctdb->pages_high = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE / ctdb->patch_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE / ctdb->patch_incr;

    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);
    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);

    bcopy(hdr->cache_size_map, ctdb->cache_size_map,
	  16 * sizeof(int32));

    ctdb->origin_northing = hdr->origin_northing;
    ctdb->origin_easting = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum = hdr->datum;
    bcopy(hdr->name, ctdb->name, sizeof(hdr->name));
    ctdb->version = hdr->version;
    bcopy(hdr->date, ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date, "\n");
    
    return;
}

static void 
ctdb_read_features_2(ctdb, f)
    CTDB_2 *ctdb;
    FILE   *f;
{
    int32 patches = ctdb->patches_wide * ctdb->patches_high;
    int32 groups = patches/CTDB_PATCHES_PER_GROUP;
    int32 i;

    ctdb->patch_groups =
      (CTDB_PATCH_GROUP_2 *)ctdb_alloc(NULL,
				       sizeof(CTDB_PATCH_GROUP_2) * groups,
				       "Fmt. 2 PG headers");
    fread(ctdb->patch_groups, sizeof(CTDB_PATCH_GROUP_2), groups, f);

    ctdb->features =
      (CTDB_FEATURE_DATA_2 *)ctdb_alloc(NULL,
					sizeof(CTDB_FEATURE_DATA_2) *
					ctdb->num_features,
					"Fmt. 2 feature data");

    fread(ctdb->features, sizeof(CTDB_FEATURE_DATA_2),
	  ctdb->num_features, f);

    return;
}

void 
ctdb_read_elevations_2(ctdb, f)
    CTDB_2 *ctdb;
    FILE   *f;
{
    uint32 memory, pages;
    int32  i;
    int    file_pos;

    pages = ctdb->pages_wide * ctdb->pages_high;

    memory = PAGE_SIZE * pages;

    ctdb->elevations_resident = 1;
    ctdb->elevations = (int32 *)ctdb_alloc(NULL,
					   memory, "elevation data");

    /* Note the file pointer value */
    file_pos = ftell(f);

    /* Load as many pages as the caller will let us */
    if (fread(ctdb->elevations, memory, 1, f) != 1)
      printf("CTDB: Read failed.  First %d bytes of post data.\n",
	     memory), exit(1);

    /* Position the file pointer at the end of the elevations */
    file_pos += PAGE_SIZE * pages;
    fseek(f, file_pos, 0);
}


static void
ctdb_reread_2(fname, ctdb)
    char   *fname;
    CTDB_2 *ctdb;
{
    FILE *f;
    CTDB_FILE_HEADER_2 hdr;
    int32 total_mem;

    if (!fname)
    {
	printf("Invalid input database.\n");
	exit(-1);
    }

    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	exit(-1);
    }

    if (hdr.format != 2)
    {
	printf("Attempted to read %d format db with fmt 2 reader!\n",
	       hdr.format);	
	fclose(f);
	exit(-1);
    }

    ctdb->file = f;

    ctdb_unpack_header_2(ctdb, &hdr);

    ctdb_read_features_2(ctdb, f);
    
    ctdb_read_elevations_2(ctdb, f);
    
    return;
}

extern void *
ctdb_read_2(fname)
    char *fname;
{
    CTDB_2 *ctdb;
    
    ctdb = (CTDB_2 *)ctdb_alloc(NULL, sizeof(CTDB_2), "Creating fmt 2 ctdb");
    bzero(ctdb, sizeof(CTDB_2));

    printf("Attempting to read CTDB datafiles... \n");

    ctdb_reread_2(fname, ctdb);
    return ctdb; 
}


/***************************************************************************
 * WRITE ROUTINES                                                          *
 ***************************************************************************/

static void
ctdb_pack_header_2(ctdb, hdr)
    CTDB_2             *ctdb;
    CTDB_FILE_HEADER_2 *hdr;
{
    bzero(hdr, sizeof(CTDB_FILE_HEADER_2));
    hdr->format = ctdb->input_format;
    hdr->is_curved = ctdb->is_curved;
    hdr->is_sw_ne = ctdb->is_sw_ne;
    hdr->fixed_point_basis = ctdb->fixed_point_basis;
    bcopy(ctdb->soil_map, hdr->soil_map, sizeof(hdr->soil_map));

    hdr->post_increment = ctdb->incr;
    hdr->posts_per_patch_side = ctdb->patch_incr;
    hdr->pages_wide = ctdb->pages_wide;
    hdr->pages_high = ctdb->pages_high;
    hdr->min_x = ctdb->min_x;
    hdr->max_x = ctdb->max_x;
    hdr->min_y = ctdb->min_y;
    hdr->max_y = ctdb->max_y;
    hdr->num_features = ctdb->num_features;
    bcopy(ctdb->cache_size_map, hdr->cache_size_map,
	  sizeof(ctdb->cache_size_map));

    hdr->origin_northing = ctdb->origin_northing;
    hdr->origin_easting = ctdb->origin_easting;
    hdr->origin_zone_number = ctdb->origin_zone_number;
    hdr->origin_zone_letter = ctdb->origin_zone_letter;
    hdr->datum = ctdb->datum;
    bcopy(ctdb->name, hdr->name, sizeof(ctdb->name));
    hdr->version = ctdb->version;
    bcopy(ctdb->date, hdr->date, sizeof(ctdb->date));

    return;
}




/* Return the size, in bytes, of all data, including the data header, in
 * pg
 */
static int32
find_pg_size_2(ctdb, pg)
    CTDB_2             *ctdb;
    CTDB_PATCH_GROUP_2 *pg;
{
    int32 size;

    /* Find the last occupied patch.  Add its offset to the number of bytes
     * of data stored for it to get the total PG data size
     */
    CTDB_FEATURE_DATA_2 *feat = NULL;
    
    if (pg->patch3_occupied)
    {
	size = pg->patch3_offset * sizeof(CTDB_FEATURE_DATA_2);
	feat = ctdb->features + pg->feature_start + pg->patch3_offset;
    }
    else if (pg->patch2_occupied)
    {
	size = pg->patch2_offset * sizeof(CTDB_FEATURE_DATA_2);
	feat = ctdb->features + pg->feature_start + pg->patch2_offset;
    }
    else if (pg->patch1_occupied)
    {
	size = pg->patch2_offset * sizeof(CTDB_FEATURE_DATA_2);
	feat = ctdb->features + pg->feature_start + pg->patch2_offset;
    }
    else if (pg->patch0_occupied)
    {
	size = 0;
	feat = ctdb->features + pg->feature_start;
    }
    if (feat == NULL)
      return 0;
    
    size += (feat->hdr.size * sizeof(CTDB_FEATURE_DATA_2 *));
    return size;
}

/* Generate the caching scheme for the patch data and fill in the
 * necessary data structures
 */
static void
generate_caching_scheme_2(ctdb)
    CTDB_2 *ctdb;
{
    int32 i;
    int32 group;
    int32 num_groups;
    int32 max_size = 0;
    CTDB_PATCH_GROUP_2 *pg;
    int32 *pg_size_map;		/* pg_size_map[i] is the size, in words, of
				 * patch group i
				 */
    int32 *map;			/* Current entry in pg_size_map array */
    int32 *counts;		/* counts[i] is the number of patch groups
				 * with i words of data
				 */
    int32 size;			/* Current PG size we're looking at */
    int32 last_size;		/* Last non-empty PG size we looked at */
    int32 *sizes = ctdb->cache_size_map; /* Sizes of the patch caches */
    int32 burdens[16];	        /* Number of PG's mapped to each 
				 * cache 
				 */
    int32 n_mapped;		/* Ideal number of groups per cache */
    int32 cur_count;		/* Groups mapped to current cache so far */
    int32 cur_cache;		/* Which cache we're currently filling */
    int32 min_burden;		/* Number of PGs mapped to least used cache */
    int32 smallest;
    int32 max_burden;		/* Number of PGs mapped to most used cache */
    int32 largest;		/* Most used cache */
    int32 best_cache;		/* "Best" cache for current PG */

    num_groups = (ctdb->patches_wide * ctdb->patches_high) /
      CTDB_PATCHES_PER_GROUP_2;

    pg_size_map = (int32 *)ctdb_alloc(NULL, num_groups * sizeof(int32),
				      "Patch group size map");


    /* 1.  Find the size of each patch group.  While we're at it, keep track
     * of how big the largest group is.
     */
    for(group = 0; group < num_groups; group++)
    {
	pg_size_map[group] = find_pg_size_2(ctdb, ctdb->patch_groups + group) 
	  / 4;

	if (pg_size_map[group] > max_size)
	  max_size = pg_size_map[group];
    }

    max_size++;

    counts = (int32 *)ctdb_alloc(NULL, sizeof(int32) * max_size,
				"patch group size counts");

    bzero(counts, max_size * sizeof(int32));
    bzero(sizes, 16 * sizeof(int32));
    bzero(burdens, 16 * sizeof(int32));

    /* 2. Find distribution of group sizes. 
     */
    map = pg_size_map;
    for(group=0;group<num_groups;group++,map++)
      counts[*map]++;

    /* 3. Choose the ideal number of groups to be mapped to each cache.
     */
    n_mapped =
      ceil((float64)num_groups/16.0) + 1;

    printf("Ideal groups/cache = %d\n",n_mapped);

    /* 4. Choose cache sizes such that AT LEAST n_mapped groups land
     * in each cache.
     */
    cur_count = cur_cache = 0;
    for(size=0;size<max_size;size++)
    {
	if (counts[size])
	{
	    cur_count += counts[size];
	    last_size = size;
	}

	if (cur_count > n_mapped)
	{
	    do
	    {
		sizes[cur_cache] = size;
		if (cur_count >= 2*n_mapped)
		  burdens[cur_cache] = n_mapped;
		else
		  burdens[cur_cache] = cur_count;
		cur_cache++;
		cur_count -= n_mapped;
	    } while (cur_count >= n_mapped);
	    cur_count = 0;
	}
    }
    sizes[cur_cache] = last_size;
    burdens[cur_cache] = cur_count;
    cur_cache++;

    /* 5. If not all the caches were used, split the most burdened
     * caches in two.
     */
    while(cur_cache<16)
    {
	/* Find the worst off */
	max_burden = largest = 0;
	for(i=0;i<cur_cache;i++)
	  if (burdens[i] > max_burden)
	  {
	      max_burden = burdens[i];
	      largest = i;
	  }
	/* Make room for the split */
	for(i=cur_cache;i>largest;i--)
	{
	    sizes[i] = sizes[i-1];
	    burdens[i] = burdens[i-1];
	}
	/* Split it in two */
	burdens[largest] /= 2;
	burdens[largest+1] -= burdens[largest];
	cur_cache += 1;
    }

    /* 6. Tag the groups with their cache number.
     */
    bzero(burdens,sizeof(burdens));
    for(group=0;group<num_groups;group++)
    {
	pg = ctdb->patch_groups + group;
	map = pg_size_map + group;

	/* Search through the caches and find which ones we can put this
	 * patch into.  Choose the least burdened of those.
	 */
	min_burden = smallest = 0x7FFFFFF;
	best_cache = 16 - 1;
	for(cur_cache=0; cur_cache < 16; cur_cache++)
	{
	    /* Have we passed the range of caches for which this size
	     * is intended?
	     */
	    if (sizes[cur_cache] > smallest)
	      break;
	      
	    /* Will the patch fit in this cache? */
	    if (sizes[cur_cache] < *map)
	      continue;
	    smallest = sizes[cur_cache];
	    if (burdens[cur_cache] < min_burden)
	    {
		best_cache = cur_cache;
		min_burden = burdens[cur_cache];
	    }
	}
	burdens[best_cache]++;
	pg->cache = best_cache;
    }

    /* 7. Print the results.
     */
    size = 0;
    for (cur_cache=0; cur_cache < 16; cur_cache++)
    {
	if (cur_cache && (sizes[cur_cache] != sizes[cur_cache-1]))
	  size = sizes[cur_cache-1];
	printf("Cache %x: [%d->%d].  %d patch groups\n",
	       cur_cache, size + 1, sizes[cur_cache], burdens[cur_cache]);
    }

    free(counts);    
    return;
}

static void
write_physical_features_2(ctdb, output_file)
    CTDB_2 *ctdb;
    FILE *output_file;
{
    CTDB_PATCH_GROUP_2 pg;
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_2 *inter_pg;
    
    groups = ctdb->patches_wide * ctdb->patches_high/
      CTDB_PATCHES_PER_GROUP_2;

    /* Write out the patch group headers */
    fwrite(ctdb->patch_groups, sizeof(CTDB_PATCH_GROUP_2), groups, 
	   output_file);
    
    /* Write out the feature data itself */
    if (ctdb->num_features)
      fwrite(ctdb->features, sizeof(CTDB_FEATURE_DATA_2), ctdb->num_features,
	     output_file);
}

static void 
write_posts_2(ctdb, output_file)
    CTDB_2  *ctdb;
    FILE    *output_file;
{
    int32 post_bytes;
    
    post_bytes = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
        
    fwrite(ctdb->elevations, post_bytes,1, output_file);
}

/* The input database is a void pointer to make it easier to call this routine
 * through a generic interface; it is the caller's responsibility to ensure
 * that it is, in fact, a format 3 CTDB.
 * NOTE: since the compiler does not use caching, the inter-format conversion
 * routines do NOT generate the caching scheme correctly; therefore it is the
 * responsibility of the write routine to do so.
 */

extern void
ctdb_write_2(fmt2_db, file)
    void *fmt2_db;
    FILE *file;
{
    CTDB_2 *ctdb = (CTDB_2 *)fmt2_db;
    CTDB_FILE_HEADER_2 hdr;
    
    generate_caching_scheme_2(ctdb);

    ctdb_pack_header_2(ctdb, &hdr);

    fwrite(&hdr, sizeof(hdr), 1, file);
	
    write_physical_features_2(ctdb, file);
	
    write_posts_2(ctdb, file);
	
    return;
}



/***************************************************************************
 * CLEANUP ROUTINES                                                          *
 ***************************************************************************/
extern void
ctdb_free_fmt2_db(fmt2_db)
    void *fmt2_db;
{
    CTDB_2 *ctdb = (CTDB_2 *)fmt2_db;
    
    if (!ctdb)
      return;
    
    if (ctdb->patch_groups)
      STDDEALLOC(ctdb->patch_groups);
    
    if (ctdb->features)
      STDDEALLOC(ctdb->features);
    
    if (ctdb->elevations)
      STDDEALLOC(ctdb->elevations);
    
    STDDEALLOC(ctdb);
    
    return;    
}

/***************************************************************************
 * 2 => 1 CONVERSION ROUTINES                                              *
 ***************************************************************************/

extern void *
ctdb_convert_2_to_1(fmt2_db)
    void *fmt2_db;
{
    CTDB_2 *ctdb = (CTDB_2 *)fmt2_db;
    CTDB_1 *ctdb_out;
    
    ctdb_out = (CTDB_1 *)ctdb_alloc(NULL, sizeof(CTDB_1), 
				      "Fmt. 1 database");
    
    bzero(ctdb_out, sizeof(CTDB_1));

    /* First, copy the easy stuff over */
#define COPY_2_1(x) (ctdb_out->x = ctdb->x)

    COPY_2_1(fixed_point_basis);
    COPY_2_1(inv_fixed_point_basis);
    
    COPY_2_1(incr);
    COPY_2_1(inv_incr);
    COPY_2_1(incrD);

    COPY_2_1(patch_incr);
    COPY_2_1(patch_inv_incr);
    COPY_2_1(patch_incrD);

    COPY_2_1(pages_wide);
    COPY_2_1(pages_high);
    COPY_2_1(patches_wide);
    COPY_2_1(patches_high);
    COPY_2_1(min_x);
    COPY_2_1(max_x);
    COPY_2_1(min_y);
    COPY_2_1(max_y);
    COPY_2_1(min_x_post);
    COPY_2_1(max_x_post);
    COPY_2_1(min_y_post);
    COPY_2_1(max_y_post);
    
    ctdb_out->input_format = 1;

    COPY_2_1(min_z);
    COPY_2_1(max_z);
    COPY_2_1(origin_northing);
    COPY_2_1(origin_easting);
    COPY_2_1(origin_zone_number);
    COPY_2_1(origin_zone_letter);
    COPY_2_1(datum);

    COPY_2_1(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_2_1(is_curved);
    COPY_2_1(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    /* Now for the real work */

    /* Soil tables */
    bcopy(ctdb->soil_map, ctdb_out->soil_map,
	  CTDB_N_SOILS_1 * sizeof(int32));

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));
    
    /* Features */
    {
	int32 groups;
	
	groups = (ctdb->patches_wide * ctdb->patches_high) /
	  CTDB_PATCHES_PER_GROUP_2;
	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_1 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_1),
		     "Fmt. 1 PG headers");
	
	bcopy(ctdb->patch_groups, ctdb_out->patch_groups,
	      groups * sizeof(CTDB_PATCH_GROUP_1));
	
	ctdb_out->num_features = ctdb->num_features;
	ctdb_out->features = (CTDB_FEATURE_DATA_1 *)
	  ctdb_alloc(NULL, ctdb->num_features * sizeof(CTDB_FEATURE_DATA_1),
		     "Fmt. 1 feature data");
	bcopy(ctdb->features, ctdb_out->features,
	      ctdb->num_features * sizeof(CTDB_FEATURE_DATA_1));
    }
    
    /* Elevations */
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 4 elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
    }
    
    return ctdb_out;    
}

/***************************************************************************
 * 2 => 3 CONVERSION ROUTINES                                              *
 ***************************************************************************/

extern void *
ctdb_convert_2_to_3(fmt2_db)
    void *fmt2_db;
{
    CTDB_2 *ctdb = (CTDB_2 *)fmt2_db;
    CTDB_3 *ctdb_out;
    
    ctdb_out = (CTDB_3 *)ctdb_alloc(NULL, sizeof(CTDB_3), 
				      "Fmt. 3 Database");
    
    bzero(ctdb_out, sizeof(CTDB_3));

    /* First, copy the easy stuff over */
#define COPY_2_3(x) (ctdb_out->x = ctdb->x)

    COPY_2_3(fixed_point_basis);
    COPY_2_3(inv_fixed_point_basis);
    
    COPY_2_3(incr);
    COPY_2_3(inv_incr);
    COPY_2_3(incrD);

    COPY_2_3(patch_incr);
    COPY_2_3(patch_inv_incr);
    COPY_2_3(patch_incrD);

    COPY_2_3(pages_wide);
    COPY_2_3(pages_high);
    COPY_2_3(patches_wide);
    COPY_2_3(patches_high);
    COPY_2_3(min_x);
    COPY_2_3(max_x);
    COPY_2_3(min_y);
    COPY_2_3(max_y);
    COPY_2_3(min_x_post);
    COPY_2_3(max_x_post);
    COPY_2_3(min_y_post);
    COPY_2_3(max_y_post);
    
    ctdb_out->input_format = 3;

    COPY_2_3(min_z);
    COPY_2_3(max_z);
    COPY_2_3(origin_northing);
    COPY_2_3(origin_easting);
    COPY_2_3(origin_zone_number);
    COPY_2_3(origin_zone_letter);
    COPY_2_3(datum);

    COPY_2_3(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_2_3(is_curved);
    COPY_2_3(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    /* Now for the real work */

    /* Soil tables */
    bcopy(ctdb->soil_map, ctdb_out->soil_map,
	  CTDB_N_SOILS_3 * sizeof(int32));

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));
    
    /* Features */
    {
	int32 groups;
	
	groups = (ctdb->patches_wide * ctdb->patches_high) /
	  CTDB_PATCHES_PER_GROUP_2;
	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_3 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_3),
		     "Fmt. 3 PG headers");
	
	bcopy(ctdb->patch_groups, ctdb_out->patch_groups,
	      groups * sizeof(CTDB_PATCH_GROUP_3));
	
	ctdb_out->num_features = ctdb->num_features;
	ctdb_out->features = (CTDB_FEATURE_DATA_3 *)
	  ctdb_alloc(NULL, ctdb->num_features * sizeof(CTDB_FEATURE_DATA_3),
		     "Fmt. 3 feature data");
	bcopy(ctdb->features, ctdb_out->features,
	      ctdb->num_features * sizeof(CTDB_FEATURE_DATA_3));
    }
    
    /* Elevations */
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 4 elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
    }
    
    return ctdb_out;    
}
