/*
 * $RCSfile: ct_cmp_fmt3.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt3.c                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 3 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/

#include "ct_cmplr.h"

/***************************************************************************
 * READ ROUTINES                                                           *
 ***************************************************************************/

static void 
ctdb_unpack_header_3(ctdb, hdr)
    CTDB_3             *ctdb;
    CTDB_FILE_HEADER_3 *hdr;
{
    ctdb->input_format = hdr->format;

    ctdb->is_curved  = hdr->is_curved;
    ctdb->is_sw_ne   = hdr->is_sw_ne;

    ctdb->num_features = hdr->num_features;

    ctdb->fixed_point_basis = hdr->fixed_point_basis;
    ctdb->inv_fixed_point_basis = 1.0/hdr->fixed_point_basis;

    bcopy(hdr->soil_map, ctdb->soil_map, sizeof(ctdb->soil_map));

    ctdb->incr = hdr->post_increment;
    ctdb->incrD = ctdb->incr;
    ctdb->inv_incr = 1.0/ctdb->incrD;

    ctdb->patch_incr = hdr->posts_per_patch_side;
    ctdb->patch_incrD = ctdb->patch_incr;
    ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide = hdr->pages_wide;
    ctdb->pages_high = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE / ctdb->patch_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE / ctdb->patch_incr;

    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);
    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);

    bcopy(hdr->cache_size_map, ctdb->cache_size_map,
	  16 * sizeof(int32));

    ctdb->origin_northing = hdr->origin_northing;
    ctdb->origin_easting = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum = hdr->datum;
    bcopy(hdr->name, ctdb->name, sizeof(hdr->name));
    ctdb->version = hdr->version;
    bcopy(hdr->date, ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date, "\n");

    ctdb->num_nodes = hdr->num_nodes;
    ctdb->node_size = hdr->node_size;
    ctdb->num_edges = hdr->num_edges;
    ctdb->edge_size = hdr->edge_size;
    ctdb->num_quad_data = hdr->num_quad_data;
}

static void 
ctdb_read_features_3(ctdb, f)
    CTDB_3 *ctdb;
    FILE   *f;
{
    int32 patches = ctdb->patches_wide * ctdb->patches_high;
    int32 groups = patches/CTDB_PATCHES_PER_GROUP;
    int32 i;

    ctdb->patch_groups =
      (CTDB_PATCH_GROUP_3 *)ctdb_alloc(NULL,
				       sizeof(CTDB_PATCH_GROUP_3) * groups,
				       "Fmt. 3 PG headers");
    fread(ctdb->patch_groups, sizeof(CTDB_PATCH_GROUP_3), groups, f);

    ctdb->features =
      (CTDB_FEATURE_DATA_3 *)ctdb_alloc(NULL,
					sizeof(CTDB_FEATURE_DATA_3) *
					ctdb->num_features,
					"Fmt. 3 feature data");

    fread(ctdb->features, sizeof(CTDB_FEATURE_DATA_3),
	  ctdb->num_features, f);

    return;
}

static void 
init_null_topology_3(ctdb)
    CTDB_3 *ctdb;
{
    ctdb->node_storage = (int8 *)
      ctdb_alloc(NULL, sizeof(CTDB_NODE_3), "initialization");
    bzero(ctdb->node_storage, sizeof(CTDB_NODE_3));
    ctdb->edge_storage = (int8 *)
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_3), "initialization");
    bzero(ctdb->edge_storage, sizeof(CTDB_EDGE_3));
    ctdb->nodes = (CTDB_NODE_3 **)
      ctdb_alloc(NULL, sizeof(CTDB_NODE_3 *), "initialization");
    ctdb->nodes[0] = (CTDB_NODE_3 *)ctdb->node_storage;
    ctdb->edges = (CTDB_EDGE_3 **)
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_3 *), "initialization");
    ctdb->edges[0] = (CTDB_EDGE_3 *)ctdb->edge_storage;
}

static CTDB_STATUS 
ctdb_read_topology_3(ctdb, fileptr)
    CTDB_3 *ctdb;
    FILE   *fileptr;
{
    int32 i, offset;
    
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces: ctdb_read_topology((CTDB *)ctdb,fileptr);*/
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    if (ctdb->num_nodes)
    {
        ctdb->nodes = NULL;
	ctdb->nodes =
	  (CTDB_NODE_3 **)ctdb_alloc(NULL,
				     sizeof(CTDB_NODE_3 *) *
				     ctdb->num_nodes,
				     "node index table");
	ctdb->node_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->node_size, "nodes");

	if (fread(ctdb->node_storage, ctdb->node_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Nodes\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_NODES_FAILED;
	}
    }

    if (ctdb->num_edges)
    {
        ctdb->edges = NULL;
	ctdb->edges =
	  (CTDB_EDGE_3 **)ctdb_alloc(NULL,
				     sizeof(CTDB_EDGE_3 *) *
				     ctdb->num_edges,
				     "edge index table");
	ctdb->edge_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->edge_size, "edges");

	if (fread(ctdb->edge_storage, ctdb->edge_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Edges\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_EDGES_FAILED;
	}
    }

    /* Fill in the node index table */
    bzero(ctdb->nodes, sizeof(CTDB_NODE_3 *) * ctdb->num_nodes);
    for(i=offset=0;(i<ctdb->num_nodes) && (offset<ctdb->node_size);i++)
    {
	ctdb->nodes[i] = (CTDB_NODE_3 *)(ctdb->node_storage + offset);
	offset += sizeof(CTDB_NODE_3) + sizeof(struct ctdb_connection) *
	  (ctdb->nodes[i]->num_connections - 1);
    }

    /* Fill in the edge index table */
    bzero(ctdb->edges, sizeof(CTDB_EDGE_3 *) * ctdb->num_edges);
    for(i=offset=0;(i<ctdb->num_edges) && (offset<ctdb->edge_size);i++)
    {
	ctdb->edges[i] = (CTDB_EDGE_3 *)(ctdb->edge_storage + offset);
	offset += sizeof(CTDB_EDGE_3) + sizeof(struct ctdb_connection) *
	  (ctdb->edges[i]->num_features - 1);
    }

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_quad_tree_3(ctdb, fileptr)
    CTDB_3 *ctdb;
    FILE   *fileptr;
{
    if (ctdb->num_quad_data == 0 )
      return CTDB_STATUS_OK;
        
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    ctdb->quad_data =
      (CTDB_FEATURE_DATA_3 *)ctdb_alloc(NULL,
					sizeof(CTDB_FEATURE_DATA_3) *
					ctdb->num_quad_data,
					"abstract data and quadtree");

    if (fread(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_3),
	      ctdb->num_quad_data, fileptr) != ctdb->num_quad_data)
    {
	printf("CTDB: Read failed.  Quad tree and abstract data\n");
	/* Survive this error */
	ctdb->num_quad_data = 0;
    }      
    return CTDB_STATUS_OK;
}

void 
ctdb_read_elevations_3(ctdb, f)
    CTDB_3 *ctdb;
    FILE   *f;
{
    uint32 memory, pages;
    int32  i;
    int    file_pos;

    pages = ctdb->pages_wide * ctdb->pages_high;

    memory = PAGE_SIZE * pages;

    ctdb->elevations_resident = 1;
    ctdb->elevations = (int32 *)ctdb_alloc(NULL,
					   memory, "elevation data");

    /* Note the file pointer value */
    file_pos = ftell(f);

    /* Load as many pages as the caller will let us */
    if (fread(ctdb->elevations, memory, 1, f) != 1)
      printf("CTDB: Read failed.  First %d bytes of post data.\n",
	     memory), exit(1);

    /* Position the file pointer at the end of the elevations */
    file_pos += PAGE_SIZE * pages;
    fseek(f, file_pos, 0);
}


static void
ctdb_reread_3(fname, ctdb)
    char   *fname;
    CTDB_3 *ctdb;
{
    FILE *f;
    CTDB_FILE_HEADER_3 hdr;
    int32 total_mem;

    if (!fname)
    {
	printf("Invalid input database.\n");
	exit(-1);
    }

    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	exit(-1);
    }

    if (hdr.format != 3)
    {
	printf("Attempted to read %d format db with fmt 3 reader!\n",
	       hdr.format);	
	fclose(f);
	exit(-1);
    }

    ctdb->file = f;

    ctdb_unpack_header_3(ctdb, &hdr);

    ctdb_read_features_3(ctdb, f);
    
    ctdb_read_elevations_3(ctdb, f);
    
    ctdb_read_topology_3(ctdb, f);
    
    ctdb_read_quad_tree_3(ctdb, f);
    
    return;
}

extern void *
ctdb_read_3(fname)
    char *fname;
{
    CTDB_3 *ctdb;
    
    ctdb = (CTDB_3 *)ctdb_alloc(NULL, sizeof(CTDB_3), "Creating fmt 3 ctdb");
    bzero(ctdb, sizeof(CTDB_3));

    printf("Attempting to read CTDB datafiles... \n");

    ctdb_reread_3(fname, ctdb);
    return ctdb; 
}




/***************************************************************************
 * WRITE ROUTINES                                                          *
 ***************************************************************************/

static void
ctdb_pack_header_3(ctdb, hdr)
    CTDB_3             *ctdb;
    CTDB_FILE_HEADER_3 *hdr;
{
    bzero(hdr, sizeof(CTDB_FILE_HEADER_3));
    hdr->format = ctdb->input_format;
    hdr->is_curved = ctdb->is_curved;
    hdr->is_sw_ne = ctdb->is_sw_ne;
    hdr->fixed_point_basis = ctdb->fixed_point_basis;
    bcopy(ctdb->soil_map, hdr->soil_map, sizeof(hdr->soil_map));

    hdr->post_increment = ctdb->incr;
    hdr->posts_per_patch_side = ctdb->patch_incr;
    hdr->pages_wide = ctdb->pages_wide;
    hdr->pages_high = ctdb->pages_high;
    hdr->min_x = ctdb->min_x;
    hdr->max_x = ctdb->max_x;
    hdr->min_y = ctdb->min_y;
    hdr->max_y = ctdb->max_y;
    hdr->num_features = ctdb->num_features;
    bcopy(ctdb->cache_size_map, hdr->cache_size_map,
	  sizeof(ctdb->cache_size_map));

    hdr->origin_northing = ctdb->origin_northing;
    hdr->origin_easting = ctdb->origin_easting;
    hdr->origin_zone_number = ctdb->origin_zone_number;
    hdr->origin_zone_letter = ctdb->origin_zone_letter;
    hdr->datum = ctdb->datum;
    bcopy(ctdb->name, hdr->name, sizeof(ctdb->name));
    hdr->version = ctdb->version;
    bcopy(ctdb->date, hdr->date, sizeof(ctdb->date));
    hdr->num_nodes = ctdb->num_nodes;
    hdr->node_size = ctdb->node_size;
    hdr->num_edges = ctdb->num_edges;
    hdr->edge_size = ctdb->edge_size;
    hdr->num_quad_data = ctdb->num_quad_data;

    return;
}




/* Return the size, in bytes, of all data, including the data header, in
 * pg
 */
static int32
find_pg_size_3(ctdb, pg)
    CTDB_3             *ctdb;
    CTDB_PATCH_GROUP_3 *pg;
{
    int32 size;

    /* Find the last occupied patch.  Add its offset to the number of bytes
     * of data stored for it to get the total PG data size
     */
    CTDB_FEATURE_DATA_3 *feat = NULL;
    
    if (pg->patch3_occupied)
    {
	size = pg->patch3_offset * sizeof(CTDB_FEATURE_DATA_3);
	feat = ctdb->features + pg->feature_start + pg->patch3_offset;
    }
    else if (pg->patch2_occupied)
    {
	size = pg->patch2_offset * sizeof(CTDB_FEATURE_DATA_3);
	feat = ctdb->features + pg->feature_start + pg->patch2_offset;
    }
    else if (pg->patch1_occupied)
    {
	size = pg->patch2_offset * sizeof(CTDB_FEATURE_DATA_3);
	feat = ctdb->features + pg->feature_start + pg->patch2_offset;
    }
    else if (pg->patch0_occupied)
    {
	size = 0;
	feat = ctdb->features + pg->feature_start;
    }
    if (feat == NULL)
      return 0;
    
    size += (feat->hdr.size * sizeof(CTDB_FEATURE_DATA_3 *));
    return size;
}

/* Generate the caching scheme for the patch data and fill in the
 * necessary data structures
 */
static void
generate_caching_scheme_3(ctdb)
    CTDB_3 *ctdb;
{
    int32 i;
    int32 group;
    int32 num_groups;
    int32 max_size = 0;
    CTDB_PATCH_GROUP_3 *pg;
    int32 *pg_size_map;		/* pg_size_map[i] is the size, in words, of
				 * patch group i
				 */
    int32 *map;			/* Current entry in pg_size_map array */
    int32 *counts;		/* counts[i] is the number of patch groups
				 * with i words of data
				 */
    int32 size;			/* Current PG size we're looking at */
    int32 last_size;		/* Last non-empty PG size we looked at */
    int32 *sizes = ctdb->cache_size_map; /* Sizes of the patch caches */
    int32 burdens[16];	        /* Number of PG's mapped to each 
				 * cache 
				 */
    int32 n_mapped;		/* Ideal number of groups per cache */
    int32 cur_count;		/* Groups mapped to current cache so far */
    int32 cur_cache;		/* Which cache we're currently filling */
    int32 min_burden;		/* Number of PGs mapped to least used cache */
    int32 smallest;
    int32 max_burden;		/* Number of PGs mapped to most used cache */
    int32 largest;		/* Most used cache */
    int32 best_cache;		/* "Best" cache for current PG */

    num_groups = (ctdb->patches_wide * ctdb->patches_high) /
      CTDB_PATCHES_PER_GROUP_3;

    pg_size_map = (int32 *)ctdb_alloc(NULL, num_groups * sizeof(int32),
				      "Patch group size map");


    /* 1.  Find the size of each patch group.  While we're at it, keep track
     * of how big the largest group is.
     */
    for(group = 0; group < num_groups; group++)
    {
	pg_size_map[group] = find_pg_size_3(ctdb, ctdb->patch_groups + group) 
	  / 4;

	if (pg_size_map[group] > max_size)
	  max_size = pg_size_map[group];
    }

    max_size++;

    counts = (int32 *)ctdb_alloc(NULL, sizeof(int32) * max_size,
				"patch group size counts");

    bzero(counts, max_size * sizeof(int32));
    bzero(sizes, 16 * sizeof(int32));
    bzero(burdens, 16 * sizeof(int32));

    /* 2. Find distribution of group sizes. 
     */
    map = pg_size_map;
    for(group=0;group<num_groups;group++,map++)
      counts[*map]++;

    /* 3. Choose the ideal number of groups to be mapped to each cache.
     */
    n_mapped =
      ceil((float64)num_groups/16.0) + 1;

    printf("Ideal groups/cache = %d\n",n_mapped);

    /* 4. Choose cache sizes such that AT LEAST n_mapped groups land
     * in each cache.
     */
    cur_count = cur_cache = 0;
    for(size=0;size<max_size;size++)
    {
	if (counts[size])
	{
	    cur_count += counts[size];
	    last_size = size;
	}

	if (cur_count > n_mapped)
	{
	    do
	    {
		sizes[cur_cache] = size;
		if (cur_count >= 2*n_mapped)
		  burdens[cur_cache] = n_mapped;
		else
		  burdens[cur_cache] = cur_count;
		cur_cache++;
		cur_count -= n_mapped;
	    } while (cur_count >= n_mapped);
	    cur_count = 0;
	}
    }
    sizes[cur_cache] = last_size;
    burdens[cur_cache] = cur_count;
    cur_cache++;

    /* 5. If not all the caches were used, split the most burdened
     * caches in two.
     */
    while(cur_cache<16)
    {
	/* Find the worst off */
	max_burden = largest = 0;
	for(i=0;i<cur_cache;i++)
	  if (burdens[i] > max_burden)
	  {
	      max_burden = burdens[i];
	      largest = i;
	  }
	/* Make room for the split */
	for(i=cur_cache;i>largest;i--)
	{
	    sizes[i] = sizes[i-1];
	    burdens[i] = burdens[i-1];
	}
	/* Split it in two */
	burdens[largest] /= 2;
	burdens[largest+1] -= burdens[largest];
	cur_cache += 1;
    }

    /* 6. Tag the groups with their cache number.
     */
    bzero(burdens,sizeof(burdens));
    for(group=0;group<num_groups;group++)
    {
	pg = ctdb->patch_groups + group;
	map = pg_size_map + group;

	/* Search through the caches and find which ones we can put this
	 * patch into.  Choose the least burdened of those.
	 */
	min_burden = smallest = 0x7FFFFFF;
	best_cache = 16 - 1;
	for(cur_cache=0; cur_cache < 16; cur_cache++)
	{
	    /* Have we passed the range of caches for which this size
	     * is intended?
	     */
	    if (sizes[cur_cache] > smallest)
	      break;
	      
	    /* Will the patch fit in this cache? */
	    if (sizes[cur_cache] < *map)
	      continue;
	    smallest = sizes[cur_cache];
	    if (burdens[cur_cache] < min_burden)
	    {
		best_cache = cur_cache;
		min_burden = burdens[cur_cache];
	    }
	}
	burdens[best_cache]++;
	pg->cache = best_cache;
    }

    /* 7. Print the results.
     */
    size = 0;
    for (cur_cache=0; cur_cache < 16; cur_cache++)
    {
	if (cur_cache && (sizes[cur_cache] != sizes[cur_cache-1]))
	  size = sizes[cur_cache-1];
	printf("Cache %x: [%d->%d].  %d patch groups\n",
	       cur_cache, size + 1, sizes[cur_cache], burdens[cur_cache]);
    }

    free(counts);    
    return;
}

static void
write_physical_features_3(ctdb, output_file)
    CTDB_3 *ctdb;
    FILE *output_file;
{
    CTDB_PATCH_GROUP_3 pg;
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_3 *inter_pg;
    
    groups = ctdb->patches_wide * ctdb->patches_high/
      CTDB_PATCHES_PER_GROUP_3;

    /* Write out the patch group headers */
    fwrite(ctdb->patch_groups, sizeof(CTDB_PATCH_GROUP_3), groups, 
	   output_file);
    
    /* Write out the feature data itself */
    if (ctdb->num_features)
      fwrite(ctdb->features, sizeof(CTDB_FEATURE_DATA_3), ctdb->num_features,
	     output_file);
}

static void 
write_posts_3(ctdb, output_file)
    CTDB_3  *ctdb;
    FILE    *output_file;
{
    int32 post_bytes;
    
    post_bytes = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
        
    fwrite(ctdb->elevations, post_bytes,1, output_file);
}

/* This function writes the nodes and edges to the output file.
 */
static void 
write_nodes_and_edges_3(ctdb, output_file)
    CTDB_3  *ctdb;
    FILE    *output_file;
{
    /* Just dump the stored nodes and edges on out */
    if (ctdb->node_size)
      fwrite(ctdb->node_storage, 1, ctdb->node_size, output_file);
    if (ctdb->edge_size)
      fwrite(ctdb->edge_storage, 1, ctdb->edge_size, output_file);
}

/* This function writes the abstract features to the output file. */
static void 
write_abstract_features_3(ctdb, output_file)
    CTDB_3  *ctdb;
    FILE    *output_file;
{
    /* Just dump the quad tree on out */
    if (ctdb->num_quad_data)
      fwrite(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_3),
	     ctdb->num_quad_data, output_file);
}

/* The input database is a void pointer to make it easier to call this routine
 * through a generic interface; it is the caller's responsibility to ensure
 * that it is, in fact, a format 3 CTDB.
 * NOTE: since the compiler does not use caching, the inter-format conversion
 * routines do NOT generate the caching scheme correctly; therefore it is the
 * responsibility of the write routine to do so.
 */

extern void
ctdb_write_3(fmt3_db, file)
    void *fmt3_db;
    FILE *file;
{
    CTDB_3 *ctdb = (CTDB_3 *)fmt3_db;
    CTDB_FILE_HEADER_3 hdr;
    
    generate_caching_scheme_3(ctdb);

    ctdb_pack_header_3(ctdb, &hdr);

    fwrite(&hdr, sizeof(hdr), 1, file);
	
    write_physical_features_3(ctdb, file);
	
    write_posts_3(ctdb, file);
	
    write_nodes_and_edges_3(ctdb, file);
	
    write_abstract_features_3(ctdb, file);
   
    return;
}



/***************************************************************************
 * CLEANUP ROUTINES                                                          *
 ***************************************************************************/
extern void
ctdb_free_fmt3_db(fmt3_db)
    void *fmt3_db;
{
    CTDB_3 *ctdb = (CTDB_3 *)fmt3_db;
    
    if (!ctdb)
      return;
    
    if (ctdb->patch_groups)
      STDDEALLOC(ctdb->patch_groups);
    
    if (ctdb->features)
      STDDEALLOC(ctdb->features);
    
    if (ctdb->elevations)
      STDDEALLOC(ctdb->elevations);
    
    if (ctdb->quad_data)
      STDDEALLOC(ctdb->quad_data);
    
    if (ctdb->nodes)
      STDDEALLOC(ctdb->nodes);
    
    if (ctdb->node_storage)
      STDDEALLOC(ctdb->node_storage);
    
    if (ctdb->edges)
      STDDEALLOC(ctdb->edges);
    
    if (ctdb->edge_storage)
      STDDEALLOC(ctdb->edge_storage);
    
    STDDEALLOC(ctdb);
    
    return;    
}



/***************************************************************************
 * 3 => 2 CONVERSION ROUTINES                                              *
 ***************************************************************************/
extern void *
ctdb_convert_3_to_2(fmt3_db)
    void *fmt3_db;
{
    CTDB_3 *ctdb = (CTDB_3 *)fmt3_db;
    CTDB_2 *ctdb_out;
    
    ctdb_out = (CTDB_2 *)ctdb_alloc(NULL, sizeof(CTDB_2), 
				      "Fmt. 4 database");
    
    bzero(ctdb_out, sizeof(CTDB_2));

    /* First, copy the easy stuff over */
#define COPY_3_2(x) (ctdb_out->x = ctdb->x)

    COPY_3_2(fixed_point_basis);
    COPY_3_2(inv_fixed_point_basis);
    
    COPY_3_2(incr);
    COPY_3_2(inv_incr);
    COPY_3_2(incrD);

    COPY_3_2(patch_incr);
    COPY_3_2(patch_inv_incr);
    COPY_3_2(patch_incrD);

    COPY_3_2(pages_wide);
    COPY_3_2(pages_high);
    COPY_3_2(patches_wide);
    COPY_3_2(patches_high);
    COPY_3_2(min_x);
    COPY_3_2(max_x);
    COPY_3_2(min_y);
    COPY_3_2(max_y);
    COPY_3_2(min_x_post);
    COPY_3_2(max_x_post);
    COPY_3_2(min_y_post);
    COPY_3_2(max_y_post);
    
    ctdb_out->input_format = 2;

    COPY_3_2(min_z);
    COPY_3_2(max_z);
    COPY_3_2(origin_northing);
    COPY_3_2(origin_easting);
    COPY_3_2(origin_zone_number);
    COPY_3_2(origin_zone_letter);
    COPY_3_2(datum);

    COPY_3_2(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_3_2(is_curved);
    COPY_3_2(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    /* Now for the real work */

    /* Soil tables */
    bcopy(ctdb->soil_map, ctdb_out->soil_map,
	  CTDB_N_SOILS_2 * sizeof(int32));

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));

    /* Features */
    {
	int32 groups;
	
	groups = (ctdb->patches_wide * ctdb->patches_high) /
	  CTDB_PATCHES_PER_GROUP_2;
	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_2 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_2),
		     "Fmt. 2 PG headers");
	
	bcopy(ctdb->patch_groups, ctdb_out->patch_groups,
	      groups * sizeof(CTDB_PATCH_GROUP_2));
	
	ctdb_out->num_features = ctdb->num_features;
	ctdb_out->features = (CTDB_FEATURE_DATA_2 *)
	  ctdb_alloc(NULL, ctdb->num_features * sizeof(CTDB_FEATURE_DATA_2),
		     "Fmt. 2 feature data");
	bcopy(ctdb->features, ctdb_out->features,
	      ctdb->num_features * sizeof(CTDB_FEATURE_DATA_2));
    }

    /* Elevations */
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 4 elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
    }
    
    return ctdb_out;    
}



/***************************************************************************
 * 3 => 4 CONVERSION ROUTINES                                              *
 ***************************************************************************/

/* For soil type conversions */

/* Mapping from pre-4 to post-3 database soil types.  The annotations are
 * CCTT-related.
 */
static int32 fmt4_soils[] =
{
    /* Undefined */
    (0 << 4) | 0,

    /* (Dry Pavement: Concrete, Asphalt) */
    CTDB_ROAD_FLAG_4 | (20 << 4) | 1,

    /* (Dry Loose Surface Road; GW, GP, GM, HC, Rock) */
    CTDB_ROAD_FLAG_4 | (17 << 4) | 2,

    /* (Soft; GW, GP, GM, GC, SM, SC, CL, ML, CH, MH, OL, OH) */
    (5 << 4) | 3,

    /* (Depth 60 inches; Water w/ (Silts and Clays) Bottom) */
    CTDB_WATER_FLAG_4 | CTDB_NO_GO_FLAG_4 | (28 << 4) | 4,

    /* (Depth 16 inches; Water w/ (Silts and Clays) Bottom) */
    CTDB_WATER_FLAG_4 | CTDB_SLOW_GO_FLAG_4 | (26 << 4) | 5,

    /* (Very Soft; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG_4 | (4 << 4) | 6,

    /* (Wet Loose Surface Road; GW, GP, GM, HC, Rock) */
    CTDB_ROAD_FLAG_4 | CTDB_SLOW_GO_FLAG_4 | (18 << 4) | 7,

    /* (Very Hard (Slippery); CH, OH) */
    CTDB_SLOW_GO_FLAG_4 | (15 << 4) | 8,

    /* (Swamps, Bogs, Etc.; NO-GO) */
    CTDB_SLOW_GO_FLAG_4 | (19 << 4) | 9,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG_4 | (7 << 4) | 10,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    (7 << 4) | 11,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    (7 << 4) | 12,

    /* (Brushland - Medium; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG_4 | (22 << 4) | 13,

    /* (Very Hard; GW, GP, GM, GC, SM, SC, CL, ML, MH, OL) */
    CTDB_NO_GO_FLAG_4 | (9 << 4) | 14,

    /* (Swamps, Bogs, Etc.; NO-GO) */
    CTDB_NO_GO_FLAG_4 | (19 << 4) | 15
};

extern void *
ctdb_convert_3_to_4(fmt3_db)
    void *fmt3_db;
{
    CTDB_3 *ctdb = (CTDB_3 *)fmt3_db;
    CTDB_4 *ctdb_out;
    
    ctdb_out = (CTDB_4 *)ctdb_alloc(NULL, sizeof(CTDB_4), 
				      "Fmt. 4 database");
    
    bzero(ctdb_out, sizeof(CTDB_4));

    /* First, copy the easy stuff over */
#define COPY_3_4(x) (ctdb_out->x = ctdb->x)

    COPY_3_4(fixed_point_basis);
    COPY_3_4(inv_fixed_point_basis);
    
    COPY_3_4(incr);
    COPY_3_4(inv_incr);
    COPY_3_4(incrD);

    COPY_3_4(patch_incr);
    COPY_3_4(patch_inv_incr);
    COPY_3_4(patch_incrD);

    COPY_3_4(pages_wide);
    COPY_3_4(pages_high);
    COPY_3_4(patches_wide);
    COPY_3_4(patches_high);
    COPY_3_4(min_x);
    COPY_3_4(max_x);
    COPY_3_4(min_y);
    COPY_3_4(max_y);
    COPY_3_4(min_x_post);
    COPY_3_4(max_x_post);
    COPY_3_4(min_y_post);
    COPY_3_4(max_y_post);
    
    ctdb_out->input_format = 4;

    COPY_3_4(min_z);
    COPY_3_4(max_z);
    COPY_3_4(origin_northing);
    COPY_3_4(origin_easting);
    COPY_3_4(origin_zone_number);
    COPY_3_4(origin_zone_letter);
    COPY_3_4(datum);

    COPY_3_4(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_3_4(is_curved);
    COPY_3_4(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_3_4(num_quad_data);
    COPY_3_4(num_nodes);
    COPY_3_4(node_size);
    COPY_3_4(num_edges);
    COPY_3_4(edge_size);

    ctdb_out->meters_to_patch = ctdb_out->patch_inv_incr * ctdb_out->inv_incr;
    ctdb_out->is_gridded = TRUE;
    ctdb_out->is_tinned = FALSE;
    ctdb_out->max_patch_number = ctdb_out->patches_wide * 
      ctdb_out->patches_high - 1;
    ctdb_out->min_patch_number = 0;
    /* Now for the real work */

    /* Soil tables */
    /* The soil table entries changed from 3 to 4.  Fix 'em. */
    {
	int32 i;
	
	for(i=0;i<16;i++)
	{
	    ctdb_out->soil_map[i] = fmt4_soils[ctdb->soil_map[i]];
	}
    }

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));
    
    /* Convert feature data over */
    ctdb_out->n_patch_bytes = convert_features_3_4(ctdb, ctdb_out);

    /* Elevations */
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 4 elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
    }
    
    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_4 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data * 
		     sizeof(CTDB_FEATURE_DATA_4),
		     "Fmt. 4 quadtree");
	
	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_4));
    }
    
    /* Topology */
    if (!ctdb->node_size)
    {
	ctdb_out->node_storage = NULL;
    }
    else
    {
	ctdb_out->node_storage = (int8 *)
	  ctdb_alloc(NULL, ctdb->node_size, "Fmt. 4 nodes");
	bcopy(ctdb->node_storage, ctdb_out->node_storage,
	      ctdb->node_size);
    }
    if (!ctdb->edge_size)
    {
	ctdb_out->edge_storage = NULL;
    }
    else
    {
	ctdb_out->edge_storage = (int8 *)
	  ctdb_alloc(NULL, ctdb->edge_size, "Fmt. 4 edges");
	bcopy(ctdb->edge_storage, ctdb_out->edge_storage,
	      ctdb->edge_size);
    }

    return ctdb_out;   
}


/* Return total feature words stored in a format 3 PG */
static int32
find_words_in_pg_3(ctdb, pg)
    CTDB_3             *ctdb;
    CTDB_PATCH_GROUP_3 *pg;
{
    int32 base;
    CTDB_FEATURE_DATA_3 *feat;

    if (pg->patch3_occupied)
    {
	base = pg->patch3_offset;
	feat = ctdb->features + pg->feature_start + base;
    }
    else if (pg->patch2_occupied)
    {
	base = pg->patch2_offset;
	feat = ctdb->features + pg->feature_start + base;
    }
    else if (pg->patch1_occupied)
    {
	base = pg->patch1_offset;
	feat = ctdb->features + pg->feature_start + base;
    }
    else if (pg->patch0_occupied)
    {
	base = 0;
	feat = ctdb->features + pg->feature_start;
    }
    else
    {
	return 0;
    }

    base += feat->hdr.size;

    return base;
}

/* Copy features from fmt. 3 database to fmt 3 database and return total
 * patch data bytes in fmt 4 db.
 */
convert_features_3_4(ctdb_3, ctdb_4)
    CTDB_3 *ctdb_3;
    CTDB_4 *ctdb_4;
{
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_4 *pg;
    CTDB_PATCH_GROUP_3 *patch_group;
    CTDB_GRID_PG_DATA_4 *data;
    int32 feat_offset = 0;
    int32 nfeat, nte;
    int32 size;
    int32 hdr_size;    
    int32 hdr_size_words;    
    
    hdr_size = sizeof(CTDB_GRID_PG_DATA_4);
    hdr_size_words = hdr_size / sizeof(CTDB_FEATURE_DATA_4);

    groups = (ctdb_3->patches_wide*ctdb_3->patches_high)/
      CTDB_PATCHES_PER_GROUP_4;
    
    ctdb_4->patch_groups = (CTDB_PATCH_GROUP_4 *)
      ctdb_alloc(NULL, sizeof(CTDB_PATCH_GROUP_4) * groups,
		 "Fmt. 4 PG headers.");
    
    bzero(ctdb_4->patch_groups, sizeof(CTDB_PATCH_GROUP_4) * groups);

    ctdb_4->n_patch_bytes = 
      (ctdb_3->num_features * sizeof(CTDB_FEATURE_DATA_3))
      + (groups * sizeof(CTDB_GRID_PG_DATA_4));
    
    ctdb_4->patch_data = (CTDB_FEATURE_DATA_4 *)
      ctdb_alloc(NULL, ctdb_4->n_patch_bytes, "Fmt. 4 patch data");
    
    pg = ctdb_4->patch_groups;
    patch_group = ctdb_3->patch_groups;
    
    for (i=0; i<groups; i++, pg++, patch_group++)
    {
	/* This will clear all tin-related fields as desired */
	bzero(pg, sizeof(pg));
	
	pg->data_start.grid_data = (CTDB_GRID_PG_DATA_4 *)
	  ((CTDB_FEATURE_DATA_4 *)ctdb_4->patch_data +
	   feat_offset);

	pg->patch0_occupied = patch_group->patch0_occupied;
	pg->patch0_dyn_terr = patch_group->patch0_dyn_terr;

	pg->mapped = patch_group->mapped;
	pg->cache = patch_group->cache;

	pg->patch1_occupied = patch_group->patch1_occupied;
	pg->patch1_dyn_terr = patch_group->patch1_dyn_terr;

	pg->patch2_occupied = patch_group->patch2_occupied;
	pg->patch2_dyn_terr = patch_group->patch2_dyn_terr;

	pg->patch3_occupied = patch_group->patch3_occupied;
	pg->patch3_dyn_terr = patch_group->patch3_dyn_terr;
	
	data = pg->data_start.grid_data;
	
	bzero(data, sizeof(CTDB_GRID_PG_DATA_4));
	
	data->patch1_offset = patch_group->patch1_offset + hdr_size_words;
	data->patch2_offset = patch_group->patch2_offset + hdr_size_words;
	data->patch3_offset = patch_group->patch3_offset + hdr_size_words;

	nfeat = find_words_in_pg_3(ctdb_3, patch_group);

	size = nfeat + hdr_size_words;
	
	bcopy(ctdb_3->features + patch_group->feature_start,
	      (int8 *)pg->data_start.grid_data + hdr_size,
	      nfeat * sizeof(CTDB_FEATURE_DATA_3));
	
	feat_offset += size;
    }
    
    return (feat_offset * sizeof(CTDB_FEATURE_DATA_4));

}
