/*
 * $RCSfile: ct_cmp_fmt4.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt4.c                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 4 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/

#include "ct_cmplr.h"
#include <libgcs.h>

/* Forward declarations */
static int32 convert_features_4_5(    
				  CTDB_4 *ctdb_4,
				  CTDB_5 *ctdb_5
				  );

static int32
convert_group_4_5(
		  CTDB_4  *ctdb_4,
		  CTDB_5  *ctdb_5,
		  int32    group,
		  ADDRESS *base,
		  int32   *num_lmodels
		  );

static void
convert_pg_4_5(
	       CTDB_4  *ctdb_4,
	       CTDB_5  *ctdb_5,
	       int32    group,
	       ADDRESS  base
	       );


static int32
convert_tin_pg_data_4_5(
			CTDB_4 *ctdb_4,
			CTDB_5 *ctdb_5,
			int32   group,
			ADDRESS base,
			int32  *num_lmodels
			);

static int32
convert_grid_pg_data_4_5(
			 CTDB_4 *ctdb_4,
			 CTDB_5 *ctdb_5,
			 int32   group,
			 ADDRESS base,
			 int32  *num_lmodels
			 );

static int32 
convert_feat_4_5(
		 CTDB_5              *ctdb_5,
		 CTDB_FEATURE_DATA_4 *feat_4,
		 CTDB_FEATURE_DATA_5 *feat_5,
		 int32               *num_lmodels
		 );

static void
convert_micro_4_5(
		  CTDB_5               *ctdb_5,
		  CTDB_FEATURE_DATA_4  *feat_4,
		  CTDB_FEATURE_DATA_5  *feat_5,
		  CTDB_FEATURE_DATA_4 **next_4,
		  CTDB_FEATURE_DATA_5 **next_5
		  );

static void
convert_building_4_5(
		     CTDB_5               *ctdb_5,
		     CTDB_FEATURE_DATA_4  *feat_4,
		     CTDB_FEATURE_DATA_5  *feat_5,
		     CTDB_FEATURE_DATA_4 **next_4,
		     CTDB_FEATURE_DATA_5 **next_5
		     );

static void
convert_tree_4_5(
		 CTDB_5               *ctdb_5,
		 CTDB_FEATURE_DATA_4  *feat_4,
		 CTDB_FEATURE_DATA_5  *feat_5,
		 CTDB_FEATURE_DATA_4 **next_4,
		 CTDB_FEATURE_DATA_5 **next_5,
		 int32                *num_lmodels
		 );

static void
convert_canopy_4_5(
		   CTDB_5               *ctdb_5,
		   CTDB_FEATURE_DATA_4  *feat_4,
		   CTDB_FEATURE_DATA_5  *feat_5,
		   CTDB_FEATURE_DATA_4 **next_4,
		   CTDB_FEATURE_DATA_5 **next_5
		   );

static void
convert_linear_4_5(
		   CTDB_5               *ctdb_5,
		   CTDB_FEATURE_DATA_4  *feat_4,
		   CTDB_FEATURE_DATA_5  *feat_5,
		   CTDB_FEATURE_DATA_4 **next_4,
		   CTDB_FEATURE_DATA_5 **next_5
		   );

static int32
find_matching_linear_model_5(
			     CTDB_5              *ctdb_5,
			     CTDB_LINEAR_MODEL_5 *model
			     );

static void 
ctdb_pack_header_4(
		   CTDB_4             *ctdb,
		   CTDB_FILE_HEADER_4 *hdr
		   );

static void
generate_caching_scheme_4(CTDB_4 *ctdb);

static int32
find_pg_size_4(CTDB_PATCH_GROUP_4 *pg);

static void
write_patch_group_headers_4(
			    CTDB_4             *ctdb,
			    CTDB_PATCH_GROUP_4 *patch_groups,
			    int32               groups,
			    FILE               *output_file
			    );

static void 
write_nodes_and_edges_4(
			CTDB_4  *ctdb,
			FILE    *output_file
			);

static void 
write_abstract_features_4(
			  CTDB_4  *ctdb,
			  FILE    *output_file
			  );

static void 
write_elevation_posts_4(
			CTDB_4  *ctdb,
			FILE    *output_file
			);

static void
write_patch_data_4(
		   CTDB_4             *ctdb,
		   FILE               *output_file
		   );

static int32
convert_features_4_3(
		     CTDB_4 *ctdb_4,
		     CTDB_3 *ctdb_3
		     );




/***************************************************************************
 * READ ROUTINES                                                           *
 ***************************************************************************/

/* $$$ I'm not going to find z-limits here; I'll wait until I get everything
 * into the compiler format so I only have to write the code once.
 */

static void 
ctdb_unpack_header_4(ctdb, hdr)
    CTDB_4             *ctdb;
    CTDB_FILE_HEADER_4 *hdr;
{
    ctdb->input_format = hdr->format;

    ctdb->is_curved  = hdr->is_curved;
    ctdb->is_sw_ne   = hdr->is_sw_ne;
    ctdb->is_gridded = hdr->is_gridded;
    ctdb->is_tinned  = hdr->is_tinned;

    ctdb->fixed_point_basis = hdr->fixed_point_basis;
    ctdb->inv_fixed_point_basis = 1.0/hdr->fixed_point_basis;

    bcopy(hdr->soil_map, ctdb->soil_map, sizeof(ctdb->soil_map));

    ctdb->incr = hdr->post_increment;
    ctdb->incrD = ctdb->incr;
    ctdb->inv_incr = 1.0/ctdb->incrD;

    ctdb->patch_incr = hdr->posts_per_patch_side;
    ctdb->patch_incrD = ctdb->patch_incr;
    ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide = hdr->pages_wide;
    ctdb->pages_high = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE / ctdb->patch_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE / ctdb->patch_incr;

    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);
    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);

    bcopy(hdr->patch_cache_size_map, ctdb->patch_cache_size_map,
	  CTDB_PATCH_CACHES_4 * sizeof(int32));

    ctdb->origin_northing = hdr->origin_northing;
    ctdb->origin_easting = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum = hdr->datum;
    bcopy(hdr->name, ctdb->name, sizeof(hdr->name));
    ctdb->version = hdr->version;
    bcopy(hdr->date, ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date, "\n");

    ctdb->num_nodes = hdr->num_nodes;
    ctdb->node_size = hdr->node_size;
    ctdb->num_edges = hdr->num_edges;
    ctdb->edge_size = hdr->edge_size;
    ctdb->num_quad_data = hdr->num_quad_data;
    ctdb->n_patch_bytes = hdr->n_patch_bytes;
    ctdb->meters_to_patch = ctdb->patch_inv_incr * ctdb->inv_incr;
    ctdb->max_patch_number = ctdb->patches_wide * ctdb->patches_high - 1;
    ctdb->min_patch_number = 0;
}

static CTDB_STATUS
read_patch_group_headers_4(ctdb, fileptr)
    CTDB_4 *ctdb ;
    FILE   *fileptr;    
{
  int32 groups;

  /* Read in all the patch group headers */
  groups=(ctdb->patches_wide*ctdb->patches_high)/CTDB_PATCHES_PER_GROUP_4;
  ctdb->patch_groups=(CTDB_PATCH_GROUP_4 *)
    ctdb_alloc(NULL,
	       sizeof(CTDB_PATCH_GROUP_4)*groups,
	       "patch group data");
  fread(ctdb->patch_groups,sizeof(CTDB_PATCH_GROUP_4),groups,fileptr);

  return CTDB_STATUS_OK;
}
 
static void 
init_null_topology_4(ctdb)
    CTDB_4 *ctdb;
{
    ctdb->node_storage = (int8 *)
      ctdb_alloc(NULL, sizeof(CTDB_NODE_4), "initialization");
    bzero(ctdb->node_storage, sizeof(CTDB_NODE_4));
    ctdb->edge_storage = (int8 *)
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_4), "initialization");
    bzero(ctdb->edge_storage, sizeof(CTDB_EDGE_4));
    ctdb->nodes = (CTDB_NODE_4 **)
      ctdb_alloc(NULL, sizeof(CTDB_NODE_4 *), "initialization");
    ctdb->nodes[0] = (CTDB_NODE_4 *)ctdb->node_storage;
    ctdb->edges = (CTDB_EDGE_4 **)
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_4 *), "initialization");
    ctdb->edges[0] = (CTDB_EDGE_4 *)ctdb->edge_storage;
}

static CTDB_STATUS 
ctdb_read_topology_4(ctdb, fileptr)
    CTDB_4 *ctdb;
    FILE   *fileptr;
{
    int32 i, offset;
    
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces: ctdb_read_topology((CTDB *)ctdb,fileptr);*/
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    if (ctdb->num_nodes)
    {
        ctdb->nodes = NULL;
	ctdb->nodes =
	  (CTDB_NODE_4 **)ctdb_alloc(NULL,
				     sizeof(CTDB_NODE_4 *) *
				     ctdb->num_nodes,
				     "node index table");
	ctdb->node_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->node_size, "nodes");

	if (fread(ctdb->node_storage, ctdb->node_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Nodes\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_NODES_FAILED;
	}
    }

    if (ctdb->num_edges)
    {
        ctdb->edges = NULL;
	ctdb->edges =
	  (CTDB_EDGE_4 **)ctdb_alloc(NULL,
				     sizeof(CTDB_EDGE_4 *) *
				     ctdb->num_edges,
				     "edge index table");
	ctdb->edge_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->edge_size, "edges");

	if (fread(ctdb->edge_storage, ctdb->edge_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Edges\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_EDGES_FAILED;
	}
    }

    /* Fill in the node index table */
    bzero(ctdb->nodes, sizeof(CTDB_NODE_4 *) * ctdb->num_nodes);
    for(i=offset=0;(i<ctdb->num_nodes) && (offset<ctdb->node_size);i++)
    {
	ctdb->nodes[i] = (CTDB_NODE_4 *)(ctdb->node_storage + offset);
	offset += sizeof(CTDB_NODE_4) + sizeof(struct ctdb_connection) *
	  (ctdb->nodes[i]->num_connections - 1);
    }

    /* Fill in the edge index table */
    bzero(ctdb->edges, sizeof(CTDB_EDGE_4 *) * ctdb->num_edges);
    for(i=offset=0;(i<ctdb->num_edges) && (offset<ctdb->edge_size);i++)
    {
	ctdb->edges[i] = (CTDB_EDGE_4 *)(ctdb->edge_storage + offset);
	offset += sizeof(CTDB_EDGE_4) + sizeof(struct ctdb_connection) *
	  (ctdb->edges[i]->num_features - 1);
    }

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_quad_tree_4(ctdb, fileptr)
    CTDB_4 *ctdb;
    FILE   *fileptr;
{
    if (ctdb->num_quad_data == 0 )
      return CTDB_STATUS_OK;
        
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    ctdb->quad_data =
      (CTDB_FEATURE_DATA_4 *)ctdb_alloc(NULL,
					sizeof(CTDB_FEATURE_DATA_4) *
					ctdb->num_quad_data,
					"abstract data and quadtree");

    if (fread(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_4),
	      ctdb->num_quad_data, fileptr) != ctdb->num_quad_data)
    {
	printf("CTDB: Read failed.  Quad tree and abstract data\n");
	/* Survive this error */
	ctdb->num_quad_data = 0;
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS
ctdb_read_all_patch_data_4(ctdb, f)
    CTDB_4 *ctdb;
    FILE   *f;
{
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_4 *pg;

    ctdb->patch_data =
      (ADDRESS)ctdb_alloc(NULL,
			  ctdb->n_patch_bytes,
			  "DB Patch data");

    /* Read _all_ the patch data for the cell in in one big gulp */
    fread(ctdb->patch_data,1,ctdb->n_patch_bytes,f);

    pg = ctdb->patch_groups;

    /* Now that we have the data, we need to set up the pointers in all the
     * patch headers
     */
    groups = (ctdb->patches_wide*ctdb->patches_high) / 
      CTDB_PATCHES_PER_GROUP_4;

    for (i=0;i<groups;i++)
    {
	/* Doesn't matter which elt of the union we choose here... */
	pg[i].data_start.grid_data = 
	  (CTDB_GRID_PG_DATA_4*)((int8*)ctdb->patch_data +
				 (int32)pg[i].data_start.grid_data);
	pg[i].mapped = 1;
    }

    return CTDB_STATUS_OK;    
}

void 
ctdb_read_elevations_4(ctdb, f)
    CTDB_4 *ctdb;
    FILE   *f;
{
    uint32 memory, pages;
    int    file_pos;

    pages = ctdb->pages_wide * ctdb->pages_high;

    memory = PAGE_SIZE * pages;

    ctdb->elevations_resident = 1;
    ctdb->elevations = (int32 *)ctdb_alloc(NULL,
					   memory, "elevation data");

    /* Note the file pointer value */
    file_pos = ftell(f);

    /* Load as many pages as the caller will let us */
    if (fread(ctdb->elevations, memory, 1, f) != 1)
      printf("CTDB: Read failed.  First %d bytes of post data.\n",
	     memory), exit(1);

    /* Position the file pointer at the end of the elevations */
    file_pos += PAGE_SIZE * pages;
    fseek(f, file_pos, 0);
}


static void
ctdb_reread_4(fname, ctdb)
    char   *fname;
    CTDB_4 *ctdb;
{
    FILE *f;
    CTDB_FILE_HEADER_4 hdr;
    int32 total_mem;

    if (!fname)
    {
	printf("Invalid input database.\n");
	exit(-1);
    }

    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	exit(-1);
    }

    if (hdr.format != 4)
    {
	printf("Attempted to read %d format db with fmt 4 reader!\n",
	       hdr.format);	
	fclose(f);
	exit(-1);
    }

    ctdb->file = f;

    ctdb_unpack_header_4(ctdb, &hdr);

    read_patch_group_headers_4(ctdb, f);

    ctdb_read_topology_4(ctdb, f);

    ctdb_read_quad_tree_4(ctdb, f);

    if ( ctdb->is_gridded )
      ctdb_read_elevations_4(ctdb, f);

    ctdb->patch_data_seek = ftell(f);

    ctdb_read_all_patch_data_4(ctdb, f);

    return;
}

extern void *
ctdb_read_4(fname)
    char *fname;
{
    CTDB_4 *ctdb;
    
    ctdb = (CTDB_4 *)ctdb_alloc(NULL, sizeof(CTDB_4), "Creating fmt 4 ctdb");
    bzero(ctdb, sizeof(CTDB_4));

    printf("Attempting to read CTDB datafiles... \n");

    ctdb_reread_4(fname, ctdb);
    return ctdb;  
}

/***************************************************************************
 * WRITE ROUTINES                                                          *
 ***************************************************************************/

/* The input database is a void pointer to make it easier to call this routine
 * through a generic interface; it is the caller's responsibility to ensure
 * that it is, in fact, a format 4 CTDB.
 * NOTE: since the compiler does not use caching, the inter-format conversion
 * routines do NOT generate the caching scheme correctly; therefore it is the
 * responsibility of the write routine to do so.
 */
void ctdb_write_4(fmt4_db, output_file)
    void *fmt4_db;
    FILE *output_file;
{
    CTDB_4 *ctdb = (CTDB_4 *)fmt4_db;
    CTDB_FILE_HEADER_4 hdr;

    int32 groups = (ctdb->patches_wide *
		    ctdb->patches_high)/CTDB_PATCHES_PER_GROUP_4;

    /* Generate the caching scheme for the patch data and fill in the
     * necessary data structures
     */
    generate_caching_scheme_4(ctdb);

    /**
     ** Write the header to the output file.
     **/
    
    ctdb_pack_header_4(ctdb, &hdr);
    fwrite(&hdr, sizeof(hdr), 1, output_file);

    /**
     ** Write patch group headers
     **/
    write_patch_group_headers_4(ctdb, ctdb->patch_groups, groups, output_file);

    /**
     ** Write topology: the nodes and edges to the output file.
     **/
    write_nodes_and_edges_4(ctdb, output_file);

    /** 
     ** Write quad tree: the abstract feature quad tree to the output
     **          file.
     **/
    write_abstract_features_4(ctdb, output_file);

    /**
     ** Write elevation Posts: the elevations if there are gridded patches.
     **/
    if ( ctdb->is_gridded )
      write_elevation_posts_4(ctdb, output_file);
    
    /**
     ** Write patch group data: the patch group pointers, features, terrain
     ** elements, if tinned; the virtual grids, if tinned. 
     **/
    write_patch_data_4(ctdb, output_file);

    return;
}

/* The ctdb_pack_header_4 routine will create a file header (hdr) from the
 * fmt 5 ctdb data structure. 
 */
static void 
ctdb_pack_header_4(ctdb, hdr)
    CTDB_4             *ctdb;
    CTDB_FILE_HEADER_4 *hdr;
{
    bzero(hdr, sizeof(CTDB_FILE_HEADER_4));
    hdr->format = ctdb->input_format;
    hdr->is_curved = ctdb->is_curved;
    hdr->is_sw_ne = ctdb->is_sw_ne;
    hdr->is_gridded = ctdb->is_gridded;
    hdr->is_tinned  = ctdb->is_tinned;
    hdr->fixed_point_basis = ctdb->fixed_point_basis;
    bcopy(ctdb->soil_map, hdr->soil_map, sizeof(hdr->soil_map));

    hdr->post_increment = ctdb->incr;
    hdr->posts_per_patch_side = ctdb->patch_incr;
    hdr->pages_wide = ctdb->pages_wide;
    hdr->pages_high = ctdb->pages_high;
    hdr->min_x = ctdb->min_x;
    hdr->max_x = ctdb->max_x;
    hdr->min_y = ctdb->min_y;
    hdr->max_y = ctdb->max_y;
    bcopy(ctdb->patch_cache_size_map, hdr->patch_cache_size_map,
	  sizeof(ctdb->patch_cache_size_map));

    hdr->origin_northing = ctdb->origin_northing;
    hdr->origin_easting = ctdb->origin_easting;
    hdr->origin_zone_number = ctdb->origin_zone_number;
    hdr->origin_zone_letter = ctdb->origin_zone_letter;
    hdr->datum = ctdb->datum;
    bcopy(ctdb->name, hdr->name, sizeof(ctdb->name));
    hdr->version = ctdb->version;
    bcopy(ctdb->date, hdr->date, sizeof(ctdb->date));
    hdr->num_nodes = ctdb->num_nodes;
    hdr->node_size = ctdb->node_size;
    hdr->num_edges = ctdb->num_edges;
    hdr->edge_size = ctdb->edge_size;
    hdr->num_quad_data = ctdb->num_quad_data;
    hdr->n_patch_bytes = ctdb->n_patch_bytes;
}

/* Generate the caching scheme for the patch data and fill in the
 * necessary data structures
 */
static void
generate_caching_scheme_4(ctdb)
    CTDB_4 *ctdb;
{
    int32 i;
    int32 group;
    int32 num_groups;
    int32 max_size = 0;
    CTDB_PATCH_GROUP_4 *pg;
    int32 *pg_size_map;		/* pg_size_map[i] is the size, in words, of
				 * patch group i
				 */
    int32 *map;			/* Current entry in pg_size_map array */
    int32 *counts;		/* counts[i] is the number of patch groups
				 * with i words of data
				 */
    int32 size;			/* Current PG size we're looking at */
    int32 last_size;		/* Last non-empty PG size we looked at */
    int32 *sizes = ctdb->patch_cache_size_map; /* Sizes of the patch caches */
    int32 burdens[CTDB_PATCH_CACHES_4];	/* Number of PG's mapped to each 
					 * cache 
					 */
    int32 n_mapped;		/* Ideal number of groups per cache */
    int32 cur_count;		/* Groups mapped to current cache so far */
    int32 cur_cache;		/* Which cache we're currently filling */
    int32 min_burden;		/* Number of PGs mapped to least used cache */
    int32 smallest;
    int32 max_burden;		/* Number of PGs mapped to most used cache */
    int32 largest;		/* Most used cache */
    int32 best_cache;		/* "Best" cache for current PG */

    num_groups = (ctdb->patches_wide * ctdb->patches_high) /
      CTDB_PATCHES_PER_GROUP_4;

    pg_size_map = (int32 *)ctdb_alloc(NULL, num_groups * sizeof(int32),
				      "Patch group size map");


    /* 1.  Find the size of each patch group.  While we're at it, keep track
     * of how big the largest group is.
     */
    for(group = 0; group < num_groups; group++)
    {
	pg_size_map[group] = find_pg_size_4(ctdb->patch_groups + group) / 
	  CTDB_CACHE_BKT_UNITS_4;
	if (pg_size_map[group] > max_size)
	  max_size = pg_size_map[group];
    }

    max_size++;             /* This MUST be > actual max size for this code
			     * to work
			     */

    counts = (int32 *)ctdb_alloc(NULL, sizeof(int32) * max_size,
				"patch group size counts");

    bzero(counts, max_size * sizeof(int32));
    bzero(sizes, CTDB_PATCH_CACHES_4 * sizeof(int32));
    bzero(burdens, CTDB_PATCH_CACHES_4 * sizeof(int32));

    /* 2. Find distribution of group sizes. 
     */
    map = pg_size_map;
    for(group=0;group<num_groups;group++,map++)
      counts[*map]++;

    /* 3. Choose the ideal number of groups to be mapped to each cache.
     */
    n_mapped =
      ceil((float64)num_groups/(float64)CTDB_PATCH_CACHES_4) + 1;

    printf("Ideal groups/cache = %d\n",n_mapped);

    /* 4. Choose cache sizes such that AT LEAST n_mapped groups land
     * in each cache.
     */
    cur_count = cur_cache = 0;
    for(size=0;size<max_size;size++)
    {
	if (counts[size])
	{
	    cur_count += counts[size];
	    last_size = size;
	}

	if (cur_count > n_mapped)
	{
	    do
	    {
		sizes[cur_cache] = size;
		if (cur_count >= 2*n_mapped)
		  burdens[cur_cache] = n_mapped;
		else
		  burdens[cur_cache] = cur_count;
		cur_cache++;
		cur_count -= n_mapped;
	    } while (cur_count >= n_mapped);
	    cur_count = 0;
	}
    }
    sizes[cur_cache] = last_size;
    burdens[cur_cache] = cur_count;
    cur_cache++;

    /* 5. If not all the caches were used, split the most burdened
     * caches in two.
     */
    while(cur_cache<CTDB_PATCH_CACHES_4)
    {
	/* Find the worst off */
	max_burden = largest = 0;
	for(i=0;i<cur_cache;i++)
	  if (burdens[i] > max_burden)
	  {
	      max_burden = burdens[i];
	      largest = i;
	  }
	/* Make room for the split */
	for(i=cur_cache;i>largest;i--)
	{
	    sizes[i] = sizes[i-1];
	    burdens[i] = burdens[i-1];
	}
	/* Split it in two */
	burdens[largest] /= 2;
	burdens[largest+1] -= burdens[largest];
	cur_cache += 1;
    }

    /* 6. Tag the groups with their cache number.
     */
    bzero(burdens,sizeof(burdens));
    for(group=0;group<num_groups;group++)
    {
	pg = ctdb->patch_groups + group;
	map = pg_size_map + group;

	/* Search through the caches and find which ones we can put this
	 * patch into.  Choose the least burdened of those.
	 */
	min_burden = smallest = 0x7FFFFFF;
	best_cache = CTDB_PATCH_CACHES_4 - 1;
	for(cur_cache=0; cur_cache < CTDB_PATCH_CACHES_4; cur_cache++)
	{
	    /* Have we passed the range of caches for which this size
	     * is intended?
	     */
	    if (sizes[cur_cache] > smallest)
	      break;
	      
	    /* Will the patch fit in this cache? */
	    if (sizes[cur_cache] < *map)
	      continue;
	    smallest = sizes[cur_cache];
	    if (burdens[cur_cache] < min_burden)
	    {
		best_cache = cur_cache;
		min_burden = burdens[cur_cache];
	    }
	}
	burdens[best_cache]++;
	pg->cache = best_cache;
    }

    /* 7. Print the results.
     */
    size = 0;
    for (cur_cache=0; cur_cache < CTDB_PATCH_CACHES_4; cur_cache++)
    {
	if (cur_cache && (sizes[cur_cache] != sizes[cur_cache-1]))
	  size = sizes[cur_cache-1];
	printf("Cache %x: [%d->%d].  %d patch groups\n",
	       cur_cache, size + 1, sizes[cur_cache], burdens[cur_cache]);
    }

    free(counts);    
    return;
}

/* Return the size, in bytes, of all data, including the data header, in
 * pg
 */
static int32
find_pg_size_4(pg)
    CTDB_PATCH_GROUP_4 *pg;
{
    int32 size;

    if (pg->contains_tins)
    {
	int32 mem = -1;
	int32 num_vg;
	CTDB_TIN_PG_DATA_4 *pgt = pg->data_start.tin_data;

	size = sizeof(CTDB_TIN_PG_DATA_4);

	if (!pg->patch3_no_tes)
	{
	    mem = 3;
	}
	else if (!pg->patch2_no_tes)
	{
	    mem = 2;
	}
	else if (!pg->patch1_no_tes)
	{
	    mem = 1;
	}
	else
	{
	    mem = 0;
	}

	/* This gives us the number of bytes from the start of the data header
	 * to the start of the last set of vgrid data
	 */
	size = pgt->te_vgrid_offset[mem] * sizeof(CTDB_FEATURE_DATA_4);

	num_vg = pgt->te_vgrid_dim[mem] * pgt->te_vgrid_dim[mem];

	/* Round up to word boundary, remembering that te_vgrid entries
	 * are uint16s
	 */
	num_vg += (num_vg % 2);

	size += (num_vg * sizeof(uint16));
	return size;
    }
    else
    {
	/* Find the last occupied patch.  Add its offset to the number of bytes
	 * of data stored for it to get the total PG data size
	 */
	CTDB_FEATURE_DATA_4 *feat = NULL;

	if (pg->patch3_occupied)
	{
	    size = pg->data_start.grid_data->patch3_offset *
	      sizeof(CTDB_FEATURE_DATA_4);
	    feat = (CTDB_FEATURE_DATA_4 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch3_offset;
	}
	else if (pg->patch2_occupied)
	{
	    size = pg->data_start.grid_data->patch2_offset *
	      sizeof(CTDB_FEATURE_DATA_4);
	    feat = (CTDB_FEATURE_DATA_4 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch2_offset;
	}
	else if (pg->patch1_occupied)
	{
	    size = pg->data_start.grid_data->patch1_offset *
	      sizeof(CTDB_FEATURE_DATA_4);
	    feat = (CTDB_FEATURE_DATA_4 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch1_offset;
	}
	else if (pg->patch0_occupied)
	{
	    size = sizeof(CTDB_GRID_PG_DATA);

	    feat = (CTDB_FEATURE_DATA_4 *)pg->data_start.grid_data +
	      (sizeof(CTDB_GRID_PG_DATA_4) / sizeof(CTDB_FEATURE_DATA_4));
	}
	if (feat == NULL)
	  return sizeof(CTDB_GRID_PG_DATA_4);

	size += (feat->hdr.size * sizeof(CTDB_FEATURE_DATA_4 *));
	return size;
    }
}

/* This routine writes the complete set of patch group headers. 
 * $$$ WARRNING:  as written, this changes the data_start values to be
 * offsets rather than pointers, rendering the database unusable in memory.
 * If you need to be able to use the database after writing it, this routine
 * will need to change
 */
static void
write_patch_group_headers_4(ctdb, patch_groups, groups, output_file)
    CTDB_4             *ctdb;
    CTDB_PATCH_GROUP_4 *patch_groups;
    int32               groups;
    FILE               *output_file;
{
    int32 i;

    for(i=0;i<groups;i++)
    {
	ctdb->patch_groups[i].data_start.offset = 
	  (int8 *)ctdb->patch_groups[i].data_start.grid_data -
	  (int8 *)ctdb->patch_data;
    }
	
    fwrite(patch_groups, sizeof(CTDB_PATCH_GROUP_4),groups, output_file);    
}

/* This function writes the nodes and edges to the output file.
 */
static void 
write_nodes_and_edges_4(ctdb, output_file)
    CTDB_4  *ctdb;
    FILE    *output_file;
{
    /* Just dump the stored nodes and edges on out */
    if (ctdb->node_size)
      fwrite(ctdb->node_storage, 1, ctdb->node_size, output_file);
    if (ctdb->edge_size)
      fwrite(ctdb->edge_storage, 1, ctdb->edge_size, output_file);
}

/* This function writes the abstract features to the output file. */
static void 
write_abstract_features_4(ctdb, output_file)
    CTDB_4  *ctdb;
    FILE    *output_file;
{
    /* Just dump the quad tree on out */
    if (ctdb->num_quad_data)
      fwrite(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_4),
	     ctdb->num_quad_data, output_file);
}

/* This function writes the elevation post data out to the data file.
   Note that this assumes all elevation posts are already in memory.
 */
static void 
write_elevation_posts_4(ctdb, output_file)
    CTDB_4  *ctdb;
    FILE    *output_file;
{
    int32 post_bytes;
    
    post_bytes = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
        
    fwrite(ctdb->elevations, post_bytes,1, output_file);
}

/* Writes the patch data out to the file */
static void
write_patch_data_4(ctdb, output_file)
    CTDB_4             *ctdb;
    FILE               *output_file;
{
    int32 i;
    int8 pad_byte = 0;
    int32 num_pad_bytes;

    /* Dump the patch data on out */
    fwrite(ctdb->patch_data, 1, ctdb->n_patch_bytes, output_file);

    /* To be perfectly safe, we'll pad the database out by the size of the
     * largest cache; this will ensure that the caching scheme does not
     * read past the end of the database.
     */
    num_pad_bytes = ctdb->patch_cache_size_map[CTDB_PATCH_CACHES_4 - 1] *
      CTDB_CACHE_BKT_UNITS_4;

    for(i = 0; i<num_pad_bytes; i++)
      fwrite(&pad_byte, sizeof(int8), 1, output_file);
    
    return;
}

/***************************************************************************
 * CLEANUP ROUTINES                                                          *
 ***************************************************************************/
extern void
ctdb_free_fmt4_db(fmt4_db)
    void *fmt4_db;
{
    CTDB_4 *ctdb = (CTDB_4 *)fmt4_db;
    
#define CTDB_FREE_4(x)        \
    if (ctdb->x)              \
        STDDEALLOC(ctdb->x)

    if (!ctdb)
      return;

    CTDB_FREE_4(patch_groups);
    CTDB_FREE_4(cache_page_numbers);
    CTDB_FREE_4(elevation_pages);
    CTDB_FREE_4(elevations);
    CTDB_FREE_4(quad_data);
    CTDB_FREE_4(nodes);
    CTDB_FREE_4(node_storage);
    CTDB_FREE_4(edges);
    CTDB_FREE_4(edge_storage);
    CTDB_FREE_4(patch_data);
    
    STDDEALLOC(ctdb);
    return;
}

/***************************************************************************
 * 4 => 3 CONVERSION ROUTINES                                              *
 ***************************************************************************/

extern void *
ctdb_convert_4_to_3(fmt4_db)
    void *fmt4_db;
{
    CTDB_4 *ctdb = (CTDB_4 *)fmt4_db;
    CTDB_3 *ctdb_out;
    
    ctdb_out = (CTDB_3 *)ctdb_alloc(NULL, sizeof(CTDB_3), 
				      "Fmt. 3 database");
    
    bzero(ctdb_out, sizeof(CTDB_3));

    /* First, copy the easy stuff over */
#define COPY_4_3(x) (ctdb_out->x = ctdb->x)

    COPY_4_3(fixed_point_basis);
    COPY_4_3(inv_fixed_point_basis);
    
    COPY_4_3(incr);
    COPY_4_3(inv_incr);
    COPY_4_3(incrD);

    COPY_4_3(patch_incr);
    COPY_4_3(patch_inv_incr);
    COPY_4_3(patch_incrD);

    COPY_4_3(pages_wide);
    COPY_4_3(pages_high);
    COPY_4_3(patches_wide);
    COPY_4_3(patches_high);
    COPY_4_3(min_x);
    COPY_4_3(max_x);
    COPY_4_3(min_y);
    COPY_4_3(max_y);
    COPY_4_3(min_x_post);
    COPY_4_3(max_x_post);
    COPY_4_3(min_y_post);
    COPY_4_3(max_y_post);
    
    ctdb_out->input_format = 3;    

    COPY_4_3(min_z);
    COPY_4_3(max_z);
    COPY_4_3(origin_northing);
    COPY_4_3(origin_easting);
    COPY_4_3(origin_zone_number);
    COPY_4_3(origin_zone_letter);
    COPY_4_3(datum);

    COPY_4_3(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_4_3(is_curved);
    COPY_4_3(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_4_3(num_quad_data);
    COPY_4_3(num_nodes);
    COPY_4_3(node_size);
    COPY_4_3(num_edges);
    COPY_4_3(edge_size);

    /* Now for the real work */

    /* Soil tables */
    {
      int32 i;
      
      /* Format 4 added lots of stuff here.  All we want is the SIMNET 
       * index
       */
      for(i=0;i<CTDB_N_SOILS_3;i++)
        ctdb_out->soil_map[i] = ctdb->soil_map[i] & 0xF;
    }    

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));
    
    /* Convert feature data over */
    ctdb_out->num_features = convert_features_4_3(ctdb, ctdb_out);

    /* Elevations */
    if (!ctdb->is_gridded)
    {
	printf("Error:  attempting to do a simple conversion from format 4\n");
	printf("tinned to format 3--the compiler should not allow this \n");
	printf("operation.\n");
	exit(-1);
    }
    else
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 3 elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
    }
    
    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_3 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data * 
		     sizeof(CTDB_FEATURE_DATA_3),
		     "Fmt. 5 quadtree");
	
	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_3));
    }
    
    /* Topology */
    if (!ctdb->node_size)
    {
	ctdb_out->node_storage = NULL;
    }
    else
    {
	ctdb_out->node_storage = (int8 *)
	  ctdb_alloc(NULL, ctdb->node_size, "Fmt. 3 nodes");
	bcopy(ctdb->node_storage, ctdb_out->node_storage,
	      ctdb->node_size);
    }
    if (!ctdb->edge_size)
    {
	ctdb_out->edge_storage = NULL;
    }
    else
    {
	ctdb_out->edge_storage = (int8 *)
	  ctdb_alloc(NULL, ctdb->edge_size, "Fmt. 3 edges");
	bcopy(ctdb->edge_storage, ctdb_out->edge_storage,
	      ctdb->edge_size);
    }

    return ctdb_out;
}

/* Return total feature words stored in a format 4 PG */
static int32
find_words_in_pg_4(pg)
    CTDB_PATCH_GROUP_4 *pg;
{
    int32 base;
    CTDB_GRID_PG_DATA_4 *data;
    int32 hdr_size;
    CTDB_FEATURE_DATA_4 *feat;

    data = pg->data_start.grid_data;
    hdr_size = sizeof(CTDB_GRID_PG_DATA_4) / sizeof(CTDB_FEATURE_DATA_4);

    if (pg->patch3_occupied)
    {
	base = data->patch3_offset - hdr_size;
	feat = (CTDB_FEATURE_DATA_4 *)data + data->patch3_offset;
    }
    else if (pg->patch2_occupied)
    {
	base = data->patch2_offset - hdr_size;
	feat = (CTDB_FEATURE_DATA_4 *)data + data->patch2_offset;
    }
    else if (pg->patch1_occupied)
    {
	base = data->patch1_offset - hdr_size;
	feat = (CTDB_FEATURE_DATA_4 *)data + data->patch1_offset;
    }
    else if (pg->patch0_occupied)
    {
	base = 0;
	feat = (CTDB_FEATURE_DATA_4 *)data + hdr_size;
    }
    else
    {
	return 0;
    }

    base += feat->hdr.size;

    return base;
}

/* Move all feature data from fmt 4 db into fmt 3 db.  Returns number of
 * feature words in fmt 3 db.
 */
static int32
convert_features_4_3(ctdb_4, ctdb_3)
    CTDB_4 *ctdb_4;
    CTDB_3 *ctdb_3;
{
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_3 *pg;
    CTDB_PATCH_GROUP_4 *patch_group;
    CTDB_GRID_PG_DATA_4 *data;
    int32 current, next;
    int32 feat_offset = 0;
    int32 te_offset = 0;
    int32 nfeat, nte;
    int32 size;
    int32 hdr_size;		/* In words */
    
    /* All fmt. 4 headers are assumed to be gridded since we're converting
     * to format 3.
     */
    hdr_size = sizeof(CTDB_GRID_PG_DATA_4) / sizeof(CTDB_FEATURE_DATA_4);

    groups = (ctdb_4->patches_wide*ctdb_4->patches_high)/
      CTDB_PATCHES_PER_GROUP_3;
    
    ctdb_3->patch_groups = (CTDB_PATCH_GROUP_3 *)
      ctdb_alloc(NULL, sizeof(CTDB_PATCH_GROUP_3) * groups,
		 "Fmt. 3 feature data");
    
    bzero(ctdb_3->patch_groups, sizeof(CTDB_PATCH_GROUP_3) * groups);

    /* We're overestimating here. */
    ctdb_3->features = (CTDB_FEATURE_DATA_3 *)
      ctdb_alloc(NULL, ctdb_4->n_patch_bytes, "Fmt. 3 features");
    
    pg = ctdb_3->patch_groups;
    patch_group = ctdb_4->patch_groups;
    
    for (i=0; i<groups; i++, pg++, patch_group++)
    {
	if (patch_group->contains_tins)
	{
	    printf("Error: attempting to convert tinned database to \n");
	    printf("format 3.  The compiler should not allow this.\n");
	    exit(-1);
	}

	data = patch_group->data_start.grid_data;

	bzero(pg, sizeof(pg));
	
	pg->feature_start = feat_offset;

	pg->patch0_occupied = patch_group->patch0_occupied;
	pg->patch0_dyn_terr = patch_group->patch0_dyn_terr;

	pg->mapped = patch_group->mapped;
	pg->cache = patch_group->cache;

	pg->patch1_occupied = patch_group->patch1_occupied;
	pg->patch1_dyn_terr = patch_group->patch1_dyn_terr;
	pg->patch1_offset = data->patch1_offset - hdr_size;

	pg->patch2_occupied = patch_group->patch2_occupied;
	pg->patch2_dyn_terr = patch_group->patch2_dyn_terr;
	pg->patch2_offset = data->patch2_offset - hdr_size;

	pg->patch3_occupied = patch_group->patch3_occupied;
	pg->patch3_dyn_terr = patch_group->patch3_dyn_terr;
	pg->patch3_offset = data->patch3_offset - hdr_size;
	
	if (i < (groups - 1))
	{
	    current = (int32)patch_group->data_start.grid_data;
	    next = (int32)((patch_group + 1)->data_start.grid_data);
	    size = (next - current)/sizeof(CTDB_FEATURE_DATA);
	    nfeat = size - (sizeof(CTDB_GRID_PG_DATA_4) / 
			    sizeof(CTDB_FEATURE_DATA_4));
	}
	else
	{
	    /* We'll calculate size the hard way only if we have to */
	    nfeat = find_words_in_pg_4(patch_group);
	}
	
	bcopy(((int8 *)patch_group->data_start.grid_data + 
	       sizeof(CTDB_GRID_PG_DATA_4)),
	      ctdb_3->features+feat_offset,
	      nfeat * sizeof(CTDB_FEATURE_DATA_4));
	
	feat_offset += nfeat;
    }
    
    return feat_offset;
}


/***************************************************************************
 * 4 => 5 CONVERSION ROUTINES                                              *
 ***************************************************************************/

extern void *
ctdb_convert_4_to_5(fmt4_db)
    void *fmt4_db;
{
    CTDB_4 *ctdb = (CTDB_4 *)fmt4_db;
    CTDB_5 *ctdb_out;
    
    ctdb_out = (CTDB_5 *)ctdb_alloc(NULL, sizeof(CTDB_5), 
				      "Fmt. 5 database");
    
    bzero(ctdb_out, sizeof(CTDB_5));

    /* First, copy the easy stuff over */
#define COPY_4_5(x) (ctdb_out->x = ctdb->x)

    COPY_4_5(fixed_point_basis);
    COPY_4_5(inv_fixed_point_basis);
    
    
    /* Leave water_char stuff at 0 */
    COPY_4_5(incr);
    COPY_4_5(inv_incr);
    COPY_4_5(incrD);
    COPY_4_5(patch_incr);
    COPY_4_5(patch_inv_incr);
    COPY_4_5(patch_incrD);

    COPY_4_5(pages_wide);
    COPY_4_5(pages_high);
    COPY_4_5(patches_wide);
    COPY_4_5(patches_high);
    COPY_4_5(min_x);
    COPY_4_5(max_x);
    COPY_4_5(min_y);
    COPY_4_5(max_y);
    COPY_4_5(min_x_post);
    COPY_4_5(max_x_post);
    COPY_4_5(min_y_post);
    COPY_4_5(max_y_post);
    
    ctdb_out->input_format = 5;
    
    COPY_4_5(min_z);
    COPY_4_5(max_z);
    COPY_4_5(origin_northing);
    COPY_4_5(origin_easting);
    COPY_4_5(origin_zone_number);
    COPY_4_5(origin_zone_letter);
    COPY_4_5(datum);
    ctdb_out->cell_id = GCS_ILLEGAL_CELL;
    ctdb_out->origin_x = -(ctdb->min_x + ctdb->max_x) / 2;
    ctdb_out->origin_y = -(ctdb->min_y + ctdb->max_y) / 2;

    ctdb_out->gcs_mode = CTDB_MODE_SIMNET;

    /* $$$ Leaving lat/long bounds blank for now, since the compiler
     * won't use them anyway
     */

    COPY_4_5(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_4_5(is_curved);
    COPY_4_5(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_4_5(num_quad_data);
    COPY_4_5(num_nodes);
    COPY_4_5(num_edges);

    /* $$$ Leaving the models NULLed out for now, I don't *think* this is
     * a problem.
     */

    COPY_4_5(meters_to_patch);
    COPY_4_5(is_gridded);
    COPY_4_5(is_tinned);
    COPY_4_5(max_patch_number);
    COPY_4_5(min_patch_number);
    
    /* Now for the real work */

    /* Soil tables */
    ctdb_out->num_soil_tables = 1;
    ctdb_out->soil_tables = (int32 **)ctdb_alloc(NULL, sizeof(int32 *),
						 "Fmt.4 soil table ptr");
    ctdb_out->soil_table_size = CTDB_N_SOILS_4 * sizeof(int32);
    ctdb_out->soil_table_storage = 
      (int8 *)ctdb_alloc(NULL, ctdb_out->soil_table_size,
			 "Fmt. 4 soil table storage");
    
    ctdb_out->soil_tables[0] = (int32 *)ctdb_out->soil_table_storage;
    
    bcopy(ctdb->soil_map, ctdb_out->soil_tables[0],
	  CTDB_N_SOILS_4 * sizeof(int32));

    /* Create the default tidal zones */
    ctdb_out->num_tidal_zones = 2;
    ctdb_out->tidal_zones = (CTDB_TIDAL_ZONE_5 *)
      ctdb_alloc(NULL, sizeof(CTDB_TIDAL_ZONE_5) * ctdb_out->num_tidal_zones,
		 "Fmt. 5 tidal zones.");
    bzero(ctdb_out->tidal_zones, (sizeof(CTDB_TIDAL_ZONE_5) * 
				  ctdb_out->num_tidal_zones));

    /* Create the default water char entry */
    ctdb_out->num_water_chars = 1;
    ctdb_out->water_chars = (CTDB_WATER_CHARS_5 *)
      ctdb_alloc(NULL, sizeof(CTDB_WATER_CHARS_5),
		 "Fmt. 5 water chars.");
    bzero(ctdb_out->water_chars, sizeof(CTDB_WATER_CHARS_5));
    /* Make default water char tidal */
    ctdb_out->water_chars[0].tidal_zone_index = 1;

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));
    
    /* Unfortunately, feature sizes changed from fmt. 4 to fmt. 5.  We
     * could go through and calculate the new patch data size, but that
     * would involve iterating over all features, which we would then
     * have to do again when we copied the feature data.  Instead, we'll
     * guess high, improving performance at the cost of some memory.
     */
    {
	int32 groups = (ctdb_out->patches_wide * ctdb_out->patches_high) /
	  CTDB_PATCHES_PER_GROUP_5;

	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_5 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_5),
		     "Fmt. 5 patch groups");

	ctdb_out->n_patch_bytes = convert_features_4_5(ctdb, ctdb_out);
    }

    /* Elevations */
    if (!ctdb->is_gridded)
    {
	ctdb_out->elevations = NULL;
    }
    else
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 5 elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
    }
    
    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_5 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data * 
		     sizeof(CTDB_FEATURE_DATA_5),
		     "Fmt. 5 quadtree");
	
	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_5));

	/* $$$We added a soil map index to soil_defrag data in format 5
	 * in a spot that used to be reserved.  To be perfectly safe,
	 * we should go through and zero all those out, but the recompiler
	 * will hopefully have the sense to ignore those values anyway,
	 * so I'm not going to fix this until I become convinced that it's
	 * a problem...
	 */
    }
    
    /* The compiler rep does not read from the topology section, so
     * we won't bother populating it, especially since it would be
     * painful.
     */
    
    return ctdb_out;
}


/* Converts PGs and features (including TINs) from format 4 to format 5.
 * Returns total patch data bytes in format 5 result.
 */
static int32
convert_features_4_5(ctdb_4, ctdb_5)
    CTDB_4 *ctdb_4;
    CTDB_5 *ctdb_5;    
{
    int32 groups;
    int32 i;
    int32 total_bytes = 0;
    ADDRESS base;
    int32 num_lmodels = 0;	/* Total linear models *allocated* so far
				 * for format 5 db.
				 */
	
    groups = ctdb_4->patches_wide * ctdb_4->patches_high /
      CTDB_PATCHES_PER_GROUP_4;
	
    /* Assume that all features were buildings (buildings are the only
     * feature whose size increased), and that each had only 3 vertices.
     * In that case, feature bytes will increase by a factor of 9/7.
     * Additionally, assume that all patches were tinned.  The patch
     * group data header for tinned PG's increased in size by 60 bytes.
     */
    ctdb_5->n_patch_bytes = (9.0/7.0) * ctdb_4->n_patch_bytes +
      (groups * 60);
	
    ctdb_5->patch_data = (ADDRESS)ctdb_alloc(NULL,
					     ctdb_5->n_patch_bytes,
					     "Fmt. 5 patch data");

    base = ctdb_5->patch_data;

    for(i=0;i<groups;i++)
    {
	total_bytes += convert_group_4_5(ctdb_4, ctdb_5, i, &base,
					 &num_lmodels);
    }

#ifdef DEBUG
    if (total_bytes > ctdb_5->n_patch_bytes)
      printf("ERROR:  too few bytes allocated for format 5 features!\n");
#endif /* DEBUG */

    return total_bytes;
    
}

/* Converts PG header and all PG data from format 4 to format 5.  group is
 * the patch group number to convert, base is an in-out variable passing
 * in the address where the format 5 data should start, and passing out
 * where the format 5 data for the next PG should start.
 */
static int32
convert_group_4_5(ctdb_4, ctdb_5, group, base, num_lmodels)
    CTDB_4  *ctdb_4;
    CTDB_5  *ctdb_5;
    int32    group;
    ADDRESS *base;
    int32   *num_lmodels;
{
    int32 bytes;
    
    /* First, convert the header. */
    convert_pg_4_5(ctdb_4, ctdb_5, group, *base);

    /* Now convert the data and the data headers */
    if (ctdb_5->patch_groups[group].contains_tins)
    {
	bytes = convert_tin_pg_data_4_5(ctdb_4, ctdb_5, group, *base,
					num_lmodels);
    }
    else
    {
	bytes = convert_grid_pg_data_4_5(ctdb_4, ctdb_5, group, *base,
					 num_lmodels);
    }
    
    *base = (int8 *)*base + bytes;
    return bytes;
}

/* Convert just the patch group header from format 4 to format 5.  base is
 * where the new PG should point.
 */
static void
convert_pg_4_5(ctdb_4, ctdb_5, group, base)
    CTDB_4  *ctdb_4;
    CTDB_5  *ctdb_5;
    int32    group;
    ADDRESS  base;
{
    CTDB_PATCH_GROUP_4 *pg_4;
    CTDB_PATCH_GROUP_5 *pg_5;

    pg_4 = ctdb_4->patch_groups + group;
    pg_5 = ctdb_5->patch_groups + group;
    
    bzero(pg_5, sizeof(CTDB_PATCH_GROUP_5));

    pg_5->data_start.grid_data = (CTDB_GRID_PG_DATA_5 *)base;
    
    /* We'll leave the cache stuff blank, since the compiler won't use
     * it anyway.
     */

    pg_5->contains_tins = pg_4->contains_tins;

    pg_5->patch0_occupied = pg_4->patch0_occupied;
    pg_5->patch0_dyn_terr = pg_4->patch0_dyn_terr;
    pg_5->patch0_te16 = pg_4->patch0_te16;
    pg_5->patch0_no_tes = pg_4->patch0_no_tes;
    pg_5->patch0_water = FALSE;	/* No patch water in format 4 */

    pg_5->patch1_occupied = pg_4->patch1_occupied;
    pg_5->patch1_dyn_terr = pg_4->patch1_dyn_terr;
    pg_5->patch1_te16 = pg_4->patch1_te16;
    pg_5->patch1_no_tes = pg_4->patch1_no_tes;
    pg_5->patch1_water = FALSE;	/* No patch water in format 4 */

    pg_5->patch2_occupied = pg_4->patch2_occupied;
    pg_5->patch2_dyn_terr = pg_4->patch2_dyn_terr;
    pg_5->patch2_te16 = pg_4->patch2_te16;
    pg_5->patch2_no_tes = pg_4->patch2_no_tes;
    pg_5->patch2_water = FALSE;	/* No patch water in format 4 */

    pg_5->patch3_occupied = pg_4->patch3_occupied;
    pg_5->patch3_dyn_terr = pg_4->patch3_dyn_terr;
    pg_5->patch3_te16 = pg_4->patch3_te16;
    pg_5->patch3_no_tes = pg_4->patch3_no_tes;
    pg_5->patch3_water = FALSE;	/* No patch water in format 4 */

    return;
}


/* Convert all data, including data header, for a tinned PG from format 4 to
 * format 5.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_tin_pg_data_4_5(ctdb_4, ctdb_5, group, base, num_lmodels)
    CTDB_4 *ctdb_4;
    CTDB_5 *ctdb_5;
    int32   group;
    ADDRESS base;
    int32  *num_lmodels;
{
    int32 i;
    int32 size;
    CTDB_TIN_PG_DATA_4 *data_4;
    CTDB_TIN_PG_DATA_5 *data_5;
    CTDB_PATCH_GROUP_5 *pg;
    CTDB_FEATURE_DATA_4 *feat_4;
    CTDB_FEATURE_DATA_5 *feat_5;
    int32 patch_bytes = 0;
    int32 delta;
    int32 te_offset_4;
    int32 te_offset_5;

    data_4 = ctdb_4->patch_groups[group].data_start.tin_data;
    data_5 = (CTDB_TIN_PG_DATA_5 *)base;
    size = sizeof(CTDB_TIN_PG_DATA_5);
    
    pg = ctdb_5->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_5, size);
    bcopy(data_4->max_elev, data_5->max_elev, 
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(float32));
    
    /* Leave water_elev and tables zeroed out. */

    /* Can't fill in offsets yet... */

    /* The feature present field grew in format 5, but this copy is still
     * correct--we're implicitly padding the end with 0's.
     */
    bcopy(data_4->feature_present, data_5->feature_present,
	  CTDB_MAX_POSTS_PER_PATCH_4 * 2);
    
    bcopy(data_4->num_te, data_5->num_te, 
	  CTDB_PATCHES_PER_GROUP_4 * sizeof(uint16));
    
    /* Can't fill in te_offsets yet */

    bcopy(data_4->te_vgrid_dim, data_5->te_vgrid_dim,
	  CTDB_PATCHES_PER_GROUP_4 * sizeof(uint8));
    
    /* Can't fill in vgrid_offsets yet */

    /* 2.  Convert the features for each patch */
    feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((int8 *)data_4 +
					 sizeof(CTDB_TIN_PG_DATA_4));
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch1_offset = (patch_bytes + sizeof(CTDB_TIN_PG_DATA_5)) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }
    else
    {
	data_5->patch1_offset = sizeof(CTDB_TIN_PG_DATA_5) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }

    if (pg->patch1_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((CTDB_FEATURE_DATA_4 *)data_4 +
					 data_4->patch1_offset);
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch2_offset = data_5->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch2_offset = data_5->patch1_offset;
    }

    if (pg->patch2_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((CTDB_FEATURE_DATA_4 *)data_4 +
					 data_4->patch2_offset);
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch3_offset = data_5->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch3_offset = data_5->patch2_offset;
    }

    if (pg->patch3_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((CTDB_FEATURE_DATA_4 *)data_4 +
					 data_4->patch3_offset);
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }	  

    /* 3. Calculate how far the TE's for the patch have moved relative to
     * where they used to be
     */
    if (!pg->patch0_no_tes)
    {
	te_offset_4 = data_4->te_offset[0];
    }
    else if (!pg->patch1_no_tes)
    {
	te_offset_4 = data_4->te_offset[1];
    }
    else if (!pg->patch2_no_tes)
    {
	te_offset_4 = data_4->te_offset[2];
    }
    else
    {
	te_offset_4 = data_4->te_offset[3];
    }
    
    te_offset_5 = feat_5 - (CTDB_FEATURE_DATA_5 *)data_5;
    
    /* Now, calculate the offset, in words */
    delta = te_offset_5 - te_offset_4;
    
#ifdef DEBUG
    if (delta <= 0)
      printf("Unreasonable data offset in 4->5 conversion.\n");
#endif /* DEBUG */
    
    /* 4.  Fill out the rest of the offset values. */
    for(i=0;i<CTDB_PATCHES_PER_GROUP_5;i++)
    {
	data_5->te_offset[i] = data_4->te_offset[i] + delta;
	data_5->te_vgrid_offset[i] = data_4->te_vgrid_offset[i] + delta;
    }
    
    /* 5.  Copy over all the TE data. */
    {
	int32 data_size;
	int32 te_data_size;	/* In bytes */

	if (group < ctdb_5->max_patch_number)
	{
	    
	    /* Total bytes of data stored for this PG */
	    data_size = 
	      (int8 *)ctdb_4->patch_groups[group+1].data_start.grid_data -
	      (int8 *)data_4;
	    
	    te_data_size = data_size - 
	      (data_4->te_offset[0] * sizeof(CTDB_FEATURE_DATA_4));
	}
	else
	{
	    int32 vgrid_size;
	    
	    if (!pg->patch3_no_tes)
	    {
		data_size = data_4->te_vgrid_offset[3] * 
		  sizeof(CTDB_FEATURE_DATA_4);
		vgrid_size = data_4->te_vgrid_dim[3] * data_4->te_vgrid_dim[3];
	    }
	    else if (!pg->patch2_no_tes)
	    {
		data_size = data_4->te_vgrid_offset[2] * 
		  sizeof(CTDB_FEATURE_DATA_4);
		vgrid_size = data_4->te_vgrid_dim[2] * data_4->te_vgrid_dim[2];
	    }
	    else if (!pg->patch1_no_tes)
	    {
		data_size = data_4->te_vgrid_offset[1] * 
		  sizeof(CTDB_FEATURE_DATA_4);
		vgrid_size = data_4->te_vgrid_dim[1] * data_4->te_vgrid_dim[1];
	    }
	    else
	    {
		data_size = data_4->te_vgrid_offset[0] * 
		  sizeof(CTDB_FEATURE_DATA_4);
		vgrid_size = data_4->te_vgrid_dim[0] * data_4->te_vgrid_dim[0];
	    }

	    /* Add padding as necessary */
	    vgrid_size += (vgrid_size % sizeof(CTDB_FEATURE_DATA_4));
	    data_size += vgrid_size;		
	    te_data_size = (data_size -
			    (te_offset_4 * sizeof(CTDB_FEATURE_DATA_4)));
	}

	feat_4 = (CTDB_FEATURE_DATA_4 *)data_4 + te_offset_4;
	bcopy(feat_4, feat_5, te_data_size);
	
	/* Calculate total bytes in this PG */
	size = ((int8 *)feat_5 - (int8 *)data_5) + te_data_size;
    }

    return size;
}

/* Convert all data, including data header, for a tinned PG from format 4 to
 * format 5.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_grid_pg_data_4_5(ctdb_4, ctdb_5, group, base, num_lmodels)
    CTDB_4 *ctdb_4;
    CTDB_5 *ctdb_5;
    int32   group;
    ADDRESS base;
    int32  *num_lmodels;
{
    int32 size;
    CTDB_GRID_PG_DATA_4 *data_4;
    CTDB_GRID_PG_DATA_5 *data_5;
    CTDB_PATCH_GROUP_5 *pg;
    CTDB_FEATURE_DATA_4 *feat_4;
    CTDB_FEATURE_DATA_5 *feat_5;
    int32 patch_bytes = 0; 

    data_4 = ctdb_4->patch_groups[group].data_start.grid_data;
    data_5 = (CTDB_GRID_PG_DATA_5 *)base;
    size = sizeof(CTDB_GRID_PG_DATA_5);
    
    pg = ctdb_5->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_5, size);
    bcopy(data_4->max_elev, data_5->max_elev, 
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(float32));
    
    /* Leave water_elev and tables zeroed out. */

    /* Can't fill in offsets yet... */

    /* 2.  Convert the features for each patch */
    feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((int8 *)data_4 +
					 sizeof(CTDB_GRID_PG_DATA_4));
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch1_offset = (patch_bytes + sizeof(CTDB_GRID_PG_DATA_5)) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }
    else
    {
	data_5->patch1_offset = sizeof(CTDB_GRID_PG_DATA_5) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }

    if (pg->patch1_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((CTDB_FEATURE_DATA_4 *)data_4 +
					 data_4->patch1_offset);
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch2_offset = data_5->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch2_offset = data_5->patch1_offset;
    }

    if (pg->patch2_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((CTDB_FEATURE_DATA_4 *)data_4 +
					 data_4->patch2_offset);
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch3_offset = data_5->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

    }
    else
    {
	data_5->patch3_offset = data_5->patch2_offset;
    }

    if (pg->patch3_occupied)
    {
	feat_4 = (CTDB_FEATURE_DATA_4 *)((CTDB_FEATURE_DATA_4 *)data_4 +
					 data_4->patch3_offset);
	
	patch_bytes = convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels);
	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }	  

    size = (int8 *)feat_5 - (int8 *)data_5;
    return size;    
}

/* Converts features for a patch from format4 to format5.  feat4 must point
 * to the start of the features in the source patch, and feat5 must point
 * to where the features should start in the destination patch.  Returns
 * number of bytes written into dest. patch.
 */
static int32 
convert_feat_4_5(ctdb_5, feat_4, feat_5, num_lmodels)
    CTDB_5              *ctdb_5;
    CTDB_FEATURE_DATA_4 *feat_4;
    CTDB_FEATURE_DATA_5 *feat_5;
    int32               *num_lmodels;
{
    CTDB_FEATURE_DATA_4 *last, *next_4;
    CTDB_FEATURE_DATA_5 *cur_5, *next_5;
    CTDB_FEAT_OBJ_PTR obj;
    int32 ftype, verts;

    last = feat_4 + feat_4->hdr.size;
    cur_5 = feat_5;
    feat_4++;
    cur_5++;
    
    for(;feat_4<last;(feat_4 = next_4), (cur_5 = next_5))
    {
	ftype = feat_4->info.feature_type;

	switch(ftype)
	{
	  case FT_MICRO_4:
	    convert_micro_4_5(ctdb_5, feat_4, cur_5, &next_4, &next_5);
	    break;
	  case FT_BUILDING_4:
	    convert_building_4_5(ctdb_5, feat_4, cur_5, &next_4, &next_5);
	    break;
	  case FT_TREE_4:
	    convert_tree_4_5(ctdb_5, feat_4, cur_5, &next_4, &next_5, 
			     num_lmodels);
	    break;
	  case FT_CANOPY_4:
	    convert_canopy_4_5(ctdb_5, feat_4, cur_5, &next_4, &next_5);
	    break;
	  case FT_LINEAR_4:
	    convert_linear_4_5(ctdb_5, feat_4, cur_5, &next_4, &next_5);
	    break;
	  default:
	    printf("ERROR: unknown feature type encountered in format 4 db.\n");
	    exit(-1);
	}
    }

    feat_5->hdr.size = next_5 - feat_5;
    return (feat_5->hdr.size * sizeof(CTDB_FEATURE_DATA_5));
}

static void
convert_micro_4_5(ctdb_5, feat_4, feat_5, next_4, next_5)
    CTDB_5               *ctdb_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4 **next_4;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* No change from format 4 to format 5, just copy it over */

    *next_4 = feat_4 + 4;
    *next_5 = feat_5 + 4;

    bcopy(feat_4, feat_5, 4 * sizeof(CTDB_FEATURE_DATA_4));
    return;    
}

static void
convert_building_4_5(ctdb_5, feat_4, feat_5, next_4, next_5)
    CTDB_5               *ctdb_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4 **next_4;
    CTDB_FEATURE_DATA_5 **next_5;
{
    int32 verts = feat_4->info.vertices;

    *next_4 = feat_4 + 1 + 2*verts;
    *next_5 = feat_5 + 3 + 2*verts;
    
    feat_5->info.feature_class = feat_4->info.feature_type;
    feat_5->info.vertices = feat_4->info.vertices;
    feat_5->info.post_mask = feat_4->info.post_mask;
    
    feat_5++;
    
    /* Format 4 doesn't have models */
    feat_5->volume.reference = CTDB_INVALID_VMODEL_5;
    feat_5->volume.type = CTDB_FT_BUILDING_5;
    feat_5->volume.avoid = FALSE;
    feat_5->volume.reserved = 0;	/* Just to be safe... */

    feat_5++;

    feat_5->integer = CTDB_INVALID_VOL_ID;

    feat_5++;
    feat_4++;
    
    /* Copy the vertices, and we're done */
    bcopy(feat_4, feat_5, 2 * verts * sizeof(CTDB_FEATURE_DATA_4));
    return;
}

#define MODELS_PER_ALLOC 100	/* Number of linear models to alloc every time
				 * we run out
				 */

static void
convert_tree_4_5(ctdb_5, feat_4, feat_5, next_4, next_5, num_lmodels)
    CTDB_5               *ctdb_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4 **next_4;
    CTDB_FEATURE_DATA_5 **next_5;
    int32                *num_lmodels;
{
    /* "Trees" and "treelines" in format 4 turned into "linears" in format 5,
     * with some critical information moving into a linear model.
     */
    int32 verts = feat_4->info.vertices;
    CTDB_LINEAR_MODEL_5 model;
    int32 model_index;

    *next_4 = feat_4 + 2 + 2*verts;
    *next_5 = feat_5 + 2 + 2*verts;
    
    feat_5->info.feature_class = feat_4->info.feature_type;
    feat_5->info.vertices = feat_4->info.vertices;
    feat_5->info.post_mask = feat_4->info.post_mask;
    
    feat_4++;
    feat_5++;
    
    feat_5->linear.first = feat_4->tree.first;
    feat_5->linear.last = feat_4->tree.last;
    
    /* Fill out a model structure based on the input feature */
    model.type = CTDB_FT_TREELINE_5;
    model.pad = 0;		/* Just to be safe */
    model.width = 2.0 * feat_4->tree.radius;
    
    model.linear_data.treeline_info.trunk_radius =
      ctdb_5->inv_incr * POST_TO_XYO(CTDB_DEFAULT_TRUNK_RADIUS, 
		  0, ctdb_5->patch_inv_incr);
    
    model.linear_data.treeline_info.foliage_height = feat_4->tree.height;
    
    model.linear_data.treeline_info.pad1 = 0; /* Just to be safe */

    model.linear_data.treeline_info.fullness = 
      CTDB_DEFAULT_FULLNESS * (1 << CTDB_FULLNESS_BITS_5);

    model.linear_data.treeline_info.pad2 = 0; /* Just to be safe */

    /* See if there's already a matching model.  If so, reference it, if
     * not, create a new one.
     */
    model_index = find_matching_linear_model_5(ctdb_5, &model);

    if (model_index == CTDB_INVALID_LMODEL_5)
    {
	if (ctdb_5->num_linear_models == 0)
	{
	    /* The first CTDB_MAX_CANOPY_MODELS slots are reserved for 
	     * canopy models */
	    ctdb_5->num_linear_models = CTDB_MAX_CANOPY_MODELS;
	    *num_lmodels = ctdb_5->num_linear_models + MODELS_PER_ALLOC;
	    ctdb_5->linear_models = (CTDB_LINEAR_MODEL_5 *)
	      ctdb_alloc(NULL, *num_lmodels * sizeof(CTDB_LINEAR_MODEL_5),
			 "Fmt. 5 linear models");
	    /* Clear out the canopy models in case someone looks at them */
	    bzero(ctdb_5->linear_models, (CTDB_MAX_CANOPY_MODELS *
					  sizeof(CTDB_LINEAR_MODEL_5)));
	}
	else
	{
	    if (ctdb_5->num_linear_models >= *num_lmodels)
	    {
		*num_lmodels = *num_lmodels + MODELS_PER_ALLOC;

		ctdb_5->linear_models = (CTDB_LINEAR_MODEL_5 *)
		  ctdb_alloc(ctdb_5->linear_models, 
			     *num_lmodels * sizeof(CTDB_LINEAR_MODEL_5),
			     "Fmt. 5 linear models");
	    }
	}
	
	bcopy(&model, ctdb_5->linear_models + ctdb_5->num_linear_models,
	      sizeof(model));
	
	model_index = ctdb_5->num_linear_models++;
    }
    
    feat_5->linear.reference = model_index;

    feat_5++;
    feat_4++;
    
    /* Copy over the vertices, and we're done */
    bcopy(feat_4, feat_5, 2 * verts * sizeof(CTDB_FEATURE_DATA_4));

    return;
}

static void
convert_canopy_4_5(ctdb_5, feat_4, feat_5, next_4, next_5)
    CTDB_5               *ctdb_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4 **next_4;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* No change from format 4 to format 5, just copy it over */

    *next_4 = feat_4 + 4;
    *next_5 = feat_5 + 4;

    bcopy(feat_4, feat_5, 4 * sizeof(CTDB_FEATURE_DATA_4));
    return;
}

static void
convert_linear_4_5(ctdb_5, feat_4, feat_5, next_4, next_5)
    CTDB_5               *ctdb_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4 **next_4;
    CTDB_FEATURE_DATA_5 **next_5;
{
    int32 verts = feat_4->info.vertices;

    /* No change from format 4 to format 5, just copy it over */

    *next_4 = feat_4 + 2 + verts;
    *next_5 = feat_5 + 2 + verts;

    bcopy(feat_4, feat_5, (verts + 2) * sizeof(CTDB_FEATURE_DATA_4));
    return;
}


#define ACCEPTABLE_WIDTH_ERROR    33  /* XYO (~= 0.5 meters) */  
#define ACCEPTABLE_RADIUS_ERROR   16  /* XYO (~= 0.25 meters) */  
#define ACCEPTABLE_HEIGHT_ERROR   82  /* CTDB_TREE_HEIGHT_BASIS (~= 1 meter) */
#define ACCEPTABLE_FULLNESS_ERROR 10
#define ACCEPTABLE_SPACING_ERROR  0.1 /* meters */

/* Find an existing linear model that matches the passed model.  It doesn't
 * have to match exactly; it only has to match within acceptable error
 * tolerances.  Note that for our purposes here, it only needs to handle
 * treeline models.
 */
static int32
find_matching_linear_model_5(ctdb_5, model)
    CTDB_5              *ctdb_5;
    CTDB_LINEAR_MODEL_5 *model;
{
    int32                  i, start, end, delta;
    float32                deltaF;
    CTDB_LINEAR_MODEL_5   *m;

    start = CTDB_MAX_CANOPY_MODELS;
    end = ctdb_5->num_linear_models;

    for(i = start; i < end; i++)
    {
	m = ctdb_5->linear_models + i;
	
	/* Compare the types.
	 */
	if(m->type != model->type)
	  continue;

	/* Compare the widths.
	 */
	delta = m->width - model->width;
	if(delta < -ACCEPTABLE_WIDTH_ERROR ||
	   delta > ACCEPTABLE_WIDTH_ERROR)
	  continue;

	/* Compare tree trunk radii.
	 */
	delta = m->linear_data.treeline_info.trunk_radius -
	  model->linear_data.treeline_info.trunk_radius;
	if(delta < -ACCEPTABLE_RADIUS_ERROR ||
	   delta > ACCEPTABLE_RADIUS_ERROR)
	  continue;
	
	/* Compare tree foliage heights.
	 */
	delta = m->linear_data.treeline_info.foliage_height -
	  model->linear_data.treeline_info.foliage_height;
	if(delta < -ACCEPTABLE_HEIGHT_ERROR ||
	   delta > ACCEPTABLE_HEIGHT_ERROR)
	  continue;

	/* Compare tree fullness.
	 */
	delta = m->linear_data.treeline_info.fullness -
	  model->linear_data.treeline_info.fullness;
	if(delta < -ACCEPTABLE_FULLNESS_ERROR ||
	   delta > ACCEPTABLE_FULLNESS_ERROR)
	  continue;

	return i;
    }
    return CTDB_INVALID_LMODEL_5;
}

