/*
 * $RCSfile: ct_cmp_fmt4.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt4.h                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 4 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/
#ifndef _CMP_FMT4_H_INCLUDED
#define _CMP_FMT4_H_INCLUDED

#define CTDB_PATCHES_PER_GROUP_4       4  /* Num of patches per patch group */

#define CTDB_MAX_POSTS_PER_PATCH_4    16   /* maximum number of posts per
					    * patch
					    */

#define CTDB_N_SOILS_4                16   /* Number of soil types per 
					    * database 
					    */

#define CTDB_NAME_LENGTH_4            14   /* same as Simnet */
#define CTDB_DATE_LENGTH_4            28   /* Enough room for ctime() */

#define CTDB_PATCH_CACHES_4           16   /* Number of patch data caches */
#define CTDB_CACHE_BKT_UNITS_4         4

/* SW soil is bits 24-27 */
#define GET_POST_SW_SOIL_4(p) (((p) >> 24) & 0xF)
#define BUILD_POST_SW_SOIL_4(s) (((s) & 0xF) << 24)

/* SE soil is bits 24-27 */
#define GET_POST_SE_SOIL_4(p) (((p) >> 24) & 0xF)
#define BUILD_POST_SE_SOIL_4(s) (((s) & 0xF) << 24)

/* NE soil is bits 28-31 */
#define GET_POST_NE_SOIL_4(p) (((p) >> 28) & 0xF)
#define BUILD_POST_NE_SOIL_4(s) (((s) & 0xF) << 28)

/* NW soil is bits 28-31 */
#define GET_POST_NW_SOIL_4(p) (((p) >> 28) & 0xF)
#define BUILD_POST_NW_SOIL_4(s) (((s) & 0xF) << 28)


/* The feature types each imply certain intervisibility semantics
 * and storage techniques:
 *
 * FT_TREE is an individual tree or a treeline (depending upon the
 * number of vertices).
 *
 * An individual tree is described
 *	INFO RADHT XY Z
 * RADHT identifies the radius of the tree and the height of the
 * foliage part of the tree; XY identifies the location of the tree
 * within the patch; Z identifies the elevation of the top of the
 * tree.  The reduction of visibility due to tree blockage is done
 * using a simple light transmittance model.
 *
 * A treeline is described
 *	INFO RADHT XY Z XY Z XY Z...
 * RADHT identifies the radius of the trees at the ends of the
 * treeline and the height of the foliage part of the treeline; (note
 * that we do not need to store the end trees as individuals since
 * they are implied by the treeline); each XY point identifies the
 * location of one vertex in the treeline; the corresponding Z
 * identifies the elevation of the top of the treeline at that vertex.
 *
 * FT_BUILDING is a building or a building-like structure which blocks
 * intervisibility.  Telephone poles, water towers, and other such
 * structures which do not interrupt intervisibility are omitted.
 *
 * A building is described
 *	INFO XY Z XY Z ...
 * Each XY point identifies the location of one vertex in the bounding
 * box of the building; the corresponding Z identifies the elevation
 * of the roof line of the building at that point.  Intervisibility is
 * interrupted by the line segments connecting these points exactly the
 * same way as the segments connecting grid posts.
 *
 * FT_CANOPY is a canopy polygon.  See-through canopies are omitted.
 * FT_MICRO is a microterrain triangle.
 *
 * Canopies and microterrain are described the same way,
 *	INFO XY Z MICRO
 * The XY and Z combination identify the location of one vertex of
 * the microterrain triangle.  Up to three subsequent vertices are listed
 * in the MICRO structure.  There are edges connecting this point to
 * each of the other listed vertices.  If there are two vertices listed
 * the two edges from this point form a triangle.  If there are three
 * listed, the three edges form two triangles (the first edge being
 * shared between the two).  For microterrain, intervisibility is
 * interrupted by the edges connecting these points exactly the same
 * way as the segments connecting grid posts.  For canopies, visibility
 * is degraded as with treelines.
 *
 * FT_LINEAR is a road or river segment.
 *
 * Road segments and river segments are described the same way
 *	INFO WDSOIL XY XY...
 * WDSOIL identifies the width at the beginning and end of the segment
 * (linear taper is assumed), and the soil type of the segment; each
 * XY point identifies the location of one vertex in the segment.  Linear
 * features do not impact intervisibility.
 */

#define FT_MICRO_4    0
#define FT_BUILDING_4 1
#define FT_TREE_4     2
#define FT_CANOPY_4   3
#define FT_LINEAR_4   4

#define M_FEATURE_TYPE_4    0xE0000000
#define S_FEATURE_TYPE_4    29
#define M_VERTICES_4        0x1E000000
#define M_MICRO_VERTICES_4  0x06000000
#define S_VERTICES_4        25

#define MICRO_VERTS_4         0x3
#define MICRO_TYPE_4          0xC
#define MICRO_TYPE_BASE_4     0x0
#define MICRO_TYPE_ML_4       0x4
#define MICRO_TYPE_DEFAULT_4  0x8
#define MICRO_TYPE_RESERVED_4 0xC /* Reserved for future expansion */

typedef struct ctdb_feature_info_4
{
    /* One of the FT_x values listed above.
     */
    unsigned feature_type : 3;

    /* For FT_TREE, FT_BUILDING, FT_CANOPY, FT_ROAD, and FT_WATER this
     * is the number of vertices.  If the actual objects has more than
     * 16 vertices, it must be broken into two or more objects.  For
     * FT_MICRO this is the number of positions which are connected to
     * this point bitwise-OR'd with a microterrain type.
     */
    unsigned vertices : 4;

    /* This is used as a high resolution bitwise bounding volume test.
     * Each post in a patch has a bit.  This mask identifies which
     * posts' polygons were touched by this feature.  Although Simnet
     * terrain databases only have 16 posts per patch, CCTT database
     * may have 25 (given the 100m interval required in the spec).
     */
    unsigned post_mask : 25;
} CTDB_FEATURE_INFO_4;

typedef struct ctdb_xy_patch_offset_4
{
    /* These offsets identify the location of the feature with respect
     * to the SW corner of the patch in units of (patch_size/2^15)
     * meters.  For Simnet databases, this yields about 16mm resolution.
     * We use 2^15 rather than 2^16 because it allows coordinates to
     * be multiplied as integers without the risk of overflow.  Note that
     * the more traditional storage method (SP floating point from the
     * SW corner of the database) takes twice as much room and yields
     * only 43mm resolution at the eastern extreme of the SAKI (360km
     * wide) database.  Note that since these units are used for all
     * feature data, feature intensive operations (such as intervisibility)
     * can be performed exclusively in these units, requiring only that
     * the non-feature locations (such as intervis eye- and target-location)
     * be converted (instead of converting each feature to meters).
     */
    uint16 x;
    uint16 y;
} CTDB_XY_PATCH_OFFSET_4;

#define CTDB_TREE_HEIGHT_BASIS_4 (200.0 / (1<<14))

#define M_TREE_FIRST_4 0x8000
#define M_TREE_LAST_4  0x4000

typedef struct ctdb_treeline_info_4
{
    /* These parameters identify characteristics of trees and treelines.
     * first & last are 1 only if a tree should be put at the
     * start and end, respectively, of the treeline.  Radius is measured in
     * the same units as x,y in the XY_PATCH_OFFSET struct:
     * (patch_size/2^15) meters.  Height is measured in the unit
     * CTDB_TREE_HEIGHT_BASIS (defined above).  Height is the height
     * of the foliage, not the height of the tree.
     */
    uint16 radius;
    unsigned       first : 1;
    unsigned       last : 1;
    unsigned       height : 14;
} CTDB_TREELINE_INFO_4;

/* For now we store feature elevations as simple floats, since there
 * is no obvious way to use the extra 13 bits if we were to instead
 * store it using the fixed point convention used for grid posts.
 * Floating point will, of course, be faster than storing them in
 * fixed point since only a SP->DP conversion will be necessary to use
 * the data.
 */
typedef float32 CTDB_Z_4;

#define M_VERT0_OUTER_4 0x01000000
#define M_VERT1_OUTER_4 0x02000000
#define M_VERT2_OUTER_4 0x04000000

typedef struct ctdb_micro_pts_4
{
    /* Each microterrain vertex is connected to 0,1,2, or 3 other
     * microterrain vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then the soil
     * fields will be 0; vert0 will contain the index of that vertex
     * relative to patch start; vert1, vert2 will be 0.  If it
     * connected to 2 others, soil0 will contain the soil type of the
     * triangle connecting this vertex to vert0 to vert1; soil1 and
     * vert2 will be 0.  If it is connected to 3 others, soil0 will
     * contain the soil type of the triangle connecting this vertex to
     * vert0 to vert1; soil1 will contain the soil type of the
     * triangle connecting this vertex to vert0 to vert2.
     * The same rules apply when this structure is used to describe
     * canopy polygons, except the soil fields are used by bits in a
     * mask identifying whether the edge to each vertex is an outer
     * edge of the canopy.
     */
    unsigned soil0 : 4;
    unsigned soil1 : 4;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
} CTDB_MICRO_PTS_4;

typedef struct ctdb_linear_info_4
{
    /* Each linear segment has a width and a soil type.  The width is
     * in the same units as x,y in the XY_PATCH_OFFSET structure.
     * The edge_key provides the low-order 12 bits of the topologic
     * edge (of type CTDB_EDGE) which uses this linear feature.
     */
    uint16    width;
    unsigned  soil : 4;
#define CTDB_EDGE_KEY_4 4096 /* 2^12 */
    unsigned  edge_key : 12;
} CTDB_LINEAR_INFO_4;

typedef struct ctdb_patch_hdr_4
{
    /* At the beginning of each patch, we store the size (the number
     * of CTDB_FEATURE_DATA things in the patch).  There will never be
     * more than 2^16 of those, so that leaves 16 bits of padding.
     * I'm sure we'll find a use for those bits eventually.
     */
    uint16 size;
    uint16 _pad;
} CTDB_PATCH_HDR_4;

#define CTDB_ABSTRACT_CANOPY_4                1
#define CTDB_ABSTRACT_SOIL_DEFRAG_4           2
#define CTDB_ABSTRACT_STEEP_SLOPE_4           3
#define CTDB_ABSTRACT_RAILROAD_4              4
#define CTDB_ABSTRACT_PIPELINE_4              5
#define CTDB_ABSTRACT_POLITICAL_BOUNDARY_4    6
#define CTDB_ABSTRACT_POWERLINE_4             7
#define CTDB_ABSTRACT_LABEL_4                 8
#define CTDB_ABSTRACT_TACTICAL_SIGN_4         9
#define CTDB_ABSTRACT_OFFROAD_SEGMENT_4       10
#define CTDB_ABSTRACT_MAX_4 /* Always Last */ 11


/* CTDB_ABSTRACT_FEATURE_INFO:
 *
 * Structure at the start of each abstract feature.
 */
#define CTDB_ABSTRACT_QUAD_NODE_4             0
typedef struct ctdb_abstract_feature_info_4
{
    /* Type code for the abstract feature.  The code 0 is reserved
     * to indicate that this "feature" is actually a reference to
     * four child quad nodes.
     */
    uint8 type_code;

    /* The size (in 4-byte units) of the CTDB_ABSTRACT_DATA associated
     * with the abstract feature.
     */
    uint8 data_size;

    /* The number of vertices in feature.  There must be at least one,
     * except for "features" which are of the type CTDB_ABSTRACT_QUAD_NODE.
     */
    uint16 num_vertices;

} CTDB_ABSTRACT_FEATURE_INFO_4;

/* CTDB_ABSTRACT_VERTEX:
 *
 * Structure used to represent a vertex of an abstract feature.
 * Note that for efficient access, the format used for vertices
 * is 32 bit floating point, not fix-point patch offsets.  This
 * results in a loss of precision at the far corners of the database,
 * but avoids a great deal of translation at run time.
 */
typedef float32 CTDB_ABSTRACT_VERTEX_4;

/* CTDB_ABSTRACT_DATA:
 *
 * Format of data associated with so-called abstract features.  Note
 * that this structure is analogous to a protocol definition, in that
 * the final record of each union substructure can be variable length.
 *
 * This goes into a 4-byte union: NO MEMBER STRUCTURE CAN EXCEED 32 BITS!
 *
 */
typedef union ctdb_abstract_data_4
{
    struct ctdb_quad_data_4
    {
	/* Array of 4 offsets to child quad nodes (in 4-byte units)
	 * from the beginning of the quad tree data storage.
	 */
	uint32 child_start[1];  /* really 4 members, not 1 */
    } quad_data;
    struct ctdb_canopy_data_4
    {
	unsigned impenetrable  : 1;
	unsigned padding       : 31;
    } canopy;
    struct ctdb_soil_defrag_data_4
    {
	unsigned soil_index    : 4;
	unsigned level         : 4;
	unsigned padding       : 24;
    } soil_defrag;
    struct ctdb_steep_slope_data_4
    {
	unsigned slope_degrees : 7;
	unsigned padding       : 25;
    } steep_slope;
    struct ctdb_label_data_4
    {
	char text[4]; /* variable length, NULL terminated */
    } label;
    struct ctdb_offroad_data_4
    {
	uint32 edge_number;
    } offroad;
} CTDB_ABSTRACT_DATA_4;

typedef union ctdb_feature_data_4
{
    uint32                       integer;
    CTDB_PATCH_HDR_4             hdr;
    CTDB_FEATURE_INFO_4          info;
    CTDB_TREELINE_INFO_4         tree;
    CTDB_LINEAR_INFO_4           linear;
    CTDB_XY_PATCH_OFFSET_4       xy;
    CTDB_Z_4                     z;
    CTDB_MICRO_PTS_4             micro;
    CTDB_ABSTRACT_FEATURE_INFO_4 abstract;
    CTDB_ABSTRACT_DATA_4         abstract_data;
    CTDB_ABSTRACT_VERTEX_4       abstract_vertex;
} CTDB_FEATURE_DATA_4;

/* Definitions of the fields in a 32-bit soil type.
 */
#define CTDB_SIMNET_SOIL_4(x)  ((x) & 0x0000000f)
#define CTDB_CCTT_SOIL_4(x)    (((x) >> 4) & 0x0000001f)
#define CTDB_MOBILITY_FLAG_SHIFT_4 9
#define CTDB_WATER_FLAG_4   (0x1 << CTDB_MOBILITY_FLAG_SHIFT_4)
#define CTDB_NO_GO_FLAG_4   (0x2 << CTDB_MOBILITY_FLAG_SHIFT_4)
#define CTDB_SLOW_GO_FLAG_4 (0x4 << CTDB_MOBILITY_FLAG_SHIFT_4)
#define CTDB_ROAD_FLAG_4    (0x8 << CTDB_MOBILITY_FLAG_SHIFT_4)
#define CTDB_WATER_4(x)   ((x) & CTDB_WATER_FLAG_4)
#define CTDB_NO_GO_4(x)   ((x) & CTDB_NO_GO_FLAG_4)
#define CTDB_SLOW_GO_4(x) ((x) & CTDB_SLOW_GO_FLAG_4)
#define CTDB_ROAD_4(x)    ((x) & CTDB_ROAD_FLAG_4)

/* CTDB_NODE:
 *
 * Represents the topologic connection between CTDB_EDGE structures.
 * The `connection' member is actually variable length.
 */
typedef struct ctdb_node_4
{
    /* Approximate location of the node (useful for planning) */
    float32 x, y;

    /* Whether the node corresponds to a bridge */
    unsigned is_bridge : 1;

    unsigned padding   : 23;

    uint8 num_connections;
    struct ctdb_connection_4
    {
	/* The node which this connects to */
	int32 node;

	/* The edge used to connect these nodes */
	int32 edge;

    } connection[1];

} CTDB_NODE_4;

/* CTDB_EDGE:
 *
 * Structure used to collect a list of linear features which form an
 * uninterrupted "edge" between two intersections.  The `feature'
 * member is actually variable length.
 */
typedef struct ctdb_edge_4
{
    /* Total distance along the segment (useful for planning) */
    int32 distance;

    /* Indices of the nodes at the beginning and end of the segment */
    int32 first_node;
    int32 last_node;

    int32 num_features;
    struct ctdb_feature_reference_4
    {
	/* The patch where this feature is stored */
	uint32 patch_number;

	/* Byte offset into the patch */
	unsigned patch_offset : 24;

	/* Whether the vertices in the patch are in reverse order */
	unsigned reverse      : 1;

	unsigned padding      : 7;
    } feature[1];
} CTDB_EDGE_4;

/**********************************************************************
 * Patch data headers                                                 *
 **********************************************************************/
/* The feature and TE data for each patch group is stored with a data header
 * at its beginning.  Because patch groups containing TINned patches need
 * substantially more data, there are two formats used, one for purely
 * gridded PGs, and one for PGs containing at least one TINned patch.
 */

/* Data header for a completely gridded patch group.  This will *always*
 * be immediately followed by the feature data for the patch group.  The
 * starting address for the features for the first patch is
 * (address of this structure) + (sizeof(this structure)).
 * The starting address for the features for the other patches in the
 * PG can be found by
 * (address of this structure) + (patch*_offset, in 32-bit words)
 */
typedef struct ctdb_grid_pg_data_4
{
    float32       max_elev[CTDB_PATCHES_PER_GROUP_4];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;
} CTDB_GRID_PG_DATA_4;

/* Data header for patch group that contains at least one TINned patch.
 * This is *always* followd by the following (in order):
 *    Feature data for each patch in PG
 *    TE data for each patch in PG
 *    Virtual grid data for each patch in PG
 * The various addresses can all be calculated as described above for the
 * CTDB_GRID_PG_DATA structure (note that the offsets for the everything
 * but features are explicitly stored for all 4 patches rather than just
 * the last 3).
 */
typedef struct ctdb_tin_pg_data_4
{
    float32       max_elev[CTDB_PATCHES_PER_GROUP_4];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;

    /* BITS             
     *	    3 : Features
     *	    2 : Buildings
     *	    1 : TREES
     *	    0 : Microterrain  
     */
    uint8   feature_present[CTDB_MAX_POSTS_PER_PATCH_4 * 2];/* 4 bits*16 posts 
							     * * 4 patches
							     * = 32 bytes
							     */
    /* Array[patches] numb TE in patch */
    uint16  num_te[CTDB_PATCHES_PER_GROUP_4];
    /* Where the TEs for each patch start.  These are in units of 32-bit
     * words and relative the starting address of this (ctdb_tin_pg_data)
     * data structure
     */
    uint32       te_offset[CTDB_PATCHES_PER_GROUP_4];
    /* SQRT(# of v. grid squares) for each patch */
    uint8        te_vgrid_dim[CTDB_PATCHES_PER_GROUP_4];
    /* Where the virtual grid starts for each patch.  These are in units
     * of 32-bit words and relative to the starting address of this
     * (ctdb_tin_pg_data) data structure
     */
    uint32       te_vgrid_offset[CTDB_PATCHES_PER_GROUP_4];
} CTDB_TIN_PG_DATA_4;


typedef union ctdb_pg_data_hdr_ptr_4
{
    struct ctdb_grid_pg_data_4 *grid_data;
    struct ctdb_tin_pg_data_4  *tin_data;
    int32                       offset;
} CTDB_PG_DATA_HDR_PTR_4;


/* A PATCH_GROUP is a collection of information about four contiguous
 * patches (contiguous in the sense x+y*w).  For each patch there are
 * four flags: occupied indicates there are features in the patch,
 * dyn_terr indicates that there are dynamic terrain changes to the
 * patch (future use), te16 indicates whether the patch's TEs use 8 or
 * 16 bit indices, and no_tes indicates whether or not a patch contains
 * TEs at all.  In addition, there is data about the patch group as a
 * whole:
 *     data_start points to the actual data for the cache.  Specifically,
 * it references either a CTDB_GRID_PG_DATA structure (contains_tins ==
 * FALSE) or a CTDB_TIN_PG_DATA structure (contains_tins == TRUE).  In either
 * case, that structure contains offsets for the feature/TE data itself,
 * which immediately follows the structure.
 *     cache specifies which of the 16 patch data caches this PG's data
 * should be placed in
 *     mapped is TRUE iff data_start points to actual data.  If it is FALSE,
 * the data must be read in from disk before being accessed.
 *     contains_tins is TRUE iff at least one patch in the PG is TINned
 */    
typedef struct ctdb_patch_group_4
{
    /* Bytes 0-3 */
    CTDB_PG_DATA_HDR_PTR_4     data_start;
    
    /* Bytes 4-7 */
    unsigned cache           : 4;  /* Which size cache bucket this PG uses */
    unsigned mapped          : 1;  /* 1 => PG data is in memory */
    unsigned contains_tins   : 1;  /* 1 => at least one patch in this PG
				    * is TINNed.  This determines what 
				    * data_start points to */
    unsigned patch0_occupied : 1;
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch0_no_tes   : 1;  /* 1 => patch 0 has 0 TEs */
    unsigned patch1_occupied : 1;
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch1_no_tes   : 1;  /* 1 => patch 1 has 0 TEs */
    unsigned patch2_occupied : 1;
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch2_no_tes   : 1;
    unsigned patch3_occupied : 1;
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch3_no_tes   : 1;
    unsigned unused          : 10;
} CTDB_PATCH_GROUP_4;

/**********************************************************************
 * Patch data cache                                                   *
 **********************************************************************/
typedef struct ctdb_cache_bucket_4
{
    CTDB_PATCH_GROUP_4 *user;	/* Patch group whose data is stored here */
    int32               patch_data_start_disk; /* The offset (in bytes)
						* from the start of the
						* patch data section in the
						* database file on disk at
						* which this PGs data starts
						*/
    union ctdb_feature_data_4 data[1];	/* Actual data.  This will be malloced
					 * to be longer than 1 word, but 
					 * declaring it this way saves us a 
					 * 4 bytes per bucket.
					 */
} CTDB_CACHE_BUCKET_4;


typedef struct ctdb_4
{
    /* Fixed point basis for elevation points.  Can be changed to deal
     * with future databases which have extreme elevations, or extreme
     * accuracy requirements.
     */
    float64         fixed_point_basis;
    float64         inv_fixed_point_basis;
    
    /* Soil type map.  The 4 bit number stored on posts and in microterrain
     * is actually an index into this map.
     */
    int32           soil_map[CTDB_N_SOILS_4];
    
    /* Spacing between grid posts */
    int32           incr;
    float64         inv_incr;
    float64         incrD;
    
    /* Spacing between patches */
    int32           patch_incr    ;/* number of posts per patch */
    float64         patch_inv_incr;/* 1 over patch_incr */
    float64         patch_incrD   ;/* double patch_incr */
    
    /* Database size */
    int32           pages_wide;
    int32           pages_high;
    int32           patches_wide;
    int32           patches_high;
    int32           min_x;
    int32           max_x;
    int32           min_y;
    int32           max_y;
    int32           min_x_post;
    int32           max_x_post;
    int32           min_y_post;
    int32           max_y_post;
    
    /* Useful trivia */
    int32           input_format;
    float64         min_z;
    float64         max_z;
    float64         origin_northing;
    float64         origin_easting;
    int32           origin_zone_number;
    int32           origin_zone_letter;
    int32           datum;
    char            name[CTDB_NAME_LENGTH_4];
    int16           version;
    char            date[CTDB_DATE_LENGTH_4];
    int32           total_memory;
    uint32          is_curved;
    uint32          is_sw_ne;
    CTDB_PATCH_GROUP_4 *patch_groups;
    int32           elevations_resident;
    int32           elevation_seek;
    int32           elevation_cache_pages;
    int32           direct_map_modulus;
    int32           elevation_hit_count;
    int32           elevation_miss_count;
    int32          *cache_page_numbers;
    int32         **elevation_pages;

    /* Caches sizes are in units of CTDB_CACHE_BKT_UNITS */
    int32           patch_cache_size_map[CTDB_PATCH_CACHES_4];

    int32           patch_data_seek;
    int32           patch_cache_entries;
    int32           patch_hit_count;
    int32           patch_miss_count;
    struct ctdb_cache_bucket_4 *patch_data_cache[CTDB_PATCH_CACHES_4];

    /* Database file */
    FILE           *file;

    int32          *elevations;

    /* Quadtree of abstract features */
    int32           num_quad_data;
    union ctdb_feature_data_4 *quad_data;

    /* Topologic relationships */
    int32           num_nodes;
    CTDB_NODE_4   **nodes;
    int32           node_size;
    int8           *node_storage;
    int32           num_edges;
    CTDB_EDGE_4   **edges;
    int32           edge_size;
    int8           *edge_storage;

    int32           n_patch_bytes;     /* Total amt of patch data for this 
					* cell, in bytes  (patch data = TEs +
					* features + virtual grids)
					*/
    ADDRESS         patch_data;	       /* Pointer to chunk o' memory with all
					* the patch data
					*/
    float64         meters_to_patch;   /* = patch_inv_incr * inv_incr  */
    uint32          is_gridded;
    uint32          is_tinned;
    int32           max_patch_number;
    int32           min_patch_number;
} CTDB_4;

#define CTDB_HEADER_RESERVED_4             10
typedef struct ctdb_file_header_4
{
    int32    format;
    unsigned is_curved : 1;
    unsigned is_sw_ne  : 1;
    unsigned is_gridded: 1;
    unsigned is_tinned : 1;
    unsigned pad1      : 28;
    float64  fixed_point_basis;
    int32    soil_map[CTDB_N_SOILS_4];
    int32    post_increment;
    int32    posts_per_patch_side;
    int32    pages_wide;
    int32    pages_high;
    int32    min_x;
    int32    max_x;
    int32    min_y;
    int32    max_y;
    int32    num_features;
    int32    patch_cache_size_map[CTDB_PATCH_CACHES_4];
    int32    pad2;
    float64  origin_northing;
    float64  origin_easting;
    int32    origin_zone_number;
    int32    origin_zone_letter;
    int32    datum;
    char     name[CTDB_NAME_LENGTH_4];
    int16    version;
    char     date[CTDB_DATE_LENGTH_4];
    int32    num_nodes;
    int32    node_size;
    int32    num_edges;
    int32    edge_size;
    int32    num_quad_data;
    int32    n_patch_bytes;
    int32    reserved[CTDB_HEADER_RESERVED_4]; /* For future use, if we 
						* need it */
} CTDB_FILE_HEADER_4;

extern void *
ctdb_convert_4_to_3(void *fmt4_db);

extern void *
ctdb_convert_4_to_5(void *fmt4_db);

extern void
ctdb_free_fmt4_db(void *fmt4_db);

extern void *
ctdb_read_4(char *fname);

extern void
ctdb_write_4(
	     void *fmt4_db,
	     FILE *file
	     );

#endif /* _CMP_FMT4_H_INCLUDED */




