/*
 * $RCSfile: ct_cmp_fmt5.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt5.c                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 5 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/

#include "ct_cmplr.h"
#include <libgcs.h>


/* Forward declarations */

static int32
convert_features_5_6(
		     CTDB_5 *ctdb_5,
		     CTDB_6 *ctdb_6
		     );

static int32
convert_group_5_6(
		  CTDB_5  *ctdb_5,
		  CTDB_6  *ctdb_6,
		  int32    group,
		  ADDRESS *base
		  );

static void
convert_pg_5_6(
	       CTDB_5  *ctdb_5,
	       CTDB_6  *ctdb_6,
	       int32    group,
	       ADDRESS  base
	       );

static int32
convert_tin_pg_data_5_6(
			CTDB_5 *ctdb_5,
			CTDB_6 *ctdb_6,
			int32   group,
			ADDRESS base
			);

static int32
convert_grid_pg_data_5_6(
			 CTDB_5 *ctdb_5,
			 CTDB_6 *ctdb_6,
			 int32   group,
			 ADDRESS base
			 );

static int32 
convert_feat_5_6(
		 CTDB_6              *ctdb_6,
		 CTDB_FEATURE_DATA_5 *feat_5,
		 CTDB_FEATURE_DATA_6 *feat_6
		 );

static void
convert_micro_5_6(
		  CTDB_6               *ctdb_6,
		  CTDB_FEATURE_DATA_5  *feat_5,
		  CTDB_FEATURE_DATA_6  *feat_6,
		  CTDB_FEATURE_DATA_5 **next_5,
		  CTDB_FEATURE_DATA_6 **next_6
		  );

static void
convert_mwater_5_6(
		   CTDB_6               *ctdb_6,
		   CTDB_FEATURE_DATA_5  *feat_5,
		   CTDB_FEATURE_DATA_6  *feat_6,
		   CTDB_FEATURE_DATA_5 **next_5,
		   CTDB_FEATURE_DATA_6 **next_6
		   );

static void
convert_volume_5_6(
		   CTDB_6               *ctdb_6,
		   CTDB_FEATURE_DATA_5  *feat_5,
		   CTDB_FEATURE_DATA_6  *feat_6,
		   CTDB_FEATURE_DATA_5 **next_5,
		   CTDB_FEATURE_DATA_6 **next_6
		   );

static void
convert_linear_5_6(
		   CTDB_6               *ctdb_6,
		   CTDB_FEATURE_DATA_5  *feat_5,
		   CTDB_FEATURE_DATA_6  *feat_6,
		   CTDB_FEATURE_DATA_5 **next_5,
		   CTDB_FEATURE_DATA_6 **next_6
		   );

static void
convert_canopy_5_6(
		   CTDB_6               *ctdb_6,
		   CTDB_FEATURE_DATA_5  *feat_5,
		   CTDB_FEATURE_DATA_6  *feat_6,
		   CTDB_FEATURE_DATA_5 **next_5,
		   CTDB_FEATURE_DATA_6 **next_6
		   );

static void
convert_laid_linear_5_6(
			CTDB_6               *ctdb_6,
			CTDB_FEATURE_DATA_5  *feat_5,
			CTDB_FEATURE_DATA_6  *feat_6,
			CTDB_FEATURE_DATA_5 **next_5,
			CTDB_FEATURE_DATA_6 **next_6
			);

static void
convert_elevation_posts_5_6(
			    CTDB_5 *ctdb_5,
			    CTDB_6 *ctdb_6
			    );

static void
generate_caching_scheme_5(CTDB_5 *ctdb);


static int32
find_pg_size_5(CTDB_PATCH_GROUP_5 *pg);

static void
write_patch_group_headers_5(
			    CTDB_5             *ctdb,
			    CTDB_PATCH_GROUP_5 *patch_groups,
			    int32               groups,
			    FILE               *output_file
			    );

static void 
write_nodes_and_edges_5(
			CTDB_5  *ctdb,
			FILE    *output_file
			);

static void 
write_abstract_features_5(
			  CTDB_5  *ctdb,
			  FILE    *output_file
			  );

static void 
write_linear_models_5(
		      CTDB_5  *ctdb,
		      FILE    *output_file
		      );

static void
write_volume_models_5(
		      CTDB_5  *ctdb,
		      FILE    *output_file
		      );

static void
write_soil_tables_5(
		    CTDB_5  *ctdb,
		    FILE    *output_file
		    );

static void
write_water_chars_5(
		    CTDB_5  *ctdb,
		    FILE    *output_file
		    );

static void
write_tidal_zones_5(
		    CTDB_5  *ctdb,
		    FILE    *output_file
		    );

static void 
write_elevation_posts_5(
			CTDB_5  *ctdb,
			FILE    *output_file
			);

static void
write_patch_data_5(
		   CTDB_5             *ctdb,
		   FILE               *output_file
		   );

static void
regen_table_ptrs(
		 int8   *old_base,
		 int8  **old_ptrs,
		 int8   *new_base,
		 int8  **new_ptrs,
		 int32   num_entries
		 );

static void 
ctdb_pack_header_5(
		   CTDB_5             *ctdb,
		   CTDB_FILE_HEADER_5 *hdr
		   );

static void
ctdb_pack_aux_header_5(
		       CTDB_5                 *ctdb,
		       CTDB_AUX_FILE_HEADER_5 *aux_hdr
		       );

static int32
convert_features_5_4(
		     CTDB_5 *ctdb_5,
		     CTDB_4 *ctdb_4,    
		     int32  *patch_deltas
		     );

static void
convert_post_soils_5_4(
		       CTDB_5 *ctdb_5,
		       CTDB_4 *ctdb_4
		       );

static void
fix_topo_refs_5_4(
		  CTDB_5 *ctdb_5,
		  CTDB_4 *ctdb_4,
		  int32  *patch_deltas
		  );


/***************************************************************************
 * READ ROUTINES                                                           *
 ***************************************************************************/

/* $$$ I'm not going to find z-limits here; I'll wait until I get everything
 * into the compiler format so I only have to write the code once.
 */

static void 
ctdb_unpack_header_5(CTDB_5 *ctdb, CTDB_FILE_HEADER_5 *hdr)
{
    ctdb->input_format          = hdr->format;
    ctdb->is_curved             = hdr->is_curved;
    ctdb->is_sw_ne              = hdr->is_sw_ne;
    ctdb->is_gridded            = hdr->is_gridded;
    ctdb->is_tinned             = hdr->is_tinned;
    ctdb->fixed_point_basis     = hdr->fixed_point_basis;

    if (ctdb->fixed_point_basis) 
      ctdb->inv_fixed_point_basis = 1.0/ctdb->fixed_point_basis;

    ctdb->num_soil_tables = hdr->num_soil_tables;
    ctdb->soil_table_size = hdr->soil_table_size;
    ctdb->num_water_chars = hdr->num_water_chars;
    ctdb->num_tidal_zones = hdr->num_tidal_zones;

    ctdb->incr           = hdr->post_increment;
    ctdb->incrD          = ctdb->incr;
    ctdb->patch_incr     = hdr->posts_per_patch_side;
    ctdb->patch_incrD    = ctdb->patch_incr;

    if (ctdb->incrD)
      ctdb->inv_incr = 1.0/ctdb->incrD;

    if (ctdb->patch_incrD)
      ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide     = hdr->pages_wide;
    ctdb->pages_high     = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE * ctdb->patch_inv_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE * ctdb->patch_inv_incr;
    
    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);
    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);

    bcopy(hdr->patch_cache_size_map, ctdb->patch_cache_size_map,
	  CTDB_PATCH_CACHES * sizeof(int32));

    ctdb->origin_northing    = hdr->origin_northing;
    ctdb->origin_easting     = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum              = hdr->datum;

    bcopy(hdr->name,ctdb->name, sizeof(hdr->name));

    ctdb->version            = hdr->version;

    bcopy(hdr->date,ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date,       "\n");

    ctdb->num_nodes        = hdr->num_nodes;
    ctdb->node_size        = hdr->node_size;
    ctdb->num_edges        = hdr->num_edges;
    ctdb->edge_size        = hdr->edge_size;
    ctdb->num_quad_data    = hdr->num_quad_data;
    ctdb->n_patch_bytes    = hdr->n_patch_bytes;
    ctdb->num_linear_models = hdr->num_linear_models;
    ctdb->num_volume_models = hdr->num_volume_models;
    ctdb->meters_to_patch  = ctdb->patch_inv_incr * ctdb->inv_incr;
    ctdb->max_patch_number = ctdb->patches_wide * ctdb->patches_high - 1;
    ctdb->min_patch_number = 0;
    return;
}

static void
ctdb_unpack_aux_header_5(CTDB_5 *ctdb, CTDB_AUX_FILE_HEADER_5 *aux_hdr)
{
    ctdb->gcs_mode  = aux_hdr->gcs_mode;
    ctdb->cell_id   = aux_hdr->cell_id;
    ctdb->origin_x  = aux_hdr->origin_x;
    ctdb->origin_y  = aux_hdr->origin_y;
    ctdb->west_long = aux_hdr->west_long;
    ctdb->east_long = aux_hdr->east_long;
    ctdb->south_lat = aux_hdr->south_lat;
    ctdb->north_lat = aux_hdr->north_lat;
    return;
}

static CTDB_STATUS
read_patch_group_headers_5(CTDB_5 *ctdb, FILE *fileptr)
{
  int32 groups;

  /* Read in all the patch group headers */
  groups=(ctdb->patches_wide*ctdb->patches_high)/CTDB_PATCHES_PER_GROUP_5;
  ctdb->patch_groups=(CTDB_PATCH_GROUP_5 *)
    ctdb_alloc(NULL,
	       sizeof(CTDB_PATCH_GROUP_5)*groups,
	       "patch group data");
  fread(ctdb->patch_groups,sizeof(CTDB_PATCH_GROUP_5),groups,fileptr);

  return CTDB_STATUS_OK;
}
 
static void 
init_null_topology_5(CTDB_5 *ctdb)
{
    ctdb->node_storage = (int8 *) 
      ctdb_alloc(NULL, sizeof(CTDB_NODE_5), "initialization");
    bzero(ctdb->node_storage, sizeof(CTDB_NODE_5));
    ctdb->edge_storage = (int8 *) 
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_5), "initialization");
    bzero(ctdb->edge_storage, sizeof(CTDB_EDGE_5));
    ctdb->nodes = (CTDB_NODE_5 **) 
      ctdb_alloc(NULL, sizeof(CTDB_NODE_5 *), "initialization");
    ctdb->nodes[0] = (CTDB_NODE_5 *)ctdb->node_storage;
    ctdb->edges = (CTDB_EDGE_5 **) 
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_5 *), "initialization");
    ctdb->edges[0] = (CTDB_EDGE_5 *)ctdb->edge_storage;
}

static CTDB_STATUS 
ctdb_read_topology_5(CTDB_5 *ctdb, FILE *fileptr)
{
    int32 i, offset;
    
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces: ctdb_read_topology((CTDB *)ctdb,fileptr);*/
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    if (ctdb->num_nodes)
    {
        ctdb->nodes = NULL;
	ctdb->nodes =
	  (CTDB_NODE_5 **)ctdb_alloc(NULL,
				     sizeof(CTDB_NODE_5 *) *
				     ctdb->num_nodes,
				     "node index table");
	ctdb->node_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->node_size, "nodes");

	if (fread(ctdb->node_storage, ctdb->node_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Nodes\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_NODES_FAILED;
	}
    }

    if (ctdb->num_edges)
    {
        ctdb->edges = NULL;
	ctdb->edges =
	  (CTDB_EDGE_5 **)ctdb_alloc(NULL,
				     sizeof(CTDB_EDGE_5 *) *
				     ctdb->num_edges,
				     "edge index table");
	ctdb->edge_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->edge_size, "edges");

	if (fread(ctdb->edge_storage, ctdb->edge_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Edges\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_EDGES_FAILED;
	}
    }

    /* Fill in the node index table */
    bzero(ctdb->nodes, sizeof(CTDB_NODE_5 *) * ctdb->num_nodes);
    for(i=offset=0;(i<ctdb->num_nodes) && (offset<ctdb->node_size);i++)
    {
	ctdb->nodes[i] = (CTDB_NODE_5 *)(ctdb->node_storage + offset);
	offset += sizeof(CTDB_NODE_5) + sizeof(struct ctdb_connection) *
	  (ctdb->nodes[i]->num_connections - 1);
    }

    /* Fill in the edge index table */
    bzero(ctdb->edges, sizeof(CTDB_EDGE_5 *) * ctdb->num_edges);
    for(i=offset=0;(i<ctdb->num_edges) && (offset<ctdb->edge_size);i++)
    {
	ctdb->edges[i] = (CTDB_EDGE_5 *)(ctdb->edge_storage + offset);
	offset += sizeof(CTDB_EDGE_5) + sizeof(struct ctdb_connection) *
	  (ctdb->edges[i]->num_features - 1);
    }

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_quad_tree_5(CTDB_5 *ctdb, FILE *fileptr)
{
    if (ctdb->num_quad_data == 0 )
      return CTDB_STATUS_OK;
        
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces ctdb_read_quad_tree((CTDB *)ctdb,fileptr); */
    ctdb->quad_data =
      (CTDB_FEATURE_DATA_5 *)ctdb_alloc(NULL,
					sizeof(CTDB_FEATURE_DATA_5) *
					ctdb->num_quad_data,
					"abstract data and quadtree");

    if (fread(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_5),
	      ctdb->num_quad_data, fileptr) != ctdb->num_quad_data)
    {
	printf("CTDB: Read failed.  Quad tree and abstract data\n");
	/* Survive this error */
	ctdb->num_quad_data = 0;
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_linear_models_5(CTDB_5 *ctdb, FILE *fileptr)
{
    if (ctdb->num_linear_models == 0 )
    {
	ctdb->linear_models = NULL;
	return CTDB_STATUS_OK;
    }
        
    /* Allocate space for the linear models.  Then read the data.
     */
    ctdb->linear_models = NULL;
    ctdb->linear_models =
      (CTDB_LINEAR_MODEL_5 *)ctdb_alloc(NULL,
					sizeof(CTDB_LINEAR_MODEL_5) *
					ctdb->num_linear_models,
					"linear models");

    if (fread(ctdb->linear_models, sizeof(CTDB_LINEAR_MODEL_5),
	      ctdb->num_linear_models, fileptr) != ctdb->num_linear_models)
    {
	printf("CTDB: Read failed.  Linear models\n");
	/* Survive this error */
	ctdb->num_linear_models = 0;
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_volume_models_5(CTDB_5 *ctdb, FILE *fileptr)
{
    int32                i, model_size;
    CTDB_VOLUME_MODEL_5 *model;

    if (ctdb->num_volume_models == 0)
    {
	ctdb->volume_models = NULL;
	return CTDB_STATUS_OK;
    }

    /* Allocate space for the array of pointers to the volume models.  Then
     * read in each model.
     */
    ctdb->volume_models = NULL;
    ctdb->volume_models =
      (CTDB_VOLUME_MODEL_5 **)ctdb_alloc(NULL,
					 sizeof(CTDB_VOLUME_MODEL_5 *) *
					 ctdb->num_volume_models,
					 "volume models");

    /* Allocate space for a temporary model (large).
     */
    model = (CTDB_VOLUME_MODEL_5 *)STDALLOC(sizeof(CTDB_VOLUME_MODEL_5) +
					    1024 * 3 * sizeof(float32));

    for(i = 0; i < ctdb->num_volume_models; i++)
    {
	/* Read in all of the model except the vertices.
	 */
	if(fread(model, sizeof(CTDB_VOLUME_MODEL_5) - sizeof(float32), 1,
		 fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Volume models\n");
	    /* Survive this error */
	    ctdb->num_volume_models = 0;
	    return CTDB_STATUS_OK;
	}

	/* Now that we know how many vertices are in this model, read them in.
	 */
	if(fread(model->vertices, model->num_vertices * 3 * sizeof(float32),
		 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Volume models\n");
	    /* Survive this error */
	    ctdb->num_volume_models = 0;
	    return CTDB_STATUS_OK;
	}   

	model_size = sizeof(CTDB_VOLUME_MODEL_5) - sizeof(float32) +
	  model->num_vertices * 3 * sizeof(float32);

	/* Now that the entire model has been read in, store it.
	 */
	ctdb->volume_models[i] =
	  (CTDB_VOLUME_MODEL_5 *)ctdb_alloc(NULL, model_size,
					    "volume model");

	bcopy(model, ctdb->volume_models[i], model_size);
    }

    STDDEALLOC(model);

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_soil_tables_5(CTDB_5 *ctdb, FILE *fileptr)
{
    int32 i;

    ctdb->soil_tables = (int32 **)ctdb_alloc(NULL,
					     ctdb->num_soil_tables *
					     sizeof(int32 *),
					     "Soil table index");
    
    ctdb->soil_table_storage = (int8 *)ctdb_alloc(NULL,
						  ctdb->soil_table_size,
						  "Soil tables");

    /* Get the indices for the soil tables */
    fread(ctdb->soil_tables, sizeof(int32 *), ctdb->num_soil_tables, fileptr);

    /* These are kept in memory as pointers, but stored on disk as offsets.
     * Convert them to pointers.
     */
    for(i=0;i<ctdb->num_soil_tables;i++)
    {
	ctdb->soil_tables[i] = (int32 *)((int8 *)ctdb->soil_table_storage +
					 (int32)(ctdb->soil_tables[i]));
    }
    
    /* Read the soil tables themselves */
    fread(ctdb->soil_table_storage, 1, ctdb->soil_table_size, fileptr);
}

static CTDB_STATUS 
ctdb_read_water_chars_5(CTDB_5 *ctdb, FILE *fileptr)
{
    ctdb->water_chars = (struct ctdb_water_chars_5 *)
      ctdb_alloc(NULL, ctdb->num_water_chars * 
		 sizeof(struct ctdb_water_chars_5), "Water characteristics");
    
    fread(ctdb->water_chars, sizeof(struct ctdb_water_chars),
	  ctdb->num_water_chars, fileptr);    
}

static CTDB_STATUS 
ctdb_read_tidal_zones_5(CTDB_5 *ctdb, FILE *fileptr)
{
    ctdb->tidal_zones = (struct ctdb_tidal_zone_5 *)
      ctdb_alloc(NULL, ctdb->num_tidal_zones *
		 sizeof(struct ctdb_tidal_zone_5), "Tidal zones");
    
    fread(ctdb->tidal_zones, sizeof(struct ctdb_tidal_zone_5), 
	  ctdb->num_tidal_zones, fileptr);
}

static CTDB_STATUS
ctdb_read_all_patch_data_5(CTDB_5 *ctdb, FILE *f)
{
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_5 *pg;

    ctdb->patch_data =
      (ADDRESS)ctdb_alloc(NULL,
			  ctdb->n_patch_bytes,
			  "DB Patch data");

    /* Read _all_ the patch data for the cell in in one big gulp */
    fread(ctdb->patch_data,1,ctdb->n_patch_bytes,f);

    pg = ctdb->patch_groups;

    /* Now that we have the data, we need to set up the pointers in all the
     * patch headers
     */
    groups = (ctdb->patches_wide*ctdb->patches_high) / 
      CTDB_PATCHES_PER_GROUP_5;

    for (i=0;i<groups;i++)
    {
	/* Doesn't matter which elt of the union we choose here... */
	pg[i].data_start.grid_data = (CTDB_GRID_PG_DATA_5 *)
	  ((int8*)ctdb->patch_data + pg[i].data_start.offset);
	pg[i].mapped = 1;
    }

    return CTDB_STATUS_OK;    
}

void 
ctdb_read_elevations_5(CTDB_5 *ctdb, FILE *f)
{
    uint32 memory, pages;
    int32  i;
    int    file_pos;

    pages = ctdb->pages_wide * ctdb->pages_high;

    memory = PAGE_SIZE * pages;

    ctdb->elevations_resident = 1;
    ctdb->elevations = (int32 *)ctdb_alloc(NULL,
					   memory, "elevation data");

    /* Note the file pointer value */
    file_pos = ftell(f);

    /* Load as many pages as the caller will let us */
    if (fread(ctdb->elevations, memory, 1, f) != 1)
      printf("CTDB: Read failed.  First %d bytes of post data.\n",
	     memory), exit(1);

    /* Position the file pointer at the end of the elevations */
    file_pos += PAGE_SIZE * pages;
    fseek(f, file_pos, 0);
}


/* Actually reads both the header and the aux header */
static int32
ctdb_reread_hdr_5(char *fname, CTDB_5 *ctdb)
{
    CTDB_FILE_HEADER_5 hdr;
    CTDB_AUX_FILE_HEADER_5 aux_hdr;
    FILE *f;
    
    if (!fname)
      return 0;
    
    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	return 0;
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	return 0;
    }

    ctdb_unpack_header_5(ctdb, &hdr);

    if (fread(&aux_hdr, hdr.aux_hdr_bytes, 1, f) != 1)
    {	
	printf("CTDB: Error reading %s AUX header.\n",fname);
	fclose(f);
	return 0;
    }

    ctdb_unpack_aux_header_5(ctdb, &aux_hdr);
    return 1;
}

static void
ctdb_reread_5(char *fname, CTDB_5 *ctdb)
{
    FILE *f;
    CTDB_FILE_HEADER_5 hdr;
    CTDB_AUX_FILE_HEADER_5 aux_hdr;
    int32 total_mem;

    if (!fname)
    {
	printf("No terrain database name specified.\n");
	exit(-1);	
    }

    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	exit(-1);
    }

    if (hdr.format != 5)
    {
	printf("CTDB: Attempted to read fmt %d database using fmt 5 reader",
	       hdr.format);
	fclose(f);
	exit(-1);
    }

    ctdb->file = f;

    ctdb_unpack_header_5(ctdb, &hdr);

    /* Read in the auxiliary header info */
    if (fread(&aux_hdr, hdr.aux_hdr_bytes, 1, f) != 1)
    {
	printf("CTDB: Error reading %s AUX header.\n", fname);
	fclose(f);
	exit(-1);
    }
    ctdb_unpack_aux_header_5(ctdb, &aux_hdr);

    read_patch_group_headers_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_topology_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_quad_tree_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_linear_models_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_volume_models_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_soil_tables_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_water_chars_5(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_tidal_zones_5(ctdb, f);

    if ( ctdb->is_gridded )
      ctdb_read_elevations_5(ctdb, f);
    
    ctdb->patch_data_seek = ftell(f);

    /* Since we're in the compiler, we're not caching. */
    ctdb_read_all_patch_data_5(ctdb, f);

    /* We'll find the z-limits later (once we've converted to the compiler
     * representation.
     */

    return;
}

void *
ctdb_read_5(fname)
    char *fname;
{
    CTDB_5 *ctdb;
    
    ctdb = (CTDB_5 *)ctdb_alloc(NULL, sizeof(CTDB_5), "Creating fmt 5 ctdb");
    bzero(ctdb, sizeof(CTDB_5));

    printf("Attempting to read CTDB datafiles... \n");

    ctdb_reread_5(fname, ctdb);
    return ctdb;
}

/***************************************************************************
 * WRITE ROUTINES                                                           *
 ***************************************************************************/

/* The input database is a void pointer to make it easier to call this routine
 * through a generic interface; it is the caller's responsibility to ensure
 * that it is, in fact, a format 5 CTDB.
 * NOTE: since the compiler does not use caching, the inter-format conversion
 * routines do NOT generate the caching scheme correctly; therefore it is the
 * responsibility of the write routine to do so.
 */
void ctdb_write_5(fmt5_db, output_file)
    void *fmt5_db;
    FILE *output_file;
{
    CTDB_5 *ctdb = (CTDB_5 *)fmt5_db;
    CTDB_FILE_HEADER_5 hdr;
    CTDB_AUX_FILE_HEADER_5 aux_hdr;

    int32 groups = (ctdb->patches_wide *
		    ctdb->patches_high)/CTDB_PATCHES_PER_GROUP;

    int32 max_cache = 0;

    /* Generate the caching scheme for the patch data and fill in the
     * necessary data structures
     */
    generate_caching_scheme_5(ctdb);

    /**
     ** Write the header to the output file.
     **/
    
    ctdb_pack_header_5(ctdb, &hdr);
    fwrite(&hdr, sizeof(hdr), 1, output_file);

    /**
     ** Write AUX header to output file
     **/
    ctdb_pack_aux_header_5(ctdb, &aux_hdr);
    fwrite(&aux_hdr, sizeof(aux_hdr), 1, output_file);

    /**
     ** Write patch group headers
     **/
    write_patch_group_headers_5(ctdb, ctdb->patch_groups, groups, output_file);

    /**
     ** Write topology: the nodes and edges to the output file.
     **/
    write_nodes_and_edges_5(ctdb, output_file);

    /** 
     ** Write quad tree: the abstract feature quad tree to the output
     **          file.
     **/
    write_abstract_features_5(ctdb, output_file);

    /** 
     ** Write linear models: the models referenced by linear features
     **          to the output file.
     **/
    write_linear_models_5(ctdb, output_file);

    /** 
     ** Write volume models: the models referenced by volume features
     **          to the output file.
     **/
    write_volume_models_5(ctdb, output_file);

    /** 
     ** Write soil tables: the tables mapping local soil values to global ones
     ** to the output file.
     **/
    write_soil_tables_5(ctdb, output_file);

    /** 
     ** Write water chars: the sets of characteristics used for water areas
     ** to the output file.
     **/
    write_water_chars_5(ctdb, output_file);

    /** 
     ** Write tidal zones: the structures defining tidal variations to the
     ** output file.
     **/
    write_tidal_zones_5(ctdb, output_file);    

    /**
     ** Write elevation Posts: the elevations if there are gridded patches.
     **/
    if ( ctdb->is_gridded )
      write_elevation_posts_5(ctdb, output_file);
    
    /**
     ** Write patch group data: the patch group pointers, features, terrain
     ** elements, if tinned; the virtual grids, if tinned. 
     **/
    write_patch_data_5(ctdb, output_file);

    return;
}

/* The ctdb_pack_header_5 routine will create a file header (hdr) from the
 * fmt 5 ctdb data structure. 
 */
static void 
ctdb_pack_header_5(ctdb, hdr)
    CTDB_5             *ctdb;
    CTDB_FILE_HEADER_5 *hdr;
{
    bzero(hdr, sizeof(hdr));
    hdr->format = ctdb->input_format;
    hdr->is_curved = ctdb->is_curved;
    hdr->is_sw_ne = ctdb->is_sw_ne;
    hdr->is_gridded = ctdb->is_gridded;
    hdr->is_tinned  = ctdb->is_tinned;
    hdr->fixed_point_basis = ctdb->fixed_point_basis;

    hdr->num_soil_tables = ctdb->num_soil_tables;
    hdr->soil_table_size = ctdb->soil_table_size;
    hdr->num_water_chars = ctdb->num_water_chars;
    hdr->num_tidal_zones = ctdb->num_tidal_zones;

    hdr->post_increment = ctdb->incr;
    hdr->posts_per_patch_side = ctdb->patch_incr;
    hdr->pages_wide = ctdb->pages_wide;
    hdr->pages_high = ctdb->pages_high;
    hdr->min_x = ctdb->min_x;
    hdr->max_x = ctdb->max_x;
    hdr->min_y = ctdb->min_y;
    hdr->max_y = ctdb->max_y;
    bcopy(ctdb->patch_cache_size_map, hdr->patch_cache_size_map,
	  sizeof(ctdb->patch_cache_size_map));

    hdr->origin_northing = ctdb->origin_northing;
    hdr->origin_easting = ctdb->origin_easting;
    hdr->origin_zone_number = ctdb->origin_zone_number;
    hdr->origin_zone_letter = ctdb->origin_zone_letter;
    hdr->datum = ctdb->datum;
    bcopy(ctdb->name, hdr->name, sizeof(ctdb->name));
    hdr->version = ctdb->version;
    bcopy(ctdb->date, hdr->date, sizeof(ctdb->date));
    hdr->num_nodes = ctdb->num_nodes;
    hdr->node_size = ctdb->node_size;
    hdr->num_edges = ctdb->num_edges;
    hdr->edge_size = ctdb->edge_size;
    hdr->num_quad_data = ctdb->num_quad_data;
    hdr->n_patch_bytes = ctdb->n_patch_bytes;
    hdr->num_linear_models = ctdb->num_linear_models;
    hdr->num_volume_models = ctdb->num_volume_models;
    hdr->aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER_5);
}

/* Generate the caching scheme for the patch data and fill in the
 * necessary data structures
 */
static void
generate_caching_scheme_5(ctdb)
    CTDB_5 *ctdb;
{
    int32 i;
    int32 group;
    int32 num_groups;
    int32 max_size = 0;
    CTDB_PATCH_GROUP_5 *pg;
    int32 *pg_size_map;		/* pg_size_map[i] is the size, in words, of
				 * patch group i
				 */
    int32 *map;			/* Current entry in pg_size_map array */
    int32 *counts;		/* counts[i] is the number of patch groups
				 * with i words of data
				 */
    int32 size;			/* Current PG size we're looking at */
    int32 last_size;		/* Last non-empty PG size we looked at */
    int32 *sizes = ctdb->patch_cache_size_map; /* Sizes of the patch caches */
    int32 burdens[CTDB_PATCH_CACHES_5];	/* Number of PG's mapped to each 
					 * cache 
					 */
    int32 n_mapped;		/* Ideal number of groups per cache */
    int32 cur_count;		/* Groups mapped to current cache so far */
    int32 cur_cache;		/* Which cache we're currently filling */
    int32 min_burden;		/* Number of PGs mapped to least used cache */
    int32 smallest;
    int32 max_burden;		/* Number of PGs mapped to most used cache */
    int32 largest;		/* Most used cache */
    int32 best_cache;		/* "Best" cache for current PG */

    num_groups = (ctdb->patches_wide * ctdb->patches_high) /
      CTDB_PATCHES_PER_GROUP_5;

    pg_size_map = (int32 *)ctdb_alloc(NULL, num_groups * sizeof(int32),
				      "Patch group size map");


    /* 1.  Find the size of each patch group.  While we're at it, keep track
     * of how big the largest group is.
     */
    for(group = 0; group < num_groups; group++)
    {
	pg_size_map[group] = find_pg_size_5(ctdb->patch_groups + group) / 
	  CTDB_CACHE_BKT_UNITS_5;
	if (pg_size_map[group] > max_size)
	  max_size = pg_size_map[group];
    }

    max_size++;

    counts = (int32 *)ctdb_alloc(NULL, sizeof(int32) * max_size,
				 "patch group size counts");

    bzero(counts, max_size * sizeof(int32));
    bzero(sizes, CTDB_PATCH_CACHES_5 * sizeof(int32));
    bzero(burdens, CTDB_PATCH_CACHES_5 * sizeof(int32));

    /* 2. Find distribution of group sizes. 
     */
    map = pg_size_map;
    for(group=0;group<num_groups;group++,map++)
    {
      counts[*map]++;
}
    /* 3. Choose the ideal number of groups to be mapped to each cache.
     */
    n_mapped =
      ceil((float64)num_groups/(float64)CTDB_PATCH_CACHES_5) + 1;

    printf("Ideal groups/cache = %d\n",n_mapped);

    /* 4. Choose cache sizes such that AT LEAST n_mapped groups land
     * in each cache.
     */
    cur_count = cur_cache = 0;
    for(size=0;size<max_size;size++)
    {
	if (counts[size])
	{
	    cur_count += counts[size];
	    last_size = size;
	}

	if (cur_count > n_mapped)
	{
	    do
	    {
		sizes[cur_cache] = size;
		if (cur_count >= 2*n_mapped)
		  burdens[cur_cache] = n_mapped;
		else
		  burdens[cur_cache] = cur_count;
		cur_cache++;
		cur_count -= n_mapped;
	    } while (cur_count >= n_mapped);
	    cur_count = 0;
	}
    }
    sizes[cur_cache] = last_size;
    burdens[cur_cache] = cur_count;
    cur_cache++;

    /* 5. If not all the caches were used, split the most burdened
     * caches in two.
     */
    while(cur_cache<CTDB_PATCH_CACHES_5)
    {
	/* Find the worst off */
	max_burden = largest = 0;
	for(i=0;i<cur_cache;i++)
	  if (burdens[i] > max_burden)
	  {
	      max_burden = burdens[i];
	      largest = i;
	  }
	/* Make room for the split */
	for(i=cur_cache;i>largest;i--)
	{
	    sizes[i] = sizes[i-1];
	    burdens[i] = burdens[i-1];
	}
	/* Split it in two */
	burdens[largest] /= 2;
	burdens[largest+1] -= burdens[largest];
	cur_cache += 1;
    }

    /* 6. Tag the groups with their cache number.
     */
    bzero(burdens,sizeof(burdens));
    for(group=0;group<num_groups;group++)
    {
	pg = ctdb->patch_groups + group;
	map = pg_size_map + group;

	/* Search through the caches and find which ones we can put this
	 * patch into.  Choose the least burdened of those.
	 */
	min_burden = smallest = 0x7FFFFFF;
	best_cache = CTDB_PATCH_CACHES_5 - 1;
	for(cur_cache=0; cur_cache < CTDB_PATCH_CACHES_5; cur_cache++)
	{
	    /* Have we passed the range of caches for which this size
	     * is intended?
	     */
	    if (sizes[cur_cache] > smallest)
	      break;
	      
	    /* Will the patch fit in this cache? */
	    if (sizes[cur_cache] < *map)
	      continue;
	    smallest = sizes[cur_cache];
	    if (burdens[cur_cache] < min_burden)
	    {
		best_cache = cur_cache;
		min_burden = burdens[cur_cache];
	    }
	}
	burdens[best_cache]++;
	pg->cache = best_cache;
    }

    /* 7. Print the results.
     */
    size = 0;
    for (cur_cache=0; cur_cache < CTDB_PATCH_CACHES_5; cur_cache++)
    {
	if (cur_cache && (sizes[cur_cache] != sizes[cur_cache-1]))
	  size = sizes[cur_cache-1];
	printf("Cache %x: [%d->%d].  %d patch groups\n",
	       cur_cache, size + 1, sizes[cur_cache], burdens[cur_cache]);
    }

    free(counts);    
    return;
}

/* Return the size, in bytes, of all data, including the data header, in
 * pg
 */
static int32
find_pg_size_5(pg)
    CTDB_PATCH_GROUP_5 *pg;
{
    int32 size;

    if (pg->contains_tins)
    {
	int32 mem = -1;
	int32 num_vg;
	CTDB_TIN_PG_DATA_5 *pgt = pg->data_start.tin_data;

	size = sizeof(CTDB_TIN_PG_DATA_5);

	if (!pg->patch3_no_tes)
	{
	    mem = 3;
	}
	else if (!pg->patch2_no_tes)
	{
	    mem = 2;
	}
	else if (!pg->patch1_no_tes)
	{
	    mem = 1;
	}
	else
	{
	    mem = 0;
	}

	/* This gives us the number of bytes from the start of the data header
	 * to the start of the last set of vgrid data
	 */
	size = pgt->te_vgrid_offset[mem] * sizeof(CTDB_FEATURE_DATA_5);

	num_vg = pgt->te_vgrid_dim[mem] * pgt->te_vgrid_dim[mem];

	/* Round up to word boundary, remembering that te_vgrid entries
	 * are uint16s
	 */
	num_vg += (num_vg % 2);

	size += (num_vg * sizeof(uint16));
	return size;
    }
    else
    {
	/* Find the last occupied patch.  Add its offset to the number of bytes
	 * of data stored for it to get the total PG data size
	 */
	CTDB_FEATURE_DATA_5 *feat = NULL;

	if (pg->patch3_occupied)
	{
	    size = pg->data_start.grid_data->patch3_offset *
	      sizeof(CTDB_FEATURE_DATA_5);
	    feat = (CTDB_FEATURE_DATA_5 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch3_offset;
	}
	else if (pg->patch2_occupied)
	{
	    size = pg->data_start.grid_data->patch2_offset *
	      sizeof(CTDB_FEATURE_DATA_5);
	    feat = (CTDB_FEATURE_DATA_5 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch2_offset;
	}
	else if (pg->patch1_occupied)
	{
	    size = pg->data_start.grid_data->patch1_offset *
	      sizeof(CTDB_FEATURE_DATA_5);
	    feat = (CTDB_FEATURE_DATA_5 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch1_offset;
	}
	else if (pg->patch0_occupied)
	{
	    size = sizeof(CTDB_GRID_PG_DATA);

	    feat = (CTDB_FEATURE_DATA_5 *)pg->data_start.grid_data +
	      (sizeof(CTDB_GRID_PG_DATA_5) / sizeof(CTDB_FEATURE_DATA_5));
	}
	if (feat == NULL)
	  return sizeof(CTDB_GRID_PG_DATA_5);

	size += (feat->hdr.size * sizeof(CTDB_FEATURE_DATA_5 *));
	return size;
    }
}

/* The ctdb_pack_aux_header_5 routine will create an aux file header 
 * (aux_hdr) from the fmt. 5 ctdb data structure. 
 */
static void 
ctdb_pack_aux_header_5(ctdb, aux_hdr)
    CTDB_5                 *ctdb;
    CTDB_AUX_FILE_HEADER_5 *aux_hdr;
{
    aux_hdr->gcs_mode = ctdb->gcs_mode;
    aux_hdr->cell_id = ctdb->cell_id;
    aux_hdr->origin_x = ctdb->origin_x;
    aux_hdr->origin_y = ctdb->origin_y;
    aux_hdr->west_long = ctdb->west_long;
    aux_hdr->east_long = ctdb->east_long;
    aux_hdr->south_lat = ctdb->south_lat;
    aux_hdr->north_lat = ctdb->north_lat;

    return;
}

/* This routine writes the complete set of patch group headers. 
 * $$$ WARRNING:  as written, this changes the data_start values to be
 * offsets rather than pointers, rendering the database unusable in memory.
 * If you need to be able to use the database after writing it, this routine
 * will need to change
 */
static void
write_patch_group_headers_5(ctdb, patch_groups, groups, output_file)
    CTDB_5             *ctdb;
    CTDB_PATCH_GROUP_5 *patch_groups;
    int32               groups;
    FILE               *output_file;
{
    int32 i;

    for(i=0;i<groups;i++)
    {
	ctdb->patch_groups[i].data_start.offset = 
	  (int8 *)ctdb->patch_groups[i].data_start.grid_data -
	  (int8 *)ctdb->patch_data;
    }
	
    fwrite(patch_groups, sizeof(CTDB_PATCH_GROUP_5),groups, output_file);    
}

/* This function writes the nodes and edges to the output file.
 */
static void 
write_nodes_and_edges_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    /* Just dump the stored nodes and edges on out */
    if (ctdb->node_size)
      fwrite(ctdb->node_storage, 1, ctdb->node_size, output_file);
    if (ctdb->edge_size)
      fwrite(ctdb->edge_storage, 1, ctdb->edge_size, output_file);
}

/* This function writes the abstract features to the output file. */
static void 
write_abstract_features_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    /* Just dump the quad tree on out */
    if (ctdb->num_quad_data)
      fwrite(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_5),
	     ctdb->num_quad_data, output_file);
}

/* This function writes the linear models to the output file.
 */
static void 
write_linear_models_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    /* Just dump the linear models on out */
    if (ctdb->num_linear_models)
      fwrite(ctdb->linear_models, sizeof(CTDB_LINEAR_MODEL_5),
	     ctdb->num_linear_models, output_file);
}

/* This function writes the volume models to the output file.
 */
static void
write_volume_models_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    int32              i, j;

    /* Just dump the volume models on out */
    for(i = 0; i < ctdb->num_volume_models; i++)
    {
	/* Write the model, excluding the vertices, to the file.
	 */
	fwrite(ctdb->volume_models[i], sizeof(CTDB_VOLUME_MODEL_5) -
	       sizeof(float32), 1, output_file);

	/* Now loop through and write each vertex to the file.
	 */
	for(j = 0; j < ctdb->volume_models[i]->num_vertices; j++)
	{
	    fwrite(&ctdb->volume_models[i]->vertices[3 * j],
		   sizeof(float32), 3, output_file);
	}
    }
}

/* This function writes the soil tables to the output file */
static void
write_soil_tables_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    int32 *offsets;
    int32 i;

    if (ctdb->num_soil_tables)
    {
	/* First, write out the offsets for the soil tables.  These are 
	 * kept as pointers in memory, but stored as byte offsets on disk
	 */
	offsets = (int32 *)ctdb_alloc(NULL, 
				      ctdb->num_soil_tables * sizeof(int32),
				      "Soil table offsets");

	for(i=0;i<ctdb->num_soil_tables;i++)
	{
	    offsets[i] = (int8 *)ctdb->soil_tables[i] - 
	      ctdb->soil_table_storage;
	}

	fwrite(offsets, sizeof(int32), ctdb->num_soil_tables, output_file);
	
	/* Now, write out the tables themselves */
	fwrite(ctdb->soil_table_storage, sizeof(int8), ctdb->soil_table_size, 
	       output_file);
	
	STDDEALLOC(offsets);
    }	    
}

/* This function writes the water characteristics to the output file */
static void
write_water_chars_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    if (ctdb->num_water_chars)
      fwrite(ctdb->water_chars, sizeof(CTDB_WATER_CHARS_5),
	     ctdb->num_water_chars, output_file);    
}


/* This function writes the tidal zones to the output file */
static void
write_tidal_zones_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    if (ctdb->num_tidal_zones)
      fwrite(ctdb->tidal_zones, sizeof(CTDB_TIDAL_ZONE_5),
	     ctdb->num_tidal_zones, output_file);    
}

/* This function writes the elevation post data out to the data file.
   Note that this assumes all elevation posts are already in memory.
 */
static void 
write_elevation_posts_5(ctdb, output_file)
    CTDB_5  *ctdb;
    FILE    *output_file;
{
    int32 post_bytes;
    
    post_bytes = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
        
    fwrite(ctdb->elevations, post_bytes,1, output_file);
}

/* Writes the patch data out to the file */
static void
write_patch_data_5(ctdb, output_file)
    CTDB_5             *ctdb;
    FILE               *output_file;
{
    int32 i;
    int8 pad_byte = 0;
    int32 num_pad_bytes;

    /* Dump the patch data on out */
    fwrite(ctdb->patch_data, 1, ctdb->n_patch_bytes, output_file);

    /* To be perfectly safe, we'll pad the database out by the size of the
     * largest cache; this will ensure that the caching scheme does not
     * read past the end of the database.
     */
    num_pad_bytes = ctdb->patch_cache_size_map[CTDB_PATCH_CACHES_5 - 1] *
      CTDB_CACHE_BKT_UNITS_5;

    for(i = 0; i<num_pad_bytes; i++)
      fwrite(&pad_byte, sizeof(int8), 1, output_file);
    
    return;
}

/* old_base is the start address for all tables in an exisiting
 * database, and old_ptrs is an array of pointers, all of which point
 * into data stored in the data chunk pointed to by old_base.  new_base
 * points to a copy of the database pointed to by old_base; set up 
 * num_tables pointers in new_ptrs to point at the mirror of the data
 * pointed at by the corresponding pointers in old_ptrs
 */
static void
regen_table_ptrs(old_base, old_ptrs, new_base, new_ptrs, num_entries)
    int8   *old_base;
    int8  **old_ptrs;
    int8   *new_base;
    int8  **new_ptrs;
    int32   num_entries;
{
    int32 i;
    int32 offset;
    
    for(i=0;i<num_entries;i++)
    {
	offset = (int8 *)old_ptrs[i] - old_base;
	new_ptrs[i] = new_base + offset;
    }
    
    return;
}

/***************************************************************************
 * FMT 5 FREE ROUTINE                                                      *
 ***************************************************************************/

void
ctdb_free_fmt5_db(db)
    void *db;
{
    int32 i;
    CTDB_5 *ctdb = (CTDB_5 *)db;

#define CTDB_FREE_5(x)       \
    if (ctdb->x)             \
        STDDEALLOC(ctdb->x)

    if (!ctdb)
      return;
    
    CTDB_FREE_5(soil_tables);
    CTDB_FREE_5(soil_table_storage);
    CTDB_FREE_5(water_chars);
    CTDB_FREE_5(tidal_zones);
    
    CTDB_FREE_5(patch_groups);
    CTDB_FREE_5(cache_page_numbers);
    CTDB_FREE_5(elevation_pages);

    for(i = 0; i < CTDB_PATCH_CACHES_5; i++)
      CTDB_FREE_5(patch_data_cache[i]);

    CTDB_FREE_5(elevations);
    CTDB_FREE_5(quad_data);
    CTDB_FREE_5(nodes);
    CTDB_FREE_5(node_storage);
    CTDB_FREE_5(edges);
    CTDB_FREE_5(edge_storage);
    CTDB_FREE_5(linear_models);
    CTDB_FREE_5(volume_models);
    CTDB_FREE_5(patch_data);
    
    STDDEALLOC(ctdb);
    return;
}

/***************************************************************************
 * 5 => 6 CONVERSION ROUTINES                                              *
 ***************************************************************************/

extern void *
ctdb_convert_5_to_6(fmt5_db)
    void *fmt5_db;
{
    CTDB_5 *ctdb = (CTDB_5 *)fmt5_db;
    CTDB_6 *ctdb_out;
    int32   i, offset;
    
    ctdb_out = (CTDB_6 *)ctdb_alloc(NULL, sizeof(CTDB_6),
				      "Fmt. 6 database");
    
    bzero(ctdb_out, sizeof(CTDB_6));

    /* First, copy the easy stuff over */
#define COPY_5_6(x) (ctdb_out->x = ctdb->x)

    ctdb_out->fixed_point_basis = CTDB_STANDARD_FIXED_POINT_BASIS_6;
    ctdb_out->inv_fixed_point_basis = 1.0 / ctdb_out->fixed_point_basis;

    COPY_5_6(num_soil_tables);
    COPY_5_6(soil_table_size);
    COPY_5_6(num_water_chars);
    COPY_5_6(num_tidal_zones);
    COPY_5_6(incr);
    COPY_5_6(inv_incr);
    COPY_5_6(incrD);
    COPY_5_6(patch_incr);
    COPY_5_6(patch_inv_incr);
    COPY_5_6(patch_incrD);

    COPY_5_6(pages_wide);
    COPY_5_6(pages_high);
    COPY_5_6(patches_wide);
    COPY_5_6(patches_high);
    COPY_5_6(min_x);
    COPY_5_6(max_x);
    COPY_5_6(min_y);
    COPY_5_6(max_y);
    COPY_5_6(min_x_post);
    COPY_5_6(max_x_post);
    COPY_5_6(min_y_post);
    COPY_5_6(max_y_post);
    
    ctdb_out->input_format = 6;
    
    COPY_5_6(min_z);
    COPY_5_6(max_z);
    COPY_5_6(origin_northing);
    COPY_5_6(origin_easting);
    COPY_5_6(origin_zone_number);
    COPY_5_6(origin_zone_letter);
    COPY_5_6(datum);
    COPY_5_6(cell_id);
    
    COPY_5_6(origin_x);
    COPY_5_6(origin_y);
    
    COPY_5_6(west_long);
    COPY_5_6(east_long);
    COPY_5_6(south_lat);
    COPY_5_6(north_lat);

    COPY_5_6(gcs_mode);

    COPY_5_6(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_5_6(is_curved);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_5_6(num_quad_data);
    COPY_5_6(num_nodes);
    COPY_5_6(num_edges);
    COPY_5_6(num_linear_models);
    COPY_5_6(num_volume_models);

    COPY_5_6(meters_to_patch);
    COPY_5_6(is_gridded);
    COPY_5_6(is_tinned);
    COPY_5_6(max_patch_number);
    COPY_5_6(min_patch_number);
    
    /* Now for the real work */

    /* Soil tables */
    if (!ctdb->num_soil_tables)
    {
	ctdb_out->soil_table_storage = NULL;
	ctdb_out->soil_tables = NULL;
	ctdb_out->num_soil_tables = 0;
	ctdb_out->soil_table_size = 0;
    }
    else
    {
	ctdb_out->num_soil_tables = ctdb->num_soil_tables;
	ctdb_out->soil_tables =
	  (int32 **)ctdb_alloc(NULL, ctdb->num_soil_tables *
			       sizeof(int32 *),
			       "Fmt.6 soil table ptrs");
	ctdb_out->soil_table_size = ctdb->soil_table_size;
	ctdb_out->soil_table_storage = 
	  (int8 *)ctdb_alloc(NULL, ctdb_out->soil_table_size,
			     "Fmt. 6 soil table storage");

	bcopy(ctdb->soil_table_storage, ctdb_out->soil_table_storage,
	      ctdb->soil_table_size);
	
	regen_table_ptrs((int8 *)ctdb->soil_table_storage, 
			 (int8 **)ctdb->soil_tables,
			 (int8 *)ctdb_out->soil_table_storage, 
			 (int8 **)ctdb_out->soil_tables,
			 ctdb_out->num_soil_tables);
    }

    /* Copy the tidal zones */
    if (!ctdb_out->num_tidal_zones)
    {
	ctdb_out->tidal_zones = NULL;
	ctdb_out->num_tidal_zones = 0;
    }
    else
    {
	ctdb_out->num_tidal_zones = ctdb->num_tidal_zones;
	ctdb_out->tidal_zones = (CTDB_TIDAL_ZONE_6 *)
	  ctdb_alloc(NULL, sizeof(CTDB_TIDAL_ZONE_6) * ctdb->num_tidal_zones,
		     "Fmt. 6 tidal zones.");
	bcopy(ctdb->tidal_zones, ctdb_out->tidal_zones,
	      sizeof(CTDB_TIDAL_ZONE_6) * ctdb_out->num_tidal_zones);
    }

    /* Copy the water char entries */
    if (!ctdb->num_water_chars)
    {
	ctdb_out->water_chars = NULL;
	ctdb_out->num_water_chars = 0;
    }
    else
    {
	ctdb_out->num_water_chars = ctdb->num_water_chars;
	ctdb_out->water_chars = (CTDB_WATER_CHARS_6 *)
	  ctdb_alloc(NULL, sizeof(CTDB_WATER_CHARS_6) * ctdb->num_water_chars,
		     "Fmt. 6 water chars.");
	bcopy(ctdb->water_chars, ctdb_out->water_chars, ctdb->num_water_chars *
	      sizeof(CTDB_WATER_CHARS_6));
    }

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));

    /* Patch features */
    {
	int32 groups = (ctdb_out->patches_wide * ctdb_out->patches_high) /
	  CTDB_PATCHES_PER_GROUP_6;

	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_6 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_6),
		     "Fmt. 6 patch groups");

	ctdb_out->n_patch_bytes = convert_features_5_6(ctdb, ctdb_out);
    }

    /* Elevations */
    if (!ctdb->is_gridded)
    {
	ctdb_out->elevations = NULL;
    }
    else
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 6 elevation data");

	convert_elevation_posts_5_6(ctdb, ctdb_out);
    }
    
    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_6 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data * 
		     sizeof(CTDB_FEATURE_DATA_6),
		     "Fmt. 6 quadtree");

	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_6));
    }

    /* The compiler rep does not read from the topology section, so
     * we won't bother populating it, especially since it would be
     * painful.
     */

    /* Linear models */
    if (!ctdb->num_linear_models)
    {
	ctdb_out->linear_models = NULL;
    }
    else
    {
	ctdb_out->linear_models = 
	  (CTDB_LINEAR_MODEL_6 *)ctdb_alloc(NULL,
					    sizeof(CTDB_LINEAR_MODEL_6) *
					    ctdb->num_linear_models,
					    "Fmt. 6 linear models");

	bcopy(ctdb->linear_models, ctdb_out->linear_models,
	      ctdb->num_linear_models * sizeof(CTDB_LINEAR_MODEL_6));
    }

    /* Volume models */
    if (!ctdb->num_volume_models)
    {
	ctdb_out->volume_models = NULL;
    }
    else
    {
	ctdb_out->volume_models = (CTDB_VOLUME_MODEL_6 **)
	  ctdb_alloc(NULL,
		     sizeof(CTDB_VOLUME_MODEL_6 *) * ctdb->num_volume_models,
		     "Fmt. 6 volume model pointers");
    }

    {
	int32 model_size;
	CTDB_VOLUME_MODEL_5 *old_model;
	int32 i;

	for(i=0;i<ctdb->num_volume_models;i++)
	{
	    old_model = ctdb->volume_models[i];
	    model_size = sizeof(CTDB_VOLUME_MODEL_5) - sizeof(float32) +
	      old_model->num_vertices * 3 * sizeof(float32);

	    ctdb_out->volume_models[i] = (CTDB_VOLUME_MODEL_6 *)
	      ctdb_alloc(NULL, model_size, "Fmt. 6 volume model");

	    bcopy(old_model, ctdb_out->volume_models[i], model_size);
	}
    }    

    return ctdb_out;
}

/* Converts PGs and features (including TINs) from format 5 to format 6.
 * Returns total patch data bytes in format 6 result.
 */
static int32
convert_features_5_6(ctdb_5, ctdb_6)
    CTDB_5 *ctdb_5;
    CTDB_6 *ctdb_6;    
{
    int32 groups;
    int32 i;
    int32 total_bytes = 0;
    ADDRESS base;
	
    groups = ctdb_5->patches_wide * ctdb_5->patches_high /
      CTDB_PATCHES_PER_GROUP_5;
	
    ctdb_6->n_patch_bytes = ctdb_5->n_patch_bytes;

    ctdb_6->patch_data = (ADDRESS)ctdb_alloc(NULL,
					     ctdb_6->n_patch_bytes,
					     "Fmt. 6 patch data");

    base = ctdb_6->patch_data;

    for(i=0;i<groups;i++)
    {
	total_bytes += convert_group_5_6(ctdb_5, ctdb_6, i, &base);
    }

#ifdef DEBUG
    if (total_bytes > ctdb_6->n_patch_bytes)
      printf("ERROR:  too few bytes allocated for format 6 features!\n");
#endif /* DEBUG */

    return total_bytes;
}

/* Converts PG header and all PG data from format 5 to format 6.  group is
 * the patch group number to convert, base is an in-out variable passing
 * in the address where the format 6 data should start, and passing out
 * where the format 6 data for the next PG should start.
 */
static int32
convert_group_5_6(ctdb_5, ctdb_6, group, base)
    CTDB_5  *ctdb_5;
    CTDB_6  *ctdb_6;
    int32    group;
    ADDRESS *base;
{
    int32 bytes;
    
    /* First, convert the header. */
    convert_pg_5_6(ctdb_5, ctdb_6, group, *base);

    /* Now convert the data and the data headers */
    if (ctdb_6->patch_groups[group].contains_tins)
    {
	bytes = convert_tin_pg_data_5_6(ctdb_5, ctdb_6, group, *base);
    }
    else
    {
	bytes = convert_grid_pg_data_5_6(ctdb_5, ctdb_6, group, *base);
    }
    
    *base = (int8 *)*base + bytes;
    return bytes;
}

/* Convert just the patch group header from format 5 to format 6.  base is
 * where the new PG should point.
 */
static void
convert_pg_5_6(ctdb_5, ctdb_6, group, base)
    CTDB_5  *ctdb_5;
    CTDB_6  *ctdb_6;
    int32    group;
    ADDRESS  base;
{
    CTDB_PATCH_GROUP_5 *pg_5;
    CTDB_PATCH_GROUP_6 *pg_6;

    pg_5 = ctdb_5->patch_groups + group;
    pg_6 = ctdb_6->patch_groups + group;
    
    bzero(pg_6, sizeof(CTDB_PATCH_GROUP_6));

    pg_6->data_start.grid_data = (CTDB_GRID_PG_DATA_6 *)base;
    
    /* We'll leave the cache stuff blank, since the compiler won't use
     * it anyway.
     */

    pg_6->contains_tins = pg_5->contains_tins;

    pg_6->patch0_occupied = pg_5->patch0_occupied;
    pg_6->patch0_dyn_terr = pg_5->patch0_dyn_terr;
    pg_6->patch0_te16 = pg_5->patch0_te16;
    pg_6->patch0_no_tes = pg_5->patch0_no_tes;
    pg_6->patch0_water = pg_5->patch0_water;

    pg_6->patch1_occupied = pg_5->patch1_occupied;
    pg_6->patch1_dyn_terr = pg_5->patch1_dyn_terr;
    pg_6->patch1_te16 = pg_5->patch1_te16;
    pg_6->patch1_no_tes = pg_5->patch1_no_tes;
    pg_6->patch1_water = pg_5->patch1_water;

    pg_6->patch2_occupied = pg_5->patch2_occupied;
    pg_6->patch2_dyn_terr = pg_5->patch2_dyn_terr;
    pg_6->patch2_te16 = pg_5->patch2_te16;
    pg_6->patch2_no_tes = pg_5->patch2_no_tes;
    pg_6->patch2_water = pg_5->patch2_water;

    pg_6->patch3_occupied = pg_5->patch3_occupied;
    pg_6->patch3_dyn_terr = pg_5->patch3_dyn_terr;
    pg_6->patch3_te16 = pg_5->patch3_te16;
    pg_6->patch3_no_tes = pg_5->patch3_no_tes;
    pg_6->patch3_water = pg_5->patch3_water;

    return;
}


/* Convert all data, including data header, for a tinned PG from format 5 to
 * format 6.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_tin_pg_data_5_6(ctdb_5, ctdb_6, group, base)
    CTDB_5 *ctdb_5;
    CTDB_6 *ctdb_6;
    int32   group;
    ADDRESS base;
{
    int32 i;
    int32 size;
    CTDB_TIN_PG_DATA_5 *data_5;
    CTDB_TIN_PG_DATA_6 *data_6;
    CTDB_PATCH_GROUP_6 *pg;
    CTDB_FEATURE_DATA_5 *feat_5;
    CTDB_FEATURE_DATA_6 *feat_6;
    int32 patch_bytes = 0;
    int32 delta;
    int32 te_offset_5;
    int32 te_offset_6;

    data_5 = ctdb_5->patch_groups[group].data_start.tin_data;
    data_6 = (CTDB_TIN_PG_DATA_6 *)base;
    size = sizeof(CTDB_TIN_PG_DATA_6);
    
    pg = ctdb_6->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_6, size);
    bcopy(data_5->max_elev, data_6->max_elev, 
	  CTDB_PATCHES_PER_GROUP_6 * sizeof(float32));
    
    bcopy(data_5->water_elev, data_6->water_elev,
	  CTDB_PATCHES_PER_GROUP_6 * sizeof(float32));

    bcopy(data_5->tables, data_6->tables,
	  CTDB_PATCHES_PER_GROUP_6 * sizeof(CTDB_PATCH_TABLE_REF_6));

    /* Can't fill in offsets yet... */

    bcopy(data_5->feature_present, data_6->feature_present,
	  CTDB_MAX_POSTS_PER_PATCH_5 * 2);
    
    bcopy(data_5->num_te, data_6->num_te, 
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(uint16));
    
    /* Can't fill in te_offsets yet */

    bcopy(data_5->te_vgrid_dim, data_6->te_vgrid_dim,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(uint8));
    
    /* Can't fill in vgrid_offsets yet */

    /* 2.  Convert the features for each patch */
    feat_6 = (CTDB_FEATURE_DATA_6 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)data_5 +
					 sizeof(CTDB_TIN_PG_DATA_5));
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));

	data_6->patch1_offset = (patch_bytes + sizeof(CTDB_TIN_PG_DATA_6)) /
	  sizeof(CTDB_FEATURE_DATA_6);
    }
    else
    {
	data_6->patch1_offset = sizeof(CTDB_TIN_PG_DATA_6) /
	  sizeof(CTDB_FEATURE_DATA_6);
    }

    if (pg->patch1_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch1_offset);
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));

	data_6->patch2_offset = data_6->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));
    }
    else
    {
	data_6->patch2_offset = data_6->patch1_offset;
    }

    if (pg->patch2_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch2_offset);
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));

	data_6->patch3_offset = data_6->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));
    }
    else
    {
	data_6->patch3_offset = data_6->patch2_offset;
    }

    if (pg->patch3_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch3_offset);
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));
    }	  

    /* 3. Calculate how far the TE's for the patch have moved relative to
     * where they used to be
     */
    if (!pg->patch0_no_tes)
    {
	te_offset_5 = data_5->te_offset[0];
    }
    else if (!pg->patch1_no_tes)
    {
	te_offset_5 = data_5->te_offset[1];
    }
    else if (!pg->patch2_no_tes)
    {
	te_offset_5 = data_5->te_offset[2];
    }
    else
    {
	te_offset_5 = data_5->te_offset[3];
    }
    
    te_offset_6 = feat_6 - (CTDB_FEATURE_DATA_6 *)data_6;
    
    /* Now, calculate the offset, in words */
    delta = te_offset_6 - te_offset_5;
    
#ifdef DEBUG
    if (delta < 0)
      printf("Unreasonable data offset in 5->6 conversion.\n");
#endif /* DEBUG */
    
    /* 4.  Fill out the rest of the offset values. */
    for(i=0;i<CTDB_PATCHES_PER_GROUP_6;i++)
    {
	data_6->te_offset[i] = data_5->te_offset[i] + delta;
	data_6->te_vgrid_offset[i] = data_5->te_vgrid_offset[i] + delta;
    }
    
    /* 5.  Copy over all the TE data. */
    {
	int32 data_size;
	int32 te_data_size;	/* In bytes */

	if (group < ctdb_6->max_patch_number)
	{
	    
	    /* Total bytes of data stored for this PG */
	    data_size = 
	      (int8 *)ctdb_5->patch_groups[group+1].data_start.grid_data -
	      (int8 *)data_5;
	    
	    te_data_size = data_size - 
	      (data_5->te_offset[0] * sizeof(CTDB_FEATURE_DATA_5));
	}
	else
	{
	    int32 vgrid_size;
	    
	    if (!pg->patch3_no_tes)
	    {
		data_size = data_5->te_vgrid_offset[3] *
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[3] * data_5->te_vgrid_dim[3];
	    }
	    else if (!pg->patch2_no_tes)
	    {
		data_size = data_5->te_vgrid_offset[2] *
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[2] * data_5->te_vgrid_dim[2];
	    }
	    else if (!pg->patch1_no_tes)
	    {
		data_size = data_5->te_vgrid_offset[1] *
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[1] * data_5->te_vgrid_dim[1];
	    }
	    else
	    {
		data_size = data_5->te_vgrid_offset[0] *
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[0] * data_5->te_vgrid_dim[0];
	    }

	    /* Add padding as necessary */
	    vgrid_size += (vgrid_size % sizeof(CTDB_FEATURE_DATA_5));
	    data_size += vgrid_size;		
	    te_data_size = (data_size -
			    (te_offset_5 * sizeof(CTDB_FEATURE_DATA_5)));
	}

	feat_5 = (CTDB_FEATURE_DATA_5 *)data_5 + te_offset_5;
	bcopy(feat_5, feat_6, te_data_size);
	
	/* Calculate total bytes in this PG */
	size = ((int8 *)feat_6 - (int8 *)data_6) + te_data_size;
    }

    return size;
}

/* Convert all data, including data header, for a tinned PG from format 5 to
 * format 6.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_grid_pg_data_5_6(ctdb_5, ctdb_6, group, base)
    CTDB_5 *ctdb_5;
    CTDB_6 *ctdb_6;
    int32   group;
    ADDRESS base;
{
    int32 size;
    CTDB_GRID_PG_DATA_5 *data_5;
    CTDB_GRID_PG_DATA_6 *data_6;
    CTDB_PATCH_GROUP_6 *pg;
    CTDB_FEATURE_DATA_5 *feat_5;
    CTDB_FEATURE_DATA_6 *feat_6;
    int32 patch_bytes = 0; 

    data_5 = ctdb_5->patch_groups[group].data_start.grid_data;
    data_6 = (CTDB_GRID_PG_DATA_6 *)base;
    size = sizeof(CTDB_GRID_PG_DATA_6);
    
    pg = ctdb_6->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_6, size);
    bcopy(data_5->max_elev, data_6->max_elev, 
	  CTDB_PATCHES_PER_GROUP_6 * sizeof(float32));
    
    bcopy(data_5->water_elev, data_6->water_elev,
	  CTDB_PATCHES_PER_GROUP_6 * sizeof(float32));

    bcopy(data_5->tables, data_6->tables,
	  CTDB_PATCHES_PER_GROUP_6 * sizeof(CTDB_PATCH_TABLE_REF_6));

    /* Can't fill in offsets yet... */

    /* 2.  Convert the features for each patch */
    feat_6 = (CTDB_FEATURE_DATA_6 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)data_5 +
					 sizeof(CTDB_GRID_PG_DATA_5));
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));

	data_6->patch1_offset = (patch_bytes + sizeof(CTDB_GRID_PG_DATA_6)) /
	  sizeof(CTDB_FEATURE_DATA_6);
    }
    else
    {
	data_6->patch1_offset = sizeof(CTDB_GRID_PG_DATA_6) /
	  sizeof(CTDB_FEATURE_DATA_6);
    }

    if (pg->patch1_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch1_offset);
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));

	data_6->patch2_offset = data_6->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));
    }
    else
    {
	data_6->patch2_offset = data_6->patch1_offset;
    }

    if (pg->patch2_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch2_offset);
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));

	data_6->patch3_offset = data_6->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));
    }
    else
    {
	data_6->patch3_offset = data_6->patch2_offset;
    }

    if (pg->patch3_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch3_offset);
	
	patch_bytes = convert_feat_5_6(ctdb_6, feat_5, feat_6);
	feat_6 = feat_6 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_6));
    }	  

    size = (int8 *)feat_6 - (int8 *)data_6;
    return size;    
}

/* Converts features for a patch from format5 to format6.  feat5 must point
 * to the start of the features in the source patch, and feat6 must point
 * to where the features should start in the destination patch.  Returns
 * number of bytes written into dest. patch.
 */
static int32 
convert_feat_5_6(ctdb_6, feat_5, feat_6)
    CTDB_6              *ctdb_6;
    CTDB_FEATURE_DATA_5 *feat_5;
    CTDB_FEATURE_DATA_6 *feat_6;
{
    CTDB_FEATURE_DATA_5 *last, *next_5;
    CTDB_FEATURE_DATA_6 *cur_6, *next_6;
    CTDB_FEAT_OBJ_PTR obj;
    int32 ftype, verts;

    last = feat_5 + feat_5->hdr.size;
    cur_6 = feat_6;
    feat_5++;
    cur_6++;
    
    for(;feat_5<last;(feat_5 = next_5), (cur_6 = next_6))
    {
	ftype = feat_5->info.feature_class;

	switch(ftype)
	{
	  case FC_MICRO_5:
	    if((feat_5->info.vertices & MICRO_TYPE) == MICRO_TYPE_WATER_5)
	      convert_mwater_5_6(ctdb_6, feat_5, cur_6, &next_5, &next_6);
	    else
	      convert_micro_5_6(ctdb_6, feat_5, cur_6, &next_5, &next_6);
	    break;
	  case FC_VOLUME_5:
	    convert_volume_5_6(ctdb_6, feat_5, cur_6, &next_5, &next_6);
	    break;
	  case FC_LINEAR_5:
	    convert_linear_5_6(ctdb_6, feat_5, cur_6, &next_5, &next_6);
	    break;
	  case FC_CANOPY_5:
	    convert_canopy_5_6(ctdb_6, feat_5, cur_6, &next_5, &next_6);
	    break;
	  case FC_LAID_LINEAR_5:
	    convert_laid_linear_5_6(ctdb_6, feat_5, cur_6, &next_5, &next_6);
	    break;
	  default:
	    printf("ERROR: unknown feature type encountered in format 5 db.\n");
	    exit(-1);
	}
    }

    feat_6->hdr.size = next_6 - feat_6;
    return (feat_6->hdr.size * sizeof(CTDB_FEATURE_DATA_6));
}

static void
convert_micro_5_6(ctdb_6, feat_5, feat_6, next_5, next_6)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_6 **next_6;
{
    /* No change from format 5 to format 6, just copy it over */

    *next_5 = feat_5 + 4;
    *next_6 = feat_6 + 4;

    bcopy(feat_5, feat_6, 4 * sizeof(CTDB_FEATURE_DATA_5));
    return;    
}

static void
convert_mwater_5_6(ctdb_6, feat_5, feat_6, next_5, next_6)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_6 **next_6;
{
    /* No change from format 5 to format 6, just copy it over */

    *next_5 = feat_5 + 5;
    *next_6 = feat_6 + 5;

    bcopy(feat_5, feat_6, 5 * sizeof(CTDB_FEATURE_DATA_5));
    return;    
}

static void
convert_volume_5_6(ctdb_6, feat_5, feat_6, next_5, next_6)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_6 **next_6;
{
    /* No change from format 5 to format 6, just copy it over */

    int32 verts = feat_5->info.vertices;
    int32 size = 3 + 2 * verts;

    *next_5 = feat_5 + size;
    *next_6 = feat_6 + size;

    bcopy(feat_5, feat_6, size * sizeof(CTDB_FEATURE_DATA_5));
    return;    
}


static void
convert_linear_5_6(ctdb_6, feat_5, feat_6, next_5, next_6)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_6 **next_6;
{
    /* No change from format 5 to format 6, just copy it over */

    int32 verts = feat_5->info.vertices;
    int32 size = 2 + 2 * verts;

    *next_5 = feat_5 + size;
    *next_6 = feat_6 + size;

    bcopy(feat_5, feat_6, size * sizeof(CTDB_FEATURE_DATA_5));
    return;    
}

static void
convert_canopy_5_6(ctdb_6, feat_5, feat_6, next_5, next_6)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_6 **next_6;
{
    /* No change from format 5 to format 6, just copy it over */

    *next_5 = feat_5 + 4;
    *next_6 = feat_6 + 4;

    bcopy(feat_5, feat_6, 4 * sizeof(CTDB_FEATURE_DATA_5));
    return;
}

static void
convert_laid_linear_5_6(ctdb_6, feat_5, feat_6, next_5, next_6)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_6 **next_6;
{
    int32 verts = feat_5->info.vertices;

    /* No change from format 5 to format 6, just copy it over */

    *next_5 = feat_5 + 2 + verts;
    *next_6 = feat_6 + 2 + verts;

    bcopy(feat_5, feat_6, (verts + 2) * sizeof(CTDB_FEATURE_DATA_5));
    return;
}

static int32
get_post_index_6(ctdb, norm_x, norm_y)
    CTDB_6 *ctdb;
    int32   norm_x;
    int32   norm_y;
{
    int32 page, index;
    
    page = (norm_x>>SIDE_SHIFT) +  
      ctdb->pages_wide * (norm_y>>SIDE_SHIFT);
    index = (norm_x&SIDE_MASK) + POSTS_PER_SIDE * 
      (norm_y&SIDE_MASK);
    return (page*POSTS_PER_PAGE + index);
}

static void
convert_elevation_posts_5_6(ctdb_5, ctdb_6)
    CTDB_5 *ctdb_5;
    CTDB_6 *ctdb_6;
{
    int32 patch;
    int32 px;
    int32 py;
    int32 base_px;
    int32 base_py;
    int32 *post_5, *post_6;
    int32 index;
    float64 elevation;

    for(patch=0;patch<ctdb_5->max_patch_number;patch++)
    {
	base_py = (patch / ctdb_5->patches_wide) * ctdb_5->patch_incr;
	base_px = (patch % ctdb_5->patches_wide) * ctdb_5->patch_incr;

	for(py=0;py<ctdb_5->patch_incr;py++)
	  for(px=0;px<ctdb_5->patch_incr;px++)
	  {
	      index = get_post_index_6(ctdb_6, base_px + px, base_py + py);
	      post_5 = ctdb_5->elevations + index;
	      post_6 = ctdb_6->elevations + index;

	      /* Copy the post data over and reassign the elevation.
	       * For format 6, one bit of elevation data was stolen and used
	       * to indicate the diagonalization for that post.
	       */
	      *post_6 = *post_5;

	      /* Clear the elevation data.
	       */
	      *post_6 &= ~0x0007FFFF;

	      elevation = GET_POST_ELEVATION_5(*post_5,
					       ctdb_5->fixed_point_basis);

	      *post_6 |= BUILD_POST_ELEVATION_6(elevation,
						ctdb_6->inv_fixed_point_basis);

	      if(ctdb_5->is_sw_ne)
		*post_6 |= POST_SW_NE_6;
	      else
		*post_6 &= ~POST_SW_NE_6;

	  }
    }
    return;    
}


/***************************************************************************
 * 5 -> 4 CONVERSION                                                       *
 ***************************************************************************/

void *
ctdb_convert_5_to_4(fmt5_db)
    void *fmt5_db;
{
    CTDB_5 *ctdb = (CTDB_5 *)fmt5_db;
    CTDB_4 *ctdb_out;
    int32 *pre_ll_patch_delta; /* Array specifying how far back the laid 
				* linears for each patch moved from fmt.5 to
				* fmt. 4--we need this to update the topology
				*/

    ctdb_out = (CTDB_4 *)ctdb_alloc(NULL, sizeof(CTDB_4),
				    "Fmt. 4 database");
    
    bzero(ctdb_out, sizeof(CTDB_4));

    /* First, copy the easy stuff over */
#define COPY_5_4(x) (ctdb_out->x = ctdb->x)

    COPY_5_4(fixed_point_basis);
    COPY_5_4(inv_fixed_point_basis);

    COPY_5_4(incr);
    COPY_5_4(inv_incr);
    COPY_5_4(incrD);

    COPY_5_4(patch_incr);
    COPY_5_4(patch_inv_incr);
    COPY_5_4(patch_incrD);

    COPY_5_4(pages_wide);
    COPY_5_4(pages_high);
    COPY_5_4(patches_wide);
    COPY_5_4(patches_high);
    COPY_5_4(min_x);
    COPY_5_4(max_x);
    COPY_5_4(min_y);
    COPY_5_4(max_y);
    COPY_5_4(min_x_post);
    COPY_5_4(max_x_post);
    COPY_5_4(min_y_post);
    COPY_5_4(max_y_post);
    
    ctdb_out->input_format = 4;
    
    COPY_5_4(min_z);
    COPY_5_4(max_z);
    COPY_5_4(origin_northing);
    COPY_5_4(origin_easting);
    COPY_5_4(origin_zone_number);
    COPY_5_4(origin_zone_letter);
    COPY_5_4(datum);

    COPY_5_4(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_5_4(is_curved);
    COPY_5_4(is_sw_ne);

    ctdb_out->elevations_resident = TRUE;
    
    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_5_4(num_quad_data);
    COPY_5_4(num_nodes);
    COPY_5_4(node_size);
    COPY_5_4(num_edges);
    COPY_5_4(edge_size);

    COPY_5_4(meters_to_patch);
    COPY_5_4(is_gridded);
    COPY_5_4(is_tinned);
    COPY_5_4(max_patch_number);
    COPY_5_4(min_patch_number);
    
    /* Now for the real work */

    /* Soil tables -- we'll assume that the first soil table is the "right"
     * one to use
     */
    bcopy(ctdb->soil_tables[0], ctdb_out->soil_map, 
	  sizeof(ctdb_out->soil_map));
    
    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));
    

    /* Unfortunately, feature sizes changed from fmt. 4 to fmt. 5.  We
     * could go through and calculate the new patch data size, but that
     * would involve iterating over all features, which we would then
     * have to do again when we copied the feature data.  Instead, we'll
     * guess high, improving performance at the cost of some memory.
     */
    {
	int32 groups = (ctdb_out->patches_wide * ctdb_out->patches_high) /
	  CTDB_PATCHES_PER_GROUP_4;

	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_4 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_4),
		     "Fmt. 4 patch groups");

	pre_ll_patch_delta = (int32 *)
	  ctdb_alloc(NULL, groups * CTDB_PATCHES_PER_GROUP * sizeof(int32),
		     "Patch feature deltas 5 -> 4");

	bzero(pre_ll_patch_delta, 
	      groups * CTDB_PATCHES_PER_GROUP * sizeof(int32));

	ctdb_out->n_patch_bytes = convert_features_5_4(ctdb, ctdb_out,
						       pre_ll_patch_delta);
    }

    /* Elevations */
    if (!ctdb->is_gridded)
    {
	ctdb_out->elevations = NULL;
    }
    else
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Compiler rep. elevation data");

	bcopy(ctdb->elevations, ctdb_out->elevations, elev_mem);
	
	/* The soil maps may have changed.  If so, we have to fix all
	 * the soil references
	 */
	if (ctdb->num_soil_tables > 1)
	  convert_post_soils_5_4(ctdb, ctdb_out);
    }

    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_4 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data * 
		     sizeof(CTDB_FEATURE_DATA_4),
		     "Fmt. 4 quadtree");
	
	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_4));

    }
    
    /* Topology */
    ctdb_out->node_storage = (int8 *)
      ctdb_alloc(NULL, ctdb->node_size, "Fmt.4 nodes");
    bcopy(ctdb->node_storage, ctdb_out->node_storage, ctdb->node_size);
    
    ctdb_out->edge_storage = (int8 *)
      ctdb_alloc(NULL, ctdb->edge_size, "Fmt. 4 edges");
    bcopy(ctdb->edge_storage, ctdb_out->edge_storage, ctdb->edge_size);

    /* In copying the features, we've offset some of them from where
     * they used to be.  We need to fix the feature references
     * in the feature topology section.
     */
    fix_topo_refs_5_4(ctdb, ctdb_out, pre_ll_patch_delta);

    return ctdb_out;
}


/* Return a pointer to the soil map for a patch */
static int32 *
get_soil_map_5(ctdb_5, patch_number)
    CTDB_5 *ctdb_5;
    int32   patch_number;
{
    CTDB_PATCH_GROUP_5 *pg;
    int32 member;
    int32 index;
    
    pg = ctdb_5->patch_groups + PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);
    
    index = pg->data_start.grid_data->tables[member].soil_table;

    return (ctdb_5->soil_tables[index]);
}

/* See if there is a entry for the given global soil type in the fmt. 4
 * ctdb's soil map.  If so, return the appropriate local soil type.  If
 * not, return an unknown soil type
 */
static int32
convert_soil_5_4(ctdb_4, soil)
    CTDB_4 *ctdb_4;
    int32   soil;
{
    int32 i;
    
    for(i=0;i<CTDB_N_SOILS_4;i++)
    {
	if (ctdb_4->soil_map[i] == soil)
	  return i;
    }
    return CTDB_ILLEGAL_SOIL_5;    
}

static void
convert_micro_5_4(ctdb_5, ctdb_4, feat_5, feat_4, next_5, next_4, patch_number)
    CTDB_5               *ctdb_5;
    CTDB_4               *ctdb_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_4 **next_4;
    int32                 patch_number;
{
    int32 micro_type;
    int32 *soil_map_4;
    int32 *soil_map_5;

    /* The only change to microterrain format from format 4 to 5 was the
     * addition of microwater (we'll just throw microwater polygons out).
     * However, the soil types could have changed...
     */
    micro_type = (feat_5->info.vertices & MICRO_TYPE_5);
    if (micro_type == MICRO_TYPE_WATER_5)
    {
	*next_5 = feat_5 + 5;
	*next_4 = feat_4;
	return;
    }
    
    *next_4 = feat_4 + 4;
    *next_5 = feat_5 + 4;
    
    bcopy(feat_5, feat_4, 4 * sizeof(CTDB_FEATURE_DATA_4));

    soil_map_4 = ctdb_4->soil_map;
    soil_map_5 = get_soil_map_5(ctdb_5, patch_number);

    feat_4 = feat_4 + 3;

    if (soil_map_4[feat_4->micro.soil0] != soil_map_5[feat_4->micro.soil0])
    {
	feat_4->micro.soil0 = 
	  convert_soil_5_4(ctdb_4, soil_map_5[feat_4->micro.soil0]);
    }

    if (soil_map_4[feat_4->micro.soil1] != soil_map_5[feat_4->micro.soil1])
    {
	feat_4->micro.soil1 = 
	  convert_soil_5_4(ctdb_4, soil_map_5[feat_4->micro.soil1]);
    }

    return;    
}

static void
convert_volume_5_4(ctdb_4, feat_5, feat_4, next_5, next_4)
    CTDB_4               *ctdb_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_4 **next_4;
{
    int32 verts = feat_5->info.vertices;
    
    *next_4 = feat_4 + 1 + 2*verts;
    *next_5 = feat_5 + 3 + 2*verts;

    feat_4->info.feature_type = feat_5->info.feature_class;
    feat_4->info.vertices = feat_5->info.vertices;
    feat_4->info.post_mask = feat_5->info.post_mask;
    
    feat_5++;
    feat_5++;
    feat_5++;
    
    feat_4++;
    /* Copy the vertices, and we're done */
    bcopy(feat_5, feat_4, 2 * verts * sizeof(CTDB_FEATURE_DATA_4));
    return;    
}

static void
convert_linear_5_4(ctdb_5, ctdb_4, feat_5, feat_4, next_5, next_4)
    CTDB_5               *ctdb_5;
    CTDB_4               *ctdb_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_4 **next_4;
{
    int32 verts = feat_5->info.vertices;
    CTDB_LINEAR_MODEL_5 *model;
    
    *next_4 = feat_4 + 2 + 2*verts;
    *next_5 = feat_5 + 2 + 2*verts;

    feat_4->info.feature_type = feat_5->info.feature_class;
    feat_4->info.vertices = feat_5->info.vertices;
    feat_4->info.post_mask = feat_5->info.post_mask;
    
    feat_4++;
    feat_5++;
    
    model = ctdb_5->linear_models + feat_5->linear.reference;

    feat_4->tree.first = feat_5->linear.first;
    feat_4->tree.last = feat_5->linear.last;
    feat_4->tree.radius = model->width / 2.0;
    feat_4->tree.height = model->linear_data.treeline_info.foliage_height;
    
    feat_4++;
    feat_5++;
    
    /* Copy over the vertices, and we're done */
    bcopy(feat_5, feat_4, 2 * verts * sizeof(CTDB_FEATURE_DATA_4));

    return;
}

static void
convert_canopy_5_4(ctdb_4, feat_5, feat_4, next_5, next_4)
    CTDB_4               *ctdb_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_4 **next_4;
{
    /* No change from format 4 to format 5, just copy it over */

    *next_4 = feat_4 + 4;
    *next_5 = feat_5 + 4;

    bcopy(feat_5, feat_4, 4 * sizeof(CTDB_FEATURE_DATA_5));
    return;    
}

static void
convert_laid_linear_5_4(ctdb_5, ctdb_4, feat_5, feat_4, next_5, next_4, 
			patch_number)
    CTDB_5               *ctdb_5;
    CTDB_4               *ctdb_4;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_4  *feat_4;
    CTDB_FEATURE_DATA_5 **next_5;
    CTDB_FEATURE_DATA_4 **next_4;
    int32                 patch_number;
{
    int32 verts = feat_5->info.vertices;
    int32 *soil_map_4;
    int32 *soil_map_5;

    /* There was no change from format 4 to format 5, but we need to make
     * sure we don't have a soil type problem...
     */

    *next_4 = feat_4 + 2 + verts;
    *next_5 = feat_5 + 2 + verts;

    bcopy(feat_5, feat_4, (verts + 2) * sizeof(CTDB_FEATURE_DATA_4));

    feat_4++;
    
    soil_map_4 = ctdb_4->soil_map;
    soil_map_5 = get_soil_map_5(ctdb_5, patch_number);

    if (soil_map_4[feat_4->linear.soil] != soil_map_5[feat_4->linear.soil])
    {
	feat_4->linear.soil = 
	  convert_soil_5_4(ctdb_4, soil_map_5[feat_4->linear.soil]);
    }
    
    return;
}

/* Converts features for a patch from format4 to format5.  feat4 must point
 * to the start of the features in the source patch, and feat5 must point
 * to where the features should start in the destination patch.  Returns
 * number of bytes written into dest. patch.
 */
static int32 
convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, patch_deltas, patch_number)
    CTDB_5              *ctdb_5;
    CTDB_4              *ctdb_4;
    CTDB_FEATURE_DATA_5 *feat_5;
    CTDB_FEATURE_DATA_4 *feat_4;
    int32               *patch_deltas;
    int32                patch_number;
{
    CTDB_FEATURE_DATA_5 *last, *next_5;
    CTDB_FEATURE_DATA_4 *cur_4, *next_4;
    CTDB_FEAT_OBJ_PTR obj;
    int32 fclass, verts;
    int32 size_5;

    size_5 = feat_5->hdr.size;
    last = feat_5 + size_5;
    cur_4 = feat_4;
    feat_5++;
    cur_4++;
    
    for(;feat_5<last;(feat_5 = next_5), (cur_4 = next_4))
    {
	fclass = feat_5->info.feature_class;

	switch(fclass)
	{
	  case FC_MICRO_5:
	    convert_micro_5_4(ctdb_5, ctdb_4, feat_5, cur_4, &next_5, &next_4,
			      patch_number);
	    break;
	  case FC_VOLUME_5:
	    convert_volume_5_4(ctdb_4, feat_5, cur_4, &next_5, &next_4);
	    break;
	  case FC_LINEAR_5:
	    convert_linear_5_4(ctdb_5, ctdb_4, feat_5, cur_4, &next_5, 
			       &next_4);
	    break;
	  case FC_CANOPY_5:
	    convert_canopy_5_4(ctdb_4, feat_5, cur_4, &next_5, &next_4);
	    break;
	  case FC_LAID_LINEAR_5:
	    convert_laid_linear_5_4(ctdb_5, ctdb_4, feat_5, cur_4, &next_5, 
				    &next_4, patch_number);
	    break;
	  default:
	    printf("ERROR: unknown feature class encountered in format 5 db.\n");
	    exit(-1);
	}
    }

    feat_4->hdr.size = next_4 - feat_4;
    patch_deltas[patch_number] = feat_4->hdr.size - size_5;

    return (feat_4->hdr.size * sizeof(CTDB_FEATURE_DATA_4));
}

/* Convert all data, including data header, for a gridded PG from format 5 to
 * format 4.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_grid_pg_data_5_4(ctdb_5, ctdb_4, group, base, patch_deltas)
    CTDB_5 *ctdb_5;
    CTDB_4 *ctdb_4;
    int32   group;
    ADDRESS base;
    int32  *patch_deltas;
{
    int32 i;
    int32 size;
    CTDB_GRID_PG_DATA_5 *data_5;
    CTDB_GRID_PG_DATA_4 *data_4;
    CTDB_PATCH_GROUP_4 *pg;
    CTDB_FEATURE_DATA_5 *feat_5;
    CTDB_FEATURE_DATA_4 *feat_4;
    int32 patch_bytes = 0;
    int32 patch_number;

    data_5 = ctdb_5->patch_groups[group].data_start.grid_data;
    data_4 = (CTDB_GRID_PG_DATA_4 *)base;
    size = sizeof(CTDB_GRID_PG_DATA_4);
    patch_number = group * CTDB_PATCHES_PER_GROUP_4;
    
    pg = ctdb_4->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_4, size);
    bcopy(data_5->max_elev, data_4->max_elev, 
	  CTDB_PATCHES_PER_GROUP_4 * sizeof(float32));
    
    /* Can't fill in offsets yet... */

    /* 2.  Convert the features for each patch */
    feat_4 = (CTDB_FEATURE_DATA_4 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)data_5 +
					 sizeof(CTDB_GRID_PG_DATA_5));
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));

	data_4->patch1_offset = (patch_bytes + sizeof(CTDB_GRID_PG_DATA_4)) /
	  sizeof(CTDB_FEATURE_DATA_4);
    }
    else
    {
	data_4->patch1_offset = sizeof(CTDB_GRID_PG_DATA_4) /
	  sizeof(CTDB_FEATURE_DATA_4);
    }

    patch_number++;
    if (pg->patch1_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch1_offset);
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);

	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));

	data_4->patch2_offset = data_4->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));
    }
    else
    {
	data_4->patch2_offset = data_4->patch1_offset;
    }

    patch_number++;
    if (pg->patch2_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch2_offset);
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));

	data_4->patch3_offset = data_4->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));
    }
    else
    {
	data_4->patch3_offset = data_4->patch2_offset;
    }

    patch_number++;
    if (pg->patch3_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch3_offset);
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));
    }	  

    size = (int8 *)feat_4 - (int8 *)data_4;
    return size;
}


/* Convert all data, including data header, for a tinned PG from format 5 to
 * format 4.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_tin_pg_data_5_4(ctdb_5, ctdb_4, group, base, patch_deltas)
    CTDB_5 *ctdb_5;
    CTDB_4 *ctdb_4;
    int32   group;
    ADDRESS base;
    int32  *patch_deltas;
{
    int32 i;
    int32 size;
    CTDB_TIN_PG_DATA_5 *data_5;
    CTDB_TIN_PG_DATA_4 *data_4;
    CTDB_PATCH_GROUP_4 *pg;
    CTDB_FEATURE_DATA_5 *feat_5;
    CTDB_FEATURE_DATA_4 *feat_4;
    int32 patch_bytes = 0;
    int32 delta;
    int32 te_offset_5;
    int32 te_offset_4;
    int32 patch_number;

    data_5 = ctdb_5->patch_groups[group].data_start.tin_data;
    data_4 = (CTDB_TIN_PG_DATA_4 *)base;
    size = sizeof(CTDB_TIN_PG_DATA_4);
    patch_number = group * CTDB_PATCHES_PER_GROUP_4;

    pg = ctdb_4->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_4, size);
    bcopy(data_5->max_elev, data_4->max_elev, 
	  CTDB_PATCHES_PER_GROUP_4 * sizeof(float32));
    
    /* Can't fill in offsets yet... */

    /* The feature present field grew in format 5, but this copy is still
     * correct, as long as we have a database that can be converted for fmt. 4
     */
    bcopy(data_5->feature_present, data_4->feature_present,
	  CTDB_MAX_POSTS_PER_PATCH_4 * 2);
    
    bcopy(data_5->num_te, data_4->num_te, 
	  CTDB_PATCHES_PER_GROUP_4 * sizeof(uint16));
    
    /* Can't fill in te_offsets yet */

    bcopy(data_5->te_vgrid_dim, data_4->te_vgrid_dim,
	  CTDB_PATCHES_PER_GROUP_4 * sizeof(uint8));
    
    /* Can't fill in vgrid_offsets yet */

    /* 2.  Convert the features for each patch */
    feat_4 = (CTDB_FEATURE_DATA_4 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)data_5 +
					 sizeof(CTDB_TIN_PG_DATA_5));
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));

	data_4->patch1_offset = (patch_bytes + sizeof(CTDB_TIN_PG_DATA_4)) /
	  sizeof(CTDB_FEATURE_DATA_4);
    }
    else
    {
	data_4->patch1_offset = sizeof(CTDB_TIN_PG_DATA_4) /
	  sizeof(CTDB_FEATURE_DATA_4);
    }

    patch_number++;
    if (pg->patch1_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch1_offset);
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));

	data_4->patch2_offset = data_4->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));
    }
    else
    {
	data_4->patch2_offset = data_4->patch1_offset;
    }

    patch_number++;
    if (pg->patch2_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch2_offset);
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));

	data_4->patch3_offset = data_4->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));
    }
    else
    {
	data_4->patch3_offset = data_4->patch2_offset;
    }

    patch_number++;
    if (pg->patch3_occupied)
    {
	feat_5 = (CTDB_FEATURE_DATA_5 *)((CTDB_FEATURE_DATA_5 *)data_5 +
					 data_5->patch3_offset);
	
	patch_bytes = convert_feat_5_4(ctdb_5, ctdb_4, feat_5, feat_4, 
				       patch_deltas, patch_number);
	feat_4 = feat_4 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_4));
    }	  

    /* 3. Calculate how far the TE's for the patch have moved relative to
     * where they used to be
     */
    if (!pg->patch0_no_tes)
    {
	te_offset_5 = data_5->te_offset[0];
    }
    else if (!pg->patch1_no_tes)
    {
	te_offset_5 = data_5->te_offset[1];
    }
    else if (!pg->patch2_no_tes)
    {
	te_offset_5 = data_5->te_offset[2];
    }
    else
    {
	te_offset_5 = data_5->te_offset[3];
    }
    
    te_offset_4 = feat_4 - (CTDB_FEATURE_DATA_4 *)data_4;
    
    /* Now, calculate the offset, in words */
    delta = te_offset_4 - te_offset_5;
    
#ifdef DEBUG
    if (delta >= 0)
      printf("Unreasonable data offset in 5->4 conversion.\n");
#endif /* DEBUG */
    
    /* 4.  Fill out the rest of the offset values. */
    for(i=0;i<CTDB_PATCHES_PER_GROUP_4;i++)
    {
	data_4->te_offset[i] = data_5->te_offset[i] + delta;
	data_4->te_vgrid_offset[i] = data_5->te_vgrid_offset[i] + delta;
    }
    
    /* 5.  Copy over all the TE data. */
    {
	int32 data_size;
	int32 te_data_size;	/* In bytes */

	if (group < ctdb_5->max_patch_number)
	{
	    
	    /* Total bytes of data stored for this PG */
	    data_size = 
	      (int8 *)ctdb_5->patch_groups[group+1].data_start.grid_data -
	      (int8 *)data_5;
	    
	    te_data_size = data_size - 
	      (data_5->te_offset[0] * sizeof(CTDB_FEATURE_DATA_5));
	}
	else
	{
	    int32 vgrid_size;
	    
	    if (!pg->patch3_no_tes)
	    {
		data_size = data_5->te_vgrid_offset[3] * 
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[3] * data_5->te_vgrid_dim[3];
	    }
	    else if (!pg->patch2_no_tes)
	    {
		data_size = data_5->te_vgrid_offset[2] * 
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[2] * data_5->te_vgrid_dim[2];
	    }
	    else if (!pg->patch1_no_tes)
	    {
		data_size = data_5->te_vgrid_offset[1] * 
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[1] * data_5->te_vgrid_dim[1];
	    }
	    else
	    {
		data_size = data_5->te_vgrid_offset[0] * 
		  sizeof(CTDB_FEATURE_DATA_5);
		vgrid_size = data_5->te_vgrid_dim[0] * data_5->te_vgrid_dim[0];
	    }

	    /* Add padding as necessary */
	    vgrid_size += (vgrid_size % sizeof(CTDB_FEATURE_DATA_5));
	    data_size += vgrid_size;		
	    te_data_size = (data_size -
			    (te_offset_5 * sizeof(CTDB_FEATURE_DATA_5)));
	}

	feat_5 = (CTDB_FEATURE_DATA_5 *)data_5 + te_offset_5;
	bcopy(feat_5, feat_4, te_data_size);
	
	/* Calculate total bytes in this PG */
	size = ((int8 *)feat_4 - (int8 *)data_4) + te_data_size;
    }

    return size;
}



/* Convert patch group header from fmt5 to 4.  base is where the new PG
 * should point.
 */
static void
convert_pg_5_4(ctdb_5, ctdb_4, group, base)
    CTDB_5 *ctdb_5;
    CTDB_4 *ctdb_4;
    int32   group;
    ADDRESS base;
{
    CTDB_PATCH_GROUP_4 *pg_4;
    CTDB_PATCH_GROUP_5 *pg_5;

    pg_4 = ctdb_4->patch_groups + group;
    pg_5 = ctdb_5->patch_groups + group;
    
    bzero(pg_4, sizeof(CTDB_PATCH_GROUP_5));

    pg_4->data_start.grid_data = (CTDB_GRID_PG_DATA_4 *)base;
    
    /* We'll leave the cache stuff blank for now
     */

    pg_4->contains_tins = pg_5->contains_tins;

    pg_4->patch0_occupied = pg_5->patch0_occupied;
    pg_4->patch0_dyn_terr = pg_5->patch0_dyn_terr;
    pg_4->patch0_te16 = pg_5->patch0_te16;
    pg_4->patch0_no_tes = pg_5->patch0_no_tes;

    pg_4->patch1_occupied = pg_5->patch1_occupied;
    pg_4->patch1_dyn_terr = pg_5->patch1_dyn_terr;
    pg_4->patch1_te16 = pg_5->patch1_te16;
    pg_4->patch1_no_tes = pg_5->patch1_no_tes;

    pg_4->patch2_occupied = pg_5->patch2_occupied;
    pg_4->patch2_dyn_terr = pg_5->patch2_dyn_terr;
    pg_4->patch2_te16 = pg_5->patch2_te16;
    pg_4->patch2_no_tes = pg_5->patch2_no_tes;

    pg_4->patch3_occupied = pg_5->patch3_occupied;
    pg_4->patch3_dyn_terr = pg_5->patch3_dyn_terr;
    pg_4->patch3_te16 = pg_5->patch3_te16;
    pg_4->patch3_no_tes = pg_5->patch3_no_tes;

    return;
}

/* Converts PG header and all PG data from format 5 to format 4.  group is
 * the patch group number to convert, base is an in-out variable passing
 * in the address where the format 4 data should start, and passing out
 * where the format 4 data for the next PG should start.
 */
static int32
convert_group_5_4(ctdb_5, ctdb_4, group, base, patch_deltas)
    CTDB_5  *ctdb_5;
    CTDB_4  *ctdb_4;
    int32    group;
    ADDRESS *base;
    int32   *patch_deltas;
{
    int32 bytes;
    
    /* First, convert the header. */
    convert_pg_5_4(ctdb_5, ctdb_4, group, *base);

    /* Now convert the data and the data headers */
    if (ctdb_4->patch_groups[group].contains_tins)
    {
	bytes = convert_tin_pg_data_5_4(ctdb_5, ctdb_4, group, *base,
					patch_deltas);
    }
    else
    {
	bytes = convert_grid_pg_data_5_4(ctdb_5, ctdb_4, group, *base,
					 patch_deltas);
    }
    
    *base = (int8 *)*base + bytes;
    return bytes;
}



/* Converts PGs and features (including TINs) from format 5 to format 4.
 * Returns total patch data bytes in format 5 result.
 */
static int32
convert_features_5_4(ctdb_5, ctdb_4, patch_deltas)
    CTDB_5 *ctdb_5;
    CTDB_4 *ctdb_4;    
    int32  *patch_deltas;
{
    int32 groups;
    int32 i;
    int32 total_bytes = 0;
    ADDRESS base;
	
    groups = ctdb_4->patches_wide * ctdb_4->patches_high /
      CTDB_PATCHES_PER_GROUP_4;
	
    /* Going from 5 to 4, features can only shrink */
    ctdb_4->n_patch_bytes = ctdb_5->n_patch_bytes;
	
    ctdb_4->patch_data = (ADDRESS)ctdb_alloc(NULL,
					     ctdb_4->n_patch_bytes,
					     "Fmt. 4 patch data");

    base = ctdb_4->patch_data;

    for(i=0;i<groups;i++)
    {
	total_bytes += convert_group_5_4(ctdb_5, ctdb_4, i, &base,
					 patch_deltas);
    }

    return total_bytes;    
}


static int32
get_post_index_4(ctdb, norm_x, norm_y)
    CTDB_4 *ctdb;
    int32   norm_x;
    int32   norm_y;
{
    int32 page, index;
    
    page = (norm_x>>SIDE_SHIFT) +  
      ctdb->pages_wide * (norm_y>>SIDE_SHIFT);
    index = (norm_x&SIDE_MASK) + POSTS_PER_SIDE * 
      (norm_y&SIDE_MASK);
    return (page*POSTS_PER_PAGE + index);
}

static void
convert_post_soils_5_4(ctdb_5, ctdb_4)
    CTDB_5 *ctdb_5;
    CTDB_4 *ctdb_4;
{
    int32 patch;
    int32 px;
    int32 py;
    int32 *soil_map_5;
    int32 base_px;
    int32 base_py;
    int32 *post;
    int32 index;
    int32 soil1, soil2;

    for(patch=0;patch<ctdb_4->max_patch_number;patch++)
    {
	soil_map_5 = get_soil_map_5(ctdb_5, patch);
	if (soil_map_5 == ctdb_5->soil_tables[0])
	  /* This is our default map anyway, so skip this patch */
	  continue;
	
	base_py = (patch / ctdb_4->patches_wide) * ctdb_4->patch_incr;
	base_px = (patch % ctdb_4->patches_wide) * ctdb_4->patch_incr;

	for(py=0;py<ctdb_4->patch_incr;py++)
	  for(px=0;px<ctdb_4->patch_incr;px++)
	  {
	      index = get_post_index_4(ctdb_4, base_px + px, base_py + py);
	      post = ctdb_4->elevations + index;
	      soil1 = GET_POST_SW_SOIL_4(*post);
	      soil2 = GET_POST_NE_SOIL_4(*post);
	      
	      /* NOTE:  it really doesn't matter here which way the
	       * posts are diagonalized, so we'll ignore it.
	       */
	      if (ctdb_4->soil_map[soil1] != soil_map_5[soil1])
	      {
		  soil1 = convert_soil_5_4(ctdb_4, soil_map_5[soil1]);
		  /* Clear out the old soil */
		  (*post) &= ~BUILD_POST_SW_SOIL_4(15);
		  /* Put in the new one */
		  (*post) |= BUILD_POST_SW_SOIL_4(soil1);
	      }
	      if (ctdb_4->soil_map[soil2] != soil_map_5[soil2])
	      {
		  soil2 = convert_soil_5_4(ctdb_4, soil_map_5[soil2]);
		  /* Clear out the old soil */
		  (*post) &= ~BUILD_POST_NE_SOIL_4(15);
		  /* Put in the new one */
		  (*post) |= BUILD_POST_NE_SOIL_4(soil2);
	      }
	  }
    }
    return;    
}

/* Update feature references in the topology for the new format 4 database. */
static void
fix_topo_refs_5_4(ctdb_5, ctdb_4, patch_deltas)
    CTDB_5 *ctdb_5;
    CTDB_4 *ctdb_4;
    int32  *patch_deltas;
{
    CTDB_EDGE_4 *edge;
    int32 offset = 0;
    int32 i;
    int32 j;
    
    for(i=offset=0;(i<ctdb_5->num_edges) && (offset<ctdb_5->edge_size);i++)
    {
	edge = (CTDB_EDGE_4 *)(ctdb_4->edge_storage + offset);
	offset += sizeof(CTDB_EDGE_4) + sizeof(struct ctdb_connection_4) *
	  (edge->num_features - 1);

	for(j=0;j<edge->num_features;j++)
	{
	    if (!CTDB_IS_QUAD_NODE(edge->feature[j].patch_number))
	    {
		edge->feature[j].patch_offset +=
		  patch_deltas[edge->feature[j].patch_number];
	    }
	}
    }
    
    return;
}

