/*
 * $RCSfile: ct_cmp_fmt5.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt5.h                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 4 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/
#ifndef _CMP_FMT5_H_INCLUDED
#define _CMP_FMT5_H_INCLUDED

#define CTDB_PATCHES_PER_GROUP_5       4  /* Num of patches per patch group */

#define CTDB_MAX_FEATURE_PRESENT_PER_PG_5 60 

#define CTDB_MAX_POSTS_PER_PATCH_5    25   /* maximum number of posts per
					    * patch
					    */

#define CTDB_NAME_LENGTH_5            14   /* same as Simnet */
#define CTDB_DATE_LENGTH_5            28   /* Enough room for ctime() */

#define CTDB_PATCH_CACHES_5           16   /* Number of patch data caches */
#define CTDB_CACHE_BKT_UNITS_5         4

#define CTDB_ILLEGAL_SOIL_5 0;

/* Standard basis gives -5000.0m thru +5000.0m with cm accuracy. */
#define CTDB_STANDARD_FIXED_POINT_BASIS_5 (5000.0 / (1 << 19))

/* Elevation is bits 0-19 (sign 19, magnitude 0-18) */
#define GET_POST_ELEVATION_5(p,basis)                           \
  ((((p) & 0x00080000) ? -((p) & 0x0007FFFF) : ((p) & 0x0007FFFF)) * (basis))

#define BUILD_POST_ELEVATION_5(e,inv_basis)                     \
  (((e) >= 0) ?                                                 \
   (((int32)((e) * (inv_basis))) & 0x0007FFFF) :                \
   ((((int32)((-e) * (inv_basis))) & 0x0007FFFF) | 0x00080000))

/* The feature classes each imply certain intervisibility semantics
 * and storage techniques:
 *
 * FC_LINEAR is an individual tree or a treeline (depending upon the
 * number of vertices), or any non-laid linear feature (fences, wire,
 * dragon's teeth).
 *
 * An individual tree is described
 *	FEAT_INFO LINEAR_INFO XY Z
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature; XY identifies the location of the tree
 * within the patch; Z identifies the elevation of the top of the
 * tree.  The reduction of visibility due to tree blockage is done
 * using a simple light transmittance model.
 *
 * A treeline (or linear feature) is described
 *	FEAT_INFO LINEAR_INFO XY Z XY Z XY Z...
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature and whether or not there are trees at each end
 * of a treeline; (note that we do not need to store the end trees as
 * individuals since they are implied by the treeline); each XY point
 * identifies the location of one vertex in the treeline; the corresponding Z
 * identifies the elevation of the top of the treeline at that vertex.
 *
 * FC_VOLUME is a building or a building-like structure which blocks
 * intervisibility.  Telephone poles, water towers, and other such
 * structures which do not interrupt intervisibility are omitted.
 *
 * A volume is described
 *	FEAT_INFO VOLUME_INFO XY Z XY Z ...
 * Each XY point identifies the location of one vertex in the bounding
 * box of the building; the corresponding Z identifies the elevation
 * of the roof line of the building at that point.  Intervisibility is
 * interrupted by the line segments connecting these points exactly the
 * same way as the segments connecting grid posts.  The VOLUME_INFO
 * contains the type of feature, a reference to a model set, and the damage
 * state.
 *
 * FC_CANOPY is a canopy polygon.  See-through canopies are omitted.
 * FC_MICRO is a microterrain triangle.
 *
 * Canopies and microterrain are described the same way,
 *	INFO XY Z MICRO
 * Microwater is identical to microterrain, except that it has an MWATER_CHARS
 * at the end:
 *      INFO XY Z MICRO MWATER_CHARS
 * The XY and Z combination identify the location of one vertex of
 * the microterrain triangle.  Up to three subsequent vertices are listed
 * in the MICRO structure.  There are edges connecting this point to
 * each of the other listed vertices.  If there are two vertices listed
 * the two edges from this point form a triangle.  If there are three
 * listed, the three edges form two triangles (the first edge being
 * shared between the two).  For microterrain, intervisibility is
 * interrupted by the edges connecting these points exactly the same
 * way as the segments connecting grid posts.  For canopies, visibility
 * is degraded as with treelines.
 *
 * FC_LAID_LINEAR is a road or river segment.
 *
 * Road segments and river segments are described the same way
 *	INFO WDSOIL XY XY...
 * WDSOIL identifies the width at the beginning and end of the segment
 * (linear taper is assumed), and the soil type of the segment; each
 * XY point identifies the location of one vertex in the segment.  Linear
 * features do not impact intervisibility.
 */

#define FC_CLASS_MIN_5     0
#define FC_MICRO_5         0
#define FC_VOLUME_5        1
#define FC_LINEAR_5        2
#define FC_CANOPY_5        3
#define FC_LAID_LINEAR_5   4
#define FC_CLASS_MAX_5     5

#define M_FEATURE_CLASS_5   0xE0000000
#define S_FEATURE_CLASS_5   29
#define M_VERTICES_5        0x1E000000
#define M_MICRO_VERTICES_5  0x06000000
#define S_VERTICES_5        25

#define MICRO_VERTS_5         0x3
#define MICRO_TYPE_5          0xC
#define MICRO_TYPE_BASE_5     0x0
#define MICRO_TYPE_ML_5       0x4
#define MICRO_TYPE_DEFAULT_5  0x8
#define MICRO_TYPE_WATER_5    0xC

typedef struct ctdb_feature_info_5
{
    /* One of the FC_x values listed above.
     */
    unsigned feature_class : 3;

    /* For FC_LINEAR, FC_VOLUME, FC_CANOPY, and FC_LAID_LINEAR this
     * is the number of vertices.  If the actual objects has more than
     * 16 vertices, it must be broken into two or more objects.  For
     * FC_MICRO this is the number of positions which are connected to
     * this point bitwise-OR'd with a microterrain type.
     */
    unsigned vertices : 4;

    /* This is used as a high resolution bitwise bounding volume test.
     * Each post in a patch has a bit.  This mask identifies which
     * posts' polygons were touched by this feature.  Although Simnet
     * terrain databases only have 16 posts per patch, CCTT database
     * may have 25 (given the 100m interval required in the spec).
     */
    unsigned post_mask : 25;
} CTDB_FEATURE_INFO_5;

#define CTDB_TREE_HEIGHT_BASIS_5 (200.0 / (1<<14))

/* Number of bits used to define fullness.
 */
#define CTDB_FULLNESS_BITS_5  8

#define M_TREE_FIRST_5 0x8000
#define M_TREE_LAST_5  0x4000

typedef struct ctdb_volume_info_5
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.
     */
    unsigned reference    : 16;
    unsigned type         : 5;
    unsigned avoid        : 1;
    unsigned reserved     : 10;
} CTDB_VOLUME_INFO_5;


typedef struct ctdb_linear_info_5
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.  `first' and `last' are only used if
     * the feature is a treeline, and specify whether or not individual
     * trees should be placed at each end.
     */
    unsigned reference        : 16;
    unsigned first            : 1;
    unsigned last             : 1;
    unsigned reserved         : 14;
} CTDB_LINEAR_INFO_5;


/* For now we store feature elevations as simple floats, since there
 * is no obvious way to use the extra 13 bits if we were to instead
 * store it using the fixed point convention used for grid posts.
 * Floating point will, of course, be faster than storing them in
 * fixed point since only a SP->DP conversion will be necessary to use
 * the data.
 */

#define M_VERT0_OUTER_5  0x01000000
#define M_VERT1_OUTER_5  0x02000000
#define M_VERT2_OUTER_5  0x04000000
#define M_CAN_EDGE_REF_5 0xF8000000
#define S_CAN_EDGE_REF_5 27

typedef struct ctdb_micro_pts_5
{
    /* Each microterrain vertex is connected to 0,1,2, or 3 other
     * microterrain vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then the soil
     * fields will be 0; vert0 will contain the index of that vertex
     * relative to patch start; vert1, vert2 will be 0.  If it
     * connected to 2 others, soil0 will contain the soil type of the
     * triangle connecting this vertex to vert0 to vert1; soil1 and
     * vert2 will be 0.  If it is connected to 3 others, soil0 will
     * contain the soil type of the triangle connecting this vertex to
     * vert0 to vert1; soil1 will contain the soil type of the
     * triangle connecting this vertex to vert0 to vert2.
     * The same rules apply when this structure is used to describe
     * canopy polygons, except the soil fields are used by bits in a
     * mask identifying whether the edge to each vertex is an outer
     * edge of the canopy.
     */
    unsigned soil0 : 4;
    unsigned soil1 : 4;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
} CTDB_MICRO_PTS_5;

/* Microwater characteristic references */
typedef struct ctdb_mwater_chars_5
{
    unsigned char0 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
    unsigned char1 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
} CTDB_MWATER_CHARS_5;

typedef struct ctdb_laid_linear_info_5
{
    /* Each linear segment has a width and a soil type.  The width is
     * in the same units as x,y in the XY_PATCH_OFFSET structure.
     * The edge_key provides the low-order 12 bits of the topologic
     * edge (of type CTDB_EDGE) which uses this linear feature.
     */
    uint16    width;
    unsigned  soil : 4;
#define CTDB_EDGE_KEY_5 4096 /* 2^12 */
    unsigned  edge_key : 12;
} CTDB_LAID_LINEAR_INFO_5;

typedef struct ctdb_patch_hdr_5
{
    /* At the beginning of each patch, we store the size (the number
     * of CTDB_FEATURE_DATA things in the patch).  There will never be
     * more than 2^16 of those, so that leaves 16 bits of padding.
     * I'm sure we'll find a use for those bits eventually.
     */
    uint16 size;
    uint16 _pad;
} CTDB_PATCH_HDR_5;

#define CTDB_ABSTRACT_CANOPY_5                1
#define CTDB_ABSTRACT_SOIL_DEFRAG_5           2
#define CTDB_ABSTRACT_STEEP_SLOPE_5           3
#define CTDB_ABSTRACT_RAILROAD_5              4
#define CTDB_ABSTRACT_PIPELINE_5              5
#define CTDB_ABSTRACT_POLITICAL_BOUNDARY_5    6
#define CTDB_ABSTRACT_POWERLINE_5             7
#define CTDB_ABSTRACT_LABEL_5                 8
#define CTDB_ABSTRACT_TACTICAL_SIGN_5         9
#define CTDB_ABSTRACT_OFFROAD_SEGMENT_5       10
#define CTDB_ABSTRACT_MAX_5 /* Always Last */ 11


/* CTDB_ABSTRACT_FEATURE_INFO:
 *
 * Structure at the start of each abstract feature.
 */
#define CTDB_ABSTRACT_QUAD_NODE_5             0
typedef struct ctdb_abstract_feature_info_5
{
    /* Type code for the abstract feature.  The code 0 is reserved
     * to indicate that this "feature" is actually a reference to
     * four child quad nodes.
     */
    uint8 type_code;

    /* The size (in 4-byte units) of the CTDB_ABSTRACT_DATA associated
     * with the abstract feature.
     */
    uint8 data_size;

    /* The number of vertices in feature.  There must be at least one,
     * except for "features" which are of the type CTDB_ABSTRACT_QUAD_NODE.
     */
    uint16 num_vertices;

} CTDB_ABSTRACT_FEATURE_INFO_5;

/* CTDB_XYO_COORD2D
 * 
 * Structure used to represent a vertex point (x,y) in XYO coordinates 
 * in either a feature or terrain element data structure.
 */
typedef struct ctdb_xyo_coord2d_5  
{
    uint16 x;
    uint16 y;
} CTDB_XYO_COORD2D_5;

/* CTDB_ABSTRACT_VERTEX:
 *
 * Structure used to represent a vertex of an abstract feature.
 * Note that for efficient access, the format used for vertices
 * is 32 bit floating point, not fix-point patch offsets.  This
 * results in a loss of precision at the far corners of the database,
 * but avoids a great deal of translation at run time.
 */
typedef float32 CTDB_ABSTRACT_VERTEX_5;

/* CTDB_ABSTRACT_DATA:
 *
 * Format of data associated with so-called abstract features.  Note
 * that this structure is analogous to a protocol definition, in that
 * the final record of each union substructure can be variable length.
 *
 * This goes into a 4-byte union: NO MEMBER STRUCTURE CAN EXCEED 32 BITS!
 *
 */
typedef union ctdb_abstract_data_5
{
    struct ctdb_quad_data_5
    {
	/* Array of 4 offsets to child quad nodes (in 4-byte units)
	 * from the beginning of the quad tree data storage.
	 */
	uint32 child_start[1];  /* really 4 members, not 1 */
    } quad_data;
    struct ctdb_canopy_data_5
    {
	unsigned impenetrable  : 1;
	unsigned padding       : 31;
    } canopy;
    struct ctdb_soil_defrag_data_5
    {
	unsigned soil_index    : 4;
	unsigned level         : 4;
	unsigned padding       : 8;
	unsigned soilmap_index : 16;
    } soil_defrag;
    struct ctdb_steep_slope_data_5
    {
	unsigned slope_degrees : 7;
	unsigned padding       : 25;
    } steep_slope;
    struct ctdb_label_data_5
    {
	char text[4]; /* variable length, NULL terminated */
    } label;
    struct ctdb_offroad_data_5
    {
	uint32 edge_number;
    } offroad;

} CTDB_ABSTRACT_DATA_5;


typedef union ctdb_feature_data_5
{
    uint32                       integer;
    CTDB_PATCH_HDR_5             hdr;
    CTDB_FEATURE_INFO_5          info;
    CTDB_VOLUME_INFO_5           volume;
    CTDB_LINEAR_INFO_5           linear;
    CTDB_LAID_LINEAR_INFO_5      laid_linear;
    CTDB_XYO_COORD2D_5           xy;
    float32                      z;
    CTDB_MICRO_PTS_5             micro;
    CTDB_ABSTRACT_FEATURE_INFO_5 abstract;
    CTDB_ABSTRACT_DATA_5         abstract_data;
    CTDB_ABSTRACT_VERTEX_5       abstract_vertex;
    CTDB_MWATER_CHARS_5          mwater;
} CTDB_FEATURE_DATA_5;

/* Definitions of feature types.
 */

/* Types of volume features.
 */
#define CTDB_FT_BUILDING_5          0
#define CTDB_FT_ROCK_DROP_5         1

/* Types of linear features.
 */
#define CTDB_FT_TREELINE_5          0
#define CTDB_FT_CONCERTINA_WIRE_5   1
#define CTDB_FT_FENCE_5             2
#define CTDB_FT_MINEFIELD_FENCE_5   3
#define CTDB_FT_DRAGONS_TEETH_5     4

/* Types of dragon's tooth components.
 */
#define CTDB_DT_STEEL_HEDGEHOG_5       0
#define CTDB_DT_STEEL_TETRAHEDRON_5    1
#define CTDB_DT_CONCRETE_BLOCK_5       2
#define CTDB_DT_CONCRETE_TETRAHEDRON_5 3

/* CTDB_EDGE:
 *
 * Structure used to collect a list of linear features which form an
 * uninterrupted "edge" between two intersections.  The `feature'
 * member is actually variable length.
 */
typedef struct ctdb_edge_5
{
    /* Total distance along the segment (useful for planning) */
    int32 distance;

    /* Indices of the nodes at the beginning and end of the segment */
    int32 first_node;
    int32 last_node;

    int32 num_features;
    struct ctdb_feature_reference_5
    {
	/* The patch where this feature is stored */
	uint32 patch_number;

	/* Byte offset into the patch */
	unsigned patch_offset : 24;

	/* Whether the vertices in the patch are in reverse order */
	unsigned reverse      : 1;

	unsigned padding      : 7;
    } feature[1];
} CTDB_EDGE_5;

/* CTDB_NODE:
 *
 * Represents the topologic connection between CTDB_EDGE structures.
 * The `connection' member is actually variable length.
 */
typedef struct ctdb_node_5
{
    /* Approximate location of the node (useful for planning) */
    float32 x, y;

    /* Whether the node corresponds to a bridge */
    unsigned is_bridge : 1;

    unsigned padding   : 23;

    uint8 num_connections;
    struct ctdb_connection_5
    {
	unsigned node_in_other_cell : 1;  /* TRUE indicates that the node
					   * index references a node in 
					   * another cell, FALSE indicates
					   * that it is in this cell.
					   */
	unsigned padding            : 7;
	
	/* The node which this connects to */
	unsigned node_index         : 24;
	
	union edge_union_5
	{
	    int32 edge_index; /* The edge used to connect these nodes,
			       * if they are in the same database. 
			       */

	    int32 other_cell; /* If node_in_other_cell is TRUE,
			       * the cell in which node_index is valid.
			       */
	} edge;
    } connection[1];

} CTDB_NODE_5;

/* CTDB_LINEAR_DESC:
 *
 * This structure is used to return information about linear models to the
 * application.
 *
 * It is also used to pass a linear feature or an individual tree
 * (n_verts == 1) to the feature compression algorithms.
 *
 * `trunk_radius' specifies the radius of the tree trunk.
 *
 * `foliage_height' specifies the height of the foliage of the tree.
 *
 * `fullness' specifies the fullness of the leaves or wire.  0.0 represents
 * zero fullness (nothing there), and 1.0 represents maximum fullness
 * (completely blocks light).
 *
 * `spacing' represents how far apart from each other the dragon's teeth are
 * placed along the linear segment.
 *
 * `component' specifies the type of dragon's tooth, specified in libctdb.h
 *
 * `type' specifies the type of linear feature, specified in libctdb.h
 *
 * `width' specifies the foliage width for trees, or the width of dragon's
 * teeth.
 */
typedef struct ctdb_linear_desc_5
{
    union
    {
	struct
	{
            float32  trunk_radius;
            float32  foliage_height;
            float32  fullness;
	} treeline_info;

	struct
	{
            float32  fullness;
            int32    pad1;
            int32    pad2;
	} var_opacity_info;

	struct
	{
            float32  spacing;
            int32    component;
            int32    pad1;
	} dragons_tooth_info;
    } linear_data;

    int32   type;
    float32 width;

} CTDB_LINEAR_DESC_5;


typedef struct ctdb_linear_model_5
{
    /* This defines the models for the various linear features.
     */
    union
    {
	struct
	{
	    /* `treeline_info' is used for trees and treelines.
	     * 
	     * `trunk_radius' is measured in the same units as x,y in the
	     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
	     *
	     *  `height' is the distance from the top of the tree to the
	     * lowest branches measured in the unit `CTDB_TREE_HEIGHT_BASIS':
	     * 
	     * #define CTDB_TREE_HEIGHT_BASIS (200.0 / (1<<14))
	     *
	     * `fullness' refers to the "density" of the leaves, wire, or
	     * whatever the feature is made of.  This value ranges from zero
	     * (nothing there) to the maximum value (lets no light through).
	     */
	    unsigned trunk_radius     : 16;
	    unsigned foliage_height   : 14;
	    unsigned pad1             : 2;
	    unsigned fullness         : 8;
	    unsigned pad2             : 24;
	} treeline_info;

	/* `var_opacity_info' is used for wire and fences.
	 */
	struct
	{
	    unsigned fullness         : 8;
	    unsigned pad1             : 24;
	    unsigned pad2             : 32;
	} var_opacity_info;

	/* `dragons_tooth_info' is used for dragon's teeth.
	 * 
	 * `spacing' represents how far apart from each other the dragon's
	 * teeth are placed along the linear segment.
	 *
	 * Dragon's teeth can be made of steel hedgehogs, steel tetrahedrons,
	 * concrete blocks, or concrete tetrahedrons.  Each of these affects
	 * intervisibility differently (hence the need for the `component'
	 * element).
	 */
	struct
	{
	    float32  spacing;
	    unsigned component        : 4;
	    unsigned pad              : 28;
	} dragons_tooth_info;
    } linear_data;

    /* `type' specifies the specific type of linear feature (i.e. treeline,
     * minefield fence, etc.).
     * 
     * The width of the foliage, the thickness of the concertina wire or
     * minefield fence, or the width of the dragon's teeth is stored in
     * `width', which is measured in the same units as X and Y in the
     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
     */
    uint8  type;
    uint8  pad;
    uint16 width;

} CTDB_LINEAR_MODEL_5;


typedef struct ctdb_volume_model_5
{
    char     model_name[32];
    uint32   model_id;
    uint32   next_state_index;
    float32  percent_damaged;
    uint16   num_vertices;
    uint16   pad;
    float32  vertices[1];   /* variable-length */

} CTDB_VOLUME_MODEL_5;

/* Defined JUST for use by the compiler */
#define CTDB_INVALID_VMODEL_5 -1
#define CTDB_INVALID_LMODEL_5 -1
#define CTDB_INVALID_VOL_ID   -1

/**********************************************************************
 * Patch data headers                                                 *
 **********************************************************************/
/* The feature and TE data for each patch group is stored with a data header
 * at its beginning.  Because patch groups containing TINned patches need
 * substantially more data, there are two formats used, one for purely
 * gridded PGs, and one for PGs containing at least one TINned patch.
 */

typedef struct ctdb_patch_table_ref_5
{
    unsigned soil_table : 16;	/* Index of patch soil table */
    unsigned water_char : 16;	/* Index of patch water characteristics */
} CTDB_PATCH_TABLE_REF_5;


/* Wave characteristics */
typedef struct ctdb_wave_chars_5
{
    float32 height;		/* In meters */
    float32 period;		/* In seconds */
    float32 speed;		/* In meters per second */
    float32 direction;		/* $$$ units??? Tangent of some angle...*/
} CTDB_WAVE_CHARS_5;

/* Structure used to store surface characteristics for water polygons */
typedef struct ctdb_water_chars_5
{
    int16 tidal_zone_index;
    uint16          pad;
    float32         surface_temp;	/* In degrees C */
    float32         surf_height;	/* In meters */
    CTDB_WAVE_CHARS primary_wave;
    CTDB_WAVE_CHARS secondary_wave;
} CTDB_WATER_CHARS_5;

typedef struct ctdb_tidal_zone_5
{
    float32 offset;		/* Current tidal offset for this zone */
    /* $$$ add a region or identifier of some kind??? */
} CTDB_TIDAL_ZONE_5;

/* Data header for a completely gridded patch group.  This will *always*
 * be immediately followed by the feature data for the patch group.  The
 * starting address for the features for the first patch is
 * (address of this structure) + (sizeof(this structure)).
 * The starting address for the features for the other patches in the
 * PG can be found by
 * (address of this structure) + (patch*_offset, in 32-bit words)
 */
typedef struct ctdb_grid_pg_data_5
{
    float32                max_elev[CTDB_PATCHES_PER_GROUP_5];

    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32                water_elev[CTDB_PATCHES_PER_GROUP_5];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF_5 tables[CTDB_PATCHES_PER_GROUP_5];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;
} CTDB_GRID_PG_DATA_5;

/* Data header for patch group that contains at least one TINned patch.
 * This is *always* followd by the following (in order):
 *    Feature data for each patch in PG
 *    TE data for each patch in PG
 *    Virtual grid data for each patch in PG
 * The various addresses can all be calculated as described above for the
 * CTDB_GRID_PG_DATA structure (note that the offsets for everything
 * but features are explicitly stored for all 4 patches rather than just
 * the last 3).
 */
typedef struct ctdb_tin_pg_data_5
{
    float32       max_elev[CTDB_PATCHES_PER_GROUP_5];

    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32                water_elev[CTDB_PATCHES_PER_GROUP_5];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF_5 tables[CTDB_PATCHES_PER_GROUP_5];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;

    /* BITS             
     *	    3 : Features
     *	    2 : Buildings
     *	    1 : TREES
     *	    0 : Microterrain  
     */
    uint8   feature_present[CTDB_MAX_FEATURE_PRESENT_PER_PG_5];

    /* Array[patches] numb TE in patch */
    uint16  num_te[CTDB_PATCHES_PER_GROUP_5];
    /* Where the TEs for each patch start.  These are in units of 32-bit
     * words and relative to the starting address of this (ctdb_tin_pg_data)
     * data structure
     */
    uint32       te_offset[CTDB_PATCHES_PER_GROUP_5];
    /* SQRT(# of v. grid squares) for each patch */
    uint8        te_vgrid_dim[CTDB_PATCHES_PER_GROUP_5];
    /* Where the virtual grid starts for each patch.  These are in units
     * of 32-bit words and relative to the starting address of this
     * (ctdb_tin_pg_data) data structure
     */
    uint32       te_vgrid_offset[CTDB_PATCHES_PER_GROUP_5];
} CTDB_TIN_PG_DATA_5;


typedef union ctdb_pg_data_hdr_ptr_5
{
    struct ctdb_grid_pg_data_5 *grid_data;
    struct ctdb_tin_pg_data_5  *tin_data;
    int32                       offset;
} CTDB_PG_DATA_HDR_PTR_5;

/* A PATCH_GROUP is a collection of information about four contiguous
 * patches (contiguous in the sense x+y*w).  For each patch there are
 * four flags: occupied indicates there are features in the patch,
 * dyn_terr indicates that there are dynamic terrain changes to the
 * patch (future use), te16 indicates whether the patch's TEs use 8 or
 * 16 bit indices, and no_tes indicates whether or not a patch contains
 * TEs at all.  In addition, there is data about the patch group as a
 * whole:
 *     data_start points to the actual data for the cache.  Specifically,
 * it references either a CTDB_GRID_PG_DATA structure (contains_tins ==
 * FALSE) or a CTDB_TIN_PG_DATA structure (contains_tins == TRUE).  In either
 * case, that structure contains offsets for the feature/TE data itself,
 * which immediately follows the structure.
 *     cache specifies which of the 16 patch data caches this PG's data
 * should be placed in
 *     mapped is TRUE iff data_start points to actual data.  If it is FALSE,
 * the data must be read in from disk before being accessed.
 *     contains_tins is TRUE iff at least one patch in the PG is TINned
 */    
typedef struct ctdb_patch_group_5
{
    /* Bytes 0-3 */
    CTDB_PG_DATA_HDR_PTR_5     data_start;
    
    /* Bytes 4-7 */
    unsigned cache           : 4;  /* Which size cache bucket this PG uses */
    unsigned mapped          : 1;  /* 1 => PG data is in memory */
    unsigned contains_tins   : 1;  /* 1 => at least one patch in this PG
				    * is TINNed.  This determines what 
				    * data_start points to */
    unsigned patch0_occupied : 1;
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch0_no_tes   : 1;  /* 1 => patch 0 has 0 TEs */
    unsigned patch0_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch1_occupied : 1;
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch1_no_tes   : 1;  /* 1 => patch 1 has 0 TEs */
    unsigned patch1_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch2_occupied : 1;
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch2_no_tes   : 1;
    unsigned patch2_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch3_occupied : 1;
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch3_no_tes   : 1;
    unsigned patch3_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned unused          : 6;
} CTDB_PATCH_GROUP_5;

/**********************************************************************
 * Patch data cache                                                   *
 **********************************************************************/
typedef struct ctdb_cache_bucket_5
{
    CTDB_PATCH_GROUP_5 *user;	/* Patch group whose data is stored here */
    int32               patch_data_start_disk; /* The offset (in bytes)
						* from the start of the
						* patch data section in the
						* database file on disk at
						* which this PGs data starts
						*/
    union ctdb_feature_data_5 data[1];	/* Actual data.  This will be malloced
					 * to be longer than 1 word, but 
					 * declaring it this way saves us a 
					 * 4 bytes per bucket.
					 */
} CTDB_CACHE_BUCKET_5;

typedef struct ctdb_5
{
    /* Fixed point basis for elevation points.  Can be changed to deal
     * with future databases which have extreme elevations, or extreme
     * accuracy requirements.
     */
    float64         fixed_point_basis;
    float64         inv_fixed_point_basis;
    
    /* Tables mapping local 4 or 8 bit soil types to the global list */
    int32             num_soil_tables;
    int32           **soil_tables;
    int32             soil_table_size;
    int8             *soil_table_storage;

    /* Table of sea-state combinations */
    int32                      num_water_chars;
    struct ctdb_water_chars_5 *water_chars;
    
    /* Table of tidal zones */
    int32                      num_tidal_zones;
    struct ctdb_tidal_zone_5  *tidal_zones;

    /* Spacing between grid posts */
    int32           incr;
    float64         inv_incr;
    float64         incrD;
    
    /* Spacing between patches */
    int32           patch_incr    ;/* number of posts per patch */
    float64         patch_inv_incr;/* 1 over patch_incr */
    float64         patch_incrD   ;/* double patch_incr */
    
    /* Database size */
    int32           pages_wide;
    int32           pages_high;
    int32           patches_wide;
    int32           patches_high;
    int32           min_x;
    int32           max_x;
    int32           min_y;
    int32           max_y;
    int32           min_x_post;
    int32           max_x_post;
    int32           min_y_post;
    int32           max_y_post;
    
    /* Useful trivia */
    int32           input_format;
    float64         min_z;
    float64         max_z;
    float64         origin_northing;
    float64         origin_easting;
    int32           origin_zone_number;
    int32           origin_zone_letter;
    int32           datum;
    int32           cell_id; /* added for gcs; 
			      * GCS_ILLEGAL_CELL implies non-gcs database 
			      */

    /* origin_x and origin_y were added for GCS.  They specify the offset
     * from the center of the cell at which x and y should be 0, respectively
     */
    int32           origin_x;
    int32           origin_y;

    /* The following specify, repsectively, the left and right longitudinal
     * boundaries of the database and the bottom and top latitudinal boundaries
     * of the database
     */
    float64         west_long;
    float64         east_long;
    float64         south_lat;
    float64         north_lat;

    CTDB_TDB_TYPE   gcs_mode;
    char            name[CTDB_NAME_LENGTH_5];
    int16           version;
    char            date[CTDB_DATE_LENGTH_5];
    int32           total_memory;
    uint32          is_curved;
    uint32          is_sw_ne;
    CTDB_PATCH_GROUP_5 *patch_groups;
    int32           elevations_resident;
    int32           elevation_seek;
    int32           elevation_cache_pages;
    int32           direct_map_modulus;
    int32           elevation_hit_count;
    int32           elevation_miss_count;
    int32          *cache_page_numbers;
    int32         **elevation_pages;

    /* Caches sizes are in units of CTDB_CACHE_BKT_UNITS */
    int32           patch_cache_size_map[CTDB_PATCH_CACHES_5];

    int32           patch_data_seek;
    int32           patch_cache_entries;
    int32           patch_hit_count;
    int32           patch_miss_count;
    struct ctdb_cache_bucket_5 *patch_data_cache[CTDB_PATCH_CACHES_5];

    /* Database file */
    FILE           *file;

    int32          *elevations;

    /* Quadtree of abstract features */
    int32           num_quad_data;
    union ctdb_feature_data_5 *quad_data;

    /* Topologic relationships */
    int32           num_nodes;
    CTDB_NODE_5   **nodes;
    int32           node_size;
    int8           *node_storage;
    int32           num_edges;
    CTDB_EDGE_5   **edges;
    int32           edge_size;
    int8           *edge_storage;

    /* Linear models */
    int32                 num_linear_models;
    CTDB_LINEAR_MODEL_5  *linear_models;

    /* Volume models */
    int32                 num_volume_models;
    CTDB_VOLUME_MODEL_5 **volume_models;

    int32           n_patch_bytes;     /* Total amt of patch data for this 
					* cell, in bytes  (patch data = TEs +
					* features + virtual grids)
					*/
    ADDRESS         patch_data;	       /* Pointer to chunk o' memory with all
					* the patch data
					*/
    float64         meters_to_patch;   /* = patch_inv_incr * inv_incr  */
    uint32          is_gridded;
    uint32          is_tinned;
    int32           max_patch_number;
    int32           min_patch_number;
} CTDB_5;

#define CTDB_HEADER_RESERVED_5             3
typedef struct ctdb_file_header_5
{
    int32    format;
    unsigned is_curved : 1;
    unsigned is_sw_ne  : 1;
    unsigned is_gridded: 1;
    unsigned is_tinned : 1;
    unsigned pad1      : 28;
    float64  fixed_point_basis;
    int32    reserved1[16];
    int32    post_increment;
    int32    posts_per_patch_side;
    int32    pages_wide;
    int32    pages_high;
    int32    min_x;
    int32    max_x;
    int32    min_y;
    int32    max_y;
    int32    num_features;
    int32    patch_cache_size_map[CTDB_PATCH_CACHES_5];
    int32    pad2;
    float64  origin_northing;
    float64  origin_easting;
    int32    origin_zone_number;
    int32    origin_zone_letter;
    int32    datum;
    char     name[CTDB_NAME_LENGTH_5];
    int16    version;
    char     date[CTDB_DATE_LENGTH_5];
    int32    num_nodes;
    int32    node_size;
    int32    num_edges;
    int32    edge_size;
    int32    num_quad_data;
    int32    n_patch_bytes;
    int32    num_linear_models;
    int32    num_volume_models;
    int32    aux_hdr_bytes;	/* Size of auxiliary header */
    int32    num_soil_tables;
    int32    soil_table_size;
    int32    num_water_chars;
    int32    num_tidal_zones;
    int32    reserved[CTDB_HEADER_RESERVED_5]; /* For future use, if we 
						* need it */
} CTDB_FILE_HEADER_5;

/* Additional data we didn't want to try to cram into the file header */
typedef struct ctdb_aux_file_header_5
{
    unsigned gcs_mode : 2;
    unsigned pad1     : 30;
    int32 cell_id;
    int32 origin_x;
    int32 origin_y;
    float64 west_long;		/* Left-most longiitude in database */
    float64 east_long;		/* Right-most longitude in database */
    float64 south_lat;		/* Bottom-most latitude in database */
    float64 north_lat;		/* Top-most latitude in database */
} CTDB_AUX_FILE_HEADER_5;


extern void *
ctdb_convert_5_to_4(void *fmt5_db);

extern void *
ctdb_convert_5_to_6(void *fmt5_db);

extern void
ctdb_free_fmt5_db(void *fmt5_db);

/* Mallocs the needed memory, populates it, and returns it. */
extern void *
ctdb_read_5(char *fname);

extern void
ctdb_write_5(
	     void *fmt5_db,
	     FILE *file
	     );

#endif /* _CMP_FMT5_H_INCLUDED */




