/*
 * $RCSfile: ct_cmp_fmt6.c,v $ $Revision: 1.1.1.1 $ $State: Exp $ 
 */
/****************************************************************************
 *   File: cmp_fmt6.c                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 6 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/

#include "ct_cmplr.h"


/* Forward declarations */

static void
generate_caching_scheme_6(CTDB_6 *ctdb); 


static int32
find_pg_size_6(CTDB_PATCH_GROUP_6 *pg);

static void
write_patch_group_headers_6(
			    CTDB_6             *ctdb,
			    CTDB_PATCH_GROUP_6 *patch_groups,
			    int32               groups,
			    FILE               *output_file
			    );

static void 
write_nodes_and_edges_6(
			CTDB_6  *ctdb,
			FILE    *output_file
			);

static void 
write_abstract_features_6(
			  CTDB_6  *ctdb,
			  FILE    *output_file
			  );

static void 
write_linear_models_6(
		      CTDB_6  *ctdb,
		      FILE    *output_file
		      );

static void
write_volume_models_6(
		      CTDB_6  *ctdb,
		      FILE    *output_file
		      );

static void
write_aggregate_models_6(
			 CTDB_6  *ctdb,
			 FILE    *output_file
			 );

static void
write_mes_6(
		       CTDB_6 *ctdb,
		       FILE   *output_file
		      );

static void
write_soil_tables_6(
		    CTDB_6  *ctdb,
		    FILE    *output_file
		    );

static void
write_water_chars_6(
		    CTDB_6  *ctdb,
		    FILE    *output_file
		    );

static void
write_tidal_zones_6(
		    CTDB_6  *ctdb,
		    FILE    *output_file
		    );

static void 
write_elevation_posts_6(
			CTDB_6  *ctdb,
			FILE    *output_file
			);

static void
write_patch_data_6(
		   CTDB_6             *ctdb,
		   FILE               *output_file
		   );

static void
regen_table_ptrs(
		 int8   *old_base,
		 int8  **old_ptrs,
		 int8   *new_base,
		 int8  **new_ptrs,
		 int32   num_entries
		 );

static void
convert_soils_6_7(
                  CTDB_6 *ctdb_6,
                  CTDB_7 *CTDB_7
                  );

static void
convert_elevations_6_7(
                       CTDB_6   *ctdb,
                       CTDB_7   *ctdb_out
                       );

static void
get_features_and_tes_6_to_7(
			      CTDB_6   *ctdb,
			      CTDB_7 *ctdb_out
			      );

static int32
convert_features_6_7(
                      CTDB_6 *ctdb_6,
                      CTDB_7 *ctdb_7
                      );

static int32
translate_tinned_pg_6_to_7(
			     CTDB_PATCH_GROUP_6   *pg_6,
			     CTDB_PATCH_GROUP_7 *pg_7,
			     int32 feat_offset,
			     int32 te_offset
			     );

static void
translate_gridded_pg_6_to_7(
			      CTDB_PATCH_GROUP_6 *pg_6,
			      CTDB_PATCH_GROUP_7 *pg_7,
			      int32 feat_offset
			      );

static int32
get_tes_6_to_7(
		 CTDB_PATCH_GROUP_6   *pg_6,
		 CTDB_PATCH_GROUP_7 *pg_7,
		 TERRAIN_ELEMENT_16   *tes
		 );

static void
copy_mes_templates_6_7(
			 CTDB_6   *ctdb,
			 CTDB_7 *ctdb_out
			 );

static void
copy_mes_volumes_6_7(
		       CTDB_6   *ctdb,
		       CTDB_7 *ctdb_out
		       );

static int32 
patch_inside_dbase_6(
		     CTDB_6        *ctdb,
		     int32          patch_number
		     );

static void 
ctdb_pack_header_6(
		   CTDB_6             *ctdb,
		   CTDB_FILE_HEADER_6 *hdr
		   );

static void
ctdb_pack_aux_header_6(
		       CTDB_6                 *ctdb,
		       CTDB_AUX_FILE_HEADER_6 *aux_hdr
		       );

static int32
find_words_in_gridded_pg_6(
			   CTDB_PATCH_GROUP_6 *pg
			   );

static int32
convert_features_6_5(
		     CTDB_6 *ctdb_6,
		     CTDB_5 *ctdb_5
		     );

static void
convert_elevation_posts_6_5(
			    CTDB_6 *ctdb_6,
			    CTDB_5 *ctdb_5
			    );

/***************************************************************************
 * READ ROUTINES                                                           *
 ***************************************************************************/

/* $$$ I'm not going to find z-limits here; I'll wait until I get everything
 * into the compiler format so I only have to write the code once.
 */

static void 
ctdb_unpack_header_6(CTDB_6 *ctdb, CTDB_FILE_HEADER_6 *hdr)
{
    ctdb->input_format          = hdr->format;
    ctdb->is_curved             = hdr->is_curved;
    ctdb->is_gridded            = hdr->is_gridded;
    ctdb->is_tinned             = hdr->is_tinned;
    ctdb->fixed_point_basis     = hdr->fixed_point_basis;

    if (ctdb->fixed_point_basis) 
      ctdb->inv_fixed_point_basis = 1.0/ctdb->fixed_point_basis;

    ctdb->num_soil_tables = hdr->num_soil_tables;
    ctdb->soil_table_size = hdr->soil_table_size;
    ctdb->num_water_chars = hdr->num_water_chars;
    ctdb->num_tidal_zones = hdr->num_tidal_zones;

    ctdb->incr           = hdr->post_increment;
    ctdb->incrD          = ctdb->incr;
    ctdb->patch_incr     = hdr->posts_per_patch_side;
    ctdb->patch_incrD    = ctdb->patch_incr;

    if (ctdb->incrD)
      ctdb->inv_incr = 1.0/ctdb->incrD;

    if (ctdb->patch_incrD)
      ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide     = hdr->pages_wide;
    ctdb->pages_high     = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE * ctdb->patch_inv_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE * ctdb->patch_inv_incr;
    
    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);
    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);

    bcopy(hdr->patch_cache_size_map, ctdb->patch_cache_size_map,
	  CTDB_PATCH_CACHES * sizeof(int32));

    ctdb->origin_northing    = hdr->origin_northing;
    ctdb->origin_easting     = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum              = hdr->datum;

    bcopy(hdr->name,ctdb->name, sizeof(hdr->name));

    ctdb->version            = hdr->version;

    bcopy(hdr->date,ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date,       "\n");

    ctdb->num_nodes        = hdr->num_nodes;
    ctdb->node_size        = hdr->node_size;
    ctdb->num_edges        = hdr->num_edges;
    ctdb->edge_size        = hdr->edge_size;
    ctdb->num_quad_data    = hdr->num_quad_data;
    ctdb->n_patch_bytes    = hdr->n_patch_bytes;
    ctdb->num_linear_models = hdr->num_linear_models;
    ctdb->num_volume_models = hdr->num_volume_models;
    ctdb->num_aggregate_models = hdr->num_aggregate_models;
    ctdb->num_mes_volumes   = hdr->num_mes_volumes;
    ctdb->num_mes_templates = hdr->num_mes_templates;
    ctdb->meters_to_patch  = ctdb->patch_inv_incr * ctdb->inv_incr;
    ctdb->max_patch_number = ctdb->patches_wide * ctdb->patches_high - 1;
    ctdb->min_patch_number = 0;
    return;
}

static void
ctdb_unpack_aux_header_6(CTDB_6 *ctdb, CTDB_AUX_FILE_HEADER_6 *aux_hdr)
{
    ctdb->gcs_mode  = aux_hdr->gcs_mode;
    ctdb->cell_id   = aux_hdr->cell_id;
    ctdb->origin_x  = aux_hdr->origin_x;
    ctdb->origin_y  = aux_hdr->origin_y;
    ctdb->west_long = aux_hdr->west_long;
    ctdb->east_long = aux_hdr->east_long;
    ctdb->south_lat = aux_hdr->south_lat;
    ctdb->north_lat = aux_hdr->north_lat;
    return;
}

static CTDB_STATUS
read_patch_group_headers_6(CTDB_6 *ctdb, FILE *fileptr)
{
  int32 groups;

  /* Read in all the patch group headers */
  groups=(ctdb->patches_wide*ctdb->patches_high)/CTDB_PATCHES_PER_GROUP_6;
  ctdb->patch_groups=(CTDB_PATCH_GROUP_6 *)
    ctdb_alloc(NULL,
	       sizeof(CTDB_PATCH_GROUP_6)*groups,
	       "patch group data");
  fread(ctdb->patch_groups,sizeof(CTDB_PATCH_GROUP_6),groups,fileptr);

  return CTDB_STATUS_OK;
}
 
static void 
init_null_topology_6(CTDB_6 *ctdb)
{
    ctdb->node_storage = (int8 *) 
      ctdb_alloc(NULL, sizeof(CTDB_NODE_6), "initialization");
    bzero(ctdb->node_storage, sizeof(CTDB_NODE_6));
    ctdb->edge_storage = (int8 *) 
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_6), "initialization");
    bzero(ctdb->edge_storage, sizeof(CTDB_EDGE_6));
    ctdb->nodes = (CTDB_NODE_6 **) 
      ctdb_alloc(NULL, sizeof(CTDB_NODE_6 *), "initialization");
    ctdb->nodes[0] = (CTDB_NODE_6 *)ctdb->node_storage;
    ctdb->edges = (CTDB_EDGE_6 **) 
      ctdb_alloc(NULL, sizeof(CTDB_EDGE_6 *), "initialization");
    ctdb->edges[0] = (CTDB_EDGE_6 *)ctdb->edge_storage;
}

static CTDB_STATUS 
ctdb_read_topology_6(CTDB_6 *ctdb, FILE *fileptr)
{
    int32 i, offset;
    
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces: ctdb_read_topology((CTDB *)ctdb,fileptr);*/
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    if (ctdb->num_nodes)
    {
        ctdb->nodes = NULL;
	ctdb->nodes =
	  (CTDB_NODE_6 **)ctdb_alloc(NULL,
				     sizeof(CTDB_NODE_6 *) *
				     ctdb->num_nodes,
				     "node index table");
	ctdb->node_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->node_size, "nodes");

	if (fread(ctdb->node_storage, ctdb->node_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Nodes\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_NODES_FAILED;
	}
    }

    if (ctdb->num_edges)
    {
        ctdb->edges = NULL;
	ctdb->edges =
	  (CTDB_EDGE_6 **)ctdb_alloc(NULL,
				     sizeof(CTDB_EDGE_6 *) *
				     ctdb->num_edges,
				     "edge index table");
	ctdb->edge_storage =
	  (int8 *)ctdb_alloc(NULL, ctdb->edge_size, "edges");

	if (fread(ctdb->edge_storage, ctdb->edge_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Edges\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_EDGES_FAILED;
	}
    }

    /* Fill in the node index table */
    bzero(ctdb->nodes, sizeof(CTDB_NODE_6 *) * ctdb->num_nodes);
    for(i=offset=0;(i<ctdb->num_nodes) && (offset<ctdb->node_size);i++)
    {
	ctdb->nodes[i] = (CTDB_NODE_6 *)(ctdb->node_storage + offset);
	offset += sizeof(CTDB_NODE_6) + sizeof(struct ctdb_connection) *
	  (ctdb->nodes[i]->num_connections - 1);
    }

    /* Fill in the edge index table */
    bzero(ctdb->edges, sizeof(CTDB_EDGE_6 *) * ctdb->num_edges);
    for(i=offset=0;(i<ctdb->num_edges) && (offset<ctdb->edge_size);i++)
    {
	ctdb->edges[i] = (CTDB_EDGE_6 *)(ctdb->edge_storage + offset);
	offset += sizeof(CTDB_EDGE_6) + sizeof(struct ctdb_connection) *
	  (ctdb->edges[i]->num_features - 1);
    }

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_quad_tree_6(CTDB_6 *ctdb, FILE *fileptr)
{
    if (ctdb->num_quad_data == 0 )
      return CTDB_STATUS_OK;
        
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces ctdb_read_quad_tree((CTDB *)ctdb,fileptr); */
    ctdb->quad_data =
      (CTDB_FEATURE_DATA_6 *)ctdb_alloc(NULL,
					sizeof(CTDB_FEATURE_DATA_6) *
					ctdb->num_quad_data,
					"abstract data and quadtree");

    if (fread(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_6),
	      ctdb->num_quad_data, fileptr) != ctdb->num_quad_data)
    {
	printf("CTDB: Read failed.  Quad tree and abstract data\n");
	/* Survive this error */
	ctdb->num_quad_data = 0;
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_linear_models_6(CTDB_6 *ctdb, FILE *fileptr)
{
    if (ctdb->num_linear_models == 0 )
    {
	ctdb->linear_models = NULL;
	return CTDB_STATUS_OK;
    }
        
    /* Allocate space for the linear models.  Then read the data.
     */
    ctdb->linear_models = NULL;
    ctdb->linear_models =
      (CTDB_LINEAR_MODEL_6 *)ctdb_alloc(NULL,
					sizeof(CTDB_LINEAR_MODEL_6) *
					ctdb->num_linear_models,
					"linear models");

    if (fread(ctdb->linear_models, sizeof(CTDB_LINEAR_MODEL_6),
	      ctdb->num_linear_models, fileptr) != ctdb->num_linear_models)
    {
	printf("CTDB: Read failed.  Linear models\n");
	/* Survive this error */
	ctdb->num_linear_models = 0;
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_volume_models_6(CTDB_6 *ctdb, FILE *fileptr)
{
    int32                i, model_size;
    CTDB_VOLUME_MODEL_6 *model;

    if (ctdb->num_volume_models == 0)
    {
	ctdb->volume_models = NULL;
	return CTDB_STATUS_OK;
    }

    /* Allocate space for the array of pointers to the volume models.  Then
     * read in each model.
     */
    ctdb->volume_models = NULL;
    ctdb->volume_models =
      (CTDB_VOLUME_MODEL_6 **)ctdb_alloc(NULL,
					 sizeof(CTDB_VOLUME_MODEL_6 *) *
					 ctdb->num_volume_models,
					 "volume models");

    /* Allocate space for a temporary model (large).
     */
    model = (CTDB_VOLUME_MODEL_6 *)STDALLOC(sizeof(CTDB_VOLUME_MODEL_6) +
					    1024 * 3 * sizeof(float32));

    for(i = 0; i < ctdb->num_volume_models; i++)
    {
	/* Read in all of the model except the vertices.
	 */
	if(fread(model, sizeof(CTDB_VOLUME_MODEL_6) - sizeof(float32), 1,
		 fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Volume models\n");
	    /* Survive this error */
	    ctdb->num_volume_models = 0;
	    return CTDB_STATUS_OK;
	}

	/* Now that we know how many vertices are in this model, read them in.
	 */
	if(fread(model->vertices, model->num_vertices * 3 * sizeof(float32),
		 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Volume models\n");
	    /* Survive this error */
	    ctdb->num_volume_models = 0;
	    return CTDB_STATUS_OK;
	}   

	model_size = sizeof(CTDB_VOLUME_MODEL_6) - sizeof(float32) +
	  model->num_vertices * 3 * sizeof(float32);

	/* Now that the entire model has been read in, store it.
	 */
	ctdb->volume_models[i] =
	  (CTDB_VOLUME_MODEL_6 *)ctdb_alloc(NULL, model_size,
					    "volume model");

	bcopy(model, ctdb->volume_models[i], model_size);
    }

    STDDEALLOC(model);

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_aggregate_models_6(CTDB_6 *ctdb, FILE *fileptr)
{
    int32                   i, model_size;
    CTDB_AGGREGATE_MODEL_6 *model;

    if (ctdb->num_aggregate_models == 0)
    {
	ctdb->aggregate_models = NULL;
	return CTDB_STATUS_OK;
    }

    /* Allocate space for the array of pointers to the aggregate models.  Then
     * read in each model.
     */
    ctdb->aggregate_models =
      (CTDB_AGGREGATE_MODEL_6 **)ctdb_alloc(NULL,
					    sizeof(CTDB_AGGREGATE_MODEL_6 *) *
					    ctdb->num_aggregate_models,
					    "aggregate models");

    /* Allocate space for a temporary model (large).
     */
    model = (CTDB_AGGREGATE_MODEL_6 *)STDALLOC(sizeof(CTDB_AGGREGATE_MODEL_6) +
					       512 *
					     sizeof(CTDB_AGG_MODEL_FEATURE_6));

    for(i = 0; i < ctdb->num_aggregate_models; i++)
    {
	/* Read in all of the model except the features.
	 */
	if(fread(model, sizeof(CTDB_AGGREGATE_MODEL_6) -
		 sizeof(CTDB_AGG_MODEL_FEATURE_6), 1,
		 fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Aggregate models\n");
	    /* Survive this error */
	    ctdb->num_aggregate_models = 0;
	    return CTDB_STATUS_OK;
	}

	/* Now that we know how many features are in this model, read them in.
	 */
	if(fread(model->features, model->n_features *
		 sizeof(CTDB_AGG_MODEL_FEATURE_6), 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Aggregate models\n");
	    /* Survive this error */
	    ctdb->num_aggregate_models = 0;
	    return CTDB_STATUS_OK;
	}

	model_size = sizeof(CTDB_AGGREGATE_MODEL_6) +
	  (model->n_features - 1) * sizeof(CTDB_AGG_MODEL_FEATURE_6);

	/* Now that the entire model has been read in, store it.
	 */
	ctdb->aggregate_models[i] =
	  (CTDB_AGGREGATE_MODEL_6 *)ctdb_alloc(NULL, model_size,
					       "aggregate model");

	bcopy(model, ctdb->aggregate_models[i], model_size);
    }

    STDDEALLOC(model);

    return CTDB_STATUS_OK;
}

/* read the MES volume templates from the database */
static CTDB_STATUS 
ctdb_read_mes_templates_6(CTDB_6 *ctdb, FILE *fileptr)
{
    int                   i, j;
    CTDB_MES_TEMPLATE_6  *cur_temp;
    CTDB_MES_ENCLOSURE_6 *cur_encl;
    CTDB_MES_APERTURE_6  *cur_ap;
    CTDB_MES_GRID_BOX_6  *cur_box;
    int32                 n_ap;

    if(ctdb->num_mes_templates == 0)
    {
	ctdb->mes_templates = NULL;
	return CTDB_STATUS_OK;
    }

    /* Assume that if there were MES's here before, they've been deallocated
     * by now.  We can't just realloc here because there are pointers hanging
     * off these pointers.
     */
    ctdb->mes_templates = (CTDB_MES_TEMPLATE_6 *)
      ctdb_alloc(NULL, ctdb->num_mes_templates * sizeof(CTDB_MES_TEMPLATE_6),
		 "MES TEMPLATES");

    /* Just to be safe */
    bzero(ctdb->mes_templates, 
	  ctdb->num_mes_templates * sizeof(CTDB_MES_TEMPLATE_6));    

    /* Now get each MES template in order */
    for(i = 0; i < ctdb->num_mes_templates; i++)
    {
	cur_temp = ctdb->mes_templates + i;

	fread(cur_temp, sizeof(CTDB_MES_TEMPLATE_6), 1, fileptr);

	if(cur_temp->num_vertices == 0)
	{
	    cur_temp->all_vertices = NULL;
	}
	else
	{
	    cur_temp->all_vertices = (CTDB_MES_VERTEX *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_VERTEX) * (cur_temp->num_vertices),
			 "MES template vertices");

	    fread(cur_temp->all_vertices, sizeof(int32)*
		  XYZ*(cur_temp->num_vertices), 1, fileptr);
	}

	if(cur_temp->num_triangles == 0)
	{
	    cur_temp->all_triangles = NULL;
	}
	else
	{
	    cur_temp->all_triangles = (CTDB_MES_TRIANGLE_6 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_TRIANGLE_6) *
			 cur_temp->num_triangles,
			 "MES template triangles");

	    fread(cur_temp->all_triangles,
		  sizeof(CTDB_MES_TRIANGLE_6) *
		  cur_temp->num_triangles,
		  1, fileptr);
	}

	/* Need to get the enclosures for this template */
	if(cur_temp->num_enclosures == 0)
	{
	    cur_temp->all_enclosures = NULL;
	}
	else
	{
	    cur_temp->all_enclosures = (CTDB_MES_ENCLOSURE_6 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_ENCLOSURE_6) *
			 cur_temp->num_enclosures,
			 "MES enclosures");

	    for(j = 0; j < cur_temp->num_enclosures; j++)
	    {
		cur_encl = cur_temp->all_enclosures + j;

		fread(cur_encl, 
		      sizeof(CTDB_MES_ENCLOSURE_6),
		      1, fileptr);

		/* Need to get vertices and aperture id arrays for this 
		 * enclosure 
		 */
		if(cur_encl->footprint.num_vertices == 0)
		{
		    cur_encl->footprint.vertex_ids = NULL;
		}
		else
		{
		    cur_encl->footprint.vertex_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32) * 
				 cur_encl->footprint.num_vertices,
				 "MES enclosure vertices");

		    fread(cur_encl->footprint.vertex_ids,
			  sizeof(int32)*cur_encl->footprint.num_vertices,
			  1 , fileptr);
		}

		if(cur_encl->num_apertures == 0)
		{
		    cur_encl->aperture_ids = NULL;
		}
		else
		{
		    n_ap = cur_encl->num_apertures;
		    /* Round to nearest even number so we have an integral
		     * number of words
		     */
		    n_ap = n_ap + (n_ap % 2);

		    cur_encl->aperture_ids = (uint16 *)
		      ctdb_alloc(NULL,
				 sizeof(uint16) * n_ap,
				 "MES enclosure aperture ids");

		    fread(cur_encl->aperture_ids,
			  sizeof(uint16) * n_ap,
			  1, fileptr);
		}
	    }
	}

	/* Need to get the apertures for this template */
	if(cur_temp->num_apertures == 0)
	{
	    cur_temp->all_apertures = NULL;
	}
	else
	{
	    cur_temp->all_apertures = (CTDB_MES_APERTURE_6 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_APERTURE_6)*cur_temp->num_apertures,
			 "MES apertures");

	    for(j = 0; j < cur_temp->num_apertures; j++)
	    {
		cur_ap = cur_temp->all_apertures + j;

		fread(cur_ap, sizeof(CTDB_MES_APERTURE_6),
		      1, fileptr);

		/* Need to get the vertices for this aperture */
		if(cur_ap->outline.num_vertices == 0)
		{
		    cur_ap->outline.vertex_ids = NULL;
		}
		else
		{
		    cur_ap->outline.vertex_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32) * 
				 cur_ap->outline.num_vertices,
				 "MES aperture vertices");

		    fread(cur_ap->outline.vertex_ids,
			  sizeof(int32)*cur_ap->outline.num_vertices,
			  1, fileptr);
		}
	    }
	}

	/* Need to get the boxes for this MES */
	if(cur_temp->num_boxes ==0)
	{
	    cur_temp->all_boxes = NULL;
	}
	else
	{
	    cur_temp->all_boxes = (CTDB_MES_GRID_BOX_6 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_GRID_BOX_6)*cur_temp->num_boxes,
			 "MES boxes");

	    for(j = 0; j < cur_temp->num_boxes; j++)
	    {
		cur_box = cur_temp->all_boxes + j;

		fread(cur_box, sizeof(CTDB_MES_GRID_BOX_6), 1, fileptr);

		/* Lets see if there are any boxes inferior to this one */
		if( (cur_box->dimensions_x*
		     cur_box->dimensions_y*
		     cur_box->dimensions_z) > 1 )
		{
		    cur_box->box_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32)*(cur_box->dimensions_x*
						cur_box->dimensions_y*
						cur_box->dimensions_z),
				 "MES interior box ids");

		    fread(cur_box->box_ids,
			  sizeof(int32)*(cur_box->dimensions_x*
					 cur_box->dimensions_y*
					 cur_box->dimensions_z),
			  1, fileptr);
		}

		if(cur_box->num_apertures == 0)
		{
		    cur_box->aperture_ids = NULL;
		}
		else
		{
		    n_ap = cur_box->num_apertures;
		    
		    /* Round to nearest even number so we have an integral
		     * number of words
		     */
		    n_ap = n_ap + (n_ap % 2);
		    
		    cur_box->aperture_ids = (uint16 *)
		      ctdb_alloc(NULL,
				 sizeof(uint16) * n_ap,
				 "MES box aperture ids");

		    fread(cur_box->aperture_ids,
			  sizeof(uint16) * n_ap,
			  1, fileptr);
		}

		if(cur_box->num_triangles == 0)
		{
		    cur_box->wall_triangle_ids = NULL;
		}
		else
		{
		    cur_box->wall_triangle_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32)*cur_box->num_triangles,
				 "MES box triangle ids");

		    fread(cur_box->wall_triangle_ids,
			  sizeof(int32)*cur_box->num_triangles,
			  1, fileptr);
		}
	    }
	}
    }
    return CTDB_STATUS_OK;
}

/* Read in the MES volumes for this database */
static CTDB_STATUS
ctdb_read_mes_volumes_6(CTDB_6 *ctdb, FILE *fileptr)
{
    int32               i;
    CTDB_MES_VOLUME_6  *cur_vol;

    if (!ctdb->num_mes_volumes)
      return CTDB_STATUS_OK;
    
    ctdb->mes_volumes = (CTDB_MES_VOLUME_6 *)
      ctdb_alloc(NULL, ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME_6),
		 "MES Volumes");

    /* Just to be safe */
    bzero(ctdb->mes_volumes, 
	  ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME_6));    

    /* Now get each MES volume in order */
    for(i = 0; i < ctdb->num_mes_volumes; i++)
    {
        cur_vol = ctdb->mes_volumes + i;
	fread(cur_vol, sizeof(CTDB_MES_VOLUME_6), 1, fileptr);
	
	if(cur_vol->roofline.num_vertices == 0)
	  {
	    cur_vol->roofline.vertices = NULL;
	  }
	else
	  {
	    cur_vol->roofline.vertices = (CTDB_MES_VERTEX_F64_6 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_VERTEX_F64_6) *
			 cur_vol->roofline.num_vertices,
			 "MES volume roofline vertices");
	    
	    fread(cur_vol->roofline.vertices,
		  sizeof(CTDB_MES_VERTEX_F64_6) *
		  (cur_vol->roofline.num_vertices),
		  1,
		  fileptr);
	  }
	if(cur_vol->floorline.num_vertices == 0)
	  {
	    cur_vol->floorline.vertices = NULL;
	  }
	else
	  {
	    cur_vol->floorline.vertices = (CTDB_MES_VERTEX_F64_6 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_VERTEX_F64_6) *
			 cur_vol->floorline.num_vertices,
			 "MES volume floorline vertices");
	    
	    fread(cur_vol->floorline.vertices,
		  sizeof(CTDB_MES_VERTEX_F64_6) *
		  (cur_vol->floorline.num_vertices),
		  1,
		  fileptr);
	  }
	
	/* Now fill in the template pointers.  Until we do this, they
	 * contain indices rather than pointers
	 */
	ctdb->mes_volumes[i].template = ctdb->mes_templates + 
	  (int32)ctdb->mes_volumes[i].template;
    }
	
    return CTDB_STATUS_OK;
}


/* Read the MES data for the database */
CTDB_STATUS
ctdb_read_mes_6(CTDB_6 *ctdb, FILE *fileptr)
{
    CTDB_STATUS stat;

    /* The MES data is broken into two parts--generic templates, and
     * volumes (instances) which reference them.
     */

    if ((stat = ctdb_read_mes_templates_6(ctdb, fileptr)) != CTDB_STATUS_OK)
      return stat;
    
    return ctdb_read_mes_volumes_6(ctdb, fileptr);
}


static CTDB_STATUS 
ctdb_read_soil_tables_6(CTDB_6 *ctdb, FILE *fileptr)
{
    int32 i;

    ctdb->soil_tables = (int32 **)ctdb_alloc(NULL,
					     ctdb->num_soil_tables *
					     sizeof(int32 *),
					     "Soil table index");
    
    ctdb->soil_table_storage = (int8 *)ctdb_alloc(NULL,
						  ctdb->soil_table_size,
						  "Soil tables");

    /* Get the indices for the soil tables */
    fread(ctdb->soil_tables, sizeof(int32 *), ctdb->num_soil_tables, fileptr);

    /* These are kept in memory as pointers, but stored on disk as offsets.
     * Convert them to pointers.
     */
    for(i=0;i<ctdb->num_soil_tables;i++)
    {
	ctdb->soil_tables[i] = (int32 *)((int8 *)ctdb->soil_table_storage +
					 (int32)(ctdb->soil_tables[i]));
    }
    
    /* Read the soil tables themselves */
    fread(ctdb->soil_table_storage, 1, ctdb->soil_table_size, fileptr);
}

static CTDB_STATUS 
ctdb_read_water_chars_6(CTDB_6 *ctdb, FILE *fileptr)
{
    ctdb->water_chars = (struct ctdb_water_chars_6 *)
      ctdb_alloc(NULL, ctdb->num_water_chars * 
		 sizeof(struct ctdb_water_chars_6), "Water characteristics");
    
    fread(ctdb->water_chars, sizeof(struct ctdb_water_chars),
	  ctdb->num_water_chars, fileptr);    
}

static CTDB_STATUS 
ctdb_read_tidal_zones_6(CTDB_6 *ctdb, FILE *fileptr)
{
    ctdb->tidal_zones = (struct ctdb_tidal_zone_6 *)
      ctdb_alloc(NULL, ctdb->num_tidal_zones *
		 sizeof(struct ctdb_tidal_zone_6), "Tidal zones");
    
    fread(ctdb->tidal_zones, sizeof(struct ctdb_tidal_zone_6), 
	  ctdb->num_tidal_zones, fileptr);
}

static CTDB_STATUS
ctdb_read_all_patch_data_6(CTDB_6 *ctdb, FILE *f)
{
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP_6 *pg;

    ctdb->patch_data =
      (ADDRESS)ctdb_alloc(NULL,
			  ctdb->n_patch_bytes,
			  "DB Patch data");

    /* Read _all_ the patch data for the cell in in one big gulp */
    fread(ctdb->patch_data,1,ctdb->n_patch_bytes,f);

    pg = ctdb->patch_groups;

    /* Now that we have the data, we need to set up the pointers in all the
     * patch headers
     */
    groups = (ctdb->patches_wide*ctdb->patches_high) / 
      CTDB_PATCHES_PER_GROUP_6;

    for (i=0;i<groups;i++)
    {
	/* Doesn't matter which elt of the union we choose here... */
	pg[i].data_start.grid_data = (CTDB_GRID_PG_DATA_6 *)
	  ((int8*)ctdb->patch_data + pg[i].data_start.offset);
	pg[i].mapped = 1;
    }

    return CTDB_STATUS_OK;    
}

void 
ctdb_read_elevations_6(CTDB_6 *ctdb, FILE *f)
{
    uint32 memory, pages;
    int32  i;
    int    file_pos;

    pages = ctdb->pages_wide * ctdb->pages_high;

    memory = PAGE_SIZE * pages;

    ctdb->elevations_resident = 1;
    ctdb->elevations = (int32 *)ctdb_alloc(NULL,
					   memory, "elevation data");

    /* Note the file pointer value */
    file_pos = ftell(f);

    /* Load as many pages as the caller will let us */
    if (fread(ctdb->elevations, memory, 1, f) != 1)
      printf("CTDB: Read failed.  First %d bytes of post data.\n",
	     memory), exit(1);

    /* Position the file pointer at the end of the elevations */
    file_pos += PAGE_SIZE * pages;
    fseek(f, file_pos, 0);
}


/* Actually reads both the header and the aux header */
static int32
ctdb_reread_hdr_6(char *fname, CTDB_6 *ctdb)
{
    CTDB_FILE_HEADER_6 hdr;
    CTDB_AUX_FILE_HEADER_6 aux_hdr;
    FILE *f;
    
    if (!fname)
      return 0;
    
    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	return 0;
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	return 0;
    }

    ctdb_unpack_header_6(ctdb, &hdr);

    if (fread(&aux_hdr, hdr.aux_hdr_bytes, 1, f) != 1)
    {	
	printf("CTDB: Error reading %s AUX header.\n",fname);
	fclose(f);
	return 0;
    }

    ctdb_unpack_aux_header_6(ctdb, &aux_hdr);
    return 1;
}


static void
ctdb_reread_6(char *fname, CTDB_6 *ctdb)
{
    FILE *f;
    CTDB_FILE_HEADER_6 hdr;
    CTDB_AUX_FILE_HEADER_6 aux_hdr;
    int32 total_mem;

    if (!fname)
    {
	printf("No terrain database name specified.\n");
	exit(-1);	
    }

    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	exit(-1);
    }

    if (hdr.format != 6)
    {
	printf("CTDB: Attempted to read fmt %d database using fmt 6 reader",
	       hdr.format);
	fclose(f);
	exit(-1);
    }

    ctdb->file = f;

    ctdb_unpack_header_6(ctdb, &hdr);

    /* Read in the auxiliary header info */
    if (fread(&aux_hdr, hdr.aux_hdr_bytes, 1, f) != 1)
    {
	printf("CTDB: Error reading %s AUX header.\n", fname);
	fclose(f);
	exit(-1);
    }
    ctdb_unpack_aux_header_6(ctdb, &aux_hdr);

    read_patch_group_headers_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_topology_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_quad_tree_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_linear_models_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_volume_models_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_aggregate_models_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_mes_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_soil_tables_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_water_chars_6(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_tidal_zones_6(ctdb, f);

    if ( ctdb->is_gridded )
      ctdb_read_elevations_6(ctdb, f);
    
    ctdb->patch_data_seek = ftell(f);

    /* Since we're in the compiler, we're not caching. */
    ctdb_read_all_patch_data_6(ctdb, f);

    /* We'll find the z-limits later (once we've converted to the compiler
     * representation.
     */

    return;
}

void *
ctdb_read_6(fname)
    char *fname;
{
    CTDB_6 *ctdb;
    
    ctdb = (CTDB_6 *)ctdb_alloc(NULL, sizeof(CTDB_6), "Creating fmt 6 ctdb");
    bzero(ctdb, sizeof(CTDB_6));

    printf("Attempting to read CTDB datafiles... \n");

    ctdb_reread_6(fname, ctdb);
    return ctdb;
}

/***************************************************************************
 * WRITE ROUTINES                                                           *
 ***************************************************************************/

/* The input database is a void pointer to make it easier to call this routine
 * through a generic interface; it is the caller's responsibility to ensure
 * that it is, in fact, a format 6 CTDB.
 * NOTE: since the compiler does not use caching, the inter-format conversion
 * routines do NOT generate the caching scheme correctly; therefore it is the
 * responsibility of the write routine to do so.
 */
void ctdb_write_6(fmt6_db, output_file)
    void *fmt6_db;
    FILE *output_file;
{
    CTDB_6 *ctdb = (CTDB_6 *)fmt6_db;
    CTDB_FILE_HEADER_6 hdr;
    CTDB_AUX_FILE_HEADER_6 aux_hdr;

    int32 groups = (ctdb->patches_wide *
		    ctdb->patches_high)/CTDB_PATCHES_PER_GROUP;

    int32 max_cache = 0;

    /* Generate the caching scheme for the patch data and fill in the
     * necessary data structures
     */
    generate_caching_scheme_6(ctdb);

    /**
     ** Write the header to the output file.
     **/
    
    ctdb_pack_header_6(ctdb, &hdr);
    fwrite(&hdr, sizeof(hdr), 1, output_file);

    /**
     ** Write AUX header to output file
     **/
    ctdb_pack_aux_header_6(ctdb, &aux_hdr);
    fwrite(&aux_hdr, sizeof(aux_hdr), 1, output_file);

    /**
     ** Write patch group headers
     **/
    write_patch_group_headers_6(ctdb, ctdb->patch_groups, groups, output_file);

    /**
     ** Write topology: the nodes and edges to the output file.
     **/
    write_nodes_and_edges_6(ctdb, output_file);

    /** 
     ** Write quad tree: the abstract feature quad tree to the output
     **          file.
     **/
    write_abstract_features_6(ctdb, output_file);

    /** 
     ** Write linear models: the models referenced by linear features
     **          to the output file.
     **/
    write_linear_models_6(ctdb, output_file);

    /** 
     ** Write volume models: the models referenced by volume features
     **          to the output file.
     **/
    write_volume_models_6(ctdb, output_file);

    /** 
     ** Write aggregate models: the models referenced by aggregate features
     **          to the output file.
     **/
    write_aggregate_models_6(ctdb, output_file);

    /**
     ** Write mes structures: the structures refernced by mes features to
     **         the output file.
     **/
    write_mes_6(ctdb, output_file);

    /** 
     ** Write soil tables: the tables mapping local soil values to global ones
     ** to the output file.
     **/
    write_soil_tables_6(ctdb, output_file);

    /** 
     ** Write water chars: the sets of characteristics used for water areas
     ** to the output file.
     **/
    write_water_chars_6(ctdb, output_file);

    /** 
     ** Write tidal zones: the structures defining tidal variations to the
     ** output file.
     **/
    write_tidal_zones_6(ctdb, output_file);    

    /**
     ** Write elevation Posts: the elevations if there are gridded patches.
     **/
    if ( ctdb->is_gridded )
      write_elevation_posts_6(ctdb, output_file);
    
    /**
     ** Write patch group data: the patch group pointers, features, terrain
     ** elements, if tinned; the virtual grids, if tinned. 
     **/
    write_patch_data_6(ctdb, output_file);

    return;
}

/* The ctdb_pack_header_6 routine will create a file header (hdr) from the
 * fmt 6 ctdb data structure. 
 */
static void 
ctdb_pack_header_6(ctdb, hdr)
    CTDB_6             *ctdb;
    CTDB_FILE_HEADER_6 *hdr;
{
    bzero(hdr, sizeof(CTDB_FILE_HEADER_6));
    hdr->format = ctdb->input_format;
    hdr->is_curved = ctdb->is_curved;
    hdr->is_gridded = ctdb->is_gridded;
    hdr->is_tinned  = ctdb->is_tinned;
    hdr->fixed_point_basis = ctdb->fixed_point_basis;

    hdr->num_soil_tables = ctdb->num_soil_tables;
    hdr->soil_table_size = ctdb->soil_table_size;
    hdr->num_water_chars = ctdb->num_water_chars;
    hdr->num_tidal_zones = ctdb->num_tidal_zones;

    hdr->post_increment = ctdb->incr;
    hdr->posts_per_patch_side = ctdb->patch_incr;
    hdr->pages_wide = ctdb->pages_wide;
    hdr->pages_high = ctdb->pages_high;
    hdr->min_x = ctdb->min_x;
    hdr->max_x = ctdb->max_x;
    hdr->min_y = ctdb->min_y;
    hdr->max_y = ctdb->max_y;
    bcopy(ctdb->patch_cache_size_map, hdr->patch_cache_size_map,
	  sizeof(ctdb->patch_cache_size_map));

    hdr->origin_northing = ctdb->origin_northing;
    hdr->origin_easting = ctdb->origin_easting;
    hdr->origin_zone_number = ctdb->origin_zone_number;
    hdr->origin_zone_letter = ctdb->origin_zone_letter;
    hdr->datum = ctdb->datum;
    bcopy(ctdb->name, hdr->name, sizeof(ctdb->name));
    hdr->version = ctdb->version;
    bcopy(ctdb->date, hdr->date, sizeof(ctdb->date));
    hdr->num_nodes = ctdb->num_nodes;
    hdr->node_size = ctdb->node_size;
    hdr->num_edges = ctdb->num_edges;
    hdr->edge_size = ctdb->edge_size;
    hdr->num_quad_data = ctdb->num_quad_data;
    hdr->n_patch_bytes = ctdb->n_patch_bytes;
    hdr->num_linear_models = ctdb->num_linear_models;
    hdr->num_volume_models = ctdb->num_volume_models;

    hdr->num_aggregate_models = ctdb->num_aggregate_models;

    hdr->num_mes_volumes = ctdb->num_mes_volumes;
    hdr->num_mes_templates = ctdb->num_mes_templates;

    hdr->aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER_6);
}

/* Generate the caching scheme for the patch data and fill in the
 * necessary data structures
 */
static void
generate_caching_scheme_6(ctdb)
    CTDB_6 *ctdb;
{
    int32 i;
    int32 group;
    int32 num_groups;
    int32 max_size = 0;
    CTDB_PATCH_GROUP_6 *pg;
    int32 *pg_size_map;		/* pg_size_map[i] is the size, in words, of
				 * patch group i
				 */
    int32 *map;			/* Current entry in pg_size_map array */
    int32 *counts;		/* counts[i] is the number of patch groups
				 * with i words of data
				 */
    int32 size;			/* Current PG size we're looking at */
    int32 last_size;		/* Last non-empty PG size we looked at */
    int32 *sizes = ctdb->patch_cache_size_map; /* Sizes of the patch caches */
    int32 burdens[CTDB_PATCH_CACHES_6];	/* Number of PG's mapped to each 
					 * cache 
					 */
    int32 n_mapped;		/* Ideal number of groups per cache */
    int32 cur_count;		/* Groups mapped to current cache so far */
    int32 cur_cache;		/* Which cache we're currently filling */
    int32 min_burden;		/* Number of PGs mapped to least used cache */
    int32 smallest;
    int32 max_burden;		/* Number of PGs mapped to most used cache */
    int32 largest;		/* Most used cache */
    int32 best_cache;		/* "Best" cache for current PG */

    num_groups = (ctdb->patches_wide * ctdb->patches_high) /
      CTDB_PATCHES_PER_GROUP_6;

    pg_size_map = (int32 *)ctdb_alloc(NULL, num_groups * sizeof(int32),
				      "Patch group size map");


    /* 1.  Find the size of each patch group.  While we're at it, keep track
     * of how big the largest group is.
     */
    for(group = 0; group < num_groups; group++)
    {
	pg_size_map[group] = find_pg_size_6(ctdb->patch_groups + group) / 
	  CTDB_CACHE_BKT_UNITS_6;
	if (pg_size_map[group] > max_size)
	  max_size = pg_size_map[group];
    }

    max_size++;

    counts = (int32 *)ctdb_alloc(NULL, sizeof(int32) * max_size,
				 "patch group size counts");

    bzero(counts, max_size * sizeof(int32));
    bzero(sizes, CTDB_PATCH_CACHES_6 * sizeof(int32));
    bzero(burdens, CTDB_PATCH_CACHES_6 * sizeof(int32));

    /* 2. Find distribution of group sizes. 
     */
    map = pg_size_map;
    for(group=0;group<num_groups;group++,map++)
    {
	counts[*map]++;
    }

    /* 3. Choose the ideal number of groups to be mapped to each cache.
     */
    n_mapped =
      ceil((float64)num_groups/(float64)CTDB_PATCH_CACHES_6) + 1;

    printf("Ideal groups/cache = %d\n",n_mapped);

    /* 4. Choose cache sizes such that AT LEAST n_mapped groups land
     * in each cache.
     */
    cur_count = cur_cache = 0;
    for(size=0;size<max_size;size++)
    {
	if (counts[size])
	{
	    cur_count += counts[size];
	    last_size = size;
	}

	if (cur_count > n_mapped)
	{
	    do
	    {
		sizes[cur_cache] = size;
		if (cur_count >= 2*n_mapped)
		  burdens[cur_cache] = n_mapped;
		else
		  burdens[cur_cache] = cur_count;
		cur_cache++;
		cur_count -= n_mapped;
	    } while (cur_count >= n_mapped);
	    cur_count = 0;
	}
    }
    sizes[cur_cache] = last_size;
    burdens[cur_cache] = cur_count;
    cur_cache++;

    /* 5. If not all the caches were used, split the most burdened
     * caches in two.
     */
    while(cur_cache<CTDB_PATCH_CACHES_6)
    {
	/* Find the worst off */
	max_burden = largest = 0;
	for(i=0;i<cur_cache;i++)
	  if (burdens[i] > max_burden)
	  {
	      max_burden = burdens[i];
	      largest = i;
	  }
	/* Make room for the split */
	for(i=cur_cache;i>largest;i--)
	{
	    sizes[i] = sizes[i-1];
	    burdens[i] = burdens[i-1];
	}
	/* Split it in two */
	burdens[largest] /= 2;
	burdens[largest+1] -= burdens[largest];
	cur_cache += 1;
    }

    /* 6. Tag the groups with their cache number.
     */
    bzero(burdens,sizeof(burdens));
    for(group=0;group<num_groups;group++)
    {
	pg = ctdb->patch_groups + group;
	map = pg_size_map + group;

	/* Search through the caches and find which ones we can put this
	 * patch into.  Choose the least burdened of those.
	 */
	min_burden = smallest = 0x7FFFFFF;
	best_cache = CTDB_PATCH_CACHES_6 - 1;
	for(cur_cache=0; cur_cache < CTDB_PATCH_CACHES_6; cur_cache++)
	{
	    /* Have we passed the range of caches for which this size
	     * is intended?
	     */
	    if (sizes[cur_cache] > smallest)
	      break;
	      
	    /* Will the patch fit in this cache? */
	    if (sizes[cur_cache] < *map)
	      continue;
	    smallest = sizes[cur_cache];
	    if (burdens[cur_cache] < min_burden)
	    {
		best_cache = cur_cache;
		min_burden = burdens[cur_cache];
	    }
	}
	burdens[best_cache]++;
	pg->cache = best_cache;
    }

    /* 7. Print the results.
     */
    size = 0;
    for (cur_cache=0; cur_cache < CTDB_PATCH_CACHES_6; cur_cache++)
    {
	if (cur_cache && (sizes[cur_cache] != sizes[cur_cache-1]))
	  size = sizes[cur_cache-1];
	printf("Cache %x: [%d->%d].  %d patch groups\n",
	       cur_cache, size + 1, sizes[cur_cache], burdens[cur_cache]);
    }

    free(counts);    
    return;
}

/* Return the size, in bytes, of all data, including the data header, in
 * pg
 */
static int32
find_pg_size_6(pg)
    CTDB_PATCH_GROUP_6 *pg;
{
    int32 size;

    if (pg->contains_tins)
    {
	int32 mem = -1;
	int32 num_vg;
	CTDB_TIN_PG_DATA_6 *pgt = pg->data_start.tin_data;

	size = sizeof(CTDB_TIN_PG_DATA_6);

	if (!pg->patch3_no_tes)
	{
	    mem = 3;
	}
	else if (!pg->patch2_no_tes)
	{
	    mem = 2;
	}
	else if (!pg->patch1_no_tes)
	{
	    mem = 1;
	}
	else
	{
	    mem = 0;
	}

	/* This gives us the number of bytes from the start of the data header
	 * to the start of the last set of vgrid data
	 */
	size = pgt->te_vgrid_offset[mem] * sizeof(CTDB_FEATURE_DATA_6);

	num_vg = pgt->te_vgrid_dim[mem] * pgt->te_vgrid_dim[mem];

	/* Round up to word boundary, remembering that te_vgrid entries
	 * are uint16s
	 */
	num_vg += (num_vg % 2);

	size += (num_vg * sizeof(uint16));
	return size;
    }
    else
    {
	/* Find the last occupied patch.  Add its offset to the number of bytes
	 * of data stored for it to get the total PG data size
	 */
	CTDB_FEATURE_DATA_6 *feat = NULL;

	if (pg->patch3_occupied)
	{
	    size = pg->data_start.grid_data->patch3_offset *
	      sizeof(CTDB_FEATURE_DATA_6);
	    feat = (CTDB_FEATURE_DATA_6 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch3_offset;
	}
	else if (pg->patch2_occupied)
	{
	    size = pg->data_start.grid_data->patch2_offset *
	      sizeof(CTDB_FEATURE_DATA_6);
	    feat = (CTDB_FEATURE_DATA_6 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch2_offset;
	}
	else if (pg->patch1_occupied)
	{
	    size = pg->data_start.grid_data->patch1_offset *
	      sizeof(CTDB_FEATURE_DATA_6);
	    feat = (CTDB_FEATURE_DATA_6 *)pg->data_start.grid_data + 
	      pg->data_start.grid_data->patch1_offset;
	}
	else if (pg->patch0_occupied)
	{
	    size = sizeof(CTDB_GRID_PG_DATA);

	    feat = (CTDB_FEATURE_DATA_6 *)pg->data_start.grid_data +
	      (sizeof(CTDB_GRID_PG_DATA_6) / sizeof(CTDB_FEATURE_DATA_6));
	}
	if (feat == NULL)
	  return sizeof(CTDB_GRID_PG_DATA_6);

	size += (feat->hdr.size * sizeof(CTDB_FEATURE_DATA_6 *));
	return size;
    }
}

/* The ctdb_pack_aux_header_6 routine will create an aux file header 
 * (aux_hdr) from the fmt. 6 ctdb data structure. 
 */
static void 
ctdb_pack_aux_header_6(ctdb, aux_hdr)
    CTDB_6                 *ctdb;
    CTDB_AUX_FILE_HEADER_6 *aux_hdr;
{
    aux_hdr->gcs_mode = ctdb->gcs_mode;
    aux_hdr->cell_id = ctdb->cell_id;
    aux_hdr->origin_x = ctdb->origin_x;
    aux_hdr->origin_y = ctdb->origin_y;
    aux_hdr->west_long = ctdb->west_long;
    aux_hdr->east_long = ctdb->east_long;
    aux_hdr->south_lat = ctdb->south_lat;
    aux_hdr->north_lat = ctdb->north_lat;

    return;
}

/* This routine writes the complete set of patch group headers. 
 * $$$ WARRNING:  as written, this changes the data_start values to be
 * offsets rather than pointers, rendering the database unusable in memory.
 * If you need to be able to use the database after writing it, this routine
 * will need to change
 */
static void
write_patch_group_headers_6(ctdb, patch_groups, groups, output_file)
    CTDB_6             *ctdb;
    CTDB_PATCH_GROUP_6 *patch_groups;
    int32               groups;
    FILE               *output_file;
{
    int32 i;

    for(i=0;i<groups;i++)
    {
	ctdb->patch_groups[i].data_start.offset = 
	  (int8 *)ctdb->patch_groups[i].data_start.grid_data -
	  (int8 *)ctdb->patch_data;
    }
	
    fwrite(patch_groups, sizeof(CTDB_PATCH_GROUP_6),groups, output_file);    
}

/* This function writes the nodes and edges to the output file.
 */
static void 
write_nodes_and_edges_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    /* Just dump the stored nodes and edges on out */
    if (ctdb->node_size)
      fwrite(ctdb->node_storage, 1, ctdb->node_size, output_file);
    if (ctdb->edge_size)
      fwrite(ctdb->edge_storage, 1, ctdb->edge_size, output_file);
}

/* This function writes the abstract features to the output file. */
static void 
write_abstract_features_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    /* Just dump the quad tree on out */
    if (ctdb->num_quad_data)
      fwrite(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA_6),
	     ctdb->num_quad_data, output_file);
}

/* This function writes the linear models to the output file.
 */
static void 
write_linear_models_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    /* Just dump the linear models on out */
    if (ctdb->num_linear_models)
      fwrite(ctdb->linear_models, sizeof(CTDB_LINEAR_MODEL_6),
	     ctdb->num_linear_models, output_file);
}

/* This function writes the volume models to the output file.
 */
static void
write_volume_models_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    int32              i, j;

    /* Just dump the volume models on out */
    for(i = 0; i < ctdb->num_volume_models; i++)
    {
	/* Write the model, excluding the vertices, to the file.
	 */
	fwrite(ctdb->volume_models[i], sizeof(CTDB_VOLUME_MODEL_6) -
	       sizeof(float32), 1, output_file);

	/* Now loop through and write each vertex to the file.
	 */
	for(j = 0; j < ctdb->volume_models[i]->num_vertices; j++)
	{
	    fwrite(&ctdb->volume_models[i]->vertices[3 * j],
		   sizeof(float32), 3, output_file);
	}
    }
}

/* This function writes the aggregate models to the output file.
 */
static void
write_aggregate_models_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    int32              i, j;

    /* Just dump the aggregate models on out */
    for(i = 0; i < ctdb->num_aggregate_models; i++)
    {
	/* Write the model, excluding the features, to the file.
	 */
	fwrite(ctdb->aggregate_models[i], sizeof(CTDB_AGGREGATE_MODEL_6) -
	       sizeof(CTDB_AGG_MODEL_FEATURE_6), 1, output_file);

	/* Now loop through and write each feature to the file.
	 */
	for(j = 0; j < ctdb->aggregate_models[i]->n_features; j++)
	{
	    fwrite(&ctdb->aggregate_models[i]->features[j],
		   sizeof(CTDB_AGG_MODEL_FEATURE_6), 1, output_file);
	}
    }
}

#define SWAP(entity) (byte_swap(&entity, sizeof(entity)))

/*
** Procedure: byte_swap
** Purpose:   Reverse order of bytes in situ.
*/
static void byte_swap(p, size)
    uint8 *p;
    int32 size;
{
    uint8 n;
    int32 i, j;

    for(i=0,j=size-1; i<j; i++,j--)
    {
        n = p[i];
        p[i] = p[j];
        p[j] = n;
    }
}


/* This function writes the MES volumes (which instantiate MES templates)
 * to the output file.  If little_endian is TRUE, all generated offsets
 * will be byte-swapped before being written out if we are running on 
 * a big-endian machine
 */
static void
write_mes_volumes_6(ctdb, output_file, little_endian)
    CTDB_6 *ctdb;
    FILE   *output_file;
    int32   little_endian;
{
    CTDB_MES_VOLUME_6        temp_vol, *curr_vol;
    CTDB_MES_LIST_VERTEX_F64 *roofline, floorline;
    int32                     i;
    int32                     roof_verts, floor_verts;
    
    /* We need to turn pointers into offsets, so we'll copy over the
     * volumes
     */

    for(i=0;i<ctdb->num_mes_volumes;i++)
    {
	curr_vol = ctdb->mes_volumes + i;
	bcopy(curr_vol, &temp_vol, sizeof(CTDB_MES_VOLUME_6));
    	temp_vol.template = (CTDB_MES_TEMPLATE_6 *)
	  (temp_vol.template - ctdb->mes_templates);

	roof_verts = temp_vol.roofline.num_vertices;
	floor_verts = temp_vol.floorline.num_vertices;

	if (little_endian)
	{
#ifndef USE_LITTLE_ENDIAN
	    SWAP(temp_vol.template);
	    
	    /* These fields will have been flipped for little-endian storage,
	     * so we'll have to flip them back in order to use them
	     */
	    SWAP(roof_verts);
	    SWAP(floor_verts);
#endif /* USE_LITTLE_ENDIAN */
	}
	
	temp_vol.roofline.vertices = NULL;
	temp_vol.floorline.vertices = NULL;

	fwrite(&temp_vol, sizeof(CTDB_MES_VOLUME_6), 1, output_file);

	fwrite(curr_vol->roofline.vertices, 
	       roof_verts * sizeof(CTDB_MES_VERTEX_F64),
	       1, output_file);

	fwrite(curr_vol->floorline.vertices, 
	       floor_verts * sizeof(CTDB_MES_VERTEX_F64),
	       1, output_file);
    }
    
    return;
}

/* This function writes the list of mes volume templates to the output file 
 * If little_endian is TRUE, all generated offsets
 * will be byte-swapped before being written out if we are running on 
 * a big-endian machine
 */
static void
write_mes_templates_6(ctdb, output_file, little_endian)
    CTDB_6 *ctdb;
    FILE   *output_file;
    int32   little_endian;
{
    CTDB_MES_TEMPLATE_6  temp_templ, *cur_templ;
    CTDB_MES_ENCLOSURE_6 temp_enclosure, *cur_encl;
    CTDB_MES_APERTURE_6  temp_aperture;
    CTDB_MES_GRID_BOX_6  temp_box;

    int32 i,j; /* for loop indices */

    int32 boxes; /* use this to check if any boxes exist */

    int32 n_vert, n_tri, n_encl, n_ap, n_box;
    int32 bx, by, bz;

    uint32 *temp_int;

    for(i=0; i < ctdb->num_mes_templates; i++)
    {
	cur_templ = ctdb->mes_templates + i;

	bcopy(cur_templ, &temp_templ, sizeof(CTDB_MES_TEMPLATE_6));

	temp_templ.all_vertices = NULL;
	temp_templ.all_triangles = NULL;
	temp_templ.all_enclosures = NULL;
	temp_templ.all_apertures = NULL;
	temp_templ.all_boxes = NULL;

	fwrite(&temp_templ, sizeof(temp_templ), 1, output_file);

	n_vert = cur_templ->num_vertices;
	n_tri  = cur_templ->num_triangles;
	n_encl = cur_templ->num_enclosures;
	if (little_endian)
	{
#ifndef USE_LITTLE_ENDIAN
	    SWAP(n_vert);
	    SWAP(n_tri);
	    
	    /* Convert n_encl the hard way, because it's 16 bits */
	    temp_int = (uint32 *)(&cur_templ->all_triangles + 1);
	    SWAP(*temp_int);
	    n_encl = cur_templ->num_enclosures;
	    SWAP(*temp_int);
#endif /* USE_LITTLE_ENDIAN */
	}

	fwrite(cur_templ->all_vertices,
	       sizeof(int32)*XYZ*n_vert,
	       1, output_file);

	fwrite(cur_templ->all_triangles,
	       sizeof(CTDB_MES_TRIANGLE_6)*n_tri,
	       1, output_file);

	for(j=0; j < n_encl; j++)
	{
	    cur_encl = cur_templ->all_enclosures + j;
	    bcopy(&cur_templ->all_enclosures[j], &temp_enclosure,
		  sizeof(CTDB_MES_ENCLOSURE_6));

	    temp_enclosure.footprint.vertex_ids = NULL;
	    temp_enclosure.aperture_ids = NULL;

	    fwrite(&temp_enclosure, sizeof(CTDB_MES_ENCLOSURE_6), 1, 
		   output_file);

	    n_vert = cur_encl->footprint.num_vertices;
	    n_ap   = cur_encl->num_apertures;
	    
	    if (little_endian)
	    {
#ifndef USE_LITTLE_ENDIAN
		SWAP(n_vert);

		temp_int = (uint32 *)(&cur_encl->first_wall_triangle_id + 1);
		SWAP(*temp_int);
		n_ap = cur_encl->num_apertures;
		SWAP(*temp_int);
#endif /* USE_LITTLE_ENDIAN */
	    }	    

	    fwrite(cur_encl->footprint.vertex_ids,
		   sizeof(int32) * n_vert,
		   1, output_file);

	    /* Round up to nearest even number to avoid writing out fractional
	     * words
	     */
	    n_ap = n_ap + (n_ap % 2);

	    fwrite(cur_encl->aperture_ids,
		   sizeof(uint16) * n_ap,
		   1, output_file);
	}

	n_ap = cur_templ->num_apertures;
	if (little_endian)
	{
#ifndef USE_LITTLE_ENDIAN
	    temp_int = (uint32 *)(&cur_templ->all_enclosures + 1);
	    SWAP(*temp_int);
	    n_ap = cur_templ->num_apertures;
	    SWAP(*temp_int);
#endif /* USE_LITTLE_ENDIAN */
	}
	
	for(j=0; j < n_ap; j++)
	{
	    bcopy(&cur_templ->all_apertures[j], &temp_aperture,
		  sizeof(CTDB_MES_APERTURE_6));

	    temp_aperture.outline.vertex_ids = NULL;

	    fwrite(&temp_aperture, sizeof(CTDB_MES_APERTURE_6),
		   1, output_file);

	    n_vert = cur_templ->all_apertures[j].outline.num_vertices;

	    if (little_endian)
	    {
#ifndef USE_LITTLE_ENDIAN
		SWAP(n_vert);
#endif /* USE_LITTLE_ENDIAN */
	    }

	    fwrite(cur_templ->all_apertures[j].outline.vertex_ids,
		   sizeof(int32) * n_vert,
		   1, output_file);
	}

	n_box =  cur_templ->num_boxes;

	if (little_endian)
	{
#ifndef USE_LITTLE_ENDIAN
	    SWAP(n_box);
#endif /* USE_LITTLE_ENDIAN */
	}

	for(j=0; j < n_box; j++)
	{
	    bcopy(&cur_templ->all_boxes[j], &temp_box,
		  sizeof(CTDB_MES_GRID_BOX_6));

	    temp_box.box_ids = NULL;
	    temp_box.aperture_ids = NULL;
	    temp_box.wall_triangle_ids = NULL;

	    fwrite(&temp_box, sizeof(CTDB_MES_GRID_BOX_6), 1 , output_file);

	    bx = cur_templ->all_boxes[j].dimensions_x;
	    by = cur_templ->all_boxes[j].dimensions_y;
	    bz = cur_templ->all_boxes[j].dimensions_z;
	    n_ap = cur_templ->all_boxes[j].num_apertures;
	    n_tri = cur_templ->all_boxes[j].num_triangles;

	    if (little_endian)
	    {
#ifndef USE_LITTLE_ENDIAN
		temp_int = (uint32 *)
		  (&cur_templ->all_boxes[j].outer_box_id + 1);
		SWAP(*temp_int);
		bx = cur_templ->all_boxes[j].dimensions_x;
		SWAP(*temp_int);
		
		temp_int = (uint32 *)
		  (&cur_templ->all_boxes[j].outer_box_id + 2);
		SWAP(*temp_int);
		by = cur_templ->all_boxes[j].dimensions_y;
		bz = cur_templ->all_boxes[j].dimensions_z;
		SWAP(*temp_int);
		
		temp_int = (uint32 *)
		  (&cur_templ->all_boxes[j].box_ids + 1);
		SWAP(*temp_int);
		n_ap = cur_templ->all_boxes[j].num_apertures;
		SWAP(*temp_int);
		
		temp_int = (uint32 *)
		  (&cur_templ->all_boxes[j].aperture_ids + 1);
		SWAP(*temp_int);
		n_tri = cur_templ->all_boxes[j].num_triangles;
		SWAP(*temp_int);
#endif /* USE_LITTLE_ENDIAN */
	    }

	    /* lets see if there are any boxes inferior to this box */
	    boxes = bx * by * bz;
	    
	    if(boxes > 1)
	    {
		fwrite(cur_templ->all_boxes[j].box_ids,
		       sizeof(int32)*boxes, 1, output_file);
	    }

	    /* Round up to nearest even number to avoid writing out fractional
	     * words
	     */
	    n_ap = n_ap + (n_ap % 2);

	    fwrite(cur_templ->all_boxes[j].aperture_ids,
		   sizeof(uint16) * n_ap,
		   1, output_file);

	    fwrite(cur_templ->all_boxes[j].wall_triangle_ids,
		   sizeof(int32) * n_tri,
		   1, output_file);
	}
    } /* for(i=0; i < ctdb->num_mes_templates; i++) */
    printf("\nNumber of MES templates written to file %d\n", 
	   ctdb->num_mes_volumes);
}

/* This function writes all MES-related structures to the output file */
static void
write_mes_6(ctdb, output_file)
    CTDB_6 *ctdb;
    FILE   *output_file;
{
    write_mes_templates_6(ctdb, output_file, FALSE);
    write_mes_volumes_6(ctdb, output_file, FALSE);
    return;
}

/* Writes just the MES models (file must be positioned at start of MES model
 * data) from the provided CTDB.  If little_endian is TRUE, the offsets
 * generated will be byte-swapped before being written if the compiler
 * is running on a big-endian machine.
 */
void
ctdb_write_just_mes_6(CTDB_6 *ctdb,
		      FILE   *output_file,
		      int32   little_endian)
{
    write_mes_templates_6(ctdb, output_file, little_endian);
    write_mes_volumes_6(ctdb, output_file, little_endian);
    return;
}

/* This function writes the soil tables to the output file */
static void
write_soil_tables_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    int32 *offsets;
    int32 i;

    if (ctdb->num_soil_tables)
    {
	/* First, write out the offsets for the soil tables.  These are 
	 * kept as pointers in memory, but stored as byte offsets on disk
	 */
	offsets = (int32 *)ctdb_alloc(NULL, 
				      ctdb->num_soil_tables * sizeof(int32),
				      "Soil table offsets");

	for(i=0;i<ctdb->num_soil_tables;i++)
	{
	    offsets[i] = (int8 *)ctdb->soil_tables[i] - 
	      ctdb->soil_table_storage;
	}

	fwrite(offsets, sizeof(int32), ctdb->num_soil_tables, output_file);
	
	/* Now, write out the tables themselves */
	fwrite(ctdb->soil_table_storage, sizeof(int8), ctdb->soil_table_size, 
	       output_file);
	
	STDDEALLOC(offsets);
    }	    
}

/* This function writes the water characteristics to the output file */
static void
write_water_chars_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    if (ctdb->num_water_chars)
      fwrite(ctdb->water_chars, sizeof(CTDB_WATER_CHARS_6),
	     ctdb->num_water_chars, output_file);    
}


/* This function writes the tidal zones to the output file */
static void
write_tidal_zones_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    if (ctdb->num_tidal_zones)
      fwrite(ctdb->tidal_zones, sizeof(CTDB_TIDAL_ZONE_6),
	     ctdb->num_tidal_zones, output_file);    
}

/* This function writes the elevation post data out to the data file.
   Note that this assumes all elevation posts are already in memory.
 */
static void 
write_elevation_posts_6(ctdb, output_file)
    CTDB_6  *ctdb;
    FILE    *output_file;
{
    int32 post_bytes;
    
    post_bytes = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
        
    fwrite(ctdb->elevations, post_bytes,1, output_file);
}

/* Writes the patch data out to the file */
static void
write_patch_data_6(ctdb, output_file)
    CTDB_6             *ctdb;
    FILE               *output_file;
{
    int32 i;
    int8 pad_byte = 0;
    int32 num_pad_bytes;

    /* Dump the patch data on out */
    fwrite(ctdb->patch_data, 1, ctdb->n_patch_bytes, output_file);

    /* To be perfectly safe, we'll pad the database out by the size of the
     * largest cache; this will ensure that the caching scheme does not
     * read past the end of the database.
     */
    num_pad_bytes = ctdb->patch_cache_size_map[CTDB_PATCH_CACHES_6 - 1] *
      CTDB_CACHE_BKT_UNITS_6;

    for(i = 0; i<num_pad_bytes; i++)
      fwrite(&pad_byte, sizeof(int8), 1, output_file);
    
    return;
}

/***************************************************************************
 * FMT 6 FREE ROUTINE                                                      *
 ***************************************************************************/

void
ctdb_free_fmt6_db(db)
    void *db;
{
    int32 i;
    CTDB_6 *ctdb = (CTDB_6 *)db;

#define CTDB_FREE_6(x)       \
    if (ctdb->x)             \
        STDDEALLOC(ctdb->x)

    if (!ctdb)
      return;
    
    CTDB_FREE_6(soil_tables);
    CTDB_FREE_6(soil_table_storage);
    CTDB_FREE_6(water_chars);
    CTDB_FREE_6(tidal_zones);
    
    CTDB_FREE_6(patch_groups);
    CTDB_FREE_6(cache_page_numbers);
    CTDB_FREE_6(elevation_pages);

    for(i = 0; i < CTDB_PATCH_CACHES_6; i++)
      CTDB_FREE_6(patch_data_cache[i]);

    CTDB_FREE_6(elevations);
    CTDB_FREE_6(quad_data);
    CTDB_FREE_6(nodes);
    CTDB_FREE_6(node_storage);
    CTDB_FREE_6(edges);
    CTDB_FREE_6(edge_storage);
    CTDB_FREE_6(linear_models);

    for(i = 0; i < ctdb->num_volume_models; i++)
      CTDB_FREE_6(volume_models[i]);
    CTDB_FREE_6(volume_models);

    for(i = 0; i < ctdb->num_aggregate_models; i++)
      CTDB_FREE_6(aggregate_models[i]);
    CTDB_FREE_6(aggregate_models);
    
    CTDB_FREE_6(patch_data);

    STDDEALLOC(ctdb);
    return;
}


/***************************************************************************
 * 6 -> 7 CONVERSION                                                       *
 ***************************************************************************/

void *
ctdb_convert_6_to_7(fmt6_db)
    void *fmt6_db;
{
    int32   i;
    CTDB_6 *ctdb = (CTDB_6 *)fmt6_db;
    CTDB_7 *ctdb_out;

    ctdb_out = (CTDB_7 *)ctdb_alloc(NULL, sizeof(CTDB_7),
				    "Fmt. 7 database");
    
    bzero(ctdb_out, sizeof(CTDB_7));

    /* First, copy the easy stuff over */
#define COPY_6_7(x) (ctdb_out->x = ctdb->x)

    ctdb_out->fixed_point_basis = CTDB_STANDARD_FIXED_POINT_BASIS_7;
    ctdb_out->inv_fixed_point_basis = 1.0 / ctdb_out->fixed_point_basis;

    COPY_6_7(num_soil_tables);
    COPY_6_7(soil_table_size);
    COPY_6_7(num_water_chars);
    COPY_6_7(num_tidal_zones);
    COPY_6_7(incr);
    COPY_6_7(inv_incr);
    COPY_6_7(incrD);

    COPY_6_7(patch_incr);
    COPY_6_7(patch_inv_incr);
    COPY_6_7(patch_incrD);

    COPY_6_7(pages_wide);
    COPY_6_7(pages_high);
    COPY_6_7(patches_wide);
    COPY_6_7(patches_high);
    COPY_6_7(min_x);
    COPY_6_7(max_x);
    COPY_6_7(min_y);
    COPY_6_7(max_y);
    COPY_6_7(min_x_post);
    COPY_6_7(max_x_post);
    COPY_6_7(min_y_post);
    COPY_6_7(max_y_post);
    
    ctdb_out->input_format = 7;
    
    COPY_6_7(min_z);
    COPY_6_7(max_z);
    COPY_6_7(origin_northing);
    COPY_6_7(origin_easting);
    COPY_6_7(origin_zone_number);
    COPY_6_7(origin_zone_letter);
    COPY_6_7(datum);
    COPY_6_7(cell_id);
    
    COPY_6_7(origin_x);
    COPY_6_7(origin_y);
    
    COPY_6_7(west_long);
    COPY_6_7(east_long);
    COPY_6_7(south_lat);
    COPY_6_7(north_lat);

    COPY_6_7(gcs_mode);

    COPY_6_7(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_6_7(is_curved);

    ctdb_out->elevations_resident = TRUE;

    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_6_7(num_quad_data);
    COPY_6_7(num_nodes);
    COPY_6_7(node_size);
    COPY_6_7(num_edges);
    COPY_6_7(edge_size);
    COPY_6_7(num_linear_models);
    COPY_6_7(num_volume_models);

    COPY_6_7(meters_to_patch);
    COPY_6_7(is_gridded);
    COPY_6_7(is_tinned);
    COPY_6_7(max_patch_number);
    COPY_6_7(min_patch_number);
    
    /* Now for the real work */

    /* Soil tables */
    if (!ctdb->num_soil_tables)
    {
	ctdb_out->soil_table_storage = NULL;
	ctdb_out->soil_tables = NULL;
	ctdb_out->num_soil_tables = 0;
	ctdb_out->soil_table_size = 0;
        ctdb_out->pat_table_ptr = NULL;
    }
    else
    {
	ctdb_out->num_soil_tables = ctdb->num_soil_tables;
	ctdb_out->soil_tables =
	  (int32 **)ctdb_alloc(NULL, ctdb->num_soil_tables *
			       sizeof(int32 *),
			       "Fmt. 7 soil table ptrs");
	ctdb_out->soil_table_size = ctdb->soil_table_size;
	ctdb_out->soil_table_storage = 
	  (int8 *)ctdb_alloc(NULL, ctdb_out->soil_table_size,
			     "Fmt. 7 soil table storage");

	regen_table_ptrs((int8 *)ctdb->soil_table_storage, 
			 (int8 **)ctdb->soil_tables,
			 (int8 *)ctdb_out->soil_table_storage, 
			 (int8 **)ctdb_out->soil_tables,
			 ctdb_out->num_soil_tables);
    }

    /* Construct the Polygon Attribute Table */
    convert_soils_6_7(ctdb, ctdb_out);

    /* Copy the tidal zones */
    if (!ctdb_out->num_tidal_zones)
    {
	ctdb_out->tidal_zones = NULL;
	ctdb_out->num_tidal_zones = 0;
    }
    else
    {
	ctdb_out->num_tidal_zones = ctdb->num_tidal_zones;
	ctdb_out->tidal_zones = (CTDB_TIDAL_ZONE_7 *)
	  ctdb_alloc(NULL, sizeof(CTDB_TIDAL_ZONE_7) * ctdb->num_tidal_zones,
		     "Fmt. 7 tidal zones.");
	bcopy(ctdb->tidal_zones, ctdb_out->tidal_zones,
	      sizeof(CTDB_TIDAL_ZONE_7) * ctdb_out->num_tidal_zones);
    }

    /* Copy the water char entries */
    if (!ctdb->num_water_chars)
    {
	ctdb_out->water_chars = NULL;
	ctdb_out->num_water_chars = 0;
    }
    else
    {
	ctdb_out->num_water_chars = ctdb->num_water_chars;
	ctdb_out->water_chars = (CTDB_WATER_CHARS_7 *)
	  ctdb_alloc(NULL, sizeof(CTDB_WATER_CHARS_7) * ctdb->num_water_chars,
		     "Fmt. 7 water chars.");
	bcopy(ctdb->water_chars, ctdb_out->water_chars, ctdb->num_water_chars *
	      sizeof(CTDB_WATER_CHARS_6));
    }

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));

    /* Patch features */
    {
	int32 groups = (ctdb_out->patches_wide * ctdb_out->patches_high) /
	  CTDB_PATCHES_PER_GROUP_7;

	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_7 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_7),
		     "Fmt. 7 patch groups");

	ctdb_out->n_patch_bytes = convert_features_6_7(ctdb, ctdb_out);
    }

    /* Elevations */
    convert_elevations_6_7(ctdb, ctdb_out);


    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_7 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data *
		     sizeof(CTDB_FEATURE_DATA_7),
		     "Fmt. 7 quadtree");
	
	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_6));

    }

    /* Topology */
    ctdb_out->node_storage = (int8 *)
      ctdb_alloc(NULL, ctdb->node_size, "Fmt. 7 nodes");
    bcopy(ctdb->node_storage, ctdb_out->node_storage, ctdb->node_size);
    
    ctdb_out->edge_storage = (int8 *)
      ctdb_alloc(NULL, ctdb->edge_size, "Fmt. 7 edges");
    bcopy(ctdb->edge_storage, ctdb_out->edge_storage, ctdb->edge_size);

    /* Linear models */
    if (!ctdb->num_linear_models)
    {
	ctdb_out->linear_models = NULL;
    }
    else
    {
	ctdb_out->linear_models = 
	  (CTDB_LINEAR_MODEL_7 *)ctdb_alloc(NULL,
					    sizeof(CTDB_LINEAR_MODEL_7) *
					    ctdb->num_linear_models,
					    "Fmt. 7 linear models");

	bcopy(ctdb->linear_models, ctdb_out->linear_models,
	      ctdb->num_linear_models * sizeof(CTDB_LINEAR_MODEL_6));
    }
    
    /* Volume models */

    if (!(ctdb->num_volume_models + ctdb->num_mes_volumes))
    {
	ctdb_out->volume_models = NULL;
	ctdb_out->mes_volume_models = NULL;	
    }
    else
    {
	int32 model_size;
	CTDB_VOLUME_MODEL_6 *old_model;
	CTDB_VOLUME_MODEL_7 *model;
	int32 i, j, index;
	int32 max_model_id = -1;
	CTDB_VOLUME_MODEL_7 **temp_volume_models_ptr;

	ctdb_out->volume_models = NULL;
	ctdb_out->mes_volume_models = NULL;
	temp_volume_models_ptr = (CTDB_VOLUME_MODEL_7 **)
	  ctdb_alloc(NULL,
		     sizeof(CTDB_VOLUME_MODEL_7 *) * 
		     (ctdb->num_mes_volumes + ctdb->num_volume_models),
		     "Fmt 7 volume model pointers");

	if (ctdb->num_volume_models >0) 
	{
	    ctdb_out->volume_models = temp_volume_models_ptr;

	    /* Copy over the volume models */
	    for(i=0; i<ctdb->num_volume_models; i++)
	    {
		old_model = ctdb->volume_models[i];

		if (old_model->model_id > max_model_id)
		  max_model_id = old_model->model_id;
	    
		model_size = sizeof(CTDB_VOLUME_MODEL_7) - sizeof(float32) +
		  old_model->num_vertices * 3 * sizeof(float32);

		ctdb_out->volume_models[i] = (CTDB_VOLUME_MODEL_7 *)
		  ctdb_alloc(NULL, model_size, "Fmt. 7 volume model");
	    
		bcopy(old_model, ctdb_out->volume_models[i], model_size);
	    }
	} 
	if (ctdb->num_mes_volumes) 
	{
	    if (ctdb->num_volume_models == 0)
	      ctdb_out->mes_volume_models = temp_volume_models_ptr;
	    else
	      ctdb_out->mes_volume_models =
		&(ctdb_out->volume_models[ctdb->num_volume_models]);
	    
	    /* Set the max_model_id to the next current model_id */
	    max_model_id++;

	    /* Allocate space for a temporary model (large) */
	    model = (CTDB_VOLUME_MODEL_7 *)
	      ctdb_alloc(NULL, (sizeof(CTDB_VOLUME_MODEL_7) +
			    1024 * 3 * sizeof(float32)),
		     "Temporary volume model");
	
	    /* Create volume models for MES volumes */   
	    for (i=0; i <ctdb->num_mes_volumes; i++)
	    {
		sprintf(model->model_name, "Mes #%d", i);
		model->model_id = max_model_id++;
		model->next_state_index = -1; /* since it's invalid */
		model->percent_damaged = 0.0;
		model->mes_volume_index = i;

		model->num_vertices = 
		  ctdb->mes_volumes[i].roofline.num_vertices;
		for (j=0; j < model->num_vertices; j++)
		{
		    model->vertices[3*j + X] =
		      ctdb->mes_volumes[i].roofline.vertices[j][X];
		    model->vertices[3*j + Y] =
		      ctdb->mes_volumes[i].roofline.vertices[j][Y];
		    model->vertices[3*j + Z] =
		  ctdb->mes_volumes[i].roofline.vertices[j][Z];
		}

		model_size = sizeof(CTDB_VOLUME_MODEL_7) - sizeof(float32) +
		  model->num_vertices * 3 * sizeof(float32);

		/* Now that the entire model has been converted, store it.
		 */
		ctdb_out->mes_volume_models[i] = 
		  (CTDB_VOLUME_MODEL_7 *)STDALLOC(model_size);
		bcopy(model, ctdb_out->mes_volume_models[i], model_size);
	    }
	    STDDEALLOC(model);
	}
    }  
	
   /* MES templates and volumes */
    if(!ctdb->num_mes_templates)
    {
        ctdb_out->mes_templates = NULL;
        ctdb_out->num_mes_templates = 0;
      }
    else
      {
        copy_mes_templates_6_7(ctdb, ctdb_out);
      }

    if(!ctdb->num_mes_volumes)
      {
        ctdb_out->mes_volumes = NULL;
        ctdb_out->num_mes_volumes = 0;
      }
    else
      {
        copy_mes_volumes_6_7(ctdb, ctdb_out);
      }

    return ctdb_out;
}

static void
convert_soils_6_7(CTDB_6 *ctdb_6, CTDB_7 *ctdb_7)
{
    int32 i, *soil_table_storage;
    CTDB_PAT_COLUMN_HDR_7 *column;
    CTDB_PAT_DATA_7       *data;


    ctdb_7->pat_table_ptr = (CTDB_PAT_TABLE_7 *)
      ctdb_alloc(NULL,
		 sizeof(CTDB_PAT_TABLE_7),
		 "Format 7 Polygon Attribute Table");
    ctdb_7->pat_table_ptr->alloced_entries =
      ctdb_7->pat_table_ptr->num_entries =
      ctdb_6->soil_table_size/sizeof(int32);
    ctdb_7->pat_table_ptr->num_columns = 4;
    ctdb_7->pat_table_ptr->columns = (CTDB_PAT_COLUMN_HDR_7 *)
      ctdb_alloc(NULL, ctdb_7->pat_table_ptr->num_columns *
		 sizeof(CTDB_PAT_COLUMN_HDR_7),
		 "Format 7 PAT column headers");

    column = ctdb_7->pat_table_ptr->columns;
    column[0].facc = CTDB_SIMNET_MOB_ATTR;
    column[1].facc = ctdb_lookup_attr_code( "STP" );
    column[2].facc = ctdb_lookup_attr_code( "SMC" );
    column[3].facc = ctdb_lookup_attr_code( "SWC" );

    for(i=0; i<4; i++)
    {
	column[i].data = (CTDB_PAT_DATA_7 *)
	  ctdb_alloc(NULL, ctdb_7->soil_table_size,
		     "Format 7 PAT data");
    }
	
    bcopy(ctdb_6->soil_table_storage, column[0].data,
	  ctdb_6->soil_table_size);
	
    for(i=0; i<ctdb_7->pat_table_ptr->num_entries; i++)
      {
	column[1].data[i].integer = CTDB_DEFAULT_STP;
	column[2].data[i].integer = CTDB_DEFAULT_SMC;
	if( CTDB_WATER( column[0].data[i].integer ) )
	  column[3].data[i].integer = CTDB_DEFAULT_WET_SWC;
	else
	  column[3].data[i].integer = CTDB_DEFAULT_DRY_SWC;
      }
	
    soil_table_storage = (int32 *) ctdb_7->soil_table_storage;
    for(i=0; i<ctdb_7->pat_table_ptr->num_entries; i++)
      soil_table_storage[i] = i;
}


static void
convert_elevations_6_7(CTDB_6 *ctdb, CTDB_7 *ctdb_out)
{
    int32 i, pages, elev_mem;

    if (!ctdb->is_gridded)
    {
	ctdb_out->elevations = NULL;
    }
    else
    {
	pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)ctdb_alloc(NULL, elev_mem, 
						   "Fmt. 7 elevation data");
        ctdb_out->soils      = (int32 *)ctdb_alloc(NULL, elev_mem,
						   "Format 7 soil grid data");

        for(i=0; i<elev_mem/4; i++)
	{
	    /* Strip out the soils */
	    ctdb_out->elevations[i] = ctdb->elevations[i] &
	      ~BUILD_POST_SW_SOIL_6(0xF) & ~BUILD_POST_NE_SOIL_6(0xF);
	    ctdb_out->soils[i] =
	      BUILD_POST_SW_SOIL_7(GET_POST_SW_SOIL_6(ctdb->elevations[i])) |
	      BUILD_POST_NE_SOIL_7(GET_POST_NE_SOIL_6(ctdb->elevations[i]));
	}
    }
    
}


static void
convert_micro_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{
   /* Format 7 Microterrain is described by:
     *      INFO MICRO_SOIL XY Z MICRO 
     * Format 6 Microterrain is described by:
     *      INFO XY Z MICRO 
     */  

    *next_6 = feat_6 + 4;
    *next_7 = feat_7 + 5;

    feat_7[0].info.feature_class = feat_6[0].info.feature_class;
    feat_7[0].info.vertices      = feat_6[0].info.vertices;
    feat_7[0].info.post_mask     = feat_6[0].info.post_mask;
    
    feat_7[1].micro_soil.poly_char_index0 = feat_6[3].micro.soil0;
    feat_7[1].micro_soil.poly_char_index1 = feat_6[3].micro.soil1;
    
    feat_7[2].xy.x = feat_6[1].xy.x;
    feat_7[2].xy.y = feat_6[1].xy.y;
    
    feat_7[3].z = feat_6[2].z;
    
    feat_7[4].micro.vert0 = feat_6[3].micro.vert0;
    feat_7[4].micro.vert1 = feat_6[3].micro.vert1;
    feat_7[4].micro.vert2 = feat_6[3].micro.vert2;
    
    return;    
}

static void
convert_mwater_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{  
    /* Format 7 Microwater is described by:
     *      INFO MICRO_SOIL XY Z MICRO MWATER_CHARS 
     * Format 6 Microwater is described by:
     *      INFO XY Z MICRO MWATER_CHARS
     */

    *next_6 = feat_6 + 5;
    *next_7 = feat_7 + 6;

    feat_7[0].info.feature_class = feat_6[0].info.feature_class;
    feat_7[0].info.vertices      = feat_6[0].info.vertices;
    feat_7[0].info.post_mask     = feat_6[0].info.post_mask;
    
    feat_7[1].micro_soil.poly_char_index0 = feat_6[3].micro.soil0;
    feat_7[1].micro_soil.poly_char_index1 = feat_6[3].micro.soil1;
    
    feat_7[2].xy.x = feat_6[1].xy.x;
    feat_7[2].xy.y = feat_6[1].xy.y;
    
    feat_7[3].z = feat_6[2].z;
    
    feat_7[4].micro.vert0 = feat_6[3].micro.vert0;
    feat_7[4].micro.vert1 = feat_6[3].micro.vert1;
    feat_7[4].micro.vert2 = feat_6[3].micro.vert2;

    feat_7[5].mwater.char0 = feat_6[4].mwater.char0;
    feat_7[5].mwater.char1 = feat_6[4].mwater.char1;
   
    return;    
}

static void
convert_volume_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{
    /* Comment left over from 6 to 5 conversion. */
    /* The only change between format 5 and 6 to volume models was the
     * addition of MES structures.  We'll count on another part of the
     * conversion process to create a traditional "volume" model for each
     * MES model, and to put all such models sequentially after any
     * traditional volume models in the format 6 database.
     */

    int32 verts = feat_6->info.vertices;
    int32 size;

    feat_7[1].volume.reference = feat_6[1].volume.reference + 
      ctdb_6->num_volume_models;
    feat_7[1].volume.type = CTDB_FS_BUILDING; /* Best guess */ 

    size = 3 + 2 * verts;

    *next_7 = feat_7 + size;
    *next_6 = feat_6 + size;

    bcopy(feat_6, feat_7, size * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_linear_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{
    /* No change from format 6 to format 7, just copy it over */

    int32 verts = feat_6->info.vertices;
    int32 size = 2 + 2 * verts;

    *next_7 = feat_7 + size;
    *next_6 = feat_6 + size;

    bcopy(feat_6, feat_7, size * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_canopy_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{
    /* No change from format 6 to format 7, just copy it over */

    *next_7 = feat_7 + 4;
    *next_6 = feat_6 + 4;

    bcopy(feat_6, feat_7, 4 * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_laid_linear_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{
    int32 verts = feat_6->info.vertices;
    int32 i;
    
    /* Format 7 Laid Linear is described by:
     *      INFO LL_SOIL LAID_LINEAR XY XY ...
     * Format 6 Laid Linear is described by:
     *      INFO LAID_LINEAR XY XY ...  
     */   

    *next_7 = feat_7 + 3 + verts; /* verts == 0, means 0 vertices, not 1 */
    *next_6 = feat_6 + 2 + verts;

    feat_7[0].info.feature_class = feat_6[0].info.feature_class;
    feat_7[0].info.vertices      = feat_6[0].info.vertices;
    feat_7[0].info.post_mask     = feat_6[0].info.post_mask;
    
    feat_7[1].ll_soil.poly_char_index = feat_6[1].laid_linear.soil;

    feat_7[2].laid_linear.width = feat_6[1].laid_linear.width;
    feat_7[2].laid_linear.edge_key= feat_6[1].laid_linear.edge_key;
    
    for (i=0; i<verts; i++)
    {
	feat_7[3+i].xy.x = feat_6[2+i].xy.x;
	feat_7[3+i].xy.y = feat_6[2+i].xy.y;
    }
    
    return;
}

/* This function should take each feature in the aggregate feature and
 * instantiate it in the format 5 database.  This would be easy except for
 * one thing:  elevation lookup.  Since the absolute Z has to be stored
 * for individual trees in format 5, and Z is not stored in aggregate
 * features, we must do an elevation lookup here.  However, the elevation
 * lookup routines in the compiler only work for the compiler representation
 * of databases (CTDB_CMP), not CTDB_6.  In order to get them to work
 * with CTDB_6 dbs, all of the elevation lookup code would have to be
 * duplicated (yuck!).  An alternative might be to convert the CTDB_6 into
 * a CTDB_CMP and pass that to the internal compiler representation elevation
 * lookup routines (double yuck, and wasteful of memory too!).  I see no
 * solution, and there is not pressing need to hang onto the abstract
 * features when converting from a post-5 to a pre-6 format (the resulting
 * database would probably be too large anyway), so I'm throwing out
 * aggregate features when converting to a format that does not support them.
 */
static void
convert_aggregate_6_7(ctdb_6, feat_6, feat_7, next_6, next_7)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_7  *feat_7;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_7 **next_7;
{
    /* Just advance and ignore the feature.
     */
    *next_7 = feat_7;
    *next_6 = feat_6 + 3;

    return;
}

/* Converts features for a patch from format6 to format5.  feat6 must point
 * to the start of the features in the source patch, and feat5 must point
 * to where the features should start in the destination patch.  Returns
 * number of bytes written into dest. patch.
 */
static int32 
convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7)
    CTDB_6              *ctdb_6;
    CTDB_7              *ctdb_7;
    CTDB_FEATURE_DATA_6 *feat_6;
    CTDB_FEATURE_DATA_7 *feat_7;
{
    CTDB_FEATURE_DATA_6 *last, *next_6;
    CTDB_FEATURE_DATA_7 *cur_7, *next_7;
    CTDB_FEAT_OBJ_PTR obj;
    int32 fclass, verts;
    int32 size_6;

    size_6 = feat_6->hdr.size;
    last = feat_6 + size_6;
    cur_7 = feat_7;
    feat_6++;
    cur_7++;
    
    for(;feat_6<last;(feat_6 = next_6), (cur_7 = next_7))
    {
	fclass = feat_6->info.feature_class;

	switch(fclass)
	{
	  case FC_MICRO_6:
	    if((feat_6->info.vertices & MICRO_TYPE_6) == MICRO_TYPE_WATER_6)
	      convert_mwater_6_7(ctdb_6, feat_6, cur_7, &next_6, &next_7);
	    else
	      convert_micro_6_7(ctdb_6, feat_6, cur_7, &next_6,	&next_7);
	    break;
	  case FC_VOLUME_6:
	    convert_volume_6_7(ctdb_6, feat_6, cur_7, &next_6, &next_7);
	    break;
	  case FC_LINEAR_6:
	    convert_linear_6_7(ctdb_6, feat_6, cur_7, &next_6, &next_7);
	    break;
	  case FC_CANOPY_6:
	    convert_canopy_6_7(ctdb_6, feat_6, cur_7, &next_6, &next_7);
	    break;
	  case FC_LAID_LINEAR_6:
	    convert_laid_linear_6_7(ctdb_6, feat_6, cur_7, &next_6, &next_7);
	    break;
	  case FC_AGGREGATE_6:
	    convert_aggregate_6_7(ctdb_6, feat_6, cur_7, &next_6, &next_7);
	    break;
	  default:
	    printf("ERROR: unknown feature class found in format 6 db.\n");
	    exit(-1);
	}
    }

    feat_7->hdr.size = next_7 - feat_7;

    return (feat_7->hdr.size * sizeof(CTDB_FEATURE_DATA_7));
}


static void
copy_mes_templates_6_7(ctdb, ctdb_out)
    CTDB_6 *ctdb;
    CTDB_7 *ctdb_out;
{
    int32 i,j;
    CTDB_MES_TEMPLATE_6 *templ_6;
    CTDB_MES_TEMPLATE_7 *templ_7;
    CTDB_MES_ENCLOSURE_6 *encl_6;
    CTDB_MES_ENCLOSURE_7 *encl_7;
	
    ctdb_out->num_mes_templates = ctdb->num_mes_templates;	
	
    ctdb_out->mes_templates = (CTDB_MES_TEMPLATE_7 *)
      ctdb_alloc(NULL,
		 sizeof(CTDB_MES_TEMPLATE_7)*ctdb->num_mes_templates,
		 "Fmt 7 MES template pointers");

    for(i = 0; i < ctdb->num_mes_templates; i++)
    {
	templ_7 = ctdb_out->mes_templates + i;
	templ_6 = ctdb->mes_templates + i;

	bcopy(templ_6, templ_7, sizeof(CTDB_MES_TEMPLATE_6));

	/* Copy over the vertices for this MES */
	templ_7->all_vertices =
	  (CTDB_MES_VERTEX_7*)ctdb_alloc(NULL,
			     sizeof(int32)*XYZ*templ_6->num_vertices,
			     "Fmt 7 MES vertices");

	bcopy(templ_6->all_vertices,
	      templ_7->all_vertices,
	      sizeof(int32)*XYZ*templ_6->num_vertices);


	/* Copy over the triangles for this MES */
	templ_7->all_triangles =
	  (CTDB_MES_TRIANGLE_7*)ctdb_alloc(NULL,
					   sizeof(CTDB_MES_TRIANGLE_7)*
					   templ_6->num_triangles,
					   "Fmt 7 MES triangles");

	bcopy(templ_6->all_triangles,
	      templ_7->all_triangles,
	      sizeof(CTDB_MES_TRIANGLE_7)*templ_6->num_triangles);


	/* Copy over the enclosures for this MES */
	templ_7->all_enclosures =
	  (CTDB_MES_ENCLOSURE_7*)ctdb_alloc(NULL,
					    sizeof(CTDB_MES_ENCLOSURE_7)*
					    templ_6->num_enclosures,
					    "Fmt 7 MES enclosure");

	bcopy(templ_6->all_enclosures, 
	      templ_7->all_enclosures,
	      sizeof(CTDB_MES_ENCLOSURE_7)*templ_6->num_enclosures);	    

	for(j = 0; j < templ_6->num_enclosures; j++)
	{
	    encl_6 = templ_6->all_enclosures + j;
	    encl_7 = templ_7->all_enclosures + j;

	    encl_7->aperture_ids = 
	      (uint16*)ctdb_alloc(NULL,
				 sizeof(uint16)*
				 encl_6->num_apertures,
				 "Fmt 7 MES aperture for the enclosures");
		
	    bcopy(encl_6->aperture_ids,
		  encl_7->aperture_ids,
		  sizeof(uint16)*encl_6->num_apertures);


	    encl_7->footprint.vertex_ids =
	      (int32*)ctdb_alloc(NULL,
				 sizeof(int32)*
				 encl_6->footprint.num_vertices,
				 "Fmt 7 MES footprint for this enclosure");
		
	    bcopy(encl_6->footprint.vertex_ids,
		  encl_7->footprint.vertex_ids,
		  sizeof(int32)*encl_6->footprint.num_vertices);
	}


	/* Copy over the apertures */
	templ_7->all_apertures =
	  (CTDB_MES_APERTURE_7*)ctdb_alloc(NULL, 
					     sizeof(CTDB_MES_APERTURE_7)*
					     templ_6->num_apertures,
					     "Fmt 7 MES apertures");

	bcopy(templ_6->all_apertures, 
	      templ_7->all_apertures,
	      sizeof(CTDB_MES_APERTURE_7)*templ_6->num_apertures);

	for(j = 0; j < templ_6->num_apertures; j++)
	{
	    templ_7->all_apertures[j].outline.vertex_ids = NULL;

	    templ_7->all_apertures[j].outline.vertex_ids =
	      (int32*)
	      ctdb_alloc(NULL,
			 sizeof(int32)*
			 templ_6->all_apertures[j].outline.num_vertices,
			 "Fmt 6 MES aperture vertices");

	    bcopy(templ_6->all_apertures[j].outline.vertex_ids,
		  templ_7->all_apertures[j].outline.vertex_ids,
		  sizeof(int32)*
		  templ_6->all_apertures[j].outline.num_vertices);
	}


	/* Copy over the boxes */
	templ_7->all_boxes =
	  (CTDB_MES_GRID_BOX_7*)ctdb_alloc(NULL,
					     sizeof(CTDB_MES_GRID_BOX_7)*
					     templ_6->num_boxes,
					     "Fmt 7 boxes");

	bcopy(templ_6->all_boxes,
	      templ_7->all_boxes,
	      sizeof(CTDB_MES_GRID_BOX_7)*

	      templ_6->num_boxes);

	for(j = 0; j < templ_6->num_boxes; j++)
	{
	    templ_7->all_boxes[j].box_ids = NULL;
	    templ_7->all_boxes[j].aperture_ids = NULL;
	    templ_7->all_boxes[j].wall_triangle_ids = NULL;

	    if(templ_6->all_boxes[j].dimensions_x*
	       templ_6->all_boxes[j].dimensions_y*
	       templ_6->all_boxes[j].dimensions_z > 1)
	    {
		templ_7->all_boxes[j].box_ids =
		  (int32*)
		  ctdb_alloc(NULL,
			     sizeof(int32)*
			     (templ_6->all_boxes[j].dimensions_x*
			      templ_6->all_boxes[j].dimensions_y*
			      templ_6->all_boxes[j].dimensions_z),
			     "Fmt 7 box inner boxes");

		bcopy(templ_6->all_boxes[j].box_ids,
		      templ_7->all_boxes[j].box_ids,
		      sizeof(int32)*
		      (templ_6->all_boxes[j].dimensions_x*
		       templ_6->all_boxes[j].dimensions_y*
		       templ_6->all_boxes[j].dimensions_z));
	    }

	    templ_7->all_boxes[j].aperture_ids =
	      (uint16*)ctdb_alloc(NULL,
				 sizeof(uint16)*
				 templ_6->all_boxes[j].num_apertures,
				 "Fmt 7 box apertures");


	    bcopy(templ_6->all_boxes[j].aperture_ids,
		  templ_7->all_boxes[j].aperture_ids,
		  sizeof(uint16)*templ_6->all_boxes[j].num_apertures);

	    templ_7->all_boxes[j].wall_triangle_ids =
	      (int32*)ctdb_alloc(NULL,
				 sizeof(int32)*
				 templ_6->all_boxes[j].num_triangles,
				 "Fmt 7 box apertures");


	    bcopy(templ_6->all_boxes[j].wall_triangle_ids,
		  templ_7->all_boxes[j].wall_triangle_ids,
		  sizeof(int32)*templ_6->all_boxes[j].num_triangles);
	}
    } /*	for(i = 0; i < ctdb->num_mes_templates; i++) */
    
    return;
}

static void
copy_mes_volumes_6_7(ctdb, ctdb_out)
    CTDB_6 *ctdb;
    CTDB_7 *ctdb_out;
{
    int32 i;
    CTDB_MES_VOLUME_6 *vol_6;
    CTDB_MES_VOLUME_7 *vol_7;
    CTDB_MES_LIST_VERTEX_F64_7 *roofline, *floorline;

    ctdb_out->num_mes_volumes = ctdb->num_mes_volumes;

    if (!ctdb->num_mes_volumes)
    {
	ctdb_out->mes_volumes = NULL;
	return;
    }
    
    ctdb_out->mes_volumes = (CTDB_MES_VOLUME_7 *)
      ctdb_alloc(NULL, ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME_7),
		 "Format 7 MES volumes");
    
    for(i=0;i<ctdb->num_mes_volumes;i++)
    {
	vol_7 = ctdb_out->mes_volumes + i;
	vol_6 = ctdb->mes_volumes + i;

	bcopy(vol_6, vol_7, sizeof(CTDB_MES_VOLUME_6));
	
	/* Reset the template pointers to point to the new templates */
	vol_7->template = 
	  ctdb_out->mes_templates + 
	  (ctdb->mes_volumes[i].template - ctdb->mes_templates);
	
	/* Copy the roofline: */
	roofline = &(vol_7->roofline);
	roofline->vertices =
	  (CTDB_MES_VERTEX_F64_7*)
	  ctdb_alloc(NULL,
		     roofline->num_vertices * sizeof(CTDB_MES_VERTEX_F64_7),
		     "Fmt 7 MES roofline");
	bcopy(vol_6->roofline.vertices, 
	      roofline->vertices,
	      roofline->num_vertices * sizeof(CTDB_MES_VERTEX_F64_7));

	/* Copy the floorline: */
	floorline = &(vol_7->floorline);
	floorline->vertices =
	  (CTDB_MES_VERTEX_F64_7*)
	  ctdb_alloc(NULL,
		     floorline->num_vertices * sizeof(CTDB_MES_VERTEX_F64_7),
		     "Fmt 7 MES floorline");
	bcopy(vol_6->floorline.vertices, 
	      floorline->vertices,
	      floorline->num_vertices * sizeof(CTDB_MES_VERTEX_F64_7));
    } /*	for(i = 0; i < ctdb->num_mes_volumes; i++) */
    
    return;    
}

/* Convert all data, including data header, for a gridded PG from format 6 to
 * format 7.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_grid_pg_data_6_7(ctdb_6, ctdb_7, group, base)
    CTDB_6 *ctdb_6;
    CTDB_7 *ctdb_7;
    int32   group;
    ADDRESS base;
{
    int32 i;
    int32 size;
    CTDB_GRID_PG_DATA_6 *data_6;
    CTDB_GRID_PG_DATA_7 *data_7;
    CTDB_PATCH_GROUP_7 *pg;
    CTDB_FEATURE_DATA_6 *feat_6;
    CTDB_FEATURE_DATA_7 *feat_7;
    int32 patch_bytes = 0;
    int32 patch_number;

    data_6 = ctdb_6->patch_groups[group].data_start.grid_data;
    data_7 = (CTDB_GRID_PG_DATA_7 *)base;
    size = sizeof(CTDB_GRID_PG_DATA_7);
    patch_number = group * CTDB_PATCHES_PER_GROUP_7;
    
    pg = ctdb_7->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_7, size);
    bcopy(data_6->max_elev, data_7->max_elev, 
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(float32));

    bcopy(data_6->water_elev, data_7->water_elev,
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(float32));

    bcopy(data_6->tables, data_7->tables,
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(CTDB_PATCH_TABLE_REF_7));
    
    /* Can't fill in offsets yet... */

    /* 2.  Convert the features for each patch */
    feat_7 = (CTDB_FEATURE_DATA_7 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((int8 *)data_6 +
					 sizeof(CTDB_GRID_PG_DATA_6));
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));

	data_7->patch1_offset = (patch_bytes + sizeof(CTDB_GRID_PG_DATA_7)) /
	  sizeof(CTDB_FEATURE_DATA_7);
    }
    else
    {
	data_7->patch1_offset = sizeof(CTDB_GRID_PG_DATA_7) /
	  sizeof(CTDB_FEATURE_DATA_7);
    }

    patch_number++;
    if (pg->patch1_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch1_offset);
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));

	data_7->patch2_offset = data_7->patch1_offset +
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));
    }
    else
    {
	data_7->patch2_offset = data_7->patch1_offset;
    }

    patch_number++;
    if (pg->patch2_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch2_offset);
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));

	data_7->patch3_offset = data_7->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));
    }
    else
    {
	data_7->patch3_offset = data_7->patch2_offset;
    }

    patch_number++;
    if (pg->patch3_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch3_offset);
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));
    }	  

    size = (int8 *)feat_7 - (int8 *)data_7;
    return size;
}


/* Convert all data, including data header, for a tinned PG from format 6 to
 * format 7.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_tin_pg_data_6_7(ctdb_6, ctdb_7, group, base)
    CTDB_6 *ctdb_6;
    CTDB_7 *ctdb_7;
    int32   group;
    ADDRESS base;
{
    int32 i;
    int32 size;
    CTDB_TIN_PG_DATA_6 *data_6;
    CTDB_TIN_PG_DATA_7 *data_7;
    CTDB_PATCH_GROUP_7 *pg;
    CTDB_FEATURE_DATA_6 *feat_6;
    CTDB_FEATURE_DATA_7 *feat_7;
    int32 patch_bytes = 0;
    int32 delta;
    int32 te_offset_6;
    int32 te_offset_7;
    int32 patch_number;

    data_6 = ctdb_6->patch_groups[group].data_start.tin_data;
    data_7 = (CTDB_TIN_PG_DATA_7 *)base;
    size = sizeof(CTDB_TIN_PG_DATA_7);
    patch_number = group * CTDB_PATCHES_PER_GROUP_7;

    pg = ctdb_7->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_7, size);
    bcopy(data_6->max_elev, data_7->max_elev, 
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(float32));
    
    bcopy(data_6->water_elev, data_7->water_elev,
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(float32));

    bcopy(data_6->tables, data_7->tables,
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(CTDB_PATCH_TABLE_REF_7));

    /* Can't fill in offsets yet... */

    bcopy(data_6->feature_present, data_7->feature_present,
	  CTDB_MAX_POSTS_PER_PATCH_7 * 2);
    
    bcopy(data_6->num_te, data_7->num_te,
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(uint16));
    
    /* Can't fill in te_offsets yet */

    bcopy(data_6->te_vgrid_dim, data_7->te_vgrid_dim,
	  CTDB_PATCHES_PER_GROUP_7 * sizeof(uint8));
    
    /* Can't fill in vgrid_offsets yet */

    /* 2.  Convert the features for each patch */
    feat_7 = (CTDB_FEATURE_DATA_7 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((int8 *)data_6 +
					 sizeof(CTDB_TIN_PG_DATA_6));
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));

	data_7->patch1_offset = (patch_bytes + sizeof(CTDB_TIN_PG_DATA_7)) /
	  sizeof(CTDB_FEATURE_DATA_7);
    }
    else
    {
	data_7->patch1_offset = sizeof(CTDB_TIN_PG_DATA_7) /
	  sizeof(CTDB_FEATURE_DATA_7);
    }

    patch_number++;
    if (pg->patch1_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch1_offset);
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));

	data_7->patch2_offset = data_7->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));
    }
    else
    {
	data_7->patch2_offset = data_7->patch1_offset;
    }

    patch_number++;
    if (pg->patch2_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch2_offset);
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));

	data_7->patch3_offset = data_7->patch2_offset +
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));
    }
    else
    {
	data_7->patch3_offset = data_7->patch2_offset;
    }

    patch_number++;
    if (pg->patch3_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch3_offset);
	
	patch_bytes = convert_feat_6_7(ctdb_6, ctdb_7, feat_6, feat_7);

	feat_7 = feat_7 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_7));
    }	  

    /* 3. Calculate how far the TE's for the patch have moved relative to
     * where they used to be
     */
    if (!pg->patch0_no_tes)
    {
	te_offset_6 = data_6->te_offset[0];
    }
    else if (!pg->patch1_no_tes)
    {
	te_offset_6 = data_6->te_offset[1];
    }
    else if (!pg->patch2_no_tes)
    {
	te_offset_6 = data_6->te_offset[2];
    }
    else
    {
	te_offset_6 = data_6->te_offset[3];
    }
    
    te_offset_7 = feat_7 - (CTDB_FEATURE_DATA_7 *)data_7;
    
    /* Now, calculate the offset, in words */
    delta = te_offset_7 - te_offset_6;
    
    /* 4.  Fill out the rest of the offset values. */
    for(i=0;i<CTDB_PATCHES_PER_GROUP_7;i++)
    {
	data_7->te_offset[i] = data_6->te_offset[i] + delta;
	data_7->te_vgrid_offset[i] = data_6->te_vgrid_offset[i] + delta;
    }
    
    /* 5.  Copy over all the TE data. */
    {
	int32 data_size;
	int32 te_data_size;	/* In bytes */

	if (group < ctdb_6->max_patch_number)
	{
	    
	    /* Total bytes of data stored for this PG */
	    data_size = 
	      (int8 *)ctdb_6->patch_groups[group+1].data_start.grid_data -
	      (int8 *)data_6;
	    
	    te_data_size = data_size - 
	      (data_6->te_offset[0] * sizeof(CTDB_FEATURE_DATA_6));
	}
	else
	{
	    int32 vgrid_size;
	    
	    if (!pg->patch3_no_tes)
	    {
		data_size = data_6->te_vgrid_offset[3] *
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[3] * data_6->te_vgrid_dim[3];
	    }
	    else if (!pg->patch2_no_tes)
	    {
		data_size = data_6->te_vgrid_offset[2] * 
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[2] * data_6->te_vgrid_dim[2];
	    }
	    else if (!pg->patch1_no_tes)
	    {
		data_size = data_6->te_vgrid_offset[1] * 
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[1] * data_6->te_vgrid_dim[1];
	    }
	    else
	    {
		data_size = data_6->te_vgrid_offset[0] * 
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[0] * data_6->te_vgrid_dim[0];
	    }

	    /* Add padding as necessary */
	    vgrid_size += (vgrid_size % sizeof(CTDB_FEATURE_DATA_6));
	    data_size += vgrid_size;		
	    te_data_size = (data_size -
			    (te_offset_6 * sizeof(CTDB_FEATURE_DATA_6)));
	}

	feat_6 = (CTDB_FEATURE_DATA_6 *)data_6 + te_offset_6;
	bcopy(feat_6, feat_7, te_data_size);
	
	/* Calculate total bytes in this PG */
	size = ((int8 *)feat_7 - (int8 *)data_7) + te_data_size;
    }

    return size;
}



/* Convert patch group header from fmt6 to 5.  base is where the new PG
 * should point.
 */
static void
convert_pg_6_7(ctdb_6, ctdb_7, group, base)
    CTDB_6 *ctdb_6;
    CTDB_7 *ctdb_7;
    int32   group;
    ADDRESS base;
{
    CTDB_PATCH_GROUP_7 *pg_7;
    CTDB_PATCH_GROUP_6 *pg_6;

    pg_7 = ctdb_7->patch_groups + group;
    pg_6 = ctdb_6->patch_groups + group;
    
    bzero(pg_7, sizeof(CTDB_PATCH_GROUP_6));

    pg_7->data_start.grid_data = (CTDB_GRID_PG_DATA_7 *)base;
    
    /* We'll leave the cache stuff blank for now
     */

    pg_7->contains_tins = pg_6->contains_tins;

    pg_7->patch0_occupied = pg_6->patch0_occupied;
    pg_7->patch0_dyn_terr = pg_6->patch0_dyn_terr;
    pg_7->patch0_te16 = pg_6->patch0_te16;
    pg_7->patch0_no_tes = pg_6->patch0_no_tes;
    pg_7->patch0_water = pg_6->patch0_water;

    pg_7->patch1_occupied = pg_6->patch1_occupied;
    pg_7->patch1_dyn_terr = pg_6->patch1_dyn_terr;
    pg_7->patch1_te16 = pg_6->patch1_te16;
    pg_7->patch1_no_tes = pg_6->patch1_no_tes;
    pg_7->patch1_water = pg_6->patch1_water;

    pg_7->patch2_occupied = pg_6->patch2_occupied;
    pg_7->patch2_dyn_terr = pg_6->patch2_dyn_terr;
    pg_7->patch2_te16 = pg_6->patch2_te16;
    pg_7->patch2_no_tes = pg_6->patch2_no_tes;
    pg_7->patch2_water = pg_6->patch2_water;

    pg_7->patch3_occupied = pg_6->patch3_occupied;
    pg_7->patch3_dyn_terr = pg_6->patch3_dyn_terr;
    pg_7->patch3_te16 = pg_6->patch3_te16;
    pg_7->patch3_no_tes = pg_6->patch3_no_tes;
    pg_7->patch3_water = pg_6->patch3_water;

    return;
}

/* Converts PG header and all PG data from format 6 to format 7.  group is
 * the patch group number to convert, base is an in-out variable passing
 * in the address where the format 6 data should start, and passing out
 * where the format 6 data for the next PG should start.
 */
static int32
convert_group_6_7(ctdb_6, ctdb_7, group, base)
    CTDB_6  *ctdb_6;
    CTDB_7  *ctdb_7;
    int32    group;
    ADDRESS *base;
{
    int32 bytes;
    
    /* First, convert the header. */
    convert_pg_6_7(ctdb_6, ctdb_7, group, *base);

    /* Now convert the data and the data headers */
    if (ctdb_7->patch_groups[group].contains_tins)
    {
	bytes = convert_tin_pg_data_6_7(ctdb_6, ctdb_7, group, *base);
    }
    else
    {
	bytes = convert_grid_pg_data_6_7(ctdb_6, ctdb_7, group, *base);
    }
    
    *base = (int8 *)*base + bytes;
    return bytes;
}


/* old_base is the start address for all tables in an existing
 * database, and old_ptrs is an array of pointers, all of which point
 * into data stored in the data chunk pointed to by old_base.  new_base
 * points to a copy of the database pointed to by old_base; set up 
 * num_tables pointers in new_ptrs to point at the mirror of the data
 * pointed at by the corresponding pointers in old_ptrs
 */
static void
regen_table_ptrs(old_base, old_ptrs, new_base, new_ptrs, num_entries)
    int8   *old_base;
    int8  **old_ptrs;
    int8   *new_base;
    int8  **new_ptrs;
    int32   num_entries;
{
    int32 i;
    int32 offset;
    
    for(i=0;i<num_entries;i++)
    {
	offset = (int8 *)old_ptrs[i] - old_base;
	new_ptrs[i] = new_base + offset;
    }
    
    return;
}


/* Converts PGs and features (including TINs) from format 6 to format 7.
 * Returns total patch data bytes in format 7 result.
 */
static int32
convert_features_6_7(ctdb_6, ctdb_7)
    CTDB_6 *ctdb_6;
    CTDB_7 *ctdb_7;    
{
    int32 groups;
    int32 i;
    int32 total_bytes = 0;
    ADDRESS base;
	
    groups = ctdb_7->patches_wide * ctdb_7->patches_high /
      CTDB_PATCHES_PER_GROUP_7;
	
    /* Going from 6 to 7, some features increase in size.  Worst case is a
     * database with nothing but 1 vertex laid linears, which would grow 
     * 33%.  Assume that case.
     */
    ctdb_7->n_patch_bytes = ceil(1.34 * (float64)ctdb_6->n_patch_bytes);
	
    ctdb_7->patch_data = (ADDRESS)ctdb_alloc(NULL,
					     ctdb_7->n_patch_bytes,
					     "Fmt. 6 patch data");

    base = ctdb_7->patch_data;

    for(i=0;i<groups;i++)
    {
	total_bytes += convert_group_6_7(ctdb_6, ctdb_7, i, &base);
    }

    return total_bytes;    
}


/***************************************************************************
 * 6 -> 5 CONVERSION                                                       *
 ***************************************************************************/

void *
ctdb_convert_6_to_5(fmt6_db)
    void *fmt6_db;
{
    CTDB_6 *ctdb = (CTDB_6 *)fmt6_db;
    CTDB_5 *ctdb_out;

    ctdb_out = (CTDB_5 *)ctdb_alloc(NULL, sizeof(CTDB_5),
				    "Fmt. 5 database");
    
    bzero(ctdb_out, sizeof(CTDB_5));

    /* First, copy the easy stuff over */
#define COPY_6_5(x) (ctdb_out->x = ctdb->x)

    ctdb_out->fixed_point_basis = CTDB_STANDARD_FIXED_POINT_BASIS_5;
    ctdb_out->inv_fixed_point_basis = 1.0 / ctdb_out->fixed_point_basis;

    COPY_6_5(num_soil_tables);
    COPY_6_5(soil_table_size);
    COPY_6_5(num_water_chars);
    COPY_6_5(num_tidal_zones);
    COPY_6_5(incr);
    COPY_6_5(inv_incr);
    COPY_6_5(incrD);

    COPY_6_5(patch_incr);
    COPY_6_5(patch_inv_incr);
    COPY_6_5(patch_incrD);

    COPY_6_5(pages_wide);
    COPY_6_5(pages_high);
    COPY_6_5(patches_wide);
    COPY_6_5(patches_high);
    COPY_6_5(min_x);
    COPY_6_5(max_x);
    COPY_6_5(min_y);
    COPY_6_5(max_y);
    COPY_6_5(min_x_post);
    COPY_6_5(max_x_post);
    COPY_6_5(min_y_post);
    COPY_6_5(max_y_post);
    
    ctdb_out->input_format = 5;
    
    COPY_6_5(min_z);
    COPY_6_5(max_z);
    COPY_6_5(origin_northing);
    COPY_6_5(origin_easting);
    COPY_6_5(origin_zone_number);
    COPY_6_5(origin_zone_letter);
    COPY_6_5(datum);
    COPY_6_5(cell_id);
    
    COPY_6_5(origin_x);
    COPY_6_5(origin_y);
    
    COPY_6_5(west_long);
    COPY_6_5(east_long);
    COPY_6_5(south_lat);
    COPY_6_5(north_lat);

    COPY_6_5(gcs_mode);

    COPY_6_5(version);
    ctdb_out->total_memory = 0;	/* Noone should be using this anyway */
    COPY_6_5(is_curved);

    /* Format 5 doesn't support per-post variable diagonalization, and there
     * is no per-database diagonalization information in format 6 databases.
     * The best we can do here is guess.  I'll guess NW-SE.  I probably have
     * a 95% chance of being right.
     */
    ctdb_out->is_sw_ne = FALSE;

    ctdb_out->elevations_resident = TRUE;

    /* We'll leave the rest of the elevation cache at 0 */

    ctdb_out->file = NULL;	/* We don't have a file... */

    COPY_6_5(num_quad_data);
    COPY_6_5(num_nodes);
    COPY_6_5(node_size);
    COPY_6_5(num_edges);
    COPY_6_5(edge_size);
    COPY_6_5(num_linear_models);
    COPY_6_5(num_volume_models);

    COPY_6_5(meters_to_patch);
    COPY_6_5(is_gridded);
    COPY_6_5(is_tinned);
    COPY_6_5(max_patch_number);
    COPY_6_5(min_patch_number);
    
    /* Now for the real work */

    /* Soil tables */
    if (!ctdb->num_soil_tables)
    {
	ctdb_out->soil_table_storage = NULL;
	ctdb_out->soil_tables = NULL;
	ctdb_out->num_soil_tables = 0;
	ctdb_out->soil_table_size = 0;
    }
    else
    {
	ctdb_out->num_soil_tables = ctdb->num_soil_tables;
	ctdb_out->soil_tables =
	  (int32 **)ctdb_alloc(NULL, ctdb->num_soil_tables *
			       sizeof(int32 *),
			       "Fmt. 5 soil table ptrs");
	ctdb_out->soil_table_size = ctdb->soil_table_size;
	ctdb_out->soil_table_storage = 
	  (int8 *)ctdb_alloc(NULL, ctdb_out->soil_table_size,
			     "Fmt. 5 soil table storage");

	bcopy(ctdb->soil_table_storage, ctdb_out->soil_table_storage,
	      ctdb->soil_table_size);
	
	regen_table_ptrs((int8 *)ctdb->soil_table_storage, 
			 (int8 **)ctdb->soil_tables,
			 (int8 *)ctdb_out->soil_table_storage, 
			 (int8 **)ctdb_out->soil_tables,
			 ctdb_out->num_soil_tables);
    }

    /* Copy the tidal zones */
    if (!ctdb_out->num_tidal_zones)
    {
	ctdb_out->tidal_zones = NULL;
	ctdb_out->num_tidal_zones = 0;
    }
    else
    {
	ctdb_out->num_tidal_zones = ctdb->num_tidal_zones;
	ctdb_out->tidal_zones = (CTDB_TIDAL_ZONE_5 *)
	  ctdb_alloc(NULL, sizeof(CTDB_TIDAL_ZONE_5) * ctdb->num_tidal_zones,
		     "Fmt. 5 tidal zones.");
	bcopy(ctdb->tidal_zones, ctdb_out->tidal_zones,
	      sizeof(CTDB_TIDAL_ZONE_5) * ctdb_out->num_tidal_zones);
    }

    /* Copy the water char entries */
    if (!ctdb->num_water_chars)
    {
	ctdb_out->water_chars = NULL;
	ctdb_out->num_water_chars = 0;
    }
    else
    {
	ctdb_out->num_water_chars = ctdb->num_water_chars;
	ctdb_out->water_chars = (CTDB_WATER_CHARS_5 *)
	  ctdb_alloc(NULL, sizeof(CTDB_WATER_CHARS_5) * ctdb->num_water_chars,
		     "Fmt. 5 water chars.");
	bcopy(ctdb->water_chars, ctdb_out->water_chars, ctdb->num_water_chars *
	      sizeof(CTDB_WATER_CHARS_5));
    }

    /* Name & date */
    bcopy(ctdb->name, ctdb_out->name, sizeof(ctdb_out->name));
    bcopy(ctdb->date, ctdb_out->date, sizeof(ctdb_out->date));

    /* Patch features */
    {
	int32 groups = (ctdb_out->patches_wide * ctdb_out->patches_high) /
	  CTDB_PATCHES_PER_GROUP_5;

	ctdb_out->patch_groups = (CTDB_PATCH_GROUP_5 *)
	  ctdb_alloc(NULL, groups * sizeof(CTDB_PATCH_GROUP_5),
		     "Fmt. 5 patch groups");

	ctdb_out->n_patch_bytes = convert_features_6_5(ctdb, ctdb_out);
    }

    /* Elevations */
    if (!ctdb->is_gridded)
    {
	ctdb_out->elevations = NULL;
    }
    else
    {
	int32 pages = ctdb_out->pages_wide * ctdb_out->pages_high;
	int32 elev_mem = pages * PAGE_SIZE;
	
	ctdb_out->elevations = (int32 *)
	  ctdb_alloc(NULL, elev_mem, 
		     "Fmt. 5 elevation data");

	convert_elevation_posts_6_5(ctdb, ctdb_out);
    }

    /* Quad tree */
    if (!ctdb->num_quad_data)
    {
	ctdb_out->quad_data = NULL;
    }
    else
    {
	ctdb_out->quad_data = (CTDB_FEATURE_DATA_5 *)
	  ctdb_alloc(NULL, ctdb->num_quad_data *
		     sizeof(CTDB_FEATURE_DATA_5),
		     "Fmt. 5 quadtree");
	
	bcopy(ctdb->quad_data, ctdb_out->quad_data, 
	      ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_5));

    }

    /* Topology */
    ctdb_out->node_storage = (int8 *)
      ctdb_alloc(NULL, ctdb->node_size, "Fmt.5 nodes");
    bcopy(ctdb->node_storage, ctdb_out->node_storage, ctdb->node_size);
    
    ctdb_out->edge_storage = (int8 *)
      ctdb_alloc(NULL, ctdb->edge_size, "Fmt. 5 edges");
    bcopy(ctdb->edge_storage, ctdb_out->edge_storage, ctdb->edge_size);

    /* Linear models */
    if (!ctdb->num_linear_models)
    {
	ctdb_out->linear_models = NULL;
    }
    else
    {
	ctdb_out->linear_models = 
	  (CTDB_LINEAR_MODEL_5 *)ctdb_alloc(NULL,
					    sizeof(CTDB_LINEAR_MODEL_5) *
					    ctdb->num_linear_models,
					    "Fmt. 5 linear models");

	bcopy(ctdb->linear_models, ctdb_out->linear_models,
	      ctdb->num_linear_models * sizeof(CTDB_LINEAR_MODEL_6));
    }
    
    /* Volume models */
    if (!ctdb->num_volume_models && !ctdb->num_mes_volumes)
    {
	ctdb_out->volume_models = NULL;
    }
    else
    {
	int32 model_size;
	CTDB_VOLUME_MODEL_6 *old_model;
	int32 i, j, index;
	
	int32 max_model_id = 0;

	/* $$$ In theory, we should probably make one volume model per
	 * MES *template*.  However, templates do not (and should not!)
	 * store rooflines, and deriving them is non-trivial.  Therefore,
	 * I have chosen to make one volume model per MES *volume*, which
	 * is wasteful of space but not wrong and probably not that important
	 */
	ctdb_out->volume_models = (CTDB_VOLUME_MODEL_5 **)
	  ctdb_alloc(NULL,
		     sizeof(CTDB_VOLUME_MODEL_5 *) * 
		     (ctdb->num_volume_models + ctdb->num_mes_volumes),
		     "Fmt 5 volume model pointers");
	
	/* Copy over the volume models */
	for(i=0;i<ctdb->num_volume_models;i++)
	{
	    old_model = ctdb->volume_models[i];
	    if (old_model->model_id > max_model_id)
	    {
		/* Keep track of the highest existing model id so
		 * we can assign unique ids to any MES models we need
		 * to convert
		 */
		max_model_id = old_model->model_id;
	    }
	    
	    model_size = sizeof(CTDB_VOLUME_MODEL_6) - sizeof(float32) +
	      old_model->num_vertices * 3 * sizeof(float32);

	    ctdb_out->volume_models[i] = (CTDB_VOLUME_MODEL_5 *)
	      ctdb_alloc(NULL, model_size, "Fmt. 5 volume model");
	    
	    bcopy(old_model, ctdb_out->volume_models[i], model_size);
	}	
	
	/* Convert MES volumes over to volume models */
	for(i=0;i<ctdb->num_mes_volumes;i++)
	{
	    index = ctdb_out->num_volume_models + i;

	    model_size = sizeof(CTDB_VOLUME_MODEL_6) - sizeof(float32) +
	      ctdb->mes_volumes[i].roofline.num_vertices * 3 * sizeof(float32);
	    
	    ctdb_out->volume_models[index] = (CTDB_VOLUME_MODEL_5 *)
	      ctdb_alloc(NULL, model_size, "Fmt. 5 volume model");
	    
	    /* Set up new model with reasonable values, then copy in the
	     * appropriate vertices
	     */
	    strcpy(ctdb_out->volume_models[index]->model_name, 
		   "Fmt. 6 MES model");
	    ctdb_out->volume_models[index]->model_id = ++max_model_id;
	    ctdb_out->volume_models[index]->next_state_index = -1;
	    ctdb_out->volume_models[index]->percent_damaged = 0.0;
	    ctdb_out->volume_models[index]->num_vertices = 
	      ctdb->mes_volumes[i].roofline.num_vertices;
	    
	    /* Just to be safe... */
	    ctdb_out->volume_models[index]->pad = 0;
	    
	    for(j=0;j<ctdb_out->volume_models[index]->num_vertices;j++)
	    {
		ctdb_out->volume_models[index]->vertices[3*j] =
		  ctdb->mes_volumes[i].roofline.vertices[i][X];

		ctdb_out->volume_models[index]->vertices[3*j + 1] =
		  ctdb->mes_volumes[i].roofline.vertices[i][Y];

		ctdb_out->volume_models[index]->vertices[3*j + 2] =
		  ctdb->mes_volumes[i].roofline.vertices[i][Z];
	    }
	}

	ctdb_out->num_volume_models += ctdb->num_mes_volumes;

    }  

    return ctdb_out;
}


static void
convert_micro_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* No change from format 5 to format 6, just copy it over */

    *next_6 = feat_6 + 4;
    *next_5 = feat_5 + 4;

    bcopy(feat_6, feat_5, 4 * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_mwater_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* No change from format 5 to format 6, just copy it over */

    *next_6 = feat_6 + 5;
    *next_5 = feat_5 + 5;

    bcopy(feat_6, feat_5, 5 * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_volume_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* The only change between format 5 and 6 to volume models was the
     * addition of MES structures.  We'll count on another part of the
     * conversion process to create a traditional "volume" model for each
     * MES model, and to put all such models sequentially after any
     * traditional volume models in the format 6 database.
     */

    /* $$$ Why are we setting reference and type fields and then doing
     * a bcopy which will overwrite anything we set before it?
     */

    int32 verts = feat_6->info.vertices;
    int32 size;

    feat_5[1].volume.reference = feat_6[1].volume.reference + 
      ctdb_6->num_volume_models;
    feat_5[1].volume.type = CTDB_FS_BUILDING; /* Best guess */

    size = 3 + 2 * verts;

    *next_5 = feat_5 + size;
    *next_6 = feat_6 + size;

    bcopy(feat_6, feat_5, size * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_linear_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* No change from format 5 to format 6, just copy it over */

    int32 verts = feat_6->info.vertices;
    int32 size = 2 + 2 * verts;

    *next_5 = feat_5 + size;
    *next_6 = feat_6 + size;

    bcopy(feat_6, feat_5, size * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_canopy_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* No change from format 5 to format 6, just copy it over */

    *next_5 = feat_5 + 4;
    *next_6 = feat_6 + 4;

    bcopy(feat_6, feat_5, 4 * sizeof(CTDB_FEATURE_DATA_6));
    return;    
}

static void
convert_laid_linear_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    int32 verts = feat_6->info.vertices;

    /* No change from format 5 to format 6, just copy it over */

    *next_5 = feat_5 + 2 + verts;
    *next_6 = feat_6 + 2 + verts;

    bcopy(feat_6, feat_5, (verts + 2) * sizeof(CTDB_FEATURE_DATA_6));
    return;
}

/* This function should take each feature in the aggregate feature and
 * instantiate it in the format 5 database.  This would be easy except for
 * one thing:  elevation lookup.  Since the absolute Z has to be stored
 * for individual trees in format 5, and Z is not stored in aggregate
 * features, we must do an elevation lookup here.  However, the elevation
 * lookup routines in the compiler only work for the compiler representation
 * of databases (CTDB_7), not CTDB_6.  In order to get them to work
 * with CTDB_6 dbs, all of the elevation lookup code would have to be
 * duplicated (yuck!).  An alternative might be to convert the CTDB_6 into
 * a CTDB_7 and pass that to the internal compiler representation elevation
 * lookup routines (double yuck, and wasteful of memory too!).  I see no
 * solution, and there is not pressing need to hang onto the abstract
 * features when converting from a post-5 to a pre-6 format (the resulting
 * database would probably be too large anyway), so I'm throwing out
 * aggregate features when converting to a format that does not support them.
 */
static void
convert_aggregate_6_5(ctdb_6, feat_6, feat_5, next_6, next_5)
    CTDB_6               *ctdb_6;
    CTDB_FEATURE_DATA_6  *feat_6;
    CTDB_FEATURE_DATA_5  *feat_5;
    CTDB_FEATURE_DATA_6 **next_6;
    CTDB_FEATURE_DATA_5 **next_5;
{
    /* Just advance and ignore the feature.
     */
    *next_5 = feat_5;
    *next_6 = feat_6 + 3;

    return;
}

/* Converts features for a patch from format6 to format5.  feat6 must point
 * to the start of the features in the source patch, and feat5 must point
 * to where the features should start in the destination patch.  Returns
 * number of bytes written into dest. patch.
 */
static int32 
convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5)
    CTDB_6              *ctdb_6;
    CTDB_5              *ctdb_5;
    CTDB_FEATURE_DATA_6 *feat_6;
    CTDB_FEATURE_DATA_5 *feat_5;
{
    CTDB_FEATURE_DATA_6 *last, *next_6;
    CTDB_FEATURE_DATA_5 *cur_5, *next_5;
    CTDB_FEAT_OBJ_PTR obj;
    int32 fclass, verts;
    int32 size_6;

    size_6 = feat_6->hdr.size;
    last = feat_6 + size_6;
    cur_5 = feat_5;
    feat_6++;
    cur_5++;
    
    for(;feat_6<last;(feat_6 = next_6), (cur_5 = next_5))
    {
	fclass = feat_6->info.feature_class;

	switch(fclass)
	{
	  case FC_MICRO_6:
	    if((feat_6->info.vertices & MICRO_TYPE_6) == MICRO_TYPE_WATER_6)
	      convert_mwater_6_5(ctdb_6, feat_6, cur_5, &next_6, &next_5);
	    else
	      convert_micro_6_5(ctdb_6, feat_6, cur_5, &next_6,	&next_5);
	    break;
	  case FC_VOLUME_6:
	    convert_volume_6_5(ctdb_6, feat_6, cur_5, &next_6, &next_5);
	    break;
	  case FC_LINEAR_6:
	    convert_linear_6_5(ctdb_6, feat_6, cur_5, &next_6, &next_5);
	    break;
	  case FC_CANOPY_6:
	    convert_canopy_6_5(ctdb_6, feat_6, cur_5, &next_6, &next_5);
	    break;
	  case FC_LAID_LINEAR_6:
	    convert_laid_linear_6_5(ctdb_6, feat_6, cur_5, &next_6, &next_5);
	    break;
	  case FC_AGGREGATE_6:
	    convert_aggregate_6_5(ctdb_6, feat_6, cur_5, &next_6, &next_5);
	    break;
	  default:
	    printf("ERROR: unknown feature class encountered in format 6 db.\n");
	    exit(-1);
	}
    }

    feat_5->hdr.size = next_5 - feat_5;

    return (feat_5->hdr.size * sizeof(CTDB_FEATURE_DATA_5));
}

/* Convert all data, including data header, for a gridded PG from format 6 to
 * format 5.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_grid_pg_data_6_5(ctdb_6, ctdb_5, group, base)
    CTDB_6 *ctdb_6;
    CTDB_5 *ctdb_5;
    int32   group;
    ADDRESS base;
{
    int32 i;
    int32 size;
    CTDB_GRID_PG_DATA_6 *data_6;
    CTDB_GRID_PG_DATA_5 *data_5;
    CTDB_PATCH_GROUP_5 *pg;
    CTDB_FEATURE_DATA_6 *feat_6;
    CTDB_FEATURE_DATA_5 *feat_5;
    int32 patch_bytes = 0;
    int32 patch_number;

    data_6 = ctdb_6->patch_groups[group].data_start.grid_data;
    data_5 = (CTDB_GRID_PG_DATA_5 *)base;
    size = sizeof(CTDB_GRID_PG_DATA_5);
    patch_number = group * CTDB_PATCHES_PER_GROUP_5;
    
    pg = ctdb_5->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_5, size);
    bcopy(data_6->max_elev, data_5->max_elev, 
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(float32));

    bcopy(data_6->water_elev, data_5->water_elev,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(float32));

    bcopy(data_6->tables, data_5->tables,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(CTDB_PATCH_TABLE_REF_5));
    
    /* Can't fill in offsets yet... */

    /* 2.  Convert the features for each patch */
    feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((int8 *)data_6 +
					 sizeof(CTDB_GRID_PG_DATA_6));
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch1_offset = (patch_bytes + sizeof(CTDB_GRID_PG_DATA_5)) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }
    else
    {
	data_5->patch1_offset = sizeof(CTDB_GRID_PG_DATA_5) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }

    patch_number++;
    if (pg->patch1_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch1_offset);
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch2_offset = data_5->patch1_offset +
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch2_offset = data_5->patch1_offset;
    }

    patch_number++;
    if (pg->patch2_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch2_offset);
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch3_offset = data_5->patch2_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch3_offset = data_5->patch2_offset;
    }

    patch_number++;
    if (pg->patch3_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch3_offset);
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }	  

    size = (int8 *)feat_5 - (int8 *)data_5;
    return size;
}


/* Convert all data, including data header, for a tinned PG from format 6 to
 * format 5.  base is the address where the data header should start.
 * Returns number of bytes used for patch group.
 */
static int32
convert_tin_pg_data_6_5(ctdb_6, ctdb_5, group, base)
    CTDB_6 *ctdb_6;
    CTDB_5 *ctdb_5;
    int32   group;
    ADDRESS base;
{
    int32 i;
    int32 size;
    CTDB_TIN_PG_DATA_6 *data_6;
    CTDB_TIN_PG_DATA_5 *data_5;
    CTDB_PATCH_GROUP_5 *pg;
    CTDB_FEATURE_DATA_6 *feat_6;
    CTDB_FEATURE_DATA_5 *feat_5;
    int32 patch_bytes = 0;
    int32 delta;
    int32 te_offset_6;
    int32 te_offset_5;
    int32 patch_number;

    data_6 = ctdb_6->patch_groups[group].data_start.tin_data;
    data_5 = (CTDB_TIN_PG_DATA_5 *)base;
    size = sizeof(CTDB_TIN_PG_DATA_5);
    patch_number = group * CTDB_PATCHES_PER_GROUP_5;

    pg = ctdb_5->patch_groups + group;

    /* 1.  Fill in as much of the data header as we can. */
    bzero(data_5, size);
    bcopy(data_6->max_elev, data_5->max_elev, 
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(float32));
    
    bcopy(data_6->water_elev, data_5->water_elev,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(float32));

    bcopy(data_6->tables, data_5->tables,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(CTDB_PATCH_TABLE_REF_5));

    /* Can't fill in offsets yet... */

    bcopy(data_6->feature_present, data_5->feature_present,
	  CTDB_MAX_POSTS_PER_PATCH_5 * 2);
    
    bcopy(data_6->num_te, data_5->num_te,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(uint16));
    
    /* Can't fill in te_offsets yet */

    bcopy(data_6->te_vgrid_dim, data_5->te_vgrid_dim,
	  CTDB_PATCHES_PER_GROUP_5 * sizeof(uint8));
    
    /* Can't fill in vgrid_offsets yet */

    /* 2.  Convert the features for each patch */
    feat_5 = (CTDB_FEATURE_DATA_5 *)((int8 *)base + size);

    if (pg->patch0_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((int8 *)data_6 +
					 sizeof(CTDB_TIN_PG_DATA_6));
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch1_offset = (patch_bytes + sizeof(CTDB_TIN_PG_DATA_5)) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }
    else
    {
	data_5->patch1_offset = sizeof(CTDB_TIN_PG_DATA_5) /
	  sizeof(CTDB_FEATURE_DATA_5);
    }

    patch_number++;
    if (pg->patch1_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch1_offset);
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch2_offset = data_5->patch1_offset + 
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch2_offset = data_5->patch1_offset;
    }

    patch_number++;
    if (pg->patch2_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch2_offset);
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));

	data_5->patch3_offset = data_5->patch2_offset +
	  (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }
    else
    {
	data_5->patch3_offset = data_5->patch2_offset;
    }

    patch_number++;
    if (pg->patch3_occupied)
    {
	feat_6 = (CTDB_FEATURE_DATA_6 *)((CTDB_FEATURE_DATA_6 *)data_6 +
					 data_6->patch3_offset);
	
	patch_bytes = convert_feat_6_5(ctdb_6, ctdb_5, feat_6, feat_5);

	feat_5 = feat_5 + (patch_bytes / sizeof(CTDB_FEATURE_DATA_5));
    }	  

    /* 3. Calculate how far the TE's for the patch have moved relative to
     * where they used to be
     */
    if (!pg->patch0_no_tes)
    {
	te_offset_6 = data_6->te_offset[0];
    }
    else if (!pg->patch1_no_tes)
    {
	te_offset_6 = data_6->te_offset[1];
    }
    else if (!pg->patch2_no_tes)
    {
	te_offset_6 = data_6->te_offset[2];
    }
    else
    {
	te_offset_6 = data_6->te_offset[3];
    }
    
    te_offset_5 = feat_5 - (CTDB_FEATURE_DATA_5 *)data_5;
    
    /* Now, calculate the offset, in words */
    delta = te_offset_5 - te_offset_6;
    
#ifdef DEBUG
    if (delta > 0)
      printf("Unreasonable data offset in 6->5 conversion.\n");
#endif /* DEBUG */
    
    /* 4.  Fill out the rest of the offset values. */
    for(i=0;i<CTDB_PATCHES_PER_GROUP_5;i++)
    {
	data_5->te_offset[i] = data_6->te_offset[i] + delta;
	data_5->te_vgrid_offset[i] = data_6->te_vgrid_offset[i] + delta;
    }
    
    /* 5.  Copy over all the TE data. */
    {
	int32 data_size;
	int32 te_data_size;	/* In bytes */

	if (group < ctdb_6->max_patch_number)
	{
	    
	    /* Total bytes of data stored for this PG */
	    data_size = 
	      (int8 *)ctdb_6->patch_groups[group+1].data_start.grid_data -
	      (int8 *)data_6;
	    
	    te_data_size = data_size - 
	      (data_6->te_offset[0] * sizeof(CTDB_FEATURE_DATA_6));
	}
	else
	{
	    int32 vgrid_size;
	    
	    if (!pg->patch3_no_tes)
	    {
		data_size = data_6->te_vgrid_offset[3] *
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[3] * data_6->te_vgrid_dim[3];
	    }
	    else if (!pg->patch2_no_tes)
	    {
		data_size = data_6->te_vgrid_offset[2] * 
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[2] * data_6->te_vgrid_dim[2];
	    }
	    else if (!pg->patch1_no_tes)
	    {
		data_size = data_6->te_vgrid_offset[1] * 
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[1] * data_6->te_vgrid_dim[1];
	    }
	    else
	    {
		data_size = data_6->te_vgrid_offset[0] * 
		  sizeof(CTDB_FEATURE_DATA_6);
		vgrid_size = data_6->te_vgrid_dim[0] * data_6->te_vgrid_dim[0];
	    }

	    /* Add padding as necessary */
	    vgrid_size += (vgrid_size % sizeof(CTDB_FEATURE_DATA_6));
	    data_size += vgrid_size;		
	    te_data_size = (data_size -
			    (te_offset_6 * sizeof(CTDB_FEATURE_DATA_6)));
	}

	feat_6 = (CTDB_FEATURE_DATA_6 *)data_6 + te_offset_6;
	bcopy(feat_6, feat_5, te_data_size);
	
	/* Calculate total bytes in this PG */
	size = ((int8 *)feat_5 - (int8 *)data_5) + te_data_size;
    }

    return size;
}



/* Convert patch group header from fmt6 to 5.  base is where the new PG
 * should point.
 */
static void
convert_pg_6_5(ctdb_6, ctdb_5, group, base)
    CTDB_6 *ctdb_6;
    CTDB_5 *ctdb_5;
    int32   group;
    ADDRESS base;
{
    CTDB_PATCH_GROUP_5 *pg_5;
    CTDB_PATCH_GROUP_6 *pg_6;

    pg_5 = ctdb_5->patch_groups + group;
    pg_6 = ctdb_6->patch_groups + group;
    
    bzero(pg_5, sizeof(CTDB_PATCH_GROUP_6));

    pg_5->data_start.grid_data = (CTDB_GRID_PG_DATA_5 *)base;
    
    /* We'll leave the cache stuff blank for now
     */

    pg_5->contains_tins = pg_6->contains_tins;

    pg_5->patch0_occupied = pg_6->patch0_occupied;
    pg_5->patch0_dyn_terr = pg_6->patch0_dyn_terr;
    pg_5->patch0_te16 = pg_6->patch0_te16;
    pg_5->patch0_no_tes = pg_6->patch0_no_tes;
    pg_5->patch0_water = pg_6->patch0_water;

    pg_5->patch1_occupied = pg_6->patch1_occupied;
    pg_5->patch1_dyn_terr = pg_6->patch1_dyn_terr;
    pg_5->patch1_te16 = pg_6->patch1_te16;
    pg_5->patch1_no_tes = pg_6->patch1_no_tes;
    pg_5->patch1_water = pg_6->patch1_water;

    pg_5->patch2_occupied = pg_6->patch2_occupied;
    pg_5->patch2_dyn_terr = pg_6->patch2_dyn_terr;
    pg_5->patch2_te16 = pg_6->patch2_te16;
    pg_5->patch2_no_tes = pg_6->patch2_no_tes;
    pg_5->patch2_water = pg_6->patch2_water;

    pg_5->patch3_occupied = pg_6->patch3_occupied;
    pg_5->patch3_dyn_terr = pg_6->patch3_dyn_terr;
    pg_5->patch3_te16 = pg_6->patch3_te16;
    pg_5->patch3_no_tes = pg_6->patch3_no_tes;
    pg_5->patch3_water = pg_6->patch3_water;

    return;
}

/* Converts PG header and all PG data from format 6 to format 5.  group is
 * the patch group number to convert, base is an in-out variable passing
 * in the address where the format 5 data should start, and passing out
 * where the format 5 data for the next PG should start.
 */
static int32
convert_group_6_5(ctdb_6, ctdb_5, group, base)
    CTDB_6  *ctdb_6;
    CTDB_5  *ctdb_5;
    int32    group;
    ADDRESS *base;
{
    int32 bytes;
    
    /* First, convert the header. */
    convert_pg_6_5(ctdb_6, ctdb_5, group, *base);

    /* Now convert the data and the data headers */
    if (ctdb_5->patch_groups[group].contains_tins)
    {
	bytes = convert_tin_pg_data_6_5(ctdb_6, ctdb_5, group, *base);
    }
    else
    {
	bytes = convert_grid_pg_data_6_5(ctdb_6, ctdb_5, group, *base);
    }
    
    *base = (int8 *)*base + bytes;
    return bytes;
}



/* Converts PGs and features (including TINs) from format 6 to format 5.
 * Returns total patch data bytes in format 6 result.
 */
static int32
convert_features_6_5(ctdb_6, ctdb_5)
    CTDB_6 *ctdb_6;
    CTDB_5 *ctdb_5;    
{
    int32 groups;
    int32 i;
    int32 total_bytes = 0;
    ADDRESS base;
	
    groups = ctdb_5->patches_wide * ctdb_5->patches_high /
      CTDB_PATCHES_PER_GROUP_5;
	
    /* Going from 6 to 5, features can only shrink */
    ctdb_5->n_patch_bytes = ctdb_6->n_patch_bytes;
	
    ctdb_5->patch_data = (ADDRESS)ctdb_alloc(NULL,
					     ctdb_5->n_patch_bytes,
					     "Fmt. 5 patch data");

    base = ctdb_5->patch_data;

    for(i=0;i<groups;i++)
    {
	total_bytes += convert_group_6_5(ctdb_6, ctdb_5, i, &base);
    }

    return total_bytes;    
}

static int32
get_post_index_6(ctdb, norm_x, norm_y)
    CTDB_6 *ctdb;
    int32   norm_x;
    int32   norm_y;
{
    int32 page, index;
    
    page = (norm_x>>SIDE_SHIFT) +  
      ctdb->pages_wide * (norm_y>>SIDE_SHIFT);
    index = (norm_x&SIDE_MASK) + POSTS_PER_SIDE * 
      (norm_y&SIDE_MASK);
    return (page*POSTS_PER_PAGE + index);
}

static void
convert_elevation_posts_6_5(ctdb_6, ctdb_5)
    CTDB_6 *ctdb_6;
    CTDB_5 *ctdb_5;
{
    int32 patch;
    int32 px;
    int32 py;
    int32 base_px;
    int32 base_py;
    int32 *post_5, *post_6;
    int32 index;
    float64 elevation;

    for(patch=0;patch<ctdb_6->max_patch_number;patch++)
    {
	base_py = (patch / ctdb_6->patches_wide) * ctdb_6->patch_incr;
	base_px = (patch % ctdb_6->patches_wide) * ctdb_6->patch_incr;

	for(py=0;py<ctdb_6->patch_incr;py++)
	  for(px=0;px<ctdb_6->patch_incr;px++)
	  {
	      /* The index for format 5 is the same as that for format 6.
	       */
	      index = get_post_index_6(ctdb_6, base_px + px, base_py + py);
	      post_5 = ctdb_5->elevations + index;
	      post_6 = ctdb_6->elevations + index;

	      /* Copy the post data over and reassign the elevation.
	       * For format 6, one bit of elevation data was stolen and used
	       * to indicate the diagonalization for that post.
	       */
	      *post_5 = *post_6;

	      /* Clear the elevation data.
	       */
	      *post_5 &= ~0x0007FFFF;

	      elevation = GET_POST_ELEVATION_6(*post_6,
					       ctdb_6->fixed_point_basis);

	      *post_5 |= BUILD_POST_ELEVATION_5(elevation,
						ctdb_5->inv_fixed_point_basis);
	  }
    }
    return;    
}

