/*
 * $RCSfile: ct_cmp_fmt7.h,v $ $Revision: 1.1.1.1 $ $State: Exp $ 
 */
/****************************************************************************
 *   File: cmp_fmt7.h                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 7 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/
#ifndef _CMP_FMT7_H_INCLUDED
#define _CMP_FMT7_H_INCLUDED

#define CTDB_PATCHES_PER_GROUP_7       4  /* Num of patches per patch group */

#define CTDB_MAX_FEATURE_PRESENT_PER_PG_7 60 

#define CTDB_MAX_POSTS_PER_PATCH_7    25   /* maximum number of posts per
					    * patch
					    */

#define CTDB_NAME_LENGTH_7            14   /* same as Simnet */
#define CTDB_DATE_LENGTH_7            28   /* Enough room for ctime() */

#define CTDB_PATCH_CACHES_7           16   /* Number of patch data caches */
#define CTDB_CACHE_BKT_UNITS_7         4

#define CTDB_ILLEGAL_SOIL_7 0;
#define CTDB_SIMNET_SOIL_7(x)  ((x) & 0x0000000f);

#define CTDB_STANDARD_FIXED_POINT_BASIS_7 (5000.0 / (1 << 18))

/* Elevation is bits 0-18 (sign 18, magnitude 0-17) */
#define GET_POST_ELEVATION_7(p,basis)                           \
  ((((p) & 0x00040000) ? -((p) & 0x0003FFFF) : ((p) & 0x0003FFFF)) * (basis))

#define BUILD_POST_ELEVATION_7(e,inv_basis)                     \
  (((e) >= 0) ?                                                 \
   (((int32)((e) * (inv_basis))) & 0x0003FFFF) :                \
   ((((int32)((-e) * (inv_basis))) & 0x0003FFFF) | 0x00040000))

/* Diagonalization flag is bit 19.  This bit is set if the diagonalization
 * for this post is SW-NE. */
#define POST_SW_NE_7        0x00080000

/* The data that will be retained in the elevation post structure
 * in format 7.
 */
#define POST_ELEV_MASK_7 0x00FFFFFF

/* SW soil is bits 0-15 */
#define GET_POST_SW_SOIL_7(p) ((p) & 0xFFFF)
#define BUILD_POST_SW_SOIL_7(s) ((s) & 0xFFFF)

/* SE soil is bits 0-15 */
#define GET_POST_SE_SOIL_7(p) ((p) & 0xFFFF)
#define BUILD_POST_SE_SOIL_7(s) ((s) & 0xFFFF)

/* NE soil is bits 16-31 */
#define GET_POST_NE_SOIL_7(p) (((p) >> 16) & 0xFFFF)
#define BUILD_POST_NE_SOIL_7(s) (((s) & 0xFFFF) << 16)

/* NW soil is bits 16-31 */
#define GET_POST_NW_SOIL_7(p) (((p) >> 16) & 0xFFFF)
#define BUILD_POST_NW_SOIL_7(s) (((s) & 0xFFFF) << 16)

/* 3 necessary macros for ctdb_lookup_soilmap_7 */
#define PG_NUMBER_7(patch) ((patch) >> 2)
#define PG_MEMBER_7(patch) ((patch) & 0x3)

/* CTDB_GET_GRID_SOILMAP and CTDB_GET_TIN_SOILMAP are used by CTDB_GET_SOILMAP
 * to extract the soilmap for the patch specified by the patch group pointer
 * and the member number.  They are defined in libctdb_local.h.
 */
#define CTDB_GET_SOILMAP_7(ctdb,pg,mem) \
    (int32 *)((pg)->contains_tins ? \
             ( CTDB_GET_TIN_SOILMAP((ctdb),(pg),(mem)) ) : \
             ( CTDB_GET_GRID_SOILMAP((ctdb),(pg),(mem)) ) )  

/* The feature classes each imply certain intervisibility semantics
 * and storage techniques:
 *
 * FC_LINEAR is an individual tree or a treeline (depending upon the
 * number of vertices), or any non-laid linear feature (fences, wire,
 * dragon's teeth).
 *
 * An individual tree is described
 *	FEAT_INFO LINEAR_INFO XY Z
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature; XY identifies the location of the tree
 * within the patch; Z identifies the elevation of the top of the
 * tree.  The reduction of visibility due to tree blockage is done
 * using a simple light transmittance model.
 *
 * A treeline (or linear feature) is described
 *	FEAT_INFO LINEAR_INFO XY Z XY Z XY Z...
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature and whether or not there are trees at each end
 * of a treeline; (note that we do not need to store the end trees as
 * individuals since they are implied by the treeline); each XY point
 * identifies the location of one vertex in the treeline; the corresponding Z
 * identifies the elevation of the top of the treeline at that vertex.
 *
 * FC_VOLUME is a building or a building-like structure which blocks
 * intervisibility.  Telephone poles, water towers, and other such
 * structures which do not interrupt intervisibility are omitted.
 *
 * A volume is described
 *	FEAT_INFO VOLUME_INFO XY Z XY Z ...
 * Each XY point identifies the location of one vertex in the bounding
 * box of the building; the corresponding Z identifies the elevation
 * of the roof line of the building at that point.  Intervisibility is
 * interrupted by the line segments connecting these points exactly the
 * same way as the segments connecting grid posts.  The VOLUME_INFO
 * contains the type of feature, a reference to a model set, and the damage
 * state.
 *
 * FC_MICRO is a microterrain triangle.
 *
 * microterrain is described
 *	INFO MICRO_SOIL XY Z MICRO
 * Microwater is identical to microterrain, except that it has an MWATER_CHARS
 * at the end:  
 *      INFO MICRO_SOIL XY Z MICRO MWATER_CHARS 
 * The XY and Z combination identify the location of one vertex of
 * the microterrain triangle.  Up to three subsequent vertices are listed
 * in the MICRO structure.  There are edges connecting this point to the
 * each of the other listed vertices.  If there are two vertices listed
 * the two edges from this point form a triangle.  If there are three
 * listed, the three edges form two triangles (the first edge being
 * shared between the two).  The MICRO_SOIL structure identifies either
 * one or two soil types according to the number of vertices listed. 
 * For microterrain, intervisibility is interrupted by the edges connecting
 * these points exactly the same way as the segments connecting grid posts.
 *
 * FC_CANOPY is a canopy polygon.  See-through canopies are omitted.
 *
 * Canopies are described
 *      INFO XY Z CANOPY
 * Canopies are basically microterrain, but store no soil information, and
 * instead store bits specifying whether each edge is an exterior edge
 * and a model reference.
 * For canopies, visibility is degraded as with treelines.  
 *
 * FC_LAID_LINEAR is a road or river segment.
 *
 * Road segments and river segments are described the same way
 *	INFO LL_SOIL LAID_LINEAR XY XY...
 * LL_SOIL stores the soil type of the road or river segment.  LAID_LINEAR
 * identifies the width at the beginning and end of the segment
 * (linear taper is assumed).  It also stores an edge key for the
 * segment which provides the low-order 12 bits of the topologic edge 
 * (of type CTDB_EDGE) which uses this linear feature.  Each XY point
 * identifies the location of one vertex in the segment.  Linear
 * features do not impact intervisibility.
 */

#define FC_CLASS_MIN_7     0
#define FC_MICRO_7         0
#define FC_VOLUME_7        1
#define FC_LINEAR_7        2
#define FC_CANOPY_7        3
#define FC_LAID_LINEAR_7   4
#define FC_AGGREGATE_7     5
#define FC_CLASS_MAX_7     6

/*  HACK!  These are included just to make inter-format conversion easier */
#define M_FEATURE_CLASS_7   0xE0000000
#define S_FEATURE_CLASS_7   29
#define M_VERTICES_7        0x1E000000
#define M_MICRO_VERTICES_7  0x06000000
#define S_VERTICES_7        25

#define MICRO_VERTS_7         0x3
#define MICRO_TYPE_7          0xC
#define MICRO_TYPE_BASE_7     0x0
#define MICRO_TYPE_ML_7       0x4
#define MICRO_TYPE_DEFAULT_7  0x8
#define MICRO_TYPE_WATER_7    0xC

typedef struct ctdb_feature_info_7
{
    /* One of the FC_x values listed above.
     */
    unsigned feature_class : 3;

    /* For FC_LINEAR, FC_VOLUME, FC_CANOPY, and FC_LAID_LINEAR this
     * is the number of vertices.  If the actual objects has more than
     * 16 vertices, it must be broken into two or more objects.  For
     * FC_MICRO this is the number of positions which are connected to
     * this point bitwise-OR'd with a microterrain type.
     */
    unsigned vertices : 4;

    /* This is used as a high resolution bitwise bounding volume test.
     * Each post in a patch has a bit.  This mask identifies which
     * posts' polygons were touched by this feature.  Although Simnet
     * terrain databases only have 16 posts per patch, CCTT database
     * may have 25 (given the 100m interval required in the spec).
     */
    unsigned post_mask : 25;
} CTDB_FEATURE_INFO_7; 

#define CTDB_TREE_HEIGHT_BASIS_7 (200.0 / (1<<14))

/* Number of bits used to define fullness.
 */
#define CTDB_FULLNESS_BITS_7  8

#define M_TREE_FIRST_7 0x8000
#define M_TREE_LAST_7  0x4000

typedef struct ctdb_volume_info_7
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.
     */
    unsigned reference    : 16;
    unsigned type         : 5;
    unsigned avoid        : 1;
    unsigned reserved     : 10;
} CTDB_VOLUME_INFO_7;


typedef struct ctdb_linear_info_7
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.  `first' and `last' are only used if
     * the feature is a treeline, and specify whether or not individual
     * trees should be placed at each end.
     */
    unsigned reference        : 16;
    unsigned first            : 1;
    unsigned last             : 1;
    unsigned reserved         : 14;
} CTDB_LINEAR_INFO_7;


/* For now we store feature elevations as simple floats, since there
 * is no obvious way to use the extra 13 bits if we were to instead
 * store it using the fixed point convention used for grid posts.
 * Floating point will, of course, be faster than storing them in
 * fixed point since only a SP->DP conversion will be necessary to use
 * the data.
 */

#define M_VERT0_OUTER_7  0x01000000
#define M_VERT1_OUTER_7  0x02000000
#define M_VERT2_OUTER_7  0x04000000
#define M_CAN_EDGE_REF_7 0xF8000000
#define S_CAN_EDGE_REF_7 27

typedef struct ctdb_micro_pts_7
{
    /* Each microterrain vertex is connected to 0,1,2, or 3 other
     * microterrain vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, vert0 will
     * contain the index of that vertex relative to patch start; vert1,
     * vert2 will be 0.  If it is connected to 2 others, then vert0 
     * and vert1 will contain the indices of those vertices relative 
     * to patch start and so on.
     */
    uint8 reserved;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
} CTDB_MICRO_PTS_7;

/* For use with FC_MICRO.  We'll use the per-patch soil table like we do for
 * TINs.  Will be used for variable feature attributes in a future AFA 
 * integration. 
 */
typedef struct ctdb_var_attr_two_soils_7 
{
    unsigned reserved          : 16; /* Reserved for use by advanced features 
				      * & attributes
				      */
    unsigned poly_char_index0  : 8; 
    unsigned poly_char_index1  : 8;
} CTDB_VAR_ATTR_TWO_SOILS_7;

/* Microwater characteristic references */
typedef struct ctdb_mwater_chars_7
{
    unsigned char0 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
    unsigned char1 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
} CTDB_MWATER_CHARS_7;

typedef struct ctdb_canopy_pts_7
{
    /* Each canopy vertex is connected to 0,1,2, or 3 other
     * canopy vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then vert0 will 
     * contain the index of that vertex relative to patch start; vert1, 
     * vert2 will be 0.  If it is connected to 2 others, then vert0 
     * and vert1 will contain the indices of those vertices relative 
     * to patch start and so on.
     */
#ifdef USE_LITTLE_ENDIAN
    uint8 vert2;   
    uint8 vert1; 
    uint8 vert0;
    unsigned vert0_outer  : 1;
    unsigned vert1_outer  : 1;
    unsigned vert2_outer  : 1;
    unsigned can_edge_ref : 5;
#else
    unsigned can_edge_ref : 5;
    unsigned vert2_outer  : 1;
    unsigned vert1_outer  : 1;
    unsigned vert0_outer  : 1;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
#endif
} CTDB_CANOPY_PTS_7;

typedef struct ctdb_laid_linear_info_7 
{
    /* Each linear segment has a width.  The width is in the same
     * units as x,y in the XY_PATCH_OFFSET structure.  The edge_key
     * provides the low-order 12 bits of the topologic edge 
     * (of type CTDB_EDGE) which uses this linear feature.
     */
    uint16    width;
    unsigned  reserved : 4;
#define CTDB_EDGE_KEY_7 4096 /* 2^12 */
    unsigned  edge_key : 12;
} CTDB_LAID_LINEAR_INFO_7;

/* For use with laid_linears.  We will use all 16 bits for a single index
 * to avoid the indirection.  Will be used for variable feature attributes
 * in a future AFA integration. 
 */
typedef struct ctdb_var_attr_one_soil_7
{
    unsigned var_attr_index    : 16;
    unsigned poly_char_index   : 8;    
    unsigned reserved          : 8;
} CTDB_VAR_ATTR_ONE_SOIL_7;

typedef struct ctdb_patch_hdr_7
{
    /* At the beginning of each patch, we store the size (the number
     * of CTDB_FEATURE_DATA things in the patch).  There will never be
     * more than 2^16 of those, so that leaves 16 bits of padding.
     * I'm sure we'll find a use for those bits eventually.
     */
    uint16 size;
    uint16 _pad;
} CTDB_PATCH_HDR_7;

#define CTDB_ABSTRACT_CANOPY_7                1
#define CTDB_ABSTRACT_SOIL_DEFRAG_7           2
#define CTDB_ABSTRACT_STEEP_SLOPE_7           3
#define CTDB_ABSTRACT_RAILROAD_7              4
#define CTDB_ABSTRACT_PIPELINE_7              5
#define CTDB_ABSTRACT_POLITICAL_BOUNDARY_7    6
#define CTDB_ABSTRACT_POWERLINE_7             7
#define CTDB_ABSTRACT_LABEL_7                 8
#define CTDB_ABSTRACT_TACTICAL_SIGN_7         9
#define CTDB_ABSTRACT_OFFROAD_SEGMENT_7       10
#define CTDB_ABSTRACT_MAX_7 /* Always Last */ 11


/* CTDB_ABSTRACT_FEATURE_INFO:
 *
 * Structure at the start of each abstract feature.
 */
#define CTDB_ABSTRACT_QUAD_NODE_7             0
typedef struct ctdb_abstract_feature_info_7
{
    /* Type code for the abstract feature.  The code 0 is reserved
     * to indicate that this "feature" is actually a reference to
     * four child quad nodes.
     */
    uint8 type_code;

    /* The size (in 4-byte units) of the CTDB_ABSTRACT_DATA associated
     * with the abstract feature.
     */
    uint8 data_size;

    /* The number of vertices in feature.  There must be at least one,
     * except for "features" which are of the type CTDB_ABSTRACT_QUAD_NODE.
     */
    uint16 num_vertices;

} CTDB_ABSTRACT_FEATURE_INFO_7;

/* CTDB_XYO_COORD2D
 * 
 * Structure used to represent a vertex point (x,y) in XYO coordinates 
 * in either a feature or terrain element data structure.
 */
typedef struct ctdb_xyo_coord2d_7  
{
    uint16 x;
    uint16 y;
} CTDB_XYO_COORD2D_7;

/* CTDB_ABSTRACT_VERTEX:
 *
 * Structure used to represent a vertex of an abstract feature.
 * Note that for efficient access, the format used for vertices
 * is 32 bit floating point, not fix-point patch offsets.  This
 * results in a loss of precision at the far corners of the database,
 * but avoids a great deal of translation at run time.
 */
typedef float32 CTDB_ABSTRACT_VERTEX_7;

/* CTDB_ABSTRACT_DATA:
 *
 * Format of data associated with so-called abstract features.  Note
 * that this structure is analogous to a protocol definition, in that
 * the final record of each union substructure can be variable length.
 *
 * This goes into a 4-byte union: NO MEMBER STRUCTURE CAN EXCEED 32 BITS!
 *
 */
typedef union ctdb_abstract_data_7
{
    struct ctdb_quad_data_7
    {
	/* Array of 4 offsets to child quad nodes (in 4-byte units)
	 * from the beginning of the quad tree data storage.
	 */
	uint32 child_start[1];  /* really 4 members, not 1 */
    } quad_data;
    struct ctdb_canopy_data_7
    {
	unsigned impenetrable  : 1;
	unsigned padding       : 31;
    } canopy;
    struct ctdb_soil_defrag_data_7
    {
	unsigned soil_index    : 4;
	unsigned level         : 4;
	unsigned padding       : 8;
	unsigned soilmap_index : 16;
    } soil_defrag;
    struct ctdb_steep_slope_data_7
    {
	unsigned slope_degrees : 7;
	unsigned padding       : 25;
    } steep_slope;
    struct ctdb_label_data_7
    {
	char text[4]; /* variable length, NULL terminated */
    } label;
    struct ctdb_offroad_data_7
    {
	uint32 edge_number;
    } offroad;

} CTDB_ABSTRACT_DATA_7;


typedef union ctdb_feature_data_7
{
    uint32                       integer;
    CTDB_PATCH_HDR_7             hdr;
    CTDB_FEATURE_INFO_7          info;
    CTDB_VOLUME_INFO_7           volume;
    CTDB_LINEAR_INFO_7           linear;
    CTDB_LAID_LINEAR_INFO_7      laid_linear;
    CTDB_VAR_ATTR_ONE_SOIL_7     ll_soil;
    CTDB_XYO_COORD2D_7           xy;
    float32                      z;
    CTDB_MICRO_PTS_7             micro;
    CTDB_CANOPY_PTS_7            canopy;
    CTDB_VAR_ATTR_TWO_SOILS_7    micro_soil;
    CTDB_ABSTRACT_FEATURE_INFO_7 abstract;
    CTDB_ABSTRACT_DATA_7         abstract_data;
    CTDB_ABSTRACT_VERTEX_7       abstract_vertex;
    CTDB_MWATER_CHARS_7          mwater;
} CTDB_FEATURE_DATA_7;

/* Definitions of feature types.
 */

/* Types of volume features.
 */
#define CTDB_FT_BUILDING_7          0
#define CTDB_FT_ROCK_DROP_7         1

/* Types of linear features.
 */
#define CTDB_FT_TREELINE_7          0
#define CTDB_FT_CONCERTINA_WIRE_7   1
#define CTDB_FT_FENCE_7             2
#define CTDB_FT_MINEFIELD_FENCE_7   3
#define CTDB_FT_DRAGONS_TEETH_7     4

/* Types of dragon's tooth components.
 */
#define CTDB_DT_STEEL_HEDGEHOG_7       0
#define CTDB_DT_STEEL_TETRAHEDRON_7    1
#define CTDB_DT_CONCRETE_BLOCK_7       2
#define CTDB_DT_CONCRETE_TETRAHEDRON_7 3

/* CTDB_EDGE:
 *
 * Structure used to collect a list of linear features which form an
 * uninterrupted "edge" between two intersections.  The `feature'
 * member is actually variable length.
 */
typedef struct ctdb_edge_7
{
    /* Total distance along the segment (useful for planning) */
    int32 distance;

    /* Indices of the nodes at the beginning and end of the segment */
    int32 first_node;
    int32 last_node;

    int32 num_features;
    struct ctdb_feature_reference_7
    {
	/* The patch where this feature is stored */
	uint32 patch_number;

	/* Byte offset into the patch */
	unsigned patch_offset : 24;

	/* Whether the vertices in the patch are in reverse order */
	unsigned reverse      : 1;

	unsigned padding      : 7;
    } feature[1];
} CTDB_EDGE_7;

/* CTDB_NODE:
 *
 * Represents the topologic connection between CTDB_EDGE structures.
 * The `connection' member is actually variable length.
 */
typedef struct ctdb_node_7
{
    /* Approximate location of the node (useful for planning) */
    float32 x, y;

    /* Whether the node corresponds to a bridge */
    unsigned is_bridge : 1;

    unsigned padding   : 23;

    uint8 num_connections;
    struct ctdb_connection_7
    {
	unsigned node_in_other_cell     : 1;  /* TRUE indicates that the node
					       * index references a node in 
					       * another cell, FALSE indicates
					       * that it is in this cell.
					       */
	unsigned connection_crosses_mes : 1;  /* TRUE indicates that either the
					       * current node or the node 
					       * connected to is in an MES
					       * structure.  This dictates
					       * special usage of the edge
					       * structure below.
					       */
	unsigned padding            : 6;
	
	
	/* The node which this connects to */
	unsigned node_index         : 24;
	
	union edge_union_7
	{
	    int32 edge_index; /* The edge used to connect these nodes,
			       * if they are in the same database. 
			       */

	    int32 other_cell; /* If node_in_other_cell is TRUE,
			       * the cell in which node_index is valid.
			       */
	    uint32 aperture_key; /* If connection_crosses_mes is TRUE,
				  * this specifies which aperture in which
				  * MES structure must be traversed to move 
				  * from the current node to the connected
				  * node.
				  */
	} edge;
    } connection[1];

} CTDB_NODE_7;

/* CTDB_LINEAR_DESC:
 *
 * This structure is used to return information about linear models to the
 * application.
 *
 * It is also used to pass a linear feature or an individual tree
 * (n_verts == 1) to the feature compression algorithms.
 *
 * `trunk_radius' specifies the radius of the tree trunk.
 *
 * `foliage_height' specifies the height of the foliage of the tree.
 *
 * `fullness' specifies the fullness of the leaves or wire.  0.0 represents
 * zero fullness (nothing there), and 1.0 represents maximum fullness
 * (completely blocks light).
 *
 * `total_height' specifies the height of the feature.
 *
 * `spacing' represents how far apart from each other the dragon's teeth are
 * placed along the linear segment.
 *
 * `component' specifies the type of dragon's tooth, specified in libctdb.h
 *
 * `type' specifies the type of linear feature, specified in libctdb.h
 *
 * `width' specifies the foliage width for trees, or the width of dragon's
 * teeth.
 */
typedef struct ctdb_linear_desc_7
{
    union
    {
	struct
	{
            float32  trunk_radius;
            float32  foliage_height;
            float32  fullness;
	    float32  total_height;
	} treeline_info;

	struct
	{
            float32  fullness;
            int32    force_id;
	} var_opacity_info;

	struct
	{
            float32  spacing;
            int32    component;
	} dragons_tooth_info;
    } linear_data;

    int32   type;
    float32 width;

} CTDB_LINEAR_DESC_7;


typedef struct ctdb_linear_model_7
{
    /* This defines the models for the various linear features.
     */
    union
    {
	struct
	{
	    /* `treeline_info' is used for trees and treelines.
	     * 
	     * `trunk_radius' is measured in the same units as x,y in the
	     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
	     *
	     *  `foliage_height' is the distance from the top of the tree to
	     * the lowest branches measured in the unit
	     * `CTDB_TREE_HEIGHT_BASIS':
	     * 
	     * #define CTDB_TREE_HEIGHT_BASIS (200.0 / (1<<14))
	     *
	     *  `total_height' is the distance from the top of the tree to
	     * the ground measured in `CTDB_TREE_HEIGHT_BASIS'.  This is
	     * only used for trees in aggregate models.  Otherwise, their
	     * heights are derived from their Z values (tree tops).  Z
	     * values aren't stored with aggregate features.
	     *
	     * `fullness' refers to the "density" of the leaves, wire, or
	     * whatever the feature is made of.  This value ranges from zero
	     * (nothing there) to the maximum value (lets no light through).
	     */
	    unsigned trunk_radius     : 16;
	    unsigned foliage_height   : 14;
	    unsigned pad1             : 2;
	    unsigned fullness         : 8;
	    unsigned total_height     : 14;
	    unsigned pad2             : 10;
	} treeline_info;

	/* `var_opacity_info' is used for wire and fences.
	 */
	struct
	{
	    unsigned fullness         : 8;
	    unsigned pad1             : 24;
	    unsigned pad2             : 32;
	} var_opacity_info;

	/* `dragons_tooth_info' is used for dragon's teeth.
	 * 
	 * `spacing' represents how far apart from each other the dragon's
	 * teeth are placed along the linear segment.
	 *
	 * Dragon's teeth can be made of steel hedgehogs, steel tetrahedrons,
	 * concrete blocks, or concrete tetrahedrons.  Each of these affects
	 * intervisibility differently (hence the need for the `component'
	 * element).
	 */
	struct
	{
	    float32  spacing;
	    unsigned component        : 4;
	    unsigned pad              : 28;
	} dragons_tooth_info;
    } linear_data;

    /* `type' specifies the specific type of linear feature (i.e. treeline,
     * minefield fence, etc.).
     * 
     * The width of the foliage, the thickness of the concertina wire or
     * minefield fence, or the width of the dragon's teeth is stored in
     * `width', which is measured in the same units as X and Y in the
     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
     */
    uint8  type;
    uint8  pad;
    uint16 width;

} CTDB_LINEAR_MODEL_7;


typedef struct ctdb_volume_model_7
{
    char     model_name[32];
    uint32   model_id;
    uint32   next_state_index;
    float32  percent_damaged;
    uint16   num_vertices;
    uint16   mes_volume_index;
    float32  vertices[1];   /* variable-length */

} CTDB_VOLUME_MODEL_7;



/*============MULTI-ELEVATION SURFACE STRUCTURES============*/

#define CTDB_MAX_SCALING_7 1.073741e+09

typedef enum 
{
     CTDB_MES_FIRST_TYPE_7 = 0,
     CTDB_MES_UNKNOWN_7    = 0,
     CTDB_MES_BUILDING_7   = 1,
     CTDB_MES_CAVE_7       = 2,
     CTDB_MES_TUNNEL_7     = 3,
     CTDB_MES_BRIDGE_7     = 4,
     CTDB_MES_LAST_TYPE_7  = 5
} CTDB_MES_TYPE_7;


typedef int32    CTDB_MES_VERTEX_7[XYZ];
typedef float64  CTDB_MES_VERTEX_F64_7[XYZ];

typedef struct ctdb_mes_list_vertex_f64_7
{
    int32                 num_vertices;
    CTDB_MES_VERTEX_F64  *vertices;
} CTDB_MES_LIST_VERTEX_F64_7;

typedef struct ctdb_mes_triangle_7
{
#ifdef USE_LITTLE_ENDIAN
    uint16        padding;
    int16         material;        /* FACC MCC attributes codes 0-999 */
#else
    int16         material;        /* FACC MCC attributes codes 0-999 */
    uint16        padding;
#endif
    float64       normal[XYZ];
    int32         vertex_ids[3];
} CTDB_MES_TRIANGLE_7;

typedef struct ctdb_mes_aperture_7
{
    int32      num_triangles;
    int32      first_triangle_id;  /* The first in the set of triangles for 
                                    * this aperture */
    
    struct
    {
	int32   num_vertices; 
	int32  *vertex_ids;        /* Vertex ids that make up the exterior
			            * boundary of the aperture, clockwise order
			            */
    } outline;
    struct
    {
#ifdef USE_LITTLE_ENDIAN
	uint8      padding;
	uint8      step_to_enclosure;
	uint16     enclosure_id;
#else
        uint16     enclosure_id; /* An enclosure the aperture connects to */

    /* Whether there is a step from the bottom of this aperture to the floor
     * of the corresponding enclosure.  A value of TRUE means that the bottom 
     * of this aperture is at a different level from the floor of the 
     * enclosure, FALSE means that they are at the same level. */
        uint8      step_to_enclosure; 
	uint8      padding;
#endif
    /* The thickness of the wall from the plane of the aperture to the 
     * corresponding enclosure.  The two wall-thicknesses together add up
     * to the total thickness of the wall. */
        int32      wall_thickness;
    } connections[2];

#ifdef USE_LITTLE_ENDIAN
    int16      open;   
    int16      visible;
#else
    int16      visible;           /* 1 = transparent, 0 = opaque */
    int16      open;              /* 1 = open, 0 = closed */
#endif
} CTDB_MES_APERTURE_7;

typedef struct ctdb_mes_grid_box_7
{
   int32           left_front_bottom_corner[XYZ];
   int32           right_back_top_corner[XYZ];
   int32           outer_box_id;      /* parent of this box */

#ifdef USE_LITTLE_ENDIAN
   uint16          dimensions_x;      /* boxes in x direction */
   uint16          grid_index;        /* which child is this box */
   uint16          dimensions_z;      /* boxes in z direction */
   uint16          dimensions_y;      /* boxes in y direction */
#else
   uint16          grid_index;        /* which child is this box */
   uint16          dimensions_x;      /* boxes in x direction */
   uint16          dimensions_y;      /* boxes in y direction */
   uint16          dimensions_z;      /* boxes in z direction */
#endif
   int32          *box_ids;           /* array of children of this box */

#ifdef USE_LITTLE_ENDIAN
   uint16          padding1;
   uint16          num_apertures;  
#else
   uint16          num_apertures;  
   uint16          padding1;
#endif

   uint16         *aperture_ids;      /* ids of apertures intersecting 
	 	    		       * this box */
#ifdef USE_LITTLE_ENDIAN
   uint16          padding2;
   uint16          num_triangles;  
#else
   uint16          num_triangles;  
   uint16          padding2;
#endif
   int32          *wall_triangle_ids; /* ids of enclosure wall triangles
				       * intersecting this box */
} CTDB_MES_GRID_BOX_7;

typedef struct ctdb_mes_enclosure_7
{ 
    int32   average_height;
    struct
    {
	int32   num_vertices; 
	int32  *vertex_ids;       /* Array of vertex ids that outline the floor
			           * of the enclosure, in clockwise order */
    } footprint;

    int32 num_triangles; 
    int32 first_wall_triangle_id; /* The first in the list of triangles
                	 	   * for the walls of this enclosure 
	 			   */
#ifdef USE_LITTLE_ENDIAN
    uint16 padding;
    uint16 num_apertures; 
#else
    uint16 num_apertures; 
    uint16 padding;
#endif
    uint16 *aperture_ids;         /* Array of the ids of the apertures 
			           * connecting to this enclosure */
} CTDB_MES_ENCLOSURE_7;

typedef struct ctdb_mes_template_7
{    
    CTDB_MES_TYPE_7    type;
    float64            fixed_point_basis;   /* Scaling of local coordinates */
    int32              num_vertices;
    CTDB_MES_VERTEX_7 *all_vertices;        /* Array of all vertices:
					     * x,y,z, x,y,z, x,y,z, etc. */

    /* We choose to group the list of all triangles so that all 
     * wall triangles are at the end of this array of triangles.
     * Furthermore, the list of triangles will be grouped by enclosure,
     * so that all wall triangles of an enclosure are grouped together.
     * This allows us to specify the list of triangles for an enclosure 
     * merely by specifying the ID of the first triangle in the list.
     * Similarly, the list of aperture triangles is grouped by aperture.
     */
    int32                num_triangles;
    CTDB_MES_TRIANGLE_7 *all_triangles;     /* Array of all triangles, 
					    * both wall and aperture */
#ifdef USE_LITTLE_ENDIAN
    uint16              padding1;
    uint16              num_enclosures;
#else
    uint16              num_enclosures;
    uint16              padding1;
#endif
    CTDB_MES_ENCLOSURE_7 *all_enclosures;  /* Array of all enclosures */
#ifdef USE_LITTLE_ENDIAN
    uint16              padding2;
    uint16              num_apertures;
#else
    uint16              num_apertures;
    uint16              padding2;
#endif
    CTDB_MES_APERTURE_7 *all_apertures;     /* Array of all apertures */
    int32                num_boxes;
    CTDB_MES_GRID_BOX_7 *all_boxes;         /* Array of all gridboxes */
} CTDB_MES_TEMPLATE_7;

typedef struct ctdb_mes_volume_7
{
    float64                   origin[XYZC]; /* Origin of structure in GCS */
 /* Rotation matrix from GCS world space to mes local space*/  
    float64                   tfm_world_to_mes[XYZ][XYZ]; 
    CTDB_MES_LIST_VERTEX_F64_7 roofline;    /* Array of vertices in roofline */
    CTDB_MES_LIST_VERTEX_F64_7 floorline;   /* Array of vertices in floorline*/
    CTDB_MES_TEMPLATE_7       *template;    /* The mes-volume's description */
} CTDB_MES_VOLUME_7;



/* Defined JUST for use by the compiler */
#define CTDB_INVALID_VMODEL_7 -1
#define CTDB_INVALID_LMODEL_7 -1
#define CTDB_INVALID_VOL_ID   -1

/*********************************************************************
 * AGGREGATE MODEL DEFINITIONS                                       *
 *********************************************************************/

typedef struct ctdb_agg_model_ref_7
{
    uint8                    class; /* VOLUME, LINEAR */
    uint8                    reserved;
    uint16                   ref;   /* REFERENCE LINEAR, VOLUME,
				     * MES OR AGGREGATE MODEL LIBRARY
				     */
} CTDB_AGG_MODEL_REF_7;

typedef struct ctdb_agg_model_feature_7
{
    CTDB_AGG_MODEL_REF_7     model;
    uint16                   offset_x;
    uint16                   offset_y;
} CTDB_AGG_MODEL_FEATURE_7;

typedef struct ctdb_aggregate_model_7
{
    uint16                     n_features;
    uint16                     reserved;
    CTDB_AGG_MODEL_FEATURE_7   features[1];  /* variable-length array */
} CTDB_AGGREGATE_MODEL_7;

/**********************************************************************
 * Patch data headers                                                 *
 **********************************************************************/
/* The feature and TE data for each patch group is stored with a data header
 * at its beginning.  Because patch groups containing TINned patches need
 * substantially more data, there are two formats used, one for purely
 * gridded PGs, and one for PGs containing at least one TINned patch.
 */

typedef struct ctdb_patch_table_ref_7
{
    unsigned soil_table : 16;	/* Index of patch soil table */
    unsigned water_char : 16;	/* Index of patch water characteristics */
} CTDB_PATCH_TABLE_REF_7;


/* Wave characteristics */
typedef struct ctdb_wave_chars_7
{
    float32 height;		/* In meters */
    float32 period;		/* In seconds */
    float32 speed;		/* In meters per second */
    float32 direction;		/* $$$ units??? Tangent of some angle...*/
} CTDB_WAVE_CHARS_7;

/* Structure used to store surface characteristics for water polygons */
typedef struct ctdb_water_chars_7
{
    int16 tidal_zone_index;
    uint16          pad;
    float32         surface_temp;	/* In degrees C */
    float32         surf_height;	/* In meters */
    CTDB_WAVE_CHARS primary_wave;
    CTDB_WAVE_CHARS secondary_wave;
} CTDB_WATER_CHARS_7;

typedef struct ctdb_tidal_zone_7
{
    float32 offset;		/* Current tidal offset for this zone */
    /* $$$ add a region or identifier of some kind??? */
} CTDB_TIDAL_ZONE_7;

/* Data header for a completely gridded patch group.  This will *always*
 * be immediately followed by the feature data for the patch group.  The
 * starting address for the features for the first patch is
 * (address of this structure) + (sizeof(this structure)).
 * The starting address for the features for the other patches in the
 * PG can be found by
 * (address of this structure) + (patch*_offset, in 32-bit words)
 */
typedef struct ctdb_grid_pg_data_7
{
    float32                max_elev[CTDB_PATCHES_PER_GROUP_7];

    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32                water_elev[CTDB_PATCHES_PER_GROUP_7];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF_7 tables[CTDB_PATCHES_PER_GROUP_7];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;
} CTDB_GRID_PG_DATA_7;

/* Data header for patch group that contains at least one TINned patch.
 * This is *always* followd by the following (in order):
 *    Feature data for each patch in PG
 *    TE data for each patch in PG
 *    Virtual grid data for each patch in PG
 * The various addresses can all be calculated as described above for the
 * CTDB_GRID_PG_DATA structure (note that the offsets for everything
 * but features are explicitly stored for all 4 patches rather than just
 * the last 3).
 */
typedef struct ctdb_tin_pg_data_7
{
    float32       max_elev[CTDB_PATCHES_PER_GROUP_7];

    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32                water_elev[CTDB_PATCHES_PER_GROUP_7];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF_7 tables[CTDB_PATCHES_PER_GROUP_7];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;

    /* BITS             
     *	    3 : Features
     *	    2 : Buildings
     *	    1 : TREES
     *	    0 : Microterrain  
     */
    uint8   feature_present[CTDB_MAX_FEATURE_PRESENT_PER_PG_7];

    /* Array[patches] numb TE in patch */
    uint16  num_te[CTDB_PATCHES_PER_GROUP_7];
    /* Where the TEs for each patch start.  These are in units of 32-bit
     * words and relative to the starting address of this (ctdb_tin_pg_data)
     * data structure
     */
    uint32       te_offset[CTDB_PATCHES_PER_GROUP_7];
    /* SQRT(# of v. grid squares) for each patch */
    uint8        te_vgrid_dim[CTDB_PATCHES_PER_GROUP_7];
    /* Where the virtual grid starts for each patch.  These are in units
     * of 32-bit words and relative to the starting address of this
     * (ctdb_tin_pg_data) data structure
     */
    uint32       te_vgrid_offset[CTDB_PATCHES_PER_GROUP_7];
} CTDB_TIN_PG_DATA_7;


typedef union ctdb_pg_data_hdr_ptr_7
{
    struct ctdb_grid_pg_data_7 *grid_data;
    struct ctdb_tin_pg_data_7  *tin_data;
    int32                       offset;
} CTDB_PG_DATA_HDR_PTR_7;

/* A PATCH_GROUP is a collection of information about four contiguous
 * patches (contiguous in the sense x+y*w).  For each patch there are
 * four flags: occupied indicates there are features in the patch,
 * dyn_terr indicates that there are dynamic terrain changes to the
 * patch (future use), te16 indicates whether the patch's TEs use 8 or
 * 16 bit indices, and no_tes indicates whether or not a patch contains
 * TEs at all.  In addition, there is data about the patch group as a
 * whole:
 *     data_start points to the actual data for the cache.  Specifically,
 * it references either a CTDB_GRID_PG_DATA structure (contains_tins ==
 * FALSE) or a CTDB_TIN_PG_DATA structure (contains_tins == TRUE).  In either
 * case, that structure contains offsets for the feature/TE data itself,
 * which immediately follows the structure.
 *     cache specifies which of the 16 patch data caches this PG's data
 * should be placed in
 *     mapped is TRUE iff data_start points to actual data.  If it is FALSE,
 * the data must be read in from disk before being accessed.
 *     contains_tins is TRUE iff at least one patch in the PG is TINned
 */    
typedef struct ctdb_patch_group_7
{
    /* Bytes 0-3 */
    CTDB_PG_DATA_HDR_PTR_7     data_start;
    
    /* Bytes 4-7 */
    unsigned cache           : 4;  /* Which size cache bucket this PG uses */
    unsigned mapped          : 1;  /* 1 => PG data is in memory */
    unsigned contains_tins   : 1;  /* 1 => at least one patch in this PG
				    * is TINNed.  This determines what 
				    * data_start points to */
    unsigned patch0_occupied : 1;
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch0_no_tes   : 1;  /* 1 => patch 0 has 0 TEs */
    unsigned patch0_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch1_occupied : 1;
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch1_no_tes   : 1;  /* 1 => patch 1 has 0 TEs */
    unsigned patch1_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch2_occupied : 1;
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch2_no_tes   : 1;
    unsigned patch2_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch3_occupied : 1;
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch3_no_tes   : 1;
    unsigned patch3_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned unused          : 6;
} CTDB_PATCH_GROUP_7;


/**********************************************************************
 * PAT data structure                                                *
 **********************************************************************/
/* Different ways of looking at the data stored for each attribute */

/* Data structure used to store a column of the PAT table */
typedef struct ctdb_pat_column_hdr_7
{
  uint32                 facc;             /* FACC code for this column */
  union ctdb_pat_data_7 *data;             /* Attribute data */
} CTDB_PAT_COLUMN_HDR_7;	

/* PAT table data structure */
typedef struct ctdb_pat_table_7
{
    uint32 num_columns;
    CTDB_PAT_COLUMN_HDR_7 *columns;  /* array of column header structures,
				    * one for each column.
				    */
    uint32 num_entries;            /* Number of valid entries in table */
    uint32 alloced_entries;        /* Number of entries each column currently
				    * has alloced. (We will enforce the rule
				    * that all columns have the same # of 
				    * entries alloced at all times.) This will 
				    * support efficient run-time addition
				    * of entries.
				    */
} CTDB_PAT_TABLE_7;  

/* Different ways of looking at the data stored for each attribute in PAT*/
typedef union ctdb_pat_data_7
{
    uint32  uinteger;
    int32   integer;
    int32   boolean;
    char    character;
    float32 fp;
    int32   string_index;  /* Index into global string array */
    /* Add others as needed - none may exceed 32 bits */
} CTDB_PAT_DATA_7; 

/**********************************************************************
 * Patch data cache                                                   *
 **********************************************************************/
typedef struct ctdb_cache_bucket_7
{
    CTDB_PATCH_GROUP_7 *user;	/* Patch group whose data is stored here */
    int32               patch_data_start_disk; /* The offset (in bytes)
						* from the start of the
						* patch data section in the
						* database file on disk at
						* which this PGs data starts
						*/
    union ctdb_feature_data_7 data[1];	/* Actual data.  This will be malloced
					 * to be longer than 1 word, but 
					 * declaring it this way saves us a 
					 * 4 bytes per bucket.
					 */
} CTDB_CACHE_BUCKET_7;

typedef struct ctdb_7
{
    /* Fixed point basis for elevation points.  Can be changed to deal
     * with future databases which have extreme elevations, or extreme
     * accuracy requirements.
     */
    float64         fixed_point_basis; 
    float64         inv_fixed_point_basis;  
    
    /* Tables mapping local 4 or 8 bit soil types to the global list */
    int32             num_soil_tables;
    int32           **soil_tables;
    int32             soil_table_size;
    int8             *soil_table_storage;
    
    /* Pointer to the PAT table */
    struct ctdb_pat_table_7 *pat_table_ptr;  

    /* Table of sea-state combinations */
    int32                      num_water_chars;
    struct ctdb_water_chars_7 *water_chars;
    
    /* Table of tidal zones */
    int32                      num_tidal_zones;
    struct ctdb_tidal_zone_7  *tidal_zones;

    /* Spacing between grid posts */
    int32           incr;
    float64         inv_incr;
    float64         incrD;
    
    /* Spacing between patches */
    int32           patch_incr    ;/* number of posts per patch */
    float64         patch_inv_incr;/* 1 over patch_incr */
    float64         patch_incrD   ;/* double patch_incr */
    
    /* Database size */
    int32           pages_wide;
    int32           pages_high;
    int32           patches_wide;
    int32           patches_high;
    int32           min_x;
    int32           max_x;
    int32           min_y;
    int32           max_y;
    int32           min_x_post;
    int32           max_x_post;
    int32           min_y_post;
    int32           max_y_post;
    
    /* Useful trivia */
    int32           input_format;
    float64         min_z;
    float64         max_z;
    float64         origin_northing;
    float64         origin_easting;
    int32           origin_zone_number;
    int32           origin_zone_letter;
    int32           datum;
    int32           cell_id; /* added for gcs; 
			      * GCS_ILLEGAL_CELL implies non-gcs database 
			      */

    /* origin_x and origin_y were added for GCS.  They specify the offset
     * from the center of the cell at which x and y should be 0, respectively
     */
    int32           origin_x;
    int32           origin_y;

    /* The following specify, repsectively, the left and right longitudinal
     * boundaries of the database and the bottom and top latitudinal boundaries
     * of the database
     */
    float64         west_long;
    float64         east_long;
    float64         south_lat;
    float64         north_lat;

    CTDB_TDB_TYPE   gcs_mode;
    char            name[CTDB_NAME_LENGTH_7];
    int16           version;
    char            date[CTDB_DATE_LENGTH_7];
    int32           total_memory;
    uint32          is_curved;
    CTDB_PATCH_GROUP_7 *patch_groups;
    int32           elevations_resident;
    int32           elevation_seek;
    int32           elevation_cache_pages;
    int32           direct_map_modulus;
    int32           elevation_hit_count;
    int32           elevation_miss_count;
    int32          *cache_page_numbers;
    int32         **elevation_pages;

    /* Caches sizes are in units of CTDB_CACHE_BKT_UNITS */
    int32           patch_cache_size_map[CTDB_PATCH_CACHES_7];

    int32           patch_data_seek;
    int32           patch_cache_entries;
    int32           patch_hit_count;
    int32           patch_miss_count;
    struct ctdb_cache_bucket_7 *patch_data_cache[CTDB_PATCH_CACHES_7];

    /* Database file */
    FILE           *file;

    int32          *elevations;
    int32          *soils;

    /* Quadtree of abstract features */
    int32           num_quad_data;
    union ctdb_feature_data_7 *quad_data;

    /* Topologic relationships */
    int32           num_nodes;
    CTDB_NODE_7   **nodes;
    int32           node_size;
    int8           *node_storage;
    int32           num_edges;
    CTDB_EDGE_7   **edges;
    int32           edge_size;
    int8           *edge_storage;

    /* Linear models */
    int32                 num_linear_models;
    CTDB_LINEAR_MODEL_7  *linear_models;

    /* Volume models */
    int32                 num_volume_models;
    CTDB_VOLUME_MODEL_7 **volume_models;
    CTDB_VOLUME_MODEL_7 **mes_volume_models;

    /* Aggregate models */
    int32                      num_aggregate_models;
    CTDB_AGGREGATE_MODEL_7   **aggregate_models;

    /* MES structures (buildings, caves, tunnels, bridges) */
    int32                num_mes_volumes;
    CTDB_MES_VOLUME_7   *mes_volumes;
    int32                num_mes_templates;
    CTDB_MES_TEMPLATE_7 *mes_templates;

    int32           n_patch_bytes;     /* Total amt of patch data for this 
					* cell, in bytes  (patch data = TEs +
					* features + virtual grids)
					*/
    ADDRESS         patch_data;	       /* Pointer to chunk o' memory with all
					* the patch data
					*/
    float64         meters_to_patch;   /* = patch_inv_incr * inv_incr  */
    uint32          is_gridded;
    uint32          is_tinned;
    uint32          is_short;
    int32           max_patch_number;
    int32           min_patch_number;
} CTDB_7;

typedef struct ctdb_file_header_7
{
    int32    format;
    unsigned is_curved : 1;
    unsigned is_sw_ne  : 1;
    unsigned is_gridded: 1;
    unsigned is_tinned : 1;
    unsigned is_short  : 1;
    unsigned pad1      : 27;
    float64  fixed_point_basis;
    int32    reserved1[16];
    int32    post_increment;
    int32    posts_per_patch_side;
    int32    pages_wide;
    int32    pages_high;
    int32    min_x;
    int32    max_x;
    int32    min_y;
    int32    max_y;
    int32    num_features;
    int32    patch_cache_size_map[CTDB_PATCH_CACHES_7];
    int32    pad2;
    float64  origin_northing;
    float64  origin_easting;
    int32    origin_zone_number;
    int32    origin_zone_letter;
    int32    datum;
    char     name[CTDB_NAME_LENGTH_7];
    int16    version;
    char     date[CTDB_DATE_LENGTH_7];
    int32    num_nodes;
    int32    node_size;
    int32    num_edges;
    int32    edge_size;
    int32    num_quad_data;
    int32    n_patch_bytes;
    int32    num_linear_models;
    int32    num_volume_models;
    int32    aux_hdr_bytes;	/* Size of auxiliary header */
    int32    num_soil_tables;
    int32    soil_table_size;
    int32    num_water_chars;
    int32    num_tidal_zones;
    int32    num_aggregate_models;
    int32    num_mes_volumes;
    int32    num_mes_templates;
} CTDB_FILE_HEADER_7;

/* Additional data we didn't want to try to cram into the file header */
typedef struct ctdb_aux_file_header_7
{
    unsigned gcs_mode : 2;
    unsigned pad1     : 30;
    int32 cell_id;
    int32 origin_x;
    int32 origin_y;
    float64 west_long;		/* Left-most longiitude in database */
    float64 east_long;		/* Right-most longitude in database */
    float64 south_lat;		/* Bottom-most latitude in database */
    float64 north_lat;		/* Top-most latitude in database */
} CTDB_AUX_FILE_HEADER_7;


extern void *
ctdb_convert_7_to_cmp(void *fmt7_db);

extern void *
ctdb_convert_cmp_to_7(void *fmt7_db);

extern void *
ctdb_convert_7_to_6(void *fmt7_db);

extern void
ctdb_free_fmt7_db(void *fmt7_db);

/* Mallocs the needed memory, populates it, and returns it. */
extern void *
ctdb_read_7(char *fname);

/* Reads just the MES models (file must be positioned at start of MES model
 * data) into the provided CTDB
 */
extern CTDB_STATUS 
ctdb_read_mes_7(CTDB_7 *ctdb,
		FILE   *file);

extern void
ctdb_write_7(void *fmt7_db,
	     FILE *file);

/* Writes just the MES models (file must be positioned at start of MES model
 * data) from the provided CTDB.  If little_endian is TRUE, the offsets
 * generated will be byte-swapped before being written if the compiler
 * is running on a big-endian machine.
 */
extern void
ctdb_write_just_mes_7(CTDB_7 *ctdb,
		      FILE   *file,
		      int32   little_endian);

#endif /* _CMP_FMT7_H_INCLUDED */




