#ifndef lint
static char rcsid [] = "$RCSfile: ct_cmprep.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_cmprep.c                                                       *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : This file contains routines used for accessing the      *
 *                   compiler representation, and for converting from the    *
 *                   compiler representation to the latest CTDB format       *
 *       Created   : February 1996                                           *
 *       Author    : flchamberlain                                           *
 *       Remarks   :                                                         *
 *                                                                           *
 ****************************************************************************/
#if !defined(SUN) && !defined(mips) && !defined(__alpha)
#include <bstring.h>
#endif
#include <stdio.h>      /* fwrite routine */
#include <math.h> 
#include <stdext.h>
#include <stdstring.h>
#include <stdcoord.h>
#include <libgcs.h>
#include <libgeometry.h>
#include "ct_cmprep.h"

/* Forward declarations */
static float64
get_posted_elevation(
		     CTDB_CMP      *ctdb,
		     float64        norm_x,
		     float64        norm_y
		     );

static CTDB_STATUS
ctdb_intern_lookup_terrain_element_elevation16(
					    ELEVATION_LOOKUP_STORAGE_CMP *els
					    );

static void
ctdb_intern_point_in_terrain_element16(ELEVATION_LOOKUP_STORAGE_CMP *els);



static int32	
ctdb_intern_point_inside_te16(
			      CTDB_COORD2D_I32             *a,
			      CTDB_COORD2D_I32             *b,
			      int32                         side,
			      ELEVATION_LOOKUP_STORAGE_CMP *els
			      );

static int32
ctdb_intern_validate_te_index16(
				uint32  te_index,
				uint32 *front,
				uint32 *back
				);

static float64
ctdb_intern_set_elevation_and_soil16(ELEVATION_LOOKUP_STORAGE_CMP *els);


static float64
ctdb_intern_compute_te_normal16(ELEVATION_LOOKUP_STORAGE_CMP *els);

static float64
ctdb_intern_compute_elevation_on_edge16(
					TERRAIN_ELEMENT_16   *base_te,
					TERRAIN_ELEMENT_16   *te,
					CTDB_COORD2D_I32     *pt,
					CTDB_POINT_STATE      edge
					);

static CTDB_STATUS
ctdb_intern_look_at_adjacent_triangles16(ELEVATION_LOOKUP_STORAGE_CMP *els);

static CTDB_STATUS
ctdb_intern_look_at_all_triangles16(ELEVATION_LOOKUP_STORAGE_CMP *els);

CTDB_STATUS 
ctdb_intern_get_te_edge_index16(
				TERRAIN_ELEMENT_16  *te,
				uint32               index,
				int32                side,
				int32 	         edge[]
				);

static void 
initialize_gcs_search_spaces(
			     CTDB_CMP                 *ctdb,
			     CTDB_SEARCH_SPACE_PTR_CMP space
			     );

static void 
copy_search_db(
	       CTDB_SEARCH_DB_CMP *src,
	       CTDB_SEARCH_DB_CMP *sink
	       );

static int32 
advance_gcs_search_space(
			 CTDB_SEARCH_SPACE_PTR_CMP space,
			 CTDB_SEARCH_DB_CMP       *current
			 );

static void 
ctdb_add_quad(
	      CTDB_SEARCH_DB_CMP *search_db,
	      int32               offset,
	      int32               lo_x,
	      int32               lo_y,
	      int32               hi_x,
	      int32               hi_y,
	      int32              *alloced
	      );

static void 
ctdb_intern_find_search_quads(
			      CTDB_SEARCH_DB_CMP *search_db,
			      int32               min_x,
			      int32               min_y, 
			      int32               max_x,
			      int32               max_y
			      );

static CTDB_FEATURE_DATA_CMP *
current_patch(
	      CTDB_SEARCH_SPACE_PTR_CMP space,
	      CTDB_SEARCH_DB_CMP       *current,
	      int32                    *number,
	      int32                    *zero_offset,
	      int32                    *patch_x,
	      int32                    *patch_y
	      );





/**********************************************************************
 * INTERNAL REPRESENTATION ACCESS ROUTINES                            *
 **********************************************************************/

/* Return TRUE iff specified patch is a transition patch */
int32
ctdb_intern_is_transition_patch(ctdb, patch_number)
    CTDB_CMP *ctdb;
    int32     patch_number;
{
    int32 patch_x, patch_y;

    /* A patch is a transition patch iff it is TINNED and at least one of
     * its neighbors is not
     */

    if (!ctdb_intern_number_tes(ctdb, patch_number))
      return FALSE;
    
    patch_x = patch_number % ctdb->patches_wide;
    patch_y = patch_number / ctdb->patches_wide;
    
    if (patch_x < (ctdb->patches_wide - 1))
    {
	if (!ctdb_intern_number_tes(ctdb, patch_number+1))
	  return TRUE;
    }
    
    if (patch_x > 0)
    {
	if (!ctdb_intern_number_tes(ctdb, patch_number-1))
	  return TRUE;
    }
    
    if (patch_y < (ctdb->patches_high - 1))
    {
	if (!ctdb_intern_number_tes(ctdb, patch_number + ctdb->patches_wide))
	  return TRUE;
    }
    
    if (patch_y > 0)
    {
	if (!ctdb_intern_number_tes(ctdb, patch_number - ctdb->patches_wide))
	  return TRUE;
    }
    
    return FALSE;    
}
 

/* Return the soil map for the specified patch */
int32 *
ctdb_intern_get_soil_map(ctdb, patch_number)
    CTDB_CMP *ctdb;
    int32     patch_number;
{
    CTDB_PATCH_GROUP_CMP *pg;
    
    pg = ctdb->patch_groups + PG_NUMBER(patch_number);
    return ctdb->soil_tables[pg->tables[PG_MEMBER(patch_number)].soil_table];
}

void ctdb_intern_unpack_header(ctdb, hdr)
    CTDB_CMP             *ctdb;
    CTDB_FILE_HEADER_CMP *hdr;
{
    /* Just to be safe... */
    bzero(ctdb, sizeof(CTDB_CMP));

    ctdb->input_format = hdr->format;

    ctdb->is_curved  = hdr->is_curved;
    ctdb->is_sw_ne  = hdr->is_sw_ne;

    if(hdr->format < 4)
    {
	ctdb->is_gridded = TRUE;
	ctdb->is_tinned = FALSE;
    }
    else
    {
	ctdb->is_gridded = hdr->is_gridded;
	ctdb->is_tinned  = hdr->is_tinned;
    }

    if(hdr->format < 7)
      {
	ctdb->is_short = FALSE;
      }
    else
      {
	ctdb->is_short = hdr->is_short;
      }

    ctdb->fixed_point_basis = hdr->fixed_point_basis;
    ctdb->inv_fixed_point_basis = 1.0/hdr->fixed_point_basis;

    ctdb->num_soil_tables = hdr->num_soil_tables;
    ctdb->soil_table_size = hdr->soil_table_size;
    ctdb->num_water_chars = hdr->num_water_chars;
    ctdb->num_tidal_zones = hdr->num_tidal_zones;

    ctdb->incr = hdr->post_increment;
    ctdb->incrD = ctdb->incr;
    ctdb->inv_incr = 1.0/ctdb->incrD;

    ctdb->patch_incr = hdr->posts_per_patch_side;
    ctdb->patch_incrD = ctdb->patch_incr;
    ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide = hdr->pages_wide;
    ctdb->pages_high = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE / ctdb->patch_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE / ctdb->patch_incr;

    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);
    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);
    
    ctdb->num_features = hdr->num_features;
 
    ctdb->origin_northing = hdr->origin_northing;
    ctdb->origin_easting = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum = hdr->datum;
    bcopy(hdr->name, ctdb->name, sizeof(hdr->name));
    ctdb->version = hdr->version;
    bcopy(hdr->date, ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date, "\n");

    ctdb->num_nodes = hdr->num_nodes;
    ctdb->node_size = hdr->node_size;
    ctdb->num_edges = hdr->num_edges;
    ctdb->edge_size = hdr->edge_size;
    ctdb->num_quad_data = hdr->num_quad_data;
    ctdb->num_linear_models = hdr->num_linear_models;
    ctdb->num_volume_models = hdr->num_volume_models;

    ctdb->num_aggregate_models = hdr->num_aggregate_models;

    ctdb->num_mes_volumes = hdr->num_mes_volumes;
    ctdb->num_mes_templates = hdr->num_mes_templates;

    ctdb->meters_to_patch = ctdb->patch_inv_incr * ctdb->inv_incr;
    ctdb->max_patch_number = ctdb->patches_wide * ctdb->patches_high - 1;
    ctdb->min_patch_number = 0;
}

void ctdb_intern_unpack_aux_header(ctdb, aux_hdr)
    CTDB_CMP                 *ctdb;
    CTDB_AUX_FILE_HEADER_CMP *aux_hdr;
{
    ctdb->gcs_mode = aux_hdr->gcs_mode;
    ctdb->origin_x = aux_hdr->origin_x;
    ctdb->origin_y = aux_hdr->origin_y;
    ctdb->west_long = aux_hdr->west_long;
    ctdb->east_long = aux_hdr->east_long;
    ctdb->south_lat = aux_hdr->south_lat;
    ctdb->north_lat = aux_hdr->north_lat;
    return;
}

/* The ctdb_intern_pack_header routine will create a file header (hdr) from the
 * compiler ctdb data structure. 
 */
void 
ctdb_intern_pack_header(ctdb, hdr)
    CTDB_CMP             *ctdb;
    CTDB_FILE_HEADER_CMP *hdr;
{
    bzero(hdr, sizeof(hdr));
    hdr->format = ctdb->input_format;
    hdr->is_curved = ctdb->is_curved;
    hdr->is_sw_ne = ctdb->is_sw_ne;
    hdr->is_gridded = ctdb->is_gridded;
    hdr->is_tinned  = ctdb->is_tinned;
    hdr->fixed_point_basis = ctdb->fixed_point_basis;

    hdr->num_soil_tables = ctdb->num_soil_tables;
    hdr->soil_table_size = ctdb->soil_table_size;
    hdr->num_water_chars = ctdb->num_water_chars;
    hdr->num_tidal_zones = ctdb->num_tidal_zones;

    hdr->post_increment = ctdb->incr;
    hdr->posts_per_patch_side = ctdb->patch_incr;
    hdr->pages_wide = ctdb->pages_wide;
    hdr->pages_high = ctdb->pages_high;
    hdr->min_x = ctdb->min_x;
    hdr->max_x = ctdb->max_x;
    hdr->min_y = ctdb->min_y;
    hdr->max_y = ctdb->max_y;
    hdr->num_features = ctdb->num_features;

    hdr->origin_northing = ctdb->origin_northing;
    hdr->origin_easting = ctdb->origin_easting;
    hdr->origin_zone_number = ctdb->origin_zone_number;
    hdr->origin_zone_letter = ctdb->origin_zone_letter;
    hdr->datum = ctdb->datum;
    bcopy(ctdb->name, hdr->name, sizeof(ctdb->name));
    hdr->version = ctdb->version;
    bcopy(ctdb->date, hdr->date, sizeof(ctdb->date));
    hdr->num_nodes = ctdb->num_nodes;
    hdr->node_size = ctdb->node_size;
    hdr->num_edges = ctdb->num_edges;
    hdr->edge_size = ctdb->edge_size;
    hdr->num_quad_data = ctdb->num_quad_data;
    hdr->num_linear_models = ctdb->num_linear_models;
    hdr->num_volume_models = ctdb->num_volume_models;

    hdr->num_aggregate_models = ctdb->num_aggregate_models;

    hdr->num_mes_volumes = ctdb->num_mes_volumes;
    hdr->num_mes_templates = ctdb->num_mes_templates;
    
    hdr->aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER_CMP);
}

/* The ctdb_intern_pack_aux_header routine will create an aux file header 
 * (aux_hdr) from the compiler ctdb data structure.
 */
void 
ctdb_intern_pack_aux_header(ctdb, aux_hdr)
    CTDB_CMP                 *ctdb;
    CTDB_AUX_FILE_HEADER_CMP *aux_hdr;
{
    aux_hdr->gcs_mode = ctdb->gcs_mode;
    aux_hdr->cell_id = ctdb->cell_id;
    aux_hdr->origin_x = ctdb->origin_x;
    aux_hdr->origin_y = ctdb->origin_y;
    aux_hdr->west_long = ctdb->west_long;
    aux_hdr->east_long = ctdb->east_long;
    aux_hdr->south_lat = ctdb->south_lat;
    aux_hdr->north_lat = ctdb->north_lat;

    return;
}

int32 
ctdb_intern_point_on_database(ctdb, x, y)
    CTDB_CMP *ctdb;
    float64   x;
    float64   y;
{
    return ( ctdb &&
	    (x >= ctdb->min_x) &&
	    (x <= ctdb->max_x) &&
	    (y >= ctdb->min_y) &&
	    (y <= ctdb->max_y));
}

/* get_post_index returns the index into the post elevations array given 
 * the location of post in norm_x, norm_y coordinates. 
 */
int32 
ctdb_intern_get_post_index(ctdb, norm_x, norm_y)
    CTDB_CMP      *ctdb;
    int32          norm_x;
    int32          norm_y;
{
    int32 page, index;
    
    page = (norm_x>>SIDE_SHIFT_CMP) +  
      ctdb->pages_wide * (norm_y>>SIDE_SHIFT_CMP);
    index = (norm_x&SIDE_MASK_CMP) + POSTS_PER_SIDE_CMP * 
      (norm_y&SIDE_MASK_CMP);
    return (page*POSTS_PER_PAGE_CMP + index);
}

/* The ctdb_intern_lookup_elev_post routine will find and return the elevation
 * post as specified by the given post coordinates. 
 */
int32 
ctdb_intern_lookup_elev_post(ctdb, norm_x, norm_y)
    CTDB_CMP      *ctdb;
    int32          norm_x;
    int32          norm_y;
{
    int32 index;
    int32 page;

    if (!((norm_x >= ctdb->min_x_post) &&
	  (norm_y >= ctdb->min_y_post) &&
	  (norm_x < ctdb->max_x_post) &&
	  (norm_y < ctdb->max_y_post)))
      return 0;

    page = (norm_x>>SIDE_SHIFT_CMP) + 
      ctdb->pages_wide * (norm_y>>SIDE_SHIFT_CMP);
    index = (norm_x&SIDE_MASK_CMP) + 
      POSTS_PER_SIDE_CMP * (norm_y&SIDE_MASK_CMP);

    return ctdb->elevations[page*POSTS_PER_PAGE_CMP + index];
}

/* The ctdb_intern_lookup_soil_post routine will find and return the soil
 * post as specified by the given post coordinates. 
 */
int32 
ctdb_intern_lookup_soil_post(ctdb, norm_x, norm_y)
    CTDB_CMP      *ctdb;
    int32          norm_x;
    int32          norm_y;
{
    int32 index;
    int32 page;
    int16 *soils = (int16 *)ctdb->soils;

    if (!((norm_x >= ctdb->min_x_post) &&
	  (norm_y >= ctdb->min_y_post) &&
	  (norm_x < ctdb->max_x_post) &&
	  (norm_y < ctdb->max_y_post)))
      return 0;

    if ( ctdb->is_short )
      {
	index = ctdb_intern_get_patch_number(ctdb, norm_x, norm_y);
        return BUILD_POST_SW_SOIL_CMP(soils[index]) |
	  BUILD_POST_NE_SOIL_CMP(soils[index]);
      }
    else
      {
	page = (norm_x>>SIDE_SHIFT_CMP) + 
	  ctdb->pages_wide * (norm_y>>SIDE_SHIFT_CMP);
	index = (norm_x&SIDE_MASK_CMP) + 
	  POSTS_PER_SIDE_CMP * (norm_y&SIDE_MASK_CMP);
	return ctdb->soils[page*POSTS_PER_PAGE_CMP + index];
      }
}


/* Grab the post (from elevations, not from ctdb->elevations) at 
 * normx, normy.
 */
int32
get_post(ctdb, elevations, norm_x, norm_y)
    CTDB_CMP      *ctdb;
    int32         *elevations;
    int32          norm_x, norm_y;
{
    int32 page, index;

    if (!((norm_x >= ctdb->min_x_post) &&
	  (norm_y >= ctdb->min_y_post) &&
	  (norm_x < ctdb->max_x_post) &&
	  (norm_y < ctdb->max_y_post)))
      return 0;

    page = (norm_x>>SIDE_SHIFT_CMP) + ctdb->pages_wide * 
      (norm_y>>SIDE_SHIFT_CMP);
    index = (norm_x&SIDE_MASK_CMP) + POSTS_PER_SIDE_CMP * 
      (norm_y&SIDE_MASK_CMP);
	
    return elevations[page*POSTS_PER_PAGE_CMP + index];
}

/* This function returns the elevation of the post defined by post 
 * coordinates (norm_x, norm_y). 
 */
float64 
ctdb_intern_lookup_post_elevation(ctdb, norm_x, norm_y)
    CTDB_CMP       *ctdb;
    int32           norm_x, norm_y;
{
    int32 post;
    float64 elev;

    if (ctdb->elevations)
    {
	post = ctdb_intern_lookup_elev_post(ctdb, norm_x, norm_y);
	return GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
    }
    else
    {
	ctdb_intern_lookup_te_info((float64)norm_x, (float64)norm_y, ctdb, 
				   &elev, NULL, NULL, NULL);
	return elev;
    }
}

#define VERY_NEGATIVE_NUMBER (-1.0E+100)
int32 
ctdb_intern_lookup_microterrain_elevation(CTDB_CMP *ctdb, int32 px, int32 py,
					  float64 norm_x, float64 norm_y,
					  float64 *elev_ret,
					  float64 reference_elevation,
					  int32 *is_ml_micro)
{
    int32 patchx, patchy, patch_number;
    int32 post_mask;
    CTDB_POINT_STATE hit;
    CTDB_FEATURE_DATA_CMP *feature, *last;
    int32 xu, yu;
    float64 xuD, yuD;
    float64 a,b,c,magic;
    uint8 type;
    float64 elev, default_elev, base_elev, ml_elev, backup_elev;
    int32 soil, default_soil, base_soil, ml_soil, backup_soil;
    float64 diff, ml_diff;
    int32 next;
    CTDB_PATCH_GROUP_CMP *pg;
    int32 mem;

    /* There are currently four types of microterrain.  We search
     * for all our intersections with each type.  The point may be
     * in more than one of each type of polygon, if there is overlap
     * in the database.  Return the elevation as follows:
     *
     * MICRO_TYPE_DEFAULT: Find the highest elevation among these polygons.
     *                     Return it only if there is nothing better to use.
     * MICRO_TYPE_BASE:    Find the highest elevation among these polygons.
     *                     Return it if we don't find any acceptable
     *                     polygons of the next type.
     * MICRO_TYPE_ML:      
     * MICRO_TYPE_WATER:   Find the elevation among these polygons
     *                     which is greater than or equal to the seed
     *                     elevation, and most similar to the seed elevation.
     *                     If we find such an elevation, return it.
     */
    
    if(is_ml_micro)
      *is_ml_micro = FALSE;

    /* What patch is this post on? */
    patchx = PATCH_X(norm_x*ctdb->patch_inv_incr);
    patchy = PATCH_Y(norm_y*ctdb->patch_inv_incr);
    patch_number = patchx + patchy * ctdb->patches_wide;
    
    feature = ctdb_intern_lookup_patch(ctdb, patch_number);
    
    if (!feature)
    {
	*elev_ret = CTDB_ILLEGAL_ELEVATION;
	return 0;
    }
    
    /* The post_mask gives us a bounding box check in one easy operation */
    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);
    
    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can do some comparison
     * operations in integer.
     */
    xuD = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
    yuD = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);
    xu = xuD;
    yu = yuD;
    
    default_elev = base_elev = ml_elev = backup_elev = VERY_NEGATIVE_NUMBER;
    default_soil = base_soil = ml_soil = backup_soil = -1;
    ml_diff = CTDB_ACCEPT_ML;
    
    last = feature + feature->hdr.size;
    for(feature++; feature<last; feature+=next)
    {
	int32 verts=feature->info.vertices;
	int32 fclass=feature->info.feature_class;
	/* Microterrain comes first in the patch, so as soon as we see
	 * some non-microterrain we're effectively done.
	 */
	if ( fclass != FC_MICRO_CMP )
	  break;
	
	/* Does this microterrain point pass the post_mask test, and
	 * does it have enough vertices (2 or 3) to describe a triangle?
	 */
	next   = NEXT_MICRO_CMP(verts);
	verts &= MICRO_VERTS_CMP;
	
	if (!(post_mask & feature->integer) || (verts < 2))
	  continue;
	
	/* Are we in this one? */
	if ((hit = ctdb_point_in_micro_feature((CTDB_FEATURE_DATA *)feature, 
					       xu, yu,
					       next, &elev)) == OUTSIDE)
	  continue;

	/* Optimization note:  if-else-if chains are still faster than
	 * switch-case, even when the values look like they could drive
	 * a lookup table.
	 */
	
	if ((hit == TRIANGLE0) || (hit == TRIANGLE1))
	{
	    /* We're inside this one.
	     * Compute a normal <a b c> and a magic number ax0 + by0 + cz0
	     */
	    if (hit == TRIANGLE0)
	    {
		ctdb_compute_micro_feature_normal((CTDB_FEATURE_DATA *)feature,
						  0, next, &a, &b, &c, &magic);
	    }
	    else		/* hit == TRIANGLE1 */
	    {
		ctdb_compute_micro_feature_normal((CTDB_FEATURE_DATA *)feature,
						  1, next, &a, &b, &c, &magic);
	    }

	    /* First make sure this isn't a vertical polygon. */
	    if (c == 0.0)
	    {
		elev = feature[2].z;
	    }
	    else
	    {
		/* Compute our elevation by the equation:
		 * ax + by + cz = ax0 + by0 + cz0
		 */
		elev = (magic - a * xuD - b * yuD) / c;
	    }
	}
	
	/* Potentially keep this elevation/soil, depending upon its
	 * height.
	 */
	type = feature->info.vertices & MICRO_TYPE_CMP;
	if (type == MICRO_TYPE_BASE_CMP)
	{
	    if (elev > base_elev)
	    {
		base_elev = elev;
		base_soil = soil;
	    }
	}
	else if (type == MICRO_TYPE_DEFAULT_CMP)
	{
	    if (elev > default_elev)
	    {
		default_elev = elev;
		default_soil = soil;
	    }
	}
	else if ((type == MICRO_TYPE_ML_CMP) || (type == MICRO_TYPE_WATER_CMP))
	{
	    if(is_ml_micro)
	      *is_ml_micro = TRUE;

	    if (elev <= reference_elevation)
	    {
		/* How far is this elevation from the reference? */
		diff = reference_elevation - elev;
		
		/* Select the closest */
		if (diff < ml_diff)
		{
		    ml_elev = elev;
		    ml_soil = soil;
		    ml_diff = diff;
		}
	    }
	    else
	    {
		/* Hang on to this elevation, in case it is the only
		 * polygon we find.
		 */
		backup_elev = elev;
		backup_soil = soil;
	    }
	}
    }
    
    /* Select one of the elevations, as explained in the comment at the
     * top of this function.
     */
    if (ml_soil != -1)
    {
	*elev_ret = ml_elev;
	return 1;
    }
    if (base_soil != -1)
    {
	*elev_ret = base_elev;
	return 1;
    }
    if (default_soil != -1)
    {
	*elev_ret = default_elev;
	return 1;
    }

#if 0
    /* This segment of code says it's OK if we did not find any microterrain
     * underneath the reference elevation (in that case, just use the only
     * elevation we found).  This, however, makes it not possible to place
     * ML micro polys on grids or TINs without covering the whole patch
     * with microterrain.
     * 
     * We do not want this restriction, so this code is being disabled.
     */
    if (backup_soil != -1)
    {
	*elev_ret = backup_elev;
	return 1;
    }
#endif

    /* We failed to find a supporting polygon.  It is expected
     * that the calling routine will find an appropriate elevation
     * from the underlying grid.
     */
    *elev_ret = CTDB_ILLEGAL_ELEVATION;
    return 0;
}

/* This function will return the elevation at the specified (x,y) location
 * (in database coordinates).  If mode is set to FAVOR_GRID, it will return
 * the elevation specified by the grid posts if possible, and if not, revert
 * to the TEs.  If mode is set to FAVOR_TES, it will do the opposite.
 */
float64
ctdb_intern_get_ground_elevation(ctdb, x, y, mode)
    CTDB_CMP      *ctdb;
    float64        x;
    float64        y;
    int32          mode;
{
    float64 ret = CTDB_ILLEGAL_ELEVATION;
    float64 norm_x = x * ctdb->inv_incr;
    float64 norm_y = y * ctdb->inv_incr;
    int32   px = POST_X(x * ctdb->inv_incr);
    int32   py = POST_Y(y * ctdb->inv_incr);

    if (!ctdb)
      return CTDB_ILLEGAL_ELEVATION;

    if (mode == FAVOR_TES)
    {
        ctdb_intern_lookup_te_info(norm_x, norm_y, ctdb, &ret, 
				   NULL, NULL, NULL);
	if (ret != CTDB_ILLEGAL_ELEVATION)
	{
	    return ret;
	}
    }

    /* Look at microterrain first.
     */
    ctdb_intern_lookup_microterrain_elevation(ctdb, px, py, norm_x, norm_y,
					      &ret, CTDB_ACCEPT_ML, NULL);

    /* If we're not on microterrain, look at the grid.
     */
    if(ret == CTDB_ILLEGAL_ELEVATION)
      ret = get_posted_elevation(ctdb, norm_x, norm_y);

    if ((ret == CTDB_ILLEGAL_ELEVATION) && (mode == FAVOR_GRID))
    {
        ctdb_intern_lookup_te_info(norm_x, norm_y, ctdb, &ret, 
				   NULL, NULL, NULL);
    }

    return ret;
}

/* Returns the elevation at a point based *only* on the gridposts.  Note,
 * however, that it does interpolate--i.e., it does not simply return
 * the elevation at the closest post
 */
static float64
get_posted_elevation(ctdb, norm_x, norm_y)
    CTDB_CMP      *ctdb;
    float64        norm_x, norm_y;
{
    int32 px, py;
    float64 pxD, pyD, pz, pzS, pzE, pzSE, dx, dy;
    int32 post;

    if (!ctdb->elevations)
    {
       return CTDB_ILLEGAL_ELEVATION;
    }

    /* If asked about a point off the database, return the elevation at the
     * closest point on the database.
     */
    norm_x = NS_LIMITS(ctdb->min_x_post, norm_x, (ctdb->max_x_post - 1));
    norm_y = NS_LIMITS((ctdb->min_y_post + 1), norm_y, (ctdb->max_y_post-1));

    px = POST_X(norm_x);
    pyD = POST_Y(norm_y);

    pxD = px;
    py = pyD;

    post = ctdb_intern_lookup_elev_post(ctdb, px, py);
    pz = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);

    dx = (norm_x - pxD);
    dy = (pyD - norm_y);

    /* Is the point on a vertical edge? */
    if (!dx)
    {
	/* Is the point on the post? */
	if (!dy)
	  return pz;

        /* Find the elevation of the next post South */
	post = ctdb_intern_lookup_elev_post(ctdb, px, py-1);
	pzS = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
  	/* Interpolate linearly along the y axis */
	return pz + (pzS - pz) * dy;
    }

    /* Is the point on a horizontal edge? */
    if (!dy)
    {
	/* Find the elevation of the next post East */
	post = ctdb_intern_lookup_elev_post(ctdb, px+1, py);
	pzE = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);

	/* Interpolate linearly along the x axis */
	return pz + (pzE - pz) * dx;
    }

    /* Find the elevation of the next post Southeast */
    post = ctdb_intern_lookup_elev_post(ctdb, px+1, py-1);
    pzSE = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
    
    /* Is the point in the leftmost poly? */
    if (dx <= dy)
    {
	/* Find the elevation of the next post South */
	post = ctdb_intern_lookup_elev_post(ctdb, px, py-1);
	pzS = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
	
	return pz + (pzSE - pzS) * dx + (pzS - pz) * dy;
    }
    else
    {
	/* Find the elevation of the next post East */
	post = ctdb_intern_lookup_elev_post(ctdb, px+1, py);
	pzE = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);

	    return pz + (pzE - pz) * dx + (pzSE - pzE) * dy;
    }
}


/* This function returns the soil type at point norm_x, norm_y 
 * $$$ NOTE THAT THIS RETURNS AN INDEX INTO THE SOIL TABLE FOR THE
 * PATCH IN WHICH THIS POST RESIDES, NOT THE ACTUAL SOIL TYPE.  WHEN
 * CHANGING SOIL MAPS, THE CALLER WILL HAVE TO DEAL WITH THIS...
 */
uint8
ctdb_intern_lookup_soil(ctdb, norm_x, norm_y)
    CTDB_CMP      *ctdb;
    float64   norm_x;
    float64   norm_y;
{
    float64 pxD, pyD, dx, dy;
    int32 px, py, post;
    float64 elev;
    int32 sNE, sSW, sHere;
    int32 soil_index;

    px = POST_X(norm_x);
    pyD = POST_Y(norm_y);

    pxD = px;
    py = pyD;


    /* Bounds check */
    if (!(((px >= ctdb->min_x_post) &&
	   (px+1 < ctdb->max_x_post) ||
	   (norm_x == ctdb->max_x_post)) &&
	  ((py-1 >= ctdb->min_y_post) &&
	   (py < ctdb->max_y_post) ||
	   (norm_y == ctdb->min_y_post))))
      return 0.0;

    if (ctdb->elevations)
	post = ctdb_intern_lookup_soil_post(ctdb, px, py);
    else
    {
	ctdb_intern_lookup_te_info(POST_X(norm_x), POST_Y(norm_y), 
				   ctdb, &elev, NULL, NULL, &sHere);
	return sHere;
    }

    dx = (norm_x - pxD);
    dy = (pyD - norm_y);


    /* Is the point on a vertical edge? */
    if (!dx)
      if (!(post & POST_SW_NE))
      {
	  return GET_POST_SW_SOIL(post);
      }
      else
      {
	  return GET_POST_NW_SOIL(post);
      }
    
    /* Is the point on a horizontal edge? */
    if (!dy)
      if (!(post & POST_SW_NE))
	return GET_POST_NE_SOIL(post);
      else
	return GET_POST_NW_SOIL(post);

    if (!(post & POST_SW_NE))
	/* Is the point in the leftmost poly? */
      if (dx <= dy)
	return GET_POST_SW_SOIL(post);
      else
	return GET_POST_NE_SOIL(post);
    else
	/* Is the point in the leftmost poly? */
      if (dx <= (1.0-dy))
	return GET_POST_NW_SOIL(post);
      else
	return GET_POST_SE_SOIL(post);
}

/* The ctdb_intern_get_patch_number will return the patch number containing
   the point specified by (norm_x, norm_y). */

int32
ctdb_intern_get_patch_number (
			      CTDB_CMP *ctdb,
			      int32     norm_x,
			      int32     norm_y
			      )
{
  int32 patch_number, patch_x, patch_y;

  patch_x = PATCH_X( norm_x * ctdb->patch_inv_incr );
  patch_y = PATCH_Y( norm_y * ctdb->patch_inv_incr );
  patch_number = patch_y * ctdb->patches_wide + patch_x;
  return patch_number;
}

/* ctdb_intern_test_patch_soil determines whether all the soil posts of a
   patch point to the same soil type. */

int32
ctdb_intern_test_patch_soil(
			    CTDB_CMP *ctdb,
			    int32     patch_x,
			    int32     patch_y
			    )
{
  int32 min_x, min_y, norm_x, norm_y, max_x, max_y;
  int32 soil0, soil1, soil;

  min_x = patch_x * ctdb->patch_incr;
  min_y = patch_y * ctdb->patch_incr + 1;
  max_x = min_x + ctdb->patch_incr;
  max_y = min_y + ctdb->patch_incr;
  soil  = GET_POST_SW_SOIL(ctdb_intern_lookup_soil_post(ctdb, min_x, min_y));
  soil0 = soil1 = soil;

  for( norm_x=min_x; norm_x<max_x && soil==soil0 && soil==soil1; norm_x++ )
    for( norm_y=min_y; norm_y<max_y && soil==soil0 && soil==soil1; norm_y++ )
      {
	soil0 = GET_POST_SW_SOIL(ctdb_intern_lookup_soil_post(ctdb, norm_x, norm_y));
	soil1 = GET_POST_NE_SOIL(ctdb_intern_lookup_soil_post(ctdb, norm_x, norm_y));
      }
  return (soil == soil0) && (soil == soil1);
}

/* ctdb_intern_lookup_patch returns a pointer to the feature list
 * belonging to the patch specified by the given patch_number. 
 */
CTDB_FEATURE_DATA_CMP *
ctdb_intern_lookup_patch (ctdb, patch_number)
    CTDB_CMP      *ctdb;
    int32          patch_number;
{
    CTDB_PATCH_GROUP_CMP *group;
    int32                 member;
    uint32                offset;
    
    group = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);
    offset = group->feature_start;
    if (member == 0)
      if (group->patch0_occupied)
	;
      else
	return NULL;
    if (member == 1)
      if (group->patch1_occupied)
	offset += group->patch1_offset;
      else
	return NULL;
    else if (member == 2)
      if (group->patch2_occupied)
	offset += group->patch2_offset;
      else
	return NULL;
    else if (member == 3)
      if (group->patch3_occupied)
	offset += group->patch3_offset;
      else
	return NULL;
    return ((CTDB_FEATURE_DATA_CMP *)ctdb->features + offset);
}

/* This routine will return the number of terrain elements for the patch
 * with the given patch number. 
 */
int32 
ctdb_intern_number_tes(ctdb, patch_number)
    CTDB_CMP          *ctdb;
    int32              patch_number;
{
    CTDB_PATCH_GROUP_CMP *pg;
    int32 member;

    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);
    return(pg->num_te[member]);
}

/* This routine will return a pointer to the terrain element list for the patch
 * with the given patch number. 
 */
TERRAIN_ELEMENT_16 * 
ctdb_intern_lookup_terrain_element(ctdb, patch_number)
    CTDB_CMP          *ctdb;
    int32              patch_number;
{
    CTDB_PATCH_GROUP_CMP *pg;
    int32 member;
    int32 te;

    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);

    te=(int32)ctdb->terrain_elements;

    te += pg->te_offset[0];
    
    if (! pg->num_te[member] ) /* if no terrain elements */
      return NULL;
    
    if ( member )
      te += pg->te_offset[member];
    
    return((TERRAIN_ELEMENT_16 *) te) ;
}

/* Return 0 for success, 1 for failure */
static int32 
ctdb_intern_initialize_element_lookup(rdbase, els)
    CTDB_CMP                     *rdbase;
    ELEVATION_LOOKUP_STORAGE_CMP *els;
{
    CTDB_COORD2D frac;
    int32        delta, vg_dim, vgrid, vg_max,vg_x , vg_y ;
    int32        patch_number = els->num; 
    int32        patches_wide = rdbase->patches_wide;
    uint16      *vgrid_map;
    float64      meters_to_patch = rdbase->patch_inv_incr * rdbase->inv_incr;
    CTDB_PATCH_GROUP_CMP *pg;
    int32 member;

    frac.x          = els->dbpt->x * meters_to_patch;
    frac.y          = els->dbpt->y * meters_to_patch; 
    els->coord.x    = PATCH_X(frac.x);
    els->coord.y    = PATCH_Y(frac.y);
    els->num        = els->coord.x + els->coord.y*patches_wide;
    frac.x         -= els->coord.x;
    frac.y         -= els->coord.y;
    delta           = patch_number - els->num; 

    if ( patch_number >= 0 && delta )
    {
	if ( delta == -1 )     /* west */
	    frac.x = 1.0 ;
	else if ( delta == 1 ) /* east */
	    frac.x = 0 ;
	else if ( delta == patches_wide ) /* north */
	    frac.y = 0 ;
	else if ( delta == -patches_wide ) /* south */
	    frac.y = 1.0 ;
	else if ( delta == patches_wide-1 )
	{ /* northwest */
	    frac.x = 1.0 ;
	    frac.y = 0;
	}
	else if ( delta == patches_wide+1 )
	{ /* northeast */
	    frac.x = 0;
	    frac.y = 0;
	}
	else if ( delta == -patches_wide+1 ) 
	{ /* southeast */
	    frac.x = 0 ;
	    frac.y = 1.0 ;
	}
	else if ( delta == -patches_wide-1 )
	{ /* southwest */
	    frac.x = 1.0;
	    frac.y = 1.0;
	}
	else /* incorrect patch number */
	  patch_number    = els->num ;
	    
	els->num      = patch_number;
    }
    
    pg = rdbase->patch_groups+PG_NUMBER(els->num);
    member = PG_MEMBER(els->num);
    els->frac.x   = frac.x;
    els->frac.y   = frac.y;
    els->npt.x    = frac.x*CTDB_PATCH_TO_XYO_SCALE;
    els->npt.y    = frac.y*CTDB_PATCH_TO_XYO_SCALE;

    els->pg       = pg;

    els->member   = member;

    els->te16  = TRUE;
    els->soil   = CTDB_ILLEGAL_SOIL_TYPE;
    els->where  = OUTSIDE;
    els->front  = els->back = 0 ; /* initialize queue pointers */

    els->base = (ADDRESS *)
      ctdb_intern_lookup_terrain_element(rdbase, els->num);

    if (els->base == NULL)
      return 1;

    vg_dim = pg->te_vgrid_dim[member];
    if (vg_dim == 0)
      return 1;

    /* Generate the virtual grid index and find the te that it maps
     * to 
     */
    vg_max = vg_dim-1;
    vg_x = (int32)(uint8)(els->frac.x*vg_dim); 
    vg_y = (int32)(uint8)(els->frac.y*vg_dim);
    vgrid   = NS_MIN(vg_x,vg_max) + NS_MIN(vg_y,vg_max) * vg_dim ;
    if (pg->te_vgrid_offset == NULL)
      return 1;
    vgrid_map = (uint16*)pg->te_vgrid_offset[member];
    if (vgrid_map == NULL)
      return 1;
    els->mapped_te=els->te_index= vgrid_map[vgrid];
    return 0;
}


/* The ctdb_intern_lookup_te_info will return the soil and elevation 
 * information for the grid post area specified by post_x, post_y (i.e., the 
 * grid square whose upper left corner is at POST_X(post_x), POST_Y(post_y).  
 * sSW returns a best guess at the soil type in the SW triangle, sNE returns
 * a best guess at the soil type in the NE triangle, and sHere returns
 * the soil type (not a guess) at (post_x, post_y).  Any and all of these
 * pointers may be NULL if no value is desired.  The elevation returned
 * is at the specified point (*not* the grid post location, unless the
 * point happens to be on the post)
 * NOTE:  the soils returned are actually indices into the soil map for the
 * patch containing the point.  If the soil maps are changing, the caller
 * will have to handle that...
 */
void 
ctdb_intern_lookup_te_info(post_x, post_y, ctdb, z, sNE, sSW, sHere)
    float64        post_x;
    float64        post_y;
    CTDB_CMP      *ctdb;
    float64       *z;
    int32         *sNE;
    int32         *sSW;
    int32         *sHere;
{
    CTDB_COORD2D pt;
    int32 px = POST_X(post_x);
    int32 py = POST_Y(post_y);
    int32 patch_number = PATCH_X(px * ctdb->patch_inv_incr) + 
      ctdb->patches_wide * PATCH_Y(py * ctdb->patch_inv_incr);
    int32 post = 0;
    ELEVATION_LOOKUP_STORAGE_CMP els;/* Elevation Lookup Storage Data 
				      * Structure
				      */
    float64 dx, dy;
    CTDB_STATUS status=CTDB_STATUS_OK;

    /* If we don't have elevation posts, just assume NW-SE diagonalization.
     */
    if(ctdb->elevations)
      post = ctdb_intern_lookup_elev_post(ctdb, px, py);
    else
      post = 0;

    if (ctdb_intern_number_tes(ctdb, patch_number) == 0)
    {
        *z = CTDB_ILLEGAL_ELEVATION;
	if (sNE)
	  *sNE = CTDB_ILLEGAL_SOIL_TYPE;
	if (sSW)
	  *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	if (sHere)
	  *sHere = CTDB_ILLEGAL_SOIL_TYPE;
	return;
    }

    /* determine elevation of point location */
    pt.x = post_x * ctdb->incrD;
    pt.y = post_y * ctdb->incrD;

    /* Better safe than sorry */
    if ( !ctdb_intern_point_on_database(ctdb, pt.x, pt.y) )
    {
	*z = CTDB_ILLEGAL_ELEVATION;
	if (sNE)
	  *sNE = CTDB_ILLEGAL_SOIL_TYPE;
	if (sSW)
	  *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	if (sHere)
	  *sHere = CTDB_ILLEGAL_SOIL_TYPE;
	return;
    }
    
    /* if edge of database is on a patch boundary then work with
       previous patch */
    if (((int32)pt.x == ctdb->max_x) && 
	(((int32)pt.x % (ctdb->incr * ctdb->patch_incr)) == 0))
      els.num = patch_number -1;
    else
      els.num = 0;

    els.ctdb  = ctdb;
    els.dbpt   = &pt;

    if (ctdb_intern_initialize_element_lookup(ctdb, &els))
    {
	*z = CTDB_ILLEGAL_ELEVATION;
	if (sNE)
	  *sNE = CTDB_ILLEGAL_SOIL_TYPE;
	if (sSW)
	  *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	if (sHere)
	  *sHere = CTDB_ILLEGAL_SOIL_TYPE;
	return;
    }

    status = ctdb_intern_lookup_terrain_element_elevation16(&els);

    *z = els.elevation;
    if ( status &&
	status != CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT)
    {
	*z = CTDB_ILLEGAL_ELEVATION;
	if (sNE)
	  *sNE = CTDB_ILLEGAL_SOIL_TYPE;
	if (sSW)
	  *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	if (sHere)
	  *sHere = CTDB_ILLEGAL_SOIL_TYPE;
	return;
    }
    if (sHere)
      *sHere = els.soil;

    /* pick two points: one in south corner of post and one in north
     * corner depending on diagonalization; and find their soil values 
     */

    /* NE or NW soil */

    if (sNE)
    {
        dy = py - 0.25;
	if (post & POST_SW_NE)
	  dx = px + 0.25;
	else
	  dx = px + 0.75;

	pt.x = dx * ctdb->incrD;
	pt.y = dy * ctdb->incrD;

	if (!ctdb_intern_point_on_database(ctdb,pt.x,pt.y) )
	  *sNE = CTDB_ILLEGAL_SOIL_TYPE;
	else
	{
	    els.dbpt   = &pt;
	    els.num    = -1;/* Don't override computed patch number */
	    if (ctdb_intern_initialize_element_lookup(ctdb, &els))
	    {
	        *sNE = CTDB_ILLEGAL_SOIL_TYPE;
	    }
	    else
	    {
	        status = ctdb_intern_lookup_terrain_element_elevation16(&els);

		*sNE = els.soil;
		if ( status &&
		     status !=
		     CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT)
		  *sNE= CTDB_ILLEGAL_SOIL_TYPE;
	    }
	}
    }

    /* SW or SE soil */
    if (sSW)
    {
        dy = post_y - 0.75;
	if (post & POST_SW_NE)
	  dx = post_x + 0.75;
	else
	  dx = post_x + 0.25;
	
	pt.x = dx * ctdb->incrD;
	pt.y = dy * ctdb->incrD;
	
	if (!ctdb_intern_point_on_database(ctdb, pt.x, pt.y) )
	  *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	else
	{
	    els.dbpt   = &pt;
	    els.num    = -1;/* Don't override computed patch number */
	    if (ctdb_intern_initialize_element_lookup(ctdb, &els))
	    {
	        *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	    }
	    else
	    {
	        status = ctdb_intern_lookup_terrain_element_elevation16(&els);

		*sSW = els.soil;
		if ( status &&
		     status != 
		     CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT)
		  *sSW = CTDB_ILLEGAL_SOIL_TYPE;
	    }
	}
    }
    return;
}    

int32
ctdb_intern_get_cell_id(ctdb)
    CTDB_CMP *ctdb;
{
    return ctdb->cell_id;
}

/******************
 * TE lookup code *
 ******************/

#define MARKLENGTH               10000
static uint32  checked_array[MARKLENGTH];
static uint32 *checked_f;/* front point of the checked TE array */
static uint32 *checked_b;/* back pointer of the checked TE array */

static CTDB_STATUS
ctdb_intern_lookup_terrain_element_elevation16(els)
    ELEVATION_LOOKUP_STORAGE_CMP  *els;
{   
    checked_b = checked_f = &checked_array[0];/* Global list of TEs checked */

    while (1)
    {
	ctdb_intern_point_in_terrain_element16(els);
	
	if (els->where != OUTSIDE)
	{
	    ctdb_intern_set_elevation_and_soil16(els);
	    return CTDB_STATUS_OK;
	}
	
	if ( ctdb_intern_look_at_adjacent_triangles16(els) != CTDB_STATUS_OK )
        { 
	    if ( ctdb_intern_look_at_all_triangles16(els) != CTDB_STATUS_OK )
	    {
		/* No Elevation found so a TE must be missing: approximate
		 * the elevation by returning the elevation for the TE
		 * associated with the Vgrid for the given point 
		 */
		els->where = TRIANGLE0;
		els->te_index = els->mapped_te;
		els->te = (TERRAIN_ELEMENT_16 *) els->base +
		  els->te_index;
		ctdb_intern_set_elevation_and_soil16(els);
		
		return CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT;
	    }
	}     
    }
#ifdef lint
    return CTDB_STATUS_OK;
#endif
}


#define AdotB(x0,y0,x1,y1) ( (x0)*(x1) + (y0)*(y1) )
#define AxB(x0,y0,x1,y1)   ((x0)*(y1) - (y0)*(x1))

static void
ctdb_intern_point_in_terrain_element16(els)
    ELEVATION_LOOKUP_STORAGE_CMP *els;
{
    /* COORDINATE SYSTEM : TE COORDINATES */
    /* patch's TE base address */
    TERRAIN_ELEMENT_16   *base_te= (TERRAIN_ELEMENT_16 *)els->base; 
    TERRAIN_ELEMENT_16   *vert_te= base_te + els->te_index ;
    int32 	          i;
    int32 	          bound  = vert_te->info.vertices + 1;
    CTDB_COORD2D_I32      P[4];
    int32                 e1,e2,e3,e4;
    
    /* Determine if the point is inside the triangle.  To do this, we
     * draw a line from the point to each vertex (V0, V1, V2).
     * If we are inside the triangle, going from V0 to V1 to V2 will
     * either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     * (A side effect of finding each vector is that we can test
     * for a direct hit of our point on a polygon vertex, so we
     * exploit it.)
     *
     * When one of the cross products produces a 0, we test dot products
     * to determine if the point is on the line connecting the vertices.
     *
     * The data may describe 0, 1, or 2 microterrain triangles.
     *
     * NOTE <xu yu> is assumed to be in the funny units used to store
     * features, NOT meters or posts.
     */
    if (bound < 3)
      return;

    els->te = (ADDRESS)vert_te ;
    
    for (i=0;i<bound;i++)
    {
	/* P0 is the vector from our point to vertex 0 */
	/* P1 is the vector from our point to vertex 1 */
	/* P2 is the vector from our point to vertex 2 */
	/* P3 is the vector from our point to vertex 3 */
	
	P[i].x = vert_te->pt.x - els->npt.x;  
	P[i].y = vert_te->pt.y - els->npt.y ;
	if (!P[i].x  && !P[i].y)
	{
	    els->where = VERTEX0+i;
	    return;
	}
	vert_te = base_te + ((TERRAIN_ELEMENT_16 *)els->te)->vert.index[i] ;
    }
    
    if ( bound == 3 )
    {
	e1 = ctdb_intern_point_inside_te16(&P[0],&P[1],(int32)0,els);
	e2 = ctdb_intern_point_inside_te16(&P[1],&P[2],(int32)1,els);
	e3 = ctdb_intern_point_inside_te16(&P[2],&P[0],(int32)2,els);
	
	if ( e1 && e2 && e3 )
	  els->where  = TRIANGLE0;
    }
    else
    {
	e1 = ctdb_intern_point_inside_te16(&P[0],&P[2],(int32)0,els);
	e2 = ctdb_intern_point_inside_te16(&P[2],&P[1],(int32)1,els);
	e3 = ctdb_intern_point_inside_te16(&P[1],&P[3],(int32)2,els);
	e4 = ctdb_intern_point_inside_te16(&P[3],&P[0],(int32)3,els);
	
	if ( AxB(P[0].x,P[0].y,P[1].x,P[1].y) > 0 )
	{
	    if ( e1 && e2) 
	      els->where = TRIANGLE0;
	}
	else
	{
	    if ( e3 && e4 )
	      els->where = TRIANGLE1 ;
	}
    }
    
    return;
}

/* ctdb_intern_point_inside_te
 *
 * Note:  This routine only changes the data parameter _where_ if the point
 * is colinear.  Any other setting of this parameter will affect callers of 
 * this routine since they do not check for a change in the state of _where_ 
 * after each call to this routine.
 */
static int32	
ctdb_intern_point_inside_te16(a, b, side, els)
    CTDB_COORD2D_I32             *a; /* vector a from pt(x,y) to a(x,y) */
    CTDB_COORD2D_I32             *b; /* vector a from pt(x,y) to b(x,y) */
    int32                         side; /* side define by the two TE vertices 
					 * used                   in a,b 
					 */
    ELEVATION_LOOKUP_STORAGE_CMP *els;   
{
    uint32 index;		/* potential preferred candidate te */
    int32  cross=AxB(a->x,a->y,b->x,b->y);
    
    index = (uint32)(((TERRAIN_ELEMENT_16 *)(els->te))->adj[side]);
    
    if ( cross < 0 ) /* Point inside half plane */	
      return(TRUE);
    
    if ( cross > 0 ) /* Point outside half plane */
    {
	if ( ctdb_intern_validate_te_index16(index,checked_f,checked_b) )
	  els->te_queue[els->back++] = index;  /* add to queue */
	return FALSE;
    }
    
    if ( AdotB(a->x,a->y,b->x,b->y) < 0 )
    { /* point on the edge */
	/* if -1 then the vectors are opposite in direction and the point 
	 * will lye between the two end points.  If the result is +1 then 
	 * the point is outside the segment since the vectors are in the 
	 * same direction
	 */
	els->where=COLINEAR0+side ;
    }
    
    return(FALSE);
}

/* This routine will return TRUE if the index is valid and has not already
 * been inspected 
 */
static int32
ctdb_intern_validate_te_index16(te_index, front, back)
    uint32  te_index;  /* te index to validate */
    uint32 *front;     /* address of list to be validated against */
    uint32 *back ;     /* end address of the list (count not address) */
{
    if ( te_index > MAX_ALLOWED_TE_INDEX_16 )
      return FALSE;
    
    while ( (back--) > front )
    {	
	if ( *back == te_index ) /* te already processed */
	  return FALSE; 
    }
    return TRUE;
}

static float64
ctdb_intern_set_elevation_and_soil16(els)
    ELEVATION_LOOKUP_STORAGE_CMP *els;
{
    switch (els->where)
    {
      case TRIANGLE0 :
	els->elevation = ctdb_intern_compute_te_normal16(els) ;
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil0;
	break;
      case TRIANGLE1 :
	els->elevation = ctdb_intern_compute_te_normal16(els) ;
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil1;
	break;
      case VERTEX0 :
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil0;
	els->elevation = ((TERRAIN_ELEMENT_16 *)els->te)->z;
	break;
      case VERTEX1 :
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil0;
	els->elevation = 
	  ((TERRAIN_ELEMENT_16 *)els->base)[((TERRAIN_ELEMENT_16 *)els->te)->
					    vert.index[els->where-VERTEX1]].z;
	break;
      case VERTEX2 :
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil0;
	els->elevation = 
	  ((TERRAIN_ELEMENT_16 *)els->base)[((TERRAIN_ELEMENT_16 *)els->te)->
					    vert.index[els->where-VERTEX1]].z;
	break;
      case VERTEX3 :
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil1;
	els->elevation = 
	  ((TERRAIN_ELEMENT_16 *)els->base)[((TERRAIN_ELEMENT_16 *)els->te)->
					    vert.index[2]].z;
	break;
      case COLINEAR0 :
	els->elevation=
	  ctdb_intern_compute_elevation_on_edge16((TERRAIN_ELEMENT_16 *)
						  els->base,
						  (TERRAIN_ELEMENT_16 *)
						  els->te,
						  &(els->npt),
						  els->where);
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil0;
	break;
      case COLINEAR1 :
	els->elevation=
	  ctdb_intern_compute_elevation_on_edge16((TERRAIN_ELEMENT_16 *)
						  els->base,
						  (TERRAIN_ELEMENT_16 *)
						  els->te,
						  &(els->npt),
						  els->where);
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil0;
	break;
      case COLINEAR2 :
	els->elevation=
	  ctdb_intern_compute_elevation_on_edge16((TERRAIN_ELEMENT_16 *)
						  els->base,
						  (TERRAIN_ELEMENT_16 *)
						  els->te,
						  &(els->npt),
						  els->where);
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil1;
	break;
      case COLINEAR3 :
	els->elevation=
	  ctdb_intern_compute_elevation_on_edge16((TERRAIN_ELEMENT_16 *)
						  els->base,
						  (TERRAIN_ELEMENT_16 *)
						  els->te,
						  &(els->npt),
						  els->where);
	els->soil = ((TERRAIN_ELEMENT_16 *)els->te)->info.soil1;
	break;
      default:
	break;
    }
#ifdef lint
    return 0.0;
#endif
}

/* Computes a normal to the microterrain polygon.  Note that the
 * units of *a and *b are the funny units used to store features,
 * NOT meters or posts.
 */
static float64
ctdb_intern_compute_te_normal16(els)
    ELEVATION_LOOKUP_STORAGE_CMP *els;
{
    /* patch's TE base address */
    TERRAIN_ELEMENT_16 *base_te = (TERRAIN_ELEMENT_16 *)els->base; 
    /* TE which contains the point */
    TERRAIN_ELEMENT_16 *te      = (TERRAIN_ELEMENT_16 *)els->te;   
    TERRAIN_ELEMENT_16 *assoc_te;
    float64             a, b, c, magic;
    CTDB_COORD3D        v1, v2;
    
    assoc_te = base_te + te->vert.index[0];
    
    v1.x = (int32)assoc_te->pt.x - (int32)te->pt.x ;
    v1.y = (int32)assoc_te->pt.y - (int32)te->pt.y ;
    v1.z = assoc_te->z - te->z ;
    
    assoc_te = base_te + te->vert.index[els->where-TRIANGLE0+1] ;
    
    v2.x = (int32)assoc_te->pt.x - (int32)te->pt.x;
    v2.y = (int32)assoc_te->pt.y - (int32)te->pt.y;
    v2.z = assoc_te->z - te->z;
    
    a = v1.y*v2.z - v1.z*v2.y;
    b = v1.z*v2.x - v1.x*v2.z;
    c = v1.x*v2.y - v1.y*v2.x;
    
    /* The magic number = a*x + b*y + c*z, where <x y z> is a vertex
     * of the polygon.  This can be used to compute the elevation on the
     * the polygon's plane at a particular <x y>, or the intersection of a
     * line with the polygon's plane.
     */
    
    magic = ((int32)te->pt.x * a + (int32)te->pt.y * b + te->z * c);
    
    if ( c == 0.0 )
      return te->z ;
    else
      return (magic-a*(float64)els->npt.x-b*
	      (float64)els->npt.y)/c ;
}

static float64
ctdb_intern_compute_elevation_on_edge16(base_te, te, pt, edge)
    TERRAIN_ELEMENT_16   *base_te; /* patch's TE base address */
    TERRAIN_ELEMENT_16   *te;      /* TE which contains the point */
    CTDB_COORD2D_I32     *pt;      /* Point in TE coordinates */
    CTDB_POINT_STATE      edge;    /* Edge the point is on */
{
    CTDB_XYO_COORD2D start, stop;
    float64          dy;    
    float32          z, dz;
    
    if ( te->info.vertices == 2 )
    { /* triangle */
	if ( edge == COLINEAR0 ) 
	{ /* 0 -> 1 */
	    start = te->pt;
	    stop = base_te[ te->vert.index[0] ].pt;	    
	    z = te->z;
	    dz = base_te[ te->vert.index[0] ].z - te->z;
	}
	else if ( edge == COLINEAR1 )
	{ /* 1 -> 2 */
	    start = base_te[ te->vert.index[0] ].pt;
	    stop = base_te[ te->vert.index[1] ].pt;
	    z =  base_te[ te->vert.index[0] ].z; 	    
	    dz = base_te[ te->vert.index[1] ].z - 
	      base_te[ te->vert.index[0] ].z;
	}
	else 
	{ /* 2 -> 0 */
	    start = base_te[ te->vert.index[1] ].pt;
	    stop = te->pt;
	    z = base_te[ te->vert.index[1] ].z; 	    
	    dz = te->z - base_te[ te->vert.index[1] ].z;
	}		
    }
    else
    { /* quadrilateral */
	if ( edge == COLINEAR0 )
	{ /* 0 -> 2 */
	    start = te->pt;
	    stop =  base_te[ te->vert.index[1] ].pt;
	    z = te->z; 	    
	    dz = base_te[ te->vert.index[1] ].z - te->z;
	}
	else if ( edge == COLINEAR1 )
	{ /* 2 -> 1 */
	    start = base_te[ te->vert.index[1] ].pt;
	    stop = base_te[ te->vert.index[0] ].pt;
	    z = base_te[ te->vert.index[1] ].z; 	    
	    dz = base_te[ te->vert.index[0] ].z - 
	      base_te[ te->vert.index[1] ].z;
	}
	else if (edge == COLINEAR2 )
	{ /* 1 -> 3 */
	    start = base_te[ te->vert.index[0] ].pt;
	    stop = base_te[te->vert.index[2] ].pt;
	    z = base_te[ te->vert.index[0] ].z;
	    dz = base_te[te->vert.index[2] ].z - 
	      base_te[ te->vert.index[0] ].z;
	}
	else 
	{ /* 3 -> 0 */
	    start = base_te[ te->vert.index[2] ].pt;
	    stop = te->pt;
	    z = base_te[ te->vert.index[2] ].z;
	    dz = te->z - base_te[ te->vert.index[2] ].z;
	}
    }
    
    if ((dy = ((float64)stop.y - (float64)start.y)) != 0.0)
      return((float64)(z + dz*(((float64)pt->y -
				(float64)start.y)/dy)));
    else
    {	
	/* Use dx instead for horizontal edges */
	dy = (float64)stop.x - (float64)start.x;
	if (dy == 0.0)
	  return (float64)z;
	else
	  return((float64)(z + dz*(((float64)pt->x -
				    (float64)start.x)/dy)));
    }
} 

static CTDB_STATUS
ctdb_intern_look_at_adjacent_triangles16(els)
ELEVATION_LOOKUP_STORAGE_CMP *els;
{
    *checked_b++ = els->te_index ;/* Add TE to the list of checked TEs */
    
    do
    {
	/* If the TO BE SEARCHED queue is empty, then resort to a brute force
	 * search of all remaining terrain elements 
	 */
	if ( els->front == els->back)
	{ 
	    return CTDB_STATUS_FAILED;
	}
	/* Take next TE to look at from the list */
	els->te_index = els->te_queue[ els->front++ ];
    } while ( !ctdb_intern_validate_te_index16(els->te_index, checked_f,
					       checked_b) );  
    els->te = (ADDRESS)(((TERRAIN_ELEMENT_16 *)els->base) + els->te_index);
    return CTDB_STATUS_OK;
}

static CTDB_STATUS
ctdb_intern_look_at_all_triangles16(els)
    ELEVATION_LOOKUP_STORAGE_CMP *els;
{
    int32 found=FALSE;
    TERRAIN_ELEMENT_16 *lte;
    int32 n_te;
    
    n_te = ctdb_intern_number_tes(els->ctdb, els->num);
    
    els->te=els->base;
    lte = (TERRAIN_ELEMENT_16 *)els->base;
    for (els->te_index=0; els->te_index < n_te; els->te_index++, lte++)
    {
	els->te = (ADDRESS)lte;
	ctdb_intern_point_in_terrain_element16(els);
	
	if (els->where != OUTSIDE)
        {
	    ctdb_intern_set_elevation_and_soil16(els);
	    /* if not inside then on an edge and elevation is already
	       set by ctdb_intern_point_in_terrain_element16 */
	    return CTDB_STATUS_OK;
	}
    }
    return CTDB_STATUS_NO_ELEVATION_FOUND ;
}

#define CTDB_MAX_EDGES_PER_TE        5 /* max edges per terrain element */

/* This routine will return the next edge to be evaluated given the side/
 * edge last evaluated, the index of the terrain element and the list of 
 * terrain elements.
 */
CTDB_STATUS 
ctdb_intern_get_te_edge_index16(
    TERRAIN_ELEMENT_16  *te,
    uint32               index ,/*vertex number */
    int32                side  ,/*side of interest*/
    int32 	         edge[])
{
    int32 v;
    int32 geometry = ((int32)te[index].info.vertices - 2);
    
    /* TRIANGLE traversal : V -> V0 ; V0 -> V1 ; V1 -> V          
     * POLYGON  traversal : V -> V1 ; V1 -> V0 ; V0 -> V2 ; V2 -> V ; V -> V1
     */
    static int32 next[2][CTDB_MAX_EDGES_PER_TE][2]=
    { {{-1,0},{0,1},{1,-1},{-1,-1},{-1,-1}},
      {{-1,1},{1,0},{0,2},{2,-1},{-1,0}} };
    
    if (geometry < 0)
      return(CTDB_STATUS_NOT_A_TERRAIN_ELEMENT);              
    
    v = next[geometry][side][0];
    if ( v < 0 ) 
      edge[0] = index;
    else
      edge[0] = te[index].vert.index[v];
    
    v = next[geometry][side][1];
    
    if ( v < 0 ) 
      edge[1] = index;
    else
      edge[1] = te[index].vert.index[v];
    
    return(CTDB_STATUS_OK);              
}

/*********************
 * Search space code *
 *********************/

/* The compiler representation currently supports only a very limited
 * form of search space traversal.  Namely, it will only properly handle
 * iteration over the abstracts by patch, using ctdb_intern_next_quad_patch
 * and ctdb_intern_next_abstract.  This could obviously be extended, but it
 * would take some work.
 */

/* Create a search space for application-driven iteration */
CTDB_SEARCH_SPACE_PTR_CMP 
ctdb_intern_create_search_space(ctdb, min_x, min_y, max_x, max_y,
				control_flags)
    CTDB_CMP *ctdb; 
    int32     min_x;
    int32     min_y;
    int32     max_x;
    int32     max_y;
    uint32    control_flags;
{
    CTDB_SEARCH_SPACE_PTR_CMP space = (CTDB_SEARCH_SPACE_PTR_CMP)
      ctdb_alloc(NULL, sizeof(struct ctdb_search_space),
		 "Creating search space");
    float64 lat_start, lng_start, lat_stop, lng_stop;
    float64 h;
    int32 success;
    
    bzero(space, sizeof(struct ctdb_search_space_cmp));
    space->control_flags = control_flags;
    
    /* GCS: INITIALIZATION OF SEARCH SPACE CELLS */
    space->call_ctdb = ctdb;

    if (!ctdb)
    {
	space->min_x = space->max_x = space->min_y = space->max_y = 0;
	space->initial_frame = GCS_ILLEGAL_CELL;
    }
    else
    {
	space->min_x = min_x;
	space->min_y = min_y;
	space->max_x = max_x;
	space->max_y = max_y;
	space->initial_frame = ctdb_intern_get_cell_id(ctdb);
    }
    
    if (!ctdb_is_tiled || !ctdb)
    {
	space->n_cells = 1;
	space->cell_id_list = (int32 *)ctdb_alloc(NULL,
						  sizeof(int32),
						  "Search space cell list ");
	space->cell_id_list[0] = space->initial_frame;
        
    }
    else
    {
	/* find the list of cells to be searched */
	gcs_gcs_to_llz(space->initial_frame, 
		       (float64)min_x, (float64)min_y, (float64)0.0, 
		       &lat_start, &lng_start, &h, DATUM_WGS84); 
	gcs_gcs_to_llz(space->initial_frame, 
		       (float64)max_x, (float64)max_y,
		       (float64)0.0, &lat_stop, &lng_stop, &h, DATUM_WGS84);
	space->n_cells = 100; /* initial maximum number of cells */
	space->cell_id_list = 
	  (int32 *)ctdb_alloc(NULL,
			      sizeof(int32)*space->n_cells, 
			      "Search space cell list ");
	success = gcs_extent_to_cell_list(lat_start, lng_start, 
					  lat_stop, lng_stop, 
					  &space->n_cells, 
					  space->cell_id_list);
	/* retry if cell_id_list not big enough */
	if (!success)
	{
	    space->cell_id_list = (int32 *)
	      ctdb_alloc(space->cell_id_list,
			 sizeof(int32)*space->n_cells, 
			 "Search space cell list ");
	    success = gcs_extent_to_cell_list(lat_start, lng_start, 
					      lat_stop, lng_stop, 
					      &space->n_cells, 
					      space->cell_id_list);
	}
    }
    
    /* Now that we know how many cells there are, allocate space for the
     * list of database iterators
     */
    space->cell_data_list = (CTDB_SEARCH_DB_CMP *)
      ctdb_alloc(NULL, space->n_cells * sizeof(CTDB_SEARCH_DB_CMP),
		 "Search space iterators");
    
    initialize_gcs_search_spaces(ctdb, space);
    
    return space;
}

static void  
transform_abstract_vertices(space, index, nverts, src, sink)
    CTDB_SEARCH_SPACE_PTR_CMP space; 
    int32                     index;
    int32                     nverts;
    float32                  *src;
    float32                  *sink;
{
    int32 i;
    float64 tempz = 0.0;
    float64 newx, newy;
    
    for (i=0; i<nverts; i++)
    {
	gcs_gcs3e_to_gcscs(space->cell_id_list[index],
			 (float64)src[i*2], (float64)src[i*2 +1], 0.0,
			 space->initial_frame,&newx, &newy, &tempz);
	sink[i*2] = newx;
	sink[i*2+1] = newy;
    }
}

static void 
transform_reference_frame(space, index, x, y, z)
    CTDB_SEARCH_SPACE_PTR_CMP space;
    int32                     index;
    float64                  *x;
    float64                  *y;
    float64                  *z;
{
    float64 new_z;
    float64 new_x;
    float64 new_y;
    
    
    if (space->initial_frame == space->cell_id_list[index])
      return;
    else
      if (z)
      {
	  gcs_gcs3e_to_gcscs(space->cell_id_list[index], *x, *y, *z, 
			   space->initial_frame, &new_x, &new_y, &new_z);
	  *x = new_x;
	  *y = new_y;
	  *z = new_z;
      }
      else
      {
	  gcs_gcs3e_to_gcscs(space->cell_id_list[index], *x, *y, 0.0,  
			   space->initial_frame, &new_x, &new_y, &new_z);
	  *x = new_x;
	  *y = new_y;
      }
}

/* Set up a search_db to allow iterating over the specified cell */
static void
init_gcs_search_db(space, search_db, index)
    CTDB_SEARCH_SPACE_PTR_CMP     space;
    CTDB_SEARCH_DB_CMP           *search_db;
    int32                         index;
{
    float64   min_x, min_y, max_x, max_y;
    float64   temp_z;
    int32     success;
    int32     temp;
    CTDB_CMP *ctdb;
    
    bzero(search_db, sizeof(CTDB_SEARCH_DB_CMP));
    
    search_db->current_index = index;
    
    if (space->cell_id_list[index] == GCS_ILLEGAL_CELL)
    {
	search_db->ctdb = space->call_ctdb;
    }
    else
    {
	search_db->ctdb = (CTDB_CMP *)gcs_get_tdb(space->cell_id_list[index],
						  FALSE);
    }

    if (search_db->ctdb == NULL)
    {
	return;
    }
    
    /* Find min_x, min_y, max_x , max_y for this frame of reference */
    
    if ((ctdb_is_tiled) &&
	(space->cell_id_list[index] != space->initial_frame))
    {
	gcs_gcs3e_to_gcscs(space->initial_frame, (float64)space->min_x, 
			 (float64)space->min_y, (float64)0.0,	
			 space->cell_id_list[search_db->current_index],
			 &min_x, &min_y, &temp_z);
	
	gcs_gcs3e_to_gcscs(space->initial_frame,(float64)space->max_x, 
			 (float64)space->max_y,(float64)0.0,
			 space->cell_id_list[search_db->current_index],
			 &max_x, &max_y, &temp_z);
    }
    else
    {
	min_x = space->min_x;
	min_y = space->min_y;
	max_x = space->max_x;
	max_y = space->max_y;
    }
    
    if (min_x < search_db->ctdb->min_x)
      min_x = search_db->ctdb->min_x;
    if (min_y < search_db->ctdb->min_y)
      min_y = search_db->ctdb->min_y;
    if (max_x > (search_db->ctdb)->max_x)
      max_x = search_db->ctdb->max_x;
    if (max_y > (search_db->ctdb)->max_y)
      max_y = search_db->ctdb->max_y;
    
    /* Compute iteration bounds */
    search_db->min_patch_x = PATCH_X(min_x*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    search_db->max_patch_x = PATCH_X(max_x*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    search_db->min_patch_y = PATCH_Y(min_y*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    search_db->max_patch_y = PATCH_Y(max_y*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    
    if (!(space->control_flags & CTDB_NO_ABSTRACT))
      ctdb_intern_find_search_quads(search_db, (int32)min_x, (int32)min_y, 
				    (int32)max_x, (int32)max_y);
    
    /* Bound search to database limits, and make sure min<max */
    search_db->min_patch_x=NS_LIMITS(0, search_db->min_patch_x, 
				     search_db->ctdb->patches_wide-1);
    search_db->min_patch_y=NS_LIMITS(0, search_db->min_patch_y, 
				     search_db->ctdb->patches_high-1);
    search_db->max_patch_x=NS_LIMITS(0, search_db->max_patch_x, 
				     search_db->ctdb->patches_wide-1);
    search_db->max_patch_y=NS_LIMITS(0, search_db->max_patch_y, 
				     search_db->ctdb->patches_high-1);

    /* Obscure case of an empty database, where the last bound is -1 */
    if (!search_db->ctdb->patches_wide || !search_db->ctdb->patches_high)
    {
	search_db->min_patch_x = search_db->min_patch_y = 
	  search_db->max_patch_x = search_db->max_patch_y = 0;
    }
    else
    {
	if (search_db->min_patch_x > search_db->max_patch_x)
	  search_db->min_patch_x = search_db->max_patch_x;
	if (search_db->min_patch_y > search_db->max_patch_y)
	  search_db->min_patch_y = search_db->max_patch_y;
    }
    return;
}

static void 
initialize_gcs_search_spaces(ctdb, space)
    CTDB_CMP                 *ctdb;
    CTDB_SEARCH_SPACE_PTR_CMP space;
{
    float64 min_x, min_y, max_x, max_y;
    float64 temp_z;
    int32 success;
    int32 i;
    
    /* Set up the boilerplate iterators for each cell; whenever we need
     * to move to a new cell hereafter, we'll just copy the appropriate
     * structure in
     */
    for(i=0;i<space->n_cells;i++)
    {
	init_gcs_search_db(space, &(space->cell_data_list[i]), i);
    }
    
    /* We want to start all of our iterators off in the first valid
     * cell.  If we set the index to -1 and call advance_gcs_search_space,
     * it will find that for us, and we can copy that one everywhere.
     */
    space->search_patch.current_index = -1;
    
    advance_gcs_search_space(space, &space->search_patch);
    
    /* If we're traversing by patch, then space->search_patch is the
     * only CTDB_SEARCH_DB we'll use, so we won't fill out any of the
     * others.  Otherwise, we need to populate all of them.
     */
    if (!(space->control_flags & CTDB_BY_PATCH))
    {
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.tree.search_patch));
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.building.search_patch));
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.linear.search_patch));
	
	for(i=0; i<CTDB_ABSTRACT_MAX; i++)
	{
	    copy_search_db(&(space->search_patch), 
			   &(space->iteration.abstract[i].search_patch));
	}
    }
}

static void 
copy_search_db(src, sink)
    CTDB_SEARCH_DB_CMP *src;
    CTDB_SEARCH_DB_CMP *sink;
{
    int32 min_x, min_y, max_x, max_y;
    
    bcopy(src, sink, sizeof(CTDB_SEARCH_DB_CMP));
    
    /* The one thing the bcopy won't catch is quad_nodes.  Set it up
     * appropriately--it's really tempting to just copy the pointer,
     * but I don't want to have to debug this when somebody decides to
     * free what we were looking at...
     */
    if (src->num_quad_nodes)
    {
	sink->quad_nodes = (uint32 *)
	  ctdb_alloc(NULL, 
		     src->num_quad_nodes * sizeof(uint32),
		     "Quad node list for search space");
	
	bcopy(src->quad_nodes, sink->quad_nodes,
	      src->num_quad_nodes * sizeof(uint32));
    }
}

static int32 
advance_gcs_search_space(space, current)
    CTDB_SEARCH_SPACE_PTR_CMP space;
    CTDB_SEARCH_DB_CMP       *current;
{
    int32 next_index;
    
    next_index = current->current_index + 1;
    
    /* Check for tiled GCS mode */
    if ((!ctdb_is_tiled) && (next_index > 0))
    {
	bzero(current, sizeof(CTDB_SEARCH_DB_CMP));
	return FALSE;
    }
    
    /* The search_db structures for each cell are all precalculated, we'll
     * just find the first interesting one and return it.
     */
    do
    {
	if (next_index >= space->n_cells)
	{
	    bzero(current, sizeof(CTDB_SEARCH_DB_CMP));
	    current->current_index = next_index;
	    return FALSE;
	}
	copy_search_db(&space->cell_data_list[next_index++], 
		       current);
    } while (current->ctdb == NULL);
    return(TRUE);
}

/* Free a search space */
void 
ctdb_intern_destroy_search_space(space)
    CTDB_SEARCH_SPACE_PTR_CMP space;
{
    CTDB_SEARCH_DB_CMP current;    
    int32 i;
    
    if (space->cell_id_list)
      STDDEALLOC(space->cell_id_list);
    
    if (space->cell_data_list)
      STDDEALLOC(space->cell_data_list);
    
    current = space->search_patch;
    if (current.quad_nodes)
      STDDEALLOC(current.quad_nodes);
    
    current = space->iteration.tree.search_patch;
    if (current.quad_nodes)
      STDDEALLOC(current.quad_nodes);
    
    current = space->iteration.building.search_patch;
    if (current.quad_nodes)
      STDDEALLOC(current.quad_nodes);
    
    current = space->iteration.linear.search_patch;
    if (current.quad_nodes)
      STDDEALLOC(current.quad_nodes);
    
    for(i=0;i<CTDB_ABSTRACT_MAX;i++)
    {
	current = space->iteration.abstract[i].search_patch;
	if (current.quad_nodes)
	  STDDEALLOC(current.quad_nodes);
    }
    
    STDDEALLOC(space);
}

/* Chunk size used when allocating quad nodes */
#define ALLOC_CHUNK 32

/* Globals used during search */
static int32 q_x0, q_y0, q_x1, q_y1;

/* Add the current quad to the list, and recurse on its children
 * as needed.
 */
static void 
ctdb_add_quad(search_db, offset, lo_x, lo_y, hi_x, hi_y, alloced)
    CTDB_SEARCH_DB_CMP *search_db;
    int32               offset;
    int32               lo_x;
    int32               lo_y;
    int32               hi_x;
    int32               hi_y;
    int32              *alloced;
{
    CTDB_CMP *ctdb = search_db->ctdb;
    CTDB_FEATURE_DATA_CMP *patch;
    struct ctdb_quad_data_cmp *qdata;
    int32 mid_x, mid_y;
    uint32 to_do;
    
    /* Error check */
    if ((ctdb == NULL) || (offset >= ctdb->num_quad_data))
      return;
    
    /* Make room for more nodes */
    if (*alloced == search_db->num_quad_nodes)
    {
	int32 size;
	
	*alloced += ALLOC_CHUNK;
	size = (*alloced) * sizeof(uint32);
	if (search_db->quad_nodes)
	  search_db->quad_nodes = (uint32 *)
	    STDREALLOC(search_db->quad_nodes, size);
	else
	  search_db->quad_nodes = (uint32 *)STDALLOC(size);
    }
    
    /* Stuff in this quad node */
    search_db->quad_nodes[search_db->num_quad_nodes] = CTDB_QUAD_PATCH(offset);
    search_db->num_quad_nodes++;
    
    /* Get this node's data */
    patch = search_db->ctdb->quad_data + offset;
    if (!patch->hdr.size)
      return;
    
    /* See if this is a leaf quad node */
    if (patch[1].abstract.type_code != CTDB_ABSTRACT_QUAD_NODE)
      return;
    
    /* Get the starting addresses of the leaves */
    qdata = &patch[2].abstract_data.quad_data;
    
    /* The leaves are stored: 23
     *                        01
     * See which ones we need to include (actually we test for which
     * ones we *don't* have to include, because that's easier).
     */
    mid_x = (lo_x + hi_x) / 2;
    mid_y = (lo_y + hi_y) / 2;
    
    /* Make a bit mask with bits 0-3 on, then turn off those which are
     * eliminated.
     */
    to_do = 0xf;
    if (q_x0 > mid_x) /* Eliminate 0 & 2, leave 1010 = 0xA */
      to_do &= 0xA;
    if (q_y0 > mid_y) /* Eliminate 0 & 1, leave 1100 = 0xC */
      to_do &= 0xC;
    if (q_x1 < mid_x) /* Eliminate 1 & 3, leave 0101 = 0x5 */
      to_do &= 0x5;
    if (q_y1 < mid_y) /* Eliminate 2 & 3, leave 0011 = 0x3 */
      to_do &= 0x3;
    
    /* Go down and add the leaf nodes which are present (a node is
     * allowed to have only some of its leaves used).
     */
    if ((to_do & 0x1) && qdata->child_start[0])
      ctdb_add_quad(search_db, qdata->child_start[0],
		    lo_x, lo_y, mid_x-1, mid_y-1, alloced);
    if ((to_do & 0x2) && qdata->child_start[1])
      ctdb_add_quad(search_db, qdata->child_start[1],
		    mid_x, lo_y, hi_x, mid_y-1, alloced);
    if ((to_do & 0x4) && qdata->child_start[2])
      ctdb_add_quad(search_db, qdata->child_start[2],
		    lo_x, mid_y, mid_x-1, hi_y, alloced);
    if ((to_do & 0x8) && qdata->child_start[3])
      ctdb_add_quad(search_db, qdata->child_start[3],
		    mid_x, mid_y, hi_x, hi_y, alloced);
}

/* Make a list of quads to be searched */
static void 
ctdb_intern_find_search_quads(search_db, min_x, min_y, max_x, max_y)
    CTDB_SEARCH_DB_CMP *search_db;
    int32               min_x;
    int32               min_y; 
    int32               max_x;
    int32               max_y;
{
    CTDB_CMP *ctdb = search_db->ctdb;
    int32 alloced;
    
    /* Stick the search bounds into some static globals to avoid
     * argument-passing overhead.
     */
    q_x0 = min_x;
    q_y0 = min_y;
    q_x1 = max_x;
    q_y1 = max_y;
    
    /* Start the recursive search */
    alloced = 0;
    ctdb_add_quad(search_db, (int32)0,
		  ctdb->min_x, ctdb->min_y, ctdb->max_x, ctdb->max_y,
		  &alloced);
}

/* Get a pointer to the current patch data */
static CTDB_FEATURE_DATA_CMP *
current_patch(space, current, number, zero_offset, patch_x, patch_y)
    CTDB_SEARCH_SPACE_PTR_CMP space;
    CTDB_SEARCH_DB_CMP       *current;
    int32                    *number;
    int32                    *zero_offset;
    int32                    *patch_x;
    int32                    *patch_y;
{
    /* If we are traversing by patch, use the global `current' structure */
    if (space->control_flags & CTDB_BY_PATCH)
    {
	current = &space->search_patch;
	/* Make sure the application is paying attention */
	if (!current->last_patch)
	  return NULL;
    }
    else
    {
	printf("Programmer error--compiler rep only supports search by patch.\n");
	exit(-1);
    }
    
    /* Since we don't cache in the compiler, the current patch is guaranteed
     * to be in memory, so we won't bother checking, we'll just return the
     * answer
     */
    if (number)
      *number = current->last_number;
    if (patch_x)
      *patch_x = current->x,
      *patch_y = current->y;
    return current->last_patch;
    
}    





/* Return the next abstract feature */
int32 
ctdb_intern_next_abstract(space, type_code, vertices, data, data_size)
    CTDB_SEARCH_SPACE_PTR_CMP space;
    int32                     type_code;
    float32                 **vertices;
    CTDB_ABSTRACT_DATA_CMP  **data;
    int32                    *data_size;
{
    struct ctdb_search_state_cmp *state;
    CTDB_FEATURE_DATA_CMP *patch, *feature;
    int32 patch_number;
    int32 verts, atype, dsize, offset, next, last;
    float32 *local_verts = NULL;	/* Buffer for returning vertices */
    
    /* Make sure the type_code is known */
    if ((type_code < 1) || (type_code >= CTDB_ABSTRACT_MAX))
      return 0;
    
    state = space->iteration.abstract+type_code;
    
    /* Make sure we aren't finished */
    if (state->all_done)
      return 0;
    
    /* Abstract features live in quad nodes, so if we are traversing
     * by patch, make sure we are in a quad node.
     */
    if (space->control_flags & CTDB_BY_PATCH)
    {
	if (!CTDB_IS_QUAD_NODE(space->search_patch.last_number))
	  return 0;
    }
    else 
    {
	printf("Programmer error:  compiler rep only supports searching by patch.\n");
	exit(-1);
    }
    
    /* Iterate until we find something, or exhaust the list of patches */
    for(;;)
    {
	/* Make sure the patch is in memory */
	if (!(patch = current_patch(space, &state->search_patch, &patch_number,
				    &state->offset,(int32*)NULL,(int32*)NULL)))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	
	/* The following test should always return TRUE because we
	 * skipped the real patches at the top of this function.  Thus
	 * the following is just an error check.
	 */
	/* Now This could happen if we changed cells */
	if (!CTDB_IS_QUAD_NODE(patch_number))
	{
	    /*state->all_done = TRUE;
	      return 0;*/
	    state->search_patch.real_patches = 0;
	    state->search_patch.q = -1;
	}
	else
	{
	    /* Note the location of the last feature in the patch */
	    last = patch->hdr.size;
	    
	    /* See if we just started this patch */
	    if (!state->offset)
	    {
		/* Skim forward to the first abstract of this type */
		for (offset=1;offset<last;offset=next)
		{
		    feature = patch+offset;
		    verts = feature->abstract.num_vertices;
		    atype = feature->abstract.type_code;
		    dsize = feature->abstract.data_size;
		    
		    if (atype != type_code)
		    {
			next = offset+1+dsize+verts*2;
			continue;
		    }
		    else
		      break;
		}
		if (offset < last)
		  state->offset = offset;
		else
		  atype = -1;
	    }
	    else
	    {
		/* Go on to the next abstract. */
		feature = patch+state->offset;
		verts = feature->abstract.num_vertices;
		atype = feature->abstract.type_code;
		dsize = feature->abstract.data_size;
		state->offset += 1+dsize+verts*2;
		if (state->offset < last)
		{
		    feature = patch+state->offset;
		    verts = feature->abstract.num_vertices;
		    atype = feature->abstract.type_code;
		    dsize = feature->abstract.data_size;
		}
		else
		  atype = -1;
	    }
	    
	    /* Is the thing we are looking at the right type? */
	    if (atype == type_code)
	    {
		/* Return the abstract feature.
		 */
		if (vertices)
		{
		    if (ctdb_is_tiled)
		    {
			/* We need to transform the stored data into the
			 * desired frame of reference.  We obviously can't
			 * do that in place, so we'll do it in a static buffer
			 * that we'll expand as necessary
			 */
			/* $$$ This is, of course, a memory leak when
			 * operating in GCS multi-cell mode.  Some day we
			 * should fix this, but it requires modifying the
			 * interface, so we won't do it now.
			 */
			local_verts = 
			  (float32 *)ctdb_alloc(NULL,
						2*verts*sizeof(float32),
						"abstract vertices");
			*vertices = local_verts;
			
			transform_abstract_vertices(
					    space,
					    state->search_patch.current_index,
					    verts, (float32 *)(feature +1),
					    *vertices);
		    }
		    else 	
		      *vertices = (float32 *)(feature+1);
		}
		
		if (data)
		  *data = (CTDB_ABSTRACT_DATA_CMP *)(feature+1+verts+verts);
		if (data_size)
		  *data_size = dsize * sizeof(int32);
		
		return verts;
	    }
	}
	/* We've exhausted the abstracts in this patch, so advance to the
	 * next patch (if allowed).
	 */
	if (space->control_flags & CTDB_BY_PATCH)
        {
	    state->all_done = TRUE;
	    return 0;
	}
	else
	{
	    printf("Programmer error--compiler rep only supports by patch.\n");
	}
    }
}

/* ctdb_next_quad_patch allows the search space to advance to the next node of
 * the quad tree.  This function must be called only if the control flag
 * CTDB_BY_PATCH was passed to ctdb_create_search_space().  
 */
int32 
ctdb_intern_next_quad_patch(space)
    CTDB_SEARCH_SPACE_PTR_CMP space;
{
    CTDB_SEARCH_DB_CMP *current;
    CTDB_FEATURE_DATA_CMP *found;
    int32 initializing;
    int32 trash;

   /* This should only be called if CTDB_BY_PATCH is being used. */
    if (!(space->control_flags & CTDB_BY_PATCH))
    {
	printf("Error:  ctdb_next_quad_patch called on non by-patch space.\n");
	return 0;
    }

    /* Initialize all the iteration variables */
    bzero(&space->iteration, sizeof(space->iteration));

    /* Move on to the next patch (which may be the first patch) */
    current = &space->search_patch;
    if (!current->last_patch)
    {
	current = &space->search_patch;
	current->real_patches = 0;
	current->q = -1;
    }
    do
    {
	/* Increment to next quad */
	current->q++;
	if (current->q >= current->num_quad_nodes)
	{
	    /* We've exhausted all patches */
	    return NULL;
	}
	current->last_number = current->quad_nodes[current->q];
	if (CTDB_IS_QUAD_NODE(current->last_number))
	  found = (current->ctdb)->quad_data + 
	    CTDB_QUAD_OFFSET(current->last_number);
    } while (!found);

    current->last_patch = found;
    return (current->last_patch != NULL); /* Success */
}


/* This routine will determine the maximum elevation within the given 
 * gridded patch by checking the elevation of each post within the patch 
 */    
void 
ctdb_intern_compute_grid_max_z(ctdb, patchx, patchy, max_elev)
    CTDB_CMP       *ctdb;
    int32           patchx, patchy;
    float64        *max_elev;    
{
    int32 postx, posty;
    float32 pz;
    int32 i,j;

    postx = patchx * ctdb->patch_incr; /* convert patch coords to post
					      coords */
    posty = patchy * ctdb->patch_incr;
      
    /* loop over the posts within the given patch and
     * return the maximum post elevation.
     */
    for (j = 0; j < (ctdb->patch_incr+1); j++)
      for (i = 0; i< (ctdb->patch_incr+1); i++)
	{
	    pz = ctdb_intern_lookup_post_elevation(ctdb, postx + i,
						posty + j);
	    if (pz > *max_elev)
	      *max_elev = pz;
	}
}    


/* Traverses the grid squares and patches, calling a user-defined function 
 * for each.  Note that this version does not have all the functionality of
 * of the one in libctdb.  In particular, it doesn't handle rays that start
 * at anything but a point, it cannot generate patch feature masks for
 * later use (the feature present bits in the compiler rep are generally not
 * set correctly when this function is called), and it does not bother
 * performing elevation lookups.
 */
int32 
ctdb_intern_intervis(ctdb, start, end, post_fcn, patch_fcn, 
		     feature_mask)
    CTDB_CMP          *ctdb;
    float64            start[3];
    float64            end[4];
    int32            (*post_fcn)();
    int32            (*patch_fcn)();/* Ignored if NULL */
    int32              feature_mask;
{
    CTDB_PATCH_GROUP_CMP *pg;
    int32 member;
    float64 zSW, zNE, zSE;
    float64 x, y, zatx1, zatx2, zaty1, zaty2;
    float64 xdiag, ydiag, zdiag1, zdiag2;
    float64 dydx, dxdy, dzdx1, dzdx2, dzdy1, dzdy2, not_fetched;
    float64 xintr, yintr, denom;
    float64 zxintr1, zxintr2, zyintr1, zyintr2;
    float64 xpery, yperx, zpery1, zpery2, zperx1, zperx2;
    float64 perp_dydx_nw_se, perp_dydx_sw_ne;
    float64 xlow, xhigh, ylow, yhigh;
    int32 postx, posty;
    float64 postxD, postyD;
    int32 ptx, pty, patchx, patchy, patch_number, post_bit;
    int32 x0, y0, x1, y1;
    float64 x1m, x1p, y1m, y1p;
    int32 xstep, ystep;
    float64 xstepD, ystepD;
    int32 move_toward_x;
    int32 horizontal, vertical;
    int32 post_value;
    float64 post_z;
    float64 fpb = ctdb->fixed_point_basis;
    int32 patch_mask, post_mask;
    int32 quadrant;
    int32 patches_wide = ctdb->patches_wide;
    int32 patch_incr = ctdb->patch_incr;
    int32 top_to_bottom = (patch_incr * (patch_incr - 1));
#define SET_NEW_PATCH_DATA                                              \
    {                                                                   \
       pg       = ctdb->patch_groups+PG_NUMBER(patch_number);           \
       member   = PG_MEMBER(patch_number);                              \
    }                  

    ctdb_n_feat = 0;

    /* A bunch of variables are kept in two formats: integer & double.
     * The convention is
     *		int32 variable;
     *		float64 variableD;
     * Keeping around both formats reduces the number of type conversions
     * needed, reducing the cost of the routine by about 10%.
     */

    /* Through the miracle of floating point, we can work in units of
     * incr meters without losing any accuracy.  This allows us to use
     * the constants 1 and -1 in lots of places we would otherwise
     * have to use incr and -incr (yielding about an 8% speed
     * improvement).  It also has the nice effect that the
     * post_fcn() gets called with post numbers so it can do its
     * lookup without needing to do a divide (about a 30% cost
     * savings).  Hence, start[X],start[Y],end[X],end[Y] ARE ASSUMED
     * TO BE IN UNITS OF ctdb->incr!
     */

    /* Find min/max of supplied line for boundary checks */
    /* Figure out which direction we will be traversing the grid posts. */
    if (start[X]<end[X])
      xlow = start[X], xhigh = end[X], xstep = 1, xstepD = 1.0;
    else
      xlow = end[X], xhigh = start[X], xstep = -1, xstepD = -1.0;

    if (start[Y]<end[Y])
      ylow = start[Y], yhigh = end[Y], ystep = 1, ystepD = 1.0;
    else
      ylow = end[Y], yhigh = start[Y], ystep = -1, ystepD = -1.0;

    /* Compensate for lack of numerical accuracy */
    if (xhigh - xlow < 1.0E-7)
      xhigh = xlow = end[X] = start[X], xstep = 1, xstepD = 1.0;
    if (yhigh - ylow < 1.0E-7)
      yhigh = ylow = end[Y] = start[Y], ystep = 1, ystepD = 1.0;

    /* We will test these later to deal with special slopes */
    horizontal = (ylow == yhigh);
    vertical = (xlow == xhigh);

    /* Catch zero distance. */
    if (horizontal && vertical)
      return 1;

    /* Find our starting and ending post */
    x0 = POST_X(start[X]);
    y0 = POST_Y(start[Y]);
    x1 = POST_X(end[X]);
    y1 = POST_Y(end[Y]);

    x1m = x1 - 1.0E-7;
    x1p = x1 + 1.0E-7;
    y1m = y1 - 1.0E-7;
    y1p = y1 + 1.0E-7;

    /* Find our starting patch and our location within it */
    ptx = PATCH_X(x0*ctdb->patch_inv_incr);
    pty = PATCH_Y(y0*ctdb->patch_inv_incr);
    patch_number = ptx + pty * patches_wide;
    SET_NEW_PATCH_DATA;
    patchx = x0 - ptx * patch_incr;
    patchy = y0 - pty * patch_incr - 1;
    post_bit = (1<<patchx) << (patchy*patch_incr);

    /* Find the slopes with respect to each axis.  Set infinite slopes to
     * zero (they won't be used anyway) to speed up the math (infinities in
     * equations slow the fp hardware way down).  I pull out the repeated
     * denominators as an optimization (only about 2% faster since this
     * code only runs once per call).
     */
    if (horizontal)
    {
	dxdy = dydx = dzdy1 = dzdy2 = 0;
	denom = 1/(end[X]-start[X]);
	dzdx1 = (end[ZL]-start[Z])*denom;
	dzdx2 = (end[ZH]-start[Z])*denom;
    }
    else if (vertical)
    {
	dxdy = dydx = dzdx1 = dzdx2 = 0;
	denom = 1/(end[Y]-start[Y]);
	dzdy1 = (end[ZL]-start[Z])*denom;
	dzdy2 = (end[ZH]-start[Z])*denom;
    }
    else
    {
	denom = 1/(end[Y]-start[Y]);
	dxdy = (end[X]-start[X])*denom;
	dzdy1 = (end[ZL]-start[Z])*denom;
	dzdy2 = (end[ZH]-start[Z])*denom;

	denom = 1/(end[X]-start[X]);
	dydx = (end[Y]-start[Y])*denom;
	dzdx1 = (end[ZL]-start[Z])*denom;
	dzdx2 = (end[ZH]-start[Z])*denom;
    }

    /* Find the x- y- and z-intercepts */
    xintr = start[X] - dxdy * start[Y];
    yintr = start[Y] - dydx * start[X];
    zxintr1 = start[Z] - dzdx1 * start[X];
    zyintr1 = start[Z] - dzdy1 * start[Y];
    zxintr2 = start[Z] - dzdx2 * start[X];
    zyintr2 = start[Z] - dzdy2 * start[Y];

    /* Figure out where the supplied line crosses y0 */
    x = dxdy * y0 + xintr;

    /* Figure out where the supplied line crosses x0 */
    y = dydx * x0 + yintr;

    /* Figure out where z is at x0 and y0 */
    zatx1 = dzdx1 * x0 + zxintr1;
    zaty1 = dzdy1 * y0 + zyintr1;
    zatx2 = dzdx2 * x0 + zxintr2;
    zaty2 = dzdy2 * y0 + zyintr2;

    /* Find the change in X along the line for each step in the Y direction.
     * We can add this to the previous value to find the next intersection.
     */
    xpery = ystepD * dxdy;

    /* Find the change in Y along the line for each step in the X direction
     * We can add this to the previous value to find the next intersection.
     */
    yperx = xstepD * dydx;

    /* Find the change in Z along the line for each kind of step */
    zperx1 = xstepD * dzdx1;
    zpery1 = ystepD * dzdy1;
    zperx2 = xstepD * dzdx2;
    zpery2 = ystepD * dzdy2;

    /* To find the intersection of our line with a NE-to-SW diagonal we
     * use the equations:
     * X-on-diagonal = (y-intercept - Ypost - Xpost) / (-1 - dydx)
     * Y-on-diagonal =  - X-on-diagonal + Ypost + Xpost
     * These are easy to derive (the slopes of the diagonals == -1,
     * the intercepts == Ypost + Xpost, the rest is plug&chug).
     *
     * Of course, We have to use different equations to find the
     * intersection of our line with a SW-to-NE diagonal:
     * X-on-diagonal = (y-intercept - Ypost + Xpost + 1) / (1 - dydx)
     * Y-on-diagonal = X-on-diagonal + Ypost - Xpost - 1
     * (derived from diagonal slope == 1, intercept = Ypost - Xpost - 1).
     */

    /* Figure the denominator of the diagonal intercept equation.  For
     * vertical lines, set this to the magic value 0.0 (which cannot be
     * generated by 1/-1-slope, or 1/1-slope) which we can check later
     * and deal with as a special case.  For slope=-1 or slope=1 lines
     * (depending on diagonalization) set the denominator to an
     * almost-infinte-but-a-little-less-to-not-slow-the-hardware-down
     * value.
     */
    if (vertical)
      perp_dydx_nw_se = perp_dydx_sw_ne = 0.0;
    else
    {
	if (dydx == -1.0)
	  perp_dydx_nw_se = -1.0E+127;
	else
	  perp_dydx_nw_se = 1 / (-1 - dydx);

	if (dydx == 1.0)
	  perp_dydx_sw_ne = 1.0E+127;
	else
	  perp_dydx_sw_ne = 1 / (1 - dydx);
    }

    /* WITHIN_LINE tests whether a point lies on the user-supplied
     * segment.  We need this mostly to test for boundary conditions.
     * Since the x,y is guaranteed to fall along the line, we can get
     * away with only testing one dimension as long as that dimension
     * is meaningful.  Therefore we test x, and in the special case
     * of a vertical line, we also test y.
     */

#define WITHIN_LINE(x,y) \
    (((x)>=xlow) && ((x)<=xhigh) && \
     (!vertical || (((y)>=ylow) && ((y)<=yhigh))))

    /* IS_POST_X determines if POST_X(x) == p without actually
     * calling POST_X(x).  This is a real win speed wise.  Similarly
     * for IS_POST_Y.
     */

#define IS_POST_X(x,p) (((x)>=(p)) && ((x)<((p)+1.0)))
#define IS_POST_Y(x,p) (((x)>((p)-1.0)) && ((x)<=(p)))

    /* NOT_FETCHED is a flag value used to identify elevations which
     * have not yet been retrieved.
     */
#define NOT_FETCHED -1234567.8

    /* CROSS_DIAG_* are called whenever we cross a diagonal.  The
     * post coordinates postx,posty are at the northeast corner of the
     * diagonal.  We compute the z value based upon xdiag, except
     * in the case of a vertical line where we must base it on ydiag.
     */

#define CROSS_DIAG_NW_SE						\
    {									\
	if (perp_dydx_nw_se)						\
	  xdiag = perp_dydx_nw_se * (yintr - posty - postx);		\
	else								\
	  xdiag = start[X];						\
									\
	if (IS_POST_X(xdiag, postxD))					\
	{								\
	    ydiag = posty + postx - xdiag;				\
	    if (WITHIN_LINE(xdiag, ydiag))				\
	    {								\
		if (!vertical)						\
		{							\
		    zdiag1 = dzdx1 * xdiag + zxintr1;			\
		    zdiag2 = dzdx2 * xdiag + zxintr2;			\
		}							\
		else							\
		{							\
		    zdiag1 = dzdy1 * ydiag + zyintr1;			\
		    zdiag2 = dzdy2 * ydiag + zyintr2;			\
		}							\
									\
		if (zSE==NOT_FETCHED)					\
	          zSE=ctdb_lookup_safe_post_elevation(ctdb,postx+1,posty-1);\
		if (!post_fcn(ctdb, xdiag, ydiag, zdiag1, zdiag2,	\
			      post_z + (xdiag - postx) * (zSE - post_z)))\
		  return 0;						\
	    }								\
	}								\
    }
#define CROSS_DIAG_SW_NE						\
    {									\
	if (perp_dydx_sw_ne)						\
	  xdiag = perp_dydx_sw_ne * (yintr - posty + postx + 1);	\
	else								\
	  xdiag = start[X];						\
									\
	if (IS_POST_X(xdiag, postxD))					\
	{								\
	    ydiag = posty - postx + xdiag - 1;				\
	    if (WITHIN_LINE(xdiag, ydiag))				\
	    {								\
		if (!vertical)						\
		{							\
		    zdiag1 = dzdx1 * xdiag + zxintr1;			\
		    zdiag2 = dzdx2 * xdiag + zxintr2;			\
		}							\
		else							\
		{							\
		    zdiag1 = dzdy1 * ydiag + zyintr1;			\
		    zdiag2 = dzdy2 * ydiag + zyintr2;			\
		}							\
									\
		if (zSW == NOT_FETCHED)					\
		  zSW=ctdb_lookup_safe_post_elevation(ctdb,postx,posty-1);\
		if (zNE == NOT_FETCHED)					\
		  zNE=ctdb_lookup_safe_post_elevation(ctdb,postx+1,posty);\
		if (!post_fcn(ctdb, xdiag, ydiag, zdiag1, zdiag2,	\
			      zSW + (xdiag - postx) * (zNE - zSW)))	\
		  return 0;						\
	    }								\
	}								\
    }

    /* NEW_POST is called whenever we change the value of posty or
     * postx.  x contains the x coordinate of the intersection of the
     * actual line with the posty horizontal.  y contains the y
     * coordinate of the intersection of the actual line with the
     * postx vertical.  The point postx,posty is guaranteed to be
     * northwest of these intersections, and will be found only once.
     * First we check the patch to see if we left an area with features
     * which need to be checked.  Then we check the horizontal, vertical,
     * and diagonal edges for blockage.  Note that we do not check
     * horizontal edges against horizontal lines, nor vertical edges
     * against vertical lines (this isn't a speed optimization, for
     * Flat lines, we cannot guarantee that <x,postyD> or <postxD,y>
     * will lie on the line, therefore the WITHIN_LINE routine doesn't
     * work).
     */

#define NEW_POST							\
    {	                                                                \
        patch_mask |= post_value;                                       \
        post_mask |= post_bit;						\
    }
    /* CHECK_PATCH checks whether we have crossed into a new patch.
     * If so, we call the patchwise visibility function to test for
     * interference from microterrain and such.
     */
#define CHECK_PATCH						         \
    {								         \
       if (patch_mask & feature_mask)				         \
       {							         \
	  ctdb_feat_list[ctdb_n_feat].patch_number = patch_number;       \
	  ctdb_feat_list[ctdb_n_feat].mask = patch_mask & feature_mask;  \
	  ctdb_feat_list[ctdb_n_feat].post_mask = post_mask;             \
	  ctdb_n_feat++;                                                 \
          if (patch_fcn)                                                 \
            if (!patch_fcn(ctdb,                                         \
			   patch_number,	  		         \
			   patch_mask & feature_mask,		         \
			   post_mask,				         \
			   start, end))				         \
	      return 0;						         \
        }							         \
        patch_mask = 0;						         \
        post_mask = 0;						         \
    }

    /* postx and posty are iteration variables.  They start at x0,y0 and
     * wander up to x1,y1 using a funny little algorithm.  I came
     * up with the rules for whether you move in the X direction or the
     * Y direction by inspection.  Each quadrant seems to have a very
     * different rule.  Go figure.
     * (By the way, the first thing I tried was the standard +dx -dy line
     * drawing algorithm.  Bad idea.  It refused to generate the posts
     * along the line.  I think this is simpler anyway, except for the
     * strange rules.)
     */
    postx=x0;
    postxD = postx;
    posty=y0;
    postyD = posty;

    /* Initialize the patch_mask.  All the features associated with
     * the posts we pass on our way through the patch get or'd into here.
     * We then check those features as we leave the patch.
     */
    patch_mask = 0;

    /* Initialize the post_mask.  Each box within the patch has a bit
     * in this mask.  If a bit is set, it means the corresponding box
     * was crossed.
     */
    post_mask = 0;

    /* Figure out what direction we're moving in terms of quadrants.
     * Flag funny slopes with the magic quadrant 0.
     */
    quadrant = 0;
    if (x0 == x1)
      move_toward_x = FALSE;
    else if (y0 == y1)
      move_toward_x = TRUE;
    else if (ystep > 0)
    {
	if (xstep > 0)
	  quadrant = 1;
	else
	  quadrant = 4;
    }
    else
    {
	if (xstep > 0)
	  quadrant = 2;
	else
	  quadrant = 3;
    }

    /* We cache the three posts around the current post, to avoid
     * redundant lookups.  How much this saves depends on direction
     * of traversal, but we are guaranteed to avoid at least one
     * elevation lookup each step.  Note also, that we rig our flag
     * value so that if we aren't supposed to do elevation lookups,
     * we won't.
     */
    post_value = -1;
    not_fetched = 0.0;
    zSW = zNE = zSE = not_fetched;

    /* Start by checking the first post. */
    NEW_POST;

    while ((postx!=x1) || (posty!=y1))
    {
	/* Optimization note: if-else-if chains are still faster than
	 * switch-case, even when the values look like they could drive
	 * a lookup table.  My guess is that the mips compiler doesn't
	 * generate lookup tables, because the pipelining is too messy.
	 */
	if (quadrant == 1)
	  move_toward_x = (postxD + 1.0 <= x);
	else if (quadrant == 2)
	  move_toward_x = (postyD - 1.0 - dydx < y);
	else if (quadrant == 3)
	  move_toward_x = (postyD - 1.0 < y);
	else if (quadrant == 4)
	  move_toward_x = (postxD > x);

	if (move_toward_x)
	{
	    if (xstep > 0)
	    {
		postx++;
		postxD += 1.0;
		patchx++;
		if (patchx == patch_incr)
		{
		    CHECK_PATCH;
		    patch_number++;
		    SET_NEW_PATCH_DATA;
		    patchx   = 0;
		    post_bit = post_bit >> patch_incr-1;
		}
		else
		  post_bit = post_bit << 1;

		zSW = zSE;
		zNE = zSE = not_fetched;
	    }
	    else
	    {
		postx--;
		postxD -= 1.0;
		patchx--;
		if (patchx < 0)
		{
		    CHECK_PATCH;
		    patch_number--;
		    SET_NEW_PATCH_DATA;
		    patchx = patch_incr - 1;
		    post_bit = post_bit << patch_incr-1;
		}
		else
		  post_bit = post_bit >> 1;

		zNE = GET_POST_ELEVATION(post_value, fpb);
		zSE = zSW;
		zSW = not_fetched;
	    }

	    y += yperx;
	    zatx1 += zperx1;
	    zatx2 += zperx2;

	    /* Correct for numerical accuracy */
	    if ((y > y1m) && (y < y1p))
	      y = y1;
	}
	else	/* move toward y */
	{
	    if (ystep > 0)
	    {
		posty++;
		postyD += 1.0;
		patchy++;
		if (patchy == patch_incr)
		{
		    CHECK_PATCH;
		    patch_number += patches_wide;
		    SET_NEW_PATCH_DATA;
		    patchy = 0;
		    post_bit = post_bit >> top_to_bottom;
		}
		else
		  post_bit = post_bit << patch_incr;

		zSW = GET_POST_ELEVATION(post_value, fpb);
		zSE = zNE;
		zNE = not_fetched;
	    }
	    else
	    {
		posty--;
		postyD -= 1.0;
		patchy--;
		if (patchy < 0)
		{
		    CHECK_PATCH;
		    patch_number -= patches_wide;
		    SET_NEW_PATCH_DATA;
		    patchy = patch_incr - 1;
		    post_bit = post_bit << top_to_bottom;
		}
		else
		  post_bit = post_bit >> patch_incr;

		zNE = zSE;
		zSW = zSE = not_fetched; 
	    }

	    x += xpery;
	    zaty1 += zpery1;
	    zaty2 += zpery2;

	    /* Correct for numerical accuracy */
	    if ((x > x1m) && (x < x1p))
	      x = x1;
	}

	NEW_POST;
    }
    /* Force it to check the patch we ended in. */
    CHECK_PATCH;

    return 1;
}



static char *datum_names[] = { "ED50" ,
			       "ED79" ,
			       "WGS72" ,
			       "WGS84" ,
			       "CONUSNAD27" ,
			       "NASBNAD27" ,
			       "NASANAD27" ,
			       "NAD83" ,
			       "?8?" , "?9?" , "?10?" , "?11?" , "?12?"
			     };
#define KAY (1024.0)
void 
ctdb_intern_print_description(ctdb)
    CTDB_CMP      *ctdb;
{
    int32 pages = ctdb->pages_wide * ctdb->pages_high;
    int32 i, total;
    
    printf("Database %s-%04d created %s\n",
	   ctdb->name, ctdb->version, ctdb->date);
    
    printf("Database %s curved.  ", ctdb->is_curved ? "is" : "is not");
    if ( !ctdb->is_gridded )
      printf("\nTerrain Format is PURE TIN.\n");
    else if ( !ctdb->is_tinned ) 
      printf("\nTerrain Format is PURE GRID POST.  ");
    else
      printf("\nTerrain Format is MIXED TIN&GP.  ");

    printf("Grid spacing = %d meters.  Fixed point basis = %1.8f\n",
	   ctdb->incr, ctdb->fixed_point_basis);

    printf("Origin at %0.0fN %0.0fE in UTM zone %d%c (datum %s)\n",
	   ctdb->origin_northing, ctdb->origin_easting,
	   ctdb->origin_zone_number, ctdb->origin_zone_letter,
	   datum_names[ctdb->datum]);
    
    printf("min_x = %10.2f, max_x = %10.2f\n",
	   (float)ctdb->min_x, (float)ctdb->max_x);
    printf("min_y = %10.2f, max_y = %10.2f\n",
	   (float)ctdb->min_y, (float)ctdb->max_y);
    
    printf("%d nodes (%2.0fKB), %d edges (%2.0fKB), %d abstract data \
(%2.0fKB)\n",
	   ctdb->num_nodes, ctdb->node_size / KAY,
	   ctdb->num_edges, ctdb->edge_size / KAY,
	   ctdb->num_quad_data,
	   ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_CMP) / KAY);

    if ( ctdb->is_gridded ) 
    {
	if (ctdb->elevations_resident)
	  printf("Grid posts completely in memory (%2.0fKB)\n",
		 pages * PAGE_SIZE_CMP / KAY);
	else
	  printf("%2.2f%% of grid posts in memory (%2.0fKB)\n",
		 100.0 * ctdb->elevation_cache_pages / pages,
		 ctdb->elevation_cache_pages * PAGE_SIZE_CMP / KAY);
    }

}

/* Return relevant patch water data for a patch in pw.  Returns TRUE if
 * the patch has patch water, else FALSE (if there is no patch water,
 * the contents of pw should be ignored)
 */
int32
ctdb_intern_get_patch_water(ctdb, patch_x, patch_y, pw)
    CTDB_CMP             *ctdb;
    int32                 patch_x;
    int32                 patch_y;
    CTDB_CMP_PATCH_WATER *pw;
{
    int32 patch_number;
    int32 member;
    CTDB_PATCH_GROUP_CMP *pg;
    
    patch_number = patch_y * ctdb->patches_wide + patch_x;
    pg = ctdb->patch_groups + PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);
    
    pw->patch_number = patch_number;

    switch(member)
    {
      case 0:
	if (!pg->patch0_water)
	  return FALSE;
	else
	  break;
      case 1:
	if (!pg->patch1_water)
	  return FALSE;
	else
	  break;
      case 2:
	if (!pg->patch2_water)
	  return FALSE;
	else
	  break;
      case 3:
	if (!pg->patch3_water)
	  return FALSE;
	else
	  break;
    }
    
    pw->water_char_index = pg->tables[member].water_char;
    pw->elev = pg->water_elev[member];
    return TRUE;    
}

/* Return a vector consisting of the OR of the post_mask for each land
 * microterrain (as opposed to microwater) feature in the specified patch.
 */
static int32
get_land_micro_post_mask_for_patch(ctdb, patch_number)
    CTDB_CMP *ctdb;
    int32 patch_number;
{
    CTDB_FEATURE_DATA_CMP *feature, *next, *last;
    int32 result = 0;
    int32 verts;
    int32 fclass;
    
    feature = ctdb_intern_lookup_patch(ctdb, patch_number);
    if (!feature)
      return 0;

    last = feature->hdr.size + feature;
    for(feature++;feature<last;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	if (fclass == FC_MICRO_CMP)
	{
	    next = feature + NEXT_MICRO_CMP(verts);
	    if ((feature->info.vertices & MICRO_TYPE) == MICRO_TYPE_WATER)
	      continue;
	}
	else
	{
	    /* Microterrain comes first, so when we run out, we're done. */
	    return result;
	}
	
	result |= feature->info.post_mask;
    }

    return result;
}


/* Sets each element of post array to TRUE if the corresponding post has
 * land microterrain (as opposed to microwater) on it, otherwise to FALSE.
 * The caller is responsible for allocating post_array--it must have at
 * least as many elements as there are posts in the database.
 */
void
ctdb_intern_get_land_micro_feature_present(ctdb, post_array)
    CTDB_CMP *ctdb;
    int8     *post_array;
{
    int32 num_posts;
    int32 px, py;
    int32 base_px, base_py;
    int32 patch_x, patch_y;
    int32 patch_number;
    int32 patch_mask;
    int32 post_mask;
    int32 index;
    int32 width;    
    
    num_posts = (ctdb->patches_wide + 1) *(ctdb->patches_high + 1) *
      ctdb->patch_incr * ctdb->patch_incr;
    
    width = ctdb->patches_wide * ctdb->patch_incr + 1;

    bzero(post_array, num_posts);
    
    for(patch_y=0;patch_y<ctdb->patches_high;patch_y++)
      for(patch_x=0;patch_x<ctdb->patches_wide;patch_x++)
      {
	  patch_number = patch_y * ctdb->patches_wide + patch_x;
	  base_px = patch_x * ctdb->patch_incr;
	  base_py = patch_y * ctdb->patch_incr;

	  patch_mask = get_land_micro_post_mask_for_patch(ctdb, patch_number);
	  
	  for(py=1;py<=ctdb->patch_incr;py++)
	    for(px=0;px<ctdb->patch_incr;px++)
	    {
		post_mask = POST_MASK(base_px + px, base_py + py,
				      patch_x, patch_y, ctdb->patch_incr);
		if (post_mask & patch_mask)
		{
		    index = (base_py + py) * width + (base_px + px);
		    post_array[index] = TRUE;
		}
	    }	  
      }
    
    return;
}


void
ctdb_intern_copy_mes_templates(ctdb, ctdb_out)
    CTDB_CMP *ctdb;
    CTDB_CMP *ctdb_out;
{
    int32 i;
	
    ctdb_out->num_mes_templates = ctdb->num_mes_templates;	
	
    if (!ctdb->num_mes_templates)
    {
	ctdb_out->mes_templates = NULL;
	return;
    }
    
    ctdb_out->mes_templates = (CTDB_MES_TEMPLATE_CMP *)
      ctdb_alloc(NULL,
		 sizeof(CTDB_MES_TEMPLATE_CMP) * ctdb->num_mes_templates,
		 "MES template pointers");

    for(i = 0; i < ctdb->num_mes_templates; i++)
    {
	ctdb_intern_copy_mes_template(ctdb->mes_templates + i,
				      ctdb_out->mes_templates + i);
    }
    return;
}

void
ctdb_intern_copy_mes_template(in, out)
    CTDB_MES_TEMPLATE_CMP *in;
    CTDB_MES_TEMPLATE_CMP *out;
{    
    int32 j;
    CTDB_MES_ENCLOSURE_CMP *encl_in;
    CTDB_MES_ENCLOSURE_CMP *encl_out;

    bcopy(in, out, sizeof(CTDB_MES_TEMPLATE_CMP));

    /* Copy over the vertices for this MES template */
    out->all_vertices =
      (int32 (*)[3])ctdb_alloc(NULL,
			       sizeof(int32)*XYZ*in->num_vertices,
			       "MES vertices");

    bcopy(in->all_vertices,
	  out->all_vertices,
	  sizeof(int32)*XYZ*in->num_vertices);

    /* Copy over the triangles for this MES template */
    out->all_triangles =
      (CTDB_MES_TRIANGLE_CMP*)ctdb_alloc(NULL,
					 sizeof(CTDB_MES_TRIANGLE_CMP)*
					 in->num_triangles,
					 "MES triangle");

    bcopy(in->all_triangles,
	  out->all_triangles,
	  sizeof(CTDB_MES_TRIANGLE_CMP)*in->num_triangles);

    /* Copy over the enclosures for this MES template */
    out->all_enclosures =
      (CTDB_MES_ENCLOSURE_CMP*)ctdb_alloc(NULL,
					  sizeof(CTDB_MES_ENCLOSURE_CMP)*
					  in->num_enclosures,
					  "MES enclosure");

    bcopy(in->all_enclosures, 
	  out->all_enclosures,
	  sizeof(CTDB_MES_ENCLOSURE_CMP)*in->num_enclosures);	    

    for(j = 0; j < in->num_enclosures; j++)
    {
	encl_in = in->all_enclosures + j;
	encl_out = out->all_enclosures + j;

	encl_out->aperture_ids = 
	  (uint16*)ctdb_alloc(NULL,
			     sizeof(uint16)*
			     encl_in->num_apertures,
			     "MES aperture for the enclosures");
		
	bcopy(encl_in->aperture_ids,
	      encl_out->aperture_ids,
	      sizeof(uint16)*encl_in->num_apertures);


	encl_out->footprint.vertex_ids =
	  (int32*)ctdb_alloc(NULL,
			     sizeof(int32)*
			     encl_in->footprint.num_vertices,
			     "MES floor outline for this enclosure");
		
	bcopy(encl_in->footprint.vertex_ids,
	      encl_out->footprint.vertex_ids,
	      sizeof(int32)*encl_in->footprint.num_vertices);
    }


    /* Copy over the apertures */
    out->all_apertures =
      (CTDB_MES_APERTURE_CMP*)ctdb_alloc(NULL, 
					 sizeof(CTDB_MES_APERTURE_CMP)*
					 in->num_apertures,
					 "MES apertures");

    bcopy(in->all_apertures, 
	  out->all_apertures,
	  sizeof(CTDB_MES_APERTURE_CMP)*in->num_apertures);

    for(j = 0; j < in->num_apertures; j++)
    {
	out->all_apertures[j].outline.vertex_ids = NULL;

	out->all_apertures[j].outline.vertex_ids =
	  (int32*)
	  ctdb_alloc(NULL,
		     sizeof(int32)*
		     in->all_apertures[j].outline.num_vertices,
		     "MES aperture vertices");

	bcopy(in->all_apertures[j].outline.vertex_ids,
	      out->all_apertures[j].outline.vertex_ids,
	      sizeof(int32)*
	      in->all_apertures[j].outline.num_vertices);
    }

    /* Copy over the boxes */
    out->all_boxes =
      (CTDB_MES_GRID_BOX_CMP*)ctdb_alloc(NULL,
					 sizeof(CTDB_MES_GRID_BOX_CMP)*
					 in->num_boxes,
					 "boxes");

    bcopy(in->all_boxes,
	  out->all_boxes,
	  sizeof(CTDB_MES_GRID_BOX_CMP)*
	  in->num_boxes);

    for(j = 0; j < in->num_boxes; j++)
    {
	out->all_boxes[j].box_ids = NULL;
	out->all_boxes[j].aperture_ids = NULL;
	out->all_boxes[j].wall_triangle_ids = NULL;

	if(in->all_boxes[j].dimensions_x*
	   in->all_boxes[j].dimensions_y*
	   in->all_boxes[j].dimensions_z > 1)
	{
	    out->all_boxes[j].box_ids =
	      (int32*)
	      ctdb_alloc(NULL,
			 sizeof(int32)*
			 (in->all_boxes[j].dimensions_x*
			  in->all_boxes[j].dimensions_y*
			  in->all_boxes[j].dimensions_z),
			 "box inner boxes");

	    bcopy(in->all_boxes[j].box_ids,
		  out->all_boxes[j].box_ids,
		  sizeof(int32)*
		  (in->all_boxes[j].dimensions_x*
		   in->all_boxes[j].dimensions_y*
		   in->all_boxes[j].dimensions_z));
	}

	if (in->all_boxes[j].num_apertures)
	{
	    out->all_boxes[j].aperture_ids =
	      (uint16*)ctdb_alloc(NULL,
				 sizeof(uint16)*
				 in->all_boxes[j].num_apertures,
				 "box apertures");
		
	    bcopy(in->all_boxes[j].aperture_ids,
		  out->all_boxes[j].aperture_ids,
		  sizeof(uint16)*in->all_boxes[j].num_apertures);
	}
	else
	{
	    out->all_boxes[j].aperture_ids = NULL;
	}
	    
	if (in->all_boxes[j].num_triangles)
	{
	    out->all_boxes[j].wall_triangle_ids =
	      (int32*)ctdb_alloc(NULL,
				 sizeof(int32)*
				 in->all_boxes[j].num_triangles,
				 "box apertures");
		
	    bcopy(in->all_boxes[j].wall_triangle_ids,
		  out->all_boxes[j].wall_triangle_ids,
		  sizeof(int32)*in->all_boxes[j].num_triangles);
	}
	else
	{
	    out->all_boxes[j].wall_triangle_ids = NULL;
	}
    }
    
    return;
}

void
ctdb_intern_copy_mes_volumes(ctdb, ctdb_out)
    CTDB_CMP *ctdb;
    CTDB_CMP *ctdb_out;
{
    int32 i;

    ctdb_out->num_mes_volumes = ctdb->num_mes_volumes;

    if (!ctdb->num_mes_volumes)
    {
	ctdb_out->mes_volumes = NULL;
	return;
    }
    
    ctdb_out->mes_volumes = (CTDB_MES_VOLUME_CMP *)
      ctdb_alloc(NULL, 
		 ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME_CMP),
		 "MES volumes");
    
    for(i = 0; i < ctdb->num_mes_volumes; i++)
    {
	ctdb_intern_copy_mes_volume(ctdb->mes_volumes + i,
				      ctdb_out->mes_volumes + i);
    }
    bcopy(ctdb->mes_volumes, ctdb_out->mes_volumes,
	  ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME_CMP));
    
    /* Now, we just need to reset the template pointers to point to the new
     * templates, and we're done
     */
    for(i=0;i<ctdb->num_mes_volumes;i++)
    {
	ctdb_out->mes_volumes[i].template = 
	  ctdb_out->mes_templates + 
	  (ctdb->mes_volumes[i].template - ctdb->mes_templates);
    }
    
    return;    
}

void
ctdb_intern_copy_mes_volume(in, out)
    CTDB_MES_VOLUME_CMP *in;
    CTDB_MES_VOLUME_CMP *out;
{    
    bcopy(in, out, sizeof(CTDB_MES_VOLUME_CMP));

    /* Copy over the roofline vertices for this MES template */
    out->roofline.vertices = (CTDB_MES_VERTEX_F64_CMP*)
      ctdb_alloc(NULL,
		 sizeof(CTDB_MES_VERTEX_F64_CMP) * in->roofline.num_vertices,
		 "MES roofline vertices");
    
    bcopy(in->roofline.vertices,
	  out->roofline.vertices,
	  sizeof(CTDB_MES_VERTEX_F64_CMP) * in->roofline.num_vertices);

     /* Copy over the floorline vertices for this MES template */
    out->floorline.vertices = (CTDB_MES_VERTEX_F64_CMP*)
      ctdb_alloc(NULL,
		 sizeof(CTDB_MES_VERTEX_F64_CMP) * in->floorline.num_vertices,
		 "MES floorline vertices");
    
    bcopy(in->floorline.vertices,
	  out->floorline.vertices,
	  sizeof(CTDB_MES_VERTEX_F64_CMP) * in->floorline.num_vertices);

 }


/***************************************************************************
 * COMPILER REPRESENTATION CLEANUP ROUTINES                                *
 ***************************************************************************/

void
ctdb_free_cmp_db(ctdb)
    CTDB_CMP *ctdb;
{
    int32 groups, i, j;
    CTDB_PATCH_GROUP_CMP *pg;

#define CTDB_FREE_CMP(x)    \
    if (ctdb->x)           \
        STDDEALLOC(ctdb->x)

    if (!ctdb)
      return;   
    
    CTDB_FREE_CMP(soil_tables);
    CTDB_FREE_CMP(soil_table_storage);
    CTDB_FREE_CMP(water_chars);
    CTDB_FREE_CMP(tidal_zones);

    /* $$$ This is usually mmapped.  Should we unmap it???
     * CTDB_FREE_CMP(features);
     */
    CTDB_FREE_CMP(elevation_cache_page_numbers);
    CTDB_FREE_CMP(elevation_pages);
    CTDB_FREE_CMP(elevations);
    CTDB_FREE_CMP(soil_cache_page_numbers);
    CTDB_FREE_CMP(soil_pages);
    CTDB_FREE_CMP(soils);
    CTDB_FREE_CMP(quad_data);
    CTDB_FREE_CMP(nodes);
    CTDB_FREE_CMP(node_storage);
    CTDB_FREE_CMP(edges);
    CTDB_FREE_CMP(edge_storage);
    CTDB_FREE_CMP(linear_models);
    CTDB_FREE_CMP(terrain_elements);
    
    for(i = 0; i < ctdb->num_volume_models; i++)
      CTDB_FREE_CMP(volume_models[i]);
    CTDB_FREE_CMP(volume_models);

    for(i = 0; i < ctdb->num_aggregate_models; i++)
      CTDB_FREE_CMP(aggregate_models[i]);
    CTDB_FREE_CMP(aggregate_models);
    
    /* Each patch group stores a pointer to its own virtual grid data.  We
     * need to go through and get rid of them one by one.
     */
    
    groups = ctdb->patches_wide * ctdb->patches_high /
      CTDB_PATCHES_PER_GROUP_CMP;
    pg = ctdb->patch_groups;
    
    for(i=0;i<groups;i++, pg++)
    {
	    for(j=0;j<CTDB_PATCHES_PER_GROUP_CMP;j++)
	    {
		if (pg->te_vgrid_offset[j])
		  STDDEALLOC(pg->te_vgrid_offset[j]);
	    }
    }

    ctdb_mes_free_cmp(ctdb);

    STDDEALLOC(ctdb->patch_groups);
    STDDEALLOC(ctdb);
    return;
}


/* MES Coordinate Conversion */



/* ctdb_gcsvec_to_mesvec
 *
 * This function will take a vector defined in GCS, and convert it 
 * to the local coordinate system defined by the mes.
 */
static int32
ctdb_intern_gcsvec_to_mesvec(
			     CTDB_MES_VOLUME_CMP *mes,
			     int32               cell,
			     float64             gcsvec[XYZ],
			     float64             mesvec[XYZ]
			     )
{
    float64            tmp_vec[XYZ];
    CTDB_MES_TEMPLATE_CMP *template;

    if (!mes)
      return 0;

    template = mes->template;
    
    if (cell != mes->origin[CELL3D])
      gcs_vec_to_vec(cell, gcsvec, (int32) mes->origin[CELL3D], tmp_vec);
    else
      VMAT3_VEC_COPY64(gcsvec, tmp_vec);

    vmat3_scal_vec_mul64(1/template->fixed_point_basis, tmp_vec, tmp_vec);
    vmat3_mat_vec_mul64(mes->tfm_world_to_mes, tmp_vec, mesvec);
    return 1;
}

/* ctdb_mesvec_to_gcsvec
 *
 * This function takes a vector defined by the mes's local coordinate
 * system and transforms it to a gcs vector in its native cell.
 * The native cell is also returned.
 */
static int32
ctdb_intern_mesvec_to_gcsvec(
			     CTDB_MES_VOLUME_CMP *mes,
			     float64              mesvec[XYZ],
			     int32               *cell,
			     float64              gcsvec[XYZ]
			     )
{        
    float64            tmp[3][3];
    float64            tmp_vec[XYZ];
    CTDB_MES_TEMPLATE_CMP *template;
    float64            inv_matrix[3][3];

    if (!mes)
      return 0;
    template = mes->template;

    VMAT3_VEC_COPY64(mesvec, tmp_vec);
    vmat3_scal_vec_mul64(template->fixed_point_basis, tmp_vec, tmp_vec);
    vmat3_inverse64(mes->tfm_world_to_mes, inv_matrix);
    vmat3_mat_vec_mul64(inv_matrix, tmp_vec, gcsvec);
    *cell = (int32) mes->origin[CELL3D];
}   


/* ctdb_gcs_to_mes
 *
 * This function will take a point in GCS (cell, x0, y0, z0)  
 * and convert it to the local coordinate system defined by 
 * the mes (x1, y1, z1).
 * 
 * If the gcs point lies outside the given mes, return 0.
 * Otherwise, we return 1.
 */
static int32
ctdb_intern_gcs_to_mes(
		       CTDB_MES_VOLUME_CMP *mes,
		       int32                cell,
		       float64              x0,
		       float64              y0,
		       float64              z0,
		       float64             *x1,
		       float64             *y1,
		       float64             *z1
		       )
{
    float64 offset[XYZ], tmp[XYZ];
    int32 mes_cell;
    CTDB_MES_TEMPLATE_CMP *templ;

    if (!mes)
      return 0;

    mes_cell = mes->origin[CELL3D];
    templ = mes->template;

    if (cell != mes_cell)
    {
	gcs_gcs3e_to_gcscs(cell, x0, y0, z0, mes_cell,
			 &offset[X], &offset[Y], &offset[Z]);
	offset[X] -= mes->origin[X];
	offset[Y] -= mes->origin[Y];
	offset[Z] -= mes->origin[Z];
    }
    else
    {
	offset[X] = x0 - mes->origin[X];
	offset[Y] = y0 - mes->origin[Y];
	offset[Z] = z0 - mes->origin[Z];
    }
    ctdb_intern_gcsvec_to_mesvec(mes, (int32) mes->origin[CELL3D],
				 offset, tmp);

    *x1 = tmp[X];
    *y1 = tmp[Y];
    *z1 = tmp[Z];
    if (((tmp[X] >= 
	  templ->all_boxes[CTDB_MES_EXTERIOR_CMP].left_front_bottom_corner[X]) &&
	 (tmp[X] <= 
	  templ->all_boxes[CTDB_MES_EXTERIOR_CMP].right_back_top_corner[X])) &&
	((tmp[Y] >= 
	  templ->all_boxes[CTDB_MES_EXTERIOR_CMP].left_front_bottom_corner[Y]) &&
	 (tmp[Y] <= 
	  templ->all_boxes[CTDB_MES_EXTERIOR_CMP].right_back_top_corner[Y])) &&
	((tmp[Z] >= 
	  templ->all_boxes[CTDB_MES_EXTERIOR_CMP].left_front_bottom_corner[Z]) &&
	 (tmp[Z] <= 
	  templ->all_boxes[CTDB_MES_EXTERIOR_CMP].right_back_top_corner[Z])))
      return 1;
    return 0;
}


/* ctdb_mes_to_gcs
 *
 * This function will take a defined by the mes's local coordinate 
 * system (x0, y0, z0), and converts it to a GCS point (x1, y1, z1)
 * in a specified cell.
 */
int32
ctdb_intern_mes_to_gcscs(
			 CTDB_MES_VOLUME_CMP *mes,
			 int32                x0,
			 int32                y0,
			 int32                z0,
			 int32                cell,
			 float64             *x1,
			 float64             *y1,
			 float64             *z1
			 )
{
    float64          origin[XYZ];
    float64          mesvec[XYZ], gcsvec[XYZ];
    int32            gcscell;
    
    
    if (!mes)
      return 0;

    mesvec[X] = (float64) x0;
    mesvec[Y] = (float64) y0;
    mesvec[Z] = (float64) z0;

    ctdb_intern_mesvec_to_gcsvec(mes, mesvec, &gcscell, gcsvec);

    if ((int32) mes->origin[CELL3D] != cell)
      gcs_gcs3_to_gcscs((int32) mes->origin[CELL3D], mes->origin,
			cell, origin);
    else
    {
	origin[X] = mes->origin[X];
	origin[Y] = mes->origin[Y];
	origin[Z] = mes->origin[Z];
    }
    
    if (gcscell != cell)
      gcs_gcs3_to_gcscs(gcscell, gcsvec, cell, gcsvec);

    *x1 = origin[X] + gcsvec[X];
    *y1 = origin[Y] + gcsvec[Y];
    *z1 = origin[Z] + gcsvec[Z];
    return 1;
}



/* Assorted MES functions */
int32
ctdb_intern_mes_get_aperture_centroid(
				      CTDB_CMP              *ctdb,
				      uint32             aperture,
				      float64           *x,
				      float64           *y,
				      float64           *z
				      )
{
    uint16             mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID_CMP(aperture);
    uint16             aperture_id =CTDB_MES_EDGE_KEY_TO_APERTURE_ID_CMP(aperture);
    CTDB_MES_TEMPLATE_CMP *template;
    CTDB_MES_APERTURE_CMP *apert;
    int32 i;
    float64 center_x = 0;
    float64 center_y = 0;
    float64 center_z = 0;
    float64 gcs_x, gcs_y, gcs_z;
    int32 vertex_id;

    if (!ctdb || !CTDB_MES_VALID_MES_ID_CMP(ctdb, mes_id) ||
	!CTDB_MES_VALID_APERTURE_ID_CMP(ctdb, mes_id, aperture_id))
      return 0;

    template = ctdb->mes_volumes[mes_id].template;
    apert = &(template->all_apertures[aperture_id]);

    for (i=0; i < apert->outline.num_vertices; i++)
    {
	vertex_id = apert->outline.vertex_ids[i];
	
	ctdb_intern_mes_to_gcscs(&(ctdb->mes_volumes[mes_id]),
				 template->all_vertices[vertex_id][X],
				 template->all_vertices[vertex_id][Y],
				 template->all_vertices[vertex_id][Z],
				 ctdb_intern_get_cell_id(ctdb),
				 &gcs_x, &gcs_y, &gcs_z);
	center_x += gcs_x;
	center_y += gcs_y;
	center_z += gcs_z;
    }
    *x = center_x / apert->outline.num_vertices;
    *y = center_y / apert->outline.num_vertices;
    *z = center_z / apert->outline.num_vertices;
    return 1;
}

#define CLOSE_ENOUGH 0.2

static int32
point_on_line(float64 pt[XY], 
	      float64 start[XY],
	      float64 end[XY])
{
    float64 int_pt[XY];
    float64 distance_sq;

    distance_sq = geo_distance_sq_point_to_segment(start, end, pt, int_pt);
    if (distance_sq > CLOSE_ENOUGH)
      return 0;
    return 1;
}
	      

int32
point_in_polygon(float64 pt[XY],
		 float64 polygon[][XY],
		 int32   num_pts_in_poly)
{
    int32 i;

  /* Check if point is inside polygon */
    if (geo_is_point_in_convex_polygon(pt, polygon, num_pts_in_poly))
      return TRUE;

  /* Check if point lies on the boundary */
    for (i=0; i < num_pts_in_poly; i++)
      if (point_on_line(pt, polygon[i], polygon[(i+1)%num_pts_in_poly]))
	return TRUE;

    /* Otherwise, return false */
    return FALSE;
}



