/*
 * $RCSfile: ct_cmprep.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: ct_cmprep.h                                                       *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations for input-format-independent compiler      *
 *                   representation, and conversions to "real" ctdb formats. *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *       Remarks:  This file contains the declarations for use by the        *
 *                 recompiler and compiler backend.                          *
 *                                                                           *
 ****************************************************************************/
#ifndef _CMPREP_H_INCLUDED
#define _CMPREP_H_INCLUDED

#include <stdtypes.h>
#include "../libctdb_local.h"

typedef struct ctdb_cmp_patch_water
{
    int32   patch_number;
    int32   water_char_index;
    float32 elev;
} CTDB_CMP_PATCH_WATER;


/***************************************************************************
 * COMPILER REPRESENTATION                                                 *
 ***************************************************************************/

#define CMP_FORMAT (CTDB_FILE_FORMAT + 1)/* Unofficial format number for
					 * compiler rep--always greater
					 * than highest supported format
					 */

#define CTDB_PATCHES_PER_GROUP_CMP      4  /* Num of patches per patch group 
					    */

#define CTDB_MAX_FEATURE_PRESENT_PER_PG_CMP 60 

#define CTDB_MAX_POSTS_PER_PATCH_CMP    25 /* maximum number of posts per
					    * patch
					    */

#define CTDB_NAME_LENGTH_CMP            14   /* same as Simnet */
#define CTDB_DATE_LENGTH_CMP            28   /* Enough room for ctime() */

#define CTDB_PATCH_CACHES_CMP           16   /* Number of patch data caches */

#define CTDB_FILENAME_DIMENSION_CMP 80

/* Posts are tiled in square pages, as follows:
 */
#define PAGE_SIZE_CMP      4096
#define SQRT_PAGE_CMP      64
#define POST_SIZE_CMP      sizeof(int32)
#define SQRT_POST_CMP      2
#define POSTS_PER_PAGE_CMP (PAGE_SIZE/POST_SIZE)
#define POSTS_PER_SIDE_CMP (SQRT_PAGE/SQRT_POST)
#define SIDE_SHIFT_CMP     5 /* log2(POSTS_PER_SIDE) */
#define SIDE_MASK_CMP      ((1<<SIDE_SHIFT)-1)

/* SW soil is bits 0-15 */
#define GET_POST_SW_SOIL_CMP(p) ((p) & 0xFFFF)
#define BUILD_POST_SW_SOIL_CMP(s) ((s) & 0xFFFF)

/* SE soil is bits 0-15 */
#define GET_POST_SE_SOIL_CMP(p) ((p) >> 24 & 0xFFFF)
#define BUILD_POST_SE_SOIL_CMP(s) ((s) & 0xFFFF)

/* NE soil is bits 16-31 */
#define GET_POST_NE_SOIL_CMP(p) (((p) >> 16) & 0xFFFF)
#define BUILD_POST_NE_SOIL_CMP(s) (((s) & 0xFFFF) << 16)

/* NW soil is bits 16-31 */
#define GET_POST_NW_SOIL_CMP(p) (((p) >> 16) & 0xFFFF)
#define BUILD_POST_NW_SOIL_CMP(s) (((s) & 0xFFFF) << 16)


#define CTDB_SOIL_TYPE_WATER_CMP 5

/* The feature classes each imply certain intervisibility semantics
 * and storage techniques:
 *
 * FC_LINEAR is an individual tree or a treeline (depending upon the
 * number of vertices), or any non-laid linear feature (fences, wire,
 * dragon's teeth).
 *
 * An individual tree is described
 *	FEAT_INFO LINEAR_INFO XY Z
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature; XY identifies the location of the tree
 * within the patch; Z identifies the elevation of the top of the
 * tree.  The reduction of visibility due to tree blockage is done
 * using a simple light transmittance model.
 *
 * A treeline (or linear feature) is described  
 *	FEAT_INFO LINEAR_INFO XY Z XY Z XY Z...
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature and whether or not there are trees at each end
 * of a treeline; (note that we do not need to store the end trees as
 * individuals since they are implied by the treeline); each XY point
 * identifies the location of one vertex in the treeline; the corresponding Z
 * identifies the elevation of the top of the treeline at that vertex.
 *
 * FC_VOLUME is a building or a building-like structure which blocks
 * intervisibility.  Telephone poles, water towers, and other such
 * structures which do not interrupt intervisibility are omitted.
 *
 * A volume is described
 *	FEAT_INFO VOLUME_INFO XY Z XY Z ...
 * Each XY point identifies the location of one vertex in the bounding
 * box of the building; the corresponding Z identifies the elevation
 * of the roof line of the building at that point.  Intervisibility is
 * interrupted by the line segments connecting these points exactly the
 * same way as the segments connecting grid posts.  The VOLUME_INFO
 * contains the type of feature, a reference to a model set, and the damage
 * state.
 *
 * FC_MICRO is a microterrain triangle.
 *
 * microterrain is described
 *	INFO MICRO_SOIL XY Z MICRO
 * Microwater is identical to microterrain, except that it has an MWATER_CHARS
 * at the end:  
 *      INFO MICRO_SOIL XY Z MICRO MWATER_CHARS 
 * The XY and Z combination identify the location of one vertex of
 * the microterrain triangle.  Up to three subsequent vertices are listed
 * in the MICRO structure.  There are edges connecting this point to the
 * each of the other listed vertices.  If there are two vertices listed
 * the two edges from this point form a triangle.  If there are three
 * listed, the three edges form two triangles (the first edge being
 * shared between the two).  The MICRO_SOIL structure identifies either
 * one or two soil types according to the number of vertices listed. 
 * For microterrain, intervisibility is interrupted by the edges connecting
 * these points exactly the same way as the segments connecting grid posts.
 *
 * FC_CANOPY is a canopy polygon.  See-through canopies are omitted.
 *
 * Canopies are described
 *      INFO XY Z CANOPY
 * Canopies are basically microterrain, but store no soil information, and
 * instead store bits specifying whether each edge is an exterior edge
 * and a model reference.
 * For canopies, visibility is degraded as with treelines.  
 *
 * FC_LAID_LINEAR is a road or river segment.
 *
 * Road segments and river segments are described the same way
 *	INFO LL_SOIL LAID_LINEAR XY XY...
 * LL_SOIL stores the soil type of the road or river segment.  LAID_LINEAR
 * identifies the width at the beginning and end of the segment
 * (linear taper is assumed).  It also stores an edge key for the
 * segment which provides the low-order 12 bits of the topologic edge 
 * (of type CTDB_EDGE) which uses this linear feature.  Each XY point
 * identifies the location of one vertex in the segment.  Linear
 * features do not impact intervisibility.
 */

#define FC_CLASS_MIN_CMP     0
#define FC_MICRO_CMP         0
#define FC_VOLUME_CMP        1
#define FC_LINEAR_CMP        2
#define FC_CANOPY_CMP        3
#define FC_LAID_LINEAR_CMP   4
#define FC_AGGREGATE_CMP     5
#define FC_CLASS_MAX_CMP     6

#define M_FEATURE_CLASS_CMP   0xE0000000
#define S_FEATURE_CLASS_CMP   29
#define M_VERTICES_CMP        0x1E000000
#define M_MICRO_VERTICES_CMP  0x06000000
#define S_VERTICES_CMP        25

#define MICRO_VERTS_CMP         0x3
#define MICRO_TYPE_CMP          0xC
#define MICRO_TYPE_BASE_CMP     0x0
#define MICRO_TYPE_ML_CMP       0x4
#define MICRO_TYPE_DEFAULT_CMP  0x8
#define MICRO_TYPE_WATER_CMP    0xC

#define NEXT_MICRO_CMP(v)  \
    (( ((v)&MICRO_TYPE_CMP) == MICRO_TYPE_WATER_CMP)?6:5)
#define NEXT_VOLUME_CMP(v)      (3 + ( (v) << 1) )
#define NEXT_CANOPY_CMP(v)      (4)
#define NEXT_LINEAR_CMP(v)      (2 + ( (v) << 1) )
#define NEXT_LAID_LINEAR_CMP(v) (3 + ( (v) ) )
#define NEXT_AGGREGATE_CMP(v)   (3)


typedef struct ctdb_feature_info_cmp
{
    /* One of the FC_x values listed above.
     */
    unsigned feature_class : 3;

    /* For FC_LINEAR, FC_VOLUME, FC_CANOPY, and FC_LAID_LINEAR this
     * is the number of vertices.  If the actual objects has more than
     * 16 vertices, it must be broken into two or more objects.  For
     * FC_MICRO this is the number of positions which are connected to
     * this point bitwise-OR'd with a microterrain type.
     */
    unsigned vertices : 4;

    /* This is used as a high resolution bitwise bounding volume test.
     * Each post in a patch has a bit.  This mask identifies which
     * posts' polygons were touched by this feature.  Although Simnet
     * terrain databases only have 16 posts per patch, CCTT database
     * may have 25 (given the 100m interval required in the spec).
     */
    unsigned post_mask : 25;
} CTDB_FEATURE_INFO_CMP;

#define CTDB_TREE_HEIGHT_BASIS_CMP (200.0 / (1<<14))

/* Number of bits used to define fullness.
 */
#define CTDB_FULLNESS_BITS_CMP  8

#define M_TREE_FIRST_CMP 0x8000
#define M_TREE_LAST_CMP  0x4000

typedef struct ctdb_volume_info_cmp
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.
     */
    unsigned reference    : 16;
    unsigned type         : 5;
    unsigned avoid        : 1;
    unsigned force_id     : 2;
    unsigned reserved     : 8;
} CTDB_VOLUME_INFO_CMP;


typedef struct ctdb_linear_info_cmp
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.  `first' and `last' are only used if
     * the feature is a treeline, and specify whether or not individual
     * trees should be placed at each end.
     */
    unsigned reference        : 16;
    unsigned first            : 1;
    unsigned last             : 1;
    unsigned reserved         : 14;
} CTDB_LINEAR_INFO_CMP;


typedef struct ctdb_aggregate_info_cmp
{
      unsigned reference        : 16;
      unsigned reserved         : 16;
} CTDB_AGGREGATE_INFO_CMP;

/* For now we store feature elevations as simple floats, since there
 * is no obvious way to use the extra 13 bits if we were to instead
 * store it using the fixed point convention used for grid posts.
 * Floating point will, of course, be faster than storing them in
 * fixed point since only a SP->DP conversion will be necessary to use
 * the data.
 */

typedef struct ctdb_micro_pts_cmp
{
    /* Each microterrain vertex is connected to 0,1,2, or 3 other
     * microterrain vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, vert0 will 
     * contain the index of that vertex relative to patch start; vert1,
     * vert2 will be 0.  If it is connected to 2 others, then vert0 
     * and vert1 will contain the indices of those vertices relative 
     * to patch start and so on.
     */
    uint8 reserved; 
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
} CTDB_MICRO_PTS_CMP;

/* For use with FC_MICRO.  We'll use the per-patch soil table like we do for
 * TINs.  Will be used for variable feature attributes in a future AFA 
 * integration. 
 */
typedef struct ctdb_var_attr_two_soils_cmp
{
    unsigned reserved          : 16; /* Reserved for use by advanced features 
				      * & attributes
				      */
    unsigned poly_char_index0  : 8; 
    unsigned poly_char_index1  : 8;
} CTDB_VAR_ATTR_TWO_SOILS_CMP;

/* Microwater characteristic references */
typedef struct ctdb_mwater_chars_cmp
{
    unsigned char0 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
    unsigned char1 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
} CTDB_MWATER_CHARS_CMP;

typedef struct ctdb_canopy_pts_cmp
{
    /* Each canopy vertex is connected to 0,1,2, or 3 other
     * canopy vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then vert0 will 
     * contain the index of that vertex relative to patch start; vert1, 
     * vert2 will be 0.  If it is connected to 2 others, then vert0 
     * and vert1 will contain the indices of those vertices relative 
     * to patch start and so on.
     */
    unsigned can_edge_ref : 5;
    unsigned vert2_outer  : 1;
    unsigned vert1_outer  : 1;
    unsigned vert0_outer  : 1;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
} CTDB_CANOPY_PTS_CMP;

typedef struct ctdb_laid_linear_info_cmp
{
    /* Each linear segment has a width.  The width is in the same
     * units as x,y in the XY_PATCH_OFFSET structure.  The edge_key
     * provides the low-order 12 bits of the topologic edge 
     * (of type CTDB_EDGE) which uses this linear feature.
     */
    uint16    width;
    unsigned  reserved : 4;
#define CTDB_EDGE_KEY_CMP 4096 /* 2^12 */
    unsigned  edge_key : 12;
} CTDB_LAID_LINEAR_INFO_CMP;

/* For use with laid_linears.  We will use all 16 bits for a single index
 * to avoid the indirection.  Will be used for variable feature attributes
 * in a future AFA integration. 
 */
typedef struct ctdb_var_attr_one_soil_cmp
{
    unsigned var_attr_index    : 16;
    unsigned poly_char_index   : 8;
    unsigned reserved          : 8;
} CTDB_VAR_ATTR_ONE_SOIL_CMP;

typedef struct ctdb_patch_hdr_cmp
{
    /* At the beginning of each patch, we store the size (the number
     * of CTDB_FEATURE_DATA things in the patch).  There will never be
     * more than 2^16 of those, so that leaves 16 bits of padding.
     * I'm sure we'll find a use for those bits eventually.
     */
    uint16 size;
    uint16 dynamic_patch_id;
} CTDB_PATCH_HDR_CMP;

#define CTDB_ABSTRACT_CANOPY_CMP                1
#define CTDB_ABSTRACT_SOIL_DEFRAG_CMP           2
#define CTDB_ABSTRACT_STEEP_SLOPE_CMP           3
#define CTDB_ABSTRACT_RAILROAD_CMP              4
#define CTDB_ABSTRACT_PIPELINE_CMP              5
#define CTDB_ABSTRACT_POLITICAL_BOUNDARY_CMP    6
#define CTDB_ABSTRACT_POWERLINE_CMP             7
#define CTDB_ABSTRACT_LABEL_CMP                 8
#define CTDB_ABSTRACT_TACTICAL_SIGN_CMP         9
#define CTDB_ABSTRACT_OFFROAD_SEGMENT_CMP       10
#define CTDB_ABSTRACT_MAX_CMP /* Always Last */ 11


/* CTDB_ABSTRACT_FEATURE_INFO:
 *
 * Structure at the start of each abstract feature.
 */
#define CTDB_ABSTRACT_QUAD_NODE_CMP             0
typedef struct ctdb_abstract_feature_info_cmp
{
    /* Type code for the abstract feature.  The code 0 is reserved
     * to indicate that this "feature" is actually a reference to
     * four child quad nodes.
     */
    uint8 type_code;

    /* The size (in 4-byte units) of the CTDB_ABSTRACT_DATA associated
     * with the abstract feature.
     */
    uint8 data_size;

    /* The number of vertices in feature.  There must be at least one,
     * except for "features" which are of the type CTDB_ABSTRACT_QUAD_NODE.
     */
    uint16 num_vertices;

} CTDB_ABSTRACT_FEATURE_INFO_CMP;

/* CTDB_XYO_COORD2D
 * 
 * Structure used to represent a vertex point (x,y) in XYO coordinates 
 * in either a feature or terrain element data structure.
 */
typedef struct ctdb_xyo_coord2d_cmp  
{
    uint16 x;
    uint16 y;
} CTDB_XYO_COORD2D_CMP;

/* CTDB_ABSTRACT_VERTEX:
 *
 * Structure used to represent a vertex of an abstract feature.
 * Note that for efficient access, the format used for vertices
 * is 32 bit floating point, not fix-point patch offsets.  This
 * results in a loss of precision at the far corners of the database,
 * but avoids a great deal of translation at run time.
 */
typedef float32 CTDB_ABSTRACT_VERTEX_CMP;

/* CTDB_ABSTRACT_DATA:
 *
 * Format of data associated with so-called abstract features.  Note
 * that this structure is analogous to a protocol definition, in that
 * the final record of each union substructure can be variable length.
 *
 * This goes into a 4-byte union: NO MEMBER STRUCTURE CAN EXCEED 32 BITS!
 *
 */
typedef union ctdb_abstract_data_cmp
{
    struct ctdb_quad_data_cmp
    {
	/* Array of 4 offsets to child quad nodes (in 4-byte units)
	 * from the beginning of the quad tree data storage.
	 */
	uint32 child_start[1];  /* really 4 members, not 1 */
    } quad_data;
    struct ctdb_canopy_data_cmp
    {
	unsigned impenetrable  : 1;
	unsigned padding       : 31;
    } canopy;
    struct ctdb_soil_defrag_data_cmp
    {
	unsigned soil_index    : 4;
	unsigned level         : 4;
	unsigned padding       : 8;
	unsigned soilmap_index  : 16;
    } soil_defrag;
    struct ctdb_steep_slope_data_cmp
    {
	unsigned slope_degrees : 7;
	unsigned padding       : 25;
    } steep_slope;
    struct ctdb_label_data_cmp
    {
	char text[4]; /* variable length, NULL terminated */
    } label;
    struct ctdb_offroad_data_cmp
    {
	uint32 edge_number;
    } offroad;

} CTDB_ABSTRACT_DATA_CMP;

typedef union ctdb_feature_data_cmp
{
    uint32                         integer;
    CTDB_PATCH_HDR_CMP             hdr;
    CTDB_FEATURE_INFO_CMP          info;
    CTDB_VOLUME_INFO_CMP           volume;
    CTDB_LINEAR_INFO_CMP           linear;
    CTDB_LAID_LINEAR_INFO_CMP      laid_linear;
    CTDB_VAR_ATTR_ONE_SOIL_CMP     ll_soil;
    CTDB_AGGREGATE_INFO_CMP        aggregate;
    CTDB_XYO_COORD2D_CMP           xy;
    float32                        z;
    CTDB_MICRO_PTS_CMP             micro;
    CTDB_CANOPY_PTS_CMP            canopy;
    CTDB_VAR_ATTR_TWO_SOILS_CMP    micro_soil;
    CTDB_ABSTRACT_FEATURE_INFO_CMP abstract;
    CTDB_ABSTRACT_DATA_CMP         abstract_data;
    CTDB_ABSTRACT_VERTEX_CMP       abstract_vertex;
    CTDB_MWATER_CHARS_CMP          mwater;
} CTDB_FEATURE_DATA_CMP;

/* Definitions of feature types.
 */

/* Types of volume features.
 */
#define CTDB_FT_BUILDING_CMP          0
#define CTDB_FT_ROCK_DROP_CMP         1

/* Types of linear features.
 */
#define CTDB_FT_TREELINE_CMP          0
#define CTDB_FT_CONCERTINA_WIRE_CMP   1
#define CTDB_FT_FENCE_CMP             2
#define CTDB_FT_MINEFIELD_FENCE_CMP   3
#define CTDB_FT_DRAGONS_TEETH_CMP     4

/* Types of dragon's tooth components.
 */
#define CTDB_DT_STEEL_HEDGEHOG_CMP       0
#define CTDB_DT_STEEL_TETRAHEDRON_CMP    1
#define CTDB_DT_CONCRETE_BLOCK_CMP       2
#define CTDB_DT_CONCRETE_TETRAHEDRON_CMP 3

/* CTDB_EDGE:
 *
 * Structure used to collect a list of linear features which form an
 * uninterrupted "edge" between two intersections.  The `feature'
 * member is actually variable length.
 */
typedef struct ctdb_edge_cmp
{
    /* Total distance along the segment (useful for planning) */
    int32  distance;

    /* Indices of the nodes at the beginning and end of the segment */
    int32  first_node;
    int32  last_node;

    int32  num_features;
    struct ctdb_feature_reference_cmp
    {
	/* The patch where this feature is stored */
	uint32 patch_number;

	/* Byte offset into the patch */
	unsigned patch_offset : 24;

	/* Whether the vertices in the patch are in reverse order */
	unsigned reverse      : 1;

	unsigned padding      : 7;
    } feature[1];
} CTDB_EDGE_CMP;

/* CTDB_NODE:
 *
 * Represents the topologic connection between CTDB_EDGE structures.
 * The `connection' member is actually variable length.
 */
typedef struct ctdb_node_cmp
{
    /* Approximate location of the node (useful for planning) */
    float32 x, y;

    /* Whether the node corresponds to a bridge */
    unsigned is_bridge : 1;

    unsigned padding   : 23;

    uint8 num_connections;
    struct ctdb_connection_cmp
    {
	unsigned node_in_other_cell : 1;  /* TRUE indicates that the node
					   * index references a node in 
					   * another cell, FALSE indicates
					   * that it is in this cell.
					   */
	unsigned connection_crosses_mes : 1;  /* TRUE indicates that either the
					       * current node or the node 
					       * connected to is in an MES
					       * structure.  This dictates
					       * special usage of the edge
					       * structure below.
					       */
	unsigned padding            : 6;
	
	/* The node which this connects to */
	unsigned node_index         : 24;
	
	union edge_union_cmp
	{
	    int32 edge_index; /* The edge used to connect these nodes,
			       * if they are in the same database. 
			       */

	    int32 other_cell; /* If node_in_other_cell is TRUE,
			       * the cell in which node_index is valid.
			       */
	    uint32 aperture_key; /* If connection_crosses_mes is TRUE,
				  * this specifies which aperture in which
				  * MES structure must be traversed to move 
				  * from the current node to the connected
				  * node.
				  */
	} edge;
    } connection[1];

} CTDB_NODE_CMP;

/* CTDB_LINEAR_DESC:
 *
 * This structure is used to return information about linear models to the
 * application.
 *
 * It is also used to pass a linear feature or an individual tree
 * (n_verts == 1) to the feature compression algorithms.
 *
 * `trunk_radius' specifies the radius of the tree trunk.
 *
 * `foliage_height' specifies the height of the foliage of the tree.
 *
 * `fullness' specifies the fullness of the leaves or wire.  0.0 represents
 * zero fullness (nothing there), and 1.0 represents maximum fullness
 * (completely blocks light).
 *
 * `total_height' specifies the height of the feature.
 *
 * `spacing' represents how far apart from each other the dragon's teeth are
 * placed along the linear segment.
 *
 * `component' specifies the type of dragon's tooth, specified in libctdb.h
 *
 * `type' specifies the type of linear feature, specified in libctdb.h
 *
 * `width' specifies the foliage width for trees, or the width of dragon's
 * teeth.
 */
typedef struct ctdb_linear_desc_cmp
{
    union
    {
	struct
	{
            float32  trunk_radius;
            float32  foliage_height;
            float32  fullness;
	    float32  total_height;
	} treeline_info;

	struct
	{
            float32  fullness;
            int32    force_id;
	} var_opacity_info;

	struct
	{
            float32  spacing;
            int32    component;
	} dragons_tooth_info;
    } linear_data;

    int32   type;
    float32 width;

} CTDB_LINEAR_DESC_CMP;


typedef struct ctdb_linear_model_cmp
{
    /* This defines the models for the various linear features.
     */
    union
    {
	struct
	{
	    /* `treeline_info' is used for trees and treelines.
	     * 
	     * `trunk_radius' is measured in the same units as x,y in the
	     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
	     *
	     *  `foliage_height' is the distance from the top of the tree to
	     * the lowest branches measured in the unit
	     * `CTDB_TREE_HEIGHT_BASIS':
	     * 
	     * #define CTDB_TREE_HEIGHT_BASIS (200.0 / (1<<14))
	     *
	     *  `total_height' is the distance from the top of the tree to
	     * the ground measured in `CTDB_TREE_HEIGHT_BASIS'.  This is
	     * only used for trees in aggregate models.  Otherwise, their
	     * heights are derived from their Z values (tree tops).  Z
	     * values aren't stored with aggregate features.
	     *
	     * `fullness' refers to the "density" of the leaves, wire, or
	     * whatever the feature is made of.  This value ranges from zero
	     * (nothing there) to the maximum value (lets no light through).
	     */
	    unsigned trunk_radius     : 16;
	    unsigned foliage_height   : 14;
	    unsigned pad1             : 2;
	    unsigned fullness         : 8;
	    unsigned total_height     : 14;
	    unsigned pad2             : 10;
	} treeline_info;

	/* `var_opacity_info' is used for wire and fences.
	 */
	struct
	{
	    unsigned fullness         : 8;
	    unsigned force_id         : 2;
	    unsigned pad1             : 22;
	    unsigned pad2             : 32;
	} var_opacity_info;

	/* `dragons_tooth_info' is used for dragon's teeth.
	 * 
	 * `spacing' represents how far apart from each other the dragon's
	 * teeth are placed along the linear segment.
	 *
	 * Dragon's teeth can be made of steel hedgehogs, steel tetrahedrons,
	 * concrete blocks, or concrete tetrahedrons.  Each of these affects
	 * intervisibility differently (hence the need for the `component'
	 * element).
	 */
	struct
	{
	    float32  spacing;
	    unsigned component        : 4;
	    unsigned pad              : 28;
	} dragons_tooth_info;
    } linear_data;

    /* `type' specifies the specific type of linear feature (i.e. treeline,
     * minefield fence, etc.).
     * 
     * The width of the foliage, the thickness of the concertina wire or
     * minefield fence, or the width of the dragon's teeth is stored in
     * `width', which is measured in the same units as X and Y in the
     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
     */
    uint8  type;
    uint8  pad;
    uint16 width;

} CTDB_LINEAR_MODEL_CMP;


typedef struct ctdb_volume_model_cmp
{
    char     model_name[32];
    uint32   model_id;
    uint32   next_state_index;
    float32  percent_damaged;
    uint16   num_vertices;
    uint16   mes_volume_index;
    float32  vertices[1];   /* variable-length */

} CTDB_VOLUME_MODEL_CMP;

/*********************************************************************
 * AGGREGATE MODEL DEFINITIONS                                       *
 *********************************************************************/

typedef struct ctdb_agg_model_ref_cmp
{
    uint8                    class; /* VOLUME, LINEAR */
    uint8                    reserved;
    uint16                   ref;   /* REFERENCE LINEAR, VOLUME,
				     * MES OR AGGREGATE MODEL LIBRARY
				     */
} CTDB_AGG_MODEL_REF_CMP;

typedef struct ctdb_agg_model_feature_cmp
{
    CTDB_AGG_MODEL_REF_CMP   model;
    uint16                   offset_x;
    uint16                   offset_y;
} CTDB_AGG_MODEL_FEATURE_CMP;

typedef struct ctdb_aggregate_model_cmp
{
    uint16                     n_features;
    uint16                     reserved;
    CTDB_AGG_MODEL_FEATURE_CMP features[1];  /* variable-length array */
} CTDB_AGGREGATE_MODEL_CMP;


/*********************************************************************
 * MES MODEL DEFINITIONS                                             *
 *********************************************************************/


typedef int32    CTDB_MES_VERTEX_CMP[XYZ];
typedef float64  CTDB_MES_VERTEX_F64_CMP[XYZ];

typedef struct ctdb_mes_list_vertex_f64_cmp
{
    int32                 num_vertices;
    CTDB_MES_VERTEX_F64  *vertices;
} CTDB_MES_LIST_VERTEX_F64_CMP;

typedef struct ctd_mes_polygon_f64_cmp
{
  CTDB_MES_LIST_VERTEX_F64_CMP boundary;
  int32                        soil;
} CTDB_MES_POLYGON_F64_CMP;

typedef struct ctdb_mes_list_polygon_f64_cmp
{
  int32                      num_polygons;
  CTDB_MES_POLYGON_F64_CMP  *polygons;
} CTDB_MES_LIST_POLYGON_F64_CMP;
/* eg, bridge.polygons[i].boundary.vertices[j][X] */

typedef struct ctdb_mes_triangle_cmp
{
    int16         material;      /* FACC MCC attributes codes 0-999 */
    float64       normal[XYZ];
    int32         vertex_ids[3];
} CTDB_MES_TRIANGLE_CMP;

typedef struct ctdb_mes_aperture_cmp
{
  int32      num_triangles;
  int32      first_triangle_id; /* The first in the set of triangles for 
                                 * this aperture */

  struct
  {
    int32      num_vertices;
    int32     *vertex_ids;      /* Vertices that make up the exterior 
				 * boundary of the aperture            
				 */
  } outline;

  struct
  {
    uint16     enclosure_id; /* An enclosure the aperture connects to */
    /* Whether there is a step from the bottom of this aperture to the floor
     * of the corresponding enclosure.  A value of TRUE means that the bottom 
     * of this aperture is at a different level from the floor of the 
     * enclosure, FALSE means that they are at the same level. */
    uint8      step_to_enclosure; 
    /* The thickness of the wall from the plane of the aperture to the 
     * corresponding enclosure.  The two wall-thicknesses together add up
     * to the total thickness of the wall. */
    int32      wall_thickness;
  } connections[2];

  int16      visible;           /* 1 = transparent, 0 = opaque */
  int16      open;              /* 1 = open, 0 = closed */
} CTDB_MES_APERTURE_CMP;

typedef struct ctdb_mes_grid_box_cmp
{
  int32           left_front_bottom_corner[XYZ];
  int32           right_back_top_corner[XYZ];
  int32           outer_box_id;      /* parent of this box */
  uint16          grid_index;        /* which child is this box */
  uint16          dimensions_x;
  uint16          dimensions_y;
  uint16          dimensions_z;
  int32          *box_ids;           /* array of children of this box */
  uint16          num_apertures;  
  uint16         *aperture_ids;      /* ids of apertures intersecting 
				      * this box */
  uint16          num_triangles;  
  int32          *wall_triangle_ids; /* ids of enclosure wall triangles
				      * intersecting this box */
} CTDB_MES_GRID_BOX_CMP;

typedef struct ctdb_mes_enclosure_cmp
{
  int32       average_height;
 struct
  {
    int32    num_vertices;
    int32   *vertex_ids;  /* Array of vertex ids that make up the floor
			   * of the enclosure, in clockwise order
			   */
  } footprint;

  int32      num_triangles; 
  int32      first_wall_triangle_id; /* The first in the list of triangles
			    	      * for the walls of this enclosure 
				      */
  uint16     num_apertures; 
  uint16    *aperture_ids; /* Array of the ids of the apertures connecting to 
		   	    * this enclosure */
} CTDB_MES_ENCLOSURE_CMP;

typedef enum 
{
     CTDB_MES_UNKNOWN_CMP,
     CTDB_MES_BUILDING_CMP,
     CTDB_MES_CAVE_CMP,
     CTDB_MES_TUNNEL_CMP,
     CTDB_MES_BRIDGE_CMP,
     CTDB_MES_OTHER_CMP
} CTDB_MES_TYPE_CMP;

#define CTDB_MES_FIRST_TYPE_CMP CTDB_MES_UNKNOWN_CMP
#define CTDB_MES_LAST_TYPE_CMP  CTDB_MES_OTHER_CMP

#define CTDB_MES_TYPE_LIST_CMP \
	 "unknown",            \
	 "building",           \
	 "cave",               \
	 "tunnel",             \
	 "bridge",             \
	 "other"

typedef struct ctdb_mes_template_cmp
{    
    CTDB_MES_TYPE_CMP    type;
    float64              fixed_point_basis; /* Scaling of local coordinates */
    int32                num_vertices;
    CTDB_MES_VERTEX_CMP *all_vertices;      /* Array of all vertices:
				             * x,y,z, x,y,z, x,y,z, etc. 
				             */
    
    /* We choose to group the list of all triangles so that all 
     * wall triangles are at the end of this array of triangles.
     * Furthermore, the list of triangles will be grouped by enclosure,
     * so that all wall triangles of an enclosure are grouped together.
     * This allows us to specify the list of triangles for an enclosure 
     * merely by specifying the ID of the first triangle in the list.
     * Similarly, the list of aperture triangles is grouped by aperture.
     */
    int32                   num_triangles;
    CTDB_MES_TRIANGLE_CMP  *all_triangles;     /* Array of all triangles, 
						* both wall and aperture */
    uint16                  num_enclosures;
    CTDB_MES_ENCLOSURE_CMP *all_enclosures;    /* Array of all enclosures */
    uint16                  num_apertures;
    CTDB_MES_APERTURE_CMP *all_apertures;     /* Array of all apertures */
    int32                   num_boxes;
    CTDB_MES_GRID_BOX_CMP  *all_boxes;         /* Array of all gridboxes */
} CTDB_MES_TEMPLATE_CMP;

typedef struct ctdb_mes_volume_cmp
{
    float64                origin[XYZC];           /* Origin of structure in 
						    * GCS 
						    */
    float64                tfm_world_to_mes[3][3]; /* Rotation matrix from GCS 
						    * world space to mes local 
						    * space
						    */
    CTDB_MES_LIST_VERTEX_F64_CMP roofline;  /* Array of vertices in roofline */
    CTDB_MES_LIST_VERTEX_F64_CMP floorline; /* Array of vertices in floorline*/
    CTDB_MES_TEMPLATE_CMP       *template;
} CTDB_MES_VOLUME_CMP;


typedef struct ctdb_patch_table_ref_cmp
{
    unsigned soil_table : 16;	/* Index of patch soil table */
    unsigned water_char : 16;	/* Index of patch water characteristics */
} CTDB_PATCH_TABLE_REF_CMP;

/* Wave characteristics */
typedef struct ctdb_wave_chars_cmp
{
    float32 height;		/* In meters */
    float32 period;		/* In seconds */
    float32 speed;		/* In meters per second */
    float32 direction;		/* $$$ units??? Tangent of some angle...*/
} CTDB_WAVE_CHARS_CMP;


/* Structure used to store surface characteristics for water polygons */
typedef struct ctdb_water_chars_cmp
{
    int16               tidal_zone_index;
    uint16              pad;
    float32             surface_temp;	/* In degrees C */
    float32             surf_height;	/* In meters */
    CTDB_WAVE_CHARS_CMP primary_wave;
    CTDB_WAVE_CHARS_CMP secondary_wave;
} CTDB_WATER_CHARS_CMP;

typedef struct ctdb_tidal_zone_cmp
{
    float32 offset;		/* Current tidal offset for this zone */
    /* $$$ add a region or identifier of some kind??? */
} CTDB_TIDAL_ZONE_CMP;


/* A PATCH_GROUP is a collection of information about four contiguous
 * patches (contiguous in the sense x+y*w).  For each patch there are
 * three flags: occupied indicates there are features in the patch,
 * dyn_terr indicates that there are dynamic terrain changes to the
 * patch (future use), reserved is for future use (you never know when
 * you might want just one bit!).  When not using a feature cache, the
 * fields feature_start and *_offset together describe the start of
 * each patch's features as follows:
 * 	Patch 0 at (ctdb->features + feature_start)
 * 	Patch 1 at (ctdb->features + feature_start + patch1_offset).
 * 	Patch 2 at (ctdb->features + feature_start + patch2_offset).
 * 	Patch 3 at (ctdb->features + feature_start + patch3_offset).
 * When using a feature cache, the mapped bit indicates that the
 * feature is currently mapped into memory (cache hit), and the cache
 * field identifies which of the 16 feature caches this PATCH_GROUP
 * belongs in.  Once in the cache, the cache_bucket field identifies
 * where in the cache the PATCH_GROUP has been placed.  The *_offset
 * fields are then relative to the start of the bucket.
 */
typedef struct ctdb_patch_group_cmp
{
    unsigned patch0_occupied : 1;
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_te16     : 1;/* 0=8bit TEs 1=16bit TEs */
    unsigned mapped          : 1;
    unsigned cache           : 4;
    unsigned feature_start   : 24;
    
    unsigned patch1_occupied : 1;
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_te16     : 1;/* 0=8bit TEs 1=16bit TEs */
    unsigned patch1_offset   : 13;
    
    unsigned patch2_occupied : 1;
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_te16     : 1;/* 0=8bit TEs 1=16bit TEs */
    unsigned patch2_offset   : 13;
    
    unsigned patch3_occupied : 1;
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_te16     : 1;/* 0=8bit TEs 1=16bit TEs */
    unsigned patch3_offset   : 13;
    
    uint16 cache_entry;

    unsigned patch0_water    : 1; /* 1 => patch 0 has a patch water level */
    unsigned patch1_water    : 1; /* 1 => patch 1 has a patch water level */
    unsigned patch2_water    : 1; /* 1 => patch 2 has a patch water level */
    unsigned patch3_water    : 1; /* 1 => patch 3 has a patch water level */
    unsigned pad1            : 28; /* Free space! Feel free to find some clever
				    * use for it!
				    */

    /* BITS             
	    3 : Features
	    2 : Buildings
	    1 : TREES
	    0 : Microterrain  */

    uint8   feature_present[CTDB_MAX_FEATURE_PRESENT_PER_PG_CMP]; 
					
    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32                water_elev[CTDB_PATCHES_PER_GROUP_CMP];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF_CMP tables[CTDB_PATCHES_PER_GROUP_CMP];

    /* Array[patches] numb TE in patch */
    uint16  num_te[CTDB_PATCHES_PER_GROUP];
    /* te_offset[0] is relative to ctdb->terrain_elements and the remaining
       are relative to ctdb->terrain_elements+te_offset[0] in bytes*/
    uint32       te_offset[CTDB_PATCHES_PER_GROUP];
    uint8        te_vgrid_dim[CTDB_PATCHES_PER_GROUP];
    uint16       *te_vgrid_offset[CTDB_PATCHES_PER_GROUP];
    float32       max_elev[CTDB_PATCHES_PER_GROUP];
    int32        total_patch_feature_words;
} CTDB_PATCH_GROUP_CMP;



/**********************************************************************
 * PAT data structure                                                *
 **********************************************************************/
/* Different ways of looking at the data stored for each attribute */

/* Data structure used to store a column of the PAT table */
typedef struct ctdb_pat_column_hdr_cmp
{
  uint32                 facc;             /* FACC code for this column */
  union ctdb_pat_data_cmp *data;             /* Attribute data */
} CTDB_PAT_COLUMN_HDR_CMP;	

/* PAT table data structure */
typedef struct ctdb_pat_table_cmp
{
  uint32 num_columns;
  CTDB_PAT_COLUMN_HDR_CMP *columns; /* array of column header structures,
				    * one for each column.
				    */
  uint32 num_entries;            /* Number of valid entries in table */
  uint32 alloced_entries;        /* Number of entries each column currently
				  * has alloced. (We will enforce the rule
				  * that all columns have the same # of 
				  * entries alloced at all times.) This will 
				  * support efficient run-time addition
				  * of entries.
				  */
  uint32 alloced_columns;        /* Number of columns allocated.
				  * This is different than the number used
				  * in the compiler representation 
				  * (though not the ctdb representation)
				  * because columns may be added at compile
				  * time.
				  */
} CTDB_PAT_TABLE_CMP;  

/* Different ways of looking at the data stored for each attribute in PAT*/
typedef union ctdb_pat_data_cmp
{
    uint32  uinteger;
    int32   integer;
    int32   boolean;
    char    character;
    float32 fp;
    int32   string_index;  /* Index into global string array */
    /* Add others as needed - none may exceed 32 bits */
} CTDB_PAT_DATA_CMP; 


typedef struct ctdb_cmp
{
    float64              fixed_point_basis;
    float64              inv_fixed_point_basis;
    
    /* Tables mapping local 4 or 8 bit soil types to the global list.  Used
     * only for format 6+
     */
    int32             num_soil_tables;
    int32           **soil_tables;
    int32             soil_table_size;
    int8             *soil_table_storage;    
    
    /* Pointer to the PAT table */
    struct ctdb_pat_table_cmp *pat_table_ptr;  

    /* Table of sea-state combinations */
    int32                        num_water_chars;
    struct ctdb_water_chars_cmp *water_chars;
    
    /* Table of tidal zones */
    int32                        num_tidal_zones;
    struct ctdb_tidal_zone_cmp  *tidal_zones;
    
    /* Spacing between grid posts */
    int32                incr;
    float64              inv_incr;
    float64              incrD;
    
    /* Spacing between patches */
    int32                patch_incr    ;/* number of posts per patch */
    float64              patch_inv_incr;/* 1 over patch_incr */
    float64              patch_incrD   ;/* double patch_incr */
    
    /* Database size */
    int32                pages_wide;
    int32                pages_high;
    int32                patches_wide;
    int32                patches_high;
    int32                min_x;
    int32                max_x;
    int32                min_y;
    int32                max_y;
    int32                min_x_post;
    int32                max_x_post;
    int32                min_y_post;
    int32                max_y_post;
    
    /* Useful trivia */
    int32                input_format;
    float64              min_z;
    float64              max_z;
    float64              origin_northing;
    float64              origin_easting;
    int32                origin_zone_number;
    int32                origin_zone_letter;
    int32                datum;
    int32                gcs_mode;
    int32                cell_id;

    int32                origin_x;
    int32                origin_y;

    float64              west_long;
    float64              east_long;
    float64              south_lat;
    float64              north_lat;

    char                 name[CTDB_NAME_LENGTH_CMP];
    int16                version;
    char                 date[CTDB_DATE_LENGTH_CMP];
    int32                total_memory;
    uint32               is_curved;
    uint32               is_sw_ne;
    CTDB_PATCH_GROUP_CMP *patch_groups;

    /* Feature memory if resident */
    int32                     num_features;
    union ctdb_feature_data_cmp  *features;
    
    /* Grid Post/Feature Present cache information */
    int32           elevations_resident;
    int32           elevation_seek;
    int32           soil_seek;
    int32           elevation_cache_pages;
    int32           soil_cache_pages;
    int32           elevation_direct_map_modulus;
    int32           soil_direct_map_modulus;
    int32           elevation_hit_count;
    int32           soil_hit_count;
    int32           elevation_miss_count;
    int32           soil_miss_count;
    int32          *elevation_cache_page_numbers;
    int32          *soil_cache_page_numbers;
    int32         **elevation_pages;
    int32         **soil_pages;

    /* Database file */
    FILE         *file;

    /* Grid posts/feature present */
    int32          *elevations;
    int32          *soils;


    /* Quadtree of abstract features */
    int32                    num_quad_data;
    union ctdb_feature_data_cmp *quad_data;

    /* Topologic relationships */
    int32           num_nodes;
    CTDB_NODE_CMP **nodes;
    int32           node_size;
    int8           *node_storage;
    int32           num_edges;
    CTDB_EDGE_CMP **edges;
    int32           edge_size;
    int8           *edge_storage;

    /* Linear models */
    int32                   num_linear_models;
    CTDB_LINEAR_MODEL_CMP  *linear_models;

    /* Volume models */
    int32                   num_volume_models;
    CTDB_VOLUME_MODEL_CMP **volume_models;
    CTDB_VOLUME_MODEL_CMP **mes_volume_models; 

    /* Aggregate models */
    int32                      num_aggregate_models;
    CTDB_AGGREGATE_MODEL_CMP **aggregate_models;

    /* MES volumes */
    int32                     num_mes_volumes;  
    CTDB_MES_VOLUME_CMP      *mes_volumes;
    int32                     num_mes_templates;  
    CTDB_MES_TEMPLATE_CMP    *mes_templates;

     /* Terrain Element Memory */
    int32           n_te_bytes;           /* total number of TEs in bytes */
    
    /* $$$ WARNING:  THIS SHOULD REALLY BE DEFINED PER-FORMAT, IN CASE WE
     * EVER CHANGE IT.  FOR NOW, I WON'T BOTHER, BUT IF IT EVER CHANGES,
     * THESE NEED TO BE EXPLICITLY DEFINED HERE AND FOR FORMATS 4+
     */
    TERRAIN_ELEMENT_16 *terrain_elements; /* point to array of TEs */

    uint32 is_gridded;
    uint32 is_tinned;
    uint32 is_short;
    float64 meters_to_patch;
    int32 max_patch_number;
    int32 min_patch_number;
} CTDB_CMP;

typedef struct ctdb_file_header_cmp
{  
    int32    format;
    unsigned is_curved : 1;
    unsigned is_sw_ne  : 1;
    unsigned is_gridded: 1;   
    unsigned is_tinned : 1;
    unsigned is_short  : 1;
    unsigned pad1      : 27;
    float64  fixed_point_basis;

    int32    reserved1[16];

    int32    post_increment;
    int32    posts_per_patch_side;
    int32    pages_wide;
    int32    pages_high;
    int32    min_x;
    int32    max_x;
    int32    min_y;
    int32    max_y;
    int32    num_features;
    int32    patch_cache_size_map[CTDB_PATCH_CACHES_CMP];
    int32    pad2;
    float64  origin_northing;
    float64  origin_easting;
    int32    origin_zone_number;
    int32    origin_zone_letter;
    int32    datum;
    char     name[CTDB_NAME_LENGTH_CMP];
    int16    version;
    char     date[CTDB_DATE_LENGTH_CMP];
    int32    num_nodes;
    int32    node_size;
    int32    num_edges;
    int32    edge_size;
    int32    num_quad_data;
    int32    n_patch_bytes;
    int32    num_linear_models;
    int32    num_volume_models;
    int32    aux_hdr_bytes;	/* Size of auxiliary header */
    int32    num_soil_tables;
    int32    soil_table_size;
    int32    num_water_chars;
    int32    num_tidal_zones;
    int32    num_aggregate_models;
    int32    num_mes_volumes;
    int32    num_mes_templates;
} CTDB_FILE_HEADER_CMP;

/* Additional data we didn't want to try to cram into the file header */
typedef struct ctdb_aux_file_header_cmp
{
    unsigned gcs_mode : 2;
    unsigned pad1     : 30;
    int32 cell_id;
    int32 origin_x;
    int32 origin_y;
    float64 west_long;		/* Left-most longiitude in database */
    float64 east_long;		/* Right-most longitude in database */
    float64 south_lat;		/* Bottom-most latitude in database */
    float64 north_lat;		/* Top-most latitude in database */
} CTDB_AUX_FILE_HEADER_CMP;



#define CAND_TE_QUEUE_LENGTH_CMP 65536  /* 2^16 where 16 is num of bits in 
					 * uint16 
					 */

/* This data structure has been defined as a collection of data required
 * by the elevation lookup processing.
 */
typedef struct elevation_lookup_storage_cmp
{
    CTDB_CMP              *ctdb;    /* pointer to the database data structure 
				     */
    int32		   num;     /* patch number */
    int32                  member;  /* patch member */
    CTDB_COORD2D          *dbpt;    /* db coordinate system */
    CTDB_COORD2D_I32       npt;     /* coordinates normalized */
    CTDB_COORD2D_I32       coord;   /* patch indices */         
    CTDB_PATCH_GROUP_CMP  *pg;      /* pointer to the patch_group data */
    CTDB_COORD2D           frac;    /* frac part within patch */
    int32                  te16;    /* 1=patch has 16bit TE indicies */    
/* OUTPUT DATA */
    float64                elevation;
    uint32                 te_index;/* TE being evaluated */
    int32                  soil;
    CTDB_POINT_STATE       where;
/* TERRAIN ELEMENT DATA */
    int32                  front;   /* queue pointers */
    int32                  back;
    int32                  mark_back;
    int32                  mapped_te;
    uint32                 check_mark[CAND_TE_QUEUE_LENGTH]; /* Checked TE list
							      */
    uint32                 te_queue[CAND_TE_QUEUE_LENGTH_CMP];
    /* Both of these point to TERRAIN_ELEMENT structures. */
    ADDRESS                te;      /* address of TE being evaluated */
    ADDRESS                base;    /* Base address of TE list for patch
				     * (i.e. &TE[0]) 
				     */
} ELEVATION_LOOKUP_STORAGE_CMP;



typedef struct ctdb_search_db_cmp
{
    /* Flag indicating whether we are looking at real patches
     * (x, y), or quad nodes (q).  We look at real patches
     * first, then quad nodes.
     */
    uint16 real_patches;
    uint16 x, y;
    int16 q;
    
    /* The last patch we fetched, and the cache_miss count after
     * we fetched it (if the feature_miss_count doesn't change, then
     * we can assume it is still in memory, and thus we won't
     * need to re-fetch it; if the miss_count has increased since
     * the last time we fetched a patch, then we would have to
     * re-fetch).
     */
    int32 last_number;
    CTDB_FEATURE_DATA_CMP *last_patch;
    int32 last_feature_miss_count;
    
    CTDB_CMP *ctdb;          /* GCS: current cell for search space */
    int32     current_index; /* GCS: current cell index in list */
    
    /* Bounds of the search in terms of the *real* patches (not quad
     * nodes) to be covered.  These are inclusive: min <= patch <= max.
     */
    int32 min_patch_x, max_patch_x;
    int32 min_patch_y, max_patch_y;
    
    /* Quad nodes which are also to be covered in the search */
    int32   num_quad_nodes;
    uint32 *quad_nodes;
    
} CTDB_SEARCH_DB_CMP;

/* ctdb_search_space_cmp:
 *
 * This is a private data structure used to track the progress of
 * application terrain searches.  The application first creates the
 * search space by specifying a bounding rectangle and some search
 * behavior control flags.  A representation of the search space
 * in terms of CTDB patches is determined from this information, and
 * various iteration variables are initialized.  The application
 * may search the space in any order desired, so many iteration variables
 * are needed to properly track which features have been returned
 * and which have not, without additional searching.
 */
struct ctdb_search_space_cmp
{
    /* GCS:  The Search space is defined in terms global search extents and a
     * list of cells (ctdbs) that make up the search space.  Each cell is
     * searched in turn and is maintained for the search routines in the ctdb
     * field.  When search is exhausted within the current cell, then the next
     * cell is assigned and this search space is reinitialized with respect to
     * this new current cell.
     */
     
    int32 min_x, min_y, max_x, max_y; /* search extents */
    int32 initial_frame;  /* cell_id in which the search extents are defined */
    int32 n_cells;       /* GCS: number of cells */
    int32 *cell_id_list; /* list of cells (ctdbs) for search space */
    CTDB_SEARCH_DB_CMP *cell_data_list; /* Initial iteration data for each
					 * cell
					 */


    /* Application-provided CTDB, used in subsequent calls.  Since
     * the internal representation of the search is dependent on the
     * particular CTDB being explored, it would be inappropriate to
     * require/allow the application to provide the CTDB in subsequent
     * calls.
     */
    CTDB_CMP *call_ctdb;

    /* Search control flags specified by application. */
    uint32 control_flags;

    /* Current patch being investigated by the application
     * (used if control_flags contains CTDB_BY_PATCH).
     */
    /* Since now all search capability is over multiple CTDB's then the CTDB d
       relavent information must be included with the current patch information
       data structure for each type of item that can be searched for 
       (i.e., by patch, tree, building, linear, abstract)
       */
    CTDB_SEARCH_DB_CMP search_patch;

    /* We group all the iteration variables into a structure for
     * easy initialization.
     */
    struct
    {
	/* Iteration variables associated with each feature */
	struct ctdb_search_state_cmp
	{
	    /* Patch being traversed (used if control_flags does not
	     * contain CTDB_BY_PATCH).
	     */
	    CTDB_SEARCH_DB_CMP search_patch;

	    /* Flag indicating we've finished everything */
	    int32 all_done;

	    /* The offset into the patch for the *next* feature to be
	     * returned.  (Note: If last_patch is NULL, then we have not
	     * looked at the patch yet, so this value cannot be valid
	     * in that case.)
	     */
	    int32 offset;

	    /* Keeps track of the last feature within an aggregation that
	     * has been returned.
	     */
	    int32 feature_number;

	} tree, building, linear, abstract[CTDB_ABSTRACT_MAX];

	/* When control_flags contains CTDB_TREELINE_TREES, we need to
	 * return the optional first & last trees separately from the
	 * tree line.  We use this variable to remember what we've
	 * returned so far on each tree.
	 */
	enum
	{
	    /* The tree-line, and the first and last tree */
	    CTDB_RETURNED_ALL_CMP,
	    /* Just the tree-line part */
	    CTDB_RETURNED_LINE_CMP,
	    /* Just the tree-line and the first tree */
	    CTDB_RETURNED_FIRST_CMP,
	    /* Done with the linears, now looking through the aggregates */
	    CTDB_PROCESSING_AGGREGATES_CMP
	} last_tree;

    } iteration;
};


/* CTDB_SEARCH_SPACE_PTR_CMP:
 *
 * Used to establish bounds on a search, and used to
 * track local iteration variables.
 */
typedef struct ctdb_search_space_cmp *CTDB_SEARCH_SPACE_PTR_CMP;


/***************************************************************************
 * INTERNAL REPRESENTATION ACCESS FUNCTIONS                                *
 ***************************************************************************/

/* Return the soil map for the specified patch */
extern int32 *
ctdb_intern_get_soil_map(
			 CTDB_CMP *ctdb,
			 int32     patch_number
			 );

extern void
ctdb_intern_unpack_header(
			  CTDB_CMP             *ctdb,
			  CTDB_FILE_HEADER_CMP *hdr
			  );

extern void
ctdb_intern_unpack_aux_header(
			      CTDB_CMP                 *ctdb,
			      CTDB_AUX_FILE_HEADER_CMP *aux_hdr
			      );

/* The ctdb_intern_pack_header routine will create a file header (hdr) from the
 * compiler ctdb data structure. 
 */
extern void 
ctdb_intern_pack_header(
			CTDB_CMP             *ctdb,
			CTDB_FILE_HEADER_CMP *hdr
			);

/* The ctdb_intern_pack_aux_header routine will create an aux file header 
 * (aux_hdr) from the compiler ctdb data structure.
 */
extern void 
ctdb_intern_pack_aux_header(
			    CTDB_CMP                 *ctdb,
			    CTDB_AUX_FILE_HEADER_CMP *aux_hdr
			    );

extern int32 
ctdb_intern_point_on_database(
			      CTDB_CMP *ctdb,
			      float64   x,
			      float64   y
			      );

/* get_post_index returns the index into the post elevations array given 
 * the location of post in norm_x, norm_y coordinates. 
 */
extern int32 
ctdb_intern_get_post_index(
			   CTDB_CMP      *ctdb,
			   int32          norm_x,
			   int32          norm_y
			   );

/* The ctdb_intern_lookup_elev_post routine will find and return the elevation 
 * post as specified by the given post coordinates. 
 */
extern int32 
ctdb_intern_lookup_elev_post(CTDB_CMP      *ctdb,
			     int32          norm_x,
			     int32          norm_y
			     );

/* The ctdb_intern_lookup_soil_post routine will find and return the soil
 * post as specified by the given post coordinates. 
 */
extern int32 
ctdb_intern_lookup_soil_post(CTDB_CMP      *ctdb,
			     int32          norm_x,
			     int32          norm_y
			     );

/* Grab the post (from elevations, not from ctdb->elevations) at 
 * normx, normy.
 */
int32
get_post(
	 CTDB_CMP      *ctdb,
	 int32         *elevations,
	 int32          norm_x,
	 int32          norm_y
	 );

/* This function returns the elevation of the post defined by post coordinates
   (norm_x, norm_y). */

extern float64 
ctdb_intern_lookup_post_elevation (
				      CTDB_CMP       *ctdb,
				      int32           norm_x, 
				      int32           norm_y
				      );

/* The ctdb_intern_get_ground_elevation function will return the elevation at 
 * the specified (x,y) location (in database coordinates).  If mode is set to 
 * FAVOR_GRID, it will return the elevation specified by the grid posts if 
 * possible, and if not, revert to the TEs.  If mode is set to FAVOR_TES, it 
 * will do the opposite.
 */
#define FAVOR_TES 0
#define FAVOR_GRID 1

extern float64
ctdb_intern_get_ground_elevation (    
				     CTDB_CMP      *ctdb,
				     float64        x,
				     float64        y,
				     int32          mode
				     );

/* This function access the soil information for the post specified by the post
   coordinates (norm_x, norm_y). */

extern uint8 
ctdb_intern_lookup_soil (
			    CTDB_CMP      *ctdb,
			    float64   norm_x,
			    float64   norm_y
			    );

/* The ctdb_intern_get_patch_number will return the patch number containing
   the point specified by (norm_x, norm_y). */

extern int32
ctdb_intern_get_patch_number (
			      CTDB_CMP *ctdb,
			      int32     norm_x,
			      int32     norm_y
			      );

/* ctdb_intern_test_patch_soil determines whether all the soil posts of a
   patch point to the same soil type. */

extern int32
ctdb_intern_test_patch_soil(
			    CTDB_CMP *ctdb,
			    int32     patch_x,
			    int32     patch_y
			    );

/* The ctdb_intern_lookup_patch will return a pointer to the feature list
   belonging to the patch specified by the given patch_number. */

extern CTDB_FEATURE_DATA_CMP *
ctdb_intern_lookup_patch (
			     CTDB_CMP      *ctdb,
			     int32          patch_number
			     );

/* This routine will return the number of terrain elements for the patch
   with the given patch number. */

extern int32
ctdb_intern_number_tes (
			  CTDB_CMP          *ctdb,
			  int32              patch_number
			  );

/* This routine will return a pointer to the terrain element list for the patch
   with the given patch number. */

extern TERRAIN_ELEMENT_16 * 
ctdb_intern_lookup_terrain_element (
				       CTDB_CMP          *ctdb,
				       int32              patch_number
				       );

/* The ctdb_intern_lookup_te_info will return the soil and elevation 
 * information for the grid post area specified by post_x, post_y (i.e., the 
 * grid square whose upper left corner is at POST_X(post_x), POST_Y(post_y).  
 * sSW returns a best guess at the soil type in the SW triangle, sNE returns
 * a best guess at the soil type in the NE triangle, and sHere returns
 * the soil type (not a guess) at (post_x, post_y).  Any and all of these
 * pointers may be NULL if no value is desired.  The elevation returned
 * is at the specified point (*not* the grid post location, unless the
 * point happens to be on the post)
 */
extern void 
ctdb_intern_lookup_te_info (
			       float64        post_x,
			       float64        post_y,
			       CTDB_CMP      *ctdb,
			       float64       *z,
			       int32         *sNE,
			       int32         *sSW,
			       int32         *sHere
			       );

/* This routine will return the next edge to be evaluated given the side/
 * edge last evaluated, the index of the terrain element and the list of 
 * terrain elements.
 */
extern CTDB_STATUS 
ctdb_intern_get_te_edge_index16(
				TERRAIN_ELEMENT_16  *te,
				uint32               index,/*vertex number */
				int32                side, /*side of interest*/
				int32 	             edge[]
				);

extern int32
ctdb_intern_get_cell_id(
			CTDB_CMP *ctdb
			);


/* Create a search space for application-driven iteration */
extern CTDB_SEARCH_SPACE_PTR_CMP 
ctdb_intern_create_search_space(
				CTDB_CMP *ctdb,
				int32     min_x,
				int32     min_y,
				int32     max_x,
				int32     max_y,
				uint32    control_flags
				);

/* Free a search space */
extern void 
ctdb_intern_destroy_search_space(
				 CTDB_SEARCH_SPACE_PTR_CMP space
				 );

/* Return the next abstract feature */
extern int32 
ctdb_intern_next_abstract(
			  CTDB_SEARCH_SPACE_PTR_CMP space,
			  int32                     type_code,
			  float32                 **vertices,
			  CTDB_ABSTRACT_DATA_CMP  **data,
			  int32                    *data_size
			  );

/* ctdb_next_quad_patch allows the search space to advance to the next node of
 * the quad tree.  This function must be called only if the control flag
 * CTDB_BY_PATCH was passed to ctdb_create_search_space().  
 */
extern int32 
ctdb_intern_next_quad_patch(
			    CTDB_SEARCH_SPACE_PTR_CMP space
			    );

/* This routine will determine the maximum elevation within the given 
 * gridded patch by checking the elevation of each post within the patch 
 */    
extern void 
ctdb_intern_compute_grid_max_z(
			       CTDB_CMP       *ctdb,
			       int32           patchx, 
			       int32           patchy,
			       float64        *max_elev
			       );

/* Traverses the grid squares and patches, calling a user-defined function 
 * for each.  Note that this version does not have all the functionality of
 * of the one in libctdb.  In particular, it doesn't handle rays that start
 * at anything but a point, it cannot generate patch feature masks for
 * later use (the feature present bits in the compiler rep are generally not
 * set correctly when this function is called), and it does not bother
 * performing elevation lookups.
 */
extern int32
ctdb_intern_intervis(
		     CTDB_CMP          *ctdb,
		     float64            start[3],
		     float64            end[4],
		     int32            (*post_fcn)(),
		     int32            (*patch_fcn)(),/* Ignored if NULL */
		     int32              feature_mask
		     );

extern void 
ctdb_intern_print_description(
			      CTDB_CMP      *ctdb
			      );

/* Return relevant patch water data for a patch in pw.  Returns TRUE if
 * the patch has patch water, else FALSE (if there is no patch water,
 * the contents of pw should be ignored)
 */
extern int32
ctdb_intern_get_patch_water(
			    CTDB_CMP             *ctdb,
			    int32                 patch_x,
			    int32                 patch_y,
			    CTDB_CMP_PATCH_WATER *pw
			    );


/* Sets each element of post array to TRUE if the corresponding post has
 * land microterrain (as opposed to microwater) on it, otherwise to FALSE.
 * The caller is responsible for allocating post_array--it must have at
 * least as many elements as there are posts in the database.
 */
extern void
ctdb_intern_get_land_micro_feature_present(
					   CTDB_CMP *ctdb,
					   int8     *post_array
					   );


/* Copy the contents of in into out.  The routine assumes that the space
 * for out itself has been allocated, but allocates all subordinate
 * pointers as needed.
 */
extern void
ctdb_intern_copy_mes_template(
			      CTDB_MES_TEMPLATE_CMP *in,
			      CTDB_MES_TEMPLATE_CMP *out
			      );

/* Copy all the MES templates from ctdb to ctdb_out.  This routine does ALL
 * necessary memory allocation on the output side.
 */
extern void
ctdb_intern_copy_mes_templates(
			       CTDB_CMP *ctdb,
			       CTDB_CMP *ctdb_out
			       );

/* Copy the contents of in into out.  The routine assumes that the space
 * for out itself has been allocated, but allocates all subordinate
 * pointers as needed.
 */
extern void
ctdb_intern_copy_mes_volume(
			      CTDB_MES_VOLUME_CMP *in,
			      CTDB_MES_VOLUME_CMP *out
			      );

/* Copy the MES volumes from ctdb to ctdb_out.  This routine does ALL
 * necessary memory allocation on the output side.
 */
extern void
ctdb_intern_copy_mes_volumes(
			     CTDB_CMP *ctdb,
			     CTDB_CMP *ctdb_out
			     );

/* Dellocate all storage used by compiler representation database */
extern void
ctdb_free_cmp_db(
		 CTDB_CMP *ctdb
		 );


/* MES Stuff */
extern int32
ctdb_intern_mes_to_gcscs(
                         CTDB_MES_VOLUME_CMP *mes,
                         int32                x0,
                         int32                y0,
                         int32                z0,
                         int32                cell,
                         float64             *x1,
                         float64             *y1,
                         float64             *z1
                         );

extern int32
ctdb_intern_mes_get_aperture_centroid(
				      CTDB_CMP              *ctdb,
				      uint32 aperture,
				      float64           *x,
				      float64           *y,
				      float64           *z
				      );

#define CTDB_MES_EXTERIOR_CMP 0
#define CTDB_INVALID_MES_ID_CMP            ((uint16) 0xffff)
#define CTDB_INVALID_MES_APERTURE_ID_CMP   ((uint16) 0xffff)
#define CTDB_INVALID_MES_EDGE_KEY_CMP      ((uint32) RT_INVALID_NODE_KEY)

#define CTDB_MES_IDS_TO_EDGE_KEY_CMP(mes_id, aperture_id, node_key)           \
  ((node_key) = (mes_id << 16) | (aperture_id) & 0xffff)

#define CTDB_MES_EDGE_KEY_TO_MES_ID_CMP(edge_key)                             \
( (uint16) ((uint32)(edge_key) >> 16) )

#define CTDB_MES_EDGE_KEY_TO_APERTURE_ID_CMP(edge_key)                        \
( (uint16) ((uint32)(edge_key) & 0x0000ffff) )

#define CTDB_MES_VALID_MES_ID_CMP(ctdb, mes_id)                                 \
(((mes_id) != CTDB_INVALID_MES_ID_CMP)                                       && \
 ((mes_id) < (ctdb)->num_mes_volumes))

#define CTDB_MES_VALID_APERTURE_ID_CMP(ctdb, mes_id, aperture_id)               \
(((aperture_id) != CTDB_INVALID_MES_APERTURE_ID_CMP )                        && \
((aperture_id) < (ctdb)->mes_volumes[(mes_id)].template->num_apertures))

#define CTDB_MES_EDGE_KEY_EQUALS_INVALID_CMP(edge_key)                      \
  ((edge_key) == CTDB_INVALID_MES_EDGE_KEY_CMP)

extern int32
point_in_polygon(float64 pt[XY],
		 float64 polygon[][XY],
		 int32   num_pts_in_poly);
#endif /* _CMPREP_H_INCLUDED */





