#ifndef lint
static char rcsid [] = "$RCSfile: ct_coalesce.c,v $ $Revision: 1.1.1.1 $ $State: Exp $"; 
#endif
/****************************************************************************
*   File: ct_coalesce.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: code to coalesce a group of 3- and 4-sided polygons       *
*       Created: Thu Aug  4 1994                                            *
*       Author: mlongtin                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "ct_coalesce.h"
#include <libvecmat.h>
#include <libgeometry.h>
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <stdalloc.h>
#include <stdio.h>

/* Maximum number of segments */
#define MAX_SEGMENTS 300000

typedef struct segment
{
    float32 points[2][2];
    int8    used;
} SEGMENT;

static int32 n_polygons;
static SEGMENT segments[MAX_SEGMENTS];
static int32 n_segments;
static int32 map3[][2] = { {0, 1}, {1, 2}, {2, 0} };
static int32 map4[][2] = { {0, 1}, {1, 2}, {2, 3}, {3, 0} };

/* Returns TRUE if the passed segments are equal.
 */
static int32 segments_are_equal(point1a, point1b, point2a, point2b)
    float32   point1a[2];
    float32   point1b[2];
    float32   point2a[2];
    float32   point2b[2];
{
    float32   seg1[2][2], seg2[2][2];

    bcopy(point1a, seg1[0], 2 * sizeof(float32));
    bcopy(point1b, seg1[1], 2 * sizeof(float32));
    bcopy(point2a, seg2[0], 2 * sizeof(float32));
    bcopy(point2b, seg2[1], 2 * sizeof(float32));

    if((!bcmp(seg1[0], seg2[0], 2 * sizeof(float32)) &&
	!bcmp(seg1[1], seg2[1], 2 * sizeof(float32))) ||
       (!bcmp(seg1[0], seg2[1], 2 * sizeof(float32)) &&
	!bcmp(seg1[1], seg2[0], 2 * sizeof(float32))))
      return TRUE;

    return FALSE;
}

#define NUM_SIDES(x) (src_polys[x].is_3_sided ? 3 : 4)

static int32 segment_is_unique(current_polygon, src_polys, n_src_polys,
			       pointa, pointb)
    int32                   current_polygon;
    int32                   n_src_polys;
    CTDB_COALESCE_SRC_POLY  src_polys[];
    float32                 pointa[2];
    float32                 pointb[2];
{
    int32        i, j;

    for(i = 0; i < n_src_polys; i++)
    {
	if(NUM_SIDES(i) == 3)
	{
	    for(j = 0; j < 3; j++)
	      if(segments_are_equal(pointa, pointb,
				    src_polys[i].verts[map3[j][0]],
				    src_polys[i].verts[map3[j][1]]) &&
		 i != current_polygon)
		return FALSE;
	}
	else
	{
	    for(j = 0; j < 4; j++)
	      if(segments_are_equal(pointa, pointb,
				    src_polys[i].verts[map4[j][0]],
				    src_polys[i].verts[map4[j][1]]) &&
		 i != current_polygon)
		return FALSE;
	}
    }

    return TRUE;
}

static void store_segment(pointa, pointb)
    float32      pointa[2];
    float32      pointb[2];
{
    bcopy(pointa, segments[n_segments].points[0], 2 * sizeof(float32));
    bcopy(pointb, segments[n_segments].points[1], 2 * sizeof(float32));

    segments[n_segments].used = FALSE;

    n_segments++;
}

static int32 find_segment_containing_point(search_point, last_point,
					   next_point)
    float32      search_point[2];
    float32      last_point[2];
    float32      next_point[2];
{
    int32        i, j, n_angles = 0, max_angle_segment_index;
    float32      angles[100], max_angle = -1.0;
    float32      possible_next_point[100][2];
    int32        segment_indices[100];

    for(i = 0; i < n_segments; i++)
      for(j = 0; j < 2; j++)
	/* If multiple segments contain `search_point', pick the one such that
	 * the angle between it and the current segment is a maximum.
	 */
	if(!segments[i].used &&
	   !bcmp(search_point, segments[i].points[j], 2 * sizeof(float32)))
	{
	    if(!last_point)
	      if(j == 0)
	      {
		  bcopy(segments[i].points[1], next_point, 2*sizeof(float32));
		  segments[i].used = TRUE;
		  return TRUE;
	      }
	    else
	      {
		  bcopy(segments[i].points[0], next_point, 2*sizeof(float32));
		  segments[i].used = TRUE;
		  return TRUE;
	      }

	    if(j == 0)
	    {
		angles[n_angles] =
		  vmat2e_angle_between_vectors32(
					       last_point[X] - search_point[X],
					       last_point[Y] - search_point[Y],
					       segments[i].points[1][X] -
					       segments[i].points[0][X],
					       segments[i].points[1][Y] -
					       segments[i].points[0][Y]);

		bcopy(segments[i].points[1], possible_next_point[n_angles],
		      2 * sizeof(float32));
	    }
	    else
	    {
		angles[n_angles] =
		  vmat2e_angle_between_vectors32(
					       last_point[X] - search_point[X],
					       last_point[Y] - search_point[Y],
					       segments[i].points[0][X] -
					       segments[i].points[1][X],
					       segments[i].points[0][Y] -
					       segments[i].points[1][Y]);

		bcopy(segments[i].points[0], possible_next_point[n_angles],
		      2 * sizeof(float32));
	    }

	    if(angles[n_angles] < 0.0)
	      angles[n_angles] += TWO_PI;

	    segment_indices[n_angles] = i;

	    n_angles++;
	}

    if(!n_angles)
      return FALSE;

    for(i = 0; i < n_angles; i++)
      if(angles[i] > max_angle)
      {
	  max_angle = angles[i];
	  max_angle_segment_index = segment_indices[i];
	  bcopy(possible_next_point[i], next_point, 2 * sizeof(float32));
      }

    segments[max_angle_segment_index].used = TRUE;

    return TRUE;
}

static void add_point_to_polygon(output_polys, point, slot)
    CTDB_COALESCE_POLYGON *output_polys[];
    float32                point[2];
    int32                  slot;
{
    CTDB_COALESCE_POLYGON *poly = output_polys[slot];

    /* Check to see if this is the first point to be added to the linked list.
     */
    if(!poly)
    {
        output_polys[slot] = (CTDB_COALESCE_POLYGON *)
	  STDALLOC(sizeof(CTDB_COALESCE_POLYGON));
        bzero(output_polys[slot], sizeof(CTDB_COALESCE_POLYGON));
        bcopy(point, (output_polys[slot])->vertex, 2 * sizeof(float32));
        return;
    }

    while(poly->next)
      poly = poly->next;

    poly->next = (CTDB_COALESCE_POLYGON *)
      STDALLOC(sizeof(CTDB_COALESCE_POLYGON));
    bzero(poly->next, sizeof(CTDB_COALESCE_POLYGON));
    bcopy(point, poly->next->vertex, 2 * sizeof(float32));
}

void ctdb_coalesce_deallocate_polygon(output_polys, polygon_number)
    CTDB_COALESCE_POLYGON *output_polys[];
    int32                  polygon_number;
{
    CTDB_COALESCE_POLYGON *poly, *next;

    poly = output_polys[polygon_number];

    while(poly)
    {
	next = poly->next;
	STDDEALLOC(poly);
	poly = next;
    }

    output_polys[polygon_number] = NULL;
}

/* A polygon is considered "valid" if not all of the X coordinates of its
 * vertices are the same.  The same goes for the Y coordinates.
 */
#define DIFFERENCE_THRESHOLD  1.0
static int32 polygon_is_valid(output_polys)
    CTDB_COALESCE_POLYGON *output_polys[];
{
    CTDB_COALESCE_POLYGON *poly;
    float32                value;

    poly = output_polys[n_polygons];
    value = poly->vertex[X];
    poly = poly->next;

    while(poly)
    {
	if(fabs(value - poly->vertex[X]) > DIFFERENCE_THRESHOLD)
	  break;
	poly = poly->next;
    }

    if(!poly)
      return FALSE;

    poly = output_polys[n_polygons];
    value = poly->vertex[Y];
    poly = poly->next;

    while(poly)
    {
	if(fabs(value - poly->vertex[Y]) > DIFFERENCE_THRESHOLD)
	  return TRUE;
	poly = poly->next;
    }

    return FALSE;
}

static void create_polygon_from_segments(output_polys, starting_point)
    CTDB_COALESCE_POLYGON *output_polys[];
    float32                starting_point[2];
{ 
    float32                next_point[2];
    float32                last_point[2];
    float32                current_point[2];

    if(!find_segment_containing_point(starting_point, NULL, next_point))
      return;

    add_point_to_polygon(output_polys, starting_point, n_polygons);

    bcopy(starting_point, last_point, 2 * sizeof(float32));
    bcopy(next_point, current_point, 2 * sizeof(float32));

    do
    {
	add_point_to_polygon(output_polys, current_point, n_polygons);

	if(!find_segment_containing_point(current_point, last_point,
					  next_point))
	  /* This polygon is not closed, so close it. */
	{
	    if(polygon_is_valid(output_polys))
	      n_polygons++;
	    else
	      ctdb_coalesce_deallocate_polygon(output_polys, n_polygons);
	    return;
	}

	bcopy(current_point, last_point, 2 * sizeof(float32));
	bcopy(next_point, current_point, 2 * sizeof(float32));
    } while(bcmp(starting_point, current_point, 2 * sizeof(float32)));

    if(polygon_is_valid(output_polys))
      n_polygons++;
    else
      ctdb_coalesce_deallocate_polygon(output_polys, n_polygons);
}

/* Take a float64 array of vertices and store that polygon at the specified
 * slot in the `output_polys' list.
 */
void ctdb_coalesce_store_polygon(output_polys, slot, poly, n_verts)
    CTDB_COALESCE_POLYGON *output_polys[];
    int32                  slot;
    float64                poly[];
    int32                  n_verts;
{
    int32         i;
    float32       point[2];

    for(i = 0; i < n_verts; i++)
    {
	point[X] = poly[2 * i + X];
	point[Y] = poly[2 * i + Y];
	add_point_to_polygon(output_polys, point, slot);
    }
}

/* Merge the 3-sided or 4-sided polygons that border each other into larger
 * n-sided polygons.
 *
 * This is done using the following algorithm:
 *
 * 1) Compute and store all segments that belong to only one polygon.  These
 *    are the segments which will make up the coalesced polygons.
 *
 * 2) Select any unused segment and create a polygon.  To create a polygon, a
 *    search is made for all segments containing a point which is the same as
 *    the other point in the segment containing the starting point.  Once those
 *    segments are found, the angle between each of them and the last segment
 *    to be placed in the polygon is found.  The segment with the greatest
 *    angle is selected as the next segment in the polygon.  Next, a search is
 *    made for a segment which contains the other point in the current segment.
 *    This process is repeated until a segment that contains the starting point
 *    is found.  Each segment that is added to the polygon is marked as "used",
 *    and will never be used again.
 *
 * 3) After a polygon has been constructed, another unused segment is found
 *    and step 2 repeats, until there are no more unused segments.
 *
 * Arguments:  src_polys, n_src_polys - specify the polygons to coalesce
 *             output_polys - array of pointers to the coalesced polys
 *             n_output_polys - number of coalesced polys
 * 
 * NOTE:  This function returns pointers to dynamically-allocated memory.
 *        This memory must be freed after it is used with
 *        `ctdb_coalesce_deallocate_polygon()':
 *
 *        for(i = 0; i < n_output_polys; i++)
 *          ctdb_coalesce_deallocate_polygon(output_polys, i);
 *
 */
void ctdb_coalesce_polygons(src_polys, n_src_polys, output_polys,
			    n_output_polys)
    CTDB_COALESCE_SRC_POLY  src_polys[];
    int32                   n_src_polys;
    CTDB_COALESCE_POLYGON  *output_polys[];
    int32                  *n_output_polys;
{
    int32                   src_poly, segment, i;
    int32                   max_output_polys = *n_output_polys;

    n_segments = 0;
    n_polygons = 0;

    for(i = 0; i < max_output_polys; i++)
      output_polys[i] = NULL;

    /* Compute and store all segments that belong to only one polygon.
     */
    printf("Finding edges...");  fflush(stdout);

    for(src_poly = 0; src_poly < n_src_polys; src_poly++)
    {
	if(NUM_SIDES(src_poly) == 3)
	{
	    for(segment = 0; segment < 3; segment++)
	      if(segment_is_unique(src_poly, src_polys, n_src_polys,
				   src_polys[src_poly].verts[map3[segment][0]],
				  src_polys[src_poly].verts[map3[segment][1]]))
		store_segment(src_polys[src_poly].verts[map3[segment][0]],
			      src_polys[src_poly].verts[map3[segment][1]]);
	}
	else
	{
	    for(segment = 0; segment < 4; segment++)
	      if(segment_is_unique(src_poly, src_polys, n_src_polys,
				   src_polys[src_poly].verts[map4[segment][0]],
				  src_polys[src_poly].verts[map4[segment][1]]))
		store_segment(src_polys[src_poly].verts[map4[segment][0]],
			      src_polys[src_poly].verts[map4[segment][1]]);
	}

	if (!(src_poly % 100))
	  putchar('.'); fflush(stdout);
    }

    printf("\n");

    /* Create polygons until there are no unused segments.
     */
    for(segment = 0; segment < n_segments; segment++)
    {
	if(segments[segment].used)
	  continue;

	create_polygon_from_segments(output_polys,
				     segments[segment].points[0]);

	if(n_polygons >= max_output_polys)
	{
	    printf("ct_coalesce.c: `max_output_polys' exceeded.  Increase your array size.\n");
	    exit(1);
	}
    }

    *n_output_polys = n_polygons;
}

