#ifndef lint
static char rcsid [] = "$RCSfile: ct_gcs.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_gcs.c                                                          *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *                                                                           *
 *       Contents  : This file contains those utils required by compiler     *
 *                   front-ends to generate GCS outputs.                     *
 *       Created   : December 1995                                           *
 *       Author    : flchamberlain                                           *
 ****************************************************************************/

#include <libcoordinates.h>
#include <string.h>
#include <libctdb.h>
#include <sys/stat.h>
#include "ct_cmplr.h"
#include "ct_reform.h"

/* Return the minimal bounding box in GCS local coordinates that 
 * contains every point in the specified lat/long range.
 */
void
find_gcs_bbox_from_ll(south_lat, north_lat, west_long, east_long,
		      cell, min_x, max_x, min_y, max_y)
    float64 south_lat, north_lat, west_long, east_long;
    int32   cell;
    float64 *min_x, *max_x, *min_y, *max_y;
{
#define UPDATE_BOUNDS \
{                     \
    if (lx < *min_x)  \
      *min_x = lx;    \
    if (lx > *max_x)  \
      *max_x = lx;    \
    if (ly < *min_y)  \
      *min_y = ly;    \
    if (ly > *max_y)  \
      *max_y = ly;    \
}

    /* Find the coordinates of the four lat/long corners, plus the top 
     * and bottom center points, and take the extrema to generate the bbox
     */
    float64 gx, gy, gz;
    float64 lx, ly, lz;

    /* Upper left corner */
    if (coord_convert(COORD_LATLON,
		      south_lat, west_long, 0.0, FALSE,
		      COORD_GCC,
		      &gx, &gy, &gz))
    {
	printf("Error converting lat/long to GCC\n");
	exit(4);
    }


    if (coord_convert(COORD_GCC, gx, gy, gz,
		      COORD_GCSCS,
		      cell, &lx, &ly, &lz))
    {
	printf("Error converting from GCC to GCS.\n");
	exit(-1);
    }
    
    *min_x = *max_x = lx;
    *min_y = *max_y = ly;

    /* Lower right corner */
    if (coord_convert(COORD_LATLON,
		      north_lat, west_long, 0.0, FALSE,
		      COORD_GCC,
		      &gx, &gy, &gz))
    {
	printf("Error converting lat/long to GCC\n");
	exit(4);
    }    

    if (coord_convert(COORD_GCC, gx, gy, gz,
		      COORD_GCSCS, cell, &lx, &ly, &lz))
    {
	printf("Error converting from GCC to GCS.\n");
	exit(-1);
    }
    
    UPDATE_BOUNDS;
    
    /* Upper left corner */
    if (coord_convert(COORD_LATLON,
		      south_lat, east_long, 0.0, FALSE,
		      COORD_GCC,
		      &gx, &gy, &gz))
    {
	printf("Error converting lat/long to GCC\n");
	exit(4);
    }
    if (coord_convert(COORD_GCC, gx, gy, gz,
		      COORD_GCSCS, cell, &lx, &ly, &lz))
    {
	printf("Error converting from GCC to GCS.\n");
	exit(-1);
    }
    UPDATE_BOUNDS;
 
    /* Upper right corner */
    if (coord_convert(COORD_LATLON,
		      north_lat, east_long, 0.0, FALSE,
		      COORD_GCC,
		      &gx, &gy, &gz))
    {
	printf("Error converting lat/long to GCC\n");
	exit(4);
    }
    if (coord_convert(COORD_GCC, gx, gy, gz,
		      COORD_GCSCS, cell, &lx, &ly, &lz))
    {
	printf("Error converting from GCC to GCS.\n");
	exit(-1);
    }
    UPDATE_BOUNDS;
 
    /* Bottom center */
    if (coord_convert(COORD_LATLON,
		      south_lat, (east_long + west_long)/2.0, 0.0, FALSE,
		      COORD_GCC,
		      &gx, &gy, &gz))
    {
	printf("Error converting lat/long to GCC\n");
	exit(4);
    }
    if (coord_convert(COORD_GCC, gx, gy, gz,
		      COORD_GCSCS, cell, &lx, &ly, &lz))
    {
	printf("Error converting from GCC to GCS.\n");
	exit(-1);
    }
    UPDATE_BOUNDS;

    /* Top center */
    if (coord_convert(COORD_LATLON,
		      north_lat, (east_long + west_long)/2.0, 0.0, FALSE,
		      COORD_GCC,
		      &gx, &gy, &gz))
    {
	printf("Error converting lat/long to GCC\n");
	exit(4);
    }
    if (coord_convert(COORD_GCC, gx, gy, gz,
		      COORD_GCSCS, cell, &lx, &ly, &lz))
    {
	printf("Error converting from GCC to GCS.\n");
	exit(-1);
    }
    UPDATE_BOUNDS;

    return;
}

/* Generate the bounds on patch x (min_px, max_px) and patch y (min_py,
 * max_py) in ctdb (which must be TCC-based) that fully contain (x1, y1)
 * (x2, y2) in cell cell.
 */
void
get_tcc_patch_list_from_gcs(ctdb, cell, tcc, x1, x2, y1, y2, 
			    min_px, max_px, min_py, max_py)
    CTDB_CMP      *ctdb;
    int32          cell;
    COORD_TCC_PTR  tcc;
    float64        x1, x2, y1, y2;    
    int32         *min_px, *max_px, *min_py, *max_py;
{
    float64 ux1, ux2, uy1, uy2;
    int32 minpx, maxpx, minpy, maxpy;
    float64 patch_conv = ctdb->inv_incr * ctdb->patch_inv_incr;
    int32 dbminpx, dbmaxpx, dbminpy, dbmaxpy;

    find_tcc_bbox_from_gcs(cell, x1, x2, y1, y2,
			   tcc, &ux1, &ux2, &uy1, &uy2);

    minpx = PATCH_X(ux1 * patch_conv);
    maxpx = PATCH_X(ux2 * patch_conv);
    minpy = PATCH_Y(uy1 * patch_conv);
    maxpy = PATCH_Y(uy2 * patch_conv);


    dbminpx = ctdb->min_x_post/ctdb->patch_incr;
    dbmaxpx = ctdb->max_x_post/ctdb->patch_incr;
    dbminpy = ctdb->min_y_post/ctdb->patch_incr;
    dbmaxpy = ctdb->max_y_post/ctdb->patch_incr;

    /* If we're off the input database, set values s.t. anything that
     * loops through them won't do anything...
     */
    if ((maxpx < dbminpx) ||
	(minpx > dbmaxpx) ||
	(maxpy < dbminpy) ||
	(minpy > dbmaxpy))
    {
	*min_px = 0;
	*max_px = -1;
	*min_py = 0;
	*max_py = -1;
	return;
    }
 
    *min_px = NS_MAX(minpx, dbminpx);
    *max_px = NS_MIN(maxpx, dbmaxpx);
    *min_py = NS_MAX(minpy, dbminpy);
    *max_py = NS_MIN(maxpy, dbmaxpy);
    return;
}

void
find_tcc_bbox_from_gcs(cell, min_gx, max_gx, min_gy, max_gy,
		       tcc, min_ux, max_ux, min_uy, max_uy)
    int32         cell;
    float64       min_gx, max_gx, min_gy, max_gy;
    COORD_TCC_PTR tcc;
    float64      *min_ux, *max_ux, *min_uy, *max_uy;
{
    float64 x1, y1, z1;		/* TCC llc */
    float64 x2, y2, z2;		/* TCC ulc */
    float64 x3, y3, z3;		/* TCC lrc */
    float64 x4, y4, z4;		/* TCC urc */

    /* This is not perfectly accurate, but it'll do */
    if (coord_convert(COORD_GCS, cell, min_gx, min_gy, (float64)0.0,
		      COORD_TCC, tcc, &x1, &y1, &z1))
    {
	printf("Error converting GCS to TCC.\n");
	exit(-1);
    }
    
    /* This is not perfectly accurate, but it'll do */
    if (coord_convert(COORD_GCS, cell, min_gx, max_gy, (float64)0.0,
		      COORD_TCC, tcc, &x2, &y2, &z2))
    {
	printf("Error converting GCS to TCC.\n");
	exit(-1);
    }

    /* This is not perfectly accurate, but it'll do */
    if (coord_convert(COORD_GCS, cell, max_gx, min_gy, (float64)0.0,
		      COORD_TCC, tcc, &x3, &y3, &z3))
    {
	printf("Error converting GCS to TCC.\n");
	exit(-1);
    }

    /* This is not perfectly accurate, but it'll do */
    if (coord_convert(COORD_GCS, cell, max_gx, max_gy, (float64)0.0,
		      COORD_TCC, tcc, &x4, &y4, &z4))
    {
	printf("Error converting GCS to TCC.\n");
	exit(-1);
    }

    *min_ux = NS_MIN(x1, x2);
    *max_ux = NS_MAX(x3, x4);
    *min_uy = NS_MIN(y1, y3);
    *max_uy = NS_MAX(y2, y4);
    return;
}

/* Convert points to GCS.  Note that since these are 2D, we need to
 * to get a Z value (we'll use the terrain surface, since that's usually
 * what was meant)
 */
void
translate_verts_2D_tcc_to_gcs32(tcc, cell, points, n_points, lookup_fn)
    COORD_TCC_PTR tcc;
    int32         cell;
    float32      *points;
    int32         n_points;
    int32       (*lookup_fn)(/* float32, float32, float32 * */);
{
    int32 i;
    int32 index;
    float64 x, y, z;
    float32 z32;

    for(i=0;i<n_points;i++)
    {
	index = 2*i;

	if (lookup_fn(points[index + X], points[index + Y], &z32))
	{
	    z = z32;
	}
	else
	{
	    z = 0.0;
	}
	
	if (coord_convert(COORD_TCC,
			  tcc, (float64)points[index + X], 
			  (float64)points[index + Y],
			  z,
			  COORD_GCSCS,
			  cell, &x, &y, &z))
        {
	    printf("Error in translate poly TCC=>GCS\n");
	    exit(7);
	}

	points[index + X] = x;
	points[index + Y] = y;
    }
    return;
}

/* Convert points to GCS.  Note that since these are 2D, we need to
 * to get a Z value (we'll use the terrain surface, since more than likely,
 * that's what was meant)
 */
void
translate_verts_2D_tcc_to_gcs(tcc, cell, points, n_points, lookup_fn)
    COORD_TCC_PTR tcc;
    int32         cell;
    float64       points[][2];
    int32         n_points;
    int32       (*lookup_fn)(/* float32, float32, float32 * */);
{
    float32 z;
    int32 i;
    float64 trash;

    for(i=0;i<n_points;i++)
    {
	if (!lookup_fn(points[i][X], points[i][Y], &z))
	{
	    z = 0.0;
	}

	if (coord_convert(COORD_TCC,
			  tcc, (float64)points[i][X], 
			  (float64)points[i][Y],
			  (float64)z,
			  COORD_GCSCS,
			  cell, &points[i][X], &points[i][Y], &trash))
        {
	    printf("Error in translate poly TCC=>GCS\n");
	    exit(7);
	}
    }
    return;
}

/* Convert elements of points list from TCC to GCS in the specified cell */
void
translate_verts_tcc_to_gcs(tcc, cell, points, n_points)
    COORD_TCC_PTR tcc;
    int32         cell;
    float64       points[][3];
    int32         n_points;
{
    int32 i;
    float64 gx, gy, gz;

    for(i=0;i<n_points;i++)
    {
	if (coord_convert(COORD_TCC,
			  tcc, (float64)points[i][X], 
			  (float64)points[i][Y],
			  (float64)points[i][Z],
			  COORD_GCC,
			  &gx, &gy, &gz))
        {
	    printf("Error in translate poly TCC=>GCC\n");
	    exit(7);
	}

	if (coord_convert(COORD_GCC, gx, gy, gz,
			  COORD_GCSCS,
			  cell, &points[i][X], &points[i][Y], 
			  &points[i][Z]))
	{
	    printf("Error converting GCC to GCS.\n");
	    exit(-1);
	}
    }
    return;
}

/* Converts min/max x/y in TCC or GCS into lat/long boundaries.  Returns TRUE 
 * for success, FALSE for failure.  Based on gcs_ll_bounds_from_gcs in libgcs.
 * mode is one of {COORD_GCS, COORD_TCC}
 */
int32
get_ll_bounds(mode, cell, tcc, min_x, min_y, max_x, max_y, 
	      west_lng, south_lat, east_lng, north_lat)
    COORD_SYSTEM  mode;
    int32         cell;
    COORD_TCC_PTR tcc;
    float64       min_x;
    float64       min_y;
    float64       max_x;
    float64       max_y;
    float64      *west_lng;
    float64      *south_lat;
    float64      *east_lng;
    float64      *north_lat;
{
    float64 lat1, lat2, avg_lat;
    float64 lng1, lng2;
    float64 dummy;
    int32 ret;

    if ((mode != COORD_TCC) && (mode != COORD_GCS))
    {
	printf("Error:  get_ll_bounds called with illegal source system.\n");
	exit(-1);
    }

#define CONVERT(x,y,wlat,wlong)                               \
    if (mode == COORD_GCS)                                    \
    {                                                         \
	ret = coord_convert(COORD_GCS,                        \
			    cell, (x), (y), (float64)0.0,     \
			    COORD_LATLON,                     \
			    &(wlat), &(wlong), &dummy, FALSE);\
    }                                                         \
    else                                                      \
    {                                                         \
        ret = coord_convert(COORD_TCC,                        \
			    tcc, (x), (y), (float64)0.0,      \
			    COORD_LATLON,                     \
			    &(wlat), &(wlong), &dummy, FALSE);\
    }                                                         \
    if (ret)                                                  \
    {                                                         \
        printf("Error converting to lat/long.\n");            \
	exit(2);                                              \
    }


    /* SW corner */
    CONVERT(min_x, min_y, lat1, lng1);
    *south_lat = lat1;

    /* NW corner */
    CONVERT(min_x, max_y, lat2, lng2);
    *north_lat = lat2;
    
    /* The way the GCS coordinate systems are set up, the easiest way to
     * determine which points will specify the min and max longitudes is
     * to see whether we're mostly north or south of the equator
     */
    avg_lat = (*north_lat + *south_lat)/2.0;
    if (avg_lat >= 0.0)
    {
	*west_lng = lng2;

	/* SE corner*/
	CONVERT(max_x, min_y, lat1, lng1);
	*south_lat = NS_MIN(*south_lat, lat1);

	/* NE corner */
	CONVERT(max_x, max_y, lat1, lng1);
	*east_lng = lng1;
	*north_lat = NS_MAX(*north_lat, lat1);
    }
    else
    {
	*west_lng = lng1;

	/* SE corner*/
	CONVERT(max_x, min_y, lat1, lng1);
	*east_lng = lng1;
	*south_lat = NS_MIN(*south_lat, lat1);

	/* NE corner */
	CONVERT(max_x, max_y, lat1, lng1);
	*north_lat = NS_MAX(*north_lat, lat1);
    }

    return TRUE;
}

/* Given the input file name, make a best guess at a desirable output file
 * name, and put that in out_name.
 */
void
generate_name_base(in_name, out_name)
    char *in_name;
    char *out_name;
{
    char *base;
    char *cur;
    char *last;
    
    /* Figure out the base terrain name */
    last = cur = base = strdup(in_name);
    
    cur = strtok(cur, "/");
    while(cur)
    {
	last = cur;
	cur = strtok(NULL, "/");
    }
    
    strtok(last, ".");

    strcpy(out_name, last);

    free(base);
    
    return;
}

/* Given a 2D point in GCS, we want to find the elevation in GCS--i.e., the
 * z value at which a line pointing straight up ("up" defined as parallel to
 * the z axis) intersects the terrain.  Since the z vector in GCS is not
 * the same as that in UTM, we have to actually look for this intersection 
 * point rather than performing a trivial elevation lookup.  This algorithm is 
 * based on two assumptions:
 *    1.  The terrain is relatively flat
 *    2.  The difference between the GCS and UTM z-vectors is not great
 *
 * Using those assumptions, we do the following:
 *    1.  Generate a line segment pointing straight up at the desired point
 *        in GCS, and translate the endpoints to UTM
 *    2.  Do a (UTM) elevation lookup at the x,y of one endpoint of the
 *        segment.
 *    3.  Determine the x,y of the segment at that z.
 *    4.  Perform an elevation lookup at that x,y.  If the result is 
 *        "close enough" the last z we saw, we're done.
 *    5.  If the result is not good enough, do another lookup at the x,y
 *        along the segment of the new z.  Iterate as necessary.
 *
 */
int32
get_elev_at_gcs_pt(tcc, cell, x, y, z, lookup_fn)
    COORD_TCC_PTR   tcc;
    int32           cell;
    float64         x;
    float64         y;
    float64        *z;
    int32         (*lookup_fn)(/* float32, float32, float32 * */);
{
    float64 height = 1000.0;	/* Height, in meters of the segment */
    float64 x1, y1, z1, x2, y2, z2;
    float64 dx, dy, dz;
    float64 t;
    float32 elev, last_elev;
    float64 curx, cury;
    int32 ret;
    float64 delta;
    int32 count = 0;

    /* Translate (x,y,0) and (x,y,height) into TCC */
    if (ret = coord_convert(COORD_GCS,
			    cell, x, y, (float64)0.0,
			    COORD_TCC,
			    tcc, &x1, &y1, &z1))
    {
	printf("Error converting from GCS to TCC.\n");
	exit(-1);
    }
    
    if (ret = coord_convert(COORD_GCS,
			    cell, x, y, height,
			    COORD_TCC,
			    tcc, &x2, &y2, &z2))
    {
	printf("Error converting from GCS to TCC.\n");
	exit(-1);
    }

    /* Find parametric equations for segment x = tdx + x1, y = tdy + y1,
     * t = (z - z1)/dz (we can get away with this because the segment will
     * always have a non-zero z component)
     */
    dx = x2 - x1;
    dy = y2 - y1;
    dz = z2 - z1;

    if (!lookup_fn(x1, y1, &last_elev))
    {
	return FALSE;
    }
    
    delta = MAX_DELTA + 1.0;
    while(delta > MAX_DELTA)
    {
	t = (last_elev - z1)/dz;
	curx = x1 + (t * dx);
	cury = y1 + (t * dy);
	if (!lookup_fn(curx, cury, &elev))
	{
	    return FALSE;
	}

	/* Avoid spinning indefinitely at the edge of the database.  In this
	 * case we return last_elev to generate as smooth a boundary as
	 * possible
	 */
	if ((elev == 0.0) && (last_elev > MAX_DELTA))
	{
	    *z = last_elev;
	    return TRUE;
	}
	
	delta = fabs((float64)(elev - last_elev));
	last_elev = elev;
	if (count++ > 10)
        {
	    printf("z lookup failed to converge at (%.3f, %.3f)!\n",
		   x, y);
	    return FALSE;
	}
    }

    /* The original length of the vector was height, the intersection occurred
     * at t * vector length, so the desired z value is t*height.  This
     * ignores length warping due to the coordinate transformation, but so
     * does the rest of this routine
     */
    *z = t * height;
    return TRUE;
}

/* Special case of above where the TCC elevation is known to be constant in
 * in the region where we're looking--this is primarily useful when dealing
 * with oceans
 */
float64
get_elev_at_gcs_pt_plane(tcc, cell, x, y, tcc_z)
    COORD_TCC_PTR   tcc;
    int32           cell;
    float64         x;
    float64         y;
    float64         tcc_z;
{
    float64 height = 1000.0;	/* Height, in meters of the segment */
    float64 x1, y1, z1, x2, y2, z2;
    float64 dx, dy, dz;
    float64 t;
    float32 elev, last_elev;
    float64 curx, cury;
    int32 ret;
    float64 delta;
    int32 count = 0;

    /* Translate (x,y,0) and (x,y,height) into TCC */
    if (ret = coord_convert(COORD_GCS,
			    cell, x, y, (float64)0.0,
			    COORD_TCC,
			    tcc, &x1, &y1, &z1))
    {
	printf("Error converting from GCS to TCC.\n");
	exit(-1);
    }
    
    if (ret = coord_convert(COORD_GCS,
			    cell, x, y, height,
			    COORD_TCC,
			    tcc, &x2, &y2, &z2))
    {
	printf("Error converting from GCS to TCC.\n");
	exit(-1);
    }

    dz = z2 - z1;

    t = (tcc_z - z1) / dz;
    
    return (t * height);
}


/* Generates the directory used to store multi-cell terrain
 * from the multi-cell header name.  
 */
void
ctdb_generate_multi_directory(dbname)
  char *dbname;
{
  char *mp_ptr,multi_path[256];

  strcpy(multi_path,dbname);
  
  /* find last / */
  mp_ptr = (char *)strrchr(multi_path,'/');
  /* add a null inplace of / to generate the directory
   * name to be created.
   */
  *mp_ptr='\0';
  printf("%s\n",multi_path);
  mkdir(multi_path,S_ISUID | S_ISGID | S_IRWXU | S_IRWXG | S_IRWXO);
}

