/*
 * $RCSfile: ct_reform.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: ct_reform.h                                                       *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations for format 4+ compiler backend
 *       Created   : January 1995                                            *
 *       Author    : wkbraudway      (TASC)                                  *
 *                                                                           *
 *       Remarks:  This file contains the declarations for use by the        *
 *                 recompiler and compiler backend.                          *
 *                                                                           *
 ****************************************************************************/
#ifndef _REFORM_H_INCLUDED
#define _REFORM_H_INCLUDED
#include <stdtypes.h>
#include <libcoordinates.h>
#include <libreader.h>
#include "../libctdb_local.h"
#include "ct_cmprep.h"

#define CTDB_REFORM_DEBUG 0 /* DEBUGGING CODE  0=OFF 1=ON */
#define CTDB_TIN_POLY 1               /* TE must be able to represent a TIN */
#define CTDB_TIN_VERTEX 0             /* TE only has to represent a vertex */
#define CTDB_MAX_FEATURE_WORDS_PER_PATCH  2000  
#define CTDB_RND(x)  (int32)((x) + 0.5)  /* Macro function to round a float */
#define CTDB_MAX_INTERSECTION_PTS    8 /* Really 7 but one for good measure */
#define CTDB_PTS_PER_QUADRALATERAL   4 /* vertices per quad */
#define CTDB_PTS_PER_TRIANGLE        3 /* vertices per triangle */
#define CTDB_UTM_DATABASE -1

/* CTDB_AxB define returns cross product of two vectors where (a.x,a.y) and 
 * (b.x,b.y) define the endpoints of the vectors which contain the same 
 * anchor point (i.e. sin of the angle between them).  If the result
 * is negative then the second endpoint is to the left of the first 
 * point, if 0 then the point is colinear, else the second point lays 
 * to the left of the first point.
 * 
 */
#define CTDB_AxB(a,b)   (a.x*b.y - a.y*b.x)

#define CTDB_POST_IN_CELL(db, px, py) (((px) >= db->min_x_post) && \
				       ((px) <= db->max_x_post) && \
				       ((py) >= db->min_y_post) && \
				       ((py) <= db->max_y_post))

#define CTDB_EXCLUSIVE_BETWEEN(a, x, b) ((a) < (b) ? (((a) < (x)) && \
						      ((x) < (b))) \
				     : (((b) < (x)) && ((x) < (a))))

typedef struct area            
{                              
    int32         count;         /* number of points in area */
    int32         limit;
    CTDB_COORD2D  extent_upperleft; 
    CTDB_COORD2D  extent_lowerright;
    CTDB_COORD2D* pts;           /* points making up area */
} AREA; 
                        

typedef struct transition_patch_info
{
    int32 patch_number;
    int32 trans_patch;    /* flag set if patch is a transition patch */
    int32 delta_te;       /* number of added or removed tes */
} TRANSITION_PATCH_INFO;


typedef enum ctdb_boundary_type
{
    PB_NORTH,
    PB_SOUTH,
    PB_EAST,
    PB_WEST
} CTDB_BOUNDARY_TYPE;

typedef enum ctdb_dbase_type
{
  CTDB_GRIDDED,
  CTDB_TINNED,
  CTDB_HYBRID
} CTDB_DBASE_TYPE;

/* Working data structure for working it patch vertices of a TIN */
#define CTDB_MAX_PTS_PER_PATCH 2048
typedef struct ctdb_vertex
{
    int32               number;  /* number for TE assignment */
    CTDB_XYO_COORD2D    pt;      /* vertex point */
    float32             z;       /* vertex elevation */
    uint8               soil0;   /* Triangle soils */
    uint8               soil1;
    int32               n_edges; /* number edges out of vertex */
    int32               edges[CTDB_MAX_PTS_PER_PATCH];   /* list of edges */
    int32               visited; /* marker for vertex/edge search routine */
    int32               count;   /* number of unnumbered edges */
} CTDB_VERTEX;

/* Triangulatization area */
typedef struct area_2d
{                              
    int32             count;   
    int32             size;
    CTDB_COORD2D      extent_upperleft;
    CTDB_COORD2D      extent_lowerright;
    CTDB_COORD2D     *pts; 
} AREA_2D;            

/* Triangulization area list */
typedef struct area_elem 
{ 
    AREA_2D           anarea; 
    struct area_elem *next; 
} AREA_ELEM;

/* Triangulatization overlays of areas */
typedef struct overlay_2d
{
    AREA_ELEM    *areas;    
    int32         count;
    CTDB_COORD2D  extent_upperleft;
    CTDB_COORD2D  extent_lowerright;
} OVERLAY_2D;

typedef CTDB_COORD2D CTDB_DB_EDGE[2]; 

/* TERRAIN ELEMENT STRUCTURES */

/* Working TE adjacency - topology information */
typedef struct te_adj_info
{
  uint16 minx;   /* TE edge Min x */
  uint16 miny;   /* TE edge Min Y */
  uint16 maxx;   /* TE edge Max x */
  uint16 maxy;   /* TE edge Max y */
  uint16 te_index;  /* TE index */
  uint16 entry_side;      /* TE edge side number */
} TE_ADJ_INFO;

typedef struct patch_adj_info
{
  TE_ADJ_INFO  *pa_east;   /* TE adj info list */
  TE_ADJ_INFO  *pa_west;   /* TE adj info list */
  TE_ADJ_INFO  *pa_north;   /* TE adj info list */
  TE_ADJ_INFO  *pa_south;   /* TE adj info list */
  uint16        east_end;   /* TE adj info list length */
  uint16        west_end;   /* TE adj info list length */
  uint16        north_end;   /* TE adj info list length */
  uint16        south_end;   /* TE adj info list length */
} PATCH_ADJ_INFO;

#define CTDB_TE_08_TO_16_BIT_CROSSOVER      (uint8)-3
#define CTDB_INCLUSIVE_MAX_POSTS_PER_PATCH 36

typedef struct ctdb_post_info
{
    int32 x;
    int32 y;
    float64 z;
    int32 soil0;
    int32 soil1;
    int32 tinned;
    int32 is_sw_ne;
}CTDB_POST_INFO;

/* -------------------------------------------------------------------
   FUNCTION PROTOTYPES
   ------------------------------------------------------------------- */

/* --------------------------------------------------------------------
 *         ROUTINES DEFINED IN ct_tecmplr.c
 * -------------------------------------------------------------------- */

/* The ctdb_covert_to_tins routine will reformat the list of microterrain
   features for a single patch into a list of terrain elements and record 
   that list in a specified temporary file.   This routine will upon
   request correct the tinning the current patch.   This routine will also
   flag this patch (transition) if the patch is not completely tinned
   (i.e., some post grid cell does not contain microterrain).
 */

extern void 
ctdb_convert_to_tins (
		      CTDB_CMP              *ctdb,
		      int32                  patch_number,
		      CTDB_FEATURE_DATA_CMP     *feature,
		      int32                  n_feat,
		      int32                 *n_te,
		      TRANSITION_PATCH_INFO *trans_patches,
		      int32                 *trans_end,
		      float64               *max_elev,
		      int32                  corrections,
		      int32                  new_micro,
		      FILE                  *fptr
		      );

/* 
   The ctdb_set_te_patch_header routine updates the patch_group header
   information for the terrain elements added to the indicated patch.  This
   routine determines if the patch is special (either a transition patch or a
   gridded patch of a database to be completely tinned), sets the patch te
   offsets, determines if the current patch is an 8 or 16 bit indexed patch,
   and updates the feature present bits for the current patch. */

extern void
ctdb_set_te_patch_header (
			     CTDB_CMP         *ctdb,
			     int32                  patch_x,
			     int32                  patch_y,
			     int32                  offset,
			     int32                 *num_te,
			     float64                max_elev,
			     TRANSITION_PATCH_INFO *trans_patches,
			     int32                 *trans_end,
			     CTDB_FEATURE_DATA_CMP *features,
			     int32                  n_features,
			     CTDB_DBASE_TYPE        dbtype
			     );


/* After all patches have been processed for terrain elements, the
 * ctdb_te_dbase_processing routine loads the tes from their temporary file,
 * creates transition patches where needed and constructs the te topology. 
 * If fix_holes is set to TRUE, it will attempt to fill grid squares that
 * are partially covered by TIN; this is not recommended on non-GCS databases.
 */
extern void 
ctdb_te_dbase_processing (
			     CTDB_CMP         *ctdb,
			     int32                  corrections,
			     TRANSITION_PATCH_INFO *trans_patches,
			     int32                  trans_end,
			     char                  *fname,
			     int32                  fix_holes
			     );

/*
 * The ctdb_format_terrain_element routine formats a terrain element into a 
 * structure which can be used by geometry intersection routines.
 */
extern void
ctdb_format_terrain_element (
				TERRAIN_ELEMENT_16 *te,
				int32               offset,
				AREA               *A,
				int32               A_index
				);

/* This routine redefines the terrain elements for a complete patch as a set of
 * terrain elements that use 8 bit indices.
 */
void 
convert_16bit_to_8bit_te(
			 TERRAIN_ELEMENT_16  *te16,
			 int32                n_te,
			 TERRAIN_ELEMENT_08  *te8
			 );

/* This routine converts a SINGLE 8-bit TE to a 16-bit TE */
void 
convert_08_to_16te(
		   TERRAIN_ELEMENT_08 *te08,
		   TERRAIN_ELEMENT_16 *te16
		   );

/* Store the specified TEs (which are assumed to be 16 bit) as-is to the
 * output TE file, and update the ctdb as necessary.
 */
void
ctdb_store_patch_tes(
		     CTDB_CMP           *ctdb,
		     int32               patch_number,
		     TERRAIN_ELEMENT_16 *tes,
		     int32               n_te,
		     float64            *max_elev,
		     FILE               *f
		     );

/* --------------------------------------------------------------------
 *         ROUTINES DEFINED IN ct_geom.c 
 * -------------------------------------------------------------------- */
/* 
   This routine determines if the point c is to the left of points a and b;
   that is, that pts a, b, and c, in that order, are traverted counter-
   clockwise.
   */
extern int32 
ctdb_pt_left_of_edge (
			 CTDB_COORD2D a,
			 CTDB_COORD2D b,
			 CTDB_COORD2D c
			 );
/* 
   This routine determines if square area surrounding areas A and B overlap.
   This routine returns 0 is not overlap, 1 if overlap, and -1 if one or
   obth area extents have not been set.
   */
extern int32 
ctdb_compare_area_extents (
			      AREA *A,
			      AREA *B
			      );

/*
 * Exclusive or: true iff exactly one argument is true.
 * The arguments are negated to ensure that they are 0/1
 * values. Then the bitwise Xor operator may apply.
 * (This idea is due to Michael Baldwin.)
 */
extern int32 
ctdb_Xor ( 
	     int32 x,
	     int32 y
	     );

/*
 * Returns true iff (a,b,c) are collinear and point c lines on the
 * closed segment ab.
 */
extern int32 
ctdb_pt_on_closed_segment (
			      CTDB_COORD2D a,
			      CTDB_COORD2D b,
			      CTDB_COORD2D c
			      );
/*
 * Returns T iff (a,b,c) are collinear and point c lines on the
 * open segment ab.
 */
extern int32 
ctdb_collinear (
		   CTDB_COORD2D a,
		   CTDB_COORD2D b,
		   CTDB_COORD2D c
		   );

extern int32 
ctdb_diagonal (    
		  int32   i,
		  int32   j,
		  AREA_2D p
		  );

/* ctdb_point_in_te return TRUE is the given point pt is within the given
   terrain element te0. */

extern int32 
ctdb_point_in_te (
		     TERRAIN_ELEMENT_16 *tes,
		     int32               te0,
		     CTDB_XYO_COORD2D    pt
		     );

/* ctdb_point_on_te_edge returns TRUE if the given point pt is on an edge of the
   given terrain element te0. */

extern int32 
ctdb_point_on_te_edge (
			  TERRAIN_ELEMENT_16 *tes,
			  int32               te0,
			  CTDB_XYO_COORD2D    pt
			  );


/*
   This routine will calculate the area of a polygon
*/
extern float64
ctdb_calculate_area (
			AREA *A
			);
/*
 * This routine will determine the intersection point of two edges.
 */
extern int32 
ctdb_intersection (
		      CTDB_DB_EDGE  e0,
		      CTDB_DB_EDGE  e1,
		      CTDB_COORD2D *p
		      );
/*
 * This routine will loop through all edges of two input polygons 
 * and determine where the edge intersections points are.
 */
extern void
ctdb_find_edge_intersection (
				AREA *A,
				AREA *B,
				AREA *C
			       );
/*
 * This routine will add a point to the area structure in a clockwise 
 * order 
 */
extern void
ctdb_add_point (
		   AREA    *A,
		   float64  x,
		   float64  y
		   );

/* Initialize a, then populate it with verts */
void 
copy_verts32_to_area(
		     float32 *verts,
		     int32    num_verts,
		     AREA    *a
		     );

/*
 * This routine will free the dynamic memory (points) assigned to the area
 * structure.  It will also reset the extent items and set the count to 0.  
 * Note the area structure will remain.
 */
extern void
ctdb_free_area_struct (
			  AREA *A
			  );
/*
 * This routine will initialize the Area data structure and allocate a points
 * array of size c.
 */
extern int32 
ctdb_init_area_struct (
			  int32  c,
			  AREA  *P
			  );
/*
 * This routine will set the area extents of for the input polygon.  
 * The extents are defined as the bounding box in x and y that completely
 * enclose the given geometry.
 */
extern void 
ctdb_find_area_extents (
			   AREA *A
			   );


/* ctdb_point_in_tin returns TRUE if the given point pt is in or on the edge
 * of any terrain element in the list tes, whose length is given by n_te
 */
extern int32
ctdb_point_in_tin (
		      TERRAIN_ELEMENT_16 *tes,
		      int32               n_te,
		      CTDB_XYO_COORD2D    pt
		      );

/* ctdb_inside_square returns TRUE iff point a is inside square A. */
extern int32 
ctdb_inside_square (
		       CTDB_COORD2D a,
		       AREA *A
		       );

/*
 *  Returns true iff segments ab and cd intersect, properly or improperly.
 */
extern int32 
ctdb_seg_intersection (
			  CTDB_COORD2D a,
			  CTDB_COORD2D b,
			  CTDB_COORD2D c,
			  CTDB_COORD2D d
			  );


/* ctdb_pt_in_area takes a point and an arbitrary (i.e., not required to
 * be convex) area, and returns TRUE if the point is inside the area, else
 * FALSE
 */
int32
ctdb_pt_in_area(
		CTDB_COORD2D *point,
		AREA         *area
		);

/* Returns one of CTDB_CMP_NO_OVERLAP, CTDB_CMP_SOME_OVERLAP, 
 * CTDB_CMP_FULL_OVERLAP, depending on whether the square and the area
 * are disjoint, partially overlapping, or the square is fully contained in
 * the area, respectively
 */
typedef enum ctdb_overlap_state
{
  CTDB_CMP_NO_OVERLAP,
  CTDB_CMP_SOME_OVERLAP,
  CTDB_CMP_FULL_OVERLAP
} CTDB_OVERLAP_STATE;


int32
ctdb_square_poly_overlap(
			 AREA   *area,
			 float64 min_x,
			 float64 min_y,
			 float64 max_x,
			 float64 max_y
			 );



/* --------------------------------------------------------------------
 *         ROUTINES DEFINED IN ct_tin.c
 * -------------------------------------------------------------------- */
    
/* This routine creates a terrain element for the polygon speficied using the
   ith vertex of the given list. */
extern void
ctdb_make_te (
		 int32               i,
		 CTDB_VERTEX        *vertices,
		 TERRAIN_ELEMENT_16 *tes,
		 int32              *n_te,
		 int32              *new_list,
		 int32              *llength
		 );


/* This routine creates a new vertex in the given vertices list whose location
   is specified by (x,y,z). */

extern void 
ctdb_create_vertex (
		       int32        x,
		       int32        y,
		       float64      z,
		       CTDB_VERTEX *vertices,
		       int32       *num_verts
		       );

/* This routine added the specified edge to the given vertex */

extern void 
ctdb_add_edge (
	     CTDB_VERTEX  *vertex,
	     int32         edge
	     );



/* This routine traverses the transition patch list and reformats each 
 * indicated patch into a tinned transition patch between gridded patches 
 * and tinned patches.  If fix_holes is TRUE, it will attempt to fill 
 * partially TINned grid squares
 */
extern void
ctdb_create_transition_patches (
				   CTDB_CMP         *ctdb,
				   char                  *dbname,
				   TRANSITION_PATCH_INFO *trans_patches,
				   int32                 *tp_end,
				   int32                  fix_holes
				   );

/* This add transition patch adds this given patch number to list of patches
   that are to be generated as transition patches in a hybrid database */

extern void 
ctdb_add_transition_patch (
			      CTDB_CMP         *ctdb,
			      int32                  patch_number,
			      TRANSITION_PATCH_INFO *trans_patches,
			      int32                 *tp_end,
			      int32                  gridded_only
			      );

/* This routine removes the given patches from the list of patches to be
   formatted as transition patches. */
extern void 
ctdb_remove_transition_patch (
				 int32                  patch_number,
				 TRANSITION_PATCH_INFO *trans_patches,
				 int32                 *tp_end
				 );

/* This routine will initialize a new terrain element and extend the terrain
   element list. */

extern void 
ctdb_create_te (
		   TERRAIN_ELEMENT_16 *te,
		   int32              *n_te,
		   int32               x,
		   int32               y,
		   float64             z
		   );


    
/* This routine updates the patch headers of each patch to account for the
   insertion  of new terrain elements as part of the transition patch
   construction. */

extern void 
ctdb_update_patch_headers (
			      CTDB_CMP         *ctdb,
			      TRANSITION_PATCH_INFO *trans_patches
			      );

/* This routine converts the given terrain element (tes) into a polygon
   specification with the given soil and polygon vertex information. */
extern void
ctdb_define_polygon (
			TERRAIN_ELEMENT_16  *tes,
			int32                num_vertices,
			int32                soil0, 
			int32                soil1,
			int32               *vertices
			);



/* --------------------------------------------------------------------
 *         ROUTINES DEFINED IN ct_valid.c 
 * -------------------------------------------------------------------- */
/* CT_VALID.C routines */

/* The ctdb_validate_point_ordering routine will check the list of n_te terrain
   elements to determine that each polygon's vertex ordering is correct and 
   will correct the ordering if needed. */

extern void
ctdb_validate_point_ordering (
				  TERRAIN_ELEMENT_16 *te,
				  int32               n_te 
				  );



/*
   This routine will place a full word template over the (x,y) point and 
   compare the two edge points to determine if the edge is identical 
   */
extern int32 
ctdb_check_edge (
		    int32                e0[],
		    int32                e1[],
		    TERRAIN_ELEMENT_16  *te
		    );

/* This routine alternatively computes the soil of given set of triangulated
   vertices by looking of the post soil value using the center of mass for each
   triangle. */
extern void
ctdb_determine_triangle_soil (
				 CTDB_CMP      *ctdb,
				 int32          patch_number,
				 CTDB_VERTEX   *vertex,
				 int32          n_verts
				 );

/* Initialize an overlay */
extern void ctdb_init_overlay (
				  OVERLAY_2D *q
				  );
/* This routine constructs and initializes a new polygonal area (list of
   vertices). */

extern void ctdb_init_area (
			       int32    n,
			       AREA_2D *p
			       );

/* This routine adds the given point to the given area. */

extern void ctdb_extend_area (
				 AREA_2D      *a,
				 CTDB_COORD2D *pt
				 );

/*
  creates out n-3 diagonals (as pairs of integer indices)
  which form a triangulation of P.
*/
extern void ctdb_triangulate (
				 AREA_2D    *p,
				 OVERLAY_2D *q
				 );
/* This routine returns the vertex index whose position is specified by the
   given point pt. */
extern int32 ctdb_find_vertex (
				  CTDB_COORD2D *pt,
				  CTDB_VERTEX  *vertices,
				  int32         n_verts
				  );



/* This routine check the ordering of vertices define for the given terrain
   element i and reorders them when appropriate to achieve the required
   ordering: for triangles, the ordering is clockwise, of quadralaterals the 
   ordering is internal edge 1st, 1st triangle's remaining vertex and 2nd
   triangles remaining vertex where 2nd triangle is clockwise oriented from the
   1st triangle. */

extern void ctdb_check_te_order (
				    TERRAIN_ELEMENT_16 *te,
				    int32               i
				    );

/*  The ctdb_tin_defined_hole function takes a counter clockwise ordered list 
    of vertices (of length n_verts) the defines the perimeter of a hole to be
    filled.  This routine fills the hole with triangles and returns the
    triangles as an extension to the given list of tes (given length is n_te).
    The n_te returned is the total number of terrain elements after the
    addition.  This routine returns TRUE if successfull else FALSE;
*/


extern int32
ctdb_tin_defined_hole (
			  CTDB_CMP                  *ctdb,
			  int32                      patch_number,
			  CTDB_VERTEX               *vertex,
			  int32                      n_verts,
			  int32                     *n_te,
			  TERRAIN_ELEMENT_16        *te,
			  int32                     *new_list,
			  int32                     *llength
			  );
	      
/* This routine traverses all patches to align the vertices between patch
   boundaries. */
extern void ctdb_align_patch_boundaries (
					    CTDB_CMP        *ctdb,
					    char            *dbname
					    );


extern void ctdb_correct_patch (
				  CTDB_CMP           *ctdb,
				  int32               patch_number,
				  TERRAIN_ELEMENT_16 *tes,
				  int32              *n_te
				  );


/* This routine countes the number of tes that are adjacent to the given edge
   and returns that number. */

extern int32 ctdb_count_edge_faces (
				       TERRAIN_ELEMENT_16 *te,
				       int32               n_te,
				       int32               edge0[]
				       );

/* This routine determines if the given te0 overlaps with only other terrain 
   element from start through to end in tes list. It returns the index of the
   terrain element that overlaps with te0; Otherwise -1 is returned.
   */

extern int32 ctdb_determine_te_overlaps (
					    TERRAIN_ELEMENT_16 *tes,
					    int32               te0,
					    int32               start,
					    int32               end
					    );


/* --------------------------------------------------------------------
 *       ct_gcs.c
 * -------------------------------------------------------------------- */
/* Used by get_elev_at_gcs_pt.  This is the maximum allowable error in the
 * z value stored at a post.  Currently set at .5cm because this is < 1/2
 * the best resolution of post z values.
 */
#define MAX_DELTA 0.005

/* Initial guess at how many cells this assembly covers.  We'll be generous,
 * because it's cheap to guess high
 */
#define START_CELLS 50



/* Return the minimal bounding box in GCS local coordinates that 
 * contains every point in the specified lat/long range.
 */
extern void
find_gcs_bbox_from_ll (
			  float64  south_lat, 
			  float64  north_lat, 
			  float64  west_long, 
			  float64  east_long,
			  int32    cell,
			  float64 *min_x, 
			  float64 *max_x, 
			  float64 *min_y, 
			  float64 *max_y
			  );

/* Generate the bounds on patch x (min_px, max_px) and patch y (min_py,
 * max_py) in ctdb (which must be TCC-based) that fully contain (x1, y1)
 * (x2, y2) in cell cell.
 */
extern void
get_tcc_patch_list_from_gcs (
				CTDB_CMP      *ctdb,
				int32          cell,
				COORD_TCC_PTR  tcc,
				float64        x1, 
				float64        x2, 
				float64        y1, 
				float64        y2,    
				int32         *min_px, 
				int32         *max_px, 
				int32         *min_py, 
				int32         *max_py
				);

/* Return, in minux, max_ux, min_uy, max_uy, the TCC bounds of the
 * of the GCS box specified by cell, min/max g*
 */
extern void
find_tcc_bbox_from_gcs (
			   int32         cell,
			   float64       min_gx, 
			   float64       max_gx, 
			   float64       min_gy, 
			   float64       max_gy,
			   COORD_TCC_PTR tcc,
			   float64      *min_ux, 
			   float64      *max_ux, 
			   float64      *min_uy, 
			   float64      *max_uy
			   );


/* Convert points to GCS.  Note that since these are 2D, we need to
 * to get a Z value (we'll use the terrain surface, since that's usually
 * what was meant)
 */
extern void
translate_verts_2D_tcc_to_gcs32 (
				    COORD_TCC_PTR tcc,
				    int32         cell,
				    float32      *points,
				    int32         n_points,
				    int32       (*lookup_fn)(/*float32, 
							       float32, 
							       float32 * */)
				    );


/* Convert points to GCS.  Note that since these are 2D, we need to
 * to get a Z value (we'll use the terrain surface, since more than likely,
 * that's what was meant)
 */
extern void
translate_verts_2D_tcc_to_gcs (
				  COORD_TCC_PTR tcc,
				  int32         cell,
				  float64       points[][2],
				  int32         n_points,
				  int32       (*lookup_fn)(/* float32, 
							      float32, 
							      float32 * */)
				  );


/* Convert elements of points list from TCC to GCS in the specified cell */
extern void
translate_verts_tcc_to_gcs (
			       COORD_TCC_PTR tcc,
			       int32         cell,
			       float64       points[][3],
			       int32         n_points
			       );

/* Converts min/max x/y in TCC or GCS into lat/long boundaries.  Returns TRUE 
 * for success, FALSE for failure.  Based on gcs_ll_bounds_from_gcs in libgcs.
 * mode is one of {COORD_GCS, COORD_TCC}
 */
extern int32
get_ll_bounds (
		  COORD_SYSTEM  mode,
		  int32         cell,
		  COORD_TCC_PTR tcc,
		  float64       min_x,
		  float64       min_y,
		  float64       max_x,
		  float64       max_y,
		  float64      *west_lng,
		  float64      *south_lat,
		  float64      *east_lng,
		  float64      *north_lat
		  );

/* Given the input file name, make a best guess at a desirable output file
 * name, and put that in out_name.
 */
extern void
generate_name_base (
		       char *in_name,
		       char *out_name
		       );

/* Given a 2D point in GCS, we want to find the elevation in GCS--i.e., the
 * z value at which a line pointing straight up ("up" defined as parallel to
 * the z axis) intersects the terrain.  Since the z vector in GCS is not
 * the same as that in UTM, we have to actually look for this intersection 
 * point rather than performing a trivial elevation lookup.  This algorithm is 
 * based on two assumptions:
 *    1.  The terrain is relatively flat
 *    2.  The difference between the GCS and UTM z-vectors is not great
 *
 * Using those assumptions, we do the following:
 *    1.  Generate a line segment pointing straight up at the desired point
 *        in GCS, and translate the endpoints to UTM
 *    2.  Do a (UTM) elevation lookup at the x,y of one endpoint of the
 *        segment.
 *    3.  Determine the x,y of the segment at that z.
 *    4.  Perform an elevation lookup at that x,y.  If the result is 
 *        "close enough" the last z we saw, we're done.
 *    5.  If the result is not good enough, do another lookup at the x,y
 *        along the segment of the new z.  Iterate as necessary.
 *
 */
extern int32
get_elev_at_gcs_pt (
		       COORD_TCC_PTR   tcc,
		       int32           cell,
		       float64         x,
		       float64         y,
		       float64        *z,
		       int32         (*lookup_fn)(/*float32, 
						    float32, 
						    float32 * */)
		       );

/* Special case of above where the TCC elevation is known to be constant in
 * in the region where we're looking--this is primarily useful when dealing
 * with oceans
 */
float64
get_elev_at_gcs_pt_plane(
			 COORD_TCC_PTR   tcc,
			 int32           cell,
			 float64         x,
			 float64         y,
			 float64         tcc_z
			 );

/* Given a multi-cell header name, ctdb_generate_multi_directory
 * will create the appropriately named directory in which
 * the multi-cell header file and the appropriate cell databases
 * will be placed.
 */
extern void
ctdb_generate_multi_directory (
			       char *dbname
			       );

	
/* --------------------------------------------------------------------
 *         ROUTINES DEFINED IN ct_mes_reader.c 
 * -------------------------------------------------------------------- */

extern int32 
ctdb_read_mes_libreader(
			CTDB_CMP*                ctdb,
			int32                    num_corrections,
			READER_UNION            *corrections,
			int32                   *num_mes_volumes,
			CTDB_MES_VOLUME_CMP*    *mes_volumes,
			int32                   *num_mes_templates,
			CTDB_MES_TEMPLATE_CMP*  *mes_templates,
			int32                    elev_relative_to_terrain
			);
extern int32
ctdb_mes_free_cmp(CTDB_CMP *ctdb);


extern CTDB_MES_VOLUME_CMP*
ctdb_mes_from_bridge_feature(
		        /* Input Parameters */
		        CTDB_CMP                      *ctdb,
		        int32                          num_footprint_vertices,
		        float32                       *footprint_vertices,
		        CTDB_MES_LIST_POLYGON_F64_CMP *bridge,
		        /* Output Parameters */
		        int32                         *num_volumes,
		        CTDB_MES_VOLUME_CMP          **volumes,
		        int32                         *num_templates,
		        CTDB_MES_TEMPLATE_CMP        **templates
		        );

#endif /*_REFORM_H_INCLUDED*/


