#ifndef lint
static char rcsid [] = "$RCSfile: ct_tecmplr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_tecmplr.c                                                      *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : This file contains the Terrain element specific routines*
 *                   of the compiler backend for format 4 CTDB databases.    *
 *       Created   : August 1995                                             *
 *       Author    : wkbraudaway                                             *
 *       Remarks   :                                                         *
 *                                                                           *
 ****************************************************************************/
#if !defined(SUN) && !defined(mips) && !defined(__alpha)
#include <bstring.h>
#endif
#include <stdio.h>      /* fwrite routine */
#include <math.h> 
#include <stdext.h>
#include <stdstring.h>
#include "../libctdb_local.h"
#include "ct_reform.h"
#include "ct_cmplr.h"

/* Forward Declarations */
static void integrate_feature_present();
static int32 patch_feature_present();
static void evaluate_patch_tinned();
static void record_feature_present();
static void record_micro_present();
static void convert_micro_to_16bit_te();
static void set_16bit_te();
static void mark_patch_16_bit();
static float64 get_posted_elevation();

/*** Form Terrain Element Topology Routines ***/
static void construct_te_adj();
static void check_te_adj();
static void associate_edge();
static void set_adj_info();
static void resolve_patch_neighbors();
static CTDB_BEARING determine_db_adj();
static int32 get_side_index();
static int32 check_edge_by_index();
static int te_adj_cmp();

/* For Virtual Grid Routines */

static void create_te_mapping();

/* The ctdb_covert_to_tins routine will reformat the list of microterrain
   features for a single patch into a list of terrain elements and record 
   that list in a specified temporary file.   This routine will upon
   request correct the tinning the current patch.   This routine will also
   flag this patch (transition) if the patch is not completely tinned
   (i.e., some post grid cell does not contain microterrain).
 */

void ctdb_convert_to_tins(ctdb, patch_number, feature, 
			  n_feat, n_te, trans_patches, trans_end,
			  max_elev, corrections, new_micro, fptr)
    CTDB_CMP              *ctdb;
    int32                  patch_number;
    CTDB_FEATURE_DATA_CMP *feature;
    int32                  n_feat;
    int32                 *n_te;
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *trans_end;
    float64               *max_elev;
    int32                  corrections;
    int32                  new_micro;
    FILE                  *fptr;
{
    int32 i;
    TERRAIN_ELEMENT_16 tes[MAX_ALLOWED_TE_INDEX_16];


    *n_te = n_feat / NEXT_MICRO_CMP(3);
    for(i = 0; i<*n_te; i++, feature += NEXT_MICRO_CMP(3))
    {
	convert_micro_to_16bit_te(feature, &tes[i], i);
	if (tes[i].z > *max_elev)
	  *max_elev = tes[i].z;
    }
    
    ctdb_validate_point_ordering(tes, *n_te);
    if (corrections)
      ctdb_correct_patch(ctdb, patch_number, tes, n_te);

    if (ctdb->elevations)
      record_micro_present(ctdb, patch_number, tes, *n_te);

    if (new_micro)
      evaluate_patch_tinned (ctdb, patch_number, tes, *n_te, trans_patches, 
			     trans_end);

    fwrite(&tes[0],sizeof(TERRAIN_ELEMENT_16), *n_te, fptr);
}

/* 
 * The ctdb_set_te_patch_header routine updates the patch_group header
 * information for the terrain elements added to the indicated patch.  This
 * routine determines if the patch is special (either a transition patch or a
 * gridded patch of a database to be completely tinned), sets the patch te
 * offsets, determines if the current patch is an 8 or 16 bit indexed patch,  
 * and updates the feature present bits for the current patch. 
 */

void
ctdb_set_te_patch_header(ctdb, patch_x, patch_y, offset, num_te,
			 max_elev, trans_patches, trans_end, features,
			 n_features, dbtype)
    CTDB_CMP              *ctdb;
    int32                  patch_x;
    int32                  patch_y;
    int32                  offset;
    int32                 *num_te;
    float64                max_elev;
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *trans_end;
    CTDB_FEATURE_DATA_CMP *features;
    int32                  n_features;
    CTDB_DBASE_TYPE        dbtype;
{
    int32 patch_number = patch_x + patch_y * ctdb->patches_wide;
    int32 member= PG_MEMBER(patch_number);
    CTDB_PATCH_GROUP_CMP *pg = ctdb->patch_groups + PG_NUMBER(patch_number);
    int32 post_x, post_y;
    int32 px,py;
    int32 packing =0;
    int32 patch_fp_size = ctdb->patch_incr * ctdb->patch_incr * 0.5;
    uint32 te_threshold=CTDB_TE_08_TO_16_BIT_CROSSOVER;	    
    uint8 *pr;
    uint8 byte;
    int32 post;
    static int32 pg_features=0;
    static int32 last_group = -1;

    /* If we're starting a new group now, reset the feature count.
     */
    if(PG_NUMBER(patch_number) != last_group)
      pg_features = 0;

    last_group = PG_NUMBER(patch_number);
    
    pg_features += n_features;
    /* This patch must be tinned so mark it */
    if (*num_te == 0) 
      if (dbtype == CTDB_TINNED)
	ctdb_add_transition_patch(ctdb, patch_number, trans_patches, trans_end,
				  (int32)FALSE);

    if (ctdb->elevations)
      ctdb_intern_compute_grid_max_z(ctdb, patch_x, patch_y, &max_elev);
    pg->max_elev[member] = max_elev;
    pg->num_te[member] = *num_te;

    /* if Patch has too many terrain elements, we'll have to use 16 bit
     * indices
     */
    if ( *num_te > te_threshold )
      mark_patch_16_bit(ctdb,patch_number);
    
    /* set terrain element position for this patch */
    if (member == 0)
      pg->te_offset[0] = offset;
    else 
      pg->te_offset[member] = offset - pg->te_offset[0];

    pg->total_patch_feature_words = pg_features;

    if (n_features)
	/* update patch header feature present bits */
      integrate_feature_present(ctdb, patch_x, patch_y, pg, features);
}


/* This routine reads the temporary terrain element data file and then removes
 * it.
 */

static CTDB_STATUS
read_terrain_element_data(ctdb,dbname)
    CTDB_CMP *ctdb;
    char     *dbname;    
{
    char filename[CTDB_FILENAME_DIMENSION_CMP];
    FILE *fileptr;
    
    
    ctdb->terrain_elements = 
      (TERRAIN_ELEMENT_16 *)ctdb_alloc(ctdb->terrain_elements,
				     ctdb->n_te_bytes,
				     "DB Terrain element data");
    
    sprintf(filename,"%s.te",dbname);
    
    if ( (fileptr=fopen(filename,"rb")) == (FILE *) NULL )
      return CTDB_STATUS_FAILED;
    
    fread(ctdb->terrain_elements,1,ctdb->n_te_bytes,fileptr);
    
    fclose(fileptr);
    unlink(filename);
    return CTDB_STATUS_OK;
}


/* This function takes the terrain elements file and reorders the terrain
 * elements such that they are stored beginning with the southwesternmost
 * patch, increasing from west to east, then from south to north.  We must
 * do this because the code that processes the terrain elements assumes
 * that they are stored in that order.  This is not always true, especially
 * when processing merge files, where the patches can be encoded in any
 * order.
 */
static CTDB_STATUS
reorder_terrain_elements(ctdb, fname)
    CTDB_CMP              *ctdb;
    char                  *fname;
{
    char filename[CTDB_FILENAME_DIMENSION_CMP];
    FILE *fileptr;
    CTDB_PATCH_GROUP_CMP *pg;
    int32 patch_number;
    int32 member, te_offset[CTDB_PATCHES_PER_GROUP], i;
    int32 patch_x, patch_y, n_tes, current_offset = 0;
    TERRAIN_ELEMENT_16 *in_tes, *out_tes;

    out_tes = (TERRAIN_ELEMENT_16 *)ctdb_alloc(NULL, ctdb->n_te_bytes,
					       "terrain element reorder");

    ctdb->terrain_elements = 
      (TERRAIN_ELEMENT_16 *)ctdb_alloc(ctdb->terrain_elements,
				       ctdb->n_te_bytes,
				       "DB Terrain element data");

    sprintf(filename,"%s.te",fname);

    if ( (fileptr=fopen(filename,"rb")) == (FILE *) NULL )
      return CTDB_STATUS_FAILED;

    fread(ctdb->terrain_elements,1,ctdb->n_te_bytes,fileptr);

    fclose(fileptr);
    unlink(filename);

    if ((fileptr=fopen(filename,"wb")) == (FILE *) NULL )
    {
	printf("unable to open temporary file in current\n");
	exit(0);
    }

    for(patch_y = 0; patch_y < ctdb->patches_high; patch_y++)
      for(patch_x = 0; patch_x < ctdb->patches_wide; patch_x++)
      {
	  patch_number = (patch_y * ctdb->patches_wide) + patch_x;
	  pg = ctdb->patch_groups+PG_NUMBER(patch_number);
	  member = PG_MEMBER(patch_number);

	  /* Look up the number of terrain elements for this patch and
	   * where they're stored.
	   */
	  n_tes = ctdb_intern_number_tes(ctdb, patch_number);
	  in_tes = ctdb_intern_lookup_terrain_element(ctdb, patch_number);

	  /* Save the new offsets into a temporary array.  We don't
	   * actually assign the new offsets to the patch group data
	   * headers until we're done processing the entire patch group.
	   * We have to wait because we're still doing lookups for the
	   * old terrain elements.
	   */
	  if(!member)
	    te_offset[0] = current_offset * sizeof(TERRAIN_ELEMENT_16);
	  else
	    te_offset[member] = current_offset * sizeof(TERRAIN_ELEMENT_16) -
	                        te_offset[0];

	  /* If we're done processing the entire group, it's OK to
	   * copy in the new offsets into the patch group data header.
	   */
	  if(member == CTDB_PATCHES_PER_GROUP - 1)
	    for(i = 0; i < CTDB_PATCHES_PER_GROUP; i++)
	      pg->te_offset[i] = te_offset[i];

	  if(!in_tes)
	    continue;

	  /* Copy the terrain elements for this patch from the original
	   * list of terrain elements to the reordered list.
	   */
	  bcopy(in_tes, out_tes + current_offset, n_tes *
		sizeof(TERRAIN_ELEMENT_16));

	  /* `current_offset' is in units of TERRAIN_ELEMENT_16.
	   */
	  current_offset += n_tes;
      }

    /* Now that the terrain elements have been reordered in memory, write
     * them out to a file.
     */
    fwrite(out_tes,1,ctdb->n_te_bytes,fileptr);
    
    fclose(fileptr);

    STDDEALLOC(out_tes);

    return CTDB_STATUS_OK;
}

/* After all patches have been processed for terrain elements, the
 * ctdb_te_dbase_processing routine loads the tes from their temporary file,
 * creates transition patches where needed and constructs the te topology. 
 * If fix_holes is set to TRUE, it will attempt to fill grid squares that
 * are partially covered by TIN; this is not recommended on non-GCS databases.
 */

void ctdb_te_dbase_processing(ctdb, corrections, trans_patches,
			      trans_end, fname, fix_holes)
    CTDB_CMP              *ctdb;
    int32                  corrections;
    TRANSITION_PATCH_INFO *trans_patches;
    int32                  trans_end;
    char                  *fname;
    int32                  fix_holes;
{
    int32 patch_x, patch_y;
    
    reorder_terrain_elements(ctdb, fname);

    if ( read_terrain_element_data(ctdb, fname) != CTDB_STATUS_OK )
      fprintf(stderr,"error reading terrain elements");

    if (!(ctdb->elevations) && (trans_end != 0))
    {
	printf("Warning attempted to tin gridded region when no grid posts exist\n");
	trans_end = 0;
    }

    /* If transition patches are needed then create them */
    if (trans_end != 0)
      {
	  printf("\nCreating Transition Patches\n");
	  ctdb_create_transition_patches(ctdb, fname, trans_patches, 
					 &trans_end, fix_holes);
	  if ( read_terrain_element_data(ctdb, fname) != CTDB_STATUS_OK )
	    fprintf(stderr,"error reading terrain elements");
      }

    /* Match vertices on all patch boundaries */
    if (ctdb->n_te_bytes)
    {
	ctdb_align_patch_boundaries(ctdb, fname);
	if ( read_terrain_element_data(ctdb, fname) != CTDB_STATUS_OK )
	  fprintf(stderr,"error reading terrain elements");
    }

    /* update post feature present bits - for the precise micro terrain 
       available with each patch in the feature list.
       */

    if (ctdb->elevations)
      for (patch_y=0; patch_y < ctdb->patches_high; patch_y++)
	for (patch_x=0; patch_x < ctdb->patches_wide; patch_x++)
	  record_feature_present(ctdb, patch_x, patch_y);

    if (ctdb->n_te_bytes)
    {
	printf("\nCreating Polygon Topology\n");
	construct_te_adj(ctdb); 
    }
}

static void integrate_feature_present(ctdb, patch_x, patch_y, pg, features)
    CTDB_CMP              *ctdb;
    int32                  patch_x;
    int32                  patch_y;
    CTDB_PATCH_GROUP_CMP  *pg;
    CTDB_FEATURE_DATA_CMP *features;    
{
    int32 post_x, post_y;
    uint8 *pr;
    uint8 byte=0;
    int32 post;
    int32 patch_number = patch_x + patch_y * ctdb->patches_wide;
    int32 member = PG_MEMBER(patch_number);
    int32 packing =0; /* status: 0 = high bits packed, 1 = low bits packed */
    int32 patch_fp_size = ctdb->patch_incr * ctdb->patch_incr * 0.5;
    int32 px, py;

    post_x = patch_x * ctdb->patch_incrD;
    post_y = patch_y * ctdb->patch_incrD + 1.0;
    pr = pg->feature_present + (member * patch_fp_size);

    /* loop through posts of patch and collect feature present information */
    for (py =0;py<ctdb->patch_incr;py++)
    {
	for (px = 0;px<ctdb->patch_incr;px++)
	{
	    post = patch_feature_present(ctdb, px+post_x, py+post_y, 
					 patch_x, patch_y, features, pg);
	    byte  |= ((post&POST_FLAGS)>>20);
	    if (packing == 1) /* low bits packed - store off */
	    {
		*pr = byte; /* move packed byte into array */
		byte = 0;
		pr++; /* advance array */
		packing = 0;
	    }
	    else
	    {
		byte  <<= 4; /* high bits packed - shift into position */
		packing = 1;
	    }
	    
	}	
	if (packing == 1) /* only high bits packed - store result */
	{
	    *pr = byte;
	    byte = 0;
	    pr++;
	    packing = 0;
	}
    }
}

static int32 patch_feature_present(ctdb, px, py, patchx, patchy, feature, pg)
    CTDB_CMP              *ctdb;
    int32                  px; 
    int32                  py; 
    int32                  patchx;
    int32                  patchy;
    CTDB_FEATURE_DATA_CMP *feature;
    CTDB_PATCH_GROUP_CMP  *pg;
{
    int32 result = 0;
    int32 post_mask; 
    CTDB_FEATURE_DATA_CMP *last, *next; 
    int32 verts, fclass; 

    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);

    last = feature->hdr.size + feature;
    for(feature++;feature<last;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	if (fclass == FC_MICRO_CMP)
	  next = feature + NEXT_MICRO_CMP(verts);
	else if (fclass == FC_VOLUME_CMP)
	  next = feature + NEXT_VOLUME_CMP(verts);
	else if (fclass == FC_CANOPY_CMP)
	  next = feature + NEXT_CANOPY_CMP(verts);
	else if (fclass == FC_LINEAR_CMP)
	  next = feature + NEXT_LINEAR_CMP(verts);
	else if (fclass == FC_AGGREGATE_CMP)
	  next = feature + NEXT_AGGREGATE_CMP(verts);
	else if (fclass == FC_LAID_LINEAR_CMP)
	  next = feature + NEXT_LAID_LINEAR_CMP(verts);

	if (!(post_mask & feature->integer))
	  continue;

	if (fclass == FC_MICRO_CMP)
	  result |= POST_MICROTERRAIN;
	else if (fclass == FC_VOLUME_CMP)
	  result |= POST_BUILDINGS;
	else if ((fclass == FC_LINEAR_CMP) || (fclass == FC_CANOPY_CMP))
	  result |= POST_TREES;
	else
	  result |= POST_FEATURES;
    }
    return result;
}

static void 
evaluate_patch_tinned(ctdb, patch_number, tes, n_te, trans_patches, tp_end)
    CTDB_CMP                   *ctdb;
    int32                       patch_number;
    TERRAIN_ELEMENT_16         *tes;
    int32                       n_te;
    TRANSITION_PATCH_INFO      *trans_patches;
    int32                      *tp_end;
{
    /* Given that only those patches to the south and west of the correct patch
     * have been formulated and that the current patch is at least partically
     * tinned, this routine evaluates the possible transition between this
     * tinned patch and its neighboring patches. 
     */


    /* if all posts are currently tinned, then remove this patch from the
     * transition list 
     */

    int32 posts_per_patch_side = ctdb->patch_incr;
    int32 ordinal_x = patch_number%ctdb->patches_wide *
      posts_per_patch_side;
    int32 ordinal_y = patch_number/ctdb->patches_wide * 
      posts_per_patch_side;
    int32 post_x, post_y, post;
    int32 i,j,k;
    int32 x, y;
    int32 transition_patch;
    int32 post_conversion = CTDB_PATCH_TO_XYO_SCALE / ctdb->patch_incr;
    int32 delta = post_conversion * 0.5;
    int32 delta_quarter = post_conversion * 0.25;
    int32 delta_3quarter = post_conversion * 0.75;
    int32 bdy_post;
    CTDB_XYO_COORD2D    pt;
    int32 inside;

    transition_patch = FALSE;
    
    /* loop through each post */
    for (j = 0; j < posts_per_patch_side; j++)
    {
	post_y = ordinal_y + j + 1;
	y = post_y * ctdb->incr;
	for (i = 0; i < posts_per_patch_side; i++)
	{    
	    /* note: (i,j) is lower left corner of grid cell */
	    bdy_post = ((i == 0) || (i == posts_per_patch_side - 1) ||
			(j == 0) || (j == posts_per_patch_side -1));

	    post_x = ordinal_x + i;
	    x = post_x * ctdb->incr;
	    
	    inside = FALSE;
	    /* If post does not contain micro terrain, or if post is not fully
	     * covered by microterrain and it is on the patch boundary, then 
	     * the current patch is a transition patch 
	     */

	    /* $$$ This should be unnecessary. */
/*
	    if (ctdb_point_on_database((CTDB *)ctdb,(float64)x,(float64)y) && 
		ctdb_point_on_database((CTDB *)ctdb,(float64)(x + ctdb->incr), 
				       (float64)(y - ctdb->incr)))
*/
	    {
		
		/* check center of post */
		pt.x = i * post_conversion + delta;
		pt.y = j * post_conversion + delta;
		/* determine if post inside some te */
		for (k = 0; k<n_te; k++)
		  /* check center of post */
		  if ((ctdb_point_on_te_edge(tes, k, pt)) ||
		      (ctdb_point_in_te(tes, k,  pt)))
		  {
		      inside = TRUE;
		      break;
		  }

		if ((!inside && !bdy_post) || (inside && bdy_post))
		{
		    if (bdy_post)
		      inside = FALSE;
		    /* check upper left of post */
		    pt.x = i * post_conversion + delta_quarter;
		    pt.y = j * post_conversion + delta_3quarter;
		    for (k = 0; k<n_te; k++)
		      if ((ctdb_point_on_te_edge(tes, k, pt)) ||
			  (ctdb_point_in_te(tes, k,  pt)))
		      {
			  inside = TRUE;
			  break;
		      }
		}

		if ((!inside && !bdy_post) || (inside && bdy_post))
		{
		    if (bdy_post)
		      inside = FALSE;
		    /* check upper right of post */
		    pt.x = i * post_conversion + delta_3quarter;
		    pt.y = j * post_conversion + delta_3quarter;
		    for (k = 0; k<n_te; k++)
		      if ((ctdb_point_on_te_edge(tes, k, pt)) ||
			  (ctdb_point_in_te(tes, k,  pt)))
		      {
			  inside = TRUE;
			  break;
		      }
		}

		if ((!inside && !bdy_post) || (inside && bdy_post))
		{
		    if (bdy_post)
		      inside = FALSE;
		    /* check lower left of post */
		    pt.x = i * post_conversion + delta_quarter;
		    pt.y = j * post_conversion + delta_quarter;
		    for (k = 0; k<n_te; k++)
		      if ((ctdb_point_on_te_edge(tes, k, pt)) ||
			  (ctdb_point_in_te(tes, k,  pt)))
		      {
			  inside = TRUE;
			  break;
		      }
		}
		
		if ((!inside && !bdy_post) || (inside && bdy_post))
		{
		    if (bdy_post)
		      inside = FALSE;
		    /* check lower right of post */
		    pt.x = i * post_conversion + delta_3quarter;
		    pt.y = j * post_conversion + delta_quarter;
		    for (k = 0; k<n_te; k++)
		      if ((ctdb_point_on_te_edge(tes, k, pt)) ||
			  (ctdb_point_in_te(tes, k,  pt)))
		      {
			  inside = TRUE;
			  break;
		      }
		}
		
		if (!inside)
		{
		    if (!transition_patch)
		    {
			ctdb_add_transition_patch(ctdb, patch_number, 
						  trans_patches,tp_end,
						  (int32)TRUE);
			transition_patch = TRUE;
		    }
		}
		else /* Post contains micro terrain */
		{   /* if a boundary post then evaluate boundary type */
		    /* west boundary */
		    if ((i == 0) && (patch_number%ctdb->patches_wide > 0)) 
		      if (ctdb_intern_number_tes(ctdb, patch_number - 1)
			  == 0)
			ctdb_add_transition_patch(ctdb, patch_number - 1,
						  trans_patches,tp_end,
						  (int32)TRUE);
		    /* south boundary */
		    if ((j == 0) && (patch_number > ctdb->patches_wide)) 
		      if (ctdb_intern_number_tes(ctdb, patch_number -
						 ctdb->patches_wide) == 0)
			ctdb_add_transition_patch(ctdb, patch_number -
						  ctdb->patches_wide,
						  trans_patches, tp_end,
						  (int32)TRUE);
		    if (i == (posts_per_patch_side - 1))  /* east boudary */
		      ctdb_add_transition_patch(ctdb, patch_number + 1, 
						trans_patches, tp_end,
						(int32)TRUE);
		    if (j == (posts_per_patch_side - 1))  /* north boudary */
		      ctdb_add_transition_patch(ctdb, patch_number +
						ctdb->patches_wide,
						trans_patches, tp_end,
						(int32)TRUE);
		} /* post covered */
	    } /* inside patch */
	} /* px */
    } /* py */
    
    if (!transition_patch)
      ctdb_remove_transition_patch(patch_number, trans_patches, tp_end);
}


/* The record_feature_present routine will fill in the feature present bits
   for each post that is contained in the patch defined by patch_x, patch_y */
static void 
record_feature_present(ctdb, patch_x, patch_y)
    CTDB_CMP      *ctdb;
    int32          patch_x;
    int32          patch_y;
{
    float64 post_x, post_y;
    int32 i, j;
    int32 post_index;
    float64 norm_x, norm_y;

    post_x = patch_x * ctdb->patch_incrD;
    post_y = patch_y * ctdb->patch_incrD + 1.0;

    for(i=0; i<ctdb->patch_incr; i++)
      for(j=0; j<ctdb->patch_incr; j++) 
      {
	  norm_x = post_x + j;
	  norm_y = post_y + i;
	  if (CTDB_POST_IN_CELL(ctdb, norm_x, norm_y))
	  {
	      post_index = ctdb_intern_get_post_index(ctdb, 
						      (int32)norm_x, 
						      (int32)norm_y);
	      /* zero out post flag bits */
	      ctdb->elevations[post_index] &= ~POST_FLAGS;
	      ctdb->elevations[post_index] |= 
		ctdb_features_of_post(ctdb, norm_x, norm_y);
	  }
      }
}

static void record_micro_present(ctdb, patch_number, tes, n_te)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
{

    float64 post_x, post_y;
    int32 i, j;
    int32 post_index;
    float64 norm_x, norm_y;
    int32 patch_x = patch_number%ctdb->patches_wide;
    int32 patch_y = patch_number/ctdb->patches_wide;
    CTDB_XYO_COORD2D    pt;
    int32 k;
    int32 post_conversion = CTDB_PATCH_TO_XYO_SCALE / ctdb->patch_incr;
    int32 delta = post_conversion * 0.5;

    post_x = patch_x * ctdb->patch_incrD;
    post_y = patch_y * ctdb->patch_incrD + 1.0;

    for(i=0; i<ctdb->patch_incr; i++)
      for(j=0; j<ctdb->patch_incr; j++) 
      {
	  norm_x = post_x + j;
	  norm_y = post_y + i;
	  if (CTDB_POST_IN_CELL(ctdb, norm_x, norm_y))
	  {
	      post_index = ctdb_intern_get_post_index(ctdb, 
						      (int32)norm_x, 
						      (int32)norm_y);
	      /* zero out post flag bits */
	      ctdb->elevations[post_index] &= ~POST_FLAGS;
	      pt.x = POST_TO_XYO(norm_x,patch_x,ctdb->patch_inv_incr) + delta;
	      pt.y = POST_TO_XYO(norm_y,patch_y,ctdb->patch_inv_incr) - delta;

	      /* determine if post inside some te */
	      for (k = 0; k<n_te; k++)
		if ((ctdb_point_on_te_edge(tes, k, pt)) ||
		    (ctdb_point_in_te(tes, k,  pt)))
		  {
		      ctdb->elevations[post_index] |= POST_MICROTERRAIN;
		      break;
		  }
	  }
      }
}


/* This routine will reform the ctdb relative offset format to absolute format
   that is be used in TINs.
   */
static void 
convert_micro_to_16bit_te(feature, te, n_te)
    CTDB_FEATURE_DATA_CMP *feature;
    TERRAIN_ELEMENT_16    *te;
    int32                  n_te;
{
    int32 i;
    
    te->info.vertices = feature->info.vertices & 0x3; /* transfer the lower two
							 bits only*/
    te->info.patch_adj = 0;
    feature++;

    te->info.soil0 = feature->micro_soil.poly_char_index0;
    te->info.soil1 = feature->micro_soil.poly_char_index1;
    feature++;

    bcopy(&(feature->xy),&(te->pt),sizeof(feature->xy));
    feature++;

    te->z = feature->z;
    feature++;

    te->vert.index[0] = feature->micro.vert0 + n_te;
    te->vert.index[1] = feature->micro.vert1 + n_te;
    te->vert.index[2] = feature->micro.vert2 + n_te;
    for (i=0;i<TE_ADJ_TRI_MAP_DIMENSION; i++)
      te->adj[i] = EDGE_NOT_MAPPED_16;
    te->info.adj_sides = 0;
    return;
}

/* This routine creates a Terrain element for the given micro terrain 
 * feature 
 */
static void
set_16bit_te(pg,patch_number)
    CTDB_PATCH_GROUP_CMP *pg;
    int32                 patch_number;
{
    switch ( PG_MEMBER(patch_number) )
    {
      case 0 :
	pg->patch0_te16 = 1;
	break;
      case 1 :
	pg->patch1_te16 = 1;
	break;
      case 2 :
	pg->patch2_te16 = 1;
	break;
      case 3 :
	pg->patch3_te16 = 1;
	break;
    }
}


/* The ctdb_mark_patch_16_bit records that the given patch must use 16 BIT
   terrain element indices since the number of terrain elements in the patch 
   is larger than can be represented using 08 bit terrain element indices. */

static void
mark_patch_16_bit(ctdb, patch_number)
    CTDB_CMP      *ctdb;
    int32          patch_number;
{
    CTDB_PATCH_GROUP_CMP *pg;
    int32 lpn;
    
    /* Center */
    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    set_16bit_te(pg,patch_number);
    /* East */
    lpn = patch_number+1;   
    if ( (lpn%ctdb->patches_wide) )
    {
	pg = ctdb->patch_groups+PG_NUMBER(lpn);
	set_16bit_te(pg,lpn);
    }
    /* West */
    lpn = patch_number-1;
    if ( ( lpn%ctdb->patches_wide) != (ctdb->patches_wide-1) )
    {
	pg = ctdb->patch_groups+PG_NUMBER(lpn);
	set_16bit_te(pg,lpn);
    }
    /* North */
    lpn = patch_number + ctdb->patches_wide ;
    if ( ( lpn/ctdb->patches_wide) < ctdb->patches_high )
    {
	pg = ctdb->patch_groups+PG_NUMBER(lpn);
	set_16bit_te(pg,lpn);
    }
    /* South */
    lpn = patch_number - ctdb->patches_wide ;
    if ( lpn >= 0 )
    {
	pg = ctdb->patch_groups+PG_NUMBER(lpn);
	set_16bit_te(pg,lpn);
    }
}


/*** Form Terrain Element Topology Routines ***/

#define X_AXIS 0
#define Y_AXIS 1

/* The ctdb_construct_te_adj function will construct the terrain element
   topology within and between patches for the given reformatted ctdb
   database */

static void construct_te_adj(ctdb)
    CTDB_CMP      *ctdb;
{
    int32 py, px;
    int32 patch_number, nbor_num;
    TERRAIN_ELEMENT_16 *te, *nbor_te;
    int32 n_te, e;
    uint32 i;
    int32  edge[2];
    CTDB_BEARING db_adj;
    int32 wx,wy;
    int te_adj_cmp ();
    PATCH_ADJ_INFO *patch_adj, *current_pa, *nbor_pa;
    int32 patch_size = ctdb->patch_incr*ctdb->incr;
    
    patch_adj = (PATCH_ADJ_INFO *) /* create array of patch adj info stucts */
      ctdb_alloc(patch_adj = NULL,
		  sizeof(PATCH_ADJ_INFO)*ctdb->patches_wide*
		  ctdb->patches_high,
		  "patch adjacency information");
    
    /* loop through each patch to setup internal te topology and collect 
       information about external te topology (i.e., tes between patches) */
    
    for (py=0; py < ctdb->patches_high; py++)
    {
	putchar('.');fflush(stdout);
	for (px=0; px < ctdb->patches_wide; px++)
	{
	    patch_number = py*ctdb->patches_wide + px;
	    te = (TERRAIN_ELEMENT_16 *)
	      ctdb_intern_lookup_terrain_element(ctdb,patch_number);
	    n_te   = ctdb_intern_number_tes(ctdb, patch_number);
	    db_adj = determine_db_adj(px, py, ctdb);
	    current_pa = patch_adj + patch_number;
	    if (!te)
	    {
#if CTDB_REFORM_DEBUG
		fprintf(stderr,
			"Warning: Patch Number %d has no terrain data\n",
			patch_number);	
#endif
		create_te_mapping(patch_number,0,0,ctdb);
		current_pa->pa_east = NULL;
		current_pa->pa_west = NULL;
		current_pa->pa_north = NULL;
		current_pa->pa_south = NULL;
		continue;
	    }
	    
	    /* allocate te adj information areas for patch */
	    current_pa->pa_east = (TE_ADJ_INFO *) 
	      ctdb_alloc(current_pa->pa_east = NULL, sizeof(TE_ADJ_INFO)*n_te,
			  "te adjacency information");
	    current_pa->east_end = 0;
	    current_pa->pa_west = (TE_ADJ_INFO *) 
	      ctdb_alloc(current_pa->pa_west = NULL, sizeof(TE_ADJ_INFO)*n_te,
			  "te adjacency information");
	    current_pa->west_end = 0;
	    current_pa->pa_north = (TE_ADJ_INFO *) 
	      ctdb_alloc(current_pa->pa_north = NULL, sizeof(TE_ADJ_INFO)*n_te,
			  "te adjacency information");
	    current_pa->north_end = 0;
	    current_pa->pa_south = (TE_ADJ_INFO *) 
	      ctdb_alloc(current_pa->pa_south = NULL, sizeof(TE_ADJ_INFO)*n_te,
			  "te adjacency information");
	    current_pa->south_end = 0;
	    
	    for (i=0;i<n_te;i++)
	    {   /* For each polygonal te determine adjaceny */
		if (te[i].info.vertices > 1 ) 
		{
		    te[i].info.patch_adj = 0;
#if CTDB_REFORM_DEBUG
		    printf("Looking at terrain element %d\n",i);
#endif
		    for (e=0;e<=te[i].info.vertices;e++) /* look at each side 
							  */
		    {
#if CTDB_REFORM_DEBUG
			printf("\tLooking at side %d\n",e);
#endif			
			 /* edge for side e */
			if ( ctdb_intern_get_te_edge_index16(te,i,e,edge) )
			  continue;
			
			if (te[i].adj[e] == EDGE_NOT_MAPPED_16)
			  associate_edge(edge,te,i+1,n_te, e, current_pa,
					 db_adj);
			
		    } /* for */
#if CTDB_REFORM_DEBUG
		    check_te_adj(te,i,patch_number);
#endif
		}
#if CTDB_REFORM_DEBUG
		else
		  printf("line or point\n");
#endif
		
	    } /* i */
	    
	    /* sort adjacent information lists for this patch */
	    qsort(current_pa->pa_east, current_pa->east_end,
		  sizeof(TE_ADJ_INFO), 
		  te_adj_cmp);
	    qsort(current_pa->pa_west, current_pa->west_end,
		  sizeof(TE_ADJ_INFO),
		  te_adj_cmp);
	    qsort(current_pa->pa_north, current_pa->north_end,
		  sizeof(TE_ADJ_INFO),
		  te_adj_cmp);
	    qsort(current_pa->pa_south, current_pa->south_end,
		  sizeof(TE_ADJ_INFO),
		  te_adj_cmp);
	    
	    create_te_mapping(patch_number,te,n_te,ctdb);
	    
	} /* px */
    } /* py */
    printf("\nCreating Patch boundary Adjacencies\n");
    /* loop back through patches, this time to resolve patch adjacencies */
    for (py=0; py < ctdb->patches_high; py++)
    {
	putchar('.');fflush(stdout);
	for (px=0; px < ctdb->patches_wide; px++)
	{
	    patch_number = py*ctdb->patches_wide + px;
#if CTDB_REFORM_DEBUG
	    printf("Patch Number %d\n", patch_number);
#endif
	    te = (TERRAIN_ELEMENT_16 *)
	      ctdb_intern_lookup_terrain_element(ctdb,patch_number);
	    if (!te)
	      /* No terrain elements -> patch is off the database */
	      continue;
	    wx = px*patch_size;
	    wy = py*patch_size;
	    current_pa = patch_adj + patch_number;
	    
	    /* if not west dbase bounded, process west patch edge */
	    if (wx > ctdb->min_x)
	    {
		nbor_num = py*ctdb->patches_wide + px - 1;
		nbor_pa = patch_adj + nbor_num;
		nbor_te = (TERRAIN_ELEMENT_16 *)
		  ctdb_intern_lookup_terrain_element(ctdb,nbor_num);
		if (nbor_te)
		  resolve_patch_neighbors(current_pa->pa_west,
					  current_pa->west_end,te,
					  nbor_pa->pa_east,nbor_pa->east_end,
					  nbor_te,Y_AXIS);
	    }
	    /* if not east dbase bounded, process east patch edge */
	    if (wx + patch_size < ctdb->max_x)
	    {
		nbor_num = py*ctdb->patches_wide + px + 1;
		nbor_pa = patch_adj + nbor_num;
		nbor_te = (TERRAIN_ELEMENT_16 *)
		  ctdb_intern_lookup_terrain_element(ctdb,nbor_num);
		if (nbor_te)
		  resolve_patch_neighbors(current_pa->pa_east,
					  current_pa->east_end,te,
					  nbor_pa->pa_west,nbor_pa->west_end,
					  nbor_te, Y_AXIS);
	    }
	    /* if not north dbase bounded, process north patch edge */
	    if (wy + patch_size < ctdb->max_y)
	    {
		nbor_num = (py + 1)*ctdb->patches_wide + px;
		nbor_pa = patch_adj + nbor_num;
		nbor_te = (TERRAIN_ELEMENT_16 *)
		  ctdb_intern_lookup_terrain_element(ctdb,nbor_num);
		if (nbor_te)
		  resolve_patch_neighbors(current_pa->pa_north,
					  current_pa->north_end,te,
					  nbor_pa->pa_south,nbor_pa->south_end,
					  nbor_te,X_AXIS);
	    }
	    /* if not south dbase bounded, process south patch edge */
	    if (wy > ctdb->min_y)
	    {
		nbor_num = (py - 1)*ctdb->patches_wide + px;
		nbor_pa = patch_adj + nbor_num;
		nbor_te = (TERRAIN_ELEMENT_16 *)
		  ctdb_intern_lookup_terrain_element(ctdb,nbor_num);
		if (nbor_te)
		  resolve_patch_neighbors(current_pa->pa_south,
					  current_pa->south_end,te,
					  nbor_pa->pa_north,nbor_pa->north_end,
					  nbor_te,X_AXIS);
	    }
	    
	    
	} /* px */
    } /* py */
    printf("\n");
    
    /* free memory for adjacency storage */
    current_pa = patch_adj;
    for (py=0; py < ctdb->patches_high; py++)
      for (px=0; px < ctdb->patches_wide; px++)
      {
	  if (current_pa->pa_east != NULL)
	    free(current_pa->pa_east);
	  if (current_pa->pa_north != NULL)
	    free(current_pa->pa_north);
	  if (current_pa->pa_west != NULL)
	    free(current_pa->pa_west);
	  if (current_pa->pa_south != NULL)
	    free(current_pa->pa_south);
	  current_pa++;
      }
    free(patch_adj);
}

/* This routine checks the completeness of the terrain topology for the given
   patch and reports any polygon that does not have its adjacencies completely
   defined. */
static void
check_te_adj(te,i,patch_number)
    TERRAIN_ELEMENT_16  *te;
    int32                i;
    int32                patch_number;
{
    int32 j;
    
    for(j=0; j<=te[i].info.vertices; j++)
    {
	if ((te[i].adj[j] == EDGE_NOT_MAPPED_16) &&  /* not mapped and */
	    !(te[i].info.patch_adj & (1<<j)))      /* not adj to patch bound */
	{
	    printf("\nError with adjacencies of patch %d element %d\n",
		   patch_number,i);
	    return;
	}
    }
}

#define ON_DB_XEDGE(x, db) (x == db->min_x || x == db->max_x)
#define ON_DB_YEDGE(y, db) (y == db->min_y || y == db->max_y)
/*
   This routine will take the input edge and compare it with the other
   edges to be found in the remaining terrain elements.                
   If a match is found both the input terrain element and the 
   matching terrain element egde are mapped to one another.  If
   an edge doesn't get mapped it is because the edge maps to the edge
   of the patch or the mapping has already been performed.
   */

#define BOUNDRY_WEST DB_BOUNDRY_16
#define BOUNDRY_EAST DB_BOUNDRY_16
#define BOUNDRY_NORTH DB_BOUNDRY_16
#define BOUNDRY_SOUTH DB_BOUNDRY_16
static void 
associate_edge(edge0,te,start,n_te, side, pa, db_adj)
    int32                edge0[];
    TERRAIN_ELEMENT_16  *te;
    int32                start;
    int32                n_te;
    int32                side;
    PATCH_ADJ_INFO      *pa;
    CTDB_BEARING         db_adj;
{
    int32         edge1[2];
    int32         s,bs;
    uint32        i;
    uint32        base = start-1;
    int32 failed=TRUE;
    int32         te0 = edge0[0];
    int32         te1 = edge0[1];
    int32         cside;
    
    
    /* Check for patch and database boundry conditions */
    if ( te[te0].pt.x == te[te1].pt.x )
    {
	if ( (db_adj & WEST) &&
	    (te[te0].pt.x == 0))
	{
	    te[base].adj[cside = get_side_index(te,base,edge0) ] = 
	      BOUNDRY_WEST;
	    te[base].info.patch_adj |= 1<<cside; /* set patch adj bit */
	    return;
	}
	else if ( (db_adj & EAST ) &&
		 ( te[te0].pt.x == (int32)CTDB_PATCH_TO_XYO_SCALE ))
	{
	    te[base].adj[cside=get_side_index(te,base,edge0) ] = BOUNDRY_EAST;
	    te[base].info.patch_adj |= 1<<cside; /* set patch adj bit */
	    return;
	}
	else if (te[te0].pt.x == 0)
	{ 
	    set_adj_info(pa->pa_west, &(pa->west_end), &te[te0].pt, 
			 &te[te1].pt, base, side);
	    te[base].info.patch_adj |= 1<<side; /* set patch adj bit */
	    return;
	}
	else if ( te[te0].pt.x == (int32)CTDB_PATCH_TO_XYO_SCALE )
	{
	    set_adj_info(pa->pa_east, &(pa->east_end), &te[te0].pt, 
			 &te[te1].pt, base, side);
	    te[base].info.patch_adj |= 1<<side; /* set patch adj bit */
	    return;
	}
    }
    else if ( te[te0].pt.y == te[te1].pt.y )
    {
	if ( ( db_adj & SOUTH ) &&
	    (te[te0].pt.y ==  0))
	{
	    te[base].adj[cside=get_side_index(te,base,edge0) ] = BOUNDRY_SOUTH;
	    te[base].info.patch_adj |= 1<<cside; /* set patch adj bit */
	    return;
	}
	else if ( (db_adj & NORTH ) &&
		 ( te[te0].pt.y == (int32)CTDB_PATCH_TO_XYO_SCALE ))
	{
	    te[base].adj[cside=get_side_index(te,base,edge0) ] = BOUNDRY_NORTH;
	    te[base].info.patch_adj |= 1<<cside; /* set patch adj bit */
	    return;
	}
	else if ( te[te0].pt.y == 0 )
	{ 
	    set_adj_info(pa->pa_south, &(pa->south_end), &te[te0].pt, 
			 &te[te1].pt, base, side);
	    te[base].info.patch_adj |= 1<<side; /* set patch adj bit */
	    return;
	}
        else if ( te[te0].pt.y ==  (int32)CTDB_PATCH_TO_XYO_SCALE )
	{ 
	    set_adj_info(pa->pa_north, &(pa->north_end), &te[te0].pt, 
			 &te[te1].pt, base, side);
	    te[base].info.patch_adj |= 1<<side; /* set patch adj bit */
	    return;
	}
    }
    
    for (i=start;i<n_te;i++)
    {
	if (te[i].info.vertices < 2)  /* don't map pts and edges */
	  continue;
	for (s=0;s<=te[i].info.vertices;s++)
	  if ( !ctdb_intern_get_te_edge_index16(te,i,s,edge1) )
	    if ( ctdb_check_edge(edge0,edge1,te) )
	      if ( te[i].adj[s] == EDGE_NOT_MAPPED_16 )
	      {
		  te[i].adj[s] = base ;
		  te[base].adj[side] = i ;	    	      
		  te[i].info.adj_sides |= (side << (s * 2));
		  te[base].info.adj_sides |= (s << (side * 2));
		  return;
		}
    }
}

/* This routine records the adjacency information at the patch boundary */

static void
set_adj_info(adj_info, adj_index, pt1 , pt2, te_index, side)
    TE_ADJ_INFO         *adj_info;
    uint16              *adj_index;
    CTDB_XYO_COORD2D     *pt1, *pt2;
    int32                te_index;
    int32                side;
{
    adj_info[*adj_index].minx = NS_MIN(pt1->x,pt2->x) ;
    adj_info[*adj_index].miny = NS_MIN(pt1->y,pt2->y) ;
    adj_info[*adj_index].maxx = NS_MAX(pt1->x,pt2->x) ;
    adj_info[*adj_index].maxy = NS_MAX(pt1->y,pt2->y) ;
    adj_info[*adj_index].te_index = (uint16)te_index;
    adj_info[*adj_index].entry_side = (uint16)side;
    (*adj_index)++;
}

/* This routine loops through the two edge lists corresponding to the TE sides
   that are on a shared patch boundary, and resolves the adjacency information
   between them. That is, this routine records the patch boundary TE 
   adjacencies between two adjacent patches. */

static void
resolve_patch_neighbors(side1,num1,te1,side2,num2,te2, axis)
    TE_ADJ_INFO        *side1;
    int32               num1;
    TERRAIN_ELEMENT_16 *te1;
    TE_ADJ_INFO        *side2;
    int32               num2;
    TERRAIN_ELEMENT_16 *te2;
    int32               axis;
{
    int32 i,j;
    
    i=j=0;
    while(i<num1 && j <num2)
    {
	/* check side1 = side2 */
	if (axis == Y_AXIS)
	{
	    if ((side1->miny == side2->miny) &&
		(side1->maxy == side2->maxy))
	    {
		te1[side1->te_index].adj[side1->entry_side] = side2->te_index;
		te2[side2->te_index].adj[side2->entry_side] = side1->te_index;
		te1[side1->te_index].info.adj_sides |= 
		  (side2->entry_side << (side1->entry_side * 2));
		te2[side2->te_index].info.adj_sides |= 
		  (side1->entry_side << (side2->entry_side * 2));
		i++;
		j++;
		side1++;
		side2++;
	    }
	    else if ((side1->miny == side2->miny) &&
		     !(side1->maxy == side2->maxy))
	    { 
		i++;
		j++;
		side1++;
		side2++;
	    }
	    else if (side1->miny < side2->miny)
	    {
		side1++;
		i++;
	    }
	    else 
	    {
		j++;
		side2++;
	    }
	}
	else
	{ /* axis = X_AXIS */
	    if ((side1->minx == side2->minx) &&
		(side1->maxx == side2->maxx))
	    {
		te1[side1->te_index].adj[side1->entry_side] = side2->te_index;
		te2[side2->te_index].adj[side2->entry_side] = side1->te_index;
		te1[side1->te_index].info.adj_sides |= 
		  (side2->entry_side << (side1->entry_side * 2));
		te2[side2->te_index].info.adj_sides |= 
		  (side1->entry_side << (side2->entry_side * 2));
		i++;
		j++;
		side1++;
		side2++;
	    }
	    else if ((side1->minx == side2->minx) &&
		     !(side1->maxx == side2->maxx))
	    { 
		i++;
		j++;
		side1++;
		side2++;
	    }
	    else if (side1->minx < side2->minx)
	    {
		i++;
		side1++;
	    }
	    else 
	    {
		side2++;
		j++;
	    }
	}
    }
}


/* This routine determines if the given point is on a patch boundary and 
   returns the specification (CTDB_BEARING) of which patch boundary(ies) it 
   resides. */
static CTDB_BEARING 
determine_db_adj(px, py, ctdb)
    int32           px;
    int32           py;
    CTDB_CMP       *ctdb;
{
    int32 wx,wy;
    int32 patch_size = ctdb->patch_incr*ctdb->incr;
    
    wx = px*patch_size;
    wy = py*patch_size;
    
    /* determine patch to db-base boundaries */
    if (wx == ctdb->min_x && wy == ctdb->min_y)
      return SOUTHWEST;
    if (wx == ctdb->min_x && wy + patch_size == ctdb->max_y)
      return NORTHWEST;
    else if ((wx + patch_size == ctdb->max_x) &&
	     (wy + patch_size == ctdb->max_y))
      return NORTHEAST;
    else if ((wx + patch_size == ctdb->max_x) &&
	     (wy == ctdb->min_y))
      return SOUTHEAST;
    else if (wx == ctdb->min_x) 
      return WEST;
    else if (wy == ctdb->min_y)
      return SOUTH;
    else if (wx + patch_size == ctdb->max_x)
      return EAST;
    else if  (wy + patch_size == ctdb->max_y)
      return NORTH;
    else return ILLEGAL;
}


/*
 * This routine will return the side index for the input edge and terrain
 * element.
 */
static int32 
get_side_index(te,index,e0)
    TERRAIN_ELEMENT_16 *te;/* terrain element which contains the edge */
    uint32              index; /* vertex number of input te */
    int32               e0[];  /* edge to evaluate */
{
    int32 s;
    int32 n_side=te[index].info.vertices+1;
    int32 e1[2];
    
    for (s=0;s<n_side;s++)
    {
	ctdb_intern_get_te_edge_index16(te,index,s,e1);
	if ( check_edge_by_index(e0,e1) )
	  return(s);
    }
    
    fprintf(stderr,"Error: Side not found\n");
    return(0); 
}

/*
   This routine will place a full word template over the (x,y) point and 
   compare the two edge points to determine if the edge is identical 
   */

static int32 
check_edge_by_index(e0,e1)
    int32 e0[];
    int32 e1[];
{
    return ((int32)( ( (e0[0] == e1[0]) && ( e0[1] == e1[1]) ) ||
			    ( (e0[1] == e1[0]) && (e0[0] == e1[1]) ) ) );
}


/* This compare routine is used by the search routine to order the list by 
   min x then min y values. */
static int 
te_adj_cmp(te_adj1, te_adj2)
    TE_ADJ_INFO *te_adj1;
    TE_ADJ_INFO *te_adj2;
{
    if (te_adj1->minx < te_adj2->minx)
      return -1;
    else if ((te_adj1->minx == te_adj2->minx) &&
	     (te_adj1->miny < te_adj2->miny))
      return -1;
    else if ((te_adj1->minx == te_adj2->minx) &&
	     (te_adj1->miny == te_adj2->miny))
      return 0;
    else return 1;
}


/* For Virtual Grid Routines */


/* The create_te_mapping routine constructs and assigns values to the virtual
   grid for the given patch. */
static void 
create_te_mapping(patch_number,te,n_te,ctdb)
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    CTDB_CMP           *ctdb;
{
    CTDB_PATCH_GROUP_CMP *patch_group=ctdb->patch_groups+
      PG_NUMBER(patch_number);
    int32 member=PG_MEMBER(patch_number);
    int32 n_polygons=0; 
    int32 max_area=0;
    int32 area=0;
    int32 t,i,dim,sqdim;
    int32 n_tri,r,c;
    int32 spacing;
    CTDB_COORD2D pt;
    
    AREA G ; /* Initial Virtual Grid Square structure */
    AREA I; 
    AREA T[2];
    uint16 *te_map;
    
    if ( n_te == 0 )
    {
	patch_group->te_vgrid_dim[member] = 0;
	patch_group->te_vgrid_offset[member] = 0;
	return ;
    }
    
    for (i=0;i<n_te;i++)
      if (te[i].info.vertices > 1 )  
        n_polygons++;
    
#ifdef __alpha
    dim = CTDB_RND( fsqrt((float)n_polygons) );
#else
    dim = CTDB_RND( sqrt(n_polygons) );
#endif
    if (dim == 0)
      dim = 1;
    sqdim = dim*dim;
    spacing=CTDB_PATCH_TO_XYO_SCALE/dim;
    
    /*    ALLOCATE PERMINANT MEMORY FOR TE MAPPING IN PATCH GROUP, AND 
	  TEMPORARY MEMORY FOR VIRTUAL_GRID AND THE TERRAIN_ELEMENT 
	  */
    te_map = 
      (uint16 *)ctdb_alloc(0,sizeof(uint16)*sqdim,"te map");
    
    patch_group->te_vgrid_dim[member] = dim;   
    patch_group->te_vgrid_offset[member] = te_map;
    
    /* continue to try and allocate space on failure */
    while ( !ctdb_init_area_struct(CTDB_PTS_PER_QUADRALATERAL+3,&G) );
    while ( !ctdb_init_area_struct(CTDB_PTS_PER_TRIANGLE+3,&T[0]) );
    while ( !ctdb_init_area_struct(CTDB_PTS_PER_TRIANGLE+3,&T[1]) );
    while ( !ctdb_init_area_struct(CTDB_MAX_INTERSECTION_PTS+3,&I) ) ;  
    
    /* look through each virtual grid square */
    
    for (r=0;r<dim;r++)
    {
	for (c=0;c<dim;c++,te_map++)
	{
	    
	    /* create an explicit grid square */
	    
	    G.count=0;/* Reset Point Counter */ 
	    pt.x   = CTDB_RND(c*spacing);   
	    pt.y   = CTDB_RND(r*spacing);
	    ctdb_add_point(&G,pt.x, pt.y);
	    
	    pt.x   = CTDB_RND(c*spacing);
	    pt.y   = CTDB_RND((r+1)*spacing);
	    ctdb_add_point(&G,pt.x, pt.y);
	    
	    pt.x   = CTDB_RND((c+1)*spacing);
	    pt.y   = CTDB_RND((r+1)*spacing);
	    ctdb_add_point(&G,pt.x, pt.y);
	    
	    pt.x   = CTDB_RND((c+1)*spacing);
	    pt.y   = CTDB_RND(r*spacing);
	    ctdb_add_point(&G,pt.x, pt.y);
	    ctdb_find_area_extents(&G);
	    
	    max_area = 0;
	    *te_map  = ADJT_NOT_MAPPED_16;
	    
	    /* for each terrain element, if the element intersects the 
	       grid square, then determine it area covering the grid */
	    
	    for (t=0;t<n_te;t++)
	    {
		area = 0 ;
		n_tri=0 ;
		
		switch (te[t].info.vertices)
		{
		  case 3 :
		    I.count=0;/* Reset Point Counters */
		    ctdb_format_terrain_element(te,t,&T[n_tri],n_tri);
		    ctdb_find_edge_intersection(&G,
					        &T[n_tri],
						&I );
		    area = ctdb_calculate_area(&I);
		    n_tri++;
		    
		  case 2 : 
		    I.count=0; /* Reset Point Counters */
		    ctdb_format_terrain_element(te,t,&T[n_tri],n_tri);
		    ctdb_find_edge_intersection(&G,
						&T[n_tri],
						&I );
		    area += ctdb_calculate_area(&I) ;
		    break;
		  default : 
		    area = 0 ;
		}
		
		if (area > max_area)
		{
		    max_area = area ;
		    *te_map = t ;
		}
	    }
	}
    }
    
    ctdb_free_area_struct(&T[0]);
    ctdb_free_area_struct(&T[1]);
    ctdb_free_area_struct(&G);
    ctdb_free_area_struct(&I);
}
/*
 * This routine formats a terrain element into a structure which can be used by
 * geometry intersection routines.
 */
void
ctdb_format_terrain_element(te,offset,A,A_index)
    TERRAIN_ELEMENT_16 *te;
    int32               offset;
    AREA               *A;
    int32               A_index;    
{  
    CTDB_COORD2D pt;
    
    A->count = 0;
    
    if ( A_index == 0 )
    {
	pt.x = te[offset].pt.x;
	pt.y = te[offset].pt.y;
	ctdb_add_point(A,pt.x, pt.y);
	
	pt.x = te[ te[offset].vert.index[0] ].pt.x;
	pt.y = te[ te[offset].vert.index[0] ].pt.y;
	ctdb_add_point(A,pt.x, pt.y);
	
	pt.x = te[ te[offset].vert.index[1] ].pt.x;
	pt.y = te[ te[offset].vert.index[1] ].pt.y;
	ctdb_add_point(A,pt.x, pt.y);
	ctdb_find_area_extents(A);
    }
    else
    {
	pt.x = te[offset].pt.x;
	pt.y = te[offset].pt.y;
	ctdb_add_point(A,pt.x, pt.y);
	
	pt.x = te[ te[offset].vert.index[0] ].pt.x;
	pt.y = te[ te[offset].vert.index[0] ].pt.y;
	ctdb_add_point(A,pt.x, pt.y);
	
	pt.x = te[ te[offset].vert.index[2] ].pt.x;
	pt.y = te[ te[offset].vert.index[2] ].pt.y;
	ctdb_add_point(A,pt.x, pt.y);
    }
    ctdb_find_area_extents(A);
}

/* This routine redefines the terrain elements for a complete patch as a set of
   terrain elements that use 8 bit indices. */
void 
convert_16bit_to_8bit_te(te16, n_te, te8)
    TERRAIN_ELEMENT_16  *te16;
    int32                n_te;
    TERRAIN_ELEMENT_08  *te8;
{
    int32 i,j;
    
    for (i=0; i < n_te ; i++ )
    {
	bcopy(&te16[i].pt,&te8[i].pt,sizeof(CTDB_XYO_COORD2D));
	te8[i].z = te16[i].z;
	for (j=0;j<TE_ADJ_TRI_MAP_DIMENSION; j++)
	  te8[i].adj[j] = te16[i].adj[j];
	for (j=0;j<TE_VERT_DIMENSION; j++ )
	  te8[i].vert.index[j] =  te16[i].vert.index[j];
	bcopy(&te16[i].info,&te8[i].info,sizeof(TE_INFO)) ;
    }
}

void 
convert_08_to_16te(te08, te16)
    TERRAIN_ELEMENT_08 *te08;
    TERRAIN_ELEMENT_16 *te16;
{
    int32 j;

    bcopy(&te08->pt,&te16->pt,sizeof(CTDB_XYO_COORD2D));
    te16->z = te08->z;
    for (j=0;j<TE_ADJ_TRI_MAP_DIMENSION; j++)
      te16->adj[j] = te08->adj[j];
    for (j=0;j<TE_VERT_DIMENSION; j++ )
      te16->vert.index[j] =  te08->vert.index[j];
    bcopy(&te08->info,&te16->info,sizeof(TE_INFO)) ;
}

/* Store the specified TEs (which are assumed to be 16 bit) as-is to the
 * output TE file, and update the ctdb as necessary.
 */
void
ctdb_store_patch_tes(ctdb, patch_number, tes, n_te, max_elev, f)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    float64            *max_elev;
    FILE               *f;
{
    int32 i;

    if (ctdb->elevations)
      record_micro_present(ctdb, patch_number, tes, n_te);

    *max_elev = tes[0].z;
    
    for(i=1;i<n_te;i++)
      if (tes[i].z > *max_elev)
	*max_elev = tes[i].z;
    
    fwrite(&tes[0],sizeof(TERRAIN_ELEMENT_16), n_te, f);

    return;
}

