#ifndef lint
static char rcsid [] = "$RCSfile: ctdbdiff.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ctdbdiff.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to compare two databases                             *
*       Created: Mon Dec  5 1994                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libctdb.h>
#include <ct_post.h>
#include <stdext.h>

static CTDB ctdb[2];

static void print_lines(label,
			n_verts0, verts0, tree_tops0,
			n_verts1, verts1, tree_tops1)
    char   *label;
    int32   n_verts0;
    float64 verts0[CTDB_VERTICES_MAX][2];
    float64 tree_tops0[CTDB_VERTICES_MAX];
    int32   n_verts1;
    float64 verts1[CTDB_VERTICES_MAX][2];
    float64 tree_tops1[CTDB_VERTICES_MAX];
{
    int32 i;

    printf("%s\n", label);
    for (i=0;(i<n_verts0)||(i<n_verts1);i++)
    {
	if (i<n_verts0)
	  printf("\t<%9.2f %9.2f %7.2f>",
		 verts0[i][0], verts0[i][1], tree_tops0[i]);
	else
	  printf("\t                             ");

	if (i<n_verts1)
	  printf("\t<%9.2f %9.2f %7.2f>\n",
		 verts1[i][0], verts1[i][1], tree_tops1[i]);
	else
	  printf("\t                             \n");
    }

    printf("\n");
}

static void compare_trees()
{
    char buf[256];
    struct
    {
	CTDB_SEARCH_SPACE_PTR space;
	int32   n_verts;
	float64 verts[CTDB_VERTICES_MAX][2];
	float64 tree_tops[CTDB_VERTICES_MAX];
	float64 foliage_height;
	float64 radius;
    } per[2];
    int32 i, sanity;
    int32 pn0, pn1;

    for (i=0;i<2;i++)
      per[i].space =
	ctdb_create_search_space(&ctdb[i],
				 0, 0, ctdb[i].max_x, ctdb[i].max_y,
				 /*CTDB_TREELINE_TREES |*/ CTDB_BY_PATCH);

    while (ctdb_next_patch(per[0].space, &pn0) &&
	   ctdb_next_patch(per[1].space, &pn1))
    {
	sanity = 0;
	while ((pn0 != pn1) && (sanity < 10))
	{
	    printf("Traversal out of sync <%d %d> <%d %d>\n",
		   pn0 % ctdb[0].patches_wide, pn0 / ctdb[0].patches_wide,
		   pn1 % ctdb[1].patches_wide, pn1 / ctdb[1].patches_wide);

	    if (pn0 > pn1)
	      ctdb_next_patch(per[1].space, &pn1);
	    else
	      ctdb_next_patch(per[0].space, &pn0);

	    sanity++;
	}
	if (pn0 != pn1)
	{
	    printf("Unable to resync traversal\n");
	    break;
	}

	for(;;)
	{
	    for (i=0;i<2;i++)
	      per[i].n_verts =
		ctdb_next_tree(per[i].space, per[i].verts, per[i].tree_tops,
			       &per[i].foliage_height, &per[i].radius);

	    if (!per[0].n_verts && !per[1].n_verts)
	      break;

	    if (per[0].n_verts != per[1].n_verts)
	    {
		if (!per[0].n_verts && per[1].n_verts)
		{
		    printf("%s has more trees, but %s is finished <%d %d>\n",
			   ctdb[0].name, ctdb[1].name,
			   pn0 % ctdb[0].patches_wide,
			   pn0 / ctdb[0].patches_wide);
		    break;
		}
		if (per[0].n_verts && !per[1].n_verts)
		{
		    printf("%s has more trees, but %s is finished <%d %d>\n",
			   ctdb[1].name, ctdb[0].name,
			   pn0 % ctdb[0].patches_wide,
			   pn0 / ctdb[0].patches_wide);
		    break;
		}

		print_lines("Treelines with differing # verts:",
			    per[0].n_verts, per[0].verts, per[0].tree_tops,
			    per[1].n_verts, per[1].verts, per[1].tree_tops);
		break;
	    }

	    if (bcmp((void*)per[0].verts, (void*)per[1].verts,
		     per[0].n_verts*2*sizeof(float64)))
	    {
		print_lines("Treelines with differing XY:",
			    per[0].n_verts, per[0].verts, per[0].tree_tops,
			    per[1].n_verts, per[1].verts, per[1].tree_tops);
		break;
	    }

	    if (bcmp((void*)per[0].tree_tops, (void*)per[1].tree_tops,
		     per[0].n_verts*sizeof(float64)))
	    {
		print_lines("Treelines with differing Z:",
			    per[0].n_verts, per[0].verts, per[0].tree_tops,
			    per[1].n_verts, per[1].verts, per[1].tree_tops);
		break;
	    }

	    if (per[0].radius != per[1].radius)
	    {
		sprintf(buf, "Treelines with differing Radius (%g != %g):",
			per[0].radius, per[1].radius);
		print_lines(buf,
			    per[0].n_verts, per[0].verts, per[0].tree_tops,
			    per[1].n_verts, per[1].verts, per[1].tree_tops);
		continue;
	    }

	    if (per[0].foliage_height != per[1].foliage_height)
	    {
		sprintf(buf, "Treelines with differing Foliage (%g != %g):",
			per[0].foliage_height, per[1].foliage_height);
		print_lines(buf,
			    per[0].n_verts, per[0].verts, per[0].tree_tops,
			    per[1].n_verts, per[1].verts, per[1].tree_tops);
		continue;
	    }
	}
    }

    for (i=0;i<2;i++)
      ctdb_destroy_search_space(per[i].space);
}

static void compare_buildings()
{
    char buf[256];
    struct
    {
	CTDB_SEARCH_SPACE_PTR space;
	int32   n_verts;
	float64 verts[CTDB_VERTICES_MAX][2];
	float64 roof_line[CTDB_VERTICES_MAX];
    } per[2];
    int32 i;

    for (i=0;i<2;i++)
      per[i].space =
	ctdb_create_search_space(&ctdb[i],
				 0, 0, ctdb[i].max_x, ctdb[i].max_y,
				 0);

    for(;;)
    {
	for (i=0;i<2;i++)
	  per[i].n_verts =
	    ctdb_next_building(per[i].space, per[i].verts, per[i].roof_line);

	if (!per[0].n_verts && !per[1].n_verts)
	  break;

	if (per[0].n_verts != per[1].n_verts)
	{
	    if (!per[0].n_verts && per[1].n_verts)
	    {
		printf("%s has more buildings, but %s is finished\n",
		       ctdb[0].name, ctdb[1].name);
		break;
	    }
	    if (per[0].n_verts && !per[1].n_verts)
	    {
		printf("%s has more buildings, but %s is finished\n",
		       ctdb[1].name, ctdb[0].name);
		break;
	    }

	    print_lines("Buildings with differing # verts:",
			per[0].n_verts, per[0].verts, per[0].roof_line,
			per[1].n_verts, per[1].verts, per[1].roof_line);
	    continue;
	}

	if (bcmp((void*)per[0].verts, (void*)per[1].verts,
		 per[0].n_verts*2*sizeof(float64)))
	{
	    print_lines("Buildings with differing XY:",
			per[0].n_verts, per[0].verts, per[0].roof_line,
			per[1].n_verts, per[1].verts, per[1].roof_line);
	    continue;
	}

	if (bcmp((void*)per[0].roof_line, (void*)per[1].roof_line,
		 per[0].n_verts*sizeof(float64)))
	{
	    print_lines("Buildings with differing Z:",
			per[0].n_verts, per[0].verts, per[0].roof_line,
			per[1].n_verts, per[1].verts, per[1].roof_line);
	    continue;
	}
    }

    for (i=0;i<2;i++)
      ctdb_destroy_search_space(per[i].space);
}

static void compare_flags()
{
    int32 i, posts;
    int32 page, x, y;
    uint32 p0, p1;

    if (!ctdb[0].elevations_resident ||
	!ctdb[1].elevations_resident)
    {
	printf("Require resident elevations for post comparison test\n");
	return;
    }

    posts = ctdb[0].pages_wide * ctdb[0].pages_high * POSTS_PER_PAGE;
    for (i=0;i<posts;i++)
    {
	p0 = ctdb[0].elevations[i];
	p1 = ctdb[1].elevations[i];

	if ((p0 & POST_FLAGS) != (p1 & POST_FLAGS))
	{
	    page = i / POSTS_PER_PAGE;
	    x = ctdb->incr *
	      ((page % ctdb[0].pages_wide) * POSTS_PER_SIDE +
	       (i % POSTS_PER_PAGE));
	    y = ctdb->incr *
	      ((page / ctdb[0].pages_wide) * POSTS_PER_SIDE +
	       ((i - page * POSTS_PER_PAGE) / POSTS_PER_SIDE));

	    printf("Differing Flags at %6d %6d: |%c%c%c%c| |%c%c%c%c|\n",
		   x, y,
		   (p0 & POST_MICROTERRAIN) ? 'M' : ' ',
		   (p0 & POST_BUILDINGS)    ? 'B' : ' ',
		   (p0 & POST_TREES)        ? 'T' : ' ',
		   (p0 & POST_FEATURES)     ? 'F' : ' ',
		   (p1 & POST_MICROTERRAIN) ? 'M' : ' ',
		   (p1 & POST_BUILDINGS)    ? 'B' : ' ',
		   (p1 & POST_TREES)        ? 'T' : ' ',
		   (p1 & POST_FEATURES)     ? 'F' : ' ');
	}
    }
}

void main (
    int32 argc,
    argv_t argv)
{
    if (argc != 3)
    {
	char *temp;
	char name1[CTDB_NAME_LENGTH + 4], name2[CTDB_NAME_LENGTH + 4];
	
	temp = ctdb_generate_tdb_filename(CTDB_MODE_SIMNET, "<path>", "<db1>",
					  CTDB_FILE_FORMAT, 0               , 
					  FALSE);

	strcpy(name1, temp);

	temp = ctdb_generate_tdb_filename(CTDB_MODE_SIMNET, "<path>", "<db2>",
					  CTDB_FILE_FORMAT, 0,
					  FALSE);
	
	strcpy(name2, temp);

	printf("Usage: %s %s %s\n", argv[0], name1, name2);
	exit(0);
    }

    ctdb_read(argv[1],&ctdb[0],0);
    ctdb_print_description(&ctdb[0]);
    ctdb_read(argv[2],&ctdb[1],0);
    ctdb_print_description(&ctdb[1]);

    compare_trees();
    compare_buildings();
    compare_flags();
}
