#ifndef lint
static char rcsid [] = "$RCSfile: ictopo.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ictopo.c                                                          *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *                                                                           *
 *       Contents  : Compilation post-processor which interconnects linear   *
 *                   feature topology between GCS cells                      *
 *       Created   : November 1995                                           *
 *       Author    : flchamberlain                                           *
 ****************************************************************************/

#include <stdext.h>
#include "ct_cmplr.h"
#include "ct_reform.h"

/* Update as latest format changes */
#include "ct_cmp_fmt6.h"

#include <libgcs.h>
#include <math.h>

/* NOTE:  This routine currently keeps all cells in memory at all times.  If
 * ever becomes impractical, the alternative is to do a multi-pass system,
 * reading in databases one at a time, and saving intermediate results to
 * a file.
 */

#define DEBUG

#define DATA_PATH "../../../data"

#define MAX_FNAME_LEN 255
#define MAX_BDY_DELTA 1.0		/* How far a point can be from the
					 * boundary (in meters) and still be
					 * considered to be "on" the boundary.
					 */

typedef struct db
{
    CTDB  ctdb;
    char  fname[MAX_FNAME_LEN];
    FILE *out_file;
} DB;

typedef struct boundary_pt
{
    float64 x,y,z;		/* GCC point location */
    int32   from_cell;	        /* Cell where bdy point occurs.  Note that this
				 * actually stores the index into the cell_dbs
				 * array, not the cell id. 
				 */
    int32   to_cell;		/* Cell with corresponding node, not 
				 * necessarily on boundary.  Note that this
				 * actually stores the index into the cell_ids
				 * array, not the cell_id
				 */
    int32   node_index;		/* Index of node in to_cell */
} BOUNDARY_PT;

typedef struct plist
{
    BOUNDARY_PT    *points;
    int32           num_points;	/* Current number of points */
    int32           max_points;	/* Max points */
} PLIST;


/* Globals */
DB *cell_dbs;			/* List of cell databases we're dealing with */
int32 ndb;			/* Length of above list */
PLIST bp_list;			/* List of all points on boundaries of all
				 * cells
				 */
#define START_NUM_PTS 1000	/* Initial length of bp_list */
#define PLIST_INCR 1000		/* How much to increment list length by when
				 * we run over current allocation*/

PLIST *cell_pts;		/* Per-cell list of boundary points, filled 
				 * out with node indices in neighoring cells
				 * corresponding to boundary nodes in the 
				 * given cell.
				 */


#define TDB_MSIZE 0       	/* Amount of memory to allow each db.  0
				 * disables caching entirely (this code may
				 * not currently work with caching enabled...)
				 */

/* $$$ Code that still needs to be written */

/* Finds all abstract features for which topology is stored (currently only
 * offroad segments) along which the point (x,y,z) lies, breaks them at
 * the point, and updates the database topology.  This routine does *not*
 * update the abstract data for the ctdb in place--instead, it updates two
 * global structures, abstracts_to_delete and abstracts_to_add, listing the
 * abstracts to replace, and what replaces them, respectively.
 */
static void
add_pt_to_abstract_linears(ctdb, x, y, z)
    CTDB   *ctdb;
    float64 x;
    float64 y;
    float64 z;
{

}

/* $$$ END CODE THAT NEEDS TO BE WRITTEN */

static void
add_bp(list, pt)
    PLIST          *list;
    BOUNDARY_PT    *pt;
{
    if (list->num_points >= list->max_points)
    {
	list->max_points += PLIST_INCR;
	list->points = 
	  (BOUNDARY_PT *)ctdb_alloc(list->points,
				    list->max_points * 
				    sizeof(BOUNDARY_PT),
				    "Boundary point list");
    }
    list->points[(list->num_points)++] = *pt;
    return;
}

/* Initialize global variables */
static void
ictopo_init(num_cells, data_path)
    int32 num_cells;    
    char *data_path;
{
    int32 i;

    ndb = num_cells;
    cell_dbs = (DB *)ctdb_alloc(NULL, ndb * sizeof(DB), 
				"Cell list");

    bp_list.max_points = START_NUM_PTS;
    bp_list.num_points = 0;
    bp_list.points = (BOUNDARY_PT *)ctdb_alloc(NULL, bp_list.max_points * 
					       sizeof(BOUNDARY_PT),
					       "Boundary point list");

    cell_pts = (PLIST *)ctdb_alloc(NULL, ndb * sizeof(PLIST),
				   "Cell point list array");
    
    for(i=0;i<ndb;i++)
    {
	cell_pts[i].num_points = 0;
	cell_pts[i].max_points = START_NUM_PTS;
	cell_pts[i].points = 
	  (BOUNDARY_PT *)ctdb_alloc(NULL, cell_pts[i].max_points * 
				    sizeof(BOUNDARY_PT),
				    "Cell point list array");
    }
						
    reader_init(0);
    
    gcs_init(coord_get_datum_info);

    return;
}

/* Read in the cell databases.  Assumes memory has been allocated for global
 * cell array
 */
static void
read_cells(names, num_names)
    char **names;
    int32 num_names;
{
    int32 i;
    int32 cell_id;
    float64 origin_x, origin_y;
    
    printf("Reading input databases.  Please be patient...\n");
    
    for(i=0;i<num_names;i++)
    {
	strncpy(cell_dbs[i].fname, names[i], MAX_FNAME_LEN);
	ctdb_read(names[i], &cell_dbs[i].ctdb, TDB_MSIZE);
	ctdb_print_description(&cell_dbs[i].ctdb);
	cell_id = ctdb_get_cell_id(&cell_dbs[i].ctdb);
	gcs_init_cell(cell_id, NULL);
	gcs_set_tdb(cell_id, (ADDRESS)&cell_dbs[i].ctdb, FALSE);
	ctdb_get_origin(&cell_dbs[i].ctdb, &origin_x, &origin_y);
	gcs_offset_origin(cell_id, origin_x, origin_y);	
    }
    
    return;
}


/* Return TRUE iff node is on a boundary of ctdb */
static int32 
node_on_boundary(ctdb, node)
    CTDB *ctdb;
    CTDB_NODE *node;
{
    /* You would think that we could test for (rough) equality with
     * the database boundary values, but we can't for two reasons:
     * 1.  Abstracts don't get clipped to database boundaries
     * 2.  To simplify intervis and lookup, there are patches beyond the
     *     min/max x/y.  It is these patches that physical features get
     *     clipped to, not the official database "boundaries"
     */
    return ((node->x <= (ctdb->min_x + MAX_BDY_DELTA)) ||
	    (node->x >= (ctdb->max_x - MAX_BDY_DELTA)) ||
	    (node->y <= (ctdb->min_y + MAX_BDY_DELTA)) ||
	    (node->y >= (ctdb->max_y - MAX_BDY_DELTA)));
}

/* Find all nodes on boundary of database cell_dbs[index] and add them to 
 * the global bp_list
 */
static void
find_boundary_nodes(index)
    int32 index;
{
    int32 i;
    CTDB *ctdb = &cell_dbs[index].ctdb;
    CTDB_NODE *node;
    BOUNDARY_PT bp;
    int32 cell_id = ctdb_get_cell_id(ctdb);

    for(i=0;i<ctdb->num_nodes;i++)
    {
	node = ctdb->nodes[i];
	if (node_on_boundary(ctdb, node))
	{
	    /* We'd really like to do an elevation lookup here in order to
	     * do a more accurate coordinate conversion, but, by definition,
	     * we're off the database, so we can't.
	     */
	    bp.from_cell = index;
	    gcs_gcs3e_to_gcc(cell_id,
			   (float64)node->x, (float64)node->y, 
			   (float64)0.0,
			   &bp.x, &bp.y, &bp.z);

	    /* Fill these last two out just to be safe */
	    bp.to_cell = -1;
	    bp.node_index = -1;
	    add_bp(&bp_list, &bp);
	}
    }
    return;
}


/* Returns TRUE if the point is "in" the specified database.  Note that "in"
 * is defined based on the *patch* boundaries, *not* min/max x/y, since
 * linear feature inclusion is based on patch boundaries
 */
static int32
pt_in_db(ctdb, x, y)
    CTDB        *ctdb;
    float64      x;
    float64      y;
{
    float64 minx, maxx, miny, maxy;
    int32 cell_id = ctdb_get_cell_id(ctdb);


    /* Calculate the database bounds we want to use */
    minx = ctdb->min_x;		/* Compiler currently pads only east and north
				 */
    miny = ctdb->min_y;
    maxx = ctdb->patches_wide * ctdb->incr * ctdb->patch_incr;
    maxy = ctdb->patches_high * ctdb->incr * ctdb->patch_incr;

    return ((x >= minx) && (x <= maxx) && (y >= miny) && (y <= maxy));
}

/* Returns TRUE if x1 and y1 are within MAX_BDY_DELTA of x2 and y2,
 * respectively
 */
static int32 
pts_equiv(x1, y1, x2, y2)
    float64 x1;
    float64 y1;
    float64 x2;
    float64 y2;
{
    return((fabs(x1 - x2) < MAX_BDY_DELTA) &&
	   (fabs(y1 - y2) < MAX_BDY_DELTA));
}

#define ILLEGAL_NODE -1
#define ILLEGAL_EDGE -1

/* Return the node index, if any, of the node in the given database at (x,y).
 * If no such node exists, return ILLEGAL_NODE.
 */
static int32
find_node_index(ctdb, x, y)
    CTDB   *ctdb;
    float64 x;
    float64 y;
{
    int32 i;

    for(i=0;i<ctdb->num_nodes;i++)
    {
	if (pts_equiv(x, y, 
		      (float64)ctdb->nodes[i]->x, (float64)ctdb->nodes[i]->y))
	{
	    return i;
	}
    }
    return ILLEGAL_NODE;
}


/* This function is a wrapper around ctdb_point_intersection_t to give
 * it the interface I want
 */
static CTDB_POINT_STATE
point_on_seg(s0, e0, pt, radius, tseg)
    CTDB_XYO_COORD2D s0;
    CTDB_XYO_COORD2D e0;
    CTDB_XYO_COORD2D pt;
    int32            radius;
    float64         *tseg;
{
    float64 ns0[2];
    float64 ne0[2];
    float64 px, py;
    float64 nrad;
    float64 tdummy;
    
    ns0[X] = s0.x;
    ns0[Y] = s0.y;
    
    ne0[X] = e0.x;
    ne0[Y] = e0.y;
    
    px = pt.x;
    py = pt.y;
    
    nrad = radius;
    
    return ctdb_point_intersection_t(ns0, ne0, px, py, nrad, 0.0, tseg,
				     &tdummy);
}

/* Return a pointer to the feature data for the first linear feature at or 
 * after start that contains the point (norm_x, norm_y).  (If there is no such
 * feature, return end).  This is based largely on the function 
 * ctdb_lookup_soil_feature in ct_micro.c.  Note that start is assumed to
 * point to the start of a valid feature
 * vert returns the index of the first vertex of the segment within the
 * feature at which the intersection occurs, and tseg returns the fractional
 * distance along that segment at which the point intersects.
 */
static CTDB_FEATURE_DATA*
next_linear_at_pt(ctdb, start, end, norm_x, norm_y, vert, tseg)
    CTDB              *ctdb;
    CTDB_FEATURE_DATA *start;
    CTDB_FEATURE_DATA *end;
    float64            norm_x;
    float64            norm_y;
    int32             *vert;
    float64           *tseg;
{
    int32 patchx, patchy, patch_number;
    int32 post_mask;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *next;
    CTDB_XYO_COORD2D pt;
    int32 i,radius,fudge;
    int32 px, py;
    float64 tdummy;

    px = POST_X(norm_x);
    py = POST_Y(norm_y);

    /* What patch is this post on? */
    patchx = PATCH_X(norm_x*ctdb->patch_inv_incr);
    patchy = PATCH_Y(norm_y*ctdb->patch_inv_incr);
    patch_number = patchx + patchy * ctdb->patches_wide;

    if (!start)
      return end;

    /* The post_mask gives us a bounding box check in one easy operation */
    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);

    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can work in integer.
     */
    pt.x = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
    pt.y = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);

    feature = start;

    for(;feature<end;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	if (fclass == FC_MICRO)
	{
	    next = feature + NEXT_MICRO(verts);
	    continue;
	}
	else if (fclass == FC_CANOPY)
	{
	    next = feature + NEXT_CANOPY(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME)
	{
	    next = feature + NEXT_VOLUME(verts);
	    continue;
	}
	else if (fclass == FC_LINEAR)
	{
	    next = feature + NEXT_LINEAR(verts);
	    continue;
	}
	else if (fclass == FC_AGGREGATE)
	{
	    next = feature + NEXT_AGGREGATE(verts);
	    continue;
	}
	/* If it falls through to here, it's a laid linear */
	next = feature + NEXT_LAID_LINEAR(verts);

	/* NOTE:  do NOT do a test against the post mask here--remember that
	 * we have some "slop" in our point location--a little "slop" can put
	 * us on the wrong post, invalidating the test
	 */

	radius = feature[2].laid_linear.width / 2;
	/* Take into account conversion problems we may have had in getting 
	 * here
	 */
	fudge = rint(MAX_BDY_DELTA * (float64)ctdb->meters_to_patch * 
		     (float64)CTDB_PATCH_TO_XYO_SCALE);
	
	radius += fudge;

	for (i=0;i<verts-1;i++)
	{
	    /* Are we on this one? */
	    if (point_on_seg(feature[i+3].xy, feature[i+4].xy, pt, radius,
			     tseg)
		== INTERSECTION)
	    {
		*vert = i;
		return feature;
	    }
	}
    }
    return end;
}


#define T_SLOP .0001		/* How close (fractionally) a point must be
				 * to the beginning or end of segment to be
				 * considered "on" the point
				 */

/* Break the linear feature pointed to by feat at its vert'th index, a
 * tseg'th of the way along it.  Create (this includes mallocing space)
 * two new features if necessary, one for everything before the break,
 * the other for everything after it, putting pointers to the new features
 * in new_feat1 and new_feat2, respectively.  If the point happens to
 * fall on an endpoint of the segment, set *new_feat1 and *new_feat2 to
 * NULL and return.
 */
static void
break_linear_feature(ctdb, feat, vert, tseg, new_feat1, new_feat2)
    CTDB               *ctdb;
    CTDB_FEATURE_DATA  *feat;
    int32               vert;
    float64             tseg;
    CTDB_FEATURE_DATA **new_feat1;
    CTDB_FEATURE_DATA **new_feat2;
{
    int32 f1_verts;
    int32 f2_verts;
    CTDB_XYO_COORD2D new_pt;
    CTDB_FEATURE_DATA *cur;
    CTDB_FEATURE_DATA_CMP *cur_cmp; /* We need to call a compiler routine that
				     * demands this type
				     */
    int32 i;
    
    if (((vert == 0) && (tseg <= T_SLOP)) ||
	((vert == (feat->info.vertices - 2)) && (tseg >= (1.0 - T_SLOP))))
    {
	/* We're on an endpoint */
	*new_feat1 = NULL;
	*new_feat2 = NULL;
	return;
    }

    /* Fill in new point */
    new_pt.x = (uint16)(feat[vert+2].xy.x + 
			(tseg * (feat[vert+3].xy.x - feat[vert+2].xy.x)));
    new_pt.y = (uint16)(feat[vert+2].xy.y + 
			(tseg * (feat[vert+3].xy.y - feat[vert+2].xy.y)));

    /* Allocate space for new features */
    f1_verts = vert + 2;
    f2_verts = feat->info.vertices - vert;
    
    *new_feat1 = 
      (CTDB_FEATURE_DATA *)ctdb_alloc(NULL, 
				      (f1_verts+2) * sizeof(CTDB_FEATURE_DATA),
				      "New linear feature");

    *new_feat2 = 
      (CTDB_FEATURE_DATA *)ctdb_alloc(NULL, 
				      (f2_verts+2) * sizeof(CTDB_FEATURE_DATA),
				      "New linear feature");
    
    /* Fill in first new feature */
    cur = *new_feat1;
    cur->info.feature_class = feat->info.feature_class;
    cur->info.vertices = f1_verts;
    cur->info.post_mask = 0;
    cur++;
    
    cur->ll_soil = feat[1].ll_soil;
    cur++;
    
    cur->laid_linear = feat[2].laid_linear;
    /* We'll fill in the edge_key later */
    cur++;
    
    for(i=0;i<=vert;i++,cur++)
    {
	cur->xy = feat[i+3].xy;
    }
    
    cur->xy = new_pt;

    /* Now fill in the post mask */
    cur = *new_feat1;
    cur_cmp = (CTDB_FEATURE_DATA_CMP *)cur;
    for(i=0;i<f1_verts-1;i++)
    {	
	(*new_feat1)->info.post_mask |=
	  ctdb_cmp_compute_line_post_mask((CTDB_CMP *)ctdb, 
					  cur_cmp[i+3], cur_cmp[i+4]);
    }

    /* Fill in second new feature */
    cur = *new_feat2;
    cur->info.feature_class = feat->info.feature_class;
    cur->info.vertices = f2_verts;
    cur->info.post_mask = 0;
    cur++;
    
    cur->ll_soil = feat[1].ll_soil;
    cur++;

    cur->laid_linear = feat[2].laid_linear;
    /* We'll fill in the edge_key later */
    cur++;
    
    cur->xy = new_pt;

    cur++;
    
    for(i=1;i<f2_verts;i++,cur++)
    {
	cur->xy = feat[i+3+vert].xy;
    }

    
    /* Now fill in the post mask */
    cur = *new_feat2;
    cur_cmp = (CTDB_FEATURE_DATA_CMP *)cur;
    for(i=0;i<f2_verts-1;i++)
    {
	(*new_feat2)->info.post_mask |=
	  ctdb_cmp_compute_line_post_mask((CTDB_CMP *)ctdb, 
					  cur_cmp[i+3], cur_cmp[i+4]);
    }

    return;
}

/* Fill out new_pg based on new_pg, except that the feature size for patch
 * member member should increase by size_delta bytes.  Return the address
 * immediately following the data header.
 */
static ADDRESS
fill_out_tin_pg_header(old_pg, new_pg, member, size_delta)
    struct ctdb_tin_pg_data  *old_pg;
    struct ctdb_tin_pg_data  *new_pg;
    int32                     member;
    int32                     size_delta;
{
    int32 offset = 0;
    int32 i;

    bcopy(old_pg, new_pg, sizeof(CTDB_TIN_PG_DATA));
    
    /* Fix patch offsets */
    if (member == 0)
    {
	offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    }
    
    new_pg->patch1_offset = old_pg->patch1_offset + offset;
    
    if (member == 1)
    {
	offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    }

    new_pg->patch2_offset = old_pg->patch2_offset + offset;

    if (member == 2)
    {
	offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    }

    new_pg->patch3_offset = old_pg->patch3_offset + offset;

    offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    
    /* Fix te offsets */
    for(i=0;i<CTDB_PATCHES_PER_GROUP;i++)
    {
	new_pg->te_offset[i] = old_pg->te_offset[i] + offset;
    }
    
    /* Fix te vgrid offsets */
    for(i=0;i<CTDB_PATCHES_PER_GROUP;i++)
    {
	new_pg->te_vgrid_offset[i] = old_pg->te_vgrid_offset[i] + offset;
    }
    
    return (ADDRESS)(&new_pg->te_vgrid_offset[CTDB_PATCHES_PER_GROUP]);    
}


/* Fill out new_pg based on old_pg, except that the feature size for patch
 * member member should increase by size_delta bytes.  Return the address
 * immediately following the data header.
 */
static ADDRESS
fill_out_grid_pg_header(old_pg, new_pg, member, size_delta)
    struct ctdb_grid_pg_data  *old_pg;
    struct ctdb_grid_pg_data  *new_pg;
    int32                     member;
    int32                     size_delta;
{
    int32 offset = 0;

    bcopy(old_pg, new_pg, sizeof(CTDB_GRID_PG_DATA));
    
    /* Fix patch offsets */
    if (member == 0)
    {
	offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    }
    
    new_pg->patch1_offset = old_pg->patch1_offset + offset;
    
    if (member == 1)
    {
	offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    }

    new_pg->patch2_offset = old_pg->patch2_offset + offset;

    if (member == 2)
    {
	offset = size_delta / sizeof(CTDB_FEATURE_DATA);
    }

    new_pg->patch3_offset = old_pg->patch3_offset + offset;

    return (ADDRESS)(&new_pg->tables[CTDB_PATCHES_PER_GROUP + 2]);
}

/* Copy the features from in (in will always point to a patch header) to
 * out, replacing the feature at feat with new_feat1 and new_feat2 (feat
 * must always be among the features in patch that in represents).  Update
 * *in and *out to point the start of the new patches in their respective
 * data arrays.  Note:  this routine *does* update the size field in the
 * new patch header.
 * $$$ for now, this routine assumes that old_feat, new_feat1, and new_feat2
 * are all linear features.  This would be easy to fix...
 */
static void
copy_feat_insert(in, out, old_feat, new_feat1, new_feat2, new_nf1, new_nf2)
    CTDB_FEATURE_DATA  **in;
    CTDB_FEATURE_DATA  **out;
    CTDB_FEATURE_DATA  *old_feat;
    CTDB_FEATURE_DATA  *new_feat1;
    CTDB_FEATURE_DATA  *new_feat2;
    CTDB_FEATURE_DATA **new_nf1;
    CTDB_FEATURE_DATA **new_nf2;
{
    int32               patch_len;
    int32 len;
    int32 left;
    int32 size_delta;

    patch_len = (*in)->hdr.size * sizeof(CTDB_FEATURE_DATA);

    size_delta = 2 + (new_feat1->info.vertices + new_feat2->info.vertices -
		      old_feat->info.vertices);
    
    (*in)->hdr.size += size_delta;

    len = (int8 *)old_feat - (int8 *)(*in);
    bcopy((void *)(*in), (void *)(*out), len);
    *in = (CTDB_FEATURE_DATA *)((int8 *)(*in) + len);
    *out = (CTDB_FEATURE_DATA *)((int8 *)(*out) + len);
    left = patch_len - len;
    
    *new_nf1 = *out;
    len = (2 + new_feat1->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    bcopy((void *)new_feat1, (void *)(*out), len);
    *out = (CTDB_FEATURE_DATA *)((int8 *)(*out) + len);
    
    *new_nf2 = *out;
    len = (2 + new_feat2->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    bcopy((void *)new_feat2, (void *)(*out), len);
    *out = (CTDB_FEATURE_DATA *)((int8 *)(*out) + len);
    
    left -= ((2 + old_feat->info.vertices) * sizeof(CTDB_FEATURE_DATA));
    (*in) += (2 + old_feat->info.vertices);
    bcopy((void *)(*in), (void *)(*out), left);
    *in = (CTDB_FEATURE_DATA *)((int8 *)(*in) + left);
    *out = (CTDB_FEATURE_DATA *)((int8 *)(*out) + left);
    
    return;
}


/* Transfers the PG data pointed to by old_pg_hdr to new_pg_hdr, replacing
 * old_feat in patch member with new_feat1 and new_feat2.  old_size specifies
 * the total size of the data pointed to by old_pg_hdr.  Returns the change
 * in size (+ => new is bigger than old).  The absolute addresses where
 * the new features end up are returned in *new_nf1 and *new_nf2
 */
static int32
update_tin_pg_data(pg, old_pg_hdr, new_pg_hdr, member, old_feat, new_feat1,
		   new_feat2, new_nf1, new_nf2, old_size)
    CTDB_PATCH_GROUP          *pg;
    struct ctdb_tin_pg_data   *old_pg_hdr;
    struct ctdb_tin_pg_data  **new_pg_hdr;
    int32                      member;
    CTDB_FEATURE_DATA         *old_feat;
    CTDB_FEATURE_DATA         *new_feat1;
    CTDB_FEATURE_DATA         *new_feat2;
    CTDB_FEATURE_DATA        **new_nf1;
    CTDB_FEATURE_DATA        **new_nf2;
    int32                      old_size;
{
    CTDB_FEATURE_DATA *cur_in_pos, *cur_out_pos;
    int32 feat_size, nf1_size, nf2_size;
    int32 size_delta;	/* # of bytes by which PG data will increase */
    CTDB_TIN_PG_DATA *npg;
    int32 len;

    feat_size = (2 + old_feat->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    nf1_size = (2 + new_feat1->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    nf2_size = (2 + new_feat2->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    /* I think this should always be 16, but we'll calculate it, just to be
     * safe.
     */
    size_delta = nf1_size + nf2_size - feat_size;
    
    npg = (CTDB_TIN_PG_DATA *)ctdb_alloc(NULL, old_size + size_delta, 
					 "New PG data");    

    *new_pg_hdr = npg;

    cur_out_pos = (CTDB_FEATURE_DATA *)fill_out_tin_pg_header(old_pg_hdr, 
							      npg, member, 
							      size_delta);
    
    cur_in_pos = (CTDB_FEATURE_DATA *)
      (&(old_pg_hdr->te_vgrid_offset[CTDB_PATCHES_PER_GROUP]));
    
    /* Copy features over, inserting where necessary */
    if (pg->patch0_occupied)
    {
	if (member != 0)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (CTDB_FEATURE_DATA *)((int8 *)cur_in_pos + len);
	    cur_out_pos = (CTDB_FEATURE_DATA *)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos, old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }
    
    
    if (pg->patch1_occupied)
    {
	if (member != 1)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (ADDRESS)((int8 *)cur_in_pos + len);
	    cur_out_pos = (ADDRESS)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos,  old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }
    
    if (pg->patch2_occupied)
    {
	if (member != 2)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (ADDRESS)((int8 *)cur_in_pos + len);
	    cur_out_pos = (ADDRESS)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos, old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }
    
    if (pg->patch3_occupied)
    {
	if (member != 3)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (ADDRESS)((int8 *)cur_in_pos + len);
	    cur_out_pos = (ADDRESS)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos, old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }

    /* Copy remaining data */
    len = old_size - ((int8 *)cur_in_pos - (int8 *)old_pg_hdr);
    bcopy(cur_in_pos, cur_out_pos, len);

    return size_delta;
}

/* Transfers the PG data pointed to by old_pg_hdr to new_pg_hdr, replacing
 * old_feat in patch member with new_feat1 and new_feat2.  old_size specifies
 * the total size of the data pointed to by old_pg_hdr.  Returns the change
 * in size (+ => new is bigger than old).  The absolute addresses where the
 * new features end up are put in *new_nf1 and *new_nf2.
 */
static int32
update_grid_pg_data(pg, old_pg_hdr, new_pg_hdr, member, old_feat, new_feat1,
		    new_feat2, new_nf1, new_nf2, old_size)
    CTDB_PATCH_GROUP           *pg;
    struct ctdb_grid_pg_data   *old_pg_hdr;
    struct ctdb_grid_pg_data  **new_pg_hdr;
    int32                       member;
    CTDB_FEATURE_DATA          *old_feat;
    CTDB_FEATURE_DATA          *new_feat1;
    CTDB_FEATURE_DATA          *new_feat2;
    CTDB_FEATURE_DATA         **new_nf1;
    CTDB_FEATURE_DATA         **new_nf2;
    int32                       old_size;
{
    CTDB_FEATURE_DATA *cur_in_pos, *cur_out_pos;
    int32 feat_size, nf1_size, nf2_size;
    int32 size_delta;	/* # of bytes by which PG data will increase */
    CTDB_GRID_PG_DATA *npg;
    int32 len;

    feat_size = (2 + old_feat->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    nf1_size = (2 + new_feat1->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    nf2_size = (2 + new_feat2->info.vertices) * sizeof(CTDB_FEATURE_DATA);
    /* I think this should always be 16, but we'll calculate it, just to be
     * safe.
     */
    size_delta = nf1_size + nf2_size - feat_size;
    
    npg = (CTDB_GRID_PG_DATA *)ctdb_alloc(NULL, old_size + size_delta, 
					  "New PG data");    

    *new_pg_hdr = npg;

    cur_out_pos = (CTDB_FEATURE_DATA *)fill_out_grid_pg_header(old_pg_hdr, 
							       npg, member, 
							       size_delta);
    
    cur_in_pos = (CTDB_FEATURE_DATA *)
      ((int8 *)old_pg_hdr + sizeof(CTDB_GRID_PG_DATA));
    
    /* Copy features over, inserting where necessary */
    if (pg->patch0_occupied)
    {
	if (member != 0)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (CTDB_FEATURE_DATA *)((int8 *)cur_in_pos + len);
	    cur_out_pos = (CTDB_FEATURE_DATA *)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos, old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }
    
    
    if (pg->patch1_occupied)
    {
	if (member != 1)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (ADDRESS)((int8 *)cur_in_pos + len);
	    cur_out_pos = (ADDRESS)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos,  old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }
    
    if (pg->patch2_occupied)
    {
	if (member != 2)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (ADDRESS)((int8 *)cur_in_pos + len);
	    cur_out_pos = (ADDRESS)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos, old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }
    
    if (pg->patch3_occupied)
    {
	if (member != 3)
	{
	    len = cur_in_pos->hdr.size * sizeof(CTDB_FEATURE_DATA);
	    bcopy(cur_in_pos, cur_out_pos, len);
	    cur_in_pos = (ADDRESS)((int8 *)cur_in_pos + len);
	    cur_out_pos = (ADDRESS)((int8 *)cur_out_pos + len);
	}
	else
	{
	    copy_feat_insert(&cur_in_pos, &cur_out_pos, old_feat, 
			     new_feat1, new_feat2, new_nf1, new_nf2);
	}
    }

    return size_delta;
}


/* Calculate how many bytes of data are pointed to by pg->data_start.  pg_num
 * is the patch group index of PG.  WARNING:  This routine only works if
 * caching is diabled.
 */
static int32 
pg_bytes(ctdb, pg, pg_num)
    CTDB             *ctdb;
    CTDB_PATCH_GROUP *pg;
    int32             pg_num;
{
    int32 size;

    if (pg->contains_tins)
    {
	CTDB_TIN_PG_DATA *pgt = pg->data_start.tin_data;
	int32 mem = -1;
	int32 num_vg;

	if (!pg->patch3_no_tes)
	{
	    mem = 3;
	}
	else if (!pg->patch2_no_tes)
	{
	    mem = 2;
	}
	else if (!pg->patch1_no_tes)
	{
	    mem = 1;
	}
	else
	{
	    mem = 0;
	}

	size = pgt->te_vgrid_offset[mem] * sizeof(CTDB_FEATURE_DATA);
	num_vg = pgt->te_vgrid_dim[mem] * pgt->te_vgrid_dim[mem];
	/* Round up to word boundary, remembering that te_vgrid entries
	 * are uint16s
	 */
	num_vg += (num_vg % 2);
	/* te_vgrid entries are uint16's */
	size += (num_vg * sizeof(uint16));
	return size;
    }
    else
    {
	/* Find the last occupied patch.  Add its offset to the number of bytes
	 * of data stored for it to get the total PG data size
	 */
	CTDB_FEATURE_DATA *feat = NULL;
	if (pg->patch3_occupied)
	{
	    feat = ctdb_lookup_feature(ctdb, 
				       (pg_num * CTDB_PATCHES_PER_GROUP) + 3);
	    size = pg->data_start.grid_data->patch3_offset *
	      sizeof(CTDB_FEATURE_DATA);
	}
	else if (pg->patch2_occupied)
	{
	    feat = ctdb_lookup_feature(ctdb, 
				       (pg_num * CTDB_PATCHES_PER_GROUP) + 2);
	    size = pg->data_start.grid_data->patch2_offset *
	      sizeof(CTDB_FEATURE_DATA);
	}
	else if (pg->patch1_occupied)
	{
	    feat = ctdb_lookup_feature(ctdb, 
				       (pg_num * CTDB_PATCHES_PER_GROUP) + 1);
	    size = pg->data_start.grid_data->patch1_offset *
	      sizeof(CTDB_FEATURE_DATA);
	}
	else if (pg->patch0_occupied)
	{
	    feat = ctdb_lookup_feature(ctdb, 
				       (pg_num * CTDB_PATCHES_PER_GROUP));

	    size = sizeof(CTDB_GRID_PG_DATA);
	}
	if (feat == NULL)
	  return sizeof(CTDB_GRID_PG_DATA);

	size += (feat->hdr.size * sizeof(CTDB_FEATURE_DATA));
	return size;
    }
}


/* Regenerate the PG data for the PG in which patch resides, replacing
 * feat in patch_number with new_feat1 and new_feat2.  The PG pointer
 * in ctdb will be set to point to the new data.  Currently assumes that
 * all three features are linears, though it wouldn't be hard to update
 * it to be more general.  This will update ctdb->n_patche_bytes and, if
 * necessary, some caching structures.  The absolute addresses where the
 * two new features end up will be returned in *new_nf1 and *new_nf2.
 */
static void
update_pg_data(ctdb, patch_number, feat, new_feat1, new_feat2, 
	       new_nf1, new_nf2)
    CTDB              *ctdb;
    int32              patch_number;
    CTDB_FEATURE_DATA *feat;
    CTDB_FEATURE_DATA *new_feat1;
    CTDB_FEATURE_DATA *new_feat2;
    CTDB_FEATURE_DATA **new_nf1;
    CTDB_FEATURE_DATA **new_nf2;
{
    CTDB_PATCH_GROUP *pg;
    int32 pg_num = PG_NUMBER(patch_number);
    int32 pg_mem = PG_MEMBER(patch_number);
    int32 pg_data_size;
    int32 i;
    CTDB_PG_DATA_HDR_PTR new_pg;
    int32 cache_size;
    int32 size_delta;

    pg = ctdb->patch_groups + pg_num;

    pg_data_size = pg_bytes(ctdb, pg, pg_num);

    if (pg->contains_tins)
    {
	size_delta = update_tin_pg_data(pg, pg->data_start.tin_data, 
					&new_pg.tin_data, pg_mem,
					feat, new_feat1, new_feat2, 
					new_nf1, new_nf2, pg_data_size);
    }
    else
    {
	size_delta =  update_grid_pg_data(pg, pg->data_start.grid_data, 
					  &new_pg.grid_data, 
					  pg_mem, feat, new_feat1, new_feat2, 
					  new_nf1, new_nf2, pg_data_size);
    }
    
    ctdb->n_patch_bytes += size_delta;
    
    cache_size = (pg_data_size + size_delta) / sizeof(CTDB_CACHE_BUCKET *);

    /* By increasing the amount of data in the PG, we may have bumped it up
     * to the next cache size
     */
    for(i = pg->cache;
	(i < CTDB_PATCH_CACHES) && 
	  (cache_size > ctdb->patch_cache_size_map[i]);
	i++);
    
    if (i == CTDB_PATCH_CACHES)
    {
	/* If this was the biggest patch and it just fit into its cache,
	 * we'll have to increase the biggest cache size
	 */
	i--;
	ctdb->patch_cache_size_map[i] = cache_size;
    }
    
    pg->cache = i;

    /* Finally, make the PG pointer in ctdb point to our new data */
    pg->data_start = new_pg;

    return;
}

/* Update the edge that currently points to the feature at patch_num,
 * offset to point to patch_num, new_offset.  edge_key is the last 12
 * bits of the edge's index.
 */
static void
change_edge_reference(ctdb, patch_num, offset, new_offset, edge_key)
    CTDB *ctdb;
    int32 patch_num;
    int32 offset;
    int32 new_offset;
    int32 edge_key;
{
    int32 edge_index;
    int32 edge_sub_index;

    if (!ctdb_find_edge(ctdb, edge_key, patch_num, offset, 
			&edge_index, &edge_sub_index))
    {
	printf("Warning: failed to find edge when changing reference\n");
	return;
    }

    ctdb->edges[edge_index]->feature[edge_sub_index].patch_offset = new_offset;
    
    return;
}


/* Loop through any linear features from base to end, updating the
 * references to them in the topology section of ctdb (the reason we're
 * here is because we added some stuff right before base, so the references
 * to the features following it will be wrong.
 */
static void
offset_edge_feature_references(ctdb, patch_number, start_offset,
			       incr, base, end)
    CTDB              *ctdb;
    int32              patch_number;
    int32              start_offset;
    int32              incr;
    CTDB_FEATURE_DATA *base;
    CTDB_FEATURE_DATA *end;
{
    int32 offset = start_offset;
    int32 next_offset;
    CTDB_FEATURE_DATA *cur, *next;
    
    for(cur = base;cur < end;cur = next, offset = next_offset)
    {
	if (cur->info.feature_class != FC_LAID_LINEAR)
	{
	    /* We're done with the linears, so we're done */
	    break;
	}
	
	next_offset = NEXT_LAID_LINEAR(cur->info.vertices);
	next = cur + next_offset;
	
	change_edge_reference(ctdb, patch_number, offset, offset + incr,
			      (int32)cur[2].laid_linear.edge_key);
	
    }

    return;    
}


/* Return TRUE iff the first vertex of the linear feature in patch 
 * patch_number at offset offset is at (x,y) (in database coordinates)
 */
static int32
check_pt_at_beg(ctdb, patch_number, offset, x, y)
    CTDB   *ctdb;
    int32   patch_number;
    int32   offset;
    float64 x;
    float64 y;
    
{
    CTDB_FEATURE_DATA *feat;
    int32 patchx, patchy;
    float64 fx, fy;
    
    
    feat = ctdb_lookup_feature(ctdb, patch_number);
    feat += offset;
    
    patchx = patch_number % ctdb->patches_wide;
    patchy = patch_number / ctdb->patches_wide;
    
    /* Convert first vertex of feature to DB coordinates */
    fx = CTDB_XYO_TO_POST(feat[2].xy.x, patchx, ctdb->patch_incrD) *
      ctdb->incrD;

    fy = CTDB_XYO_TO_POST(feat[2].xy.y, patchy, ctdb->patch_incrD) *
      ctdb->incrD;
    
    return pts_equiv(x, y, fx, fy);
}

/* Calculate the total distance along edge # index in ctdb */
static float64
calc_edge_dist(ctdb, index)
    CTDB *ctdb;
    int32 index;
{
    int32 i;
    float64 dist = 0.0;
    int32 nverts;
    float64 *verts;
    float64 dx, dy, dz;
    float64 dist_sq;
    int32 vert;
    
    nverts = ctdb_expand_edge(ctdb, index, FALSE, &verts, NULL, NULL);

    for(i=0;i<nverts-1;i++)
    {
	vert = 2*i;
	
	dx = verts[vert+2] - verts[vert];
	dy = verts[vert+3] - verts[vert+1];
	dist_sq = dx*dx + dy*dy;
	dist += sqrt(dist_sq);
    }
    
    return dist;    
}

/* Iterate through the features referenced by e1 and set them to point back
 * to edge number index
 */
static void
reset_feature_backpointers(ctdb, e1, index)
    CTDB *ctdb;
    CTDB_EDGE *e1;
    int32      index;
{
    int32 i;
    CTDB_FEATURE_DATA *feat;
    int32 is_abstract;
    int32 verts;
	
    for(i=0;i<e1->num_features;i++)
    {
	is_abstract = CTDB_IS_QUAD_NODE(e1->feature[i].patch_number);

	if (!(feat = ctdb_lookup_feature(ctdb, e1->feature[i].patch_number)))
	{
	    printf("Warning:  could not find feature data for edge %d\n",
		   index);
	    continue;
	}

	feat += e1->feature[i].patch_offset;
	if (!is_abstract && feat->info.feature_class != FC_LAID_LINEAR)
	{
	    printf("Warning: bogus feature reference in edge %d\n", index);
	    continue;
	}
	
	if (is_abstract)
	{
	    verts = feat->abstract.num_vertices;
	    if (feat->abstract.type_code != CTDB_ABSTRACT_OFFROAD_SEGMENT)
	    {
		printf("Warning: topology references non linear abstract\n");
		continue;
	    }
            ((CTDB_ABSTRACT_DATA *)(&feat[1+verts+verts].abstract_data))->
              offroad.edge_number = index;
	}
	else
	{
	    feat[2].laid_linear.edge_key = (index & CTDB_EDGE_KEY - 1);
	}
    }
}


#define ILLEGAL_PATCH -1

/* Create a new node in ctdb at position (x,y).  This node occurs somewhere
 * along the edge with index edge_index and subindex edge_sub_index.  (Note
 * that we will have to break this edge in two at that subindex).  The 
 * subindex'th element of the edge will be replaced by the two features
 * specified by (f1_patch, f1_offset) and (f2_patch, f2_offset).  f1 must
 * must be the start of the original feature, and f2 must be the end of it.
 * If f1_patch is ILLEGAL_PATCH, then this is the special case where we're
 * not removing or adding any sub-edges, just splitting an edge at an 
 * existing break point.  In this case, the break should occur at 
 * whichever end of the sub-edge has a point at (x,y).
 * Note that must be careful not to change the indices of any existing edges
 * in the topology (except the one we're splitting), because doing so
 * would invalidate the back pointers in the linear features.
 * Note:  this function assumes that someone has thoughtfully realloced
 * the ctdb nodes and edges arrays so that there is guaranteed to be room
 * for the new references (this should speed things up considerably if
 * you end up calling this a lot).
 */
static void
add_topo_node(ctdb, x, y, edge_index, edge_sub_index, 
	      f1_patch, f1_offset, f2_patch, f2_offset)
    CTDB *ctdb;
    float64 x;
    float64 y;
    int32 edge_index;
    int32 edge_sub_index;
    int32 f1_patch;
    int32 f1_offset;
    int32 f2_patch;
    int32 f2_offset;
{
    int32 new_edge_index;
    int32 new_node_index;
    CTDB_NODE *n1;
    int32 n1_size;		/* In bytes */
    CTDB_EDGE *e1, *e2;
    int32 e1_size, e2_size;
    int32 i;
    CTDB_EDGE *in_edge;
    
    /* Here just for convenience */
    in_edge = ctdb->edges[edge_index];

    /* We're going to create one new edge index, and reuse the old one */
    new_edge_index = ctdb->num_edges;

    new_node_index = ctdb->num_nodes;

    /* Allocate space for new node and edges */
    n1_size = sizeof(CTDB_NODE) + sizeof(struct ctdb_connection);
    n1 = (CTDB_NODE *)ctdb_alloc(NULL, n1_size, "New node");
    
    /* e1 goes from the first node of the original to the new break point */
    /* We may overestimate the sizes of either or both of these, but that
     * doesn't matter
     */
    e1_size = sizeof(CTDB_EDGE) + 
      ((in_edge->num_features + 1) * 
       sizeof(struct ctdb_feature_reference));

    e1 = (CTDB_EDGE *)ctdb_alloc(NULL, e1_size, "New edge");

    e2_size = sizeof(CTDB_EDGE) + 
      ((in_edge->num_features + 1) * 
       sizeof(struct ctdb_feature_reference));
    e2 = (CTDB_EDGE *)ctdb_alloc(NULL, e2_size, "New edge");
    
    /* Reset the edge sizes so that we can calculate them as we go along */
    e2_size = e1_size = 
      sizeof(CTDB_EDGE) - sizeof(struct ctdb_feature_reference);

    /* Fill in the node.  This is easy, since all we need to know are the
     * edge indices
     */
    n1->x = x;
    n1->y = y;
    n1->is_bridge = FALSE;	/* If there were a bridge here, there would
				 * already have been a node, and we wouldn't
				 * be in this routine
				 */
    n1->num_connections = 2;
    
    /* Modify for new structure.  DO NOT succumb to the temptation to
     * put a link to the adjacent patch here--we're avoiding doing so
     * intentionally
     */
    n1->connection[0].node_in_other_cell = FALSE;
    n1->connection[0].node_index         = in_edge->first_node;
    n1->connection[0].edge.edge_index    = edge_index;
    n1->connection[1].node_in_other_cell = FALSE;
    n1->connection[1].node_index         = in_edge->last_node;
    n1->connection[1].edge.edge_index    = new_edge_index;

    /* Now we'll fill in the easy parts of the edges.  We'll figure out which
     * parts of the boundary go to which edges afterwards
     */
    e1->first_node = in_edge->first_node;
    e1->last_node = new_node_index;
    e1->num_features = 0;	/* We'll increment this as we add features */

#define ADD_TOPO_FEATURE(edge, feat)               \
    (edge)->feature[(edge)->num_features++] = feat

    for(i=0;i<edge_sub_index;i++)
    {
	ADD_TOPO_FEATURE(e1, in_edge->feature[i]);
	e1_size += sizeof(struct ctdb_feature_reference);
    }

    e2->first_node = in_edge->last_node;
    e2->last_node = new_node_index;
    e2->num_features = 0;
    
    for(i=in_edge->num_features-1;i>edge_sub_index;i--)
    {
	ADD_TOPO_FEATURE(e2, in_edge->feature[i]);
	e2_size += sizeof(struct ctdb_feature_reference);
    }
    
    /* Assign the last few features.  What we do here depends on whether
     * we have new features, or are just juggling old ones
     */
    if (f1_patch != ILLEGAL_PATCH)
    {
	struct ctdb_feature_reference f1, f2;
	
	f1.patch_number = f1_patch;
	f1.patch_offset = f1_offset;
	f1.reverse = in_edge->feature[edge_sub_index].reverse;

	f2.patch_number = f2_patch;
	f2.patch_offset = f2_offset;
	f2.reverse = in_edge->feature[edge_sub_index].reverse;
	
	if (in_edge->feature[edge_sub_index].reverse)
	{
	    ADD_TOPO_FEATURE(e1, f2);
	    e1_size += sizeof(struct ctdb_feature_reference);
	    ADD_TOPO_FEATURE(e2, f1);
	    e2_size += sizeof(struct ctdb_feature_reference);
	}
	else
	{
	    ADD_TOPO_FEATURE(e1, f1);
	    e1_size += sizeof(struct ctdb_feature_reference);
	    ADD_TOPO_FEATURE(e2, f2);
	    e2_size += sizeof(struct ctdb_feature_reference);
	}
    }
    else
    {
	int32 pt_at_beginning;
	
	pt_at_beginning = 
	  check_pt_at_beg(ctdb, in_edge->feature[edge_sub_index].patch_number,
			  in_edge->feature[edge_sub_index].patch_offset,
			  x, y);

	if ((in_edge->feature[edge_sub_index].reverse && pt_at_beginning) ||
	    (!in_edge->feature[edge_sub_index].reverse && !pt_at_beginning))
	{
	    ADD_TOPO_FEATURE(e1, in_edge->feature[edge_sub_index]);
	    e1_size += sizeof(struct ctdb_feature_reference);
	}
	else
	{
	    ADD_TOPO_FEATURE(e2, in_edge->feature[edge_sub_index]);
	    e2_size += sizeof(struct ctdb_feature_reference);
	}
    }

    /* Since we have reversed the feature ordering for e2, we have to go
     * back and flip all the "reverse" bits for its features
     */
    for(i=0;i<e2->num_features;i++)
    {
	e2->feature[i].reverse = !e2->feature[i].reverse;
    }
    
    /* Put pointers to the new edges and node into ctdb.  Once they're there, 
     * we can use API routines to simplify calculating the distance along
     * the edges.
     */
    ctdb->nodes[ctdb->num_nodes++] = n1;
    ctdb->node_size += n1_size;

    ctdb->edge_size -= (sizeof(CTDB_EDGE) + 
			((in_edge->num_features - 1) * 
			 sizeof(struct ctdb_feature_reference)));
    ctdb->edges[edge_index] = e1;
    ctdb->edge_size += e1_size;
    ctdb->edges[ctdb->num_edges++] = e2;
    ctdb->edge_size += e2_size;
    
    /* Figure out the lengthgs of our new edges */
    e1->distance = rint(calc_edge_dist(ctdb, edge_index));
    e2->distance = rint(calc_edge_dist(ctdb, ctdb->num_edges-1));

    /* Features have backpointers into the topology, may of which we have
     * broken.  To be safe, we'll go in and reset all of them for the edges
     * we changed/generated
     */
    reset_feature_backpointers(ctdb, e1, edge_index);
    reset_feature_backpointers(ctdb, e2, ctdb->num_edges - 1);

    return;
}



/* Finds all physical linear features that the point (x,y,z) lies along,
 * breaks them at the point, and updates the topology.  This will generate
 * a new patch group data structure for patch group in which the feature
 * lies.  The PG header in ctdb will be modified to point to the new data,
 * and the PG index will be added to the pgs_modified list.  Similarly,
 * this will result in significant modifactions to the topology edges.
 * Any edges that we need to store outside the "normal" array will be
 * be registered in the modified_edges array.
 * x, y, and z are in GCS local coordinates to start off.
 */
static void
add_pt_to_physical_linears(ctdb, x, y, z, pgs_modified, pgm_len, 
			   modified_edges, em_len)
    CTDB   *ctdb;
    float64 x;
    float64 y;
    float64 z;
    int32  *pgs_modified;
    int32  *pgm_len;
    int32  *modified_edges;
    int32  *em_len;
{
    int32 patch_number;
    int32 pg_mem;
    int32 pg_num;
    int32 patch_x, patch_y;
    float64 norm_x, norm_y;
    CTDB_FEATURE_DATA *feat, *last, *start;
    CTDB_FEATURE_DATA *new_feat1, *new_feat2;
    CTDB_FEATURE_DATA *new_nf1, *new_nf2;
    CTDB_FEATURE_DATA *base, *end;
    CTDB_FEATURE_DATA *next;
    int32 vert;
    float64 tseg;
    int32 feat_offset;
    int32 incr;

    norm_x = x * ctdb->inv_incr;
    norm_y = y * ctdb->inv_incr;

    patch_x = PATCH_X(x * ctdb->meters_to_patch);
    patch_y = PATCH_Y(y * ctdb->meters_to_patch);

    patch_number = (patch_y * ctdb->patches_wide) + patch_x;
    pg_num = PG_NUMBER(patch_number);
    pg_mem = PG_MEMBER(patch_number);

    feat = start = ctdb_lookup_feature(ctdb, patch_number);
    last = feat + feat->hdr.size;
    /* We need feat to point to the first feature in the list */
    feat++;

    while(TRUE) 
    {
	feat = next_linear_at_pt(ctdb, feat, last, norm_x, norm_y, &vert, 
				 &tseg);
	if (feat >= last)
	{
	    break;
	}
	next = feat + NEXT_LAID_LINEAR(feat->info.vertices);
	new_feat1 = new_feat2 = NULL;
	
	/* Break the feature at point, unless it happens to be on one
	 * of the endpoints of the feature already
	 */
	break_linear_feature(ctdb, feat, vert, tseg, &new_feat1, &new_feat2);
	
	/* Create new feature data for the patch group this feature
	 * lies in.
	 */
	if (new_feat1)
	{
	    pgs_modified[(*pgm_len)++] = pg_num;
	    update_pg_data(ctdb, patch_number, feat, new_feat1, new_feat2,
			   &new_nf1, &new_nf2);

	    /* Adding new features will break the feature references in the
	     * topology for any features after the ones we added.  Go
	     * fix 'em
	     */
	    /* These next few calculations give answers in terms of
	     * sizeof(CTDB_FEATURE_DATA), which is what we want
	     */
	    /* Old offset for the feature we changed */
	    feat_offset = feat - start;
	    /* Amount that and all subsequent offsets should be incremented by
	     */
	    incr = ((2 + new_feat1->info.vertices) + 
		    (2 + new_feat2->info.vertices) -
		    (2 + feat->info.vertices));
	    /* Base should point just past the end of the second new feature,
	     * and end should point just past the end of the features for
	     * this (newly generated) patch
	     */
	    base = new_nf2 + NEXT_LAID_LINEAR(new_nf2->info.vertices);

	    end = new_nf1 + (last - feat) + incr;
	    
	    /* Those happen to be the values we'll want to use on our next 
	     * iteration through
	     */
	    next = base;
	    last = end;	    
	    start = new_nf1 - (feat - start);

	    /* Since there probably won't be huge numbers of linears left 
	     * in the patch, we'll go through them and fix their edges 
	     * directly.  The alternative would be to loop through all 
	     * the edges looking for ones that referenced changed features, 
	     * but I think that would be slower
	     */
	    offset_edge_feature_references(ctdb, patch_number, 
					 feat_offset +
					 NEXT_LAID_LINEAR(feat->info.vertices),
					 incr, base, end);
					   
	}
	
	{	    
	    /* It's finally time to add our new node and stitch it into the
	     * exisiting topology.
	     */
	    int32 eindex;
	    int32 esubindex;
	    int32 offset1, offset2;
	    
	    /* Find the edge that points to the feature we split. */
	    if (!ctdb_find_edge(ctdb, feat[2].laid_linear.edge_key, 
				patch_number, feat_offset, &eindex,
				&esubindex))
	    {
		printf(": Couldn't find edge for split feature.\n");
		continue;
	    }
	    
	    /* Note that because of the way we added them in, the offset of
	     * the first new feature is identical to that for the old feature,
	     * and for the second feature we just need to add the size of
	     * the first one
	     */
	    modified_edges[(*em_len)++] = eindex;
	    modified_edges[(*em_len)++] = ctdb->num_edges;

	    if (new_feat1)
	    {		
		add_topo_node(ctdb, x, y, eindex, esubindex,
			      patch_number, feat_offset,
			      patch_number, 
			      feat_offset + 
			      NEXT_LAID_LINEAR(new_feat1->info.vertices));
	    }
	    else
	    {
		add_topo_node(ctdb, x, y, eindex, esubindex,
			      ILLEGAL_PATCH, 0,
			      ILLEGAL_PATCH, 0);
	    }
	}
	feat = next;
    }
    return;
}


/* Copy all the data pointed to by pg->data_start to where, and update
 * pg->data_start.  Return a pointer to the end of the new data.
 */
static ADDRESS
copy_pg_data(ctdb, pg, pg_num, where)
    CTDB_PATCH_GROUP *pg;
    int32             pg_num;
    ADDRESS           where;
{
    int32 data_size;

    data_size = pg_bytes(ctdb, pg, pg_num);

    bcopy((void *)pg->data_start.grid_data, (void *)where, data_size);

    pg->data_start.grid_data = (CTDB_GRID_PG_DATA *)where;

    return((ADDRESS)((int8 *)where + data_size));
}


/* Regenerate the patch_data array, get rid of the old one, and put the new 
 * one in ctdb. This routine assumes that ctdb->n_patch_bytes has been 
 * updated to reflect the sizes of the new PG's.
 */
static void
regen_db_patch_features(ctdb)
    CTDB  *ctdb;
{
    int32 i;
    int32 num_pg;
    ADDRESS start, cur;

    num_pg = (ctdb->patches_wide*ctdb->patches_high)/
      CTDB_PATCHES_PER_GROUP;

    start = (ADDRESS)ctdb_alloc(NULL, ctdb->n_patch_bytes, 
				"New PG data array");
    cur = start;

    for(i=0;i<num_pg;i++)
    {
	cur = copy_pg_data(ctdb, ctdb->patch_groups + i, i, cur);
    }

    /* Now there should be nothing left that references the old data block,
     * so we'll get rid of it.
     */
    free(ctdb->patch_data);
    ctdb->patch_data = start;
    return;
}

/* Copy the data from nin to nout and return the address immediately following
 * the newly copied data in nout.  The node index is index.
 */
static CTDB_NODE *
copy_node(ctdb, index, nin, nout)
    CTDB      *ctdb;
    int32      index;
    CTDB_NODE *nin;
    CTDB_NODE *nout;
{
    int32 size;
    
    size = sizeof(CTDB_NODE) + ((nin->num_connections - 1) * 
				sizeof(struct ctdb_connection));
    
    bcopy((void *)nin, (void *)nout, size);

    ctdb->nodes[index] = (CTDB_NODE *)nout;

    return ((CTDB_NODE *)((int8 *)nout + size));
}

/* Allocate a single block of memory, put all of ctdb's nodes in it, and
 * replace ctdb->node_storage with it. 
 */
static void
regen_db_nodes(ctdb)
    CTDB *ctdb;
{
    int8 *node_storage, *cur;
    int32 i;
    
    cur = node_storage = (int8 *)ctdb_alloc(NULL, ctdb->node_size, 
					    "New node data");
    
    for(i=0;i<ctdb->num_nodes;i++)
    {
	cur = (int8 *)copy_node(ctdb, i, ctdb->nodes[i], (CTDB_NODE *)cur);
    }
    
    free(ctdb->node_storage);
    ctdb->node_storage = node_storage;
    return;
}   


/* Copy the data from ein to eout and return the address immediately following
 * the newly copied data in eout
 */
static CTDB_EDGE *
copy_edge(ctdb, index, ein, eout)
    CTDB      *ctdb;
    int32      index;
    CTDB_EDGE *ein;
    CTDB_EDGE *eout;
{
    int32 size;
    
    size = sizeof(CTDB_EDGE) + ((ein->num_features - 1) * 
				sizeof(struct ctdb_feature_reference));
    
    bcopy((void *)ein, (void *)eout, size);
    
    ctdb->edges[index] = (CTDB_EDGE *)eout;

    return ((CTDB_EDGE *)((int8 *)eout + size));
}



/* Allocate a single block of memory, put all of ctdb's edges in it, and
 * replace ctdb->edge_storage with it.  
 */
static void
regen_db_edges(ctdb)
    CTDB  *ctdb;
{
    int8 *edge_storage, *cur;
    int32 i;
    
    cur = edge_storage = (int8 *)ctdb_alloc(NULL, ctdb->edge_size, 
					    "New edge data");
    
    for(i=0;i<ctdb->num_edges;i++)
    {
	cur = (int8 *)copy_edge(ctdb, i, ctdb->edges[i], (CTDB_EDGE *)cur);
    }
    
    free(ctdb->edge_storage);
    ctdb->edge_storage = edge_storage;
    return;
}


/* Look through all the points in the bp_list (except the ones generated by
 * this cell).  If any of them fall within this database, make sure this
 * database has a node in the right place.  If not, break all features that
 * point lies on at the point, and add a node there.  In either case, add
 * an entry with the appropriate node index for the requesting database to
 * use.
 */
static void
add_points_to_linears(index)
    int32 index;
{
    int32 i, j;
    CTDB *ctdb = &cell_dbs[index].ctdb;
    int32 cur_cell = ctdb_get_cell_id(ctdb);
    float64 x,y,z;
    BOUNDARY_PT *pt;
    BOUNDARY_PT tbp;
    int32 node_index;
    int32 *pgs_modified;
    int32 pgm_len = 0;
    int32 *edges_modified;
    int32 em_len = 0;
    int32 ni;
    int32 last_node;
    
    /* There can't be more modified patch groups than there are points, so
     * that gives us a safe upper bound on the length of this array.
     */
    pgs_modified = (int32 *)ctdb_alloc(NULL, 
				       bp_list.num_points * sizeof(int32),
				       "Modified PG list");

    /* There can't be more modified edges than twice the number of points,
     * so that gives us a safe upper bound on the length of this array.
     */
    edges_modified = 
      (int32 *)ctdb_alloc(NULL,
			  bp_list.num_points * sizeof(int32) * 2,
			  "Modified edge list");
    
    /* While we're at it, we'll increase the size of the nodes and edges
     * arrays for ctdb so we won't have to worry about those, either
     */
    ctdb->nodes = 
      (CTDB_NODE **)ctdb_alloc(ctdb->nodes, 
			       (ctdb->num_nodes + bp_list.num_points) *
			       sizeof(CTDB_NODE *),
			       "CTDB Nodes array");
    
    ctdb->edges =
      (CTDB_EDGE **)ctdb_alloc(ctdb->edges,
			       (ctdb->num_edges + 2*bp_list.num_points) *
			       sizeof(CTDB_EDGE *),
			       "CTDB Edges array");

    for(i=0;i<bp_list.num_points;i++)
    {
	if (bp_list.points[i].from_cell == index)
	{
	    /* This point came from this cell, so we should ignore it */
	    continue;
	}
	else
	{
	    pt = &bp_list.points[i];
	    gcs_gcc3e_to_gcscs(pt->x, pt->y, pt->z, cur_cell, &x, &y, &z);
	    
	    if (pt_in_db(ctdb, x, y))
	    {
		/* First, check to see if we already have a node at the point
		 * (we will, for example, if the node is the endpoint of an
		 * abstract feature).  If so, can ignore this point on this
		 * iteration.
		 */
		/* $$$ This isn't quite right--at the moment, anyway, we're
		 * allowing the creation of multiple nodes at a single point,
		 * one per feature found there (assuming that for some reason
		 * this did not result in the presence of a node in the 
		 * original).  This call will only find one of those nodes.
		 */
		if ((ni = find_node_index(ctdb, x, y)) != ILLEGAL_NODE)
		{
		    tbp = *pt;
		    tbp.to_cell = index;
		    tbp.node_index = ni;
		    add_bp(&cell_pts[tbp.from_cell], &tbp);		    
		    continue;
		}

		/* Otherwise, put it where it needs to go */
		last_node = ctdb->num_nodes;

		add_pt_to_physical_linears(ctdb, x, y, z, 
					   pgs_modified, &pgm_len,
					   edges_modified, &em_len);

		add_pt_to_abstract_linears(ctdb, x, y, z);

		/* Store off the node(s) we added for this point so that the
		 * requester can use them
		 */
		tbp = *pt;
		tbp.to_cell = index;
		for(j=last_node;j<ctdb->num_nodes;j++)
		{
		    tbp.node_index = j;
		    add_bp(&cell_pts[tbp.from_cell], &tbp);
		}
	    }
	}
    }

    /* $$$ Regenerate the quad tree ??? */
    /* $$$ fix quad tree references in the topology */

    return;
}


/* Add a connection to node dest_index in cell dest_cell at node src_index in 
 * ctdb.
 */
static void
add_connection(ctdb, src_index, dest_cell, dest_index)
    CTDB *ctdb;
    int32 src_index;
    int32 dest_cell;
    int32 dest_index;
{
    CTDB_NODE *old_node, *new_node;
    int32 old_size, new_size;

    old_node = ctdb->nodes[src_index];
    
    old_size = sizeof(CTDB_NODE) + ((old_node->num_connections - 1) *
				    sizeof(struct ctdb_connection));
    new_size = old_size + sizeof(struct ctdb_connection);
    
    ctdb->node_size += sizeof(struct ctdb_connection);;

    new_node = (CTDB_NODE *)ctdb_alloc(NULL, new_size, 
					"Inter-cell node connection");
    
    bcopy((void *)old_node, (void *)new_node, old_size);

    new_node->connection[new_node->num_connections].node_in_other_cell = TRUE;
    new_node->connection[new_node->num_connections].node_index = dest_index;
    new_node->connection[new_node->num_connections].edge.other_cell = 
      dest_cell;
    
    (new_node->num_connections)++;
    
    ctdb->nodes[src_index] = new_node;

    return;
}


/* Process the cell_pts[index] list, adding references to nodes in neighboring
 * databases.
 */
static void
update_boundary_nodes(index)
    int32 index;
{
    CTDB *ctdb = &cell_dbs[index].ctdb;
    int32 cur_cell = ctdb_get_cell_id(ctdb);
    int32 i;
    int32 ni;
    PLIST *list = &cell_pts[index];    
    BOUNDARY_PT *pt;
    float64 x,y,z;
    
    
    for(i=0;i<list->num_points;i++)
    {
	pt = &list->points[i];
	gcs_gcc3e_to_gcscs(pt->x, pt->y, pt->z, cur_cell, &x, &y, &z);
	
	if ((ni = find_node_index(ctdb, x, y)) == ILLEGAL_NODE)
	{
	    printf("WARNING: cell %d could not find its own boundary node\n",
		   cur_cell);
	}
	else
	{
	    add_connection(ctdb, ni, pt->to_cell, pt->node_index);
	}
    }

    return;
}

/* Write the databases in the cell_dbs list back out to disk. */
static void 
write_cells()
{
    int32 i;
    CTDB *ctdb;
    CTDB_FILE_HEADER hdr;
    FILE *out_file;
    char name[MAX_FNAME_LEN + 4];
    
#define TMP_EXT ".ict"

    for(i=0;i<ndb;i++)
    {
	strcpy(name, cell_dbs[i].fname);
	strcat(name, TMP_EXT);

	if (!(out_file = fopen(name, "w")))
	{
	    perror("Error opening file:");
	    exit(-1);
	}

	ctdb = &cell_dbs[i].ctdb;

	/* Update whenever latest format changes */
	ctdb_write_7((void *)ctdb, out_file);

	fclose(out_file);
    }
    return;
}


main(argc, argv)
    int argc;
    char **argv;
{
    int32 i;
    CTDB *cur_ctdb;
    
    /* $$$ A note on memory usage:  We frequently allocate new blocks
     * of memory to replace chunks of bigger blocks.  Since we make
     * very little effort to keep track of where these chunks occur,
     * we cannot free them.  This is an intentional and (hopefully)
     * acceptable memory leak that could be fixed with additional effort
     */

    if (argc <= 2)
    {
	printf("Syntax: %s fn1 fn2 [fn3...]\n",argv[0]);
	printf("(Where fn* are the names of cell databases)\n");
	exit(1);
    }
    
    ictopo_init(argc-1, DATA_PATH);
    
    read_cells(argv+1, argc-1);
    
    /* Find nodes at boundaries of all databases, and put them in the global
     * bp_list
     */
    for(i=0;i<ndb;i++)
    {
	find_boundary_nodes(i);
    }

    /* Iterate through the databases, breaking linear features (both physical
     * and abstract) and adding nodes at all locations where other databases
     * have boundary nodes.  The locations and indices of these new nodes
     * are stored off in the per-cell new_node_list
     */
    for(i=0;i<ndb;i++)
    {
	add_points_to_linears(i);
    }
    
    /* Iterate through all the databases again, this time updating boundary
     * nodes to reference the corresponding nodes in neighboring databases
     */
    for(i=0;i<ndb;i++)
    {
	update_boundary_nodes(i);
    }

    
    /* A lot of stuff that used to be stored in single large chunks of memory
     * is now fragmented.  We need to put it back together before writing
     * out the database.
     */
    for(i=0;i<ndb;i++)
    {
	cur_ctdb = &cell_dbs[i].ctdb;
	regen_db_patch_features(cur_ctdb);
	regen_db_nodes(cur_ctdb);
	regen_db_edges(cur_ctdb);
	/* $$$ Fix other stuff, too. */
    }


    /* Write out the modified cell databases */
    /* $$$ Make sure to copy any changes from the ctdbs back into the headers 
     */
    write_cells();    
}






