#ifndef lint
static char rcsid [] = "$RCSfile: quadtocorr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: quadtocorr.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Utility which generates compiler input from a libquad DB  *
*       Created: Tue Nov  1 1994                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <libquad.h>
#include <libctdb.h>
#include <stdext.h>

static void dump_libquad_verts();

int main(argc, argv)
    int32 argc;
    char *argv[];
{
    FILE *fout;
    QUAD_DATA quad_data;
    int32 i, j, count;
    CTDB_ABSTRACT_DATA adata[1024]; /* Big to handle big labels */
    int32 *adata_as_int = (int32 *)adata;

    if ((argc < 3) ||
	((argc == 4) && strcmp(argv[3], "linear")) ||
	(argc > 4))
    {
	printf("Usage: %s <terrain-directory> <output.rdr> [linear]\n", argv[0]);
	exit(0);
    }

    quad_read_quadtree_database(&quad_data, argv[1], 0, QUAD_FULL_FILLED);

    if (!(fout = fopen(argv[2], "w")))
    {
	perror("Opening output file");
	exit(-1);
    }

    fprintf(fout, "(\n");

    /* Canopies */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_CANOPY);
    bzero(adata, sizeof(adata[0]));
    for (i=0;i<quad_data.canopies.num_models;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_CANOPY);
	dump_libquad_verts(fout,
			   quad_data.canopies.model_ptr[i].num_pts,
			   quad_data.canopies.model_ptr[i].points);
	adata->canopy.impenetrable =
	  quad_data.canopies.model_ptr[i].impenetrable;
	fprintf(fout, "(0x%x))\n", *adata_as_int);
    }

    /* Soil defrags ("lakes") */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_SOIL_DEFRAG);
    bzero(adata, sizeof(adata[0]));
    for (i=0;i<quad_data.lakes.num_lakes;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_SOIL_DEFRAG);
	dump_libquad_verts(fout,
			   quad_data.lakes.lake_ptr[i].num_pts,
			   quad_data.lakes.lake_ptr[i].points);
	adata->soil_defrag.soil_index =
	  quad_data.lakes.lake_ptr[i].poly_type;
	adata->soil_defrag.level =
	  quad_data.lakes.lake_ptr[i].level;
	fprintf(fout, "(0x%x))\n", *adata_as_int);
    }

    /* Railroads */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_RAILROAD);
    for (i=0;i<quad_data.rails.num_rails;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_RAILROAD);
	dump_libquad_verts(fout,
			   quad_data.rails.rail_ptr[i].num_pts,
			   quad_data.rails.rail_ptr[i].points);
	fprintf(fout, ")\n");
    }

    /* Pipelines */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_PIPELINE);
    for (i=0;i<quad_data.pipelines.num_rails;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_PIPELINE);
	dump_libquad_verts(fout,
			   quad_data.pipelines.rail_ptr[i].num_pts,
			   quad_data.pipelines.rail_ptr[i].points);
	fprintf(fout, ")\n");
    }

    /* Political Boundaries */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_POLITICAL_BOUNDARY);
    for (i=0;i<quad_data.political_boundaries.num_rails;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_POLITICAL_BOUNDARY);
	dump_libquad_verts(fout,
			   quad_data.political_boundaries.rail_ptr[i].num_pts,
			   quad_data.political_boundaries.rail_ptr[i].points);
	fprintf(fout, ")\n");
    }

    /* Labels */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_LABEL);
    for (i=0;i<quad_data.labels.num_labels;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_LABEL);
	fprintf(fout, "(%d.0 %d.0)",
		quad_data.labels.label_ptr[i].x,
		quad_data.labels.label_ptr[i].y);
	count = strlen(quad_data.labels.label_ptr[i].name)+1;
	strncpy(adata->label.text, quad_data.labels.label_ptr[i].name,
		count+3); /* zero-fill up to 3 places further */
	fprintf(fout, "(");
	for (j=0;j<count;j+=4)
	  fprintf(fout, "0x%x ", adata_as_int[j]);
	fprintf(fout, "))\n");
    }

    /* Tactical Signs */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_TACTICAL_SIGN);
    for (i=0;i<quad_data.tact_signs.num_models;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_TACTICAL_SIGN);
	dump_libquad_verts(fout,
			   quad_data.tact_signs.model_ptr[i].num_pts,
			   quad_data.tact_signs.model_ptr[i].points);
	fprintf(fout, ")\n");
    }

    /* Offroad Segments */
    fprintf(fout, "(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_OFFROAD_SEGMENT);
    for (i=0;i<quad_data.offroads.num_segs;i++)
    {
	fprintf(fout, "(ADD_ABSTRACT %d ", CTDB_ABSTRACT_OFFROAD_SEGMENT);
	dump_libquad_verts(fout,
			   quad_data.offroads.segments[i].num_pts,
			   quad_data.offroads.segments[i].points);
	fprintf(fout, "(0x0))\n");
    }

    /* Roads & Rivers */
    if (argc == 4)
    {
	fprintf(fout, "(REPLACE_LINEAR\n");
	for (i=0;i<quad_data.roads.num_segs;i++)
	{
	    fprintf(fout, "(%d.0 1 ",
		    quad_data.roads.segments[i].width);
	    dump_libquad_verts(fout,
			       quad_data.roads.segments[i].num_pts,
			       quad_data.roads.segments[i].points);
	    fprintf(fout, ")\n");
	}
	for (i=0;i<quad_data.river.num_segs;i++)
	{
	    fprintf(fout, "(%d.0 %d ",
		    quad_data.river.segments[i].width,
		    quad_data.river.segments[i].fordable ? 5 : 4);
	    dump_libquad_verts(fout,
			       quad_data.river.segments[i].num_pts,
			       quad_data.river.segments[i].points);
	    fprintf(fout, ")\n");
	}
	fprintf(fout, ")\n");
    }

    fprintf(fout, ")\n");
    fclose(fout);

    return 0;
}

static void dump_libquad_verts(fout, num_pts, points)
    FILE  *fout;
    int32  num_pts;
    int32 *points;
{
    int32 i;

    /* Don't repeat the first point at the end (closed polys) */
    if ((points[0] == points[num_pts-2]) &&
	(points[1] == points[num_pts-1]))
      num_pts -= 2;

    fprintf(fout, "(");
    for (i=0;i<num_pts;i++)
      fprintf(fout, "%d.0 ", points[i]);
    fprintf(fout, ")");
}
