#ifndef lint
static char rcsid [] = "$RCSfile: s1kctdb.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
 *   File: s1kctdb.c                                                         *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Wrapper around s1kctdb and recompile                      *
 *       Created: March 1996                                                 *
 *       Author: flchamberlain                                               *
 *       Remarks: This is a wrapper around s1kctdb and (sometimes) recompile *
 *                that allows the user to issue a single command to generate *
 *                any flavor CTDB database from s1000.                       *
 *                                                                           *
 *       NOTE:  You may have to "limit stacksize unlimited" before running   *
 *              this on a Sun.                                               *
 ****************************************************************************/

#include <libcmdline.h>
#include <stdext.h>
#include <stdcoord.h>
#include <libctdb.h>
#include <libgcs.h>

#define TMP_PREFIX "CMP_TMP_"

/* Command-line options */
struct
{
    CMD_STRING_OPTION  project_location;
    CMD_STRING_OPTION  project;
    CMD_STRING_OPTION  assembly;
    CMD_STRING_OPTION  output_base_name;
    CMD_INTEGER_OPTION version;
    CMD_STRING_OPTION  zone_letter;
    CMD_INTEGER_OPTION datum;
    CMD_INTEGER_OPTION format;
    CMD_STRING_OPTION  exclude_regexp;
    CMD_STRING_OPTION  road;
    CMD_INTEGER_OPTION political;
    CMD_INTEGER_OPTION pipeline;
    CMD_INTEGER_OPTION powerline;
    CMD_TOGGLE_OPTION  microterrain;
    CMD_BOOLEAN_OPTION sparse;
    CMD_BOOLEAN_OPTION large_patches;
    CMD_TOGGLE_OPTION  gridded; /* gridded database requested */
    CMD_TOGGLE_OPTION  tinned;  /* tinned database requested */
    CMD_STRING_OPTION  data_path;
    CMD_STRING_OPTION  gcs_mode;
    CMD_INTEGER_OPTION cell_id;
    CMD_BOOLEAN_OPTION gen_mc_header;
    CMD_BOOLEAN_OPTION list_cells;
    CMD_STRING_OPTION  tdb_path;  
    CMD_BOOLEAN_OPTION gen_ml_water;
    CMD_BOOLEAN_OPTION bathymetry;
    CMD_BOOLEAN_OPTION list_holes;
    CMD_BOOLEAN_OPTION short_soil;
    CMD_BOOLEAN_OPTION long_soil;
    CMD_BOOLEAN_OPTION pend_hack;
} options = {
    {
	"Project Location" , "Specifies directory where projects are stored" ,
	NULL ,
	CMD_STRING ,	"s1kproj" ,	NULL ,	"/proj/ModSAF/s1000/proj",
/* $$$ We'd like this to be "../../../s1000/proj", but the latest version of
 * s1k doesn't like path names that aren't absolute :-( 
 */
    } ,
    {
	"Project" , "Specifies project name (REQUIRED)" ,
	NULL ,
	CMD_STRING ,	"project" ,	NULL ,	"" ,
    } ,
    {
	"Assembly" , "Specifies assembly name (REQUIRED)" ,
	NULL ,
	CMD_STRING ,	"assembly" ,	NULL ,	"" ,
    } ,
    {
	"Output File" , "Specifies name of output CTDB file" ,
	NULL ,
	CMD_STRING ,	"output_base_name" ,	NULL ,	"" ,
    } ,
    {
	"Version" , "Specifies database version (REQUIRED)" ,
	NULL ,
	CMD_INTEGER ,	"version" ,	NULL ,	0 ,
    } ,
    {
	"Zone Letter" , "Specifies letter part of UTM zone" ,
	NULL ,
	CMD_STRING ,	"zone" ,	NULL ,	"" ,
    } ,
    {
	"Datum" , "Specifies datum number" ,
	NULL ,
	CMD_INTEGER ,	"datum" ,	NULL ,	DATUM_WGS84 ,
    } ,
    {
	"Format" , "Specifies CTDB output format" ,
	NULL ,
	CMD_INTEGER ,	"format" ,	NULL ,	CTDB_FILE_FORMAT ,
    } ,
    {
	"Exclude Regular Expression" , 
	"Regular expression which identifies models to be excluded" ,
	NULL ,
	CMD_STRING ,	"exclude" ,	NULL ,	"pole_" ,
    } ,
    {
	"Road/River/Rail Networks" , 
	"Numbers of the network files for roads, rivers, railroads" ,
	NULL ,
	CMD_STRING ,	"networks" ,	NULL ,	"0" ,
    } ,
    {
	"Political Boundary Network" , 
	"Number of the network file for political boundaries" ,
	NULL ,
	CMD_INTEGER ,	"political" ,	NULL ,	0 ,
    } ,
    {
	"Pipeline Network" , "Number of the network file for pipelines" ,
	NULL ,
	CMD_INTEGER ,	"pipeline" ,	NULL ,	0 ,
    } ,
    {
	"Powerline Network" , "Number of the network file for powerlines" ,
	NULL ,
	CMD_INTEGER ,	"powerline" ,	NULL ,	0 ,
    } ,
    {
	"Microterrain" , "Specifies whether to include microterrain" ,
	NULL ,
	CMD_TOGGLE ,	"micro" ,	"nomicro" ,	TRUE ,
    } ,
    {
	"Sparse" , "Hint that the database is sparsely populated" ,
	NULL ,
	CMD_BOOLEAN ,	"sparse" ,	"dense" ,	FALSE ,
    } ,
    {
	"Large Patches" , "Use the largest possible patch size" ,
	NULL ,
	CMD_BOOLEAN ,	"large_patch" ,	"default_patch" ,	FALSE ,
    } ,
    {
	"Gridded" , "Specifies that the database should have gridded patches " ,
	NULL ,
	CMD_TOGGLE ,	"gridded" ,	"nogridded" ,	FALSE ,
    },
    {
	"Tinned" , 
	"Specifies that the database should have tinned patches (note: if both gridded and tinned specified, the database will be hybrid (as by default))" ,
	NULL ,
	CMD_TOGGLE ,	"tinned" ,	"notinned" ,	FALSE ,
    },
    {
	"data path", " Select the data path" ,
	NULL ,
	CMD_STRING ,    "datapath" ,     NULL , "../../../data" ,
    },
    {
	"GCS mode" , "Specifies coord system to compile in / tiling mode.\n\tLegal options are simnet, single, and multi",
	NULL ,
	CMD_STRING ,	"gcs_mode" ,	NULL ,	"simnet"
    },
    {
	"cell number" , "Specifies cell number to load" ,
	NULL ,
	CMD_INTEGER ,	"cell_id" ,	NULL ,	GCS_ILLEGAL_CELL ,
    },
    {
      "Generate Multi-cell header" , "Specifies that a multi-cell header file should be generated",
	NULL ,
	CMD_BOOLEAN ,	"gen_mc_header" ,	"no_mc_hdr" ,	FALSE,
    },
    {
	"List the Cells" , "Specifies  the Cells given these parameters" ,
	NULL ,
	CMD_BOOLEAN ,	"list_cells" ,	"no_list" ,	FALSE,
    },   
    {
	"tdb path", " Select the terrain data path" ,
	NULL ,
	CMD_STRING ,    "tdbpath" ,     NULL , "../../../terrain" ,
    },
    {
	"Generate ML water" , "Transform single-level water to multi-level" ,
	NULL ,
	CMD_BOOLEAN ,	"gen_ml_water" ,	"no_ml_water" ,	FALSE,
    },
    {
	"Bathymetry flag" , "Specifies that source data has bathymetry which should be used" ,
	NULL ,
	CMD_BOOLEAN ,	"bathymetry" ,	"no_bathymetry" ,	FALSE,
    },
    {
      "List terrain holes" , "Print location of each hole found in terrain skin",
	NULL ,
	CMD_BOOLEAN ,	"list_holes" ,	"no_hole_list" ,	FALSE,
    },
    {
      "Use short soil format" , "Assign attribute data to patches",
      NULL,
      CMD_BOOLEAN , "short_soil", "no_short_soil", FALSE,
    },
    {
      "Use long soil format", "Assign attribute data to grid posts",
      NULL,
      CMD_BOOLEAN, "long_soil", "no_long_soil", FALSE,
    },
    {
      "Pendleton hack" , "Use hack to fix ocean abstract when compiling Camp Pendleton DB",
	NULL ,
	CMD_BOOLEAN ,	"pend_hack" ,	"no_pend_hack" ,	FALSE,
    }
};

main(int argc, char **argv)
{
#define COMMAND_SIZE 5000
    char s1k_command[COMMAND_SIZE];
    char recomp_command[COMMAND_SIZE];
    char temp[COMMAND_SIZE], *temp2;
    char s1k_name_base[200];
    char recomp_name_base[200];
    int32 run_recompile = FALSE;
    int32 i;
    int32 leftover_argc;
    char **leftover_argv;

    /* Parse the argument list */
    cmd_process_options(argc, argv, /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized */
			(CMD_OPTION *)&options, /* options array */
			sizeof(options), /* size of options array */
			TRUE	/* Verbose mode */
			);
    cmd_gripe(leftover_argc, leftover_argv);

    /* Make sure all the required fellers were specified */
    if (!*options.project.value)
      printf("Project name must be specified\n"), exit(-1);
    if (!*options.assembly.value)
      printf("Assembly name must be specified\n"), exit(-1);
    if (!options.version.value)
      printf("Database version number must be specified\n"), exit(-1);

    /* To properly emulate the behavior of s1kctdb by itself, we need to
     * take into account the value of the environment variable for this
     * program.  The easiest way to do this is to just build up a command
     * line from our options values.
     */
    bzero(s1k_command, COMMAND_SIZE);
    sprintf(s1k_command,"s1kprep ");

    bzero(recomp_command, COMMAND_SIZE);
    sprintf(recomp_command, "recompile ");
    
    sprintf(temp, "-s1kproj %s ", options.project_location.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-project %s ", options.project.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-assembly %s ", options.assembly.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-version %d ", options.version.value);
    strcat(s1k_command, temp);

    if (strlen(options.zone_letter.value))
    {
	sprintf(temp, "-zone %s ", options.zone_letter.value);
	strcat(s1k_command, temp);
    }
    
    sprintf(temp, "-datum %d ", options.datum.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-format %d ", options.format.value);
    strcat(s1k_command, temp);
    strcat(recomp_command, temp);

    sprintf(temp, "-exclude \"%s\" ", options.exclude_regexp.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-networks \"%s\" ", options.road.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-political %d ", options.political.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-pipeline %d ", options.pipeline.value);
    strcat(s1k_command, temp);

    sprintf(temp, "-powerline %d ", options.powerline.value);
    strcat(s1k_command, temp);

    if (!options.microterrain.value)
    {
	sprintf(temp, "-nomicro ");
	strcat(s1k_command, temp);
    }
    
    if (options.sparse.value)
    {
	sprintf(temp, "-sparse ");
	strcat(s1k_command, temp);
    }
    
    if (options.large_patches.value)
    {
	sprintf(temp, "-large_patch ");
	strcat(s1k_command, temp);
    }
    
    if (options.gridded.value)
    {
	sprintf(temp, "-gridded ");
	strcat(s1k_command, temp);
	strcat(recomp_command, temp);
    }
    
    if (options.tinned.value)
    {
	sprintf(temp, "-tinned ");
	strcat(s1k_command, temp);
	strcat(recomp_command, temp);
    }
    
    sprintf(temp, "-datapath %s ", options.data_path.value);
    strcat(s1k_command, temp);
    strcat(recomp_command, temp);
    
    /* Don't use s1kctdb to directly generate a fully tinned GCS database */
    if (((!strcmp(options.gcs_mode.value, "single") ) || 
	 (!strcmp(options.gcs_mode.value, "multi"))) &&
	(!options.gridded.value) &&
	(!options.list_cells.value) &&
	(!options.gen_mc_header.value))
    {
	run_recompile = TRUE;
	sprintf(temp, "-gcs_mode simnet ");
	strcat(s1k_command, temp);
	sprintf(temp, "-gcs_mode %s ", options.gcs_mode.value);
	strcat(recomp_command, temp);
    }
    else
    {
	sprintf(temp, "-gcs_mode %s ", options.gcs_mode.value);
	strcat(s1k_command, temp);
    }
    
    sprintf(temp, "-cell_id %d ", options.cell_id.value);
    strcat(s1k_command, temp);
    strcat(recomp_command, temp);

    if (options.gen_mc_header.value)
    {
	sprintf(temp, "-gen_mc_header ");
	strcat(s1k_command, temp);
    }
    
    if (options.list_cells.value)
    {
	sprintf(temp, "-list_cells ");
	strcat(s1k_command, temp);
    }

    if (options.short_soil.value)
    {
        sprintf(temp, "-short_soil ");
        strcat(recomp_command, temp);
    }
    else if (options.long_soil.value)
    {
        sprintf(temp, "-long_soil ");
        strcat(recomp_command, temp);
    }

    if (options.pend_hack.value)
    {
	sprintf(temp, "-pend_hack ");
	strcat(s1k_command, temp);
    }
    
    sprintf(temp, "-tdbpath %s ", options.tdb_path.value);
    strcat(s1k_command, temp);
    strcat(recomp_command, temp);

    if (options.gen_ml_water.value)
    {
	sprintf(temp, "-gen_ml_water ");
        strcat(s1k_command, temp);
    }

    if (options.bathymetry.value)
    {
	sprintf(temp, "-bathymetry ");
        strcat(s1k_command, temp);
    }

    if (options.list_holes.value)
    {
	sprintf(temp, "-list_holes ");
	strcat(s1k_command, temp);
    }

    /* Figure out what names to use for databases */

    if (!strlen(options.output_base_name.value))
      sprintf(recomp_name_base,"%s-%04d",options.assembly.value, 
	      options.version.value);
    else
    {
	strcpy(recomp_name_base, options.output_base_name.value);
    }

    if (!run_recompile)
    {
	sprintf(temp, "-output_base_name %s ", recomp_name_base);
	strcat(s1k_command, temp);
    }
    else
    {
	sprintf(s1k_name_base, "%s%s", TMP_PREFIX, recomp_name_base);
	sprintf(temp, "-output_base_name %s ", s1k_name_base);	
	strcat(s1k_command, temp);
	sprintf(temp, "-output_base_name %s ", recomp_name_base);
	strcat(recomp_command, temp);
	
	/* Figure out what s1kprep will name its output file */
	temp2 = ctdb_generate_tdb_filename(CTDB_MODE_SIMNET,
					   options.tdb_path.value,
					   s1k_name_base,
					   options.format.value,
					   GCS_ILLEGAL_CELL, FALSE);

	sprintf(temp, "-input_ctdb %s ", temp2);
	strcat(recomp_command, temp);
    }
    
    sprintf(temp, "\n");
    strcat(s1k_command, temp);

    /* First, call s1kctdb */
    system(s1k_command);

    if (run_recompile)
    {
	/* The user has asked us to do something we'd prefer not to do
	 * in one step.  Complete the requested operation with recompile
	 */
	system(recomp_command);

	/* Get rid of temp file */
	sprintf(recomp_command, "rm -f %s\n", temp2);
	system(recomp_command);
    }
}







