#ifndef lint
static char rcsid [] = "$RCSfile: tdbtoctdb.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tdbtoctdb.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to generate a CTDB file from TDB+libquad data
*       Created: Thu Oct 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks: This code is based upon legacy SIMNET libraries and data
*                files which are not likely to be found at most installations.
*                Therefore, don't expect to be able to build this program!
****************************************************************************/

#include <libreader.h>
#include <tdb.h>
#include <pvis_call.h>
#include <libquad.h>
#include <libctdb.h>
#include "../libctdb_local.h"
#include "ct_cmplr.h"

#define INCR           125
#define POSTS_PER_PATCH_SIDE 4

/* Used when data is missing from libtdb or libquad */
#define DEFAULT_BUILDING_HEIGHT 15
#define DEFAULT_TREE_HEIGHT     15
#define DEFAULT_TREE_CLEARANCE   1
#define DEFAULT_TREE_RADIUS      1.5

CTDB_FEATURE_DATA data[1024*1024];

/*
#define TDB_INIT_READ_TYPE tdb_init_memory
#define TERRAIN_READ_TYPE_INQUIRE terrain_memory_inquire
*/
#define TDB_INIT_READ_TYPE tdb_init_cache
#define TERRAIN_READ_TYPE_INQUIRE terrain_cache_inquire

static void write_features();
static void get_post_soils();

main (argc, argv)
    int32 argc;
    char *argv[];
{
    float64 x,y,dx,dy,pos[3];
    float32 e;
    int32 i;
    int32 p,sNE,sSW;
    CTDB_FILE_HEADER hdr;
    CTDB ctdb;
    QUAD_DATA quad_data;
    float64 inv_basis;
    FILE *f;
    char poly_file[100];
    int32 tloc;
    DATA_UNION milgrid_data;
    int32 ret;
    extern float64 simnet_origin_northing;
    extern float64 simnet_origin_easting;
    extern int32    simnet_origin_zone;
    extern char   simnet_origin_zone_letter;
    extern int32    simnet_datum;

    if (argc != 3)
    {
	printf("Usage: %s <dbdir/> <ctdb-file>\n", argv[0]);
	exit(0);
    }

    strcpy(poly_file, argv[1]);
    strcat(poly_file, "/polygons");
    if (TDB_INIT_READ_TYPE(poly_file,100) == -1)
    {
	printf("Unable to init terrain: %s\n",
	       tdb_error());
	exit(1);
    }

    quad_read_quadtree_database(&quad_data, argv[1], FALSE, LEAF_FILLED);
    printf("\n");

    if (reader_read_file("/saf/config.factory/utm.lisp", &milgrid_data) != 0)
    {
	printf("Unable to read utm.lisp\n");
	exit(0);
    }
    coord_set_utm_table(milgrid_data.array);
    if (ret = set_simnet_coord_params(quad_datap->quadtree.latitude,
				      quad_datap->quadtree.longitude, 
				      quad_datap->quadtree.UTM_SW))
      printf("Error setting simnet params: %s\n",
	     convert_error(ret));

    f = fopen(argv[2],"w");

    hdr.format = CTDB_FILE_FORMAT;
    hdr.fixed_point_basis = CTDB_STANDARD_FIXED_POINT_BASIS;
    hdr.post_increment = INCR;
    hdr.posts_per_patch_side = POSTS_PER_PATCH_SIDE;
    inv_basis = 1.0/hdr.fixed_point_basis;
    hdr.pages_wide = 1 +
      tdb_info.db_header.db_info.max_x/(INCR*POSTS_PER_SIDE);
    hdr.pages_high = 1 +
      tdb_info.db_header.db_info.max_y/(INCR*POSTS_PER_SIDE);
    hdr.min_x = 0;
    hdr.max_x = tdb_info.db_header.db_info.max_x;
    hdr.min_y = 0;
    hdr.max_y = tdb_info.db_header.db_info.max_y;
    hdr.num_features = 0;
    for(i=0;i<CTDB_HEADER_RESERVED;i++)
      hdr.soil_map[i] = i;
    hdr.origin_northing = simnet_origin_northing;
    hdr.origin_easting = simnet_origin_easting;
    hdr.origin_zone_number = simnet_origin_zone;
    hdr.origin_zone_letter = simnet_origin_zone_letter;
    hdr.datum = simnet_datum;
    strcpy(hdr.name, tdb_info.db_header.db_info.db_name);
    hdr.version = tdb_info.db_header.db_info.version;
    time(&tloc);
    strcpy(hdr.date, ctime(&tloc));

    ctdb_cmplr_unpack_header(&ctdb, &hdr);

    hdr.num_features = generate_features(&ctdb);

    ctdb_choose_feature_cache(&ctdb, hdr.num_features, (int *)NULL);
    bcopy(ctdb.cache_size_map,hdr.cache_size_map,sizeof(hdr.cache_size_map));

    fwrite(&hdr, sizeof(hdr), 1, f);

    write_features(f, &ctdb, hdr.num_features);

    e = 0.0;

#define REASONABLE_ELEVATION 5000.0

    for (y=0.0;y<=hdr.max_y;y+=INCR*POSTS_PER_SIDE)
      for (x=0.0;x<=hdr.max_x;x+=INCR*POSTS_PER_SIDE)
	for (dy=0.0;dy<INCR*POSTS_PER_SIDE;dy+=INCR)
	  for (dx=0.0;dx<INCR*POSTS_PER_SIDE;dx+=INCR)
	  {
	      pos[0] = x+dx;
	      pos[1] = y+dy;

	      if ((pos[0] > hdr.max_x) ||
		  (pos[1] > hdr.max_y))
		p = 0;
	      else
	      {
		  if ((tdb_get_z(pos, TDB_GROUND_ONLY) == -1) ||
		      (pos[2] > REASONABLE_ELEVATION))
		    ;		/* Use last value of e */
		  else
		    e = pos[2];

		  get_post_soils(pos, &sNE, &sSW);

		  p = BUILD_POST_ELEVATION(e, inv_basis);

		  /* The bottom row of posts describes no polygons */
		  if (pos[1] != 0)
		  {
		      p |= BUILD_POST_NE_SOIL(sNE);
		      p |= BUILD_POST_SW_SOIL(sSW);

		      p |= ctdb_features_of_post(&ctdb,
						 (x+dx)/INCR, (y+dy)/INCR);
		  }
	      }

	      fwrite(&p, sizeof(int32), 1, f);
	  }

    fclose(f);
}

static void get_post_soils(pos, sNE, sSW)
    float64  pos[3];
    int32   *sNE, *sSW;
{
    int32 sanity;
    int32 NE, SW;
    float64 test[3];

    /* Find the soil of the post with a test.  If we hit road or water,
     * test again to make sure it really covers the whole post.
     */

    bcopy(pos,test,sizeof(test));
    sanity = 0;
    do
    {
	test[0] += 10.0;
	test[1] = pos[1] - 5.0 - (sanity % 2) * sanity * 5.0;
	NE = tdb_get_z(test, TDB_GROUND_ONLY);

	if (sanity++ > 10)
	  break;
    } while ((NE == SOIL_ROAD) || (NE == SOIL_WATER) || (NE == SOIL_MUCK));

    bcopy(pos,test,sizeof(test));
    sanity = 0;
    do
    {
	test[0] = pos[0] + 5.0 + (sanity % 2) * sanity * 5.0;
	test[1] -= 10.0;
	SW = tdb_get_z(test, TDB_GROUND_ONLY);

	if (sanity++ > 10)
	  break;
    } while ((SW == SOIL_ROAD) || (SW == SOIL_WATER) || (SW == SOIL_MUCK));

    *sNE = NE;
    *sSW = SW;
}

static void write_features(f, ctdb, num_features)
    FILE   *f;
    CTDB   *ctdb;
    int32   num_features;
{
    fwrite(ctdb->patch_groups, sizeof(CTDB_PATCH_GROUP),
	   ctdb->patches_wide*ctdb->patches_high/CTDB_PATCHES_PER_GROUP, f);
    fwrite(ctdb->features, sizeof(CTDB_FEATURE_DATA), num_features, f);
}

static void complete_micro_poly(poly)
    CTDB_MC_POLYGON *poly;
{
    poly->soil = tdb_get_z(poly->verts[0], TDB_GROUND_ONLY);
    tdb_get_z(poly->verts[1], TDB_GROUND_ONLY);
    tdb_get_z(poly->verts[2], TDB_GROUND_ONLY);
}

static int32 generate_features(ctdb)
    CTDB *ctdb;
{
    int32 npatches;
    int32 start_poly_index;
    int32 end_poly_index;
    int32 grid_number;
    GRID_MAP * grid_map;
    TERRAIN_POINT * vertices;
    POLYGON_DESCRIPTOR * polys;
    PATCH_HEADER * patch_header;
    POLYGON_DESCRIPTOR * current_poly;
    char *patch;
    int16 * poly_vertices;
    OBJECT_DESCRIPTOR *objects;
    int32 no_objects;
    int32 i,j,v;
    uint32 start, end;
    int16 vert_list[1000][3];
    int32 vert_count, dup;
    int32 patches = ctdb->patches_wide * ctdb->patches_high;
    CTDB_MC_POLYGON be_microterrain[256];
    CTDB_BUILDING_POLYGON be_buildings[256];
    CTDB_TREELINE be_trees[2048];
    CTDB_MC_POLYGON be_canopies[256];
    CTDB_CAN_EDGE be_cedges[1024];
    CTDB_LINEAR be_linear[256];
    int32 n_micro, n_build, n_trees, n_cans, n_cedge, n_linear;
    int32 f, k, ke, added;
    int32 patch_x, patch_y;
    int32 tdb_patch_index, ctdb_patch_index;
    LIST *feature_index_list;
    LIST *cons;
    QUAD_FEATURES *feat;
    int32 *ptr, num;
    MODEL *build, *tree;
    LINE_SEG *seg;
    int32 no_canopies, no_polygons, no_boundaries, no_b_verts;
    CANOPY_HEADER *canopies;
    TREELINE_HEADER *bounds;
    TERRAIN_POINT *b_verts;

    ctdb->patch_groups =
      (CTDB_PATCH_GROUP *)malloc(sizeof(CTDB_PATCH_GROUP)*
				 (patches/CTDB_PATCHES_PER_GROUP));
    bzero(ctdb->patch_groups,
	  sizeof(CTDB_PATCH_GROUP)* (patches/CTDB_PATCHES_PER_GROUP));
    ctdb->features = data;

    npatches = tdb_info.max_number_patches;

    f = 0;
    for (patch_y=0;patch_y<tdb_info.y_dim;
	 patch_y++,putchar('.'),fflush(stdout))
      for (patch_x=0;patch_x<tdb_info.x_dim;patch_x++)
      {
	  tdb_patch_index = patch_y * tdb_info.x_dim + patch_x;
	  ctdb_patch_index = patch_y * ctdb->patches_wide + patch_x;

	  patch = (char *) TERRAIN_READ_TYPE_INQUIRE (tdb_patch_index);
	  patch_header = (PATCH_HEADER *)patch;

	  /* COLLECT MICROTERRAIN */
	  k = 0;

	  grid_map = patch_header->poly_grid;
	  vertices = (TERRAIN_POINT *)(patch + patch_header->vertices_offset);
	  polys=(POLYGON_DESCRIPTOR *)(patch + patch_header->polygons_offset);

	  start = -1;
	  end = 0;
	  for (grid_number=0;grid_number<16;grid_number++)
	  {

	      start_poly_index = grid_map[grid_number].start_micro_polys;
	      end_poly_index =
		start_poly_index + grid_map[grid_number].num_micro_polys;

	      if (start_poly_index < start)
		start = start_poly_index;
	      if (end_poly_index > end)
		end = end_poly_index;
	  }

	  vert_count = 0;
	  for (i = start; i < end; i++)
	  {
	      current_poly = polys + i;
	      poly_vertices = (int16 *)current_poly->vertices;

	      dup = 0;
	      for (j=0;j<vert_count;j++)
		if ((vert_list[j][0] == poly_vertices[0]) &&
		    (vert_list[j][1] == poly_vertices[1]) &&
		    (vert_list[j][2] == poly_vertices[2]))
		{
		    dup = 1;
		    break;
		}

	      if (dup)
		continue;

	      vert_list[vert_count][0] = poly_vertices[0];
	      vert_list[vert_count][1] = poly_vertices[1];
	      vert_list[vert_count][2] = poly_vertices[2];
	      vert_count++;

	      be_microterrain[k].soil =
		current_poly->poly_info.poly_bits.soil_type;
	      be_microterrain[k].n_verts =
		current_poly->poly_info.poly_bits.number_vertices;
	      for(v=0;v<be_microterrain[k].n_verts;v++)
	      {
		  be_microterrain[k].verts[v][0] =
		    vertices[poly_vertices[v]].x;
		  be_microterrain[k].verts[v][1] =
		    vertices[poly_vertices[v]].y;
		  be_microterrain[k].verts[v][2] =
		    vertices[poly_vertices[v]].z;
	      }

	      k++;
	  }
	  n_micro = k;

	  /* COLLECT BUILDINGS */
	  k = 0;

	  feature_index_list =
	    quad_get_quad_nodes (quad_nodes, 0,
			    quadtree.x, quadtree.y, quadtree.size,
			    patch_x*ctdb->incr*ctdb->patch_incr+1,
			    patch_y*ctdb->incr*ctdb->patch_incr+1,
			    (patch_x+1)*ctdb->incr*ctdb->patch_incr-1,
			    (patch_y+1)*ctdb->incr*ctdb->patch_incr-1);

	  for (cons = feature_index_list; cons != NULL_PTR; cons = cons->cdr)
	  {
	      feat = (QUAD_FEATURES *) cons->car;
	      ptr = feat->building_indecies + 1;
	      num = *(feat->building_indecies);
	      for (i=1; i<= num; i++, ptr++)
	      {
		  build = &(buildings.model_ptr[*ptr]);

		  be_buildings[k].n_verts = build->num_pts/2;

		  if (be_buildings[k].n_verts >= MAX_BUILDING_POLYGON_VERTS)
		  {
		      printf("Building with %d verts\n",
			     be_buildings[k].n_verts);
		      abort();
		  }

		  for (j=0;j<build->num_pts/2;j++)
		  {
		      be_buildings[k].verts[j][0] = build->points[j*2];
		      be_buildings[k].verts[j][1] = build->points[j*2+1];
		      be_buildings[k].verts[j][2] = 0.0;
		      tdb_get_z(be_buildings[k].verts[j], TDB_GROUND_ONLY);
		      be_buildings[k].verts[j][2] +=
			(build->height ?
			 build->height : DEFAULT_BUILDING_HEIGHT);
		  }

		  k++;
	      }
	  }
	  n_build = k;

	  /* COLLECT TREES */
	  k = 0;

	  for (cons = feature_index_list; cons != NULL_PTR; cons = cons->cdr)
	  {
	      feat = (QUAD_FEATURES *) cons->car;
	      ptr = feat->tree_indecies + 1;
	      num = *(feat->tree_indecies);
	      for (i=1; i<= num; i++, ptr++)
	      {
		  tree = &(trees.model_ptr[*ptr]);

		  be_trees[k].n_verts = tree->num_pts/2;
		  be_trees[k].height = (tree->height ?
				        tree->height : DEFAULT_TREE_HEIGHT);
		  be_trees[k].radius = DEFAULT_TREE_RADIUS;

		  if (be_trees[k].n_verts >= MAX_TREELINE_VERTS)
		  {
		      printf("Treeline with %d verts\n",
			     be_trees[k].n_verts);
		      abort();
		  }

		  for (j=0;j<tree->num_pts/2;j++)
		  {
		      be_trees[k].verts[j][0] = tree->points[j*2];
		      be_trees[k].verts[j][1] = tree->points[j*2+1];
		      be_trees[k].verts[j][2] = 0.0;
		      tdb_get_z(be_trees[k].verts[j], TDB_GROUND_ONLY);
		      be_trees[k].verts[j][2] +=
			be_trees[k].height + DEFAULT_TREE_CLEARANCE;
		  }

		  k++;
	      }
	  }
	  n_trees = k;

	  /* COLLECT CANOPIES */
	  k = 0;
	  ke = 0;

	  no_canopies = patch_header->number_canopies;
	  canopies = (CANOPY_HEADER *)(patch + patch_header->canopies_offset);
	  vertices =
	    (TERRAIN_POINT *) (patch + patch_header->canopy_vertices_offset);

	  for (i=0;i<no_canopies;i++,
	       canopies = (CANOPY_HEADER *)
	       ((char *) canopies + canopies->canopy_size))
	  {
	      no_polygons = canopies->number_polygons;
	      polys = (POLYGON_DESCRIPTOR *)
		((char *) canopies + canopies->polygons_offset);

	      for (j=0;j<no_polygons;j++)
	      {
		  poly_vertices = (int16 *)polys[j].vertices;

		  be_canopies[k].n_verts =
		    GET_CANOPY_POLY_NUM_VERTS (polys[j].poly_info);

		  for(v=0;v<be_canopies[k].n_verts;v++)
		  {
		      be_canopies[k].verts[v][0]=vertices[poly_vertices[v]].x;
		      be_canopies[k].verts[v][1]=vertices[poly_vertices[v]].y;
		      be_canopies[k].verts[v][2]=vertices[poly_vertices[v]].z;
		  }

		  k++;
	      }

	      no_boundaries = canopies->number_boundaries;
	      bounds = (TREELINE_HEADER *)
		((char *) canopies + canopies->boundaries_offset);
	      for (j=0;j<no_boundaries;j++,
		   bounds = (TREELINE_HEADER *)(b_verts+no_b_verts))
	      {
		  no_b_verts = bounds[j].number_vertices;
		  b_verts = (TERRAIN_POINT *)(bounds+1);
		  for (v=0;v<no_b_verts/**/-1/**/;v++)
		  {

		      be_cedges[ke].start[0] = b_verts[v].x;
		      be_cedges[ke].start[1] = b_verts[v].y;
		      be_cedges[ke].end[0] = b_verts[v+1].x;
		      be_cedges[ke].end[1] = b_verts[v+1].y;
		      ke++;
		  }
	      }
	  }
	  n_cans = k;
	  n_cedge = ke;

	  /* COLLECT ROADS and RIVERS */
	  k = 0;

	  for (cons = feature_index_list; cons != NULL_PTR; cons = cons->cdr)
	  {
	      feat = (QUAD_FEATURES *) cons->car;
	      ptr = feat->road_indecies + 1;
	      num = *(feat->road_indecies);
	      for (i=1; i<= num; i++, ptr++)
	      {
		  seg = &(roads.segments[*ptr]);

		  be_linear[k].n_verts = seg->num_pts/2;
		  be_linear[k].width = seg->width;
		  be_linear[k].soil = SOIL_ROAD;

		  if (be_linear[k].n_verts >= MAX_LINEAR_VERTS)
		  {
		      printf("Road with %d verts\n",
			     be_linear[k].n_verts);
		      abort();
		  }

		  for (j=0;j<seg->num_pts/2;j++)
		  {
		      be_linear[k].verts[j][0] = seg->points[j*2];
		      be_linear[k].verts[j][1] = seg->points[j*2+1];
		  }

		  k++;
	      }
	  }
	  for (cons = feature_index_list; cons != NULL_PTR; cons = cons->cdr)
	  {
	      feat = (QUAD_FEATURES *) cons->car;
	      ptr = feat->water_indecies + 1;
	      num = *(feat->water_indecies);
	      for (i=1; i<= num; i++, ptr++)
	      {
		  seg = &(river.segments[*ptr]);

		  be_linear[k].n_verts = seg->num_pts/2;
		  be_linear[k].width = seg->width;
		  if (seg->fordable)
		    be_linear[k].soil = SOIL_SHALLOW_WATER;
		  else
		    be_linear[k].soil = SOIL_DEEP_WATER;

		  if (be_linear[k].n_verts >= MAX_LINEAR_VERTS)
		  {
		      printf("River with %d verts\n",
			     be_linear[k].n_verts);
		      abort();
		  }

		  for (j=0;j<seg->num_pts/2;j++)
		  {
		      be_linear[k].verts[j][0] = seg->points[j*2];
		      be_linear[k].verts[j][1] = seg->points[j*2+1];
		  }

		  k++;
	      }
	  }
	  n_linear = k;

	  if (n_micro || n_build || n_trees || n_cans)
	  {
	      added = ctdb_store_features(ctdb, patch_x, patch_y, data+f,
					  be_microterrain, n_micro,
					  be_buildings, n_build,
					  be_trees, n_trees,
					  be_canopies, n_cans,
					  be_cedges, n_cedge,	
					  be_linear, n_linear,
					  complete_micro_poly);

	      if (added)
		ctdb_set_patch_feature_start(ctdb, ctdb_patch_index, f);

	      f += added;
	  }

	  if (feature_index_list)
	    quad_free_list (feature_index_list);
      }

    ctdb_print_feature_stats();

    ctdb->num_features = f;

    return f;
}
