#ifndef lint
static char rcsid [] = "$RCSfile: thin.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: this.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Compiler which generates CTDB *from* CTDB                 *
*       Created: March 1996                                                 *
*       Author: mlongtin                                                    *
*       Remarks: This compiler generates a thinned-out version of a CTDB    *
*                file by throwing out treelines, trees, buildings,          *
*                 microterrain, and terrain elements.                       *
*                                                                           *
*       NOTE:  You may have to "limit stacksize unlimited" before running   *
*              this on a Sun.  This works for SIMNET databases only.        *
****************************************************************************/
#define DEBUG

/* For mmap() in encode_physical_features */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <libcmdline.h>

/* Solaris & SunOS <sys/mman.h> lacks MAP_AUTOGROW ... maybe it is SGI-only? */
#ifndef sgi
# ifndef MAP_AUTOGROW
#  define MAP_AUTOGROW 0
# endif
#endif
#define TMPFILE "recompile.tmp"

/* This may have to be increased for really large databases.
 */
#define PHYSICAL_FEATURE_SIZE    40*1024*1024

/* The maximum number of linear models.  Increase this as needed.
 */
#define MAX_NUM_LINEAR_MODELS   400000

/* Pre-format-5 databases didn't store some information that format 5 databases
 * do store.  Use default values for unstored data.
 */
#define DEFAULT_TRUNK_RADIUS    0.5   /* meters */
#define DEFAULT_FULLNESS        0.85

#include <stdio.h>
#include <errno.h>
#include <stdstring.h> /*common/include/global*/
#include <libctdb.h>
#include <ct_feat.h>
#include "ct_cmplr.h"
#include "ct_reform.h"
#include <ct_post.h>
#include <libreader.h>
#include <stdalloc.h>
#include <stdext.h>
#include <time.h>
#include <string.h>
#include <libcoordinates.h>
#include <libgcs.h>


#define INPUT_MEMORY 0 /* the internal working representation is
			  not cache-able */

/* Initial guess at how many cells this assembly covers.  We'll be generous,
 * because it's cheap to guess high
 */
#define START_CELLS 50

/* Compiler Globals */
static CTDB_CMP *ctdb_in, ctdb_out;
int32 output_format;
int32 input_format;

float64 indb_south_lat, indb_north_lat, indb_west_long, indb_east_long;

static int32 gcs_mode;

static COORD_TCC_PTR tcc;	/* TCC pointer for s1k assembly */
static int32 cur_cell;		/* ID of current cell */

CTDB_DBASE_TYPE dbase_type = CTDB_GRIDDED; /*default type */


/* Forward Declarations */
static int32 encode_physical_features();
static void collect_nodes_and_edges();
static int32 encode_abstract_features();
static void collect_linear();
static int32 *generate_elevation_posts();
static int32 convert_soil_index();
static void complete_micro_poly();


#define CMP_MAX_REF 1000	/* Bound on how many tidal zones, water chars,
				 * etc. can be added in one compilation
				 */

#define CTDB_INVALID_REF -1

typedef struct cmp_ref
{
    int32 user_id;
    int32 index;
} CMP_REF;


struct
{
    CMD_STRING_OPTION  input_ctdb;
    CMD_STRING_OPTION  output_base_name;    
    CMD_INTEGER_OPTION format;
    CMD_STRING_OPTION  data_path;
    CMD_STRING_OPTION  tdb_path;
} options = {
    {
	"Input Ctdb", "Absolute pathname to the input CTDB" ,
	NULL ,
	CMD_STRING ,    "input_ctdb" ,     NULL , "knox-0311.ctb" ,
    },
    {
	"Output Name Base", "Output database base name (no extensions \
and path since the path is defined by tdb_path)" ,
	NULL ,
	CMD_STRING ,    "output_base_name" ,     NULL , NULL ,
    },
    {
	"Output Format","Selects the output format" ,
	NULL ,
	CMD_INTEGER ,    "format" ,    NULL ,  CTDB_FILE_FORMAT ,
    },   
    {
	"data_path", " Select the data path" ,
	NULL ,
	CMD_STRING ,    "data_path" ,     NULL , "../../../data" ,
    },
    {
	"tdb_path", " Select the terrain data path for the output database" ,
	NULL ,
	CMD_STRING ,    "tdb_path" ,     NULL , "../../../terrain" ,
    }
};

static char *
generate_tdb_filename(kind,tdb_path,outname,inname,format,cell)
    CTDB_TDB_TYPE kind;
    char         *tdb_path;
    char         *outname;
    char         *inname;
    int32         format;
    int32         cell;
{
    static char dbname[256];
    
    if (outname)
      return ctdb_generate_tdb_filename(kind, tdb_path, outname,format,cell,
                                        TRUE);
    else
    {
	char base[256];
	
	generate_name_base(inname, base);
	return ctdb_generate_tdb_filename(kind, tdb_path, base, format, cell,
                                          TRUE);
    }
}

int main(
    int    argc,
    argv_t argv)
{
    CTDB_FILE_HEADER_CMP hdr;
    int32 i;
    extern int atoi();
    char *dbname;
    int32 leftover_argc;
    argv_t leftover_argv;
    char  *rest_of_str, *cpy_rest_of_str, *current_file;
    int32 err, chars_processed;
    char sp = '\040';

    /* Parse the argument list */
    if (argc < 2)
     {
	 printf("\nINCORRECT USAGE: Restart with option -help for usage.\n");
	 exit(1);
     }

    cmd_process_options(argc, argv, /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized */
			(CMD_OPTION *)&options, /* options array */
			sizeof(options), /* size of options array */
			TRUE /* Verbose mode */
			);

    cmd_gripe(leftover_argc, leftover_argv);

    printf("Compiling in SIMNET mode.\n");
    gcs_mode = CTDB_MODE_SIMNET;

    /* Initialize support libraries */
    reader_init(0);
    
    if (coord_init(options.data_path.value,0))
    {
	printf("Error initializing libcoordinates.\n");
	exit(1);
    }
    coord_set_tcc_awareness(TRUE);
    
    gcs_init(coord_get_datum_info);
    gcs_set_cell_awareness(GCS_MULTI_CELL);

    /* Initialize compiler routines */
    ctdb_cmplr_init();
 
    /* Read the input source.
     */
    printf("\nReading %s:\n", options.input_ctdb.value);
    ctdb_in = ctdb_cmplr_read(options.input_ctdb.value, &input_format);

    /* The input ctdb (for now, anyway) will never be tiled */
    ctdb_set_tiling_mode(FALSE);
    
    ctdb_intern_print_description(ctdb_in);

    /* Check for read errors */
    if (!input_format)
      exit(-1);
    
    output_format = options.format.value;

    if (input_format == 4 && output_format > 4)
      printf("\n\nWarning: If the format 4 database was created prior to the release of\nModSAF 2.1, then there may be some problems with it.  We recommend that you\ngenerate the new database from a previous version, or directly from the s1000\nor other source format.\n\n");

    printf("Output format is %d\n", output_format);

    /* Setup TCC ptr for future coordinate conversions */
    tcc = coord_define_tcc(COORD_UTM_NE,
			   ctdb_in->origin_northing,
			   ctdb_in->origin_easting,
			   ctdb_in->origin_zone_number,
			   ctdb_in->origin_zone_letter,
			   ctdb_in->datum,
			   (ctdb_in->max_x - ctdb_in->min_x),
			   (ctdb_in->max_y - ctdb_in->min_y));

    /* Find the lat-long extents of the input database, stuff them where they
     * need to go, and convert them into a list of cells if we're doing a
     * GCS multi-cell compilation.
     */
    {
	get_ll_bounds(COORD_TCC, GCS_ILLEGAL_CELL, tcc, 
		      (float64)ctdb_in->min_x, (float64)ctdb_in->min_y, 
		      (float64)ctdb_in->max_x, (float64)ctdb_in->max_y,
		      &indb_west_long, &indb_south_lat, 
		      &indb_east_long, &indb_north_lat);
	
#ifdef DEBUG
	printf("Lat/long bounds = (%6.3f, %6.3f) to (%6.3f, %6.3f)\n",
	       indb_south_lat, indb_west_long, 
	       indb_north_lat, indb_east_long);
#endif /* DEBUG */
    }

    if (gcs_mode == CTDB_MODE_SIMNET)
    {
	dbname = generate_tdb_filename(CTDB_MODE_SIMNET, 
				       options.tdb_path.value,
				       options.output_base_name.value,
				       options.input_ctdb.value,
				       output_format, GCS_ILLEGAL_CELL);
    }

    printf("\n\nCreating %s\n",dbname);

    /**
     ** Step  1: Fill in the header.  Include everything but feature counts
     **          (which are not known yet).
     **/
    bzero((void*)&hdr, sizeof(hdr)); /* So padding has zeros */
    ctdb_intern_pack_header(ctdb_in, &hdr);
    hdr.format = output_format;
    {
	int tloc;
	time((time_t *)&tloc);
	strcpy(hdr.date, ctime((time_t *)&tloc));
    }

    /* Post-format-4 databases have at least `CTDB_MAX_CANOPY_MODELS'
     * linear models.  The first `CTDB_MAX_CANOPY_MODELS' are for canopies,
     * the rest are for other linear features.
     */
    if(output_format < 5)
      hdr.num_linear_models = 0;
    else
      hdr.num_linear_models = CTDB_MAX_CANOPY_MODELS;

    /**
     ** Step  2: Unpack the header into a CTDB structure, and copy any
     **          necessary additional data from the old ctdb to the new
     **          one.
     **/
    bzero((void*)&ctdb_out, sizeof(ctdb_out));
    ctdb_intern_unpack_header(&ctdb_out, &hdr);

    /* Zero out fields we'll probably fill in later just to be safe */
    ctdb_out.num_features = 0;
    ctdb_out.num_nodes = 0;
    ctdb_out.node_size = 0;
    ctdb_out.num_edges = 0;
    ctdb_out.edge_size = 0;
    ctdb_out.num_quad_data = 0;

    /* Zero out counts for things we don't put into thinned databases */
    ctdb_out.num_volume_models    = 0;
    ctdb_out.num_aggregate_models = 0;
    ctdb_out.num_mes_volumes      = 0;
    ctdb_out.num_mes_templates    = 0;

    /* Copy the soil tables from the input database to the output
     * database.
     */
    {
	int32 i, j;

	ctdb_out.soil_tables = (int32 **)
	  ctdb_alloc(NULL, ctdb_in->num_soil_tables * sizeof(int32 *),
		     "Compiler rep. soil table");
	ctdb_out.soil_table_storage = (int8 *)
	  ctdb_alloc(NULL, ctdb_in->soil_table_size,
		     "Compiler rep. soil table data");
	    
	bcopy((void*)ctdb_in->soil_table_storage,
              (void*)ctdb_out.soil_table_storage,
	      ctdb_in->soil_table_size);
	    
	ctdb_out.soil_tables[0] = (int32 *)ctdb_out.soil_table_storage;
	for(i=1;i<ctdb_in->num_soil_tables;i++)
	{
	    ctdb_out.soil_tables[i] = 
	      (int32 *)(ctdb_out.soil_table_storage +
			((int8 *)ctdb_in->soil_tables[i] - 
			 ctdb_in->soil_table_storage));
	}

	/* Copy over the PAT, too */
	ctdb_out.pat_table_ptr = (CTDB_PAT_TABLE_CMP *)
	  ctdb_alloc(NULL, sizeof(CTDB_PAT_TABLE_CMP),
		     "Compiler Polygon Attribute Table");
	ctdb_out.pat_table_ptr->num_columns =
	  ctdb_in->pat_table_ptr->num_columns;
	ctdb_out.pat_table_ptr->alloced_columns =
	  ctdb_in->pat_table_ptr->num_columns;
	ctdb_out.pat_table_ptr->num_entries =
	  ctdb_in->pat_table_ptr->num_entries;
	ctdb_out.pat_table_ptr->alloced_entries =
	  ctdb_in->pat_table_ptr->alloced_entries;
	ctdb_out.pat_table_ptr->columns = (CTDB_PAT_COLUMN_HDR_CMP *)
	  ctdb_alloc(NULL, ctdb_out.pat_table_ptr->alloced_columns *
		     sizeof(CTDB_PAT_COLUMN_HDR_CMP),
		     "Compiler PAT column headers");
	for(i=0; i<ctdb_out.pat_table_ptr->num_columns; i++)
	{
	    ctdb_out.pat_table_ptr->columns[i].facc =
	      ctdb_in->pat_table_ptr->columns[i].facc;
	    ctdb_out.pat_table_ptr->columns[i].data =
	      (CTDB_PAT_DATA_CMP *)
	      ctdb_alloc(NULL, ctdb_out.pat_table_ptr->alloced_entries *
			 sizeof(CTDB_PAT_DATA_CMP),
			 "Compiler PAT data");
	    bcopy((void*)ctdb_in->pat_table_ptr->columns[i].data,
		  (void*)ctdb_out.pat_table_ptr->columns[i].data,
		  ctdb_in->pat_table_ptr->num_entries *
		  sizeof(CTDB_PAT_DATA_CMP));
	    for(j=ctdb_out.pat_table_ptr->num_entries;
		j<ctdb_out.pat_table_ptr->alloced_entries;
		j++)
	      ctdb_out.pat_table_ptr->columns[i].data[j] =
		ctdb_out.pat_table_ptr->columns[i].data[0];
	}
    }
    

    /* Copy over the water characteristics and tidal zones from the 
     * input
     */
    ctdb_out.water_chars = (CTDB_WATER_CHARS_CMP *)
      ctdb_alloc(NULL, ctdb_in->num_water_chars *
		 sizeof(CTDB_WATER_CHARS_CMP),
		 "Cmp. rep. water chars");

    bcopy((void*)ctdb_in->water_chars, (void*)ctdb_out.water_chars,
	  ctdb_in->num_water_chars * sizeof(CTDB_WATER_CHARS_CMP));

    /* Ensure that the default water char points at the default tidal
     * zone
     */
    ctdb_out.water_chars[0].tidal_zone_index = 1;

    ctdb_out.tidal_zones = (CTDB_TIDAL_ZONE_CMP *)
      ctdb_alloc(NULL, ctdb_in->num_tidal_zones *
		 sizeof(CTDB_TIDAL_ZONE_CMP),
		 "Cmp. rep. water chars");

    bcopy((void*)ctdb_in->tidal_zones, (void*)ctdb_out.tidal_zones,
	  ctdb_in->num_tidal_zones * sizeof(CTDB_TIDAL_ZONE_CMP));

    /* Add the data that would normally come from the aux header */
    if (gcs_mode == CTDB_MODE_SIMNET)
    {
	/* The origin offset is measured as (X, Y) where
	 * X is the offset towards the east, and Y is the offset
	 * towards the north.  Therefore, the offset to the 
	 * SW corner is negative 
	 */
	ctdb_out.origin_x = -(ctdb_out.min_x + ctdb_out.max_x) / 2;
	ctdb_out.origin_y = -(ctdb_out.min_y + ctdb_out.max_y) / 2;
	ctdb_out.cell_id = GCS_ILLEGAL_CELL;
    }
	
    {
	float64 min_x, min_y, max_x, max_y;

	ctdb_out.gcs_mode = gcs_mode;
	    
	ctdb_out.west_long = indb_west_long;
	ctdb_out.south_lat = indb_south_lat;
	ctdb_out.east_long = indb_east_long;
	ctdb_out.north_lat = indb_north_lat;
    }	

    /**
     ** Step  3: Generate the elevation posts (if we need them)
     **/	
    if ((output_format < 4) || (dbase_type != CTDB_TINNED))
    {
	if (!ctdb_in->elevations)
	{
	    ctdb_out.elevations = generate_elevation_posts();

	    ctdb_in->elevations = (int32 *)
	      ctdb_alloc(NULL, ctdb_in->pages_wide *
			 ctdb_in->pages_high * PAGE_SIZE,
			 "New elevation posts");
	    ctdb_in->soils = (int32 *)
	      ctdb_alloc(NULL, ctdb_in->pages_wide *
			 ctdb_in->pages_high * PAGE_SIZE,
			 "New soil posts");

	    bcopy((void*)ctdb_out.elevations, (void*)ctdb_in->elevations, 
		  ctdb_in->pages_wide * ctdb_in->pages_high * PAGE_SIZE);
	    bcopy((void*)ctdb_out.soils, (void*)ctdb_in->soils, 
		  ctdb_in->pages_wide * ctdb_in->pages_high * PAGE_SIZE);
	}
	else
	{
	    ctdb_out.elevations = generate_elevation_posts();
	}
    }	

    /**
     ** Step  4: Generate the quad tree of abstract features.
     **          After doing so, we know the count, so put that
     **          into the header.  We need to do this *before* generating
     **          physical features in case we're generating multi-level
     **          water, which uses the ocean outlines.
     **/
#ifdef DEBUG
	printf("\nGenerating quadtree...\n");
#endif /* DEBUG */
    if (output_format > 2) /* These were added in format 3 */
      hdr.num_quad_data = encode_abstract_features();

    /**
     ** Step  5: Generate a list of physical features in memory.
     **          After doing so, we know the feature count, so put that
     **          into the header.
     **/
    printf("\nProcessing physical features...\n");
    hdr.num_features = encode_physical_features(output_format, dbname, 
						dbase_type,
						&hdr.num_linear_models);
	
    /**
     ** Step  6: Generate the lists of nodes and edges in memory.
     **          This routine also fills in size and count fields
     **          in the header.
     **/
#ifdef DEBUG
    printf("\nGenerating nodes and edges...\n");
#endif /* DEBUG */
    if (output_format > 2) /* These were added in format 3 */
      collect_nodes_and_edges(&hdr);	
	
    /**
     ** Step  7: Write the database to disk
     **/
    ctdb_cmplr_write(&ctdb_out, dbname, output_format);

    ctdb_print_feature_stats();
	
    if(ctdb_out.linear_models)
      STDDEALLOC(ctdb_out.linear_models);

#ifdef sgi
    if (ctdb_out.features)
      munmap(ctdb_out.features, 
	     hdr.num_features * sizeof(CTDB_FEATURE_DATA_CMP));
#endif

    return 0;
}

/* This function extracts the physical features from the input CTDB
 * file, puts them into the form expected by the compiler back end,
 * then passes them down to the back end for encoding.  The return
 * value is the number of feature data (4-byte nuggets) encoded.
 */
static int32 encode_physical_features(format, dbname, dbase_type,
				      num_linear_models)
    int32 format;
    char *dbname;
    CTDB_DBASE_TYPE dbase_type;   
    int32 *num_linear_models;
{
    int32 patches;
    int32 patch_meters;
    int32 patch_x, patch_y;
    float64 corner_x, corner_y;
    int32 feature_offset;
    int32 added;
    CTDB_LAID_LINEAR *linear, *all_linear;
    int32 suppress_linear, all_n_lines;
    int32 n_lines = 0;
    TRANSITION_PATCH_INFO *trans_patches;  /* used for hybrid databases:
					    * identifies gridded patches that
					    * must converted to tinned patches
					    * to provide a transition patch
					    * adjacent tinned patches. 
					    */
    int32 trans_end = 0;
    char temp_fname[CTDB_FILENAME_DIMENSION_CMP];
    FILE *fptr;
    float64 max_elev;
    int32 num_te;
    int32 te_offset=0;
    int32 total_tes=0;
    int32 corrections=FALSE;
    int32 new_micro;
    int32 min_px, max_px, min_py, max_py; /* Bounds on patch_x, patch_y
					   * covered on input db by current
					   * (GCS) patch 
					   */
    int32 gcs_patch_num;	/* Current patch num for GCS db */

    int32 i,j;
    
    static CTDB_LAID_LINEAR local;

    int8 *post_array = NULL;

    int32 skip_land_polygons;

    /* used to record tinning patch information for format 4+ */
    trans_patches = (TRANSITION_PATCH_INFO *)
      ctdb_alloc(NULL,sizeof(TRANSITION_PATCH_INFO)*
		  ctdb_out.patches_wide * ctdb_out.patches_high,
		 "trans patches");
    bzero((void*)trans_patches,
          sizeof(TRANSITION_PATCH_INFO)*ctdb_out.patches_wide *
	  ctdb_out.patches_high);

    ctdb_zero_out_patch_buffer();

    /* for format 4+, record tinned terrain element into a temporary file */
    if (format >3)
    {
	sprintf(temp_fname, "%s.te",dbname);
	if ((fptr=fopen(temp_fname,"wb")) == (FILE *) NULL )
	{
	    printf("Unable to open temporary file %s in current dir.\n",
		   dbname);
	    exit(0);
	}
    }

    /* Allocate memory to hold the patch groups (global information
     * about each patch).
     */
    patches = ctdb_out.patches_wide * ctdb_out.patches_high;
    ctdb_out.patch_groups =
      (CTDB_PATCH_GROUP_CMP *)STDALLOC(sizeof(CTDB_PATCH_GROUP_CMP)*
				       (patches/CTDB_PATCHES_PER_GROUP_CMP));
    bzero((void*)ctdb_out.patch_groups,
	  sizeof(CTDB_PATCH_GROUP_CMP)*(patches/CTDB_PATCHES_PER_GROUP_CMP));

/* mmap() doesn't support MAP_AUTOGROW on the sun and mips.  Do it the
 * old-fashioned way and use malloc() in this case.
 */
#ifndef sgi
    if(!(ctdb_out.features =
	 (CTDB_FEATURE_DATA_CMP *)STDALLOC(PHYSICAL_FEATURE_SIZE)))
    {
	printf("Memory allocation error.\n");
	exit(-1);
    }

#else
    /** Create a memory-mapped file which we can use as virtual
     ** address space (we could just malloc a big hunk of memory,
     ** but that requires too much swap).
     **/
    {
	int fd = open(TMPFILE, O_RDWR|O_CREAT, 0666);
	if (fd == -1)
	{
	    char errbuf[1024];
	    sprintf(errbuf, "Unable to open temporary file %s", TMPFILE);
	    perror(errbuf);
	    exit(-1);
	}
	ctdb_out.features = (CTDB_FEATURE_DATA_CMP *)
	  mmap(0, 96*1024*1024,
	       PROT_READ|PROT_WRITE, MAP_SHARED|MAP_AUTOGROW, fd, 0);
	if (ctdb_out.features == (CTDB_FEATURE_DATA_CMP *)-1)
	{
	    perror("Unable to memory-map temporary file");
	    exit(-1);
	}
	close(fd);
	unlink(TMPFILE);
    }
#endif
    
    /* Allocate space for the linear models.
     */
    ctdb_out.linear_models = (CTDB_LINEAR_MODEL_CMP *)
                             STDALLOC(MAX_NUM_LINEAR_MODELS *
			     sizeof(CTDB_LINEAR_MODEL_CMP));
    bzero((void*)ctdb_out.linear_models, MAX_NUM_LINEAR_MODELS *
	  sizeof(CTDB_LINEAR_MODEL_CMP));

    feature_offset = 0;

    if ((input_format < 4) ||
	(dbase_type == CTDB_TINNED) ||
	(gcs_mode != CTDB_MODE_SIMNET))
      new_micro = TRUE;
    else
      new_micro = FALSE;


    /** Traverse the patches in order.
     **/
    patch_meters = ctdb_out.incr * ctdb_out.patch_incr;
    for (patch_y = 0; patch_y < ctdb_out.patches_high; patch_y++,
	 putchar('.'),fflush(stdout))
    {
	for (patch_x = 0; patch_x < ctdb_out.patches_wide; patch_x++)
	{
	    max_elev = -1e10;
	    num_te = 0;
	    skip_land_polygons = FALSE;

	    gcs_patch_num = patch_y * ctdb_out.patches_wide + patch_x;

	    /* Init the dynamic patch buffer */
	    ctdb_init_patch_buffer();

	    /* Compute the lower left corner of this patch in GCS coords */
	    corner_x = patch_x * patch_meters;
	    corner_y = patch_y * patch_meters;

	    /* Figure out which patches on the input database overlap the
	     * current patch--in SIMNET mode, this is trivial, in GCS modes,
	     * it may require some work to figure out.
	     */
	    min_px = max_px = patch_x;
	    min_py = max_py = patch_y;

	    /* Collect the linear features
	     */
	    for(i=min_py;i<=max_py;i++)
	      for(j=min_px;j<=max_px;j++)
		collect_linear(j, i);

	    /* Add them starting at the current offset into the
	     * feature array.
	     */
	    added = ctdb_store_patch_buffer(&ctdb_out, 
					    patch_x, patch_y,
					    ctdb_out.features+feature_offset,
					    ctdb_out.linear_models,
					    &ctdb_out.num_linear_models,
					    NULL, NULL,
					    complete_micro_poly, &num_te, 
					    trans_patches, &trans_end,
					    &max_elev, format, dbase_type, 
					    corrections, new_micro, fptr,
					    FALSE);

	    if(ctdb_out.num_linear_models >= MAX_NUM_LINEAR_MODELS)
	    {
		printf("fatal error:  MAX_NUM_LINEAR_MODELS exceeded.  Increase it.\n");
		exit(-1);
	    }

	    /* Note the starting address in the patch group structure */
	    if (added)
	      ctdb_set_patch_feature_start(&ctdb_out,
					   patch_y * ctdb_out.patches_wide +
					   patch_x,
					   feature_offset);

	    ctdb_set_te_patch_header(&ctdb_out, patch_x, patch_y, 
				     te_offset, &num_te, max_elev, 
				     trans_patches, &trans_end, 
				     ctdb_out.features+feature_offset,
				     added, dbase_type);
	    te_offset += num_te * sizeof(TERRAIN_ELEMENT_16);
	    total_tes += num_te;
	    
	    /* Increment the offset for the next set of features */
	    feature_offset += added;
	}
    }

    free(ctdb_in->features);
    if (ctdb_in->elevations)
      STDDEALLOC(ctdb_in->elevations);
    if (ctdb_in->terrain_elements)
      STDDEALLOC(ctdb_in->terrain_elements);

    if (post_array)
      STDDEALLOC(post_array);

    /* For format 4+, inter-patch terrain element relationships must also be
     * resolved. 
     */
    if (format > 3)
    {
	fclose(fptr);
	ctdb_out.n_te_bytes = total_tes * sizeof(TERRAIN_ELEMENT_16);
	/* Don't try to fill any holes for SIMNET mode */
	ctdb_te_dbase_processing(&ctdb_out, corrections, trans_patches,
				 trans_end, dbname, FALSE);
    }

    *num_linear_models = ctdb_out.num_linear_models;

    return feature_offset;
}

/* Callback function provided to the back end to fill in holes in the input.
 */
static void complete_micro_poly(poly)
    CTDB_MC_POLYGON *poly;
{
;
}

#define START_VERTS 1000

/* This routine collects the abstract features from the input CTDB file,
 * adds in abstract features from the corrections files, and hands it
 * all down to the compiler back-end for encoding.  The return value is
 * the number of feature data (4-byte nuggets) encoded.
 */
static int32 encode_abstract_features()
{
    CTDB_SEARCH_SPACE_PTR_CMP search_space; 
    int32 i, j;
    int32 count;
    int32 code;
    float32 *verts;
    CTDB_ABSTRACT_DATA_CMP *data;
    int32 data_size;
    CTDB_COMPILER_QUAD quad_root;
    READER_UNION *c;
    char *add_abstract = reader_get_symbol("ADD_ABSTRACT");
    float32 *lverts = NULL;
    int32 num_lverts = 0;
    int32 *soil_table;
    int32 soil;
    
    /** Initialize the root of the compiler's quad tree.
     **/
    bzero((void*)&quad_root, sizeof(quad_root));

    /** Read in all the features from the original database. Traverse
     ** "by patch" for speed.
     **/
    search_space =
      ctdb_intern_create_search_space(ctdb_in,
				      ctdb_in->min_x, ctdb_in->min_y, 
				      ctdb_in->max_x, ctdb_in->max_y, 
				      CTDB_BY_PATCH);
    if (!search_space)
    {
	printf("Unable to create search space. Probably out of memory\n");
	return 0;
    }
    /** Traverse all the patches.
     **/
    while (ctdb_intern_next_quad_patch(search_space))
    {
	/* Look at every known code value, and every feature */
	for (code=1;code<CTDB_ABSTRACT_MAX;code++)
	  while (count = ctdb_intern_next_abstract(search_space, code,
						   &verts, &data, 
						   &data_size))
	  {
	      /* If it's an ocean and we're trying to "fix" the water
	       * for format 5+, grab it and update the list of patch
	       * water states
	       */
	      ctdb_store_abstract(&ctdb_out, &quad_root,
				  code, count, verts,
				  CTDB_ABSTRACT_DATA_SIZE(data_size),
				  (int32 *)data,
				  ctdb_out.min_x, ctdb_out.min_y,
				  ctdb_out.max_x, ctdb_out.max_y);
	  }
    }
    
    /** Free up the search space.
     **/
    ctdb_intern_destroy_search_space(search_space);
    
    if (num_lverts > 0)
      free(lverts);

    /** Tell the compiler to encode the entire quad tree.
     **/
    ctdb_encode_abstract(&ctdb_out, &quad_root);

    return ctdb_out.num_quad_data;
}

/* The compiler back-end has a routine to connect up linear features
 * into networks.  We use that routine to regenerate the nodes and
 * edges from scratch.
 */
static void collect_nodes_and_edges(hdr)
    CTDB_FILE_HEADER *hdr;
{
    /* Leave creation up to the back end */
    ctdb_create_networks(&ctdb_out, output_format);

    /* Fill in the various header fields */
    hdr->num_nodes = ctdb_out.num_nodes;
    hdr->node_size = ctdb_out.node_size;
    hdr->num_edges = ctdb_out.num_edges;
    hdr->edge_size = ctdb_out.edge_size;
}

static int32 *generate_elevation_posts()
{
    float64 x,y,dx,dy;
    float64 loc_x, loc_y;
    float64 elev;
    float64 page_meters, post_meters;
    uint32 p, s;
    int32 sNE, sSW;
    int32 *new_posts, *new_soils;
    int32 norm_x, norm_y;
    int32 post_index;
    
    new_posts = (int32 *)ctdb_alloc(NULL, ctdb_in->pages_wide *
				    ctdb_in->pages_high * PAGE_SIZE,
				    "New elevation posts");
    new_soils = (int32 *)ctdb_alloc(NULL, ctdb_in->pages_wide *
                                    ctdb_in->pages_high * PAGE_SIZE,
                                    "New soil indices");
    
    post_meters = ctdb_out.incrD;
    page_meters = POSTS_PER_SIDE * post_meters;

    /** Traverse the database in the funny order used by CTDB to store
     ** the grid posts.
     **/

    for (y=0; y<=ctdb_out.max_y; y+=page_meters)
      for (x=0; x<=ctdb_out.max_x; x+=page_meters)
	for (dy=0.0; dy<page_meters; dy+=post_meters)
	  for (dx=0.0; dx<page_meters; dx+=post_meters)
	  {
	      loc_x = x+dx;
	      loc_y = y+dy;

	      norm_x = (int32)POST_X(loc_x * ctdb_in->inv_incr);
	      norm_y = (int32)POST_Y(loc_y * ctdb_in->inv_incr);
	      
	      /* If the input database has a grid, then use it.  Otherwise,
	       * get the elevation data from the terrain elements.
	       */
	      if(!ctdb_in->elevations)
	      {
		  ctdb_intern_lookup_te_info((float64)norm_x, (float64)norm_y, 
					     ctdb_in, &elev, 
					     &sNE, &sSW, NULL);

		  p = BUILD_POST_ELEVATION(elev,
					   ctdb_in->inv_fixed_point_basis);
		  p &= ~POST_FLAGS;
		  s = BUILD_POST_NE_SOIL(sNE) | BUILD_POST_SW_SOIL(sSW);
	      }
	      else
	      {
		  /* CTDB does not provide any interfaces to get exactly what
		   * we need here.  Basically we want to get the elevations
		   * and soil types from the grid *ONLY*, ignoring features,
		   * microterrain, etc.  This is not the sort of interface we
		   * want to provide in libctdb, so instead we require the
		   * caller to know a bit about how stuff is stored.
		   */
		  p = ctdb_intern_lookup_elev_post(ctdb_in, norm_x, norm_y);
		  p &= ~POST_FLAGS;
		  s = ctdb_intern_lookup_soil_post(ctdb_in, norm_x, norm_y);

		  /* Before version 3, the soils in SW-NE diagonalized dbs
		   * were stored in the next-post down from the post we use
		   * now (don't ask why).  Correct this.
		   */
		  if ((p & POST_SW_NE) && (ctdb_in->input_format < 3)) 
		  { 
		      s = ctdb_intern_lookup_soil_post(ctdb_in, norm_x, 
						       norm_y-1);
		  } 
		  
		  post_index = ctdb_intern_get_post_index(ctdb_in, norm_x, 
							  norm_y);
		  new_posts[post_index] = p;
		  new_soils[post_index] = s;
	      } 
	      
	      ctdb_out.soils = new_soils;
	      return new_posts;
	  }
}


static void collect_linear(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    CTDB_LAID_LINEAR line;
    float64 verts[CTDB_VERTICES_MAX][2];
    float64 width;
    int32 soil;
    int32 n_verts;
    int32 i;
    CTDB_FEATURE_DATA_CMP *feature, *last;
    int32 patch_number = patch_x + patch_y * ctdb_in->patches_wide;
    int32 next, fclass;

    feature = ctdb_intern_lookup_patch(ctdb_in, patch_number);
    if (feature)
    {
	last = feature + feature->hdr.size;
	feature++;
	for(; feature < last; feature += next)
	{
	    n_verts = feature->info.vertices;
	    fclass = feature->info.feature_class;
	    
	    if (fclass == FC_MICRO_CMP)
	    {
		next = NEXT_MICRO(n_verts);
		continue;
	    }
	    else if (fclass == FC_VOLUME_CMP)
	    {
		next = NEXT_VOLUME_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_CANOPY_CMP)
	    {
		next = NEXT_CANOPY_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_AGGREGATE_CMP)
	    {
		next = NEXT_AGGREGATE_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LINEAR_CMP)
	    {
		next = NEXT_LINEAR_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LAID_LINEAR_CMP)
	      next = NEXT_LAID_LINEAR_CMP(n_verts);

	    if (n_verts)
	    {
		if (n_verts >= MAX_LAID_LINEAR_VERTS)
		{
		    printf("MAX_LAID_LINEAR_VERTS (ct_cmplr.h) exceeded.  Increase it to %d\n",
			   n_verts+1);
		    n_verts = MAX_LAID_LINEAR_VERTS-1;
		}		    
		line.n_verts = n_verts;
		line.width = 
		  ctdb_in->incrD *
		    XYO_TO_POST(feature[2].laid_linear.width,0,
				ctdb_in->patch_incrD);
		line.soil = feature[1].ll_soil.poly_char_index;

		for (i=0;i<n_verts;i++)
		{
		    line.verts[i][X] = 
		      ctdb_in->incrD *
			XYO_TO_POST(feature[3+i].xy.x,patch_x,
				    ctdb_in->patch_incrD);
		    line.verts[i][Y] =
		      ctdb_in->incrD *
			XYO_TO_POST(feature[3+i].xy.y,patch_y,
				    ctdb_in->patch_incrD);
		}

		ctdb_buffer_laid_linear(&line);			
	    } /* n_verts */
	} /* feature list */
    } /* feature */
}


/* Mapping from pre-4 to post-3 database soil types.  The annotations are
 * CCTT-related.
 */
static int32 augmented_soil_types[] =
{
    /* Undefined */
    (0 << 4) | 0,

    /* (Dry Pavement: Concrete, Asphalt) */
    CTDB_ROAD_FLAG | (20 << 4) | 1,

    /* (Dry Loose Surface Road; GW, GP, GM, HC, Rock) */
    CTDB_ROAD_FLAG | (17 << 4) | 2,

    /* (Soft; GW, GP, GM, GC, SM, SC, CL, ML, CH, MH, OL, OH) */
    (5 << 4) | 3,

    /* (Depth 60 inches; Water w/ (Silts and Clays) Bottom) */
    CTDB_WATER_FLAG | CTDB_NO_GO_FLAG | (28 << 4) | 4,

    /* (Depth 16 inches; Water w/ (Silts and Clays) Bottom) */
    CTDB_WATER_FLAG | CTDB_SLOW_GO_FLAG | (26 << 4) | 5,

    /* (Very Soft; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG | (4 << 4) | 6,

    /* (Wet Loose Surface Road; GW, GP, GM, HC, Rock) */
    CTDB_ROAD_FLAG | CTDB_SLOW_GO_FLAG | (18 << 4) | 7,

    /* (Very Hard (Slippery); CH, OH) */
    CTDB_SLOW_GO_FLAG | (15 << 4) | 8,

    /* (Swamps, Bogs, Etc.; NO-GO) */
    CTDB_SLOW_GO_FLAG | (19 << 4) | 9,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG | (7 << 4) | 10,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    (7 << 4) | 11,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    (7 << 4) | 12,

    /* (Brushland - Medium; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG | (22 << 4) | 13,

    /* (Very Hard; GW, GP, GM, GC, SM, SC, CL, ML, MH, OL) */
    CTDB_NO_GO_FLAG | (9 << 4) | 14,

    /* (Swamps, Bogs, Etc.; NO-GO) */
    CTDB_NO_GO_FLAG | (19 << 4) | 15
};

static int32 convert_soil_index(input_format, output_format, soil_index)
    int32          input_format;
    int32          output_format;
    int32          soil_index;
{
    /* If we're converting from a post-3 format to a pre-4 format, we
     * just need to extract the SIMNET constant from the soil type and
     * use that as the index, since for pre-4 databases, the following
     * holds:
     *
     * soil_map[i] = i
     */

    /* $$$ This is not right if we have multiple soil tables... */
    if(input_format > 3 && output_format < 4)
      return CTDB_SIMNET_SOIL(ctdb_in->soil_tables[0][soil_index]);

    return soil_index;
}


