#ifndef lint
static char rcsid [] = "$RCSfile: ct_8or16.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_8or16.c                                                        *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: This file contains all the routines which are dependant   *
 *                 on the size of TERRAIN ELEMENTS (either 8 or 16 bits).    *
 *       Created:  February 1995                                             *
 *       Author:   cbuettner                                                 *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdext.h>
#include "libctdb_local.h"
#include <libgcs.h>
#include <libvecmat.h>

#define MARKLENGTH               10000
#define CTDB_MAX_EDGES_PER_TE        5 /* max edges per terrain element */


static uint32  checked_array[MARKLENGTH];
static uint32 *checked_f;/* front point of the checked TE array */
static uint32 *checked_b;/* back pointer of the checked TE array */
#ifdef CTDB_INTERVIS_METRICS
extern CTDB_COLLECTION ctdb_data;
#endif 
static int32
validate_te_indexXX (
		     uint32  te_index, 
		     uint32 *front, 
		     uint32 *back    
		     );

static float64
compute_te_normalXX ( 
		     ELEVATION_LOOKUP_STORAGE *els 
		     );

static float64
compute_elevation_on_edgeXX (
			     TERRAIN_ELEMENT_XX   *base,
			     TERRAIN_ELEMENT_XX   *te,      
			     CTDB_COORD2D_I32     *pt,      
			     CTDB_POINT_STATE      edge 
			     );     

static void
set_elevation_and_soilXX ( 
			  ELEVATION_LOOKUP_STORAGE *els 
			  );

static int32	         
point_inside_teXX (
		   CTDB_COORD2D_I32         *a,
		   CTDB_COORD2D_I32         *b,
		   int32                     side,           
		   ELEVATION_LOOKUP_STORAGE *els
		   );

static void
point_in_terrain_elementXX ( 
			    ELEVATION_LOOKUP_STORAGE *els 
			    );

static CTDB_STATUS
look_at_adjacent_trianglesXX ( 
			      ELEVATION_LOOKUP_STORAGE *els 
			      );
    
static CTDB_STATUS
look_at_all_trianglesXX ( 
			 ELEVATION_LOOKUP_STORAGE *els 
			 );   

#ifdef CTDB_INTERVIS_METRICS                                      
#define COUNT_EDGE_CHECK ctdb_data.te_edges_crossed++
#define COUNT_EDGE_INTERSECTION ctdb_data.te_edges_intersected++
#define COUNT_VERTEX           ctdb_data.vertices_crossed++
#define TEST_VERTEX_AND_COUNT \
{                             \
    if ( vtx_flg )            \
      COUNT_VERTEX;           \
}                          

#else
#define COUNT_EDGE_CHECK
#define COUNT_EDGE_INTERSECTION
#define COUNT_VERTEX 
#define TEST_VERTEX_AND_COUNT
#endif

/* VMAT2E_DOT_PROD will return if -1 then the vectors are opposite in 
 * direction 
 * and the point will lye between the two end points.  If the result is 
 * +1 then the point is outside the segment since the vectors are in the 
 * same direction
 */

/* VMAT2E_CROSS_PROD define returns cross product of two vectors where 
 * (x0,y0) and 
 * (x1,y1) define the endpoints of the vectors which contain the same 
 * anchor point (i.e. sin of the angle between them).  If the result
 * is negative then the second endpoint is to the left of the first 
 * point, if 0 then the point is colinear, else the second point lays 
 * to the left of the first point.
 * 
 */

/* Evaluates to TRUE iff los intersects segment from te1 to te2 */
#define ANY_INTERSECTION(te1, te2, s, shape)                             \
(((status = ctdb_segment_intersection_t(&(los->xyo),((te1)->pt),         \
					((te2)->pt),                     \
					&tray, &tedge)) != DISJOINT ) && \
 ( tray > ptop->tray ) )

/* Save only the closest intersection: Find tray for our start point into
 * the patch.  We want the edge whose intersection is closest to this
 * point but beyond it. 
 */
#define FIRST_TE_CHECK_EDGE(te1, te2, s, shape)                 \
{                                                               \
    CTDB_POINT_STATE status;                                    \
    COUNT_EDGE_CHECK;                                           \
    if ( ANY_INTERSECTION(te1,te2,s,shape) )                    \
    {                                                           \
	COUNT_EDGE_INTERSECTION ;                               \
	  if (tray<best_tray)                                   \
	  {                                                     \
	      side = (s);                                       \
	      adj_side = ((lte->info.adj_sides>>((s)<<1))&0x3); \
	      best_tray = tray;                                 \
	      best_edge = tedge;                                \
	      best_te1  = (te1);                                \
	      best_te2  = (te2);                                \
	      horiz = ((te1)->pt.y == (te2)->pt.y);             \
	      vtx_flg = (status!=INTERSECTION);                 \
	      TEST_VERTEX_AND_COUNT                             \
	  }                                                     \
    }                                                           \
}

#define PROCESS_INTERSECTION(h, x, y, s )                    \
{                                                            \
    ptop->this_te    = lte->adj[s];                          \
    if ( ptop->this_te > MAX_ALLOWED_TE_INDEX_XX )           \
      return CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING;      \
    if ( lte->info.patch_adj & (0x1<<s) )                    \
    { /* patch boundry TE */                                 \
	static CTDB_BEARING direction ;                      \
	if ( h ) /* horizontal edge */                       \
	{ /* NORTH or SOUTH */                               \
	    if ( !(y) )                                      \
	      direction = SOUTH ;                            \
	    else                                             \
	      direction = NORTH ;                            \
	}                                                    \
	else                                                 \
	{ /* EAST or WEST */                                 \
	    if ( !(x) )                                      \
	      direction = WEST ;                             \
	    else                                             \
	      direction = EAST ;                             \
	}                                                    \
	if ( direction & ptop->motion )                      \
	{                                                    \
	    ptop->patch_to   = direction  ;                  \
	    return CTDB_STATUS_INTERVIS_EXIT_PATCH;          \
	}                                                    \
    }                                                        \
}

/* Macros to test an edge (consisting of TERRAIN_ELEMENT_XX* v1 and v2, plus
 * edge number s) for intersection with the line of sight.  These macros
 * depends on local variables defined in ptop_te_compare.
 */

/* Check quad edge 4 (the internal edge) for effect on intervis */
#define CHECK_QUAD_INTERNAL_EDGE                                      \
{                                                                     \
    CTDB_POINT_STATE status;                                          \
    COUNT_EDGE_CHECK;                                                 \
    if ( ANY_INTERSECTION(lte,te+lte->vert.index[0],3,0) )            \
    {                                                                 \
        COUNT_EDGE_INTERSECTION ;                                     \
	if (!vis_func(ctdb, los,(float64)lte->z,                      \
		      (float64)te[lte->vert.index[0]].z,              \
		      tedge,tray))                                    \
	  return CTDB_STATUS_INTERVIS_COMPLETE;                       \
	ptop->tray = tray;                                            \
    }                                                                 \
}                                                                     \

/* Don't call on the inside edge of a quadrilateral! */
#define CHECK_EDGE(te1, te2, s, shape)                                     \
{                                                                          \
    CTDB_POINT_STATE status;                                               \
    COUNT_EDGE_CHECK;                                                      \
    if ( ANY_INTERSECTION( te1, te2, s, shape ) )                          \
    {                                                                      \
	COUNT_EDGE_INTERSECTION ;                                          \
	if (!vis_func(ctdb, los,(float64)((te1)->z),                       \
		      (float64)((te2)->z),tedge,tray))                     \
	  return CTDB_STATUS_INTERVIS_COMPLETE;                            \
	if (status == INTERSECTION)                                        \
	{                                                                  \
	    ptop->entry_side  = ((lte->info.adj_sides >> (s << 1)) & 0x3); \
	    if ( tray > ptop->tray )                                       \
	    {                                                              \
		ptop->tray = tray ;                                        \
		PROCESS_INTERSECTION( ((te1)->pt.y==(te2)->pt.y),          \
				     ((te1)->pt.x), ((te1)->pt.y), s)      \
		  continue ;                                               \
	    }                                                              \
	}                                                                  \
	else                                                               \
	{ /* Intersection was on a vertex */                               \
	    /* vertex case */                                              \
	    COUNT_VERTEX;                                                  \
	    on_vertex = TRUE;                                              \
	    /* If we find an edge intersection for the current TE, we're   \
	     * going to ignore this vertex intersection, so we can't do    \
	     * anything permanent here.  We'll change all the iteration    \
	     * variables only if we _do_ use this intersection             \
	     */                                                            \
	    vtx_tray = tray;                                               \
	}                                                                  \
    }                                                                      \
}

#define CTDB_TRAY_OUT_OF_BOUNDS 1.25
CTDB_STATUS
ctdb_ptop_te_compare_initXX(
    CTDB                       *ctdb,
    CTDB_LOS_IN_XYO_AND_DB     *los,
    TERRAIN_ELEMENT_XX         *te,
    CTDB_PTOP_TE_DATA          *ptop,
    int32                     (*vis_func)())/* function pointer to visibility
					     * routine
					     */
{
    TERRAIN_ELEMENT_XX *lte;      /* ptr to the current te data structure */
    int32               side;     /* side of this_te that the valid 
				   * intersection is tied to 
			           */
    uint8               adj_side; /* side of the adjacent te that will be 
				   * evaluated next 
				   */
    float64             best_tray; /* Stored as fraction of dist along ray,0 
				    * to 1
				    */
    float64             best_edge; /* Stored as fraction of dist along edge,
				    * 0 to 1
				    */
    TERRAIN_ELEMENT_XX *best_te1,*best_te2;/* terrain elements that make
					    * up the edge associated with
					    * other best parameters 
					    */
    int32               vtx_flg;
    int32               soil0_valid, soil1_valid;
    int32               horiz;      /* 1=horizontal edge */
    float64             tray, tedge;
    
#ifdef CTDB_INTERVIS_METRICS
    ctdb_data.tes_checked++;
#endif
    /* initialize to a value just outside legal range */    
    best_tray  = CTDB_TRAY_OUT_OF_BOUNDS;
    lte = te + ptop->this_te;
    ptop->tray += 1e-14 ;/* move the starting tray point off of the first
			  * edge.  This is needed because the comparison in 
			  * ANY_INTERSECTION of tray>ptop.tray returns true
			  * when tray=ptop.tray (as view from within the 
			  * debugger to the 15th decimal place). 
			  */
    
    soil0_valid = (lte->info.soil0 != CTDB_ILLEGAL_SOIL_TYPE);
    soil1_valid = (lte->info.soil1 != CTDB_ILLEGAL_SOIL_TYPE);

    /* 1.) Find all Intersections with TE sides */
    if (lte->info.vertices == 3)
    {
	/* quadrilateral */
	/* Quad edge 0 */
	if(soil0_valid)
	  FIRST_TE_CHECK_EDGE(lte,
			      te + lte->vert.index[1],
			      0,
			      1);
	/* Quad edge 1 */
	if(soil0_valid)
	  FIRST_TE_CHECK_EDGE(te + lte->vert.index[1],
			      te + lte->vert.index[0],
			      1,
			      1);
	/* Quad edge 2 */
	if(soil1_valid)
	  FIRST_TE_CHECK_EDGE(te + lte->vert.index[0],
			      te + lte->vert.index[2],
			      2,
			      1);
	/* Quad edge 3 */
	if(soil1_valid)
	  FIRST_TE_CHECK_EDGE(te + lte->vert.index[2],
			      lte,
			      3,
			      1);
	/* Quad edge 4 */
	if(soil0_valid && soil1_valid)
	  CHECK_QUAD_INTERNAL_EDGE
    }
    else
    {
	/* triangle */
	/* Triangle edge 0 */
	if(soil0_valid)
	{
	    FIRST_TE_CHECK_EDGE(lte,
				te + lte->vert.index[0],
				0,
				0);
	    /* Triangle edge 1 */
	    FIRST_TE_CHECK_EDGE(te + lte->vert.index[0],
				te + lte->vert.index[1],
				1,
				0);
	    /* Triangle edge 2 */
	    FIRST_TE_CHECK_EDGE(te + lte->vert.index[1],
				lte,
				2,
				0);
	}
    }

    /* No intersections that were considered valid where handled */
    if ( best_tray == CTDB_TRAY_OUT_OF_BOUNDS )
      return CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING;
    ptop->tray = best_tray;
    
    /* Check this edge to see if it blocks visibility */
    if ( !vis_func(ctdb, los, (float64)best_te1->z,
		   (float64)best_te2->z,best_edge,best_tray) )
      return CTDB_STATUS_INTERVIS_COMPLETE;
    
    /* If closest intersection is a vertex return result of 
       vertex processing */
    if (  vtx_flg )
      return CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING;
    
    /* We went through all that so we could initialize ptop->prev_te and
     * and start the body of the algorithm.  Our "current" te will
     * now be the closest adjacency.  Note that doing this via
     * smart- edge is obviously _much_ faster... */
    ptop->entry_side = adj_side;
    /* check smart edge of next side */
    PROCESS_INTERSECTION( horiz,
			 best_te1->pt.x,
			 best_te1->pt.y,
			 side);
    return CTDB_STATUS_OK;
}

CTDB_STATUS
ctdb_ptop_te_compareXX(
    CTDB                    *ctdb,
    CTDB_LOS_IN_XYO_AND_DB  *los,
    TERRAIN_ELEMENT_XX      *te,
    CTDB_PTOP_TE_DATA       *ptop,
    int32                  (*vis_func)())/* function pointer to visibility
					  * routine 
					  */
{
    TERRAIN_ELEMENT_XX *lte;
    float64             tray,tedge;
    float64             vtx_tray = 0.0;
    int32               on_vertex;
    int32               soil0_valid, soil1_valid;
    
    while ( 1 )
    {
#ifdef CTDB_INTERVIS_METRICS
	ctdb_data.tes_checked++;
#endif
	lte = te + ptop->this_te;

	soil0_valid = (lte->info.soil0 != CTDB_ILLEGAL_SOIL_TYPE);
	soil1_valid = (lte->info.soil1 != CTDB_ILLEGAL_SOIL_TYPE);

	if ( lte->info.vertices == 3 )
	{ /* Quadrilateral */	    
	    if ( ptop->entry_side == 0 )
	    {
		/* Quad edge 1 */
		if(soil0_valid)
		  CHECK_EDGE(te + lte->vert.index[1],
			     te + lte->vert.index[0],
			     1,
			     1);

		if(soil0_valid && soil1_valid)
		  CHECK_QUAD_INTERNAL_EDGE;

		/* Quad edge 2 */
		if(soil1_valid)
		  CHECK_EDGE(te + lte->vert.index[0],
			     te + lte->vert.index[2],
			     2,
			     1);

		/* Quad edge 3 */
		if(soil1_valid)
		  CHECK_EDGE(te + lte->vert.index[2],
			     lte,
			     3,
			     1);
	    }
	    else if (ptop->entry_side == 1)
	    {
		/* Quad edge 0 */
		if(soil0_valid)
		  CHECK_EDGE(lte,
			     te + lte->vert.index[1],
			     0,
			     1);

		if(soil0_valid && soil1_valid)
		  CHECK_QUAD_INTERNAL_EDGE;

		/* Quad edge 2 */
		if(soil1_valid)
		  CHECK_EDGE(te + lte->vert.index[0],
			     te + lte->vert.index[2],
			     2,
			     1);

		/* Quad edge 3 */
		if(soil1_valid)
		  CHECK_EDGE(te + lte->vert.index[2],
			     lte,
			     3,
			     1);
	    }
	    else if (ptop->entry_side == 2)
	    {
		/* Quad edge 3 */
		if(soil1_valid)
		  CHECK_EDGE(te + lte->vert.index[2],
			     lte,
			     3,
			     1);

		if(soil0_valid && soil1_valid)
		  CHECK_QUAD_INTERNAL_EDGE;

		/* Quad edge 0 */
		if(soil0_valid)
		  CHECK_EDGE(lte,
			     te + lte->vert.index[1],
			     0,
			     1);

		/* Quad edge 1 */
		if(soil0_valid)
		  CHECK_EDGE(te + lte->vert.index[1],
			     te + lte->vert.index[0],
			     1,
			     1);
	    }
	    else
	    {
		/* Quad edge 2 */
		if(soil1_valid)
		  CHECK_EDGE(te + lte->vert.index[0],
			     te + lte->vert.index[2],
			     2,
			     1);

		if(soil0_valid && soil1_valid)
		  CHECK_QUAD_INTERNAL_EDGE;

		/* Quad edge 0 */
		if(soil0_valid)
		  CHECK_EDGE(lte,
			     te + lte->vert.index[1],
			     0,
			     1);

		/* Quad edge 1 */
		if(soil0_valid)
		  CHECK_EDGE(te + lte->vert.index[1],
			     te + lte->vert.index[0],
			     1,
			     1);
	    }
	}    
	else
	{
	    /* Triangle */
	    /* Triangle edge 0 */
	    if (ptop->entry_side != 0 && soil0_valid)
	    {
		CHECK_EDGE(lte,
			   te+lte->vert.index[0],
			   0,
			   0)
	    }

	    /* Triangle edge 1 */
	    if (ptop->entry_side != 1 && soil0_valid)
	    {
		CHECK_EDGE(te+lte->vert.index[0],
			   te+lte->vert.index[1],
			   1,
			   0)
	    }
	    /* Triangle edge 2 */
	    if (ptop->entry_side != 2 && soil0_valid)
	    {
		CHECK_EDGE(te+lte->vert.index[1],
			   lte,
			   2,
			   0)
	    }
	}
	/* If we fall through to here, we are either done or we've hit a
	 * a vertex.  Deal with both.
	 */
	if ( on_vertex )
	{
	    ptop->tray = NS_MAX(vtx_tray,ptop->tray);
	    return CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING;
	}
	
	return CTDB_STATUS_NO_VALID_INTERSECTIONS;
    }
#ifdef lint
    return CTDB_STATUS_OK;
#endif
}


/* This routine will return the next edge to be evaluated given the side/
 * edge last evaluated, the index of the terrain element and the list of 
 * terrain elements.
 */
CTDB_STATUS 
ctdb_get_te_edge_indexXX(
    TERRAIN_ELEMENT_XX  *te,
    uint32               index ,/*vertex number */
    int32                side  ,/*side of interest*/
    int32 	         edge[])
{
    int32 v;
    int32 geometry = ((int32)te[index].info.vertices - 2);
    
    /* TRIANGLE traversal : V -> V0 ; V0 -> V1 ; V1 -> V          
     * POLYGON  traversal : V -> V1 ; V1 -> V0 ; V0 -> V2 ; V2 -> V ; V -> V1
     */
    static int32 next[2][CTDB_MAX_EDGES_PER_TE][2]=
    { {{-1,0},{0,1},{1,-1},{-1,-1},{-1,-1}},
      {{-1,1},{1,0},{0,2},{2,-1},{-1,0}} };
    
    if (geometry < 0)
      return(CTDB_STATUS_NOT_A_TERRAIN_ELEMENT);              
    
    v = next[geometry][side][0];
    if ( v < 0 ) 
      edge[0] = index;
    else
      edge[0] = te[index].vert.index[v];
    
    v = next[geometry][side][1];
    
    if ( v < 0 ) 
      edge[1] = index;
    else
      edge[1] = te[index].vert.index[v];
    
    return(CTDB_STATUS_OK);              
}

/* This routine will return TRUE if the index is valid and has not already
 * been inspected 
 */
static int32
validate_te_indexXX(
    uint32  te_index ,  /* te index to validate */
    uint32 *front    ,  /* address of list to be validated against */
    uint32 *back    )   /* end address of the list (count not address) */
{
    if ( te_index > MAX_ALLOWED_TE_INDEX_XX )
      return FALSE;
    
    while ( (back--) > front )
    {	
	if ( *back == te_index ) /* te already processed */
	  return FALSE; 
    }
    return TRUE;
}

/* Computes a normal to the microterrain polygon.  Note that the
 * units of *a and *b are the funny units used to store features,
 * NOT meters or posts.
 */
static float64
compute_te_normalXX(ELEVATION_LOOKUP_STORAGE *els)
{
    /* patch's TE base address */
    TERRAIN_ELEMENT_XX *base_te = (TERRAIN_ELEMENT_XX *)els->base; 
    /* TE which contains the point */
    TERRAIN_ELEMENT_XX *te      = (TERRAIN_ELEMENT_XX *)els->te;   
    TERRAIN_ELEMENT_XX *assoc_te;
    float64             magic, normal[XYZ];
    CTDB_COORD3D        v1, v2;
    
    if (!els->normal)
      els->normal = normal;
    
    assoc_te = base_te + te->vert.index[0];
    
    v1.x = (int32)assoc_te->pt.x - (int32)te->pt.x ;
    v1.y = (int32)assoc_te->pt.y - (int32)te->pt.y ;
    v1.z = assoc_te->z - te->z ;
    
    assoc_te = base_te + te->vert.index[els->where-TRIANGLE0+1] ;
    
    v2.x = (int32)assoc_te->pt.x - (int32)te->pt.x;
    v2.y = (int32)assoc_te->pt.y - (int32)te->pt.y;
    v2.z = assoc_te->z - te->z;
    
    VMAT3E_CROSS_PROD(v1.x, v1.y, v1.z, v2.x, v2.y, v2.z, els->normal);
    
    /* The magic number = a*x + b*y + c*z, where <x y z> is a vertex
     * of the polygon.  This can be used to compute the elevation on the
     * the polygon's plane at a particular <x y>, or the intersection of a
     * line with the polygon's plane.
     */
    
    magic = ((int32)te->pt.x * els->normal[X] + 
	     (int32)te->pt.y * els->normal[Y] + 
	     te->z * els->normal[Z]);
    
    if ( els->normal[Z] == 0.0 )
      return te->z ;
    else
      return (magic-els->normal[X]*(float64)els->npt.x-els->normal[Y]*
	      (float64)els->npt.y)/els->normal[Z] ;
}

#define TE_VERT(a) base[te->vert.index[(a)]]

static float64
compute_elevation_on_edgeXX(
    TERRAIN_ELEMENT_XX   *base, /* patch's TE base address */
    TERRAIN_ELEMENT_XX   *te,      /* TE which contains the point */
    CTDB_COORD2D_I32     *pt,      /* Point in TE coordinates */
    CTDB_POINT_STATE      edge)  
{
    CTDB_XYO_COORD2D start, stop;
    float64          dy;    
    float32          z, dz;
    
    if ( te->info.vertices == 2 )
    { /* triangle */
	if ( edge == COLINEAR0 ) 
	{ /* 0 -> 1 */
	    start = te->pt;
	    stop  = TE_VERT(0).pt;	    
	    z     = te->z;
	    dz    = TE_VERT(0).z - z;
	}
	else if ( edge == COLINEAR1 )
	{ /* 1 -> 2 */
	    start = TE_VERT(0).pt;
	    stop  = TE_VERT(1).pt;
	    z     = TE_VERT(0).z; 	    
	    dz    = TE_VERT(1).z - z;
	}
	else 
	{ /* 2 -> 0 */
	    start = TE_VERT(1).pt;
	    stop  = te->pt;
	    z     = TE_VERT(1).z; 	    
	    dz    = te->z - z;
	}
    }
    else
    { /* quadrilateral */
	if ( edge == COLINEAR0 )
	{ /* 0 -> 2 */
	    start = te->pt;
	    stop  = TE_VERT(1).pt;
	    z     = te->z; 	    
	    dz    = TE_VERT(1).z - z;
	}
	else if ( edge == COLINEAR1 )
	{ /* 2 -> 1 */
	    start = TE_VERT(1).pt;
	    stop  = TE_VERT(0).pt;
	    z     = TE_VERT(1).z; 	    
	    dz    = TE_VERT(0).z - z;
	}
	else if (edge == COLINEAR2 )
	{ /* 1 -> 3 */
	    start = TE_VERT(0).pt;
	    stop  = TE_VERT(2).pt;
	    z     = TE_VERT(0).z;
	    dz    = TE_VERT(2).z - z;
	}
	else 
	{ /* 3 -> 0 */
	    start = TE_VERT(2).pt;
	    stop  = te->pt;
	    z     = TE_VERT(2).z;
	    dz    = te->z - z;
	}
    }
    
    if ((dy = ((float64)stop.y - (float64)start.y)) != 0.0)
      return((float64)(z + dz*(((float64)pt->y -
				(float64)start.y)/dy)));
    else
    {	
	/* Use dx instead for horizontal edges */
	dy = (float64)stop.x - (float64)start.x;
	if (dy == 0.0)
	  return (float64)z;
	else
	  return((float64)(z + dz*(((float64)pt->x -
				    (float64)start.x)/dy)));
    }
} 

static void
set_normal_vectorXX(CTDB                 *ctdb,
		    TERRAIN_ELEMENT_XX   *base,
		    TERRAIN_ELEMENT_XX   *te,
		    CTDB_POINT_STATE      which,
		    float64              *normal)
{
    float64 v0[XYZ], v1[XYZ];
    
    if (!normal)
      return;

    v0[X] = TE_VERT(0).pt.x - te->pt.x;
    v0[Y] = TE_VERT(0).pt.y - te->pt.y;
    v0[Z] = TE_VERT(0).z    - te->z;
    
    if (which == TRIANGLE0)
    {
	v1[X] = TE_VERT(1).pt.x - te->pt.x;
	v1[Y] = TE_VERT(1).pt.y - te->pt.y;
	v1[Z] = TE_VERT(1).z    - te->z;
    }
    else
    {
	v1[X] = TE_VERT(2).pt.x - te->pt.x;
	v1[Y] = TE_VERT(2).pt.y - te->pt.y;
	v1[Z] = TE_VERT(2).z    - te->z;
    }
    vmat3_cross_prod64(v0, v1, normal);
}

static void
set_elevation_and_soilXX(ELEVATION_LOOKUP_STORAGE *els)
{
    switch (els->where)
    {
      case TRIANGLE0 :
	els->elevation = compute_te_normalXX(els) ;
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil0;
	break;
      case TRIANGLE1 :
	els->elevation = compute_te_normalXX(els) ;
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil1;
	break;
      case VERTEX0 :
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil0;
	els->elevation = ((TERRAIN_ELEMENT_XX *)els->te)->z;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE0, 
			    els->normal);
	break;
      case VERTEX1 :
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil0;
	els->elevation = 
	  ((TERRAIN_ELEMENT_XX *)els->base)[((TERRAIN_ELEMENT_XX *)els->te)->
					    vert.index[els->where-VERTEX1]].z;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE0, 
			    els->normal);
	break;
      case VERTEX2 :
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil0;
	els->elevation = 
	  ((TERRAIN_ELEMENT_XX *)els->base)[((TERRAIN_ELEMENT_XX *)els->te)->
					    vert.index[els->where-VERTEX1]].z;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE0, 
			    els->normal);
	break;
      case VERTEX3 :
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil1;
	els->elevation = 
	  ((TERRAIN_ELEMENT_XX *)els->base)[((TERRAIN_ELEMENT_XX *)els->te)->
					    vert.index[2]].z;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE1, 
			    els->normal);
	break;
      case COLINEAR0 :
	els->elevation=
	  compute_elevation_on_edgeXX((TERRAIN_ELEMENT_XX *)els->base,
				      (TERRAIN_ELEMENT_XX *)els->te,
				      &(els->npt),
				      els->where);
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil0;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE0, 
			    els->normal);
	break;
      case COLINEAR1 :
	els->elevation=
	  compute_elevation_on_edgeXX((TERRAIN_ELEMENT_XX *)els->base,
				      (TERRAIN_ELEMENT_XX *)els->te,
				      &(els->npt),
				      els->where);
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil0;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE0, 
			    els->normal);
	break;
      case COLINEAR2 :
	els->elevation=
	  compute_elevation_on_edgeXX((TERRAIN_ELEMENT_XX *)els->base,
				      (TERRAIN_ELEMENT_XX *)els->te,
				      &(els->npt),
				      els->where);
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil1;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE1, 
			    els->normal);
	break;
      case COLINEAR3 :
	els->elevation=
	  compute_elevation_on_edgeXX((TERRAIN_ELEMENT_XX *)els->base,
				      (TERRAIN_ELEMENT_XX *)els->te,
				      &(els->npt),
				      els->where);
	els->soil = ((TERRAIN_ELEMENT_XX *)els->te)->info.soil1;
	set_normal_vectorXX(els->ctdb,els->base, els->te, TRIANGLE1, 
			    els->normal);
	break;
      default:
	break;
    }

    if(els->soil == CTDB_ILLEGAL_SOIL_TYPE)
      els->elevation = 0.0;

    if (els->normal)
    {
	els->normal[Z] *= els->ctdb->incrD * els->ctdb->patch_incrD *
	  CTDB_XYO_TO_PATCH_SCALE;

	vmat3_unit64(els->normal,els->normal);
	/* All terrain surfaces are considered to be positive.  Enforce this
	 * assumption instead ensuring that the vertices are in the proper
	 * order.
	 */
	if (els->normal[Z] < 0.0)
	  vmat3_negate64(els->normal,els->normal);
    }
}

/* point_inside_te
 *
 * Note:  This routine only changes the data parameter where only if the point
 * is colinear.  Any other setting of this parameter will affect callers of 
 * this routine since they do not check for a change in the state of where 
 * after each call to this routine.
 */
static int32	
point_inside_teXX(
    CTDB_COORD2D_I32         *a, /* vector a from pt(x,y) to a(x,y) */
    CTDB_COORD2D_I32         *b, /* vector a from pt(x,y) to b(x,y) */
    int32                     side, /* side define by the two TE vertices 
				     * used in a,b 
				     */
    ELEVATION_LOOKUP_STORAGE *els)   
{
    uint32 index;		/* potential preferred candidate te */
    int32  cross=VMAT2E_CROSS_PROD(a->x,a->y,b->x,b->y);
    
    index = (uint32)(((TERRAIN_ELEMENT_XX *)(els->te))->adj[side]);
    
    if ( cross < 0 ) /* Point inside half plane */	
      return(TRUE);
    
    if ( cross > 0 ) /* Point outside half plane */
    {
	if ( validate_te_indexXX(index,checked_f,checked_b) )
	  els->te_queue[els->back++] = index;  /* add to queue */
	return FALSE;
    }
    
    if ( VMAT2E_DOT_PROD(a->x,a->y,b->x,b->y) < 0 )
    { /* point on the edge */
	/* if -1 then the vectors are opposite in direction and the point 
	 * will lie between the two end points.  If the result is +1 then 
	 * the point is outside the segment since the vectors are in the 
	 * same direction
	 */
	els->where=COLINEAR0+side ;
    }
    
    return(FALSE);
}

static void
point_in_terrain_elementXX(ELEVATION_LOOKUP_STORAGE *els)
{
    /* COORDINATE SYSTEM : TE COORDINATES */
    /* patch's TE base address */
    TERRAIN_ELEMENT_XX   *base_te= (TERRAIN_ELEMENT_XX *)els->base; 
    TERRAIN_ELEMENT_XX   *vert_te= base_te + els->te_index ;
    int32 	          i;
    int32 	          bound  = vert_te->info.vertices + 1;
    CTDB_COORD2D_I32      P[4];
    int32                 e1,e2,e3,e4;
    
    /* Determine if the point is inside the triangle.  To do this, we
     * draw a line from the point to each vertex (V0, V1, V2).
     * If we are inside the triangle, going from V0 to V1 to V2 will
     * either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     * (A side effect of finding each vector is that we can test
     * for a direct hit of our point on a polygon vertex, so we
     * exploit it.)
     *
     * When one of the cross products produces a 0, we test dot products
     * to determine if the point is on the line connecting the vertices.
     *
     * The data may describe 0, 1, or 2 microterrain triangles.
     *
     * NOTE <xu yu> is assumed to be in the funny units used to store
     * features, NOT meters or posts.
     */
    if ( bound < 3 )
      return;

    els->te = (ADDRESS)vert_te ;
    
    for (i=0;i<bound;i++)
    {
	/* P0 is the vector from our point to vertex 0 */
	/* P1 is the vector from our point to vertex 1 */
	/* P2 is the vector from our point to vertex 2 */
	/* P3 is the vector from our point to vertex 3 */
	
	P[i].x = vert_te->pt.x - els->npt.x;  
	P[i].y = vert_te->pt.y - els->npt.y ;
	if (!P[i].x  && !P[i].y)
	{
	    els->where = VERTEX0+i;
	    return;
	}
	vert_te = base_te + ((TERRAIN_ELEMENT_XX *)els->te)->vert.index[i] ;
    }
    
    if ( bound == 3 )
    {
	e1 = point_inside_teXX(&P[0],&P[1],(int32)0,els);
	e2 = point_inside_teXX(&P[1],&P[2],(int32)1,els);
	e3 = point_inside_teXX(&P[2],&P[0],(int32)2,els);
	
	if ( e1 && e2 && e3 )
	  els->where  = TRIANGLE0;
    }
    else
    {
	e1 = point_inside_teXX(&P[0],&P[2],(int32)0,els);
	e2 = point_inside_teXX(&P[2],&P[1],(int32)1,els);
	e3 = point_inside_teXX(&P[1],&P[3],(int32)2,els);
	e4 = point_inside_teXX(&P[3],&P[0],(int32)3,els);
	
	if ( VMAT2E_CROSS_PROD(P[0].x,P[0].y,P[1].x,P[1].y) > 0 )
	{
	    if ( e1 && e2) 
	      els->where = TRIANGLE0;
	}
	else
	{
	    if ( e3 && e4 )
	      els->where = TRIANGLE1 ;
	}
    }
    
    return;
}

static CTDB_STATUS
look_at_adjacent_trianglesXX(ELEVATION_LOOKUP_STORAGE *els)
{
    *checked_b++ = els->te_index ;/* Add TE to the list of checked TEs */
    
    do
    {
	/* Is the TO BE SEARCH queue empty, if then resort to a brute force
	 * search of all remaining terrain elements 
	 */
	if ( els->front == els->back)
	{ 
#ifdef CTDB_ELEV_DEBUG
	    printf("\nFailed directed search - going to Linear search\n");
#endif
	    return CTDB_STATUS_FAILED;
	}
	/* Take next TE to look at from the list */
	els->te_index = els->te_queue[ els->front++ ];
    } while ( !validate_te_indexXX(els->te_index,checked_f,checked_b) );  
    els->te = (ADDRESS)(((TERRAIN_ELEMENT_XX *)els->base) + els->te_index);
    return CTDB_STATUS_OK;
}


static CTDB_STATUS
look_at_all_trianglesXX(ELEVATION_LOOKUP_STORAGE *els)
{
    int32 found=FALSE;
    TERRAIN_ELEMENT_XX *lte;
    int32 n_te;
    
    if (!els->pg->mapped)
      ctdb_fetch_patch_group(els->ctdb, els->pg);
    else
      els->ctdb->patch_hit_count++;
    
    n_te = CTDB_NUM_TE(els->pg, els->member);
    
#ifdef CTDB_ELEV_DEBUG
    printf("linear search info:\n");
    printf("Patch %d, normalized point %d %d\n", 
	   els->num, els->npt.x, els->npt.y);
    printf("                                ");
#endif
    
    els->te=els->base;
    lte = (TERRAIN_ELEMENT_XX *)els->base;
    for (els->te_index=0; els->te_index < n_te; els->te_index++, lte++)
    {
#ifdef CTDB_ELEV_DEBUG
	count++;
#endif
	els->te = (ADDRESS)lte;
	point_in_terrain_elementXX(els);
	
	if (els->where != OUTSIDE)
        {
#ifdef CTDB_ELEV_DEBUG
	    printf("post count %d ", count);
#endif
	    set_elevation_and_soilXX(els);
	    /* if not inside then on an edge and elevation is already
	       set by point_in_terrain_element */
#ifdef CTDB_ELEV_DEBUG
	    printf("Elevation found\n"); 
#endif
	    return CTDB_STATUS_OK;
	}
    }
#ifdef CTDB_ELEV_DEBUG
    printf("NO Elevation found count %d\n",i);
#endif
    return CTDB_STATUS_NO_ELEVATION_FOUND ;
}

CTDB_STATUS
ctdb_lookup_terrain_element_elevationXX(ELEVATION_LOOKUP_STORAGE    *els)
{   
    checked_b = checked_f = &checked_array[0];/* Global list of TEs checked */
#ifdef CTDB_ELEV_DEBUG
    int32 count = 0;
    printf("\n Our point %d %d\n", els->npt->x, els->npt->y);
#endif    
    while (1)
    {
#ifdef CTDB_ELEV_DEBUG
	printf("Considering te %d\n",els->te_index);
	count++;
#endif
	point_in_terrain_elementXX(els);
	
	if (els->where != OUTSIDE)
	{
#ifdef CTDB_ELEV_DEBUG
	    printf("count %d ", count);
#endif
	    set_elevation_and_soilXX(els);
	    /* Change the virtual grid mapping to point to this TE.  This
	     * will improve processing time for lookups since chances are 
	     * that more hits to the TE will be performed and thin cut and
	     * fill roads will start to get mapped instead of the large polys
	     * that are to both sides of the road that will probably not 
	     * be used.
	     */
	    *(els->vg_ptr) = (uint16)els->te_index;
	    return CTDB_STATUS_OK;
	}
#ifdef CTDB_ELEV_DEBUG
	{
	    int32 i;	    
	    printf("QUEUE ");
	    for (i=els->front;i<els->back;i++)
	      printf("%d ",els->te_queue[i]);
	    printf("\n");
	}	
#endif	 
	
	if ( look_at_adjacent_trianglesXX(els) != CTDB_STATUS_OK )
        { 
#ifdef CTDB_ELEV_DEBUG 
	    printf("prior count %d \n", count);
#endif
	    if ( look_at_all_trianglesXX(els) != CTDB_STATUS_OK )
	    {
#ifdef CTDB_ELEV_DEBUG 
		printf("Missing TE - using vgrid default te %d patch %d\n", 
		       els->mapped_te, els->num);       
#endif
		/* No Elevation found so a TE must be missing: approximate
		 * the elevation by returning the elevation for the TE
		 * associated with the Vgrid for the given point 
		 */
		els->where = TRIANGLE0;
		els->te_index = els->mapped_te;
		els->te = (TERRAIN_ELEMENT_XX *) els->base +
		  els->te_index;
		set_elevation_and_soilXX(els);
		
		return CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT;
	    }
	}     
    }
#ifdef lint
    return CTDB_STATUS_OK;
#endif
}

void
ctdb_print_te_elementXX(TERRAIN_ELEMENT_XX *te, uint32 index)
{
    int32        i;
    int32        edge[2];
    CTDB_STATUS  no_edge;
    
    printf("%d\t%d\t%d\t%d\t",
	   index,te[index].pt.x,te[index].pt.y,(int32)te[index].z);
    
    for (i=0;i< (int32)te[index].info.vertices;i++)
      printf("%6d ",te[index].vert.index[i]);
    
    for (i=te[index].info.vertices;i<TE_VERT_DIMENSION;i++)
      printf("      ");
    
    printf("\t");
    
    for (i=0;i<TE_ADJ_TRI_MAP_DIMENSION;i++)
    {
	no_edge = ctdb_get_te_edge_indexXX(te,index,i,edge) ;
	
	if (te[index].adj[i] == DB_BOUNDRY_XX)
	  printf("%8s ","   BND  ");
	else if (te[index].adj[i] == EDGE_NOT_MAPPED_XX)
	  printf("%8s ","   ENM  ");
	else 
	{
	    printf("%6d",te[index].adj[i]);
	    
	    if ( te[index].info.patch_adj & (0x1<<i) && !no_edge )
	    {
		if ( te[ edge[0] ].pt.x  == te[ edge[1] ].pt.x )
		{
		    if ( !te[ edge[0] ].pt.x )
		      printf("%1c",'w');
		    else
		      printf("%1c",'e');
		}
		else
		{
		    if ( !te[ edge[0] ].pt.y )
		      printf("%1c",'s');
		    else
		      printf("%1c",'n');
		}
	    }
	    else
	      printf("%c",' ');	
	    printf(" %1d", ((te[index].info.adj_sides >> (2 * i)) & 0x3));
	} 
	printf("%c",' ');	
    }
    printf("\t%3d\t%3d\n",te[index].info.soil0,te[index].info.soil1);    
}

int32
ctdb_terrain_element_densityXX(CTDB *ctdb, int32 patch_number)
{
    int32               count=0;
    TERRAIN_ELEMENT_XX *te;
    int32               n_te;
    int32               i;
    int32               member      =  PG_MEMBER(patch_number);
    CTDB_PATCH_GROUP   *patch_group =  ((CTDB_PATCH_GROUP *)
					ctdb->patch_groups + 
					PG_NUMBER(patch_number));
    
    te = (TERRAIN_ELEMENT_XX *)ctdb_lookup_te_ptr(ctdb,patch_number);
    
    if (te)
    {
	/* We can't safely call CTDB_NUM_TE unless we know the PG
	 * data is in memory.  Fortunately, the call to lookup_
	 * terrain_element above guarantees that.
	 */
	n_te = CTDB_NUM_TE(patch_group, member);
	for(i=0;i<n_te;i++)
	{
	    if ( te[i].info.vertices == 2 )
	      count++;
	    else if (  te[i].info.vertices == 3 )
	      count+=2  ;
	}
    }   
    return(count);
}

/* This function changes the soil type of all terrain element triangles
 * that share all three of its vertices with the specified input
 * polygon (which may have more than three vertices).  It has to search
 * through the terrain elements.  The index and information about which
 * polygon is changed is returned so that the next time the same polygon
 * is changed, we won't have to do another search from scratch.
 */
#define ACCEPTABLE_ERROR   0.5
void
ctdb_change_terrain_element_soilXX(CTDB                  *ctdb,
				   int32                  patch_number,
				   CTDB_TS_POLYGON       *poly,
				   int32                  soil,
				   int32                 *num_tes,
				   int32                  which[],
				   uint32                 te_index[])
{
    int32 i, j, ii, jj, n_te;
    TERRAIN_ELEMENT_XX   *te;
    float32 v[3][XY], delta_x, delta_y;
    int32 px=(patch_number%ctdb->patches_wide);
    int32 py=(patch_number/ctdb->patches_wide);
    CTDB_PATCH_GROUP *pg=(ctdb->patch_groups+PG_NUMBER(patch_number));
    int32 member=PG_MEMBER(patch_number);
    int32 max_tes = *num_tes;

    *num_tes = 0;

    te = (TERRAIN_ELEMENT_XX *)ctdb_lookup_te_ptr(ctdb, patch_number);
    
    n_te = CTDB_NUM_TE(pg, member);

    for(i = 0; i < n_te; i++)
    {
	int32 n_vert = te[i].info.vertices;

	/* if this TE index defines a line or point then there is no triangle.
	 */
	if ( n_vert < 2 )
	  continue;

	/* Save the anchor vertex which is always defined */
	v[0][X] = CTDB_XYO_TO_METERS(te[i].pt.x,px,ctdb->patch_incrD,
				     ctdb->incrD) ;
	v[0][Y] = CTDB_XYO_TO_METERS(te[i].pt.y,py,ctdb->patch_incrD,
				     ctdb->incrD) ;
    
	j = te[i].vert.index[1];

	v[1][X] = CTDB_XYO_TO_METERS(te[j].pt.x,px,ctdb->patch_incrD,
				     ctdb->incrD);
	v[1][Y] = CTDB_XYO_TO_METERS(te[j].pt.y,py,ctdb->patch_incrD,
				     ctdb->incrD);
    
	j = te[i].vert.index[0];

	v[2][X] = CTDB_XYO_TO_METERS(te[j].pt.x,px,ctdb->patch_incrD,
				     ctdb->incrD);
	v[2][Y] = CTDB_XYO_TO_METERS(te[j].pt.y,py,ctdb->patch_incrD,
				     ctdb->incrD);

	/* Check to see if all three vertices of this triangle are
	 * included in the input polygon.
	 */
	for(ii = 0; ii < 3; ii++)
	{
	    for(jj = 0; jj < poly->n_verts; jj++)
	    {
		delta_x = v[ii][X] - poly->verts[jj][X];
		delta_y = v[ii][Y] - poly->verts[jj][Y];

		if(delta_x > -ACCEPTABLE_ERROR && delta_x < ACCEPTABLE_ERROR &&
		   delta_y > -ACCEPTABLE_ERROR && delta_y < ACCEPTABLE_ERROR)
		  break;
	    }

	    /* If this point is not part of the input polygon, then skip it.
	     */
	    if(jj == poly->n_verts)
	      break;
	}

	if(ii == 3)
	{
	    te[i].info.soil0 = soil;

	    which[*num_tes] = 0;
	    te_index[*num_tes] = i;
	    if(++(*num_tes) == max_tes)
	    {
		printf("ct_8or16.c: max_tes exceeded\n");
		return;
	    }
	}

	if ( n_vert != 3 )
	  continue;
    
	/* Get the next triangle.
	 */
	v[1][X] = v[2][X];
	v[1][Y] = v[2][Y];

	j = te[i].vert.index[2];

	v[2][X] = CTDB_XYO_TO_METERS(te[j].pt.x,px,ctdb->patch_incrD,
				     ctdb->incrD);
	v[2][Y] = CTDB_XYO_TO_METERS(te[j].pt.y,py,ctdb->patch_incrD,
				     ctdb->incrD);
    
	/* Check to see if all three vertices of this triangle are
	 * included in the input polygon.
	 */
	for(ii = 0; ii < 3; ii++)
	{
	    for(jj = 0; jj < poly->n_verts; jj++)
	    {
		delta_x = v[ii][X] - poly->verts[jj][X];
		delta_y = v[ii][Y] - poly->verts[jj][Y];

		if(delta_x > -ACCEPTABLE_ERROR && delta_x < ACCEPTABLE_ERROR &&
		   delta_y > -ACCEPTABLE_ERROR && delta_y < ACCEPTABLE_ERROR)
		  break;
	    }

	    /* If this point is not part of the input polygon, then skip it.
	     */
	    if(jj == poly->n_verts)
	      break;
	}

	if(ii == 3)
	{
	    te[i].info.soil1 = soil;

	    which[*num_tes] = 1;
	    te_index[*num_tes] = i;
	    if(++(*num_tes) == max_tes)
	    {
		printf("ct_8or16.c: max_tes exceeded\n");
		return;
	    }
	}
    }
}

/* This function updates the soil type of the specified TE.  The TE is
 * completely specified by the arguments, so no search is required to
 * find the right one.  Call this only if you know the index and which
 * TE you're changing.
 */
void
ctdb_update_terrain_element_soilXX(CTDB   *ctdb,
				   int32   patch_number,
				   uint32  index,
				   int32   which,
				   int32   soil)
{
    TERRAIN_ELEMENT_XX *te;

    te = (TERRAIN_ELEMENT_XX *)ctdb_lookup_te_ptr(ctdb, patch_number);

    switch(which)
    {
      case 0:
	(te + index)->info.soil0 = soil;
	break;

      case 1:
	(te + index)->info.soil1 = soil;
	break;

      default:
	break;
    }
}

static void 
find_te_z_boundsXX(
    TERRAIN_ELEMENT_XX   *base_te,
    TERRAIN_ELEMENT_XX   *te,
    float64              *min_z,
    float64              *max_z)
{
    float64 z;
    
    *max_z = *min_z = te->z;
    
    if (te->info.vertices> 2)
    {
	z = base_te[te->vert.index[2]].z;
	if (z < *min_z) *min_z = z;
	else if (z > *max_z) *max_z = z;
    }
    if (te->info.vertices > 1)
    {
	z = base_te[te->vert.index[1]].z;
	if (z < *min_z) *min_z = z;
	else if (z > *max_z) *max_z = z;
    }
    if (te->info.vertices)
    {
	z = base_te[te->vert.index[0]].z;
	if (z < *min_z) *min_z = z;
	else if (z > *max_z) *max_z = z;
    }
}

/* Go through all TEs in member'th patch of pg and draw contour lines as
 * specified.
 */
void
ctdb_find_te_contoursXX(
    CTDB               *ctdb,
    CTDB               *src_ctdb,
    int32               patch_num,
    float64             fudge_factor,
    CTDB_CONTOUR_STATE *state)
{
    CTDB_PATCH_GROUP *pg=(ctdb->patch_groups+PG_NUMBER(patch_num));
    int32 member=PG_MEMBER(patch_num);
    int32 px=(patch_num%ctdb->patches_wide);
    int32 py=(patch_num/ctdb->patches_wide);
    int32 i;
    TERRAIN_ELEMENT_XX *base_te, *v, *v0, *v1, *v2;
    float64 x, y, z, min_z, max_z;
    int32 found;
    int32 nte;
    int32 local = (ctdb == src_ctdb);
    int32 src_cell = ctdb_get_cell_id(src_ctdb);
    float64 src_x, src_y, src_z;
    int32 incr = ctdb->incrD;
    
#define STORE(x,y,z)						       \
    {								       \
	if (!local)                                                    \
	{                                                              \
	    gcs_gcs2e_to_gcscs(ctdb_get_cell_id(ctdb),               \
				 (float64) (x)*incr,                   \
				 (float64) (y)*incr,                   \
				 src_cell,                             \
				 &src_x, &src_y);                      \
	}                                                              \
	else                                                           \
	{                                                              \
	    src_x = (float64) (x)*incr;                                \
	    src_y = (float64) (y)*incr;                                \
	}						               \
	src_z = (float64) (z);                                         \
	if (!(state->seg_cnt & 0x2))				       \
	  state->elevs[state->seg_cnt>>2] = src_z;		       \
	state->segs[state->seg_cnt] = (src_x);		               \
	state->segs[state->seg_cnt+1] = (src_y);		       \
	state->seg_cnt += 2;					       \
	if (state->seg_cnt >= CTDB_MAX_CONTOUR_PTS-4)		       \
	{							       \
	    state->draw_fcn(state->seg_cnt, state->segs, state->elevs);\
	    state->seg_cnt = 0;					       \
	}							       \
    }
    
      base_te = (TERRAIN_ELEMENT_XX*)ctdb_lookup_te_ptr(ctdb, patch_num);
    if (!base_te)
      return;
    
    nte = CTDB_NUM_TE(pg, member);
    
    for(i=0,v=base_te;i<nte;i++,v++)
    {
	if (v->info.vertices < 2)
	  continue;
	find_te_z_boundsXX(base_te, v, &min_z, &max_z);
	
	v0 = base_te + v->vert.index[0];
	v1 = base_te + v->vert.index[1];
	if (v->info.vertices == 3)
	  v2 = base_te + v->vert.index[2];
	
	if ((max_z - min_z) < state->max_delta)
	  for (z = state->interval * floor(max_z * state->inv_interval) + 
	       fudge_factor;
	       z >= min_z; z -= state->interval)
	  {
	      if (ctdb_find_z_on_line(v->pt,v->z,v1->pt,v1->z,
				      px, py, ctdb->patch_incrD,
				      z, &x, &y))
		STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
	      
	      if (ctdb_find_z_on_line(v1->pt,v1->z,v0->pt,v0->z,
				      px, py, ctdb->patch_incrD,
				      z, &x, &y))
		STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
	      
	      /* This is the shared edge, and _must_ be the last one
	       * we check for this to work on quads
	       */
	      if (found = ctdb_find_z_on_line(v0->pt,v0->z,v->pt,v->z,
					      px, py,
					      ctdb->patch_incrD,
					      z, &x, &y))
		STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
	      
	      /* Back up if we only found one point */
	      if (state->seg_cnt & 0x2)
		state->seg_cnt &= ~0x2;
	      
	      if (v->info.vertices < 3)
		continue;
	      
	      /* Store the intersection point again */
	      if (found)
		STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
	      
	      
	      if (ctdb_find_z_on_line(v0->pt,v0->z,v2->pt,v2->z,
				      px, py, ctdb->patch_incrD,
				      z, &x, &y))
		STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
	      
	      if (ctdb_find_z_on_line(v2->pt,v2->z,v->pt,v->z,
				      px, py, ctdb->patch_incrD,
				      z, &x, &y))
		STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
	      
	      /* Back up if we only found one point */
	      if (state->seg_cnt & 0x2)
		state->seg_cnt &= ~0x2;
	  }
    }
}

/* Return, in *min_z and *max_z, the minimum and maximum z values for
 * TE's in the patch defined by pg and member.
 */
void
ctdb_find_te_z_bounds_in_patchXX(
    CTDB             *ctdb,
    int32             patch_num,
    float64          *min_z,
    float64          *max_z)
{
    TERRAIN_ELEMENT_XX *te;
    int32 i;
    float64 lmin, lmax;
    CTDB_PATCH_GROUP *pg=(ctdb->patch_groups+PG_NUMBER(patch_num));
    int32 member=PG_MEMBER(patch_num);
    int32 nte;
    
    lmax = -1.0E+9;
    lmin = 1.0E+9;
    
    te =  (TERRAIN_ELEMENT_XX *)ctdb_lookup_te_ptr(ctdb, patch_num);
    
    if (te)
    {
	nte = CTDB_NUM_TE(pg, member);
	for(i=0;i<nte;i++)
	{
	    if (te[i].z < lmin)
	      lmin = te[i].z;
	    if (te[i].z > lmax)
	      lmax = te[i].z;
	}
    }
    *min_z = lmin;
    *max_z = lmax;
    return;
}

/* Set up patch map for patch specified by patch_num (and for its neighbors
 * if cross_patch_bdy is TRUE.  Returns total number of triangles in this
 * patch.
 */
int32
ctdb_get_patch_maps_XX(
    CTDB             *ctdb,
    int32             patch_num,
    CTDB_PATCH_GROUP *pg,
    int32             mem,
    CTDB_PATCH_MAP   *pm)
{
    TERRAIN_ELEMENT_XX *te;
    int32 n_te;
    int32 i;
    int32 count;
    
    if (!(te = ctdb_lookup_te_ptr(ctdb, patch_num)))
      return 0;
    
    /* Must be called after lookup_te_ptr to ensure that pg data is in
     * memory
     */
    n_te = CTDB_NUM_TE(pg, mem);
    
    if (n_te > pm->len)
    {
	/* Dynamically resize */
	pm->map = (int32 (*)[2])STDREALLOC(pm->map, 
					   n_te * 2 * sizeof(int32));
	pm->len = n_te;
    }
    
    count = 0;
    for(i=0;i<n_te;i++)
    {
	if (te[i].info.vertices < 2)
        {
	    /* No triangles here */
	    pm->map[i][0] = pm->map[i][1] = CTDB_NO_ADJACENT_TIN;
	}
	else
        {
	    pm->map[i][0] = count++;
	    if (te[i].info.vertices == 2)
	    {
		pm->map[i][1] = pm->map[i][0];
	    }
	    else
	    {
		pm->map[i][1] = count++;
	    }
	}
    }
    return count;
}


/* Starting array length for patch maps */
#define CTDB_MIN_PM_SIZE 256

/* ctdb_lookup_tin_triangles_XX generates a list of all TIN triangles in the
 * patch specified by patch_number.  The _triangles_ array is used to return
 * the triangle data (see below for more details).  *_num_triangles_ is used
 * to pass in the length of the _triangles_ array (and hence the maximum number
 * of triangles that can be returned), and to pass back the actual number
 * of triangles in the patch.  If more triangles are found than will fit in the
 * _triangles_ array, the routine returns FALSE; otherwise it returns TRUE.
 * _cross_patch_bdy_ is a boolean value which, if TRUE, specifies that TIN 
 * adjacency information should be generated for triangle edges at the patch 
 * boundary.
 *
 * The fields of each element of the _triangles_ array are filled out as
 * follows:
 *
 *    verts:          The three vertices of the triangle in clockwise order.
 *
 *    soil:           The soil type of the triangle (final global values)
 *
 *    adj_triangles:  This references the triangle adjacent at each edge 
 *                    (i.e., adj_triangles[0] has the adjacency for edge 0, 
 *                    etc.)  If the edge is not on a patch boundary, 
 *                    adj_triangles stores the index in the triangles list for 
 *                    this patch at which the adjacent triangle can be found.
 *                    Otherwise it stores the index in the triangles list 
 *                    (as would be returned on another call to 
 *                    ctdb_lookup_tin_triangles) for the patch on the other 
 *                    side of the boundary.  The special value 
 *                    CTDB_NO_ADJACENT_TIN is used when no valid adjacency 
 *                    can be found for an edge.  Note that if cross_patch_bdy 
 *                    was set to FALSE in the original call, all edges on the
 *                    patch boundary will have their adjacency set to 
 *                    CTDB_NO_ADJACENT_TIN.
 *
 *    adj_patches:    This stores the patch in which each adjacent triangle 
 *                    resides--thus, adj_patches[] will be set to the current 
 *                    patch number for all internal adjacencies, and to the 
 *                    number of the appropriate patch for other adjacencies.
 *                    Note that when adj_triangles[i] is set to 
 *                    CTDB_NO_ADJACENT_TIN, the value of adj_patches[i] is
 *                    *not* guaranteed to be a valid patch number.
 */
int32
ctdb_lookup_tin_trianglesXX(
			    CTDB                  *ctdb,
			    int32                  patch_number,
			    CTDB_TIN_TRIANGLE      triangles[],
			    int32                 *num_triangles,
			    int32                  cross_patch_bdy,
			    CTDB_PATCH_GROUP      *pg,
			    int32                  mem)
{
    TERRAIN_ELEMENT_XX *te;
    TERRAIN_ELEMENT_XX *cur;
    TERRAIN_ELEMENT_XX *prev;
    int32 n_te;
    int32 count;
    int32 i;
    int32 adj_te;
    int32 adj_edge;
    int32 adj_side;
    int32 total_triangles;
    int32 n_tri;		/* Number of triangles in a given TE 
				 * (0,1, or 2) */
    int32 size_ok;
    int32 px, py;
    int32 *soilmap;
    int32 row_in_pat;
    
    /* These statics are used to implement the mapping from TE indices to
     * triangle indices.  They are static only to avoid the overhead of
     * mallocing and freeing every time
     */
    static CTDB_PATCH_MAP patch_map;
    static CTDB_PATCH_MAP north_patch_map;
    static CTDB_PATCH_MAP south_patch_map;
    static CTDB_PATCH_MAP east_patch_map;
    static CTDB_PATCH_MAP west_patch_map;
    static int32 patch_maps_inited = FALSE;
    
    /* Find the *triangle* index (as opposed to the TE index) of the triangle
     * adjacent to edge, and put that and the appropriate patch number in
     * triangles[count].{adj_triangles[index], adj_patches[index]}
     * This does not work for internal edges of quadrilaterals.
     */
#define GET_TIN_ADJ_INFO(edge, index)                                       \
    {                                                                       \
	adj_edge = te[i].adj[(edge)];                                       \
	if (adj_edge > (int32)MAX_ALLOWED_TE_INDEX_XX)                      \
	{                                                                   \
	    triangles[count].adj_triangles[(index)] =                       \
	      CTDB_NO_ADJACENT_TIN;                                         \
	}                                                                   \
	else                                                                \
	{                                                                   \
	    adj_side = ((te[i].info.adj_sides >> ((edge)<<1)) & 0x3);       \
	    if (te[i].info.patch_adj & (1 << (edge)))                       \
	    {                                                               \
		/* Patch boundary */                                        \
		if (!cross_patch_bdy)                                       \
		{                                                           \
		    triangles[count].adj_triangles[(index)] =               \
		      CTDB_NO_ADJACENT_TIN;                                 \
		}                                                           \
		else if ((prev->pt.x == 0) && (cur->pt.x == 0))             \
		{                                                           \
		    /* West boundary */                                     \
		    triangles[count].adj_triangles[(index)] =               \
		      west_patch_map.map[adj_edge][adj_side / 2];           \
		    triangles[count].adj_patches[(index)] =                 \
		      patch_number - 1;                                     \
		}                                                           \
		else if ((prev->pt.y == 0) && (cur->pt.y == 0))             \
		{                                                           \
		    /* South boundary */                                    \
		    triangles[count].adj_triangles[(index)] =               \
		      south_patch_map.map[adj_edge][adj_side / 2];          \
		    triangles[count].adj_patches[(index)] =                 \
		      patch_number - ctdb->patches_wide;                    \
		}                                                           \
		else if (prev->pt.x == cur->pt.x)                           \
		{                                                           \
		    /* East boundary */                                     \
		    triangles[count].adj_triangles[(index)] =               \
		      east_patch_map.map[adj_edge][adj_side / 2];           \
		    triangles[count].adj_patches[(index)] =                 \
		      patch_number + 1;                                     \
		}                                                           \
		else                                                        \
		{                                                           \
		    /* North boundary */                                    \
		    triangles[count].adj_triangles[(index)] =               \
		      north_patch_map.map[adj_edge][adj_side / 2];          \
		    triangles[count].adj_patches[(index)] =                 \
		      patch_number + ctdb->patches_wide;                    \
		}                                                           \
	    }                                                               \
	    else                                                            \
	    {                                                               \
		triangles[count].adj_triangles[(index)] =                   \
		  patch_map.map[adj_edge][adj_side / 2];                    \
		triangles[count].adj_patches[(index)] = patch_number;       \
	    }                                                               \
	}                                                                   \
    }
       
       
       if (!patch_maps_inited)
       {
	   patch_map.map = (int32 (*)[2])(STDALLOC(CTDB_MIN_PM_SIZE * 
						   2 *sizeof(int32)));
	   north_patch_map.map = (int32 (*)[2])(STDALLOC(CTDB_MIN_PM_SIZE * 
							 2 *sizeof(int32)));
	   south_patch_map.map = (int32 (*)[2])(STDALLOC(CTDB_MIN_PM_SIZE * 
							 2 *sizeof(int32)));
	   east_patch_map.map = (int32 (*)[2])(STDALLOC(CTDB_MIN_PM_SIZE * 
							2 *sizeof(int32)));
	   west_patch_map.map = (int32 (*)[2])(STDALLOC(CTDB_MIN_PM_SIZE * 
							2 *sizeof(int32)));
	   patch_map.len = CTDB_MIN_PM_SIZE;
	   north_patch_map.len = CTDB_MIN_PM_SIZE;
	   south_patch_map.len = CTDB_MIN_PM_SIZE;
	   east_patch_map.len = CTDB_MIN_PM_SIZE;
	   west_patch_map.len = CTDB_MIN_PM_SIZE;
	   patch_maps_inited = TRUE;
       }
    
    count = 0;
    size_ok = TRUE;
    
    if (!(te = (TERRAIN_ELEMENT_XX *)ctdb_lookup_te_ptr(ctdb, patch_number)))
    {
	*num_triangles = 0;
	return TRUE;
    }
    
    /* This *must* be done after tha call to lookup_te_ptr to guarantee that
     * the PG data is in memory
     */
    n_te = CTDB_NUM_TE(pg, mem);
    soilmap = CTDB_GET_SOILMAP(ctdb,pg,mem);
    
    total_triangles = ctdb_get_patch_maps(ctdb,
					  patch_number,
					  cross_patch_bdy,
					  &patch_map,
					  &north_patch_map,
					  &south_patch_map,
					  &east_patch_map,
					  &west_patch_map);
    if (total_triangles > *num_triangles)
    {
	*num_triangles = total_triangles;
	return FALSE;
    }
    px = patch_number%ctdb->patches_wide;
    py = patch_number/ctdb->patches_wide;
    
    for(i=0;i<n_te;i++)
    {
	n_tri = te[i].info.vertices - 1;
	if (n_tri < 1)
	  /* TE is a point or a line, so ignore it */
	  continue;
	
	/* Grab first triangle */
	
	triangles[count].verts[0][X] = 
	  CTDB_XYO_TO_METERS(te[i].pt.x,px,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[0][Y] = 
	  CTDB_XYO_TO_METERS(te[i].pt.y,py,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[0][Z] = te[i].z;
	
	/* To ensure a clockwise point ordering, we must choose subsequent
	 * vertices based on whether this is a triangle or a quad
	 */
	prev = te + i;
	cur = (n_tri == 1) ? (te + te[i].vert.index[0]) : 
	  (te + te[i].vert.index[1]);
	triangles[count].verts[1][X] = 
	  CTDB_XYO_TO_METERS(cur->pt.x,px,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[1][Y] = 
	  CTDB_XYO_TO_METERS(cur->pt.y,py,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[1][Z] = (float64)cur->z;
	GET_TIN_ADJ_INFO(0, 0);
	
	prev = cur;
	cur = (n_tri == 1) ? (te + te[i].vert.index[1]) : 
	  (te + te[i].vert.index[0]);
	triangles[count].verts[2][X] = 
	  CTDB_XYO_TO_METERS(cur->pt.x,px,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[2][Y] = 
	  CTDB_XYO_TO_METERS(cur->pt.y,py,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[2][Z] = (float64)cur->z;
	GET_TIN_ADJ_INFO(1, 1);
	
	row_in_pat = soilmap[te[i].info.soil0];
	triangles[count].soil =
	  ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
	
	if (n_tri != 2) 
        {
	    /* Set adjacency info for third edge */
	    prev = cur;
	    cur = (TERRAIN_ELEMENT_XX *) (te + i);
	    GET_TIN_ADJ_INFO(2, 2);	    
	    count++;
	    continue;
	}
	
	count++;
	
	/* Grab second triangle */
	
	/* Set adjacency info for internal edge */
	triangles[count-1].adj_triangles[2] = count;
	triangles[count-1].adj_patches[2] = patch_number;
	triangles[count].adj_triangles[0] = count - 1;
	triangles[count].adj_patches[0] = patch_number;
	
	bcopy(triangles[count-1].verts[0], triangles[count].verts[0],
	      CTDB_PTS_IN_TRIANGLE * sizeof(float64));
	
	bcopy(triangles[count-1].verts[2], triangles[count].verts[1],
	      CTDB_PTS_IN_TRIANGLE * sizeof(float64));
	
	prev = cur;
	cur = te + te[i].vert.index[2];
	triangles[count].verts[2][X] = 
	  CTDB_XYO_TO_METERS(cur->pt.x,px,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[2][Y] = 
	  CTDB_XYO_TO_METERS(cur->pt.y,py,ctdb->patch_incrD, ctdb->incrD);
	triangles[count].verts[2][Z] = (float64)cur->z;
	GET_TIN_ADJ_INFO(2, 1);
	
	/* Set adjacency info for last edge */
	prev = cur;
	cur = (TERRAIN_ELEMENT_XX *) (te + i);
	GET_TIN_ADJ_INFO(3, 2);

	row_in_pat = soilmap[te[i].info.soil1];
	triangles[count].soil =
	  ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;

	count++;
    }
    *num_triangles = count;
    return TRUE;
}

/* This routine will populate the CTDB_TERRAIN_TRIANGLE data structure by
 * unpacking each terrain element into one or two triangles.  The output
 * from this routine is the number of triangles generated.
 */
int32
ctdb_lookup_tin_polygonsXX(
			   CTDB                  *ctdb,
			   int32                  patch_number,
			   TERRAIN_ELEMENT_XX    *te,
			   int32                  index,
			   CTDB_TERRAIN_TRIANGLE *pt0,
			   CTDB_TERRAIN_TRIANGLE *pt1)
{
    int32 i,n_pts,j;
    int32 px=(patch_number%ctdb->patches_wide);
    int32 py=(patch_number/ctdb->patches_wide);
    int32 n_vert=te[index].info.vertices;
    
    /* if this TE index defines a line or point then there is no triangle 
     * which should be added to the list.  return a negative one to 
     * prevent moving to the next element in the output list.
     */
    if ( n_vert < 2 )
      return 0;
    
    /* Save the anchor vertex which is always defined */
    pt0->vert[0][X] = CTDB_XYO_TO_METERS(te[index].pt.x,px,ctdb->patch_incrD,
					 ctdb->incrD) ;
    pt0->vert[0][Y] = CTDB_XYO_TO_METERS(te[index].pt.y,py,ctdb->patch_incrD,
					 ctdb->incrD) ;
    pt0->vert[0][Z] = te[index].z;
    
    j = te[index].vert.index[1];
    pt0->vert[1][X] = CTDB_XYO_TO_METERS(te[j].pt.x,px,ctdb->patch_incrD,
					 ctdb->incrD);
    pt0->vert[1][Y] = CTDB_XYO_TO_METERS(te[j].pt.y,py,ctdb->patch_incrD,
					 ctdb->incrD);
    pt0->vert[1][Z] = te[j].z;
    
    j = te[index].vert.index[0];
    pt0->vert[2][X] = CTDB_XYO_TO_METERS(te[j].pt.x,px,ctdb->patch_incrD,
					 ctdb->incrD);
    pt0->vert[2][Y] = CTDB_XYO_TO_METERS(te[j].pt.y,py,ctdb->patch_incrD,
					 ctdb->incrD);
    pt0->vert[2][Z] = te[j].z;
    
    /* store soil information */
    pt0->soil = te[index].info.soil0;
    pt0->index = index;
    
    if ( n_vert != 3 )
    {
	pt0->assoc_triangle = -1;
	return 1;
    }
    pt0->assoc_triangle = 0;/* will be mapped by the caller */
    pt1->assoc_triangle = 0;/* will be mapped by the caller */
    
    pt1->vert[0][X] = pt0->vert[0][X] , pt1->vert[0][Y] = pt0->vert[0][Y] ;
    pt1->vert[0][Z] = pt0->vert[0][Z] ;
    pt1->vert[1][X] = pt0->vert[2][X] , pt1->vert[1][Y] = pt0->vert[2][Y] ;
    pt1->vert[1][Z] = pt0->vert[2][Z] ;
    j = te[index].vert.index[2];
    pt1->vert[2][X] = CTDB_XYO_TO_METERS(te[j].pt.x,px,ctdb->patch_incrD,
					 ctdb->incrD);
    pt1->vert[2][Y] = CTDB_XYO_TO_METERS(te[j].pt.y,py,ctdb->patch_incrD,
					 ctdb->incrD);
    pt1->vert[2][Z] = te[j].z;
    
    /* store soil information */
    pt1->soil = te[index].info.soil1;
    pt1->index = index;
    return 2;
}
