#ifndef lint
static char rcsid [] = "$RCSfile: ct_access.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_access.c                                                       *
*                                                                           *
*       Copyright 1995 by TASC & SAIC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Driver               *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC and SAIC under U. S. Government *
*       contracts 
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Accessor functions for data in the CTDB structure         *
*       Created: Thurs. Oct 19 1995                                         *
*       Author: flchamberlain                                               *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#define SE3_OCEAN_DEMO
#include "libctdb_local.h"
#include <stdstring.h>
#include <stdtypes.h>
#include <libgcs.h>
#include <stdarg.h>

int32
ctdb_get_utm_data(CTDB *ctdb, CTDB_UTM_DATA *utm)
{
    if ( !ctdb )
      return 0;
    /* utm->source:
     *              1 = UTM_NE source data
     *              0 = GCC source data
     */
    utm->source       = !ctdb->is_curved;
    utm->datum        = ctdb->datum;
    utm->zone_number  = ctdb->origin_zone_number;
    utm->zone_letter  = ctdb->origin_zone_letter;
    utm->northing     = ctdb->origin_northing;
    utm->easting      = ctdb->origin_easting;
    return 1;
}    
    
/* Return the cell id for the given ctdb. */
int32
ctdb_get_cell_id(CTDB *ctdb)
{
    if ( ctdb )
      return ctdb->cell_id;
    else
      return GCS_ILLEGAL_CELL;

}

/* Return the extent of a ctdb in meters where (0,0) is defined to be at 
 * the lower left hand corner of the database.
 */
void
ctdb_get_extent(CTDB *ctdb, float64 *mnx, float64 *mny,
		float64 *mxx, float64 *mxy)
{
    if ( ctdb )
    {
	*mnx = ctdb->min_x;
	*mny = ctdb->min_y;
	*mxx = ctdb->max_x;
	*mxy = ctdb->max_y;
    }
    else
      *mnx = *mny = *mxx = *mxy = 0;
}

/* Return the offset of the lower left-hand corner of ctdb from the center
 * of its cell
 */
void
ctdb_get_origin(CTDB *ctdb, float64 *x, float64 *y)
{
    if (ctdb)
    {
	*x = ctdb->origin_x;
	*y = ctdb->origin_y;
    }
    else
    {
	*x = 0.0;
	*y = 0.0;
    }
}

void
ctdb_get_geographic_extent(CTDB *ctdb, float64 *south_lat, float64 *west_lng, 
			   float64 *north_lat, float64 *east_lng)
{
    if ( ctdb )
    {
	*south_lat = ctdb->south_lat;
	*west_lng  = ctdb->west_long;
	*north_lat = ctdb->north_lat;
	*east_lng  = ctdb->east_long;
    }
    else
      *south_lat = *west_lng = *north_lat = *east_lng = 0;
}

void
ctdb_get_tdb_identifier(CTDB *ctdb, char *name, int32 *version)
{
    if ( ctdb )
    {
	strcpy(name,ctdb->name);
	*version = ctdb->version;
    }
    else
    {
	strcpy(name,"Invalid\0");
	*version = 0;
    }  
}

/*VARARGS*/   
int32
ctdb_ocean_point_query(CTDB *ctdb,
		       float64 x,
		       float64 y,
		       CTDB_OCEAN_QUALIFIER qual,
		       ...
		       )
{
    va_list ap;
    float64 *surf_height;
    float64 *tide_height;
    float64 *temp;
    CTDB_WAVE_CHARS *pwave;
    CTDB_WAVE_CHARS *swave;
    CTDB_WATER_CHARS *water_ptr;

    if ( ctdb_is_tiled && !ctdb_point_on_database(ctdb,x,y))
    {
	CTDB *tdb;
	int32 ocell,cell;
	/*
	 * There is inherent error
	 * since normal vector in that cell is not parallel to
	 * normal in current cell. We make the assumption that 
	 * elevation lookup will only occur within "reasonable"
	 * range of the current cell.
	 */
	ocell = ctdb->cell_id;
	gcs_gcs2e_to_gcs(ocell,x,y,&cell,&x,&y);
	tdb = (CTDB *)gcs_get_tdb(cell, FALSE);
	
	if (!tdb || (ocell != cell))
	  return FALSE; /* There is no valid ctdb defined here */
    }

    water_ptr = ctdb_lookup_water_ptr(ctdb,x,y);
    if (!water_ptr)
      return FALSE;

    /* start processing the variable entry list */
    va_start(ap, qual);

    do
    {
	switch (qual)
	{
	  case CTDB_SURF_HEIGHT:
	    surf_height = va_arg(ap,float64 *);
	    *surf_height = water_ptr->surf_height;
	    break;
	  case CTDB_TEMP:
	    temp = va_arg(ap,float64 *);
	    *temp = water_ptr->surface_temp;
	    break;
	  case CTDB_PRIMARY:
	    pwave = va_arg(ap,CTDB_WAVE_CHARS *);
	    bcopy(&water_ptr->primary_wave,pwave,sizeof(CTDB_WAVE_CHARS));
	    break;
	  case CTDB_SECONDARY:
	    swave = va_arg(ap,CTDB_WAVE_CHARS *);
	    bcopy(&water_ptr->secondary_wave,swave,sizeof(CTDB_WAVE_CHARS));
	    break;
	  case CTDB_TIDAL_HEIGHT:
	    tide_height = va_arg(ap, float64 *);
	    *tide_height =
	      ctdb->tidal_zones[water_ptr->tidal_zone_index].offset;
	    break;
	  default:
	    return TRUE;
	}

    } while ( qual =va_arg(ap, CTDB_OCEAN_QUALIFIER) );

    return TRUE;    
}

/*VARARGS*/   
void
ctdb_ocean_point_write(CTDB *ctdb,
		       float64 x,
		       float64 y,
		       CTDB_OCEAN_QUALIFIER qual,
		       ...
		       )
{
    va_list ap;
    CTDB_WAVE_CHARS *pwave;
    CTDB_WAVE_CHARS *swave;
    CTDB_WATER_CHARS *water_ptr;

    /* $$$ this processing will current handle only SIMNET and SINGLE cell
     * databases.  If using a GCS_MULTI_CELL database the implication of this
     * API are uncertain.
     */
    if ( ctdb_is_tiled )
      return;
   
    water_ptr = ctdb_lookup_water_ptr(ctdb,x,y);
    if ( !water_ptr )
      return;

    /* start processing the variable entry list */
    va_start(ap, qual);

    do
    {
	switch (qual)
	{
	  case CTDB_SURF_HEIGHT:
	    water_ptr->surf_height = va_arg(ap,float64);
	    break;
	  case CTDB_TEMP:
	    water_ptr->surface_temp = va_arg(ap,float64);
	    break;
	  case CTDB_PRIMARY:
	    pwave = va_arg(ap,CTDB_WAVE_CHARS *);
	    bcopy(pwave,&water_ptr->primary_wave,sizeof(CTDB_WAVE_CHARS));
	    break;
	  case CTDB_SECONDARY:
	    swave = va_arg(ap,CTDB_WAVE_CHARS *);
	    bcopy(swave,&water_ptr->secondary_wave,sizeof(CTDB_WAVE_CHARS));
	    break;
	  case CTDB_TIDAL_HEIGHT:
	    ctdb->tidal_zones[water_ptr->tidal_zone_index].offset = 
	      va_arg(ap,float64);
	    break;
	  default:
	    return;
	}

    } while ( qual =va_arg(ap, CTDB_OCEAN_QUALIFIER) );
}

	    
float64
ctdb_lookup_micro_tide(CTDB *ctdb,
		       int32 patch_number,
		       int32 micro_water_index)
{
    CTDB_PATCH_GROUP *pg;
    int32 member;
    
    /* Protect against BOGUS patch numbers */
    if ( !CTDB_INCLUSIVE_BETWEEN(patch_number,
				 ctdb->min_patch_number,
				 ctdb->max_patch_number) )
      return 0.0;
    
    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);

    if (!pg)
      return 0.0;

    if (pg->mapped)
      ctdb->patch_hit_count++;
    else
      ctdb_fetch_patch_group(ctdb, pg);
    
    return CTDB_GET_MICRO_TIDE(ctdb,patch_number,index);
}

int32 
ctdb_compute_patch_num(CTDB   *ctdb, 
		       float64 x, 
		       float64 y,
		       int32  *patch_x, 
		       int32  *patch_y)
{
    int32 intern_patch_x;
    int32 intern_patch_y;
    
    intern_patch_x = PATCH_X(x * ctdb->meters_to_patch);
    if (patch_x)
      *patch_x = intern_patch_x;
    
    intern_patch_y = PATCH_Y(y * ctdb->meters_to_patch);
    if (patch_y)
      *patch_y = intern_patch_y;
    
    return (intern_patch_x + intern_patch_y * ctdb->patches_wide);
}

