#ifndef lint
static char rcsid [] = "$RCSfile: ct_clip.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_clip.c                                                         *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to clip a line to a bounding box                     *
 *       Created: Mon Oct 21 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include "libctdb_local.h"

#define MEMCPY_STD(a,b,c) bcopy(b,a,c)

#define MAX_DIM 10

int32 ctdb_clip_intersections(float64 *v0, float64 *v1, 
			      float64 min_x, float64 max_x, 
			      float64 min_y, float64 max_y, 
			      float64 *clipped, int32 dim)
{
    int32   c;
    int32   l[2],r[2],b[2],t[2];
    int32   out;
    int32   i;
    float64 pts[2][MAX_DIM], x0,y0,z0,x1,y1,z1, d;
    
    /* This algorithm finds intersections of the line from v0 to v1 with
     * box specified by min_x, max_x, min_y, max_y.  The code is
     * my own implementation of the Cohen-Sutherland algorithm.
     * Basically, you just keep clipping the line from both ends
     * until it is within the box.  I also will truncate up to MAX_DIM
     * other variables simultaneously (such as a Z coordinate).
     * This is used by ctdb_clip_line for regular line clipping, and
     * within the ctdb compiler for polygon clipping.
     */
    
    for(i=0;i<dim;i++)
      pts[0][i] = v0[i], pts[1][i] = v1[i];
    
    while (1)
    {
	for(i=0;i<2;i++)
	{
	    l[i] = pts[i][0] < min_x;
	    r[i] = pts[i][0] > max_x;
	    b[i] = pts[i][1] < min_y;
	    t[i] = pts[i][1] > max_y;
	}
	
	if (l[0]&&l[1] || r[0]&&r[1] || b[0]&&b[1] || t[0]&&t[1])
	  return 0;
	
	out = 0;
	if (!(l[0] | r[0] | t[0] | b[0]))
	{
	    out = 1;
	    if (!(l[1] | r[1] | t[1] | b[1]))
	    {
		c = 0;
		if (!((pts[0][0] == v0[0]) && (pts[0][1] == v0[1]) ||
		      (pts[0][0] == v1[0]) && (pts[0][1] == v1[1])))
		{
		    for(i=0;i<dim;i++)
		      clipped[c++] = pts[0][i];
		}
		if (!((pts[1][0] == v0[0]) && (pts[1][1] == v0[1]) ||
		      (pts[1][0] == v1[0]) && (pts[1][1] == v1[1])))
		{
		    for(i=0;i<dim;i++)
		      clipped[c++] = pts[1][i];
		}
		return c / dim;
	    }
	}
	
	x0 = pts[0][0]; y0 = pts[0][1];
	x1 = pts[1][0]; y1 = pts[1][1];
	
	if (l[out])
	  d = (min_x - x0) / (x1 - x0);
	else if (r[out])
	  d = (max_x - x0) / (x1 - x0);
	else if (b[out])
	  d = (min_y - y0) / (y1 - y0);
	else if (t[out])
	  d = (max_y - y0) / (y1 - y0);
	else
	  d = 0.0;
	
	pts[out][0] = x0 + (x1 - x0) * d;
	pts[out][1] = y0 + (y1 - y0) * d;
	
	/* Because of rounding error, the above two lines are not
	 * guaranteed to clip as they are supposed to.  Make sure the
	 * points do get clipped.
	 */
	if (l[out])
	  pts[out][0] = min_x;
	else if (r[out])
	  pts[out][0] = max_x;
	else if (b[out])
	  pts[out][1] = min_y;
	else if (t[out])
	  pts[out][1] = max_y;
	
	for(i=2;i<dim;i++)
	{
	    z0 = pts[0][i];
	    z1 = pts[1][i];
	    pts[out][i] = z0 + (z1 - z0) * d;
	}
    }
#ifdef lint
    return 1;
#endif
}

/*VARARGS6*/
int32 ctdb_clip_line(float64 *v0, float64 *v1, float64 min_x, float64 max_x,
		     float64 min_y, float64 max_y, float64 *clipped, int32 dim)
{
    int32 c=0;
    int32 i;
    
    /* Copy point 0 if it is OK */
    if ((v0[0] >= min_x) && (v0[0] <= max_x) &&
	(v0[1] >= min_y) && (v0[1] <= max_y))
    {
	for (i=0;i<dim;i++)
	  clipped[c++] = v0[i];
    }
    
    /* Find clipping points */
    c += dim *
      ctdb_clip_intersections(v0,v1,min_x,max_x,min_y,max_y,clipped+c,dim);
    
    /* Copy point 1 if it is OK */
    if ((v1[0] >= min_x) && (v1[0] <= max_x) &&
	(v1[1] >= min_y) && (v1[1] <= max_y))
    {
	for (i=0;i<dim;i++)
	  clipped[c++] = v1[i];
    }
    
    /* Never return less than two points */
    if (c > dim)
      return c/dim;
    else
      return 0;
}


/* This function clips a 3D polygon to a ground rectangle.  The
assumption is made that a polygon may be convex, but it may not be
convex across rectangle boundaries.  I.e., when the polygon is clipped
to a rectangle, either 0 or one polygon portion remains.  This may
later be relaxed.

This is an implementation of the Liang-Barsky clipping algorithm.
*/

static CTDB_CLIP_RESULT
ctdb_clip_3d_polygon_to_rect(
    int32    n_verts,  	
    float64 *input_verts, 
    float64  rect_min_x, 
    float64  rect_max_x, 
    float64  rect_min_y, 
    float64  rect_max_y,
    int32   **out_n_verts,
    float64 **out_verts)
{
    int32    	l1;		/* Standard local loop variable */

    float64    *tmp_f64_p; 	/* Temporary convenience variables */
    float64  	tmp_f64;

    float64  	last_x, last_y,	/* Variables for Liang-Barsky Algorithm */
	     	last_z;
    float64  	delta_x, 
	     	delta_y;
    float64  	xIn, yIn, 
	     	xOut, yOut;
    float64  	tOut1, tIn2, 
	     	tOut2;
    float64  	tInX, tOutX, 
	       	tInY, tOutY;

    CTDB_CLIP_RESULT result;		/* Output clipping result */
    int32	original_vertices_reused;


    /* Set up memory for returned values */
    /* We are going to allocate maximum memory needed: input vertices * 3,
	* 3 components of vertices, * 8 bytes per float64 */
    *out_n_verts = (int32 *)   STDALLOC(12 + ((n_verts * 9) << 3));
    *out_verts   = (float64 *) ((*out_n_verts)+3);

    **out_n_verts = 0;
    original_vertices_reused = 0;

    /* Start Liang-Barsky */

    last_x = *(tmp_f64_p = input_verts + (n_verts-1) * 3);
    last_y = *(tmp_f64_p + 1);
    last_z = *(tmp_f64_p + 2);

    tmp_f64_p = *out_verts;

    for (l1=0; l1<n_verts; l1++)
    {
	delta_x = *input_verts - last_x;
	delta_y = *(input_verts+1) - last_y;
	
	if (delta_x > 0 || (!delta_x && last_x > rect_max_x))
	{
	    xIn  = rect_min_x;
	    xOut = rect_max_x;
	}
	else
	{
            xOut = rect_min_x;
            xIn  = rect_max_x;
        }

        if (delta_y > 0 || (!delta_y && last_y > rect_max_y))
        {
            yIn  = rect_min_y;
            yOut = rect_max_y;
        }
        else
        {
            yOut = rect_min_y;
            yIn  = rect_max_y;
        }

	if (delta_x)
	    tOutX = (xOut - last_x) / delta_x;
	else if (last_x <= rect_max_x && rect_min_x <= last_x)
	    tOutX = HUGE;
	else
	    tOutX = -HUGE;

        if (delta_y)
            tOutY = (yOut - last_y) / delta_y;
        else if (last_y <= rect_max_y && rect_min_y <= last_y)
            tOutY = HUGE;
        else
            tOutY = -HUGE;

	if (tOutX < tOutY)
	{
	    tOut1 = tOutX;
	    tOut2 = tOutY;
	}
	else
	{
            tOut1 = tOutY;
            tOut2 = tOutX;
        }

	if (tOut2 > 0)
	{
	    if (delta_x)
		tInX = (xIn - last_x) / delta_x;
	    else
		tInX = -HUGE;

	    if (delta_y)
                tInY = (yIn - last_y) / delta_y;
            else
                tInY = -HUGE;

	    tIn2 = (tInX < tInY) ? tInY : tInX;

	    if (tOut1 < tIn2)
	    {
		if (tOut1 > 0 && tOut1 <= 1)
		{
		    tmp_f64 = (*(input_verts+2) - last_z);
		    if (tInX < tInY)
		    {
			*((*out_verts)++) = xOut;
			*((*out_verts)++) = yIn;
			*((*out_verts)++) = last_z + tmp_f64 * tInY;
		    }
		    else
		    {   
                        *((*out_verts)++) = xIn;
                        *((*out_verts)++) = yOut;
                        *((*out_verts)++) = last_z + tmp_f64 * tInX;
                    }
		    (**out_n_verts)++;
		}
	    }
	    else
	    {
		if (tOut1 > 0 && tIn2 <= 1)
		{
		    if (tIn2 > 0 && tIn2 < 1)
		    {
			tmp_f64 = (*(input_verts+2) - last_z);
			if (tInX > tInY)
			{
			    *((*out_verts)++) = xIn;
			    *((*out_verts)++) = last_y + tInX * delta_y;
			    *((*out_verts)++) = last_z + tmp_f64 * tInX;
			}
			else
			{
                            *((*out_verts)++) = last_x + tInY * delta_x;
                            *((*out_verts)++) = yIn;
                            *((*out_verts)++) = last_z + tmp_f64 * tInY;
                        }
			(**out_n_verts)++;
		    }

		    if (tOut1 <= 1)
		    {
			tmp_f64 = (*(input_verts+2) - last_z);
			if (tOutX < tOutY)
                        {       
                            *((*out_verts)++) = xOut;
                            *((*out_verts)++) = last_y + tOutX * delta_y;
                            *((*out_verts)++) = last_z + tmp_f64 * tOutX;
                        }
                        else
                        {
                            *((*out_verts)++) = last_x + tOutY * delta_x;
                            *((*out_verts)++) = yOut;
                            *((*out_verts)++) = last_z + tmp_f64 * tOutY;
                        }
                        (**out_n_verts)++;
                    }
		    else
		    {
		        *((*out_verts)++) = *input_verts;
		    	*((*out_verts)++) = *(input_verts+1);
		    	*((*out_verts)++) = *(input_verts+2);
		    	(**out_n_verts)++;
			original_vertices_reused++;
		    }
		}
	    }
	   
	    if (tOut2 < 1)
	    {
		*((*out_verts)++) = xOut;
                *((*out_verts)++) = yOut;
		*((*out_verts)++) = last_z;

		/* I have been told that this z value here doesn't really
		 * matter, since it is an interior point on the model.  The
		 * way to correctly calculate this (assuming that the roof of
		 * the model is planer) is to calculate the equation of the
		 * roof plane from the vertices and evaluate at (xOut, yOut).
		 * Note that if the roof is horizontal, the above works anyway.
		 */ 

                (**out_n_verts)++;
	    }
	}
	last_x = *(input_verts++);
	last_y = *(input_verts++);
	last_z = *(input_verts++);
    }
    /* END Liang-Barsky */

    *out_verts = tmp_f64_p;

    if (original_vertices_reused == n_verts)
	return(SUBJ_IN_CLIP);
    else if (**out_n_verts > 2)
    {
	if (original_vertices_reused)
	    return(INTERSECT);
	else
	    return(CLIP_IN_SUBJ);
    }
    else
    {
	/* No intersection with rectangle */
	STDDEALLOC(*out_n_verts);
	return(SEPARATE);
    }
}

/*============================== CTDB_CLIP_POLYGON ============================*/
int32 *
ctdb_clip_polygon(CTDB *ctdb, int32 n_verts, float64 *verts)
/*******************************************************************************
 Find which patches the polygon may intersect by calculating a bounding
 box and compare to the patch size.  Then, call another function to clip to
 those boundaries. 
 
 The function passes back a block of data, representing:

	int32		number of patches the polygon covers
	int32(s)	list of patch number(s) covered by the polygon (x, y)
	int32(s) 	number of vertices on each covered patch
	float64s 	output vertex information (x, y, z; x, y, z)

 Note that an initial pointer to the beginning of information is:

	number of patches  = beginning of block
	list of patches    = beginning of block + 1
	number of vertices = beginning of block + 1 + 2 * number of patches
	vertices           = beginning of block + 1 + 3 * number of patches

 A "number of patches" value of -1 indicates that the polygon remained 
 entirely in the original patch.  In this case, the calling routine can use the 
 original set of vertices.  But since we had to figure out the information 
 anyway, we include the patch number, in case the calling routine wants it.  
 The format is therefore:

	int32	"in original" data flag (-1)
	int32	patch number containing the feature

 Return value of 0 is reserved for errors.

 It is up to the calling routine to delete this memory block after it is 
 returned.
*******************************************************************************/
{
    int32    l1, l2;		/* Standard loop variables */
    float64  tmp_f64;		/* Temporary convenience variable */

    float64  min_x, max_x,	/* Building bounding box limits */
	     min_y, max_y;

    int32    min_x_patch,	/* Bounding box in units of patch */
	     min_y_patch,
	     max_x_patch,
	     max_y_patch;
    float64  patch_min_x, 	/* Min values of patch boundaries */
	     patch_min_y;
    float64  patch_meters;	/* Size of a patch */
    int32    current_vertex;	/* index into vertex array */

    int32   *output;		/* Output memory block */
    int32    clip_result;	/* Return value from patch clipping */
    int32   *patch_data[16];	/* Pointers to handle individual patch info */
    float64 *patch_verts[16];
    int32    total_vertices;	/* Total # of verts in output */
    int32    info_block;	/* Index into array of patch info */

    int32   *out_patch_p,	/* Pointers to parts of the patch info array */
	    *out_n_verts;
    float64 *out_verts;


    /* Calculate polygon bounding box, and convert to units of "patch" */
    min_x = *verts;
    min_y = verts[1];
    max_x = min_x;
    max_y = min_y;

    for (l1=n_verts-1, current_vertex = 3; l1; l1--, current_vertex += 2)
    {
	if (verts[current_vertex] > max_x)
	    max_x = verts[current_vertex];
	else if (verts[current_vertex] < min_x)
	    min_x = verts[current_vertex];

	current_vertex++;

	if (verts[current_vertex] > max_y)
            max_y = verts[current_vertex];
	else if (verts[current_vertex] < min_y)
            min_y = verts[current_vertex];
    }

    min_x_patch = min_x * (tmp_f64 = ctdb->meters_to_patch);
    min_y_patch = min_y * tmp_f64;
    max_x_patch = max_x * tmp_f64;
    max_y_patch = max_y * tmp_f64;

    /* Trivial: everything stays in the same patch */

    if (min_x_patch == max_x_patch && min_y_patch == max_y_patch)
    {
	output = (int32 *) STDALLOC(12);
	*output   = -1;
	output[1] = min_x_patch;
	output[2] = min_y_patch;
    }
    else /* Need to break up on each patch */
    {
	/* Bounds check! */

	if (min_x_patch < 0)
	    min_x_patch = 0;
	if (min_y_patch < 0)
            min_y_patch = 0;
        if (max_x_patch >= ctdb->patches_wide)
            max_x_patch = (ctdb->patches_wide) - 1;
        if (max_y_patch >= ctdb->patches_high)
            max_y_patch = (ctdb->patches_high) - 1;

	info_block = 0;
	patch_meters = ctdb->incr * ctdb->patch_incr;
	patch_min_y = min_y_patch * patch_meters;
	tmp_f64 = min_x_patch * patch_meters;

	for (l1=min_y_patch; l1<=max_y_patch; l1++, patch_min_y += patch_meters)
	{
	    if (info_block >> 5)
	    {
		/* More than 16 patches used ?!?  Transfer this data now! 
		   This is not currently implemented because it is not yet
		   considered possible to happen 
		*/
		info_block = 0;
	    }

	    patch_min_x = tmp_f64;

	    for (l2=min_x_patch; l2 <= max_x_patch; l2++)
	    {
		clip_result = ctdb_clip_3d_polygon_to_rect(n_verts, verts, 
					patch_min_x, 
					patch_min_x + patch_meters,
					patch_min_y, 
                                        patch_min_y + patch_meters,
					&patch_data[info_block],
					&patch_verts[info_block]);

		patch_min_x += patch_meters;
		
		if (clip_result != SEPARATE)
		{
		    *(patch_data[info_block]+1) = l2;
		    *(patch_data[info_block]+2) = l1;
		    info_block++;
		}
	    }
	}

	/* Put it all together for return! (Will need to be modified if we
		expect over 16 patches to be covered) */

	total_vertices = 0;
	for (l1=0; l1<info_block; l1++)
	    total_vertices += *patch_data[l1];

	/* Total output block:
		int32 + 2 int32 / patch + 1 int32 / patch + 3 float64 / vertex
	   or (4 + 12/patch + 24/vertex) bytes
	*/
				   
	output = (int32 *) STDALLOC((1 + 3 * (info_block + 
					(total_vertices << 1))) << 2);
	
	*output = info_block;	/* number patches covered */ 
	
 	/* Set up pointers to the datastructure to contain the patch number,
	   number of vertices on the patch, and the list of output vertices 
	 */

	out_patch_p = output+1;
	out_n_verts = out_patch_p+(info_block << 1);
	out_verts   = (float64 *) (out_n_verts+info_block);
	
	for (l1=0; l1<info_block; l1++)
	{
	    *out_n_verts     = *patch_data[l1];
	    *(out_patch_p++) = patch_data[l1][1];
	    *(out_patch_p++) = patch_data[l1][2];

	    MEMCPY_STD(out_verts, patch_verts[l1], *out_n_verts*24);
	    
	    out_verts += *(out_n_verts++)*3;

	    STDDEALLOC(patch_data[l1]);
	}
    }
    return(output);
}
