#ifndef lint
static char rcsid [] = "$RCSfile: ct_contour.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_contour.c                                                      *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to generate contour lines                            *
 *       Created: Wed Jun  5 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include "libctdb_local.h"
#include <stdext.h>
#include <libgcs.h>
#include <stdcoord.h>


static void gcs_box_to_ll_box(int32 cell0, float64 xmin, float64 ymin,
			      float64 xmax, float64 ymax, 
			      float64 *latmin, float64 *lngmin,
			      float64 *latmax, float64 *lngmax);
static void gcs_box_to_gcs_box(int32 cell0, float64 x0_min, float64 y0_min,
			       float64 x0_max, float64 y0_max,
			       int32 cell1, float64 *x1_min, float64 *y1_min,
			       float64 *x1_max, float64 *y1_max);

/* Largest supported screen size (pixels) */
#define LARGEST_SCREEN (1300*1100)

#define BOGUS_ELEVATION (-1.0E+100)

void ctdb_contour_image(CTDB *ctdb, int32 low_x, int32 low_y, 
			int32 width, int32 height, 
			float64 scale_mpp, int32 dirt_pixel, 
			int32 contour_pixel, char *img)
{
    int32 xp, yp;
    float64 xn, yn, low_xn, low_yn;
    float64 z, scale;
    int32 w2, h2;
    int32 pixel, base, row, img_row, index;
    float64 inv_interval = 0.05;
    int16 zmap[LARGEST_SCREEN];
    
    if (!ctdb)
    {
	bzero(img, height*width);
	return;
    }
    /* Scale is in units of posts/pixels */
    scale = scale_mpp * ctdb->inv_incr;
    w2 = width;
    h2 = height;
    low_xn = low_x * ctdb->inv_incr;
    low_yn = low_y * ctdb->inv_incr;
    
    /* We do this in two passes.  First we store the Z value of every
     * other pixel.  Then we go back and find the contours.
     */
    
    for (yp=0,yn=low_yn; (yp<h2); yp++,yn+=scale)
    {
	row = w2*yp;
	for (xp=0,xn=low_xn; (xp<w2); xp++,xn+=scale)
	{
	    /* Note:  The old ctdb routine for this used an approx flag to
	     * determine if we needed to interpolate between posts 
	     * (scale >= 0.5), or if finding the post was enough.  This was 
	     * an efficiency hack for gridded databases being displayed at 
	     * coarse resolution.   Since it ignored microterrain, etc., 
	     * I don't feel _too_ bad getting rid
	     * of it.  If performance is really, truly needed, try using the 
	     * grid posts if they're there, and the vgrid-mapped tes otherwise.
	     */
	    z = ctdb_lookup_norm_elevation_amsl(ctdb, xn, yn,
						CTDB_MAX_LAND_NO_WATER);
	    zmap[xp+row] = (int32)(z*inv_interval);
	}
    }
    
    for(yp=0;yp<h2;yp++)
    {
	row = w2*yp;
	img_row = width*(h2-yp-1);
	for(xp=0,base=img_row;xp<w2;xp++,base+=1)
	{
	    if (!xp || !yp || (xp==w2-1) || (yp==h2-1))
	      pixel = dirt_pixel;
	    else
	    {
		index = xp+row;
		z = zmap[index];
		
		if ((z > zmap[index-w2-1]) ||
		    (z > zmap[index-w2  ]) ||
		    (z > zmap[index-w2+1]) ||
		    (z > zmap[index   -1]) ||
		    (z > zmap[index   +1]) ||
		    (z > zmap[index+w2-1]) ||
		    (z > zmap[index+w2  ]) ||
		    (z > zmap[index+w2+1]))
		  pixel = contour_pixel;
		else
		  pixel = dirt_pixel;
	    }
	    
	    img[base]= pixel;
	}
    }
}

#define FUDGE 0.0001

int32 
ctdb_find_z_on_line(
		    CTDB_XYO_COORD2D xy0, 
		    float64 z0, 
		    CTDB_XYO_COORD2D xy1, 
		    float64 z1, 
		    int32 px, 
		    int32 py, 
		    float64 patch_incrD, 
		    float64 z,
		    float64 *x_res, 
		    float64 *y_res
		    )
{
    float64 min_z, max_z;
    float64 dxdz, dydz;
    float64 x, y, dz;
    
    if (z0 < z1)
      min_z = z0, max_z = z1;
    else
      min_z = z1, max_z = z0;
    
    /* Is there such a Z on this line? */
    if ((max_z - min_z < FUDGE) || (z < min_z) || (z > max_z))
      return FALSE;
    
    /* Find x & y */
    dxdz = ((int32)xy1.x - (int32)xy0.x) / (z1 - z0);
    dydz = ((int32)xy1.y - (int32)xy0.y) / (z1 - z0);
    
    dz = (z - z0);
    x = xy0.x + dxdz * dz;
    y = xy0.y + dydz * dz;
    
    *x_res = XYO_TO_POST(x, px, patch_incrD);
    *y_res = XYO_TO_POST(y, py, patch_incrD);
    
    return TRUE;
}

static void 
find_feature_z_bounds(
		      CTDB_FEATURE_DATA *data,
		      int32    incr,
		      float64 *min_z, 
		      float64 *max_z
		      )
{
    float64 z;
    int32 verts = (data->info.vertices&MICRO_VERTS);

    *max_z = *min_z = data[2].z;
    
    if ( verts > 2 )
    {
	z = data[incr*data[3].micro.vert2+2].z;

	if (z < *min_z) 
	  *min_z = z;
	else if (z > *max_z) 
	  *max_z = z;
    }

    /* $$$ if there is a tidal break you must check the same edge
     * as above but apply the second tidal value.  Extend Ocean Rep.
     */


    if (verts > 1)
    {
	z = data[incr*data[3].micro.vert1+2].z;

	if (z < *min_z) 
	  *min_z = z;
	else if (z > *max_z) 
	  *max_z = z;
    }

    if (verts)
    {
	z = data[incr*data[3].micro.vert0+2].z;

	if (z < *min_z) 
	  *min_z = z;
	else if (z > *max_z) 
	  *max_z = z;
    }
}

/* Calculate patch number from postx, posty */
#define PATCH_NUM(px, py) (((px)/ctdb->patch_incr) + \
			   ((((py)-1)/ctdb->patch_incr) * ctdb->patches_wide))


static void ctdb_contour_segs_in_ctdb(CTDB *ctdb, CTDB *src_ctdb,
				      float64 interval, 
				      int32 x_low, int32 y_low, 
				      int32 x_high, int32 y_high, 
				      void (*draw_fcn) (/* ??? */))
{
    int32 xp_low, yp_low, xp_high, yp_high;
    int32 xp, yp;
    float64 xpD, ypD;
    int32 xpatch_low, ypatch_low, xpatch_high, ypatch_high;
    int32 patch_x, patch_y;
    CTDB_FEATURE_DATA *feature, *last;
    int32 g0, g1, g2, next;
    int32 found;
    float64 zNW, zNE, zSW, zSE;
    int32 post, next_post;
    float64 x, y, z, idzW, idzN, idzS, idzE, idzD;
    float64 min_z, max_z, dx, dy;
    float64 ii_basis;
    float64 incr = ctdb->incrD;
    CTDB_CONTOUR_STATE state;
    float64 elev_buf[10000]; /* Accomodate 1250km wide db */
    ADDRESS te_patch;
    int32 dynamic_patch;
    int32 patch_num;
    float64 src_x, src_y, src_z;
    int32 local = (ctdb == src_ctdb);
    int32 src_cell = ctdb_get_cell_id(src_ctdb);
    
    if (!ctdb)
      return;
    
    state.inv_interval = 1.0/interval;
    state.max_delta = interval * CTDB_MAX_CONTOURS_PER_EDGE;
    ii_basis = ctdb->fixed_point_basis * state.inv_interval;
    
    state.seg_cnt = 0;
    state.interval = interval;
    state.draw_fcn = draw_fcn;
    
    xp_low = POST_X(x_low * ctdb->inv_incr);
    xp_high = POST_X(x_high * ctdb->inv_incr);
    yp_low = POST_Y(y_low * ctdb->inv_incr);
    yp_high = POST_Y(y_high * ctdb->inv_incr);
    
    if (xp_low <= 0) xp_low = 1;
    if (yp_low <= 0) yp_low = 1;
    
    if (!(xp_low % ctdb->patch_incr)) xp_low--;
    
    if (xp_high >= ctdb->max_x_post-1) xp_high = ctdb->max_x_post-2;
    if (yp_high >= ctdb->max_y_post-1) yp_high = ctdb->max_y_post-2;
    
    if ((xp_high < xp_low) || (yp_high < yp_low))
      return;
    
    for (xp=xp_low;xp<=xp_high+1;xp++)
      elev_buf[xp] = BOGUS_ELEVATION;
    
    /* Find the contours for the regular grid, if there is one */
    if ( ctdb->is_gridded )
      for (yp=yp_low,ypD=yp;yp<=yp_high;yp++,ypD+=1.0)
      {
	  patch_num = PATCH_NUM(xp_low, yp);
	  
	  /* Patch is gridded iff there are no TEs in it */
	  te_patch = ctdb_lookup_te_ptr(ctdb,patch_num);

	  if ( !te_patch )
	  {
	      post = ctdb_lookup_elev_post(ctdb, xp_low, yp);

	      zNE = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
	      zNE = ctdb_get_norm_elevation_amsl(ctdb, (float64) xp_low, 
						 (float64) yp, 
						 (float64) zNE);
	      zNE *= state.inv_interval;
	      
	      if ((zSE = elev_buf[xp_low]) == BOGUS_ELEVATION)
	      {
		  /* elev_buf only gets valid values when gridposts are 
		   * available.  We'll have to look this one up the hard way,
		   * remembering that we have to renormalize (hence the 
		   * multiply)
		   */
		  zSE = 
		    ctdb_lookup_norm_elevation_amsl(ctdb, (float64)xp_low,
						    (float64)(yp-1),
						    CTDB_MAX_LAND_NO_WATER) * 
						      state.inv_interval;
	      }
	      elev_buf[xp_low] = zNE;
	  }
	  else
	  {
	      /* Just in case we're on a boundary with a posted patch
	       * which will want to use this value
	       */
	      elev_buf[xp_low] = BOGUS_ELEVATION;
	  }
	  
	  for (xp=xp_low,xpD=xp;xp<=xp_high;xp++,xpD+=1.0)
	  {
	      if (!(xp % ctdb->patch_incr))
		te_patch = ctdb_lookup_te_ptr(ctdb,++patch_num) ;
	      
	      if ( te_patch )
	      {
		  /* skip any patch w/o grid posts 
		   * To do this, put xp at the end of this patch, and let the
		   * main loop handle the transition
		   */
		  while((xp+1)%ctdb->patch_incr)
		    elev_buf[++xp] = BOGUS_ELEVATION;
		  elev_buf[xp+1] = BOGUS_ELEVATION;
		  xpD = xp;
		  zNE = zSE = BOGUS_ELEVATION;
		  continue;
	      }
	      
	      /* If we just left a non-gridded patch, these won't be set
	       * properly, so we have to go get them
	       */
	      if ((zNW = zNE) == BOGUS_ELEVATION)
		zNW = 
		  ctdb_lookup_norm_elevation_amsl(ctdb, (float64)(xp),
						  (float64)(yp),
						  CTDB_MAX_LAND_NO_WATER) *
						    state.inv_interval;
	      if ((zSW = zSE) == BOGUS_ELEVATION)
		zSW= ctdb_lookup_norm_elevation_amsl(ctdb, (float64)(xp), 
						     (float64)(yp-1),
						     CTDB_MAX_LAND_NO_WATER) *
						       state.inv_interval;
	      
	      next_post = ctdb_lookup_elev_post(ctdb, xp+1, yp);

	      zNE = GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
	      zNE = ctdb_get_norm_elevation_amsl(ctdb, (float64)(xp+1), 
						 (float64) yp,(float64) zNE);
	      zNE *= state.inv_interval;

	      if ((zSE = elev_buf[xp+1]) == BOGUS_ELEVATION)
	      {
		  /* elev_buf only gets valid values when gridposts are 
		   * available.  We'll have to look this one up the hard way,
		   * remembering that we have to renormalize (hence the 
		   * multiply)
		   */
		  zSE = ctdb_lookup_norm_elevation_amsl(ctdb, (float64)(xp+1),
							(float64)(yp-1),
							CTDB_MAX_LAND_NO_WATER
							)*state.inv_interval;
	      }
	      elev_buf[xp+1] = zNE;
	      
	      /* If this post has microterrain on it, then don't process
	       * it.  Go on to the next post.  We'll handle this area when
	       * we process the microterrain.
	       *
	       * If it's a dynamic patch that has microterrain, we'll
	       * assume that the whole patch is covered.  This very well
	       * may be an incorrect assumption, but it's the best we can
	       * do.  You see, if it's a dynamic patch, the post flags
	       * will not necessarily be correct, since the posts are cached
	       * and there's no "dynamic post buffer" like there is for
	       * patches.  So we'll check for the presence of any microterrain
	       * if this is a dynamic patch.  If there is some, then we'll
	       * skip it and let the microterrain processing handle it.
	       */
	      dynamic_patch = CTDB_DPB_HAS_PATCH(ctdb_dpb_has_patch(
                              PATCH_NUM(xp, yp)));
	      if (post & POST_MICROTERRAIN)
	      {
		  post = next_post;
		  continue;
	      }
	      else if(dynamic_patch)
	      {
		  /* This patch is dynamic.  We'll check for the presence
		   * of microterrain.  Since microterrain is always stored
		   * first, all we have to do is check the type of the
		   * first feature.
		   */
		  CTDB_FEATURE_DATA *feat;
		  feat = ctdb_lookup_feature(ctdb, patch_num);

		  if(feat)
		  {
		      feat++;
		      if(feat->info.feature_class == FC_MICRO)
		      {
			  post = next_post;
			  continue;
		      }
		  }
	      }

	      post = next_post;
	      
	      /* Filter out flat-ish boxes */
	      /*
		if (((int32)zNW == (int32)zNE) &&
		((int32)zSW == (int32)zSE) &&
		((int32)zNW == (int32)zSE))
		*/

#define FIND_SLOPES(idz, p0, p1)					\
	      if (p0 == p1)						\
		idz = 0.0;						\
	      else							\
		idz = 1.0/(p0-p1);
		
	      FIND_SLOPES(idzW,zNW,zSW);
	      FIND_SLOPES(idzN,zNW,zNE);
	      FIND_SLOPES(idzS,zSW,zSE);
	      FIND_SLOPES(idzE,zNE,zSE);
	      if(post & POST_SW_NE)
	      {
		  FIND_SLOPES(idzD,zSW,zNE);
	      }
	      else
	      {
		  FIND_SLOPES(idzD,zNW,zSE);
	      }
	      
#define CHECK_Z(c)							\
	      if (c < min_z) min_z = c;					\
	      else if (c > max_z) max_z = c
		
		min_z = max_z = zNW;
	      CHECK_Z(zNE);
	      CHECK_Z(zSW);
	      CHECK_Z(zSE);
	      
	      /* This is a slight problem.  Really want to do a 3D coord 
	       * convert from x, y, and whatever z' corresponds to the 
	       * point (x,y,z') that has an altitude above the ellipsoid of z
	       *
	       * Instead, we just do a 2d coord convert.  Don't think the error
	       * is all that much.
	       */
	      
#define STORE(x,y,z)							  \
	      {								  \
		  if (!local)                                             \
		  {                                                       \
		      gcs_gcs2e_to_gcscs(ctdb_get_cell_id(ctdb),        \
					   (float64) (x)*incr,            \
					   (float64) (y)*incr,            \
					   src_cell,                      \
					   &src_x, &src_y);               \
		  }                                                       \
		  else                                                    \
		  {                                                       \
		      src_x = (float64) (x)*incr;                         \
		      src_y = (float64) (y)*incr;                         \
		  }						          \
		  src_z = (float64) (z);                                  \
		  if (!(state.seg_cnt & 0x2))				  \
		    state.elevs[state.seg_cnt>>2] = src_z;		  \
		  state.segs[state.seg_cnt] = (src_x);		          \
		  state.segs[state.seg_cnt+1] = (src_y);                  \
		  state.seg_cnt += 2;				          \
		  if (state.seg_cnt >= CTDB_MAX_CONTOUR_PTS-4)		  \
		  {							  \
		      draw_fcn(state.seg_cnt, state.segs, state.elevs);	  \
		      state.seg_cnt = 0;			          \
		  }							  \
	      }
		
		if ((max_z - min_z) < CTDB_MAX_CONTOURS_PER_EDGE)
		  for (z = (int32)(max_z) + FUDGE;
		       z >= min_z; z -= 1.0)
		  {
		      if(post & POST_SW_NE)
		      {
			  /* Northwest poly */
			  if ((zSW != zNW) || (zSW != zNE))
			  {
			      /* Direct hit */
			      if (z == zSW)
				STORE(xpD, ypD-1.0, z * interval);
			      
			      /* West line */
			      if (idzW)
			      {
				  dy = (zNW-z)*idzW;
				  if ((dy >= 0.0) && (dy <= 1.0))
				    STORE(xpD, ypD-dy, z * interval);
			      }
			      else if (z == zNW)
			      {
				  STORE(xpD, ypD, z * interval);
			      }
			      
			      /* North line */
			      if (idzN && (z != zNW))
			      {
				  dx = (zNW-z)*idzN;
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD, z * interval);
			      }
			      
			      /* Diagonal line */
			      if (idzD)
			      {
				  dx = (zSW-z)*idzD;
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD-1.0+dx, z * interval);
			      }
			      else if (z == zSW)
			      {
				  STORE(xpD, ypD-1.0, z * interval);
				  STORE(xp+1.0, ypD, z * interval);
			      }
			      
			      /* Back up if we only found one point */
			      if (state.seg_cnt & 0x2)
				state.seg_cnt &= ~0x2;
			  }
			  
			  /* Southeast poly */
			  if ((zSW != zSE) || (zSW != zNE))
			  {
			      /* Direct hit */
			      if (z == zSW)
				STORE(xpD, ypD-1.0, z * interval);
			      
			      /* Diagonal line (again) */
			      if (idzD)
			      {
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD-1.0+dx, z * interval);
			      }
			      
			      /* South line */
			      if (idzS)
			      {
				  dx = (zSW-z)*idzS;
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD-1.0, z * interval);
			      }
			      else if (z == zSW)
			      {
				  STORE(xpD, ypD-1.0, z * interval);
			      }
			      
			      /* East line */
			      if (idzE && (z != zNE))
			      {
				  dy = (zNE-z)*idzE;
				  if ((dy >= 0.0) && (dy <= 1.0))
				    STORE(xpD+1.0, ypD-dy, z * interval);
			      }
			      
			      /* Back up if we only found one point */
			      if (state.seg_cnt & 0x2)
				state.seg_cnt &= ~0x2;
			  }
		      }
		      else
		      {
			  /* Southwest poly */
			  if ((zNW != zSW) || (zNW != zSE))
			  {
			      /* Direct hit */
			      if (z == zNW)
				STORE(xpD, ypD, z * interval);
			      
			      /* West line */
			      if (idzW)
			      {
				  dy = (zNW-z)*idzW;
				  if ((dy >= 0.0) && (dy <= 1.0))
				    STORE(xpD, ypD-dy, z * interval);
			      }
			      else if (z == zSW)
			      {
				  STORE(xpD, ypD-1.0, z * interval);
			      }
			      
			      /* South line */
			      if (idzS && (z != zSW))
			      {
				  dx = (zSW-z)*idzS;
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD-1.0, z * interval);
			      }
			      
			      /* Diagonal line */
			      if (idzD)
			      {
				  dx = (zNW-z)*idzD;
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD-dx, z * interval);
			      }
			      else if (z == zNW)
			      {
				  STORE(xpD, ypD, z * interval);
				  STORE(xp+1.0, yp-1.0, z * interval);
			      }
			      
			      /* Back up if we only found one point */
			      if (state.seg_cnt & 0x2)
				state.seg_cnt &= ~0x2;
			  }
			  
			  /* Northeast poly */
			  if ((zNW != zNE) || (zNW != zSE))
			  {
			      /* Direct hit */
			      if (z == zNW)
				STORE(xpD, ypD, z * interval);
			      
			      /* Diagonal line (again) */
			      if (idzD)
			      {
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD-dx, z * interval);
			      }
			      
			      /* North line */
			      if (idzN)
			      {
				  dx = (zNW-z)*idzN;
				  if ((dx >= 0.0) && (dx <= 1.0))
				    STORE(xpD+dx, ypD, z * interval);
			      }
			      else if (z == zNE)
			      {
				  STORE(xpD+1.0,ypD, z * interval);
			      }
			      
			      /* East line */
			      if (idzE && (z != zNE))
			      {
				  dy = (zNE-z)*idzE;
				  if ((dy >= 0.0) && (dy <= 1.0))
				    STORE(xpD+1.0, ypD-dy, z * interval);
			      }
			      
			      /* Back up if we only found one point */
			      if (state.seg_cnt & 0x2)
				state.seg_cnt &= ~0x2;
			  }
		      }
		  }
	      post = next_post;
	  }
      }
    
    /* Find the contours for microterrain and TEs */
    xpatch_low = xp_low * ctdb->patch_inv_incr;
    ypatch_low = yp_low * ctdb->patch_inv_incr;
    if (ypatch_low)
      ypatch_low--;
    xpatch_high = xp_high * ctdb->patch_inv_incr;
    ypatch_high = yp_high * ctdb->patch_inv_incr;
    for(patch_y=ypatch_low;patch_y<=ypatch_high;patch_y++)
      for(patch_x=xpatch_low;patch_x<=xpatch_high;patch_x++)
      {
	  patch_num = (patch_y * ctdb->patches_wide) + patch_x;
	  
	  feature =
	    ctdb_lookup_feature(ctdb, patch_num);
	  
	  if (feature)
	  {
	      last = feature + feature->hdr.size;
	      for(feature++;feature<last;feature+=next)
	      {
		  int32 verts = feature->info.vertices;
		  int32 type  = (verts & MICRO_TYPE);
		  int32 incr = ctdb->incrD;
		  
		  if (feature->info.feature_class != FC_MICRO)
		    break;
		  
		  next = NEXT_MICRO(verts);
		  
		  if ((type == MICRO_TYPE_WATER) || 
		      ((verts &= MICRO_VERTS) < 2))
		    continue;

		  find_feature_z_bounds(feature,next,&min_z,&max_z);
		  
		  if ((max_z - min_z) < state.max_delta)
		    for (z = interval * floor(max_z * state.inv_interval) + 
			 FUDGE;
			 z >= min_z; z -= interval)
		    {
			g0 = next*feature[4].micro.vert0;
			g1 = next*feature[4].micro.vert1;
		        
			if (ctdb_find_z_on_line(feature[2].xy,
						feature[3].z,
						feature[g1+2].xy,
						feature[g1+3].z,
						patch_x, patch_y, 
						ctdb->patch_incrD,
						z, &x, &y))
			  STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
			
			if (ctdb_find_z_on_line(feature[g1+2].xy,
						feature[g1+3].z,
						feature[g0+2].xy,
						feature[g0+3].z,
						patch_x, patch_y, 
						ctdb->patch_incrD,
						z, &x, &y))
			  STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
			
			if (found = ctdb_find_z_on_line(feature[g0+2].xy,
						       feature[g0+3].z,
						       feature[2].xy,
						       feature[3].z,
						       patch_x, patch_y,
						       ctdb->patch_incrD,
						       z, &x, &y))
			  STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
			
			/* Back up if we only found one point */
			if (state.seg_cnt & 0x2)
			  state.seg_cnt &= ~0x2;
			
			if (verts < 3)
			  continue;
			
			g2 = next*feature[4].micro.vert2;
			
			/* Store the point between f & v0 again */
			if (found)
			  STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
			
			if (ctdb_find_z_on_line(feature[g0+2].xy,
						feature[g0+3].z,
						feature[g2+2].xy,
						feature[g2+3].z,
						patch_x, patch_y, 
						ctdb->patch_incrD,
						z, &x, &y))
			  STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
			
			if (ctdb_find_z_on_line(feature[g2+2].xy,
						feature[g2+3].z,
						feature[2].xy,	
						feature[3].z,
						patch_x, patch_y, 
						ctdb->patch_incrD,
						z, &x, &y))
			  STORE(x,y,ctdb_get_norm_elevation_amsl(ctdb,x,y,z));
			
			/* Back up if we only found one point */
			if (state.seg_cnt & 0x2)
			  state.seg_cnt &= ~0x2;
			
		    }
	      }
	  }
	  
	  /* Draw Contours using terrain elements */
	  if ( ctdb_lookup_te_ptr(ctdb,patch_num) )
	  {
	      int32 member = PG_MEMBER(patch_num);
	      CTDB_PATCH_GROUP *pg = (ctdb->patch_groups + 
				      PG_NUMBER(patch_num));
	      
	      if (((member==0) && (pg->patch0_te16)) ||
		  ((member==1) && (pg->patch1_te16)) ||
		  ((member==2) && (pg->patch2_te16)) ||
		  ((member==3) && (pg->patch3_te16)))
		ctdb_find_te_contours16(ctdb, src_ctdb, patch_num, 
					(float64)FUDGE, &state);
	      else
		ctdb_find_te_contours08(ctdb, src_ctdb, patch_num, 
					(float64)FUDGE, &state);
	  }
      }
    
    draw_fcn(state.seg_cnt, state.segs, state.elevs);
}

/* gcs_box_to_ll_box will take a bounding box defined by two gcs2d
 * points, (xmin,ymin) and (xmax, ymax), and generate a bounding
 * box that encompasses the gcs bounding box, defined by two lat/long
 * points, (latmin,lngmin) and (latmax,lngmax)
 */
static void gcs_box_to_ll_box(int32 cell0, float64 xmin, float64 ymin,
			      float64 xmax, float64 ymax,
			      float64 *latmin, float64 *lngmin, 
			      float64 *latmax, float64 *lngmax)
{
    float64 nw_lat, nw_lng, ne_lat, ne_lng, sw_lat, sw_lng, se_lat, se_lng;
    float64 h;
    
    gcs_gcs_to_llz(cell0, xmin, ymin, 0.0,
		   &sw_lat, &sw_lng, &h, DATUM_WGS84);
    gcs_gcs_to_llz(cell0, xmin, ymax, 0.0,
		   &nw_lat, &nw_lng, &h, DATUM_WGS84);
    gcs_gcs_to_llz(cell0, xmax, ymin, 0.0,
		   &se_lat, &se_lng, &h, DATUM_WGS84);
    gcs_gcs_to_llz(cell0, xmax, ymax, 0.0,
		   &ne_lat, &ne_lng, &h, DATUM_WGS84);
    
    *latmin = (sw_lat < nw_lat) ? sw_lat : nw_lat;
    *lngmin = (sw_lng < se_lng) ? sw_lng : se_lng;
    *lngmax = (nw_lng > ne_lng) ? nw_lng : ne_lng;
    *latmax = (ne_lat > se_lat) ? ne_lat : se_lat;
}

/* gcs_box_to_gcs_box will take a bounding box defined by two gcs2d
 * points in cell0, (x0_min,y0_min) and (x0_max, y0_max), and generates
 * a bounding box that encompasses it, defined by two gcs2d points in cell1,
 * (x1_min, y1_min) and (x1_max,y1_max)
 */
static void gcs_box_to_gcs_box(int32 cell0, float64 x0_min, float64 y0_min,
			       float64 x0_max, float64 y0_max, 
			       int32 cell1, float64 *x1_min, float64 *y1_min, 
			       float64 *x1_max, float64 *y1_max)
{
    float64 nw_x, nw_y, ne_x, ne_y, sw_x, sw_y, se_x, se_y;
    
    if (cell0 == cell1)
    {
	*x1_min = x0_min;
	*x1_max = x0_max;
	*y1_min = y0_min;
	*y1_max = y0_max;
	return;
    }
    
    gcs_gcs2e_to_gcscs(cell0, x0_min, y0_min,
			 cell1, &sw_x, &sw_y);
    gcs_gcs2e_to_gcscs(cell0, x0_min, y0_max,
			 cell1, &nw_x, &nw_y);
    gcs_gcs2e_to_gcscs(cell0, x0_max, y0_min,
			 cell1, &se_x, &se_y);
    gcs_gcs2e_to_gcscs(cell0, x0_max, y0_max,
			 cell1, &ne_x, &ne_y);
    
    *x1_min = (sw_x < nw_x) ? sw_x : nw_x;
    *y1_min = (sw_y < se_y) ? sw_y : se_y;
    *y1_max = (nw_y > ne_y) ? nw_y : ne_y;
    *x1_max = (ne_x > se_x) ? ne_x : se_x;
}

#define MAX_CELLS 100

#define MAX(a,b)   \
  (a > b) ? a : b  \

#define MIN(a,b)   \
  (a < b) ? a : b  \

void ctdb_contour_segments(CTDB *ctdb, float64 interval,
			   int32 x_low, int32 y_low, 
			   int32 x_high, int32 y_high,
			   void (*draw_fcn) (/* ??? */))
{
    if (!ctdb_is_tiled)
      ctdb_contour_segs_in_ctdb(ctdb, ctdb, interval, x_low, y_low, 
				x_high, y_high, draw_fcn);
    else
    {
	int32 n_cell = MAX_CELLS;
	int32 cell_list[MAX_CELLS];
	int32 src_cell = ctdb_get_cell_id(ctdb);
	float64 latmin, lngmin, latmax, lngmax;
	int32 i;
	
	gcs_box_to_ll_box(src_cell, (float64) x_low, (float64) y_low, 
			  (float64) x_high, (float64) y_high,
			  &latmin, &lngmin, &latmax, &lngmax);
	if (!gcs_extent_to_cell_list(latmin, lngmin, latmax, lngmax, &n_cell,
				     cell_list))
	{
	    printf("ctdb_contour_segments: unable to load that many cells\n");
	    return;	
	}
	
	for(i=0; i<n_cell; i++)
	{
	    float64 dest_minx, dest_miny, dest_maxx, dest_maxy;
	    CTDB* my_ctdb = gcs_get_tdb(cell_list[i], FALSE);
	    
	    if (my_ctdb)
	    {
		gcs_box_to_gcs_box(src_cell, (float64) x_low, (float64) y_low, 
				   (float64) x_high, (float64) y_high,
				   cell_list[i], &dest_minx, &dest_miny,
				   &dest_maxx, &dest_maxy);
		
		dest_minx = MAX(my_ctdb->min_x, dest_minx);
		dest_miny = MAX(my_ctdb->min_y, dest_miny);
		dest_maxx = MIN(my_ctdb->max_x, dest_maxx);
		dest_maxy = MIN(my_ctdb->max_y, dest_maxy);
		
		if ((dest_minx > my_ctdb->max_x) ||
		    (dest_miny > my_ctdb->max_y) ||
		    (dest_maxx < my_ctdb->min_x) ||
		    (dest_maxy < my_ctdb->min_y))
		  break;
		
		ctdb_contour_segs_in_ctdb(my_ctdb, ctdb, interval,
					  (int32) dest_minx,
					  (int32) dest_miny,
					  (int32) dest_maxx,
					  (int32) dest_maxy, draw_fcn);
	    }
	}
    }
}
