
#ifndef lint
static char rcsid [] = "$RCSfile: ct_elev.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_elev.c                                                          *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: This file contains all elevation lookup processing.       *
 *       Created:  February 1996                                             *
 *       Author:   cbuettner                                                 *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdext.h> 
#include <libgcs.h>
#include "libctdb_local.h"

#define MICROTYPE_DONE       0
#define MICROTYPE_NOT_DONE   1

#define VERY_NEGATIVE_NUMBER (-1.0E+100)
#define VERY_POSITIVE_NUMBER (1.0E+100)
#define ROAD_BOOST 0.1 /* meters */
/*#define ELEVATION_DEBUG*/

static float64 ctdb_dummy_normal[XYZ];

typedef struct ctdb_lookup_data
{
    int32               featmask;
    CTDB_ELEV_QUALIFIER qual;
    /* On if a default or base micro poly is found.  This indicator
     * prevents the evaluation of the grid or tin surface.
     */    
    int32               terrain_skin_found;
    float64             norm_x; 
    float64             norm_y;
    /* XYO coordinates and valid only if features are evaluated, since it
     * is initialized in the feature loop routine.
     */
    float64             xyo_to_meters;
    CTDB_XYO_COORD2D    xyo;
    /* Data structure contain patch information associated with the current
     * elevation lookup query.
     */
    CTDB_PATCH_DATA     patch;
    CTDB_POST_DATA      post;
    /* start of the local elevation lookup storage array */
    int32               n_elev;
    CTDB_ELEV_DATA     *elev_head;    
    /* pointer to the next element in the elevation array to receive
     * an elevation entry.
     */
    CTDB_ELEV_DATA     *elev;    
} CTDB_LOOKUP_DATA;

static CTDB_ELEV_DATA   ctdb_elev[CTDB_MAX_ELEVS];
static CTDB_FEATURE_OBJ elev_obj[FC_CLASS_MAX];
static CTDB_FEAT_TYPE   
ctdb_elev_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];
static CTDB_FEAT_TYPE   
ctdb_elev_volume_subclass_obj[CTDB_NUM_VOLUME_SUBCLASSES];

static int32
elev_standard_micro(CTDB               *ctdb,
		    CTDB_FEAT_TYPE_PTR  sub,
		    CTDB_FEATURE_DATA  *feature, 
		    CTDB_LOOKUP_DATA   *data
		    );

static int32
elev_water_micro(CTDB               *ctdb,
		 CTDB_FEAT_TYPE_PTR  sub,
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_LOOKUP_DATA   *data
		 );
static void
lookup_tes(
	   CTDB             *ctdb, 
	   CTDB_LOOKUP_DATA *data
	   );

static void
lookup_grid(
	    CTDB             *ctdb, 
	    CTDB_LOOKUP_DATA *data
	    );
    
static void 
ctdb_lookup_engine(
		   CTDB               *ctdb,
		   float64             norm_x, 
		   float64             norm_y,
		   CTDB_ELEV_QUALIFIER qual,
		   CTDB_ELEV_DATA     *elev
		   );

static CTDB_STATUS
init_local_storage_data(ELEVATION_LOOKUP_STORAGE *els)
{
    CTDB_COORD2D frac;
    int32        delta, vg_dim, vgrid, vg_max,vg_x , vg_y ;
    int32        patch_number = els->num; 
    int32        patches_wide = els->ctdb->patches_wide;
    uint16      *vgrid_map;
    
    frac.x          = els->dbpt->x * els->ctdb->meters_to_patch;
    frac.y          = els->dbpt->y * els->ctdb->meters_to_patch; 
    els->coord.x    = PATCH_X(frac.x);
    els->coord.y    = PATCH_Y(frac.y);
    els->num        = els->coord.x + els->coord.y*patches_wide;
    frac.x         -= els->coord.x;
    frac.y         -= els->coord.y;
    delta           = patch_number - els->num; 
    
    if ( patch_number >= 0 && delta )
    {
	if ( delta == -1 )     /* west */
	  frac.x = 1.0 ;
	else if ( delta == 1 ) /* east */
	  frac.x = 0 ;
	else if ( delta == patches_wide ) /* north */
	  frac.y = 0 ;
	else if ( delta == -patches_wide ) /* south */
	  frac.y = 1.0 ;
	else if ( delta == patches_wide-1 )
	{ /* northwest */
	    frac.x = 1.0 ;
	    frac.y = 0;
	}
	else if ( delta == patches_wide+1 )
	{ /* northeast */
	    frac.x = 0;
	    frac.y = 0;
	}
	else if ( delta == -patches_wide+1 ) 
	{ /* southeast */
	    frac.x = 0 ;
	    frac.y = 1.0 ;
	}
	else if ( delta == -patches_wide-1 )
	{ /* southwest */
	    frac.x = 1.0;
	    frac.y = 1.0;
	}
	else /* incorrect patch number */
	  patch_number    = els->num ;
	
	els->num      = patch_number;
    }
    
    els->frac.x   = frac.x;
    els->frac.y   = frac.y;
    els->npt.x    = frac.x*CTDB_PATCH_TO_XYO_SCALE;
    els->npt.y    = frac.y*CTDB_PATCH_TO_XYO_SCALE;
    els->pg       = els->ctdb->patch_groups + PG_NUMBER(els->num);
    els->member   = PG_MEMBER(els->num);
    
    els->te16  = (( els->member == 0 && els->pg->patch0_te16 ) ||
		  ( els->member == 1 && els->pg->patch1_te16 ) ||
		  ( els->member == 2 && els->pg->patch2_te16 ) ||
		  ( els->member == 3 && els->pg->patch3_te16 ));
    els->soil   = CTDB_ILLEGAL_SOIL_TYPE;
    els->where  = OUTSIDE;
    els->front  = els->back = 0 ; /* initialize queue pointers */
    
    /* Do I have terrain elements */
    if (! (els->base = ctdb_lookup_te_ptr(els->ctdb,els->num)) )
      return CTDB_STATUS_NO_TERRAIN_DATA;
    
    /* Note: if TE data available then ctdb_lookup_te_ptr has fetched
     * the patch group data from the cache.
     */
    vg_dim = els->pg->data_start.tin_data->te_vgrid_dim[els->member];
    
    if ( !vg_dim )
      return CTDB_STATUS_NO_VIRTUAL_GRID;
    
    /* Generate the virtual grid index and find the te that it maps
     * to 
     */
    vg_max = vg_dim-1;
    vg_x = (int32)(uint8)(els->frac.x*vg_dim); 
    vg_y = (int32)(uint8)(els->frac.y*vg_dim);
    vgrid   = NS_MIN(vg_x,vg_max) + NS_MIN(vg_y,vg_max) * vg_dim ;
    vgrid_map = (uint16*)((CTDB_FEATURE_DATA*)
			  els->pg->data_start.grid_data +
			  els->pg->data_start.tin_data->
			  te_vgrid_offset[els->member]);
    
    els->mapped_te=els->te_index= vgrid_map[vgrid];
    els->vg_ptr = (uint16 *)&vgrid_map[vgrid];
    
    if ( !els->base )
      return CTDB_STATUS_NO_TERRAIN_DATA;
    
    if ( els->te16 )
    {
	if ( (uint16)els->te_index == ADJT_NOT_MAPPED_16 )
	  return CTDB_STATUS_ILLEGAL_VIRTUAL_GRID_MAPPING ;
    }
    else if ( (uint8)els->te_index == ADJT_NOT_MAPPED_08 )
      return CTDB_STATUS_ILLEGAL_VIRTUAL_GRID_MAPPING ;
   
    return CTDB_STATUS_OK;
    /* Note: els->te will be set later */
}



static CTDB_STATUS
lookup_terrain_element_elevation(ELEVATION_LOOKUP_STORAGE *els)
{
    CTDB_STATUS status=init_local_storage_data(els) ;
    
    if ( !status )
    { 
	if ( els->te16 )
	  return ctdb_lookup_terrain_element_elevation16(els);
	else
	  return ctdb_lookup_terrain_element_elevation08(els);
    }
    return status;
} 

static CTDB_STATUS
ctdb_lookup_terrain_element_elevation(
				      CTDB         *ctdb, 
				      CTDB_COORD2D *pt, 
				      float64      *elevation, 
				      int32        *soil,
				      float64      *normal
				      )
{
    CTDB_STATUS              status=CTDB_STATUS_OK;
    ELEVATION_LOOKUP_STORAGE els;/* Elevation Lookup Storage Data Structure */
    
    if ( !ctdb_point_on_database(ctdb,pt->x,pt->y) )
      status = CTDB_STATUS_POINT_NOT_IN_CELL ; 
    else
    {
	els.ctdb  = ctdb;
	els.dbpt   = pt;
	els.num    = -1;/* Don't override computed patch number */
	els.normal = normal;
	status     = lookup_terrain_element_elevation(&els);
	*elevation = els.elevation;
	*soil      = els.soil;
    }
    
    if ( status &&
	status != CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT)
    {
	*elevation = CTDB_ILLEGAL_ELEVATION;
	*soil      = CTDB_ILLEGAL_SOIL_TYPE;
    }
    
    return(status);
}


CTDB_STATUS
ctdb_lookup_terrain_element_index(
				  CTDB *ctdb,         /* database */
				  CTDB_COORD2D *pt,   /* point in database 
							 coordinates */
				  int32 patch_number, /* force point to be 
							 in this patch.  
							 If negative then 
							 allow the routine to 
							 determine   the patch
							 number */
				  uint32 *te_index    /* OUTPUT parameter - 
							 TE which containes 
							 the pt */
				  )
{
    CTDB_STATUS status=CTDB_STATUS_OK;
    ELEVATION_LOOKUP_STORAGE els;/* Elevation Lookup Storage Data Structure */
    
    if ( !ctdb_point_on_database(ctdb,pt->x,pt->y) )
      status = CTDB_STATUS_POINT_NOT_IN_CELL;
    else 
    {
	els.ctdb   = ctdb;
	els.dbpt   = pt;
	els.num    = patch_number;
	els.normal = NULL;
	status     = lookup_terrain_element_elevation(&els);
	*te_index  = els.te_index;
    }
    
    if ( status &&
	status != CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT )
      *te_index = ADJT_NOT_MAPPED_INT ; 
    
    return(status);
}


static void
ctdb_lookup_feature_elevation_loop(
				   CTDB *ctdb, 
				   CTDB_LOOKUP_DATA *data
				   )
{
    static CTDB_TYPE_FUNC *ml_func=
      &ctdb_elev_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_ML)].func;
    static CTDB_TYPE_FUNC *h2o_func=
      &ctdb_elev_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_WATER)].func;
    static CTDB_TYPE_FUNC *def_func=
     &ctdb_elev_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_DEFAULT)].func;
    static CTDB_TYPE_FUNC *base_func=
      &ctdb_elev_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_BASE)].func;
    CTDB_FEATURE_DATA *feature, *last;
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_ELEV_QUALIFIER feat_qual;
    int32 next;

    feature = ctdb_lookup_feature(ctdb, data->patch.number);

    if (!feature)
      return;

    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can work in integer.
     */
    data->xyo.x = POST_TO_XYO(data->norm_x,data->patch.x,
			      ctdb->patch_inv_incr);
    data->xyo.y = POST_TO_XYO(data->norm_y,data->patch.y,
			      ctdb->patch_inv_incr);

    /* Generate a mask of only the feature class bits */
    feat_qual = (data->qual&CTDB_ALL_FEATURES);  

    obj       = &elev_obj[FC_MICRO];
    /* If you are checking water or the terrain skin make sure that you have 
     * included MICRO terrain in your search.
     */
    if ( data->qual & CTDB_INCL_MICRO )
    {
	obj->valid = TRUE;
	
	if ( data->qual & CTDB_INCL_MICRO_ML )
	  *ml_func = (CTDB_TYPE_FUNC)elev_standard_micro ;
	else
	  *ml_func = (CTDB_TYPE_FUNC)ctdb_invalid_type;
	
	if (data->qual & CTDB_INCL_MICRO_WATER)
	  *h2o_func = (CTDB_TYPE_FUNC)elev_water_micro ;
	else
	  *h2o_func = (CTDB_TYPE_FUNC)ctdb_invalid_type;
	
	if ( data->qual & CTDB_INCL_MICRO_DEF_BASE )
	  *base_func = *def_func  = (CTDB_TYPE_FUNC)elev_standard_micro ;
	else
	  *base_func = *def_func  = (CTDB_TYPE_FUNC)ctdb_invalid_type;
    }
    else
      obj->valid = FALSE; 

    last      = feature + feature->hdr.size;
  
    for(feature++;feature<last;feature+=next)
    {
	int32 fclass=feature->info.feature_class;
	
	/* if the current class doesn't match the previous class it is time 
	 * to change the class object pointer.
	 */
	if ( fclass != obj->sensClass )
	{
	    /* Remove the feature class from the feat_qual list of 
	     * features to be processed.  If there are not more features
	     * to be processed in other feature classes then terminate the
	     * feature elevation lookup processing.  If there are more in
	     * the list then move to the next feature class in the feature
	     * list.
	     */
	    if ( !( feat_qual = (feat_qual & ~obj->inclmask)) )
	      return;

	    /* Switch to the next feature class */
	    obj        = (CTDB_FEAT_OBJ_PTR)&elev_obj[fclass];
	    /* determine if this feature is to be included in this elevation
	     * lookup search.
	     */
	    obj->valid = feat_qual & obj->inclmask;
	}

	/* Call the feature class processing function */
	(void)obj->func(ctdb, obj, feature,(ADDRESS)data,&next);
    }   
}

#define SMIDGE  0.01

static int32
ctdb_gcs_lookup(CTDB               *ctdb,		   
		float64             norm_x, 
		float64             norm_y,
		CTDB_ELEV_QUALIFIER qual,
		CTDB_ELEV_DATA     *elev)
{
    float64 x,y,x_off, y_off, z_off,x_out,y_out;
    int32   id0,id1;
    CTDB *new_ctdb;
    CTDB_ELEV_DATA *elevptr;
    
    x = norm_x * ctdb->incrD;
    y = norm_y * ctdb->incrD;
    id0 = ctdb_get_cell_id(ctdb);	
    /*
     * Guess at elevation 0.0 in my cell frame, put this in GCC
     * and convert to correct cell. There is inherent error
     * since normal vector in that cell is not parallel to
     * normal in current cell. We make the assumption that 
     * elevation lookup will only occur within "reasonable"
     * range of the current cell.
     */
    gcs_gcs3e_to_gcs(id0, x, y, 0.0, &id1, &x_off, &y_off, &z_off);
    new_ctdb = (CTDB *)gcs_get_tdb(id1,FALSE);
    
    if (new_ctdb && (id0 != id1))
    {
	/* convert all elevation into the proper frame of reference */
	for (elevptr = elev;elevptr->sensClass != CTDB_FC_ILLEGAL;elevptr++)
	{
	    ctdb_lookup_engine(new_ctdb, x_off, y_off, qual, elev);
	
	    gcs_gcs3e_to_gcscs(id1, x_off, y_off, z_off,
			       id0, &x_out, &y_out, &(elevptr->z));
	}
	return 1;/* elevation lookup complete */
    }
    return 0;/* elevation lookup not performed */
}

/* Static variable.  When searching for negative surfaces only
   need to lookup at volumes */
static CTDB_ELEV_QUALIFIER ctdb_negative_qualifier =
  (~CTDB_INCL_MICRO_ML & ~CTDB_INCL_LINEAR & ~CTDB_INCL_CANOPY & 
   ~CTDB_INCL_LAID_LINEAR & ~CTDB_INCL_TERRAIN_SKIN & 
   ~CTDB_INCL_MICRO_DEF_BASE & ~CTDB_INCL_MICRO_WATER & 
   ~CTDB_INCL_VEHICLES & ~CTDB_INCL_SINGLE_TREES & 
   ~CTDB_INCL_PATCH_WATER);

static void
ctdb_lookup_engine(CTDB               *ctdb,
		   float64             norm_x, 
		   float64             norm_y,
		   CTDB_ELEV_QUALIFIER qual,
		   CTDB_ELEV_DATA     *elev
		   )
{
    CTDB_LOOKUP_DATA data;
    int32     *soilmap;
    int32     care_about_soils; 
    uint32    i;
				 
    data.n_elev = 0;
    /* Protection from init as nowhere processing */
    if ( !ctdb->patch_groups )
    {
	CTDB_ADD_ELEVATION_DATA(data.n_elev,elev,CTDB_FC_MICRO,CTDB_FS_SKIN,
				0.0,CTDB_ILLEGAL_SOIL_TYPE,ctdb_dummy_normal);
	return;  
    }   
    
    /* Only volumes can have negative surfaces */
    if ( qual & CTDB_ONLY_NEGATIVE_SURFACES)
      qual &= ctdb_negative_qualifier;

    data.norm_x  = norm_x;
    data.norm_y  = norm_y;
    data.post.x  = POST_X(norm_x);
    data.post.y  = POST_Y(norm_y);    
    data.patch.x = PATCH_X(data.post.x * ctdb->patch_inv_incr);
    data.patch.y = PATCH_Y(data.post.y * ctdb->patch_inv_incr);
    
    /* Bounds check */
    if (!((norm_x >= ctdb->min_x_post) &&
	  (norm_x <= (ctdb->max_x_post)) &&
	  (norm_y >= ctdb->min_y_post) &&
	  (norm_y <= (ctdb->max_y_post))))
    {
	float64 adj_norm_x = data.norm_x;
	float64 adj_norm_y = data.norm_y;
	
	if ((norm_x != norm_x) || (norm_y != norm_y))
	{
	    printf("WARNING:  elevation lookup attempted w. NAN coords.\n");
	    CTDB_ADD_ELEVATION_DATA(data.n_elev,elev,CTDB_FC_MICRO,
				    CTDB_FS_SKIN,0.0,
				    CTDB_ILLEGAL_SOIL_TYPE,ctdb_dummy_normal);
	    return; 
	}

	if ((ctdb_is_tiled) && 
	    ctdb_gcs_lookup(ctdb,norm_x,norm_y,qual,elev))
	  return;
	
	/* If we fall through to here, we haven't been able to find any
	 * terrain that covers the point.  Clip it to the nearest point
	 * on the current database, and return the elevation there.
	 */

	if(NS_MIN(norm_x, data.post.x) < ctdb->min_x_post)
	  adj_norm_x = SMIDGE;
	if(NS_MAX(data.post.x, norm_x) >= ctdb->max_x_post)
	  adj_norm_x = ctdb->max_x_post - SMIDGE;
	if(NS_MIN((data.post.y-1), (norm_y-1) < ctdb->min_y_post))
	  adj_norm_y = SMIDGE;
	if((NS_MAX(data.post.y, norm_y) >= ctdb->max_y_post))
	  adj_norm_y = ctdb->max_y_post - SMIDGE;
	
	ctdb_lookup_engine(ctdb, adj_norm_x, adj_norm_y, qual, 
			   elev);
	return;
    }
    
    data.featmask     = 
      ctdb_lookup_feature_mask(ctdb, data.post.x, data.post.y);
    data.patch.number = data.patch.y * ctdb->patches_wide + data.patch.x;
    data.patch.pg     = ctdb->patch_groups+PG_NUMBER(data.patch.number);
    data.patch.member = PG_MEMBER(data.patch.number);
    data.elev_head    = data.elev         = elev;
    data.post.mask    = POST_MASK(data.post.x,data.post.y,data.patch.x,
				  data.patch.y,ctdb->patch_incr);
    data.qual         = qual;

    if (qual & CTDB_COMPUTE_NORMAL)
      data.xyo_to_meters = ctdb->incrD * ctdb->patch_incrD *
	CTDB_XYO_TO_PATCH_SCALE;

    /* Ask if we want to look at soil types, and if so lookup_soilmap. 
     * If we are interested in laid linear, then we are also definitely 
     * interested in soil types.
     */
    if ((data.qual & CTDB_INCL_LAID_LINEAR) || 
	(data.qual & CTDB_INCL_MATERIAL)) 
    {
	care_about_soils = TRUE;
	soilmap = ctdb_lookup_soilmap(ctdb,data.patch.number);
    }
    else 
      care_about_soils = FALSE;

    /* if the micro terrain found is either base or default then this is
     * the terrain skin that will be accepted. (old processing)
     */
    data.terrain_skin_found = 0;    
    /* Note: data.xyo is set in ctdb_lookup_feature_elevation_loop if 
     * there are features in the feature list.  
     */

    /* If there are no features under this post then don't even attempt 
     * a call to process features.
     */
    if ( data.featmask )
    {
	uint32 mask=0;
      
	/* reconstruct a feature mask from the input qualifiers */
	if ( data.qual & CTDB_INCL_VOLUME )
	  mask |= POST_BUILDINGS;
	
	if ( data.qual & CTDB_INCL_CANOPY )
	  mask |= POST_TREES;
	
	if ( data.qual & CTDB_INCL_MICRO )
	  mask |= POST_MICROTERRAIN;
	
	if ( data.qual & CTDB_INCL_LAID_LINEAR )
	{
	    mask |= POST_FEATURES;
	    /* If there is a request for laid linears then always include
	     * the material since a request without materials doesn't 
	     * make sense.
	     */
	    data.qual |= CTDB_INCL_MATERIAL;
	}
	
	/* are any of the features of interest in the post region */
	if ( mask & data.featmask )
	  ctdb_lookup_feature_elevation_loop(ctdb, &data);
	
	if (care_about_soils)
	{
	    for (i=0; i<data.n_elev; i++)
	    {
		if (!((data.elev[i].sensClass == CTDB_FC_VOLUME) && 
		      (data.elev[i].subclass == CTDB_FS_MES)))
		{
		    /* MES store indices directly into the PAT; all other
		     * features go through the patch soil tables
		     */
		    elev[i].material = soilmap[elev[i].material];
		}
	    }
	    
	}
    } 
      
    /* Process patch water */  
    if ( data.qual & CTDB_INCL_WATER )
    {
	static float32 z;
	if ( ctdb_lookup_patch_water_level(ctdb,
					   data.patch.number,
					   data.norm_x,
					   data.norm_y,
					   &z) )
	{
	    float64 *normal, tmp_normal[XYZ];
	    
	    /* Note that this assumes that PAT index 4 is deep water,
	     * which should continue to be true for the foreseeable future.
	     */
	    if ( data.qual & CTDB_COMPUTE_NORMAL)
	    {
		gcs_lookup_gravity_vector(ctdb->cell_id, 
					  data.norm_x*ctdb->incr,
					  data.norm_y*ctdb->incr,
					  z,
					  tmp_normal);
		if (tmp_normal[Z] < 0.0)
		  vmat3_negate64(tmp_normal,tmp_normal);

		normal = tmp_normal;
	    }
	    else
	      normal = ctdb_dummy_normal;
	    
	    CTDB_ADD_ELEVATION_DATA(data.n_elev,data.elev,CTDB_FC_MICRO,
				    CTDB_FS_WATER,z,/*SOIL_DEEP_WATER*/4,
				    normal);
	}
    } 
    
    /* If there was no default or base micro terrain found in the feature
     * list under this point, look at the terrain skin.  A patch is either
     * TINned or gridded.  
     */
    if (!data.terrain_skin_found &&
	(data.qual & CTDB_INCL_TERRAIN_SKIN) )
    {
	if ( ctdb->is_tinned && 
	     CTDB_PATCH_HAS_TES(data.patch.pg,data.patch.member))
	{
	  lookup_tes(ctdb, &data);
	  if (care_about_soils)
	    elev[data.n_elev-1].material =
	      soilmap[elev[data.n_elev-1].material]; 
	}
	else 
	  lookup_grid(ctdb, &data);
    }
}

static void
lookup_laid_linear(
		   CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj,
		   CTDB_FEATURE_DATA  *feature,
		   CTDB_LOOKUP_DATA   *data,
		   int32              *next
		   )
{   
    static float64 normal[XYZ]={0,0,1};/* straight up for now*/    
    int32 radius,i;
    int32 verts = feature->info.vertices;
    *next = NEXT_LAID_LINEAR(verts);

    if ( !obj->valid || 
	!(data->post.mask & feature->integer))
       return;

    /* Check the segments connecting the two */
    radius = feature[2].laid_linear.width>>1;
    for (i=0;i<verts-1;i++)
    {
	/* Are we on this one? */
	if (ctdb_point_intersection(feature[i+3].xy, feature[i+4].xy,
				    data->xyo, radius) == INTERSECTION)
	{
	    CTDB_ADD_ELEVATION_DATA(data->n_elev,data->elev, 
				    CTDB_FC_LAID_LINEAR,  
				    CTDB_FS_NA, 0,
				    feature[1].ll_soil.poly_char_index,
				    normal);
	}
    }
    /* Although there may be multiple laid linears at one point (road and 
     * river), we are assured by the compiler that roads will preceed other
     * laid linears in the feature list.  Since the only time we care about
     * laid linears is when we are looking for soil types and past processing
     * has stated that roads are considered more interesting than rivers, 
     * further processing to find other laid linears is wasteful and
     * unnecessary. 
     */
    obj->valid = FALSE;
}

static void
lookup_linear(
	      CTDB               *ctdb, 
	      CTDB_FEAT_OBJ_PTR   obj,
              CTDB_FEATURE_DATA  *feature,
	      CTDB_LOOKUP_DATA   *data,
	      int32              *next
	      )
{
    /* Treelines will not be evaluated in this routine but will be addressed
     * in the canopy processing.
     */
    static int32 verts;
    verts = feature->info.vertices;
    *next = NEXT_LINEAR(verts);
    return;
}

static void
lookup_volume(
	      CTDB               *ctdb, 
	      CTDB_FEAT_OBJ_PTR   obj,
              CTDB_FEATURE_DATA  *feature,
	      CTDB_LOOKUP_DATA   *data,
	      int32              *next
	      )
{
    static int32 verts;
    CTDB_FEAT_TYPE    *sub;

    verts = feature->info.vertices;
    *next = NEXT_VOLUME(verts);

    /* if the caller doesn't want this type of feature or the feature
     * is not resident at this post then return.
     */
    if (!( obj->valid ) || !( data->post.mask & feature->integer ))
       return;

    sub = &ctdb_elev_volume_subclass_obj[ feature[1].volume.type ];
    if ( sub->func(ctdb,sub,feature,data) == MICROTYPE_DONE )
      obj->valid = FALSE;
	
    return;
}

static int32
lookup_standard_volume(CTDB               *ctdb,
		       CTDB_FEAT_TYPE_PTR  sub,
		       CTDB_FEATURE_DATA  *feature, 
		       CTDB_LOOKUP_DATA   *data
		       )
{
    CTDB_POINT_STATE hit;
    float64 z;
    static float64 normal[XYZ]={0,0,1};

    if ((hit = ctdb_point_in_building(feature, data->xyo.x, data->xyo.y, &z))
	== OUTSIDE)
      return MICROTYPE_NOT_DONE;
    
    /* No need to find the exact elevation.  The elevation of the
     * first point should be fine. If the point was inside of the 
     * building.  If the point is on a vertex the item z will
     * have been set by ctdb_point_in_building.
     */
    if (hit == TRIANGLE0 )
      z = feature[4].z;

    CTDB_ADD_ELEVATION_DATA(data->n_elev,data->elev,sub->tcid, 
			    sub->tsid,z,CTDB_ILLEGAL_SOIL_TYPE,normal);	
    
    /* Once you have found the feature below this point, don't look 
     * for anymore of this class.
     */
    return MICROTYPE_NOT_DONE;
}

/* This route expects a sorted list of elevation intersections.  The 
 * reason sorted is required is to allow filtering out of identical
 * surface intersection that may occur at vertices and/or shared edges.
 */
static int32  
mes_store_surfaces(CTDB                  *ctdb,
		   int32                 *n_elev,
		   CTDB_ELEV_DATA       **elev,
		   int32                  n_mes_inter, 
		   CTDB_MES_INTERSECTION *mes_inter,
		   CTDB_MES_VOLUME       *mes_volume,
		   float64                z_start,
		   float64                z_delta,
		   int32                  pos_flag, 
		   int32                  neg_flag,
		   float64                cosine_limit)
{
    int32    stored, add_elevation, i;
    CTDB_MES_TRIANGLE  *triangle;
    float64             normal[XYZ];
    
    stored = FALSE;
    
    for ( i=0; i<n_mes_inter; i++ )
    {
	/* Filter out normals which are not requested by the caller */
	triangle = 
	  &(mes_volume->template->all_triangles[mes_inter[i].triangle]);
	vmat3_vec_mat_mul64 (triangle->normal, mes_volume->tfm_world_to_mes, 
			     normal);
    
	if (pos_flag && !neg_flag)
	  add_elevation = ( normal[Z] >= cosine_limit );
	else if (!pos_flag && neg_flag)
	  add_elevation = ( normal[Z] <= cosine_limit );
	else
	  add_elevation = TRUE;
	
	if ( add_elevation )
	{
	    float64 z,last_z=1e-200;
	    
	    stored = TRUE;
	    z = z_start + mes_inter[i].distance*z_delta;
#define MILLIMETER 0.001
	    if (fabs(z-last_z) > MILLIMETER)
	    {
		CTDB_ADD_ELEVATION_DATA((*n_elev), (*elev), CTDB_FC_VOLUME, 
					CTDB_FS_MES, z, 
					triangle->material, normal );
		last_z = z;
	    }
	}
    }

    return stored;
}

static int32
lookup_mes_volume(CTDB               *ctdb,
		  CTDB_FEAT_TYPE_PTR  sub,
		  CTDB_FEATURE_DATA  *feature, 
		  CTDB_LOOKUP_DATA   *data)
{
    uint32              which_volume_model; 
    uint32              which_mes_volume;
    CTDB_VOLUME_MODEL  *volume_model;
    CTDB_MES_VOLUME    *mes_volume;
    CTDB_MES_TEMPLATE  *mes_template;
    float64             mes_top[XYZ];
    CTDB_COORD3D        top;
    float64             mes_bottom[XYZ]; 
    CTDB_COORD3D        bottom;
    CTDB_COORD3D        start;
    float64             ray_origin[XYZ];
    CTDB_COORD3D        end;
    float64             ray_destination[XYZ];
    int32               positive_surfaces, negative_surfaces;
    int32               n_mes_inter;
    CTDB_MES_INTERSECTION *mes_inter;

    positive_surfaces = data->qual & CTDB_ONLY_POSITIVE_SURFACES;
    negative_surfaces = data->qual & CTDB_ONLY_NEGATIVE_SURFACES;
    which_volume_model = feature[1].volume.reference;

    if ((data->n_elev >= (CTDB_MAX_ELEVS-1)) ||
	(which_volume_model >= ctdb->num_mes_volumes) ||
	(positive_surfaces && negative_surfaces))
      return MICROTYPE_NOT_DONE;

    volume_model = ctdb->mes_volume_models[which_volume_model];
    which_mes_volume = volume_model->mes_volume_index;

    if ( which_mes_volume >= ctdb->num_mes_volumes ) /* ERROR*/
      return MICROTYPE_NOT_DONE;  

    mes_volume = &(ctdb->mes_volumes[which_mes_volume]);
    mes_template = mes_volume->template;
    mes_top[X] = mes_top[Y] = 0.0;
    mes_top[Z] = 
      (float64)(mes_template->all_boxes[0].right_back_top_corner[Z]);
    ctdb_mes_to_xyo(ctdb, mes_volume, mes_top[X],mes_top[Y],mes_top[Z],
		    data->patch.x, data->patch.y, 
		    &top.x, &top.y, &top.z );

    mes_bottom[X] = mes_bottom[Y] = 0.0;
    mes_bottom[Z] = 
      (float64)(mes_template->all_boxes[0].left_front_bottom_corner[Z]);
    ctdb_mes_to_xyo(ctdb, mes_volume, 
		    mes_bottom[X],mes_bottom[Y],mes_bottom[Z],
		    data->patch.x, data->patch.y, 
		    &bottom.x, &bottom.y, &bottom.z );

    start.x = (float64)(data->xyo.x);
    start.y = (float64)(data->xyo.y);
    start.z = top.z + ( top.z - bottom.z );
    ctdb_xyo_to_mes(ctdb, mes_volume, data->patch.x, data->patch.y,
		    start.x, start.y, start.z,
	            &ray_origin[X],&ray_origin[Y],&ray_origin[Z]);

    end.x = (float64)(data->xyo.x);
    end.y = (float64)(data->xyo.y);
    end.z = bottom.z + ( bottom.z - top.z );
    ctdb_xyo_to_mes(ctdb, mes_volume, data->patch.x, data->patch.y,
                    end.x, end.y, end.z, &ray_destination[X],
                    &ray_destination[Y],&ray_destination[Z]);

    ctdb_mes_line_intersection(mes_volume,
			       ray_origin,
			       ray_destination,
			       TRUE,
			       CTDB_MES_ALL_ENCLOSURES,
			       FALSE,
			       TRUE,
			       FALSE,
			       &n_mes_inter,
			       &mes_inter);

    if (!mes_store_surfaces(ctdb, &data->n_elev, &data->elev, n_mes_inter, 
			    mes_inter, mes_volume, start.z, (end.z - start.z),
			    positive_surfaces, negative_surfaces, 0.0))
      return MICROTYPE_NOT_DONE;

    /* Once you have found the feature below this point, don't look
     * for anymore of this class.
     */
    return MICROTYPE_DONE;
}

static void
lookup_canopy(
	      CTDB               *ctdb, 
	      CTDB_FEAT_OBJ_PTR   obj,
	      CTDB_FEATURE_DATA  *feature,
	      CTDB_LOOKUP_DATA   *data,
	      int32              *next
	      )
{
    static int32 incr, verts;
    CTDB_POINT_STATE hit;
    float64 z,magic;
    float64 normal[XYZ];

    verts = feature->info.vertices;
    incr  = *next = NEXT_CANOPY(verts);

    /* if the caller doesn't want this type of feature or the feature
     * is not resident at this post then return.
     */
    if (!( obj->valid ) || 
	!( data->post.mask & feature->integer ))
       return;

    /* This works just like microterrain.  Look there for
     * comments.
     */
    if ((hit = ctdb_point_in_canopy_feature(feature, data->xyo.x, 
					    data->xyo.y, incr, &z)) == OUTSIDE)
      return;
    
    if ( (hit == TRIANGLE0) || (hit == TRIANGLE1))
    {
	ctdb_compute_canopy_feature_normal(feature, (hit == TRIANGLE1), incr,
					   &normal[X], &normal[Y], 
					   &normal[Z], &magic);
	/* First make sure this isn't a vertical polygon. */
	if (normal[Z] == 0.0)
	  z = feature[2].z;
	else
	  /* Compute our elevation by the equation:
	   * n(x)*x + n(y)*y + n(z)*z = n(x)*x0 + n(y)*y0 + n(z)*z0
	   */
	  z = (magic - normal[X] * (float64)data->xyo.x - 
	       normal[Y]*(float64)data->xyo.y) / normal[Z];
    }

    if (data->qual & CTDB_COMPUTE_NORMAL)
    {
	/* transform the vector from xyo frame of reference to the database
	 * frame of reference. 
	 */
	normal[Z] *= data->xyo_to_meters;
	vmat3_unit64(normal,normal);
	/* All canopy surfaces are considered to be positive.  Enforce this
	 * assumption instead ensuring that the vertices are in the proper
	 * order.
	 */
	if (normal[Z] < 0.0)
	  vmat3_negate64(normal,normal);
    }
    
    
    CTDB_ADD_ELEVATION_DATA(data->n_elev,data->elev,CTDB_FC_CANOPY,CTDB_FS_NA,
			    z,CTDB_ILLEGAL_SOIL_TYPE,normal);  
    /* Once you have found the feature below this point, don't look 
     * for anymore of this class.
     */
    obj->valid = FALSE;
}

static void
lookup_aggregate(
		 CTDB               *ctdb, 
		 CTDB_FEAT_OBJ_PTR   obj,
		 CTDB_FEATURE_DATA  *feature,
		 CTDB_LOOKUP_DATA   *data,
		 int32              *next
		 )
{
    /* Currently, aggregate features only contain linears, which are not
     * evaluated.
     */
    *next = NEXT_AGGREGATE(0);
    return;
}

static int32
elev_water_micro(CTDB               *ctdb,
		 CTDB_FEAT_TYPE_PTR  sub,
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_LOOKUP_DATA   *data
		 )
{ 
    float32 offset;
    CTDB_POINT_STATE hit;
    int32 soil;
    float64 z,normal[XYZ],magic;

    /* note: ctdb_point_in_micro_feature will exit early if the number of 
     * points are less than 2.
     */

    /* Are we in this one? */
    if ((hit = ctdb_point_in_micro_feature(feature, data->xyo.x, data->xyo.y, 
					   sub->incr, &z)) == OUTSIDE)
      return MICROTYPE_NOT_DONE;
    
    /* Optimization note:  if-else-if chains are still faster than
     * switch-case, even when the values look like they could drive
     * a lookup table.
     */
    
    /* If there was a direct hit, the raw elev is all set and we need to 
     * find out how much this point should be shifted by due to tides.
     */

    /* It is assumed that the soil for micro water is 4 (SOIL_DEEP_WATER)
     * as defined by the s1000_def.h global include file.
     */
    if ((hit == TRIANGLE1) || (hit == VERTEX3))
    {
	soil = feature[1].micro_soil.poly_char_index1;
	offset = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
				     feature[5].mwater.char1);
    }
    else
    {
	soil = feature[1].micro_soil.poly_char_index0;
	offset = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
				     feature[5].mwater.char0);
    }

    /* We're inside this one.
     * Compute a normal <n(x) n(y) n(z)> and a magic number 
     * n(x)*x0 + n(y)*y0 + n(z)*z0
     */
    if (hit == TRIANGLE0)
    {
	ctdb_compute_micro_feature_normal(feature, 0, sub->incr,
					  &normal[X], &normal[Y], 
					  &normal[Z], &magic);
	/* First make sure this isn't a vertical polygon. */
	if (normal[Z] == 0.0)
	  z = feature[3].z;
	else
	{
	    /* Compute our elevation by the equation:
	     * n(x)*x + n(y)*y + n(z)*z = n(x)*x0 + n(y)*y0 + n(z)*z0
	     */
	    z = (magic - normal[X] * (float64)data->xyo.x - 
		 normal[Y] * (float64)data->xyo.y) / normal[Z];
	}
    }
    else if ( hit == TRIANGLE1 )
    {
	ctdb_compute_micro_feature_normal(feature, 1, sub->incr,
					  &normal[X], &normal[Y], 
					  &normal[Z], &magic);
	/* First make sure this isn't a vertical polygon. */
	if (normal[Z] == 0.0)
	  z = feature[3].z;
	else
	{
	    /* Compute our elevation by the equation:
	     * n(x)*x + n(y)*y + n(z)*z = n(x)*x0 + n(y)*y0 + n(z)*z0
	     */
	    z = (magic - normal[X] * (float64)data->xyo.x - 
		 normal[Y] * (float64)data->xyo.y) / normal[Z];
	}
    }
    
    if (data->qual & CTDB_COMPUTE_NORMAL)
    {
	/* transform the vector from xyo frame of reference to the database
	 * frame of reference. 
	 */
	normal[Z] *= data->xyo_to_meters;
	vmat3_unit64(normal,normal);
	/* All canopy surfaces are considered to be positive.  Enforce this
	 * assumption instead ensuring that the vertices are in the proper
	 * order.
	 */
	if (normal[Z] < 0.0)
	  vmat3_negate64(normal,normal);
    }
    
    CTDB_ADD_ELEVATION_DATA(data->n_elev,data->elev, sub->tcid, sub->tsid,
			    (z+offset),soil,normal);

    /* We assume that there is only one level of micro water at a single
     * point in the database.  By removing the water bit from the input 
     * qualifier we can potentially terminate out of the microterrain
     * processing early if there are no other micro terrain types to 
     * process.
     */
    data->qual &= ~CTDB_INCL_MICRO_WATER;
    /* unmap the micro water processing */
    return MICROTYPE_DONE;
}


static int32
elev_standard_micro(CTDB               *ctdb,
		    CTDB_FEAT_TYPE_PTR  sub,
		    CTDB_FEATURE_DATA  *feature, 
		    CTDB_LOOKUP_DATA   *data
		    )
{ 
    CTDB_POINT_STATE hit;
    int32 soil;
    float64 z,magic;
    float64 normal[XYZ];

    /* note: ctdb_point_in_micro_feature will exit early if the number of 
     * points are less than 2.
     */

    /* Are we in this one? */
    if ((hit = ctdb_point_in_micro_feature(feature, data->xyo.x, data->xyo.y, 
					   sub->incr, &z)) == OUTSIDE)
      return MICROTYPE_NOT_DONE;
    
    /* Optimization note:  if-else-if chains are still faster than
     * switch-case, even when the values look like they could drive
     * a lookup table.
     */
    
    if ((hit != TRIANGLE0) && (hit != TRIANGLE1))
      soil = feature[1].micro_soil.poly_char_index0;
    else
    {
	/* We're inside this one.
	 * Compute a normal <n(x) n(y) n(z)> and a magic number 
	 * n(x)*x0 + n(y)*y0 + n(z)*z0
	 */
	if (hit == TRIANGLE0)
	{
	    soil = feature[1].micro_soil.poly_char_index0;
	    ctdb_compute_micro_feature_normal(feature, 0, sub->incr,
					      &normal[X], &normal[Y], 
					      &normal[Z], &magic);
	}
	else		/* hit == TRIANGLE1 */
	{
	    soil = feature[1].micro_soil.poly_char_index1;
	    ctdb_compute_micro_feature_normal(feature, 1, sub->incr,
					      &normal[X], &normal[Y], 
					      &normal[Z], &magic);
	}
	
	/* We treat microterrain polys with the illegal soil type as if it's
	 * not there.
	 */
	if(soil == CTDB_ILLEGAL_SOIL_TYPE)
	  return MICROTYPE_NOT_DONE;

	/* First make sure this isn't a vertical polygon. */
	if (normal[Z] == 0.0)
	  z = feature[3].z;
	else
	{
	    /* Compute our elevation by the equation:
	     * n(x)*x + n(y)*y + n(z)*z = n(x)*x0 + n(y)*y0 + n(z)*z0
	     */
	    z = (magic - normal[X] * (float64)data->xyo.x - 
		    normal[Y] * (float64)data->xyo.y) / normal[Z];
	}
    }
    
    if (data->qual & CTDB_COMPUTE_NORMAL)
    {
	/* transform the vector from xyo frame of reference to the database
	 * frame of reference. 
	 */
	normal[Z] *= data->xyo_to_meters;
	vmat3_unit64(normal,normal);
	/* All canopy surfaces are considered to be positive.  Enforce this
	 * assumption instead ensuring that the vertices are in the proper
	 * order.
	 */
	if (normal[Z] < 0.0)
	  vmat3_negate64(normal,normal);
    }
    
    CTDB_ADD_ELEVATION_DATA(data->n_elev, data->elev, sub->tcid, sub->tsid, 
			    z, soil, normal);

    /* If the micro type was default or base then we have completed the 
     * search for this microtype.  If it is multilevel microterrain then
     * keep looking since microterrain is used for a various purposes and 
     * multiple microtype multi can be defined at one point.
     */

    if (sub->tsid == CTDB_FS_MULTI)
      return MICROTYPE_NOT_DONE;
    else
    {
	data->terrain_skin_found = TRUE;
	return MICROTYPE_DONE;
    }
}


static void
lookup_micro(
	      CTDB               *ctdb, 
	      CTDB_FEAT_OBJ_PTR   obj,
	      CTDB_FEATURE_DATA  *feature,
	      CTDB_LOOKUP_DATA   *data,
	      int32              *next
	      )
{ 
    static CTDB_FEAT_TYPE *sub;
    int32 verts = feature->info.vertices;

    /* check if the micro type has changed */
    sub = &ctdb_elev_micro_subclass_obj[MICRO_TYPE_INDEX(verts)];

    *next = sub->incr;
    
    if (!obj->valid || 
	!(data->post.mask & feature->integer) ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return;

    /* call the micro type processing function - if we found a micro
     * terrain element of this subtype then map the subtype function
     * pointer to invalid to prevent lookup at other microterrain of
     * the same subtype.
     */
    if ( sub->func(ctdb,sub,feature,data) == MICROTYPE_DONE)
    {
	/* If there are other types of micro to look at continue this 
	 * processing.
	 */
	if ( data->qual & CTDB_INCL_MICRO )
	  sub->func = (CTDB_TYPE_FUNC)ctdb_invalid_type;
	else
	  obj->valid = FALSE;
    }
}


static void
lookup_tes(CTDB             *ctdb, 
	   CTDB_LOOKUP_DATA *data
	   )
{
    float64 normal[XYZ];
    float64 *norm;
    
    if ( ctdb_lookup_te_ptr(ctdb,data->patch.number) )
    {
	float64 z;
	int32   material;
	CTDB_COORD2D pt;
	pt.x = data->norm_x * ctdb->incr; 
	pt.y = data->norm_y * ctdb->incr;
	
	if (data->qual & CTDB_COMPUTE_NORMAL)
	  norm = normal;
	else
	  norm = NULL;
	
	if ( !ctdb_lookup_terrain_element_elevation(ctdb,&pt,&z,&material,
						    norm))
	{
	    if (!norm)
	      norm = ctdb_dummy_normal;
	    
	    CTDB_ADD_ELEVATION_DATA(data->n_elev,data->elev,CTDB_FC_MICRO,
				    CTDB_FS_SKIN, z, material, norm);
	}
    }    
}

#define S_VECTOR  0.0, -ctdb->incrD, (pzS-pz)
#define SE_VECTOR ctdb->incrD, -ctdb->incrD, (pzSE-pz)
#define E_VECTOR  ctdb->incrD, 0, (pzE-pz)

static void
lookup_grid(CTDB             *ctdb, 
	    CTDB_LOOKUP_DATA *data
	    )
{
    int32 elev_post, soil_post;
    float64 pz,dx,dy,pzS,pzE,pzSE,normal[XYZ];
    int32 soil;
    int32 do_soil;

    elev_post = ctdb_lookup_elev_post(ctdb, data->post.x, data->post.y);
    soil_post = ctdb_lookup_soil_post(ctdb, data->post.x, data->post.y);

    pz   = GET_POST_ELEVATION(elev_post, ctdb->fixed_point_basis);
    
    dx = (data->norm_x - (float64)data->post.x);
    dy = ((float64)data->post.y - data->norm_y);
    
    if (!dx)
    {
	/* Is the point on a vertical edge? */
	if ( data->qual & CTDB_INCL_MATERIAL )  
	{
	    if (!(elev_post & POST_SW_NE))
	      soil = GET_POST_SW_SOIL(soil_post);
	    else
	      soil = GET_POST_NW_SOIL(soil_post);
	}

	if (data->qual & CTDB_COMPUTE_NORMAL)
	{
	    pzS = ctdb_lookup_safe_post_elevation(ctdb, data->post.x, 
						   data->post.y-1);
	    if (elev_post & POST_SW_NE)
	      vmat3e_cross_prod64(S_VECTOR, E_VECTOR, normal);
	    else
	      vmat3e_cross_prod64(S_VECTOR, SE_VECTOR, normal);
	}
	
	/* Is the point on the post? */
	if ( dy )
	{
	    /* Find the elevation of the next post South */
	    pzS = ctdb_lookup_safe_post_elevation(ctdb, data->post.x, 
						  data->post.y-1);
	    /* Interpolate linearly along the y axis */

	    pz += ((pzS - pz) * dy);
	}
    }
    else if (!dy)
    {
	/* Is the point on a horizontal edge? */
	if ( data->qual & CTDB_INCL_MATERIAL )
	{
	    if (!(elev_post & POST_SW_NE))
	      soil = GET_POST_NE_SOIL(soil_post);
	    else
	      soil = GET_POST_NW_SOIL(soil_post);
	}

	/* Find the elevation of the next post East */
	pzE = ctdb_lookup_safe_post_elevation(ctdb, data->post.x+1, 
					      data->post.y);
	if (data->qual & CTDB_COMPUTE_NORMAL)
	{
	    if (elev_post & POST_SW_NE)
	      vmat3e_cross_prod64(S_VECTOR , E_VECTOR, normal);
	    else
	      vmat3e_cross_prod64(SE_VECTOR, E_VECTOR, normal);
	}
	/* Interpolate linearly along the x axis */
	pz += ((pzE - pz) * dx);	
    }
    else if (!(elev_post & POST_SW_NE))
    {
	/* Find the elevation of the next post Southeast */
	pzSE = ctdb_lookup_safe_post_elevation(ctdb, data->post.x+1, 
					       data->post.y-1);
	
	/* Is the point in the leftmost poly? */
	if (dx <= dy)
	{
	    if ( data->qual & CTDB_INCL_MATERIAL )
	      soil = GET_POST_SW_SOIL(soil_post);

	    /* Find the elevation of the next post South */
	    pzS = ctdb_lookup_safe_post_elevation(ctdb, data->post.x, 
						  data->post.y-1);
	    if (data->qual & CTDB_COMPUTE_NORMAL)
	      vmat3e_cross_prod64(S_VECTOR, SE_VECTOR, normal);

	    pz += ((pzSE - pzS) * dx + (pzS - pz) * dy);
	}
	else
	{
	    if ( data->qual & CTDB_INCL_MATERIAL )
	      soil = GET_POST_NE_SOIL(soil_post);

	    /* Find the elevation of the next post East */
	    pzE = ctdb_lookup_safe_post_elevation(ctdb, data->post.x+1, 
						  data->post.y);
	    if (data->qual & CTDB_COMPUTE_NORMAL)
	      vmat3e_cross_prod64(SE_VECTOR, E_VECTOR, normal);
	    
	    pz += ((pzE - pz) * dx + (pzSE - pzE) * dy);
	}
    }
    else
    {
	/* Find the elevations of the next post South & East */
	pzS = ctdb_lookup_safe_post_elevation(ctdb, data->post.x, 
					      data->post.y-1);
	pzE = ctdb_lookup_safe_post_elevation(ctdb, data->post.x+1, 
					      data->post.y);
	
	/* Is the point in the leftmost poly? */
	if (dx <= (1.0-dy))
	{
	    if ( data->qual & CTDB_INCL_MATERIAL )
	      soil = GET_POST_NW_SOIL(soil_post);

	    if (data->qual & CTDB_COMPUTE_NORMAL)
	      vmat3e_cross_prod64(S_VECTOR, E_VECTOR, normal);
	    
	    pz += ((pzE - pz) * dx + (pzS - pz) * dy);
	}
	else
	{
	    if ( data->qual & CTDB_INCL_MATERIAL )
	      soil = GET_POST_SE_SOIL(soil_post);
	    
	    /* Find the elevation of the next post Southeast */
	    pzSE = ctdb_lookup_safe_post_elevation(ctdb, data->post.x+1, 
						   data->post.y-1);
	    if (data->qual & CTDB_COMPUTE_NORMAL)
	      vmat3e_cross_prod64(/*SW*/-ctdb->incrD, -ctdb->incrD, (pzS-pzE),
				  /*S*/ 0, -ctdb->incrD, (pzSE-pzE),
				normal);
	    pz = pzSE + (pzS - pzSE) * (1.0-dx) + (pzE - pzSE) * (1.0-dy);
	}
    }
    
    if (data->qual & CTDB_COMPUTE_NORMAL)
      vmat3_unit64(normal,normal);

    CTDB_ADD_ELEVATION_DATA(data->n_elev, data->elev,
			    CTDB_FC_MICRO, CTDB_FS_SKIN, pz, soil, normal);
}


int32
ctdb_elev_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static CTDB_TYPE_FUNC all_micro_func[CTDB_NUM_MICRO_SUBCLASSES] =
    {
	(CTDB_TYPE_FUNC)elev_standard_micro,   /* MICRO_TYPE_BASE */ 
	(CTDB_TYPE_FUNC)elev_standard_micro,   /* MICRO_TYPE_ML */
	(CTDB_TYPE_FUNC)elev_standard_micro,   /* MICRO_TYPE_DEFAULT */
	(CTDB_TYPE_FUNC)elev_water_micro       /* MICRO_TYPE_WATER */
    };

    static CTDB_TYPE_FUNC all_vol_func[CTDB_NUM_VOLUME_SUBCLASSES] =
    {
	(CTDB_TYPE_FUNC)lookup_standard_volume,   
	(CTDB_TYPE_FUNC)lookup_standard_volume,   
	(CTDB_TYPE_FUNC)lookup_mes_volume,
	(CTDB_TYPE_FUNC)lookup_standard_volume,   
    };

    ctdb_init_micro_subclass(ctdb_elev_micro_subclass_obj,all_micro_func);
    obj[FC_MICRO].func         = (CTDB_OBJ_FUNC)lookup_micro; 

    ctdb_init_volume_subclass(ctdb_elev_volume_subclass_obj,all_vol_func);
    obj[FC_VOLUME].func        = (CTDB_OBJ_FUNC)lookup_volume;

    obj[FC_LINEAR].func        = (CTDB_OBJ_FUNC)lookup_linear;
    obj[FC_CANOPY].func        = (CTDB_OBJ_FUNC)lookup_canopy;
    obj[FC_LAID_LINEAR].func   = (CTDB_OBJ_FUNC)lookup_laid_linear;
    obj[FC_AGGREGATE].func     = (CTDB_OBJ_FUNC)lookup_aggregate;
    bcopy(obj,elev_obj,sizeof(elev_obj));
    
    return 1;
}

static CTDB_ELEV_DATA *
ctdb_lookup_norm_elevation_ref(
			       CTDB               *ctdb, 
			       float64             norm_x, 
			       float64             norm_y, 
			       CTDB_ELEV_QUALIFIER qual,
			       float64             reference_elevation)
{
    CTDB_ELEV_DATA *elevptr,*ml,*ll_ptr;
    int32   material;
    float64 ml_diff;

    /* Initial the condition of not getting a valid elevation */
    ctdb_elev[0].sensClass    = CTDB_FC_ILLEGAL;
    ctdb_elev[0].material = 0;   
    
    ctdb_lookup_engine(ctdb,norm_x,norm_y,qual,ctdb_elev);

    ml_diff = VERY_POSITIVE_NUMBER;
    ll_ptr  = (CTDB_ELEV_DATA *)NULL;
 
    /* initialize the ml elevation pointer to the first element in the 
     * list.  This will guarentee that this routine will always exit
     * with the first element in the list even if it is mapped to 
     * an element that contain a class of CTDB_FC_ILLEGAL (null elevation
     * list i.e. no valid elevation found in this query).
     */
    for (ml=elevptr=ctdb_elev;elevptr->sensClass>CTDB_FC_ILLEGAL;elevptr++)
    {
	/* We are guarenteed that there will only be one laid linear 
	 * in the returned list.  See laid linear feature processing
	 * for explanation.
	 */
	if (elevptr->sensClass == CTDB_FC_LAID_LINEAR)
	{
	    ll_ptr = elevptr;
	    continue;
	}

	/* Return the smallest delta range from the reference elevation
	 * that is a positive delta.
	 */
	if (elevptr->z <= reference_elevation)
	{
	    static float64 diff;
	    /* Give me the first surface below this reference point */
	    diff = reference_elevation - elevptr->z;
	    
	    /* Select the closest positive elevation */
	    if ( diff < ml_diff )
	    {
		ml      = elevptr;
		ml_diff = diff;
	    }
	}
    }

    /* Special processing to handle laid linears (roads) since they
     * don't have an elevation associated with them and they area assumed
     * to be placed on the terrain skin
     */

    if ( ll_ptr    &&                    /* laid linears encountered */
	(ml->sensClass == CTDB_FC_MICRO) &&  /* highest pt is terrain */
	(ml->subclass == CTDB_FS_SKIN )) /* highest pt is terrain skin*/
    {
	/* change the soil type of the terrain skin to be that of the
	 * laid linear.
	 */
	ml->material = ll_ptr->material;

	/* Minor kludge. Roads actually sit 0.1 meters ABOVE 
	 * terrain skin.
	 */
	material = ctdb->pat_table_ptr->columns[0].data[ml->material].integer;
	if ( CTDB_ROAD(material) )
	{
	    ml->subclass = CTDB_FS_ROAD;
	    ml->z += ROAD_BOOST;
	}
	else if ( CTDB_WATER(material) )
	  ml->subclass = CTDB_FS_WATER;
    }
    else if (!(qual & CTDB_INCL_MATERIAL))
      ml->material = CTDB_ILLEGAL_SOIL_TYPE;
    
    return ml;
} /* end of lookup_norm_elevation_ref */

static void
ctdb_lookup_norm_elevation_mes(
			       CTDB               *ctdb, 
			       float64             norm_x, 
			       float64             norm_y,
			       CTDB_ELEV_QUALIFIER qual,
			       CTDB_ELEV_DATA     *elev
			       )
{
    CTDB_ELEV_DATA *elevptr, *ll_ptr,*skin;
    int32  material;
    
    /* $$$ Currently don't populate the elev->norm_vec field in the 
     * output data structure.
     */

    /* for mes lookups always include soils/materials */
    ctdb_lookup_engine(ctdb, norm_x, norm_y, qual|CTDB_INCL_MATERIAL, 
		       elev);

    ll_ptr = (CTDB_ELEV_DATA *)NULL;
    /* initialize the skin elevation pointer to the first element in the 
     * list.  This will guarentee that this routine will always exit
     * with the first element in the list even if it is mapped to 
     * an element that contain a class of CTDB_FC_ILLEGAL (null elevation
     * list i.e. no valid elevation found in this query).
     */
    /* find all laid linears and the terrain skin */
    for (skin=elevptr=elev;elevptr->sensClass>CTDB_FC_ILLEGAL;elevptr++)
    {
	/* We are guarenteed that there will only be one laid linear 
	 * in the returned list.  See laid linear feature processing
	 * for explanation.
	 */
	if (elevptr->sensClass == CTDB_FC_LAID_LINEAR)
	{
	    ll_ptr = elevptr;
	    continue;
	}

	/* Find the highest terrain skin element */
	if ((elevptr->sensClass == CTDB_FC_MICRO) && 
	    (elevptr->subclass == CTDB_FS_SKIN) &&
	    (skin->z < elevptr->z))
	  skin  = elevptr;    
    }

    /* Add the elevation information to the laid linear */  
    if ( ll_ptr )
    {
	material =
	  ctdb->pat_table_ptr->columns[0].data[ll_ptr->material].integer;
	if ( CTDB_ROAD(material) )
	{
	    ll_ptr->subclass = CTDB_FS_ROAD;
	    ll_ptr->z        = skin->z + ROAD_BOOST;
	}
	else if ( CTDB_WATER(material) ) 
	{
	    ll_ptr->subclass = CTDB_FS_WATER;  
	    ll_ptr->z        = skin->z;
	}
    }
} /* end of lookup_norm_elevation_mes */
 

CTDB_ELEV_DATA *
ctdb_lookup_norm_elevation_max(
			       CTDB               *ctdb, 
			       float64             norm_x, 
			       float64             norm_y,
			       CTDB_ELEV_QUALIFIER qual
			       )
{
    CTDB_ELEV_DATA *elevptr, *maxi,*ll_ptr;
    int32 material;
 
    /* Initial the condition of not getting a valid elevation */
    ctdb_elev[0].sensClass = CTDB_FC_ILLEGAL;
    ctdb_elev[0].material = 0;   

    ctdb_lookup_engine(ctdb, norm_x, norm_y, qual, ctdb_elev);
    
    ll_ptr = (CTDB_ELEV_DATA *)NULL;
    
    /* initialize the maxi elevation pointer to the first element in the 
     * list.  This will guarentee that this routine will always exit
     * with the first element in the list even if it is mapped to 
     * an element that contain a class of CTDB_FC_ILLEGAL (null elevation
     * list i.e. no valid elevation found in this query).
     */
    for (maxi=elevptr=ctdb_elev;elevptr->sensClass>CTDB_FC_ILLEGAL;elevptr++)
    {
	/* We are guarenteed that there will only be one laid linear 
	 * in the returned list.  See laid linear feature processing
	 * for explanation.
	 */
	if (elevptr->sensClass == CTDB_FC_LAID_LINEAR)
	{
	    ll_ptr = elevptr;
	    continue;
	}
	
	if ( elevptr->z > maxi->z )
	  maxi = elevptr;
    }
    
    /* Special processing to handle laid linears (roads) since they
     * don't have an elevation associated with them and they area assumed
     * to be placed on the terrain skin
     */
    if (ll_ptr &&
	(maxi->sensClass  == CTDB_FC_MICRO) && 
	(maxi->subclass == CTDB_FS_SKIN ))
    {
	/* Minor kludge. Roads actually sit 0.1 meters ABOVE terrain
	 * skin.  Only boost terrain skin elevation for the road.
	 */
	material =
	  ctdb->pat_table_ptr->columns[0].data[ll_ptr->material].integer;
        if (CTDB_ROAD(material))         
	{
	    maxi->z += ROAD_BOOST; 
	    maxi->subclass = CTDB_FS_ROAD;
	}
	else if (CTDB_WATER(material))
	  maxi->subclass = CTDB_FS_WATER;

	maxi->material = ll_ptr->material;
    }
    else if (!(qual & CTDB_INCL_MATERIAL))
      maxi->material = CTDB_ILLEGAL_SOIL_TYPE; 

    return maxi;
} /* end of lookup_norm_elevation_max */

float64 
ctdb_lookup_norm_elevation(
			   CTDB   *ctdb, 
			   float64 norm_x, 
			   float64 norm_y,
			   int32   buildings,
			   int32   canopies,
			   int32  *soil,
			   float64 reference
			   )
{
    CTDB_ELEV_QUALIFIER qual;
    static CTDB_ELEV_DATA *elev;
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_QUALIFIER canopy=(CTDB_INCL_CANOPY |
				       CTDB_INCL_LINEAR );
    static CTDB_ELEV_QUALIFIER soils=(CTDB_INCL_MATERIAL|
				      CTDB_INCL_LAID_LINEAR);
    
     
    if (ctdb)
    {
	qual = (dflt | 
		(buildings?CTDB_INCL_VOLUME:0) |
		(canopies?canopy:0)  |
		(soil?soils:0));

	elev = ctdb_lookup_norm_elevation_ref(ctdb,norm_x,norm_y,qual,
					      reference);

	if (soil)
	  *soil = elev->material;

	return elev->z;
    }
    else
    {
	if (soil)
	  *soil = CTDB_ILLEGAL_SOIL_TYPE;

	return 0.0;
    }   
}



    


/* This function computes the elevation above mean sea level.
 * First, it computes the elevation z at the normalized point
 * (xn, yn) and then calculates the distance from that point to
 * the ellipsoid.
 */
float64 
ctdb_lookup_norm_elevation_amsl(
				CTDB               *ctdb, 
				float64             xn, 
				float64             yn,
				CTDB_ELEV_QUALIFIER qual
				)
{
    static CTDB_ELEV_DATA *elev;
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    
    elev = ctdb_lookup_norm_elevation_max(ctdb, 
					  xn, 
					  yn,
					  (qual?qual:dflt));
					  
    return(ctdb_get_norm_elevation_amsl(ctdb, xn, yn, elev->z));
}


/* This function compoutes the elevation above mean sea level
 * of a point (xn, yn, z), where xn and yn are normalized
 * coordinates.  Elevation amsl is along the gravity vector.
 *
 * Note: If we are running under SIMNET mode, z is also the
 * elevation amsl.  Otherwise, we calculate this elevation
 * by converting to lat/long/z.
 */

float64
ctdb_get_norm_elevation_amsl(
			     CTDB *ctdb, 
			     float64 xn, 
			     float64 yn, 
			     float64 z
			     )
{
    float64 lat, lng, h;
    
    if ( ctdb->gcs_mode == CTDB_MODE_SIMNET )
      return z; /* Flat earth UTM database where z is the height above 
		 * mean sea level.
		 */
    else
    {
	gcs_gcs_to_llz(ctdb_get_cell_id(ctdb), (float64)xn*ctdb->incrD, 
		       (float64)yn*ctdb->incrD, z, 
		       &lat, &lng, &h, ctdb->datum);
	return h;
    }
}



/* --------------------------------------------------------------------------
 *                    PUBLIC API FUNCTIONS 
 * --------------------------------------------------------------------------
 */
float64 
ctdb_lookup_elevation_ml(
			 CTDB *ctdb, 
			 float64 x, 
			 float64 y, 
			 float64 reference_elevation
			 )
{   
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_DATA *elev;
    
    if ( !ctdb )
      return 0;
    else
    {
	elev = ctdb_lookup_norm_elevation_ref(ctdb,
					      (float64)(x*ctdb->inv_incr),
					      (float64)(y*ctdb->inv_incr),
					      dflt,
					      reference_elevation);
	return elev->z;
    }    
}


float64 
ctdb_lookup_elevation(CTDB *ctdb, float64 x, float64 y)
{
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_DATA *elev;
    
    if (ctdb)
    {
	elev = ctdb_lookup_norm_elevation_max(ctdb,
					      (float64)(x*ctdb->inv_incr),
					      (float64)(y*ctdb->inv_incr),
					      dflt);
	return elev->z;
    }
    else
      return 0.0;
}

float64 
ctdb_lookup_max_elevation(
			  CTDB   *ctdb, 
			  float64 x, 
			  float64 y, 
			  int32   check_canopy
			  )
{   
    static CTDB_ELEV_DATA *elev;
    static CTDB_ELEV_QUALIFIER dflt=(CTDB_MAX_LAND_AND_WATER |
				     CTDB_INCL_VOLUME);
    static CTDB_ELEV_QUALIFIER canopy=(CTDB_INCL_LINEAR |
				       CTDB_INCL_CANOPY);
    
    if (ctdb)
    {
	elev =  ctdb_lookup_norm_elevation_max(ctdb, 
					       (float64)(x * ctdb->inv_incr),
					       (float64)(y * ctdb->inv_incr),
					       (dflt |(check_canopy?canopy:0))
					       );
	return elev->z;				       
    }    
    else
      return 0.0;    
} 


int32 
ctdb_lookup_soil(CTDB *ctdb, float64 x, float64 y)
{
    static CTDB_ELEV_DATA *elev;
    static CTDB_ELEV_QUALIFIER dflt=(CTDB_MAX_LAND_AND_WATER   |
				     CTDB_INCL_LAID_LINEAR     |
				     CTDB_INCL_MATERIAL);
    if (ctdb)
    {
	int32 row_in_pat;
	elev = ctdb_lookup_norm_elevation_max(ctdb, 
					      (float64)(x*ctdb->inv_incr),
					      (float64)(y*ctdb->inv_incr),
					      dflt);
	row_in_pat = (int32)elev->material;
	return ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
    }
    else
      return CTDB_ILLEGAL_SOIL_TYPE;
}

/* Fill out an array of attribute values, one for each attribute specified 
 * by attr_list, whose length is num_attributes.  ret_list[i] is the value
 * for attribute attr_list[i], and it's interpretation is attribute-specific.
 * It is the *caller's* responsibility to allocate space for ret_list.  If
 * any requested attribute is not in the PAT, the routine will return
 * FALSE , otherwise it will return TRUE.  Callers making this call repeatedly
 * are thus expected to remember which attributes are available to avoid
 * repeated failures.  
 */
int32   
ctdb_lookup_attributes( 
		       CTDB          *ctdb,
		       int32          row_in_pat,
		       uint32         attr_list[],
		       uint32         num_attributes,
		       CTDB_PAT_DATA  ret_list[]
		       )
{
    int32     false_flag, *found_attribute;
    uint32    attribute, i, j;

    if (ctdb)
    {	
	found_attribute = ctdb_alloc(NULL, sizeof(int32) * num_attributes,
				     "Found Attribute Array");
	for (i=0; i < num_attributes; i++)
        {
	    attribute = attr_list[i];
	    found_attribute[i] = 0;    
	    for (j=0; j < ctdb->pat_table_ptr->num_columns; j++)
	    {
		if (attribute == ctdb->pat_table_ptr->columns[j].facc)
		{
		    found_attribute[i] = 1;  /* attribute was found in PAT */
		    ret_list[i] = ctdb->pat_table_ptr->columns[j].
		      data[row_in_pat];
		    j = ctdb->pat_table_ptr->num_columns; /* so I exit loop */
		}
	    }
	}

	false_flag = 0;
	for (i=0; i < num_attributes; i++)
	{
	    if (found_attribute[i] == 0)
	    {
		false_flag = 1;
		i = num_attributes;  /* so I exit loop */
	    }
	}
	if (false_flag == 1)
	  return FALSE;
	else 
	  return TRUE;
    }
    else 
      return FALSE;
}   
 
/* Fills out the attr_list array with the FACC code for each attribute
 * stored in the PAT of ctdb.  The caller is responsible for allocating the
 * space for attr_list, and should fill out *num_attributes with the length
 * of the allocated array.  The routine will overwrite this value with the
 * actual number of attributes present.  If there are too many attributes to
 * fit in the allocated space, the routine returns FALSE, otherwise it 
 * returns TRUE. 
 */
int32
ctdb_lookup_pat_attributes(
			   CTDB   *ctdb,
			   uint32  attr_list[],
			   uint32 *num_attributes
			   )  
{  
    if (*num_attributes < ctdb->pat_table_ptr->num_columns)
      return FALSE;
    else
    {
	uint32 i;
	for (i=0; i<ctdb->pat_table_ptr->num_columns; i++)
	    attr_list[i] = ctdb->pat_table_ptr->columns[i].facc;
	*num_attributes = ctdb->pat_table_ptr->num_columns;
	return TRUE;
    }
}

int32 
ctdb_lookup_soil_ml(
		    CTDB *ctdb, 
		    float64 x, 
		    float64 y, 
		    float64 reference_elevation
		    )
{
    static CTDB_ELEV_DATA *elev;    
    static CTDB_ELEV_QUALIFIER dflt=(CTDB_MAX_LAND_AND_WATER   |
				     CTDB_INCL_LAID_LINEAR     |
				     CTDB_INCL_MATERIAL );    
    if ( ctdb )
    {
	int32 row_in_pat;
	elev = ctdb_lookup_norm_elevation_ref(ctdb,
					      (float64)(x*ctdb->inv_incr),
					      (float64)(y*ctdb->inv_incr),
					      dflt,
					      reference_elevation);
	row_in_pat = (int32)elev->material;
	return ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
    }
    else
      return CTDB_ILLEGAL_SOIL_TYPE;
}

float64
ctdb_lookup_elevation_amsl(
			   CTDB *ctdb, 
			   float64 x, 
			   float64 y
			   )
{
    float64 z;
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    
    z = ctdb_lookup_norm_elevation_amsl(ctdb,
					(float64)(x*ctdb->inv_incr),
					(float64)(y*ctdb->inv_incr),
					dflt
					);
    
    return ctdb_get_norm_elevation_amsl(ctdb, x, y, z);
}

 
void
ctdb_lookup_elevation_mes(
			  CTDB               *ctdb, 
			  float64             x, 
			  float64             y,
			  CTDB_ELEV_QUALIFIER qual,
			  CTDB_ELEV_DATA      elev[CTDB_MAX_ELEVS]
			  )
{
    int32  row_in_pat;
    
    /* These are the defualt flags for this API routine */
    static CTDB_ELEV_QUALIFIER dflt=(CTDB_LAND_AND_WATER   |
				     CTDB_INCL_VOLUME      |
				     CTDB_INCL_MATERIAL );    
    /* These are the flags which are valid for this API routine */
    static CTDB_ELEV_QUALIFIER valid=(CTDB_ALL_FEATURES     |
				      CTDB_LAND_AND_WATER   |
				      CTDB_INCL_MATERIAL    |
				      CTDB_ALL_SURFACES     |
				      CTDB_COMPUTE_NORMAL);
    /* These are the flags which are maniditory for this API routine */
    static CTDB_ELEV_QUALIFIER must=(CTDB_INCL_MATERIAL |
				     CTDB_COMPUTE_NORMAL);

    elev[0].sensClass = CTDB_FC_ILLEGAL;

    if (ctdb)
      ctdb_lookup_norm_elevation_mes(ctdb, 
				     (float64)(x * ctdb->inv_incr),
				     (float64)(y * ctdb->inv_incr),
				     (qual?((qual|must)&valid):dflt),
				     elev
				     );

    /* This routine must return the old mobility word in the material field
     * for each elevation, so go retrieve it based on the PAT index.
     */

    for (;
	 elev->sensClass != CTDB_FC_ILLEGAL;
	 elev++)
    {
	row_in_pat = elev->material;
	elev->material =
	  ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
    }
}

/* This API is the same as ctdb_lookup_elevation_mes except for the
 * fact that the material fields of elev contain the row in the PAT
 * and not the value in the PAT.
 */
void
ctdb_lookup_attributes_mes(
			   CTDB               *ctdb, 
			   float64             x, 
			   float64             y,
			   CTDB_ELEV_QUALIFIER qual,
			   CTDB_ELEV_DATA      elev[CTDB_MAX_ELEVS]
			   )
{  
    /* These are the defualt flags for this API routine */
    static CTDB_ELEV_QUALIFIER dflt=(CTDB_LAND_AND_WATER   |
				     CTDB_INCL_VOLUME      |
				     CTDB_INCL_MATERIAL );    
    /* These are the flags which are valid for this API routine */
    static CTDB_ELEV_QUALIFIER valid=(CTDB_ALL_FEATURES     |
				      CTDB_LAND_AND_WATER   |
				      CTDB_INCL_MATERIAL    |
				      CTDB_ALL_SURFACES     |
				      CTDB_COMPUTE_NORMAL);
    /* These are the flags which are maniditory for this API routine */
    static CTDB_ELEV_QUALIFIER must=(CTDB_INCL_MATERIAL     |
				     CTDB_COMPUTE_NORMAL);

    elev[0].sensClass = CTDB_FC_ILLEGAL;

    if (ctdb)
      ctdb_lookup_norm_elevation_mes(ctdb, 
				     (float64)(x * ctdb->inv_incr),
				     (float64)(y * ctdb->inv_incr),
				     (qual?((qual|must)&valid):dflt),
				     elev
				     );
}

int32
ctdb_lookup_enclosure_elevation(CTDB                *ctdb,
				CTDB_MES_NODE_KEY    enclosure,
				float64              x,
				float64              y,
				float64              maximum_angle,
				CTDB_ELEV_QUALIFIER  qualifier,
				CTDB_ELEV_DATA       elevations[
						   CTDB_MAX_ENCLOSURE_ELEVS] )
{
    uint16              mes_id;
    uint16              enclosure_id;
    CTDB_MES_VOLUME    *mes_volume;
    CTDB_MES_TEMPLATE  *mes_template;
    float64             ray_origin[XYZ];
    float64             ray_destination[XYZ];
    CTDB_MES_INTERSECTION *mes_inter=NULL;
    int32               n_mes_inter;
    CTDB_ELEV_DATA      elev[CTDB_MAX_ELEVS], *elevptr;
    int32               positive_surfaces;
    int32               negative_surfaces;
    int32 n_elev;

    enclosure_id = CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure);
    mes_id       = CTDB_MES_NODE_KEY_TO_MES_ID(enclosure);
    positive_surfaces = qualifier & CTDB_ONLY_POSITIVE_SURFACES;
    negative_surfaces = qualifier & CTDB_ONLY_NEGATIVE_SURFACES;

    /* Initial the condition of not getting a valid elevation */
    elev[0].sensClass    = elevations[0].sensClass    = CTDB_FC_ILLEGAL;
    elev[0].material = elevations[0].material = CTDB_ILLEGAL_SOIL_TYPE;
    n_elev = 0;
    
    if (( mes_id >= ctdb->num_mes_volumes ) ||
	( positive_surfaces && negative_surfaces ))
      return n_elev;

    mes_volume   = &(ctdb->mes_volumes[mes_id]);
    mes_template = mes_volume->template;
#define Z_START 2000.0
#define Z_END  -2000.0    
    ctdb_gcs_to_mes(mes_volume, ctdb_get_cell_id(ctdb), x, y, Z_START,
		    &ray_origin[X], &ray_origin[Y], &ray_origin[Z]);
    ctdb_gcs_to_mes(mes_volume, ctdb_get_cell_id(ctdb), x, y, Z_END,
		    &ray_destination[X], &ray_destination[Y], 
		    &ray_destination[Z]);
		    
    ctdb_mes_line_intersection(mes_volume, ray_origin, ray_destination,
			       TRUE, enclosure_id, FALSE, TRUE, TRUE,
			       &n_mes_inter, &mes_inter);

    /* If there are no intersection default processing to return the terrain 
     * skin.  This is required for bridges that don't have floors defined in 
     * the lower enclosure.
     */
    n_elev = 0;
    elevptr = elev;
    mes_store_surfaces(ctdb, &n_elev, &elevptr, n_mes_inter, mes_inter, 
		       mes_volume, Z_START, (Z_END - Z_START),
		       positive_surfaces, negative_surfaces, 
		       cos(DEG_TO_RAD(maximum_angle)));
    
    /* Failure condition processing: too many intersections or no 
     * intersections.
     */
    if ( n_elev > CTDB_MAX_ENCLOSURE_ELEVS )
    {
	ctdb_sort_elev_data(elev);
	n_elev = CTDB_MAX_ENCLOSURE_ELEVS;
	elev[n_elev].sensClass = CTDB_FC_ILLEGAL;
	elev[n_elev].material = CTDB_ILLEGAL_SOIL_TYPE;
    }
    else if (( n_elev == 0 ) && 
	     (CTDB_MES_NODE_IS_EXTERIOR(enclosure) ||
	      (mes_template->type == CTDB_MES_BRIDGE)))
    {
	static int32 bridge_qual=(CTDB_COMPUTE_NORMAL |
		                  CTDB_MAX_LAND_AND_WATER);
	
	qualifier |= bridge_qual;
        ctdb_lookup_qual_elevation (ctdb, x, y, qualifier, CTDB_NO_REF_ELEV, 
				    elevations );
	elevations[1].sensClass = CTDB_FC_ILLEGAL;
	return (elevations[0].sensClass != CTDB_FC_ILLEGAL);
    }
    
    if (n_elev)
      bcopy(&elev[0],&elevations[0],(n_elev+1)*sizeof(CTDB_ELEV_DATA));
    return n_elev;
}

float64
ctdb_lookup_qual_elevation(
			   CTDB               *ctdb,
			   float64             x,
			   float64             y,
			   CTDB_ELEV_QUALIFIER qual,
			   float64             ref,
			   CTDB_ELEV_DATA      elev[1]
			   )
{
    int32  row_in_pat;
    
    static CTDB_ELEV_DATA *lcl_elev;
    static CTDB_ELEV_QUALIFIER dflt  = CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_QUALIFIER valid = (CTDB_LAND_AND_WATER      |
					CTDB_RETURN_MAX          |
					CTDB_RETURN_MIN          |
					CTDB_ALL_FEATURES        |
					CTDB_INCL_MATERIAL       |
					CTDB_ALL_SURFACES        |
					CTDB_COMPUTE_NORMAL);

    if (!ctdb)
      return 0.0;   
    else
    {
	/* set up the appropriate processing.  Note: If MINIMUM is not 
	 * requested then MAXIMUM is assumed.
	 */
	if ( qual & CTDB_RETURN_MIN )
	{
	    CTDB_ELEV_DATA *elevptr;
	    
	    ctdb_lookup_norm_elevation_mes(ctdb,
					   (x*ctdb->inv_incr),
					   (y*ctdb->inv_incr),
					   (qual?(qual&valid):dflt),
					   ctdb_elev);

	    /* The material fields of ctdb_elev
	     * need to be set to the value in the PAT, not the 
	     * row in the PAT, which the internal routines return.
	     */ 
	    for (lcl_elev=elevptr=ctdb_elev;
		 elevptr->sensClass != CTDB_FC_ILLEGAL;
		 elevptr++)
	    {
		if ( elevptr->z < lcl_elev->z )
		{
		    row_in_pat = elevptr->material;
		    elevptr->material = 
		      ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
		    lcl_elev = elevptr;
		}
		
	    }
	}
	else
	{
	    if ( qual & CTDB_RETURN_MAX )
	      ref = CTDB_ACCEPT_ML;
	    
	    lcl_elev = 
		  ctdb_lookup_norm_elevation_ref(ctdb,
						 (x*ctdb->inv_incr),
						 (y*ctdb->inv_incr),
						 (qual?(qual&valid):dflt),
						 ref);

	    /* The material fields of lcl_elev
	     * need to be set to the value in the PAT, not the 
	     * row in the PAT which the internal routines return.
	     */
	    row_in_pat = lcl_elev->material;
	    lcl_elev->material =
	      ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;   
	}
	
	/* If an elev pointer is passed then copy the final elev data
	 * structure into the output pointers address.
	 */
	if ( elev != (CTDB_ELEV_DATA *)NULL )
	  bcopy(lcl_elev,elev,sizeof(CTDB_ELEV_DATA));
	
	return lcl_elev->z;
    }
}

static int32
cmp_elev(CTDB_ELEV_DATA *elev0, CTDB_ELEV_DATA *elev1)
{
    if (elev0->z < elev1->z)
      return 1;
    else if (elev1->z < elev0->z)
      return -1;
    else
      return 0;
}

/* In the future, we might want to have a key passed for the type of sorting 
 * required 
 */
void
ctdb_sort_elev_data(CTDB_ELEV_DATA elev[])
{
    int32 n_elev;

    /* Determine the number of elements in the array to sort */
    for (n_elev=0;elev[n_elev].sensClass!=CTDB_FC_ILLEGAL;n_elev++);

    /* sort'em */
    (void)qsort(elev, n_elev,sizeof(CTDB_ELEV_DATA),(CTDB_QSORT)cmp_elev);
}



