/*
 * $RCSfile: ct_export.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: ct_export.h                                                       *
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Driver               *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7612                                                        *
*                                                                           *
*       This software was developed by TASC and SAIC under U. S. Government *
*       contracts 
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Exported "private" definitions used by libxcig            *
*       Created: Thu Aug. 17 1995                                           *
*       Author: flchamberlain    (TASC)                                     *
*       Author: cbuettner        (SAIC)                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libctdb.h>
#include <ct_feat.h>
#include <ct_post.h>
#ifndef _CT_EXPORT_H_INCLUDED
#define _CT_EXPORT_H_INCLUDED
#define CTDB_ACCEPT_ML 1.0E+6
#define CTDB_REJECT_ML -1.0E+6
#define CTDB_ILLEGAL_ELEVATION 0         
#define CTDB_ILLEGAL_SOIL_TYPE 0         

#define PG_NUMBER(patch) ((patch) >> 2)  

#define PG_MEMBER(patch) ((patch) & 0x3) 


/* This macro will check the patch group header for patch water and return
 * true if the patch, defined by member in the patch group, has patch water
 * polygons.
 */
#define CTDB_HAS_WATER_PATCH(pg, member)        \
    ((((member)==0) && ((pg)->patch0_water)) || \
     (((member)==1) && ((pg)->patch1_water)) || \
     (((member)==2) && ((pg)->patch2_water)) || \
     (((member)==3) && ((pg)->patch3_water)))

/* This macro will convert a value in patch relative xyo coordinates 
 * to database coordinates.  patch_incrD and incrD are defined in the 
 * ctdb data structure, x is the point in xyo coordinates and patchx is
 * the patch coordinate given by the macro PATCH_X or PATCH_Y.  
 */
#define CTDB_XYO_TO_METERS(x,patch_x,patch_incrD,incrD) \
((incrD * patch_incrD) * ((patch_x) +                   \
			  ((float64)(x) * CTDB_XYO_TO_PATCH_SCALE)))

/* This macro will convert a value in database coordinates to patch relative 
 * xyo coordinates (patch offset coordinates).  meters_to_patch is defined in 
 * the ctdb data structure, x is the point in db coordinates (meters) and 
 * patchx is the patch coordinate given by the macro PATCH_X or PATCH_Y.  
 */
#define CTDB_METERS_TO_XYO(x,patchx,meters_to_patch)     \
(((x)*(meters_to_patch)-patchx)*CTDB_PATCH_TO_XYO_SCALE)

/*============MULTI-ELEVATION SURFACE STRUCTURES============*/

#define CTDB_MAX_SCALING 1.073741e+09

typedef int32    CTDB_MES_VERTEX[XYZ];
typedef float64  CTDB_MES_VERTEX_F64[XYZ];

typedef struct ctdb_mes_list_vertex_f64
{
    int32                 num_vertices;
    CTDB_MES_VERTEX_F64  *vertices;
} CTDB_MES_LIST_VERTEX_F64 ;

typedef struct ctdb_mes_triangle
{
#ifdef USE_LITTLE_ENDIAN
    uint16        padding;
    int16         material;	/* Index into CTDB PAT */
#else
    int16         material;	/* Index into CTDB PAT */
    uint16        padding;
#endif
    float64       normal[XYZ];
    int32         vertex_ids[3];
} CTDB_MES_TRIANGLE;

typedef struct ctdb_mes_aperture
{
    int32      num_triangles;
    int32      first_triangle_id;  /* The first in the set of triangles for 
                                    * this aperture */
    
    struct
    {
	int32   num_vertices; 
	int32  *vertex_ids;        /* Vertex ids that make up the exterior
			            * boundary of the aperture, clockwise order
			            */
    } outline;
    struct
    {
#ifdef USE_LITTLE_ENDIAN
	uint8      padding;
	uint8      step_to_enclosure;
	uint16     enclosure_id;
#else
        uint16     enclosure_id; /* An enclosure the aperture connects to */

    /* Whether there is a step from the bottom of this aperture to the floor
     * of the corresponding enclosure.  A value of TRUE means that the bottom 
     * of this aperture is at a different level from the floor of the 
     * enclosure, FALSE means that they are at the same level. */
        uint8      step_to_enclosure; 
	uint8      padding;
#endif
    /* The thickness of the wall from the plane of the aperture to the 
     * corresponding enclosure.  The two wall-thicknesses together add up
     * to the total thickness of the wall. */
        int32      wall_thickness;
    } connections[2];

#ifdef USE_LITTLE_ENDIAN
    int16      open;   
    int16      visible;
#else
    int16      visible;           /* 1 = transparent, 0 = opaque */
    int16      open;              /* 1 = open, 0 = closed */
#endif
} CTDB_MES_APERTURE; 

typedef struct ctdb_mes_grid_box
{
   int32           left_front_bottom_corner[XYZ];
   int32           right_back_top_corner[XYZ];
   int32           outer_box_id;      /* parent of this box */

#ifdef USE_LITTLE_ENDIAN
   uint16          dimensions_x;      /* boxes in x direction */
   uint16          grid_index;        /* which child is this box */
   uint16          dimensions_z;      /* boxes in z direction */
   uint16          dimensions_y;      /* boxes in y direction */
#else
   uint16          grid_index;        /* which child is this box */
   uint16          dimensions_x;      /* boxes in x direction */
   uint16          dimensions_y;      /* boxes in y direction */
   uint16          dimensions_z;      /* boxes in z direction */
#endif
   int32          *box_ids;           /* array of children of this box */

#ifdef USE_LITTLE_ENDIAN
   uint16          padding1;
   uint16          num_apertures;  
#else
   uint16          num_apertures;  
   uint16          padding1;
#endif

   uint16         *aperture_ids;      /* ids of apertures intersecting 
	 	    		       * this box */
#ifdef USE_LITTLE_ENDIAN
   uint16          padding2;
   uint16          num_triangles;  
#else
   uint16          num_triangles;  
   uint16          padding2;
#endif
   int32          *wall_triangle_ids; /* ids of enclosure wall triangles
				       * intersecting this box */
} CTDB_MES_GRID_BOX;

typedef struct ctdb_mes_enclosure
{ 
    int32   average_height;
    struct
    {
	int32   num_vertices; 
	int32  *vertex_ids;       /* Array of vertex ids that outline the floor
			           * of the enclosure, in clockwise order */
    } footprint;

    int32 num_triangles; 
    int32 first_wall_triangle_id; /* The first in the list of triangles
                	 	   * for the walls of this enclosure 
	 			   */
#ifdef USE_LITTLE_ENDIAN
    uint16 padding;
    uint16 num_apertures; 
#else
    uint16 num_apertures; 
    uint16 padding;
#endif
    uint16 *aperture_ids;         /* Array of the ids of the apertures 
			           * connecting to this enclosure */
} CTDB_MES_ENCLOSURE;

typedef struct ctdb_mes_template
{    
    CTDB_MES_TYPE     type;
    float64           fixed_point_basis;   /* Scaling of local coordinates */
    int32             num_vertices;
    CTDB_MES_VERTEX  *all_vertices;        /* Array of all vertices:
					    * x,y,z, x,y,z, x,y,z, etc. */

    /* We choose to group the list of all triangles so that all 
     * wall triangles are at the end of this array of triangles.
     * Furthermore, the list of triangles will be grouped by enclosure,
     * so that all wall triangles of an enclosure are grouped together.
     * This allows us to specify the list of triangles for an enclosure 
     * merely by specifying the ID of the first triangle in the list.
     * Similarly, the list of aperture triangles is grouped by aperture.
     */
    int32               num_triangles;
    CTDB_MES_TRIANGLE  *all_triangles;     /* Array of all triangles, 
					    * both wall and aperture */
#ifdef USE_LITTLE_ENDIAN
    uint16              padding1;
    uint16              num_enclosures;
#else
    uint16              num_enclosures;
    uint16              padding1;
#endif
    CTDB_MES_ENCLOSURE *all_enclosures;    /* Array of all enclosures */
#ifdef USE_LITTLE_ENDIAN
    uint16              padding2;
    uint16              num_apertures;
#else
    uint16              num_apertures;
    uint16              padding2;
#endif
    CTDB_MES_APERTURE  *all_apertures;     /* Array of all apertures */
    int32               num_boxes;
    CTDB_MES_GRID_BOX  *all_boxes;         /* Array of all gridboxes */
} CTDB_MES_TEMPLATE;

typedef struct ctdb_mes_volume 
{
    float64                   origin[XYZC]; /* Origin of structure in GCS */
 /* Rotation matrix from GCS world space to mes local space*/  
    float64                   tfm_world_to_mes[XYZ][XYZ]; 
    CTDB_MES_LIST_VERTEX_F64  roofline;     /* Array of vertices in roofline */
    CTDB_MES_LIST_VERTEX_F64  floorline;    /* Array of vertices in floorline*/
    CTDB_MES_TEMPLATE        *template;     /* The mes-volume's description */
} CTDB_MES_VOLUME;


/**********************************************************************
 * PAT data structure                                                *
 **********************************************************************/
/* Different ways of looking at the data stored for each attribute */

/* Data structure used to store a column of the PAT table */
typedef struct ctdb_pat_column_hdr
{
  uint32                 facc;             /* FACC code for this column */
  union ctdb_pat_data   *data;             /* Attribute data */
} CTDB_PAT_COLUMN_HDR;	

/* PAT table data structure */
typedef struct ctdb_pat_table
{
    uint32 num_columns;
    CTDB_PAT_COLUMN_HDR *columns;  /* array of column header structures,
				    * one for each column.
				    */
    uint32 num_entries;            /* Number of valid entries in table */
    uint32 alloced_entries;        /* Number of entries each column currently
				    * has alloced. (We will enforce the rule
				    * that all columns have the same # of 
				    * entries alloced at all times.) This will 
				    * support efficient run-time addition
				    * of entries.
				    */
} CTDB_PAT_TABLE;

/**********************************************************************
 * ct_lookup.c                                                     *
 *********************************************************************/

/* ctdb_lookup_feature will return the starting address of the feature list
 * for the input patch number.  If there are no features in the patch, NULL
 * will be returned.
 */
extern CTDB_FEATURE_DATA *                                             
ctdb_lookup_feature (
		     CTDB  *ctdb,
		     int32  patch_number
		     );

/* ctdb_lookup_feature_mask will find the features present mask associated 
 * with the input post position (px,py).  The output from this function
 * will have the bits in the POST_FLAGS position 
 */
extern int32
ctdb_lookup_feature_mask ( 
			  CTDB *ctdb,
			  int32 px,
			  int32 py
			  );

/* ctdb_lookup_norm_elevation finds the elevations of the passed location
 * (normalized to grid post spacing), possibly calling on
 * ctdb_lookup_microterrain_elevation in the process.  External users
 * are encouraged to use ctdb_lookup_elevation, which takes meters.
 * If buildings or canopies is non-zero the elevation returned will be above
 * any of those features at the point, if 0, the elevation will
 * be at ground level. The soil argument, if non-null, will be filled in
 * with the correct soil type.  The reference elevation specifies the
 * minimum acceptable elevation for multi-level microterrain.  Use
 * CTDB_ACCEPT_ML to accept any multi-level polygon; use CTDB_REJECT_ML
 * to reject all multi-level polygons.  There are two versions, one
 * for NW_SE databases, one for SW_NE.
 */
extern float64 
ctdb_lookup_norm_elevation ( 
			    CTDB   *ctdb,
			    float64 norm_x,
			    float64 norm_y,
			    int32   buildings,
			    int32   canopies,
			    int32  *soil,
			    float64 reference_elevation
			    );

/* ctdb_lookup_te_ptr will return the address of the first terrain 
 * element for the input patch number.  If there are no TEs in the patch,
 * NULL will be returned.
 */
extern ADDRESS 
ctdb_lookup_te_ptr (
		    CTDB  *ctdb,
		    int32  patch_number
		    );

/* ctdb_lookup_soilmap will return the starting address of the soilmap
 * for the given ctdb and patch_number.  This map is then to be used 
 * inconjunction with the soilmap stored in the internal representation
 * to determine the final soilmap information.
 */
extern int32 *
ctdb_lookup_soilmap(CTDB *ctdb, int32 patch_number);

/* ctdb_lookup_micro_tide will return the tidal offset for the microterrain 
 * index in patch_number.  This offset should be added to the microterrain
 * z vertices that make up the triangle.
 */
extern float64
ctdb_lookup_micro_tide(CTDB *ctdb,
		       int32 patch_number,
		       int32 micro_water_index);

extern int32
ctdb_find_best_enclosure(
			 CTDB*              ctdb,
			 CTDB_MES_NODE_KEY  depart_mes,
			 float64            depart_point[XYZ],
			 float64            goal_point[XYZ],
			 float64            traversal_height,
			 CTDB_MES_NODE_KEY *arrive_mes,
			 float64            arrive_point[XYZ]
			 );
/* Return the patch number for a given position. 
 * If patch_x and/or patch_y are non-NULL, they will return the patch coords.
 */
extern int32 
ctdb_compute_patch_num(CTDB   *ctdb, 
		       float64 x, 
		       float64 y,
		       int32  *patch_x, 
		       int32  *patch_y);


#endif /*_CT_EXPORT_H_INCLUDED*/



