#ifndef lint
static char rcsid [] = "$RCSfile: ct_feat.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_feat.c                                                         *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: This file contains all elevation lookup processing.       *
 *       Created:  March 1996                                                *
 *       Author:   cbuettner                                                 *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/
#include "libctdb_local.h"

int32
ctdb_invalid_micro(CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj, 
		   CTDB_FEATURE_DATA  *feature, 
		   ADDRESS             data,
		   int32              *next
		   )
{
    *next = NEXT_MICRO(feature->info.vertices);
    return 1;
}


int32
ctdb_invalid_volume(CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj, 
		   CTDB_FEATURE_DATA  *feature, 
		   ADDRESS             data,
		   int32              *next
		   )
{
    *next = NEXT_VOLUME(feature->info.vertices);
    return 1;
}


int32
ctdb_invalid_linear(CTDB               *ctdb, 
		    CTDB_FEAT_OBJ_PTR   obj, 
		    CTDB_FEATURE_DATA  *feature, 
		    ADDRESS             data,
		    int32              *next
		    )
{
    *next = NEXT_LINEAR(feature->info.vertices);
    return 1;
}

int32
ctdb_invalid_canopy(CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj, 
		   CTDB_FEATURE_DATA  *feature, 
		   ADDRESS             data,
		   int32              *next
		   )
{
    *next = NEXT_CANOPY(feature->info.vertices);
    return 1;
}


int32
ctdb_invalid_laid_linear(
			 CTDB               *ctdb,
			 CTDB_FEAT_OBJ_PTR   obj, 
			 CTDB_FEATURE_DATA  *feature, 
			 ADDRESS             data,
			 int32              *next
			 )
{
    *next = NEXT_LAID_LINEAR(feature->info.vertices);
    return 1;
}


int32
ctdb_invalid_aggregate(
		       CTDB               *ctdb,
		       CTDB_FEAT_OBJ_PTR   obj, 
		       CTDB_FEATURE_DATA  *feature, 
		       ADDRESS             data,
		       int32              *next
		       )
{
    *next = NEXT_AGGREGATE(feature->info.vertices);
    return 1;
}


int32 
ctdb_invalid_type(CTDB               *ctdb,
		  CTDB_FEAT_TYPE_PTR  sub,
		  CTDB_FEATURE_DATA  *feature, 
		  ADDRESS            *data)
{
    return 1;
}

void
ctdb_init_micro_subclass(
			 CTDB_FEAT_TYPE *sc, 
			 CTDB_TYPE_FUNC *func
			 )
{
    int32 i; 
    static int32 types[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	MICRO_TYPE_BASE , 
	MICRO_TYPE_ML ,
	MICRO_TYPE_DEFAULT , 
	MICRO_TYPE_WATER
    };
    static CTDB_TERRAIN_SUBCLASS ttid[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	CTDB_FS_SKIN  , 
	CTDB_FS_MULTI ,
	CTDB_FS_SKIN  , 
	CTDB_FS_WATER
    };

    for (i=0;i<CTDB_NUM_MICRO_SUBCLASSES;i++)
    {
	sc[i].tcid     = CTDB_FC_MICRO;
	sc[i].tsid     = ttid[i];
	sc[i].type     = types[i];
	sc[i].incr     = NEXT_MICRO(types[i]);
	sc[i].func     = (CTDB_TYPE_FUNC)func[i];
    }
}

void
ctdb_init_volume_subclass(CTDB_FEAT_TYPE *sc,
                          CTDB_TYPE_FUNC *func)
{
    int32 i;
    static CTDB_TERRAIN_SUBCLASS ttid[CTDB_NUM_VOLUME_SUBCLASSES] = 
    {
	CTDB_FS_BUILDING  , 
	CTDB_FS_ROCK_DROP ,
	CTDB_FS_MES
    };

    for (i=0;i<CTDB_NUM_VOLUME_SUBCLASSES;i++)
    {
	sc[i].tcid     = CTDB_FC_VOLUME;
        sc[i].tsid     = ttid[i];
        sc[i].incr     = -1;      /* Not static */
        sc[i].type     = 0;       /* Not relevant */
        sc[i].func     = (CTDB_TYPE_FUNC)func[i];
    }
}
