#ifndef lint
static char rcsid [] = "$RCSfile: ct_gi.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_gi.c                                                           *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to find the first Ground Intersection with a 3D line *
 *       Created: Fri Nov 15 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdext.h>
#include <libgcs.h>
#include <libvecmat.h>
#include "libctdb_local.h"

#define LARGE_HEIGHT    1000.0

/* Global Local Variables */
static int32 ctdb_hit_dirt;
static int32 ctdb_hit_te;
static int32 ctdb_tested_dirt;
static int32 ctdb_hit_dirt_px, ctdb_hit_dirt_py,ctdb_hit_dirt_dist ;
static int32 ctdb_hit_te_px, ctdb_hit_te_py, ctdb_hit_te_dist;
static int32 ctdb_hit_feat;
static float64 ctdb_feat_t;	/* t along *entire* intervis ray at which we
				 * hit a feature
				 */
static CTDB_COORD3D ctdb_start;
static CTDB_COORD3D ctdb_dirt_limit;
static CTDB_COORD3D ctdb_te_limit;
static int32 ctdb_treeline_patch_x, ctdb_treeline_patch_y;
static CTDB_COORD2D ctdb_treeline_segment_start, ctdb_treeline_segment_end;
static CTDB_BACKGROUND background;
static int32 do_background = 0;
static int32 ctdb_last_cell;	/* Need to keep track of this in case someone
				 * calls ctdb_get_treeline_segment */
static float64 ctdb_best_point[3];
static int32 ctdb_result;
static int32 ctdb_post_flags;

static CTDB_ELEV_QUALIFIER gi_qual;/* qualifier that was passed by the caller.
				    */
static int32 gi_check_dirt;
static int32 gi_no_water;
static int32 gi_first_water; /* this is the first water polygon evaluated*/
static CTDB_FEATURE_OBJ gi_obj[FC_CLASS_MAX];
static CTDB_FEAT_TYPE   gi_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];
static CTDB_FEAT_TYPE   gi_volume_subclass_obj[CTDB_NUM_VOLUME_SUBCLASSES];
static int32 from_above,from_below;
static int32 patch_x, patch_y;
static float64 total_tree_height;

typedef struct ctdb_gi_obj_data
{
    CTDB_PATCH_DATA        patch;
    CTDB_LOS_IN_XYO_AND_DB los;
    int32                  postmask;
} CTDB_GI_OBJ_DATA;


static void
gi_from_above_or_below(float64 zl,
		       float64 zh,
		       float64 z_cross)
{
    /* If the zh=zl=zcross then the point is starting at the water
     * surface or a ray of 0 height has entered the water at the
     * very edge of the first water occurance.  If this is the 
     * case then differ judgement as to where the ray originated.
     * Otherwise if the LOS top is above the water surface then the
     * ray originated from above or if the bottom of the LOS is below 
     * the z_cross then the ray originated from below.
     */
    if ( zh > z_cross )
    {
	from_below       = FALSE;
	from_above       = TRUE;
	gi_first_water   = FALSE;
    }	
    else if ( zl < z_cross )
    {
	from_above       = FALSE;
	from_below       = TRUE;
	gi_first_water   = FALSE;
    }
    else
    {
	from_below = TRUE;
	from_above = TRUE;
    }
}


static int32 
gi_water_compare(CTDB *ctdb, 
		 float64 x, 
		 float64 y, 
		 float64 zl, 
		 float64 zh,
		 float64 z_cross)
{
    float64 precent_reduction,from_top,from_bottom,area;
    float64 ix, iy, dx, dy, dist;

    if ( gi_no_water )
      return VIS_KEEP_GOING;
    /* Compare the z at the intersection with a line connecting
     * two posts to the two z values of the line-of-sight rate at this
     * same intersection, and determine if they go above or below that
     * line.  In order 
     */

    if ( gi_first_water )
    {
	ctdb_tested_dirt = TRUE;
	gi_from_above_or_below(zl,zh,z_cross);
    }
    
    /* Once we hit dirt, keep going until we pass into another patch.
     * Note that we need to compute the patch number for xNW,yNW because
     * that is the only variable guaranteed to move positively along
     * the line of sight.  I also allow one patch of slack to avoid
     * messy problems at patch boundaries.
     */
    if (ctdb_hit_dirt)
    {
	int32 dpx, dpy;
	
	dpx = PATCH_X(x*ctdb->patch_inv_incr) - ctdb_hit_dirt_px;
	dpy = PATCH_Y(y*ctdb->patch_inv_incr) - ctdb_hit_dirt_py;
	
	if ((dpx < -1) || (dpx > 1) || (dpy < -1) || (dpy > 1))
	  return VIS_STOP;
    }
    
    /* If the line-of-sight is running totally below the water 
     * (z_cross >= zh) or above the water (z_cross <= zl), then continue
     * running.
     */
    if (((z_cross >= zh) && from_below) || 
	((z_cross <= zl) && from_above))
      return VIS_KEEP_GOING;
    
    
    /* The ground level at this point is higher (if view from above water)
     * or lower (if viewing from below), than our line at
     * this point.  The exact point of intersection lies somewhere
     * between this point and the eye point, no more than ctdb->incr
     * meters from here.  We will store this ctdb_limiting point, and
     * use a simple (and very fast) sampling to narrow down to the
     * exact location later.
     */
    ix = x * ctdb->incrD;
    iy = y * ctdb->incrD;
    
    /* Find the distance between this point and our starting point */
    dx = ix - ctdb_start.x;
    dy = iy - ctdb_start.y;
    dist = dx * dx + dy * dy;
    
    /* If we have hit dirt before, compare the distance to see if this
     * is closer.
     */
    if (ctdb_hit_dirt  && (dist > ctdb_hit_dirt_dist))
      return VIS_KEEP_GOING;
   
    ctdb_hit_dirt      = CTDB_HIT_WATER;
    ctdb_hit_dirt_px   = PATCH_X(x*ctdb->patch_inv_incr);
    ctdb_hit_dirt_py   = PATCH_Y(y*ctdb->patch_inv_incr);
    ctdb_hit_dirt_dist = dist;
    ctdb_dirt_limit.x  = ix;
    ctdb_dirt_limit.y  = iy;
    ctdb_dirt_limit.z  = zl;
    
    return VIS_KEEP_GOING;
}

/* x,y are in post coordinates in ctdb's native coordinate system */
static int32 
gi_post_compare(
		CTDB *ctdb, 
		float64 x, 
		float64 y, 
		float64 zl, 
		float64 zh, 
		float64 zcross,
		int32   water
		)
{
    float64 ix, iy, dx, dy, dist;
    
    if ( water )
      return gi_water_compare(ctdb,x,y,zl,zh,zcross);
   
    if (!gi_check_dirt)
      return VIS_KEEP_GOING;
    
    /* Figure out where the posts cross the line and check the elevation.
     * Note that since we are concerned with both the posts and the micro-
     * terrain, and since the microterrain will not be processed until
     * we reach the far side of the patch, we cannot simply stop when
     * we find an intersection.  Instead, we need to keep looking until
     * we cross into another patch.
     */
    
    ctdb_tested_dirt = TRUE;
    
    /* Once we hit dirt, keep going until we pass into another patch.
     * Note that we need to compute the patch number for xNW,yNW because
     * that is the only variable guaranteed to move positively along
     * the line of sight.  I also allow one patch of slack to avoid
     * messy problems at patch boundaries.
     */
    if (ctdb_hit_dirt)
    {
	int32 dpx, dpy;
	
	dpx = PATCH_X(x*ctdb->patch_inv_incr) - ctdb_hit_dirt_px;
	dpy = PATCH_Y(y*ctdb->patch_inv_incr) - ctdb_hit_dirt_py;
	
	if ((dpx < -1) || (dpx > 1) || (dpy < -1) || (dpy > 1))
	  return VIS_STOP;
    }
    
    /* If the ground is below the line at this point, keep looking */
    if (zcross < zl)
      return VIS_KEEP_GOING;
    
    /* The ground level at this point is higher than our line at
     * this point.  The exact point of intersection lies somewhere
     * between this point and the eye point, no more than ctdb->incr
     * meters from here.  We will store this ctdb_limiting point, and
     * use a simple (and very fast) sampling to narrow down to the
     * exact location later.
     */
    ix = x * ctdb->incrD;
    iy = y * ctdb->incrD;
    
    /* Find the distance between this point and our starting point */
    dx = ix - ctdb_start.x;
    dy = iy - ctdb_start.y;
    dist = dx * dx + dy * dy;
    
    /* If we have hit dirt before, compare the distance to see if this
     * is closer.
     */
    if (ctdb_hit_dirt  && (dist > ctdb_hit_dirt_dist))
      return VIS_KEEP_GOING;

    ctdb_hit_dirt      = CTDB_HIT_GROUND;
    ctdb_hit_dirt_px   = PATCH_X(x*ctdb->patch_inv_incr);
    ctdb_hit_dirt_py   = PATCH_Y(y*ctdb->patch_inv_incr);
    ctdb_hit_dirt_dist = dist;
    ctdb_dirt_limit.x  = ix;
    ctdb_dirt_limit.y  = iy;
    ctdb_dirt_limit.z  = zl;
    
    return VIS_KEEP_GOING;
}

/* los is the *entire* LOS, not just the portion in this cell */
static int32 
gi_te_edge_compare(
		   CTDB   *ctdb, 
		   CTDB_LOS_IN_XYO_AND_DB *los, 
		   float64 inz0, 
		   float64 inz1, 
		   float64 tedge, 
		   float64 tray,
		   int32   hit
		   )
{
    float64 z, zl;
    float64 ix, iy, dx, dy, dist;
    
    /* this routine would not have been called if we weren't concerned about
       tes */
    
    ctdb_tested_dirt = TRUE;

    z = (inz1 - inz0) * tedge + inz0;
    zl = los->db.start.z + tray * (los->db.end.z - los->db.start.z);
    
    /* If the segment is below the line at this point, keep looking */
    if (z < zl)
      return VIS_KEEP_GOING;
    
    /* The ground level at this point is higher than our line at
     * this point.  The exact point of intersection lies somewhere
     * between this point and the eye point, no more than (ctdb->incr *
     * ctdb->patch_incr) meters from here.  We will store this limiting point, 
     * and use a simple (and very fast) sampling to narrow down to the
     * exact location later.
     */
    ix = los->db.start.x + tray * (los->db.end.x - los->db.start.x);
    iy = los->db.start.y + tray * (los->db.end.y - los->db.start.y);
    
    /* Find the distance between this point and our starting point */
    dx = ix - ctdb_start.x;
    dy = iy - ctdb_start.y;
    dist = dx * dx + dy * dy;

    ctdb_hit_te      = CTDB_HIT_GROUND;
    ctdb_hit_te_px   = PATCH_X(ix*ctdb->meters_to_patch);
    ctdb_hit_te_py   = PATCH_Y(iy*ctdb->meters_to_patch);
    ctdb_hit_te_dist = dist;
    ctdb_te_limit.x  = ix;
    ctdb_te_limit.y  = iy;
    ctdb_te_limit.z  = zl;

    /* no reason to continue - we hit at TE and the TE intervis tranversal 
       only
       checks for ground hits.  All other feature type hits and dirt hits on
       post patches are checked later. */
    
    return VIS_STOP;    
}

/* LOS is entire intervis ray stored in current coord system */
static void 
gi_edge_compare(
		CTDB                   *ctdb, 
		CTDB_FEATURE_DATA      *v0, 
		CTDB_FEATURE_DATA      *v1, 
		CTDB_LOS_IN_XYO_AND_DB *los, 
		int32                   hit_type
		)
{
    float64 tray, tedge;
    float64 z, zl;
    float64 ix, iy, dx, dy, dist;
    
    /* Figure out where the posts cross the line and store that location
     * and elevation.  See comments in gi_post_compare for more...
     */
    
    if (ctdb_segment_intersection_t(&los->xyo, v0->xy, v1->xy, 
				    &tray, &tedge) == DISJOINT)
      return;
    
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    zl = los->db.start.z + tray * (los->db.end.z - los->db.start.z);
    
    /* If the segment is below the line at this point, keep looking */
    if (z < zl)
      return;
    
    /* When we hit a feature, we know exactly where.  That isn't true
     * with ground.
     */
    if (hit_type != CTDB_HIT_GROUND)
    {
	/* Ignore farther-away features */
	if (ctdb_hit_feat && (tray > ctdb_feat_t))
	  return;
	
	ctdb_hit_feat = hit_type;
	ctdb_feat_t = tray;
	return;
    }
    
    /* The ground level at this point is higher than our line at
     * this point.  The exact point of intersection lies somewhere
     * between this point and the eye point, no more than ctdb->incr
     * meters from here.  We will store this limiting point, and
     * use a simple (and very fast) sampling to narrow down to the
     * exact location later.
     */
    ix = los->db.start.x + tray * (los->db.end.x - los->db.start.x);
    iy = los->db.start.y + tray * (los->db.end.y - los->db.start.y);
    
    /* Find the distance between this point and our starting point */
    dx = ix - ctdb_start.x;
    dy = iy - ctdb_start.y;
    dist = dx * dx + dy * dy;
    
    /* If we have hit dirt before, compare the distance to see if this
     * is closer.
     */
    if ((ctdb_hit_dirt) && (dist > ctdb_hit_dirt_dist))
      return;
    
    ctdb_hit_dirt      = CTDB_HIT_GROUND;
    ctdb_hit_dirt_px   = PATCH_X(ix*ctdb->meters_to_patch);
    ctdb_hit_dirt_py   = PATCH_Y(iy*ctdb->meters_to_patch);
    ctdb_hit_dirt_dist = dist;
    ctdb_dirt_limit.x  = ix;
    ctdb_dirt_limit.y  = iy;
    ctdb_dirt_limit.z  = zl;
}

/* LOS is entire intervis ray stored in current coord system */
static void 
gi_tree_compare(CTDB                   *ctdb,
		CTDB_FEATURE_DATA      *pt, 
		float64                 height, 
		float64                 radius, 
		CTDB_LOS_IN_XYO_AND_DB *los,
		int32                   check_trunk,
		int32                   z_provided
		)
{
    static CTDB_ELEV_DATA *elev;
    float64 tray, tedge;
    float64 z, zl, zf;
    /* This works just like gi_treeline_compare, except we need to
     * find the intersection with a different routine, and width is
     * significant.
     */
    
    if (ctdb_point_intersection_t(&(los->xyo.start.x), &(los->xyo.end.x),
				  (float64)pt->xy.x, (float64)pt->xy.y,
				  radius, 0.0, &tray, &tedge) ==
	DISJOINT)
      return;
    
    /* If we've made it here, the line of sight crosses the tree in X&Y.
     * If the tree is part of an aggregate feature, we must do an
     * elevation lookup.  It's better to do the lookup here, after
     * we've found that the LOS does cross the tree in X&Y than to do
     * it in `gi_check_aggregate()' and pass the Z in.  Only calculate
     * the Z (tree top) if we don't already know it.
     */
    if(!z_provided)
    {
	elev =
	  ctdb_lookup_norm_elevation_max(ctdb,
					 (float64)CTDB_XYO_TO_POST(pt[0].xy.x,
								   patch_x,
							    ctdb->patch_incrD),
					 (float64)CTDB_XYO_TO_POST(pt[0].xy.y,
								   patch_y,
							    ctdb->patch_incrD),
					 CTDB_MAX_LAND_AND_WATER);
	pt[1].z = elev->z + total_tree_height;
    }
    
    z  = pt[1].z;
    zl = los->db.start.z + tray * (los->db.end.z - los->db.start.z);
    
    if(check_trunk)
      zf = z - LARGE_HEIGHT;
    else
      zf = z - height;

    /* If the tree top is below the line, or the tree bottom is above the
     * line at this point, keep looking
     */
    if ((z < zl) || (zf > zl))
      return;
    
    /* Ignore farther-away features */
    if (ctdb_hit_feat && (tray > ctdb_feat_t))
      return;
    
    ctdb_hit_feat = CTDB_HIT_TREE;
    ctdb_feat_t = tray;
}

/* LOS is entire intervis ray stored in current coord system */
static void 
gi_treeline_compare(
		    CTDB_FEATURE_DATA *v0, 
		    CTDB_FEATURE_DATA *v1, 
		    float64 height,
		    CTDB_GI_OBJ_DATA *data
		    )
{
    CTDB_LOS_IN_XYO_AND_DB *los=&(data->los);
    float64 tray, tedge;
    float64 z, zl;
    
    /* Figure out where the posts cross the line and store that location
     * and elevation.  See comments in gi_post_compare for more...
     */
    
    if (ctdb_segment_intersection_t(&los->xyo,v0->xy, v1->xy, 
				    &tray, &tedge) == DISJOINT)
      return;
    
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    zl = los->db.start.z + tray * (los->db.end.z - los->db.start.z);
    
    /* If the tree top is below the line, or the tree bottom is above the
     * line at this point, keep looking
     */
    if ((z < zl) || (z-height > zl))
      return;
    
    /* Ignore farther-away features */
    if (ctdb_hit_feat && (tray > ctdb_feat_t))
      return;
    
    ctdb_hit_feat                 = CTDB_HIT_TREELINE;
    ctdb_feat_t                   = tray;
    ctdb_treeline_segment_start.x = v0->xy.x;
    ctdb_treeline_segment_start.y = v0->xy.y;
    ctdb_treeline_segment_end.x   = v1->xy.x;
    ctdb_treeline_segment_end.y   = v1->xy.y;
    
    ctdb_treeline_patch_x         = data->patch.x;
    ctdb_treeline_patch_y         = data->patch.y;
}


static int32
gi_micro_compare(CTDB               *ctdb,
		 CTDB_FEAT_TYPE_PTR  sub,
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_GI_OBJ_DATA   *data
		 )
{
    feature += 2; 

    gi_edge_compare(ctdb,
		    feature,
		    feature+(sub->incr*feature[2].micro.vert0), 
		    &(data->los),
		    (int32)CTDB_HIT_GROUND);
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    gi_edge_compare(ctdb,
		    feature,
		    feature+(sub->incr*feature[2].micro.vert1),
		    &(data->los),
		    (int32)CTDB_HIT_GROUND);
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    gi_edge_compare(ctdb,
		    feature,
		    feature+(sub->incr*feature[2].micro.vert2), 
		    &(data->los),
		    (int32)CTDB_HIT_GROUND);    
    return VIS_KEEP_GOING; 
}
static int32
gi_water_edge_compare(CTDB *ctdb, 
		      CTDB_FEATURE_DATA      *v0, 
		      CTDB_FEATURE_DATA      *v1,
		      CTDB_LOS_IN_XYO_AND_DB *los, 
		      float32                 tide
		      )
{
    float64 tray, tedge;
    float64 z, zl;
    float64 ix, iy, dx, dy, dist;
    
    /* We treat DISJOINT and COLINEAR the same way:
     * DISJOINT means our line-of-sight cannot be blocked by this
     * segment.
     * COLINEAR segments may or may not block our line of sight, but
     * it doesn't matter: if we are dealing with closed polygons (such
     * as microterrain or a house) the tips of the other segments in
     * the polygon will block or not block our visibility as
     * appropriate; if we are dealing with an open list of segments
     * (such as a treeline) a colinear segment would have 0 width, and
     * hence could not block intervisibility of an object with >0 width.
     */
    if (ctdb_segment_intersection_t(&los->xyo,v0->xy, v1->xy,
				    &tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;

    ctdb_tested_dirt = TRUE;

    z = (v1[1].z - v0[1].z) * tedge + v0[1].z + tide;
    zl = los->db.start.z + tray * (los->db.end.z - los->db.start.z);

    if ( gi_first_water )
    {
	ctdb_tested_dirt = TRUE;
	/* zl and zh are the same in ground intersection */
	gi_from_above_or_below(zl,zl,z);
    }   
    
    /* If the line-of-sight is running totally below the water 
     * (z_cross >= zl) or above the water (z_cross <= zl), then continue
     * running.
     */
    if (((z >= zl) && from_below) || 
	((z <= zl) && from_above))
      return VIS_KEEP_GOING;

    /* The ground level at this point is higher than our line at
     * this point.  The exact point of intersection lies somewhere
     * between this point and the eye point, no more than ctdb->incr
     * meters from here.  We will store this limiting point, and
     * use a simple (and very fast) sampling to narrow down to the
     * exact location later.
     */
    ix = los->db.start.x + tray * (los->db.end.x - los->db.start.x);
    iy = los->db.start.y + tray * (los->db.end.y - los->db.start.y);
    
    /* Find the distance between this point and our starting point */
    dx = ix - ctdb_start.x;
    dy = iy - ctdb_start.y;
    dist = dx * dx + dy * dy;
    
    /* If we have hit dirt before, compare the distance to see if this
     * is closer.
     */
    if ((ctdb_hit_dirt) && (dist > ctdb_hit_dirt_dist))
      return VIS_KEEP_GOING;

    ctdb_hit_dirt      = CTDB_HIT_WATER;
    ctdb_hit_dirt_px   = PATCH_X(ix*ctdb->meters_to_patch);
    ctdb_hit_dirt_py   = PATCH_Y(iy*ctdb->meters_to_patch);
    ctdb_hit_dirt_dist = dist;
    ctdb_dirt_limit.x  = ix;
    ctdb_dirt_limit.y  = iy;
    ctdb_dirt_limit.z  = zl;
    return VIS_KEEP_GOING;    
}

static int32
gi_micro_water_compare(CTDB               *ctdb,
		       CTDB_FEAT_TYPE_PTR  sub, 
		       CTDB_FEATURE_DATA  *feature, 
		       CTDB_GI_OBJ_DATA   *data
		       )
{
    float32 tide1,tide2;
    feature += 2;
    
    tide1 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char0);
    
    if (!gi_water_edge_compare(ctdb,
			       feature,
			       feature+(sub->incr*feature[2].micro.vert0), 
			       &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    if (!gi_water_edge_compare(ctdb,
			       feature,
			       feature+(sub->incr*feature[2].micro.vert1),
			       &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    tide2 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[4].mwater.char1);

    /* Since the tidal level may shift at the center edge of a polygon
     * check it again with the new tidal level if the tidal level is 
     * different from the first.
     */
    if ((tide2 != tide1) &&
	(!gi_water_edge_compare(ctdb,
				feature,
				feature+(sub->incr*feature[2].micro.vert0),
				&(data->los),tide2)))
	  return VIS_STOP;
    
    if (!gi_water_edge_compare(ctdb,
			       feature,
			       feature+(sub->incr*feature[2].micro.vert2), 
			       &(data->los),tide2))
      return VIS_STOP; 

    return VIS_KEEP_GOING;
}

static int32
gi_check_micro(
	       CTDB               *ctdb,
	       CTDB_FEAT_OBJ_PTR   obj, 
	       CTDB_FEATURE_DATA  *feature, 
	       CTDB_GI_OBJ_DATA   *data,
	       int32              *next
	       )
{
    static CTDB_FEAT_TYPE *sub=gi_micro_subclass_obj;
    int32 verts = feature->info.vertices;

    /* check if the micro type has changed */
    if ((verts & MICRO_TYPE) != sub->type)
      sub = &gi_micro_subclass_obj[MICRO_TYPE_INDEX(verts)];

    *next = sub->incr;
    
    if (!obj->valid || 
	!(data->postmask & feature->integer) ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return VIS_KEEP_GOING;

    /* call the micro type processing function */
    return sub->func(ctdb,sub,feature,data);
}

static int32 
gi_volume_compare(
		  CTDB               *ctdb,
		  CTDB_FEAT_TYPE     *sub,
		  CTDB_FEATURE_DATA  *feature, 
		  CTDB_GI_OBJ_DATA   *data
		  )
{
    int32 j;
    int32 verts = feature->info.vertices;
    
    /* move past INFO, VOLUME_INFO, ID */ 
    feature += 3;

    gi_edge_compare(ctdb, 
		    feature, 
		    feature+((verts-1)<<1), 
		    &(data->los),
		    (int32)CTDB_HIT_BUILDING);

    for(j=0;j<verts-1;j++,feature+=2)
    {
	gi_edge_compare(ctdb, 
			feature, 
			feature+2, 
			&(data->los),
			(int32)CTDB_HIT_BUILDING);
    }
    
    return VIS_KEEP_GOING;
}
static
void cell_to_mes ( CTDB             *ctdb,
                   CTDB_MES_VOLUME  *mes,
                   int32             patch_x,
                   int32             patch_y,
                   float64           cell_x,
                   float64           cell_y,
	           float64           cell_z,
                   float64          *mes_x,
                   float64          *mes_y,
                   float64          *mes_z )                

{
   float64 mes_position[XYZ];
   int32 mes_cell = mes->origin[CELL3D];

   float64 offset[XYZ];

   if ( ctdb_get_cell_id(ctdb) != mes_cell )
   {
       gcs_gcs3e_to_gcscs ( ctdb->cell_id, cell_x, cell_y, cell_z, mes_cell,
                            &offset[X], &offset[Y], &offset[Z] );
       offset[X] -= mes->origin[X];
       offset[Y] -= mes->origin[Y];
       offset[Z] -= mes->origin[Z];
   }
   else
   {
       offset[X] = cell_x - mes->origin[X];
       offset[Y] = cell_y - mes->origin[Y];
       offset[Z] = cell_z - mes->origin[Z];
   }

   ctdb_gcsvec_to_mesvec ( mes, mes_cell, offset, mes_position );
   *mes_x = mes_position[X];
   *mes_y = mes_position[Y];
   *mes_z = mes_position[Z];
}

static int32
gi_mes_volume_compare(
		      CTDB               *ctdb,
		      CTDB_FEAT_TYPE     *sub,
		      CTDB_FEATURE_DATA  *feature, 
		      CTDB_GI_OBJ_DATA   *data
		      )
{
    uint32              which_volume_model; 
    uint32              which_mes_volume;
    CTDB_MES_VOLUME    *mes;
    CTDB_VOLUME_MODEL  *volume_model;

    float64 ray_origin[XYZ];
    float64 ray_destination[XYZ];

    int32   number;
    int32   intercepted_triangle;
    float64 intercept_distance;
    CTDB_MES_INTERSECTION *mes_inter;
    int32 n_mes_inter=0;

    which_volume_model = feature[1].volume.reference;
    if ( which_volume_model >= ctdb->num_mes_volumes ) /* ERROR */
      return VIS_KEEP_GOING;

    volume_model = ctdb->mes_volume_models[which_volume_model];
    which_mes_volume = volume_model->mes_volume_index;

    if ( which_mes_volume >= ctdb->num_mes_volumes ) /* ERROR*/
      return VIS_KEEP_GOING;

    mes = &(ctdb->mes_volumes[which_mes_volume]);

    cell_to_mes ( ctdb, mes,
                  data->patch.x, data->patch.y,
                  data->los.db.start.x,
                  data->los.db.start.y,
                  data->los.db.start.z,
                  &ray_origin[X],
                  &ray_origin[Y],
                  &ray_origin[Z] );

    cell_to_mes ( ctdb, mes,
                  data->patch.x, data->patch.y,
                  data->los.db.end.x,
                  data->los.db.end.y,
                  data->los.db.end.z,
                  &ray_destination[X],
                  &ray_destination[Y],
                  &ray_destination[Z] );

   ctdb_mes_line_intersection(mes,
			      ray_origin,
			      ray_destination,
			      FALSE, CTDB_MES_ALL_ENCLOSURES, 
			      TRUE, FALSE,
			      FALSE, &n_mes_inter,
			      &mes_inter);
    
    if ((n_mes_inter) && (mes_inter->distance <= ctdb_feat_t))
    {
	ctdb_feat_t   = mes_inter->distance;
	ctdb_hit_feat = (int32)CTDB_HIT_BUILDING;
    }

    return VIS_KEEP_GOING;    
}


static int32
gi_check_volume(
		CTDB               *ctdb,
		CTDB_FEAT_OBJ_PTR   obj, 
		CTDB_FEATURE_DATA  *feature, 
		CTDB_GI_OBJ_DATA   *data,
		int32              *next
		)
{
    CTDB_FEAT_TYPE *sub;
    int32 verts = feature->info.vertices;

    *next = NEXT_VOLUME(verts);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

    sub = &gi_volume_subclass_obj[ feature[1].volume.type];
    return sub->func(ctdb, sub, feature, data);
}


static int32
gi_check_canopy(CTDB               *ctdb,
		CTDB_FEAT_OBJ_PTR   obj, 
		CTDB_FEATURE_DATA  *feature, 
		CTDB_GI_OBJ_DATA   *data,
		int32              *next
		)
{
    static int32 ray_info;
    static int32 verts, incr;

    verts = feature->info.vertices;
    incr  = *next = NEXT_CANOPY(verts);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

    feature++;
    
    if (feature[2].canopy.vert0_outer)
    {
	gi_edge_compare(ctdb,
			feature, 
			feature+(incr*feature[2].canopy.vert0), 
			&(data->los),
			(int32)CTDB_HIT_TREELINE);
    }
    
    if (verts == 1)
      return VIS_KEEP_GOING;
    
    if (feature[2].canopy.vert1_outer)
    {
	gi_edge_compare(ctdb,
			feature, 
			feature+(incr*feature[2].canopy.vert1), 
			&(data->los),
			(int32)CTDB_HIT_TREELINE);
    }
    
    if (verts == 2)
      return VIS_KEEP_GOING;
    
    if (feature[2].canopy.vert2_outer)
    {
	gi_edge_compare(ctdb, 
			feature, 
			feature+(incr*feature[2].canopy.vert2), 
			&(data->los),
			(int32)CTDB_HIT_TREELINE);
    }
    
    return VIS_KEEP_GOING;
}

static int32
gi_check_linear(CTDB               *ctdb, 
		CTDB_FEAT_OBJ_PTR   obj, 
		CTDB_FEATURE_DATA  *feature, 
		CTDB_GI_OBJ_DATA   *data,
		int32              *next
		)
{
    static CTDB_LINEAR_MODEL *model;
    float64                   height, radius;
    int32                     j;

    int32 verts = feature->info.vertices;
    *next = NEXT_LINEAR(verts);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

    model = &ctdb->linear_models[feature[1].linear.reference];

    /* if the model type is a wire or dragons tooth obstacle, then
     * adjust the height.  otherwise use the standard tree line
     * computation below..
     *
     * the following chunk of code should _not_ be removed!!!
     */
    switch(model->type)
    {
      case CTDB_FS_TREELINE:
          height = model->linear_data.treeline_info.foliage_height*
            CTDB_TREE_HEIGHT_BASIS;
          break;

        case CTDB_FS_CONCERTINA_WIRE:
        case CTDB_FS_WIRE_ROAD_BLOCK:
        case CTDB_FS_FENCE:
        case CTDB_FS_MINEFIELD_FENCE:
          height = LARGE_HEIGHT;
          break;
          
        case CTDB_FS_DRAGONS_TEETH:
          height = LARGE_HEIGHT;
          break;

    }
    
    radius = model->width * 0.5;

    /* Move past CTDB_FEATURE_INFO and CTDB_LINEAR_INFO */
    feature+=2;
    
    if (verts==1)
    {
	/* Check the foliage. */
	gi_tree_compare(ctdb, feature, height, radius, &(data->los), FALSE,
			TRUE);

	/* Check the trunk. */
	gi_tree_compare(ctdb, feature, height, radius, &(data->los), TRUE,
			TRUE);
    }
    else
      for(j=0;j<verts-1;j++,feature+=2)
	gi_treeline_compare(feature, feature+2, height, data);
    
    return VIS_KEEP_GOING;
}

static int32
gi_check_aggregate(CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj, 
		   CTDB_FEATURE_DATA  *feature, 
		   CTDB_GI_OBJ_DATA   *data,
		   int32              *next
		   )
{
    int32                        i;
    static CTDB_AGGREGATE_MODEL *model;
    static CTDB_LINEAR_MODEL    *linear_model;
    static CTDB_FEATURE_DATA     pt[2];
    float64                      height, foliage_radius, trunk_radius;

    *next = NEXT_AGGREGATE(0);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

    model = ctdb->aggregate_models[feature[1].aggregate.reference];

    for(i = 0; i < model->n_features; i++)
    {
	/* Grab the linear model associated with this feature.
	 */
	linear_model = &ctdb->linear_models[model->features[i].model.ref];

	/* Assign pt[0], which is a `CTDB_XYO_COORD2D' structure.
	 */
	pt[0].xy.x = feature[2].xy.x + model->features[i].offset_x;
	pt[0].xy.y = feature[2].xy.y + model->features[i].offset_y;

	/* Extract needed values from the model.
	 */
	total_tree_height = linear_model->linear_data.treeline_info.
	                    total_height * CTDB_TREE_HEIGHT_BASIS;
	height         = linear_model->linear_data.treeline_info.
	                 foliage_height * CTDB_TREE_HEIGHT_BASIS;
	foliage_radius = linear_model->width * 0.5;
	trunk_radius   = linear_model->linear_data.treeline_info.trunk_radius;

	/* We assign these to global variables to avoid function argument
	 * passing overhead.
	 */
	patch_x = data->patch.x;
	patch_y = data->patch.y;

	/* Check for intersection with the foliage part of the tree.
	 */
	gi_tree_compare(ctdb, pt, height, foliage_radius, &(data->los),
			FALSE, FALSE);

	/* Check for intersection with the trunk.
	 *
	 * We pass TRUE for `z_provided' here because the z was calculated
	 * and filled in during the last call to `gi_tree_compare()',
	 * and there's no need to do the elevation lookup twice.  If
	 * it didn't get to the elevation lookup during the foliage check,
	 * then it certainly won't during the trunk check.  Have you
	 * ever seen a tree whose foliage width is less than its trunk
	 * width?  I didn't think so.
	 */
	gi_tree_compare(ctdb, pt, height, trunk_radius, &(data->los),
			TRUE, TRUE);
    }

    return VIS_KEEP_GOING;
}

int32
ctdb_gi_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static CTDB_TYPE_FUNC gi_micro_func[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)gi_micro_compare,      /* MICRO_TYPE_BASE */ 
	(CTDB_TYPE_FUNC)gi_micro_compare,      /* MICRO_TYPE_ML */
	(CTDB_TYPE_FUNC)gi_micro_compare,      /* MICRO_TYPE_DEFAULT */
	(CTDB_TYPE_FUNC)gi_micro_water_compare /* MICRO_TYPE_WATER */
    };

    static CTDB_TYPE_FUNC gi_volume_func[CTDB_NUM_VOLUME_SUBCLASSES]=
    {
	(CTDB_TYPE_FUNC)gi_volume_compare,      /* CTDB_FT_BUILDING */
	(CTDB_TYPE_FUNC)gi_volume_compare,      /* CTDB_FT_ROCK_DROP */
	(CTDB_TYPE_FUNC)gi_mes_volume_compare,  /* CTDB_FT_MES */
	(CTDB_TYPE_FUNC)gi_volume_compare       /* CTDB_FT_PLUG */
    };

    /* CLASS FC_MICRO */
    obj[FC_MICRO].func         = (CTDB_OBJ_FUNC)gi_check_micro;
    ctdb_init_micro_subclass(gi_micro_subclass_obj,gi_micro_func);

    /* CLASS FC_VOLUME */
    obj[FC_VOLUME].func        = (CTDB_OBJ_FUNC)gi_check_volume;
    ctdb_init_volume_subclass(gi_volume_subclass_obj,gi_volume_func);

    obj[FC_LINEAR].func        = (CTDB_OBJ_FUNC)gi_check_linear;
    obj[FC_CANOPY].func        = (CTDB_OBJ_FUNC)gi_check_canopy;
    obj[FC_LAID_LINEAR].func   = (CTDB_OBJ_FUNC)ctdb_invalid_laid_linear;
    obj[FC_AGGREGATE].func     = (CTDB_OBJ_FUNC)gi_check_aggregate;

    /* Store Initialized local feature objects locally */
    bcopy(obj,gi_obj,sizeof(gi_obj));    
    return 1;
}


/* start, end are the endpoints of the *entire* intervis ray, represented in
 * the frame of reference specified by ctdb.
 */
static int32 
gi_patch_compare(
		 CTDB *ctdb, 
		 int32 patch_number, 
		 int32 feature_mask,
		 int32 post_mask, 
		 float64 *start, 
		 float64 *end
		 )
{
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_GI_OBJ_DATA data;
    CTDB_FEATURE_DATA *feature, *last;
    int32 fclass, next;
    
    /* Note that since we selected the things we want to test for
     * in the feature_mask we passed to ctdb_intervis, we don't
     * need to explicitly check those flags here.
     */
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
      return VIS_KEEP_GOING;
    
    /* It's faster to compute these than pass them in.
     */
    data.postmask     = post_mask;
    data.patch.number = patch_number;
    data.patch.x = data.patch.number % ctdb->patches_wide;
    data.patch.y = data.patch.number / ctdb->patches_wide;
    
    data.los.xyo.start.x = POST_TO_XYO(start[X], data.patch.x,
				  ctdb->patch_inv_incr);
    data.los.xyo.start.y = POST_TO_XYO(start[Y], data.patch.y,
				  ctdb->patch_inv_incr);
    
    data.los.xyo.end.x = POST_TO_XYO(end[X], data.patch.x, 
				     ctdb->patch_inv_incr);
    data.los.xyo.end.y = POST_TO_XYO(end[Y], data.patch.y, 
				     ctdb->patch_inv_incr);
    data.los.db.start.x = start[X] * ctdb->incrD;
    data.los.db.start.y = start[Y] * ctdb->incrD;
    data.los.db.start.z = start[Z];
    data.los.db.end.x = end[X] * ctdb->incrD;
    data.los.db.end.y = end[Y] * ctdb->incrD;
    data.los.db.end.z = end[ZL];
    
    /* Run through all the microterrain on this patch.
     */
    
    /* Note that bit masks are used in some places rather than
     * bitfield accessors to avoid unnecessary shifting (a big
     * performance win, it turns out).
     */
    obj        = (CTDB_FEAT_OBJ_PTR)gi_obj;
    obj->valid = feature_mask & obj->featmask;

    last = feature + feature->hdr.size;
    for(feature++;feature<last;feature+=next)
    {
	/* Only interested in microterrain and buildings */
	fclass = feature->info.feature_class;

	/* if the current class doesn't match the previos class it is time 
	 * to change the class object pointer.
	 */
	if ( fclass != obj->sensClass )
	{
	    /* If we are done with the feature class then remove from the 
	     * list and check if we are done.  If the class is LINEAR then
	     * don't remove the feature flag POST_TREES since FC_CANOPY may
	     * also exist in the list of features to evaluate.
	     */
	    if ( obj->sensClass != FC_LINEAR )
	      feature_mask &= ~obj->featmask;

	    if ( !feature_mask )
	      break;

	    obj        = (CTDB_FEAT_OBJ_PTR)&gi_obj[fclass];
	    obj->valid = feature_mask & obj->featmask;
	}

	if ( obj->func(ctdb,obj,feature,(ADDRESS)&data,&next) == VIS_STOP)
	  return VIS_STOP;
    }	
    return VIS_KEEP_GOING;
}

static int32 
gi_sample(CTDB         *ctdb, 
	  float64      *gi_pt, 
	  int32         px, 
	  int32         py, 
	  CTDB_COORD3D *limit,
	  int32        *hit)
{
    int32 i, dist;
    float64 x, y, z, dx, dy, dz, hx, hy;
    float64 dirt;
    float64 max_dist;
    int32 patch_num;
    int32 get_max;
    CTDB_ELEV_DATA elev;
    CTDB_ELEV_QUALIFIER qual;
    static CTDB_ELEV_QUALIFIER valid_dirt=(CTDB_MAX_LAND_NO_WATER |
					   CTDB_INCL_LAID_LINEAR);
    static CTDB_ELEV_QUALIFIER valid_h2o=(CTDB_MAX_LAND_AND_WATER);
    
    /* Don't do anything if it is farther than this. */
    dx = gi_pt[0] - ctdb_start.x;
    dy = gi_pt[1] - ctdb_start.y;
    max_dist = dx * dx + dy * dy;

    /* Choose a sampling rate */    
    dx = limit->x - ctdb_start.x;
    if (dx < 0) dx = -dx;
    dy = limit->y - ctdb_start.y;
    if (dy < 0) dy = -dy;
    
    /* If we hit the ground then we will only want to look at terrain 
     * skin elevation in the lookup.  If the intersection was due to water
     * then we can be either above or below the water surface.  The item 
     * from_above will be TRUE if the ray originated from above the 
     * water surface (this means the same intersection criteria as terrain
     * intersection but we should include water in the elevation lookup.
     * If the ray originated from below then we need to invert the intersection
     * test since we want to stop when the z of the line is above the water 
     * surface.
     */
    if ( *hit == CTDB_HIT_GROUND )
    {
	get_max = TRUE;
	qual = gi_qual & valid_dirt;
    }
    else
    {
	get_max = from_above;
	qual = gi_qual & valid_h2o;
    }
     
    if ((dx < 1.0) && (dy < 1.0))
    {
	/* If the ground intersection is good enough and the intersection
	 * is closer then the best feature intersection, then return the 
	 * best terrain intersection, otherwise leave the best interesection
	 * set to the feature intersection.
	 */
	if ((dx * dx) + (dy * dy) <= max_dist)
	{
	    gi_pt[0] = limit->x;
	    gi_pt[1] = limit->y;
	    gi_pt[2] = ctdb_lookup_qual_elevation(ctdb, limit->x, limit->y,
						  qual, CTDB_NO_REF_ELEV, 
						  &elev);
	    if ( elev.subclass == CTDB_FS_WATER )
	      *hit = CTDB_HIT_WATER;
	    else
	      *hit = CTDB_HIT_GROUND;
	    
	    return 1;
	}
	else
	  return 0;
    }
    
    /* For a gridded patch, the last point we are guaranteed not to have
     * intersected the ground at was the last grid edge we crossed.  For
     * a TINNed patch, it was the last *patch* edge we crossed...
     */
    patch_num = (py * ctdb->patches_wide) + px;
    if ( ctdb_lookup_te_ptr(ctdb,patch_num) )
      dist = ctdb->patch_incr * ctdb->incr;
    else
      dist = ctdb->incr;
    
    if (dx > dy)
    {
	if (dx < dist)
	  dist = dx;
	
	dy = (limit->y - ctdb_start.y) / dx;
	dz = (limit->z - ctdb_start.z) / dx;
	
	if (limit->x > ctdb_start.x)
	  dx = 1.0;
	else
	  dx = -1.0;
    }
    else
    {
	if (dy < dist)
	  dist = dy;
	
	dx = (limit->x - ctdb_start.x) / dy;
	dz = (limit->z - ctdb_start.z) / dy;
	
	if (limit->y > ctdb_start.y)
	  dy = 1.0;
	else
	  dy = -1.0;
    }
    
    /* The intersection must be within incr meters of the limiting edge.
     * Hence, we sample the elevation of our line of sight starting
     * that far back.
     */
    x = limit->x - dist * dx;
    y = limit->y - dist * dy;
    z = limit->z - dist * dz;

    for(i=0;i<=dist;i++)
    {
	dirt = ctdb_lookup_qual_elevation(ctdb, x, y, qual, CTDB_NO_REF_ELEV, 
					  (CTDB_ELEV_DATA *)NULL);
	
	if ((get_max && (dirt > z)) ||
	    (!get_max && (z > dirt)))
	  break;
	
	x += dx;
	y += dy;
	z += dz;
    }
    
    hx = x-dx-ctdb_start.x;
    hy = y-dy-ctdb_start.y;
    
    if (((hx*hx) + (hy*hy)) < max_dist)
    {
	gi_pt[0] = x-dx;
	gi_pt[1] = y-dy;
	gi_pt[2] = ctdb_lookup_qual_elevation(ctdb, gi_pt[0] , gi_pt[1],
					      qual, CTDB_NO_REF_ELEV, &elev);
	
	if ( elev.subclass == CTDB_FS_WATER )
	  *hit = CTDB_HIT_WATER;
	else
	  *hit = CTDB_HIT_GROUND;

	return 1;
    }
    return 0;
}

static int32 
vehicle_intersect(float64 x0, 
		  float64 y0, 
		  float64 z0, 
		  float64 x1, 
		  float64 y1, 
		  float64 z1, 
		  int32 n_veh, 
		  CTDB_VEHICLE_LOCATION *veh, 
		  int32 ignore0, 
		  int32 ignore1, 
		  int32 *hit_veh, 
		  float64 *min_t)
{
    CTDB_VEHICLE_LOCATION *v;
    float64 eye[2], targ[2];
    float64 tray, tveh, tray2;
    float64 zray, ztop, zbot, zray2;
    float64 radius;
    int32 missed_all;
    float64 closest;
    int32 i, know_unit=0;
    float64 ux, uy, mag;
    
    eye[0] = x0;
    eye[1] = y0;
    targ[0] = x1;
    targ[1] = y1;
    missed_all = 1;
    closest = 2.0;		/* Bigger than 1.0 */
    
    for (i=0;i<n_veh;i++)
      if ((i != ignore0) && (i != ignore1) && veh[i].used)
      {
	  v = veh+i;
	  
	  radius = v->width * 0.5;
	  
	  /* Does this vehicle cross our line of sight in 2D? */
	  if (ctdb_point_intersection_t(eye, targ,
					v->location[X] , v->location[Y],
					radius, 0.0,
					&tray, &tveh) ==
	      DISJOINT)
	    continue;
	  
	  /* If this vehicle is farther than the closest one, skip it. */
	  if (tray > closest)
	    continue;
	  
	  /* Find elevation of line-of-sight ray at this point */
	  zray = (z1 - z0) * tray + z0;
	  
	  /* Find top and bottom of obstruction */
	  zbot = v->location[Z];
	  ztop = zbot + v->height;
	  
	  /* When the ray is very steep (such as an airplane looking
	   * down on a tank), the approximation that height (which,
	   * if ctdb_apparent_size is used, is actually the dimension
	   * perpendicular to both width & line of sight) is the dimension
	   * perpendicular to the XY plane isn't very good.
	   */
	  if ((z1 - z0)*(z1 - z0) > (x1 - x0)*(x1 - x0) + (y1 - y0)*(y1 - y0))
	  {
	      /* The method we use for steep lines works as follows:
	       * zray is the altitude at which our line of sight first
	       * crossed the target; compute the altitude where cross
	       * the target again, and see if the two altitudes are
	       * on opposite sides of the target.
	       */
	      
	      if (!know_unit)
	      {
		  know_unit = 1;
		  ux = x1 - x0;
		  uy = y1 - y0;
		  mag = fsqrt(ux*ux + uy*uy);
		  if (mag != 0.0)
		  {
		      ux /= mag;
		      uy /= mag;
		  }
		  else
		  {
		      ux = 0;
		      uy = 0;
		  }
	      }
	      
	      if (ctdb_point_intersection_t(eye, targ,
					    v->location[X] +v->height*ux*0.5,
					    v->location[Y]+v->height*uy*0.5,
					    radius, 0.0,
					    &tray2, &tveh) ==
		  DISJOINT)
	      {
		  /* The ray must have stopped halfway across the
		   * target.
		   */
		  tray2 = 1.0;
	      }
	      zray2 = (z1 - z0) * tray2 + z0;
	      
	      if (!((zray >= zbot) && (zray2 <= ztop) ||
		    (zray2 >= zbot) && (zray <= ztop)))
		continue;
	      
	  }
	  else
	  {
	      /* Traditional case.  Can we see completely above or
	       * below it?
	       */
	      if ((ztop <= zray) || (zbot >= zray))
		continue;
	  }
	  
	  missed_all = 0;
	  
	  /* I know this vehicle is closest because of the > closest check
	   * up above.
	   */
	  *hit_veh = i;
	  *min_t = tray;
      }
    
    return !missed_all;
}

static int32 
gi_cell_fcn(CTDB *ctdb, float64 *ray_start, float64 *ray_end, 
	    float64 *cur_start, float64 *cur_end, float64 width)
{
    float64 ray_start_post[3];
    float64 ray_end_post[4];
    float64 cur_start_post[4];
    float64 cur_end_post[4];
    CTDB_COORD3D start, end, cell_start, cell_end;
    
    /* Store this off so the caller can find out where we ended. */
    ctdb_last_cell = ctdb_get_cell_id(ctdb);
    
    /* Scale from meters to grid spacing */
    ray_start_post[X] = ray_start[X] * ctdb->inv_incr;
    ray_start_post[Y] = ray_start[Y] * ctdb->inv_incr;
    ray_start_post[ZL] = ray_start[ZL];
    
    ray_end_post[X] = ray_end[X] * ctdb->inv_incr;
    ray_end_post[Y] = ray_end[Y] * ctdb->inv_incr;
    ray_end_post[ZL] = ray_end[ZL];
    ray_end_post[ZH] = ray_end[ZH];
    
    cur_start_post[X] = cur_start[X] * ctdb->inv_incr;
    cur_start_post[Y] = cur_start[Y] * ctdb->inv_incr;
    cur_start_post[ZL] = cur_start[ZL];
    cur_start_post[ZH] = cur_start[ZH];
    
    cur_end_post[X] = cur_end[X] * ctdb->inv_incr;
    cur_end_post[Y] = cur_end[Y] * ctdb->inv_incr;
    cur_end_post[ZL] = cur_end[ZL];
    cur_end_post[ZH] = cur_end[ZH];
    
    /* Copy endpoints into structure used by terrain_element_intervis */
    start.x = ray_start[X];
    start.y = ray_start[Y];
    start.z = ray_start[Z];
    end.x   = ray_end[X];
    end.y   = ray_end[Y];
    end.z   = ray_end[ZL];
    cell_start.x = cur_start[X];
    cell_start.y = cur_start[Y];
    cell_start.z = cur_start[ZL];
    cell_end.x = cur_end[X];
    cell_end.y = cur_end[Y];
    cell_end.z = cur_end[ZL];
    
    /* Initialize 'global'ish variables */
    ctdb_tested_dirt = FALSE;
    ctdb_start.x = cur_start[X];
    ctdb_start.y = cur_start[Y];
    ctdb_start.z = cur_start[ZL];
    ctdb_dirt_limit.x = cur_end[X];
    ctdb_dirt_limit.y = cur_end[Y];
    ctdb_dirt_limit.z = cur_end[ZL];
    ctdb_te_limit.x = cur_end[X];
    ctdb_te_limit.y = cur_end[Y];
    ctdb_te_limit.z = cur_end[ZL];
    ctdb_feat_t = 2.0;
    
    /* This intervis look only checks for dirt hits of TINNED patches */
    if ( gi_check_dirt && ctdb->is_tinned ) 
      (void) ctdb_terrain_element_intervis(ctdb,&start,&end,ray_end[ZH],
					   &cell_start, &cell_end,
					   gi_te_edge_compare,
					   FALSE);
    
    /* This intervis looks for gridded patches and features if desired */
    /* Invoke the traversal engine to find limiting point */
    if ( ctdb->is_gridded || ctdb_post_flags || !gi_no_water )
      (void) ctdb_intervis(ctdb, ray_start_post, ray_end_post,
			   cur_start_post, cur_end_post, TRUE,
			   gi_post_compare, gi_patch_compare, 
			   ctdb_post_flags);
    
    /* If the line is going severely in the down direction, or it is
     * really short, the intervis ray may not cross *any* edges.  In
     * this case, the ctdb_tested_dirt flag will still be zero.
     */
    if (gi_check_dirt && !ctdb_tested_dirt)
#if 0
    {
	CTDB_ELEV_DATA elev;
	static CTDB_ELEV_QUALIFIER valid=(CTDB_MAX_LAND_AND_WATER|
					  CTDB_INCL_LAID_LINEAR);
	
	if ((ctdb_lookup_qual_elevation(ctdb, cur_start[X], cur_start[Y],
					gi_qual&valid, CTDB_NO_REF_ELEV, 
					&elev) 
	     > cur_start[ZL]) ||
	    (ctdb_lookup_qual_elevation(ctdb, cur_end[X], cur_end[Y],
					gi_qual&valid, CTDB_NO_REF_ELEV, 
					&elev) > 
	     cur_end[ZL]))

	  ctdb_hit_dirt = ((elev.subclass ==CTDB_FS_WATER)?CTDB_HIT_WATER:
			   CTDB_HIT_GROUND);
    }
#else
   {
       CTDB_ELEV_DATA elev;
       CTDB_ELEV_QUALIFIER qual;
       float64 t_ray=-1.0;
	
       /* look for anything that crosses ray below the highest point on ray*/
       qual = gi_qual & ~CTDB_RETURN_MAX & ~CTDB_RETURN_MIN;

       if (!gi_no_water)
	 qual |= CTDB_INCL_LAID_LINEAR;
       
       if ( cur_start[Z] < cur_end[Z] )
       {
	   /* straight up */
	   ctdb_lookup_qual_elevation(ctdb,cur_start[X],cur_start[Y],
				      qual,cur_end[Z],&elev);
	   t_ray = (elev.z-cur_end[Z])/(cur_start[Z]-cur_end[Z]);
       }
       else if ( cur_end[Z] < cur_start[Z] )
       {
	   /* straight down */
	   ctdb_lookup_qual_elevation(ctdb,cur_start[X],cur_start[Y],
				      qual,cur_start[Z],&elev);
	   t_ray = (elev.z-cur_start[Z])/(cur_end[Z]-cur_start[Z]);
       }
       
       if ((elev.sensClass != CTDB_FC_ILLEGAL) && (t_ray > 0.0) && (t_ray < 1.0))
       {
	 if (elev.sensClass == CTDB_FC_MICRO)
	       ctdb_hit_dirt = ((elev.subclass ==CTDB_FS_WATER)?CTDB_HIT_WATER:
				CTDB_HIT_GROUND);
	   else
	   {
	       ctdb_feat_t   = t_ray;
	       ctdb_hit_feat = TRUE;
	   }
       }
   }
#endif
    
    ctdb_best_point[X] = cur_end[X];
    ctdb_best_point[Y] = cur_end[Y];
    ctdb_best_point[Z] = cur_end[ZL];
    
    if (!ctdb_hit_dirt && !ctdb_hit_feat && !ctdb_hit_te)
      return VIS_KEEP_GOING;
    
    if (ctdb_hit_feat)
    {
	/* ctdb_feat_t is t along *entire* LOS ray */
	ctdb_best_point[X] = ray_start[X] + 
	  ctdb_feat_t * (ray_end[X] - ray_start[X]);
	ctdb_best_point[Y] = ray_start[Y] + 
	  ctdb_feat_t * (ray_end[Y] - ray_start[Y]);
	ctdb_best_point[Z] = ray_start[ZL] + 
	  ctdb_feat_t * (ray_end[ZL] - ray_start[ZL]);
    }
    
    /* Sample for ground intersection (we could actually compute this,
     * but it would be a lot of work, and this sampling technique is
     * quite fast).
     */
    if (gi_check_dirt && (ctdb_hit_dirt || ctdb_hit_te))
    {
	if (!ctdb_hit_te || 
	    (ctdb_hit_dirt && (ctdb_hit_dirt_dist < ctdb_hit_te_dist)))
	{
	    if (gi_sample(ctdb, ctdb_best_point, ctdb_hit_dirt_px,
			  ctdb_hit_dirt_py, &ctdb_dirt_limit,
			  &ctdb_hit_dirt))
	    {
		ctdb_result = ctdb_hit_dirt;
		return VIS_STOP;
	    }
	}
	else
	{
	    if (gi_sample(ctdb, ctdb_best_point, ctdb_hit_te_px, 
			  ctdb_hit_te_py, &ctdb_te_limit,
			  &ctdb_hit_te))
	    {
		ctdb_result = ctdb_hit_te;
		return VIS_STOP;
	    }
	}
    }
    
    if (ctdb_hit_feat)
    {
	ctdb_result = ctdb_hit_feat;
	return VIS_STOP;
    }
    else
    {
	ctdb_result = ctdb_hit_dirt;
	return VIS_STOP;
    }
}

int32 
ctdb_find_ground_intersection(CTDB *ctdb, 
			      float64 x0, 
			      float64 y0, 
			      float64 z0,
			      float64 x1, 
			      float64 y1, 
			      float64 z1,
			      float64 *gi_pt, 
			      CTDB_ELEV_QUALIFIER qual, 
			      int32 n_veh,
			      CTDB_VEHICLE_LOCATION *veh, 
			      int32 ignore0,
			      int32 ignore1, 
			      int32 *hit_id)
{
    /* These obj function pointers are going to be modified on each call
     * so in an effort to save time the address of each objs function pointer
     * will be remembered to avoid the need to index into the obj data 
     * strcuture each time.
     */
    static CTDB_TYPE_FUNC *micro_h2o_func = 
      &gi_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_WATER)].func;
    static CTDB_ELEV_QUALIFIER
      check_trees=(CTDB_INCL_SINGLE_TREES|CTDB_INCL_LINEAR);

   /* These are the default flags for this API routine */
    static CTDB_ELEV_QUALIFIER dflt =(CTDB_MAX_LAND_AND_WATER |
				      CTDB_ALL_FEATURES       |
				      CTDB_INCL_SINGLE_TREES  |
				      CTDB_INCL_VEHICLES);


   /* These are the valid/support flags for this API routine --
    *
    * INCL_LINEAR is included here so that 
    * ctdb_find_ground_intersection can check for intersections with
    * wire obstacles.   
    */
    static CTDB_ELEV_QUALIFIER valid=(CTDB_ALL_FEATURES      |
				      CTDB_INCL_VEHICLES     |
				      CTDB_LAND_AND_WATER    |
				      CTDB_INCL_SINGLE_TREES |
				      CTDB_INCL_LINEAR       |
				      CTDB_RETURN_MAX); 
    /* These are the flags which are maniditory for this API routine */
    static CTDB_ELEV_QUALIFIER must = CTDB_RETURN_MAX;

    /* This is the mask that is used to determine the qualifer mask that 
     * should be used in the elevation lookup call.
     */
    static CTDB_ELEV_QUALIFIER valid_elev_qual = CTDB_MAX_LAND_AND_WATER;

    float64 veh_t;
    int32 hit_veh;
    
    if (!ctdb)
    {
	VMAT3_VEC_ZERO64(gi_pt);
	return CTDB_HIT_NOTHING;
    }   

    gi_qual    = (qual?(qual&valid)|must:dflt);
    
    /* Check for vehicle intersections before line clipping so we
     * can use this to detect collisions off-database.
     */
    if ( gi_qual & CTDB_INCL_VEHICLES )
      hit_veh = vehicle_intersect(x0, y0, z0, x1, y1, z1,
				  n_veh, veh, ignore0, ignore1,
				  hit_id, &veh_t);
    else
      hit_veh = 0;
    
    /* Initialize 'global'ish variables */
    ctdb_hit_dirt   = CTDB_HIT_NOTHING;
    ctdb_hit_te     = CTDB_HIT_NOTHING;
    ctdb_hit_feat   = CTDB_HIT_NOTHING;
    ctdb_result     = CTDB_HIT_NOTHING;
 
    if ( gi_qual & CTDB_INCL_WATER )
    {
	gi_first_water  = TRUE;
	ctdb_post_flags = POST_MICROTERRAIN;
	gi_no_water     = FALSE;
	*micro_h2o_func = (CTDB_TYPE_FUNC)gi_micro_water_compare;
	
    }
    else
    {
	ctdb_post_flags = 0;
	gi_no_water     = TRUE;
	*micro_h2o_func = (CTDB_TYPE_FUNC)ctdb_invalid_type;
    }    

    /* MICROTERRAIN AND TERRAIN PROCESSING */
    if (gi_check_dirt =  ( gi_qual & CTDB_INCL_LAND ))
      ctdb_post_flags |= POST_MICROTERRAIN;     
    
    /* LAID LINEAR PROCESSING will be handled by finding the ground 
     * intersection (CTDB_HIT_GROUND) and then performing and elevation
     * look up to see if the intersection point occurred with in a 
     * laid linear feature.  This is more efficient that checking 
     * each laid linear since the an elevation loookup is required 
     * for each edge crossing since laid linears are assumed to lay 
     * on the terrain.
     */

    /* VOLUME PROCESSING */
    if ( gi_qual & CTDB_INCL_VOLUME)
      ctdb_post_flags |= POST_BUILDINGS;
	
    /* LINEARS PROCESSING */
    if ( gi_qual & check_trees )
      ctdb_post_flags |= POST_TREES;
    else if ( gi_qual & CTDB_INCL_CANOPY )
      ctdb_post_flags |= POST_TREES;
    
    ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, z1, z1, 0.0, gi_cell_fcn);
    
    if (ctdb_result == CTDB_HIT_NOTHING)
    {
	if (hit_veh)
	{
	    if (gi_pt)
	    {
		gi_pt[0] = x0 + veh_t * (x1 - x0);
		gi_pt[1] = y0 + veh_t * (y1 - y0);
		gi_pt[2] = z0 + veh_t * (z1 - z0);
	    }
	    return CTDB_HIT_VEHICLE;
	}
	else
	{
	    if (gi_pt)
	      VMAT3_VEC_ZERO64(gi_pt);
	    return CTDB_HIT_NOTHING;
	}
    }
    else
    {
	float64 best_t, dx;
	
        if (ctdb_is_tiled)
	  gcs_gcs3_to_gcscs(ctdb_last_cell, ctdb_best_point, 
			     ctdb_get_cell_id(ctdb), ctdb_best_point);
	
	dx = x1 - x0;
	if (dx)
	{
	    best_t = (ctdb_best_point[X] - x0) / dx;
	}
	else
	{
	    best_t = (ctdb_best_point[Y] - y0) / (y1 - y0);
	}
	
	if ((! hit_veh) || (best_t <= veh_t))
	{
	    if (gi_pt)
	    {
		VMAT3_VEC_COPY64(ctdb_best_point, gi_pt);
	    }
	    return ctdb_result;
	}
	else
	{
	    if (gi_pt)
	    {
		gi_pt[0] = x0 + veh_t * (x1 - x0);
		gi_pt[1] = y0 + veh_t * (y1 - y0);
		gi_pt[2] = z0 + veh_t * (z1 - z0);
	    }
	    return CTDB_HIT_VEHICLE;
	}
    }
}



/* This public function must be called *just after* a call to
 * ctdb_find_ground_intersection().  It returns the endpoints of the treeline
 * segment intersected in ctdb_find_ground_intersection().  If no treeline
 * segment was intersected in the last call to ctdb_ground_intersection(),
 * then zeros are returned.
 */
void 
ctdb_get_treeline_segment(CTDB *ctdb, float64 *pt0, float64 *pt1)
{
    if (ctdb && ( ctdb_hit_feat == CTDB_HIT_TREELINE ))
    {
	float64 t1[3], t2[3];
	CTDB *pt_ctdb;
	int32 ret_cell;
	
	if (!ctdb_is_tiled)
	{
 	    pt_ctdb = ctdb;
	    /* ret_cell is ignored in this case */
	}
	else
	{
	    pt_ctdb = (CTDB *)gcs_get_tdb(ctdb_last_cell, FALSE);
	    ret_cell = ctdb_get_cell_id(ctdb);
	}
	
	t1[X] = CTDB_XYO_TO_METERS(ctdb_treeline_segment_start.x, 
				   ctdb_treeline_patch_x,
				   pt_ctdb->patch_incrD,
				   pt_ctdb->incrD);
	t1[Y] = CTDB_XYO_TO_METERS(ctdb_treeline_segment_start.y, 
				   ctdb_treeline_patch_y,
				   pt_ctdb->patch_incrD,
				   pt_ctdb->incrD);
	t1[Z] = ctdb_lookup_qual_elevation(pt_ctdb, t1[X], t1[Y],
					   CTDB_MAX_LAND_AND_WATER, 
					   CTDB_NO_REF_ELEV, 
					   (CTDB_ELEV_DATA *)NULL);
	
	t2[X] = CTDB_XYO_TO_METERS(ctdb_treeline_segment_end.x, 
				   ctdb_treeline_patch_x,
				   pt_ctdb->patch_incrD,
				   pt_ctdb->incrD);
	t2[Y] = CTDB_XYO_TO_METERS(ctdb_treeline_segment_end.y, 
				   ctdb_treeline_patch_y,
				   pt_ctdb->patch_incrD,
				   pt_ctdb->incrD);
	t2[Z] = ctdb_lookup_qual_elevation(pt_ctdb, t2[X], t2[Y],
					   CTDB_MAX_LAND_AND_WATER, 
					   CTDB_NO_REF_ELEV, 
					   (CTDB_ELEV_DATA *)NULL);
	
	if ((ctdb_is_tiled) &&
	    (ret_cell != ctdb_last_cell))
	{
	    gcs_gcs3e_to_gcscs(ctdb_last_cell, t1[X], t1[Y], t1[Z],
			     ret_cell, &pt0[X], &pt0[Y], &t1[Z]);
	    gcs_gcs3e_to_gcscs(ctdb_last_cell, t2[X], t2[Y], t2[Z],
			     ret_cell, &pt1[X], &pt1[Y], &t2[Z]);
	    return;
	}
	else
	{
	    VMAT2_VEC_COPY64(t1, pt0);
	    VMAT2_VEC_COPY64(t2, pt1);
	}
    }
    else
    {
	pt0[X] = 0.0;  pt0[Y] = 0.0;
	pt1[X] = 0.0;  pt1[Y] = 0.0;
    }
}

/* ctdb_point_to_background projects a ray from an observer through 
 * a target and returns information about the target's background.
 */
int32 
ctdb_find_background(CTDB *ctdb, 
		     float64 x0, 
		     float64 y0, 
		     float64 z0, 
		     float64 x1, 
		     float64 y1, 
		     float64 zl, 
		     float64 zh, 
		     float64 width, 
		     float64 range,
		     CTDB_ELEV_QUALIFIER qual,
		     CTDB_BACKGROUND *output_ptr)
{
   /* These are the default flags for this API routine */
    static CTDB_ELEV_QUALIFIER dflt =(CTDB_MAX_LAND_AND_WATER |
				      CTDB_ALL_FEATURES       |
				      CTDB_INCL_SINGLE_TREES);
    /* These are the queries that are valid and supported in this API */
    static CTDB_ELEV_QUALIFIER valid=(CTDB_ALL_FEATURES      |
				      CTDB_LAND_AND_WATER    |
				      CTDB_INCL_SINGLE_TREES |
				      CTDB_RETURN_MAX); 
    /* These are the flags which are maniditory for this API routine */
    /*static CTDB_ELEV_QUALIFIER must = ;*/

    float64 dx, dy, dzl, dzh, dz;
    float64 dist, t;
    float64 z1;
    float64 px, py, pz, pzh, pw; /* projected values */
    
    /*
     * This routine is actually only interested in the stuff 
     * *behind* the target. We will use ctdb_find_ground_intersection(),
     * but we will use the target as the viewpoint, and the 
     * projection from the original viewpoint through the target
     * to the end of the range as the new target. 
     *
     * This approach will be insufficient is we want to model
     * composite backgrounds (e.g, half ground/half sky). In that
     * case, we will need to base this functionality on 
     * ctdb_point_to_point().
     *
     * We do the projection using parametric equations.
     * The parameter (t) is defined to yield the eye point at t=0, and
     * passed target at t=1.  We compute the parameter for the range
     * passed, and plug it back into the equations to find the target
     * at that range.
     */
    
    /* validate the input qualifier */
    qual = (qual?(valid&qual):dflt);
    
    /* We calculate new eyepoint at half-height on the target */
    z1 = (zl + zh)*0.5;
    
    dx = x1 - x0; dy = y1 - y0; dzl = zl - z0; dzh = zh - z0;
    dz = (dzl + dzh) * 0.5;
    
    dist = fsqrt(dx * dx + dy * dy + dz * dz);
    
    t = range / dist;
    
    px = x0 + dx * t;
    py = y0 + dy * t;
    pz = z0 + dzl * t;
    pzh = z0 + dzh * t;
    
    /* Calculate new target for only a single ray: */
    /* Actually, this is only useful if we convert to ctdb_point_to_point() */
    pz = (pz + pzh)*0.5;
    pzh = pz + .001;
    pw = 0.001; /* pw = width * t; */
    
    background.type = 0;
    background.location[0] = background.location[1] = background.location[2] = 
      0.0;
    background.u.terrain_type = 0;
    do_background = 1;
    background.type = 
      ctdb_find_ground_intersection(ctdb, x1, y1, z1, px, py, pz, 
				    background.location, qual, (int32)0, 
				    (CTDB_VEHICLE_LOCATION *)NULL, 
				    (int32)0, (int32)0, (int32*)NULL);
    do_background = 0;
    
    /* Look up the subtype, if necessary: */
    switch (background.type)
    {
      case CTDB_HIT_GROUND:
	background.u.terrain_type = 
	  ctdb_lookup_soil_ml(ctdb, background.location[X], 
			      background.location[Y],background.location[Z]);
	break;
    }
    
    if (output_ptr)
      *output_ptr = background;
    
    return 0;
}

/* void ctdb_print_background prints a text description of the passed
 * CTDB_BACKGROUND info onto stdout.
 */
void ctdb_print_background(CTDB_BACKGROUND *background)
{
    char *type_str;
    char *subtype_str;
    char buffer[100];
    static char *none_string = "None";
    
    buffer[0] = 0;
    subtype_str = buffer;
    switch(background->type)
    {
      case 0:
	type_str = "Sky";
	subtype_str = none_string;
	break;
      case CTDB_HIT_GROUND:
	type_str = "Terrain";
	sprintf(buffer, "%d", background->u.terrain_type);
	break;
      case CTDB_HIT_BUILDING:
	type_str = "Building";
	subtype_str = none_string;
	break;
      case CTDB_HIT_TREELINE:
	type_str = "Treeline";
	subtype_str = none_string;
	break;
      case CTDB_HIT_TREE:
	type_str = "Tree";
	subtype_str = none_string;
	break;
      case CTDB_HIT_VEHICLE:
	type_str = "Vehicle";
	subtype_str = none_string;
	break;
      case CTDB_HIT_WATER:
	type_str = "Water";
	subtype_str = none_string;
	break;
      default:
	type_str = "Unknown";
	subtype_str = none_string;
	break;
    }
    
    printf("Background type: %s... subtype: %s\n", type_str, subtype_str);
    printf("    at <%.2f, %.2f, %.2f>\n", background->location[0],
	   background->location[1], background->location[2]);
}

