#ifndef lint
static char rcsid [] = "$RCSfile: ct_hg.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_hg.c                                                           *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to find the highest point along a ray                *
 
 *       Author: jesmith                                                     *
 *       Modified for GCS by flchamberlain 11/6/95                           *
 *       Modified for OceanRep by cbuettner 3/96                             *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/
#include <stdext.h>
#include <libgcs.h>
#include "libctdb_local.h"

/* $$$ The meaning of this routine changes when we move to GCS.  Before
 * GCS, this routine answered at least two distinct questions whose
 * answers happened to be identical:
 *
 * 1.  What is the point on the terrain along the ray furthest above the
 * the reference ellipsoid?
 * 2.  What is the point on the terrain along the ray with the highest z-value
 * 
 * Unfortunately, under GCS the answers to these two questions will not always
 * be the same.  The current implementation answers question 2 but not question
 * 1.  If the distinction is critical to some caller, it may be necessary to
 * split this routine in two, one version for each question.
 */

/* Global Local Variables */
static int32            ctdb_no_water;
static CTDB_FEAT_TYPE   hg_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];
static CTDB_FEATURE_OBJ hg_obj[FC_CLASS_MAX];
static CTDB_ELEV_QUALIFIER hg_qual;

static CTDB_COORD3D ctdb_te_hg_pt;
static CTDB_COORD3D ctdb_hg_pt;
static CTDB_COORD3D ctdb_best_so_far;
static int32 ctdb_result_cell;	/* Frame of reference in which answer should
				 * be returned
				 */
typedef struct ctdb_hg_obj_data
{
    CTDB_LOS_IN_XYO_AND_DB los ;  
    CTDB_PATCH_DATA        patch ;
    int32                  postmask;
} CTDB_HG_OBJ_DATA;

/* x,y are in post coordinates in ctdb's native coordinate system */
static int32 
hg_post_compare(
		CTDB *ctdb, 
		float64 x, 
		float64 y, 
		float64 zl, 
		float64 zh,		
		float64 z_cross,
		int32   water)
{
    if ( water && ctdb_no_water )
      return VIS_KEEP_GOING;
    
    /* Figure out where the posts cross the line and check the elevation.
     */
    if (z_cross > ctdb_hg_pt.z)
    {
	ctdb_hg_pt.x = x * ctdb->incrD;
	ctdb_hg_pt.y = y * ctdb->incrD;
	ctdb_hg_pt.z = z_cross;
    }
    
    return VIS_KEEP_GOING;
} 

/* LOS is entire intervis ray stored in current coord system */
static int32 
hg_te_edge_compare(
		   CTDB *ctdb, 
		   CTDB_LOS_IN_XYO_AND_DB *los, 
		   float64 inz0, 
		   float64 inz1, 
		   float64 tedge, 
		   float64 tray)
{
    float64 z;
    
    z = (inz1 - inz0) * tedge + inz0;
    
    if (z > ctdb_te_hg_pt.z)
    {
	ctdb_te_hg_pt.x =
	  los->db.start.x + tray * (los->db.end.x - los->db.start.x);
	ctdb_te_hg_pt.y =
	  los->db.start.y + tray * (los->db.end.y - los->db.start.y);
	ctdb_te_hg_pt.z = z;
    }
    
    return VIS_KEEP_GOING;
}

/* LOS is entire intervis ray stored in current coord system */
static int32 
hg_edge_compare(
		CTDB_FEATURE_DATA      *v0, 
		CTDB_FEATURE_DATA      *v1, 
		CTDB_LOS_IN_XYO_AND_DB *los
		)
{
    float64 tray, tedge;
    float64 z;
    
    /* Figure out where the posts cross the line and store that location
     * and elevation.
     */
    
    if (ctdb_segment_intersection_t(&los->xyo,v0->xy, v1->xy, 
				    &tray, &tedge) == DISJOINT)
      return VIS_KEEP_GOING;
    
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    
    if (z > ctdb_hg_pt.z)
    {
	ctdb_hg_pt.x = los->db.start.x + tray * (los->db.end.x - 
						 los->db.start.x);
	ctdb_hg_pt.y = los->db.start.y + tray * (los->db.end.y - 
						 los->db.start.y);
	ctdb_hg_pt.z = z;
    }
    return VIS_KEEP_GOING;
}

static int32 
hg_micro_invalid(CTDB               *ctdb,
		 CTDB_FEAT_TYPE_PTR  sub,
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_HG_OBJ_DATA *data)
{
    return VIS_KEEP_GOING;
}

static int32
hg_micro_compare(CTDB               *ctdb,
		 CTDB_FEAT_TYPE_PTR  sub,
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_HG_OBJ_DATA   *data
		 )
{
#ifdef WATER_DEBUG
    printf("MCR ,");
#endif
    feature += 2;

    hg_edge_compare(feature,
		    feature+(sub->incr*feature[2].micro.vert0), 
		    &(data->los));
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    hg_edge_compare(feature,
		    feature+(sub->incr*feature[2].micro.vert1),
		    &(data->los));
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    hg_edge_compare(feature,
		    feature+(sub->incr*feature[2].micro.vert2), 
		    &(data->los));    
    return VIS_KEEP_GOING; 
}
static int32
hg_water_edge_compare(
		      CTDB_FEATURE_DATA      *v0, 
		      CTDB_FEATURE_DATA      *v1,
		      CTDB_LOS_IN_XYO_AND_DB *los, 
		      float32                 tide
		      )
{
    float64 tray, tedge;
    float64 z_cross;
    
    /* We treat DISJOINT and COLINEAR the same way:
     * DISJOINT means our line-of-sight cannot be blocked by this
     * segment.
     * COLINEAR segments may or may not block our line of sight, but
     * it doesn't matter: if we are dealing with closed polygons (such
     * as microterrain or a house) the tips of the other segments in
     * the polygon will block or not block our visibility as
     * appropriate; if we are dealing with an open list of segments
     * (such as a treeline) a colinear segment would have 0 width, and
     * hence could not block intervisibility of an object with >0 width.
     */
    if (ctdb_segment_intersection_t(&los->xyo,v0->xy, v1->xy,&tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;

    /* This is a little inefficient but the number of micro water polygons
     * that will be processed in a database are very small.
     */
    z_cross = (v1[1].z - v0[1].z) * tedge + v0[1].z + tide;

    if (z_cross > ctdb_te_hg_pt.z)
    {
	ctdb_te_hg_pt.x =
	  los->db.start.x + tray * (los->db.end.x - los->db.start.x);
	ctdb_te_hg_pt.y =
	  los->db.start.y + tray * (los->db.end.y - los->db.start.y);
	ctdb_te_hg_pt.z = z_cross;
    }
    return VIS_KEEP_GOING;
}

static int32
hg_micro_water_compare(CTDB               *ctdb,
		       CTDB_FEAT_TYPE_PTR  sub, 
		       CTDB_FEATURE_DATA  *feature, 
		       CTDB_HG_OBJ_DATA   *data
		       )
{
    float32 tide1,tide2;
#ifdef WATER_DEBUG
    printf("WMCR ,");
#endif
    
    feature+= 2;
    
    tide1 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char0);
    
    if (!hg_water_edge_compare(feature,
			       feature+(sub->incr*feature[2].micro.vert0), 
			       &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    if (!hg_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert1),
				 &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    tide2 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char1);

    /* Since the tidal level may shift at the center edge of a polygon
     * check it again with the new tidal level if the tidal level is 
     * different from the first.
     */
    if ((tide2 != tide1) &&
	(!hg_water_edge_compare(feature,
				  feature+(sub->incr*feature[2].micro.vert0),
				  &(data->los),tide2)))
	  return VIS_STOP;
    
    if (!hg_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert2), 
				 &(data->los),tide2))
      return VIS_STOP; 
    
    return VIS_KEEP_GOING;
}

static int32
hg_check_micro(
		 CTDB               *ctdb,
		 CTDB_FEAT_OBJ_PTR   obj, 
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_HG_OBJ_DATA   *data,
		 int32              *next
		 )
{
    static CTDB_FEAT_TYPE *sub=hg_micro_subclass_obj;
    int32 verts = feature->info.vertices;

    /* check if the micro type has changed */
    if ((verts & MICRO_TYPE) != sub->type)
      sub = &hg_micro_subclass_obj[MICRO_TYPE_INDEX(verts)];

    *next = sub->incr;
    
    if (!obj->valid || 
	!(data->postmask & feature->integer) ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return VIS_KEEP_GOING;
#ifdef CTDB_OBJ_DEBUG
    printf("micro ,");
#endif

    /* call the micro type processing function */
    return sub->func(ctdb,sub,feature,data);
}

int32
ctdb_hg_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static CTDB_TYPE_FUNC micro_func[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)hg_micro_compare,      /* MICRO_TYPE_BASE */ 
	(CTDB_TYPE_FUNC)hg_micro_compare,      /* MICRO_TYPE_ML */
	(CTDB_TYPE_FUNC)hg_micro_compare,      /* MICRO_TYPE_DEFAULT */
	(CTDB_TYPE_FUNC)hg_micro_water_compare /* MICRO_TYPE_WATER */
    };
    
    /* for high ground processing microterrain is the only valid feature
     * to process.
     */
    obj[FC_MICRO].valid   = TRUE;
    obj[FC_MICRO].func = (CTDB_OBJ_FUNC)hg_check_micro;
    ctdb_init_micro_subclass(hg_micro_subclass_obj,micro_func);

    obj[FC_VOLUME].valid          = FALSE;
    obj[FC_VOLUME].func           = 0;
    obj[FC_LINEAR].valid          = FALSE;
    obj[FC_LINEAR].func           = 0;
    obj[FC_CANOPY].valid          = FALSE;
    obj[FC_CANOPY].func           = 0;
    obj[FC_LAID_LINEAR].valid     = FALSE;
    obj[FC_LAID_LINEAR].func      = 0;
    obj[FC_AGGREGATE].valid       = FALSE;
    obj[FC_AGGREGATE].func        = 0;

    /* save the initialized structure locally */
    bcopy(obj,hg_obj,sizeof(hg_obj));
    
    return 1;
}


/* start, end are the endpoints of the *entire* intervis ray, represented in
 * the frame of reference specified by ctdb.
 */
static int32 
hg_patch_compare(
		 CTDB    *ctdb, 
		 int32    patch_number, 
		 int32    feature_mask, 
		 int32    post_mask, 
		 float64 *start, 
		 float64 *end
		 )
{
    int32 ray_info,i,next;
    CTDB_FEATURE_DATA *feature, *last;
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_HG_OBJ_DATA   data;

    feature = (CTDB_FEATURE_DATA *)ctdb_lookup_feature(ctdb, patch_number);

    if (!feature)
      return VIS_KEEP_GOING;
	
    /* It's faster to compute these than pass them in.
     */
    data.postmask     = post_mask;
    data.patch.number = patch_number;
    data.patch.x      = data.patch.number % ctdb->patches_wide;
    data.patch.y      = data.patch.number / ctdb->patches_wide;
    
    data.los.xyo.start.x = 
      POST_TO_XYO(start[X],data.patch.x,ctdb->patch_inv_incr);
    data.los.xyo.start.y = 
      POST_TO_XYO(start[Y],data.patch.y,ctdb->patch_inv_incr);
    data.los.xyo.end.x   = 
      POST_TO_XYO(end[X],data.patch.x,ctdb->patch_inv_incr);
    data.los.xyo.end.y   = 
      POST_TO_XYO(end[Y],data.patch.y,ctdb->patch_inv_incr);
    data.los.db.start.x = start[X] * ctdb->incrD;
    data.los.db.start.y = start[Y] * ctdb->incrD;
    data.los.db.end.x   = end[X] * ctdb->incrD;
    data.los.db.end.y   = end[Y] * ctdb->incrD;
    
#ifdef CTDB_OBJ_DEBUG
    printf("\npatch number %d (%d,%d)\t",patch_number,
	   data.patch.x,data.patch.y);
#endif    

    last = feature + feature->hdr.size;
    feature++;

    obj = &hg_obj[FC_MICRO];

    for(;feature<last; feature+=next)
    {
	int32 fclass=feature->info.feature_class;	
	/* if the current class doesn't match the previos class (FC_MICRO) 
	 * then return since only terrain is to be evaluted in this routine.
	 */
	if ( fclass != FC_MICRO )
	  return VIS_KEEP_GOING;

	obj->func(ctdb,obj,feature,(ADDRESS)&data,&next);
	
    }
    return VIS_KEEP_GOING;
}


static int32
hg_cell_fcn(CTDB *ctdb, float64 *ray_start, float64 *ray_end,
	    float64 *cur_start, float64 *cur_end, float64 width)
{
    int32 cur_cell;
    float64 ray_start_post[3];
    float64 ray_end_post[4];
    float64 cur_start_post[4];
    float64 cur_end_post[4];
    CTDB_COORD3D start, end, cell_start, cell_end, cell_high_pt;
    
    /* Scale from meters to grid spacing */
    ray_start_post[X] = ray_start[X] * ctdb->inv_incr;
    ray_start_post[Y] = ray_start[Y] * ctdb->inv_incr;
    ray_start_post[ZL] = ray_start[ZL];
    
    ray_end_post[X] = ray_end[X] * ctdb->inv_incr;
    ray_end_post[Y] = ray_end[Y] * ctdb->inv_incr;
    ray_end_post[ZL] = ray_end[ZL];
    ray_end_post[ZH] = ray_end[ZH];
    
    cur_start_post[X] = cur_start[X] * ctdb->inv_incr;
    cur_start_post[Y] = cur_start[Y] * ctdb->inv_incr;
    cur_start_post[ZL] = cur_start[ZL];
    cur_start_post[ZH] = cur_start[ZH];
    
    cur_end_post[X] = cur_end[X] * ctdb->inv_incr;
    cur_end_post[Y] = cur_end[Y] * ctdb->inv_incr;
    cur_end_post[ZL] = cur_end[ZL];
    cur_end_post[ZH] = cur_end[ZH];
    
    /* Copy endpoints into structure used by terrain_element_intervis */
    start.x = ray_start[X];
    start.y = ray_start[Y];
    start.z = ray_start[Z];
    end.x   = ray_end[X];
    end.y   = ray_end[Y];
    end.z   = ray_end[ZL];
    cell_start.x = cur_start[X];
    cell_start.y = cur_start[Y];
    cell_start.z = cur_start[ZL];
    cell_end.x = cur_end[X];
    cell_end.y = cur_end[Y];
    cell_end.z = cur_end[ZL];
    
    
    
    /* Seed with start point */
    ctdb_hg_pt.x = ctdb_te_hg_pt.x = cur_start[X];
    ctdb_hg_pt.y = ctdb_te_hg_pt.y = cur_start[Y];
    
    ctdb_hg_pt.z = ctdb_te_hg_pt.z =
      ctdb_lookup_qual_elevation(ctdb, cur_start[X], cur_start[Y], hg_qual,
				 CTDB_NO_REF_ELEV,(CTDB_ELEV_DATA *)NULL);
    
    /* call terrain element intervis. if there are tinned patches */
    if ( ctdb->is_tinned ) 
      (void) ctdb_terrain_element_intervis(ctdb,&start,&end,ray_end[ZH],
					   &cell_start, &cell_end,
					   hg_te_edge_compare,
					   FALSE);
    
    /* if gridded patches or water (patch) should be looked at, then 
     * traverse gridded patches 
     */
    if ( ctdb->is_gridded || !ctdb_no_water) 
      /* Invoke the traversal engine for point in between */
      (void) ctdb_intervis(ctdb, ray_start_post, ray_end_post, 
			   cur_start_post, cur_end_post, TRUE,
			   hg_post_compare, hg_patch_compare, 
			   (ctdb->is_gridded?POST_MICROTERRAIN:0));
    
    /* Those two calls generated separate high points.  If the heights are
     * the same, choose the one closest to the start of the ray
     */
    if (ctdb_hg_pt.z > ctdb_te_hg_pt.z)
    {
        cell_high_pt.x = ctdb_hg_pt.x;
	cell_high_pt.y = ctdb_hg_pt.y;
	cell_high_pt.z = ctdb_hg_pt.z;
    }
    else if (ctdb_te_hg_pt.z > ctdb_hg_pt.z)
    {
	cell_high_pt.x = ctdb_te_hg_pt.x;
	cell_high_pt.y = ctdb_te_hg_pt.y;
	cell_high_pt.z = ctdb_te_hg_pt.z;
    }
    else 
    {
	float64 dx, dy, dxte, dyte;
	
	dx = fabs(ctdb_hg_pt.x - ray_start[X]);
	dy = fabs(ctdb_hg_pt.y - ray_start[Y]);
	dxte = fabs(ctdb_te_hg_pt.x - ray_start[X]);
	dyte = fabs(ctdb_te_hg_pt.y - ray_start[Y]);
	
	if ((dx < dxte) || ((dx == dxte) && (dy < dyte)))
	{
	    cell_high_pt.x = ctdb_hg_pt.x;
	    cell_high_pt.y = ctdb_hg_pt.y;
	    cell_high_pt.z = ctdb_hg_pt.z;
	}
	else
	{
	    cell_high_pt.x = ctdb_te_hg_pt.x;
	    cell_high_pt.y = ctdb_te_hg_pt.y;
	    cell_high_pt.z = ctdb_te_hg_pt.z;
	}
    }
    
    cur_cell = ctdb_get_cell_id(ctdb);
    if (cur_cell != ctdb_result_cell)
    {
	gcs_gcs3e_to_gcscs(cur_cell, cell_high_pt.x, cell_high_pt.y,
			 cell_high_pt.z,
			 ctdb_result_cell,&cell_high_pt.x, &cell_high_pt.y,
			 &cell_high_pt.z);
    }
    
    if (cell_high_pt.z > ctdb_best_so_far.z)
    {
	/* We're counting on the fact that cells are traversed in order here--
	 * if the z's are equal, we can safely assume that the first one we
	 * found was the one we wanted.
	 */
	ctdb_best_so_far = cell_high_pt;
    }
    
    return VIS_KEEP_GOING;
}


float64 
ctdb_find_high_ground(CTDB               *ctdb, 
		      float64             x0, 
		      float64             y0,
		      float64             x1, 
		      float64             y1, 
		      CTDB_ELEV_QUALIFIER qual,
		      float64            *pt_ret)
{
    /* pointer is needed to allow quick remapping instead of always reindexing
     */
    static CTDB_TYPE_FUNC *water_func=
      &hg_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_WATER)].func;
    /* Define the DEFAULT and VALID ctdb_elev_qualifier bits */
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_QUALIFIER valid=CTDB_MAX_LAND_AND_WATER;
    float64 z0, z1;
    
    if (!ctdb)
    {
	if (pt_ret)
	{
	    pt_ret[X] = 0.0;
	    pt_ret[Y] = 0.0;
	}
	return 0.0;
    }
    
    ctdb_result_cell = ctdb_get_cell_id(ctdb);
    
    /* check the qualifier, validate it or set the default */
    hg_qual = (qual?(qual&valid):dflt)&CTDB_MAX_LAND_AND_WATER;
    
    /* Remap function pointer and control flag for water or no water 
     * processing
     */
    if ( hg_qual & CTDB_INCL_WATER )
    {
	ctdb_no_water = FALSE;
	*water_func   = (CTDB_TYPE_FUNC)hg_micro_water_compare;
    }
    else
    {
	ctdb_no_water = TRUE;
	*water_func   = (CTDB_TYPE_FUNC)ctdb_invalid_type;
    }
	  
    /* Grab z-values for the endpoints from the terrain */
    z0 = ctdb_lookup_qual_elevation(ctdb, x0, y0, hg_qual, 
				    CTDB_NO_REF_ELEV, (CTDB_ELEV_DATA *)NULL);
    z1 = ctdb_lookup_qual_elevation(ctdb, x1, y1, hg_qual, 
				    CTDB_NO_REF_ELEV, (CTDB_ELEV_DATA *)NULL);
    
    /* Seed with starting point */
    ctdb_best_so_far.x = x0;
    ctdb_best_so_far.y = y0;
    ctdb_best_so_far.z = -1.0e18; /* Very negative number.  We can't be
				   * completely safe using the lookup-up
				   * elevation for the start point because
				   * of the possibility that it's off the
				   * database*/
    
    /* Run through all cells crossed by ray looking for highest point */
    ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, z1, 0.0, 0.0, hg_cell_fcn);
    
    
    
    /* Those two calls generated separate high points.  If the heights are
     * the same, choose the one closest to the start of the ray
     */
    if (pt_ret)
    {
	pt_ret[X] = ctdb_best_so_far.x;
	pt_ret[Y] = ctdb_best_so_far.y;
    }
    return ctdb_best_so_far.z;
}
