#ifndef lint
static char rcsid [] = "$RCSfile: ct_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_init.c                                                         *
*                                                                           *
*       Copyright 1995 TASC/SAIC                                            *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Driver               *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7612                                                        *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libctdb                                *
*       Created: January 1995                                               *
*       Author: flchamberlain                                               *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/ 
#include "libctdb_local.h"
#include <stdalloc.h>  /*common/include/global*/
#include <stdstring.h> /*common/include/global*/
#include <stdcoord.h>  /*common/include/global*/
#include <stdext.h>    /*common/include/global*/
#include <libgcs.h>

/* Library global */
int32 ctdb_is_tiled = FALSE;	/* Disable tiling by default */
static int32          ctdb_total_malloced;

/* The following global is only used to save the pointer to the roof array
 * for play(), the temporary hack which forces the roofline of a mes volume
 * into the patch table.
 */
static CTDB_MES_LIST_VERTEX_F64  *g_roof_array;

static void 
ctdb_init( void )
{
   static int32 ctdb_initted=0;
   static CTDB_FEATURE_OBJ feat_obj[FC_CLASS_MAX];

   if ( ctdb_initted )
     return;
   
   ctdb_initted=1;
   
   /* Class MICRO */
   feat_obj[FC_MICRO].sensClass      = FC_MICRO;
   feat_obj[FC_MICRO].featmask   = POST_MICROTERRAIN;
   feat_obj[FC_MICRO].inclmask   = CTDB_INCL_MICRO;
   feat_obj[FC_MICRO].valid      = 1;
   /* Class VOLUME */
   feat_obj[FC_VOLUME].sensClass     = FC_VOLUME;
   feat_obj[FC_VOLUME].featmask  = POST_BUILDINGS;
   feat_obj[FC_VOLUME].inclmask  = CTDB_INCL_VOLUME;
   feat_obj[FC_VOLUME].valid     = 1;
   /* Class LINEAR */
   feat_obj[FC_LINEAR].sensClass     = FC_LINEAR;
   feat_obj[FC_LINEAR].featmask  = POST_TREES;
   feat_obj[FC_LINEAR].inclmask  = CTDB_INCL_LINEAR;
   feat_obj[FC_LINEAR].valid     = 1;
   /* Class CANOPY */
   feat_obj[FC_CANOPY].sensClass     = FC_CANOPY;
   feat_obj[FC_CANOPY].featmask  = POST_TREES;
   feat_obj[FC_CANOPY].inclmask  = CTDB_INCL_CANOPY;
   feat_obj[FC_CANOPY].valid     = 1;
   /* Class LAID_LINEAR */
   feat_obj[FC_LAID_LINEAR].sensClass     = FC_LAID_LINEAR;
   feat_obj[FC_LAID_LINEAR].featmask  = POST_FEATURES;
   feat_obj[FC_LAID_LINEAR].inclmask  = CTDB_INCL_LAID_LINEAR;
   feat_obj[FC_LAID_LINEAR].valid     = 1;
   /* Class CANOPY */
   feat_obj[FC_AGGREGATE].sensClass     = FC_AGGREGATE;
   feat_obj[FC_AGGREGATE].featmask  = POST_TREES;
   /* Currently, aggregate features consist of only trees. */
   feat_obj[FC_AGGREGATE].inclmask  = CTDB_INCL_LINEAR;
   feat_obj[FC_AGGREGATE].valid     = 1;

   /* initialize the processing function pointers */
   ctdb_ptop_init_feature_class( feat_obj );
   ctdb_print_init_feature_class( feat_obj );
   ctdb_hg_init_feature_class( feat_obj );
   ctdb_gi_init_feature_class( feat_obj );
   ctdb_mr_init_feature_class( feat_obj );
   ctdb_prof_init_feature_class( feat_obj );
   ctdb_elev_init_feature_class( feat_obj );
   /* add more functions here */
}
  
ADDRESS
ctdb_alloc(ADDRESS p, uint32 size, char *reason)
{
    char *result;
    
    /* You can't alloc 0 bytes, so we bump it up here. */
    if (size == 0)
      size = 1;
    
    if (p == NULL)
      result = (char *) STDALLOC(size);
    else
    {
#ifdef _GT_
	STDDEALLOC (p);
	result = (char *) STDALLOC(size);
#else
	result = (char *) STDREALLOC(p, size);
#endif
    }
    
    if (!result)
    {
	printf("Allocation of %d bytes for %s failed.\n",size,reason);
	exit(1);
    }
    ctdb_total_malloced += size;
    return result;
}

/* Loops through _all_ of the cell data to find min and
 * max Z for the cell. 
 */
static void 
ctdb_find_z_limits(CTDB *ctdb)
{
    int32 post;
    CTDB_FEATURE_DATA *feature, *last;
    float64 z, min_z, max_z, p_minz, p_maxz;
    int32 px, py, px_min, py_min, patch_x, patch_y, patch_num;
    int32 i, max_patch_x, max_patch_y;
    int32 incr = ctdb->patch_incr;

    /* In order to find the min/max Z values, this routine has to scan
     * the entire database.  Presumably, bad data will be discovered here
     * rather than during a demo for some VIP.  Therefore, do not succumb
     * to the urge to simply store min/max Z in the header.  That would
     * save a teeny amount of time at startup and forfeit this wonderful
     * chance to detect problems early!
     */

    max_z = -1.0E+9;
    min_z = 1.0E+9;

    /* Avoid wasting time on empty patches */
    max_patch_x = (ctdb->max_x_post / incr);
    max_patch_y = (ctdb->max_y_post / incr);

    for(patch_y=0,py_min=1;patch_y<max_patch_y;
	patch_y++,py_min+=incr)
    {
	for(patch_x=0,px_min=0;patch_x<max_patch_x;
	    patch_x++,px_min+=incr)
	{
	    CTDB_PATCH_GROUP *pg;
	    int32 member;
	    patch_num = (patch_y * ctdb->patches_wide) + patch_x;
	    pg = ctdb->patch_groups + PG_NUMBER(patch_num);
	    member = PG_MEMBER(patch_num);

	    if ( ctdb_lookup_te_ptr(ctdb,patch_num) )
	    {
		/* Check TEs for patch */
		if (((member==0) && (pg->patch0_te16)) ||
		    ((member==1) && (pg->patch1_te16)) ||
		    ((member==2) && (pg->patch2_te16)) ||
		    ((member==3) && (pg->patch3_te16)))
		  ctdb_find_te_z_bounds_in_patch16(ctdb, patch_num, &p_minz, 
						   &p_maxz);
		else
		  ctdb_find_te_z_bounds_in_patch08(ctdb, patch_num, &p_minz, 
						   &p_maxz);
		if (p_minz < min_z) min_z = p_minz;
		if (p_maxz > max_z) max_z = p_maxz;

		if (((member == 0) && (pg->patch0_water)) || 
		    ((member == 1) && (pg->patch1_water)) || 
		    ((member == 2) && (pg->patch2_water)) || 
		    ((member == 3) && (pg->patch3_water)))
		{
		    z = pg->data_start.tin_data->max_elev[member];
		    if (z < min_z) min_z = z;
		    if (z > max_z) max_z = z;
		    
		    z = pg->data_start.tin_data->water_elev[member];
		    if (z < min_z) min_z = z;
		    if (z > max_z) max_z = z;
		}
	    }
	    else
	    {
		
		/* Ensure that the patch is in physical memory.
		 */
		if (pg->mapped)
		  ctdb->patch_hit_count++;
		else
		  ctdb_fetch_patch_group(ctdb, pg);

		if (((member == 0) && (pg->patch0_water)) || 
		    ((member == 1) && (pg->patch1_water)) || 
		    ((member == 2) && (pg->patch2_water)) || 
		    ((member == 3) && (pg->patch3_water)))
		{
		    z = pg->data_start.grid_data->max_elev[member];
		    if (z < min_z) min_z = z;
		    if (z > max_z) max_z = z;
		    
		    z = pg->data_start.grid_data->water_elev[member];
		    if (z < min_z) min_z = z;
		    if (z > max_z) max_z = z;
		}

		/* Check gridposts for patch */
		for(py=py_min;py<py_min+incr;py++)
		{
		    for(px=px_min;px<px_min+incr;px++)
		    {
			if ((px < ctdb->max_x_post) && (py < ctdb->max_y_post))
			{
			    post = ctdb_lookup_elev_post(ctdb, px, py);
			    z = GET_POST_ELEVATION(post, 
						   ctdb->fixed_point_basis);
			    if (z>max_z) max_z = z;
			    if (z<min_z) min_z = z;
			}
		    }
		}
	    }

	    /* Check features for patch */
	    if (feature=ctdb_lookup_feature(ctdb, patch_num))
	    {
		int32 next;
		last = feature + feature->hdr.size;
		for(feature++;feature<last;feature+=next)
		{
		    int32 verts=feature->info.vertices;
			
		    switch (feature->info.feature_class)
		    {
		      case FC_MICRO:
			z = feature[3].z;
			if (z>max_z) max_z = z;
			if (z<min_z) min_z = z;
			next = NEXT_MICRO(verts);
			break;
		      case FC_CANOPY:
			next = NEXT_CANOPY(verts);
			break;
		      case FC_VOLUME:
			next = NEXT_VOLUME(verts);
			break;
		      case FC_LINEAR:
			next = NEXT_LINEAR(verts);
			break;
		      case FC_LAID_LINEAR:
			next = NEXT_LAID_LINEAR(verts);
			break;
		      case FC_AGGREGATE:
			next = NEXT_AGGREGATE(verts);
			break;
		    }
		}
	    }
	}
    }
    ctdb->min_z = min_z;
    ctdb->max_z = max_z;
} 

void 
ctdb_unpack_header(CTDB *ctdb, CTDB_FILE_HEADER *hdr)
{
    ctdb->input_format          = hdr->format;
    ctdb->is_curved             = hdr->is_curved; 
    ctdb->is_gridded            = hdr->is_gridded;
    ctdb->is_tinned             = hdr->is_tinned;
    ctdb->is_short              = hdr->is_short;
    ctdb->fixed_point_basis     = hdr->fixed_point_basis;

    if (ctdb->fixed_point_basis) 
      ctdb->inv_fixed_point_basis = 1.0/ctdb->fixed_point_basis;

    ctdb->num_soil_tables = hdr->num_soil_tables;
    ctdb->soil_table_size = hdr->soil_table_size;
    ctdb->num_water_chars = hdr->num_water_chars;
    ctdb->num_tidal_zones = hdr->num_tidal_zones;

    ctdb->incr           = hdr->post_increment;
    ctdb->incrD          = ctdb->incr;
    ctdb->patch_incr     = hdr->posts_per_patch_side;
    ctdb->patch_incrD    = ctdb->patch_incr;

    if (ctdb->incrD)
      ctdb->inv_incr = 1.0/ctdb->incrD;

    if (ctdb->patch_incrD)
      ctdb->patch_inv_incr = 1.0/ctdb->patch_incrD;

    ctdb->pages_wide     = hdr->pages_wide;
    ctdb->pages_high     = hdr->pages_high;

    ctdb->patches_wide =
      ctdb->pages_wide * POSTS_PER_SIDE * ctdb->patch_inv_incr;
    ctdb->patches_high =
      ctdb->pages_high * POSTS_PER_SIDE * ctdb->patch_inv_incr;
    
    ctdb->min_x = hdr->min_x;
    ctdb->max_x = hdr->max_x;
    ctdb->min_y = hdr->min_y;
    ctdb->max_y = hdr->max_y;

    ctdb->min_x_post = POST_X(ctdb->min_x*ctdb->inv_incr);
    ctdb->max_x_post = POST_X(ctdb->max_x*ctdb->inv_incr);

    ctdb->min_y_post = POST_Y(ctdb->min_y*ctdb->inv_incr);
    ctdb->max_y_post = POST_Y(ctdb->max_y*ctdb->inv_incr);

    bcopy((char *)hdr->patch_cache_size_map, 
	  (char *)ctdb->patch_cache_size_map,
	  CTDB_PATCH_CACHES * sizeof(int32));

    ctdb->origin_northing    = hdr->origin_northing;
    ctdb->origin_easting     = hdr->origin_easting;
    ctdb->origin_zone_number = hdr->origin_zone_number;
    ctdb->origin_zone_letter = hdr->origin_zone_letter;
    ctdb->datum              = hdr->datum;

    bcopy((char *)hdr->name, (char *)ctdb->name, sizeof(hdr->name));

    ctdb->version            = hdr->version;

    bcopy((char *)hdr->date, (char *)ctdb->date, sizeof(hdr->date));
    strtok(ctdb->date,       "\n");

    ctdb->num_nodes        = hdr->num_nodes;
    ctdb->node_size        = hdr->node_size;
    ctdb->num_edges        = hdr->num_edges;
    ctdb->edge_size        = hdr->edge_size;
    ctdb->num_quad_data    = hdr->num_quad_data;
    ctdb->n_patch_bytes    = hdr->n_patch_bytes;
    ctdb->num_linear_models = hdr->num_linear_models;
    ctdb->num_volume_models = hdr->num_volume_models;

    ctdb->num_aggregate_models = hdr->num_aggregate_models;

    ctdb->num_mes_volumes    = hdr->num_mes_volumes;
    ctdb->num_mes_templates  = hdr->num_mes_templates;

    ctdb->meters_to_patch  = ctdb->patch_inv_incr * ctdb->inv_incr;
    ctdb->max_patch_number = ctdb->patches_wide * ctdb->patches_high - 1;
    ctdb->min_patch_number = 0;
    return;
}

static void
ctdb_unpack_aux_header(CTDB *ctdb, CTDB_AUX_FILE_HEADER *aux_hdr)
{
    ctdb->gcs_mode  = aux_hdr->gcs_mode;
    ctdb->cell_id   = aux_hdr->cell_id;
    ctdb->origin_x  = aux_hdr->origin_x;
    ctdb->origin_y  = aux_hdr->origin_y;
    ctdb->west_long = aux_hdr->west_long;
    ctdb->east_long = aux_hdr->east_long;
    ctdb->south_lat = aux_hdr->south_lat;
    ctdb->north_lat = aux_hdr->north_lat;
    return;
}


static CTDB_STATUS
read_patch_group_headers(CTDB *ctdb, FILE *fileptr)
{
  int32 groups;

  /* Read in all the patch group headers */
  groups=(ctdb->patches_wide*ctdb->patches_high)/CTDB_PATCHES_PER_GROUP;
  ctdb->patch_groups=(CTDB_PATCH_GROUP *)
    ctdb_alloc(ctdb->patch_groups,
	       sizeof(CTDB_PATCH_GROUP)*groups,
	       "patch group data");
  fread(ctdb->patch_groups,sizeof(CTDB_PATCH_GROUP),groups,fileptr);

  return CTDB_STATUS_OK;
}
 
static void 
init_null_topology(CTDB *ctdb)
{
    ctdb->node_storage = (int8 *) 
      ctdb_alloc(ctdb->node_storage, sizeof(CTDB_NODE), "initialization");
    bzero((char *)ctdb->node_storage, sizeof(CTDB_NODE));
    ctdb->edge_storage = (int8 *) 
      ctdb_alloc(ctdb->edge_storage, sizeof(CTDB_EDGE), "initialization");
    bzero((char *)ctdb->edge_storage, sizeof(CTDB_EDGE));
    ctdb->nodes = (CTDB_NODE **) 
      ctdb_alloc(ctdb->nodes, sizeof(CTDB_NODE *), "initialization");
    ctdb->nodes[0] = (CTDB_NODE *)ctdb->node_storage;
    ctdb->edges = (CTDB_EDGE **) 
      ctdb_alloc(ctdb->edges, sizeof(CTDB_EDGE *), "initialization");
    ctdb->edges[0] = (CTDB_EDGE *)ctdb->edge_storage;
}

static void
init_null_quadtree(CTDB *ctdb)
{
    ctdb->num_quad_data = 1;
    ctdb->quad_data = (CTDB_FEATURE_DATA *)
      ctdb_alloc(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA), "initialization");
    ctdb->quad_data[0].hdr.dynamic_patch_id = 0;
    ctdb->quad_data[0].hdr.size = 1;
}

/* $$$ Someday, this should generate enough of a database to work for
 * unmapped cells in a playbox
 */
static void 
init_as_nowhere(CTDB *ctdb)
{
    /* Note that the structure has already been zero'd out.  We just
     * set things which should not be zero here.
     */
    ctdb->input_format        = CTDB_FILE_FORMAT;
    ctdb->cell_id             = GCS_ILLEGAL_CELL;
    ctdb->fixed_point_basis   = ctdb->inv_fixed_point_basis = 1.0;
    ctdb->incr                = ctdb->incrD = 125.0;
    ctdb->inv_incr            = 1.0 / ctdb->incrD;
    ctdb->patch_incr          = ctdb->patch_incrD = 1.0;
    ctdb->patch_inv_incr      = 1.0 / ctdb->patch_incrD;
    ctdb->origin_northing     = 3105000; /* SAKI area */
    ctdb->origin_easting      = 508000;
    ctdb->origin_zone_number  = 38;
    ctdb->origin_zone_letter  = 'R';
    ctdb->datum               = DATUM_WGS84;
    strcpy(ctdb->name,"Nowhere");
    strcpy(ctdb->date,"Never");
    ctdb->is_gridded          = 1;
    ctdb->elevations_resident = 1;
    ctdb->elevations          = 
      (int32 *)ctdb_alloc(ctdb->elevations, sizeof(int32),"initialization");
    ctdb->elevations[0]       = 0.0;

    init_null_topology(ctdb);
    init_null_quadtree(ctdb);

    /* Generate a minimal soil database.*/
    ctdb->pat_table_ptr =
      (CTDB_PAT_TABLE *)ctdb_alloc(NULL,
				   sizeof(CTDB_PAT_TABLE),
				   "Polygon Attribute Table");
    ctdb->pat_table_ptr->num_columns = 1;
    ctdb->pat_table_ptr->num_entries =
      ctdb->pat_table_ptr->alloced_entries = 1;
    ctdb->pat_table_ptr->columns =
      (CTDB_PAT_COLUMN_HDR *)ctdb_alloc(NULL,
					sizeof(CTDB_PAT_COLUMN_HDR),
					"PAT Column Header");
    ctdb->pat_table_ptr->columns[0].facc = CTDB_SIMNET_MOB_ATTR;
    ctdb->pat_table_ptr->columns[0].data =
      (CTDB_PAT_DATA *)ctdb_alloc(NULL,
				  sizeof(CTDB_PAT_DATA),
				  "Soil Data");
    ctdb->pat_table_ptr->columns[0].data[0].integer = 0;
    ctdb->soil_table_size = sizeof(int32);
    ctdb->soil_table_storage = (int8 *)ctdb_alloc(NULL, sizeof(int32),
						  "Soil Table Storage");
    ctdb->num_soil_tables = 1;
    ctdb->soil_tables = (int32 **)ctdb_alloc(NULL, sizeof(int32 *),
					     "Soil Table Pointer");
    ctdb->soil_tables[0] = (int32 *) &ctdb->soil_table_storage;
    ctdb->soil_tables[0][0] = 0;
}

CTDB_STATUS 
ctdb_read_topology(CTDB *ctdb, FILE *fileptr)
{
    int32 i, offset;
    
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces: ctdb_read_topology((CTDB *)ctdb,fileptr);*/
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    if (ctdb->num_nodes)
    {
        ctdb->nodes = NULL;
	ctdb->nodes =
	  (CTDB_NODE **)ctdb_alloc(ctdb->nodes,
				   sizeof(CTDB_NODE *) *
				   ctdb->num_nodes,
				   "node index table");
	ctdb->node_storage =
	  (int8 *)ctdb_alloc(ctdb->node_storage, ctdb->node_size, "nodes");

	if (fread(ctdb->node_storage, ctdb->node_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Nodes\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_NODES_FAILED;
	}
    }

    if (ctdb->num_edges)
    {
        ctdb->edges = NULL;
	ctdb->edges =
	  (CTDB_EDGE **)ctdb_alloc(ctdb->edges,
				   sizeof(CTDB_EDGE *) *
				   ctdb->num_edges,
				   "edge index table");
	ctdb->edge_storage =
	  (int8 *)ctdb_alloc(ctdb->edge_storage, ctdb->edge_size, "edges");

	if (fread(ctdb->edge_storage, ctdb->edge_size, 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Edges\n");
	    /* Survive this error */
	    ctdb->num_nodes = ctdb->num_edges = ctdb->num_quad_data = 0;
	    return CTDB_STATUS_READ_EDGES_FAILED;
	}
    }

    /* Fill in the node index table */
    bzero((char *)ctdb->nodes, sizeof(CTDB_NODE *) * ctdb->num_nodes);
    for(i=offset=0;(i<ctdb->num_nodes) && (offset<ctdb->node_size);i++)
    {
	ctdb->nodes[i] = (CTDB_NODE *)(ctdb->node_storage + offset);
	offset += sizeof(CTDB_NODE) + sizeof(struct ctdb_connection) *
	  (ctdb->nodes[i]->num_connections - 1);
    }

    /* Fill in the edge index table */
    bzero((char *)ctdb->edges, sizeof(CTDB_EDGE *) * ctdb->num_edges);
    for(i=offset=0;(i<ctdb->num_edges) && (offset<ctdb->edge_size);i++)
    {
	ctdb->edges[i] = (CTDB_EDGE *)(ctdb->edge_storage + offset);
	offset += sizeof(CTDB_EDGE) + sizeof(struct ctdb_connection) *
	  (ctdb->edges[i]->num_features - 1);
    }

    return CTDB_STATUS_OK;
}

CTDB_STATUS 
ctdb_read_quad_tree(CTDB *ctdb, FILE *fileptr)
{
    if (ctdb->num_quad_data == 0 )
    {
	init_null_quadtree(ctdb);
	return CTDB_STATUS_OK;
    }
        
    /* Allocate space for the nodes, edges and index tables into
     * each (these records are variable length).  Then read the data.
     */

    /* replaces ctdb_read_quad_tree((CTDB *)ctdb,fileptr); */
    ctdb->quad_data =
      (CTDB_FEATURE_DATA *)ctdb_alloc(ctdb->quad_data,
				      sizeof(CTDB_FEATURE_DATA) *
				      ctdb->num_quad_data,
				      "abstract data and quadtree");

    if (fread(ctdb->quad_data, sizeof(CTDB_FEATURE_DATA),
	      ctdb->num_quad_data, fileptr) != ctdb->num_quad_data)
    {
	printf("CTDB: Read failed.  Quad tree and abstract data\n");
	/* Survive this error */
	init_null_quadtree(ctdb);
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_linear_models(CTDB *ctdb, FILE *fileptr)
{
    if (ctdb->num_linear_models == 0 )
    {
	ctdb->linear_models = NULL;
	return CTDB_STATUS_OK;
    }
        
    /* Allocate space for the linear models.  Then read the data.
     */
    ctdb->linear_models = NULL;
    ctdb->linear_models =
      (CTDB_LINEAR_MODEL *)ctdb_alloc(ctdb->linear_models,
				      sizeof(CTDB_LINEAR_MODEL) *
				      ctdb->num_linear_models,
				      "linear models");

    if (fread(ctdb->linear_models, sizeof(CTDB_LINEAR_MODEL),
	      ctdb->num_linear_models, fileptr) != ctdb->num_linear_models)
    {
	printf("CTDB: Read failed.  Linear models\n");
	/* Survive this error */
	ctdb->num_linear_models = 0;
    }      
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_volume_models(CTDB *ctdb, FILE *fileptr)
{
    int32               i, model_size;
    CTDB_VOLUME_MODEL  *model;
    
    ctdb->volume_models = NULL; 
    ctdb->mes_volume_models = NULL;

    if ( (ctdb->num_volume_models == 0) && (ctdb->num_mes_volumes == 0) )
	return CTDB_STATUS_OK;	

    /* Allocate space for the array of pointers to the volume models and
     * to the mes volume models.  Then read in each model.
     */
    ctdb->volume_models = 
      (CTDB_VOLUME_MODEL **)ctdb_alloc(ctdb->volume_models,
				       sizeof(CTDB_VOLUME_MODEL *) *
				       (ctdb->num_volume_models +
					ctdb->num_mes_volumes),
				       "volume models");
    
    /* Allocate space for a temporary model (large).
     */
    model = (CTDB_VOLUME_MODEL *)STDALLOC(sizeof(CTDB_VOLUME_MODEL) +
					  1024 * 3 * sizeof(float32));

    /* Read in volume models */
    for(i = 0; i < (ctdb->num_volume_models + ctdb->num_mes_volumes); i++)
    {
	/* Read in all of the model except the vertices.
	 */
	if(fread(model, sizeof(CTDB_VOLUME_MODEL) - sizeof(float32), 1,
		 fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Volume models\n");
	    /* Survive this error */
	    ctdb->num_volume_models = 0;
	    ctdb->num_mes_volumes = 0;
	    return CTDB_STATUS_OK;
	}

	/* Now that we know how many vertices are in this model, read them in.
	 */
	if(fread(model->vertices, model->num_vertices * 3 * sizeof(float32),
		 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Volume models\n");
	    /* Survive this error */
/*
	    ctdb->num_volume_models = 0;
	    ctdb->num_mes_volumes = 0;
	    return CTDB_STATUS_OK;
*/
		model->num_vertices = 3;
		bzero(model->vertices, 9 * sizeof(float32));
	}   

	model_size = sizeof(CTDB_VOLUME_MODEL) - sizeof(float32) +
	             model->num_vertices * 3 * sizeof(float32);

	/* Now that the entire model has been read in, store it.
	 */
	ctdb->volume_models[i] = NULL;
	ctdb->volume_models[i] =
	  (CTDB_VOLUME_MODEL *)ctdb_alloc(ctdb->volume_models[i], model_size,
					  "volume model");

	bcopy((char *)model, (char *)ctdb->volume_models[i], model_size);
    }
    if (ctdb->num_volume_models > 0)
      ctdb->mes_volume_models = 
	&(ctdb->volume_models[ctdb->num_volume_models]);
    else
    {
	ctdb->mes_volume_models = ctdb->volume_models;
	ctdb->volume_models = NULL;
    }
    STDDEALLOC(model);

    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_aggregate_models(CTDB *ctdb, FILE *fileptr)
{
    int32                   i, model_size;
    CTDB_AGGREGATE_MODEL   *model;

    if (ctdb->num_aggregate_models == 0)
    {
	ctdb->aggregate_models = NULL;
	return CTDB_STATUS_OK;
    }

    /* Allocate space for the array of pointers to the aggregate models.  Then
     * read in each model.
     */
    ctdb->aggregate_models =
      (CTDB_AGGREGATE_MODEL **)ctdb_alloc(NULL,
					  sizeof(CTDB_AGGREGATE_MODEL *) *
					  ctdb->num_aggregate_models,
					  "aggregate models");

    /* Allocate space for a temporary model (large).
     */
    model = (CTDB_AGGREGATE_MODEL *)STDALLOC(sizeof(CTDB_AGGREGATE_MODEL) +
					     512 *
					     sizeof(CTDB_AGG_MODEL_FEATURE));

    for(i = 0; i < ctdb->num_aggregate_models; i++)
    {
	/* Read in all of the model except the features.
	 */
	if(fread(model, sizeof(CTDB_AGGREGATE_MODEL) -
		 sizeof(CTDB_AGG_MODEL_FEATURE), 1,
		 fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Aggregate models\n");
	    /* Survive this error */
	    ctdb->num_aggregate_models = 0;
	    return CTDB_STATUS_OK;
	}

	/* Now that we know how many features are in this model, read them in.
	 */
	if(fread(model->features, model->n_features *
		 sizeof(CTDB_AGG_MODEL_FEATURE), 1, fileptr) != 1)
	{
	    printf("CTDB: Read failed.  Aggregate models\n");
	    /* Survive this error */
	    ctdb->num_aggregate_models = 0;
	    return CTDB_STATUS_OK;
	}

	model_size = sizeof(CTDB_AGGREGATE_MODEL) +
	  (model->n_features - 1) * sizeof(CTDB_AGG_MODEL_FEATURE);

	/* Now that the entire model has been read in, store it.
	 */
	ctdb->aggregate_models[i] =
	  (CTDB_AGGREGATE_MODEL *)ctdb_alloc(NULL, model_size,
					     "aggregate model");

	bcopy(model, ctdb->aggregate_models[i], model_size);
    }

    STDDEALLOC(model);

    return CTDB_STATUS_OK;
}

/* read the MES volume templates from the database */
static CTDB_STATUS 
ctdb_read_mes_templates(CTDB *ctdb, FILE *fileptr)
{
    int                 i, j;
    CTDB_MES_TEMPLATE  *cur_temp;
    CTDB_MES_ENCLOSURE *cur_encl;
    CTDB_MES_APERTURE  *cur_ap;
    CTDB_MES_GRID_BOX  *cur_box;
    int32               n_ap;

    if(ctdb->num_mes_templates == 0)
    {
	ctdb->mes_templates = NULL;
	return CTDB_STATUS_OK;
    }

    /* Assume that if there were MES's here before, they've been deallocated
     * by now.  We can't just realloc here because there are pointers hanging
     * off these pointers.
     */
    ctdb->mes_templates = (CTDB_MES_TEMPLATE *)
      ctdb_alloc(NULL, ctdb->num_mes_templates * sizeof(CTDB_MES_TEMPLATE),
		 "MES TEMPLATES");

    /* Just to be safe */
    bzero(ctdb->mes_templates, 
	  ctdb->num_mes_templates * sizeof(CTDB_MES_TEMPLATE));    

    /* Now get each MES template in order */
    for(i = 0; i < ctdb->num_mes_templates; i++)
    {
	cur_temp = ctdb->mes_templates + i;

	fread(cur_temp, sizeof(CTDB_MES_TEMPLATE), 1, fileptr);

	if(cur_temp->num_vertices == 0)
	{
	    cur_temp->all_vertices = NULL;
	}
	else
	{
	    cur_temp->all_vertices = (CTDB_MES_VERTEX *)
	      ctdb_alloc(NULL,
			 sizeof(int32)*XYZ*
			 (cur_temp->num_vertices),
			 "MES template vertices");

	    fread(cur_temp->all_vertices, sizeof(int32)*
		  XYZ*(cur_temp->num_vertices), 1, fileptr);
	}

	if(cur_temp->num_triangles == 0)
	{
	    cur_temp->all_triangles = NULL;
	}
	else
	{
	    cur_temp->all_triangles = (CTDB_MES_TRIANGLE *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_TRIANGLE) *
			 cur_temp->num_triangles,
			 "MES template triangles");

	    fread(cur_temp->all_triangles,
		  sizeof(CTDB_MES_TRIANGLE) *
		  cur_temp->num_triangles,
		  1, fileptr);
	}

	/* Need to get the enclosures for this template */
	if(cur_temp->num_enclosures == 0)
	{
	    cur_temp->all_enclosures = NULL;
	}
	else
	{
	    cur_temp->all_enclosures = (CTDB_MES_ENCLOSURE *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_ENCLOSURE) *
			 cur_temp->num_enclosures,
			 "MES enclosures");

	    for(j = 0; j < cur_temp->num_enclosures; j++)
	    {
		cur_encl = cur_temp->all_enclosures + j;

		fread(cur_encl, 
		      sizeof(CTDB_MES_ENCLOSURE),
		      1, fileptr);

		/* Need to get vertices and aperture id arrays for this 
		 * enclosure 
		 */
		if(cur_encl->footprint.num_vertices == 0)
		{
		    cur_encl->footprint.vertex_ids = NULL;
		}
		else
		{
		    cur_encl->footprint.vertex_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32) * 
				 cur_encl->footprint.num_vertices,
				 "MES enclosure vertices");

		    fread(cur_encl->footprint.vertex_ids,
			  sizeof(int32)*cur_encl->footprint.num_vertices,
			  1 , fileptr);
		}

		if(cur_encl->num_apertures == 0)
		{
		    cur_encl->aperture_ids = NULL;
		}
		else
		{
		    n_ap = cur_encl->num_apertures;
		    /* Round to nearest even number so we have an integral
		     * number of words
		     */
		    n_ap = n_ap + (n_ap % 2);


		    cur_encl->aperture_ids = (uint16 *)
		      ctdb_alloc(NULL,
				 sizeof(uint16) * n_ap,
				 "MES enclosure aperture ids");

		    fread(cur_encl->aperture_ids,
			  sizeof(uint16) * n_ap,
			  1, fileptr);
		}
	    }
	}

	/* Need to get the apertures for this template */
	if(cur_temp->num_apertures == 0)
	{
	    cur_temp->all_apertures = NULL;
	}
	else
	{
	    cur_temp->all_apertures = (CTDB_MES_APERTURE *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_APERTURE)*cur_temp->num_apertures,
			 "MES apertures");

	    for(j = 0; j < cur_temp->num_apertures; j++)
	    {
		cur_ap = cur_temp->all_apertures + j;

		fread(cur_ap, sizeof(CTDB_MES_APERTURE),
		      1, fileptr);

		/* Need to get the vertices for this aperture */
		if(cur_ap->outline.num_vertices == 0)
		{
		    cur_ap->outline.vertex_ids = NULL;
		}
		else
		{
		    cur_ap->outline.vertex_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32) * 
				 cur_ap->outline.num_vertices,
				 "MES aperture vertices");

		    fread(cur_ap->outline.vertex_ids,
			  sizeof(int32)*cur_ap->outline.num_vertices,
			  1, fileptr);
		}
	    }
	}

	/* Need to get the boxes for this MES */
	if(cur_temp->num_boxes ==0)
	{
	    cur_temp->all_boxes = NULL;
	}
	else
	{
	    cur_temp->all_boxes = (CTDB_MES_GRID_BOX *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_GRID_BOX)*cur_temp->num_boxes,
			 "MES boxes");

	    for(j = 0; j < cur_temp->num_boxes; j++)
	    {
		cur_box = cur_temp->all_boxes + j;

		fread(cur_box, sizeof(CTDB_MES_GRID_BOX), 1, fileptr);

		/* Lets see if there are any boxes inferior to this one */
		if( (cur_box->dimensions_x*
		     cur_box->dimensions_y*
		     cur_box->dimensions_z) > 1 )
		{
		    cur_box->box_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32)*(cur_box->dimensions_x*
						cur_box->dimensions_y*
						cur_box->dimensions_z),
				 "MES child box ids");

		    fread(cur_box->box_ids,
			  sizeof(int32)*(cur_box->dimensions_x*
					 cur_box->dimensions_y*
					 cur_box->dimensions_z),
			  1, fileptr);
		}

		if(cur_box->num_apertures == 0)
		{
		    cur_box->aperture_ids = NULL;
		}
		else
		{
		    n_ap = cur_box->num_apertures;
		    
		    /* Round to nearest even number so we have an integral
		     * number of words
		     */
		    n_ap = n_ap + (n_ap % 2);

		    cur_box->aperture_ids = (uint16 *)
		      ctdb_alloc(NULL,
				 sizeof(uint16) * n_ap,
				 "MES box aperture ids");

		    fread(cur_box->aperture_ids,
			  sizeof(uint16) * n_ap,
			  1, fileptr);
		}

		if(cur_box->num_triangles == 0)
		{
		    cur_box->wall_triangle_ids = NULL;
		}
		else
		{
		    cur_box->wall_triangle_ids = (int32 *)
		      ctdb_alloc(NULL,
				 sizeof(int32)*cur_box->num_triangles,
				 "MES box triangle ids");

		    fread(cur_box->wall_triangle_ids,
			  sizeof(int32)*cur_box->num_triangles,
			  1, fileptr);
		}
	    }
	}
    }
    return CTDB_STATUS_OK;
}

/* Read in the MES volumes for this database */
static CTDB_STATUS
ctdb_read_mes_volumes(CTDB *ctdb, FILE *fileptr)
{
    int32             i;
    CTDB_MES_VOLUME  *cur_vol;

    if (!ctdb->num_mes_volumes)
      return CTDB_STATUS_OK;
    
    ctdb->mes_volumes = (CTDB_MES_VOLUME *)
      ctdb_alloc(NULL, ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME),
		 "MES Volumes");

    /* Just to be safe */
    bzero(ctdb->mes_volumes, 
	  ctdb->num_mes_volumes * sizeof(CTDB_MES_VOLUME));    

    /* Now get each MES volume in order */
    for(i = 0; i < ctdb->num_mes_volumes; i++)
    {
	cur_vol = ctdb->mes_volumes + i;
	fread(cur_vol, sizeof(CTDB_MES_VOLUME), 1, fileptr);

	if(cur_vol->roofline.num_vertices == 0)
	{
	    cur_vol->roofline.vertices = NULL;
	}
	else
	{
	    cur_vol->roofline.vertices = (CTDB_MES_VERTEX_F64 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_VERTEX_F64) *
			 cur_vol->roofline.num_vertices,
			 "MES volume roofline vertices");

	    fread(cur_vol->roofline.vertices,
		  sizeof(CTDB_MES_VERTEX_F64) *
		  (cur_vol->roofline.num_vertices),
		  1,
		  fileptr);
	}
	if(cur_vol->floorline.num_vertices == 0)
	{
	    cur_vol->floorline.vertices = NULL;
	}
	else
	{
	    cur_vol->floorline.vertices = (CTDB_MES_VERTEX_F64 *)
	      ctdb_alloc(NULL,
			 sizeof(CTDB_MES_VERTEX_F64) *
			 cur_vol->floorline.num_vertices,
			 "MES volume floorline vertices");

	    fread(cur_vol->floorline.vertices,
		  sizeof(CTDB_MES_VERTEX_F64) *
		  (cur_vol->floorline.num_vertices),
		  1,
		  fileptr);
	}

    /* Now fill in the template pointers.  Until we do this, they
     * contain indices rather than pointers
     */
      ctdb->mes_volumes[i].template = ctdb->mes_templates + 
	(int32)ctdb->mes_volumes[i].template;
    }
    
    return CTDB_STATUS_OK;
}


#define CTDB_FREE(x)    \
    if (x)              \
      STDDEALLOC(x)

/* Free up all storage associated with the MES templates and volumes attached
 * to CTDB.  Since we tend to do this in the middle of reading a new database,
 * use nvol and ntempl as the number of existing volumes and templates,
 * resppectively, rather than the values stored in ctdb.
 */
static void
ctdb_free_mes(CTDB *ctdb,
	      int32 nvol,
	      int32 ntempl)
{
    int32              i,j;
    CTDB_MES_VOLUME   *volume;
    CTDB_MES_TEMPLATE *template;   

    for (i=0; i<nvol; ++i)
    {
	volume = &(ctdb->mes_volumes[i]);
	CTDB_FREE(volume->roofline.vertices);
	CTDB_FREE(volume->floorline.vertices);
    }

    CTDB_FREE(ctdb->mes_volumes);
    ctdb->mes_volumes = NULL;

    for (i=0; i<ntempl; ++i)
    {
	template = &(ctdb->mes_templates[i]);

	for (j=0; j < template->num_boxes; ++j)
	{
	    CTDB_FREE(template->all_boxes[j].box_ids);
	    CTDB_FREE(template->all_boxes[j].aperture_ids);
	    CTDB_FREE(template->all_boxes[j].wall_triangle_ids);
	}

	CTDB_FREE(template->all_boxes);

	for (j=0; j < template->num_apertures; ++j)
	{
	    CTDB_FREE(template->all_apertures[j].outline.vertex_ids);
	}

	CTDB_FREE(template->all_apertures);

	for (j=0; j < template->num_enclosures; ++j)
	{
	    CTDB_FREE(template->all_enclosures[j].footprint.vertex_ids);
	    CTDB_FREE(template->all_enclosures[j].aperture_ids);
	}

	CTDB_FREE(template->all_enclosures);
	CTDB_FREE(template->all_triangles);
	CTDB_FREE(template->all_vertices);
    }

    CTDB_FREE(ctdb->mes_templates);
    ctdb->mes_templates = NULL;
    
    return;
}
/* Read the MES data for the database */
static CTDB_STATUS
ctdb_read_mes(CTDB *ctdb, 
	      FILE *fileptr, 
	      int32 orig_num_vol,
	      int32 orig_num_templ)
{
    CTDB_STATUS stat;


    if (ctdb->mes_volumes || ctdb->mes_templates)
    {
	/* Free up old MES before reading in new ones.  MES are so complicated
	 * that it's too much trouble trying to realloc the space used
	 * by the old ones
	 */
	ctdb_free_mes(ctdb, orig_num_vol, orig_num_templ);
    }

    /* The MES data is broken into two parts--generic templates, and
     * volumes (instances) which reference them.
     */

    if ((stat = ctdb_read_mes_templates(ctdb, fileptr)) != CTDB_STATUS_OK)
      return stat;
    
    return ctdb_read_mes_volumes(ctdb, fileptr);
}


static CTDB_STATUS 
ctdb_read_soil_tables(CTDB *ctdb, FILE *fileptr)
{
    int32 i;

    ctdb->soil_tables = (int32 **)ctdb_alloc(ctdb->soil_tables,
					     ctdb->num_soil_tables *
					     sizeof(int32 *),
					     "Soil table index");
    
    ctdb->soil_table_storage = (int8 *)ctdb_alloc(ctdb->soil_table_storage,
						  ctdb->soil_table_size,
						  "Soil tables");

    /* Get the indices for the soil tables */
    fread(ctdb->soil_tables, sizeof(int32 *), ctdb->num_soil_tables, fileptr);

    /* These are kept in memory as pointers, but stored on disk as offsets.
     * Convert them to pointers.
     */
    for(i=0;i<ctdb->num_soil_tables;i++)
    {
	ctdb->soil_tables[i] = (int32 *)((int8 *)ctdb->soil_table_storage +
					 (int32)(ctdb->soil_tables[i]));
    }
    
    /* Read the soil tables themselves */
    fread(ctdb->soil_table_storage, ctdb->soil_table_size, 1, fileptr);
    return CTDB_STATUS_OK;
}

#define CTDB_PAT_COLUMN_PADDING 100

static CTDB_STATUS
ctdb_read_pat_table(CTDB *ctdb, FILE *fileptr)
{
    uint32 i;  
    
    /* Allocate space and read in PAT table */
    ctdb->pat_table_ptr = (CTDB_PAT_TABLE *)ctdb_alloc(ctdb->pat_table_ptr,
                           sizeof(CTDB_PAT_TABLE),
                           "Polygon Attribute Table");
    fread(ctdb->pat_table_ptr, sizeof(CTDB_PAT_TABLE), 1, fileptr);

    ctdb->pat_table_ptr->alloced_entries =
      ctdb->pat_table_ptr->num_entries + CTDB_PAT_COLUMN_PADDING;    

    /* Allocate space and read in PAT column headers */
    ctdb->pat_table_ptr->columns = (CTDB_PAT_COLUMN_HDR *)
      ctdb_alloc(NULL,
		 ctdb->pat_table_ptr->num_columns *
		 sizeof(CTDB_PAT_COLUMN_HDR),
		 "PAT column headers");
    fread(ctdb->pat_table_ptr->columns, sizeof(CTDB_PAT_COLUMN_HDR),
                           ctdb->pat_table_ptr->num_columns, fileptr);

    /* Allocate space and read in PAT column data */
    for (i=0; i<ctdb->pat_table_ptr->num_columns; i++)
    {	
	ctdb->pat_table_ptr->columns[i].data = (CTDB_PAT_DATA *)
	  ctdb_alloc(NULL,
		     (ctdb->pat_table_ptr->alloced_entries * 
		      sizeof(CTDB_PAT_DATA)),
		     "PAT data");
	fread(ctdb->pat_table_ptr->columns[i].data, sizeof(CTDB_PAT_DATA),
		  ctdb->pat_table_ptr->num_entries, fileptr);
    }   
    return CTDB_STATUS_OK; 
}

static CTDB_STATUS 
ctdb_read_water_chars(CTDB *ctdb, FILE *fileptr)
{
    ctdb->water_chars = (struct ctdb_water_chars *)
      ctdb_alloc(ctdb->water_chars, ctdb->num_water_chars * 
		 sizeof(struct ctdb_water_chars), "Water characteristics");
    
    fread(ctdb->water_chars, sizeof(struct ctdb_water_chars),
	  ctdb->num_water_chars, fileptr);    
    return CTDB_STATUS_OK;
}

static CTDB_STATUS 
ctdb_read_tidal_zones(CTDB *ctdb, FILE *fileptr)
{
    ctdb->tidal_zones = (struct ctdb_tidal_zone *)
      ctdb_alloc(ctdb->tidal_zones, ctdb->num_tidal_zones *
		 sizeof(struct ctdb_tidal_zone), "Tidal zones");
    
    fread(ctdb->tidal_zones, sizeof(struct ctdb_tidal_zone), 
	  ctdb->num_tidal_zones, fileptr);
    return CTDB_STATUS_OK;
}


static CTDB_STATUS
ctdb_read_all_patch_data(CTDB *ctdb, FILE *f)
{
    int32 groups;
    int32 i;
    CTDB_PATCH_GROUP *pg;

    ctdb->patch_data =
      (ADDRESS)ctdb_alloc(ctdb->patch_data,
			  ctdb->n_patch_bytes,
			  "DB Patch data");

    /* Read _all_ the patch data for the cell in in one big gulp */
    fread(ctdb->patch_data,1,ctdb->n_patch_bytes,f);

    pg = ctdb->patch_groups;

    /* Now that we have the data, we need to set up the pointers in all the
     * patch headers
     */
    groups = (ctdb->patches_wide*ctdb->patches_high) / 
      CTDB_PATCHES_PER_GROUP;

    for (i=0;i<groups;i++)
    {
	/* Doesn't matter which elt of the union we choose here... */
	pg[i].data_start.grid_data = 
	  (CTDB_GRID_PG_DATA*)((int8*)ctdb->patch_data +
			       pg[i].data_start.offset);
	pg[i].mapped = 1;
    }

    return CTDB_STATUS_OK;    
}

static CTDB_STATUS
ctdb_init_patch_data_cache(CTDB *ctdb, int32 memory_limit)
{
    int32 i;
    int32 cache_breadth;	/* Total memory consumption of one bucket
				 * of each size */
    int32 size;
    
    for (i=0, cache_breadth=0; i<CTDB_PATCH_CACHES;i++)
      /* Subtract 1 from size_map entry because bucket itself contains first
       * word of data
       */
      cache_breadth += sizeof(CTDB_CACHE_BUCKET) +
	((ctdb->patch_cache_size_map[i] - 1) * CTDB_CACHE_BKT_UNITS);

    /* Note that this will probably push us a little over the specified
     * limit, but not by much.
     */
    ctdb->patch_cache_entries = ceil((float64)memory_limit/
				     (float64)cache_breadth);
    
    for(i=0; i<CTDB_PATCH_CACHES; i++)
    {
	/* We subtract 1 from the map size because the cache bucket 
	 * structure has space allocated for the first word
	 */
	size = ctdb->patch_cache_entries * 
	  (sizeof(CTDB_CACHE_BUCKET) + 
	   ((ctdb->patch_cache_size_map[i]-1) * CTDB_CACHE_BKT_UNITS));

	ctdb->patch_data_cache[i] = (CTDB_CACHE_BUCKET *)
	  ctdb_alloc(ctdb->patch_data_cache[i],
		      size,
		      "Patch data cache");
	/* $$$This may be very slow for large caches.  The alternative is to
	 * go through and zero out the user fields, since those are all we
	 * really need to clear
	 */
	bzero((char *)ctdb->patch_data_cache[i], size);
    }
    return CTDB_STATUS_OK;
}

/* Actually reads both the header and the aux header */
static int32
ctdb_reread_hdr(char *fname, CTDB *ctdb)
{
    CTDB_FILE_HEADER hdr;
    CTDB_AUX_FILE_HEADER aux_hdr;
    FILE *f;
    
    if (!fname)
      return 0;
    
    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	return 0;
    }

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	return 0;
    }

    ctdb_unpack_header(ctdb, &hdr);

    if (fread(&aux_hdr, hdr.aux_hdr_bytes, 1, f) != 1)
    {	
	printf("CTDB: Error reading %s AUX header.\n",fname);
	fclose(f);
	return 0;
    }

    ctdb_unpack_aux_header(ctdb, &aux_hdr);
    return 1;
}

int32
ctdb_reread(char *fname, 
	    CTDB *ctdb, 
	    int32 memory_limit)
{
    FILE *f;
    CTDB_FILE_HEADER hdr;
    CTDB_AUX_FILE_HEADER aux_hdr;
    int32 overhead_mem;
    int32 feature_ml, elevation_ml;
    int32 feature_mem, elevation_mem, soil_mem;
    int32 total_mem;
    int32 looking_for_thinned = FALSE;
    char *period = NULL;
    int32 orig_num_mes_vol, orig_num_mes_templ;

    period = strrchr(fname, '.');
    if(period)
      looking_for_thinned = (*(period + 1) == 't');

    ctdb_total_malloced = 0;

    if (!fname)
    {
	printf("Initializing null terrain database.\n");
	init_as_nowhere(ctdb);
	return 1;
    }

    if (!(f = fopen(fname,"r")))
    {
	if(!looking_for_thinned)
	{
	    printf("CTDB: Error opening %s.\n",fname);
	    init_as_nowhere(ctdb);
	}

	return 1;
    }

    if(looking_for_thinned)
      printf("Found thinned version:  %s\n", fname);

    if (fread(&hdr, sizeof(hdr), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	init_as_nowhere(ctdb);
	return 1;
    }

    if (hdr.format != CTDB_FILE_FORMAT)
    {
	printf("CTDB: File %s is format %d.  Libctdb supports format %d.\n",
	       fname, hdr.format,
	       CTDB_FILE_FORMAT);
	fclose(f);
	init_as_nowhere(ctdb);
	return 1;
    }

    strcpy(ctdb->fname, fname);

    ctdb->memory_limit = memory_limit;

    ctdb->file = f;

    /* Before we unpack the new header, store off the number of mes templates
     * and volumes in the old database so we can free any pre-existing 
     * allocated memory
     */
    orig_num_mes_vol   = ctdb->num_mes_volumes;
    orig_num_mes_templ = ctdb->num_mes_templates;

    ctdb_unpack_header(ctdb, &hdr);

    /* Read in the auxiliary header info */
    if (fread(&aux_hdr, hdr.aux_hdr_bytes, 1, f) != 1)
    {
	printf("CTDB: Error reading %s AUX header.\n", fname);
	fclose(f);
	init_as_nowhere(ctdb);
	return 1;
    }
    ctdb_unpack_aux_header(ctdb, &aux_hdr);

    /* Figure out how much memory will go to overhead of non-cached
     * information.
     */
    overhead_mem = ((hdr.node_size + hdr.num_nodes * sizeof(CTDB_NODE *)) +
		    (hdr.edge_size + hdr.num_edges * sizeof(CTDB_EDGE *)) +
		    (hdr.num_quad_data * sizeof(CTDB_FEATURE_INFO)) +
		    (hdr.num_linear_models * sizeof(CTDB_LINEAR_MODEL)) +
		    /* This is an approximation (it excludes the vertices) */
		    ((hdr.num_volume_models + hdr.num_mes_volumes) * 
		     sizeof(CTDB_VOLUME_MODEL)) +
		    /* Another approximation (assumes 30 features per model) */
		    (hdr.num_aggregate_models * sizeof(CTDB_AGGREGATE_MODEL) +
		     29 * sizeof(CTDB_AGG_MODEL_FEATURE)) +
		    /* Soil tables */
		    (hdr.soil_table_size) +
		    /* Water Characteristics Table */
		    (hdr.num_water_chars * sizeof(CTDB_WATER_CHARS)) +
		    /* Tidal Characteristics Table */
		    (hdr.num_tidal_zones * sizeof(CTDB_TIDAL_ZONE)));

    if (memory_limit)
    {
	if (memory_limit <= overhead_mem)
	  memory_limit = 1; /* Use as little as possible */
	else
	  memory_limit -= overhead_mem;
    }    

    /* Distribute the available memory fairly between features and
     * elevations.
     */
    feature_mem   = ctdb->n_patch_bytes;
    elevation_mem = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
    soil_mem      = ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE;
    total_mem     = feature_mem + elevation_mem + soil_mem;
    /* Avoid caching if we were given more memory than we can use */
    if (memory_limit > total_mem)
      memory_limit = 0;

    feature_ml = (int32)(memory_limit * 
			 ((float64)feature_mem/
			 (float64)(feature_mem + elevation_mem)));
    elevation_ml = (int32)(memory_limit * 
			   ((float64)elevation_mem/
			    (float64)(feature_mem + elevation_mem)));

    /* Catch the case where the user has given us an almost-zero
     * memory_limit to indicate he wants as little memory as possible
     * to be used.
     */
    if (memory_limit && !feature_ml)
      feature_ml = 1;
    if (memory_limit && !elevation_ml)
      elevation_ml = 1;

    read_patch_group_headers(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_topology(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_quad_tree(ctdb, f); 

    /* Not cached, at least for now */
    ctdb_read_linear_models(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_volume_models(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_aggregate_models(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_mes(ctdb, f, orig_num_mes_vol, orig_num_mes_templ);

    /* Not cached, at least for now */
    ctdb_read_soil_tables(ctdb, f);

    
    ctdb_read_pat_table(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_water_chars(ctdb, f);

    /* Not cached, at least for now */
    ctdb_read_tidal_zones(ctdb, f);

    if ( ctdb->is_gridded )
      ctdb_read_elevations(ctdb, f, elevation_ml);
    
    ctdb->patch_data_seek = ftell(f);

    /* If we're not caching, read in all the patch data, otherwise set
     * up the cache
     */
    if (!memory_limit)
      ctdb_read_all_patch_data(ctdb, f);
    else
      ctdb_init_patch_data_cache(ctdb, memory_limit);      

    ctdb_find_z_limits(ctdb);

    ctdb->total_memory = ctdb_total_malloced;
    
    return 0;
}

/* Actually reads both the header and the aux header */
int32
ctdb_read_hdr(char *fname, CTDB *ctdb)
{
    if (!ctdb)
      return 0;
    bzero((char *)ctdb, sizeof(CTDB));
    return ctdb_reread_hdr(fname, ctdb);
}

/* Set the ctdb tiling mode.  TRUE => tiling is enabled, FALSE => tiling is
 * disabled.  If this routine is not called, tiling mode defaults 
 * to off.
 */
void
ctdb_set_tiling_mode(int32 mode)
{
    ctdb_is_tiled = mode;
}


int32
ctdb_read(char *fname, 
	  CTDB *ctdb, 
	  int32 memory_limit)
{
    bzero((char *)ctdb, sizeof(CTDB));
    ctdb_init();
    printf("Attempting to read %s\n",fname);

#ifdef CTDB_INTERVIS_METRICS
    printf("TEST MODE ... ");
    printf("METRICS COLLECTION\t");
    printf("\n");
#endif

    ctdb_dpb_init(1024);
    return ctdb_reread(fname, ctdb, memory_limit);
}

char *
ctdb_generate_tdb_filename(CTDB_TDB_TYPE kind,
			   char *tdb_path,
			   char *dbname,
			   int32 format,
			   int32 cell,
			   int32 thinned)
{
    static char filename[256];
    char *period;
    int32 failed=FALSE;
    
    if ( kind == CTDB_MODE_SIMNET )
    {
	if ( format >= 4 )
	  sprintf(filename,"%s/%s.%s%d%s",tdb_path,dbname,
		  CTDB_SIMNET_ID,format,CTDB_ENDIAN_ID);
	else if ( format == 3 )
	  sprintf(filename,"%s/%s.%s",tdb_path,dbname,
		  CTDB_FMT3_EXTENSION);
	else if ( format == 2 )
	  sprintf(filename,"%s/%s.%s",tdb_path,dbname,
		  CTDB_FMT2_EXTENSION);
	else if ( format == 1 )
	  sprintf(filename,"%s/%s.%s",tdb_path,dbname,
		  CTDB_FMT2_EXTENSION);
	else
	  failed = TRUE;
    }
    else if ( format >= 5 )
    {
	if (  kind == CTDB_MODE_MULTI_HDR )
	{
	    sprintf(filename,"%s/%s.%s/%s.%s%d%s",tdb_path,dbname,
		    CTDB_MULTI_CELL_DIR,dbname,CTDB_MULTI_CELL_ID,
		    CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
	}
	else if ( kind == CTDB_MODE_SINGLE_CELL )
	{
	    sprintf(filename,"%s/%s.%s%d%s",tdb_path,dbname,
		    CTDB_SINGLE_CELL_ID,CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
	}
	else if ( kind == CTDB_MODE_MULTI_CELL )
	{
	    if ( cell != GCS_ILLEGAL_CELL )
	      sprintf(filename,"%s/%s.%s/cell%05d.%s%d%s",tdb_path,dbname,
		      CTDB_MULTI_CELL_DIR,cell,CTDB_SINGLE_CELL_ID,
		      CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
	    else
	      failed = TRUE;
	}
    }
    else
      failed = TRUE;
    
    /* If we are generating the filename for a thinned database, change the
     * ID to a "t".
     */
    if(thinned)
      if(!(period = strrchr(filename, '.')))
	failed = TRUE;
      else
	*(period + 1) = CTDB_THINNED_ID;

    if (failed)
    {
	sprintf(filename,"%s/Nowhere.%s%d%s",tdb_path,
		CTDB_SIMNET_ID,CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
	printf("Failed to generate a proper filename: using %s\n",filename);
    }
    
    return filename;
}
