#ifndef lint
static char rcsid [] = "$RCSfile: ct_lookup.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_lookup.c                                                       *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                          *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts*
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to look up elevations                                *
 *       Created: Fri May 24 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                            *
 ****************************************************************************/
#include <stdext.h>
#include <stdcoord.h>
#include <libgcs.h>
#include <ctype.h>
#include <string.h>
#include "libctdb_local.h"

static int32 
ctdb_lookup_safe_feature_mask(
			      CTDB *ctdb, 
			      int32 norm_x, 
			      int32 norm_y
			      );

static CTDB_STATUS
ctdb_lookup_terrain_element_elevation(
				      CTDB           *ctdb,
				      CTDB_COORD2D   *pt,
				      float64        *elevation,
				      int32          *soil
				      );
static int32 
ctdb_point_on_ground_in_ctdb(
			     CTDB   *ctdb,
			     float64 x, 
			     float64 y,
			     float64 radius,
			     int32  *soil_ret
	 		     );


int32 
ctdb_lookup_elev_post(CTDB *ctdb, 
		      int32 norm_x, 
		      int32 norm_y)
{
    int32 page, index;
    int32 *page_memory;
    
    if ( ctdb->is_gridded )
    {
	if(norm_x < 0 || norm_y < 0)
	  return 0;
	
	page = (norm_x>>SIDE_SHIFT) + ctdb->pages_wide * (norm_y>>SIDE_SHIFT);
	index = (norm_x&SIDE_MASK) + POSTS_PER_SIDE * (norm_y&SIDE_MASK);
	
	/* It would work just fine if we ran the code after this every
	 * time; this test is purely a speed enhancer (8us/intervis call)
	 */
	if (ctdb->elevations_resident)
	  return ctdb->elevations[page*POSTS_PER_PAGE + index];
	
	if (page_memory = ctdb->elevation_pages[page])
	  ctdb->elevation_hit_count++;
	else
	{
	    ctdb->elevation_miss_count++;
	    page_memory = ctdb_fetch_elev_page(ctdb, page);
	}
	return page_memory[index];
    }
    return 0 ;
}

int32 
ctdb_lookup_soil_post(CTDB *ctdb, 
		      int32 norm_x, 
		      int32 norm_y)
{
    int32 page, index;
    int32 *page_memory;
    int32 patch_x, patch_y;
    int16 soil_post, *soils;
    
    if ( ctdb->is_gridded )
      if( !ctdb->is_short )
	{
	  if(norm_x < 0 || norm_y < 0)
	    return 0;
	
	  page = (norm_x>>SIDE_SHIFT) + ctdb->pages_wide * (norm_y>>SIDE_SHIFT);
	  index = (norm_x&SIDE_MASK) + POSTS_PER_SIDE * (norm_y&SIDE_MASK);
	
	  /* It would work just fine if we ran the code after this every
	   * time; this test is purely a speed enhancer (8us/intervis call)
	   */
	  if (ctdb->elevations_resident)
	    return ctdb->soils[page*POSTS_PER_PAGE + index];
	
	  if (page_memory = ctdb->soil_pages[page])
	    ctdb->soil_hit_count++;
	  else
	    {
	      ctdb->soil_miss_count++;
	      page_memory = ctdb_fetch_soil_page(ctdb, page);
	    }
	  return page_memory[index];
	}
      else
	{
	  if(norm_x < 0 || norm_y < 0)
	    return 0;

	  if(norm_y != 0) norm_y--;
	  patch_x = norm_x * ctdb->patch_inv_incr;
	  patch_y = norm_y * ctdb->patch_inv_incr;
	  soils = (int16 *) ctdb->soils;
	  soil_post = soils[patch_x + patch_y * ctdb->patches_wide];
	  return BUILD_POST_SW_SOIL(soil_post) | BUILD_POST_NE_SOIL(soil_post);
	}
    return 0 ;
}

/* Only call this if you KNOW norm_x and norm_y are within proper range.
 * Calling this with norm_x or norm_y off the database will most certainly
 * segv.  If you are at all unsure, call ctdb_lookup_post_elevation.
 */
float64 
ctdb_lookup_safe_post_elevation(CTDB *ctdb, int32 norm_x, int32 norm_y)
{
    int32 post;
    
    post = ctdb_lookup_elev_post(ctdb, norm_x, norm_y);
    
    return GET_POST_ELEVATION(post, ctdb->fixed_point_basis);
}

float64 
ctdb_lookup_post_elevation(CTDB *ctdb, int32 norm_x, int32 norm_y)
{
    if (!((norm_x >= ctdb->min_x_post) &&
	  (norm_y >= ctdb->min_y_post) &&
	  (norm_x <= ctdb->max_x_post) &&
	  (norm_y <= ctdb->max_y_post)))
      return 0.0;
    
    return ctdb_lookup_safe_post_elevation(ctdb, norm_x, norm_y);
}

static int32
ctdb_lookup_safe_feature_mask(CTDB *ctdb, int32 norm_x, int32 norm_y)
{
    int32 feature_mask;
    CTDB_PATCH_GROUP *pg;

    if ( !ctdb->is_gridded )
    {
	int32 patch_x, patch_y, patch_number, mem, ppx, ppy;
	/* variables used to access feature present bits stored in 
	   CTDB_TIN_PG_DATA for each patch. */
	int32 pps = ctdb->patch_incr;
	int32 pp_row = PP_ROW(pps);
	int32 pp_patch = PP_PATCH(pps, pp_row);
	
	CTDB_PATCH_GROUP *pg;
	
	/* norm_y of 0 defines an illegal patch number and 
	 * a region which is physically not defined by the 
	 * database.  return 0. 
	 */
	patch_x = PATCH_X(norm_x * ctdb->patch_inv_incr);
	patch_y = PATCH_Y(norm_y * ctdb->patch_inv_incr);
        patch_number = (patch_y*ctdb->patches_wide+patch_x);
	mem = PG_MEMBER(patch_number);
	pg=(ctdb->patch_groups+PG_NUMBER(patch_number));
	ppx=POST_WITHIN_PATCH_X(norm_x,ctdb->patch_incr);
	ppy=POST_WITHIN_PATCH_Y(norm_y,ctdb->patch_incr);
	
	if (!pg->mapped)
	  /* Cache miss--go get the data */
	  ctdb_fetch_patch_group(ctdb, pg);
	else
	  ctdb->patch_hit_count++;
	
	/*  Because the post flags aren't cached in the dpbuff, then we've
	 *  got to make sure it will included any "added" features
	 */
	feature_mask=CTDB_LOOKUP_POST_FLAGS(pg,ppx,ppy,mem, pp_row, pp_patch);
	if ((mem == 0 && pg->patch0_dyn_terr) ||
	    (mem == 1 && pg->patch1_dyn_terr) ||
	    (mem == 2 && pg->patch2_dyn_terr) ||
	    (mem == 3 && pg->patch3_dyn_terr))
	  feature_mask |= POST_FLAGS;
	return(feature_mask);
    } 
    else
      /* PURE GRIDDED or MIXED TERRAIN DATABASE */
    {
	/*l we've got to compute the patch_x and patch_y in order
	 *  to determine whether or not this patch is in the DPB
	 */
	int32 patch_x, patch_y, patch_number;
	feature_mask = ctdb_lookup_elev_post(ctdb,norm_x,norm_y)&POST_FLAGS;
	patch_x = PATCH_X(norm_x * ctdb->patch_inv_incr);
	patch_y = PATCH_Y(norm_y * ctdb->patch_inv_incr);
        patch_number = (patch_y*ctdb->patches_wide+patch_x);
	if (CTDB_DPB_HAS_PATCH(ctdb_dpb_has_patch(patch_number)))
	  feature_mask |= POST_FLAGS;
	return(feature_mask);
    }
}    

int32
ctdb_lookup_feature_mask(
			 CTDB *ctdb, 
			 int32 norm_x, 
			 int32 norm_y
			 )
{
    if (!((norm_x >= ctdb->min_x_post) &&
	  (norm_y >= ctdb->min_y_post) &&
	  (norm_x <= ctdb->max_x_post) &&
	  (norm_y <= ctdb->max_y_post)))
      return 0;
    
    return ctdb_lookup_safe_feature_mask(ctdb,norm_x,norm_y);    
}

/* ctdb_lookup_soil_feature checks if the passed point lies on top
 * of a linear feature (road or river).  If so, *soil is filled in
 * with the feature's soil type, and 1 is returned.  Otherwize the
 * routine returns 0.  px, py, norm_x, and norm_y are as in
 * ctdb_lookup_microterrain_elevation().  *width_ret, if non-NULL
 * returns the width of the linear feature which yielded the soil.
 * direction_ret, if non-NULL returns the direction of the linear
 * feature which yielded the soil.
 */
int32 
ctdb_lookup_soil_feature(CTDB    *ctdb, 
			 int32    px, 
			 int32    py, 
			 float64  norm_x, 
			 float64  norm_y, 
			 int32   *soil_ret, 
			 float64 *soil_width, 
			 float64 *direction)
{
    int32 patchx, patchy, patch_number;
    int32 post_mask;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *last;
    CTDB_XYO_COORD2D pt;
    int32 i,radius, *soilmap, next;
    uint32 row_in_pat;
    
    /* What patch is this post on? */
    patchx = PATCH_X(norm_x*ctdb->patch_inv_incr);
    patchy = PATCH_Y(norm_y*ctdb->patch_inv_incr);
    patch_number = patchx + patchy * ctdb->patches_wide;
    
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
      return 0;
   
    soilmap = ctdb_lookup_soilmap(ctdb,patch_number);
    
    /* The post_mask gives us a bounding box check in one easy operation */
    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);
    
    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can work in integer.
     */
    pt.x = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
    pt.y = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);
    
    last = feature + feature->hdr.size;
    for(feature++;feature<last;feature+=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;
	
	if (fclass == FC_MICRO)
	{
	    next = NEXT_MICRO(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME)
	{
	    next = NEXT_VOLUME(verts);
	    continue;
	}
	else if (fclass == FC_LINEAR)
	{
	    next = NEXT_LINEAR(verts);
	    continue;
	}
	else if (fclass == FC_CANOPY )
	{
	    next = NEXT_CANOPY(verts);
	    continue;
	}
	else if (fclass == FC_AGGREGATE )
	{
	    next = NEXT_AGGREGATE(verts);
	    continue;
	}
	  
	next = NEXT_LAID_LINEAR(verts);
	
	if (!(post_mask & feature->integer))
	  continue;
	
	radius =  CTDB_1OVER2x(feature[2].laid_linear.width);
	
	for (i=0;i<verts-1;i++)
	{
	    /* Are we on this one? */
	    if (ctdb_point_intersection(feature[i+3].xy, feature[i+4].xy,
					pt, radius) == INTERSECTION)
	    {
		if (soil_ret)
		{
		    row_in_pat = soilmap[feature[1].ll_soil.poly_char_index];
		    *soil_ret =
		      ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
		}
		if (soil_width)
		  *soil_width = CTDB_XYO_TO_METERS(CTDB_2x(radius), 0, 
						   ctdb->patch_incrD,
						   ctdb->incrD);
		
		if (direction)
		{
		    direction[X] =
		      (int)feature[i+4].xy.x - (int)feature[i+3].xy.x;
		    direction[Y] =
		      (int)feature[i+4].xy.y - (int)feature[i+3].xy.y;
		}
		return 1;
	    }
	}
    }
    return 0;
}


/* ctdb_lookup_feature_info takes a point and looks to see if there
 * is a linear feature under it. If so it returns 1 and fills in the 
 * the type, width, and direction variables passed to it. Otherwise it returns
 * 0 and type and width are undefined.
 */
int32 
ctdb_lookup_feature_info
(
			 CTDB    *ctdb, 
			 float64  x, 
			 float64  y, 
			 int32   *soil_ret, 
			 float64 *width_ret, 
			 float64  dir_ret[XY]
			 )
{
    float64 norm_x, norm_y;
    int32 px, py, post;
    
    if (!ctdb)
    {
	if(soil_ret)
	  *soil_ret  = 0;
	*dir_ret   = 0;
	*width_ret = 0;
	return 0;
    }
    
    /* If we're not cell aware, we won't worry about conversions */
    if (ctdb_is_tiled)
    {	
	if (!ctdb_point_on_database(ctdb, x, y))
	{
	    int32 home_cell;
	    CTDB* cur_ctdb;
	    
	    /* Determine which cell the point is in */
	    gcs_gcs2e_to_gcs((int32)ctdb_get_cell_id(ctdb), x, y, 
			       &home_cell, &x, &y);	
    
	    cur_ctdb = (CTDB *)gcs_get_tdb(home_cell, FALSE);

	    if (cur_ctdb != ctdb)
	      return ctdb_lookup_feature_info(cur_ctdb, x, y, soil_ret,
					      width_ret, dir_ret);
	    else 
	    {
		printf("Warning: infinite recursion in ctdb_lookup_feature_info.\n");
		return 0;
	    }
	}
    }
    
    
    /* Normalize x & y to incr units */
    norm_x = x * ctdb->inv_incr;
    norm_y = y * ctdb->inv_incr;
    
    px = POST_X(norm_x);
    py = POST_Y(norm_y);
    
    /* Bounds check */
    if (!(((px >= ctdb->min_x_post) &&
	   (px+1 <= ctdb->max_x_post)) &&
	  ((py-1 >= ctdb->min_y_post) &&
	   (py <= ctdb->max_y_post))))
      return 0;
    
    post = ctdb_lookup_feature_mask(ctdb, px, py);
    
    /* If there are "features" under this post (meaning roads or rivers),
     * see if we are on top of one.  If so, get its soil type.
     */
    
    return (post &&
	    ctdb_lookup_soil_feature(ctdb, px, py, norm_x, norm_y,
				     soil_ret, width_ret, dir_ret));
}

int32 
ctdb_point_on_database(CTDB *ctdb, float64 x, float64 y)
{
    return ( ctdb &&
	    (x >= ctdb->min_x) &&
	    (x <= ctdb->max_x) &&
	    (y >= ctdb->min_y) &&
	    (y <= ctdb->max_y));
}

#define CTDB_POINT_START_CELLS 50 /* Starting length to use for list of cells
				   * to look at when checking in point is
				   * on the ground
				   */
int32 
ctdb_point_on_ground(
		     CTDB    *ctdb, 
		     float64  x, 
		     float64  y, 
		     float64  radius, 
		     int32   *soil_ret
		     )
{
    int32 cur_cell;
    CTDB *cur_ctdb = ctdb;
    int32 ret;
    
    
    /*  In ctdb_point_on_ground_in_ctdb, points off the database
     *  were defined to be on the ground.  Similarly, a point
     *  on a null database is defined to be on the ground.
     */
    if (!ctdb)
      return 1;
    
    
    cur_cell = ctdb_get_cell_id(ctdb);
    
    /* If we're not cell aware, we won't worry about conversions */
    if (ctdb_is_tiled)
    {	
	if (!ctdb_point_on_database(ctdb, x, y))
	{
	    /* Determine which cell the point is in */
	    gcs_gcs2e_to_gcs(cur_cell, x, y, &cur_cell, &x, &y);	    
	    cur_ctdb = (CTDB *)gcs_get_tdb(cur_cell, FALSE);
	}
    }
    
    ret = ctdb_point_on_ground_in_ctdb(cur_ctdb, x, y, radius, soil_ret);
    
    /* If we are not cell-aware, we are done */
    if (!ctdb_is_tiled)
      return ret;
    
    /* If search locus is not on ground in this ctdb, we are done.
     * otherwise, we have to check other cells, if our search
     * locus extends to other cells
     */
    
    if (!ret)
      return 0;
    
    if (((x - radius) < cur_ctdb->min_x) ||
	((y - radius) < cur_ctdb->min_y) ||
	((x + radius) > cur_ctdb->max_x) ||
	((y + radius) > cur_ctdb->max_y))
    {
	/* The search locus extends beyond the cell.  Search
	 * neighboring cells.
	 */
	int32 *cell_list;
	int32 ncells = CTDB_POINT_START_CELLS;
	int32 i;
	float64 newx, newy;
	
	/* Get cells within radius */
	cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					"Checking point on ground");
	
	if (!gcs_cells_within_radius(cur_cell, x, y, 0.0, radius,
				     &ncells, cell_list))
	{
	    cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					    "Checking point on ground");
	    gcs_cells_within_radius(cur_cell, x, y, 0.0, radius,
				    &ncells, cell_list);
	}
	
	for(i=0;i<ncells;i++)
	{
	    if (cell_list[i] != cur_cell)
	    {
		/* If ctdb is not defined... define the point to be
		 * on the ground
		 */
		ret = 1;
		cur_ctdb = (CTDB *)gcs_get_tdb(cell_list[i], FALSE);
		if (cur_ctdb)
		{
		    gcs_gcs2e_to_gcscs(cur_cell, x, y, 
					 cell_list[i], &newx, &newy);
		    ret = ctdb_point_on_ground_in_ctdb(cur_ctdb, 
						       x, y, radius,
						       soil_ret);
		    if (!ret)
		      /* The search locus is not entirely on the
		       * ground in this ctdb.  
		       */
		      return 0;
		}
	    }
	}
    }
    return ret;
}


static int32 
ctdb_point_on_ground_in_ctdb(
			     CTDB   *ctdb, 
			     float64 x, 
			     float64 y, 
			     float64 radius, 
			     int32  *soil_ret
			     )
{
    float64 norm_x, norm_y, norm_r;
    int32 post;
    int32 pxl, pyl, pxh, pyh;
    /* Determine if the circle around this point has any
     * trees, water, or buildings in it.
     */ 
    
    norm_x = x * ctdb->inv_incr;
    norm_y = y * ctdb->inv_incr;
    norm_r = radius * ctdb->inv_incr;
    
    pxl = POST_X(norm_x-norm_r);
    pyl = POST_Y(norm_y-norm_r);
    
    pxh = POST_X(norm_x+norm_r);
    pyh = POST_Y(norm_y+norm_r);
    
    /* Bounds check
     * I declare points off the terrain database to be on the ground.
     */
    if (!((pxl >= ctdb->min_x_post) &&
	  (pyl >= ctdb->min_y_post) &&
	  (pxh <= ctdb->max_x_post) &&
	  (pyh <= ctdb->max_y_post)))
      return 1;
    
    /* Make a post which has the flags of all related posts combined.
     * Note that this code assumes a radius of no more than incr/2 meters.
     */
    post = ctdb_lookup_elev_post(ctdb, pxl, pyl);
    if (pxl != pxh)
      post |= ctdb_lookup_elev_post(ctdb, pxh, pyl);
    if (pyl != pyh)
    {
	post |= ctdb_lookup_elev_post(ctdb, pxl, pyh);
	if (pxl != pxh)
	  post |= ctdb_lookup_elev_post(ctdb, pxh, pyh);
    }
    
    /* Get the soil type, so the caller can check for water */
    if (soil_ret)
    {
	int32  row_in_pat;
	static CTDB_ELEV_DATA *elev;
	elev = ctdb_lookup_norm_elevation_max(ctdb, norm_x, norm_y, 
					      CTDB_MAX_LAND_AND_WATER);
	row_in_pat = (int32)elev->material;
	*soil_ret = ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
    }
    
    
    /* See if this super-post has any buildings or trees.
     * If so, see if any are on the point.
     */
    if (post & (POST_BUILDINGS | POST_TREES))
      return ctdb_check_clear(ctdb, pxl, pyl, pxh, pyh,
			      norm_x, norm_y, norm_r, post);
    else
      return 1;
}

int32 *
ctdb_lookup_soilmap(CTDB *ctdb, int32 patch_number)
{
    CTDB_PATCH_GROUP  *pg;
    int32              member;
    static int32       def_soils[256];

     /* Guard ourselves against bogus patch numbers.  */
    if ( patch_number < ctdb->min_patch_number )
      return def_soils;
    
    /* Hold harmless for nowhere */
    if (!ctdb->patch_groups)
      return def_soils;        
    
    if ( patch_number > ctdb->max_patch_number)
      return def_soils;
    
    pg     = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);
   
    if (pg->mapped)  
      ctdb->patch_hit_count++;
    else
      ctdb_fetch_patch_group(ctdb, pg);
    
    return CTDB_GET_SOILMAP(ctdb,pg,member);  
} 

#define SQUARE(x) (x*x)
static float64 alpha_sq = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS;
static float64 gamma_sq = WGS84_MINOR_AXIS*WGS84_MINOR_AXIS;

static float64 gamma_alpha_sq = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS *
                                WGS84_MINOR_AXIS*WGS84_MINOR_AXIS;

float64
ctdb_find_z_from_gcs_patch_water(CTDB   *ctdb, 
				 float64 x,
				 float64 y)
{
    static CTDB    *lastctdb=NULL;    
    static float64  origin[XYZ], x_off, y_off, rot[3][3];
    static float64 tmp[XYZ];
    static float64 R, T;
    float64 x0, y0, z;
    float64 A, B, C;
    float64 S, discriminant;

    if (lastctdb != ctdb)
    {
	/* $$$ future have GCS store these values for 
	 * easy retrieval 
	 */
	(void)gcs_get_origin_in_gcc(ctdb->cell_id,origin);
	(void)gcs_get_rot_mat(ctdb->cell_id, rot);
	gcs_get_origin_offsets(ctdb->cell_id, &x_off,
			       &y_off);


	/* Term for our quadratic equation (see below).  
	 * We save off the two terms that only depend on the cell,
	 * and not x or y.
	 */
	R = gamma_sq * (SQUARE(rot[0][2])+SQUARE(rot[1][2])) + 
	    alpha_sq * SQUARE(rot[2][2]);
	
	lastctdb = ctdb;

	tmp[X] = gamma_sq * rot[0][2];
	tmp[Y] = gamma_sq * rot[1][2];
	tmp[Z] = alpha_sq * rot[2][2];
    }

    /* shift the GCS (x,y) point into the proper frame of reference
     * were (0,0) is at the cell tangent point.
     */
    x0 = x + x_off;
    y0 = y + y_off;


    /* Calculate the same GCS point in GCC: pull out some constants:
     * 
     * U = A + rot[0][2]*z
     * V = B + rot[1][2]*z
     * W = C + rot[2][2]*z
     */
    A = rot[0][0]*x0 + rot[0][1]*y0 + origin[X];
    B = rot[1][0]*x0 + rot[1][1]*y0 + origin[Y];
    C = rot[2][0]*x0 + rot[2][1]*y0 + origin[Z];

    /* Solving the ellipsoid equation: (U^2 + V^2)/alpha_sq + (W^2)/gamma_sq
     * in terms of z:
     *
     * R*z^2 + S*z + T = 0
     */

    S = tmp[X]* A + tmp[Y]*B + tmp[Z]*C;
    S *= 2;
    

    T = gamma_sq * (SQUARE(A) + SQUARE(B)) + 
        alpha_sq * SQUARE(C) -
	gamma_alpha_sq;

    discriminant = SQUARE(S) - 4 * R * T;
    if (discriminant >= 0)
    {
	z = (-S + sqrt(discriminant)) / (R *2);
	return z;
    }
    else
    {
	printf("Negative Discriminant: no solutions\n");
	return 0;
    }
}

static float32 
ctdb_get_gcs_patch_water_level(CTDB             *ctdb,
			       CTDB_PATCH_GROUP *pg,
			       int32             mem,
			       float64           x,
			       float64           y)
{
    float64 patchz,tide_z;
    float64 z;

    if ( pg->contains_tins )
    {
	patchz = pg->data_start.tin_data->water_elev[mem];
	tide_z = ctdb->tidal_zones[ctdb->water_chars
				   [pg->data_start.
				    tin_data->tables[mem].water_char]
				   .tidal_zone_index].offset;
    }
    else
    {
	patchz = pg->data_start.grid_data->water_elev[mem];
	tide_z = ctdb->tidal_zones[ctdb->water_chars
				   [pg->data_start.
				    grid_data->tables[mem].water_char]
				   .tidal_zone_index].offset;
    }
    
    z = ctdb_find_z_from_gcs_patch_water(ctdb,x,y);

    return (float32)(z + patchz + tide_z);
}
    
    
#define CTDB_GET_TIN_WATER_PATCH_LEVEL(ctdb,pg,mem) \
   ((pg)->data_start.tin_data->water_elev[(mem)] +        \
   (ctdb)->tidal_zones[(ctdb)->water_chars[(pg)->data_start.tin_data->tables[(mem)].water_char].tidal_zone_index].offset)

#define CTDB_GET_GRID_WATER_PATCH_LEVEL(ctdb,pg,mem) \
   ((pg)->data_start.grid_data->water_elev[(mem)] +        \
   (ctdb)->tidal_zones[(ctdb)->water_chars[(pg)->data_start.grid_data->tables[(mem)].water_char].tidal_zone_index].offset)

/* CTDB_GET_WATER_PATCH_LEVEL will return the water level for a patch water
 * square shifted to account for the tide.
 */
#define CTDB_GET_WATER_PATCH_LEVEL(ctdb,pg,mem) \
    (float32)((pg)->contains_tins ? \
         (CTDB_GET_TIN_WATER_PATCH_LEVEL((ctdb),(pg),(mem))) : \
         (CTDB_GET_GRID_WATER_PATCH_LEVEL((ctdb),(pg),(mem))))

int32
ctdb_lookup_patch_water_level(CTDB    *ctdb, 
			      int32    patch_number,
			      float64  norm_x,
			      float64  norm_y,
			      float32 *level)
{
    CTDB_PATCH_GROUP  *pg;
    int32              member;

    /* Guard ourselves against bogus patch numbers or nowhere database */
    if (( patch_number < ctdb->min_patch_number ) ||
	( patch_number > ctdb->max_patch_number ) ||
	(!ctdb->patch_groups))
    {
	*level = 0;
	return FALSE;
    }
        
    pg     = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);

    if ( !CTDB_HAS_WATER_PATCH( pg, member ) )
    {
	*level = 0;
	return FALSE;
    }
   
    if (pg->mapped)
      ctdb->patch_hit_count++;
    else
      ctdb_fetch_patch_group(ctdb, pg);

    if ( ctdb->gcs_mode == CTDB_MODE_SIMNET )
    {
	*level = CTDB_GET_WATER_PATCH_LEVEL(ctdb,pg,member);
    }
    else
    {
	*level = ctdb_get_gcs_patch_water_level(ctdb,pg,member,
						norm_x*ctdb->incrD,
						norm_y*ctdb->incrD);
    }
    
    return TRUE;
}

/*- the following macro checks to see if patch_number is in the patch buffer,
 *  and if it is, returns it in 'patch'
 */
#define CTDB_CHECK_FOR_DYNAMIC_TERRAIN					     \
{									     \
    CTDB_FEATURE_DATA *patch;						     \
    if (CTDB_DPB_HAS_PATCH(ctdb_dpb_get_patch(patch_number,		     \
					      &patch)))			     \
      return(patch);		                          		     \
}

CTDB_FEATURE_DATA *
ctdb_lookup_feature(CTDB *ctdb, int32 patch_number)
{
    CTDB_PATCH_GROUP  *group;
    int32              member;
    uint32             offset;
    
    /* Guard ourselves against bogus patch numbers.  */
    if ( patch_number < ctdb->min_patch_number )
      return NULL;
    
    /* Hold harmless for nowhere */
    if (!ctdb->patch_groups)
      return NULL;
    
    if (CTDB_IS_QUAD_NODE(patch_number))
      return(ctdb_get_quad_node(ctdb, CTDB_QUAD_OFFSET(patch_number)));
    
    if ( patch_number > ctdb->max_patch_number)
      return NULL;
    
    group = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);
    
    if (!member)
    {	
	if (group->patch0_occupied)
	{
	    if (group->patch0_dyn_terr)
	      CTDB_CHECK_FOR_DYNAMIC_TERRAIN
	    /* Make sure the patch data is in memory.  If not, fetch it */
	    if (!group->mapped)
	      ctdb_fetch_patch_group(ctdb, group);
	    else
	      ctdb->patch_hit_count++;
	    
	    offset = group->contains_tins ?
	      (sizeof(CTDB_TIN_PG_DATA) / sizeof(CTDB_FEATURE_DATA)) :
		(sizeof(CTDB_GRID_PG_DATA) / sizeof(CTDB_FEATURE_DATA));
	}
	else
	  return NULL;
    }
    else if (member == 1)
    {
	if (group->patch1_occupied)
	{
	    if (group->patch1_dyn_terr)
	      CTDB_CHECK_FOR_DYNAMIC_TERRAIN
	    /* Make sure the patch data is in memory.  If not, fetch it */
	    if (!group->mapped)
	      ctdb_fetch_patch_group(ctdb, group);
	    else
	      ctdb->patch_hit_count++;
	    
	    offset = group->contains_tins ?
	      group->data_start.tin_data->patch1_offset :
		group->data_start.grid_data->patch1_offset;
	}	
	else
	  return NULL;
    }
    else if (member == 2)
    {
	if (group->patch2_occupied)
	{
	    if (group->patch2_dyn_terr)
	      CTDB_CHECK_FOR_DYNAMIC_TERRAIN
	    /* Make sure the patch data is in memory.  If not, fetch it */
	    if (!group->mapped)
	      ctdb_fetch_patch_group(ctdb, group);
	    else
	      ctdb->patch_hit_count++;
	    
	    offset = group->contains_tins ?
	      group->data_start.tin_data->patch2_offset :
		group->data_start.grid_data->patch2_offset;
	}
	else
	  return NULL;
    }
    else if (member == 3)
    {
	if (group->patch3_occupied)
	{
	    if (group->patch3_dyn_terr)
	      CTDB_CHECK_FOR_DYNAMIC_TERRAIN
	    /* Make sure the patch data is in memory.  If not, fetch it */
	    if (!group->mapped)
	      ctdb_fetch_patch_group(ctdb, group);
	    else
	      ctdb->patch_hit_count++;
	    
	    offset = group->contains_tins ?
	      group->data_start.tin_data->patch3_offset :
		group->data_start.grid_data->patch3_offset;
	}
	else
	  return NULL;
    }
    
    return ((CTDB_FEATURE_DATA *)group->data_start.grid_data + offset);
}


ADDRESS
ctdb_lookup_te_ptr(CTDB *ctdb, int32 patch_number)
{
    CTDB_PATCH_GROUP *pg;
    int32             member;
    
    /* Protect against BOGUS patch numbers */
    if ( !CTDB_INCLUSIVE_BETWEEN(patch_number,
				 ctdb->min_patch_number,
				 ctdb->max_patch_number) )
      return NULL;
    
    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);

    if ((!pg) || (!CTDB_PATCH_HAS_TES(pg, member)))
      return NULL;
    
    if (pg->mapped)
    {
	ctdb->patch_hit_count++;
	return ((ADDRESS)((CTDB_FEATURE_DATA *)pg->data_start.tin_data +
			  pg->data_start.tin_data->te_offset[member]));
    }
    else
    {
	ctdb_fetch_patch_group(ctdb, pg);
	return ((ADDRESS)((CTDB_FEATURE_DATA *)pg->data_start.tin_data +
			  pg->data_start.tin_data->te_offset[member]));
	
    }
}

#define CTDB_METERS_TO_PATCH_NUMBER(ctdb,x,y)\
((int32)(PATCH_X((x)*(ctdb)->meters_to_patch) + \
	 PATCH_Y((y)*(ctdb)->meters_to_patch) * (ctdb)->patches_wide ))

CTDB_WATER_CHARS *
ctdb_lookup_water_ptr(CTDB *ctdb, 
		      float64 x, 
		      float64 y)
{
    CTDB_PATCH_GROUP *pg;
    int32             member,patch_number;
    
    patch_number = CTDB_METERS_TO_PATCH_NUMBER(ctdb,x,y);
    /* Protect against BOGUS patch numbers */
    if ( !CTDB_INCLUSIVE_BETWEEN(patch_number,
				 ctdb->min_patch_number,
				 ctdb->max_patch_number) )
      return NULL;
    
    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    member = PG_MEMBER(patch_number);

    if (!pg)
      return NULL;

     /* $$$ check for water micro under point x,y.  If none exist then 
      * check for patch water.
      */
    if (!CTDB_HAS_WATER_PATCH(pg, member))
      return NULL;

    if (pg->mapped)
      ctdb->patch_hit_count++;
    else
      ctdb_fetch_patch_group(ctdb, pg);
    
    if ( pg->contains_tins )
      return (ADDRESS)&ctdb->water_chars[pg->data_start.
					 tin_data->tables[member].water_char];
    
    return (ADDRESS)&ctdb->water_chars[pg->data_start.
				       grid_data->tables[member].water_char];
}


int32
ctdb_lookup_tin_triangles(
			  CTDB *ctdb, 
			  int32 patch_number, 
			  CTDB_TIN_TRIANGLE *triangles, 
			  int32 *num_triangles, 
			  int32 cross_patch_bdy
			  )
{
    int32 mem;
    CTDB_PATCH_GROUP *pg;
    
    if (!ctdb || 
	(patch_number < ctdb->min_patch_number) ||
	(patch_number > ctdb->max_patch_number))
    {
	/* Illegal patch number--we'll just say there are no TEs and return */
	*num_triangles = 0;
	return TRUE;
    }
    mem = PG_MEMBER(patch_number);
    pg=(ctdb->patch_groups+PG_NUMBER(patch_number));
    
    if (((mem == 0)&&(pg->patch0_te16)) ||                 
	((mem == 1)&&(pg->patch1_te16)) ||
	((mem == 2)&&(pg->patch2_te16)) ||                 
	((mem == 3)&&(pg->patch3_te16)) )
      return ctdb_lookup_tin_triangles16(ctdb,
					 patch_number,
					 triangles,
					 num_triangles,
					 cross_patch_bdy,
					 pg,
					 mem);
    else
      return ctdb_lookup_tin_triangles08(ctdb,
					 patch_number,
					 triangles,
					 num_triangles,
					 cross_patch_bdy,
					 pg,
					 mem);
    
}

int32
ctdb_lookup_tin_polygons(
			 CTDB *ctdb, 
			 int32 patch_number, 
			 CTDB_TERRAIN_TRIANGLE *pts, 
			 int32 pt_dim, 
			 int32 *n_pts
			 )
{
    int32 mem = PG_MEMBER(patch_number);
    CTDB_PATCH_GROUP *pg=(ctdb->patch_groups+PG_NUMBER(patch_number));
    int32 i,j;
    int32 n_te;
    int32 size_ok;
    ADDRESS te;
    int32 (*te_pts_func)();
    
    *n_pts=0;
    size_ok = TRUE;
    
    if ( !( te=ctdb_lookup_te_ptr(ctdb,patch_number) ) )
      return size_ok;
    
    /* This *must* be done after tha call to lookup_te_ptr to guarantee that
     * the PG data is in memory
     */
    n_te = CTDB_NUM_TE(pg, mem);
    
    if (((mem == 0)&&(pg->patch0_te16)) ||                 
	((mem == 1)&&(pg->patch1_te16)) ||
	((mem == 2)&&(pg->patch2_te16)) ||                 
	((mem == 3)&&(pg->patch3_te16)) )
      te_pts_func = ctdb_lookup_tin_polygons16;
    else
      te_pts_func = ctdb_lookup_tin_polygons08;
    
    
    for (i=0;i<n_te;i++)
    {
	if ( (*n_pts+2) > pt_dim )
	{
	    size_ok = FALSE;
	    j = 0;
	}
	else
	  j = *n_pts;
	
	*n_pts += te_pts_func(ctdb,patch_number,(TERRAIN_ELEMENT_16 *)te,
			      i,&pts[j],&pts[j+1]);
	if ( pts[j].assoc_triangle != -1 )
	{
	    pts[j].assoc_triangle   = j+1;
	    pts[j+1].assoc_triangle = j  ;
	}    
    }
    /* make sure that you have one more that the actual
     * count to prevent infinite looping when the last
     * terrain element is only a triangle and the bounds
     * check assumes worst case of a polygon.
     */    
    if ( !size_ok )
      *n_pts += 2; 
    
    return size_ok;
}

/* Set up patch map for patch specified by patch_num (and for its neighbors
 * if cross_patch_bdy is TRUE.  Returns total number of triangles in 
 * patch_num.
 */
int32
ctdb_get_patch_maps(
		    CTDB *ctdb, 
		    int32 patch_num, 
		    int32 cross_patch_bdy, 
		    CTDB_PATCH_MAP *pm, 
		    CTDB_PATCH_MAP *npm, 
		    CTDB_PATCH_MAP *spm, 
		    CTDB_PATCH_MAP *epm, 
		    CTDB_PATCH_MAP *wpm
		    )
{
    int32 pnum;
    int32 mem;
    CTDB_PATCH_GROUP *pg;
    int32 n_tri;
    
    if ((patch_num < ctdb->min_patch_number) || 
	(patch_num > ctdb->max_patch_number))
      return 0;
    
    mem = PG_MEMBER(patch_num);
    pg = ctdb->patch_groups+PG_NUMBER(patch_num);
    
    if (((mem == 0)&&(pg->patch0_te16)) ||                 
	((mem == 1)&&(pg->patch1_te16)) ||
	((mem == 2)&&(pg->patch2_te16)) ||                 
	((mem == 3)&&(pg->patch3_te16)) )
      n_tri = ctdb_get_patch_maps_16(ctdb, patch_num, pg, mem, pm);
    else
      n_tri = ctdb_get_patch_maps_08(ctdb, patch_num, pg, mem, pm);
    
    if (!cross_patch_bdy)
    {
	return n_tri;
    }
    
    /* North */
    pnum = patch_num + ctdb->patches_wide;
    if (pnum <= ctdb->max_patch_number)
    {
	mem = PG_MEMBER(pnum);
	pg = ctdb->patch_groups+PG_NUMBER(pnum);
	
	if (((mem == 0)&&(pg->patch0_te16)) ||                 
	    ((mem == 1)&&(pg->patch1_te16)) ||
	    ((mem == 2)&&(pg->patch2_te16)) ||                 
	    ((mem == 3)&&(pg->patch3_te16)) )
	  ctdb_get_patch_maps_16(ctdb, pnum, pg, mem, npm);
	else
	  ctdb_get_patch_maps_08(ctdb, pnum, pg, mem, npm);
    }
    
    /* South */
    pnum = patch_num - ctdb->patches_wide;
    if (pnum >= ctdb->min_patch_number)
    {
	mem = PG_MEMBER(pnum);
	pg = ctdb->patch_groups+PG_NUMBER(pnum);
	
	if (((mem == 0)&&(pg->patch0_te16)) ||                 
	    ((mem == 1)&&(pg->patch1_te16)) ||
	    ((mem == 2)&&(pg->patch2_te16)) ||                 
	    ((mem == 3)&&(pg->patch3_te16)) )
	  ctdb_get_patch_maps_16(ctdb, pnum, pg, mem, spm);
	else
	  ctdb_get_patch_maps_08(ctdb, pnum, pg, mem, spm);
    }
    
    /* East */
    pnum = patch_num + 1;
    if (pnum <= ctdb->max_patch_number)
    {
	mem = PG_MEMBER(pnum);
	pg = ctdb->patch_groups+PG_NUMBER(pnum);
	
	if (((mem == 0)&&(pg->patch0_te16)) ||                 
	    ((mem == 1)&&(pg->patch1_te16)) ||
	    ((mem == 2)&&(pg->patch2_te16)) ||                 
	    ((mem == 3)&&(pg->patch3_te16)) )
	  ctdb_get_patch_maps_16(ctdb, pnum, pg, mem, epm);
	else
	  ctdb_get_patch_maps_08(ctdb, pnum, pg, mem, epm);
    }
    
    /* West */
    pnum = patch_num - 1;
    if (pnum >= ctdb->min_patch_number)
    {
	mem = PG_MEMBER(pnum);
	pg = ctdb->patch_groups+PG_NUMBER(pnum);
	
	if (((mem == 0)&&(pg->patch0_te16)) ||                 
	    ((mem == 1)&&(pg->patch1_te16)) ||
	    ((mem == 2)&&(pg->patch2_te16)) ||                 
	    ((mem == 3)&&(pg->patch3_te16)) )
	  ctdb_get_patch_maps_16(ctdb, pnum, pg, mem, wpm);
	else
	  ctdb_get_patch_maps_08(ctdb, pnum, pg, mem, wpm);
    }
    return n_tri;
}


static uint16
ctdb_lookup_attr_digit(char c)
{
    if( isdigit(c) ) 
      return (uint16) ( c - '0' );
    if( isupper(c) ) 
      return (uint16) ( c - 'A' + 10 );
    return CTDB_INVALID_FACC_ATTR;
}

uint16
ctdb_lookup_attr_code(char *attr_name)
{
    int32 i, j;
    uint16 code=0, k;
    
    i = strlen( attr_name );
    if( i<CTDB_LEN_FACC_NAME ) return CTDB_INVALID_FACC_ATTR;
    
    if( i>CTDB_LEN_FACC_NAME )
    {
	if( strcmp(attr_name, "SIMNET") == 0 )
	  return CTDB_SIMNET_MOB_ATTR;
	else
	  return CTDB_INVALID_FACC_ATTR;
    }
    
    for(j=0; j<CTDB_LEN_FACC_NAME; j++)
    {
	k = ctdb_lookup_attr_digit( attr_name[j] );
	if( k==CTDB_INVALID_FACC_ATTR ) return k;
	code = 36 * code + k;
    }
    return code;
}

void
ctdb_lookup_attr_name (uint16 code, 
		       char   attr_name[] )
{
    int32  i;
    uint16 j;
    
    if( code>CTDB_MAX_FACC_ATTR )
      if( code == CTDB_SIMNET_MOB_ATTR )
      {
	  strcpy( attr_name, "SIMNET" );
	  return;
      }
      else
      {
	  strcpy( attr_name, "INVALID" );
	  return;
      }
    else
    {
	attr_name[CTDB_LEN_FACC_NAME] = 0;
	for(i=CTDB_LEN_FACC_NAME-1; i>-1; i--)
	{
	    j = code%36;
	    if( j<10 ) 
	      attr_name[i] = j + '0';
	    else       
	      attr_name[i] = j + 'A' - 10;
	    code /= 36;
	}
    }
}
