#ifndef lint
static char rcsid [] = "$RCSfile: ct_mar.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_mar.c                                                           *
 *                                                                           *
 *       Copyright 1995 by TASC & SAIC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7600                                                        *
 *                                                                           *
 *	This software was developed by TASC and SAIC under U. S. Government *
 *      contracts 
 *	and may be reproduced by or for the U. S. Government pursuant to    *
 *	the copyright license under the clause at DFARS 252.227-7013        *
 *	(OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to find all MES structures along a ray.              *
 *       Created: Aug 23 1996
 *       Author: flchamberlain                                               *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdext.h>
#include <libgcs.h>
#include <libvecmat.h>
#include "libctdb_local.h"

typedef struct ctdb_mar_obj_data
{
  CTDB_LOS_IN_XYO_AND_DB los;
  CTDB_PATCH_DATA        patch;
  uint32                 postmask;
} CTDB_MAR_OBJ_DATA;

static int32
mar_mes_volume_compare(CTDB               *ctdb,
		       CTDB_FEAT_TYPE     *sub,
		       CTDB_FEATURE_DATA  *feature, 
		       CTDB_MAR_OBJ_DATA  *data);


static CTDB_MES_INTERSECT_INFO intersects[CTDB_MAX_MES_ON_RAY];
static int32 num_ints;

static float64 climb_threshold, fall_threshold;	/* Parameters used for finding
						 * intersection apertures
						 */


/* Do nothing for posts */
static int32
mar_post_compare(CTDB   *ctdb, 
		 float64 x, 
		 float64 y, 
		 float64 zl, 
		 float64 zh, 
		 float64 zcross,
		 int32   water)
{
    return VIS_KEEP_GOING;
}

/* start, end are the endpoints of the *entire* intervis ray, represented in
 * the frame of reference specified by ctdb.
 */
static int32 
mar_patch_compare(CTDB    *ctdb, 
		  int32    patch_number, 
		  int32    feature_mask,
		  int32    post_mask, 
		  float64 *start, 
		  float64 *end)
{
    CTDB_MAR_OBJ_DATA data;
    CTDB_FEATURE_DATA *feature, *last;
    int32 fclass, next;
    
    /* Note that since we selected the things we want to test for
     * in the feature_mask we passed to ctdb_intervis, we don't
     * need to explicitly check those flags here.
     */
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
      return VIS_KEEP_GOING;
    
    /* It's faster to compute these than pass them in.
     */
    data.postmask     = post_mask;
    data.patch.number = patch_number;
    data.patch.x = data.patch.number % ctdb->patches_wide;
    data.patch.y = data.patch.number / ctdb->patches_wide;
    
    data.los.xyo.start.x = POST_TO_XYO(start[X], data.patch.x,
				  ctdb->patch_inv_incr);
    data.los.xyo.start.y = POST_TO_XYO(start[Y], data.patch.y,
				  ctdb->patch_inv_incr);
    
    data.los.xyo.end.x = POST_TO_XYO(end[X], data.patch.x, 
				     ctdb->patch_inv_incr);
    data.los.xyo.end.y = POST_TO_XYO(end[Y], data.patch.y, 
				     ctdb->patch_inv_incr);
    data.los.db.start.x = start[X] * ctdb->incrD;
    data.los.db.start.y = start[Y] * ctdb->incrD;
    data.los.db.start.z = start[Z];
    data.los.db.end.x = end[X] * ctdb->incrD;
    data.los.db.end.y = end[Y] * ctdb->incrD;
    data.los.db.end.z = end[ZL];
    
    last = feature + feature->hdr.size;
    for(feature++;feature<last;feature+=next)
    {
	/* Only interested in buildings */
	fclass = feature->info.feature_class;

	if ( fclass == CTDB_FC_MICRO )
	  next = NEXT_MICRO(feature->info.vertices);
	else if ( fclass == CTDB_FC_VOLUME )
	{
	    next = NEXT_VOLUME(feature->info.vertices);
	    if ( feature[1].volume.type == CTDB_FS_MES )
	    {
		if ( mar_mes_volume_compare(ctdb, NULL, feature,
					    (ADDRESS)&data) == VIS_STOP)
		  return VIS_STOP;
	    }
	}
	else
	  return VIS_KEEP_GOING;
    }	
    return VIS_KEEP_GOING;
    
}


static int32
mar_cell_fcn(CTDB    *ctdb,
	     float64 *ray_start,
	     float64 *ray_end,
	     float64 *cur_start,
	     float64 *cur_end,
	     float64  width)
{
    float64 ray_start_post[3];
    float64 ray_end_post[4];
    float64 cur_start_post[4];
    float64 cur_end_post[4];

    /* Scale from meters to grid spacing */
    ray_start_post[X] = ray_start[X] * ctdb->inv_incr;
    ray_start_post[Y] = ray_start[Y] * ctdb->inv_incr;
    ray_start_post[ZL] = ray_start[ZL];
    
    ray_end_post[X] = ray_end[X] * ctdb->inv_incr;
    ray_end_post[Y] = ray_end[Y] * ctdb->inv_incr;
    ray_end_post[ZL] = ray_end[ZL];
    ray_end_post[ZH] = ray_end[ZH];
    
    cur_start_post[X] = cur_start[X] * ctdb->inv_incr;
    cur_start_post[Y] = cur_start[Y] * ctdb->inv_incr;
    cur_start_post[ZL] = cur_start[ZL];
    cur_start_post[ZH] = cur_start[ZH];
    
    cur_end_post[X] = cur_end[X] * ctdb->inv_incr;
    cur_end_post[Y] = cur_end[Y] * ctdb->inv_incr;
    cur_end_post[ZL] = cur_end[ZL];
    cur_end_post[ZH] = cur_end[ZH];


    (void) ctdb_intervis(ctdb, ray_start_post, ray_end_post,
			 cur_start_post, cur_end_post, FALSE,
			 mar_post_compare, mar_patch_compare, 
			 POST_BUILDINGS);


    /* Do any necessary post-processing */

    return VIS_KEEP_GOING;
}



static void 
cell_to_mes (CTDB             *ctdb,
	     CTDB_MES_VOLUME  *mes,
	     int32             patch_x,
	     int32             patch_y,
	     float64           cell_x,
	     float64           cell_y,
	     float64           cell_z,
	     float64          *mes_x,
	     float64          *mes_y,
	     float64          *mes_z )                

{
   float64 mes_position[XYZ];
   int32 mes_cell = mes->origin[CELL3D];

   float64 offset[XYZ];

   if ( ctdb_get_cell_id(ctdb) != mes_cell )
   {
       gcs_gcs3e_to_gcscs ( ctdb->cell_id, cell_x, cell_y, cell_z, mes_cell,
                            &offset[X], &offset[Y], &offset[Z] );
       offset[X] -= mes->origin[X];
       offset[Y] -= mes->origin[Y];
       offset[Z] -= mes->origin[Z];
   }
   else
   {
       offset[X] = cell_x - mes->origin[X];
       offset[Y] = cell_y - mes->origin[Y];
       offset[Z] = cell_z - mes->origin[Z];
   }

   ctdb_gcsvec_to_mesvec ( mes, mes_cell, offset, mes_position );
   *mes_x = mes_position[X];
   *mes_y = mes_position[Y];
   *mes_z = mes_position[Z];
}

/* Return the number of intersections between los and the roofline of mes.
 * If there is at least one intersections, the fractional distance along
 * the ray of the first intersection is returned in t0, and the fractional
 * distance along the ray of the last intersection is returned in t1.
 * Note that los must be in GCS coordinates in the same cell as ctdb.
 */
static int32
find_mes_ray_int(CTDB               *ctdb,
		 CTDB_MES_VOLUME    *mes,
		 CTDB_LINE_OF_SIGHT *los,
		 float64            *t0,
		 float64            *t1)
{
    int32 i;
    float64 t_low, t_high;
    float32 tt1, tt2;
    float32 start[XY], end[XY], temp1[XY], temp2[XY];
    float32 int_pt[XY];
    int32 intersections = 0;
    int32 n_vert = mes->roofline.num_vertices;
    
    start[X] = los->start.x;
    start[Y] = los->start.y;
    end[X]   = los->end.x;
    end[Y]   = los->end.y;
    
    t_low = 2.0;
    t_high = -1.0;
    
    for(i=0;i<(n_vert-1);i++)
    {
	temp1[X] = mes->roofline.vertices[i][X];
	temp1[Y] = mes->roofline.vertices[i][Y];
	temp2[X] = mes->roofline.vertices[i+1][X];
	temp2[Y] = mes->roofline.vertices[i+1][Y];
	if (ctdb_intersection_2_linesegs(start, end,
					 temp1, temp2,
					 int_pt, &tt1, &tt2))
	{
	    /* don't accept points that start or end ray to avoid
	     * the intesections that start at or end on MES walls.
	     */
	    if ((tt1 == 0.0) || (tt1 == 1.0))
	      continue;
	    
	    intersections++;
	    
	    if (tt1 < t_low)
	    {
		t_low = tt1;
	    }
	    if (tt1 > t_high)
	    {
		t_high = tt1;
	    }
	}
    }
    
    /* Check the last edge */
    temp1[X] = mes->roofline.vertices[n_vert-1][X];
    temp1[Y] = mes->roofline.vertices[n_vert-1][Y];
    temp2[X] = mes->roofline.vertices[0][X];
    temp2[Y] = mes->roofline.vertices[0][Y];
    if (ctdb_intersection_2_linesegs(start, end,
				     temp1, temp2,
				     int_pt, &tt1, &tt2))
    {
	intersections++;
	
	if (tt1 < t_low)
	{
	    t_low = tt1;
	}
	if (tt1 > t_high)
	{
	    t_high = tt1;
	}
    }	
    
    if (intersections)
    {
	*t0 = t_low;
	*t1 = t_high;
    }

    return intersections;
}


static int32
mar_mes_volume_compare(CTDB               *ctdb,
		       CTDB_FEAT_TYPE     *sub,
		       CTDB_FEATURE_DATA  *feature, 
		       CTDB_MAR_OBJ_DATA  *data)
{
    uint32              which_volume_model; 
    uint32              which_mes_volume;
    CTDB_VOLUME_MODEL  *volume_model;
    CTDB_MES_VOLUME    *mes;

    float64 t0, t1;

    float64 ray_delta[XYZ];
    float64 intercept_delta[XYZ];

    which_volume_model = feature[1].volume.reference;
    if ( which_volume_model >= ctdb->num_mes_volumes ) /* ERROR */
      return VIS_KEEP_GOING;

    volume_model = ctdb->mes_volume_models[which_volume_model];
    which_mes_volume = volume_model->mes_volume_index;

    if ( which_mes_volume >= ctdb->num_mes_volumes ) /* ERROR*/
      return VIS_KEEP_GOING;
   
    mes = &(ctdb->mes_volumes[which_mes_volume]);

    if (find_mes_ray_int(ctdb, mes, &data->los.db, &t0, &t1))
    {	
	/* We hit this one; fill out an entry for it. */
	intersects[num_ints].ctdb = ctdb;
	intersects[num_ints].type = mes->template->type;
	
	/* Find entry point */
	ray_delta[X] = data->los.db.end.x - data->los.db.start.x;
	ray_delta[Y] = data->los.db.end.y - data->los.db.start.y;
	ray_delta[Z] = data->los.db.end.z - data->los.db.start.z;
	
	vmat3_scal_vec_mul64(t0, ray_delta, intercept_delta);
	
	intersects[num_ints].entry_pt[X] = data->los.db.start.x + 
	  intercept_delta[X];
	intersects[num_ints].entry_pt[Y] = data->los.db.start.y + 
	  intercept_delta[Y];
	intersects[num_ints].entry_pt[Z] = data->los.db.start.z + 
	  intercept_delta[Z];

	intersects[num_ints].entry = 
	  ctdb_find_close_ext_ap(ctdb, which_volume_model, 
				 intersects[num_ints].entry_pt,
				 climb_threshold, fall_threshold);
	
	/* Find exit point */
	vmat3_scal_vec_mul64(t1, ray_delta, intercept_delta);
	    
	intersects[num_ints].exit_pt[X] = data->los.db.start.x + 
	  intercept_delta[X];
	intersects[num_ints].exit_pt[Y] = data->los.db.start.y + 
	  intercept_delta[Y];
	intersects[num_ints].exit_pt[Z] = data->los.db.start.z + 
	  intercept_delta[Z];

	intersects[num_ints].exit = 
	  ctdb_find_close_ext_ap(ctdb, which_volume_model, 
				 intersects[num_ints].exit_pt,
				 climb_threshold, fall_threshold);
	
	num_ints++;
    }

    return VIS_KEEP_GOING;    
}

void
ctdb_find_mes_along_ray(CTDB                     *ctdb, 
			float64                   x0, 
			float64                   y0, 
			float64                   z0,
			float64                   x1, 
			float64                   y1, 
			float64                   z1,
			float64                   climb,
			float64                   fall,
			int32                    *num_found,
			CTDB_MES_INTERSECT_INFO **mes_list)
{
    if (!ctdb)
    {
	*num_found = 0;
	return;
    }   

    /* Initialize globals */
    num_ints = 0;
    climb_threshold = climb;
    fall_threshold  = fall;
    
    /* Run the ray */
    ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, z1, z1, 0.0, mar_cell_fcn);

    *num_found = num_ints;
    *mes_list = intersects;
    return;
}
