#ifndef lint
static char rcsid [] = "$RCSfile: ct_mes.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_mes.c                                                        *
*                                                                           *
*       Copyright 1996 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*       This software was developed by TASC and SAIC  under U. S. Government*
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013 (OCT 1988).                                   *
*                                                                           *
*       Contents: functions for Multi-elevation surface                     *
*       Created: Mon Feb 12 15:31:26 EST 1996                               *
*       Author: hlu                                                         *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb_local.h"
#include <libgcs.h>
#include <stdio.h>

#define LARGE_NUMBER    5000.0
#define MAX_ENCLOSURES  50


#define CTDB_MES_UNIT_TO_GCS(mes, x) \
(x * mes->template->fixed_point_basis)

#define CTDB_GCS_UNIT_TO_MES(mes, x) \
(x / mes->template->fixed_point_basis)

#define CTDB_MES_VALID_ENCLOSURE_ID(ctdb, mes_id, enclosure_id)             \
(((enclosure_id) != CTDB_INVALID_MES_ENCLOSURE_ID )                      && \
 ((enclosure_id) < (ctdb)->mes_volumes[(mes_id)].template->num_enclosures))

#define CTDB_MES_VALID_APERTURE_ID(ctdb, mes_id, aperture_id)               \
(((aperture_id) != CTDB_INVALID_MES_APERTURE_ID )                        && \
((aperture_id) < (ctdb)->mes_volumes[(mes_id)].template->num_apertures))

#define CTDB_MES_VALID_EDGE_KEY(ctdb, edge_key)                             \
(CTDB_MES_VALID_MES_ID((ctdb), CTDB_MES_EDGE_KEY_TO_MES_ID(edge_key))    && \
 CTDB_MES_VALID_APERTURE_ID((ctdb),                                        \
			     CTDB_MES_EDGE_KEY_TO_MES_ID(edge_key),         \
			     CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edge_key)))

/* ctdb_mes_mesvec_to_gcsvec
 *
 * This function will take a vector defined by the mes's local coordinate
 * system and transform it to a gcs vector.
 */
static void
ctdb_mesvec_to_gcsvec(CTDB_MES_VOLUME  *mes,
		      float64           mesvec[XYZ],
		      int32            *cell,
		      float64           gcsvec[XYZ]
		      );

void
ctdb_mesvec_to_gcscsvec(
			CTDB_MES_VOLUME  *mes,
			float64           mesvec[XYZ],
			int32             cell,
			float64           gcsvec[XYZ]
			);



int32
ctdb_mes_get_triangle_normal(CTDB   *ctdb,
			     uint16  mes_id,
			     int32   triangle_id,
			     float64 normal[XYZ]
			     )
{
    CTDB_MES_VOLUME   *mes;
    CTDB_MES_TEMPLATE *template;
    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
      return 0;
    
    mes      = &ctdb->mes_volumes[mes_id];
    template = mes->template;

    ctdb_mesvec_to_gcscsvec(mes, template->all_triangles[triangle_id].normal,
			    ctdb_get_cell_id(ctdb), normal);
    vmat3_unit64(normal, normal);
    return 1;
}



static int32
ctdb_mes_ptop(CTDB   *ctdb,
	      uint16  mes_id,
	      float64 origin[XY],
	      float64 z,
	      float64 destination[XY],
	      float64 dest_z,
	      float64 location[XYZ])
{
    int32 n_mes_inter;
    CTDB_MES_INTERSECTION *mes_inter;
    CTDB_MES_VOLUME   *mes;
    float64 vec[XYZ];
    float64 mes_origin[XYZ];
    float64 mes_destination[XYZ];
    float64 extended_destination[XYZ];
    float64 delta[XYZ];
    
    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
      return 0;

    delta[X] = destination[X] - origin[X];
    delta[Y] = destination[Y] - origin[Y];
    delta[Z] = dest_z - z;

    vmat3_scal_vec_mul64(1.1, delta, delta);
    vmat3e_vec_add64(origin[X], origin[Y], z, delta[X], delta[Y], delta[Z],
		     extended_destination);

    mes = &ctdb->mes_volumes[mes_id];

    ctdb_gcs_to_mes(mes, ctdb_get_cell_id(ctdb),
		    extended_destination[X], extended_destination[Y],
		    extended_destination[Z],
		    &mes_destination[X],  &mes_destination[Y],
		    &mes_destination[Z]);
    ctdb_gcs_to_mes(mes, ctdb_get_cell_id(ctdb),
		    origin[X], origin[Y], z,
		    &mes_origin[X], &mes_origin[Y], &mes_origin[Z]);

    ctdb_mes_line_intersection(mes, mes_origin, mes_destination, FALSE,
			       CTDB_MES_ALL_ENCLOSURES, FALSE, TRUE,
			       FALSE, &n_mes_inter, &mes_inter);
    if (!n_mes_inter)
    {
	location[X] = destination[X];
	location[Y] = destination[Y];
	location[Z] = dest_z;
	return 1;
    }

    VMAT3_VEC_ZERO64(location);

    /* Compute point of intersection, less a smidgen */
    vmat3_vec_sub64(mes_destination, mes_origin, vec);

    if (mes_inter[0].distance > 0.1)
      vmat3_scal_vec_mul64(mes_inter[0].distance-0.1, vec, vec);
    else
      VMAT3_VEC_ZERO64(vec);

    vmat3_vec_add64(mes_origin, vec, vec);

    ctdb_mes_to_gcscs(mes, (int32)vec[X], (int32)vec[Y], (int32)vec[Z],
		      ctdb_get_cell_id(ctdb),
		      &location[X], &location[Y], &location[Z]);
    return 0;
}


/* ===================  MES Geometry Functions ================== */
int32 
ctdb_mes_find_mes_from_xy(CTDB              *ctdb, 
			  float64            x,
			  float64            y,
			  int32              size_mes_volumes,
			  CTDB_MES_NODE_KEY  mes_volumes[]
			  )
{
    int32 patch_x, patch_y;
    int32 patch_number;
    CTDB_FEATURE_DATA *feature, *last;
    int32  verts, fclass, next;
    uint32 post_mask;
    int32  post_x, post_y;
    int32 cnt = 0;

    /* $$$ This routine has no GCS processing */
    if (!ctdb)
      return 0;

    patch_number = ctdb_compute_patch_num(ctdb, x, y, &patch_x, &patch_y);

    post_x = POST_X((float64)(x*ctdb->inv_incr));
    post_y = POST_Y((float64)(y*ctdb->inv_incr));
    post_mask = POST_MASK(post_x, post_y, 
			  patch_x, patch_y, ctdb->patch_incr);

    feature = ctdb_lookup_feature(ctdb, patch_number);
    if (!feature)
      return 0;
    
    last = feature + feature->hdr.size;
    for (feature++; feature < last; feature += next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;
	
	if (fclass == FC_MICRO)
	{
	    next = NEXT_MICRO(verts);
	    continue;
	}

	if (fclass != FC_VOLUME)
	  break;
	
	next = NEXT_VOLUME(verts);
	
	if ((post_mask & feature->integer) && 
	    (feature[1].volume.type == CTDB_FS_MES))
	{
	    uint32              which_volume_model; 
	    uint32              which_mes_volume;
	    CTDB_VOLUME_MODEL  *volume_model;
	    
	    which_volume_model = feature[1].volume.reference;
	    if (which_volume_model >= ctdb->num_mes_volumes)
	    {
		printf("LibCTDB: invalid MES feature reference\n");
		continue;/* Error: skip this MES */
	    }

	    volume_model = ctdb->mes_volume_models[which_volume_model];
	    which_mes_volume = volume_model->mes_volume_index;
	    
	    if (which_mes_volume >= ctdb->num_mes_volumes)
	    {
		printf("LibCTDB: invalid MES volume reference\n");
		continue;/* Error: skip this MES */
	    }
	    
	    if (cnt >= size_mes_volumes)
	      break;
	    
	    CTDB_MES_IDS_TO_NODE_KEY(which_mes_volume,CTDB_MES_EXTERIOR,
				     mes_volumes[cnt]);
	    cnt++;
	}
    }
    return cnt;
}


static uint16 
ctdb_mes_enclosure_containing_triangle(CTDB   *ctdb, 
				       uint16  mes_id, 
				       int32   triangle_id,
				       float64 ray[XYZ]
				       )
{
    int32 i;
    CTDB_MES_ENCLOSURE *enclosure;
    CTDB_MES_APERTURE  *aperture;
    CTDB_MES_TEMPLATE  *templ;
    float64 *normal;

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
      return CTDB_INVALID_MES_ENCLOSURE_ID;

    templ  = ctdb->mes_volumes[mes_id].template;
    normal = (float64 *) templ->all_triangles[triangle_id].normal;

    for (i=0; i < templ->num_enclosures; i++)
    {
	enclosure = &templ->all_enclosures[i];
	if ((enclosure->first_wall_triangle_id <= triangle_id) && 
	    (triangle_id < enclosure->first_wall_triangle_id + 
	                   enclosure->num_triangles))
	{
	  /* We have hit a wall triangle in this enclosure */
	    if (vmat3_dot_prod64(normal, ray) < 0)
	      return i;
	    else
	      /* we are between walls */
	      return CTDB_INVALID_MES_ENCLOSURE_ID;
	}
    }
	
    for (i=0; i < templ->num_apertures; i++)
    {
	aperture = &templ->all_apertures[i];
	if ((aperture->first_triangle_id <= triangle_id) && 
	    (triangle_id < aperture->first_triangle_id + 
	     aperture->num_triangles))
	{
	    /* We have hit an aperture triangle */
	    if (vmat3_dot_prod64(normal, ray) < 0)
	      return aperture->connections[0].enclosure_id;
	    else
	      return aperture->connections[1].enclosure_id;
	}
    }
    return CTDB_INVALID_MES_ENCLOSURE_ID;
}

/* This function finds all enclosures that lie along the line x-y
 * in GCS space for the given MES.
 *
 * $$$ In the future, probably want to extend this to work for all MES's,
 * to allow for the possibility of MES's on top of each other.
 */

int32 
ctdb_mes_find_enclosures_from_xy(CTDB              *ctdb,
				 CTDB_MES_NODE_KEY  search_mes,
				 float64            x,
				 float64            y,
				 int32              array_size,
				 CTDB_MES_NODE_KEY  enclosures[]
				 )
{
    uint16             mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(search_mes);
    CTDB_MES_VOLUME   *mes;
    float64            ray_origin[XYZ], ray_destination[XYZ];
    int32              i, j, cnt = 0;
    CTDB_MES_INTERSECTION *mes_inter;
    int32                  n_mes_inter;

    if (!ctdb || 
	!CTDB_MES_VALID_MES_ID(ctdb,mes_id) ||
	!array_size)
      return cnt;

    mes   = &(ctdb->mes_volumes[mes_id]);

    /* Some very out-of-there kind of z's */
    ctdb_gcs_to_mes(mes, ctdb->cell_id, x, y, (float64)LARGE_NUMBER,
		    &ray_origin[X], &ray_origin[Y], &ray_origin[Z]);
    ctdb_gcs_to_mes(mes, ctdb->cell_id, x, y, (float64)(-LARGE_NUMBER),
		    &ray_destination[X],  &ray_destination[Y],
		    &ray_destination[Z]);

    ctdb_mes_line_intersection(mes, ray_origin, ray_destination, TRUE,
			       CTDB_MES_ALL_ENCLOSURES, FALSE, TRUE,
			       TRUE, &n_mes_inter,&mes_inter);

    if ( n_mes_inter )
    {
	float64 ray[XYZ];

	vmat3_vec_sub64(ray_destination, ray_origin, ray);
	
	for (i=0; ((i < n_mes_inter) && (cnt < array_size)); i++)
	{
	    int32 found;
	    enclosures[cnt] = 
	      ctdb_mes_enclosure_containing_triangle(ctdb, mes_id,
						     mes_inter[i].triangle, 
						     ray);
	    /* remove invalid enclosures */
	    if (enclosures[cnt] == CTDB_INVALID_MES_ENCLOSURE_ID)
	      continue;
	    
	    /* thin out redundant enclosure keys */
	    found = FALSE;
	    for (j=0;j<cnt;j++)
	      if (enclosures[cnt] == enclosures[j])
	      {
		  found = TRUE;
		  break;
	      }
	    
	    
	    if (!found)
	      cnt++;	    
	}

	/* convert enclosure keys to mes keys */
	for (j=0;j<cnt;j++)
	  enclosures[j] = CTDB_MES_KEY(mes_id, enclosures[j]);
    }
    else
      enclosures[cnt++] = CTDB_MES_KEY(mes_id, CTDB_MES_EXTERIOR);
 
    return cnt;
}

/* This function determines if a given x,y location is in a given 
 * enclosure, and returns TRUE if so.
 */
int32
ctdb_mes_xy_in_enclosure(
			 CTDB              *ctdb,
			 float64            x,
			 float64            y,
			 CTDB_MES_NODE_KEY  enclosure
			 )
{
    int32             i; 
    int32             num_enclosures;
    CTDB_MES_NODE_KEY enclosure_keys[MAX_ENCLOSURES];
    CTDB_MES_NODE_KEY mes_key;

    if (ctdb_mes_valid_node_key(ctdb, enclosure))
    {
        CTDB_NODE_KEY_TO_MES_KEY(enclosure, mes_key);
	num_enclosures = 
	  ctdb_mes_find_enclosures_from_xy(ctdb, mes_key,
					   x, y,
					   MAX_ENCLOSURES,
					   enclosure_keys);

	for (i = 0; i < num_enclosures; i++)
	  if (CTDB_MES_NODE_KEYS_EQUAL( enclosure, enclosure_keys[i]))
	    return TRUE;
    }
    else
    {
	/* If the MES key is invalid but the point is not contained within
	 * a MES then the MES key is correct.
	 */
	CTDB_ELEV_DATA elev;
	static CTDB_ELEV_QUALIFIER qual=(CTDB_INCL_VOLUME|CTDB_RETURN_MAX);
	
	ctdb_lookup_qual_elevation(ctdb, x, y, qual, CTDB_NO_REF_ELEV, &elev);
	if ((elev.sensClass == CTDB_FC_ILLEGAL) || (elev.subclass != CTDB_FS_MES))
	  return TRUE;
    }

    return FALSE;    
}

static int32
ctdb_mes_pt_in_mes(CTDB   *ctdb,
		   uint16 *mes_id,
		   float64 x,
		   float64 y,
		   float64 z,
		   CTDB  **out_ctdb
		   )
{
    float64 mes_x, mes_y, mes_z, max_z, min_z;
    float64 ray_origin[XYZ], ray_dest[XYZ];
    
    int32 patch_x, patch_y;
    int32 patch_number;
    CTDB_FEATURE_DATA *feature, *last;
    int32               verts, fclass, next;
    uint32 post_mask;
    int32 post_x, post_y;

    if (!ctdb)
      return 0;

    if (ctdb_is_tiled)
      if (!ctdb_point_on_database(ctdb, x, y))
      {
	  float64 new_x, new_y, new_z;
	  int32 cell;
	  CTDB *new_ctdb;

	  gcs_gcs3e_to_gcs(ctdb_get_cell_id(ctdb), x, y, z,
			   &cell, &new_x, &new_y, &new_z);
	  new_ctdb = (CTDB *) gcs_get_tdb(cell, FALSE);
	  if (new_ctdb == ctdb)
	  {
	      printf("Infinite recursion in ctdb_mes_pt_in_mes\n");
	      return 0;
	  }
	  return ctdb_mes_pt_in_mes(new_ctdb, mes_id, new_x, new_y, new_z,
				    out_ctdb);
      }

    patch_number = ctdb_compute_patch_num(ctdb, x, y, &patch_x, &patch_y);
    *mes_id = CTDB_INVALID_MES_ID;
    
    post_x = POST_X((float64)(x*ctdb->inv_incr));
    post_y = POST_Y((float64)(y*ctdb->inv_incr));
    post_mask = POST_MASK(post_x, post_y, 
			  patch_x, patch_y, ctdb->patch_incr);

    feature = ctdb_lookup_feature(ctdb, patch_number);
    if (!feature)
      return 0;
    
    last = feature + feature->hdr.size;
    for (feature++; feature < last; feature += next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;
	
	if (fclass == FC_MICRO)
	{
	    next = NEXT_MICRO(verts);
	    continue;
	}
	
	if (fclass != FC_VOLUME)
	  return 0;
	
	next = NEXT_VOLUME(verts);
	
	if ((post_mask & feature->integer) && 
	    (feature[1].volume.type == CTDB_FS_MES))
	{
	    CTDB_MES_VOLUME    *mes;
	    CTDB_VOLUME_MODEL  *volume_model;
	    CTDB_MES_TEMPLATE  *templ;
	    uint32 which_volume_model; 
	    uint32 which_mes_volume;
	    
	    which_volume_model = feature[1].volume.reference;
	    if (which_volume_model >= ctdb->num_mes_volumes) 
	    {
		printf("LibCTDB: invalid MES feature reference\n");
		continue;/* Error: skip this MES */
	    }

	    volume_model = ctdb->mes_volume_models[which_volume_model];
	    which_mes_volume = volume_model->mes_volume_index;

	    if (which_mes_volume >= ctdb->num_mes_volumes)
	    {
		printf("LibCTDB: invalid MES volume reference\n");
		continue;/* Error: skip this MES */
	    }

	    mes = &(ctdb->mes_volumes[which_mes_volume]);
	    
	    templ = mes->template;
	    
	    /* ctdb_gcs_to_mes returns 0 if x,y,z lies outside of the mes */
	    if (ctdb_gcs_to_mes(mes, ctdb_get_cell_id(ctdb), x, y, z,
				 &mes_x, &mes_y, &mes_z))
	    {
		int32 n_mes_inter;
		CTDB_MES_INTERSECTION *mes_inter;
		/* Run a ray up in GCS z */
		ctdb_gcs_to_mes(mes, ctdb_get_cell_id(ctdb), x, y, 
				z + LARGE_NUMBER,
				&ray_origin[X],&ray_origin[Y],&ray_origin[Z]);
		ctdb_gcs_to_mes(mes, ctdb_get_cell_id(ctdb), x, y, 
				z - LARGE_NUMBER,
				&ray_dest[X], &ray_dest[Y], &ray_dest[Z]);
		ctdb_mes_line_intersection(mes, ray_origin, ray_dest, FALSE, 
					   CTDB_MES_ALL_ENCLOSURES,
					   FALSE, TRUE, TRUE, 
					   &n_mes_inter, &mes_inter);
		
		if ( n_mes_inter )
		{
		    *mes_id = feature[1].volume.reference;
		    *out_ctdb = ctdb;
		    return 1;
		}
	    }
	}
    }
    return 0;
}

/* Calculates the distance square between two points (point1, point2) of <n> 
   dimension.  Stolen from ligeometry.
 */
static float64 
distance_sq_2_points(float64 point1[],
		     float64 point2[],
		     int32   n)
{
    int32 i;
    float64 temp;
    float64 dist2 = 0;
    
    for (i = 0; i < n; i++)
    {
	temp = (point1[i] - point2[i]);
	dist2 += temp * temp;
    }
    return(dist2);
}

/* Calculates the distance square between a point and a line.  Stolen from
 * libgeometry.
 */
static float64 
distance_sq_point_to_line(float64 line_pnt1[2],
			  float64 line_pnt2[2],
			  float64 point[2],
			  float64 line_inter[2])
{
    float64 to_pnt2[2];
    float64 to_point[2];
    float64 to_inter[2];
    
    VMAT2_VEC_SUB(line_pnt2, line_pnt1, to_pnt2);
    VMAT2_VEC_SUB(point, line_pnt1, to_point);
    vmat2_project64(to_point, to_pnt2, to_inter);
    vmat2_vec_add64(to_inter, line_pnt1, line_inter);
    return(distance_sq_2_points(point, line_inter, 2));
}

/* Calculates the distance square between a point and a line segment.  Stolen
 * from libgeometry.
 */
static float64 
distance_sq_point_to_segment(float64 seg_start[2],
			     float64 seg_end[2],
			     float64 point[2],
			     float64 seg_inter[2])
{
    float64 start_to_end[2];
    float64 start_to_point[2];
    float64 point_to_end[2];

    VMAT2_VEC_SUB(seg_end, seg_start, start_to_end);
    VMAT2_VEC_SUB(point, seg_start, start_to_point);
    VMAT2_VEC_SUB(seg_end, point, point_to_end);
    
    if (VMAT2_DOT_PROD(start_to_end, start_to_point) < 0.0)
    {
	VMAT2_VEC_COPY64(seg_start, seg_inter);
	return(distance_sq_2_points(seg_start, point, 2));
    }
    else if (VMAT2_DOT_PROD(start_to_end, point_to_end) < 0.0)
    {
	VMAT2_VEC_COPY64(seg_end, seg_inter);
	return(distance_sq_2_points(seg_end, point, 2));
    }
    else
      return(distance_sq_point_to_line(seg_start, seg_end, point, 
				       seg_inter));
}

#define CTDB_MAX_VERTS 20

/* Return the square of distance from the pt to the specified aperture */
static float64
ctdb_dist_sq_to_aperture(CTDB              *ctdb,
			 float64            loc[XY],
			 CTDB_MES_EDGE_KEY  aperture)
{
    int32 i, j, vertex_id, skip, num = 0;
    float64 int_pt[XY];
    float64 outline[CTDB_MAX_VERTS][XY], z;
    float64 dist_sq, best_dist_sq = 9999999;
    CTDB_MES_APERTURE *ap;
    CTDB_MES_TEMPLATE *template;
    
    uint16 mes_id, ap_id;
    mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID(aperture);
    ap_id  = CTDB_MES_EDGE_KEY_TO_APERTURE_ID(aperture);

    template = ctdb->mes_volumes[mes_id].template;
    ap = &template->all_apertures[ap_id];

    for (i=0; i < ap->outline.num_vertices; i++)
    {
	skip = 0;
	vertex_id = ap->outline.vertex_ids[i];
	ctdb_mes_to_gcscs(&(ctdb->mes_volumes[mes_id]),
			  template->all_vertices[vertex_id][X],
			  template->all_vertices[vertex_id][Y],
			  template->all_vertices[vertex_id][Z],
			  ctdb_get_cell_id(ctdb),
			  &outline[num][X], &outline[num][Y], &z);

	/* Check for duplicates first */
	for (j=0; j < num; j++)
	  if (VMAT2_VEC_EQUAL64(outline[j], outline[num]))
	  {
	      skip = 1;
	      break;
	  }
	if (skip)
	  continue;

	num++;
    }

    for (i=0; i < num; i++)
    {
	dist_sq = distance_sq_point_to_segment(outline[i], 
					       outline[(i+1)%num],
					       loc, int_pt);
	if (dist_sq < best_dist_sq)
	  best_dist_sq = dist_sq;
    }
    return best_dist_sq;
}


/* ctdb_mes_find_mes
 *
 * If there is a mes-volume within "threshold" meters of the given point,
 * this function returns TRUE, and "mes_id" points to the id of the nearest
 * mes-volume.  Otherwise, this function returns FALSE, and the value of 
 * *mes_id is undefined.
 */
int32
ctdb_mes_find_mes(CTDB*               ctdb,
		  float64             x, 
		  float64             y, 
		  float64             z,
		  float64             threshold,
		  CTDB*              *out_ctdb,
		  CTDB_MES_NODE_KEY  *mes_found
		  )
{
    CTDB_SEARCH_SPACE_PTR search_space_ptr;
    int32 patch_number;
    CTDB_TERRAIN_SUBCLASS type;
    int32 avoid, force_id, num_vertices;
    float64 verts[CTDB_VERTICES_MAX][2];
    float64 roof_elevation[CTDB_VERTICES_MAX];
    float64 dist_sq, best_dist_sq;
    int32 current_mes_id;
    uint16             found_mes_id;
    CTDB_MES_NODE_KEY  current_mes_key;
    int32              found;
    
    CTDB_MES_MAKE_NODE_KEY_INVALID(*mes_found);

    if (!ctdb)
      return FALSE;

    if (ctdb_mes_pt_in_mes(ctdb, &found_mes_id, x, y, z, out_ctdb))
    {
	/* Point in MES */
        CTDB_MES_ID_TO_MES_KEY(found_mes_id, *mes_found);
	return TRUE;
    }
    else if ( threshold == 0.0)
    {
	/* The point and MES found is the outside MES structure */
	*mes_found = CTDB_INVALID_MES_NODE_KEY;
	return FALSE;
    }

    found = FALSE;
    best_dist_sq = threshold*threshold;    
    search_space_ptr = ctdb_create_search_space(ctdb, 
						(int32)(x-threshold),
						(int32)(y-threshold), 
						(int32)(x+threshold),
						(int32)(y+threshold),
						CTDB_BY_PATCH);
    while (ctdb_next_patch(search_space_ptr, &patch_number))
      while (num_vertices = ctdb_next_volume(search_space_ptr, 
					     &type, 
					     &avoid, 
					     &force_id, 
					     &current_mes_key,
					     verts, roof_elevation))
	if (type == CTDB_FS_MES)
	{
	    int32 i, closest_vertex_dist_sq = threshold*threshold + 1;

	    current_mes_id =  CTDB_MES_NODE_KEY_TO_MES_ID(current_mes_key);
	    if (current_mes_id ==  CTDB_INVALID_MES_ID)
	    {
		printf("Error in next_volume: received MES with invalid id\n");
		return 0;
	    }

	    /* Need a measure to determine which MES is closest. 
	       Using distance to the vertices of the roofline at half
	       the elevation 
	     */
	    for (i=0; i < num_vertices; i++)
	    {
		/* Compute distance to the vertices.  We assume that
		   the MES has a lower z value of 0, and we compute
		   the distance to the vertices at the "ave height" level
		 */
		dist_sq = ((verts[i][X] - x)*(verts[i][X] - x) + 
			   (verts[i][Y] - y)*(verts[i][Y] - y) + 
			   (roof_elevation[i]/2-z)*(roof_elevation[i]/2-z));
		if (dist_sq < closest_vertex_dist_sq)
		  closest_vertex_dist_sq = dist_sq;
	    }
	    dist_sq = closest_vertex_dist_sq;

#if 0
	    /* Quick approximation to distance to MES */
	    mes = &(ctdb->mes_volumes[current_mes_id]);
	    if ((int32)mes->origin[CELL3D] != ctdb_get_cell_id(ctdb))
	    {
		gcs_gcs3e_to_gcscs((int32) mes->origin[CELL3D],
				   mes->origin[X], mes->origin[Y], 
				   mes->origin[Z],
				   ctdb_get_cell_id(ctdb),
				   &mes_x, &mes_y, &mes_z);
	    }
	    else
	    {
		mes_x = mes->origin[X];
		mes_y = mes->origin[Y];
		mes_z = mes->origin[Z];
	    }

	    dist_sq = ((mes_x - x)*(mes_x -x) + 
		       (mes_y - y)*(mes_y -y) + 
		       (mes_z - z)*(mes_z -z));
#endif

	    if (dist_sq <= best_dist_sq)
	    {
	        found = TRUE;
		CTDB_MES_ID_TO_MES_KEY(current_mes_id, *mes_found);
		if (ctdb_point_on_database(ctdb, x, y))
		  *out_ctdb = ctdb;
		else
		{
		    float64 x1, y1, z1;
		    int32 cell;
		    gcs_gcs3e_to_gcs(ctdb_get_cell_id(ctdb), x, y, z,
				     &cell, &x1, &y1, &z1);
		    *out_ctdb = (CTDB *) gcs_get_tdb(cell, FALSE);
		}
		best_dist_sq = dist_sq;
	    }
	}
    
    ctdb_destroy_search_space(search_space_ptr);
    return found;
}

/* ctdb_mes_get_mes
 *
 * This routine will return the list of the vertices that make up the
 * roofline of the mes-volume.  The item num_vertices must be set on input
 * to the size of the allocated roofline_vertices array. 
 * If the number of vertices of the roofline exceeds num_vertices,
 * then num_vertices is set to the number of vertices of the roofline,
 * and the function returns 0.  Otherwise, this routine will return
 * the list of roofline vertices in the roofline_vertices array,
 * set num_vertices to this number, and return 1.
 */
int32
ctdb_mes_get_mes(
		 CTDB              *ctdb,
		 CTDB_MES_NODE_KEY  mes_volume,
		 CTDB_MES_TYPE     *type,
		 int32             *num_roof_vertices,
		 float64            roofline_vertices[][XYZ],
		 int32             *num_floor_vertices,
		 float64            floorline_vertices[][XYZ]
		 )
{    
    uint16           mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(mes_volume);
    CTDB_MES_VOLUME *mes;
    int32            i;
    int32            roofline_limit;
    int32            floorline_limit;
    
    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
    {
	if (num_floor_vertices)
	  *num_floor_vertices = 0;
	if (num_roof_vertices)
	  *num_roof_vertices = 0;
	if (type)
	  *type = CTDB_MES_UNKNOWN;
	return 1;
    }

    mes = &(ctdb->mes_volumes[mes_id]);

    if (type)
      *type = mes->template->type;

    if ( roofline_vertices )
      roofline_limit  = *num_roof_vertices;
    else
      roofline_limit  = 0;
    
    if ( num_roof_vertices )
      *num_roof_vertices  = mes->roofline.num_vertices;

    if ( floorline_vertices )
      floorline_limit = *num_floor_vertices;
    else
      floorline_limit = 0;
      
    if ( num_floor_vertices )
      *num_floor_vertices = mes->floorline.num_vertices;

    if (roofline_vertices  && (mes->roofline.num_vertices  > roofline_limit))
      return (roofline_vertices != NULL);
    
    if (floorline_vertices && (mes->floorline.num_vertices > floorline_limit))
      return (floorline_vertices != NULL);
    
    if (roofline_vertices)
    {
        if (roofline_limit > mes->roofline.num_vertices)
	  roofline_limit = mes->roofline.num_vertices;
        for (i=0; i < roofline_limit; i++)
	{
	    roofline_vertices[i][X] = mes->roofline.vertices[i][X];
	    roofline_vertices[i][Y] = mes->roofline.vertices[i][Y];
	    roofline_vertices[i][Z] = mes->roofline.vertices[i][Z];
	}
    }

    if (floorline_vertices)
    {
        if (floorline_limit > mes->floorline.num_vertices)
	  floorline_limit = mes->floorline.num_vertices;
        for (i=0; i < floorline_limit; i++)
        {
	    floorline_vertices[i][X] = mes->floorline.vertices[i][X];
	    floorline_vertices[i][Y] = mes->floorline.vertices[i][Y];
	    floorline_vertices[i][Z] = mes->floorline.vertices[i][Z];
	}
    }
    return 1;
}


/* ctdb_mes_find_enclosure
 *
 * If there is an enclosure at this point x, y, z in the given mes,
 * *enclosure equals the node_key of the enclosure; if the point is outside
 * of the mes, *enclosure equals the key of the "outside" enclosure.  
 * 
 * Otherwise, this function returns FALSE, and the 
 * value of *enclosure is undefined.
 * This only happens if you are between walls 
 * 
 * This function is a wrapper around ctdb_mes_find_node.
 */
int32
ctdb_mes_find_enclosure(
			CTDB*               ctdb,
			CTDB_MES_NODE_KEY   mes,
			float64             x, 
			float64             y, 
			float64             z,
			CTDB_MES_NODE_KEY  *enclosure
			)
{
    return ctdb_mes_find_node(ctdb, mes, x, y, z, enclosure);
}

int32
ctdb_mes_global_find_enclosure(CTDB*               ctdb,
			       float64             x, 
			       float64             y, 
			       float64             z,
			       CTDB*              *mes_ctdb,
			       CTDB_MES_NODE_KEY  *enclosure
			       )

{
    CTDB_MES_NODE_KEY  mes_key;

    /* Initialize output variables */
    *enclosure = CTDB_INVALID_MES_NODE_KEY;
    *mes_ctdb  = ctdb;
    
    if (!ctdb)
      return 0;

    if (!ctdb_mes_find_mes(ctdb, x, y, z, 0.0, mes_ctdb, &mes_key))
      return 0;

    if (!ctdb_mes_find_enclosure(ctdb, mes_key, x, y, z, enclosure))
      return 0;

    return 1;
}


int32
ctdb_mes_find_enclosure_with_guess(CTDB*               ctdb,
				   CTDB_MES_NODE_KEY   guess_enclosure, 
				   float64             x, 
				   float64             y, 
				   float64             z,
				   CTDB*              *found_ctdb,
				   CTDB_MES_NODE_KEY  *found_enclosure)
{
    CTDB_MES_NODE_KEY    mes_key;

    *found_ctdb      = ctdb;
    *found_enclosure = CTDB_INVALID_MES_NODE_KEY;
#ifdef WHEN_THIS_FUNCTION_GETS_WRITTEN
    if (ctdb_mes_xyz_in_enclosure(ctdb, x, y, z, guess_enclosure))
      return TRUE;
#endif
    
    CTDB_NODE_KEY_TO_MES_KEY(guess_enclosure, mes_key);    
    if (ctdb_mes_find_enclosure(ctdb, mes_key, x, y, z, found_enclosure))
      return TRUE;

    return ctdb_mes_global_find_enclosure(ctdb,
					  x, y, z,
					  found_ctdb,
					  found_enclosure);
}

/* ctdb_mes_get_enclosure_polygons
 * Return a list of the polygons making up an enclosure.
 */
CTDB_RESULT_CODE
ctdb_mes_get_enclosure_polygons(CTDB*              ctdb,
				CTDB_MES_NODE_KEY  enclosure,
				int32             *num_polygons,
				CTDB_MES_POLYGON **polygons
				)
{
    int32              i, j, k, count;
    int32             *vertex;
    uint16             mes_id;
    uint16             enclosure_id;
    int32              start_id, end_id, triangle_id;
    CTDB_MES_VOLUME   *mes;
    CTDB_MES_TEMPLATE *template;
    CTDB_RESULT_CODE   status;

    mes_id       = CTDB_MES_NODE_KEY_TO_MES_ID(enclosure);

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
    {
	*num_polygons = 0;
	return CTDB_RESULT_INVALID_PARAMETERS;
    }
    
    enclosure_id = CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure);
    mes          = &(ctdb->mes_volumes[mes_id]);
    template     = mes->template;

    if (enclosure_id == CTDB_INVALID_MES_ENCLOSURE_ID)
    {
	/* Return all polygons */
	start_id = 0;
	end_id = template->num_enclosures;
    }
    else if (CTDB_MES_VALID_ENCLOSURE_ID(ctdb, mes_id, enclosure_id))
    {
	/* Return polygons of this enclosure only */
	start_id = end_id = enclosure_id;
	end_id++;
    }
    else
    {
	/* Oops, invalid enclosure specified */
	*num_polygons = 0;
	return CTDB_RESULT_INVALID_PARAMETERS;
    }
    
    /* If the user has passed in a NULL pointer then the intent was to have
     * this routine determine the size of the output array and allocate
     * sufficient storage space to hold the data.
     */
    if (*polygons == NULL)
    {
	int32 total = 0;
	for (i=start_id; i < end_id; i++)
	  total += template->all_enclosures[i].num_triangles;
	
	*polygons = ctdb_alloc(*polygons, total*sizeof(CTDB_MES_POLYGON),
			       "Polygon array");
	*num_polygons = total;
    }
    
    count = 0;
    status =  CTDB_RESULT_NO_ERROR;

    for (i=start_id; i < end_id; i++)
      for (j=0; j < (template->all_enclosures)[i].num_triangles; j++,count++)
      {
	  triangle_id = CTDB_MES_GET_ENCLOSURE_TRIANGLE_ID(template, i, j);
	  if ( count < *num_polygons )
	  {
	      for (k=0; k < 3; k++)
	      {
		  vertex = CTDB_MES_GET_VERTEX(template, triangle_id, k);
		  ctdb_mes_to_gcscs(mes, vertex[X], vertex[Y], vertex[Z],
				    ctdb_get_cell_id(ctdb),
				    &(*polygons)[count].vertices[k][X],
				    &(*polygons)[count].vertices[k][Y],
				    &(*polygons)[count].vertices[k][Z]);
		  
	      }
	      (*polygons)[count].num_vertices = 3;
	      ctdb_mes_get_triangle_normal(ctdb, mes_id, triangle_id,
					   (*polygons)[count].normal);
	      (*polygons)[count].material =
		template->all_triangles[triangle_id].material;
	  }
	  else
	    status = CTDB_RESULT_ARRAY_TOO_SMALL;
      }

    *num_polygons = count;
    return status;
}

int32
ctdb_mes_get_aperture_enclosures(
				 CTDB*              ctdb,
				 CTDB_MES_EDGE_KEY  aperture,
				 CTDB_MES_NODE_KEY  enclosures[2],
				 float64            directions[2][XYZ]
				 )
{
    uint16             mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID(aperture);
    uint16             aperture_id =CTDB_MES_EDGE_KEY_TO_APERTURE_ID(aperture);
    CTDB_MES_TEMPLATE *template;
    CTDB_MES_APERTURE *ap;
    int32 i;

    if (!ctdb || 
	(mes_id == CTDB_INVALID_MES_ID) || 
	(mes_id >= ctdb->num_mes_volumes))
      return 0;

    template = ctdb->mes_volumes[mes_id].template;
    
    if ((aperture_id == CTDB_INVALID_MES_APERTURE_ID) || 
	(aperture_id >= template->num_apertures))
      return 0;
    
    ap = &template->all_apertures[aperture_id];

    for (i=0; i < 2; i++)
      CTDB_MES_IDS_TO_NODE_KEY(mes_id, ap->connections[i].enclosure_id,
			       enclosures[i]);

    if (directions)
    {
	ctdb_mes_get_triangle_normal(ctdb, mes_id, 
				     ap->first_triangle_id,
				     directions[0]);
	vmat3_negate64(directions[0], directions[1]);
    }
    return 1;
}

int32
ctdb_mes_get_aperture_centroid(
			       CTDB              *ctdb,
			       CTDB_MES_EDGE_KEY  aperture,
			       float64           *x,
			       float64           *y,
			       float64           *z
			       )
{
    uint16             mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID(aperture);
    uint16             aperture_id =CTDB_MES_EDGE_KEY_TO_APERTURE_ID(aperture);
    CTDB_MES_TEMPLATE *template;
    CTDB_MES_APERTURE *apert;
    int32 i;
    float64 center_x = 0;
    float64 center_y = 0;
    float64 center_z = 0;
    float64 gcs_x, gcs_y, gcs_z;
    int32 vertex_id;

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id) ||
	!CTDB_MES_VALID_APERTURE_ID(ctdb, mes_id, aperture_id))
      return 0;

    template = ctdb->mes_volumes[mes_id].template;
    apert = &(template->all_apertures[aperture_id]);

    for (i=0; i < apert->outline.num_vertices; i++)
    {
	vertex_id = apert->outline.vertex_ids[i];
	
	ctdb_mes_to_gcscs(&(ctdb->mes_volumes[mes_id]),
			  template->all_vertices[vertex_id][X],
			  template->all_vertices[vertex_id][Y],
			  template->all_vertices[vertex_id][Z],
			  ctdb_get_cell_id(ctdb),
			  &gcs_x, &gcs_y, &gcs_z);
	center_x += gcs_x;
	center_y += gcs_y;
	center_z += gcs_z;
    }
    *x = center_x / apert->outline.num_vertices;
    *y = center_y / apert->outline.num_vertices;
    *z = center_z / apert->outline.num_vertices;
    return 1;
}





/* ===================  MES Topology Functions ================== */


/* ctdb_mes_find_node
 *
 * returns the node corresponding to the location x, y, z in the CTDB
 * for the multi-elevation structure mes.  If this point lies outside
 * of the multi-elevation structure, the "Outside" node for this
 * multi-elevaiton structure is returned.  If the mes is invalid
 * node is set to NULL.
 *
 * Returns 0 if you are between walls.  1 otherwise.
 */
int32
ctdb_mes_find_node(
                   CTDB               *ctdb,
		   CTDB_MES_NODE_KEY   mes,
		   float64             x, 
		   float64             y, 
		   float64             z, 
		   CTDB_MES_NODE_KEY  *node
		   )
{
    uint16             mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(mes);
    CTDB_MES_VOLUME   *mes_volume;
    CTDB_MES_TEMPLATE *templ;
    float64            rbt[XYZ];
    float64            mes_loc[XYZ];
    int32              j;
    uint16             encl_id;
    int32              n_mes_inter1,n_mes_inter2;
    CTDB_MES_INTERSECTION *mes_inter;
    static float64     gcs_up[XYZ]={0,0,1};
    float64            mes_dir[XYZ];
    float64            mes_offset[XYZ], max_scale;
    CTDB_MES_GRID_BOX  *outside_box;

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
    {
	CTDB_MES_MAKE_NODE_KEY_INVALID(*node);
	return 0;
    }

    mes_volume = &(ctdb->mes_volumes[mes_id]);
    templ = mes_volume->template;
    outside_box = &templ->all_boxes[CTDB_MES_EXTERIOR];
    
    /* Transform the start point to MES coordinates */
    if (!ctdb_gcs_to_mes(mes_volume, ctdb->cell_id, x, y, z,
			 &mes_loc[X], &mes_loc[Y], &mes_loc[Z]))
    {
	/* Point lies outside of mes */
        CTDB_MES_ID_TO_MES_KEY(mes_id, *node);
	return 1;
    }
#if 0
    /* Find the maximum distance in any direction and scale the MES unit 
     * vector by this.  This will ensure a vector that is moving up in the
     * GCS frame of reference and out of the MES box.
     */
    mes_offset[X] = outside_box->right_back_top_corner[X] -
      outside_box->left_front_bottom_corner[X];
    mes_offset[Y] = outside_box->right_back_top_corner[Y] -
      outside_box->left_front_bottom_corner[Y];
    mes_offset[Z] = outside_box->right_back_top_corner[Z] -
      outside_box->left_front_bottom_corner[Z];

    max_scale = NS_MAX(NS_MAX(mes_offset[X],mes_offset[Y]),mes_offset[Z]);
    ctdb_gcsvec_to_mesvec(mes_volume, ctdb->cell_id, gcs_up, mes_dir);
    vmat3_scal_vec_mul64(max_scale, mes_dir, mes_dir);

    /* end of the up ray */
    vmat3_vec_add64(mes_loc, mes_dir, rbt);
#else
    ctdb_gcs_to_mes(mes_volume, ctdb->cell_id, x, y, z + LARGE_NUMBER,
		    &rbt[X], &rbt[Y], &rbt[Z]);
    vmat3_vec_sub64(rbt, mes_loc, mes_dir);
#endif
    /* run the ray up through the MES structure */
    ctdb_mes_line_intersection(mes_volume, mes_loc, rbt, TRUE,
			       CTDB_MES_ALL_ENCLOSURES, FALSE, TRUE,
			       TRUE, &n_mes_inter1, &mes_inter);

    for (j=0;j<n_mes_inter1;j++)
    {
	/* Since there may be more than one intersection at the same distance
	 * and the list is sorted from closest to further intersection, only
	 * look at points that are at the same distance as the first inter-
	 * section returned.
	 */
#if 0
	if (mes_inter[0].distance != mes_inter[j].distance)
	  break;
#endif	
	encl_id = ctdb_mes_enclosure_containing_triangle(ctdb, mes_id,
							 mes_inter[j].triangle,
							 mes_dir);
	if (encl_id != CTDB_INVALID_MES_ENCLOSURE_ID)
	{
	    CTDB_MES_IDS_TO_NODE_KEY(mes_id, encl_id, *node);
	    return 1;
	}
    }
    
    /* At this point we will run a ray downward and try to hit the floor 
     * of the MES enclosure.  Generate a point that is straight down from 
     * the current (x,y) position. 
     */

    /* Change the direction of the MES ray */
    vmat3_negate64(mes_dir, mes_dir);
    vmat3_vec_add64(mes_loc, mes_dir, rbt);

    ctdb_mes_line_intersection(mes_volume, mes_loc, rbt, TRUE,
			       CTDB_MES_ALL_ENCLOSURES, FALSE, TRUE,
			       TRUE, &n_mes_inter2, &mes_inter);

    /* If there were no intersections in the first and second running of the 
     * ray then we must be outside of the MES structure.
     */
    if (!n_mes_inter1 && !n_mes_inter2)
    {
	CTDB_MES_ID_TO_MES_KEY(mes_id, *node);
	return 1;
    }
    

    for (j=0;j<n_mes_inter2;j++)
    {
	/* Since there may be more than one intersection at the same distance
	 * and the list is sorted from closest to further intersection, only
	 * look at points that are at the same distance as the first inter-
	 * section returned.
	 */
#if 0
	if (mes_inter[0].distance != mes_inter[j].distance)
	  break;
#endif
	
	encl_id = ctdb_mes_enclosure_containing_triangle(ctdb, mes_id,
							 mes_inter[j].triangle,
							 mes_dir);
	if (encl_id != CTDB_INVALID_MES_ENCLOSURE_ID)
	{
	    CTDB_MES_IDS_TO_NODE_KEY(mes_id, encl_id, *node);
	    return 1;
	}
    }
#define CTDB_DEBUG
#ifdef CTDB_DEBUG
    printf("libCTDB: FindNode Failed - between walls or bad normals\n");
#endif
    CTDB_MES_MAKE_NODE_KEY_INVALID(*node);
    return 0;
}

/* ctdb_mes_get_topology
 *
 * Returns the "outside" node for the given mes.
 * Sets node to NULL if mes is invalid.
 */
void 
ctdb_mes_get_topology(
		      CTDB               *ctdb,
		      CTDB_MES_NODE_KEY   mes,
		      CTDB_MES_NODE_KEY  *node
		      )
{
    uint16 mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(mes);

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
    {
        CTDB_MES_MAKE_NODE_KEY_INVALID(*node);
	return;
    }

    CTDB_MES_ID_TO_MES_KEY(mes_id,*node);
}

/* ctdb_mes_get_connected_edges
 * 
 * This routine will return the list of edges connected to the 
 * given node.  The item num_edges must be set on input to the
 * size of the allocated array edges.  If the number of edges connected
 * to the node exceeds num_edges, this function will return 0, and
 * num_edges will be set to the number of edges connected to the node.
 * This number should be used to reallocate the edges list.
 *
 * Otherwise, the function will return 1, and place all the connected
 * edges into the edges array, and set num_edges to the number of elts
 * in the array
 */
int32
ctdb_mes_get_connected_edges(
			     CTDB              *ctdb,
			     CTDB_MES_NODE_KEY  node, 
			     int32             *num_edges,
			     CTDB_MES_EDGE_KEY  edges[]
			     )
{
    uint16               mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(node);
    CTDB_MES_ENCLOSURE  *encl; 
    int32               i;
    
    if (!ctdb || !ctdb_mes_valid_node_key(ctdb, node))
    {
	*num_edges = 0;
	return 1;
    }

    encl = CTDB_MES_NODE_KEY_TO_ENCLOSURE_PTR(ctdb,node);

    if (*num_edges < encl->num_apertures)
    {
	*num_edges = encl->num_apertures;
	return 0;
    }
    
    for (i=0; i < encl->num_apertures; i++)
      CTDB_MES_IDS_TO_EDGE_KEY(mes_id, encl->aperture_ids[i], edges[i]);
    *num_edges = encl->num_apertures;
    return 1;
}


#define ADD_CONNECTION_ENTRY \
{\
    if (cnt >= *num_connections)\
      ret = 0;\
    else\
    {\
       if (connected_edges) \
         CTDB_MES_IDS_TO_EDGE_KEY(mes_id, edge_id, connected_edges[cnt]);  \
       if (connected_nodes) \
         CTDB_MES_IDS_TO_NODE_KEY(mes_id,connected_node_id,connected_nodes[cnt]);\
    }\
    cnt++;\
}

static int32 
ctdb_mes_check_feasible_connection(CTDB*   ctdb, 
				   uint16  mes_id,  
				   uint16  src, 
				   uint16  edge, 
				   uint16  dest,
				   float64 climb_threshold, 
				   float64 drop_threshold)
{
    float64 min[XYZ], delta_z;
    CTDB_MES_TEMPLATE *template;
    CTDB_MES_VOLUME *mes;
    CTDB_MES_APERTURE *aperture;
    float64 x, y, z;
    float64 offset, point[XYZ], delta_vec[XYZ];
    float64 elev_lookup;
    int32 i;
    int32 t_id;
    static CTDB_ELEV_QUALIFIER qual = (CTDB_INCL_VOLUME | 
				       CTDB_ONLY_POSITIVE_SURFACES);
    CTDB_ELEV_DATA elevs[CTDB_MAX_ELEVS];
    float64 unit_normal[XYZ];
    int32 src_scale, dest_scale, which;
    CTDB_MES_NODE_KEY  mes_key;

    template = ctdb->mes_volumes[mes_id].template;
    mes      = &ctdb->mes_volumes[mes_id];
    aperture = &template->all_apertures[edge];

    /* Do not allow passage through a closed aperture */
    if (!aperture->open)
      return 0; 

    /* Allow passage through any continuous apertures */
    which = aperture->connections[0].enclosure_id == src;
    if (!aperture->connections[which].step_to_enclosure)
      return 1;

    /* Do not allow passage if the aperture is not continous and
       there is no climb/fall thresholds */
    if ((climb_threshold == drop_threshold) && (climb_threshold == 0.0))
      return 0;
    
    src_scale = (src == aperture->connections[0].enclosure_id) ? 1 : -1;
    dest_scale = (dest == aperture->connections[0].enclosure_id) ? 1 : -1;

    /* Find the lowest point in the MES aperture.  
     * NOTE: we have to convert each vertex to GCS to determine which is
     * the "lowest" point, since our definition of "lowest" is in the 
     * GCS frame of reference
     */
    min[Z] = LARGE_NUMBER;

    for (i=0; i < aperture->outline.num_vertices; i++)
    {
	int32 vtx_id = aperture->outline.vertex_ids[i];

	ctdb_mes_to_gcscs(mes, 
			  template->all_vertices[vtx_id][X],
			  template->all_vertices[vtx_id][Y],
			  template->all_vertices[vtx_id][Z],
			  ctdb_get_cell_id(ctdb),
			  &x, &y, &z);
	if (z < min[Z])
	{
	    min[X] = x;
	    min[Y] = y;
	    min[Z] = z;
	}
    }


    /* The aperture outline is defined to be in the center of the 
     * aperture.  
     * 
     * In order to do elevation lookups to determine the climb/drop from
     * going from and enclosure to an aperture, or aperture to an enclosure,
     * we have to move our point beyond the thickness of the aperture,
     * before doing the lookup.  We determine the direction of moving
     * away from the aperture using the normal of a triangle of the aperture.
     * This assumes that the apertures are planar.
     */

    t_id = CTDB_MES_GET_APERTURE_TRIANGLE_ID(template, edge, 0);
    ctdb_mes_get_triangle_normal(ctdb, mes_id, t_id, unit_normal);
    
    /* Check movement from the src enclosure to the aperture */
    if (src != CTDB_INVALID_MES_ENCLOSURE_ID )
    {
	int32 which = (src == aperture->connections[1].enclosure_id);
	offset = aperture->connections[which].wall_thickness * 
	         template->fixed_point_basis + 0.01;
	
	vmat3_scal_vec_mul64(offset*src_scale, unit_normal, delta_vec);
	/* Check if first enclosure or not! */
	vmat3_vec_add64(delta_vec, min, point);
	
	/* Find positive surface */
	CTDB_MES_IDS_TO_NODE_KEY(mes_id, src, mes_key);
	if (!ctdb_lookup_enclosure_elevation(ctdb, mes_key,
					     point[X], point[Y],
					     CTDB_ALL_POSITIVE_ANGLES, 
					     qual, elevs))
	{
	    printf("Error in check_feasible_connections.  No surfaces\n");
	    return 0;
	}
	elev_lookup = elevs->z;
	
	delta_z = min[Z] - elev_lookup;
	if (delta_z < 0)
	{
	    if ((-delta_z) > drop_threshold)
	      return 0;
	}
	else
	  if (delta_z > climb_threshold)
	    return 0;
    }

    /* Check movement from the aperture to the destination enclosure */
    if ( dest != CTDB_INVALID_MES_ENCLOSURE_ID )
    {
	int32 which = (dest == aperture->connections[1].enclosure_id);
	offset = aperture->connections[which].wall_thickness * 
	         template->fixed_point_basis + 0.01;
	
	vmat3_scal_vec_mul64(offset*dest_scale, unit_normal, delta_vec);
	/* Check if first enclosure or not! */
	vmat3_vec_add64(delta_vec, min, point);
	CTDB_MES_IDS_TO_NODE_KEY(mes_id, dest, mes_key);
	if (!ctdb_lookup_enclosure_elevation(
					     ctdb, mes_key,
					     point[X], point[Y],
					     CTDB_ALL_POSITIVE_ANGLES, 
					     qual, elevs))
	{
	    printf("Error in check_feasible_connections.  No surfaces\n");
	    return 0;
	}
	elev_lookup = elevs->z;

	delta_z = elev_lookup - min[Z];
	if (delta_z < 0)
	{
	    if (-delta_z > drop_threshold)
	      return 0;
	}
	else
	  if (delta_z > climb_threshold)
	    return 0;
    }
    return 1;
}

int32
ctdb_mes_filtered_connected_edges(
				CTDB             *ctdb,
				CTDB_MES_NODE_KEY node, 
				float64           climb_threshold,
				float64           drop_threshold,
				int32            *num_edges,
				CTDB_MES_EDGE_KEY edges[]
				)
{
    uint16              mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(node);
    CTDB_MES_ENCLOSURE  *enclosure;
    int32               i, cnt = 0;
    CTDB_MES_TEMPLATE  *templ;
    uint16              from_id;

    if (!ctdb || !ctdb_mes_valid_node_key(ctdb, node))
    {
	*num_edges = 0;
	return 1;
    }

    enclosure = CTDB_MES_NODE_KEY_TO_ENCLOSURE_PTR(ctdb, node);
    templ     = ctdb->mes_volumes[mes_id].template;
    from_id   = CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node);

    for (i=0; i < enclosure->num_apertures; i++)
    {
	uint16 via_id, to_id;
	int32  which;
	CTDB_MES_APERTURE  *aperture;
	
	via_id   = enclosure->aperture_ids[i];
	aperture = &templ->all_apertures[via_id];
	which    = (aperture->connections[0].enclosure_id == from_id);
	to_id    = aperture->connections[which].enclosure_id;

	if (ctdb_mes_check_feasible_connection(ctdb, mes_id, 
					       from_id, via_id, to_id,
					       climb_threshold, 
					       drop_threshold))
	{
	    if (cnt < *num_edges)
	      CTDB_MES_IDS_TO_EDGE_KEY(mes_id, 
				       enclosure->aperture_ids[i],
				       edges[cnt]);
	    cnt++;
	}
    }
    
    if (cnt >= *num_edges)
    {
	*num_edges = cnt;
	return 0;
    }

    *num_edges = cnt;
    return 1;
}

void
ctdb_mes_filtered_connected_nodes(CTDB             *ctdb,
				  CTDB_MES_EDGE_KEY edge,
				  float64           climb_threshold,
				  float64           drop_threshold,
				  int32            *num_nodes,
				  CTDB_MES_NODE_KEY nodes[2]
				  )
{
    uint16             mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID(edge);
    CTDB_MES_APERTURE *aperture;
    int32 i, cnt = 0;
    uint16 current_edge_id;

    if (!ctdb || !CTDB_MES_VALID_EDGE_KEY(ctdb, edge))
    {
	*num_nodes = 0;
	return;
    }

    aperture = CTDB_MES_EDGE_KEY_TO_APERTURE_PTR(ctdb,edge);
    current_edge_id = CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edge);

    for (i=0; i < 2; i++)
    {
	if (ctdb_mes_check_feasible_connection(
					 ctdb, 
					 mes_id,
				         CTDB_INVALID_MES_ENCLOSURE_ID,
				         current_edge_id, 
				         aperture->connections[i].enclosure_id,
				         climb_threshold,drop_threshold))
	{
	    CTDB_MES_IDS_TO_NODE_KEY(mes_id, 
				     aperture->connections[i].enclosure_id,  
				     nodes[cnt]);
	    cnt++;
	}
    }
    *num_nodes = cnt;
}

/* ctdb_mes_get_connected_nodes
 *
 * This routine will return the two nodes that are connected by 
 * the given edge.
 */
int32
ctdb_mes_get_connected_nodes(
			     CTDB*              ctdb,
			     CTDB_MES_EDGE_KEY  edge,
			     CTDB_MES_NODE_KEY  nodes[2]
			     )

{
    uint16             mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID(edge);
    CTDB_MES_APERTURE *aperture;
    int32              i;
    CTDB_MES_TEMPLATE *template;

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
      return 0;

    template = ctdb->mes_volumes[mes_id].template;
    aperture = &(template->
		 all_apertures[CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edge)]);

    for (i=0; i < 2; i++)
      CTDB_MES_IDS_TO_NODE_KEY(mes_id, 
			       aperture->connections[i].enclosure_id, 
			       nodes[i]);
    return 1;
}
/* ctdb_mes_get_edge_info
 * 
 * Since topologic MES edges correspond to physical apertures,
 * this function will return the outline of the aperture corresponding
 * to the given edge.
 */
int32
ctdb_mes_get_edge_info(
		       CTDB             *ctdb,
		       CTDB_MES_EDGE_KEY edge,
		       int16            *visible,
		       int16            *open,
		       int16            *material,
		       float64           aperture_thickness[2],
		       float64           thickness_direction[2][XYZ],
		       int32            *num_vertices,
		       float64           vertices[][XYZ]
		       )
{
    uint16             mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID(edge);
    CTDB_MES_APERTURE *aperture;
    int32              i,cell;
    int32             *vertex;
    CTDB_MES_TEMPLATE *template;

    if (!ctdb || 
	!CTDB_MES_VALID_EDGE_KEY(ctdb, edge))
      return 0;

    aperture = CTDB_MES_EDGE_KEY_TO_APERTURE_PTR(ctdb,edge);
    template = ctdb->mes_volumes[mes_id].template;
    
    if (visible)
      *visible = aperture->visible;

    if (open)
      *open    = aperture->open;

    if (material)
      *material
	= template->all_triangles[aperture->first_triangle_id].material;

    if (*num_vertices < aperture->outline.num_vertices)
    {
	*num_vertices = aperture->outline.num_vertices;
	return 0;
    }
    *num_vertices = aperture->outline.num_vertices;
    cell = ctdb_get_cell_id(ctdb);

    for (i=0; i < aperture->outline.num_vertices; i++)
    {
	vertex = template->all_vertices[(aperture->outline.vertex_ids)[i]];

	/* This is the center of the aperture.  The aperture has two faces 
	 * and these two faces will be returned.
	 */
	ctdb_mes_to_gcscs(&(ctdb->mes_volumes[mes_id]),
			  vertex[X], vertex[Y], vertex[Z], cell,
			  &vertices[i][X], &vertices[i][Y], &vertices[i][Z]);
    }    


    /* $$$ This assumes that the MES instance's rotation matrix is unitary */
    if (aperture_thickness)
    {
	int32 i;
	for (i=0; i < 2; i++)
	  aperture_thickness[i] = aperture->connections[i].wall_thickness * 
	    template->fixed_point_basis;
    }
    
    if (thickness_direction)
    {
	ctdb_mes_get_triangle_normal(ctdb, mes_id, 
				     aperture->first_triangle_id,
				     thickness_direction[0]);
	vmat3_negate64(thickness_direction[0], thickness_direction[1]);
    }

    return 1;
   
}

/* ctdb_mes_get_connections
 *
 * This API function will take a node in a mes given by mes_id
 * in a ctdb, and return a list of all the edges (and the connected nodes)
 * that satisfy the criteria that when traversing from this node through
 * this edge into the connected node, the entity would never have to 
 * ascend a vertical distance greater than climb_threshold, and never
 * have to descend a vertical distance greater than drop_threshold.
 *
 * On input, num_connections must be set the number of elements allocated
 * for connected_edges and connected_nodes.  If the allocation is insufficient,
 * 0 is returned.
 *
 * Otherwise, num_connections is set to the number of connections satisfying
 * the criteria, and connected_edges and connected nodes return the edges
 * and nodes respecively.  NOTE: for any i >= 0, i < num_connections,
 * an entity can traverse from node through connected_edges[i] to 
 * connected_nodes[i] without exceeding either climb/drop thresholds.
 * A value of 1 is returned.
 */
int32
ctdb_mes_get_connections(
			 CTDB             *ctdb,
			 CTDB_MES_NODE_KEY node,
			 float64           climb_threshold,
			 float64           drop_threshold,
			 int32            *num_connections,
			 CTDB_MES_EDGE_KEY connected_edges[],
			 CTDB_MES_NODE_KEY connected_nodes[]
			 )
{
    uint16              mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(node);
    CTDB_MES_ENCLOSURE *enclosure;
    CTDB_MES_TEMPLATE  *template;
    CTDB_MES_APERTURE  *aperture;
    uint16              current_node_id, connected_node_id;
    int32               i, which, ret=1, cnt = 0;
    uint16              edge_id;

    if (!ctdb || !ctdb_mes_valid_node_key(ctdb, node))
    {
	*num_connections = 0;
	return 1;
    }

    enclosure = CTDB_MES_NODE_KEY_TO_ENCLOSURE_PTR(ctdb, node);   
    template = ctdb->mes_volumes[mes_id].template;
    current_node_id = (uint16) (enclosure - template->all_enclosures);

    for (i=0; i < enclosure->num_apertures; i++)
    {
	edge_id = enclosure->aperture_ids[i];
	aperture= &template->all_apertures[edge_id];
	which = 
	  (aperture->connections[0].enclosure_id == current_node_id);
	connected_node_id = aperture->connections[which].enclosure_id;
	
	if (ctdb_mes_check_feasible_connection(ctdb, mes_id, 
					       current_node_id, edge_id,
					       connected_node_id, 
					       climb_threshold,
					       drop_threshold))
	  ADD_CONNECTION_ENTRY
    }
    *num_connections = cnt;
    return ret;
}


CTDB_MES_EDGE_KEY
ctdb_find_close_ext_ap(CTDB            *ctdb,
		       int32            mes_index,
		       float64          pt[XYZ],
		       float64          climb_threshold,
		       float64          fall_threshold)
{
    CTDB_MES_NODE_KEY exterior_node;
    int32 i;
    float64 best_dist, dist;
    CTDB_MES_EDGE_KEY best_ap;
    int32 num_connections;
    static int32 num_conn_alloced = 0;
    static CTDB_MES_EDGE_KEY *connected_edges = NULL;
    static CTDB_MES_NODE_KEY *connected_nodes = NULL;
    
    exterior_node = CTDB_MES_KEY(mes_index, CTDB_MES_EXTERIOR);
    num_connections = num_conn_alloced;
    
    /* Find all exterior apertures (i.e., those connected to the exterior node)
     * meeting the climb/fall criteria
     */
    if (!ctdb_mes_get_connections(ctdb, exterior_node, climb_threshold,
				  fall_threshold, &num_connections,
				  connected_edges, connected_nodes))
    {
	/* Realloc so that all the data will fit */
	connected_edges = ctdb_alloc(connected_edges,
				     num_connections * 
				     sizeof(CTDB_MES_EDGE_KEY),
				     "Closest ext ap connections");

	connected_nodes = ctdb_alloc(connected_nodes,
				     num_connections * 
				     sizeof(CTDB_MES_NODE_KEY),
				     "Closest ext ap nodes");

	num_conn_alloced = num_connections;
	
	ctdb_mes_get_connections(ctdb, exterior_node, climb_threshold,
				 fall_threshold, &num_connections,
				 connected_edges, connected_nodes);
    }
    
    if (!num_connections)
    {
	return CTDB_INVALID_MES_EDGE_KEY;
    }
    
    /* Find closest aperture to point in connections list */
    best_dist = ctdb_dist_sq_to_aperture(ctdb, pt, connected_edges[0]);
    best_ap   = connected_edges[0];
    
    for(i=1; i<num_connections;i++)
    {
	if ((dist = ctdb_dist_sq_to_aperture(ctdb, pt, connected_edges[i])) < 
	    best_dist)
	{
	    best_dist = dist;
	    best_ap   = connected_edges[i];
	}
    }
    
    return best_ap;    
}


/* ==========MES coordinate conversion utility functions========== */

/* ctdb_xyo_to_mes
 * 
 * This function will take a ctdb and patch coordinates
 * (patch_x, patch_y, x, y, and z) and convert it to the local 
 * coordinate system defined by the given mes (mes_x, mes_y, mes_z)
 */
int32
ctdb_xyo_to_mes(
		CTDB             *ctdb, 
		CTDB_MES_VOLUME  *mes,
		int32             patch_x, 
		int32             patch_y,
		float64           x, 
		float64           y, 
		float64           z,
		float64          *mes_x,
		float64          *mes_y,
		float64          *mes_z
		)		

{
   float64 x0, y0, mes_position[XYZ];
   float64 offset[XYZ];
   int32 mes_cell;

   if (!ctdb || !mes)
     return 0;

   mes_cell = mes->origin[CELL3D];
   
   x0 = CTDB_XYO_TO_METERS(x, patch_x, ctdb->patch_incrD, ctdb->incrD);
   y0 = CTDB_XYO_TO_METERS(y, patch_y, ctdb->patch_incrD, ctdb->incrD);

   if ( ctdb_get_cell_id(ctdb) != mes_cell )
   {
       gcs_gcs3e_to_gcscs (ctdb->cell_id, x0, y0, z, mes_cell,
			   &offset[X], &offset[Y], &offset[Z]);
       offset[X] -= mes->origin[X];
       offset[Y] -= mes->origin[Y];
       offset[Z] -= mes->origin[Z];
   }
   else
   {
       offset[X] = x0 - mes->origin[X];
       offset[Y] = y0 - mes->origin[Y];
       offset[Z] = z - mes->origin[Z];
   }

   ctdb_gcsvec_to_mesvec(mes, mes_cell, offset, mes_position);
   *mes_x = mes_position[X];
   *mes_y = mes_position[Y];
   *mes_z = mes_position[Z];
   return 1;
}

int32
ctdb_mes_to_xyo ( 
		 CTDB            *ctdb, 
		 CTDB_MES_VOLUME *mes,
		 float64          mes_x, 
		 float64          mes_y, 
		 float64          mes_z, 
		 int32            patch_x, 
		 int32            patch_y,
		 float64         *x, 
		 float64         *y, 
		 float64         *z 
		 )
{
    int32   mes_cell;
    int32   gcscell;
    float64 gcsvec[XYZ], mes_position[XYZ];
    float64 scale;

    if (!ctdb || !mes)
      return 0;

    mes_cell = mes->origin[CELL3D];

    mes_position[X] = mes_x;
    mes_position[Y] = mes_y;
    mes_position[Z] = mes_z;
    
    ctdb_mesvec_to_gcsvec(mes, mes_position, &gcscell, gcsvec);

    /* This should never be true */
    if ( gcscell != mes_cell );
    {
        gcs_gcs3_to_gcscs(gcscell, gcsvec, mes_cell, gcsvec);
    }

    scale = ctdb->inv_incr * ctdb->patch_inv_incr;
    *x = CTDB_METERS_TO_XYO ((mes->origin[X] + gcsvec[X]), patch_x, scale);
    *y = CTDB_METERS_TO_XYO ((mes->origin[Y] + gcsvec[Y]), patch_y, scale);
    *z = mes->origin[Z] + gcsvec[Z];
    return 1;
}


/* ctdb_gcs_to_mes
 *
 * This function will take a point in GCS (cell, x0, y0, z0)  
 * and convert it to the local coordinate system defined by 
 * the mes (x1, y1, z1).
 * 
 * If the gcs point lies outside the given mes, return 0.
 * Otherwise, we return 1.
 */
int32
ctdb_gcs_to_mes(
		CTDB_MES_VOLUME *mes,
		int32            cell,
		float64          x0,
		float64          y0,
		float64          z0,
		float64         *x1,
		float64         *y1,
		float64         *z1
		)
{
    float64 offset[XYZ], tmp[XYZ];
    int32 mes_cell;
    CTDB_MES_TEMPLATE *templ;

    if (!mes)
      return 0;

    mes_cell = mes->origin[CELL3D];
    templ = mes->template;

    if (cell != mes_cell)
    {
	gcs_gcs3e_to_gcscs(cell, x0, y0, z0, mes_cell,
			 &offset[X], &offset[Y], &offset[Z]);
	offset[X] -= mes->origin[X];
	offset[Y] -= mes->origin[Y];
	offset[Z] -= mes->origin[Z];
    }
    else
    {
	offset[X] = x0 - mes->origin[X];
	offset[Y] = y0 - mes->origin[Y];
	offset[Z] = z0 - mes->origin[Z];
    }
    ctdb_gcsvec_to_mesvec(mes, (int32) mes->origin[CELL3D],
			 offset, tmp);

    *x1 = tmp[X];
    *y1 = tmp[Y];
    *z1 = tmp[Z];
    if (((tmp[X] >= 
	  templ->all_boxes[CTDB_MES_EXTERIOR].left_front_bottom_corner[X]) &&
	 (tmp[X] <= 
	  templ->all_boxes[CTDB_MES_EXTERIOR].right_back_top_corner[X])) &&
	((tmp[Y] >= 
	  templ->all_boxes[CTDB_MES_EXTERIOR].left_front_bottom_corner[Y]) &&
	 (tmp[Y] <= 
	  templ->all_boxes[CTDB_MES_EXTERIOR].right_back_top_corner[Y])) &&
	((tmp[Z] >= 
	  templ->all_boxes[CTDB_MES_EXTERIOR].left_front_bottom_corner[Z]) &&
	 (tmp[Z] <= 
	  templ->all_boxes[CTDB_MES_EXTERIOR].right_back_top_corner[Z])))
      return 1;
    return 0;
}

/* ctdb_mes_to_gcs
 *
 * This function will take a defined by the mes's local coordinate 
 * system (x0, y0, z0), and converts it to a GCS point (cell, x1, y1, z1).
 */
int32
ctdb_mes_to_gcs(
		CTDB_MES_VOLUME *mes,
		int32            x0,
		int32            y0,
		int32            z0,
		int32           *cell,
		float64         *x1,
		float64         *y1,
		float64         *z1
		)
{
    int32 gcscell;
    float64 mesvec[XYZ], gcsvec[XYZ];

    if (!mes)
      return 0;

    mesvec[X] = (float64) x0;
    mesvec[Y] = (float64) y0;
    mesvec[Z] = (float64) z0;

    ctdb_mesvec_to_gcsvec(mes, mesvec, &gcscell, gcsvec);
    
    /* This should never be true */
    if (gcscell != (int32) mes->origin[CELL3D])
      gcs_gcs3_to_gcscs(gcscell, gcsvec, 
			 (int32) mes->origin[CELL3D], gcsvec);
    *x1 = mes->origin[X] + gcsvec[X];
    *y1 = mes->origin[Y] + gcsvec[Y];
    *z1 = mes->origin[Z] + gcsvec[Z];
    *cell = (int32) mes->origin[CELL3D];
    return 1;
}

/* ctdb_mes_to_gcs
 *
 * This function will take a defined by the mes's local coordinate 
 * system (x0, y0, z0), and converts it to a GCS point (x1, y1, z1)
 * in a specified cell.
 */
int32
ctdb_mes_to_gcscs(
		  CTDB_MES_VOLUME *mes,
		  int32            x0,
		  int32            y0,
		  int32            z0,
		  int32            cell,
		  float64         *x1,
		  float64         *y1,
		  float64         *z1
		  )
{
    float64          origin[XYZ];
    float64          mesvec[XYZ], gcsvec[XYZ];
    int32            gcscell;
    
    
    if (!mes)
      return 0;

    mesvec[X] = (float64) x0;
    mesvec[Y] = (float64) y0;
    mesvec[Z] = (float64) z0;

    ctdb_mesvec_to_gcsvec(mes, mesvec, &gcscell, gcsvec);

    if ((int32) mes->origin[CELL3D] != cell)
      gcs_gcs3_to_gcscs((int32) mes->origin[CELL3D], mes->origin,
			cell, origin);
    else
    {
	origin[X] = mes->origin[X];
	origin[Y] = mes->origin[Y];
	origin[Z] = mes->origin[Z];
    }
    
    if (gcscell != cell)
      gcs_gcs3_to_gcscs(gcscell, gcsvec, cell, gcsvec);

    *x1 = origin[X] + gcsvec[X];
    *y1 = origin[Y] + gcsvec[Y];
    *z1 = origin[Z] + gcsvec[Z];
    return 1;
}

/* ctdb_gcsvec_to_mesvec
 *
 * This function will take a vector defined in GCS, and convert it 
 * to the local coordinate system defined by the mes.
 */
int32
ctdb_gcsvec_to_mesvec(
		      CTDB_MES_VOLUME   *mes,
		      int32              cell,
		      float64            gcsvec[XYZ],
		      float64            mesvec[XYZ]
		      )
{
    float64            tmp_vec[XYZ];
    CTDB_MES_TEMPLATE *template;

    if (!mes)
      return 0;

    template = mes->template;
    
    if (cell != mes->origin[CELL3D])
      gcs_vec_to_vec(cell, gcsvec, (int32) mes->origin[CELL3D], tmp_vec);
    else
      VMAT3_VEC_COPY64(gcsvec, tmp_vec);

    vmat3_scal_vec_mul64(1/template->fixed_point_basis, tmp_vec, tmp_vec);
    vmat3_mat_vec_mul64(mes->tfm_world_to_mes, tmp_vec, mesvec);
    return 1;
}

/* ctdb_mesvec_to_gcsvec
 *
 * This function takes a vector defined by the mes's local coordinate
 * system and transforms it to a gcs vector in its native cell.
 * The native cell is also returned.
 */
static void
ctdb_mesvec_to_gcsvec(
		      CTDB_MES_VOLUME   *mes,
		      float64            mesvec[XYZ],
		      int32             *cell,
		      float64            gcsvec[XYZ]
		      )
{        
    float64            tmp_vec[XYZ];
    CTDB_MES_TEMPLATE *template;
    float64            inv_matrix[3][3];

    template = mes->template;

    VMAT3_VEC_COPY64(mesvec, tmp_vec);
    vmat3_scal_vec_mul64(template->fixed_point_basis, tmp_vec, tmp_vec);
    vmat3_inverse64(mes->tfm_world_to_mes, inv_matrix);
    vmat3_mat_vec_mul64(inv_matrix, tmp_vec, gcsvec);
    *cell = (int32) mes->origin[CELL3D];
}   

/* ctdb_mesvec_to_gcsvec
 *
 * This function will take a vector defined by the mes's local coordinate
 * system and transforms it to vector in a specific GCS cell.
 */
void 
ctdb_mesvec_to_gcscsvec(
			CTDB_MES_VOLUME *mes,
			float64          mesvec[XYZ],
			int32            cell,
			float64          gcsvec[XYZ]
			)
{
    float64 tmp_vec[XYZ];
    int32 mes_cell;
    CTDB_MES_TEMPLATE *template;
    float64 inv_matrix[3][3];

    template = mes->template;

    VMAT3_VEC_COPY64(mesvec, tmp_vec);
    vmat3_scal_vec_mul64(template->fixed_point_basis, tmp_vec, tmp_vec);
    
    vmat3_inverse64(mes->tfm_world_to_mes, inv_matrix);
    vmat3_mat_vec_mul64(inv_matrix, tmp_vec, gcsvec);
    mes_cell = (int32) mes->origin[CELL3D];

    if (cell != mes_cell)
      gcs_vec_to_vec(mes_cell, gcsvec, cell, gcsvec);
}   

/* Conversion from GCS coordinates to PO enclosure point coordinates.
 * x, y, z are GCS coordinates in the given ctdb.  
 *
 * this returns mes_id, the index into out_ctdb for the given mes.
 * The enclosure that x,y,z are in is returned in enclosure_id.
 * mes_x is the delta between the x and the origin of the mes in x in cms.
 * mes_y is the delta between the y and the origin of the mes in y in cms.
 * mes_z is the delta between the z and the origin of the mes in z.
 */
int32
ctdb_gcs_to_mes_origin_offsets(
			       CTDB*              ctdb,      /* in */
			       float64            x,         /* in */
			       float64            y,         /* in */
			       float64            z,         /* in */
			       int32             *mes_cell,  /* out */
			       CTDB_MES_NODE_KEY *enclosure, /* in-out */
			       int16             *mes_x,     /* out */
			       int16             *mes_y,     /* out */
			       float32           *mes_z      /* out */
			       )
{
    int32              cell;
    CTDB_MES_VOLUME   *mes;
    float64            offset[XYZ];
    uint32             mes_id;

    if (!ctdb)
    {
	*mes_cell = GCS_ILLEGAL_CELL;
	*mes_x = *mes_y = *mes_z = 0;
	return 0;
    }
    
    cell = ctdb->cell_id;

    if (ctdb_is_tiled)
      if (!ctdb_point_on_database(ctdb, x, y)) 
      {
	  float64 new_x, new_y, new_z;
	  int32 new_cell;

	  gcs_gcs3e_to_gcs(cell, x, y, z,
			   &new_cell, &new_x, &new_y, &new_z);

	  if (cell == new_cell)
	  {
	      printf("Infinite recursion in ctdb_gcs_to_mes_origin_offsets\n");
	      *mes_cell = GCS_ILLEGAL_CELL;
	      *mes_x = *mes_y = *mes_z = 0;
	      return 0;
	  }

	  return ctdb_gcs_to_mes_origin_offsets(gcs_get_tdb(new_cell, FALSE), 
						new_x, new_y, new_z,
						mes_cell, enclosure,
						mes_x, mes_y, mes_z);
      }
 
    /* If the enclosure is invalid then try and find the mes that contains
     * the point.  If none exists then set the output parameters to some
     * safe default values and return a failure status.
     */
    if ( !ctdb_mes_valid_node_key(ctdb, *enclosure) )
    {
	*enclosure = CTDB_INVALID_MES_NODE_KEY;

	if (!ctdb_mes_global_find_enclosure(ctdb, x,y,z, &ctdb, enclosure))
	{
	    *mes_cell = GCS_ILLEGAL_CELL;
	    *mes_x = *mes_y = *mes_z = 0;
	    return 0;
	}
    }

    /* At this point there is a valid MES and the point will be converted into
     * the MES frame of reference and coordinate system.
     */
    mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(*enclosure);    
    mes = &(ctdb->mes_volumes[mes_id]);
    *mes_cell = (int32) mes->origin[CELL3D];
    
    if (cell != *mes_cell)
    {
	gcs_gcs3e_to_gcscs(cell, x, y, z, *mes_cell,
			   &offset[X], &offset[Y], &offset[Z]);
	offset[X] -= mes->origin[X];
	offset[Y] -= mes->origin[Y];
	offset[Z] -= mes->origin[Z];
    }
    else
    {
	offset[X] = x - mes->origin[X];
	offset[Y] = y - mes->origin[Y];
	offset[Z] = z - mes->origin[Z];
    }
    
    *mes_x = (int16)  (100*offset[X]);
    *mes_y = (int16)  (100*offset[Y]);
    *mes_z = (float32) (100*offset[Z]);
    return 1;
}

/* Conversion from PO enclosure point to GCS point.  mes_id is the mes 
 * index into ctdb.  mes_x, mes_y are deltas (measured in cms) from the
 * mes origin in the x and y directions (east-west, north-south).  
 * mes_z is delta from the mes_origin in the z-direction (up-down).
 */
int32
ctdb_mes_origin_offsets_to_gcscs(CTDB             *ctdb,
				 CTDB_MES_NODE_KEY mes,
				 int16             mes_x,
				 int16             mes_y,
				 float32           mes_z,
				 float64          *x,
				 float64          *y,
				 float64          *z
				 )
{
    uint16           mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(mes);
    CTDB_MES_VOLUME *mes_volume;
    float64          point[XYZ];

    if (!ctdb || !ctdb_mes_valid_node_key(ctdb, mes))
      return 0;
    
    mes_volume = &ctdb->mes_volumes[mes_id];
    point[X] = ((float64) mes_x) / 100 + mes_volume->origin[X];
    point[Y] = ((float64) mes_y) / 100 + mes_volume->origin[Y];
    point[Z] = ((float64) mes_z) / 100 + mes_volume->origin[Z];

    if ((int32) mes_volume->origin[CELL3D] != ctdb->cell_id)
      gcs_gcs3e_to_gcscs((int32)mes_volume->origin[CELL3D], point[X],
			 point[Y], point[Z], ctdb->cell_id, x, y, z);
    else
    {
	*x = point[X];
	*y = point[Y];
	*z = point[Z];
    }
    return 1;
}


#define MAX_CONNECTED_EDGES 50
static int32
ctdb_connected_enclosures(CTDB  *ctdb,
			  uint16 mes_id,
			  uint16 src,
			  uint16 dest
			  )
{
    CTDB_MES_TEMPLATE  *template;
    CTDB_MES_EDGE_KEY   edges[MAX_CONNECTED_EDGES];
    int32               edge_id, num_edges = MAX_CONNECTED_EDGES;
    int32               i, j;
    CTDB_MES_NODE_KEY   node;

    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
      return 0;

    template = ctdb->mes_volumes[mes_id].template;
    
    if ((src  == CTDB_INVALID_MES_ENCLOSURE_ID ) || 
	(dest == CTDB_INVALID_MES_ENCLOSURE_ID ))
      return 0;

    /* $$$ In the future, can use get_connections to filter out 
     * impossible ones */
    CTDB_MES_IDS_TO_NODE_KEY(mes_id, src, node);
    if (!ctdb_mes_get_connected_edges(ctdb, node, &num_edges, edges))
      printf("Error in ctdb_connected_enclosures: insufficient edges\n");

    for (i=0; i < num_edges; i++)
      for (j=0; j < 2; j++)
      {
	  edge_id = CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edges[i]);
	  if (template->all_apertures[edge_id].connections[j].enclosure_id == 
	      dest)
	    return 1;
      }
    return 0;
}
#undef MAX_CONNECTED_EDGES


#define find_best_dest(check_clear)\
{\
    float64 best_z_diff, tmp_z_diff;\
    best = CTDB_INVALID_MES_ENCLOSURE_ID ;\
    for (i=0; i < cnt; i++)\
    {\
	if (!check_clear || valid_enclosures[i].clear_path)\
	  if (best == CTDB_INVALID_MES_ENCLOSURE_ID )\
	  {\
	      best = i;\
	      best_z_diff\
		= fabs(valid_enclosures[i].location[Z] - departure_point[Z]);\
	  }\
	  else\
	  {\
	      tmp_z_diff\
                = fabs(valid_enclosures[i].location[Z] - departure_point[Z]);\
	      if (tmp_z_diff < best_z_diff)\
	      {\
		  best = i;\
		  best_z_diff = tmp_z_diff;\
	      }\
	  }\
    }\
}

/* Inputs: ctdb, a 3-D point to start from, and a 2-D x,y destination pt.
 * Outputs: mes_key representing the MES and enclosure to go to.
 */

/* In meters */
#define DELTA_THRESHOLD 0.01
#define DEMO_CLIMB_THRESHOLD 2.5
int32
ctdb_find_best_enclosure(
			 CTDB*              ctdb,
			 CTDB_MES_NODE_KEY  departure_mes,
			 float64            departure_point[XYZ],
			 float64            goal_point[XYZ],
			 float64            traversal_height,
			 CTDB_MES_NODE_KEY *arrival_mes,
			 float64            arrival_point[XYZ]
			 )
{
    int32               src_in_mes;
    int32               i, dest_in_mes;
    uint16              dest_mes_id;
    CTDB_MES_NODE_KEY   dest_mes_key;
    CTDB_MES_NODE_KEY   src_enclosure_key;
    CTDB_MES_NODE_KEY   dest_enclosure_keys[MAX_ENCLOSURES];
    uint16              src_mes_id, src_enclosure_id;
    static CTDB_ELEV_QUALIFIER qual = (CTDB_INCL_VOLUME | 
				       CTDB_ONLY_POSITIVE_SURFACES);
    CTDB_ELEV_DATA      elevs[CTDB_MAX_ELEVS];
    float64             z;
    int32               cnt, is_clear = 0;
    uint16              best = CTDB_INVALID_MES_ENCLOSURE_ID;
    int32               num_enclosures =0;
    int32               same_enclosure;
    float64 delta_z;
    int32 closest;
    float64 closest_distance, land_z;


    struct {
	int32               clear_path;
	CTDB_MES_NODE_KEY   enclosure;
	float64             location[XYZ];
    } valid_enclosures[MAX_ENCLOSURES];

    src_mes_id       = CTDB_MES_NODE_KEY_TO_MES_ID(departure_mes);
    src_enclosure_id = CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(departure_mes);
    src_in_mes       = CTDB_MES_VALID_MES_ID(ctdb, src_mes_id) &&
                       !CTDB_MES_NODE_IS_EXTERIOR(departure_mes);

    if (src_in_mes)
      if (!CTDB_MES_VALID_MES_ID(ctdb, src_mes_id) ||
	  !CTDB_MES_VALID_ENCLOSURE_ID(ctdb, src_mes_id, src_enclosure_id))
      {
	  CTDB_MES_MAKE_NODE_KEY_INVALID(*arrival_mes);
	  VMAT3_VEC_COPY64(departure_point,arrival_point);
	  return 0;
      }

    /* $$$ For demo: I'm assuming there's only one mes per x-y */
    if (ctdb_mes_find_mes_from_xy(ctdb, goal_point[X], 
				  goal_point[Y], 1, &dest_mes_key))
      num_enclosures = ctdb_mes_find_enclosures_from_xy(ctdb, dest_mes_key,
							goal_point[X], goal_point[Y],
							MAX_ENCLOSURES,
							dest_enclosure_keys);

    dest_in_mes = 0;
    for (i=0; i < num_enclosures; i++)
      if (!CTDB_MES_NODE_IS_EXTERIOR(dest_enclosure_keys[i]))
      {
	  dest_in_mes = 1;
	  break;
      }

    dest_mes_id = CTDB_MES_NODE_KEY_TO_MES_ID(dest_mes_key);

    if (!src_in_mes && !dest_in_mes)
    {
	int32 reachable = (num_enclosures == 0);

	for (i=0; i < num_enclosures; i++)
	{
	    if (CTDB_MES_NODE_IS_EXTERIOR(dest_enclosure_keys[i]))
	    {
		ctdb_lookup_enclosure_elevation(ctdb, dest_enclosure_keys[i], 
						goal_point[X], goal_point[Y],
						CTDB_ALL_POSITIVE_ANGLES,
						qual, elevs);

		
		if ((elevs[0].z - departure_point[Z]) < DEMO_CLIMB_THRESHOLD)
		{
		    reachable = TRUE;
		    break;
		}
	    }
	}
	
	if (reachable)
	{
	    /* Not travelling to or from a MES.
	     * Assuming that there is no intervening MES between
	     * your start and end points (i.e. distance between
	     * src and dest is small.
	     */
	    CTDB_MES_MAKE_NODE_KEY_INVALID(*arrival_mes);
	    VMAT3_VEC_COPY64(goal_point, arrival_point);
	    return 1;
	}
    }

    if (num_enclosures == 0)
    {
	num_enclosures = 1;
	dest_enclosure_keys[i] = CTDB_MES_KEY(src_mes_id, CTDB_MES_EXTERIOR);
    }

    land_z = ctdb_lookup_elevation(ctdb, goal_point[X], goal_point[Y]);

    closest_distance = 999999;
    for (i=0; i < num_enclosures; i++)
    {
	
	ctdb_lookup_enclosure_elevation(ctdb, dest_enclosure_keys[i], 
					goal_point[X], goal_point[Y],
					CTDB_ALL_POSITIVE_ANGLES,
					qual, elevs);
	valid_enclosures[i].enclosure = dest_enclosure_keys[i];

	if (elevs[0].z < land_z)
	  elevs[0].z = land_z;

	delta_z = elevs[0].z - departure_point[Z];
	if (delta_z < 0)
	  delta_z = -delta_z;
	
	if (delta_z < closest_distance)
	{
	    closest_distance = delta_z;
	    closest = i;
	}
	valid_enclosures[i].clear_path = 
	  ctdb_mes_ptop(ctdb, dest_mes_id, departure_point,
			departure_point[Z] + traversal_height,
			goal_point, elevs[0].z + traversal_height,
			valid_enclosures[i].location);
	if (!valid_enclosures[i].clear_path)
	  ctdb_mes_global_find_enclosure(ctdb, 
					 valid_enclosures[i].location[X],
					 valid_enclosures[i].location[Y],
					 valid_enclosures[i].location[Z],
					 &ctdb, &valid_enclosures[i].enclosure);
    }
    valid_enclosures[closest].location[Z] -= traversal_height;

    *arrival_mes = valid_enclosures[closest].enclosure;
    VMAT3_VEC_COPY64(valid_enclosures[closest].location, 
		     arrival_point);
    return 1;
}


#define CTDB_MES_VALID_NODE_KEY(ctdb, node_key)                             \
(CTDB_MES_VALID_MES_ID((ctdb), CTDB_MES_NODE_KEY_TO_MES_ID(node_key))    && \
 CTDB_MES_VALID_ENCLOSURE_ID((ctdb),                                        \
			     CTDB_MES_NODE_KEY_TO_MES_ID(node_key),         \
			     CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_key)))
int32
ctdb_mes_valid_node_key(CTDB             *ctdb,
			CTDB_MES_NODE_KEY key)
{
    if ( ctdb )
    {
	/* Optimization since in most cases the test is for an explicit
	 * invalid node key.  The remaining processing will handle the 
	 * strange case that the node key is not within range.
	 */
	if ( key == CTDB_INVALID_MES_NODE_KEY )
	  return FALSE;
	else
	  return CTDB_MES_VALID_NODE_KEY(ctdb, key);
    }
    else
      return FALSE;
}



int32
ctdb_mes_same_enclosure(
                        CTDB              *ctdb1,
                        CTDB_MES_NODE_KEY  node_key1,
                        CTDB              *ctdb2,
                        CTDB_MES_NODE_KEY  node_key2
                        )
{
    /* If both node_keys are invalid i.e. not in a MES,
     * or exterior node keys, then treat them as the same enclosure
     */
    if ((!ctdb_mes_valid_node_key(ctdb1, node_key1) || 
         CTDB_MES_NODE_IS_EXTERIOR(node_key1)) && 
        (!ctdb_mes_valid_node_key(ctdb2, node_key2) || 
         CTDB_MES_NODE_IS_EXTERIOR(node_key2)))
      return 1;

    /* If the node keys belong to the same ctdb and are equal,
     * then they are the same enclosure.
     */
    if ((ctdb1 == ctdb2) && (CTDB_MES_NODE_KEYS_EQUAL(node_key1, node_key2)))
      return 1;
    
    /* Otherwise, the node_keys do not point to the same enclosure */
    return 0;
}

