#ifndef lint
static char rcsid [] = "$RCSfile: ct_mes_boxes.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
 *   File: ct_mes_boxes.c                                                     *
 *                                                                            *
 *       Copyright 1996 by SAIC & TASC                                        *
 *                                                                            *
 *       SAIC                          TASC                                   *
 *       20 Burlington Mall Road       55 Walkers Brook Drive                 *
 *       Suite 130                     Reading, MA 01867                      *
 *       Burlington, MA 01803          617-942-2000                           *
 *       617-221-7600                                                         *
 *                                                                            *
 *       This software was developed by TASC and SAIC under U. S. Government  *
 *       contracts and may be reproduced by or for the U. S. Government       *
 *       pursuant to the copyright license under the clause                   *
 *       at DFARS 252.227-7013 (OCT 1988).                                    *
 *                                                                            *
 *       Contents: Code to handle grid-boxes for MES                          *
 *       Created:  March 1996                                                 *
 *       Author: lemabius                                                     *
 *       Remarks:                                                             *
 *                                                                            *
 *****************************************************************************/

#include "libctdb_local.h"
#include <libvecmat.h>
#include <stdio.h>
#include <stdstring.h>

/*#define CTDB_BOXES_VERBOSE*/
#define TOLERANCE 1

#define transform_vertex(int32_ver, f64_ver) \
    (f64_ver)[X] = (float64)int32_ver[X], \
    (f64_ver)[Y] = (float64)int32_ver[Y], \
    (f64_ver)[Z] = (float64)int32_ver[Z];

static void 
calculate_triangle_normal(int32              vertices[][XYZ], 
			  CTDB_MES_TRIANGLE *triangle)
{
    float64 vertex0[XYZ];
    float64 vertex1[XYZ];
    float64 vertex2[XYZ];
    
    float64 side0[XYZ];
    float64 side1[XYZ];
    
    float64 normal[XYZ];
    
    transform_vertex ( vertices[triangle->vertex_ids[0]], vertex0 );
    transform_vertex ( vertices[triangle->vertex_ids[1]], vertex1 );
    transform_vertex ( vertices[triangle->vertex_ids[2]], vertex2 );
    
    vmat3_vec_sub64 ( vertex1, vertex0, side0 );
    vmat3_vec_sub64 ( vertex2, vertex0, side1 );
    vmat3_cross_prod64 ( side0, side1, normal );
    vmat3_unit64 ( normal, triangle->normal );
}

static void 
calculate_grid_indices(int32  grid_index,
		       uint16 dimensions_y,
		       uint16 dimensions_z,
		       int32  grid_indices[XYZ])
{
    int32 index;
    
    if ( grid_index > 0 )
    {
	grid_indices[Z] = grid_index % dimensions_z;
	index = grid_index / dimensions_z;
	grid_indices[Y] = index % dimensions_y;
	grid_indices[X] = index / dimensions_y;
    }
    else
    {
	grid_indices[X] = 0;
	grid_indices[Y] = 0;
	grid_indices[Z] = 0;
    }
}

static int32 
subdivided(CTDB_MES_GRID_BOX *box)
{
    return (   ( box->dimensions_x > 1 )
	    || ( box->dimensions_y > 1 )
	    || ( box->dimensions_z > 1 ) );
}

static int32 
last_box(CTDB_MES_GRID_BOX *boxes, 
	 int32              box_id, 
	 int32              face )
{
    int32  grid_indices[XYZ];
    int32  grid_index = boxes[box_id].grid_index;
    uint16 dimensions[XYZ];
    dimensions[X] = boxes[boxes[box_id].outer_box_id].dimensions_x;
    dimensions[Y] = boxes[boxes[box_id].outer_box_id].dimensions_y;
    dimensions[Z] = boxes[boxes[box_id].outer_box_id].dimensions_z;

    calculate_grid_indices ( grid_index, dimensions[Y], dimensions[Z],
			    grid_indices );

    if ( face < 3 )
    {
	return ( grid_indices[face] == 0 );
    }
    else
    {
	return ( grid_indices[face-3] == (dimensions[face-3]-1) );
    }
}

static void 
advance_to_next_box(CTDB_MES_GRID_BOX *boxes,
		    int32             *box_id,
		    int32             *face,
		    float64            normalized_position[XYZ])
{
    uint16 dimensions[XYZ];
    dimensions[X] = boxes[boxes[*box_id].outer_box_id].dimensions_x;
    dimensions[Y] = boxes[boxes[*box_id].outer_box_id].dimensions_y;
    dimensions[Z] = boxes[boxes[*box_id].outer_box_id].dimensions_z;

    if ( *face < 3 )
    {
	if ( *face == 0 )
	{
	    *box_id -= dimensions[Y]*dimensions[Z];
	    normalized_position[X] = 1.0;
	}
	else if ( *face == 1 )
	{
	    *box_id -= dimensions[Z];
	    normalized_position[Y] = 1.0;
	}
	else
	{
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "advance (down) from %d", *box_id );
#endif
	    *box_id -= 1;
#ifdef CTDB_BOXES_VERBOSE
	    printf ( " to %d\n", *box_id );
#endif
	    normalized_position[Z] = 1.0;
	}
	*face += 3;
    }
    else
    {
	*face -= 3;
	if ( *face == 0 )
	{
	    *box_id += dimensions[Y]*dimensions[Z];
	    normalized_position[X] = 0.0;
	}
	else if ( *face == 1 )
	{
	    *box_id += dimensions[Z];
	    normalized_position[Y] = 0.0;
	}
	else
	{
	    *box_id += 1;
	    *box_id++;
	    normalized_position[Z] = 0.0;
	}
    }
}

static void 
calculate_box_id(CTDB_MES_GRID_BOX *boxes,
		 float64            position[XYZ],
		 float64            normalized_position[XYZ],
		 int32             *box_id)
{
    int32   axis;
    int32   x_index;
    int32   y_index;
    int32   z_index;
    float64 value;
    int32   index;

    /* based on face and normalized position, select inner box index
       and modify normalized position */
    
    if ( ( position[X] < (float64)(boxes[0].left_front_bottom_corner[X]) )
	|| ( position[Y] < (float64)(boxes[0].left_front_bottom_corner[Y]) )
	|| ( position[Z] < (float64)(boxes[0].left_front_bottom_corner[Z]) )
	|| ( position[X] > (float64)(boxes[0].right_back_top_corner[X]) )
	|| ( position[Y] > (float64)(boxes[0].right_back_top_corner[Y]) )
	|| ( position[Z] > (float64)(boxes[0].right_back_top_corner[Z]) ) )
    {
	*box_id = -1;
	return;
    }
    for ( axis=0; axis<XYZ; axis++ )
    {
	normalized_position[axis]
	  = ( position[axis]
	     - (float64)(boxes[0].left_front_bottom_corner[axis]) )
	    / ( (float64)(boxes[0].right_back_top_corner[axis])
	       - (float64)(boxes[0].left_front_bottom_corner[axis]) );
	*box_id = 0;
    }
    while ( subdivided ( &(boxes[*box_id]) ) )
    {
	value = normalized_position[X] * boxes[*box_id].dimensions_x;
	x_index = value;
	if ( x_index >= boxes[*box_id].dimensions_x )
	{
	    x_index = boxes[*box_id].dimensions_x - 1;
	}
	normalized_position[X] = value - x_index;
	
	value = normalized_position[Y] * boxes[*box_id].dimensions_y;
	y_index = value;
	if ( y_index >= boxes[*box_id].dimensions_y )
	{
	    y_index = boxes[*box_id].dimensions_y - 1;
	}
	normalized_position[Y] = value - y_index;
	
	value = normalized_position[Z] * boxes[*box_id].dimensions_z;
	z_index = value;
	if ( z_index >= boxes[*box_id].dimensions_z )
	{
	    z_index = boxes[*box_id].dimensions_z - 1;
	}
	normalized_position[Z] = value - z_index;
	index = z_index + boxes[*box_id].dimensions_z
	  * ( y_index + boxes[*box_id].dimensions_y * x_index );
	*box_id = boxes[*box_id].box_ids[index];
    }
}

static void 
transform_to_inner_box(CTDB_MES_GRID_BOX box,
		       int32             face,
		       float64           normalized_position[XYZ],
		       int32            *box_id)
{
    int32   x_index;
    int32   y_index;
    int32   z_index;
    float64 value;
    int32   index;
	   
    /* based on face and normalized position, select inner box index
       and modify normalized position */
    
    if ( face == X )
    {
	x_index = 0;
    }
    else if ( face == (X+3) )
    {
	x_index = box.dimensions_x - 1;
    }
    else
    {
	value = normalized_position[X] * box.dimensions_x;
	x_index = value;
	if ( x_index >= box.dimensions_x )
	{
	    x_index = box.dimensions_x - 1;
	}
	normalized_position[X] = value - x_index;
    }
    
    if ( face == Y )
    {
	y_index = 0;
    }
    else if ( face == (Y+3) )
    {
	y_index = box.dimensions_y - 1;
    }
    else
    {
	value = normalized_position[Y] * box.dimensions_y;
	y_index = value;
	if ( y_index >= box.dimensions_y )
	{
	    y_index = box.dimensions_y - 1;
	}
	normalized_position[Y] = value - y_index;
    }
    
    if ( face == Z )
    {
	z_index = 0;
    }
    else if ( face == (Z+3) )
    {
	z_index = box.dimensions_z - 1;
    }
    else
    {
	value = normalized_position[Z] * box.dimensions_z;
	z_index = value;
	if ( z_index >= box.dimensions_z )
	{
	    z_index = box.dimensions_z - 1;
	}
	normalized_position[Z] = value - z_index;
    }
    index = z_index + box.dimensions_z
      * ( y_index + box.dimensions_y * x_index );
#ifdef CTDB_BOXES_VERBOSE
    printf ( "transform to inner -- index = %d %d %d %d\n",
            x_index, y_index, z_index, index );
#endif
    *box_id = box.box_ids[index];
}

static void 
transform_to_parent(CTDB_MES_GRID_BOX *boxes,
		    int32              face,
		    int32             *box_id,
		    float64            normalized_position[XYZ])
{
    /* Transforms to box which is in outer_box_id */
    
    int32 grid_index = boxes[*box_id].grid_index;
    int32 grid_indices[XYZ];
    int32 axis;
    uint16 dimensions[XYZ];

    *box_id = boxes[*box_id].outer_box_id;
    dimensions[X] = boxes[*box_id].dimensions_x;
    dimensions[Y] = boxes[*box_id].dimensions_y;
    dimensions[Z] = boxes[*box_id].dimensions_z;

    calculate_grid_indices ( grid_index, dimensions[Y], dimensions[Z],
			    grid_indices);
    
    for ( axis=0; axis<XYZ; axis++ )
    {
	if ( axis != (face%XYZ) )
	{
	    normalized_position[axis]
	      = ( normalized_position[axis] + grid_indices[axis] )
		/ dimensions[axis];
	}
    }
}

static void 
calculate_starting_point(CTDB_MES_GRID_BOX box,
			 float64           ray_origin[XYZ],
			 float64           ray_direction[XYZ],
			 float64           normalized_position[XYZ],
			 int32            *face)
{
    /* Calculates ray intercept with edge of outer most box and face of
     * intercept.
     * face = 0 left      3 right
     * 1 front     4 back
     * 2 bottom    5 top
     * -1 outside   6 inside
     * Ignores "face" side
     * Uses intercept with shortest distance 
     */
    
    float64 maximum_ray;
    float64 tolerance;
    float64 minimum_distance;
    int32   entry_face;
    float64 distance;
    int32   axis, axis1, axis2;
    float64 position1, position2;
    
    maximum_ray = fabs(ray_direction[X]);
    for ( axis=1; axis<XYZ; axis++ )
    {
	if ( fabs(ray_direction[axis]) > maximum_ray )
	{
	    maximum_ray = fabs(ray_direction[axis]);
	}
    }
    tolerance = TOLERANCE / maximum_ray;
    
    minimum_distance = 1.0 + 2.0 * tolerance;
    for ( axis=0; axis<XYZ; axis++ )
    {
	if ( ray_direction[axis] > TOLERANCE )
	{
	    distance = ( (float64)(box.left_front_bottom_corner[axis])
			- ray_origin[axis] )
	      / ray_direction[axis];
	    axis1 = ( axis + 1 ) % XYZ;
	    position1 = ( ray_origin[axis1] + distance * ray_direction[axis1]
			 - (float64)(box.left_front_bottom_corner[axis1]) )
	      / ( (float64)(box.right_back_top_corner[axis1])
		 - (float64)(box.left_front_bottom_corner[axis1]) );
	    axis2 = ( axis + 2 ) % XYZ;
	    position2 = ( ray_origin[axis2] + distance * ray_direction[axis2]
			 - (float64)(box.left_front_bottom_corner[axis2]) )
	      / ( (float64)(box.right_back_top_corner[axis2])
		 - (float64)(box.left_front_bottom_corner[axis2]) );
	    if ( ( position1 >= 0.0 ) && ( position1 <= 1.0 )
		&& ( position2 >= 0.0 ) && ( position2 <= 1.0 )
		&& ( distance <= (1.0+tolerance) ) )
	    {
		if ( distance < minimum_distance )
		{
		    minimum_distance = distance;
		    entry_face = axis;
		    normalized_position[axis] = 0.0;
		    normalized_position[axis1] = position1;
		    normalized_position[axis2] = position2;
		}
	    }
	}
	else if ( ray_direction[axis] < (-TOLERANCE) )
	{
	    distance = ( (float64)(box.right_back_top_corner[axis])
			- ray_origin[axis] )
	      / ray_direction[axis];
	    axis1 = ( axis + 1 ) % XYZ;
	    position1 = ( ray_origin[axis1] + distance * ray_direction[axis1]
			 - (float64)(box.left_front_bottom_corner[axis1]) )
	      / ( (float64)(box.right_back_top_corner[axis1])
		 - (float64)(box.left_front_bottom_corner[axis1]) );
	    axis2 = ( axis + 2 ) % XYZ;
	    position2 = ( ray_origin[axis2] + distance * ray_direction[axis2]
			 - (float64)(box.left_front_bottom_corner[axis2]) )
	      / ( (float64)(box.right_back_top_corner[axis2])
		 - (float64)(box.left_front_bottom_corner[axis2]) );
	    if ( ( position1 >= 0.0 ) && ( position1 <= 1.0 )
		&& ( position2 >= 0.0 ) && ( position2 <= 1.0 )
		&& ( distance <= (1.0+tolerance) ) )
	    {
		if ( distance < minimum_distance )
		{
		    minimum_distance = distance;
		    entry_face = axis+3;
		    normalized_position[axis] = 1.0;
		    normalized_position[axis1] = position1;
		    normalized_position[axis2] = position2;
		}
	    }
	}
    }
    
    if ( minimum_distance <= 1.0 )
    {
	*face = entry_face;
    }
    else
    {
	*face = -1;
    }
}

static void 
calculate_box_intercept (CTDB_MES_GRID_BOX box,
			 float64           ray_origin[XYZ],
			 float64           ray_direction[XYZ],
			 float64           normalized_position[XYZ],
			 int32            *face,
			 float64          *distance)
{
    /* Calculates ray intercept with edge of box and face of intercept.
     * face = 0 left      3 right
     * 1 front     4 back
     * 2 bottom    5 top
     * -1 outside   6 inside
     * Ignores "face" side
     * Uses intercept with shortest distance 
     */
    
    int32   exit_face;
    float64 minimum_distance;
    int32   axis;
    float64 edge_distance;
    static float64 large_number=1e99;
    
    exit_face = -1;
    minimum_distance = large_number;
    for ( axis=0; axis<XYZ; axis++ )
    {
	if ( ray_direction[axis] > TOLERANCE )
	{
	    edge_distance = ( (float64)(box.right_back_top_corner[axis])
			     - ray_origin[axis] ) / ray_direction[axis];
	    if ( ( exit_face == -1 ) || ( edge_distance < minimum_distance ) )
	    {
		minimum_distance = edge_distance;
		exit_face = axis+3;
	    }
	}
	else if ( ray_direction[axis] < (-TOLERANCE) )
	{
	    edge_distance = ( (float64)(box.left_front_bottom_corner[axis])
			     - ray_origin[axis] ) / ray_direction[axis];
	    if ( ( exit_face == -1 ) || ( edge_distance < minimum_distance ) )
	    {
		minimum_distance = edge_distance;
		exit_face = axis;
	    }
	}
    }
    
    for ( axis=0; axis<XYZ; axis++ )
    {
	if ( exit_face == axis )
	{
	    normalized_position[axis] = 0.0;
	}
	else if ( exit_face == (axis+3) )
	{
	    normalized_position[axis] = 1.0;
	}
	else
	{
	    normalized_position[axis]
	      = ( ray_origin[axis]
		 - (float64)(box.left_front_bottom_corner[axis])
		 + minimum_distance * ray_direction[axis] )
		/ ( (float64)(box.right_back_top_corner[axis])
		   - (float64)(box.left_front_bottom_corner[axis]) );
	    if ( normalized_position[axis] < 0.0 )
	    {
		normalized_position[axis] = 0.0;
	    }
	    else if ( normalized_position[axis] > 1.0 )
	    {
		normalized_position[axis] = 1.0;
	    }
	}
    }
    *face = exit_face;
    *distance = minimum_distance;
}

static void 
calculate_triangle_intercept(float64              ray_origin[XYZ],
			     float64              ray_direction[XYZ],
			     CTDB_MES_TRIANGLE    triangle,
			     int32                vertices[][XYZ],
			     float64             *intercept_distance,
			     int32               *triangle_intercept)
{
    /* If ray and triangle normal have a positive or zero dot product
     * ray does not intercept triangle.
     * If they have a negative dot product check if intercept point
     * lies inside triangle.
     * If ray intercepts triangle return
     * cosine of angle bewteen ray and negative normal
     * distance from ray origin to intercept point
     * If ray does not intercept triangle return
     * cosine = 0.0 
     */
    
    float64 v0[XYZ];
    float64 v1[XYZ];
    float64 v2[XYZ];
    
    float64 relative_ray_origin[XYZ];
    float64 numerator;
    float64 denomenator;
    float64 distance;
    float64 relative_ray_direction[XYZ];
    float64 intercept[XYZ];
    
    float64 side01[XYZ];
    float64 side12[XYZ];
    float64 side20[XYZ];
    float64 side01_normal[XYZ];
    float64 side12_normal[XYZ];
    float64 side20_normal[XYZ];
    float64 relative_intercept0[XYZ];
    float64 relative_intercept1[XYZ];
    
    transform_vertex ( vertices[triangle.vertex_ids[0]], v0 );
    transform_vertex ( vertices[triangle.vertex_ids[1]], v1 );
    transform_vertex ( vertices[triangle.vertex_ids[2]], v2 );
    
    vmat3_vec_sub64 ( ray_origin, v0, relative_ray_origin );
    numerator = vmat3_dot_prod64 ( triangle.normal, relative_ray_origin );
    denomenator = vmat3_dot_prod64 ( triangle.normal, ray_direction );
    if ( fabs(denomenator) > fabs(numerator) )
    {
	distance = - numerator / denomenator;
#ifdef CTDB_BOXES_VERBOSE
	printf ( "distance = %lf\n", distance );
#endif
    }
    else
    {
	distance = -1.0;
    }
    if ( ( distance > 0.0 ) && ( distance < 1.0 ) )
    {
	vmat3_scal_vec_mul64 ( distance, ray_direction, 
			       relative_ray_direction );
	vmat3_vec_add64 ( ray_origin, relative_ray_direction, intercept );
#ifdef CTDB_BOXES_VERBOSE
	printf ( "intercept = %10.3f %10.3f %10.3f\n",
		intercept[0], intercept[1], intercept[2] );
#endif
	
	vmat3_vec_sub64 ( v1, v0, side01 );
	vmat3_cross_prod64 ( triangle.normal, side01, side01_normal );
	vmat3_vec_sub64 ( intercept, v0, relative_intercept0 );
	if ( vmat3_dot_prod64 ( side01_normal, relative_intercept0 ) < 0.0 )
	{
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "outside 0-1\n" );
#endif
	    *intercept_distance = 0.0;
	    *triangle_intercept = FALSE;
	    return;
	}
	
	vmat3_vec_sub64 ( v2, v1, side12 );
	vmat3_cross_prod64 ( triangle.normal, side12, side12_normal );
	vmat3_vec_sub64 ( intercept, v1, relative_intercept1 );
	if ( vmat3_dot_prod64 ( side12_normal, relative_intercept1 ) < 0.0 )
	{
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "outside 1-2\n" );
#endif
	    *intercept_distance = 0.0;
	    *triangle_intercept = FALSE;
	    return;
	}
	
	vmat3_vec_sub64 ( v0, v2, side20 );
	/*
	  vmat3_cross_prod64 ( triangle.normal, side20, side20_normal );
	  vmat3_vec_sub64 ( intercept, v2, relative_intercept2 );
	  if ( vmat3_dot_prod64 ( side20_normal, relative_intercept2 ) < 0.0 )
	  */
	vmat3_vec_add64 ( side01_normal, side12_normal, side20_normal );
	if ( vmat3_dot_prod64 ( side20_normal, relative_intercept0 ) > 0.0 )
	{
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "outside 2-0\n" );
#endif
	    *intercept_distance = 0.0;
	    *triangle_intercept = FALSE;
	    return;
	}
#ifdef CTDB_BOXES_VERBOSE
	printf ( "inside triangle\n" );
#endif
	*intercept_distance = distance;
	*triangle_intercept = TRUE;
    }
    else
    {
	*intercept_distance = 0.0;
	*triangle_intercept = FALSE;
    }
}


void 
ctdb_calculate_triangle_normals(CTDB_MES_VOLUME *mes)
{
    int32               id;
    CTDB_MES_TEMPLATE  *template = mes->template;
    
    for ( id=0; id<(template->num_triangles); id++ )
    {
	calculate_triangle_normal ( template->all_vertices,
				   &(template->all_triangles[id]) );
    }
}

#define ADD_INTERSECTION(_tri, _dist, array, size)                   \
{                                                                    \
   if (n_alloc_intercept < ((size)+1))	                             \
   {                                                                 \
       n_alloc_intercept = (size)+10;                                \
       intercept = (array) = (CTDB_MES_INTERSECTION *)               \
                                           ctdb_alloc ((array),      \
                                       n_alloc_intercept *           \
				       sizeof(CTDB_MES_INTERSECTION),\
				       "intersection list");         \
   }                                                                 \
                                                                     \
   (array)[(size)].triangle = (_tri);                                \
   (array)[(size)].distance = (_dist);                               \
   (size)++;                                                         \
}

static int32
cmp_mes_inter(CTDB_MES_INTERSECTION *data0, 
	      CTDB_MES_INTERSECTION *data1)
{
    if (data0->distance < data1->distance)
      return -1;
    else if (data1->distance < data0->distance)
      return 1;
    else
      return 0;
}

/* Return number of intercepted triangles */
void 
ctdb_mes_line_intersection (CTDB_MES_VOLUME        *mes,
			    float64                 ray_origin[XYZ],
			    float64                 ray_destination[XYZ],
			    int32                   all_intercepts,
			    int32                   enclosure_id,
			    int32                   check_mobility,
			    int32                   check_visibility,
			    int32                   aperture_intercepts,
			    int32                  *n_intercepts,
			    CTDB_MES_INTERSECTION **intercepts)    
{
    static CTDB_MES_INTERSECTION *intercept=NULL;
    static int32             n_alloc_intercept=0;
    static int8         *triangles_checked=NULL;
    static int32         n_alloc_triangles_checked=0;
    int32               (*vertices)[XYZ];
    CTDB_MES_TRIANGLE    *triangles;
    CTDB_MES_APERTURE    *apertures;
    CTDB_MES_GRID_BOX    *boxes;
    CTDB_MES_TEMPLATE    *template = mes->template;
    CTDB_MES_ENCLOSURE   *enclosure;
    int32                blocked;
    
    /* Box Coordinates - x(0) left(0.0) to right(1.0)   (x) = Coordinate Index
     * - y(1) front(0.0) to back(1.0)   (x.x) = Normalized Value
     * - z(2) bottom(0.0) to top(1.0)  
     */
    
     int32       offset_index;
    
    float64     ray_direction[XYZ];
    float64     normalized_position[XYZ]; /* position of ray within box */
    int32       box_id = 0;               /* Outer-most or defining box */
    
    int32       face = 0;
    
    int32       triangle_intercept;
    
    int32       triangle_index;
    int32       triangle_id;
    int32       num_triangles;
    int32      *wall_triangle_ids;
    
    int32       aperture_index;
    uint16      aperture_id;
    float64     distance;
    
    (*n_intercepts) = 0;
    (*intercepts) = intercept;
    vertices   = template->all_vertices;
    triangles  = template->all_triangles;
    apertures  = template->all_apertures;
    boxes      = template->all_boxes;
    
    /* If the allocated space is not large enough then reallocate the 
     * memeory block to allow storage of the triangle data.
     */
    if ( n_alloc_triangles_checked < template->num_triangles)
    {
	n_alloc_triangles_checked = template->num_triangles;
	triangles_checked = (int8 *)ctdb_alloc(triangles_checked,
					       n_alloc_triangles_checked *
					       sizeof(int8), 
					       "Checked triangles list" );
    }
    
    if ((enclosure_id == CTDB_MES_ALL_ENCLOSURES)
	|| ( enclosure_id >= template->num_enclosures ) )
      enclosure = NULL;
    else
      enclosure = &(template->all_enclosures[enclosure_id]);

    /* Initialize the flag array to all TRUE or all FALSE */
    memset(triangles_checked, (enclosure != NULL), template->num_triangles);
    
    if ( enclosure )
    {
	offset_index = enclosure->first_wall_triangle_id;
	num_triangles = enclosure->num_triangles;
	memset(&triangles_checked[offset_index], FALSE, num_triangles);
	
	if ( aperture_intercepts )
	{
	    for ( aperture_index=0; aperture_index<(enclosure->num_apertures);
		 aperture_index++ )
	    {
		aperture_id = enclosure->aperture_ids[aperture_index];
		offset_index = apertures[aperture_id].first_triangle_id;
		num_triangles = apertures[aperture_id].num_triangles;
		memset(&triangles_checked[offset_index], FALSE,
		       num_triangles);
	    }
	}
    }
    
    vmat3_vec_sub64 ( ray_destination, ray_origin, ray_direction );
    
    /* Calculate normalized position and face of outer-most
       box where ray first intercepts */
    calculate_box_id (boxes, ray_origin, normalized_position, &box_id );
#ifdef CTDB_BOXES_VERBOSE
    printf ( "initial box id = %d\n", box_id );
#endif
    if ( box_id < 0 )
    {
	box_id = 0;
	
	calculate_starting_point ( boxes[box_id], ray_origin, ray_direction,
				  normalized_position, &face );
	if ( face < 0 )
	  return;
    }
    else
    {
	face = 6;
    }
    
#ifdef CTDB_BOXES_VERBOSE
    printf ( "box id =       face =  " );
    printf ( "   position = %10.3f %10.3f %10.3f\n",
            ray_origin[X], ray_origin[Y], ray_origin[Z] );
#endif
    
    while ( 1 )
    {
#ifdef CTDB_BOXES_VERBOSE
	printf ( "box id = %3d   face = %1d", box_id, face );
	printf ( "   position = %10.3f %10.3f %10.3f\n",
		boxes[box_id].left_front_bottom_corner[X]
		+ normalized_position[X]
		* ( boxes[box_id].right_back_top_corner[X]
		   - boxes[box_id].left_front_bottom_corner[X] ),
		boxes[box_id].left_front_bottom_corner[Y]
		+ normalized_position[Y]
		* ( boxes[box_id].right_back_top_corner[Y]
		   - boxes[box_id].left_front_bottom_corner[Y] ),
		boxes[box_id].left_front_bottom_corner[Z]
		+ normalized_position[Z]
		* ( boxes[box_id].right_back_top_corner[Z]
		   - boxes[box_id].left_front_bottom_corner[Z] ) );
#endif
	if ( subdivided ( &(boxes[box_id]) ) )
	{        /* Box contains subboxes */
	    
	    /* Transform to first level of subboxes.  Outer loop repeats
	       this until a box with the triangles is reached. */
	    
	    transform_to_inner_box ( boxes[box_id],
				    face, normalized_position,
				    &box_id );
	}
	else
	{        /* Box contains triangles not subboxes */
	    
	    /* Calculate which triangle is intercepted by ray and cosine
	       of Intercept -  This is more than is needed but would allow
	       for of displaying the result in an image */
	    for ( aperture_index=-1; 
		 aperture_index<boxes[box_id].num_apertures;
		 aperture_index++ )
	    {
		if ( aperture_index == -1 )
		{
		    num_triangles = boxes[box_id].num_triangles;
		    wall_triangle_ids = boxes[box_id].wall_triangle_ids;
		}
		else
		{
		    aperture_id = boxes[box_id].aperture_ids[aperture_index];
		    
		    if ((aperture_intercepts) ||
			((check_visibility) &&
			 (apertures[aperture_id].visible == 0 )) ||
			((check_mobility) && 
			 (apertures[aperture_id].open == 0)))
		      blocked = TRUE;
		    else
		      blocked = FALSE;
		    
		    if (blocked)
		    {
			num_triangles = apertures[aperture_id].num_triangles;
#ifdef CTDB_BOXES_VERBOSE
			printf ( "checking aperture %d\n", aperture_id );
#endif
		    }
		    else
		    {
			/* do not check aperture triangles as they do not 
			 * block 
			 */
			num_triangles = 0;
		    }
		}
		
		for ( triangle_index=0; triangle_index<num_triangles;
		     triangle_index++ )
		{
		    if ( aperture_index == -1 )
		    {
			triangle_id = wall_triangle_ids[triangle_index];
		    }
		    else
		    {
			triangle_id = apertures[aperture_id].first_triangle_id
			  + triangle_index;
		    }
		    if ( ! triangles_checked[triangle_id] )
		    {
#ifdef CTDB_BOXES_VERBOSE
			printf ( "checking triangle %d\n", triangle_id );
#endif
			calculate_triangle_intercept(ray_origin, 
						     ray_direction,
						     triangles[triangle_id],
						     vertices,
						     &distance,
						     &triangle_intercept );
			
			triangles_checked[triangle_id] = TRUE;
			
			if ( triangle_intercept )
			{
			    if (n_alloc_intercept < ((*n_intercepts)+1))
			    { 
				n_alloc_intercept = (*n_intercepts)+10; 
				intercept = (*intercepts) = 
				  (CTDB_MES_INTERSECTION *) 
				    ctdb_alloc ((*intercepts),
						n_alloc_intercept * 
						sizeof(CTDB_MES_INTERSECTION),
						"intersection list");
			    }
			    (*intercepts)[(*n_intercepts)].triangle = 
			      triangle_id;
			    (*intercepts)[(*n_intercepts)].distance = 
			      distance;
			    (*n_intercepts)++;
			}
		    }
		}
	    }

	    if ( (*n_intercepts) ) 
	      qsort(intercept, (*n_intercepts), sizeof(CTDB_MES_INTERSECTION),
		    (CTDB_QSORT)cmp_mes_inter);
	    
	    if ((!all_intercepts) && ((*n_intercepts) > 0))
	      return;
	    
	    /* Calculate where ray leaves box */
	    calculate_box_intercept(boxes[box_id], ray_origin, ray_direction,
				    normalized_position, &face, &distance );
	    
	    if ( distance >= 1.0 )
	      return;
	    
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "box id = %3d I face = %1d", box_id, face );
	    printf ( "   position = %10.3f %10.3f %10.3f %10.3f\n",
		    
		    boxes[box_id].left_front_bottom_corner[X]
		    + normalized_position[X]
		    * ( boxes[box_id].right_back_top_corner[X]
		       - boxes[box_id].left_front_bottom_corner[X] ),
		    
		    boxes[box_id].left_front_bottom_corner[Y]
		    + normalized_position[Y]
		    * ( boxes[box_id].right_back_top_corner[Y]
		       - boxes[box_id].left_front_bottom_corner[Y] ),
		    
		    boxes[box_id].left_front_bottom_corner[Z]
		    + normalized_position[Z]
		    * ( boxes[box_id].right_back_top_corner[Z]
		       - boxes[box_id].left_front_bottom_corner[Z] ),
		    
		    distance );
#endif
	    while ( box_id > 0 )   /* box is not outer-most box */
	    {
		if ( last_box ( boxes, box_id, face ) )
		{                      /* face of box is on outer-box face */
		    /* transform to next outer box */
		    transform_to_parent ( boxes, face, &box_id,
					 normalized_position );
		}
		else                   /* Move to next sub-box of outer-box */
		{
		    advance_to_next_box ( boxes, &box_id, &face,
					 normalized_position );
		    break;
		}
	    }
	}

	if ( box_id == 0 ) 
	  break;
    }
#ifdef CTDB_BOXES_VERBOSE
    printf ( "box id = %3d   face = %1d", box_id, face );
    printf ( "   position = %10.3f %10.3f %10.3f\n",
            boxes[box_id].left_front_bottom_corner[X] + normalized_position[X]
	    * ( boxes[box_id].right_back_top_corner[X]
	       - boxes[box_id].left_front_bottom_corner[X] ),
            boxes[box_id].left_front_bottom_corner[Y] + normalized_position[Y]
	    * ( boxes[box_id].right_back_top_corner[Y]
	       - boxes[box_id].left_front_bottom_corner[Y] ),
            boxes[box_id].left_front_bottom_corner[Z] + normalized_position[Z]
	    * ( boxes[box_id].right_back_top_corner[Z]
	       - boxes[box_id].left_front_bottom_corner[Z] ) );
#endif
}

static void 
truncate_polygon (int32  face,
		  int32  limit,
		  int32  *points,
		  int32  polygon[][XYZ])
{
    int32   limit_axis = face % 3;
    int32   first, last;             /* points within limits */
    int32   p;
    int32   count;
    int32   after_last;
    int32   axis;
    float64 scale;
    int32   new_points[10][XYZ];
    int32   before_first;
    
    first = -1;
    for ( p=0; p<(*points); p++ )
    {
	if ( ( ( face < 3 ) && ( polygon[p][limit_axis] >= limit ) )
	    || ( ( face > 2 ) && ( polygon[p][limit_axis] <= limit ) ) )
	{
	    if ( first == -1 )
	    {
		first = p;
		last = p;
	    }
	    else if ( first == -2 )
	    {
		first = p;
		break;
	    }
	    else
	    {
		last = p;
	    }
	}
	else if ( first > 0 )
	{
	    break;
	}
	else if ( first == 0 )
	{
	    first = -2;
	}
    }
    if ( first == -2 )
    {
	first = 0;
    }
    
    if ( first == -1 )
    {
	*points = 0;
	return;
    }
    else if ( ( first == 0 ) && ( last == ((*points)-1) ) )
    {
	return;
    }
    
    count = 0;
    if ( polygon[last][limit_axis] != limit )
    {
	after_last = (last+1) % (*points);
	scale = (float64)( limit - polygon[last][limit_axis] )
	  / (float64)( polygon[after_last][limit_axis]
		      - polygon[last][limit_axis] );
	for ( axis=0; axis<XYZ; axis++ )
	{
	    if ( axis == limit_axis )
	    {
		new_points[count][axis] = limit;
	    }
	    else
	    {
		new_points[count][axis] = polygon[last][axis]
                  + ( polygon[after_last][axis] - polygon[last][axis] ) * 
		  scale;
	    }
	}
	count++;
    }
    if ( polygon[first][limit_axis] != limit )
    {
	if ( first > 0 )
	{
	    before_first = first - 1;
	}
	else
	{
	    before_first = (*points) - 1;
	}
	scale = (float64)( limit - polygon[first][limit_axis] )
	  / (float64)( polygon[before_first][limit_axis]
		      - polygon[first][limit_axis] );
	for ( axis=0; axis<XYZ; axis++ )
	{
	    if ( axis == limit_axis )
	    {
		new_points[count][axis] = limit;
	    }
	    else
	    {
		new_points[count][axis] = polygon[first][axis]
		  + ( polygon[before_first][axis] - polygon[first][axis] ) * 
		  scale;
	    }
	}
	count++;
    }
    
    if ( first > last )  /* append split points onto end of new points */
    {
	for ( p=first; p<(*points); p++ )
	{
	    for ( axis=0; axis<XYZ; axis++ )
	    {
		new_points[count][axis] = polygon[p][axis];
	    }
	    count++;
	}
    }
    
    *points = 0;
    for ( p=first; p<=last; p++ )
    {
	for ( axis=0; axis<XYZ; axis++ )
	{
	    polygon[(*points)][axis] = polygon[p][axis];
	}
	*points += 1;
    }
    for ( p=0; p<count; p++ )
    {
	for ( axis=0; axis<XYZ; axis++ )
	{
	    polygon[(*points)][axis] = new_points[p][axis];
	}
	*points += 1;
    }
}

static void 
assign_triangles (int32                 vertices[][XYZ],
		  CTDB_MES_TRIANGLE    *triangles,
		  uint16                num_triangles,
		  int32                *triangle_ids,
		  int32                 right_back_top_corner[XYZ],
		  int32                 left_front_bottom_corner[XYZ],
		  uint16               *num_assigned_triangles,
		  int32                *assigned_triangle_ids)
{
    int32             count;
    int32             index;
    CTDB_MES_TRIANGLE triangle;
    int32             v;
    int32             outside;
    int32             axis;
    int32             shape[10][XYZ];
    int32             points;
    int32             face;
    
    count = 0;
    for ( index=0; index<num_triangles; index++ )
    {
	triangle = triangles[triangle_ids[index]];
	for ( v=0; v<3; v++ )
	{
	    outside = 0;
	    for ( axis=0; axis<XYZ; axis++ )
	    {
		shape[v][axis] = vertices[triangle.vertex_ids[v]][axis];     
		if ( ( shape[v][axis] > 
		       (right_back_top_corner[axis]+TOLERANCE) )
		    || ( shape[v][axis]
			< (left_front_bottom_corner[axis]-TOLERANCE) ) )
		{
		    outside = 1;
		}
	    }
	    if ( outside == 0 )
	    {
		break;
	    }
	}
	if ( outside == 1 )
	{
	    points = 3;
	    outside = 0;
	    for ( face=0; face<6; face++ )
	    {
		if ( face < 3 )
		{
		    truncate_polygon ( face, left_front_bottom_corner[face],
				      &points, shape );
		}
		else
		{
		    truncate_polygon ( face, right_back_top_corner[face-3],
				      &points, shape );
		}
		if ( points == 0 )
		{
		    outside = 1;
		    break;
		}
		/*
		  else
		  {
		  if ( face < 3 )
		  {
                  printf ( "face = %d  limit = %d  points = %d\n",
		  face, left_front_bottom_corner[face], points );
		  }
		  else
		  {
                  printf ( "face = %d  limit = %d  points = %d\n",
		  face, right_back_top_corner[face-3], points );
		  }
		  for ( v=0; v<points; v++ )
		  {
                  printf ( "%8d %8d %8d\n", shape[v][X],
		  shape[v][Y], shape[v][Z] );
		  }
		  }
		  */
	    }
	}
	if ( outside == 0 )
	{
	    assigned_triangle_ids[count++] = triangle_ids[index];
	}
    }
    *num_assigned_triangles = count;
}

static void 
assign_apertures(CTDB_MES_APERTURE    *apertures,
		 uint16                num_apertures,
		 uint16               *aperture_ids,
		 int32                 num_triangles,
		 int32                *wall_triangle_ids,
		 uint16               *num_assigned_apertures,
		 uint16               *assigned_aperture_ids)
{
    int32    count;
    int32    index;
    CTDB_MES_APERTURE aperture;
    int32    i;
    int32    offset;
    
    count = 0;
    for ( index=0; index<num_apertures; index++ )
    {
	aperture = apertures[aperture_ids[index]];
#ifdef CTDB_BOXES_VERBOSE
	printf ( "check aperture id = %d", aperture_ids[index] );
#endif
	for ( i=0; i<num_triangles; i++ )
	{
	    offset = wall_triangle_ids[i] - aperture.first_triangle_id;
	    if ( ( offset >= 0 ) && ( offset < aperture.num_triangles ) )
	    {
#ifdef CTDB_BOXES_VERBOSE
		printf ( " yes" );
#endif
		assigned_aperture_ids[count++] = aperture_ids[index];
		break;
	    }
	}
#ifdef CTDB_BOXES_VERBOSE
	printf ( "\n" );
#endif
    }
    *num_assigned_apertures = count;
}

static void 
remove_aperture_triangles(CTDB_MES_APERTURE    *apertures,
			  uint16                num_apertures,
			  uint16               *aperture_ids,
			  uint16               *num_triangles,
			  int32                *wall_triangle_ids)
{
    int32             index;
    CTDB_MES_APERTURE aperture;
    int32             i;
    int32             offset;
    int32             count;
    
    for ( index=0; index<num_apertures; index++ )
    {
	aperture = apertures[aperture_ids[index]];
#ifdef CTDB_BOXES_VERBOSE
	printf ( "triangles in aperture %d removed ", aperture_ids[index] );
#endif
	for ( i=0; i<(*num_triangles); i++ )
	{
	    if ( wall_triangle_ids[i] >= 0 )
	    {
		offset = wall_triangle_ids[i] - aperture.first_triangle_id;
		if ( ( offset >= 0 ) && ( offset < aperture.num_triangles ) )
		{
#ifdef CTDB_BOXES_VERBOSE
		    printf ( " %d", wall_triangle_ids[i] );
#endif
		    wall_triangle_ids[i] = -1;
		}
	    }
	}
#ifdef CTDB_BOXES_VERBOSE
	printf ( "\n" );
#endif
    }
    count = 0;
    for ( i=0; i<(*num_triangles); i++ )
    {
	if ( wall_triangle_ids[i] >= 0 )
	{
	    wall_triangle_ids[count++] = wall_triangle_ids[i];
	}
    }
    
    *num_triangles = count;
}

void 
ctdb_generate_boxes (CTDB_MES_TEMPLATE *template,
		     int32              maximum_triangles_per_box )
{
    int32              num_vertices;
    int32            (*vertices)[XYZ];
    int32              num_triangles;
    CTDB_MES_TRIANGLE *triangles;
    int32              num_apertures;
    CTDB_MES_APERTURE *apertures;
    CTDB_MES_GRID_BOX *boxes;
    
    int32       axis;
    float64     range[2][XYZ];
    int32       vertex_id;
    int32       distance;
    int32       power2;
    int32       power;
    int32       total;
    int32       box_id;
    int32       length, maximum_length;
    int32       count, i;
    int32       id, r, s, c;
    
    num_vertices  = template->num_vertices;
    vertices  = template->all_vertices;
    num_triangles = template->num_triangles;
    triangles = template->all_triangles;
    num_apertures = template->num_apertures;
    apertures = template->all_apertures;
    
    boxes = (CTDB_MES_GRID_BOX*)ctdb_alloc(NULL, sizeof(CTDB_MES_GRID_BOX),
					   "Grid box array");
    
    /* Calculate corners of outer most grid box */
    
    for ( axis=0; axis<XYZ; axis++ )
    {
	range[0][axis] = vertices[0][axis];
	range[1][axis] = vertices[0][axis];
    }
    for ( vertex_id=1; vertex_id<num_vertices; vertex_id++ )
    {
	for ( axis=0; axis<XYZ; axis++ )
	{
	    if ( vertices[vertex_id][axis] < range[0][axis] )
	    {
		range[0][axis] = vertices[vertex_id][axis];
	    }
	    else if ( vertices[vertex_id][axis] > range[1][axis] )
	    {
		range[1][axis] = vertices[vertex_id][axis];
	    }
	}
    }
    for ( axis=0; axis<XYZ; axis++ )
    {
	distance = range[1][axis] - range[0][axis] + 1;
#ifdef CTDB_BOXES_VERBOSE
	printf ( "axis = %d  distance = %d\n", axis, distance );
#endif
	power2 = 2;
	for ( power=2; power<31; power++ )
	{
	    power2 = power2 * 2;
	    if ( distance < power2 )
	    {
		if ( ( range[0][axis] > -(power2/2) )
		    && ( range[1][axis] < (power2/2) ) )
		{
		    boxes[0].left_front_bottom_corner[axis] = -(power2/2);
		}
		else
		{
		    boxes[0].left_front_bottom_corner[axis] = 
		      range[0][axis] - 1;
		}
		boxes[0].right_back_top_corner[axis]
		  = boxes[0].left_front_bottom_corner[axis] + power2;
		break;
	    }
	    else if ( power == 30 )
	    {
		boxes[0].left_front_bottom_corner[axis] = range[0][axis];
		boxes[0].right_back_top_corner[axis]
		  = boxes[0].left_front_bottom_corner[axis] + power2;
	    }
	}
#ifdef CTDB_BOXES_VERBOSE
	printf ( "   power2 = %d corners = %d %d\n", power2,
		boxes[0].left_front_bottom_corner[axis],
		boxes[0].right_back_top_corner[axis] );
#endif
    }
    boxes[0].outer_box_id = 0;
    boxes[0].grid_index = 0;
    
    boxes[0].num_triangles = num_triangles;
    boxes[0].wall_triangle_ids = (int32 *)
      ctdb_alloc(NULL, sizeof(int32)*num_triangles, "Wall triangles");
    
    for ( i=0; i<num_triangles; i++ )
    {
	boxes[0].wall_triangle_ids[i] = i;
    }
    
    boxes[0].num_apertures = num_apertures;
    boxes[0].aperture_ids = (uint16 *)
      ctdb_alloc(NULL, sizeof(uint16)*num_apertures, "Aperture ids");
    
    for ( i=0; i<num_apertures; i++ )
    {
	boxes[0].aperture_ids[i] = i;
    }
    
    total = 1;
    for ( box_id=0; box_id<total; box_id++ )
    {
#ifdef CTDB_BOXES_VERBOSE
	printf ( "box id = %d  triangles = %d\n",
		box_id, boxes[box_id].num_triangles );
#endif
	if ( boxes[box_id].num_triangles > maximum_triangles_per_box )
	{
	    for ( axis=0; axis<XYZ; axis++ )
	    {
		length = boxes[box_id].right_back_top_corner[axis]
		  - boxes[box_id].left_front_bottom_corner[axis];
		if ( axis == 0 )
		{
		    maximum_length = length;
		}
		else
		{
		    if ( length > maximum_length )
		    {
			maximum_length = length;
		    }
		}
	    }
	}
	else
	{
	    maximum_length = 1;
	}
	
	
	if ( maximum_length > 1 )
	{
	    uint16 dimensions[XYZ];
	    dimensions[X] = boxes[box_id].dimensions_x;
	    dimensions[Y] = boxes[box_id].dimensions_y;
	    dimensions[Z] = boxes[box_id].dimensions_z;

	    count = 1;
	    for ( axis=0; axis<XYZ; axis++ )
	    {
		length = boxes[box_id].right_back_top_corner[axis]
		  - boxes[box_id].left_front_bottom_corner[axis];
		if ( length == maximum_length )
		{
		    dimensions[axis] = 2;
		    count = count * 2;
		}
		else
		{
		    dimensions[axis] = 1;
		}
#ifdef CTDB_BOXES_VERBOSE
		printf ( "axis = %d  dimension = %d\n", axis,
			dimensions[axis] );
#endif
	    }
	    boxes[box_id].dimensions_x = dimensions[X];	
	    boxes[box_id].dimensions_y = dimensions[Y]; 
	    boxes[box_id].dimensions_z = dimensions[Z];
   
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "count = %d\n", count );
#endif
	    boxes[box_id].box_ids = (int32 *)ctdb_alloc(NULL, 
							sizeof(int32)*count,
							"Box ids");
	    
#ifdef CTDB_BOXES_VERBOSE
	    printf ( " box ids allocated\n" );
#endif
	    boxes = (CTDB_MES_GRID_BOX*) 
	      ctdb_alloc(boxes, (total+count) * sizeof(CTDB_MES_GRID_BOX),
			 "Boxes");
	    
	    id = 0;
	    for ( c=0; c<boxes[box_id].dimensions_x; c++ )
	      for ( s=0; s<boxes[box_id].dimensions_y; s++ )
		for ( r=0; r<boxes[box_id].dimensions_z; r++ )
		{
		    boxes[total].left_front_bottom_corner[X]
		      = boxes[box_id].left_front_bottom_corner[X]
			+ ( boxes[box_id].right_back_top_corner[X]
			   - boxes[box_id].left_front_bottom_corner[X] )
			  / boxes[box_id].dimensions_x * c;
		    boxes[total].left_front_bottom_corner[Y]
		      = boxes[box_id].left_front_bottom_corner[Y]
			+ ( boxes[box_id].right_back_top_corner[Y]
			   - boxes[box_id].left_front_bottom_corner[Y] )
			  / boxes[box_id].dimensions_y * s;
		    boxes[total].left_front_bottom_corner[Z]
		      = boxes[box_id].left_front_bottom_corner[Z]
			+ ( boxes[box_id].right_back_top_corner[Z]
			   - boxes[box_id].left_front_bottom_corner[Z] )
			  / boxes[box_id].dimensions_z * r;
		    boxes[total].right_back_top_corner[X]
		      = boxes[box_id].left_front_bottom_corner[X]
			+ ( boxes[box_id].right_back_top_corner[X]
			   - boxes[box_id].left_front_bottom_corner[X] )
			  / boxes[box_id].dimensions_x * (c+1);
		    boxes[total].right_back_top_corner[Y]
		      = boxes[box_id].left_front_bottom_corner[Y]
			+ ( boxes[box_id].right_back_top_corner[Y]
			   - boxes[box_id].left_front_bottom_corner[Y] )
			  / boxes[box_id].dimensions_y * (s+1);
		    boxes[total].right_back_top_corner[Z]
		      = boxes[box_id].left_front_bottom_corner[Z]
			+ ( boxes[box_id].right_back_top_corner[Z]
			   - boxes[box_id].left_front_bottom_corner[Z] )
			  / boxes[box_id].dimensions_z * (r+1);
		    boxes[total].outer_box_id = box_id;
		    boxes[total].grid_index = id;
		    
		    boxes[total].wall_triangle_ids = (int32 *)
		      ctdb_alloc(NULL, 
				 boxes[box_id].num_triangles * sizeof(int32),
				 "Wall triangle ids");
		    
		    assign_triangles ( vertices, triangles,
				      boxes[box_id].num_triangles,
				      boxes[box_id].wall_triangle_ids,
				      boxes[total].right_back_top_corner,
				      boxes[total].left_front_bottom_corner,
				      &(boxes[total].num_triangles),
				      boxes[total].wall_triangle_ids );
		    
		    boxes[total].aperture_ids = (uint16 *)
		      ctdb_alloc(NULL, 
				 boxes[box_id].num_apertures * sizeof(uint16),
				 "Aperture ids");
		    
		    assign_apertures ( apertures,
				      boxes[box_id].num_apertures,
				      boxes[box_id].aperture_ids,
				      boxes[total].num_triangles,
				      boxes[total].wall_triangle_ids,
				      &(boxes[total].num_apertures),
				      boxes[total].aperture_ids );
		    
#ifdef CTDB_BOXES_VERBOSE
		    printf ( "total = %d", total );
		    printf ( "  triangles = %d", boxes[total].num_triangles );
		    printf ( "  apertures = %d\n", 
			     boxes[total].num_apertures );
#endif
		    boxes[box_id].box_ids[id] = total++;
		    id++;
		}
	    boxes[box_id].num_triangles = 0;
	    boxes[box_id].num_apertures = 0;
	}
	else
	{
	    boxes[box_id].dimensions_x = 1;
	    boxes[box_id].dimensions_y = 1;
	    boxes[box_id].dimensions_z = 1;
#ifdef CTDB_BOXES_VERBOSE
	    printf ( "box id = %d", box_id );
	    printf ( "  triangles = %d", boxes[box_id].num_triangles );
	    printf ( "  apertures = %d\n", boxes[box_id].num_apertures );
#endif
	    
	    boxes[box_id].box_ids = NULL;
	    
	    remove_aperture_triangles ( apertures,
				       boxes[box_id].num_apertures,
				       boxes[box_id].aperture_ids,
				       &(boxes[box_id].num_triangles),
				       boxes[box_id].wall_triangle_ids );
	}      
    }
    template->num_boxes = total;
    template->all_boxes = boxes;
}
