#ifndef lint
static char rcsid [] = "$RCSfile: ct_mr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_mr.c                                                           *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: find a grade that clears all points along a segment.      *
 *       Created: 5/23/94                                                    *
 *       Author: Scott Swarts                                                *
 *       Modified for GCS by flchamberlain 11/6/95                           *
 *       Modified for OceanRep by cbuettner 3/96                             *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <math.h>
#include <stdext.h>
#include <libgcs.h>
#include "libctdb_local.h"

/* Global Local Variables */
static int32 ctdb_base_cell;
static CTDB_COORD3D ctdb_seg_start;
static CTDB_COORD3D ctdb_real_start; /* Where the vehicle actually starts out,
				      * stored in coord system specified by
				      * ctdb_base_cell
				      */
static float64 ctdb_dist_to_start;
static float64 start_z;
static float64 max_ratio;
static int32 ctdb_post_mask;
static int32            mr_no_water;
static CTDB_FEATURE_OBJ mr_obj[FC_CLASS_MAX];
static CTDB_FEAT_TYPE   mr_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];
static CTDB_FEAT_TYPE   mr_volume_subclass_obj[CTDB_NUM_VOLUME_SUBCLASSES];

typedef struct ctdb_mr_obj_data
{
    CTDB_PATCH_DATA        patch;
    CTDB_LOS_IN_XYO_AND_DB los;
    int32                  postmask;
} CTDB_MR_OBJ_DATA;


#define MAG_SQ(a,b) ((float64)(a)*(float64)(a)+(float64)(b)*(float64)(b))

#define MR_UPDATE_RATIO(distance,delta_z)            \
       if (distance > 0.0)                           \
       {                                             \
	    float64 ratio = delta_z / distance;      \
	    if (ratio > max_ratio)                   \
	      max_ratio = ratio;                     \
        }    

/* x,y are in post coordinates in ctdb's native coordinate system */
static int32 
mr_post_compare(CTDB   *ctdb, 
		float64 x, 
		float64 y, 
		float64 zl, 
		float64 zh, 
		float64 z_cross,
		int32   water)
{
    float64 distance;
    float64 delta_z;
    

    if ( water && mr_no_water )
      return VIS_KEEP_GOING;
      
    /* Figure out where the posts cross the line and check the elevation.
     */
    
    delta_z = z_cross - start_z;
    
    distance = ctdb_dist_to_start +
      fsqrt(MAG_SQ(x * ctdb->incrD - ctdb_seg_start.x,
		   y * ctdb->incrD - ctdb_seg_start.y) );

    MR_UPDATE_RATIO(distance,delta_z);
    return VIS_KEEP_GOING;
}

/* LOS is entire intervis ray stored in current coord system */
static int32 
mr_te_edge_compare(CTDB *ctdb, 
		   CTDB_LOS_IN_XYO_AND_DB *los, 
		   float64 inz0, 
		   float64 inz1, 
		   float64 tedge, 
		   float64 tray)
{
    float64 z;
    float64 delta_z, distance;
    
    z = (inz1 - inz0) * tedge + inz0;
    
    delta_z = z - start_z;
    
    if (los->length < 0.0)
    {
	los->length = sqrt(MAG_SQ(los->db.end.x - los->db.start.x,
				  los->db.end.y - los->db.start.y));
	los->distance_to_start =
	  sqrt(MAG_SQ(los->db.start.x - ctdb_seg_start.x,
		      los->db.start.y - ctdb_seg_start.y));
    }
    
    distance = ctdb_dist_to_start + los->distance_to_start + los->length * 
      tray;

    MR_UPDATE_RATIO(distance,delta_z);
    return VIS_KEEP_GOING;
}

/* LOS is entire intervis ray stored in current coord system */
static int32 
mr_edge_compare(CTDB_FEATURE_DATA      *v0, 
		CTDB_FEATURE_DATA      *v1,
		CTDB_LOS_IN_XYO_AND_DB *los
		)
{
    float64 tray, tedge;
    float64 z;
    float64 delta_z, distance;
    
    /* Figure out where the posts cross the line and store that location
     * and elevation.
     */
    
    if (ctdb_segment_intersection_t(&los->xyo,v0->xy, v1->xy, &tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;
    
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    
    delta_z = z - start_z;
    
    if (los->length < 0.0)
    {
	los->length = sqrt(MAG_SQ(los->db.end.x - los->db.start.x,
				  los->db.end.y - los->db.start.y));
	los->distance_to_start =
	  sqrt(MAG_SQ(los->db.start.x - ctdb_seg_start.x,
		      los->db.start.y - ctdb_seg_start.y));
    }
    
    distance = ctdb_dist_to_start + los->distance_to_start + los->length * 
      tray;
    
    MR_UPDATE_RATIO(distance,delta_z);
    return VIS_KEEP_GOING;
}

static int32
mr_micro_compare(CTDB               *ctdb,
		 CTDB_FEAT_TYPE_PTR  sub,
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_MR_OBJ_DATA   *data
		 )
{
    feature += 2; 

    mr_edge_compare(feature,
		    feature+(sub->incr*feature[2].micro.vert0), 
		    &(data->los));
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    mr_edge_compare(feature,
		    feature+(sub->incr*feature[2].micro.vert1),
		    &(data->los));
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    mr_edge_compare(feature,
		    feature+(sub->incr*feature[2].micro.vert2), 
		    &(data->los));    
    return VIS_KEEP_GOING; 
}


static int32
mr_water_edge_compare(
		      CTDB_FEATURE_DATA      *v0, 
		      CTDB_FEATURE_DATA      *v1,
		      CTDB_LOS_IN_XYO_AND_DB *los,
		      float32                tide)
{
    float64 tray, tedge;
    float64 z_cross;
    
    /* We treat DISJOINT and COLINEAR the same way:
     * DISJOINT means our line-of-sight cannot be blocked by this
     * segment.
     * COLINEAR segments may or may not block our line of sight, but
     * it doesn't matter: if we are dealing with closed polygons (such
     * as microterrain or a house) the tips of the other segments in
     * the polygon will block or not block our visibility as
     * appropriate; if we are dealing with an open list of segments
     * (such as a treeline) a colinear segment would have 0 width, and
     * hence could not block intervisibility of an object with >0 width.
     */
    if (ctdb_segment_intersection_t(&los->xyo,v0->xy, v1->xy,&tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;

    return mr_te_edge_compare((CTDB *)NULL,los,v0[1].z+tide,v1[1].z+tide,
			      tedge,tray);
}

static int32
mr_micro_water_compare(CTDB               *ctdb,
		       CTDB_FEAT_TYPE_PTR  sub, 
		       CTDB_FEATURE_DATA  *feature, 
		       CTDB_MR_OBJ_DATA   *data
		       )
{
    float32 tide1,tide2;
    
    feature += 2;
    
    tide1 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char0);
    
    if (!mr_water_edge_compare(feature,
			       feature+(sub->incr*feature[2].micro.vert0), 
			       &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    if (!mr_water_edge_compare(feature,
			       feature+(sub->incr*feature[2].micro.vert1),
			       &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    tide2 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char1);

    /* Since the tidal level may shift at the center edge of a polygon
     * check it again with the new tidal level if the tidal level is 
     * different from the first.
     */
    if ((tide2 != tide1) &&
	(!mr_water_edge_compare(feature,
				  feature+(sub->incr*feature[2].micro.vert0),
				  &(data->los),tide2)))
	  return VIS_STOP;
    
    if (!mr_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert2), 
				 &(data->los),tide2))
      return VIS_STOP; 

    return VIS_KEEP_GOING;
}


/* LOS is entire intervis ray stored in current coord system */
static int32 
mr_building_compare(CTDB_FEATURE_DATA *pt, 
		    int32 verts,
		    CTDB_LOS_IN_XYO_AND_DB *los)
{
    int32 i;
    CTDB_FEATURE_DATA *c0, *c1;
    
    i  = 0;
    c0 = pt;
    c1 = &pt[2];
    while(i<verts)
    {
	if (mr_edge_compare(c0, c1, los) == VIS_STOP)
	  return VIS_STOP;
	i++;
	c0 = c1;
	if (i == verts-1)
	  c1 = pt;
	else
	  c1 += 2;
    }
    
    return VIS_KEEP_GOING;
}

static int32
mr_check_micro(
	       CTDB               *ctdb,
	       CTDB_FEAT_OBJ_PTR   obj, 
	       CTDB_FEATURE_DATA  *feature, 
	       CTDB_MR_OBJ_DATA   *data,
	       int32              *next
	       )
{
    static int32 verts;
    static CTDB_FEAT_TYPE *sub=mr_micro_subclass_obj;

    verts = feature->info.vertices;

    /* check if the micro type has changed */
    if ((verts & MICRO_TYPE) != sub->type)
      sub = &mr_micro_subclass_obj[MICRO_TYPE_INDEX(verts)];

    *next = sub->incr;
    
    if (!obj->valid || 
	!(data->postmask & feature->integer) ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return VIS_KEEP_GOING;
#ifdef CTDB_OBJ_DEBUG
    printf("micro ,");
#endif

    /* call the micro type processing function */
    return sub->func(ctdb,sub,feature,data);
}


static int32
mr_volume_compare(
		  CTDB               *ctdb,
		  CTDB_FEAT_TYPE_PTR  sub,
		  CTDB_FEATURE_DATA  *feature,
		  CTDB_MR_OBJ_DATA *data
		  )
{
    int32 verts = feature->info.vertices;
    
    if (!mr_building_compare(feature+3, verts, &data->los))
      return VIS_STOP;
    return VIS_KEEP_GOING;
}

static int32
mr_mes_volume_compare(
		  CTDB               *ctdb,
		  CTDB_FEAT_TYPE_PTR  sub,
		  CTDB_FEATURE_DATA  *feature,
		  CTDB_MR_OBJ_DATA *data
		  )
{
    int32 verts = feature->info.vertices;
    
    if (!mr_building_compare(feature+3, verts, &data->los))
      return VIS_STOP;
    return VIS_KEEP_GOING;
}

static int32
mr_check_volume(
		CTDB               *ctdb,
		CTDB_FEAT_OBJ_PTR   obj, 
		CTDB_FEATURE_DATA  *feature, 
		CTDB_MR_OBJ_DATA *data,
		int32              *next
		)
{
    CTDB_FEAT_TYPE    *sub;    
    int32 verts = feature->info.vertices;
    
    *next = NEXT_VOLUME(verts);

    if (!obj->valid || 
	!(data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;
    
    sub = &mr_volume_subclass_obj[ feature[1].volume.type ];
    return sub->func(ctdb,sub,feature,data);
}


static int32
mr_check_canopy(CTDB               *ctdb,
		CTDB_FEAT_OBJ_PTR   obj, 
		CTDB_FEATURE_DATA  *feature, 
		CTDB_MR_OBJ_DATA   *data,
		int32              *next
		)
{
    static int32 g, verts, incr;

    verts = feature->info.vertices;
    incr  = *next = NEXT_CANOPY(verts);

    if (!obj->valid || 
	!(data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

#ifdef CTDB_OBJ_DEBUG    
    printf("canopy, ");    
#endif   
    feature++;

    /* Note that the if-continue contstruct used below
     * runs faster than an equivalent fall-through switch-case.
     */
    
    g = incr*feature[2].micro.vert0;
    if (!mr_edge_compare(feature, feature+g, &data->los))
      return VIS_STOP;
    
    if (verts == 1)
      return VIS_KEEP_GOING;
		
    g = incr*feature[2].micro.vert1;
    if (!mr_edge_compare(feature, feature+g, &data->los))
      return VIS_STOP;
    
    if (verts == 2)
      return VIS_KEEP_GOING;
    
    g = incr*feature[2].micro.vert2;
    if (!mr_edge_compare(feature, feature+g, &data->los))
      return VIS_STOP;
    
    return VIS_KEEP_GOING;
}

static int32
mr_check_linear(CTDB               *ctdb, 
		CTDB_FEAT_OBJ_PTR   obj, 
		CTDB_FEATURE_DATA  *feature, 
		CTDB_MR_OBJ_DATA   *data,
		int32              *next
		)
{
    int32 verts = feature->info.vertices;
    *next = NEXT_LINEAR(verts);

    if (!obj->valid || 
	!(data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

    verts--;
    feature += 2;
    
    while(verts--)
    {
	if (!mr_edge_compare(feature,feature+2,&data->los))
	  return VIS_STOP;
	feature += 2;
    }
    
    return VIS_KEEP_GOING;
}

static int32
mr_check_aggregate(CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj, 
		   CTDB_FEATURE_DATA  *feature, 
		   CTDB_MR_OBJ_DATA   *data,
		   int32              *next
		   )
{
    /* Currently, there are only individual trees in aggregate features.
     * We ignore individual trees.
     */
    *next = NEXT_AGGREGATE(0);
    
    return VIS_KEEP_GOING;
}

int32
ctdb_mr_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static CTDB_TYPE_FUNC mr_micro_func[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)mr_micro_compare,      /* MICRO_TYPE_BASE */ 
	(CTDB_TYPE_FUNC)mr_micro_compare,      /* MICRO_TYPE_ML */
	(CTDB_TYPE_FUNC)mr_micro_compare,      /* MICRO_TYPE_DEFAULT */
	(CTDB_TYPE_FUNC)mr_micro_water_compare /* MICRO_TYPE_WATER */
    };

    static CTDB_TYPE_FUNC mr_volume_func[CTDB_NUM_VOLUME_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)mr_volume_compare,      /* CTDB_FT_BUILDING */
	(CTDB_TYPE_FUNC)mr_volume_compare,      /* CTDB_FT_ROCK_DROP */
	(CTDB_TYPE_FUNC)mr_mes_volume_compare,  /* CTDB_FT_MES */
	(CTDB_TYPE_FUNC)mr_volume_compare       /* CTDB_FT_PLUG */
    };

    /* CLASS FC_MICRO */
    obj[FC_MICRO].func         = (CTDB_OBJ_FUNC)mr_check_micro;
    ctdb_init_micro_subclass(mr_micro_subclass_obj,mr_micro_func);

    /* CLASS FC_VOLUME */
    obj[FC_VOLUME].func        = (CTDB_OBJ_FUNC)mr_check_volume;
    ctdb_init_volume_subclass(mr_volume_subclass_obj, 
			      mr_volume_func);

    obj[FC_LINEAR].func        = (CTDB_OBJ_FUNC)mr_check_linear;
    obj[FC_CANOPY].func        = (CTDB_OBJ_FUNC)mr_check_canopy;
    obj[FC_LAID_LINEAR].func   = (CTDB_OBJ_FUNC)ctdb_invalid_laid_linear;
    obj[FC_AGGREGATE].func     = (CTDB_OBJ_FUNC)mr_check_aggregate;

    /* Store Initialized local feature objects locally */
    bcopy(obj,mr_obj,sizeof(mr_obj));    
    return 1;
}

/* start, end are the endpoints of the *entire* intervis ray, represented in
 * the frame of reference specified by ctdb.
 */

static int32 
mr_patch_compare(
		 CTDB    *ctdb, 
		 int32    patch_number, 
		 int32    feature_mask, 
		 int32    post_mask, 
		 float64 *start, 
		 float64 *end
		 )
{
    int32 ray_info,i,next;
    CTDB_FEATURE_DATA *feature, *last;
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_MR_OBJ_DATA data;

    feature = (CTDB_FEATURE_DATA *)ctdb_lookup_feature(ctdb, patch_number);

    if (!feature)
      return VIS_KEEP_GOING;
	
    data.postmask              = post_mask;
    data.patch.number          = patch_number;
    data.patch.x               = data.patch.number % ctdb->patches_wide;
    data.patch.y               = data.patch.number / ctdb->patches_wide;
    
    data.los.xyo.start.x       = POST_TO_XYO(start[X], data.patch.x, 
				  ctdb->patch_inv_incr);
    data.los.xyo.start.y       = POST_TO_XYO(start[Y], data.patch.y, 
				  ctdb->patch_inv_incr);
    data.los.xyo.end.x         = POST_TO_XYO(end[X], data.patch.x, 
				     ctdb->patch_inv_incr);
    data.los.xyo.end.y         = POST_TO_XYO(end[Y], data.patch.y, 
				     ctdb->patch_inv_incr);
    data.los.db.start.x        = start[X] * ctdb->incrD;
    data.los.db.start.y        = start[Y] * ctdb->incrD;
    data.los.db.end.x          = end[X] * ctdb->incrD;
    data.los.db.end.y          = end[Y] * ctdb->incrD;
    data.los.length            = -1.0; /* Calculated if needed */
    data.los.distance_to_start = -1.0; /* Calculated if needed */
    
    /* Run through all the features on this patch looking for
     * things which might cross our line-of-sight (post_mask check).
     */
    
    /* Note that bit masks are used in some places rather than
     * bitfield accessors to avoid unnecessary shifting (a big
     * performance win, it turns out).
     */

    last = feature + feature->hdr.size;
    feature++;

    /* initialize the feature object valid bits for this patch */
    obj = (CTDB_FEAT_OBJ_PTR)mr_obj;
    obj->valid = feature_mask & obj->featmask;

    for(;feature<last; feature+=next)
    {
	int32 fclass=feature->info.feature_class;
	
	/* if the current class doesn't match the previos class it is time 
	 * to change the class object pointer.
	 */
	if ( fclass != obj->sensClass )
	{	    
	    /* If we are done with the feature class then remove from the 
	     * list and check if we are done.  If the class is LINEAR then
	     * don't remove the feature flag POST_TREES since FC_CANOPY may
	     * also exist in the list of features to evaluate.
	     */
	    if ( obj->sensClass != FC_LINEAR )
	      feature_mask &= ~obj->featmask;

	    if ( !feature_mask )
	      break;

	    obj = (CTDB_FEAT_OBJ_PTR)&mr_obj[fclass];	
	    obj->valid = feature_mask & obj->featmask;
	}
	
	if ( obj->func(ctdb,obj,feature,(ADDRESS)&data,&next) == VIS_STOP)
	   return VIS_STOP;
    }
    return VIS_KEEP_GOING;
}


static int32
mr_cell_fcn(CTDB *ctdb, 
	    float64 *ray_start, 
	    float64 *ray_end, 
	    float64 *cur_start, 
	    float64 *cur_end, 
	    float64 width)
{
    float64 ray_start_post[3];
    float64 ray_end_post[4];
    float64 cur_start_post[4];
    float64 cur_end_post[4];
    CTDB_COORD3D start, end, cell_start, cell_end;
    int32 cur_cell;
    
    /* Scale from meters to grid spacing */
    ray_start_post[X] = ray_start[X] * ctdb->inv_incr;
    ray_start_post[Y] = ray_start[Y] * ctdb->inv_incr;
    ray_start_post[ZL] = ray_start[ZL];
    
    ray_end_post[X] = ray_end[X] * ctdb->inv_incr;
    ray_end_post[Y] = ray_end[Y] * ctdb->inv_incr;
    ray_end_post[ZL] = ray_end[ZL];
    ray_end_post[ZH] = ray_end[ZH];
    
    cur_start_post[X] = cur_start[X] * ctdb->inv_incr;
    cur_start_post[Y] = cur_start[Y] * ctdb->inv_incr;
    cur_start_post[ZL] = cur_start[ZL];
    cur_start_post[ZH] = cur_start[ZH];
    
    cur_end_post[X] = cur_end[X] * ctdb->inv_incr;
    cur_end_post[Y] = cur_end[Y] * ctdb->inv_incr;
    cur_end_post[ZL] = cur_end[ZL];
    cur_end_post[ZH] = cur_end[ZH];
    
    /* Copy endpoints into structure used by terrain_element_intervis */
    start.x = ray_start[X];
    start.y = ray_start[Y];
    start.z = ray_start[Z];
    end.x   = ray_end[X];
    end.y   = ray_end[Y];
    end.z   = ray_end[ZL];
    cell_start.x = cur_start[X];
    cell_start.y = cur_start[Y];
    cell_start.z = cur_start[ZL];
    cell_end.x = cur_end[X];
    cell_end.y = cur_end[Y];
    cell_end.z = cur_end[ZL];
    
    cur_cell = ctdb_get_cell_id(ctdb);
    
    /* Set up necessary static globals */
    if (cur_cell != ctdb_base_cell)
    {
	float64 dummy1, dummy2;
	
	gcs_gcs3e_to_gcscs(ctdb_base_cell, ray_start[X], ray_start[Y], 
			 ray_start[Z], cur_cell,
			 &ctdb_seg_start.x, &ctdb_seg_start.y,
			 &ctdb_seg_start.z);
	
	gcs_gcs3e_to_gcscs(ctdb_base_cell, 
			 ctdb_real_start.x, ctdb_real_start.y, 
			 ctdb_real_start.z, cur_cell,
			 &dummy1, &dummy2, &start_z);
    }
    else
    {
	ctdb_seg_start = start;
	start_z = ctdb_real_start.z;
    }
    
    /* call terrain element intervis. if there are tinned patches */
    if ( ctdb->is_tinned ) 
    {
	(void) ctdb_terrain_element_intervis(ctdb, &start, &end, ray_start[ZH],
					     &cell_start, &cell_end,
					     mr_te_edge_compare,
					     FALSE);
    }
    
    /* Check features, grid posts and patch water (if applicable) */
    /* Invoke the intervis engine to find the ratio. */
    if ( ctdb->is_gridded || ctdb_post_mask || !mr_no_water )
      (void)ctdb_intervis(ctdb, ray_start_post, ray_end_post, 
			  cur_start_post, cur_end_post, TRUE,
			  mr_post_compare, mr_patch_compare, ctdb_post_mask);
    
    return VIS_KEEP_GOING;
}

float64 
ctdb_find_max_ratio(CTDB    *ctdb, 
		    float64  st_z,           /* starting altitude of seg */
		    float64  distance_to_st, /* dist of veh from segment */
		    CTDB_ELEV_QUALIFIER qual,
		    float64 *start,          /* segment start */
		    float64 *end)            /* segment end */
{
    /* pointer is needed to allow quick remapping instead of always reindexing
     */
    static CTDB_TYPE_FUNC *water_func=
      &mr_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_WATER)].func;

   /* These are the default flags for this API routine */
    static CTDB_ELEV_QUALIFIER dflt =(CTDB_MAX_LAND_AND_WATER);
    static CTDB_ELEV_QUALIFIER valid=(CTDB_ALL_FEATURES     |
				      CTDB_MAX_LAND_AND_WATER);
    /* These are the flags which are maniditory for this API routine */
    static CTDB_ELEV_QUALIFIER must=CTDB_MAX_LAND_NO_WATER;

    /* Automatic Variables */
    float64 z;
    float64 ratio;
    float64 length;
    float64 z0, z1;
    float64 dx, dy, mag;
    CTDB_ELEV_QUALIFIER elev_qual;
    
    if (!ctdb)
      return 0.0;
  
    /* Apply manditory arguments and validate passed in mask */
    qual = (qual?((qual|must)&valid):dflt);

    /* If the search is to include water but the starting z (st_z) is 
     * underwater then remove the water surface (since this will not contribute
     * to the max ratio.  If st_z is at or above the water surface then 
     * include the water surface if the input qualifier requests water.
     */
    if ((qual & CTDB_INCL_WATER)==CTDB_INCL_WATER)
    {
	static CTDB_ELEV_DATA elev[CTDB_MAX_ELEVS],*elevptr;
	
	ctdb_lookup_elevation_mes(ctdb,start[X],start[Y],
				  CTDB_LAND_AND_WATER,elev);
	mr_no_water  = FALSE;
	*water_func  = (CTDB_TYPE_FUNC)mr_micro_water_compare;
	elev_qual    = CTDB_MAX_LAND_AND_WATER;

	for (elevptr=elev; elevptr->sensClass != CTDB_FC_ILLEGAL; elevptr++ )
	{
	    if ((elevptr->subclass == CTDB_FS_WATER ) &&
		(st_z < elevptr->z))
	    {
		qual &= ~CTDB_INCL_WATER;
		mr_no_water  = TRUE;
		*water_func  = (CTDB_TYPE_FUNC)ctdb_invalid_type;
		elev_qual    = CTDB_MAX_LAND_NO_WATER;
		break;
	    }
	}
	
    }
    else
    {
	elev_qual    = CTDB_MAX_LAND_NO_WATER;
	mr_no_water  = FALSE;
	*water_func  = (CTDB_TYPE_FUNC)mr_micro_water_compare;
    }
	    
    z0 = st_z;
    /* Put z1 on the database based on the qualifiers supplied on input */
    z1 = ctdb_lookup_qual_elevation(ctdb, end[X], end[Y], elev_qual, 
				    CTDB_NO_REF_ELEV, (CTDB_ELEV_DATA *)NULL);
       
    /* Calculate where the vehicle is actually starting from */
    dx = end[X] - start[X];
    dy = end[Y] - start[Y];
    mag = sqrt(dx*dx + dy*dy);
    dx /= mag;
    dy /= mag;
    ctdb_real_start.x = start[X] - (distance_to_st * dx);
    ctdb_real_start.y = start[Y] - (distance_to_st * dy);
    ctdb_real_start.z = st_z;
    
    /* Set up static data */
    ctdb_dist_to_start = distance_to_st;
    max_ratio          = -E_GIGA;	/* Big negative number */
    ctdb_base_cell     = ctdb_get_cell_id(ctdb);
    
    /* This routine assumes that CTDB_INCL_TERRAIN_SKIN which means that 
     * CTDB_INCL_MICRO is assumed.
     */
    ctdb_post_mask = POST_MICROTERRAIN;
    /* Initialize the feature objects that will be evaluated this go
     * around.   POST_FEATURE was not checked since these are laid linears
     * which have no height component associated with them and therefore
     * would not contribute to the max ration calculation.
     */    
    if ( qual & CTDB_INCL_LINEAR )
      ctdb_post_mask |= POST_TREES;

    if ( qual & CTDB_INCL_CANOPY )
      ctdb_post_mask |= POST_TREES;
    
    if ( qual & CTDB_INCL_VOLUME )
      ctdb_post_mask |= POST_BUILDINGS;

    /* Run through all cells hit by ray looking for max ratio */
    ctdb_traverse_ray(ctdb, start[X], start[Y], z0, end[X], end[Y], z1,
		      0.0, 0.0, mr_cell_fcn);
    
    /* Now look at the segment start and end. */
    if (distance_to_st > 0.0)
    {
	/* If the distance_to_st is 0.0 then we don't look at the start. */
	z = ctdb_lookup_qual_elevation(ctdb, start[X], start[Y], 
				       elev_qual, CTDB_NO_REF_ELEV, 
				       (CTDB_ELEV_DATA *)NULL);
	ratio = (z - st_z) / distance_to_st;
	if (ratio > max_ratio)
	  max_ratio = ratio;
    }
    
    /* Look at the end. */
    z =  ctdb_lookup_qual_elevation(ctdb, end[X], end[Y], elev_qual, 
				    CTDB_NO_REF_ELEV, (CTDB_ELEV_DATA *)NULL);
    length = fsqrt(MAG_SQ(end[X] - start[X], end[Y] - start[Y]));
    ratio = (z - st_z) / (distance_to_st + length);

    if (ratio > max_ratio)
      max_ratio = ratio;
    
    return max_ratio;
}
