#ifndef lint
static char rcsid [] = "$RCSfile: ct_print.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_print.c                                                       *
 *                                                                          *
 *       Copyright 1995 by SAIC & TASC                                      *
 *                                                                          *
 *       SAIC                          TASC                                 *
 *       20 Burlington Mall Road       55 Walkers Brook Driver              *
 *       Suite 130                     Reading, MA 01867                    *
 *       Burlington, MA 01803          617-942-2000                         *
 *       617-221-7612                                                       *
 *                                                                          *
 *       This software was developed by TASC and SAIC under U. S. Government*
 *       contracts                                                          *
 *       and may be reproduced by or for the U. S. Government pursuant to   *
 *       the copyright license under the clause at DFARS 252.227-7013       *
 *       (OCT 1988).                                                        *
 *                                                                          *
 *       Contents: Routines to ASCII print information in the database      *
 *       Created:  January 1995                                             *
 *       Author:   buettner                                                 *
 *       Remarks:                                                           *
 *                                                                          *
 ****************************************************************************/
#include <stdstring.h>
#include <stdio.h>
#include <libgcs.h>
#include "libctdb_local.h"


#define KAY (1024.0)
static CTDB_FEATURE_OBJ print_obj[FC_CLASS_MAX];

extern CTDB_COLLECTION ctdb_data;
CTDB_COLLECTION ctdb_data;
static CTDB_COLLECTION total;
    
typedef struct ctdb_print_obj_data
{
    int32 n_feat;
} CTDB_PRINT_OBJ_DATA; 
    
static void
ResetMetrics(CTDB_COLLECTION *data)
{
    data->patch_number         = 0;
    data->te_edges_crossed     = 0;
    data->te_edges_intersected = 0;
    data->tes_checked          = 0;
    data->tes_intersected      = 0;
    data->vertices_crossed     = 0;
    data->missed_init_te       = 0;
    data->status               = 0;
    data->do_lookup            = 0;   
    data->visibility           = 1.0;
}

void
ctdb_print_intervis_metrics(CTDB *db, CTDB_PRINT_MODE mode)
{
    if ( mode == CTDB_PRINT_HEADER ) /* header */
    {
	ResetMetrics(&total);
	ResetMetrics(&ctdb_data);
	printf("\n");
	printf("patch\tedges\tedges\tpolys\tpolys \tverts \
\tmissed\tdensity\tvis\tstatus\tlookup\n");
	printf("number\teval- \tinter- \teval-\tinter-\tcrossed\tinit \t\t\n");
	printf("      \tuated\tsected\tuated\tsected\t       \t  te  \t\t\n");
    }
    else if ( mode == CTDB_PRINT_LINE )
    {
	printf("%5d\t%5d\t%5d\t%5d\t%5d\t%5d\t%5d\t%5d\t%3.2f\t%3d\t%1d\n"
	       ,ctdb_data.patch_number
	       ,ctdb_data.te_edges_crossed
	       ,ctdb_data.te_edges_intersected
	       ,ctdb_data.tes_checked
	       ,ctdb_data.tes_intersected
	       ,ctdb_data.vertices_crossed
	       ,ctdb_data.missed_init_te
	       ,ctdb_terrain_element_density(db,ctdb_data.patch_number)
	       ,ctdb_data.visibility
	       ,ctdb_data.status
	       ,ctdb_data.do_lookup);
	total.te_edges_crossed     += ctdb_data.te_edges_crossed;
	total.te_edges_intersected += ctdb_data.te_edges_intersected;
	total.tes_checked          += ctdb_data.tes_checked;
	total.tes_intersected      += ctdb_data.tes_intersected;
	total.vertices_crossed     += ctdb_data.vertices_crossed;
	total.missed_init_te       += ctdb_data.missed_init_te;
	total.visibility           = ctdb_data.visibility;
	
	ResetMetrics(&ctdb_data);
    }
    else 
    {
	int32 i;
	
	for (i=0;i<80;i++)
	  printf("-");
	
	printf("\nTotals\t%5d\t%5d\t%5d\t%5d\t%5d\t%5d\t\t%3.2f\n"
	       ,total.te_edges_crossed
	       ,total.te_edges_intersected
	       ,total.tes_checked
	       ,total.tes_intersected
	       ,total.vertices_crossed
	       ,total.missed_init_te
	       ,total.visibility);
    }    
}

int32 
ctdb_terrain_element_density(CTDB *ctdb, int32 patch_number)
{
    int32 member=PG_MEMBER(patch_number);
    CTDB_PATCH_GROUP *patch_group=
      ( ctdb->patch_groups+PG_NUMBER(patch_number) );
    if  ( (member == 0 && patch_group->patch0_te16 ) ||
	 (member == 1 && patch_group->patch1_te16 ) ||
	 (member == 2 && patch_group->patch2_te16 ) ||
	 (member == 3 && patch_group->patch3_te16 ) )
      return ctdb_terrain_element_density16(ctdb,patch_number);
    else
      return ctdb_terrain_element_density08(ctdb,patch_number);
}

void 
ctdb_print_te_data(CTDB *ctdb, int32 patch_number, CTDB_PRINT_MODE mode)
{
    int32 member=PG_MEMBER(patch_number);
    ADDRESS te=ctdb_lookup_te_ptr(ctdb,patch_number);
    CTDB_PATCH_GROUP *pg=ctdb->patch_groups+PG_NUMBER(patch_number);
    uint32 i;
    
    if (te)
      if ( mode == CTDB_PRINT_TABLE )
      {
	  int32 n_te = CTDB_NUM_TE(pg, member);
	  printf("Vert#\tx\ty\tz\t\tverts\t\t\t\tadj\t\t\t\tsoil0\tsoil1\n");
	  if (( member == 0 && pg->patch0_te16 ) ||
	      ( member == 1 && pg->patch1_te16 ) ||
	      ( member == 2 && pg->patch2_te16 ) ||
	      ( member == 3 && pg->patch3_te16 ))
	  {	
	      for (i=0;i<n_te;i++)
		ctdb_print_te_element16((TERRAIN_ELEMENT_16 *)te,i);
	  }
	  else
	  {
	      for (i=0;i<n_te;i++)
		ctdb_print_te_element08((TERRAIN_ELEMENT_08 *)te,i);
	  }
      }
}

void 
ctdb_print_te_map(CTDB *ctdb, int32 patch_number)
{
    int32             s;
    CTDB_PATCH_GROUP *pg = ctdb->patch_groups+PG_NUMBER(patch_number);
    int32             member = PG_MEMBER(patch_number);
    uint16           *map;   
    int32             sqdim;
    
    /* Map is only valid if there are terrain elements in this 
     * patch
     */
    if ( !ctdb_lookup_te_ptr(ctdb,patch_number) )
      return;
    
    if (!pg->mapped)
      ctdb_fetch_patch_group(ctdb, pg);
    
    sqdim = pg->data_start.tin_data->te_vgrid_dim[member] * 
      pg->data_start.tin_data->te_vgrid_dim[member];
    
    for (s=0;s<sqdim;s++,map++)
    {
	if (s == 0)
	{
	    map = (uint16*)((CTDB_FEATURE_DATA*)pg->data_start.grid_data +
			    pg->data_start.tin_data->
			    te_vgrid_offset[member]);
	    printf("\nPatch Number: %d \tMaximum Elevation = %9.3f\n",
		   patch_number,CTDB_LOOKUP_MAX_Z(pg, member));
	    printf("\t\tVirtual Grid\tVertex Number\n");
	}
	printf("\t\t%d\t\t%d\n",s,*map);
    }
}

static void 
print_patch_hdr(CTDB_FEATURE_DATA *f)
{
    CTDB_PATCH_HDR *h = &f->hdr;
    printf("Patch Header: size %d\n", h->size);
}

static void 
print_feature_info(CTDB_FEATURE_DATA *f, int32 n)
{
    CTDB_FEATURE_INFO *info = &f->info;
    char buffer[15];
    uint32 mask=0xf;
    
    printf(" Feature Info for #%d: ", n);
    switch (info->feature_class)
    {
      case FC_MICRO:
	printf("FC_MICRO: "); mask=MICRO_VERTS; break;
      case FC_VOLUME:
	printf("FC_VOLUME: "); break;
      case FC_LINEAR:
	printf("FC_LINEAR: "); break;
      case FC_CANOPY:
	printf("FC_CANOPY: "); break;
      case FC_AGGREGATE:
	printf("FC_AGGREGATE: "); break;
      case FC_LAID_LINEAR:
	printf("FC_LAID_LINEAR: "); break;
      default:
	printf("Unknown");
    }
    
    printf(", %d vertices, post_mask  0x%x\n",(info->vertices&mask),
	   info->post_mask);
}

void 
ctdb_print_posts(CTDB           *ctdb,
		 int32           patch_x,
		 int32           patch_y,
		 CTDB_POST_STAT *stat
		 )
{
    int32 posts_per_patch_side = ctdb->patch_incr;
    int32 x = patch_x * posts_per_patch_side;
    int32 y = patch_y * posts_per_patch_side;
    int32 j;
    int32 i;

    printf("\n---Posts in patch (%d, %d)---\n", patch_x, patch_y);
    printf("---  implied patch lower left is at (%.2f, %.2f)---\n",
	   (float64)(x * ctdb->incr), (float64)(y * ctdb->incr));

    for (j = 0; j < posts_per_patch_side; j++)
      for (i = 0; i < posts_per_patch_side; i++)
      {
	  int32 norm_x = x + i;
	  int32 norm_y = y + j;
	  int32 world_x = norm_x * ctdb->incr;
	  int32 world_y = norm_y * ctdb->incr;
	  int32 elev_post = ctdb_lookup_elev_post(ctdb, norm_x, norm_y);
	  int32 soil_post = ctdb_lookup_soil_post(ctdb, norm_x, norm_y);

	  printf("Post at (%d, %d): NE soil %d, SE soil %d\n",
		 world_x, world_y, GET_POST_NE_SOIL(soil_post),
		 GET_POST_SW_SOIL(soil_post));

	  stat->posts++;

	  if (elev_post & POST_FEATURES)
	  {
	      stat->features++;
	      printf(" post has features\n");
	  }
	  if (elev_post & POST_BUILDINGS)
	  {
	      stat->buildings++;
	      printf(" post has buildings\n");
	  }
	  if (elev_post & POST_TREES)
	  {
	      stat->trees++;
	      printf(" post has trees\n");
	  }
	  if (elev_post & POST_MICROTERRAIN)
	  {
	      stat->micro++;
	      printf(" post has microterrain\n");
	  }

	  printf(" elevation 0x%x (%.2f)\n",
		 (elev_post & 0x0007FFFF),
		 GET_POST_ELEVATION(elev_post, ctdb->fixed_point_basis));
      }
}

static void 
ctdb_print_linear(CTDB *ctdb, CTDB_FEATURE_DATA *f)
{
    CTDB_LINEAR_INFO *t = &f->linear;
    
    printf("  LINEAR: type %d  width 0x%x (%.2f), ",
	   ctdb->linear_models[t->reference].type,
	   ctdb->linear_models[t->reference].width,
	   CTDB_XYO_TO_METERS(ctdb->linear_models[t->reference].width, 0,
			      ctdb->patch_incrD,ctdb->incrD));
    if(ctdb->linear_models[t->reference].type == CTDB_FS_TREELINE)
    {
	if (t->first)
	  printf("is first, ");
	if (t->last)
	  printf("is last, ");
	printf("height 0x%x (%.2f)\n",
	    ctdb->linear_models[t->reference].linear_data.treeline_info.
	    foliage_height,
	    ctdb->linear_models[t->reference].linear_data.treeline_info.
	    foliage_height * CTDB_TREE_HEIGHT_BASIS);
    }
}

static void 
ctdb_print_aggregate(CTDB *ctdb, CTDB_FEATURE_DATA *f)
{
    CTDB_AGGREGATE_INFO *t = &f->aggregate;

    printf("  AGGREGATE:  %d features",
	   ctdb->aggregate_models[t->reference]->n_features);
}

static void
ctdb_print_laid_linear(CTDB *ctdb, CTDB_FEATURE_DATA *f)
{
    CTDB_VAR_ATTR_ONE_SOIL *ll_soil = &f->ll_soil;
    CTDB_LAID_LINEAR_INFO *ll = &f[1].laid_linear;
    
    printf("  LAID LINEAR: width 0x%x (%.2f), soil %d\n",
	   ll->width,
	   CTDB_XYO_TO_METERS(ll->width, 0, ctdb->patch_incrD,ctdb->incrD),
	   ll_soil->poly_char_index); 
}

static void 
print_xy_patch_offset(CTDB *ctdb, CTDB_FEATURE_DATA *f)
{
    CTDB_XYO_COORD2D pt;
    
    pt = f->xy;
    
    printf("  XY Patch relative offset: 0x%x, 0x%x (%.2f, %.2f)\n",
	   pt.x,
	   pt.y,
	   CTDB_XYO_TO_METERS(pt.x, 0, ctdb->patch_incrD,ctdb->incrD),
	   CTDB_XYO_TO_METERS(pt.y, 0, ctdb->patch_incrD,ctdb->incrD) );
}

static void 
print_z(CTDB_FEATURE_DATA *f)
{
    float32 z = f->z;
    
    printf("  Z Elevation: %.2f\n", z);
}

#define NEWLINE	 	printf("\n")

static void 
print_micro_pts(CTDB_FEATURE_DATA *f, int32 verts)
{
    CTDB_MICRO_PTS *m = &f->micro;
    CTDB_VAR_ATTR_TWO_SOILS *ms = &f->micro_soil;
    
    printf("  MICROTERRAIN: ");
    switch (verts & MICRO_TYPE)
    {
      case MICRO_TYPE_BASE: printf("BASE "); break;
      case MICRO_TYPE_ML: printf("ML "); break;
      case MICRO_TYPE_DEFAULT: printf("DEFAULT "); break;
      case MICRO_TYPE_WATER: printf("WATER "); break;
      default: printf("(?) "); return;
    }
    if ((verts & MICRO_VERTS) > 1)
      printf(", soil0 %d ", ms->poly_char_index0);
    if ((verts & MICRO_VERTS) > 2)
      printf(", soil1 %d ", ms->poly_char_index1);
    if ((verts & MICRO_VERTS) > 0)
      printf(", vert0 %d ", m->vert0);
    if ((verts & MICRO_VERTS) > 1)
      printf(", vert1 %d ", m->vert1);
    if ((verts & MICRO_VERTS) > 2)
      printf(", vert2 %d ", m->vert2);
    printf("\n");
}

static void 
print_micro_characteristics(CTDB_FEATURE_DATA *f)
{
    CTDB_MWATER_CHARS *m = &f->mwater;
    
    printf("  MWATER: %d\t%d\n",m->char0,m->char1);
}


static void 
ctdb_print_canopy_pts(CTDB_FEATURE_DATA *data)
{
    printf("  CANOPY: ");
    
    if (data->canopy.vert0_outer)
      printf("  edge 0 is outer,");
    if (data->canopy.vert1_outer)
      printf("edge 1 is outer,");
    if (data->canopy.vert2_outer)
      printf("edge 2 is outer,");
    
    printf("vert%d= %d,",0,data->micro.vert0);
    printf("vert%d= %d,",1,data->micro.vert1);
    printf("vert%d= %d,",2,data->micro.vert2);
        
    NEWLINE     ;  
}

/* Some MES printing routines */
#define MES2DB(a) ((float32)(a)*mes_template->fixed_point_basis)
static void
print_boxes (CTDB            *ctdb,
	     CTDB_MES_VOLUME *mes )
{
    CTDB_MES_TEMPLATE *mes_template;
    CTDB_MES_GRID_BOX *boxes;
    int32              box_id;
    int32              count, i;
    int32              num_boxes;
    
    mes_template = mes->template;
    boxes        = mes_template->all_boxes;
    num_boxes    = mes_template->num_boxes;
    
    for ( box_id=0; box_id<num_boxes; box_id++ )
    {
	printf("box #%d\n  outer box id =%3d   grid index =%3d\n",
	       box_id, boxes[box_id].outer_box_id, boxes[box_id].grid_index);
	printf("  bottom corner =%.3f %.3f %.3f\n",
	       MES2DB(boxes[box_id].left_front_bottom_corner[X]),
	       MES2DB(boxes[box_id].left_front_bottom_corner[Y]),
	       MES2DB(boxes[box_id].left_front_bottom_corner[Z]));
	printf("  top corner =%.3f %.3f %.3f\n",
	       MES2DB(boxes[box_id].right_back_top_corner[X]),
	       MES2DB(boxes[box_id].right_back_top_corner[Y]),
	       MES2DB(boxes[box_id].right_back_top_corner[Z]));
	printf("  dimensions =%.3f %.3f %.3f\n",
	       MES2DB(boxes[box_id].dimensions_x), 
	       MES2DB(boxes[box_id].dimensions_y),
	       MES2DB(boxes[box_id].dimensions_z));
	printf("  grid box ids = " );
	count = boxes[box_id].dimensions_x * boxes[box_id].dimensions_y *
	  boxes[box_id].dimensions_z;
	if ( count > 1 )
	{
	    for ( i=0; i<count; i++ )
	    {
		printf(" %3d", boxes[box_id].box_ids[i]);
	    }
	}
	printf("\n");
	printf("  triangle ids = ");
	for (i=0; i<boxes[box_id].num_triangles; i++)
	{
	    printf(" %3d", boxes[box_id].wall_triangle_ids[i]);
	}
	printf("\n");
	printf("  aperture ids = ");
	for (i=0; i<boxes[box_id].num_apertures; i++)
	{
	    printf(" %3d", boxes[box_id].aperture_ids[i]);
	}
	printf("\n");
    }
}

static void
print_triangles (CTDB *ctdb,
		 CTDB_MES_VOLUME *mes)
{
    int32             (*vertices)[XYZ];
    int32              num_triangles;
    CTDB_MES_TRIANGLE *triangles;
    int32              num_apertures;
    CTDB_MES_APERTURE *apertures;
    CTDB_MES_TEMPLATE *mes_template;
    
    int32              id;
    CTDB_MES_TRIANGLE *triangle;
    CTDB_MES_APERTURE  aperture;
    float64            normal[XYZ];
    int32              v;
    int32             *vertex;
    
    mes_template  = mes->template;
    vertices      = mes_template->all_vertices;
    num_triangles = mes_template->num_triangles;
    triangles     = mes_template->all_triangles;
    num_apertures = mes_template->num_apertures;
    apertures     = mes_template->all_apertures;
    
    for (id=0; id<num_triangles; id++)
    {
	printf("triangle #%d\n", id);
	triangle = &triangles[id];
	for (v=0; v<3; v++)
	{
	    vertex = vertices[triangle->vertex_ids[v]];
	    printf("\tvertex%1d = %.3f %.3f %.3f\n", v,
		   MES2DB(vertex[X]), MES2DB(vertex[Y]), MES2DB(vertex[Z]));
	}

	ctdb_mesvec_to_gcscsvec(mes, triangle->normal, ctdb->cell_id,
				normal);
	vmat3_unit64(normal,normal);
	
	printf("\tnormal = %.3f %.3f %.3f\n\n", normal[X], normal[Y], 
	       normal[Z]);
    }
    
    for (id=0; id<num_apertures; id++)
    {
	printf("aperture #%d\n", id);
	aperture = apertures[id];
	printf("   triangle ids = " );
	for (v=0; v<(aperture.num_triangles); v++)
	{
	    printf ("%12d ", aperture.first_triangle_id+v);
	}
	printf("\n   visible = %1d  open = %1d\n\n",
	       aperture.visible, aperture.open );
    }
}

static void
print_mes (CTDB             *ctdb,
	   CTDB_VOLUME_INFO *volume)
{
    CTDB_VOLUME_MODEL  *volume_model;
    CTDB_MES_VOLUME    *mes_volume;
    CTDB_MES_TEMPLATE  *mes_template;

    volume_model = ctdb->mes_volume_models[volume->reference];
    mes_volume   = &(ctdb->mes_volumes[volume_model->mes_volume_index]);

    print_triangles ( ctdb, mes_volume);
#if 0
    print_boxes ( ctdb, mes_volume );
#endif
}

static CTDB_TERRAIN_SUBCLASS
print_volume_info(CTDB_FEATURE_DATA *f)
{
   printf("ref=%5d\ttype=%2d\tavoid=%1d\n",f->volume.reference,
	  f->volume.type,f->volume.avoid);
   return f->volume.type;
}

static int32
print_volume(
	     CTDB                   *ctdb, 
	     CTDB_FEAT_OBJ_PTR       obj,
	     CTDB_FEATURE_DATA      *feature,
	     CTDB_PRINT_OBJ_DATA    *pdata,
	     int32                  *next
	     )
{
    int32 verts,i;
    CTDB_TERRAIN_SUBCLASS type;

    verts = feature[0].info.vertices;   
    *next = NEXT_VOLUME(verts);

    print_feature_info(&feature[0], pdata->n_feat);
    type = print_volume_info(&feature[1]);

    if (type == CTDB_FS_MES)
      printf("MES Volume ID = %10d\n",feature[2]);
    else
      printf("    Volume ID = %10d\n",feature[2]);
      
    for (i = 0; i < verts; i++)
    {
	print_xy_patch_offset(ctdb, &feature[3+i]);
	print_z(&feature[4+i]);
    }

    if( type == CTDB_FS_MES )
      print_mes(ctdb, &feature[1].volume);

    return 1;
}

static int32
print_linear(
	     CTDB                   *ctdb, 
	     CTDB_FEAT_OBJ_PTR      obj,
	     CTDB_FEATURE_DATA      *feature,
	     CTDB_PRINT_OBJ_DATA    *pdata,
	     int32                  *next
	     )
{
    int32 verts,i;
    
    verts = feature->info.vertices;   
    *next = NEXT_LINEAR(verts);
    print_feature_info(feature++, pdata->n_feat);
    ctdb_print_linear(ctdb, feature++);
    
    for (i = 0; i < verts; i++)
    {
	print_xy_patch_offset(ctdb, feature++);
	print_z(feature++);
    }
    return 1;
}

static int32
print_canopy(
	     CTDB                   *ctdb, 
	     CTDB_FEAT_OBJ_PTR      obj,
	     CTDB_FEATURE_DATA      *feature,
	     CTDB_PRINT_OBJ_DATA    *pdata,
	     int32                  *next
	     )
{
    int32 verts,i;

    verts = feature->info.vertices;   
    *next = NEXT_CANOPY(verts);

    print_feature_info(&feature[0], pdata->n_feat);
    ctdb_print_canopy_pts(&feature[3]);
    print_xy_patch_offset(ctdb, &feature[1]);
    print_z(&feature[2]);
    return 1;
}

static int32
print_aggregate(
		CTDB                   *ctdb, 
		CTDB_FEAT_OBJ_PTR      obj,
		CTDB_FEATURE_DATA      *feature,
		CTDB_PRINT_OBJ_DATA    *pdata,
		int32                  *next
		)
{
    int32 verts,i;

    verts = feature->info.vertices;   
    *next = NEXT_AGGREGATE(verts);
    print_feature_info(feature++, pdata->n_feat);
    ctdb_print_aggregate(ctdb, feature++);
    print_xy_patch_offset(ctdb, feature++);

    return 1;
}

static int32
print_laid_linear(
		  CTDB                   *ctdb, 
		  CTDB_FEAT_OBJ_PTR       obj,
		  CTDB_FEATURE_DATA      *feature,
		  CTDB_PRINT_OBJ_DATA    *pdata,
		  int32                  *next
		  )
{
    int32 verts,i;
    
    verts = feature->info.vertices;   
    *next = NEXT_LAID_LINEAR(verts);

    print_feature_info(feature++, pdata->n_feat);
    ctdb_print_laid_linear(ctdb, feature++);
    feature++; /* New int32 here */

    for (i = 0; i < verts; i++)
      print_xy_patch_offset(ctdb, feature++);

    return 1;
}


static int32
print_micro(CTDB                   *ctdb, 
	    CTDB_FEAT_OBJ_PTR      obj,
	    CTDB_FEATURE_DATA      *feature,
	    CTDB_PRINT_OBJ_DATA    *pdata,
	    int32                  *next
	    )
{
    int32 verts = feature->info.vertices;
    
    *next = NEXT_MICRO(verts);

    print_feature_info(feature++, pdata->n_feat);
    feature++;   /* New int32 here */

    print_xy_patch_offset(ctdb, feature++);
    print_z(feature++);
    print_micro_pts(feature++, verts);

    if ( verts & MICRO_TYPE_WATER)
      print_micro_characteristics(feature++);

    return 1;
}


int32
ctdb_print_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
   obj[FC_MICRO].func         = (CTDB_OBJ_FUNC)print_micro;
   obj[FC_VOLUME].func        = (CTDB_OBJ_FUNC)print_volume;
   obj[FC_LINEAR].func        = (CTDB_OBJ_FUNC)print_linear;
   obj[FC_CANOPY].func        = (CTDB_OBJ_FUNC)print_canopy;
   obj[FC_LAID_LINEAR].func   = (CTDB_OBJ_FUNC)print_laid_linear;
   obj[FC_AGGREGATE].func     = (CTDB_OBJ_FUNC)print_aggregate;
   bcopy(obj,print_obj,sizeof(print_obj));
   return 1;
}

void 
ctdb_print_features(CTDB *ctdb, int32 patch_x, int32 patch_y)
{
    CTDB_FEATURE_DATA *feature, *last;
    int32 i = 0;
    int32 n_feature = 0;
    int32 patch_number = patch_x + (patch_y * ctdb->patches_wide);
    int32 x = patch_x * ctdb->patch_incr * ctdb->incr;
    int32 y = patch_y * ctdb->patch_incr * ctdb->incr;
    int32 next;
    CTDB_PRINT_OBJ_DATA data;
    CTDB_FEAT_OBJ_PTR obj;
    
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
    {
	printf("No features in patch\n");
	return;
    }
    
    printf("\n---Features in patch %d (%d, %d)---\n", 
	   patch_number, patch_x, patch_y);
    printf("---  implied patch lower left is at (%.2f, %.2f)---\n",
	   (float64)x, (float64)y);
    
    last = feature + feature->hdr.size;
    
    print_patch_hdr(feature);

    /* Initialize the object pointer to the beginning of the object class
     * structure.
     */
    obj = print_obj;
    data.n_feat = 0;
    
    for(feature++;feature<last;feature+=next,data.n_feat++)
    {
	int32 fclass = feature->info.feature_class;
    
	if ( obj->sensClass != fclass )
	  obj = &print_obj[fclass];
		
	(void)obj->func(ctdb, obj, feature, (ADDRESS)&data, &next);
    }
    
    ctdb_print_te_map(ctdb,patch_number);
}

static char *datum_names[] = { "ED50" ,
				 "ED79" ,
				 "WGS72" ,
				 "WGS84" ,
				 "CONUSNAD27" ,
				 "NASBNAD27" ,
				 "NASANAD27" ,
				 "NAD83" ,
				 "?8?" , "?9?" , "?10?" , "?11?" , "?12?" ,
			   };

void
ctdb_print_description(CTDB *ctdb)
{
    int32 pages = ctdb->pages_wide * ctdb->pages_high;
    int32 i, total, cell;
    float64 min_x,min_y,max_x,max_y;
    
    cell = ctdb_get_cell_id(ctdb);
    
    if (!ctdb)
    {
	printf("Database is not defined\n");
	return;
    }
    /* ---------------------------------------------------------------------
     *               TERRAIN CREATION INFORMATION           
     * ---------------------------------------------------------------------
     */
    if ( cell != GCS_ILLEGAL_CELL )
    {
	printf("\nCELL %05d\n",cell);
	printf("-------------------------------------------------------\n");
	printf("Generation Date    : %s\n",ctdb->date);

	/* either recompiler or s1000 compiler */
/*	printf("Generation Tool    : %s\n","recompile");*/
	printf("Source Data Set    : %s-%04d\n",ctdb->name, ctdb->version);
    }
    else
      printf("\nDatabase %s-%04d created %s\n",ctdb->name, ctdb->version, 
	     ctdb->date);
    
    /* ---------------------------------------------------------------------
     *               TERRAIN FORMAT AND OPTIONS             
     * ---------------------------------------------------------------------
     */
    printf("-------------------------------------------------------\n");
    printf("Terrain Format %d with the following features:\n",
	   ctdb->input_format);
    
    if ( cell != GCS_ILLEGAL_CELL )
      printf("GCS,");
    else
      printf("UTM %s,", ctdb->is_curved ? "curved" : "flat");
    
    printf("%s%s%s%s,",
	   (ctdb->is_gridded&&ctdb->is_tinned)?" MIXED":" PURE",
	   (ctdb->is_tinned)?" TIN":"",
	   (ctdb->is_gridded&&ctdb->is_tinned)?" &":"",
	   (ctdb->is_gridded)?" GRID POST":"");
    
    printf("\n");
    
    /* ---------------------------------------------------------------------
     *               DATABASE EXTENT AND ORIGIN INFORMATION
     * ---------------------------------------------------------------------
     */
    printf("-------------------------------------------------------\n");
    if ( ctdb->is_gridded )
      printf("Grid Spacing     (METERS): %d\n",ctdb->incr);
    
    printf("Fixed point basis        : %1.8f\n",ctdb->fixed_point_basis);
    
    if ( cell != GCS_ILLEGAL_CELL )
    {
	float64 x,y;
	
	ctdb_get_origin(ctdb,&x,&y);
	printf("Tangency Point (X,Y)     : (%10.2f, %10.2f, %05d)\n",-x,-y,cell);
    }
    else
    {
	printf("Origin at %0.0fN %0.0fE in UTM zone %d%c (datum %s)\n",
	       ctdb->origin_northing, ctdb->origin_easting,
	       ctdb->origin_zone_number, ctdb->origin_zone_letter,
	       datum_names[ctdb->datum]);
    }
    
    ctdb_get_extent(ctdb,&min_x,&min_y,&max_x,&max_y);
    
    printf("Minimum (SW Corner)(X,Y) : (%10.2f, %10.2f)\n",min_x,min_y);
    printf("Maximum (NE Corner)(X,Y) : (%10.2f, %10.2f)\n",max_x,max_y);
    printf("Minimum Elevation        : (%10.2f)\n",ctdb->min_z);
    printf("Maximum Elevation        : (%10.2f)\n",ctdb->max_z);
    
    /* ---------------------------------------------------------------------
     *               DATABASE MEMORY USAGE INFORMATION      
     * ---------------------------------------------------------------------
     */
    printf("-------------------------------------------------------\n");
    printf("%d nodes (%2.0fKB), %d edges (%2.0fKB), %d abstract data \
(%2.0fKB)\n",
	   ctdb->num_nodes, ctdb->node_size / KAY,
	   ctdb->num_edges, ctdb->edge_size / KAY,
	   ctdb->num_quad_data,
	   ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA) / KAY);

    if ( ctdb->is_gridded ) 
    {
	if (ctdb->elevations_resident || !pages)
	  printf("Grid posts completely in memory (%2.0fKB)\n",
		 pages * PAGE_SIZE / KAY);
	else
	  printf("%2.2f%% of grid posts in memory (%2.0fKB)\n",
		 100.0 * ctdb->elevation_cache_pages / pages,
		 ctdb->elevation_cache_pages * PAGE_SIZE / KAY);

	if (ctdb->elevations_resident || !pages)
	  printf("Poly Attrib posts completely in memory (%2.0fKB)\n",
		 pages * PAGE_SIZE / KAY);
	else
	  printf("%2.2f%% of poly attrib posts in memory (%2.0fKB)\n",
		 100.0 * ctdb->soil_cache_pages / pages,
		 ctdb->soil_cache_pages * PAGE_SIZE / KAY);
    }

    if (ctdb->patch_data || !ctdb->n_patch_bytes)
      printf("Patch group data completely in memory (%2.0fKB)\n",
	     ctdb->n_patch_bytes/KAY);
    else
    {
        total = 0;
	for(i=0;i<CTDB_PATCH_CACHES;i++)
	  total += (ctdb->patch_cache_size_map[i] * ctdb->patch_cache_entries *
                    CTDB_CACHE_BKT_UNITS);

	printf("%2.2f%% of patch data in memory (%2.0fKB)\n%d entries/cache\n",
	       100.0 * total / ctdb->n_patch_bytes,
	       total / KAY,
	       ctdb->patch_cache_entries);
    }

    printf("%2.0fKB total memory usage\n", ctdb->total_memory / KAY);
}

void 
ctdb_print_stats(CTDB *ctdb)
{
    if (!ctdb)
    {
	printf("ctdb is NULL.  No hit rate data available\n");
	return;
    }
    if (ctdb->elevation_hit_count || ctdb->elevation_miss_count)
    {
        printf("Elevation hit rate: %2.2f%%\n",
               100.0 * ctdb->elevation_hit_count /
               (ctdb->elevation_hit_count + ctdb->elevation_miss_count));
    }
    ctdb->elevation_hit_count = ctdb->elevation_miss_count = 0;
    if (ctdb->patch_hit_count || ctdb->patch_miss_count)
    {
        printf("Patch data hit rate: %2.2f%%\n",
               100.0 * ctdb->patch_hit_count /
               (ctdb->patch_hit_count + ctdb->patch_miss_count));
    }
    ctdb->patch_hit_count = ctdb->patch_miss_count = 0;
}

void
ctdb_elev_data_to_string(CTDB_ELEV_DATA *elev,
			 char text[132])
{
    char *subclass;
    char tmp_string[132];
    
    switch (elev->sensClass)
    {
      case CTDB_FC_MICRO:
	{
	    static char sub[CTDB_NUM_MICRO_SUBCLASSES+1][20]=
	    {
		"Unknown",
		"Terrain Skin",
		"Multilevel Terrain",
		"Water",
		"Road"     
	    };
	    
	    subclass = &sub[elev->subclass][0];
	    break;
	}
      case CTDB_FC_VOLUME:
	{
	    static char sub[CTDB_NUM_VOLUME_SUBCLASSES][20]=
	    {
		"Generic Building",
		"Rock Drop",
		"MultiElev",
		"Plug"
	    };
	    static char mes_sub[CTDB_MES_LAST_TYPE][20]=
	    {
		"Unknown",
		"Building",
		"Cave",
		"Tunnel",
		"Bridge"
	    };
	    
	    if (elev->subclass == CTDB_FS_MES)
	    {
		/* $$$ current only handle MES buildings */
		sprintf(tmp_string,"%s %s",sub[elev->subclass],
			mes_sub[CTDB_MES_BUILDING]);
		subclass = tmp_string;
	    }
	    else
	      subclass = &sub[elev->subclass][0];
	    break;
	}
      case CTDB_FC_LINEAR:
	{
	    static char sub[CTDB_NUM_LINEAR_SUBCLASSES][20]=
	    {
		"Treeline",
		"Concertina Wire",
		"Wire Road Block",
		"Generic Fence",
		"Minefield Fence",
		"Dragons Teeth"
	    };
	    
	    subclass = &sub[elev->subclass][0];
	    break;
	}
      case CTDB_FC_CANOPY:
	{
	    static char sub[1][20]=
	    {
		"Canopy"
	    };
	    
	    subclass = &sub[elev->subclass][0];
	    break;
	}
      case CTDB_FC_LAID_LINEAR:
	{
	    static char sub[5][20]=
	    {
		"N/A",
		"N/A",
		"N/A",
		"Water",
		"Road",
	    };
	    
	    subclass = &sub[elev->subclass][0];
	    break;
	}
      default:
	{   
	    static char sub[1][20]=
	    {
		"Unknown"
	    };
	    
	    subclass = &sub[elev->subclass][0];
	    break;
	}
    }

    sprintf(text,"%7.2lf (%-15s)\0",elev->z,subclass);
}
