#ifndef lint
static char rcsid [] = "$RCSfile: ct_prof.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_prof.c                                                         *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to find the elevation profile of a vector            *
 *       Created: Mon Oct 21 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/
#include <stdext.h>
#include <libgcs.h>
#include <libvecmat.h>
#include "libctdb_local.h"

/* Global Local Variables */
static CTDB *ctdb_ptr;
static CTDB_FEATURE_OBJ prof_obj[FC_CLASS_MAX];
static CTDB_FEAT_TYPE   prof_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];
static int32 ctdb_prof_limit;
static int32 ctdb_n_prof;
static int32 ctdb_cell_list_start;	/* Starting index in list for this 
					 * cell */
static int32 ctdb_result_cell;	/* Cell in whose coordinate system all 
				 * results should be returned
				 */
static int32 ctdb_include;
static CTDB_ELEV_QUALIFIER prof_qual;
static int32 prof_no_water;
static int32 prof_water_found;


static CTDB_COORD3D ctdb_start;
typedef struct prof_pt
{
    float64 pt[XYZ];
    enum
    {       
	PROF_TERRAIN       = 1, /* SKIN, BASE, DEFAULT */
	PROF_MICRO_WATER   = 2,
	PROF_PATCH_WATER   = 4,
	PROF_WATER         = 6
    } subtype ;
} PROF_PT;

typedef struct ctdb_prof_obj_data
{
  CTDB_LOS_IN_XYO_AND_DB los;
  CTDB_PATCH_DATA        patch;
  uint32                 postmask;
} CTDB_PROF_OBJ_DATA;


static PROF_PT *prof_pts,*prof_pts_head;


static int32
cmp_prof(PROF_PT *pt0, PROF_PT *pt1)
{
    float64 dx, dy, dist0, dist1;
    
    dx = (pt0->pt[X] - ctdb_start.x);
    dy = (pt0->pt[Y] - ctdb_start.y);
    dist0 = dx * dx + dy * dy;
    
    dx = (pt1->pt[X] - ctdb_start.x);
    dy = (pt1->pt[Y] - ctdb_start.y);
    dist1 = dx * dx + dy * dy;

    return (dist0-dist1);
}

#define VERY_NEGATIVE_NUMBER -1e100
static int32
consolidate_list(void)
{
    float64 z;
    CTDB_ELEV_QUALIFIER qual;
    static CTDB_ELEV_QUALIFIER lqual=CTDB_MAX_LAND_NO_WATER;
    static CTDB_ELEV_QUALIFIER wqual=(CTDB_RETURN_MAX|CTDB_INCL_WATER);
    int32 last_i,i,tossed;
    static PROF_PT inval_pt=
    {
	VERY_NEGATIVE_NUMBER,VERY_NEGATIVE_NUMBER,VERY_NEGATIVE_NUMBER,
	PROF_TERRAIN
    };   
    PROF_PT *last_pt=&inval_pt;
    
    
    /* This routine will loop through all of the terrain points and toss
     * any unneeded points as either the final step before returning the
     * list to the caller or as a method of thinning the list when there
     * are to many points.
     */

    (void)qsort(prof_pts_head, ctdb_n_prof, sizeof(PROF_PT), 
		(CTDB_QSORT)cmp_prof);

    for (i=0,tossed=0,last_i=0;i<ctdb_n_prof;i++,(last_i=i-tossed))
    {
	/* If there was no water found during this profile then there is 
	 * no need to process the list to determine switch over points.
	 */
	if ( prof_water_found )
	{
	    /* $$$This is probably not the most efficient way of thinning the 
	     * data points but it is the only one that seemed to work for
	     * patch water in both SIMNET and GCS.
	     */
	    qual = (prof_pts_head[i].subtype & PROF_WATER )?lqual:wqual;
	    z = ctdb_lookup_qual_elevation(ctdb_ptr,
					   prof_pts_head[i].pt[X],
					   prof_pts_head[i].pt[Y],qual,
					   CTDB_NO_REF_ELEV,
					   (CTDB_ELEV_DATA *)NULL);

	    /* If the other type of terrain is above this point then toss 
	     * the point.  
	     */
	    if ( z > prof_pts_head[i].pt[Z] )
	    {
		tossed++;
		continue;
	    }
	}
	
    
	/* Check if this is a duplicate point */
	if ( VMAT3_VEC_EQUAL64(&last_pt->pt[X],&prof_pts_head[i].pt[X]) )
	{
	    tossed++;
	    continue;
	}
	
	/* Save the last point added */
	last_pt = &prof_pts_head[i];
	
	/* Don't start copying points around unless a point has been 
	 * tossed 
	 */
	if ( tossed )
	  bcopy(&prof_pts_head[i],&prof_pts_head[last_i],sizeof(PROF_PT));
    }    
    
    /* Reset the point count and the pointer to the next point to add */
    ctdb_n_prof = last_i;
    prof_pts    = &prof_pts_head[ctdb_n_prof];   
    
    if (ctdb_n_prof < ctdb_prof_limit)
      return VIS_KEEP_GOING;
    else
      return VIS_STOP;
}

/* x,y are in post coordinates in ctdb's native coordinate system */
static int32 
prof_post_compare(
		  CTDB *ctdb, 
		  float64 x, 
		  float64 y, 
		  float64 zl, 
		  float64 zh,
		  float64 z_cross,
		  int32   water
		  )
                   /* argument retained to maintain consistent interface
		    * for calling out of ctdb_intervis 
		    */
                       
                     
{
    if ( water && prof_no_water )
      return VIS_KEEP_GOING;
 
    /* Figure out where the posts cross the line and store that location
     * and elevation.
     */
    prof_pts->pt[X] = x * ctdb->incrD;
    prof_pts->pt[Y] = y * ctdb->incrD;
    prof_pts->pt[Z] = z_cross;

    if ( water )
    {
	prof_water_found  = TRUE;
	prof_pts->subtype = PROF_PATCH_WATER;
    }
    else
      prof_pts->subtype = PROF_TERRAIN;
    
    ctdb_n_prof++; prof_pts++;
    
    if (ctdb_n_prof < ctdb_prof_limit)
      return VIS_KEEP_GOING;
    else
      return consolidate_list();
}

/* LOS is entire intervis ray stored in current coord system */
static int32 
prof_te_edge_compare(
		     CTDB *ctdb, 
		     CTDB_LOS_IN_XYO_AND_DB *los,
		     float64 inz0, 
		     float64 inz1, 
		     float64 tedge, 
		     float64 tray
		     )
{
    float64 z;
    
    z = (inz1 - inz0) * tedge + inz0;
    
    prof_pts->pt[X] =
      los->db.start.x + tray * (los->db.end.x - los->db.start.x);
    prof_pts->pt[Y] =
      los->db.start.y + tray * (los->db.end.y - los->db.start.y);
    prof_pts->pt[Z]    = z;
    prof_pts->subtype = PROF_TERRAIN;
    ctdb_n_prof++; prof_pts++;
    
    if (ctdb_n_prof < ctdb_prof_limit)
      return VIS_KEEP_GOING;
    else
      return consolidate_list();
}

/* LOS is entire intervis ray stored in current coord system */
static int32 
prof_edge_compare(
		  CTDB_FEATURE_DATA      *v0, 
		  CTDB_FEATURE_DATA      *v1, 
		  CTDB_LOS_IN_XYO_AND_DB *los 
		  )
{
    float64 tray, tedge;
    float64 z;
    
    /* Figure out where the posts cross the line and store that location
     * and elevation.
     */
    
    if (ctdb_segment_intersection_t(&los->xyo, v0->xy, v1->xy, &tray, &tedge) 
	== DISJOINT)
      return VIS_KEEP_GOING;
    
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    
    prof_pts->pt[X] =
      los->db.start.x + tray * (los->db.end.x - los->db.start.x);
    prof_pts->pt[Y] =
      los->db.start.y + tray * (los->db.end.y - los->db.start.y);
    prof_pts->pt[Z] = z;
    prof_pts->subtype = PROF_TERRAIN;
    
    ctdb_n_prof++; prof_pts++;
    
    if (ctdb_n_prof < ctdb_prof_limit)
      return VIS_KEEP_GOING;
    else
      return consolidate_list();
}


/* LOS is entire intervis ray stored in current coord system */
static int32 
prof_water_edge_compare(
			CTDB_FEATURE_DATA      *v0, 
			CTDB_FEATURE_DATA      *v1, 
			CTDB_LOS_IN_XYO_AND_DB *los,
			float32                 tide
			)
{
    float64 tray, tedge;
    float64 z;
    
    /* Figure out where the posts cross the line and store that location
     * and elevation.
     */
    
    if (ctdb_segment_intersection_t(&los->xyo, v0->xy, v1->xy, &tray, &tedge) 
	== DISJOINT)
      return VIS_KEEP_GOING;
    
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z + tide;
    
    prof_pts->pt[X] =
      los->db.start.x + tray * (los->db.end.x - los->db.start.x);
    prof_pts->pt[Y] =
      los->db.start.y + tray * (los->db.end.y - los->db.start.y);
    prof_pts->pt[Z] = z;
    prof_water_found  = TRUE;
    prof_pts->subtype = PROF_MICRO_WATER;
    
    ctdb_n_prof++; prof_pts++;
    
    if (ctdb_n_prof < ctdb_prof_limit)
      return VIS_KEEP_GOING;
    else
      return consolidate_list();
}

static int32
prof_check_micro(
		 CTDB               *ctdb,
		 CTDB_FEAT_OBJ_PTR   obj, 
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_PROF_OBJ_DATA *data,
		 int32              *next
		 )
{
    static CTDB_FEAT_TYPE *sub=prof_micro_subclass_obj;
    int32 verts = feature->info.vertices;

    /* check if the micro type has changed */
    if ((verts & MICRO_TYPE) != sub->type)
      sub = &prof_micro_subclass_obj[MICRO_TYPE_INDEX(sub->verts)];

    *next = sub->incr;
    
    if (!obj->valid || 
	!(data->postmask & feature->integer) ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return VIS_KEEP_GOING;

    /* call the micro type processing function */
    return sub->func(ctdb,sub,feature,data);
}



static int32
prof_micro_edge_compare(CTDB               *ctdb,
			CTDB_FEAT_TYPE_PTR sub,
			CTDB_FEATURE_DATA  *feature, 
			CTDB_PROF_OBJ_DATA *data
			)
{
#ifdef WATER_DEBUG
    printf("MCR ,");
#endif
     feature += 2;
    if (!prof_edge_compare(feature,feature+(sub->incr*feature[2].micro.vert0), 
			   &(data->los)))
      return VIS_STOP; 
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    if (!prof_edge_compare(feature,feature+(sub->incr*feature[2].micro.vert1),
			   &(data->los)))
      return VIS_STOP; 
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    if (!prof_edge_compare(feature,feature+(sub->incr*feature[2].micro.vert2), 
			   &(data->los)))
      return VIS_STOP; 

    return VIS_KEEP_GOING;    
}



static int32
prof_micro_water_compare(CTDB               *ctdb,
			 CTDB_FEAT_TYPE_PTR  sub, 
			 CTDB_FEATURE_DATA  *feature, 
			 CTDB_PROF_OBJ_DATA *data
			 )
{
    float32 tide1,tide2;
#ifdef WATER_DEBUG
    printf("WMCR ,");
#endif
    
    feature += 2;
    
    tide1 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char0);
    
    if (!prof_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert0), 
				 &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    if (!prof_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert1),
				 &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    tide2 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[3].mwater.char1);

    /* Since the tidal level may shift at the center edge of a polygon
     * check it again with the new tidal level if the tidal level is 
     * different from the first.
     */
    if ((tide2 != tide1) &&
	(!prof_water_edge_compare(feature,
				  feature+(sub->incr*feature[2].micro.vert0),
				  &(data->los),tide2)))
	  return VIS_STOP;
    
    if (!prof_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert2), 
				 &(data->los),tide2))
      return VIS_STOP; 

    return VIS_KEEP_GOING;    
}

int32
ctdb_prof_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static CTDB_TYPE_FUNC micro_func[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)prof_micro_edge_compare,      /* MICRO_TYPE_BASE */ 
	(CTDB_TYPE_FUNC)ctdb_invalid_type   ,         /* MICRO_TYPE_ML */
	(CTDB_TYPE_FUNC)prof_micro_edge_compare,      /* MICRO_TYPE_DEFAULT */
	(CTDB_TYPE_FUNC)prof_water_edge_compare       /* MICRO_TYPE_WATER */
    };

    /* CLASS FC_MICRO */
    obj[FC_MICRO].func         = (CTDB_OBJ_FUNC)prof_check_micro;
    ctdb_init_micro_subclass(prof_micro_subclass_obj,micro_func);

    obj[FC_VOLUME].valid       = 0;
    obj[FC_VOLUME].func        = (CTDB_OBJ_FUNC)ctdb_invalid_volume;
    obj[FC_LINEAR].valid       = 0;
    obj[FC_LINEAR].func        = (CTDB_OBJ_FUNC)ctdb_invalid_linear;
    obj[FC_CANOPY].valid       = 0;
    obj[FC_CANOPY].func        = (CTDB_OBJ_FUNC)ctdb_invalid_canopy;
    obj[FC_LAID_LINEAR].valid  = 0;
    obj[FC_LAID_LINEAR].func   = (CTDB_OBJ_FUNC)ctdb_invalid_laid_linear;
    obj[FC_AGGREGATE].valid    = 0;
    obj[FC_AGGREGATE].func     = (CTDB_OBJ_FUNC)ctdb_invalid_aggregate;

    bcopy(obj,prof_obj,sizeof(prof_obj));
    
    return 1;
}



/* start, end are the endpoints of the *entire* intervis ray, represented in
 * the frame of reference specified by ctdb.
 */
   
static int32 
prof_patch_compare(
		   CTDB    *ctdb, 
		   int32    patch_number, 
		   int32    feature_mask, 
		   int32    post_mask, 
		   float64 *start, 
		   float64 *end
		   )
{
    int32 ray_info,i,next;
    CTDB_FEATURE_DATA *feature, *last;
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_PROF_OBJ_DATA data;
    
    /* If there is no microterrain to be processed, return since that is 
     * the only type of feature that will be processed here.
     */
    if (!(feature_mask & prof_obj[FC_MICRO].featmask))
      return VIS_KEEP_GOING;
    
    feature = (CTDB_FEATURE_DATA *)ctdb_lookup_feature(ctdb, patch_number);

    if ( !feature )
      return VIS_KEEP_GOING;

    data.postmask     = post_mask;
    data.patch.number = patch_number;
    data.patch.x      = patch_number % ctdb->patches_wide;
    data.patch.y      = patch_number / ctdb->patches_wide;

    data.los.xyo.start.x = POST_TO_XYO(start[X], data.patch.x, 
				       ctdb->patch_inv_incr);
    data.los.xyo.start.y  = POST_TO_XYO(start[Y], data.patch.y, 
					ctdb->patch_inv_incr);
    data.los.xyo.end.x = POST_TO_XYO(end[X], data.patch.x, 
				     ctdb->patch_inv_incr);
    data.los.xyo.end.y = POST_TO_XYO(end[Y], data.patch.y, 
				     ctdb->patch_inv_incr);
    data.los.db.start.x = start[X] * ctdb->incrD;
    data.los.db.start.y = start[Y] * ctdb->incrD;
    data.los.db.end.x = end[X] * ctdb->incrD;
    data.los.db.end.y = end[Y] * ctdb->incrD;

#ifdef CTDB_OBJ_DEBUG
    printf("\npatch number %d (%d,%d)\t",patch_number,
	   data.patch.x,data.patch.y);
#endif    

    last = feature + feature->hdr.size;
    feature++;

    /* initialize the feature valid bit, initialize micro terrain processing
     * only since processing will be terminated after non-microterrain is 
     * encountered.
     */
    prof_obj[FC_MICRO].valid = TRUE;
    obj = prof_obj;
    
    for(;feature<last; feature+=next)
    {
	int32 fclass=feature->info.feature_class;
	
	/* if the current class doesn't match the previous class
	 * then return (note only micro terrain is viewed here).
	 */
	if ( fclass != FC_MICRO )
	  return VIS_KEEP_GOING;

	if ( obj->func(ctdb,obj,feature,(ADDRESS)&data,&next) == VIS_STOP)
	  return VIS_STOP;
    }
    return VIS_KEEP_GOING;
}



static int32 
prof_cell_fcn(
	      CTDB *ctdb, 
	      float64 *ray_start, 
	      float64 *ray_end,
	      float64 *cur_start, 
	      float64 *cur_end, 
	      float64 width
	      )
{
    int32 ret;
    int32 i;
    float64 ray_start_post[3];
    float64 ray_end_post[4];
    float64 cur_start_post[4];
    float64 cur_end_post[4];
    CTDB_COORD3D start, end, cell_start, cell_end;
    
    /* Scale from meters to grid spacing */
    ray_start_post[X] = ray_start[X] * ctdb->inv_incr;
    ray_start_post[Y] = ray_start[Y] * ctdb->inv_incr;
    ray_start_post[ZL] = ray_start[ZL];
    
    ray_end_post[X] = ray_end[X] * ctdb->inv_incr;
    ray_end_post[Y] = ray_end[Y] * ctdb->inv_incr;
    ray_end_post[ZL] = ray_end[ZL];
    ray_end_post[ZH] = ray_end[ZH];
    
    cur_start_post[X] = cur_start[X] * ctdb->inv_incr;
    cur_start_post[Y] = cur_start[Y] * ctdb->inv_incr;
    cur_start_post[ZL] = cur_start[ZL];
    cur_start_post[ZH] = cur_start[ZH];
    
    cur_end_post[X] = cur_end[X] * ctdb->inv_incr;
    cur_end_post[Y] = cur_end[Y] * ctdb->inv_incr;
    cur_end_post[ZL] = cur_end[ZL];
    cur_end_post[ZH] = cur_end[ZH];
    
    /* Copy endpoints into structure used by terrain_element_intervis */
    start.x = ray_start[X];
    start.y = ray_start[Y];
    start.z = ray_start[Z];
    end.x   = ray_end[X];
    end.y   = ray_end[Y];
    end.z   = ray_end[ZL];
    cell_start.x = cur_start[X];
    cell_start.y = cur_start[Y];
    cell_start.z = cur_start[ZL];
    cell_end.x = cur_end[X];
    cell_end.y = cur_end[Y];
    cell_end.z = cur_end[ZL];
    
    ctdb_ptr = ctdb;
    /* call terrain element intervis. if there are tinned patches and we have
     * still have room in the points list 
     */
    if ( ctdb->is_tinned &&
	(ctdb_n_prof < ctdb_prof_limit) )
    {
	ret = ctdb_terrain_element_intervis(ctdb,&start,&end,ray_end[ZH],
					    &cell_start, &cell_end,
					    prof_te_edge_compare,
					    FALSE);
    }
    
    /* if still room for points and the database had gridded patches, then 
       traverse gridded patches */
    if ( ctdb_n_prof < ctdb_prof_limit )
    {
	/* Profile between the first and last points */
        /* if the database contains grids or looking for water (patch water)
         * then enter into the grid intervis engine.
         */
        if ( ctdb->is_gridded || !prof_no_water )
	  ret = ctdb_intervis(ctdb, ray_start_post, ray_end_post,
			      cur_start_post, cur_end_post, TRUE,
			      prof_post_compare, prof_patch_compare,
			      (ctdb->is_gridded?POST_MICROTERRAIN:0));
    }
    
    /* Convert points we just generated into calling frame of reference */
    if (ctdb_is_tiled)
    {
        int32 cur_cell = ctdb_get_cell_id(ctdb);
	if (cur_cell != ctdb_result_cell)
	{
	    for(i=ctdb_cell_list_start;i<ctdb_n_prof;i++)
	    {
	        gcs_gcs3e_to_gcscs(cur_cell, 
				 prof_pts[i].pt[X], 
				 prof_pts[i].pt[Y], 
				 prof_pts[i].pt[Z], 
				 ctdb_result_cell,
				 &prof_pts[i].pt[X], 
				 &prof_pts[i].pt[Y],
				 &prof_pts[i].pt[Z]);
	    }
	}
	/* Update for next cell */
	ctdb_cell_list_start = ctdb_n_prof;
    }
    
    return ret;
}

int32 
ctdb_profile_vector(
		    CTDB               *ctdb, 
		    float64             x0, 
		    float64             y0, 
		    float64             x1, 
		    float64             y1,
		    CTDB_ELEV_QUALIFIER qual,
		    int32              *n_prof,
		    float64             prof[][3]
		    )
{
    static CTDB_TYPE_FUNC *water_func=
      &prof_micro_subclass_obj[MICRO_TYPE_INDEX(MICRO_TYPE_WATER)].func;

    /* Define the DEFAULT and VALID ctdb_elev_qualifier bits */
    static CTDB_ELEV_QUALIFIER dflt=CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_QUALIFIER valid=CTDB_MAX_LAND_AND_WATER;
    static CTDB_ELEV_QUALIFIER must=CTDB_RETURN_MAX;

    int32   ret,i;
    static PROF_PT *lcl_prof_pts;
    static int32 n_lcl_prof_pts = 0;
    CTDB_ELEV_DATA elev0,elev1;
   
    if (!ctdb || (*n_prof < 2))
    {
	*n_prof = 0;
	return 0;
    }

    ctdb_ptr = ctdb;

    prof_qual  = (qual?(qual&valid)|must:dflt);
    ctdb_prof_limit = *n_prof - 1;
    ctdb_n_prof     = 0;
    
    /* Determine if the size of the internal profile points array is large
     * enough to store all of the points.
     */
    if ( ctdb_prof_limit > n_lcl_prof_pts )
    {
	n_lcl_prof_pts = (ctdb_prof_limit+100);
	if ( lcl_prof_pts )
	  lcl_prof_pts = (PROF_PT *)STDREALLOC(lcl_prof_pts,n_lcl_prof_pts 
						 * sizeof(PROF_PT));
	else
	  lcl_prof_pts = (PROF_PT *)STDALLOC(n_lcl_prof_pts*sizeof(PROF_PT));
    }
    
    prof_pts_head = prof_pts = lcl_prof_pts;
    
    /* We do not need to worry about checking if the points are
     * within the ctdb.  This is because the results of the
     * elevation lookups will be in ctdb's frame of reference
     */
    
    /* Store off result cell for use by later calls */
    ctdb_result_cell = ctdb_get_cell_id(ctdb);
    
    /* Grab z-values for the endpoints from the terrain */
    ctdb_lookup_qual_elevation(ctdb, x0, y0,prof_qual,CTDB_NO_REF_ELEV,&elev0);
    ctdb_lookup_qual_elevation(ctdb, x1, y1,prof_qual,CTDB_NO_REF_ELEV,&elev1);
    
    /* Store the first point */
    prof_pts->pt[X]   = x0;
    prof_pts->pt[Y]   = y0;
    prof_pts->pt[Z]   = elev0.z;
    if ( elev0.subclass == CTDB_FS_WATER )
    {
	prof_water_found  = TRUE;
	prof_pts->subtype = PROF_PATCH_WATER;
    }
    else
    {
	prof_water_found  = FALSE;
	prof_pts->subtype = PROF_TERRAIN;
    }
    
    ctdb_n_prof++; prof_pts++;
    ctdb_cell_list_start = *n_prof;

    /* If water is to be included in the profile then set the appropriate
     * processing routine into action.
     */
    if ( prof_qual & CTDB_INCL_WATER )
    {
	prof_no_water = FALSE;
	*water_func   = (CTDB_TYPE_FUNC)prof_water_edge_compare;
    }
    else
    {
	prof_no_water = TRUE;
	*water_func   = (CTDB_TYPE_FUNC)ctdb_invalid_type;
    }     
    
    /* Run through all cells hit by the ray and generate the profile vector
     * (all points will be stored in the return frame of reference)
     */
    
    ctdb_traverse_ray(ctdb, x0, y0, elev0.z, 
		      x1, y1, elev1.z, 0.0, 0.0, prof_cell_fcn);
    
    /* We rigged the ctdb_prof_limit to guarantee there will be room for
     * the end point.
     */
    prof_pts--;/* look at the last point */
    if ((prof_pts->pt[X] != x1) || (prof_pts->pt[Y] != y1))
    {
	prof_pts++;
	prof_pts->pt[X] = x1;
	prof_pts->pt[Y] = y1;
	prof_pts->pt[Z] = elev1.z;
	ctdb_n_prof++; prof_pts++;
    }
    
    /* Sort the points. */
    ctdb_start.x = x0;
    ctdb_start.y = y0;

    ctdb_prof_limit = *n_prof;
    ret = consolidate_list();

    /* Copy points from the local format into the output format required 
     * by this API.
     */
    for (i=0;i<ctdb_n_prof;i++)
      VMAT3_VEC_COPY64(prof_pts_head[i].pt,&prof[i][X]);
    
    *n_prof = ctdb_n_prof;

    return ret;
}
