#ifndef lint
static char rcsid [] = "$RCSfile: ct_quad.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_quad.c                                                         *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts*
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Quadtree traversal code                                   *
 *       Created: Sun Oct 23 1994                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdio.h>
#include "libctdb_local.h"
#include <stdalloc.h>
#include <stdext.h>

/* Chunk size used when allocating quad nodes */
#define ALLOC_CHUNK 32

/* Globals used during search */
static int32 q_x0, q_y0, q_x1, q_y1;

/* Add the current quad to the list, and recurse on its children
 * as needed.
 */
static void 
ctdb_add_quad(CTDB_SEARCH_DB *search_db, int32 offset, int32 lo_x, int32 lo_y,
	      int32 hi_x, int32 hi_y, int32 *alloced)
{
    CTDB *ctdb = search_db->ctdb;
    CTDB_FEATURE_DATA *patch;
    struct ctdb_quad_data *qdata;
    int32 mid_x, mid_y;
    uint32 to_do;
    
    /* Error check */
    if ((ctdb == NULL) || (offset >= ctdb->num_quad_data))
      return;
    
    /* Make room for more nodes */
    if (*alloced == search_db->num_quad_nodes)
    {
	int32 size;
	
	*alloced += ALLOC_CHUNK;
	size = (*alloced) * sizeof(uint32);
	if (search_db->quad_nodes)
	  search_db->quad_nodes = (uint32 *)
	    STDREALLOC(search_db->quad_nodes, size);
	else
	  search_db->quad_nodes = (uint32 *)STDALLOC(size);
    }
    
    /* Stuff in this quad node */
    search_db->quad_nodes[search_db->num_quad_nodes] = CTDB_QUAD_PATCH(offset);
    search_db->num_quad_nodes++;
    
    /* Get this node's data */
    patch = search_db->ctdb->quad_data + offset;
    if (!patch->hdr.size)
      return;
    
    /* See if this is a leaf quad node */
    if (patch[1].abstract.type_code != CTDB_ABSTRACT_QUAD_NODE)
      return;
    
    /* Get the starting addresses of the leaves */
    qdata = &(((CTDB_ABSTRACT_DATA *)(&patch[2].abstract_data))->quad_data);
    
    /* The leaves are stored: 23
     *                        01
     * See which ones we need to include (actually we test for which
     * ones we *don't* have to include, because that's easier).
     */
    mid_x = (lo_x + hi_x) / 2;
    mid_y = (lo_y + hi_y) / 2;
    
    /* Make a bit mask with bits 0-3 on, then turn off those which are
     * eliminated.
     */
    to_do = 0xf;
    if (q_x0 > mid_x) /* Eliminate 0 & 2, leave 1010 = 0xA */
      to_do &= 0xA;
    if (q_y0 > mid_y) /* Eliminate 0 & 1, leave 1100 = 0xC */
      to_do &= 0xC;
    if (q_x1 < mid_x) /* Eliminate 1 & 3, leave 0101 = 0x5 */
      to_do &= 0x5;
    if (q_y1 < mid_y) /* Eliminate 2 & 3, leave 0011 = 0x3 */
      to_do &= 0x3;
    
    /* Go down and add the leaf nodes which are present (a node is
     * allowed to have only some of its leaves used).
     */
    if ((to_do & 0x1) && qdata->child_start[0])
      ctdb_add_quad(search_db, qdata->child_start[0],
		    lo_x, lo_y, mid_x-1, mid_y-1, alloced);
    if ((to_do & 0x2) && qdata->child_start[1])
      ctdb_add_quad(search_db, qdata->child_start[1],
		    mid_x, lo_y, hi_x, mid_y-1, alloced);
    if ((to_do & 0x4) && qdata->child_start[2])
      ctdb_add_quad(search_db, qdata->child_start[2],
		    lo_x, mid_y, mid_x-1, hi_y, alloced);
    if ((to_do & 0x8) && qdata->child_start[3])
      ctdb_add_quad(search_db, qdata->child_start[3],
		    mid_x, mid_y, hi_x, hi_y, alloced);
}

/* Make a list of quads to be searched */
void 
ctdb_find_search_quads(CTDB_SEARCH_DB *search_db, int32 min_x, int32 min_y, 
		       int32 max_x, int32 max_y)
{
    CTDB *ctdb = search_db->ctdb;
    int32 alloced;
    
    /* Stick the search bounds into some static globals to avoid
     * argument-passing overhead.
     */
    q_x0 = min_x;
    q_y0 = min_y;
    q_x1 = max_x;
    q_y1 = max_y;
    
    /* Start the recursive search */
    alloced = 0;
    ctdb_add_quad(search_db, (int32)0,
		  ctdb->min_x, ctdb->min_y, ctdb->max_x, ctdb->max_y,
		  &alloced);
}
