#ifndef lint
static char rcsid [] = "$RCSfile: ct_search.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_search.c                                                       *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts*
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to iterate over features                             *
 *       Created: Sun Oct 23 1994                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdlib.h>
#include "libctdb_local.h"
#include <libgcs.h>
#include <math.h>
#include <stdalloc.h>
#include <stdext.h>
#include <stdcoord.h>
#include <libgcs.h>

/* Forward declarations */
static int32 advance_gcs_search_space(CTDB_SEARCH_SPACE_PTR space, 
				      CTDB_SEARCH_DB *current);
static void copy_search_db(CTDB_SEARCH_DB *src, CTDB_SEARCH_DB *sink);
static void initialize_gcs_search_spaces(CTDB *ctdb, 
					 CTDB_SEARCH_SPACE_PTR space);

static float64 last_road_dist;

/* Reset the iterators to the beginning of the search space */
void ctdb_reset_search_space(CTDB_SEARCH_SPACE_PTR space)
{
    int32     i;

    space->iteration.tree.offset = 0;
    space->iteration.tree.all_done = FALSE;
    space->iteration.building.offset = 0;
    space->iteration.building.all_done = FALSE;
    space->iteration.linear.offset = 0;
    space->iteration.linear.all_done = FALSE;
    for(i = 0; i < CTDB_ABSTRACT_MAX; i++)
    {
	space->iteration.abstract[i].offset = 0;
	space->iteration.abstract[i].all_done = FALSE;
    }

    /* We want to start all of our iterators off in the first valid
     * cell.  If we set the index to -1 and call advance_gcs_search_space,
     * it will find that for us, and we can copy that one everywhere.
     */
    space->search_patch.current_index = -1;
    
    advance_gcs_search_space(space, &space->search_patch);
    
    /* If we're traversing by patch, then space->search_patch is the
     * only CTDB_SEARCH_DB we'll use, so we won't fill out any of the
     * others.  Otherwise, we need to populate all of them.
     */
    if (!(space->control_flags & CTDB_BY_PATCH))
    {
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.tree.search_patch));
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.building.search_patch));
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.linear.search_patch));
	
	for(i=0; i<CTDB_ABSTRACT_MAX; i++)
	{
	    copy_search_db(&(space->search_patch), 
			   &(space->iteration.abstract[i].search_patch));
	}
    }
}

/* Create a search space for application-driven iteration */
CTDB_SEARCH_SPACE_PTR ctdb_create_search_space(CTDB *ctdb, 
					       int32 min_x, 
					       int32 min_y,
					       int32 max_x,
					       int32 max_y,
					       uint32 control_flags)
{
    CTDB_SEARCH_SPACE_PTR space = (CTDB_SEARCH_SPACE_PTR)
      ctdb_alloc(NULL, sizeof(struct ctdb_search_space),
		 "Creating search space");
    float64 lat_start, lng_start, lat_stop, lng_stop;
    float64 h;
    int32 success;
    
    bzero(space, sizeof(struct ctdb_search_space));
    space->control_flags = control_flags;
    
    /* GCS: INITIALIZATION OF SEARCH SPACE CELLS */
    space->call_ctdb = ctdb;

    if (!ctdb)
    {
	space->min_x = space->max_x = space->min_y = space->max_y = 0;
	space->initial_frame = GCS_ILLEGAL_CELL;
    }
    else
    {
	space->min_x = min_x;
	space->min_y = min_y;
	space->max_x = max_x;
	space->max_y = max_y;
	space->initial_frame = ctdb_get_cell_id(ctdb);
    }
    
    if (!ctdb_is_tiled || !ctdb)
    {
	space->n_cells = 1;
	space->cell_id_list = (int32 *)ctdb_alloc(NULL,
						  sizeof(int32),
						  "Search space cell list ");
	space->cell_id_list[0] = space->initial_frame;
        
    }
    else
    {
	/* find the list of cells to be searched */
	gcs_gcs_to_llz(space->initial_frame, 
		       (float64)min_x, (float64)min_y, (float64)0.0, 
		       &lat_start, &lng_start, &h, DATUM_WGS84); 
	gcs_gcs_to_llz(space->initial_frame, 
		       (float64)max_x, (float64)max_y,
		       (float64)0.0, &lat_stop, &lng_stop, &h, DATUM_WGS84);
	space->n_cells = 100; /* initial maximum number of cells */
	space->cell_id_list = 
	  (int32 *)ctdb_alloc(NULL,
			      sizeof(int32)*space->n_cells, 
			      "Search space cell list ");
	success = gcs_extent_to_cell_list(lat_start, lng_start, 
					  lat_stop, lng_stop, 
					  &space->n_cells, 
					  space->cell_id_list);
	/* retry if cell_id_list not big enough */
	if (!success)
	{
	    space->cell_id_list = (int32 *)
	      ctdb_alloc(space->cell_id_list,
			 sizeof(int32)*space->n_cells, 
			 "Search space cell list ");
	    success = gcs_extent_to_cell_list(lat_start, lng_start, 
					      lat_stop, lng_stop, 
					      &space->n_cells, 
					      space->cell_id_list);
	}
    }
    
    /* Now that we know how many cells there are, allocate space for the
     * list of database iterators
     */
    space->cell_data_list = (CTDB_SEARCH_DB *)
      ctdb_alloc(NULL, space->n_cells * sizeof(CTDB_SEARCH_DB),
		 "Search space iterators");
    
    initialize_gcs_search_spaces(ctdb, space);
    
    return space;
}

static void  transform_abstract_vertices(CTDB_SEARCH_SPACE_PTR space, 
					 int32 index, int32 nverts, 
					 float32 *src, float32 *sink)
{
    int32 i;
    float64 tempz = 0.0;
    float64 newx, newy;
    
    for (i=0; i<nverts; i++)
    {
	gcs_gcs3e_to_gcscs(space->cell_id_list[index],
			 (float64)src[(i<<1)], (float64)src[(i<<1)+1], 0.0,
			 space->initial_frame,&newx, &newy, &tempz);
	sink[(i<<1)] = newx;
	sink[(i<<1)+1] = newy;
    }
}

static void transform_reference_frame(CTDB_SEARCH_SPACE_PTR space, int32 index,
				      float64 *x, float64 *y, float64 *z)
{
    float64 new_z;
    float64 new_x;
    float64 new_y;
    
    
    if (space->initial_frame == space->cell_id_list[index])
      return;
    else
      if (z)
      {
	  gcs_gcs3e_to_gcscs(space->cell_id_list[index], *x, *y, *z, 
			   space->initial_frame, &new_x, &new_y, &new_z);
	  *x = new_x;
	  *y = new_y;
	  *z = new_z;
      }
      else
      {
	  gcs_gcs3e_to_gcscs(space->cell_id_list[index], *x, *y, 0.0,  
			   space->initial_frame, &new_x, &new_y, &new_z);
	  *x = new_x;
	  *y = new_y;
      }
}

/* Set up a search_db to allow iterating over the specified cell */
static void
init_gcs_search_db(CTDB_SEARCH_SPACE_PTR space, CTDB_SEARCH_DB *search_db,
		   int32 index)
{
    float64 min_x, min_y, max_x, max_y;
    float64 temp_z;
    int32 success;
    int32 temp;
    CTDB *ctdb;
    
    bzero(search_db, sizeof(CTDB_SEARCH_DB));
    
    search_db->current_index = index;
    
    if (space->cell_id_list[index] == GCS_ILLEGAL_CELL)
    {
	search_db->ctdb = space->call_ctdb;
    }
    else
    {
	search_db->ctdb = (CTDB *)gcs_get_tdb(space->cell_id_list[index],
					      FALSE);
    }

    if (search_db->ctdb == NULL)
    {
	return;
    }
    
    /* Find min_x, min_y, max_x , max_y for this frame of reference */
    
    if ((ctdb_is_tiled) &&
	(space->cell_id_list[index] != space->initial_frame))
    {
	gcs_gcs3e_to_gcscs(space->initial_frame, (float64)space->min_x, 
			 (float64)space->min_y, (float64)0.0,	
			 space->cell_id_list[search_db->current_index],
			 &min_x, &min_y, &temp_z);
	
	gcs_gcs3e_to_gcscs(space->initial_frame,(float64)space->max_x, 
			 (float64)space->max_y,(float64)0.0,
			 space->cell_id_list[search_db->current_index],
			 &max_x, &max_y, &temp_z);
    }
    else
    {
	min_x = space->min_x;
	min_y = space->min_y;
	max_x = space->max_x;
	max_y = space->max_y;
    }
    
    if (min_x < search_db->ctdb->min_x)
      min_x = search_db->ctdb->min_x;
    if (min_y < search_db->ctdb->min_y)
      min_y = search_db->ctdb->min_y;
    if (max_x > (search_db->ctdb)->max_x)
      max_x = search_db->ctdb->max_x;
    if (max_y > (search_db->ctdb)->max_y)
      max_y = search_db->ctdb->max_y;
    
    /* Compute iteration bounds */
    search_db->min_patch_x = PATCH_X(min_x*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    search_db->max_patch_x = PATCH_X(max_x*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    search_db->min_patch_y = PATCH_Y(min_y*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    search_db->max_patch_y = PATCH_Y(max_y*search_db->ctdb->inv_incr*
				     search_db->ctdb->patch_inv_incr);
    
    if (!(space->control_flags & CTDB_NO_ABSTRACT))
      ctdb_find_search_quads(search_db, (int32)min_x, (int32)min_y, 
			     (int32)max_x, (int32)max_y);
    
    /* Bound search to database limits, and make sure min<max */
    search_db->min_patch_x=NS_LIMITS(0, search_db->min_patch_x, 
				     search_db->ctdb->patches_wide-1);
    search_db->min_patch_y=NS_LIMITS(0, search_db->min_patch_y, 
				     search_db->ctdb->patches_high-1);
    search_db->max_patch_x=NS_LIMITS(0, search_db->max_patch_x, 
				     search_db->ctdb->patches_wide-1);
    search_db->max_patch_y=NS_LIMITS(0, search_db->max_patch_y, 
				     search_db->ctdb->patches_high-1);
    /* Obscure case of an empty database, where the last bound is -1 */
    if (!search_db->ctdb->patches_wide || !search_db->ctdb->patches_high)
    {
	search_db->min_patch_x = search_db->min_patch_y = 
	  search_db->max_patch_x = search_db->max_patch_y = 0;
    }
    else
    {
	if (search_db->min_patch_x > search_db->max_patch_x)
	  search_db->min_patch_x = search_db->max_patch_x;
	if (search_db->min_patch_y > search_db->max_patch_y)
	  search_db->min_patch_y = search_db->max_patch_y;
    }
    return;
}


static void initialize_gcs_search_spaces(CTDB *ctdb, 
					 CTDB_SEARCH_SPACE_PTR space)
{
    float64 min_x, min_y, max_x, max_y;
    float64 temp_z;
    int32 success;
    int32 i;
    
    /* Set up the boilerplate iterators for each cell; whenever we need
     * to move to a new cell hereafter, we'll just copy the appropriate
     * structure in
     */
    for(i=0;i<space->n_cells;i++)
    {
	init_gcs_search_db(space, &(space->cell_data_list[i]), i);
    }
    
    /* We want to start all of our iterators off in the first valid
     * cell.  If we set the index to -1 and call advance_gcs_search_space,
     * it will find that for us, and we can copy that one everywhere.
     */
    space->search_patch.current_index = -1;
    
    advance_gcs_search_space(space, &space->search_patch);
    
    /* If we're traversing by patch, then space->search_patch is the
     * only CTDB_SEARCH_DB we'll use, so we won't fill out any of the
     * others.  Otherwise, we need to populate all of them.
     */
    if (!(space->control_flags & CTDB_BY_PATCH))
    {
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.tree.search_patch));
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.building.search_patch));
	copy_search_db(&(space->search_patch), 
		       &(space->iteration.linear.search_patch));
	
	for(i=0; i<CTDB_ABSTRACT_MAX; i++)
	{
	    copy_search_db(&(space->search_patch), 
			   &(space->iteration.abstract[i].search_patch));
	}
    }
}

static void copy_search_db(CTDB_SEARCH_DB *src, CTDB_SEARCH_DB *sink)
{
    int32                      min_x, min_y, max_x, max_y;
    uint32                    *orig_sink_quad_nodes = sink->quad_nodes;
    
    bcopy(src, sink, sizeof(CTDB_SEARCH_DB));
    
    /* The one thing the bcopy won't catch is quad_nodes.  Set it up
     * appropriately--it's really tempting to just copy the pointer,
     * but I don't want to have to debug this when somebody decides to
     * free what we were looking at...
     */
    if (src->num_quad_nodes)
    {
	sink->quad_nodes = (uint32 *)
	  ctdb_alloc(orig_sink_quad_nodes,
		     src->num_quad_nodes * sizeof(uint32),
		     "Quad node list for search space");
	
	bcopy(src->quad_nodes, sink->quad_nodes,
	      src->num_quad_nodes * sizeof(uint32));
    }
}

static int32 advance_gcs_search_space(CTDB_SEARCH_SPACE_PTR space,
				      CTDB_SEARCH_DB *current)
{
    int32 next_index;
    
    next_index = current->current_index + 1;
    
    /* Check for tiled GCS mode */
    if ((!ctdb_is_tiled) && (next_index > 0))
    {
	bzero(current, sizeof(CTDB_SEARCH_DB));
	return FALSE;
    }
    
    /* The search_db structures for each cell are all precalculated, we'll
     * just find the first interesting one and return it.
     */
    do
    {
	if (next_index >= space->n_cells)
	{
	    bzero(current, sizeof(CTDB_SEARCH_DB));
	    current->current_index = next_index;
	    return FALSE;
	}
	copy_search_db(&space->cell_data_list[next_index++], 
		       current);
    } while (current->ctdb == NULL);
    return(TRUE);
}

/* Free a search space */
void ctdb_destroy_search_space(CTDB_SEARCH_SPACE_PTR space)
{
    CTDB_SEARCH_DB current;    
    int32 i;
    
    if (space->cell_id_list)
    {
	STDDEALLOC(space->cell_id_list);
	space->cell_id_list = NULL;
    }

    if (space->cell_data_list)
    {
        for(i=0;i<space->n_cells;i++)
            STDDEALLOC(space->cell_data_list[i].quad_nodes);
	
        STDDEALLOC(space->cell_data_list);
	space->cell_data_list = NULL;
    }

    current = space->search_patch;
    if (current.quad_nodes)
    {
	STDDEALLOC(current.quad_nodes);
	current.quad_nodes = NULL;
    }

    current = space->iteration.tree.search_patch;
    if (current.quad_nodes)
    {
	STDDEALLOC(current.quad_nodes);
	current.quad_nodes = NULL;
    }
    current = space->iteration.building.search_patch;
    if (current.quad_nodes)
    {
	STDDEALLOC(current.quad_nodes);
	current.quad_nodes = NULL;
    }
    current = space->iteration.linear.search_patch;
    if (current.quad_nodes)
    {
	STDDEALLOC(current.quad_nodes);
	current.quad_nodes = NULL;
    }
    for(i=0;i<CTDB_ABSTRACT_MAX;i++)
    {
	current = space->iteration.abstract[i].search_patch;
	if (current.quad_nodes)
	{
	    STDDEALLOC(current.quad_nodes);
	    current.quad_nodes = NULL;
	}
    }
    
    STDDEALLOC(space);
    space = NULL;
}

/* Return the maximum number of patches which might be included in a
 * search space.
 */
/* GCS Note that this routine will provide the max number of patches found
   in the current cell of search space; not all cells.  Unknown where this is 
   a problem. 
   */
int32 ctdb_max_space_patches(CTDB_SEARCH_SPACE_PTR space)
{
    int32 i;
    int32 count = 0;
    CTDB_SEARCH_DB *cur;
    
    for(i=0;i<space->n_cells;i++)
    {
	cur = &space->cell_data_list[i];
	
	count += (cur->num_quad_nodes +
		  ((1 + cur->max_patch_x - cur->min_patch_x) *
		   (1 + cur->max_patch_y - cur->min_patch_y)));
    }
    return count;
}

/* Get the next patch (or the first patch), and return its number and
 * a pointer to it.
 */
static CTDB_FEATURE_DATA *next_patch(CTDB_SEARCH_SPACE_PTR space,
				     CTDB_SEARCH_DB *current, int32 *number,
				     int32 *patch_x, int32 *patch_y)
{
    CTDB_FEATURE_DATA *found = NULL;
    int32 initializing;
    int32 success;
    
    /* See if it is the first patch */
    if (!current->last_patch)
    {
	current->real_patches = 1;
	current->x = current->min_patch_x;
	current->y = current->min_patch_y;
	current->q = 0;
	initializing = TRUE;
    }
    else
      initializing = FALSE;
    
    do
    {
	if (!initializing)
	{
	    /* Go to the next real (not quad) patch, if applicable */
	    if (current->real_patches)
	    {
		/* increment (row major) */
		current->x++;
		if (current->x > (uint16)current->max_patch_x)
		{
		    current->x = current->min_patch_x;
		    current->y++;
		    if (current->y > (uint16)current->max_patch_y)
		    {
			current->real_patches = 0;
			current->q = -1; /* This'll be incremented below */
		    }
		}
	    }
	    /* Go to the next quad patch, if applicable */
	    if (!current->real_patches)
	    {
		/* Increment to next quad */
		current->q++;
		if (current->q >= current->num_quad_nodes)
		{
		    /* We've exhausted all patches */
		    /* go to next GCS cell */
		    success = advance_gcs_search_space(space, current);
		    if (success)
		      return (next_patch(space, current, number, patch_x,
					 patch_y)); 
		    else
		    {
			if (number)
			  *number = 0;
			if (patch_x)
			  *patch_x = *patch_y = 0;
			return NULL;
		    }
		}
	    }
	}
	
	if (current->ctdb == NULL)
	  return NULL;
	
	/* Look up this patch */
	if (current->real_patches)
	  current->last_number =
	    current->x + (current->y * current->ctdb->patches_wide);
	else
	  current->last_number = current->quad_nodes[current->q];
	if (number)
	  *number = current->last_number;
	if (patch_x)
	  *patch_x = current->x,
	  *patch_y = current->y;
	
	found = ctdb_lookup_feature(current->ctdb, current->last_number);
	
	/* Iterate until we find a patch, or we exhaust the space */
	initializing = FALSE;
	
    } while (!found);
    
    current->last_patch = found;
    current->last_feature_miss_count = current->ctdb->patch_miss_count;
    return current->last_patch; /* Success */
}

/* Return the next patch */
int32 ctdb_next_patch(CTDB_SEARCH_SPACE_PTR space, int32 *patch_number)
{
    /* This should only be called if CTDB_BY_PATCH is being used. */
    if (!(space->control_flags & CTDB_BY_PATCH))
    {
	*patch_number = 0;
	return 0;
    }
    
    /* Initialize all the iteration variables (we need to start over for
     * each new patch)
     */
    bzero(&space->iteration, sizeof(space->iteration));
    
    /* Move on to the next patch (which may be the first patch) */
    return (next_patch(space, &space->search_patch, patch_number,
		       (int32*)NULL, (int32*)NULL) != NULL);
}

/* Get a pointer to the current patch data */
static CTDB_FEATURE_DATA *current_patch(CTDB_SEARCH_SPACE_PTR space, 
					CTDB_SEARCH_DB *current,
					int32 *number,
					int32 *zero_offset,
					int32 *patch_x,
					int32 *patch_y)
{
    /* If we are traversing by patch, use the global `current' structure */
    if (space->control_flags & CTDB_BY_PATCH)
    {
	current = &space->search_patch;
	/* Make sure the application is paying attention */
	if (!current->last_patch)
	  return NULL;
    }
    else
    {
	/* Make sure there is a current patch */
	if (!current->last_patch)
	{
	    /* Get and return the next patch */
	    *zero_offset = 0;
	    return next_patch(space, current, number, patch_x, patch_y);
	}
    }
    
    /* See if the current patch is guaranteed to still be in memory
     * (because nothing has been paged into the cache since the last
     * time we fetched it).
     */
    if (current->last_feature_miss_count == current->ctdb->patch_miss_count)
    {
	if (number)
	  *number = current->last_number;
	if (patch_x)
	  *patch_x = current->x,
	  *patch_y = current->y;
	return current->last_patch;
    }
    
    /* Re-fetch the current patch */
    current->last_patch = ctdb_lookup_feature(current->ctdb, 
					      current->last_number);
    current->last_feature_miss_count = current->ctdb->patch_miss_count;
    if (number)
      *number = current->last_number;
    if (patch_x)
      *patch_x = current->x,
      *patch_y = current->y;
    return current->last_patch;
}    

/* Before more linear feature types were supported by CTDB, there were only
 * treelines.  `ctdb_next_linear_ref()' used to be called `ctdb_next_tree()'.
 * The following function is provided for backward compatibility.
 */
int32 ctdb_next_tree(CTDB_SEARCH_SPACE_PTR space,
		     float64 verts[][2],
		     float64 *tree_tops,
		     float64 *foliage_height,
		     float64 *radius)
{
    int32            return_val;
    CTDB_LINEAR_DESC lm;

    do
    {
        return_val = ctdb_next_linear_ref(space, verts, tree_tops, &lm);
    } while(lm.type != CTDB_FS_TREELINE && return_val);

    if(return_val)
    {
	if(foliage_height)
	  *(foliage_height) = lm.linear_data.treeline_info.foliage_height;

	if(radius)
	  *(radius) = lm.width * 0.5;
    }

    return return_val;
}

static void extract_model_information(CTDB                *ctdb,
				      CTDB_FEATURE_DATA   *feature,
				      CTDB_LINEAR_MODEL   *model,
				      CTDB_LINEAR_DESC    *feature_info)
{
    switch( model->type )
    {
      case CTDB_FS_TREELINE:
	feature_info->linear_data.treeline_info.trunk_radius =
	  CTDB_XYO_TO_METERS(model->linear_data.treeline_info.
			     trunk_radius,0,ctdb->patch_incrD,
			     ctdb->incrD);
	feature_info->linear_data.treeline_info.foliage_height=
	  model->linear_data.treeline_info.foliage_height *
	    CTDB_TREE_HEIGHT_BASIS;

	feature_info->linear_data.treeline_info.fullness =
	  model->linear_data.treeline_info.fullness /
	    (float32)(1 << CTDB_FULLNESS_BITS - 1);

	break;

      case CTDB_FS_CONCERTINA_WIRE:
      case CTDB_FS_FENCE:
      case CTDB_FS_MINEFIELD_FENCE:

	feature_info->linear_data.var_opacity_info.fullness =
	  model->linear_data.var_opacity_info.fullness /
	    (float32)(1 << CTDB_FULLNESS_BITS - 1);

	feature_info->linear_data.var_opacity_info.force_id =
	  ctdb->linear_models[feature[1].linear.reference].
	    linear_data.var_opacity_info.force_id;

	break;

      case CTDB_FS_DRAGONS_TEETH:

	feature_info->linear_data.dragons_tooth_info.spacing =
	  model->linear_data.dragons_tooth_info.spacing;
	feature_info->linear_data.dragons_tooth_info.component=
	  model->linear_data.dragons_tooth_info.component;

	break;
    }

    feature_info->width = CTDB_XYO_TO_METERS(model->width,0,
					     ctdb->patch_incrD,
					     ctdb->incrD);
    feature_info->type = model->type;
}


/* Return the next linear feature */
int32 
ctdb_next_linear_ref(
		     CTDB_SEARCH_SPACE_PTR space,
		     float64 vertices[][2],
		     float64 *tops,
		     CTDB_LINEAR_DESC *feature_info
		     )
{
    struct ctdb_search_state *state = &space->iteration.tree;
    CTDB *ctdb;
    CTDB_FEATURE_DATA *patch, *feature;
    int32 patch_number, patch_x, patch_y;
    int32 verts, fclass, offset, next, last;
    float64 trash, v[XY];
    int32 i, lin_ref;
    CTDB_LINEAR_MODEL *model;
    CTDB_AGGREGATE_MODEL *agg_model;
    
    /* Make sure we aren't finished */
    if (state->all_done)
      return 0;
    
    /* Iterate until we find something, or exhaust the list of patches */
    for(;;)
    {
	/* Make sure the patch is in memory */
	if (!(patch = current_patch(space, &state->search_patch, &patch_number,
				    &state->offset,
				    &patch_x, &patch_y)))
	{
	    state->all_done = TRUE;
	    return 0;
	}

	if (space->control_flags & CTDB_BY_PATCH)
	{
	    /* If we're searching by patch, all the valid state, except
	     * all_done and offset, is in the search space's search_db,
	     * not in the iterators
	     */
	    ctdb = space->search_patch.ctdb;
	}
	else
	{
	    ctdb = state->search_patch.ctdb;
	}
	
	/* See if we're beyond the real patches */
	if ( !CTDB_IS_QUAD_NODE(patch_number))
	{
	    if(space->iteration.last_tree != CTDB_PROCESSING_AGGREGATES)
	    {
		/* See if we already have a linear */
		if ((space->control_flags & CTDB_TREELINE_TREES) &&
		    state->offset &&
		    (space->iteration.last_tree != CTDB_RETURNED_ALL))
		{
		    int32 return_tree = 0;

		    feature = patch + state->offset;
		    verts = feature->info.vertices;

		    /* If the feature is a treeline, and if we've returned the 
		     * line, see if there is a first tree */
		    if (space->iteration.last_tree == CTDB_RETURNED_LINE)
		    {
			space->iteration.last_tree = CTDB_RETURNED_FIRST;
			if ( ctdb->linear_models[feature[1].linear.reference].
			    type == CTDB_FS_TREELINE 
			    && feature[1].integer & M_TREE_FIRST)
			  return_tree = 1;
		    }

		    /* If we've returned the first one, see if there is a 
		     * last one */
		    if (!return_tree &&
			(space->iteration.last_tree == CTDB_RETURNED_FIRST))
		    {
			space->iteration.last_tree = CTDB_RETURNED_ALL;
			if ( ctdb->linear_models[feature[1].linear.reference].
			    type == CTDB_FS_TREELINE && (verts > 1) && 
			    feature[1].integer & M_TREE_LAST)
			  return_tree = verts;
		    }

		    /* If we just decided to return a tree, do that */
		    if (return_tree)
		    {
			if (vertices)
			{
			    vertices[0][X] = 
			     CTDB_XYO_TO_METERS(feature[(return_tree<<1)].xy.x,
						patch_x, ctdb->patch_incrD,
						ctdb->incrD);
			    vertices[0][Y] = 
			     CTDB_XYO_TO_METERS(feature[(return_tree<<1)].xy.y,
						patch_y,ctdb->patch_incrD,
						ctdb->incrD);
			}

			if (tops)
			{
			    tops[0] = feature[(return_tree<<1)+1].z;

			    if (ctdb_is_tiled)
			      transform_reference_frame(
							space,
							state->search_patch.
							current_index,
							&vertices[0][X], 
							&vertices[0][Y], 
							&tops[0]);
			}
			else if (ctdb_is_tiled)
			{
			    trash = feature[(return_tree<<1)+1].z;
			    transform_reference_frame(
						      space, 
						      state->search_patch.
						      current_index,
						      &vertices[0][X], 
						      &vertices[0][Y], 
						      &trash);
			}

			if (feature_info)
			{
			    /* Convert information about the model from CTDB's 
			     * internal representation to a more user-friendly 
			     * form.
			     * Note that all this data is essentially
			     * coordinate-system independent
			     */
			    CTDB_LINEAR_MODEL *model = 
			      &(ctdb->linear_models[feature[1].linear.
						    reference]);

			    feature_info->linear_data.treeline_info.
			      trunk_radius =
				CTDB_XYO_TO_METERS(model->linear_data.
						   treeline_info.
						   trunk_radius, 0, 
						   ctdb->patch_incrD,
						   ctdb->incrD);

			    feature_info->linear_data.treeline_info.
			      foliage_height =
				model->linear_data.treeline_info.
				  foliage_height * CTDB_TREE_HEIGHT_BASIS;

			    feature_info->linear_data.treeline_info.fullness =
			      model->linear_data.treeline_info.fullness /
				(float32)(1 << CTDB_FULLNESS_BITS - 1);

			    feature_info->width = 
			      CTDB_XYO_TO_METERS(model->width, 0,
						 ctdb->patch_incrD,
						 ctdb->incrD);

			    feature_info->type = model->type;
			}

			return 1;
		    }
		}

		/* Note the location of the last feature in the patch */
		last = patch->hdr.size;

		/* See if we just started this patch */
		if (!state->offset)
		{
		    /* Skim forward to the first linear feature */
		    for (offset=1;offset<last;offset+=next)
		    {
			feature = patch+offset;
			verts   = feature->info.vertices;
			fclass  = feature->info.feature_class;

			if (fclass == FC_MICRO)
			{
			    next = NEXT_MICRO(verts);
			    continue;
			}
			else if (fclass == FC_VOLUME)
			{
			    next = NEXT_VOLUME(verts);
			    continue;
			}
			/* If we're not returning individual trees and we 
			 * come across one, skip it.
			 */
			else if ((fclass == FC_LINEAR) && 
				 (verts == 1) &&
				 (ctdb->linear_models[feature[1].linear.
						      reference].type ==
				  CTDB_FS_TREELINE) &&
				 (space->control_flags &
				  CTDB_NO_INDIVIDUAL_TREES))
			{
			    next = NEXT_LINEAR(verts);
			    continue;
			}
			else
			  break;
		    }

		    if (offset < last)
		      state->offset = offset;
		    else
		      fclass = -1;
		}
		else
		{
		    /* We know there is a linear at state->offset.  Go on to
		     * the next linear.  If it is an individual tree, and we
		     * don't want those, then keep on looking.
		     */
		    for (;;)
		    {
			feature = patch+state->offset;
			state->offset += NEXT_LINEAR(feature->info.vertices);

			if (state->offset < last)
			{
			    feature = patch+state->offset;
			    verts = feature->info.vertices;
			    fclass = feature->info.feature_class;

			    if (fclass == FC_LINEAR)
			    {
				if ((verts == 1) && 
				    ctdb->linear_models[feature[1].linear.
							reference].
				    type == CTDB_FS_TREELINE &&
				    (space->control_flags & 
				     CTDB_NO_INDIVIDUAL_TREES))
				  continue;
				else
				  break;
			    }
			    else
			      break;
			}
			else
			{
			    fclass = -1;
			    break;
			}
		    }
		}

		/* Is the thing we are looking at a linear? */
		if (fclass == FC_LINEAR)
		{
		    /* Return the line part of this linear.
		     */
		    if (vertices)
		      for (i=0;i<verts;i++)
		      {
			  vertices[i][X] = 
			    CTDB_XYO_TO_METERS(feature[2+(i<<1)].xy.x,patch_x,
					       ctdb->patch_incrD,ctdb->incrD);
			  vertices[i][Y] = 
			    CTDB_XYO_TO_METERS(feature[2+(i<<1)].xy.y,patch_y,
					       ctdb->patch_incrD,ctdb->incrD);
		      }
		    if (tops)
		      for (i=0;i<verts;i++)
		      {
			  tops[i] = feature[3+(i<<1)].z;
			  if (ctdb_is_tiled)
			    transform_reference_frame(
						      space, 
						      state->search_patch.
						      current_index,
						      &vertices[i][X], 
						      &vertices[i][Y], 
						      &tops[i]);
		      }
		    else if (ctdb_is_tiled && vertices)
		      for (i=0;i<verts;i++)
			transform_reference_frame(space, 
						  state->search_patch.
						  current_index,
						  &vertices[i][X],
						  &vertices[i][Y], 
						  &trash);
		    if (feature_info)
		    {
			/* Convert information about the model from CTDB's
			 * internal representation to a more user-friendly
			 * form.
			 */
			model =
			  &(ctdb->linear_models[feature[1].linear.reference]);

			extract_model_information(ctdb, feature, model,
						  feature_info);
		    }

		    /* Remember whether there might be more to return in future
		     * iteration calls.
		     */
		    if ((!(space->control_flags & CTDB_TREELINE_TREES) &&
			 (verts > 1)) || ctdb->linear_models[feature[1].linear.
							     reference].type
			!= CTDB_FS_TREELINE)
		      space->iteration.last_tree = CTDB_RETURNED_ALL;
		    else
		      space->iteration.last_tree = CTDB_RETURNED_LINE;

		    return verts;
		}
		else
		{
		    /* We've exhausted all of the linears, so start looking
		     * through the aggregates.
		     */
		    space->iteration.last_tree = CTDB_PROCESSING_AGGREGATES;

		    /* Note the location of the last feature in the patch */
		    last = patch->hdr.size;

		    /* Skim forward to the first aggregate feature */
		    if(space->control_flags & CTDB_NO_INDIVIDUAL_TREES)
		      fclass = -1;
		    else
		    {
			for (offset=1;offset<last;offset+=next)
			{
			    feature = patch+offset;
			    verts   = feature->info.vertices;
			    fclass  = feature->info.feature_class;

			    if (fclass == FC_MICRO)
			    {
				next = NEXT_MICRO(verts);
				continue;
			    }
			    else if (fclass == FC_VOLUME)
			    {
				next = NEXT_VOLUME(verts);
				continue;
			    }
			    else if (fclass == FC_LINEAR)
			    {
				next = NEXT_LINEAR(verts);
				continue;
			    }
			    else if (fclass == FC_CANOPY)
			    {
				next = NEXT_CANOPY(verts);
				continue;
			    }
			    else if (fclass == FC_LAID_LINEAR)
			    {
				next = NEXT_LAID_LINEAR(verts);
				continue;
			    }
			    else
			      break;
			}

			if (offset < last)
			  state->offset = offset;
			else
			  fclass = -1;
		    }
		}
	    }
	    else
	    {
		/* We know there is an aggregate at state->offset.  Go on
		 * to the next aggregate if we've processed all of the
		 * features in the current aggregate.
		 */

		/* Note the location of the last feature in the patch */
		last = patch->hdr.size;

		if(!state->feature_number)
		  for (;;)
		  {
		      feature = patch+state->offset;
		      state->offset += NEXT_AGGREGATE(0);

		      if (state->offset < last)
		      {
			  feature = patch+state->offset;
			  fclass = feature->info.feature_class;
			  break;
		      }
		      else
		      {
			  fclass = -1;
			  break;
		      }
		  }
		else
		  fclass = FC_AGGREGATE;
	    }

	    /* Is the thing we are looking at an aggregate? */
	    if (fclass == FC_AGGREGATE)
	    {
		feature = patch+state->offset;

		agg_model =
		  ctdb->aggregate_models[feature[1].aggregate.reference];

		/* Extract the `feature_number'th feature from the
		 * aggregate feature and return it.
		 */
		v[X] = CTDB_XYO_TO_METERS(feature[2].xy.x + agg_model->
					  features[state->feature_number].
					  offset_x, patch_x,
					  ctdb->patch_incrD,ctdb->incrD);
		v[Y] = CTDB_XYO_TO_METERS(feature[2].xy.y + agg_model->
					  features[state->feature_number].
					  offset_y, patch_y,
					  ctdb->patch_incrD,ctdb->incrD);

		if(vertices)
		{
		    vertices[0][X] = v[X];
		    vertices[0][Y] = v[Y];
		}

		if (tops)
		{
		    lin_ref =
		      agg_model->features[state->feature_number].model.ref;
		    tops[0] = ctdb_lookup_elevation(ctdb, v[X], v[Y]) +
		        ctdb->linear_models[lin_ref].linear_data.
			treeline_info.total_height*CTDB_TREE_HEIGHT_BASIS;

		    if (ctdb_is_tiled)
		      transform_reference_frame(
						space, 
						state->search_patch.
						current_index,
						&vertices[0][X], 
						&vertices[0][Y], 
						&tops[0]);
		}
		else if (ctdb_is_tiled && vertices)
		  transform_reference_frame(space, 
					    state->search_patch.
					    current_index,
					    &vertices[0][X],
					    &vertices[0][Y], 
					    &trash);

		if(feature_info)
		{
		    /* Convert information about the model from CTDB's
		     * internal representation to a more user-friendly
		     * form.
		     */
		    lin_ref =
		      agg_model->features[state->feature_number].model.ref;
		    model = &(ctdb->linear_models[lin_ref]);

		    extract_model_information(ctdb, feature, model,
					      feature_info);
		}

		/* Increment the feature number within the aggregate
		 * feature.  If we're done with the features in this
		 * aggregate, set the feature number to zero so we can
		 * move on to the next aggregate.
		 */
		if(++(state->feature_number) >= agg_model->n_features)
		  state->feature_number = 0;

		return 1;
	    }
	}

	/* We've exhausted the linears and aggregates in this patch, so
	 * advance to the next patch (if allowed).
	 */
	if (space->control_flags & CTDB_BY_PATCH)
	{
	    state->all_done = TRUE;
	    return 0;
	}
	if (!next_patch(space, &state->search_patch, &patch_number,
			&patch_x, &patch_y))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	state->offset = 0;
	space->iteration.last_tree = CTDB_RETURNED_ALL;
    }
}

/* Return the next volume */
int32 
ctdb_next_volume(
		 CTDB_SEARCH_SPACE_PTR   space,
		 CTDB_TERRAIN_SUBCLASS  *type,
		 int32                  *avoid,
		 int32                  *force_id,
		 CTDB_MES_NODE_KEY      *mes, 
		 float64                 vertices[][2],
		 float64                 roof_elevation[]
		 )
{
    struct ctdb_search_state *state = &space->iteration.building;
    CTDB *ctdb;
    CTDB_FEATURE_DATA *patch, *feature;
    int32 patch_number, patch_x, patch_y;
    int32 verts, fclass, offset, next, last;
    int32 i;
    float64 trash;
    
    /* Make sure we aren't finished */
    if (state->all_done)
      return 0;
    
    /* Iterate until we find something, or exhaust the list of patches */
    for(;;)
    {
	/* Make sure the patch is in memory */
	if (!(patch = current_patch(space, &state->search_patch, &patch_number,
				    &state->offset,
				    &patch_x, &patch_y)))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	
	if (space->control_flags & CTDB_BY_PATCH)
	{
	    /* If we're searching by patch, all the valid state, except
	     * all_done and offset, is in the search space's search_db,
	     * not in the iterators
	     */
	    ctdb = space->search_patch.ctdb;
	}
	else
	{
	    ctdb = state->search_patch.ctdb;
	}
	
	/* See if we're beyond the real patches */
	if (CTDB_IS_QUAD_NODE(patch_number))
	{
	    /*state->all_done = TRUE;
	      return 0;*/
	    
	}
	else
	{
	    /* Note the location of the last feature in the patch */
	    last = patch->hdr.size;
	    
	    /* See if we just started this patch */
	    if (!state->offset)
	    {
		/* Skim forward to the first volume */
		for (offset=1;offset<last;offset+=next)
		{
		    feature = patch+offset;
		    verts   = feature->info.vertices;
		    fclass  = feature->info.feature_class;
		    
		    if (fclass == FC_MICRO)
		    {
			next = NEXT_MICRO(verts);
			continue;
		    }
		    else
		      break;
		}

		if (offset < last)
		  state->offset = offset;
		else
		  fclass = -1;
	    }
	    else
	    {
		/* We know there is a volume at state->offset.  Go on to
		 * the next volume.
		 */
		feature = patch+state->offset;
		state->offset += NEXT_VOLUME(feature->info.vertices);

		if (state->offset < last)
		{
		    feature = patch+state->offset;
		    verts   = feature->info.vertices;
		    fclass  = feature->info.feature_class;
		}
		else
		  fclass = -1;
	    }
	    
	    /* Is the thing we are looking at a volume? */
	    if (fclass == FC_VOLUME)
	    {
		/* Return the volume.
		 */
		if (vertices)
		  for (i=0;i<verts;i++)
		  {
		      vertices[i][X] = 
			CTDB_XYO_TO_METERS(feature[3+(i<<1)].xy.x,patch_x,
					   ctdb->patch_incrD,ctdb->incrD);
		      vertices[i][Y] = 
			CTDB_XYO_TO_METERS(feature[3+(i<<1)].xy.y,patch_y,
					   ctdb->patch_incrD,ctdb->incrD);
		  }

		if (roof_elevation)
		  for (i=0;i<verts;i++)
		  {
		      roof_elevation[i] = feature[4+(i<<1)].z;
		      if (ctdb_is_tiled)
			transform_reference_frame(
					  space, 
					  state->search_patch.current_index,
					  &vertices[i][X], &vertices[i][Y], 
					  &roof_elevation[i]);
		  }
		else if (ctdb_is_tiled)
		  for (i=0;i<verts;i++)
		  {
		      trash = feature[4+(i<<1)].z;
		      transform_reference_frame(
					space, 
					state->search_patch.current_index,
					&vertices[i][X], &vertices[i][Y], 
					&trash);
		  }
		
		if (mes)
		{
		  if (feature[1].volume.type == CTDB_FS_MES)
		    CTDB_MES_ID_TO_MES_KEY(feature[1].volume.reference, *mes);
		  else 
		    CTDB_MES_MAKE_NODE_KEY_INVALID(*mes);
		}
		
		if (type)
		  *type = feature[1].volume.type;

		if (avoid)
		  *avoid = feature[1].volume.avoid;

		if (force_id)
		  *force_id = feature[1].volume.force_id;

		return verts;
	    }
	}
	
	/* We've exhausted the volumes in this patch, so advance to the
	 * next patch (if allowed).
	 */
	if (space->control_flags & CTDB_BY_PATCH)
	{
	    state->all_done = TRUE;
	    return 0;
	}
	if (!next_patch(space, &state->search_patch, &patch_number,
			&patch_x, &patch_y))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	state->offset = 0;
    }
}

/* Return the next laid linear */
int32 
ctdb_next_linear_gcs(
		     CTDB_SEARCH_SPACE_PTR space, 
		     float64               vertices[][2],
		     int32                *soil, 
		     float64              *width, 
		     int32                *edge_index,
		     int32                *edge_sub_index, 
		     CTDB                **out_ctdb)
{
    struct ctdb_search_state *state = &space->iteration.linear;
    CTDB *ctdb;
    CTDB_FEATURE_DATA *patch, *feature; 
    int32 patch_number, patch_x, patch_y;
    int32 verts, fclass, offset, next, last;
    int32 i, row_in_pat;
   
    /* Make sure we aren't finished */
    if (state->all_done)
      return 0;
    
    /* Iterate until we find something, or exhaust the list of patches */
    for(;;)
    {
	/* Make sure the patch is in memory */
	if (!(patch = current_patch(space, &state->search_patch, &patch_number,
				    &state->offset,
				    &patch_x, &patch_y)))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	
	if (space->control_flags & CTDB_BY_PATCH)
	{
	    /* If we're searching by patch, all the valid state, except
	     * all_done and offset, is in the search space's search_db,
	     * not in the iterators
	     */
	    ctdb = space->search_patch.ctdb;
	}
	else
	{
	    ctdb = state->search_patch.ctdb;
	}
	
	if (out_ctdb)
	  *out_ctdb = ctdb;
	
	/* See if we're beyond the real patches */
	if (!CTDB_IS_QUAD_NODE(patch_number))
	{
	    /* Note the location of the last feature in the patch */
	    last = patch->hdr.size;
	    
	    /* See if we just started this patch */
	    if (!state->offset)
	    {
		/* Skim forward to the first laid linear */
		for (offset=1;offset<last;offset+=next)
		{
		    feature = patch+offset;
		    verts   = feature->info.vertices;
		    fclass  = feature->info.feature_class;
		    
		    if (fclass == FC_MICRO)
		    {
			next = NEXT_MICRO(verts);
			continue;
		    }
		    else if (fclass == FC_VOLUME)
		    {
			next = NEXT_VOLUME(verts);
			continue;
		    }
		    else if (fclass == FC_LINEAR)
		    {
			next = NEXT_LINEAR(verts);
			continue;
		    }
		    else if (fclass == FC_CANOPY)
		    {
			next = NEXT_CANOPY(verts);
			continue;
		    }
		    else if (fclass == FC_AGGREGATE)
		    {
			next = NEXT_AGGREGATE(verts);
			continue;
		    }
		    else 
		      break;
		}

		if (offset < last)
		  state->offset = offset;
		else
		  fclass = -1;
	    }
	    else
	    {
		/* We know there is a laid linear at state->offset.  Go on to
		 * the next laid linear.
		 */
		feature = patch+state->offset;
		state->offset += NEXT_LAID_LINEAR(feature->info.vertices);

		if (state->offset < last)
		{
		    feature = patch+state->offset;
		    verts   = feature->info.vertices;
		    fclass  = feature->info.feature_class;
		}
		else
		  fclass = -1;
	    }
	    
	    /* Is the thing we are looking at a laid linear? */
	    if (fclass == FC_LAID_LINEAR)
	    {
		/* Return the laid linear.
		 */
		if (vertices)
		  for (i=0;i<verts;i++)
		  {
		      vertices[i][X] = 
			CTDB_XYO_TO_METERS(feature[3+i].xy.x, patch_x,
					   ctdb->patch_incrD, ctdb->incrD);
		      vertices[i][Y] = 
			CTDB_XYO_TO_METERS(feature[3+i].xy.y,patch_y,
					   ctdb->patch_incrD,ctdb->incrD);
		      if (ctdb_is_tiled)
			transform_reference_frame(
					  space, 
					  state->search_patch.current_index,
					  &vertices[i][X], &vertices[i][Y], 
					  (float64*)NULL);
		  }

		if (soil)
		{
		    int32 *soilmap=ctdb_lookup_soilmap(ctdb,patch_number);
		    row_in_pat = soilmap[feature[1].ll_soil.poly_char_index];
		    *soil =
		      ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
		}

		if (width)
		  *width = CTDB_XYO_TO_METERS(feature[2].laid_linear.width,0,
					      ctdb->patch_incrD,ctdb->incrD);

		if (edge_index || edge_sub_index)
		{
		    /* Hunt down the edge # with a little search */
		    ctdb_find_edge(ctdb, feature[2].laid_linear.edge_key,
				   patch_number, state->offset,
				   edge_index, edge_sub_index);
		}
		
		return verts;
	    }
	}
	
	/* We've exhausted the laid linears in this patch, so advance to the
	 * next patch (if allowed).
	 */
	if (space->control_flags & CTDB_BY_PATCH)
	{
	    state->all_done = TRUE;
	    return 0;
	}

	if (!next_patch(space, &state->search_patch, &patch_number,
			&patch_x, &patch_y))
	{
	    state->all_done = TRUE;
	    return 0;
	}

	state->offset = 0;
    }
}



/* Return the next laid linear.  WARNING:  This function is provided only
 * for backward compatibility with pre-GCS CTDB.  In GCS multi-cell mode,
 * edge_index and edge_sub_index may not be valid in the specified ctdb.
 *
 * The soil returned is the final soil information.
 */
int32 
ctdb_next_linear(
		 CTDB_SEARCH_SPACE_PTR space, 
		 float64               vertices[][2], 
		 int32                *soil, 
		 float64              *width, 
		 int32                *edge_index, 
		 int32                *edge_sub_index)
{
    return ctdb_next_linear_gcs(space, vertices, soil, width, edge_index, 
				edge_sub_index, (CTDB**) NULL);
}

/* Return the next abstract feature */
int32 
ctdb_next_abstract(
		   CTDB_SEARCH_SPACE_PTR space, 
		   int32 type_code, 
		   float32 **vertices, 
		   CTDB_ABSTRACT_DATA **data, 
		   int32 *data_size)
{
    struct ctdb_search_state *state;
    CTDB_FEATURE_DATA *patch, *feature;
    int32 patch_number;
    int32 verts, atype, dsize, offset, next, last;
    float32 *local_verts = NULL;	/* Buffer for returning vertices */
    
    /* Make sure the type_code is known */
    if ((type_code < 1) || (type_code >= CTDB_ABSTRACT_MAX))
      return 0;
    
    state = space->iteration.abstract+type_code;
    
    /* Make sure we aren't finished */
    if (state->all_done)
      return 0;
    
    /* Abstract features live in quad nodes, so if we are traversing
     * by patch, make sure we are in a quad node.
     */
    if (space->control_flags & CTDB_BY_PATCH)
    {
	if (!CTDB_IS_QUAD_NODE(space->search_patch.last_number))
	  return 0;
    }
    else if (!CTDB_IS_QUAD_NODE(state->search_patch.last_number))
    {
	/* Skip forward to the first quad node */
	state->search_patch.real_patches = 0;
	state->search_patch.q = 0;
	if (state->search_patch.num_quad_nodes > 0)
	{
	    state->search_patch.last_number = 
	      state->search_patch.quad_nodes[0];
	    state->search_patch.last_patch =
	      ctdb_lookup_feature(state->search_patch.ctdb,
				  state->search_patch.last_number);
	    state->search_patch.last_feature_miss_count =
	      state->search_patch.ctdb->patch_miss_count;
	    state->offset = 0;
	}
	else
	{
	    space->search_patch.last_patch = NULL;
	    state->all_done = TRUE;
	    return 0;
	}
    }
    
    /* Iterate until we find something, or exhaust the list of patches */
    for(;;)
    {
	/* Make sure the patch is in memory */
	if (!(patch = current_patch(space, &state->search_patch, &patch_number,
				    &state->offset,(int32*)NULL,(int32*)NULL)))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	
	/* The following test should always return TRUE because we
	 * skipped the real patches at the top of this function.  Thus
	 * the following is just an error check.
	 */
	/* Now This could happen if we changed cells */
	if (!CTDB_IS_QUAD_NODE(patch_number))
	{
	    /*state->all_done = TRUE;
	      return 0;*/
	    state->search_patch.real_patches = 0;
	    state->search_patch.q = -1;
	}
	else
	{
	    /* Note the location of the last feature in the patch */
	    last = patch->hdr.size;
	    
	    /* See if we just started this patch */
	    if (!state->offset)
	    {
		/* Skim forward to the first abstract of this type */
		for (offset=1;offset<last;offset=next)
		{
		    feature = patch+offset;
		    verts   = feature->abstract.num_vertices;
		    atype   = feature->abstract.type_code;
		    dsize   = feature->abstract.data_size;
		    
		    if (atype != type_code)
		    {
			next = offset+1+dsize+verts*2;
			continue;
		    }
		    else
		      break;
		}
		if (offset < last)
		  state->offset = offset;
		else
		  atype = -1;
	    }
	    else
	    {
		/* Go on to the next abstract of the right type. */
		feature = patch+state->offset;
		verts = feature->abstract.num_vertices;
		dsize = feature->abstract.data_size;
		state->offset += 1+dsize+verts*2;

		for(;;)
		{
		    if (state->offset < last)
		    {
			feature = patch+state->offset;
			verts   = feature->abstract.num_vertices;
			atype   = feature->abstract.type_code;
			dsize   = feature->abstract.data_size;

			if(atype != type_code)
			{
			    /* This feature is of the wrong type.  Fetch
			     * the next feature.
			     */
			    state->offset += 1+dsize+verts*2;
			    continue;
			}
			else
			  break;
		    }
		    else
		    {
			/* We did not find another abstract feature of the
			 * right type in this quad node.  Go to the next
			 * quad node.
			 */
			atype = -1;
			break;
		    }
		}
	    }
	    
	    /* Is the thing we are looking at the right type? */
	    if (atype == type_code)
	    {
		/* Return the abstract feature.
		 */
		if (vertices)
		{
		    if (ctdb_is_tiled)
		    {
			/* We need to transform the stored data into the
			 * desired frame of reference.  We obviously can't
			 * do that in place, so we'll do it in a static buffer
			 * that we'll expand as necessary
			 */
			/* $$$ This is, of course, a memory leak when
			 * operating in GCS multi-cell mode.  Some day we
			 * should fix this, but it requires modifying the
			 * interface, so we won't do it now.
			 */
			local_verts = 
			  (float32 *)ctdb_alloc(NULL,
						2*verts*sizeof(float32),
						"abstract vertices");
			*vertices = local_verts;
			
			transform_abstract_vertices(
					    space,
					    state->search_patch.current_index,
					    verts, (float32 *)(feature +1),
					    *vertices);
		    }
		    else 	
		      *vertices = (float32 *)(feature+1);
		}
		
		if (data)
		  *data = (CTDB_ABSTRACT_DATA *)(feature+1+verts+verts);

		if (data_size)
		  *data_size = dsize * sizeof(int32);
		
		return verts;
	    }
	}
	/* We've exhausted the abstracts in this patch, so advance to the
	 * next patch (if allowed).
	 */
	if (space->control_flags & CTDB_BY_PATCH)
        {
	    state->all_done = TRUE;
	    return 0;
	}

	if (!next_patch(space, &state->search_patch, &patch_number,
			(int32*)NULL, (int32*)NULL))
	{
	    state->all_done = TRUE;
	    return 0;
	}
	state->offset = 0;
    }
}

/* ctdb_nearest_road
 * Find the road nearest (x, y) in the current cell (specified by ctdb)
 *
 * THIS FUNCTION IS PUBLICLY VISIBLE ONLY FOR BACKWARD COMPATIBILITY,
 * and does not work as desired if used directly by a GCS aware application.
 * Please use ctdb_nearest_road_gcs.
 *
 * returns 1 for success, 0 for failure
 */
int32 
ctdb_nearest_road(CTDB *ctdb, float64 x, float64 y, float64 search_radius,
		  int32 *edge_index, int32 *edge_sub_index)
{
    float64 norm_x, norm_y, norm_r;
    int32 patchx, patchy, px0, py0, pr, patch_radius;
    int32 pxl, pxh, pyl, pyh;
    float64 dist;
    int32 success, success_cnt;
    
    /* In case the caller doesn't check the return value */
    *edge_index = 0;

    if (!ctdb)
    {
	*edge_sub_index = -1;
	return FALSE;
    }

    /* Normalize x, y and radius to incr units */
    norm_x = x * ctdb->inv_incr;
    norm_y = y * ctdb->inv_incr;
    norm_r = search_radius * ctdb->inv_incr;
    
    /* Get the patch number of each corner of the search radius,
     * and determine how many patches our search needs to cover.
     * This is better than simply converting search_radius to patch
     * units, since it considers proximity of the location to patch
     * edges.
     */
    patch_radius = 0;
#define COMPUTE_PATCHES(nx,ny,px,py)			\
    px = PATCH_X((nx)*ctdb->patch_inv_incr);		\
    py = PATCH_Y((ny)*ctdb->patch_inv_incr);
    
    COMPUTE_PATCHES(norm_x, norm_y, px0, py0);
    COMPUTE_PATCHES(norm_x-norm_r, norm_y-norm_r, patchx, patchy);
    if ((pr = abs(patchx-px0)) > patch_radius)
      patch_radius = pr;
    if ((pr = abs(patchy-py0)) > patch_radius)
      patch_radius = pr;
    COMPUTE_PATCHES(norm_x+norm_r, norm_y+norm_r, patchx, patchy);
    if ((pr = abs(patchx-px0)) > patch_radius)
      patch_radius = pr;
    if ((pr = abs(patchy-py0)) > patch_radius)
      patch_radius = pr;
    
    /* Now search ever-expanding rectangles, first 0x0, then 3x3, then
     * 5x5, etc.
     */
    dist = XYO_TO_POST(0x3fffffff, 0, ctdb->patch_incrD); /* Big number */
    success_cnt = 0;
    for (pr = 0; pr <= patch_radius; pr++)
    {
	pxl = px0-pr;
	pxh = px0+pr;
	pyl = py0-pr;
	pyh = py0+pr;
	
	success = 0;
	for (patchy=pyl; patchy<=pyh; patchy++)
	{
	    /* Bounds check */
	    if ((patchy < 0) || (patchy >= ctdb->patches_high))
	      continue;
	    
	    for (patchx=pxl; patchx<=pxh; patchx++)
	    {
		/* Bounds check */
		if ((patchx < 0) || (patchx >= ctdb->patches_wide))
		  continue;
		
		/* Only do the boundary on each pass (the interior was already
		 * covered on a previous pass).
		 */
		if ((patchy!=pyl) && (patchy!=pyh) &&
		    (patchx!=pxl) && (patchx!=pxh))
		  continue;
		
		success |=
		  ctdb_find_nearest_road_in_patch(ctdb, norm_x, norm_y,
						  patchx, patchy,
						  &dist,
						  edge_index, edge_sub_index);
		last_road_dist = dist;
	    }
	}
	
	/* Since the point may be near an patch boundary, we always need
	 * to go out one patch further than our first hit (unless the
	 * point was directly on top of the edge).
	 */
	if (success)
	  success_cnt++;
	if ((success_cnt == 2) || (dist == 0.0))
	  break;
    }
    if (success_cnt)
      return TRUE;
    else
      return FALSE;
}

/* ctdb_nearest_road_gcs
 *
 * Finds the road closest to x,y in the playbox.  in_ctdb specifies the
 * coordinate system in which x,y is set, out_ctdb will be set to the ctdb
 * containing the bridge (node_index *must* be used with out_ctdb for proper
 * behavior).  Like ctdb_nearest road, returns 1 for success, zero for failure.
 */

#define CTDB_ROAD_START_CELLS 50   /* Starting length to use for list of cells
				    * to look at when finding nearest road
				    */
int32
ctdb_nearest_road_gcs(CTDB *in_ctdb, float64 x, float64 y, 
		      float64 search_radius, int32 *edge_index, 
		      int32 *edge_sub_index, CTDB **out_ctdb)
{
    int32 home_cell = ctdb_get_cell_id(in_ctdb);
    int32 cur_cell = home_cell;
    CTDB *cur_ctdb = in_ctdb;
    
    float64 min_dist;
    int32 ret;
    int32 best_road = -1;
    int32 best_sub_road = -1;
    
    float64 best_dist;
    float64 max_radius = search_radius;
    
    best_dist = 1e18;             /* Very far away */
    *out_ctdb = in_ctdb;
    
    if (!in_ctdb)
    {
	*edge_index = -1;
	*edge_sub_index = -1;
	*out_ctdb = NULL;
	return 0;
    }
    
    /* If we're not GCS aware, we won't worry about conversions */
    if (ctdb_is_tiled)
    {
	if (!ctdb_point_on_database(in_ctdb, x, y))
	{
	    gcs_gcs2e_to_gcs(home_cell, x, y,&home_cell, &x, &y);
	    cur_ctdb = (CTDB *)gcs_get_tdb(home_cell, FALSE);
	    cur_cell = home_cell;
	}
    }
    
    if ((ret = ctdb_nearest_road(cur_ctdb, x, y, search_radius, &best_road,
				 &best_sub_road)) != 0)
    {
	/* We found one.  We're done if we're not GCS-aware, otherwise,
	 * we may need to look in more cells.
	 */
	if (!ctdb_is_tiled)
	{
	    *edge_index = best_road;
	    *edge_sub_index = best_sub_road;
	    return ret;
	}
	
	best_dist = last_road_dist;
	/* Only look in other cells if it's possible that there is
	 * a road in another cell that's closer than the one we just
	 * found
	 */	
	if (((x - best_dist) < cur_ctdb->min_x) ||
	    ((y - best_dist) < cur_ctdb->min_y) ||
	    ((x + best_dist) > cur_ctdb->max_x) ||
	    ((y + best_dist) > cur_ctdb->max_y))
	{
	    max_radius = best_dist;
	}
	else
	{
	    *edge_index = best_road;
	    *edge_sub_index = best_sub_road;
	    return ret;
	}	
    }
    
    {
	/* If we fall through to here, we need to look at more cells */
	int32 *cell_list;
	int32 ncells = CTDB_ROAD_START_CELLS;
	int32 i;
	float64 newx, newy;
	int32 cur_road, cur_sub_road;
	
	cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					"Find nearest road index");
	
	if (!gcs_cells_within_radius(cur_cell, x, y, 0.0, max_radius,
				     &ncells, cell_list))
	{
	    cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					    "Find nearest road index");
	    gcs_cells_within_radius(cur_cell, x, y, 0.0, max_radius,
				    &ncells, cell_list);
	}
	
	for(i=0;i<ncells;i++)
	{
	    if (cell_list[i] != home_cell)
	    {
		cur_ctdb = (CTDB *)gcs_get_tdb(cell_list[i], FALSE);
		if (cur_ctdb)
		{
		    gcs_gcs2e_to_gcscs(home_cell, x, y, 
					 cell_list[i], &newx, &newy);
		    
		    if ((ret = ctdb_nearest_road(cur_ctdb, newx, newy, 
						 max_radius, &cur_road,
						 &cur_sub_road)) != 0)
		    {
			if (last_road_dist < best_dist)
			{
			    best_dist = last_road_dist;
			    best_road = cur_road;
			    best_sub_road = cur_sub_road;
			    *out_ctdb = cur_ctdb;
			}
		    }
		}
	    }  
	}
	STDDEALLOC(cell_list);
	
	*edge_index = best_road;
	*edge_sub_index = best_sub_road;
	return (best_road != -1);
    }
    
}

int32
ctdb_lookup_soil_from_defrag(CTDB *ctdb, CTDB_ABSTRACT_DATA *defrag)
{
    if (defrag)
    {
	int32  row_in_pat;
	
	int32 *soilmap = ctdb->soil_tables[defrag->soil_defrag.soilmap_index];
	row_in_pat = soilmap[defrag->soil_defrag.soil_index];
	return ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
    }
    else
      return CTDB_ILLEGAL_SOIL_TYPE;
} 
 
