#ifndef lint
static char rcsid [] = "$RCSfile: ct_veh.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_veh.c                                                          *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to compute intervisibility wrt vehicles              *
*       Created: Thu Oct 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libctdb_local.h"

/*ARGSUSED*/
float64 
ctdb_vehicle_raster(CTDB *ctdb, 
		    float64 x0, 
		    float64 y0, 
		    float64 z0, 
		    float64 x1, 
		    float64 y1, 
		    float64 zl, 
		    float64 zh,
		    float64 width, 
		    int32 n_veh, 
		    CTDB_VEHICLE_LOCATION *veh,
		    int32 ignore0, 
		    int32 ignore1, 
		    uint32 *raster)
{
    CTDB_VEHICLE_LOCATION *v;
    float64 eye[2], targ[2];
    float64 half_width;
    float64 tray, tveh;
    float64 zrayl, zrayh, ztop, zbot;
    float64 radius;
    float64 top_block, bottom_block, left_block, right_block;
    int32 hit_veh;
    int32 i;

    eye[0] = x0;
    eye[1] = y0;
    targ[0] = x1;
    targ[1] = y1;
    half_width = width * 0.5;
    hit_veh = 0;

    for (i=0;i<n_veh;i++)
      if ((i != ignore0) && (i != ignore1) && veh[i].used)
      {
	  v = veh+i;

	  radius = v->width * 0.5;

	  /* Does this vehicle cross our line of sight in 2D? */
	  if (ctdb_point_intersection_t(eye, targ,
					v->location[X], v->location[Y],
					radius, half_width,
					&tray, &tveh) ==
	      DISJOINT)
	    continue;

	  /* Find elevation of line-of-sight rays at this point */
	  zrayl = (zl - z0) * tray + z0;
	  zrayh = (zh - z0) * tray + z0;

	  /* Find top and bottom of obstruction */
	  zbot = v->location[Z];
	  ztop = zbot + v->height;

	  /* Can we see completely above or below it? */
	  if ((ztop <= zrayl) || (zbot >= zrayh))
	    continue;

	  hit_veh = 1;

	  /* Find how much of the target is blocked by the
	   * obstruction.
	   */

	  ctdb_width_obstruction(tray * width, radius, tveh,
				 ((x1 - x0) * (v->location[Y] - y0) >
				  (v->location[X] - x0) * (y1 - y0)),
				 &left_block, &right_block);


	  /* How much is visible under the vehicle? */
	  if (zrayl < zbot)
	    bottom_block = (zbot - zrayl) / (zrayh - zrayl);
	  else
	    bottom_block = 0.0;

	  /* How much is visible above the vehicle? */
	  if (zrayh > ztop)
	    top_block = (ztop - zrayl) / (zrayh - zrayl);
	  else
	    top_block = 1.0;

	  /* OR this block onto our rasterized image */
	  ctdb_rasterize_box(left_block,right_block,bottom_block,top_block,
			     raster);
      }

    if (!hit_veh)
      return 1.0;
    else
      return 1.0 - ctdb_raster_density(raster);
}

float64 
ctdb_vehicle_blockage(CTDB *ctdb, 
		      float64 x0, 
		      float64 y0, 
		      float64 z0,
		      float64 x1, 
		      float64 y1, 
		      float64 zl, 
		      float64 zh,
		      float64 width, 
		      int32 n_veh, 
		      CTDB_VEHICLE_LOCATION *veh,
		      int32 ignore0, 
		      int32 ignore1)
{
    CTDB_RASTER raster;
    
    if (!ctdb)
    {
	printf("Warning: null ctdb in ctdb_vehicle_blockage\n");
	return 1.0;
    }

    bzero(raster, sizeof(raster));

    return ctdb_vehicle_raster(ctdb, x0, y0, z0, x1, y1, zl, zh, width,
			       n_veh, veh, ignore0, ignore1, raster);
}  
