/*
 * $RCSfile: libctdb.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libctdb.h                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Header file for libctdb                                   *
*       Created: Mon May 20 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                             *
*                                                                           *
****************************************************************************/

#ifndef _LIBCTDB_H_INCLUDED
#define _LIBCTDB_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>    /* Needed for reference to FILE later */
#include <stdtypes.h> /* common/include/global */
#include <stdroute.h> /* common/include/global */
#include <stdext.h>

#define CTDB_FEATURE_TREELINES		0x1
#define CTDB_FEATURE_CANOPIES		0x2 /* Including surrounding 
					     * treeline 
					     */
#define CTDB_FEATURE_BUILDINGS		0x4

/* Misc constants */
#define CTDB_NAME_LENGTH     14   /* same as Simnet */
#define CTDB_FNAME_LENGTH    256  /* Enough room for entire path */
#define CTDB_DATE_LENGTH     28   /* Enough room for ctime() */
#define CTDB_PATCH_CACHES    16   /* Number of patch data caches */
#define CTDB_CACHE_BKT_UNITS  4   /* Unit (i.e. #of bytes) cache buckets
				   * are stored in
				   */

/* Definitions of the fields in a 32-bit soil type.
 */
#define CTDB_SIMNET_SOIL(x)  ((x) & 0x0000000f)
#define CTDB_CCTT_SOIL(x)    (((x) >> 4) & 0x0000001f)
#define CTDB_MOBILITY_FLAG_SHIFT 9
#define CTDB_WATER_FLAG   (0x1 << CTDB_MOBILITY_FLAG_SHIFT)
#define CTDB_NO_GO_FLAG   (0x2 << CTDB_MOBILITY_FLAG_SHIFT)
#define CTDB_SLOW_GO_FLAG (0x4 << CTDB_MOBILITY_FLAG_SHIFT)
#define CTDB_ROAD_FLAG    (0x8 << CTDB_MOBILITY_FLAG_SHIFT)
#define CTDB_WATER(x)   ((x) & CTDB_WATER_FLAG)
#define CTDB_NO_GO(x)   ((x) & CTDB_NO_GO_FLAG)
#define CTDB_SLOW_GO(x) ((x) & CTDB_SLOW_GO_FLAG)
#define CTDB_ROAD(x)    ((x) & CTDB_ROAD_FLAG)

/* Definitions of constants for the Polygon Attribute Table.
 */
#define CTDB_INVALID_FACC_ATTR   50000
#define CTDB_SIMNET_MOB_ATTR     50001
#define CTDB_MAX_FACC_ATTR       46656
#define CTDB_LEN_FACC_NAME       3
#define CTDB_MAX_FACC_NAME       8

typedef enum ctdb_terrain_subclass
{
    /* FC_MICRO and TERRAIN subclasses */
    CTDB_FS_NA                = 0,  /* Unknown or not applicable */
    CTDB_FS_SKIN              = 1,  /* Terrain Skin Subclass */
    CTDB_FS_MULTI             = 2,  /* Multilevel Terrain Subclass */
    CTDB_FS_WATER             = 3,  /* Terrain Water subclass */
    CTDB_NUM_MICRO_SUBCLASSES = 4,
    CTDB_FS_ROAD              = 4,  /* Laid Linear Road */
    
    /* FC_VOLUME subclasses */
    CTDB_FS_BUILDING        = 0,
    CTDB_FS_ROCK_DROP       = 1,
    CTDB_FS_MES             = 2,  /* Multi Elevation structure */
    CTDB_FS_PLUG            = 3,
    /* add more FC_VOLUME subclasses here and change 
     * CTDB_NUM_VOLUME_SUBCLASSES accordingly
     */
    CTDB_NUM_VOLUME_SUBCLASSES = 4,
    
    /* FC_LINEAR subclasses */
    CTDB_FS_TREELINE        = 0,
    CTDB_FS_CONCERTINA_WIRE = 1,
    CTDB_FS_WIRE_ROAD_BLOCK = 2,
    CTDB_FS_FENCE           = 3,
    CTDB_FS_MINEFIELD_FENCE = 4,
    CTDB_FS_DRAGONS_TEETH   = 5,
    /* add more FC_LINEAR subclasses here and change 
     * CTDB_NUM_LINEAR_SUBCLASSES accordingly
     */
    CTDB_NUM_LINEAR_SUBCLASSES = 6
} CTDB_TERRAIN_SUBCLASS;

typedef float64 (*CTDB_OPACITY_MAPPING_FCN)(float64 fullness);

typedef enum ctdb_terrain_class
{
    CTDB_FC_ILLEGAL          = -1,
    CTDB_FC_MICRO            = 0,
    CTDB_FC_VOLUME           = 1,
    CTDB_FC_LINEAR           = 2,
    CTDB_FC_CANOPY           = 3,
    CTDB_FC_LAID_LINEAR      = 4,
    CTDB_FC_AGGREGATE        = 5
} CTDB_TERRAIN_CLASS;

typedef enum
{
     CTDB_MES_FIRST_TYPE = 0,
     CTDB_MES_UNKNOWN    = 0,
     CTDB_MES_BUILDING   = 1,
     CTDB_MES_CAVE       = 2,
     CTDB_MES_TUNNEL     = 3,
     CTDB_MES_BRIDGE     = 4,
     CTDB_MES_LAST_TYPE  = 5
} CTDB_MES_TYPE;

/* Types of dragon's tooth components.
 */
#define CTDB_DT_STEEL_HEDGEHOG        0
#define CTDB_DT_STEEL_TETRAHEDRON     1
#define CTDB_DT_CONCRETE_BLOCK        2
#define CTDB_DT_CONCRETE_TETRAHEDRON  3



#define CTDB_MAX_POSTS_PER_PATCH      25   /*maximum number of posts per
					     patch*/

#define CTDB_PATCHES_PER_GROUP         4  /* Num of patches per patch group */

/* CTDB_RASTER_SIZE cannot exceed the number of bits in an uint32 int! */
#define CTDB_RASTER_SIZE 32

/* Valid values for gcs_mode, below.  Specifies coordinate system and tiling
 * options used in database generation.  Note that multi-cell database headers
 * will not contain any terrain data, but rather will implicitly reference
 * one or more "single-cell" databases.
 */
typedef enum
{
    CTDB_MODE_UNKNOWN      = -1,
    CTDB_MODE_SIMNET       = 0,
    CTDB_MODE_SINGLE_CELL  = 1,
    CTDB_MODE_MULTI_CELL   = 2,
    CTDB_MODE_MULTI_HDR    = 3
} CTDB_TDB_TYPE;

/* DATABASE NAMING CONVENSION defines */
#ifdef USE_LITTLE_ENDIAN
#define CTDB_ENDIAN_ID          "l"
#else
#define CTDB_ENDIAN_ID          "b"
#endif
#define CTDB_MULTI_CELL_ID      "m"
#define CTDB_SINGLE_CELL_ID     "s"
#define CTDB_SIMNET_ID          "c"
#define CTDB_THINNED_ID         't'

#define CTDB_FMT1_EXTENSION     "ctdb"
#define CTDB_FMT2_EXTENSION     "ctdb"
#define CTDB_FMT3_EXTENSION     "ctb"
#define CTDB_MULTI_CELL_DIR     "dir"

#define CTDB_OLDEST_SUPPORTED_FILE_FORMAT 7
#define CTDB_FILE_FORMAT                  7

typedef uint32 CTDB_RASTER[CTDB_RASTER_SIZE];

typedef struct ctdb_vehicle_location
{
    int32   used;
    float64 location[3];
    float64 width;
    float64 height;
} CTDB_VEHICLE_LOCATION;

/*
 * CTDB OTHER DATA STRUCTURES            
 */

/* CTDB_ABSTRACT_DATA:
 *
 * Format of data associated with so-called abstract features.  Note
 * that this structure is analogous to a protocol definition, in that
 * the final record of each union substructure can be variable length.
 */
typedef union ctdb_abstract_data
{
    struct ctdb_quad_data
    {
	/* Array of 4 offsets to child quad nodes (in 4-byte units)
	 * from the beginning of the quad tree data storage.
	 */
	uint32 child_start[1];  /* really 4 members, not 1 */
    } quad_data;

    struct ctdb_canopy_data
    {
#ifdef USE_LITTLE_ENDIAN
	unsigned padding         : 23;
	unsigned thickness       : 4;
	unsigned vegetation_type : 4;
	unsigned impenetrable    : 1;
#else
	unsigned impenetrable    : 1;
	unsigned vegetation_type : 4;
	unsigned thickness       : 4;
	unsigned padding         : 23;
#endif
    } canopy;

    struct ctdb_soil_defrag_data
    {
#ifdef USE_LITTLE_ENDIAN
	unsigned soilmap_index :16;
	unsigned padding       : 8;
	unsigned level         : 4;
	unsigned soil_index    : 4;
#else
	unsigned soil_index    : 4;
	unsigned level         : 4;
	unsigned padding       : 8;
	unsigned soilmap_index :16;
#endif
    } soil_defrag;

    struct ctdb_steep_slope_data
    {
#ifdef USE_LITTLE_ENDIAN
    unsigned padding       : 25;
    unsigned slope_degrees : 7;
#else
    unsigned slope_degrees : 7;
    unsigned padding       : 25;
#endif
    } steep_slope;

    struct ctdb_label_data
    {
	char text[4]; /* variable length, NULL terminated */
    } label;

    struct ctdb_offroad_data
    {
	uint32 edge_number;
    } offroad;

    struct ctdb_contaminant_data
    {
        uint32 info;
    } contaminant;

    struct ctdb_bridge_data
    {
#define CTDB_MASS_BASIS (300000.0 / (1 << 16))
	uint32 max_mass;
    } bridge;

    struct ctdb_crater_data
    {
#ifdef USE_LITTLE_ENDIAN
        unsigned padding      : 15;
        unsigned avoid        : 1;
        unsigned crater_type  : 8;
        unsigned force_id     : 8;
#else
        unsigned force_id     : 8;
        unsigned crater_type  : 8;
        unsigned avoid        : 1;
        unsigned padding      : 15;
#endif
        float32 depth;
    } crater;

    struct ctdb_ditch_data
    {
#ifdef USE_LITTLE_ENDIAN
        unsigned padding      : 14;
        unsigned avoid        : 1;
	unsigned parapet      : 1;
        unsigned atd_type     : 8;
        unsigned force_id     : 8;
#else
        unsigned force_id     : 8;
        unsigned atd_type     : 8;
	unsigned parapet      : 1;
        unsigned avoid        : 1;
        unsigned padding      : 14;
#endif

	float32 depth;
    } ditch;

    struct ctdb_fighting_pos_data
    {
#define MAX_ENTRY_PATH_PT 10
	float32  entry_path_points[MAX_ENTRY_PATH_PT][2];
#ifdef USE_LITTLE_ENDIAN
	uint16   padding; 
	uint8    entry_pt_index;
	uint8    hide_pos_index;     
        float32  fp_width;
        float32  turret_depth;
        float32  hull_depth;             
        float32  platform_len;          /* Defilade base length */ 
#else
        float32  platform_len;          /* Defilade base length */
        float32  hull_depth;             
        float32  turret_depth;
        float32  fp_width;
	uint8    hide_pos_index;     
	uint8    entry_pt_index;
	uint16   padding; 
#endif
    } fighting_pos;

    struct ctdb_dragons_teeth_data
    {
	struct _unique_id
	{
	    uint16  site;
	    uint16  host;
	    uint16  pid;
	    uint16  sequence;
        } unique_id; /* used to group sets of dragon's teeth together */
        uint8    force_id;
        uint8    dt_type;
        uint8    line_type; /* center line, or center line attachment */
        uint8    _pad1;
        uint32   total_teeth; /* per row */
        float32  teeth_distance;
        /* These are only used if this line_type is a center line */
        uint8    obst_intent;
        uint8    num_rows;
        int16    _pad2;
        float32  row_distance;  
        float32  template_location[XYC];
        uint32   template_orientation;
        char     template_name[20];
    } dragons_teeth;

    struct ctdb_berm_data
    {
	float32 height;
    } berm;

    struct ctdb_camouflage_data
    {
	uint32 pad;
    } camouflage;

    struct ctdb_rubble_data
    {
	unsigned makeup      : 4;
	unsigned coarseness  : 4;
	unsigned padding     : 24;
    } rubble;

    struct ctdb_snow_data
    {
#define CTDB_PRESSURE_BASIS (10000.0 / (1 << 16))
	uint32 max_pressure;
    } snow;

    struct ctdb_miclic_data
    {
        uint32 detonated;
        float32 segment_width; /* meters */
	float32 scorch_width; /* meters */
    } miclic;

    struct ctdb_hydro_surface_data
    {
#ifdef USE_LITTLE_ENDIAN
        unsigned padding       : 31;
        unsigned wetness       : 1;
#else
        unsigned wetness       : 1;
        unsigned padding       : 31;
#endif
    } hydro_surface;

    struct ctdb_hydro_subsurface_data
    {
#ifdef USE_LITTLE_ENDIAN
        unsigned padding       : 31;
        unsigned wetness       : 1;
#else
        unsigned wetness       : 1;
        unsigned padding       : 31;
#endif
    } hydro_subsurface;

} CTDB_ABSTRACT_DATA;

/* CTDB_EDGE:
 *
 * Structure used to collect a list of linear features which form an
 * uninterrupted "edge" between two intersections.  The `feature'
 * member is actually variable length.
 */
typedef struct ctdb_edge
{
    /* Total distance along the segment (useful for planning) */
    int32 distance;

    /* Indices of the nodes at the beginning and end of the segment */
    int32 first_node;
    int32 last_node;

    int32 num_features;
    struct ctdb_feature_reference
    {
	/* The patch where this feature is stored */
	uint32 patch_number;

#ifdef USE_LITTLE_ENDIAN
    unsigned padding      : 7;
    unsigned reverse      : 1;
    unsigned patch_offset : 24;
#else
    /* Byte offset into the patch */
    unsigned patch_offset : 24;

    /* Whether the vertices in the patch are in reverse order */
    unsigned reverse      : 1;

    unsigned padding      : 7;
#endif
    } feature[1];

} CTDB_EDGE;

/* CTDB_NODE:
 *
 * Represents the topologic connection between CTDB_EDGE structures.
 * The `connection' member is actually variable length.
 */
typedef struct ctdb_node
{
    /* Approximate location of the node (useful for planning) */
    float32 x, y;

#ifdef USE_LITTLE_ENDIAN
    unsigned num_connections : 8;
    unsigned padding         : 23;
    unsigned is_bridge       : 1;
#else
    /* Whether the node corresponds to a bridge */
    unsigned is_bridge       : 1;
    unsigned padding         : 23;
    unsigned num_connections : 8;
#endif

    struct ctdb_connection
    {
#ifdef USE_LITTLE_ENDIAN
	unsigned node_index             : 24;
	unsigned padding                : 6;
	unsigned connection_crosses_mes : 1;
	unsigned node_in_other_cell     : 1;
#else
	unsigned node_in_other_cell     : 1;  /* TRUE indicates that the node
					       * index references a node in 
					       * another cell, FALSE 
					       * indicates that it is in 
					       * this cell.
					       */
	unsigned connection_crosses_mes : 1;  /* TRUE indicates that the
					       * either the current node
					       * or the node connected to
					       * is in an MES structure.
					       * This dictates special
					       * usage of the edge structure
					       * below.
					       */
	unsigned padding                : 6;
	
	/* The node which this connects to */
	unsigned node_index             : 24;
#endif
	
	union edge_union
	{
	    int32 edge_index;    /* The edge used to connect these nodes,
			          * if they are in the same database. 
			          */

	    int32 other_cell;    /* If node_in_other_cell is TRUE,
			          * the cell in which node_index is valid.
			          */
	    uint32 aperture_key; /* If connection_crosses_mes is TRUE,
				  * this specifies which aperture in which
				  * MES structure must be traversed to move
				  * from the current node to the connected
				  * node
				  */
	} edge;
    } connection[1];

} CTDB_NODE;

/* CTDB_LINEAR_DESC:
 *
 * This structure is used to return information about linear models to the
 * application.
 *
 * It is also used to pass a linear feature or an individual tree
 * (n_verts == 1) to the feature compression algorithms.
 *
 * `trunk_radius' specifies the radius of the tree trunk.
 *
 * `foliage_height' specifies the height of the foliage of the tree.
 *
 * `fullness' specifies the fullness of the leaves or wire.  0.0 represents
 * zero fullness (nothing there), and 1.0 represents maximum fullness
 * (completely blocks light).
 *
 * `total_height' specifies the height of the feature.
 *
 * `spacing' represents how far apart from each other the dragon's teeth are
 * placed along the linear segment.
 *
 * `component' specifies the type of dragon's tooth, specified in libctdb.h
 *
 * `type' specifies the type of linear feature, specified in libctdb.h
 *
 * `width' specifies the foliage width for trees, or the width of dragon's
 * teeth.
 */
typedef struct ctdb_linear_desc
{
    union
    {
	struct
	{
            float32  trunk_radius;
            float32  foliage_height;
            float32  fullness;
	    float32  total_height;
	} treeline_info;

	struct
	{
            float32  fullness;
            int32    force_id;
            int32    pad2;
	} var_opacity_info;

	struct
	{
            float32  spacing;
            int32    component;
            int32    pad1;
	} dragons_tooth_info;
    } linear_data;

    CTDB_TERRAIN_SUBCLASS type;
    float32 width;

} CTDB_LINEAR_DESC;


typedef struct ctdb_linear_model
{
    /* This defines the models for the various linear features.
     */
    union
    {
	struct
	{
	    /* `treeline_info' is used for trees and treelines.
	     * 
	     * `trunk_radius' is measured in the same units as x,y in the
	     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
	     *
	     *  `foliage_height' is the distance from the top of the tree to
	     * the lowest branches measured in the unit
	     * `CTDB_TREE_HEIGHT_BASIS':
	     * 
	     * #define CTDB_TREE_HEIGHT_BASIS (200.0 / (1<<14))
	     *
	     *  `total_height' is the distance from the top of the tree to
	     * the ground measured in `CTDB_TREE_HEIGHT_BASIS'.  This is
	     * only used for trees in aggregate models.  Otherwise, their
	     * heights are derived from their Z values (tree tops).  Z
	     * values aren't stored with aggregate features.
	     *
	     * `fullness' refers to the "density" of the leaves, wire, or
	     * whatever the feature is made of.  This value ranges from zero
	     * (nothing there) to the maximum value (lets no light through).
	     */
#ifdef USE_LITTLE_ENDIAN
	    unsigned pad1             : 2;
	    unsigned foliage_height   : 14;
	    unsigned trunk_radius     : 16;
	    unsigned pad2             : 10;
	    unsigned total_height     : 14;
	    unsigned fullness         : 8;
#else
	    unsigned trunk_radius     : 16;
	    unsigned foliage_height   : 14;
	    unsigned pad1             : 2;
	    unsigned fullness         : 8;
	    unsigned total_height     : 14;
	    unsigned pad2             : 10;
#endif
	} treeline_info;

	/* `var_opacity_info' is used for wire and fences.
	 */
	struct
	{
#ifdef USE_LITTLE_ENDIAN
	    unsigned fullness         : 8;
            unsigned force_id         : 2;
	    unsigned pad1             : 22;
	    unsigned pad2             : 32;
#else
	    unsigned pad1             : 22;
            unsigned force_id         : 2;
	    unsigned fullness         : 8;
	    unsigned pad2             : 32;
#endif
	} var_opacity_info;

	/* `dragons_tooth_info' is used for dragon's teeth.
	 * 
	 * `spacing' represents how far apart from each other the dragon's
	 * teeth are placed along the linear segment.
	 *
	 * Dragon's teeth can be made of steel hedgehogs, steel tetrahedrons,
	 * concrete blocks, or concrete tetrahedrons.  Each of these affects
	 * intervisibility differently (hence the need for the `component'
	 * element).
	 */
	struct
	{
#ifdef USE_LITTLE_ENDIAN
	    float32  spacing;
	    unsigned component        : 4;
	    unsigned pad              : 28;
#else
	    float32  spacing;
	    unsigned pad              : 28;
	    unsigned component        : 4;
#endif
	} dragons_tooth_info;
    } linear_data;

    /* `type' specifies the specific type of linear feature (i.e. treeline,
     * minefield fence, etc.).
     * 
     * The width of the foliage, the thickness of the concertina wire or
     * minefield fence, or the width of the dragon's teeth is stored in
     * `width', which is measured in the same units as X and Y in the
     * `CTDB_XYO_COORD2D' struct: (patch_size/2^15) meters.
     */
#ifdef USE_LITTLE_ENDIAN
    uint16 width;
    uint8  pad;
    uint8  type;
#else
    uint8  type;
    uint8  pad;
    uint16 width;
#endif

} CTDB_LINEAR_MODEL;


typedef struct ctdb_volume_model
{
    char     model_name[32];
    uint32   model_id;
    uint32   next_state_index;
    float32  percent_damaged;
#ifdef USE_LITTLE_ENDIAN
    uint16   mes_volume_index; /* index into an array of mes_volumes
				* this is invalid if not a mes model
				*/
    uint16   num_vertices;
#else
    uint16   num_vertices;
    uint16   mes_volume_index;
#endif
    float32  vertices[1];   /* variable-length */

} CTDB_VOLUME_MODEL;


/*********************************************************************
 * AGGREGATE MODEL DEFINITIONS                                       *
 *********************************************************************/

typedef struct ctdb_agg_model_ref
{
/*changed class to sensClass 30 Jul 97 AUSTIN*/
#ifdef USE_LITTLE_ENDIAN
    uint16                   ref;   /* REFERENCE LINEAR, VOLUME,
				     * MES OR AGGREGATE MODEL LIBRARY
				     */
    uint8                    reserved;
    uint8                    sensClass; /* VOLUME, LINEAR */
#else
    uint8                    sensClass; /* VOLUME, LINEAR */
    uint8                    reserved;
    uint16                   ref;   /* REFERENCE LINEAR, VOLUME,
				     * MES OR AGGREGATE MODEL LIBRARY
				     */
#endif
} CTDB_AGG_MODEL_REF;

typedef struct ctdb_agg_model_feature
{
    CTDB_AGG_MODEL_REF       model;
#ifdef USE_LITTLE_ENDIAN
    uint16                   offset_y;
    uint16                   offset_x;
#else
    uint16                   offset_x;
    uint16                   offset_y;
#endif
} CTDB_AGG_MODEL_FEATURE;

typedef struct ctdb_aggregate_model
{
#ifdef USE_LITTLE_ENDIAN
    uint16                   reserved;
    uint16                   n_features;
#else
    uint16                   n_features;
    uint16                   reserved;
#endif
    CTDB_AGG_MODEL_FEATURE   features[1];  /* variable-length array */
} CTDB_AGGREGATE_MODEL;


/**********************************************************************
 * Patch group headers                                                *
 **********************************************************************/

typedef union ctdb_pg_data_hdr_ptr
{
    struct ctdb_grid_pg_data *grid_data;
    struct ctdb_tin_pg_data  *tin_data;
    int32                     offset;
} CTDB_PG_DATA_HDR_PTR;

/* A PATCH_GROUP is a collection of information about four contiguous
 * patches (contiguous in the sense x+y*w).  For each patch there are
 * four flags: occupied indicates there are features in the patch,
 * dyn_terr indicates that there are dynamic terrain changes to the
 * patch (future use), te16 indicates whether the patch's TEs use 8 or
 * 16 bit indices, and no_tes indicates whether or not a patch contains
 * TEs at all.  In addition, there is data about the patch group as a
 * whole:
 *     data_start points to the actual data for the cache.  Specifically,
 * it references either a CTDB_GRID_PG_DATA structure (contains_tins ==
 * FALSE) or a CTDB_TIN_PG_DATA structure (contains_tins == TRUE).  In either
 * case, that structure contains offsets for the feature/TE data itself,
 * which immediately follows the structure.
 *     cache specifies which of the 16 patch data caches this PG's data
 * should be placed in
 *     mapped is TRUE iff data_start points to actual data.  If it is FALSE,
 * the data must be read in from disk before being accessed.
 *     contains_tins is TRUE iff at least one patch in the PG is TINned
 */    
typedef struct ctdb_patch_group
{
    /* Bytes 0-3 */
    CTDB_PG_DATA_HDR_PTR       data_start;
    
    /* Bytes 4-7 */
#ifdef USE_LITTLE_ENDIAN
    unsigned unused          : 6;
    unsigned patch3_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch3_no_tes   : 1;
    unsigned patch3_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_occupied : 1;
    unsigned patch2_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch2_no_tes   : 1;
    unsigned patch2_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_occupied : 1;
    unsigned patch1_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch1_no_tes   : 1;  /* 1 => patch 1 has 0 TEs */
    unsigned patch1_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_occupied : 1;
    unsigned patch0_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch0_no_tes   : 1;  /* 1 => patch 0 has 0 TEs */
    unsigned patch0_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_occupied : 1;
    unsigned contains_tins   : 1;  /* 1 => at least one patch in this PG
				    * is TINNed.  This determines what 
				    * data_start points to */
    unsigned mapped          : 1;  /* 1 => PG data is in memory */
    unsigned cache           : 4;  /* Which size cache bucket this PG uses */
#else
    unsigned cache           : 4;  /* Which size cache bucket this PG uses */
    unsigned mapped          : 1;  /* 1 => PG data is in memory */
    unsigned contains_tins   : 1;  /* 1 => at least one patch in this PG
				    * is TINNed.  This determines what 
				    * data_start points to */
    unsigned patch0_occupied : 1;
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch0_no_tes   : 1;  /* 1 => patch 0 has 0 TEs */
    unsigned patch0_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch1_occupied : 1;
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch1_no_tes   : 1;  /* 1 => patch 1 has 0 TEs */
    unsigned patch1_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch2_occupied : 1;
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch2_no_tes   : 1;
    unsigned patch2_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned patch3_occupied : 1;
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_te16     : 1;  /* 0=8bit TEs 1=16bit TEs */
    unsigned patch3_no_tes   : 1;
    unsigned patch3_water    : 1;  /* 1 => patch 0 has a patch water level */
    unsigned unused          : 6;
#endif
} CTDB_PATCH_GROUP; 

/* Different ways of looking at the data stored for each attribute in PAT*/
typedef union ctdb_pat_data
{
    uint32  uinteger;
    int32   integer;
    int32   boolean;
    char    character;
    float32 fp;
    int32   string_index;  /* Index into global string array */
    /* Add others as needed - none may exceed 32 bits */
} CTDB_PAT_DATA; 

typedef struct ctdb
{
    /* Fixed point basis for elevation points.  Can be changed to deal
     * with future databases which have extreme elevations, or extreme
     * accuracy requirements.
     */
    float64         fixed_point_basis;
    float64         inv_fixed_point_basis;
    
    /* Tables mapping local 4 or 8 bit soil types to the global list */    
    int32             num_soil_tables;
    int32           **soil_tables;      
    int32             soil_table_size;
    int8             *soil_table_storage;   
    
    /* Pointer to the PAT table which is defined in libctdb_local.h */
    struct ctdb_pat_table *pat_table_ptr;  
 
    /* Table of sea-state combinations */
    int32                    num_water_chars;
    struct ctdb_water_chars *water_chars;
    
    /* Table of tidal zones */
    int32                    num_tidal_zones;
    struct ctdb_tidal_zone  *tidal_zones;
    
    /* Spacing between grid posts */
    int32           incr;
    float64         inv_incr;
    float64         incrD;
    
    /* Spacing between patches */
    int32           patch_incr    ;/* number of posts per patch */
    float64         patch_inv_incr;/* 1 over patch_incr */
    float64         patch_incrD   ;/* double patch_incr */
    
    /* Database size */
    int32           pages_wide;
    int32           pages_high;
    int32           patches_wide;
    int32           patches_high;
    int32           min_x;
    int32           max_x;
    int32           min_y;
    int32           max_y;
    int32           min_x_post;
    int32           max_x_post;
    int32           min_y_post;
    int32           max_y_post;
    
    /* Useful trivia */
    int32           input_format;
    float64         min_z;
    float64         max_z;
    float64         origin_northing;
    float64         origin_easting;
    int32           origin_zone_number;
    int32           origin_zone_letter;
    int32           datum;
    int32           cell_id; /* added for gcs; 
			      * GCS_ILLEGAL_CELL implies non-gcs database 
			      */

    /* origin_x and origin_y were added for GCS.  They specify the offset
     * from the center of the cell at which x and y should be 0, respectively
     */
    int32           origin_x;
    int32           origin_y;

    /* The following specify, repsectively, the left and right longitudinal
     * boundaries of the database and the bottom and top latitudinal boundaries
     * of the database
     */
    float64         west_long;
    float64         east_long;
    float64         south_lat;
    float64         north_lat;

    CTDB_TDB_TYPE   gcs_mode;

    char            name[CTDB_NAME_LENGTH];
    int16           version;
    char            date[CTDB_DATE_LENGTH];
    int32           total_memory;
    uint32          is_curved;
    CTDB_PATCH_GROUP *patch_groups;

    /* Elevation/soil post data */
    int32           elevations_resident;
    int32           elevation_seek;
    int32           soil_seek;
    int32           elevation_cache_pages;
    int32           soil_cache_pages;
    int32           elevation_direct_map_modulus;
    int32           soil_direct_map_modulus;
    int32           elevation_hit_count;
    int32           soil_hit_count;
    int32           elevation_miss_count;
    int32           soil_miss_count;
    int32          *elevation_cache_page_numbers;
    int32          *soil_cache_page_numbers;
    int32         **elevation_pages;
    int32         **soil_pages;

    /* Caches sizes are in units of CTDB_CACHE_BKT_UNITS */
    int32           patch_cache_size_map[CTDB_PATCH_CACHES];

    int32           patch_data_seek;
    int32           patch_cache_entries;
    int32           patch_hit_count;
    int32           patch_miss_count;
    struct ctdb_cache_bucket *patch_data_cache[CTDB_PATCH_CACHES];

    /* Database file */
    FILE           *file;
    char            fname[CTDB_FNAME_LENGTH];
    int32           memory_limit;

    int32          *elevations;
    int32          *soils;

    /* Quadtree of abstract features */
    int32           num_quad_data;
    union ctdb_feature_data *quad_data;

    /* Topologic relationships */
    int32           num_nodes;
    CTDB_NODE     **nodes;
    int32           node_size;
    int8           *node_storage;
    int32           num_edges;
    CTDB_EDGE     **edges;
    int32           edge_size;
    int8           *edge_storage;

    /* Linear models */
    int32               num_linear_models;
    CTDB_LINEAR_MODEL  *linear_models;

    /* Volume models */
    int32               num_volume_models;
    CTDB_VOLUME_MODEL **volume_models;
    CTDB_VOLUME_MODEL **mes_volume_models;

    /* Aggregate models */
    int32                  num_aggregate_models;
    CTDB_AGGREGATE_MODEL **aggregate_models;

    /* MES volumes */
    int32                     num_mes_volumes;  
    struct ctdb_mes_volume   *mes_volumes;
    int32                     num_mes_templates;  
    struct ctdb_mes_template *mes_templates;

    int32           n_patch_bytes;     /* Total amt of patch data for this 
					* cell, in bytes  (patch data = TEs +
					* features + virtual grids)
					*/
    ADDRESS         patch_data;	       /* Pointer to chunk o' memory with all
					* the patch data
					*/
    float64         meters_to_patch;   /* = patch_inv_incr * inv_incr  */
    uint32          is_gridded;
    uint32          is_tinned;
    uint32          is_short;
    int32           max_patch_number;
    int32           min_patch_number;
} CTDB;

#define CTDB_HEADER_RESERVED 16	/* Number of reserved words in header */

typedef struct ctdb_file_header  
{
    int32    format;
#ifdef USE_LITTLE_ENDIAN
    unsigned pad1      : 27;
    unsigned is_short  : 1;
    unsigned is_tinned : 1;
    unsigned is_gridded: 1;
    unsigned pad0      : 1;
    unsigned is_curved : 1;
#else
    unsigned is_curved : 1;
    unsigned pad0      : 1;
    unsigned is_gridded: 1;
    unsigned is_tinned : 1;
    unsigned is_short  : 1;
    unsigned pad1      : 27;
#endif
    float64  fixed_point_basis;
    int32    reserved1[CTDB_HEADER_RESERVED];
    int32    post_increment;
    int32    posts_per_patch_side;
    int32    pages_wide;
    int32    pages_high;
    int32    min_x;
    int32    max_x;
    int32    min_y;
    int32    max_y;
    int32    num_features;
    int32    patch_cache_size_map[CTDB_PATCH_CACHES];
    int32    pad2;
    float64  origin_northing;
    float64  origin_easting;
    int32    origin_zone_number;
    int32    origin_zone_letter;
    int32    datum;
    char     name[CTDB_NAME_LENGTH];
    int16    version;
    char     date[CTDB_DATE_LENGTH];
    int32    num_nodes;
    int32    node_size;
    int32    num_edges;
    int32    edge_size;
    int32    num_quad_data;
    int32    n_patch_bytes;
    int32    num_linear_models;
    int32    num_volume_models;
    int32    aux_hdr_bytes;	/* Size of auxiliary header */
    int32    num_soil_tables;
    int32    soil_table_size;
    int32    num_water_chars;
    int32    num_tidal_zones;
    int32    num_aggregate_models;
    int32    num_mes_volumes;
    int32    num_mes_templates;
} CTDB_FILE_HEADER;

/* Additional data we didn't want to try to cram into the file header */
typedef struct ctdb_aux_file_header
{
#ifdef USE_LITTLE_ENDIAN
    unsigned pad1     : 30;
    unsigned gcs_mode : 2;
#else
    unsigned gcs_mode : 2;
    unsigned pad1     : 30;
#endif
    int32 cell_id;
    int32 origin_x;
    int32 origin_y;
    float64 west_long;		/* Left-most longiitude in database */
    float64 east_long;		/* Right-most longitude in database */
    float64 south_lat;		/* Bottom-most latitude in database */
    float64 north_lat;		/* Top-most latitude in database */
} CTDB_AUX_FILE_HEADER;

/* Note: Combinations of these are achieved by or-ing the individuals 
 * qualifiers together (i.e. CTDB_INCL_LAND |  CTDB_INCL_WATER would 
 * include both terrain and water in your search.
 */
typedef enum ctdb_elev_qualifier
{
    CTDB_USE_DEFAULT        = 0,
    /* include multilevel microterrain in the search */
    CTDB_INCL_MICRO_ML      = 1, 
    /* include all volumes in the search (i.e. building, multielevation 
     * structures, etc.)
     */
    CTDB_INCL_VOLUME        = (1<<1),
    /* include all linears in the search (i.e. treelines, dragon teeth, etc.)
     */
    CTDB_INCL_LINEAR        = (1<<2),
    /* include all canopies in the search not including treelines */
    CTDB_INCL_CANOPY        = (1<<3),
    /* include laid linears in the search (i.e. roads, rivers, etc.) */
    CTDB_INCL_LAID_LINEAR   = (1<<4),
    /* include the surface material in the search (i.e. soil ) */
    CTDB_INCL_MATERIAL      = (1<<5),
    /* include the terrain skin (grid and or tin surfaces).  In most cases
     * this would not be used but the CTDB_INCL_LAND would be the query
     * that the user would want.
     */
    CTDB_INCL_TERRAIN_SKIN  = (1<<6),
    /* include default or base microterrain in the search.  In most cases
     * this would not be used but the CTDB_INCL_LAND would be the query
     * that the user would want.
     */
    CTDB_INCL_MICRO_DEF_BASE= (1<<7),
    /* include water microterrain in the search.   In most cases
     * this would not be used but the CTDB_INCL_WATER would be the query
     * that the user would want.
     */ 
    CTDB_INCL_MICRO_WATER   = (1<<8), 

    /* include individual vehicles in the search */
    CTDB_INCL_VEHICLES      = (1<<9),


    /* INSERT NEW QUALIFIERS HERE */



    /* ------  SPECIAL DEFINES  ------- */
    /* Set this to ensure that valid normal vectors are returned for
     * all surface intersections.
     */
    CTDB_COMPUTE_NORMAL         = (1<<24),

    /* Used for MES only processing */
    CTDB_ONLY_POSITIVE_SURFACES = (1<<25),
    CTDB_ONLY_NEGATIVE_SURFACES = (1<<26),
    CTDB_ALL_SURFACES           = (CTDB_ONLY_POSITIVE_SURFACES |
				   CTDB_ONLY_NEGATIVE_SURFACES),

    /* include individual tree models in the search */
    CTDB_INCL_SINGLE_TREES  = (1<<27),
    /* include patch water in the search.  this mask should probably not 
     * be used individually since the user probably is looking for all
     * water and should then use CTDB_INCL_WATER instead.
     */
    CTDB_INCL_PATCH_WATER   = (1<<28), /* PATCH WATER */
    CTDB_RETURN_MAX         = (1<<29),
    CTDB_RETURN_MIN         = (1<<30), 

    /* SPECIAL COMBINATIONS */
    /* include anything that is considered to represent water surfaces */
    CTDB_INCL_WATER         = (CTDB_INCL_PATCH_WATER | 
			       CTDB_INCL_MICRO_WATER)           ,
    /* include anything type of micro terrain (default, multilevel, base and
     * water.
     */
    CTDB_INCL_MICRO         = (CTDB_INCL_MICRO_DEF_BASE |
			       CTDB_INCL_MICRO_ML       |
			       CTDB_INCL_MICRO_WATER)           ,
    /* include all elements in the terrain database that are considered to
     * constitute the terrain surface (terrain skin).
     */
    CTDB_INCL_LAND          = (CTDB_INCL_TERRAIN_SKIN   | 
			       CTDB_INCL_MICRO_DEF_BASE |
			       CTDB_INCL_MICRO_ML)              ,
    
    /* include all classes of features in the search */
    CTDB_ALL_FEATURES       = (CTDB_INCL_MICRO      |
			       CTDB_INCL_VOLUME     |
			       CTDB_INCL_LINEAR     |
			       CTDB_INCL_CANOPY     |
			       CTDB_INCL_LAID_LINEAR  )         ,
    /* include all land and water, see the individual elements for more 
     * information.
     */
    CTDB_LAND_AND_WATER     = (CTDB_INCL_LAND |
			       CTDB_INCL_WATER)                 ,
    /* include all land and water and maintain only the maximum */
    CTDB_MAX_LAND_AND_WATER = (CTDB_LAND_AND_WATER |
			       CTDB_RETURN_MAX)                 ,
    /* include all land maintain only the maximum */
    CTDB_MAX_LAND_NO_WATER  = (CTDB_INCL_LAND |
			       CTDB_RETURN_MAX)
} CTDB_ELEV_QUALIFIER;

/*
 * CTDB PUBLIC FUNCTIONS                 
 */

/* ctdb_print_stats prints cache performance statistics.
 */
extern void 
ctdb_print_stats ( CTDB *ctdb );

/* ctdb_get_cell_id returns the GCS cell id for the given ctdb */
extern int32
ctdb_get_cell_id (
		     CTDB *ctdb
		     );
/* ctdb_get_extent returns the extent box that contains the database (i.e.
 * min(X,Y) and max(X,Y).
 */
extern void
ctdb_get_extent (
		 CTDB *ctdb,
		 float64 *mn_x,
		 float64 *mn_y,
		 float64 *mx_x,
		 float64 *mx_y
		 );

/* Return the offset of the lower left-hand corner of ctdb from the center
 * of its cell
 */
extern void
ctdb_get_origin (
		 CTDB    *ctdb,
		 float64 *x,
		 float64 *y
		 );


/* ctdb_contour_image generates a Z-pixmap format image of the contours
 * of the terrain bound by low_x,low_y to low_x+width,low_y+height, using
 * dirt_pixel as a background, and contour_pixel as the forground.  The
 * size of the image is computed from scale_mpp (meters per pixel).
 * This routine may be faster than ctdb_contour_segments for extremely
 * large terrain areas, although its use at application side is much
 * more complicated.
 */
extern void 
ctdb_contour_image (
		    CTDB   *ctdb,
		    int32   low_x,
		    int32   low_y,
		    int32   width,
		    int32   height,
		    float64 scale_mpp,
		    int32   dirt_pixel,
		    int32   contour_pixel,
		    char    img[]
		    );

/* ctdb_contour_segments calls the passed draw_fcn for each contour line
 * segment in the area bound by x_low,y_low to x_high,y_high at the
 * specified interval. seg_cnt is the number of elements in the segs[]
 * array (x0, y0, x1, y1, in meters).  seg_cnt/4 is the the number of
 * elements in the elevs[] array (one elevation for each line segment
 * int32 the segs[] array).
 */
#define CTDB_MAX_CONTOUR_PTS 10000
extern void 
ctdb_contour_segments (
		       CTDB    *ctdb,
		       float64  interval,
		       int32    x_low,
		       int32    y_low,
		       int32    x_high,
		       int32    y_high,
		       void   (*draw_fcn)()
		       );

/* ctdb_contour_route:
 *
 * Generates a route between points x0,y0 and x1,y1 which follows
 * the contour of the terrain.  The route either attempts to stay
 * high (direction==CTDB_CREST) or low (direction==CTDB_VALLEY), but
 * will not deviate more than max_deviation from the original route.
 * The route is returned in route[] and the length of the route is
 * returned in n_route.  Upon invokation, n_route should
 * contain the maximum number of points which will fit in route[].
 * When the routine returns, route[0] will be at x0,y0, and
 * route[(*n_route)-1] will be at x1,y1.
 */
#define CTDB_CREST  1
#define CTDB_VALLEY 2
extern void 
ctdb_contour_route (
		       CTDB       *ctdb,
		       float64     x0, 
		       float64     y0,
		       float64     x1, 
		       float64     y1,
		       float64     max_deviation,
		       int32       direction,
		       float64     sin_significant_slope,
		       int32      *n_route,
		       ROUTE_POINT route[]
		       );

/* ctdb_hypso_image generates a Z-pixmap format hypsometric image of the
 * terrain bound by low_x,low_y to low_x+width,low_y+height, using
 * the pixel values specified in the cmap[0..ncells-1] array.  Points
 * off the terrain database are set to the off_terrain_pixel value.
 * The size of the image is computed from scale_mpp (meters per pixel).
 * *min_z and *max_z return the elevations corresponding to cmap[0] and
 * cmap[ncells-1], respectively.  It is expected that the caller can
 * use this information to change the color map appropriately.
 *
 * NOTE: This function conditionally compiled only if the the CFLAGS has
 * -DHYPSO.  This avoids a dependency on libdither if libctdb is not being
 * used for hypsometric map drawing.
 */
extern void 
ctdb_hypso_image (
		     CTDB    *ctdb,
		     int32    low_x,
		     int32    low_y,
		     int32    width,
		     int32    height,
		     float64  scale_mpp,
		     int32    ncells,
		     unsigned long cmap[],
		     int32    off_terrain_pixel,
		     char     img[],
		     float64 *min_z,
		     float64 *max_z
		     );

/* ctdb_hypso_bitmap generates an XY-bitmap format hypsometric image
 * of the terrain bound by low_x,low_y to low_x+width,low_y+height,
 * using dithering.  Points off the terrain database are set to 0.
 * The size of the image is computed from scale_mpp (meters per
 * pixel).  min_z and max_z are passed in by the caller.  points at or
 * below min_z use an empty dither, and points at or above max_z use a
 * solid dither.  min_z and max_z are measured as height above the ellipsoid.
 * Points in between use dither patterns with density
 * proportional to elevation.  The returned bitmap is suitable for
 * a call to XPutImage() or XCreatePixmapFromBitmapData().
 *
 * NOTE: This function is conditionally compiled only if the the CFLAGS has
 * -DHYPSO.  This avoids a dependency on libdither if libctdb is not being
 * used for hypsometric map drawing.
 */
extern void 
ctdb_hypso_bitmap (
		      CTDB   *ctdb,
		      int32   low_x,
		      int32   low_y,
		      int32   width,
		      int32   height,
		      float64 scale_mpp,
		      float64 min_z,
		      float64 max_z,
		      char    bitmap[]
		      );

/* ctdb_lookup_elevation finds the elevation at the specified point on
 * the terrain database based upon the elevation grid and microterrain.
 * x and y are in meters.  The point is tested to make sure it is not
 * off the terrain database (it is not necessary for the application
 * to make this check), and 0.0 is returned in that case.
 */
extern float64 
ctdb_lookup_elevation (
		       CTDB    *ctdb,
		       float64  x,
		       float64  y
		       );

typedef struct ctdb_elev_data
{
/*changed class to sensClass 30 Jul 97 AUSTIN */
    float64               z;
    CTDB_TERRAIN_CLASS    sensClass;
    CTDB_TERRAIN_SUBCLASS subclass;    
    int32                 material;/* only valid if MES or CTDB_INCL_MATERIAL
				    * was specified.
				    */
    float64               norm_vec[XYZ];/* only valid for CTDB_FS_MES surfaces
					 */
} CTDB_ELEV_DATA; 

/* ctdb_lookup_qual_elevation finds the elevation at the specified point on
 * the terrain database based upon the the qualfier.  The point is tested 
 * to make sure it is not off the terrain database (it is not necessary for 
 * the application to make this check), and 0.0 is returned in that case where
 * the elevation is undefined.
 *
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_MAX_LAND_AND_WATER
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_LAND_AND_WATER      |
 *           CTDB_RETURN_MAX          |
 *           CTDB_RETURN_MIN          |
 *           CTDB_ALL_FEATURES        |
 *           CTDB_INCL_MATERIAL       |
 *           CTDB_ALL_SURFACES
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = 
 *
 */
#define CTDB_NO_REF_ELEV 0.0
extern float64
ctdb_lookup_qual_elevation(
			   CTDB                *ctdb,
			   float64              x,
			   float64              y,
			   CTDB_ELEV_QUALIFIER  qualifier,
			   float64              reference_elevation,
			   CTDB_ELEV_DATA       elev[1]
			   );

/* ctdb_lookup_elevation_ml is the multi-level terrain equivalent of
 * ctdb_lookup_elevation.
 */
extern float64 
ctdb_lookup_elevation_ml (
			  CTDB    *ctdb,
			  float64  x,
			  float64  y,
			  float64  reference_elevation
			  );

/* ctdb_lookup_max_elevation finds the elevation at the specified
 * point on the terrain database based upon the elevation grid,
 * microterrain, buildings, and if check_canopies is TRUE, tree
 * canopies.  x and y are in meters.  The point is tested to make sure
 * it is not off the terrain database (it is not necessary for the
 * application to make this check), and 0.0 is returned in that case.
 */
extern float64
ctdb_lookup_max_elevation (
			   CTDB   *ctdb,
			   float64 x,
			   float64 y,
			   int32   check_canopies
			   ); 
/* ctdb_lookup_elevation_mes finds all elevation at the specified point on
 * the terrain database based upon the the qualfier.  The point is tested 
 * to make sure it is not off the terrain database (it is not necessary for 
 * the application to make this check).  If no elevations have been found
 * for the specific qualifier passed then the elev[0].class = CTDB_FC_ILLEGAL.
 * The list of intersection found will not be in any specific order but the 
 * list of elevations stored in elev will be terminated by 
 * elev[nth].class = CTDB_FC_ILLEGAL.
 *
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_LAND_AND_WATER |
 *           CTDB_INCL_VOLUME    |
 *           CTDB_INCL_MATERIAL
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_LAND_AND_WATER      |
 *           CTDB_ALL_FEATURES        |
 *           CTDB_INCL_MATERIAL       |
 *           CTDB_ALL_SURFACES
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = CTDB_INCL_MATERIAL
 *
 */

#define CTDB_MAX_ELEVS 100

#define CTDB_MES_NODE_KEY uint32 
#define CTDB_MES_EDGE_KEY uint32

extern void
ctdb_lookup_elevation_mes(
                          CTDB               *ctdb, 
                          float64             x, 
                          float64             y,
                          CTDB_ELEV_QUALIFIER qual,
                          CTDB_ELEV_DATA      elev[CTDB_MAX_ELEVS]
                          );

/* ctdb_lookup_enclosure_elevation finds up to 2 elevations at the specified
 * point on the surface of the specifed enclosure in the specified mes volume
 * based upon the the qualfier.  The point is tested 
 * to make sure it is not off the terrain database (it is not necessary for 
 * the application to make this check).  If no elevations have been found
 * for the specific qualifier passed then the elevations[0].class
 * = CTDB_FC_ILLEGAL.  The return value is number of elevations.
 * The list of intersection found will not be in any specific order but the 
 * list of elevations stored in elev will be terminated by 
 * elevations[nth].class = CTDB_FC_ILLEGAL.
 *
 * This routine has the following qualifiers defined:
 * 
 * CTDB_ONLY_POSITIVE_SURFACES = (1<<25),
 * CTDB_ONLY_NEGATIVE_SURFACES = (1<<26),
 * 
 * The qualifier should be set to only one of these values but not both.
 * If both surfaces are desired set the qualifier = CTDB_USE_DEFAULT
 */
#define CTDB_ALL_POSITIVE_ANGLES 90.0
#define CTDB_ALL_NEGATIVE_ANGLES 90.0
#define CTDB_MAX_ENCLOSURE_ELEVS 3
extern int32
ctdb_lookup_enclosure_elevation(CTDB                *ctdb,
				CTDB_MES_NODE_KEY    enclosure,
				float64              x,
				float64              y,
				float64              maximum_angle,
				CTDB_ELEV_QUALIFIER  qualifier,
				CTDB_ELEV_DATA       
				     elevations[CTDB_MAX_ENCLOSURE_ELEVS] );

#define CTDB_PTS_IN_TRIANGLE  3
#define CTDB_NO_ADJACENT_TIN -1

typedef struct ctdb_tin_triangle
{
    float64  verts[CTDB_PTS_IN_TRIANGLE][3]; /* Vertices of triangle */
    int32    soil;		/* Soil type in triangle */
    int32    adj_triangles[3];	/* Indices of triangles adjacent at each 
				 * edge */
    int32    adj_patches[3];	/* Patches in which adjacent triangles reside 
				 */
} CTDB_TIN_TRIANGLE;

/* ctdb_lookup_tin_triangles generates a list of all TIN triangles in the
 * patch specified by patch_number.  The _triangles_ array is used to return
 * the triangle data (see below for more details).  *_num_triangles_ is used
 * to pass in the length of the _triangles_ array (and hence the maximum number
 * of triangles that can be returned), and to pass back the actual number
 * of triangles in the patch.  If more triangles are found than will fit in the
 * _triangles_ array, the routine returns FALSE; otherwise it returns TRUE.
 * _cross_patch_bdy_ is a boolean value which, if TRUE, specifies that TIN 
 * adjacency information should be generated for triangle edges at the patch 
 * boundary.
 *
 * The fields of each element of the _triangles_ array are filled out as
 * follows:
 *
 *    verts:          The three vertices of the triangle in clockwise order.
 *
 *    soil:           The soil information of the triangle
 *
 *    adj_triangles:  This references the triangle adjacent at each edge 
 *                    (i.e., adj_triangles[0] has the adjacency for edge 0, 
 *                    etc.)  If the edge is not on a patch boundary, 
 *                    adj_triangles stores the index in the triangles list for 
 *                    this patch at which the adjacent triangle can be found.
 *                    Otherwise it stores the index in the triangles list 
 *                    (as would be returned on another call to 
 *                    ctdb_lookup_tin_triangles) for the patch on the other 
 *                    side of the boundary.  The special value 
 *                    CTDB_NO_ADJACENT_TIN is used when no valid adjacency 
 *                    can be found for an edge.  Note that if cross_patch_bdy 
 *                    was set to FALSE in the original call, all edges on the
 *                    patch boundary will have their adjacency set to 
 *                    CTDB_NO_ADJACENT_TIN.
 *
 *    adj_patches:    This stores the patch in which each adjacent triangle 
 *                    resides--thus, adj_patches[] will be set to the current 
 *                    patch number for all internal adjacencies, and to the 
 *                    number of the appropriate patch for other adjacencies.
 *                    Note that when adj_triangles[i] is set to 
 *                    CTDB_NO_ADJACENT_TIN, the value of adj_patches[i] is
 *                    *not* guaranteed to be a valid patch number.
 */
extern int32
ctdb_lookup_tin_triangles (
			   CTDB                  *ctdb,
			   int32                  patch_number,
			   CTDB_TIN_TRIANGLE      triangles[],
			   int32                 *num_triangles,
			   int32                  cross_patch_bdy
			   );



/* ctdb_lookup_soil finds the soil type at the specified point on
 * the terrain database based upon the elevation grid and microterrain.
 * x and y are in meters.  The point is tested to make sure it is not
 * off the terrain database (it is not necessary for the application
 * to make this check), and 0 is returned in that case.
 */
extern int32 
ctdb_lookup_soil ( 
		     CTDB   *ctdb,
		     float64 x,
		     float64 y
		     );

/* ctdb_lookup_soil_ml is the multi-level terrain equivalent of
 * ctdb_lookup_soil.
 */
extern int32 
ctdb_lookup_soil_ml (
			CTDB    *ctdb,
			float64  x,
			float64  y,
			float64  reference_elevation
			);

/* ctdb_lookup_attributes is used to determine the attribute values in a given
 * PAT entry (as returned by ctdb_lookup_attributes_mes).  The PAT entry is
 * specified by the _row_in_pat_ argument.  The routine fills out an array 
 * of attribute values, one for each attribute specified by attr_list, whose 
 * length is num_attributes.  ret_list[i] is the value for attribute 
 * attr_list[i], and its interpretation is attribute-specific.  It is the 
 * caller's responsibility to allocate space for ret_list.  If any requested
 * attribute is not in the PAT, the routine will return FALSE , otherwise
 * it will return TRUE.  Callers making this call repeatedly are thus
 * expected to remember which attributes are available to avoid
 * repeated failures.  (Note that ctdb_lookup_pat_attributes() can be used
 * to determine what attribute values are available in the database PAT).
 */
extern int32  
ctdb_lookup_attributes(
		       CTDB          *ctdb,
		       int32          row_in_pat,  
		       uint32         attr_list[],
		       uint32         num_attributes,
		       CTDB_PAT_DATA  ret_list[]
		       );

/* ctdb_lookup_pat_attributes fills out the attr_list array with the FACC 
 * code for each attribute stored in the PAT of ctdb.  The caller is 
 * responsible for allocating the space for attr_list, and should fill out 
 * *num_attributes with the length of the allocated array.  The routine will 
 * overwrite this value with the actual number of attributes present.  If 
 * there are too many attributes to fit in the allocated space, the routine 
 * returns FALSE, otherwise it returns TRUE. 
 */
extern int32
ctdb_lookup_pat_attributes(
			   CTDB   *ctdb,
			   uint32  attr_list[],
			   uint32 *num_attributes
			   );

/* ctdb_lookup_attributes_mes is the same as ctdb_lookup_elevation_mes 
 * except for the fact that the material fields of elev contain the row 
 * in the PAT rather than the SIMNET mobility word.  The row in PAT
 * value can then be used as token for retrieving terrain attributes
 * through the ctdb_lookup_attributes() interface.
 */
extern void
ctdb_lookup_attributes_mes(
			   CTDB               *ctdb, 
			   float64             x, 
			   float64             y,
			   CTDB_ELEV_QUALIFIER qual,
			   CTDB_ELEV_DATA      elev[CTDB_MAX_ELEVS]
			   );

/* ctdb_lookup_attr_code converts the three character FACC attribute
 * code to a 16 bit integer value. The digits 0 through 9 are assigned
 * their actual value and the upper case characters A through Z are
 * assigned the value 10 through 36. The FACC is then converted as if
 * it were a base 36 number.
 */
extern
uint16 ctdb_lookup_attr_code ( char *facc_name );

/* ctdb_lookup_attr_name converts the 16 bit integer facc value to a
 * three character FACC attribute name. If the code is unrecognized, the 
 * character string "INVALID" is returned. If the code is the SIMNET mobility
 * attribute (CTDB_SIMNET_MOB_ATTR), the string "SIMNET" is returned. The 
 * user is responsible for allocating adequate space in the character buffer.
 * Note the maximum possible string size (in characters) returned by this 
 * routine is  CTDB_MAX_FACC_NAME.
 */
extern void
ctdb_lookup_attr_name ( uint16 code, char facc_name[] );

/* ctdb_print_description prints a description of the passed ctdb
 * database including min/max information and memory usage statistics.
 */
extern void 
ctdb_print_description ( CTDB *ctdb );


/* ctdb_place_vehicle finds the elevation, hull-to-world rotation
 * matrix, and soil type for a vehicle siting at the specified point,
 * facing the specified direction.  x, y, length, and width are in
 * meters.  <dx dy> is assumed to be a unit vector pointing in the
 * direction of the vehicle.  The elevation is returned in *z, the
 * rotation matrix is filled in, and the soil type is returned in
 * *soil.  The point is tested to make sure it is not off the
 * terrain database (it is not necessary for the application to make
 * this check); points off the database have an elevation of 0.0, a
 * flat rotation matrix pointing in the specified direction, and a
 * soil type of 0.
 */
extern void         
ctdb_place_vehicle (CTDB    *ctdb,
		    float64  x,
		    float64  y,
		    float64  length,
		    float64  width,
		    float64  dx,
		    float64  dy,
		    float64 *z,
		    float64  rotation[3][3],
		    int32   *soil
		    );

/* ctdb_place_vehicle_ml is the multi-level terrain equivalent of
 * ctdb_place_vehicle.  The value of *z is used as the reference
 * elevation.
 */
extern void         
ctdb_place_vehicle_ml (
		       CTDB             *ctdb,
		       float64           x,
		       float64           y,
		       float64           length,
		       float64           width,
		       float64           mass,
		       float64           dx,
		       float64           dy,
		       float64           *z,
		       float64            rotation[3][3],
		       int32             *soil,
		       CTDB_MES_NODE_KEY *enclosure
		       );

/* ctdb_point_on_database returns 1 if the point is on the database,
 * 0 if it is not.  All libctdb functions make this check internally.
 */
extern int32 
ctdb_point_on_database (CTDB   *ctdb,
			float64 x,
			float64 y
			);

/* ctdb_point_on_ground returns 1 if the circle centered at the
 * passed location, with the specified radius is completely ground
 * (no trees or buildings cross the circle).  The soil type at the
 * passed point is also returned so the caller can check for undesired
 * soil types.  This function is used for initializing vehicles in
 * reasonable locations.
 */ 
extern int32 
ctdb_point_on_ground (CTDB   *ctdb,
		      float64 x, 
		      float64 y,
		      float64 radius,
		      int32  *soil_ret
		      );

/* This function determines whether the passed point is within an abstract
 * feature of the passed type.
 */
extern int32
ctdb_point_within_abstract(CTDB   *ctdb,
                           int32   type_code,
                           float64 x,
                           float64 y,
			   CTDB_ABSTRACT_DATA **data);

/* ctdb_norm_within_canopy tests whether the point at pos[X],pos[Y] is 
 * within a canopy polygon.  Returns 0 if false, non-zero if true.
 */
extern int32 
ctdb_point_within_canopy (CTDB   *ctdb,
			  float64 x,
			  float64 y
			  );

/* ctdb_get_buildings returns the number of buildings within the specified
 * radius of the passed point.  If non-NULL, buildings will be filled with
 * the list of buildings (where max_count specifies the number which
 * will fit in the passed array).
 */
typedef struct ctdb_feature_outline
{
    int32   n_verts;
    float64 verts[10][3];
} CTDB_FEATURE_OUTLINE;

extern int32 
ctdb_get_buildings (CTDB                *ctdb,
		    float64              x, 
		    float64              y,
		    float64              radius,
		    int32                max_count,
		    CTDB_FEATURE_OUTLINE buildings[]
		    );

/* ctdb_point_to_point performs an intervisibility check starting at
 * the point x0,y0 and proceeding to the point x1,y1.  z0 is the eye
 * point of the viewer and zl and zh are the bottom and top of the
 * target.  width is the width of the target, and is used when
 * comparing against individual trees and buildings, otherwise a
 * zero-width target is assumed.  All these values are in meters.
 * The visible target area is returned as a floating point number
 * in the range 0.0 to 1.0 (0.0 for complete blockage, 1.0 for complete
 * visibility).  The effective opacity is returned in an argument.  Since
 * visibility can only get smaller as more features are tested,
 * knowing the minimum visibility interesting to the application can
 * greatly enhance the speed of calculation.  Even very small values
 * (such as 0.05) can greatly increase speed.  If the visibility
 * measure drops below this visibility, 0.0 will be returned.  The ray
 * from x0,y0 to x1,y1 is clipped to terrain boundaries, visibility is
 * 1.0 in areas off the terrain database.
 *
 * The last set of parameters are only examined if n_veh is non-zero;
 * hence they need not be passed unless vehicle blocking vehicle
 * intervisibility calculations are required.  If n_veh is non-zero,
 * ctdb_point_to_point first calls ctdb_vehicle_blockage with the
 * arguments specified (see ctdb_vehicle_blockage, below, for an
 * explanation of these arguments).  Note that since a target could
 * be partially blocked by terrain and partially blocked by vehicles,
 * two separate calls (one to ctdb_point_to_point with n_veh=0, and one
 * to ctdb_vehicle_blockage) will not necessarily yield the same result
 * as one call to ctdb_point_to_point with n_veh>0.
 */
extern float64 
ctdb_point_to_point (CTDB                 *ctdb,
		     float64               x0, 
		     float64               y0, 
		     float64               z0,
		     float64               x1, 
		     float64               y1, 
		     float64               zl, 
		     float64               zh,
		     float64               width,
		     CTDB_OPACITY_MAPPING_FCN map_fcn[],
		     float64              *effective_opacity,
		     float64               minimum_visibility,
		     int32                 n_veh,
		     CTDB_VEHICLE_LOCATION veh[],
		     int32                 ignore0, 
		     int32                 ignore1 
		     );

/* ctdb_ptop_raster is similar to ctdb_point_to_point except that it
 * takes an additional argument, CTDB_RASTER grid, a pre-computed
 * raster supplied by the client. It also requires the frame of reference
 * (x_min, x_max, z_min, z_max) in "projection" coordinates of the
 * raster. These are adjusted according to the result of the
 * intervis computation and passed back to the caller to reticulate the
 * object. There are two additional arguments, ftype, the
 * feature type of the target, and object_correction. Object correction
 * must be computed by the caller and is used as a scale factor to
 * "pull back" the shape of the target along parametric lines to the
 * eye point to enable intervis calculations in two steps for terrain
 * features as targets.
 */
extern float64 
ctdb_ptop_raster (
		  CTDB                 *ctdb,
		  float64               x0, 
		  float64               y0, 
		  float64               z0,
		  float64               x1, 
		  float64               y1, 
		  float64               zl, 
		  float64               zh,
		  float64               width,
		  int32                 ftype,
		  float64               object_correction,
		  float64               minimum_visibility,
		  CTDB_RASTER           grid,
		  float64              *x_min,
		  float64              *x_max,
		  float64              *z_min,
		  float64              *z_max,
		  float64              *actual_visible_area,
		  int32                 n_veh,
		  CTDB_VEHICLE_LOCATION veh[],
		  int32                 ignore0, 
		  int32                 ignore1 
		  );

/* ctdb_point_thru_point is a special version of intervisibility for
 * use with radar.  Within the passed range, objects behind the target
 * block visibility to the same extent as objects in front of it.
 */
extern float64                           
ctdb_point_thru_point (
		       CTDB                 *ctdb,
		       float64               x0, 
		       float64               y0, 
		       float64               z0,
		       float64               x1, 
		       float64               y1, 
		       float64               zl, 
		       float64               zh,
		       float64               width,
		       CTDB_OPACITY_MAPPING_FCN map_fcn[],
		       float64              *effective_opacity,
		       float64               minimum_visibility,
		       float64               range,
		       int32                 n_veh,
		       CTDB_VEHICLE_LOCATION veh[],
		       int32                 ignore0, 
		       int32                 ignore1 
		       );

/* ctdb_profile_vector generates a sequence of 3D points along the
 * specified ray from x0,y0 to x1,y1 which exactly follow the contour
 * of the terrain (as specified by the input flag include).  The points 
 * are returned in prof[] and the number
 * of points is returned in n_prof.  Upon invokation, n_prof should
 * contain the maximum number of points which will fit in prof[].  If
 * more profile points are necessary than will fit in prof[], the
 * function returns 0, otherwise it returns 1.  ctdb_profile_vector
 * clips the specified ray to terrain database boundaries, hence the
 * points x0,y0 and x1,y1 will not necessarily be the first and last
 * points in the prof[] array (although they usually will be).
 * 
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_MAX_LAND_AND_WATER
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_MAX_LAND_AND_WATER
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = 
 *
 */
extern int32 
ctdb_profile_vector (CTDB               *ctdb,
		     float64             x0, 
		     float64             y0,
		     float64             x1, 
		     float64             y1,
		     CTDB_ELEV_QUALIFIER include,
		     int32              *n_prof,
		     float64             prof[][3]
		     );

/* ctdb_find_high_ground returns the elevation of the highest point
 * between x0,y0 and x1,y1.  If non-NULL, xy_at_high will be filled in
 * with the x and y values of this point.  Hence, to get the 3D point,
 * usage would be: vec[Z] = ctdb_find_high_ground(ctdb,x0,y0,x1,y1,vec);
 * The ray from x0,y0 to x1,y1 is clipped to terrain boundaries,
 * if the ray does not cross the terrain, the returned elevation will
 * be 0.0, and the returned point will by x0,y0.
 * 
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_MAX_LAND_AND_WATER
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_MAX_LAND_AND_WATER
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = 
 *
 */
extern float64     
ctdb_find_high_ground (CTDB               *ctdb,
		       float64             x0, 
		       float64             y0,
		       float64             x1, 
		       float64             y1,
		       CTDB_ELEV_QUALIFIER qual,
		       float64            *xy_at_high
		       );

/* ctdb_vehicle_blockage performs a visibility check from the eye point
 * to the target point, comparing against the vehicles passed in the
 * veh[] array.  The locations and width are as in ctdb_point_to_point.
 * n_veh is the size of the veh[] array.  Vehicles in the veh[] array
 * are only checked if the 'used' field is set.  The height and width
 * are assumed to be projected onto the viewing plane of the observer.
 * Two vehicle can be eliminated from the search by setting ignore0
 * and ignore1 to their indecies in the veh[] array.  To avoid eliminating
 * vehicles, pass CTDB_DONT_IGNORE for ignore0 and ignore1.  As in
 * ctdb_point_to_point, the return value is in the range 0.0 for full
 * blockage, to 1.0 for full visibility.
 */
extern float64                           
ctdb_vehicle_blockage (
		       CTDB                 *ctdb,
		       float64               x0, 
		       float64               y0, 
		       float64               z0,
		       float64               x1, 
		       float64               y1, 
		       float64               zl, 
		       float64               zh,
		       float64               width,
		       int32                 n_veh,
		       CTDB_VEHICLE_LOCATION veh[],
		       int32                 ignore0,
		       int32                 ignore1 
		       );

#define CTDB_DONT_IGNORE -1
/* ctdb_apparent_size finds the apparent size of the perpendicular
 * parallelepiped described by length, width, height, given the eye
 * point x0,y0,z0, the target location x1,y1,z1 and target rotatation.
 * The apparent width and height of the object are returned, as well
 * as a location corrected to be at the bottom center of the apparent
 * location.
 *
 * Note that all inputs are required to be in the same GCS frame of
 * reference, and that the answer will be returned in that frame of
 * reference as well.
 *
 * This function is provided to allow extremely accurate intervisibility
 * test.  Ideally this should be run for the target, as well as for
 * each potentially blocking vehicle.  For applications needing a
 * more coarse measure of visibility, it is probably not necessary
 * to use this function.
 */
extern void                   
ctdb_apparent_size (
		    float64 x0, 
		    float64 y0,
		    float64 z0,
		    float64 x1,
		    float64 y1,
		    float64 z1,
		    float64 rotation[3][3],
		    float64 length, 
		    float64 width, 
		    float64 height,
		    float64 app_loc[3],
		    float64 *app_width,
		    float64 *app_height
		    );  

/* ctdb_apparent_shape is similar to ctdb_apparent_size, except that,
 * in addition to returning the "size" of the distant object, it
 * returns the raster giving a bitmap of the outline of the vehicle
 * as seen from the eye point. In addition, it returns the apparent
 * leftmost, rightmost, top and bottom of the object in "projection"
 * coordinates. This is the frame of reference for the raster.
 *
 * Note that all inputs are required to be in the same GCS frame of
 * reference, and that the answer will be returned in that frame of
 * reference as well.
 */		     
extern void  
ctdb_apparent_shape (
		     float64  x0, 
		     float64  y0,
		     float64  z0,
		     float64  x1,
		     float64  y1,
		     float64  z1,
		     float64  rotation[3][3],
		     float64  length, 
		     float64  width, 
		     float64  height,
		     float64  app_loc[3],
		     float64 *zh,
		     float64 *x_min,
		     float64 *x_max,
		     float64 *z_min,
		     float64 *z_max,
		     float64  proj[3][3],
		     float64 *dist,
		     float64 *object_correction,
		     uint32   raster[32]
		     );

/* ctdb_find_ground_intersection locates the point at which the ray
 * from <x0 y0 z0> to <x1 y1 z1> crosses through the terrain grid, a
 * microterrain polygon, a building, or (if test_treelines is
 * non-zero) a tree line (including those which surround canopies).
 * The point of intersection (+/- 1 meter along the ray when testing
 * ground polygons) is returned in pt_ret.  If no such intersection
 * can be found, the routine returns 0, otherwise one of the
 * constants: CTDB_HIT_GROUND, CTDB_HIT_BUILDING, CTDB_HIT_TREELINE,
 * or CTDB_HIT_VEHICLE, CTDB_HIT_WATER. The endpoints are clipped to
 * database boundaries.  If the return value is CTDB_HIT_VEHICLE, the
 * index of that vehicle in the veh array is returned in *hit_veh.
 * Note that individual trees are not modeled accurately enough for
 * tests of them to be meaningful.
 *
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_MAX_LAND_AND_WATER  |
 *           CTDB_ALL_FEATURES        |
 *           CTDB_INCL_SINGLE_TREES   
 *           CTDB_INCL_VEHICLES  
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_ALL_FEATURES        |
 *           CTDB_INCL_VEHICLES       |
 *           CTDB_LAND_AND_WATER      |
 *           CTDB_INCL_SINGLE_TREES   |
 *           CTDB_INCL_LINEAR |
 *           CTDB_RETURN_MAX 
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = CTDB_MAX_LAND_NO_WATER
 *
 */

/* Used for output status only */
#define CTDB_HIT_NOTHING  0x00
#define CTDB_HIT_TREELINE 0x01
#define CTDB_HIT_BUILDING 0x02
#define CTDB_HIT_GROUND   0x04
#define CTDB_HIT_VEHICLE  0x08
#define CTDB_HIT_WATER    0x10
#define CTDB_HIT_TREE     0x20
extern int32      
ctdb_find_ground_intersection (CTDB                 *ctdb,
			       float64               x0, 
			       float64               y0,
			       float64               z0,
			       float64               x1,
			       float64               y1,
			       float64               z1,
			       float64               gi_pt[3],
			       CTDB_ELEV_QUALIFIER   qual,
			       int32                 n_veh,
			       CTDB_VEHICLE_LOCATION veh[],
			       int32                 ignore0,
			       int32                 ignore1,
			       int32                *hit_veh
			       );

/* ctdb_get_treeline_segment() must be called *after* a call to
 * ctdb_find_ground_intersection().  If a treeline was intersected in the call
 * to ctdb_find_ground_intersection(), then the endpoints of that segment of
 * the treeline (in meters from the lower left coner of the database) are
 * returned in pt0[] and pt1[].  If there was no treeline intersection in the
 * last call to ctdb_find_ground_intersection(), zeros are returned in
 * pt0[] and pt1[].
 */
extern void 
ctdb_get_treeline_segment (CTDB     *ctdb,
			   float64   pt0[2],
			   float64   pt1[2]
			   );

typedef struct ctdb_background
{
    int32 type;             /* e.g, CTDB_HIT_GROUND, CTDB_HIT_TREE */
    float64 location[XYZ];  /* location, if not CTDB_HIT_NOTHING */
    union {
        int32 terrain_type; /* if CTDB_HIT_GROUND, soil_type */
    } u;
} CTDB_BACKGROUND;

/* ctdb_point_to_background projects a ray from an observer through 
 * a target and returns information about the target's background.
 *
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_MAX_LAND_AND_WATER  |
 *           CTDB_ALL_FEATURES        |
 *           CTDB_INCL_SINGLE_TREES   
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_ALL_FEATURES        |
 *           CTDB_LAND_AND_WATER      |
 *           CTDB_INCL_SINGLE_TREES   |
 *           CTDB_RETURN_MAX 
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = 
 *
 */
extern int32 
ctdb_find_background (CTDB   *ctdb,
		      float64 x0, /* observer location */
		      float64 y0, 
		      float64 z0, 
		      float64 x1, /* target location */
		      float64 y1, 
		      float64 zl,
		      float64 zh,
		      float64 width, /* target width */
		      float64 range, /* range beyond target to check */
		      CTDB_ELEV_QUALIFIER tests,/* types of backgrounds to 
						 * test for */
		      CTDB_BACKGROUND *output_ptr
		      );    /* output background information  */

/* ctdb_print_background prints a text description of the passed
 * CTDB_BACKGROUND info onto stdout.
 */
extern void 
ctdb_print_background ( CTDB_BACKGROUND *background );

/* ctdb_find_max_ratio returns a ratio of rise to run that will just clear
 * the terrain and features along the segment.  start_z is the starting
 * altitude and distance_to_start is the distance to the start of the segment.
 * A vehicle that starts at an altitude of start_z distance_to_start from
 * the start of the segment, climbs at the returned ratio, and travels to the
 * start and then along the segment would just clear the terrain and features
 * along the segment.  The signature of this function is set up this way so
 * it can be used to calculate ratios along paths.  Features specifies which
 * features to take into account.  It is the OR of the CTDB_FEATURE_XXX bits.
 * ctdb_find_max_ratio does not look at individual trees.
 *
 * This routine has the following qualifiers defined:
 * 
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * DEFAULT = CTDB_MAX_LAND_AND_WATER
 *
 * VALID is the qualifier that are considered valid for this API routine
 * 
 * VALID   = CTDB_ALL_FEATURES       |
 *           CTDB_MAX_LAND_AND_WATER
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 *
 * MUST    = CTDB_MAX_LAND_NO_WATER
 *
 */
extern float64 
ctdb_find_max_ratio (
		     CTDB               *ctdb,
		     float64             start_z,
		     float64             distance_to_start,
		     CTDB_ELEV_QUALIFIER qual,
		     float64             start[2],
		     float64             end[2]
		     );

typedef struct ctdb_mes_intersect_info
{
    CTDB              *ctdb;	/* Database where MES resides */ 
    CTDB_MES_TYPE      type;
    CTDB_MES_EDGE_KEY  entry;   /* Nearest aperture to ray entry point */
    CTDB_MES_EDGE_KEY  exit;    /* Nearest aperture to ray exit point  */
    float64            entry_pt[XYZ];
    float64            exit_pt[XYZ];
} CTDB_MES_INTERSECT_INFO;


#define CTDB_MAX_MES_ON_RAY 100	/* Can't handle more than this many 
				 * intersections */

	       
/* ctdb_find_mes_along_ray returns a pointer to an array describing 
 * intersections of the specified ray with MES structures.  climb and fall
 * specify which exterior apertures should be considered when finding the
 * closest ones to the exit and entry point--any aperture exceeding either
 * of these threholds will be eliminated from consideration.  The number of
 * MES structures intersected is returned in *num_found, and information about
 * each intersection is returned in *mes_list (the value returned is a pointer
 * into libctdb static memory--if information needs to be retained, it must
 * be copied before the next call to this routine, and the caller must not
 * attempt to free the array).  Each intersection is described by the
 * CTDB_MES_INTERESECT_INFO structure, the fields of which contain the 
 * following:
 *
 * ctdb:     The database on which the intersected MES resides.
 * type:     The type of MES structure intersected (e.g. MES_BUILDING, 
 *           MES_BRIDGE, etc.
 * entry:    The MES_EDGE_KEY describing the nearest exterior aperture to the
 *           point at which the ray enters the MES structure.
 * exit:     The MES_EDGE_KEY describing the nearest exterior aperture to the
 *           point at which the ray exits the MES structure.
 * entry_pt: The point at which the ray enters the MES.
 * exit_pt:  The point at which the ray exits the MES.
 */
extern void
ctdb_find_mes_along_ray(CTDB                     *ctdb, 
			float64                   x0, 
			float64                   y0, 
			float64                   z0,
			float64                   x1, 
			float64                   y1, 
			float64                   z1,
			float64                   climb,
			float64                   fall,
			int32                    *num_found,
			CTDB_MES_INTERSECT_INFO **mes_list);


/* ctdb_lookup_feature_info takes a point and looks to see if there
 * is a linear feature under it. If so it returns 1 and fills in the 
 * the type, width, and direction variables passed to it. Otherwise it returns
 * 0 and type and width are undefined.
 */
extern int32      
ctdb_lookup_feature_info (
			  CTDB     *ctdb,
			  float64   x, 
			  float64   y,
			  int32    *soil,
			  float64  *width,
			  float64   direction[2]
			  );

/* CTDB_SEARCH_SPACE_PTR:
 *
 * Anonymous type used to establish bounds on a search, and used to
 * track local iteration variables.
 */
typedef struct ctdb_search_space *CTDB_SEARCH_SPACE_PTR;

/* ctdb_create_search_space:
 *
 * Create a search space which will be used to bound a search of CTDB
 * feature information.  The locations specify a bounding rectangle
 * for the search.  It is not guaranteed that any returned feature
 * will actually fall inside this area, but almost all will.  It is
 * guaranteed that every feature which falls inside the search area
 * will be accessed.  The control flags * specify the search
 * conventions which should be followed:
 *
 * CTDB_BY_PATCH		Flag indicating that the search should
 *				proceed on a per patch basis, instead
 *				of traversing the entire search space
 *				uniformly.  If this flag is specified,
 *				the application must call
 *				ctdb_next_patch() to advance to the
 *				next patch, before any new features
 *				will be returned.  This functionality
 *				is provided to facilitate feature
 *				traversal on a patch-by-patch basis,
 *				to improve performance.  Note that
 *				abstract features (canopies, lakes,
 *				town names, etc.)  can traverse
 *				multiple patches, and are thus stored
 *				separately in a quad tree.  The nodes
 *				of this quad tree are treated as
 *				another kind of "patch" for the
 *				purposes of this search.
 *
 * CTDB_TREELINE_TREES		Flag indicating that the implicit
 *				trees which fall at the ends of tree
 *				lines should be returned as individual
 *				trees.  If this flag is not specified,
 *				these trees will not be returned.
 *
 * CTDB_NO_ABSTRACT		Flag indicating that no abstract features
 *				will be requested during the search.
 *				Providing this flag allows the application
 *				to avoid some initialization overhead
 *				involved in preparing for a search of
 *				abstract features.
 *
 * CTDB_NO_INDIVIDUAL_TREES	Flag indicating that no individual trees
 *				should be returned by ctdb_next_tree -- only
 *				treelines will be returned.
 */
#define CTDB_BY_PATCH            0x00000001
#define CTDB_TREELINE_TREES      0x00000002
#define CTDB_NO_ABSTRACT         0x00000004
#define CTDB_NO_INDIVIDUAL_TREES 0x00000008
extern CTDB_SEARCH_SPACE_PTR 
ctdb_create_search_space (
			     CTDB  *ctdb,
			     int32  min_x,
			     int32  min_y,
			     int32  max_x,
			     int32  max_y,
			     uint32 control_flags
			     );

/* ctdb_destroy_search_space:
 *
 * Frees the local memory used by a search space.
 */
extern void 
ctdb_destroy_search_space (
			      CTDB_SEARCH_SPACE_PTR space 
			      );

/* ctdb_reset_search_space:
 *
 * Resets the search space iterators so that the space can be retraversed
 * without having to destroy it and create another.
 */
extern void
ctdb_reset_search_space(CTDB_SEARCH_SPACE_PTR space);

/* ctdb_max_space_patches:
 *
 * Returns the maximum number of patches which might be included in the
 * search space.
 */
extern int32 
ctdb_max_space_patches (
			   CTDB_SEARCH_SPACE_PTR space 
			   );

/* CTDB_VERTICES_MAX:
 *
 * Defines the maximum number of vertices which may be used by a
 * linear feature.  Physical features which include more vertices are
 * subdivided by libctdb.
 */
#define CTDB_VERTICES_MAX 15

/* ctdb_next_linear_ref:
 *
 * Returns the next tree or linear feature in the search space.  The return
 * value indicates how many vertices are in the linear feature (or 1 for an
 * individual tree).  A return value of zero indicates that there are
 * no more trees or linear features.  In addition to the X,Y locations of
 * each linear vertex, the absolute elevations of the tops which
 * correspond to each vertex location, and feature-specific information,
 * such as foliage height and radius are returned.  NULL pointers may be
 * passed for any undesired return value.
 */
extern int32 
ctdb_next_linear_ref(CTDB_SEARCH_SPACE_PTR space,
		     float64               verts[CTDB_VERTICES_MAX][2],
		     float64               tops[CTDB_VERTICES_MAX],
		     CTDB_LINEAR_DESC     *feature_info
		     );

/* Before more linear feature types were supported by CTDB, there were only
 * treelines.  This function used to be called `ctdb_next_tree()'.  The
 * following function is provided for backward compatibility.
 */
extern int32
ctdb_next_tree(CTDB_SEARCH_SPACE_PTR space,
	       float64               verts[CTDB_VERTICES_MAX][2],
	       float64               tree_tops[CTDB_VERTICES_MAX],
	       float64              *foliage_height,
	       float64              *radius
	       );

/* ctdb_next_volume:
 *
 * Returns the next volume in the search space.  The return value
 * indicates how many vertices are in the volume.  A return value of
 * zero indicates that there are no more volumes.  In addition to
 * the X,Y locations of each vertex, the absolute elevations of the
 * roof line which correspond to each vertex location are returned.
 * NULL pointers may be passed for any undesired return value.
 */
extern int32 
ctdb_next_volume(CTDB_SEARCH_SPACE_PTR   space,
		 CTDB_TERRAIN_SUBCLASS  *type,
		 int32                  *avoid,
		 int32                  *force_id,
		 CTDB_MES_NODE_KEY      *mes, 
		 float64                 vertices[CTDB_VERTICES_MAX][2],
		 float64                 roof_elevation[CTDB_VERTICES_MAX]
		 );

/* Before more volume feature types were supported by CTDB, there were only
 * buildings.  This function used to be called `ctdb_next_building()'.  The
 * following macro is provided for backward compatibility.
 */
#define ctdb_next_building(space, verts, roof_elevation) \
  ctdb_next_volume(space, NULL, NULL, NULL, NULL, verts, roof_elevation)

/* ctdb_next_linear_gcs:
 *
 * Returns the next laid linear feature (road or river) in the search
 * space.  The return value indicates how many vertices are in the
 * feature.  A return value of zero indicates that there are no more
 * laid linear features.  In addition to the X,Y locations of each vertex,
 * the soil type, feature width, the index of the topologic edge
 * which uses this feature, the sub-index of this feature within
 * that edge, and the ctdb in which that edge and feature reside are 
 * returned.  NULL pointers may be passed for any undesired return value.
 * Note that edge_index and edge_sub_index are valid only in out_ctdb.
 */
extern int32
ctdb_next_linear_gcs (CTDB_SEARCH_SPACE_PTR space,
		      float64               vertices[CTDB_VERTICES_MAX][2],
		      int32                *soil,
		      float64              *width,
		      int32                *edge_index,
		      int32                *edge_sub_index,
		      CTDB                 **out_ctdb
		      );


/* ctdb_next_linear:
 *
 * Returns the next laid linear feature (road or river) in the search
 * space.  The return value indicates how many vertices are in the
 * feature.  A return value of zero indicates that there are no more
 * laid linear features.  In addition to the X,Y locations of each vertex,
 * the soil type, feature width, and the index of the topologic edge
 * which uses this feature and the sub-index of this feature within
 * that edge are returned.  NULL pointers may be passed for any
 * undesired return value.
 * THIS FUNCTION IS PUBLICALLY VISIBLE ONLY FOR BACKWARD COMPATIBILITY,
 * and does not work as desired if used directly by a GCS aware application.
 * Please use ctdb_next_linear_gcs; failure to do so may result in the
 * edge indices and sub-indices being invalid.
 */
extern int32 
ctdb_next_linear (
		     CTDB_SEARCH_SPACE_PTR space,
		     float64               verts[CTDB_VERTICES_MAX][2],
		     int32                *soil,
		     float64              *width,
		     int32                *edge_index,
		     int32                *edge_sub_index
		     );

/* ctdb_next_abstract:
 *
 * Returns the next "abstract" feature of a particular type.  These
 * are features which are either not tied to a real physical feature
 * represented in CTDB (such as a political boundary), or are
 * aggregations of polygonal features (such as a lake or a tree
 * canopy).  Since these features can have many vertices, the method
 * used to return them is different from that used for other features.
 * The application is returned a pointer to private memory within
 * libctdb (this should be treated as read-only).  NOTE: For
 * efficiency, the vertices in this call are returned as 32-BIT
 * FLOATING POINT.  The return value indicates how many vertices are
 * in the feature.  Note that a return value of N means that a pointer
 * to a buffer of 2*N values is returned.  A return value of zero
 * indicates that there are no more abstract features of the specified
 * type.  In addition to the X,Y location of each vertex, a pointer to
 * associated data is returned.  The format and content of the
 * associated data depends upon the type code.  The size of the data
 * (in bytes) is returned in data_size.  NULL pointers may be
 * passed for any undesired return value.
 *
 * NOTE: Make sure to update the `ctdb_is_areal()' function whenever you add a
 *       new abstract type!
 */
#define CTDB_ABSTRACT_CANOPY                1
#define CTDB_ABSTRACT_SOIL_DEFRAG           2
#define CTDB_ABSTRACT_STEEP_SLOPE           3
#define CTDB_ABSTRACT_RAILROAD              4
#define CTDB_ABSTRACT_PIPELINE              5
#define CTDB_ABSTRACT_POLITICAL_BOUNDARY    6
#define CTDB_ABSTRACT_POWERLINE             7
#define CTDB_ABSTRACT_LABEL                 8
#define CTDB_ABSTRACT_TACTICAL_SIGN         9
#define CTDB_ABSTRACT_OFFROAD_SEGMENT       10
#define CTDB_ABSTRACT_BRIDGE                11
#define CTDB_ABSTRACT_CRATER                12
#define CTDB_ABSTRACT_DITCH                 13
#define CTDB_ABSTRACT_FIGHTING_POS          14
#define CTDB_ABSTRACT_DRAGONS_TEETH         15
#define CTDB_ABSTRACT_BERM                  16
#define CTDB_ABSTRACT_CAMOUFLAGE            17
#define CTDB_ABSTRACT_RUBBLE                18
#define CTDB_ABSTRACT_SNOW                  19
#define CTDB_ABSTRACT_MICLIC                20
#define CTDB_ABSTRACT_CONTAMINANT           21
#define CTDB_ABSTRACT_HYDRO_SURFACE         22
#define CTDB_ABSTRACT_HYDRO_SUBSURFACE      23
#define CTDB_ABSTRACT_MAX /* Always Last */ 24
extern int32 
ctdb_next_abstract (
		       CTDB_SEARCH_SPACE_PTR space,
		       int32                 type_code,
		       float32             **verts,  /* -- x, y, x, y, ... */
		       CTDB_ABSTRACT_DATA  **data,
		       int32                *data_size
		       );

/* ctdb_next_patch:
 *
 * Allows the search space to advance to the next patch of feature
 * data.  This function must be called *only* if the control flag
 * CTDB_BY_PATCH was passed to ctdb_create_search_space().  If this
 * flag was not passed, then the features will be returned as though
 * they all came from a single patch.  Note that if used,
 * ctdb_next_patch() must be called prior to accessing information in
 * a patch.  Thus, if CTDB_BY_PATCH is passed to
 * ctdb_create_search_space(), ctdb_next_patch() must be called
 * *before* any features can be extracted.  A non-zero return value
 * indicates that a patch was found.  A return value of zero indicates
 * that there are no more patches.  The returned patch_number is an
 * encoding used by CTDB to uniquely identify a set of geographically
 * related data.  Every feature returned by the ctdb_next_* routines
 * exists in exactly one "patch".  In the case of abstract features, the
 * patches refer to nodes in a quad tree.
 */
extern int32 
ctdb_next_patch (
		    CTDB_SEARCH_SPACE_PTR space,
		    int32                *patch_number
		    );

/* ctdb_expand_edge:
 *
 * Expands an edge into an array of locations.  The application is
 * returned a pointer to private memory within libctdb.  This common
 * pool of memory will be reused by libctdb the next time this
 * function is called; therefore the caller should operate on the data
 * immediately.  The return value indicates how many vertices are in
 * the edge.  Note that a return value of N means that a pointer to a
 * buffer of 2*N values is returned.  The reverse flag indicates
 * whether the vertices should be extracted in last-to-first order
 * (instead of first-to-last).  NULL pointers may be passed for any
 * undesired return value.
 */
extern int32 
ctdb_expand_edge (
		     CTDB     *ctdb,
		     int32     edge_index,
		     int32     reverse,
		     float64 **verts, /* -- x, y, x, y, ... */
		     float64  *width,
		     int32    *soil
		     );

/* ctdb_edge_soil:
 *
 * Returns the soil type of the first feature in an edge (the soils of
 * all the other features will be the same).
 */
extern int32 
ctdb_edge_soil (
		   CTDB *ctdb,
		   int32 edge_index
		   );

/* ctdb_edge_type:
 *
 * Returns the abstract type of an edge, or 0 if the edge is a physical
 * feature.
 */
extern int32 
ctdb_edge_type (
		   CTDB *ctdb,
		   int32 edge_index
		   );

/* ctdb_edge_width:
 *
 * Returns the width of the first feature in an edge (the widths of all
 * the other features will be the same).
 */
extern float64 
ctdb_edge_width (
		    CTDB *ctdb,
		    int32 edge_index
		    );

/* ctdb_nearest_road:
 *
 * Finds the nearest linear feature with a road soil, and returns the
 * index of the topological edge which uses that road.  A search is
 * made of ever-widening rectangular regions of patches surrounding
 * the passed location until a road has been found, or the search
 * radius is exceeded.  A zero return value indicates failure to find
 * a road.
 *
 * THIS FUNCTION IS PUBLICALLY VISIBLE ONLY FOR BACKWARD COMPATIBILITY,
 * and does not work as desired if used directly by a GCS aware application.
 * Please use ctdb_nearest_bridge_gcs.
 */
extern int32 
ctdb_nearest_road (
		      CTDB   *ctdb,
		      float64 x, 
		      float64 y,
		      float64 search_radius,
		      int32  *edge_index,
		      int32  *edge_sub_index
		      );
/* ctdb_nearest_road_gcs:
 *
 * Finds the nearest linear feature with a road soil, and returns the
 * index of the topological edge which uses that road.  A search is
 * made of ever-widening rectangular regions of patches surrounding
 * the passed location until a road has been found, or the search
 * radius is exceeded.  in_ctdb specifies the coordinate system in which
 * x,y is set, out_ctdb will be set to the ctdb containing the linear 
 * feature (edge_index and edge_sub_index *must* be used with out_ctdb
 * for proper behavior).  A zero return value indicates failure to find
 * a road.
 *
 * The algorithm used here is just as naive as that used in 
 * ctdb_nearest_road:  we first determine the cell in which x,y sits,
 * and call ctdb_nearest_road in that database.  If, upon calculating the
 * distance from the point to that road, we find that a circle of that
 * radius extends outside the current database, we find all cells within the
 * radius, and call ctdb_nearest_road in them, too, choosing the closest
 * resultant road.  
 */
extern int32 
ctdb_nearest_road_gcs (
		      CTDB   *in_ctdb,
		      float64 x, 
		      float64 y,
		      float64 search_radius,
		      int32  *edge_index,
		      int32  *edge_sub_index,
		      CTDB  **out_ctdb
		      );

/* ctdb_nearest_bridge
 *
 * Finds the bridge closest to (x,y) in the current cell (specified by ctdb)
 * and set node_index to be the offset into ctdb->nodes of that closest 
 * bridge. Currently does a linear search of all intersections, 
 * it could probably use a better algorithm.  Returns non-zero for success,
 * 0 for failure.
 *
 * THIS FUNCTION IS PUBLICALLY VISIBLE ONLY FOR BACKWARD COMPATIBILITY,
 * and does not work as desired if used directly by a GCS aware application.
 * Please use ctdb_nearest_bridge_gcs.
 */
extern int32 
ctdb_nearest_bridge (
			CTDB    *ctdb,
			float64  x,
			float64  y,
			int32   *node_index
			);

/* ctdb_nearest_bridge_gcs
 *
 * Finds the bridge closest to x,y in the playbox.  in_ctdb specifies the
 * coordinate system in which x,y is set, out_ctdb will be set to the ctdb
 * containing the bridge (node_index *must* be used with out_ctdb for proper
 * behavior).  Returns 0 for success, non-zero for failure.
 *
 * The algorithm used here is just as naive as that used in 
 * ctdb_nearest_bridge:  we first determine the cell in which x,y sits,
 * and call ctdb_nearest_bridge in that database.  If, upon calculating the
 * distance from the point to that bridge, we find that a circle of that
 * radius extends outside the current database, we find all cells within the
 * radius, and call ctdb_nearest_bridge in them, too, choosing the closest
 * resultant bridge.  If there are no bridges in the initial cell, things
 * get uglier.  We don't have access in libctdb to a list of all the cells
 * that have ctdb's defined for them.  For now, we'll punt here and look
 * within a pre-defined radius (CTDB_MAX_BRIDGE_DIST).
 */

#define CTDB_MAX_BRIDGE_DIST 100000.0 /* Never say there is no nearest bridge
				       * unless there are no bridges within 
				       * this radius of the query point.  This
				       * value is in meters.
				       */

extern int32 
ctdb_nearest_bridge_gcs (
			    CTDB     *in_ctdb,
			    float64   x,
			    float64   y,
			    int32    *node_index,
			    CTDB    **out_ctdb
			    );

/* ctdb_bridge_info
 *
 * Fills in the dimensions and orientation of the bridge closest to x,y in the
 * coordinate system specified by ctdb.  dx and dy are the unnormalized 
 * orientation of the bridge in that coordinate system.  If there is no 
 * bridge at (x,y) in the specified ctdb, then nothing is done.
 * If any of the last 4 parameters are not needed, NULL can
 * be passed into them.
 */
extern void 
ctdb_bridge_info (
				 CTDB       *ctdb,
				 float64     x,
				 float64     y,
				 float64    *length,
				 float64    *width,
				 float64    *dx,
				 float64    *dy
				 );

/* ctdb_bridge_info_by_index
 *
 * Same as ctdb_bridge_info, but the bridge is identified by
 * an index into ctdb->nodes, rather than by its coordinates.
 * This is much more efficient if you already have the index
 * into the array. If the index is out of bounds or if the
 * intersection is not a bridge nothing is done.  Note that
 * this routine assumes that node_index refers to the specified ctdb,
 * and that all return values should be in that ctdb's frame of
 * reference.
 */
extern void 
ctdb_bridge_info_by_index (
			      CTDB       *ctdb,
			      int32	  node_index,
			      float64    *length,
			      float64    *width,
			      float64    *dx,
			      float64    *dy
			      );

/* ctdb_read reads in the ctdb format terrain database from the file
 * named fname, and puts the resulting information into the user-supplied
 * structure ctdb.  If non-zero, memory_limit specifies the maximum
 * number of bytes which the application can afford for storing elevation
 * and feature information.  If zero, memory_limit specifies that the
 * entire database should be loaded, regardless of size.  If the
 * database size exceeds the memory_limit, disk-based caching is used
 * to keep only a portion of the database resident.
 */
extern int32     
ctdb_read (char            *filename,
	   CTDB            *ctdb,
	   int32            memory_limit
	   );

/* ctdb_reread reads in the specified database, but assumes that the
 * passed ctdb structure already contains a database which should be
 * replaced.  It is *not* safe to call this with a CTDB pointer which has
 * not been passed to ctdb_read.
 */
extern int32
ctdb_reread (char            *filename,
	     CTDB            *ctdb,
	     int32            memory_limit
	     );


/*l prototypes from ct_dto.c */
extern int32
ctdb_delete_volume_feature(CTDB *ctdb,
			   char *model_name,
			   float32 last_state,
			   float64 position[3]
			   );

extern int32 
ctdb_modify_volume_feature(CTDB   *ctdb,
			   char   *model_name,
                           int32   model_type,
                           int16   force_id,
                           int16   avoid,
			   float32 prev_state,
			   float32 next_state,
			   float64 position[3],
			   float32 rotation[3]
			   );

extern int32 
ctdb_create_volume_feature(CTDB   *ctdb,
			   char   *model_name,
                           int32   model_type,
                           int16   force_id,
                           int16   avoid,
			   float32 model_state,
			   float64 position[3],
			   float32 rotation[3]
			   );

extern int32 
ctdb_create_linear_feature(
			   CTDB             *ctdb, 
			   int32             type, 
			   CTDB_LINEAR_DESC *desc_p, 
			   int32             n_verts,
			   float32          *vert    
			   );
extern int32 
ctdb_delete_linear_feature(
			   CTDB             *ctdb, 
			   int32             type, 
                           float32          *first_vert
			   );

#define MAX_TS_POLYGON_VERTS 5
typedef struct
{
    int32 soil;
    int16 reserved; /* THIS MUST BE SET TO 0 */
    int16 n_verts;
    float64 verts[MAX_TS_POLYGON_VERTS][3];
} CTDB_TS_POLYGON;

extern int32 
ctdb_create_terrain_skin(CTDB             *ctdb,
			 CTDB_TS_POLYGON   polys[],
			 int32             n_polys
			 );

extern int32 
ctdb_delete_terrain_skin(CTDB             *ctdb,
			 CTDB_TS_POLYGON   polys[],
			 int32             n_polys
			 );

extern int32 
ctdb_modify_terrain_skin(CTDB             *ctdb,
			 CTDB_TS_POLYGON   polys[],
			 int32             n_polys
			 );

/*
 * This function dynamically adds an abstract feature to the
 * existing ctdb quadtree.
 */
extern int32
ctdb_create_abstract_feature(CTDB    *ctdb,
                             int32    abstract_feature_type,
                             int32    n_verts,
                             float32 *verts,
                             int32    exdata_size,
                             int32   *exdata
                             );

/* This function deletes an abstract feature from the ctdb quadtree.
 */
extern int32
ctdb_delete_abstract_feature(CTDB    *ctdb,
                             int32    abstract_feature_type,
                             float32 *first_vtx,
                             float32 *point_inside
                             );

/* This function modifies the attributes of an existing abstract 
 * feature in the ctdb quadtree.
 */
extern int32
ctdb_modify_abstract_feature(CTDB    *ctdb,
                             int32    abstract_feature_type,
                             float32 *first_vtx,
                             float32 *point_inside,
                             int32    exdata_size,
                             int32   *exdata
                             );

/* This function will reset all the terrain patches in the ctdb
 * back to their original states. 
 * Returns zero for success, nonzero for failure.
 */
extern int32 
ctdb_dpb_reset(CTDB *ctdb, int32 new_max_patches);

/*** prototypes for functions in ct_dqbuff.c ***/

/* This function removes all dynamic quad buffers, restoring the
 * quadtree to the state found in the compiled ctdb.
 */
extern void
ctdb_clear_dynamic_abstract_features(CTDB *ctdb
                                    );

extern CTDB_VOLUME_MODEL *
ctdb_mlib_model_by_index(CTDB *ctdb,     
			 int32 model_index,
			 int32 is_mes
			 );

extern CTDB_VOLUME_MODEL *
ctdb_mlib_model_by_name(CTDB *ctdb,    
			char *model_name
			);


extern int32 
ctdb_mlib_switch(CTDB   *ctdb,
		 char   *model_name,
		 float32 prev_state,
		 float32 next_state
		 );

extern void 
ctdb_mlib_describe(CTDB_VOLUME_MODEL *model,   
		   int print_verts             
		   );

extern void 
ctdb_mlib_printtree(
		    CTDB *ctdb
		    );


/* ctdb_set_tiling_mode sets the ctdb tiling mode.  TRUE => tiling is 
 * enabled, FALSE => tiling is disabled.  If this routine is not called, 
 * tiling mode defaults to off.
 */
extern void
ctdb_set_tiling_mode ( int32 mode );


/* ctdb_get_utm reads the utm data from the ctdb data structure and returns
 * the information in the CTDB_UTM_DATA data structure.  If the request is 
 * illegal the routine will return 0 otherwise it returns 1 and the output
 * data is valid.
 */
typedef struct ctdb_utm_data
{
    int32   source       ;
    int32   datum        ;
    int32   zone_number  ;
    int32   zone_letter  ;
    float64 northing     ;
    float64 easting      ;
} CTDB_UTM_DATA;

extern int32
ctdb_get_utm_data (
		      CTDB            *ctdb,
		      CTDB_UTM_DATA   *utm
		      );

/* ctdb_read_hdr will read the ctdb database file header and auz header, and 
 * store the contents of these headers in the appropriate fields in the ctdb 
 * data structure.  Returns 1 for success, and 0 for failure.
 */
extern int32
ctdb_read_hdr (
		  char  *filename,
		  CTDB  *ctdb
		  );

/* This routine will return the southern and northern most latitudes and 
 * the eastern and western most longitudes of the terrain database pointed
 * to by CTDB *.
 */
extern void
ctdb_get_geographic_extent (
			       CTDB            *ctdb,
			       float64 *south_lat,
			       float64 *west_lng,
			       float64 *north_lat,
			       float64 *east_lng
			       );
/* This routine will retrieve the database identification information */
extern void
ctdb_get_tdb_identifier (
			    CTDB  *ctdb,
			    char  *name,
			    int32 *version
			    );
/* This routine will generate the ctdb filename based on the input */
extern char *
ctdb_generate_tdb_filename (CTDB_TDB_TYPE kind,
			    char         *tdb_path,
			    char         *tdb_name,
			    int32         format,
			    int32         cell,
			    int32         thinned
			    );
/* ctdb_lookup_soil_from_defag will return the soil value for the defrag in 
 * question.
 */
extern int32
ctdb_lookup_soil_from_defrag(CTDB *ctdb, CTDB_ABSTRACT_DATA *defrag);

/* ctdb_ocean_point_query will query the ctdb database at a the point defined
 * by (ctdb,x,y).  If the output value is 0 then the output data is invalid.
 *
 * ctdb_ocean_point_write will change the ocean attributes table that is 
 * mapped to the input point (ctdb,x,y),  this means that all ocean areas that
 * map to the same attributes table as the point will have the new
 * ocean attributes.  
 *
 * The types of output required by the user are specified 
 * by a combination of the following input parameters and NULL terminated. 
 *                         
 *                         output type       input type
 * qualifier               for query         for write
 * ---------------------   -----------       ----------
 * CTDB_SURF_HEIGHT        *float64          float64
 * CTDB_TEMP               *float64          float64
 * CTDB_PRIMARY            *CTDB_WAVE_CHARS  *CTDB_WAVE_CHARS
 * CTDB_SECONDARY          *CTDB_WAVE_CHARS  *CTDB_WAVE_CHARS
 * CTDB_TIDAL_HEIGHT       *float64          float64
 */

typedef struct ctdb_wave_chars /* Wave characteristics */
{
    float32 height;    /* In meters */
    float32 period;    /* In seconds */
    float32 speed;     /* In meters per second */
    float32 direction; /* MILS */
} CTDB_WAVE_CHARS;


typedef enum ctdb_ocean_qualifier
{
    CTDB_NULL         = 0, /* VARIABLE LIST TERMINATOR */
    CTDB_SURF_HEIGHT  = 1,
    CTDB_TEMP         = 2,
    CTDB_PRIMARY      = 3, 
    CTDB_SECONDARY    = 4,
    CTDB_TIDAL_HEIGHT = 5
} CTDB_OCEAN_QUALIFIER;

extern int32
ctdb_ocean_point_query(CTDB *ctdb,
		       float64 x,
		       float64 y,
		       CTDB_OCEAN_QUALIFIER qual,
		       ...
		       );
extern void
ctdb_ocean_point_write(CTDB *ctdb,
		       float64 x,
		       float64 y,
		       CTDB_OCEAN_QUALIFIER qual,
		       ...
		       );

/* ctdb_elev_data_to_string will convert a elevation data structure into a
 * formated string of format:
 *
 *       "Elev xxxx.xx (<terrain surface intersected>)"
 */
extern void
ctdb_elev_data_to_string(CTDB_ELEV_DATA *elev,
			 char            text[132]);
/* ctdb_sort_elev_data will sort the elevation data from highest to lowest 
 * elevation.
 *
 * Important: the array must be terminated with a 
 * CTDB_ELEV_DATA structure containing a class of CTDB_FC_ILLEGAL.
 * 
 */
extern void
ctdb_sort_elev_data(CTDB_ELEV_DATA elev[]);

#ifdef _GT_
#define fread(a,b,c,d) ctdb_init_fread(a,b,c,d)
#endif /* _GT_ */




/*************************************************
*                                                *
*   Definitions for Multi-Elevation Surfaces     *
*                                                *
*************************************************/

/* The following declarations are for exporting polygon information 
 * from libctdb.
 * NOTE: All vertices of a polygon are assumed to be coplanar.
 */
#define CTDB_MES_MAX_VERTICES_PER_POLYGON 3
typedef float64    CTDB_MES_POLYGON_VERTEX[XYZ];
typedef struct ctdb_mes_polygon
{
#ifdef USE_LITTLE_ENDIAN
    int16                    num_vertices;
    int16                    material;   /* FACC MCC attributes codes 0-999 */
#else
    int16                    material;   /* FACC MCC attributes codes 0-999 */
    int16                    num_vertices;
#endif
    float64                  normal[XYZ];
    CTDB_MES_POLYGON_VERTEX  vertices[CTDB_MES_MAX_VERTICES_PER_POLYGON];
} CTDB_MES_POLYGON;

/* ====================== MES Geometry Functions ===================== */

/* ctdb_mes_find_mes
 *
 * If there is a mes-volume within "threshold" meters of the given point,
 * this function returns TRUE, and "mes_found" points to the id of the nearest
 * mes-volume in out_ctdb.  Otherwise, this function returns FALSE, and 
 * the value of mes_found is undefined.
 */
extern int32
ctdb_mes_find_mes(
		  CTDB*               ctdb,
		  float64             x, 
		  float64             y, 
		  float64             z,
		  float64             threshold,
		  CTDB*              *out_ctdb,
		  CTDB_MES_NODE_KEY  *mes_found
		  );

/* ctdb_mes_find_mes_from_xy
 *
 * This function searches for all mes_volumes which intersect the vertical 
 * line through the given x and y.  If any are found, this function returns
 * how many, and mes_volumes[] contains their keys.  If none are found, this
 * function returns 0 and mes_volumes[] is unaffected.
 */
extern int32 
ctdb_mes_find_mes_from_xy(CTDB              *ctdb, 
			  float64            x,
			  float64            y,
			  int32              n_mes_volumes,	
			  CTDB_MES_NODE_KEY  mes_volumes[]
			  );

/* ctdb_mes_get_mes
 *
 * This routine will return the list of the vertices that make up the
 * roofline and floorline of the mes-volume.  The arguments 
 * num_floor_vertices and num_roof_vertices must be set on input
 * to the size of the allocated floorline_vertices and roofline_vertices
 * array, respectively.
 * 
 * On output, the number of vertices of the roofline and floorline are 
 * num_roof_vertices and num_floor_vertices.
 *
 * If the number of vertices of the roofline exceeds num_roof_vertices,
 * or the number of vertices of the floorline exceeds num_floor_vertices,
 * the function returns 0.
 *
 * Otherwise, this routine will return the list of roofline and floorline
 * vertices in the roofline_vertices and floorline_vertices respectively,
 * and returns 1.
 *
 * Note: NULL input for any of the vertice interfaces means that the user
 * is uninterested in the output generated for that interface.
 */
extern int32
ctdb_mes_get_mes(
		 CTDB              *ctdb,
		 CTDB_MES_NODE_KEY  mes_volume,
		 CTDB_MES_TYPE     *type,
		 int32             *num_roof_vertices,
		 float64            roofline_vertices[][XYZ],
		 int32             *num_floor_vertices,
		 float64            floorline_vertices[][XYZ]
		 );


/* ctdb_mes_find_enclosure
 *
 * This function locates the enclosure, in the given mes-volume, containing 
 * the point x,y,z.
 * 
 * If the point lies outside the mes, *enclosure is set equal to the 
 * CTDB_MES_EXTERIOR
 *
 * If the point does not lie in an enclosure (i.e. between walls), 
 * this function returns 0.  Otherwise, it returns non-zero.
 */
extern int32
ctdb_mes_find_enclosure(
			CTDB*               ctdb,
			CTDB_MES_NODE_KEY   mes,
			float64             x, 
			float64             y, 
			float64             z,
			CTDB_MES_NODE_KEY  *enclosure
			);

/* ctdb_mes_global_find_enclosure
 *
 * This function locates the mes_volume and the enclosure containing 
 * the point x,y,z.
 * 
 * If the point does not lie in an enclosure, this function returns 0. 
 * Otherwise, it returns non-zero.
 */
extern int32
ctdb_mes_global_find_enclosure(
			       CTDB*               ctdb,
			       float64             x, 
			       float64             y, 
			       float64             z,
			       CTDB*              *mes_ctdb,
			       CTDB_MES_NODE_KEY  *enclosure
			       );

/* ctdb_mes_global_find_enclosure
 *
 * This function locates the mes_volume and the enclosure containing 
 * the point (x,y,z), using guess_enclosure as an initial guess.
 * 
 * If the point does not lie in an enclosure, this function returns 0. 
 * Otherwise, it returns non-zero.
 */
extern int32
ctdb_mes_find_enclosure_with_guess(CTDB*               ctdb,
				   CTDB_MES_NODE_KEY   guess_enclosure, 
				   float64             x, 
				   float64             y, 
				   float64             z,
				   CTDB*              *found_ctdb,
				   CTDB_MES_NODE_KEY  *found_enclosure);


/* ctdb_mes_find_enclosures_from_xy
 *
 * This function searches for all mes_enclosures in the given mes_volume 
 * which intersect the vertical line through the given x and y.  
 * If any are found, this function returns how many, 
 * and found_enclosures[] contains their mes_keys.  If none are found, 
 * this function returns 0 and found_enclosures[] is unaffected. 
 */
extern int32 
ctdb_mes_find_enclosures_from_xy(CTDB              *ctdb,
				 CTDB_MES_NODE_KEY  search_mes,
				 float64            x,
				 float64            y,
				 int32              array_size,
				 CTDB_MES_NODE_KEY  found_enclosures[]
				 );

/* ctdb_mes_xy_in_enclosure
 * 
 * This function determines if a given x,y location is in a given 
 * enclosure, and returns TRUE if so.
 */
extern int32
ctdb_mes_xy_in_enclosure(
			 CTDB              *ctdb,
			 float64            x,
			 float64            y,
			 CTDB_MES_NODE_KEY  enclosure
			 );


/* ctdb_mes_get_enclosure_polygons
 *
 * Return a list of the polygons making up an enclosure.
 * If enclosure_key.enclosure_id equals CTDB_MES_INVALID_ENCLOSURE_ID,
 * all of the triangles in the mes-volume are returned.
 * If (*polygons) equals NULL, this routine will allocate sufficient space
 * for the polygons.  If (*polygons) is not NULL, this routine will assume
 * that (*polygons) points to sufficient space to store (*num_polygons)
 * number of polygons.  In either case, the caller is responsible for
 * deallocating the space when done.
 *
 * Return values:
 *    CTDB_RESULT_NO_ERROR:
 * Routine completed successfully. (*num_polygons) equals the number of
 * polygons found, and (*polygons) points to the array of polygons.
 *    CTDB_RESULT_ARRAY_TOO_SMALL: 
 * The actual number of polygons is greater than the passed value of
 * (*num_polygons).  In this case the returned value of (*num_polygons) will
 * equal the number of polygons found, and (*polygons) will be unchanged.
 *    CTDB_RESULT_INVALID_PARAMETERS:
 * One or more of the input parameters are invalid.  In this case, 
 * (*num_polygons) will be 0, and (*polygons) will be unchanged.
 */
 
typedef enum ctdb_result_code
{
    CTDB_RESULT_NO_ERROR,
    CTDB_RESULT_ARRAY_TOO_SMALL,
    CTDB_RESULT_INVALID_PARAMETERS
} CTDB_RESULT_CODE;

extern CTDB_RESULT_CODE
ctdb_mes_get_enclosure_polygons(CTDB*              ctdb,
				CTDB_MES_NODE_KEY  enclosure,
				int32             *num_polygons,
				CTDB_MES_POLYGON **polygons
				);

/* ctdb_mes_get_aperture_enclosures
 *
 * This routine will return the ids of the two enclosures joined by the
 * passed in aperture_id for the mes.  
 *
 * directions is an optional array.  If it is not-null, directions[i] will 
 * return the unit vector pointing in the direction from the aperture to the 
 * enclosure[i]
 *
 * returns 1 if successful, 0 otherwise.
 */

extern int32
ctdb_mes_get_aperture_enclosures(CTDB*              ctdb,
				 CTDB_MES_EDGE_KEY  aperture,
				 CTDB_MES_NODE_KEY  enclosures[2],
				 float64            directions[2][XYZ]
				 );

/* ctdb_mes_get_aperture_centroid
 *
 * This routine will compute the centroid of the given aperture in the
 * given mes.
 *
 * returns 1 if successful, 0 otherwise.
 */

extern int32
ctdb_mes_get_aperture_centroid(CTDB              *ctdb,
			       CTDB_MES_EDGE_KEY  aperture,
			       float64           *x,
			       float64           *y,
			       float64           *z
			       );

/* ====================== MES Topology Functions ===================== */

typedef struct ctdb_mes_aperture  CTDB_MES_EDGE;
typedef struct ctdb_mes_enclosure CTDB_MES_NODE;

/* ctdb_mes_find_node
 *
 * returns the node corresponding to the location x, y, z in the CTDB
 * for the multi-elevation structure mes.  If this point lies outside
 * of the multi-elevation structure, the "Outside" node for this
 * multi-elevation structure is returned.  If the mes is invalid
 * node is set to NULL.
 *
 * Returns 0 if this routine failed (e.g. invalid arguments, point is
 * between walls), 1 otherwise.
 */
extern int32
ctdb_mes_find_node(CTDB               *ctdb,
		   CTDB_MES_NODE_KEY   mes,
		   float64             x, 
		   float64             y, 
		   float64             z, 
		   CTDB_MES_NODE_KEY  *node
		   );

/* ctdb_mes_get_topology
 *
 * Returns the "outside" node for the given mes.
 * If the mes_id or ctdb are invalid, node is set to NULL
 */
extern void 
ctdb_mes_get_topology(CTDB               *ctdb,
		      CTDB_MES_NODE_KEY   mes,
		      CTDB_MES_NODE_KEY  *node
		      );

/* ctdb_mes_get_connected_edges
 * 
 * This routine will return the list of edges connected to the 
 * given node.  The item num_edges must be set on input to the
 * size of the allocated array edges.  If the number of edges connected
 * to the node exceeds num_edges, this function will return 0, and
 * num_edges will be set to the number of edges connected to the node.
 * This number should be used to reallocate the edges list.
 *
 * Otherwise, the function will return 1, and place all the connected
 * edges into the edges array, and set num_edges to the number of elements
 * in the array
 */
extern int32
ctdb_mes_get_connected_edges(CTDB              *ctdb,
			     CTDB_MES_NODE_KEY  node, 
			     int32             *num_edges,
			     CTDB_MES_EDGE_KEY  edges[]
			     );

/* ctdb_mes_get_connected_nodes
 *
 * This routine will return the two nodes that are connected by 
 * the given edge.
 *
 * Returns 1 if successful, 0 otherwise.
 */
extern int32
ctdb_mes_get_connected_nodes(CTDB              *ctdb,
			     CTDB_MES_EDGE_KEY  edge,
			     CTDB_MES_NODE_KEY  nodes[2]
			     );

/* ctdb_mes_get_edge_info
 * 
 * Since topological MES edges correspond to physical apertures,
 * this function will return the planar outline of the aperture corresponding
 * to the given edge.
 *
 * On input, num_vertices should be set to the size of the allocated vertices
 * array.  On output, num_vertices will report the number of vertices of the
 * outline of the aperture/edge.
 *
 * If the number of vertices of the aperture exceeds num_vertices, this
 * function returns 0, otherwise 1.
 *
 * Reports physical characteristics about the aperture in visible, open,
 * and material_composition arguments.
 *
 * Function also returns the thickness extent of the aperture from its planar
 * outline in a given direction.  Each entry of aperture_thickness array
 * corresponds to an entry in the thickness_direction array.
 */
extern int32
ctdb_mes_get_edge_info(CTDB             *ctdb,
		       CTDB_MES_EDGE_KEY edge,
		       int16            *visible,
		       int16            *open,
		       int16            *material,
		       float64           aperture_thickness[2],
		       float64           thickness_direction[2][XYZ],
		       int32            *num_vertices,
		       float64           vertices[][XYZ]
		       );


/* ctdb_mes_get_connections
 *
 * This function, when given a node, will return a list of nodes
 * and a list of edges in the parallel arrays: connected_edges and
 * connected_nodes.  From node, it is possible to traverse from an
 * edge in the connected_edges array to the corresponding node in the 
 * connected_node array, without climbing more than climb_threshold
 * or falling more than drop_threshold.
 *
 * On input, num_connections should be set to the size of the allocated
 * parallel arrays connected_edges and connected_nodes.  On output,
 * it returns the number of filtered connections found.
 * 
 * If the number of connections exceeds num_connections, this routine returns
 * 0.  Otherwise it returns 1.
 */
extern int32
ctdb_mes_get_connections(CTDB             *ctdb,
			 CTDB_MES_NODE_KEY node,
			 float64           climb_threshold,
			 float64           drop_threshold,
			 int32            *num_connections,
			 CTDB_MES_EDGE_KEY connected_edges[],
			 CTDB_MES_NODE_KEY connected_nodes[]
			 );

/* ctdb_mes_filtered_connected_edges (apertures)
 *
 * Given a node for a given mes in the given ctdb, this function will
 * return a list of connected edges to the node that do not require
 * a climb greater than climb_threshold and a fall greater than 
 * drop_threshold.
 * 
 * On input, num_edges should be set to the size of the allocated edges
 * array.  On output, num_edges reports the number of filtered connected
 * edges.
 *
 * If the number of filtered connected edges exceeds num_edges, this function
 * returns 0, 1 otherwise.
 */
extern int32
ctdb_mes_filtered_connected_edges(CTDB             *ctdb,
				  CTDB_MES_NODE_KEY node, 
				  float64           climb_threshold,
				  float64           drop_threshold,
				  int32            *num_edges,
				  CTDB_MES_EDGE_KEY edges[]
				  );

/* ctdb_mes_filtered_connected_nodes (enclosures)
 *
 * Given an edge for a given mes of a given ctdb, this function will
 * return the list of nodes connected to that node that do not
 * require a climb greater than climb_threshold and a fall greater
 * than drop_threshold.
 *
 * Since an edge can connect at most 2 nodes, nodes should be allocated
 * for a size of 2 CTDB_MES_NODES.  
 *
 * num_nodes will return the number of filtered nodes.
 */
extern void
ctdb_mes_filtered_connected_nodes(CTDB             *ctdb,
				  CTDB_MES_EDGE_KEY edge,
				  float64           climb_threshold,
				  float64           drop_threshold,
				  int32            *num_nodes,
				  CTDB_MES_NODE_KEY nodes[2]
				  );



/* ==================== MES Coordinate Conversions ================= */

/* Conversion from GCS coordinates to PO enclosure point coordinates.
 * x, y, z are GCS coordinates in the given ctdb.  
 *
 * this returns mes_id, the index into out_ctdb for the given mes.
 * The enclosure that x,y,z are in is returned in enclosure_id.
 * mes_x is the delta between the x and the origin of the mes in x in cms.
 * mes_y is the delta between the y and the origin of the mes in y in cms.
 * mes_z is the delta between the z and the origin of the mes in z.
 */

/* ctdb_gcs_to_mes_origin_offsets
 *
 * Given a x,y,z in GCS coordinates in the frame of reference of the ctdb,
 * returns the mes_id and enclosure_id (with respect to the returned out_ctdb)
 * of the mes/enclosure that contains x,y,z.  Also returns the offset,
 * measured in centimeters, between this point and the origin of the MES.
 * 
 * Returns 1 if successful, 0 otherwise.
 */

/* Use this define when user wishes to fudge the elevation up when trying 
 * to ensure that the point will be defined in the enclosure and not below
 * the floor surface.
 */
#define CTDB_MES_POSITIVE_SURFACE_FUDGE 0.010
extern int32
ctdb_gcs_to_mes_origin_offsets(CTDB*              ctdb,
			       float64            x,
			       float64            y,
			       float64            z,
			       int32             *cell,
			       CTDB_MES_NODE_KEY *enclosure,
			       int16             *mes_x,
			       int16             *mes_y,
			       float32           *mes_z
			       );

/* ctdb_mes_origin_offset_to_gcscs
 *
 * Given a mes_id and a ctdb, converts the offset of a point
 * from the mes origin (measured in centimeters), and converts
 * it to GCS coordinates with respect to the frame of reference of ctdb.
 *
 * Returns 1 if successful, 0 otherwise.
 */
extern int32
ctdb_mes_origin_offsets_to_gcscs(CTDB             *ctdb,
				 CTDB_MES_NODE_KEY mes,
				 int16             mes_x,
				 int16             mes_y,
				 float32           mes_z,
				 float64          *x,
				 float64          *y,
				 float64          *z
				 );

/* =============== MES utility functions and macros ================ */
/* Macros to convert between topology/external view of the world and the
 * internal view of the world.  Consumers must insure that the 
 * mes_id, node_id, enclosure_id, node, and edge are all valid. 
 */

/* There are three different kinds of keys:
 * an ENCLOSURE_KEY (same thing as a NODE_KEY), which consists of a mes_id
 * and an enclosure_id;
 * an APERTURE_KEY (same thing as a EDGE_KEY), which consists of a mes_id
 * and an aperture_id;
 * and a MES_KEY, which is an ENCLOSURE_KEY whose enclosure_id equals
 * CTDB_INVALID_MES_ENCLOSURE_ID.
 */

#define CTDB_MES_EXTERIOR              ((uint16) 0x0000)
#define CTDB_INVALID_MES_ENCLOSURE_ID  ((uint16) 0xffff)
#define CTDB_INVALID_MES_APERTURE_ID   ((uint16) 0xffff)
#define CTDB_INVALID_MES_ID            ((uint16) 0xffff)
#define CTDB_INVALID_MES_NODE_KEY      ((uint32) RT_INVALID_NODE_KEY)
#define CTDB_INVALID_MES_EDGE_KEY      ((uint32) RT_INVALID_NODE_KEY)
#define CTDB_INVALID_MES_KEY           ((uint32) RT_INVALID_NODE_KEY)

#define CTDB_MES_KEY(mes_id, enclosure_id)                                \
  ((mes_id << 16) | enclosure_id)

#define CTDB_MES_NODE_KEY_TO_MES_ID(node_key)                             \
( (uint16) ((uint32)(node_key) >> 16) ) 

#define CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_key)                       \
( (uint16) ((uint32)(node_key) & 0x0000ffff) )

#define CTDB_MES_NODE_IS_EXTERIOR(node_key)                               \
( CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_key)  == CTDB_MES_EXTERIOR )

#define CTDB_MES_EDGE_KEY_TO_MES_ID(edge_key)                             \
( (uint16) ((uint32)(edge_key) >> 16) )

#define CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edge_key)                        \
( (uint16) ((uint32)(edge_key) & 0x0000ffff) )

#define CTDB_MES_NODE_KEYS_EQUAL(node_key1, node_key2)                    \
  ((node_key1) == (node_key2))

#define CTDB_MES_EDGE_KEYS_EQUAL(edge_key1, edge_key2)                    \
  ((edge_key1) == (edge_key2))

#define CTDB_MES_IDS_TO_NODE_KEY(mes_id, enclosure_id, node_key)          \
  ((node_key) = (mes_id << 16) | ((enclosure_id) & 0xffff))

#define CTDB_MES_KEYS_TO_NODE_KEY(mes_key, enclosure_key, node_key)       \
  ((node_key) = ((mes_key) & 0xffff0000) | ((enclosure_key) & 0x0000ffff))

#define CTDB_MES_IDS_TO_EDGE_KEY(mes_id, aperture_id, node_key)           \
  ((node_key) = (mes_id << 16) | (aperture_id) & 0xffff)

#define CTDB_MES_KEYS_TO_EDGE_KEY(mes_key, aperture_key, edge_key)       \
  ((edge_key) = ((mes_key) & 0xffff0000) | ((aperture_key) & 0x0000ffff))

#define CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key)                           \
  CTDB_MES_IDS_TO_NODE_KEY(mes_id, CTDB_MES_EXTERIOR, mes_key)

#define CTDB_NODE_KEY_TO_MES_KEY(node_key, mes_key)                       \
CTDB_MES_ID_TO_MES_KEY(CTDB_MES_NODE_KEY_TO_MES_ID(node_key),mes_key)

#define CTDB_MES_IDS_TO_ENCLOSURE_PTR(ctdb, mes_id, enclosure_id)         \
  &((ctdb)->mes_volumes[(mes_id)].template->all_enclosures[(enclosure_id)])

#define CTDB_MES_NODE_KEY_TO_ENCLOSURE_PTR(ctdb, node_key)                \
  CTDB_MES_IDS_TO_ENCLOSURE_PTR(ctdb,                                     \
	      		     CTDB_MES_NODE_KEY_TO_MES_ID(node_key),       \
		 	     CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_key))

#define CTDB_MES_IDS_TO_APERTURE_PTR(ctdb, mes_id, aperture_id)           \
  &((ctdb)->mes_volumes[(mes_id)].template->all_apertures[(aperture_id)])

#define CTDB_MES_EDGE_KEY_TO_APERTURE_PTR(ctdb, edge_key)                 \
  CTDB_MES_IDS_TO_APERTURE_PTR(ctdb,                                      \
	      		     CTDB_MES_EDGE_KEY_TO_MES_ID(edge_key),       \
		 	     CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edge_key))


/* Definitions for invalid keys and ids
 */
#define CTDB_MES_MAKE_NODE_KEY_INVALID(node_key)                            \
  ((node_key)  = CTDB_INVALID_MES_NODE_KEY)

#define CTDB_MES_NODE_KEY_EQUALS_INVALID(node_key)                          \
  ((node_key) == CTDB_INVALID_MES_NODE_KEY)

#define CTDB_MES_MAKE_EDGE_KEY_INVALID(edge_key)                            \
  ((edge_key)  = CTDB_INVALID_MES_EDGE_KEY)

#define CTDB_MES_EDGE_KEY_EQUALS_INVALID(edge_key)                          \
  ((edge_key) == CTDB_INVALID_MES_EDGE_KEY)

#define CTDB_MES_VALID_MES_ID(ctdb, mes_id)                                 \
(((mes_id) != CTDB_INVALID_MES_ID)                                       && \
 ((mes_id) < (ctdb)->num_mes_volumes))

/* ctdb_mes_valid_node_key
 *
 * Given a mes key this API routine will check if the mes key is valid for the 
 * input ctdb database.
 * 
 * Returns 1 if VALID, 0 otherwise.
 */
extern int32
ctdb_mes_valid_node_key(CTDB             *ctdb,
			CTDB_MES_NODE_KEY key);


/* ctdb_mes_same_enclosure
 * 
 * Given a pair of mes_keys, this API routine will check if these mes keys
 * point to the same enclosure.  Returns TRUE if they are the same,
 * FALSE otherwise.
 * 
 * for the purposes of this API, the EXTERIOR of a MES is equivalent
 * to being not in a MES.  For example, this API would return TRUE
 * for an invalid mes_key and an exterior_mes_key.
 */
extern int32
ctdb_mes_same_enclosure(
			CTDB              *ctdb1,
			CTDB_MES_NODE_KEY  node_key1,
			CTDB              *ctdb2,
			CTDB_MES_NODE_KEY  node_key2
			);

#ifdef __cplusplus
}
#endif

#endif /*_LIBCTDB_H_INCLUDED*/
