/*
 * $RCSfile: libctdb_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libctdb_local.h                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Local header file for libctdb                             *
*       Created: Tue Jul  2 1990                                            *
*       Author: jesmith                                                     *
*       Author: flchamberlain    (TASC)                                     *
*       Author: cbuettner        (SAIC)                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libctdb.h>      
#include <ct_export.h>    
#include <stdlib.h>
#include <stdext.h>
#ifndef _LIBCTDB_LOCAL_H_INCLUDED
#define _LIBCTDB_LOCAL_H_INCLUDED


#if defined(hpux) || defined(_GT_) || defined(sun5)
#define random rand
#endif



/* CTDB_VISIBILITY_DEBUG is used to track which edges along an intervis
 * ray are obscuring the line of sight.  Also reported is how much the 
 * line of sight is degraded 
 */
/*#define CTDB_VISIBILITY_DEBUG*/

/* CTDB_ELEV_DEBUG is used to add allow tty debugging of the terrain element
 * elevation lookup processing
 */
/*#define CTDB_ELEV_DEBUG*/

/* CTDB_INTERVIS_METRICS is used to collect metrics about the number of 
 * patches and edges that has been intersected during an intervisibility
 * ray run.  The output is printed to the tty in a table format.
 */
/*#define CTDB_INTERVIS_METRICS*/

#define DEBUG

/* Library global variables */
extern int32 ctdb_is_tiled;		/* TRUE iff tiling is enabled */

/* Used in place of (2 * a) as a means of speeding the operations within 
 * libctdb.  Note: a must be an integer.
 */
#define CTDB_1OVER2x(a) (a>>1)
#define CTDB_2x(a) (a<<1)
#define CTDB_4x(a) (a<<2)

#define CTDB_MAX_POSTS_PER_PATCH_SIDE 5

/* Return 1 if patch contains TEs, 0 else.  pg is a patch group pointer,
 * member is the patch number within that PG.  The local routine 
 * ctdb_lookup_te_ptr should be used over this to determine if the 
 * patch contains terrain elements.
 */
#define CTDB_PATCH_HAS_TES(pg, member)            \
    ((((member)==0) && (!(pg)->patch0_no_tes)) || \
     (((member)==1) && (!(pg)->patch1_no_tes)) || \
     (((member)==2) && (!(pg)->patch2_no_tes)) || \
     (((member)==3) && (!(pg)->patch3_no_tes)))

/* Return the maximum Z value in the patch.  Assumes PG data is in memory */
#define CTDB_LOOKUP_MAX_Z(pg, member)                  \
    ((pg)->contains_tins ?                             \
     ((pg)->data_start.tin_data->max_elev[(member)]) : \
     ((pg)->data_start.grid_data->max_elev[(member)]))

/* Returns 1 if x is between lower and upper inclusive 
 * (lower must be <= upper)
 */
#define CTDB_INCLUSIVE_BETWEEN(x,lower,upper) (((x)>=(lower))&&((x)<=(upper)))

/* This macro requires a pointer to the patch group and the member number
 * within the patch group.  The output will be the number of terrain 
 * elements contained with in the patch.  Note that this assumes the
 * patch group data is in memory
 */
#define CTDB_NUM_TE(pg, member)                      \
    ((pg)->contains_tins ?                           \
     ((pg)->data_start.tin_data->num_te[(member)]) : \
     0)

/**********************************************************************
 * Terrain Element Defines                                            *
 **********************************************************************/
#define MAX_TE_INDEX_08                (uint8)-1
#define MAX_TE_INDEX_16                (uint16)-1
#define DB_BOUNDRY_08                  (uint8)-2
#define DB_BOUNDRY_16                  (uint16)-2
#define ADJT_NOT_MAPPED_INT            (uint32)-1
#define ADJT_NOT_MAPPED_08             (uint8)ADJT_NOT_MAPPED_INT
#define ADJT_NOT_MAPPED_16             (uint16)-1
#define EDGE_NOT_MAPPED_08             (uint8)-1
#define EDGE_NOT_MAPPED_16             (uint16)-1
#define MAX_ALLOWED_TE_INDEX_08        (uint8)-3
#define MAX_ALLOWED_TE_INDEX_16        (uint16)-3
#define TE_ADJ_TRI_MAP_DIMENSION        4
#define TE_VERT_DIMENSION               3

/*  CELL READER DEFINES */
#define CTDB_GEOREF_ID_LEN 6
#define CTDB_MAX_GEOREF_ID 60

/**********************************************************************
 *  CTDB LOCAL REPRESENTATION DATA STRUCTURES                         *
 **********************************************************************/
/*  GENERIC COORDINATE SYSTEM DECLARATIONS */
/* --- 64 BIT FLOAT FLAVOR --- */

typedef struct ctdb_coord2d      /* Generic two-dimensional quantity */ 
{
    float64 x;
    float64 y;
} CTDB_COORD2D;

typedef struct ctdb_coord3d      /* Generic three-dimensional quantity */
{
    float64  x;
    float64  y;
    float64  z;
} CTDB_COORD3D;

/* --- 32 BIT INTEGER FLAVOR --- */
typedef struct ctdb_coord2d_i32
{
    int32 x;
    int32 y;
} CTDB_COORD2D_I32;

typedef struct ctdb_coord3d_i32
{
    int32 x;
    int32 y;
    int32 z;
} CTDB_COORD3D_I32 ;


typedef struct te_info
{
#ifdef USE_LITTLE_ENDIAN
    uint8    adj_sides;		/* Two bits per edge */
    unsigned soil1        : 8;
    unsigned soil0        : 8;
    unsigned patch_adj    : 4;	
    unsigned vertices     : 2;
    unsigned blank1       : 2;	/* place holder */
#else
    unsigned blank1       : 2;	/* place holder */
    unsigned vertices     : 2;
    unsigned patch_adj    : 4;	
    unsigned soil0        : 8;
    unsigned soil1        : 8;
    uint8    adj_sides;		/* Two bits per edge */
#endif
} TE_INFO;      

typedef struct assoc_vertex_16
{
    uint16              index[TE_VERT_DIMENSION];
    uint16              padding1;
} ASSOC_VERTEX_16;

typedef struct assoc_vertex_08
{
    uint8               index[TE_VERT_DIMENSION];
    uint8               padding1;
} ASSOC_VERTEX_08;

typedef struct terrain_element_16
{
    CTDB_XYO_COORD2D    pt    ;  /* 4 bytes */
    float32             z     ;  /* 4 bytes */
    uint16              adj[TE_ADJ_TRI_MAP_DIMENSION]   ;  /* 8 bytes */ 
    ASSOC_VERTEX_16     vert  ;  /* 8 bytes */ 
    TE_INFO             info  ;  /* 4 bytes */
} TERRAIN_ELEMENT_16;     

typedef struct terrain_element_08
{
    CTDB_XYO_COORD2D 	pt    ;  /* 4 bytes */
    float32             z     ;  /* 4 bytes */
    uint8               adj[TE_ADJ_TRI_MAP_DIMENSION]   ;  /* 4 bytes */ 
    ASSOC_VERTEX_08    	vert  ;  /* 4 bytes */ 
    TE_INFO        	info  ;  /* 4 bytes */
} TERRAIN_ELEMENT_08;      

				      
/**********************************************************************
 * Patch data headers                                                 *
 **********************************************************************/
/* The feature and TE data for each patch group is stored with a data header
 * at its beginning.  Because patch groups containing TINned patches need
 * substantially more data, there are two formats used, one for purely
 * gridded PGs, and one for PGs containing at least one TINned patch.
 */

typedef struct ctdb_patch_table_ref
{
#ifdef USE_LITTLE_ENDIAN
    unsigned water_char : 16;	/* Index of patch water characteristics */
    unsigned soil_table : 16;	/* Index of patch soil table */
#else
    unsigned soil_table : 16;	/* Index of patch soil table */
    unsigned water_char : 16;	/* Index of patch water characteristics */
#endif
} CTDB_PATCH_TABLE_REF;

/* Data header for a completely gridded patch group.  This will *always*
 * be immediately followed by the feature data for the patch group.  The
 * starting address for the features for the first patch is
 * (address of this structure) + (sizeof(this structure)).
 * The starting address for the features for the other patches in the
 * PG can be found by
 * (address of this structure) + (patch*_offset, in 32-bit words)
 */
typedef struct ctdb_grid_pg_data
{
    float32              max_elev[CTDB_PATCHES_PER_GROUP];

    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32              water_elev[CTDB_PATCHES_PER_GROUP];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF tables[CTDB_PATCHES_PER_GROUP];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */
#ifdef USE_LITTLE_ENDIAN
    unsigned patch2_offset : 16;
    unsigned patch1_offset : 16;
    unsigned unused        : 16;
    unsigned patch3_offset : 16;
#else
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;
#endif
} CTDB_GRID_PG_DATA;

/* Data header for patch group that contains at least one TINned patch.
 * This is *always* followd by the following (in order):
 *    Feature data for each patch in PG
 *    TE data for each patch in PG
 *    Virtual grid data for each patch in PG
 * The various addresses can all be calculated as described above for the
 * CTDB_GRID_PG_DATA structure (note that the offsets for the everything
 * but features are explicitly stored for all 4 patches rather than just
 * the last 3).
 */
#define CTDB_MAX_FEATURE_PRESENT_PER_PG 60

typedef struct ctdb_tin_pg_data         
{
    float32       max_elev[CTDB_PATCHES_PER_GROUP];

    /* Water elevation in each patch.  Valid iff patch*_water is set in
     * the PG header
     */
    float32              water_elev[CTDB_PATCHES_PER_GROUP];
    
    /* Soil and water characteristic table indices for each patch */
    CTDB_PATCH_TABLE_REF tables[CTDB_PATCHES_PER_GROUP];

    /* Feature data indices, all in units of 32-bit words relative to the
     * the address of this structure */ 
#ifdef USE_LITTLE_ENDIAN 
    unsigned patch2_offset : 16;
    unsigned patch1_offset : 16;
    unsigned unused        : 16;
    unsigned patch3_offset : 16;
#else
    unsigned patch1_offset : 16;
    unsigned patch2_offset : 16;
    unsigned patch3_offset : 16;
    unsigned unused        : 16;
#endif

    /* BITS             
     *	    3 : Features
     *	    2 : Buildings
     *	    1 : TREES
     *	    0 : Microterrain  
     */
    uint8   feature_present[CTDB_MAX_FEATURE_PRESENT_PER_PG];

    /* Array[patches] numb TE in patch */
    uint16  num_te[CTDB_PATCHES_PER_GROUP];
    /* Where the TEs for each patch start.  These are in units of 32-bit
     * words and relative the starting address of this (ctdb_tin_pg_data)
     * data structure
     */
    uint32       te_offset[CTDB_PATCHES_PER_GROUP];
    /* SQRT(# of v. grid squares) for each patch */
    uint8        te_vgrid_dim[CTDB_PATCHES_PER_GROUP];
    /* Where the virtual grid starts for each patch.  These are in units
     * of 32-bit words and relative to the starting address of this
     * (ctdb_tin_pg_data) data structure
     */
    uint32       te_vgrid_offset[CTDB_PATCHES_PER_GROUP];
} CTDB_TIN_PG_DATA;


/**********************************************************************
 * Patch data cache                                                   *
 **********************************************************************/
typedef struct ctdb_cache_bucket
{
    CTDB_PATCH_GROUP *user;	/* Patch group whose data is stored here */
    int32             patch_data_start_disk; /* The offset (in bytes)
					      * from the start of the
					      * patch data section in the
					      * database file on disk at
					      * which this PGs data starts
					      */
    CTDB_FEATURE_DATA data[1];	/* Actual data.  This will be malloced
				 * to be longer than 1 word, but declaring
				 * it this way saves us a 4 bytes per bucket 
				 */
} CTDB_CACHE_BUCKET;

/* CTDB_PATCH_DATA Generic structure used to pass patch related information
 * between routines.
 */
typedef struct ctdb_patch_data
{
    int32             x;
    int32             y;
    int32             number;
    int32             member;
    CTDB_PATCH_GROUP *pg;
} CTDB_PATCH_DATA;

/* CTDB_GET_GRID_SOILMAP and CTDB_GET_TIN_SOILMAP are used by CTDB_GET_SOILMAP
 * (defined in ct_export.h) to extract the soilmap for the patch specified by 
 * the patch group pointer and the member number.
 *
 * THESE macros should not be used*/
#define CTDB_GET_GRID_SOILMAP(ctdb,pg,mem) \
    (ctdb)->soil_tables[(pg)->data_start.grid_data->tables[(mem)].soil_table]

#define CTDB_GET_TIN_SOILMAP(ctdb,pg,mem) \
    (ctdb)->soil_tables[(pg)->data_start.tin_data->tables[(mem)].soil_table]

/* CTDB_GET_SOILMAP will point to the proper soil map table for the 
 * patch specified by the patch group pointer and the member number
 */
#define CTDB_GET_SOILMAP(ctdb,pg,mem) \
    (int32 *)((pg)->contains_tins ? \
             ( CTDB_GET_TIN_SOILMAP((ctdb),(pg),(mem)) ) : \
             ( CTDB_GET_GRID_SOILMAP((ctdb),(pg),(mem)) ) )  

/* qsort typecast */
typedef int (*CTDB_QSORT) (const void *, const void *);

/**********************************************************************
 * Local Feature Grid Post Macros and Data Structures                 *
 **********************************************************************/
#define FEATURE_MASK_BIT_POSITION 20 
#define FEATURE_MASK_BIT_POSITION_MINUS_BYTE 16

/* POST_BITS formats the feature present bits from the patch group 
 * data structures
 */
#define POST_BITS(byte, odd)                                      \
  (odd ? (int32)(byte&0x0F)<<FEATURE_MASK_BIT_POSITION :          \
   (int32)(byte&0xF0)<<FEATURE_MASK_BIT_POSITION_MINUS_BYTE)

/* PP_ROW computes the number of feature present bytes per patch row */
#define PP_ROW(pps) ((pps) >> (1 + ((pps) & 0x1)))

/* PP_PATCH computes the number of feature present bytes per patch */
#define PP_PATCH(pps, pp_row) ((pps) * (pp_row))


/* POST_INDEX determines the index into the feature_present array in the 
   CTDB_TIN_PG_DATA structure.  This macro is used to determine the "byte"
   in the POST_BITS macro.
*/
#define POST_INDEX(postx, posty, member, pp_row, pp_patch)        \
  (postx>>1) + (posty * pp_row) + (member * pp_patch)

/* POST_WITHIN_PATCH_n will return the x,y that define the lower left corner
 * of the post grid 
 */
#define POST_WITHIN_PATCH_X(post,pps) (post%pps)
#define POST_WITHIN_PATCH_Y(post,pps) ((post-1)%pps)

/*
 * px,py - is the post within the patch (obtained by using the 
 *      POST_WITHIN_PATCH macro - where px,py define the low left corner of 
 *      the post box
 * member - is the patch_group member
 * pg - is the patch_group pointer
 * WARNING:  this macro assumes that
 *       1.  The PG is not purely gridded
 *       2.  The PG data is resident
 */
#define CTDB_LOOKUP_POST_FLAGS(pg,px,py,member, pp_row, pp_patch) \
    POST_BITS(pg->data_start.tin_data->         \
	      feature_present[POST_INDEX(px,py,member,pp_row, pp_patch)],\
              (px)&1)

/*********************************************************************
 * Ocean surface data structures                                     *
 *********************************************************************/

#define CTDB_NO_TIDE_ZONE 0	/* Tidal zone with no tidal offset */

#define CTDB_GET_MICRO_TIDE(ctdb,patch_number,index) \
  (ctdb)->tidal_zones[(ctdb)->water_chars[((ctdb)->patch_groups+PG_NUMBER(patch_number))->data_start.grid_data->tables[(PG_MEMBER(patch_number))].water_char].tidal_zone_index].offset;

/* Structure used to store surface characteristics for water polygons */
typedef struct ctdb_water_chars
{
    int16           tidal_zone_index;
    uint16          pad;
    float32         surface_temp;       /* In degrees C */
    float32         surf_height;        /* In meters */
    CTDB_WAVE_CHARS primary_wave;
    CTDB_WAVE_CHARS secondary_wave;
} CTDB_WATER_CHARS;

typedef struct ctdb_tidal_zone
{
    float32 offset;		/* Current tidal offset for this zone */
    /* Someday we may add a region or identifier of some kind here. */
} CTDB_TIDAL_ZONE;

/*********************************************************************
 * General use declarations                                          *
 *********************************************************************/

/* The next type is the result from clipping.  Either (a) the polygon and the
 * clipping region are totally disjoin, (b) the polygon partially intersects
 * the clip region, (c) the polygon is wholly within the clip region, or
 * (d) the clip region is wholly within the polygon.
 *
 * NOTE NOTE NOTE: The first name below, SEPARATE, SHOULD be "DISJOINT",
 * but that conflicts with the next enum type.  Should this code ever be
 * moved to its own library, change all the SEPARATEs to DISJOINTs.
 */
typedef enum 
{
    SEPARATE, 
    INTERSECT, 
    SUBJ_IN_CLIP, 
    CLIP_IN_SUBJ
} CTDB_CLIP_RESULT;

typedef enum 
{ /* NOTE: Order is important */
    OUTSIDE      = 0 ,
    INSIDE       = 1 ,
    VERTEX0      = 2 ,
    VERTEX1      = 3 ,
    VERTEX2      = 4 ,
    VERTEX3      = 5 ,
    COLINEAR0    = 6 ,
    COLINEAR1    = 7 ,
    COLINEAR2    = 8 ,
    COLINEAR3    = 9 ,
    TRIANGLE0    = 10 ,
    TRIANGLE1    = 11 ,
    DISJOINT     = 12 ,
    INTERSECTION = 13
} CTDB_POINT_STATE;

/*
 * CTDB STATUS DEFINITIONS               
 */
typedef enum
{
    CTDB_STATUS_OK                                    = 0,
    CTDB_STATUS_FAILED                                = 1,
    CTDB_STATUS_QTREE_ERROR                           = 2,
    CTDB_STATUS_FEATURE_NOT_FOUND                     = 3,
    CTDB_STATUS_TOO_MANY_FEATURES                     = 4,
    CTDB_STATUS_AREA_NOT_IN_QTREE                     = 5,
    CTDB_STATUS_NO_FILENAME                           = 6,
    CTDB_STATUS_READ_DB_FAIL                          = 7,
    CTDB_STATUS_READ_HDR_FAIL                         = 8,
    CTDB_STATUS_CANNOT_WRITE_DB                       = 9,
    CTDB_STATUS_ASSOC_TRIANGLE_MAP_MAX_EXCEEDED       = 10,
    CTDB_STATUS_ILLEGAL_VIRTUAL_GRID_MAPPING          = 11,
    CTDB_STATUS_NO_TERRAIN_DATA                       = 12,
    CTDB_STATUS_NO_VIRTUAL_GRID                       = 13,
    CTDB_STATUS_NO_ELEVATION_FOUND                    = 14,
    CTDB_STATUS_POINT_NOT_IN_CELL                     = 15,
    CTDB_STATUS_TE_CHECKED                            = 16,
    CTDB_STATUS_NOT_A_TERRAIN_ELEMENT                 = 17,
    CTDB_STATUS_VERTEX_ENCOUNTERED                    = 18,
    CTDB_STATUS_INVALID_TERRAIN_ELEMENT_INDEX_FOUND   = 19,
    CTDB_STATUS_NO_TERRAIN_ELEMENTS_IN_PATCH          = 20,
    CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT = 21,
    CTDB_STATUS_PREMATURE_TERMINATION                 = 22,
    CTDB_STATUS_POINT_OUTSIDE_PATCH                   = 23,
    CTDB_STATUS_INTERVIS_COMPLETE                     = 24,
    CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING         = 25,
    CTDB_STATUS_MAXIMUM_Z_FILTERING                   = 26,
    CTDB_STATUS_INTERVIS_EXIT_PATCH                   = 27,
    CTDB_STATUS_NO_VALID_INTERSECTIONS                = 28,
    CTDB_STATUS_READ_NODES_FAILED                     = 29,
    CTDB_STATUS_READ_EDGES_FAILED                     = 30
} CTDB_STATUS;

/**********************************************************************
 *      CTDB ELEVATION LOOKUP STORAGE                                *
 **********************************************************************/
#define CAND_TE_QUEUE_LENGTH 65536  /* 2^16 where 16 is num of bits in 
				     * uint16 
				     */
/* This data structure has been defined as a collection of data required
 * by the elevation lookup processing.
 */
typedef struct elevation_lookup_storage
{
    CTDB                  *ctdb;    /* pointer to the database data structure 
				     */
    int32		   num;     /* patch number */
    int32                  member;  /* patch member */
    CTDB_COORD2D          *dbpt;    /* db coordinate system */
    CTDB_COORD2D_I32       npt;     /* coordinates normalized */
    CTDB_COORD2D_I32       coord;   /* patch indices */         
    CTDB_PATCH_GROUP      *pg;      /* pointer to the patch_group data */
    CTDB_COORD2D           frac;    /* frac part within patch */
    int32                  te16;    /* 1=patch has 16bit TE indicies */    
/* OUTPUT DATA */
    float64                elevation;
    uint32                 te_index;/* TE being evaluated */
    int32                  soil;
    CTDB_POINT_STATE       where;
    float64               *normal; /* if !NULL then a normal is to be computed
				    * and returned to the caller.
				    */
/* TERRAIN ELEMENT DATA */
    int32                  front;   /* queue pointers */
    int32                  back;
    int32                  mark_back;
    int32                  mapped_te;
    uint32                 check_mark[CAND_TE_QUEUE_LENGTH]; /* Checked TE list
							      */
    uint32                 te_queue[CAND_TE_QUEUE_LENGTH];
    /* Both of these point to TERRAIN_ELEMENT structures. */
    ADDRESS                te;      /* address of TE being evaluated */
    ADDRESS                base;    /* Base address of TE list for patch
				     * (i.e. &TE[0]) 
				     */
    uint16                *vg_ptr;/* This is used to dynamically modify the 
				   * virtual grid mapping based on the triangle
				   * found for the lookup that referenced this 
				   * virual grid.
				   */
} ELEVATION_LOOKUP_STORAGE;

/* This structure contains the line of sight data used in intervis 
 * processing. 
 */
typedef struct ctdb_line_of_sight
{
    CTDB_COORD3D start;
    CTDB_COORD3D end;
    float64      top;
} CTDB_LINE_OF_SIGHT;

/* This structure contains the line of sigth data used in intervis 
 * processing. xyo defines the los in xyo coordinates and db
 * defines los in meters.  length and distance_to_start are required
 * by some of the co-users of the intervis engine.
 */
typedef struct ctdb_los_in_xyo_and_db
{
    CTDB_LINE_OF_SIGHT xyo;
    CTDB_LINE_OF_SIGHT db;
    float64 length;
    float64 distance_to_start;
} CTDB_LOS_IN_XYO_AND_DB;

typedef struct ctdb_post_data
{
    int32  x;
    int32  y;
    uint32 mask;
} CTDB_POST_DATA;


#define CTDB_ADD_ELEVATION_DATA(count,elev,_class,_subclass,_elevation,_material, _normal)\
  {                                                                          \
    if ((count++) < (CTDB_MAX_ELEVS-2))                                      \
    {                                                                        \
      (elev)->z        = (_elevation);                                       \
      (elev)->sensClass    = (_class);                                           \
      (elev)->subclass = (_subclass);                                        \
      (elev)->material = (_material);                                        \
      VMAT3_VEC_COPY64(_normal,(elev)->norm_vec);                            \
      (elev)++;                                  /* Move to next structure */\
      (elev)->sensClass    = CTDB_FC_ILLEGAL;/* defines end of list */           \
    }                                                                        \
 }


/*********************************************************************
 * miscellaneous constants                                           *
 *********************************************************************/
#define FUDGE_FACTOR 0.000001	/* Used when we want to move just inside
				 * a boundary.
				 */
#define CTDB_MAX_CANOPY_MODELS    32  /* The first `CTDB_MAX_CANOPY_MODELS'
				       * linear models are reserved for
				       * canopies.
				       */



/**********************************************************************
 * ct_init.c                                                          *
 *********************************************************************/
/* ctdb_alloc is a wrapper around realloc which prints a message and
 * exits if the allocation fails.  libctdb only uses realloc when
 * reading in the terrain database (presumably at program startup),
 * so this behavior is generally correct.
 */
extern ADDRESS
ctdb_alloc (
		ADDRESS  p,
		uint32   size,
		char    *reason
		);

/* ctdb_unpack_header takes the relevant information from the hdr structure
 * and translates it into the format used internally by the ctdb structure.
 */
extern void
ctdb_unpack_header (
			CTDB             *ctdb,
			CTDB_FILE_HEADER *hdr
			);

/**********************************************************************
 * ct_cache.c                                                      *
 *********************************************************************/

/* ctdb_read_elevations allocates memory for, then reads from disk, the
 * elevation data.  If memory_limit is non-zero, a cache will be enabled
 * using as little memory over memory_limit as possible.  If it is zero, or
 * it is larger than would be needed to load the data completely in memory,
 * no cache will be used.  If a cache is used, as many pages of data as will
 * fit in the cache will be pre-fetched.
 */
extern void 
ctdb_read_elevations ( 
			 CTDB *ctdb,
			 FILE *f,
			 int32 memory_limit
			 );

/* ctdb_fetch_elev_page loads the specified page of elevation data into the
 * cache.
 */
extern int32 *
ctdb_fetch_elev_page(CTDB *ctdb,
		     int32 page);

/* ctdb_fetch_soil_page loads the specified page of posted soil data into the
 * cache.
 */
extern int32 *
ctdb_fetch_soil_page(CTDB *ctdb,
		     int32 page);

/* Read the data for the specified patch group from disk and put it in
 * the patch data cache.
 */
extern CTDB_STATUS 
ctdb_fetch_patch_group (
			    CTDB             *ctdb,
			    CTDB_PATCH_GROUP *pg
			    );

/**********************************************************************
 * ct_lookup.c                                                     *
 *********************************************************************/
/*
 * ctdb_lookup_terrain_element_index returns the terrain element found
 * at the input database coordinate.  If patch number is non-negative,
 * it specifies that the returned TE must be in the specified patch
 * (this is primarily useful when the input point is on a patch boundary).
 * If patch_number is negative, it is ignored.
 */
extern CTDB_STATUS
ctdb_lookup_terrain_element_index (
				      CTDB         *ctdb,
				      CTDB_COORD2D *pt,
				      int32         patch_number,
				      uint32       *te_index 
				      );


/* Structure used to store mapping between terrain element indices in a
 * patch and triangle indices (as returned by ctdb_lookup_tin_triangles)
 * in that patch.  len stores the current allocated length of map, which
 * in turn is an array of two elements for each TE in the patch.  
 * For each i, 0 <= i < number of TEs in patch, map[i][0] is triangle index
 * for the first triangle in TE i, and map[i][1] is the triangle index for
 * the second triangle in TE i.  However, if there is only one triangle
 * in TE i, *both* will be set to the index of the triangle, and if there
 * are no triangles in the TE, both will be set to CTDB_NO_ADJACENT_TIN.
 * Thus, if one knows the TE index of an adjacent polygon, and the side
 * of the adjacent polygon at which the adjacency occurs, the triangle
 * index is patch_map.map[te index][(side index) / 2].
 */
typedef struct ctdb_patch_map
{
    int32 len;
    int32 (*map)[2];
} CTDB_PATCH_MAP;

/* ctdb_get_patch_maps sets up patch map for patch specified by patch_num 
 * (and for its neighbors if cross_patch_bdy is TRUE. If there are no TEs
 * in any of these patches, the corresponding array will not be modified.
 * The return value is the number of triangles in the patch.
 */
extern int32
ctdb_get_patch_maps (
			CTDB           *ctdb,
			int32           patch_num,
			int32           cross_patch_bdy,
			CTDB_PATCH_MAP *pm,
			CTDB_PATCH_MAP *npm,
			CTDB_PATCH_MAP *spm,
			CTDB_PATCH_MAP *epm,
			CTDB_PATCH_MAP *wpm
			);

typedef struct ctdb_terrain_triangle
{
    float64  vert[CTDB_PTS_IN_TRIANGLE][3];
    int32    soil;
    int32    assoc_triangle;/* -1=none else index in list */
    int32    index;/* original terrain element index */
} CTDB_TERRAIN_TRIANGLE;

/*
 * ctdb_lookup_tin_polygons will take as input a patch number and a
 * allocated array of CTDB_TERRAIN_TRIANGLE structures of length pts_dim.  This
 * routine will then populate the passed array with all of the terrain 
 * triangles that cover the patch.  If the allocated array
 * dimension is too small the function will return false and set n_pts to the 
 * array dimension that is required.  Otherwise true is returned and n_pts is
 * set to the number of triangles provided.
 *
 * The output CTDB_TERRAIN_TRIANGLE structure will contain the vertices which
 * define the triangle in TCC coordinates (meters) and the soil type that is 
 * contained in the triangle.
 */
extern int32
ctdb_lookup_tin_polygons (
			     CTDB                  *ctdb,
			     int32                  patch_number,
			     CTDB_TERRAIN_TRIANGLE  pts[],
			     int32                  pts_dim,
			     int32                 *n_pts
			     );

/*********************************************************************
 * ct_interv.c                                                       *
 *********************************************************************/

#ifndef X
#define X          0
#endif
#ifndef Y
#define Y          1
#endif
#ifndef Z
#define Z          2
#endif
#define EYE_VERTS  3
#define ZL         2
#define ZH         3
#define TARG_VERTS 4
#define VIS_STOP       0
#define VIS_KEEP_GOING 1

/* Structure used to store off post masks asoociated with each patch traversed
 * by ctdb_intervis
 */
typedef struct ctdb_feature_storage
{
    int32   mask;
    int32   post_mask;
    int32   patch_number;
} CTDB_FEATURE_STORAGE;


/* To make full use of ctdb_intervis(), these three need to be visible outside
 * this file.
 */
extern int32 ctdb_te_found;	/* set to TRUE by ctdb_intervis() if at least
				 * one TINned patch intersects LOS 
				 */
extern CTDB_FEATURE_STORAGE ctdb_feat_list[400]; /* Feature and post masks 
						  * found during intervis 
						  * traversal.
						  */

extern int32 ctdb_n_feat;		/* Length of above list */


/* ctdb_traverse ray breaks up the given ray at cell boundaries, and calls
 * cell_fcn on each ray segment.  If any call to cell_fcn returns VIS_STOP,
 * then so will ctdb_traverse_ray, otherwise it returns VIS_KEEP_GOING
 *
 * The format for cell_fcn is:
 *    int32 cell_fcn(CTDB   *ctdb;
 *                   float64 ray_start[EYE_VERTS],
 *                   float64 ray_end[4],
 *                   float64 cur_start[4],
 *                   float64 cur_end[4],
 *                   float64 width);
 *    where ray_start and ray_end are the end points of the entire ray in
 *    in the coordinate system specified by ctdb, and cur_start and cur_end
 *    are the coordinates of the ray clipped to the current cell.
 */
extern int32
ctdb_traverse_ray (
		      CTDB                   *ctdb,
		      float64                 x0, 
		      float64                 y0, 
		      float64                 z0,
		      float64                 x1, 
		      float64                 y1, 
		      float64                 zl, 
		      float64                 zh,
		      float64                 width,
		      int32                  (*cell_fcn)()
		      );
		      
/* ctdb_invervis traverses the posts and patches starting at start[]
 * and proceeding to end[].  lookup_fcn is called to fetch each post
 * (generally ctdb_lookup_post will be used).  post_fcn is called each
 * time a new post is traversed.  Upon exiting a patch in which at least one
 * post in the patch had features which were in the feature_mask (see
 * ct_post.h), two things happen:
 *    1.  If patch_fcn is non-NULL, it is called on the patch.
 *    2.  The post mask for the patch is stored off in the global 
 *        ctdb_feat_list list and ctdb_n_feat is incremented (this
 *        allows a post-processing routine to look at the effects of
 *        features after ctdb_intervis is done)
 *
 * Both post_fcn and patch_fcn must return either VIS_KEEP_GOING or VIS_STOP, 
 * which affect whether the algorithm makes it from start[] to end[].  If all 
 * calls return VIS_KEEP_GOING, ctdb_invervis returns a non-zero value.  If any
 * call return VIS_STOP, ctdb_invervis return a zero value.  See ct_ptop.c for 
 * example post_fcn and patch_fcn.
 */
extern int32 
ctdb_intervis ( 
		  CTDB            *ctdb,
		  float64          start[EYE_VERTS],
		  float64          end[TARG_VERTS],
		  float64          cell_start[TARG_VERTS],
		  float64          cell_end[TARG_VERTS],
		  int32            do_lookups,
		  int32           (*post_fcn)(),
		  int32           (*patch_fcn)(),
		  int32           feature_mask
		  );

/* ctdb_terrain_element_intervis will walk along the terrain skin checking
 * for edge blockage of the intervisibility ray.  This routine will only
 * process terrain elements.  Grid post processing will be handled in 
 * ctdb_intervis().
 */
extern int32
ctdb_terrain_element_intervis (
				  CTDB           *ctdb,
				  CTDB_COORD3D   *start,
				  CTDB_COORD3D   *end,
				  float64         top,
				  CTDB_COORD3D   *cell_start,
				  CTDB_COORD3D   *cell_end,
				  int32         (*vis_func)(),
				  int32           test_max_z
				  );

#ifdef CTDB_INTERVIS_METRICS
extern CTDB_COLLECTION ctdb_data;
#define PRINT_METRICS_BUFFER(a,b,c)    \
{                                    \
   ctdb_data.visibility = ctdb_visible_area; \
   ctdb_data.patch_number = c;   \
   ctdb_print_intervis_metrics(a,b);        \
}
#else
#define PRINT_METRICS_BUFFER(a,b,c)
#endif

/**********************************************************************
 * ct_8or16.c                                                      *
 *********************************************************************/
/* NOTE:  The following routines are generated from ctdb_8or16.c and are 
   stored in ctdb_16bitTE.c or ctdb_8bitTE.c.  Any routine which manipulates
   terrain elements is a candidate for incorporation into the ctdb_8or16.c
   file.  Note that every function name in ct_8or16.c must end in "XX",
   and must be declared twice here, once with "16" in place of "XX",
   and once with "08" in place of "XX".
*/

typedef enum
{
    ILLEGAL    = 0 ,  /* 0x00000000 */
    WEST       = 1 ,  /* 0x00000001 */
    EAST       = 2 ,  /* 0x00000002 */
    EASTWEST   = 3 ,  /* 0x00000003 = (0x00000001 & 0x00000002) */
    SOUTH      = 4 ,  /* 0x00000004 */
    SOUTHWEST  = 5 ,  /* 0x00000005 = (0x00000004 & 0x00000001) */
    SOUTHEAST  = 6 ,  /* 0x00000006 = (0x00000004 & 0x00000002)  */
    NORTH      = 8 ,  /* 0x00000008 */
    NORTHWEST  = 9 ,  /* 0x00000009 = (0x00000008 & 0x00000001) */
    NORTHEAST  = 10,  /* 0x0000000a = (0x00000008 & 0x00000002) */
    NORTHSOUTH = 12   /* 0x0000000c = (0x00000008 & 0x00000004) */
} CTDB_BEARING;

typedef struct ctdb_ptop_te_data
{
    CTDB_BEARING      patch_to;  /* direction to the next patch to process */
    CTDB_BEARING      motion;  /* overall direction of the intervis ray */
    uint32            this_te;  /* the terrain element being processed */
    int32             entry_side;  /* side of entry in this_te */
    float64           tray;  /* current tray location */
} CTDB_PTOP_TE_DATA;

typedef struct ctdb_contour_state
{
    int32    seg_cnt;
    float64  segs[CTDB_MAX_CONTOUR_PTS];
    float64  elevs[CTDB_MAX_CONTOUR_PTS/4];
    float64  interval;
    float64  inv_interval;
    float64  max_delta;
    void     (*draw_fcn)();
} CTDB_CONTOUR_STATE;

/* ctdb_print_te_elementXX prints information about the index'th terrain
 * element, where the "0th" terrain element is defined to start at te.
 */
extern void
ctdb_print_te_element16 (
			     TERRAIN_ELEMENT_16  *te,
			     uint32               index
			     );


extern void
ctdb_print_te_element08 (
			     TERRAIN_ELEMENT_08  *te,
			     uint32               index
			     );


/* ctdb_lookup_terrain_element_elevation16 finds the elevation on the terrain
 * skin (as defined by the TIN), given that els has been correctly set up
 * (as by InitLocalStorageData(), a routine private to ct_lookup.c.  
 * This routine should not be called externally to ct_lookup.c; it is defined
 * here only because it must reside in ct_8or16.c because it deals with terrain
 * elements.
 */
extern CTDB_STATUS
ctdb_lookup_terrain_element_elevation16 ( 
					    ELEVATION_LOOKUP_STORAGE *els 
					    );


extern CTDB_STATUS
ctdb_lookup_terrain_element_elevation08 ( 
					    ELEVATION_LOOKUP_STORAGE *els 
					    );


/* ctdb_get_te_edge_indexXX is used to perform an ordered (clockwise) traversal
 * of the edges of a TE.  It takes as inputs a pointer to the start of the TEs
 * for the patch of interest (te), the specific TE under consideration (index),
 * and the last side we considered (side), and returns in *edge the next edge 
 * to look at.
 */
extern CTDB_STATUS
ctdb_get_te_edge_index16 (
			     TERRAIN_ELEMENT_16  *te,
			     uint32               index,
			     int32                side,
			     int32                edge[]
			     );

extern CTDB_STATUS
ctdb_get_te_edge_index08 (
			     TERRAIN_ELEMENT_08  *te,
			     uint32               index,
			     int32                side,
			     int32                edge[]
			     );


/* ctdb_ptop_te_compare_initXX takes a pointer to a ctdb database, a line
 * of sight, the starting address for the TEs in a patch, a pointer 
 * to a partially initialized CTDB_PTOP_TE_DATA structure, and a function
 * to call on any intersection found between the LOS and TE segments, and 
 * finds a valid starting TE within the patch (specified by ptop), at or 
 * past the starting location along the intervis ray.  For an example of 
 * how to use this, see the function ctdb_terrain_element_intervis() in 
 * ct_interv.c
 */
extern CTDB_STATUS
ctdb_ptop_te_compare_init16 (
				CTDB                   *ctdb,
				CTDB_LOS_IN_XYO_AND_DB *los,
				TERRAIN_ELEMENT_16     *te,
				CTDB_PTOP_TE_DATA      *ptop,
				int32                 (*vis_func)()
				);

extern CTDB_STATUS
ctdb_ptop_te_compare_init08 (
				CTDB                   *ctdb,
				CTDB_LOS_IN_XYO_AND_DB *los,
				TERRAIN_ELEMENT_08     *te,
				CTDB_PTOP_TE_DATA      *ptop,
				int32                 (*vis_func)()
				);


/* ctdb_ptop_te_compareXX checks intervisibility due to terrain elements in a 
 * patch.  The arguments are the same as for ctdb_ptop_te_compare_initXX, 
 * above.
 */
extern CTDB_STATUS
ctdb_ptop_te_compare16 (
			   CTDB                   *ctdb,
			   CTDB_LOS_IN_XYO_AND_DB *los,
			   TERRAIN_ELEMENT_16     *te,
			   CTDB_PTOP_TE_DATA      *ptop,
			   int32                 (*vis_func)()
			   );

extern CTDB_STATUS
ctdb_ptop_te_compare08 (
			   CTDB                   *ctdb,
			   CTDB_LOS_IN_XYO_AND_DB *los,
			   TERRAIN_ELEMENT_08     *te,
			   CTDB_PTOP_TE_DATA      *ptop,
			   int32                 (*vis_func)()
			   );

/* ctdb_terrain_element_densityXX returns the total number of TIN triangles
 * in a patch.  Note that this is different from the number of terrain elements
 * in a patch because TEs need not be triangles.
 */
extern int32
ctdb_terrain_element_density16 (
				  CTDB  *ctdb,
				  int32  patch_number
				  );

extern int32
ctdb_terrain_element_density08 (
				  CTDB  *ctdb,
				  int32  patch_number
				  );

/* ctdb_find_te_contoursXX generates contour lines in a patch due to the TIN
 * surface.  For an example of how to use this, see ct_contour.c
 *
 * The returned values in state are all with respect to src_ctdb
 */
extern void
ctdb_find_te_contours16 (
			    CTDB               *ctdb,
			    CTDB               *src_ctdb,
			    int32               patch_number,
			    float64             fudge_factor,
			    CTDB_CONTOUR_STATE *state
			    );

extern void
ctdb_find_te_contours08 (
			    CTDB               *ctdb,
			    CTDB               *src_ctdb,
			    int32               patch_number,
			    float64             fudge_factor,
			    CTDB_CONTOUR_STATE *state
			    );

/* ctdb_find_te_z_bounds_in_patchXX returns, in *min_z and *max_z, the minimum 
 * and maximum z values for TE's in the patch defined by pg and member.
 */
extern void
ctdb_find_te_z_bounds_in_patch16 (
				     CTDB             *ctdb,
				     int32             patch_number,
				     float64          *min_z,
				     float64          *max_z
				     );

extern void
ctdb_find_te_z_bounds_in_patch08 (
				     CTDB             *ctdb,
				     int32             patch_number,
				     float64          *min_z,
				     float64          *max_z
				     );

/* Set up patch map for patch specified by patch_num (and for its neighbors
 * if cross_patch_bdy is TRUE.  Note that if the patch has no TEs, pm will
 * not be changed.  The return value is the number of triangles in the patch.
 */
extern int32
ctdb_get_patch_maps_16 (
			   CTDB             *ctdb,
			   int32             patch_num,
			   CTDB_PATCH_GROUP *pg,
			   int32             mem,
			   CTDB_PATCH_MAP   *pm
			   );

extern int32
ctdb_get_patch_maps_08 (
			   CTDB             *ctdb,
			   int32             patch_num,
			   CTDB_PATCH_GROUP *pg,
			   int32             mem,
			   CTDB_PATCH_MAP   *pm
			   );

/* ctdb_lookup_tin_triangles_XX generates a list of all TIN triangles in the
 * patch specified by patch_number.  The _triangles_ array is used to return
 * the triangle data (see below for more details).  *_num_triangles_ is used
 * to pass in the length of the _triangles_ array (and hence the maximum number
 * of triangles that can be returned), and to pass back the actual number
 * of triangles in the patch.  If more triangles are found than will fit in the
 * _triangles_ array, the routine returns FALSE; otherwise it returns TRUE.
 * _cross_patch_bdy_ is a boolean value which, if TRUE, specifies that TIN 
 * adjacency information should be generated for triangle edges at the patch 
 * boundary.
 *
 * The fields of each element of the _triangles_ array are filled out as
 * follows:
 *
 *    verts:          The three vertices of the triangle in clockwise order.
 *
 *    soil:           The soil type of the triangle
 *
 *    adj_triangles:  This references the triangle adjacent at each edge 
 *                    (i.e., adj_triangles[0] has the adjacency for edge 0, 
 *                    etc.)  If the edge is not on a patch boundary, 
 *                    adj_triangles stores the index in the triangles list for 
 *                    this patch at which the adjacent triangle can be found.
 *                    Otherwise it stores the index in the triangles list 
 *                    (as would be returned on another call to 
 *                    ctdb_lookup_tin_triangles) for the patch on the other 
 *                    side of the boundary.  The special value 
 *                    CTDB_NO_ADJACENT_TIN is used when no valid adjacency 
 *                    can be found for an edge.  Note that if cross_patch_bdy 
 *                    was set to FALSE in the original call, all edges on the
 *                    patch boundary will have their adjacency set to 
 *                    CTDB_NO_ADJACENT_TIN.
 *
 *    adj_patches:    This stores the patch in which each adjacent triangle 
 *                    resides--thus, adj_patches[] will be set to the current 
 *                    patch number for all internal adjacencies, and to the 
 *                    number of the appropriate patch for other adjacencies.
 *                    Note that when adj_triangles[i] is set to 
 *                    CTDB_NO_ADJACENT_TIN, the value of adj_patches[i] is
 *                    *not* guaranteed to be a valid patch number.
 */
extern int32
ctdb_lookup_tin_triangles16 (
				CTDB                  *ctdb,
				int32                  patch_number,
				CTDB_TIN_TRIANGLE      triangles[],
				int32                 *num_triangles,
				int32                  cross_patch_bdy,
				CTDB_PATCH_GROUP      *pg,
				int32                  mem);

extern int32
ctdb_lookup_tin_triangles08 (
				CTDB                  *ctdb,
				int32                  patch_number,
				CTDB_TIN_TRIANGLE      triangles[],
				int32                 *num_triangles,
				int32                  cross_patch_bdy,
				CTDB_PATCH_GROUP      *pg,
				int32                  mem);


/* ctdb_lookup_tin_polygonsXX will populate the CTDB_TERRAIN_TRIANGLE data 
 * structure by unpacking each terrain element into one or two triangles.  The 
 * output from this routine is the number of triangles generated.
 *
 */
extern int32
ctdb_lookup_tin_polygons16 (
					 CTDB                  *ctdb,
					 int32                  patch_number,
					 TERRAIN_ELEMENT_16    *te,
					 int32                  index,
					 CTDB_TERRAIN_TRIANGLE *pt0,
					 CTDB_TERRAIN_TRIANGLE *pt1
					 );

extern int32
ctdb_lookup_tin_polygons08 (
					 CTDB                  *ctdb,
					 int32                  patch_number,
					 TERRAIN_ELEMENT_08    *te,
					 int32                  index,
					 CTDB_TERRAIN_TRIANGLE *pt0,
					 CTDB_TERRAIN_TRIANGLE *pt1
					 );


/* This function changes the soil type of a terrain element that contains
 * a specified point.  It has to search through the terrain elements.
 * The index and information about which polygon is changed is returned
 * so that the next time the same polygon is changed, we won't have to
 * do another search from scratch.  We can just use
 * ctdb_update_terrain_element_soilXX() instead.
 */
extern void
ctdb_change_terrain_element_soil08(CTDB *ctdb,
				   int32 patch_number,
				   CTDB_TS_POLYGON *poly,
				   int32 soil,
				   int32 *max_tes,
				   int32 which[],
				   uint32 te_index[]);

extern void
ctdb_change_terrain_element_soil16(CTDB *ctdb,
				   int32 patch_number,
				   CTDB_TS_POLYGON *poly,
				   int32 soil,
				   int32 *max_tes,
				   int32 which[],
				   uint32 te_index[]);

/* This function updates the soil type of the specified TE.  The TE is
 * completely specified by the arguments, so no search is required to
 * find the right one.  Call this only if you know the index and which
 * TE you're changing.
 */
extern void
ctdb_update_terrain_element_soil08(CTDB   *ctdb,
				   int32   patch_number,
				   uint32  index,
				   int32   which,
				   int32   soil);

extern void
ctdb_update_terrain_element_soil16(CTDB   *ctdb,
				   int32   patch_number,
				   uint32  index,
				   int32   which,
				   int32   soil);


/***********************************************************************
 * ct_print.c                                                          *
 ***********************************************************************/
typedef struct ctdb_collection
{
    int32 patch_number;
    int32 te_edges_crossed;/* number of edges inspected for intersection*/
    int32 te_edges_intersected;/* num edges intersected by ray */
    int32 tes_checked;/* num of terrain elements evaluated */
    int32 tes_intersected;/* of the tes evaluated how many had an edge that
			     was intersected by the ray */
    int32 vertices_crossed;/* num of vertices intersected */
    int32 missed_init_te;/* num of times the initial pt was not in the te
			    index returned by the te index routine - this
			    is a item precision issue */
    CTDB_STATUS status;
    CTDB_STATUS do_lookup;/* was elevation lookup required?*/
    float64 visibility;/* visibility after processing tes in patch */
} CTDB_COLLECTION;

typedef struct ctdb_post_stat
{
    int32 posts;
    int32 trees;
    int32 micro;
    int32 features;
    int32 buildings;
} CTDB_POST_STAT;

typedef enum
{
    CTDB_PRINT_LINE   = 0,
    CTDB_PRINT_TABLE  = 1,
    CTDB_PRINT_HEADER = 2,
    CTDB_PRINT_TOTAL  = 3
} CTDB_PRINT_MODE;

/* ctdb_print_te_data prints out information about all TIN terrain elements in
 * the specified patch.  Note that this function will only generate output
 * if mode is set to CTDB_PRINT_TABLE
 */
extern void 
ctdb_print_te_data (
		       CTDB           *ctdb,
		       int32           patch_number,
		       CTDB_PRINT_MODE mode
		       );

/* ctdb_print_te_map prints out the virtual grid mapping for a TINned patch. */
extern void 
ctdb_print_te_map (
		      CTDB   *ctdb,
		      int32   patch_number
		      );

/* ctdb_terrain_element_density returns the total number of TIN triangles
 * in a patch.  Note that this is different from the number of terrain elements
 * in a patch because TEs need not be triangles.  This is a wrapper for
 * ctdb_terrain_element_densityXX.
 */
extern int32 
ctdb_terrain_element_density (
				 CTDB *ctdb,
				 int32 patch_number
				 );

/* ctdb_print_intervis_metrics displays data about the most recent intervis
 * traversal (the data is stored in global ctdb_data).
 */
extern void
ctdb_print_intervis_metrics (
				CTDB            *ctdb,
				CTDB_PRINT_MODE  mode
				);
			     
/* ctdb_print_features displays information about the features in a patch. */
extern void 
ctdb_print_features (
			 CTDB              *ctdb,
			 int32              patch_x,
			 int32              patch_y
			 );

/* ctdb_print_posts prints posts in the database and counts the feature status
 */ 
extern void
ctdb_print_posts(CTDB           *ctdb,
		 int32           patch_x,
		 int32           patch_y,
		 CTDB_POST_STAT *stat
		 );

/* ctdb_print_stats prints the feature status generated by running
 * ctdb_print_posts on a ctdb database.
 */
extern void 
ctdb_print_stats(CTDB *ctdb);

/****************
 * ct_contour.c *
 ****************/
/* Max number of contour lines to allow on a single terrain edge.  This
 * is used to prevent near-infinite looping when the database contains
 * obviously bogus elevation values
 */
#define CTDB_MAX_CONTOURS_PER_EDGE 50


/* If there is a point along segment xy0=>xy1 with altitude z, return
 * TRUE and put its coordinates in (*xres, *yres), else return FALSE
 */
extern int32 
ctdb_find_z_on_line (
			CTDB_XYO_COORD2D       xy0,
			float64                z0,
			CTDB_XYO_COORD2D       xy1,
			float64                z1,
			int32                  px,
			int32                  py,
			float64                patch_incrD,
			float64                z,
			float64               *x_res,
			float64               *y_res
			);


/***************************
 * ct_search.c             *
 ***************************/


typedef struct ctdb_search_db
{
    /* Flag indicating whether we are looking at real patches
     * (x, y), or quad nodes (q).  We look at real patches
     * first, then quad nodes.
     */
    uint16 real_patches;
    uint16 x, y;
    int16 q;
    
    /* The last patch we fetched, and the cache_miss count after
     * we fetched it (if the feature_miss_count doesn't change, then
     * we can assume it is still in memory, and thus we won't
     * need to re-fetch it; if the miss_count has increased since
     * the last time we fetched a patch, then we would have to
     * re-fetch).
     */
    int32 last_number;
    CTDB_FEATURE_DATA *last_patch;
    int32 last_feature_miss_count;
    
    CTDB *ctdb;          /* GCS: current cell for search space */
    int32 current_index; /* GCS: current cell index in list */
    
    /* Bounds of the search in terms of the *real* patches (not quad
     * nodes) to be covered.  These are inclusive: min <= patch <= max.
     */
    int32 min_patch_x, max_patch_x;
    int32 min_patch_y, max_patch_y;
    
    /* Quad nodes which are also to be covered in the search */
    int32   num_quad_nodes;
    uint32 *quad_nodes;
    
} CTDB_SEARCH_DB;


/* ctdb_search_space:
 *
 * This is a private data structure used to track the progress of
 * application terrain searches.  The application first creates the
 * search space by specifying a bounding rectangle and some search
 * behavior control flags.  A representation of the search space
 * in terms of CTDB patches is determined from this information, and
 * various iteration variables are initialized.  The application
 * may search the space in any order desired, so many iteration variables
 * are needed to properly track which features have been returned
 * and which have not, without additional searching.
 */
struct ctdb_search_space
{
    /* GCS:  The Search space is defined in terms global search extents and a
     * list of cells (ctdbs) that make up the search space.  Each cell is
     * searched in turn and is maintained for the search routines in the ctdb
     * field.  When search is exhausted within the current cell, then the next
     * cell is assigned and this search space is reinitialized with respect to
     * this new current cell.
     */
     
    int32 min_x, min_y, max_x, max_y; /* search extents */
    int32 initial_frame;  /* cell_id in which the search extents are defined */
    int32 n_cells;       /* GCS: number of cells */
    int32 *cell_id_list; /* list of cells (ctdbs) for search space */
    CTDB_SEARCH_DB *cell_data_list; /* Initial iteration data for each
				     * cell
				     */


    /* Application-provided CTDB, used in subsequent calls.  Since
     * the internal representation of the search is dependent on the
     * particular CTDB being explored, it would be inappropriate to
     * require/allow the application to provide the CTDB in subsequent
     * calls.
     */
    CTDB *call_ctdb;

    /* Search control flags specified by application. */
    uint32 control_flags;

    /* Current patch being investigated by the application
     * (used if control_flags contains CTDB_BY_PATCH).
     */
    /* Since now all search capability is over multiple CTDB's then the CTDB-
       relevant information must be included with the current patch information
       data structure for each type of item that can be searched for 
       (i.e., by patch, tree, building, linear, abstract)
       */
    CTDB_SEARCH_DB search_patch;

    /* We group all the iteration variables into a structure for
     * easy initialization.
     */
    struct
    {
	/* Iteration variables associated with each feature */
	struct ctdb_search_state
	{
	    /* Patch being traversed (used if control_flags does not
	     * contain CTDB_BY_PATCH).
	     */
	    CTDB_SEARCH_DB search_patch;

	    /* Flag indicating we've finished everything */
	    int32 all_done;

	    /* The offset into the patch for the *next* feature to be
	     * returned.  (Note: If last_patch is NULL, then we have not
	     * looked at the patch yet, so this value cannot be valid
	     * in that case.)
	     */
	    int32 offset;

	    /* Keeps track of the last feature within an aggregation that
	     * has been returned.
	     */
	    int32 feature_number;

	} tree, building, linear, abstract[CTDB_ABSTRACT_MAX];

	/* When control_flags contains CTDB_TREELINE_TREES, we need to
	 * return the optional first & last trees separately from the
	 * tree line.  We use this variable to remember what we've
	 * returned so far on each tree.
	 */
	enum
	{
	    /* The tree-line, and the first and last tree */
	    CTDB_RETURNED_ALL,
	    /* Just the tree-line part */
	    CTDB_RETURNED_LINE,
	    /* Just the tree-line and the first tree */
	    CTDB_RETURNED_FIRST,
	    /* Done with the linears, now looking through the aggregates */
	    CTDB_PROCESSING_AGGREGATES
	} last_tree;

    } iteration;
};

/****************************
 * Functions from ct_poly.c *
 ****************************/

/* ctdb_point_in_feature tests whether the point xu,yu (in patch-relative
 * units, see ct_feat.h) is within the polygonal feature starting
 * at data.  If PT_ON_VERTEX is returned, result is set to the Z value
 * of that vertex, otherwise result is unchanged.  One of the following
 * values is returned:
 */
extern CTDB_POINT_STATE 
ctdb_point_in_micro_feature(CTDB_FEATURE_DATA data[],
			    int32             xu,
			    int32             yu,
			    int32             incr,/* size of feature in
						    * CTDB_FEATURE_DATA 
						    * words 
						    */
			    float64          *elev);


/* Same as above, for canopies */
extern CTDB_POINT_STATE 
ctdb_point_in_canopy_feature(CTDB_FEATURE_DATA data[],
			    int32             xu,
			    int32             yu,
			    int32             incr,/* size of feature in
						    * CTDB_FEATURE_DATA 
						    * words 
						    */
			    float64          *elev);

/* ctdb_point_in_area uses the crossings test to determine if a point
 * is contained within a general closed area, as defined by an array of
 * single precision vertices.  A test ray is assumed to eminate from
 * the test point in the positive X direction.  If this test ray crosses
 * the area an odd number of times, the point is inside.  Trivial
 * elimination of non-intersecting area line segments is done whenever
 * both area line segment endpoints are on the same side of the test
 * ray.
 * This algorithm works for concave areas including areas which are
 * self intersecting.
 */
extern int32
ctdb_point_in_area(float32 *verts,
                   uint32   num_verts,
                   float32  x, 
                   float32  y);

/* Given a 3-D plane defined by the three 3-D vertices of a feature, and (x,y)
 * of a point on the plane, this routine calculates the z of the point. 
 * The function returns 1 if z is defined and calculated, 0 if z doesn't exist.
 */
extern int32
ctdb_point_on_micro_plane(CTDB_FEATURE_DATA data[],
			  int32             xu,
			  int32             yu,
			  int32             incr,/*size of feature in
						   CTDB_FEATURE_DATA words */
			  float64          *result,
			  float64          *max);

/* Same as above for canopies */
extern int32
ctdb_point_on_canopy_plane(CTDB_FEATURE_DATA data[],
			   int32             xu,
			   int32             yu,
			   int32             incr,/*size of feature in
						    CTDB_FEATURE_DATA words */
			   float64          *result,
			   float64          *max);



/* ctdb_point_in_building works just like ctdb_point_in_feature,
 * except it is for buildings instead of microterrain/canopies.  Note
 * that this does assume the building is convex!
 */
extern CTDB_POINT_STATE 
ctdb_point_in_building (
			   CTDB_FEATURE_DATA data[],
			   int32             xu,
			   int32             yu,
			   float64          *elev
			   );

/* This function tests whether a specified point is inside or outside a
 * specified convex polygon.
 */
extern CTDB_POINT_STATE
ctdb_is_point_in_convex_polygon(
				float32               point[2],
				float32               polygon[][2],
				int32                 n_points
				);

/* ctdb_segment_intersection_t determines if the segment from s1 to e1
 * intersects the segment from s0 to e0 (defined in los structure).
 * If so, the parameters of each segment at the point of intersection 
 * is returned (see ct_poly.c for more details).
 *
 * ctdb_point_intersection_t determines if a segment perpendicular to
 * the segment from s0 to e0 with length radius centered at location
 * px,py intersects the sector starting with width 0 at s0 and widening
 * to 2*width at e0.  If so, the parameters of each segment at the
 * point of intersection is returned (see ct_poly.c for more details).
 *
 * ctdb_point_intersection does the same thing as ctdb_point_intersection_t
 * except width is ignored, and the parameters are not returned.
 *
 * ctdb_point_proximity is another variation which simply returns the
 * radius value, so the caller can make judgements about it.
 *
 * All units are patch-relative (see ct_feat.h).  One of the following
 * values is returned:
 */
extern CTDB_POINT_STATE 
ctdb_segment_intersection_t (
				CTDB_LINE_OF_SIGHT *los,
				CTDB_XYO_COORD2D    s1,
				CTDB_XYO_COORD2D    e1,
				float64            *t0_ret,
				float64            *t1_ret
				);


/* To a first order of approximation, ctdb_point_intersection_t determines
 * whether the segment from s0 to e0 intersects the circle with radius radius
 * and centered at (px,py).  As above, *t0_ret and *t1_ret return the distance
 * along the given rays at which intersection occurs.  This function also
 * takes into account the width of the target when determining whether an
 * an intersection occurs.  See the comments for this routine in ct_poly.c for 
 * more detail.
 */
extern CTDB_POINT_STATE 
ctdb_point_intersection_t (
			      float64  s0[2],
			      float64  e0[2],
			      float64  px, 
			      float64  py,
			      float64  radius,
			      float64  width,
			      float64 *t0_ret,
			      float64 *t1_ret
			      );


/* ctdb_point_intersection is a simpler version of ctdb_point_intersection_t
 * that ignore target width and does not return t values.
 */
extern CTDB_POINT_STATE 
ctdb_point_intersection (
			    CTDB_XYO_COORD2D s0,
			    CTDB_XYO_COORD2D e0,
			    CTDB_XYO_COORD2D pt,
			    int32            radius
			    );

/* ctdb_point_proximity returns in dist_sq the square of the distance from 
 * (ptx, pty) to the segment defined by s0 and e0
 */
extern void 
ctdb_point_proximity (
			 CTDB_XYO_COORD2D   s0,
			 CTDB_XYO_COORD2D   e0,
			 int32              ptx, 
			 int32              pty,
			 float64           *dist_sq
			 );

/* ctdb_compute_micro_feature_normal finds a normal to the feature
 * triangle starting at data (either data,vert0,vert1 or data,vert0,vert2
 * depending on whether which_t is 0 or 1, respectively).  The normal
 * in patch-relative coordinates (see ct_feat.h) is returned in
 * *pa,*pb,*pc, and *magic contains the dot-product of this normal
 * with one vertex of the triangle.
 */
extern void 
ctdb_compute_micro_feature_normal(CTDB_FEATURE_DATA data[],
				  int32             which_t,
				  int32             incr,/* size of feature in
							  * CTDB_FEATURE_DATA 
							  *  words 
							  */
				  float64          *pa, 
				  float64          *pb, 
				  float64          *pc,
				  float64          *magic);

/* Same as above, for canopies */
extern void 
ctdb_compute_canopy_feature_normal(CTDB_FEATURE_DATA data[],
				   int32             which_t,
				   /* size of feature in CTDB_FEATURE_DATA
				    * words */
				   int32             incr,
				   float64          *pa, 
				   float64          *pb, 
				   float64          *pc,
				   float64          *magic);

/* ctdb_width_obstruction computes the left and right boundaries of
 * the blockage caused by a perpendicular to line of sight.  width is
 * the width of the target SCALED BACK to the width at the point of
 * intersection.  radius is the radius of the obstruction.  tedge is
 * the parameter of the obstruction intersection (0 if the center of
 * the obstruction is on the line, 1 if it is exactly radius distance
 * away).  on_left is TRUE if the obstruction center is left of the
 * line of sight (use a cross product).  The left and right boundaries
 * in the range 0..1 are returned in lbp and rbp.
 */
extern void 
ctdb_width_obstruction(float64  width,
		       float64  radius,
		       float64  tedge,
		       int32    on_left,
		       float64 *lbp,
		       float64 *rbp
		       );

/* Sutherland-Hodgman polygon clipping algorithm.
 */
extern int32
ctdb_clip_poly(float32  input_poly[][XY],
	       int32    n_input_verts,
	       float32  output_poly[][XY],
	       float32  clipping_window[][XY],
	       int32    n_clipping_win_verts,
	       int32    max_output_verts,
	       int32    is_closed);

/* Returns TRUE iff the 2 line segments intersect.  If they do, returns
 * the fractional distance along the segs in t0 and t1, each of which
 * may be NULL if they are not needed.
 */
extern int32
ctdb_intersection_2_linesegs(float32        s0[2],
			     float32        e0[2],
			     float32        s1[2],
			     float32        e1[2],
			     float32        int_pt[2],
			     float32        *t0,
			     float32        *t1);

/*****************************
 * Functions from ct_micro.c *
 *****************************/

/* ctdb_norm_point_within_canopy tests whether the point at norm_x,norm_y
 * (normalized to grid post spacing) is within a canopy polygon.  If
 * so, the z is compared to the height of the polygon and two of the
 * following bitmasks are inclusive-or'd into *ray_info:
 */
#define UPPER_OVER  0x1
#define UPPER_UNDER 0x2
#define LOWER_OVER  0x4
#define LOWER_UNDER 0x8
extern int32
ctdb_norm_point_within_canopy ( 
			       CTDB         *ctdb,
			       CTDB_COORD3D *norm,
			       int32        *ray_info
			       );

/* ctdb_lookup_soil_feature checks if the passed point lies on top
 * of a linear feature (road or river).  If so, *soil is filled in
 * with the feature's soil type, and 1 is returned.  Otherwize the
 * routine returns 0.  px, py, norm_x, and norm_y are as in
 * ctdb_lookup_microterrain_elevation().  *width_ret, if non-NULL
 * returns the width of the linear feature which yielded the soil.
 * direction_ret, if non-NULL returns the direction of the linear
 * feature which yielded the soil.
 */
extern int32 
ctdb_lookup_soil_feature ( 
			  CTDB     *ctdb,
			  int32     px,
			  int32     py,
			  float64   norm_x,
			  float64   norm_y,
			  int32    *soil_ret,
			  float64  *width_ret,
			  float64  *direction_ret
			  );

/* ctdb_lookup_microterrain_elevation determines the elevation of the
 * microterrain at point norm_x,norm_y (normalized to grid post
 * spacing), and returns it in *elevation.  If non-null, *soil will be
 * filled with the soil type of the microterrain polygon.
 * Reference_elevation gives the maximum elevation acceptable for
 * multi-level microterrain (see comments at beginning of function for
 * an explanation).  px,py (POST_X(norm_x) and POST_Y(norm_y),
 * respectively) are passed in because they are generally already
 * known to the caller.  A non-zero return values indicates success.
 * A zero return value indicates that microterrain was not found at
 * this point.  It is assumed that the caller has already checked the
 * post_mask for the post at px,py to determine that a call to this
 * routine is appropriate.
 */
extern int32 
ctdb_lookup_microterrain_elevation ( 
				       CTDB    *ctdb,
				       int32    px,
				       int32    py,
				       float64  norm_x,
				       float64  norm_y,
				       float64 *elevation,
				       int32   *soil,
				       float64  reference_elevation,
				       int32   *is_ml_micro
				       );

/* ctdb_lookup_feature_elevation determines the elevation of the
 * features (canopies and buildings) which are selected in the
 * post_flags mask.
 */
extern int32 
ctdb_lookup_feature_elevation ( 
				  CTDB     *ctdb,
				  int32     px,
				  int32     py,
				  float64   norm_x,
				  float64   norm_y,
				  int32     post_flags,
				  float64  *elev_ret
				  );

/* ctdb_check_clear determines if the area within the radius of the
 * passed center point is "clear" (there are none of the features selected
 * in post_flags in the area).  The passed post coordinates bound the
 * search.
 */
extern int32 
ctdb_check_clear ( 
		     CTDB     *ctdb,
		     int32     pxl,
		     int32     pyl,
		     int32     pxh,
		     int32     pyh,
		     float64   norm_x,
		     float64   norm_y,
		     float64   norm_r,
		     int32     post_flags
		     );

/* find_nearest_road_in_patch finds any road-soil linear feature in the
 * patch which is closer to the passed point (in post coordinates) than
 * the current value of *dist.  If one is found the distance is updated,
 * the edge_index is filled in, and 1 is returned.  If not, 0 is returned.
 * If the point actually fall within the bounds of a road segment, a
 * distance of 0.0 is returned, otherwise the closest distance between
 * the point and the centerline of the feature is returned.
 */
extern int32 
ctdb_find_nearest_road_in_patch ( 
				    CTDB     *ctdb,
				    float64   norm_x,
				    float64   norm_y,
				    int32     patchx,
				    int32     patchy,
				    float64  *dist,
				    int32    *edge_index,
				    int32    *edge_sub_index
				    );


/* ctdb_compute_rotation finds the rotation matrix for a body sitting
 * at the passed location (normalize to grid post spacing), facing the
 * passed direction (dx, dy).  <dx, dy> is assumed to be a unit vector.
 * The reference_elevation is used for multi-level microterrain
 * (see ctdb_lookup_microterrain_elevation).
 */
extern void 
ctdb_compute_rotation ( 
			  CTDB   *ctdb,
			  float64 norm_x,
			  float64 norm_y,
			  float64 norm_hl,
			  float64 norm_hw,
			  float64 dx,
			  float64 dy,
			  float64 rotation[3][3],
			  float64 reference_elevation
			  );

/* ctdb_compute_mes_rotation finds the rotation matrix for a vehicle
 * sitting on a MES triangle with the passed unit_normal vector, 
 * facing the passed direction (dx, dy).  <dx, dy> is assumed to be a unit 
 * vector.
 */
extern void 
ctdb_compute_mes_rotation(
			  float64 dx,
			  float64 dy,
			  float64 unit_normal[XYZ],
			  float64 rotation[3][3]
			  );


/****************************
 * Functions from ct_clip.c *
 ****************************/

/* ctdb_clip_intersections finds the intersections of the line
 * connecting v0 to v1 with the 2D bounding box specified.  The
 * number of intersections * is returned.
 * The dimensions of the vectors are passed in dim (>= 2).
 */
extern int32 
ctdb_clip_intersections ( 
			    float64 v0[],
			    float64 v1[],
			    float64 min_x,
			    float64 max_x,
			    float64 min_y,
			    float64 max_y,
			    float64 clipped[],
			    int32   dim
			    );

/* ctdb_clip_line clips the line connecting v0 to v1 to be within the
 * 2D bounding box specified.  The clipped line is returned in clipped[],
 * the number of points in the clipped line (either 0 or 2) is returned.
 * The dimensions of the vectors are passed in dim (>= 2).
 */
extern int32 
ctdb_clip_line ( 
		   float64 v0[],
		   float64 v1[],
		   float64 min_x,
		   float64 max_x,
		   float64 min_y,
		   float64 max_y,
		   float64 clipped[],
		   int32   dim
		   );

/* ctdb_clip_polygon clips a polygon to patch boundaries.  
 * This function should be named better and return something more clear than
 * a bunch of data pointed to by a int32 *  (editorial opinion)
 */
extern int32 *
ctdb_clip_polygon(CTDB *ctdb, int32 n_verts, float64 *verts);


/******************************
 * Functions from ct_raster.c *
 ******************************/

/* ctdb_rasterize_box ORs a sequence of bits into the passed raster
 * array.  left, right, bottom, and top are all in the range 0..1,
 * left <= right, bottom <= top.
 */
extern void 
ctdb_rasterize_box ( 
		       float64 left,
		       float64 right,
		       float64 bottom,
		       float64 top,
		       uint32  raster[32]
		       );

/* ctdb_rasterize_polygon ORs bits matching the shape of a polygon
 * into the passed raster array. It requires the (ordered) list of
 * vertices of the polygon, the number of vertices, as well as the
 * indicies of the points with max and min X and Z values.
 * Note that all calculations are done in the X-Z plane, ignoring
 * Y components of all vectors.
 */
extern void 
ctdb_rasterize_polygon ( 
			   int32    min_x_index,
			   int32    max_x_index,
			   int32    min_z_index,
			   int32    max_z_index,
			   float64 *vertices[],
			   int32    num_vertices,
			   float64  interior_point[3],
			   uint32   raster[32]
			   );

/* ctdb_raster_density counts the number of ON bits in the passed raster
 * array, and returns a number in the range 0..1, indicating the portion
 * which are on.
 */
extern float64 
ctdb_raster_density ( 
			uint32 raster[32]
			);

/* ctdb_raster_density_and_size counts the number of ON bits in the passed 
 * raster array, and returns a number in the range 0..1, indicating the 
 * portion  which are on. In addition, it returns the indices of the 
 * left-most, right-most, bottom and top bits which are on.
 */
extern float64 
ctdb_raster_density_and_size ( 
				 uint32 raster[32],
				 int32 *left,
				 int32 *right,
				 int32 *bottom,
				 int32 *top
				 );

/* ctdb_raster_diff performs the bitwise operation c = a & ~b
 * on the passed rasters.
 */
extern void 
ctdb_raster_diff ( 
		     CTDB_RASTER a, 
		     CTDB_RASTER b, 
		     CTDB_RASTER c 
		     );

/***************************
 * Functions from ct_veh.c *
 ***************************/

/* ctdb_vehicle_raster works just like ctdb_vehicle_blockage, except
 * no clipping is done, and the caller provides the raster.  We use
 * this function from ctdb_point_to_point in order to combine the
 * effects of vehicles with those of the terrain.
 */
extern float64 
ctdb_vehicle_raster ( 
			CTDB                 *ctdb,
			float64               x0, 
			float64               y0, 
			float64               z0,
			float64               x1, 
			float64               y1, 
			float64               zl,
			float64               zh,
			float64               width,
			int32                 n_veh,
			CTDB_VEHICLE_LOCATION veh[],
			int32                 ignore0, 
			int32                 ignore1,
			CTDB_RASTER           raster
			);

/******************************
 * Functions from ct_appear.c *
 ******************************/

/* ctdb_planar_hull returns the convex hull of the polygon of max_vertices
 * described by the list vertices. These points are flagged as visible or
 * not. For now, ctdb_planar_hull is a fake, and assumes that all the
 * visible points are actually vertices of the polygon, not interior
 * points, and simply returns the points in clockwise order. The list hull
 * contains these ordered points the the number of points is the return
 * value of the function. Also needed as argument is a point known to
 * be in the interior.
 */
extern int32 
ctdb_planar_hull ( 
		     float64  vertices[][3],
		     int32    max_vertices,
		     int32    visible[],
		     float64  interior_point[3],
		     float64 *hull[]
		     );

/****************************
 * Functions from ct_quad.c *
 ****************************/

/* ctdb_read_quad_tree:
 *
 * Read the abstract data (the quadtree and contents) for initialization.
 */
extern CTDB_STATUS 
ctdb_read_quad_tree (
			 CTDB *ctdb,
			 FILE *fileptr
			 );
/* ctdb_find_search_quads:
 *
 * Construct a list of quad node starting addresses within the search
 * space which provides coverage of the entire area bounded by the
 * passed parameters.
 */
extern void 
ctdb_find_search_quads ( 
			   CTDB_SEARCH_DB            *state,
			   int32                      min_x, 
			   int32                      min_y, 
			   int32                      max_x, 
			   int32                      max_y
			   );

/****************************
 * Functions from ct_topo.c *
 ****************************/

/* ctdb_read_topology:
 *
 * Read the topology (NODEs and EDGEs) for initialization.
 */
extern CTDB_STATUS 
ctdb_read_topology (
			CTDB *ctdb,
			FILE *fileptr
			);

/* ctdb_find_edge:
 *
 * Finds the index of the topologic edge which includes the
 * linear feature which starts at the passed offset within the specified
 * patch.  The edge_key provides a key for the search by specifying the
 * low-order 12 bits of the index of the edge (in most cases, this
 * key will be unique, since there are rarely > 4095 edges).  The return
 * values indicates success (1) or failure (0).  NOTE: Failure probably
 * indicates a corrupted database!
 */
extern int32 
ctdb_find_edge ( 
		   CTDB  *ctdb,
		   uint32 edge_key,
		   int32  patch_number,
		   int32  offset,
		   int32 *edge_index,
		   int32 *edge_sub_index
		   );

#define CTDB_DPB_OK           0
#define CTDB_DPB_FAILED      -1
#define CTDB_DPB_NO_PATCH    -1
#define CTDB_DPB_BUFFER_FULL -2

#define CTDB_DPB_HAS_PATCH(ii)  (ii >= CTDB_DPB_OK)
#define CTDB_DPB_PATCH_SIZE(p) ((p)->hdr.size * sizeof(CTDB_FEATURE_DATA))

extern int32 
ctdb_dpb_init(
	      int32 max_patches
	      );

extern int32 
ctdb_dpb_has_patch(
		   int32 patch_num
		   );

extern int32 
ctdb_dpb_get_patch(
		   int32 patch_num,
		   CTDB_FEATURE_DATA **patch
		   );

extern int32 
ctdb_dpb_dup_patch(
		   int32 patch_num,
		   CTDB_FEATURE_DATA **patch
		   );

extern int32 
ctdb_dpb_put_patch(
		   CTDB *ctdb,
		   int32 patch_num,
		   CTDB_FEATURE_DATA *patch
		   );

extern CTDB_FEATURE_DATA*
ctdb_dpb_get_feature(
		     CTDB *ctdb,
		     int32 patch_number
		     );

extern int32
ctdb_dpb_change_terrain_element(CTDB             *ctdb,
				CTDB_TS_POLYGON  *poly,
				float32           centroid[XY],
				int32             soil);

extern void
ctdb_apply_te_deltas(CTDB *ctdb,
		     int32 patch_number);

/*** prototype for functions moved into ct_encode.c ***/

extern int32 
ctdb_compute_line_post_mask(CTDB *ctdb, 
			    CTDB_XYO_COORD2D xy0,
			    CTDB_XYO_COORD2D xy1);

extern int32 
ctdb_compute_point_post_mask(CTDB *ctdb,
			     CTDB_XYO_COORD2D xy,
			     float64 radius);

extern int32 
ctdb_compute_triangle_post_mask(CTDB *ctdb,
				CTDB_XYO_COORD2D xy0,
				CTDB_XYO_COORD2D xy1,
				CTDB_XYO_COORD2D xy2);

extern int32 
ctdb_fill_post_gaps(int32 mask, 
		    int32 width);


#define CTDB_MLIB_VALID_VOLUME_INDEX(ctdb,ii)  \
    (((ii>=0) && (ii<(ctdb)->num_volume_models)) ? 1 : 0) 

#define CTDB_MLIB_VALID_MES_VOLUME_INDEX(ctdb,ii) \
    (((ii>=0) && (ii<(ctdb)->num_mes_volumes)) ? 1 : 0)

extern int32 
ctdb_mlib_model_index(
    CTDB *ctdb,           /*v pointer to the database */
    char *model_name,     /*v network identifier for the model */
    float32 target_state, /*v target state (between 0.0 and 1.0) */
    int32 *is_mes
		       );

#define CTDB_DESC_TO_MODEL 0
#define CTDB_MODEL_TO_DESC 1

extern int32 
ctdb_convert_linear_desc_to_model(
    CTDB *ctdb,
    int32 desc_to_model,
    int32 type,
    CTDB_LINEAR_DESC *desc_p,
    CTDB_LINEAR_MODEL *model_p
		  );
extern int32 
ctdb_create_linear_model(
    CTDB *ctdb,               
    int32 type,
    CTDB_LINEAR_DESC *desc_p  
			 );

#define CTDB_MODEL_NOT_FOUND           -1
extern int32 
ctdb_find_matching_linear_model(
    CTDB *ctdb, 
    CTDB_LINEAR_MODEL *model, 
    int32 is_canopy
			    );

/* ctdb_lookup_elevation_amsl finds all elevation at the specified point on
 * the terrain database.  The point is tested to make sure it is not off 
 * the terrain database (it is not necessary for 
 * the application to make this check).  
 */
extern float64
ctdb_lookup_elevation_amsl(
			   CTDB *ctdb, 
			   float64 x, 
			   float64 y
			   );


/* ctdb_lookup_norm_elevation_amsl:
 *
 * This function computes the elevation above mean sea level.
 * First, it computes the elevation z at the normalized point
 * (xn, yn) and then calculates the distance from that point to
 * the ellipsoid.
 */
extern float64 
ctdb_lookup_norm_elevation_amsl (
				 CTDB    *ctdb,
				 float64  xn,
				 float64  yn,
				 CTDB_ELEV_QUALIFIER qual
				 );

/* ctdb_get_norm_elevation_amsl
 *
 * This function computes the elevation above mean sea level
 * for the point (xn, yn, z), where xn and yn are normalized
 * coordinates.  This elevation is returned, in meters.
 */
extern float64
ctdb_get_norm_elevation_amsl (
			    CTDB   *ctdb,
			    float64 xn,
			    float64 yn,
			    float64 z
			    );

/*******************************
 * Functions from ct_dqbuff.c  *
 *******************************/

/* Given an offset into the quad data buffer, this function
 * returns a pointer to the current quad data for that node.
 * It returns dynamic quad data if it exists.
 */
extern CTDB_FEATURE_DATA *
ctdb_get_quad_node (
                   CTDB *ctdb,
                   int32 quad_offset
                   );

extern void 
ctdb_add_dynamic_quad_buffer(
                            CTDB *ctdb, 
                            int32 quad_offset, 
                            CTDB_FEATURE_DATA *new_quad
                            );

/* ctdb_lookup_norm_elevation_max
 *
 * This function computes the maximum elevation of all items
 * defined by "qual".
 *
 * This routine has no DEFAULT, VALID or MUST qualifier defined since 
 * it is assumed that libctdb coders will define these appropriately 
 * prior to calling.
 *
 * DEFAULT is the qualifier that is set for the call if they specify
 * CTDB_USE_DEFAULT as their input qualifier.
 *
 * VALID is the qualifier that are considered valid for this API routine
 *
 * MUST  is the qualifier that is automatically included in the input 
 *       qualifier and defines the absolute minimum case for processing 
 *       in this API.
 */
extern CTDB_ELEV_DATA *
ctdb_lookup_norm_elevation_max(
			       CTDB               *ctdb, 
			       float64             norm_x, 
			       float64             norm_y,
			       CTDB_ELEV_QUALIFIER qual
			       );

/* ctdb_lookup_water_ptr will return the water chars pointer associated with 
 * the point (x,y). 
 */
extern CTDB_WATER_CHARS *
ctdb_lookup_water_ptr(CTDB *ctdb, 
		      float64 x, 
		      float64 y);
 
/* ctdb_lookup_patch_water_level will return the patch water level in addition
 * validating the patch number and ensuring that the patch data is in memory.
 * If the level is invalid (i.e. bad patch number, no patch group pointer or
 * the patch doesn't contain patch water) the ouput value is FALSE otherwise
 * the value is TRUE.
 */
extern int32
ctdb_lookup_patch_water_level(CTDB    *ctdb, 
			      int32    patch_number,
			      float64  norm_x,
			      float64  norm_y,
			      float32 *level);

extern int32
ctdb_is_areal(int32 abstract_feature_type);


/*************************************************
*                                                *
*   Definitions for Multi-Elevation Surfaces     *
*                                                *
*************************************************/

#define CTDB_MES_ALL_ENCLOSURES  -1

/* ===========INTERNAL FUNCTIONS for Multi-elevation Surfaces====== */

/* These functions are probably going to either be encapsulated in the
 * ctdb functions, or else they will be local variables that will be
 * called from those ctdb functions when the intervis ray enters a
 * MES.
 */

extern int32
ctdb_mes_get_triangle_normal(CTDB   *ctdb,
			     uint16  mes_id,
			     int32   triangle_id,
			     float64 normal[XYZ]
			     );
extern void
ctdb_calculate_triangle_normals ( CTDB_MES_VOLUME *mes );


/* ctdb_mes_line_intersection
 *
 * Traverses ray from ray_origin to ray_destination, defined in MES coordinates
 * 
 * all_intercepts is a flag (TRUE/FALSE).  Set this to true if you are 
 *     interested in all intersections the ray has with the MES.
 *
 * enclosure_id: if set, function will return all intersections in this
 *     enclosure.  Otherwise, set to CTDB_MES_ALL_ENCLOSURES.
 *
 * apertures_intercepts: flag to indicate treating intersections with 
 *     apertures as an intersection.
 *
 */
typedef struct ctdb_mes_intersection
{
    int32   triangle;/* triangle id of all intersected triangles */
    float64 distance;/* distance of all intersected triangles in normalized 
		      * coordinates (0.0 at origin, 1.0 at destination). 
		      */
} CTDB_MES_INTERSECTION;

extern void
ctdb_mes_line_intersection (CTDB_MES_VOLUME        *mes,
			    float64                 ray_origin[XYZ],
			    float64                 ray_destination[XYZ],
			    int32                   all_intercepts,
			    int32                   enclosure_id,
			    int32                   check_mobility,
			    int32                   check_visibility,
			    int32                   aperture_intercepts,
			    int32                  *n_intercepts,
			    CTDB_MES_INTERSECTION **intercepts);

extern void
ctdb_generate_boxes ( CTDB_MES_TEMPLATE  *template,
                      int32               maximum_triangles_per_box );


/* ctdb_find_close_ext_ap returns the closest exterior aperture in mes
 * mes_index to the point pt whose climb and fall heights do not violate
 * climb_threshold and fall_threshold.
 */
extern CTDB_MES_EDGE_KEY
ctdb_find_close_ext_ap(CTDB            *ctdb,
		       int32            mes_index,
		       float64          pt[XYZ],
		       float64          climb_threshold,
		       float64          fall_threshold);

/* NOTE: The following macros take a mes-template as their first operand.
 * If you have a mes-volume instead, use <macro> (mes->template, ....).
 */

#define CTDB_MES_GET_ENCLOSURE_TRIANGLE_ID(templ, encl_id, triangle_number) \
  (templ->all_enclosures)[encl_id].first_wall_triangle_id + triangle_number

#define CTDB_MES_GET_APERTURE_TRIANGLE_ID(templ, aperture_id, triangle_number)\
  (templ->all_apertures)[aperture_id].first_triangle_id + triangle_number

#define CTDB_MES_GET_VERTEX_ID(templ, tri_id, vtx_number)                   \
  templ->all_triangles[tri_id].vertex_ids[vtx_number]

#define CTDB_MES_GET_VERTEX(templ, tri_id, vtx_number)                      \
(templ->all_vertices)[(templ->all_triangles[tri_id].vertex_ids[vtx_number])]

#define CTDB_MES_GET_ENCLOSURE_TRIANGLE(templ, encl_id, triangle_number)    \
(templ->all_triangles)                                                      \
[(templ->all_enclosures)[encl_id].first_wall_triangle_id + triangle_number]

#define CTDB_MES_GET_APERTURE_TRIANGLE(templ, aperture_id, triangle_number) \
(templ->all_triangles)                                                      \
[(templ->all_apertures)[aperture_id].first_triangle_id + triangle_number]


/* ============= Multi-Elevation Coordinate Conversion ========== */

#define CTDB_MES_ROTATE_WORLD_TO_MES(mes, vector)                     \
    vmat3_mat_vec_mul64((mes)->tfm_world_to_mes, (vector), (vector));   


#define CTDB_MES_ROTATE_MES_TO_WORLD(mes, vector)                     \
{                                                                     \
    float64            inv_matrix[XYZ][XYZ];                          \
    vmat3_inverse64((mes)->tfm_world_to_mes, inv_matrix);             \
    vmat3_mat_vec_mul64(inv_matrix, (vector), (vector));              \
}

#define CTDB_MES_OFFSET_WORLD_TO_MES(mes, vector)                     \
    VMAT3_VEC_ADD(-(mes)->origin, (vector), (vector))


#define CTDB_MES_OFFSET_MES_TO_WORLD(mes, vector)                     \
    VMAT3_VEC_ADD((mes)->origin, (vector), (vector))


#define CTDB_MES_SCALE_WORLD_TO_MES(mes, vector)                      \
    vmat3_scal_vec_mul64(1/(mes)->template->fixed_point_basis, (vector)); 


#define CTDB_MES_SCALE_MES_TO_WORLD(mes, vector)                      \
    vmat3_scal_vec_mul64((mes)->template->fixed_point_basis, (vector));


/* ctdb_gcs_to_mes
 *
 * This function will take a point in GCS (cell, x0, y0, z0)  
 * and convert it to the local coordinate system defined by 
 * the mes (x1, y1, z1).
 * 
 * If the gcs point lies outside the given mes, return 0.
 * Otherwise, we return 1.
 */
extern int32
ctdb_gcs_to_mes(
		CTDB_MES_VOLUME  *mes,
		int32             cell,
		float64           x0,
		float64           y0,
		float64           z0,
		float64          *x1,
		float64          *y1,
		float64          *z1
		);

/* ctdb_mes_to_gcs
 *
 * This function will take a defined by the mes's local coordinate 
 * system (x0, y0, z0), and converts it to a GCS point (cell, x1, y1, z1),
 * where cell is the cell that the MES origin is in.
 */
extern int32
ctdb_mes_to_gcs(
		CTDB_MES_VOLUME  *mes,
		int32             x0,
		int32             y0,
		int32             z0,
		int32            *cell,
		float64          *x1,
		float64          *y1,
		float64          *z1
		);

/* ctdb_mes_to_gcs
 *
 * This function will take a defined by the mes's local coordinate 
 * system (x0, y0, z0), and converts it to a GCS point (cell, x1, y1, z1).
 */
extern int32
ctdb_mes_to_gcscs(
		  CTDB_MES_VOLUME  *mes,
		  int32             x0,
		  int32             y0,
		  int32             z0,
		  int32             cell,
		  float64          *x1,
		  float64          *y1,
		  float64          *z1
		  );

extern int32
ctdb_xyo_to_mes(
		CTDB             *ctdb, 
		CTDB_MES_VOLUME  *mes,
		int32             patch_x, 
		int32             patch_y,
		float64           x, 
		float64           y, 
		float64           z,
		float64          *mes_x,
		float64          *mes_y,
		float64          *mes_z
		);

extern int32
ctdb_mes_to_xyo ( 
		 CTDB            *ctdb, 
		 CTDB_MES_VOLUME *mes,
		 float64          mes_x, 
		 float64          mes_y, 
		 float64          mes_z, 
		 int32            patch_x, 
		 int32            patch_y,
		 float64         *x, 
		 float64         *y, 
		 float64         *z 
		 );

/* ctdb_gcsvec_to_mesvec
 *
 * This function will take a vector defined in GCS, and convert it 
 * to the local coordinate system defined by the mes.
 */
extern int32
ctdb_gcsvec_to_mesvec(CTDB_MES_VOLUME    *mes,
		      int32               cell,
		      float64             gcsvec[XYZ],
		      float64             mesvec[XYZ]
		      );

/* ctdb_mesvec_to_gcsvec
 *
 * This function will take a vector defined by the mes's local coordinate
 * system and transforms it to vector in a specific GCS cell.
 */
extern void 
ctdb_mesvec_to_gcscsvec(
			CTDB_MES_VOLUME *mes,
			float64          mesvec[XYZ],
			int32            cell,
			float64          gcsvec[XYZ]
			);

extern float64
ctdb_find_z_from_gcs_patch_water(CTDB   *ctdb, 
				 float64 x,
				 float64 y);

#endif /*_LIBCTDB_LOCAL_H_INCLUDED*/
