#include <stdio.h>
#include <math.h>
/****************************************************************************
 *   File: compare_elev.c                                                    *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *                                                                           *
 *       Contents: Program to print the database header                      *
 *       Created:  1995                                                      *
 *       Author:   cbuettner flchamberlain wkbraudaway                       *
 *       Remarks:  This routine is used to compare the output from two       *
 *       different ctdb elevation sampling runs with the driver util         *
 *       ( query points (generated by driver) must be identical for the      *
 *       output of this routine to have any meaning)                         *
 *                                                                           *
 ****************************************************************************/
/* Min difference in elevations to flag */
#define THRESHOLD .01
#define FNAME_LEN 50
#define MAX_LINE_LEN 200


main(argc, argv)
    int argc;
    char **argv;
{
    char f1_name[FNAME_LEN], f2_name[FNAME_LEN];
    FILE *f1, *f2;    
    char line1[MAX_LINE_LEN], line2[MAX_LINE_LEN];    
    double x1, y1, e1, x2, y2, e2;
    int p1, p2,v0,v1;
    int i=0, ok_count=0;
    int ignoring_points,n_points=0;
    double max_x, max_y, limit_x, limit_y, increment;
   
    if (argc == 3)
    {
	strcpy(f1_name, argv[1]);
	strcpy(f2_name, argv[2]);	
	ignoring_points = 0;
    }
    else
    {
	if (argc == 6 )
	{
	    printf( "\n %s %s %s", argv[3], argv[4], argv[5] );
	    
	    strcpy(f1_name, argv[1]);
	    strcpy(f2_name, argv[2]);	
	    max_x     = atof(argv[3]);
	    max_y     = atof(argv[4]);
	    increment = atof(argv[5]);
	    limit_x = max_x - increment;
	    limit_y = max_y - increment;
	    ignoring_points = 1 ;
	}
	else
	{
	    printf("Usage:\n");
	    printf("%s <file1> <file2> ",argv[0]);
	    printf("[<max_x> <max_y> <post_increment>]\n");
	    printf("where file 1 & 2 are outputs from the driver utility\n");
	    printf("labeled <dbname>.elev.samples.<ctb format>\n");
	    printf("and (optional) <max_x> <max_y> <post_increment> are\n");
	    printf("the maximum x and y in the database, ");
	    printf("and the increment from post to post.\n");
	    exit(1);
	}
    }

    /* Assumes each file is of the format
     * x y elev status
     * and that they are line-to-line correlated.
     */
    
    if ((f1 = fopen(f1_name,"r")) == (FILE *)NULL)
    {
	printf("Error opening file %s\n",f1_name);
	exit(1);
    }
    
    if ((f2 = fopen(f2_name,"r")) == (FILE *)NULL)
    {
	printf("Error opening file %s\n",f2_name);
	exit(1);
    }
    
    printf("Output columns:\n");
    printf("1.  Elevation difference\n");
    printf("2.  %s Elevation\n",f1_name);
    printf("3.  %s Elevation\n",f2_name);
    printf("4.  X\n");
    printf("5.  Y\n");
    printf("6.  Patch number\n");

    printf("Working");
    while((!feof(f1)) && (!feof(f2)))
    {
	
	if (!((i++)%100))
        {
	    printf(".");
	    fflush(stdout);
	}

	fgets(line1, MAX_LINE_LEN, f1);
	fgets(line2, MAX_LINE_LEN, f2);

	v0 = sscanf(line1,"%lf %lf %lf %d",&x1,&y1,&e1,&p1);
	v1 = sscanf(line2,"%lf %lf %lf %d",&x2,&y2,&e2,&p2);

	if ( v0 != 4 || v1 != 4 )
	  continue;
	
	if ( fabs( e1-e2 ) > THRESHOLD )
	  printf("\n%9.3f\t%9.3f\t%9.3f\t%9.3f\t%9.3f\t%9d", 
		 fabs(e1-e2),
		 e1,
		 e2,
		 x1,
		 y1,
		 p1);

	else
	  ok_count++;

	if ((x1 != x2) || (y1 != y2))
	  printf("\nWARNING:  x and y values do not correlate between files!\n");
	else
	  n_points++;	
    }
    printf("\n\nTotal samples = %d\n",n_points);
    printf("Total correct results = %d\n",ok_count);
    printf("Success rate = %d\n",(100*ok_count)/n_points);
}



