#ifndef lint
static char rcsid [] = "$RCSfile: compare_intervis.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: compare_intervis.c                                                *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *                                                                           *
 *       Contents: Compare driver intervis datafiles                         *
 *       Created:  1995                                                      *
 *       Author:   cbuettner                                                 *
 *       Remarks:  This routine is used to compare the output from two       *
 *       different ctdb intervis runs with the driver utility (the same      *
 *       intervis rays should be used for this output to have any meaning).  *
 *                                                                           *
 ****************************************************************************/
#include <stdio.h>
#include <math.h>
/* Min difference in intervisibility to flag */
#define THRESHOLD .01
#define FNAME_LEN 100
#define MAX_LINE_LEN 300

main(argc, argv) 
    int argc;
    char **argv;
{
    char f1_name[FNAME_LEN], f2_name[FNAME_LEN];
    FILE *f1, *f2;    
    char line1[MAX_LINE_LEN], line2[MAX_LINE_LEN];    
    float x11, x12, y11, y12, z11, z12;
    float x21, x22, y21, y22, z21, z22;
    float v1, v2;
    int s1, s2;
    int total = 0;
    int good = 0;
     int ignoring_points;
    double max_x, max_y, limit_x, limit_y, increment;
   
    if (argc == 3)
    {
	strcpy(f1_name, argv[1]);
	strcpy(f2_name, argv[2]);	
	ignoring_points = 0;
    }
    else
    {
	if (argc == 6 )
	{
	    printf( "\n %s %s %s", argv[3], argv[4], argv[5] );
	    
	    strcpy(f1_name, argv[1]);
	    strcpy(f2_name, argv[2]);	
	    max_x     = atof(argv[3]);
	    max_y     = atof(argv[4]);
	    increment = atof(argv[5]);
	    limit_x = max_x - increment;
	    limit_y = max_y - increment;
	    ignoring_points = 1 ;
	}
	else
	{
	    printf("Usage:\n");
	    printf("%s <file1> <file2> ",argv[0]);
	    printf("[<max_x> <max_y> <post_increment>]\n");
	    printf("where file 1 & 2 are outputs from the driver utility\n");
	    printf("labeled <dbname>.intervis.samples.<ctb format>\n");
	    printf("and (optional) <max_x> <max_y> <post_increment> are\n");
	    printf("the maximum x and y in the database, ");
	    printf("and the increment from post to post.\n");
	    exit(1);
	}
    }
  
    /* Assumes each file is of the format
     * x y elev status
     * and that they are line-to-line correlated.
     */
    
    if ((f1 = fopen(f1_name,"r")) == (FILE *)NULL)
    {
	printf("Error opening file %s\n",f1_name);
	exit(1);
    }
    
    if ((f2 = fopen(f2_name,"r")) == (FILE *)NULL)
    {
	printf("Error opening file %s\n",f2_name);
	exit(1);
    }
    
    printf("Output format:\n");
    printf("DelVis\tV.ctb\tV.ict\teye.x\t\teye.y\t\teye.z\t\ttarg.x\t\ttarg.y\t\ttarg.z\t\tStatus\n");

    while((!feof(f1)) && (!feof(f2)))
    {
	fgets(line1, MAX_LINE_LEN, f1);
	fgets(line2, MAX_LINE_LEN, f2);

	if ( sscanf(line1,"%f\t%d\t%f\t%f\t%f\t%f\t%f\t%f",
		    &v1,&s1,&x11,&y11,&z11,&x12,&y12,&z12) == 8 )
	{
	    if ( sscanf(line2,"%f\t%d\t%f\t%f\t%f\t%f\t%f\t%f",
			&v2,&s2,&x21,&y21,&z21,&x22,&y22,&z22)==8)
		total++;
	    
	    if ( x11 < limit_x &&  x12 < limit_x &&
		 y11 < limit_y &&  y12 < limit_y && 
		 fabs( v1-v2 ) > THRESHOLD )
	      printf("%5.3f\t%5.3f\t%5.3f\t%9.3f\t%9.3f\t%9.3f\t%9.3f\t%9.3f\t%9.3f\t%d\n",
		     fabs(v1-v2),
		     v1,
		     v2,
		     x11,
		     y11,
		     z11,
		     x12,
		     y12,
		     z12,
		     s2);
	    else
	      good++;
	    
	    if ((x11 != x21) || (y11 != y21) || (z11 != z21) ||
		(x12 != x22) || (y12 != y22) || (z12 != z22))
	      printf("WARNING:  points do not correspond between files!\n");
	}
    }

    fprintf(stderr,"Total tests: %d\n",total);
    fprintf(stderr,"Succesful tests: %d\n",good);
    fprintf(stderr,"Success rate: %.0f%%\n",
	    100.0*((double)good/(double)total));
}



