#ifndef lint
static char rcsid [] = "$RCSfile: compare_time_elev.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: compare_time_elev.c                                               *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).          
 *                                                                           *
 *       Contents: Compare driver elevation timing datafile                  *
 *       Created:  1996                                                      *
 *       Author:   kcomanor                                                  *
 *       Remarks:  This routine is used to compare the output from two       *
 *       different ctdb elevation timing runs with the driver util (the      *
 *       same elevation los datafile should be used for this output to       *
 *       be significant.                                                     *
 *                                                                           *
 ****************************************************************************/
#include <stdio.h>
#include <math.h>

/* Min difference in elevations to flag */
#define THRESHOLD .01
#define FNAME_LEN 50
#define MAX_LINE_LEN 200


main(argc, argv)
    int argc;
    char **argv;
{
    char fname[2][FNAME_LEN];
    FILE *f[2],*fptr;    
    char line[MAX_LINE_LEN];    
    float time[2][10000],sum[2];
    int j,i;
    float te_count, elev_finds;
    int patch_num;
        
    if (argc == 3)
    {
	strcpy(fname[0], argv[1]);
	strcpy(fname[1], argv[2]);	
    }
    else
    {
	printf("Usage:\n");
	printf("%s <file1> <file2>\n",argv[0]);
	printf("\nwhere file 1 & 2 are outputs from the driver utility\n");
	printf("labeled <dbname>.<elev>.profile.<ctb format>\n");
	exit(1);
    }

       printf("Output columns:\n");
    printf("1.  run number\n");
    printf("2.  file1 time\n");
    printf("3.  file2 time\n");
    printf("4.  Percent Diff (i.e. negative=file2 faster than file 1)\n");    
    
    for (i=0;i<2;i++)
    {
	j=0;

	if ((f[i] = fopen(fname[i],"r")) == (FILE *)NULL)
	{
	    printf("Error opening file %s\n",fname[i]);
	    exit(1);
	}

	fptr = f[i];
	
	while ( !feof(fptr) )
        {
	    fgets(line, MAX_LINE_LEN, fptr);
	    
	    if ( sscanf(line,"%f\t%f\t%f\t%d",&te_count,&time[i][j],&elev_finds,&patch_num) != 4 )
	      continue;
	    j++;
	}
    }
    sum[0] = sum[1] = 0;

    fclose(f[0]);
    fclose(f[1]);
    

    for (i=0;i<j;i++)
    {
	sum[0] += time[0][i];
	sum[1] += time[1][i];
	printf("%d\t%9.3f\t%9.3f\t%9.3f\n",i,time[0][i],time[1][i],
	       ((time[1][i]-time[0][i])/time[0][i])*100.0);
    }
    printf("Average file1 : %9.3f\n",sum[0]/j);
    printf("Average file2 : %9.3f\n",sum[1]/j);
    printf("Percent Diff (file1-file2)/file1 : %9.3f\n",(sum[0]-sum[1])/sum[0]*100);
    exit(0);    
}
