#ifndef lint
static char rcsid [] = "$RCSfile: driver.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: driver.c                                                         *
 *                                                                          *
 *       Copyright 1995 by SAIC/TASC                                        *
 *                                                                          *
 *       SAIC                          TASC                                 *
 *       20 Burlington Mall Road       55 Walkers Brook Drive               *
 *       Suite 130                     Reading, MA 01867                    *
 *       Burlington, MA 01803          617-942-2000                         *
 *       617-221-7612                                                       *
 *                                                                          *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *       Contents: Used to profile the ctdb local terrain representation    *
 *                 and test various ctdb component functionality            *
 *       Created:  January 1995                                             *
 *       Author:   cbuettner wkbraudaway flchamberlain                       *
 *       Remarks:                                                           *
 *                                                                          *
 ***************************************************************************/
#include <stdext.h> /*common/include/global*/
#include <stdcoord.h>
#include "../libctdb_local.h"
#include "../compiler/ct_reform.h"
#include <libcoordinates.h>
#include <libworld.h>
#include <libntpif.h>
#include <libtime.h>
#include <libgcs.h>
#include <libreader.h>
#include <libgeometry.h>
#include <math.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdalloc.h>

struct main_options_struct
{
    CMD_STRING_OPTION  data;
} main_options;

struct main_options_struct main_options = {
    {
	"Data Directory" , "Specifies directory where data files are stored" ,
	NULL ,
	CMD_STRING ,	"data" ,	NULL ,		"../../../data" ,
    } ,
};

/* #define CTDB_OBJ_DEBUG */
/* Amount of memory terrain db is allowed to use (0 => infinite) */
#define TDB_MSIZE 0
#define RAND_0_TO_N(n) ((((float64)rand())/(float64)0x8FFF)*(float64)n)
#define FILENAME_DIMENSION 	80
#define MODE 0
#define N_ELEVATIONS 0
#define VIEW 0
#define MIN_VIS 0.05
#define N_VEH 0
#define VEH_PTR (CTDB_VEHICLE_LOCATION *)NULL
#define IGNORE0 0
#define IGNORE1 0
#define STEP 249
#define MAX_INTERV_RAY 5000 /* 5 km in x and y - maximum length 7.071 km */
#define MAX_HEIGHT 10   /* 10 meters */
#define MAX_WIDTH  10   /* 10 meters */
#ifdef CTDB_INTERVIS_METRICS
#define MAX_POINTS 1
#else
#define MAX_POINTS 1000
#endif
#define TARGET_WIDTH 4.0
#define TARGET_HEIGHT 10.0 
#define MAX_Z_OFFSET 2.0 /* Max height off ground of both points for 
			  * intervis
			  */
#define MAX_DIST 5000.0	 /* Max change in x and y between intervis points */

CTDB *ctdb=(CTDB *)NULL;

typedef struct ctdb_dim2d
{
    float64  width;
    float64  height;
} CTDB_DIM2D;

typedef struct test_los
{
    CTDB_COORD3D start;
    CTDB_COORD3D end;
    CTDB_DIM2D   shape;
} TEST_LOS;

void perror(const char *s );
int32   entered_priority=0;
#define PATCH_TO_XY_COORDINATES(a,b) \
((float64)(b)*a->incr*a->patch_incr)

#define NUM_TIMES                       1550
#define EDGE_OFFSET     		3500
#define PROFILER_POINTS_DIMENSION 	100
#define PROFILE_THRESHOLD		1000
#define PROFILE_PRIORITY  		-18 
#define NUM_POINTS 			2
#define CTDB_VISIBILITY_UNDEFINED      0

typedef uint32 TIME;

typedef struct 
{
    int32   x0;
    int32   x1;
    int32   y0;
    int32   y1;
    float32 n_te;
    float32 elev_lookup;
    float32 elev_hit_ratio;
    int32   patch_number;
} METRICS;

static METRICS *metric;
static int32   default_priority=0;
static int32 priority_error_displayed = FALSE;

static void
set_system_priority(lpri)
    int32 lpri ;
{
    errno = 0;         /* reset error indicator */
    default_priority = getpriority(PRIO_PROCESS,0);
    
    if (( setpriority(PRIO_PROCESS,0,lpri) != 0 ) &&
	(!priority_error_displayed))
    {
	priority_error_displayed = TRUE;
	perror("ERROR: Priority Setting FAILURE");
    }
    
}

static void 
reset_system_priority()
{
    set_system_priority(default_priority);
}


static int32 parse_num(char *s, int32 c)
{
    int32 i, result = 0;

    for(i=0;i<c;i++)
      result = result * 10 + (s[i] - '0');
    return result;
}

static int32 get_node_id(CTDB_MES_VOLUME *mes, CTDB_MES_NODE *node)
{
    return ((int32)(node - mes->template->all_enclosures));
}

static int32 get_edge_id(CTDB_MES_VOLUME *mes, CTDB_MES_EDGE *edge)
{
    return ((int32) (edge - mes->template->all_apertures));
}


#ifdef NEVER
static int32
ctdb_mes_get_triangle_normal(CTDB *ctdb,
			     int16 mes_id,
			     int16 triangle_id,
			     float64 normal[XYZ]
			     )
{
    CTDB_MES_VOLUME   *mes;
    CTDB_MES_TEMPLATE *template;
    if (!ctdb || !CTDB_MES_VALID_MES_ID(ctdb, mes_id))
      return 0;
    
    mes      = &ctdb->mes_volumes[mes_id];
    template = mes->template;
    ctdb_mesvec_to_gcscsvec(mes, template->all_triangles[triangle_id].normal,
			    ctdb_get_cell_id(ctdb), normal);
    vmat3_unit64(normal, normal);
    return 1;
}
#endif


static void 
write_elevation_profiling_data(ctdb,metric,n_metric)
    CTDB    *ctdb;
    METRICS *metric;
    int32    n_metric;
{
    int32 i ;
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
 
    sprintf(filename,"%s-%04d.elev.profile.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ( (fileptr = fopen(filename,"w")) == (FILE *) NULL )
    {
	printf("Couldn't Open Output Datafile %s\n",filename);
	return;
    }
    
    fprintf(fileptr,"Note: If any of the timing data points exceeds ");
    fprintf(fileptr,"%d then the data will not be written since this is ",
	    PROFILE_THRESHOLD);
    fprintf(fileptr,"assumed that the data point has been corrupted by a ");
    fprintf(fileptr,"system interrupt.  Also if no");
    fprintf(fileptr,"terrain element data was found in the patch the data");
    fprintf(fileptr,"point will not be displayed\n\n");
    fprintf(fileptr,"Column\n");
    fprintf(fileptr,"1 : Terrain Element Count (tri/patch) where patch_side=%d\n",ctdb->incr*ctdb->patch_incr);
    fprintf(fileptr,"2 : Elevation Lookup (usec)\n");
    fprintf(fileptr,"3 : Elevation Finds (hits/attempts)\n");
    fprintf(fileptr,"4 : Patch Number\n");

    for (i=0;i<n_metric;i++,metric++)
    {
	if (metric->elev_lookup < PROFILE_THRESHOLD)
	{
	    fprintf(fileptr,"%f\t%f\t%f\t%d\n",
		    metric->n_te,
		    metric->elev_lookup,
		    metric->elev_hit_ratio,
		    metric->patch_number); 
	}	
    }
    
    fclose(fileptr);    
}

static float64 
dummy_fcn()
{
    return 0;
}

static void
rollover() 
{
#ifdef _GT_
  TIME dummy;
  dummy = time_realtime_clock();
#endif
}



static float32
set_time(start,middle,end,samples)
    TIME  start;
    TIME  middle;
    TIME  end;
    int32 samples;
{
    return (1000.0*((float32)((end - middle)-
			      (middle-start)))/((float32)samples));
}


static FILE *
generate_elevation_datafile(ctdb)
    CTDB *ctdb;
{
  char filename[FILENAME_DIMENSION];
  CTDB_COORD2D pt;
  FILE * fileptr;
  int32 i;
  int32 patch_size=ctdb->incr * ctdb->patch_incr;
  
  printf("Generating Elevation Points Datafile ... Please Wait\n");
  
  sprintf(filename,"%s-%04d.elev.pts",ctdb->name,ctdb->version);
  
  if ((fileptr=fopen(filename,"w"))==(FILE *) NULL)
  {
      printf("Couldn't Write to File %s\n",filename);
      return (FILE *)NULL;
  }
  /*$$$ Function should be initialized with time - later */ 
  srand(time_realtime_clock());

  for (i=0;i<PROFILER_POINTS_DIMENSION;i++)
  {
      pt.x = RAND_0_TO_N(patch_size);
      pt.y = RAND_0_TO_N(patch_size);
      fprintf(fileptr,"%lf\t%lf\n",pt.x,pt.y);
  }
  
  fclose(fileptr);
  
  if ((fileptr=fopen(filename,"r"))==(FILE *) NULL)
  {
      printf("Couldn't Open File %s to Read\n",filename);
      return (FILE *)NULL;
  }
  
  return(fileptr);
  
}


/*
 * Read the GCC coords (x,y,z) from an elevations file, store them
 * in a block of sizeof(CTDB_COORD3D) * num_pt and return a ptr to
 * them
 */
static CTDB_COORD3D *
read_3D_elevation_datafile(num_pt, filename)
    int32 *num_pt;
    char  *filename;
{
    FILE * datafile;
    CTDB_COORD3D *pt,*base_pt;
    CTDB_COORD3D tmp;
    int32 n_pt=0;
    int32 answer;
    
    
    /*  open the file in case closed */
    datafile = fopen(filename,"r");
    
    if (datafile == (FILE *)NULL)
    {
	fprintf(stderr, " File could not be opened \n");
	exit(1);
    }
    
    while ( !feof(datafile) )
    {
	fscanf(datafile,"%lf %lf %lf",&tmp.x,&tmp.y, &tmp.z);
	n_pt++;
    }
    
    rewind(datafile);
    
    base_pt = pt = (CTDB_COORD3D *)ctdb_alloc(0,
					      sizeof(CTDB_COORD3D)*n_pt,
					      "profiling elev points");
    
    while ( !feof(datafile) )
    {
	fscanf(datafile,"%lf %lf %lf",&pt->x,&pt->y, &pt->z);
	pt++;
    }
    
    fclose(datafile);
    
    *num_pt = n_pt-1;
    return(base_pt);
}

/* function that actually compares the two elevations-stored 
 * and the ones that are returned by elev_lookup
 */
void
compare_ground_loc (fname)
    char * fname;
{
    CTDB_COORD3D  pt;
    float64       z_lookup;
    float64       gcs[3];
    int32         cell_id,cell;
    CTDB          *ctdb;
    CTDB_COORD3D  *gcc_pts;
    int32         size, i;
    char          filename[FILENAME_DIMENSION];
    float64       min_lng, min_lat, max_lat, max_lng, origin_lng, origin_lat;
    float64       x,y,z, dist;
    COORD_TCC_PTR driver_tcc; 
    
    sprintf(filename,"%s.grd_cor.pts", fname);

    bzero((char*)gcs, sizeof(float64)*3);

    /* now the data */
    gcc_pts = read_3D_elevation_datafile(&size, filename);

    /* set it up for coord_convert */
   if (coord_init(main_options.data.value,0))
    {
	printf("Error initializing libcoordinates.\n");
	exit(1);
    }

    /* this to get a ctb and init driver_tcc only once */
    pt = *gcc_pts;
    /* get the cell id first */
    cell_id = gcs_lookup_cell_id(pt.x, 
				 pt.y,
				 pt.z);
    /* now get the ctdb */
    ctdb = (CTDB *)gcs_get_tdb(cell_id, FALSE);
    driver_tcc = coord_define_tcc(COORD_UTM_NE,
				  ctdb->origin_northing,
				  ctdb->origin_easting,
				  ctdb->origin_zone_number,
				  ctdb->origin_zone_letter,
				  ctdb->datum,
				  (ctdb->max_x - ctdb->min_x),
				  (ctdb->max_y - ctdb->min_y));
    
    

    for(i=0;i<size;i++)
    {
	pt = *(gcc_pts+i);
 
	/* gotta do it for each pt-get the cell id first */
	cell_id = gcs_lookup_cell_id(pt.x, 
				     pt.y,
				     pt.z);
	/* now get the ctdb */
	ctdb = (CTDB *)gcs_get_tdb(cell_id, FALSE);

	/* convert GCC into GCS z value */
	gcs_gcc3e_to_gcscs(pt.x, pt.y, pt.z, cell_id, 
			 &gcs[0], &gcs[1], &gcs[2]);

	z_lookup = ctdb_lookup_elevation(ctdb, gcs[0], gcs[1]);

	/* compare z values from file and lookup */

/* 	printf(" TCC: %lf\t GCS: %lf Difference %lf\n", 
	       gcs[2], z_lookup, fabs(gcs[2] - z_lookup));
*/
	gcs_lookup_cell_extents(cell_id, &min_lat, &max_lat, 
				&min_lng, &max_lng);
	
	origin_lat = min_lat + max_lat / 2.0;
	origin_lng = min_lng + max_lng / 2.0;

	coord_convert(COORD_LATLON, origin_lat, origin_lng, 0.0, FALSE,
		      COORD_GCSCS, cell, &x, &y, &z);
#define __square(x) x * x
	/* dist in x,y plane */
	dist = fsqrt(__square(fabs(gcs[0] - x)) + __square(fabs(gcs[1] - y)));
	printf(" GCS cell=%d x=%lf y=%lf Difference= %lf Dist from Cell Origin= %lf\n" ,
	       cell_id, gcs[0], gcs[1],  fabs(gcs[2] - z_lookup), dist);
    }
}

/* Top level function to read the gcc files
 */
void 
read_gcc_file_and_compare(db_name, mode)
    char * db_name;
    char * mode;
{

    compare_ground_loc(db_name);
}

static CTDB_COORD2D *
read_elevation_datafile(ctdb,num_pt)
    CTDB  *ctdb;
    int32 *num_pt;
{
    FILE *datafile;
    CTDB_COORD2D *pt,*base_pt;
    CTDB_COORD2D tmp;
    int32 n_pt=0;
    char filename[FILENAME_DIMENSION];
    int32 answer;

    sprintf(filename,"%s-%04d.elev.pts",ctdb->name,ctdb->version);

    printf("Generate elevation points data file? (1 or 0)\n");
    scanf("%d",&answer);
    if (( answer ) || ((datafile=fopen(filename,"r")) == (FILE *) NULL))
      datafile=generate_elevation_datafile(ctdb);
    
    while ( !feof(datafile) )
    {
	fscanf(datafile,"%lf %lf",&tmp.x,&tmp.y);
	n_pt++;
    }
    
    rewind(datafile);

    base_pt = pt = (CTDB_COORD2D *)ctdb_alloc(0,
						sizeof(CTDB_COORD2D)*n_pt,
						"profiling points");
    
    while ( !feof(datafile) )
    {
	fscanf(datafile,"%lf %lf",&pt->x,&pt->y);
	pt++;
    }

    fclose(datafile);

    *num_pt = n_pt-1;
    return(base_pt);
}

static void
generate_intervisibility_datafile(ctdb)
    CTDB *ctdb;
{
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    TEST_LOS *pt;
    int32 i,n_pts;
    float64 elev;
    CTDB_COORD2D max;
    CTDB_COORD2D min;
    CTDB_COORD3D range;
    float64 max_dist = MAX_DIST;    /* Max x or y dist between pt */

    printf("Number of test points to generate > ");
    scanf("%d",&n_pts);

    max.x = ctdb->max_x , max.y = ctdb->max_y;
    min.x = ctdb->min_x , min.y = ctdb->min_y;
    range.x = max.x-min.x , range.y = max.y-min.y , range.z = MAX_Z_OFFSET;  

    sprintf(filename,"%s-%04d.intervis.pts",ctdb->name, ctdb->version);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }

    pt = (TEST_LOS *)ctdb_alloc(0,n_pts * sizeof(TEST_LOS),"intervis_pts");

    printf("Generating %d test points...",n_pts);
    fflush(stdout);
   
    srand(time_realtime_clock());

    for(i=0;i<n_pts;i++)
    {
	/* Find start point */
	pt[i].start.x = min.x + RAND_0_TO_N(range.x);
	pt[i].start.y = min.y + RAND_0_TO_N(range.y);

	/* Find legal end point */
	do
	{
	    /* Generate end point at (startx +- max_dist, starty +- max_dist)*/
	    pt[i].end.x = pt[i].start.x - max_dist + 
	      RAND_0_TO_N(2.0*max_dist);
	    pt[i].end.y = pt[i].start.y - max_dist + 
	      RAND_0_TO_N(2.0*max_dist);
	} while ( !ctdb_point_on_database( ctdb,pt[i].end.x,pt[i].end.y ) );

	/* Retrieve elevations at points and offset from there */
	elev = ctdb_lookup_elevation(ctdb,pt[i].start.x,pt[i].start.y);
	pt[i].start.z = elev +  RAND_0_TO_N(range.z);
	elev = ctdb_lookup_elevation(ctdb,pt[i].end.x,pt[i].end.y);
	pt[i].end.z = elev +  RAND_0_TO_N(range.z);
	pt[i].shape.width = RAND_0_TO_N(MAX_WIDTH);
	pt[i].shape.height= RAND_0_TO_N(MAX_HEIGHT);

	/* Write out to file */
	fprintf(fileptr,"%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f\n",
		pt[i].start.x,
		pt[i].start.y,
		pt[i].start.z,
		pt[i].end.x,
		pt[i].end.y,
		pt[i].end.z,
		pt[i].shape.width,
		pt[i].shape.height);
    }
    fclose(fileptr);
    STDDEALLOC(pt);
    printf("Done.\n");
    return;
}

static TEST_LOS *
read_intervisibility_datafile(ctdb,count)
    CTDB  *ctdb;
    int32 *count;
{
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    TEST_LOS temp;
    int32 i,n_pt,answer ;
    TEST_LOS *pt;

    printf("Generate intervis points data file? (1 or 0)\n");
    scanf("%d",&answer);
    if ( answer )
	generate_intervisibility_datafile(ctdb);

    sprintf(filename,"%s-%04d.intervis.pts",ctdb->name, ctdb->version);
    
    if ((fileptr = fopen(filename, "r")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to read.\n",filename);
	return (TEST_LOS *)NULL;
    }
    
    n_pt = 0;
    printf("Counting input points...");

    while ( !feof(fileptr) )
    {
	if ( fscanf(fileptr, "%lf %lf %lf %lf %lf %lf %lf %lf",
		    &temp.start.x,
		    &temp.start.y,
		    &temp.start.z,
		    &temp.end.x,
		    &temp.end.y,
		    &temp.end.z,
		    &temp.shape.width,
		    &temp.shape.height) == 8 )
	n_pt++;
    }

    rewind(fileptr);
    pt = (TEST_LOS *)ctdb_alloc(0,n_pt*sizeof(TEST_LOS),"intervis_pts");

    printf("Done.\nReading input points...");

    for(i=0;i<n_pt;i++)
      fscanf(fileptr, "%lf %lf %lf %lf %lf %lf %lf %lf",
	     &pt[i].start.x,
	     &pt[i].start.y,
	     &pt[i].start.z,
	     &pt[i].end.x,
	     &pt[i].end.y,
	     &pt[i].end.z,
	     &pt[i].shape.width,
	     &pt[i].shape.height);

    fclose(fileptr);
    *count=n_pt;
    printf("Done.\n%d points read.\n",*count);
    return pt;
}

static void 
sample_intervis_across_database(ctdb)
    CTDB *ctdb;
{
    float64 percent_vis, opacity;
    int32 n_pts;
    TEST_LOS *pt;
    CTDB_STATUS status;
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    int32 i;

    pt = read_intervisibility_datafile(ctdb,&n_pts);

    sprintf(filename, "%s-%04d.intervis.samples.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }

    fprintf(fileptr,"%%vis\tstatus\t\tx1\t\ty1\t\tz1\t\tx2\t\ty2\t\tz2\n");

    for(i=0;i<n_pts;i++)
    {
#ifdef CTDB_OBJ_DEBUG
	printf("Ray %4d\n",i);
#endif     
	percent_vis= ctdb_point_to_point(ctdb,
					 pt[i].start.x,
					 pt[i].start.y,
					 pt[i].start.z,
					 pt[i].end.x,
					 pt[i].end.y,
					 pt[i].end.z,
					 pt[i].end.z+pt[i].shape.height,
					 pt[i].shape.width,
					 NULL, &opacity,
					 MIN_VIS,
					 N_VEH,
					 VEH_PTR,
					 IGNORE0,
					 IGNORE1);

	percent_vis *= opacity;

#ifdef CTDB_OBJ_DEBUG
	printf("\n");	
#endif     
	
			    

	fprintf(fileptr,
		"%5.3f\t%2d\t%9.3f\t%9.3f\t%9.3f\t%9.3f\t%9.3f\t%9.3f\n",
		percent_vis,
		0/*status*/,
		pt[i].start.x,
		pt[i].start.y,
		pt[i].start.z,
		pt[i].end.x,
		pt[i].end.y,
		pt[i].end.z);
    }
    fclose(fileptr);
}

static void 
sample_ground_across_database(ctdb)
    CTDB *ctdb;
{
    int32 hit_status;
    int32 n_pts;
    TEST_LOS *pt;
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    int32 i;
    float64 pt_ref[3];
    CTDB_VEHICLE_LOCATION temp[1];
    int32 temp2;

    pt = read_intervisibility_datafile(ctdb,&n_pts);

    sprintf(filename, "%s-%04d.ground.samples.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }

    fprintf(fileptr,"ground hit\tx1\t\ty1\t\tz1\t\tx2\t\ty2\t\tz2\thit:\t\tx\t\ty\t\tz\tstatus\n");

    for(i=0;i<n_pts;i++)
    {
	hit_status = ctdb_find_ground_intersection
	  (ctdb,
	   pt[i].start.x,
	   pt[i].start.y,
	   pt[i].start.z,
	   pt[i].end.x,
	   pt[i].end.y,
	   pt[i].end.z,
	   pt_ref,
	   CTDB_USE_DEFAULT,
	   0,temp,0,0,&temp2);
			    

	fprintf(fileptr,
		"\t\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%d\n",
		pt[i].start.x,
		pt[i].start.y,
		pt[i].start.z,
		pt[i].end.x,
		pt[i].end.y,
		pt[i].end.z,
		pt_ref[0],pt_ref[1],pt_ref[2],
		hit_status);
    }
    fclose(fileptr);
}

static void
sample_profile_across_database(ctdb)
    CTDB *ctdb;
{
    int32 n_pts;
    TEST_LOS *pt;
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    int32 i,j;
    int32 prof_complete;
    int32 n_prof=1000;
    float64 prof[1000][3];
    int32 water;

    pt = read_intervisibility_datafile(ctdb,&n_pts);
    printf("Shall we include WATER (0/1)");
    scanf("%d",&water);

    sprintf(filename, "%s-%04d.prof_vector.samples.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }
    
    fprintf(fileptr,"\tstatus\t\tx1\t\ty1\t\tx2\t\ty2\n");
    
    for(i=0;i<n_pts;i++)
    {
	n_prof=100;
	prof_complete = ctdb_profile_vector(ctdb,
					    pt[i].start.x,
					    pt[i].start.y,
					    pt[i].end.x,
					    pt[i].end.y,
					    (water?CTDB_MAX_LAND_AND_WATER:
					    CTDB_MAX_LAND_NO_WATER),
					    &n_prof,
					    prof);
	
	
	fprintf(fileptr,
		"\t%2d\t%9.0f\t%9.0f\t%9.0f\t%9.0f\t%d:\n",
		prof_complete,
		pt[i].start.x,
		pt[i].start.y,
		pt[i].end.x,
		pt[i].end.y,
		n_prof
		);
	
	for(j=0; j<n_prof; j++)
	  fprintf(fileptr,
		  "\t%9.0f\t%9.0f\t%9.0f\n",
		  prof[j][0], prof[j][1], prof[j][2]);
    }
    fclose(fileptr);
}

static void 
check_profile_vector(ctdb)
    CTDB *ctdb;
{
    CTDB_COORD3D start,end;
    int32 j;
    int32 prof_complete;
    int32 n_prof=1000;
    float32 x,y;
    float64 prof[1000][3];
    int32 water;
    
    while (1)
    {
	printf("Starting Point (x,y) > ");
	scanf("%f %f",&x,&y);
	start.x = x;
	start.y = y;
	printf("End Point (x,y) > ");
	scanf("%f %f",&x,&y);
	end.x = x;
	end.y = y;
	printf("Shall we include WATER (0/1)");
	scanf("%d",&water);
	
	prof_complete = ctdb_profile_vector(ctdb,
					    start.x,
					    start.y,
					    end.x,
					    end.y,
					    (water?CTDB_MAX_LAND_AND_WATER:
					     CTDB_MAX_LAND_NO_WATER),
					    &n_prof,
					    prof);
	
	printf("\tstatus\t\tx1\t\ty1\t\tx2\t\ty2\n");
	printf("\t%2d\t%9.1lf\t%9.1lf\t%9.1lf\t%9.1lf\t%d:\n",
	       prof_complete,
	       start.x,
	       start.y,
	       end.x,
	       end.y,
	       n_prof
	       );
	
	for(j=0; j<n_prof; j++)
	  printf("\t%9.1lf\t%9.1lf\t%9.1lf\n",
		 prof[j][0], prof[j][1], prof[j][2]);
    }
    
}

static void 
sample_highest_across_database(ctdb)
    CTDB *ctdb;
{
    int32 n_pts;
    TEST_LOS *pt;
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    int32 i;
    float64 highest;
    float64 xy_at_high[2];

    pt = read_intervisibility_datafile(ctdb,&n_pts);

    sprintf(filename, "%s-%04d.highest.samples.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }

    fprintf(fileptr,"\tx1\t\ty1\t\tx2\t\ty2\t\tHigh\t\tx\t\ty\t\tz\n");

    for(i=0;i<n_pts;i++)
    {
	highest= ctdb_find_high_ground(ctdb,
				       pt[i].start.x,
				       pt[i].start.y,
				       pt[i].end.x,
				       pt[i].end.y,
				       CTDB_MAX_LAND_AND_WATER,
				       xy_at_high);
    
			    

	fprintf(fileptr,
		"\t%9.1lf\t%9.1lf\t%9.1lf\t%9.1lf\t\t%9.1lf\t%9.1lf\t%9.1lf\n",
		pt[i].start.x,
		pt[i].start.y,
		pt[i].end.x,
		pt[i].end.y,
		xy_at_high[0], xy_at_high[1], highest);
	
    }
    fclose(fileptr);
}

static void 
sample_mratio_across_database(ctdb)
    CTDB *ctdb;
{
    int32 n_pts;
    TEST_LOS *pt;
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    int32 i;
    float64 ratio;
    float64 start[2];
    float64 end[2];
    float64 distance_to_start = 0.0;
    
    pt = read_intervisibility_datafile(ctdb,&n_pts);

    sprintf(filename, "%s-%04d.mratio.samples.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }

    fprintf(fileptr,"ratio\t\tx1\t\ty1\t\tz1\t\tx2\t\ty2\t\tz2\n");

    for(i=0;i<n_pts;i++)
    {
	start[0] = pt[i].start.x;
	start[1] = pt[i].start.y;
	end[0] = pt[i].end.x;
	end[1] = pt[i].end.y;
	ratio = ctdb_find_max_ratio(ctdb, pt[i].start.z, distance_to_start,
				    CTDB_MAX_LAND_AND_WATER | CTDB_INCL_LINEAR |
				    CTDB_INCL_CANOPY | CTDB_INCL_VOLUME, start, end);

	fprintf(fileptr,
		"%9.3lf\t%9.1lf\t%9.1lf\t%9.1lf\t%9.1lf\t%9.1lf\t%9.1lf\n",
		ratio,
		pt[i].start.x,
		pt[i].start.y,
		pt[i].start.z,
		pt[i].end.x,
		pt[i].end.y,
		pt[i].end.z
		);
    }
    fclose(fileptr);
}

static void
profile_intervisibility(ctdb)
    CTDB *ctdb;
{
    uint32 t_start, t_end, t_middle;
    float64 opacity;
    int32 j;
    int32 count=0;
    CTDB_STATUS status=0;
    CTDB_STATUS v = 0;
    FILE *fileptr,*infileptr;
    TEST_LOS pt;
    char filename[FILENAME_DIMENSION];
#ifdef CTDB_INTERVIS_METRICS
    fileptr=stdout; 
#else    
    sprintf(filename,"%s-%04d.intervis.profile.%d",
	    ctdb->name,
	    ctdb->version,
	    ctdb->input_format);

    if ((fileptr = fopen(filename, "w")) == (FILE *)NULL)
    {
	fprintf(stderr,"Couldn't open file %s to write.\n",filename);
	return;
    }
#endif
    
    fprintf(fileptr,"Run Number\tInterv Time\tTotal Status\n");
    fflush(fileptr);

    /* Read in Data Points */
    (void)read_intervisibility_datafile(ctdb,&count);
    count=0;
    sprintf(filename,"%s-%04d.intervis.pts",ctdb->name, ctdb->version);
    
    if ((infileptr = fopen(filename, "r")) == (FILE *)NULL)
    {
	fprintf(stderr,"Couldn't open file %s to read.\n",filename);
	return;
    }

    while ( !feof(infileptr) )
    {
	fscanf(infileptr, "%lf %lf %lf %lf %lf %lf %lf %lf",
	       &pt.start.x,
	       &pt.start.y,
	       &pt.start.z,
	       &pt.end.x,
	       &pt.end.y,
	       &pt.end.z,
	       &pt.shape.width,
	       &pt.shape.height);
#ifdef CTDB_INTERVIS_METRICS
	printf("\n\nIntervis Ray\nStart(x,y,z)\t= (%9.3f , %9.3f , %9.3f)\n",
	       pt.start.x,pt.start.y,pt.start.z);
	printf("End(x,y,z)\t= (%9.3f , %9.3f , %9.3f)\n",
	       pt.end.x,pt.end.y,pt.end.z);
	printf("Shape(w,h)\t= (%9.3f , %9.3f)\n",pt.shape.width,
	       pt.shape.height);
#endif
	
	t_start = time_realtime_clock();
	rollover();
	for (j=0;j<MAX_POINTS;j++)
	{
	    v +=  dummy_fcn(ctdb,pt.start.x,pt.start.y,pt.start.z,
			    pt.end.x,pt.end.y,pt.end.z,
			    pt.end.z+pt.shape.height,
			    pt.shape.width,NULL,&opacity,MIN_VIS,N_VEH,
			    VEH_PTR,IGNORE0,IGNORE1);	      
	}
	
	t_middle = time_realtime_clock();	      
	rollover();
	for (j=0;j<MAX_POINTS;j++)
	{
	    v +=  ctdb_point_to_point(ctdb,pt.start.x,pt.start.y,
				      pt.start.z,pt.end.x,pt.end.y,pt.end.z,
				      pt.end.z+pt.shape.height,pt.shape.width,
				      NULL,&opacity,MIN_VIS,N_VEH,VEH_PTR,
				      IGNORE0,IGNORE1);	      
	}  
	t_end = time_realtime_clock();
#ifndef CTDB_INTERVIS_METRICS 
	fprintf(fileptr,"%d\t\t%9.3f\t\t%d\n",count++,
		set_time(t_start,t_middle,t_end,MAX_POINTS),status);
#endif
	fflush(fileptr);
    }
    fclose(fileptr);
}

static void
test_intervisibility_ray(ctdb)
    CTDB *ctdb;
{
    CTDB_COORD3D start,end;
    CTDB_STATUS status;
    CTDB_DIM2D shape;
    float64 percent_vis, opacity;
    float32 x,y,z,zl,zh;

    while (1)
    {	
	printf("Starting Point (x,y,z) > ");
	scanf("%f %f %f",&x,&y,&z);
	start.x = x;
	start.y = y;
	start.z = z;
	printf("End Point (x,y,zl,zh) > ");
	scanf("%f %f %f %f",&x,&y,&zl,&zh);
	end.x = x;
	end.y = y;
	end.z = zl;
	shape.height = zh-zl;
	shape.width = TARGET_WIDTH;
	
	printf("\tTarget Shape (W=%9.3f,H=%9.3f)\n",shape.width,shape.height);
	
	percent_vis = ctdb_point_to_point(ctdb,
					  start.x,
					  start.y,
					  start.z,
					  end.x,
					  end.y,
					  end.z,
					  end.z+shape.height,
					  shape.width,
					  NULL, &opacity,
					  MIN_VIS,
					  N_VEH,
					  VEH_PTR,
					  IGNORE0,
					  IGNORE1);
	percent_vis *= opacity;
					  
					  
	printf("\nStart X\t\tStart Y\t\tStart Z");
	printf("\t\tEnd X\t\tEnd Y\t\tEnd Z");
	printf("\t\tVisibility\tStatus\n");
	printf("%9.3f\t%9.3f\t%9.3f\t",start.x,start.y,start.z);	    
	printf("%9.3f\t%9.3f\t%9.3f\t",end.x,end.y,end.z);
	printf("%9.3f\t%d\n",percent_vis,status);
    }    
}


static void
limit_number_of_patches(ctdb,min_p,max_p)
    CTDB             *ctdb;
    CTDB_COORD2D_I32 *min_p;
    CTDB_COORD2D_I32 *max_p;
{   
    int32 answer;
    
    printf("Would you like to like to perform sampling/profiling on every\n");
    printf("patch in the database? (1 or 0) ");
    scanf("%d",&answer);
    if ( !answer )
    {
	printf("\tMinimum patch x and y coordinates > ");
	scanf("%d %d",&min_p->x,&min_p->y);
	printf("\tMaximum patch x and y coordinates > ");
	scanf("%d %d",&max_p->x,&max_p->y);
	max_p->x++;
	max_p->y++;
    }
    else
    {
	min_p->x = ctdb->min_x_post * ctdb->patch_inv_incr;
	min_p->y = ctdb->min_y_post * ctdb->patch_inv_incr;
	max_p->x = ctdb->max_x_post * ctdb->patch_inv_incr;
	max_p->y = ctdb->max_y_post * ctdb->patch_inv_incr;
    }
}


static void 
sample_elevations_across_database(ctdb, qual)
    CTDB *ctdb;
    CTDB_ELEV_QUALIFIER qual;
{
    CTDB_COORD2D_I32 min_p,max_p;
    float64 x0,y0;
    char filename[FILENAME_DIMENSION];    
    FILE *fileptr;    
    int32 patch_number;    
    CTDB_COORD2D *deltas;
    int32 n_deltas=0;
    int32 px,py;
    int32 point_indx;
    CTDB_COORD3D pt;    
 
    deltas = read_elevation_datafile(ctdb,&n_deltas);

    sprintf(filename, "%s-%04d.elev.samples.%d",
	    ctdb->name,ctdb->version,ctdb->input_format);

    if ((fileptr = fopen(filename,"w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }
    
    fprintf(fileptr, "Columns:\n1. X\n2. Y\n3. Elev.\n4. Patch #\n");

    limit_number_of_patches(ctdb,&min_p,&max_p);
    
    for (py=min_p.y;py<max_p.y;py++)
    {
	y0 = PATCH_TO_XY_COORDINATES(ctdb,py)  ;
	
	for (px=min_p.x;px<max_p.x;px++)
	{
	    x0=PATCH_TO_XY_COORDINATES(ctdb,px)  ;
	    patch_number  = py*ctdb->patches_wide + px;
	    
	    for(point_indx=0; point_indx<n_deltas;point_indx++)
	    {
		pt.x = x0 + deltas[point_indx].x;
		pt.y = y0 + deltas[point_indx].y;
		pt.z = ctdb_lookup_qual_elevation(ctdb,pt.x,pt.y,
						  qual,NULL,NULL);
		fprintf(fileptr, "%9.3f\t%9.3f\t%9.3f\t%d\n",
			pt.x,
			pt.y,
			pt.z,
			patch_number);		
	    }
	}
    }
    
    STDDEALLOC(deltas);
    fclose(fileptr);    
}


static void 
sample_max_elevations_across_database(CTDB *ctdb,int32 canopies)
{
    CTDB_COORD2D_I32 min_p,max_p;
    float64 x0,y0;
    char filename[FILENAME_DIMENSION];    
    FILE *fileptr;    
    int32 patch_number;    
    CTDB_COORD2D *deltas;
    int32 n_deltas=0;
    int32 px,py;
    int32 point_indx;
    CTDB_COORD3D pt;    
 
    deltas = read_elevation_datafile(ctdb,&n_deltas);

    sprintf(filename, "%s-%04d.maxelev%1d.samples.%d",
	    ctdb->name,ctdb->version,canopies,ctdb->input_format);

    if ((fileptr = fopen(filename,"w")) == (FILE *)NULL)
    {
	printf("Couldn't open file %s to write.\n",filename);
	return;
    }
    
    fprintf(fileptr, "Columns:\n1. X\n2. Y\n3. Elev.\n4. Patch #\n");

    limit_number_of_patches(ctdb,&min_p,&max_p);
    
    for (py=min_p.y;py<max_p.y;py++)
    {
	y0 = PATCH_TO_XY_COORDINATES(ctdb,py)  ;
	
	for (px=min_p.x;px<max_p.x;px++)
	{
	    x0=PATCH_TO_XY_COORDINATES(ctdb,px)  ;
	    patch_number  = py*ctdb->patches_wide + px;
	    
	    for(point_indx=0; point_indx<n_deltas;point_indx++)
	    {
		pt.x = x0 + deltas[point_indx].x;
		pt.y = y0 + deltas[point_indx].y;
		pt.z = ctdb_lookup_max_elevation(ctdb,pt.x,pt.y,canopies);
		fprintf(fileptr, "%9.3f\t%9.3f\t%9.3f\t%d\n",
			pt.x,
			pt.y,
			pt.z,
			patch_number);		
	    }
	}
    }
    
    STDDEALLOC(deltas);
    fclose(fileptr);    
}


static void
time_elevation_lookup(ctdb,pcx,pcy,d,n_d)
    CTDB          *ctdb;
    float64        pcx;
    float64        pcy;   
    CTDB_COORD2D  *d;
    int32          n_d;
{
    int32 j,i;
    uint32 start, end, middle;
    float64 checksum = 0;
    float64 v = 0;
    CTDB_COORD2D pt;
    
    set_system_priority(entered_priority);
    start = time_realtime_clock();
    
    for (j=0,i=0;j<NUM_TIMES;j++,i++)
    {
	rollover();
	i=i%n_d ;
	pt.x = pcx+d[i].x;
	pt.y = pcy+d[i].y;
	v += (dummy_fcn(ctdb,pt.x,pt.y)>0);
    }
    
    
    middle = time_realtime_clock();
    
    for (j=0,i=0;j<NUM_TIMES;j++,i++)
    {
	rollover();
	i=i%n_d;
	pt.x = pcx+d[i].x;
	pt.y = pcy+d[i].y;
	/* 0 elevation not necessarily bad */
	checksum += (ctdb_lookup_elevation(ctdb,pt.x,pt.y) >= 0); 
    }
    
    end = time_realtime_clock();
    reset_system_priority();
    fflush(stdout);
    metric->elev_hit_ratio = ((float32)checksum)/((float32)NUM_TIMES);
    metric->elev_lookup    = set_time(start,middle,end,NUM_TIMES);
}

static void
print_usage_message(routine)
  char *routine;
{
  printf("Usage: %s <dbname> <none|single|multi>\n", routine);
  exit(0);
}


static void
profile_elevation_lookup(ctdb)
    CTDB *ctdb;
{
    CTDB_COORD2D_I32 min_p, max_p;
    float64 x0,y0,x1,y1;
    METRICS *base_metric;
    int32 n_metric = 0;
    CTDB_COORD2D *deltas;
    int32 n_deltas=0;
    int32 px,py;
   
    time_init(0);
    deltas = read_elevation_datafile(ctdb,&n_deltas);
    base_metric = metric =  (METRICS *)
      ctdb_alloc(0,
		  sizeof(METRICS)*ctdb->patches_wide*ctdb->patches_high,
		  "metrics collection buffer");
    
    limit_number_of_patches(ctdb,&min_p,&max_p);
    for (py=min_p.y;py<max_p.y;py++)
    {
	y0 = PATCH_TO_XY_COORDINATES(ctdb,py)  ;
	y1 = PATCH_TO_XY_COORDINATES(ctdb,py+1);
	
	for (px=min_p.x;px<max_p.x;px++)
	{
	    x0=PATCH_TO_XY_COORDINATES(ctdb,px)  ;
	    x1=PATCH_TO_XY_COORDINATES(ctdb,px+1);
	    metric->x0            = (int)x0 ; 
	    metric->x1            = (int)x1 ;
	    metric->y0            = (int)y0 ; 
	    metric->y1            = (int)y1 ;
	    metric->patch_number  = py*ctdb->patches_wide + px;
	    metric->n_te          = 
	      ctdb_terrain_element_density(ctdb,metric->patch_number);
	    time_elevation_lookup(ctdb,x0,y0,deltas,n_deltas);
	    metric++;
	    n_metric++;
        }
    }
    
    write_elevation_profiling_data(ctdb,base_metric,n_metric);

    STDDEALLOC(metric);
    STDDEALLOC(deltas);
}

static void
dump_geometry_data(A,S)
    AREA  *A;
    char  *S;
{
    int32 i;
    
    printf("Dump of %s Data\n",S);
    printf("Number of Points           : %d\n",A->count);
    printf("Square Area of %s:\t%lf\n",S,ctdb_calculate_area(A) );

    for (i=0;i<A->count;i++)
      printf("\t( %8.2f , %8.2f )\n",A->pts[i].x,A->pts[i].y);   

    printf("Extent : upper (%f,%f)\tlower (%f,%f)\n",
	   A->extent_upperleft.x , A->extent_upperleft.y,
	   A->extent_lowerright.x, A->extent_lowerright.y);
}


static void
test_virtual_grid_logic()
{
    AREA T,G,I;
    int32 i;
    CTDB_COORD2D pt;
    int32 again=1;
    
    ctdb_init_area_struct(CTDB_PTS_PER_QUADRALATERAL,&G);
    printf("Enter Grid Points in Patch Coordinates\n ");
    for (i=0;i<CTDB_PTS_PER_QUADRALATERAL;i++)
    {
	printf("\tx > ");
	scanf("%lf",&pt.x);
        printf("\ty > ");
	scanf("%lf",&pt.y);
        ctdb_add_point(&G,pt.x,pt.y);
    }

    ctdb_find_area_extents(&G);      
    dump_geometry_data(&G,"grid");

    ctdb_init_area_struct(CTDB_MAX_INTERSECTION_PTS,&I);
    ctdb_init_area_struct(CTDB_PTS_PER_TRIANGLE,&T);

    while ( again )
    {
	T.count=0;
	printf("Enter Triangle Points in Patch Coordinates\n ");
	for (i=0;i<3;i++)
	{
	    printf("\tx > ");
	    scanf("%lf",&pt.x);
	    printf("\ty > ");
	    scanf("%lf",&pt.y);
	    ctdb_add_point(&T,pt.x,pt.y);
	}
	ctdb_find_area_extents(&T);          
	dump_geometry_data(&T,"triangle");
	
	ctdb_find_edge_intersection(&G,&T,&I);
	dump_geometry_data(&I,"intersection");
	
	printf("Would you like to check another triangle? 1=yes 0=no");
	scanf("%d",&again);
    }
    ctdb_free_area_struct(&T);
    ctdb_free_area_struct(&G);
    ctdb_free_area_struct(&I);
}

/* This isn't much of a test, but it at least makes sure the routine runs... */
static void
run_raster(ctdb)
    CTDB *ctdb;
{
    float64 x0,y0,z0,x1,y1,zl,zh,width;
    float64 xmin, xmax, zmin, zmax;
    int32 ix0,iy0,ix1,iy1,ih,iw;
    int32 ftype = FC_VOLUME;
    float64 actual_vis_area;
    float64 result;
    CTDB_RASTER my_grid;
    int32 i;
    
    for(i=0;i<CTDB_RASTER_SIZE;i++)
      if (i<16)
	my_grid[i] = 0x00ffff00;
      else
	my_grid[i] = 0;
    
    printf("Enter start point (x y):  ");
    scanf("%d %d",&ix0,&iy0);
    x0 = ix0;
    y0 = iy0;
    z0 = ctdb_lookup_elevation(ctdb, x0, y0);
    
    printf("Enter end point (x y):  ");
    scanf("%d %d",&ix1,&iy1);
    x1 = ix1;
    y1 = iy1;
    zl = ctdb_lookup_elevation(ctdb, x1, y1);

    printf("Enter target width and height:  ");
    scanf("%d %d",&iw,&ih);
    width = iw;
    zh = zl + ih;
    xmin = x1 - width;
    xmax = x1 + width;



    zmin = zl - width;
    zmax = zh + width;

    result = ctdb_ptop_raster(ctdb, x0, y0, z0, x1, y1, zl, zh,
			      width, ftype, 1.0, 0.0, my_grid,
			      &xmin, &xmax, &zmin, &zmax, 
			      &actual_vis_area, 0,
			      (CTDB_VEHICLE_LOCATION *)NULL, 0, 0);

    printf("Result = %f\n",result);
    printf("Actual vis area = %f\n",actual_vis_area);
}

/* Routine to test ctdb_lookup_tin_triangles */
static void
get_triangles(ctdb)
    CTDB *ctdb;
{
    char filename[FILENAME_DIMENSION];
    FILE *fileptr;
    int32 patch_num;
    int32 max_triangles;
    int32 cross_bdy;
    CTDB_TIN_TRIANGLE *data;
    int32 ret;
    int32 i;

    sprintf(filename,"%s-%04d.triangles",
	    ctdb->name,
	    ctdb->version);

    if ( (fileptr = fopen(filename,"w")) == (FILE *) NULL )
    {
	printf("Couldn't Open Output Datafile %s\n",filename);
	return;
    }

    printf("Enter patch number:  ");
    scanf("%d",&patch_num);
    printf("Enter max # of triangles:  ");
    scanf("%d",&max_triangles);
    printf("Generate inter-patch adjacenies? (0 = no, 1 = yes):  ");
    scanf("%d",&cross_bdy);

    data = (CTDB_TIN_TRIANGLE *)STDALLOC(max_triangles * 
					 sizeof(CTDB_TIN_TRIANGLE));
    
    ret = ctdb_lookup_tin_triangles(ctdb, patch_num, data, &max_triangles,
				    cross_bdy);

    if (!ret)
    {
	printf("\nToo many triangles (%d) in patch.\n",max_triangles);
    }
    else
    {
	printf("\nFound %d triangles:\n\n",max_triangles);
	for(i=0;i<max_triangles;i++)
        {
	    fprintf(fileptr,"Triangle %d:\n",i);
	    fprintf(fileptr,"Soil type: %d\n",data[i].soil);
	    fprintf(fileptr,"Verts:\n");
	    fprintf(fileptr,"\t%6.2f\t%6.2f\t%6.2f\n",
		   data[i].verts[0][X],
		   data[i].verts[0][Y],
		   data[i].verts[0][Z]);
	    fprintf(fileptr,"\t%6.2f\t%6.2f\t%6.2f\n",
		   data[i].verts[1][X],
		   data[i].verts[1][Y],
		   data[i].verts[1][Z]);
	    fprintf(fileptr,"\t%6.2f\t%6.2f\t%6.2f\n",
		   data[i].verts[2][X],
		   data[i].verts[2][Y],
		   data[i].verts[2][Z]);
	    fprintf(fileptr,"Adjacenies:\n");
	    fprintf(fileptr,"\tEdge 0 triangle: %4d\tPatch: %d\n",
		   data[i].adj_triangles[0],
		   data[i].adj_patches[0]);
	    fprintf(fileptr,"\tEdge 1 triangle: %4d\tPatch: %d\n",
		   data[i].adj_triangles[1],
		   data[i].adj_patches[1]);
	    fprintf(fileptr,"\tEdge 2 triangle: %4d\tPatch: %d\n",
		   data[i].adj_triangles[2],
		   data[i].adj_patches[2]);		   
	}
    }
    STDDEALLOC(data);
    fclose(fileptr);
    return;
}

static void
database_handler(ctdb,dbname,ctdb_resident)
    CTDB            *ctdb;
    char            *dbname;
    CTDB_STATUS     *ctdb_resident;
{
    char filename[FILENAME_DIMENSION],depth[FILENAME_DIMENSION];

    FILE *fptr;
    
    ctdb_print_description(ctdb);

/*    if ( (fptr=fopen("make.depth","rb")) != (FILE *)NULL )
      fscanf(fptr,"DEPTH = %s",depth);
    else
    {
	fprintf(stdout,"run `config_dir make.depth` and try again\n");
	exit(1);
    }
    
    sprintf(filename,"%s/terrain/%s.ctb",depth,dbname);
    
    if ( *ctdb_resident != CTDB_STATUS_OK )
    {
	*ctdb_resident = CTDB_STATUS_OK;
	ctdb_read(filename,ctdb,TDB_MSIZE);
	ctdb_print_description(ctdb);
    }
*/
}

static CTDB_ELEV_QUALIFIER
determine_elev_qualifier(void)
{    
    static CTDB_ELEV_QUALIFIER qualifier[11]=
    {
	CTDB_USE_DEFAULT,
	CTDB_INCL_LAND,
	CTDB_INCL_WATER,
	CTDB_INCL_VOLUME,
	CTDB_INCL_LINEAR,
	CTDB_INCL_CANOPY,
	CTDB_INCL_LAID_LINEAR,
	CTDB_INCL_MATERIAL,
	CTDB_RETURN_MAX,
	CTDB_RETURN_MIN,
	CTDB_USE_DEFAULT
    }; 
      
    CTDB_ELEV_QUALIFIER qual=CTDB_USE_DEFAULT;
    int32 answer;
       
    printf("what is your qualifier?\n");
    printf("\t\t%2d. LAND\n",1);
    printf("\t\t%2d. WATER\n",2);
    printf("\t\t%2d. VOLUMES\n",3);
    printf("\t\t%2d. LINEARS\n",4);
    printf("\t\t%2d. CANOPIES\n",5);
    printf("\t\t%2d. LAID LINEARS\n",6);
    printf("\t\t%2d. MATERIAL\n",7);
    printf("\t\t%2d. MAXIMUM\n",8);
    printf("\t\t%2d. MINIMUM\n",9);
    printf("\t\t%2d. USE DEFAULT\n",10);
    printf("\t\t%2d. EVERYTHING\n",11);
    printf("\t\t%2d. DONE\n",0);
    
    while (1)
    {
	scanf("%d",&answer);
	
	if ( !answer )
	  return qual;
	
	if ( answer == 8 )
	  return 0;
	else if (answer == 11)
	  return (qual | 
		  CTDB_ALL_FEATURES   |
		  CTDB_LAND_AND_WATER |
		  CTDB_ONLY_POSITIVE_SURFACES);
	
	qual |= qualifier[answer];
    }
}


static CTDB_OCEAN_QUALIFIER
determine_ocean_qualifier(char *pick)
{   
    static char answer_list[6][20]=
    {
	{
	    "CTDB_NULL"
	}
	,
	{
	    "CTDB_SURF_HEIGHT"
	}
	,
	{
	    "CTDB_TEMP"
	}
	,
	{
	    "CTDB_PRIMARY"
	}
	,
	{
	    "CTDB_SECONDARY"
	}
	,
	{
	    "CTDB_TIDAL_HEIGHT"
	}
    };
    
    static CTDB_OCEAN_QUALIFIER qualifier[6]=
    {
	NULL,
	CTDB_SURF_HEIGHT,
	CTDB_TEMP,
	CTDB_PRIMARY,
	CTDB_SECONDARY,
	CTDB_TIDAL_HEIGHT
    }; 
      
    CTDB_OCEAN_QUALIFIER qual=NULL;
    int32 answer;
       
    printf("what is your qualifier?\n");
    printf("\t\t%2d. SURF HEIGHT\n",1);
    printf("\t\t%2d. TEMPERATURE\n",2);
    printf("\t\t%2d. PRIMARY WAVE\n",3);
    printf("\t\t%2d. SECONDARY WAVE\n",4);
    printf("\t\t%2d. TIDAL HEIGHT\n",5);
    printf("\t\t%2d. DONE\n",0);
    scanf("%d",&answer);
    sprintf(pick,answer_list[answer]);
    return answer;
    
#ifdef NEVER    
    while (1)
    {
	scanf("%d",&answer);
	
	if ( !answer )
	  return qual;
	
	if ( answer == 8 )
	  return 0;
	
	qual |= qualifier[answer];
    }
#endif
}
#define CTDB_METERS_TO_PATCH_NUMBER(ctdb,x,y)\
((int32)(PATCH_X((x)*(ctdb)->meters_to_patch) + \
	 PATCH_Y((y)*(ctdb)->meters_to_patch) * (ctdb)->patches_wide ))

test_it()
{
    
    typedef enum octdb_terrain_subclass
    {
	OCTDB_FS_NA              = 0,  /* Unknown or not applicable */
	OCTDB_FS_SKIN            = 1,  /* Terrain Skin Subclass */
	OCTDB_FS_MULTI           = 2,  /* Multilevel Terrain Subclass */
	OCTDB_FS_WATER           = 3,  /* Terrain Water subclass */
	OCTDB_FS_ROAD            = 4,  /* Laid Linear Road */
	OCTDB_FS_MES             = 5,  /* Multi Elevation structure */
	
	/* FC_VOLUME subclasses */
	OCTDB_FS_BUILDING        = 0,
	OCTDB_FS_ROCK_DROP       = 1,
	/* add more FC_VOLUME subclasses here */
	
	/* FC_LINEAR subclasses */
	OCTDB_FS_TREELINE        = 0,
	OCTDB_FS_CONCERTINA_WIRE = 1,
	OCTDB_FS_FENCE           = 2,
	OCTDB_FS_MINEFIELD_FENCE = 3,
	OCTDB_FS_DRAGONS_TEETH   = 4
	  /* add more FC_LINEAR subclasses here */
    } OCTDB_TERRAIN_SUBCLASS;
    printf("LINEAR %d,%d,%d,%d,%d\n",
	   OCTDB_FS_TREELINE,
	   OCTDB_FS_CONCERTINA_WIRE,
	   OCTDB_FS_FENCE,
	   OCTDB_FS_MINEFIELD_FENCE,
	   OCTDB_FS_DRAGONS_TEETH);
    printf("BUILDINGS %d,%d\n",
	   OCTDB_FS_BUILDING,
	   OCTDB_FS_ROCK_DROP);
}



get_gcs_xy_from_tcc(CTDB         *ctdb,
		    COORD_TCC_PTR tcc,
		    float64       x0,
		    float64       y0,
		    float64       z0,
		    int32        *cell_out,
		    float64      *x_out,
		    float64      *y_out,
		    float64      *z_out)
{
    float64 x,y,z;
    int32 cell;   
    
    cell = ctdb->cell_id;
    printf("TCC: x = %lf\t\ty= %lf\t\tz= %lf\n",x0, y0, z0);
    
    /* On the ocean TCC */
    coord_convert(COORD_TCC,tcc,x0,y0,z0,
		  COORD_GCC,&x,&y,&z);
    
    printf("GCC: x = %lf\ty= %lf\tz= %lf\n",
	   x, y, z);
    
    coord_convert(COORD_GCC,x,y,z,
		  COORD_GCSCS,cell,&x,&y,&z);
    
    printf("GCS: x = %lf\t\ty= %lf\t\tz= %lf\tcell= %d\n",
	   x, y, z, cell);

    *x_out = x;
    *y_out = y;
    *z_out = z;
    *cell_out = cell;
    
    coord_convert(COORD_GCS,cell,x,y,z,
		  COORD_GCC,&x,&y,&z);
    
    printf("GCC: x = %lf\ty= %lf\tz= %lf\n",
	   x, y, z);
    
    coord_convert(COORD_GCC,x,y,z,
		  COORD_TCC,tcc,&x,&y,&z);
    
    printf("TCC: x = %lf\t\ty= %lf\t\tz= %lf\n",
	   x0, y0, z0);
}

static float64
find_z_from_gcs_patch_water(CTDB   *ctdb, 
			    float64 x,
			    float64 y)
{
    static CTDB    *lastctdb=NULL;    
    float64         offset[XYZ], x_sq, z_sq, radius;
    static float64  one_over_twice_radius;
    static float64  direction[XY];
    static float64  x_offset, y_offset;
    float64         d_squared;
    
    if (lastctdb != ctdb)
    {
	/* $$$ future have GCS store these values for 
	 * easy retrieval 
	 */
	gcs_get_origin_offsets(ctdb->cell_id, &x_offset,&y_offset);
#if 0
	printf("origin offset (%lf , %lf)\n",x_offset,y_offset);
#endif
	(void)gcs_get_origin_in_gcc(ctdb->cell_id, offset);
	x_sq = offset[X]*offset[X] + offset[Y]*offset[Y];
	z_sq = offset[Z]*offset[Z];
	radius = WGS84_MINOR_AXIS*x_sq/(x_sq+z_sq)
	  +WGS84_MAJOR_AXIS*z_sq/(x_sq+z_sq);
	one_over_twice_radius = 0.5/radius;
	lastctdb = ctdb;
    }

    /* shift the GCS (x,y) point into the proper frame of reference
     * were (0,0) is at the cell tangent point.
     */
    x += x_offset , y += y_offset;
    d_squared = x * x + y * y;
#if 0
    printf("extrapolated (x,y,z) = (%lf , %lf , %lf )\n",x-x_offset,y-y_offset,
	   -d_squared*one_over_twice_radius);
#endif
    return -d_squared*one_over_twice_radius;
}

static float64 alpha_sq = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS;
static float64 gamma_sq = WGS84_MINOR_AXIS*WGS84_MINOR_AXIS;

static float64 gamma_alpha_sq = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS *
                                WGS84_MINOR_AXIS*WGS84_MINOR_AXIS;
#define SQUARE(x) (x*x)

static float64
find_z_from_gcs_patch_water_ellipsoid(CTDB   *ctdb, 
				      float64 x,
				      float64 y)
{
    static CTDB    *lastctdb=NULL;    
    static float64  origin[XYZ], x_off, y_off, rot[3][3];
    static float64 tmp[XYZ];
    static float64 R, T;
    float64 x0, y0, z;
    float64 A, B, C;
    float64 S, discriminant;

    if (lastctdb != ctdb)
    {
	/* $$$ future have GCS store these values for 
	 * easy retrieval 
	 */
	(void)gcs_get_origin_in_gcc(ctdb->cell_id,origin);
	(void)gcs_get_rot_mat(ctdb->cell_id, rot);
	gcs_get_origin_offsets(ctdb->cell_id, &x_off,
			       &y_off);


	/* Term for our quadratic equation (see below).  
	 * We save off the two terms that only depend on the cell,
	 * and not x or y.
	 */
	R = gamma_sq * (SQUARE(rot[0][2])+SQUARE(rot[1][2])) + 
	    alpha_sq * SQUARE(rot[2][2]);
	
	lastctdb = ctdb;

	tmp[X] = gamma_sq * rot[0][2];
	tmp[Y] = gamma_sq * rot[1][2];
	tmp[Z] = alpha_sq * rot[2][2];
    }

    /* shift the GCS (x,y) point into the proper frame of reference
     * were (0,0) is at the cell tangent point.
     */
    x0 = x + x_off;
    y0 = y + y_off;


    /* Calculate the same GCS point in GCC: pull out some constants:
     * 
     * U = A + rot[0][2]*z
     * V = B + rot[1][2]*z
     * W = C + rot[2][2]*z
     */
    A = rot[0][0]*x0 + rot[0][1]*y0 + origin[X];
    B = rot[1][0]*x0 + rot[1][1]*y0 + origin[Y];
    C = rot[2][0]*x0 + rot[2][1]*y0 + origin[Z];

    /* Solving the ellipsoid equation: (U^2 + V^2)/alpha_sq + (W^2)/gamma_sq
     * in terms of z:
     *
     * R*z^2 + S*z + T = 0
     */

    S = tmp[X]* A + tmp[Y]*B + tmp[Z]*C;
    S *= 2;
    

    T = gamma_sq * (SQUARE(A) + SQUARE(B)) + 
        alpha_sq * SQUARE(C) -
	gamma_alpha_sq;

    discriminant = SQUARE(S) - 4 * R * T;
    if (discriminant >= 0)
    {
	z = (-S + sqrt(discriminant)) / (R *2);
	return z;
    }
    else
    {
	printf("Negative Discriminant: no solutions\n");
	return 0;
    }
}

static float64 beta_sq = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS;

static float64
find_z_from_gcs_patch_water_ellipsoid2(CTDB   *ctdb, 
				       float64 x,
				       float64 y)
{
    static CTDB    *lastctdb=NULL;    
    static float64  origin[XYZ], x_off, y_off, rot[3][3];
    float64 x0, y0, z;
    float64 A, B, C;
    float64 R, S, T, discriminant;

    if (lastctdb != ctdb)
    {
	/* $$$ future have GCS store these values for 
	 * easy retrieval 
	 */
	(void)gcs_get_origin_in_gcc(ctdb->cell_id,origin);
	(void)gcs_get_rot_mat(ctdb->cell_id, rot);
	gcs_get_origin_offsets(ctdb->cell_id, &x_off,
			       &y_off);
	lastctdb = ctdb;
    }

    /* shift the GCS (x,y) point into the proper frame of reference
     * were (0,0) is at the cell tangent point.
     */
    x0 = x + x_off;
    y0 = y + y_off;

    A = rot[0][0]*x0 + rot[0][1]*y0 + origin[X];
    B = rot[1][0]*x0 + rot[1][1]*y0 + origin[Y];
    C = rot[2][0]*x0 + rot[2][1]*y0 + origin[Z];
    
    R = SQUARE(rot[0][2]) * beta_sq  * gamma_sq +
        SQUARE(rot[1][2]) * alpha_sq * gamma_sq + 
        SQUARE(rot[2][2]) * alpha_sq * beta_sq;
    
    S = A * rot[0][2] * beta_sq  * gamma_sq +  
	B * rot[1][2] * alpha_sq * gamma_sq + 
        C * rot[2][2] * alpha_sq * beta_sq;
    S *= 2;
    
    T = SQUARE(A) * beta_sq  * gamma_sq + 
        SQUARE(B) * alpha_sq * gamma_sq + 
        SQUARE(C) * alpha_sq * beta_sq -
	  alpha_sq * beta_sq * gamma_sq;
	
    
    discriminant = SQUARE(S) - 4 * R * T;
    if (discriminant > 0)
    {
	z = (-S + sqrt(discriminant)) / (2 * R);
	return z;
    }
    else
    {
	printf("Negative Discriminant: no solutions\n");
	return 0;
    }
}







static void
compare_elevation(int32 cell, float64 x,float64 y, 
		  float64 real_z, float64 *extra_z)
{
    float64 z;
    
    /* real */
    printf("\nGCS %d (%lf,%lf,%lf)\n",cell,x,y,real_z);

    /* extrapolated */
    z = find_z_from_gcs_patch_water(gcs_get_tdb(cell, FALSE),x,y);

    printf("Ext %d (%lf,%lf,%lf)\n",cell,x,y,z);

    printf("diff = %lf \npercent error %6.3lf \%\n",
	   real_z-z,(real_z-z)/real_z*100.0);
    *extra_z = z;
}


main (argc,argv)
    int32 argc;
    char *argv[];
{
    char dbname[FILENAME_DIMENSION];
    int32 selection = 1;
    CTDB_STATUS ctdb_resident=CTDB_STATUS_READ_DB_FAIL;
    CTDB  *ctdb;
    int32 cell[100];
    int32 n_cell=100;
    int32 i, cell_id, in_list = 0;
    int32 mode = 0;

    {
	CMD_OPTION *options;
	uint32      sizeof_options = 0;
	CMD_TYPES **map;
	int32       leftover_argc;
	char *     *leftover_argv;

	/* parse terrain related options */
	wld_tdb_aggregate_options(&options, &sizeof_options, &map);
	
	cmd_aggregate(&options, &sizeof_options,
		      (CMD_OPTION *)&main_options, /* options array */
		      sizeof(main_options), /* size of options array */
		      &map);
	cmd_process_aggregate_options(argc, argv, /* args to main() */
		      &leftover_argc, &leftover_argv, /* unrecognized */
		      options, sizeof_options, map,
		      TRUE /* Verbose mode */
		      );
    }

    time_init(0);
    /* do this stuff */
    reader_init(0);

    wld_load_playbox(1, 0, NULL, NULL, main_options.data.value, 
		     NULL, NULL, NULL, NULL);
    wld_lookup_playbox_cells(&n_cell,cell);
    
    if ( n_cell > 1 )
    {
	printf("List of cells which define database extent: ");
	for (i=0;i<n_cell;i++)
	  printf("%06d, ",cell[i]);
	printf("\n");

	do
	{
	    printf("\nEnter cell id:  ");
	    scanf("%d",&cell_id);
	    
	    for (i=0; i < n_cell; i++)
	      if (cell[i] == cell_id)
	      {
		  in_list = 1;
		  break;
	      }

	    if (!in_list)
	      printf("Sorry, that cell is not defined for the database\n");
	} while (!in_list);
    }

    ctdb = (CTDB *)gcs_get_tdb(cell_id, FALSE);
    test_it();

    while (selection)
    {	
	printf("What would you like to run on cell %05d\n\n",cell_id);
	printf("\tSelect Option:\n");
	printf("\n\t\t\t\tELEVATION OR SOIL LOOKUP\n");
	printf("\t\t\t%2d. Check Elevation At Specific Point\n",1);
	printf("\t\t\t%2d. Look up Soil At Specific Point\n", 216);
	printf("\t\t\t%2d. Sample Elevation Across database\n",2);
	printf("\t\t\t%2d. Profile Elev Lookup Across database\n",3);

	printf("\n\t\t\t\tINTERVISIBILITY\n");
	printf("\t\t\t%2d. Check Intervis Ray\n",4);
	printf("\t\t\t%2d. Sample Intervis Across database\n",5);
	printf("\t\t\t%2d. Profile Intervis Across database\n",6);

	printf("\n\t\t\t\tTERRAIN PROFILE VECTOR\n");
	printf("\t\t\t%2d. Check Profile Intervis Vector\n",9);
	printf("\t\t\t%2d. Sample Profile Across database\n",10);

	printf("\n\t\t\t\tMAX. ELEVATION RATIO\n");	
	printf("\t\t\t%2d. Check Max Elevation Ratio Vector\n",36);
	printf("\t\t\t%2d. Sample Max Elevation Ratio Across database\n",11);

	printf("\n\t\t\t\tMAXIMUM ELEVATION LOOKUP\n");
	printf("\t\t\t%2d. Check Elevation At Specific Point\n",18);
	printf("\t\t\t%2d. Sample Elevation Across database\n",19);
	printf("\t\t\t%2d. Profile Elev Lookup Across database\n",20);

	printf("\n\t\t\t\tML ELEVATION LOOKUP\n");
	printf("\t\t\t%2d. Check Elevation At Specific Point\n",21);
	printf("\t\t\t%2d. Sample Elevation Across database\n",22);
	printf("\t\t\t%2d. Profile Elev Lookup Across database\n",23);

	printf("\n\t\t\t\tMES ELEVATION LOOKUP\n");
	printf("\t\t\t%2d. Check Elevation At Specific Point\n",24);
	printf("\t\t\t%2d. Sample Elevation Across database\n",25);
	printf("\t\t\t%2d. Profile Elev Lookup Across database\n",26);

	printf("\n\t\t\t\tHIGH GROUND INTERVIS\n");
	printf("\t\t\t%2d. Check Intervis Ray\n",27);
	printf("\t\t\t%2d. Sample Intervis Across database\n",28);
	printf("\t\t\t%2d. Profile Intervis Across database\n",29);

	printf("\n\t\t\t\tGROUND INTERSECTION\n");
	printf("\t\t\t%2d. Check Ground Intersection Ray\n",33);
	printf("\t\t\t%2d. Sample Ground Intersection Across database\n",8);
	printf("\t\t\t%2d. Profile Ground Intersection Across database\n",35);

	printf("\n\t\t\t\tOTHER\n");
	printf("\t\t\t%2d. Test Virtual Grid processing\n",7);
	printf("\t\t\t%2d. Find intervis raster\n",13);
	printf("\t\t\t%2d. List TIN triangles in a patch\n",14);
	printf("\t\t\t%2d. Ground correlation comparison \n",15);
	printf("\t\t\t%2d. Read and Test Database Correlation \n",16);
	printf("\t\t\t%2d. Change the Tide Level\n",100);
	printf("\t\t\t%2d. Change Ocean State\n",101);
	printf("\t\t\t%2d. Retrieve Ocean State\n",102);
	printf("\t\t\t%2d. Test Patch Water in extrapolation\n",103);
	printf("\t\t\t%2d. Test Patch Water in extrapolation, old ellipsoid model\n",104);
	printf("\t\t\t%2d. Test Patch Water in extrapolation, new ellipsoid model\n",105);
	printf("\t\t\t%2d. Test Patch Water in extrapolation, new ellipsoid model, function\n",106);
	 
	printf("\n\t\t\t\tMulti-Elevation Surfaces\n");
	printf("\t\t\t%2d. Find MES structure\n",201);
	printf("\t\t\t%2d. Get MES topology\n",202);	
	printf("\t\t\t%2d. Lookup triangles for a MES enclosure\n",203);
	printf("\t\t\t%2d. Find MES enclosure\n",204);	
	printf("\t\t\t%2d. Get Enclosure Elevation At Specific Point\n",205);
	printf("\t\t\t%2d. Place a vehicle in a mes\n",206);	
	printf("\t\t\t%2d. 2D MES structure + enclosure lookup\n",207);	
	printf("\t\t\t%2d. find_best_enclosure\n",208);	
	
	printf("\t\t\t%2d. Look up data for particular vertex\n",209);
	printf("\t\t\t%2d. Search for bridges\n",210);	

	printf("\t\t\t%2d. Test feasible connections\n",211);	
	printf("\t\t\t%2d. Get aperture info\n",212);	
	
	printf("\t\t\t%2d. Test conversion GCS to/from MES PO coords\n",213);
	printf("\t\t\t%2d. Look up data for particular triangle\n",214);
	printf("\t\t\t%2d. Check aperture triangle normals for a MES\n",215);
	printf("\t\t\t%2d. Lookup soil at a point\n",216);
	printf("\t\t\t%2d. Find Node in a MES\n",217);
	
	if (mode == 2)
	  printf("\t\t\t%2d. Select new cell \n",17);

	printf("\t\t\t%2d. Exit\n\n",0);
	printf("\n\t\t\t\tEnter Selection :");
	scanf("%d",&selection);

	if ((selection == 17) && (mode != 2))
	  selection = 0;

	switch (selection)
	{
	  case 0 :
	    break ;
	  case 1:
	    {
		CTDB_COORD2D pt;		
		float32 elev;
		float32 a,b;		
		int32 main_menu=FALSE;

		while (!main_menu)
		{
		    printf("Point <x> <y> :");
		    scanf("%f %f",&a, &b);
		    pt.x = a;
		    pt.y = b;	
		    elev = ctdb_lookup_elevation(ctdb,pt.x,pt.y);
		    printf("Elevation = %f\n",elev);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }	
	  case 2 :
	    sample_elevations_across_database(ctdb,CTDB_MAX_LAND_AND_WATER);
	    break;
	  case 3 :
	    profile_elevation_lookup(ctdb); 
	    break;
	  case 4 :
	    test_intervisibility_ray(ctdb);
	    break;
	  case 5 :
	    sample_intervis_across_database(ctdb);
	    break;
	  case 6:
	    profile_intervisibility(ctdb);
	    break;
	  case 7 :
	    test_virtual_grid_logic();
	    break;
	  case 8 :
	    sample_ground_across_database(ctdb);
	    break;
	  case 9 :
	    check_profile_vector(ctdb);
	    break;
	  case 10 :
	    sample_profile_across_database(ctdb);
	    break;
	  case 11 :
	    sample_mratio_across_database(ctdb);
	    break;
	  case 13:
	    run_raster(ctdb);
	    break;
	  case 14:
	    get_triangles(ctdb);
	    break;
	  case 15:
	    get_triangles(ctdb);
	    break;
	  case 16:
	    /* database_handler(ctdb,dbname,&ctdb_resident);*/
	    read_gcc_file_and_compare(argv[1]);
	    break;
	  case 17:
	    printf("List of cells which define database extent: ");
	    for (i=0;i<n_cell;i++)
	      printf("%06d, ",cell[i]);
	    printf("\nEnter cell id:  ");
	    scanf("%d",&cell_id);
	    ctdb = (CTDB *)gcs_get_tdb(cell_id, FALSE);
	    break;
	  case 18:
	    {
		float32 elev;
		float32 a,b;		
		int32 canopies;
		int32 main_menu=0;

		while (!main_menu)
		{
		    printf("Point <x> <y> :");
		    scanf("%f %f",&a, &b);
		    printf("With Canopies? (0/1) > ");
		    scanf("%d",&canopies);
		    elev = ctdb_lookup_max_elevation(ctdb,
						     (float64)a,
						     (float64)b,
						     canopies);
		    printf("Elevation = %f\n",elev);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		
		break;
	    }	
	  case 19:
	    {
		int32 canopies;
		printf("With Canopies? (0/1) > ");
		scanf("%d",&canopies);
		sample_max_elevations_across_database(ctdb,canopies);
		break;
	    }	
	  case 21:
	    {
		float64 x,y,r;
		int32 main_menu=0;
		
		while (!main_menu)
		{		    
		    printf("Point <x> <y> <reference elevation>:");
		    scanf("%lf %lf %lf",&x, &y, &r);
		    printf("Elevation = %lf\n",ctdb_lookup_elevation_ml(ctdb,x,y,r));
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 24:
	    {
		CTDB_ELEV_QUALIFIER qual;
		CTDB_ELEV_DATA elev[CTDB_MAX_ELEVS],*elevptr;
		float64 x,y;
		int32 i=0;
		int32 main_menu=0;
		char text[132];

		while (!main_menu)
		{
		    printf("Point <x> <y> :");
		    scanf("%lf %lf",&x, &y);
		    
		    qual = determine_elev_qualifier();
		    ctdb_lookup_elevation_mes(ctdb,x,y,qual,elev);
		    ctdb_sort_elev_data(elev);
		    for (elevptr=elev;elevptr->class!=CTDB_FC_ILLEGAL;
			 elevptr++ )
		    {
			ctdb_elev_data_to_string(elevptr,text);
			printf("%s material: %d\t",text, elevptr->material);
			printf("(%.2f, %.2f, %.2f)\n",elevptr->norm_vec[0],
			       elevptr->norm_vec[1],elevptr->norm_vec[2]);
		    }
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 27:
	    {
		float64 elev;
		float64 xs,ys,xe,ye,pt[XYZ];
		int32 main_menu=0;
		
		while (!main_menu)
		{		    
		    printf("Start <start_x> <start_y> <end_x> <end_y>:");
		    scanf("%lf %lf %lf %lf",&xs, &ys, &xe, &ye);
		    pt[Z] = ctdb_find_high_ground(ctdb,xs,ys,xe,ye,
						  CTDB_MAX_LAND_AND_WATER,pt);
		    printf("Highest Point ( %lf , %lf , %lf )\n",
			   pt[X],pt[Y],pt[Z]);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 28 :
	    sample_highest_across_database(ctdb);
	    break;
	  case 33:
	    {
		float64 elev;
		float64 xs,ys,zs,xe,ye,ze,pt[XYZ];
		int32 main_menu=0;
		CTDB_VEHICLE_LOCATION temp[1];
		int32 hit,temp2;
		
		while (!main_menu)
		{		    
		    printf("Start <start_x> <start_y> <start_z> <end_x> <end_y> <end_z>:");
		    scanf("%lf %lf %lf %lf %lf %lf",&xs, &ys, &zs, &xe, &ye, &ze);
		    hit = ctdb_find_ground_intersection(ctdb,xs,ys,zs,xe,ye,ze,
							pt,CTDB_USE_DEFAULT,0,temp,
							0,0,&temp2);
		    printf("Hit = %2d/%4x \t( %lf , %lf , %lf )\n",
			   hit,hit,pt[X],pt[Y],pt[Z]);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 36:
	    {
		float64 ratio;
		float64 s[XY],e[XY],z;
		int32 main_menu=0;
		
		while (!main_menu)
		{		    
		    printf("<start_x> <start_y> <start_z> <end_x> <end_y>:");
		    scanf("%lf %lf %lf %lf %lf %lf",&s[X], &s[Y], &z, 
			  &e[X], &e[Y]);
		    ratio = ctdb_find_max_ratio(ctdb,z,0.0,CTDB_USE_DEFAULT,
					      s,e);
		    printf("Ratio = %2lf\n",ratio);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	    
	  case 100 :
	  {
	      float32 level;
	      
	      printf("Enter the new tide level > ");
	      scanf("%f",&level);
	      
	      ctdb->tidal_zones[1].offset = level;
	      break;
	  } 
	  case 101 :
	  {
	      float64 x,y;
	      CTDB_WAVE_CHARS primary,secondary;
	      
	      printf("Enter the point to change <x,y> ");
	      scanf("%lf %lf",&x,&y);
	      primary.height    = 2;
	      primary.period    = 3;
	      primary.direction = 90;
	      primary.speed     = 10;
		  
	      secondary.height    = 1;
	      secondary.period    = 5;
	      secondary.direction = 270;
	      secondary.speed     = 5;

	      ctdb_ocean_point_write(ctdb,x,y,
				     CTDB_SURF_HEIGHT,5.0,
				     CTDB_TEMP,33.0,
				     CTDB_PRIMARY,&primary,
				     CTDB_SECONDARY,&secondary,
				     CTDB_TIDAL_HEIGHT,20.0,
				     CTDB_NULL);
	      break;
	  }                       
	  case 102:
	    {
		CTDB_OCEAN_QUALIFIER qual;
		CTDB_WAVE_CHARS wave;
		float64 out,x,y;
		char pick[132];
		int32 main_menu=0;
		
		printf("What is the point location <x,y>");
		scanf("%lf %lf",&x,&y);
		
		while (!main_menu)
		{
		    qual = determine_ocean_qualifier(pick);
		    if (!qual)
		      break;
		    
		    if ( qual == CTDB_PRIMARY || qual == CTDB_SECONDARY)
		    {
			ctdb_ocean_point_query(ctdb,x,y,qual,&wave);
			printf("%s\n",pick);
			printf("height\t%9.2lf\n",wave.height);
			printf("period\t%9.2lf\n",wave.period);
			printf("direction\t%9.2lf\n",wave.direction);
			printf("speed\t%9.2lf\n",wave.speed);
		    }
		    else
		    {
			ctdb_ocean_point_query(ctdb,x,y,qual,&out);
			printf("%s \t=\t%lf\n",pick, out);
		    }
		}
		break;
	    }
	  case 103:
	    {
		int32           main_menu=0;
		float64         x0,y0,z0;
		static CTDB    *lastctdb=NULL;
		COORD_TCC_PTR   tcc=NULL;
		float64         x, y, real_z, extra_z; 
		int32           cell;
		
#ifdef NEVER
		if (( ctdb->gcs_mode != CTDB_MODE_SINGLE_CELL ) &&
		    ( ctdb->gcs_mode != CTDB_MODE_MULTI_CELL ))
		{
		    printf("database is not GCS\n");
			break;
		}
#endif
		coord_set_tcc_awareness(TRUE);
		tcc = coord_define_tcc(COORD_UTM_NE,
				       ctdb->origin_northing,
				       ctdb->origin_easting,
				       ctdb->origin_zone_number,
				       ctdb->origin_zone_letter,
				       ctdb->datum,
				       ctdb->max_x - ctdb->min_x,
				       ctdb->max_y - ctdb->min_y);
		
		while (!main_menu)
		{
		    printf("Enter the point to change (TCC) <x,y> ");
		    scanf("%lf %lf",&x0,&y0);
		    
		    
		    get_gcs_xy_from_tcc(ctdb,tcc,x0,y0,z0,
					&cell,&x,&y,&real_z);
		    
		    printf("TCC (%lf,%lf,0.0) -> GCS (%d,%lf,%lf,%lf)\n",
			   x0,y0,cell,x,y,real_z);
		    
		    extra_z = find_z_from_gcs_patch_water(ctdb, x, y);
		    printf("Extra_z calculated: %f\n", extra_z);
		    printf("diff = %lf \npercent error %6.3lf \%\n",
			   real_z-extra_z,(real_z-extra_z)/real_z*100.0);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 104:
	    {
		float64 x0,y0,z,zamsl;
		int32 mem,patch_number;
		CTDB_PATCH_GROUP *pg;
		int32 water_index;
		CTDB_WATER_CHARS *water_ptr;
		int32 main_menu=0;
#ifdef NEVER		
		if ( !ctdb->is_curved )
		{
		    printf("database is not GCS\n");
		    break;
		}
#endif       
		while (!main_menu)
		{
		    printf("Enter the point to change <x,y> ");
		    scanf("%lf %lf",&x0,&y0);
		    patch_number = CTDB_METERS_TO_PATCH_NUMBER(ctdb,x0,y0);
		    pg = ctdb->patch_groups+PG_NUMBER(patch_number);
		    mem = PG_MEMBER(patch_number);
		    
		    if ( CTDB_HAS_WATER_PATCH( pg, mem )  || 1)
		    {
			/* If the database is curved then the flat patch water 
			 * has increasing error across the 500 meter patch
			 * (between 4 and 6 meters at 50 km from the point
			 * of tangency of the database to the ellipsoid).
			 */
			static float64 rot[3][3],dd,rr,rr4,irr_over2;
			static CTDB *lastctdb=NULL;
			float64 aa[XYZ],RR,ss,tt,offset[XYZ],bb[XYZ],x,y;
			float64 s_x, s_y, s_c;
			float64 t_x2, t_y2, t_xy, t_x, t_y, t_c;
			float64 discriminant;
			float64 ss_old;
			float64 tt_old;
			float64 alpha,gamma;
			static float64 x_off, y_off;

			if ( lastctdb != ctdb )
			{
			    /* $$$ future have GCS store these values for 
			     * easy retrieval 
			     */
			    alpha = sqrt(sqrt(WGS84_MAJOR_QUAD));
			    gamma = sqrt(WGS84_MAJOR_SQ_X_MINOR_SQ)/alpha;
		    
			    (void)gcs_get_origin_in_gcc(ctdb->cell_id,offset);
			    (void)gcs_get_rot_mat(ctdb->cell_id, rot);
			    gcs_get_origin_offsets(ctdb->cell_id, &x_off,
						   &y_off);

			    vmat3_transpose64(rot, rot);


			    bb[X]=rot[2][0]*WGS84_MAJOR_SQ_X_MINOR_SQ;
			    bb[Y]=rot[2][1]*WGS84_MAJOR_SQ_X_MINOR_SQ;
			    bb[Z]=rot[2][2]*WGS84_MAJOR_QUAD;
			    dd = WGS84_MAJOR_SQ_X_MINOR_SQ * 
			      WGS84_MAJOR_SQUARE;
			    rr=(rot[2][0]*rot[2][0]+rot[2][1]*rot[2][1])
			      *WGS84_MAJOR_SQ_X_MINOR_SQ + 
				rot[2][2]*rot[2][2]*WGS84_MAJOR_QUAD;
			    rr4 = rr*4.0;
			    irr_over2 = 1/rr*0.5;
			    s_x = 2*((rot[2][0]*rot[0][0]+rot[2][1]*rot[0][1])
				     *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[2][2]*rot[0][2]*WGS84_MAJOR_QUAD);
			    s_y = 2*((rot[2][0]*rot[1][0]+rot[2][1]*rot[1][1])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[2][2]*rot[1][2]*WGS84_MAJOR_QUAD);
			    s_c = 2*((rot[2][0]*offset[X]+rot[2][1]*offset[Y])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[2][2]*offset[Z]*WGS84_MAJOR_QUAD);
			    t_x2= (rot[0][0]*rot[0][0]+rot[0][1]*rot[0][1])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[0][2]*rot[0][2]*WGS84_MAJOR_QUAD;
			    t_y2= (rot[1][0]*rot[1][0]+rot[1][1]*rot[1][1])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[1][2]*rot[1][2]*WGS84_MAJOR_QUAD;
			    t_xy= 2*((rot[0][0]*rot[1][0]+rot[0][1]*rot[1][1])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[0][2]*rot[1][2]*WGS84_MAJOR_QUAD);
			    t_x = 2*((rot[0][0]*offset[X]+rot[0][1]*offset[Y])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[0][2]*offset[Z]*WGS84_MAJOR_QUAD);
			    t_y = 2*((rot[1][0]*offset[X]+rot[1][1]*offset[Y])
				      *WGS84_MAJOR_SQ_X_MINOR_SQ
				     +rot[1][2]*offset[Z]*WGS84_MAJOR_QUAD);
			    t_c = (offset[X]*offset[X]+offset[Y]*offset[Y])
			              *WGS84_MAJOR_SQ_X_MINOR_SQ
					+offset[Z]*offset[Z]*WGS84_MAJOR_QUAD
					-dd;
			    lastctdb = ctdb;
			}
#ifdef NEVER
			x = data.norm_x * ctdb->incr;
			y = data.norm_y * ctdb->incr;
#endif
/*			x = x0 - ctdb->origin_x;
			y = y0 - ctdb->origin_y; */
			x = x0 - x_off;
			y = y0 - y_off;
			
/*#ifdef NEVER */
			aa[X] = (rot[0][0]*x + rot[1][0]*y + offset[X]);
			aa[Y] = (rot[0][1]*x + rot[1][1]*y + offset[Y]);
			aa[Z] = (rot[0][2]*x + rot[1][2]*y + offset[Z]);
			ss_old = 2.0*(aa[X]*bb[X]+aa[Y]*bb[Y]+aa[Z]*bb[Z]);
			tt_old = WGS84_MAJOR_SQ_X_MINOR_SQ * 
			  (aa[X] * aa[X] +aa[Y]*aa[Y]) + 
			    WGS84_MAJOR_QUAD*aa[Z]*aa[Z] - dd;
/*#endif */
			ss = s_x*x + s_y*y + s_c;
			tt = (t_x2*x+t_x)*x + (t_y2*y+t_y)*y + t_xy*x*y + t_c;
			discriminant =  ss*ss - rr4*tt;
			discriminant =  ss_old*ss_old - rr4*tt_old;
			/* Negative discriminant corresponds to dropping a 
                         * vertical ray which misses the earth completely;
			 * ie, the point (x,y) was too far away from the center
			 * of the cell.  So, we replace it with a ray which
			 * just grazes the surface of the earth, as far from
			 * the center of the cell as possible.
			 */
			if(discriminant < 0)
			  discriminant = 0;
			RR = sqrt( discriminant );
			z = (-ss+RR)*irr_over2;   
		    }
		    zamsl = ctdb_lookup_elevation_amsl(ctdb,x0,y0);
		    printf("z = %lf\tzamsl= %lf\n",z);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }

	  case 106:
	    {
		float64 x0,y0,z;
		int32 main_menu=0;
		float64 z1;

		while (!main_menu)
		{
		    
		    printf("Enter the point to change <x,y> ");
		    scanf("%lf %lf",&x0,&y0);
		    
		    z1 = find_z_from_gcs_patch_water_ellipsoid(ctdb, x0, y0);
		    z = find_z_from_gcs_patch_water_ellipsoid2(ctdb, x0, y0);
		    
		    printf("z is %f   new: %f\n", z1, z);
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 105:
	    {
		float64 x0,y0,z;
		int32 main_menu=0;

		while (!main_menu)
		{
		    static float64 rot1[3][3], rot[3][3];
		    static CTDB *lastctdb=NULL;
		    float64 origin[XYZ], x_off, y_off;

		    printf("Enter the point to change <x,y> ");
		    scanf("%lf %lf",&x0,&y0);
		    
		    if ( lastctdb != ctdb )
		    {
			float64 A, B, C;
			float64 R, S, T, discriminant;

			float64 alpha_sq = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS;
			float64 gamma_sq = WGS84_MINOR_AXIS*WGS84_MINOR_AXIS;
			int32 i, j;


			/* $$$ future have GCS store these values for 
			 * easy retrieval 
			 */
			(void)gcs_get_origin_in_gcc(ctdb->cell_id,origin);
			(void)gcs_get_rot_mat(ctdb->cell_id, rot1);

			for (i=0; i < 3; i++)
			  for (j=0; j < 3; j++)
/*			    rot[i][j] = rot1[j][i]; */
			    rot[i][j] = rot1[i][j];

			gcs_get_origin_offsets(ctdb->cell_id, &x_off,
					       &y_off);

			x0 += x_off;
			y0 += y_off;

			A = rot[0][0]*x0 + rot[0][1]*y0 + origin[X];
			B = rot[1][0]*x0 + rot[1][1]*y0 + origin[Y];
			C = rot[2][0]*x0 + rot[2][1]*y0 + origin[Z];
			
			R = gamma_sq * (SQUARE(rot[0][2])+SQUARE(rot[1][2])) + 
			    alpha_sq * SQUARE(rot[2][2]);

			S = gamma_sq * ((A * rot[0][2]) + (B * rot[1][2])) + 
			    alpha_sq * C * rot[2][2];
			S *= 2;
			
			T = gamma_sq * (SQUARE(A) + SQUARE(B)) + 
			    alpha_sq * SQUARE(C) -
			    alpha_sq * gamma_sq;

			discriminant = SQUARE(S) - 4 * R * T;
			
			if (discriminant > 0)
			{
			    z = (-S + sqrt(discriminant)) / (2 * R);
			    printf("z is %f\n", z);
			}
			else
			{
			    printf("Negative Discriminant: no solutions\n");
			}
			printf("RETURN to MAIN MENU (0/1)");
			scanf("%d",&main_menu);
		    }

		}
		break;
	    }
	  case 201:
	    {
		float64 x, y, z, threshold;
		uint16 mes_id;
		int32 i;
		CTDB *out_ctdb;
		CTDB_MES_NODE_KEY  mes_key;

		printf("Enter the center of your search <x, y, z>: ");
		scanf("%lf %lf %lf",&x, &y, &z);
		printf("Enter the radius of your search: ");
		scanf("%lf",&threshold);
		
		if (ctdb_mes_find_mes(ctdb, 
				      x, y, z, 
				      threshold, 
				      &out_ctdb,
				      &mes_key))
		{
		    int32 choice;
		    printf("Found mes number %d in cell %d\n", 
			   CTDB_MES_NODE_KEY_TO_MES_ID(mes_key),
			   ctdb_get_cell_id(out_ctdb));

		    printf("\nGet information about this MES (0/1)? ");
		    scanf("%d", &choice);
		    if (choice == 1)
		    {
			float64  roof_vertices[100][XYZ];
			int32    num_roof_vertices = 100;
			float64  floor_vertices[100][XYZ];
			int32    num_floor_vertices = 100;
			CTDB_MES_TYPE type;
			if (!ctdb_mes_get_mes(ctdb, 
					      mes_key, 
					      &type,
					      &num_roof_vertices, 
					      roof_vertices,
					      &num_floor_vertices, 
					      floor_vertices))
			  printf("driver cannot handle such a complicated MES\n");
			else
			{

			    switch (type)
			    {
			      case CTDB_MES_UNKNOWN:
				printf("Unknown MES\n");
				break;
			      case CTDB_MES_BUILDING:
				printf("MES Building\n");
				break;
			      case CTDB_MES_CAVE:
				printf("MES Cave\n");
				break;
			      case CTDB_MES_TUNNEL:
				printf("MES Tunnel\n");
				break;
			      case CTDB_MES_BRIDGE:
				printf("MES Bridge\n");
				break;
			      default: 
				printf("Bad MES type!\n");
				break;
			    }
			    printf("Roofline:\n");
			    for (i=0; i < num_roof_vertices; i++)
			    {
			        printf("\t<%f,%f,%f>\n", 
				       roof_vertices[i][X], 
				       roof_vertices[i][Y], 
				       roof_vertices[i][Z]);
			    }
			    printf("Floorline:\n");
			    for (i=0; i < num_floor_vertices; i++)
			    {
			        printf("\t<%f,%f,%f>\n", 
				       floor_vertices[i][X], 
				       floor_vertices[i][Y], 
				       floor_vertices[i][Z]);
			    }
			}
		    }
		    
		}
		else
		  printf("No MES found!\n");
		break;
	    }
	  case 202:
	    {
		uint16             mes_id;
		CTDB_MES_NODE     *node = NULL;
		CTDB_MES_EDGE     *edge = NULL;
		CTDB_MES_NODE     *(nodes[2]);
		CTDB_MES_EDGE     *(edges[40]);
		float64            vertices[200][XYZ];
		int32              done = 0;
		int32              i;
		CTDB_MES_VOLUME   *mes;
		CTDB_MES_TEMPLATE *templ;
		CTDB_MES_NODE_KEY  mes_key;
		CTDB_MES_NODE_KEY  node_key;
		CTDB_MES_EDGE_KEY  edge_key;
		CTDB_MES_NODE_KEY  node_keys[3];
		CTDB_MES_EDGE_KEY  edge_keys[40];

		do
		{
		    printf("Enter the mes_id: ");
		    scanf("%hd",&mes_id);
		    if ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes))
		      printf("Bad MES id entered\n");
		    else
		      done = 1;
		} while (!done);

		done = 0;
		CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key);
		mes = &(ctdb->mes_volumes[mes_id]);
		templ = mes->template;
		ctdb_mes_get_topology(ctdb, mes_key, &node_key);

		printf("The world exterior to the MES structure is node %d\n",
		       CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_key));
		do 
		{
		    int32 num_edges = 40;
		    int32 num_vertices = 200;
		    int32 node_id;
		    int32 edge_id;
		    int16 visible, open, material;
		    float64 ap_thickness[2];
		    float64 thick_direction[2][XYZ];
		
		    printf("\n\t1. Get edges from a node\n");
		    printf("\t2. Get info on a node\n");
		    printf("\t3. Get nodes from an edge\n"); 
		    printf("\t4. Get info on an edge\n");
		    printf("\t5. Quit\n");
		    printf("\tEnter selection: ");
		    scanf("%d", &selection);
		    switch(selection)
		    {
		      case 1:
			printf("\t  Enter a node id: ");
			scanf("%d", &node_id);
			if ((node_id < 0) ||
			    (node_id >= templ->num_enclosures))
			{
			    printf("Invalid node id\n");
			    break;
			}
			CTDB_MES_IDS_TO_NODE_KEY(mes_id, node_id, node_key);
			if (!ctdb_mes_get_connected_edges(ctdb, 
							  node_key,
							  &num_edges,
							  edge_keys))
			{
			    printf("Driver cannot load %s", 
				   "such a complicated topology\n");
			    break;
			}
			printf("\t  Current node %d %s",
			       node_id, 
			       "connected to the following edges: ");

			for (i=0; i < num_edges; i++)
			{
			    printf("%d ", CTDB_MES_EDGE_KEY_TO_APERTURE_ID(edge_keys[i]));
			}
			printf("\n");
			break;

		      case 2:
			printf("\t  Enter a node id: ");
			scanf("%d", &node_id);
			if ((node_id < 0) || 
			    (node_id >= templ->num_enclosures))
			  {
			    printf("Invalid node id\n");
			    break;
			  }
			CTDB_MES_IDS_TO_NODE_KEY(mes_id, node_id, node_key);
#if 0
			/* $$$ */
			if (!ctdb_mes_get_node_info(ctdb, 
						    node_key, 
						    &num_vertices, 
						    vertices))
			    {
			      printf("Cannot handle so complicated a node\n");
			      break;
			    }
			
			for (i=0; i < num_vertices; i++)
			  printf("\t<%f,%f,%f>\n", 
				 vertices[i][X], 
				 vertices[i][Y], 
				 vertices[i][Z]);
#endif
			break;

		      case 3:
			printf("\t  Enter an edge id: ");
			scanf("%d", &edge_id);
			if ((edge_id < 0) || 
			    (edge_id >= templ->num_apertures))
			{
			    printf("Invalid edge id\n");
			    break;
			}
			CTDB_MES_IDS_TO_EDGE_KEY(mes_id, edge_id, edge_key);
			ctdb_mes_get_connected_nodes(ctdb, 
						     edge_key, 
						     node_keys);
			printf("\t  Current edge %d %s %d %d\n", 
			      node_id,
			      "connected to the following nodes:",
			      CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_keys[0]),
			      CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node_keys[1]));
			break;
		      case 4:
			printf("\t  Enter an edge id: ");
			scanf("%d", &edge_id);
			if ((edge_id < 0) || 
			    (edge_id >= templ->num_apertures))
			{
			    printf("Invalid edge id\n");
			    break;
			}

			CTDB_MES_IDS_TO_EDGE_KEY(mes_id, edge_id, edge_key);

			if (!ctdb_mes_get_edge_info(ctdb,
						    edge_key, 
						    &visible, 
						    &open, 
						    &material,
						    ap_thickness,
						    thick_direction,
						    &num_vertices, 
						    vertices))
			{
			    printf("Cannot handle such a complicated edge\n");
			    break;
			}
			printf("Edge- Visible:%d, Open:%d Material:%d\n",
			       visible, open, material);

			for (i=0; i < 2; i++)
			  printf("Aperture is %f meters thick in <%f, %f, %f>"\
				 " direction\n", ap_thickness[i],
				 thick_direction[i][X],thick_direction[i][Y],
				 thick_direction[i][Z]);

			for (i=0; i < num_vertices; i++)
			  printf("\t<%f,%f,%f>\n",
				 vertices[i][X], 
				 vertices[i][Y], 
				 vertices[i][Z]);
			break;
		      case 5:
			done = 1;
			break;
		      default:
			printf("Invalid selection\n");
			break;
		    }
		} while (!done);
		break;
	    }
	  case 203:
	    {
	        /* enclosure_id has to be permitted the value -1 */
		int32              mes_id, enclosure_id, choice;
		float64   triangle[3][XYZ], normal[XYZ];
		int32     mes_triangle[3][XYZ];
		CTDB_MES_VOLUME   *mes;
		CTDB_MES_TEMPLATE *templ;
		CTDB_MES_ENCLOSURE *enclosure;
		CTDB_MES_TRIANGLE *tri;
		int32 start_id, end_id, j, k;

		printf("Enter the index of your MES: ");
		scanf("%d",&mes_id);
		
		if ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes))
		{
		    printf("Invalid mes id\n");
		    break;
		}

		printf("Enter the index of your enclosure (-1 for all): ");
		scanf("%d",&enclosure_id);
		mes = &(ctdb->mes_volumes[mes_id]);
		templ = mes->template;
		if (enclosure_id>=templ->num_enclosures)
		{
		    printf("Invalid enclosure id\n");
		    break;
		}
		
		printf("Do you want: 1. Normals and Materials   2. Vertices    3. Both   4. Vertex ids: ");
		scanf("%d",&choice);
		if ((choice < 1) || (choice > 4))
		{
		    printf("Invalid choice\n");
		    break;
		}
		
		if (enclosure_id == -1)
		{
		    start_id = 0;
		    end_id = templ->num_enclosures;
		}
		else
		  start_id = end_id = enclosure_id;

		for (i=start_id; i <= end_id; i++)
		{
		    printf("Enclosure %d\n", i);
		    printf("(Triangle#) ");
		    if ((choice == 1) || (choice == 3))
		      printf("<Normal> [Material] ");
		    if ((choice ==2) || (choice == 3))
		      printf("<Vertices>");
		    if (choice == 4)
		      printf("<Vertex IDS>");
		    printf("\n");

		    enclosure = &templ->all_enclosures[i];
		    for (j=enclosure->first_wall_triangle_id;
			 j < enclosure->first_wall_triangle_id + enclosure->num_triangles;
			 j++)
		    {
			tri = &templ->all_triangles[j];
			
			printf("( %d )", j);
			if ((choice == 1) || (choice == 3))
			{
			    ctdb_mes_get_triangle_normal(ctdb, mes_id, j, normal);
			    printf("<%9.3f,%9.3f,%9.3f>   ",
				   normal[X], normal[Y], normal[Z]);
			    printf("[ %d ]\n", tri->material);
			}
			if ((choice == 2) || (choice == 3))
			{
			    for (k=0; k < 3; k++)
			    {
				ctdb_mes_to_gcscs(mes,
						  templ->all_vertices[tri->vertex_ids[k]][X],
						  templ->all_vertices[tri->vertex_ids[k]][Y],
						  templ->all_vertices[tri->vertex_ids[k]][Z],
						  ctdb_get_cell_id(ctdb),
						  &triangle[k][X],
						  &triangle[k][Y],
						  &triangle[k][Z]);
			    }
			    printf("<%9.3f,%9.3f,%9.3f> <%9.3f,%9.3f,%9.3f> <%9.3f,%9.3f,%9.3f>\n",
				   triangle[0][X],triangle[0][Y],triangle[0][Z],
				   triangle[1][X],triangle[1][Y],triangle[1][Z],
				   triangle[2][X],triangle[2][Y],triangle[2][Z]
				   );
			}
			if (choice == 4)
			  printf("<%d %d %d>\n",
				 tri->vertex_ids[0], tri->vertex_ids[1],
				 tri->vertex_ids[2]);
		    }
		}
		break;
	    }
	  case 204:
	    {
		float64            x, y, z;
		CTDB_MES_NODE_KEY  enclosure_key;
		CTDB_MES_NODE_KEY  mes_key;
		int32              i;
		uint16             mes_id;

		printf("Enter your point <x, y, z>: ");
		scanf("%lf %lf %lf",&x, &y, &z);

		printf("Enter the mes id: ");
		scanf("%hd",&mes_id);
		
		if ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes))
		{
		    printf("Invalid mes_id\n");
		    break;
		}
		CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key);
		if (ctdb_mes_find_enclosure(ctdb,
					    mes_key,
					    x, y, z,
					    &enclosure_key))
		  printf("In enclosure %d for mes %d\n", 
			 CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure_key), 
			 CTDB_MES_NODE_KEY_TO_MES_ID(enclosure_key));
		break;
	    }
	  case 205:
	    {
		CTDB_ELEV_DATA      elevations[3];
		float64             x,y;		
		int32               i,main_menu=FALSE;
                uint16              mes_id;
                uint16              enclosure_id;
                float64             maximum_angle;
                int32               num_elevations, answer;
		CTDB_MES_NODE_KEY   mes_key;

		CTDB_ELEV_QUALIFIER qualifier;

		while (!main_menu)
		{
		    CTDB_ELEV_DATA *elevptr;
		    printf("Point <x> <y> :");
		    scanf("%lf %lf",&x, &y);
		    printf("IDs <mes> <enclosure> :");
		    scanf("%hd %hd",&mes_id, &enclosure_id);
		    printf("Maximum Angle (deg) :");
		    scanf("%lf",&maximum_angle);
   
                    printf("what is your qualifier?\n");

                    printf("\t\t%2d. ALL_SURFACES\n",0);
                    printf("\t\t%2d. CTDB_ONLY_POSITIVE_SURFACES\n",1);
                    printf("\t\t%2d. CTDB_ONLY_NEGATIVE_SURFACES\n",2);
                    scanf("%d",&answer);
                    qualifier = (answer<<25);
		    CTDB_MES_IDS_TO_NODE_KEY( mes_id, enclosure_id, mes_key );
                    num_elevations = ctdb_lookup_enclosure_elevation
                                       ( ctdb, mes_key, x, y,
                                         maximum_angle, qualifier, 
					 elevations );

                    printf ( "Number of Elevations = %d\n", num_elevations );
		    for (elevptr=elevations;elevptr->class != CTDB_FC_ILLEGAL ;elevptr++)
		    {
			char text[80];
			
			ctdb_elev_data_to_string(elevptr,text);
			printf("%s material: %d\t",text, elevptr->material);
			printf("(%.2f, %.2f, %.2f)\n",elevptr->norm_vec[0],
			       elevptr->norm_vec[1],elevptr->norm_vec[2]);
		    }
		    
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 206:
	    {
		uint16             mes_id, enclosure_id;
		float64            x, y, z;
		float64            rotation[3][3];
		int32              soil;
		CTDB_MES_NODE_KEY  mes_key;
		float64            heading;
		float64            direction[XY];
		printf("Enter your point <x, y>: ");
		scanf("%lf %lf",&x, &y);

		printf("Enter the mes id: ");
		scanf("%hd",&mes_id);

		printf("Enter the enclosure id: ");
		scanf("%hd", &enclosure_id);
		
		printf("Enter heading (degrees cw from north): ");
		scanf("%lf", &heading);
		
		direction[X] = sin (heading * 0.017453292519943295);
		direction[Y] = cos (heading * 0.017453292519943295);

		CTDB_MES_IDS_TO_NODE_KEY(mes_id, enclosure_id, mes_key);
		ctdb_place_vehicle_ml(ctdb, x, y, 2.0, 1.0,
				      50.0, direction[X], direction[Y], &z,
				      rotation, &soil,
				      &mes_key);
		printf("Placed vehicle at %f %f %f\n", x, y, z);
		printf("  on soil %d\n", soil);
		printf(" with rotation:   %f   %f   %f\n", rotation[0][0],
		       rotation[0][1], rotation[0][2]);
		printf("                  %f   %f   %f\n", rotation[1][0],
		       rotation[1][1], rotation[1][2]);
		printf("                  %f   %f   %f\n", rotation[2][0],
		       rotation[2][1], rotation[2][2]);

		break;
	    }
	  case 207:
	    {
#define SIZEOF_ARRAY 50
		CTDB_MES_NODE_KEY  mes_keys[SIZEOF_ARRAY];
		CTDB_MES_NODE_KEY  enclosure_keys[SIZEOF_ARRAY];
		int32              mes_cnt, enclosure_cnt;
		int32              main_menu = 0;
		int32              i, size_mes_volumes = 50; 
		uint16             mes_id;
		CTDB_MES_NODE_KEY  mes_key;
		float64  x, y;

		while (!main_menu)
		{
		    printf("Enter your point <x, y>: ");
		    scanf("%lf %lf",&x, &y);
		      
		    mes_cnt = ctdb_mes_find_mes_from_xy(ctdb, x, y, 
							SIZEOF_ARRAY, 
							mes_keys);
		    
		    if (!mes_cnt)
		      printf("No MES found\n");
		    else
		    {
			printf("Found %d MES structures: \n", mes_cnt);
			for (i=0; i < mes_cnt; i++)
			  printf("\tMes id %hd\n",
				 CTDB_MES_EDGE_KEY_TO_MES_ID(mes_keys[i]));

			printf("Enter your choice of MES: ");
			scanf("%hd",&mes_id);
			CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key);
			
			enclosure_cnt = 
			  ctdb_mes_find_enclosures_from_xy(ctdb,
							   mes_key,
							   x, y,
							   SIZEOF_ARRAY, 
							   enclosure_keys);
			printf("Found %d Enclosures: \n", enclosure_cnt);
			for (i=0; i < enclosure_cnt; i++)
			  printf("\tEnclosure id %hd\n", 
				 CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure_keys[i]));
		    }
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 208:
	    {
		uint16   enclosure_id;
		int32    main_menu = 0;
		int32    i; 
		uint16   mes_id;
		float64  x, y, z;
		float64  src[XYZ], dest[XYZ];
		float64  actual_dest[XYZ];

		while (!main_menu)
		{
		    CTDB_MES_NODE_KEY  old_key;
		    CTDB_MES_NODE_KEY  new_mes_key;

		    printf("Enter your point <x, y, z>: ");
		    scanf("%lf %lf %lf",&src[X], &src[Y], &src[Z]);

		    printf("Enter your MES id: ");
		    scanf("%hd",&mes_id);

		    printf("Enter your enclosure id: ");
		    scanf("%hd",&enclosure_id);
		    CTDB_MES_IDS_TO_NODE_KEY(mes_id, enclosure_id, old_key);

		    printf("Enter your desired destination <x, y>: ");
		    scanf("%lf %lf", &dest[X], &dest[Y]);

		    ctdb_find_best_enclosure(ctdb, old_key, src, 
					     dest, 2.0, &new_mes_key,
					     actual_dest);
		    printf("Best Approximation: %f %f %f\n",
			   actual_dest[X], actual_dest[Y], actual_dest[Z]);
		    printf("In mes %hd   enclosure %hd\n", 
			   CTDB_MES_NODE_KEY_TO_MES_ID(new_mes_key),
			   CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(new_mes_key));
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}	
		break;
	    }


	  case 209:
	    {
		int32              main_menu = 0;
		int32              i; 
		uint16             mes_id;
		int32              vertex_id;
		CTDB_MES_TEMPLATE *temp;
		float64            x, y, z;
		CTDB_MES_VOLUME   *mes;

		while (!main_menu)
		{
		    printf("Enter your MES id: ");
		    scanf("%hd",&mes_id);

		    printf("Enter your vertex_id: ");
		    scanf("%d",&vertex_id);
		    mes = &ctdb->mes_volumes[mes_id];
		    temp = ctdb->mes_volumes[mes_id].template;
		    
		    ctdb_mes_to_gcscs(mes, temp->all_vertices[vertex_id][X],
				      temp->all_vertices[vertex_id][Y],
				      temp->all_vertices[vertex_id][Z],
				      ctdb_get_cell_id(ctdb),
				      &x, &y, &z);
		    printf("Vertex is %f %f %f\n", x, y, z);
		    
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);

		}
		break;
	    }
	  case 210:
	    {
		CTDB_ABSTRACT_DATA    *data;
		CTDB_SEARCH_SPACE_PTR  space;
		int32                  vertex_count;
 		float32               *vertex; 
		float32               *vertices;  /* -- x, y, x, y, ... */
		
		
		space = ctdb_create_search_space(ctdb,
						 ctdb->min_x, ctdb->min_y,
						 ctdb->max_x, ctdb->max_y,
						 FALSE
						 );					 
		while (vertex_count = ctdb_next_abstract(space,
							   CTDB_ABSTRACT_BRIDGE,
							   &vertices,
							   &data, NULL))
		{
		    for (i=0; i<vertex_count; ++i)
		    {
			vertex = &( vertices[XY * i] );
			printf ("vertex #%d = < %9.3f, %9.3f, %9.3f >.\n",
				vertex[X], vertex[Y], vertex[Z]);
		    }
		}
		break;
	    }
	    
	  case 211:
	    {
		int32              main_menu = 0;
		int32              i; 
		uint16             mes_id;
		float64            fall, climb;
		int32              num_connections;
		CTDB_MES_NODE     *node;
		CTDB_MES_NODE_KEY  mes_key;
		CTDB_MES_NODE_KEY  node_key;
		CTDB_MES_NODE_KEY  cnode_keys[50];
		CTDB_MES_EDGE_KEY  cedge_keys[50];


		while (!main_menu)
		{
		    int32 done = 0;

		    printf("Enter your MES id: ");
		    scanf("%hd",&mes_id);
		    CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key);

		    printf("Enter your maximum climb: ");
		    scanf("%lf",&climb);
		    printf("Enter your maximum fall distance: ");
		    scanf("%lf",&fall);

		    ctdb_mes_get_topology(ctdb, mes_key, &node_key);
		    		    
		    while (!done)
		    {
			num_connections = 50;
			ctdb_mes_get_connections(ctdb, node_key,
						 climb, fall, &num_connections,
						 cedge_keys, cnode_keys);
			printf("Found %d connected nodes\n", num_connections);
			for (i=0; i < num_connections; i++)
			{
			    printf("Through edge %d to node %d\n",
				   CTDB_MES_EDGE_KEY_TO_APERTURE_ID(cedge_keys[i]),
				   CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(cnode_keys[i]));

			}
			printf("Continue 0/1: ");
			scanf("%d", &done);
			done = !done;
			if (!done)
			{
			    int32 node_num;
			    
			    printf("Select node to go to: ");
			    scanf("%d", &node_num);
			    CTDB_MES_IDS_TO_NODE_KEY(mes_id, 
						     node_num, 
						     node_key);
			}
		    }
		    printf("Return to main menu? 0/1\n");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 212:
	    {
		int32              main_menu = 0;
		int32              i, vtx_id;
		CTDB_MES_EDGE_KEY  aperture_key;
		CTDB_MES_NODE_KEY  enclosure_keys[2];
		float64            x, y, z;
		uint16             mes_id;
		uint16             aperture_id;
                CTDB_MES_APERTURE *ap;
                CTDB_MES_TEMPLATE *templ;

		while (!main_menu)
		{
		    printf("Enter your MES id: ");
		    scanf("%hd",&mes_id);
		    printf("Enter your Aperture id: ");
		    scanf("%hd",&aperture_id);
		    CTDB_MES_IDS_TO_EDGE_KEY(mes_id,
					     aperture_id,
					     aperture_key);

                    templ = ctdb->mes_volumes[mes_id].template;
                    ap = &templ->all_apertures[aperture_id];


 
		    if (!ctdb_mes_get_aperture_centroid(ctdb,
							aperture_key, 
							&x, &y, &z))
		      printf("Error getting centroid\n");
		    else
		      printf("Aperture centroid is located: %f %f %f\n", 
			     x,y,z);

                    printf("Aperture outline: \n");
                    for (i=0; i < ap->outline.num_vertices; i++)
                    {
                        vtx_id = ap->outline.vertex_ids[i];
                        ctdb_mes_to_gcscs(&(ctdb->mes_volumes[mes_id]),
                                          templ->all_vertices[vtx_id][X],
                                          templ->all_vertices[vtx_id][Y],
                                          templ->all_vertices[vtx_id][Z],
                                          ctdb_get_cell_id(ctdb),
                                          &x, &y, &z);
                        printf("<%f %f %f>   ", x, y, z);
                    }
                    printf("\n");

		    if (!ctdb_mes_get_aperture_enclosures(ctdb, 
							  aperture_key,
							  enclosure_keys,
							  NULL))
		      printf("Error!\n");
		    else
		      printf("Aperture %hd %s %hd and %hd\n",
			     CTDB_MES_EDGE_KEY_TO_APERTURE_ID(aperture_key), 
			     "is attached to enclosures",
			     CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure_keys[0]), 
			     CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure_keys[1]));

		    printf("Return to main menu? 0/1\n");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 213:
	    {
		float64            x,y,z;
		int32              out_cell;
		CTDB_MES_NODE_KEY  enclosure_key;
		int16              mes_x, mes_y;
		float32            mes_z;

		printf("enter x, y, z: ");
		scanf("%lf %lf %lf", &x, &y, &z);
		
		if (!ctdb_gcs_to_mes_origin_offsets(ctdb, x, y, z, &out_cell,
						    &enclosure_key,
						    &mes_x, &mes_y, &mes_z))
		  printf("error in gcs->mes origin offsets\n");
		else
		  printf("Converted to MES %hd, Enclosure %hd: %hd %hd %f\n", 
			 CTDB_MES_NODE_KEY_TO_MES_ID(enclosure_key), 
			 CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(enclosure_key), 
			 mes_x, mes_y, mes_z);
		
		if (!ctdb_mes_origin_offsets_to_gcscs(ctdb, enclosure_key,
						      mes_x, mes_y, mes_z, 
						      &x, &y, &z))
		  printf("error in mes_origin_offsets->gcs\n");
		else
		  printf("Converted back to %f %f %f\n", x, y, z);
		break;
	    }
	  case 214:
	    {
		uint16  mes_id;
		int32   tri_id;
		int32   vtx_id;

		float64 vertex[3][XYZ];
		float64 normal[XYZ];
		CTDB_MES_VOLUME   *mes;
		CTDB_MES_TEMPLATE *templ;
		
		printf("Enter the index of your MES: ");
		scanf("%hd",&mes_id);
		
		if ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes))
		{
		    printf("Invalid mes id\n");
		    break;
		}
		printf("Enter your triangle id: ");
		scanf("%d", &tri_id);

		mes = &(ctdb->mes_volumes[mes_id]);
		templ = mes->template;
		
		if ((tri_id < 0) || (tri_id >= templ->num_triangles))
		{
		    printf("Invalid triangle id\n");
		    break;
		}

		ctdb_mes_get_triangle_normal(ctdb, mes_id, tri_id, normal);

		for (i=0; i < 3; i++)
		{
		    vtx_id = templ->all_triangles[tri_id].vertex_ids[i];
		    ctdb_mes_to_gcscs(mes, 
				      templ->all_vertices[vtx_id][X],
				      templ->all_vertices[vtx_id][Y],
				      templ->all_vertices[vtx_id][Z],
				      ctdb_get_cell_id(ctdb),
				      &vertex[i][X],
				      &vertex[i][Y],
				      &vertex[i][Z]);
		}
		printf("Triangle %d has vertices: \n", tri_id);
		printf("<<%f %f %f> <%f %f %f> <%f %f %f>\n",
		       vertex[0][X], vertex[0][Y], vertex[0][Z], 
		       vertex[1][X], vertex[1][Y], vertex[1][Z], 
		       vertex[2][X], vertex[2][Y], vertex[2][Z]);
		printf("            has normal: %f %f %f\n",
		       normal[X], normal[Y], normal[Z]);

		if (tri_id < templ->all_enclosures[0].first_wall_triangle_id)
		  printf("Triangle is an aperture triangle\n");
		else
		  {
		      for (i=0; i < templ->num_enclosures; i++)
			if (
			    tri_id >= 
			    templ->all_enclosures[i].first_wall_triangle_id 
			    &&
			    tri_id < 
			    templ->all_enclosures[i].first_wall_triangle_id
			    + templ->all_enclosures[i].num_triangles
			    )
			{
			    printf("Belongs to enclosure %d\n", i);
			    break;
			}
		  }

		printf("In Model space: \n");
		for (i=0; i < 3; i++)
		{
		    vtx_id = templ->all_triangles[tri_id].vertex_ids[i];
		    printf("<%f %f %f>   ", 
			   templ->all_vertices[vtx_id][X]*templ->fixed_point_basis,
			   templ->all_vertices[vtx_id][Y]*templ->fixed_point_basis,
			   templ->all_vertices[vtx_id][Z]*templ->fixed_point_basis);
		}
		printf("\n");
		
		break;		       

	    }
	  case 215:
	    {
		uint16 mes_id;
		CTDB_MES_TEMPLATE *templ;
		float64 vect[XY];
		float64 ave_x, ave_y, ave_z;
		int32 enclosure_id, vtx_id;
		int32 ok, i, j, tri_id;
		int32 negate;
		CTDB_MES_NODE_KEY mes_key, node_key, new_key;
		float64 location[XYZ], gcs[XYZ];
		printf("Enter the index of your MES: ");
		scanf("%hd",&mes_id);

		if ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes))
		{
		    printf("Invalid mes id\n");
		    break;
		}

		templ = ctdb->mes_volumes[mes_id].template;
		
		for (i=0; i < templ->num_apertures; i++)
		{
		    printf("Checking aperture %d: ", i);
		    
		    ok = 1;
		    negate = 1;

		    ave_x = 0;
		    ave_y = 0;
		    ave_z = 0;
		    for (j=0; j < templ->all_apertures[i].outline.num_vertices;
			 j++)
		    {
			vtx_id = 
			  templ->all_apertures[i].outline.vertex_ids[j];
			ave_x += templ->all_vertices[vtx_id][X];
			ave_y += templ->all_vertices[vtx_id][Y];
			ave_z += templ->all_vertices[vtx_id][Z];
		    }

		    ave_x /= templ->all_apertures[i].outline.num_vertices;
		    ave_y /= templ->all_apertures[i].outline.num_vertices;
		    ave_z /= templ->all_apertures[i].outline.num_vertices;

		    
		    enclosure_id = templ->all_apertures[i].connections[0].enclosure_id;
		    if (enclosure_id == 0)
		    {
			enclosure_id = templ->all_apertures[i].connections[1].enclosure_id;
			negate = -1;
		    }
		    
			CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key);
			CTDB_MES_IDS_TO_NODE_KEY(mes_id, enclosure_id, node_key);
		    for (tri_id = templ->all_apertures[i].first_triangle_id;
			 tri_id < templ->all_apertures[i].first_triangle_id +
			          templ->all_apertures[i].num_triangles; 
			          tri_id++)
		    {
			location[X] = ave_x + templ->all_triangles[tri_id].normal[X] * negate / templ->fixed_point_basis;
			location[Y] = ave_y + templ->all_triangles[tri_id].normal[Y] * negate / templ->fixed_point_basis;
			location[Z] = ave_z + templ->all_triangles[tri_id].normal[Z] * negate / templ->fixed_point_basis;
			
			ctdb_mes_to_gcscs(&ctdb->mes_volumes[mes_id], (int32) location[X], (int32) location[Y], 
					  (int32) location[Z], ctdb_get_cell_id(ctdb), 
					  &gcs[X], &gcs[Y], &gcs[Z]);
			

			ctdb_mes_find_enclosure(ctdb, mes_key, gcs[X], gcs[Y], gcs[Z], &new_key);

			if (node_key != new_key)
			{
			    if (ok)
			      printf("\n");
			    ok = 0;
			    printf("Bad triangle %d.  Normal = %f %f %f\n", tri_id, templ->all_triangles[tri_id].normal[X], templ->all_triangles[tri_id].normal[Y], templ->all_triangles[tri_id].normal[Z]);
			    
			}
			
		    }
		    if (ok)
		      printf("OK!\n");
		}
		

	    }

	  case 216:
	  {
	      CTDB_COORD2D pt;	
	      int32 main_menu=FALSE;
	      float32 a,b;
	      int32 soil;
	      
	      while (!main_menu)
	      {
		  CTDB_ELEV_DATA elev;
		  CTDB_ELEV_QUALIFIER qual=CTDB_RETURN_MAX | 
		    CTDB_INCL_MATERIAL | CTDB_INCL_TERRAIN_SKIN;;
		  printf("Point <x> <y> :");
		  scanf("%f %f",&a, &b);
		  pt.x = a;
		  pt.y = b;
		 
		  ctdb_lookup_qual_elevation(ctdb,a,b,qual,CTDB_NO_REF_ELEV,
					       &elev);
		  printf("Soil = %x\n", elev.material);
		  /* soil = ctdb_lookup_soil(ctdb,pt.x,pt.y);
		   * printf("Soil = %x\n",soil);
		   */
		  printf("RETURN to MAIN MENU (0/1)");
		  scanf("%d",&main_menu);
	      }
	      break;
	  } /* end of case 216 */
	  case 217:
	  {
	      int32 main_menu = FALSE;
	      int32 mes_id;
	      CTDB_MES_NODE_KEY node_key, node;
	      float64 x, y ,z;

	      while (!main_menu)
	      {
		  do 
		  {
		      printf("Enter MES id: ");
		      scanf("%d", &mes_id);
		      if ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes))
			printf("Invalid MES id\n");
		  } while ((mes_id < 0) || (mes_id >= ctdb->num_mes_volumes));

		  CTDB_MES_ID_TO_MES_KEY(mes_id, node_key);

		  printf("Enter x, y, z: ");
		  scanf("%lf%lf%lf", &x, &y, &z);
		  
		  if (!ctdb_mes_find_node(ctdb, node_key,
					  x, y, z, &node))
		    printf("ctdb_mes_find_node failed!\n");
		  else
		    printf("In MES %d Enclosure %d\n", 
			   mes_id, CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(node));

		  printf("Node is %d\n", node);
		  printf("RETURN to MAIN MENU (0/1)");
		  scanf("%d",&main_menu);
	      }
	      break;
	  }
	  case 500:
	    {
		static float64 data[8][4]=
		{
		    {/* WESTERN EDGE */
			0,50000,-25000,0
		    }
		    ,
		    {/* NORTHERN EDGE */
			50000,100000,0,25000
		    }
		    ,
		    {/* EASTERN EDGE */
			100000,50000,25000,0
		    }
		    ,
		    {/* SOUTHERN EDGE */
			50000,0,0,-25000
		    }
		    ,
		    {/* SOUTHWEST CORNER */
			0,0,-25000,-25000
		    }
		    ,
		    {/* NORTHWEST CORNER */
			0,100000,-25000,25000
		    }
		    ,
		    {/* NORTHEAST CORNER */
			100000,100000,25000,25000
		    }
		    ,
		    {/* SOUTHEAST CORNER */
			100000,0,25000,-25000
		    }
		};		
		int32           main_menu=0;
		float64         x0,y0,z0;
		static CTDB    *lastctdb=NULL;
		COORD_TCC_PTR   tcc=NULL;
		float64         lat,lng,z,lat_z,trash,nx,ny,nz;
		float64         n_extra_z,x, y, real_z, extra_z; 
		int32           ccell,ncell,i,cell[100],n_cell=100;
		int32 direction;
		float64 tmp[XYC];
		CTDB *ctdb;
		
		if ( !wld_lookup_playbox_cells (&n_cell,cell) )
		  printf("larger cell array dimension needed\n");
		
		
		while (!main_menu)
		{
		    printf("Select a cell : ");
		    for (i=0;i<n_cell;i++)
		      printf("%d, ",cell[i]);
		    printf("\nCell Selection > ");
		    scanf("%d",&ccell);
		    
		    for (i=0;i<8;i++)
		    {
			printf("\n************ Run %d ****************",i);
			/* find the real and extrapolated z for a point */
			x = data[i][X], y=data[i][Y];

			/* What is the real z of the x,y point on the ocean 
			 * surface?
			 */
			coord_convert(COORD_GCS,ccell,x,y,(float64)0.0,
				      COORD_LATLON,&lat,&lng,&real_z,0);
			coord_convert(COORD_LATLON,lat,lng,(float64)0.0,0,
				      COORD_GCSCS,ccell,&x,&y,&real_z);

			/* Compare real elevation to the extrapolated elevation
			 * data.
			 */
			compare_elevation(ccell,x,y,real_z,&extra_z);


			/* Get the neighboring cell */
			coord_convert(COORD_GCS,ccell,(x+data[i][X+2]),
				      (y+data[i][Y+2]),z,
				      COORD_GCS,&ncell,&nx,&ny,&nz);

			/* Convert GCS extrapolated point into the neighboring 
			 * coordinate system.
			 */
			coord_convert(COORD_GCS,ccell,x,y,extra_z,
				      COORD_GCSCS,ncell,&nx,&ny,&nz);

			compare_elevation(ncell,nx,ny,nz,&n_extra_z);
#if 0
			/* convert the extrapolated z into the original
			 * frame of reference and compare the difference
			 * between the two.
			 */
			coord_convert(COORD_GCS,ncell,nx,ny,n_extra_z,
				      COORD_GCSCS,ccell,&trash,&trash,
				      &n_extra_z);
			
			/* compare the extrapolated z values */
			printf("\norig (%d , %lf)\tneigh (%d , %lf)\ndiff = %lf \tpercent error %6.3lf \%\n",
			       ccell,extra_z,ncell,n_extra_z,
			       (extra_z-n_extra_z),
			       (extra_z-n_extra_z)/extra_z*100.0);
#endif
		    }
		    
		    printf("RETURN to MAIN MENU (0/1)");
		    scanf("%d",&main_menu);
		}
		break;
	    }
	  case 501:
	    {
		int32 i;
		int32 zone = COORD_DEFAULT_ZONE;
		char buf[80];
		int32 cell = 0;
		float64 x, y, z;
		
		int32 mgrs_zone_number;
		char mgrs_zone_letter;
		char map_sheet[2];
		int32 mgrs_east;
		int32 mgrs_north;
		int32 eres = 5;
		int32 nres = 5;
		float64 mgrs_z;
		x = 0.0;
		y = 0.0;
		z = 0.0;
		
		for (i=0; i < 50; i++)
		{
		    coord_convert(COORD_GCS, cell, x, y, z,
				  COORD_UTM_GRID, &zone, 5, buf, &mgrs_z);
		    mgrs_zone_number =  (buf[0]-'0')*10+(buf[1]-'0');
		    mgrs_zone_letter = buf[2];
		    map_sheet[0] = buf[3];
		    map_sheet[1] = buf[4];
		    mgrs_east = parse_num(buf+5, 5);
		    mgrs_north = parse_num(buf+10, 5);   
		    coord_convert(COORD_UTM_GRID, COORD_DEFAULT_ZONE, 5, buf,  mgrs_z,
				  COORD_GCS, &cell, &x, &y, &z);
		}
		break;
	    }	  
	  case 502:
	    {
		int32 i;

		int32 zone = 38;
		float64 northing = 3105000.1226121401;
		float64 easting =   507999.9994500574;
		float64 utm_z = 0.067864864133298397;

		int32 grid_zone = COORD_DEFAULT_ZONE;
		char buf[80];
		float64 grid_z;
		
		for (i=0; i < 50; i++)
		{
		    int32 mgrs_zone_number;
		    char mgrs_zone_letter;
		    char map_sheet[2];
		    int32 mgrs_east;
		    int32 mgrs_north;
		    int32 eres = 5;
		    int32 nres = 5;
		    float64 mgrs_z;
		    
		    coord_convert(COORD_UTM_NE, zone, northing, easting, utm_z,
				  COORD_UTM_GRID, &grid_zone, 5, buf, &grid_z);
		    
		    mgrs_zone_number =  (buf[0]-'0')*10+(buf[1]-'0');
		    mgrs_zone_letter = buf[2];
		    map_sheet[0] = buf[3];
		    map_sheet[1] = buf[4];
		    mgrs_east = parse_num(buf+5, 5);
		    mgrs_north = parse_num(buf+10, 5);   
		    
		    coord_convert(COORD_UTM_GRID, COORD_DEFAULT_ZONE, 5, buf, grid_z,
				  COORD_UTM_NE, &zone, &northing, &easting, &utm_z);
		}
		break;
	    }
	  default :
	    printf("NOT IMPLEMENTED YET ... FEEL FREE :-)\n");
	    selection = 0;
	    break;
	}
	}
	
	if ( ctdb != NULL )
      STDDEALLOC(ctdb);
}

