#ifndef lint
static char rcsid [] = "$RCSfile: hdredit.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: hdredit.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Program to edit a CTDB file header                        *
*       Created: Fri Jul  8 1994                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb.h"
#include <stdio.h>

static char *help_string =
  "\nEach field in the CTDB file header will be presented.\nAt each prompt, type <RETURN> to keep the current value, or type a new value.\n\n";

main(argc, argv)
    int   argc;
    char *argv[];
{
    FILE *fin, *fout;
    CTDB_FILE_HEADER hdr;
    CTDB_AUX_FILE_HEADER aux;
    int ival;
    double fval;
    char cval;
    char instring[1024];
    char buf[4096];
    int c,i;
    int tloc;

    if (argc < 3)
    {
	printf("Usage: %s <input database.ctb> <output database.ctb>\n", 
	       argv[0]);
	exit(0);
    }

    if (!(fin = fopen(argv[1], "r")))
    {
	perror("Unable to open input file");
	exit(-1);
    }
    if (!(fout = fopen(argv[2], "w")))
    {
	perror("Unable to open output file");
	exit(-1);
    }

    puts(help_string);

    /* Read the header */
    if (fread(&hdr, sizeof(hdr), 1, fin) != 1)
    {
	perror("Unable to read header");
	exit(-1);
    }    

    /* Read the aux header */
    if (fread(&aux, sizeof(aux), 1, fin) != 1)
    {
	perror("Unable to read aux header");
	exit(-1);
    }    

#define GETARRAY(src,var,index)	      			\
    for(;;)						\
    {							\
	printf("var[%d] (%d): ",index,src.var[index]);  \
	fflush(stdout);					\
	gets(instring);					\
	if (instring[0])				\
	{						\
	    if (sscanf(instring, "%d", &ival) == 1)	\
	    {                                           \
	        src.var[index] = ival;			\
		break;                                  \
	    }                                           \
	}						\
	else						\
	  break;					\
    }

#define GETINT(src,var)					\
    for(;;)						\
    {							\
	printf("var (%d): ", src.var);			\
	fflush(stdout);					\
	gets(instring);					\
	if (instring[0])				\
	{						\
	    if (sscanf(instring, "%d", &ival) == 1)	\
	    {                                           \
	        src.var = ival;				\
		break;                                  \
	    }                                           \
	}						\
	else						\
	  break;					\
    }

#define GETCHAR(src,var)				\
    for(;;)						\
    {							\
	printf("var (%c): ", src.var);			\
	fflush(stdout);					\
	gets(instring);					\
	if (instring[0])				\
	{						\
	    if (sscanf(instring, "%c", &cval) == 1)	\
	    {                                           \
	        src.var = cval;				\
		break;                                  \
	    }                                           \
	}						\
	else						\
	  break;					\
    }

#define GETFLOAT(src,var)				\
    for(;;)						\
    {							\
	printf("var (%f): ", src.var);			\
	fflush(stdout);					\
	gets(instring);					\
	if (instring[0])				\
	{						\
	    if (sscanf(instring, "%lf", &fval) == 1)	\
	    {                                           \
	        src.var = fval;				\
		break;                                  \
	    }                                           \
	}						\
	else						\
	  break;					\
    }

#define GETSTRING(src, var, len)			\
    for(;;)						\
    {							\
	printf("var (%s): ", src.var);			\
	fflush(stdout);					\
	gets(instring);					\
	if (instring[0])				\
	{                                               \
	    strncpy(src.var, instring, len);		\
	    break;                                      \
	}                                               \
	else						\
	  break;					\
    }

    GETINT(hdr, format);
    GETINT(hdr, is_curved);
    GETINT(hdr, pad0);
    GETINT(hdr, is_gridded);
    GETINT(hdr, is_tinned);
    GETINT(hdr, pad1);
    for (i=0;i<CTDB_HEADER_RESERVED;i++)
    {
	GETARRAY(hdr, reserved1,i);
    }
    GETFLOAT(hdr, origin_northing);
    GETFLOAT(hdr, origin_easting);
    GETINT(hdr, origin_zone_number);
    GETCHAR(hdr, origin_zone_letter);
    GETINT(hdr, datum);
    GETSTRING(hdr, name, CTDB_NAME_LENGTH);
    GETINT(hdr, version);
    time(&tloc);
    strcpy(hdr.date, ctime(&tloc));

    printf("\nLegal gcs_mode values:\n");
    printf("0.\tSIMNET mode\n");
    printf("1.\tGCS single-cell mode\n");
    printf("2.\tGCS multi-cell mode\n");
    
    GETINT(aux, gcs_mode);
    GETINT(aux, cell_id);
    GETINT(aux, origin_x);
    GETINT(aux, origin_y);

    printf("\nWriting header\n");
    if (fwrite(&hdr, sizeof(hdr), 1, fout) != 1)
    {
	perror("Error writing header");
	exit(-1);
    }

    printf("Writing aux header\n");
    if (fwrite(&aux, sizeof(aux), 1, fout) != 1)
    {
	perror("Error writing aux header");
	exit(-1);
    }

    printf("Copying data..."); fflush(stdout);
    
    while (c = fread(buf, 1, sizeof(buf), fin))
    {
	if (fwrite(buf, 1, c, fout) != c)
	{
	    perror("Error writing data");
	    exit(-1);
	}
    }

    if (fclose(fin) || fclose(fout))
    {
	perror("Error closing files");
	exit(-1);
    }

    printf("Done\n");
    return 0;
}
