#ifndef lint
static char rcsid [] = "$RCSfile: posts.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: posts.c                                                          *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Program to print statistics about grid posts              *
*       Created: Tue Jul 13 1993                                            *
*       Author: ajc                                                         *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "../libctdb_local.h"
#include <ct_post.h>

CTDB ctdb;

int32 posts = 0;
int32 trees = 0;
int32 micro = 0;
int32 features = 0;
int32 buildings = 0;

static void analyze_posts(ctdb, patch_x, patch_y)
    CTDB *ctdb;
    int32 patch_x;
    int32 patch_y;
{
    int32 posts_per_patch_side = ctdb->patch_incr;
    int32 x = patch_x * posts_per_patch_side;
    int32 y = patch_y * posts_per_patch_side;
    int32 j;
    int32 i;

    for (j = 0; j < posts_per_patch_side; j++)
      for (i = 0; i < posts_per_patch_side; i++)      
      {
	  int32 norm_x = x + i;
	  int32 norm_y = y + j;
	  int32 world_x = norm_x * ctdb->incr;
	  int32 world_y = norm_y * ctdb->incr;
	  int32 post = ctdb_lookup_post_elevation(ctdb, norm_x, norm_y);

	  posts++;
	  if (post & POST_TREES)
	    trees++;
	  if (post & POST_MICROTERRAIN)
	    micro++;
	  if (post & POST_FEATURES)
	    features++;
	  if (post & POST_BUILDINGS)
	    buildings++;
      }
}

main (argc,argv)
    int32 argc;
    char *argv[];

{
    int32 i;
    int32 j;

    if (argc != 2)
    {
	printf("Usage: %s <database.ctdb>\n", argv[0]);
	exit(0);
    }

    ctdb_read(argv[1],&ctdb,0);

    ctdb_print_description(&ctdb);

    for (j = 0; j < ctdb.patches_high; j++)
      for (i = 0; i < ctdb.patches_wide; i++)      
	analyze_posts(&ctdb, i, j);

    printf("%d posts, of which\n %d have features, %d have buildings, %d have trees, %d have micro\n",
	   posts, features, buildings, trees, micro);
    printf(" % microterrain = %.4f %%\n",
	   (((float64) micro) * 100.0) / (float64) posts);
}
