#ifndef lint
static char rcsid [] = "$RCSfile: print-db.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
 *   File: printdb.c                                                         *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts                                                           *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Used print various information about the CTDB database    *
 *       Created:  April 1995                                                *
 *       Author:   braudaway                                                 *
 *       Remarks:                                                            *
 *                                                                           *
 *****************************************************************************/

#include <libtime.h>
#include "../libctdb_local.h"

#define CTDB_FEATURE_MASK_SHIFT 20
void 
ctdb_print_feature_present(ctdb, patch_number)
    CTDB *ctdb;
    int32 patch_number; 
    
{
    int32 px=POST_X((int32)(patch_number%ctdb->patches_wide)*ctdb->patch_incr);
    /* add one to patch y because posts are being evaluated from top to bottom
     * of the patch 
     */
    int32 py=POST_Y((patch_number/ctdb->patches_wide+1)*ctdb->patch_incr);
    int32 i,j;
      
    printf("Post Feature present Flags\n");
    /* print post flags top down */
    for (j=0;j<ctdb->patch_incr;j++)
    {
	for (i=0;i<ctdb->patch_incr;i++)
	  printf("%01x",(uint8)(ctdb_lookup_feature_mask(ctdb,px+i,py-j)
				>>CTDB_FEATURE_MASK_SHIFT));
	printf("\n");
    }
}

#ifdef OLD_ROUTINE
void 
ctdb_print_feature_present(ctdb, patch_number)
    CTDB *ctdb;
    int32 patch_number; 
    
{
    CTDB_PATCH_GROUP *pg=ctdb->patch_groups+PG_NUMBER(patch_number);
    int32 posty, postx, post;
    unsigned short *posts;
    int32 index = PG_MEMBER(patch_number) * (POSTS_PER_PATCH>>1);
      
    printf("Post Feature present Flags\n");
    /* print post flags top down */
    for (posty=POSTS_PER_PATCH_SIDE-1;posty>=0;posty--)
      {  
	  for (postx =0; postx <(POSTS_PER_PATCH_SIDE>>1);postx++)
	    {
		post = posty*(POSTS_PER_PATCH_SIDE>>1) + postx;
		printf("%02x",pg->feature_present[post+index]);	  
	    }
	  printf("\n");	  
      }
}
#endif   

void
PrintPatchInfo(ctdb)
    CTDB *ctdb;
{
    int32 patch_number,spatch_number,epatch_number;
    CTDB_STATUS status;
    int32 mode;
    int32 patch_x, patch_y;

    printf("Starting Patch Number > ");
    scanf("%d",&spatch_number);
    printf("Ending Patch Number > ");
    scanf("%d",&epatch_number);
    epatch_number++;
    
    
    for (patch_number=spatch_number;patch_number<epatch_number;patch_number++)
    {
	patch_x = patch_number%ctdb->patches_wide;
	patch_y = (int32)(patch_number/ctdb->patches_wide);
	printf("\n Patch number %d (%d, %d)\n", patch_number, patch_x, patch_y);
	ctdb_print_feature_present(ctdb, patch_number);
	ctdb_print_features(ctdb, patch_x, patch_y);
	ctdb_print_te_data(ctdb, patch_number, 1);
    }
}

main (argc,argv)
    int32 argc;
    char *argv[];
{
    CTDB *ctdb = NULL;
    FILE *fileptr;
    char filename[80];
    char dbname[80];
    int32 selection = 1;
    CTDB_STATUS ctdb_resident;

    ctdb = (CTDB *)ctdb_alloc(ctdb,sizeof(CTDB),"Main Database Structure");
    if ((argc != 2) && (argc != 3) && (argc != 7))
    {
	printf("Usage: %s <file,ctb>\n", argv[0]);
	exit(0);
    }
    sprintf(dbname,"%s",argv[1]);
    ctdb_read(dbname,ctdb,0);
    while (selection)
    {	
	printf("What would you like to run?\n\n");
	printf("\tSelect Option:\n");
	printf("\t\t\t%2d. Print Database Patches (Feature bits, Features, TEs\n",1);
	printf("\t\t\t%2d. Exit\n\n",0);
	printf("\t\t\t\tEnter Selection :");
	scanf("%d",&selection);
	
	switch (selection)
        {
	  case 0 :
	    break ;
	  case 1 :
	    ctdb_print_description(ctdb);
	    PrintPatchInfo(ctdb);
	    break;
	  default :
	    selection = 0;
	    ctdb_print_description((CTDB *)ctdb);
	    break;
	}
    }
    if ( ctdb != NULL )
      free(ctdb);
}
