#ifndef lint
static char rcsid [] = "$RCSfile: profile.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: profile.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Profiling test program                                    *
*       Created: Thu May 23 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
*	On the GT100 series, the "realtime" clock rolls over if it's not    *
*	called at least every 400ms.  The call to obtain the time on the    *
*	GT is fairly cheap, since it's a lookup in a hardware register.     *
*	Making frequent calls under Unix to check the time, however, can    *
*	be fairly expensive.  For this reason, there are two ways of	    *
*	tracking time in this profiling program -- one of the GT100s, and   *
*	one for everything else.			pjm 20 Jan 94	    *
                                                                           *
****************************************************************************/
#include <libntpif.h>
#include <libtime.h>
#include <libctdb.h>

CTDB ctdb;
float64 min_x, max_x, min_y, max_y;

#define STEP 249

float64 dummy_fcn()
{
    return 0;
}

time_ptop()
{
    int32 i;
    float64 x,y,z,zl,zh;
    float64 dx, dy, opacity;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;

    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
          avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  z = ctdb_lookup_elevation(&ctdb, x, y) + 2.0;
	  zl = ctdb_lookup_elevation(&ctdb, x+dx, y+dy) + 0.5;
	  zh = zl + 1.5;
	  v += dummy_fcn(&ctdb,x,y,z,x+dx,y+dy,zl,zh,
			 4.0,NULL,&opacity,0.05,0);
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  z = ctdb_lookup_elevation(&ctdb, x, y) + 2.0;
	  zl = ctdb_lookup_elevation(&ctdb, x+dx, y+dy) + 0.5;
	  zh = zl + 1.5;
	  checksum += ctdb_point_to_point(&ctdb,x,y,z,x+dx,y+dy,zl,zh,
					  4.0,NULL,&opacity,0.05, 0,
					  (CTDB_VEHICLE_LOCATION *)0,
					  0,0);
      }

    end = time_realtime_clock();

    printf("Point to point intervis:\n");
    printf("    \(zh-zl=1.5, target width=4.0, \n");
    printf("      minimum_visibility = 0.05\)\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

time_ptop_bare_minimum()
{
    int32 i;
    float64 x,y,z,zl,zh;
    float64 dx, dy, opacity;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;

    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
          avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  z = ctdb_lookup_elevation(&ctdb, x, y) + 2.0;
	  zl = ctdb_lookup_elevation(&ctdb, x+dx, y+dy) + 0.5;
	  zh = zl;
	  v += dummy_fcn(&ctdb,x,y,z,x+dx,y+dy,zl,zh,
			 0.0 /* zero width target */,
			 0.0 /* trees have no effect */,
			 1.00 /* return if visibility drops below 100% */,
			 0 /* no checking of vehicle-in-front-of-vehicle blockage */);
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  z = ctdb_lookup_elevation(&ctdb, x, y) + 2.0;
	  zl = ctdb_lookup_elevation(&ctdb, x+dx, y+dy) + 0.5;
	  zh = zl;
	  checksum += ctdb_point_to_point(&ctdb,x,y,z,x+dx,y+dy,zl,zh,
					  0.0  /* zero width target */,
					  NULL, &opacity,
					  1.00 /* return if visibility drops below 100% */,
					  0    /* no checking of vehicle-in-front-of-vehicle
						  blockage */,
					  (CTDB_VEHICLE_LOCATION *)NULL ,
					  0,
					  0);
      }

    end = time_realtime_clock();

    printf ("Point to Point intervis BARE MINIMUM:\n");
    printf ("    \(zl==zh, zero target width,\n");
    printf ("      return if visibility drops below 100%%\):\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

time_pthrup()
{
    int32 i;
    float64 x,y,z,zl,zh;
    float64 dx, dy;
    uint32 start, end, middle;
    float64 opacity;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  z = ctdb_lookup_elevation(&ctdb, x, y) + 2.0;
	  zl = ctdb_lookup_elevation(&ctdb, x+dx, y+dy) + 100.0;
	  zh = zl + 1.5;
	  v += dummy_fcn(&ctdb,x,y,z,x+dx,y+dy,zl,zh,
			 4.0,0.5,0.05,3500.0);
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  z = ctdb_lookup_elevation(&ctdb, x, y) + 2.0;
	  zl = ctdb_lookup_elevation(&ctdb, x+dx, y+dy) + 20.0;
	  zh = zl + 1.5;
	  checksum += ctdb_point_thru_point(&ctdb,x,y,z,
					    x+dx,y+dy,zl,zh,
					    4.0 /*width*/,
					    NULL,
					    &opacity,
					    0.05/*min_vis*/,
					    3500.0 /*range*/,
					    0 /* number of vehicles*/,
					    (CTDB_VEHICLE_LOCATION *)NULL,
					    0,
					    0);
      }

    end = time_realtime_clock();

    printf("Point thru point (RADAR) intervis:\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

time_get_gl()
{
    int32 i;
    float64 x,y,z,zl,zh;
    float64 dx, dy;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  v += dummy_fcn(&ctdb, x+dx, y+dy);
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  checksum += ctdb_lookup_elevation(&ctdb, x+dx, y+dy);
      }

    end = time_realtime_clock();

    printf("Lookup elevation:\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

time_get_bl()
{
    int32 i;
    float64 x,y,z,zl,zh;
    float64 dx, dy;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  v += dummy_fcn(&ctdb, x+dx, y+dy, 1);
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  checksum += ctdb_lookup_max_elevation(&ctdb, x+dx, y+dy, 1);
      }

    end = time_realtime_clock();

    printf("Lookup elevation (+ buildings & canopies):\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

time_get_hg()
{
    int32 i;
    float64 x,y,z,zl,zh;
    float64 dx, dy;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  v += dummy_fcn(&ctdb, x, y ,x+dx, y+dy, NULL);
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  checksum += ctdb_find_high_ground(&ctdb, x, y ,x+dx, y+dy, 
					    CTDB_USE_DEFAULT, NULL);
      }

    end = time_realtime_clock();

    printf("Find high ground:\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

#define CS45 0.70710678118654752440

time_pv()
{
    int32 i;
    float64 x,y,z=0.0,zl,zh;
    float64 dx, dy;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 htow[3][3];
    int32 soil;
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dummy_fcn(&ctdb, x+dx, y+dy, 5.8, 2.5, CS45, CS45, &z,
		    htow, &soil);
	  v += z;
      }

    middle = time_realtime_clock();

    i = 0;
    srand(0);
    for (x=min_x;x<max_x;x+=STEP)
      for (y=min_y;y<max_y;y+=STEP,i++)
      {
#ifdef _GT_
	  avoiding_rollover = time_realtime_clock();
#endif
	  dx = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  dy = 1500.0-3000.0*(rand()&0xFFFF)/0xFFFF;
	  ctdb_place_vehicle(&ctdb, x+dx, y+dy, 5.8, 2.5, CS45, CS45, &z,
			     htow, &soil);
	  checksum += z;
      }

    end = time_realtime_clock();

    printf("Place vehicle:\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

float64 rot[3][3] = { { 1.0 , 0.0 , 0.0 } ,
		     { 0.0 , 1.0 , 0.0 } ,
		     { 0.0 , 0.0 , 1.0 } ,
		   };
time_as()
{
    int32 i;
    uint32 start, end, middle;
#ifdef _GT_
    uint32 avoiding_rollover;
#endif
    float64 loc[3], width, height;
    float64 checksum = 0;
    float64 v = 0;

    start = time_realtime_clock();

    for (i=0;i<100000;i++)
    {
#ifdef _GT_
	avoiding_rollover = time_realtime_clock();
#endif
	dummy_fcn(10.0, 10.0, 10.0,
		  20.0, 20.0, 20.0,
		  rot, 2.0, 2.0, 2.0,
		  loc, &width, &height);
	v += height;
    }

    middle = time_realtime_clock();

    for (i=0;i<100000;i++)
    {
#ifdef _GT_
	avoiding_rollover = time_realtime_clock();
#endif
	ctdb_apparent_size(10.0, 10.0, 10.0,
			   20.0, 20.0, 20.0,
			   rot, 2.0, 2.0, 2.0,
			   loc, &width, &height);
	checksum += height;
    }


    end = time_realtime_clock();

    printf("Apparent size:\n");
    printf("\tChecksum = %f\n",checksum);
    printf("\t%f usec (%0.0f ops / second)\n",
	   1000.0*(end-middle - (middle-start))/i,
	   1000.0*i/(end-middle - (middle-start)));
}

main (argc,argv)
    int32 argc;
    char *argv[];
{
    if ((argc != 2) && (argc != 3) && (argc != 7))
    {
	printf("Usage: %s <file.ctb>\n", argv[0]);
	printf("       %s <file.ctb> [memory_limit]\n", argv[0]);
	printf("       %s <file.ctb> [memory_limit min_x max_x min_y max_y]\n", argv[0]);
	exit(0);
    }

    ctdb_read(argv[1],&ctdb,(argc == 2) ? 0 : atoi(argv[2]));
    ctdb_print_description(&ctdb);

    if ((argc == 2) || (argc == 3))
    {
        min_x = ctdb.min_x;
        min_y = ctdb.min_y;
        max_x = ctdb.max_x;
        max_y = ctdb.max_y;
    }
    else
    {
        min_x = atoi(argv[3]);
        max_x = atoi(argv[4]);
        min_y = atoi(argv[5]);
        max_y = atoi(argv[6]);
    }

    min_x += 3500; max_x -= 3500; min_y += 3500; max_y -= 3500;

    printf("\nTest Area: (%d < X < %d) (%d < Y %d)\n\n",
	   (int)min_x, (int)max_x, (int)min_y, (int)max_y);

    time_init(0);

    time_ptop_bare_minimum();
    time_ptop_bare_minimum();
    time_ptop();
    time_ptop();
/*  time_pthrup(); */
/*  time_pthrup(); */
    time_get_gl();
    time_get_gl();
    time_get_bl();
    time_get_bl();
    time_get_hg();
    time_get_hg();
    time_pv();
    time_pv();
    time_as();
    time_as();
}
