#ifdef ALPHA_SHORT_PTR
#pragma pointer_size long
#endif
#ifndef lint
static char rcsid [] = "$RCSfile: viewer.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: viewer.c                                                           *
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Database viewing utility
*       Created: Tue Jun  4 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
/*#define SE3_OCEAN_DEMO */
#include <stdio.h>
#include <math.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <Xm/ToggleB.h>
#include <Xm/Scale.h>
#include <Xm/RowColumn.h>
#include <sys/types.h>
#include <varargs.h>
#include <stdext.h>
#include <libgcs.h>
#include <libworld.h>
#include <libcmdline.h>
#include "../compiler/ct_reform.h"
/*#define CTDB_VIEWER_DEBUG */
/*#define DO_COLOR */
#define NCELLS 16
#define WINDOW pixmap
#define SIZEOF_LIST 1000
#define NCOLORS XtNumber(color_names)
#define ARC_CONST_360x64 23040

static CTDB_ELEV_QUALIFIER viewer_qual=0;
static CTDB_FEATURE_OBJ view_obj[FC_CLASS_MAX];
static CTDB_FEAT_TYPE   view_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];

typedef struct ctdb_view_obj_data
{
    CTDB_PATCH_DATA patch;
} CTDB_VIEW_OBJ_DATA;


struct main_options_struct
{
    CMD_STRING_OPTION   data;
    CMD_BOOLEAN_OPTION  tide;
    CMD_INTEGER_OPTION  soil_sample;
} main_options;

struct main_options_struct main_options = {
    {
	"Data Directory" ,"Specifies directory where data files are stored" ,
	NULL ,
	CMD_STRING ,	"data" ,	NULL ,		"../../../data" ,
    } ,
    {
	"Tidal Demo" ,"Use the slide to change the tide level if in 2D mode",
	NULL ,
	CMD_BOOLEAN,	"tide" ,		"notide" ,	0 ,
    } ,
    {
        "Soil Sampling Rate" ,"Specify the soil sampling rate (in pixels) for rendering soils",
        NULL ,
        CMD_INTEGER,    "soil_sample" ,         NULL ,  5 ,
    } ,
};

/* Feature type priority used when polys overlap.  In increasing order. */
typedef enum
{
    LAND_PRI     = 1,
    INTERV_PRI   = 2,
    TREE_PRI     = 3,
    BUILDING_PRI = 4
} viewer_priority;
#define MAX_POLY_PTS 4
typedef struct
{
    viewer_priority priority;
    char always_outline;
    float64 min_dist;
    float64 max_dist;
    int32 n_pts;
    float64 pts[MAX_POLY_PTS][3];
    GC gc;
} SCR_POLY;
#define MAX_POLYS (1<<14)
typedef enum 
{
    NO_VISIBILITY      = 0,
    PARTIAL_VISIBILITY = 1,
    FULL_VISIBILITY    = 2
} viewer_visibility;


/* polygon soil types */
#define SOIL_NA         0                /* Unknown type of soil             */
#if 0
#define SOIL_ROAD       1                /* Asphalt or other hard surface    */
#define SOIL_RCI250     2                /* Packed Soil                      */
#define SOIL_RCI050     3                /* Sandy Soil                       */
#endif
#define SOIL_WATER      5                /* River that is passable           */
#define SOIL_MUCK       4                /* River that is impassable         */

/* New polygon soil types (for NTC) */

#define SOIL_NON_SOIL		0	/* Unknown type of soil	             */
#ifndef SOIL_ROAD
#define SOIL_ROAD		1	/* Asphalt or other hard surface     */
#define SOIL_RCI250		2	/* Packed soil or dirt road          */
#define SOIL_RCI050		3	/* Soft sandy soil                   */
#endif
#define SOIL_DEEP_WATER		4	/* Impassable deep water             */
#define SOIL_SHALLOW_WATER	5	/* Passable shallow water            */
#define SOIL_MUD		6	/* Muddy soil                        */
#define SOIL_MUDDY_ROAD 	7	/* Wet dirt road                     */
#define SOIL_ICE		8	/* Slick ice surface                 */
#define SOIL_SWAMP		9	/* Very soft surface                 */
#define SOIL_FORRESTED  	10	/* Canopy or forrested area          */
#define SOIL_US_RAILROAD	11	/* Railroad w/ US specifications     */
#define SOIL_EURO_RAILROAD	12	/* Railroad w/ European specs.       */
#define SOIL_ROCKY		13	/* Small rocks <= 18 inches          */
#define SOIL_BOULDERS		14	/* Large boulders 6 ft. high         */
#define SOIL_FLIMSY		15	/* Indoor surface for DI             */
#define SOIL_NO_GO		15	/* General No Go Terrain for NTC     */
                         
/* ----------------------------------------------------------------
 *                    Function prototypes: 
 *                Prototypes required for ANSI 
 * ----------------------------------------------------------------*/
static void
draw_grid(
	  int32 x0,
	  int32 y0,
	  int32 x1,
	  int32 y1
	  );
static void
add_feature_polygons(
		     int32 px,
		     int32 py,
		     CTDB_FEATURE_DATA data[],
		     int32 incr
		     );
static int32 
shade_val(
	  float64 xnorm, float64 ynorm, 
	  float64 znorm, float64 dist2
	  );
static void    
compute_3D_transform(
		     int32 low_x, 
		     int32 low_y, 
		     int32 width_m, 
		     int32 height_m
		     );
static void
transform_point(
		float64 pt_in[3],
		float64 pt_out[3]
		);
static void
squash_point(
	     float64 pt_in[3],
	     float64 pt_out[2]
	     );
static int32
to_pixels(
	  float64 x,
	  float64 y,
	  float64 z,
	  float64 inv_scale,
	  int16 *xresult,
	  int16 *yresult
	  );
static int32 
clip_poly(
	  int32    n_pts_in,
	  float64 pts_in[][2],
	  int32   *n_pts_out,
	  int16  pts_out[][2]
	  );
static void
transform_poly(
	       int32    n_pts_in,
	       float64 pts_in[][3],
	       int32   *n_pts_out,
	       int16  pts_out[][2]
	       );
static void
interpolate_color(
		  float64  elev,
		  XColor *color
		  );
static void
setup_cmap(
	   float64 low_elev,
	   float64 high_elev
	   );

static void
setup_gcs();

static void
draw_contour_segs(
		  int32    nsegs,
		  float64 segs[],
		  float64 elevs[]
		  );
static void
draw_contours(
	      int32 x0,
	      int32 y0,
	      int32 x1,
	      int32 y1
	      );
static void
set_dist(
	 SCR_POLY *poly 
	 );

static void
draw_iv_line(
	     int32    x0, 
	     int32    y0,
	     float64  z0,
	     int32    x1, 
	     int32 y1,
	     float64 z1,
	     float64 x,
	     float64 y,
	     float64 zl,
	     float64 zh,
	     float64 z,
	     int32    vis
	     );

static void
draw_iv_mline(
	      int32 px, 
	      int32 py,
	      CTDB_FEATURE_DATA v0[2],
	      CTDB_FEATURE_DATA v1[2]
	      );
static void
circle_point(
	     float64 x,
	     float64 y,
	     float64 z
	     );
static void
circle_post(
	    int32 x,
	    int32 y
	    );
static void
set_interv_frac( 
		float64 fraction
		);

static void
draw_intervis_pole(
		   int32 xl,
		   int32 yl,
		   int32 xh,
		   int32 yh
		   );
static void
draw_sun();

static void
draw_los(
	 int32 x0,
	 int32 y0,
	 int32 x1,
	 int32 y1
	 );
static void
draw_intervis_ray(
		  int32 x0,
		  int32 y0,
		  int32 x1,
		  int32 y1
		  );
static void 
compute_triangle_normal(
			CTDB_COORD3D *pts,
			int32         which_t,
			float64      *pa, 
			float64       *pb, 
			float64 *pc
			);
static void
add_terrain_polygon(
		    CTDB_COORD3D *db_pts,
		    int32 n_pts,
		    int32 soil
		    );
    
static void
draw_microterrain3D(
		    int32 x0,
		    int32 y0,
		    int32 x1,
		    int32 y1
		    );

static void
draw_micro_line(
		GC gc,
		int16 px0,
		int16 py0,
		int16 px1,
		int16 py1,
		int16 px2,
		int16 py2,
		int32 mask
		);
static void
draw_microterrain2D(
		    int32 x0,
		    int32 y0,
		    int32 x1,
		    int32 y1
		    );
static void
draw_microterrain(
		  int32 x0,
		  int32 y0,
		  int32 x1,
		  int32 y1
		  );
static void
draw_terrain_elements_markings2D(
				 int32 x0,
				 int32 y0,
				 int32 x1,
				 int32 y1
				 );
static void
draw_terrain_elements(
		      int32 x0,
		      int32 y0,
		      int32 x1,
		      int32 y1,
		      int32 do_soil
		      );
static void
draw_assoc2D(
	     int32 x0,
	     int32 y0,
	     int32 x1,
	     int32 y1
	     );

static void
draw_gridded_terrain(
		     int32 x0,
		     int32 y0,
		     int32 x1,
		     int32 y1,
		     int32 do_soil
		     );

/* ----------------------------------------------------------------- */
CTDB *ctdb;
Widget toplevel, form, area, g_dialog_shell;
Pixmap pixmap;
Colormap cmap = NULL;
Dimension win_width, win_height;
float64 scale_mpp;
XImage *image = NULL;
uint32 local_cmap[NCELLS];
char *color_names[] = {
    "orange red",
    "tan",
    "gray50" ,
    "hotpink",
    "Violet" ,
    "brown" ,
    "dark violet",
    "darkolivegreen4" ,
    "black" ,
    "gray75" ,
    "gray25" ,
    "green" ,
    "yellow" ,
    "red" ,
    "skyblue" ,
};
XColor colors[NCOLORS];
static int32 old_do_3D=0;
static float64 zoom=1.0;
static int32 low_y=0, low_x=0;
static int32 dir3D = 0;
static int32 fov3D = 120;
static int32 alt3D = 1;
static int32 contour = 40;
static int32 do_3D = FALSE;
static int32 do_contour = FALSE;
static int32 do_grid = FALSE;
static int32 do_terrain = FALSE;
static int32 do_assoc = FALSE;
static int32 do_micro = FALSE;
static int32 do_water = FALSE;
static int32 do_ml_micro = FALSE;
static int32 do_linear = FALSE;
static int32 do_hypso = FALSE;
static int32 do_masks = FALSE;
static int32 do_intervis = FALSE;
static int32 do_slow = FALSE;
static int32 do_outline = FALSE;
static int32 do_drive = FALSE;
static int32 do_build = FALSE;
static int32 do_tree = FALSE;
static int32 do_abstract = FALSE;
static int32 do_nodes = FALSE;
static int32 do_edges = FALSE;
static int32 do_soil = FALSE;
static int32 do_road_lookups = TRUE;
static int32 do_elevation_lookups = TRUE;
static int32 do_lookups = FALSE;
static int32 do_damage = FALSE;
static int32 do_markings = FALSE;
static int32 got_click = FALSE;
static CTDB_COORD3D pt1;
char iv_frac_buf[20];
static CTDB_TERRAIN_TRIANGLE *te_pts;
static int32 pts_dim=SIZEOF_LIST;

SCR_POLY poly_list[MAX_POLYS];
int32 pl_cnt = 0;
SCR_POLY mes_poly_list[MAX_POLYS];
int32 mes_pl_cnt = 0;

float64 sun_vec[] = { 1.0 , 0.0 , 0.0 };
float64 origin_x, origin_y, origin_z;
float64 sin_dir, cos_dir, sin_alt, cos_alt;
float64 eye_dist, proj_dist;
float64 view_x, view_y, view_z;
float64 my_x, my_y, my_z;
#define CTDB_N_SOILS 16
GC contour_gc, micro_gc, grid1_gc, grid2_gc, line_gc, tree_gc,
  iv_gcs[4], shade_gcs[16], sky_gc, black_gc, hyp_gc, abstract_gc,
  node_gc, bridge_gc,te_gc,soil_gcs[CTDB_N_SOILS],intervis_gc;
int32 iv_clr[4];
#ifndef PI
#define PI 3.141592654
#endif

/* cmp_poly
 * 
 * If the return is positive then the second  element is placed in the 
 * list to be rendered ealier then the first.  If negative then the reverse
 * is true.  If 0, then no change will occur.
 */
#define CP_ERR 100.0
static int32
cmp_poly(SCR_POLY *poly0,
	 SCR_POLY *poly1)
{
    float64 dx,dn;

    /* If one is way behind the other, put the farther one first */
    if (poly1->min_dist > poly0->max_dist)
      return 1;

    if (poly0->min_dist > poly1->max_dist)
      return -1;

    /* If they are about the same, use priority */
    dx = poly1->max_dist - poly0->max_dist;
    dn = poly1->min_dist - poly0->min_dist;
    if ((dx > -CP_ERR) && (dx < CP_ERR) &&
	(dn > -CP_ERR) && (dn < CP_ERR))
      return poly0->priority - poly1->priority;

    /* If one is inside the other, put the expansive one first */
/*
    if ((poly1->min_dist < poly0->min_dist) &&
	(poly1->max_dist > poly0->max_dist))
      return 1;
    if ((poly0->min_dist < poly1->min_dist) &&
	(poly0->max_dist > poly1->max_dist))
      return -1;
*/
    /* Simple overlap.  Draw the farther one first */
    if (poly1->max_dist > poly0->max_dist)
      return 1;
    if (poly0->max_dist > poly1->max_dist)
      return -1;

    /* Back edges line up.  How about the front edges */
    if (poly1->min_dist > poly0->min_dist)
      return 1;
    if (poly0->min_dist > poly1->min_dist)
      return -1;

    return 0;
}


static void    
compute_3D_transform (low_x, low_y, width_m, height_m)
    int32 low_x;
    int32 low_y;
    int32 width_m;
    int32 height_m;
{
    float64 dir_rad;

    dir_rad = (dir3D + 90) * PI / 180.0;
    sin_dir = sin(dir_rad);
    cos_dir = cos(dir_rad);

    origin_x = low_x + width_m * 0.5;
    origin_y = low_y + height_m * 0.5;
    origin_z = alt3D;

    /* Point on projection plane */
    my_x = height_m * cos(dir_rad) * 0.5;
    my_y = -height_m * sin(dir_rad) * 0.5;
    my_z = 0;

    if (do_drive)
    {
	origin_z =
	  ctdb_lookup_elevation(ctdb,my_x+origin_x,my_y+origin_y) + alt3D;
    }

    /* Distance to simulated eye point */
    eye_dist = sqrt(my_x * my_x + my_y * my_y + my_z * my_z);

    /* Distance back to projection plane to give desired field of view */
    proj_dist = -(width_m * 0.5) / (tan(fov3D * PI / 360.0));

    /* Sin/Cos of the angle between positive Z and the ray to my position */
    sin_alt = sqrt(my_x * my_x + my_y * my_y) / eye_dist;
    cos_alt = my_z / eye_dist;
}
static void
transform_point(pt_in, pt_out)
    float64 pt_in[3];
    float64 pt_out[3];
{
    float64 x,y,z;

    /* Transform the points so that z goes away from the eye, x goes
     * left-right, y goes up-down.
     */

    x = pt_in[X] - origin_x;
    y = pt_in[Y] - origin_y;
    z = pt_in[Z] - origin_z;

    pt_out[X] = x * sin_dir + y * cos_dir;
    pt_out[Y] = -x * cos_dir * cos_alt + y * sin_dir * cos_alt + z * sin_alt;
    pt_out[Z] = eye_dist +
      -x * cos_dir * sin_alt + y * sin_dir * sin_alt - z * cos_alt;
}
static void
squash_point(pt_in, pt_out)
    float64 pt_in[3];
    float64 pt_out[2];
{
    float64 z;

    if (pt_in[Z] < 0.01)
      z = 0.01;
    else
      z = pt_in[Z];

    /* Like the projection on the back of the eyeball, this image is
     * upside down and backwards.  Compensate here.
     */
    pt_out[X] = win_width * 0.5 - proj_dist * pt_in[X] / z;
    pt_out[Y] = win_height * 0.5 + proj_dist * pt_in[Y] / z;
}


static int32
to_pixels(float64 x, 
	  float64 y, 
	  float64 z,
	  float64 inv_scale, 
	  int16 *xresult, 
	  int16 *yresult)
{
    float64 passed[3], eye[3], screen[2];

    if (do_3D)
    {
	passed[X] = x; passed[Y] = y; passed[Z] = z;

	transform_point(passed, eye);
	squash_point(eye, screen);
    }
    else
    {
	screen[X] = (x - low_x) * inv_scale;
	screen[Y] = win_height - 1 - (y - low_y) * inv_scale;
	eye[Z] = 0;
    }

    *xresult = screen[X];
    *yresult = screen[Y];

    if (eye[Z] < 0)
      return 0;

    return ((screen[X] >= 0) && (screen[X] <= win_width) &&
	    (screen[Y] >= 0) && (screen[Y] <= win_height));
}

#define PIXELX_TO_METERS(sx,inv_scale,low) (sx)/(inv_scale) + (low);
#define PIXELY_TO_METERS(sy,inv_scale,low) (low_y - ((sy) - win_height - 1)/(inv_scale));

static void
draw_soils(int32 x0,
	   int32 y0,
	   int32 x1,
	   int32 y1)
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    XRectangle tmp;
    float64 halfw,halfh,mx,my;
    int16 x,y,sx0,sy0,sx1,sy1;
    CTDB_ELEV_DATA elev;
    CTDB_ELEV_QUALIFIER qual=viewer_qual|CTDB_RETURN_MAX;

    tmp.width = tmp.height = main_options.soil_sample.value;    
    halfw = tmp.width/2;
    halfh = tmp.height/2;
    
    to_pixels(x0, y0, 0.0, inv_scale, &sx0, &sy0);
    to_pixels(x1, y1, 0.0, inv_scale, &sx1, &sy1);

    for (x=sx0;x<sx1;x+=tmp.width)
    {
	mx = PIXELX_TO_METERS(x+halfw,inv_scale,low_x);
	
	for (y=sy1;y<sy0;y+=tmp.height)
	{
	    my = PIXELY_TO_METERS(y+halfh,inv_scale,low_y);
	    ctdb_lookup_qual_elevation(ctdb,mx,my,qual,CTDB_NO_REF_ELEV,
				       &elev);
	    XFillRectangle(XtDisplay(area),WINDOW,
			   soil_gcs[CTDB_SIMNET_SOIL(elev.material)],
			   x,y,tmp.width,tmp.height);
	}
    }
}

static void
draw_damage(CTDB *ctdb,    
	    float64 x,
	    float64 y)
{
    int32 patch_number, patchx, patchy, x0, y0, next;
    CTDB_POINT_STATE result = 0; /* really int32 */
    float64 z_result;
    CTDB_TERRAIN_CLASS fclass;
    CTDB_FEATURE_DATA *last, *feature;

    /*find out if we're in a mes or non mes volume*/
    /*if not in a volume, don't do anything*/

    if ( !ctdb_point_on_database(ctdb,x,y) ) 
    {
	printf("(%g,%g,%f) patch number = ILLEGAL\n", x, y,
	       ctdb_lookup_elevation( ctdb, x, y ));		    
    }
    else
    {
	patch_number = ctdb_compute_patch_num(ctdb,x,y,&patchx,&patchy);
	feature = ctdb_lookup_feature(ctdb, patch_number);
    
	if (!feature) 
	  return;
    
	last = feature + feature->hdr.size;
	for(feature++;feature<last;feature+=next)
	{
	    /* Only interested in buildings */
	    fclass = feature->info.feature_class;

	    if ( fclass == CTDB_FC_MICRO )
	      next = NEXT_MICRO(feature->info.vertices);
	    else if ( fclass == CTDB_FC_VOLUME )
	    {
		next = NEXT_VOLUME(feature->info.vertices);
		x0 = CTDB_METERS_TO_XYO(x,patchx,ctdb->meters_to_patch);
		y0 = CTDB_METERS_TO_XYO(y,patchy,ctdb->meters_to_patch);
	    
		result = ctdb_point_in_building(feature, x0, y0, &z_result);
		
		if (result != OUTSIDE) /* then we're in a building */
		{ 
		    if (feature[1].volume.type == CTDB_FS_MES)
		    {
			/* change state */
			int32 model_type; /* really it's CTDB_TERRAIN_SUBCLASS
					   * value of 2 is CTDB_FS_MES
					   */ 
			char *model_name;
			int32 model_index;
			int32 next_model_index; 
			int32 is_mes;
			int16 force_id;
			int16 avoid;
			float32 prev_state;
			float32 next_state;
			float64 position[3];
			float32 rotation[3];
			float32 yaw, pitch, roll;
			float64 *tfm_world_to_mes; 
			CTDB_VOLUME_MODEL *model;
			uint32 which_mes_volume;
		    

			model_type = CTDB_FS_MES;
			is_mes = 1;

			model_index = feature[1].volume.reference;
			model = (CTDB_VOLUME_MODEL *) 
			  ctdb_alloc(NULL, sizeof(CTDB_VOLUME_MODEL),
				     "Volume Model");
			model = ctdb_mlib_model_by_index(ctdb,model_index,
							 is_mes);
			model_name = model->model_name;

			force_id = feature[1].volume.force_id;
			avoid = feature[1].volume.avoid;

			next_model_index = model->next_state_index;
			if (next_model_index == -1)
			{
			    printf("This MES volume is already fully damaged.\n");
			    continue;
			} 
			
			else 
			{
			    if (ctdb->volume_models) 
			      next_state =
				ctdb->volume_models[next_model_index]->percent_damaged;
			    else
			      next_state =  
				ctdb->mes_volume_models[next_model_index]->percent_damaged;
			}
		    
			
			which_mes_volume = model->mes_volume_index;
			if (CTDB_MLIB_VALID_MES_VOLUME_INDEX(ctdb, 
							     which_mes_volume))
			{
			    position[0] = 
			      ctdb->mes_volumes[which_mes_volume].origin[0];
			    position[1] = 
			      ctdb->mes_volumes[which_mes_volume].origin[1];
			    position[2] = 
			      ctdb->mes_volumes[which_mes_volume].origin[2];

			    tfm_world_to_mes =
			      ctdb->mes_volumes[which_mes_volume].tfm_world_to_mes;

			    vmat3_rotation64_to_zdown_angles32(tfm_world_to_mes, 
							   &yaw,&pitch,&roll);
		    
			    rotation[0] = yaw;
			    rotation[1] = pitch;
			    rotation[2] = roll;
			}
			else continue;
		    

			/* prev_state isn't used in modify_volume_feature
			 * but it is printed out if model == NULL.
			 * It is passed to delete_volume_feature but not
			 * used there. I'll set it to 0 for now.
			 */
			prev_state = 0.0; /* it represents a percent damaged 
				       */
			    
			ctdb_modify_volume_feature(ctdb, model_name, 
						   model_type, force_id, 
						   avoid, prev_state,    
						   next_state, position, 
						   rotation);

		    } /*end of if type is CTDB_FS_MES */
		    else
		      printf("Viewer doesn't currently damage non MES buildings.\n");

		} /* end of if (!OUTSIDE) */

	    } /* end of if class == volume */
	    else
	      return;
	} /* end of feature loop */
    }
}

static int32 
clip_poly(
	  int32    n_pts_in,
	  float64 pts_in[][2],
	  int32   *n_pts_out,
	  int16  pts_out[][2])
{
    int32 j,this_in,next_in;
    float64 df, dc;
    float64 left[MAX_POLY_PTS*2+1][2];
    float64 right[MAX_POLY_PTS*2+1][2];
    float64 top[MAX_POLY_PTS*2+1][2];
    int32 l,r,t,p;

#define MIN_X 10.0
#define MAX_X (win_width - 10.0)
#define MIN_Y 10.0
#define MAX_Y (win_height - 10.0)

#define CLIP_SIDE(TEST,n_source,source,n_sink,sink,fixed,fval,changed)	\
    {									\
	for (j=0,n_sink=0;j<n_source;j++)				\
	{								\
	    this_in = source[j][fixed] TEST fval;			\
	    next_in = source[j+1][fixed] TEST fval;			\
									\
	    if (!this_in && !next_in)					\
	      continue;							\
									\
	    if (this_in)						\
	    {								\
		sink[n_sink][0] = source[j][0];				\
		sink[n_sink][1] = source[j][1];				\
		n_sink++;						\
	    }								\
									\
	    if (this_in && next_in)					\
	      continue;							\
									\
	    df = source[j+1][fixed] - source[j][fixed];			\
	    dc = source[j+1][changed] - source[j][changed];		\
									\
	    sink[n_sink][fixed] = fval;					\
	    sink[n_sink][changed] =					\
	      source[j][changed] + (dc / df) * (fval - source[j][fixed]); \
	    n_sink++;							\
	}								\
	if (n_sink < 3)							\
	{								\
	    *n_pts_out = 0;						\
	    return;							\
	}								\
	sink[n_sink][0] = sink[0][0];					\
	sink[n_sink][1] = sink[0][1];					\
    }

    CLIP_SIDE(>, n_pts_in,pts_in, l,left,    0,MIN_X,1);
    CLIP_SIDE(<, l,left,          r,right,   0,MAX_X,1);
    CLIP_SIDE(<, r,right,         t,top,     1,MAX_Y,0);
    CLIP_SIDE(>, t,top,           p,pts_out, 1,MIN_Y,0);

    *n_pts_out = p;
}

static void
transform_poly(int32    n_pts_in,float64 pts_in[][3],int32   *n_pts_out,
	       int16  pts_out[][2])
{
    float64 xformed[MAX_POLY_PTS+1][3];
    float64 clipped[MAX_POLY_PTS*2+1][2];
    float64 dx, dy, dz, inter[3];
    int32 i,j;

    for(i=0;i<n_pts_in;i++)
      transform_point(pts_in[i], xformed[i]);
    bcopy(xformed[X],xformed[n_pts_in],3*sizeof(float64));

    /* Clip negative z values */
    for(i=0,j=0;i<n_pts_in;i++)
    {
	if ((xformed[i][Z] < 0) && (xformed[i+1][Z] < 0))
	  continue;

	if (xformed[i][Z] >= 0)
	{
	    squash_point(xformed[i],clipped[j]);
	    j++;
	}

	if ((xformed[i][Z] >= 0) && (xformed[i+1][Z] >= 0))
	  continue;

	dx = xformed[i+1][X] - xformed[i][X];
	dy = xformed[i+1][Y] - xformed[i][Y];
	dz = xformed[i+1][Z] - xformed[i][Z];

	inter[X] = xformed[i][X] + (dx/dz) * (-xformed[i][Z]);
	inter[Y] = xformed[i][Y] + (dy/dz) * (-xformed[i][Z]);
	inter[Z] = 0.0;

	squash_point(inter,clipped[j]);
	j++;
    }
    if (j < 3)
    {
	*n_pts_out = 0;
	return;
    }

    clipped[j][X] = clipped[0][X];
    clipped[j][Y] = clipped[0][Y];
    clip_poly(j,clipped,n_pts_out,pts_out);

}
static void
interpolate_color(elev, color)
    float64  elev;
    XColor *color;
{
    int32 low;
    float64 cell;
    int32 red, green, blue, bright;

    cell = NCOLORS * (elev - ctdb->min_z) / (ctdb->max_z - ctdb->min_z);
    low = (int32)cell;

    if (low<0) low = 0;
    if (low>NCOLORS-1) low = NCOLORS-1;

    bright = 0x3F00*(cell-low)-0x1F00;

    red = colors[low].red;
    green = colors[low].green;
    blue = colors[low].blue;

#define CCLIP(rgb) \
    if (rgb+bright < 0) color->rgb = 0; \
    else if (rgb+bright > 0xFFFF) color->rgb = 0xFFFF; \
    else color->rgb = rgb+bright;

    CCLIP(red);
    CCLIP(green);
    CCLIP(blue);

    color->flags = DoRed | DoGreen | DoBlue;
}
static void
setup_cmap(low_elev, high_elev)
    float64 low_elev;
    float64 high_elev;
{
    int32 i;
    XColor local_colors[NCELLS];
    float64 step = (high_elev-low_elev)/NCELLS;

    for(i=0;i<NCELLS;i++)
    {
	interpolate_color(low_elev + i*step, &local_colors[i]);
	local_colors[i].pixel = local_cmap[i];
    }
    XStoreColors(XtDisplay(toplevel), cmap, local_colors, NCELLS);
}

static void
setup_gcs()
{
    XColor clr, ign;
    XGCValues gcv;
    int32 i;
    int ign2,ign3;
    Pixmap tree_stipple;
    unsigned int w,h;

    XReadBitmapFile(XtDisplay(area), XtWindow(area),
		    "/usr/include/X11/bitmaps/flipped_gray",
		    &w,&h,&tree_stipple,&ign2,&ign3);

    gcv.foreground = 1;
    gcv.line_width = 0;
    contour_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			   GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "orange red", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 0;
    intervis_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			 GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "Tan", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 0;
    micro_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			 GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "gray50", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 0;
    grid1_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			 GCForeground | GCLineWidth, &gcv);

    gcv.foreground = clr.pixel;
    gcv.line_width = 3;
    grid2_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			 GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "HotPink", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 3;
    abstract_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			    GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "Violet", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 2;
    node_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "Brown", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 0;
    te_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			GCForeground | GCLineWidth, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "Dark Violet", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 2;
    bridge_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			  GCForeground | GCLineWidth, &gcv);

    line_gc = contour_gc;

    XAllocNamedColor(XtDisplay(area), cmap, "darkOliveGreen4", &clr, &ign);
    gcv.foreground = clr.pixel;
    gcv.line_width = 0;
    tree_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			GCForeground | GCLineWidth, &gcv);
    XSetStipple(XtDisplay(area), tree_gc, tree_stipple);

    XAllocNamedColor(XtDisplay(area), cmap, "Black", &clr, &ign);
    gcv.foreground = clr.pixel;
    black_gc = XCreateGC(XtDisplay(area), XtWindow(area),
			 GCForeground | GCBackground, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "gray75", &clr, &ign);
    gcv.foreground = clr.pixel;
    XAllocNamedColor(XtDisplay(area), cmap, "gray25", &clr, &ign);
    gcv.background = clr.pixel;
    hyp_gc = XCreateGC(XtDisplay(area), XtWindow(area),
		       GCForeground | GCBackground, &gcv);


    XAllocNamedColor(XtDisplay(area), cmap, "Green", &clr, &ign);
    gcv.foreground = clr.pixel;
    iv_clr[FULL_VISIBILITY] = clr.pixel;
    iv_gcs[FULL_VISIBILITY] = XCreateGC(XtDisplay(area), XtWindow(area),
					GCForeground, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "Yellow", &clr, &ign);
    gcv.foreground = clr.pixel;
    iv_clr[PARTIAL_VISIBILITY] = clr.pixel;
    iv_gcs[PARTIAL_VISIBILITY] = XCreateGC(XtDisplay(area), XtWindow(area),
					   GCForeground, &gcv);

    XAllocNamedColor(XtDisplay(area), cmap, "Red", &clr, &ign);
    gcv.foreground = clr.pixel;
    iv_clr[NO_VISIBILITY] = clr.pixel;
    iv_gcs[NO_VISIBILITY] = XCreateGC(XtDisplay(area), XtWindow(area),
				      GCForeground, &gcv);

#define MAP_COLOR(_NAME,_GC)\
    XAllocNamedColor(XtDisplay(area), cmap, _NAME, &clr, &ign);\
    gcv.foreground = clr.pixel;\
    XAllocNamedColor(XtDisplay(area), cmap, _NAME, &clr, &ign);\
    gcv.background = clr.pixel;\
    (_GC) = XCreateGC(XtDisplay(area), XtWindow(area),\
		       GCForeground | GCBackground, &gcv);

    /* Unknown type of soil	             */
    MAP_COLOR("dark gray",soil_gcs[SOIL_NON_SOIL]);
    /* Asphalt or other hard surface     */
    MAP_COLOR("gray20",soil_gcs[SOIL_ROAD]);
    /* Packed soil or dirt road          */;
    MAP_COLOR("burlywood4",soil_gcs[SOIL_RCI250]) 
    /* Soft sandy soil                   */;
    MAP_COLOR("burlywood",soil_gcs[SOIL_RCI050]) 
    /* Impassable deep water             */
    MAP_COLOR("blue4",soil_gcs[SOIL_DEEP_WATER]) ;
    /* Passable shallow water            */
    MAP_COLOR("blue2",soil_gcs[SOIL_SHALLOW_WATER]) ;
    /* Muddy soil                        */
    MAP_COLOR("sienna4",soil_gcs[SOIL_MUD]) ;
    /* Wet dirt road                     */
    MAP_COLOR("gray30",soil_gcs[SOIL_MUDDY_ROAD]) ;
    /* Slick ice surface                 */
    MAP_COLOR("aliceblue",soil_gcs[SOIL_ICE]) ;
    /* Very soft surface                 */
    MAP_COLOR("dark green",soil_gcs[SOIL_SWAMP]) ;
    /* Canopy or forrested area          */
    MAP_COLOR("chartreuse3",soil_gcs[SOIL_FORRESTED]) ;
    /* Railroad w/ US specifications     */
    MAP_COLOR("gold1",soil_gcs[SOIL_US_RAILROAD]) ;
    /* Railroad w/ European specs.       */
    MAP_COLOR("gold4",soil_gcs[SOIL_EURO_RAILROAD]) ;
    /* Small rocks <= 18 inches          */
    MAP_COLOR("dark slate gray",soil_gcs[SOIL_ROCKY]) ;
    /* Large boulders 6 ft. high         */
    MAP_COLOR("gray40",soil_gcs[SOIL_BOULDERS]) ;
    /* General No Go Terrain for NTC     */
    MAP_COLOR("crimson",soil_gcs[SOIL_NO_GO]) ;

    for(i=0;i<16;i++)
    {
	clr.red = ((i<<2) + 139) << 8;
	clr.green = ((i<<2) + 69) << 8;
	clr.blue = ((i<<2) + 19) << 8;
	clr.flags = DoRed | DoGreen | DoBlue;
	XAllocColor(XtDisplay(area), cmap, &clr);

	gcv.foreground = clr.pixel;
	shade_gcs[i] = XCreateGC(XtDisplay(area), XtWindow(area),
				GCForeground, &gcv);
    }

    XAllocNamedColor(XtDisplay(area), cmap, "SkyBlue", &clr, &ign);
    gcv.foreground = clr.pixel;
    sky_gc = XCreateGC(XtDisplay(area), XtWindow(area), GCForeground, &gcv);
}
static void
draw_contour_segs(nsegs, segs, elevs)
    int32    nsegs;
    float64 segs[];
    float64 elevs[];
{
    int32 i,p,l;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int16 x0,x1,y0,y1,dx,dy;
    int16 pts[CTDB_MAX_CONTOUR_PTS];
    int16 lines[CTDB_MAX_CONTOUR_PTS];
    GC gc;

    if (do_3D)
      gc = black_gc;
    else
      gc = contour_gc;

    p = 0;
    l = 0;
    for (i=0;i<nsegs;i+=4)
      if (to_pixels(segs[i],segs[i+1],elevs[i>>2],inv_scale,&x0,&y0) |
	  to_pixels(segs[i+2],segs[i+3],elevs[i>>2],inv_scale,&x1,&y1))
      {
	  dx = x1-x0;
	  dy = y1-y0;

	  if ((dx == 0) && (dy == 0))
	    continue;

	  if (dx < 0) dx = -dx;
	  if (dy < 0) dy = -dy;

	  if ((dx<2) && (dy<2))
	  {
	      pts[p] = x0;
	      pts[p+1] = y0;
	      pts[p+2] = x1;
	      pts[p+3] = y1;
	      p += 4;
	  }
	  else
	  {
	      lines[l] = x0;
	      lines[l+1] = y0;
	      lines[l+2] = x1;
	      lines[l+3] = y1;
	      l += 4;
	  }
      }

    XDrawSegments(XtDisplay(area), WINDOW, gc, (XSegment *)lines, l/4);
    XDrawPoints(XtDisplay(area), WINDOW, gc, (XPoint *)pts, p/2,
		CoordModeOrigin);
}
static void
draw_contours(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    ctdb_contour_segments(ctdb,(float64)contour,
			  x0, y0, x1, y1, draw_contour_segs);
}


static void
set_dist(SCR_POLY *poly)
{
    float64 x, y;
    float64 dx, dy;
    float64 dist;
    int32 i;

    for(i=0;i<poly->n_pts;i++)
    {
	x = poly->pts[i][X] - origin_x;
	y = poly->pts[i][Y] - origin_y;

	dx = x - my_x;
	dy = y - my_y;
	dist = (dx * dx + dy * dy);
 
	if (!i || dist < poly->min_dist)
	  poly->min_dist = dist;
	if (!i || dist > poly->max_dist)
	  poly->max_dist = dist;
    }
}    


static void
draw_iv_line(x0, y0, z0, x1, y1, z1, x, y, zl, zh, z, vis)
    int32    x0, y0;
    float64 z0;
    int32    x1, y1;
    float64 z1;
    float64 x,y,zl,zh;
    float64 z;
    int32    vis;
{
    char buf[20];
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int16 px0, py0, px1, py1;

    to_pixels(x0*ctdb->incrD,y0*ctdb->incrD,z0,inv_scale,&px0,&py0);
    to_pixels(x1*ctdb->incrD,y1*ctdb->incrD,z1,inv_scale,&px1,&py1);

    XDrawLine(XtDisplay(area), WINDOW, iv_gcs[vis],
	      px0,py0,px1,py1);

    if (!vis)
    {
	sprintf(buf,"%f",z0);
	XDrawString(XtDisplay(area), WINDOW, micro_gc,
		    px0, py0, buf, strlen(buf));

	sprintf(buf,"%f",z1);
	XDrawString(XtDisplay(area), WINDOW, micro_gc,
		    px1, py1, buf, strlen(buf));

	to_pixels(x*ctdb->incrD,y*ctdb->incrD,zl,inv_scale,&px0,&py0);
	to_pixels(x*ctdb->incrD,y*ctdb->incrD,zh,inv_scale,&px1,&py1);

	XDrawLine(XtDisplay(area), WINDOW, micro_gc,
		  px0,py0,px1,py1);

	sprintf(buf,"%f",zl);
	XDrawString(XtDisplay(area), WINDOW, micro_gc,
		    px0, py0, buf, strlen(buf));

	sprintf(buf,"%f",zh);
	XDrawString(XtDisplay(area), WINDOW, micro_gc,
		    px1, py1, buf, strlen(buf));

	to_pixels(x*ctdb->incrD,y*ctdb->incrD,z,inv_scale,&px0,&py0);
	XDrawArc(XtDisplay(area), WINDOW, micro_gc,
		 px0-5,py0-5,10,10,0,ARC_CONST_360x64);
    }
}
static void
draw_iv_mline(px, py, v0, v1)
    int32               px, py;
    CTDB_FEATURE_DATA v0[2];
    CTDB_FEATURE_DATA v1[2];
{
    float64 x0, y0, z0, x1, y1, z1;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int16 px0, py0, px1, py1;

    x0 = CTDB_XYO_TO_METERS(v0[0].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    y0 = CTDB_XYO_TO_METERS(v0[0].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    z0 = v0[1].z;
    x1 = CTDB_XYO_TO_METERS(v1[0].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    y1 = CTDB_XYO_TO_METERS(v1[0].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    z1 = v1[1].z;

    to_pixels(x0,y0,z0,inv_scale,&px0,&py0);
    to_pixels(x1,y1,z1,inv_scale,&px1,&py1);

    XDrawLine(XtDisplay(area), WINDOW, iv_gcs[NO_VISIBILITY],
	      px0,py0,px1,py1);
}
static void
circle_point(x,y,z)
    float64 x,y,z;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int16 px,py;

    to_pixels(x, y, z, inv_scale, &px, &py);

    XDrawArc(XtDisplay(area), WINDOW, micro_gc,
	     px-5,py-5,10,10,0,ARC_CONST_360x64);
}

static void
circle_post(x,y)
    int32 x,y;
{
    float64 xD, yD, z;

    printf("Post %d %d\n",x,y);

    xD = x*ctdb->incrD;
    yD = y*ctdb->incrD;
    z = ctdb_lookup_elevation(ctdb, xD, yD);
    circle_point(xD,yD,z);
}

#define POLE_HEIGHT 3
#define POLE_WIDTH  4
static void
set_interv_frac(fraction)
    float64 fraction;
{
    sprintf(iv_frac_buf,"%3.2f%%",fraction*100.0);
}


static void
draw_intervis_pole(xl,yl,xh,yh)
    int32 xl,yl,xh,yh;
{
    viewer_visibility v;
    float64 x, y;
    float64 x0,y0, z0, x1, y1, z1;
    float64 fraction, opacity;

    if (!do_3D)
      return;

    if (ctdb_lookup_elevation(ctdb, my_x+origin_x, my_y+origin_y) >=
	origin_z)
    {
	printf("\007Below ground\n");
	return;
    }

    x = (xl + xh) * 0.5;
    y = (yl + yh) * 0.5;
    
    x0 = x + sin_dir * (POLE_WIDTH/2);
    x1 = x - sin_dir * (POLE_WIDTH/2);
    y0 = y + cos_dir * (POLE_WIDTH/2);
    y1 = y - cos_dir * (POLE_WIDTH/2);

    z0 = ctdb_lookup_elevation(ctdb, x, y) + 0.5;
    z1 = z0 + POLE_HEIGHT-0.5;

    fraction = ctdb_point_to_point(ctdb,
				   my_x+origin_x,
				   my_y+origin_y,
				   my_z+origin_z,
				   x, y, z0, z1, 
				   (float64)POLE_WIDTH /* width */,
				   NULL,
				   &opacity,
				   0.05                /* min vis */, 
				   0                   /* num_veh */,
				   (CTDB_VEHICLE_LOCATION *)NULL,
				   0,0);
    fraction += opacity;

    if (fraction == 0.0)
      v = NO_VISIBILITY;
    else if (fraction == 1.0)
      v = FULL_VISIBILITY;
    else
      v = PARTIAL_VISIBILITY;
    set_interv_frac(fraction);

    poly_list[pl_cnt].n_pts = 4;
    poly_list[pl_cnt].pts[0][X] = x0;
    poly_list[pl_cnt].pts[0][Y] = y0;
    poly_list[pl_cnt].pts[0][Z] = z0;
    poly_list[pl_cnt].pts[1][X] = x0;
    poly_list[pl_cnt].pts[1][Y] = y0;
    poly_list[pl_cnt].pts[1][Z] = z1;
    poly_list[pl_cnt].pts[2][X] = x1;
    poly_list[pl_cnt].pts[2][Y] = y1;
    poly_list[pl_cnt].pts[2][Z] = z1;
    poly_list[pl_cnt].pts[3][X] = x1;
    poly_list[pl_cnt].pts[3][Y] = y1;
    poly_list[pl_cnt].pts[3][Z] = z0;

    set_dist(poly_list+pl_cnt);
    poly_list[pl_cnt].gc = iv_gcs[v];
    poly_list[pl_cnt].always_outline = 0;
    poly_list[pl_cnt].priority = INTERV_PRI;
    pl_cnt = (pl_cnt+1)%MAX_POLYS;
}

#define REAL_FAR 1.0E+20

static void
draw_sun()
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 x, y, z;
    int16 px0, py0;
    int16 pts[8];

    if (!do_3D)
      return;

    /* Make sure the sun is in front of us */
    if ((sun_vec[X]*my_x + sun_vec[Y]*my_y + sun_vec[Z]*my_z) > 0)
      return;

    x = sun_vec[X] * REAL_FAR;
    y = sun_vec[Y] * REAL_FAR;
    z = sun_vec[Z] * REAL_FAR;

    to_pixels(x,y,z,inv_scale,&px0,&py0);

    pts[0] = px0-4; pts[1] = py0;
    pts[2] = px0;   pts[3] = py0+4;
    pts[4] = px0+4; pts[5] = py0;
    pts[6] = px0;   pts[7] = py0-4;

    XFillPolygon(XtDisplay(area),WINDOW,iv_gcs[PARTIAL_VISIBILITY],
		 (XPoint *)pts,4, Convex,CoordModeOrigin);
}

static void
draw_los(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 x, y, dx, dy;
    float64 z, dz;
    float64 i;
    int16 px0, py0;

    x = (x0 + x1) * 0.5;
    y = (y0 + y1) * 0.5;
    
    dx = my_x+origin_x - x;
    dy = my_y+origin_y - y;

    for(i=0.0;i<1.0;i+=0.01)
    {
	z = ctdb_lookup_elevation(ctdb, x+i*dx, y+i*dy);
	to_pixels(x+i*dx,y+i*dy,z,inv_scale,&px0,&py0);
	XDrawPoint(XtDisplay(area),WINDOW,grid2_gc, px0, py0);
    }

    if (!do_3D)
      return;
    
    z = ctdb_lookup_elevation(ctdb, x, y) + 0.5;
    for (dz=0;dz<POLE_HEIGHT;dz+=POLE_HEIGHT/5.0)
    {
	to_pixels(x,y,z+dz,inv_scale, &px0, &py0);
	XDrawLine(XtDisplay(area),WINDOW,intervis_gc,px0-2,py0,px0+2,py0);
    }
    XDrawString(XtDisplay(area), WINDOW, black_gc,
		px0+4, py0, iv_frac_buf, strlen(iv_frac_buf));
}

static void
draw_intervis_ray(int32 x0, int32 y0, int32 x1, int32 y1)
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 cx, cy, angle, d, x, y;
    int16 px0, py0, px1, py1;

    cx = (x0 + x1) * 0.5;
    cy = (y0 + y1) * 0.5;
    d = (y1 - y0) * 0.5;

    angle = (dir3D * PI / 180.0);

    x = cx+d*cos(angle);
    y = cy+d*sin(angle);

    to_pixels(cx,cy,1000.0,inv_scale,&px0,&py0);
    to_pixels(x,y,1000.0,inv_scale,&px1,&py1);

    XDrawLine(XtDisplay(area), WINDOW, intervis_gc,
	      px0,py0,px1,py1);
}

static void 
compute_triangle_normal(pts, which_t, pa, pb, pc)
    CTDB_COORD3D *pts;
    int32         which_t;
    float64      *pa, *pb, *pc;
{
    float64 v1x, v1y, v1z;
    float64 v2x, v2y, v2z;

    /* Computes a normal to the microterrain polygon.  Note that the
     * units of *a and *b are the funny units used to store features,
     * NOT meters or posts.
     */
    v1x = (int32)pts[1+which_t].x - (int32)pts[0].x; 
    v1y = (int32)pts[1+which_t].y - (int32)pts[0].y;
    v1z = pts[1+which_t].z - pts[0].z; 
    v2x = (int32)pts[2+which_t].x - (int32)pts[0].x; 
    v2y = (int32)pts[2+which_t].y - (int32)pts[0].y;
    v2z = pts[2+which_t].z - pts[0].z; 
    *pa = v1y*v2z - v1z*v2y;
    *pb = v1z*v2x - v1x*v2z;
    *pc = v1x*v2y - v1y*v2x;
}

static void
add_terrain_polygon(CTDB_COORD3D  *db_pts,
                    int32          n_pts,
                    int32          soil)    
{
    int32 n_poly =(n_pts - 2);
    int32 i;
    float64 a,b,c;
    static viewer_priority pri=1;
    
    if ( n_pts < CTDB_PTS_PER_TRIANGLE )
      return;
    
    for (i=0;i<n_poly;i++)
    {
	poly_list[pl_cnt].pts[0][X] = db_pts[0].x;
	poly_list[pl_cnt].pts[0][Y] = db_pts[0].y;
	poly_list[pl_cnt].pts[0][Z] = db_pts[0].z;
	poly_list[pl_cnt].pts[1][X] = db_pts[2].x;
	poly_list[pl_cnt].pts[1][Y] = db_pts[2].y;
	poly_list[pl_cnt].pts[1][Z] = db_pts[2].z;
	if ( !i )
	{
	    poly_list[pl_cnt].pts[2][X] = db_pts[1].x;
	    poly_list[pl_cnt].pts[2][Y] = db_pts[1].y;
	    poly_list[pl_cnt].pts[2][Z] = db_pts[1].z;
	}
	else
	{
	    poly_list[pl_cnt].pts[2][X] = db_pts[3].x;
	    poly_list[pl_cnt].pts[2][Y] = db_pts[3].y;
	    poly_list[pl_cnt].pts[2][Z] = db_pts[3].z;
	}
	compute_triangle_normal(db_pts,i,&a,&b,&c);
	a /= ctdb->meters_to_patch;
	b /= ctdb->meters_to_patch;
	c /= ctdb->meters_to_patch;
	
	if( c < 0 )
	  c = -c;

	poly_list[pl_cnt].n_pts = CTDB_PTS_PER_TRIANGLE;
	set_dist(poly_list+pl_cnt);
        
        switch (soil)
        {
          case 4:
          case 5:
            pri = LAND_PRI /*WATER_PRI*/;
            poly_list[pl_cnt].gc = soil_gcs[soil];
            break;
            
          default:
            pri = LAND_PRI;
            poly_list[pl_cnt].gc =
              shade_gcs[shade_val(a, b, c, poly_list[pl_cnt].min_dist)];
            break;
        }
        
        poly_list[pl_cnt].always_outline = 0;
        poly_list[pl_cnt].priority = pri;
	pl_cnt = (pl_cnt++)%MAX_POLYS;
    }
}

static void
draw_micro_line(GC gc,int16 px0,int16 py0,int16 px1,int16 py1, int16 px2, int16 py2,
		int32   mask)
{
    char buf[20];
    int32 x,y;

    XDrawLine(XtDisplay(area), WINDOW, gc, px0, py0, px2, py2);
    if (do_masks)
    {
	x = px0 + (px1 - px0)/10 + (px2 - px0)/10;
	y = py0 + (py1 - py0)/10 + (py2 - py0)/10;
	sprintf(buf,"%x",mask);
	XDrawString(XtDisplay(area), WINDOW, micro_gc,
		    x, y, buf, strlen(buf));
    }
}

static void
draw_soil_legend(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 box_start,box_height,squ_width,y_offset,soil_height;
    int32 soil_box_width,soil_box_height,width_off,height_off,squ_height;
    int16 box[CTDB_PTS_PER_QUADRALATERAL][2];
    int16 soil[CTDB_PTS_PER_QUADRALATERAL][2];
    int32 i;
    float64 x,y;
    char buf[20];
    
#define OFFSET_FROM_TOP_OF_AREA 0.05
#define SOIL_FROM_BOX_BOUNDRY 0.005
#define SOIL_BOX_SEPARATION   0.005
    /* Draw Legend Box */
    box[0][X] = box_start = win_width * 0.95 ;
    box[0][Y] = win_height;

    box[1][X] = win_width ;
    box[1][Y] = box[0][Y];

    box[2][X] = box[1][X];
    box[2][Y] = 0;

    box[3][X] = box[0][X];
    box[3][Y] = 0;

    XFillPolygon(XtDisplay(area),
		 WINDOW,
		 sky_gc,
		 (XPoint *)box,CTDB_PTS_PER_QUADRALATERAL, 
		 Convex, CoordModeOrigin);
    XDrawLine(XtDisplay(area), WINDOW , black_gc,
	      box[0][X], box[0][Y],
	      box[1][X], box[1][Y]);
    XDrawLine(XtDisplay(area), WINDOW , black_gc,
	      box[1][X], box[1][Y],
	      box[2][X], box[2][Y]);
    XDrawLine(XtDisplay(area), WINDOW , black_gc,
	      box[2][X], box[2][Y],
	      box[3][X], box[3][Y]);
    XDrawLine(XtDisplay(area), WINDOW , black_gc,
	      box[3][X], box[3][Y],
	      box[0][X], box[0][Y]);

    /* Add the TITLE text */
    sprintf(buf,"SOIL");
    XDrawString(XtDisplay(area), WINDOW, black_gc,
		box_start+5, 20, buf, strlen(buf));

    soil_height= ( win_height ) / (CTDB_N_SOILS+2) ;
    
    soil[0][X] = box_start+5;
    soil[0][Y] = win_height-5 ;
    soil[1][X] = soil[0][X] + (win_width-box_start)* 0.30 ;
    soil[1][Y] = soil[0][Y] ;

    /* Draw Soil Boxes */    
    for (i=0;i<CTDB_N_SOILS;i++,y_offset+=soil_height)
    {
	soil[2][X] = soil[1][X];
	soil[2][Y] = soil[0][Y] - soil_height;
	soil[3][X] = soil[0][X];
	soil[3][Y] = soil[2][Y];
	
	/* Draw soil information */
	XFillPolygon(XtDisplay(area),
		     WINDOW,
		     soil_gcs[i],
		     (XPoint *)soil,CTDB_PTS_PER_QUADRALATERAL, 
		     Convex, CoordModeOrigin);

	XDrawLine(XtDisplay(area), WINDOW , te_gc,
		  soil[0][X], soil[0][Y],
		  soil[1][X], soil[1][Y]);
	XDrawLine(XtDisplay(area), WINDOW , te_gc,
		  soil[1][X], soil[1][Y],
		  soil[2][X], soil[2][Y]);
	XDrawLine(XtDisplay(area), WINDOW , te_gc,
		  soil[2][X], soil[2][Y],
		  soil[3][X], soil[3][Y]);
	XDrawLine(XtDisplay(area), WINDOW , te_gc,
		  soil[3][X], soil[3][Y],
		  soil[0][X], soil[0][Y]);

	/* Add soil number */
	sprintf(buf,"%3d",i);
	XDrawString(XtDisplay(area), WINDOW, te_gc,
		    soil[1][X], (soil[0][Y]+soil[2][Y])/2,
		    buf, strlen(buf));

	soil[0][X] = soil[3][X] ;
	soil[0][Y] = soil[3][Y] ; 
	soil[1][X] = soil[2][X] ;
	soil[1][Y] = soil[2][Y] ;
    } 	  
}    
    
    


static void
draw_terrain_element_markings2D(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 i, j, n_te, patch_number, px, py;
    int32 xpatch_low, ypatch_low, xpatch_high, ypatch_high;
    int16 pts[4*4][2];
    int32 on_screen[CTDB_PTS_PER_TRIANGLE];
    GC gc;
    
    xpatch_low = PATCH_X(x0 * ctdb->meters_to_patch);
    ypatch_low = NS_MAX(PATCH_Y(y0 * ctdb->meters_to_patch),0);
    xpatch_high = NS_MIN(PATCH_X(x1 * ctdb->meters_to_patch + 1), ctdb->patches_wide);
    ypatch_high = NS_MIN(PATCH_Y(y1 * ctdb->meters_to_patch + 1), ctdb->patches_high);
        
    for(py=ypatch_low;py<ypatch_high;py++)
      for(px=xpatch_low;px<xpatch_high;px++)
      {
	  patch_number=py*ctdb->patches_wide+px;
	  while (!ctdb_lookup_tin_polygons(ctdb,patch_number,
					   te_pts,pts_dim,
					   &n_te))
	  {
	      printf("Allocating a larger list for terrain triangles %d\n",
		     n_te);
	      pts_dim = n_te;
	      te_pts = (CTDB_TERRAIN_TRIANGLE *)
		ctdb_alloc(te_pts,sizeof(CTDB_TERRAIN_TRIANGLE)*n_te,
			   "larger terrain triangle list");
	  }
	  
	  for (i=0;i<n_te;i++)
	  {
	      /* convert all points to display coordinate points and
	       * determine if the point is on or off the display
	       */
	      for (j=0;j<CTDB_PTS_PER_TRIANGLE;j++)
	      {
		  on_screen[j] = to_pixels(te_pts[i].vert[j][X], 
					   te_pts[i].vert[j][Y], 
					   0.0, 
					   inv_scale, 
					   &pts[j][X], 
					   &pts[j][Y]);
	      }
	      
		  
	      
	      /* if this is a quadralateral (two triangles) then 
	       * draw the markings in yellow (iv_gcs[PARTIAL_VISIBILITY])
	       * else draw them using bridge_gc
	       */
	      if ( te_pts[i].assoc_triangle != -1 )
		gc = iv_gcs[PARTIAL_VISIBILITY];
	      else
		gc = bridge_gc;

	      if ( on_screen[0] || on_screen[1] )
		  XDrawLine(XtDisplay(area), WINDOW , gc,
			    pts[0][X], pts[0][Y],
			    pts[0][X]+(pts[1][X]-pts[0][X])/4,
			    pts[0][Y]+(pts[1][Y]-pts[0][Y])/4);

	      if ( on_screen[0] || on_screen[2] )  
		XDrawLine(XtDisplay(area), WINDOW , gc, 
			  pts[0][X], pts[0][Y],
			  pts[0][X]+(pts[2][X]-pts[0][X])/4,
			  pts[0][Y]+(pts[2][Y]-pts[0][Y])/4);
	  }
      }
}

static void
draw_terrain_elements(x0,y0,x1,y1,do_soil)
    int32 x0,y0,x1,y1,do_soil;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 i, j, n_te, patch_number, px, py;
    int32 xpatch_low, ypatch_low, xpatch_high, ypatch_high;
    int16 pts[4*4][2];
    int32 on_screen;
    CTDB_COORD3D dbpt[CTDB_PTS_PER_TRIANGLE];

    xpatch_low = PATCH_X(x0 * ctdb->meters_to_patch);
    ypatch_low = NS_MAX(PATCH_Y(y0 * ctdb->meters_to_patch),0);
    xpatch_high = NS_MIN(PATCH_X(x1 * ctdb->meters_to_patch + 1), 
			 ctdb->patches_wide);
    ypatch_high = NS_MIN(PATCH_Y(y1 * ctdb->meters_to_patch + 1), 
			 ctdb->patches_high);

    for(py=ypatch_low;py<ypatch_high;py++)
      for(px=xpatch_low;px<xpatch_high;px++)
      {
          int32 *soilmap;
          
          patch_number = py*ctdb->patches_wide+px;
          soilmap      = ctdb_lookup_soilmap(ctdb,patch_number);
	  
	  while (!ctdb_lookup_tin_polygons(ctdb,patch_number,
					   te_pts,pts_dim,
					   &n_te))
	  {
	      printf("Allocating a larger list for terrain triangles %d\n",
		     n_te);
	      pts_dim = n_te;
	      te_pts = (CTDB_TERRAIN_TRIANGLE *)
		ctdb_alloc(te_pts,sizeof(CTDB_TERRAIN_TRIANGLE)*n_te,
			   "larger terrain triangle list");
	  }

	  if ( !n_te )
	    continue;

	  for (i=0;i<n_te;i++)
	  {
	      /* if any point of the triangle is on the screen render
	       * the polygon.
	       */
	      on_screen = 0;
	      for (j=0;j<CTDB_PTS_PER_TRIANGLE;j++)
	      {
		  if (do_3D)
		  {
		      dbpt[j].x = te_pts[i].vert[j][X];
		      dbpt[j].y = te_pts[i].vert[j][Y];
		      dbpt[j].z = te_pts[i].vert[j][Z];
		  }
		  else
		  {
		      on_screen |= to_pixels(te_pts[i].vert[j][X], 
					     te_pts[i].vert[j][Y], 
					     0.0, 
					     inv_scale, 
					     &pts[j][X], 
					     &pts[j][Y]);
		  }		      
	      }
	      
	      if ( do_3D )
	      {
		  add_terrain_polygon(&dbpt[0],CTDB_PTS_PER_TRIANGLE,
				      te_pts[i].soil);
		  continue;
	      }
	      
	      if ( !on_screen )
		continue;

	      /* Draw terrain element markings */
	      XDrawLine(XtDisplay(area), WINDOW , te_gc,
			pts[0][X], pts[0][Y], pts[1][X], pts[1][Y]);
	      XDrawLine(XtDisplay(area), WINDOW , te_gc,
			pts[1][X], pts[1][Y], pts[2][X], pts[2][Y]);
	      XDrawLine(XtDisplay(area), WINDOW , te_gc,
			pts[2][X], pts[2][Y], pts[0][X], pts[0][Y]);
	  }
      }
}
#define MAX_TRI_PER_QUAD 2

static void
draw_assoc2D(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    int32 n_te;
    int32 patch_number;
    int32 member;
    CTDB_PATCH_GROUP *patch_group;
    int32 px, py;
    int32 xpatch_low, ypatch_low, xpatch_high, ypatch_high;  
    int32 i,k,x,y,tt,n_tri;
    int16 pts[MAX_POLY_PTS*4][2];
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 on_screen=0;
    int32 dim,association;
    float64 vg_space_m;
    uint16 *vg;
    AREA I,VG,T[MAX_TRI_PER_QUAD];

    xpatch_low  = PATCH_X(x0 * ctdb->meters_to_patch);
    ypatch_low  = NS_MAX(PATCH_Y(y0 * ctdb->meters_to_patch),0);
    xpatch_high = NS_MIN(PATCH_X(x1 * ctdb->meters_to_patch + 1), 
                         ctdb->patches_wide);
    ypatch_high = NS_MIN(PATCH_Y(y1 * ctdb->meters_to_patch + 1), 
                         ctdb->patches_high);

    (void) ctdb_init_area_struct(CTDB_PTS_PER_QUADRALATERAL, &VG);
    (void) ctdb_init_area_struct(CTDB_MAX_INTERSECTION_PTS+10, &I);
    (void) ctdb_init_area_struct(CTDB_PTS_PER_TRIANGLE, &T[0]);
    (void) ctdb_init_area_struct(CTDB_PTS_PER_TRIANGLE, &T[1]);
    for(py=ypatch_low;py<ypatch_high;py++)
      for(px=xpatch_low;px<xpatch_high;px++)
      {
          patch_number=py*ctdb->patches_wide+px;
	  while ( !ctdb_lookup_tin_polygons(ctdb,patch_number,
					    te_pts,pts_dim,
					    &n_te))
	  {
	      printf("Allocating a larger list for terrain triangles %d\n",
		     n_te);
	      pts_dim = n_te;
	      te_pts = (CTDB_TERRAIN_TRIANGLE *)
		ctdb_alloc(te_pts,sizeof(CTDB_TERRAIN_TRIANGLE)*pts_dim,
			   "larger terrain triangle list");
	  }

	  if (!n_te)
	    continue;

	  /* Gain access to the virtual grid map */
          member=PG_MEMBER(patch_number);
          patch_group= ctdb->patch_groups + PG_NUMBER(patch_number);
	  dim = patch_group->data_start.tin_data->te_vgrid_dim[member];
	  vg_space_m=(ctdb->incrD*ctdb->patch_incrD)/dim;
	  vg = (uint16*)((CTDB_FEATURE_DATA*)
			 patch_group->data_start.tin_data +
			 patch_group->data_start.tin_data->
			 te_vgrid_offset[member]);

	  for(y=0,i=0;y<dim;y++)
	  {
	      float64 y_m = py*ctdb->patch_incr*ctdb->incr + y*vg_space_m;
	      for(x=0;x<dim;x++,i++)
	      {
		  float64 x_m = px*ctdb->patch_incr*ctdb->incr + x*vg_space_m ;
		  int32 kk;
		  
		  /* Generate the virtual grid square */
		  VG.count = 0;
		  ctdb_add_point(&VG, x_m, y_m);
		  ctdb_add_point(&VG, x_m, y_m+vg_space_m);
		  ctdb_add_point(&VG, x_m+vg_space_m, y_m+vg_space_m);
		  ctdb_add_point(&VG, x_m+vg_space_m, y_m);
		  ctdb_find_area_extents(&VG);
		  
		  /* find the terrain triangle in the list which maps
		   * to the virtual grid
		   */
		  for (kk=0;kk<n_te;kk++)
		    if ( te_pts[kk].index == vg[i] )
		      break;
		  
		  /* terrain element mapped is always either a triangle or
		     polygon, don't need to check point count */

		  /* Generate the first triangle */
		  T[0].count = 0;
		  ctdb_add_point(&T[0], te_pts[kk].vert[0][X], te_pts[kk].vert[0][Y]);
		  ctdb_add_point(&T[0], te_pts[kk].vert[1][X], te_pts[kk].vert[1][Y]);
		  ctdb_add_point(&T[0], te_pts[kk].vert[2][X], te_pts[kk].vert[2][Y]);
		  ctdb_find_area_extents(&T[0]);

		  /* Generate the quadralaterial (if applicable) */
		  kk = te_pts[kk].assoc_triangle;
		  /* Note: The only point which is new is the last pt */
		  if ( kk != -1 )
		  {
		      T[1].count = 0;
		      ctdb_add_point(&T[1], te_pts[kk].vert[0][X], 
				     te_pts[kk].vert[0][Y]);
		      ctdb_add_point(&T[1], te_pts[kk].vert[1][X], 
				     te_pts[kk].vert[1][Y]);
		      ctdb_add_point(&T[1], te_pts[kk].vert[2][X], 
				     te_pts[kk].vert[2][Y]);
		      ctdb_find_area_extents(&T[1]);
		      n_tri = 2 ;
		  }
		  else
		    n_tri = 1;
		  
		  association = 0;
		  for (tt=0;tt<n_tri;tt++)
		  {
		      /* Find intersection between virtual grid and the first
			 triangle */
		      ctdb_find_edge_intersection(&VG, &T[tt], &I);
		      association += I.count;
		      /* if any of the points is on the screen render the poly
		       */
		      on_screen = 0;
		      for (k=0;k<I.count;k++)
			on_screen |= to_pixels(I.pts[k].x, I.pts[k].y, 0.0, 
					       inv_scale,&pts[k][X],
					       &pts[k][Y]);
		      
		      if (on_screen)
			XFillPolygon(XtDisplay(area),WINDOW,shade_gcs[7],
				     (XPoint *)pts,I.count, 
				     Convex, CoordModeOrigin);
		  }
		  if ( !association )
		  {
		      printf("no assoc intersections patch_number=%d, vg=%d\n",
			     patch_number,vg[i]);
		      continue; /* should never happen */
		  }
	  
	      }
	  }
      }
    ctdb_free_area_struct(&I);
    ctdb_free_area_struct(&VG);
    ctdb_free_area_struct(&T[0]);
    ctdb_free_area_struct(&T[1]);
}



static void
add_wall(
	 int32               px, 
	 int32 py,
	 CTDB_FEATURE_DATA v0[],
	 CTDB_FEATURE_DATA v1[]
	 )
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int16 xp0,yp0,xp1,yp1;
    CTDB_COORD3D pt0,pt1;

    pt0.x = CTDB_XYO_TO_METERS(v0->xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    pt0.y = CTDB_XYO_TO_METERS(v0->xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    pt0.z = v0[1].z;
    pt1.x = CTDB_XYO_TO_METERS(v1->xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    pt1.y = CTDB_XYO_TO_METERS(v1->xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    pt1.z = v1[1].z;

    if (!do_3D)
    {
	to_pixels(pt0.x,pt0.y,pt0.z,inv_scale,&xp0,&yp0);
	to_pixels(pt1.x,pt1.y,pt1.z,inv_scale,&xp1,&yp1);
	XDrawLine(XtDisplay(area), WINDOW, black_gc,
		  xp0,yp0,xp1,yp1);
	return;
    }

    poly_list[pl_cnt].n_pts = 4;
    poly_list[pl_cnt].pts[0][X] = pt0.x;
    poly_list[pl_cnt].pts[0][Y] = pt0.y;
    poly_list[pl_cnt].pts[0][Z] = ctdb_lookup_elevation(ctdb, pt0.x, pt0.y);
    poly_list[pl_cnt].pts[1][X] = pt0.x;
    poly_list[pl_cnt].pts[1][Y] = pt0.y;
    poly_list[pl_cnt].pts[1][Z] = pt0.z;
    poly_list[pl_cnt].pts[2][X] = pt1.x;
    poly_list[pl_cnt].pts[2][Y] = pt1.y;
    poly_list[pl_cnt].pts[2][Z] = pt1.z;
    poly_list[pl_cnt].pts[3][X] = pt1.x;
    poly_list[pl_cnt].pts[3][Y] = pt1.y;
    poly_list[pl_cnt].pts[3][Z] = ctdb_lookup_elevation(ctdb, pt1.x, pt1.y);

    set_dist(poly_list+pl_cnt);
    poly_list[pl_cnt].gc = black_gc;
    poly_list[pl_cnt].always_outline = 1;
    poly_list[pl_cnt].priority = BUILDING_PRI;
    pl_cnt = (pl_cnt+1)%MAX_POLYS;
}

static void
add_tree(
	 int32               px, 
	 int32 py,
	 float64             radius, 
	 float64 height,
	 CTDB_FEATURE_DATA   v[],
	 int32               on_line
	 )
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 x,y,x0,x1,y0,y1,zl,zh;
    int16 xp0,yp0, w;

    x = CTDB_XYO_TO_METERS(v->xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    y = CTDB_XYO_TO_METERS(v->xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    zl = v[1].z;
    zh = v[1].z-height;

    if (!do_3D)
    {
	w = radius * inv_scale;
	to_pixels(x,y,zl,inv_scale,&xp0,&yp0);
	if (w < 1)
	  XDrawPoint(XtDisplay(area), WINDOW, tree_gc, xp0, yp0);
	else if (w < 3)
	  XFillRectangle(XtDisplay(area), WINDOW, tree_gc, xp0-w, yp0-w,
			 w<<1,w<<1);
	else if (on_line)
	  XDrawArc(XtDisplay(area), WINDOW, tree_gc, xp0-w, yp0-w,
		   w<<1,w<<1,0,ARC_CONST_360x64);
	else
	  XFillArc(XtDisplay(area), WINDOW, tree_gc, xp0-w, yp0-w,
		   w<<1,w<<1,0,ARC_CONST_360x64);
	return;
    }

    x0 = x + sin_dir * radius;
    x1 = x - sin_dir * radius;
    y0 = y + cos_dir * radius;
    y1 = y - cos_dir * radius;

    poly_list[pl_cnt].n_pts = 4;
    poly_list[pl_cnt].pts[0][X] = x0;
    poly_list[pl_cnt].pts[0][Y] = y0;
    poly_list[pl_cnt].pts[0][Z] = zl;
    poly_list[pl_cnt].pts[1][X] = x0;
    poly_list[pl_cnt].pts[1][Y] = y0;
    poly_list[pl_cnt].pts[1][Z] = zh;
    poly_list[pl_cnt].pts[2][X] = x1;
    poly_list[pl_cnt].pts[2][Y] = y1;
    poly_list[pl_cnt].pts[2][Z] = zh;
    poly_list[pl_cnt].pts[3][X] = x1;
    poly_list[pl_cnt].pts[3][Y] = y1;
    poly_list[pl_cnt].pts[3][Z] = zl;

    set_dist(poly_list+pl_cnt);
    poly_list[pl_cnt].gc = tree_gc;
    poly_list[pl_cnt].always_outline = 0;
    poly_list[pl_cnt].priority = TREE_PRI;
    pl_cnt = (pl_cnt+1)%MAX_POLYS;
}

static void
add_laid_linear(int32 px, 
		int32 py, 
		float64 width, 
		int32 vert,
		CTDB_FEATURE_DATA v[]
		)
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 i;
    float64 x0,y0;
    int16 pts[32];

    for(i=0;i<vert;i++)
    {
	x0 = CTDB_XYO_TO_METERS(v[i].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	y0 = CTDB_XYO_TO_METERS(v[i].xy.y,py,ctdb->patch_incrD,ctdb->incrD);

	to_pixels(x0,y0,0.0,inv_scale,pts+(i<<1),pts+(i<<1)+1);
    }
    XDrawLines(XtDisplay(area), WINDOW, grid1_gc,
	       (XPoint *)pts, vert, CoordModeOrigin);
    XFillArc(XtDisplay(area), WINDOW, grid1_gc,
	     pts[0]-1, pts[1]-1, 3, 3, 0, ARC_CONST_360x64);
}

static void
add_abstract(
	     int32               code, 
	     int32               count, 
	     float32            *verts, 
	     int32               data_size, 
	     CTDB_ABSTRACT_DATA *data
	     )
{
    int32 i;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    char buf[20];
    int16 pts[10240];

    /* Draw the outline */
    for(i=0;i<count;i++)
    {
	int32 i2=i<<1;
	to_pixels(verts[i2],verts[(i2)+1],0.0,inv_scale,pts+i2,pts+i2+1);
    }
    
    if (count == 1)
      XFillArc(XtDisplay(area), WINDOW, abstract_gc,
	       pts[0]-3, pts[1]-3, 6, 6, 0, ARC_CONST_360x64);
    else
      XDrawLines(XtDisplay(area), WINDOW, abstract_gc,
		 (XPoint *)pts, count, CoordModeOrigin);

    sprintf(buf,"(%d)",code);
    XDrawString(XtDisplay(area), WINDOW, abstract_gc,
		pts[0], pts[1], buf, strlen(buf));
}

static void
add_node(node, which)
    CTDB_NODE *node;
    int32 which;
{
    int32 i, v;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    GC gc;
    float64 dx,dy;
    char buf[20];
    int16 pts[256];

    if (node->is_bridge)
      gc = bridge_gc;
    else
      gc = node_gc;

    for(i=v=0;i<node->num_connections;i++)
    {
	dx = (ctdb->nodes[node->connection[i].node_index]->x - node->x) * 0.1;
	dy = (ctdb->nodes[node->connection[i].node_index]->y - node->y) * 0.1;

	if (to_pixels(node->x,node->y,0.0,inv_scale,pts+v,pts+v+1) &
	    to_pixels(node->x+dx,node->y+dy,0.0,inv_scale,pts+v+2,pts+v+3))
	  v+=4;
    }
    if (v)
    {
	XDrawSegments(XtDisplay(area), WINDOW, gc, (XSegment *)pts, v/4);
	sprintf(buf,"%d",which);
	XDrawString(XtDisplay(area), WINDOW, gc,
		    pts[0], pts[1], buf, strlen(buf));
    }
}

static void
add_edge(which)
    int32 which;
{
    int32 i, v;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    GC gc;
    float64 x, y;
    char buf[20];
    int16 pts[1024];
    int32 n_verts;
    float64 *verts;

    gc = bridge_gc;

    n_verts = ctdb_expand_edge(ctdb, which, 0, &verts, NULL, NULL);

    for(i=v=0;i<(n_verts<<1);i+=2)
    {
	x = verts[i];
	y = verts[i+1];

	to_pixels(x,y,0.0,inv_scale,pts+v,pts+v+1);
	v+=2;
    }
    if (v)
    {
	XDrawLines(XtDisplay(area), WINDOW, gc, 
		   (XPoint *)pts, v/2, CoordModeOrigin);
	sprintf(buf,"%d",which);
	XDrawString(XtDisplay(area), WINDOW, gc,
		    pts[2], pts[3], buf, strlen(buf));
    }

    /* Note: The following should never happen.  This is displayed
     *       to help find errors in encoding.
     */

    gc = iv_gcs[NO_VISIBILITY];

    x = ctdb->nodes[ctdb->edges[which]->first_node]->x;
    y = ctdb->nodes[ctdb->edges[which]->first_node]->y;
    if ((x != verts[0]) || (y != verts[1]))
    {
	to_pixels(x,y,0.0,inv_scale,pts+v,pts+v+1);
	XDrawLine(XtDisplay(area), WINDOW, gc,
		  pts[0], pts[1], pts[v], pts[v+1]);
    }

    x = ctdb->nodes[ctdb->edges[which]->last_node]->x;
    y = ctdb->nodes[ctdb->edges[which]->last_node]->y;
    if ((x != verts[2*n_verts-2]) || (y != verts[2*n_verts-1]))
    {
	to_pixels(x,y,0.0,inv_scale,pts+v,pts+v+1);
	XDrawLine(XtDisplay(area), WINDOW, gc,
		  pts[v-2], pts[v-1], pts[v], pts[v+1]);
    }
}

static void
add_treeline(px,py,height,v0,v1, in_canopy, canopy_edge)
    int32               px, py;
    float64             height;
    CTDB_FEATURE_DATA      v0[];
    CTDB_FEATURE_DATA      v1[];
    int32               canopy_edge;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 x0,y0,z0,x1,y1,z1,g0,g1;
    int16 xp0,yp0,xp1,yp1;

    x0 = CTDB_XYO_TO_METERS(v0->xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    y0 = CTDB_XYO_TO_METERS(v0->xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    z0 = v0[1].z;
    x1 = CTDB_XYO_TO_METERS(v1->xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    y1 = CTDB_XYO_TO_METERS(v1->xy.y,py,ctdb->patch_incrD,ctdb->incrD);
    z1 = v1[1].z;

    if (!do_3D)
    {
	to_pixels(x0,y0,z0,inv_scale,&xp0,&yp0);
	to_pixels(x1,y1,z1,inv_scale,&xp1,&yp1);
	if (canopy_edge)
	{
	    XDrawLine(XtDisplay(area), WINDOW, iv_gcs[NO_VISIBILITY],
		      xp0,yp0,xp1,yp1);
	}
	else
	  XDrawLine(XtDisplay(area), WINDOW, tree_gc,
		    xp0,yp0,xp1,yp1);
	return;
    }

    if (in_canopy && !canopy_edge)
      return;

    g0 = z0 - height;
    g1 = z1 - height;

    poly_list[pl_cnt].n_pts = 4;
    poly_list[pl_cnt].pts[0][X] = x0;
    poly_list[pl_cnt].pts[0][Y] = y0;
    poly_list[pl_cnt].pts[0][Z] = g0;
    poly_list[pl_cnt].pts[1][X] = x0;
    poly_list[pl_cnt].pts[1][Y] = y0;
    poly_list[pl_cnt].pts[1][Z] = z0;
    poly_list[pl_cnt].pts[2][X] = x1;
    poly_list[pl_cnt].pts[2][Y] = y1;
    poly_list[pl_cnt].pts[2][Z] = z1;
    poly_list[pl_cnt].pts[3][X] = x1;
    poly_list[pl_cnt].pts[3][Y] = y1;
    poly_list[pl_cnt].pts[3][Z] = g1;

    set_dist(poly_list+pl_cnt);
    poly_list[pl_cnt].gc = tree_gc;
    poly_list[pl_cnt].always_outline = (canopy_edge != 0);
    poly_list[pl_cnt].priority = TREE_PRI;
    pl_cnt = (pl_cnt+1)%MAX_POLYS;
}

static void
add_mask(
	 int32               px, 
	 int32               py,
	 CTDB_FEATURE_DATA   v[],
	 int32               mask,
	 int32               yoff
	 )
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 x,y;
    int16 xp,yp;
    char buf[20];

    x = CTDB_XYO_TO_METERS(v->xy.x,px,ctdb->patch_incrD,ctdb->incrD);
    y = CTDB_XYO_TO_METERS(v->xy.y,py,ctdb->patch_incrD,ctdb->incrD);

    to_pixels(x,y,0.0,inv_scale,&xp,&yp);

    sprintf(buf,"%x",mask);
    XDrawString(XtDisplay(area), WINDOW, micro_gc,
		xp, yp+yoff, buf, strlen(buf));
}

static int32
view_draw_water_micro3d(CTDB               *ctdb,
                        CTDB_FEAT_TYPE_PTR  sub, 
                        CTDB_FEATURE_DATA  *feature, 
                        CTDB_VIEW_OBJ_DATA *data 
                        )
{
    int32 px=data->patch.x;
    int32 py=data->patch.y;
    int16 on_screen;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 g;
    int16 px0,py0,px1,py1,px2,py2;
    float64 x,y;
    GC gc;
    CTDB_COORD3D pt[3];
    float64 tide;
    CTDB_FEATURE_DATA *featptr;
    
    if ( sub->verts < 2 )
      return VIS_KEEP_GOING;

    tide = ctdb_lookup_micro_tide(ctdb,data->patch.number,
                                  feature[5].mwater.char0);
    
    pt[0].x = CTDB_XYO_TO_METERS(feature[2].xy.x,px,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[0].y = CTDB_XYO_TO_METERS(feature[2].xy.y,py,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[0].z = feature[3].z + tide;

    g = sub->incr*feature[4].micro.vert0;
    
    pt[1].x = CTDB_XYO_TO_METERS(feature[g+2].xy.x,px,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[1].y = CTDB_XYO_TO_METERS(feature[g+2].xy.y,py,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[1].z = feature[g+3].z + tide;
    
    g = sub->incr*feature[4].micro.vert1;
    
    pt[2].x = CTDB_XYO_TO_METERS(feature[g+2].xy.x,px,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[2].y = CTDB_XYO_TO_METERS(feature[g+2].xy.y,py,ctdb->patch_incrD,
                           ctdb->incrD);
    pt[2].z = feature[g+3].z + tide;

    add_terrain_polygon(&pt[0],CTDB_PTS_PER_TRIANGLE,SOIL_DEEP_WATER);
    
    if ( sub->verts < 3 )
      return VIS_KEEP_GOING;
    
    tide = ctdb_lookup_micro_tide(ctdb,data->patch.number,
                                                    feature[5].mwater.char1);
    pt[0].z = feature[3].z + tide;
    pt[1].z = feature[3+g].z + tide;
    
    g = sub->incr*feature[4].micro.vert2;
    
    pt[2].x = CTDB_XYO_TO_METERS(feature[g+2].xy.x,px,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[2].y = CTDB_XYO_TO_METERS(feature[g+2].xy.y,py,ctdb->patch_incrD,
                                 ctdb->incrD);
    pt[2].z = feature[3+g].z + tide;

    add_terrain_polygon(&pt[0],CTDB_PTS_PER_TRIANGLE,SOIL_DEEP_WATER);
    
    return VIS_KEEP_GOING;   
}

static int32 
view_draw_standard_micro3d(CTDB               *ctdb,
                           CTDB_FEAT_TYPE_PTR  sub, 
                           CTDB_FEATURE_DATA  *feature, 
                           CTDB_VIEW_OBJ_DATA *data
                           )
{
    add_feature_polygons(data->patch.x,data->patch.y,feature,sub->incr);
    return VIS_KEEP_GOING;
}

static int32 
view_draw_standard_micro2d(CTDB               *ctdb,
                           CTDB_FEAT_TYPE_PTR  sub, 
                           CTDB_FEATURE_DATA  *feature, 
                           CTDB_VIEW_OBJ_DATA *data
                           )
{
    int32 px=data->patch.x;
    int32 py=data->patch.y;
    int16 on_screen;
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int32 g;
    int16 px0,py0,px1,py1,px2,py2;
    float64 x,y;
    GC gc;
    uint32  postmask = feature->info.post_mask;

    switch (feature->info.vertices & MICRO_TYPE)
    {
      case MICRO_TYPE_BASE: 
      case MICRO_TYPE_DEFAULT: 
        gc = micro_gc;
        if (!do_micro) 
          return VIS_KEEP_GOING;
        break;
      case MICRO_TYPE_ML: 
        gc = iv_gcs[FULL_VISIBILITY];
        if (!do_ml_micro) 
          return VIS_KEEP_GOING;
        break;
      case MICRO_TYPE_WATER: 
        gc = iv_gcs[PARTIAL_VISIBILITY];
        if (!do_water) 
          return VIS_KEEP_GOING;
        break;
    }
        
    x = CTDB_XYO_TO_METERS(feature[2].xy.x,px,ctdb->patch_incrD,
                           ctdb->incrD);
    y = CTDB_XYO_TO_METERS(feature[2].xy.y,py,ctdb->patch_incrD,
                           ctdb->incrD);
    
    on_screen = to_pixels(x, y, 0.0, inv_scale, &px0, &py0);
    
    g = sub->incr*feature[4].micro.vert0;
    
    x = CTDB_XYO_TO_METERS(feature[g+2].xy.x,px,ctdb->patch_incrD,
                           ctdb->incrD);
    y = CTDB_XYO_TO_METERS(feature[g+2].xy.y,py,ctdb->patch_incrD,
                           ctdb->incrD);
    
    if (on_screen | to_pixels(x, y, 0.0, inv_scale, &px1, &py1))
    {
	XDrawLine(XtDisplay(area), WINDOW, gc,
		  px0, py0, px1, py1);
	XDrawLine(XtDisplay(area), WINDOW, iv_gcs[NO_VISIBILITY],
		  px0, py0, px0+((px1-px0)>>2), py0+((py1-py0)>>2));
    }
    
    if ( sub->verts < 2 )
      return VIS_KEEP_GOING;
    
    g = sub->incr*feature[4].micro.vert1;
    
    x = CTDB_XYO_TO_METERS(feature[g+2].xy.x,px,ctdb->patch_incrD,
                           ctdb->incrD);
    y = CTDB_XYO_TO_METERS(feature[g+2].xy.y,py,ctdb->patch_incrD,
                           ctdb->incrD);
    
    if (on_screen | to_pixels(x, y, 0.0, inv_scale, &px2, &py2))
    {
        draw_micro_line(gc, px0, py0, px1, py1, px2, py2, postmask);
        XDrawLine(XtDisplay(area), WINDOW, iv_gcs[NO_VISIBILITY],
                  px0, py0, px0+((px2-px0)>>2), py0+((py2-py0)>>2));
        
    }
    
    if ( sub->verts < 3 )
      return VIS_KEEP_GOING;
    
    g = sub->incr*feature[4].micro.vert2;
    
    x = CTDB_XYO_TO_METERS(feature[g+2].xy.x,px,ctdb->patch_incrD,
                           ctdb->incrD);
    y = CTDB_XYO_TO_METERS(feature[g+2].xy.y,py,ctdb->patch_incrD,
                           ctdb->incrD);
    
    if (on_screen | to_pixels(x, y, 0.0, inv_scale, &px2, &py2))
    {
	draw_micro_line(gc, px0, py0, px1, py1, px2, py2, postmask);
	XDrawLine(XtDisplay(area), WINDOW, iv_gcs[NO_VISIBILITY],
		  px0, py0, px0+((px2-px0)>>2), py0+((py2-py0)>>2));
    }
    return VIS_KEEP_GOING;
}

static int32 
view_draw_water_micro2d(CTDB               *ctdb,
                        CTDB_FEAT_TYPE_PTR  sub, 
                        CTDB_FEATURE_DATA  *feature, 
                        CTDB_VIEW_OBJ_DATA *data
                        )
{
    
    if (do_water)
      return view_draw_standard_micro2d(ctdb,sub,feature,data);

    return VIS_KEEP_GOING;
}

static int32
view_draw_micro(CTDB               *ctdb,
                CTDB_FEAT_OBJ_PTR   obj, 
                CTDB_FEATURE_DATA  *feature, 
                CTDB_VIEW_OBJ_DATA *data,
                int32              *next
                )
{
    static CTDB_FEAT_TYPE *sub=view_micro_subclass_obj;
    int32 verts = feature->info.vertices;

    /* check if the micro type has changed */
    if ((verts & MICRO_TYPE) != sub->type)
      sub = &view_micro_subclass_obj[MICRO_TYPE_INDEX(verts)];

    *next = sub->incr;
    
    if (!obj->valid ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return VIS_KEEP_GOING;

    /* call the micro type processing function */
    return sub->func(ctdb,sub,feature,data);
}

static void
add_mes_polygons(CTDB_MES_POLYGON *poly,
		 int32   n_poly, 
		 GC      gc)
{    
    int32 i,j;
    float64 ray[XYZ],eye[XYZ];
    eye[X] = my_x, eye[Y] = my_y,eye[Z] = my_z;
    
    /* If the eye can see the point then add it to the poly list. Otherwise
     * discard the triangle.
     */
    
    for (i=0;i<n_poly;i++)
    {
	if ( vmat3_dot_prod64(poly[i].normal, eye) >= 0 ) 
	{
	    for (j=0;j<poly[i].num_vertices;j++)
	    {
		poly_list[pl_cnt].pts[j][X] = poly[i].vertices[j][X];
		poly_list[pl_cnt].pts[j][Y] = poly[i].vertices[j][Y];
		poly_list[pl_cnt].pts[j][Z] = poly[i].vertices[j][Z];
	    }

	    poly_list[pl_cnt].n_pts          = poly[i].num_vertices;
	    poly_list[pl_cnt].gc             = gc;
	    poly_list[pl_cnt].always_outline = 0;
	    poly_list[pl_cnt].priority       = BUILDING_PRI;
	    set_dist(&poly_list[pl_cnt]);
	    pl_cnt                             = (pl_cnt++)%MAX_POLYS;
	}
    }
}


static void
add_mes(CTDB               *ctdb,
	CTDB_VIEW_OBJ_DATA *data,
	CTDB_FEATURE_DATA  *feature)
{
    CTDB_MES_VOLUME   *mes;
    uint16             mes_id,encl_id,i;
    CTDB_MES_POLYGON  *poly=NULL;
    int32              n_poly=0;
    CTDB_MES_NODE_KEY  mes_key;
    CTDB_MES_NODE_KEY  encl_key;
    
    mes_id = feature[1].volume.reference;
    CTDB_MES_ID_TO_MES_KEY(mes_id, mes_key);

    /* If the walking point is within the walls of the MES print a message
     * but continue processing.
     */
    if (!ctdb_mes_find_enclosure(ctdb,mes_key,
				 my_x, my_y, my_z,
				 &encl_key))
    {
	printf("Currently between WALLS - placing into default enclosure\n");
	encl_id = 0;
    }
    
    encl_id = CTDB_MES_NODE_KEY_TO_ENCLOSURE_ID(encl_key);
    mes     = &ctdb->mes_volumes[mes_id];
#if 1
    /* render all of the enclosure triangles */
    for (i=0;i<mes->template->num_enclosures;i++)
    {
	if (poly)
	  STDDEALLOC(poly), poly=NULL;
	n_poly=0;
	
	if ( i != encl_id )
	{
	    CTDB_MES_NODE_KEY node_key = CTDB_MES_KEY(mes_id, i);
	    ctdb_mes_get_enclosure_polygons(ctdb, node_key, &n_poly, &poly);
	    add_mes_polygons(poly,n_poly,soil_gcs[SOIL_ROAD]);
	}
    }
#endif
    /* render the current enclosure that contains the point */
    {
	int32 save_cnt=mes_pl_cnt;
	int32 del_cnt;
	CTDB_MES_NODE_KEY node_key = CTDB_MES_KEY(mes_id, encl_id);
	
	if (poly)
	  STDDEALLOC(poly), poly=NULL;
	n_poly=0;
	
	ctdb_mes_get_enclosure_polygons(ctdb, node_key, &n_poly, &poly);
					
	add_mes_polygons(poly,n_poly,soil_gcs[SOIL_DEEP_WATER]);
    }
}

static int32
view_draw_volume_2d(
		     CTDB               *ctdb,
		     CTDB_FEAT_OBJ_PTR   obj, 
		     CTDB_FEATURE_DATA  *feature, 
		     CTDB_VIEW_OBJ_DATA *data,
		     int32              *next
		     )
{
    static int32 verts,j;

    verts = feature->info.vertices;
    *next = NEXT_VOLUME(verts);

    if ( !obj->valid )
       return VIS_KEEP_GOING;

    if ((feature[1].volume.type == CTDB_FS_MES) && do_3D)
    {
	add_mes(ctdb,data,feature);
    }
    else
    {
	feature += 3;
	add_wall(data->patch.x,data->patch.y,feature,feature+((verts-1)<<1));
	for(j=0;j<verts-1;j++,feature+=2)
	  add_wall(data->patch.x,data->patch.y,feature,feature+2);
    }
    
    return VIS_KEEP_GOING;
}

static int32
view_draw_laid_linear_2d(
		     CTDB               *ctdb,
		     CTDB_FEAT_OBJ_PTR   obj, 
		     CTDB_FEATURE_DATA  *feature, 
		     CTDB_VIEW_OBJ_DATA *data,
		     int32              *next
		     )
{
    int32 verts = feature->info.vertices;
    float64 width;

    *next = NEXT_LAID_LINEAR(verts);

    if ( !obj->valid )
      return VIS_KEEP_GOING;
    
    width = CTDB_XYO_TO_METERS(feature[2].laid_linear.width,0,
			       ctdb->patch_incrD,ctdb->incrD);

    add_laid_linear(data->patch.x, data->patch.y, width, verts, feature+3);
    return VIS_KEEP_GOING;
}

static int32
view_draw_canopy_2d(
		     CTDB               *ctdb,
		     CTDB_FEAT_OBJ_PTR   obj, 
		     CTDB_FEATURE_DATA  *feature, 
		     CTDB_VIEW_OBJ_DATA *data,
		     int32              *next
		     )
{
    static int32 ray_info,g;
    static int32 verts, incr;
    static float64 height;
    

    verts = feature->info.vertices;
    incr  = *next = NEXT_CANOPY(verts);

    if ( !obj->valid || !verts)
       return VIS_KEEP_GOING;

    if (do_3D)
    {
	add_feature_polygons(data->patch.x,data->patch.y,feature, incr);
    }

    height = 20.0;
    
    g = incr*feature[3].canopy.vert0;
    add_treeline(data->patch.x,data->patch.y,height,feature+1, feature+g+1,1,
		 feature[3].canopy.vert0_outer);
    
    if (verts == 1) 
      return VIS_KEEP_GOING;
    
    g = incr*feature[3].canopy.vert1;
    add_treeline(data->patch.x,data->patch.y,height,feature+1,feature+g+1,1,
		 feature[3].canopy.vert1_outer);
    
    if (verts == 2)
      return VIS_KEEP_GOING;
    
    g = incr*feature[3].canopy.vert2;
    add_treeline(data->patch.x,data->patch.y,height,feature+1,feature+g+1,1,
			   feature[3].canopy.vert2_outer);
    return VIS_KEEP_GOING;
}

static int32
view_draw_linear_2d(
		     CTDB               *ctdb, 
		     CTDB_FEAT_OBJ_PTR   obj, 
		     CTDB_FEATURE_DATA  *feature, 
		     CTDB_VIEW_OBJ_DATA *data,
		     int32              *next
		     )
{
    static CTDB_LINEAR_MODEL *model;
    int32 firstT, lastT;
    float64 radius, height;
    int32 px = data->patch.x;
    int32 py = data->patch.y;
    int32 verts;

    verts = feature->info.vertices;
    *next = NEXT_LINEAR(verts);
    model = &ctdb->linear_models[feature[1].linear.reference];
    
    /* $$$ this logic only does trees */    
    if ( !obj->valid || (model->type == CTDB_FS_DRAGONS_TEETH) )
       return VIS_KEEP_GOING;

    if ( !do_3D && do_masks )
      add_mask(px, py, feature+2,feature->info.post_mask,(verts==1 ? 0 : 20));

    radius = CTDB_XYO_TO_METERS(model->width * 0.5 , 0,
				ctdb->patch_incrD,ctdb->incrD);
    height = model->linear_data.treeline_info.foliage_height *
	CTDB_TREE_HEIGHT_BASIS;

    firstT = feature[1].linear.first;
    lastT  = feature[1].linear.last;
    feature   += 2;
    
    if (firstT)
      add_tree(px,py,radius,height,feature,(verts>1));
    
    if (verts > 1)
    {
	int32 j;
	
	if (lastT)
	  add_tree(px,py,radius,height,feature+((verts-1)<<1),1);
	
	for(j=0;j<verts-1;j++,feature+=2)
	  add_treeline(px,py,height,feature,feature+2, 0,0);
    }

    return VIS_KEEP_GOING;
}

  
static void
feature_loop(
	     int32 x0,
	     int32 y0,
	     int32 x1,
	     int32 y1
	     )
{
    CTDB_FEATURE_DATA *feature, *last;
    int32 next;
    int32 xpatch_low, ypatch_low, xpatch_high, ypatch_high;
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_VIEW_OBJ_DATA data;
    float64 to_meters=ctdb->incrD*ctdb->patch_incr;

    xpatch_low = x0 * ctdb->inv_incr * ctdb->patch_inv_incr;
    ypatch_low = y0 * ctdb->inv_incr * ctdb->patch_inv_incr;
    xpatch_high = x1 * ctdb->inv_incr * ctdb->patch_inv_incr + 1;
    ypatch_high = y1 * ctdb->inv_incr * ctdb->patch_inv_incr + 1;

    for(data.patch.y=ypatch_low;data.patch.y<ypatch_high;data.patch.y++)
      for(data.patch.x=xpatch_low;data.patch.x<xpatch_high;data.patch.x++)
      {
	  data.patch.number = data.patch.y * ctdb->patches_wide 
	    + data.patch.x;
	  
          data.patch.member = PG_MEMBER(data.patch.number);
          data.patch.pg     = ctdb->patch_groups+PG_NUMBER(data.patch.number);

	  /* Only rendering patch water in 3D since the water information
	   * can be seen through the use of the soils button.
	   */
	  if (do_3D && 
	      CTDB_HAS_WATER_PATCH(data.patch.pg,data.patch.member) )
	  {
	      CTDB_COORD3D pt[3];
	      float32 level;
	      
	      pt[0].x = data.patch.x * to_meters;
	      pt[0].y = data.patch.y * to_meters;
	      ctdb_lookup_patch_water_level(ctdb,
					    data.patch.member,
					    pt[0].x*ctdb->patch_inv_incr,
					    pt[0].y*ctdb->patch_inv_incr,
					    &level);
	      pt[0].z=level;

	      pt[1].x = pt[0].x + to_meters;
	      pt[1].y = pt[0].y + to_meters;
	      ctdb_lookup_patch_water_level(ctdb,
					    data.patch.member,
					    pt[1].x*ctdb->patch_inv_incr,
					    pt[1].y*ctdb->patch_inv_incr,
					    &level);
	      pt[1].z=level;
	      
	      pt[2].x = pt[0].x;
	      pt[2].y = pt[0].y + to_meters;
	      ctdb_lookup_patch_water_level(ctdb,
					    data.patch.member,
					    pt[2].x*ctdb->patch_inv_incr,
					    pt[2].y*ctdb->patch_inv_incr,
					    &level);
	      pt[2].z=level;

	      add_terrain_polygon(&pt[0],CTDB_PTS_PER_TRIANGLE,
				  SOIL_DEEP_WATER);
	      
	      pt[2].x = pt[0].x + to_meters;
	      pt[2].y = pt[0].y;
	      ctdb_lookup_patch_water_level(ctdb,
					    data.patch.member,
					    pt[2].x*ctdb->patch_inv_incr,
					    pt[2].y*ctdb->patch_inv_incr,
					    &level);
	      pt[2].z=level;

	      add_terrain_polygon(&pt[0],CTDB_PTS_PER_TRIANGLE,
				  SOIL_DEEP_WATER);
	  }  
	  
	  feature = ctdb_lookup_feature(ctdb,data.patch.number);

	  if (!feature)
	    continue;

	  last = feature + feature->hdr.size;
	  feature++;
	  
	  obj = (CTDB_FEAT_OBJ_PTR)view_obj;

	  for(;feature<last; feature+=next)
	  {
	      int32 fclass=feature->info.feature_class;
	
	      /* if the current class doesn't match the previos class 
	       * it is time to change the class object pointer.
	       */
	      if ( fclass != obj->class )
		obj = (CTDB_FEAT_OBJ_PTR)&view_obj[fclass];	

	      if (obj->func)
		obj->func(ctdb,obj,feature,(ADDRESS)&data,&next); 
	  }
      }
}

static void
add_feature_polygons(
		     int32 px,
		     int32 py,
		     CTDB_FEATURE_DATA data[],
		     int32 incr
		     )
{
    int32 g;
    float64 vx0, vy0, vz0;
    float64 vx1, vy1, vz1;
    float64 vx2, vy2, vz2;
    float64 a,b,c;
    int32 fclass; 

    if ((data[0].info.vertices & MICRO_VERTS) < 2)
      return;

    fclass = data[0].info.feature_class;
    
    if (fclass == FC_MICRO)
    {
	vx0 = 
	  CTDB_XYO_TO_METERS(data[2].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy0 = 
	  CTDB_XYO_TO_METERS(data[2].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz0 = data[3].z;
    }
    else /* it's fclass == FC_CANOPY */ 
    {
	vx0 = 
	  CTDB_XYO_TO_METERS(data[1].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy0 = 
	  CTDB_XYO_TO_METERS(data[1].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz0 = data[2].z;
    }

    poly_list[pl_cnt].pts[0][X] = vx0;
    poly_list[pl_cnt].pts[0][Y] = vy0;
    poly_list[pl_cnt].pts[0][Z] = vz0;

    if (fclass == FC_MICRO)
    {
	g = incr*data[4].micro.vert0;
	vx1 = 
	  CTDB_XYO_TO_METERS(data[g+2].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy1 = 
	  CTDB_XYO_TO_METERS(data[g+2].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz1 = data[g+3].z;	
    }   
    else 
    {
	g = incr*data[3].canopy.vert0;
	vx1 = 
	  CTDB_XYO_TO_METERS(data[g+1].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy1 = 
	  CTDB_XYO_TO_METERS(data[g+1].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz1 = data[g+2].z;
    }

    poly_list[pl_cnt].pts[1][X] = vx1;
    poly_list[pl_cnt].pts[1][Y] = vy1;
    poly_list[pl_cnt].pts[1][Z] = vz1;

    if (fclass == FC_MICRO)
    {
	g = incr*data[4].micro.vert1;
	vx2 = 
	  CTDB_XYO_TO_METERS(data[g+2].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy2 = 
	  CTDB_XYO_TO_METERS(data[g+2].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz2 = data[g+3].z;
    }   
    else 
    {
	g = incr*data[3].canopy.vert1;
	vx2 = 
	  CTDB_XYO_TO_METERS(data[g+1].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy2 = 
	  CTDB_XYO_TO_METERS(data[g+1].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz2 = data[g+2].z;
    }	

    poly_list[pl_cnt].pts[2][X] = vx2;
    poly_list[pl_cnt].pts[2][Y] = vy2;
    poly_list[pl_cnt].pts[2][Z] = vz2;

    poly_list[pl_cnt].n_pts = 3;
    set_dist(poly_list+pl_cnt);

    if (fclass == FC_MICRO)
    {
#ifdef CTDB_VIEWER_TINTING
	ctdb_compute_normal(data, 0, &a, &b, &c);
	if (c<0) 
	c = -c;
#endif
	a = b = 0.0;
	c = 1.0;
	poly_list[pl_cnt].gc =
	  shade_gcs[shade_val(a,b,c, poly_list[pl_cnt].min_dist)];
	poly_list[pl_cnt].always_outline = 0;
	poly_list[pl_cnt].priority = LAND_PRI;
    }
    else
    {
	poly_list[pl_cnt].gc = tree_gc;
	poly_list[pl_cnt].always_outline = 0;
	poly_list[pl_cnt].priority = TREE_PRI;
    }

    pl_cnt = (pl_cnt+1)%MAX_POLYS;

    if ((data[0].info.vertices & MICRO_VERTS) < 3)
      return;

    poly_list[pl_cnt].pts[0][X] = vx0;
    poly_list[pl_cnt].pts[0][Y] = vy0;
    poly_list[pl_cnt].pts[0][Z] = vz0;
    poly_list[pl_cnt].pts[1][X] = vx1;
    poly_list[pl_cnt].pts[1][Y] = vy1;
    poly_list[pl_cnt].pts[1][Z] = vz1;

    if (fclass == FC_MICRO)
    {
	g = incr*data[4].micro.vert2;
	vx2 = 
	  CTDB_XYO_TO_METERS(data[g+2].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy2 = 
	  CTDB_XYO_TO_METERS(data[g+2].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz2 = data[g+3].z;
    }
    else 
    {
	g = incr*data[3].canopy.vert2;
	vx2 = 
	  CTDB_XYO_TO_METERS(data[g+1].xy.x,px,ctdb->patch_incrD,ctdb->incrD);
	vy2 = 
	  CTDB_XYO_TO_METERS(data[g+1].xy.y,py,ctdb->patch_incrD,ctdb->incrD);
	vz2 = data[g+2].z;
    }	
    
    poly_list[pl_cnt].pts[2][X] = vx2;
    poly_list[pl_cnt].pts[2][Y] = vy2;
    poly_list[pl_cnt].pts[2][Z] = vz2;

    poly_list[pl_cnt].n_pts = 3;
    set_dist(poly_list+pl_cnt);

    if (fclass == FC_MICRO)
    {
#ifdef CTDB_VIEWER_TINTING
	ctdb_compute_normal(data, 1, &a, &b, &c);
	if (c<0) c = -c;
#endif
	a = b = 0.0;
	c = 1.0;
	poly_list[pl_cnt].gc =
	  shade_gcs[shade_val(a,b,c, poly_list[pl_cnt].min_dist)];
	poly_list[pl_cnt].always_outline = 0;
	poly_list[pl_cnt].priority = LAND_PRI;
    }
    else
    {
	poly_list[pl_cnt].gc = tree_gc;
	poly_list[pl_cnt].always_outline = 0;
	poly_list[pl_cnt].priority = TREE_PRI;
    }

    pl_cnt = (pl_cnt+1)%MAX_POLYS;
}

static int32 
shade_val(xnorm, ynorm, znorm, dist2)
    float64 xnorm, ynorm, znorm;
    float64 dist2;
{
    float64 dot, cost;
    int32 ret;

    dot = xnorm * sun_vec[0] + ynorm * sun_vec[Y] + znorm * sun_vec[Z];

    cost = dot / (sqrt(xnorm*xnorm + ynorm*ynorm + znorm*znorm));

    ret = 8 - 8 * (dist2 / (7000 * 7000)) + 3 * sun_vec[Z] + 4 * cost;

    if (ret < 0)
      ret = 0;
    else if (ret > 15)
      ret = 15;

    return ret;
}      

static void
draw_grid(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;/* db coordinates */
{
    CTDB_COORD2D_I32 plow,phigh;
    CTDB_COORD2D pt1;
    float64 post,step;
    float64 dim;
    int32 patch_number,px,py;
    float64 patch_size=(ctdb->patch_incr * ctdb->incr);
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    int16 xp0, yp0, xp1, yp1;
    char buf[20];
    plow.x = NS_MAX(x0 * ctdb->meters_to_patch,0);
    plow.y = NS_MAX(y0 * ctdb->meters_to_patch,0);
    phigh.x = NS_MIN(ceil(x1 * ctdb->meters_to_patch),ctdb->patches_wide);
    phigh.y = NS_MIN(ceil(y1 * ctdb->meters_to_patch),ctdb->patches_high);
    for (px=plow.x; px < phigh.x; px++)
    {
	pt1.x = px * patch_size;
	for (py=plow.y; py < phigh.y; py++)
	{
	    pt1.y = py * patch_size;
	    patch_number = py*ctdb->patches_wide + px;
	    if ( ctdb_lookup_te_ptr(ctdb,patch_number) )
	    {
		CTDB_PATCH_GROUP *pg;
		pg = ctdb->patch_groups + PG_NUMBER(patch_number);
		dim = pg->data_start.tin_data->te_vgrid_dim[ PG_MEMBER(patch_number) ] ;
	    }
	    else 
	      dim = ctdb->patch_incr;
	    
	    step = ctdb->patch_incr * ctdb->incr / dim ;
	    for (post=pt1.x;post<=(pt1.x+patch_size);post+=step)
	    { /* Draw grid lines in the patch */
		/* Ensure that both start and end point are defined
		   in the database */
		if ( !CTDB_INCLUSIVE_BETWEEN(pt1.y,ctdb->min_y,ctdb->max_y) ||
		    !CTDB_INCLUSIVE_BETWEEN(pt1.y,ctdb->min_y,ctdb->max_y))
		  continue;
		
		to_pixels(post,pt1.y,0.0,
			  inv_scale,&xp0,&yp0);
		to_pixels(post,(pt1.y+patch_size),0.0,
			  inv_scale,&xp1,&yp1);
		if ( post==pt1.x )
		{   /* drawing the vertical lines */ 
		    XDrawLine(XtDisplay(area),WINDOW, grid2_gc,
			      xp0, yp0, xp1, yp1);
		    sprintf(buf,"%d",(int32)(post/patch_size));
		    XDrawString(XtDisplay(area), WINDOW, black_gc,
				xp0, 20, buf, strlen(buf));
		}
		else
		  XDrawLine(XtDisplay(area),WINDOW, grid1_gc,
			    xp0, yp0, xp1, yp1);
	    }

	    for (post=pt1.y;post<=(pt1.y+patch_size);post+=step)
	    { /* drawing the horizontal lines */
		/* Ensure that both start and end point are defined
		   in the database */
		if ( !CTDB_INCLUSIVE_BETWEEN(pt1.x,ctdb->min_x,ctdb->max_x) ||
		    !CTDB_INCLUSIVE_BETWEEN(pt1.x,ctdb->min_x,ctdb->max_x))
		  continue;
		to_pixels(pt1.x,post,0.0,
			  inv_scale,&xp0,&yp0);
		to_pixels((pt1.x+patch_size),post,0.0,
			  inv_scale,&xp1,&yp1);
		if ( post==pt1.y )
		{
		    XDrawLine(XtDisplay(area),WINDOW, grid2_gc,
			      xp0, yp0, xp1, yp1);
		    sprintf(buf,"%d",(int32)(post/patch_size));
		    XDrawString(XtDisplay(area), WINDOW, black_gc,
				0, yp0, buf, strlen(buf));
		}
		else
		  XDrawLine(XtDisplay(area),WINDOW, grid1_gc,
			    xp0, yp0, xp1, yp1);
	    }
	}
    }  
} 


static void
draw_gridded_terrain(x0,y0,x1,y1,do_soil)
    int32 x0;
    int32 y0;
    int32 x1;
    int32 y1;
    int32 do_soil;
{
/* ADD 3D Processing in this routine and remove the other */
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 x, y;
    CTDB_COORD3D dbpt[6];
    int16 pts[6][2];/* SOUTH=0, POST=1, EAST=2, SE=3, SOUTH=4, POST=5 */
    int32 i,j, post;
    int32 on_screen[6];
    int32 loop_start[2][2]=
    {/* Northwest->Southeast */
	{/* SOUTHEAST->SOUTH->POST */ 3,
	 /* POST->EAST->SOUTHEAST  */ 1 } ,
     /* Northeast->Southwest */
	{/* EAST->SOUTHEAST->SOUTH */ 2,
	   /* SOUTH->POST->EAST */    0 }	
    };
    
    x0 = ctdb->incr * (x0/ctdb->incr-1);
    y0 = ctdb->incr * (y0/ctdb->incr-1);
    x1 = ctdb->incr * (x1/ctdb->incr+1);
    y1 = ctdb->incr * (y1/ctdb->incr+1);
#define LIMIT_TO_CELL(pt,min,max) ( pt<min ? min :( pt>max ? max : pt ))
    x0 = LIMIT_TO_CELL(x0,0,ctdb->max_x);
    x1 = LIMIT_TO_CELL(x1,0,ctdb->max_x);
    /* Note: The Post Y = 0 is a row of posts with nothing defined to the 
     * south of the row.  Omit the generation of triangles for this row.
     */
    y0 = LIMIT_TO_CELL(y0,1,ctdb->max_y);
    y1 = LIMIT_TO_CELL(y1,1,ctdb->max_y);
    
    /* This loop moves by posts but in units of meters */
    for (y=y1;y>=y0;y-=ctdb->incrD)
      for (x=x0;x<x1;x+=ctdb->incrD)
      {
	  int32 patch_number=(PATCH_Y(y*ctdb->meters_to_patch)*
			      ctdb->patches_wide +
			      PATCH_X(x*ctdb->meters_to_patch));
	  
	  /* patch contains terrain elements instead of grid post data */
	  if ( ctdb_lookup_te_ptr(ctdb,patch_number) )
	    continue;
    
	  /* SOUTH=0, POST=1, EAST=2, SE=3, SOUTH=4, POST=5 */
	  dbpt[0].x = x,            dbpt[0].y = y-ctdb->incrD;
	  dbpt[1].x = x,            dbpt[1].y = y;
	  dbpt[2].x = x+ctdb->incrD, dbpt[2].y = y;
	  dbpt[3].x = x+ctdb->incrD, dbpt[3].y = y-ctdb->incrD;
	  dbpt[4].x = x           , dbpt[4].y = y-ctdb->incrD;
	  dbpt[5].x = x           , dbpt[5].y = y;

	  if (do_3D)
	  {
	      for (j=0;j<6;j++)
	      {
		  int32 norm_x = (dbpt[j].x*ctdb->inv_incr);
		  int32 norm_y = (dbpt[j].y*ctdb->inv_incr);
	
		  dbpt[j].z = ctdb_lookup_post_elevation(ctdb,norm_x,norm_y);
	      }	      
	  }
	  
	  for (i=0;i<6;i++)
	    on_screen[i] = to_pixels(dbpt[i].x,dbpt[i].y,0.0,
				 inv_scale,&pts[i][X],&pts[i][Y]);

	  for (j=0; j<2; j++ )
	  {
	      int32 norm_x = x * ctdb->inv_incr;
	      int32 norm_y = y * ctdb->inv_incr;

	      post = ctdb_lookup_elev_post(ctdb, norm_x, norm_y);

	      i = loop_start[post & POST_SW_NE ? 1 : 0][j];

	      if ( !(on_screen[i] | on_screen[i+1] | on_screen[i+2] ) )
		continue;

	      if (do_3D)
	      {
		  CTDB_COORD2D soil;
		  int32 sindex;
		  
		  soil.x = (dbpt[i].x + dbpt[i+1].x + dbpt[i+2].x)/3 ;
		  soil.y = (dbpt[i].y + dbpt[i+1].y + dbpt[i+2].y)/3 ;
		  sindex = CTDB_SIMNET_SOIL(ctdb_lookup_soil(ctdb,
							     soil.x,
							     soil.y));
		  add_terrain_polygon(&dbpt[i],CTDB_PTS_PER_TRIANGLE,
				      sindex);
		  continue;
	      }

	      /* generate triangles (line art only) */
	      XDrawLine(XtDisplay(area), WINDOW , black_gc,
			pts[i][X], pts[i][Y], 
			pts[i+1][X], pts[i+1][Y]);
	      XDrawLine(XtDisplay(area), WINDOW , black_gc,
			pts[i+1][X], pts[i+1][Y], 
			pts[i+2][X], pts[i+2][Y]);
	      XDrawLine(XtDisplay(area), WINDOW , black_gc,
			pts[i+2][X], pts[i+2][Y], 
			pts[i][X], pts[i][Y]);
	  }
      }
}

static void
render_poly_list()
{
    int32 i;
    int16 pts[MAX_POLY_PTS*4][2];
    int32 n_pts;

    XSetFillStyle(XtDisplay(area), tree_gc, FillStippled);

    qsort(poly_list, pl_cnt, sizeof(SCR_POLY), (CTDB_QSORT)cmp_poly);

    for(i=0;i<pl_cnt;i++)
    {
	transform_poly(poly_list[i].n_pts, poly_list[i].pts,
		       &n_pts, pts);

	if (n_pts < 3)
	  continue;

	if (poly_list[i].gc == tree_gc)
	  XSetTSOrigin(XtDisplay(area), tree_gc, pts[0][X], pts[0][Y]);

	XFillPolygon(XtDisplay(area),WINDOW,poly_list[i].gc,
		     (XPoint *)pts,n_pts, Convex,CoordModeOrigin);

	if (do_outline || poly_list[i].always_outline)
	  XDrawLines(XtDisplay(area),WINDOW,line_gc,
		     (XPoint *)pts,n_pts+1, CoordModeOrigin);
    }

    for(i=0;i<mes_pl_cnt;i++)
    {
	transform_poly(mes_poly_list[i].n_pts, mes_poly_list[i].pts,
		       &n_pts, pts);

	if (n_pts < 3)
	  continue;

	XFillPolygon(XtDisplay(area),WINDOW,mes_poly_list[i].gc,
		     (XPoint *)pts,n_pts, Convex,CoordModeOrigin);

	if (do_outline || mes_poly_list[i].always_outline)
	  XDrawLines(XtDisplay(area),WINDOW,line_gc,
		     (XPoint *)pts,n_pts+1, CoordModeOrigin);
    }

    XSetFillStyle(XtDisplay(area), tree_gc, FillSolid);
}

static void
draw_abstract(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    CTDB_SEARCH_SPACE_PTR space;
    int32 code, count, data_size;
    CTDB_ABSTRACT_DATA *data;
    float32 *verts;

    if (do_3D)
      return;

    space = 
      ctdb_create_search_space(ctdb, x0, y0, x1, y1, CTDB_BY_PATCH);
    while (ctdb_next_patch(space, NULL))
    {
	for (code=1;code<CTDB_ABSTRACT_MAX;code++)
	  while (count = ctdb_next_abstract(space, code, &verts, &data,
					    &data_size))
	    add_abstract(code, count, verts, data_size, data);
    }
    ctdb_destroy_search_space(space);
}

static void
draw_nodes(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    int32 i;
    CTDB_NODE *n;

    if (do_3D)
      return;

    x0 -= 1;
    y0 -= 1;
    x1 += 1;
    y1 += 1;

    for (i=0;i<ctdb->num_nodes;i++)
    {
	n = ctdb->nodes[i];
	if ((x0 < n->x) && (n->x < x1) &&
	    (y0 < n->y) && (n->y < y1))
	  add_node(n, i);
    }
}

static void
draw_edges(x0,y0,x1,y1)
    int32 x0,y0,x1,y1;
{
    int32 i;
    CTDB_EDGE *e;
    CTDB_NODE *fn, *ln;

    if (do_3D)
      return;

    x0 -= 1;
    y0 -= 1;
    x1 += 1;
    y1 += 1;

    for (i=0;i<ctdb->num_edges;i++)
    {
	e = ctdb->edges[i];
	fn = ctdb->nodes[e->first_node];
	ln = ctdb->nodes[e->last_node];
	if ((x0 < fn->x) && (fn->x < x1) &&
	    (y0 < fn->y) && (fn->y < y1) &&
	    (x0 < ln->x) && (ln->x < x1) &&
	    (y0 < ln->y) && (ln->y < y1))
	  add_edge(i);
    }
}



int32
ctdb_view_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static int32 first=TRUE;
    static CTDB_TYPE_FUNC view_micro_func[2][CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	{
	    (CTDB_TYPE_FUNC)view_draw_standard_micro2d, /* MICRO_TYPE_BASE */ 
	    (CTDB_TYPE_FUNC)view_draw_standard_micro2d, /* MICRO_TYPE_ML */
	    (CTDB_TYPE_FUNC)view_draw_standard_micro2d, /* MICRO_TYPE_DEFAULT*/
	    (CTDB_TYPE_FUNC)view_draw_water_micro2d     /* MICRO_TYPE_WATER */
	}
	,
	{
	    (CTDB_TYPE_FUNC)view_draw_standard_micro3d, /* MICRO_TYPE_BASE */ 
	    (CTDB_TYPE_FUNC)view_draw_standard_micro3d, /* MICRO_TYPE_ML */
	    (CTDB_TYPE_FUNC)view_draw_standard_micro3d, /* MICRO_TYPE_DEFAULT*/
	    (CTDB_TYPE_FUNC)view_draw_water_micro3d     /* MICRO_TYPE_WATER */
	}
    };
    
    if (first)
    {
	first = FALSE;
	
	/* Class MICRO */
	obj[FC_MICRO].class      = FC_MICRO;
	obj[FC_MICRO].featmask   = POST_MICROTERRAIN;
	obj[FC_MICRO].inclmask   = CTDB_INCL_MICRO;
	obj[FC_MICRO].valid      = 1;
	obj[FC_MICRO].func       = (CTDB_OBJ_FUNC)view_draw_micro; 
	
	/* Class VOLUME */
	obj[FC_VOLUME].class     = FC_VOLUME;
	obj[FC_VOLUME].featmask  = POST_BUILDINGS;
	obj[FC_VOLUME].inclmask  = CTDB_INCL_VOLUME;
	obj[FC_VOLUME].valid     = 1;
	/* Class LINEAR */
	obj[FC_LINEAR].class     = FC_LINEAR;
	obj[FC_LINEAR].featmask  = POST_TREES;
	obj[FC_LINEAR].inclmask  = CTDB_INCL_LINEAR;
	obj[FC_LINEAR].valid     = 1;
	/* Class CANOPY */
	obj[FC_CANOPY].class     = FC_CANOPY;
	obj[FC_CANOPY].featmask  = POST_TREES;
	obj[FC_CANOPY].inclmask  = CTDB_INCL_CANOPY;
	obj[FC_CANOPY].valid     = 1;
	/* Class LAID_LINEAR */
	obj[FC_LAID_LINEAR].class     = FC_LAID_LINEAR;
	obj[FC_LAID_LINEAR].featmask  = POST_FEATURES;
	obj[FC_LAID_LINEAR].inclmask  = CTDB_INCL_LAID_LINEAR;
	obj[FC_LAID_LINEAR].valid     = 1;
    }
    
    ctdb_init_micro_subclass(view_micro_subclass_obj,view_micro_func[do_3D]);
    
    if ( do_3D )
    {
	obj[FC_VOLUME].func     = (CTDB_OBJ_FUNC)view_draw_volume_2d;
	obj[FC_LINEAR].func     = (CTDB_OBJ_FUNC)view_draw_linear_2d;
	obj[FC_CANOPY].func     = (CTDB_OBJ_FUNC)view_draw_canopy_2d;
	obj[FC_LAID_LINEAR].func= (CTDB_OBJ_FUNC)ctdb_invalid_laid_linear;
    }
    else
    {
	obj[FC_VOLUME].func     = (CTDB_OBJ_FUNC)view_draw_volume_2d;
	obj[FC_LINEAR].func     = (CTDB_OBJ_FUNC)view_draw_linear_2d;
	obj[FC_CANOPY].func     = (CTDB_OBJ_FUNC)view_draw_canopy_2d;
	obj[FC_LAID_LINEAR].func= (CTDB_OBJ_FUNC)view_draw_laid_linear_2d;
    }

    return 1;
}

static void
draw_everything()
{
    int32 x0,y0,x1,y1;

    x0 = low_x;
    y0 = low_y;
    x1 = low_x+win_width*scale_mpp*zoom;
    y1 = low_y+win_height*scale_mpp*zoom;

    /* Complete database is off of the viewing screen */
    if ((x0 > ctdb->max_x) || (y0 > ctdb->max_y) ||
	(x1 < ctdb->min_x) || (y1 < ctdb->min_y))
      return;
    /* Clipping the viewing area to be within the database bounds */
    if (x0 < ctdb->min_x) x0 = ctdb->min_x;
    if (y0 < ctdb->min_y) y0 = ctdb->min_y;
    if (x1 > ctdb->max_x) x1 = ctdb->max_x;
    if (y1 > ctdb->max_y) y1 = ctdb->max_y;

    pl_cnt = 0;

    ctdb_view_init_feature_class(view_obj);

    CTDB_VALID_FEATURES(view_obj,(do_micro || do_water || do_ml_micro),
                        do_build,do_tree,do_tree,do_linear);

    if (do_3D)
    {
	draw_sun();
	draw_gridded_terrain(x0,y0,x1,y1,0);
	draw_terrain_elements(x0,y0,x1,y1,0);
	
    }    
    else
    {
        viewer_qual = CTDB_INCL_TERRAIN_SKIN | CTDB_COMPUTE_NORMAL; 
	
        if ( do_micro )
          viewer_qual |= CTDB_INCL_MICRO_DEF_BASE;

        if (do_water)
          viewer_qual |= CTDB_INCL_WATER;
        
        if (do_ml_micro)
          viewer_qual |= CTDB_INCL_MICRO_ML;
        
        if (do_build)
          viewer_qual |= CTDB_INCL_VOLUME;
        
        if (do_tree)
          viewer_qual |= (CTDB_INCL_LINEAR|CTDB_INCL_CANOPY);
        
        if (do_linear)
          viewer_qual |= CTDB_INCL_LAID_LINEAR;
        
	if ( do_soil )
	{
            viewer_qual |= CTDB_INCL_MATERIAL;
            draw_soils(x0,y0,x1,y1);
	}
	
	if ( do_assoc )
	  draw_assoc2D(x0,y0,x1,y1);/* draw association */
	
	if ( do_grid )
        {
            viewer_qual |= CTDB_INCL_TERRAIN_SKIN;
            draw_grid(x0,y0,x1,y1);
        }
	
	if ( do_terrain )
	{
            viewer_qual |= CTDB_INCL_TERRAIN_SKIN;
	    draw_terrain_elements(x0,y0,x1,y1,0);/*draw terrain element lines*/
	    draw_gridded_terrain(x0,y0,x1,y1,0) ;/*draw gridded terrain lines*/
	}
	
	if ( do_markings )
	  draw_terrain_element_markings2D(x0,y0,x1,y1);
    }
    
    CTDB_VALID_FEATURES(view_obj,(do_micro || do_water || do_ml_micro),
			do_build,do_tree,do_tree,do_linear);
    
    feature_loop(x0,y0,x1,y1);

    if (do_abstract)
      draw_abstract(x0,y0,x1,y1);

    if (do_nodes)
      draw_nodes(x0,y0,x1,y1);

    if (do_edges)
      draw_edges(x0,y0,x1,y1);

    if (do_intervis)
      draw_intervis_pole(x0,y0,x1,y1);

    render_poly_list();

    if (do_contour)
      draw_contours(x0,y0,x1,y1);

    if (do_intervis)
    {
	draw_intervis_pole(x0,y0,x1,y1);	
	draw_los(x0,y0,x1,y1);
    } 

    if ( do_soil && !do_3D )
      draw_soil_legend(x0,y0,x1,y1);
}

static void
resize_area(w)
    Widget w;
{
    Arg size_args[2];
    char *img;

    if (!XtIsRealized(w))
      return;

    XtSetArg(size_args[0], XmNwidth, &win_width);
    XtSetArg(size_args[1], XmNheight, &win_height);
    XtGetValues(w, size_args, XtNumber(size_args));

    scale_mpp = ((float64)(ctdb->max_x-ctdb->min_x)/win_width);

    if (do_3D)
      compute_3D_transform(low_x, low_y,
			   (int32)(win_width * scale_mpp * zoom),
			   (int32)(win_height * scale_mpp * zoom));

    if (do_hypso && !do_3D)
    {
	if (image)
	{
	    free(image->data);
	    image->data = NULL;
	    XDestroyImage(image);
	    image = NULL;
	}
	    
#ifdef DO_COLOR
	img = (char *) malloc(win_width*win_height);
	image = XCreateImage(XtDisplay(w), DefaultVisualOfScreen(XtScreen(w)),
			     8,ZPixmap,0,img, win_width,win_height,8,win_width);

	ctdb_hypso_image(ctdb, low_x, low_y, win_width, win_height, 
			 scale_mpp*zoom, NCELLS, local_cmap, 1, img, 
			 &minz, &maxz);

	setup_cmap(minz, maxz);
#else
	img = (char *) malloc(win_width*win_height);
	image = XCreateImage(XtDisplay(w), DefaultVisualOfScreen(XtScreen(w)),
			     1,XYBitmap,0,img, win_width,win_height,8,(win_width+7)/8);

	ctdb_hypso_bitmap(ctdb, low_x, low_y, win_width, win_height, 
			  scale_mpp*zoom, ctdb->min_z, ctdb->max_z, img);

#endif
	XPutImage(XtDisplay(area), WINDOW, hyp_gc,
		  image,0,0,0,0,win_width,win_height);
    }
    else
    {
	XFillRectangle(XtDisplay(area), WINDOW, sky_gc,
		       0,0,
		       WidthOfScreen(XtScreen(area))-1,
		       HeightOfScreen(XtScreen(area))-1);
    }


    draw_everything();

    XClearArea(XtDisplay(w), XtWindow(w), 0,0,0,0, TRUE);
}

static void
expose_area(w, ignored, call_data)
    Widget                         w;
    int32                          ignored;
    XmDrawingAreaCallbackStruct   *call_data;
{
    if (call_data && call_data->event->xexpose.count)
      return;

    if (do_drive)
    {
	low_x = low_x - my_x * 0.05;
	low_y = low_y - my_y * 0.05;
	resize_area(w);
    }
}
static void
draw_intervis_ray2D(pt1,pt2)
    CTDB_COORD3D pt1;
    CTDB_COORD3D pt2;
{
    float64 inv_scale = 1.0/(scale_mpp * zoom);
    float64 i, x, y, dx, dy;
    float64 visibility, opacity;
    int16 px0, py0;
    dx = (pt2.x - pt1.x)/100.0;
    dy = (pt2.y - pt1.y)/100.0;
    
    /* First, draw the LOS ray from pt1 to pt2 */
    x = pt1.x , y = pt1.y;
    for(i=0;i<100;i++,x+=dx,y+=dy)
    {
	to_pixels(x,y,0.0,inv_scale,&px0,&py0);
	XDrawPoint(XtDisplay(area),WINDOW,intervis_gc, px0, py0);
    }
    visibility = ctdb_point_to_point(ctdb, pt1.x,pt1.y,
				     pt1.z,pt2.x,pt2.y,
				     pt2.z,pt2.z+POLE_HEIGHT,POLE_WIDTH,
				     NULL, &opacity, 0.05, 0, 
				     (CTDB_VEHICLE_LOCATION *)NULL,
				     0, 0 ) * 100.0 ;
    visibility *= opacity;

    printf("From (%g, %g, %g) ",pt1.x,pt1.y,pt1.z);
    printf("to (%g, %g, %g) ",pt2.x,pt2.y,pt2.z);
    printf("visibility = %g%%\n",visibility);
    
    /* Generate intervis data on the viewer display next to the 
     * endpoint.
     */
    to_pixels(pt2.x,pt2.y,0.0,inv_scale, &px0, &py0);
    sprintf(iv_frac_buf,"%5.2f%%",visibility);
    XDrawString(XtDisplay(area), WINDOW, black_gc,
		px0+4, py0+4, iv_frac_buf, strlen(iv_frac_buf));
}

static void
click(
      Widget    w,
      int32     ignored, 
      XEvent   *event
      )
{
    float64 cx, cy; 
    int32   x;
    uint32  i, num_attributes=100, attr_list[100];
    char    facc_name[CTDB_MAX_FACC_NAME];
    CTDB_PAT_DATA  *ret_list;

    if (do_3D)
      return;

    cx = event->xbutton.x * scale_mpp * zoom + low_x;
    cy = (win_height-event->xbutton.y-1) * scale_mpp * zoom + low_y;

    switch (event->xbutton.button)
    {
      case 1:
        if (do_intervis)
        {
            if (got_click == TRUE)
            {
		CTDB_COORD3D pt2;
	       
		/* This is the second click. Do the intervis. */
		pt2.x = cx, pt2.y = cy;
		pt2.z = ctdb_lookup_elevation(ctdb, pt2.x, pt2.y) + 0.5 ;
                got_click = FALSE;
		draw_intervis_ray2D(pt1,pt2);
		XClearArea(XtDisplay(w), XtWindow(w), 0,0,0,0, TRUE);
	    }
            else
	    {
                /* This is the first click. Wait to do the intervis. */
                pt1.x = cx , pt1.y = cy ;
                pt1.z = ctdb_lookup_elevation(ctdb, pt1.x, pt1.y) + 0.5 ;
                got_click = TRUE;
	    }
	}
        else if ( do_lookups )
	{	    
	    if ( do_road_lookups ) 
	    {
		int32 ei, esi;
		if (ctdb_nearest_road(ctdb, cx, cy, 2000.0, &ei, &esi))
		  printf("Near road %d[%d]\n", ei, esi);
		
	    }

	    if ( do_elevation_lookups )
	    {
		int32 xp = PATCH_X(cx * ctdb->meters_to_patch);
		int32 yp = PATCH_Y(cy * ctdb->meters_to_patch);
		CTDB *tctdb = ctdb;

		if ( !ctdb_point_on_database(tctdb,cx,cy) )
		{
		    printf("(%g,%g,%f) patch number = ILLEGAL\n", cx, cy,
			   ctdb_lookup_elevation( ctdb, cx, cy ));
		    
		}
		else
		{
		    char output[132];
		    int32 patch_number=yp*ctdb->patches_wide+xp;
		    static CTDB_ELEV_DATA *elevptr,elev[CTDB_MAX_ELEVS];
		    
		    ctdb_lookup_attributes_mes(ctdb, cx, cy,
					       viewer_qual,elev);
		    ctdb_sort_elev_data(elev);
			
		    if (do_soil)
		    {
			ctdb_lookup_pat_attributes(ctdb,attr_list,
						   &num_attributes);
			
			ret_list = 
			  ctdb_alloc(NULL,
				     sizeof(CTDB_PAT_DATA)*
				     num_attributes,
				     "Returned List of Attribute Values");
		    }
		    

		    for (elevptr=elev;elevptr->class!=CTDB_FC_ILLEGAL;
			 elevptr++)
		    {
			ctdb_elev_data_to_string(elevptr,output);
			printf("%lf %lf norm (%.4lf,%.4lf,%.4lf) %s \t",
			       cx, cy, elevptr->norm_vec[X], 
			       elevptr->norm_vec[Y], elevptr->norm_vec[Z], 
			       output);
			printf("patch number = %d\n",patch_number);

			if (!do_soil)
			  continue;

			ctdb_lookup_attributes(ctdb, 
					       (uint32)elevptr->material,
					       attr_list,
					       num_attributes, 
					       ret_list);
			x = ret_list[0].integer;
			printf("SIMNET : %2d\t",CTDB_SIMNET_SOIL(x));
			printf("CCTT   : %2d\t",CTDB_CCTT_SOIL(x));
		        printf("WATER  : %2d\t",(CTDB_WATER(x)&&1));
			printf("NO_GO  : %2d\t",(CTDB_NO_GO(x)&&1));
			printf("SLOW_GO: %2d\t",(CTDB_SLOW_GO(x)&&1));
			printf("ROAD   : %d \t",(CTDB_ROAD(x)&&1));

			for (i=1; i<num_attributes; i++)
			{
			    ctdb_lookup_attr_name(attr_list[i], facc_name);
			    printf("%s: %d\t", facc_name, ret_list[i]);
			}
			printf("\n");
		    }
		}
	    }    
	    printf("\n");
	}
	else if (do_damage)
	{
	    draw_damage(ctdb,cx,cy);
            resize_area(w);
	}
	else
	{
            low_x = (int32)cx - 0.5 * win_width * scale_mpp * zoom;
            low_y = (int32)cy - 0.5 * win_height * scale_mpp * zoom;
            resize_area(w);
	}
	break;
      case 2:
	zoom *= 0.5;
        low_x = (int32)cx - 0.5 * win_width * scale_mpp * zoom;
        low_y = (int32)cy - 0.5 * win_height * scale_mpp * zoom;
        resize_area(w);
	break;
      case 3:
	zoom *= 2.0;
        low_x = (int32)cx - 0.5 * win_width * scale_mpp * zoom;
        low_y = (int32)cy - 0.5 * win_height * scale_mpp * zoom;
        resize_area(w);
	break;
    }
}

change_var(
	   Widget                          w,
	   int32                          *var_ptr,
	   XmToggleButtonCallbackStruct   *call_data
	   )
{
    *var_ptr = call_data->set;
    resize_area(area);
}


static int32
get_valid_tide_point(CTDB *ctdb,
                     float64 *x,
                     float64 *y)
{
    static int32 tide_found=FALSE;
    static float64 tide_x,tide_y,tidal;
    int32 i,j;

    if ( !tide_found )
    {
	float64 to_meters=(ctdb->incr*ctdb->patch_incr);
	float64 mid=(to_meters*0.5);
	for (i=0;i<ctdb->patches_high;i++)
	{
	    tide_y=i*to_meters+mid;
	    
	    for (j=0;j<ctdb->patches_wide;j++)
	    {
		tide_x=j*to_meters+mid;
		
			/* Looking for a valid tidal point */
		if (ctdb_ocean_point_query(ctdb,tide_x,tide_y,
					   CTDB_TIDAL_HEIGHT,&tidal,
					   CTDB_NULL))
		{
		    tide_found = TRUE;
		    *x = tide_x;
		    *y = tide_y;
		    return tide_found;
		}
	    }
	}
    }

    *x = tide_x;
    *y = tide_y;
    return tide_found;
}
      
change_alt(Widget                   w,
	   int32                    ignored,
	   XmScaleCallbackStruct   *call_data
	   )
{
    alt3D = call_data->value;
    
    if ( old_do_3D != do_3D )
    {
	old_do_3D = do_3D;
	
	if ( main_options.tide.value && !do_3D)
	{
            static float64 tidal,x,y;

	    if ( get_valid_tide_point(ctdb,&x,&y) )
	      (void)ctdb_ocean_point_query(ctdb,x,y,
					   CTDB_TIDAL_HEIGHT,&tidal,
					   CTDB_NULL);
	    else
	      tidal = 0;
	    
	    /* change the altitude resource to reflect tidal needs */
	    XtVaSetValues(w, XmNminimum , -50, XmNmaximum , 50, 
			  XmNvalue , (int)tidal ,
			  XmNtitleString, 
			  XmStringCreateLtoR("Tidal Level",
					     XmSTRING_DEFAULT_CHARSET), 
			  NULL);
	}
	else
	{
#ifdef SE3_OCEAN_DEMO
	    XtVaSetValues(w, 
			  XmNminimum , (int)(-560), 
			  XmNmaximum , (int)(800), 
			  XmNvalue , (int)(0) ,
			  XmNtitleString, 
			  XmStringCreateLtoR("AGL",
					     XmSTRING_DEFAULT_CHARSET), 
			  NULL);
#else
	    XtVaSetValues(w, 
			  XmNminimum , (int)(ctdb->min_z), 
			  XmNmaximum , (int)(ctdb->max_z), 
			  XmNvalue , (int)((ctdb->max_z-ctdb->min_z)/2) ,
			  XmNtitleString, 
			  XmStringCreateLtoR("Altitude",
					     XmSTRING_DEFAULT_CHARSET), 
			  NULL);
#endif
	}
    }

    if ( !do_3D && main_options.tide.value )
    {
	float64 tidal,x,y;
        
	if (get_valid_tide_point(ctdb,&x,&y) &&
	    ctdb_ocean_point_query(ctdb,x,y,
				   CTDB_TIDAL_HEIGHT,&tidal,CTDB_NULL))
	{
	    /* change the tide level in the ctdb */
	    printf("Current Tidal Height %.2lf\tChange to %.2lf\n",
		   tidal,(float64)call_data->value);
	    ctdb_ocean_point_write(ctdb,x,y,
				   CTDB_TIDAL_HEIGHT,
				   (float64)call_data->value,
				   CTDB_NULL);
	    resize_area(area);
	}
    } 
    else if ( do_3D )
      resize_area(area);
}


change_fov(
	   Widget                   w,
	   int32                    ignored,
	   XmScaleCallbackStruct   *call_data
	   )
{
    fov3D = call_data->value;
    if (do_3D)
      resize_area(area);
}

change_dir(w,ignored,call_data)
    Widget                 w;
    int32                    ignored;
    XmScaleCallbackStruct *call_data;
{
    dir3D = call_data->value;
    if (do_3D)
      resize_area(area);
}

change_sun(w,ignored,call_data)
    Widget                 w;
    int32                    ignored;
    XmScaleCallbackStruct *call_data;
{
    float64 theta = PI * call_data->value / 180.0 - PI/2;

    sun_vec[0] = cos(theta);
    sun_vec[Y] = 0.0;
    sun_vec[Z] = sin(theta);
#ifdef CTDB_VIEWER_DEBUG
    printf("%f %f %f\n",sun_vec[0],sun_vec[Y],sun_vec[Z]);
#endif

    if (do_3D)
      resize_area(area);
}

change_con(w,ignored,call_data)
    Widget                 w;
    int32                    ignored;
    XmScaleCallbackStruct *call_data;
{
    contour = call_data->value;
    resize_area(area);
}

static Arg form_args[] = { { XmNresizePolicy , XmRESIZE_NONE } ,
			   { XmNwidth , 10 } ,
			   { XmNheight , 10 } ,
			 };
static Arg box_args[] = { { XmNorientation , XmHORIZONTAL } ,
			  { XmNpacking , XmPACK_TIGHT } ,
#define BOX_HEIGHT 80
			  { XmNheight , BOX_HEIGHT } ,
			  { XmNresizeHeight , FALSE } ,
			 };
static Arg fov_args[] = { { 0 , 0 } ,
			  { XmNorientation , XmHORIZONTAL } ,
			  { XmNminimum , 1 } ,
			  { XmNmaximum , 179 } ,
			  { XmNvalue , 120 } ,
			  { XmNshowValue , TRUE },
			};
static Arg dir_args[] = { { 0 , 0 } ,
			  { XmNorientation , XmHORIZONTAL } ,
			  { XmNminimum , 0 } ,
			  { XmNmaximum , 360 } ,
			  { XmNvalue , 0 } ,
			  { XmNshowValue , TRUE },
			};
static Arg sun_args[] = { { 0 , 0 } ,
			  { XmNorientation , XmHORIZONTAL } ,
			  { XmNminimum , 0 } ,
			  { XmNmaximum , 360 } ,
 			  { XmNvalue , 90 } ,
			};
static Arg con_args[] = { { 0 , 0 } ,
			  { XmNorientation , XmHORIZONTAL } ,
			  { XmNminimum , 1 } ,
			  { XmNmaximum , 160 } ,
			  { XmNvalue , 40 } ,
			  { XmNshowValue , TRUE },
			};
static Arg alt_args[] = { { 0 , 0 } ,
			  { XmNorientation , XmHORIZONTAL } ,
			  { XmNminimum , 1 } ,
			  { XmNmaximum , 2000 } ,
			  { XmNvalue , 1 } ,
			  { XmNshowValue , TRUE },
			};

void XswLayoutWidget(base, width, height, va_alist)
    Widget base;
    int width;
    int height;
    va_dcl
{
    int att_val, offset_val;
    Widget att_wid;
    Arg organize_args[20];
    int n = 0;
    va_list ap;

    if (width)
      XtSetArg(organize_args[n], XmNwidth, width), n++;
    if (height)
      XtSetArg(organize_args[n], XmNheight, height), n++;

    va_start(ap);

#define GET_ATT(att_name, offset_name, position_name, wid_name) \
    att_val = va_arg(ap, int);                                  \
    XtSetArg(organize_args[n], att_name, att_val), n++;         \
    offset_val = va_arg(ap, int);                               \
    if (att_val == XmATTACH_POSITION)                                   \
      XtSetArg(organize_args[n], position_name, offset_val), n++;       \
    else                                                                \
      XtSetArg(organize_args[n], offset_name, offset_val), n++; \
    if (att_val == XmATTACH_WIDGET)                             \
    {                                                           \
        att_wid = va_arg(ap, Widget);                           \
        XtSetArg(organize_args[n], wid_name, att_wid), n++;     \
    }

    GET_ATT(XmNtopAttachment, XmNtopOffset,
            XmNtopPosition, XmNtopWidget);
    GET_ATT(XmNleftAttachment, XmNleftOffset,
            XmNleftPosition, XmNleftWidget);
    GET_ATT(XmNrightAttachment, XmNrightOffset,
            XmNrightPosition, XmNrightWidget);
    GET_ATT(XmNbottomAttachment, XmNbottomOffset,
            XmNbottomPosition, XmNbottomWidget);

    va_end(ap);

    XtSetValues(base, organize_args, n);
}

main (argc,argv)
    int32 argc;
    char *argv[];
{

    int32 copy_argc = argc;
    char **copy_argv;
    Widget box;
    Widget grid, micro, linear, contour, hypso, masks, outline, te, assoc;
    Widget build, tree, abstract, nodes, edges, interv, soil, lookups, markings, damage;
    Widget view_3d, alt, fov, dir, sunw, con, drive;
    int32 ignore,j;
    Arg bp_arg;
    int32 cell[100];
    int32 n_cell=100;
    int32 i,cell_id, in_list = 0;
    char name[256],namever[256];
    int32 version;
    
    reader_init(0);
    {
	CMD_OPTION *options;
	uint32      sizeof_options = 0;
	CMD_TYPES **map;
	int32       leftover_argc;
	char *     *leftover_argv;

	/* parse terrain related options */
	wld_tdb_aggregate_options(&options, &sizeof_options, &map);
	
	cmd_aggregate(&options, &sizeof_options,
		      (CMD_OPTION *)&main_options, /* options array */
		      sizeof(main_options), /* size of options array */
		      &map);
	cmd_process_aggregate_options(argc, argv, /* args to main() */
		      &leftover_argc, &leftover_argv, /* unrecognized */
		      options, sizeof_options, map,
		      TRUE /* Verbose mode */
		      );
    }

    wld_load_playbox(1, 0, 0, NULL, main_options.data.value, 
		     NULL, NULL, NULL, NULL);

    wld_lookup_playbox_cells(&n_cell,cell);

    if (n_cell == 0)
      printf("Error:  no cells defining database.\n"), exit(-1);
    else if (n_cell == 1)
      cell_id = cell[0];
    else
    {
	printf("List of cells which define database extent: ");
	
	for (i=0;i<n_cell;i++)
	  printf("%06d, ",cell[i]);
	
	do
	{
	    printf("\nEnter cell id:  ");
	    scanf("%d",&cell_id);
	    
	    for (i=0; i < n_cell; i++)
	      if (cell[i] == cell_id)
	      {
		  in_list = 1;
		  break;
	      }
	    if (!in_list)
	      printf("Sorry, that cell is not defined for the database\n");
	    
	} while (!in_list);
    }
    
    ctdb = (CTDB *)gcs_get_tdb(cell_id, FALSE);

    {
	/* process the X Display command line */
	copy_argv = (char **)malloc((argc+5)*sizeof(char *));
	bcopy(argv, copy_argv, argc*sizeof(char *));
	copy_argv[copy_argc++] = "-xrm";
	copy_argv[copy_argc++] = "*.fontList: 6x10";
	/* Name of window displayed is the path entered for the database */
	copy_argv[copy_argc++] = "-name";
	ctdb_get_tdb_identifier (ctdb,name,&version);
	sprintf(namever,"%s-%04d/cell %5d",name,version,cell_id);
	copy_argv[copy_argc++] = namever;
    }

    toplevel = XtInitialize(argv[0],"viewer", NULL, 0,
			    &copy_argc,copy_argv);

    /* allocate points for terrain element rendering */
    te_pts = (CTDB_TERRAIN_TRIANGLE *)
      ctdb_alloc((te_pts=(CTDB_TERRAIN_TRIANGLE *)NULL),
		 sizeof(CTDB_TERRAIN_TRIANGLE)*pts_dim,
		 "Initial allocation of terrain triangle list");

    g_dialog_shell = toplevel;

    form = XtCreateManagedWidget("form", xmFormWidgetClass,
				 toplevel,
				 form_args, XtNumber(form_args));

    area = XtCreateManagedWidget("area", xmDrawingAreaWidgetClass,
				 form, NULL, 0);

    XtAddCallback(area, XmNresizeCallback, 
		  (XtCallbackProc)resize_area, NULL);
    XtAddCallback(area, XmNexposeCallback, 
		  (XtCallbackProc)expose_area, NULL);
    XtAddEventHandler(area, ButtonPressMask, FALSE, click, 0);

    
    box = XtCreateManagedWidget("box", xmRowColumnWidgetClass,
				form, box_args, XtNumber(box_args));

    view_3d = XtCreateManagedWidget("3D", xmToggleButtonWidgetClass,
				    box, NULL, 0);
    XtAddCallback(view_3d, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_3D);

    grid = XtCreateManagedWidget("Grid", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(grid, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_grid);

    te = XtCreateManagedWidget("Terrain", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(te, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_terrain);

    markings = XtCreateManagedWidget("Terrain Markings", 
				     xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(markings, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_markings);

    soil = XtCreateManagedWidget("Soil", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(soil, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_soil);

    assoc = XtCreateManagedWidget("Assoc", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(assoc, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_assoc);

    micro = XtCreateManagedWidget("Micro", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(micro, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_micro);

    micro = XtCreateManagedWidget("Water", xmToggleButtonWidgetClass,
				  box, NULL, 0);
    XtAddCallback(micro, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_water);

    micro = XtCreateManagedWidget("ML Micro", xmToggleButtonWidgetClass,
				  box, NULL, 0);
    XtAddCallback(micro, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_ml_micro);

    contour = XtCreateManagedWidget("Contour", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(contour, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_contour);

    hypso = XtCreateManagedWidget("Hypso", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(hypso, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_hypso);

    masks = XtCreateManagedWidget("Masks", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(masks, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_masks);

    interv = XtCreateManagedWidget("Intervis", xmToggleButtonWidgetClass,
				   box, NULL, 0);
    XtAddCallback(interv, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_intervis);

    lookups = XtCreateManagedWidget("Lookups", xmToggleButtonWidgetClass,
				   box, NULL, 0);
    XtAddCallback(lookups, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_lookups);

    damage =  XtCreateManagedWidget("Damage", xmToggleButtonWidgetClass,
				   box, NULL, 0);
    XtAddCallback(damage, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_damage);

    outline = XtCreateManagedWidget("Outline", xmToggleButtonWidgetClass,
				    box, NULL, 0);
    XtAddCallback(outline, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_outline);

    linear = XtCreateManagedWidget("Linear", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(linear, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_linear);

    build = XtCreateManagedWidget("Buildings", xmToggleButtonWidgetClass,
				  box, NULL, 0);
    XtAddCallback(build, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_build);

    tree = XtCreateManagedWidget("Trees", xmToggleButtonWidgetClass,
				 box, NULL, 0);
    XtAddCallback(tree, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_tree);

    abstract = XtCreateManagedWidget("Abstract", xmToggleButtonWidgetClass,
				     box, NULL, 0);
    XtAddCallback(abstract, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_abstract);

    nodes = XtCreateManagedWidget("Nodes", xmToggleButtonWidgetClass,
				  box, NULL, 0);
    XtAddCallback(nodes, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_nodes);

    edges = XtCreateManagedWidget("Edges", xmToggleButtonWidgetClass,
				  box, NULL, 0);
    XtAddCallback(edges, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_edges);

    XtSetArg(alt_args[0], XmNtitleString, 
	     XmStringCreateLtoR("Altitude", XmSTRING_DEFAULT_CHARSET));    
    alt = XtCreateManagedWidget("Altitude", xmScaleWidgetClass,
				form, alt_args, XtNumber(alt_args));
    XtAddCallback(alt, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_alt, NULL);

    if ( main_options.tide.value )
    {
	XtVaSetValues(alt, 
		      XmNminimum , -50, 
		      XmNmaximum ,  50, 
		      XmNvalue   ,  0,
		      XmNtitleString, 
		      XmStringCreateLtoR("Tidal Level",
					 XmSTRING_DEFAULT_CHARSET), 
		      NULL);
    }

    XtSetArg(fov_args[0], XmNtitleString, 
	     XmStringCreateLtoR("Field of View", XmSTRING_DEFAULT_CHARSET));
    fov = XtCreateManagedWidget("Field of View", xmScaleWidgetClass,
				form, fov_args, XtNumber(fov_args));
    XtAddCallback(fov, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_fov, NULL);

    XtSetArg(dir_args[0], XmNtitleString, 
	     XmStringCreateLtoR("Direction", XmSTRING_DEFAULT_CHARSET));
    dir = XtCreateManagedWidget("Direction", xmScaleWidgetClass,
				form, dir_args, XtNumber(dir_args));
    XtAddCallback(dir, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_dir, NULL);

    XtSetArg(sun_args[0], XmNtitleString, 
	     XmStringCreateLtoR("Sun", XmSTRING_DEFAULT_CHARSET));
    sunw = XtCreateManagedWidget("Sun", xmScaleWidgetClass,
				form, sun_args, XtNumber(sun_args));
    XtAddCallback(sunw, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_sun, NULL);

    XtSetArg(con_args[0], XmNtitleString, 
	     XmStringCreateLtoR("Contour", XmSTRING_DEFAULT_CHARSET));
    con = XtCreateManagedWidget("Contour", xmScaleWidgetClass,
				form, con_args, XtNumber(con_args));
    XtAddCallback(con, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_con, NULL);

    drive = XtCreateManagedWidget("Drive", xmToggleButtonWidgetClass,
				  form, NULL, 0);
    XtAddCallback(drive, XmNvalueChangedCallback, 
		  (XtCallbackProc)change_var, &do_drive);

    XswLayoutWidget(drive, 0, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_FORM, 5,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(con, 132, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_WIDGET, 5, drive,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(fov, 132, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_WIDGET, 5, con,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(dir, 132, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_WIDGET, 5, fov,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(alt, 132, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_WIDGET, 5, dir,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(sunw, 132, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_NONE, 0,
		    XmATTACH_WIDGET, 5, alt,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(box, 0, BOX_HEIGHT,
		    XmATTACH_NONE, 0,
		    XmATTACH_FORM, 5,
		    XmATTACH_WIDGET, 5, sunw,
		    XmATTACH_FORM, 5);
    XswLayoutWidget(area, 0, 0,
		    XmATTACH_FORM, 5,
		    XmATTACH_FORM, 5,
		    XmATTACH_FORM, 5,
		    XmATTACH_WIDGET, 5, box);

    XtRealizeWidget(toplevel);

    pixmap = XCreatePixmap(XtDisplay(toplevel), XtWindow(toplevel),
			   WidthOfScreen(XtScreen(toplevel)),
			   HeightOfScreen(XtScreen(toplevel)),
			   DefaultDepthOfScreen(XtScreen(toplevel)));
    XtSetArg(bp_arg, XmNbackgroundPixmap, pixmap);
    XtSetValues(area, &bp_arg, 1);

    cmap = DefaultColormapOfScreen(XtScreen(toplevel));

    setup_gcs();

    if (!XAllocColorCells(XtDisplay(toplevel), cmap,
			  FALSE, &ignore, 1, local_cmap, NCELLS))
      printf("XAllocColorCells failed\n");

    for(j=0;j<XtNumber(colors);j++)
    {
	if (!XParseColor(XtDisplay(toplevel), cmap,
			 color_names[j], &colors[j]))
	  printf("Did not find color %s\n",color_names[j]);
    }

    XtResizeWidget(toplevel, ctdb->max_x/100, ctdb->max_y/100, 0);

    XtMainLoop();
    /* Free allocated space for terrain element rendering */
    STDDEALLOC(te_pts);
}
