#ifndef lint
static char rcsid [] = "$RCSfile: dc_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: dc_init.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to read and assemble translation databases           *
*       Created: Mon May  3 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libdc_local.h"
#include <libotmatch.h>
#include <libreader.h>
#include <libprotocol.h>
#include <p_dis.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/
#include <stdio.h>
#include <stdlib.h>

static READER_UNION *traversal;
int32 disconst_dis_version;
static READER_UNION *warheads;
READER_UNION disconst_artic_trans;
READER_UNION *disconst_entity_list;

/* Tree of DIS->SIMNET translations */
DISCONST_NODE *disconst_translation_tree[N_DIS_VERSIONS] = 
{(DISCONST_NODE *)NULL, (DISCONST_NODE *)NULL };
DISCONST_FUZE disconst_fuze_translation[N_DIS_VERSIONS][MAX_FUZES];
int32 disconst_number_fuzes;


static void disconst_read_aggregate(
    READER_UNION *trans, 
    DISCONST_UNIT *simnet_to_dis,
    DISCONST_UNIT *dis_to_simnet);


DISCONST_UNIT disconst_unitEnvironment_simnet_to_dis;
DISCONST_UNIT disconst_unitCountry_simnet_to_dis;
DISCONST_UNIT disconst_unitSize_simnet_to_dis;
DISCONST_UNIT disconst_unitRoleLight_simnet_to_dis;
DISCONST_UNIT disconst_unitRoleHeavy_simnet_to_dis;

DISCONST_UNIT disconst_unitEnvironment_dis_to_simnet;
DISCONST_UNIT disconst_unitCountry_dis_to_simnet;
DISCONST_UNIT disconst_unitSize_dis_to_simnet;
DISCONST_UNIT disconst_unitRoleLight_dis_to_simnet;
DISCONST_UNIT disconst_unitRoleHeavy_dis_to_simnet;

/* libotmatch key for SIMNET->DIS object type translations */
int32 disconst_key;
int32 disconst_fuze_key;

/* Flag whether to use Atlas DIS protocol variations or not: */
int32 disconst_use_atlas;

static int32 proof_trans(
    READER_UNION *entry,
    int32 which);
static void add_trans(
    READER_UNION *entry);

int32 disconst_init(
    char *data_path,
    uint32 flags,
    int32 version,
    int32 use_atlas_variations)
{
    int32 i, j;
    READER_UNION translations;

    int32 ret = reader_read("disconst.rdr", data_path, &translations,
			    flags & ~READER_TYPING);

    if (ret != 0)
      return ret;

    ret = reader_read("artic.rdr", data_path, &disconst_artic_trans,
		      flags & ~READER_TYPING);

    if (ret != 0)
      return ret;

    disconst_use_atlas = use_atlas_variations;
    disconst_set_dis_version(version);

    /* Grab the traversal instructions, the list of countries, and
     * the list of warheads
     */
    traversal = translations.array[1].array;
    warheads = translations.array[2].array;

    /* Check the warhead list for errors */
    for (i=1;i<warheads[0].integer;i++)
    {
	if (warheads[i].array[0].integer != 3)
	{
	    fprintf(stderr, "LIBDISCONST ERROR: Warhead #%d not of form:\n  (<index macro> <value>)\n",
		    i);
	    return READER_READ_ERROR;
	}
	  
	if (warheads[i].array[1].integer != i-1)
	{
	    fprintf(stderr, "LIBDISCONST ERROR: Warhead #%d (%d) out of order\n",
		    i, warheads[i].array[1].integer);
	    return READER_READ_ERROR;
	}
    }

    disconst_entity_list = translations.array[3].array;

    /* Build the translation tree */
    for (i=1;i<translations.array[3].array[0].integer;i++)
    {
	if (proof_trans(translations.array[3].array[i].array, i))
	  add_trans(translations.array[3].array[i].array);
    }    

    /* Make the otmatch simnet->dis conversion database */
    disconst_key = otm_incorporate_db((OTM_DB *)translations.array[3].array,
				      TRUE, NULL);



    /* Build the translation tree for fuzes (detonators) */
    disconst_number_fuzes = 0;
    
    for (i=1;i<translations.array[4].array[0].integer;i++)
    {
	if (translations.array[4].array[i].array[2].array[0].integer != 
	    N_DIS_VERSIONS + 1)    
	  fprintf(stderr, "LIBDISCONST ERROR: Translation #%d (%s) not of form:\n  (SIMNET_fuze DIS100 DIS203 DIS204)\n", 
		  protocol_find_name(translations.array[4].array[i].array[2].array[1].integer));
	else
	{
	    if (disconst_number_fuzes < MAX_FUZES-1)
	    {		
		for (j = 0; j < N_DIS_VERSIONS; j++)
		{
		    disconst_fuze_translation[j][i-1].simnet_name = 
		      translations.array[4].array[i].array[1].integer;
		    disconst_fuze_translation[j][i-1].dis_value =
		    translations.array[4].array[i].array[2].array[j+1].integer;
		}	 
		disconst_number_fuzes++;

	    }
	}	
    }    
    
    /* Make the otmatch simnet->dis conversion database */
    disconst_fuze_key = 
      otm_incorporate_db((OTM_DB *)translations.array[4].array,
			 TRUE, NULL);

    /*
    ** Read the aggregrate type field translations.
    */

    /* Read Environment */
    disconst_read_aggregate(&(translations.array[5].array[1]),
			    &disconst_unitEnvironment_simnet_to_dis, 
			    &disconst_unitEnvironment_dis_to_simnet);

    /* Read Country */
    disconst_read_aggregate(&(translations.array[5].array[2]),
			    &disconst_unitCountry_simnet_to_dis, 
			    &disconst_unitCountry_dis_to_simnet);

    /* Read Size */
    disconst_read_aggregate(&(translations.array[5].array[3]),
			    &disconst_unitSize_simnet_to_dis, 
			    &disconst_unitSize_dis_to_simnet);

    /* Read Role */
    disconst_read_aggregate(&(translations.array[5].array[4]),
			    &disconst_unitRoleLight_simnet_to_dis, 
			    &disconst_unitRoleLight_dis_to_simnet);

    disconst_read_aggregate(&(translations.array[5].array[4]),
			    &disconst_unitRoleHeavy_simnet_to_dis, 
			    &disconst_unitRoleHeavy_dis_to_simnet);

    return 0;
}

/* Look at the translation entry and confirm that it looks OK
 */
static int32 proof_trans(
    READER_UNION *entry,
    int32 which)
{
    int32 i, j;
    READER_UNION *dis;

    if (entry[0].integer != 3)
    {
	fprintf(stderr, "LIBDISCONST ERROR: Translation #%d (%s) not of form:\n  (SIMNET ((Kind Domain Country Category Subcategory Specific Extra Warhead) ...))\n",
		which, protocol_find_name(entry[1].integer));
	return FALSE;
    }

    /* Make sure there are enough DIS version datums */
    if (entry[2].array[0].integer != N_DIS_VERSIONS + 1)
    {
	fprintf(stderr, "LIBDISCONST ERROR: Translation #%d doesn't have %d DIS versions.\n",
		which, N_DIS_VERSIONS);
	return FALSE;
    }

    /* Make sure each DIS version datum has the right number of fields 
     * and that the numbers are in the right range.
     */
    for (j = 0; j < N_DIS_VERSIONS; j++)
    {
	dis = entry[2].array[j+1].array;

	if (dis[0].integer != 9)
	{
	    fprintf(stderr, "LIBDISCONST ERROR: Translation #%d (%s) DIS version #%d not of form:\n  (Kind Domain Country Category Subcategory Specific Extra Warhead)\n",
		    which, protocol_find_name(entry[1].integer));
	    return FALSE;
	}

	/* Make sure all the numeric fields have good small numbers in them */
	for (i=1;i<dis[0].integer;i++)
	  if ((uint32)dis[i].integer > 0xFF)
	  {
	      fprintf(stderr, "LIBDISCONST ERROR: DIS field %d of translation #%d out of range\n",
		      i, which);
	      return FALSE;
	  }
    }


    return TRUE;
}

/* Add this translation to the global translation tree.
 */
static void add_trans(
    READER_UNION *entry)
{
    int32 j;
    DISCONST_NODE *node, *last_node;
    int32 kind;
    READER_UNION *dis, *order;
    int32 i, value, width;
    ObjectType object;

    for (j = 0; j < N_DIS_VERSIONS; j++)
    {
	/* Make sure there is a root */
	if (!(node = disconst_translation_tree[j]))
	{
	    disconst_translation_tree[j] = (DISCONST_NODE *)
	      STDALLOC(sizeof(DISCONST_NODE));
	    bzero((char *)disconst_translation_tree[j], sizeof(DISCONST_NODE));
	    node = disconst_translation_tree[j];
	}

	object = entry[1].integer;
	dis = entry[2].array[j+1].array;
	kind = dis[1].integer;

	/* Get the field traversal order for entries of this kind */
	if (!(order = reader_find_tag_int(kind, traversal,
					  READER_UNTAGGED, READER_NO_ERRORS)))
	  order = traversal[1].array; /* Use the default */

	/* Iterate through the fields of the entry in order, building the
	 * tree down to this object.  Every time we create a new node,
	 * make this object the default for that node.
	 */
	for (i=2;i<order[0].integer;i++)
	{
	    value = dis[node->field+1].integer;

	    /* Make sure there is room in this node's array for this value.
	     */
	    if (value >= node->children)
	    {
		width = value+1;

		if (!node->child)
		  node->child = (union disconst_value *)
		    STDALLOC(sizeof(union disconst_value) * width);
		else
		  node->child = (union disconst_value *)
		    STDREALLOC(node->child,
			       sizeof(union disconst_value) * width);

		/* Zero out all the new slots */
		bzero((char *)(node->child + node->children),
		      (int)((width-node->children)*sizeof(union disconst_value)));
		node->children = width;
	    }

	    /* If we are at a leaf node, store the object type and move on */
	    if (node->leaf)
	    {
		/* Only store the first such object */
		if (!node->child[value].value)
		  node->child[value].value = object;
		continue;
	    }

	    /* If no node exists with the current value, make a new one */
	    last_node = node;
	    if (!(node = last_node->child[value].node))
	    {
		last_node->child[value].node =
		  (DISCONST_NODE *)STDALLOC(sizeof(DISCONST_NODE));
		bzero((char *)(last_node->child[value].node),
		      sizeof(DISCONST_NODE));
		node = last_node->child[value].node;

		/* Initialize the default of the new node to the current
		 * object (if an object gets down to this new node, but
		 * then cannot match any further, it will default to the
		 * current object).
		 */
		node->defalt = object;

		/* Note which field is the key for this new node */
		node->field = order[i+1].integer;

		/* Note if this new node is a leaf node */
		node->leaf = ((i+2) == order[0].integer);
	    }
	}
    }
}

/* Translate a warhead index to its DIS code number
 */
uint16 disconst_warhead_code(
    int32 index)
{
    if (!index)
      return 0;

    return warheads[index+1].array[2].integer;
}

/* Translate a DIS code number to its warhead index
 */
uint8 disconst_warhead_index(
    uint16 code)
{
    uint8 i;

    for(i=1;i<warheads[0].integer;i++)
      if (warheads[i].array[2].integer == code)
	return i-1;
    
    return 0;
}

/*
** int disconst_item_compare()
**
** Comparison function for quicksort (qsort).
*/
static int disconst_item_compare(
    const void *a,
    const void *b)
{
    DISCONST_ITEM *i = (DISCONST_ITEM *)a;
    DISCONST_ITEM *j = (DISCONST_ITEM *)b;

    return (i->key - j->key);
}

/*
** void disconst_read_aggregate()
**
** Read the simnet to dis and dis to simnet aggregate type
** field translations from the READER specificed by "trans".
**
** Store the mappings in the sorted arrays "simnet_to_dis"
** and "dis_to_simnet".
**
** The first pair in the mapping ALWAYS specifies
** the default value to use if the mapping is unresolved.
*/
static void disconst_read_aggregate(
    READER_UNION *trans,
    DISCONST_UNIT *simnet_to_dis,
    DISCONST_UNIT *dis_to_simnet)
{
    int i, size;

    size = trans->array[0].integer - 1;

    /* Number of elements in array (subtract one for default entry). */
    simnet_to_dis->size = dis_to_simnet->size = size - 1;
 

    simnet_to_dis->item =
      (DISCONST_ITEM *)STDALLOC(sizeof(DISCONST_ITEM)*size);
    dis_to_simnet->item =
      (DISCONST_ITEM *)STDALLOC(sizeof(DISCONST_ITEM)*size);
    
    simnet_to_dis->default_value = trans->array[1].array[2].integer;
    dis_to_simnet->default_value = trans->array[1].array[1].integer;
    
    for (i = 2; i <= size; i++)
    {
      simnet_to_dis->item[i-2].key = trans->array[i].array[1].integer;
      simnet_to_dis->item[i-2].value = trans->array[i].array[2].integer;
      dis_to_simnet->item[i-2].key = trans->array[i].array[2].integer;
      dis_to_simnet->item[i-2].value = trans->array[i].array[1].integer;
    }
    qsort(simnet_to_dis->item, simnet_to_dis->size, 
	  sizeof(DISCONST_ITEM), disconst_item_compare);
    qsort(dis_to_simnet->item, dis_to_simnet->size, 
	  sizeof(DISCONST_ITEM), disconst_item_compare);
}
