#ifndef lint
static char rcsid [] = "$RCSfile: dc_lookup.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: dc_lookup.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to do translations                                   *
*       Created: Tue May  4 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libdc_local.h"
#include <stdio.h>
#include <libreader.h>
#include <libconstants.h>
#include <libprotocol.h>
#include <libotmatch.h>
#include <p_dis.h>
#include <dis203_values.h>
#include <veh_appear.h>
#include <obj_type.h>
#include <unit_type.h>
#include <stdext.h>



static void print_node(
    DISCONST_NODE *node);
static void disconst_appearance_simnet_to_atlas(
    DIS_ENTITY_TYPE *dis_type,
    uint32 *simnet_p,
    DIS_ENTITY_APPEARANCE *dis_p);
static int32 disconst_lookup(
    DISCONST_UNIT *table_unit,
    uint32 key);


static int32 dis_table_index;

/*
 * Choose the DIS protocol, and figure out what index into the tables
 * will be used for this protocol.
 */
void disconst_set_dis_version(
    int32 dis_version)
{
    disconst_dis_version = dis_version;

    if (disconst_dis_version == DIS_PROTOCOL_VERSION_JUL92)
      dis_table_index = 0;
    else if (disconst_dis_version == DIS_PROTOCOL_VERSION_NOV93)
      dis_table_index = 1;
    else
      dis_table_index = 2;
}

/* Convert a SIMNET guise to its DIS equivalent.
 */
void disconst_guise_simnet_to_dis(
    ObjectType *simnet,
    DIS_ENTITY_TYPE *dis,
    DIS_WARHEAD_TYPE *warhead)
{
    READER_UNION *dis_data;

    if (!(dis_data = (READER_UNION *)otm_query(disconst_key,
					       (uint32)(*simnet))))
      bzero(dis, sizeof(DIS_ENTITY_TYPE));
    else
    {
	dis->entity_kind  = dis_data[dis_table_index+1].array[1].integer;
	dis->domain       = dis_data[dis_table_index+1].array[2].integer;
	dis->country      = dis_data[dis_table_index+1].array[3].integer;
	dis->category     = dis_data[dis_table_index+1].array[4].integer;
	dis->sub_category = dis_data[dis_table_index+1].array[5].integer;
	dis->specific     = dis_data[dis_table_index+1].array[6].integer;
	dis->extra        = dis_data[dis_table_index+1].array[7].integer;
	if (warhead)
	  *warhead        = disconst_warhead_code(dis_data[dis_table_index+1].array[8].integer);
    }
}

/* Convert a SIMNET fuze to its DIS equivalent.
 */
void disconst_fuze_simnet_to_dis(
    ObjectType *simnet,
    uint16     *dis)
{
    READER_UNION *dis_data;

    if (!(dis_data = (READER_UNION *)otm_query(disconst_fuze_key,
					       (uint32)(*simnet))))
      bzero(dis, sizeof(uint16));
    else
      *dis = dis_data[dis_table_index+1].integer;

}

/* Convert a DIS fuze to its SIMNET equivalent.
 */
void disconst_fuze_dis_to_simnet(
    uint16     *dis,
    ObjectType *simnet)
{
    uint32  i;

    for (i=0; i < disconst_number_fuzes; i++)
    {
	if (disconst_fuze_translation[dis_table_index][i].dis_value == *dis)
	{	    
	    *simnet = disconst_fuze_translation[dis_table_index][i].simnet_name;
	    break;
	}
    }
    return;
    

}

/* Converts a DIS guise to its SIMNET equivalent.
 */
void disconst_guise_dis_to_simnet(
    DIS_ENTITY_TYPE *dis,
    DIS_WARHEAD_TYPE *warhead,
    ObjectType *simnet)
{
    DISCONST_NODE *node;
    uint16 keys[8];
    uint16 value;

    keys[0]   = dis->entity_kind;
    keys[1]   = dis->domain;
    keys[2]   = dis->country;
    keys[3]   = dis->category;
    keys[4]   = dis->sub_category;
    keys[5]   = dis->specific;
    keys[6]   = dis->extra;
    if (warhead)
      keys[7] = disconst_warhead_index(*warhead);

    node = disconst_translation_tree[dis_table_index];
    for(;;)
    {
	value = keys[node->field];

	/* Check for unrecognized values */
	if ((value >= node->children) || !node->child[value].value)
	{
	    *simnet = node->defalt;
	    return;
	}

	/* If we're at a leaf node, return the result */
	if (node->leaf)
	{
	    *simnet = node->child[value].value;
	    return;
	}

	/* Go down the tree one step */
	node = node->child[value].node;
    }

    /*NOTREACHED*/
}

/* Converts a SIMNET appearance to its DIS 1.0 equivalent
 */
static void appearance_simnet_to_dis1(
    DIS_ENTITY_TYPE *dis_type,
    uint32 *simnet_p,
    DIS_ENTITY_APPEARANCE *dis_p)
{
    uint32 simnet, dis;

    simnet = *simnet_p;
    dis = 0;

    switch (dis_type->entity_kind)
    {
      case DIS_ENTITY_KIND_PLATFORM:
	switch(dis_type->domain)
	{
	  case DIS_ENTITY_DOMAIN_LAND:
	    dis = ((DIS_ENTITY_APPEARANCE_LAND_DESTROYED &
		    -((simnet & vehDestroyed) != 0)) |
		   (DIS_ENTITY_APPEARANCE_LAND_SMOKE_PLUME &
		    -((simnet & vehSmokePlume) != 0)) |
		   (DIS_ENTITY_APPEARANCE_LAND_FLAMING &
		    -((simnet & vehFlaming) != 0)) |
		   (DIS_ENTITY_APPEARANCE_LAND_LAUNCHER_RAISED &
		    -((simnet & vehLauncherUp) != 0)) |
		   (DIS_ENTITY_APPEARANCE_LAND_ENGINE_SMOKE &
		    -((simnet & vehEngineSmoke) != 0)));
	    if ((simnet & vehDustCloudMask) == vehDustCloudSmall)
	      dis |= DIS_ENTITY_APPEARANCE_LAND_SMALL_DUST_CLOUD;
	    else if ((simnet & vehDustCloudMask) == vehDustCloudMedium)
	      dis |= DIS_ENTITY_APPEARANCE_LAND_MEDIUM_DUST_CLOUD;
 	    else if ((simnet & vehDustCloudMask) == vehDustCloudLarge)
	      dis |= DIS_ENTITY_APPEARANCE_LAND_LARGE_DUST_CLOUD;
	    break;

	  case DIS_ENTITY_DOMAIN_AIR:
	    dis = ((DIS_ENTITY_APPEARANCE_AIR_DESTROYED &
		    -((simnet & vehDestroyed) != 0)) |
		   (DIS_ENTITY_APPEARANCE_AIR_FLAMING &
		    -((simnet & vehFlaming) != 0)));
	    break;

	  case DIS_ENTITY_DOMAIN_SURFACE:
	    dis = ((DIS_ENTITY_APPEARANCE_SURFACE_DESTROYED &
		    -((simnet & vehDestroyed) != 0)) |
		   (DIS_ENTITY_APPEARANCE_SURFACE_FLAMING &
		    -((simnet & vehFlaming) != 0)));
	    if ((simnet & vehWakeMask) == vehWakeSmall)
	      dis |= DIS_ENTITY_APPEARANCE_SURFACE_SMALL_WAKE;
	    else if ((simnet & vehWakeMask) == vehWakeMedium)
	      dis |= DIS_ENTITY_APPEARANCE_SURFACE_MEDIUM_WAKE;
	    else if ((simnet & vehWakeMask) == vehWakeLarge)
	      dis |= DIS_ENTITY_APPEARANCE_SURFACE_LARGE_WAKE;
	    break;

	  case DIS_ENTITY_DOMAIN_SUBSURFACE:
	    dis = ((DIS_ENTITY_APPEARANCE_SUB_SURFACE_DESTROYED &
		    -((simnet & vehDestroyed) != 0)));
	    if ((simnet & vehWakeMask) == vehWakeSmall)
	      dis |= DIS_ENTITY_APPEARANCE_SUB_SURFACE_SMALL_WAKE;
	    else if ((simnet & vehWakeMask) == vehWakeMedium)
	      dis |= DIS_ENTITY_APPEARANCE_SUB_SURFACE_MEDIUM_WAKE;
	    else if ((simnet & vehWakeMask) == vehWakeLarge)
	      dis |= DIS_ENTITY_APPEARANCE_SUB_SURFACE_LARGE_WAKE;
	    break;

	  case DIS_ENTITY_DOMAIN_SPACE:
	    dis = ((DIS_ENTITY_APPEARANCE_SPACE_DESTROYED &
		    -((simnet & vehDestroyed) != 0)));
	    break;
	}
	break;
      case DIS_ENTITY_KIND_MUNITION:
	/* No equivalent SIMNET bits */
	break;
      case DIS_ENTITY_KIND_LIFE_FORM:
	if (simnet & vehDestroyed)
	  dis = 0;
	else if ((simnet & lfPositionMask) == lfPositionStanding)
	  dis |= DIS_ENTITY_APPEARANCE_LIFE_UPRIGHT;
	else if ((simnet & lfPositionMask) == lfPositionKneeling)
	  dis |= DIS_ENTITY_APPEARANCE_LIFE_KNEELING;
	else if ((simnet & lfPositionMask) == lfPositionProne)
	  dis |= DIS_ENTITY_APPEARANCE_LIFE_PRONE;
	break;
      case DIS_ENTITY_KIND_ENVIRONMENTAL:
	/* No equivalent SIMNET bits */
	break;
      case DIS_ENTITY_KIND_CULTURAL_FEATURE:
	dis = ((DIS_ENTITY_APPEARANCE_CULTURAL_DESTROYED &
		-((simnet & vehDestroyed) != 0)) |
	       (DIS_ENTITY_APPEARANCE_CULTURAL_SMOKE_PLUME &
		-((simnet & vehSmokePlume) != 0)) |
	       (DIS_ENTITY_APPEARANCE_CULTURAL_FLAMING &
		-((simnet & vehFlaming) != 0)));
	break;
    }
    *dis_p = dis;
}

static void disconst_appearance_simnet_to_atlas(
    DIS_ENTITY_TYPE *dis_type,
    uint32 *simnet_p,
    DIS_ENTITY_APPEARANCE *dis_p)
{
    uint32 simnet, dis;

    simnet = *simnet_p;
    dis = *dis_p;

    /* These bits were overlayed for AE on top of the SIMET protocol */
    switch(simnet & lfAEUseWeaponsMask)
    {
    case lfAEUseWeapon1:
	dis |=  DIS203_ENTITY_APPEARANCE_LIFE_WEAPON_1_DEPLOYED;
	break;
    case lfAEUseWeapon2:
	dis |=  DIS203_ENTITY_APPEARANCE_LIFE_WEAPON_2_DEPLOYED;
	break;
    case lfAEUseBothWeapons:
	dis |=  DIS203_ENTITY_APPEARANCE_LIFE_WEAPON_1_DEPLOYED;
	dis |=  DIS203_ENTITY_APPEARANCE_LIFE_WEAPON_2_DEPLOYED;
	break;
    case lfAEUseNoWeapons:
    default:
	break;
    }

    if ((simnet & lfAEDeactivateMask) == lfAEDeactivated)
      dis |=  DIS203AE_ENTITY_APPEARANCE_LIFE_DEACTIVATED;

    switch(simnet & lfAEMountMask)
    {
    case lfAEMountDismounting:
	dis |= DIS203AE_ENTITY_APPEARANCE_LIFE_MOUNT_DISMOUNTING;
	break;
    case lfAEMountMounting:
	dis |= DIS203AE_ENTITY_APPEARANCE_LIFE_MOUNT_MOUNTING;
	break;
    case lfAEMountMounted:
	dis |= DIS203AE_ENTITY_APPEARANCE_LIFE_MOUNT_MOUNTED;
	break;
    case lfAEMountInvalid:
    default:
	break;
    }

    switch(simnet & lfAEFormationMask)
    {
    case lfAEFormationReihe:
	dis |= DIS203AE_ENTITY_APPEARANCE_LIFE_FORMATION_REIHE;
	break;
    case lfAEFormationRudel:
	dis |= DIS203AE_ENTITY_APPEARANCE_LIFE_FORMATION_RUDEL;
	break;
    case lfAEFormationDefense:
	dis |= DIS203AE_ENTITY_APPEARANCE_LIFE_FORMATION_DEFENSE;
	break;
    case lfAEFormationOther:
    default:
	break;
    }

    *dis_p = dis;
}

/* Converts a SIMNET appearance to its DIS 2.xx equivalent.  DIS 2.03
 * and DIS 2.04 are the same.
 */
static void appearance_simnet_to_dis2xx(
    DIS_ENTITY_TYPE *dis_type,
    uint32 *simnet_p,
    DIS_ENTITY_APPEARANCE *dis_p)
{
    uint32 simnet, dis;

    simnet = *simnet_p;
    dis = 0;

    dis = ((DIS203_ENTITY_APPEARANCE_GENERAL_MOBILITY_KILL_YES &
	    -((simnet & vehMobilityDisabled) != 0)) |
	   (DIS203_ENTITY_APPEARANCE_GENERAL_FIRE_POWER_KILL_YES &
	    -((simnet & vehFirepowerDisabled) != 0)) |
	   (DIS203_ENTITY_APPEARANCE_GENERAL_DAMAGE_DESTROYED &
	    -((simnet & vehDestroyed) != 0)) |
	   (DIS203_ENTITY_APPEARANCE_GENERAL_SMOKE_PLUME &
	    -((simnet & vehSmokePlume) != 0)) |
	   (DIS203_ENTITY_APPEARANCE_GENERAL_SMOKE_ENGINE &
	    -((simnet & vehEngineSmoke) != 0)) |
	   (DIS203_ENTITY_APPEARANCE_GENERAL_FLAMING_FLAMES_PRESENT &
	    -((simnet & vehFlaming) != 0)));

    if ((simnet & vehDustCloudMask) == vehDustCloudSmall)
      dis |= DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_SMALL;
    else if ((simnet & vehDustCloudMask) == vehDustCloudMedium)
      dis |= DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_MEDIUM;
    else if ((simnet & vehDustCloudMask) == vehDustCloudLarge)
      dis |= DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_LARGE;

    switch (dis_type->entity_kind)
    {
      case DIS_ENTITY_KIND_PLATFORM:
	switch(dis_type->domain)
	{
	  case DIS_ENTITY_DOMAIN_LAND:
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_LAUNCHER_RAISED &
		     -((simnet & vehLauncherUp) != 0)));
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_BRIDGE_DISABLED_YES &
		     -((simnet & vehBridgeDisabled) != 0))); 
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_SPOIL_DISPLAYED &
		     -((simnet & vehBladeSpoil) != 0))); 
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_SHOVEL_DISABLED_YES &
		     -((simnet & vehShovelDisabled) != 0))); 
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_PRIMARY_ROCKET_PRESENT &
		     -((simnet & vehPrimMICLICpresent) != 0))); 	    
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_SECONDARY_ROCKET_PRESENT &
		     -((simnet & vehSecMICLICpresent) != 0))); 	    
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_PRIMARY_ROCKET_DISABLED_YES &
		     -((simnet & vehPrimMiclicDisabled) != 0))); 	    
	    dis |= ((DIS203_ENTITY_APPEARANCE_LAND_SECONDARY_ROCKET_DISABLED_YES &
		     -((simnet & vehSecMiclicDisabled) != 0))); 	    
	    break; 
	  case DIS_ENTITY_DOMAIN_AIR:
	    /* No equivalent SIMNET bits */
	    break;
	}
	break;
      case DIS_ENTITY_KIND_MUNITION:
	dis |= ((DIS203_ENTITY_APPEARANCE_MUNITION_CHARGE_ATTACHED &
		 -((simnet & munTailLineAttached) != 0)));
	break;
      case DIS_ENTITY_KIND_LIFE_FORM:
	if (simnet & vehDestroyed)
	  ;
	else if ((simnet & lfPositionMask) == lfPositionKneeling)
	  dis |= DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_KNEELING;
	else if ((simnet & lfPositionMask) == lfPositionProne)
	  dis |= DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_PRONE;
	else
	  dis |= DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_UPRIGHT;
	if (disconst_use_atlas)
	    disconst_appearance_simnet_to_atlas(dis_type, simnet_p, &dis);
	break;
      case DIS_ENTITY_KIND_ENVIRONMENTAL:
	/* No equivalent SIMNET bits */
	break;
    }
    *dis_p = dis;
}

/* Converts a SIMNET appearance to its DIS equivalent.
 */
void disconst_appearance_simnet_to_dis(
    DIS_ENTITY_TYPE *dis_type,
    uint32 *simnet_p,
    DIS_ENTITY_APPEARANCE *dis_p)
{
    if (disconst_dis_version == DIS_PROTOCOL_VERSION_JUL92)
      appearance_simnet_to_dis1(dis_type, simnet_p, dis_p);
    else  /* DIS 2.03 and DIS 2.04 are the same */
      appearance_simnet_to_dis2xx(dis_type, simnet_p, dis_p);
}

/* Converts a DIS 1.0 appearance to its SIMNET equivalent.
 */
static void appearance_dis1_to_simnet(
    DIS_ENTITY_TYPE *dis_type,
    DIS_ENTITY_APPEARANCE *dis_p,
    uint32 *simnet_p)
{
    uint32 simnet, dis;

    dis = *dis_p;
    simnet = 0;

    switch (dis_type->entity_kind)
    {
      case DIS_ENTITY_KIND_PLATFORM:
	switch(dis_type->domain)
	{
	  case DIS_ENTITY_DOMAIN_LAND:
	    simnet =
	      ((vehDestroyed &
		-((dis & DIS_ENTITY_APPEARANCE_LAND_DESTROYED) != 0)) |
	       (vehSmokePlume &
		-((dis & DIS_ENTITY_APPEARANCE_LAND_SMOKE_PLUME) != 0)) |
	       (vehFlaming &
		-((dis & DIS_ENTITY_APPEARANCE_LAND_FLAMING) != 0)) |
	       (vehLauncherUp &
		-((dis & DIS_ENTITY_APPEARANCE_LAND_LAUNCHER_RAISED) != 0)) |
	       (vehEngineSmoke &
		-((dis & DIS_ENTITY_APPEARANCE_LAND_ENGINE_SMOKE) != 0)));
	    if ((dis & DIS_ENTITY_APPEARANCE_LAND_LARGE_DUST_CLOUD) ==
		DIS_ENTITY_APPEARANCE_LAND_SMALL_DUST_CLOUD)
	      simnet |= vehDustCloudSmall;
	    else if ((dis & DIS_ENTITY_APPEARANCE_LAND_LARGE_DUST_CLOUD) ==
		     DIS_ENTITY_APPEARANCE_LAND_MEDIUM_DUST_CLOUD)
	      simnet |= vehDustCloudMedium;
	    else if ((dis & DIS_ENTITY_APPEARANCE_LAND_LARGE_DUST_CLOUD) ==
		     DIS_ENTITY_APPEARANCE_LAND_LARGE_DUST_CLOUD)
	      simnet |= vehDustCloudLarge;
	    break;

	  case DIS_ENTITY_DOMAIN_AIR:
	    simnet =
	      ((vehDestroyed &
		-((dis & DIS_ENTITY_APPEARANCE_AIR_DESTROYED) != 0)) |
	       (vehFlaming &
		-((dis & DIS_ENTITY_APPEARANCE_AIR_FLAMING) != 0)));
	    break;

	  case DIS_ENTITY_DOMAIN_SURFACE:
	    simnet =
	      ((vehDestroyed &
		-((dis & DIS_ENTITY_APPEARANCE_SURFACE_DESTROYED) != 0)) |
	       (vehFlaming &
		-((dis & DIS_ENTITY_APPEARANCE_SURFACE_FLAMING) != 0)));
	    if ((dis & DIS_ENTITY_APPEARANCE_SURFACE_LARGE_WAKE) ==
		DIS_ENTITY_APPEARANCE_SURFACE_SMALL_WAKE)
	      simnet |= vehWakeSmall;
	    else if ((dis & DIS_ENTITY_APPEARANCE_SURFACE_LARGE_WAKE) ==
		     DIS_ENTITY_APPEARANCE_SURFACE_MEDIUM_WAKE)
	      simnet |= vehWakeMedium;
	    else if ((dis & DIS_ENTITY_APPEARANCE_SURFACE_LARGE_WAKE) ==
		     DIS_ENTITY_APPEARANCE_SURFACE_LARGE_WAKE)
	      simnet |= vehWakeLarge;
	    break;

	  case DIS_ENTITY_DOMAIN_SUBSURFACE:
	    simnet =
	      ((vehDestroyed &
		-((dis & DIS_ENTITY_APPEARANCE_SUB_SURFACE_DESTROYED) != 0)));
	    if ((dis & DIS_ENTITY_APPEARANCE_SUB_SURFACE_LARGE_WAKE) ==
		DIS_ENTITY_APPEARANCE_SUB_SURFACE_SMALL_WAKE)
	      simnet |= vehWakeSmall;
	    else if ((dis & DIS_ENTITY_APPEARANCE_SUB_SURFACE_LARGE_WAKE) ==
		     DIS_ENTITY_APPEARANCE_SUB_SURFACE_MEDIUM_WAKE)
	      simnet |= vehWakeMedium;
	    else if ((dis & DIS_ENTITY_APPEARANCE_SUB_SURFACE_LARGE_WAKE) ==
		     DIS_ENTITY_APPEARANCE_SUB_SURFACE_LARGE_WAKE)
	      simnet |= vehWakeLarge;
	    break;

	  case DIS_ENTITY_DOMAIN_SPACE:
	    simnet =
	      ((vehDestroyed &
		-((dis & DIS_ENTITY_APPEARANCE_SPACE_DESTROYED) != 0)));
	    break;
	}
	break;
      case DIS_ENTITY_KIND_MUNITION:
	/* No equivalent SIMNET bits */
	break;
      case DIS_ENTITY_KIND_LIFE_FORM:
	if ((dis & DIS_ENTITY_APPEARANCE_LIFE_PRONE) ==
	    DIS_ENTITY_APPEARANCE_LIFE_UPRIGHT)
	  simnet = lfPositionStanding;
	else if ((dis & DIS_ENTITY_APPEARANCE_LIFE_PRONE) ==
		 DIS_ENTITY_APPEARANCE_LIFE_KNEELING)
	  simnet = lfPositionKneeling;
	else if ((dis & DIS_ENTITY_APPEARANCE_LIFE_PRONE) ==
		 DIS_ENTITY_APPEARANCE_LIFE_PRONE)
	  simnet = lfPositionProne;
	else
	  simnet = vehDestroyed;
	break;
      case DIS_ENTITY_KIND_ENVIRONMENTAL:
	/* No equivalent SIMNET bits */
	break;
      case DIS_ENTITY_KIND_CULTURAL_FEATURE:
	simnet =
	  ((vehDestroyed &
	    -((dis & DIS_ENTITY_APPEARANCE_CULTURAL_DESTROYED) != 0)) |
	   (vehSmokePlume &
	    -((dis & DIS_ENTITY_APPEARANCE_CULTURAL_SMOKE_PLUME) != 0)) |
	   (vehFlaming &
	    -((dis & DIS_ENTITY_APPEARANCE_CULTURAL_FLAMING) != 0)));
	break;
    }
    *simnet_p = simnet;
}

/* Converts a DIS 2.xx appearance to its SIMNET equivalent.
 * DIS 2.03 and DIS 2.04 are the same. 
*/
static void appearance_dis2xx_to_simnet(
    DIS_ENTITY_TYPE *dis_type,
    DIS_ENTITY_APPEARANCE *dis_p,
    uint32 *simnet_p)
{
    uint32 simnet, dis;

    dis = *dis_p;
    simnet = 0;

    simnet =
      ((vehMobilityDisabled &
	-((dis & DIS203_ENTITY_APPEARANCE_GENERAL_MOBILITY_KILL_YES) != 0)) |
       (vehFirepowerDisabled &
	-((dis & DIS203_ENTITY_APPEARANCE_GENERAL_FIRE_POWER_KILL_YES) != 0)) |
       (vehDestroyed &
	-((dis & DIS203_ENTITY_APPEARANCE_GENERAL_DAMAGE_DESTROYED) != 0)) |
       (vehSmokePlume &
	-((dis & DIS203_ENTITY_APPEARANCE_GENERAL_SMOKE_PLUME) != 0)) |
       (vehEngineSmoke &
	-((dis & DIS203_ENTITY_APPEARANCE_GENERAL_SMOKE_ENGINE) != 0)) |
       (vehFlaming &
	-((dis & DIS203_ENTITY_APPEARANCE_GENERAL_FLAMING_FLAMES_PRESENT) != 0)));

    if ((dis & DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_LARGE) ==
	DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_SMALL)
      simnet |= vehDustCloudSmall;
    else if ((dis & DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_LARGE) ==
	     DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_MEDIUM)
      simnet |= vehDustCloudMedium;
    else if ((dis & DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_LARGE) ==
	     DIS203_ENTITY_APPEARANCE_GENERAL_TRAILING_EFFECTS_LARGE)
      simnet |= vehDustCloudLarge;

    switch (dis_type->entity_kind)
    {
      case DIS_ENTITY_KIND_PLATFORM:
	switch(dis_type->domain)
	{
	  case DIS_ENTITY_DOMAIN_LAND:
	    simnet |=
	      ((vehLauncherUp &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_LAUNCHER_RAISED) != 0)));
	    simnet |=
	      ((vehBridgeDisabled &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_BRIDGE_DISABLED_YES) != 0)));
	    simnet |=
	      ((vehBladeSpoil &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_SPOIL_DISPLAYED) != 0)));
	    simnet |=
	      ((vehShovelDisabled &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_SHOVEL_DISABLED_YES) != 0)));
	    simnet |=
	      ((vehPrimMICLICpresent &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_PRIMARY_ROCKET_PRESENT) != 0)));
	    simnet |=
	      ((vehSecMICLICpresent &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_SECONDARY_ROCKET_PRESENT) != 0)));
	    simnet |=
	      ((vehPrimMiclicDisabled &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_PRIMARY_ROCKET_DISABLED_YES) != 0)));
	    simnet |=
	      ((vehSecMiclicDisabled &
		-((dis & DIS203_ENTITY_APPEARANCE_LAND_SECONDARY_ROCKET_DISABLED_YES) != 0)));

	    break;

	  case DIS_ENTITY_DOMAIN_AIR:
	    /* No equivalent SIMNET bits */
	    break;
	}
	break;
      case DIS_ENTITY_KIND_MUNITION:
	simnet |=
	  ((munTailLineAttached &
	    -((dis & DIS203_ENTITY_APPEARANCE_MUNITION_CHARGE_ATTACHED) != 0)));

	break;
      case DIS_ENTITY_KIND_LIFE_FORM:
	if ((dis & DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_PRONE) ==
	    DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_UPRIGHT)
	  simnet |= lfPositionStanding;
	else if ((dis & DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_PRONE) ==
		 DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_KNEELING)
	  simnet |= lfPositionKneeling;
	else if ((dis & DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_PRONE) ==
		 DIS203_ENTITY_APPEARANCE_LIFE_LIFE_FORM_STATE_PRONE)
	  simnet |= lfPositionProne;
	else
	  simnet |= vehDestroyed;
	break;
      case DIS_ENTITY_KIND_ENVIRONMENTAL:
	/* No equivalent SIMNET bits */
	break;
    }
    *simnet_p = simnet;
}

/* Converts a DIS appearance to its SIMNET equivalent.
 */
void disconst_appearance_dis_to_simnet(
    DIS_ENTITY_TYPE *dis_type,
    DIS_ENTITY_APPEARANCE *dis_p,
    uint32 *simnet_p)
{
    if (disconst_dis_version == DIS_PROTOCOL_VERSION_JUL92)
      appearance_dis1_to_simnet(dis_type, dis_p, simnet_p);
    else   /* DIS 2.03 and DIS 2.04 are the same */
      appearance_dis2xx_to_simnet(dis_type, dis_p, simnet_p);
}

/* Converts a SIMNET capabilities record to its DIS equivalent.
 */
void disconst_capabilities_simnet_to_dis(
    VehicleCapabilities *simnet,
    DIS_ENTITY_CAPABILITIES *dis)
{
    dis->ammunition_supply = simnet->ammunitionSupply;
    dis->fuel_supply = simnet->fuelSupply;
    dis->repair = simnet->repair;
}

/* Converts a DIS capabilities record to its SIMNET equivalent.
 */
void disconst_capabilities_dis_to_simnet(
    DIS_ENTITY_CAPABILITIES *dis,
    VehicleCapabilities *simnet)
{
    simnet->ammunitionSupply = dis->ammunition_supply;
    simnet->fuelSupply = dis->fuel_supply;
    simnet->repair = dis->repair;
}

/* Print the DIS->SIMNET translation tree
 */
void disconst_print(void)
{
    printf("Kind\tDom\tCount\tCat\tScat\tSpec\tExtr\n");

    print_node(disconst_translation_tree[dis_table_index]);
}

/* Code to print a single node of the tree */
static void print_node(
    DISCONST_NODE *node)
{
    int32 i, j;
    char *name;

    for (i=0;i<node->children;i++)
      if (node->child[i].value)
      {
	  for (j=0;j<node->field;j++)
	    putchar('\t');

	  if (i && (node->field == WARHEAD_FIELD))
	    printf("%4d", disconst_warhead_code(i));
	  else
	    printf("%3d", i);

	  if (node->leaf)
	  {
	      if ((name = protocol_find_name((int32)node->child[i].value)) ==
		  const_error)
		printf("  0x%08x\n", node->child[i].value);
	      else
		printf("  %s\n", name);
	  }
	  else
	  {
	      putchar('\n');
	      print_node(node->child[i].node);
	  }
      }
}


/* disconst_articulation_enum:
 * Translates the symbolic name for articulated parts to the appropriate 
 * DIS constant value.  This assumes the name is a libreader symbol.
 * Returns -l, if error.
 */
int32 disconst_articulation_enum(
    char *name)
{
    READER_UNION   *artic_table;
    
    /* Search for name in reader file */
    artic_table = reader_find_tag(name, disconst_artic_trans.array, 
				  READER_UNTAGGED, 
				  READER_NO_ERRORS);

    /* If name found, get associated enumerated value */
    if (artic_table)
      if (disconst_dis_version == DIS_PROTOCOL_VERSION_JUL92)
      {	  
	  return artic_table[2].integer;
      }
      else if (disconst_dis_version == DIS_PROTOCOL_VERSION_NOV93)
      {
	  return artic_table[3].integer;
      }	
      else
      {
	  return artic_table[4].integer;
      }	

    /* Can't find it, return error */
    return -1;
}



/* disconst_articulation_name:
 * Translates the DIS constant value to a symbolic name for articulated 
 * parts.  Returns error, if it couldn't find enum constant.
 */
char *disconst_articulation_name(
    int32 enum_const,
    int32 *error)
{
    uint32  i;
    uint32  col;
    
    /* Default the error to no error */
    *error = FALSE;
    
    /* Determine which column to search */
    if (disconst_dis_version == DIS_PROTOCOL_VERSION_JUL92)
      col = 2;
    else if (disconst_dis_version == DIS_PROTOCOL_VERSION_NOV93)
      col = 3;
    else
      col = 4;
       
    /* Search for the user specified enum_constant, and return the name */
    for (i=0; i < disconst_artic_trans.array[0].integer - 1; i++)
    {
	if (disconst_artic_trans.array[i+1].array[col].integer ==
	    enum_const)
	  return disconst_artic_trans.array[i+1].array[1].charptr;
    }

    /* Can't find it, return error */
    *error = TRUE;
    return NULL;
    
}	


/*
** void disconst_aggregate_simnet_to_dis_type()
**
** Return the equivalent DIS aggregate type translation
** for a given SIMNET aggregate type.
**
** If the simnet object represents a single vehicle
** (indicated by the unitSize being either Squad or Section),
** set the entity_kind in the DIS field to DISAggregateKindOther
** and the category field (specifing EchelonLevel) to IndividualVehicle.
*/
void disconst_aggregate_simnet_to_dis_type(
    ObjectType *simnet,
    DIS_ENTITY_TYPE *dis)
{
    if (((*simnet & unitSizeMask) == unitSizeSquad) ||
	(*simnet & unitSizeMask) == unitSizeSquad)
    {
	/* Single vehicle */
	dis->entity_kind = DISAggregateKindOther;
	dis->category = DISEchelonLevelIndVeh;
    }
    else
    {
	dis->entity_kind = DISAggregateKindHierarchy;
	dis->category = disconst_lookup(&disconst_unitSize_simnet_to_dis,
					*simnet & unitSizeMask);
    }

    dis->domain = (uint8) disconst_lookup(
                                  &disconst_unitEnvironment_simnet_to_dis,
		                  *simnet & unitEnvironmentMask);
	

    dis->country = (uint16) disconst_lookup(
                                        &disconst_unitCountry_simnet_to_dis,
				        *simnet & unitCountryMask);

    dis->category = disconst_lookup(&disconst_unitSize_simnet_to_dis,
				    *simnet & unitSizeMask);
    /*
    ** Use a different LUT depending on the modifer of the 
    ** simnet object type.
    */
    if ((*simnet & unitModifierMask) == unitModifierHeavy)
    {
	dis->sub_category = disconst_lookup(
				       &disconst_unitRoleHeavy_simnet_to_dis,
				       *simnet & unitRoleMask);
    }
    else
    {
	dis->sub_category = disconst_lookup(
					&disconst_unitRoleLight_simnet_to_dis,
					*simnet & unitRoleMask);
    }
    dis->specific = (*simnet & unitHQMask) ? 1 : 0;
    dis->extra = 0;
}


/*
** void disconst_aggregate_dis_to_simnet()
**
** Return the equivalent SIMNET aggregate type translation
** for a given DIS aggregate type.
**
** If the DIS object indicates a single vehicle (as specified
** by the entity_kind field being DISAggregateKindOther), set
** the unitSize of the simnet object to Squad.
**
** Note: Single vehicles are not being handled correctly,
** it should be a mapping between objectDomainVehicle
** and DIS, not objectDomainUnit.
*/
void disconst_aggregate_dis_to_simnet_type(
    DIS_ENTITY_TYPE *dis,
    ObjectType *simnet)
{
    *simnet = 0;

    *simnet |= objectDomainUnit;
    
    *simnet |= disconst_lookup(&disconst_unitEnvironment_dis_to_simnet,
			       dis->domain);

    *simnet |= disconst_lookup(&disconst_unitCountry_dis_to_simnet,
			       dis->country);

    if (dis->entity_kind == DISAggregateKindOther)
    {
	/* Single vehicle */
	*simnet |= 0x0;
    }
    else
    {
	*simnet |= disconst_lookup(&disconst_unitSize_dis_to_simnet,
				   dis->category);
    }

    /*
    ** If the DIS echelon type would be considered with
    ** a modifier of "heavy" under simnet, set that field
    ** and use the appropriate LUT.
    */
    if ((dis->sub_category == DISEchelonTypeMechanHeavyTaskForce) ||
	(dis->sub_category == DISEchelonTypeArmorHeavyTaskForce))
    {
	*simnet |= disconst_lookup(&disconst_unitRoleHeavy_dis_to_simnet,
				   dis->sub_category);
	*simnet |= unitModifierHeavy;
    }
    else 
    {
	*simnet |= disconst_lookup(&disconst_unitRoleLight_dis_to_simnet,
				   dis->sub_category);
    }
	
    *simnet |= (dis->specific) ? unitHQ : 0;
}

/*
** int32 disconst_lookup()
**
** Return the value associated a key in the table.
** If the key is not found, return the default mapping value.
**
** Table is assumed to be a sorted array of type DISCONST_ITEM.
** Table is searched using a iterative binary search.
*/
static int32 disconst_lookup(
    DISCONST_UNIT *table_unit,
    uint32 key)
{
    int32 pivot;
    int32 head = 0;
    int32 tail = table_unit->size-1;
    DISCONST_ITEM *table_item = table_unit->item;
    
    while (1)
    {
	if (head > tail) return (table_unit->default_value);

	pivot = (head + tail)/2;

	if (key == table_item[pivot].key) 
	{ 
	    return (table_item[pivot].value);
	}
	if (key < table_item[pivot].key)
	  tail = pivot-1;
	else
	  head = pivot+1;
    }
}

