/*
 * $RCSfile: libdc_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libdc_local.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libdisconst                       *
*       Created: Mon May  3 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libdisconst.h"
#include <libreader.h>
#include <basic.h>

/* Structure of dis->simnet object type translation tree nodes */
typedef struct disconst_node
{
    /* Which field is the key at this node */
    uint8 field;

    /* Note the field where country and warhead are stored (we need to
     * translate these using a lookup table).
     */
#define COUNTRY_FIELD 2
#define WARHEAD_FIELD 7

    /* Is this a leaf node? */
    uint8 leaf;

    /* How wide is this node */
    uint8 children;

    /* Default object type if the input key does not match any of the
     * defined sub-tree values.
     */
    ObjectType defalt;

    /* Array of sub-nodes (internal nodes) or object types (leaf nodes)
     * keyed directly by field of input key.
     */
    union disconst_value
    {
	struct disconst_node *node;
	ObjectType value;
    } *child;

} DISCONST_NODE;



/* Structure of dis->simnet fuze type translation */
typedef struct disconst_fuze
{
    ObjectType  simnet_name;
    uint16      dis_value;
} DISCONST_FUZE;


typedef struct disconst_item
{
    uint32 key;
    uint32 value;
} DISCONST_ITEM;

typedef struct disconst_unit
{
  uint32 size;           /* Size of mapping array (not including defaults) */
  uint32 default_value;
  DISCONST_ITEM *item;
} DISCONST_UNIT;

/*
 * This constant defines the number of different DIS enumerations supported.
 * This constant must match the number of translations provided for each
 * SIMNET object type in the file "disconst.rdr".
 */
#define N_DIS_VERSIONS 3

#define MAX_FUZES 25

/* Private declarations */
extern int32 disconst_use_atlas;
extern DISCONST_NODE *disconst_translation_tree[];
extern DISCONST_FUZE disconst_fuze_translation[N_DIS_VERSIONS][MAX_FUZES];
extern int32 disconst_number_fuzes;
extern int32 disconst_key;
extern int32 disconst_fuze_key;
extern READER_UNION disconst_artic_trans;
extern READER_UNION *disconst_entity_list;
extern int32 disconst_dis_version;
extern DISCONST_UNIT disconst_unitEnvironment_simnet_to_dis;
extern DISCONST_UNIT disconst_unitCountry_simnet_to_dis;
extern DISCONST_UNIT disconst_unitSize_simnet_to_dis;
extern DISCONST_UNIT disconst_unitRoleLight_simnet_to_dis;
extern DISCONST_UNIT disconst_unitRoleHeavy_simnet_to_dis;

extern DISCONST_UNIT disconst_unitEnvironment_dis_to_simnet;
extern DISCONST_UNIT disconst_unitCountry_dis_to_simnet;
extern DISCONST_UNIT disconst_unitSize_dis_to_simnet;
extern DISCONST_UNIT disconst_unitRoleLight_dis_to_simnet;
extern DISCONST_UNIT disconst_unitRoleHeavy_dis_to_simnet;

