/*
 * $RCSfile: libdisconst.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libdisconst.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header for libdisconst                             *
*       Created: Mon May  3 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBDISCONST_H_INCLUDED
#define _LIBDISCONST_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> /*common/include/global*/
#include <p_dis.h>
#include <p_dis203.h>
#include <p_sim.h>
#include <libconstants.h>
#undef Time
#undef Boolean

/* disconst_init:
 *
 * Initializes libdisconst, causing it to read its data file (disconst.rdr)
 * from the passed directory (or .) using the specified reader_flags.
 * The return value is 0 for success, or one of the libreader return values.
 * The version of DIS being used should be passed in dis_version.
 */
extern int32 disconst_init(
    char *data_path,
    uint32 flags,
    int32 version,
    int32 use_atlas_variations);

/* disconst_set_dis_version:
 *
 * Set the version of DIS constants to be used during translations.
 */
extern void disconst_set_dis_version(
    int32 dis_version);

/* disconst_print:
 *
 * Prints the DIS->SIMNET translation tree in a manner similar to the
 * DIS specification.
 */
extern void disconst_print(void);

/* disconst_guise_simnet_to_dis:
 *
 * Converts a SIMNET guise to its DIS equivalent.  If the warhead is
 * not desired, pass NULL.
 */
extern void disconst_guise_simnet_to_dis(
    ObjectType *simnet,
    DIS_ENTITY_TYPE *dis,
    DIS_WARHEAD_TYPE *warhead);

/* disconst_fuze_simnet_to_dis:
 *
 * Converts a SIMNET fuze to its DIS equivalent.
 */
extern void disconst_fuze_simnet_to_dis(
    ObjectType *simnet,
    uint16 *dis);

/* disconst_fuze_dis_to_simnet:
 *
 * Converts a DIS fuze to its SIMNET equivalent.
 */
extern void disconst_fuze_dis_to_simnet(
    uint16 *dis,
    ObjectType *simnet);

/* disconst_guise_dis_to_simnet:
 *
 * Converts a DIS guise to its SIMNET equivalent.  If the warhead is
 * not relevant, pass NULL for 'warhead'.
 */
extern void disconst_guise_dis_to_simnet(
    DIS_ENTITY_TYPE *dis,
    DIS_WARHEAD_TYPE *warhead,
    ObjectType *simnet);

/* disconst_appearance_simnet_to_dis:
 *
 * Converts a SIMNET appearance to its DIS equivalent.
 */
extern void disconst_appearance_simnet_to_dis(
    DIS_ENTITY_TYPE *dis_type,
    uint32 *simnet_p,
    DIS_ENTITY_APPEARANCE *dis_p);

/* disconst_appearance_dis_to_simnet:
 *
 * Converts a DIS appearance to its SIMNET equivalent.
 */
extern void disconst_appearance_dis_to_simnet(
    DIS_ENTITY_TYPE *dis_type,
    DIS_ENTITY_APPEARANCE *dis_p,
    uint32 *simnet_p);

/* disconst_capabilities_simnet_to_dis:
 *
 * Converts a SIMNET capabilities record to its DIS equivalent.
 */
extern void disconst_capabilities_simnet_to_dis(
    VehicleCapabilities *simnet,
    DIS_ENTITY_CAPABILITIES *dis);

/* disconst_capabilities_dis_to_simnet:
 *
 * Converts a DIS capabilities record to its SIMNET equivalent.
 */
extern void disconst_capabilities_dis_to_simnet(
    DIS_ENTITY_CAPABILITIES *dis,
    VehicleCapabilities *simnet);


/* disconst_articulation_enum:
 * Translates the symbolic name for articulated parts to the DIS constant 
 * value.  Returns -1 if error.
 */
extern int32 disconst_articulation_enum(
    char *name);

/* disconst_articulation_name:
 * Translates the DIS constant value to a symbolic name for articulated 
 * parts.  Returns TRUE in error, if error.
 */
extern char *disconst_articulation_name(
    int32 enum_const,
    int32 *error);


/* disconst_warhead_code:
 * Translates the warhead index to its DIS warhead type code. 
 */
extern uint16 disconst_warhead_code(
    int32 index);

/* disconst_warhead_index:
 * Translates the DIS warhead type code to its wahead index.
 */
extern uint8 disconst_warhead_index(
    uint16 code);


#define DISAggregateKindOther 0
#define DISAggregateKindHierarchy 1
#define DISEchelonLevelIndVeh 1
#define DISEchelonTypeArmorHeavyTaskForce 17
#define DISEchelonTypeMechanHeavyTaskForce 19

/*
** disconst_aggregate_simnet_to_dis_type()
**
** Translate a SIMNET aggregate type to a DIS aggregate type.
*/
extern void disconst_aggregate_simnet_to_dis_type(
    ObjectType *simnet,
    DIS_ENTITY_TYPE *dis);
/*
** disconst_aggregate_dis_to_simnet_type()
**
** Translate a DIS aggregate type to a SIMNET aggregate type.
*/
extern void disconst_aggregate_dis_to_simnet_type(
    DIS_ENTITY_TYPE *dis,
    ObjectType *simnet);

#ifdef __cplusplus
}
#endif

#endif /*_LIBDISCONST_H_INCLUDED*/
