#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libdisconst                              *
*       Created: Tue May  4 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libreader.h>
#include <libprotocol.h>
#include <libconstants.h>
#include <libotmatch.h>
#include "libdisconst.h"
#include "libdc_local.h"
#include <basic.h>
#include <p_dis.h>
#include <stdext.h> /*common/include/global*/
#include <stdio.h>
#include <stdlib.h>
#ifdef PYTHON_PARSER
#    include <Py/Python.h>
#    include <py_stdpython.h>
#endif

static void check_unique(void);

int main(
    int argc,
    char **argv)
{
    ObjectType ot;
    DIS_ENTITY_TYPE et;
    DIS_WARHEAD_TYPE wh;
    int32 ver;

#ifdef ALPHA_SHORT_PTR
    {
      int i;
      for (i=1; i<argc; i++)
	argv[i] = argv[2*i];
    }
#endif /* ALPHA_SHORT_PTR */

#ifdef PYTHON_PARSER
    Py_Initialize(); /* Need this because libreader has a python module */
#endif

    /* Initialize libreader */
    reader_init(0);

    /* Initialize libconstants */
    const_init("../../data/constants");

    /* Initialize libprotocol */
    protocol_init();

    /* Initialize libotmatch */
    if (otm_init("../../data", 0))
      exit(0);

    if ((argc == 1) ||
	((ver = atoi(argv[1])) < 1) || (ver > 4))
    {
	printf("\nUsage:\t%s <DIS Version> -print\n\t%s <DIS Version> -unique\n\t%s <DIS Version> <SIMNET ObjectType>\n\t%s <DIS Version> <Kind> <Dom> <Count> <Cat> <Scat> <Spec> <Extr> <Warhead>\n",
	       argv[0], argv[0], argv[0], argv[0]);
	exit(0);
    }

    /* Initialize libdisconst */
    if (disconst_init("", 0, ver, FALSE))
      exit(0);
{

    ObjectType simnet = 0xe8120040;
    DIS_ENTITY_TYPE dis;

    bzero(&dis, sizeof(int32)*2);
    
    disconst_aggregate_simnet_to_dis_type(&simnet, &dis);
    simnet = 0;
    disconst_aggregate_dis_to_simnet_type(&dis, &simnet);
    
    simnet = 0xe8120380;
    bzero(&dis, sizeof(int32)*2);

    disconst_aggregate_simnet_to_dis_type(&simnet, &dis);
    simnet = 0;
    disconst_aggregate_dis_to_simnet_type(&dis, &simnet);
    disconst_aggregate_simnet_to_dis_type(&simnet, &dis);
}

    if (!strcmp(argv[2],"-print"))
      disconst_print();
    else if (!strcmp(argv[2],"-unique"))
      check_unique();
    else if (argc == 3)
    {
	ot = const_name_to_value(reader_get_symbol(argv[2]));
	disconst_guise_simnet_to_dis(&ot, &et, &wh);
	printf("0x%x == %d %d %d %d %d %d %d %d\n",
	       ot,
	       et.entity_kind,
	       et.domain,
	       et.country,
	       et.category,
	       et.sub_category,
	       et.specific,
	       et.extra,
	       wh);
    }
    else if (argc == 10)
    {
	et.entity_kind  = atoi(argv[2]);
	et.domain       = atoi(argv[3]);
	et.country      = atoi(argv[4]);
	et.category     = atoi(argv[5]);
	et.sub_category = atoi(argv[6]);
	et.specific     = atoi(argv[7]);
	et.extra        = atoi(argv[8]);
	wh              = atoi(argv[9]);

	disconst_guise_dis_to_simnet(&et, &wh, &ot);
	printf("%d %d %d %d %d %d %d %d == 0x%x (%s)\n",
	       et.entity_kind,
	       et.domain,
	       et.country,
	       et.category,
	       et.sub_category,
	       et.specific,
	       et.extra,
	       wh,
	       ot, protocol_find_name(ot));
    }

    return(0);
}


/* This function checks for unique values in the disconst.rdr file, and
   reports any conflicts.
 */
static void check_unique(void)
{
    ObjectType ot, new_ot;
    DIS_ENTITY_TYPE et;
    DIS_WARHEAD_TYPE wh;
    int32 i;
    int32 err = FALSE;

    printf("Checking for unique DIS %d values in disconst.rdr.\n",
	   disconst_dis_version);

    /* for each entity ... */
    for (i=1;i< disconst_entity_list[0].integer; i++)
    {
	ot = (ObjectType) disconst_entity_list[i].array[1].integer;

	disconst_guise_simnet_to_dis(&ot, &et, &wh);

	disconst_guise_dis_to_simnet(&et, &wh, &new_ot);

	/* Does the original object type match the new one? */
	if (ot != new_ot)
	{
	    err = TRUE;
	    printf("Error: Conflict beteween %s and %s == ",
		   protocol_find_name(disconst_entity_list[i].array[1].integer),
		   protocol_find_name(new_ot));

	    printf("%d %d %d %d %d %d %d %d\n",
		   et.entity_kind,
		   et.domain,
		   et.country,
		   et.category,
		   et.sub_category,
		   et.specific,
		   et.extra,
		   wh);
	}
    }    

    if (!err)
      printf("There are no conflicts in disconst.rdr for DIS version %d.\n",
	     disconst_dis_version);
}
