#ifndef lint
static char rcsid [] = "$RCSfile: ddbo_rtest.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ddbo_rtest.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to test locals and remotes in libdisdbobj
*       Created: Mon Jan 30 17:12:18 EST 1995
*       Author: ajc
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libdisdbobj.h"

#include <stdtypes.h> /*common/include/global*/
#include <stdext.h>   /*common/include/global*/
#include <stdalloc.h> /*common/include/global*/

#include <libtime.h>
#include <libsched.h>
#include <libcmdline.h>
#include <libvtab.h>
#include <libexcept.h>
#include <librandom.h>
#include <libreader.h>
#include <librdrconst.h>
#include <libparmgr.h>
#include <libpo.h>

#include <stdio.h>

/*
 * Options structure to parse command line arguments.
 */
struct main_options_struct
{
    CMD_INTEGER_OPTION remotes;
    CMD_INTEGER_OPTION locals;
    CMD_INTEGER_OPTION remote_tick_rate;
    CMD_INTEGER_OPTION local_tick_rate;
    CMD_INTEGER_OPTION duration;
    CMD_INTEGER_OPTION delete_prob;
    CMD_BOOLEAN_OPTION verbose;
    CMD_BOOLEAN_OPTION benchmark;
};

struct main_options_struct main_options = {
    {
	"Number of Remotes" , "Specifies the number of remote vehicles" ,
	NULL,
	CMD_INTEGER ,	"remotes" ,	NULL ,		10 ,
    } ,
    {
	"Number of Locals" , "Specifies the number of local vehicles" ,
	NULL,
	CMD_INTEGER ,	"locals" ,	NULL ,		10 ,
    } ,
    {
	"Remote Tick Rate" , "Specifies tick rate (msecs) of remote vehicles" ,
	NULL,
	CMD_INTEGER ,	"rtick" ,	NULL ,		67 ,
    } ,
    {
	"Local Tick Rate" , "Specifies tick rate (msecs) of local vehicles" ,
	NULL,
	CMD_INTEGER ,	"ltick" ,	NULL ,		67 ,
    } ,
    {
	"Test Duration" , "Specifies the duration of the test in minutes" ,
	NULL,
	CMD_INTEGER ,	"duration" ,	NULL ,		3 ,
    },
    {
	"Deletion Probability" , "Specifies the probability of deleting a vehicle" ,
	NULL,
	CMD_INTEGER ,	"delete" ,	NULL ,		0 ,
    },
    {
	"Verbose" , "Specifies whether to print out each tick" ,
	NULL,
	CMD_BOOLEAN ,	"verbose" ,	"silent",		FALSE,
    },
    {
	"Benchmark" , "Specifies whether to print out benchmark stats" ,
	NULL,
	CMD_BOOLEAN ,	"bench" ,	"nobench",		FALSE,
    },
};

/* Forward references */
static void maybe_delete_vehicle();
static void print_random_vehicle();
static void subclass1_init();
static void subclass2_init();

/*
 * Utility function to init all the local libraries needed to initialize
 * parmgr for the creation of local entities.
 */
static void init_local_libraries()
{
    TerrainDatabaseID   terrain_id;
    PO_DATABASE        *po_db_p;
    SimulationAddress   simaddr;
    READER_UNION        data_file;

    /* We need to init libparmgr.  To do that we need to open:
     * libreader, and librdrconst.
     */
    reader_init(0);
    rdc_init("../data", READER_OVERRIDE);

    /* Normally libparmgr is initialized with an opened PO database, but
     * it can accept a NULL.  Also, if libdisdbobj and libparmgr are
     * compiled with NO_PO defined, than libpo can avoid being
     * linked in completely.
     */
    pm_init(NULL);

    /* Read the data file */
    reader_read("ddbo_rtest.rdr", NULL, &data_file, 0);

    /* Give the data file to libparmgr */
    pm_set_data_table(data_file.array);
}

int main(argc, argv)
    int   argc;
    char *argv[];
{
    int32               leftover_argc;
    char *             *leftover_argv;
    int32               i;
    VehicleID           simnet_id;
    int32               vehicle_id;
    VTAB_USER_DATA_TYPE unused_user_data;

    /* Herald */
    printf("Libdisdbobj Test Program.\n\n");

    /* Initialize libtime/libsched first, because we need to start posting
     * tick functions (such as tick_X) right away.
     */
    time_init(5000);
    sched_init(2, 67, 29000);

    /* First turn on exceptions. */
    except_init(TRUE, TRUE);

    /* Initialize related libraries */
    cmd_process_options(argc, argv,                     /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized   */
			(CMD_OPTION *)&main_options,    /* options array  */
			sizeof(main_options),           /* size of array  */
			TRUE                            /* Verbose mode   */
			);

    /* Complain about any unrecognized options */
    cmd_gripe(leftover_argc, leftover_argv);

    /* Initialize the vehicle table with site 1, host 1 */
    vtab_init(1, 1, 1);

    /* Initialize the stuff needed to create locals */
    init_local_libraries();    

    /* Initialize the disdbobj */
    disdbobj_init();

    /* Init some fake subclasses */
    subclass1_init();
    subclass2_init();

    /* Create some remote objects */
    printf("Creating %d remotes...\n", main_options.remotes.value);

    for (i = 0; i < main_options.remotes.value; i++)
    {
	/* Just make up a simnet id (pretend it came off the net) */
	vtab_create_simnet_id(&simnet_id);

	/* What would this hash to? */
	vehicle_id = vtab_vehicle_number(&simnet_id);
	
	/* Create the remote */
	disdbobj_create_remote(vehicle_id, &simnet_id,
			       VTAB_REMOTE_VEHICLE,
			       main_options.remote_tick_rate.value);

	/* Print each remote as it is created */
	if (main_options.verbose.value)
	{
	    printf("Created remote vehicle %d\n", vehicle_id);
	    class_show((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
		       "all");
	}
    }

    /* Create some local objects */
    printf("Creating %d locals...\n", main_options.locals.value);

    for (i = 0; i < main_options.locals.value; i++)
    {
	char *type;

	/* Request a simnet id */
	vtab_create_simnet_id(&simnet_id);

	/* Figure out what type of vehicle to make */
	if (i % 2)
	  type = reader_get_symbol("B52");
	else
	  type = reader_get_symbol("AWACS");
	
	/* Create the local */
	vehicle_id = 
	  disdbobj_create_local(type,
				VTAB_LOCAL_VEHICLE,
				main_options.local_tick_rate.value,
				&simnet_id,
				NULL);           /* Unit entry */

	/* Print each local as it is created */
	if (main_options.verbose.value)
	{
	    printf("Created local vehicle %d\n", vehicle_id);
	    class_show((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
		       "all");
	}
    }

    /* Schedule a function to randomly delete objects */
    sched_periodic_fncl(maybe_delete_vehicle, 1000, 1000, 0, A_END);

    /* Schedule a function to randomly print an object */
    if (main_options.verbose.value)
      sched_periodic_fncl(print_random_vehicle, 10000, 10000, 0, A_END);

    /*
     * Fire up the scheduler for N minutes
     */
    printf("Starting simulation...\n");

    if (main_options.benchmark.value)
      sched_perf_monitor_on(249000);

    sched_invoke_functions_until(main_options.duration.value * 60 * 1000);
    printf("End of simulation.\n");

    if (main_options.benchmark.value)
      sched_perf_monitor_flush();

    /*
     * When done, optionally print out lots of stuff.
     */
    if (main_options.verbose.value)
    {
	printf("Printing all remaining vehicles.\n");

	for (vehicle_id = vtab_get_first(vtab_vehicle_table, VTAB_ALL,
					 &unused_user_data);
	     vehicle_id;
	     vehicle_id = vtab_get_next(vtab_vehicle_table, VTAB_ALL,
					&unused_user_data))
	  class_show((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
		     "all");
    }

    /* Print summary statistics */
    printf("There were %d vehicles at the end of the simulation.\n",
	   vtab_count_vehicles(vtab_vehicle_table, VTAB_ALL));

    /* End of test */
    printf("Test complete.\n");
}

/*
 * Utility to randomly delete vehicles during execution.
 */
static void maybe_delete_vehicle()
{
    int32               n_vehicles;
    int32               unlucky_vehicle;
    VTAB_USER_DATA_TYPE unused_user_data;
    int32               is_local;

    if (rnd_probability(main_options.delete_prob.value/100.0))
    {
	unlucky_vehicle = 
	  vtab_get_first(vtab_vehicle_table, VTAB_ALL, &unused_user_data);

	if (unlucky_vehicle)
	{
	    is_local = 
	      VTAB_TYPES_MATCH(vtab_vehicle_type(unlucky_vehicle),
			       VTAB_LOCAL_VEHICLE);

	    printf("Randomly deleting %s vehicle %d\n",
		   is_local ? "local" : "remote",
		   unlucky_vehicle);

	    /* Dispatch on whether to delete the vehicle as a local/remote */
	    if (is_local)
	      disdbobj_destroy_local(unlucky_vehicle, FALSE);
	    else
	      disdbobj_destroy_remote(unlucky_vehicle);

	    /* A little something to sound out when I delete a vehicle */
	    putchar(7); fflush(stdout);
	}
	
	/* See if all are gone now */
	n_vehicles = vtab_count_vehicles(vtab_vehicle_table, VTAB_ALL);

	if (n_vehicles == 0)
	{
	    /* There don't exist any more vehicles.  There is no
	     * use in waiting.
	     */
	    printf("All vehicles have been deleted.  Ending test early.\n");
	    exit(0);
	}
    }
}

/*
 * Utility to randomly delete print a vehicle during execution.
 */
static void print_random_vehicle()
{
    int32               n_vehicles = vtab_count_vehicles(vtab_vehicle_table,
							 VTAB_ALL);
    int32               nth_vehicle = rnd_uniform_dist() * n_vehicles;
    int32               random_vehicle;
    int32               i;
    VTAB_USER_DATA_TYPE unused_user_data;

    /* Sometimes this can be called with no vehicles in the table.
     * in that case, bail out early.
     */
    if (!n_vehicles)
      return;

    /* Iterate nth_vehicle times, until we get the random vehicle we want. */
    random_vehicle = vtab_get_first(vtab_vehicle_table, 
				    VTAB_ALL, &unused_user_data);
    i = 1;

    while (i < nth_vehicle)
    {
	random_vehicle = vtab_get_next(vtab_vehicle_table,
				       VTAB_ALL, &unused_user_data);
	i++;
    }

    /* Make sure i really found the vehicle */
    if (random_vehicle)
    {
	/* Call the show routine on this random vehicle */
	printf("Randomly printing vehicle %d, which is the %dth vehicle out of %d\n",
	       random_vehicle, nth_vehicle+1, n_vehicles);
	
	class_show((CLASS_USER_DATA_TYPE)vtab_get_vehicle(random_vehicle),
		   "all");
    }
    else
    {
	/* This is wierd.  I am supposed to always find a vehicle. */
	printf("TEST ERROR:  print_random_vehicle() saw %d vehicles and\n tried to pick the %dth one, but it wasn't there!\n",
	       n_vehicles, nth_vehicle);
	abort();
    }
}

/*
 * Subclass 1 simulation.
 *
 * Remote subclass which post an event on MODELING every second.
 */

static int32 subclass1_user_data_handle;

typedef struct subclass1_vars
{
    char    name[100];
    int32   sb1;
    float64 sb2;
    int32   n_modeling_ticks;
} SUBCLASS1_VARS;

static void subclass1_print(subclass1)
    SUBCLASS1_VARS *subclass1;
{
    printf("Subclass1:\n  Name: \"%s\", variable 1 = %d, variable 2 = %.4f\n",
	   subclass1->name, subclass1->sb1, subclass1->sb2);
    printf("  %d modeling ticks\n",
	   subclass1->n_modeling_ticks);
}

static void subclass1_physical_modeling(vehicle_id, interval_since_last_tick, 
					user_data)
    int32   vehicle_id;
    uint32  interval_since_last_tick;
    int32   user_data;
{
    SUBCLASS1_VARS *subclass1 =
      (SUBCLASS1_VARS *) class_get_user_data(vtab_get_vehicle(vehicle_id), 
					     subclass1_user_data_handle);

    if (!subclass1) /* Passive error detection */
      return;
    if (main_options.verbose.value)
      printf("Vehicle %d MODELING tick #%d.  Last tick was %.3f seconds ago.\n",
	     vehicle_id,
	     subclass1->n_modeling_ticks,
	     (interval_since_last_tick / 1000.0));
    
    /* Up my little input counter */
    subclass1->n_modeling_ticks++;
}

static void subclass1_create(vehicle_id)
    int32 vehicle_id;
{
    SUBCLASS1_VARS *subclass1 =
      (SUBCLASS1_VARS *)STDALLOC(sizeof(SUBCLASS1_VARS));
    static int32 var1 = 0;

    bzero((char *)subclass1, sizeof(SUBCLASS1_VARS));

    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			subclass1_user_data_handle,
			(CLASS_USER_DATA_TYPE)subclass1);
    /* 
     * Give this subclass some data including the vehicle number and
     * a random digit between 0 and 1000.
     */
    sprintf(subclass1->name,
	    "SUB 1 VEH %d DIGIT %d",
	    vehicle_id,
	    ((int) (rnd_uniform_dist() * 1000)));
    /*
     * Give this subclass some interesting variables.
     */
    subclass1->sb1 = var1;
    subclass1->sb2 = var1 + (((float64) var1)/100.0);
    var1++;
    /*
     * Register callbacks to tick during certain phases at certain rates.
     */
    callback_register_handler(
                 disdbobj_get_tick_event(vehicle_id,
					 DISDBOBJ_PHYSICAL_HULL_MODELING, 
					 1000),
			      subclass1_physical_modeling,
			      (ADDRESS)50); /* Dummy user data */

}

static void subclass1_add_remote_subclass(vehicle_id, user_data)
    int32   vehicle_id;
    int32   user_data;
{
    /* Only add the subclass if this is actually a remote vehicle */
    if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_REMOTE_VEHICLE))
      subclass1_create(vehicle_id);
}

static void subclass1_destroy(vehicle_id, is_migration)
    int32 vehicle_id;
    int32 is_migration;
{
    SUBCLASS1_VARS *subclass1 =
      (SUBCLASS1_VARS *) class_get_user_data(vtab_get_vehicle(vehicle_id), 
					     subclass1_user_data_handle);

    if (!subclass1) /* Passive error detection */
      return;

    /* Clear out the data */
    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			subclass1_user_data_handle,
			(CLASS_USER_DATA_TYPE)0);

    /* Free the memory */
    STDDEALLOC(subclass1);
}

static void subclass1_remove_remote_subclass(vehicle_id, is_migration,
					     user_data)
    int32   vehicle_id;
    int32   is_migration;
    int32   user_data;
{
    if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_REMOTE_VEHICLE))
      subclass1_destroy(vehicle_id, is_migration);
}
	
static void subclass1_init()
{
    /*
     * Declare my membership in the disdbobj class.
     */
    subclass1_user_data_handle =
      class_reserve_user_data(disdbobj_class, "subclass1", subclass1_print);

    /*
     * Register callbacks to create and destroy subclasses.
     */
    callback_register_handler(disdbobj_create_event,
			      subclass1_add_remote_subclass, 
			      (ADDRESS)41); /* Dummy user data */
    callback_register_handler(disdbobj_destroy_event,
			      subclass1_remove_remote_subclass, 
			      (ADDRESS)42); /* Dummy user data */
}
 
/*
 * Subclass 2 simulation.
 *
 * Local subclass, which posts events on physical modeling
 */
static int32 subclass2_user_data_handle;

typedef struct subclass2_vars
{
    char    name[100];
    int32   aa1;
    int32   parm1;
    int32   n_modeling_ticks;
    int32   n_output_ticks;
    int32   n_output_pdus;
} SUBCLASS2_VARS;

static void subclass2_print(subclass2)
    SUBCLASS2_VARS *subclass2;
{
    printf("Subclass2:\n  Name: \"%s\", parm1 = %d, variable 1 = %d.\n",
	   subclass2->name, subclass2->parm1, subclass2->aa1);
    printf("  %d input ticks, %d output ticks, %d output PDUs\n",
	   subclass2->n_modeling_ticks,
	   subclass2->n_output_ticks,
	   subclass2->n_output_pdus);
}

#define Model2 1

typedef struct
{
    int32 foo;
} MODEL2_PARAMETRIC_DATA;

static MODEL2_PARAMETRIC_DATA * 
subclass2_parser(libreader_format, model_num, size)
    READER_UNION *libreader_format;
    uint32 model_num;	/* ignored for now, but required */
    int32 *size;
{
    READER_UNION *my_entry;
    MODEL2_PARAMETRIC_DATA *result;

    my_entry = reader_find_tag(reader_get_symbol("model2"),
			       libreader_format,
			       READER_TAGGED,
			       READER_ERRORS_WITH_CONTEXT);
    
    if (!my_entry)
      return NULL;

    result = (MODEL2_PARAMETRIC_DATA *)malloc(sizeof(MODEL2_PARAMETRIC_DATA));

    *size = sizeof(MODEL2_PARAMETRIC_DATA);

    result->foo = my_entry[2].array[1].integer;

    return result;
}

static void subclass2_physical_modeling(vehicle_id, interval_since_last_tick, 
				user_data)
    int32   vehicle_id;
    uint32  interval_since_last_tick;
    int32   user_data;
{
    SUBCLASS2_VARS *subclass1 =
      (SUBCLASS2_VARS *) class_get_user_data(vtab_get_vehicle(vehicle_id), 
					     subclass2_user_data_handle);

    if (!subclass1) /* Passive error detection */
      return;

    if (main_options.verbose.value)
      printf("Vehicle %d MODELING tick #%d.  Last tick was %.3f seconds ago.\n",
	     vehicle_id,
	     subclass1->n_modeling_ticks,
	     (interval_since_last_tick / 1000.0));
    
    /* Up my little input counter */
    subclass1->n_modeling_ticks++;
}

static void subclass2_pdu_output(vehicle_id, interval_since_last_tick,
				 user_data)
    int32   vehicle_id;
    uint32  interval_since_last_tick;
    int32   user_data;
{
    SUBCLASS2_VARS *subclass2 =
      (SUBCLASS2_VARS *) class_get_user_data(vtab_get_vehicle(vehicle_id), 
					     subclass2_user_data_handle);
    int32 do_send = rnd_probability(.10);

    if (!subclass2) /* Passive error detection */
      return;

    /* Print that I am ticking */
    if (main_options.verbose.value)
      printf("Vehicle %d PDU OUTPUT tick #%d.  Last tick %.3f seconds ago. %s send.\n",
	     vehicle_id,
	     subclass2->n_output_ticks,
	     (interval_since_last_tick / 1000.0),
	     do_send ? "Did" : "Didn't");
    
    /* Up my little input counter */
    subclass2->n_output_ticks++;

    if (do_send)
      subclass2->n_output_pdus++;
}

static void subclass2_create(vehicle_id, params, unit_entry)
    int32                   vehicle_id;
    MODEL2_PARAMETRIC_DATA *params;
    PO_DB_ENTRY            *unit_entry;
{
    SUBCLASS2_VARS *subclass2 =
      (SUBCLASS2_VARS *)STDALLOC(sizeof(SUBCLASS2_VARS));
    static int32 var1 = 0;

    bzero((char *)subclass2, sizeof(SUBCLASS2_VARS));

    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			subclass2_user_data_handle,
			(CLASS_USER_DATA_TYPE)subclass2);
    /* 
     * Give this subclass some data including the vehicle number and
     * a random digit between 0 and 1000.
     */
    sprintf(subclass2->name,
	    "SUB 2 VEH %d DIGIT %d",
	    vehicle_id,
	    ((int) (rnd_uniform_dist() * 1000)));
    /*
     * Give this subclass some interesting variables.
     */
    subclass2->aa1 = var1;
    subclass2->parm1 = params->foo;
    var1++;

    /*
     * Register callbacks to tick during certain phases at certain rates.
     */
    callback_register_handler(
            disdbobj_get_tick_event(vehicle_id,
				    DISDBOBJ_PHYSICAL_HULL_MODELING, 
				    0),
			      subclass2_physical_modeling,
			      (ADDRESS)43); /* Dummy user data */
    callback_register_handler(
	    disdbobj_get_tick_event(vehicle_id,
				    DISDBOBJ_PDU_OUTPUT, 
				    2500),
			      subclass2_pdu_output,
			      (ADDRESS)44); /* Dummy user data */
}

static void subclass2_destroy(vehicle_id, is_migration)
    int32 vehicle_id;
    int32 is_migration;
{
    SUBCLASS2_VARS *subclass2 =
      (SUBCLASS2_VARS *) class_get_user_data(vtab_get_vehicle(vehicle_id), 
					     subclass2_user_data_handle);

    if (!subclass2) /* Passive error detection */
      return;

    /* Clear out the data */
    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			subclass2_user_data_handle,
			(CLASS_USER_DATA_TYPE)0);

    /* Free the memory */
    STDDEALLOC(subclass2);
}

static void subclass2_init()
{
    /*
     * Declare my membership in the disdbobj class.
     */
    subclass2_user_data_handle =
      class_reserve_user_data(disdbobj_class, "subclass2", subclass2_print);

    /*
     * Tell parmgr about myself.
     */
    pm_register_parm_fcns(Model2, 
			  subclass2_parser, 
			  NULL,
			  subclass2_create,
			  subclass2_destroy);
}
