/*
 * $RCSfile: libdisdbobj.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libdisdbobj.h
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libdisdbobj
*       Created: Mon Jan 30 17:12:18 EST 1995
*       Author: ajc
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBDISDBOBJ_H_INCLUDED
#define _LIBDISDBOBJ_H_INCLUDED

#include <basic.h>
#include <libclass.h>
#include <libcallback.h>
#include <libpo.h>

/* Order of this enumeration is important.  This order determine the
 * order of ticks.
 */
typedef enum disdbobj_tick_phase
{
    DISDBOBJ_PDU_INPUT                 = 0,
    DISDBOBJ_PHYSICAL_HULL_MODELING    = 1,
    DISDBOBJ_PHYSICAL_TURRET_MODELING  = 2,
    DISDBOBJ_PHYSICAL_GUN_MODELING     = 3,
    DISDBOBJ_PDU_OUTPUT                = 4,
    DISDBOBJ_SENSORY_PROCESSING        = 5,
    DISDBOBJ_TASK_PROCESSING           = 6,
    DISDBOBJ_GUI                       = 7
} DISDBOBJ_TICK_PHASE;

#define DISDBOBJ_N_PHASES 8

/*
 * Global class pointer for objects of this class.
 */
extern CLASS_PTR disdbobj_class;

/* Public global events for when objects of this class are created and
 * destroyed
 */
extern CALLBACK_EVENT_PTR disdbobj_create_event;
extern CALLBACK_EVENT_PTR disdbobj_destroy_event;
extern CALLBACK_EVENT_PTR disdbobj_frozen_tick_event;

/* disdbobj_init:
 *
 * Initializes libdisdbobj.  Call this before calling any other libdisdbobj
 * functions.
 */
extern void disdbobj_init();

/****************************************
 * Routines to create and destroy locals.
 ****************************************/

/*
 * Public function to create a local object of this class.
 */
extern int32 disdbobj_create_local(/* char        *name_symbol,
				      uint16       vtab_type,
				      int32        tick_rate,
				      VehicleID   *simnet_id,
				      PO_DB_ENTRY *unit_entry
				      */);
/*
 * Public function to destroy local objects of this class.
 */
extern void disdbobj_destroy_local(/* int32 vehicle_id,
				      int32 is_migration
				      */);

/************************************************
 * Routines to create and destroy "empty" locals.
 ***********************************************/

/*
 * Public function to create a local empty object of this class.
 */
extern int32 disdbobj_create_empty_local(/* uint16       vtab_type */);

/*
 * Public function to destroy local empty objects of this class.
 */
extern void disdbobj_destroy_empty_local(/* int32 vehicle_id */);

/****************************************
 * Routines to create and destroy remotes.
 ****************************************/

/*
 * Public function to create remote objects of this class.
 */
extern void disdbobj_create_remote(/* int32      vehicle_id,
				      VehicleID *simnet_id,
				      uint16     vtab_type,
				      int32      tick_rate,
				      */);

/*
 * Public function to destroy remote objects of this class.
 */
extern void disdbobj_destroy_remote(/* int32 vehicle_id */);

/*
 * Public function to get the requested callback event from a
 * disdbobj.  Used by simulation code to post tick routines for
 * appropriate phases at appropriate rates.
 */
extern CALLBACK_EVENT_PTR 
disdbobj_get_tick_event(/* int32                  vehicle_id,
                           DISDBOBJ_TICK_PHASE    phase,
			   uint32                 rate_in_ms
			   */);
/*
 * Public function to start a tick for a vehicle that has 
 * previously cancelled it.  ie. libshipvls does not need
 * missiles ticked until they are launched.
 */
extern void disdbobj_activate_tick(/* int32 vehicle_id, 
                                     int32 delay, 
				     int32 tick_rate
				   */);

#endif /*_LIBDISDBOBJ_H_INCLUDED*/

