#ifndef lint
static char rcsid [] = "$RCSfile: dt_copy.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: dt_copy.c                                                         *
*                                                                           *
*       Copyright 1990 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Systems and Technology Corporation                    *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to copy a  dither into a bitmap                      *
*       Created: Fri Nov  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libdither.h"

void dither_or(size, d, mat_2D, bitmap, bitmap_width, bitmap_height,
	       x0, y0, w, h)
    int32 size;
    int32 d;
    int32 **mat_2D; /* Dimension vary */
    char *bitmap;
    int32 bitmap_width;
    int32 bitmap_height;
    int32 x0, y0;
    int32 w, h;
{
    int32 bwidth;
    int32 i,j;
    int32 x, y;
    int32 *mat = (int32 *)mat_2D;

    /* This code OR's part of a dither into the passed bitmap.
     * The code is similar to that in dither_bitmaps(), except the
     * bitmap is provided by the caller.
     */

    /* No need to OR in 0's */
    if (!d)
      return;

    /* Compute the width of the bitmap in bytes */
    bwidth = (bitmap_width+7)/8;

    /* Make sure we don't overrun our bounds */
    if (x0 + w > bitmap_width)
      w = bitmap_width-x0;
    if (y0 + h > bitmap_height)
      h = bitmap_height-y0;

#define MAT(x,y) mat[(x&(size-1))+(y&(size-1))*size]

    /* Set those bits indicated in the generating matrix */
    for(j=0;j<h;j++)
      for(i=0;i<w;i++)
      {
	  x = x0 + i;
	  y = y0 + j;
	  if (MAT(x,y) < d)
	    bitmap[y*bwidth + (x>>3)] |= 1 << (7-(x&7));
      }
}

void dither_copy(size, d, mat_2D, bitmap, bitmap_width, bitmap_height,
		 x0, y0, w, h)
    int32 size;
    int32 d;
    int32 **mat_2D; /* Dimensions vary */
    char *bitmap;
    int32 bitmap_width;
    int32 bitmap_height;
    int32 x0, y0;
    int32 w, h;
{
    int32 bwidth;
    int32 i,j;
    int32 x, y;
    int32 *mat = (int32 *)mat_2D;

    /* This is basically the same as dither_or(), except 0 bits in
     * the dither are set to 0 in the bitmap.
     */

    /* Compute the width of the bitmap in bytes */
    bwidth = (bitmap_width+7)/8;

    /* Make sure we don't overrun our bounds */
    if (x0 + w > bitmap_width)
      w = bitmap_width-x0;
    if (y0 + h > bitmap_height)
      h = bitmap_height-y0;

    /* Set those bits indicated in the generating matrix, and unset
     * all others.
     */
    for(j=0;j<h;j++)
      for(i=0;i<w;i++)
      {
	  x = x0 + i;
	  y = y0 + j;
	  if (MAT(x,y) < d)
	    bitmap[y*bwidth + (x>>3)] |= 1 << (7-(x&7));
	  else
	    bitmap[y*bwidth + (x>>3)] &= ~(1 << (7-(x&7)));
      }
}
