#ifndef lint
static char rcsid [] = "$RCSfile: dt_matrix.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: dt_matrix.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to generate a dither matrix                          *
*       Created: Fri Nov  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libdither.h>
#include <stdio.h>
#include <stdstring.h>
#include <stdalloc.h>

void dither_matrix(size, mat_2D)
    int32    size;
    int32  **mat_2D; /* Dimensions vary */
{
    int32 *recur;
    int32 half;
    int32 i,j,r;
    int32 *mat = (int32 *)mat_2D;

    /* This code generates a NxN dither matrix where N is a power
     * of 2.  The algorithm is based on the Judice, Jarvice, Ninke
     * recurrence relation detailed in Foley, vanDam, et. al.
     * (II edition, p. 571).  A dither matrix can be used to generate
     * dither bitmaps.  For example, the 2x2 dither matrix:
     *
     * 0 2                                  00 10 10 11 11
     * 3 1  leads to the following bitmaps: 00 00 01 01 11
     *
     * the 00/00 bitmap is implied, the rest come from adding the
     * next bit indicated in the dither matrix.
     */

    /* Guarantee size is a power of two > 1.  ffs() is a little known
     * libc function which finds the first 1 bit in an integer.
     */
    if ((size < 2) || ((1 << (ffs((int)size)-1)) != size))
    {
	printf("Cannot generate %dx%d dither matrix\n", size,size);
	return;
    }

#define MAT(x,y) mat[(x)+((y)*size)]
#define RECUR(x,y) recur[(x)+((y)*half)]

    if (size == 2)
    {
	MAT(0,0) = 0;
	MAT(1,1) = 1;
	MAT(1,0) = 2;
	MAT(0,1) = 3;
    }
    else
    {
	half = size>>1;
	recur = (int32 *)STDALLOC(half*half*sizeof(int32));
	dither_matrix(half, (int32 **)recur);

	for (j=0;j<half;j++)
	  for (i=0;i<half;i++)
	  {
	      r = RECUR(i,j);

	      /* Upper Left */
	      MAT(i,j) = 4 * r;

	      /* Lower Right */
	      MAT(half+i,half+j) = 4 * r + 1;

	      /* Upper Right */
	      MAT(half+i,j) = 4 * r + 2;

	      /* Lower Left */
	      MAT(i,half+j) = 4 * r + 3;
	  }
	STDDEALLOC(recur);
    }
}
