/*
 * $RCSfile: libdither.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libdither.h                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: libdither header file                                     *
*       Created: Fri Nov  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBDITHER_H_INCLUDED
#define _LIBDITHER_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> /*common/include/global*/

/* DITHER_WIDTH(size) computes the number of bytes needed to store
 * size bytes (where size is a legal dither size)
 */
#define DITHER_WIDTH(size) (((size)<8)?1:((size)>>3))

/* DITHER_COUNT(size) computes the number of bitmaps which can be
 * generated from a size x size dither matrix.
 */
#define DITHER_COUNT(size) (((size)*(size))+1)

/* DITHER_MATRIX(name, size) declares a size x size dither matrix
 * with the given name.  name can then be passed as the matrix argument
 * to libdither functions.
 */
#define DITHER_MATRIX(name,size) int32 name[size][size]

/* DITHER_BITMAPS(name, size) declares an array of bitmaps, which
 * can be generated by dither_bitmaps().
 */
#define DITHER_BITMAPS(name,size) \
  char name[DITHER_COUNT(size)][size][DITHER_WIDTH(size)]

/* dither_matrix() creates a dither matrix which can then be passed
 * to dither generating functions.
 */
extern void dither_matrix(/* int32  size,
			     DITHER_MATRIX(mat, size)
			     */);

/* dither_bitmaps() creates DITHER_COUNT(size) bitmaps from the
 * dither generating matrix (created with dither_matrix()).
 * bits[i] 0<=i<DITHER_COUNT(size), is the i'th bitmap, and can
 * be passed to Xwindows functions such as XPutImage(), or
 * XCreatePixmapFromBitmapData().
 */
extern void dither_bitmaps(/* int32 size,
			      DITHER_MATRIX(mat, size),
			      DITHER_BITMAPS(bits, size)
			      */);

/* dither_or() OR's in an area_width x area_height section of the
 * which_dither'th dither generated from the matrix mat.  The upper
 * left corner of the dither is at x0,y0 of the
 * bitmap_width x bitmap_height bitmap.
 */
extern void dither_or(/* int32 size,
			 int32 which_dither,
			 DITHER_MATRIX(mat, size),
			 char *bitmap,
			 int32 bitmap_width,
			 int32 bitmap_height,
			 int32 x0, y0,
			 int32 area_width, area_height
			 */);

/* dither_copy() works just like dither_or(), except that zero bits
 * in the dither are set to zero in the bitmap.  Hence, dither_copy
 * runs only about half the speed of dither_or().
 */
extern void dither_copy(/* int32 size,
			   int32 which_dither,
			   DITHER_MATRIX(mat, size),
			   char *bitmap,
			   int32 bitmap_width,
			   int32 bitmap_height,
			   int32 x0, y0,
			   int32 area_width, area_height
			   */);


/* EXAMPLE:  see xtest.c for an example.
 */
#ifdef __cplusplus
}
#endif

#endif
