/*
 * $RCSfile: libenvcloud.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libenvcloud.h
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libenvcloud
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBENVCLOUD_H_INCLUDED
#define _LIBENVCLOUD_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <libctdb.h>


/* void ecloud_init(char *data_path, uint32 flags)
 * Initializes the libenvcloud model of water clouds, for use
 * in ModSAF ray visibility calculations.
 * Call this before calling any other libenvcloud functions. 
 *
 * The cloud reader file, containing model parameters and cloud
 * layer type descriptions, is read (using data_path and flags) and parsed.
 * A ecloud_clouds data structure is built and stored in a file static variable.
 * The appropriate tags are registered with libenvironment. 
 * The terrain database is needed to determine the finite 2-D simulation area
 * in which to model clouds, via fractals inside of each 2-D cloud grid cell */
extern void ecloud_init(
    char  *data_path,
    uint32 flags);

/* int32 ecloud_model_exists_p(void)
 * Returns true if the libenvcloud model has been properly initialized,
 * and is currently active. */
extern int32 ecloud_model_exists_p(void);

/* float32 ecloud_get_2d_value(float32 position[3])
 * Given a 2-D position, returns a number proportional to the
 * the total 2-D water density (which is the 3-D water density 
 * integrated vertically) at that position.
 * This is intended primarily displaying clouds on the PVD.
 * So a muliplicative (fudge) factor is also included, so that the
 * result is near one, when the clouds are heavy, but less than
 * one when they are thin enough that you can (statiscally) see
 * through them, when looking straight down. */
extern float32 ecloud_get_2d_value(float32 position[3]);

#ifdef USE_MOTIF
typedef struct ecloud_info
{
    uint32      time;
    float64     cloud_cover;
    float64     cloud_ceiling;
    float64     cloud_height;
    uint32      cloud_type;
} ECLOUD_INFO;

extern uint32 ecloud_refresh_display(ECLOUD_INFO *old_cloud,
					ECLOUD_INFO *new_cloud);
#endif

#endif /*_LIBENVCLOUD_H_INCLUDED*/
