#ifndef lint
static char rcsid [] = "$RCSfile: tst_3D.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tst_3D.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: test of cloud value functions, produces values on a grid
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libecloud_local.h"

/*************************************************************/

/* String describing how to use this program. */
char *usage = 
"Usage: tst_3D ecloud_rdr_file seed time cloudiness_fraction two_d_flag {{xstart,xstop,xstep},...} >outputfile\n";

/* Local versions of environment variables, for this simple test program. */
uint32  tst_time;
float32 tst_cloudiness;

/* Local versions of environment variable querying functions, used by 
 * the cloud library.  Used only for this simple test program. 
 * In a real simulation, the simulation would be supplying the environmental
 * variable values. */
uint32 tst_get_time(uint32 time_default);
void   tst_query_env(ECLOUD_QUERY_ENV_ARGS);

int main(int argc, char *argv[])
{
    int32    iarg;
    char    *ecloud_rdr_filename;
    int32    iseed;
    int32    two_d;
    float32  xstart[3], xstop[3], xstep[3];
    int32    d, size, dims[3], i, i2;
    ecloud_clouds *clouds;
    float32 *values;

    /* Parse the command line arguments. */
    if ((iarg=0, 0)
        ||(argc < 7)
	||(!(ecloud_rdr_filename = argv[++iarg]))
        ||(1!=sscanf(argv[++iarg],"%d",&iseed))
        ||(1!=sscanf(argv[++iarg],"%d", &tst_time))
        ||(1!=sscanf(argv[++iarg],"%f", &tst_cloudiness))
        ||(1!=sscanf(argv[++iarg],"%d", &two_d))
        ||(9!=sscanf(argv[++iarg],"{{%f,%f,%f},{%f,%f,%f},{%f,%f,%f}}",
		     &xstart[0], &xstop [0], &xstep [0],
		     &xstart[1], &xstop [1], &xstep [1],
		     &xstart[2], &xstop [2], &xstep [2])) )
    {
	if (iarg > 0) 
	  printf("ERROR: bad arg %d = \"%s\"\n", iarg, argv[iarg]);
	printf("%s", usage);
	return(0);
    }

    /* Initialize the cloud model by creating the main clouds
     * data structure. */
    if (!(clouds = 
	  ecloud_create(ecloud_rdr_filename, "", 0, (float32 *)0,
		       tst_get_time, tst_query_env)))
    {
	fprintf(stderr, "tst_3D: Creation of clouds failed.\n");
	return(1);
    }

    /* Redo the set of random numbers used in the horizontal fractal,
     * using the given random seed. */
    ecloud_redo_lattice(clouds, iseed);

    /* Update the cloud model time: this forces the passed in time 
     * function to be called, and the result is stored in the clouds struct.*/
    ecloud_time_update(clouds);

    /* Allocate an array to store the resulting cloud density values into. */
    for (d=0, size = 1; d<3; d++)  
      size *= (dims[d] = 1 + (xstop[d]-xstart[d])/xstep[d]);
    values = (float32*)STDCALLOC(1,size*sizeof(float32));

    /* Compute the cloud densities at the specified regular cartesian
     * grid points. */
    ecloud_densities(clouds, 1, xstart, xstep, dims, values, two_d);

    /* Print the resulting values to standard out. */
    for (i=0; i<size; i++)
    {
	printf("%f ", values[i]);
	for (d=0, i2=i; (d<3) && ((i2 % dims[d])==dims[d]-1) ; 
	     i2 /= dims[d],d++)
	  printf("\n");
    }
}

/* Function (passed as a pointer to ecloud_create), that services 
 * time queries made by the cloud library, during this test program. */
uint32 tst_get_time(uint32 time_default)
{
    return(tst_time);
}

/* Function (passed as a pointer to ecloud_create), that services 
 * simulation environmental variable queries made by the cloud library.
 * For this test program only. */
void   tst_query_env(ECLOUD_QUERY_ENV_ARGS)
{
    *cloudiness = tst_cloudiness;
    return;
}

/* Hack. Put here so that libreader files can easily be linked with this
 * test program. */
int yywrap(void)
{
   return(1);
}

/*************************************************************/
