#ifndef lint
static char rcsid [] = "$RCSfile: tst_image2d.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/* tst_image2d.c */
/* graphics interface, by Stephen A. Kukolich,
 * rewritten starting from John P. Beale's utilities in gforge */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include "tst_image2d.h"

#define DOS  yuck                     /* compiling on DOS system */
#define MAX_PIXVAL 65535              /* ascii integer output max value */
#define NUM_FMT 0       /* 0 IEEE Little Endian (80x86, DEC Risc) */
                        /* 1 IEEE Big Endian (SPARC, SGI, Motorola, etc.) */
#define PREC 1          /* storing Matlab file as floats rather than longs */
#define PTYPE float     /* internal representation of float numbers */

                        /* Matlab Level 1.0 format data file header */
typedef struct matlab_header_struct {
        long type;      /* type */
        long mrows;     /* row dimension */
        long ncols;     /* column dimension */
        long imagf;     /* flag indicating imag part */
        long namelen;    /* name length (including NULL) */
} Fmatrix;

#define TYPECMP(T1,T2) \
  ((T1[0]==T2[0])&&(T1[1]==T2[1])&&(T1[2]==T2[2]))
#define FPC(CH)  fputc((int)(CH),image->fpout)
#define FPS(SHORT1)  \
  FPC((SHORT1) % 255); /* Low byte */ \
  FPC((SHORT1) / 255)  /* High byte */

/* image2d *image2d_open(char* filename, char *type, int meshsize)
 * Open a 2-D image file for writing, returning the constructed
 * image2d object.  meshsize describes the number pixels high
 * and wide the image will be.  type should be one of the string
 * constants supplied in the tst_image2d.h header file. */
image2d *image2d_open(char* filename, char *type, int meshsize)
{
    image2d *image;
    char    *opentype;
    Fmatrix xh;                           /* matlab format header */  
    char mname[80] = "land";

    image = (image2d*)malloc(sizeof(struct image2d));
    memset(image, 0, sizeof(struct image2d));
    image->type = type;
    image->filename = filename;
    image->meshsize = meshsize;

    opentype = ((TYPECMP(image->type,PGM)) || (TYPECMP(image->type,OCT)) 
		|| (TYPECMP(image->type,PLT))) 
               ? ("w" /*text*/) : ("wb" /*binary*/ );
    if ((image->fpout = fopen(filename,opentype)) == NULL )
    {
        fprintf(stderr,"IMAGE2D ERROR: Couldn't open output file \"%s\"\n",
		filename);
	return((image2d*)0);
    }

    if (TYPECMP(image->type,TGA))
    {
	FPC(0); /*IDLength*/
	FPC(0); /*CoMapType*/
	FPC(2); /*ImgType*/
	FPS(0); /*Index*/
	FPS(0); /*Length*/
	FPC(0); /*CoSize*/
	FPS(0); /*X_org*/
	FPS(0); /*Y_org*/
	FPS(image->meshsize); /*Width*/
	FPS(image->meshsize); /*Height*/
	FPC(24); /*PixelSize*/
	FPC((unsigned char) 0x20);        /* descriptor byte b17 */
    }
    else if (TYPECMP(image->type,PGM))
    {
        fprintf(image->fpout,"P2\n");       /* PGM type 2: 16-bit ASCII data */
	fprintf(image->fpout,"# created using image2d, meshsize=%d\n",
		image->meshsize);
	fprintf(image->fpout,"%d %d\n",image->meshsize,image->meshsize);
	fprintf(image->fpout,"%u\n",MAX_PIXVAL);
    }
    else if (TYPECMP(image->type,PG8))
    {
        fprintf(image->fpout,"P5\n");       /* PGM type 2: 8-bit binary data */
	fprintf(image->fpout,"# created using image2d, meshsize=%d\n",
		image->meshsize);
	fprintf(image->fpout,"%d %d\n",image->meshsize,image->meshsize);
	fprintf(image->fpout,"255\n");
    }
    else if ((TYPECMP(image->type,OCT))||(TYPECMP(image->type,PLT))) 
    {     
        /* Octave ascii format, or Gnuplot ascii format */
        fprintf(image->fpout,"# name: land\n");
	fprintf(image->fpout,"# type: matrix\n");
	fprintf(image->fpout,"# rows: %d\n",image->meshsize);
	fprintf(image->fpout,"# columns: %d\n",image->meshsize);
    }
    else if (TYPECMP(image->type,MAT)) 
    {     /* Matlab binary format */
        xh.type = NUM_FMT*1000 + PREC*10;
	xh.mrows = image->meshsize;
	xh.ncols = image->meshsize;
	xh.imagf = 0;
	xh.namelen = strlen(mname) + 1;
	fwrite(&xh, sizeof(Fmatrix),(size_t)1, image->fpout);
	fwrite(mname,sizeof(char),(size_t)xh.namelen,image->fpout);
    }
    else
    {
	fprintf(stderr,"IMAGE2D ERROR: bad type \"%s\"\n", image->type);
	return((image2d*)0);
    }
    return(image);
}

/* void image2d_writepixel(image2d *image, float pf)
 * write a greyscale pixel. */
void image2d_writepixel(image2d *image, float pf)
{
    unsigned char r,g,b;
    unsigned int p;

    r = 0; g =0; b = 0;
    p = MAX_PIXVAL * pf;

    if (TYPECMP(image->type,PGM))
      fprintf(image->fpout,"%u ",p);
    else if(TYPECMP(image->type,PG8))
      fprintf(image->fpout,"%c",(unsigned char)(p>>8));
    else if(TYPECMP(image->type,TGA))  /* TGA pixel order is BGR. R=hi, G=lo */
    {
        r = p / 256;
        g = p % 256;
        fprintf(image->fpout,"%c%c%c",b,g,r);
    }
    else if (TYPECMP(image->type,OCT)) 
      fprintf(image->fpout,"%1.5f ",pf);
    else if (TYPECMP(image->type,PLT))
      fprintf(image->fpout,"%1.5f\n",pf);
    else if (TYPECMP(image->type,MAT))
      fwrite(&pf,sizeof(PTYPE),1,image->fpout);
    else if(TYPECMP(image->type,PNG))
      ;
}

/* void image2d_writenewline(image2d *image)
 * Write a newline (typically at the end of a row).
 * Does nothing unless the output image is an ascii representation of 
 * the numbers. */
void image2d_writenewline(image2d *image)
{  
   /* newline if ascii data */
   if ((TYPECMP(image->type,OCT)) || (TYPECMP(image->type,PLT)) 
       || (TYPECMP(image->type,PGM))) 
     fprintf(image->fpout,"\n");
}

/* int image2d_close(image2d *image)
 * Close the output file, and free the image object. */
int image2d_close(image2d *image)
{
  int  retval;

  if(TYPECMP(image->type,PNG))
  {
      fprintf(image->fpout,"END-PNG\n");
      return(0);
  }
  retval = fclose(image->fpout);
  free(image);
  return(retval);
}


