/*
 * $RCSfile: tst_image2d.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */

/* tst_image2d.h */
/* graphics interface, by Stephen A. Kukolich,
 * rewritten from John P. Beale's utilities in gforge */

typedef struct image2d 
{
    char *type;
    char *filename;
    int   meshsize;
    FILE *fpout;                          /* pointer to image output file */
} image2d;

/* image types */
#define PG8  "pg8"          /* 8-bit binary PGM format */
#define PLT  "plt"          /* Gnuplot data format */
#define TGA  "tga"          /* POV-Ray 2.2 TGA heightfield format */
#define PGM  "pgm"          /* PGM ascii type */
#define OCT  "oct"          /* Octave ascii format */
#define MAT  "mat"          /* Matlab Level 1.0 binary format */
#define PNG  "png"          /* PNG format (not yet supported) */

/* functions */

/* image2d *image2d_open(char* filename, char *type, int meshsize)
 * Open a 2-D image file for writing, returning the constructed
 * image2d object.  meshsize describes the number pixels high
 * and wide the image will be.  type should be one of the string
 * constants supplied in the tst_image2d.h header file. */
image2d *image2d_open(char* filename, char *type, int meshsize);

/* void image2d_writepixel(image2d *image, float pf)
 * write a greyscale pixel. */
void image2d_writepixel(image2d *image, float pf);

/* void image2d_writenewline(image2d *image)
 * Write a newline (typically at the end of a row).
 * Does nothing unless the output image is an ascii representation of 
 * the numbers. */
void image2d_writenewline(image2d *image);

/* int image2d_close(image2d *image)
 * Close the output file, and free the image object. */
int image2d_close(image2d *image);

