#ifndef lint
static char rcsid [] = "$RCSfile: env_get.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_get.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Handles queries to libenvironment                         *
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: rg
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <stdext.h>
#include <stdalloc.h>
#include <stdio.h>
#include "libenv_local.h"
#include "libenvironment.h"

int32 env_last_get_error = ENV_ERROR_UNKNOWN;

/*
 * env_get_explicit
 *
 * Call the appropriate function for the input tag, and pass
 * the i/o structures through.
 */

int32 env_get_explicit(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    ENV_TAG_DEF_T	*td;
    int32		*m;
    int32		n;
    int32		resolver;
    int32		have_resolver;

    env_last_get_error = ENV_ERROR_UNKNOWN;

    /* 1) Called with a single model.  Use this model.
     * 2) Called with a list.  Look up resolver, pass list.
     *    If no resolver, call 1st model in list.
     * 3) Called with a null list.  Look up list, look up resolver, pass list.
     *    If no resolver, call 1st model in list.
     */

    if (models == NULL)
    {
	if (env_tags[tag].cached_get_models)
	{
	    n = env_tags[tag].cached_num_get_models;
	    m = env_tags[tag].cached_get_models;
	}
	else
	{
	    n = env_num_models_for_tag(tag, 
				       ENV_AMBIENT_MODEL|ENV_LOCAL_MODEL,
				       ENV_GET);
	    m = env_alloc_model_array(tag, n, 
				      ENV_AMBIENT_MODEL|ENV_LOCAL_MODEL, 
				      ENV_GET);
	    env_find_models(tag, m, n, 
			    ENV_AMBIENT_MODEL|ENV_LOCAL_MODEL, 
			    ENV_GET);

	    env_tags[tag].cached_num_get_models = n;
	    env_tags[tag].cached_get_models = m;
	}
    }
    else
    {
	n = num_models;
	m = models;
    }

    if ((n < 1) || (m == NULL))
    {
	env_last_get_error = ENV_ERROR_NO_MODEL;
	return(-1);
    }


    if (n == 1)
      have_resolver = FALSE;
    else
      have_resolver = 
	(env_find_models(tag, &resolver, 1, ENV_RESOLVER_MODEL, ENV_GET) == 1);


    if (have_resolver)
    {
	/* Refuse to call a resolver model when it is our only model. */

	if (n == 1)
	{
	    env_last_get_error = ENV_ERROR_NO_MODEL;
	    return(-1);
	}

	td = env_find_handler_for_model(resolver, tag);
	if (td && td->getfcn)
	  return ((*(td->getfcn))(tag, interface, m, n));
	else
	{
	    env_last_get_error = ENV_ERROR_NO_MODEL;
	    return(-1);
	}
    }
    else
    {
	td = env_find_handler_for_model(*m, tag);
	if (td && td->getfcn)
	  return ((*(td->getfcn))(tag, interface));
	else
	{
	    env_last_get_error = ENV_ERROR_NO_MODEL;
	    return(-1);
	}
    }
}

int32 env_get(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface)
{
    int32 return_val;

    return_val = env_get_explicit(tag, interface, NULL, 0);

    if (env_keep_statistics)
    {
	env_statistics.total_gets++;
	env_statistics.gets_by_tag[tag]++;
	if (return_val == -1)
	{
	    env_statistics.total_errors++;
	    env_statistics.errors_by_code[env_last_get_error]++;
	}
	else
	{
	    switch(tag)
	    {
	      case ENV_RAY_VISIBILITY:
		if (interface->u.ray_visibility.output_value < 1.0)
		  env_statistics.intervis_obscured++;
		break;
	      default:
		break;
	    }
	}
    }

    return return_val;
}

/*
 * Returns the ambient value of the queried parameter. Does not
 * factor in any effects from local models.
 */
int32 env_get_ambient(
    int32		tag,
    ENV_INTERFACE	*interface)
{
    int32		*models;
    int32		num_models;

    num_models = env_num_models_for_tag(tag, ENV_AMBIENT_MODEL, ENV_GET);
    models = env_alloc_model_array(tag, num_models, ENV_AMBIENT_MODEL,
				   ENV_GET);
    env_find_models(tag, models, num_models, ENV_AMBIENT_MODEL, ENV_GET);

    return(env_get_explicit(tag, interface, models, num_models));
}

