#ifndef lint
static char rcsid [] = "$RCSfile: env_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_init.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libenvironment
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: rg
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <stdext.h>
#include <string.h>
#include <stdalloc.h>
#include "libenv_local.h"
#include <libsched.h>
#include <libreader.h>
#include <libctdb.h>
#include <libgcs.h>

READER_UNION env_param_table;
CTDB *env_ctdb;

#define ENV_SETUP(TAG, TAG_IN_QUOTES)  \
  strncpy(env_tags[(TAG)].name, TAG_IN_QUOTES, ENV_MAX_NAME)

#define ENV_FX_SETUP(EFFECT, EFFECT_IN_QUOTES) \
  strncpy(env_effects[(EFFECT)].effectname, EFFECT_IN_QUOTES, ENV_MAX_NAME)

void env_init(
    char	*directory,
    uint32	flags)
{

    env_ctdb = (CTDB *)gcs_get_tdb(GCS_ILLEGAL_CELL, FALSE);

    bzero(&env_param_table, sizeof(READER_UNION));

    if (!directory)
    {
	printf("Environment: Skipping environment.rdr.\n");
	printf("             (No initialized models will be ignored.)\n");
    }

    else if (reader_read("environment.rdr", directory, &env_param_table,
		    flags & ~READER_TYPING))
    {
	printf("Environment: Unable to read environment.rdr!");
	printf("             (No initialized models will be ignored.)\n");
    }
    
/*    sched_periodic_fncl(env_tick, 67, 67, 0, A_END); */

    env_tags = (struct env_tag_ctl *)
      STDALLOC(ENV_NUM_TAGS * sizeof(struct env_tag_ctl));

    bzero(env_tags, ENV_NUM_TAGS * sizeof(struct env_tag_ctl));

    ENV_SETUP(ENV_INTERNAL_STATS,"ENV_INTERNAL_STATS");
    ENV_SETUP(ENV_TEMPERATURE,"ENV_TEMPERATURE");
    ENV_SETUP(ENV_DEWPOINT,"ENV_DEWPOINT");
    ENV_SETUP(ENV_RELATIVE_HUMIDITY,"ENV_RELATIVE_HUMIDITY");
    ENV_SETUP(ENV_BAROMETRIC_PRESSURE,"ENV_BAROMETRIC_PRESSURE");
    ENV_SETUP(ENV_WIND_VELOCITY,"ENV_WIND_VELOCITY"); 
    ENV_SETUP(ENV_PRECIPITATION_TYPE,"ENV_PRECIPITATION_TYPE");
    ENV_SETUP(ENV_PRECIPITATION_RATE,"ENV_PRECIPITATION_RATE");
    ENV_SETUP(ENV_EXTINCTION_TYPE,"ENV_EXTINCTION_TYPE");
    ENV_SETUP(ENV_EXTINCTION_AMOUNT,"ENV_EXTINCTION_AMOUNT");
    ENV_SETUP(ENV_EXTINCTION_COEFF,"ENV_EXTINCTION_COEFF");
    ENV_SETUP(ENV_RAY_VISIBILITY,"ENV_RAY_VISIBILITY");
    ENV_SETUP(ENV_VISUAL_RANGE,"ENV_VISUAL_RANGE");
    ENV_SETUP(ENV_ILLUMINATION,"ENV_ILLUMINATION");
    ENV_SETUP(ENV_SOLAR_ILLUMINATION, "ENV_SOLAR_ILLUMINATION");
    ENV_SETUP(ENV_LUNAR_ILLUMINATION, "ENV_LUNAR_ILLUMINATION");
    ENV_SETUP(ENV_SKY_ILLUMINATION, "ENV_SKY_ILLUMINATION");
    ENV_SETUP(ENV_SKY_OVER_GROUND,"ENV_SKY_OVER_GROUND"); 
    ENV_SETUP(ENV_CLOUD_COVER,"ENV_CLOUD_COVER"); 
    ENV_SETUP(ENV_CLOUD_CEILING,"ENV_CLOUD_CEILING"); 
    ENV_SETUP(ENV_CLOUD_HEIGHT, "ENV_CLOUD_HEIGHT");
    ENV_SETUP(ENV_CLOUD_TYPE, "ENV_CLOUD_TYPE");
    ENV_SETUP(ENV_SIM_TIME,"ENV_SIM_TIME"); 
    ENV_SETUP(ENV_SUN_POSITION,"ENV_SUN_POSITION");
    ENV_SETUP(ENV_MOON_POSITION,"ENV_MOON_POSITION");
    ENV_SETUP(ENV_MOON_ELONGATION,"ENV_MOON_ELONGATION");
    ENV_SETUP(ENV_CONTRAST,"ENV_CONTRAST");
    ENV_SETUP(ENV_SEA_STATE, "ENV_SEA_STATE");
    ENV_SETUP(ENV_SEA_SALINITY, "ENV_SEA_SALINITY");
    ENV_SETUP(ENV_SEA_NUM_WAVES, "ENV_SEA_NUM_WAVES");
    ENV_SETUP(ENV_SEA_NUM_FREQUENCIES, "ENV_SEA_NUM_FREQUENCIES");
    ENV_SETUP(ENV_SEA_TIDE_LEVEL, "ENV_SEA_TIDE_LEVEL");
    ENV_SETUP(ENV_SEA_SURFACE_HEIGHT, "ENV_SEA_SURFACE_HEIGHT");
    ENV_SETUP(ENV_SEA_SURFACE_TEMP, "ENV_SEA_SURFACE_TEMP");
    ENV_SETUP(ENV_SEA_WAVE_DIRECTION, "ENV_SEA_WAVE_DIRECTION");
    ENV_SETUP(ENV_SEA_WAVE_PERIOD, "ENV_SEA_WAVE_PERIOD");
    ENV_SETUP(ENV_SEA_WAVE_HEIGHT, "ENV_SEA_WAVE_HEIGHT");
    ENV_SETUP(ENV_SEA_CURRENT_VELOCITY, "ENV_SEA_CURRENT_VELOCITY");
    ENV_SETUP(ENV_SEA_BREAKER_ANGLE, "ENV_SEA_BREAKER_ANGLE");
    ENV_SETUP(ENV_SEA_BREAKER_TYPE, "ENV_SEA_BREAKER_TYPE");
    ENV_SETUP(ENV_SEA_IN_SURF_ZONE, "ENV_SEA_IN_SURF_ZONE");
    ENV_SETUP(ENV_TOTAL_PRECIP_IN_HR, "ENV_TOTAL_PRECIP_IN_HR");
    ENV_SETUP(ENV_SUB_SOIL_SATURATION, "ENV_SUB_SOIL_SATURATION");
    ENV_SETUP(ENV_LAST_RAIN_TIME, "ENV_LAST_RAIN_TIME");
    
    ENV_SETUP(ENV_BATCH,"ENV_BATCH");
    ENV_SETUP(ENV_FLUSH,"ENV_FLUSH");

    bzero(env_effects, ENV_NUM_EFFECTS * sizeof(struct env_effect_def));

    ENV_FX_SETUP(ENV_SMOKE_EFFECT,"ENV_SMOKE_EFFECT");
    ENV_FX_SETUP(ENV_FLARE_EFFECT,"ENV_FLARE_EFFECT");
    ENV_FX_SETUP(ENV_DUST_EFFECT,"ENV_DUST_EFFECT");
}
