#ifndef lint
static char rcsid [] = "$RCSfile: env_resolvers.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_resolvers.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Misc resolver models for general purpose use.
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: rg
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* 
 * This multiply resolver can handle any interface that is properly padded
 * with value/max/min 64 bytes into the structure.
 */

#include <stdio.h>
#include <stdext.h>
#include <stdio.h>
#include "libenv_local.h"
typedef enum env_ops 
{
    ENV_MULTIPLY, ENV_ADD, ENV_VEC_ADD
} ENV_OPS;

/*
 * Forward declarations:
 */
static int32 env_vmm_assoc(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models,
    ENV_OPS		op);

static int32 in_surf_zone(
    float64 loc[3]);			 

int32 env_vmm_multiply(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    return(env_vmm_assoc(tag, interface, models, num_models, ENV_MULTIPLY));
}

int32 env_vmm_add(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    return(env_vmm_assoc(tag, interface, models, num_models, ENV_ADD));
}
    
int32 env_vmm_vadd(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    return(env_vmm_assoc(tag, interface, models, num_models, ENV_VEC_ADD));
}
    
static int32 env_vmm_assoc(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models,
    ENV_OPS		op)
{
    ENV_INTERFACE	local_iface;
    int32 		i, n;
    float64		value, max, min;
    int32		found = FALSE;
    float64             vvalue[3];

    switch(op)
    {
      case ENV_MULTIPLY:	value = 1.0;	break;
      case ENV_VEC_ADD:         vvalue[0] = vvalue[1] = vvalue[2] = 0.0;
      case ENV_ADD:
      default:			value = 0.0;	break;
    }

    max = 0.0;
    min = 0.0;
    n = 0;


	
    for (i = 0; i < num_models; i++)
    {
	bcopy(interface, &local_iface, sizeof(ENV_INTERFACE));
	local_iface.u.generic_value.output_value = 0.0;

	if (models[i] != -1)
	{
	    if (env_get_explicit(tag, &local_iface, 
				 &models[i], 1) != -1)
	    {
		if (op == ENV_MULTIPLY)
		  value *= 
		    local_iface.u.generic_value.output_value;
		else if (op == ENV_ADD)
		  value +=
		    local_iface.u.generic_value.output_value;
		else if (op == ENV_VEC_ADD)
		{
		    vvalue[0] +=
		      local_iface.u.generic_vector_to_vector.output_vector[0];
		    vvalue[1] +=
		      local_iface.u.generic_vector_to_vector.output_vector[1];
		    vvalue[2] +=
		      local_iface.u.generic_vector_to_vector.output_vector[2];
		}
		
		
		/* $$$ What was this supposed to do? */
		if (local_iface.u.generic_value.output_max > max)
		  max = local_iface.u.generic_value.output_max;
		if (local_iface.u.generic_value.output_min > min)
		  max = local_iface.u.generic_value.output_min;
		found = TRUE;
		n++;
	    }
	}
    }
    if (!found)
      return(-1);
    else
    {
	if (op == ENV_VEC_ADD)
	{
	    interface->u.generic_vector_to_vector.output_vector[0] = vvalue[0];
	    interface->u.generic_vector_to_vector.output_vector[1] = vvalue[1];
	    interface->u.generic_vector_to_vector.output_vector[2] = vvalue[2];
	}
	else

	  interface->u.generic_value.output_value = value;

	return(n);
    }
}

/* This resolver calls only the first local registered, or ambient
 * if no other choice.  If multiple ambient models, call the first.
 */
int32 env_use_priority(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    ENV_TAG_DEF_T  *td;
    int32           index = 0;	/* Call the 0th model by default */
    int32	    i;

    if (num_models == 0)
      return(-1);
    if (num_models > 1)
    {
	for (i = 0; i < num_models; i++)
	{
	    td = env_find_handler_for_model(models[i], tag);
	    if (td->type == ENV_LOCAL_MODEL)
	    {
		index = i;
		break;
	    }
	}
    }
    return(env_get_explicit(tag, interface, &models[index], 1));
}


/*
 *  This resolver selects a value based on whether the location is in
 *  a surf zone.
 */
    
int32 env_surf_zone(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{

    int32 in_surf;
    static int32 complained = 0;
    ENV_TAG_DEF_T  *td;
    int32           chosen;
    int32	    idx;

    if (num_models <= 0)
      return(-1);

    if (num_models == 1)
      return(env_get_explicit(tag, interface, &models[0], 1));

    /* 
     * Input location field varies with tag.  Although the structure is
     * organized so that the location occurs in the same position in
     * each member, we'll take the time here to correctly identify the
     * field.
     */
    switch(tag)
    {
	/*vector-to-value */
      case ENV_SEA_SURFACE_HEIGHT:
      case ENV_SEA_SURFACE_TEMP:
      case ENV_SEA_WAVE_DIRECTION:
      case ENV_SEA_WAVE_PERIOD:
      case ENV_SEA_WAVE_HEIGHT:
      case ENV_SEA_BREAKER_ANGLE:
	in_surf =
	  in_surf_zone(interface->u.generic_vector_to_value.input_vector);
	break;

	/* vector-to-vector */
      case ENV_SEA_CURRENT_VELOCITY:
      case ENV_SEA_BREAKER_TYPE:
	in_surf =
	  in_surf_zone(interface->u.generic_vector_to_vector.input_vector);
	break;

	/* vector-to-type */
      case ENV_SEA_IN_SURF_ZONE:
	in_surf =
	  in_surf_zone(interface->u.generic_vector_to_type.input_vector);
	break;

      default:
	if (!complained)
	{
	    fprintf(stderr,
		    "env_in_surf_zone: called for unsupported tag: %d\n",
		    tag);
	    complained = 1;
	}
	return -1;

    }

    /*
     * Now that we've determined whether or not we're in surf zone,
     * locate the ambient (open ocean) or local (surf zone) model and
     * return its value.
     *
     * $$$ We need to update this scheme to support lakes and rivers.
     */

    chosen = 0;	/* Call the 0th model by default */

    for (idx = 0; idx < num_models; idx++)
    {
	td = env_find_handler_for_model(models[idx], tag);
	if ((in_surf && (td->type == ENV_LOCAL_MODEL)) ||
	  (!in_surf && (td->type == ENV_AMBIENT_MODEL)))
	{
	    chosen = idx;
	    break;
	}
    }

    return(env_get_explicit(tag, interface, &models[chosen], 1));

}


static int32 in_surf_zone (loc)
float64 loc[3];
{
    CTDB_ELEV_DATA           elev[CTDB_MAX_ELEVS];
    int32 water_found, ground_found, eidx;
    float64 ground, water, depth;
    float64 surf_threshold = 3.0; /* $$$ Make this external/configurable. */
    
    ctdb_lookup_elevation_mes(env_ctdb, loc[0], loc[1],
			      CTDB_USE_DEFAULT, elev);

    water_found = ground_found = 0;

    for (eidx = 0; eidx < CTDB_MAX_ELEVS; ++eidx)
    {

	if (elev[eidx].sensClass == CTDB_FC_ILLEGAL) /* denotes end */
	  break;
	
	if (elev[eidx].sensClass == CTDB_FC_MICRO)
	{
	    if (elev[eidx].subclass == CTDB_FS_SKIN)
	    {
		ground = elev[eidx].z;
		ground_found = 1;
	    }
	    else if (elev[eidx].subclass == CTDB_FS_WATER)
	    {
		water = elev[eidx].z;
		water_found = 1;
	    }
	}

	if (water_found && ground_found)
	  break;

    }

    if (water_found && ground_found)
    {
	depth = water - ground; /* water == 0. ground < 0 */
	
#if 0
	return (depth <= eui_get_surf_depth(EUI_BASE_PREFS_PTR));
#else
	return (depth <= surf_threshold);
#endif
    }
    
    return 0;

} 

