#ifndef lint
static char rcsid [] = "$RCSfile: env_set.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_set.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Handles queries to libenvironment                         *
*       Created: Fri Aug 15 1994
*       Author: gonzalez
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libenv_local.h"
#include <stdext.h>
#include <stdalloc.h>


/*
 * env_set_explicit
 *
 * Call the appropriate function for the input tag, and pass
 * the i/o structures through.
 */
int32 env_set_explicit(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    ENV_TAG_DEF_T	*td;
    int32		*m;
    int32		n;
    int32		resolver;
    int32		have_resolver;
    int32		i;
    int32		found = FALSE;
    int32		ret = -1;

    /* 1) Called with a single model.  Use this model.
     * 2) Called with a list.  Call each in list except resolvers.
     * 3) Called with a null list.  Look up list, call all but resolvers.
     */

    if (models == NULL)
    {
	if (env_tags[tag].cached_set_models)
	{
	    n = env_tags[tag].cached_num_set_models;
	    m = env_tags[tag].cached_set_models;
	}
	else
	{
	    n = env_num_models_for_tag(tag, 
				       ENV_AMBIENT_MODEL
				       |ENV_LOCAL_MODEL
				       |ENV_CONTROL_MODEL,
				       ENV_SET);
	    m = env_alloc_model_array(tag, n, 
				      ENV_AMBIENT_MODEL
				      |ENV_LOCAL_MODEL
				      |ENV_CONTROL_MODEL,
				      ENV_SET);
	    env_find_models(tag, m, n, 
			    ENV_AMBIENT_MODEL
			    |ENV_LOCAL_MODEL
			    |ENV_CONTROL_MODEL,
			    ENV_SET);
	    env_tags[tag].cached_num_set_models = n;
	    env_tags[tag].cached_set_models = m;
	}
    }
    else
    {
	n = num_models;
	m = models;
    }

    if ((n < 1) || (m == NULL))
      return(-1);

    for (i = 0; i < n; i++)
    {
	td = env_find_handler_for_model(m[i], tag);
	if (td && (td->type != ENV_RESOLVER_MODEL) && (td->setfcn))
	{
	    ret = (*(td->setfcn))(tag, interface);
	    if ((ret == 0) && (found == FALSE))
	      found = TRUE;
	}
    }
    if (found)	/* If success in setting at least -one- thing, we're happy */
      return(0);
    else
      return(ret);
}


int32 env_set(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface)
{
    return(env_set_explicit(tag, interface, NULL, 0));
}
    
