#ifndef lint
static char rcsid [] = "$RCSfile: env_utils.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_utils.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Various utility functions                                 *
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: rg
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <stdext.h>
#include <stdalloc.h>
#include <stdstring.h>
#include <stdio.h>
#include <libreader.h>
#include "libenvironment.h"
#include "libenv_local.h"

extern READER_UNION env_param_table;

/*
 * Static function declarations: 
 */
static void env_print_tag_defs(ENV_TAG_DEF_T *td);
static void env_print_model_label(
    char  *label,
    int32 *first_used_ptr);

char *env_model_name(int32 md)
{
    if ((md < 0) || (md >= env_find_num_models()))
      return(NULL);
    
    return(env_models[md].name);
}

char *env_tag_name(ENV_TAG tag)
{
    return(env_tags[tag].name);
}

int32 env_ambient_set_models_exist(ENV_TAG tag)
{
    if (env_num_models_for_tag(tag, 
			       ENV_AMBIENT_MODEL, 
			       ENV_SET) > 0)
      return(TRUE);
    else
      return(FALSE);
}

int32 env_ambient_get_models_exist(ENV_TAG tag)
{
    if (env_num_models_for_tag(tag, 
			       ENV_AMBIENT_MODEL, 
			       ENV_GET) > 0)
      return(TRUE);
    else
      return(FALSE);
}

int32 env_num_models_for_tag(
    ENV_TAG	tag,
    int32	types,
    ENV_GETSET	getset)
{
    return(env_find_models(tag, NULL, 0, types, getset));
}

int32 env_find_models(
    ENV_TAG	tag,
    int32	*buf,
    int32	bsize,
    int32	types,
    ENV_GETSET	getset)
{
    int		num_models = 0;
    ENV_TAG_DEF_T	*td;

    if (buf && (bsize < 0))
      return(-1);

    if (buf)
      bzero(buf, bsize * sizeof(int32));
    
    if ((tag < 0) || (tag >= ENV_NUM_TAGS))
      return(-1);

    /* There can only be one resolver, and "active" status is ignored. */
    /* I can't imagine why this would ever change. */

    td = env_tags[tag].handler[ENV_RESOLVER_INDEX];

    if ((getset == ENV_GET) &&
	(types & ENV_RESOLVER_MODEL) &&
	env_tags[tag].handler[ENV_RESOLVER_INDEX])
    {
	if (buf && (bsize > num_models) && (td->getfcn))
	  buf[num_models] = env_tags[tag].handler[ENV_RESOLVER_INDEX]->md;
	num_models++;
    }

    /* There can only be one active ambient model, but there might */
    /* be some passive ones that we need to skip over. */

    if (types & ENV_AMBIENT_MODEL)
    {
	td = env_tags[tag].handler[ENV_AMBIENT_INDEX];

	while (td)
	{
	    if ((td->active || (getset == ENV_SET))
		&& (((getset == ENV_SET) && (td->setfcn))
		    || ((getset == ENV_GET) && (td->getfcn))))
	    {
		if (buf && (bsize > num_models))
		  buf[num_models] = td->md;
		num_models++;

		/* Skip out as soon as we find an active model */
		if (td->active)
		  break;
	    }
	    td = td->t_next;
	}
    }

    if (types & ENV_LOCAL_MODEL)
    {
	td = env_tags[tag].handler[ENV_LOCAL_INDEX];
	while (td)
	{
	    if ((td->active || (getset == ENV_SET))
		&& (((getset == ENV_SET) && (td->setfcn))
		    || ((getset == ENV_GET) && (td->getfcn))))
	    {
		if (buf && (bsize > num_models))
		  buf[num_models] = td->md;
		num_models++;
	    }
	    td = td->t_next;
	}
    }

    if (types & ENV_CONTROL_MODEL)
    {
	td = env_tags[tag].handler[ENV_CONTROL_INDEX];
	while(td)
	{
	    if (((getset == ENV_SET) && (td->setfcn))
		|| ((getset == ENV_GET) && (td->getfcn)))
	    {
		if (buf && (bsize > num_models))
		  buf[num_models] = td->md;
		num_models++;
	    }
	    td = td->t_next;
	}
    }
    
    return(num_models);
}

void env_active(
    int32	md,
    ENV_TAG	tag,
    int32	active)
{
    ENV_TAG_DEF_T *td;

    if ((md < 0) || (md >= env_find_num_models()))
      return;

    td = env_find_handler_for_model(md, tag);

    if (!td)
      return;

    if (active != td->active)
    {
	td->active = active;

	/* Clear any cached "get" information.  
	 * "set" isn't affected by the active flag, so we don't
	 * need to clear its cache
	 */
	if (env_tags[tag].cached_get_models)
	{
	    STDDEALLOC(env_tags[tag].cached_get_models);
	    env_tags[tag].cached_get_models = NULL;
	    env_tags[tag].cached_num_get_models = 0;    
	}
    }
}
    
int32 *env_alloc_model_array(
    ENV_TAG	tag,
    int32	n,
    int32	types,
    ENV_GETSET	getset)
{
    int32	*array;

    if (n <= 0)
      n = env_num_models_for_tag(tag, types, getset);

    if (n <= 0)
      return(NULL);

    array = (int32 *) STDALLOC(n * sizeof(int32));
    return(array);
}

void env_dealloc_model_array(
    int32 *array)
{
    STDDEALLOC((ADDRESS)array);
}

ENV_TAG_DEF_T *env_find_handler(ENV_TAG tag)
{
    int32	type = 0;

    if ((tag < 0) || (tag >= ENV_NUM_TAGS))
      return(NULL);
    
    while (ENV_MODEL_VALID(type))
    {
	if (env_tags[tag].handler[ENV_MODEL_TO_INDEX(type)])
	  return(env_tags[tag].handler[ENV_MODEL_TO_INDEX(type)]);
	type++;
    }
    return(NULL);
}

ENV_TAG_DEF_T *env_find_handler_for_model(
    int32	model,
    ENV_TAG	tag)
{
    ENV_TAG_DEF_T	*td;

    if (model < 0)
      return(NULL);
    
    td = env_models[model].handler;
    
    while (td)
    {
	if (td->tag == tag)
	  return(td);
	else
	  td = td->m_next;
    }
    return(NULL);
}

int32 env_find_model(char *mname)
{
    int32		md;
    int32		n;
    
    n = env_find_num_models();
    
    for (md = 0; md < n; md++)
    {
	if (!strncmp(mname, env_models[md].name, ENV_MAX_NAME))
	  return(md);
    }
    return(-1);
}

/*
 * Prints a description of each of the tag defs
 * in the passed tag def linked list. 
 */
static void env_print_tag_defs(ENV_TAG_DEF_T *td)
{
    
    while (td)
    {
	/* Separate tag defs with spaces. 
	 * Enclose inactive models in "<>".
	 */
	printf("%s", td->active ? " " : " <");
	    
	printf("%s:", env_models[td->md].name);
	if (td->getfcn)
	  printf("?");
	if (td->setfcn)
	  printf("!");

	printf("%s", td->active ? " " : "> ");

	td = td->t_next;
    }
	
}

/*
 * Prints the passed label as the beginning of a 
 * class of models. Uses the correct format given 
 * a pointer to an indication of whether this is the
 * first model class being labeled or not. Updates the
 * passed pointer.
 */
static void env_print_model_label(
    char  *label,
    int32 *first_used_ptr)
{
    /* If the first line has been used already,
     * we need to format over to the same column:
     */
    if (*first_used_ptr)
      printf("\n%-22s", ""); /* see format in env_print_registered_models */

    printf("%s)  ", label);
    *first_used_ptr = 1;
}

/*
 * Prints a description of all the currently registered models.
 * Prints a label for each tag regardless of whether any models
 * are registered for it or not. After each tag, models are
 * listed by class (ambient, local, control, resolver), with 
 * each class listed on a separate line. Classes with no models
 * in them are skipped.
 *
 * For each model, get support is indicated by '?', set support by '!'.
 * If the model is inactive, it is encompassed by '<' and '>'.
 */
int32 env_print_registered_models(void)
{
    ENV_TAG	tag;
    ENV_TAG_DEF_T	*td;
    int32 first_line_used;

    for (tag = 0; tag < ENV_NUM_TAGS; tag++)
    {
	/* Label the tag whose handlers we are about to
	 * print. We chop of the "ENV_" prefix by skipping
	 * the first four characters:
	 */
	printf("%-22s", env_tags[tag].name+4);
	first_line_used = FALSE;

	/* Resolver models: */
	if (td = env_tags[tag].handler[ENV_RESOLVER_INDEX])
	  env_print_model_label("R", &first_line_used);
	env_print_tag_defs(td);

	/* Ambient models: */
	if (td = env_tags[tag].handler[ENV_AMBIENT_INDEX])
	  env_print_model_label("A", &first_line_used);
	env_print_tag_defs(td);

	/* Local models: */
	if (td = env_tags[tag].handler[ENV_LOCAL_INDEX])
	  env_print_model_label("L", &first_line_used);
	env_print_tag_defs(td);

	/* Control models: */
	if (td = env_tags[tag].handler[ENV_CONTROL_INDEX])
	  env_print_model_label("C", &first_line_used);
	env_print_tag_defs(td);
	printf("\n");
    }

    printf("\n");
    return (0);
}

int32 env_print_registered_effects(void)
{
    int32	i;
    
    for (i = 0; i < ENV_NUM_EFFECTS; i++)
    {
	if (env_effects[i].reqfcn)
	  printf("%-22s    %s\n", 
		 env_effects[i].effectname+4, 
		 env_effects[i].handlername);
    }
    printf("\n");
    return TRUE;
}

char *env_find_model_for_effect(int32 effect)
{
    if (effect >= ENV_NUM_EFFECTS)
      return(NULL);
    else
      return(env_effects[effect].handlername);
}

int32 env_check_enabled(
    char *name, 
    char *tag)
{
    READER_UNION	*triple;
    int32		i;
    
    for (i = 1; env_param_table.array 
	 && (i < env_param_table.array[0].integer); i++)
    {
	triple = env_param_table.array[i].array;
	if (triple[0].integer != 4)	/* includes 0th count integer */
	  continue;
	    
	if (((reader_get_symbol(name) == triple[1].charptr) 
	     || (reader_get_symbol("ANY") == triple[1].charptr))
	    && ((reader_get_symbol(tag) == triple[2].charptr)
		|| (reader_get_symbol("ANY") == triple[2].charptr)))
	{
	    if (reader_get_symbol("on") == triple[3].charptr)
	      return(1);
	    else if (reader_get_symbol("disabled") == triple[3].charptr)
	      return(2);
	    else
	      return(0);
	}
    }
    return(3);
}


