/*
 * $RCSfile: libenv_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libenv_local.h
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libenvironment
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: rg
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libenvironment.h"
#include <libreader.h>

/*
 * Structure used to track generic statistics.
 * Generic statistics are not model-specific. 
 * For model-specific statistics, look at the 
 * use of the ENV_INTERNAL_STATISTICS tag.
 */
typedef struct _ENV_STATISTICS_TYPE {
    int32 total_gets;
    int32 gets_by_tag[ENV_NUM_TAGS];
    int32 total_errors;
    int32 errors_by_code[ENV_ERROR_LAST_VALUE + 1];
    int32 intervis_obscured;
} ENV_STATISTICS_TYPE;
extern uint16 env_keep_statistics;
extern ENV_STATISTICS_TYPE env_statistics;

/* Per-vehicle environment class information */
typedef struct environment_vars
{
    int32 TEMPLATE_dummy;
} ENVIRONMENT_VARS;

typedef struct env_tag_def 
{
    enum env_tag_query		tag;
    int32			md;
    int32			type;
    int32			active;
    ENV_HANDLER			getfcn;
    ENV_HANDLER			setfcn;
    struct env_tag_def		*m_next;
    struct env_tag_def		*t_next;
} ENV_TAG_DEF_T;

#define ENV_RESOLVER_INDEX	0
#define ENV_AMBIENT_INDEX	1
#define ENV_LOCAL_INDEX		2
#define ENV_CONTROL_INDEX	3
#define ENV_MODEL_TO_INDEX(m) (((m)<=4)?((m)>>1):(3))
#define ENV_MODEL_VALID(m)	(ENV_MODEL_TO_INDEX(m)<ENV_NUM_MODEL_TYPES)

#define ENV_MAX_NAME		32

typedef enum env_getset
{
    ENV_GET,
    ENV_SET
} ENV_GETSET;

typedef struct env_tag_ctl
{
    char			name[ENV_MAX_NAME];
    ENV_TAG_DEF_T		*handler[ENV_NUM_MODEL_TYPES];
    int32			num_models;
    int32			num_calls;
    float64			total_time;

    int32			*cached_get_models;
    int32			cached_num_get_models;
    int32			*cached_set_models;
    int32			cached_num_set_models;
} ENV_TAG_CTL;


typedef struct env_model_def 
{
    char			name[ENV_MAX_NAME];
    void			(*tickfcn)();
    int32			(*savefcn)();
    int32			(*loadfcn)();
    ENV_TAG_DEF_T		*handler;
    int32			num_tags;
} ENV_MODEL_DEF_T;

typedef struct env_effect_def
{
    char			effectname[ENV_MAX_NAME];
    char			handlername[ENV_MAX_NAME];
    int32			(*reqfcn)();
} ENV_EFFECT_DEF_T;




extern int32 env_num_models;
extern struct env_model_def	*env_models;
extern struct env_effect_def	env_effects[];
extern struct env_tag_ctl	*env_tags;
extern READER_UNION env_param_table;

/* VERY PRIVATE VARIABLES AND FUNCTIONS!  DO NOT USE OUTSIDE! */
extern int32 env_user_data_handle;

/* env_utils.c: */
extern char *env_model_name(int32 md);
extern int32 env_num_models_for_tag(
    ENV_TAG	tag,
    int32	types,
    ENV_GETSET	getset);
extern int32 env_find_models(
    ENV_TAG	tag,
    int32	*buf,
    int32	bsize,
    int32	types,
    ENV_GETSET	getset);
extern int32 *env_alloc_model_array(
    ENV_TAG	tag,
    int32	n,
    int32	types,
    ENV_GETSET	getset);
extern void env_dealloc_model_array(
    int32 	*array);
extern ENV_TAG_DEF_T *env_find_handler(ENV_TAG tag);
extern ENV_TAG_DEF_T *env_find_handler_for_model(
    int32	model,
    ENV_TAG	tag);
extern int32 env_print_registered_effects(void);

/* env_set.c: */
extern int32 env_set_explicit(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);


