/*
 * $RCSfile: libenvironment.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libenvironment.h
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libenvironment
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: rg
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBENVIRONMENT_H_INCLUDED
#define _LIBENVIRONMENT_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> 
#include <stdext.h>
#include <basic.h>
#include <libctdb.h>


typedef enum env_em_band
{
    ENV_BAND_UNKNOWN,
    ENV_BAND_47MM,	/* visual */
    ENV_BAND_49MM,	/* near infrared */
    ENV_BAND_712,
    ENV_BAND_106,
    ENV_BAND_35,
    ENV_BAND_812,	/* far infrared */
    ENV_BAND_1060,
    ENV_BAND_94,
    ENV_NUM_BANDS
} ENV_EM_BAND;

/** IMPORTANT!!! 
  * The ordering of the tags affect all the environment libraries.
  * Tags are grouped together for ease of processing by individual
  * libraries. Atmospheric tags are listed first, followed by global
  * sea tags and deep sea tags. The next group is surf only tags
  * followed by control tags. When adding tags, the groupings need to
  * be consistent!!!
  **/
typedef enum env_tag_query
{
    ENV_INTERNAL_STATS,
    ENV_TEMPERATURE,
    ENV_DEWPOINT,
    ENV_RELATIVE_HUMIDITY,
    ENV_BAROMETRIC_PRESSURE,
    ENV_WIND_VELOCITY,
    ENV_PRECIPITATION_TYPE,
    ENV_PRECIPITATION_RATE,
    ENV_EXTINCTION_TYPE,
    ENV_EXTINCTION_AMOUNT,
    ENV_EXTINCTION_COEFF,
    ENV_RAY_VISIBILITY,
    ENV_VISUAL_RANGE,
    ENV_ILLUMINATION,
    ENV_SKY_OVER_GROUND,
    ENV_CLOUD_COVER,
    ENV_CLOUD_HEIGHT,
    ENV_CLOUD_CEILING,
    ENV_CLOUD_TYPE,
    ENV_SIM_TIME,
    ENV_SUN_POSITION,
    ENV_MOON_POSITION,
    ENV_MOON_ELONGATION,
    ENV_CONTRAST,
    ENV_SOLAR_ILLUMINATION,
    ENV_LUNAR_ILLUMINATION,
    ENV_SKY_ILLUMINATION,
    ENV_SEA_STATE,
    ENV_SEA_SALINITY,
    ENV_SEA_NUM_WAVES,
    ENV_SEA_NUM_FREQUENCIES,
    ENV_SEA_TIDE_LEVEL,
    ENV_LAST_RAIN_TIME,	/* Precipitation amount in last 24 hr */
    ENV_SUB_SOIL_SATURATION,
    ENV_TOTAL_PRECIP_IN_HR,

    /* The following 6 tags are serviced by both the deep sea model 
     * and the surf zone model in net-sea
     */
    ENV_SEA_SURFACE_HEIGHT,
    ENV_SEA_SURFACE_TEMP,
    ENV_SEA_WAVE_DIRECTION,
    ENV_SEA_WAVE_PERIOD,
    ENV_SEA_WAVE_HEIGHT,
    ENV_SEA_CURRENT_VELOCITY,
    ENV_SEA_SPECTRAL_RECORD,
    /* The 3 following tags are serviced by sirf only models
     */
    ENV_SEA_BREAKER_ANGLE,
    ENV_SEA_BREAKER_TYPE,
    ENV_SEA_IN_SURF_ZONE, 



    /* Some magical tags used by env_set() to cause deferred setting.
     * This allows someone doing several env_set() calls in a row to
     * hint to the model that it can cache up a bunch of sets.  This
     * is useful if each set would result in a packet being sent out
     * on the network.  For example,
     *   env_set(ENV_BATCH, NULL);
     *   env_set(ENV_TEMPERATURE, &ei1);
     *   env_set(ENV_HUMIDITY, &ei2);
     *   env_set(ENV_PRESSURE, &ei3);
     *   env_set(ENV_FLUSH, NULL);
     * In this example, several related parameters are being set in
     * one pass.  If a model supported all three, and would normally
     * send out a packet on the network, it could instead notice the
     * ENV_BATCH and realize it can defer the network update until it
     * sees a ENV_FLUSH.
     */
     
    ENV_BATCH,
    ENV_FLUSH,

    /* Number of entries */
    ENV_NUM_TAGS
} ENV_TAG;

#define ENV_FIRST_ATMOS_TAG ENV_TEMPERATURE
#define ENV_FIRST_SEA_TAG   ENV_SEA_STATE
#define ENV_FIRST_SURF_TAG  ENV_SEA_SURFACE_HEIGHT
#define ENV_FIRST_SURF_ONLY_TAG ENV_SEA_BREAKER_ANGLE
#define ENV_FIRST_MISC_TAG  ENV_BATCH

#define ENV_NUM_ATMOS_TAGS ((int32)ENV_FIRST_SEA_TAG - (int32)ENV_FIRST_ATMOS_TAG)
#define ENV_NUM_SEA_TAGS ((int32)ENV_FIRST_SURF_ONLY_TAG-(int32)ENV_FIRST_SEA_TAG)
#define ENV_NUM_SURF_TAGS ((int32)ENV_FIRST_MISC_TAG - (int32)ENV_FIRST_SURF_TAG)
#define ENV_NUM_ALL_SEA_TAGS ((int32)ENV_FIRST_MISC_TAG - (int32)ENV_FIRST_SEA_TAG)

/* Support for testing invalid values. This can be especially 
 * useful for values transmitted inside of PDUs. Values flagged
 * "components" have the values assigned to each of the components
 * (cells in an array or fields in a structure).
 *
 * NOTE!!! To allow applications the most flexibility, use values 
 * that fit in uint16 and float32 (see /usr/include/limits.h)
 */

#define ENV_FLOAT64_INVALID -990099.0
#define ENV_INT32_INVALID -9009
#define ENV_ENUM_INVALID 9009

#define ENV_TEMPERATURE_INVALID		ENV_FLOAT64_INVALID
#define ENV_DEWPOINT_INVALID		ENV_FLOAT64_INVALID
#define ENV_RELATIVE_HUMIDITY_INVALID	ENV_FLOAT64_INVALID
#define ENV_BAROMETRIC_PRESSURE_INVALID	ENV_FLOAT64_INVALID
#define ENV_WIND_VELOCITY_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_PRECIPITATION_TYPE_INVALID	ENV_ENUM_INVALID
#define ENV_PRECIPITATION_RATE_INVALID	ENV_FLOAT64_INVALID
#define ENV_EXTINCTION_TYPE_INVALID	ENV_ENUM_INVALID
#define ENV_EXTINCTION_AMOUNT_INVALID	ENV_FLOAT64_INVALID
#define ENV_EXTINCTION_COEFF_INVALID	ENV_FLOAT64_INVALID
#define ENV_RAY_VISIBILITY_INVALID	ENV_FLOAT64_INVALID
#define ENV_VISUAL_RANGE_INVALID	ENV_FLOAT64_INVALID
#define ENV_ILLUMINATION_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_SKY_OVER_GROUND_INVALID	ENV_FLOAT64_INVALID
#define ENV_CLOUD_COVER_INVALID		ENV_FLOAT64_INVALID
#define ENV_CLOUD_CEILING_INVALID	ENV_FLOAT64_INVALID
#define ENV_CLOUD_BASE_INVALID   	ENV_FLOAT64_INVALID
#define ENV_CLOUD_TYPE_INVALID          ENV_INT32_INVALID
#define ENV_SIM_TIME_INVALID		ENV_INT32_INVALID
#define ENV_SUN_POSITION_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_MOON_POSITION_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_MOON_ELONGATION_INVALID	ENV_FLOAT64_INVALID
#define ENV_CONTRAST_INVALID		ENV_FLOAT64_INVALID
#define ENV_SOLAR_ILLUMINATION_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_LUNAR_ILLUMINATION_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_SKY_ILLUMINATION_INVALID	ENV_FLOAT64_INVALID /* components */
#define ENV_TOTAL_PRECIP_INVALID	ENV_FLOAT64_INVALID
#define ENV_LAST_RAIN_TIME_INVALID	ENV_INT32_INVALID
#define ENV_SOIL_SATURATION_INVALID	ENV_INT32_INVALID

#define ENV_FLOAT_INVALID_RANGE         0.4
#define ENV_NS_IS_VALUE_VALID(v, t)   (((v) >= ((t) + ENV_FLOAT_INVALID_RANGE)) || \
                                        ((v) <= ((t) - ENV_FLOAT_INVALID_RANGE)) )

/*
 * Put any helpful conversions here:
 */

/* Used to convert barometric pressure pascals<->atmospheres: */	
#define ENV_ATMS_PER_PASCAL 9.89233e-06


typedef enum env_effect_request
{
    ENV_SMOKE_EFFECT,
    ENV_LIGHT_EFFECT,
    ENV_SOUND_EFFECT,
    ENV_FLARE_EFFECT,
    ENV_DUST_EFFECT,

    ENV_NUM_EFFECTS
} ENV_EFFECT;

typedef enum env_source_type
{

	ENV_SOURCE_POINT_DETONATION,
	ENV_SOURCE_BARRAGE,
	ENV_SOURCE_VEHICLE

} ENV_SOURCE_TYPE;

/***************************************************************/


/* For models that want to operate on this structure in a fairly
 * generic way, input fields should be padded out to 128 bytes,
 *
 * $$ Support for max and min values will probably be removed in
 * $$ a future version.  their use has been inconsistent, and cannot
 * $$ be counted on to be set for all models.
 *
 * $$ Support for using vehicle_id's is discouraged.  Whether this
 * $$ "feature" will remain in future versions or not is being debated.
 */

typedef union env_interface
{
    union
    {
	struct env_generic_value
	{
	    char	padding[128];
	    
	    float64	output_value;
	    float64	output_max;
	    float64	output_min;
	} generic_value;

	struct env_generic_vector_to_value
	{
	    float64	input_vector[XYZC];
	    int32	input_vehicle_id;
	    char	padding[92];
	    
	    float64	output_value;
	    float64	output_max;
	    float64	output_min;
	} generic_vector_to_value, 
	temperature,
	relative_humidity,
	barometric_pressure,
	precipitation_rate,
	extinction_amount,
	dewpoint,
	moon_elongation,
	cloud_cover,
	cloud_ceiling,
	cloud_height,
	pasquill_stability,
	sea_surface_temp,
	sea_salinity,
	sea_wave_height,
	sea_wave_direction,
	sea_wave_period,
	sea_surface_height,
	sea_tide_level,
	sea_breaker_angle,
	total_precip_in_hr;

	struct env_generic_vector_to_type
	{
	    float64	input_vector[XYZC];
	    int32	input_vehicle_id;
	    char	padding[92];
	    
	    uint32	output_type;
	    float64	output_max;
	    float64	output_min;
	} generic_vector_to_type,
	precipitation_type,
	extinction_type,
	cloud_type,
	sea_state,
	sea_in_surf_zone,
	sea_num_waves,
	sea_num_frequencies,
	sea_breaker_type,
	sub_soil_saturation;
	
	struct env_generic_vector_pair_to_value
	{
	    float64	input_vector_1[XYZC];
	    float64	input_vector_2[XYZC];
	    char	padding[64];
	    
	    float64	output_value;
	    float64	output_max;
	    float64	output_min;
	} generic_vector_pair_to_value;

	struct env_generic_visual_check
	{
	    float64	input_viewer[XYZC];
	    float64	input_target[XYZC];
	    ENV_EM_BAND	input_wavelength;
	    int32	input_vid;
	    int32	input_tid;
	    char	padding[(56-sizeof(ENV_EM_BAND))];
	    
	    float64	output_value;
	    float64	output_max;
	    float64	output_min;
	} generic_visual_check,
	visual_range,
	contrast,
	ray_visibility,
	extinction_coeff,
	sky_over_ground;

	/* $$$$ This definition is temporary!!!! */
	struct env_visual_check_to_vector
	{
	    float64	input_vector[XYZC];
	    float64	input_target[XYZC];
	    ENV_EM_BAND	vector_info;
	    int32	input_vid;
	    int32	input_tid;
	    char	padding[(56-sizeof(ENV_EM_BAND))];

	    float64	output_vector[XYZC];
	    float64	output_max_mag;
	    float64	output_min_mag;
	} visual_check_to_vector,
	illumination,
	solar_illumination,
	lunar_illumination,
	sky_illumination;
		    
	struct env_generic_vector_to_vector
	{
	    float64	input_vector[XYZC];
      	    char	padding[96];
	    
	    float64	output_vector[XYZC];
	    float64	output_max_mag;
	    float64	output_min_mag;
	} generic_vector_to_vector,
	wind_velocity,
	moon_position,
	sun_position,
	sea_current_velocity;
		
	struct env_time_date 
	{
	    uint32 date;
	    uint32 time;
	    uint32 uxtime;
	} time_date,
	sim_time,
	last_precip_event_time;

	struct env_internal
	{
	    /* Only the first two fields are used with env_sets() */
	    uint16	keep_stats;	/* whether to keep statistics */
	    uint16	zero_stats;	/* whether to zero statistics */

	    uint32	num_get_calls; 
	    uint32	num_set_calls; 
	    
	    uint32	total_time;	/* Maintain using calls to 
					   env_start_timing and 
					   env_stop_timing */
	} internal;
        struct env_spectral_data
        {
            float64     input_vector[XYZC];
            int32       long_wave_num;
            int32       frequency_num;

            float32     output_frequency;
            float32     output_amplitude;
            float32     output_phase;
            float32     output_wave_number_X;
            float32     output_wave_number_Y;
        }spectral_data;
    } u;
} ENV_INTERFACE;

typedef union env_effect_interface
{
    union
    {
	struct env_smoke
	{
	    float64	location[XYZC];
	    ObjectType	source;
	    int32       num_sources;
            ENV_SOURCE_TYPE source_type;/* ENV_SOURCE_POINT_DETONATION, etc. */
            float64    detonate_velocity[3];/* angle from which detonation */
		                            /* happened.  Used for barrages. */
	} smoke, flare;
	struct env_dust
	{
	    float64	location[XYZC];
	    ObjectType	source;
	    VehicleID simnet_id;
	      /* if cause==vehicle: vehicle creating the dust */
	} dust;
	
    } u;
} ENV_EFFECT_INTERFACE;

/* The following lines define the classification scheme for models. */

#define ENV_RESOLVER_MODEL	1
#define ENV_AMBIENT_MODEL	2
#define ENV_LOCAL_MODEL		4
#define ENV_CONTROL_MODEL	8

/* We don't include CONTROL models in this */
#define ENV_ALL_MODELS		7


#define ENV_NUM_MODEL_TYPES	4

/**
 ** environment parametric data structure.
 ** (unused)
 **/

typedef struct environment_parametric_data
{
    /* Network-friendly variables.
     */
    int32 TEMPLATE_dummy;
} ENVIRONMENT_PARAMETRIC_DATA;

/* 
 * Function type of  handlers 
 */
typedef int32 (*ENV_HANDLER) (/*
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models */);

/** 
 ** Functions relating to global libenvironment operation
 **
 **/

/* env_init.c: */
extern void env_init(
    char	*directory,
    uint32	flags);

/* env_register.c: */
extern int32 env_register_model(
    char		*mname,
    void		(*tickfcn)(),
    int32               (*savefcn)(),
    int32               (*loadfcn)());
extern int32 env_register_tag(
    int32		md,
    ENV_TAG		tag,
    int32		type,
    ENV_HANDLER		getfcn,
    ENV_HANDLER		setfcn);
extern int32 env_register_tags(
    int32		md,
    ENV_TAG		*tags,
    int32		num_tags,
    int32		type,
    ENV_HANDLER		getfcn,
    ENV_HANDLER		setfcn);
extern int32 env_find_num_models(void);
extern int32 env_register_effect_handler(
    int32		effect,
    char		*handlername,
    int32		(*reqfcn)());

/* env_get.c: */
extern int32 env_get_explicit(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);
extern int32 env_set_explicit(
    ENV_TAG           tag,
    ENV_INTERFACE     *interface,
    int32             *models,
    int32             num_models);
extern int32 env_get(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface);
extern int32 env_get_ambient(
    int32		tag,
    ENV_INTERFACE	*interface);

/* env_utils.c: */
extern char *env_tag_name(ENV_TAG tag);
extern int32 env_ambient_set_models_exist(ENV_TAG tag);
extern int32 env_ambient_get_models_exist(ENV_TAG tag);
extern void env_active(
    int32	md,
    ENV_TAG	tag,
    int32	active);
extern int32 env_find_model(char *mname);
extern int32 env_print_registered_models(void);
extern int32 env_print_registered_effects(void);
extern char *env_find_model_for_effect(int32 effect);
extern int32 env_check_enabled(
    char *name, 
    char *tag);

/* env_tick.c: */
extern void env_tick(void);
    
/* env_set.c: */
extern int32 env_set(
    ENV_TAG		tag,
    ENV_INTERFACE	*interface);

/* env_effect.c: */
extern int32 env_request_effect(
    ENV_EFFECT			effect,
    ENV_EFFECT_INTERFACE	*interface);

/* env_resolvers.c: */
extern int32 env_vmm_multiply(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);
extern int32 env_vmm_add(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);
extern int32 env_vmm_vadd(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);
extern int32 env_use_priority(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);
extern int32 env_surf_zone(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);

/* env_stats.c: */
extern int32 env_clear_statistics(void);
extern int32 env_start_statistics(void);
extern int32 env_stop_statistics(void);
extern int32 env_print_statistics(void);
extern int32 env_print_internal_stats(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models);
extern void env_start_timing(uint16 keep_stats, uint32 *start_time);
extern void env_update_timing(
    uint16 keep_stats, 
    uint32 start_time, 
    uint32 *total_time);
extern int32 env_save_all(char *fname);
extern int32 env_load_all(char *fname);

/* Error codes for env_get() or env_get_explicit().
 *
 * N.B: Make sure the ENV_ERROR_LAST_VALUE reflects the
 * largest value assigned to any error code. The error codes
 * are used as indices into an array for the purposes of 
 * statistics gathering!!!
 */
#define ENV_ERROR_UNKNOWN 0
#define ENV_ERROR_NO_MODEL 1		/* no model registered */
#define ENV_ERROR_INVALID_INPUT 2	/* error in input values */
#define ENV_ERROR_UNHANDLED_INPUT 3	/* input valid; inadequate model */
#define ENV_ERROR_INTERNAL_TO_MODEL 4   /* error in model while processing */
#define ENV_ERROR_LAST_VALUE ENV_ERROR_INTERNAL_TO_MODEL


extern int32 env_last_get_error;

/* These work just like errno and sys_errlist[].  These error codes are 
 * used for the scenario save and lload functions. When an error occurs,
 * the reason is left in env_errno and a string description can be found
 * in env_errlist[env_errno].
 */

extern int32 env_errno;
extern char *env_errlist[];

#define ENV_ENOMEM		1 /* Memory allocation failed */
#define ENV_EMODINV		2 /* Invalid model number */
#define ENV_ESAVST		3 /* Error saving state for client */
#define ENV_EOPEN		4 /* Error opening file */
#define ENV_EWRITE		5 /* Error writing into file */
#define ENV_EREAD		6 /* Error reading from file */
#define ENV_ECLOSE		7 /* Error closing file */
#define ENV_ENOENV		8 /* No env entity to save from */
#define ENV_ECLLOAD		9 /* Client cannot load scenario */
#define ENV_EREMOTE_SIM		10/* Simulator already exists remotely for 
				     unique entity. Cannot simulate */


extern CTDB *env_ctdb;

#ifdef __cplusplus
}
#endif

#endif /*_LIBENVIRONMENT_H_INCLUDED*/


