#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: environment test program                                  *
*       Created: Mon Aug 16 13:52:58 EDT 1993                               *
*       Author: rg                                                          *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include <stdext.h>
#include "libenv_local.h"

static void test_tick();
static int32 test_get_local();
static int32 test_get_ambient();
static int32 test_get_resolver();
static int32 test_set_local();
static int32 test_set_ambient();
static float local_temp = 50.0;
static float ambient_temp = 20.0;

main()
{
    int32	md1, md2, md3;
    ENV_INTERFACE	temp;
    bzero(&temp, sizeof(ENV_INTERFACE));

    time_init(5000);
    sched_init(2, 67, 29000);
    env_init(".", READER_OVERRIDE);
    
    md1 = env_register_model("Test1", test_tick);
    md2 = env_register_model("Test2", test_tick);
    md3 = env_register_model("Test3", test_tick);

   
    env_register_tag(md1,
		     ENV_TEMPERATURE,
		     ENV_AMBIENT_MODEL,
		     test_get_ambient, test_set_ambient);
    env_register_tag(md1,
		     ENV_RELATIVE_HUMIDITY,
		     ENV_AMBIENT_MODEL,
		     test_get_ambient, test_set_ambient);
    env_register_tag(md2,
		     ENV_TEMPERATURE,
		     ENV_LOCAL_MODEL,
		     test_get_local, test_set_local);
    env_register_tag(md3,
		     ENV_TEMPERATURE,
		     ENV_RESOLVER_MODEL,
		     test_get_resolver, test_set_local);

    env_get(ENV_TEMPERATURE, &temp);
    printf("the temperature is %f\n", temp.u.temperature.output_value);


    temp.u.temperature.output_value = 30.0;
    env_set(ENV_TEMPERATURE, &temp);
    temp.u.temperature.output_value = 0.0;

    env_get(ENV_TEMPERATURE, &temp);
    printf("the temperature is %f\n", temp.u.temperature.output_value);

    env_get(ENV_TEMPERATURE, &temp);
    printf("the temperature is %f\n", temp.u.temperature.output_value);
}

static void test_tick()
{
    printf("I never get called.\n");
}

static int32 test_set_ambient(tag,interface)
    int32	tag;
    ENV_INTERFACE	*interface;
{

    ambient_temp = interface->u.temperature.output_value;

}

static int32 test_set_local(tag,interface)
    int32	tag;
    ENV_INTERFACE	*interface;
{

    local_temp = interface->u.temperature.output_value;

}

static int32 test_get_ambient(tag, interface)
    int32	tag;
    ENV_INTERFACE	*interface;
{
    
    
    switch(tag)
    {
      case ENV_TEMPERATURE:
	interface->u.temperature.output_value = ambient_temp;
	return(0);
	break;
      default:
	return(-1);
    }
}

static int32 test_get_local(tag, interface)
    int32		tag;
    ENV_INTERFACE 	*interface;
{
    switch(tag)
    {
      case ENV_TEMPERATURE:
	interface->u.temperature.output_value = local_temp;
	return(0);
	break;
      default:
	return(-1);
    }
}
    

static int32 test_get_resolver(tag, interface, models, num_models)
    int32		tag;
    ENV_INTERFACE	*interface;
    int32		*models;
    int32		num_models;
{
    ENV_INTERFACE	local_interface;
    int32 i;

    bcopy(interface, &local_interface, sizeof(ENV_INTERFACE));

    switch(tag)
    {
      case ENV_TEMPERATURE:

	local_interface.u.temperature.output_value = 0.0;
	
	for (i = 0; i < num_models; i++)
	{
	    if (models[i] != -1)
	    {
		if (env_get_explicit(tag, &local_interface, &models[i], 1) 
		    != -1)
		  interface->u.temperature.output_value += 
		    local_interface.u.temperature.output_value;
	    }
	}

	
	return(0);
	break;
      default:
	return(-1);
    }
}
    

