/*
 * $RCSfile: eui_parser.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
#ifndef lint
static char rcsid [] = "$RCSfile: eui_parser.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: eui_parser.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: parser interface for libenvironment                                *
*       Created: May 1995						    *
*       Author:  krobasky
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_PARSER

#include <stdio.h>
#include <libparser.h>
#include <stdalloc.h>
#include <stdext.h>

#include "libeui_local.h"

static char *command_table_callback();
static char *print_table_callback();
static char *set_table_callback();
static char *stats_table_callback();


/*ARGSUSED*/
static void set_debug_flag(argc, argv)
    int32 argc;
    int32 argv[];
{
    int32 flag = argv[0];

    eui_set_debugging(flag, EUI_BASE_PREFS_PTR);
}

/*ARGSUSED*/
static void print_em_band(argc, argv)
    int32 argc;
    int32 argv[];
{
    printf("Electromagnetic band set to:  %s\n", 
	   eui_em_band_name( eui_get_em_band(EUI_BASE_PREFS_PTR) ) );
}

/*ARGSUSED*/
static void set_em_band(argc, argv)
    int32 argc;
    int32 argv[];
{
    int32  band = argv[0];
    if( (band >= ENV_NUM_BANDS) || ( band <= ENV_BAND_UNKNOWN) )
      printf("%d unknown, bandwidth not set \n", band);
    else
      eui_set_em_band(band, EUI_BASE_PREFS_PTR);
    print_em_band(0, NULL);
}

#if 0
/*ARGSUSED*/
static void print_surf_depth(argc, argv)
    int32 argc;
    int32 argv[];
{
    printf("Surf depth threshold set to:  %d\n", 
	   eui_get_surf_depth(EUI_BASE_PREFS_PTR)); /* $$$ */
}

/*ARGSUSED*/
static void set_surf_depth(argc, argv)
    int32 argc;
    int32 argv[];
{
    int32  depth = argv[0];

    if(depth < 0)
      printf("%d less than zero, depth not set\n", depth);
    else
      eui_set_surf_depth(depth, EUI_BASE_PREFS_PTR);
    print_surf_depth(0, NULL);
}

#endif 

/*ARGSUSED*/
static void print_environment(argc, argv)
    int32 argc;
    int32 argv[];
{
    printf("Environmental models:\n");
    env_print_registered_models();

    printf("Requestable environmental effects:\n");
    env_print_registered_effects();
}

/*ARGSUSED*/
static void print_statistics(argc, argv)
    int32 argc;
    int32 argv[];
{
    (void)env_print_statistics();
}

/*ARGSUSED*/
static void on_statistics(argc, argv)
    int32 argc;
    int32 argv[];
{
    (void)env_start_statistics();
}

/*ARGSUSED*/
static void off_statistics(argc, argv)
    int32 argc;
    int32 argv[];
{
    (void)env_stop_statistics();
}

/*ARGSUSED*/
static void clear_statistics(argc, argv)
    int32 argc;
    int32 argv[];
{
    (void)env_clear_statistics();
}


#ifndef lint

static DEFINE_TABLE(set_table)

    KEYWORD_SELECT("     Generic environmental set")

        KEYWORD ("em_band", "- electromagnetic band for intervisibility")
          GETDECIMAL("um: 1:.4-.7, 2:.4-.9, 3:.7-1.2, 4:1.06, 5:3-5, 6:8-12, 7:10.6, 8:94")
          CALL (set_em_band)
        END_KEYWORD

#if 0
        KEYWORD ("surf_depth", "- depth at which surf starts")
          GETDECIMAL("(positive integer, in meters)")
          CALL (set_surf_depth)
        END_KEYWORD
#endif

        KEYWORD ("debug", "- debug for libenvui")
          GET_ON_OR_OFF
          CALL (set_debug_flag)
        END_KEYWORD

    END_KEYWORD_SELECT

END_DEFINE_TABLE


static DEFINE_TABLE(print_table)

    KEYWORD_SELECT("     Generic environmental print")

        KEYWORD ("handlers", "- registered libenvironment handlers")
            CALL (print_environment)
        END_KEYWORD

        KEYWORD ("statistics", "- generic libenvironment statistics")
            CALL (print_statistics)
        END_KEYWORD

        KEYWORD ("em_band", "- electromagnetic band for intervisibility")
            CALL (print_em_band)
        END_KEYWORD

#if 0
        KEYWORD ("surf_depth", "- depth at which surf starts")
            CALL (print_surf_depth)
        END_KEYWORD
#endif

    END_KEYWORD_SELECT

END_DEFINE_TABLE


static DEFINE_TABLE(stats_table)

    KEYWORD_SELECT("     Generic environmental statistics")

        KEYWORD ("on", "- turn on statistics collection")
            CALL (on_statistics)
        END_KEYWORD

        KEYWORD ("off", "- turn off statistics collection")
            CALL (off_statistics)
        END_KEYWORD

        KEYWORD ("clear", "- reset (zero) current statistics")
            CALL (clear_statistics)
        END_KEYWORD

        KEYWORD ("print", "- print current statistics summary")
            CALL (print_statistics)
        END_KEYWORD

    END_KEYWORD_SELECT

END_DEFINE_TABLE


/* Top level libenvironment parser table. */

static DEFINE_TABLE(command_table)

    KEYWORD_SELECT("     Environmental control commands")

        KEYWORD ("print", "- generic print menu")
          DO_DYNOBOUND_KEYWORD_TABLE(print_table_callback)
        END_KEYWORD

        KEYWORD ("set", "- generic set menu")
          DO_DYNOBOUND_KEYWORD_TABLE(set_table_callback)
        END_KEYWORD

        KEYWORD ("statistics", "- generic statistics menu")
          DO_DYNOBOUND_KEYWORD_TABLE(stats_table_callback)
        END_KEYWORD

  END_KEYWORD_SELECT

END_DEFINE_TABLE

static DEFINE_TABLE(top_table)

        KEYWORD ("environment", "- environment operations")
          DO_DYNOBOUND_KEYWORD_TABLE(command_table_callback)
        END_KEYWORD

END_DEFINE_TABLE


PARSE_TABLE **eui_command_table;
PARSE_TABLE **eui_set_table;
PARSE_TABLE **eui_print_table;
static PARSE_TABLE **eui_stats_table;

static char *command_table_callback()
{
    return (char *) *eui_command_table;
}

static char *set_table_callback()
{
    return (char *) *eui_set_table;
}

static char *print_table_callback()
{
    return (char *) *eui_print_table;
}

static char *stats_table_callback()
{
    return (char *) *eui_stats_table;
}

void eui_init_parser(main_table_ptr)
    PARSE_TABLE **main_table_ptr;
{

    ParseAddTable(main_table_ptr, top_table);

    /* models_table defines the skeletal environmental effects parser
     * interface.  As models register with libenvironment, they will
     * want to add there own tables to the interface.
     * eui_models_table starts out as just a copy of the skeleton,
     * and as new models add their components, libparser will resize
     * eui_models_table accordingly.  This is why we can't just use
     * models_table to add new components.  This is also why we need
     * to dynamically bind eui_models_table to the eui_table using
     * DO_DYNOBOUND_KEYWORD_TABLE; because eui_models_table is not
     * static */

    eui_command_table = (PARSE_TABLE **) STDALLOC(sizeof(PARSE_TABLE *));
    *eui_command_table = (PARSE_TABLE *) STDALLOC(sizeof(command_table));
    bcopy(command_table, *eui_command_table, sizeof(command_table));

    eui_set_table = (PARSE_TABLE **) STDALLOC(sizeof(PARSE_TABLE *));
    *eui_set_table = (PARSE_TABLE *) STDALLOC(sizeof(set_table));
    bcopy(set_table, *eui_set_table, sizeof(set_table));

    eui_print_table = (PARSE_TABLE **) STDALLOC(sizeof(PARSE_TABLE *));
    *eui_print_table = (PARSE_TABLE *) STDALLOC(sizeof(print_table));
    bcopy(print_table, *eui_print_table, sizeof(print_table));

    eui_stats_table = (PARSE_TABLE **) STDALLOC(sizeof(PARSE_TABLE *));
    *eui_stats_table = (PARSE_TABLE *) STDALLOC(sizeof(stats_table));
    bcopy(stats_table, *eui_stats_table, sizeof(stats_table));

}

#endif /* lint */

#endif /*USE_PARSER*/


