#ifndef lint
static char rcsid [] = "$RCSfile: eui_pref.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: eui_pref.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Operator preferences for euiironmental settings
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: krobasky
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libeui_local.h"
#include <stdalloc.h>

typedef struct eui_preferences {
    int32       debugging;
    ENV_EM_BAND	em_band;
#if 0
    int32       surf_depth;
#endif
}EUI_PREFS;

static EUI_PREFS base_prefs;

/* Always use base_prefs when the passed pointer is NULL */
#define SELECT_PREF(p) (p?(p):(&base_prefs))


EUI_PREFS_PTR eui_create_prefs()
{
    EUI_PREFS_PTR prefs;

    prefs = (EUI_PREFS_PTR) STDALLOC (sizeof(EUI_PREFS));

    /* Set defaults */
    eui_init_prefs(prefs);

    return (prefs);
}

void eui_init_prefs(prefs)
    EUI_PREFS_PTR prefs;
{
    SELECT_PREF(prefs)->em_band = 1;
#if 0
    SELECT_PREF(prefs)->surf_depth = 3;
#endif
    SELECT_PREF(prefs)->debugging = 0;

}

void eui_set_debugging(flag, prefs)
    int32 flag;
    EUI_PREFS_PTR prefs;
{
    SELECT_PREF(prefs)->debugging = flag;
}

void eui_set_em_band(band, prefs)
    ENV_EM_BAND	band;
    EUI_PREFS_PTR prefs;
{
    if( ( band > ENV_BAND_UNKNOWN) && (band < ENV_NUM_BANDS) )
      SELECT_PREF(prefs)->em_band = band;
    else if(SELECT_PREF(prefs)->debugging)
      printf("%d unknown, bandwidth not set \n", band);
}


ENV_EM_BAND eui_get_em_band(prefs)
    EUI_PREFS_PTR prefs;
{
    return(SELECT_PREF(prefs)->em_band);
}

char *eui_em_band_name(band)
    ENV_EM_BAND band;
{
    switch(band)
    {
      case ENV_BAND_47MM:
	return("DVO-Visible(.4-.7um)");
      case ENV_BAND_49MM:
	return("ENV_BAND_49MM"); /* $$$ what is this? */
      case ENV_BAND_712:
	return("Near IR(.7-1.2um)");
      case ENV_BAND_106:
	return("Near IR(1.06um)");
      case ENV_BAND_35:
	return("Near IR(3-5um)");
      case ENV_BAND_812:
	return("Far IR(8-12um)");
      case ENV_BAND_1060:
	return("Microwave(10.6um)");
      case ENV_BAND_94:
	return("Far IR(9.4um)");
      case ENV_BAND_UNKNOWN:
      default:
	return("Unknown");
    }
}

#if 0
/*
 * $$$ This code introduces a circular dependency with libenvironment
 *     and is not to be integrated.
 */
void eui_set_surf_depth(depth, prefs)
    int32 depth;
    EUI_PREFS_PTR prefs;
{
    if(depth >= 0)
      SELECT_PREF(prefs)->surf_depth = depth;
    else if(SELECT_PREF(prefs)->debugging)
      printf("%d less than zero, depth not set.\n", depth);
}


int32 eui_get_surf_depth(prefs)
    EUI_PREFS_PTR prefs;
{
    return(SELECT_PREF(prefs)->surf_depth);
}
#endif
