/*
 * $RCSfile: libenvui.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libenvui.h
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libenvui
*       Created: Fri Apr 15 09:05:09 EDT 1994
*       Author: krobasky
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBENVUI_H_INCLUDED
#define _LIBENVUI_H_INCLUDED

#include <stdtypes.h> 
#include <basic.h>
#include <libenvironment.h>

#if USE_PARSER

#include <stdio.h>
#include <libparser.h>


/* Environmental parser tables:  eui_<name>_table
 * 
 * To add a new module table, call:
 *   "ParseAddTable(eui_<name>_table, <new table>)" 
 */

/* Parser tables of general environmental commands should be added here. */
extern PARSE_TABLE **eui_command_table;

/* Parser tables of print commands should be added here. */
extern PARSE_TABLE **eui_print_table;

/* Parser tables of set commands should be added here. */
extern PARSE_TABLE **eui_set_table;

#endif /* USE_PARSER */



#define EUI_BASE_PREFS_PTR NULL
typedef struct eui_preferences *EUI_PREFS_PTR;

/* eui_init
 *
 * Points to parser table to which the environmental tables should be
 * added.
 */
 extern void eui_init();

/* eui_create_prefs
 *
 * Creates a new environmental user preferences structure.
 */
extern EUI_PREFS_PTR eui_create_prefs();

/* eui_set_em_band
 *
 * Sets the electromagnetic band for the specified preference pointer
 */
extern void eui_set_em_band(/* ENV_EM_BAND band, EUI_PREFS_PTR prefs */);

/* eui_get_band
 *
 * Returns the integer number corresponding to the current
 * electromagnetic band setting for the specified "prefs"
 * structure. 
 */
extern ENV_EM_BAND eui_get_em_band(/* EUI_PREFS_PTR prefs */);

/* eui_em_band_name
 *
 * Electromagnetic band for which to find the name
 */
extern char *eui_em_band_name(/* ENV_EM_BAND band */);

#if 0
/* eui_set_surf_depth
 *
 * Sets the surf depth for the specified preference pointer
 */
extern void eui_set_surf_depth(/* int32 depth, EUI_PREFS_PTR prefs */);

/* eui_get_surf_depth
 *
 * Returns the integer depth that defines the surf zone
 * for the specified "prefs" structure. 
 */
extern int32 eui_get_surf_depth(/* EUI_PREFS_PTR prefs */);

#endif

#endif /*_LIBENVUI_H_INCLUDED*/
