#ifndef lint
static char rcsid [] = "$RCSfile: except.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: except_fpes.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code for turning on and off aborting on floating point    *
*		  exceptions.						    *
*       Created: 14 May 94				
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libexcept.h"
#include <signal.h>


int32 except_asserts_on;

#ifdef xxx

#if defined(sgi)
int32 except_fpes_on;
#endif

void except_init(asserts_on, fpes_on)
    int32 asserts_on;
    int32 fpes_on;
{
    except_asserts_on = asserts_on;
#if defined(sgi)
    except_fpes_on = fpes_on;
#endif

#ifdef freebsd
    {
	struct sigaction sa, sa2;
	sa.sa_handler = SIG_IGN;
	sigfillset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sigaction(SIGFPE, &sa, &sa2);
    }
#endif

}

void except_set_asserts(asserts_on)
    int32 asserts_on;
{
    except_asserts_on = asserts_on;
}

void except_set_fpes(fpes_on)
    int32 fpes_on;
{
#if defined(sgi)
    except_fpes_on = fpes_on;
    if (except_fpes_on == FALSE)
    {
	except_fpes_off_internal();
    }
#endif
}

#if defined(sgi)

/* IRIX 4.0.5 has a problem in sigfpe.h, which declares several arrays as 
 * globals, instead of externs.  This means the linker will complain about
 * multiply defined symbols.  This UGLY, horrible hack will work around this
 * brokenness.
 */

#if defined(sgi4)
#define sigfpe_ ___this_is_ugly_and_horrible
#define invalidop_results_ ___dont_ask_you_dont_want_to_know
#define invalidop_operands_ ___i_told_you_not_to_ask

#include <sigfpe.h>

#undef sigfpe_
#undef invalidop_results_
#undef invalidop_operands_

extern struct sigfpe_template sigfpe_[_N_EXCEPTION_TYPES + 1];
extern int invalidop_results_[_N_INVALIDOP_RESULTS+1];
extern int invalidop_operands_[_N_INVALIDOP_OPERANDS+1];

#else

/* IRIX 5 got a brain */
#include <sigfpe.h>

#endif

#include <stdext.h>

static struct sgi_fpe_info
{
    int32 modsaf_bit;
    int32 sgi_bit;
    int32 sgi_number;
} fpe_info[] = { 
    {
	EXCEPT_OVERFLOW,
	_EN_OVERFL,
	_OVERFL
    },
    {
	EXCEPT_UNDERFLOW,
	_EN_UNDERFL,
	_UNDERFL
    },
    {
	EXCEPT_DIVZERO,
	_EN_DIVZERO,
	_DIVZERO
    },
    {
	EXCEPT_INVALID,
	_EN_INVALID,
	_INVALID
    }
};

#define NFPES (sizeof(fpe_info)/sizeof(struct sgi_fpe_info))


void except_fpes_on_internal(exception_mask)
    int32 exception_mask;
{
    int i;
    int sgi_mask = 0;

    for (i=0; i<NFPES; i++)
    {
	if (fpe_info[i].modsaf_bit & exception_mask)
	{
	    sgi_mask |= fpe_info[i].sgi_bit;
	    sigfpe_[fpe_info[i].sgi_number].abort=1;
	}
	else
	{
	    sigfpe_[fpe_info[i].sgi_number].abort=0;
	}
    }

    handle_sigfpes(_ON, sgi_mask, 0, _ABORT_ON_ERROR, 0);
}


void except_fpes_off_internal()
{
    handle_sigfpes(_OFF, 0, 0, _ABORT_ON_ERROR, 0);
}

#endif /* defined(sgi) */

#endif

