#ifndef _LIBEXCEPT_H_INCLUDED
#define _LIBEXCEPT_H_INCLUDED

/*
 * $RCSfile: libexcept.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libmovemap.h                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:							    *
*		The general error philosophy in ModSAF is to continue on an *
*		error if at all possible.  libexcept contains debugging aids *
*		that will stop the program on an error.  The program only
*		aborts if exceptions are activated so when fielded ModSAF
*		can continue on in the face of errors.
*
*       Created: May 14 1994                                                *
*       Author: swarts                                                      *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#include <stdlib.h>
#include <stdext.h>

extern void except_init(/* int32 asserts_on, int32 fpes_on */);
    /* except_init initializes the library.  If asserts_on is FALSE then
     * asserts will do nothing.  If fpes_on is FALSE then no floating
     * point exceptions will be generated.
     */

extern void except_set_asserts(/* int32 asserts_on */);
extern void except_set_fpes(/* int32 fpes_on */);


extern int32 except_asserts_on;	/* Variable that controls asserts. 
				 * This is external so EXCEPT_ASSERT can
				 * be a macro. */


/* EXCEPT_ASSERT(expression) prints a message and aborts is asserts_on is
 * TRUE and expression is FALSE.  This is like the ANSI assert except it
 * is controlled by a run-time switch.
 */
#if defined(__LINE__) && defined(__FILE__)
#define EXCEPT_ASSERT_PRINTF fprintf(stderr, \
      "Assertion failed: file \"%s\", line %d\n", __FILE__, __LINE__)
#else
#define EXCEPT_ASSERT_PRINTF fprintf(stderr, "Assertion failed\n")
#endif

#define EXCEPT_ASSERT(expression) \
   if (except_asserts_on) \
     if (expression) ; else {EXCEPT_ASSERT_PRINTF; abort();}



/*
 * EXCEPT_FPES_ON(exception_mask) turns on aborting on the floating point
 * exceptions specified in exception_mask if fpes_on is TRUE.
 *
 * EXCEPT_FPES_OFF turns off aborting on floating point exceptions.
 *
 * The intent of these macros is to bracket sections on code where floating
 * point exceptions are not expected with EXCEPT_FPES_ON and EXCEPT_FPES_OFF.
 * Unexpected FPEs will cause the program to abort and the problem to be found.
 *
 * Currently exception detection only happens on SGI machine.  In other cases
 * the macros do nothing.
 */

/* We define our own names for FPES for fortability */
#define EXCEPT_OVERFLOW 1
#define EXCEPT_UNDERFLOW 2
#define EXCEPT_DIVZERO 4
#define EXCEPT_INVALID 8

/* For convenience we define ALL and ALL_BUT_UNDERFLOW since we are not usually
 * interested in underflows.
 */
#define EXCEPT_ALL_FPES (EXCEPT_OVERFLOW|EXCEPT_UNDERFLOW|\
		       EXCEPT_DIVZERO|EXCEPT_INVALID)
#define EXCEPT_ALL_BUT_UNDERFLOW (EXCEPT_OVERFLOW|EXCEPT_DIVZERO|\
				  EXCEPT_INVALID)

 



extern int32 except_fpes_on;	/* This is external so we can have macros. */

extern void except_fpes_on_internal(/* int32 mask */);
    /* except_fpes_on turns on exception detection.  It should not be called
     * directly.  EXCEPT_FPES_ON should be used instead.
     */

extern void except_fpes_off_internal(/* void */);
    /* except_fpes_off turns off exception detection.  It should not be called
     * directly.
     */

#if defined(sgi)

#define EXCEPT_FPES_ON(mask) if (except_fpes_on) except_fpes_on_internal(mask)
#define EXCEPT_FPES_OFF except_fpes_off_internal();

#else /* defined(sgi) */

#define EXCEPT_FPES_ON(mask)
#define EXCEPT_FPES_OFF

#endif /* defined(sgi) */

#endif /*_LIBEXCEPT_H_INCLUDED*/



