#ifndef lint
static char rcsid [] = "$RCSfile: gcs_access.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: gcs_access.c
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libgcs                            *
*       Created: Tue Aug 29 17:36:21 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#include "libgcs_local.h"
#include <libvecmat.h>

extern GCS_LATMAP         gcs_latmap[180];
extern GCS_CELL_DATA_TYPE gcs_cell_to_frame[GCS_MAX_CELL];
extern int32              gcs_active_cells[GCS_MAX_CELL];
extern int32              gcs_n_cells;
extern int32              gcs_cell_aware;

int32                     gcs_cell_aware=GCS_MULTI_CELL;

/* -------------------------------------------------------------------------
 * --------------------- GET ROUTINES --------------------------------------
 * -------------------------------------------------------------------------
 */
ADDRESS
gcs_get_tdb(int32 cell, int32 thinned)
{
    if ( !gcs_valid_cell(&cell) )
      return (ADDRESS)NULL;
    else
      if(thinned)
        return gcs_cell_to_frame[cell].thinned_tdb;
      else
        return gcs_cell_to_frame[cell].tdb;
}

int32
gcs_get_rot_mat(int32 cell, float64 matrix[3][3])
{
    /* Note: rotation matrix format
     *
     *       -sin_lon  -sin_lat*cos_lon     cos_lat*cos_lon
     *        cos_lon  -sin_lat*sin_lon     cos_lat*sin_lon
     *        0.0       cos_lat             sin_lat
     */
    if ( !gcs_valid_cell(&cell) )
    {
	VMAT3_MAT_ZERO64(matrix);
	return FALSE;
    }

    if ( !gcs_cell_to_frame[cell].coord)
    {
	float64 lat,lng;
	
	gcs_get_cell_center(cell,&lat,&lng);
	gcs_init_coord_data(cell,lat,lng);

	/* Failure to allocate and populate coordinate transformation
	 * data structures
	 */
	if ( !gcs_cell_to_frame[cell].coord )
	{
	    VMAT3_MAT_ZERO64(matrix);
	    return FALSE;
	}	
    }
    

    VMAT3_MAT_COPY64(gcs_cell_to_frame[cell].coord->matrix, matrix);    
    return TRUE;
}

int32
gcs_get_origin_in_gcc(int32 cell, float64 *location)
{
    if ( !gcs_valid_cell(&cell) )
    {
        VMAT3_VEC_ZERO64(location);	
	return FALSE;
    }
    
    if ( !gcs_cell_to_frame[cell].coord) 
    {
	float64 lat,lng;
	
	gcs_get_cell_center(cell,&lat,&lng);
	gcs_init_coord_data(cell,lat,lng);

	/* Failure to allocate and populate coordinate transformation
	 * data structures
	 */
	if ( !gcs_cell_to_frame[cell].coord )
	{
	    VMAT3_VEC_ZERO64(location);	
	    return FALSE;
	}	
    }

    VMAT3_VEC_COPY64(gcs_cell_to_frame[cell].coord->origin_in_gcc, location);
    return TRUE;
}


ADDRESS
gcs_get_pbt_info(int32 cell)
{
    if ( !gcs_valid_cell(&cell) )
      return (ADDRESS)NULL;
    else
      return gcs_cell_to_frame[cell].pbt_info;
}

int32
gcs_get_origin_offsets(int32 cell, float64 *xoff, float64 *yoff)
{
    if ( !gcs_valid_cell( &cell ) )
    {
        *xoff = 0;
        *yoff = 0;
	return FALSE;
    }
    
    if (!gcs_cell_to_frame[cell].coord) 
    {
	float64 lat,lng;
	
	gcs_get_cell_center(cell,&lat,&lng);
	gcs_init_coord_data(cell,lat,lng);
	
	/* Failure to allocate and populate coordinate transformation
	 * data structures
	 */
	if (!gcs_cell_to_frame[cell].coord )
	{
	    *xoff = 0;
	    *yoff = 0;
	    return FALSE;
	}	
    }

    *xoff = gcs_cell_to_frame[cell].coord->x_off;
    *yoff = gcs_cell_to_frame[cell].coord->y_off;
    return TRUE;
 
}


/* -------------------------------------------------------------------------
 * --------------------- SET ROUTINES --------------------------------------
 * -------------------------------------------------------------------------
 */
int32 
gcs_set_tdb(int32 cell, ADDRESS tdb, int32 thinned)
{
    if ( !gcs_valid_cell( &cell ) )
      return FALSE;

    /* The user really shouldn't call this without first calling init_cell,
     * but if it happens, we'll call init_cell ourselves
     */
    if (!gcs_cell_to_frame[cell].coord)
      gcs_init_cell(cell, NULL);
	
    if(thinned)
      gcs_cell_to_frame[cell].thinned_tdb  = tdb;
    else
      gcs_cell_to_frame[cell].tdb  = tdb;
    return TRUE;
}


int32
gcs_set_pbt_info(int32 cell, ADDRESS pbt)
{
    if ( !gcs_valid_cell( &cell ) )
      return FALSE;

    /* The user really shouldn't call this without first calling init_cell,
     * but if it happens, we'll call init_cell ourselves
     */
    if ( !gcs_cell_to_frame[cell].coord )
      gcs_init_cell(cell, NULL);
	
    gcs_cell_to_frame[cell].pbt_info  = pbt;
    return TRUE;
}


int32
gcs_offset_origin(int32 cell, float64 xoff, float64 yoff)
{
    if ( !gcs_valid_cell( &cell ) )
      return FALSE;

    if (!gcs_cell_to_frame[cell].coord)
      gcs_init_cell(cell, NULL);
	
    gcs_cell_to_frame[cell].coord->x_off += xoff;
    gcs_cell_to_frame[cell].coord->y_off += yoff;

    return TRUE;
}    

int32
gcs_set_coord_data(int32 cell, float64 rot_matrix[3][3], float64 *gcc_origin)
{
    /* Note: rotation matrix format
     *
     *       -sin_lon  -sin_lat*cos_lon     cos_lat*cos_lon
     *        cos_lon  -sin_lat*sin_lon     cos_lat*sin_lon
     *        0.0       cos_lat             sin_lat
     */
    
    if ( !gcs_valid_cell(&cell) )
      return FALSE;

    if ( !gcs_cell_to_frame[cell].coord)
      gcs_init_cell(cell, NULL);

    VMAT3_MAT_COPY64(rot_matrix, gcs_cell_to_frame[cell].coord->matrix);    

    VMAT3_VEC_COPY64(gcc_origin, gcs_cell_to_frame[cell].coord->origin_in_gcc);
    return TRUE;
}


void
gcs_set_cell_awareness(int32 aware)
{
    gcs_cell_aware = aware;
}
