#ifndef lint
static char rcsid [] = "$RCSfile: gcs_conv.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/******************************************************************************
 *   File: gcs_conv.c                                                         *
 *                                                                            *
 *       Copyright 1995 by SAIC & TASC                                        *
 *                                                                            *
 *       SAIC                          TASC                                   *
 *       20 Burlington Mall Road       55 Walkers Brook Drive                 *
 *       Suite 130                     Reading, MA 01867                      *
 *       Burlington, MA 01803          617-942-2000                           *
 *       617-221-7612                                                         *
 *                                                                            *
 *       This software was developed by Loral under U. S. Government contracts*
 *       and may be reproduced by or for the U. S. Government pursuant to     *
 *       the copyright license under the clause at DFARS 252.227-7013         *
 *       (OCT 1988).                                                          *
 *                                                                            *
 *       Contents: conversion from topcentric to geocentric coordinates,      *
 *                 and vice-versa                                             *
 *                 also contains some functions compute the 4 letter cell     *
 *                 any longitude/latitude belongs in                          *
 *       Created: Thu Aug. 17 1995                                            *
 *       Author: hlu              (SAIC)                                      *
 *       Remarks:                                                             *
 *       Functions to convert coordinates between                             *
 *       gcs and other coordinate systems and vice versa.                     *
 *                                                                            *
 *****************************************************************************/

#include <math.h>
#include <basic.h>
#include <stdcoord.h>
#include "libgcs_local.h"
/* -------------------------------------------------------------------------
 * --------------------- EXTERNAL LIBRARY GLOBALS --------------------------
 * -------------------------------------------------------------------------
 */
extern int32 gcs_default_cell;
extern int32 gcs_cell_aware;
extern int32 gcs_active_cells[GCS_MAX_CELL];
/* Items needed to perform a conversion from TCC (UTM_NE based) to GCC and 
 * from GCC to TCC (UTM_NE based).
 */
extern ADDRESS                 gcs_tcc_ptr;
extern GCS_GCC_TO_TCC_FUNCTION gcs_gcc_to_tcc;
extern GCS_TCC_TO_GCC_FUNCTION gcs_tcc_to_gcc;    
extern GCS_GET_DATUM_INFO      gcs_get_datum_info;


/* -------------------------------------------------------------------------
 * ---------------------  FILE STATIC GLOBALS  -----------------------------
 * -------------------------------------------------------------------------
 */

void
gcs_get_gcc_datum_info(float64 lat,
		       float64 lng,
		       int32 *datum,
		       float64 *major,
		       float64 *minor,
		       float64 *dx,
		       float64 *dy,
		       float64 *dz)
{
    *datum = DATUM_WGS84;
    *major = WGS84_MAJOR_AXIS;
    *minor = WGS84_MINOR_AXIS;
    *dx = *dy = *dz = 0;
}

void
gcs_llz_to_gcc(float64 lat,
	       float64 lng,
	       float64 z,
	       int32 dat,
	       float64 *u,
	       float64 *v,
	       float64 *w)
{
    float64 lat_rad, lng_rad, cos_lat, cos_lng, sin_lat, sin_lng;
    float64 dx,dy,dz,major,minor,major_sq,minor_sq,E2,Rn;
    int32 datum=dat;

    if ( datum == DATUM_WGS84 || datum == DATUM_UNKNOWN )
      gcs_get_gcc_datum_info(lat, lng, &datum, &major, &minor, &dx, &dy, &dz);
    else
      gcs_get_datum_info(lat, lng, &datum, &major, &minor, &dx, &dy, &dz);
      
    lat_rad = DEG_TO_RAD(lat);
    lng_rad = DEG_TO_RAD(lng);

    major_sq = (major*major);
    minor_sq = (minor*minor);
    
    E2 = (major_sq-minor_sq)/major_sq;

    cos_lat = cos(lat_rad);
    sin_lat = sin(lat_rad);
    cos_lng = cos(lng_rad);
    sin_lng = sin(lng_rad);

    Rn = major / (sqrt((1 - (E2*sin_lat*sin_lat))));

    /* Compute the location in from_datum cartesian */
    *u = (Rn+z)*cos_lat*cos_lng + dx ;
    *v = (Rn+z)*cos_lat*sin_lng + dy ;
    *w = ((((minor_sq/major_sq)*Rn+z))*sin_lat) + dz ;
}

int32
gcs_gcs3e_to_gcs(int32 cell0,
		 float64 x0,
		 float64 y0,
		 float64 z0,
		 int32 *cell1,
		 float64 *x1,
		 float64 *y1,
		 float64 *z1)
{
    float64 u, v, w;
   
    if ( !gcs_cell_aware )
    {
	/* If the library is in cell unaware mode then the coordinate 
	 * conversion from gcs to gcs is just the returning of the same 
	 * input (but with the cell id changed to be set to the valid
	 * cell id
	 */
	*x1    = x0;
	*y1    = y0;
	*z1    = z0;
	*cell1 = gcs_default_cell;
    }
    else
    {
	if (!gcs_gcs3e_to_gcc(cell0, x0, y0, z0, &u, &v, &w) ||
	    !gcs_gcc3e_to_gcs(u, v, w, cell1 , x1, y1, z1) )
	{
	    *x1    = *y1 = *z1 = 0.0;
	    *cell1 = gcs_default_cell;
	    return FALSE;
	}
    }

    return TRUE;
}

int32
gcs_gcs3e_to_gcscs(int32   cell0,
		   float64 x0,
		   float64 y0,
		   float64 z0,
		   int32   cell1,
		   float64 *x1,
		   float64 *y1,
		   float64 *z1)
{
    float64 u, v, w;
    
    if ( !gcs_cell_aware || (cell0 == cell1) )
    {
	*x1 = x0;
	*y1 = y0;
	*z1 = z0;
    }
    else
    {
	if (!gcs_gcs3e_to_gcc(cell0, x0, y0, z0, &u, &v, &w) ||
	    !gcs_gcc3e_to_gcscs(u, v, w, cell1,x1, y1, z1) )
	{
	    *x1 = *y1 = *z1 = 0.0;
	    return FALSE;
	}
    }

    return TRUE;
}

/* Convert (x0, y0) in cell 0 to the equivalent (x1, y1) in cell1.  This
 * conversion assumes that the point in question is at height z=0.
 */
int32
gcs_gcs2e_to_gcscs(int32   cell0,
		   float64 x0,
		   float64 y0,
		   int32   cell1,
		   float64 *x1,
		   float64 *y1)
{
    float64 z;
    
    return gcs_gcs3e_to_gcscs(cell0,x0, y0, 0.0,  cell1, x1, y1, &z);
}
/* Convert (x0, y0) in cell 0 to the equivalent (x1, y1) in cell1.  This
 * conversion assumes that the point in question is at height z=0.
 */
int32
gcs_gcs2e_to_gcs(int32    cell0,
		 float64  x0,
		 float64  y0,
		 int32   *cell1,
		 float64 *x1,
		 float64 *y1)
{
    float64 gx,gy,gz,z;
    
    if ( !gcs_cell_aware )
    {
	/* If the library is in cell unaware mode then the coordinate 
	 * conversion from gcs to gcs is just the returning of the same 
	 * input (but with the cell id changed to be set to the valid
	 * cell id
	 */
	*x1    = x0;
	*y1    = y0;
	*cell1 = gcs_default_cell;
	return TRUE;
    }
    else
    {
	return (gcs_gcs3e_to_gcc(cell0, x0, y0, 0.0,  &gx, &gy, &gz) &&
		gcs_gcc3e_to_gcs(gx, gy, gz, cell1, x1, y1, &z));
    }
}

/* Convert pt0 to the equivalent pt1 in the coordinates of cell1.
 * This conversion assumes that the point pt0 is at height z=0.
 */
int32
gcs_gcs2_to_gcscspl(PointLocation *pt0,
		    int32 cell1,
		    PointLocation *pt1)
{
    float64 x,y;
    /* Note: check of cells will be handled in gcs2e_to_gcscs */
    int32 status = gcs_gcs2e_to_gcscs(pt0->cell,(float64)pt0->x,
				      (float64)pt0->y,cell1,&x,&y);
    pt1->x    = x;
    pt1->y    = y;
    pt1->cell = cell1;
    return status;
}

/* Convert xy0 in cell0 to the equivalent xy1 in cell1.  This
 * conversion assumes that the point xy0 is at height z=0.
 */
int32
gcs_gcs2_to_gcscs(int32 cell0, float64 *xy0, int32 cell1, float64 *xy1)
{
    /* Note: check of cells will be handled in gcs_gcs2e_to_gcscs */
    return gcs_gcs2e_to_gcscs(cell0, xy0[X], xy0[Y], 
			      cell1, &xy1[X], &xy1[Y] );
}


int32
gcs_gcc3e_to_gcs(float64 u,
		 float64 v,
		 float64 w,
		 int32   *cell,
		 float64 *x,
		 float64 *y,
		 float64 *z)
{ 
    float64 lat, lng, height = 0;
    
    /* If the conversion is based on the WGS84 ellipsoid then perform
     * the standard conversion.  If not then the conversion is special
     * case (COORD_UTM_NE type).
     */
    if ( !gcs_gcc_to_tcc )
    {
	/* Entry into here assumes that we are not taking into account
	 * a datum shift since the terrain is based on WGS-84 or COORD_GCC
	 * which is what the datum shift is trying to get back to.
	 */
	gcs_gcc_to_llz(u, v, w, &lat, &lng, &height, DATUM_WGS84);
    
	*cell = gcs_lookup_cell_id_ll(lat, lng);
	/* If you want to optimize, you can save on one call of 
	   gcs_lookup_cell_id_ll by just putting the code of 
	   gcs_gcc3e_to_gcscs here */
	return gcs_gcc3e_to_gcscs(u, v, w, *cell, x, y, z);
    }
    else
    {
	/* Special COORD_UTM_NE processing handled in coord_convert */
	gcs_gcc_to_tcc(u, v, w, gcs_tcc_ptr, x, y, z);
	*cell = gcs_default_cell;
	return TRUE;
    }
}

int32
gcs_gcc3e_to_gcscs(float64 u,
		   float64 v,
		   float64 w,
		   int32   cell,
		   float64 *x,
		   float64 *y,
		   float64 *z)
{
    float64 dx, dy, dz;
/*     float64 gcc_gcs_mat[3][3], gcs_gcc_mat[3][3]; */
    float64 gcs_gcc_mat[3][3];
    float64 result[XYZ], origin[XYZ];
    float64 x_off, y_off;
    
    if ( !gcs_valid_cell( &cell ) )
    {
	*x = *y = *z = 0.0;
	return FALSE;
    }
    
    /* If the conversion is based on the WGS84 ellipsoid then perform
     * the standard conversion.  If not then the conversion is special
     * case (COORD_UTM_NE type).
     */
    if ( !gcs_gcc_to_tcc )
    {
	gcs_get_rot_mat(cell,gcs_gcc_mat);
	gcs_get_origin_in_gcc(cell,origin);
	
	dx = u - origin[X];
	dy = v - origin[Y];
	dz = w - origin[Z];
	
	/* vmat3e_vec_mat_mul64 premultiplies the vector with the
         * transpose of the matrix.
         */
	vmat3e_vec_mat_mul64(dx, dy, dz, gcs_gcc_mat, result);
	gcs_get_origin_offsets(cell, &x_off, &y_off);
	
	/* Entry into here assumes that we are not taking into account
	 * a datum shift since the terrain is based on WGS-84 or COORD_GCC
	 * which is what the datum shift is trying to get back to.
	 */
	*x = result[X] - x_off;
	*y = result[Y] - y_off;
	*z = result[Z];
    }
    else
      /* Special COORD_UTM_NE processing handled in coord_convert */
      gcs_gcc_to_tcc(u, v, w, gcs_tcc_ptr, x, y, z);

    return TRUE;
}

int32 
gcs_gcs3e_to_gcc(int32 cell,
		 float64 x,
		 float64 y,
		 float64 z,
		 float64 *u,
		 float64 *v,
		 float64 *w)
{
    float64 matrix[3][3];
    float64 origin[XYZ],result[XYZ];
    float64 x_off, y_off;
  
    if ( !gcs_valid_cell( &cell ) )
    {
	*u = *v = *w = 0.0;
	return FALSE;
    }
     
    if ( !gcs_tcc_to_gcc )
    {
	gcs_get_rot_mat(cell,matrix);
	gcs_get_origin_in_gcc(cell,origin);
	
	gcs_get_origin_offsets (cell, &x_off, &y_off);
	vmat3e_mat_vec_mul64(matrix, x+x_off, y+y_off, z, result);
	
	/* Entry into here assumes that we are not taking into account
	 * a datum shift since the terrain is based on WGS-84 or COORD_GCC
	 * which is what the datum shift is trying to get back to.
	 */
	*u = result[X] + origin[X];
	*v = result[Y] + origin[Y];
	*w = result[Z] + origin[Z];
    }
    else
      gcs_tcc_to_gcc(gcs_tcc_ptr, x, y, z, u, v, w);

    return TRUE;
}

int32 
gcs_gcs_to_llz(int32 cell,
	       float64 x,
	       float64 y,
	       float64 z,
	       float64 *lat,
	       float64 *lng,
	       float64 *h,
	       int32 datum)
{
    float64 u, v, w;

    return (gcs_gcs3e_to_gcc(cell, x, y, z, &u, &v, &w) &&
	    gcs_gcc_to_llz(u, v, w, lat, lng, h, datum));
}


/* The Vector form of the same functions */
int32
gcs_gcs3_to_gcs(int32 cell0, 
		float64 xyz0[XYZ], 
		int32 *cell1, 
		float64 xyz1[XYZ])
{
    return gcs_gcs3e_to_gcs(cell0, xyz0[X],  xyz0[Y],  xyz0[Z], 
			    cell1, &xyz1[X], &xyz1[Y], &xyz1[Z]);
}


int32
gcs_gcs3_to_gcscs(int32 cell0, float64 *xyz0, int32 cell1, float64 *xyz1)
{
    return gcs_gcs3e_to_gcscs(cell0, xyz0[X],  xyz0[Y],  xyz0[Z], 
			    cell1, &xyz1[X], &xyz1[Y], &xyz1[Z]);
}

int32
gcs_gcc3_to_gcs(float64 *uvw, int32 *cell, float64 *xyz)
{
    return gcs_gcc3e_to_gcs(uvw[X], uvw[Y], uvw[Z],
			    cell, &xyz[X], &xyz[Y], &xyz[Z]);
}

int32 
gcs_gcs3_to_gcc (int32 cell, float64 *xyz, float64 *uvw)
{
    return gcs_gcs3e_to_gcc(cell, xyz[X], xyz[Y], xyz[Z], 
			    &uvw[X], &uvw[Y], &uvw[Z]);
}

int32 
gcs_gcc3_to_gcscs(float64 *uvw, int32 cell, float64 *xyz)
{
    return gcs_gcc3e_to_gcscs(uvw[X], uvw[Y], uvw[Z], 
			      cell, &xyz[X], &xyz[Y], &xyz[Z]);
}

int32
gcs_vec_to_vec(from_cell, from_vec, to_cell, to_vec)
    int32    from_cell;
    float64  from_vec[XYZ];
    int32    to_cell;
    float64  to_vec[XYZ];
{
    float64 from_matrix[3][3], trans_to_matrix[3][3];
    float64 gcc_vec[XYZ];
    
    if (!gcs_valid_cell( &from_cell ) || 
	!gcs_valid_cell( &to_cell ) )
    {
	to_vec[X] = to_vec[Y] = to_vec[Z] = 0.0;
	return FALSE;
    }

    if (from_cell == to_cell)
    {
	VMAT3_VEC_COPY64(from_vec, to_vec);
	return TRUE;
    }
    /* This is libpduapi's version of converting a tcc vector to a gcc vector
     * vmat3_vec_mat_mul32((float32 *)source, pa_tcc_gcc_rotation32,
     *                     (float32 *)sink);
     * where pa_tcc_gcc_rotation is the matrix multiplied to a gcc
     * coordinate to obtain a tcc coordinate.
     * vmat3_vec_mat_mul transposes the matrix before multiplication.
     */

    /* from_matrix and trans_to_matrix are matrices from gcs->gcc */
    gcs_get_rot_mat(from_cell, from_matrix);
    vmat3_mat_vec_mul64(from_matrix, from_vec, gcc_vec);
 
    gcs_get_rot_mat(to_cell, trans_to_matrix);
    /* vmat3_vec_mat_mul64 premultiplies the vector with the
     * transpose of the matrix.
     */
    vmat3_vec_mat_mul64(gcc_vec, trans_to_matrix, to_vec);
    return TRUE;
}

int32 
gcs_rot_to_rot(int32 from_cell,
	       float64 from_rot[3][3],
	       int32 to_cell,
	       float64 to_rot[3][3])
{
    float64 from_matrix[3][3], to_matrix[3][3], temp[3][3];
    float32 gcc_angles[XYZ];
    
    if (!gcs_valid_cell(&to_cell ) || 
	!gcs_valid_cell(&from_cell))
    {
	VMAT3_MAT_ZERO64(to_rot);
	return FALSE;
    }

    if (from_cell == to_cell)
    {
	VMAT3_MAT_COPY64(from_rot, to_rot);
	return TRUE;
    }

    /* The ModSAF libpduapi code to convert rotation matrices is: 
     * vmat3_rotation64_to_gcc_angles32((float64 **)source,
     *                                  &((float32 *)sink)[0],
     *					&((float32 *)sink)[1],
     *					&((float32 *)sink)[2],
     *					pa_tcc_gcc_rotation64);
     * where pa_tcc_gcc_rotation is a matrix whose transpose is used
     * to convert from tcc to gcc:
     * gcc_location = Transpose(pa_tcc_gcc_rotation64)*tcc_location + 
     *                          origin_in_gcc)
     */


    gcs_get_rot_mat(from_cell, temp);
    vmat3_transpose64(temp, from_matrix);
    
    gcs_get_rot_mat(to_cell, to_matrix);
    
    /* temp and to_matrix are rotation matrices used to convert
     *    from gcs to gcc.
     * from_matrix is a rotation matrix used to convert from gcc to gcs
     */
    vmat3_rotation64_to_gcc_angles32(from_rot, 
				     &gcc_angles[X], 
				     &gcc_angles[Y], 
				     &gcc_angles[Z], 
				     from_matrix);
    
    vmat3_gcc_angles32_to_rotation64(gcc_angles[X], 
				     gcc_angles[Y], 
				     gcc_angles[Z], 
				     to_rot, 
				     to_matrix);
    return TRUE;
}


static int32 last_lat, last_lng;

#define MAX_ARRAY_SIZE 110
static float64 t[MAX_ARRAY_SIZE];
static float64 tt[MAX_ARRAY_SIZE];
static float64 ang[MAX_ARRAY_SIZE];
static float64 ti[MAX_ARRAY_SIZE];
static float64 ti2[MAX_ARRAY_SIZE];


#define GCS_TABLE_SIZE 90.0
void
gcs_init_tables()
{
    int32    i;

    for (i=1;i<=GCS_TABLE_SIZE;++i)
      tt[i] = tan((2*i-1) * PI / (4*GCS_TABLE_SIZE));

    t[0] = ti[0]= ti2[0] = ang[0] = 0.0;
    
    for (i=2;i<=GCS_TABLE_SIZE+1;++i)
    {
	ang[i] = (2*i - 2.0) * PI / (4 * GCS_TABLE_SIZE);
	t[i]   = tan(ang[i]);
	ti[i]  = 1.0/t[i];
	ti2[i] = ti[i]*ti[i]+1.0;
    }
    
    t[0]  = -.1e-250;
    tt[0] = -.1e-250;

    /* Make sure that the latest entry is set correctly */
    {
	int32 index=(GCS_TABLE_SIZE+1);
	
	t[index]   = .1e250;
	ti[index]  = 0.0;
	ti2[index] = 1.0;
	tt[index]  = .179769313486231e309;
    }
}

#define compute_arctan(abs_tanx) \
{\
   if (abs_tanx == 0.0)\
     tn = 0.0;\
   else\
   {\
       /* Compute arctan */\
       if(abs_tanx >= tt[jl])\
       {\
	   jh = jl +1;\
	   while (abs_tanx >= tt[jh])\
	   {\
	       jl = jh;\
	       jh++;\
	   }\
       }\
       else\
       {\
	   jh = jl;\
	   jl--;\
	   \
	   while (abs_tanx <= tt[jl])\
	   {\
	       jh = jl;\
	       jl--;\
	   }\
       }\
       \
       ts = (jl == 0) ?\
	 abs_tanx : ti[jh] - (ti2[jh] / (ti[jh] + abs_tanx));\
       tn = ang[jh] + ts*(.9999999987 - ts*ts*(.333299));\
   }\
}   

/* Some variables we want to cache, and share with another function */
static float64 sin_phi1, cos_phi1, sin2_phi1, XN, tan_phi1, major, minor;
   
int32
gcs_gcc_to_llz(
	       float64 x, 
	       float64 y, 
	       float64 z,
	       float64 *lat, 
	       float64 *lng, 
	       float64 *height, 
	       int32 dat)
{
    /* Implementation of GCC->Geodetic conversion algorithm proposed
     * by Ralph M. Toms
     */

    static struct
    {
	float64 factor, max, val;
    } table[4] = { { 1.0065336, .43271886E14, 1.00329},
		   { 1.0051245, .7019318E14,  1.0026},
		   { 1.0034642, .15321828E15, 1.00092592},
		   { 1.0017567, .59429356E15, .999250297},
	         };
    float64 w, w_sq, inv_S0, S0, T0, test, sin_B0, cos_B0, bot, inv_bot;
    float64 temp, S1, inv_S1, S1_sq, abs_tanx, tn;
    float64 z_sq, T1, T1_sq, ts;
    float64 major_sq, minor_sq, sum_sq, inv_major, inv_minor;
    int32 i, j, jl, jh, at_pole = FALSE;

    int32 datum =dat;
    int32 loop_lim = 1;

    /* Some variables that we want to cache */
    static float64 dx, dy, dz;
    static float64 ceps2, aeps2, eps25, eps21;

    /* Just some bogus value */
    static int32 last_datum = DATUM_MAXIMUM_COUNT;

    for (j=0;j<loop_lim;j++)
    {
	if ((datum != last_datum) || (datum == DATUM_UNKNOWN))
	{
	    if ( datum == DATUM_UNKNOWN )
	    {
		int32 tmp_datum;
		
		/* if the datum is unknown then the datum to use first is 
		 * GCC (WGS-84).
		 */
		tmp_datum = DATUM_WGS84;
		gcs_get_gcc_datum_info((float64)0.0, (float64)0.0, &tmp_datum, &major, &minor, &dx, &dy, &dz);
		/* If a datum routine has been provided then first determine
		 * the lat/long based on the GCC/WGS84 model and then use the 
		 * lat/long to provide the datum shift values and then recompute
		 * the proper lat/long.
		 */
		loop_lim++;
	    }
	    else
	      gcs_get_datum_info((float64)0.0, (float64)0.0, &datum, &major, &minor, &dx, &dy, &dz);


	    /* Compute constants that we need, if not already cached */
	    sum_sq = major*major - minor*minor;
	    inv_major = 1 / major;
	    inv_minor = 1 / minor;
	    ceps2 = sum_sq * inv_minor;
	    aeps2 = sum_sq * inv_major;
	    
	    temp = aeps2 * inv_major;
	    eps25 = 0.25 * temp;
	    eps21 = temp - 1.0;
	    last_datum = datum;
	}

	x -= dx;
	y -= dy;
	z -= dz;
 
	/* Special case processing */
	if (x == 0.0)
	{
	    if (y > 0.0)
	      *lng = 90.0;
	    else    
	      if (y < 0.0)
		*lng = -90.0;
	      else    
	      {
		  at_pole = TRUE;
		  *lng = 0.0;
		  if(z > 0.0)
		    *lat = 90.0;
		  else
		  {
		      if (z < 0.0)
			*lat = -90.0;
		      else
		      {
			  *lng = 0.0;
			  *height = -major;
			  return FALSE;
		      }
		  }
	      }
	}
	
	w_sq = (x*x) + (y*y);
	w    = sqrt(w_sq);
	
	z_sq = z*z;
	
	for (i=0, T0 = 0; i < 3; i++)
	{
	    test = w_sq + table[i].factor*z_sq;
	    if (test <= table[i].max)
	    {
		T0 = table[i].val*z;
		break;
	    }
	}
	if (!T0)
	  T0=.997523508*z;
	
	
	S0     = sqrt(T0*T0 + w_sq);
	inv_S0 = 1 / S0;
	sin_B0 = T0 * inv_S0;
	cos_B0 = w * inv_S0;
	
	T1      = z + ceps2 * sin_B0 * sin_B0 * sin_B0;
	bot     = w - aeps2 * cos_B0 * cos_B0 * cos_B0;
	inv_bot = 1/bot;

	/***  PREPARE FOR NEW SINE AND COSINE ***/
	
	T1_sq = T1*T1;
	S1_sq = T1_sq + (bot*bot);
	
	S1       = sqrt(S1_sq);
	inv_S1   = 1 / S1;
	
	sin_phi1  = T1 * inv_S1;
	sin2_phi1 = sin_phi1 * sin_phi1;
	cos_phi1  = bot * inv_S1;

/*	XN = major / (temp + (temp - 0.25) / temp); */
	temp     = 0.5 - eps25 * sin2_phi1;
	XN       = temp * major / (temp*temp + temp - 0.25);
	*height  = (sin2_phi1 >= 0.85355339080) ? 
	  (z / sin_phi1) + (eps21 * XN) : (w * inv_bot * S1) - XN;
/*	  (z / sin_phi1) + (eps21 * XN) : (w / cos_phi1) - XN; */
	
	tan_phi1 = T1 * inv_bot;
	if (!at_pole)
	{
	    abs_tanx = fabs(tan_phi1);

	    jl = last_lat;
	    compute_arctan(abs_tanx);
	    last_lat = jl;

	    if(tan_phi1 < 0)
	      *lat = RAD_TO_DEG(-tn);
	    else
	      *lat = RAD_TO_DEG(tn);
	}
	
	
	if (x != 0.0)
	{
	    abs_tanx = fabs(y/x);
	    jl = last_lng;
	
	    compute_arctan(abs_tanx);
	    last_lng = jl;
	    if (y >= 0)
	      *lng = (x > 0) ? RAD_TO_DEG(tn) : RAD_TO_DEG(PI - tn);
	    else
	      *lng = (x > 0) ? RAD_TO_DEG(-tn) : RAD_TO_DEG(tn - PI);
	}

	if ( datum == DATUM_UNKNOWN )
	{
	    /* If the original datum was unknown. Check the datum for the give
	     * lat/long computed.  If the datum is DATUM_WGS84 then decrement
	     * the loop counter (since our first guess was correct) and allow
	     * exiting of this routine.
	     */
	    gcs_get_datum_info(*lat,*lng,&datum,&major,&minor,&dx,&dy,&dz); 
	    if ( datum == DATUM_WGS84 )
	    {
		last_datum = DATUM_WGS84;
		return TRUE;
	    }
	}
    }
    return TRUE;
}

    

void 
gcs_gcc_to_utm(float64 u, float64 v, float64 w, 
	       int32 *zone, float64 *northing, float64 *easting,
	       float64 *z)
{
    float64 axlat, axlng, axlon0, si2ph, co2ph, sm, tn2, tn4, al, al2;
    float64 poly1, poly2, cee, cee2;
    float64 lat, lng;
    static float64 epps2, eps2_sq, eps2_cubed, eps2;
    static float64 f, ef, ef_sq, ef_cubed, ef_fourth;
    int32 izone;

    /* Some values we can cache */
    static float64 const1, const2;
    static float64 poly1a, poly2a, poly3a, poly4a;
    static float64 poly5a, poly1b, poly2b, poly3b, poly4b;

    /* Just some bogus value */
    static float64 old_minor = -1;
    static float64 old_major = -1;

    /* Defining some constants that we will need */
#define inv_6   0.16666666666666666
#define inv_24  0.04166666666666666
#define inv_120 0.008333333333333333
#define inv_720 0.001388888888888888
#define cscale  0.9996
#define const3  9.5492965855137211   /*   30 / PI    */
#define const4  0.017453292519943295 /*   180 / PI   */

    gcs_gcc_to_llz(u, v, w, &lat, &lng, z, DATUM_UNKNOWN);
    axlng = DEG_TO_RAD(lng);
    axlat = DEG_TO_RAD(lat);

    if ((major != old_major) || (minor != old_minor)) 
    {
	/* Calculate intermediate constants */
	float64 minor_over_major = minor / major;
	
	f          = 1.0 - minor_over_major;
	ef         = f / (2. - f);
	eps2       = 1.0 - minor_over_major * minor_over_major;
	epps2      = eps2 / (1 - eps2);
	const1     = 58  * epps2;
	const2     = 330 * epps2;
	ef_sq      = ef       * ef;
	ef_cubed   = ef_sq    * ef;
	ef_fourth  = ef_cubed * ef;
	eps2_sq    = eps2    * eps2;
	eps2_cubed = eps2_sq * eps2;

	/* Compute needed constants */
	poly1a = 1 - eps2*0.25 - eps2_sq * 0.046875 - eps2_cubed*0.01953125;
	poly2a = 1.5 * ef      - ef_cubed*0.84375;
	poly3a = ef_sq*1.3125  - ef_fourth*1.71875;
	poly4a = ef_cubed * 1.5729166666666667;
	poly5a = ef_fourth * 2.142578125;
	poly1b = 1 - eps2*0.25 - eps2_sq*0.046875 - eps2_cubed*0.01953125;
	poly2b = eps2*0.375    + eps2_sq*0.09375  + eps2_cubed*0.0439453125;
	poly3b = eps2_sq*0.05859375 + eps2_cubed * 0.0439453125;
	poly4b = eps2_cubed * 0.011393229166666666;

	/* Remember for what values of major and minor we have cached data */
	old_major = major;
	old_minor = minor;
    }

    /* COMPUTE UTM COORDINATES */
    izone = 31+ (int32)(axlng*const3);
    
    if (axlng < 0.0)
      izone--;
    
    axlon0 = const4 * (float64)(6 * izone - 183); 
    tn2    = tan_phi1 * tan_phi1;
    cee    = epps2 * cos_phi1 * cos_phi1;
    al     = (axlng - axlon0) * cos_phi1;
    
    si2ph = 2.0 * sin_phi1 * cos_phi1;
    co2ph = 1.0 - 2 * sin_phi1 * sin_phi1;
    
    sm = major * (poly1b*axlat - si2ph * 
		  (poly2b - 2*co2ph * (poly3b - poly4b *
				       (1.0 - 2.0 * si2ph * si2ph))));
    tn4 = tn2 * tn2;
    cee2= cee * cee;
    al2 = al  * al;
    
    poly1  =  1.0 - tn2 + cee;
    poly2  =  5.0 - 18. * tn2 + tn4 + 72.0 * cee - const1;
    
    /* COMPUTE EASTING */
    
    *easting = cscale * XN * al * (1.0 + al2 * 
				   (inv_6 * poly1 + inv_120 * al2 * poly2));
    (*easting) += 500000.0;
    
    /* COMPUTE NORTHING */
    
    poly1  =  5.0 - tn2 + cee * (9.0 + 4.0 * cee);
    poly2  =  61.0 - 58.0 * tn2 + tn4 + 600.0 * cee - const2;
    
    *northing= cscale* (sm + XN * tan_phi1 * al2 * 
			(0.5 + al2*(inv_24*poly1 + inv_720 * al2 * poly2)));
    
    if (axlat <= 0.0)
      (*northing) += 10000000.0;
    
    *zone = izone;
    return;
}



