#ifndef lint
static char rcsid [] = "$RCSfile: gcs_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: gcs_init.c
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libGCS                                 *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgcs_local.h"
#include <stdstring.h>
#include <stdalloc.h>
#include <stdio.h>

extern int32              gcs_default_cell;
extern GCS_CELL_DATA_TYPE gcs_cell_to_frame[GCS_MAX_CELL];
extern int32              gcs_active_cells[GCS_MAX_CELL];
extern int32              gcs_n_cells;
extern int32              gcs_cell_aware;

/* Items needed to perform a conversion from TCC (UTM_NE based) to GCC and 
 * from GCC to TCC (UTM_NE based).
 */
extern GCS_GCC_TO_TCC_FUNCTION gcs_gcc_to_tcc;
extern GCS_TCC_TO_GCC_FUNCTION gcs_tcc_to_gcc;
extern GCS_GET_DATUM_INFO      gcs_get_datum_info;
extern ADDRESS                 gcs_tcc_ptr;

/* LIBGCS GLOBAL VARIABLES */
GCS_CELL_DATA_TYPE gcs_cell_to_frame[GCS_MAX_CELL];
int32              gcs_active_cells[GCS_MAX_CELL];
int32              gcs_n_cells;
int32              gcs_default_cell=GCS_ILLEGAL_CELL;

GCS_GCC_TO_TCC_FUNCTION gcs_gcc_to_tcc = NULL;
GCS_TCC_TO_GCC_FUNCTION gcs_tcc_to_gcc= NULL;
GCS_GET_DATUM_INFO      gcs_get_datum_info=NULL;
ADDRESS                 gcs_tcc_ptr = NULL;

int32 gcs_simnet;

static int32 library_init=FALSE;

void
gcs_init(GCS_GET_DATUM_INFO datum_func)
{
    if ( library_init )
       return;

    gcs_generate_latmap();
   
    /* Clear the contents of the cell_to_frame map */
    bzero(&gcs_cell_to_frame,sizeof(GCS_CELL_DATA_TYPE)*GCS_MAX_CELL);

    /* initialize the number of active cells to zero */
    gcs_n_cells = 0;

    /* set the datum function pointer needed to process datum shifts.  This 
     * logic is resident in libcoordinates.
     */
    if (datum_func)
      gcs_get_datum_info = datum_func;
    else
    {
	printf("No DATUM function provided to libGCS - WGS84 DATUM assumed\n");
	gcs_get_datum_info = gcs_get_gcc_datum_info;
    }
    library_init = TRUE;

    gcs_init_tables();
}



/* This sets up needed GCS data structures for a cell, and clears all
 * user defined ones.  If desired, it can be called more than once as a 
 * means of clearing all user-defined mappings.
 */

int32  
gcs_init_cell(int32 cell, GCS_FLAT_EARTH *flat_earth)
{
    float64 lat,lng;

    /* cell is GCS_ILLEGAL_CELL only if we are running under SIMNET
     * mode.
     */
    if (cell == GCS_ILLEGAL_CELL)
      gcs_simnet = 1;
    else 
      gcs_simnet = 0;
    
    /* The item gcs_default will be used to set the cell on either 
     * the selection of an illegal cell in multi cell mode or the 
     * default cell for single or simnet mode.  Default setting is 
     * GCS_ILLEGAL_CELL.
     */
    if ( !gcs_simnet && !gcs_cell_aware )
      gcs_default_cell = cell;
    
    /* If the flat earth pointer is non-null then we are processing 
     * a flat earth SIMNET database and special functions must be called
     * from gcs_gcs3e_to_gcc and gcs_gcc3e_to_gcscs.
     */
    if ( flat_earth )
    {
	/* Save the coordinate conversion functions and the required 
	 * tcc pointer (input to these functions).
	 */
	gcs_tcc_to_gcc = flat_earth->tccf_to_gcc;
	gcs_gcc_to_tcc = flat_earth->gcc_to_tccf;
	gcs_tcc_ptr    = flat_earth->tcc_ptr;
	return TRUE;
    }      
    
    /* To avoid making multiple copies of the same cell, check to see 
     * whether this is the first time we've initialized this cell
     */
    if (!gcs_cell_to_frame[cell].coord)
      gcs_active_cells[gcs_n_cells++] = cell;

    /* Initialize GCS-specific data items */
    gcs_get_cell_center(cell,&lat,&lng);
    gcs_init_coord_data(cell,lat,lng);

    /* NULL out user-supplied data items, so we'll notice if the user
     * asks for one that hasn't been set yet
     */
    gcs_cell_to_frame[cell].tdb = NULL;

    return TRUE;
}

