#ifndef lint
static char rcsid [] = "$RCSfile: gcs_util.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: gcs_util.c
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Drive                *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7600                                                        *
 *                                                                           *
 *	This software was developed by Loral under U. S. Government contracts*
 *	and may be reproduced by or for the U. S. Government pursuant to     *
 *	the copyright license under the clause at DFARS 252.227-7013         *
 *	(OCT 1988).                                                          *
 *                                                                           *
 *       Contents: Utility routines for libgcs                               *
 *       Created: Tue Aug 29 17:36:21 EDT 1995                               *
 *       Author: buettner                                                    *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/
#include "libgcs_local.h"
#include <stdalloc.h>
#include <stdcoord.h>
#include <stdio.h>
#include <math.h>

extern GCS_CELL_DATA_TYPE gcs_cell_to_frame[GCS_MAX_CELL];
extern int32              gcs_active_cells[GCS_MAX_CELL];
extern int32              gcs_n_cells;
extern int32              gcs_cell_aware;
extern int32              gcs_default_cell;

static GCS_LATMAP         gcs_latmap[181];

/* If operating in multi cell mode then only accept cell values between
 * GCS_MIN_CELL and GCS_MAX_CELL.  If running in single cell or simnet
 * mode then don't validate the cell field but set it to the default 
 * cell value (GCS_ILLEGAL_CELL for simnent and the cell that is mapped 
 * for the ctdb for single cell).
 */
int32 
gcs_valid_cell(int32 *cell)
{
    if ( gcs_cell_aware && (*cell >= GCS_MIN_CELL) && (*cell <= GCS_MAX_CELL))
      return TRUE;
    else
    {
	*cell = gcs_default_cell;
	return !gcs_cell_aware;
    }
}	

    
/* This function returns the center of the "real" GCS cell.  If the 
 * cell id is invalid or the database is a simnet database then the
 * answer is undefined and the output will be 0.  Otherwise the lat
 * long center point is returned.
 */
void
gcs_get_cell_center(int32 cell, float64 *lat, float64 *lng)
{
    float64 lat_min,lat_max,lng_min,lng_max;
    
    if (gcs_lookup_cell_extents(cell, &lat_min, &lat_max, &lng_min, &lng_max))
    {
	/* Point of tangency is determined to be the center of the cell */
	*lat = (lat_min + lat_max) * 0.5 ;
	*lng = (lng_min + lng_max) * 0.5 ;
    }
    else
      *lat = *lng = 0 ;
}
    
void
gcs_init_coord_data(int32 cell, float64 lat, float64 lng)
{
    if (!gcs_cell_to_frame[cell].coord)
      gcs_cell_to_frame[cell].coord = 
	(GCS_COORD_DATA *)STDALLOC(sizeof(GCS_COORD_DATA));

    bzero(gcs_cell_to_frame[cell].coord, sizeof(GCS_COORD_DATA));

    gcs_generate_gcs_gcc_tfm_data(cell,lat,lng);
}

void
gcs_generate_gcs_gcc_tfm_data(int32 cell, float64 lat, float64 lng)
{
    float64 sin_lat, sin_lng, cos_lat, cos_lng;
    float64 we, ze;
    static float64 gcs_const_mjr_sq=(WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS);
    static float64 gcs_const_f2f_sq=(2*WGS84_DEFORMATION - 
				     WGS84_DEFORMATION*WGS84_DEFORMATION);

    lat = DEG_TO_RAD(lat);
    lng = DEG_TO_RAD(lng);
    
    sin_lat = sin(lat);
    cos_lat = cos(lat);
    sin_lng = sin(lng);
    cos_lng = cos(lng);
    
    we = WGS84_MAJOR_AXIS * cos_lat / 
      (sqrt(1-gcs_const_f2f_sq*sin_lat*sin_lat));
    ze = WGS84_MINOR_AXIS * sqrt (1-we*we/(gcs_const_mjr_sq));
    
    if (lat < 0) 
      ze *= -1;                

    {
	float64 matrix[3][3];
	float64 origin_in_gcc[XYZ];

	/*
	 * This is the original equation, but h is always zero: 
	 * origin_in_gcc[0] = (we + h*cos_lat) * cos_lng;
	 * origin_in_gcc[1] = (we + h*cos_lat) * sin_lng;
	 * origin_in_gcc[2] =  ze + h*sin_lat;
	 */
		
	origin_in_gcc[X] = we * cos_lng;
	origin_in_gcc[Y] = we * sin_lng;
	origin_in_gcc[Z] = ze;
    
	/* This matrix is used to convert from GCS to GCC.
	 * position_gcc = matrix * position_gcs + position_origin_in_gcc 
	 */

	matrix[X][X] = -sin_lng;
	matrix[X][Y] = -sin_lat * cos_lng;
	matrix[X][Z] =  cos_lat * cos_lng;
	matrix[Y][X] =  cos_lng;
	matrix[Y][Y] = -sin_lat * sin_lng;
	matrix[Y][Z] =  cos_lat * sin_lng;
	matrix[Z][0] =  0;
	matrix[Z][Y] =  cos_lat;
	matrix[Z][Z] =  sin_lat;

	gcs_set_coord_data(cell,matrix,origin_in_gcc);
    }
 
    gcs_offset_origin(cell, 0.0, 0.0);
}


int32 
gcs_lookup_cell_id(float64 x, float64 y, float64 z)
                   /* GCC coordinates */
{
    float64 lat,lng,outz;

    if (!gcs_cell_aware)
      return gcs_default_cell;
    
    /* Note: a call to this routine assumes that the coordinate system
     * is based on WGS-84 ellipsoid (COORD_GCC) and that the datum shift
     * is irrelavant when converting gcc to lat/long.
     */
    gcs_gcc_to_llz(x,y,z,&lat,&lng,&outz,FALSE);   
    return gcs_lookup_cell_id_ll(lat,lng) ;   
}

#define GCS_LAT_TO_WGC(a) ( (a)==90.0?((a)+89.9):(a)+90.0 ) 
#define GCS_LNG_TO_WGC(a) ( (a)==180.0?0.0:(a)+180.0 ) 

int32 
gcs_lookup_cell_id_ll(float64 lat, float64 lng)
                 /* Geodetic coordinates range (-90..90)  */
                 /* Geodetic coordinates range (-180..180)*/
{
    int32 lat_index,lng_index,cell;
    
    if (!gcs_cell_aware)
      return gcs_default_cell;

    lat_index = GCS_LAT_TO_WGC(lat);
    lng_index = GCS_LNG_TO_WGC(lng) * ((gcs_latmap[lat_index].n_cell-1) /
				       gcs_latmap[lat_index].last_lng);
    cell = gcs_latmap[lat_index].base_cell + lng_index;
    (void)gcs_valid_cell(&cell);
    return cell ;
}

int32
gcs_extent_to_cell_list(float64 lat_start,
		        float64 lng_start,
		        float64 lat_stop,
		        float64 lng_stop,
		        int32 *n_cell,
		        int32 *cell_list)
                      /* Geodetic Coordinates (-90,+90) */
                      /* Geodetic Coordinates (-180,+180) */
                      /* Geodetic Coordinates (-90,+90) */
                      /* Geodetic Coordinates (-180,+180) */
                   
                        
{
    int32   ncell=0;
    float64 cur_lat,olng_stop,lng_end,lat_end,dummy;
    int32   stradle_idl,stradle_flg;
    int32   last_cell;
    int32   ret;
    
    /* IMPORTANT: The extent defined must be the SW corner (first input pair) 
     * and the NE corner (second pair). No correction can be done if this
     * input is incorrect. 
     */
    if ( !gcs_cell_aware )
    {
	/* if GCS single cell then there is always one cell in the cell list
	 * and that is identified as gcs_default_cell.
	 */
	if ( *n_cell < 1 )
	{
	    /* The allocated list is not large enough to hold
	     * one cell. Inform the caller of the required list
	     * size.
	     */
	    *n_cell = 1;
	    return 0;
	}
	
	*n_cell = 1 ;
	cell_list[0] = gcs_default_cell;
	return 1;
    }
    
    if ( lat_start > lat_stop )
    { /* Simple validation of latitude where start must be less than
       * or equal to stop.
       */
	*n_cell = 0;
	return 1;
    }
    
    if ( lng_stop == -180.0 )
      lng_stop = 180.0;
    
    stradle_flg = stradle_idl = ( lng_stop < lng_start );
    last_cell = gcs_lookup_cell_id_ll(lat_start,lng_start);
    
    if (*n_cell > 0)
      cell_list[ncell++] = last_cell;
    else
      ncell++;
    
    /* get the lat/lng extents for this cell from the reader file */
    if (!gcs_lookup_cell_extents(last_cell,&cur_lat,&lat_end,
				 &dummy,&lng_end))
    {
	printf("Warning:  error looking up lat long for cell id %d\n",
	       last_cell);
	ret = (*n_cell >= ncell);
	*n_cell = ncell;
	return ret;
    }

    while ( 1 )
    { /* loop through all latitudes */
	if ( stradle_idl )
	{
	    olng_stop=lng_stop;
	    lng_stop = 180.0;    
	}
	
	while ( 1 )
	{/* loop through all longitudes */
	    if ( lng_end == -180.0 )
	      lng_end = 180.0;
	    
	    if ( lng_end >= lng_stop )
	    {
		if ( stradle_idl )
		{
		    lng_end     = -180.0;
		    lng_stop    = olng_stop;
		    stradle_idl = 0;
		}
		else
		  break;
	    }
	    
	    last_cell = gcs_lookup_cell_id_ll(cur_lat,lng_end);
	    if (*n_cell > ncell)
	      cell_list[ncell++] = last_cell;
	    else
	      ncell++;
	    if (!gcs_lookup_cell_extents(last_cell,&cur_lat,
					 &lat_end,&dummy,&lng_end))
	    {
		printf("Warning:  error looking up lat long for cell id %d\n",
		       last_cell);
		ret = (*n_cell >= ncell);
		*n_cell = ncell;
		return ret;
	    }
	}
	
	stradle_idl = stradle_flg;
	
	/* move to next cell above */
	cur_lat = lat_end;
	if ( cur_lat >= lat_stop )
	  break;
	
	last_cell = gcs_lookup_cell_id_ll(cur_lat,lng_start);
	if (*n_cell > ncell)
	  cell_list[ncell++] = last_cell;
	else
	  ncell++;
	
	if ( !gcs_lookup_cell_extents(last_cell,&cur_lat,&lat_end,
				      &dummy,&lng_end))
        {
	    printf("Warning:  error looking up lat long for cell id %d\n",
		   last_cell);
	    ret = (*n_cell >= ncell);
	    *n_cell = ncell;
	    return ret;
	}
    }

    ret = (*n_cell >= ncell);    
    *n_cell = ncell;
    return ret;
}    

/* Converts min/max x/y in GCS into lat/long boundaries.  Returns TRUE for
 * success, FALSE for failure.
 */
int32
gcs_ll_bounds_from_gcs(int32 cell,
		       float64 min_x,
		       float64 min_y,
		       float64 max_x,
		       float64 max_y,
		       float64 *west_lng,
		       float64 *south_lat,
		       float64 *east_lng,
		       float64 *north_lat)
{
    float64 lat1, lat2, avg_lat;
    float64 lng1, lng2;
    float64 dummy;

    /* SW corner */
    gcs_gcs_to_llz(cell, min_x, min_y, (float64)0.0, 
		   &lat1, &lng1, &dummy, FALSE);
    *south_lat = lat1;

    /* NW corner */
    gcs_gcs_to_llz(cell, min_x, max_y, (float64)0.0, 
		   &lat2, &lng2, &dummy, FALSE);
    *north_lat = lat2;
    
    /* The way the GCS coordinate systems are set up, the easiest way to
     * determine which points will specify the min and max longitudes is
     * to see whether we're mostly north or south of the equator
     */
    avg_lat = (*north_lat + *south_lat)*0.5;
    if (avg_lat >= 0.0)
    {
	*west_lng = lng2;

	/* SE corner*/
	gcs_gcs_to_llz(cell, max_x, min_y, (float64)0.0, 
		       &lat1, &lng1, &dummy, FALSE);
	*south_lat = NS_MIN(*south_lat, lat1);

	/* NE corner */
	gcs_gcs_to_llz(cell, max_x, max_y, (float64)0.0, 
		       &lat1, &lng1, &dummy, FALSE);
	*east_lng = lng1;
	*north_lat = NS_MAX(*north_lat, lat1);
    }
    else
    {
	*west_lng = lng1;

	/* SE corner*/
	gcs_gcs_to_llz(cell, max_x, min_y, (float64)0.0, 
		       &lat1, &lng1, &dummy, FALSE);
	*east_lng = lng1;
	*south_lat = NS_MIN(*south_lat, lat1);

	/* NE corner */
	gcs_gcs_to_llz(cell, max_x, max_y, (float64)0.0, 
		       &lat1, &lng1, &dummy, FALSE);
	*north_lat = NS_MAX(*north_lat, lat1);
    }

    return TRUE;
}


/* Returns all cells within meters of (x,y,z,cell) in cell_list, assuming
 * that the number of cells is <= *n_cell, which specifies the length of
 * the (caller allocated) cell_list array.  If the number of cells is greater
 * than this, the actual number of cells within the radius is returned
 * in *n_cell, and the return value is 0, otherwise the return value is 1.  
 * Note that this routine is guaranteed to return all cells
 * within the given radius, but may also return some cells that are beyond
 * it.
 */
int32
gcs_cells_within_radius(int32 cell,
		        float64 x,
		        float64 y,
		        float64 z,
		        float64 radius,
		        int32 *n_cell,
		        int32 *cell_list)
{
    /* For now, this just creates a square and finds all cells in it. */
    float64 south_lat, north_lat;
    float64 west_lng, east_lng;
    float64 min_x, max_x, min_y, max_y;

    if (!gcs_cell_aware)
    {
	if (*n_cell < 1)
	{
	    *n_cell = 1;
	    return 0;
	}

	*n_cell = 1;
	cell_list[0] = gcs_default_cell;
	return 1;
    }

    min_x = x - radius;
    max_x = x + radius;
    min_y = y - radius;
    max_y = y + radius;
    
    if (!gcs_ll_bounds_from_gcs(cell, min_x, min_y, max_x, max_y, 
				&west_lng, &south_lat, &east_lng, &north_lat))
    {
	*n_cell = 0;
	return 1;
    }

    return gcs_extent_to_cell_list(south_lat, west_lng, north_lat, east_lng,
				   n_cell, cell_list);
}



int32
gcs_lookup_gravity_vector(int32 cell,
			  float64 x,
			  float64 y,
			  float64 z,
			  float64 *vec)
{
    float64 mag = 0.0;   

    if (gcs_simnet)
    {
	vec[X] = vec[Y] = 0.0;
	vec[Z] = -1.0;
	return TRUE;
    }

    if ( !gcs_gcs3e_to_gcc(cell, x, y, z,&vec[X], &vec[Y], &vec[Z]) )
    {
	vec[X] = vec[Y] = 0.0;
	vec[Z] = -1.0;
	return FALSE;
    }
    
    mag = -1.0/sqrt(vec[X]*vec[X] + vec[Y]*vec[Y] + vec[Z]*vec[Z]);
    /* Flip the vector and normalize it */
    vec[X] *= mag;
    vec[Y] *= mag;
    vec[Z] *= mag;    
    
    return TRUE;
}


float64 
gcs_lookup_ht_asl(int32 cell, float64 x, float64 y, float64 z)
{
    float64 lat, lng, h;
    float64 u2, v2, w2;
    float64 u, v, w;

    if (gcs_simnet)
      return z;

    if (gcs_gcs3e_to_gcc(cell, x, y, z, &u2, &v2, &w2) &&
	gcs_gcs_to_llz(cell, x, y, (float64) 0,
		       &lat, &lng, &h, FALSE))
    {
	gcs_llz_to_gcc(lat, lng, 0.0, FALSE, &u, &v, &w);
	return( sqrt(((u2-u)*(u2-u) + (v2-v)*(v2-v) + (w2-w)*(w2-w))));
    }
    else
      return 0.0;
}

static int32
gcs_ll_to_cell_id(float64 lat, float64 lng)
{
    int32 lng_index,lat_index;

    lat += 90;
    lng += 180;    
#ifdef GCS_VARIABLE_LAT    
    if ( lat < last_lat )
      lat_index = (int32)(lat/(last_lat/last_lat_index));
    else
      lat_index = last_lat_index+1;
#else
    lat_index = lat;
#endif
    
    if ( lng < gcs_latmap[lat_index].last_lng )
      lng_index =(int32)(lng/(gcs_latmap[lat_index].last_lng 	
			      / gcs_latmap[lat_index].n_cell));
    else
      lng_index = gcs_latmap[lat_index].n_cell + 1;
    
    return (int32)(gcs_latmap[lat_index].base_cell + lng_index + GCS_MIN_CELL);
}

int32
gcs_lookup_cell_extents(int32 cell, 
			float64 *lat_sw, 
			float64 *lat_ne, 
			float64 *lng_sw,
			float64 *lng_ne)
{
#ifndef GCS_VARIABLE_LAT    
    static int32 last_lat_index = 180;
#endif
    float32 lat_s,lat_n,lng_e,lng_w;
    int32 i,n_cell;
 
    if ( cell < GCS_MIN_CELL )
      return FALSE;

    /* find where the cell fits into the range of lat/lngs */
    for (i=0;i<last_lat_index;i++)
    {
	if ((cell >= gcs_latmap[i].base_cell) &&
	    (cell < gcs_latmap[i+1].base_cell))
	  break;
    }
    
    n_cell = cell - gcs_latmap[i].base_cell;

    /* cell is in the last region of longitude */
    if ( n_cell > gcs_latmap[i].n_cell )
    {
	lng_w = gcs_latmap[i].last_lng;
	lng_e = 360.0;
	
    }
    else
    {
	lng_w = gcs_latmap[i].last_lng * (n_cell/
				    ((float64)gcs_latmap[i].n_cell-1));
	lng_e = gcs_latmap[i].last_lng * ((n_cell+1)/
					    ((float64)gcs_latmap[i].n_cell-1));

	if ( lng_w > 360 )
	  lng_w = 360;

	if ( lng_e > 360 )
	  lng_e = 360;
    }
    
#ifdef GCS_VARIABLE_LAT    
    /* cell is in the last region of latitude */
    if ( i >= last_lat_index )    
    {
	lat_s = last_lat;
	lat_n = 180.0;
    }
    else
    {
	lat_s = last_lat * (i/(float64)last_lat_index);
	lat_n = last_lat * ((i+1)/(float64)last_lat_index);

	if ( lat_s > 180 )
	  lat_s = 360;

	if ( lat_n > 180 )
	  lat_n = 360;
    }
#else
    lat_n = i+1;
    lat_s = i;
#endif
    /* change the range of lat to -90 to +90 and longitude from
     * -180 to +180.  
     */    
    *lat_sw = lat_s - 90;
    *lat_ne = lat_n - 90;
    *lng_sw = lng_w - 180;
    *lng_ne = lng_e - 180;
    return TRUE;
}

int32
gcs_ll_to_georef_id(float64 s_lat, float64 w_lng, 
		    float64 n_lat, float64 e_lng,
		    int32 *n_georefs , char georefs[][GCS_GEOREF_ID_STRLEN])
{
    float32 sw_lat,sw_lng,ne_lat,ne_lng;
    int32 flag,num,r,c,srow,lrow,scol,lcol;
    static int32 sub_div=15;
    static char letter[24]=
    {  'A','B','C','D','E','F','G','H','J','K','L','M','N','P',
	 'Q','R','S','T','U','V','W','X','Y','Z' };

    if (!gcs_cell_aware)
      return FALSE;

    sw_lat = s_lat + 90  ; sw_lng = w_lng + 180;
    ne_lat = n_lat + 90  ; ne_lng = e_lng + 180;

    /* check if the lat and long are the same, if so then bump up the 
     * east bound 
     */
    ne_lat = ceil(ne_lat);
    ne_lng = ceil(ne_lng);

    if (sw_lat==ne_lat)
      ne_lat++;

    if (sw_lng==ne_lng)
      ne_lng++;


    srow = NS_MIN((int32)sw_lat,180);
    scol = NS_MIN((int32)sw_lng,360);
    lrow = NS_MIN((int32)ne_lat,181);
    lcol = NS_MIN((int32)ne_lng,361);
    
    for (num=0,r=srow;r<lrow;r++)
      for (c=scol;c<lcol;c++,num++)
      {
	  if ( num < *n_georefs )
	    sprintf(georefs[num],"%c%c%c%c\t\0",
		    letter[c/sub_div],
		    letter[r/sub_div],
		    letter[c%sub_div],
		    letter[r%sub_div] );
      }
    
    flag = ( *n_georefs >= num );
    *n_georefs = num;
    return flag;
}



int32
gcs_cell_to_georef_id(int32 cell, int32 *n_georefs , char georefs[][6])
{
    float64 min_lat, max_lat, min_lng, max_lng;
    
    return (gcs_lookup_cell_extents(cell, &min_lat, &max_lat, 
				    &min_lng, &max_lng) &&
	    gcs_ll_to_georef_id(min_lat, min_lng, max_lat, max_lng, 
				n_georefs, georefs));
    
}

void
gcs_generate_latmap( void )
{
    int32   lng2,lat2,llng,n_lng,i;
    int32   cell=GCS_MIN_CELL;
    float64 lngx,lngy,lngz,lngx1,lngy1,lngz1;
    float64 blat,blng,lng_len;

    /* Initialization */
    blat=-90;
    
    for (lat2=0;lat2<180;lat2++,blat++)
    {
	int32 status = 0;

	gcs_latmap[lat2].base_cell = cell;
	blng = -180; 
	/* Note: the length of the latitude lines remain relatively
	 * unchanged from pole to pole (111,692m->110,572m->111,692m).
	 * We will only ever fit one degree latitude in a cell 
	 */

	/* What is the length of the longitude line between
	 * lat and lat+1
	 */
	gcs_llz_to_gcc(blat,blng,0.0,DATUM_WGS84,&lngx,&lngy,&lngz);
	gcs_llz_to_gcc(blat,blng+1,0.0,DATUM_WGS84,&lngx1,&lngy1,&lngz1);
	
        lng_len = sqrt( ((lngx-lngx1)*(lngx-lngx1)) +  
                        ((lngy-lngy1)*(lngy-lngy1)) +
                        ((lngz-lngz1)*(lngz-lngz1)));
	
	gcs_llz_to_gcc(blat+1,blng,0.0,DATUM_WGS84,&lngx,&lngy,&lngz);
	gcs_llz_to_gcc(blat+1.0,blng+1.0,0.0,DATUM_WGS84,&lngx1,&lngy1,&lngz1);
	
	{
	    float64 xx;
	    
            xx =  sqrt( ((lngx-lngx1)*(lngx-lngx1)) +  
                        ((lngy-lngy1)*(lngy-lngy1)) +
                        ((lngz-lngz1)*(lngz-lngz1)));
	    /* save the maximum lng_len */
	    lng_len = NS_MAX(lng_len,xx);
	}
	
	/* calculate the number of longitude lines that you 
	 * will to traverse
	 */
	n_lng = ((float32)GCS_MAX_CELL_SIDE_METERS/lng_len);

	if (n_lng<1)
	  n_lng=1;
	
	for (lng2=0;lng2<360;cell++)
	{
	    llng = blng;
	    blng += n_lng;

	    if (blng>=180)
	    {
		cell++;
		break;
	    }
	}
	
	gcs_latmap[lat2].last_lng = llng+180;
	gcs_latmap[lat2].n_cell   = cell-gcs_latmap[lat2].base_cell;
    }

    gcs_latmap[lat2].base_cell = 
      gcs_latmap[lat2-1].base_cell + gcs_latmap[lat2-1].n_cell;

#ifdef NEVER    
    for (i=0;i<181;i++)
    {
	printf("\t(%7d\t%4d\t%f) ;; latitude %3d\n",
	       gcs_latmap[i].base_cell,
	       (int32)gcs_latmap[i].n_cell,
	       gcs_latmap[i].last_lng,(i-90));
    }
#endif    
}

