#ifndef lint
static char rcsid [] = "$RCSfile: gcs_vec.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
*   File: gcs_vec.c                                                          *
*                                                                            *
*       Copyright 1995 by SAIC & TASC                                        *
*                                                                            *
*       SAIC                          TASC                                   *
*       20 Burlington Mall Road       55 Walkers Brook Drive                 *
*       Suite 130                     Reading, MA 01867                      *
*       Burlington, MA 01803          617-942-2000                           *
*       617-221-7600                                                         *
*                                                                            *
*                                                                            *
*	This software was developed by Loral under U. S. Government contracts*
*	and may be reproduced by or for the U. S. Government pursuant to     *
*	the copyright license under the clause at DFARS 252.227-7013         *
*	(OCT 1988).                                                          *
*                                                                            *
*       Contents: Code to handle vector calculations                         *
*       Created: Tue Aug 29 17:36:21 EDT 1995                                *
*       Author: buettner                                                     *
*       Remarks:                                                             * 
*                                                                            *
*****************************************************************************/

#include "libgcs_local.h"
#include <math.h>

extern int32 gcs_default_cell;

int32
gcs_gcs2_vec_rotate64(
		      float64  from_vec[XYC],
		      int32    to_cell,
		      float64  to_vec[XYC]
		      )
{
    float64 from_matrix[3][3];
    float64 trans_to_matrix[3][3];
    
    int32   from_cell=from_vec[CELL2D];
    float64 gcc_vec[XYZ];
    
    if (!gcs_valid_cell( &from_cell ) || 
	!gcs_valid_cell( &to_cell ) )
    {
	to_vec[X] = to_vec[Y] = 0.0;
	to_vec[CELL2D] = gcs_default_cell;
	return FALSE;
    }

    if (from_cell == to_cell)
    {
	VMAT3_VEC_COPY64(from_vec, to_vec);
	return TRUE;
    }
   
    /* This is libpduapi's version of converting a tcc vector to a gcc vector
     * vmat3_vec_mat_mul32((float32 *)source, pa_tcc_gcc_rotation32,
     *                     (float32 *)sink);
     * where pa_tcc_gcc_rotation is the matrix multiplied to a gcc
     * coordinate to obtain a tcc coordinate.
     * vmat3_vec_mat_mul transposes the matrix before multiplication.
     */

    /* rotate from gcs to gcc */
    gcs_get_rot_mat(from_cell, from_matrix);
    vmat3e_mat_vec_mul64(from_matrix, from_vec[X], from_vec[Y], 0.0, gcc_vec);
 
    /* rotate from gcc to the new gcs */
    gcs_get_rot_mat(to_cell, trans_to_matrix);
    vmat3_vec_mat_mul64(gcc_vec, trans_to_matrix, to_vec);

    /* overwrite the "Z" value with the final cell id */
    to_vec[CELL2D] = to_cell;
    return TRUE;
}

int32
gcs_gcs2e_vec_rotate64(
		       float64  fromx,
		       float64  fromy,
		       int32    fromc,
		       int32    to_cell,
		       float64  to_vec[XYC]
		       )
{
    float64  from_vec[XYC];
    from_vec[X] = fromx;
    from_vec[Y] = fromy;
    from_vec[CELL2D] = fromc;
    return (gcs_gcs2_vec_rotate64(from_vec,to_cell,to_vec));
}
    
int32
gcs_gcs3_vec_rotate64(
		      float64  from_vec[XYZC],
		      int32    to_cell,
		      float64  to_vec[XYZC]
		      )
{
    if ( gcs_vec_to_vec((int32)from_vec[CELL3D],from_vec,to_cell,to_vec) )
    {
	to_vec[CELL3D] = to_cell;
	return TRUE;
    }

    to_vec[CELL3D] = gcs_default_cell;
    return FALSE;
}

int32
gcs_gcs3e_vec_rotate64(
		       float64  fromx,
		       float64  fromy,
		       float64  fromz,
		       int32    fromc,
		       int32    to_cell,
		       float64  to_vec[XYZC]
		       )
{
    float64 from_vec[XYZC];
    from_vec[X]=fromx;
    from_vec[Y]=fromy;
    from_vec[Z]=fromz;
    from_vec[CELL3D]=fromc;
    
    if ( gcs_vec_to_vec((int32)from_vec[CELL3D],from_vec,to_cell,to_vec) )
    {
	to_vec[CELL3D] = to_cell;
	return TRUE;
    }

    to_vec[CELL3D] = gcs_default_cell;
    return FALSE;
}


int32
gcs_gcs2_vec_rotate32(
		      float32  from_vec[XYC],
		      int32    to_cell,
		      float32  to_vec[XYC]
		      )
{
    float64 to_vec64[XYC];
    int32 ret = 
      gcs_gcs2e_vec_rotate64((float64)from_vec[X],(float64)from_vec[Y],
			      (int32)from_vec[CELL2D],to_cell, to_vec64);
    to_vec[X]=(float32)to_vec64[X];
    to_vec[Y]=(float32)to_vec64[Y];
    to_vec[CELL2D]=(float32)to_vec64[CELL2D];
    return ret;
}

int32
gcs_gcs2e_vec_rotate32(
			float32  fromx,
			float32  fromy,
			int32    fromc,
			int32    to_cell,
			float32  to_vec[XYC]
			)
{
    float64 to_vec64[XYC];
    int32 ret = 
      gcs_gcs2e_vec_rotate64((float64)fromx,(float64)fromy,
			      fromc,to_cell, to_vec64);
    to_vec[X]=(float32)to_vec64[X];
    to_vec[Y]=(float32)to_vec64[Y];
    to_vec[CELL2D]=(float32)to_vec64[CELL2D];
    return ret;
}

int32
gcs_gcs3_vec_rotate32(
		      float32  from_vec[XYZC],
		      int32    to_cell,
		      float32  to_vec[XYZC]
		      )
{
    float64 to_vec64[XYZC];
    int32 ret = 
      gcs_gcs3e_vec_rotate64((float64)from_vec[X],(float64)from_vec[Y],
			      (float64)from_vec[Z],
			      (int32)from_vec[CELL3D],to_cell, to_vec64);
    to_vec[X]=(float32)to_vec64[X];
    to_vec[Y]=(float32)to_vec64[Y];
    to_vec[Z]=(float32)to_vec64[Z];
    to_vec[CELL3D]=(float32)to_vec64[CELL3D];
    return ret;
}

int32
gcs_gcs3e_vec_rotate32(
			float32  fromx,
			float32  fromy,
			float32  fromz,
			int32    fromc,
			int32    to_cell,
			float32  to_vec[XYZC]
			)
{
    float64 to_vec64[XYZC];
    int32 ret = 
      gcs_gcs3e_vec_rotate64((float64)fromx, (float64)fromy,(float64)fromz,
			      (int32)fromc,to_cell, to_vec64);
    to_vec[X]=(float32)to_vec64[X];
    to_vec[Y]=(float32)to_vec64[Y];
    to_vec[Z]=(float32)to_vec64[Z];
    to_vec[CELL3D]=(float32)to_vec64[CELL3D];
    return ret;
}



/*=======================GCS VECTOR ADDITION=====================*/
#define TWO_DIM_ADDSUB(type, operator, wx, wy, wc, vx, vy, vc, r)\
{\
    float64 tmp[XYC];\
    if ((int32)wc != (int32)vc)\
    {\
        if (gcs_gcs2e_vec_rotate64((float64)vx,(float64)vy,(int32)vc,\
				    (int32)wc, tmp))\
	{\
	    r[X] = wx operator (type) tmp[X];\
	    r[Y] = wy operator (type) tmp[Y];\
	}\
	else\
	{\
	   /* Error in coordinate system */ \
	    r[X] = r[Y] = 0.0;\
	    r[CELL2D] = gcs_default_cell;\
	    return FALSE;\
	}\
    }\
    else\
    {\
	r[X] = wx operator vx;\
	r[Y] = wy operator vy;\
    }\
    r[CELL2D] = wc;\
    return TRUE;\
}

#define THREE_DIM_ADDSUB(type,operator,wx,wy,wz,wc,vx,vy,vz,vc,r)\
{\
    float64 tmp[XYZC];\
    if ((int32)wc != (int32)vc)\
    {\
        if (gcs_gcs3e_vec_rotate64((float64)vx, (float64)vy, (float64)vz,\
				    (int32)vc,(int32)wc, tmp))\
	{\
	    r[X] = wx operator (type) tmp[X];\
	    r[Y] = wy operator (type) tmp[Y];\
	    r[Z] = wz operator (type) tmp[Z];\
	}\
	else\
	{\
	   /* Error in coordinate system */ \
	    r[X] = r[Y] = r[Z] = 0.0;\
	    r[CELL3D] = gcs_default_cell;\
	    return FALSE;\
	}\
    }\
    else\
    {\
	r[X] = wx operator vx;\
	r[Y] = wy operator vy;\
	r[Z] = wz operator vz;\
    }\
    r[CELL3D] = wc;\
    return TRUE;\
}

int32
gcs_gcs2_vec_add32(
		   float32 w[XYC], 
		   float32 v[XYC], 
		   float32 r[XYC]
		   )
TWO_DIM_ADDSUB(float32,+,w[X],w[Y],w[CELL2D],v[X],v[Y],v[CELL2D],r)

int32
gcs_gcs2e_vec_add32(
		    float32 wx, 
		    float32 wy, 
		    int32   wc,
		    float32 vx, 
		    float32 vy, 
		    int32   vc, 
		    float32 r[XYC]
		    )
TWO_DIM_ADDSUB(float32, +, wx, wy, wc, vx, vy, vc, r)

int32
gcs_gcs3_vec_add32(
		   float32 w[XYZC], 
		   float32 v[XYZC], 
		   float32 r[XYZC]
		   )
THREE_DIM_ADDSUB(float32,+,w[X],w[Y],w[Z],w[CELL3D], 
		 v[X],v[Y],v[Z],v[CELL3D],r)

int32
gcs_gcs3e_vec_add32(
		    float32 wx, 
		    float32 wy, 
		    float32 wz, 
		    int32   wc,
		    float32 vx, 
		    float32 vy, 
		    float32 vz, 
		    int32   vc,
		    float32 r[XYC]
		    )
THREE_DIM_ADDSUB(float32, +, wx, wy, wz, wc, vx, vy, vz, vc, r)

int32
gcs_gcs2_vec_add64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   )
TWO_DIM_ADDSUB(float64,+,w[X],w[Y],w[CELL2D],v[X],v[Y],v[CELL2D],r)

int32
gcs_gcs2e_vec_add64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    )
TWO_DIM_ADDSUB(float64,+,wx,wy,wc,vx,vy,vc,r)


int32
gcs_gcs3_vec_add64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   )
THREE_DIM_ADDSUB(float64,+,w[X],w[Y],w[Z],w[CELL3D], 
		 v[X],v[Y],v[Z],v[CELL3D],r)

int32
gcs_gcs3e_vec_add64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYC]
		    )
THREE_DIM_ADDSUB(float64,+,wx,wy,wz,wc,vx,vy,vz,vc,r)


/*=======================GCS VECTOR SUBTRACTION=====================*/

int32
gcs_gcs2_vec_sub32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   )
TWO_DIM_ADDSUB(float32,-,w[X],w[Y],w[CELL2D], 
		   v[X],v[Y],v[CELL2D],r)
    
int32
gcs_gcs2e_vec_sub32(
		    float32 wx, 
		    float32 wy, 
		    int32   wc, 
		    float32 vx, 
		    float32 vy, 
		    int32   vc, 
		    float32 r[XYC]
		    )
TWO_DIM_ADDSUB(float32, -, wx, wy, wc, vx, vy, vc, r)
    
int32
gcs_gcs3_vec_sub32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   )
THREE_DIM_ADDSUB(float32,-,w[X],w[Y],w[Z],w[CELL3D], 
		 v[X],v[Y],v[Z],v[CELL3D],r)
    
int32
gcs_gcs3e_vec_sub32(
		    float32 wx, 
		    float32 wy, 
		    float32 wz, 
		    int32   wc,
		    float32 vx, 
		    float32 vy, 
		    float32 vz, 
		    int32   vc,
		    float32 r[XYC]
		    )
THREE_DIM_ADDSUB(float32, -, wx, wy, wz, wc, vx, vy, vz, vc, r)

int32
gcs_gcs2_vec_sub64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   )
TWO_DIM_ADDSUB(float64,-,w[X],w[Y],w[CELL2D], 
	       v[X],v[Y],v[CELL2D],r)

int32
gcs_gcs2e_vec_sub64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    )
TWO_DIM_ADDSUB(float64,-,wx,wy,wc,vx,vy,vc,r)


int32
gcs_gcs3_vec_sub64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   )
THREE_DIM_ADDSUB(float64, -, w[X], w[Y], w[Z], w[CELL3D], 
		              v[X], v[Y], v[Z], v[CELL3D], r)

int32
gcs_gcs3e_vec_sub64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYC]
		    )
THREE_DIM_ADDSUB(float64, -, wx, wy, wz, wc, vx, vy, vz, vc, r)

/*=======================GCS UNIT VECTOR =====================*/

#define TWO_DIM_UNIT(type, vx, vy, vc, r)                              \
{                                                                      \
   type mag;                                                           \
   mag = vx*vx + vy*vy;                                                \
   if (mag == 0.0)	                                               \
      r[X] = 1.0, r[Y] = 0.0;                                          \
   else				                                       \
   {				                                       \
       mag = 1.0/sqrt(mag);	                                       \
       r[X] = vx * mag;	                                               \
       r[Y] = vy * mag;	                                               \
   }                                                                   \
   r[CELL2D] = vc;                                                     \
}

#define THREE_DIM_UNIT(type, vx, vy, vz, vc, r)                        \
{                                                                      \
   type mag;                                                           \
   mag = vx*vx + vy*vy + vz*vz;                                        \
   if (mag == 0.0)	                                               \
      r[X] = 1.0, r[Y] = 0.0, r[Z] = 0.0;                              \
   else				                                       \
   {				                                       \
       mag = 1.0/sqrt(mag);	                                       \
       r[X] = vx * mag;	                                               \
       r[Y] = vy * mag;	                                               \
       r[Z] = vz * mag;	                                               \
   }                                                                   \
   r[CELL3D] = vc;                                                     \
}

void gcs_gcs2_unit32(
		      float32 v[XYC], 
		      float32 r[XYC]
		      )
TWO_DIM_UNIT(float32, v[X], v[Y], v[CELL2D], r)

void gcs_gcs2e_unit32(
		       float32 vx, 
		       float32 vy, 
		       int32   vc, 
		       float32 r[XYC])
TWO_DIM_UNIT(float32, vx, vy, vc, r)

void gcs_gcs2_unit64(
		     float64 v[XYC],
		     float64 r[XYC]
		     )
TWO_DIM_UNIT(float64, v[X], v[Y], v[CELL2D], r)

void gcs_gcs2e_unit64(
		      float64 vx,
		      float64 vy,
		      int32   vc,
		      float64 r[XYC]
		      )
TWO_DIM_UNIT(float64, vx, vy, vc, r)

void gcs_gcs3_unit32(
		      float32 v[XYZC],
		      float32 r[XYZC]
		      )
THREE_DIM_UNIT(float32, v[X], v[Y], v[Z], v[CELL3D], r)

void gcs_gcs3e_unit32(
		       float32 vx,
		       float32 vy,
		       float32 vz,
		       int32   vc,
		       float32 r[XYZC]
		       )
THREE_DIM_UNIT(float32, vx, vy, vz, vc, r)

void gcs_gcs3_unit64(
		     float64 v[XYZC],
		     float64 r[XYZC]
		     )
THREE_DIM_UNIT(float64, v[X], v[Y], v[Z], v[CELL3D], r)

void gcs_gcs3e_unit64(
		      float64 vx,
		      float64 vy,
		      float64 vz,
		      int32   vc,
		      float64 r[XYZC]
		      )
THREE_DIM_UNIT(float64, vx, vy, vz, vc, r)

/*=======================GCS NEGATE VECTOR =====================*/
#define TWO_DIM_NEGATE(type, vx, vy, vc, r)\
{\
    r[X] = -vx;\
    r[Y] = -vy;\
    r[CELL2D] = vc;\
}

#define THREE_DIM_NEGATE(type, vx, vy, vz, vc, r)\
{\
    r[X] = -vx;\
    r[Y] = -vy;\
    r[Z] = -vz;\
    r[CELL3D] = vc;\
}

void
gcs_gcs2_negate32(
		  float32 v[XYC],
		  float32 r[XYC]
		  )
TWO_DIM_NEGATE(float32, v[X], v[Y], v[CELL2D], r)
    
void
gcs_gcs2e_negate32(
		   float32 vx,
		   float32 vy,
		   int32   vc,
		   float32 r[XYC]
		   )
TWO_DIM_NEGATE(float32, vx, vy, vc, r)

void
gcs_gcs3_negate32(
		  float32 v[XYZC],
		  float32 r[XYZC]
		  )
THREE_DIM_NEGATE(float32, v[X], v[Y], v[Z], v[CELL3D], r)

void
gcs_gcs3e_negate32(
		   float32 vx,
		   float32 vy,
		   float32 vz,
		   int32   vc,
		   float32 r[XYZC]
		   )
THREE_DIM_NEGATE(float32, vx, vy, vz, vc, r)
 
void
gcs_gcs2_negate64(
		  float64 v[XYC],
		  float64 r[XYC]
		  )
TWO_DIM_NEGATE(float64, v[X], v[Y], v[CELL2D], r)

void
gcs_gcs2e_negate64(
		   float64 vx,
		   float64 vy,
		   int32   vc,
		   float64 r[XYC]
		   )
TWO_DIM_NEGATE(float64, vx, vy, vc, r)

void
gcs_gcs3_negate64(
		  float64 v[XYZC],
		  float64 r[XYZC]
		  )
THREE_DIM_NEGATE(float64, v[X], v[Y], v[Z], v[CELL3D], r)

void
gcs_gcs3e_negate64(
		   float64 vx,
		   float64 vy,
		   float64 vz,
		   int32   vc,
		   float64 r[XYZC]
		   )
THREE_DIM_NEGATE(float64, vx, vy, vz, vc, r)


/* Generic macros: 
 * TWO_DIMENSION_WITH_RETURN and THREE_DIMENSION_WITH_RETURN apply
 * some generic function func to some 2D/3D vector (wx,wy,[wz],wc) and
 * (vx,vy,[vz],vc), and returns that value.
 */
#define TWO_DIMENSION_WITH_RETURN(type, func, wx, wy, wc, vx, vy, vc)\
{\
    float64 tmp[XYC];\
    if ((int32)wc != (int32)vc)\
    {\
        if (!gcs_gcs2e_vec_rotate64((float64)vx,(float64)vy,(int32)vc,\
				    (int32)wc, tmp))\
          return 0.0;\
	return func(wx, wy, (type)tmp[X], (type)tmp[Y]);\
    }\
    else\
    {\
        return func(wx, wy, vx, vy); \
    }\
}

#define THREE_DIMENSION_WITH_RETURN(type,func,wx,wy,wz,wc,vx,vy,vz,vc)\
{\
    float64 tmp[XYZC];\
    if ((int32)wc != (int32)vc)\
    {\
        if (!gcs_gcs3e_vec_rotate64((float64)vx,(float64)vy,(float64)vz,\
				(int32)vc,(int32)wc, tmp))\
          return 0.0;\
	return func(wx,wy,wz,(type)tmp[X],(type)tmp[Y],(type)tmp[Z]); \
    }\
    else\
    {\
	return func(wx, wy, wz, vx, vy, vz); \
    }\
}

/* Generic macros: 
 * TWO_DIMENSION_WITHOUT_RETURN and THREE_DIMENSION_WITHOUT_RETURN apply
 * some generic function func to some 2D/3D GCS vector (wx,wy,[wz],wc) and
 * (vx,vy,[vz],vc), and places the result into the 2D/3D GCS vector r.
 */

#define TWO_DIMENSION_WITHOUT_RETURN(type, func, wx, wy, wc, vx, vy, vc, r)\
{\
    float64 tmp[XYC];\
    if ((int32)wc != (int32)vc)\
    {\
        if (!gcs_gcs2e_vec_rotate64((float64)vx,(float64) vy,(int32)vc,\
			            (int32)wc, tmp))\
          return FALSE;\
	func(wx, wy, (type)tmp[X], (type)tmp[Y],r);\
    }\
    else\
    {\
        func(wx, wy, vx, vy,r); \
    }\
    r[CELL2D] = wc;\
    return TRUE;\
}

#define THREE_DIMENSION_WITHOUT_RETURN(type,func,wx,wy,wz,wc,vx,vy,vz,vc,r)\
{\
    float64 tmp[XYZC];\
    if ((int32)wc != (int32)vc)\
    {\
        if (!gcs_gcs3e_vec_rotate64((float64)vx,(float64)vy,(float64)vz,\
				(int32)vc,(int32)wc, tmp))\
          return FALSE;\
	func(wx, wy, wz, (type)tmp[X], (type)tmp[Y], (type)tmp[Z],r); \
    }\
    else\
    {\
	func(wx, wy, wz, vx, vy, vz,r); \
    }\
    r[CELL3D] = wc;\
    return TRUE;\
}



/*=======================GCS VECTOR DOT PRODUCT=====================*/

float32
gcs_gcs2_dot_prod32(
		    float32  w[XYC],
		    float32  v[XYC]
		    )
TWO_DIMENSION_WITH_RETURN(float32, vmat2e_dot_prod32, w[X], w[Y], w[CELL2D],
			  v[X], v[Y], v[CELL2D])

float32
gcs_gcs2e_dot_prod32(
		     float32  wx,
		     float32  wy,
		     int32    wc,
		     float32  vx,
		     float32  vy,
		     int32    vc
		     )
TWO_DIMENSION_WITH_RETURN(float32, vmat2e_dot_prod32, wx, wy, wc,vx, vy, vc)
    
float32
gcs_gcs3_dot_prod32(
		    float32  w[XYZC],
		    float32  v[XYZC]
		    )
THREE_DIMENSION_WITH_RETURN(float32,vmat3e_dot_prod32,w[X],w[Y],w[Z],w[CELL3D],
			    v[X],v[Y],v[Z],v[CELL3D])

float32
gcs_gcs3e_dot_prod32(
		     float32  wx,
		     float32  wy,
		     float32  wz,
		     int32    wc,
		     float32  vx,
		     float32  vy,
		     float32  vz,
		     int32    vc
		     )
THREE_DIMENSION_WITH_RETURN(float32, vmat3e_dot_prod32, wx, wy, wz, wc,
			    vx, vy, vz, vc)

float64
gcs_gcs2_dot_prod64(
		    float64  w[XYC],
		    float64  v[XYC]
		    )
TWO_DIMENSION_WITH_RETURN(float64, vmat2e_dot_prod64, w[X], w[Y], w[CELL2D],
			  v[X], v[Y], v[CELL2D])

float64
gcs_gcs2e_dot_prod64(
		     float64  wx,
		     float64  wy,
		     int32    wc,
		     float64  vx,
		     float64  vy,
		     int32    vc
		     )
TWO_DIMENSION_WITH_RETURN(float64, vmat2e_dot_prod64, wx, wy, wc,vx, vy, vc)

float64
gcs_gcs3_dot_prod64(
		    float64  w[XYZC],
		    float64  v[XYZC]
		    )
THREE_DIMENSION_WITH_RETURN(float64,vmat3e_dot_prod64,w[X],w[Y],w[Z],w[CELL3D],
			    v[X],v[Y],v[Z],v[CELL3D])

float64
gcs_gcs3e_dot_prod64(
		     float64  wx,
		     float64  wy,
		     float64  wz,
		     int32    wc,
		     float64  vx,
		     float64  vy,
		     float64  vz,
		     int32    vc
		     )
THREE_DIMENSION_WITH_RETURN(float64, vmat3e_dot_prod64, wx, wy, wz, wc,
			  vx, vy, vz, vc)

/*=======================GCS VECTOR CROSS PRODUCT=====================*/
float32
gcs_gcs2_cross_prod32(
		      float32  w[XYC],
		      float32  v[XYC]
		      )
TWO_DIMENSION_WITH_RETURN(float32,vmat2e_cross_prod32, w[X],w[Y],w[CELL2D],
			  v[X],v[Y],v[CELL2D])

float32
gcs_gcs2e_cross_prod32(
		       float32  wx,
		       float32  wy,
		       int32    wc,
		       float32  vx,
		       float32  vy,
		       int32    vc
		       )
TWO_DIMENSION_WITH_RETURN(float32,vmat2e_cross_prod32,wx,wy,wc,vx,vy,vc)
    
int32
gcs_gcs3_cross_prod32(
		      float32  w[XYZC],
		      float32  v[XYZC],
		      float32  r[XYZC]
		      )
THREE_DIMENSION_WITHOUT_RETURN(float32, vmat3e_cross_prod32,w[X],w[Y],w[Z], w[CELL3D], v[X],v[Y],v[Z],v[CELL3D], r)

int32
gcs_gcs3e_cross_prod32(
		       float32  wx,
		       float32  wy,
		       float32  wz,
		       int32    wc,
		       float32  vx,
		       float32  vy,
		       float32  vz,
		       int32    vc,
		       float32  r[XYZC]
		       )
THREE_DIMENSION_WITHOUT_RETURN(float32, vmat3e_cross_prod32,wx,wy,wz,wc, 
			       vx,vy,vz,vc,r)

float64
gcs_gcs2_cross_prod64(
		      float64  w[XYC],
		      float64  v[XYC]
		      )
TWO_DIMENSION_WITH_RETURN(float64,vmat2e_cross_prod64, w[X],w[Y],w[CELL2D],
			  v[X],v[Y],v[CELL2D])

float64
gcs_gcs2e_cross_prod64(
		       float64  wx,
		       float64  wy,
		       int32    wc,
		       float64  vx,
		       float64  vy,
		       int32    vc
		       )
TWO_DIMENSION_WITH_RETURN(float64,vmat2e_cross_prod64,wx,wy,wc,vx,vy,vc)

int32
gcs_gcs3_cross_prod64(
		      float64  w[XYZC],
		      float64  v[XYZC],
		      float64  r[XYZC]
		      )
THREE_DIMENSION_WITHOUT_RETURN(float64, vmat3e_cross_prod64,w[X],w[Y],w[Z],
			       w[CELL3D], v[X],v[Y],v[Z],v[CELL3D], r)

int32
gcs_gcs3e_cross_prod64(
		       float64  wx,
		       float64  wy,
		       float64  wz,
		       int32    wc,
		       float64  vx,
		       float64  vy,
		       float64  vz,
		       int32    vc,
		       float64  r[XYZC]
		       )
THREE_DIMENSION_WITHOUT_RETURN(float64, vmat3e_cross_prod64,wx,wy,wz,wc, 
			       vx,vy,vz,vc,r)


/*=======================GCS VECTOR SCALE MULT=====================*/
void 
gcs_gcs2_scal_vec_mul32(
			float32 s,
			float32 v[XYC],
			float32 r[2]
			)
{
    VMAT2E_SCAL_VEC_MUL(s, v[X], v[Y], r);
    r[CELL2D] = v[CELL2D];
}

void 
gcs_gcs2e_scal_vec_mul32(
			 float32 s,
			 float32 vx,
			 float32 vy,
			 int32   vc,
			 float32 r[2]
			 )
{
    VMAT2E_SCAL_VEC_MUL(s, vx, vy, r);
    r[CELL2D] = vc;

}

void 
gcs_gcs3_scal_vec_mul32(
			float32 s,
			float32 v[XYZC],
			float32 r[2]
			)
{
    VMAT3E_SCAL_VEC_MUL(s, v[X], v[Y], v[Z], r);
    r[CELL3D] = v[CELL3D];
}

void 
gcs_gcs3e_scal_vec_mul32(
			 float32 s,
			 float32 vx,
			 float32 vy,
			 float32 vz,
			 int32   vc,
			 float32 r[2]
			 )
{
    VMAT3E_SCAL_VEC_MUL(s, vx, vy, vz, r);
    r[CELL3D] = vc;
}

void 
gcs_gcs2_scal_vec_mul64(
			float64 s,
			float64 v[XYC],
			float64 r[2]
			)
{
    VMAT2E_SCAL_VEC_MUL(s, v[X], v[Y], r);
    r[CELL2D] = v[CELL2D];
}

void 
gcs_gcs2e_scal_vec_mul64(
			 float64 s,
			 float64 vx,
			 float64 vy,
			 int32   vc,
			 float64 r[2]
			 )
{
    VMAT2E_SCAL_VEC_MUL(s, vx, vy, r);
    r[CELL2D] = vc;

}

void 
gcs_gcs3_scal_vec_mul64(
			float64 s,
			float64 v[XYZC],
			float64 r[2]
			)
{
    VMAT3E_SCAL_VEC_MUL(s, v[X], v[Y], v[Z], r);
    r[CELL3D] = v[CELL3D];
}

void 
gcs_gcs3e_scal_vec_mul64(
			 float64 s,
			 float64 vx,
			 float64 vy,
			 float64 vz,
			 int32   vc,
			 float64 r[2]
			 )
{
    VMAT3E_SCAL_VEC_MUL(s, vx, vy, vz, r);
    r[CELL3D] = vc;
}

/*=================GCS VECTOR ANGLE BETWEEN VECTORS==================*/

float32
gcs_gcs2_angle_between_vectors32(
				 float32  w[XYC],
				 float32  v[XYC]
				 )
TWO_DIMENSION_WITH_RETURN(float32, vmat2e_angle_between_vectors32, 
			  w[X],w[Y],w[CELL2D],v[X],v[Y],v[CELL2D])

float32
gcs_gcs2e_angle_between_vectors32(
				  float32  wx,
				  float32  wy,
				  int32    wc,
				  float32  vx,
				  float32  vy,
				  int32    vc
				  )
TWO_DIMENSION_WITH_RETURN(float32, vmat2e_angle_between_vectors32, 
			  wx, wy, wc, vx, vy, vc)

float32
gcs_gcs3_angle_between_vectors32(
				 float32  w[XYZC],
				 float32  v[XYZC]
				  )
THREE_DIMENSION_WITH_RETURN(float32, vmat3e_angle_between_vectors32, 
			    w[X],w[Y],w[Z],w[CELL3D],v[X],v[Y],v[Z],v[CELL3D])

float32
gcs_gcs3e_angle_between_vectors32(
				  float32  wx,
				  float32  wy,
				  float32  wz,
				  int32    wc,
				  float32  vx,
				  float32  vy,
				  float32  vz,
				  int32    vc
				  )
THREE_DIMENSION_WITH_RETURN(float32, vmat3e_angle_between_vectors32, 
			    wx, wy, wz, wc, vx, vy, vz, vc)

float64
gcs_gcs2_angle_between_vectors64(
				 float64  w[XYC],
				 float64  v[XYC]
				 )
TWO_DIMENSION_WITH_RETURN(float64, vmat2e_angle_between_vectors64, 
			  w[X],w[Y],w[CELL2D],v[X],v[Y],v[CELL2D])

float64
gcs_gcs2e_angle_between_vectors64(
				  float64  wx,
				  float64  wy,
				  int32    wc,
				  float64  vx,
				  float64  vy,
				  int32    vc
				  )
TWO_DIMENSION_WITH_RETURN(float64, vmat2e_angle_between_vectors64, 
			  wx, wy, wc, vx, vy, vc)

float64
gcs_gcs3_angle_between_vectors64(
				 float64  w[XYZC],
				 float64  v[XYZC]
				 )
THREE_DIMENSION_WITH_RETURN(float64, vmat3e_angle_between_vectors64, 
			    w[X],w[Y],w[Z],w[CELL3D],v[X],v[Y],v[Z],v[CELL3D])

float64
gcs_gcs3e_angle_between_vectors64(
				  float64  wx,
				  float64  wy,
				  float64  wz,
				  int32    wc,
				  float64  vx,
				  float64  vy,
				  float64  vz,
				  int32    vc
				  )
THREE_DIMENSION_WITH_RETURN(float64, vmat3e_angle_between_vectors64, 
			    wx, wy, wz, wc, vx, vy, vz, vc)


/*=====================GCS VECTOR PROJECT PARALLEL======================*/

int32
gcs_gcs2_project32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   )
TWO_DIMENSION_WITHOUT_RETURN(float32, vmat2e_project32, 
			     w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D], r)

int32
gcs_gcs2e_project32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC])
TWO_DIMENSION_WITHOUT_RETURN(float32, vmat2e_project32, 
			     wx,wy,wc,vx, vy, vc, r)

int32
gcs_gcs3_project32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   )
THREE_DIMENSION_WITHOUT_RETURN(float32, vmat3e_project32, 
			       w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D], r)
int32
gcs_gcs3e_project32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYZC]
		    )
THREE_DIMENSION_WITHOUT_RETURN(float32, vmat3e_project32, 
			       wx,wy,wz,wc,vx,vy,vz,vc,r)

int32
gcs_gcs2_project64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   )
TWO_DIMENSION_WITHOUT_RETURN(float64, vmat2e_project64, 
			     w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D], r)

int32
gcs_gcs2e_project64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    )
TWO_DIMENSION_WITHOUT_RETURN(float64, vmat2e_project64, 
			     wx,wy,wc,vx, vy, vc, r)

int32
gcs_gcs3_project64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   )
THREE_DIMENSION_WITHOUT_RETURN(float64, vmat3e_project64, 
			       w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D], r)
int32
gcs_gcs3e_project64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYZC]
		    )
THREE_DIMENSION_WITHOUT_RETURN(float64, vmat3e_project64, 
			       wx,wy,wz,wc,vx,vy,vz,vc,r)

/*=====================GCS VECTOR PROJECT PERPENDICULAR======================*/
int32
gcs_gcs2_project_perp32(
			float32 w[XYC],
			float32 v[XYC],
			float32 r[XYC]
			)
TWO_DIMENSION_WITHOUT_RETURN(float32, vmat2e_project_perp32, 
			     w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D], r)
int32
gcs_gcs2e_project_perp32(
			 float32 wx,
			 float32 wy,
			 int32   wc,
			 float32 vx,
			 float32 vy,
			 int32   vc,
			 float32 r[XYC]
			 )
TWO_DIMENSION_WITHOUT_RETURN(float32, vmat2e_project_perp32, 
			     wx,wy,wc,vx, vy, vc, r)

int32
gcs_gcs3_project_perp32(
			float32 w[XYZC],
			float32 v[XYZC],
			float32 r[XYZC]
			)
THREE_DIMENSION_WITHOUT_RETURN(float32, vmat3e_project_perp32, 
			       w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D], r)
int32
gcs_gcs3e_project_perp32(
			 float32 wx,
			 float32 wy,
			 float32 wz,
			 int32   wc,
			 float32 vx,
			 float32 vy,
			 float32 vz,
			 int32   vc,
			 float32 r[XYZC]
			 )
THREE_DIMENSION_WITHOUT_RETURN(float32, vmat3e_project_perp32, 
			       wx,wy,wz,wc,vx,vy,vz,vc,r)

int32
gcs_gcs2_project_perp64(
			float64 w[XYC],
			float64 v[XYC],
			float64 r[XYC]
			)
TWO_DIMENSION_WITHOUT_RETURN(float64, vmat2e_project_perp64, 
			     w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D], r)
int32
gcs_gcs2e_project_perp64(
			 float64 wx,
			 float64 wy,
			 int32   wc,
			 float64 vx,
			 float64 vy,
			 int32   vc,
			 float64 r[XYC]
			 )
TWO_DIMENSION_WITHOUT_RETURN(float64, vmat2e_project_perp64, 
			     wx,wy,wc,vx, vy, vc, r)

int32
gcs_gcs3_project_perp64(
			float64 w[XYZC],
			float64 v[XYZC],
			float64 r[XYZC]
			)
THREE_DIMENSION_WITHOUT_RETURN(float64, vmat3e_project_perp64, 
			       w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D], r)
int32
gcs_gcs3e_project_perp64(
			 float64 wx,
			 float64 wy,
			 float64 wz,
			 int32   wc,
			 float64 vx,
			 float64 vy,
			 float64 vz,
			 int32   vc,
			 float64 r[XYZC]
			 )
THREE_DIMENSION_WITHOUT_RETURN(float64, vmat3e_project_perp64, 
			       wx,wy,wz,wc,vx,vy,vz,vc,r)



/*==============GCS VECTOR OFFSET POINT BY VECTOR====================*/
int32 
gcs_gcs3_pos_vec_add64(
		       float64 pt0[XYZC],/* original point */
		       float64 offset[XYZC],/* offset vector */
		       float64 npt[XYZC] /* new point still in cell of point */
		       )
{
    if ( offset[CELL3D] != pt0[CELL3D] )
    {
	if( !gcs_gcs3_vec_rotate64(offset, (int32)pt0[CELL3D], offset) )
	  return FALSE;
    }
    
    npt[X]      = pt0[X] + offset[X] ;
    npt[Y]      = pt0[Y] + offset[Y] ;
    npt[Z]      = pt0[Z] + offset[Z] ;
    npt[CELL3D] = pt0[CELL3D];
    return TRUE;
}

int32 
gcs_gcs3e_pos_vec_add64(
			float64 ptx,  /* original point */
			float64 pty,
			float64 ptz,
			int32   ptc,
			float64 offx, /* offset vector */
			float64 offy,
			float64 offz,
			int32   offc,
			float64 npt[XYZC] /* new point still in cell of point*/
			)
{
    float64 offset[XYZC];
    float64 pt0[XYZC];
    offset[X]=offx;
    offset[Y]=offy;
    offset[Z]=offz;
    offset[CELL3D]=offc;
    pt0[X]=ptx;
    pt0[Y]=pty;
    pt0[Z]=ptz;
    pt0[CELL3D]=ptc;
    return gcs_gcs3_pos_vec_add64(pt0,offset,npt);
}

int32 
gcs_gcs3_pos_vec_add32(
		       float64 pt0[XYZC],    /* original point */
		       float64 offset[XYZC], /* offset vector */
		       float32 npt[XYZC]     /* new point still in cell 
						of point */
		       )
{
    float64 offset64[XYZC];
    float64 pt64[XYZC];
    int32 ret;
    float64 npt64[XYZC];

    offset64[X]=offset[X];
    offset64[Y]=offset[Y];
    offset64[Z]=offset[Z];
    offset64[CELL3D]=offset[CELL3D];
    pt64[X]=pt0[X];
    pt64[Y]=pt0[Y];
    pt64[Z]=pt0[Z];
    pt64[CELL3D]=pt0[CELL3D];
    ret =  gcs_gcs3_pos_vec_add64(pt64,offset64,npt64);
    npt[X] = (float32)npt64[X];
    npt[Y] = (float32)npt64[Y];
    npt[Z] = (float32)npt64[Z];
    npt[CELL3D] = (float32)npt64[CELL3D];
    return ret;
}

int32 
gcs_gcs3e_pos_vec_add32(
			float32 ptx,
			float32 pty,
			float32 ptz,
			int32   ptc,
			float32 offx,
			float32 offy,
			float32 offz,
			int32   offc,
			float32 npt[XYZC]/* new point still in cell of point */
			)
{
    float64 offset64[XYZC];
    float64 pt64[XYZC];
    int32 ret;
    float64 npt64[XYZC];

    offset64[X]=offx;
    offset64[Y]=offy;
    offset64[Z]=offz;
    offset64[CELL3D]=offc;
    pt64[X]=ptx;
    pt64[Y]=pty;
    pt64[Z]=ptz;
    pt64[CELL3D]=ptc;
    ret =  gcs_gcs3_pos_vec_add64(pt64,offset64,npt64);
    npt[X] = (float32)npt64[X];
    npt[Y] = (float32)npt64[Y];
    npt[Z] = (float32)npt64[Z];
    npt[CELL3D] = (float32)npt64[CELL3D];
    return ret;
}

/* Note: This processing will break when the offset is very large because
 * the approximation that the x-y plane is the surface of the earth becomes 
 * less and less accurate as we move further and further away from the point's
 * original cell.
 */
int32 
gcs_gcs2_pos_vec_add64(
		       float64 pt0[XYC],    /* original point */
		       float64 offset[XYC], /* offset vector */
		       float64 npt[XYC]     /* new point still in cell of 
					       point */
		       )
{
    if ( offset[CELL2D] != pt0[CELL2D] )
    {
	if( !gcs_gcs2_vec_rotate64(offset, (int32)pt0[CELL2D], offset) )
	  return FALSE;
    }
    npt[X]      = pt0[X] + offset[X] ;
    npt[Y]      = pt0[Y] + offset[Y] ;
    npt[CELL2D] = pt0[CELL2D];
    return TRUE;
}

int32 
gcs_gcs2e_pos_vec_add64(
			float64 ptx,  /* original point */
			float64 pty,
			int32   ptc,
			float64 offx, /* offset vector */
			float64 offy,
			int32   offc,
			float64 npt[XYC] /* new point still in cell of point */
			)
{
    float64 offset[XYC];
    float64 pt0[XYC];
    offset[X]=offx;
    offset[Y]=offy;
    offset[CELL2D]=offc;
    pt0[X]=ptx;
    pt0[Y]=pty;
    pt0[CELL2D]=ptc;
    return gcs_gcs2_pos_vec_add64(pt0,offset,npt);
}

int32 
gcs_gcs2_pos_vec_add32(
		       float64 pt0[XYC],    /* original point */
		       float64 offset[XYC], /* offset vector */
		       float32 npt[XYC]     /* new point still in 
					       cell of point */
		       )
{
    float64 offset64[XYC];
    float64 pt64[XYC];
    int32 ret;
    float64 npt64[XYC];

    offset64[X]=offset[X];
    offset64[Y]=offset[Y];
    offset64[CELL2D]=offset[CELL2D];
    pt64[X]=pt0[X];
    pt64[Y]=pt0[Y];
    pt64[CELL2D]=pt0[CELL2D];
    ret =  gcs_gcs2_pos_vec_add64(pt64,offset64,npt64);
    npt[X] = (float32)npt64[X];
    npt[Y] = (float32)npt64[Y];
    npt[CELL2D] = (float32)npt64[CELL2D];
    return ret;
}

int32 
gcs_gcs2e_pos_vec_add32(
			float32 ptx,
			float32 pty,
			int32   ptc,
			float32 offx,
			float32 offy,
			int32   offc,
			float32 npt[XYC] /* new point still in cell of point */
			)    
{
    float64 offset64[XYC];
    float64 pt64[XYC];
    int32 ret;
    float64 npt64[XYC];

    offset64[X]=offx;
    offset64[Y]=offy;
    offset64[CELL2D]=offc;
    pt64[X]=ptx;
    pt64[Y]=pty;
    pt64[CELL2D]=ptc;
    ret =  gcs_gcs3_pos_vec_add64(pt64,offset64,npt64);
    npt[X] = (float32)npt64[X];
    npt[Y] = (float32)npt64[Y];
    npt[CELL2D] = (float32)npt64[CELL2D];
    return ret;
}



/* Note: This processing will break when the offset is very large because
 * the approximation that the x-y plane is the surface of the earth becomes 
 * less and less accurate as we move further and further away from the points
 * original cell.
 */
int32 
gcs_gcs2_pos_vec_addpl(
		       PointLocation *pt0,         /* original point */
		       float64        offset[XYC],    /* offset vector */
		       PointLocation *npt          /* new point still in 
							 cell of point */
		       )
{
    if ( (int32)offset[CELL2D] != pt0->cell )
    {
	if( !gcs_gcs2_vec_rotate64(offset,(int32)pt0->cell,offset) )
	  return FALSE;
    }
    npt->x    = pt0->x + offset[X] ;
    npt->y    = pt0->y + offset[Y] ;
    npt->cell = pt0->cell;
    return TRUE;
}


/*====================GCS POSITION ADDITION======================*/
#define TWO_DIM_POSITION_WITHOUT_RETURN(type,operator,wx,wy,wc,vx,vy,vc,r)\
{\
   if ((int32)wc != (int32) vc)\
   {\
       float64 tmpx, tmpy;\
       if (gcs_gcs2e_to_gcscs((int32)vc,(float64)vx,(float64)vy,\
				(int32)wc,&tmpx,&tmpy))\
       {\
	   r[X] = wx operator tmpx;\
	   r[Y] = wy operator tmpy;\
       }\
       else\
       {\
	   r[X] = r[Y] = 0.0;\
	   r[CELL2D] = (type)gcs_default_cell;\
	   return FALSE;\
       }\
   }\
   else\
   {\
       r[X] = wx operator vx;\
       r[Y] = wy operator vy;\
   }\
   r[CELL2D] = wc;\
   return TRUE;\
}

#define THREE_DIM_POSITION_WITHOUT_RETURN(type,operator,wx,wy,wz,wc,vx,vy,vz,vc,r)\
{\
   if ((int32)wc != (int32) vc)\
   {\
       float64 tmpx, tmpy, tmpz;\
       if (gcs_gcs3e_to_gcscs((int32)vc,(float64)vx,(float64)vy,(float64)vz,\
			    (int32)wc,&tmpx,&tmpy,&tmpz))\
       {\
	   r[X] = wx operator tmpx;\
	   r[Y] = wy operator tmpy;\
	   r[Z] = wz operator tmpz;\
       }\
       else\
       {\
	   r[X] = r[Y] = r[Z] = 0.0;\
	   r[CELL3D] = (type)gcs_default_cell;\
	   return FALSE;\
       }\
   }\
   else\
   {\
       r[X] = wx operator vx;\
       r[Y] = wy operator vy;\
       r[Z] = wz operator vz;\
   }\
   r[CELL3D] = wc;\
   return TRUE;\
}

int32
gcs_gcs2_pos_addpl(
		   PointLocation *w,
		   PointLocation *v,
		   PointLocation *r
		   )
{
    if ( w->cell != v->cell )
    {
	PointLocation tmp;
	
	if ( gcs_gcs2_to_gcscspl(v, w->cell, &tmp) )
	{
	    r->x = w->x + tmp.x;
	    r->y = w->y + tmp.y;
	}
	else
	{
	    /* Invalid Coordinate conversion */
	    r->x = r->y = 0.0;
	    r->cell = gcs_default_cell;
	    return FALSE;
	}
    }
    else
    {
	r->x = w->x + v->x ;
	r->y = w->y + v->y ;
    }
     
    r->cell = w->cell;
    return TRUE;    
}


int32
gcs_gcs2_pos_add32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   )
TWO_DIM_POSITION_WITHOUT_RETURN(float32,+,w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D],r)

int32
gcs_gcs2e_pos_add32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    )
TWO_DIM_POSITION_WITHOUT_RETURN(float32,+,wx,wy,wc,vx,vy,vc,r)

int32
gcs_gcs2_pos_add64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   )
TWO_DIM_POSITION_WITHOUT_RETURN(float64,+,w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D],r)


int32
gcs_gcs2e_pos_add64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    )
TWO_DIM_POSITION_WITHOUT_RETURN(float64,+,wx,wy,wc,vx,vy,vc,r)

int32
gcs_gcs3_pos_add32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   )
THREE_DIM_POSITION_WITHOUT_RETURN(float32,+,w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D],r)
    
int32
gcs_gcs3e_pos_add32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYZC]
		    )
THREE_DIM_POSITION_WITHOUT_RETURN(float32,+,wx,wy,wz,wc,vx,vy,vz,vc,r)


int32
gcs_gcs3_pos_add64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   )
THREE_DIM_POSITION_WITHOUT_RETURN(float64,+,w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D],r)

int32
gcs_gcs3e_pos_add64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYZC]
		    )
THREE_DIM_POSITION_WITHOUT_RETURN(float64,+,wx,wy,wz,wc,vx,vy,vz,vc,r)

/*====================GCS POSITION SUBTRACTION======================*/
int32
gcs_gcs2_pos_subpl(
		   PointLocation *w,
		   PointLocation *v,
		   PointLocation *r
		   )
{
    if ( w->cell != v->cell )
    {
	PointLocation tmp;
	
	if ( gcs_gcs2_to_gcscspl(v, w->cell, &tmp) )
	{
	    r->x = w->x - tmp.x;
	    r->y = w->y - tmp.y;
	}
	else
	{
	    /* Invalid Coordinate conversion */
	    r->x = r->y = 0.0;
	    r->cell = gcs_default_cell;
	    return FALSE;
	}
    }
    else
    {
	r->x = w->x - v->x ;
	r->y = w->y - v->y ;
    }
     
    r->cell = w->cell;
    return TRUE;    
}

int32
gcs_gcs2_pos_sub32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   )
TWO_DIM_POSITION_WITHOUT_RETURN(float32,-,w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D],r)

int32
gcs_gcs2e_pos_sub32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    )
TWO_DIM_POSITION_WITHOUT_RETURN(float32,-,wx,wy,wc,vx,vy,vc,r)

int32
gcs_gcs2_pos_sub64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   )
TWO_DIM_POSITION_WITHOUT_RETURN(float64,-,w[X],w[Y],w[CELL2D],
			     v[X],v[Y],v[CELL2D],r)


int32
gcs_gcs2e_pos_sub64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    )
TWO_DIM_POSITION_WITHOUT_RETURN(float64,-,wx,wy,wc,vx,vy,vc,r)

int32
gcs_gcs3_pos_sub32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   )
THREE_DIM_POSITION_WITHOUT_RETURN(float32,-,w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D],r)
    
int32
gcs_gcs3e_pos_sub32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYZC]
		    )
THREE_DIM_POSITION_WITHOUT_RETURN(float32,-,wx,wy,wz,wc,vx,vy,vz,vc,r)


int32
gcs_gcs3_pos_sub64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   )
THREE_DIM_POSITION_WITHOUT_RETURN(float64,-,w[X],w[Y],w[Z],w[CELL3D],
			       v[X],v[Y],v[Z],v[CELL3D],r)

int32
gcs_gcs3e_pos_sub64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYZC]
		    )
THREE_DIM_POSITION_WITHOUT_RETURN(float64,-,wx,wy,wz,wc,vx,vy,vz,vc,r)

/*=================GCS DISTANCE BETWEEN POSITIONS===============*/
#define DISTANCE_ERROR_VAL -1.0

#define TWO_DIM_POSITION_DIST(type,func,wx,wy,wc,vx,vy,vc)\
{\
   if ((int32)wc != (int32) vc)\
   {\
       float64 tmpx, tmpy;\
       if (gcs_gcs2e_to_gcscs((int32)vc,(float64)vx,(float64)vy,\
				(int32)wc,&tmpx,&tmpy))\
         return func((wx-tmpx)*(wx-tmpx)+(wy-tmpy)*(wy-tmpy));\
       else\
	 return DISTANCE_ERROR_VAL; /* Invalid Coordinate Conversion */ \
   }\
   else\
   {\
     return func((wx-vx)*(wx-vx)+(wy-vy)*(wy-vy));\
   }\
}

#define THREE_DIM_POSITION_DIST(type,func,wx,wy,wz,wc,vx,vy,vz,vc)\
{\
   if ((int32)wc != (int32) vc)\
   {\
       float64 tmpx, tmpy,tmpz;\
       if (gcs_gcs3e_to_gcscs((int32)vc,(float64)vx,(float64)vy,(float64)vz,\
			    (int32)wc,&tmpx,&tmpy,&tmpz))\
         return func((wx-tmpx)*(wx-tmpx)+(wy-tmpy)*(wy-tmpy)+\
		     (wz-tmpz)*(wz-tmpz));\
       else\
	 return DISTANCE_ERROR_VAL; /* Invalid Coordinate Conversion */\
   }\
   else\
     return func((wx-vx)*(wx-vx)+(wy-vy)*(wy-vy)+(wz-vz)*(wz-vz));\
}

float32
gcs_gcs2_pos_distpl(
		    PointLocation *w,
		    PointLocation *v
		    )
{
    if ( w->cell != v->cell )
    {
	PointLocation tmp;
	
	/* Convert the points to the same frame of reference */
	if ( !gcs_gcs2_to_gcscspl(v, w->cell, &tmp) )
	  return DISTANCE_ERROR_VAL; /* Invalid Coordinate Conversion */
	else
	  /* Need to do some conversion to keep from overflow */
	  return 
	    fsqrt(((float32)w->x-(float32)tmp.x)*
		  ((float32)w->x-(float32)tmp.x)+
		  ((float32)w->y-(float32)tmp.y)*
		  ((float32)w->y-(float32)tmp.y));
    }
    else
      return 
	fsqrt(((float32)w->x-(float32)v->x)*((float32)w->x-(float32)v->x)+
	      ((float32)w->y-(float32)v->y)*((float32)w->y-(float32)v->y));
}

float32
gcs_gcs2_pos_dist32(
		    float32 w[XYC],
		    float32 v[XYC]
		    )
TWO_DIM_POSITION_DIST(float32,fsqrt,w[X],w[Y],w[CELL2D],v[X],v[Y],v[CELL2D])

float32
gcs_gcs2e_pos_dist32(
		     float32 wx,
		     float32 wy,
		     int32   wc,
		     float32 vx,
		     float32 vy,
		     int32   vc
		     )
TWO_DIM_POSITION_DIST(float32,fsqrt,wx,wy,wc,vx,vy,vc)

float64
gcs_gcs2_pos_dist64(
		    float64 w[XYC],
		    float64 v[XYC]
		    )
TWO_DIM_POSITION_DIST(float64,sqrt,w[X],w[Y],w[CELL2D],v[X],v[Y],v[CELL2D])

float64
gcs_gcs2e_pos_dist64(
		     float64 wx,
		     float64 wy,
		     int32   wc,
		     float64 vx,
		     float64 vy,
		     int32   vc
		     )
TWO_DIM_POSITION_DIST(float64,sqrt,wx,wy,wc,vx,vy,vc)
    
float32
gcs_gcs3_pos_dist32(
		    float32 w[XYZC],
		    float32 v[XYZC]
		    )
THREE_DIM_POSITION_DIST(float32,fsqrt,w[X],w[Y],w[Z],w[CELL3D],
		     v[X],v[Y],v[Z],v[CELL3D])


float32
gcs_gcs3e_pos_dist32(
		     float32 wx,
		     float32 wy,
		     float32 wz,
		     int32   wc,
		     float32 vx,
		     float32 vy,
		     float32 vz,
		     int32   vc
		     )
THREE_DIM_POSITION_DIST(float32,fsqrt,wx,wy,wz,wc,vx,vy,vz,vc)



float64
gcs_gcs3_pos_dist64(
		    float64 w[XYZC],
		    float64 v[XYZC]
		    )
THREE_DIM_POSITION_DIST(float64,sqrt,w[X],w[Y],w[Z],w[CELL3D],
		     v[X],v[Y],v[Z],v[CELL3D])

float64
gcs_gcs3e_pos_dist64(
		     float64 wx,
		     float64 wy,
		     float64 wz,
		     int32   wc,
		     float64 vx,
		     float64 vy,
		     float64 vz,
		     int32   vc
		     )
THREE_DIM_POSITION_DIST(float64,sqrt,wx,wy,wz,wc,vx,vy,vz,vc)


/*===================GCS POINT EQUALITY==================*/
#define GCS_THRESHOLD 0.001
#define GCS_GCS3_CONVERT_THRESHOLD 0.001
/* When doing 2D coord conversion, the amount of error introduced
 * is increase the farther you get away from the cell.
 * 
 * So depending on how far off a cell you are, this threshold may
 * or may not be large enough 
 */
#define GCS_GCS2_CONVERT_THRESHOLD 0.1


int32
gcs_gcs2_pos_equalpl(
		     PointLocation *w,
		     PointLocation *v
		     )
{
    float32 dist = gcs_gcs2_pos_distpl(w,v);
    float32 threshold;

    threshold = GCS_GCS2_CONVERT_THRESHOLD;

    if ((dist < threshold) && (dist != DISTANCE_ERROR_VAL))
      return TRUE;
    else 
      return FALSE;
}

/*==================GCS POSITION EQUAL============================*/

#define DIM_EQUAL(type,size,n,w, v)\
{\
    type dist = gcs_gcs##n##_pos_dist##size(w,v);\
    type threshold;\
    if ((int32)w[CELL##n##D] != (int32)v[CELL##n##D])\
      threshold = GCS_GCS##n##_CONVERT_THRESHOLD;\
    else\
      threshold = GCS_THRESHOLD;\
    if ((dist < threshold) && (dist != DISTANCE_ERROR_VAL))\
      return TRUE;\
    else \
      return FALSE;\
}

int32
gcs_gcs2_pos_equal32(
		     float32 w[XYC],
		     float32 v[XYC]
		     )
DIM_EQUAL(float32,32,2,w,v)

int32
gcs_gcs2_pos_equal64(
		     float64 w[XYC],
		     float64 v[XYC]
		     )
DIM_EQUAL(float64,64,2,w,v)

int32
gcs_gcs3_pos_equal32(
		     float32 w[XYZC],
		     float32 v[XYZC]
		     )
DIM_EQUAL(float32,32,3,w,v)

int32
gcs_gcs3_pos_equal64(
		     float64 w[XYZC],
		     float64 v[XYZC]
		     )
DIM_EQUAL(float64,64,3,w,v)


/*==================GCS VECTOR EQUAL============================*/

#define DIM_VEC_EQUAL(type,size,n,w, v)\
{\
    type threshold;\
    type dist;\
    type diff[n+1];\
    if ((int32)w[CELL##n##D] != (int32)v[CELL##n##D])\
      threshold = GCS_GCS##n##_CONVERT_THRESHOLD;\
    else\
      threshold = GCS_THRESHOLD;\
    if (!gcs_gcs##n##_vec_sub##size(w,v,diff))\
      return FALSE;\
    dist = gcs_gcs##n##_dot_prod##size(diff,diff);\
    if ((dist < threshold) && (dist != DISTANCE_ERROR_VAL))\
      return TRUE;\
    else \
      return FALSE;\
}

int32
gcs_gcs2_vec_equal32(
		     float32 w[XYC],
		     float32 v[XYC]
		     )
DIM_VEC_EQUAL(float32,32,2,w,v)

int32
gcs_gcs2_vec_equal64(
		     float64 w[XYC],
		     float64 v[XYC]
		     )
DIM_VEC_EQUAL(float64,64,2,w,v)

int32
gcs_gcs3_vec_equal32(
		     float32 w[XYZC],
		     float32 v[XYZC]
		     )
DIM_VEC_EQUAL(float32,32,3,w,v)

int32
gcs_gcs3_vec_equal64(
		     float64 w[XYZC],
		     float64 v[XYZC]
		     )
DIM_VEC_EQUAL(float64,64,3,w,v)
