/*
 * $RCSfile: libgcs.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: libgcs.h
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7600                                                        *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts*
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Public header file for libgcs
 *       Created: Tue Aug 29 17:36:21 EDT 1995
 *       Author: buettner
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#ifndef _LIBGCS_H_INCLUDED
#define _LIBGCS_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <libvecmat.h>
#include <libbasic.h>
#include <stdtypes.h> /*common/include/global*/
#include <stdext.h>
#include <basic.h>
  /*#include <po_const.h>
#include <p_po.h>*/

#define GCS_ILLEGAL_CELL          0
#define GCS_MAX_CELL_SIDE_METERS  112000

/* This macro will cast a GCS 3-Dimensional vector (X,Y,Z,C) to a 
 * GCS 2-Dimensional vector of format (X,Y,C).
 */
#define GCS_CAST_VEC_3D_TO_2D(vec3d,vec2d)         \
((vec2d)[X] = (vec3d)[X], (vec2d)[Y] = (vec3d)[Y], \
 (vec2d)[CELL2D] = (vec3d)[CELL3D])
/* This macro will cast a GCS 2-Dimensional vector (X,Y,C) to a 
 * GCS 2-Dimensional vector of format (X,Y,Z,C), where Z=0.0
 */
#define GCS_CAST_VEC_2D_TO_3D(vec2d,vec3d)         \
((vec3d)[X] = (vec2d)[X], (vec3d)[Y] = (vec2d)[Y], \
 (vec3d)[Z] = 0.0, (vec3d)[CELL3D] = (vec2d)[CELL2D])

/* File   : gcs_init.c
 * Routine: gcs_init
 *
 * This routine will initialize the appropriate libgcs data structures and 
 * environment.  It is mandatory to call this routine prior to using any
 * libgcs API routines.
 */
typedef void (*GCS_GET_DATUM_INFO) (float64  lat,
                                    float64  lng,
                                    int32   *datum,
                                    float64 *major,
                                    float64 *minor,
                                    float64 *dx,
                                    float64 *dy,
                                    float64 *dz
                                    );
extern void
gcs_init(GCS_GET_DATUM_INFO  datum_fnc);
    
/* File   : gcs_util.c
 * Routine: gcs_init_cell
 *
 * This sets up needed GCS data structures for a cell, and clears all
 * user defined ones.  If desired, it can be called more than once as a 
 * means of clearing all user-defined mappings.
 *
 * Returns TRUE if it successfully initializes the cell data.
 */
typedef void (*GCS_GCC_TO_TCC_FUNCTION) (float64       x_in,
					 float64       y_in,
					 float64       z_in,
					 ADDRESS       tcc,
					 float64      *x_out,
					 float64      *y_out,
					 float64      *z_out
					 );
typedef void (*GCS_TCC_TO_GCC_FUNCTION) (ADDRESS       tcc,
					 float64       x_in,
					 float64       y_in,
					 float64       z_in,
					 float64      *x_out,
					 float64      *y_out,
					 float64      *z_out
					 );
typedef struct gcs_flat_earth
{
    ADDRESS tcc_ptr;
    GCS_GCC_TO_TCC_FUNCTION gcc_to_tccf;/* function ptr for conversion
					 * from gcc to tcc flat earth
					 */
    GCS_TCC_TO_GCC_FUNCTION tccf_to_gcc;/* function ptr for conversion
					 * from tcc flat earth gcc to gcc
					 */
} GCS_FLAT_EARTH;

extern int32
gcs_init_cell (
	       int32 cell,
	       GCS_FLAT_EARTH *flat_earth
	       );

/* File   : gcs_access.c:
 * Routine: gcs_set_tdb
 *
 * This routine registers a pointer to a user terrain database to a cell id
 */
extern int32 
gcs_set_tdb (
	     int32   cell,
	     ADDRESS tdb,
	     int32   thinned
	     );

/* File   : gcs_access.c:
 * Routine: gcs_get_tdb
 *
 * This routine will retrieve the database pointer that is mapped to the 
 * input cell id.
 */
extern ADDRESS
gcs_get_tdb (int32 cell, 
	     int32 thinned  );


/* - - - - - NOTE ABOUT THE FOLLOWING THREE ROUTINES - - - - -
 * The "origin" specified in gcs_set_coord_data and 
 * gcs_get_origin_in_gcc is the central point in the cell.  
 * It records the location of the cell in global coordinates.
 *
 * The "origin" specified in gcs_offset_origin is the (0,0) point
 * of the cell, and the "offset" is the x and y distance from
 * the center point to the (0,0) point.
 *  
 * -----------------------------
 * \                           \
 * \                           \
 * \                           \
 * \                           \
 * \                           \
 * \                           \
 * \                           \
 * \                           \
 * \             A             \ -
 * \                           \ .
 * \                           \ .
 * \                           \ .
 * \                 .(x,y)    \ dy
 * \                           \ .
 * \                           \ .
 * \                           \ .
 * \B                          \ .
 * ----------------------------- -
 * \.....dx......\
 *
 * In the picture above, "A" is the origin specified in 
 * gcs_set_coord_data, and "B" is the origin specified in
 * gcs_offset_origin.  (-dx, -dy) are the parameters passed to
 * gcs_offset_origin.  To transform the point (x,y) into gcc,
 * we first set (x1,y1) = (x,y) - (dx,dy), to get the coordinates
 * relative to "A", the central point.  We then set 
 * (x2,y2) = M (x1,y1) + V, where M is the rotation-matrix relative
 * to "A", and V is the translation-vector for "A".
 *
 * The routines are typically called once at initialization, in 
 * the following order (cf gcs_init_coord_data):
 *      gcs_set_coord_data(cell,rot_matrix,gcc_origin);
 *      gcs_offset_origin(cell, xoff, yoff);
 */

/* File    : gcs_access.c
 * Routine : gcs_set_coord_data
 *
 * Sets the GCC coordinates of the central point of a cell,
 * used to transform between gcs coordinates and gcc.
 *
 * Sets the gcs-to-gcc rotation matrix for the given cell.
 *
 *       -sin_lon  -sin_lat*cos_lon     cos_lat*cos_lon
 *        cos_lon  -sin_lat*sin_lon     cos_lat*sin_lon
 *        0.0       cos_lat             sin_lat
 */

extern int32
gcs_set_coord_data (
		    int32    cell,
		    float64  rot_matrix[3][3],
		    float64  gcc_origin[XYZ]
		    );

/* File    : gcs_access.c
 * Routine : gcs_get_origin_in_gcc
 *
 * Gets the GCC coordinates of the central point of a cell,
 * used to transform between gcs coordinates and gcc.
 */
extern int32
gcs_get_origin_in_gcc (
		       int32    cell,
		       float64  location[XYZ]
		       );

/* File    : gcs_access.c
 * Routine : gcs_get_origin_offsets
 *
 * Obtains the GCS origin offset from the center of the cell.
 *
 * NOTE: the true origin of the cell is still
 *   at the point of tangency.  This provides
 *   an interface so that users can think of the
 *   origin as someplace else.
 */

extern int32
gcs_get_origin_offsets (int32 cell,
			float64 *xoff,
			float64 *yoff
			);

/* File    : gcs_access.c
 * Routine : gcs_offset_origin
 *
 * Sets the offset between the central point of a cell and the 
 * (0,0) point within the cell.  xoff is typically -(one-half the
 * width of the cell), and yoff is typically -(one-half the height
 * of the cell).
 *
 * NOTE: This procedure is typically called only once.
 */

extern int32
gcs_offset_origin (
		   int32 cell,
		   float64 xoff,
		   float64 yoff
		   );

/* File   : gcs_util.c:
 * Routine: gcs_lookup_cell_id
 * This routine will return the cell id associated with an input GCC point
 * in meters.
 */
extern int32
gcs_lookup_cell_id (
		    float64 x,
		    float64 y,
		    float64 z
		    );


/* File   : gcs_conv.c
 * Routine: gcs_gcc3e_to_gcscs
 *
 * This routine, given a cell and global coordinates u, v, w
 * converts those global coordinates into local coordinates is that cell.
 *
 * Returns TRUE is the conversion was successful
 */
extern int32 
gcs_gcc3e_to_gcscs (
		    float64  u,
		    float64  v,
		    float64  w,
		    int32    cell,
		    float64 *x, 
		    float64 *y, 
		    float64 *z
		    );


/* File   : gcs_conv.c
 * Routine: gcs_gcs3e_to_gcc
 * This routine, given a cell_id and local coordinates x, y, z
 * converts those coordinates into global ones
 *
 * Returns TRUE if the conversion was successful
 */
extern int32
gcs_gcs3e_to_gcc (
		  int32 cell,
		  float64 x, 
		  float64 y, 
		  float64 z,
		  float64 *u, 
		  float64 *v, 
		  float64 *w
		  );

/* File   : gcs_conv.c
 * Routine: gcs_gcs3e_to_gcscs
 * This routine converts local coordinates x0, y0, z0 in one cell 
 * reference, cell0, to local coordinates x1, y1, z1, in another cell
 * reference, cell1. 
 * Returns TRUE if the conversion was successful
 */
extern int32 
gcs_gcs3e_to_gcscs (
		    int32    cell0, 
		    float64  x0, 
		    float64  y0, 
		    float64  z0,
		    int32    cell1,
		    float64 *x1, 
		    float64 *y1, 
		    float64 *z1
		    );
/* File   : gcs_conv.c
 * Routine: gcs_gcs3e_to_gcs
 * This routine converts gcs coordinates x0, y0, z0 in one cell 
 * reference, cell0, to gcs coordinates x1, y1, z1, cell1, 
 * where the new gcs point references the cell which contains the
 * point.
 * Returns TRUE if the conversion was successful
 */
extern int32 
gcs_gcs3e_to_gcs (
		  int32    cell0, 
		  float64  x0, 
		  float64  y0, 
		  float64  z0,
		  int32   *native_cell,
		  float64 *native_x,
		  float64 *native_y,
		  float64 *native_z
		  );

/* File   : gcs_conv.c
 * Routine: gcs_gcs2e_to_gcscs
 * Convert (x0, y0) in cell 0 to the equivalent (x1, y1) in cell1.  This
 * conversion assumes that the point (x0, y0) is at height z=0.
 */
extern int32
gcs_gcs2e_to_gcscs (
		    int32    cell0,
		    float64  x0,
		    float64  y0,
		    int32    cell1,
		    float64 *x1,
		    float64 *y1
		    );

/* File   : gcs_conv.c
 * Routine: gcs_gcs2_to_gcscs
 * Convert xy0 in cell0 to the equivalent xy1 in cell1.  This
 * conversion assumes that the point xy0 is at height z=0.
 */
extern int32
gcs_gcs2_to_gcscs (
		   int32     cell0,
		   float64   xy0[XY],
		   int32     cell1,
		   float64   xy1[XY]
		   );
/* File   : gcs_conv.c
 * Routine: gcs_gcs2_to_gcscspl
 * Convert pt0 to the equivalent pt1 in cell1.  This
 * conversion assumes that pt0 is at height z=0.
 */
extern int32
gcs_gcs2_to_gcscspl (
		     PointLocation  *pt0,
		     int32           cell1,
		     PointLocation  *pt1
		     );
/* File   : gcs_conv.c
 * Routine: gcs_gcc3e_to_gcs
 * This routine converts global coordinates u, v, w into local coordinates
 * x, y, z, after calculating the appropriate cell, cell.
 */
extern int32 
gcs_gcc3e_to_gcs (
		  float64  u, 
		  float64  v, 
		  float64  w,
		  int32   *cell,
		  float64 *x, 
		  float64 *y, 
		  float64 *z
		  );
/* File   : gcs_conv.c
 * Routine: gcs_gcc3_to_gcs
 * This routine converts a geocentric coordinate system vector (X,Y,Z) to a
 * GCS vector cell,(X,Y,Z).
 */
extern int32
gcs_gcc3_to_gcs (
		 float64   uvw[XYZ],
		 int32    *cell,
		 float64   xyz[XYZ]
		 );


/* File   : gcs_conv.c
 * Routine: gcs_gcc3_to_gcscs
 * This routine is the vector form of gcs_gcc3e_to_gcscs
 */
extern int32 
gcs_gcc3_to_gcscs (
		   float64 uvw[XYZ], 
		   int32   cell, 
		   float64 xyz[XYZ]
		   );

/* File   : gcs_conv.c
 * Routine: gcs_gcs3_to_gcc
 * This routine is the vector form of gcs_gcs3e_to_gcc
 */
extern int32 
gcs_gcs3_to_gcc (
		 int32 cell, 
		 float64 xyz[XYZ], 
		 float64 uvw[XYZ]
		 );

/* File   : gcs_conv.c
 * Routine: gcs_gcs2e_to_gcs
 * This routine converts a 2d gcs point in any context into a 2d point in the
 * context of the cell that contains that point.
 */
extern int32 
gcs_gcs2e_to_gcs (
		  int32 cell0,
		  float64 x,
		  float64 y,
		  int32 *cell1,
		  float64 *x1,
		  float64 *y1
		  );
/* File   : gcs_conv.c
 * Routine: gcs_gcs2_to_gcs
 * Convert xy0 in cell0 into the equivalent point xy1 in the context of the
 * cell that contains that point.  This conversion assumes that the point xy0
 * is at height z=0.
 */
extern int32 
gcs_gcs2_to_gcs (
		  int32 cell0,
		  float64 xy0[XY],
		  int32 *cell1,
		  float64 xy1[XY]
		  );
/* File   : gcs_conv.c
 * Routine: gcs_gcs3_to_gcs
 * This routine is the vector form of gcs_gcs3e_to_gcs
 
 */
extern int32 
gcs_gcs3_to_gcs (
		 int32   cell0, 
		 float64 xyz0[XYZ],
		 int32  *cell1, 
		 float64 xyz1[XYZ]
		 );
/* File   : gcs_conv.c
 * Routine: gcs_gcs3_to_gcscs
 * This routine is the vector form of gcs_gcs3e_to_gcscs
 
 */
extern int32 
gcs_gcs3_to_gcscs (
		   int32 cell0, 
		   float64 xyz0[XYZ],
		   int32 cell1, 
		   float64 xyz1[XYZ]
		   );

/* File    : gcs_util.c
 * Routine : gcs_extent_to_cell_list
 *
 * This routine will take a geodetic extent (lat/lng) where *_start defines
 * the lower left lat/lng and *_stop defines the upper right lat/lng.  The 
 * output of this routine is a list of cells which have an intersection with
 * the input extent.  
 * *n_cell is in-out:  on input, it is the length of the cell_list array, and
 * on output it is the number of cells covered by the specified area.  If
 * the number of covered cells exceeds the list length, the return value
 * for the function is 0, otherwise it is 1.
 */
extern int32
gcs_extent_to_cell_list (
			 float64 lat_start,
			 float64 lng_start,
			 float64 lat_stop,
			 float64 lng_stop,
			 int32  *n_cell,
			 int32   cell_list[]
			 );


/* File    : gcs_util.c
 * Routine : gcs_ll_bounds_from_gcs
 *
 * Converts min/max x/y in GCS into lat/long boundaries.  Returns TRUE for
 * success, FALSE for failure.
 */
extern int32
gcs_ll_bounds_from_gcs (
			int32   cell,
			float64 min_x,
			float64 min_y,
			float64 max_x,
			float64 max_y,
			float64 *west_lng,
			float64 *south_lat,
			float64 *east_lng,
			float64 *north_lat
			);


/* File    : gcs_util.c
 * Routine : gcs_cells_within_radius
 *
 * Returns all cells within meters of (x,y,z,cell) in cell_list, assuming
 * that the number of cells is <= *n_cell, which specifies the length of
 * the (caller allocated) cell_list array.  If the number of cells is greater
 * than this, the actual number of cells within the radius is returned
 * in *n_cell, and the return value is 0, otherwise the return value is 1.  
 * Note that this routine is guaranteed to return all cells
 * within the given radius, but may also return some cells that are beyond
 * it.
 */
extern int32
gcs_cells_within_radius (
			 int32   cell,
			 float64 x,
			 float64 y,
			 float64 z,
			 float64 radius,
			 int32  *n_cell,
			 int32   cell_list[]
			 );


/* File    : gcs_rdr.c
 * Routine : gcs_rdr_lookup_cell_extents
 *
 * This routine returns the following:
 *    *min_lat:  latitude of southern-most point in cell
 *    *max_lat:  latitude of northern-most point in cell
 *    *min_lng:  longitude of western-most point in cell
 *    *max_lng:  longitude of eastern-most point in cell
 *
 * Note:  this reader file will be in memory
 * if the reader containing the cell id has been accessed once before (
 * which is the case if the playbox has been initialized - see libworld).
 */
extern int32
gcs_lookup_cell_extents (
			 int32 cell,
			 float64 *min_lat,
			 float64 *max_lat,
			 float64 *min_lng,
			 float64 *max_lng
			 );
/* File    : gcs_access.c
 * Routine : gcs_get_rot_mat
 *
 * Obtains a rotation matrix used to transform gcs coordinates
 * into gcc.  
 *
 *       -sin_lon  -sin_lat*cos_lon     cos_lat*cos_lon
 *        cos_lon  -sin_lat*sin_lon     cos_lat*sin_lon
 *        0.0       cos_lat             sin_lat
 *
 * This matrix is used for coordinate transformation as described in:
 * equation 21 in the paper:
 * "Interconversions Between Different Coordinate Systems"
 * Kuo-Chi Lin, Huat Keng Ng; IST UCF
 * 
 * This equation is of the form:
 * position_gcc = position_gcs * rotation_matrix + origin_in_gcc
 */
extern int32
gcs_get_rot_mat (
		 int32    cell,
		 float64  matrix[3][3]
		 );
/* File    : gcs_conv.c
 * Routine : gcs_vec_to_vec
 *
 * Converts a vector XYZ from "from_cell" to "to_cell"
 */
extern int32
gcs_vec_to_vec (
		int32   from_cell,
		float64 from_vec[XYZ],
		int32   to_cell,
		float64 to_vec[XYZ]
		);

/* File    : gcs_conv.c
 * Routine : gcs_rot_to_rot
 *
 * Converts rotation-matrix from cell from_cell 
 * into a rotation-matrix in cell to_cell
 */
extern int32
gcs_rot_to_rot (
		int32   from_cell,
		float64 from_rot[3][3],
		int32   to_cell,
		float64 to_rot[3][3]
		);
/* File    : gcs_util.c
 * Routine : gcs_lookup_ht_asl
 *
 * Calculates mean-sea-level (distance from
 * point on tangent plane to ellipsoid) for
 * a given cell, x, y, along the gravity vector.
 */
extern float64
gcs_lookup_ht_asl (
		   int32    cell, 
		   float64  x, 
		   float64  y,
		   float64  z
		   );
/* File    : gcs_util.c
 * Routine : gcs_lookup_gravity_vector
 *
 * Computes the normalized down vector for any x, y,
 * for any cell.
 */
extern int32
gcs_lookup_gravity_vector (
			   int32    cell, 
			   float64  x, 
			   float64  y,
			   float64  z,
			   float64  vec[XYZ]
			   );

/* File    : gcs_access.c
 * Routine : gcs_get_pbt_info
 *
 * Fetches the address of position based table info structure
 * associated with the cell
 */
extern ADDRESS
gcs_get_pbt_info (
		  int32 cell
		  );

/* File    : gcs_access.c
 * Routine : gcs_set_pbt_info
 *
 * Sets the address of the position based table 
 * information structure associated with the cell to pbt
 */
extern int32 
gcs_set_pbt_info (
		  int32   cell, 
		  ADDRESS pbt
		  );

/* File    : gcs_access.c
 * Routine : gcs_set_cell_awareness
 *
 * Sets the flag which allows libGCS to process single or multi cell.
 *
 * GCS_MULTI_CELL means that this library should treat all cell 
 * information as valid
 *
 * GCS_SINGLE_CELL means only one CTDB cell available and no cell to cell
 * conversions are possible.
 */
#define GCS_SINGLE_CELL 0
#define GCS_MULTI_CELL  1
extern void
gcs_set_cell_awareness ( int32 aware );

/* File    : gcs_conv.c
 * Routine : gcs_gcc_to_utm
 *
 * Conversion from geographics coordinates (latitude/longitude) to 
 * UTM (NE).
 */
extern void 
gcs_gcc_to_utm(
	       float64  u, 
	       float64  v, 
	       float64  w, 
	       int32   *zone, 
	       float64 *northing, 
	       float64 *easting,
	       float64 *z
	       );

/* File    : gcs_conv.c
 * Routine : gcs_gcc_to_llz
 *
 * Conversion from geographics coordinates (latitude/longitude) to 
 * geocentric coordinates.                                        
 */
extern int32
gcs_gcc_to_llz (
		float64  x,
		float64  y,
		float64  z,
		float64 *lat,
		float64 *lng, 
		float64 *height,
		int32    datum
		);    
/* File    : gcs_conv.c
 * Routine : gcs_llz_to_gcc
 *
 * Conversion from geographics coordinates (latitude/longitude) to 
 * goecentric coordinates.                                        
 */
extern void
gcs_llz_to_gcc (
		float64 lat,
		float64 lng,
		float64 z,
		int32   datum,
		float64 *u,
		float64 *v,
		float64 *w 
		);

/* File    : gcs_conv.c
 * Routine : gcs_llz_to_gcc
 *
 * Conversion from gcs coordinates to geographics coordinates.
 */
extern int32
gcs_gcs_to_llz (
		int32   cell, 
		float64 x, 
		float64 y, 
		float64 z,
		float64 *lat,
		float64 *lng,
		float64 *h,
		int32   datum
		);

/*======================GCS VECTOR FUNCTIONS=======================*/

/*=======================GCS VECTOR ADDITION=====================*/
extern int32
gcs_gcs2_vec_add32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   );
extern int32
gcs_gcs2e_vec_add32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    );

extern int32
gcs_gcs3_vec_add32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   );
extern int32
gcs_gcs3e_vec_add32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYC]
		    );
extern int32
gcs_gcs2_vec_add64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   );

extern int32
gcs_gcs2e_vec_add64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    );

extern int32
gcs_gcs3_vec_add64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   );

extern int32
gcs_gcs3e_vec_add64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYC]
		    );

/*=======================GCS VECTOR SUBTRACTION=====================*/

extern int32
gcs_gcs2_vec_sub32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   );

extern int32
gcs_gcs2e_vec_sub32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    );
extern int32
gcs_gcs3_vec_sub32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   );

extern int32
gcs_gcs3e_vec_sub32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYC]
		    );

extern int32
gcs_gcs2_vec_sub64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   );
extern int32
gcs_gcs2e_vec_sub64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    );
extern int32
gcs_gcs3_vec_sub64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   );
extern int32
gcs_gcs3e_vec_sub64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYC]
		    );
/*=======================GCS UNIT VECTOR =====================*/
extern void 
gcs_gcs2_unit32(
		float32 v[XYC],
		float32 r[XYC]
		);

extern void 
gcs_gcs2e_unit32(
		 float32 vx, 
		 float32 vy,
		 int32   vc,
		 float32 r[XYC]
		 );

extern void 
gcs_gcs2_unit64(
		float64 v[XYC],
		float64 r[XYC]
		);

extern void 
gcs_gcs2e_unit64(
		 float64 vx, 
		 float64 vy,
		 int32   vc,
		 float64 r[XYC]
		 );

extern void 
gcs_gcs3_unit32(
		float32 v[XYZC],
		float32 r[XYZC]
		);

extern void 
gcs_gcs3e_unit32(
		 float32 vx, 
		 float32 vy,
		 float32 vz,
		 int32   vc,
		 float32 r[XYZC]
		 );
extern void 
gcs_gcs3_unit64(
		float64 v[XYZC],
		float64 r[XYZC]
		);

extern void 
gcs_gcs3e_unit64(
		 float64 vx, 
		 float64 vy,
		 float64 vz,
		 int32   vc,
		 float64 r[XYZC]
		 );
/*=======================GCS NEGATE VECTOR =====================*/
extern void
gcs_gcs2_negate32(
		  float32 v[XYC],
		  float32 r[XYC]
		  );

extern void
gcs_gcs2e_negate32(
		   float32 vx,
		   float32 vy,
		   int32   vc,
		   float32 r[XYC]
		   );

extern void
gcs_gcs3_negate32(
		  float32 v[XYZC],
		  float32 r[XYZC]
		  );

extern void
gcs_gcs3e_negate32(
		   float32 vx,
		   float32 vy,
		   float32 vz,
		   int32   vc,
		   float32 r[XYZC]
		   );

extern void
gcs_gcs2_negate64(
		  float64 v[XYC],
		  float64 r[XYC]
		  );

extern void
gcs_gcs2e_negate64(
		   float64 vx,
		   float64 vy,
		   int32   vc,
		   float64 r[XYC]
		   );

extern void
gcs_gcs3_negate64(
		  float64 v[XYZC],
		  float64 r[XYZC]
		  );

extern void
gcs_gcs3e_negate64(
		   float64 vx,
		   float64 vy,
		   float64 vz,
		   int32   vc,
		   float64 r[XYZC]
		   );
/*=======================GCS VECTOR DOT PRODUCT=====================*/

extern float32
gcs_gcs2_dot_prod32(
		    float32  w[XYC],
		    float32  v[XYC]
		    );

extern float32
gcs_gcs2e_dot_prod32(
		     float32  wx,
		     float32  wy,
		     int32    wc,
		     float32  vx,
		     float32  vy,
		     int32    vc
		     );
extern float32
gcs_gcs3_dot_prod32(
		    float32  w[XYZC],
		    float32  v[XYZC]
		    );

extern float32
gcs_gcs3e_dot_prod32(
		     float32  wx,
		     float32  wy,
		     float32  wz,
		     int32    wc,
		     float32  vx,
		     float32  vy,
		     float32  vz,
		     int32    vc
		     );

extern float64
gcs_gcs2_dot_prod64(
		    float64  w[XYC],
		    float64  v[XYC]
		    );

extern float64
gcs_gcs2e_dot_prod64(
		     float64  wx,
		     float64  wy,
		     int32    wc,
		     float64  vx,
		     float64  vy,
		     int32    vc
		     );

extern float64
gcs_gcs3_dot_prod64(
		    float64  w[XYZC],
		    float64  v[XYZC]
		    );

extern float64
gcs_gcs3e_dot_prod64(
		     float64  wx,
		     float64  wy,
		     float64  wz,
		     int32    wc,
		     float64  vx,
		     float64  vy,
		     float64  vz,
		     int32    vc
		     );

/*=======================GCS VECTOR CROSS PRODUCT=====================*/
extern float32
gcs_gcs2_cross_prod32(
		      float32  w[XYC],
		      float32  v[XYC]
		      );

extern float32
gcs_gcs2e_cross_prod32(
		       float32  wx,
		       float32  wy,
		       int32    wc,
		       float32  vx,
		       float32  vy,
		       int32    vc
		       );

extern int32
gcs_gcs3_cross_prod32(
		      float32  w[XYZC],
		      float32  v[XYZC],
		      float32  r[XYZC]
		      );

extern int32
gcs_gcs3e_cross_prod32(
		       float32  wx,
		       float32  wy,
		       float32  wz,
		       int32    wc,
		       float32  vx,
		       float32  vy,
		       float32  vz,
		       int32    vc,
		       float32  r[XYZC]
		       );

extern float64
gcs_gcs2_cross_prod64(
		      float64  w[XYC],
		      float64  v[XYC]
		      );

extern float64
gcs_gcs2e_cross_prod64(
		       float64  wx,
		       float64  wy,
		       int32    wc,
		       float64  vx,
		       float64  vy,
		       int32    vc
		       );

extern int32
gcs_gcs3_cross_prod64(
		      float64  w[XYZC],
		      float64  v[XYZC],
		      float64  r[XYZC]
		      );

extern int32
gcs_gcs3e_cross_prod64(
		       float64  wx,
		       float64  wy,
		       float64  wz,
		       int32    wc,
		       float64  vx,
		       float64  vy,
		       float64  vz,
		       int32    vc,
		       float64  r[XYZC]
		       );

/*=======================GCS VECTOR MAGNITUDE SQUARED=====================*/
#define GCS_GCS2_VEC_MAG_SQ(w)               (VMAT2_VEC_MAG_SQ(w))
#define GCS_GCS3_VEC_MAG_SQ(w)               (VMAT33_VEC_MAG_SQ(w))
#define GCS_GCS2E_VEC_MAG_SQ(wx, wy, wc)     (VMAT2E_VEC_MAG_SQ(wx, wy))
#define GCS_GCS3E_VEC_MAG_SQ(wx, wy, wz, wc) (VMAT3E_VEC_MAG_SQ(wx, wy, wz))



/*=======================GCS VECTOR SCALE MULT=====================*/
extern void 
gcs_gcs2_scal_vec_mul32(
			float32 s,
			float32 v[XYC],
			float32 r[XYC]
			);

extern void 
gcs_gcs2e_scal_vec_mul32(
			 float32 s,
			 float32 vx,
			 float32 vy,
			 int32   vc,
			 float32 r[XYC]
			 );

extern void 
gcs_gcs3_scal_vec_mul32(
			float32 s,
			float32 v[XYZC],
			float32 r[XYZC]
			);

extern void 
gcs_gcs3e_scal_vec_mul32(
			 float32 s,
			 float32 vx,
			 float32 vy,
			 float32 vz,
			 int32   vc,
			 float32 r[XYZC]
			 );

extern void 
gcs_gcs2_scal_vec_mul64(
			float64 s,
			float64 v[XYC],
			float64 r[XYC]
			);


extern void 
gcs_gcs2e_scal_vec_mul64(
			 float64 s,
			 float64 vx,
			 float64 vy,
			 int32   vc,
			 float64 r[XYC]
			 );

extern void 
gcs_gcs3_scal_vec_mul64(
			float64 s,
			float64 v[XYZC],
			float64 r[XYZC]
			);

extern void 
gcs_gcs3e_scal_vec_mul64(
			 float64 s,
			 float64 vx,
			 float64 vy,
			 float64 vz,
			 int32   vc,
			 float64 r[XYZC]
			 );

/*=================GCS VECTOR ANGLE BETWEEN VECTORS==================*/

extern float32
gcs_gcs2_angle_between_vectors32(
				 float32  w[XYC],
				 float32  v[XYC]
				 );

extern float32
gcs_gcs2e_angle_between_vectors32(
				  float32  wx,
				  float32  wy,
				  int32    wc,
				  float32  vx,
				  float32  vy,
				  int32    vc
				  );

extern float32
gcs_gcs3_angle_between_vectors32(
				 float32  w[XYZC],
				 float32  v[XYZC]
				 );
extern float32
gcs_gcs3e_angle_between_vectors32(
				  float32  wx,
				  float32  wy,
				  float32  wz,
				  int32    wc,
				  float32  vx,
				  float32  vy,
				  float32  vz,
				  int32    vc
				  );

extern float64
gcs_gcs2_angle_between_vectors64(
				 float64  w[XYC],
				 float64  v[XYC]
				 );

extern float64
gcs_gcs2e_angle_between_vectors64(
				  float64  wx,
				  float64  wy,
				  int32    wc,
				  float64  vx,
				  float64  vy,
				  int32    vc
				  );

extern float64
gcs_gcs3_angle_between_vectors64(
				 float64  w[XYZC],
				 float64  v[XYZC]
				 );

extern float64
gcs_gcs3e_angle_between_vectors64(
				  float64  wx,
				  float64  wy,
				  float64  wz,
				  int32    wc,
				  float64  vx,
				  float64  vy,
				  float64  vz,
				  int32    vc
				  );


/*=====================GCS VECTOR PROJECT PARALLEL======================*/

extern int32
gcs_gcs2_project32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   );

extern int32
gcs_gcs2e_project32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    );

extern int32
gcs_gcs3_project32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   );

extern int32
gcs_gcs3e_project32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYZC]
		    );

extern int32
gcs_gcs2_project64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   );

extern int32
gcs_gcs2e_project64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    );

extern int32
gcs_gcs3_project64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   );

extern int32
gcs_gcs3e_project64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYZC]
		    );

/*=====================GCS VECTOR PROJECT PERPENDICULAR======================*/
extern int32
gcs_gcs2_project_perp32(
			float32 w[XYC],
			float32 v[XYC],
			float32 r[XYC]
			);
extern int32
gcs_gcs2e_project_perp32(
			 float32 wx,
			 float32 wy,
			 int32   wc,
			 float32 vx,
			 float32 vy,
			 int32   vc,
			 float32 r[XYC]
			 );

extern int32
gcs_gcs3_project_perp32(
			float32 w[XYZC],
			float32 v[XYZC],
			float32 r[XYZC]
			);

extern int32
gcs_gcs3e_project_perp32(
			 float32 wx,
			 float32 wy,
			 float32 wz,
			 int32   wc,
			 float32 vx,
			 float32 vy,
			 float32 vz,
			 int32   vc,
			 float32 r[XYZC]
			 );

extern int32
gcs_gcs2_project_perp64(
			float64 w[XYC],
			float64 v[XYC],
			float64 r[XYC]
			);

extern int32
gcs_gcs2e_project_perp64(
			 float64 wx,
			 float64 wy,
			 int32   wc,
			 float64 vx,
			 float64 vy,
			 int32   vc,
			 float64 r[XYC]
			 );
extern int32
gcs_gcs3_project_perp64(
			float64 w[XYZC],
			float64 v[XYZC],
			float64 r[XYZC]
			);
extern int32
gcs_gcs3e_project_perp64(
			 float64 wx,
			 float64 wy,
			 float64 wz,
			 int32   wc,
			 float64 vx,
			 float64 vy,
			 float64 vz,
			 int32   vc,
			 float64 r[XYZC]
			 );

/*=====================GCS VECTOR VECTOR COPY======================*/
#define GCS_GCS2_VEC_COPY32(w, r)  (bcopy((char *)(w),(char *)(r),(XYC  * sizeof(float32))))
#define GCS_GCS3_VEC_COPY32(w, r)  (bcopy((char *)(w),(char *)(r),(XYZC * sizeof(float32))))
#define GCS_GCS2_VEC_COPY64(w, r)  (bcopy((char *)(w),(char *)(r),(XYC  * sizeof(float64))))
#define GCS_GCS3_VEC_COPY64(w, r)  (bcopy((char *)(w),(char *)(r),(XYZC * sizeof(float64))))


/*=====================GCS VECTOR ZEROED======================*/
#define GCS_GCS2_VEC_ZERO32(w) ((bzero((char *)(w),(XYC *sizeof(float32)))),\
				(w[CELL2D]=GCS_ILLEGAL_CELL))
#define GCS_GCS3_VEC_ZERO32(w) ((bzero((char *)(w),(XYZC*sizeof(float32)))),\
				(w[CELL3D]=GCS_ILLEGAL_CELL))
#define GCS_GCS2_VEC_ZERO64(w) ((bzero((char *)(w),(XYC *sizeof(float64)))),\
				(w[CELL2D]=GCS_ILLEGAL_CELL))
#define GCS_GCS3_VEC_ZERO64(w) ((bzero((char *)(w),(XYZC*sizeof(float64)))),\
				(w[CELL3D]=GCS_ILLEGAL_CELL))
/*=====================GCS VECTOR EQUALITY======================*/


extern int32
gcs_gcs2_vec_equal32(float32 w[XYC], float32 r[XYC]);

extern int32
gcs_gcs3_vec_equal32(float32 w[XYZC], float32 r[XYZC]);

extern int32
gcs_gcs2_vec_equal64(float64 w[XYC], float64 r[XYC]);

extern int32
gcs_gcs3_vec_equal64(float64 w[XYZC], float64 r[XYZC]);


/*==============GCS VECTOR OFFSET POINT BY VECTOR====================*/
extern int32 
gcs_gcs3_pos_vec_add64(
		       float64 pt0[XYZC],
		       float64 offset[XYZC],
		       float64 npt[XYZC]
		       );
extern int32 
gcs_gcs3e_pos_vec_add64(
			float64 ptx,
			float64 pty,
			float64 ptz,
			int32   ptc,
			float64 offx,
			float64 offy,
			float64 offz,
			int32   offc,
			float64 npt[XYZC]
			);
extern int32 
gcs_gcs3_pos_vec_add32(
		       float64 pt0[XYZC]    ,
		       float64 offset[XYZC] ,
		       float32 npt[XYZC]
		       );
extern int32 
gcs_gcs3e_pos_vec_add32(
			float32 ptx,
			float32 pty,
			float32 ptz,
			int32   ptc,
			float32 offx,
			float32 offy,
			float32 offz,
			int32   offc,
			float32 npt[XYZC]
			);
extern int32 
gcs_gcs2_pos_vec_add64(
		       float64 pt0[XYC]    ,
		       float64 offset[XYC] ,
		       float64 npt[XYC]
		       );
extern int32 
gcs_gcs2e_pos_vec_add64(
			float64 ptx,
			float64 pty,
			int32   ptc,
			float64 offx,
			float64 offy,
			int32   offc,
			float64 npt[XYC]
			);
extern int32 
gcs_gcs2_pos_vec_add32(
		       float64 pt0[XYC]    ,
		       float64 offset[XYC] ,
		       float32 npt[XYC]
		       );

extern int32 
gcs_gcs2e_pos_vec_add32(
			float32 ptx,
			float32 pty,
			int32   ptc,
			float32 offx,
			float32 offy,
			int32   offc,
			float32 npt[XYC]
			);
extern int32 
gcs_gcs2_pos_vec_addpl(
		       PointLocation *pt0,
		       float64           offset[XYC],
		       PointLocation *npt
		       );


/* ------------------------------------------------------------------*/
/* ---------------- GCS VECTOR ROTATION OPERATIONS ------------------*/
/* ------------------------------------------------------------------*/
/* File    : gcs_vec.c
 * Routine : gcs_gcs3_vec_rotate64
 *
 * Converts a vector XYZC from "from_cell" to "to_cell"
 */
extern int32
gcs_gcs3_vec_rotate64 (
		       float64 from_vec[XYZC],
		       int32   to_cell,
		       float64 to_vec[XYZC]
		       );
/* File    : gcs_vec.c
 * Routine : gcs_gcs2_vec_rotate64
 *
 * Converts a vector XY from "from_cell" to "to_cell"
 */
extern int32
gcs_gcs2_vec_rotate64 (
		       float64 from_vec[XYC],
		       int32   to_cell,
		       float64 to_vec[XYC]
		       );

extern int32
gcs_gcs2e_vec_rotate64(
		       float64  fromx,
		       float64  fromy,
		       int32    fromc,
		       int32    to_cell,
		       float64  to_vec[XYC]
		       );
extern int32
gcs_gcs3e_vec_rotate64(
		       float64  fromx,
		       float64  fromy,
		       float64  fromz,
		       int32    fromc,
		       int32    to_cell,
		       float64  to_vec[XYZC]
		       );
extern int32
gcs_gcs2_vec_rotate32(
		      float32  from_vec[XYC],
		      int32    to_cell,
		      float32  to_vec[XYC]
		      );

extern int32
gcs_gcs2e_vec_rotate32(
		       float32  fromx,
		       float32  fromy,
		       int32    fromc,
		       int32    to_cell,
		       float32  to_vec[XYC]
		       );
extern int32
gcs_gcs3_vec_rotate32(
		      float32  from_vec[XYZC],
		      int32    to_cell,
		      float32  to_vec[XYZC]
		      );

extern int32
gcs_gcs3e_vec_rotate32(
		       float32  fromx,
		       float32  fromy,
		       float32  fromz,
		       int32    fromc,
		       int32    to_cell,
		       float32  to_vec[XYZC]
		       );

/*====================GCS POINT ADDITION======================*/
extern int32
gcs_gcs2_pos_addpl(
		   PointLocation *w,
		   PointLocation *v,
		   PointLocation *r
		   );
extern int32
gcs_gcs2_pos_add32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   );
extern int32
gcs_gcs2_pos_add32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   );

extern int32
gcs_gcs2e_pos_add32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    );

extern int32
gcs_gcs2_pos_add64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   );

extern int32
gcs_gcs2e_pos_add64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    );

extern int32
gcs_gcs3_pos_add32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   );
extern int32
gcs_gcs3e_pos_add32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYZC]
		    );

extern int32
gcs_gcs3_pos_add64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   );
extern int32
gcs_gcs3e_pos_add64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYZC]
		    );

/*====================GCS POINT SUBTRACTION======================*/
extern int32
gcs_gcs2_pos_subpl(
		   PointLocation *w,
		   PointLocation *v,
		   PointLocation *r
		   );
extern int32
gcs_gcs2_pos_sub32(
		   float32 w[XYC],
		   float32 v[XYC],
		   float32 r[XYC]
		   );

extern int32
gcs_gcs2e_pos_sub32(
		    float32 wx,
		    float32 wy,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    int32   vc,
		    float32 r[XYC]
		    );

extern int32
gcs_gcs2_pos_sub64(
		   float64 w[XYC],
		   float64 v[XYC],
		   float64 r[XYC]
		   );

extern int32
gcs_gcs2e_pos_sub64(
		    float64 wx,
		    float64 wy,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    int32   vc,
		    float64 r[XYC]
		    );

extern int32
gcs_gcs3_pos_sub32(
		   float32 w[XYZC],
		   float32 v[XYZC],
		   float32 r[XYZC]
		   );
extern int32
gcs_gcs3e_pos_sub32(
		    float32 wx,
		    float32 wy,
		    float32 wz,
		    int32   wc,
		    float32 vx,
		    float32 vy,
		    float32 vz,
		    int32   vc,
		    float32 r[XYZC]
		    );

extern int32
gcs_gcs3_pos_sub64(
		   float64 w[XYZC],
		   float64 v[XYZC],
		   float64 r[XYZC]
		   );
extern int32
gcs_gcs3e_pos_sub64(
		    float64 wx,
		    float64 wy,
		    float64 wz,
		    int32   wc,
		    float64 vx,
		    float64 vy,
		    float64 vz,
		    int32   vc,
		    float64 r[XYZC]
		    );

/*=================GCS DISTANCE BETWEEN POINTS===============*/
extern float32
gcs_gcs2_pos_distpl(
		    PointLocation *w,
		    PointLocation *v
		    );

extern float32
gcs_gcs2_pos_dist32(
		    float32 w[XYC],
		    float32 v[XYC]
		    );

extern float32
gcs_gcs2e_pos_dist32(
		     float32 wx,
		     float32 wy,
		     int32   wc,
		     float32 vx,
		     float32 vy,
		     int32   vc
		     );

extern float64
gcs_gcs2_pos_dist64(
		    float64 w[XYC],
		    float64 v[XYC]
		    );

extern float64
gcs_gcs2e_pos_dist64(
		     float64 wx,
		     float64 wy,
		     int32   wc,
		     float64 vx,
		     float64 vy,
		     int32   vc
		     );

extern float32
gcs_gcs3_pos_dist32(
		    float32 w[XYZC],
		    float32 v[XYZC]
		    );

extern float32
gcs_gcs3e_pos_dist32(
		     float32 wx,
		     float32 wy,
		     float32 wz,
		     int32   wc,
		     float32 vx,
		     float32 vy,
		     float32 vz,
		     int32   vc
		     );


extern float64
gcs_gcs3_pos_dist64(
		    float64 w[XYZC],
		    float64 v[XYZC]
		    );
extern float64
gcs_gcs3e_pos_dist64(
		     float64 wx,
		     float64 wy,
		     float64 wz,
		     int32   wc,
		     float64 vx,
		     float64 vy,
		     float64 vz,
		     int32   vc
		     );

/*===================GCS POINT EQUALITY==================*/
extern int32
gcs_gcs2_pos_equalpl(
		     PointLocation *w,
		     PointLocation *v
		     );

extern int32
gcs_gcs2_pos_equal32(
		     float32 w[XYC],
		     float32 v[XYC]
		     );

extern int32
gcs_gcs2e_pos_equal32(
		      float32 wx,
		      float32 wy,
		      int32   wc,
		      float32 vx,
		      float32 vy,
		      int32   vc
		      );

extern int32
gcs_gcs2_pos_equal64(
		     float64 w[XYC],
		     float64 v[XYC]
		     );

extern int32
gcs_gcs2e_pos_equal64(
		      float64 wx,
		      float64 wy,
		      int32   wc,
		      float64 vx,
		      float64 vy,
		      int32   vc
		      );

extern int32
gcs_gcs3_pos_equal32(
		     float32 w[XYZC],
		     float32 v[XYZC]
		     );
extern int32
gcs_gcs3e_pos_equal32(
		      float32 wx,
		      float32 wy,
		      float32 wz,
		      int32   wc,
		      float32 vx,
		      float32 vy,
		      float32 vz,
		      int32   vc
		      );

extern int32
gcs_gcs3_pos_equal64(
		     float64 w[XYZC],
		     float64 v[XYZC]
		     );

extern int32
gcs_gcs3e_pos_equal64(
		      float64 wx,
		      float64 wy,
		      float64 wz,
		      int32   wc,
		      float64 vx,
		      float64 vy,
		      float64 vz,
		      int32   vc
		      );

#ifdef __cplusplus
}
#endif

#endif /*_LIBGCS_H_INCLUDED*/
