/*
 * $RCSfile: libgcs_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libgcs_local.h
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Driver               *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libgcs
*       Created: Tue Aug 29 17:36:21 EDT 1995
*       Author: buettner
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#include "libgcs.h"
#ifndef _LIBGCS_LOCAL_H_INCLUDED
#define _LIBGCS_LOCAL_H_INCLUDED
/* This define should be commented out if the orgin is to be put in the 
 * lower left hand corner of the cell.  Otherwise, if defined then the origin
 * is placed in the center of a cell
 */
#define GCS_ORIGIN_AT_CENTER 

/* Short-hand debugging macro (assumes local variable vehicle_id) */
#define DEBUG_GCS(format) \
  CLASS_DEBUG((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id), \
	      gcs_user_data_handle, format)

/***************************************************************************/

/* This is the data structure that will allow quick conversion from latitude
 * and longitude to cell id.
 */
#define GCS_MAX_GEOREF_IDS        60
#define GCS_GEOREF_ID_STRLEN      6
#define GCS_MAX_CELL              49779
#define GCS_MIN_CELL              1
/*$$$ could we set these items in the gcs init routine and not have them
 * as hardcoded defines
 */
typedef struct gcs_latmap
{
    int32   base_cell ;
    int32   n_cell    ;
    float32 last_lng  ;
} GCS_LATMAP;

typedef struct gcs_coord_data
{
    float64 matrix[3][3];
    float64 origin_in_gcc[XYZ];
    float64 x_off;
    float64 y_off;
} GCS_COORD_DATA;


    
typedef struct gcs_cell_data_type 
{
    ADDRESS             pbt_info;
    ADDRESS             tdb;
    ADDRESS             thinned_tdb;
    GCS_COORD_DATA     *coord;
} GCS_CELL_DATA_TYPE;


#ifndef PI
#define RAD_TO_DEG(x) ((x) * 57.295779513082323)
#define DEG_TO_RAD(x) ((x) * 0.017453292519943295)
#define PI            3.14159265358979323846
#endif

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif

/* gcs_simnet is a flag.  If operating under simnet mode, it is
 * set to 1.  When running under gcs single or multi cell mode,
 * it is set to 0.
 */
extern int32 gcs_simnet;

/* File    : gcs.util.c
 * Routine : gcs_lookup_cell_id_ll
 *
 * Converts a geodetic (lat/long) coordinate into a cell id 
 */
extern int32 
gcs_lookup_cell_id_ll (float64 lat,
		       float64 lng
		       );
 
/* File    : gcs_rdr.c
 * Routine : gcs_latmap_reader_read
 *
 * Generates the global latmat which is used to quickly convert geodetic
 * (lat/lng) to cell ids.
 */
extern void
gcs_generate_latmap( void );


/* File    : gcs_util.c
 * Routine : gcs_cell_to_georef_id
 *
 * This routine will return the World Geographic References that are contained
 * within a cell (Reference DMA TM 8358.1 Chapter 5 Geographic Coordinate 
 * References). 
 */
extern int32
gcs_cell_to_georef_id(int32 cell, 
		      int32 *n_georefs , 
		      char georefs[][GCS_GEOREF_ID_STRLEN]);

/* File    : gcs_util.c
 * Routine : gcs_ll_to_georef_id
 *
 * This routine will return the World Geographic References that are contained
 * within the lat/lng bounding box (Reference DMA TM 8358.1 Chapter 5 
 * Geographic Coordinate References). 
 */
extern int32
gcs_ll_to_georef_id(float64 sw_lat, 
		    float64 sw_lng, 
		    float64 ne_lat, 
		    float64 ne_lng,
		    int32 *n_georefs , 
		    char georefs[][GCS_GEOREF_ID_STRLEN]);

/* File   : gcs_access.c
 * Routine: gcs_get_cell_center
 *
 * This routine will return the lat/lng of the center of the cell.
 * 
 */
extern void
gcs_get_cell_center (int32    cell,
		     float64 *lat,
		     float64 *lng
		     );

/* File   : gcs_util.c
 * Routine: gcs_generate_gcs_gcc_tfm_data
 *
 * This routine will generate a gcs to gcc transformation matrix
 * 
 */
extern void
gcs_generate_gcs_gcc_tfm_data (int32    cell,
			       float64  lat,
			       float64  lng
			       );
/* File   : gcs_init.c
 * Routine: gcs_init_coord_data
 *
 * This routine will initialize the coordinate database which is used to 
 * perform coordinate conversions.
 * 
 */
extern void
gcs_init_coord_data (int32    cell,
		     float64  lat,
		     float64  lng
		     );

/* File    : gcs_access.c
 * Routine : gcs_set_origin_in_gcc
 *
 * Sets the origin of the cell in gcc to be location.
 */
extern int32
gcs_set_origin_in_gcc (int32 cell,
		       float64 location[XYZ]
		       );

/* File    : gcs_util.c
 * Routine : gcs_valid_cell
 *
 * This routine will return TRUE if 
 *   1. the cell is valid and the software is GCS aware or 
 *   2. if the software is GCS unaware then set the cell to GCS_ILLEGAL_CELL
 *      and process a valid cell.
 */ 			     
extern int32 
gcs_valid_cell ( int32 *cell );

/* File    : gcs_conv.c
 * Routine : gcs_get_gcc_datum_info
 *
 * This routine will return GCC/WGS84 datum information.  This routine will be
 * used when only WGS84 modeling is desired.
 */ 			     
extern void
gcs_get_gcc_datum_info (float64  lat,
			float64  lng,
			int32   *datum,
			float64 *major,
			float64 *minor,
			float64 *dx,
			float64 *dy,
			float64 *dz
			);
/* File    : gcs_conv.c
 * Routine : gcs_init_tables
 *
 * This routine initializes tables needed for gcc->llz conversion
 */ 			     
extern void 
gcs_init_tables();

#endif /*_LIBGCS_LOCAL_H_INCLUDED*/
