#ifndef lint
static char rcsid [] = "$RCSfile: geo_clip.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_clip.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Generic intersection utilities                            *
*       Created: Tue Mar 29 13:39:35 EST 1994                               *
*       Author: jtan                                                        *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>
#include <stdalloc.h>
#include <stdio.h>
#include <stdstring.h> /*common/include/global*/

typedef enum {POLY1, POLY2}  WHICH_POLY;

/* This function clips a geometric object1 with object2 */
int32 geo_clip_2_objects(
    float64 vertices1[][2],
    int32   num1,
    float64 vertices2[][2],
    int32   num2,
    float64 clipped[][2])
{
    GEO_POINT int1[200];
    GEO_POINT int2[200];
    float64 int_pt[2];
    int32   n_clipped_points=0;
    int32 n_ints;
    int32 i, j;
    int32 first_v1=0;
    int32 first_v2[2];
    int32 last_v1;
    int32 last_v2;
    
    if (!(n_ints = geo_intersection_2_objects(vertices1, num1, vertices2, num2,
					      0, int1, int2)))
      return(0);
    
    for (i=0; i<n_ints; i++)
    {
	if (!(i%2)) /* even, object1 entering object2 */
	{
	    /* Copy all points on object1 before this intersection to clipped*/
	    last_v1 = int1[i].vertex - 1 + ceil(int1[i].t);
 
	    if ((int2[i].t == 0.0) || int2[i].t == 1.0)
	      first_v2[0] = first_v2[1] = int2[i].vertex + int2[i].t;
	    else
	    {
		first_v2[0] = int2[i].vertex;
		first_v2[1] = int2[i].vertex+1;
	    }
	      
	    for (j=first_v1; j<=last_v1; j++)
	      VMAT2_VEC_COPY64(vertices1[j], clipped[n_clipped_points++]);
	    VMAT2_VEC_COPY64(int1[i].point, clipped[n_clipped_points++]);
	}
	else /* odd, object1 leaving object2 */
	{
	    /* Copy all points on object2 before this intersection to clipped*/
	    if (first_v2[0] < int2[i].vertex)
	      last_v2 = int2[i].vertex - 1 + ceil(int2[i].t);
	    else
	    {
		last_v2 = first_v2[0];
		first_v2[1] = int2[i].vertex + 1 + floor(int2[i].t);
	    }
		
	    first_v1 = int1[i].vertex + 1 + floor(int1[i].t);

	    for (j=first_v2[1]; j<=last_v2; j++)
	      VMAT2_VEC_COPY64(vertices2[j], clipped[n_clipped_points++]);
	    VMAT2_VEC_COPY64(int2[i].point, clipped[n_clipped_points++]);
	}
    }

    if (!(n_ints%2)) /* Copy the rest of the object1 to clipped */
      for (j=first_v1; j<num1; j++)
	VMAT2_VEC_COPY64(vertices1[j], clipped[n_clipped_points++]);

    return(n_clipped_points);
}


/* Check to see if a point is inside *or on* a polygon.
 */
static int32 is_point_in_polygon(
    float64 point[2],
    float64 verts[][2],
    int32   n_verts)
{
    return(geo_is_point_in_polygon(point, verts, n_verts) ||
	   !(geo_distance_sq_2_objects((float64 (*)[2])point, 1, verts,
				       n_verts + 1,
				       NULL, NULL) > GEO_SIGMA));
}


#define MAX_INTS         20
static int32 select_next_point(
    WHICH_POLY which_poly,
    float64    int_pt[2],
    float64    segment[2][2],
    float64    current_poly[][2],
    int32      n_verts)
{
    int32                inside0, inside1, n_ints0, n_ints1;
    float64              seg[2][2], dir[2];


    /* Is segment[0] inside `current_poly'?
     */
    inside0 = is_point_in_polygon(segment[0], current_poly, n_verts);

    /* Is segment[1] inside `current_poly'?
     */
    inside1 = is_point_in_polygon(segment[1], current_poly, n_verts);

    /* Count the number of intersections from `int_pt' to segment[0].
     * We do NOT want to count `int_pt', but we DO want to count `segment[0]'
     * if it lies on the other polygon.  geo_intersection_2_objects() does not
     * count points that lie on polygons, so the segment is lengthened
     * a smidge to account for this.
     */
    bcopy(int_pt, seg[0], 2 * sizeof(float64));
    bcopy(segment[0], seg[1], 2 * sizeof(float64));

    /* Lengthen seg[] by 1% on the seg[1] end.
     */
    VMAT2_VEC_SUB(seg[1], seg[0], dir);
    VMAT2_SCAL_VEC_MUL(0.01, dir, dir);
    VMAT2_VEC_ADD(seg[1], dir, seg[1]);

    n_ints0 = geo_intersection_2_objects(seg, 2, current_poly, n_verts + 1,
					 MAX_INTS, NULL, NULL);

    /* Count the number of intersections from `int_pt' to segment[1].
     */
    bcopy(int_pt, seg[0], 2 * sizeof(float64));
    bcopy(segment[1], seg[1], 2 * sizeof(float64));

    /* Lengthen seg[] by 1% on the seg[1] end.
     */
    VMAT2_VEC_SUB(seg[1], seg[0], dir);
    VMAT2_SCAL_VEC_MUL(0.01, dir, dir);
    VMAT2_VEC_ADD(seg[1], dir, seg[1]);

    n_ints1 = geo_intersection_2_objects(seg, 2, current_poly, n_verts + 1,
					 MAX_INTS, NULL, NULL);

    if(which_poly == POLY1)
    {
	if(inside0 && !inside1)
	  return 0;

	if(inside1 && !inside0)
	  return 1;

	if((n_ints0 % 2) && !(n_ints1 % 2))
	{
	    if(inside0 && inside1)
	      return 1;
	    else
	      return 0;
	}

	if((n_ints1 % 2) && !(n_ints0 % 2))
	{
	    if(inside0 && inside1)
	      return 0;
	    else
	      return 1;
	}
    }

    if(which_poly == POLY2)
    {
	if(!inside0 && inside1)
	  return 0;

	if(!inside1 && inside0)
	  return 1;

	if(!(n_ints0 % 2) && (n_ints1 % 2))
	{
	    if(inside0 && inside1)
	      return 1;
	    else
	      return 0;
	}

	if(!(n_ints1 % 2) && (n_ints0 % 2))
	{
	    if(inside0 && inside1)
	      return 0;
	    else
	      return 1;
	}
    }

    /* If we get here, we're very confused.
     */
    return -1;
}


static void dealloc_polys(
    float64 *polys[],
    int32    n_polys)
{
    int32                i;

    for(i = 0; i < n_polys; i++)
      STDDEALLOC(polys[i]);
}


/* Check a polygon against a list of polygons to see if it is unique.
 */
static int32 polygon_is_unique(
    float64  check_poly[][2],
    int32    n_check_poly_verts,
    float64 *poly_list[],
    int32    n_verts_list[],
    int32    n_polys)
{
    int32                i, j, k;

    for(i = 0; i < n_polys; i++)
    {
	if(n_verts_list[i] != n_check_poly_verts)
	  continue;

	/* Does the first point of the check polygon match any points of the
	 * current polygon?
	 */
	for(j = 0; j < n_verts_list[i]; j++)
	  if((fabs(check_poly[0][X]-*(poly_list[i]+2*j+X))<GEO_SIGMA
	      &&
	      fabs(check_poly[0][Y]-*(poly_list[i]+2*j+Y))<GEO_SIGMA))
	    break;

	/* If the point was not found in the current polygon, continue.
	 */
	if(j == n_verts_list[i])
	  continue;

	/* Compare every point in the check polygon to every point in the
	 * current polygon to see if they match.
	 */
	for(k = 0; k < n_check_poly_verts; k++)
	{
	    if(!((fabs(check_poly[k][X]-*(poly_list[i]+2*j+X))<GEO_SIGMA
		  &&
		  fabs(check_poly[k][Y]-*(poly_list[i]+2*j+Y))<GEO_SIGMA)))
	      break;

	    j++;
	    if(j == n_verts_list[i])
	      j = 0;
	}

	/* If all points match, then the check polygon is not unique.
	 */
	if(k == n_check_poly_verts)
	  return FALSE;
    }

    return TRUE;
}


/* Given a polygon (poly1) and a "cutting" polygon (poly2), this function
 * returns an array of pointers to polygons which are parts of `poly1' that
 * are outside of `poly2'.
 *
 * The algoritm basically works like this:
 *
 * - start tracing poly1 until it intersects with poly2
 * - switch to poly2, pick a direction, and trace poly2 until it intersects
 *   with poly1
 * - switch to poly1, pick a direction, and trace poly1 until it intersects
 *   with poly2
 * - do this until you get to the original starting point
 * 
 * This is done for all points on poly1 until they're all used.
 */
#define MAX_VERTS        1000
int32 geo_cut_polygon(
    float64  poly1[][2],
    int32    n_verts1,
    float64  poly2[][2],
    int32    n_verts2,
    float64 *cut_polys[],
    int32    n_verts_cut[],
    int32   *n_cut_polys)
{
    int32                i, vert, n_ints, point, first_pt_index;
    int32                max_cut_polys, next, used_vertices[MAX_VERTS + 1];
    int32                next_vert[2], n_verts[2], n_temp_poly_verts;
    int32                direction[2], n_points_to_process, point_count;
    int32                n_verts_inside;
    float64              points_to_process[MAX_INTS + 1][2];
    float64              temp_poly[MAX_VERTS + 1][2], t, min_t;
    float64              temp_int_pt[2];
    float64              current_pt[2], next_pt[2], starting_pt[2];
    float64             *verts[2], segment[2][2], dir[2];
    float64              int_pt[2];
    float64              closed_poly1[MAX_VERTS + 1][2];
    float64              closed_poly2[MAX_VERTS + 1][2];
    WHICH_POLY           curr_poly, other_poly;
    GEO_POINT            int_pt1[MAX_INTS], int_pt2[MAX_INTS];

    if(n_verts1 > MAX_VERTS || n_verts2 > MAX_VERTS)
    {
	printf("libgeometry: geo_cut_polygon(): MAX_VERTS exceeded\n");
	*n_cut_polys = 0;
	return 1;  /* error */
    }

    /* Create closed versions of the input polygons (if they're not closed
     * already).
     */
    verts[POLY1] = (float64 *)closed_poly1;
    n_verts[POLY1] = n_verts1;

    bcopy(poly1[0], closed_poly1[0], 2 * sizeof(float64) * n_verts1);
    if(!VMAT2_VEC_EQUAL64(poly1[0], poly1[n_verts1 - 1]))
      bcopy(poly1[0], closed_poly1[n_verts1], 2 * sizeof(float64));
    else
      n_verts[POLY1]--;

    verts[POLY2] = (float64 *)closed_poly2;
    n_verts[POLY2] = n_verts2;

    bcopy(poly2[0], closed_poly2[0], 2 * sizeof(float64) * n_verts2);
    if(!VMAT2_VEC_EQUAL64(poly2[0], poly2[n_verts2 - 1]))
      bcopy(poly2[0], closed_poly2[n_verts2], 2 * sizeof(float64));
    else
      n_verts[POLY2]--;

    /* If poly2 is completely inside of poly1, return with `n_cut_polys'
     * equal to zero.
     */
    n_verts_inside = 0;
    for(vert = 0; vert < n_verts2; vert++)
      if(is_point_in_polygon(verts[POLY2] + 2 * vert, 
			     (float64 (*)[2])verts[POLY1],
			     n_verts[POLY1]))
	n_verts_inside++;

    if(n_verts_inside == n_verts2)
    {
	*n_cut_polys = 0;
	return 0;
    }

    /* If poly1 does not intersect with poly2, return with `n_cut_polys'
     * equal to zero.
     */
    if(!geo_intersection_2_objects((float64 (*)[2])verts[POLY1],
				   n_verts[POLY1] + 1,
				   (float64 (*)[2])verts[POLY2],
				   n_verts[POLY2] + 1,
				   0, NULL, NULL))
    {
	*n_cut_polys = 0;
	return 0;
    }

    /* Initially mark all of the vertices of poly1 as unused.
     */
    for(i = 0; i < MAX_VERTS + 1; i++)
      used_vertices[i] = 0;

    max_cut_polys = *n_cut_polys;
    *n_cut_polys = 0;

    for(vert = 0; vert < n_verts[POLY1]; vert++)
    {
	n_points_to_process = 0;

	point_count = 0;

	/* Is this vertex outside of POLY2?
	 */
	if(!is_point_in_polygon(verts[POLY1] + 2 * vert,
				(float64 (*)[2])verts[POLY2],
				n_verts[POLY2]))
	{
	    if(!used_vertices[vert])
	      bcopy(verts[POLY1] + 2 * vert,
		    points_to_process[n_points_to_process++],
		    2 * sizeof(float64));
	    point_count++;
	}

	/* Find the number of intersections with POLY2 along this segment,
	 * and add every other one to the `points_to_process' list.
	 */
	n_ints = geo_intersection_2_objects((float64 (*)[2])(verts[POLY1] +
					    2 * vert), 2,
					    (float64 (*)[2])verts[POLY2],
					    n_verts[POLY2] + 1,
					    MAX_INTS, int_pt1, NULL);

	for(i = 0; i < n_ints; i++, point_count++)
	  if(!(point_count % 2))
	    bcopy(int_pt1[i].point, points_to_process[n_points_to_process++],
		  2 * sizeof(float64));

	for(point = 0; point < n_points_to_process; point++)
	{
	    direction[POLY1] = 1;
	    direction[POLY2] = 1;

	    next_vert[POLY1] = vert + 1;
	    next_vert[POLY2] = 0;

	    /* Begin "tracing" poly1.
	     */
	    curr_poly = POLY1;
	    other_poly = POLY2;

	    n_temp_poly_verts = 0;

	    /* Assign this as the current and starting point.
	     */
	    bcopy(points_to_process[point], current_pt, 2 * sizeof(float64));
	    bcopy(current_pt, starting_pt, 2 * sizeof(float64));

	    /* Assign the next point.
	     */
	    bcopy(verts[POLY1] + 2 * next_vert[POLY1], next_pt,
		  2 * sizeof(float64));

	    /* Mark the current point as being used.
	     */
	    used_vertices[vert] = 1;
	    if(vert == 0)
	      used_vertices[n_verts[POLY1]] = 1;

	    while(TRUE)
	    {
		/* Add the current point to the polygon that's being built.
		 */
		bcopy(current_pt, temp_poly[n_temp_poly_verts],
		      2 * sizeof(float64));
		if(n_temp_poly_verts++ == MAX_VERTS)
		{
#if 0
		    printf("libgeometry: geo_cut_polygon(): MAX_VERTS exceeded\n");
#endif
		    dealloc_polys(cut_polys, *n_cut_polys);
		    *n_cut_polys = 0;
		    return 1;  /* error */
		}

		/* Find the number of intersections that the segment from the
		 * current point to the next point has with the other polygon.
		 */
		bcopy(current_pt, segment[0], 2 * sizeof(float64));
		bcopy(next_pt, segment[1], 2 * sizeof(float64));

		n_ints =
		  geo_intersection_2_objects(segment, 2,
					     (float64 (*)[2])verts[other_poly],
					     n_verts[other_poly] + 1,
					     MAX_INTS, int_pt1,
					     int_pt2);

		/* Out of all the intersections, find the first one.
		 */
		min_t = 1.1;

		for(i = 0; i < n_ints; i++)
		  if(int_pt1[i].t < min_t)
		  {
		      first_pt_index = int_pt2[i].vertex;
		      bcopy(int_pt1[i].point, int_pt, 2 * sizeof(float64));
		      min_t = int_pt1[i].t;
		  }

		/* If there are no intersections, just add the next vertex to
		 * the polygon that is being built, and increment the next
		 * vertex.
		 */
		if(!n_ints)
		{
		    /* Mark the next vertex as "used".
		     */
		    if(curr_poly == POLY1)
		    {
			used_vertices[next_vert[curr_poly]] = 1;
			if(used_vertices[n_verts[curr_poly]])
			  used_vertices[0] = 1;
		    }

		    /* Update the next vertex.
		     */
		    next_vert[curr_poly] += direction[curr_poly];
		    if(next_vert[curr_poly] == n_verts[curr_poly] + 1)
		      next_vert[curr_poly] = 1;
		    else
		      if(next_vert[curr_poly] == n_verts[curr_poly])
			next_vert[curr_poly] = 0;
		      else
			if(next_vert[curr_poly] == -1)
			  next_vert[curr_poly] = n_verts[curr_poly] - 1;

		    /* Update `current_pt' and `next_pt'.
		     */
		    bcopy(next_pt, current_pt, 2 * sizeof(float64));
		    bcopy(verts[curr_poly] + 2 * next_vert[curr_poly], next_pt,
			  2 * sizeof(float64));
		}
		else
		{
		    /* Select which point on the intersected segment of the
		     * other polygon to process next.
		     */
		    bcopy(verts[other_poly] + 2 * first_pt_index,
			  segment[0], 2 * sizeof(float64));
		    bcopy(verts[other_poly] + 2 * (first_pt_index + 1),
			  segment[1], 2 * sizeof(float64));

		    next = select_next_point(curr_poly, int_pt, segment,
					     (float64 (*)[2])verts[curr_poly],
					     n_verts[curr_poly]);

		    if(next == -1)
		    {
			dealloc_polys(cut_polys, *n_cut_polys);
			*n_cut_polys = 0;
			return 1;  /* error */
		    }

		    /* Update the direction of tracing.
		     */
		    if(next == 0)
		      direction[other_poly] = -1;
		    else
		      direction[other_poly] = 1;

		    /* Update the next vertex.
		     */
		    next_vert[other_poly] = first_pt_index + next;
		    if(next_vert[other_poly] == n_verts[other_poly])
		      next_vert[other_poly] = 0;

		    /* Update `current_pt' and `next_pt'.
		     */
		    bcopy(int_pt, current_pt, 2 * sizeof(float64));
		    bcopy(verts[other_poly] + 2 * next_vert[other_poly],
			  next_pt, 2 * sizeof(float64));

		    /* Switch the tracing to the other polygon.
		     */
		    curr_poly = other_poly;
		    other_poly = curr_poly == POLY1 ? POLY2 : POLY1;
		}

		/* If we've come across the vertex we've started with,
		 * it's time to begin constructing the next polygon.
		 */
		if((fabs(current_pt[X] - starting_pt[X]) < GEO_SIGMA &&
		    fabs(current_pt[Y] - starting_pt[Y]) < GEO_SIGMA))
		  /* If the polygon is unique, then add it to our list.
		   * Otherwise, do nothing.
		   */
		  if(polygon_is_unique(temp_poly, n_temp_poly_verts,
				       cut_polys, n_verts_cut, *n_cut_polys))
		  {
		      if(!(cut_polys[*n_cut_polys] =
			   (float64 *)STDALLOC(2 * sizeof(float64) *
					       n_temp_poly_verts)))
		      {
			  printf("libgeometry: memory allocation error\n");
			  dealloc_polys(cut_polys, *n_cut_polys);
			  *n_cut_polys = 0;
			  return 1;  /* error */
		      }

		      bcopy(temp_poly, cut_polys[*n_cut_polys],
			    2 * sizeof(float64) * n_temp_poly_verts);

		      n_verts_cut[*n_cut_polys] = n_temp_poly_verts;

		      if(++(*n_cut_polys) == max_cut_polys)
			return 0;

		      break;
		  }
		  else
		    break;
	    }
	}
    }

    return 0;
}

