#ifndef lint
static char rcsid [] = "$RCSfile: geo_clone.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_clone.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Manages class information attached to each vehicle        *
*       Created: Fri Feb  4 11:29:35 EST 1994
*       Author: jtan
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>

#include <stdalloc.h> /*common/include/global*/
#include <stdstring.h> /*common/include/global*/
#include <stdio.h>

#if DEBUG_CLONE
static FILE *fout = 0;
#endif

/* Given the unit vectors AA' and BB', this route 
 * calculate the vector CC', which bisect the angle ABC. 
 * A'C' is paralle to AC, B'C' is to BC.
 *
 *   A'-------C'
 *             \
 *   A ----C    \
 *          \    \
 *           \    \
 *    indside \    B'
 *             B    
 */
static int32 find_corner_vector(
    float64 a[2],
    float64 b[2],
    float64 c[2])
{
    float64 f;
    
    VMAT2_VEC_ADD(a, b, c);
    if (VMAT2_VEC_MAG_SQ(c) < GEO_SIGMA)
    {
	/* bad corner, cannot found a corner vector. */
	return(FALSE);
    }
    else
    {
	f = VMAT2_VEC_MAG_SQ(a) / VMAT2_DOT_PROD(c, a);
	VMAT2_SCAL_VEC_MUL(f, c, c);
    }
    
    return(TRUE);
}

static int32 polygon_center(
    float64 points[][2],
    int32   num,
    float64 clone[][2])
{
    int32 i;
    
    clone[0][X] = clone[0][Y] = 0.0;
    for (i=0; i<num; i++)
      VMAT2_VEC_ADD(points[i], clone[0], clone[0]);
    VMAT2_SCAL_VEC_MUL(1.0/num, clone[0], clone[0]);
    return(1);
}


static int32 shrink_polygon(
    float64 points[][2],
    int32   num,
    int32   clockwise,
    float64 dist,
    int32   dir,
    float64 clone[][2])
{
    GEO_POLYGONS *first_poly, *ptr, *next, *prev, *min_poly;
    float64 temp[200][2];
    int32 n_temp, n, i;
    GEO_POINT int1, int2, min_int1, min_int2;
    FILE *fp;
    float64 d, min_dist;
    int32 min_i;
    
    first_poly = (GEO_POLYGONS *) STDALLOC(sizeof(GEO_POLYGONS));
    first_poly->n_pts = num;
    first_poly->next = NULL;
    first_poly->points = 
      (float64(*)[2]) STDALLOC(sizeof(float64)*2*first_poly->n_pts);
    bcopy((char *)points, (char *)first_poly->points,
	  sizeof(float64)*2*first_poly->n_pts);
    
    if (geo_get_polygons(first_poly) == 1)
      return(geo_clone_object(points, num, dist, dir, clone));
    
#if DEBUG_CLONE    
    fp = fopen("test.dat", "w");
    for (i=0; i< num; i++)
      fprintf(fp, "%lf %lf\n", points[i][X], points[i][Y]);
    fprintf(fp, "\n");
    fflush(fp);
    
    for (ptr=first_poly; ptr; ptr=ptr->next)
    {
	for (i=0; i< ptr->n_pts; i++)
	  fprintf(fp, "%lf %lf\n", ptr->points[i][X], ptr->points[i][Y]);
	fprintf(fp, "\n");
    }
    fflush(fp);
#endif

    /* Remove all sub-polygons which have flipped direction as the input
       polygon. Those sub-polygons are the results of over shrinking */
    ptr=first_poly; 
    prev = NULL;
    while (ptr)
    {
	next = ptr->next;
	if (clockwise != geo_is_polygon_clockwise(ptr->points, ptr->n_pts))
	{
	    if (ptr == first_poly)
	      first_poly = next;
	    else
	      prev->next = next;
	    STDDEALLOC(ptr->points);
	    STDDEALLOC(ptr);
	}
	else
	  prev = ptr;
	ptr = next;
    }
    
    if (!first_poly)
      return(polygon_center(points, num, clone));
    else if (!first_poly->next)
      return(geo_clone_object(first_poly->points, first_poly->n_pts, 
			      dist, dir, clone));
      
    /* Shrink each sub-polygon. */
    for (ptr = first_poly; ptr; ptr=ptr->next)
    {
	if (ptr->n_pts <= 3)
	  continue;
	ptr->n_pts = geo_clone_object(ptr->points, ptr->n_pts, dist, dir, 
				      temp);
	ptr->points = (float64 (*)[2])STDREALLOC(ptr->points,
						 sizeof(float64)*2*ptr->n_pts);
	bcopy((char *)temp, (char *)ptr->points, sizeof(float64)*2*ptr->n_pts);
    }

    /* Connect each sub-polygons */
    ptr = first_poly;
    bcopy((char *)ptr->points, (char *)temp, sizeof(float64)*2*ptr->n_pts);
    n_temp = ptr->n_pts;
    
    first_poly = ptr->next;
    STDDEALLOC(ptr->points);
    STDDEALLOC(ptr);

    while(first_poly)
    {
	min_dist = GEO_HUGE;
	for (ptr = first_poly; ptr; ptr=ptr->next)
	{
	    if ((d = geo_distance_sq_2_objects(temp, n_temp, ptr->points, 
					       ptr->n_pts, &int1, &int2))
		< min_dist)
	    {
		min_dist = d;
		min_int1 = int1;
		min_int2 = int2;
		min_poly = ptr;
	    }
	}
	
#if DEBUG_CLONE
	for (i=0; i< min_poly->n_pts; i++)
	  fprintf(fp, "%lf %lf\n", min_poly->points[i][X], 
		  min_poly->points[i][Y]);
	fprintf(fp, "\n");
	
	printf("int1=%d, %lf, %lf, %lf\n", min_int1.vertex, min_int1.t, 
	       min_int1.point[X], min_int1.point[Y]);
	printf("int2=%d, %lf, %lf, %lf\n", min_int2.vertex, min_int2.t, 
	       min_int2.point[X], min_int2.point[Y]);
#endif	    
	/* Connect with min_poly */
	n=0;
	for (i=0; i<=min_int1.vertex; i++)
	{
	    VMAT2_VEC_COPY64(temp[i], clone[n]);
	    n++;
	}
	
	if (min_int1.t > 0)
	{
	    VMAT2_VEC_COPY64(min_int1.point, clone[n]);
	    n++;
	}
	    
	VMAT2_VEC_COPY64(min_int2.point, clone[n]);
	n++;
	    
	if (min_int2.vertex < min_poly->n_pts-2)
	  for (i=min_int2.vertex+1; i <=min_poly->n_pts-2; i++)
	  {
	      VMAT2_VEC_COPY64(min_poly->points[i], clone[n]);
	      n++;
	  }
	
	for (i=0; i<=min_int2.vertex; i++)
	{
	    VMAT2_VEC_COPY64(min_poly->points[i], clone[n]);
	    n++;
	}
	if (min_int2.t > 0)
	{
	    VMAT2_VEC_COPY64(min_int2.point, clone[n]);
	    n++;
	}
	
	VMAT2_VEC_COPY64(min_int1.point, clone[n]);
	n++;
	
	if (min_int1.vertex < n_temp-2)
	  for (i=min_int1.vertex+1; i <= n_temp-2; i++)
	  {
	      VMAT2_VEC_COPY64(temp[i], clone[n]);
	      n++;
	  }
	VMAT2_VEC_COPY64(temp[0], clone[n]);
	n++;
	
	bcopy((char *)clone, (char *)temp, sizeof(float64)*2*n);
	n_temp = n;
#if DEBUG_CLONE	
	for (i=0; i< n; i++)
	  fprintf(fp, "%lf %lf\n", clone[i][X], clone[i][Y]);
	fprintf(fp, "\n");
	fflush(fp);
#endif
	for (prev=NULL,ptr = first_poly; ptr != min_poly; 
	     prev=ptr,ptr=ptr->next);
	if (!prev)
	  first_poly = ptr->next;
	else
	  prev->next = ptr->next;
	
	STDDEALLOC(min_poly->points);
	STDDEALLOC(min_poly);
    }

#if DEBUG_CLONE	
    fclose(fp);
#endif
    
    return(n);
}

	    
	    
/* Given a geometric object specified by 'num' points of 'vertices', 
 * geo_clone_object() creates a new geometric object from the input object by
 * moving each segment of the input object 'dist' distance towards the 
 * direction specified by 'dir', 
 * where, dir = 1 (right) or -1 (left), looking from the beginning of the 
 * segment towards the end of the segment.
 * For an input polygon object, geo_clone_object() will expend or shrink the 
 * polygon. The input 'dir' value of 1 (right) or -1 (left) is valid, but not
 * sufficient to specify whether to expend or shrink the polygon (because a
 * polygon can be clockwise or counter-clockwise). An input 
 * 'dir' value of 2 (expand) or -2 (shrink) can be used instead.
 */
int32 geo_clone_object(
    float64 vertices[][2],
    int32   num,
    float64 dist,
    int32   dir,
    float64 clone[][2])
{
    int32  i, j, n_temp;
    float64 tau, min_tau = 10.0;
    float64 norm[200][2], corner[200][2], temp[200][2];
    float64 seg[2], unit[2], int_pt[2], center[2];
    int32 clockwise = geo_is_polygon_clockwise(vertices, num);
    int32 is_polygon = VMAT2_VEC_EQUAL64(vertices[0], vertices[num-1]);
    int32 polygon_shrink = is_polygon && (dir == GEO_CLONE_OBJECT_SHRINK);
    
#if DEBUG_CLONE    
    if (!fout)
      fout = fopen("clone.dat", "w");
#endif    
    if (num > 200) printf("array too small\n");
    
    if (is_polygon && 
	(dir == GEO_CLONE_OBJECT_EXPAND) || (dir == GEO_CLONE_OBJECT_SHRINK))
    {
	if (clockwise)
	{
	    if (dir == GEO_CLONE_OBJECT_SHRINK)
	      dir = GEO_CLONE_OBJECT_RIGHT;
	    else
	      dir = GEO_CLONE_OBJECT_LEFT;
	}
	else
	{
	    if (dir == GEO_CLONE_OBJECT_SHRINK)
	      dir = GEO_CLONE_OBJECT_LEFT;
	    else
	      dir = GEO_CLONE_OBJECT_RIGHT;
	}
    }
    
    /* If a line, just return the center of the line. Depending on the
     * definition of shrinking, this may not be the right solution. Other 
     * possible ways to shrink a line are: (1) do nothing if we define 
     * shrinking as moving segment in its inward normal direction, (2) shorten
     * the line at each ends by desired distance if we define shrinking as 
     * moving vertices towards medial-axis direction, and (3) delete the line 
     * completely. 
     */
    if (polygon_shrink && (num <= 3))
      return(polygon_center(vertices, num, clone));
    
    /* First calculate the norm vector for each segment */
    for (i=0; i<num-1; i++)
    {
	VMAT2_VEC_SUB(vertices[i+1], vertices[i], seg);
	vmat2_unit64(seg, seg);
	norm[i][X] = dir*seg[Y];
	norm[i][Y] = -dir*seg[X];
    }
    
    /* 
     * Now calculate the corner vector for each corner
     */
    /* First point */
    if (is_polygon)
    {
	if (!find_corner_vector(norm[0], norm[num-2], corner[0]))
	{
	    /* bad point. remove it */
	    for (j=1; j<num; j++)
	      VMAT2_VEC_COPY64(vertices[j], temp[j-1]);
	    n_temp = num - 1;
	    VMAT2_VEC_COPY64(temp[0], temp[n_temp-1]);
	    return(geo_clone_object(temp, n_temp, dist, dir, clone));
	}
    }
    else
      VMAT2_VEC_COPY64(norm[0], corner[0]);
    /* Other points */
    for (i=1; i< num-1; i++)
    {
	if (!find_corner_vector(norm[i-1], norm[i], corner[i]))
	{
	    /* bad point. remove it */
	    for (j=0; j<i; j++)
	      VMAT2_VEC_COPY64(vertices[j], temp[j]);
	    for (j=i+1; j<num; j++)
	      VMAT2_VEC_COPY64(vertices[j], temp[j-1]);
	    n_temp = num - 1;
	    return(geo_clone_object(temp, n_temp, dist, dir, clone));
	}
    }
    /* Last point */
    if (is_polygon)
      VMAT2_VEC_COPY64(corner[0], corner[num-1]);
    else
      VMAT2_VEC_COPY64(norm[num-2], corner[i]);
    
    /* Finally, construct new object by moving vertices */
    for (i=0; i<num-1; i++)
    {
	VMAT2_SCAL_VEC_MUL(dist, corner[i], seg);
	VMAT2_VEC_ADD(seg, vertices[i], temp[i]);
    }
    if (is_polygon)
    {
	VMAT2_VEC_COPY64(temp[0], temp[num-1]);
    }
    else
    {
	VMAT2_SCAL_VEC_MUL(dist, corner[num-1], seg);
	VMAT2_VEC_ADD(seg, vertices[num-1], temp[num-1]);
    }

    /* Now we need to check whether we have over shrunk the polygon, the same
       check is needed for non-polygon too.
       The idea is to test whether each pair of corner vectors intesected with 
       each other. */
    for (i=0; i<num-1; i++)
    {
	if (geo_intersection_2_linesegs(vertices[i], temp[i], 
					vertices[i+1], temp[i+1], 
					NULL, &tau, NULL))
	{
	    if (min_tau > tau)
	      min_tau = tau;
	}
    }	
    
    if (min_tau > 2.0) /* tau's range is 0-1 */
    {
	if (clockwise == geo_is_polygon_clockwise(temp, num))
	{
	    /* We didn't over shrank */
	    bcopy((char *)temp, (char *)clone, sizeof(float64)*2*num);
	    return(num);
	}
	else
	{
	    /* We over shrank it way too much */
	    return(polygon_center(temp, num, clone));
	}
    }
    
    /* We over shrunk. Redo it, shrink only what is allowed first. */
    for (i=0; i<num-1; i++)
    {
	VMAT2_SCAL_VEC_MUL(min_tau*dist, corner[i], seg);
	VMAT2_VEC_ADD(seg, vertices[i], temp[i]);
    }
    if (is_polygon)
    {
	VMAT2_VEC_COPY64(temp[0], temp[num-1]);
    }
    else
    {
	VMAT2_SCAL_VEC_MUL(min_tau*dist, corner[num-1], seg);
	VMAT2_VEC_ADD(seg, vertices[num-1], temp[num-1]);
    }
    
    /* At least one segment should shrunk to almost zero length, find it and 
       remove the second point */
    n_temp = num;
    for (i=0; i<n_temp-1; i++)
      if (geo_distance_sq_2_points(temp[i], temp[i+1], 2) < GEO_SIGMA)
      {
	  if (i<n_temp-2)
	  {
	      bcopy((char *)temp[i+2], (char *)temp[i+1], 
		    sizeof(float64)*2*(n_temp-(i+2)));
	      i--; /* recheck it */
	  }
	  else
	  {
	      VMAT2_VEC_COPY64(temp[n_temp-1], temp[n_temp-2]);
	  }
	  n_temp--;
      }
    
    if ((is_polygon && (n_temp <= 3)) || (n_temp <= 2))
    {
	bcopy((char *)temp, (char *)clone, sizeof(float64)*2*n_temp);
	return(n_temp);
    }

#if DEBUG_CLONE
    printf("shrank dist = %lf, n_temp=%d\n", min_tau*dist, n_temp);
    
    for (i=0; i<n_temp; i++)
      fprintf(fout, "%lf %lf\n", temp[i][0], temp[i][1]);
    fprintf(fout, "\n");
    fflush(fout);
    
    system("gnuplot -geometry 600x400+0+0 clone.cmd");
#endif    
    /* If it is a polygon shrink operation, we need to do more checking */
    if (polygon_shrink)
      return(shrink_polygon(temp, n_temp, clockwise, (1.0-min_tau)*dist, dir, 
			    clone));
    
    /* Now we shrink the rest of the distance */
    return (geo_clone_object(temp, n_temp, (1.0-min_tau)*dist, dir, clone));
}
