#ifndef lint
static char rcsid [] = "$RCSfile: geo_distance.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_distance.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Manages class information attached to each vehicle        *
*       Created: Fri Feb  4 11:29:35 EST 1994
*       Author: jtan
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>

#include <stdstring.h> /*common/include/global*/


/* Calculates the distance square between two points (point1, point2) of <n> 
   dimension.
 */
float64 geo_distance_sq_2_points(
    float64 point1[],
    float64 point2[],
    int32   n)
{
    int32 i;
    float64 temp;
    float64 dist2 = 0;
    
    for (i = 0; i < n; i++)
    {
	temp = (point1[i] - point2[i]);
	dist2 += temp * temp;
    }
    return(dist2);
}

/* Calculates the distance between two points (point1, point2) of <n> 
   dimension.
 */
float64 geo_distance_2_points(
    float64 point1[],
    float64 point2[],
    int32   n)
{
    float64 distance;

    distance = sqrt(geo_distance_sq_2_points(point1, point2, n));
    return(distance);
}

/* Calculates the distance square between a point and a line segment */
float64 geo_distance_sq_point_to_segment(
    float64 seg_start[2],
    float64 seg_end[2],
    float64 point[2],
    float64 seg_inter[2])
{
    float64 start_to_end[2];
    float64 start_to_point[2];
    float64 point_to_end[2];

    VMAT2_VEC_SUB(seg_end, seg_start, start_to_end);
    VMAT2_VEC_SUB(point, seg_start, start_to_point);
    VMAT2_VEC_SUB(seg_end, point, point_to_end);
    
    if (VMAT2_DOT_PROD(start_to_end, start_to_point) < 0.0)
    {
	VMAT2_VEC_COPY64(seg_start, seg_inter);
	return(geo_distance_sq_2_points(seg_start, point, 2));
    }
    else if (VMAT2_DOT_PROD(start_to_end, point_to_end) < 0.0)
    {
	VMAT2_VEC_COPY64(seg_end, seg_inter);
	return(geo_distance_sq_2_points(seg_end, point, 2));
    }
    else
      return(geo_distance_sq_point_to_line(seg_start, seg_end, point, 
					   seg_inter));
}

/* Calculates the distance square between a point and a line */
float64 geo_distance_sq_point_to_line(
    float64 line_pnt1[2],
    float64 line_pnt2[2],
    float64 point[2],
    float64 line_inter[2])
{
    float64 to_pnt2[2];
    float64 to_point[2];
    float64 to_inter[2];
    
    VMAT2_VEC_SUB(line_pnt2, line_pnt1, to_pnt2);
    VMAT2_VEC_SUB(point, line_pnt1, to_point);
    vmat2_project64(to_point, to_pnt2, to_inter);
    vmat2_vec_add64(to_inter, line_pnt1, line_inter);
    return(geo_distance_sq_2_points(point, line_inter, 2));
}

/* Calculates the distance squared between a point and a geometric object */
float64 geo_distance_sq_point_to_object(
    float64    vertices[][2],
    int32      num,
    float64    point[2],
    GEO_POINT *pt)
{
    int32 i;
    float64 dist2 = GEO_HUGE; 
    float64 tmp_dist2;
    float64 inter[2];
    
    if (pt)
      bzero((char *)pt, sizeof(GEO_POINT));

    for (i=0; i<num-1; i++)
      if ((tmp_dist2 = geo_distance_sq_point_to_segment(vertices[i], 
							vertices[i+1],
							point, inter)) 
	  < dist2)
      {
	  dist2 = tmp_dist2;
	  if (pt)
	  {
	      pt->vertex = i;
	      VMAT2_VEC_COPY64(inter, pt->point);
	  }
      }
    
    if (pt)
    {
	i = pt->vertex;
	if (fabs(vertices[i][X]-vertices[i+1][X]) > 
	    fabs(vertices[i][Y]-vertices[i+1][Y]))
	  pt->t = (pt->point[X] - vertices[i][X]) / 
	    (vertices[i+1][X] - vertices[i][X]);
	else
	  pt->t = (pt->point[Y] - vertices[i][Y]) / 
	    (vertices[i+1][Y] - vertices[i][Y]);
	if (fabs(pt->t - 1.0) < GEO_SIGMA)
	  pt->t = 1.0;
	else if (fabs(pt->t) < GEO_SIGMA)
	  pt->t = 0.0;

	if (pt->t == 1.0)
	{
	    pt->t = 0.0;
	    pt->vertex++;
	}
    }
    
    return(dist2);
}

    
static float64 geo_distance_sq_2_segments(
    float64    seg1[2][2],
    float64    seg2[2][2],
    GEO_POINT *p1,
    GEO_POINT *p2)
{
    float64 int_pt[2];
    float64 tmpd, d_min=GEO_HUGE;
    float64 t1, t2;
    int32 i;
    
    if (geo_intersection_2_linesegs(seg1[0], seg1[1], seg2[0], seg2[1], 
				    int_pt, &t1, &t2))
    {
	VMAT2_VEC_COPY64(int_pt, p1->point);
	p1->vertex = 0;
	p1->t = t1;
	VMAT2_VEC_COPY64(int_pt, p2->point);
	p2->vertex = 0;
	p2->t = t2;
	return(0.0);
    }
    
    for (i=0; i<2; i++)
      if ((tmpd = geo_distance_sq_point_to_segment(seg2[0], seg2[1], seg1[i], 
						   int_pt)) < d_min)
      {
	  d_min = tmpd;
	  VMAT2_VEC_COPY64(seg1[i], p1->point);
	  VMAT2_VEC_COPY64(int_pt, p2->point);
      }
    
    for (i=0; i<2; i++)
      if ((tmpd = geo_distance_sq_point_to_segment(seg1[0], seg1[1], seg2[i], 
						   int_pt)) < d_min)
      {
	  d_min = tmpd;
	  VMAT2_VEC_COPY64(int_pt, p1->point);
	  VMAT2_VEC_COPY64(seg2[i], p2->point);
      }
    
    if (fabs(seg1[0][X]-seg1[1][X]) > fabs(seg1[0][Y]-seg1[1][Y]))
      p1->t = (p1->point[X] - seg1[0][X]) / (seg1[1][X] - seg1[0][X]);
    else
      p1->t = (p1->point[Y] - seg1[0][Y]) / (seg1[1][Y] - seg1[0][Y]);
    if (fabs(p1->t - 1.0) < GEO_SIGMA)
      p1->t = 1.0;
    else if (fabs(p1->t) < GEO_SIGMA)
      p1->t = 0.0;

    if (fabs(seg2[0][X]-seg2[1][X]) > fabs(seg2[0][Y]-seg2[1][Y]))
      p2->t = (p2->point[X] - seg2[0][X]) / (seg2[1][X] - seg2[0][X]);
    else
      p2->t = (p2->point[Y] - seg2[0][Y]) / (seg2[1][Y] - seg2[0][Y]);
    if (fabs(p2->t - 1.0) < GEO_SIGMA)
      p2->t = 1.0;
    else if (fabs(p2->t) < GEO_SIGMA)
      p2->t = 0.0;

    return(d_min);
}


/*
 * Given two separated geometry objects, represented by two arrays of 
 * vertices. This routine returns the separation square between the two, 
 * and the two closest points.
 */
float64 geo_distance_sq_2_objects(
    float64    vertices1[][2],
    int32      num1,
    float64    vertices2[][2],
    int32      num2,
    GEO_POINT *point1,
    GEO_POINT *point2)
{
    int32 i,j;
    GEO_POINT p1, p2;
    float64 dist2 = GEO_HUGE; 
    float64 tmp_dist2;

    if ((num1 == 1) && (num2 != 1))
    {
	dist2 = geo_distance_sq_point_to_object(vertices2, num2, vertices1[0],
						point2);
	if (point1)
	{
	    VMAT2_VEC_COPY64(vertices1[0], point1->point);
	    point1->vertex = 0;
	    point1->t = 0.0;
	}
	return(dist2);
    }
    else if ((num1 != 1) && (num2 == 1))
    {
	dist2 = geo_distance_sq_point_to_object(vertices1, num1, vertices2[0],
						point1);
	if (point2)
	{
	    VMAT2_VEC_COPY64(vertices2[0], point2->point);
	    point2->vertex = 0;
	    point2->t = 0.0;
	}
	return(dist2);
    }
    else if ((num1 == 1) && (num2 == 1))
    {
	dist2 = geo_distance_sq_2_points(vertices1[0], vertices2[0], 2);
	if (point1)
	{
	    VMAT2_VEC_COPY64(vertices1[0], point1->point);
	    point1->vertex = 0;
	    point1->t = 0.0;
	}
	if (point2)
	{
	    VMAT2_VEC_COPY64(vertices2[0], point2->point);
	    point2->vertex = 0;
	    point2->t = 0.0;
	}
	return(dist2);
    }	
    
    for (i=0; i<num1-1; i++)
      for (j=0; j<num2-1; j++)
	if ((tmp_dist2=geo_distance_sq_2_segments((float64 (*)[2])vertices1[i],
						  (float64 (*)[2])vertices2[j],
						  &p1, &p2)) < dist2)
	{
	    dist2 = tmp_dist2;
	    if (point1)
	    {
		VMAT2_VEC_COPY64(p1.point, point1->point);
		point1->t = p1.t;
		point1->vertex = i;
	    }
	    if (point2)
	    {
		VMAT2_VEC_COPY64(p2.point, point2->point);
		point2->t = p2.t;
		point2->vertex = j;
	    }
	}
    
    if (point1 && (point1->t == 1.0))
    {
	point1->t = 0.0;
	point1->vertex++;
    }
    if (point2 && (point2->t == 1.0))
    {
	point2->t = 0.0;
	point2->vertex++;
    }
    return(dist2);
}

	    
