#ifndef lint
static char rcsid [] = "$RCSfile: geo_intersect.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_intersect.c                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Generic intersection utilities                            *
*       Created: Tue Mar 29 13:39:35 EST 1994                               *
*       Author: jtan                                                        *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>
#include <stdalloc.h>
#include <stdstring.h> /*common/include/global*/
#include <assert.h>
#include <stdlib.h>


#define MAXIMUM(x,y) (x)>(y) ? (x) : (y)

/* Cross-product of the line-segment from s0 to e0,
 * with the line from s1 to e1.  (signed magnitude)
 */
#define CROSS_PRODUCT(s0,e0,s1,e1)                                       \
( ( (e0)[0] - (s0)[0] ) * ( (e1)[1] - (s1)[1] )                          \
- ( (e0)[1] - (s0)[1] ) * ( (e1)[0] - (s1)[0] ) )

#define MAX_POLYGON_VERTICES 100
#define NUM_POLYGONS           2

typedef enum {VECTOR_ALONG_INCOMING_EDGE,
	      VECTOR_INSIDE_ANGLE,
	      VECTOR_ALONG_OUTGOING_EDGE,
	      VECTOR_OUTSIDE_ANGLE} ANGLE_RELATIONSHIP;

static GEO_2D_POINT  *geo_polygon[NUM_POLYGONS];
static int32          geo_num_vertices[NUM_POLYGONS];
static int32          geo_direction[NUM_POLYGONS];  /* direction of traversal*/

/* This is the type of the found point (both may be true): */
static int32          geo_point_is_intersection;
static int32          geo_point_is_vertex;

/* If geo_point_is_intersection, these are the start-points of the two
 * edges that intersect, given as indexes into their respective polygons: */
static int32        geo_intersection_edge_id[NUM_POLYGONS];

/* If geo_point_is_intersection, this is the polygon we followed to get to it */
static int32        geo_incoming_polygon;

/* If geo_point_is_vertex, this is the polygon containing the vertex: */
static int32        geo_vertex_polygon_id;
/* If geo_point_is_vertex, this is the vertex-number of the vertex: */
static int32        geo_vertex_vertex;
static int32        geo_available_vertices[NUM_POLYGONS][MAX_POLYGON_VERTICES];

static int32 just_a_touch(
    float64 p1[][2],
    int32   num1,
    int32   open1,
    float64 p2[][2],
    int32   num2,
    int32   open2,
    int32   v1,
    int32   v2,
    int32   v2num);
static int32 from_inside(
    float64  start_point[2],
    float64  end_point[2],
    int32    poly_open,
    int32    num_poly_vertices,
    float64  poly_vertices[][2],
    int32    which_edge);
static ANGLE_RELATIONSHIP compare_vector_to_angle(
    float64 vector[2],
    float64 angle_incoming[2],
    float64 angle_outgoing[2]);
static int32 choose_outgoing_edge(
    int32    incoming_polygon,
    float64 *prior_point[NUM_POLYGONS],
    float64  point[2],
    float64 *next_point[NUM_POLYGONS],
    int32   *which_polygon);
static int32 find_next_subtract_point(
    float64  point[2],
    int32    which_vertex[NUM_POLYGONS],
    float64  threshold);
static int32 find_first_intersection_edge_with_other_polygon(
    int32    edge_polygon_id,
    int32    edge_initial_vertex,
    float64  start_point[2],
    float64  new_point[2],
    int32   *new_initial_vertex,
    float64 *tau0,
    float64 *tau1);
static int32 find_start_subtract_point(
    float64  point[2],
    int32    which_vertex[NUM_POLYGONS],
    float64  threshold);
static int32 geo_polygon_contains_polygon(
    float64 outer_poly_vertices[][XY],
    int32   outer_poly_num_vertices,
    float64 inner_poly_vertices[][XY],
    int32   inner_poly_num_vertices);
static void convert_polygon(
    float64 vertices[][XY],
    int32   num_vertices,
    float64 threshold,
    int32   which_polygon);


/* This function finds the intersection of two line segments. */
int32 geo_intersection_2_linesegs(
    float64  s0[2],
    float64  e0[2],
    float64  s1[2],
    float64  e1[2],
    float64  int_pt[2],
    float64 *tau0,
    float64 *tau1)
{
    float64 a0, b0, c0, d0;
    float64 a1, b1, c1, d1;
    float64 numer, denom;
    float64 t0, t1;

    /* We do this using parametric equations. */

    a0 = e0[X] - s0[X];  b0 = s0[X];
    c0 = e0[Y] - s0[Y];  d0 = s0[Y];
    a1 = e1[X] - s1[X];  b1 = s1[X];
    c1 = e1[Y] - s1[Y];  d1 = s1[Y];

    numer = c0 * (b1 - b0) + a0 * (d0 - d1);
    denom = a0 * c1 - a1 * c0;

    if ((fabs(numer) < GEO_SIGMA * GEO_SIGMA) && 
	(fabs(denom) < GEO_SIGMA * GEO_SIGMA))
      return GEO_SEG_DISJOINTorCOLINEAR;
    if (denom == 0.0) return GEO_SEG_DISJOINTorCOLINEAR;
    t1 = numer / denom;

    if (fabs(t1) < GEO_SIGMA)
      t1 = 0.0;
    else if (fabs(t1 - 1.0) < GEO_SIGMA)
      t1 = 1.0;
      
    /* Are we within the bounds of the second segment? */
    if (t1 >= 0.0 && t1 <= 1.0)
    {
	/* Derive t0 from t1 using either the Y or the y equation.
	 */
	if (fabs(a0) > GEO_SIGMA)
	  t0 = (b1 - b0 + a1 * t1) / a0;
	else
	  t0 = (d1 - d0 + c1 * t1) / c0;

	if (fabs(t0) < GEO_SIGMA)
	  t0 = 0.0;
	else if (fabs(t0 - 1.0) < GEO_SIGMA)
	  t0 = 1.0;

	/* Are we within the bounds of the first segment? */
	if (t0 >= 0.0 && t0 <= 1.0)
	{

	    if(int_pt)
	    {
		int_pt[X] = (e0[X] - s0[X]) * t0 + s0[X];
		int_pt[Y] = (e0[Y] - s0[Y]) * t0 + s0[Y];
	    }
	    if (tau0) *tau0 = t0;
	    if (tau1) *tau1 = t1;
	    return GEO_SEG_INTERSECT;
	}
    }
    
    return GEO_SEG_DISJOINTorCOLINEAR;
}

typedef struct geo_points
{
    GEO_POINT pt1;
    GEO_POINT pt2;
} GEO_POINTS;

static int compare_index(
    const void *a1,
    const void *a2)
{
    GEO_POINTS *geop1 = (GEO_POINTS *)a1;
    GEO_POINTS *geop2 = (GEO_POINTS *)a2;
    float64 index;
    
    index = (geop1->pt1.vertex + geop1->pt1.t) - 
      (geop2->pt1.vertex + geop2->pt1.t);
    return (index > 0 ? 1 : -1);
}


/* This function returns TRUE if any points of either polygon are inside
 * the other, FALSE otherwise.
 * If such a point is found, this function returns the found point as follows:
 * which_polygopn is 0 if the point is a vertex of vertices0, 
 * and 1 if the point is a vertex of vertices1.  which_vertex is the index 
 * of the vertex, in either case.
 * NOTE: The returned point, if any, is not guaranteed to be the first,
 * the last, or any other such special point.
 */
int32 geo_polygons_overlap(
    float64 vertices0[][2],
    int32   num0,
    float64 vertices1[][2],
    int32   num1,
    int32  *which_polygon,
    int32  *which_vertex)
{
    int32       i;
    int32       limit;

    limit = MAXIMUM(num0, num1);
    for (i = 0; i < limit; ++i)
    {
        if (i < num0)
	{
	    if ( geo_is_point_in_polygon(vertices0[i], vertices1, num1)
		== GEO_PT_INSIDE)
	    {
	        if (which_vertex)
		  *which_vertex = i;
		if (which_polygon)
		  *which_polygon = 0;
		return TRUE;
	    }
	}
        if (i < num1)
	{
	    if ( geo_is_point_in_polygon(vertices1[i], vertices0, num0)
		== GEO_PT_INSIDE)
	    {
	        if (which_vertex)
		  *which_vertex = i;
		if (which_polygon)
		  *which_polygon = 1;
		return TRUE;
	    }
	}
    }
    return FALSE;
}


/* This function finds a maximum of n_ints intersections of two geometric 
 * objects
 */

int32 geo_intersection_2_objects(
    float64   vertices1[][2],
    int32     num1,
    float64   vertices2[][2],
    int32     num2,
    int32     n_ints,
    GEO_POINT int1[],
    GEO_POINT int2[])
{
    float64      int_pt[2];
    int32        n_intsect=0;
    int32        i, j;
    char       **skip;
    float64      t1;
    float64      t2;
    GEO_POINT    tmp1, tmp2;
    GEO_POINTS   geop[100];
    int32        is_inside;
    
    int32 obj1_open = !((num1>2) && 
			VMAT2_VEC_EQUAL64(vertices1[0], vertices1[num1-1]));
    int32 obj2_open = !((num2>2) &&
			VMAT2_VEC_EQUAL64(vertices2[0], vertices2[num2-1]));

    if (n_ints == 0) n_ints = 10000;
    
    skip = (char **) STDALLOC(num1 * sizeof(char*));
    for (i=0; i<num1; i++)
    {
	skip[i] = (char *) STDALLOC(num2 * sizeof(char));
	bzero((char *) skip[i], num2 * sizeof(char));
    }

    for(i=0; (i<num1-1) && (n_intsect < n_ints); i++)
      for(j=0; j<num2-1; j++)
      {
	  if (!skip[i][j] && 
	      (geo_intersection_2_linesegs(vertices1[i], vertices1[i+1], 
					   vertices2[j], vertices2[j+1], 
					   int_pt, &t1, &t2)
	       == GEO_SEG_INTERSECT))
	  {
	      /* Exclude the case where the 1st point touches the other obj */
	      if (t1 == 0.0 || t2 == 0.0)
		continue ;
	      /* If an open polygon, exclude the case where the last point
		 touches the other object too */
	      if ((obj1_open && (t1 == 1.0) && (i == num1-2)) || 
		  (obj2_open && (t2 == 1.0) && (j == num2-2)))
		continue ;
		  
	      if ((t1 == 1.0) && (t2 == 1.0))    /* vertex-vertex  case */
	      {
		  skip[i+1][j] = 1;
		  skip[i+1][j+1] = 1;
		  skip[i][j+1] = 1;
		  if ( !from_inside(vertices1[i], vertices1[i+1], 
				    obj2_open, num2, vertices2, j)
		       &&
		       !from_inside(vertices2[j], vertices2[j+1], 
				    obj1_open, num1, vertices1, i))	
		  {
		      if (just_a_touch(vertices1, num1, obj1_open,
				       vertices2, num2, obj2_open, i+1, j, 3))
			continue ;
		  }
	      }
	      else if ((t1 == 1.0) && (t2 < 1.0))   /* vertex-line case */
	      {
		  skip[i+1][j] = 1;
		  if( !from_inside(vertices1[i], vertices1[i+1], 
				   obj2_open, num2, vertices2, j))
		  {
		      if (just_a_touch(vertices1, num1, obj1_open,
				       vertices2, num2, obj2_open, i+1, j, 2))
			continue ;
		  }
	      }
	      else if ((t1 < 1.0) && (t2 == 1.0))   /* line-vertex case */
	      {
		  skip[i][j+1] = 1;
		  if( !from_inside(vertices2[j], vertices2[j+1], 
				    obj1_open, num1, vertices1, i))	
		  {
		      if (just_a_touch(vertices2, num2, obj2_open,
				       vertices1, num1, obj1_open, j+1, i, 2))
			continue ;
		  }
	      }
	      
	      VMAT2_VEC_COPY64(int_pt, geop[n_intsect].pt1.point);
	      if (t1 < 1.0)
	      {
		  geop[n_intsect].pt1.vertex = i;
		  geop[n_intsect].pt1.t = t1;
	      }
	      else
	      {
		  geop[n_intsect].pt1.vertex = i+1;
		  geop[n_intsect].pt1.t = 0.0;
		  if (!obj1_open && (geop[n_intsect].pt1.vertex == num1-1))
		    geop[n_intsect].pt1.vertex = 0;
	      }

	      VMAT2_VEC_COPY64(int_pt, geop[n_intsect].pt2.point);
	      if (t2 < 1.0)
	      {
		  geop[n_intsect].pt2.vertex = j;
		  geop[n_intsect].pt2.t = t2;
	      }
	      else
	      {
		  geop[n_intsect].pt2.vertex = j+1;
		  geop[n_intsect].pt2.t = 0.0;
		  if (!obj2_open && (geop[n_intsect].pt2.vertex == num2-1))
		    geop[n_intsect].pt2.vertex = 0;
	      }
	      if (++n_intsect >= n_ints) break;
	  }
      }

    /* sort ints array according to the direction of the first object */
    qsort(geop, n_intsect, sizeof(GEO_POINTS), compare_index);

    if (int1 || int2)
    {
	for (i=0; i<n_intsect; i++)
	{
	    if (int1) 
	      int1[i] = geop[i].pt1;
	    if (int2) 
	      int2[i] = geop[i].pt2;
	}
    }
    
    for (i=0; i<num1; i++)
      STDDEALLOC(skip[i]);
    STDDEALLOC(skip);

    return(n_intsect);
}

/* Check whether the line-segment from start_point to end_point comes from
 * inside the polygon "poly".
 */
static int32 from_inside(
    float64  start_point[2],
    float64  end_point[2],
    int32    poly_open,
    int32    num_poly_vertices,
    float64  poly_vertices[][2],
    int32    which_edge)
{
    int32       poly_start_vertex;
    int32       poly_end_vertex;
    float64     cross_product;

    if (poly_open)      /* an open curve has no inside */
      return FALSE;
    
    poly_start_vertex = which_edge % num_poly_vertices;
    poly_end_vertex = (which_edge+1) % num_poly_vertices;
    cross_product = CROSS_PRODUCT(start_point, 
				  end_point, 
				  poly_vertices[poly_start_vertex], 
				  poly_vertices[poly_end_vertex]);
    return cross_product < 0;
}



static int32 next_vertex(
    int32 n_vertices,
    int32 is_polygon,
    int32 this_vertex)
{
    if (is_polygon)
      return((this_vertex+1)%(n_vertices-1));
    else
      return((this_vertex >= n_vertices-1) ? -1 : this_vertex+1);
}

/* To test whether the two objects just touch each other, like in >| or ><,
 * we first move the vertex of the first object 1% of the length towards the 
 * start of the segment, then check to see whether the two object intersects
 * each other.
 */
static int32 just_a_touch(
    float64 p1[][2],
    int32   num1,
    int32   open1,
    float64 p2[][2],
    int32   num2,
    int32   open2,
    int32   v1,
    int32   v2,
    int32   v2num)
{
    float64 vertices1[4][2];
    float64 vertices2[3][2];
    float64 vec1[2], vec2[2];
    int32   i, next;
    int32   n1=0;
    int32   n2=0;
    
    VMAT2_VEC_COPY64(p1[v1-1], vertices1[n1++]);
    VMAT2_VEC_SUB(p1[v1-1], p1[v1], vec1);
    vmat2_unit64(vec1, vec1);
    VMAT2_SCAL_VEC_MUL(0.01, vec1, vec1);
    VMAT2_VEC_ADD(p1[v1], vec1, vertices1[n1]);
    n1++;
    if ((next = next_vertex(num1, !open1, v1)) >= 0)
    {
	VMAT2_VEC_COPY64(p1[next], vertices1[n1++]);
    
	/* We need to add one more in case the next segment (between v1 and 
	 *  v1+1) is colinear with the p2 line:
	 *           |
	 * (v1-1)*---* (v1)
	 *           |
	 *           |
	 *    (v1+1) *-----* (v1+2)
	 *           |
	 *
	 * However, we will fail if the segment between v1+1 and v1+2 is also
	 * colinear with the p2 line!
	 */
	if ((next = next_vertex(num1, !open1, next)) >= 0)
	  VMAT2_VEC_COPY64(p1[next], vertices1[n1++]);
    }
    
    for (i=0; i<v2num-1; i++)
      VMAT2_VEC_COPY64(p2[v2+i], vertices2[i]);
    n2 = v2num-1;
    
    if ((next = next_vertex(num2, !open2, v2+v2num-2)) >= 0)
      VMAT2_VEC_COPY64(p2[next], vertices2[n2++]);
	
    return(!geo_intersection_2_objects(vertices1, n1, vertices2, n2, 0, 0,0));
}


/* Determine if two points are about the same */
static int32 same_point(
    float64 *a,
    float64 *b,
    float64  threshold)
{
    return (fabs(a[X] - b[X]) < threshold) && (fabs(a[Y] - b[Y]) < threshold);
}


#define OTHER(which) (1 - (which))

/* Subtract poly1 - poly2.
 * On entry, poly1 and poly2 contain the polygons; polygon_heap is an array,
 * allocated by the caller, in which to put the results; polygon_heap_size
 * is the amount of room available in the polygon_heap; result_polygon_ptrs
 * is an array, allocated by the caller, in which to put pointers to the
 * resulting polygons; result_num_vertices is an array, allocated by the
 * caller, in which to put the sizes of the resulting polygons; and
 * result_num_polygons is a pointer to where to return the count of resulting
 * polygons.
 * On completion, result_polygon_ptrs contains pointers to the resulting
 * polygons; result_num_vertices contains the sizes of the polygons; and
 * *result_num_polygons is the number of polygons found.
 * The return value is TRUE if the routine succeeded, FALSE otherwise.
 * FALSE probably indicates insufficient space in the passed arrays.
 *
 * ALGORITHM:
 * (1) Mark all vertices as "unused".
 * (2) Select one of the unused vertices of poly1 as a starting point.
 * (3) Select the next point as described below, until we return to
 *     starting point.
 * (4) Return to (2), as long as any vertices of poly1 remain unused.
 *
 * SELECTING NEXT VERTEX:
 * (1) Select which edge to use; if the current point is a vertex, choose
 *     the outgoing edge from the vertex.  If the current point is an 
 *     intersection of an edge from each polygon, choose whichever edge
 *     goes "into" the other polygon.  (For polygon-subtraction, we want to
 *     find all areas of poly1 which are NOT in poly2; thus going "into" poly2
 *     actually means moving into the EXTERIOR of poly2.)
 * (2) Trace along the selected edge looking for an intersection of that edge
 *      with some edge of the other polygon.
 * (3) If such an intersection is found, use that intersection as the next
 *     vertex of the result.
 * (4) If no such intersection is found, use the endpoint of the current edge
 *     as the next vertex of the result.
 *
 * NOTE: geo_polygon[0] is the local data corresponding to the input poly1,
 *       and geo_polygon[1] corresponds to poly2.  Thus, all references to 
 *       "poly1" above apply to geo_polygon[0], and all references to "poly2"
 *       apply to geo_polygon[1].
 *
 * NOTE: Most of this code can be reused to find the intersection of two
 *       polygons.  The major change would be re-defining "into" poly2
 *       as going into the INTERIOR of poly2.
 */
int32 geo_subtract_polygons(
    float64       poly1_vertices[][2],
    int32         poly1_num_vertices,
    float64       poly2_vertices[][2],
    int32         poly2_num_vertices,
    float64       threshold,
    float64       polygon_heap[][2],
    int32         polygon_heap_size,
    GEO_2D_POINT *result_polygon_ptrs[],
    int32         ptr_array_size,
    int32         result_num_vertices[],
    int32        *result_num_polygons)
{
    int32         vertices_found;
    float64       current_point[XY];
    float64       previous_point[XY];
    int32         i, j;
    int32         current_polygon_id;
    float64      (*current_polygon)[XY];
    float64      (*heap_limit)[XY];
    int32          heap_end;
    int32          which_vertex[NUM_POLYGONS];
    int32          first_time;
    int32          first_vertex;
    int32          reusing_vertex;

    assert(poly1_num_vertices < MAX_POLYGON_VERTICES);
    assert(poly2_num_vertices < MAX_POLYGON_VERTICES);

    convert_polygon(poly1_vertices, poly1_num_vertices, threshold, 0);
    convert_polygon(poly2_vertices, poly2_num_vertices, threshold, 1);
    

    geo_direction[0] = +1;     /* go clockwise around polygon 0 */
    geo_direction[1] = -1;     /* go counter-clockwise around polygon 1,
			      * in order to stay on the outside of it */
    heap_end = 0;
    heap_limit = polygon_heap + polygon_heap_size;
    current_polygon_id = 0;
    
    while(find_start_subtract_point(current_point, which_vertex, threshold))
    {
        first_vertex = which_vertex[0];
        if (which_vertex[0] != -1)
	  geo_available_vertices[0][which_vertex[0]] = FALSE;
        if (which_vertex[1] != -1)
	  geo_available_vertices[1][which_vertex[1]] = FALSE;

	current_polygon = polygon_heap + heap_end;
	if (current_polygon_id >= ptr_array_size)
	  return FALSE;   /* out of room */
	result_polygon_ptrs[current_polygon_id] = current_polygon;
	vertices_found = 0;
	previous_point[X] = 0;
	previous_point[Y] = 0;
	first_time = TRUE;
	while (find_next_subtract_point(current_point,which_vertex,threshold))
	{
	    if (which_vertex[0] != -1)
	      geo_available_vertices[0][which_vertex[0]] = FALSE;
	    if (which_vertex[1] != -1)
	      geo_available_vertices[1][which_vertex[1]] = FALSE;
	    
	    if (vertices_found == 0 || 
		!same_point(current_point, previous_point, threshold))
	    {
	        if (current_polygon + vertices_found > heap_limit)
		  return FALSE; /*$$$ out of heap-space*/
		current_polygon[vertices_found][0] = current_point[0];
		current_polygon[vertices_found][1] = current_point[1];
		++ (vertices_found);
		previous_point[X] = current_point[X];
		previous_point[Y] = current_point[Y];
	    }
	    if (first_time)
	    {
	        first_time = FALSE;
		first_vertex = which_vertex[0];
	    }
	    else
	    {
	        if (which_vertex[0] == first_vertex)
		  break;
	    }
	}
	
	while (vertices_found > 1
	       && same_point(current_polygon[vertices_found - 1],
			     current_polygon[vertices_found - 2], 
			     threshold))
	  --vertices_found;
	
	if (vertices_found > 0 )
	{
	    if (current_polygon_id >= ptr_array_size)
	      return FALSE;   /* out of room */
	    result_num_vertices[current_polygon_id] = vertices_found;
	    heap_end += vertices_found;
	    ++current_polygon_id;
	}
    }
    STDDEALLOC(geo_polygon[0]);
    STDDEALLOC(geo_polygon[1]);
    *result_num_polygons = current_polygon_id;
    return TRUE;
}


static void convert_polygon(
    float64 vertices[][XY],
    int32   num_vertices,
    float64 threshold,
    int32   which_polygon)
{
    int32       i;
    int32       source, dest, current;

    geo_polygon[which_polygon] = (GEO_2D_POINT*)
      STDALLOC((num_vertices+1) * XY * sizeof(float64));

    /* Copy, and coalesce any pair of successive points which are too close */
    source = 0;
    dest = 0;
    while(source < num_vertices)
    {
        geo_polygon[which_polygon][dest][X] = vertices[source][X];
        geo_polygon[which_polygon][dest][Y] = vertices[source][Y];
	current = source;
	source++;
	dest++;
        while(source < num_vertices &&
	      same_point(vertices[current], vertices[source], threshold))
	  ++source;
    }

    /* Close the polygon, if necessary */
    if(!same_point(geo_polygon[which_polygon][dest-1],
		   geo_polygon[which_polygon][0],
		   threshold))
    {
        geo_polygon[which_polygon][dest][X] = vertices[0][X];
	geo_polygon[which_polygon][dest][Y] = vertices[0][Y];
	dest++;
    }

    geo_num_vertices[which_polygon] = dest;
    geo_make_polygon_clockwise(geo_polygon[which_polygon], 
			       geo_num_vertices[which_polygon]);

    /* Initially mark all of the vertices of poly1 as available.
     */
    for(i = 0; i < geo_num_vertices[which_polygon] + 1; i++)
      geo_available_vertices[which_polygon][i] = TRUE;

    if(!same_point(geo_polygon[which_polygon][geo_num_vertices[which_polygon] - 1],
		   geo_polygon[which_polygon][0],
		   threshold))
    {
        geo_available_vertices[which_polygon][geo_num_vertices[which_polygon] - 1]
	  = FALSE;
    }
}


static int32 find_start_subtract_point(
    float64 point[XY],
    int32   which_vertex[NUM_POLYGONS],
    float64 threshold)
{
    int32        i;
    int32        inside;
    
    which_vertex[0] = -1;              /* none so far*/
    which_vertex[1] = -1;              /* none at all*/

    for(i = 0; i < geo_num_vertices[0] - 1; i++) /*we assume polygon is closed*/
    {     
        if (geo_available_vertices[0][i])
	{
	    inside = ( geo_is_point_in_polygon(geo_polygon[0][i], 
					 geo_polygon[1], 
					 geo_num_vertices[1])
		       == GEO_PT_INSIDE);
	    /* geo_is_point_in_polygon can get confused if the point is near
	     * an edge of the polygon.  If the given point is just outside
	     * the polygon, we don't need to use it as a starting point.
	     */
	    if (!inside)
	    {
	        inside = (geo_distance_sq_point_to_object(geo_polygon[1],
							 geo_num_vertices[1],
							 geo_polygon[0][i],
							 NULL)
		           <= threshold);
	    }
	    if (inside)
	    {
	        geo_available_vertices[0][i] = FALSE;
	    }
	    else
	    {
	        geo_point_is_vertex = TRUE;
		geo_point_is_intersection = FALSE;
		geo_vertex_vertex = i;
		geo_vertex_polygon_id = 0;
		point[0] = geo_polygon[0][i][0];
		point[1] = geo_polygon[0][i][1];
	        which_vertex[0] = i;
	        return TRUE;
	    }
	}
    }
    return FALSE;
}


#define OTHER(which) (1 - (which))

/* Return next vertex for a closed polygon.
 * This routine skips the last vertex of the polygon.
 */
static int32 next_vertex_c(
    int32 vertex,
    int32 polygon_id)
{
    int32     next;

    next = (vertex + geo_direction[polygon_id] + geo_num_vertices[polygon_id])
      % geo_num_vertices[polygon_id];

    if (next == geo_num_vertices[polygon_id] - 1)   /* if at last vertex */
    {
	if (geo_direction[polygon_id] == 1)         /* if going forward */
	  return 0;                               /* then skip to zero */

	if (geo_direction[polygon_id] == -1)        /* if going backward */
	  return geo_num_vertices[polygon_id] - 2;  /* then skip last */
    }
    return next;
}

/* Return prior vertex for a closed polygon.
 * This routine skips the last vertex of the polygon.
 */
static int32 prior_vertex_c(
    int32 vertex,
    int32 polygon_id)
{
    int32     prior;

    prior = (vertex - geo_direction[polygon_id] + geo_num_vertices[polygon_id])
      % geo_num_vertices[polygon_id];

    if (prior == geo_num_vertices[polygon_id] - 1)  /* if at last vertex */
    {
	if (geo_direction[polygon_id] == -1)        /* if going forward */
	  return 0;                               /* then skip to zero */

	if (geo_direction[polygon_id] == 1)         /* if going backward */
	  return geo_num_vertices[polygon_id] - 2;  /* then skip last */
    }
    return prior;
}

/* The "edge_id" of an edge is the vertex of when moving aroung the polygon
 * in its natural "direction", ie, the edge runs from "edge_id" to
 * "edge_id" + "direction".  Note that the meaning of "edge_id" thus depends
 * on the value of geo_direction for the polygon.
 */

/* This function returns the found point in "point".  If the found point
 * is a vertex of one or both polygons, "which_vertex" contains the
 * respective vertex-id's for each applicable polygon.
 * If no further points are found, this function returns the input point
 * as output point, and leaves which_vertex unchanged.
 */

static int32 find_next_subtract_point(
    float64 point[XY],
    int32   which_vertex[NUM_POLYGONS],
    float64 threshold)
{
    int32         i,j;
    int32         outgoing_polygon;
    float64      *prior_point[NUM_POLYGONS];
    float64      *next_point[NUM_POLYGONS];
    int32         new_edge_id;
    float64       tau0, tau1;
    int32         edge_ids[NUM_POLYGONS];
    int32         other_polygon;
    int32         new_vertex[2];
    
    new_vertex[0] = -1;              /* none */
    new_vertex[1] = -1;              /* none */

    /* If the starting point is the intersection of an edge of one polygon 
     * with an edge of the other polygon */
    if (geo_point_is_intersection)
    {
        for (i=0; i < NUM_POLYGONS; ++i)
	{
	    j = geo_intersection_edge_id[i];
	    prior_point[i] = geo_polygon[i][j];
	    if (same_point(prior_point[i], point, threshold))
	    {
	        j = prior_vertex_c(j, i);
		prior_point[i] = geo_polygon[i][j];
	    }
	    j = next_vertex_c(j, i);
	    next_point[i] = geo_polygon[i][j];
	    if (same_point(next_point[i], point, threshold))
	    {
	        j = next_vertex_c(j, i);
		next_point[i] = geo_polygon[i][j];
	    }
	    edge_ids[i] = prior_vertex_c(j, i);
	}
	if(!choose_outgoing_edge(geo_incoming_polygon,
				 prior_point, 
				 point, 
				 next_point,
				 &outgoing_polygon))
	  /* there are no more points, so return "point" 
	   * and "which_vertex" unchanged. */
	  return FALSE;

	other_polygon = OTHER(outgoing_polygon);
	if ( find_first_intersection_edge_with_other_polygon(
						    outgoing_polygon,
						    edge_ids[outgoing_polygon],
					            point,
					            point,
					            &new_edge_id,
					            &tau0,
					            &tau1))
	{
	    geo_point_is_intersection = TRUE;
	    geo_intersection_edge_id[other_polygon] = new_edge_id;
	    /* the old outgoing edge will be the new incoming edge, next time*/
	    geo_incoming_polygon = outgoing_polygon;
	    
	    geo_point_is_vertex = FALSE;
	    if (tau0 == 1)
	    {
	        new_vertex[outgoing_polygon] 
		  =next_vertex_c(edge_ids[outgoing_polygon], outgoing_polygon);
		geo_point_is_vertex = TRUE;
	    }
	    if (tau1 == 0)
	    {
	        new_vertex[other_polygon] = new_edge_id;
		geo_point_is_vertex = TRUE;
	    }
	    if (tau1 == 1)
	    {
	        new_vertex[other_polygon] 
		  = next_vertex_c(new_edge_id, other_polygon);
		geo_point_is_vertex = TRUE;
	    }
	}
	else
	{
	    geo_point_is_intersection = FALSE;
	    geo_point_is_vertex = TRUE;
	    geo_vertex_polygon_id = outgoing_polygon;
	    geo_vertex_vertex 
	      = next_vertex_c(edge_ids[outgoing_polygon], outgoing_polygon);
	    new_vertex[geo_vertex_polygon_id] = geo_vertex_vertex;
	    point[X] = next_point[outgoing_polygon][X];
	    point[Y] = next_point[outgoing_polygon][Y];
	}
    }


    /* If the starting point is a vertex of a polygon */
    else if (geo_point_is_vertex)
    {
	other_polygon = OTHER(geo_vertex_polygon_id);
	if ( find_first_intersection_edge_with_other_polygon(
						       geo_vertex_polygon_id,
						       geo_vertex_vertex,
						       point,
						       point,
						       &new_edge_id,
						       &tau0,
						       &tau1))
	{
	    geo_point_is_intersection = TRUE;
	    geo_intersection_edge_id[geo_vertex_polygon_id] = geo_vertex_vertex;
	    geo_intersection_edge_id[other_polygon] = new_edge_id;
	    /* the old outgoing edge will be the new incoming edge, next time*/
	    geo_incoming_polygon = geo_vertex_polygon_id;
	    
	    geo_point_is_vertex = FALSE;
	    if (tau0 == 1)
	    {
	        new_vertex[geo_vertex_polygon_id]
		  = next_vertex_c(geo_vertex_vertex, geo_vertex_polygon_id);
		geo_point_is_vertex = TRUE;
	    }
	    if (tau1 == 0)
	    {
	        new_vertex[other_polygon] = new_edge_id;
		geo_point_is_vertex = TRUE;
	    }
	    if (tau1 == 1)
	    {
	        new_vertex[other_polygon] 
		  = next_vertex_c(new_edge_id, other_polygon);
		geo_point_is_vertex = TRUE;
	    }
	}
	else
	  {
	    geo_point_is_intersection = FALSE;
	    geo_point_is_vertex = TRUE;
	    geo_vertex_vertex = next_vertex_c(geo_vertex_vertex,geo_vertex_polygon_id);
	    new_vertex[geo_vertex_polygon_id] = geo_vertex_vertex;
	    point[X] = geo_polygon[geo_vertex_polygon_id][geo_vertex_vertex][X];
	    point[Y] = geo_polygon[geo_vertex_polygon_id][geo_vertex_vertex][Y];
	  }
    }


    else
    {
        assert(FALSE);    /* ERROR: geo_point_is_intersection and */
	return FALSE;     /* geo_point_is_vertex are both FALSE */
	    geo_vertex_vertex=next_vertex_c(geo_vertex_vertex,geo_vertex_polygon_id);
    }
    which_vertex[0] = new_vertex[0];
    which_vertex[1] = new_vertex[1];
    return TRUE;
}


/* Let A0 be the edge from "prior_point[0]" to "point", A1 be the edge from
 * "point" to "next_point[0]". Let B0 be the edge from "prior_point[1]"
 * to "point", B1 be the edge from "point" to "next_point[1]".
 * Choose whichever of A1 or B1 makes the smaller angle with the incoming edge
 * of the incoming polygon.  If we chose A1, 
 * If B1 lies within the angle from A0 to A1, then choose edge B1.  
 * Else, if A1 lies within the angle from B0 to B1, then choose A1.
 * If neither of those two, then there is no edge to follow from here.
 * The parameter *which_polygon returns 0 for polygon0, 1 for polygon1.
 */

static int32 choose_outgoing_edge(
    int32    incoming_polygon,
    float64 *prior_point[NUM_POLYGONS],
    float64  point[XY],
    float64 *next_point[NUM_POLYGONS],
    int32   *which_polygon)
{
    float64           angle_incoming[XY];
    float64           angle_outgoing[XY];
    float64           vector[XY];
    ANGLE_RELATIONSHIP relation;

    VMAT2_VEC_SUB(prior_point[incoming_polygon], point, angle_incoming);
    VMAT2_VEC_SUB(next_point[1], point, angle_outgoing);
    VMAT2_VEC_SUB(next_point[0], point, vector);
    relation = compare_vector_to_angle(vector, angle_incoming, angle_outgoing);
    if (relation == VECTOR_INSIDE_ANGLE)
    /* we want to follow polygon[0] */
    {
        VMAT2_VEC_SUB(prior_point[1], point, angle_incoming);
	relation = compare_vector_to_angle(vector, 
					   angle_incoming, 
					   angle_outgoing);
	/* Verify that polygon[0] goes into polygon[1] from this point */
	if (relation == VECTOR_INSIDE_ANGLE ||
	    relation == VECTOR_ALONG_OUTGOING_EDGE)
	{
	    *which_polygon = 0;
	    return TRUE;
        }
	else
	  return FALSE;
    }
    else
    /* we want to follow polygon[1] */
    {
        VMAT2_VEC_SUB(prior_point[0], point, angle_incoming);
	VMAT2_VEC_SUB(next_point[0], point, angle_outgoing);
	VMAT2_VEC_SUB(next_point[1], point, vector);
	relation = compare_vector_to_angle(vector, 
					   angle_incoming, 
					   angle_outgoing);
	/* Verify that polygon[1] goes into polygon[0] from this point */
	if (relation == VECTOR_INSIDE_ANGLE ||
	    relation == VECTOR_ALONG_OUTGOING_EDGE)
	{
	    *which_polygon = 1;
	    return TRUE;
        }
	else
	  return FALSE;        
    }
}


#define ANGLE_THRESHOLD  0.0

/* NOTE: All three vectors are assumed to point outward from the vertex of
 * the angle.  In particular, "angle_incoming" points from the vertex outward,
 * NOT inward into the vertex.
 */
static ANGLE_RELATIONSHIP compare_vector_to_angle(
    float64 vector[XY],
    float64 angle_incoming[XY],
    float64 angle_outgoing[XY])
{     
    float64      angle_incoming_n[XY];
    float64      angle_outgoing_n[XY];
    float64      vector_n[XY];
    float64      cos_outgoing, cos_vector;
    float64      sin_outgoing, sin_vector;
    float64      ord_outgoing, ord_vector;

    vmat2_unit64(angle_incoming, angle_incoming_n);
    vmat2_unit64(angle_outgoing, angle_outgoing_n);
    vmat2_unit64(vector, vector_n);

    cos_outgoing = VMAT2_DOT_PROD  (angle_incoming_n, angle_outgoing_n);
    cos_vector   = VMAT2_DOT_PROD  (angle_incoming_n, vector_n);

    if (cos_vector >= 1 - ANGLE_THRESHOLD)
      return VECTOR_ALONG_INCOMING_EDGE;

    sin_outgoing = VMAT2_CROSS_PROD(angle_incoming_n, angle_outgoing_n);
    sin_vector   = VMAT2_CROSS_PROD(angle_incoming_n, vector_n);
    if (sin_outgoing >= -ANGLE_THRESHOLD)
      ord_outgoing = 1 - cos_outgoing; /* 0 <= ord_outgoing <= 2 */
    else
      ord_outgoing = 3 + cos_outgoing; /* 2 <= ord_outgoing <= 4 */
    if (sin_vector >= -ANGLE_THRESHOLD)
      ord_vector = 1 - cos_vector;     /* 0 <= ord_vector <= 2 */
    else
      ord_vector = 3 + cos_vector;     /* 2 <= ord_vector <= 4 */

    /* "ord" is now a monotonically increasing function of angle */
    if (ord_vector <= ord_outgoing - ANGLE_THRESHOLD)
      return VECTOR_INSIDE_ANGLE;
    else 
    {
        if (ord_vector <= ord_outgoing + ANGLE_THRESHOLD)
	  return VECTOR_ALONG_OUTGOING_EDGE;
	else 
	  return VECTOR_OUTSIDE_ANGLE;
    }
}


/* Consider the edge running from "edge_initial_vertex" to 
 * "edge_initial_vertex" + "geo_direction" in "edge_polygon".  
 * Starting at "start_point", find the first place that edge intersects 
 * the other polygon.
 * Return the point in "point", the index of the intersected edge in
 * "new_initial_vertex", and the tau's of intersection in "tau0" and "tau1".
 */

static int32 find_first_intersection_edge_with_other_polygon(
    int32    edge_polygon_id,
    int32    edge_initial_vertex,
    float64  start_point[XY],
    float64  new_point[XY],
    int32   *new_initial_vertex,
    float64 *tau0,
    float64 *tau1)
{
    int32              search_polygon_id;
    float64          (*search_polygon)[XY];
    int32              num_search_vertices;
    float64           *end_point;
    int32              v0, v1;
    float64            min_distance_sq, new_distance_sq;
    int32              found;
    float64            intersect_point[XY];
    float64            found_point[XY];
    float64            current_tau0, current_tau1;
    float64            angle_outgoing[XY];
    float64            angle_incoming[XY];
    float64            vector[XY];
    ANGLE_RELATIONSHIP result;

    search_polygon_id = OTHER(edge_polygon_id);
    search_polygon = geo_polygon[search_polygon_id];
    num_search_vertices = geo_num_vertices[OTHER(edge_polygon_id)];
    end_point = geo_polygon[edge_polygon_id]
                         [next_vertex_c(edge_initial_vertex, edge_polygon_id)];
    VMAT2_VEC_SUB(end_point,
		  geo_polygon[edge_polygon_id][edge_initial_vertex],
		  vector);
    VMAT2_SCAL_VEC_MUL(-1, angle_outgoing, angle_incoming);
    found = FALSE;
    *new_initial_vertex = -1;
    min_distance_sq = 2 * geo_distance_sq_2_points(start_point, end_point, 2);
    for (v0=0; v0 < num_search_vertices - 1; ++v0)
    {
        v1 = next_vertex_c(v0, search_polygon_id);
        if (geo_intersection_2_linesegs(start_point,
					end_point,
					search_polygon[v0], 
					search_polygon[v1],
					intersect_point, 
					&current_tau0, &current_tau1))
	{
	    new_distance_sq = geo_distance_sq_2_points(start_point,
						       intersect_point, 2);
	    
	    VMAT2_VEC_SUB(search_polygon[v1], 
			  search_polygon[v0], 
			  angle_outgoing); 	    
	    VMAT2_VEC_SUB(search_polygon[v0], 
			  search_polygon[v1], 
			  angle_incoming); 
	    if (current_tau1 == 0)
	      VMAT2_VEC_SUB(search_polygon
			       [prior_vertex_c(v0,search_polygon_id)], 
			    search_polygon[v0], 
			    angle_incoming); 
	    else if (current_tau1 == 1)
	      VMAT2_VEC_SUB(search_polygon
			       [next_vertex_c(v1,search_polygon_id)], 
			    search_polygon[v1], 
			    angle_outgoing); 	    
	    /* If we're crossing into the search polygon, ignore the edge */
	    result =
	      compare_vector_to_angle(vector, angle_incoming, angle_outgoing);
		if (result == VECTOR_INSIDE_ANGLE ||
		    result == VECTOR_ALONG_OUTGOING_EDGE)
	      continue;
	    
	    found = TRUE;
	    if (min_distance_sq >= new_distance_sq)
	    {
	        *new_initial_vertex = v0;
		*tau0 = current_tau0;
		*tau1 = current_tau1;
		min_distance_sq = new_distance_sq;
		found_point[X] = intersect_point[X];
		found_point[Y] = intersect_point[Y];
	    }
       	}
    }    
    assert( found == (*new_initial_vertex != -1) );

    if (found)
    {
        new_point[X] = found_point[X];
        new_point[Y] = found_point[Y];
    }
    return found;
}


static int32 geo_polygon_contains_polygon(
    float64 outer_poly_vertices[][XY],
    int32   outer_poly_num_vertices,
    float64 inner_poly_vertices[][XY],
    int32   inner_poly_num_vertices)
{
    int32       i;
    
    for (i=0; i<inner_poly_num_vertices; ++i)
    {
        if (geo_is_point_in_polygon(inner_poly_vertices[i],
				    outer_poly_vertices,
				    outer_poly_num_vertices)
	    == GEO_PT_OUTSIDE)
	{
	    return FALSE;
	}
    }
    return TRUE;
}

#undef CROSS_PRODUCT
#undef MAXIMUM
#undef OTHER

/* geo_locate_xy_pos_on_segments:
 *
 * given a bunch of line segments, compute the xy locations along the 
 * segments starting at 'start_offset' such that the locations are 
 * 'spacing' meters away from each other.  
 *
 * args:
 *    line:	  the vertices of the segments.
 *    n_verts:    the number of vertices on the line.
 *    spacing:    the distance between positions.
 *    locations:  the returned xy positions.
 *    start_offset: upon entering, the amount of space left over on the last
 *                segment.  upon exit, the offset for the next segment.
 *    num_positions:  upon entering, the max number of positions to locate.  
 *                upon exit, its the actual number of positions that were 
 *                located.
 */
void geo_locate_xy_pos_on_segments(
    float64  line[][XY],
    int32    n_verts,
    float32 *start_offset,
    float32  spacing,
    int32   *num_positions,
    float64  locations[][XY])
{
    int32 seg_num;      
    int32 count=0;      
    float64 seg_length; 
    float64 direction[XY];
    float64 length;
    float64 base_location[XY];
    float64 leftover = *start_offset;
    int32 max_positions= *num_positions;

    /* foreach segment.. */
    for (seg_num=0; (seg_num<n_verts-1) && (max_positions > count); seg_num++)
    { 
        float64 t1[XY], t2[XY];

        VMAT2_RECAST_VEC(line[seg_num], t1);
        VMAT2_RECAST_VEC(line[seg_num+1], t2);

	/* make the base_location be the first vertex on this segment. */
	VMAT2_VEC_COPY64(t1, base_location);

        /* compute a unit vector in the direction of the current segment */
        vmat2_vec_sub64(t2, t1, direction);
        vmat2_unit64(direction, direction);

        /* compute the length of the current segment */
	seg_length = geo_distance_2_points(t1, t2, 2);

        /* figure out the fist point on this segment in case there was
         * leftover space on the last segment..
         */
        locations[count][X] = base_location[X] + leftover * direction[X];
        locations[count][Y] = base_location[Y] + leftover * direction[Y]; 

        length = leftover;

        if (++count >= max_positions)
	{
  	    leftover = spacing - NS_LIMITS(0.0, (seg_length-length), spacing);
	    break;
	}

        /* scale the direction by the spacing. */  
        vmat2_scal_vec_mul64((float64)spacing, direction, direction);

        /* process this segment..  loop through adding the appropriate
         * offset of the last computed point.  when the length from the
         * current location to the first location is greater than the
         * segment length, break out of this loop and move on to the
         * next segment..  
         */
	while ((seg_length >= (length + spacing - 0.1)) && 
               (count < max_positions))
        {
	    vmat2_vec_add64(locations[count-1], direction, locations[count]);
            length = geo_distance_2_points(base_location, locations[count], 2);

            count++;
        } 

        /* check to see if there is any leftover space that will effect
         * the location first position on the next segment..
         */
        
        leftover = spacing - NS_LIMITS(0.0, (seg_length-length), spacing);
    }

    *start_offset = leftover;
    *num_positions = count;
}
