#ifndef lint
static char rcsid [] = "$RCSfile: geo_param.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_param.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Generic line utilities                                    *
*       Created: Sat Feb 26 10:44:36 EST 1994                               *
*       Author: swarts                                                      *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>

void geo_init_param_line(
    GEO_PARAMETRIC_LINE *line,
    float64              point[2],
    float64              dir[2])
{
    VMAT2_VEC_COPY64(point, line->p0);
    VMAT2_VEC_COPY64(dir, line->dp_dt);
}

void geo_init_param_line_from_pts(
    GEO_PARAMETRIC_LINE *line,
    float64              start[2],
    float64              end[2])
{
    VMAT2_VEC_COPY64(start, line->p0);
    VMAT2_VEC_SUB(end, start, line->dp_dt);
}

void geo_param_line_value(
    GEO_PARAMETRIC_LINE *line,
    float64              t,
    float64              value[2])
{
    if (value)
    {
	VMAT2_SCAL_VEC_MUL(t, line->dp_dt, value);
	VMAT2_VEC_ADD(line->p0, value, value);
    }
}

/*
 * We are looking for the point where:
 *
 * l0->p0 + t0 * l0->dp_dt == l1->p0 + t1 * l1->dp_dt or
 * t0 * l0->dp_dt == (l1->p0 - l0->p0) + t1 * l1->lp_dt
 *
 * Doing a cross prod with l1->dp_dt to get rid of the t1 term gives:
 *
 * t0 * (l0->dp_dt x l1->dp_dt) = (l1->p0 - l0->p0) x l1->dp_dt
 *
 * If l0->lp_dt x l1->dp_dt == 0 then the lines are parallel or colinear,
 * otherwise, 
 *
 * t0 = ((l1->p0 - l0->p0) x l1->dp_dt) / (l0->dp_dt x l1->dp_dt)
 *
 * Simularly,
 *
 * t1 = ((l0->dp - l1->p0) x l0->dp_dt) / (l1->dp_dt x l0->dp_dt)
 *
 * Using the fact that A x B == - B x A we get
 *
 * t1 = ((l1->p0 - l0->p0) x l0->dp_dt) / (l0->dp_dt x l1->dp_dt)
 *
 * So we can calculate (l1->p0 - lp->p0) / (l0->dp_dt x l1->dp_dt)
 * and take the cross with each dp_dt to get the t's
 *
 */
int32 geo_param_lines_intersect(
    GEO_PARAMETRIC_LINE *l0,
    GEO_PARAMETRIC_LINE *l1,
    int32               *t0,
    int32               *t1)
{
    float64 denom;
    float64 temp[2];

    denom = VMAT2_CROSS_PROD(l0->dp_dt, l1->dp_dt);

    if (denom == 0.0)
    {
	return FALSE;
    }

    if (t0 || t1)
    {
	VMAT2_VEC_SUB(l1->p0, l0->p0, temp);
	VMAT2_SCAL_VEC_MUL(1.0/denom, temp, temp);
    }

    if (t0)
    {
	*t0 = VMAT2_CROSS_PROD(temp, l1->dp_dt);
    }

    if (t1)
    {
	*t1 = VMAT2_CROSS_PROD(temp, l0->dp_dt);
    }
    return TRUE;
}




