#ifndef lint
static char rcsid [] = "$RCSfile: geo_tang.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vsrch_class.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Manages class information attached to each vehicle        *
*       Created: Thu Jan 20 10:44:36 EST 1994
*       Author: jtan
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>
#include <stdio.h>
#include <stdstring.h> /*common/include/global*/

/*
 * Given a circle defined by the 'circle_center' and the 'circle_radius' and 
 * a 'point', geo_tangential_points_on_circle() calculates the two tangential 
 * points on the circle. Looking at the circle from the point, 'tang_point1' 
 * is the tangential point on the left side of the circle, and 'tang_point2' 
 * is the one on the right side.
 */
int32 geo_tangential_points_on_circle(
    float64 circle_center[2],
    float64 radius,
    float64 point[2],
    float64 tang_point1[2],
    float64 tang_point2[2])
{
    float64 dist;
    float64 sin_theta;
    float64 cos2_theta;
    float64 to_circle[2];
    float64 rotation[2][2];

    vmat2_vec_sub64(circle_center, point, to_circle);
    dist = sqrt(VMAT2_VEC_MAG_SQ(to_circle));
	
    if (dist < radius) 
    {
	/* no tangential points */
	printf("@geo_tangential_points_on_circle: point out of circle!!!\n");
	return(0);
    }
    /* First calculate the coordinates of the two tangential 
       points in the following coordinate frame
                  y |			
                   (|) <- circle at y=dist, x=0, radius=r
                    | 
                    |
               -----*------ x 
                    0 (point)
    */	  
 
    sin_theta = radius / dist;
    cos2_theta = 1 - sin_theta * sin_theta;
    tang_point1[X] = - radius * sqrt(cos2_theta);
    tang_point1[Y] = dist * cos2_theta;
    tang_point2[X] = - tang_point1[X];
    tang_point2[Y] = tang_point1[Y];
	
    /* Find the rotation matrix between the above coordinate frame
       and the following coordinate frame:
	   
                  y |    ( ) <- circle
                    |
                    |
                    |
              ------*------- x   
                    0 (point)
    */

    vmat2_unit64(to_circle, to_circle);
    vmat2_primary_rotation64(to_circle[X], to_circle[Y], rotation);
	
    /* Apply the coordinate transform to the 2 tangential points*/
    vmat2_mat_vec_mul64(rotation, tang_point1, tang_point1);
    vmat2_mat_vec_mul64(rotation, tang_point2, tang_point2);
    vmat2_vec_add64(point, tang_point1, tang_point1);
    vmat2_vec_add64(point, tang_point2, tang_point2);
    return(1);
}				
