/*
 * $RCSfile: libnet.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/* header file for most net_ functions */

#ifndef NETWORKLIB_INCLUDED

#define NETWORKLIB_INCLUDED


#if defined(_GT_)

#include <stdext.h>  /*common/include/global*/
#include "compiler.h"
#include "ifxvisi.h"
#include "stdio.h"
#include "errno.h"
#include "enpioctl.h"
#include "enparg.h"
#include "netlock.h"
#include "compat.h"
/* Must insure that data is moved no more than 16 bits at a time because the
   ENP10 is a D16 device. */
#define GETLONG(a)	get_long_as_words(a)
#define SETLONG(a, b)	set_long_as_words((a), (b))
#define DATACOPY(a,b,c)	movewords((a), (b), ((c) + 1) >> 1)
#define VOLATILE

#elif defined(SIMBFLY)
#include "public.h"
#include "multibus.h"
#include "mbregion.h"
#include "errno.h"
#include "libnetif.h"
#include "enpsvr.h"
#include "enparg.h"
#include "netlock.h"
#include "compat.h"
extern int errno;
#define GETLONG(a)	*(a)
#define SETLONG(a, b)	*(a) = (b)
/* bcopy uses PNC and is slower for intra-node copies */
#define DATACOPY(a,b,c)	movedata(a,b,c)
#define VOLATILE

#elif defined(MASSCOMP)

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <errno.h>
#include "libnetif.h"
#include "enpioctl.h"
#include "enparg.h"
#include "netlock.h"
#include "compat.h"
extern int errno;
#define GETLONG(a)	*(a)
#define SETLONG(a, b)	*(a) = (b)
#define DATACOPY(a,b,c)	bcopy(a,b,c)
#define VOLATILE

#elif defined(mips)

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>

#ifndef sgi
typedef	int		key_t;		/* IPC key type */
#include <sysv/sys/ipc.h>
#include <sysv/sys/shm.h>
#include <sysv/sys/sem.h>
#else /* sgi */
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#endif /* sgi */

#include <errno.h>
#include "libnetif.h"
#include "enpioctl.h"
#include "enparg.h"
#include "netlock.h"
#include "compat.h"
extern int errno;
#define GETLONG(a)	*(a)
#define SETLONG(a, b)	*(a) = (b)
#define DATACOPY(a,b,c)	bcopy(a,b,c)
#define VOLATILE volatile

#elif defined(SUN)

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/mman.h>
#include <errno.h>
#include "sim_macros.h"
#include "libnetif.h"
#include "enpioctl.h"
#include "enparg.h"
#include "netlock.h"
#include "compat.h"
#ifdef sun4
#include "simle_ioctl.h"
#include "simle.h"
#endif
extern int errno;
#ifdef sun4
#define GETLONG(a)	get_long_as_words(a)
#define SETLONG(a, b)	set_long_as_words((a), (b))
#define DATACOPY(a,b,c)	movewords((a), (b), ((c) + 1) >> 1)
#define VOLATILE
#else
#define GETLONG(a)	*(a)
#define SETLONG(a, b)	*(a) = (b)
#define DATACOPY(a,b,c)	bcopy(a,b,c)
#define VOLATILE
#endif

#else

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <errno.h>
#include "sim_macros.h"
#include "libnetif.h"
#include "enpioctl.h"
#include "enparg.h"
#include "netlock.h"
#include "compat.h"
extern int errno;
#define GETLONG(a)      *(a)
#define SETLONG(a, b)   *(a) = (b)
#define DATACOPY(a,b,c) bcopy(a,b,c)
#define VOLATILE

#endif



#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

#define NUM_NETHANDLES			16

#ifdef mips
#define ENP_MAPSIZE     (CMC_BUFFER_SIZE + CMC_DEVICE_SIZE)
#else
#ifndef ENP_MAPSIZE
#define ENP_MAPSIZE     (1 << 17)
#endif
#endif

struct netlib_pb {
    int used;
    int *currentinbuf;			/* current in buffer index */
    RingElement *inbuffers;		/* input buffer ring */
    int *currentoutbuf;			/* current out buffer index */
    RingElement *outbuffers;		/* output buffer ring */
    long timestamp;			/* timestamp value */
    int net_device;
    unsigned long last_time;		/* used with mapped-in timer */
    unsigned long total_elapsed_time;
    struct enp_registers *enp_ptr;	/* enp registers */
    long *reg_ptr;			/* general registers */
    char *base_ptr;			/* base of mapped in memory */
    int timeout;			/* timeout for blocking accesses */
    int inringnum;			/* input ring elements */
    int outringnum;			/* output ring elements */
    char *share_ptr;			/* pointer to shared block */
    int share_size;			/* bytes in shared block */
    NetworkAddress sndFrom;		/* source address when sending */
    int sndType;			/* packet type when sending */
    NetworkAddress rcvTo;		/* dest address of last packet rcvd */
    NetworkAddress rcvFrom;		/* source address of last packet rcvd */
    int rcvType;			/* packet type of last packet rcvd */
#ifdef SIMBFLY
    struct enpsvr_req *net_req;
    QH net_queue;
    EH net_event;
    OID net_reqoid;
    QH net_xmtlock;
    QH net_rcvlock;
    OID inringoid;			/* input ring */
    OID outringoid;			/* output ring */
    OID share_oid;			/* shared block */
#else
    char *enp_base;			/* address of base of mapped memory */
    char dev_name[40];			/* name of device */
    char *params;			/* -> parameters for ioctl call */
    int net_xmtlock;			/* semaphore ids for locks */
    int net_rcvlock;
#endif
    int fd;				/* file descriptor */
    int type;				/* network interface type */
    int enptype;			/* the type of CMC card */
    struct net_info_block nib;
#ifdef _GT_
    IFXCCB  *ccb;
    int     (*read)();
    int     (*write)();
    int     (*ioctl)();
    unsigned int    flags;
    long    base_time;
#endif
    int     inbuf_size;			/* Amount of data currently in the 
						input buffer */
    char    inbuf[1518]; 		/* 14 bytes header, maximum 1500 
						data, 4 CRC */
    char    outbuf[1514]; 		/* 14 bytes header, maximum 1500 data */
    int inbuffers_map_size;		/* input buffer ring */
    NetList *net_list;			/* pointers into list */
    int micro;				/* index into array of next pointers
						in Sim_buf */
    float	padding;		/* keeps inbuf aligned on 8 bytes */
};

struct netlib_pb_aligned 
{
    double alignment;
    struct netlib_pb npb_array[NUM_NETHANDLES];
};

#define CHECK_HANDLE(h, retval) \
if(h >= NUM_NETHANDLES || npb[h].used == FALSE) \
    { \
	errno = EBADF; \
	return(retval); \
    }

#ifdef MSC
#define XXX_OPEN(a,b)		dos_open(a,b)
#define XXX_IOCTL(a,b,c)	dos_ioctl(a,b,c)
#define XXX_CLOSE(a)		dos_close(a)
#else
#define XXX_OPEN(a,b)		open(a,b)
#define XXX_IOCTL(a,b,c)	ioctl(a,b,c)
#define XXX_READ(a,b,c)		read(a,b,c)
#define XXX_WRITE(a,b,c)	write(a,b,c)
#define XXX_LSEEK(a,b,c)	lseek(a,b,c)
#define XXX_CLOSE(a)		close(a)
#endif 

#define ENPMAPSIZE	((1 << 17) - 0x1000)	/* Bytes of card to map in */


extern char statistics_strings[][41];
extern int max_statistics_string;
extern struct netlib_pb_aligned npb_space;
extern struct netlib_pb *npb;
extern char errstr[];
extern unsigned _enpmapsize;
extern char *_enpmemory_string;

extern void movebytes();
extern void movewords();
extern int32 get_long_as_words();
extern void set_long_as_words();
extern void smovewords();
extern void movelongs();
extern void movedata();

#endif
