/*
 * $RCSfile: libnetif.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/*
 * File:	network.h
 *
 * Copyright (c) 1985, 1986, 1987, 1988 by BBN Laboratories
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This include file defines constants and declares data types related to
 * the network data transfer data structures.  
 *
 *
 */

#ifndef NETWORK_H__INCLUDED

#define NETWORK_H__INCLUDED 1


#define NETWORK_BUFFER_SIZE	1500		/* data bytes in a packet */
#define maxNetworkPDUSize 256			/* size of PDU - DEFUNCT!!!! */
#define ENETADDRSZ 6				/* number of chars in address*/
#define ENETMINSIZE 46				/* minimum data size */
#define ENETMAXSIZE 1500			/* maximum data size */
#define PROTOCOL_SIM	21000			/* simulation protocol */
#define PROTOCOL_DATA	21001			/* data link type */
#define PROTOCOL_VOICE	21002			/* voice packets */
#define PROTOCOL_GT	21003			/* GT sim/cig packets */
#define PROTOCOL_EX	21004			/* EX packets */
#define NETWORK_CODE "Enp"			/* network control code */
#ifdef BF_PLUS
#define NETWORK_DEVICE "enp0"			/* network controller */
#endif
#ifdef SIMBFLY
#define NETWORK_DEVICE "enp0"			/* network controller */
#endif
#ifdef MASSCOMP
#define NETWORK_DEVICE "/dev/enp0"		/* network controller */
#endif
#ifdef mips
#ifdef sgi
#define NETWORK_DEVICE "ec0"			/* network controller */
#else
#define NETWORK_DEVICE "/dev/enp0"		/* network controller */
#endif
#endif
#ifdef _GT_
#define NETWORK_DEVICE "enp0:"			/* network controller */
#endif

/* NOTE!!!!!!!  for many SPARC based machines running SunOS 4.1.1 the default
	network device is NOT "/dev/enp0".  It is "/dev/simle0" (for the 
	one-tap method), or "/dev/simle1" (for the two-tap method).  Hence,
	it is likely that the NETWORK_DEVICE macro may not work.  You can:
	1. change the macro at compile time -> -DNETWORK_DEVICE=\"/dev/simle1\"
	2. feed the network device in to the program as a command line option
	3. set an environment variable to contain the network device and use 
		getenv() to check it at runtime
	4. get the network device from a configuration file at runtime
*/

#if defined(SUN) && !defined(sun4) && !defined(sun5)
#define NETWORK_DEVICE "/dev/enp0"		/* network controller */
#endif
#ifdef sun4
#define NETWORK_DEVICE "/dev/simle0"
#endif
#ifdef sun5
#define NETWORK_DEVICE "le0"
#endif
#ifdef linux
#define NETWORK_DEVICE "eth0"
#endif
#ifdef freebsd
#define NETWORK_DEVICE "ed0"
#endif

/* Any more? */
#define NUMBER_OF_REGISTERS	12		/* number of registers */

/* This flag is ONLY valid for net_open.  It is used when net_open is called */
/* before the CMC card is actually executing (as in ringstart) */
#define NETINIT 0x80
#define NETNOCHECKALIVE 0x40

/* Supported network interface types. */
#define NIF_TYPE_CMC	1
#define NIF_TYPE_147	2
#define NIF_TYPE_SIMLE	3

/* These flags are used for the net_flush call */
#define NETFLUSHRCV 0x1				/* flush the receive queue */
#define NETFLUSHXMT 0x2				/* flush the transmit queue */

/* Flags on the net_send and net_recv calls. */
#define NETWAIT 0x1				/* wait for success */
#define NETLOCK 0x2				/* lock rings before access */
#define NETNOLEN 0x4				/* don't force length */
#define NETBLOCK 0x8				/* true blocking access */
#ifdef mips
#undef  NETLOCK
#define NETLOCK NETWAIT  /* On the mips, we can afford to wait on write,
			  * but all the software (i.e., libassoc, libp2p)
			  * which is used on a Masscomp cannot, so the
			  * simplest place to differentiate is here.
			  */
#endif

#define NETCLKFUDGE .99826	/* milliseconds in a tick for cmc card clock */ 
#define CMC_TICK_MS .99826	/* milliseconds in a tick for cmc card clock */ 

/* 802.3 related defines: */
#define SAP_SNAP		0xaa	/* SAP that says there's a SNAP */
#define SAP_SNAP_SIZE		8
#define MAX_DATA_SIZE_8023	(ENETMAXSIZE - SAP_SNAP_SIZE)
#define MIN_DATA_SIZE_8023	(ENETMINSIZE - SAP_SNAP_SIZE)
#define ORG_CODE_BBN		0x080008
#define UNNUMBERED_DATA		3

/* Macros used in the network library. */
#define Ringsize_in_chars(size) ((size + 1) * sizeof(RingElement))
/* Macros for getting/setting fields of the 802.2 headers.  This is ugly
   but I couldn't make the structure declarations to the right thing as
   far as alignment was concerned. */
#define GET_DSAP(p)	 	(((unsigned char *)(p))[0])
#define GET_SSAP(p) 		(((unsigned char *)(p))[1])
#define GET_CONTROL(p)		(((unsigned char *)(p))[2])
#define GET_PROTOID(p)		((((unsigned char *)(p))[3] << 16) | \
				 (((unsigned char *)(p))[4] <<  8) | \
				  ((unsigned char *)(p))[5])
#define GET_ETHER_TYPE(p)	(((unsigned short *)(p))[3])
#define GET_DATA_PTR(p)		((unsigned char *)(p) + 8)
#define SET_DSAP(p,x)	 	(((unsigned char *)(p))[0] = (x))
#define SET_SSAP(p,x) 		(((unsigned char *)(p))[1] = (x))
#define SET_CONTROL(p,x)	(((unsigned char *)(p))[2] = (x))
#define SET_PROTOID(p,x)	((unsigned char *)(p))[3] = (x) >> 16; \
				((unsigned char *)(p))[4] = (x) >>  8; \
				((unsigned char *)(p))[5] = (x)
#define SET_ETHER_TYPE(p,x)	((unsigned short *)(p))[3] = (x)


struct netdef {					/* network definition data */
    char eaddr[13];				/* ethernet address */
};

struct net_info_block {		/* information about network for net_open */
    int inringkey;		/* keys for various items */
    int inringsize;		/* size of input ring buffer (number of bufs) */
    int outringkey;
    int outringsize;
    int sharekey;		/* key for shared memory block */
    int sharesize;		/* size of shared memory block */
    int xmtsemkey;		/* transmit ring semaphore */
    int rcvsemkey;		/* receive ring semaphore */
    int inringoff;		/* offset from start of shared memory of ring */
    int outringoff;
    int shareoff;
    int new;
    int hostbufsize;		/* bytes in on-board buffer */
    int mca_num;		/* number of multicast addresses */
    int pad[34];		/* adjust as parameters are added */
};

/* Declarations of network buffer and its components.  */

/* An Ethernet address: */
typedef struct {
    unsigned char addr[ENETADDRSZ];
} NetworkAddress;

/* An Ethernet header (version2.0, not 802.3): */
typedef struct {
    NetworkAddress to;
    NetworkAddress from;
    short type;
} NetworkHeader;

/* An Ethernet header (802.3, not version2.0): */
typedef struct {
    NetworkAddress to;
    NetworkAddress from;
    short length;
} NetworkHeader8023;

/* An Ethernet packet (version2.0): */
typedef struct {
    NetworkHeader hdr;
    char data[NETWORK_BUFFER_SIZE];
} NetworkBuffer;

/* An Ethernet packet (802.3): */
typedef struct {
    union {
        NetworkHeader hdr2;
        NetworkHeader8023 hdr8023;
    } h;
    char data[NETWORK_BUFFER_SIZE];
} NetworkPacket;

/* An Ethernet packet with a byte count attached: */
typedef struct {
    int num_bytes;
    NetworkBuffer netbuf;
} ExtNetworkBuffer;

#define ExtNetBufHdrSize (sizeof(ExtNetworkBuffer) - sizeof(NetworkBuffer))

/* Declaration of the component of the rings. */
typedef struct {
    short full;				/* TRUE = full, FALSE = empty */
    short size;				/* size in bytes of data to
					   transfer (in element below) */
    long timestamp;
    NetworkBuffer buf;				
} RingElement;

typedef struct {
    short full;
    short size;
    long timestamp;
    NetworkPacket buf;
} QueueElement;

typedef struct {
    short next;
    short size;
    long timestamp;
    NetworkPacket buf;
} ListElement;

typedef struct {
    short    head;		/* start of busy list */
    short    tail;		/* end of busy list */
    short    alt_head;		/* start of busy list if buffer is in use */
    short    alt_tail;		/* end of alt busy list */
    short    in_use;		/* so driver and libnetif can synchronize */
} NetList;

#define NET_EMPTY_LIST		((short)-1)

/* Available from the network library: */
extern NetworkAddress BroadcastAddress;	/* broadcast address */
extern NetworkAddress FillinAddress;	/* use as from address to force card */
					/* to fill in rommed station address. */
extern int net_zero_statistics();
extern int net_print_statistics();
extern void net_addr_bin_to_str();
extern void net_addr_format_convert();
extern int net_zeroaddr();
extern int net_open();
extern int net_close();
extern int net_send();
extern int net_snd();
extern int net_recv();
extern int net_rcv();
extern int net_get_rcv();
extern int net_release_rcv();
extern int net_get_snd();
extern int net_release_snd();
extern int net_current_time();
extern int net_gettime();
extern int net_set_snd_type();
extern int net_set_snd_from_addr();
extern int net_get_rcv_to_addr();
extern int net_get_rcv_from_addr();
extern int net_get_rcv_type();
extern int net_addr_compare();
extern int net_add_mca();
extern int net_del_mca();
extern int net_init_type();
extern int net_add_type();
extern int net_addr_str_to_bin();
extern int net_flush();
extern int net_get_statistics();
extern void net_stat_string();
extern int net_print_statistics();
extern long net_get_timestamp();
extern int net_put_timestamp();
extern int net_stop();
extern int net_norm();
extern int net_alive();
extern int net_init_mca();
extern int net_iocontrol();

#endif

