#ifndef lint
static char rcsid [] = "$RCSfile: net_addr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_addr.c
 *
 * Copyright 1995 Loral Advanced Distributed Simulation Inc.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform control functions.
 *
 * Functions:
 *		net_addr_compare
 *		net_addr_bin_to_str
 *		net_addr_str_to_bin
 *		net_addr_format_convert
 *		net_getaddr
 *		net_zeroaddr
 */

/*LINTLIBRARY*/

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

/*
 * net_addr_compare - compare addresses
 *
 * This routine compares two addresses in binary format.  It returns TRUE
 * if the addresses are the same and FALSE if not.
 */
int net_addr_compare(a1, a2)
NetworkAddress *a1;
NetworkAddress *a2;
{
    int i;

    for(i = 0; i < ENETADDRSZ; i++) {
        if(a1->addr[i] != a2->addr[i])
	    return(FALSE);
    }

    return(TRUE);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */

/*
 * net_addr_bin_to_str - convert binary network address to string
 */

void net_addr_bin_to_str(p, ps)
NetworkAddress *p;
char *ps;
{
    int i,j;

    for(i = 0, j = 0; i < ENETADDRSZ; i++) {
        sprintf(&ps[j++], "%x", (int)p->addr[i] >> 4 & 0xf);
        sprintf(&ps[j++], "%x", (int)p->addr[i] & 0xf);
    }
}

#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

/*
 * net_addr_str_to_bin - convert string to binary network address
 *
 * All 12 hex digits must be present in the string.
 */

int net_addr_str_to_bin(p, s)
NetworkAddress *p;
char *s;
{

    int i, x;
    
    for(i = 0; i < ENETADDRSZ; i++)
        p->addr[i] = 0;

    /* check for exactly 12 digits, all hex digits */
    for (i = 0; i < ENETADDRSZ * 2; i++) {
	if(s[i] >= '0' && s[i] <= '9')
	    continue;
	else if(s[i] >= 'a' && s[i] <= 'f')
	    continue;
	else if(s[i] >= 'A' && s[i] <= 'F')
	    continue;
	else
            return FALSE;
    }

    if (s[i] != '\0')
	return FALSE;

#ifdef _GT_
    {
       char stmp[3];
       /* sscanf doesn't seem to work right! (at least at one point in time!) */
       for (i = 0; i < ENETADDRSZ; i++) {
	    stmp[0] = s[2 * i];
	    stmp[1] = s[2 * i + 1];
	    stmp[2] = '\0';

	    p->addr[i] = strtol(stmp, NULL, 16);
        }
    }
#else
    for (i = 0; i < ENETADDRSZ; i++, s += 2) {
	sscanf (s, "%2x", &x);
	p->addr[i] = x;
    }
#endif

    return TRUE;
}


/*
 * net_getaddr - get the network address
 *
 * This routine is called to get the network address of the host.
 */

int net_getaddr(h, p)
int h;
NetworkAddress *p;
{

    CHECK_HANDLE(h, -1);

#ifndef SIMBFLY
    npb[h].params = (char *)p;
#endif

    if(errno = net_access(h, ENPGETEADDR))
	return(-1);

#ifdef SIMBFLY
    strcpy(p, npb[h].net_req->a.geteaddr_req.eaddr);
#endif

    return(0);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */

/*
	convert from "XXXXXXXXXXXX" to "YY YY YY YY YY YY"
	 
	assumes p1[13] and p2[19]
*/
		  
void net_addr_format_convert(p1, p2)
register char *p1;
register char *p2;
{
    register int i;
				 
    for(i = 0; i < 12; i+=2) {
	*p2++ = *p1++;
	*p2++ = *p1++;
	*p2++ = ' ';
    }

    *p2 = '\0';
}


/*
 *
 * net_zeroaddr - zeroes a network address
 *
 */

int net_zeroaddr (p)
NetworkAddress *p;
{
    (void)bzero (p, sizeof (NetworkAddress));
    return 0;
}

#endif /* NO_ASSOC */
