#ifndef lint
static char rcsid [] = "$RCSfile: net_ctl.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_ctl.c
 *
 * Copyright 1995 Loral Advanced Distributed Simulation
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform control functions.
 *
 * Functions:
 *		net_alive
 *		net_iocontrol
 *		net_nopened
 *		net_res
 *		net_settimeout
 *		net_loopback
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "sys/types.h"
#include "libnet.h"


/*
 *	net_alive
 *
 *	Check if enp is alive.  Returns 0 if successful, -1 if not.
 */

int net_alive(h)
int h;
{
    CHECK_HANDLE(h, -1);

    if(errno = net_access(h, ENPALIVE))
      return(-1);
    else
      return(0);
}

/*
 * net_iocontrol
 *
 * IO control.  Sends buffer to filter routine and receives buffer back.
 * Returns -1 if fails, 0 if ok.
 */

int net_iocontrol(h, sendbuf, sendsize, recvbuf, recvsize)
int h;
char *sendbuf;
int sendsize;
char *recvbuf;
int recvsize;
{
    register struct netlib_pb *np;
    int ret = 0;
#ifdef SIMBFLY
    char *rbuf;
    OID rbufoid;
    char *sbuf;
    OID sbufoid;
#else
    struct enpaccess_arg arg;
#endif

    CHECK_HANDLE(h, -1);

    np = npb + h;

    if (np->type == NIF_TYPE_SIMLE)
    {
	return -1;
    }

#ifdef SIMBFLY

    if(sendsize) {
	sbufoid = Make_Obj((char)0, LOCAL_NODE, sizeof(int), RW_rw_);
	sbuf = Map_Obj(sbufoid, 0, RW_rw_);
	bcopy(sendbuf, sbuf, sendsize);
    } else
	sbufoid = NULL;

    if(recvsize) {
	rbufoid = Make_Obj((char)0, LOCAL_NODE, sizeof(int), RW_rw_);
	rbuf = Map_Obj(rbufoid, 0, RW_rw_);
    } else
	rbufoid = NULL;

    np->net_req->a.enpaccess_req.command = IOCONTROL;
    np->net_req->a.enpaccess_req.buffer_to_enp_OID = sbufoid;
    np->net_req->a.enpaccess_req.bytes_to_enp = sendsize;
    np->net_req->a.enpaccess_req.buffer_to_host_OID = rbufoid;
    np->net_req->a.enpaccess_req.bytes_to_host = recvsize;

#else

    arg.command = IOCONTROL;
    arg.buffer_to_enp = sendbuf;
    arg.bytes_to_enp = sendsize;
    arg.buffer_to_host = recvbuf;
    arg.bytes_to_host = recvsize;

    np->params = (char *)&arg;

#endif

    if(errno = net_access(h, ENPACCESS))
	ret = -1;

#ifdef SIMBFLY

    if(ret == 0) {
	if(sendsize)
	    Del_Obj(sbufoid);
	if(recvsize) {
	    bcopy(rbuf, recvbuf, recvsize);
	    Del_Obj(rbufoid);
	}
    }

#endif

    return(ret);
}

/*
 * net_nopened - return number of times interface has been opened
 *
 * Returns -1 if it fails with errno set and the number of times
 * the interface has been opened if successful.
 */

int net_nopened(h)
int h;
{
#ifdef MASSCOMP

    int id;
    struct shmid_ds buf;

    CHECK_HANDLE(h, -1);
    
    if((id = shmget((key_t) npb[h].nib.inringkey,
		    (int) ((Ringsize_in_chars(npb[h].inringnum) + 0xfff) & ~0xfff),
		    0)) < 0)
	return(-1);

    if(shmctl(id, IPC_STAT, &buf) < 0)
	return(-1);

    return(buf.shm_nattch);
#else

    printf("net_nopened: not implemented yet\n");

    return(-1);

#endif
}

/*
 * net_res - reset network controller
 *
 * This routine is called to do a hardware reset of the network 
 * controller.
 */

int net_res(h)
int h;
{
    CHECK_HANDLE(h, -1);
    
    if(errno = net_access(h, ENPRESET))
	return(-1);

    return(0);
}

/*
 *	net_settimeout
 *
 * 	Set timeout value for blocking accesses.  Value is in seconds.
 */

int net_settimeout(h, value)
int h;
int value;
{
    CHECK_HANDLE(h, -1);

    npb[h].timeout = value;

    return(0);
}

/*
 * net_loopback
 *
 * Turns loopback of packets on/off.  If the flag is 1, packets are looped.
 * If the flag is 0, they are not looped.  The default condition after 
 * initialization is that packets are not looped.  
 */

int net_loopback(h, flag)
int h;
int flag;
{
    register struct netlib_pb *np;
    int ret = 0;
    int command = flag ? ENABLE_LOOPBACK : DISABLE_LOOPBACK;
#ifndef SIMBFLY
    struct enpaccess_arg arg;
#endif

    CHECK_HANDLE(h, -1);

    np = npb + h;

#ifdef SIMBFLY

    np->net_req->a.enpaccess_req.command = command;
    np->net_req->a.enpaccess_req.bytes_to_enp = 0;
    np->net_req->a.enpaccess_req.bytes_to_host = 0;

#else

    arg.command = command;
    arg.buffer_to_enp = NULL;
    arg.bytes_to_enp = 0;
    arg.buffer_to_host = NULL;
    arg.bytes_to_host = 0;

    np->params = (char *)&arg;

#endif

    if(errno = net_access(h, ENPACCESS))
	ret = -1;

    return(ret);
}


#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
