#ifndef lint
static char rcsid [] = "$RCSfile: net_dlpi.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*
 * File:	net_dlpi.c
 */

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifdef STREAMS_NETIF

/* net_dlpi.c is used to interface to the data link provider interface.
 * It has been used on Solaris 2.4.  It may work on other SVR4 systems too.
 * 
 * Partially derived from net_sock.c.
 */


/* Portions of this file come from pcap-dlpi.c, part of the libpcap packet-
 * capture library, part of the tcpdump distribution.  These parts are under
 * the following copyright:
 */

/*
 * Copyright (c) 1993, 1994, 1995
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * This code contributed by Atanu Ghosh (atanu@cs.ucl.ac.uk),
 * University College London.
 */


#include <stdext.h>  /*common/include/global*/
#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/bufmod.h>
#include <sys/dlpi.h>
#include <sys/stream.h>
#include <sys/time.h>
#include <stropts.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include "libnetif.h"
#include <stdstring.h> /*common/include/global*/


/* 080008 is BBN specific enet addressing */
NetworkAddress fake_from =     {0x08, 0x0, 0x08, 0xff, 0xff, 0xff};
NetworkAddress FillinAddress = {0x08, 0x0, 0x08, 0xff, 0xff, 0xff};

NetworkAddress BroadcastAddress = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

#define IFNAMSIZ 15
#define MAXDLBUF        8192

struct padded_packet {
    NetworkPacket pkt;
};

static struct {
    char            interface_name[IFNAMSIZ];
    NetworkAddress  sndFrom;	/* source address when sending */
    int             sndType;	/* packet type when sending */
    NetworkAddress  rcvTo;	/* dest address of last packet rcvd */
    NetworkAddress  rcvFrom;	/* source address of last packet rcvd */
    int             rcvType;	/* packet type of last packet rcvd */
    int             n_types;	/* number of types accepted */
    int             types[256];	/* types accepted */
    int             timestamp;	/* time of last received packet */

    /* Statistics
     */
    int if_dropped;
    int sb_dropped;
    int rcv_count;
    int snd_count;

    /* These next three are really nonsense with this kind of a driver.
     * But they are supported for compatibility.
     */
    char            snd_buffer[NETWORK_BUFFER_SIZE];
    char            rcv_buffer[NETWORK_BUFFER_SIZE];
    int             rcv_released;
} sockinfo[256];    


/*
 * DLPI Support Routines
 *
 */

static int send_request(int fd, char *ptr, int len, char *what, char *ebuf)
{
    struct	strbuf	ctl;
    int	flags;
    
    ctl.maxlen = 0;
    ctl.len = len;
    ctl.buf = ptr;
    
    flags = 0;
    if (putmsg(fd, &ctl, (struct strbuf *) NULL, flags) < 0) {
	sprintf(ebuf, "putmsg \"%s\"failed: %s",
		what, strerror(errno));
	return (-1);
    }
    return (0);
}

static int dlattachreq(int fd, u_long ppa, char *ebuf)
{
    dl_attach_req_t	req;
    
    req.dl_primitive = DL_ATTACH_REQ;
    req.dl_ppa = ppa;
    
    return (send_request(fd, (char *)&req, sizeof(req), "attach", ebuf));
}

static int dlpromisconreq(int fd, u_long level, char *ebuf)
{
    dl_promiscon_req_t req;
    
    req.dl_primitive = DL_PROMISCON_REQ;
    req.dl_level = level;
    
    return (send_request(fd, (char *)&req, sizeof(req), "promiscon", ebuf));
}

static int dlokack(int fd, char *bufp, char *ebuf)
{
    union DL_primitives	*dlp;
    struct strbuf	 ctl;
    int			 flags;
    
    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;
    
    flags = 0;
    if (getmsg(fd, &ctl, (struct strbuf*)NULL, &flags) < 0) {
	sprintf(ebuf, "getmsg: %s", strerror(errno));
	return (-1);
    }
    
    dlp = (union DL_primitives *) ctl.buf;
    
    if (dlp->dl_primitive != DL_OK_ACK) {
	sprintf(ebuf, "dlokack unexpected primitive %d",
			dlp->dl_primitive);
	return (-1);
    }
    
    if (ctl.len != sizeof(dl_ok_ack_t)) {
	sprintf(ebuf, "dlokack incorrect size returned");
	return (-1);
    }
    return (0);
}


static int dladdrreq(int fd, char *ebuf)
{
    dl_phys_addr_req_t req;
    
    req.dl_primitive = DL_PHYS_ADDR_REQ;
    req.dl_addr_type = DL_FACT_PHYS_ADDR;
    
    return (send_request(fd, (char *)&req, sizeof(req), "addr", ebuf));
}

static int dladdrack(int fd, char *bufp, char *ebuf)
{
    union	DL_primitives	*dlp;
    struct	strbuf	ctl;
    int	flags;
    
    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;
    
    flags = 0;
    if (getmsg(fd, &ctl, (struct strbuf *)NULL, &flags) < 0) {
	sprintf(ebuf, "dladdrack: getmsg: %s", strerror(errno));
	return (-1);
    }
    
    dlp = (union DL_primitives *) ctl.buf;
    
    if (dlp->dl_primitive != DL_PHYS_ADDR_ACK) {
	sprintf(ebuf, "dladdrack: unexpected primitive %ld",
		dlp->dl_primitive);
	return (-1);
    }
    return (0);
}

static int dlbindreq(int fd, char *ebuf)
{
    dl_bind_req_t req;
    
    req.dl_primitive = DL_BIND_REQ;
    req.dl_sap = 0;   /* Means 802.3 */
    req.dl_max_conind = 0;
    req.dl_service_mode = DL_CLDLS;
    req.dl_conn_mgmt = 0;
    req.dl_xidtest_flg = 0;
    
    return (send_request(fd, (char *)&req, sizeof(req), "bind", ebuf));
}

static int dlbindack(int fd, char *bufp, char *ebuf)
{
    union	DL_primitives	*dlp;
    struct	strbuf	ctl;
    int	flags;
    
    ctl.maxlen = MAXDLBUF;
    ctl.len = 0;
    ctl.buf = bufp;
    
    flags = 0;
    if (getmsg(fd, &ctl, (struct strbuf *)NULL, &flags) < 0) {
	sprintf(ebuf, "dlbindack: getmsg: %s", strerror(errno));
	return (-1);
    }
    
    dlp = (union DL_primitives *) ctl.buf;

    if (dlp->dl_primitive != DL_BIND_ACK) {
	sprintf(ebuf, "dladdrack: unexpected primitive %ld",
		dlp->dl_primitive);
	return (-1);
    }
    return (0);
}

static int dlenablemultireq(int fd, NetworkAddress *pna, char *ebuf)
{
    unsigned char buf[512];
    dl_enabmulti_req_t *req = (dl_enabmulti_req_t *)buf;
    
    req->dl_primitive = DL_ENABMULTI_REQ;
    req->dl_addr_length = sizeof(NetworkAddress);
    req->dl_addr_offset = DL_ENABMULTI_REQ_SIZE;
    
    memcpy(&buf[DL_ENABMULTI_REQ_SIZE], pna->addr, sizeof(NetworkAddress));
    
    return (send_request(fd, (char *)buf, 
			 sizeof(NetworkAddress)+DL_ENABMULTI_REQ_SIZE,
			 "enable multicast", ebuf));
}



static int dldisablemultireq(int fd, NetworkAddress *pna, char *ebuf)
{
    unsigned char buf[512];
    dl_disabmulti_req_t *req = (dl_disabmulti_req_t *)buf;

    req->dl_primitive = DL_DISABMULTI_REQ;
    req->dl_addr_length = sizeof(NetworkAddress);
    req->dl_addr_offset = DL_DISABMULTI_REQ_SIZE;
    
    memcpy(&buf[DL_DISABMULTI_REQ_SIZE], pna->addr, sizeof(NetworkAddress));

    return (send_request(fd, (char *)buf, 
			 sizeof(NetworkAddress)+DL_DISABMULTI_REQ_SIZE,
			 "disable multicast", ebuf));
}


static int strioctl(int fd, int cmd, int len, char *dp)
{
    struct strioctl str;
    int rc;
    
    str.ic_cmd = cmd;
    str.ic_timout = -1;
    str.ic_len = len;
    str.ic_dp = dp;
    rc = ioctl(fd, I_STR, &str);
    
    if (rc < 0)
      return (rc);
    else
      return (str.ic_len);
}



int net_open(device, flags, ignored)
    char *device;
    int   flags;
    int   ignored;
{
    int s;
    int ppa;
    int cppa;
    dl_info_ack_t *infop;
    dl_phys_addr_ack_t *addrp;
    unsigned long ss, flag;
    int cc;
    int on = 1;
    char dname[100];
    char cbuf[BUFSIZ];
    int s2;
    char ebuf[256];
    long buf[MAXDLBUF];

    /*
     * 1) In order to get the ppa take the last character of the device
     * name if it is a number then fail the open.
     *
     * 2) If the name starts with a '/' then this is an absolute pathname,
     * otherwise prepend '/dev/'.
     *
     * 3) Remove the trailing digit and try and open the device
     * not staggeringly intuitive but it should work.
     *
     * If there are more than 9 devices this code will fail.
     */
    
    cppa = device[strlen(device) - 1];
    if (!isdigit(cppa)) 
    {
	printf("%c is not a digit, therefore not a valid ppa", cppa);
	return -1;
    }
    
    dname[0] = '\0';
    if (device[0] != '/')
      strcpy(dname, "/dev/");
    
    strcat(dname, device);
    dname[strlen(dname) - 1] = '\0';
    
    if ((s = open(dname, O_RDWR | O_NDELAY | O_NONBLOCK)) < 0) 
    {
	printf("%s: Could not open stream.\n", dname);
	return -1;
    }
    ppa = cppa - '0';
    
    /*
     * Attach.
     */
    if (dlattachreq(s, ppa, ebuf) < 0 ||
	dlokack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error opening network: %s.\n", ebuf); 
	return -1;
    }

    /*
     * enable multicast
     */
    if (dlpromisconreq(s, DL_PROMISC_SAP, ebuf) < 0 ||
	dlokack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error opening network: %s.\n", ebuf); 
	return -1;
    }
#if 0 
    if (dlpromisconreq(s, DL_PROMISC_MULTI, ebuf) < 0 ||
	dlokack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error opening network: %s.\n", ebuf); 
	return -1;
    }
#endif
    /*
     * Determine physical address of device
     */
    if (dladdrreq(s, ebuf) < 0 ||
	dladdrack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error opening network: %s.\n", ebuf); 
	return -1;
    }
    
    addrp = &((union DL_primitives *)buf)->physaddr_ack;

    memcpy(&fake_from, (char *)(&addrp) + addrp->dl_addr_offset,
	  addrp->dl_addr_length);

    
#ifdef  DLIOCRAW
    /*
     * This is a non standard SunOS hack to get the ethernet header.
     */
    if (strioctl(s, DLIOCRAW, 0, NULL) < 0) 
    {
	printf("Error setting stream to raw.\n");
	return -1;
    }
#endif

#if 0    
    /*
     * Another non standard call to get the data nicely buffered
     */
    if (ioctl(s, I_PUSH, "bufmod") != 0) 
    {
	printf("Error opening network: I_PUSH bufmod: %s", strerror(errno));
	return -1;
    }
#endif

    /* 
     * Bind to stream
     */

    if (dlbindreq(s, ebuf) < 0 ||
	dlbindack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error opening network: %s.\n", ebuf); 
	return -1;
    }

    /* Initialize local info */
    bzero(&sockinfo[s], sizeof(sockinfo[s]));
    strncpy(sockinfo[s].interface_name, device, IFNAMSIZ);
    sockinfo[s].sndFrom = fake_from;
    sockinfo[s].sndType = PROTOCOL_SIM;
    sockinfo[s].n_types = 1;
    sockinfo[s].types[0] = PROTOCOL_SIM;
    sockinfo[s].rcv_released = 1;

    
    return s;
}

int net_close(s)
    int s;
{
    return close(s);
}

int net_send(s, to, buf, len, flags)
    int             s;
    NetworkAddress *to;
    char           *buf;
    int             len;
    int             flags;
{
    static int warned = 0;

    if (!warned)
    {
	printf("Warning: Solaris net_send driver does not support ethernet 2.\n         Calling net_snd instead\n");
	warned = 1;
    }
    return net_snd(s, to, buf, len, flags);
}

int net_snd(s, to, buf, len, flags)
    int             s;
    NetworkAddress *to;
    char           *buf;
    int             len;
    int             flags;
{
    struct padded_packet pkt;
    unsigned int wsize;
    struct strbuf dbuf;

    if(len > MAX_DATA_SIZE_8023)
      len = MAX_DATA_SIZE_8023;

    pkt.pkt.h.hdr8023.to = *to;
    pkt.pkt.h.hdr8023.from = sockinfo[s].sndFrom;

    if(!(flags & NETNOLEN) && len < MIN_DATA_SIZE_8023)
      pkt.pkt.h.hdr8023.length = ENETMINSIZE;
    else
      pkt.pkt.h.hdr8023.length = len + SAP_SNAP_SIZE;

    SET_DSAP(pkt.pkt.data, SAP_SNAP);
    SET_SSAP(pkt.pkt.data, SAP_SNAP);
    SET_CONTROL(pkt.pkt.data, UNNUMBERED_DATA);
    SET_PROTOID(pkt.pkt.data, ORG_CODE_BBN);
    SET_ETHER_TYPE(pkt.pkt.data, sockinfo[s].sndType);

    memcpy(GET_DATA_PTR(pkt.pkt.data), (char *)buf, len);

    wsize = pkt.pkt.h.hdr8023.length + sizeof(NetworkHeader8023);

    dbuf.len = wsize;
    dbuf.buf = (caddr_t)&pkt.pkt;

    if (putmsg(s, NULL, &dbuf, 0) < 0)
      return -1;

    sockinfo[s].snd_count += 1;

    return 0;
}

static int check_type(type, n, arr)
    int type;
    int n;
    int arr[];
{
    int i;

    for(i=0;i<n;i++)
      if (arr[i] == type)
	return 1;
    return 0;
}

int net_recv(s, buf, bufsize, plen, flags)
    int   s;
    char *buf;		/* Put packet here (w/o header) */
    int   bufsize;	/* Size of buf */
    int  *plen;		/* Return bytes in buf here */
    int   flags;
{
    static int warned = 0;

    if (!warned)
    {
	printf("Warning: Solaris net_recv driver does not support ethernet 2.\n         Calling net_rcv instead\n");
	warned = 1;
    }
    return net_rcv(s, buf, bufsize, plen, flags);
}

int net_rcv(s, buf, bufsize, plen, flags)
    int   s;
    char *buf;		/* Put packet here (w/o header) */
    int   bufsize;	/* Size of buf */
    int  *plen;		/* Return bytes in buf here */
    int   flags;
{
    int ret = 0;                /* assume success */
    int cc;
    struct padded_packet pkt;
    int size;
    char *datap;
    int off = 0, on = 1;
    struct strbuf dat_ctl;
    int flag=0;
    long buffer[MAXDLBUF];


    for(;;)
    {
	dat_ctl.maxlen = sizeof(pkt);
	dat_ctl.buf    = (char *)&pkt.pkt;

	if (getmsg(s, NULL, &dat_ctl, &flag) < 0)
	{
	    /* Don't choke when we get ptraced */
	    if (errno == EINTR) 
	    {
		continue;
	    }
	    
	    /* If no data, return */
	    if (errno == EAGAIN)
	    {
		ret = -1;
		break;
	    }

	    printf("Error reading from network: %s", strerror(errno));
	    return -1;
	}
	cc = dat_ctl.len;

	if (cc == 0)
	{
	    /* I don't trust read never to do this */
	    errno = EWOULDBLOCK;
	    ret = -1;
	    break;
	}

	sockinfo[s].rcv_count += 1;

	/* Note the time. */
	sockinfo[s].timestamp = net_current_time(s);


	if (pkt.pkt.h.hdr8023.length > ENETMAXSIZE)
	  continue;
	else
	{
	    sockinfo[s].rcvTo = pkt.pkt.h.hdr8023.to;
	    sockinfo[s].rcvFrom = pkt.pkt.h.hdr8023.from;
	    sockinfo[s].rcvType = GET_ETHER_TYPE(pkt.pkt.data);
	    size = cc-sizeof(NetworkHeader8023)-SAP_SNAP_SIZE;
	    datap = (char *)GET_DATA_PTR(pkt.pkt.data);
	}
       
	/* Do we want this type? */
	if(GET_PROTOID(pkt.pkt.data) != ORG_CODE_BBN)
	  continue;
	if (!check_type(sockinfo[s].rcvType,
			sockinfo[s].n_types, sockinfo[s].types))
	  continue;

        /* If packet is bigger than the buffer bomb out. */
        if (size > bufsize)
	{
	    errno = E2BIG;
	    ret = -1;
	}
	else
	{
	    /* Return data to caller's buffer. */
	    memcpy(buf, datap, size);

	    /* Return size of packet == data size. */
	    *plen = size;
	}
	break;
    }

    return ret;
}

int net_get_rcv(s, buf, plen, flags)
    int    s;
    char **buf;			/* returned pointer to data */
    int   *plen;		/* bytes in *buf */
    int    flags;
{
    /* Calling this the first time gets you a new packet.
     * Successive calls give you the same packet until net_release_rcv
     * is called.
     */

    *buf = sockinfo[s].rcv_buffer;

    if (sockinfo[s].rcv_released)
    {
	sockinfo[s].rcv_released = 0;
	return net_rcv(s, sockinfo[s].rcv_buffer,
		       NETWORK_BUFFER_SIZE, plen, flags);
    }
    else
      return 0;
}

int net_release_rcv(s)
    int s;
{
    sockinfo[s].rcv_released = 1;
    return 0;
}

int net_get_snd(s, buf, plen, flags)
    int s;
    char **buf;			/* returned pointer to data */
    int *plen;			/* bytes in *buf */
    int flags;			/* should be zero for now */
{
    /* Calling this gives you a pointer back to a place to put a
     * packet.  The packet is sent with net_release_snd.
     */
    *buf = sockinfo[s].snd_buffer;
    *plen = MAX_DATA_SIZE_8023;
    return 0;
}

int net_release_snd(s, to, len, flags)
    int s;
    NetworkAddress *to;		/* destination */
    int len;			/* bytes to send */
    int flags;
{
    /* Sends the packet in the send buffer */
    return net_snd(s, to, sockinfo[s].snd_buffer, len, flags);
}

#include <sys/time.h>

int net_current_time(s)
    int s;
{
    struct timeval t;
    struct timezone tz;

    gettimeofday(&t, &tz);

    return t.tv_sec * 1000 + t.tv_usec / 1000;
}

int net_gettime(s)
    int s;
{
    return net_current_time(s);
}

int net_set_snd_type(s, type)
    int s;
    int type;
{
    sockinfo[s].sndType = type;
    return 0;
}

int net_set_snd_from_addr(s, from)
    int s;
    NetworkAddress *from;
{
    sockinfo[s].sndFrom = *from;
    return 0;
}

int net_get_rcv_to_addr(s, to)
    int s;
    NetworkAddress *to;
{
    *to = sockinfo[s].rcvTo;
    return 0;
}

int net_get_rcv_from_addr(s, from)
    int s;
    NetworkAddress *from;
{
    *from = sockinfo[s].rcvFrom;
    return 0;
}

int net_get_rcv_type(s)
    int s;
{
    return sockinfo[s].rcvType;
}

int net_addr_compare(a1, a2)
    NetworkAddress *a1;
    NetworkAddress *a2;
{
    return (!bcmp((char *)a1, (char *)a2, sizeof(NetworkAddress)));
}

int net_add_mca(s, pna)
    int s;
    NetworkAddress *pna;
{
    char ebuf[256];
    long buf[MAXDLBUF];

    if (dlenablemultireq(s, pna, ebuf) < 0 ||
	dlokack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error subscribing to multicast\n");
	return -1;
    }
    return 0;
}

int net_del_mca(s, pna)
    int s;
    NetworkAddress *pna;
{
    char ebuf[256];
    long buf[MAXDLBUF];

    if (dldisablemultireq(s, pna, ebuf) < 0 ||
	dlokack(s, (char *)buf, ebuf) < 0)
    {
	printf("Error unsubscribing from multicast\n");
	return -1;
    }
    return 0;
}


int net_init_type(s)
    int s;
{
    sockinfo[s].n_types = 0;
    return 0;
}

int net_add_type(s, type)
    int s;
    int type;
{
    sockinfo[s].types[sockinfo[s].n_types++] = type;
    return 0;
}

int net_addr_str_to_bin(p, s)
    NetworkAddress *p;
    char *s;
{
    int i, x;
    
    for(i = 0; i < ENETADDRSZ; i++)
        p->addr[i] = 0;

    /* check for exactly 12 digits, all hex digits */
    for (i = 0; i < ENETADDRSZ * 2; i++) {
	if(s[i] >= '0' && s[i] <= '9')
	    continue;
	else if(s[i] >= 'a' && s[i] <= 'f')
	    continue;
	else if(s[i] >= 'A' && s[i] <= 'F')
	    continue;
	else
            return 0;
    }

    if (s[i] != '\0')
	return 0;

    for (i = 0; i < ENETADDRSZ; i++, s += 2) {
	sscanf (s, "%2x", &x);
	p->addr[i] = x;
    }

    return 1;
}

int net_flush(s, flags)
    int s;
    int flags;
{
    char buf[NETWORK_BUFFER_SIZE];
    int len;

    if (flags & NETFLUSHRCV)
    {
	while (net_rcv(s, buf, NETWORK_BUFFER_SIZE, &len, 0) == 0)
	  ;
    }

    if (flags & NETFLUSHXMT)
    {
	/* NOP */;
    }

    return 0;
}

#define N_STATS 4		/* set to number of strings */
char statistics_strings[][41] = {
		"successful transmissions                ",
		"packets received                        ",
		"missed packets reported (hardware)      ",
		"missed packets reported (software)      ",
	};
int max_statistics_string = N_STATS;

int net_zero_statistics(s)
    int s;
{
    return 0;
}

int net_get_statistics(s, stats)
    int s;
    long *stats;
{
    return 0;
}


void net_stat_string(stat, buf)
    int stat;
    char *buf;
{
    if(stat >= max_statistics_string)
      buf[0] = '\0';
    else
      strcpy(buf, &statistics_strings[stat][0]);
}

int net_print_statistics(s)
    int s;
{
    long stats[N_STATS];
    char str[60];
    register int i;

    if (net_get_statistics (s, stats) == -1)
      return -1;

    for (i = 0; i < N_STATS; i++)
    {
	net_stat_string (i, str);
	if (str[0] != '\0')
	  printf ("%s%ld\n", str, stats[i]);
    }
    return 0;
}

long net_get_timestamp(s)
int s;
{
    return(sockinfo[s].timestamp);
}

int net_put_timestamp(s, time)
int s;
long time;
{

    sockinfo[s].timestamp = time;

    return(0);
}

#define stub(name) int name() { return 0; }

stub(net_norm)
stub(net_stop)
stub(net_alive)
stub(net_init_mca)
stub(net_iocontrol)

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */

