#ifndef lint
static char rcsid [] = "$RCSfile: net_info.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_info
 *
 * Copyright (c) 1988 by BBN Systems and Technologies, Inc.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * Functions:
 *		net_bufs
 *		net_hostbuf_info
 *		net_interface_type
 *		net_sharebuf_info
 *		net_sysstat_info
 *		net_version
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


int net_bufs(h, pxmtbufs, prcvbufs)
int h;
int *pxmtbufs;
int *prcvbufs;
{
    int bufs[2];

    CHECK_HANDLE (h, -1);

#ifndef SIMBFLY

    npb[h].params = (char *)bufs;

#endif

    if(errno = net_access(h, ENPBUFS))
	return(-1);

#ifndef SIMBFLY

    *pxmtbufs = bufs[0];
    *prcvbufs = bufs[1];

#else

    *pxmtbufs = npb[h].net_req->a.bufs_req.xmtbufs;
    *prcvbufs = npb[h].net_req->a.bufs_req.rcvbufs;

#endif

    return(0);
}

/*
 * net_hostbuf_info
 *
 * Get pointer and size (in bytes) of the host buffer.  This is a block
 * of memory on the network card accessible both by the host and the card.
 */

int net_hostbuf_info(h, pbufptr, pbufsize)
int h;
char **pbufptr;
int *pbufsize;
{
    int bufs[2];

    CHECK_HANDLE (h, -1);

#ifndef SIMBFLY

    npb[h].params = (char *)bufs;

#endif

    if (npb[h].type == NIF_TYPE_SIMLE)
    {
	return -1;
    }

    if(errno = net_access(h, ENPGETHOSTBUF))
	return(-1);

    /* Add our pointer to the base to the base of the card to the
       offset of the buffer (relative to the base of the card). */

#ifdef MASSCOMP
    *pbufptr = npb[h].base_ptr + bufs[0];
    *pbufsize = bufs[1];
#endif

#ifdef SIMBFLY
    *pbufptr = npb[h].base_ptr + npb[h].net_req->a.hostbuf_req.ofs;
    *pbufsize = npb[h].net_req->a.hostbuf_req.size;
#endif

#ifdef _GT_
    *pbufptr = npb[h].base_ptr + bufs[0];
    *pbufsize = bufs[1];
#endif

    return(0);
}

/*
 * net_interface_type
 *
 * Returns interface type for the handle.  Current types are
 *	NIF_TYPE_CMC
 *	NIF_TYPE_147
 *	NIF_TYPE_SIMLE
 */

int net_interface_type(h)
int h;
{
    CHECK_HANDLE (h, -1);

    return(npb[h].type);
}

/*
 * net_sharebuf_info
 *
 * Get pointer and size (in bytes) of the shared buffer.  This is a block
 * of memory on the host accessible both by the host and the card.
 */

int net_sharebuf_info(h, pbufptr, pbufsize)
int h;
char **pbufptr;
int *pbufsize;
{
    CHECK_HANDLE (h, -1);

    *pbufptr = npb[h].share_ptr;
    *pbufsize = npb[h].share_size;

    return(0);
}

/*
 * net_syserror_info
 *
 * Get pointer to char array of system errors and also the number of
 * system errors.
 */

int net_syserror_info(h, perr, p_num)
int h;
char **perr;
int *p_num;
{
    CHECK_HANDLE (h, -1);

    *perr = npb[h].enp_ptr->syserrors;
    *p_num = NUM_SYSERRORS;

    return(0);
}

/*
 * net_version
 *
 * Get version.  Returns nul terminated string.
 *
 * Returns 0 if successful and -1 if not with errno set as follows:
 *    EBADF	bad file handle
 *    EINVAL	not supported by underlying network interface
 */
int net_version(h, version, version_size)
int h;
char *version;
int version_size;		/* size of version buffer */
{
    register struct netlib_pb *np;
    int ret = 0;
#ifdef SIMBFLY
    char *buf;
    OID bufoid;
#else
    struct enpaccess_arg arg;
    char buf[VERSION_SIZE];
#endif

    CHECK_HANDLE(h, -1);

    np = npb + h;

#ifdef SIMBFLY

    bufoid = Make_Obj((char)0, LOCAL_NODE, VERSION_SIZE, RW_rw_);
    buf = Map_Obj(bufoid, 0, RW_rw_);

    np->net_req->a.enpaccess_req.command = VERSION_STRING;
    np->net_req->a.enpaccess_req.bytes_to_enp = 0;
    np->net_req->a.enpaccess_req.buffer_to_host_OID = bufoid;
    np->net_req->a.enpaccess_req.bytes_to_host = VERSION_SIZE;

#else

    arg.command = VERSION_STRING;
    arg.bytes_to_enp = 0;
    arg.buffer_to_host = buf;
    arg.bytes_to_host = VERSION_SIZE;

    np->params = (char *)&arg;

#endif

    if(errno = net_access(h, ENPACCESS))
	ret = -1;

    bzero(version, version_size);
    bcopy(buf, version, version_size - 1);

#ifdef SIMBFLY

    Del_Obj(bufoid);

#endif

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
