#ifndef lint
static char rcsid [] = "$RCSfile: net_mode.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_mode.c
 *
 * Copyright 1995 Loral Advanced Distributed Simulation
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform mode control functions.
 *
 * Functions:
 *		net_extloop
 *		net_intloop
 *		net_norm
 *		net_prom
 *		do_mode_cmd
 *		do_mode_cmd_cmc
 *		do_mode_cmd_147
 *		do_mode_cmd_simle
 *
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


/********************************************
 *	net_extloop
 *
 *	Begin execution in external loopback mode
 */
int net_extloop(h, na)
int h;
NetworkAddress *na;
{
    return (do_mode_cmd (ENPEXTLOOP, h, na));
}


/*********************************************
 *	net_intloop
 *
 *	Begin execution in internal loopback mode
 */
int net_intloop(h, na)
int h;
NetworkAddress *na;
{
    return (do_mode_cmd (ENPINTLOOP, h, na));
}


/**********************************************
 *	net_norm
 *
 *	Begin execution in normal mode
 */
int net_norm(h, na)
int h;
NetworkAddress *na;
{
    return (do_mode_cmd (ENPNORM, h, na));
}

/**********************************************
 *	net_prom
 *
 *	Begin execution in promiscuous mode
 */
int net_prom(h, na)
int h;
NetworkAddress *na;
{
    return (do_mode_cmd (ENPPROM, h, na));
}

/*****************************************
 *	do_mode_cmd
 *
 */

int do_mode_cmd (mode, h, na)
int mode;
int h;
NetworkAddress *na;
{
    struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = npb+h;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    ret = do_mode_cmd_cmc(mode, h, na, np);
	    break;
	case NIF_TYPE_147:
	    ret = do_mode_cmd_147(mode, h, na, np);
	    break;
	case NIF_TYPE_SIMLE:
	    ret = do_mode_cmd_simle(mode, h, na, np);
	    break;
	default:
	    printf("do_mode_cmd: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

int do_mode_cmd_cmc(mode, h, na, np)
int mode;
int h;
NetworkAddress *na;
struct netlib_pb *np;
{
#ifndef SIMBFLY
    static struct enprun_arg args;
#endif

#ifndef SUN
#ifndef _GT_
/* current* not set up when rings on board. */
/* Zero the rings to flush out any left over packets. */
    SETLONG(np->currentinbuf, 0);
    (void)bzero(np->inbuffers, (int) Ringsize_in_chars(np->inringnum - 1));
    SETLONG(np->currentoutbuf, 0);
    (void)bzero(np->outbuffers, (int) Ringsize_in_chars(np->outringnum - 1));
#endif
#endif

#ifdef SIMBFLY
    bcopy(na->addr, np->net_req->a.start_req.enetaddr, ENETADDRSZ);
#else
    bcopy (na->addr, args.enetaddr, ENETADDRSZ);
    args.inr = (char *)np->inbuffers;
    args.ins = sizeof (RingElement);
    args.inn = np->nib.inringsize;
    args.outr = (char *)np->outbuffers;
    args.outs = sizeof (RingElement);
    args.outn = np->nib.outringsize;
    args.share_address = (char *)np->share_ptr;
    args.share_size = np->share_size;
    args.hostbufsize = np->nib.hostbufsize;
    args.mcanum = np->nib.mca_num;

    np->params = (char *)&args;
#endif
    
    if(errno = net_access(h, mode))
	return(-1);

#ifndef SIMBFLY
/* This sets up the offsets to the rings.  It's a no op when the rings are
   on the host but is essential when the rings are on the card. */
    np->nib.inringoff = (int)args.inr;
    np->nib.outringoff = (int)args.outr;
#endif

    return (0);
}

int do_mode_cmd_147(mode, h, na, np)
int mode;
int h;
NetworkAddress *na;
struct netlib_pb *np;
{
    int ret = 0;

#ifdef _GT_
    if(errno = (*np->ioctl)(np->ccb, mode, na))
	ret = -1;
#else
    printf("do_mode_cmd_147: not supported\n");
    ret = -1;
#endif

    return(ret);
}


int do_mode_cmd_simle(mode, h, na, np)
int mode;
int h;
NetworkAddress *na;
struct netlib_pb *np;
{
    int ret = 0;

    if (net_access(h, mode) != 0)
	ret = -1;

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
