#ifndef lint
static char rcsid [] = "$RCSfile: net_open.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_open.c
 *
 * Copyright (c) 1987 by BBN Laboratories.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform control functions.
 *
 * Functions:
 *		net_open
 *		net_get_parameters
 *		net_set_parameters
 *		get_type
 *		get_device_number
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include <stdstring.h> /*common/include/global*/
#include "libnet.h"


/*
 *	net_open
 *
 *	Open access to enp.  Takes server name, device number
 *	and flags (none presently defined  - use 0) as arguments.
 *	Returns a handle to use or -1 if fails.
 */

#ifdef SIMBFLY

int net_open(svr, flags)
char *svr;		/* also device name */
int flags;
{
    MBMAPREGION mb_mr;
    int h;
    char s[120];
    char *p; 
    register struct netlib_pb *np;
    int i;

    for(h = 0; h < NUM_NETHANDLES; h++) {
	if(npb[h].used == FALSE)
	    break;
    }

    if(h >= NUM_NETHANDLES) {
	errno = EMFILE;
	goto error1;
    }

    np = &npb[h];

    if(strlen(svr) > 80) {
	errno = ENODEV;
	goto error1;
    }

    np->net_device = get_device_number(svr);

    p = &svr[strlen(svr)];
    while (p >= svr && *p != '/')
	p--;
    strcpy(s, ++p);
    s[strlen(s) - 1] = '\0';
    strcat(s, ENPSVRQ_NAME);

    catch

        np->net_queue = (QH)Find_Value(s, NTYPE_OBJ);

    onthrow

	when(TRUE) {
	    errno = ENODEV;
	    goto error1;
	}

    endcatch

    np->net_event = Make_Event(0, 0, 0, 0L);
    np->net_reqoid = Make_Obj((char)0, -1,sizeof(struct enpsvr_req), RW_rw_);
    np->net_req = (struct enpsvr_req *)Map_Obj(np->net_reqoid, 0, RW_rw_);

    /* Check for error on first access in case of bad device number. */
    if(get_type(np, svr, flags) < 0)
	goto error2;

    if(errno = net_access(h, ENPGETXMTLOCK))
	goto error2;
    np->net_xmtlock = np->net_req->a.getxmtlock_req.lock;

    net_access(h, ENPGETRCVLOCK);
    np->net_rcvlock = np->net_req->a.getrcvlock_req.lock;

    net_access(h, ENPGETINRING);
    np->inringoid = np->net_req->a.getinring_req.ringoid;
    np->inringnum = np->net_req->a.getinring_req.num;

    p = Map_Obj(np->inringoid, 0, RW_rw_);
    np->currentinbuf = (int *)p;
    np->inbuffers = (RingElement *)(p + sizeof(RingElement)); 

    net_access(h, ENPGETOUTRING);
    np->outringoid = np->net_req->a.getoutring_req.ringoid;
    np->outringnum = np->net_req->a.getoutring_req.num;

    p = Map_Obj(np->outringoid, 0, RW_rw_);
    np->currentoutbuf = (int *)p;
    np->outbuffers = (RingElement *)(p + sizeof(RingElement)); 

    net_access(h, ENPGETSHAREBUF);
    np->share_oid = np->net_req->a.sharebuf_req.oid;
    np->share_size = np->net_req->a.sharebuf_req.size;

    p = Map_Obj(np->share_oid, 0, RW_rw_);
    np->share_ptr = p;

    net_access(h, ENPADDRESS);
    mb_mr.mbstart    = (MBADDR)np->net_req->a.enp_addr_req.address;
    mb_mr.len        = ENP_MAPSIZE;
    mb_mr.ismemspace = TRUE;
    mb_mr.access     = RW_rw_;
    mb_mr.swap       = FALSE;

    np->base_ptr = map_mbregion(&mb_mr);
    np->reg_ptr = (long *)((char *)np->base_ptr + REG_REG_OFS);
    np->enp_ptr = (struct enp_registers *)((char *)np->base_ptr+ENP_REG_OFS); 

    np->used = TRUE;
    np->timeout = 0;

    return(h);

error2:
    Del_Obj(np->net_event);
error1:
    return(-1);
}

#else

int net_open (device, flags, nib)
char *device;			/* name of device to use */
int flags;			/* options for opening the network I/F */
struct net_info_block *nib;	/* supplied during initialization only */
{
    int h;
    register struct netlib_pb *np;
    int ret;

    for (h = 0; h < NUM_NETHANDLES; h++) {
	if (!npb[h].used)
	    break;
    }

    if (h >= NUM_NETHANDLES) {
	errno = EMFILE;
	return(-1);
    }

    np = &npb[h];

    bzero(np, sizeof (struct netlib_pb));

    if (strlen (device) > 40) {
	errno = ENODEV;
	return(-1);
    }

    if(get_type(np, device, flags) < 0)
	return(-1);

    np->used = TRUE;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    ret = open_cmc(device, flags, nib, h, np);
	    break;
	case NIF_TYPE_147:
	    ret = open_147(device, flags, nib, h, np);
	    break;
	case NIF_TYPE_SIMLE:
	    ret = open_simle(device, flags, nib, h, np);
	    break;
	default:
	    printf("net_open: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

open_cmc(device, flags, nib, h, np)
char *device;			/* name of device to use */
int flags;			/* options for opening the network I/F */
struct net_info_block *nib;	/* supplied during initialization only */
int h;				/* handle */
struct netlib_pb *np;
{
    np->enp_base = NULL;
    np->params = (char *)&np->enp_base;
    np->fd = -1;
    strcpy (np->dev_name, device);	/* must be before first access */
    if(net_access(h, ENPADDRESS) != 0)	/* first access */
	goto exit1;

    /* the device name should be /dev/enp0 or /dev/enp1 for now */
    np->net_device = get_device_number(device);
    np->timeout = 0;
    np->base_ptr = NULL;

    if (flags & NETINIT)
	bcopy (nib, &np->nib, sizeof (*nib));
    else {
	if((flags & NETNOCHECKALIVE) == 0)
	    if (net_alive (h) < 0)
		goto exit1;
        (void)bzero(&np->nib, sizeof(np->nib));
	if (net_get_parameters (h, &np->nib, TRUE) < 0)
	    goto exit1;
    }

    if (get_locks (np) < 0)
	goto exit1;

    if (map_buffers (np, flags) < 0)
	goto exit2;

    if (map_enp (np) < 0)
	goto exit3;

    return (h);

exit3:
    (void)unmap_buffers (np);
exit2:
    (void)unget_locks (np);
exit1:
    if (np->fd >= 0)
	XXX_CLOSE (np->fd);
    np->used = FALSE;
    return (-1);
}

open_147(device, flags, nib, h, np)
char *device;			/* name of device to use */
int flags;			/* options for opening the network I/F */
struct net_info_block *nib;	/* supplied during initialization only */
int h;				/* handle */
struct netlib_pb *np;
{
#ifdef _GT_

    NetworkAddress na;


    if (errno = ifx_open(device, IFXMRDWR, &np->fd))
	goto exit1;

    if (errno = ifx_ioctl(np->fd, ENPGETCCB, &np->ccb)) 
	goto exit2;

    if (errno = ifx_ioctl(np->fd, ENPGETREAD, &np->read))
	goto exit2;

    if (errno = ifx_ioctl(np->fd, ENPGETWRITE, &np->write))
	goto exit2;

    if (errno = ifx_ioctl(np->fd, ENPGETIOCTL, &np->ioctl))
	goto exit2;

    np->base_time = sc_gtime();

    /* Oh, what a hack!!! */
    net_getaddr(h, &na);
    net_add_mca(h, &na);
    net_add_mca(h, &BroadcastAddress);

    return h;

exit2:
    ifx_close(np->fd);
    np->used = FALSE;
exit1:
    return(-1);

#else

    printf("open_147: not supported\n");
    return(-1);

#endif
}

open_simle(device, flags, nib, h, np)
char *device;			/* name of device to use */
int flags;			/* options for opening the network I/F */
struct net_info_block *nib;	/* supplied during initialization only */
int h;				/* handle */
struct netlib_pb *np;
{
#ifdef sun4
int p;
char dev_name[32];
int dev_minor;

    dev_minor = atoi(strpbrk(device, "0123456789"));
    if (dev_minor >= 100)
    {
	np->fd = open (dev_name, O_RDWR);
	if (np->fd > 0)
	    np->micro = dev_minor - ((dev_minor / 100) * 100);
    }
    else
    {

	for (np->micro = 0; np->micro < SIMLE_MAX_HANDLES; np->micro++)
	{
	    sprintf(dev_name,"%s%02d", device, np->micro);

	    np->fd = open (dev_name, O_RDWR);
	    if (np->fd > 0)
		break;
	}
    }

    if (np->fd < 0)
	return -1;

    strcpy (np->dev_name, dev_name);
    net_getaddr(h, &np->sndFrom);

    if (net_create_inq(h, 512) == 0)
	return h;
    else
    {
	close(np->fd);
	return -1;
    }

#endif /* SUN */

   return -1;

}

/***********************************************
 *
 *	net_get_parameters
 *	net_set_parameters
 *
 *	These routines get or set the network parameter block.  Either
 * the local (to this opening) or card version of the parameters is used,
 * according to the setting of the flag.
 *
 *************************************************/

net_set_parameters (h, nib, flag)
int h;
struct net_info_block *nib;
int flag;
{
    CHECK_HANDLE (h, -1);

    if (flag == FALSE)
	bcopy (nib, &npb[h].nib, sizeof (struct net_info_block));
    else if (flag == TRUE) {
#ifdef mips
	char *ptr = (char *)&npb[h].nib;
	     
	npb[h].params = (char *)&ptr;
	if (net_access (h, ENPSETPARAMS) != 0)
	    return (-1);
#endif

#ifdef SUN
	char *ptr = (char *)&npb[h].nib;
	     
	npb[h].params = (char *)&ptr;
	if (net_access (h, ENPSETPARAMS) != 0)
	    return (-1);
#endif

#ifndef mips
#ifndef SUN
	npb[h].params = (char *)&npb[h].nib;
	if (net_access (h, ENPSETPARAMS) != 0)
	    return (-1);
	npb[h].params = (char *)&npb[h].nib.new;
	if (net_access (h, ENPSETPARAMS2) != 0)
	    if(errno != EINVAL)		/* backwards compatibility */
	        return (-1);
#endif
#endif
    }

    return (0);
}

net_get_parameters (h, nib, flag)
int h;
struct net_info_block *nib;
int flag;
{
    CHECK_HANDLE (h, -1);

    if (flag == FALSE)
	bcopy (&npb[h].nib, nib, sizeof (struct net_info_block));
    else if (flag == TRUE) {
#ifdef mips
	char *ptr = (char *)&npb[h].nib;
	     
	npb[h].params = (char *)&ptr;
	if (net_access (h, ENPGETPARAMS) != 0)
	    return (-1);
#endif

#ifdef SUN
	char *ptr = (char *)&npb[h].nib;
	     
	npb[h].params = (char *)&ptr;
	if (net_access (h, ENPGETPARAMS) != 0)
	    return (-1);
#endif

#ifndef mips
#ifndef SUN
	npb[h].params = (char *)nib;
	if (net_access (h, ENPGETPARAMS) != 0)
	    return (-1);
	npb[h].params = (char *)&nib->new;
	if (net_access (h, ENPGETPARAMS2) != 0)
	    if(errno != EINVAL)		/* backwards compatibility */
	        return (-1);
#endif
#endif
    }

    return (0);
}
#endif

get_type(np, dev, flags)
struct netlib_pb *np;
char *dev;			/* device name */
int flags;			/* flags from open */
{
#ifdef SIMBFLY

    /* Need to calculate handle from pointer difference: */
    if(errno = net_access(np - &npb[0], ENPTYPE)) 
	return(-1);

    np->enptype = np->net_req->a.type_req.enptype;

#endif /*  SIMBFLY */

#ifdef _GT_

    int ret;
    int desc;

    if(ret = ifx_open(dev, IFXMRDWR, &desc)) {
	errno = ret;
	return(-1);
    }
	
    if(ret = ifx_ioctl(desc, ENPTYPE, &np->enptype)) {
	if(ret == IFXENOTIMP)
	    np->enptype = ENPNOT;
	else {
	    errno = ret;
	    return(-1);
	}
    }

    ifx_close(desc);

#endif /*  _GT_ */

#ifndef _GT_
#ifndef BF_PLUS

    int h;

    if((h = XXX_OPEN(dev, 2)) < 0)
	return(-1);
	
    np->params = (char *)&np->enptype;
    if (XXX_IOCTL(h, ENPTYPE, np->params) < 0)
	return(-1);

    XXX_CLOSE(h);

#endif /*  BF_PLUS */
#endif /*  _GT_ */

    if(np->enptype == ENPNOT)
#ifdef _GT_
	np->type = NIF_TYPE_147;
#else
	np->type = NIF_TYPE_SIMLE;
#endif /*  _GT_ */
    else 
	np->type = NIF_TYPE_CMC;

    return(0);
}

get_device_number(device)
char *device;
{
    /* For now, always device 0. */
    return(0);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
