#ifndef lint
static char rcsid [] = "$RCSfile: net_orecv.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif


/*
 * File:	net_recv.c
 *
 * Copyright 1995 Loral Advanced Distributed Simulation
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform data access functions.
 *
 * Functions:
 *		net_recv 
 *		recv_cmc
 *		recv_147
 *		net_get_next_packet		(defunct)
 *		net_release_next_packet		(defunct)
 *		net_get_recv
 *		net_release_recv
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


int net_recv(h, hdr, buf, bufsize, plen, flags)
int h;
NetworkHeader *hdr;		/* Return header here if not NULL */
char *buf;			/* Put packet here (w/o header if hdr == NULL)*/
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
{
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    ret = recv_cmc(h, hdr, buf, bufsize, plen, flags, np);
	    break;
	case NIF_TYPE_147:
	    ret = recv_147(h, hdr, buf, bufsize, plen, flags, np);;
	    break;
	case NIF_TYPE_SIMLE:
	    ret = recv_simle(h, hdr, buf, bufsize, plen, flags, np);;
	    break;
	default:
	    printf("net_recv: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

int recv_cmc(h, hdr, buf, bufsize, plen, flags, np)
int h;
NetworkHeader *hdr;		/* Return header here if not NULL */
char *buf;			/* Put packet here (w/o header if hdr == NULL)*/
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
register struct netlib_pb *np;
{
    VOLATILE RingElement *rp;
    int tmp;
    int ret = 0;				/* assume success */
#ifndef SIMBFLY
    struct enprwsupport_arg blocking_args;
#endif

    switch(flags & (NETWAIT | NETLOCK | NETBLOCK)) {
	case 0:
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];
	    if(rp->full == FALSE) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case NETWAIT:
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];
	    wait_for_full_ring_element(rp);
	    break;
#ifndef mips /* NETLOCK not supported on mips */
	case NETLOCK:
	    if(!POLL_LOCK(np->net_rcvlock)) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];
	    if(rp->full == FALSE) {
		REMOVE_LOCK(np->net_rcvlock);
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case (NETWAIT | NETLOCK):
	    WAIT_FOR_LOCK(np->net_rcvlock);
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];
	    wait_for_full_ring_element(rp);
	    break;
#endif /*mips*/
	case NETBLOCK:
	    WAIT_FOR_LOCK(np->net_rcvlock);
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];
#ifndef SIMBFLY
	    blocking_args.pfull = (short *)&rp->full;
	    blocking_args.tout_val = np->timeout;
	    np->params = (char *)&blocking_args;
	    while(rp->full == FALSE) {
		if(net_access(h, ENPREADSUPPORT) != 0) {
		    REMOVE_LOCK (np->net_rcvlock);
		    errno = EWOULDBLOCK;
		    return(-1);
		}
	    }
#endif
	    break;
	default:
	    errno = EINVAL;
	    return(-1);
    }

    /* Save the timestamp for the packet. */
    np->timestamp = L_REFLECT(GETLONG(&rp->timestamp));

    /* Return NetworkHeader to caller separately if requested. */
    if (hdr != NULL) {
	/* If packet - hdr size is bigger than the buffer bomb out. */
	if((rp->size - sizeof(NetworkHeader)) > bufsize) {
	    errno = E2BIG;
	    ret = -1;
	}
	else {
	    DATACOPY((char *)&rp->buf.hdr, hdr, (int) sizeof (NetworkHeader));
	    DATACOPY((char *)rp->buf.data, buf, (int) (rp->size-sizeof(NetworkHeader)));

	    /* Return size of packet == data size - sizeof header. */
	    *plen = rp->size - sizeof (NetworkHeader);
	}
    } else {
	/* If packet is bigger than the buffer bomb out. */
	if(rp->size > bufsize) {
	    errno = E2BIG;
	    ret = -1;
	} else {
	    /* Return data to caller's buffer. */
	    DATACOPY ((char *)&rp->buf, (char *)buf, (int)rp->size);

	    /* Return size of packet == data size. */
	    *plen = rp->size;
	}
    }

    /* Release ring element. */
    rp->full = FALSE;

    /* Point to next ring element. */
    tmp = GETLONG(np->currentinbuf);
    if(++tmp >= np->inringnum)
	tmp = 0;
    SETLONG(np->currentinbuf, tmp);

    if(flags & (NETLOCK | NETBLOCK))
	REMOVE_LOCK(np->net_rcvlock);

    return(ret);
}


/*ARGSUSED*/
int recv_147(h, hdr, buf, bufsize, plen, flags, np)
int h;
NetworkHeader *hdr;		/* Return header here if not NULL */
char *buf;			/* Put packet here (w/o header if hdr == NULL)*/
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
register struct netlib_pb *np;
{
#ifdef _GT_
    int ret = 0;                /* assume success */
    int err;

    if( flags & NETBLOCK ) {
        net_access(h, ENPBLOCKING);
    }
    else {
        net_access(h, ENPPOLLED);
    }


    if (!hdr) {
        /* Read the packet directly into the users buffer */
        if((err = (*np->read) (np->ccb, buf, bufsize, plen)) != RET_OK) {
	    if(err == IFXEBUFEMPTY) {
		errno = EWOULDBLOCK;
		ret = -1;
	    } else {
		errno = E2BIG;
		ret = -1;
	    }
	}
    } else {
        int bytes_read;
        /* Seperate the header and data portion of the packet */

        err = (*np->read) (np->ccb, np->inbuf, sizeof(np->inbuf), &bytes_read);
        if (!err) {
            if (bytes_read <= sizeof(NetworkHeader) + bufsize) {
                *plen  = bytes_read - sizeof(NetworkHeader);
                bcopy((char *)np->inbuf, (char *)hdr, sizeof (NetworkHeader));
                bcopy((char *)(np->inbuf+sizeof(NetworkHeader)), buf, *plen);
            } else {
                errno = E2BIG;
                ret = -1;
            }
        } else if (err == IFXEBUFEMPTY) {
            errno = EWOULDBLOCK;
            ret = -1;
        } else {
            errno = E2BIG;
            ret = -1;
        }
    }

    return ret;

#else

    printf("recv_147: not supported\n");
    return(-1);

#endif
}

/*ARGSUSED*/
int recv_simle(h, hdr, buf, bufsize, plen, flags, np)
int h;
NetworkHeader *hdr;		/* Return header here if not NULL */
char *buf;			/* Put packet here (w/o header if hdr == NULL)*/
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
register struct netlib_pb *np;
{
    int ret = 0;                /* assume success */
    int err;
    char *bufptr;


    if (hdr == NULL)
    {
	ret = net_simle_get_packet(h, buf + sizeof(NetworkHeader), &bufptr, 
		bufsize, plen, (int *)&(np->timestamp), (NetworkAddress *)buf, 
		(NetworkAddress *)(buf + sizeof(NetworkAddress)), 
		(short *)buf + (2 *sizeof(NetworkAddress)), flags);
    }
    else
    {
	ret = net_simle_get_packet(h, buf, &bufptr, bufsize, plen, 
	    (int *)&(np->timestamp), &(hdr->to), &(hdr->from), &(hdr->type), flags);
    }

    return ret;
}


/* 
 * net_get_next_packet
 *
 * See net_get_recv below.
 */

int net_get_next_packet(h, hdr, buf, plen, flags)
int h;				/* handle */
NetworkHeader **hdr;		/* returned pointer to network header */
char **buf;			/* returned pointer to data */
int *plen;			/* bytes in *buf */
int flags;			/* should be zero for now */
{
    return(net_get_recv(h, hdr, buf, plen, flags));
}

/* 
 * net_release_next_packet
 *
 * See net_release_recv.
 */

int net_release_next_packet(h, flags)
int h;				/* handle */
int flags;			/* should be zero for now */
{
    return(net_release_recv(h, flags));
}

/* 
 * net_get_recv
 *
 * Get pointer to header and data part of next packet in the queue
 * and also return the number of bytes in the buffer.  Will not
 * try to use NULL pointers passed as arguments.  Returns 0 if
 * there's a packet waiting and -1 if not.
 */

/*ARGSUSED*/
int net_get_recv(h, hdr, buf, plen, flags)
int h;				/* handle */
NetworkHeader **hdr;		/* returned pointer to network header */
char **buf;			/* returned pointer to data */
int *plen;			/* bytes in *buf */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register RingElement *rp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];

	    if(rp->full == FALSE) {
		errno = EWOULDBLOCK;
		ret = -1;
	    } else {
		if(plen) *plen = rp->size;
		if(buf) *buf = rp->buf.data;
		if(hdr) *hdr = &rp->buf.hdr;
	    }
	    break;
	case NIF_TYPE_147:
	    printf("net_get_next_packet: NIF_TYPE_147 not supported\n");
	    ret = -1;
	    break;
	case NIF_TYPE_SIMLE:
	    printf("net_get_recv: NIF_TYPE_SIMLE not supported\n");
	    ret = -1;
	    break;
	default:
	    printf("net_get_next_packet: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

/* 
 * net_release_recv
 *
 * Release next queue element and just return.
 */

/*ARGSUSED*/
int net_release_recv(h, flags)
int h;				/* handle */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register RingElement *rp;
    int tmp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = &np->inbuffers[GETLONG(np->currentinbuf)];
	    rp->full = FALSE;

	    /* Point to next ring element. */
	    tmp = GETLONG(np->currentinbuf);
	    if(++tmp >= np->inringnum)
	        tmp = 0;
    	    SETLONG(np->currentinbuf, tmp);

	    break;
	case NIF_TYPE_147:
	    printf("net_release_recv: NIF_TYPE_147 not supported\n");
	    ret = -1;
	    break;
	case NIF_TYPE_SIMLE:
	    ret = 0;
	    break;
	default:
	    printf("net_release_recv: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
