#ifndef lint
static char rcsid [] = "$RCSfile: net_osend.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_send.c
 *
 * Copyright (c) 1989 by BBN STC, Inc.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform data access functions.
 *
 * Functions:
 *		net_send
 *		send_cmc
 *		send_147
 *		send_simle
 *		net_get_send
 *		net_release_send
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


int net_send(h, hdr, buf, len, flags)
int h;
NetworkHeader *hdr;		/* Get header from here if not NULL */
char *buf;			/* pkt from here (w/ header if hdr is NULL) */
int len;			/* bytes in buf */
int flags;
{
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    ret = send_cmc(h, hdr, buf, len, flags, np);
	    break;
	case NIF_TYPE_147:
	    ret = send_147(h, hdr, buf, len, flags, np);
	    break;
	case NIF_TYPE_SIMLE:
	    ret = send_simle(h, hdr, buf, len, flags, np);
	    break;
	default:
	    printf("net_send: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

int send_cmc(h, hdr, buf, len, flags, np)
int h;
NetworkHeader *hdr;		/* Get header from here if not NULL */
char *buf;			/* pkt from here (w/ header if hdr is NULL) */
int len;			/* bytes in buf */
int flags;
register struct netlib_pb *np;
{
    VOLATILE RingElement *rp;
    int tmp;
#ifndef SIMBFLY
    struct enprwsupport_arg blocking_args;
#endif

    switch(flags & (NETWAIT | NETLOCK | NETBLOCK)) {
	case 0:
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];
	    if(rp->full == TRUE) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case NETWAIT:
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];
	    wait_for_empty_ring_element(rp);
	    break;
#ifndef mips /* NETLOCK not supported on mips */
	case NETLOCK:
	    if(!POLL_LOCK(np->net_xmtlock)) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];
	    if(rp->full == TRUE) {
		REMOVE_LOCK(np->net_xmtlock);
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case (NETWAIT | NETLOCK):
	    WAIT_FOR_LOCK(np->net_xmtlock);
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];
	    wait_for_empty_ring_element(rp);
	    break;
#endif /*mips*/
	case NETBLOCK:
	    WAIT_FOR_LOCK(np->net_xmtlock);
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];
#ifndef SIMBFLY
	    blocking_args.pfull = (short *)&rp->full;
	    blocking_args.tout_val = np->timeout;
	    np->params = (char *)&blocking_args;
	    while(rp->full == TRUE) {
		if(net_access(h, ENPWRITESUPPORT) != 0) {
		    REMOVE_LOCK(np->net_xmtlock);
		    errno = EWOULDBLOCK;
		    return(-1);
		}
	    }
#endif
	    break;
	default:
	    errno = EINVAL;
	    return(-1);
    }
	    
    /* If the NETNOLEN flag is not specified (this flag allows the
       caller to send packets smaller than the legal enet limit) and
       the len is less than the enet limit, then use the enet limit.
       Else, use caller's length. */

    if (hdr != NULL) { 
	if(len > NETWORK_BUFFER_SIZE)
	    len = NETWORK_BUFFER_SIZE;

	if(!(flags & NETNOLEN) && len < ENETMINSIZE)
	    rp->size = ENETMINSIZE + sizeof(NetworkHeader);
	else
	    rp->size = len + sizeof(NetworkHeader);

	DATACOPY ((char *)hdr, (char *)&rp->buf.hdr, sizeof(NetworkHeader));
	DATACOPY ((char *)buf, (char *)rp->buf.data, len);
    } else {
	if(len > sizeof(NetworkBuffer))
	    len = sizeof(NetworkBuffer);

	if(!(flags & NETNOLEN) && len < ENETMINSIZE + sizeof(NetworkHeader))
	    rp->size = ENETMINSIZE + sizeof(NetworkHeader);
	else
	    rp->size = len;

	DATACOPY ((char *)buf, (char *)&rp->buf, len);
    }

    /* Pass timestamp value. */
    SETLONG(&rp->timestamp, L_REFLECT(np->timestamp));

    /* Release buffer to the card. */
    rp->full = TRUE;

    /* Increment my current buffer pointer */
    tmp = GETLONG(np->currentoutbuf);
    if(++tmp >= np->outringnum)
	tmp = 0;
    SETLONG(np->currentoutbuf, tmp);

    if(flags & (NETLOCK | NETBLOCK))
	REMOVE_LOCK(np->net_xmtlock);

    if ((np->enptype == ENPSIMULATION) &&
	(errno = net_access(h, ENPSENDNOW)))
      return(-1);

    return(0);
}

/*ARGSUSED*/
int send_147(h, hdr, buf, len, flags, np)
int h;
NetworkHeader *hdr;		/* Get header from here if not NULL */
char *buf;			/* pkt from here (w/ header if hdr is NULL) */
int len;			/* bytes in buf */
int flags;
register struct netlib_pb *np;
{
#ifdef _GT_
    int cc;

    if (!hdr) {
        if (len > sizeof(np->outbuf))
            len = sizeof(np->outbuf);
        (*np->write) (np->ccb, buf, len, &cc);
    } else {
        int plen = len + sizeof(*hdr);

        bcopy(hdr, np->outbuf, sizeof(*hdr));
        if (plen > sizeof(np->outbuf)) {
            plen = sizeof(np->outbuf);
            len = plen - sizeof(*hdr);
        }

        bcopy(buf, np->outbuf + sizeof(*hdr), len);

        (*np->write) (np->ccb, np->outbuf, plen, &cc);
    }

    return 0;

#else

    printf("send_147: not supported\n");
    return(-1);

#endif
}

/*ARGSUSED*/
int send_simle(h, hdr, buf, len, flags, np)
int h;
NetworkHeader *hdr;		/* Get header from here if not NULL */
char *buf;			/* pkt from here (w/ header if hdr is NULL) */
int len;			/* bytes in buf */
int flags;
register struct netlib_pb *np;
{
    int rc;

    if (!hdr) {
        if (len > sizeof(np->outbuf))
            len = sizeof(np->outbuf);
        rc = write (np->fd, buf, len);
    } else {
        int plen = len + sizeof(*hdr);

        bcopy(hdr, np->outbuf, sizeof(*hdr));
        if (plen > sizeof(np->outbuf)) {
            plen = sizeof(np->outbuf);
            len = plen - sizeof(*hdr);
        }

        bcopy(buf, np->outbuf + sizeof(*hdr), len);

        rc = write (np->fd, np->outbuf, plen);
    }

    if (rc >= 0)
	return 0;
    else
	return -1;
}

/* 
 * net_get_send
 *
 * Get pointer to header and data part of next packet in the queue
 * and also return the buffer size.  This call will not try to use
 * a NULL pointer passed to it as an argument.  Returns 0 if
 * there's a buffer available waiting and -1 if not.
 */

/*ARGSUSED*/
int net_get_send(h, hdr, buf, plen, flags)
int h;				/* handle */
NetworkHeader **hdr;		/* returned pointer to network header */
char **buf;			/* returned pointer to data */
int *plen;			/* bytes in *buf */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register RingElement *rp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];

	    if(rp->full == FALSE) {
		errno = EWOULDBLOCK;
		ret = -1;
	    } else {
		if(plen) *plen = NETWORK_BUFFER_SIZE;
		if(buf) *buf = rp->buf.data;
		if(hdr) *hdr = &rp->buf.hdr;
	    }
	    break;
	case NIF_TYPE_147:
	    printf("net_get_send: NIF_TYPE_147 not supported\n");
	    ret = -1;
	    break;
	case NIF_TYPE_SIMLE:
	    printf("net_get_send: NIF_TYPE_SIMLE not supported\n");
	    ret = -1;
	    break;
	default:
	    printf("net_get_send: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

/* 
 * net_release_send
 *
 * Release next queue element and just return.
 */

/*ARGSUSED*/
int net_release_send(h, flags)
int h;				/* handle */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register RingElement *rp;
    int tmp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = &np->outbuffers[GETLONG(np->currentoutbuf)];
	    rp->full = FALSE;

	    /* Point to next ring element. */
	    tmp = GETLONG(np->currentoutbuf);
	    if(++tmp >= np->outringnum)
	        tmp = 0;
    	    SETLONG(np->currentoutbuf, tmp);

	    break;
	case NIF_TYPE_147:
	    printf("net_release_send: NIF_TYPE_147 not supported\n");
	    ret = -1;
	    break;
	case NIF_TYPE_SIMLE:
	    printf("net_release_send: NIF_TYPE_SIMLE not supported\n");
	    ret = -1;
	    break;
	default:
	    printf("net_release_send: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
