#ifndef lint
static char rcsid [] = "$RCSfile: net_send.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_send.c
 *
 * Copyright (c) 1989 by BBN STC, Inc.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform data access functions.
 *
 * Functions:
 *		wait_for_empty_ring_element
 *		net_snd
 *		snd_cmc
 *		snd_147
 *		net_get_snd
 *		net_release_snd
 *		net_set_snd_from_addr
 *		net_set_snd_type
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


/* This function is necessary for the MIPS.  It is difficult to get a RISC
 * machine with a good optimizing compiler to actually wait around doing
 * nothing. 
 */

wait_for_empty_ring_element(rp)
VOLATILE RingElement *rp;
{
#ifndef lint
    int dummy = 0;
    register int i;
#endif

    /* Wait until the buffer is no longer full */
    while(rp->full == TRUE) {
	/* Do something while we wait, or else the optimizing compiler will
	 * throw away this dumb code.
	 */
#ifndef lint
	for(i = 0; i < 5; i++)
	  dummy++;
#endif
    }
}

/*
 * net_snd
 *
 * Send a packet.  "len" bytes are copied from "buf".  The packet 
 * is composed from "to", the source address set by net_set_snd_from_addr(),
 * and the type set by net_set_snd_type(). 
 *
 * "flags" are
 *	NETWAIT		spin while waiting for room in queue
 *	NETLOCK		lock queue for exclusive access
 *	NETBLOCK	sleep while waiting for room in queue
 *
 * Returns 0 if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 *	EWOULDBLOCK	output queue full, no room
 *	EINVAL		unknown flag
 *	ENETUNREACH     write() failed for simle device
 */
int net_snd(h, to, buf, len, flags)
int h;
NetworkAddress *to;
char *buf;
int len;
int flags;
{
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    ret = send_cmc_8023(h, to, buf, len, flags, np);
	    break;
	case NIF_TYPE_147:
	    ret = send_147_8023(h, to, buf, len, flags, np);
	    break;
	case NIF_TYPE_SIMLE:
	    ret = send_simle_8023(h, to, buf, len, flags, np);
	    break;
	default:
	    printf("net_snd: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

/*
 * net_set_snd_from_addr
 *
 * Set the source address for subsequent net_snd() calls.  If this 
 * is any multicast address (least significant bit of the first byte == 1)
 * then the default address of the interface is used.
 *
 * Returns 0 if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 */
int net_set_snd_from_addr(h, from)
int h;
NetworkAddress *from;
{
    CHECK_HANDLE(h, -1);

    npb[h].sndFrom = *from;

    return(0);
}

/*
 * net_set_snd_type
 *
 * Set the packet type for subsequent net_snd() calls.
 *
 * Returns 0 if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 */
int net_set_snd_type(h, type)
int h;
int type;
{
    CHECK_HANDLE(h, -1);

    npb[h].sndType = type;

    return(0);
}

int send_cmc_8023(h, to, buf, len, flags, np)
int h;
NetworkAddress *to;
char *buf;
int len;
int flags;
register struct netlib_pb *np;
{
    VOLATILE QueueElement *rp;
    int tmp;
#ifndef SIMBFLY
    struct enprwsupport_arg blocking_args;
#endif

    switch(flags & (NETWAIT | NETLOCK | NETBLOCK)) {
	case 0:
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];
	    if(rp->full == TRUE) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case NETWAIT:
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];
	    wait_for_empty_ring_element(rp);
	    break;
#ifndef mips /* NETLOCK not supported on mips */
	case NETLOCK:
	    if(!POLL_LOCK(np->net_xmtlock)) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];
	    if(rp->full == TRUE) {
		REMOVE_LOCK(np->net_xmtlock);
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case (NETWAIT | NETLOCK):
	    WAIT_FOR_LOCK(np->net_xmtlock);
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];
	    wait_for_empty_ring_element(rp);
	    break;
#endif /*mips*/
	case NETBLOCK:
	    WAIT_FOR_LOCK(np->net_xmtlock);
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];
#ifndef SIMBFLY
	    blocking_args.pfull = (short *)&rp->full;
	    blocking_args.tout_val = np->timeout;
	    np->params = (char *)&blocking_args;
	    while(rp->full == TRUE) {
		if(net_access(h, ENPWRITESUPPORT) != 0) {
		    REMOVE_LOCK(np->net_xmtlock);
		    errno = EWOULDBLOCK;
		    return(-1);
		}
	    }
#endif
	    break;
	default:
	    errno = EINVAL;
	    return(-1);
    }
	    
    /* If the NETNOLEN flag is not specified (this flag allows the caller
       to send packets smaller than the legal enet limit) and the len is
       less than the enet limit, then use the enet limit.  Else, use 
       caller's length. */
    if(len > MAX_DATA_SIZE_8023)
	len = MAX_DATA_SIZE_8023;

    /* This is done to keep the ENP10 happy - it doesn't like a long word
       move (which a structure assignment generates). */
    movewords(to, 
	      &rp->buf.h.hdr8023.to, 
	      sizeof(*to)/sizeof(short));
    movewords(&np->sndFrom, 
    	      &rp->buf.h.hdr8023.from, 
	      sizeof(np->sndFrom)/sizeof(short));

    if(!(flags & NETNOLEN) && len < MIN_DATA_SIZE_8023)
        rp->buf.h.hdr8023.length = ENETMINSIZE;
    else
	rp->buf.h.hdr8023.length = len + SAP_SNAP_SIZE;

    SET_DSAP(rp->buf.data, SAP_SNAP);
    SET_SSAP(rp->buf.data, SAP_SNAP);
    SET_CONTROL(rp->buf.data, UNNUMBERED_DATA);
    SET_PROTOID(rp->buf.data, ORG_CODE_BBN);
    SET_ETHER_TYPE(rp->buf.data, np->sndType);

    DATACOPY ((char *)buf, GET_DATA_PTR(rp->buf.data), len);

    /* Pass timestamp value. */
    SETLONG(&rp->timestamp, L_REFLECT(np->timestamp));

    /* Set byte count for card to copy: */
    rp->size = rp->buf.h.hdr8023.length + sizeof(NetworkHeader8023);

    /* Release buffer to the card. */
    rp->full = TRUE;

    /* Increment my current buffer pointer */
    tmp = GETLONG(np->currentoutbuf);
    if(++tmp >= np->outringnum)
	tmp = 0;
    SETLONG(np->currentoutbuf, tmp);

    if(flags & (NETLOCK | NETBLOCK))
	REMOVE_LOCK(np->net_xmtlock);

    if ((np->enptype == ENPSIMULATION) &&
	(errno = net_access(h, ENPSENDNOW)))
      return(-1);

    return(0);
}

/*ARGSUSED*/
int send_147_8023(h, to, buf, len, flags, np)
int h;
NetworkAddress *to;
char *buf;
int len;
int flags;
register struct netlib_pb *np;
{
#ifdef _GT_
    int cc;
    register NetworkPacket *pkt;

    pkt = (NetworkPacket *)np->outbuf;

    if(len < MIN_DATA_SIZE_8023)
	len = MIN_DATA_SIZE_8023;
    else if(len > MAX_DATA_SIZE_8023) 
	len = MAX_DATA_SIZE_8023;

    pkt->h.hdr8023.to = *to;
    pkt->h.hdr8023.from = np->sndFrom;
    pkt->h.hdr8023.length = len + SAP_SNAP_SIZE;

    SET_DSAP(pkt->data, SAP_SNAP);
    SET_SSAP(pkt->data, SAP_SNAP);
    SET_CONTROL(pkt->data, UNNUMBERED_DATA);
    SET_PROTOID(pkt->data, ORG_CODE_BBN);
    SET_ETHER_TYPE(pkt->data, np->sndType);

    bcopy(buf, GET_DATA_PTR(pkt->data), len);

    (*np->write) (np->ccb, 
		  np->outbuf, 
		  sizeof(NetworkHeader8023) + SAP_SNAP_SIZE + len, 
		  &cc);

    return 0;

#else

    printf("snd_147: not supported\n");
    return(-1);

#endif
}

/*ARGSUSED*/
int send_simle_8023(h, to, buf, len, flags, np)
int h;
NetworkAddress *to;
char *buf;
int len;
int flags;
register struct netlib_pb *np;
{
    register NetworkPacket *pkt;

    pkt = (NetworkPacket *)np->outbuf;

    if(len < MIN_DATA_SIZE_8023)
	len = MIN_DATA_SIZE_8023;
    else if(len > MAX_DATA_SIZE_8023) 
	len = MAX_DATA_SIZE_8023;

    pkt->h.hdr8023.to = *to;
    pkt->h.hdr8023.from = np->sndFrom;
    pkt->h.hdr8023.length = len + SAP_SNAP_SIZE;

    SET_DSAP(pkt->data, SAP_SNAP);
    SET_SSAP(pkt->data, SAP_SNAP);
    SET_CONTROL(pkt->data, UNNUMBERED_DATA);
    SET_PROTOID(pkt->data, ORG_CODE_BBN);
    SET_ETHER_TYPE(pkt->data, np->sndType);

    bcopy(buf, GET_DATA_PTR(pkt->data), len);

    if (write(np->fd, np->outbuf, sizeof(NetworkHeader8023) + 
	    SAP_SNAP_SIZE + len) >= 0)
	return 0;
    else
	return -1;
}
/* 
 * net_get_snd
 *
 * Get pointer to data part of next packet in the queue
 * and also return the buffer size.  Returns 0 if
 * there's a buffer available waiting and -1 if not.
 */

/*ARGSUSED*/
int net_get_snd(h, buf, plen, flags)
int h;				/* handle */
char **buf;			/* returned pointer to data */
int *plen;			/* bytes in *buf */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register QueueElement *rp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];

	    if(rp->full == TRUE) {
		errno = EWOULDBLOCK;
		ret = -1;
	    } else {
		if(np->enptype == ENP10)
		    *buf = np->outbuf;
		else
		    *buf = (char *)GET_DATA_PTR(rp->buf.data);
		*plen = MAX_DATA_SIZE_8023;
	    }
	    break;
	case NIF_TYPE_147:
	{
#ifdef _GT_

	    *buf = (char *)GET_DATA_PTR(((NetworkPacket *)np->outbuf)->data);
	    *plen = MAX_DATA_SIZE_8023;

#else
	    printf("net_get_snd: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	}
	case NIF_TYPE_SIMLE:
	    *buf = (char *)GET_DATA_PTR(((NetworkPacket *)np->outbuf)->data);
	    *plen = MAX_DATA_SIZE_8023;
	    break;

	default:
	    printf("net_get_snd: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

/* 
 * net_release_snd
 *
 * Release next queue element and just return.
 */

/*ARGSUSED*/
int net_release_snd(h, to, len, flags)
int h;				/* handle */
NetworkAddress *to;		/* destination */
int len;			/* bytes to send */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register QueueElement *rp;
    int tmp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = (QueueElement *)&np->outbuffers[GETLONG(np->currentoutbuf)];

	    /* Set the size */
	    if(len < MIN_DATA_SIZE_8023)
		len = MIN_DATA_SIZE_8023;
	    else if(len > MAX_DATA_SIZE_8023)
		len = MAX_DATA_SIZE_8023;
	    rp->size = len + sizeof(NetworkHeader8023) + SAP_SNAP_SIZE;

	    /* This is done to keep the ENP10 happy - it doesn't like a 
	       long word move (which a structure assignment generates). */
	    movewords(to, 
		      &rp->buf.h.hdr8023.to, 
		      sizeof(*to)/sizeof(short));
	    movewords(&np->sndFrom, 
		      &rp->buf.h.hdr8023.from, 
		      sizeof(np->sndFrom)/sizeof(short));
	    rp->buf.h.hdr8023.length = len + SAP_SNAP_SIZE;

	    SET_DSAP(rp->buf.data, SAP_SNAP);
	    SET_SSAP(rp->buf.data, SAP_SNAP);
	    SET_CONTROL(rp->buf.data, UNNUMBERED_DATA);
	    SET_PROTOID(rp->buf.data, ORG_CODE_BBN);
	    SET_ETHER_TYPE(rp->buf.data, np->sndType);

	    if(np->enptype == ENP10)
    		DATACOPY (np->outbuf, GET_DATA_PTR(rp->buf.data), len);

	    /* Pass timestamp value. */
	    SETLONG(&rp->timestamp, L_REFLECT(np->timestamp));

	    rp->full = TRUE;

	    /* Point to next ring element. */
	    tmp = GETLONG(np->currentoutbuf);
	    if(++tmp >= np->outringnum)
	        tmp = 0;
    	    SETLONG(np->currentoutbuf, tmp);

	    break;
	case NIF_TYPE_147:
	{
#ifdef _GT_
	    int cc;
    	    register NetworkPacket *pkt;

	    pkt = (NetworkPacket *)np->outbuf;

	    if(len < MIN_DATA_SIZE_8023)
		len = MIN_DATA_SIZE_8023;
	    else if(len > MAX_DATA_SIZE_8023)
		len = MAX_DATA_SIZE_8023;

	    pkt->h.hdr8023.to = *to;
	    pkt->h.hdr8023.from = np->sndFrom;
	    pkt->h.hdr8023.length = len + SAP_SNAP_SIZE;

	    SET_DSAP(pkt->data, SAP_SNAP);
	    SET_SSAP(pkt->data, SAP_SNAP);
	    SET_CONTROL(pkt->data, UNNUMBERED_DATA);
	    SET_PROTOID(pkt->data, ORG_CODE_BBN);
	    SET_ETHER_TYPE(pkt->data, np->sndType);

	    (*np->write) (np->ccb, 
			  np->outbuf, 
			  sizeof(NetworkHeader8023) + SAP_SNAP_SIZE + len, 
			  &cc);
#else
	    printf("net_release_snd: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	}
	case NIF_TYPE_SIMLE:
	{
    	    register NetworkPacket *pkt;

	    pkt = (NetworkPacket *)np->outbuf;

	    if(len < MIN_DATA_SIZE_8023)
		len = MIN_DATA_SIZE_8023;
	    else if(len > MAX_DATA_SIZE_8023)
		len = MAX_DATA_SIZE_8023;

	    pkt->h.hdr8023.to = *to;
	    pkt->h.hdr8023.from = np->sndFrom;
	    pkt->h.hdr8023.length = len + SAP_SNAP_SIZE;

	    SET_DSAP(pkt->data, SAP_SNAP);
	    SET_SSAP(pkt->data, SAP_SNAP);
	    SET_CONTROL(pkt->data, UNNUMBERED_DATA);
	    SET_PROTOID(pkt->data, ORG_CODE_BBN);
	    SET_ETHER_TYPE(pkt->data, np->sndType);

	    write(np->fd, np->outbuf, sizeof(NetworkHeader8023) + SAP_SNAP_SIZE
		+ len);
	    break;
	}
	default:
	    printf("net_release_snd: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
