#ifndef lint
static char rcsid [] = "$RCSfile: net_sock.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*
 * File:	net_sock.c
 */

#ifndef NO_ASSOC
#ifdef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include <stdio.h>
#include <sys/types.h>
#include <net/raw.h>
#include <net/if.h>
#include <errno.h>
#include <stdlib.h>
#include "libnetif.h"

#include <stdstring.h> /*common/include/global*/

#define ETHERHDRPAD RAW_HDRPAD(sizeof(NetworkHeader))
#define IEEE802_3PORT 1

/* 080008 is BBN specific enet addressing */
NetworkAddress fake_from =     {0x08, 0x0, 0x08, 0xff, 0xff, 0xff};
NetworkAddress FillinAddress = {0x08, 0x0, 0x08, 0xff, 0xff, 0xff};

NetworkAddress BroadcastAddress = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

struct padded_packet {
    char          pad[ETHERHDRPAD];
    NetworkPacket pkt;
};

static struct {
    char            interface_name[IFNAMSIZ];
    NetworkAddress  sndFrom;	/* source address when sending */
    int             sndType;	/* packet type when sending */
    NetworkAddress  rcvTo;	/* dest address of last packet rcvd */
    NetworkAddress  rcvFrom;	/* source address of last packet rcvd */
    int             rcvType;	/* packet type of last packet rcvd */
    int             n_types;	/* number of types accepted */
    int             types[256];	/* types accepted */
    int             timestamp;	/* time of last received packet */

    /* Statistics
     */
    int if_dropped;
    int sb_dropped;
    int rcv_count;
    int snd_count;

    /* These next three are really nonsense with this kind of a driver.
     * But they are supported for compatibility.
     */
    char            snd_buffer[NETWORK_BUFFER_SIZE];
    char            rcv_buffer[NETWORK_BUFFER_SIZE];
    int             rcv_released;
} sockinfo[256];    

int net_open(device, flags, ignored)
    char *device;
    int   flags;
    int   ignored;
{
    int s;
    struct sockaddr_raw sr;
    struct drainmap map;
    int cc;
    int on = 1;

    struct ifreq ifr;
    struct ifconf ifc;
    char cbuf[BUFSIZ];
    int s2;

    /* Create the socket */
    if ((s = socket(PF_RAW, SOCK_RAW, RAWPROTO_DRAIN)) == -1)
      return -1;

    sr.sr_family = AF_RAW;
    sr.sr_port = IEEE802_3PORT;
    bzero(sr.sr_ifname, sizeof(sr.sr_ifname));
    strncpy(sr.sr_ifname, device, sizeof(sr.sr_ifname));
    if (bind(s, &sr, sizeof(sr)) == -1)
      return -1;

/* $$$ */

    if (0 > (s2 =socket(PF_RAW, SOCK_RAW,RAWPROTO_DRAIN))) {
        perror("socket");
	exit(-1);
    }
    
    ifc.ifc_len = sizeof cbuf;
    ifc.ifc_buf = cbuf;
    if (0 > ioctl(s2, SIOCGIFCONF, (char *)&ifc)) {
	perror("SIOCGIFCONF");
	exit(1);
    }
    (void) strcpy(ifr.ifr_name, ifc.ifc_req->ifr_name);
    if (0 > ioctl(s2, SIOCGIFADDR, &ifr)) {
	perror("SIOCGIFADDR");
	exit(1);
    }
    bcopy(ifr.ifr_addr.sa_data, &fake_from, sizeof(fake_from));
    close(s2);

/* $$$ */

    /* Tell the kernel to give all 802.3 pacekts to this socket */
    map.dm_minport = 46;
    map.dm_maxport = 1500;
    map.dm_toport = IEEE802_3PORT;
    if (ioctl(s, SIOCDRAINMAP, &map) == -1)
      return -1;

    /* Crank up the input buffering */
    cc = (1024 * 512);   /* SGI 5.2 OS */
    if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&cc, sizeof(cc)) == -1)
    {
	/* failed.  try smaller buf size */
	cc = (1024 * 64) - 8;   /* SGI 4.0.5 OS */
	if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&cc, sizeof(cc)) == -1)
	{
	    printf("unable to set socket buffering\n");
	    return -1;
	}
	else
	  printf("socket buffering 64K\n");
    }
    else
      printf("socket buffering 512K\n");

    /* Make the socket non-blocking */
    if (ioctl(s, FIONBIO, &on) == -1)
      return -1;

    /* Initialize local info */
    bzero(&sockinfo[s], sizeof(sockinfo[s]));
    strncpy(sockinfo[s].interface_name, device, IFNAMSIZ);
    sockinfo[s].sndFrom = fake_from;
    sockinfo[s].sndType = PROTOCOL_SIM;
    sockinfo[s].n_types = 1;
    sockinfo[s].types[0] = PROTOCOL_SIM;
    sockinfo[s].rcv_released = 1;

    return s;
}

int net_close(s)
    int s;
{
    return close(s);
}

int net_send(s, to, buf, len, flags)
    int             s;
    NetworkAddress *to;
    char           *buf;
    int             len;
    int             flags;
{
    static int warned = 0;

    if (!warned)
    {
	printf("Warning: SGI net_send driver does not support ethernet 2.\n         Calling net_snd instead\n");
	warned = 1;
    }
    return net_snd(s, to, buf, len, flags);
}

int net_snd(s, to, buf, len, flags)
    int             s;
    NetworkAddress *to;
    char           *buf;
    int             len;
    int             flags;
{
    struct padded_packet pkt;
    unsigned int wsize;

    /* Basically copied from libnetif/net_send.c */

    if(len > MAX_DATA_SIZE_8023)
      len = MAX_DATA_SIZE_8023;

    pkt.pkt.h.hdr8023.to = *to;
    pkt.pkt.h.hdr8023.from = sockinfo[s].sndFrom;

    if(!(flags & NETNOLEN) && len < MIN_DATA_SIZE_8023)
      pkt.pkt.h.hdr8023.length = ENETMINSIZE;
    else
      pkt.pkt.h.hdr8023.length = len + SAP_SNAP_SIZE;

    SET_DSAP(pkt.pkt.data, SAP_SNAP);
    SET_SSAP(pkt.pkt.data, SAP_SNAP);
    SET_CONTROL(pkt.pkt.data, UNNUMBERED_DATA);
    SET_PROTOID(pkt.pkt.data, ORG_CODE_BBN);
    SET_ETHER_TYPE(pkt.pkt.data, sockinfo[s].sndType);

    bcopy((char *)buf, GET_DATA_PTR(pkt.pkt.data), len);

    wsize = pkt.pkt.h.hdr8023.length + sizeof(NetworkHeader8023);
    if (write(s, (char *)&pkt.pkt, wsize) != wsize)
      return -1;

    sockinfo[s].snd_count += 1;

    return 0;
}

static int check_type(type, n, arr)
    int type;
    int n;
    int arr[];
{
    int i;

    for(i=0;i<n;i++)
      if (arr[i] == type)
	return 1;
    return 0;
}

int net_recv(s, buf, bufsize, plen, flags)
    int   s;
    char *buf;		/* Put packet here (w/o header) */
    int   bufsize;	/* Size of buf */
    int  *plen;		/* Return bytes in buf here */
    int   flags;
{
    static int warned = 0;

    if (!warned)
    {
	printf("Warning: SGI net_recv driver does not support ethernet 2.\n         Calling net_rcv instead\n");
	warned = 1;
    }
    return net_rcv(s, buf, bufsize, plen, flags);
}

int net_rcv(s, buf, bufsize, plen, flags)
    int   s;
    char *buf;		/* Put packet here (w/o header) */
    int   bufsize;	/* Size of buf */
    int  *plen;		/* Return bytes in buf here */
    int   flags;
{
    int ret = 0;                /* assume success */
    int cc;
    struct padded_packet pkt;
    int size;
    char *datap;
    int off = 0, on = 1;;

    /* Real time users will always use non-blocking reads.  Hence, we
     * leave the socket non-blocking, and make it blocking only when
     * needed.
     */
    if( flags & NETBLOCK )
      ioctl(s, FIONBIO, &off);

    for(;;)
    {
	cc = read(s, &pkt, sizeof(pkt));

	if (cc == 0)
	{
	    /* I don't trust read never to do this */
	    errno = EWOULDBLOCK;
	    ret = -1;
	    break;
	}

	if (cc < 0)
	{
	    /* Nothing to read */
	    ret = -1;
	    break;
	}

#if 0
	/* Under IRIX 5.2 we get back packets from ourselves.  Filter them
	 * here.
	 */
	if (pkt.pkt.h.hdr8023.length > ENETMAXSIZE)
	{
	    if (bcmp(pkt.pkt.h.hdr2.from.addr, 
		     sockinfo[s].sndFrom.addr,
		     ENETADDRSZ) == 0)
	    {
		/* from us */
		continue;
	    }
	}
	else
	{
	    if (bcmp(pkt.pkt.h.hdr8023.from.addr, 
		     sockinfo[s].sndFrom.addr,
		     ENETADDRSZ) == 0)
	    {
		/* from us */
		continue;
	    }
	}
#endif /* sgi */

	sockinfo[s].rcv_count += 1;

	/* Note the time.  It is unfortunate that we have to get the
	 * time here, because promiscuous snoop sockets provide
	 * an in-kernel timestamping facility.  Unfortunately, drain
	 * sockets do not.
	 */
	sockinfo[s].timestamp = net_current_time(s);

	if (pkt.pkt.h.hdr8023.length > ENETMAXSIZE)
	{
	    sockinfo[s].rcvTo = pkt.pkt.h.hdr2.to;
	    sockinfo[s].rcvFrom = pkt.pkt.h.hdr2.from;
	    sockinfo[s].rcvType = pkt.pkt.h.hdr2.type;
	    size = cc-ETHERHDRPAD-sizeof(NetworkHeader);
	    datap = pkt.pkt.data;
	}
	else
	{
	    sockinfo[s].rcvTo = pkt.pkt.h.hdr8023.to;
	    sockinfo[s].rcvFrom = pkt.pkt.h.hdr8023.from;
	    sockinfo[s].rcvType = GET_ETHER_TYPE(pkt.pkt.data);
	    size = cc-ETHERHDRPAD-sizeof(NetworkHeader8023)-SAP_SNAP_SIZE;
	    datap = (char *)GET_DATA_PTR(pkt.pkt.data);
	}
       
	/* Do we want this type? */
	if (!check_type(sockinfo[s].rcvType,
			sockinfo[s].n_types, sockinfo[s].types))
	  continue;

        /* If packet is bigger than the buffer bomb out. */
        if (size > bufsize)
	{
	    errno = E2BIG;
	    ret = -1;
	}
	else
	{
	    /* Return data to caller's buffer. */
	    bcopy(datap, buf, size);

	    /* Return size of packet == data size. */
	    *plen = size;
	}
	break;
    }

    /* Re-enable non-blocking I/O */
    if( flags & NETBLOCK )
      ioctl(s, FIONBIO, &on);

    return ret;
}

int net_get_rcv(s, buf, plen, flags)
    int    s;
    char **buf;			/* returned pointer to data */
    int   *plen;		/* bytes in *buf */
    int    flags;
{
    /* Calling this the first time gets you a new packet.
     * Successive calls give you the same packet until net_release_rcv
     * is called.
     */

    *buf = sockinfo[s].rcv_buffer;

    if (sockinfo[s].rcv_released)
    {
	sockinfo[s].rcv_released = 0;
	return net_rcv(s, sockinfo[s].rcv_buffer,
		       NETWORK_BUFFER_SIZE, plen, flags);
    }
    else
      return 0;
}

int net_release_rcv(s)
    int s;
{
    sockinfo[s].rcv_released = 1;
    return 0;
}

int net_get_snd(s, buf, plen, flags)
    int s;
    char **buf;			/* returned pointer to data */
    int *plen;			/* bytes in *buf */
    int flags;			/* should be zero for now */
{
    /* Calling this gives you a pointer back to a place to put a
     * packet.  The packet is sent with net_release_snd.
     */
    *buf = sockinfo[s].snd_buffer;
    *plen = MAX_DATA_SIZE_8023;
    return 0;
}

int net_release_snd(s, to, len, flags)
    int s;
    NetworkAddress *to;		/* destination */
    int len;			/* bytes to send */
    int flags;
{
    /* Sends the packet in the send buffer */
    return net_snd(s, to, sockinfo[s].snd_buffer, len, flags);
}

#include <sys/time.h>

int net_current_time(s)
    int s;
{
    struct timeval t;
    struct timezone tz;

    gettimeofday(&t, &tz);

    return t.tv_sec * 1000 + t.tv_usec / 1000;
}

int net_gettime(s)
    int s;
{
    return net_current_time(s);
}

int net_set_snd_type(s, type)
    int s;
    int type;
{
    sockinfo[s].sndType = type;
    return 0;
}

int net_set_snd_from_addr(s, from)
    int s;
    NetworkAddress *from;
{
    sockinfo[s].sndFrom = *from;
    return 0;
}

int net_get_rcv_to_addr(s, to)
    int s;
    NetworkAddress *to;
{
    *to = sockinfo[s].rcvTo;
    return 0;
}

int net_get_rcv_from_addr(s, from)
    int s;
    NetworkAddress *from;
{
    *from = sockinfo[s].rcvFrom;
    return 0;
}

int net_get_rcv_type(s)
    int s;
{
    return sockinfo[s].rcvType;
}

int net_addr_compare(a1, a2)
    NetworkAddress *a1;
    NetworkAddress *a2;
{
    return (!bcmp((char *)a1, (char *)a2, sizeof(NetworkAddress)));
}

int net_add_mca(s, pna)
    int s;
    NetworkAddress *pna;
{
    struct ifreq ifr;

    bcopy(sockinfo[s].interface_name, ifr.ifr_name, IFNAMSIZ);
    ifr.ifr_addr.sa_family = AF_RAW;
    bcopy(pna, ifr.ifr_addr.sa_data, sizeof(NetworkAddress));
    return ioctl(s, SIOCADDMULTI, &ifr, sizeof(ifr));
}

int net_del_mca(s, pna)
    int s;
    NetworkAddress *pna;
{
    struct ifreq ifr;

    bcopy(sockinfo[s].interface_name, ifr.ifr_name, IFNAMSIZ);
    ifr.ifr_addr.sa_family = AF_RAW;
    bcopy(pna, ifr.ifr_addr.sa_data, sizeof(NetworkAddress));
    return ioctl(s, SIOCDELMULTI, &ifr, sizeof(ifr));
}


int net_init_type(s)
    int s;
{
    sockinfo[s].n_types = 0;
    return 0;
}

int net_add_type(s, type)
    int s;
    int type;
{
    sockinfo[s].types[sockinfo[s].n_types++] = type;
    return 0;
}

int net_addr_str_to_bin(p, s)
    NetworkAddress *p;
    char *s;
{
    int i, x;
    
    for(i = 0; i < ENETADDRSZ; i++)
        p->addr[i] = 0;

    /* check for exactly 12 digits, all hex digits */
    for (i = 0; i < ENETADDRSZ * 2; i++) {
	if(s[i] >= '0' && s[i] <= '9')
	    continue;
	else if(s[i] >= 'a' && s[i] <= 'f')
	    continue;
	else if(s[i] >= 'A' && s[i] <= 'F')
	    continue;
	else
            return 0;
    }

    if (s[i] != '\0')
	return 0;

    for (i = 0; i < ENETADDRSZ; i++, s += 2) {
	sscanf (s, "%2x", &x);
	p->addr[i] = x;
    }

    return 1;
}

int net_flush(s, flags)
    int s;
    int flags;
{
    char buf[NETWORK_BUFFER_SIZE];
    int len;

    if (flags & NETFLUSHRCV)
    {
	while (net_rcv(s, buf, NETWORK_BUFFER_SIZE, &len, 0) == 0)
	  ;
    }

    if (flags & NETFLUSHXMT)
    {
	/* NOP */;
    }

    return 0;
}

#define N_STATS 4		/* set to number of strings */
char statistics_strings[][41] = {
		"successful transmissions                ",
		"packets received                        ",
		"missed packets reported (hardware)      ",
		"missed packets reported (software)      ",
	};
int max_statistics_string = N_STATS;

int net_zero_statistics(s)
    int s;
{
    struct rawstats zstats;

    if (ioctl(s, SIOCRAWSTATS, &zstats, sizeof(struct rawstats)) == -1)
      return -1;

    sockinfo[s].if_dropped = zstats.rs_drain.ds_ifdrops;
    sockinfo[s].sb_dropped = zstats.rs_drain.ds_sbdrops;
    sockinfo[s].rcv_count = 0;
    sockinfo[s].snd_count = 0;

    return 0;
}

int net_get_statistics(s, stats)
    int s;
    int *stats;
{
    struct rawstats sys_stats;

    if (ioctl(s, SIOCRAWSTATS, &sys_stats, sizeof(struct rawstats)) == -1)
      return -1;

    stats[0] = sockinfo[s].snd_count;
    stats[1] = sockinfo[s].rcv_count;
    stats[2] = sys_stats.rs_drain.ds_ifdrops - sockinfo[s].if_dropped;
    stats[3] = sys_stats.rs_drain.ds_sbdrops - sockinfo[s].sb_dropped;

    return 0;
}


void net_stat_string(stat, buf)
    int stat;
    char *buf;
{
    if(stat >= max_statistics_string)
      buf[0] = '\0';
    else
      strcpy(buf, &statistics_strings[stat][0]);
}

int net_print_statistics(s)
    int s;
{
    long stats[N_STATS];
    char str[60];
    register int i;

    if (net_get_statistics (s, stats) == -1)
      return -1;

    for (i = 0; i < N_STATS; i++)
    {
	net_stat_string (i, str);
	if (str[0] != '\0')
	  printf ("%s%ld\n", str, stats[i]);
    }
    return 0;
}

long net_get_timestamp(s)
int s;
{
    return(sockinfo[s].timestamp);
}

int net_put_timestamp(s, time)
int s;
long time;
{

    sockinfo[s].timestamp = time;

    return(0);
}

#define stub(name) int name() { return 0; }

#ifdef sgi
int net_norm(s, na)
    int s;
    NetworkAddress *na;
{
    struct ifreq req;

    strncpy(req.ifr_name, sockinfo[s].interface_name, IFNAMSIZ);
    if (ioctl(s, SIOCGIFADDR, &req) == -1)
    {
	perror("net_norm ioctl:SIOCGIFADDR");
	return -1;
    }

    if (req.ifr_addr.sa_family != AF_RAW)
    {
	fprintf(stderr, "net_norm:  Address family not AF_RAW\n");
	return -1;
    }

    bcopy(req.ifr_addr.sa_data, na->addr, ENETADDRSZ);
    return 0;
}
#else /* sgi */
stub(net_norm);
#endif /* sgi */

stub(net_stop)
stub(net_alive)
stub(net_init_mca)
stub(net_iocontrol)

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */

