#ifndef lint
static char rcsid [] = "$RCSfile: net_time.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_time.c
 *
 * Copyright 1995 Loral Advanced Distributed Simulation Inc.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to access the CMC card timer.
 *
 * Functions:
 *		net_gettime
 *		net_settime
 *		net_current_time
 *		net_init_time
 *		net_stomp_time
 *		net_heartbeat
 *		net_device_base
 */


/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include "libnet.h"


/*****************************************
 *
 *	net_gettime
 *
 *	This routine gets the value of the 32-bit timer on the
 * CMC card.
 *
 ****************************************/
int net_gettime(h)
int h;
{
    int ret;

    CHECK_HANDLE(h, -1);

    switch(npb[h].type) {
	case NIF_TYPE_CMC:
	case NIF_TYPE_SIMLE:
#ifndef SIMBFLY
    	    npb[h].params = (char *)&ret;
#endif
	    if(errno = net_access(h, ENPGETTIME))
		ret = -1;
#ifdef SIMBFLY
            ret = npb[h].net_req->a.gettime_req.time;
#endif
	    break;
	case NIF_TYPE_147:
#ifdef _GT_
	    ret = sc_gtime() * 10 - npb[h].base_time;
#else
	    printf("net_gettime: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	default:
	    printf("net_gettime: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return (ret);
}

/*****************************************
 *
 *	net_settime
 *
 *	This routine sets the 32-bit timer on the CMC card.
 *
 ****************************************/
int net_settime(h, time)
int h;
int time;
{
    int ret = 0;

    CHECK_HANDLE(h,  -1);

    switch(npb[h].type) {
	case NIF_TYPE_CMC:
#ifdef SIMBFLY
    	    npb[h].net_req->a.settime_req.time = time;
#else
    	    npb[h].params = (char *)&time;
#endif
	    if(errno = net_access(h, ENPSETTIME))
		ret = -1;
	    break;
	case NIF_TYPE_147:
#ifdef _GT_
	    npb[h].base_time = time;
#else
	    printf("net_settime: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	case NIF_TYPE_SIMLE:
#ifdef sun4
	    if (time == 0)
	    {
		if(errno = net_access(h, SIMLEINITTIME))
		    ret = -1;
	    }
	    else
		ret = -1;
#else
	    ret = -1;
#endif /* sun4 */
	    break;
	default:
	    printf("net_settime: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}


/*****************************************
 *
 *	net_current_time
 *
 * This routine returns the current time as kept by
 * the 16-bit timer.  It must be called more often than
 * every 60 seconds or it will lose track of the correct
 * time.  Note that this routine returns a 32 bit number.
 *
 ****************************************/
net_current_time (h)
int h;
{
    register long diff;
    register unsigned long cur_time;
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = npb + h;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    if (np->base_ptr == NULL) {		/* if timer not mapped in */
		errno = EINVAL;
		ret = -1;
		break;
	    }

	    cur_time = np->enp_ptr->timer;
	    diff = cur_time - np->last_time;
	    np->total_elapsed_time += diff;
	    if (diff < 0)
		np->total_elapsed_time += 65536;
	    np->last_time = cur_time;
	    ret = np->total_elapsed_time;
	    break;
	case NIF_TYPE_147:
#ifdef _GT_
	    ret = net_gettime(h);
#else
	    printf("net_current_time: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	case NIF_TYPE_SIMLE:
	    ret = net_gettime(h);
	    break;
	default:
	    printf("net_current_time: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}


/*****************************************
 *
 *	net_current_time_in_ms
 *
 * Same as net_current_time except if we are using the cmc card
 * for time then we must multipy the return value by NETCLKFUDGE
 * This will convert CMC card ticks to milliseconds
 *
 ****************************************/
net_current_time_in_ms (h)
int h;
{
    register long diff;
    register unsigned long cur_time;
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = npb + h;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    if (np->base_ptr == NULL) {		/* if timer not mapped in */
		errno = EINVAL;
		ret = -1;
		break;
	    }

	    cur_time = np->enp_ptr->timer;
	    diff = cur_time - np->last_time;
	    np->total_elapsed_time += diff;
	    if (diff < 0)
		np->total_elapsed_time += 65536;
	    np->last_time = cur_time;
	    ret = (int) (np->total_elapsed_time * NETCLKFUDGE);
	    break;
	case NIF_TYPE_147:
#ifdef _GT_
	    ret = net_gettime(h);
#else
	    printf("net_current_time: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	default:
	    printf("net_current_time: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}


/****************************************
 *
 *	net_init_time
 *
 *	This routine initializes the stuff necessary to
 * use the 16-bit timer.
 *
 ***************************************/
net_init_time (h)
int h;
{
    register struct netlib_pb *np;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = npb + h;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    if (np->base_ptr == NULL) {		/* if timer not mapped in */
		errno = EINVAL;
		ret = -1;
	    } else {
		np->total_elapsed_time = 0;
		np->last_time = np->enp_ptr->timer;
	    }
	    break;
	case NIF_TYPE_147:
#ifdef _GT_
	    ret = net_settime(h, 0);
#else
	    printf("net_init_time: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	case NIF_TYPE_SIMLE:
#ifdef sun4
	    if(errno = net_access(h, SIMLEINITTIME))
		ret = -1;
#endif /* sun4 */
	    break;

	default:
	    printf("net_init_time: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

/****************************************
 *
 *	net_stomp_time
 *
 *	set timer to be timestamp of waiting
 *	xmt packet (if there is one)
 *
 ***************************************/
net_stomp_time (h)
int h;
{
    CHECK_HANDLE(h, -1);

    if(errno = net_access(h, ENPSTOMPTIME))
	return(-1);

    return(0);
}

/*****************************************
 *
 *	net_heartbeat
 *
 *	This routine returns the heartbeat value
 *
 ****************************************/
unsigned short net_heartbeat (h)
int h;
{
    register struct netlib_pb *np;
    unsigned short ret;

    CHECK_HANDLE(h, -1);

    np = npb + h;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    if (np->base_ptr == NULL) {		/* if heartbeat not mapped in */
		errno = EINVAL;
		ret = -1;
	    } else 
	        ret = np->enp_ptr->heartbeat;
	    break;
	default:
	    ret = 0;
	    break;
    }

    return (ret);
}

/*****************************************
 *
 *	net_device_base
 *
 *	This routine returns the base of the mapped in device
 *
 ****************************************/
net_device_base (h)
int h;
{
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, NULL);

    np = npb + h;

    switch(np->type) {
	case NIF_TYPE_CMC:
	    if (np->base_ptr == NULL) {		/* if device not mapped in */
		errno = EINVAL;
		ret = -1;
	    }
	    ret = (int)np->base_ptr;
	    break;
	case NIF_TYPE_147:
	    printf("net_device_base: NIF_TYPE_147 not supported\n");
	    ret = NULL;
	    break;
	default:
	    printf("net_device_base: unknown NIF type!!!\n");
	    ret = NULL;
	    break;
    }

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
