#ifndef lint
static char rcsid [] = "$RCSfile: net_type.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_type.c
 *
 * Copyright (c) 1989 by BBN Systems and Technologies Corporation.
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines for ethernet type code checking.
 *
 * Functions:
 *		net_add_type
 *		net_init_type
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


/*
 * net_add_type
 *
 * Add type code the the list of types the network interface is checking for.
 * Returns 0 if successful, -1 if not.
 */

int net_add_type(h, type)
int h;
int type;
{
    register struct netlib_pb *np;
    int ret = 0;
#ifdef SIMBFLY
    char *buf;
    OID bufoid;
#else
    struct enpaccess_arg arg;
#endif

    CHECK_HANDLE(h, -1);

    np = npb + h;

#ifdef SIMBFLY

    bufoid = Make_Obj((char)0, LOCAL_NODE, sizeof(int), RW_rw_);
    buf = Map_Obj(bufoid, 0, RW_rw_);
    bcopy(&type, buf, sizeof(type));

    np->net_req->a.enpaccess_req.command = ADDTYPE;
    np->net_req->a.enpaccess_req.buffer_to_enp_OID = bufoid;
    np->net_req->a.enpaccess_req.bytes_to_enp = sizeof(type);
    np->net_req->a.enpaccess_req.bytes_to_host = 0;

#else

    arg.command = ADDTYPE;
    arg.buffer_to_enp = (char *)&type;
    arg.bytes_to_enp = sizeof(type);
    arg.bytes_to_host = 0;

    np->params = (char *)&arg;

#endif
#ifdef sun4

    if (np->type == NIF_TYPE_SIMLE)
    {
	*(short *)&arg = (short)type;
	if(errno = net_access(h, SIMLEADDETYPE))
	    ret = -1;
    }
    else
#endif /* sun4 */
	if(errno = net_access(h, ENPACCESS))
	    ret = -1;

#ifdef SIMBFLY

    Del_Obj(bufoid);

#endif

    return(ret);
}

/*
 * net_init_type
 *
 * Init list of type codes the network interface is checking for.
 * Returns 0 if successful, -1 if not.
 */

int net_init_type(h)
int h;
{
    register struct netlib_pb *np;
    int ret = 0;
#ifdef SIMBFLY
#else
    struct enpaccess_arg arg;
#endif

    CHECK_HANDLE(h, -1);

    np = npb + h;

#ifdef SIMBFLY

    np->net_req->a.enpaccess_req.command = INITTYPE;
    np->net_req->a.enpaccess_req.bytes_to_enp = 0;
    np->net_req->a.enpaccess_req.bytes_to_host = 0;

#else

    arg.command = INITTYPE;
    arg.bytes_to_enp = 0;
    arg.bytes_to_host = 0;

    np->params = (char *)&arg;

#endif
#ifdef sun4

    if (np->type == NIF_TYPE_SIMLE)
    {
	if(errno = net_access(h, SIMLEINITETYPE))
	    ret = -1;
    }
    else
#endif /* sun4 */
	if(errno = net_access(h, ENPACCESS))
	    ret = -1;

    return(ret);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
