/*
 * $RCSfile: sim_macros.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */

#ifndef SIM_MACROS

#define SIM_MACROS	1

 /***************************************************************
 *								*
 * FILE:	sim_macros.h					*
 *								*
 * Copyright 1988 BBN Systems and Techologies Corp.		*
 * All rights reserved.						*
 *								*
 ***************************************************************/

/* Macros */

			/* Degrees to radians and back again. 	*/
#define deg_to_rad(x)	(((x) / 180.0) * PI)
#define rad_to_deg(x)	(((x) / PI) * 180.0)

			/* Mils to radians and back again 	*/
#define mil_to_rad(x)   ((((x) * 0.05625) / 180.0) * PI)
#define rad_to_mil(x)   ((((x) / PI) * 180.0) * 17.777777778)

			/* Degrees to mils and back again	*/
#define deg_to_mil(x)	(((x) / 360.0) * 6400.0)
#define mil_to_deg(x)	(((x) / 6400.0) * 360.0)

			/* An infinite loop construct. 		*/
#define FOREVER		for (;;)

			/* Returns the absolute value of x. 	*/
#define abs(x)		(((x) >= 0) ? (x) : (-(x)))

			/* Returns the square of x. 		*/
#define square(x)	((x) * (x))

			/* Returns the magnitude of a two element vector. */
#define mag2(x, y)	((REAL)sqrt((double)(square(x) + square(y))))
#define vec_mag2(vec)	mag2(vec[0], vec[1])

			/* Returns the magnitude of a three element vector. */
#define mag3(x, y, z)  ((REAL)sqrt((double)(square(x) + square(y) + square(z))))
#define vec_mag3(vec)	mag3(vec[0], vec[1], vec[2])

			/* Returns the cube of x. */
#define cube(x)		((x) * (x) * (x))

			/* Is (y-e) <= x <= (y+e), where e is an appropriate 
			   tolerance? */
#define eq(x, y, e)	((abs((x) - (y)) <= (e)) ? TRUE : FALSE)

			/* If |x-z| <= e then return z, else return x. */
#define step(x, z, e)	((abs((x) - (z)) <= (e)) ? (z) : (x))

			/* Returns 1, -1, or 0, depending on the sign of x. */
#define sign(x)		(((x) > 0) ? 1 : (((x) < 0) ? (-1) : 0))

			/* Returns the maximum of a and b. */
#define max(a, b)	(((a) >= (b)) ? (a) : (b))

			/* Returns the minimum of a and b. */
#define min(a, b)	(((a) <= (b)) ? (a) : (b))

			/* Returns the exclusive-or of two booleans. */
#define xor(a, b)	((((a) && (!(b))) || ((!(a)) && (b))) ? TRUE : FALSE)

			/* returns TRUE if a is between b and c (inclusive) */
#define between(a, b, c)	\
			(((((a) >= (b)) && ((a) <= (c))) ||	\
			  (((a) >= (c)) && ((a) <= (b))))	\
			? TRUE : FALSE)

			/* if b < a return a,				*/
			/* else if b > c return c			*/
			/* else (a < b < c) return b			*/
#define	bound(a, b, c)		\
			((b) < (a) ? (a) : ((b) > (c) ? (c) : (b)))

/*
 * p is a probability in [0, 1]
 * rand() returns an integer in [0, MAX_RAND_NUM]
 */
extern int rand();

#if 0
#    define MAX_RAND_NUM 2147483647.0	        /* on Butterfly 2^31 -1 */
#    define MAX_RAND_NUM 32767.0		/* on Masscomp 2^15 -1 */
#endif

#define MAX_RAND_NUM 32767.0		/* rand() in libfail/rand.c */
#define roll_dice(p)	((MAX_RAND_NUM * (p)) > rand())

#define frame_delay_of_sound( range_squared )	\
	( (int)((sqrt((range_squared)) / SPEED_OF_SOUND_M_PER_S) * 15.0 + 0.5))

#define	HEX2(a)		((a) & 0xff)
#define	HEX3(a)		((a) & 0xfff)

#define potval(x)	((int) (idc_values [x] & 0xff))
#define cntrl(x)	((x)-'A'+1)

#ifdef mips
#define BCOPY           bcopy
#else
#ifdef LYNX
#define BCOPY           bcopy
#else
#define BCOPY		movedata
#endif
#endif

#define SIM_ADDRESSES_EQUAL(__a_a1, __a_a2) \
	(((__a_a1)->host == (__a_a2)->host && \
	  (__a_a1)->site == (__a_a2)->site) ? 1 : 0)

#define VEHICLE_IDS_EQUAL(__vid1, __vid2) \
	((__vid1)->vehicle == (__vid2)->vehicle && \
	 SIM_ADDRESSES_EQUAL (&(__vid1)->simulator, &(__vid2)->simulator))

#define HELP_PRINT1(__help_c1, __help_s1)                        \
/* char __help_c1;	   */                                    \
/* char *__help_s1;        */                                    \
	(void) printf ("%c: %s\n", __help_c1, __help_s1)

#define HELP_PRINT2(__help_c1, __help_s1, __help_c2, __help_s2)  \
/* char __help_c1, __help_c2;	   */                            \
/* char *__help_s1,  *__help_s2;   */                            \
	(void) printf ("%c: %s	%c: %s\n", __help_c1, __help_s1, \
	__help_c2, __help_s2)

#endif /* SIM_MACROS */
