/*
 * $RCSfile: libntp_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: ntp_init.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libntpif				    *
*       Created: Fri May 19 14:30:14 EDT 1995				    *
*       Author: gdt							    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBNTPIF_NTP_H_
#define _LIBNTPIF_NTP_H_
/* This file must be K&R C */
/*
 * $Id: libntp_local.h,v 1.1.1.1 1999/04/12 18:49:08 ksnively Exp $
 *
 * libntpif.h
 * Copyright (c) 1995 BBN STD
 *
 * Gregory D. Troxel <gdt@bbn.com>
 * 28 April 1995
 *
 * This file defines the interface between the NTP daemon
 * and the routines to compute time.
 *
 * This file must NOT be included by users of libntpif;
 * it is shared between
 *   implementation of libntpif 
 *   provider of precise time (modified xntp3 daemon only, at the moment)
 * 
 * This file should satisfy both the ModSAF and xntp3 style rules.
 *
 * See libntpif.h for more information.
 */

#include "libntpif.h"

typedef struct ntp_cyclemap
{
    unsigned long cycleSnapshot;
    NTP_TIMEVAL tvSnapshot;
    unsigned long tickNominalRate;
    unsigned long usecNominalRate;
    int valid;
} NTP_CYCLEMAP;
/*
 * The above structure is the basic unit of communication between the
 * provider of precise time and libntpif.
 * cycleSnapshot and tvSnapshot are values of the cycle counter and
 * real time at a particular instant.
 * cycle- and usecNominalRate together give the relationship between
 * cycles and microseconds, specifying that for every tickNominalRate
 * cycles there are usecNominalRate microseconds.  
 * Thus, given the current value of cycles, one can compute how many
 * cycles greater than the snapshot value, compute how many usec this
 * corresponds to, and compute the timeval corresponding to the
 * current cycle value.
 * valid: This is 1 if the structure is valid.
 *  0 is a message from NTP saying "I don't know."
 */

#define MAX_CYCLEMAP 64

typedef struct ntp_shared
{
    unsigned int n_cyclemap;
    unsigned int index_cyclemap;
    NTP_CYCLEMAP cyclemap[MAX_CYCLEMAP];
} NTP_SHARED;
/*
 * This structure exists to convey information from the NTP daemon
 * to user processes.
 * n_cyclemap is the number of entries in the cyclemap array.
 * It must be at least 8 and no more than 64.
 * The writer (NTP daemon) initially sets index_cyclemap to zero,
 * indicating that the structure is invalid, and then sets all values
 * of the cyclemap array to be invalid (valid=0).
 * Once a second, NTP computes the location of the next map entry to
 * be written as (index_cyclemap % n_cyclemap); index_cyclemap is an
 * index into a 2^32 entry virtual array which is mapped into the
 * actual size.
 * Then, NTP writes the cyclemap, and AFTER writing it, increments the
 * index.
 * A reader should read the index, copy the cyclemap corresponding to
 * that index, and read the index again.  If the index has not
 * increased by n_cyclemap-2, the copy is valid because the NTP daemon
 * has 1) already written the map and 2) not started writing it again.
 * For example, if n_cyclemap is 16 and index is 23, the current map
 * entry is 23%16 = 7.  So map 7 is copied, and then the index is read
 * as 37.  The NTP daemon has written index 37/location 5 and updated
 * the index, but may already be writing index 36/location 5.
 * However, it cannot be writing index 37/location 5, so the copy is
 * acceptable.
 */

#define NTP_SHM_KEY (123 << 8 | 123)
/*
 * This is the shared memory key used to name the shared memory
 * segment.
 * It is arbitrary but the use of ftok() is unacceptable because there
 * is no pathname associated with the NTP daemon; the daemon could be
 * in various places and there are not necessarily any configuration
 * files.
 */

#define NTP_TOOLONG        2       /* seconds */
#include <stdext.h>
#ifdef USE_STDERR
#include <stdio.h>
#endif

#include "libntpif.h"

#ifdef sgi
#include <sys/ipc.h>
#include <sys/shm.h>

/* from libtime/realtime.c */
#include <sys/time.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/syssgi.h>
#include <sys/immu.h>
#endif
 
extern NTP_SHARED *ntp_shared;
#ifndef NO_VOLATILE
extern volatile unsigned *ntp_iotimer_addr;
#else
extern unsigned *ntp_iotimer_addr;
#endif

extern unsigned int ntp_cycles_toomany;
extern NTP_CYCLEMAP ntp_fallback_cyclemap;

#endif /* _libntpif_ntp_h_ */
