/*
 * $RCSfile: libntpif.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: ntp_init.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Header file for libntpif				    *	
*       Created: Fri May 19 14:30:14 EDT 1995				    *
*       Author: gdt							    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#ifndef _LIBNTPIF_H_
#define _LIBNTPIF_H_
/* This file must be K&R C */
/*
 * $Id: libntpif.h,v 1.1.1.1 1999/04/12 18:49:09 ksnively Exp $
 *
 * libntpif.h
 * Copyright (c) 1995 BBN STD
 *
 * Gregory D. Troxel
 * 28 April 1995
 *
 * This file defines the interface to precise time as computed by NTP
 * optionally via the cycle counter.
 * The points of the scheme are
 *  1) have accurate time available (requires active synchronization)
 *  2) have low overhead of getting time (requires cycle counter)
 *  3) avoid OS bugs in maintaining system clock (requires cycle counter)
 *  4) fall back to gettimeofday()/etc. if NTP daemon not running
 */

/*
 * These structures are the return value of the routine to read time.
 *
 * This MUST be the same as the struct timeval used by the NTP daemon.
 */
typedef struct ntp_timeval
{
    long sec;
    long usec;
} NTP_TIMEVAL;

/*
 * POSIX.4-style time structure.
 * For now, may be simple 'nsec = 1000 * usec'.
 */
typedef struct ntp_timespec
{
    long sec;
    long nsec;
} NTP_TIMESPEC;

/*
 * Call to initialize ntpif timing.
 * This routine may be called any number of times.
 * If the link to the daemon does not exist, it attempts to create it.
 * Then, if the link is up, a sanity check is performed.
 * If the sanity check fails, the link is declared down.
 * The return value indicates the current status.
 * This routine should execute fairly quickly, but there are no
 * hard realtime constraints.
 */
typedef enum
{
    NTP_INIT_OK =		0, /* all is well */
    NTP_INIT_CYCLE =		1, /* couldn't map cycle counter */
    NTP_INIT_NTP =		2, /* couldn't map NTP segment */
    NTP_INIT_SANITY =		3, /* sanity check ntp/cycle failed */
    NTP_INIT_GENERIC =	9 /* something else wrong */
} NTP_INIT_ENUM;
extern NTP_INIT_ENUM ntp_init();

/*
 * This is the return value of ntp_get_timeval and ntp_get_timespec
 */

typedef enum
{
    NTP_GET_OK =		0, /* all is well; value is NTP-derived */
    NTP_GET_CYCLE =		1, /* fell back to tick counter */
    NTP_GET_FALLBACK =	2  /* fell back to gettimeofday() */
} NTP_GET_ENUM;

/*
 * Fill in *tvp or *tsp with the current time.
 * Return values are NTP_GET_*
 */
extern NTP_GET_ENUM ntp_get_timeval( /* NTP_TIMEVAL *tvp */ );
extern NTP_GET_ENUM ntp_get_timespec( /* NTP_TIMESPEC *tsp */ );

#endif /* _libntpif_h_ */
