#ifndef lint
static char rcsid [] = "$RCSfile: ntp_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ntp_init.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libntpif				    *
*       Created: Fri May 19 14:30:14 EDT 1995				    *
*       Author: gdt							    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libntp_local.h"
#include </usr/include/unistd.h>

NTP_SHARED *ntp_shared = 0;

#ifndef NO_VOLATILE
volatile
#endif
        unsigned *ntp_iotimer_addr = NULL;
unsigned int ntp_cycles_toomany = 0;
NTP_CYCLEMAP ntp_fallback_cyclemap;

static void find_iotimer();

NTP_INIT_ENUM ntp_init()
{
#ifdef sgi
    int shmid;
    
    /* MAP CYCLE COUNTER */
    find_iotimer();
    if ( ntp_iotimer_addr == NULL )
      return NTP_INIT_CYCLE;

    /* MAP NTP DAEMON SHARED MEMORY */
    /* get id for shared memory segment - don't create */
    shmid = shmget(NTP_SHM_KEY, 4096, 0);
    if ( shmid < 0 )
      /* no segment present; implies daemon never started */
      return NTP_INIT_NTP;

    /* actually attach it readonly */
    ntp_shared = shmat(shmid, 0, SHM_RDONLY);
    if ( ntp_shared == (NTP_SHARED *) -1  )
      /* can't attach - should never happen */
    {
	ntp_shared = 0;
	return NTP_INIT_NTP;
    }

    if ( ntp_shared->n_cyclemap == 0 )
      /* no daemon has written to shm */
      return NTP_INIT_SANITY;

    return NTP_INIT_OK;

#else

    return(NTP_INIT_GENERIC);

#endif
}

/* stuff from libtime/realtime.c */

/* fd is opened but never closed.  is this right? */

/* Some SGI machines have a high-resolution real time clock which can
 * be read very efficiently.  Include the appropriate headers to allow
 * us to attempt to use it.
 */
#ifdef sgi


static unsigned int tick_gcd(a, b)
   unsigned int a;
   unsigned int b;
{
    if ( a < b )
      return tick_gcd(b, a);

    if ( b == 0 )
      return a;

    return tick_gcd(b, a % b);
}


static void find_iotimer()
{
    unsigned phys_addr, raddr;
    unsigned            cycleval;
    unsigned            cycle_gcd;
    int                 fd, poffmask;
    unsigned            r0, 
                        r1, 
                        r2, 
                        cnt;

    if ( ntp_iotimer_addr != NULL )
        return;			/* already attached */


    if (!(phys_addr = syssgi(SGI_QUERY_CYCLECNTR, &cycleval)))
    {
#ifdef USE_STDERR
	fprintf(stderr,
	"Hardware cycle counter not available.  Using gettimeofday.\n");
#endif
	ntp_iotimer_addr = NULL;
	return;
    }

    /* fill in fallback ntpif_cycle */
    ntp_fallback_cyclemap.tickNominalRate = 1000000;
    ntp_fallback_cyclemap.usecNominalRate = cycleval;

    /*
     * Special kludge for "174" MHz machines which are really 174.7 MHz
     * This is due to an SGI bug; the ps/cycle value they return is just
     * plain wrong.  -- gdt, 11/14/95
     */
    if ( ntp_fallback_cyclemap.usecNominalRate == 45977 )
      {
#ifdef USE_STDERR
	fprintf(stderr, "tick_counter: 174 MHz 45977 ps/cycle detected -> 45792\n");
#endif
	/*
	 * See the RITN-patched NTP source, file lib/systime.c for
	 * explanation of where this number comes from.
	 */
	ntp_fallback_cyclemap.usecNominalRate = 45792;
      }

    /* compute gcd (after the 174.7/174 kludge!) */
    cycle_gcd = tick_gcd(ntp_fallback_cyclemap.tickNominalRate,
			 ntp_fallback_cyclemap.usecNominalRate);

    /* divide by gcd */
    ntp_fallback_cyclemap.tickNominalRate /= cycle_gcd;
    ntp_fallback_cyclemap.usecNominalRate /= cycle_gcd;

    ntp_fallback_cyclemap.cycleSnapshot = 0;
    ntp_fallback_cyclemap.tvSnapshot.sec = 0; /* too old to be used */
    ntp_fallback_cyclemap.tvSnapshot.usec = 0; /* too old to be used */
    ntp_fallback_cyclemap.valid = 1;

    ntp_cycles_toomany = NTP_TOOLONG * 1000000;
    /* order of computation is less accurate but avoids overflow */
    ntp_cycles_toomany /= ntp_fallback_cyclemap.usecNominalRate;
    ntp_cycles_toomany *= ntp_fallback_cyclemap.tickNominalRate; 

    /* I don't understand any of this addressing nonsense, but the
     * 5.3 & 6.2 syssgi man pages say to do it this way...
     */
    poffmask = getpagesize() - 1;
    raddr = ((unsigned)phys_addr & ~poffmask);
    fd = open("/dev/mmem", O_RDONLY);
#ifndef NO_VOLATILE
    ntp_iotimer_addr = (volatile unsigned *)mmap(0, poffmask, PROT_READ,
					     MAP_PRIVATE, fd, (int)raddr);
#else
    ntp_iotimer_addr = ( unsigned *)mmap(0, poffmask, PROT_READ,
					 MAP_PRIVATE, fd, (int)raddr);
#endif

    if ((int)ntp_iotimer_addr == -1)
    {
#ifdef USE_STDERR
	perror("Unable to access hardware cycle counter");
	fprintf(stderr,"Using gettimeofday.\n");
#endif
	ntp_iotimer_addr = NULL;
	return;
    }
    ntp_iotimer_addr = (unsigned *)((unsigned)ntp_iotimer_addr + (phys_addr &poffmask));

    /* There are three kinds of hardware timers available on the SGI:
     * 24 bit, 32 bit, and 64 bit.  Apparently there is no way (!) to
     * find out which kind we have.  If we have a 64 bit one, we need
     * to read the second word.  However, since this is a cycle
     * counter, the lowest-order bits should change every time we read
     * them.  Thus we can be sneaky and read the clock three times.
     * If it changes every time, then we have the low order bits.
     * If it doesn't, then we have the (not always available) high
     * order 32 bits.  If we have the low order bits, make sure that
     * we have a full 32 bit counter.  A 24 bit counter will, presumably,
     * always have 0's in the high-order 8 bits.
     *
     * The manpage says that only the R3k have 24, that most R4k have 32,
     * and that challenge/onyx have 64.
     */
    r0 = *ntp_iotimer_addr;
    r1 = *ntp_iotimer_addr;
    r2 = *ntp_iotimer_addr;
    if ((r0 == r1) || (r1 == r2))
    {
#ifdef USE_STDERR
	fprintf(stderr,
		"Using low order bits of 64 bit cycle counter.\n");
#endif
	ntp_iotimer_addr += 1;
    }
    else
    {
	/* We need to execute no more than 2^24 instructions to determine
	 * whether we have a full 32 bit counter.  This simple loop will
	 * have at least 6 instructions per iteration (probably many more).
	 * Therefore we need to iterate about (2^24)/6 = 3 million times.
	 */
	for (cnt=0;cnt<3000000;cnt++)
	    if (*ntp_iotimer_addr & 0xff000000)
	    {
#ifdef USE_STDERR
	        fprintf(stderr, "Using 32 bit cycle counter.\n");
#endif
	        return;
	    }

#ifdef USE_STDERR
	fprintf(stderr,
		"Only 24 bit cycle counter available.  Using gettimeofday.\n");
#endif
	ntp_iotimer_addr = NULL;
    }
}
#endif
