#ifndef lint
static char rcsid [] = "$RCSfile: ntp_time.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ntp_time.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Time utility funcitons
*       Created: Fri May 19 14:30:14 EDT 1995
*       Author: gdt
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include <sys/time.h>
#include "libntp_local.h"

static unsigned int ntp_last_index = 0;
static unsigned int ntp_bad_index = 0;
static NTP_CYCLEMAP ntp_cyclemap;

static int compute_cycle();

/* 
 * First, get time from cycle/NTP-data.
 * If can't do, or is old, fall back to checking
 * if our data about cycle counter is recent.
 * If so, use it.
 * Else, do gettimeofday and save cycle data if possible
 *
 * This routine uses GOTO to jump to a fallback case
 * from several places.  It is always forward, and never
 * more complex than this.
 */
NTP_GET_ENUM  ntp_get_timeval( tvp )
NTP_TIMEVAL *tvp;
{
#ifdef sgi
    unsigned long cyclevalue;
    int 	  status;

    unsigned int shm_index, shm_ncycle, shm_entry, shm_index_new;

    if ( ntp_iotimer_addr == NULL )	/* cycle counter present? */
    {
        gettimeofday((struct timeval *) tvp, NULL);
        return NTP_GET_FALLBACK;
    }

    if ( ntp_shared == NULL )	/* NTP daemon present? */
      goto cycle;		/* fallback to cycle counter */

    /*
     * Try to get cyclemap from shared memory.
     * If we lose, fall back to cycle counter.
     */

    /*
     * Read index and decide to punt, get new map, or reuse old.
     * Ntp writes zero on boot.  If zero, use cycle counter.
     * Daemon should skip zero on wrap (every 48 days).
     * If this index determined to be bad before, punt also.
     */
    shm_index = ntp_shared->index_cyclemap;
    if ( shm_index == 0 || shm_index == ntp_bad_index )
      goto cycle;


    /*
     * Do we need to read shm again?
     * ntp_last_index is zero initially, and shm_index is never zero here.
     */
    if ( shm_index != ntp_last_index )
    {
        /* compute entry from index */
        shm_ncycle = ntp_shared->n_cyclemap;
        shm_entry = shm_index % shm_ncycle;
        if ( shm_entry >= MAX_CYCLEMAP ) /* unsigned, so not < 0 */
	  goto ntpcycle;			/* bad value */
      
        /* copy our version */
        bcopy(&ntp_shared->cyclemap[shm_entry],
	     &ntp_cyclemap,
	     sizeof(NTP_CYCLEMAP));
      
        /* check if entry we read is still valid
         * It isn't if the index has increased to near where it might be
         * written. or if index has decreased.
         * Note that this is a check on index, not entry.
         * The index always increases by one.
         * We assume it won't wrap by 2^32 (48 days at 1 sec update)
         */
        shm_index_new = ntp_shared->index_cyclemap;
        if ( shm_index + shm_ncycle - 2 <= shm_index_new ||
	     shm_index_new < shm_index )
	  goto ntpcycle;		/* no good; possibly corrupt data */

        /* note index corresponding to this map */
        ntp_last_index = shm_index;
    }

    /*
     * Now we have a valid cyclemap
     * We must read the cycle counter AFTER the cyclemap, to make
     * sure the value we get is GREATER than the one in the cyclemap.
     */
    cyclevalue = *ntp_iotimer_addr;	/* read cycle counter */
    status = compute_cycle(tvp, cyclevalue, &ntp_cyclemap);
    if ( status < 0 )
    {
        ntp_bad_index = shm_index; /* this cyclemap no good */
        goto ntpcycle;
    }
    return NTP_GET_OK;		/* all is ok */

    ntpcycle: 
    ntp_fallback_cyclemap.tvSnapshot.sec = 0; /* don't use previous */

    cycle:
    /* something's wrong with NTP, so compute based on fallback cyclemap */

    cyclevalue = *ntp_iotimer_addr;	/* read cycle counter */
    status = compute_cycle(tvp, cyclevalue, &ntp_fallback_cyclemap);
    if ( status < 0 )
    {
        /* Too long.  Save current relationship. */
        ntp_fallback_cyclemap.cycleSnapshot = cyclevalue;
        gettimeofday((struct timeval *) &ntp_fallback_cyclemap.tvSnapshot,
		     NULL);

        /* return gettimeofday value */
        tvp->sec = ntp_fallback_cyclemap.tvSnapshot.sec;
        tvp->usec = ntp_fallback_cyclemap.tvSnapshot.usec;
        return NTP_GET_FALLBACK;
    }
    /* was ok */
    return NTP_GET_CYCLE;
#else /* sgi */
    gettimeofday((struct timeval *) tvp, NULL);
    return NTP_GET_FALLBACK;
#endif /* sgi */
}

#ifdef sgi
/*
 * Compute time from cycle value and map.
 * Return 0 if OK, and -1 if too old.
 */

#define USEC_USE_LONGLONG
/* #define USEC_USE_DOUBLE */

static int
compute_cycle(tvp, cycleval, cyclemapp)
NTP_TIMEVAL   *tvp;
unsigned long  cycleval;
NTP_CYCLEMAP  *cyclemapp;
{
    unsigned int ticks_used;
    unsigned int usec_used;
    unsigned int new_usec, new_sec;
  
    /*
     * 32-bit int would overflow here.
     * long long is not ANSI, but both sgicc and gcc support it
     * Note that this code is inside "#ifdef sgi"
     */
#ifdef USEC_USE_LONGLONG
    unsigned long long usec_extra;
#else
#ifdef USEC_USE_DOUBLE
    double usec_extra;
#else
#error MUST DEFINE LONGLONG OR DOUBLE
#endif
#endif

    if ( cyclemapp->tvSnapshot.sec == 0 )
      return -1;

    /* ticks elapsed between snapshot and now */
    ticks_used = cycleval - cyclemapp->cycleSnapshot;

    /* if too long, refuse to use this map */
    if ( ticks_used > ntp_cycles_toomany )
      return -1;

    /* compute number of us corresponding to ticks*/
    if ( cyclemapp->usecNominalRate == 1 )
      {
	/*
	 * If 1 usec for so many ticks, just divide.
	 * This can't overflow.
	 */
	usec_used = ticks_used / cyclemapp->tickNominalRate;
      }
    else
      {
	/*
	 * Be careful about overflow here.
	 * ticks could easily be 25e6 * 2, or 26 bits.
	 * tick/usec rate could be 250k/11493 (174 mhz ip22)
	 * or 1000/21 (onyx).
	 * So tick * usecrate could be 26 bits * 16 bits, which overflows 32.
	 */

	/*
	 * assignment precedes multiply to force use of long long
	 * arithmetic.
	 */
	usec_extra = ticks_used;
	usec_extra *= cyclemapp->usecNominalRate;
	usec_extra /= cyclemapp->tickNominalRate;
	
	/*
	 * convert to long
	 * Only 2 s of ticks, which is 2e6 of usec, are allowed.
	 * Thus, this is less than 2000000, which fits in 32 bits.
	 */
	usec_used = usec_extra;
      }

    /* compute value of time; don't change map */
    new_usec = cyclemapp->tvSnapshot.usec + usec_used;
    new_sec = cyclemapp->tvSnapshot.sec;
    while ( new_usec > 1000000 )
    {
        new_usec -= 1000000;
        new_sec++;
    }
    tvp->usec = new_usec;
    tvp->sec = new_sec;

    return 0;
}

#endif /* sgi */

/*
 * Just call ntp_get_timeval and convert.
 * This routine is specified so that if better data becomes
 * available it can be used.
 */
NTP_GET_ENUM ntp_get_timespec( tsp )
NTP_TIMESPEC *tsp;
{
    NTP_TIMEVAL tv;
    int         retval;

    retval = ntp_get_timeval(&tv);

    tsp->sec = tv.sec;
    tsp->nsec = tv.usec * 1000;

    return retval;
}
