/*
 * $RCSfile: libotmatch.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libotmatch.h                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libotmatch                         *
*       Created: Mon Jun 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBOTMATCH_H_INCLUDED
#define _LIBOTMATCH_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> /*common/include/global*/

/* otm_init:
 *
 * Initializes libotmatch, causing it to read its data file
 * (domains.rdr) either from '.' or the specified data path.
 * The flags are as in reader_read.
 * The return value is zero if the read succeeds, or one of the
 * libreader return values: READER_READ_ERROR READER_FILE_NOT_FOUND
 *
 * Note that the libreader function init_symbol_table, and the
 * librdrconst function rdc_init, must be called before this function.
 */
extern int32 otm_init(
    char  *data_path,
    uint32 flags);

typedef struct otm_db_entry
{
    int32   the_number_three;
    uint32  object_type;
    ADDRESS data;
} OTM_DB_ENTRY;

typedef struct otm_db
{
    uint32        n_entries_plus_1;
    OTM_DB_ENTRY *entries[1];
} OTM_DB;

/* otm_incorporate_db:
 *
 * Incorporates the passed list of vehicle type information into the
 * object type database.  The return value is a key which can be used
 * to access this information in the future.  WARNING: libotm
 * reserves the right to sort this data.  Note that libreader
 * data files of the form
 *  ( (<object type> (data data)) (<object type> (data data)) ...)
 * may be passed in without any changes, as follows: 
 *  otm_incorporate_db((OTM_DB *)(data.array), ...)
 * The best_match flag indicates whether queries for this key which
 * cannot find a perfect match should try to find the best match in
 * the database, or should immediately return the default_data.
 * In some pathological cases, default_data will be returned even if
 * best_match is TRUE, so always pass a valid default_data.
 */
extern int32 otm_incorporate_db(
    OTM_DB *database,
    int32   best_match,
    ADDRESS default_data);

/* otm_query:
 *
 * Queries for the keyed information associated with the passed object
 * type.  A NULL return value will never be returned unless default_data
 * was specified as NULL in otm_incorporate_db.  The returned data
 * should be treated as *read only* by the caller.
 */
extern ADDRESS otm_query(
    int32  key,
    uint32 object_type);

/* otm_peek:
 *
 * Like otm_query, but will return NULL if exact match is not found,
 * regardless of "best_match" or "default_data" setting. 
 */
extern ADDRESS otm_peek(
    int32  key,
    uint32 object_type);

/* otm_describe:
 *
 * Generates a string describing the passed object.  For an object type
 * which is recognized, the returned string will be something like:
 * "M1", "T72M", "F15E", etc.  For unrecognized objects, the string
 * will be more like: "SP Armored Tracked Personnel Carrier".
 */
extern char *otm_describe(uint32 object_type);

#ifdef __cplusplus
}
#endif

#endif /*_LIBOTMATCH_H_INCLUDED*/
