#ifndef lint
static char rcsid [] = "$RCSfile: ot_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ot_init.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize/and access the databases               *
*       Created: Tue Jun 30 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libot_local.h"
#include <libreader.h>
#include <libconstants.h>
#include <libprotocol.h>
#include <obj_type.h>
#include <stdext.h> /*common/include/global*/
#include <stdio.h>
#include <stdstring.h> /*common/include/global*/

/* Arbitrarily large constant */
#define MAX_KEYS 1024

static int32 read_ok = 0;
static READER_UNION domain_db;
#define N_DOMAINS (domain_db.array[0].integer-1)
#define DOMAIN_COUNT(x) (domain_db.array[(x)+1].array[0].integer-2)
#define SHIFTED_DOMAIN(x) (domain_db.array[(x)+1].array[1].integer)
#define DOMAIN_MASK(x,y) (domain_db.array[(x)+1].array[y+2].integer)

static struct
{
    int32   n_entries;
    OTM_DB *database;
    int32   best_match;
    ADDRESS default_data;
} databases[MAX_KEYS];
static int32 n_databases = 0;

int32 otm_init(
    char   data_path[],
    uint32 flags)
{
    int32 i;
    char *new_domain;
    int32 ret = reader_read("domains.rdr", data_path, &domain_db,
			    flags & ~READER_TYPING);

    if (ret != 0)
      return ret;

    /* First, check that the domains are all listed in the right order.
     */
    for (i=0;i<N_DOMAINS;i++)
    {
	if (DOMAIN_COUNT(i) < 0)
	{
	    printf("Error in domains.rdr database: entry %d has no domain listed\n",
		   i);
	    return READER_READ_ERROR;
	}
	if (SHIFTED_DOMAIN(i) != (i << objectDomainShift))
	{
	    printf("Error in domains.rdr database: entry %d is for domain %s\n",
		   i, const_value_to_name(reader_get_symbol("ObjectType"),
					  SHIFTED_DOMAIN(i)));
	    return READER_READ_ERROR;
	}
    }

    read_ok = 1;

    /* Now snoop through the macros and see if there is a domain which
     * should be listed, but isn't.
     */
    new_domain = const_value_to_name(reader_get_symbol("ObjectType"),
				     N_DOMAINS);
    if (!strncmp(new_domain, "objectDomain", strlen("objectDomain")))
    {
	printf("Warning: The domain %s is not listed in the\n         libotmatch domains.rdr database\n",
	       new_domain);
    }

    return ret;
}

int32 otm_incorporate_db(
    OTM_DB *database,
    int32   best_match,
    ADDRESS default_data)
{
    int32 key;

    key = n_databases;
    n_databases += 1;

    databases[key].n_entries = database->n_entries_plus_1 - 1;
    databases[key].database = database;
    databases[key].best_match = best_match;
    databases[key].default_data = default_data;

    /* Sort the database.
     */
    reader_sort_tag_table((READER_UNION *)database, READER_UNTAGGED);

    return key;
}

static int32 has_mask_gap(int32 mask)
{
    int32 found_zeros = FALSE;
    int32 i;
    int32 on;

    /* Big endian search for a group of zeros surrounded by a group of ones */
    for (i = 31; i >= 0; i--)
    {
	on = mask & (1 << i);

	if (on)
	{
	    if (found_zeros)
	      return TRUE;
	}
	else
	  found_zeros = TRUE;
    }
    return FALSE;
}

static OTM_DB_ENTRY *best_match(
    int32  key,
    uint32 object_type)
{
    int32 i, j;
    uint32 d;
    uint32 shifted_domain = object_type & objectDomainMask;
    uint32 matched_mask, matched_value;
    OTM_DB_ENTRY *best_entry;
    uint32 domain = shifted_domain >> objectDomainShift;
    uint32 desired_mask, desired_value;
    uint32 this_ot, this_value;

    /* Find the first entry in the database with this domain.
     */
    for (i=0; i<databases[key].n_entries; i++)
    {
	d = (databases[key].database->entries[i]->object_type &
	     objectDomainMask);
	if (d >= shifted_domain)
	  break;
    }

    /* Did we find a domain larger than this one, meaning there is
     * nothing in this domain?
     */
    if (d != shifted_domain)
    {
	/* If there is a zero entry (it would be first), then that will
	 * do nicely.
	 */
	if (!databases[key].database->entries[0]->object_type)
	  return databases[key].database->entries[0]->data;
	return NULL;
    }

    /* Thus far, we've just matched the domain. */
    matched_mask = objectDomainMask;
    matched_value = object_type & matched_mask;
    best_entry = databases[key].database->entries[i];
    j = 0;
    desired_mask = DOMAIN_MASK(domain,j);
    desired_value = object_type & desired_mask;
    for(; i<databases[key].n_entries;)
    {
	this_ot = databases[key].database->entries[i]->object_type;

	/* Have we gone past the entries with the current mask? */
	if ((this_ot & matched_mask) != matched_value)
	  break;

	/* The upper bits match, so now see if the lower ones are what
	 * we want.
	 */
	this_value = (this_ot & desired_mask);

	if (this_value < desired_value)
	{
	    /* Not there yet... */
	    i++;
	    continue;
	}

	if (this_value == desired_value)
	{
	    /* Cool, we've found an object with the same value in
	     * this sub-type field as the value our object type has
	     * in the corresponding field.  For example, we might
	     * be looking for an M1 and just found the first armored-
	     * tracked vehicle in the database.
	     */

	    /* Remember that any better entry must include these high
	     * order bits.
	     */
	    matched_mask |= desired_mask;
	    matched_value = (object_type & matched_mask);

	    /* Save this as a good-enough value */
	    best_entry = databases[key].database->entries[i];

	    /* Now set our desires for something even more specific. */
	    j += 1;

	    /* Is there anything more specific? */
	    if (j >= DOMAIN_COUNT(domain))
	      break;

	    desired_mask = DOMAIN_MASK(domain,j);
	    desired_value = object_type & desired_mask;

	    /* Note that we don't increment i, because the thing we're
	     * looking at might be have the value we want already...
	     */
	    continue;
	}

	/* this_value > desired_value.  Since the database is sorted,
	 * we aren't going to find the value we want later in the table,
	 * unless we are looking at a field which comes after some fields
	 * that we aren't looking for (i.e., skipped some fields).
	 */
	if (has_mask_gap(matched_mask | desired_mask))
	{
	    /* Keep looking further on */
	    i++;
	    continue;
	}

	break;
    }	

    /* Return the best we found */
    return best_entry;
}

ADDRESS otm_query(
    int32  key,
    uint32 object_type)
{
    OTM_DB_ENTRY *entry;

    /* First see if we can find an exact match.
     */
    if (entry =
	(OTM_DB_ENTRY *)reader_find_tag_sorted((char *)object_type,
					       (READER_UNION *)
					       databases[key].database,
					       READER_UNTAGGED,
					       READER_NO_ERRORS))
      return entry->data;

    /* Are we allowed to grope around for a good match? */
    if (!databases[key].best_match)
      return databases[key].default_data;

    /* Go find a good match. */
    if (entry = best_match(key, object_type))
      return entry->data;

    /* No luck. */
    return databases[key].default_data;
}

ADDRESS otm_peek(
    int32  key,
    uint32 object_type)
{
    OTM_DB_ENTRY *entry;

    /* First see if we can find an exact match.
     */
    if (entry =
	(OTM_DB_ENTRY *)reader_find_tag_sorted((char *)object_type,
					       (READER_UNION *)
					       databases[key].database,
					       READER_UNTAGGED,
					       READER_NO_ERRORS))
      return entry->data;

    return NULL;
}

static char *make_readable(char *s)
{
    static char buf[2048];
    char *b, *c;
    char *orig_s = s;

    for (b=buf;*s;s++,b++)
    {
	if ((*s >= 'A') && (*s <= 'Z'))
	{
	    /* Insert spaces before caps which preceed or follow lowercase */
	    if (((*(s+1) >= 'a') && (*(s+1) <= 'z')) ||
		((s > orig_s) && (*(s-1) >= 'a') && (*(s-1) <= 'z')))
	      *b++ = ' ';
	}
	*b = *s;
    }
    *b = 0;

    /* Change '_' to ' ' */
    for (b=c=buf;*b;b++)
      if (*b != '_')
	*c++ = *b;
      else if (*(b+1) != ' ')
	*c++ = ' ';
    *c = 0;
    
    if (buf[0] == ' ')
      return buf+1;
    else
      return buf;
}

char *otm_describe(uint32 object_type)
{
    char *name;
    uint32 domain;
    int32 i;
    uint32 mask, value;
    char *translated;
    char domain_name[1024];
    char class_name[1024];
    char mask_name[1024];
    char regexp[1024];
    char buf[2048];

    /* First see if there is a constant which corresponds to this object
     * type.
     */
    if ((name = protocol_find_name((int32)object_type)) ==
	const_error)
      name = NULL;
    
    /* See if we know the domain. */
    domain = (object_type & objectDomainMask) >> objectDomainShift;

    if (domain > N_DOMAINS)
    {
	if (name)
	  return make_readable(name);
	else
	  return "Unrecognizable Object";
    }

    /* Find the name of this domain */
    translated =
      const_value_to_name(reader_get_symbol("ObjectType"),
			  (int32) (object_type & objectDomainMask));
    strcpy(domain_name, translated + 12/*strlen("objectDomain")*/);
    domain_name[0] += 'a' - 'A';    

    /* If we know what this is, lop off the domain name, and return it.
     */
    if (name)
    {
	strcpy(buf, name + 1/*_*/ + strlen(domain_name));
	return make_readable(buf);
    }

    /* Cobble together a string describing this object */
    *buf = 0;
    *class_name = 0;
    for (i=0;i<DOMAIN_COUNT(domain);i++)
    {
	mask = DOMAIN_MASK(domain,i);
	value = object_type & mask;

	/* What is this mask called? */
	translated =
	  const_value_to_name(reader_get_symbol("ObjectType"), mask);

	strcpy(mask_name, translated);

	/* Chop off the "Mask" */
	mask_name[strlen(mask_name)-4/*strlen("Mask")*/] = 0;

	/* See if we recognize the value in this spot */
	translated = const_value_to_name(reader_get_symbol("ObjectType"),
					 (int32)value);
	if (translated == const_error)
	  break;

	/* Take everything after the mask name.  For example, if
	 * the value is "vehicleClassFixedWing", take the "FixedWing".
	 */
	strcat(buf, translated+strlen(mask_name));

	/* Assume the first thing we match is a class.  Classes are
	 * sometimes used instead of domains for making mask names.
	 */
	if (!(*class_name))
	{
	    strcpy(class_name, translated+strlen(mask_name));
	    class_name[0] += 'a' - 'A';
	}
    }

    return make_readable(buf);
}
