#ifndef lint
static char rcsid [] = "$RCSfile: pm_create.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pm_create.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to make a parameter change in the PO database        *
*       Created: Thu Jan 14 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpm_local.h"
#include <libpo.h>
#include <p_size.h>
#include <po_size.h>
#include <stdext.h>

void pm_create_parameters(data, size, object_id)
    ADDRESS      data;
    uint32       size;
    ObjectID    *object_id;
{
    PO_DB_ENTRY *entry;
    PO_BUFFER(ParametricInputClass, parms, buf);
    uint32 last_size;
    ObjectID next;

    bzero((char *)object_id, sizeof(ObjectID));

    if (!size)
      return;

    /* Since the pointers go forward, we need to build it back to front.
     */

    last_size = size % maxParametricInputClassSize;
    if (!last_size)
      last_size = maxParametricInputClassSize;

    bzero((char *)&parms->chain, sizeof(ObjectID));
    parms->chainSerial = 1;
    parms->size = last_size;
    bcopy((char *)data + size - last_size,
	  parms->data,
	  (unsigned int)last_size);

    entry = po_create_object(pm_db, NULL, objectClassParametricInput,
			     FALSE, parms, PRO_PO_PI_CLASS_SIZE(last_size),
			     NULL);
    next = PO_OBJECT_ID(entry);

    for (size-=last_size;size;size-=maxParametricInputClassSize)
    {
	parms->chain = next;
	parms->size = maxParametricInputClassSize;
	bcopy((char *)data + size - maxParametricInputClassSize,
	      parms->data,
	      (unsigned int)maxParametricInputClassSize);

	entry = po_create_object(pm_db, NULL, objectClassParametricInput,
				 FALSE, parms,
				 PRO_PO_PI_CLASS_SIZE(maxParametricInputClassSize),
				 NULL);
	next = PO_OBJECT_ID(entry);
    }

    *object_id = next;
}

void pm_update_parameters(data, size, object_id)
    ADDRESS      data;
    uint32       size;
    ObjectID    *object_id;
{
    PO_DB_ENTRY *entry, *next_entry;
    PO_BUFFER(ParametricInputClass, parms, buf);
    ParametricInputClass dummy;
    int32 serial;
    int32 copied;
    ObjectID next;
    PO_DB_ENTRY *kill_list[1024];
    int32 n_kill;

    /* Find the head of the chain */
    if (!(entry = po_get_object(pm_db, object_id)) ||
	(PO_OBJECT_CLASS(entry) != objectClassParametricInput))
      return;

    /* Get the new serial number */
    serial = PO_PARAMETRIC_INPUT_DATA(entry).chainSerial + 1;

    /* Fill out a dummy object which we will need to create if the
     * size of the parameters has grown a lot.
     */
    bzero((char *)&dummy, sizeof(dummy));
    dummy.chainSerial = serial-1; /* So it will be ignored when created */

    /* Copy as much data as we can into the current objects, then
     * start creating new ones, if needed.
     */
    copied = 0;
    while (copied < size)
    {
	next = PO_PARAMETRIC_INPUT_DATA(entry).chain;

	parms->chainSerial = serial;
	if ((size - copied) > maxParametricInputClassSize)
	  parms->size = maxParametricInputClassSize;
	else
	  parms->size = size - copied;
	bcopy((char *)data+copied, 
	      parms->data,
	      (unsigned int)parms->size);

	copied += parms->size;

	if (copied == size)
	{
	    bzero((char *)&parms->chain, sizeof(ObjectID));
	    if (NS_OBJECT_ID_NULL(next))
	      next_entry = NULL;
	    else
	      next_entry = po_get_object(pm_db, &next);
	}
	else
	{
	    /* Make another one, if we'll need it.  Note that we create
	     * it with bogus info this time around, then immediately
	     * change it next time through the loop (this won't happen
	     * very often).
	     */
	    if (!(next_entry = po_get_object(pm_db, &next)) ||
		(PO_OBJECT_CLASS(next_entry) != objectClassParametricInput))
	    {
		next_entry = po_create_object(pm_db, NULL,
					      objectClassParametricInput,
					      FALSE, &dummy,
					      PRO_PO_PI_CLASS_SIZE(0),
					      NULL);
		next = PO_OBJECT_ID(next_entry);
	    }
	    parms->chain = next;
	}

	po_change_object(pm_db, entry, parms,
			 PRO_PO_PI_CLASS_SIZE(parms->size));
	entry = next_entry;
    }

    /* If the size got a lot smaller, delete the unused entries */
    for (n_kill=0;entry;n_kill++,
	 entry=((NS_OBJECT_ID_NULL(PO_PARAMETRIC_INPUT_DATA(entry).chain)) ?
		NULL :
		po_get_object(pm_db, &PO_PARAMETRIC_INPUT_DATA(entry).chain)))
      kill_list[n_kill] = entry;

    if (n_kill)
      po_delete_objects(pm_db, n_kill, kill_list);
}
