#ifndef lint
static char rcsid [] = "$RCSfile: pm_util.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pm_util.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to make a parameter change in the PO database        *
*       Created: Mon Apr 3 1994                                             *
*       Author: wrichard                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpm_local.h"
#include <libpo.h>
#include <p_size.h>
#include <po_size.h>
#include <stdext.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdio.h>

char *pm_get_params(
    PO_DATABASE *db,
    ObjectID    *head,
    int32       *p_size)
{
    PO_DB_ENTRY *params;
    ObjectID next;
    int32 first;
    uint8 serial=0;
    int32 size, last_size;
    char *p_buf = NULL;

    *p_size = 0;
    next = *head;
    first = 1;
    size = 0;
    while ((params = po_get_object(db, &next)) &&
	   (PO_OBJECT_CLASS(params) == objectClassParametricInput))
    {
	if (!first &&
	    (PO_PARAMETRIC_INPUT_DATA(params).chainSerial != serial))
	{
	    /* When the serial numbers don't match, we can't edit.
	     */
	    return NULL;
	}

	/* Append this data to the buffer.
	 */
		  
	/* Make room */
	last_size = size;
	size += PO_PARAMETRIC_INPUT_DATA(params).size;
	if (size > *p_size)
	{
	    *p_size = size;
	    if (!p_buf)
	      p_buf = (char *)STDALLOC(*p_size);
	    else
	      p_buf = (char *)STDREALLOC(p_buf, *p_size);
	}

	/* Copy it in */
	bcopy((char *)PO_PARAMETRIC_INPUT_DATA(params).data,
	      (char *)p_buf + last_size,
	      PO_PARAMETRIC_INPUT_DATA(params).size);
	
	serial = PO_PARAMETRIC_INPUT_DATA(params).chainSerial;
	next = PO_PARAMETRIC_INPUT_DATA(params).chain;
	first = 0;
    }
    return(p_buf);
    
}


/* pm_find_model:
 *
 * Given a list of models which are considered equivalent by the
 * caller, this function finds which, if any, are being used by the
 * passed unit.  The first model found which matches one listed is
 * returned, along with a copy of the parameter values which are
 * currently being used by that model.  If no such model can be found,
 * a NULL pointer is returned.  The returned pointer is dynamically
 * allocated memory and should be freed with a call to STDDEALLOC
 * after being used.
 */

ADDRESS pm_find_model(
    PO_DATABASE *db,
    ObjectID    *vehicle_unit_id,
    int32        num_known_models,
    uint32       known_models[],
    uint32      *model_found,
    uint32      *parameter_size,
    ObjectID    *parameter_object)
{
    return pm_find_model_with_name(db, NULL, vehicle_unit_id, 
				   num_known_models, 
				   known_models, model_found, parameter_size,
				   parameter_object);
}

ADDRESS pm_find_model_with_name(
    PO_DATABASE *db,
    char        *type,
    ObjectID    *vehicle_unit_id,
    int32        num_known_models,
    uint32       known_models[],
    uint32      *model_found,
    uint32      *parameter_size,
    ObjectID    *parameter_object)
{
    int32 j,k;
    PO_DB_ENTRY *unit, *holder;
    ParametricInputHolderClass *holder_data;
    ADDRESS p_data;
    int32 p_size;
    ObjectID head;
    int32 default_size;

    /* Check the unit */
    if (NS_OBJECT_ID_NULL(vehicle_unit_id) ||
        !(unit = po_get_object(db, vehicle_unit_id)) ||
        (PO_OBJECT_CLASS(unit) != objectClassUnit))
    {
	printf("PM: not a unit.\n");
	return NULL;
    }
    
    /* Get the holder for this unit, if there is one */
    if (!NS_OBJECT_ID_NULL(PO_UNIT_DATA(unit).parameters) &&
	(holder = po_get_object(db, 
				&PO_UNIT_DATA(unit).parameters)) &&
	(PO_OBJECT_CLASS(holder) == objectClassParametricInputHolder))
      holder_data = &PO_PARAMETRIC_INPUT_HOLDER_DATA(holder);
    else
      holder_data = NULL;


    /* Are there already overrides registered for this editor? */
    bzero(&head, sizeof(ObjectID));
    p_data = NULL;
    
    if (holder_data)
    {
	for (k=0;k<num_known_models;k++) /* for each given model */
	{
	    /* search through the registered models for a match. */
	    for (j=0;j<holder_data->size;j++)
	    {
		if ((holder_data->blocks[j].model == known_models[k]) &&
		    !(NS_OBJECT_ID_NULL(holder_data->blocks[j].data)))
		{
		    head = holder_data->blocks[j].data;
		    p_data = pm_get_params(db, &head, &p_size);
		    if (p_data)
		    {
			*model_found = holder_data->blocks[j].model;
			*parameter_size = p_size;
			*parameter_object = head;
			return(p_data);
		    }
		}
	    }
	}
    }

    if (!p_data)
    {
	/* Get the official type name of this unit */
	if (!type)
	  type = pm_type_symbol(PO_UNIT_DATA(unit).objectType,
				PO_UNIT_DATA(unit).methodology);

	/* Are there defauls for any of these models? */
	for (k=0;k<num_known_models;k++)
	  if ((p_data = pm_read_defaults_with_size(type, known_models[k],
						   &default_size)))
	  {
	      *model_found = known_models[k];
	      /* set the size of the parameters */
	      *parameter_size = (uint32) default_size;
	      bzero((char *)parameter_object,sizeof(ObjectID));
	      return(p_data); 
	  }
    }

    return NULL;

}

/* pm_update_model:
 *
 * After finding the parameters being used by a vehicle with
 * pm_find_model, an application can use this function to modify
 * these parameters.  A TRUE return value indicates success; FALSE
 * indicates failure.
 */
int32 pm_update_model(
    PO_DATABASE *db,
    ObjectID    *vehicle_unit_id,
    uint32       model, 
    ADDRESS      parameters, 
    uint32       parameter_size,
    ObjectID    *parameter_object)
{	      
  
    PO_DB_ENTRY *unit, *holder;
    ParametricInputHolderClass *holder_data;
    PO_BUFFER(ParametricInputHolderClass, new_holder, buf);
    UnitClass unit_copy;
    

    /* Check the unit */
    if (NS_OBJECT_ID_NULL(vehicle_unit_id) ||
        !(unit = po_get_object(db, vehicle_unit_id)) ||
        (PO_OBJECT_CLASS(unit) != objectClassUnit))
    {
	printf("PM: not a unit.\n");
	return FALSE;
    }

    if (!NS_OBJECT_ID_NULL(PO_UNIT_DATA(unit).parameters) &&
	(holder = po_get_object(db, 
				&PO_UNIT_DATA(unit).parameters)) &&
	(PO_OBJECT_CLASS(holder) == objectClassParametricInputHolder))
      holder_data = &PO_PARAMETRIC_INPUT_HOLDER_DATA(holder);
    else
      holder_data = NULL;
     
    if (NS_OBJECT_ID_NULL(*parameter_object))
    {
	pm_create_parameters(parameters, parameter_size, parameter_object);
        /* Do we need to make a holder? */
        if (!holder_data)
        {
            /* Make the holder */
            bzero(buf, sizeof(buf));
            new_holder->size = 1;
            new_holder->blocks[0].model = model;
            new_holder->blocks[0].data = *parameter_object;
            holder = po_create_object(db, NULL,
                                      objectClassParametricInputHolder,
                                      FALSE, new_holder,
                                      PRO_PO_PIH_CLASS_SIZE(1),
                                      NULL);
            if (!holder)
              return FALSE;

            /* Stick a pointer to the holder into the unit */
            bcopy(&PO_UNIT_DATA(unit), &unit_copy, PRO_PO_UNIT_CLASS_SIZE);
            po_clear_change_flags(&unit_copy);
            unit_copy.parameters = PO_OBJECT_ID(holder);
            po_change_object(db, unit,
                             &unit_copy, PRO_PO_UNIT_CLASS_SIZE);
            return TRUE;
        }

        bcopy(holder_data, new_holder,
              PRO_PO_PIH_CLASS_SIZE(holder_data->size));
        new_holder->blocks[new_holder->size].model = model;
        new_holder->blocks[new_holder->size].data = *parameter_object;
        new_holder->size += 1;
        po_change_object(db, holder,
                         new_holder, PRO_PO_PIH_CLASS_SIZE(new_holder->size));

	return TRUE;
    }
    else
    {
	pm_update_parameters(parameters, parameter_size, parameter_object); 
	return TRUE;
    }
}
