#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libparmgr                                *
*       Created: Fri May 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libparmgr.h"
#include <libreader.h>
#include <stdext.h> /*common/include/global*/

/*** Model 1 module ***/

#define Model1 1

typedef struct
{
    int32 foo;
} MODEL1_PARAMETRIC_DATA;

static MODEL1_PARAMETRIC_DATA * model1_parser(
    READER_UNION *libreader_format,
    uint32        model_num,	/* ignored for now, but required */
    int32        *size)
{
    READER_UNION *my_entry;
    MODEL1_PARAMETRIC_DATA *result;

    my_entry = reader_find_tag(reader_get_symbol("model1"),
			libreader_format,
			READER_TAGGED,
			READER_ERRORS_WITH_CONTEXT);

    if (!my_entry)
      return NULL;

    result = (MODEL1_PARAMETRIC_DATA *)malloc(sizeof(MODEL1_PARAMETRIC_DATA));

    *size = sizeof(MODEL1_PARAMETRIC_DATA);

    result->foo = my_entry[2].array[1].integer;

    return result;
}

static void model1_init()
{
    pm_register_parm_fcns(1, model1_parser, NULL, NULL, NULL);
}

/*** Model 2 module ***/

#define Model2 2

typedef struct
{
    int32 bar;
} MODEL2_PARAMETRIC_DATA;

static MODEL2_PARAMETRIC_DATA * model2_parser(libreader_format, model_num, size)
    READER_UNION *libreader_format;
    uint32 model_num;	/* ignored for now, but required */
    int32 *size;
{
    READER_UNION *my_entry;
    MODEL2_PARAMETRIC_DATA *result;

    my_entry = reader_find_tag(reader_get_symbol("model2"),
			libreader_format,
			READER_TAGGED,
			READER_ERRORS_WITH_CONTEXT);

    if (!my_entry)
      return NULL;

    result = (MODEL2_PARAMETRIC_DATA *)malloc(sizeof(MODEL2_PARAMETRIC_DATA));

    *size = sizeof(MODEL2_PARAMETRIC_DATA);

    result->bar = my_entry[2].array[1].integer;

    return result;
}

static void model2_init()
{
    pm_register_parm_fcns(2, model2_parser, NULL, NULL, NULL);
}

/*** Main program ***/

main()
{
    READER_UNION data_file;
    char *defaults;
    int32 i;

    /* Initialize stuff */
    reader_init(0);
    pm_init();

    /* Read the data file */
    reader_read("test.lisp", NULL, &data_file, 0);

    /* Give the data file to libparmgr */
    pm_set_data_table(data_file.array);

    /* Install the model functions */
    model1_init();
    model2_init();

    /* Exercise the library */

    for (i=0;i<3;i++)
    {
	defaults = pm_get_defaults(reader_get_symbol("type1"), Model1);
	printf("type1, Model1, %d (should be 1)\n",
	       ((MODEL1_PARAMETRIC_DATA *)defaults)->foo);

	defaults = pm_get_defaults(reader_get_symbol("type1"), Model2);
	printf("type1, Model2, %d (should be 2)\n",
	       ((MODEL2_PARAMETRIC_DATA *)defaults)->bar);

	defaults = pm_get_defaults(reader_get_symbol("type2"), Model1);
	printf("type2, Model1, %d (should be 3)\n",
	       ((MODEL1_PARAMETRIC_DATA *)defaults)->foo);

	defaults = pm_get_defaults(reader_get_symbol("type2"), Model2);
	printf("type2, Model2, %d (should be 4)\n",
	       ((MODEL2_PARAMETRIC_DATA *)defaults)->bar);
    }
}
