#ifndef LIBPARSER_INCLUDED
#define LIBPARSER_INCLUDED 1
/*****************************************************************************
* File: parse.h
*
*   Contents:       The parser's header file.
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation
*   Created:        Tue Jan 24 13:19:39 1984
*       Modified by Brian Anderson for multiple parsers under CMOS
*   Remarks:
*
*****************************************************************************/
/*
 * $RCSfile: libparser.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */


/* Parser constant defines. */

#define ARGSIZE          30
#define INPUTLINE       200
#define STRINGSIZE      256

/* Legal parse function return values. */

#define PARSE_ERROR             -1  /* Error with cursor positioning.       */
#define PARSE_SUCCESS           -2  /* Cleanup and start new command.       */
#define PARSE_IGNORE            -3  /* Error without cursor reposition.     */
#define PARSE_NEEDKEYWORD       -4  /* Waiting for another keyword.         */
#define PARSE_UPDATE_LINE       -5  /* Just update line and try again.      */
#define PARSE_ADDSPACE          -6  /* Add space to end of line.            */

/* Legal ParseGetToken () function returns. */

#define TOKEN_DATA              1   /* A token is returned.                 */
#define TOKEN_ESCAPE            2   /* A token returned. It ended in escape.*/
#define TOKEN_QUESTION          3   /* A token returned. It ended in '?'.   */
#define TOKEN_EXHAUSTED         4   /* No token return, input exhausted.    */
#define TOKEN_OPTIONAL_DONE     5   /* Optional input exhausted.            */

#define FIELD_ADD               1   /* Add this to current value.           */
#define FIELD_SUBTRACT          2   /* Subtract this from current value.    */
#define FIELD_REPLACE           3   /* Repleace this with current value.    */

/* The parse table structure define. */

typedef struct
{
    int     p_command;
    char   *p_arg1;
    char   *p_arg2;
} PARSE_TABLE;

#define p_arg0(x)   (char *)((x)->p_command)

/* Command tokens used in parse tables. */

#define P_CALL                  1
#define P_END_KEYWORD           2
#define P_END_KEYWORD_SELECT    3
#define P_KEYWORD               4
#define P_KEYWORD_SELECT        5
#define P_PARSE_FUNCTION        6
#define P_PARSE_KEYWORDS        7
#define P_PRINT                 8
#define P_PRINTFIELD            9
#define P_DYNAMIC_PARSE_FUNCTION        10

/* Defines to create parse tables. */

#define CALL(f)             {P_CALL, ((char *)f), NULL},
#define CALL1(f,a)          {P_CALL, ((char *)f), ((char *)a)},
#define COMMAND_DONE        {P_PARSE_FUNCTION, (char *)ParseCommandDone, NULL},
#define CONSTANT(a,b,c)     {((int)a), ((char *)b), ((char *)c)},
#define CONSTANT_TABLE(t)   PARSE_TABLE t[] = {
#define DEFINE_TABLE(t)     PARSE_TABLE t[] = {
#define DO_KEYWORD_TABLE(t) {P_PARSE_FUNCTION, (char *)ParseDoKeywordTable, ((char *)t)},
#define DO_DYNOBOUND_KEYWORD_TABLE(t) {P_DYNAMIC_PARSE_FUNCTION, (char *)ParseDoKeywordTable, ((char *)t)},
#define END_CONSTANT_TABLE  {0, NULL, NULL} };
#define END_DEFINE_TABLE    {0, NULL, NULL} };
#define END_FIELD_TABLE     {0, NULL, NULL} };
#define END_KEYWORD         {P_END_KEYWORD, NULL, NULL},
#define END_KEYWORD_SELECT  {P_END_KEYWORD_SELECT, NULL, NULL},
#define FIELD(a,b,c)        {((int)a), ((char *)b), ((char *)c)},
#define FIELD_TABLE(t)      PARSE_TABLE t[] = {
#define GETCONSTANT(f)      {P_PARSE_FUNCTION, (char *)ParseGetConstant, ((char *)f)},
#define GETDECIMAL(a)       {P_PARSE_FUNCTION, (char *)ParseGetDecimal, ((char *)a)},
#define GETFLOAT(a)       {P_PARSE_FUNCTION, (char *)ParseGetFloat, ((char *)a)},
#define GETFIELDS(f)        {P_PARSE_FUNCTION, (char *)ParseGetFields, ((char *)f)},
#define GETHEX(a)           {P_PARSE_FUNCTION, (char *)ParseGetHex, ((char *)a)},
#define GETOCTAL(a)         {P_PARSE_FUNCTION, (char *)ParseGetOctal, ((char *)a)},
#define GETSTRING(a)        {P_PARSE_FUNCTION, (char *)ParseGetString, ((char *)a)},
#define GET_ON_OR_OFF       {P_PARSE_FUNCTION, (char *)ParseDoKeywordTable, ((char *)g_ParseOnOff)},
#define GET_YES_OR_NO       {P_PARSE_FUNCTION, (char *)ParseDoKeywordTable, ((char *)g_ParseYesNo)},
#define KEYWORD(f,c)        {P_KEYWORD, ((char *)f), ((char *)c)},
#define KEYWORD_SELECT(a)   {P_KEYWORD_SELECT, NULL, ((char *)a)},
#define OPTIONAL            {P_PARSE_FUNCTION, (char *)ParseOptional, NULL},
#define PARSE_FUNCTION(f,a) {P_PARSE_FUNCTION, ((char *)f), ((char *)a)},
#define PARSE_KEYWORDS(t)   {P_PARSE_KEYWORDS, ((char *)t), NULL},
#define PRINT(s)            {P_PRINT, ((char *)s), NULL},
#define PRINT1(s,a)         {P_PRINT, ((char *)s), ((char *)a)},
#define PRINTFIELDS(f,b)    {P_PRINTFIELD, ((char *)f), ((char *)b)},
#define PUTARG(v)           {P_PARSE_FUNCTION, (char *)ParsePutArg, ((char *)v)},
#define SETPROMPT(a)   {P_PARSE_FUNCTION, (char *)ParseSetPrompt, ((char *)a)},
#define SETTOPLEVEL(a) {P_PARSE_FUNCTION, (char *)ParseSetTopLevel, ((char *)a)},

/* second level defines 	-wr */

#define POKE(d,a)           PUTARG(sizeof (*a))         \
			    PUTARG(d)                   \
			    PUTARG(a)                   \
			    CALL(ParPoke)

	/* or for the do-it-yourself-er */

#define POKE_B(d,a)         PUTARG(1)          		\
			    PUTARG(d)                   \
			    PUTARG(a)                   \
			    CALL(ParPoke)
#define POKE_W(d,a)         PUTARG(2)          		\
			    PUTARG(d)                   \
			    PUTARG(a)                   \
			    CALL(ParPoke)
#define POKE_L(d,a)         PUTARG(4)          		\
			    PUTARG(d)                   \
			    PUTARG(a)                   \
			    CALL(ParPoke)

typedef struct
{
    PARSE_TABLE *TopLevel;          /* The top level parser table.          */
    char        *TokenPtr;          /* Pointer to user tokens.              */
    int          TokenType;         /* The type of token we are getting.    */
    int          Optional;          /* Processing optional input.           */
    char       **CommandBuffer;     /* Pointer to previous commands.        */
    char       **CommandBufferEnd;  /* End of previous commands.            */
    char       **EarlierCommand;    /* Pointer to some previous command.    */
    char       **CurrentCommand;    /* Pointer to current command.          */
    char        *First;             /* Pointer to beginning of buffer.      */
    char        *Cur;               /* Pointer to next character to use.    */
    char        *Last;              /* Pointer to last char + 1.            */
    char        *End;               /* Pointer to the end of the buffer.    */
    char        *SaveCur;           /* The last cur.                        */
    char        *CursorMark;        /* Saves value of cursor.               */
    char        *Prompt;            /* The user prompt.                     */
    char        *StopPoint;         /* Used by NextChar to stop input.      */
    long         PrintArg;          /* Argument for use by printf.          */
    int         (*LogFunction)();   /* Pointer to logging function.         */
    long         Argv[ARGSIZE];     /* Arguments.                           */
    int          ArgFree[ARGSIZE];  /* TRUE if allocated item.              */
    char ParseToken[INPUTLINE];     /* A token from the user input.         */
    char InputBuffer[INPUTLINE];    /* The input buffer.                    */
    char SaveBuffer[INPUTLINE];     /* A copy of the last buffer.           */
    char StringBuffer[STRINGSIZE];  /* The text output buffer.              */
} PARSE_DATA;

/*
 * Global data
 */
extern PARSE_TABLE g_ParseOnOff[];
extern PARSE_TABLE g_ParseYesNo[];

/* Extern defines (generated automatically) */

extern void ParseUpdateTable(/* PARSE_TABLE **dst_table_p,
				     PARSE_TABLE *src_table */);
extern void ParseAddTable(/* PARSE_TABLE **dst_table_p,
			    PARSE_TABLE *src_table */);
extern int ParseCommandDone (/* int nu_argc,
				char *nu_argv[],
				register PARSE_DATA *pdp,
				char *arg2 */);
extern int ParseDoKeywordTable (/* int argc,
				   char *argv[],
				   register PARSE_DATA *pdp,
				   char *arg2 */);
extern int ParseGetConstant (/* int argc,
				char *argv[],
				register PARSE_DATA *pdp,
				register PARSE_TABLE *ftable */);
extern int ParseGetFields (/* int argc,
			      char *argv[],
			      register PARSE_DATA *pdp,
			      register PARSE_TABLE *ftable */);
extern int ParseSetPrompt (/* int argc,
			      char *argv[],
			      register PARSE_DATA *pdp,
			      char *arg2 */);
extern int ParseSetTopLevel (/* int argc,
				char *argv[],
				register PARSE_DATA *pdp,
				char *arg2 */);
extern int ParseGetDecimal (/* int argc,
			       char *argv[],
			       register PARSE_DATA *pdp,
			       char *arg2 */);
extern int ParseGetHex (/* int argc,
			   char *argv[],
			   register PARSE_DATA *pdp,
			   char *arg2 */);
extern int ParseGetOctal (/* int argc,
			     char *argv[],
			     register PARSE_DATA *pdp,
			     char *arg2 */);
extern int ParseGetFloat (/* int argc,
			     char *argv[],
			     register PARSE_DATA *pdp,
			     char *arg2 */);
extern int ParseGetString (/* int argc,
			      char *argv[],
			      register PARSE_DATA *pdp,
			      char *arg2 */);
extern int ParsePutArg (/* register int argc,
			   char *argv[],
			   PARSE_DATA *nu_pdp,
			   char *arg2 */);
extern int ParseOptional (/* register int argc,
			     char *nu_argv[],
			     register PARSE_DATA *pdp,
			     char *nu_arg2 */);
extern int ParseConvertNumber (/* int argc,
				  char *argv[],
				  register PARSE_DATA *pdp,
				  char *arg2,
				  int base,
				  char *bname */);
extern int ParseConvertFloat (/* int argc,
				 char *argv[],
				 register PARSE_DATA *pdp,
				 char *arg2 */);

/*
 * Useful commands which can be called from a KEYWORD_SELECT data structure:
 */
extern void parser_clear_screen ();
extern void parser_help();

/*
 * Following used only by libtty
 */
extern int KillLine (/* register PARSE_DATA *pdp */);
extern int KillForward (/* register PARSE_DATA *pdp */);
extern int DoCommand (/* register PARSE_DATA *pdp */);
extern int Back1Char (/* register PARSE_DATA *pdp */);
extern int Delete1Forward (/* register PARSE_DATA *pdp */);
extern int EndOfLine (/* register PARSE_DATA *pdp */);
extern int Forward1Char (/* register PARSE_DATA *pdp */);
extern int GobbleWord (/* register PARSE_DATA *pdp */);
extern int Delete1Backward (/* register PARSE_DATA *pdp */);
extern int LeftWord (/* register PARSE_DATA *pdp */);
extern int RightWord (/* register PARSE_DATA *pdp */);
extern int DeleteWord (/* register PARSE_DATA *pdp */);
extern int InsertChar (/* register PARSE_DATA *pdp, register char ch */);

extern int DoHelp (/* register PARSE_DATA *pdp */);
extern int DoEscape (/* register PARSE_DATA *pdp */);
extern int BeginningOfLine (/* register PARSE_DATA *pdp */);
extern int PreviousCommand (/* register PARSE_DATA * pdp */);
extern int ShowBindings (/* register PARSE_DATA *pdp */);

extern int NextCommand (/* register PARSE_DATA * pdp */);

#endif
