#ifndef lint
static char rcsid [] = "$RCSfile: par_base.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
* File: par_base.c                                                           *
*                                                                            *
*   Contents:       The base routines of this parser.                        *
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation        *
*   Created:        Mon Jan 23 18:11:34 1984                                 *
*                                                                            *
*   Remarks:                                                                 *
*       This forms the heart of this parser.  The functions in this file     *
*   are those parser functions that have special knowledge and therefore     *
*   cannot be added like user-addable parse function (see par_lib.c).        *
*                                                                            *
*****************************************************************************/

#include <stdext.h>

#include "libparser_l.h"

/*****************************************************************************
* ParseInput:  execute a parse table.                                        *
*                                                                            *
*   Arguments:  table,  a pointer to a parse table.                          *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*       This routine executes a parse table.  It finds the end of the table  *
*   and calls ParseList to do the work.  It then processes the parse value   *
*   returned from ParseList.                                                 *
*                                                                            *
*****************************************************************************/

ParseInput (pdp, table)
register PARSE_DATA *pdp;
register PARSE_TABLE *table;
{
    register PARSE_TABLE *end;
    register int          val;
    register int          i;
    extern   char         NextChar ();

    for (i = 0; i < ARGSIZE; i++)
	pdp->ArgFree[i] = FALSE;

    for (end = table; end->p_command!=0; end++)
	;
    BeginningOfLine (pdp);
    val = ParseList (0, (char **) pdp->Argv, pdp, table, end);
    switch (val)
    {
	case PARSE_ERROR:
	    RestoreCursor (pdp);
	    break;
	case PARSE_SUCCESS:
	    BeginningOfLine (pdp);
	    KillForward (pdp);
	    break;
	case PARSE_NEEDKEYWORD:
	    if (!EmptyLine (pdp))
	    	ParseError (pdp, (char *)NULL, "Expecting a keyword");
	    /* This falls through to PARSE_ADDSPACE. */
	case PARSE_ADDSPACE:
	    if (!EmptyLine (pdp))
	    {
		EndOfLine (pdp);
		BackChar (pdp, 1);
		if (NextChar (pdp)!=' ')
		{
		    InsertChar (pdp, ' ');
		    EndOfLine (pdp);
		}
	    }
	    break;
	case PARSE_IGNORE:
	    break;
    }

    for (i = 0; i < ARGSIZE; i++)
    {
	if (pdp->ArgFree[i])
	    Free ((char *) pdp->Argv[i]);
    }

    if (val==PARSE_UPDATE_LINE)
	UpdateLine (pdp);
    else
	RedisplayLine (pdp);
}

/*****************************************************************************
* ParseList:  process a parse table list.                                    *
*                                                                            *
*   Arguments:  argc,   the current value of argc.                           *
*               argv,   the current argv.                                    *
*               first,  the first entry to process.                          *
*               end,    one beyond the last entry to process.                *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*       This and ParseKeyword are the heart of the parser.  This routine     *
*   walks down the parse table executing all special parser functions.       *
*                                                                            *
*****************************************************************************/

int ParseList (argc, argv, pdp, first, end)
register PARSE_DATA *pdp;
int argc;
char *argv[];                   /* changed this from 'char *argv;' -wr */
register PARSE_TABLE *first;
register PARSE_TABLE *end;
{
    register PARSE_TABLE *ptr;
    register int          val;
    register PARSE_TABLE *clause_end;
    int      (*fp)();
    char *(*get_table_fp)(); /* For dynamic tables */
    char *token;
    extern   PARSE_TABLE *ParseFindEndList ();

    for (ptr = first; ptr < end; ptr++)
    {
	switch (ptr->p_command)
	{
	    case P_CALL:
		if (argc >= ARGSIZE)
		    ParsePrint ("ParseList (pdp): internal error, argc to big.\n");
		if (ParseGetToken (pdp, &token)==TOKEN_DATA)
		{
		    ParseError (pdp, (char *) NULL, "Extra data on line");
		    return (PARSE_ERROR);
		}
		if (pdp->TokenType==TOKEN_DATA)
		{
		    CommandLog (pdp);
		    fp = (int (*)())ptr->p_arg1;
		    val = (*fp)(argc, argv, pdp, ptr->p_arg2);
		    return (PARSE_SUCCESS);
		}
		ParseMessage (pdp, "Carriage-return to complete command");
		return (PARSE_IGNORE);
	    case P_DYNAMIC_PARSE_FUNCTION:
		fp = (int (*)())ptr->p_arg1;
		get_table_fp = (char * (*)())ptr->p_arg2;
		val = (*fp)(argc, argv, pdp, (*get_table_fp)());
		if (val < 0)
		  return (val);
		argc = val;
		break;
	    case P_PARSE_FUNCTION:
		fp = (int (*)())ptr->p_arg1;
		val = (*fp)(argc, argv, pdp, ptr->p_arg2);
		if (val < 0)
		    return (val);
		argc = val;
		break;
	    case P_KEYWORD_SELECT:
		clause_end = ParseFindEndList (ptr, end,
		    P_KEYWORD_SELECT, P_END_KEYWORD_SELECT);
		val = ParseKeyword (argc, argv, pdp, ptr, clause_end);
		if (val < 0)
		    return (val);
		argc = val;
		ptr = clause_end;
		break;
	    case P_PRINT:
		if (pdp->TokenType==TOKEN_DATA)
		{
		    if (ptr->p_arg2==NULL)
		        ParsePrint (ptr->p_arg1);
		    else
		        ParsePrint (ptr->p_arg1, *ptr->p_arg2);
		}
		break;
	    case P_PRINTFIELD:
		if (pdp->TokenType==TOKEN_DATA)
		{
		    if (ptr->p_arg2==NULL)
		    	ParsePrint ("ParseTypeFlag (): internal table error in PRINTFIELD.\n");
		    else
		    	TypeFields ((PARSE_TABLE *)ptr->p_arg1, 
				    (int) *ptr->p_arg2);
		}
		break;
	    default:
		ParsePrint ("ParseList (pdp): illegal command (%d).\r\n",
		    ptr->p_command);
	}
    }
    return (argc);
}

/*****************************************************************************
* ParseKeyword:  process a keyword list.                                     *
*                                                                            *
*   Arguments:  argc,   the current value of argc.                           *
*               argv,   the current argv.                                    *
*               first,  the first entry to process (points to KEYWORD entry).*
*               end,    one beyond the last entry to process.                *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*       This and ParseList are the heart of the parser.  This routine        *
*   determines the next keyword and processes the commands following it.     *
*   This code is ugly,  sorry.                                               *
*                                                                            *
*****************************************************************************/

int ParseKeyword (argc, argv, pdp, first, end)
register PARSE_DATA *pdp;
int argc;
char *argv[];
register PARSE_TABLE *first;
register PARSE_TABLE *end;
{
    char *token;
    register PARSE_TABLE *match;
    register PARSE_TABLE *end_match;
    char *arg2;
    int escape_used;
    int common;			/* Length of common part of matches */
    int first_length;		/* Length of this match */
    int token_length;		/* Length of the token */
    int unique;			/* non-zero if just one match */
    extern   PARSE_TABLE *ParseFindEndList ();
    extern   char         NextChar ();

    arg2 = first->p_arg2;
    first++;
    if (first->p_command!=P_KEYWORD)
	ParsePrint ("ParseKeyword (pdp): illegal parse table.\r\n");

    escape_used = FALSE;
    switch (ParseGetToken (pdp, &token))
    {
	case TOKEN_EXHAUSTED:
	    return (PARSE_NEEDKEYWORD);

	case TOKEN_QUESTION:
	    match = NULL;
	    if (arg2!=NULL)
	        ParsePrint ("%s:\r\n", arg2);
	    while (first < end)
	    {
	        if (first->p_command!=P_KEYWORD)
	    	    ParsePrint ("ParseKeyword (pdp): illegal parse table.\r\n");
	        if (ParseMatch (first->p_arg1, token))
	        {
	            ParsePrint ("%-20.20s %s\r\n", first->p_arg1, first->p_arg2);
	    	    match = first;
	        }
	        first = ParseFindEndList (first, end, P_KEYWORD, P_END_KEYWORD);
	        first++;
	    }
	    if (match==NULL)
	    {
	        ParseError (pdp, arg2, "Unable to find a keyword match (? for list)");
	        return (PARSE_ERROR);
	    }
	    return (PARSE_IGNORE);

	case TOKEN_ESCAPE:
	    escape_used = TRUE;
	    /* Warning: this falls through. */
	case TOKEN_DATA:
	    match = NULL;
	    token_length = strlen(token);
	    if (token_length==0)
	    {
		ParseError (pdp, arg2, "Expecting a keyword (? for list)");
		return (PARSE_ERROR);
	    }
	    while (first < end)
	    {
	    	if (first->p_command!=P_KEYWORD)
	    	    ParsePrint ("ParseKeyword (pdp): illegal parse table.\r\n");
	    	if (ParseMatch (first->p_arg1, token))
	    	{
		    first_length = strlen(first->p_arg1);
		    if (first_length == token_length)
		    {		/* Exact match always succeeds */
			match = first;
			unique = 1;
			common = first_length;
			break;
		    }
	    	    if (match==NULL)
		    {
	    	        match = first;
			unique = 1;
			common = first_length;
		    }
		    else
		    {
			if (first_length < common)
			    common = first_length;
			while (strncmp(match->p_arg1, first->p_arg1, common))
			    common--;
			if (!escape_used || common == token_length)
			{
			    ParseError (pdp, arg2,
					"'%s' is not an unique keyword", token);
			    if (!escape_used)
				return (PARSE_ERROR);
			    return (PARSE_IGNORE);
			}
			unique = 0;
		    }
		}

		first = ParseFindEndList (first, end, P_KEYWORD, P_END_KEYWORD);
	    	first++;
	    }
	    if (match==NULL)
	    {
	        ParseError (pdp, arg2,
			    "Unable to find a keyword match (? for list)");
		return (PARSE_ERROR);
	    }
	    if (escape_used)
	    {
		ParseEscapeComplete (pdp, match->p_arg1, unique, common);
	        return (PARSE_UPDATE_LINE);
	    }
	    end_match = ParseFindEndList (match, end, P_KEYWORD, P_END_KEYWORD);
	    match++;
	    return (ParseList (argc, argv, pdp, match, end_match));

	case TOKEN_OPTIONAL_DONE:
	    return (argc);

	default:
	    ParsePrint ("ParseKeyword (pdp): internal error.\r\n");
	return(ERROR);  /* for lint */
    }
}

/*****************************************************************************
* ParseCommandDone:  this routine terminates a command list.                 *
*                                                                            *
*   Arguments:  first,  the first entry to process.                          *
*               end,    one beyond the last entry to process.                *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

/*ARGSUSED*/
int ParseCommandDone (nu_argc, nu_argv, pdp, arg2)
register PARSE_DATA *pdp;
int nu_argc;                    /* not used */
char *nu_argv[];                /* not used */
char *arg2;
{
    char *token;

    if (ParseGetToken (pdp, &token)==TOKEN_DATA)
    {
	ParseError (pdp, arg2, "Extra data on line");
	return (PARSE_ERROR);
    }
    if (pdp->TokenType==TOKEN_DATA)
	return (PARSE_SUCCESS);
    else
    {
	ParseMessage (pdp, "Carriage-return to complete command");
	return (PARSE_IGNORE);
    }
}

/*****************************************************************************
* ParseDoKeywordTable:  execute a new keyword table.                         *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  the pointer to the new table.                         *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseDoKeywordTable (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2; 
{
    register PARSE_TABLE *begin;
    register PARSE_TABLE *end;

    begin = (PARSE_TABLE *) arg2;
    for (end = begin; end->p_command!=0; end++)
	;
    return (ParseList (argc, argv, pdp, begin, end));
}
