#ifndef lint
static char rcsid [] = "$RCSfile: par_const.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*****************************************************************************
* File: par_const.c                                                          *
*                                                                            *
*   Contents:       The procedures to parse constant tables.                 *
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation        *
*   Created:        Tue Mar 20 09:57:08 EST 1984                             *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

#include <stdext.h>
#include "libparser_l.h"

/*****************************************************************************
* ParseGetConstant:  get a constant value.                                   *
*                                                                            *
*   Arguments:  argc,    the argument count.                                 *
*               argv,    the argument array.                                 *
*               ftable,  the field table to use.                             *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

ParseGetConstant (argc, argv, pdp, ftable)
register PARSE_DATA *pdp;
int argc;
char *argv[];
register PARSE_TABLE *ftable;
{
    register PARSE_TABLE *ptr;
    register int  matched;
    register PARSE_TABLE *match;
    PARSE_TABLE *p_const;
    int   escape_used;
    char *token;
    extern char  NextChar ();

    escape_used = FALSE;
    switch (ParseGetToken (pdp, &token))
    {
        case TOKEN_ESCAPE:
	    escape_used = TRUE;
            /* Warning: this falls through. */
        case TOKEN_DATA:
	    if (*token=='\0')
	    {
	        ParseError (pdp, (char *) NULL, "Expecting a constant name (? for list)");
	        if (escape_used)
	            return (PARSE_IGNORE);
	        else
	            return (PARSE_ERROR);
	    }
	    matched = ParseTableFind (pdp, ftable, token, escape_used, &p_const);
	    if (matched!=PARSE_SUCCESS)
	        return (matched);
	    if (!escape_used)
	    {
                argv[argc] = (char *) p_const->p_arg1;
                return (argc + 1);
	    }
	    else
	    {
	        ParseEscapeComplete (pdp, p_arg0(p_const), TRUE, 0);
	        return (PARSE_UPDATE_LINE);
	    }
        case TOKEN_QUESTION:
	    match = NULL;
	    for (ptr = ftable; p_arg0(ptr)!=0; ptr++)
	    {
	        if (ParseMatch (p_arg0(ptr), token))
	        {
		    ParsePrint ("%-20.20s %s\r\n", p_arg0(ptr), ptr->p_arg2);
		    match = ptr;
	        }
	    }
	    if (match==NULL)
	    {
	        ParseError (pdp, (char *) NULL, "Unable to find a constant name match");
	        return (PARSE_ERROR);
	    }
	    return (PARSE_IGNORE);

        case TOKEN_EXHAUSTED:
	    ParseError (pdp, (char *) NULL, "Expecting a constant name (? for list)");
	    return (PARSE_ADDSPACE);

	case TOKEN_OPTIONAL_DONE:
	    return (argc);

        default:
	    ParsePrint ("GetFlags (): internal error.\r\n");
    }
    return(ERROR);  /* for lint */
}
