/*****************************************************************************
* File: par_edit.c
*
*   Contents:       Parser editing commands.
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation
*   Created:        Thu Aug 25 12:05:06 EDT 1983
*
*   Remarks:    The editing buffer look like this:
*
*           +---+---+---+---+---+---+---+---+---+---+
*           | a | b | c | d |   |   |   |   |   |   |
*           +---+---+---+---+---+---+---+---+---+---+
*             |       |       |               |   |
*   First ----+       |       |               |   |
*   Cur   ------------+       |               |   |
*   Last  --------------------+               |   |
*   End   ------------------------------------+   |
*   Room for final '\0' --------------------------+
*
*****************************************************************************/

#ifndef lint
static char rcsid[]= "$RCSfile: par_edit.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

#include "stdio.h" /* $$$ */
#include "libparser_l.h"

static PARSE_DATA *pdp;


/*****************************************************************************
* ChangePrompt:  reset the parser prompt.                                    *
*                                                                            *
*   Arguments:  prompt, the new prompt.                                      *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:    A hack added by pditmars                                     *
*                                                                            *
*****************************************************************************/

ChangePrompt (prompt)
char		*prompt;
{
	pdp->Prompt = prompt;
}
	

ChangeTopLevel (table)
PARSE_TABLE *table;
{
	pdp->TopLevel = table;
}
	
/*****************************************************************************
* InitParser:  initialize the parser.                                        *
*                                                                            *
*   Arguments:  pdp,    a pointer to a PARSE_DATA structure.                 *
*               table,  the initial table.                                   *
*               prompt, the initial prompt.                                  *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

PARSE_DATA *InitParser (table, prompt, logsize)
PARSE_TABLE *table;
char        *prompt;
int          logsize;
{
    extern char *Alloc();

    pdp = (PARSE_DATA *) Alloc (sizeof(PARSE_DATA));
    if (pdp!=(PARSE_DATA *) NULL)
    {	 
        pdp->TopLevel         = table;
        pdp->First            = &(pdp->InputBuffer[0]);
        pdp->Cur              = pdp->First;
        pdp->Last             = pdp->First;
        pdp->SaveCur          = pdp->Cur;
	pdp->CommandBuffer    = (char **) Alloc((int)(sizeof(char *)*logsize));
	pdp->EarlierCommand   = pdp->CommandBuffer;
	pdp->CommandBufferEnd = pdp->CommandBuffer + logsize - 1;
	pdp->CurrentCommand   = pdp->CommandBufferEnd + 1;
	pdp->LogFunction      = NULL;
	pdp->PrintArg         = NULL;

	do
	{
	    pdp->CurrentCommand--;
	    *(pdp->CurrentCommand) = NULL;
	} while (pdp->CurrentCommand > pdp->CommandBuffer);

        /* The -1 leaves room for '\0'. */
        pdp->End       = &(pdp->InputBuffer[INPUTLINE-1]);
        pdp->SaveBuffer[0] = '\0';
        if (prompt)
	    pdp->Prompt = prompt;
        else
            pdp->Prompt = "=> ";
    }

    return pdp;
}

/*****************************************************************************
* UpdateLine:  update the output line to look like input buffer.             *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*       This routine is the heart of the editor.  It takes what is on the    *
*   screen and updates it to be what is in the internal input buffer.        *
*   This version of UpdateLine assumes that it is being run on a CRT that    *
*   is capable of performing backspace (^H).  Other version for hardcopy     *
*   or smarter terminals can be written.                                     *
*                                                                            *
*****************************************************************************/

UpdateLine (pdp)
register PARSE_DATA *pdp;
{
    register char *newp;
    register char *oldp;
    register char *outstr;
    extern   char *UpdateMoveCursor ();

    outstr = pdp->StringBuffer;
    *outstr = '\0';

    /* Find first difference. */
    newp = pdp->InputBuffer;
    oldp = pdp->SaveBuffer;
    while (*newp==*oldp&&*newp!='\0')
    {
	oldp++;
	newp++;
    }

    if (*newp=='\0'&&*oldp == '\0')
        outstr = UpdateMoveCursor (outstr, pdp->SaveCur, 
                        pdp->Cur - pdp->SaveCur);
    else
    {
        /* Move to the first difference. */
        outstr = UpdateMoveCursor (outstr, pdp->Cur, 
                        newp - pdp->SaveCur);
        
        /* Now output the difference. */
        while (*newp!='\0')
	{
            *outstr++ = *newp++;
	    oldp++;
	}
        while (*oldp!='\0')
        {
            *outstr++ = ' ';
            newp++;
            oldp++;
        }
        *outstr = '\0';

        /* Now put the cursor where it belongs. */
        outstr = UpdateMoveCursor (outstr, pdp->SaveCur, 
                        pdp->Cur - newp);
    }

    /* Send it to the terminal. */
    ParsePrint ("%s", pdp->StringBuffer);
#if 0
    fflush(stdout); /* $$$ */
#endif

    /* Save current stuff as old stuff. */
    pdp->SaveCur = pdp->Cur;
    newp = pdp->InputBuffer;
    oldp = pdp->SaveBuffer;
    while (*newp)
	*oldp++ = *newp++;
    *oldp = '\0';
}

/*****************************************************************************
* UpdateMoveCursor:  the moves the cursor to where it is needed.             *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

char *UpdateMoveCursor (outbuf, linebuf, count)
register char *outbuf;
register char *linebuf;
register int count;
{
    while (count < 0)
    {
	*outbuf++ = '\b';
	count++;
    }
    while (count > 0)
    {
	*outbuf++ = *linebuf++;
	--count;
    }
    *outbuf = '\0';
    return (outbuf);
}

/*****************************************************************************
* RedisplayLine:  redisplay the current line (including user prompt).        *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

RedisplayLine (pdp)
register PARSE_DATA *pdp;
{
    ParsePrint ("%s%s", pdp->Prompt, pdp->InputBuffer);
    pdp->SaveCur = pdp->Last;
    UpdateLine (pdp);
}

/*****************************************************************************
* SetStopPoint:  inform NextChar where to stop reading.                      *
*                                                                            *
*   Arguments:  at_cursor,  TRUE if at cursor, FALSE if at end of line.      *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

SetStopPoint (pdp, at_cursor)
register PARSE_DATA *pdp;
int at_cursor;
{
    if (at_cursor)
	pdp->StopPoint = pdp->Cur;
    else
	pdp->StopPoint = pdp->Last;
}

/*****************************************************************************
* NextChar:  read the next character from the input buffer.                  *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    the next character.                                          *
*                                                                            *
*   Remarks:                                                                 *
*       This updates the buffer pointer to point to the next character.      *
*                                                                            *
*****************************************************************************/

char NextChar (pdp)
register PARSE_DATA *pdp;
{

    if (pdp->Cur==pdp->StopPoint)
	return ('\0');

    return (*pdp->Cur++);
}

/*****************************************************************************
* MarkCursor:  mark the current cursor position for later restoral.          *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

MarkCursor (pdp)
register PARSE_DATA *pdp;
{
    pdp->CursorMark = pdp->Cur;
}

/*****************************************************************************
* RestoreCursor:  restore cursor to late mark.                               *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

RestoreCursor (pdp)
register PARSE_DATA *pdp;
{
    pdp->Cur = pdp->CursorMark;
}

/*****************************************************************************
* EmptyLine:  returns TRUE if the input line is empty.                       *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    TRUE if input line is empty.                                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int EmptyLine (pdp)
register PARSE_DATA *pdp;
{
    return (pdp->InputBuffer[0]=='\0');
}

/*****************************************************************************
* BeginningOfLine:  go to the beginning of the line.                         *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

BeginningOfLine (pdp)
register PARSE_DATA *pdp;
{
    pdp->Cur = pdp->First;
}

/*****************************************************************************
* Back1Char:  move cursor back one character.                                *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

Back1Char (pdp)
register PARSE_DATA *pdp;
{
    BackChar (pdp, 1);
}

/*****************************************************************************
* Delete1Forward:  delete the next character.                                *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

Delete1Forward (pdp)
register PARSE_DATA *pdp;
{
    DeleteChar (pdp, 1);
}

/*****************************************************************************
* EndOfLine:  move cursor to the end of the line.                            *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

EndOfLine (pdp)
register PARSE_DATA *pdp;
{
    pdp->Cur = pdp->Last;
}

/*****************************************************************************
* Forward1Char:  move the cursor forward one character.                      *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

Forward1Char (pdp)
register PARSE_DATA *pdp;
{
    ForwardChar (pdp, 1);
}

/*****************************************************************************
* GobbleWord:  delete one word forward.                                      *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*       This is done in PEN's broken way. (Compatibility).                   *
*                                                                            *
*****************************************************************************/

GobbleWord (pdp)
register PARSE_DATA *pdp;
{
    register char *p;
    extern   char *LookForward ();

    p = LookForward (pdp);
    DeleteChar (pdp, p - pdp->Cur);
}

/*****************************************************************************
* Delete1Backward:  delete the last character.                               *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

Delete1Backward (pdp)
register PARSE_DATA *pdp;
{
    if (pdp->Cur > pdp->First)
    {
        BackChar (pdp, 1);
        DeleteChar (pdp, 1);
    }
}

/****************************************************************************
 * KillLine:  kill the line
 *                                                                            *
 *   Arguments:  none.                                                        *
 *                                                                            *
 *   Returns:    nothing.                                                     *
 *                                                                            *
 *   Remarks:                                                                 *
 *                                                                            *
 *****************************************************************************/
KillLine (pdp)
register PARSE_DATA *pdp;
{
    pdp->Cur = pdp->First;
    DeleteChar (pdp, pdp->Last - pdp->First);
}


/*****************************************************************************
* KillForward:  kill to the end of the line.                                 *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

KillForward (pdp)
register PARSE_DATA *pdp;
{
    DeleteChar (pdp, pdp->Last - pdp->Cur);
}

/*****************************************************************************
* LeftWord:  move the cursor one word to the left.                           *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

LeftWord (pdp)
register PARSE_DATA *pdp;
{
    register char *p;
    extern   char *ScanBackwards ();

    p = ScanBackwards (pdp);
    BackChar (pdp, pdp->Cur - p);
}

/*****************************************************************************
* RightWord:  move the cursor one word to the right.                         *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

RightWord (pdp)
register PARSE_DATA *pdp;
{
    register char *p;
    extern   char *ScanForward ();

    p = ScanForward (pdp);
    ForwardChar (pdp, p - pdp->Cur);
}

/*****************************************************************************
* DeleteWord:  delete the next word.                                         *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

DeleteWord (pdp)
register PARSE_DATA *pdp;
{
    register char *p;
    register int   count;
    extern   char *ScanBackwards ();

    p = ScanBackwards (pdp);
    count = pdp->Cur - p;
    BackChar (pdp, count);
    DeleteChar (pdp, count);
}

/*****************************************************************************
* InsertChar:  insert character at current position.                         *
*                                                                            *
*   Arguments:  ch,  the character to insert.                                *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

InsertChar (pdp, ch)
register PARSE_DATA *pdp;
register char ch;
{
    register int  i;
    register char och;
    register int  tocopy;

    if (pdp->Last > pdp->End)
	ParsePrint ("\007");
    else
    {
	if (ch < ' ')           /* Make control chars viewable. */
	{
	    InsertChar (pdp, '^');
	    ch = ch + '@';
	}
	tocopy = pdp->Last - pdp->Cur + 1;
	for (i = 0; i < tocopy; i++)
	{
	    och = *pdp->Cur;
	    *pdp->Cur++ = ch;
	    ch = och;
	}
	BackChar (pdp, tocopy-1);
	pdp->Last++;
	*pdp->Last = '\0';
    }
}

/*****************************************************************************
* ForwardChar:  move forward characters.                                     *
*                                                                            *
*   Arguments:  count,  the number of characters to move.                    *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

ForwardChar (pdp, count)
register PARSE_DATA *pdp;
register int count;
{
    register int i;

    for (i = 0; i < count; i++)
    {
	if (pdp->Cur < pdp->Last)
	    pdp->Cur++;
	else if (pdp->Cur < pdp->End)
	{
	    *pdp->Cur++ = ' ';
	    *pdp->Cur   = '\0';
	    pdp->Last = pdp->Cur;
	}
	else
    	    ParsePrint ("\007");
    }
}

/*****************************************************************************
* BackChar:  move backwards characters.                                      *
*                                                                            *
*   Arguments:  count,  the number of characters to move.                    *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

BackChar (pdp, count) 
register PARSE_DATA *pdp;
register int count;
{
    pdp->Cur -= count;
    if (pdp->Cur < pdp->First)
	pdp->Cur = pdp->First;
}

/*****************************************************************************
* DeleteChar:  delete characters.                                            *
*                                                                            *
*   Arguments:  count,  the number of characters to delete.                  *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

DeleteChar (pdp, count)
register PARSE_DATA *pdp;
register int count;
{
    register char *p;
    register int   i;
    register int   tocopy;

    if ((pdp->Cur+count) > pdp->Last)
	count = pdp->Last - pdp->Cur;

    if (count==0)
	return;

    tocopy = pdp->Last - pdp->Cur - count;
    for (i = 0, p = (pdp->Cur+count); i < tocopy; i++)
	*pdp->Cur++ = *p++;

    BackChar (pdp, tocopy);
    pdp->Last -= count;
    *pdp->Last = '\0';
}

/*****************************************************************************
* ScanForward:  return pointer to next forward word.                         *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    pointer to next word.                                        *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

char *ScanForward (pdp)
register PARSE_DATA *pdp;
{
    register char *p;

    for (p = pdp->Cur; p <= pdp->Last&&*p==' '; p++) ;
    for (         ; p <= pdp->Last&&*p!=' '; p++) ;
    if (p > pdp->Last)
	p = pdp->Last;

    return p;
}

/*****************************************************************************
* ScanForward:  return pointer to next forward word or whitespace.           *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    pointer to next word or whitespace.                          *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

char *LookForward (pdp)
register PARSE_DATA *pdp;
{
    register char *p;

    if (*pdp->Cur==' ')
        for (p = pdp->Cur; 
             p <= pdp->Last&&*p==' '; p++) ;
    else
        for (p = pdp->Cur; 
             p <= pdp->Last&&*p!=' '; p++) ;

    if (p > pdp->Last)
	p = pdp->Last;

    return p;
}

/*****************************************************************************
* ScanForward:  return pointer to next backwards word.                       *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    pointer to next backwards word.                              *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

char *ScanBackwards (pdp)
register PARSE_DATA *pdp;
{
    register char *p;

    for (p = pdp->Cur - 1 ; 
            p > pdp->First&&*p==' '  ; --p) ;
    for (           ; p > pdp->First&&*p!=' '; --p) ;

    if (p!=pdp->First)
        p++;

    return p;
}
