#ifndef lint
static char rcsid [] = "$RCSfile: par_fields.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
* File: par_fields.c                                                         *
*                                                                            *
*   Contents:       The procedures to maintain data fields.                  *
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation        *
*   Created:        Thu Jan 26 09:15:26 EST 1984                             *
*                                                                            *
*   Remarks:                                                                 *
*       This is perhaps the most hairy code in the parser.  This is          *
*   unfortunate, but I was unable to figure out a simpler structure.  If     *
*   you can come up with one I would love to hear about it.                  *
*                                                                            *
*****************************************************************************/

#include <stdext.h>
#include "libparser_l.h"

/*****************************************************************************
* ParseGetFields:  get the flag bits.                                        *
*                                                                            *
*   Arguments:  argc,    the argument count.                                 *
*               argv,    the argument array.                                 *
*               ftable,  the field table to use.                             *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*       This function takes three forms:                                     *
*           txint|rxint     would set the flags word to txint and rxint.     *
*           +txint|rxint    would add txint and rxint to flag.               *
*           -txint|rxint    would subtract txint and rxint from flag.        *
*           0               would clear all flags.                           *
*                                                                            *
*****************************************************************************/

ParseGetFields (argc, argv, pdp, ftable)
register PARSE_DATA *pdp;
int argc;
char *argv[];
register PARSE_TABLE *ftable;
{
    register char *p;
    register int  sumflag;
    register int  done;
    register PARSE_TABLE *ptr;
    register int  matched;
    PARSE_TABLE *p_const;
    PARSE_TABLE *match;
    int   escape_used;
    int   mode;
    char *token;
    extern char *FieldGet ();
    extern char  NextChar ();

    mode    = FIELD_REPLACE;
    done    = FALSE;
    sumflag = 0;
    while (!done)
    {
	escape_used = FALSE;
	switch (ParseGetToken (pdp, &token))
	{
	    case TOKEN_ESCAPE:
		escape_used = TRUE;
	        /* Warning: this falls through. */
	    case TOKEN_DATA:
		switch (*token)
		{
		    case '0':
		        argv[argc]   = (char *) FIELD_REPLACE;
			argv[argc+1] = (char *) 0;
			return (argc + 2);
		    case '+':
		        mode = FIELD_ADD;
		        continue;
		    case '-':
		        mode = FIELD_SUBTRACT;
		        continue;
		}
		if (*token=='\0')
		{
		    ParseError (pdp, (char *)NULL, "Expecting a fieldlist (? for list)");
		    if (escape_used)
		        return (PARSE_IGNORE);
		    else
		        return (PARSE_ERROR);
		}
		p = token;
		while (*p!='|'&&*p != '\0')
		    p++;
		if (*p=='\0')
		    done = TRUE;
		*p++ = '\0';
	        matched = ParseTableFind (pdp, ftable, token, escape_used, &p_const);
	        if (matched!=PARSE_SUCCESS)
	            return (matched);
		sumflag |= ((int)p_const->p_arg1);
		if (done&&escape_used)
		{
		    ParseEscapeComplete (pdp, 
		        FieldGet (ftable, (int)p_const->p_arg1), FALSE, 0);
		    return (PARSE_UPDATE_LINE);
		}
		break;
	    case TOKEN_QUESTION:
		if (*token=='+'||*token == '-')
		    token++;
		match = NULL;
		for (ptr = ftable; p_arg0(ptr)!=0; ptr++)
		{
		    if (ParseMatch (p_arg0(ptr), token))
		    {
			ParsePrint ("%-20.20s %s\r\n", p_arg0(ptr), ptr->p_arg2);
			match = ptr;
		    }
		}
		if (match==NULL)
		{
		    ParseError (pdp, (char *)NULL, "Unable to find a field match");
		    return (PARSE_ERROR);
		}
		return (PARSE_IGNORE);

	    case TOKEN_EXHAUSTED:
		ParseError (pdp, (char *)NULL, "Expecting a fieldlist (? for list)");
		return (PARSE_IGNORE);

	    case TOKEN_OPTIONAL_DONE:
	        return (argc);

	    default:
		ParsePrint ("GetFlags (): internal error.\r\n");
	}
    }
    argv[argc]   = (char *) mode;
    argv[argc+1] = (char *) sumflag;
    return (argc + 2);
}

/*****************************************************************************
* TypeFields:  type out the fields symbolically.                             *
*                                                                            *
*   Arguments:  ftable,  a pointer to the field table.                       *
*               flag,    the flag word.                                      *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

TypeFields (ftable, flag)
PARSE_TABLE *ftable;
register int flag;
{
    register int i;
    register int bit;
    register int many;
    extern char *FieldGet ();

    if (flag==0)
    {
	ParsePrint ("NULL");
	return;
    }

    many = FALSE;
    for (i = 0; i < 16; i++)
    {
	bit = 1<<i;
	if (flag & bit)
	{
	    if (many)
		ParsePrint ("|");
	    many = TRUE;
	    ParsePrint ("%s", FieldGet (ftable, bit));
	}
    }
}

/*****************************************************************************
* FieldGet:  return the field name associated with a bit pattern.            *
*                                                                            *
*   Arguments:  ftable,  a pointer to the field table.                       *
*               bit,     the bit pattern.                                    *
*                                                                            *
*   Returns:    the name of this field.                                      *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

char *FieldGet (ftable, bit)
PARSE_TABLE *ftable;
int bit;
{
    register PARSE_TABLE *ptr;

    for (ptr = ftable; p_arg0(ptr)!=0; ptr++)
	if (((int)ptr->p_arg1)==bit)
	    return (p_arg0(ptr));

    return ("<<bad field>>");
}
