#ifndef lint
static char rcsid [] = "$RCSfile: par_hist.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

#include <stdext.h>
#include    "libparser_l.h"
/*
* par_history.c contains functions that implement the "logging" of
*           commands issued by the user to the parser.
*           The routines place the commands into a queue
*           (identical to the CMOS queue). Some routines
*           NextCommand and PreviousCommand walk along the queue
*           in response to the parser command keys like the BBN-UNIX B-Shell.
*
*   written by Brian Anderson  9/2/84
*   modified for no globals by Wolfgang Rupprecht 11/9/84
*
*/

/*
* CommandLog - is called by the ParseList, the routine that issues the call
*               to the function the user invoked from the command line
*               It copies the command line into a buffer that has a QELEM
*               before the null terminated command line.
*               Then it tacks it onto the end of the parsequeue
*
*/
CommandLog (pdp)
register    PARSE_DATA * pdp;
{
    extern char *stringcopy ();

    if (pdp -> LogFunction)
		/* I'm not sure what the log function needs -wr */

	(*pdp -> LogFunction) (pdp, pdp -> First); 

    *(pdp -> CurrentCommand) = stringcopy (pdp -> First);

    pdp -> CurrentCommand++;

    if (pdp -> CurrentCommand > pdp -> CommandBufferEnd)
	pdp -> CurrentCommand = pdp -> CommandBuffer;

    if (*(pdp -> CurrentCommand)!=NULL)
	Free (*(pdp -> CurrentCommand));

    *(pdp -> CurrentCommand) = NULL;
    pdp -> EarlierCommand = pdp -> CurrentCommand;
}

/*
* CopyEarlierCommand - is used to restore the command line pointers and 
*                       buffer area to that of the "CurrentCommand"
*                       QELEM of the global structure.
*                       The "CurrentCommand" can of course be null
*                       as one walks of either end of the QUEUE.
*                       Special handling for these cases is obvious.
*/

CopyEarlierCommand (pdp)
register    PARSE_DATA * pdp;
{
    register char  *s;

    s = *(pdp -> EarlierCommand);
    BeginningOfLine (pdp);
    KillForward (pdp);
    if (s!=NULL)
    {
	while (*s)
	    InsertChar (pdp, *s++);
    }
}

/*
* PreviousCommand - walks from the tail of the queue or uses the prev
*                   of the QELEM -----same direction though.
*
*/
PreviousCommand (pdp)
register    PARSE_DATA * pdp;
{
    pdp -> EarlierCommand--;

    if (pdp -> EarlierCommand < pdp -> CommandBuffer)
	pdp -> EarlierCommand = pdp -> CommandBufferEnd;

    if (*(pdp -> EarlierCommand)==NULL)
	pdp -> EarlierCommand = pdp -> CurrentCommand;

    CopyEarlierCommand (pdp);
}
/*
* NextCommand - walks the queue using qhead of the queue and 
*               next of the QELEM  again same direction
*
*/
NextCommand (pdp)
register    PARSE_DATA * pdp;
{

    if (pdp -> EarlierCommand==pdp -> CurrentCommand)
	return;

    pdp -> EarlierCommand++;

    if (pdp -> EarlierCommand > pdp -> CommandBufferEnd)
	pdp -> EarlierCommand = pdp -> CommandBuffer;

    if (*(pdp -> EarlierCommand)==NULL)
	pdp -> EarlierCommand = pdp -> CurrentCommand;

    CopyEarlierCommand (pdp);
}
