#ifndef lint
static char rcsid [] = "$RCSfile: par_lib.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
* File: par_lib.c                                                            *
*                                                                            *
*   Contents:       A library of useful parsing routines.                    *
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation        *
*   Created:        Mon Jan 23 18:12:57 1984                                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

#include <stdext.h>
#include "libparser_l.h"
#include <stdstring.h> /*common/include/global*/

/* Used by the yes/no macro.                                                */

DEFINE_TABLE (g_ParseYesNo)
    KEYWORD_SELECT (NULL)
	KEYWORD ("yes",     "affirmitive")
	    PUTARG (TRUE)
	END_KEYWORD
	KEYWORD ("no",      "negative")
	    PUTARG (FALSE)
	END_KEYWORD
    END_KEYWORD_SELECT
END_DEFINE_TABLE

/* Used by the on/off macro.                                                */

DEFINE_TABLE (g_ParseOnOff)
    KEYWORD_SELECT (NULL)
	KEYWORD ("on",      "turn on")
	    PUTARG (TRUE)
	END_KEYWORD
	KEYWORD ("off",     "turn off")
	    PUTARG (FALSE)
	END_KEYWORD
    END_KEYWORD_SELECT
END_DEFINE_TABLE

/*****************************************************************************
* ParseSetPrompt:  Set the parser prompt.                                    *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  pointer to the prompt string.                         *
*                                                                            *
*   Returns:     the argc value.                                             *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

/*ARGSUSED*/
int ParseSetPrompt (argc, argv, pdp, arg2)
PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
  ChangePrompt(arg2);
  return (argc);
}

/*****************************************************************************
* ParseSetTopLevel:  Set the top level table.                                *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  the table to set the top level to.                    *
*                                                                            *
*   Returns:     the argc value.                                             *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

/*ARGSUSED*/
int ParseSetTopLevel (argc, argv, pdp, arg2)
PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
  ChangeTopLevel((PARSE_TABLE *)arg2);
  return (argc);
}

/*****************************************************************************
* ParseGetDecimal:  get a decimal number.                                    *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseGetDecimal (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
    return (ParseConvertNumber (argc, argv, pdp, arg2, 10, "decimal"));
}

/*****************************************************************************
* ParseGetHex:  get a hex number.                                            *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseGetHex (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
    return (ParseConvertNumber (argc, argv, pdp, arg2, 16, "hex"));
}

/*****************************************************************************
* ParseGetOctal:  get an octal number                                        *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseGetOctal (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
    return (ParseConvertNumber (argc, argv, pdp, arg2, 8, "octal"));
}

/*****************************************************************************
* ParseGetFloat:  get a floating point number.                               *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseGetFloat (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
    return (ParseConvertFloat (argc, argv, pdp, arg2));
}

/*****************************************************************************
* ParseGetString:  get a string.                                             *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseGetString (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
    char *token;
    extern char *stringcopy ();

    switch (ParseGetToken (pdp, &token))
    {
	case TOKEN_DATA:
	    argv[argc] = stringcopy (token);
	    ParseMustFree (pdp, argc);
	    return (argc + 1);

	case TOKEN_ESCAPE:
	case TOKEN_QUESTION:
	    ParseError (pdp, arg2, "Expecting a string");
	    return (PARSE_IGNORE);

	case TOKEN_EXHAUSTED:
	    ParseError (pdp, arg2, "Expecting a string");
	    return (PARSE_ADDSPACE);

	case TOKEN_OPTIONAL_DONE:
	    return (argc);

	default:
	    ParsePrint ("ParseGetString (pdp): internal error.\r\n");
	return(ERROR);  /* for lint */
    }
}

/*****************************************************************************
* ParsePutArg:  put an argument in the input stream.                         *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

/*ARGSUSED*/
int ParsePutArg (argc, argv, nu_pdp, arg2)
PARSE_DATA *nu_pdp;        /* unused */
register int argc;
char *argv[];           
char *arg2;             
{
    argv[argc] = arg2;
    return (argc + 1);
}

/*****************************************************************************
* ParseOptional:  process the optional statement.                            *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  not used.                                             *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

/*ARGSUSED*/
int ParseOptional (argc, nu_argv, pdp, nu_arg2)
register PARSE_DATA *pdp;
register int argc;
char *nu_argv[];
char *nu_arg2;
{
    pdp->Optional = TRUE;
    return (argc);
}

/*****************************************************************************
* ParseConvertNumber:  get number of specified base.                         *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  the last argument in the command.                     *
*               base,  the radix to use.                                     *
*               bname, the name of the base.                                 *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseConvertNumber (argc, argv, pdp, arg2, base, bname)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
int base;
char *bname;
{
    register int   digit;
    register char *ptr;
    register int   num;
    int   neg;
    char *token;

    neg = FALSE;
    while (TRUE)
    {
	switch (ParseGetToken (pdp, &token))
	{
	    case TOKEN_DATA:
		if (*token=='-')
		    neg = TRUE;
		else if (*token!='+')
		{
		    num = 0;
		    for (ptr = token; *ptr!='\0'; ptr++)
		    {
		    	if (*ptr >= '0'&&*ptr <= '9')
		    	    digit = *ptr - '0';
		    	else if (*ptr >= 'a'&&*ptr <= 'f')
		    	    digit = *ptr - 'a' + 10;
		    	else if (*ptr >= 'A'&&*ptr <= 'F')
		    	    digit = *ptr - 'A' + 10;
		    	else
		    	{
		    	    ParseError (pdp, arg2, "'%s' is an illegal %s number", token, bname);
		    	    return (PARSE_ERROR);
			}

		    	if (digit >= base)
		    	{
		    	    ParseError (pdp, arg2, "'%s' is an illegal %s number", token, bname);
		    	    return (PARSE_ERROR);
			}

		    	num = (num*base) + digit;
		    }
		    if (neg)
			num = -num;
		    argv[argc] = (char *) num;
		    return (argc + 1);
		}
		break;

	    case TOKEN_EXHAUSTED:
		ParseError (pdp, arg2, "Expecting a %s number", bname);
		return (PARSE_ADDSPACE);

	    case TOKEN_ESCAPE:
	    case TOKEN_QUESTION:
		ParseError (pdp, arg2, "Expecting a %s number", bname);
		return (PARSE_IGNORE);

	    case TOKEN_OPTIONAL_DONE:
		return (argc);

	    default:
		ParsePrint ("ParseConvertNumber (pdp): internal error.\r\n");
	}
    }
}


/*****************************************************************************
* ParseConvertFloat:  get floating point number.                             *
*                                                                            *
*   Arguments:  argc,  the current state of argc.                            *
*               argv,  the current state of argv.                            *
*               arg2,  the last argument in the command.                     *
*                                                                            *
*   Returns:    a parse return status or the new argc value.                 *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseConvertFloat (argc, argv, pdp, arg2)
register PARSE_DATA *pdp;
int argc;
char *argv[];
char *arg2;
{
    register char *ptr;
    float   num;
    int   neg, dpoint, exponent, exp_started;
    char *token;
    char	numstr[128];

    dpoint = FALSE;
    exponent = FALSE;
    exp_started = FALSE;
    neg = FALSE;
    while (TRUE)
    {
	switch (ParseGetToken (pdp, &token))
	{
	    case TOKEN_DATA:
		if (*token=='-' && !exponent)
		    neg = TRUE;
		else if (*token!='+' && !exponent)
		{
		    num = 0;
		    for (ptr = token; *ptr!='\0'; ptr++)
		    {
			if ((*ptr == 'e' || *ptr == 'E') && !exponent)
			  {
			    exponent = TRUE;
			    dpoint = TRUE;	/* No decimal pts in exponent*/
			  }
			else if ((*ptr == '+' || *ptr == '-') && exponent &&
				!exp_started)
			  {
			    strcpy(numstr, token);
			    goto bottom;
			  }
			else if (*ptr == '.' && !dpoint)
			    dpoint = TRUE;
		    	else if (*ptr < '0' || *ptr > '9')
		    	  {
		    	    ParseError (pdp, arg2, "'%s' is an illegal floating point number", token);
		    	    return (PARSE_ERROR);
			  }
			else if (exponent)
			  exp_started = TRUE;
		    }
		    sscanf(token,"%f",&num);
		    if (neg)
			num = -num;
		    bcopy((char *)&num, (char *)&argv[argc], sizeof(num));
		    return (argc + 1);
		}
		else if (exponent)
		{
		    for (ptr = token; *ptr!='\0'; ptr++)
		    {
			if (*ptr < '0' || *ptr > '9')
		    	{
		    	    ParseError (pdp, arg2, "'%s' is an illegal floating point number", token);
		    	    return (PARSE_ERROR);
			}
		    }
		    strcat(numstr,token);
		    sscanf(numstr,"%f",&num);
		    if (neg)
			num = -num;
		    bcopy((char *)&num, (char *)&argv[argc], sizeof(num));
		    return (argc + 1);
		}
bottom:		break;

	    case TOKEN_EXHAUSTED:
		ParseError (pdp, arg2, "Expecting a floating point number");
		return (PARSE_ADDSPACE);

	    case TOKEN_ESCAPE:
	    case TOKEN_QUESTION:
		ParseError (pdp, arg2, "Expecting a floating point number");
		return (PARSE_IGNORE);

	    case TOKEN_OPTIONAL_DONE:
		return (argc);

	    default:
		ParsePrint ("ParseConvertFloat (pdp): internal error.\r\n");
	}
    }
}
