#ifndef lint
static char rcsid [] = "$RCSfile: par_util.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
* File: par_util.c                                                           *
*                                                                            *
*   Contents:       This contains important parser auxillary routines.       *
*   Author:         Steve Geyer at Loral Advanced Distributed Simulation        *
*   Created:        Thu Jan 26 09:04:52 1984                                 *
*                                                                            *
*   Remarks:	    Paul J. Metzger	01 Febrauary 1994		     *
*		    Added parser_clear_screen() and parser_help().	     *
*                                                                            *
*****************************************************************************/


#include "stdio.h"
#include <stdext.h>
#include <stdalloc.h>
#include "libparser_l.h"
	
extern char NextChar ();

/*****************************************************************************
* DoCommand:  process a user command.                                        *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

DoCommand (pdp)
register PARSE_DATA *pdp;
{
    ParsePrint ("\r\n");
    SetStopPoint (pdp, FALSE);
    pdp->TokenType = TOKEN_DATA;
    pdp->Optional  = FALSE;
    ParseInput (pdp, pdp->TopLevel);
}

/*****************************************************************************
* DoHelp:  process a help command.                                           *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

DoHelp (pdp)
register PARSE_DATA *pdp;
{
    InsertChar (pdp, '?');
    UpdateLine (pdp);
    Delete1Backward (pdp);
    ParsePrint ("\r\n");
    SetStopPoint (pdp, TRUE);
    pdp->TokenType = TOKEN_QUESTION;
    pdp->Optional  = FALSE;
    ParseInput (pdp, pdp->TopLevel);
}

/*****************************************************************************
* DoEscape:  process an escape command.                                      *
*                                                                            *
*   Arguments:  none.                                                        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

DoEscape (pdp)
register PARSE_DATA *pdp;
{
    SetStopPoint (pdp, TRUE);
    pdp->TokenType = TOKEN_ESCAPE;
    pdp->Optional  = FALSE;
    ParseInput (pdp, pdp->TopLevel);
}

/*****************************************************************************
* ParseMustFree:  this data item must be freed when done.                    *
*                                                                            *
*   Arguments:  index,  the argument index.                                  *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*       This call is used by parsing functions to inform the parse system    *
*       that an 'allocated' item is at this argv index.  When the parse is   *
*       done, all allocated things will be freed.                            *
*                                                                            *
*****************************************************************************/

ParseMustFree (pdp, index)
register PARSE_DATA *pdp;
int index;
{
    pdp->ArgFree[index] = TRUE;
}

/*****************************************************************************
* ParseInSet:  return TRUE if ch is in set.                                  *
*                                                                            *
*   Arguments:  ch,   the character to use.                                  *
*               set,  the set to use.                                        *
*                                                                            *
*   Returns:    TRUE if ch is in set, otherwise FALSE.                       *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseInSet (ch, set)
register char ch;
register char *set;
{
    while (*set!='\0'&&*set != ch)
	set++;
    return (*set==ch);
}

/*****************************************************************************
* ParseGetToken:  read in the next user token.                               *
*                                                                            *
*   Arguments:  tok,  the address of the user token pointer.                 *
*                                                                            *
*   Returns:    a token return value.                                        *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseGetToken (pdp, tok)
register PARSE_DATA *pdp;
register char **tok;
{
    register char ch;
    register char *p;

    MarkCursor (pdp);

    p = pdp->ParseToken;
    *p = '\0';
    *tok = p;

    ch = NextChar (pdp);
    while (ch!='\0'&&ParseInSet (ch, " \t"))
	ch = NextChar (pdp);

    if (ch=='\0')
    {
	if (pdp->TokenType==TOKEN_DATA)
	{
	    if (pdp->Optional)
	        return (TOKEN_OPTIONAL_DONE);
	    else
	        return (TOKEN_EXHAUSTED);
	}
	else
	    return (pdp->TokenType);
    }

    if (ch=='"')
    {
	ch = NextChar (pdp);
	while (ch!='\0'&&!ParseInSet (ch, "\"?\033"))
	{
	    if (ch=='\\')
		ch = NextChar (pdp);
	    if (ch=='\0')     /* Catches \<NULL>. */
		break;
	    *p++ = ch;
	    ch = NextChar (pdp);
	}
    }
    else
    {
        while (ch!='\0'&&!ParseInSet (ch, " ?\033"))
        {
	    if (ch=='\\')
		ch = NextChar (pdp);
	    if (ch=='\0')     /* Catches \<NULL>. */
		break;
            *p++ = ch;
            if (ParseInSet (ch, "|+-"))
                break;
            ch = NextChar (pdp);
        }
    }
    *p = '\0';

    if (ch=='\0')
	return (pdp->TokenType);
    else
	return (TOKEN_DATA);
}

/*****************************************************************************
* ParseError:  process a error message.                                      *
*                                                                            *
*   Arguments:  arg2,     	the user's error message.                        *
*               msg1-msg5,  the error message (arguments to printf).         *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

/*VARARGS3*/

ParseError (pdp, arg2, msg1, msg2, msg3, msg4, msg5)
register PARSE_DATA *pdp;
char *arg2;
char *msg1, *msg2, *msg3, *msg4, *msg5;
{
    if (pdp->TokenType==TOKEN_ESCAPE)
		ParsePrint ("\r\n");

    if (pdp->Optional)
        ParsePrint ("[Optional: ");
    else
        ParsePrint ("[");

    if (arg2==NULL)
		printf(msg1, msg2, msg3, msg4, msg5);
    else
    {
		printf("%s (", arg2);
		printf(msg1, msg2, msg3, msg4, msg5);
        ParsePrint (")");
    }
    ParsePrint ("]\r\n");
}

/*****************************************************************************
* ParseMessage:  print user message.                                         *
*                                                                            *
*   Arguments:  message,  the user message.                                  *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

ParseMessage (pdp, message)
register PARSE_DATA *pdp;
char *message;
{
    if (pdp->TokenType==TOKEN_ESCAPE)
	ParsePrint ("\r\n");
/* $$$    ParsePrint ("[%r", &message); */
 printf("[%s", message);
    ParsePrint ("]\r\n");
}

/*****************************************************************************
* ParseFindEndList:  find the end of a parse table block.                    *
*                                                                            *
*   Arguments:  first,      the first entry in the block.                    *
*               end,        one entry beyond the last to check.              *
*               begin_tok,  the block entry token.                           *
*               end_tok,    the block end token.                             *
*                                                                            *
*   Returns:    a parse table entry pointer.                                 *
*                                                                            *
*   Remarks:                                                                 *
*       This routine will skip sub-blocks.                                   *
*                                                                            *
*****************************************************************************/

PARSE_TABLE *ParseFindEndList (first, end, begin_tok, end_tok)
register PARSE_TABLE *first;
register PARSE_TABLE *end;
register int begin_tok;
register int end_tok;
{
    int level;

    level = 0;
    if (first->p_command!=begin_tok)
	ParsePrint ("ParseFindEndList (): error in parse tables.\r\n");
    first++;
    while (first < end)
    {
	if (first->p_command==begin_tok)
	    level++;
	else if (first->p_command==end_tok)
	{
	    --level;
	    if (level < 0)
		return (first);
	}
	first++;
    }
    ParsePrint ("ParseFindEndList (): never found the end of list (%d %d).\r\n",
	begin_tok, end_tok);
    return((PARSE_TABLE *) NULL);  /* for lint */
}

/*****************************************************************************
* ParseMatch:  return TRUE if the token matches the command.                 *
*                                                                            *
*   Arguments:  command,  the command string.                                *
*               token,    the token to match.                                *
*                                                                            *
*   Returns:    TRUE if it matches, otherwise FALSE.                         *
*                                                                            *
*   Remarks:                                                                 *
*       This routine also matches substrings.                                *
*                                                                            *
*****************************************************************************/

int ParseMatch (command, token)
register char *command;
register char *token;
{
    char uc ();

    while (*command!='\0' &&
	   *token  !='\0' &&
	   uc (*command)==uc (*token))
    {
	command++;
	token++;
    }
    if (*token=='\0')
	return (TRUE);

    return (FALSE);
}

/*****************************************************************************
* ParseTableFind:  look for a token in the table.                            *
*                                                                            *
*   Arguments:  table,        pointer to the table.                          *
*               token,        the input token.                               *
*               escape_used,  TRUE if command called with escape.            *
*               entry,        where to put matching entry.                   *
*                                                                            *
*   Returns:    parser error code.                                           *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

int ParseTableFind (pdp, table, token, escape_used, entry)
register PARSE_DATA *pdp;
PARSE_TABLE *table;
char *token;
int   escape_used;
PARSE_TABLE **entry;
{
    register PARSE_TABLE *ptr;
    register PARSE_TABLE *match;
    int			 token_len;
    int			 unique = 0;
    int			 match_len;

    token_len = strlen (token);

    match = NULL;
    for (ptr = table; p_arg0(ptr)!=0; ptr++)
    {
	if (ParseMatch (p_arg0(ptr), token))
	{
	  match_len = strlen (p_arg0(ptr));
	  if (match_len == token_len)
	    {
	      match = ptr;
	      unique = 1;
	      break;
	    }

	    if (match==NULL)
	      {
	        match = ptr;
		unique = 1;
	      }
	    else
	        unique = 0;
	}
    }

    if (match==NULL)
    {
	ParseError (pdp, (char *) NULL,
		    "'%s' is not a legal name (? for list)", token);
	return (PARSE_ERROR);
    }

    if (!unique)
      {
	ParseError (pdp, (char *) NULL, 
		    "'%s' is not a unique name", token);
	if (escape_used)
	  return (PARSE_IGNORE);
	else
	  return (PARSE_ERROR);
      }

    *entry = match;
    return (PARSE_SUCCESS);
}

/*****************************************************************************
* ParseEscapeComplete:  update token in input string on escape.              *
*                                                                            *
*   Arguments:  token,      the token to update.                             *
*               addspace,   TRUE if a trailing space should be added.        *
*                                                                            *
*   Returns:    nothing.                                                     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

ParseEscapeComplete (pdp, token, addspace, nchars)
register PARSE_DATA *pdp;
register char *token;
int addspace;
int nchars;
{
    RestoreCursor (pdp);
    if (nchars == 0)
	nchars = 100000;
    while (*token!='\0'&&*token==NextChar (pdp))
    {
        token++;
	nchars--;
    }
    while (*token && nchars > 0)
    {
        InsertChar (pdp, *token++);
	nchars--;
    }
    if (addspace)
	InsertChar (pdp, ' ');
}

/*****************************************************************************
* uc:  return an uppercase character.                                        *
*                                                                            *
*   Arguments:  ch,  the input character.                                    *
*                                                                            *
*   Returns:    the uppercased character.                                    *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

char uc (ch)
register char ch;
{
    if (ch >= 'a'&&ch <= 'z')
	return (ch - 'a' + 'A');
    return (ch);
}

/*****************************************************************************
* parser_clear_screen:  clears the terminal screen			     *
*                                                                            *
*   Arguments:  none							     *
*                                                                            *
*   Returns:    nothing							     *
*                                                                            *
*   Remarks:	Should be called from within an application KEYWORD_SELECT   *
*		section as follows:					     *
*                                                                            *
*		    KEYWORD ("clear", "- Clear terminal screen")	     *
*			CALL (parser_clear_screen)			     *
*		    END_KEYWORD						     *
*                                                                            *
*****************************************************************************/

void parser_clear_screen ()
{
    fprintf (stdout, "%c[H%c[J", '\033', '\033');
    fflush (stdout);
}

/*****************************************************************************
* parser_help:  Display somewhat verbose help on 
*                                                                            *
*   Arguments:  none							     *
*                                                                            *
*   Returns:    nothing							     *
*                                                                            *
*   Remarks:	Should be called from within an application KEYWORD_SELECT   *
*		section as follows:					     *
*                                                                            *
*		    KEYWORD ("help", "- Display help on using the parser")   *
*			CALL (parser_help)				     *
*		    END_KEYWORD						     *
*                                                                            *
*****************************************************************************/

void parser_help ()
{
    parser_clear_screen();

    fprintf (stdout,
	"All features are accessed by pressing the control key with one\n");
    fprintf (stdout,
	"of the special function keys.  The available functions are:\n\n");

    fprintf (stdout, "\t?      - show list of available commands\n");
    fprintf (stdout, "\t<cr>   - execute command\n");
    fprintf (stdout, "\t<tab>  - complete current word\n");
    fprintf (stdout, "\t^n     - next command on history stack\n");
    fprintf (stdout, "\t^p     - previous command on history stack\n");
    fprintf (stdout, "\t^a     - beginning of line\n");
    fprintf (stdout, "\t^e     - end of line\n");
    fprintf (stdout, "\t^f     - forwards one character\n");
    fprintf (stdout, "\t^b     - backwards one character\n");
    fprintf (stdout, "\t^l     - left one word\n");
    fprintf (stdout, "\t^r     - right one word\n");
    fprintf (stdout, "\t^d     - delete one character forward\n");
    fprintf (stdout, "\t^h     - delete one character backward\n");
    fprintf (stdout, "\t^g     - erase rightmost segment of current word\n");
    fprintf (stdout, "\t^w     - erase leftmost segment of current word\n");
    fprintf (stdout, "\t^k     - erase remainder of line\n");
    fprintf (stdout, "\t^u     - erase entire line\n");
    fprintf (stdout, "\t^i     - escape\n");
    fprintf (stdout,
	"\t^x     - show key bindings (short version of this list)\n\n");
}

/*****************************************************************************
* ShowBindings:  Display short summary of key bindings			     *
*                                                                            *
*   Arguments:  pdp (pointer to PARSE_DATA)				     *
*                                                                            *
*   Returns:    nothing							     *
*                                                                            *
*   Remarks:                                                                 *
*                                                                            *
*****************************************************************************/

ShowBindings(pdp)
register PARSE_DATA *pdp;
{

  fprintf(stdout,"\nKey          Operation        Key          Operation\n");
  fprintf(stdout,"---          ---------          ---          ---------\n");
  fprintf(stdout," ?           Help               TAB          Complete\n");
  fprintf(stdout,"ESC          Complete           ^I           Complete\n");
  fprintf(stdout,"^F           Move +1            ^B           Move -1\n");
  fprintf(stdout,"^A           BOL                ^E           EOL\n");
  fprintf(stdout,"^G           Del +Word          ^W           Del -Word\n");
  fprintf(stdout,"^D           Del +1             ^H           Del -1\n");
  fprintf(stdout,"^L           Move -Word         ^R           Move +Word\n");
  fprintf(stdout,"^P           Prev Cmd           ^N           Next Cmd\n");
  fprintf(stdout,"^U           Kill Line          ^K           Kill Rest of Line\n");
  
  pdp->TokenType = TOKEN_DATA;
  pdp->Optional  = FALSE;
  ParseInput (pdp, pdp->TopLevel);
  UpdateLine(pdp);
  EndOfLine(pdp);
}


/*
 * Add the entries in src_table to the entries in the dst_table and
 * sort them by the appropriate tag value.
 *
 * Note that dst_table must have been allocated off of the heap (malloc)
 * and not the stack since it will be freed.
 *
 */
void ParseAddTable (dst_table_p, src_table)
    PARSE_TABLE **dst_table_p;
    PARSE_TABLE *src_table;
{
    PARSE_TABLE *dst_table = *dst_table_p;
    PARSE_TABLE *new_dst_table;
    int dst_table_size;
    int src_table_size;
    int i;

    for (i=0; dst_table [i].p_command != 0; i++);

    dst_table_size = i;

    for (i=0; src_table [i].p_command != 0; i++);

    src_table_size = i;

    new_dst_table =
      (PARSE_TABLE *)STDALLOC (sizeof (PARSE_TABLE) *
			       (dst_table_size + src_table_size + 1));

    bcopy (dst_table, new_dst_table, sizeof (PARSE_TABLE) * dst_table_size);

    if (dst_table [0].p_command == P_KEYWORD_SELECT)
    {
	bcopy (src_table, new_dst_table + dst_table_size - 1,
	       sizeof (PARSE_TABLE) * src_table_size);
	new_dst_table [dst_table_size + src_table_size - 1].p_command =
	  P_END_KEYWORD_SELECT;
    }
    else
      bcopy (src_table, new_dst_table + dst_table_size,
	     sizeof (PARSE_TABLE) * src_table_size);

    bzero (&new_dst_table [dst_table_size + src_table_size],
	   sizeof (PARSE_TABLE));

    free (dst_table);

    *dst_table_p = new_dst_table;
}

