/*
 * $RCSfile: libpbtab.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpbtab.h                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libpbtab                           *
*       Created: Mon Apr 13 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBPBTAB_H_INCLUDED
#define _LIBPBTAB_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <libvtab.h>
#include <stdext.h>

typedef void (*PBT_EXPAND_FUNC) (/* new_cell */);

/**
 ** Position-based table parametric data structure.
 ** Note that although there is no parametric data for this model,
 ** every vehicle sub-class must define such a structure.
 **/

/* Data file entry should read:
 *
 * (SM_PBTab)
 */

typedef struct pbtab_parametric_data
{
    int32 dummy;
} PBTAB_PARAMETRIC_DATA;

/** 
 ** Functions relating to global libpbtab operation
 **
 **/

/* pbt_init:
 *
 * Initializes the position based table library.  Call this before any
 * other libpbtab function.  Note that an application also has to call 
 * pbt_class_init before using libpbt.
 * 
 * This function does not allocate the table and offmap data.
 * To generate a pbtable to use, the application must invoke
 * pbt_alloc_info.  Before using that table, the application should
 * also call pbt_set_size.
 */
extern void 
pbt_init ( PBT_EXPAND_FUNC expand );

/* pbt_select_table:
 *
 * Selects which position based vehicle table is currently active.
 * There is no need to call this if using only one vehicle table.
 * This should be called whenever vtab_select_table is called.
 *
 * For GCS use: this function takes in a cell_id to select the table.
 */
extern void 
pbt_select_table ( int32 which_table );

/* pbt_set_size:
 *
 * Sets the extents of the position based vehicle table.  These should
 * probably be the same as the extents of the terrain database.  Note
 * that setting the size of the database is a non-trivial operation,
 * and is not recommended during real-time execution (in other words,
 * this is intended to be done during program initialization).
 *
 * Vehicles which fall outside the extents set here are not guaranteed
 * to appear in the position based vehicle table (although they are
 * not guaranteed to NOT appear, either).
 *
 * Call this function before adding any vehicles to the vehicle table.
 *
 * For gcs, the which_table index is simply the associated cell_id for
 * the table we are trying to select.
 */
extern void 
pbt_set_size ( 
		 int32 min_x, 
		 int32 min_y, 
		 int32 max_x, 
		 int32 max_y 
		 );

/* pbt_get_vehicles:
 *
 * Adds each vehicle which is within 'range' meters of the specified
 * x, y location to the passed vehicle list.
 *
 * For GCS use: pbt_select_table should be called prior to usage.
 */
extern void 
pbt_get_vehicles ( 
		     int32 x,
		     int32 y,
		     int32 range,
		     VTAB_LIST list
		     );

/**
 ** Functions relating to per-vehicle PBT class information.
 **
 **/


/* pbt_destroy:
 *
 * Frees the PBT class information for a vehicle, and removes it from
 * the position based table.
 */
extern void 
pbt_destroy ( 
		int32 vehicle_id, 
		int32 is_migration
		);

/* pbt_set_location:
 *
 * Updates the location of the vehicle in the position based table.
 *
 * pbt_select_table should be invoked first.
 */
extern void 
pbt_set_location ( 
		     int32 vehicle_id,
		     int32 x, 
		     int32 y
		     );

/* pbt_alloc_info:
 *
 * initializes all pbt_root and offmap information needed for one pbtable.
 */
extern ADDRESS 
pbt_alloc_info (void);

/* pbt_dealloc_info:
 *
 * frees pbt_info structure
 */
extern void 
pbt_dealloc_info ( ADDRESS info );


#endif /*_LIBPBTAB_H_INCLUDED*/
