/*
 * $RCSfile: libpbtab_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpbtab_local.h                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libpbtab                          *
*       Created: Mon Apr 13 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpbtab.h"
#include <libvtab.h>
#include <libclass.h>
#include <libdisdbobj.h>

/* 100 meter maximum resolution */
#define PBT_RESOLUTION 100

/* define the default size of the db */
#define PBT_DEFAULT_SIZE 100000

/* The table id for cells that do not have a table */
#define PBT_NO_TABLE_ID -1

extern int32 pbt_current_id;


/* Child node array indecies */
#define PBT_SW 0
#define PBT_SE 1
#define PBT_NW 2
#define PBT_NE 3

/* Special code numbers */
#define PBT_CODE_UNINITTED 0x00000001
#define PBT_CODE_OFF_MAP   0x00000002

typedef struct pbt_quad_node
{
    int32 x, y;
    int32 size;
    int32 has_vehicles;
    /* This is 30 for the root node, and decreases by two for each
     * level you go down.
     */
    uint32 child_code_shift;

    union
    {
	VTAB_LIST list;
	struct pbt_quad_node *children;
    } u;
} PBT_QUAD_NODE;

/* Per-vehicle PBT class information */
typedef struct pbt_vars
{
    PBT_QUAD_NODE *node;
    int32 x, y;
    uint32 code_number;
    int32 vehicle_id;
    struct pbt_vars *next;
} PBT_VARS;

/* Private functions */
extern void pbt_add_vehicle(PBT_QUAD_NODE *node, int32 id, uint32 code,
			    PBT_VARS *pbt);
extern void pbt_remove_vehicle(PBT_QUAD_NODE *node, int32 id);
extern int32  pbt_generate_code_number(int32 x, int32 y);
extern void pbt_rebuild(void);


extern void pbt_init_subclass(void);
extern void pbt_create(int32 vehicle_id, ADDRESS params,
		       ADDRESS unit_entry, char *name_symbol);
