#ifndef lint
static char rcsid [] = "$RCSfile: pbt_class.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pbt_class.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Manages class information attached to each vehicle        *
*       Created: Mon Apr 13 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpbtab_local.h"
#include <libclass.h>
#include <libreader.h>
#include <libparmgr.h>
#include <p_safmodels.h>
#include <stdext.h> /* common/include/global */
#include <stdalloc.h> /* common/include/global */
#include <stdio.h>
#include <libctdb.h>
#include <libgcs.h>

extern PBT_EXPAND_FUNC pbt_expand;
static int32 pbt_user_data_handle;

static void pbt_print(CLASS_USER_DATA_TYPE vars);

static ADDRESS parser(READER_UNION *libreader_format,
				     uint32 model_number, int32 *size);
static void changer(
    int32   vehicle_id,
    PBTAB_PARAMETRIC_DATA *new_params);

static void pbt_add_remote_subclass(int32 vehicle_id, ADDRESS user_data)
{
    uint16 vtab_type = vtab_vehicle_type(vehicle_id);

    /* This is a remote vehicle being added to the system.
     * Unlike locals, parmgr won't be adding subclasses to this object
     * based on parameters.  Intead, this library (libpbtab) always
     * knows to add itself to these particular objects.
     */
    if ((vtab_type != VTAB_REMOTE_STEALTH) &&
	(vtab_type != VTAB_LOCAL_STEALTH))
      pbt_create(vehicle_id, (ADDRESS) NULL, (ADDRESS) NULL, "");
}

static void pbt_remove_remote_subclass(int32 vehicle_id, int32 is_migration,
				       ADDRESS user_data)
{
    uint16 vtab_type = vtab_vehicle_type(vehicle_id);

    /* This is a remote vehicle being deleted from the system.
     * Unlike locals, parmgr won't be deleting subclasses from this object
     * based on parameters.  Intead, this library (libpbtab) always
     * knows to delete itself from these particular objects.
     */
    if (VTAB_TYPES_MATCH(vtab_type,VTAB_REMOTE))
      pbt_destroy(vehicle_id, is_migration);
}

void pbt_init_subclass(void)
{
    pbt_user_data_handle =
      class_reserve_user_data(disdbobj_class, "pbtab", pbt_print);

    pm_register_parm_fcns(SM_PBTab, parser, changer, pbt_create, pbt_destroy);

    /*
     * Register callbacks to create this subclass when remote vehicles
     * are created.
     */
    callback_register_handler(disdbobj_create_event, 
			      pbt_add_remote_subclass, NULL);
    callback_register_handler(disdbobj_destroy_event, 
			      pbt_remove_remote_subclass, NULL);
}

static void pbt_print(CLASS_USER_DATA_TYPE vars)
{
    PBT_VARS *pbt = (PBT_VARS *)vars;

    printf("x: %d\ny: %d\ncode: 0x%x\n", pbt->x, pbt->y, pbt->code_number);
    if (pbt->node)
      printf("node: 0x%x {x=%d, y=%d, size=%d, list=0x%x}\n",
	     pbt->node, pbt->node->x, pbt->node->y, pbt->node->size,
	     pbt->node->u.list);
    else
      printf("node: <nil>\n");
}

/*ARGSUSED*/
void pbt_create(int32 vehicle_id, ADDRESS params, ADDRESS unit_entry,
		char *name_symbol)
{
    PBT_VARS *pbt = (PBT_VARS *)STDALLOC(sizeof(PBT_VARS));

    bzero(pbt, sizeof(PBT_VARS));
    pbt->code_number = PBT_CODE_UNINITTED;
    pbt->vehicle_id = vehicle_id;

    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			pbt_user_data_handle, (CLASS_USER_DATA_TYPE)pbt);
}

void pbt_destroy(int32 vehicle_id, int32 is_migration)
{
    PBT_VARS *prev = NULL;
    PBT_VARS *pbt = (PBT_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id), pbt_user_data_handle);

    if (!pbt) /* Passive error detection */
      return;

    /* If we have multiple vehicle ids per safobj, we need to find the
     * appropriate slot in the pbt linked list.
     */
    while (pbt->vehicle_id != vehicle_id)
    {
	prev = pbt;
	pbt = pbt->next;
	if (!pbt)
	  return;
    }
    /* Now pbt is the correct slot, and prev is the previous slot, if any. */
    
    /* Remove vehicle from pbt. */
    pbt_remove_vehicle(pbt->node, vehicle_id);

    /* If there is a slot previous to this one, point its next to our next.
     * If we're the first one in the list, set the user data to our next.
     * In either case if we're the last one in the list, the appropriate
     * pointer will be set to zero since our next is zero.
     */
    if (prev)
      prev->next = pbt->next;
    else
      class_set_user_data((CLASS_USER_DATA_TYPE) vtab_get_vehicle(vehicle_id),
			  pbt_user_data_handle,
			  (CLASS_USER_DATA_TYPE)pbt->next);

    /* Free this slot */
    STDDEALLOC(pbt);
}    

static ADDRESS parser(READER_UNION *libreader_format,
				     uint32 model_number, int32 *size)
{
    READER_UNION *my_entry;
    PBTAB_PARAMETRIC_DATA *result;

    my_entry = reader_find_tag_int(model_number,
				   libreader_format,
				   READER_TAGGED,
				   READER_NO_ERRORS);

    if (!my_entry)
      return NULL;

    result =
      (PBTAB_PARAMETRIC_DATA *)STDALLOC(sizeof(PBTAB_PARAMETRIC_DATA));

    *size = sizeof(PBTAB_PARAMETRIC_DATA);    return (ADDRESS) result;

}

/*ARGSUSED*/
static void changer(
    int32   vehicle_id,
    PBTAB_PARAMETRIC_DATA *new_params)
{
    /* Nop */;
}

void 
pbt_set_location(int32 vehicle_id, 
		 int32 x, 
		 int32 y)
{
    PBT_VARS *pbt = (PBT_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id), pbt_user_data_handle);
    int32 old_code;
    int32 changed;
    PBT_VARS *new_pbt;

    if (!pbt) /* Passive error detection */
      return;

    if (!ctdb_point_on_database((CTDB *) gcs_get_tdb(pbt_current_id,FALSE), 
				(float64) x,
				(float64) y))
    {
	int32 new_cell, old_cell;
	float64 x1, y1;

	old_cell = pbt_current_id;
	gcs_gcs2e_to_gcs(pbt_current_id, (float64)x, (float64)y,
			 &new_cell, &x1, &y1);
	
	if (new_cell != pbt_current_id)
	{
	    pbt_destroy(vehicle_id, (int32) 0);
	    pbt_select_table(new_cell);

	    if (!pbt_current_id == PBT_NO_TABLE_ID)
	    {
		pbt_expand(new_cell);
		pbt_select_table(new_cell);
	    }

	    if (pbt_generate_code_number(x, y) == PBT_CODE_OFF_MAP)
	      printf("Error!  Location %d %d %d does not fall within any pbtable\n",
		     x, y, pbt_current_id);
      
	    pbt_set_location(vehicle_id, (int32)x1, (int32)y1);
	    pbt_select_table(old_cell);
	    return;
	}
	/* If we coord_convert, and we have the same cell back, we must
	 * be in single-cell or SIMNET mode.  In that case, the vehicle
	 * will be placed in the offmap.
	 */
    }

    /* Find the slot in the list for this vehicle */
    /* If there's only one vehicle in this safobj, this will fall through */
    while (vehicle_id != pbt->vehicle_id)
    {
	if (!pbt->next)
	{
	    new_pbt = (PBT_VARS *)STDALLOC(sizeof(PBT_VARS));
	    bzero(new_pbt, sizeof(PBT_VARS));
	    new_pbt->code_number = PBT_CODE_UNINITTED;
	    new_pbt->vehicle_id = vehicle_id;
	    pbt->next = new_pbt;
	}
	pbt = pbt->next;
    }
	

    /* Filter out non-motion early */
    if ((pbt->code_number != PBT_CODE_UNINITTED) &&
	(x == pbt->x) && (y == pbt->y))
      return;


    old_code = pbt->code_number;

    pbt->x = x;
    pbt->y = y;
    pbt->code_number = pbt_generate_code_number(x, y);

    /* See if we changed nodes.
     */

    /* Is this the first time we've been placed in the tree? */
    if (old_code == PBT_CODE_UNINITTED)
      changed = 1;
    /* If we are not in the tree, see if we should move in */
    else if (!pbt->node)
      changed = (pbt->code_number != PBT_CODE_OFF_MAP);
    /* If we should not be in the tree, see if we need to move out */
    else if (pbt->code_number == PBT_CODE_OFF_MAP)
      changed = (old_code != PBT_CODE_OFF_MAP);
    /* See if we are in a different node, and if we are, see if the
     * difference is significant at this level of the quad tree.
     */
    else
      changed = (old_code != pbt->code_number) &&
	((pbt->code_number >> (pbt->node->child_code_shift + 2)) !=
	 (old_code >> (pbt->node->child_code_shift + 2)));

    if (changed)
    {
	pbt_remove_vehicle(pbt->node, vehicle_id);
	pbt->node = (PBT_QUAD_NODE *)0;

	pbt_add_vehicle((PBT_QUAD_NODE *)0,
			vehicle_id, pbt->code_number, pbt);
    }
}

void pbt_rebuild(void)
{
    int32 id;
    VTAB_USER_DATA_TYPE user_data;
    PBT_VARS *pbt;

    /* This function runs through the master vehicle table and
     * puts everybody into the table where he belongs.  It assumes
     * the old table is trashed, so nodes are zero'd out.
     */

    for (id = vtab_get_first(vtab_vehicle_table, VTAB_ALL, &user_data);
	 id;
	 id = vtab_get_next(vtab_vehicle_table, VTAB_ALL, &user_data))
    {
	pbt = (PBT_VARS *)class_get_user_data(user_data,pbt_user_data_handle);
	pbt->node = (PBT_QUAD_NODE *)0;
	pbt->code_number = PBT_CODE_UNINITTED;

	pbt_set_location(id, pbt->x, pbt->y);
    }
}

