#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libpbtab                                 *
*       Created: Mon Apr 13 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpbtab.h"
#include <libvtab.h>
#include <libclass.h>
#include <basic.h>

#define NVEH 1000
#define NITS 1000
#define MAXX 1000
#define MAXY 1000
#define EPSI 100

CLASS_PTR vehicle_class;

void pm_register_parm_fcns() { /*Stub*/; }
void reader_find_tag() { /*Stub*/; }

main()
{
    int32 i, j, k;
    int32 ids[NVEH];
    int32 locations[NVEH][2];
    VehicleID simnet_id;
    CLASS_USER_DATA_TYPE user_data;
    VTAB_LIST mine, his;
    int32 x, y, r;
    double dx, dy, r2;

    /* Initialize everything (make the database a little smaller than
     * the range of locations, so we can test off-data as well).
     */
    vtab_init(1, 1, 1);
    pbt_init(1);
    pbt_set_size(EPSI, EPSI, MAXX-EPSI, MAXY-EPSI);

    /* Make a class for vehicles */
    vehicle_class = class_declare_class();

    /* Initialize the class stuff for libpbtab */
    pbt_class_init(vehicle_class);

    /* Make a bunch of vehicles */
    for (i=0;i<NVEH;i++)
    {
	/* Create the vehicle */
	user_data = class_alloc_user_data(vehicle_class);
	vtab_create_simnet_id(&simnet_id);
	ids[i] = vtab_vehicle_number(&simnet_id);
	vtab_new_vehicle(ids[i], &simnet_id, VTAB_LOCAL_VEHICLE,
			 (VTAB_USER_DATA_TYPE)user_data);

	/* Create the PBT part */
	pbt_create(ids[i]);

	/* Make up a location */
	locations[i][0] = random() % MAXX;
	locations[i][1] = random() % MAXY;

	/* Set the location in the vehicle table */
	pbt_set_location(ids[i], locations[i][0], locations[i][1]);
    }

    /* Drive them around a while */
    for (j=0;j<NITS;j++)
    {
	i = random() % NVEH;
	locations[i][0] += random() % 11 - 5;
	locations[i][1] += random() % 11 - 5;

	pbt_set_location(ids[i], locations[i][0], locations[i][1]);
    }

    for (k=0;k<2;k++)
    {
	/* See if the query stuff works */
	mine = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE, 0);
	his = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE, 0);
	for (j=0;j<NITS;j++)
	{
	    vtab_clear_list(mine);
	    vtab_clear_list(his);
	    x = random() % MAXX;
	    y = random() % MAXY;
	    r = random() % 200;

	    /* Find the vehicles near here using brute force */
	    r2 = r * r;
	    for (i=0;i<NVEH;i++)
	    {
		dx = locations[i][0] - x;
		dy = locations[i][1] - y;
		if (dx * dx + dy * dy <= r2)
		  vtab_add_to_list(mine, ids[i], 0);
	    }

	    /* Find them using the position based table */
	    pbt_get_vehicles(x, y, r, his);

	    /* See if they came up with the same answers */
	    if (!vtab_lists_equal(mine, his, 0))
	    {
		printf("MINE:\t");
		vtab_print_list(mine);
		printf("HIS:\t");
		vtab_print_list(his);
		abort();
	    }
	}

	/* Change the database size */
	pbt_set_size(0, 0, MAXX, MAXY);
    }

    /* Delete them */
    for (i=0;i<NVEH;i++)
    {
	pbt_destroy(ids[i]);
	class_free_user_data(vtab_get_vehicle(ids[i]));
	vtab_remove_vehicle(ids[i]);
    }

    printf("Yea Verily, everthing works.\n");
}

