/*
 * $RCSfile: libphysdb.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libphysdb.h                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libphysdb                          *
*       Created: Mon Jul  6 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBPHYSDB_H_INCLUDED 
#define _LIBPHYSDB_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> /*common/include/global*/

/* physdb_init:
 *
 * Initializes libphysdb, causing it to read its data file
 * (physdb.rdr) from the specified directory.
 * The flags are as in reader_read.
 * The return value is zero if the read succeeds, or one of the
 * libreader return values: READER_READ_ERROR READER_FILE_NOT_FOUND
 *
 * Note that the libreader function reader_init, the
 * librdrconst function rdc_init, and the libotmatch function otm_init,
 * must be called before this function.
 */
extern int32 physdb_init(   char *directory,
			    uint32 flags
			    );

/* Aspect bins are 30 degrees wide, and start centered at 0 degrees
 */
#define ASPECT_BINS	6
typedef struct contrast_data
{
    float32 hull[ASPECT_BINS];
    float32 turret[ASPECT_BINS];
} CONTRAST_DATA;

/* PHYSDB_DATA:
 *
 * This structure is the data returned by calls to otm_query with this
 * database key.
 */
typedef struct physdb_data 
{
    int32 _ignored;
    float32 dimensions[3];		/* X Y Z; width length height */
    float32 draft;			/* draft of vehicle */
    float32 collision_width;
    float32 model_base_adjustment;	/* Z */
    float32 mass_kg;
    float32 lethal_range;
    CONTRAST_DATA optical_contrast; 
    CONTRAST_DATA imagei_contrast;
    CONTRAST_DATA thermal_contrast;
    struct physdb_represented
    {
	int32 count;
	char  *names[1];
    } *represented;    
    
    struct physdb_turrets
    {
	int32 count;
	struct physdb_turret
	{
	    int32 _ignored;
	    char   *name;
	    float32 location[3];	/* X Y Z in vehicle coordinates */
	    float32 dimensions[3];	/* X Y Z; width length height */
	    float32 ccw_limit;		/* +radians (>= 2PI for no limit) */
	    float32 cw_limit;		/* +radians (= ccw_limit for fixed) */
	    struct physdb_guns
	    {
		int32 count;
		struct physdb_gun
		{
		    int32 _ignored;
		    char   *name;
		    float32 pivot[3];	/* X Y Z in turret coordinates */
		    float32 tip[3];	/* X Y Z in pivot coordinates */
		    float32 length;
		    float32 up_limit;	/* +radians (>= 2PI for no limit) */
		    float32 down_limit;	/* +radians (= up_limit is fixed) */
		    uint32  appearance; /* app. bits to turn on when active */
		} *gun[1];
	    } *guns;
	} *turret[1];
    } *turrets;
} PHYSDB_DATA;

/* physdb_key:
 *
 * Use this key value with otm_query to get the physical database
 * information for a vehicle.
 */
extern int32 physdb_key;

/* physdb_print:
 *
 * Prints the passed data on stdout.
 */
extern void physdb_print(/* PHYSDB_DATA *data */);

/*
 * physdb_nearest_angle:
 *
 * Return an angle that fits within limits imposed by physdb config
 * information.
 */
extern float64
  physdb_nearest_angle(/* float64 ccw,
			  float64 cw,
			  float64 desired
			  */);

#ifdef __cplusplus
}
#endif

#endif /*_LIBPHYSDB_H_INCLUDED*/
