#ifndef lint
static char rcsid [] = "$RCSfile: physdb_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: physdb_init.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Physdb initialization code                                *
*       Created: Mon Jul  6 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libphysdb.h"
#include <libotmatch.h>
#include <libreader.h>
#include <stdext.h> /*common/include/global*/
#include <math.h>

int32 physdb_key;

static PHYSDB_DATA default_data = 
{ 0, 
  {1.0 , 1.0 , 1.0},
  0.0, 
  0.0,
  0.0, 
  1.0,
  0.0,
  {{0.636, 0.636, 0.636, 0.636, 0.636, 0.636},
   {0.636, 0.636, 0.636, 0.636, 0.636, 0.636}},
  {{0.362, 0.362, 0.362, 0.362, 0.362, 0.362},
   {0.362, 0.362, 0.362, 0.362, 0.362, 0.362}},
  {{2.5, 2.5, 2.5, 2.5, 2.5, 2.5},
   {2.5, 2.5, 2.5, 2.5, 2.5, 2.5}},
  NULL,
  NULL
};

static struct physdb_turrets no_turrets = { 0 , NULL };
static struct physdb_guns no_guns = { 0 , NULL };
static struct physdb_represented no_represented = { 0, NULL };
    
int32 physdb_init(data_path, flags)
    char   data_path[];
    uint32 flags;
{
    READER_UNION db;
    int32 i, j, k;
    PHYSDB_DATA *data;
    struct physdb_gun *gun;
    int32 ret = reader_read("physdb.rdr", data_path, &db,
			    flags & ~READER_TYPING);

    if (ret != 0)
      return ret;

    /* Check if the data is OK */
    for (i=1;i<db.array[0].integer;i++)
    {
	/* Make sure the data is of the form:
	 * ( object_type (x y z t) )
	 */
	if (db.array[i].array[0].integer != 3)
	{
	    printf("Error in file physdb.rdr: item %d not of form (<object_type> (...))\n", i);
	    return READER_READ_ERROR;
	}

	data = (PHYSDB_DATA *)(db.array[i].array[2].array);

	/* See if we've already processed this one (this can happen
	 * if the data file contains macros).
	 */
	if (!data->_ignored)
	  continue;

	if (data->_ignored != (sizeof(PHYSDB_DATA)/sizeof(READER_UNION)))
	{
	    printf("Error in file physdb.rdr: item %d not of form \n(<x> <y> <z> <coll_width> <adjustment> <mass> \n<lethal_range> <opt_contrast> <therm_contrast> <turrets>)\n", i);
	    return READER_READ_ERROR; 
	}
	data->_ignored = 0; /* Macro protection */

	/* set collision_width to dimensions[X] (width) if width is greater */
	if (data->dimensions[X] > data->collision_width)
	  data->collision_width = data->dimensions[X];

	/* If they specified () as the turret, make that a legal array */
	if (!data->turrets)
	{
	    data->turrets = &no_turrets;
	    continue;
	}

	if (!data->represented)
	    data->represented = &no_represented;
	else
	  data->represented->count -= 1;
	
	/* 
         * Now make sure each turret is OK 
	 */

	/* Libreader counts the 'count', but we don't */
	data->turrets->count -= 1;
	for (j=0;j<data->turrets->count;j++)
	{
	    if (!data->turrets->turret[j]->_ignored) /* Macro protection */
	      continue;

	    if (data->turrets->turret[j]->_ignored !=
		(sizeof(struct physdb_turret)/sizeof(READER_UNION)))
	    {
		printf("Error in file physdb.rdr: item %d, turret %d not of form\n(<name> <off_x> <off_y> <off_z> <x> <y> <z> <ccw> <cw> <guns>)\n", i, j);
		return READER_READ_ERROR;
	    }
	    data->turrets->turret[j]->_ignored = 0; /* Macro protection */

	    if (data->turrets->turret[j]->ccw_limit < 0.0)
	      data->turrets->turret[j]->ccw_limit += 360.0;
	    data->turrets->turret[j]->ccw_limit =
	      DEG_TO_RAD(data->turrets->turret[j]->ccw_limit);
	    if (data->turrets->turret[j]->cw_limit < 0.0)
	      data->turrets->turret[j]->cw_limit += 360.0;
	    data->turrets->turret[j]->cw_limit =
	      DEG_TO_RAD(data->turrets->turret[j]->cw_limit);

	    if (!data->turrets->turret[j]->guns)
	    {
		data->turrets->turret[j]->guns = &no_guns;
		continue;
	    }

	    /* Libreader counts the 'count', but we don't */
	    data->turrets->turret[j]->guns->count -= 1;
	    for (k=0;k<data->turrets->turret[j]->guns->count;k++)
	    {
		if (!data->turrets->turret[j]->guns->gun[k]->_ignored)
		  continue;	/* Macro protection */

		if (data->turrets->turret[j]->guns->gun[k]->_ignored !=
		    (sizeof(struct physdb_gun)/sizeof(READER_UNION)))
		{
		    printf("Error in file physdb.rdr: item %d, turret %d, gun %d not of form\n(<name> <px> <py> <pz> <tx> <ty> <tz> 0.0 <up> <down>)\n", i, j, k);
		    return READER_READ_ERROR;
		}
		/* Macro protection */
		data->turrets->turret[j]->guns->gun[k]->_ignored = 0;

		gun = data->turrets->turret[j]->guns->gun[k];
		gun->length = sqrt((gun->tip[X]) * (gun->tip[X]) +
				   (gun->tip[Y]) * (gun->tip[Y]) +
				   (gun->tip[Z]) * (gun->tip[Z]));
		if (gun->up_limit < 0.0)
		  gun->up_limit += 360.0;
		gun->up_limit = DEG_TO_RAD(gun->up_limit);
		if (gun->down_limit < 0.0)
		  gun->down_limit += 360.0;
		gun->down_limit = DEG_TO_RAD(gun->down_limit);
	    }
	}
    }

    physdb_key = otm_incorporate_db((OTM_DB *)db.array, 1, &default_data);
    return (0);
}

void physdb_print(data)
    PHYSDB_DATA *data;
{
    int32 i, j;

    printf("width = %f\nlength = %f\nheight = %f\nmodel base adjustment = %f\nmass = %f kg\n",
	   data->dimensions[X], data->dimensions[Y], data->dimensions[Z],
	   data->model_base_adjustment, data->mass_kg);
    if(data->collision_width)
      printf("Collision width: %f\n", data->collision_width);
    else
      printf("Vehicle uses width for collision width\n");

    printf("optical_contrast - hull:");
    for (i=0; (i<ASPECT_BINS); ++i)
       printf(" %f",data->optical_contrast.hull[i]);
    printf("\n");
    printf("optical_contrast - turret:");
    for (i=0; (i<ASPECT_BINS); ++i)
       printf(" %f",data->optical_contrast.turret[i]);
    printf("\n");

    printf("thermal_contrast - hull:");
    for (i=0; (i<ASPECT_BINS); ++i)
       printf(" %f",data->thermal_contrast.hull[i]);
    printf("\n");
    printf("thrermal_contrast - turret:");
    for (i=0; (i<ASPECT_BINS); ++i)
       printf(" %f",data->thermal_contrast.turret[i]);
    printf("\n");

    for (i=0;i<data->represented->count; i++)
	printf("Represented on net:  %s\n", data->represented->names[i]);

    for (i=0;i<data->turrets->count;i++)
    {
	printf("Turret %2d \"%s\", location = <%f %f %f>\n",
	       i,
	       data->turrets->turret[i]->name,
	       data->turrets->turret[i]->location[X],
	       data->turrets->turret[i]->location[Y],
	       data->turrets->turret[i]->location[Z]);
	printf("           size %.3f, %.3f, %.3f\n",
	       data->turrets->turret[i]->dimensions[X],
	       data->turrets->turret[i]->dimensions[Y],
	       data->turrets->turret[i]->dimensions[Z]);

	if (data->turrets->turret[i]->ccw_limit >= TWO_PI)
	  printf("           movement unlimited\n");
	else if (data->turrets->turret[i]->ccw_limit ==
		 data->turrets->turret[i]->cw_limit)
	  printf("           movement fixed at %d\n",
		 (int)RAD_TO_DEG(data->turrets->turret[i]->ccw_limit));
	else
	  printf("           limits = %d ccw, %d cw\n",
		 (int)RAD_TO_DEG(data->turrets->turret[i]->ccw_limit),
		 (int)RAD_TO_DEG(data->turrets->turret[i]->cw_limit));
	printf("           gun count = %d\n",
	       data->turrets->turret[i]->guns->count);

	for (j=0;j<data->turrets->turret[i]->guns->count;j++)
	{
	    printf("   Gun %2d \"%s\",\n           pivot = <%f %f %f>\n           tip = <%f %f %f>\n           length = %f\n",
		   j,
		   data->turrets->turret[i]->guns->gun[j]->name,
		   data->turrets->turret[i]->guns->gun[j]->pivot[X],
		   data->turrets->turret[i]->guns->gun[j]->pivot[Y],
		   data->turrets->turret[i]->guns->gun[j]->pivot[Z],
		   data->turrets->turret[i]->guns->gun[j]->tip[X],
		   data->turrets->turret[i]->guns->gun[j]->tip[Y],
		   data->turrets->turret[i]->guns->gun[j]->tip[Z],
		   data->turrets->turret[i]->guns->gun[j]->length);
	    if (data->turrets->turret[i]->guns->gun[j]->up_limit >= TWO_PI)
	      printf("           movement unlimited\n");
	    else if (data->turrets->turret[i]->guns->gun[j]->up_limit ==
		     data->turrets->turret[i]->guns->gun[j]->down_limit)
	      printf("           movement fixed at %d\n",
		     (int)RAD_TO_DEG(data->turrets->turret[i]->
				     guns->gun[j]->down_limit));
	    else
	      printf("           limits = %d up, %d down\n",
		     (int)RAD_TO_DEG(data->turrets->turret[i]->
				     guns->gun[j]->up_limit),
		     (int)RAD_TO_DEG(data->turrets->turret[i]->
				     guns->gun[j]->down_limit));
	    if (data->turrets->turret[i]->guns->gun[j]->appearance)
	      printf("           appearance bits = 0x%x\n",
		     data->turrets->turret[i]->guns->gun[j]->appearance);
	    else
	      printf("           no appearance modifier\n");
	}
    }
}


