/*
 * $RCSfile: libpo_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpo_local.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Local header file for libpo                               *
*       Created: Wed Apr 10 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/*
 * Non-static private functions.
 */

/*
 * po_init.c
 */
extern int32 po_ready(
    PO_DATABASE *db,
    uint32       clock);

/* 
 * po_del_pdu.c:
 */
extern void
  po_send_delete_objects_pdu(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries);

/*
 * po_delete.c:
 */
extern void
  po_delete_dead_objects(
    PO_DATABASE *db);

extern int32
  po_process_delete_objects(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);
extern PO_DB_ENTRY *
  po_should_have_been_deleted(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu);

extern int32
  po_delete_entries_internal(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         object_deleted,
    int32         call_handlers);
extern void
  po_duplicate_for_world_state_deletion(
    PO_DATABASE *db,
    PO_DB_ENTRY *ws_entry,
    PO_DB_ENTRY *entry);

/*
 * po_do_pdu.c:
 */
extern PO_DB_ENTRY *
  po_send_new_describe_object_pdu(
    PO_DATABASE       *db,
    ObjectID          *id,
    ObjectID          *world_state,
    uint8              class,
    uint32             missing_from_world_state,
    ADDRESS            variants,
    uint32             vsize,
    PO_USER_DATA_TYPE  object_user_data);

extern void
  po_send_describe_object_pdu(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);
				 
extern void
  po_send_oob_describe_object_pdu(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);
extern PO_DB_ENTRY *
  po_send_changed_describe_object_pdu(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       missing_from_world_state,
    ADDRESS      variants,
    uint32       vsize);

/*
 * po_error.c:
 */
extern int32
  po_good_pdu(
    PersistentObjectPDU *pdu,
    int32                pdu_size);

extern int32
  po_associated_ok(
    PO_DATABASE *db,
    uint8        class,
    char        *variant,
    ObjectID    *ws_id);

extern int32
  po_valid_pdu(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu);

#ifdef PARANOID
extern void 
  po_check_queues(
      PO_DATABASE *db,
      PO_DB_ENTRY *entry);
#endif /*  PARANOID */

/*
 * po_hash.c:
 */
extern void
  po_choose_world_state_id(
    PO_DATABASE *db,
    ObjectID    *id_ret);

extern void
  po_choose_object_id(
    PO_DATABASE *db,
    ObjectID    *id_ret,
    ObjectID    *ws_id);

extern int32
  po_reserve_object_ids(
    PO_DATABASE *db,
    int32        num_ws,
    int32        num_obj,
    ObjectID    *id_buf);

extern void
  po_return_object_ids(
    PO_DATABASE *db,
    int32        num_ids,
    ObjectID    *id_buf,
    char        *used_vector);

extern void
  po_return_object_id(
    PO_DATABASE *db,
    ObjectID    *id);

extern PO_DB_ENTRY *
  po_find_matching_entry(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu);

extern PO_DB_ENTRY *
  po_find_entry(
    PO_DATABASE *db,
    ObjectID    *object_id,
    ObjectID    *world_state_id);

extern void
  po_destroy_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_destroy_all_entries(
    PO_DATABASE *db);

extern void
  po_destroy_all_objects(
    PO_DATABASE *db);

extern PO_DB_ENTRY *
  po_add_entry_to_db(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size,
    int32                implied,
    PO_USER_DATA_TYPE    object_user_data);

/*
 * po_load:
 */
extern uint32
  po_load(
    PO_DATABASE *db);

extern void
  po_packet_from_simulator(
    PO_DATABASE       *db,
    SimulationAddress *sim);

extern void
  po_reset_load(
    PO_DATABASE *db);

extern PO_DB_ENTRY *
  po_lowest_loaded_simulator(
    PO_DATABASE *db,
    int32        skip_me);

/*
 * po_netws:
 */
extern int32
  po_process_set_world_state(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);
extern void
  po_tick_animation(
    PO_DATABASE *db);

/*
 * po_nom_pdu.c:
 */
extern void
  po_send_nominate_pdus(
    PO_DATABASE       *db,
    SimulationAddress *nominated_addr,
    int32              n_missing,
    SimulationAddress *missing_addrs);

/*
 * po_objects.c:
 */

extern void
  po_take_ownership(
    PO_DATABASE       *db,
    int32              n_addrs,
    SimulationAddress *addrs);

extern void
  po_not_dormant(
    PO_DB_ENTRY *entry,
    PO_DATABASE *db);

extern void
  po_nothing_dormant(
    PO_DATABASE *db);

extern void
  po_send_overlays_and_world_states(
    PO_DATABASE *db);

extern void
  po_tick_objects(
    PO_DATABASE *db);

extern void
  po_timeout_objects(
    PO_DATABASE *db);

extern int32
  po_process_object_request(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);
extern int32
  po_process_objects_present(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);

extern PO_DB_ENTRY *
  po_process_describe_object(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size,
    PO_USER_DATA_TYPE    object_user_data);
extern PO_DB_ENTRY *
  po_find_or_invent_world_state(
    PO_DATABASE *db,
    ObjectID    *ws_id);

/*
 * po_op_pdu.c:
 */

extern void
  po_add_object_present(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_remove_object_present(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);
/*
 * po_or_pdu.c:
 */

extern void
  po_send_object_request_pdu(
    PO_DATABASE       *db,
    SimulationAddress *owner,
    ObjectID          *world_state,
    int32              n_ids,
    ObjectID          *obj_ids);

/*
 * po_overlay.c:
 */

extern int32
  po_get_overlay_id(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    ObjectID            *id_ret);

/*
 * po_sims.c:
 */
extern int32
  po_process_simpresent(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);

extern void
  po_set_database_sequence_number(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       new_number);

extern void
  po_set_time(
    PO_DATABASE *db,
    uint32       new_time);

extern void
  po_timeout_sims(
    PO_DATABASE *db);

extern int32
  po_process_nomination(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);

/*
 * po_sp_pdu.c:
 */

extern void
  po_send_simulator_present_pdu(
    PO_DATABASE *db);

/*
 * po_stealth.c:
 */

extern void
  po_handle_new_sc(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_handle_changed_sc(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);
/*
 * po_sws_pdu.c:
 */

extern int32
  po_send_set_world_state_pdu(
    PO_DATABASE *db,
    ObjectID    *world_state_id,
    uint32       secondsSince1970,
    float32      clock_rate);
extern int32
  po_resend_set_world_state_pdu(
    PO_DATABASE *db,
    uint32       dtime);

/*
 * po_thread.c:
 */

extern void
  po_thread_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_unthread_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern PO_OBJECT_HANDLE *
  po_find_handle(
    PO_DATABASE *db,
    ObjectID    *object_id);

extern void
  po_make_local(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_make_remote(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_describe_object_sent(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_unthread_time(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/*
 * po_ws.c:
 */
extern PO_DB_ENTRY *
  po_get_object_for_world_state(
    PO_DATABASE *db,
    ObjectID    *object_id,
    ObjectID    *world_state_id);

extern PO_DB_ENTRY *
  po_current_object(
    PO_DATABASE *db,
    ObjectID    *object_id);

extern int32
  po_world_state_dependency(
    PO_DB_ENTRY *dependent_state,
    PO_DB_ENTRY *base_state,
    ObjectID    *subsequent_state);

extern int32
  po_get_associate_id(
    PersistentObjectPDU *pdu,
    ObjectID            *id_ret);

extern void
  po_handle_new_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

extern void
  po_handle_entry_changed(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    int32        previously_missing);

extern void
  po_handle_entry_gone(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/*
 * po_exercise.c:
 */
extern ObjectID *
  po_get_exercise_select_id(void);

extern void
  po_handle_new_exercise_initializer(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);
extern void
  po_handle_changed_exercise_initializer(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* DRN enumerates starting with 1 */
#define PO_OBJECT_CLASS_INDEX(class) (class-1)

#define PO_ERROR(code) { po_errno = code; return 0; }

#define SIM_ADDR_MAG(sim) ((uint32)(((sim).site << 16) | ((sim).host)))

#define SAME_SIMULATOR(x,y) \
  (((x).site == (y).site) && \
   ((x).host == (y).host))

#define SAME_OBJECT_ID(x,y) \
  ((SAME_SIMULATOR((x).simulator,(y).simulator)) && \
   ((x).object == (y).object))

#define SAME_OBJECT(xw,xo,yw,yo) \
  (SAME_OBJECT_ID(xw,yw) && SAME_OBJECT_ID(xo,yo))

/* Insert into doubly linked list */
#define LL_INSERT(entry, head, prev, next)	\
{						\
    (entry)->prev = NULL;			\
    (entry)->next = (head);			\
    if ((head))					\
      (head)->prev = (entry);			\
    (head) = (entry);				\
}

/* Delete from doubly linked list */
#define LL_DELETE(entry, head, prev, next)	\
{						\
    if ((head) == (entry))			\
      (head) = (entry)->next;			\
    if ((entry)->next)				\
      (entry)->next->prev = (entry)->prev;	\
    if ((entry)->prev)				\
      (entry)->prev->next = (entry)->next;	\
    (entry)->prev = NULL;			\
    (entry)->next = NULL;			\
}

/* Insert into doubly linked queue */
#define Q_INSERT(entry, head, tail, prev, next)	\
{						\
    if ((entry) == (entry)->next)		\
      abort();					\
    (entry)->prev = (tail);			\
    (entry)->next = NULL;			\
    if ((tail))					\
      (tail)->next = (entry);			\
    if (!(head))				\
      (head) = (entry);				\
    (tail) = (entry);				\
    if ((entry) == (entry)->next)		\
      abort();					\
}

/* Delete from doubly linked queue */
#define Q_DELETE(entry, head, tail, prev, next)	\
{						\
    if ((entry) == (entry)->next)		\
      abort();					\
    if ((head) == (entry))			\
      (head) = (entry)->next;			\
    if ((tail) == (entry))			\
      (tail) = (entry)->prev;			\
    if ((entry)->next)				\
      (entry)->next->prev = (entry)->prev;	\
    if ((entry)->prev)				\
      (entry)->prev->next = (entry)->next;	\
    (entry)->prev = NULL;			\
    (entry)->next = NULL;			\
}
