#ifndef lint
static char rcsid [] = "$RCSfile: po_do_pdu.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_do_pdu.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to send describe object PDUs                         *
*       Created: Wed Apr 10 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <po_size.h>
#include <stdext.h> /* common/include/global */
#include <stdio.h>

/*VARARGS6*/
PO_DB_ENTRY *
po_send_new_describe_object_pdu(
    PO_DATABASE       *db,
    ObjectID          *id,
    ObjectID          *world_state,
    uint8              class,
    uint32             missing_from_world_state,
    ADDRESS            variants,
    uint32             vsize,
    PO_USER_DATA_TYPE  object_user_data)
{
    char buf[NETWORK_BUFFER_SIZE];
    PersistentObjectPDU *pdu = (PersistentObjectPDU *)buf;
    int32 pdu_size;
    PO_DB_ENTRY *entry;

    if (!db->ready)
      printf("Warning: Creating object before the database is ready\n");

    pdu->version = persistentObjectProtocolVersionCurrent;
    pdu->kind = describeObjectPDUKind;
    pdu->exercise = db->exercise_id;
    pdu->database = db->database_id;
    pdu->variant.describeObject.databaseSequenceNumber =
      db->database_sequence_number;
    pdu->variant.describeObject.objectID = *id;
    pdu->variant.describeObject.worldStateID = *world_state;
    pdu->variant.describeObject.owner = db->sim_addr;
    pdu->variant.describeObject.sequenceNumber = 1;
/*changed class to pclass 3 Nov 97 TRANG*/
    pdu->variant.describeObject.pclass = class;
    pdu->variant.describeObject.missingFromWorldState =
      missing_from_world_state;
    bcopy((char *)variants,
	  (char *)&pdu->variant.describeObject.variant,
	  (unsigned int) vsize);

    pdu_size = PRO_PO_DESC_OBJ_SIZE(pdu)+vsize;
    pdu->length =  NS_HTON16(pdu_size);

    /* Check to make sure the user gave us good info */
    if (!po_valid_pdu(db, pdu))
    {
	/* Since we will not be deleting this object explicity, 
	 * tell the hashing library it can reuse this id.
	 */
	po_return_object_id(db, id);

	return NULL;
    }

    /* Loopback to add to our database */
    if (entry =
	po_process_describe_object(db, pdu, pdu_size, object_user_data))
    {
	db->send_handler(pdu, pdu_size, db->send_user_data);
	db->packets_sent++;
    }

    return entry;
}

void po_take_possession(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    /* This function provides a public interface to 
     * po_send_describe_object_pdu which causes the simulator to
     * take possession of the entry.
     */
    po_send_describe_object_pdu(db, entry);
}

void po_send_describe_object_pdu(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    /* Ensure we own the object. */
    if (!SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
			db->sim_addr))
    {
	/* If we are taking it over, then it is not dormant! */
	po_not_dormant(entry, db);

	po_make_local(db, entry);

	entry->pdu->variant.describeObject.owner = db->sim_addr;
	entry->pdu->variant.describeObject.sequenceNumber++;
    }
    /* In case this is an UNDELETABLE object */
    entry->pdu->variant.describeObject.databaseSequenceNumber =
      db->database_sequence_number;

    /* Note send occurred */
    po_describe_object_sent(db, entry);

    db->send_handler(entry->pdu, entry->pdu_size, db->send_user_data);
    db->packets_sent++;
}

void po_send_oob_describe_object_pdu(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    /* This code sends a describe object pdu "out of band" which
     * just sends the object data, but doesn't remember we did it.
     * We do this to let other folks know about objects without
     * making long term spikes.
     */

    /* Ensure we own the object.  This function can not be used
     * to grab objects from other sims.
     */
    if (!SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
			db->sim_addr))
      return;

    /* In case this is an UNDELETABLE object */
    entry->pdu->variant.describeObject.databaseSequenceNumber =
      db->database_sequence_number;

    db->send_handler(entry->pdu, entry->pdu_size, db->send_user_data);
    db->packets_sent++;
}

/*VARARGS4*/
PO_DB_ENTRY *
po_send_changed_describe_object_pdu(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       missing_from_world_state,
    ADDRESS      variants,
    uint32       vsize)
{
    char buf[NETWORK_BUFFER_SIZE];
    PersistentObjectPDU *pdu = (PersistentObjectPDU *)buf;
    int32 pdu_size;

    /* If it is being changed, then it is not dormant! */
    po_not_dormant(entry, db);

    /* Copy the last pdu, and just change the stuff we need to. */
    bcopy((char *)entry->pdu, (char *)pdu, (unsigned int)entry->pdu_size);
    /* In case this is an UNDELETABLE object */
    pdu->variant.describeObject.databaseSequenceNumber =
      db->database_sequence_number;
    pdu->variant.describeObject.owner = db->sim_addr;
    pdu->variant.describeObject.sequenceNumber++;
    pdu->variant.describeObject.missingFromWorldState =
      missing_from_world_state;
    bcopy((char *)variants,
	  (char *)&pdu->variant.describeObject.variant,
	  (unsigned int) vsize);

    pdu_size = PRO_PO_DESC_OBJ_SIZE(pdu)+vsize;
    pdu->length =  NS_HTON16(pdu_size);

    /* Loopback to add to our database */
    if (po_process_describe_object(db, pdu, pdu_size, (PO_USER_DATA_TYPE)NULL))
    {
	db->send_handler(pdu, pdu_size, db->send_user_data);
	db->packets_sent++;
    }

    return entry;
}
