#ifndef lint
static char rcsid [] = "$RCSfile: po_error.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_error.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Error related routines                                    *
*       Created: Fri Apr 12 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <stdext.h> /* common/include/global */

int32 po_errno = 0;

/* Error Handling.
 *
 * The basic idea is that when an error occurs, the application can look
 * at po_errno to find out the reason.  It can then look at
 * po_errlist[po_errno] for a string description.
 */

char * po_errlist[] = {
    "No Error" ,
    "Illegal operation for inactive database" ,
    "Send handler required" ,
    "Memory allocation failed" ,
    "All 65536 unique object IDs have been used" ,
    "Bad base state" ,
    "Attempt to reuse an existing objectID/worldStateID pair" ,
    "" ,
    "" ,
    "" ,
    "Bad data in passed PDU" ,
    "PDU size too large" ,
    "Unknown protocol version" ,
    "Unknown or unexpected PDU kind" ,
    "Unknown or unexpected object class" ,
    "Object refers to unknown overlay" ,
    "Object can only exist in Real Time world state",
    "Object in unknown world state",
    "Object refers to unknown associated object",
    "Object will be/has been deleted",
    "Error opening file" ,
    "Error writing into file" ,
    "Error reading from file" ,
    "Error closing file" ,
    "Nothing to save or load" ,
    "Unknown file format" ,
    "File uses different Unit Database" ,
    "File uses different Terrain Database" ,
    "Data in file is corrupt" ,
    "Overlay in file conflicts with existing overlay" ,
};

int32 po_good_pdu(
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    /* Check for reasonable size */
    if (pdu_size > (NETWORK_BUFFER_SIZE-
                    (16 /* ethernet & padding */ +
                     8  /* 802.3              */ +
                     20 /* IP                 */ +
                     8  /* UDP                */ +
                     12 /* DIS                */)))
      PO_ERROR(PO_ESIZE);

    /* Check for bad version */
    if (pdu->version != persistentObjectProtocolVersionCurrent)
      PO_ERROR(PO_EVRSN);

    /* Check for unknown kind */
    switch (pdu->kind)
    {
      case simulatorPresentPDUKind:
      case deleteObjectsPDUKind:
      case objectsPresentPDUKind:
      case objectRequestPDUKind:
      case setWorldStatePDUKind:
      case nominationPDUKind:
	break;
      case describeObjectPDUKind:
	/* Check for unknown class */
/*changed class to pclass 3 Nov 97 TRANG*/
	if (pdu->variant.describeObject.pclass > PO_MAX_OBJECT_CLASS)
	  PO_ERROR(PO_ECLASS);
	break;
      default:
	PO_ERROR(PO_EKIND);
    }
    return TRUE;
}

static int32 check_for_valid_object(
    PO_DATABASE *db,
    ObjectID    *associated,
    ObjectID    *ws_id)
{
    if ((associated->simulator.site != 0) ||
	(associated->simulator.host != 0) ||
	(associated->object != 0))
    {
	if (!po_get_object_for_world_state(db, associated, ws_id))
	  return FALSE;
    }
    return TRUE;
}

int32 po_associated_ok(
    PO_DATABASE *db,
    uint8        class,
    char        *variant,
    ObjectID    *ws_id)
{
    int32 i;
    int32 result = TRUE;

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol which has an
     * associatedObject field, add it somewhere in this switch.
     */

    switch(class)
    {
      case objectClassText:
	result &=
	  check_for_valid_object(db,
				 &((TextClass *)variant)->associatedObject,
				 ws_id);
	break;
      case objectClassUnit:
	result &= check_for_valid_object(db,
				      &((UnitClass *)variant)->taskOrganizedTo,
					 ws_id);
/*	result &= check_for_valid_object(db,
 *			      &((UnitClass *)variant)->functionallyOrganizedTo,
 *					 ws_id);
 */
	result &= check_for_valid_object(db,
				       &((UnitClass *)variant)->taskFrameStack,
					 ws_id);
	result &= check_for_valid_object(db,
				      &((UnitClass *)variant)->backgroundFrame,
					 ws_id);
	result &= check_for_valid_object(db,
				      &((UnitClass *)variant)->parameters,
					 ws_id);
	result &= check_for_valid_object(db,
				      &((UnitClass *)variant)->handOffData,
					 ws_id);	
	break;
      case objectClassTask:
	result &= check_for_valid_object(db,
					 &((TaskClass *)variant)->frame,
					 ws_id);
	result &= check_for_valid_object(db,
					 &((TaskClass *)variant)->state,
					 ws_id);
	for (i = 0; i < ((TaskClass *)variant)->refcount; i++)
	  result &=
	    check_for_valid_object(db,
				   PO_REFERENCES(*(TaskClass *)variant)+i,
				   ws_id);

	break;
      case objectClassTaskState:
	for (i = 0; i < ((TaskStateClass *)variant)->refcount; i++)
	  result &=
	    check_for_valid_object(db,
			   PO_REFERENCES(*(TaskStateClass *)variant)+i,
				   ws_id);

	break;
      case objectClassTaskFrame:
/*
	result &=
	  check_for_valid_object(db,
				 &((TaskFrameClass *)variant)->unit,
				 ws_id);
	result &=
	  check_for_valid_object(db,
				 &((TaskFrameClass *)variant)->nextStackFrame,
				 ws_id);
	result &=
	  check_for_valid_object(db,
			    &((TaskFrameClass *)variant)->previousMissionFrame,
			    ws_id);
*/
	for (i = 0; i < ((TaskFrameClass *)variant)->etaskCount; i++)
	  result &=
	    check_for_valid_object(db,
				 &((TaskFrameClass *)variant)->enablingTask[i],
				 ws_id);	  
	break;
      case objectClassTaskAuthorization:
/*
	result &=
	  check_for_valid_object(db,
		 &((TaskAuthorizationClass *)variant)->proposedTaskFrame,
				 ws_id);
	result &=
	  check_for_valid_object(db,
		 &((TaskAuthorizationClass *)variant)->modifiedTaskFrame,
				 ws_id);
	result &=
	  check_for_valid_object(db,
				 &((TaskAuthorizationClass *)variant)->unit,
				 ws_id);
*/
	break;
      case objectClassParametricInput:
	result &= 
	  check_for_valid_object(db,
				 &((ParametricInputClass *)variant)->chain,
				 ws_id);
	break;
      case objectClassParametricInputHolder:
	for (i = 0; i < ((ParametricInputHolderClass *)variant)->size; i++)
	  result &=
	    check_for_valid_object(db,
		      &((ParametricInputHolderClass *)variant)->blocks[i].data,
		      ws_id);
	break;

      case objectClassFireParameters:
	result &=
	  check_for_valid_object(db,
				 &((FireParametersClass *)variant)->unit,
				 ws_id);
	break;
      case objectClassLink:
	for (i=0; i<maxLinkConnections; i++)
	    result &= 
	      check_for_valid_object(db,
				     &(((LinkClass *)variant)->memberIds[i]),
				     ws_id);
	break;
      case objectClassMinefield:
	result &=
	  check_for_valid_object(db,
				 &((MinefieldClass *)variant)->parentID,
				 ws_id);
	break;
      case objectClassHandOff:
	for (i=0; i<((HandOffClass *)variant)->numHandOff; i++)
	  result &= 
	    check_for_valid_object(db,
				   &((HandOffClass *)variant)->handOff[i].unit,
				   ws_id);
      default:
	return TRUE;
    }
    return result;
}

int32 po_valid_pdu(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu)
{
    ObjectID overlay;
    PO_DB_ENTRY *entry;

    if (pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    bzero((char *)&overlay, sizeof(overlay));

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol, add it somewhere
     * in this switch.
     */

    switch(pdu->variant.describeObject.pclass)
    {
      /*
       * Classes which must be in the real time world state:
       */
      case objectClassWorldState:
      case objectClassOverlay:
      case objectClassStealthController:
      case objectClassHHour:
	if (!SAME_OBJECT_ID(pdu->variant.describeObject.worldStateID,
			    po_real_time_world_state))
	  PO_ERROR(PO_ERTWS);
	return TRUE;

      /*
       * Classes which require the presence of an overlay/association.
       */
      case objectClassPoint:
	overlay =   pdu->variant.describeObject.variant.point.overlayID;
	break;
      case objectClassLine:
	overlay =   pdu->variant.describeObject.variant.line.overlayID;
	break;
      case objectClassMinefieldParent:
	overlay =   pdu->variant.describeObject.variant.minefieldParent.overlayID;
	break;
      case objectClassChemical:
	overlay =   pdu->variant.describeObject.variant.chemical.overlayID;
	break;
      case objectClassMinefield:
	if (!po_associated_ok(db, objectClassMinefield,
			      (char *)&(pdu->variant.describeObject.variant.minefield),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassSector:
	overlay =   pdu->variant.describeObject.variant.sector.overlayID;
	break;
      case objectClassText:
	overlay =   pdu->variant.describeObject.variant.text.overlayID;
	if (!po_associated_ok(db, objectClassText,
			      (char *)&(pdu->variant.describeObject.variant.text),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassUnit:
	overlay =   pdu->variant.describeObject.variant.unit.overlayID;
	if (!po_associated_ok(db, objectClassUnit,
			      (char *)&(pdu->variant.describeObject.variant.unit),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);

	break;
      case objectClassTask:
	if (!po_associated_ok(db, objectClassTask,
			      (char *)&(pdu->variant.describeObject.variant.task),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;

      case objectClassTaskState:
	if (!po_associated_ok(db, objectClassTaskState,
			      (char *)&(pdu->variant.describeObject.variant.taskState),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;

      case objectClassTaskFrame:
	if (!po_associated_ok(db, objectClassTaskFrame,
			      (char *)&(pdu->variant.describeObject.variant.taskFrame),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassTaskAuthorization:
	if (!po_associated_ok(db, objectClassTaskAuthorization,
			      (char *)&(pdu->variant.describeObject.variant.taskAuthorization),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassParametricInput:
	if (!po_associated_ok(db, objectClassParametricInput,
			      (char *)&(pdu->variant.describeObject.variant.parametricInput),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassParametricInputHolder:
	if (!po_associated_ok(db, objectClassParametricInputHolder,
			      (char *)&(pdu->variant.describeObject.variant.parametricInputHolder),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassFireParameters:
	if (!po_associated_ok(db, objectClassFireParameters,
			      (char *)&(pdu->variant.describeObject.variant.fireParameters),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassLink:
	overlay =   pdu->variant.describeObject.variant.link.overlayID;
	if (!po_associated_ok(db, objectClassLink,
			      (char *)&(pdu->variant.describeObject.variant.link),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);
	break;
      case objectClassHandOff:
	if (!po_associated_ok(db, objectClassHandOff,
			      (char *)&(pdu->variant.describeObject.variant.handOff),
			      &(pdu->variant.describeObject.worldStateID)))
	  PO_ERROR(PO_EASSOC);

      /*
       * Classes which have no sematic requirements:
       */
      default:
	return TRUE;
    }

    if ((overlay.simulator.site != 0) ||
	(overlay.simulator.host != 0) ||
	(overlay.object != 0))
    {
	entry = po_get_object_for_world_state(db, &overlay,
					      &po_real_time_world_state);

	if (!entry ||
	    (entry->pdu->kind != describeObjectPDUKind) ||
	    (entry->pdu->variant.describeObject.pclass !=
	     objectClassOverlay))
	  PO_ERROR(PO_EOVLY);
    }

    return TRUE;
}

#ifdef PARANOID
void po_check_queues(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *i;
    int32 on_deleted = FALSE;
    int32 on_retransmit = FALSE;
    int32 on_timeout = FALSE;

    if (entry->pdu->kind != describeObjectPDUKind)
      return;

    if (db->deleted_head && db->deleted_head->prev_time)
      abort();

    if (db->retransmit_head && db->retransmit_head->prev_time)
      abort();

    if (db->timeout_head && db->timeout_head->prev_time)
      abort();

    for(i=db->deleted_head;i;i=i->next_time)
      if (i == entry)
      {
	  on_deleted = TRUE;
	  break;
      }
    for(i=db->retransmit_head;i;i=i->next_time)
      if (i == entry)
      {
	  on_retransmit = TRUE;
	  break;
      }
    for(i=db->timeout_head;i;i=i->next_time)
      if (i == entry)
      {
	  on_timeout = TRUE;
	  break;
      }

    if (entry->implied)
    {
	if (on_deleted || on_retransmit || on_timeout)
	  abort();
    }
    else if(entry->deleted)
    {
	if (!on_deleted || on_retransmit || on_timeout)
	  abort();
    }
    else if (SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
			    db->sim_addr))
    {
	if (on_deleted || !on_retransmit || on_timeout)
	  abort();
    }
    else
    {
	if (on_deleted || on_retransmit || !on_timeout)
	  abort();
    }
}
#endif
