#ifndef lint
static char rcsid [] = "$RCSfile: po_exercise.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_exercise.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to manage the exerciseInitializerClass packets       *
*       Created: 7/7/92                                                     *
*       Author: ajc                                                         *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h> /* common/include/global */

/*
 * Common users of libpo guarentee that there is only one exercise initializer
 *   object by forcing it to have a known common id.
 */

static ObjectID g_exercise_select_id = { { 0, 0 }, 1 };

static ExerciseInitializerClass g_default_exercise_info =
{
    { "Unknown", 0 },
    battleSchemeAbsolute,
    0,
    1.0,
    0,
};

ObjectID *po_get_exercise_select_id(void)
{
    return &g_exercise_select_id;
}

ExerciseInitializerClass *po_get_exercise_initialization(
    PO_DATABASE *db)
{
    PO_DB_ENTRY *retrieved = po_get_entry(db, &g_exercise_select_id,
					  &po_real_time_world_state);

    if (retrieved)
      return &(PO_EXERCISE_INITIALIZER_DATA(retrieved));
    else
      return &g_default_exercise_info;
}

void po_set_exercise_initialization(
    PO_DATABASE              *db,
    ExerciseInitializerClass *data)
{
    PO_DB_ENTRY *retrieved;

    /* Can you modify or create objects in this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      return;
      
    /* Do we already know about an exercise initializer? */
    retrieved = po_get_entry(db, &g_exercise_select_id,
			     &po_real_time_world_state);

    if (retrieved)
      /* Since there is already one, I must change it to have the new data */
      (void) po_send_changed_describe_object_pdu(db, retrieved,
						 (int32)FALSE, (char *)data,
					     sizeof(ExerciseInitializerClass));
    else
      /* Since there is no current one, I must create one */
      (void) po_send_new_describe_object_pdu(db, &g_exercise_select_id,
					     &po_real_time_world_state, 
					     objectClassExerciseInitializer,
					     (int32)FALSE, (char *)data,
					     sizeof(ExerciseInitializerClass),
					     NULL);
}

void po_delete_exercise_initialization(
    PO_DATABASE *db)
{
    PO_DB_ENTRY *retrieved;

    /* Can you modify or create objects in this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      return;
      
    /* Do we know about an exercise initializer? */
    retrieved = po_get_entry(db, &g_exercise_select_id,
			     &po_real_time_world_state);

    if (retrieved)
      po_delete_object(db, retrieved);
}

void po_handle_new_exercise_initializer(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    callback_fire_event(db->exercise_initialization_event,
			&PO_EXERCISE_INITIALIZER_DATA(entry));
}

void po_handle_changed_exercise_initializer(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    callback_fire_event(db->exercise_initialization_event,
			&PO_EXERCISE_INITIALIZER_DATA(entry));
}
