#ifndef lint
static char rcsid [] = "$RCSfile: po_load.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_load.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to calculate a machine's load                        *
*       Created: Mon Apr  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h> /* common/include/global */

uint32 po_load(
    PO_DATABASE *db)
{
    return (db->packets_sent + db->packets_missed*db->packets_missed);
}

void po_packet_from_simulator(
    PO_DATABASE       *db,
    SimulationAddress *sim)
{
    PO_DB_ENTRY *sim_entry;

    if (sim_entry = po_find_simulator(db, sim))
    {
	sim_entry->pdu->variant.simulatorPresent.packetsSent += 1;
    }
}

void po_reset_load(
    PO_DATABASE *db)
{
    db->packets_sent = 0;
    db->packets_missed = 0;
}

PO_DB_ENTRY *
po_lowest_loaded_simulator(
    PO_DATABASE *db,
    int32        skip_me)
{
    PO_DB_ENTRY *sim_entry;
    PO_DB_ENTRY *lowest_load_entry;
    uint32 lowest_load;

    /* Unless the caller wants otherwise, choose yourself if you are
     * <= other sims.
     */
    if (!skip_me)
    {
	lowest_load_entry = db->own_sim_entry;
	lowest_load = db->own_sim_entry->pdu->variant.simulatorPresent.load;
    }
    else
    {
	lowest_load_entry = NULL;
	lowest_load = 0xFFFFFFFF; /* Very big number */
    }

    for(sim_entry = db->sim_class_head;
	sim_entry;
	sim_entry = sim_entry->next_class)
    {
	if (skip_me && (sim_entry == db->own_sim_entry))
	  continue;

	if (sim_entry->pdu->variant.simulatorPresent.load < lowest_load)
	{
	    lowest_load = sim_entry->pdu->variant.simulatorPresent.load;
	    lowest_load_entry = sim_entry;
	}
    }

    return lowest_load_entry;
}

float32 po_get_simulation_load(
    PO_DATABASE *db)
{
    return db->simulation_load;
}

void po_set_simulation_load(
    PO_DATABASE *db,
    float32      val)
{
    db->simulation_load = val;
}

PO_DB_ENTRY *po_least_simulation_loaded_simulator(
    PO_DATABASE       *db,
    SimulatorType      sim_type,
    SimulationAddress *excluded_sim)
{
    PO_DB_ENTRY *sim_entry;
    PO_DB_ENTRY *lowest_load_entry;
    float32      lowest_load;
    
    if ((db->own_sim_entry->pdu->variant.simulatorPresent.simulatorType ==
	 sim_type) &&
	(!excluded_sim ||
	 !NS_SIM_ADDRESSES_EQUAL(*excluded_sim, db->sim_addr)))
    {
	lowest_load_entry = db->own_sim_entry;
	lowest_load =
	  db->own_sim_entry->pdu->variant.simulatorPresent.simulationLoad;
    }
    else
    {
	lowest_load_entry = NULL;
	lowest_load = 1.0e+20; /* Very big number */
    }

    for(sim_entry = db->sim_class_head;
	sim_entry;
	sim_entry = sim_entry->next_class)
    {
	if ((sim_entry->pdu->variant.simulatorPresent.simulatorType == 
	     sim_type) &&
	    (!excluded_sim ||
	     !NS_SIM_ADDRESSES_EQUAL(*excluded_sim,
				     sim_entry->pdu->variant.simulatorPresent.simulator)))
	{
	    float32 other_load = 
	      sim_entry->pdu->variant.simulatorPresent.simulationLoad;

	    if (other_load < lowest_load)
	    {
		lowest_load = other_load;
		lowest_load_entry = sim_entry;
	    }
	}
    }

    return lowest_load_entry;
}
