#ifndef lint
static char rcsid [] = "$RCSfile: po_netws.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_netws.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle network worldState                         *
*       Created: Wed May  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <libtime.h>
#include <stdext.h> /* common/include/global */

/*ARGSUSED*/
int32 po_process_set_world_state(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    /* Note when this packet came in */
    if ((int32)pdu->variant.setWorldState.clockRate != 0)
      db->received_set_world_state_time = time_realtime_clock();
    else
      db->received_set_world_state_time = 0;

    /* Stop retransmitting our own animation packets */
    db->next_set_world_state_time = 0;

    /* Tell the application */
    callback_fire_event(db->animation_event, pdu);

    return TRUE;
}

int32 po_start_network_animation(
    PO_DATABASE *db,
    ObjectID    *world_state,
    uint32       secondsSince1970,
    float32      clock_rate)
{
    /* Can we do animation? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    /* Assume other players will stop when they hear us */
    db->received_set_world_state_time = 0;

    /* Remember that we have to retransmit this in a while */
    db->next_set_world_state_time = 
      time_realtime_clock() + setWorldStateTransmitTime*1000;

    /* Tell the network to go to this state */
    return po_send_set_world_state_pdu(db, world_state,
				       secondsSince1970, clock_rate);
}

int32 po_stop_network_animation(
    PO_DATABASE *db,
    ObjectID    *world_state,
    uint32       secondsSince1970)
{
    /* Assume other players will stop when they hear us */
    db->received_set_world_state_time = 0;

    /* Send out a sws packet with magic clock rate of 0.0 */
    (void) po_start_network_animation(db, world_state, secondsSince1970, 0.0);

    /* No need to retansmit */
    db->next_set_world_state_time = 0;

    return TRUE;
}

void po_tick_animation(
    PO_DATABASE *db)
{
    uint32 time;
    uint32 dtime;

    /* Is anybody doing animation? */
    if (!db->received_set_world_state_time &&
	!db->next_set_world_state_time)
      return;

    time = time_realtime_clock();

    /* Is somebody on the net doing animation? */
    if (db->received_set_world_state_time)
    {
	if (time >
	    db->received_set_world_state_time+setWorldStateTimeoutTime*1000)
	{
	    callback_fire_event(db->animation_timeout_event);
	    db->received_set_world_state_time = 0;
	}
    }

    /* Are we doing animation? */
    if (db->next_set_world_state_time)
    {
	/* Is it time to send a packet? */
	if (time > db->next_set_world_state_time)
	{
	    /* Increment the seconds factor */
	    dtime = (time - db->next_set_world_state_time +
		     setWorldStateTransmitTime*1000);
	    dtime = (float64)dtime / 1000.0 *
	      db->set_world_state_pdu->variant.setWorldState.clockRate;

	    db->next_set_world_state_time =
	      time+setWorldStateTransmitTime*1000;

	    if (!po_resend_set_world_state_pdu(db, dtime))
	      db->next_set_world_state_time = 0;
	}
    }
}

