#ifndef lint
static char rcsid [] = "$RCSfile: po_objects.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_objects.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to manage objects                                    *
*       Created: Tue Apr  9 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* Define this to get a message whenever something times out.
 */

#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <po_size.h>
#include <libtime.h>
#include <libqueue.h>
#include <stdext.h> /* common/include/global */
#include <stdstring.h> /*common/include/global*/
#include <stdlib.h>

/*VARARGS5*/
PO_DB_ENTRY *po_create_object(
    PO_DATABASE       *db,
    ObjectID          *world_state,
    uint8              class,
    uint32             missing_from_world_state,
    ADDRESS            variants,
    uint32             vsize,
    PO_USER_DATA_TYPE  object_user_data)
{
    PO_DB_ENTRY *ws;
    ObjectID new_id;

    /* Can you add objects to this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    /* Do not create world state objects with this function.  Use
     * po_create_world_state instead.  Do not create stealth
     * controller objects with this function.  Use
     * po_new_stealth instead.
     */
    if ((class == objectClassWorldState) ||
	(class == objectClassStealthController))
      PO_ERROR(PO_ECLASS);

    /* Overlays only exist in the real-time world state */
    if (class == objectClassOverlay)
    {
	if (!world_state)
	  world_state = &po_real_time_world_state;
	else if (!SAME_OBJECT_ID(*world_state, po_real_time_world_state))
	  PO_ERROR(PO_ERTWS);
    }
    else if (!world_state)
      world_state = &db->current_world_state_id;

    /* Check that the world state is actually a world state */
    if (!SAME_OBJECT_ID(*world_state, po_real_time_world_state))
    {
	ws = po_get_object_for_world_state(db, world_state,
					   &po_real_time_world_state);
/*changed class to pclass 3 Nov 97 TRANG*/
	if (!ws ||
	    (ws->pdu->kind != describeObjectPDUKind) ||
	    (ws->pdu->variant.describeObject.pclass != objectClassWorldState))
	  PO_ERROR(PO_EUKWS);
    }

    /* After normal error checking, we can go try to find an object ID. */
    po_choose_object_id(db, &new_id, world_state);

    /* Did we get an ID? */
    if (!new_id.object)
      PO_ERROR(PO_EOBJID);

    return po_send_new_describe_object_pdu(db, &new_id, world_state, class,
					   missing_from_world_state,
					   variants, vsize,
					   object_user_data);
}

PO_DB_ENTRY *po_create_world_state(
    PO_DATABASE *db,
    ObjectID    *base_state,
    char        *description,
    uint32       secondsSince1970)
{
    ObjectID new_id;
    PO_DB_ENTRY *entry;
    char buf[NETWORK_BUFFER_SIZE];
    WorldStateClass *variant = (WorldStateClass *)buf;

    /* Can you add objects to this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    /* First check the passed base state */
    if (base_state)
    {
	if ((entry = po_find_entry(db, base_state,
				   &po_real_time_world_state)) &&
	    !entry->deleted && !entry->implied)
	{
	    if ((entry->pdu->kind != describeObjectPDUKind) ||
		(entry->pdu->variant.describeObject.pclass !=
		 objectClassWorldState))
	      PO_ERROR(PO_EBASE);
	}
	else
	  PO_ERROR(PO_EBASE);
    }

    po_choose_world_state_id(db, &new_id);

    /* Did we get an ID? */
    if (!new_id.object)
      PO_ERROR(PO_EOBJID);

    /* Fill in user-supplied data */
    (void) strncpy((char *)variant->description, description,
		   maxWorldStateDescriptionLength);
    variant->description[maxWorldStateDescriptionLength-1] = 0;
    variant->secondsSince1970 = secondsSince1970;

    /* New history is base history + base itself */
    if (base_state)
    {
	variant->historyCount = 1 +
	  entry->pdu->variant.describeObject.variant.worldState.historyCount;
	bcopy((char *)entry->pdu->variant.describeObject.variant.worldState.history,
	      (char *)variant->history,
	      (unsigned int) ((variant->historyCount-1)*sizeof(ObjectID)));
	variant->history[variant->historyCount-1] = *base_state;
    }
    else
    {
	variant->historyCount = 0;
    }

    return po_send_new_describe_object_pdu(db, &new_id,
					   &po_real_time_world_state,
					   objectClassWorldState,
					   (int32)FALSE, (char *)variant,
			 PRO_PO_WORLD_STATE_CLASS_SIZE(variant->historyCount),
					   NULL);
}

/*VARARGS2*/
PO_DB_ENTRY *po_change_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    ADDRESS      variants,
    uint32       vsize)
{
    /* Can you change objects in this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    if (entry->pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    /* Do not change stealth controller objects with this function.  Use
     * po_obtain_stealth instead.
     */
    if (entry->pdu->variant.describeObject.pclass ==
	objectClassStealthController)
      PO_ERROR(PO_ECLASS);

    return po_send_changed_describe_object_pdu(db, entry,
             (uint32) entry->pdu->variant.describeObject.missingFromWorldState,
					       variants, vsize);
}

PO_DB_ENTRY *
  po_change_entry_missing_flag(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       missing_from_world_state)
{
    /* Can you change objects in this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    if (entry->pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    /* Do not change stealth controller objects with this function.  Use
     * po_obtain_stealth instead.
     */
    if (entry->pdu->variant.describeObject.pclass ==
	objectClassStealthController)
      PO_ERROR(PO_ECLASS);

    return po_send_changed_describe_object_pdu(db, entry,
					       missing_from_world_state,
					       (char *)&entry->pdu->
					       variant.describeObject.variant,
					       entry->pdu_size -
					       PRO_PO_DESC_OBJ_SIZE(entry->
								    pdu));
}

/*VARARGS4*/
static PO_DB_ENTRY *
po_change_object_and_flag_into_ws(
    PO_DATABASE *db,
    PO_DB_ENTRY *base_entry,
    ObjectID    *into_ws,
    uint32       missing_from_world_state,
    ADDRESS      variants,
    uint32       vsize)
{
    ObjectID object_id;
    ObjectID world_state_id;
    PO_DB_ENTRY *entry;

    /* Can you change objects in this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    if (base_entry->pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    /* Do not change stealth controller objects with this function.  Use
     * po_obtain_stealth instead.
     */
    if (base_entry->pdu->variant.describeObject.pclass ==
	objectClassStealthController)
      PO_ERROR(PO_ECLASS);

    /* We want to change the entry for this world state.  First check
     * the easy case where the entry is the one we want.
     */
    object_id = base_entry->pdu->variant.describeObject.objectID;
    world_state_id = base_entry->pdu->variant.describeObject.worldStateID;

    if (SAME_OBJECT_ID(world_state_id, *into_ws))
      return po_send_changed_describe_object_pdu(db, base_entry,
						 missing_from_world_state,
						 variants, vsize);

    /* If that fails, see if the one we want already exists. */
    if (entry = po_find_entry(db, &object_id, into_ws))
    {
	if (entry->deleted)
	  PO_ERROR(PO_EEXIST);
	return po_send_changed_describe_object_pdu(db, entry,
						   missing_from_world_state,
						   variants, vsize);
    }

    /* We need to create a new entry in order to make this change happen. */
    return po_send_new_describe_object_pdu(db,
					   &object_id,
					   into_ws,
					   base_entry->pdu->
					   variant.describeObject.pclass,
					   missing_from_world_state,
					   variants, vsize,
					   base_entry->object_user_data);
}

/*VARARGS3*/
static PO_DB_ENTRY *
po_change_object_and_flag(
    PO_DATABASE *db,
    PO_DB_ENTRY *base_entry,
    uint32       missing_from_world_state,
    ADDRESS      variants,
    uint32       vsize)
{
    if (PO_REAL_TIME_CLASS(base_entry->pdu->variant.describeObject.pclass))
      return po_change_object_and_flag_into_ws(db, base_entry,
					       &po_real_time_world_state,
					       missing_from_world_state,
					       variants, vsize);
    else
      return po_change_object_and_flag_into_ws(db, base_entry,
					       &db->current_world_state_id,
					       missing_from_world_state,
					       variants, vsize);
}

/*VARARGS2*/
PO_DB_ENTRY *
po_change_object(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    ADDRESS      variants,
    uint32       vsize)
{
    return po_change_object_and_flag(db, entry,
				     (uint32)entry->pdu->variant.
				     describeObject.missingFromWorldState,
				     variants, vsize);
}

PO_DB_ENTRY *
po_change_object_missing_flag(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       missing_from_world_state)
{
    return po_change_object_and_flag(db, entry,
				     missing_from_world_state,
				     &entry->pdu->
				      variant.describeObject.variant,
				     entry->pdu_size -
				      PRO_PO_DESC_OBJ_SIZE(entry->pdu));
}

PO_DB_ENTRY *
po_copy_object_into_ws(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    ObjectID    *into_ws)
{
    return
      po_change_object_and_flag_into_ws(db, entry,
					into_ws,
					(uint32)entry->pdu->variant.
					  describeObject.
					  missingFromWorldState,
					&entry->pdu->
					  variant.describeObject.variant,
					entry->pdu_size -
					 PRO_PO_DESC_OBJ_SIZE(entry->pdu));
}

/*ARGSUSED*/
void po_set_object_user_data(
    PO_DATABASE       *db,
    PO_DB_ENTRY       *entry,
    PO_USER_DATA_TYPE  val)
{
    PO_DB_ENTRY *obj;

    /* Some entries (like sims) do not thread by object, since there cannot
     * be more than one of them.
     */
    if (!entry->object_handle)
    {
	entry->object_user_data = val;
	return;
    }

    for(obj=entry->object_handle->object_head;obj;obj=obj->next_object)
      obj->object_user_data = val;
}

void po_take_ownership(
    PO_DATABASE       *db,
    int32              n_addrs,
    SimulationAddress *addrs)
{
    int32 i, j;
    PO_DB_ENTRY *entry;

    /* Run through all the packets.  For those owned by the named addr,
     * send out packets declaring ownership.
     */
    for (i=0;i<PO_HASH_TABLE_SIZE;i++)
    {
	for (entry=db->entries[i];entry;entry=entry->next_hashed)
	  if (!entry->deleted && !entry->implied &&
	      (entry->pdu->kind == describeObjectPDUKind))
	    for (j=0;j<n_addrs;j++)
	    {
		if (SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
				   addrs[j]))
		  po_send_describe_object_pdu(db, entry);
	    }
    }
}

void po_not_dormant(
    PO_DB_ENTRY *entry,
    PO_DATABASE *db)
{
    if (entry->dormant)
    {
	po_remove_object_present(db, entry);
	entry->dormant = FALSE;
	entry->retransmit_start = 0;
    }
}

void po_nothing_dormant(
    PO_DATABASE *db)
{
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_not_dormant, db);
}

static void send_if_ov_or_ws(
    PO_DB_ENTRY *entry,
    PO_DATABASE *db)
{
    if ((entry->pdu->kind == describeObjectPDUKind) &&
	((entry->pdu->variant.describeObject.pclass == objectClassWorldState) ||
	 (entry->pdu->variant.describeObject.pclass == objectClassOverlay)))
      po_send_oob_describe_object_pdu(db, entry);
}

void po_send_overlays_and_world_states(
    PO_DATABASE *db)
{
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)send_if_ov_or_ws, db);
}

void po_tick_objects(
    PO_DATABASE *db)
{
    PO_DB_ENTRY *entry;
    uint32 time = time_realtime_clock();
    float64 fraction;
    int32 min_send;

    /* We play a neat game with dormant objects here.  First, we retransmit
     * packets before checking for new dormancy, which gives us a 30 second
     * slack before we have to send the object's id in an objectsPresentPDU.
     * Next, if the object is dormant, we add its ID to the PDU being built
     * in its world state entry.  Before we add it, we check if its ID is
     * at the beginning of the packet.  If it is, we know that it has been
     * thirty seconds, and we send the packet out.  Using this trick, we
     * do not need to keep a queue of outgoing objectsPresentPDUs.
     * Pretty neat huh?
     */

    /* In order to evenly distribute packet load on the network, we will
     * send no fewer than time-since-last-tick/retansmission-rate objects
     * each tick.  We round this number down to the nearest integer, and
     * save the fraction for next time (this will even traffic when the
     * number of objects is small).
     */
    fraction = db->retransmit_last_fraction +
      ((float64)db->retransmit_size *
       (float64)(time - db->last_tick_time) /
       (float64)(describeObjectTransmitTime*1000.0));
    min_send = fraction;
    db->retransmit_last_fraction = fraction - min_send;
    db->last_tick_time = time;

    while(db->retransmit_head &&
	  ((db->retransmit_head->time+describeObjectTransmitTime*1000<time) ||
	   (min_send > 0)))
    {
	min_send--;
	entry = db->retransmit_head;
	if (!entry->dormant)
	{
	    /* Sending a describe object will push this object to the end
	     * of the time queue, thus revealing the next object in line.
	     */
	    po_send_describe_object_pdu(db, entry);
	}
	else
	{
	    /* Push to the end of the queue by hand */
	    po_describe_object_sent(db, entry);
	}

	/* Is this the first tick of the rest of this object's life? :-) */
	if (!entry->dormant && !entry->retransmit_start)
	{
	    entry->retransmit_start = time;
	    continue;
	}

	/* Did this object JUST become dormant? */
	if (entry->retransmit_start+describeObjectRetransmitTime*1000 <= time)
	{
	    entry->dormant = TRUE;
	}

	if (entry->dormant)
	  po_add_object_present(db, entry);
    }
}

void po_timeout_objects(
    PO_DATABASE *db)
{
    uint32 time = time_realtime_clock();

    while(db->timeout_head &&
	  db->timeout_head->time+describeObjectTimeoutTime*1000 < time)
    {
	/* Do not time out UNDELETABLE objects; take them over instead */
	if (PO_UNDELETABLE(db->timeout_head->pdu))
	{
	    /* This will pop the object off the timeout queue and onto
	     * the retransmit queue, hence exposing the next item for the
	     * while loop.
	     */
	    po_send_describe_object_pdu(db, db->timeout_head);
	    continue;
	}

#ifdef WARN_TIMEOUTS
	{
	    PO_DB_ENTRY *deleted_obj = db->timeout_head;
	    ObjectID *id = &deleted_obj->pdu->variant.describeObject.objectID;
	    int32 sequence = 
	      deleted_obj->pdu->variant.describeObject.sequenceNumber;
	    SimulationAddress *owner_addr = &deleted_obj->pdu->variant.describeObject.owner;
	    char *owner = po_entry_owner(db, deleted_obj);

	    printf("Timing out obj %d/%d/%d, sequence %d, owner %s (%d/%d)\n",
		   id->simulator.site,
		   id->simulator.host,
		   id->object,
		   sequence,
		   owner,
		   owner_addr->site,
		   owner_addr->host);
	}
#endif /*WARN_TIMEOUTS*/

	/* Tell the application that this guy is history */
	po_handle_entry_gone(db, db->timeout_head);

	/* Take him out of our database. This will pop the object off
	 * the timeout queue, hence exposing the next item for the
	 * while loop.
	 */
	po_destroy_entry(db, db->timeout_head);
    }
}

static PO_DB_ENTRY *po_present_is_known(
    PO_DATABASE       *db,
    ObjectID          *obj_id,
    ObjectID          *ws_id,
    SimulationAddress *owner,
    uint16             seq)
{
    PO_DB_ENTRY *entry;

    /* This looks up an ID in our database and finds out whether we
     * agree with his description of it (same owner, same or higher
     * sequence number, etc.).  Note that if the ID belongs to someone
     * we tried to delete, po_get_entry will say we've never heard of
     * him.  We will then ask for the full info about the object, at
     * which point we will tell the sender to delete it.
     */

    if ((entry = po_get_entry(db, obj_id, ws_id)) &&
	(entry->pdu->kind == describeObjectPDUKind) &&
	SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
		       *owner) &&
	entry->pdu->variant.describeObject.sequenceNumber >= seq)
      return entry;
    else
      return NULL;
}

/*ARGSUSED*/
int32 po_process_object_request(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    PO_DB_ENTRY *entry;
    int32 i;
    ObjectID ws;
    int32 count;

    /* Only pay attention if they are asking us */
    if (!SAME_SIMULATOR(pdu->variant.objectRequest.objectOwner, db->sim_addr))
      return TRUE;

    ws = pdu->variant.objectRequest.worldStateID;
    count = pdu->variant.objectRequest.objectCount;
    
    /* Undormantize these guys to clue the asker in */
    for (i=0;i<count;i++)
    {
	if (entry = po_find_entry(db,
				  &pdu->variant.objectRequest.objects[i],
				  &ws))
	{
	    po_not_dormant(entry, db);
	}
    }
    return TRUE;
}

/*ARGSUSED*/
int32 po_process_objects_present(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    int32 i, new_count;
    PO_DB_ENTRY *entry;
    ObjectID new[maxObjectRequestCount];
    ObjectID ws;
    int32 count;
    SimulationAddress owner;

    /* Very young active database should ignore this packet, since they
     * will automatically trigger describeObject PDUs with their first
     * simulator present PDU.
     */
    if ((db->db_type == PO_DATABASE_TYPE_ACTIVE) &&
	(!db->db_create_time	/* Have not yet ticked */ ||
	 (time_realtime_clock() <
	  (db->db_create_time +
	   (simulatorPresentTransmitTime+describeObjectTransmitTime) * 1000))))
      return TRUE;

    /* Cache the repeatedly used stuff */
    ws = pdu->variant.objectsPresent.worldStateID;
    count = pdu->variant.objectsPresent.objectCount;
    owner = pdu->variant.objectsPresent.owner;

    /* For each object identified, either push it to the back of the
     * timeout queue (with po_describe_object_sent) or stuff it into
     * a list of ids we want to know more about.
     */
    new_count = 0;
    for (i=0;i<count;i++)
    {
	if (entry = po_present_is_known(db,
					&pdu->variant.objectsPresent.
					objects[i].objectID,
					&ws, &owner,
					pdu->variant.objectsPresent.
					objects[i].sequenceNumber))
	{
	    po_describe_object_sent(db, entry);
	}
	else
	{
	    new[new_count] = pdu->variant.objectsPresent.objects[i].objectID;
	    new_count += 1;
	}
    }
    if (new_count)
      po_send_object_request_pdu(db, &owner, &ws, new_count, new);

    return TRUE;
}
	
PO_DB_ENTRY *
po_process_describe_object(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size,
    PO_USER_DATA_TYPE    object_user_data)
{
    PO_DB_ENTRY *entry;
    PersistentObjectPDU *new_pdu;
    PO_DB_ENTRY *deleted_entry;
    int32 change_failed_flag = FALSE;
    uint32 his_seq;
    uint32 my_seq;
    int32 variant_changed;
    int32 previously_missing;
    int32 loopback;

    /* Check for loopback.  Only loopback would send us a packet
     * which identifies this simulator as the owner.
     */
    loopback = SAME_SIMULATOR(pdu->variant.describeObject.owner,
			      db->sim_addr);

    if (!loopback)
    {
	/* Check database sequence numbers first */
	his_seq = pdu->variant.describeObject.databaseSequenceNumber;
	my_seq = db->database_sequence_number;

	/* "A simulator should immediately adopt the database
	 * sequence number in the first Simulator Present or
	 * Describe Object PDU it hears, provided that number is
	 * greater than the number it currently believes.
	 * Thereafter, only Simulator Present PDUs should be used
	 * to inherit new database sequence numbers."
	 */
	if (!db->heard_sequence_from_net)
	{
	    if (his_seq > my_seq)
	      po_set_database_sequence_number(db,
					      (PO_DB_ENTRY *)NULL, his_seq);
	    db->heard_sequence_from_net = TRUE;
	}

	/* "All objects with database numbers less than this number
	 * are invalid."
	 */
	if (his_seq < my_seq)
	  return NULL;
    }

    /* Check if we already know this guy */
    if (entry = po_find_matching_entry(db, pdu))
    {
	/* Make sure we don't want this guy dead */
	if (entry->deleted)
	{
	    po_send_delete_objects_pdu(db, (int32)1, &entry);
	    return entry;
	}

	/* When we hear a packet off the net about an entry which we
	 * have created as implied, always just accept the new
	 * info.
	 */
	if (entry->implied)
	{
	    /* Note the time of receipt */
	    entry->time = time_realtime_clock();
	    /* fall through */
	}
	else
	{
	    /* Note the time of tranmission/receipt */
	    po_describe_object_sent(db, entry);

	    if (loopback)
	    {
		/* Ensure we own the object. */
		if (!SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
				    db->sim_addr))
		  po_make_local(db, entry);

		/* fall through */
	    }
	    else		/* This is by spec */
	    {
		/* "A simulator should disregard information received from
		 * the network if the sequence number is lower than that
		 * currently stored in the simulator's memory."
		 */

		if (pdu->variant.describeObject.sequenceNumber <
		    entry->pdu->variant.describeObject.sequenceNumber)
		  return entry;

		/* "If a simulator receives a PDU regarding an object
		 * which it currently owns the new information may or may
		 * not be used, as follows:"
		 */

		if (SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
				   db->sim_addr))
		{

		    /* o "If the received PDU has a sequence number higher
		     *   than that currently known to the simulator, the
		     *   simulator must relinquish ownership of the object
		     *   by accepting the new information."
		     *
		     * o "If the received PDU has a sequence number equal
		     *   to that currently known to the simulator, the
		     *   simulator will compare its own simulator address
		     *   magnitude (MAM = host << 16 | site) to that of
		     *   the simulator claiming ownership, and if higher,
		     *   will increment the sequence number and retransmit
		     *   the currently stored information."
		     */

		    if (pdu->variant.describeObject.sequenceNumber ==
			entry->pdu->variant.describeObject.sequenceNumber)
		    {
			if (SIM_ADDR_MAG(db->sim_addr) >
			    SIM_ADDR_MAG(pdu->variant.describeObject.owner))
			{
			    entry->pdu->variant.describeObject.sequenceNumber++;
			    po_send_describe_object_pdu(db, entry);
			    return entry;
			}
			else
			{
			    /* If we get here, it means somebody else changed
			     * this object at the same time we did, and we
			     * lose.
			     */
			    po_make_remote(db, entry);
			    change_failed_flag = TRUE;
			}
		    }
		    else
		    {
			po_make_remote(db, entry);
		    }
		}

		/* If a simulator receives a PDU regarding an object
		 * which it does not own, it should ignore the PDU if
		 * the sequence number is less than that currently
		 * known, or if the sequence numbers are the same and
		 * the owners are the same.  Otherwise, if the
		 * sequence number is greater than that currently
		 * known or the sequence numbers are the same and the
		 * owner has changed, it should take the new
		 * information.
		 */

		else
		{
		    if ((pdu->variant.describeObject.sequenceNumber <
			 entry->pdu->variant.describeObject.sequenceNumber) ||
			((pdu->variant.describeObject.sequenceNumber ==
			  entry->pdu->variant.describeObject.sequenceNumber) &&
			 SAME_SIMULATOR(pdu->variant.describeObject.owner,
					entry->pdu->variant.describeObject.owner)))
		      return entry;
		    else
		      ;		/* Fall through */
		}

	    }
	}

        /* Note whether the variant has changed.  For tasks also check
         * whether the sequence number has changed because applications
         * track this value as well.
         */

if ((char *)&entry->pdu->variant.describeObject.variant ==
    (char *)&pdu->variant.describeObject.variant)
  printf("Attempted to change po object without using PO_BUFFER!  (class %d, id %d/%d/%d)\n",entry->pdu->variant.describeObject.pclass, entry->pdu->variant.describeObject.objectID.simulator.site, entry->pdu->variant.describeObject.objectID.simulator.host, entry->pdu->variant.describeObject.objectID.object);

	if (pdu_size != entry->pdu_size)
	  variant_changed = TRUE;
	else
	  variant_changed =
	    ((entry->pdu->variant.describeObject.missingFromWorldState !=
	      pdu->variant.describeObject.missingFromWorldState) ||
	     bcmp((char *)&entry->pdu->variant.describeObject.variant,
		  (char *)&pdu->variant.describeObject.variant,
		  (int) (entry->pdu_size - PRO_PO_DESC_OBJ_SIZE(pdu))) ||
                  (entry->pdu->variant.describeObject.pclass == objectClassTask) &&
                  (entry->pdu->variant.describeObject.sequenceNumber !=
                   pdu->variant.describeObject.sequenceNumber));

	previously_missing =
	  entry->pdu->variant.describeObject.missingFromWorldState;

	if (pdu_size > entry->alloced_size)
	{
	    if (!(new_pdu =
		  (PersistentObjectPDU *)queue_buffer_allocate((unsigned int)pdu_size)))
	    {
		po_destroy_entry(db, entry);
		po_errno = PO_ENOMEM;
		return NULL;
	    }

	    queue_buffer_deallocate((char *)entry->pdu);
	    entry->pdu = new_pdu;
	    /* Normally we would do a queue_buffer_up_refcnt() since
	     * we are assigning * the queue_buffer, and doing a
	     * queue_buffer_deallocate() to match with the
	     * queue_buffer_allocate().  But, the two routines combine
	     * into a no-op, so they are ommitted.
	     */
	    entry->alloced_size = pdu_size;
	}
	bcopy((char *)pdu, (char *)entry->pdu, (unsigned int)pdu_size);
	entry->pdu_size = pdu_size;

	/* We may have just received an object which was only implied before.
	 * If that is the case, we need to tell the application about it
	 * as though it were new.
	 */
	if (entry->implied)
	{
	    /* This is real now */
	    entry->implied = FALSE;

	    /* Put this on the timeout queue. */
	    Q_INSERT(entry, db->timeout_head, db->timeout_tail,
		     prev_time, next_time);

	    po_handle_new_entry(db, entry);

	    return entry;
	}

	if (variant_changed)
	{
	    if (change_failed_flag)
	      callback_fire_event(db->object_change_failed_event, entry);
	    po_handle_entry_changed(db, entry, previously_missing);
	}
	return entry;
    }

    /* It is possible that a new entry actually describes an object
     * we wanted to delete.  However, we didn't know that the object
     * existed at the time of the deletion.  To deal with this case,
     * we have to make an entry in the database so we have a handle
     * on the object, and then immediately delete it.  There is no
     * need to invoke callbacks here.
     */
    if (deleted_entry = po_should_have_been_deleted(db, pdu))
    {
	if (!(entry = po_add_entry_to_db(db, pdu, pdu_size,
					 (int32)FALSE,
					 (PO_USER_DATA_TYPE)NULL)))
	  return NULL;

	/* When we want to delete an object because its world state is
	 * gone, we first need to duplicate it into subsequent world
	 * states.
	 */
	if ((deleted_entry->pdu->kind == describeObjectPDUKind) &&
	    (deleted_entry->pdu->variant.describeObject.pclass ==
	     objectClassWorldState))
	  po_duplicate_for_world_state_deletion(db, deleted_entry, entry);

	(void) po_delete_entries_internal(db, (int32)1, &entry, 
					  (int32)FALSE, (int32)FALSE);
	PO_ERROR(PO_EDELETE);
    }

    /* If this is the first we've heard of this object, check to see if
     * it is valid.  New invalid objects may be ignored.
     */
    if (!po_valid_pdu(db, pdu))
    {
	/* Note the time */
	db->last_invalid_time = time_realtime_clock();

	return NULL;
    }

    /* This object is OK.  Add him to the database and tell the application.
     */
    if (!(entry = po_add_entry_to_db(db, pdu, pdu_size,
				     (int32)FALSE, object_user_data)))
      return NULL;

    po_handle_new_entry(db, entry);

    return entry;
}

PO_DB_ENTRY *po_find_or_invent_world_state(
    PO_DATABASE *db,
    ObjectID    *ws_id)
{
    char buf[NETWORK_BUFFER_SIZE];
    PersistentObjectPDU *pdu = (PersistentObjectPDU *)buf;
    int32 pdu_size;
    PO_DB_ENTRY *entry;

    /* First try to find this world state */
    if (entry = po_find_entry(db, ws_id, &po_real_time_world_state))
    {
	/* Make sure it is a world state, and not some other kind of
	 * object.
	 */
	if (entry->pdu->variant.describeObject.pclass != objectClassWorldState)
	  PO_ERROR(PO_ECLASS);

	return entry;
    }

    /* There is no such world state in the database, so we have to create
     * an implied world state.
     */
    pdu->version = persistentObjectProtocolVersionCurrent;
    pdu->kind = describeObjectPDUKind;
    pdu->exercise = db->exercise_id;
    pdu->database = db->database_id;
    pdu->variant.describeObject.objectID = *ws_id;
    pdu->variant.describeObject.worldStateID = po_real_time_world_state;
    pdu->variant.describeObject.owner = db->sim_addr;
    pdu->variant.describeObject.sequenceNumber = 1;
    pdu->variant.describeObject.pclass = objectClassWorldState;
    pdu->variant.describeObject.missingFromWorldState = FALSE;
    pdu->variant.describeObject.variant.worldState.description[0] = 0;
    pdu->variant.describeObject.variant.worldState.secondsSince1970 = 0;
    pdu->variant.describeObject.variant.worldState.historyCount = 0;

    pdu_size = PRO_PO_DESC_OBJ_SIZE(pdu)+ PRO_PO_WORLD_STATE_CLASS_SIZE(0);

    return po_add_entry_to_db(db, pdu, pdu_size,
			      (int32)TRUE, (PO_USER_DATA_TYPE)NULL);
}
