#ifndef lint
static char rcsid [] = "$RCSfile: po_op_pdu.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_op_pdu.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to send the objects present PDU                      *
*       Created: Thu May 16 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <po_size.h>
#include <libqueue.h>
#include <stdalloc.h> /*common/include/global*/

#define ALLOC_QUANTUM 16

void po_add_object_present(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *ws_entry;
    PersistentObjectPDU *pdu;
    struct po_objects_present *t;
    ObjectID obj_id;
    ObjectID ws_id;
    int32 pdu_size;

    /* This code takes the object ID and sequence number for the entry
     * and puts it into an objects present packet.  When 30 have passed
     * or the packet gets full, the packet is sent.
     */

    obj_id = entry->pdu->variant.describeObject.objectID;
    ws_id = entry->pdu->variant.describeObject.worldStateID;

    /* First check for things which live in the Real Time world state like
     * world states and overlays.
     */
    if (SAME_OBJECT_ID(ws_id, po_real_time_world_state))
    {
	ws_entry = po_find_entry(db,
				 &po_real_time_world_state,
				 &po_real_time_world_state);
    }
    else
      ws_entry = entry->world_state_head;

    /* Make the objects present tracking structure if this is the
     * first time it is being used
     */
    if (!ws_entry->objects_present)
    {
	if (!(ws_entry->objects_present = (struct po_objects_present *)
	      STDALLOC(sizeof(struct po_objects_present))))
	{
	    po_errno = PO_ENOMEM;
	    return;
	}
	bzero((char *)ws_entry->objects_present,
	      (unsigned int) sizeof(struct po_objects_present));

	if (!(ws_entry->objects_present->pdu =
	      (PersistentObjectPDU *)
	      queue_buffer_allocate(NETWORK_BUFFER_SIZE)))
	{
	    po_errno = PO_ENOMEM;
	    return;
	}
	pdu = ws_entry->objects_present->pdu;
	bzero((char *)pdu, (unsigned int)NETWORK_BUFFER_SIZE);

	/* Normally we would do a queue_buffer_up_refcnt() since
	 * we are assigning the queue_buffer, and doing a
	 * queue_buffer_deallocate() to match with the
	 * queue_buffer_allocate().  But, the two routines combine
	 * into a no-op, so they are ommitted.
	 */
	pdu->version = persistentObjectProtocolVersionCurrent;
	pdu->kind = objectsPresentPDUKind;
	pdu->exercise = db->exercise_id;
	pdu->database = db->database_id;
	pdu->variant.objectsPresent.owner = db->sim_addr;
	pdu->variant.objectsPresent.worldStateID = ws_id;
    }
    t = ws_entry->objects_present;

    /* If our ID is at the front of the list, then it must have been
     * sitting around for the last 30 seconds.  Better send it!
     */
    if (t->used && SAME_OBJECT_ID(t->objects[0].objectID, obj_id))
    {
	int32 start;
	int32 this;

	/* Send out a bunch of PDUs */
	for (start=0; start < t->used; start += maxObjectsPresentCount)
	{
	    /* Take as many as we can */
	    this = t->used - start;
	    if (this > maxObjectsPresentCount)
	      this = maxObjectsPresentCount;

	    /* Copy them into the PDU */
	    t->pdu->variant.objectsPresent.objectCount = this;
	    bcopy((char *)&t->objects[start],
		  (char *)t->pdu->variant.objectsPresent.objects,
		  this * sizeof(ObjectIDSequencePair));

	    /* Fire off the PDU */
	    pdu_size = PRO_PO_OBJ_PRES_SIZE(t->pdu, this);
	    t->pdu->length =  NS_HTON16(pdu_size);
	    db->send_handler(t->pdu, pdu_size, db->send_user_data);
	    db->packets_sent++;
	}

	/* Start over at the beginning */
	t->used = 0;
    }

    /* Make room for the ID */
    if (t->alloced == t->used)
    {
	t->alloced += ALLOC_QUANTUM;
	if (!t->objects)
	  t->objects = (ObjectIDSequencePair *)
	    STDALLOC(t->alloced * sizeof(ObjectIDSequencePair));
	else
	  t->objects = (ObjectIDSequencePair *)
	    STDREALLOC(t->objects, t->alloced * sizeof(ObjectIDSequencePair));
    }

    /* Stuff us into the packet */
    t->objects[t->used].objectID = obj_id;
    t->objects[t->used].sequenceNumber =
      entry->pdu->variant.describeObject.sequenceNumber;

    /* Up the count */
    t->used += 1;
}

void po_remove_object_present(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *ws_entry;
    struct po_objects_present *t;
    ObjectID obj_id;
    ObjectID ws_id;
    int32 i, state;

    /* When a dormant object is deleted, we may have to pull it out of
     * the objects present PDU.
     */

    obj_id = entry->pdu->variant.describeObject.objectID;
    ws_id = entry->pdu->variant.describeObject.worldStateID;

    /* First check for things which live in the Real Time world state like
     * world states and overlays.
     */
    if (SAME_OBJECT_ID(ws_id, po_real_time_world_state))
    {
	ws_entry = po_find_entry(db,
				 &po_real_time_world_state,
				 &po_real_time_world_state);
    }
    else
      ws_entry = entry->world_state_head;

    if (!ws_entry || !ws_entry->pdu || !ws_entry->objects_present)
      return;

    t = ws_entry->objects_present;

#define SEARCHING 0
#define SHIFTING  1

    /* Find this object in the list and squeeze it out! */

    for (i=0,state=SEARCHING;i<t->used;i++)
    {
	/* When we find this guy, start shifting those to his right
	 * leftward.
	 */
	if ((state==SEARCHING) &&
	    SAME_OBJECT_ID(t->objects[i].objectID, obj_id))
	  state = SHIFTING;
	else if (state == SHIFTING)
	  t->objects[i-1] = t->objects[i];
    }

    /* Did we find it? */
    if (state != SEARCHING)
      t->used -= 1;
}
