#ifndef lint
static char rcsid [] = "$RCSfile: po_overlay.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_overlay.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code dealing with objectClassOverlay                      *
*       Created: Tue Apr 16 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h> /* common/include/global */

int32 po_get_overlay_id(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    ObjectID            *id_ret)
{
    PO_DB_ENTRY *entry;

    if (pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol which has an overlay
     * field, add a case to this switch.  For non-overlay-grouped classes,
     * there is no need to change this switch.
     */

    switch (pdu->variant.describeObject.pclass)
    {
      case objectClassPoint:
	*id_ret = pdu->variant.describeObject.variant.point.overlayID;
	return TRUE;
      case objectClassLine:
	*id_ret = pdu->variant.describeObject.variant.line.overlayID;
	return TRUE;
      case objectClassChemical:
	*id_ret = pdu->variant.describeObject.variant.chemical.overlayID;
	return TRUE;
      case objectClassMinefieldParent:
	*id_ret = pdu->variant.describeObject.variant.minefieldParent.overlayID;
	return TRUE;
      case objectClassMinefield:
	/* To find the overlay a Minefield object is in, find the overlay its
	 * parent is in 
	 */
	entry = 
	  po_get_entry(db, 
		       &pdu->variant.describeObject.variant.minefield.parentID,
		       &pdu->variant.describeObject.worldStateID);
	if(entry && (entry->pdu->variant.describeObject.pclass == 
		     objectClassMinefieldParent))
	  return(po_get_overlay_id(db, entry->pdu, id_ret));
	else
	{
	    *id_ret = po_real_time_world_state;
	    return FALSE;
	}
      case objectClassSector:
	*id_ret = pdu->variant.describeObject.variant.sector.overlayID;
	return TRUE;
      case objectClassText:
	*id_ret = pdu->variant.describeObject.variant.text.overlayID;
	return TRUE;
      case objectClassUnit:
	*id_ret = pdu->variant.describeObject.variant.unit.overlayID;
	return TRUE;
      case objectClassLink:
	*id_ret = pdu->variant.describeObject.variant.link.overlayID;
	return TRUE;
      default:
	*id_ret = po_real_time_world_state;
	return FALSE;
    }
}

PO_DB_ENTRY *
po_find_overlay(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    ObjectID overlay;
    PO_DB_ENTRY *overlay_entry;

    if (!po_get_overlay_id(db, entry->pdu, &overlay))
      return NULL;

    overlay_entry = po_get_entry(db, &overlay, &po_real_time_world_state);

    if (!overlay_entry)
      return NULL;

    if (overlay_entry->pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);
    if (overlay_entry->pdu->variant.describeObject.pclass !=
	objectClassOverlay)
      PO_ERROR(PO_ECLASS);

    return overlay_entry;
}

uint8 po_object_color(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *overlay_entry;
    uint8 color;

    if (entry->pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol which has a color
     * field, add a case to this switch.  For non-colored classes,
     * there is no need to change this switch.
     */

    switch (entry->pdu->variant.describeObject.pclass)
    {
      case objectClassPoint:
	color = entry->pdu->variant.describeObject.variant.point.color;
	break;
      case objectClassLine:
	color = entry->pdu->variant.describeObject.variant.line.color;
	break;
      case objectClassChemical:
	color = entry->pdu->variant.describeObject.variant.chemical.color;
	break;
      case objectClassMinefieldParent:
	color = entry->pdu->variant.describeObject.variant.minefieldParent.color;
	break;
      case objectClassSector:
	color = entry->pdu->variant.describeObject.variant.sector.color;
	break;
      case objectClassText:
	color = entry->pdu->variant.describeObject.variant.text.color;
	break;
      case objectClassLink:
	color = entry->pdu->variant.describeObject.variant.link.color;
	break;
      default:
	PO_ERROR(PO_ECLASS);
    }

    if (color == OCOverlayDefault)
    {
	if (overlay_entry = po_find_overlay(db, entry))
	{
	    color = PO_OVERLAY_DATA(overlay_entry).color;
	}
	else
	  PO_ERROR(PO_EOVLY);
    }
    return color;
}

void po_clear_change_flags(
    UnitClass *unit)
{
    unit->changeFormation = 0;
    unit->changeLocation = 0;
    unit->changeDirection = 0;
    unit->changeUnitStrength = 0;
    unit->changeAppearance = 0;
}

