#ifndef lint
static char rcsid [] = "$RCSfile: po_process.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_process.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Packet handling process                                   *
*       Created: Mon Apr  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h>

int32 po_process_packet(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    if (!po_good_pdu(pdu, pdu_size))
      return FALSE;

    /* Filter different exercise ids.  Note that libassoc will most likely
     * do this for us via subscription, but this is a safety net.
     */
    if (pdu->exercise != db->exercise_id)
      return TRUE;

    /* Filter different database ids.
     */
    if (pdu->database != db->database_id)
      return TRUE;

    switch(pdu->kind)
    {
      case simulatorPresentPDUKind:
	po_packet_from_simulator(db,
				 &pdu->variant.simulatorPresent.simulator);
	return po_process_simpresent(db, pdu, pdu_size);
      case describeObjectPDUKind:
	po_packet_from_simulator(db,
				 &pdu->variant.describeObject.owner);
	return (po_process_describe_object(db, pdu, pdu_size, (PO_USER_DATA_TYPE)NULL) != NULL);
      case objectsPresentPDUKind:
	po_packet_from_simulator(db,
				 &pdu->variant.objectsPresent.owner);
	return po_process_objects_present(db, pdu, pdu_size);
      case objectRequestPDUKind:
	po_packet_from_simulator(db,
				 &pdu->
				 variant.objectRequest.requestingSimulator);
	return po_process_object_request(db, pdu, pdu_size);
      case deleteObjectsPDUKind:
	po_packet_from_simulator(db,
				 &pdu->
				 variant.deleteObjects.deletingSimulator);
	return po_process_delete_objects(db, pdu, pdu_size);
      case nominationPDUKind:
	po_packet_from_simulator(db,
				 &pdu->
				 variant.nomination.nominatingSimulator);
	return po_process_nomination(db, pdu, pdu_size);
      case setWorldStatePDUKind:
	po_packet_from_simulator(db,
				 &pdu->
				 variant.setWorldState.requestingSimulator);
	return po_process_set_world_state(db, pdu, pdu_size);
      default:
	PO_ERROR(PO_EKIND);
    }
}
