#ifndef lint
static char rcsid [] = "$RCSfile: po_query.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_query.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to query the database from the application level     *
*       Created: Sun Apr 21 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h> /* common/include/global */

PO_DB_ENTRY *po_get_object(
    PO_DATABASE *db,
    ObjectID    *object_id)
{
    PO_DB_ENTRY *entry;

    entry = po_current_object(db, object_id);

    if (!entry || entry->pdu->variant.describeObject.missingFromWorldState)
      return NULL;

    return entry;
}

PO_DB_ENTRY *po_get_entry(
    PO_DATABASE *db,
    ObjectID    *object_id,
    ObjectID    *world_state_id)
{
    PO_DB_ENTRY *entry;

    entry = po_find_entry(db, object_id, world_state_id);
    if (!entry || entry->deleted || entry->implied)
      return NULL;

    return entry;
}

char *po_entry_owner(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *sim_entry;

    switch(entry->pdu->kind)
    {
      case simulatorPresentPDUKind:
	return ((char *) entry->pdu->variant.simulatorPresent.hostname);
      case describeObjectPDUKind:
	sim_entry =
	  po_find_simulator(db, &entry->pdu->variant.describeObject.owner);

	if (!sim_entry)
	  return "Unknown Simulator";
	else
	  return ((char *) sim_entry->pdu->variant.simulatorPresent.hostname);
      default:
	PO_ERROR(PO_EKIND);
    }
}

/*VARARGS1*/
void po_query_for_current_objects(
    PO_DATABASE       *db,
    PO_QUERY_HANDLER   query_handler,
    PO_USER_DATA_TYPE  query_user_data)
{
    PO_OBJECT_HANDLE *handle;

    for(handle=db->object_handle_head;handle;handle=handle->next_handle)
      if (handle->current_entry &&
	  !handle->current_entry->implied &&
	  !handle->current_entry->deleted &&
	  !handle->current_entry->pdu->
	  variant.describeObject.missingFromWorldState)
	query_handler(handle->current_entry, query_user_data);
}

/*VARARGS1*/
void po_query_for_all_entries(
    PO_DATABASE       *db,
    PO_QUERY_HANDLER   query_handler,
    PO_USER_DATA_TYPE  query_user_data)
{
    PO_OBJECT_HANDLE *handle;
    PO_DB_ENTRY *entry;

    for(handle=db->object_handle_head;handle;handle=handle->next_handle)
      for(entry=handle->object_head;entry;entry=entry->next_object)
	if (!entry->implied && !entry->deleted)
	  query_handler(entry, query_user_data);
}

/*VARARGS1*/
void po_query_for_all_simulators(
    PO_DATABASE       *db,
    PO_QUERY_HANDLER   query_handler,
    PO_USER_DATA_TYPE  query_user_data)
{
    PO_DB_ENTRY *sim_entry;

    /* Call the query handler for all simulators including yourself. */
    for (sim_entry = db->sim_class_head; sim_entry;
         sim_entry = sim_entry->next_class)
    {
        query_handler(sim_entry, query_user_data);
    }
}

/*VARARGS1*/
void po_query_for_current_class(
    PO_DATABASE           *db,
    PersistentObjectClass  class,
    PO_QUERY_HANDLER       query_handler,
    PO_USER_DATA_TYPE      query_user_data)
{
    PO_DB_ENTRY *entry;

    for(entry=db->object_classes[PO_OBJECT_CLASS_INDEX(class)];
	entry;entry=entry->next_class)
      if ((entry==entry->object_handle->current_entry) &&
	  !entry->implied &&
	  !entry->deleted &&
	  !entry->pdu->variant.describeObject.missingFromWorldState)
        query_handler(entry, query_user_data);
}
