#ifndef lint
static char rcsid [] = "$RCSfile: po_sims.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_sims.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to manage simulators                                 *
*       Created: Tue Apr  9 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <po_size.h>
#include <libtime.h>
#include <libqueue.h>
#include <stdext.h> /* common/include/global */

static PO_DB_ENTRY *po_add_simulator_to_db(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    PO_DB_ENTRY *entry;
    
    if (!(entry = po_add_entry_to_db(db, pdu, pdu_size, (int32)FALSE, (PO_USER_DATA_TYPE)NULL)))
      return NULL;

    /* Is this me? */
    if (SAME_SIMULATOR(pdu->variant.simulatorPresent.simulator, db->sim_addr))
      db->own_sim_entry = entry;

    return entry;
}

/*ARGSUSED*/
int32 po_process_simpresent(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    PO_DB_ENTRY *entry;
    PersistentObjectPDU *new_pdu;
    uint32 my_seq, his_seq;
    uint32 received_packets;
    uint32 sent_packets;
    uint32 prev_missed;

    /* If we already know this guy, just save the new info */
    if (entry = po_find_matching_entry(db, pdu))
    {
	/* Note the numbers of packets sent/received */
	received_packets = entry->pdu->variant.simulatorPresent.packetsSent;
	sent_packets = pdu->variant.simulatorPresent.packetsSent;

	if (pdu_size > entry->alloced_size)
	{
	    if (!(new_pdu =
		  (PersistentObjectPDU *)queue_buffer_allocate((unsigned int)pdu_size)))
	    {
		po_destroy_entry(db, entry);
		po_errno = PO_ENOMEM;
		return TRUE;
	    }

	    queue_buffer_deallocate((char *)entry->pdu);
	    entry->pdu = new_pdu;
	    /* Normally we would do a queue_buffer_up_refcnt() since
	     * we are assigning the queue_buffer, and doing a
	     * queue_buffer_deallocate() to match with the
	     * queue_buffer_allocate().  But, the two routines combine
	     * into a no-op, so they are ommitted.
	     */
	    entry->alloced_size = pdu_size;
	}
	bcopy((char *)pdu, (char *)entry->pdu, (unsigned int)pdu_size);
	entry->pdu_size = pdu_size;
	entry->time = time_realtime_clock();

	/* Zero out our packet counting spot */
	entry->pdu->variant.simulatorPresent.packetsSent = 0;	

	/* If he sent more than we received, then that is how many
	 * we missed.  If, however, we think he sent more than he says
	 * he did, we most likely missed his last simulatorPresentPDU,
	 * so up the missed count by one.
	 */
	prev_missed = db->packets_missed;
	if (sent_packets > received_packets)
	{
	    db->packets_missed += (sent_packets - received_packets);
	}
	else if (sent_packets < received_packets)
	{
	    db->packets_missed += 1;
	}

	if (db->packets_missed != prev_missed)
	{
	    /* Note the time of the missed packets */
	    db->last_missed_time = entry->time;

	    /* Tell the user we missed some */
	    callback_fire_event(db->packets_missed_event,
				entry,
				db->packets_missed - prev_missed);
	}
    }
    else
    {
	/* If this is the first we've heard of this sim, add him to the
	 * database and tell the application.
	 */
	entry = po_add_simulator_to_db(db, pdu, pdu_size);

	/* Zero out our packet counting spot */
	entry->pdu->variant.simulatorPresent.packetsSent = 0;	

	if (entry != db->own_sim_entry)
	  callback_fire_event(db->new_simulator_event, entry);

	po_nothing_dormant(db);
	po_send_overlays_and_world_states(db);
    }

    if (entry != db->own_sim_entry)
    {
	/* Check clocks.  Passive simulators follow the eb and flow of
	 * other folks clocks.  Among active simulators, fastest wins.
	 */
	if ((db->db_type == PO_DATABASE_TYPE_PASSIVE) ||
	    (po_time(db) < entry->pdu->variant.simulatorPresent.time))
	  po_set_time(db, entry->pdu->variant.simulatorPresent.time);

	db->heard_sequence_from_net = TRUE;

	/* See if the database sequence number has changed */
	my_seq = db->database_sequence_number;
	his_seq = entry->pdu->variant.simulatorPresent.databaseSequenceNumber;

	/* Easy case first */
	if (his_seq == my_seq)
	  return TRUE;

	/* Correct him? */
	if (his_seq < my_seq)
	{
	    if (db->db_type == PO_DATABASE_TYPE_ACTIVE)
	      po_send_simulator_present_pdu(db);
	    return TRUE;
	}

	/* Uh oh.  We're out of date.  Time to blow everything away and start
	 * over.
	 */
	po_set_database_sequence_number(db, entry, his_seq);
    }
    return TRUE;
}

void po_set_database_sequence_number(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       new_number)
{
    callback_fire_event(db->delete_all_event, entry);

    po_destroy_all_objects(db);

    db->database_sequence_number = new_number;
    *db->database_sequence_number_backup = new_number;

    po_set_time(db, (uint32)0);

    /* Let him know we obliged */
    if (db->db_type == PO_DATABASE_TYPE_ACTIVE)
      po_send_simulator_present_pdu(db);
}

void po_set_time(
    PO_DATABASE *db,
    uint32       new_time)
{
    db->po_time_offset = (int32) (new_time - time_last_simulation_clock);
}

uint32 po_time(
    PO_DATABASE *db)
{
    int32 app_time;

    app_time = (int32) time_last_simulation_clock;
    return (uint32)(app_time + db->po_time_offset);
}

int32 po_delete_all(
    PO_DATABASE *db)
{
    /* Can you do a delete all on this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    po_set_database_sequence_number(db, db->own_sim_entry,
				    db->database_sequence_number+1);

    db->ready = TRUE;

    return TRUE;
}

/* Time out no more than this number each tick. */
#define MAX_TIMED_OUT 100

void po_timeout_sims(
    PO_DATABASE *db)
{
    PO_DB_ENTRY *sim_entry, *next;
    PO_DB_ENTRY *lowest_load_entry;
    SimulationAddress timed_out[MAX_TIMED_OUT];
    int32 timed_out_count = 0;
    uint32 time = time_realtime_clock();

    /* We do this in two phases.  First eliminate the timed out sims,
     * then do nominations, etc.  That way, we avoid nominating a sim
     * which is actually timed out (such as when two workstations crash
     * simultaneously because of some shared bug).
     */

    /* First look for timeouts */
    for(sim_entry = db->sim_class_head;
	sim_entry;
	sim_entry = next)
    {
	next = sim_entry->next_class;

	/* Don't time out yourself */
	if (sim_entry == db->own_sim_entry)
	  continue;

	/* Did he time out? */
	if (sim_entry->time + simulatorPresentTimeoutTime*1000 < time)
	{
	    /* Tell the application that this guy is history */
	    callback_fire_event(db->simulator_gone_event, sim_entry);

	    /* Note that he's gone for nomination phase */
	    timed_out[timed_out_count++] =
	      sim_entry->pdu->variant.simulatorPresent.simulator;

	    /* Take him out of our database */
	    po_destroy_entry(db, sim_entry);

	    if (timed_out_count == MAX_TIMED_OUT)
	      break;
	}
    }

    /* Now do nominations */
    if (timed_out_count)
    {
	/* Do we have to take over these packets? */
	if (db->db_type == PO_DATABASE_TYPE_ACTIVE)
	{
	    lowest_load_entry = po_lowest_loaded_simulator(db, (int32)FALSE);

	    if (lowest_load_entry == db->own_sim_entry)
	      po_take_ownership(db, timed_out_count, timed_out);
	    else
	      po_send_nominate_pdus(db,
				    &lowest_load_entry->pdu->
				     variant.simulatorPresent.simulator,
				    timed_out_count, timed_out);
	}
    }
}

/*ARGSUSED*/
int32 po_process_nomination(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    PO_DB_ENTRY *sim_entry;
    uint32 time = time_realtime_clock();

    /* Am I the one nominated? */
    if (!SAME_SIMULATOR(pdu->variant.nomination.nominatedSimulator,
			db->sim_addr))
      return TRUE;

    /* Is this someone asking for help? */
    if (SAME_SIMULATOR(pdu->variant.nomination.nominatingSimulator,
		       pdu->variant.nomination.missingSimulator))
    {
	/* Oblige */
	po_take_ownership(db, (int32)1, &pdu->variant.nomination.missingSimulator);
	return TRUE;
    }

    /* First test that the missing sim is really missing */
    if ((sim_entry =
	 po_find_simulator(db, &pdu->variant.nomination.missingSimulator)) &&
	(sim_entry->time + simulatorPresentTimeoutTime*1000 > time))
    {
	return TRUE;
    }

    /* Either he's not there or I'm about to time him out myself */
    po_take_ownership(db, (int32)1, &pdu->variant.nomination.missingSimulator);
    return TRUE;
}

char *po_simulator_name(
    PO_DATABASE       *db,
    SimulationAddress *addr)
{
    PO_DB_ENTRY *sim_entry;

    sim_entry = po_find_simulator(db, addr);

    if (!sim_entry)
      return "Unknown Simulator";
    else
      return ((char *) sim_entry->pdu->variant.simulatorPresent.hostname);
}

