
/*
 * po_size.h
 *
 * Moved from p_size.h
 * using emacs and my fingers.
 * 
 */

#ifndef _po_size_included
#define _po_size_included

/*  This file defines macros for determining the size of various PDUs. */
/*  As with all macros (potentially) all bets are off if you use args  */
/*  with side effects.                                                 */

/* These size macros have explicit padding on the end so that if allocating
 * an array of PDU's using malloc (n * XXX_SIZE()), each element in the array
 * is properly aligned.
 */

/* $RCSfile: po_size.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

/*  First, the headers.  These macros take a pointer to the            */
/*  appropriate PDU type.                                              */

#define PRO_PO_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

/*  Then the individual PDU types  */


/* **Persistent Object Protocol (fixed length) - These macros take a pointer*/
/*   to a PersistentObjectPDU.                                           */

#define PRO_PO_SIM_PRESENT_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (SimulatorPresentVariant))

#define PRO_PO_NOMINATE_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (NominationVariant))

#define PRO_PO_DESC_OBJ_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (DescribeObjectVariant) - \
	sizeof ((p)->variant.describeObject.variant))

#define PRO_PO_SET_WS_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (SetWorldStateVariant))

/* **Persistent Object Protocol (variable length) - These macros take a */
/*   pointer to a PersistentObjectPDU and the number of variants to be  */
/*   included.                                                          */

#define PRO_PO_OBJ_PRES_SIZE(p, v) \
	(PRO_PO_HDR_SIZE (p) + sizeof (ObjectsPresentVariant) + \
	((v) - 1) * sizeof (ObjectIDSequencePair))

#define PRO_PO_OBJ_REQ_SIZE(p, v) \
	(PRO_PO_HDR_SIZE (p) + sizeof (ObjectRequestVariant) + \
	((v) - 1) * sizeof (ObjectID) + ((v) % 2 ? sizeof (int16) : 0))

#define PRO_PO_DELETE_OBJS_SIZE(p, v) \
	(PRO_PO_HDR_SIZE (p) + sizeof (DeleteObjectsVariant) + \
	((v) - 1) * sizeof (ObjectIDWorldStateIDPair))

/* **Persistent Object Protocol (Describe Object).  This is a           */
/*   sub-protocol.  Macros are provided for getting the sizes of each   */
/*   class part.  The application must add the class part to size of    */
/*   PDU returned by the PRO_PO_DESC_OBJ_SIZE macro.                    */

/* Class sizes.  For variants, takes  the number of variants to be      */
/* included.                                                            */

#define PRO_PO_WORLD_STATE_CLASS_SIZE(v) \
	(sizeof (WorldStateClass) + \
	((v) - 1) * sizeof (ObjectID) + ((v) % 2 ? sizeof (int16) : 0))

#define PRO_PO_OVERLAY_CLASS_SIZE \
	(sizeof (OverlayClass))

#define PRO_PO_POINT_CLASS_SIZE \
	(sizeof (PointClass))

#define PRO_PO_LINE_CLASS_SIZE(v) \
	(sizeof (LineClass) + \
	((v) - 1) * sizeof (PointDescription))

#define PRO_PO_SECTOR_CLASS_SIZE(v) \
	(sizeof (SectorClass) + \
	((v) - 1) * sizeof (int32))

#define PRO_PO_TEXT_CLASS_SIZE(v)					     \
	(sizeof (TextClass) +						     \
	 ((v) - 4) * sizeof (char) + ((v) % 4 ? (4-(v)%4) : 0))

#define PRO_PO_UNIT_CLASS_SIZE \
	(sizeof (UnitClass))

#define PRO_PO_LINK_CLASS_SIZE \
        (sizeof (LinkClass))

#define PRO_PO_HANDOFF_CLASS_SIZE \
        (sizeof (HandOffClass))

#define PRO_PO_CIS_CLASS_SIZE \
	(sizeof (CISClass))

#define PRO_PO_STCON_CLASS_SIZE \
	(sizeof (StealthControllerClass))

#define PRO_PO_HHOUR_CLASS_SIZE \
	(sizeof (HHourClass))

#define PRO_PO_COMMO_CLASS_SIZE \
	(sizeof (CommoClass))

#define PRO_PO_TASK_CLASS_SIZE(v)					     \
        (sizeof (TaskClass) +						     \
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_TASK_STATE_CLASS_SIZE(v)					     \
        (sizeof (TaskStateClass) +					     \
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_TASK_FRAME_CLASS_SIZE(v)				             \
        (sizeof (TaskFrameClass) +					     \
	 ((v) - 1) * sizeof (ObjectID) + ((v) % 2 ? sizeof(int16) : 0))

#define PRO_PO_TASK_AUTH_CLASS_SIZE(v)					     \
	(sizeof (TaskAuthorizationClass) +				     \
	 ((v) - 4) * sizeof (char) + ((v) % 4 ? (4-(v)%4) : 0))

#define PRO_PO_PI_CLASS_SIZE(v)						     \
        (sizeof (ParametricInputClass) +				     \
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_PIH_CLASS_SIZE(v)					     \
        (sizeof (ParametricInputHolderClass) +				     \
	 ((v) - 1) * sizeof (TaggedParametricInput) + ((v) % 2 ? 0 : sizeof (int16)))

#define PRO_PO_FIRE_PARAM_CLASS_SIZE                                         \
        (sizeof (FireParametersClass))

#define PRO_PO_CHEMICAL_CLASS_SIZE \
        (sizeof (ChemicalClass))

#define PRO_PO_MINEFIELD_CLASS_SIZE(v)					\
  (sizeof (MinefieldClass) - sizeof((v)->variant) +			\
   (((v)->minefieldType == minefieldTypePoint) ?			\
    (sizeof(MinefieldPoints) +						\
     (((v)->variant.point.pointCount-1) * sizeof(PointLocation))) :	\
    (((v)->minefieldType == minefieldTypeArea) ?			\
     (sizeof(MinefieldArea) +						\
      (((v)->variant.area.size-4) * sizeof(char) +			\
       (((v)->variant.area.size+2) % 4 ? 				\
	(4-((v)->variant.area.size+2)%4) : 0))) :			\
     (sizeof(MinefieldRow) +						\
      (((v)->variant.row.size-1) * sizeof(char) +			\
       (((v)->variant.row.size) % 4 ?					\
	(4-((v)->variant.row.size) % 4) : 0))))))

#define PRO_PO_MINEFIELD_PARENT_CLASS_SIZE                              \
       (sizeof (MinefieldParentClass))

#define PRO_PO_SIMULATION_REQUEST_CLASS_SIZE(v)				\
       (sizeof (SimulationRequestClass) +				\
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_ALERT_USER_CLASS_SIZE(v) \
        (sizeof (AlertUserClass) + \
        ((v) - 4) * sizeof (char) + ((v) % 2 ? sizeof (char) : 0))


#endif /* _po_size_included */
