#ifndef lint
static char rcsid [] = "$RCSfile: po_sp_pdu.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_sp_pdu.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to send a simulator present pdu                      *
*       Created: Mon Apr  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/
#include <stdio.h>
#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <po_size.h>

void po_send_simulator_present_pdu(
    PO_DATABASE *db)
{
    char buf[NETWORK_BUFFER_SIZE];
    PersistentObjectPDU *pdu;
    int32 pdu_size;

    /* This is a sanity check to ensure that if the application trashed
     * the memory pointed to by db, it does not cause the entire world
     * to suffer.
     */
    if (db->database_sequence_number != *db->database_sequence_number_backup)
    {
	fprintf(stderr,"LIBPO FATAL ERROR:\n  Database sequence number consistency check failed.\n  database_sequence_number = %d\n  database_sequence_number_backup = %d\n",
		db->database_sequence_number,
		*db->database_sequence_number_backup);
	abort();
    }

    if (!db->own_sim_entry)
    {
	pdu = (PersistentObjectPDU *)buf;

	pdu->version = persistentObjectProtocolVersionCurrent;
	pdu->kind = simulatorPresentPDUKind;
	pdu->exercise = db->exercise_id;
	pdu->database = db->database_id;
	pdu->variant.simulatorPresent.simulator = db->sim_addr;
	pdu->variant.simulatorPresent.unitDatabaseVersion =
	  db->unit_db_version;
        pdu->variant.simulatorPresent.relativeBattleScheme = 
            db->relative_battle_scheme;
	bcopy((char *)&db->terrain_db_id, (char *)&pdu->variant.simulatorPresent.terrain,
	      sizeof(TerrainDatabaseID));
	bcopy(db->hostname, (char *)pdu->variant.simulatorPresent.hostname,
	      (unsigned int)maxSPHostnameLength);
	/* Guarantee null termination */
	pdu->variant.simulatorPresent.hostname[maxSPHostnameLength-1] = 0;
	pdu->variant.simulatorPresent.simulatorType = db->simulator_type;
    }
    else
      pdu = db->own_sim_entry->pdu;

    pdu->variant.simulatorPresent.databaseSequenceNumber =
      db->database_sequence_number;

    pdu->variant.simulatorPresent.load = po_load(db);
    pdu->variant.simulatorPresent.simulationLoad = po_get_simulation_load(db);
    pdu->variant.simulatorPresent.time = po_time(db);

    /* Loopback to ensure I know my last transmitted load, etc */
    (void) po_process_simpresent(db, pdu, (int32)PRO_PO_SIM_PRESENT_SIZE(pdu));

    /* Fill in packets sent (+1 for this packet)
     * Do this after the loopback, since po_process_simpresent changes
     * packets_sent to 0 so it can be used as a counter.
     */
    pdu->variant.simulatorPresent.packetsSent = db->packets_sent+1;
    pdu_size = PRO_PO_SIM_PRESENT_SIZE(pdu);
    pdu->length =  NS_HTON16(pdu_size);
    db->send_handler(pdu, pdu_size, db->send_user_data);

    po_reset_load(db);
}
    
