#ifndef lint
static char rcsid [] = "$RCSfile: po_stealth.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_stealth.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to manage the stealthControllerClass packets         *
*       Created: Tue May 14 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h> /* common/include/global */

void po_new_stealth(
    PO_DATABASE *db,
    ObjectID    *vehicleID)
{
    StealthControllerClass variant;

    /* Can you add objects to this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      return;

    /* Do we already know about it? */
    if (po_get_entry(db, vehicleID, &po_real_time_world_state))
      return;

    /* Create it.  Nobody is in control. */
    bzero((char *)&variant, sizeof(variant));
    (void) po_send_new_describe_object_pdu(db, vehicleID, 
					   &po_real_time_world_state,
					   objectClassStealthController,
					   (uint32)FALSE,
					   (char *)&variant, sizeof(variant), NULL);
    return;
}

void po_control_stealth(
    PO_DATABASE *db,
    ObjectID    *vehicleID,
    PO_DB_ENTRY *simulator)
{
    PO_DB_ENTRY *entry;
    StealthControllerClass variant;

    /* Can you modify objects in this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      return;

    /* Did the application tell us about it? */
    if (!(entry = po_get_entry(db, vehicleID, &po_real_time_world_state)))
      return;

    if (!simulator)
      simulator = db->own_sim_entry;
    else if (simulator->pdu->kind != simulatorPresentPDUKind)
      return;

    /* Identify the sim as the controller */
    variant.controller = simulator->pdu->variant.simulatorPresent.simulator;
    (void) po_send_changed_describe_object_pdu(db, entry, (int32)FALSE,
					       &variant, sizeof(variant));
}

void po_handle_new_sc(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    entry->locally_controlled =
      SAME_SIMULATOR(db->sim_addr,
		     entry->pdu->variant.describeObject.variant.
		     stealthController.controller);

    if (entry->locally_controlled)
      callback_fire_event(db->project_event, 
			  &entry->pdu->variant.describeObject.objectID,
			  TRUE);
}

void po_handle_changed_sc(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    int32 was_controlled = entry->locally_controlled;

    entry->locally_controlled =
      SAME_SIMULATOR(db->sim_addr,
		     entry->pdu->variant.describeObject.variant.
		     stealthController.controller);

    if (entry->locally_controlled != was_controlled)
      callback_fire_event(db->project_event,
			  &entry->pdu->variant.describeObject.objectID,
			  entry->locally_controlled);
}
