#ifndef lint
static char rcsid [] = "$RCSfile: po_sws_pdu.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_sws_pdu.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to send a SetWorldState PDU                          *
*       Created: Wed May  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <po_size.h>
#include <libqueue.h>
#include <stdext.h> /* common/include/global */

int32 po_send_set_world_state_pdu(
    PO_DATABASE *db,
    ObjectID    *world_state_id,
    uint32       secondsSince1970,
    float32      clock_rate)
{
    PersistentObjectPDU *pdu;
    int32 pdu_size;

    if (!db->set_world_state_pdu)
    {
	db->set_world_state_pdu =
	  (PersistentObjectPDU *)
	    queue_buffer_allocate(PRO_PO_SET_WS_SIZE(pdu));
	if (!db->set_world_state_pdu)
	{
	    PO_ERROR(PO_ENOMEM);
	}
	/* Normally we would do a queue_buffer_up_refcnt() since
	 * we are assigning the queue_buffer, and doing a
	 * queue_buffer_deallocate() to match with the
	 * queue_buffer_allocate().  But, the two routines combine
	 * into a no-op, so they are ommitted.
	 */
    }
    pdu = db->set_world_state_pdu;

    pdu->version = persistentObjectProtocolVersionCurrent;
    pdu->kind = setWorldStatePDUKind;
    pdu->exercise = db->exercise_id;
    pdu->database = db->database_id;
    pdu->variant.setWorldState.requestingSimulator = db->sim_addr;
    pdu->variant.setWorldState.clockRate = clock_rate;
    pdu->variant.setWorldState.secondsSince1970 = secondsSince1970;
    pdu->variant.setWorldState.worldState = *world_state_id;

    pdu_size = PRO_PO_SET_WS_SIZE(pdu);
    pdu->length =  NS_HTON16(pdu_size);
    db->send_handler(pdu, pdu_size, db->send_user_data);
    db->packets_sent++;

    return TRUE;
}

int32 po_resend_set_world_state_pdu(
    PO_DATABASE *db,
    uint32       dtime)
{
    PersistentObjectPDU *pdu;
    int32 pdu_size;

    if (!db->set_world_state_pdu)
      return FALSE;

    pdu = db->set_world_state_pdu;

    pdu->variant.setWorldState.secondsSince1970 += dtime;

    pdu_size = PRO_PO_SET_WS_SIZE(pdu);
    pdu->length = NS_HTON16(pdu_size);
    db->send_handler(pdu, pdu_size, db->send_user_data);
    db->packets_sent++;

    return TRUE;
}
