#ifndef lint
static char rcsid [] = "$RCSfile: po_thread.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_thread.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to manage object threads                             *
*       Created: Fri Apr 12 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include "libpo.h"
#include "libpo_local.h"
#include <libtime.h>
#include <stdalloc.h> /* common/include/global */
#include <stdext.h> /* common/include/global */

/*
 * Forward references:
 */
static PO_OBJECT_HANDLE *
  po_find_or_invent_object_handle(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);


void po_thread_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    int32 index;

    switch(entry->pdu->kind)
    {
      case simulatorPresentPDUKind:
	/*
	 * Thread for class
	 */
	LL_INSERT(entry, db->sim_class_head, prev_class, next_class);
	/* Simulators do not use the time thread */
	/* Simulators do not use the world state thread */
	/* Simulators do not use the object thread */
	break;
      case describeObjectPDUKind:
	/*
	 * Thread for class
	 */
/*changed class to pclass 3 Nov 97 TRANG*/
	index =PO_OBJECT_CLASS_INDEX(entry->pdu->variant.describeObject.pclass);
	LL_INSERT(entry, db->object_classes[index], prev_class, next_class);

	/*
	 * Thread for time
	 */

	/* Implied objects are not retransmitted or timed out. */
	if (!entry->implied)
	{
	    /* Locally owned objects go on the retransmit thread,
	     * unowned objects go on the timeout thread.
	     */

	    if (SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
			       db->sim_addr))
	    {
		Q_INSERT(entry, db->retransmit_head, db->retransmit_tail,
			 prev_time, next_time);
		db->retransmit_size++;
	    }
	    else
	    {
		Q_INSERT(entry, db->timeout_head, db->timeout_tail,
			 prev_time, next_time);
	    }
	}


#ifdef PARANOID
	po_check_queues(db, entry);
#endif

	if (entry->pdu->variant.describeObject.pclass != objectClassWorldState)
	{
	    /*
	     * Thread for world state
	     */

	    /* The entry is guaranteed to have a world_state_head pointer
	     * filled in by this point.
	     */
	    LL_INSERT(entry, entry->world_state_head->world_state_head,
		      prev_world_state, next_world_state);
	}

	/*
	 * Thread for object ID
	 */

	entry->object_handle = po_find_or_invent_object_handle(db, entry);
	LL_INSERT(entry, entry->object_handle->object_head,
		  prev_object, next_object);

	/* Copy object-invariant data here because it is the most
	 * convenient time to do it.
	 */
	if (entry->next_object)
	  entry->object_user_data = entry->next_object->object_user_data;
    }
}

void po_unthread_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    int32 index;
    PO_OBJECT_HANDLE *hold_handle;

    switch(entry->pdu->kind)
    {
      case simulatorPresentPDUKind:
	/*
	 * Unthread by class
	 */
	LL_DELETE(entry, db->sim_class_head, prev_class, next_class);
	break;
      case describeObjectPDUKind:
	/*
	 * Unthread by class
	 */
	index =PO_OBJECT_CLASS_INDEX(entry->pdu->variant.describeObject.pclass);
	LL_DELETE(entry, db->object_classes[index], prev_class, next_class);

	/*
	 * Unthread by time
	 */
	if (!entry->implied)
	{
	    po_unthread_time(db, entry);
	}

	if (entry->pdu->variant.describeObject.pclass != objectClassWorldState)
	{
	    /*
	     * Unthread for world state
	     */
	    LL_DELETE(entry, entry->world_state_head->world_state_head,
		      prev_world_state, next_world_state);
	}

	/*
	 * Unthread for object ID
	 */
	hold_handle = entry->object_handle;
	LL_DELETE(entry, entry->object_handle->object_head,
		  prev_object, next_object);
	/* Free the space if this is the last guy who knows about it */
	if (!hold_handle->object_head)
	{
	    LL_DELETE(hold_handle, db->object_handle_head,
		      prev_handle, next_handle);
	    callback_destroy_event(hold_handle->object_changed_event);
	    STDDEALLOC(hold_handle);
	}
    }
}
    
static PO_OBJECT_HANDLE *po_find_or_invent_object_handle(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *ws_entry;
    ObjectID obj_id, ws_id, *test_id;
    PO_DB_ENTRY *match=NULL;
    PO_OBJECT_HANDLE *result;

    obj_id = entry->pdu->variant.describeObject.objectID;
    ws_id = entry->pdu->variant.describeObject.worldStateID;

    /* Search through all the known world states to see if we can find
     * an object with the same id.  There are probably an order of magnitude
     * more objects than there are world states, so although this seems
     * a peculiar way to search, it should be much more efficient than
     * searching through all the object handles.
     */

    /* Look in the real time world state first */
    if (!SAME_OBJECT_ID(ws_id, po_real_time_world_state))
      match = po_find_entry(db, &obj_id, &po_real_time_world_state);

    /* Then look in all the user-created world states (provided the object
     * can exist in non-real-time world states.
     */
    if (!match &&
	!PO_REAL_TIME_CLASS(entry->pdu->variant.describeObject.pclass))
    {
	for (ws_entry=
	     db->object_classes[PO_OBJECT_CLASS_INDEX(objectClassWorldState)];
	     ws_entry && !match;
	     ws_entry=ws_entry->next_class)
	{
	    test_id = &ws_entry->pdu->variant.describeObject.objectID;
	    if (!SAME_OBJECT_ID(ws_id,*test_id))
	      match = po_find_entry(db, &obj_id, test_id);
	}
    }

    /* If we found another object with this ID, return his object head ptr */
    if (match)
      return match->object_handle;

    /* Otherwise, we need to allocate a new object head ptr.  This is
     * a very small malloc, and catching an error at this point in the
     * code is just too hard.  So this had better succeed.  Sorry.
     */
    result = (PO_OBJECT_HANDLE *)STDALLOC(sizeof(PO_OBJECT_HANDLE));
    bzero((char *)result, sizeof(PO_OBJECT_HANDLE));
    LL_INSERT(result, db->object_handle_head, prev_handle, next_handle);
    result->object_changed_event = callback_define_event(A_PTR, A_END);

    /* Current entry is not maintained for REAL_TIME_CLASSes, so we
     * set it here.
     */
    if (PO_REAL_TIME_CLASS(entry->pdu->variant.describeObject.pclass))
      result->current_entry = entry;

    return result;
}

PO_OBJECT_HANDLE * po_find_handle(
    PO_DATABASE *db,
    ObjectID    *object_id)
{
    PO_DB_ENTRY *ws_entry;
    ObjectID *test_id;
    PO_DB_ENTRY *entry;

    /* Given an object_id, look for the handle to that object.
     * We have two altenatives, search all the handles, or search all the
     * world states of the object.  I vote for searching the world
     * states, since there are a lot fewer world states than there
     * are objects.
     */

    /* First look in the Real Time world state */
    if (entry = po_find_entry(db, object_id, &po_real_time_world_state))
      return (entry->object_handle);

    /* Then look in the rest of the world states */
    for (ws_entry=
	 db->object_classes[PO_OBJECT_CLASS_INDEX(objectClassWorldState)];
	 ws_entry; ws_entry=ws_entry->next_class)
    {
	test_id = &ws_entry->pdu->variant.describeObject.objectID;
	if (entry = po_find_entry(db, object_id, test_id))
	  return (entry->object_handle);
    }

    /* We lose */
    return NULL;
}

void po_make_local(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    if (entry->implied)
      abort();

#ifdef PARANOID
    po_check_queues(db, entry);
#endif

    /* Local objects reside on the retransmit queue */

    Q_DELETE(entry, db->timeout_head, db->timeout_tail,
	     prev_time, next_time);
    Q_INSERT(entry, db->retransmit_head, db->retransmit_tail,
	     prev_time, next_time);
    db->retransmit_size++;
}

void po_make_remote(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    if (entry->implied)
      abort();

#ifdef PARANOID
    po_check_queues(db, entry);
#endif

    /* Remote objects reside on the timeout queue */

    db->retransmit_size--;
    Q_DELETE(entry, db->retransmit_head, db->retransmit_tail,
	     prev_time, next_time);
    Q_INSERT(entry, db->timeout_head, db->timeout_tail,
	     prev_time, next_time);

    /* We don't do dormant for stuff owned by other folks */
    po_not_dormant(entry, db);
}

void po_describe_object_sent(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    /* Note the time of transmission/receipt */
    entry->time = time_realtime_clock();

    /* Pop this object off the front of the time queue, and append
     * it to the back.
     */

#ifdef PARANOID
    po_check_queues(db, entry);
#endif

    if (SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
		       db->sim_addr))
    {
	Q_DELETE(entry, db->retransmit_head, db->retransmit_tail,
		 prev_time, next_time);
	Q_INSERT(entry, db->retransmit_head, db->retransmit_tail,
		 prev_time, next_time);
    }
    else
    {
	Q_DELETE(entry, db->timeout_head, db->timeout_tail,
		 prev_time, next_time);
	Q_INSERT(entry, db->timeout_head, db->timeout_tail,
		 prev_time, next_time);
    }
}

void po_unthread_time(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{

#ifdef PARANOID
    po_check_queues(db, entry);
#endif

    if (entry->deleted)
    {
	Q_DELETE(entry, db->deleted_head, db->deleted_tail,
		 prev_time, next_time);
    }
    else if (SAME_SIMULATOR(entry->pdu->variant.describeObject.owner,
			    db->sim_addr))
    {
	db->retransmit_size--;
	Q_DELETE(entry, db->retransmit_head, db->retransmit_tail,
		 prev_time, next_time);

	/* Dormant objects may be sitting around in an Objects
	 * Present PDU.  Clean them out, too.
	 */
	if (entry->dormant)
	  po_remove_object_present(db, entry);
    }
    else
    {
	Q_DELETE(entry, db->timeout_head, db->timeout_tail,
		 prev_time, next_time);
    }
}
