#ifndef lint
static char rcsid [] = "$RCSfile: po_tick.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_tick.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Tick based po routines                                    *
*       Created: Mon Apr  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <libtime.h>

void po_timeout_control(
    PO_DATABASE *db,
    int32        enabled)
{
    db->timeout_enabled = enabled;
}

void po_tick(
    PO_DATABASE *db)
{
    uint32 msecs = time_realtime_clock();

    /* Note the time of our first tick */
    if (!db->db_create_time)
      db->db_create_time = msecs;

    /* Simulator Present transmission.
     *
     * Send out a packet periodically indicating that you are still
     * alive.
     */

    if ((db->db_type == PO_DATABASE_TYPE_ACTIVE) &&
	(msecs > db->next_simulator_present_time))
    {
	db->next_simulator_present_time =
	  msecs + simulatorPresentTransmitTime * 1000;
	po_send_simulator_present_pdu(db);
    }

    /* Time out simulators.
     *
     * Any simulator not transmitting any packets for a long time
     * must be timed out.  His objects will be reallocated among other
     * machines as a result.
     */

    if (db->timeout_enabled)
      po_timeout_sims(db);

    /* Object transmission.
     *
     * Send out a packet periodically for each object you own.
     */

    po_tick_objects(db);

    /* Time out objects.
     *
     * Objects which have not been described in a long time must be
     * timed out.
     */

    if (db->timeout_enabled)
      po_timeout_objects(db);

    /* Time out objects.
     *
     * Destroy objects marked for deletion for a long time.
     */

    po_delete_dead_objects(db);

    /* Animation.
     *
     * Resend animation packets (if any).
     */
    po_tick_animation(db);

    /* Check for transitions to 'ready' state */
    if (!db->ready)
    {
	if (po_ready(db, msecs))
	{
	    /* Set the simulation load to zero on the transition to ready */
	    db->simulation_load = 0.0;
	}
    }
}
