#ifndef lint
static char rcsid [] = "$RCSfile: po_write.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_write.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to save libpo files                                  *
*       Created: Mon May 13 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <po_size.h>
#include <stdext.h> /* common/include/global */
#include <stdstring.h> /*common/include/global*/

/* Query structures */

struct save_struct
{
    PO_DATABASE   *db;
    FILE          *file;
    int32          mask;
    int32          write_error;
};

struct choose_struct
{
    PO_DATABASE   *db;
    uint32         mask;
    int32          save_scratch;
    int32          save_realtime;
    int32          save_non_realtime;
    PO_DB_ENTRY   *overlay;
    int32          count;
    uint32         serial;
};

/*
 * Forward references:
 */

static void po_zero_serials(
    PO_DATABASE *db);

static int32 po_preassign_ws_serials(
    PO_DATABASE *db);

static void po_choose_entry(
    PO_DB_ENTRY          *entry,
    struct choose_struct *cs);

static void po_save_entry(
    PO_DB_ENTRY        *entry,
    struct save_struct *ss);

static void po_copy_to_output_buffer(
    PO_DATABASE           *db,
    char                  *variant,
    char                  *buf,
    PersistentObjectClass  pclass,
    uint32                 vsize);

int32 po_save_all(
    PO_DATABASE *db,
    char        *fname,
    int32        save_scratch,
    int32        save_realtime,
    int32        save_non_realtime)
{
    FILE *file;
    PO_FILE_HDR hdr;
    struct save_struct ss;
    struct choose_struct cs;

    /* Initialize serial numbers */
    po_zero_serials(db);

    /* Give the world states the first serial numbers */
    hdr.num_world_states = po_preassign_ws_serials(db);

    /* Choose the entries */
    cs.db = db;
    cs.mask = (PO_FULL_CLASS_MASK);
    cs.save_scratch = save_scratch;
    cs.save_realtime = save_realtime;
    cs.save_non_realtime = save_non_realtime;
    cs.overlay = NULL;
    cs.count = 0;
    cs.serial = hdr.num_world_states+1;
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_choose_entry, &cs);
    hdr.num_entries = cs.count;
    hdr.num_objects = cs.serial-1;

    if (hdr.num_entries == 0)
      PO_ERROR(PO_EEMPTY);

    /* You cannot save more than 65535 object IDs because it would be
     * impossible for one machine to load the file.  (ObjectID only leaves
     * 16 bits for the unique part.)
     */
    if (hdr.num_objects > 0xFFFF)
      PO_ERROR(PO_EOBJID);

    file = fopen(fname, "w");

    /* Did we get it open? */
    if (!file)
      PO_ERROR(PO_EOPEN);

    /* Fill out the header */
    hdr.file_format = PO_CURRENT_FILE_FORMAT;
    hdr.file_type = PO_FILE_TYPE_COMPLETE;
    hdr.protocol_version = persistentObjectProtocolVersionCurrent;
    (void) strcpy(hdr.creator, db->hostname);
    hdr.unit_database_version = db->unit_db_version;
    bcopy((char *)&db->terrain_db_id, (char *)&hdr.terrain, sizeof(hdr.terrain));

    /* Write the header */
    if (fwrite(&hdr, sizeof(hdr), 1, file) != 1)
    {
	(void) fclose(file);
	PO_ERROR(PO_EWRITE);
    }

    ss.db = db;
    ss.file = file;
    ss.write_error = FALSE;

    /* Write the entries in a dependency friendly order:
     *   World States first
     *   HHours Second
     *   Overlays third
     *   Everything else after that
     *   Minefields last
     */
    ss.mask = PO_CLASS_MASK(objectClassWorldState);
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_save_entry, &ss);

    ss.mask = PO_CLASS_MASK(objectClassHHour);
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_save_entry, &ss);

    ss.mask = PO_CLASS_MASK(objectClassOverlay);
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_save_entry, &ss);

    ss.mask = (PO_FULL_CLASS_MASK &
	       ~(PO_CLASS_MASK(objectClassWorldState) |
		 PO_CLASS_MASK(objectClassHHour) |
		 PO_CLASS_MASK(objectClassOverlay) |
		 PO_CLASS_MASK(objectClassMinefield)));
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_save_entry, &ss);

    ss.mask = PO_CLASS_MASK(objectClassMinefield);
    po_query_for_all_entries(db, (PO_QUERY_HANDLER)po_save_entry, &ss);

    /* Did we succeed at writing the entries? */
    if (ss.write_error)
    {
	(void) fclose(file);
	PO_ERROR(PO_EWRITE);
    }

    /* Close the file */
    if (fclose(file) == EOF)
    {
	PO_ERROR(PO_ECLOSE);
    }

    return hdr.num_entries;
} 

int32 po_save_overlay(
    PO_DATABASE *db,
    char        *fname,
    PO_DB_ENTRY *overlay,
    uint32       class_mask)
{
    FILE *file;
    PO_FILE_HDR hdr;
    struct save_struct ss;
    struct choose_struct cs;

    /* Make sure he gave us an overlay */
    if (overlay->pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);
    if (overlay->pdu->variant.describeObject.pclass != objectClassOverlay)
      PO_ERROR(PO_ECLASS);

    /* Initialize serial numbers */
    po_zero_serials(db);

    /* Do not save world states in PARTIAL files.  That would not make
     * sense.  Also skip overlays in the queries, since we will save that
     * explicitly.  And hhour to prevent them from being saved in multiple
     * files, causing them to duplicate whenever the overlay is reloaded.
     */
    /* ADD NEW CLASS HERE 
     *
     * Add any class that shouldn't be saved in an overlay
     */
    class_mask &= ~(PO_CLASS_MASK(objectClassWorldState) |
		    PO_CLASS_MASK(objectClassHHour) |
		    PO_CLASS_MASK(objectClassOverlay) |
		    PO_CLASS_MASK(objectClassTask) |
		    PO_CLASS_MASK(objectClassTaskFrame) |
		    PO_CLASS_MASK(objectClassTaskAuthorization) |
		    PO_CLASS_MASK(objectClassParametricInput) |
		    PO_CLASS_MASK(objectClassParametricInputHolder) |
		    PO_CLASS_MASK(objectClassExerciseInitializer) |
		    PO_CLASS_MASK(objectClassFireParameters) |
		    PO_CLASS_MASK(objectClassSimulationRequest) |
		    PO_CLASS_MASK(objectClassMinefield) |
		    PO_CLASS_MASK(objectClassHandOff));
    
    /* Choose the entries first, because the query will zero out the
     * overlay's will_save flag.
     */
    cs.db = db;
    cs.count = 0;
    cs.serial = 1;
    cs.mask = class_mask;
    cs.save_scratch = TRUE;
    cs.save_realtime = TRUE;
    cs.save_non_realtime = TRUE;
    cs.overlay = overlay;
    po_query_for_current_objects(db, (PO_QUERY_HANDLER)po_choose_entry, &cs);

    /* Now choose the overlay */
    cs.mask = PO_CLASS_MASK(objectClassOverlay);
    cs.overlay = NULL;
    po_choose_entry(overlay, &cs);

    hdr.num_entries = cs.count;
    hdr.num_world_states = 0;
    hdr.num_objects = cs.serial-1;

    if (hdr.num_entries == 0)
      PO_ERROR(PO_EEMPTY);

    /* You cannot save more than 65535 object IDs because it would be
     * impossible for one machine to load the file.  (ObjectID only leaves
     * 16 bits for the .object part.)
     */
    if (hdr.num_objects > 0xFFFF)
      PO_ERROR(PO_EOBJID);

    file = fopen(fname, "w");

    /* Did we get it open? */
    if (!file)
      PO_ERROR(PO_EOPEN);

    /* Fill out the header */
    hdr.file_format = PO_CURRENT_FILE_FORMAT;
    hdr.file_type = PO_FILE_TYPE_PARTIAL;
    hdr.protocol_version = persistentObjectProtocolVersionCurrent;
    (void) strcpy(hdr.creator, db->hostname);
    hdr.unit_database_version = db->unit_db_version;
    bcopy((char *)&db->terrain_db_id, 
	  (char *)&hdr.terrain,
	  sizeof(hdr.terrain));

    /* Write the header */
    if (fwrite(&hdr, sizeof(hdr), 1, file) != 1)
    {
	(void) fclose(file);
	PO_ERROR(PO_EWRITE);
    }

    ss.db = db;
    ss.file = file;
    ss.write_error = FALSE;

    /* Write the entries in a dependency friendly order:
     *   Overlay first
     *   Everything else after that
     *   Minefields last
     */
    ss.mask = PO_CLASS_MASK(objectClassOverlay);
    po_save_entry(overlay, &ss);

    ss.mask = class_mask;
    po_query_for_current_objects(db, (PO_QUERY_HANDLER)po_save_entry, &ss);

    ss.mask = PO_CLASS_MASK(objectClassMinefield);
    po_save_entry(overlay, &ss);

    /* Did we succeed at writing the entries? */
    if (ss.write_error)
    {
	(void) fclose(file);
	PO_ERROR(PO_EWRITE);
    }

    /* Close the file */
    if (fclose(file) == EOF)
    {
	PO_ERROR(PO_ECLOSE);
    }

    return hdr.num_entries;
} 

static void po_zero_serials(
    PO_DATABASE *db)
{
    PO_OBJECT_HANDLE *handle;

    for(handle=db->object_handle_head;handle;handle=handle->next_handle)
      handle->save_serial = 0;
}

static int32 po_should_save(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       mask,
    int32        save_scratch,
    int32        save_realtime,
    int32        save_non_realtime,
    PO_DB_ENTRY *overlay)
{
    PersistentObjectClass pclass = entry->pdu->variant.describeObject.pclass;
    PO_DB_ENTRY *over;

    /* Is it in a class we want? */
    if (!(PO_CLASS_MASK(pclass) & mask))
      return FALSE;

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol, add it somewhere
     * in this switch.
     */
    switch(pclass)
    {
      /*
       * Things that are NEVER saved.
       */
      case objectClassStealthController:
	return FALSE;

      /*
       * Things that are ALWAYS saved if the mask identifies them.
       */
      case objectClassHHour:
      case objectClassTask:
      case objectClassTaskState:
      case objectClassTaskFrame:
      case objectClassTaskAuthorization:
      case objectClassParametricInput:
      case objectClassParametricInputHolder:
      case objectClassExerciseInitializer:
      case objectClassFireParameters:
      case objectClassSimulationRequest:
      case objectClassHandOff:
	return TRUE;
	
      /*
       * Things that are sometimes saved depending on whether we are
       *   saving the non-real time state.
       */
      case objectClassWorldState:
	if (save_non_realtime)
	  return TRUE;
	else
	  return FALSE;

      /*
       * Overlays should be included if they pass the mask test, and
       * they pass the scratch test.
       */
      case objectClassOverlay:
	return (save_scratch ||
		!entry->pdu->variant.describeObject.variant.overlay.scratch);

      /*
       * Things in overlays should be included if they are in the overlay
       * we want.
       */
      case objectClassPoint:
      case objectClassLine:
      case objectClassMinefieldParent:
      case objectClassMinefield:
      case objectClassChemical:
      case objectClassSector:
      case objectClassText:
      case objectClassUnit:
      case objectClassLink:
	/* What overlay is this in? */
	over = po_find_overlay(db, entry);

	/* These objects MUST be in a real overlay.  If they are not,
	 * we cannot save them.
	 */
	if (!over)
	  return FALSE;

	/*
	 * Don't save objects in world states that we didn't want.
	 */
	if (SAME_OBJECT_ID(entry->pdu->variant.describeObject.worldStateID,
			   po_real_time_world_state))
	{
	    if (!save_realtime)
	      return FALSE;
	}
	else
	{
	    if (!save_non_realtime)
	      return FALSE;
	}

	/* Are we saving scratch overlays? */
	if (save_scratch)
	{
	    /* If we are saving all the overlays, then having the right mask
	     * is good enough.  Otherwise make sure we want it.
	     */
	    return (!overlay || (overlay == over));
	}
	else
	{
	    /* Save it if it we want its overlay and it is not scratch */
	    return
	      ((!overlay || (overlay == over)) &&
	       !over->pdu->variant.describeObject.variant.overlay.scratch);
	}

      /*
       * Other entries should be included if they are associated with
       * included things.
       */
    }
    return TRUE;
}

static int32 po_preassign_ws_serials(PO_DATABASE *db)
{
    PO_DB_ENTRY *ws;
    int32 count = 1;

    for(ws=db->object_classes[PO_OBJECT_CLASS_INDEX(objectClassWorldState)];
	ws;ws=ws->next_class)
    {
	if (!ws->object_handle->save_serial)
	{
	    ws->object_handle->save_serial = count;
	    count += 1;
	}
    }
    return count-1;
}

static void po_choose_entry(
    PO_DB_ENTRY          *entry,
    struct choose_struct *cs)
{
    if (po_should_save(cs->db, entry, cs->mask,
		       cs->save_scratch, cs->save_realtime, 
		       cs->save_non_realtime, cs->overlay))
    {
	entry->will_save = TRUE;
	cs->count += 1;
	if (!entry->object_handle->save_serial)
	{
	    entry->object_handle->save_serial = cs->serial;
	    cs->serial += 1;
	}
    }
    else
      entry->will_save = FALSE;
}

static int32 po_get_serial(
    PO_DATABASE *db,
    ObjectID    *object_id)
{
    PO_OBJECT_HANDLE *handle;

    if (SAME_OBJECT_ID(*object_id, po_real_time_world_state))
      return 0;

    if (handle = po_find_handle(db, object_id))
      return (handle->save_serial);

    return 0;
}

static void po_make_serial(
    PO_DATABASE *db,
    ObjectID    *object_id)
{
    int32 serial = po_get_serial(db, object_id);

    object_id->simulator.site = 0;
    object_id->simulator.host = 0;
    object_id->object = serial;
}

static void po_save_entry(
    PO_DB_ENTRY        *entry,
    struct save_struct *ss)
{
    PO_FILE_ENTRY fe;
    char buf[NETWORK_BUFFER_SIZE];

    /* Don't even try if the last one failed */
    if (ss->write_error)
      return;

    /* Is this something we want to save? */
    if (!entry->will_save)
      return;

    /* Is it being filtered out this pass? */
    if (!(PO_CLASS_MASK(entry->pdu->variant.describeObject.pclass) & ss->mask))
      return;

    /* Fill out the entry */
    fe.serial_number = entry->object_handle->save_serial;
    fe.world_state_serial_number =
      po_get_serial(ss->db, &entry->pdu->variant.describeObject.worldStateID);
    fe.class = entry->pdu->variant.describeObject.pclass;
    fe.missing_from_world_state =
      entry->pdu->variant.describeObject.missingFromWorldState;
    fe.variant_size = entry->pdu_size - PRO_PO_DESC_OBJ_SIZE(entry->pdu);

    /* Write it into the file */
    if (fwrite(&fe, sizeof(fe), 1, ss->file) != 1)
    {
	ss->write_error = TRUE;
	return;
    }

    /* Convert objectIDs to serial numbers */
    po_copy_to_output_buffer(ss->db,
			     (char *)&entry->pdu->variant.describeObject.variant,
			     buf, fe.class, (uint32)fe.variant_size);

    /* Write it out */
    if (fwrite(buf, (int) fe.variant_size, 1, ss->file) != 1)
    {
	ss->write_error = TRUE;
	return;
    }
}

static void po_copy_to_output_buffer(
    PO_DATABASE           *db,
    char                  *variant,
    char                  *buf,
    PersistentObjectClass  pclass,
    uint32                 vsize)
{
    WorldStateClass *wsc;
    int32 i;

    /* First copy the raw data */
    bcopy(variant, buf, (unsigned int)vsize);

    /* Then go in and replace objectID references with serial numbers */

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol, add it somewhere
     * in this switch.
     */
    switch (pclass)
    {
      /*
       * Things which do not reference other objects.
       */
      case objectClassHHour:
	((HHourClass *)buf)->time -= (int32) po_time(db);
      case objectClassOverlay:
      case objectClassStealthController:
      case objectClassExerciseInitializer:
	return;
      case objectClassSimulationRequest:
	((SimulationRequestClass *)buf)->startTime -= (int32) po_time(db);
	((SimulationRequestClass *)buf)->simulated = FALSE;
	bzero((char *)&((SimulationRequestClass *)buf)->simulator,
	      sizeof(SimulationAddress));
	return;

      /*
       * Things which do reference other objects.
       */
      case objectClassWorldState:
	wsc = (WorldStateClass *)buf;
	for (i=0;i<wsc->historyCount;i++)
	  po_make_serial(db, &wsc->history[i]);
	return;
      case objectClassPoint:
	po_make_serial(db, &(((PointClass *)buf)->overlayID));
	return;
      case objectClassLine:
	po_make_serial(db, &(((LineClass *)buf)->overlayID));
	bzero((char *)&(((LineClass *)buf)->simulator),
	      sizeof(SimulationAddress));
	((LineClass *)buf)->simulated = FALSE;
	return;
      case objectClassMinefieldParent:
	po_make_serial(db, &(((MinefieldParentClass *)buf)->overlayID));
	bzero((char *)&(((MinefieldParentClass *)buf)->simulationID),
	      sizeof(VehicleID));
	bzero((char *)&(((MinefieldParentClass *)buf)->simulator),
	      sizeof(SimulationAddress));
	((MinefieldParentClass *)buf)->simulated = FALSE;
	return;
      case objectClassChemical:
	po_make_serial(db, &(((ChemicalClass *)buf)->overlayID));
	return;
      case objectClassMinefield:
	po_make_serial(db, &(((MinefieldClass *)buf)->parentID));
	return;
      case objectClassSector:
	po_make_serial(db, &(((SectorClass *)buf)->overlayID));
	return;
      case objectClassText:
	po_make_serial(db, &(((TextClass *)buf)->overlayID));
	po_make_serial(db, &(((TextClass *)buf)->associatedObject));
	return;
      case objectClassUnit:
	po_make_serial(db, &(((UnitClass *)buf)->overlayID));
	po_make_serial(db, &(((UnitClass *)buf)->taskOrganizedTo));
	po_make_serial(db, &(((UnitClass *)buf)->functionallyOrganizedTo));
	po_make_serial(db, &(((UnitClass *)buf)->taskFrameStack));
	po_make_serial(db, &(((UnitClass *)buf)->backgroundFrame));
	po_make_serial(db, &(((UnitClass *)buf)->parameters));
	po_make_serial(db, &(((UnitClass *)buf)->handOffData));
	bzero((char *)&(((UnitClass *)buf)->simulationID), sizeof(VehicleID));
	bzero((char *)&(((UnitClass *)buf)->simulator), sizeof(SimulationAddress));
	((UnitClass *)buf)->simulated = FALSE;
	((UnitClass *)buf)->time -= po_time(db);
	return;
      case objectClassLink:
	for (i = 0; i < maxLinkConnections; i++)
	{
	    if (!NS_OBJECT_ID_NULL(((LinkClass *)buf)->memberIds[i]))
	      bzero(&(((LinkClass *)buf)->memberVehIds[i]), sizeof(VehicleID));
	    po_make_serial(db, &(((LinkClass *)buf)->memberIds[i]));
	}
	po_make_serial(db, &(((LinkClass *)buf)->overlayID));
	return;
      case objectClassTask:
	po_make_serial(db, &(((TaskClass *)buf)->frame));
	po_make_serial(db, &(((TaskClass *)buf)->state));
	for (i = 0; i < ((TaskClass *)buf)->refcount; i++)
	  po_make_serial(db, PO_REFERENCES(*(TaskClass *)buf)+i);
	return;
      case objectClassTaskState:
	for (i = 0; i < ((TaskStateClass *)buf)->refcount; i++)
	  po_make_serial(db, PO_REFERENCES(*(TaskStateClass *)buf)+i);
	return;
      case objectClassTaskFrame:
	po_make_serial(db, &(((TaskFrameClass *)buf)->unit));
	po_make_serial(db, &(((TaskFrameClass *)buf)->commandingUnit));
	po_make_serial(db, &(((TaskFrameClass *)buf)->nextStackFrame));
	po_make_serial(db,
		       &(((TaskFrameClass *)buf)->previousMissionFrame));
	po_make_serial(db, &(((TaskFrameClass *)buf)->sponsoringTask));
	for (i = 0; i < ((TaskFrameClass *)buf)->etaskCount; i++)
	  po_make_serial(db, &(((TaskFrameClass *)buf)->enablingTask[i]));
	return;
      case objectClassTaskAuthorization:
	po_make_serial(db, &(((TaskAuthorizationClass *)buf)->unit));
	po_make_serial(db, &(((TaskAuthorizationClass *)buf)->proposedTaskFrame));
	po_make_serial(db, &(((TaskAuthorizationClass *)buf)->modifiedTaskFrame));
	return;
      case objectClassParametricInput:
	po_make_serial(db, &(((ParametricInputClass *)buf)->chain));
	return;
      case objectClassFireParameters:
	po_make_serial(db, &(((FireParametersClass *)buf)->unit));
	return;
      case objectClassParametricInputHolder:
	for (i = 0; i < ((ParametricInputHolderClass *)buf)->size; i++)
	  po_make_serial(db,
			 &(((ParametricInputHolderClass *)buf)->blocks[i].data));
	return;
      case objectClassHandOff:
	for (i = 0; i < ((HandOffClass *)buf)->numHandOff; i++)
	{
	    po_make_serial(db, &(((HandOffClass *)buf)->handOff[i].unit));
	}
	return;
    }
}
