#ifndef lint
static char rcsid [] = "$RCSfile: po_ws.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_ws.c                                                           *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code dealing with world states                            *
*       Created: Wed Apr 17 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <stdext.h> /* common/include/global */

/*
 * Forward references:
 */
static void
  po_find_current_entry(
    PO_DATABASE      *db,
    PO_OBJECT_HANDLE *handle,
    int32             sanity);

static void
  po_find_current_entries(
    PO_DATABASE *db);
			     

ObjectID po_real_time_world_state = { { 0 , 0 } , 0 };

int32 po_set_world_state(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    if (!entry)
    {
	callback_fire_event(db->world_state_changing_event, entry);
	db->current_world_state = NULL;
	db->current_world_state_id = po_real_time_world_state;
    }
    else
    {
	if (entry->pdu->kind != describeObjectPDUKind)
	  PO_ERROR(PO_EKIND);

	if (entry->pdu->variant.describeObject.pclass != objectClassWorldState)
	  PO_ERROR(PO_ECLASS);

	callback_fire_event(db->world_state_changing_event, entry);
	db->current_world_state = entry;
	db->current_world_state_id =
	  entry->pdu->variant.describeObject.objectID;
    }

    po_find_current_entries(db);
    callback_fire_event(db->world_state_changed_event, entry);
    return TRUE;
}

PO_DB_ENTRY *
po_find_base_world_state(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    PO_DB_ENTRY *skip_entry)
{
    int32 i;
    ObjectID *history;
    int32 history_count;
    PO_DB_ENTRY *base_entry;

    history = entry->pdu->variant.describeObject.variant.worldState.history;
    history_count =
      entry->pdu->variant.describeObject.variant.worldState.historyCount;

    for(i=history_count-1;i>=0;i--)
    {
	if ((base_entry = po_get_entry(db, history+i,
				       &po_real_time_world_state)) &&
	    (base_entry != skip_entry))
	{
	    return base_entry;
	}
    }
    return NULL;
}

static void po_world_state_lost(
    PO_DATABASE *db)
{
    PO_DB_ENTRY *new_world_state;

    /* This function gets called when the world state we are in
     * is deleted from the database.  If that happens, switch to
     * the most recent existing world state in our history.
     */

    if (!db->current_world_state)
      return;

    if (new_world_state =
	po_find_base_world_state(db, db->current_world_state,
				 (PO_DB_ENTRY *)NULL))
    {
	(void) po_set_world_state(db, new_world_state);
	return;
    }
    /* If we could not find one, go to the real time world state */
    (void) po_set_world_state(db, (PO_DB_ENTRY *)NULL);
}

PO_DB_ENTRY *
po_get_object_for_world_state(
    PO_DATABASE *db,
    ObjectID    *object_id,
    ObjectID    *world_state_id)
{
    PO_DB_ENTRY *entry;
    PO_DB_ENTRY *ws_entry;
    ObjectID *history;
    int32 i;

    /* First look for a perfect world state match */
    if ((entry = po_find_entry(db, object_id, world_state_id)) &&
	!entry->deleted && !entry->implied)
      ;
    else if ((ws_entry = po_find_entry(db, world_state_id,
				       &po_real_time_world_state)) &&
	     !ws_entry->deleted && !ws_entry->implied)
    {
	/* Search through the history of the world state looking for
	 * a match.
	 */
	history = ws_entry->pdu->
	  variant.describeObject.variant.worldState.history;

	for (i=ws_entry->pdu->
	     variant.describeObject.variant.worldState.historyCount-1;
	     i>=0; i--)
	{
	    if ((entry = po_find_entry(db, object_id, history+i)) &&
		!entry->deleted && !entry->implied)
	      break;
	}
    }
	
    if (!entry || entry->deleted || entry->implied)
      return NULL;

    return entry;
}

PO_DB_ENTRY *
po_current_object(
    PO_DATABASE *db,
    ObjectID    *object_id)
{
    return (po_get_object_for_world_state(db, object_id,
					  &db->current_world_state_id));
}

int32 po_world_state_dependency(
    PO_DB_ENTRY *dependent_state,
    PO_DB_ENTRY *base_state,
    ObjectID    *subsequent_state)
{
    ObjectID bs_id;
    ObjectID *history;
    int32 history_count;
    int32 i;

    bs_id = base_state->pdu->variant.describeObject.objectID;
    history =
      dependent_state->pdu->variant.describeObject.variant.worldState.history;
    history_count =
      dependent_state->pdu->
	variant.describeObject.variant.worldState.historyCount;

    for(i=0; i<history_count; i++)
      if (SAME_OBJECT_ID(bs_id, history[i]))
      {
	  if (i==history_count-1)
	    *subsequent_state =
	      dependent_state->pdu->variant.describeObject.objectID;
	  else
	    *subsequent_state = history[i+1];

	  return TRUE;
      }
    return FALSE;
}

int32 po_get_associate_id(
    PersistentObjectPDU *pdu,
    ObjectID            *id_ret)
{
    if (pdu->kind != describeObjectPDUKind)
      PO_ERROR(PO_EKIND);

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol which has an associated id
     * field, add a case to this switch.  For non-association-grouped classes,
     * there is no need to change this switch.
     */

    switch (pdu->variant.describeObject.pclass)
    {
      case objectClassText:
	*id_ret = pdu->variant.describeObject.variant.text.associatedObject;
	return TRUE;
      default:
	*id_ret = po_real_time_world_state;
	return FALSE;
    }
}

static void po_find_current_associates(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    int32        sanity)
{
    ObjectID associated;
    PO_OBJECT_HANDLE *handle;

    if (!po_get_associate_id(entry->pdu, &associated))
      return;

    if ((associated.simulator.site == 0) &&
	(associated.simulator.host == 0) &&
	(associated.object == 0))
      return;

    handle = po_find_handle(db, &associated);

    if (handle && handle->object_head &&
	(handle->object_head->pdu->kind == describeObjectPDUKind))
      po_find_current_entry(db, handle, sanity);
}

static void po_find_current_entry(
    PO_DATABASE      *db,
    PO_OBJECT_HANDLE *handle,
    int32             sanity)
{
    PO_DB_ENTRY *previous_entry;

    /* REAL_TIME_CLASSes are always known to the application.
     */
    if (PO_REAL_TIME_CLASS(handle->object_head->pdu->
			   variant.describeObject.pclass))
      return;

    /* First find the previous entry */
    if (previous_entry = handle->current_entry)
    {
	if (previous_entry->deleted ||
	    previous_entry->implied)
	  previous_entry = NULL;
    }

    /* Now note the new current entry */
    handle->current_entry =
      po_current_object(db, &handle->object_head->pdu->
			variant.describeObject.objectID);

    /* Easy case first: no change */
    if (previous_entry == handle->current_entry)
      return;

    /* If there is a current entry which the application can see,
     * it will either look like a new object, or a changed object,
     * depending upon what was there before.
     */
    if (handle->current_entry &&
	!handle->current_entry->pdu->
	variant.describeObject.missingFromWorldState)
    {
	/* Before telling the application about this object (changed or
	 * new), make sure it already knows about objects with which this
	 * one is associated.  The sanity count is used to bail out of
	 * cyclic relationships.
	 */
	if (sanity < 100)
	  po_find_current_associates(db, handle->current_entry, sanity+1);

	if (!previous_entry ||
	    previous_entry->pdu->
	    variant.describeObject.missingFromWorldState)
	  callback_fire_event(db->new_object_event, handle->current_entry);
	else
	{
	    /* If there was something there, this will look changed */
	    callback_fire_event(db->object_changed_event,
				handle->current_entry);
	    callback_fire_event(handle->object_changed_event,
				handle->current_entry);
	}
    }
    else
    {
	/* There is no current entry which the application, can see.
	 * If there was one before, then the object will disappear.
	 */
	if (previous_entry &&
	    !previous_entry->pdu->
	    variant.describeObject.missingFromWorldState)
	{
	    callback_fire_event(db->object_gone_event, previous_entry);

	    /* The user may be expecting this to be NULL */
	    po_set_object_user_data(db, previous_entry,
				    (PO_USER_DATA_TYPE)NULL);
	}
    }

}

static void po_find_current_entries(
    PO_DATABASE *db)
{
    PO_OBJECT_HANDLE *handle;

    /* This function is called when the world state changes.
     * We need to look at each object, and if it looks somehow different
     * to the application, we need to tell the application.
     */
    for(handle=db->object_handle_head;handle;handle=handle->next_handle)
    {
	po_find_current_entry(db, handle, (int32)0);
    }
}

/* Application handlers regarding entries are called through here,
 * since the object_ versions may need to be called depending upon
 * world state information.
 *
 * These all look fairly similar.  What they are doing is figuring out
 * what the change will look like to the application.
 */

void po_handle_new_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *previous_entry = NULL;

    /* Handle stealth controller objects specially */
    if (entry->pdu->variant.describeObject.pclass ==
	objectClassStealthController)
    {
	po_handle_new_sc(db, entry);
	return;
    }

    /* Handle exercise initializer objects specially */
    if (entry->pdu->variant.describeObject.pclass ==
	objectClassExerciseInitializer)
    {
	po_handle_new_exercise_initializer(db, entry);
	return;
    }

    callback_fire_event(db->new_entry_event, entry);

    /* REAL_TIME_CLASSes are simple.  They exist regardless of
     * the world state the application is interested in.
     */
    if (PO_REAL_TIME_CLASS(entry->pdu->variant.describeObject.pclass))
    {
	callback_fire_event(db->new_object_event, entry);
	return;
    }

    /* First find the previous entry */
    if (entry->object_handle &&
	(previous_entry = entry->object_handle->current_entry))
    {
	if (previous_entry->deleted ||
	    previous_entry->implied ||
	    previous_entry->pdu->variant.describeObject.missingFromWorldState)
	  previous_entry = NULL;
    }

    /* Now note the new current entry */
    entry->object_handle->current_entry =
      po_current_object(db, &entry->pdu->variant.describeObject.objectID);

    /* Is this the current entry for this object? */
    if (entry == entry->object_handle->current_entry)
    {
	/* Is it missing? */
	if (entry->pdu->variant.describeObject.missingFromWorldState)
	{
	    /* If there was an object here before, and this object is
	     * tagged as missing, this will look like the object
	     * was deleted.
	     */
	    if (previous_entry)
	    {
		callback_fire_event(db->object_gone_event, entry);

		/* The user may be expecting this to be NULL */
		po_set_object_user_data(db, entry, (PO_USER_DATA_TYPE)NULL);
	    }
	}
	else
	{
	    /* If there was something here before, this is actually going
	     * to look like a change to that object, rather than like a
	     * new object.
	     */
	    if (previous_entry)
	    {
		callback_fire_event(db->object_changed_event, entry);
		callback_fire_event(entry->object_handle->object_changed_event,
				    entry);
	    }
	    else
	      callback_fire_event(db->new_object_event, entry);
	}
    }
}

void po_handle_entry_changed(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    int32        previously_missing)
{
    /* Handle stealth controller objects specially */
    if (entry->pdu->variant.describeObject.pclass ==
	objectClassStealthController)
    {
	po_handle_changed_sc(db, entry);
	return;
    }

    /* Handle exercise initializer objects specially */
    if (entry->pdu->variant.describeObject.pclass ==
	objectClassExerciseInitializer)
    {
	po_handle_changed_exercise_initializer(db, entry);
	return;
    }

    callback_fire_event(db->entry_changed_event, entry);

    /* REAL_TIME_CLASSes are simple.  They exist regardless of
     * the world state the application is interested in.
     */
    if (PO_REAL_TIME_CLASS(entry->pdu->variant.describeObject.pclass))
    {
	callback_fire_event(db->object_changed_event, entry);
	callback_fire_event(entry->object_handle->object_changed_event, entry);
    }

    /* Changing the meat of an entry can not possibly cause which entry
     * is current to change.  Hence, we do not need to do the search,
     * merely check against the last value.
     */

    /* Is this the current entry for this object? */
    if (entry == entry->object_handle->current_entry)
    {
	/* Is it missing? */
	if (entry->pdu->variant.describeObject.missingFromWorldState)
	{
	    /* If this object was here, but is no longer, then this will
	     * look like a deletion to the application.
	     */
	    if (!previously_missing)
	    {
		callback_fire_event(db->object_gone_event, entry);

		/* The user may be expecting this to be NULL */
		po_set_object_user_data(db, entry, (PO_USER_DATA_TYPE)NULL);
	    }
	}
	else
	{
	    /* If this was missing before, then this is actually going
	     * to look like a new object, rather than like a change to
	     * that object.
	     */
	    if (previously_missing)
	      callback_fire_event(db->new_object_event, entry);
	    else
	    {
		callback_fire_event(db->object_changed_event, entry);
		callback_fire_event(entry->object_handle->object_changed_event,
				    entry);
	    }
	}
    }
}

void po_handle_entry_gone(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *previous_entry;
    int32 deleted_flag;

    callback_fire_event(db->entry_gone_event, entry);

    /* REAL_TIME_CLASSes are simple.  They exist regardless of
     * the world state the application is interested in.
     */
    if (PO_REAL_TIME_CLASS(entry->pdu->variant.describeObject.pclass))
    {
	callback_fire_event(db->object_gone_event, entry);

	/* The user may be expecting this to be NULL */
	po_set_object_user_data(db, entry, (PO_USER_DATA_TYPE)NULL);

	if (entry == db->current_world_state)
	  po_world_state_lost(db);

	return;
    }

    /* First find the previous entry */
    if (entry->object_handle &&
	(previous_entry = entry->object_handle->current_entry))
    {
	if (previous_entry->deleted ||
	    previous_entry->implied)
	  previous_entry = NULL;
    }

    /* This object will be going away, but it is probably not marked
     * as deleted just yet.  In order to figure out what we should be
     * telling the application, we set its deleted flag temporarily to
     * find out what the world will look like when it is gone.
     */

    /* Set the flag */
    deleted_flag = entry->deleted;
    entry->deleted = TRUE;

    /* Now note the new current entry */
    entry->object_handle->current_entry =
      po_current_object(db, &entry->pdu->variant.describeObject.objectID);

    /* Now change it back */
    entry->deleted = deleted_flag;

    /* Was this the current entry for this object? */
    if (entry == previous_entry)
    {
	if (entry->pdu->variant.describeObject.missingFromWorldState)
	{
	    /* If this object was missing, then it may look like a new
	     * object has shown up, depending upon what is there now.
	     */
	    if (entry->object_handle->current_entry &&
		!entry->object_handle->current_entry->pdu->
		variant.describeObject.missingFromWorldState)
	      callback_fire_event(db->new_object_event,
				  entry->object_handle->current_entry);
	}
	else
	{
	    /* When this object goes away, this will either look like
	     * a change or a deletion, depending upon what is left.
	     */
	    if (entry->object_handle->current_entry &&
		!entry->object_handle->current_entry->pdu-> 
		  variant.describeObject.missingFromWorldState)
	    {
		callback_fire_event(db->object_changed_event,
				    entry->object_handle->current_entry);
		callback_fire_event(entry->object_handle->object_changed_event,
				    entry->object_handle->current_entry);
	    }
	    else
	    {
		callback_fire_event(db->object_gone_event, entry);

		/* The user may be expecting this to be NULL */
		po_set_object_user_data(db, entry, (PO_USER_DATA_TYPE)NULL);
	    }
	}
    }
}
