#ifndef lint
static char rcsid [] = "$RCSfile: xtest.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: xtest.c                                                           *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Xwindows test program for libpo                           *
*       Created: Mon Apr 15 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#define SPEED_FACTOR         200
#define RELIABILITY          0.98 /* actual  ~ 1499/1500 = .9993333*/
#define LATENCY_MS           100
#define VARIANCE_MS          20
#define NUM_DATABASES 4
#define ACTIVE_TEST(x) (x%4)
#define FNCL_GROUP(x) (x+1)

#include "libpo.h"
#include <stdext.h>   /* common/include/global */
#include <stdalloc.h> /* common/include/global */
#include <stdio.h>
#include <p_size.h>
#include <libsched.h>
#include <libtime.h>
#include <libcallback.h>
#include <librandom.h>

/*
 * X includes:
 */
#undef Boolean
#undef Time
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/Form.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/DrawingA.h>
#ifdef USECOA
#include <libcoa.h>
#endif
#include "test_Xsw.h"

#define EXERCISE_ID 182

ObjectID null_object = { { 0 , 0 } , 0 };
ObjectID g_stealth_id = { { 6 , 6 } , 6 };

struct
{
    int32 network_up;
    PO_DATABASE *db;
    Widget shell;
    Widget form;
    Widget area;
    Widget message;
    Widget sims;
    Widget overlays;
    Widget rtws;
#ifdef USECOA
    Widget coa_controls;
    Widget coa;
#endif
    Widget clock;
    uint32 last_clock;
} host[NUM_DATABASES];

TerrainDatabaseID terrain = { "Knox" , 1 };

send_after_latency(db, pdu, pdu_size)
    PO_DATABASE *db;
    PersistentObjectPDU *pdu;
    int32 pdu_size;
{
    po_process_packet(db, pdu);
    STDDEALLOC(pdu);
}

send_packet(pdu, pdu_size, from_whom)
    PersistentObjectPDU *pdu;
    int32 pdu_size;
    int32 from_whom;
{
    PersistentObjectPDU *pdu_copy;
    int32 i;

    if (!host[from_whom].network_up)
      return;

    for(i=0;i<NUM_DATABASES;i++)
      if (i != from_whom)
      {
	  if (host[i].network_up)
	  {
	      if (rnd_probability(RELIABILITY))
	      {
		  pdu_copy = (PersistentObjectPDU *)STDALLOC(pdu_size);
		  bcopy(pdu, pdu_copy, pdu_size);
		  sched_deferred_fncl(send_after_latency,
				      LATENCY_MS - VARIANCE_MS/2 +
				      (int32)(rnd_uniform_dist() * VARIANCE_MS),
				      FNCL_GROUP(i),
				      A_PTR, host[i].db,
				      A_PTR, pdu_copy,
				      A_INT, pdu_size,
				      A_END);
	      }
	      else
	      {
		  printf("Dropping pkt type %d from %s.\n",
			 pdu->kind,
			 host[from_whom].db->hostname);
	      }
	  }
      }
}

extern give_stealth();

new_sim(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    Widget btn;

    btn = XtCreateManagedWidget(po_entry_owner(host[which].db, entry),
				xmLabelWidgetClass, host[which].sims,
				NULL, 0);

    po_set_object_user_data(host[which].db, entry, btn);
    XswSetUserData(btn, entry);

    XtAddEventHandler(btn, ButtonPressMask, FALSE, give_stealth, which);
}

gone_sim(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    XtDestroyWidget(entry->object_user_data);
}

del_all(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    printf("Delete All on Host %d.  Triggered by %s\n",which+1,
	   po_entry_owner(host[which].db, entry));
}

extern make_text();
extern change_text();
extern describe();
extern delete_object();

#define NAME_ARG             0
#define XTNUMBER_WORLD_STATE 1
#define COLOR_ARG            1
#define XTNUMBER_OVERLAY     2
#define X_ARG                2
#define Y_ARG                3
#define XTNUMBER_TEXT        4
static Arg obj_args[] = { { 0 , 0 } , /* Placeholder for name */
			  { 0 , 0 } , /* Placeholder for color */
			  { 0 , 0 } , /* Placeholder for x */
			  { 0 , 0 } , /* Placeholder for y */
			 };

int32 should_show(overlay, which)
    PO_DB_ENTRY *overlay;
    int32 which;
{
    /* Ignore other folks scratch overlays */
    return !((PO_OVERLAY_DATA(overlay).scratch &&
	      ((PO_OBJECT_ID(overlay).simulator.site !=
		host[which].db->sim_addr.site) ||
	       (PO_OBJECT_ID(overlay).simulator.host !=
		host[which].db->sim_addr.host))));
}

new_obj(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    int32 color;
    Widget btn;
    PO_DB_ENTRY *overlay;

    switch (PO_OBJECT_CLASS(entry))
    {
      case objectClassOverlay:
	if (!should_show(entry, which))
	  return;

	XtSetArg(obj_args[NAME_ARG], XmNlabelString,
		 XmStringCreate(PO_OVERLAY_DATA(entry).name,
				XmSTRING_DEFAULT_CHARSET));
	XtSetArg(obj_args[COLOR_ARG], XmNforeground,
		 pixel_of_po_color(PO_OVERLAY_DATA(entry).color));

	btn = XtCreateManagedWidget("overlay", xmPushButtonWidgetClass,
				    host[which].overlays,
				    obj_args, XTNUMBER_OVERLAY);
	XtAddCallback(btn, XmNactivateCallback, make_text, which);
	break;
      case objectClassText:
	if ((overlay = po_find_overlay(host[which].db, entry)) &&
	    !should_show(overlay, which))
	  return;

	XtSetArg(obj_args[NAME_ARG], XmNlabelString,
		 XmStringCreate(PO_TEXT_DATA(entry).text,
				XmSTRING_DEFAULT_CHARSET));
	if (!(color = po_object_color(host[which].db, entry)))
	  printf("Error: %s\n",po_errlist[po_errno]);
	XtSetArg(obj_args[COLOR_ARG], XmNforeground,
		 pixel_of_po_color(color));
	XtSetArg(obj_args[X_ARG], XmNx, PO_TEXT_DATA(entry).location.x);
	XtSetArg(obj_args[Y_ARG], XmNy, PO_TEXT_DATA(entry).location.y);

	btn = XtCreateManagedWidget("text", xmPushButtonWidgetClass,
				    host[which].area,
				    obj_args, XTNUMBER_TEXT);
	XtAddCallback(btn, XmNactivateCallback, change_text, which);
	break;
      default:
	return;
    }

    po_set_object_user_data(host[which].db, entry, btn);
    XswSetUserData(btn, entry);
    XtAddEventHandler(btn, EnterWindowMask, FALSE, describe, which);
    XtAddEventHandler(btn, ButtonPressMask, FALSE, delete_object, which);
}

obj_chg(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    int32 color;

    if (!entry->object_user_data)
      return;

    switch (PO_OBJECT_CLASS(entry))
    {
      case objectClassOverlay:
	XtSetArg(obj_args[NAME_ARG], XmNlabelString,
		 XmStringCreate(PO_OVERLAY_DATA(entry).name,
				XmSTRING_DEFAULT_CHARSET));
	XtSetArg(obj_args[COLOR_ARG], XmNforeground,
		 pixel_of_po_color(PO_OVERLAY_DATA(entry).color));

	XtSetValues(entry->object_user_data, obj_args, XTNUMBER_OVERLAY);
	break;
      case objectClassText:
	XtSetArg(obj_args[NAME_ARG], XmNlabelString,
		 XmStringCreate(PO_TEXT_DATA(entry).text,
				XmSTRING_DEFAULT_CHARSET));
	if (!(color = po_object_color(host[which].db, entry)))
	  printf("Error: %s\n",po_errlist[po_errno]);
	XtSetArg(obj_args[COLOR_ARG], XmNforeground,
		 pixel_of_po_color(color));
	XtSetArg(obj_args[X_ARG], XmNx, PO_TEXT_DATA(entry).location.x);
	XtSetArg(obj_args[Y_ARG], XmNy, PO_TEXT_DATA(entry).location.y);
	XtSetValues(entry->object_user_data, obj_args, XTNUMBER_TEXT);
	break;
      default:
	return;
    }
    XswSetUserData(entry->object_user_data, entry);
}

obj_gone(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    if (!entry->object_user_data)
      return;

    switch (PO_OBJECT_CLASS(entry))
    {
      case objectClassOverlay:
      case objectClassText:
	XtDestroyWidget(entry->object_user_data);
	break;
      default:
	return;
    }
}

obj_ch_fail(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    printf("Object change by Host %d overriden by %s\n",which,
	   po_entry_owner(host[which].db, entry));
}

ws_changed(entry, which)
    PO_DB_ENTRY *entry;
    int32 which;
{
    if (entry)
    {
	XmToggleButtonSetState((Widget)entry->object_user_data, TRUE, FALSE);
    }
    else
    {
	XmToggleButtonSetState(host[which].rtws, TRUE, FALSE);
    }
}

project(vid, do_project, which)
    ObjectID *vid;
    int32 do_project;
    int32 which;
{
    printf("%s: %s projecting to %d/%d/%d\n",
	   host[which].db->hostname,
	   do_project ? "Start" : "Stop",
	   vid->simulator.site,vid->simulator.host,vid->object);
}

pmissed(entry, num_missed, which)
    PO_DB_ENTRY *entry;
    uint32 num_missed;
    int32 which;
{
    printf("%s: Missed %d packets from %s\n",
	   host[which].db->hostname,
	   num_missed,
	   po_entry_owner(host[which].db,entry));
}

handle_new_exercise(data, which)
    ExerciseInitializerClass *data;
    int32                     which;
{
    printf("%s: New exercise info.  Terrain now %s\n",
	   host[which].db->hostname,
	   data->terrain.terrainName);
}


make_new_overlay(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    static int32 count = 0;
    OverlayClass variant;

    sprintf(variant.name,"<%d>",count);
    variant.color = (count%5)+1;
    /* Black overlays will be scratch */
    variant.scratch = (variant.color == 1);
    if (!po_create_object(host[which].db, NULL, objectClassOverlay,
			  FALSE,
			  &variant, PRO_PO_OVERLAY_CLASS_SIZE, NULL))
      printf("Error: %s\n",po_errlist[po_errno]);
    count += 1;
}

adjust_host_network(w,which,call_data)
    Widget w;
    int32 which;
    XmToggleButtonCallbackStruct *call_data;
{
    host[which].network_up = call_data->set;
}

static ObjectID last_text = { { 0 , 0 } , 0};

make_text(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    XWindowAttributes att;
    PO_BUFFER(TextClass, variant, buf);
    static char count = 'A';
    PO_DB_ENTRY *overlay_entry = (PO_DB_ENTRY *)XswGetUserData(w);
    PO_DB_ENTRY *result;

    XGetWindowAttributes(XtDisplay(host[which].area),
			 XtWindow(host[which].area),
			 &att);

    sprintf(variant->text,"{%c}",count);
    variant->overlayID = PO_OBJECT_ID(overlay_entry);
    variant->color = OCOverlayDefault;
    variant->length = strlen(variant->text) + 1;
    variant->alignment = center;
    variant->location.x = rnd_uniform_dist() * (att.width - 50);
    variant->location.y = rnd_uniform_dist() * (att.height - 50);
    variant->verticalOffset = 0;
    variant->horizontalOffset = 0;
    variant->associatedObject = last_text;
    variant->associatedPointNumber = 0;

    if (!(result =
	  po_create_object(host[which].db, NULL, objectClassText,
			   FALSE,
			   variant, PRO_PO_TEXT_CLASS_SIZE(variant->length),
			   NULL)))
    {
	if (po_errno == PO_EASSOC)
	{
	    variant->associatedObject = null_object;
	    if (!(result =
		  po_create_object(host[which].db, NULL, objectClassText,
				   FALSE,
				   variant,
				   PRO_PO_TEXT_CLASS_SIZE(variant->length),
				   NULL)))
	    {
		printf("Error: %s\n",po_errlist[po_errno]);
		return;
	    }
	}
	else
	  return;
    }
    count += 1;
    last_text = result->pdu->variant.describeObject.objectID;
}

change_text(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    PO_BUFFER(TextClass, variant, buf);
    PO_DB_ENTRY *entry = (PO_DB_ENTRY *)XswGetUserData(w);

    bcopy(&PO_TEXT_DATA(entry),variant,
	  PRO_PO_TEXT_CLASS_SIZE(PO_TEXT_DATA(entry).length));
    sprintf(variant->text,"%s%c",PO_TEXT_DATA(entry).text,'a'+which);
    variant->length += 1;

    if (!po_change_object(host[which].db, entry,
			  variant, PRO_PO_TEXT_CLASS_SIZE(variant->length)))
      printf("Error: %s\n",po_errlist[po_errno]);
}

describe(w,which)
    Widget w;
    int32 which;
{
    PO_DB_ENTRY *entry = (PO_DB_ENTRY *)XswGetUserData(w);
    char buf[1024];

    sprintf(buf,"%d/%d/%d in WS %d/%d/%d: Seq #%d, Owner %s %s",
	    entry->pdu->variant.describeObject.objectID.simulator.site,
	    entry->pdu->variant.describeObject.objectID.simulator.host,
	    entry->pdu->variant.describeObject.objectID.object,
	    entry->pdu->variant.describeObject.worldStateID.simulator.site,
	    entry->pdu->variant.describeObject.worldStateID.simulator.host,
	    entry->pdu->variant.describeObject.worldStateID.object,
	    entry->pdu->variant.describeObject.sequenceNumber,
	    po_entry_owner(host[which].db, entry),
	    entry->dormant ? "(dormant)" : "");
    XmTextSetString(host[which].message, buf);
}

new_stealth(db)
    PO_DATABASE *db;
{
    po_new_stealth(db, &g_stealth_id);
}

static ExerciseInitializerClass g_new_ex =
{
    { "T.Not.Set", 0 },
    battleSchemeAbsolute,
    0,
    1.0,
    0,
};

static void new_exercise(db, i)
    PO_DATABASE *db;
    int i;
{
    printf("%s now changing exercise to use terrain %d\n",
	   host[i].db->hostname, i);
    sprintf(g_new_ex.terrain.terrainName, "<t %d>", i);
    po_set_exercise_initialization(db, &g_new_ex);
}

static void check_exercise(db, i)
    PO_DATABASE *db;
    int i;
{
    ExerciseInitializerClass *result = po_get_exercise_initialization(db);

    printf("%s randomly checked exercise and discoverd terrain %s\n",
	   host[i].db->hostname, result->terrain.terrainName);
}

give_stealth(w,which,event)
    Widget w;
    int32 which;
    XEvent *event;
{
    PO_DB_ENTRY *entry;

    switch(event->xbutton.button)
    {
      case 1:
	entry = (PO_DB_ENTRY *)XswGetUserData(w);
	po_control_stealth(host[which].db, &g_stealth_id, entry);
    }
}

delete_object(w,which,event)
    Widget w;
    int32 which;
    XEvent *event;
{
    PO_DB_ENTRY *entry;

    switch(event->xbutton.button)
    {
      case 2:
	entry = (PO_DB_ENTRY *)XswGetUserData(w);

	if (PO_OBJECT_CLASS(entry) == objectClassWorldState)
	  break;

	if (PO_OBJECT_CLASS(entry) == objectClassOverlay)
	{
	    if (!po_save_overlay(host[which].db, "overlay",
				 entry,
				 PO_CLASS_MASK(objectClassOverlay) |
				 PO_CLASS_MASK(objectClassText)))
	      printf("Error: %s\n",po_errlist[po_errno]);
	    break;
	}
	    

	if (!po_change_object_missing_flag(host[which].db, entry, TRUE))
	  printf("Error: %s\n",po_errlist[po_errno]);

	break;
      case 3:
	entry = (PO_DB_ENTRY *)XswGetUserData(w);

	if (!po_delete_object(host[which].db, entry))
	  printf("Error: %s\n",po_errlist[po_errno]);

	break;
    }
}

kill_host(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
#ifdef USECOA
    coa_popdown_controls(host[which].coa_controls);
#endif
    sched_cancel_fncl_group(FNCL_GROUP(which));
    XtDestroyWidget(host[which].shell);
    po_destroy(host[which].db, TRUE);
    host[which].network_up = FALSE;
}

kill_db(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    po_delete_all(host[which].db);
}

#ifdef USECOA

toggle_COA(w,which,call_data)
    Widget w;
    int32 which;
    XmToggleButtonCallbackStruct *call_data;
{
    PO_DB_ENTRY *entry;

    if (call_data->set)
      coa_popup_controls(host[which].coa_controls);
    else
      coa_popdown_controls(host[which].coa_controls);
}

#endif

save_all(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    printf("%s saving realtime objects in file \'all\'\n",
	   host[which].db->hostname);
    if (!po_save_all(host[which].db, "all", FALSE, TRUE, FALSE))
      printf("Error: %s\n",po_errlist[po_errno]);
}

struct exist_struct
{
    char *name;
    int32 conflict;
};

overlay_name_match(entry, es)
    PO_DB_ENTRY *entry;
    struct exist_struct *es;
{
    if (es->conflict)
      return;

    if (PO_OBJECT_CLASS(entry) == objectClassOverlay)
      es->conflict = (strncmp(PO_OVERLAY_DATA(entry).name, es->name,
			      maxOverlayNameLength) == 0);
}

confirm_overlay(name, db)
    char *name;
    PO_DATABASE *db;
{
    struct exist_struct es;

    es.name = name;
    es.conflict = 0;
    po_query_for_current_objects(db, overlay_name_match, &es);
    return !es.conflict;
}

load_all(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    if (!po_load_file(host[which].db, "all",
		      confirm_overlay, host[which].db, NULL))
      printf("Error: %s\n",po_errlist[po_errno]);
}

load_overlay(w,which,call_data)
    Widget w;
    int32 which;
    XmAnyCallbackStruct *call_data;
{
    if (!po_load_file(host[which].db, "overlay",
		      confirm_overlay, host[which].db, NULL))
      printf("Error: %s\n",po_errlist[po_errno]);
}

update_clock(which)
    int32 which;
{
    char buf[20];
    uint32 clock = po_time(host[which].db)/1000;

    if (clock != host[which].last_clock)
    {
	sprintf(buf,"%d",clock);
	XmTextSetString(host[which].clock, buf);
	host[which].last_clock = clock;
    }
}

XtAppContext g_app_context = NULL;
Widget toplevel = NULL;
Widget g_dialog_shell;

static tick_X()
{
    XEvent event;
    int32 pend;

    while (pend = XtAppPending(g_app_context))
    {
	if (pend & (XtIMXEvent  | XtIMTimer))
	{
	    XtAppProcessEvent(g_app_context, pend);
	}
	else
	  break;
    }
}

static Arg shell_args[] = { { XmNgeometry, (XtArgVal)"700x210" } };
static Arg form_args[] = { { XmNresizePolicy , XmRESIZE_NONE } };
static Arg sims_args[] = { { XmNorientation , XmHORIZONTAL } };
static Arg ws_args[] = { { XmNorientation , XmHORIZONTAL } ,
			 { XmNradioBehavior , TRUE } ,
		       };

init_x(argc,argv)
    int32 argc;
    char *argv[];
{
    int32 copy_argc = argc;
    int32 i;
    char buf[20];
    Widget network, overlay, sa_b, la_b, lp_b, del_all_b, quitb, sb;

    toplevel = XtInitialize(argv[0],"POTest", NULL, 0,
			    &copy_argc,argv);
#ifdef USECOA
    XswInitDialogShell(toplevel);
    XswInitDisplay(XtDisplay(toplevel));
#endif
    g_app_context = XtWidgetToApplicationContext(toplevel);

    for (i=0;i<NUM_DATABASES;i++)
    {
	sprintf(buf,"Host %d (%s)",i+1,ACTIVE_TEST(i) ? "active" : "passive");
	host[i].shell = XtCreatePopupShell(buf,
					   topLevelShellWidgetClass,
					   toplevel,
					   shell_args, XtNumber(shell_args));

	host[i].form = XtCreateManagedWidget("form", xmFormWidgetClass,
					     host[i].shell,
					     form_args, XtNumber(form_args));

	host[i].area = XtCreateManagedWidget("form", xmDrawingAreaWidgetClass,
					     host[i].form,
					     NULL, 0);

	host[i].message = XmCreateText(host[i].form, "message", NULL, 0);
	XtManageChild(host[i].message);

	host[i].clock = XmCreateText(host[i].form, "clock", NULL, 0);
	XtManageChild(host[i].clock);

	host[i].sims = XtCreateManagedWidget("sims", xmRowColumnWidgetClass,
					     host[i].form,
					     sims_args, XtNumber(sims_args));
	sb = XtCreateManagedWidget("Known Simulators:",
				   xmLabelWidgetClass, host[i].sims, NULL, 0);
	XswSetUserData(sb, NULL);
	XtAddEventHandler(sb, ButtonPressMask, FALSE, give_stealth, i);

	network = XtCreateManagedWidget("Network",
					xmToggleButtonWidgetClass,
					host[i].form, NULL, 0);
	XtAddCallback(network, XmNvalueChangedCallback,
		      adjust_host_network, i);
	XmToggleButtonSetState(network, TRUE, FALSE);

	overlay = XtCreateManagedWidget("New Overlay",
					xmPushButtonWidgetClass,
					host[i].form, NULL, 0);
	XtAddCallback(overlay, XmNactivateCallback,
		      make_new_overlay, i);

#ifdef USECOA

	host[i].coa = XtCreateManagedWidget("COA",
					    xmToggleButtonWidgetClass,
					    host[i].form, NULL, 0);
	XtAddCallback(host[i].coa, XmNvalueChangedCallback,
		      toggle_COA, i);

#endif
	sa_b = XtCreateManagedWidget("Save All",
				     xmPushButtonWidgetClass,
				     host[i].form, NULL, 0);
	XtAddCallback(sa_b, XmNactivateCallback,
		      save_all, i);

	la_b = XtCreateManagedWidget("Load All",
				     xmPushButtonWidgetClass,
				     host[i].form, NULL, 0);
	XtAddCallback(la_b, XmNactivateCallback,
		      load_all, i);

	lp_b = XtCreateManagedWidget("Load Overlay",
				     xmPushButtonWidgetClass,
				     host[i].form, NULL, 0);
	XtAddCallback(lp_b, XmNactivateCallback,
		      load_overlay, i);

	del_all_b = XtCreateManagedWidget("Delete All",
					  xmPushButtonWidgetClass,
					  host[i].form, NULL, 0);
	XtAddCallback(del_all_b, XmNactivateCallback,
		      kill_db, i);

	quitb = XtCreateManagedWidget("Quit",
				     xmPushButtonWidgetClass,
				     host[i].form, NULL, 0);
	XtAddCallback(quitb, XmNactivateCallback,
		      kill_host, i);

	host[i].overlays = XtCreateManagedWidget("overlays",
						 xmRowColumnWidgetClass,
						 host[i].form,
						 sims_args,
						 XtNumber(sims_args));
	XtCreateManagedWidget("Overlays:",
			      xmLabelWidgetClass, host[i].overlays, NULL, 0);

	XswLayoutWidget(host[i].clock, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_NONE, 0,
			XmATTACH_FORM, 0,
			XmATTACH_FORM, 0);

	XswLayoutWidget(host[i].message, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_FORM, 0,
			XmATTACH_WIDGET, 5, host[i].clock,
			XmATTACH_FORM, 0);

	XswLayoutWidget(host[i].sims, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_FORM, 0,
			XmATTACH_FORM, 0,
			XmATTACH_WIDGET, 0, host[i].message);

	XswLayoutWidget(network, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_FORM, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);

	XswLayoutWidget(overlay, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, network,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);

#ifdef USECOA
	XswLayoutWidget(host[i].coa, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, overlay,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);

	XswLayoutWidget(sa_b, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].coa,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);
#else
	XswLayoutWidget(sa_b, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, overlay,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);
#endif

	XswLayoutWidget(la_b, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, sa_b,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);

	XswLayoutWidget(lp_b, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, la_b,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, host[i].sims);

	XswLayoutWidget(quitb, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_NONE, 0,
			XmATTACH_FORM, 0,
			XmATTACH_WIDGET, 0, host[i].sims);

	XswLayoutWidget(del_all_b, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_NONE, 0,
			XmATTACH_WIDGET, 0, quitb,
			XmATTACH_WIDGET, 0, host[i].sims);

	XswLayoutWidget(host[i].overlays, 0, 0,
			XmATTACH_NONE, 0,
			XmATTACH_FORM, 0,
			XmATTACH_FORM, 0,
			XmATTACH_WIDGET, 0, network);

	XswLayoutWidget(host[i].area, 0, 0,
			XmATTACH_FORM, 0,
			XmATTACH_FORM, 0,
			XmATTACH_FORM, 0,
			XmATTACH_WIDGET, 0, host[i].overlays);

	XtRealizeWidget(host[i].shell);
	XtPopup(host[i].shell, XtGrabNone);
    }

    g_dialog_shell = host[0].shell;

    sched_periodic_fncl(tick_X, 500, 10, 0, A_END);
}


static char *color_names[] = { "gray" ,
				 "black" ,
				 "yellow" ,
				 "red" ,
				 "green" ,
				 "blue" };

static Pixel color_pixels[XtNumber(color_names)];

pixel_of_po_color(color)
    int32 color;
{
    return color_pixels[color];
}

init_pixels()
{
    XColor ignored, ret;
    int32 color;

    for (color=0;color<XtNumber(color_names);color++)
    {
	XAllocNamedColor(XtDisplay(toplevel),
			 DefaultColormap(XtDisplay(toplevel), 0),
			 color_names[color],
			 &ret, &ignored);
	color_pixels[color] = ret.pixel;
    }
}

main(argc, argv)
    int32 argc;
    char *argv[];
{
    SimulationAddress addr;
    int32 i;
    int32 j;
    char buf[20];
    XColor scr, exc;
    int32 black, red;

    rnd_srandom(time_realtime_clock());

    time_init();

    sched_init(3,10,50,100);

    init_x(argc, argv);

    init_pixels();

    XAllocNamedColor(XtDisplay(toplevel),
		     DefaultColormapOfScreen(XtScreen(toplevel)),
		     "Black", &scr, &exc);
    black = scr.pixel;
    XAllocNamedColor(XtDisplay(toplevel),
		     DefaultColormapOfScreen(XtScreen(toplevel)),
		     "Red", &scr, &exc);
    red = scr.pixel;

    for (i=0;i<NUM_DATABASES;i++)
    {
#ifdef USECOA
	host[i].coa_controls = coa_make_controls(toplevel, black, red,
						 host[i].coa, NULL, 1);
#endif
	sprintf(buf,"Host %d",i+1);

	addr.site = i+1;
	addr.host = i+1;
	host[i].network_up = TRUE;
	host[i].db = po_create(ACTIVE_TEST(i) ?
			       PO_DATABASE_TYPE_ACTIVE :
			       PO_DATABASE_TYPE_PASSIVE,
			       EXERCISE_ID, 0, &addr,
			       0, &terrain, buf, simulatorUnknown,
			       send_packet, i, FALSE);

	callback_register_handler(host[i].db->new_simulator_event,
				  new_sim, i);
	callback_register_handler(host[i].db->simulator_gone_event,
				  gone_sim, i);
#ifdef USECOA
	callback_register_handler(host[i].db->new_entry_event,
				  coa_handle_new_entry, host[i].coa_controls);
	callback_register_handler(host[i].db->entry_changed_event,
				  coa_handle_entry_changed, 
				  host[i].coa_controls);
	callback_register_handler(host[i].db->entry_gone_event,
				  coa_handle_entry_gone, host[i].coa_controls);
#endif
	callback_register_handler(host[i].db->new_object_event,
				  new_obj, i);
	callback_register_handler(host[i].db->object_changed_event,
				  obj_chg, i);
	callback_register_handler(host[i].db->object_gone_event,
				  obj_gone, i);
	callback_register_handler(host[i].db->object_change_failed_event,
				  obj_ch_fail, i);
	callback_register_handler(host[i].db->delete_all_event,
				  del_all, i);
#ifdef USECOA
	/* Nothing to add to host[i].db->world_state_changing_event */

	callback_register_handler(host[i].db->world_state_changed_event,
				  coa_handle_world_state_changed,
				  host[i].coa_controls);
	callback_register_handler(host[i].db->animation_event,
				  coa_handle_animation,
				  host[i].coa_controls);
	callback_register_handler(host[i].db->animation_timeout_event,
				  coa_handle_animation_timeout,
				  host[i].coa_controls);
#endif
	callback_register_handler(host[i].db->project_event,
				  project, i);
	callback_register_handler(host[i].db->packets_missed_event,
				  pmissed, i);

	callback_register_handler(host[i].db->exercise_initialization_event,
				  handle_new_exercise, i);
	if (!host[i].db)
	{
	    printf("Error: %s\n",po_errlist[po_errno]);
	    exit(0);
	}

#ifdef USECOA
	coa_init_controls(host[i].coa_controls, host[i].db);
#endif
	sched_periodic_fncl(po_tick, 100, 100,
			    FNCL_GROUP(i), A_PTR, host[i].db, A_END);

	sched_deferred_fncl(new_stealth, 5000, 0, A_PTR, host[i].db, A_END);

	/*
	 * Cause each ACTIVE db to try making a new exercise info.
	 */
	if (ACTIVE_TEST(i))
	  sched_deferred_fncl(new_exercise, (10000 + i * 40000), 0, A_PTR,
			      host[i].db, A_INT, i, A_END);

	/*
	 * Cause each db to check on the exercise at certain points.
	 */
	for(j = 0; j < 5; j++)
	  sched_deferred_fncl(check_exercise, (5000+(i * 1000)+(j * 30000)),
			      0, A_PTR,
			      host[i].db, A_INT, i, A_END);

	sched_periodic_fncl(update_clock, 5000, 100,
			    FNCL_GROUP(i), A_INT, i, A_END);
    }

    sched_invoke_functions_until(0xFFFFFFFF);
}



#if 0
char *enp_device  =            "/dev/enp0";
char *assoc_def_file         = "/saf/bin/assoc.def";

    if ((g_simnet_handle = AssocOpen(enp_device, assoc_def_file)) ==
	-1)
    {
	printf("Error from association layer (on init):\n  %s\n",
	       AssocError());
	exit (-1);
    }


    AssocSubscribe(g_simnet_handle, EXERCISE_ID, poProtocolNumber);

    /* Get the site and host from the assoc layer */
    if (AssocGetSimAddress(po_simnet_handle, &po_sim_address) == -1)
    {
	fprintf(stderr, "Failed to get simulation address\n");
	exit(0);
    }


    AssocUnsubscribe(g_simnet_handle, EXERCISE_ID, poProtocolNumber);

    AssocClose(g_simnet_handle);
#endif
