#ifndef _LIBPRIQUEUE_H_INCLUDED
#define _LIBPRIQUEUE_H_INCLUDED

/*
 * $RCSfile: libpriqueue.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpriqueue.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libpriqueue                        *
*       Created: June 27 1994                                               *
*       Author: swarts                                                      *
*       Remarks: This library provide a heap (data structure) based priority
*		 queue.
*                                                                           *
****************************************************************************/

#include <stdext.h>

typedef struct pq_opaque PRIORITY_QUEUE;

extern PRIORITY_QUEUE* pq_create(int32 initial_max_size, int32 expand_chunk);
    /* Creates a PRIORITY_QUEUE.  Enough memory is allocated for 
     * initial_max_size elements.  expand_chunk is the amount to expand the
     * max_size by if it is exceeded.
     */

extern void pq_delete(PRIORITY_QUEUE *queue,
		      void (*data_deletion_func) (/* ??? */));
    /* Deletes the queue.   The queue memory is deallocated.  If
     * data_deletion_func is not NULL then it is called for each non-NULL
     * data on the queue.
     */

extern void pq_insert(PRIORITY_QUEUE *queue, uint32 priority, ADDRESS data);
    /* Inserts an element on the queue with priority priority and
     * data data.  If necessary the queue is expanded.
     */

extern int32 pq_empty(PRIORITY_QUEUE *queue);
    /* Returns whether or not the queue is empty.
     */

extern int32 pq_in_queue(PRIORITY_QUEUE *queue, ADDRESS data);
    /* Returns whether or not data is in the queue
     */

extern int32 pq_min(PRIORITY_QUEUE *queue, uint32 *priority, ADDRESS *data);
    /* If the queue is empty then return FALSE.  Otherwise, returns TRUE
     * and fills in priority and data from the minimum priority element.
     * Does not modify the queue.  It is safe to pass NULL for priority
     * or data.
     */


extern int32 pq_extract_min(PRIORITY_QUEUE *queue,
			    uint32 *priority,
			    ADDRESS *data);
    /* If the queue is empty then return FALSE.  Otherwise, returns TRUE,
     * fills in priority and data from the minimum priority element,
     * and removes that element.  It is safe to pass NULL for priority
     * or data.
     */

extern int32 pq_remove(PRIORITY_QUEUE *queue,
		       ADDRESS item,
		       void (*data_delete_func) (/* ??? */));
    /* Removes data_to_delete from the queue, if it is there.  If
     * data_delete_func is not NULL then it is called on data when
     * it is removed.  Returns whether or not the item was found.
     */

extern void pq_remove_matching(PRIORITY_QUEUE *queue,
			       int32 (*match_func) (/* ??? */),
			       ADDRESS passed_data,
			       void (*data_delete_func) (/* ??? */));
    /* For each item in the queue calls match_func on the item's data
     * and passed_data.  If match_func returns TRUE then the item is
     * removed from the queue.  If data_delete_func is not NULL then it
     * is called on each data as it is removed.
     */

extern int32 pq_remove_first_matching(PRIORITY_QUEUE *queue,
				      int32 (*match_func) (/* ??? */),
				      ADDRESS passed_data,
				      void (*data_delete_func) (/* ??? */));
    /* Loops through the queue looking for an item where match_func func
     * returns TRUE.  If such an item is found then it is removed, 
     * if data_delete_func is not NULL it is called on the data, and
     * TRUE is returned.  If no matching item is found then returns FALSE.
     */

extern void pq_print(PRIORITY_QUEUE *queue);
    /* Prints the queue to stdout.
     */

extern void pq_check(PRIORITY_QUEUE *queue);
    /* Checks the queue invariant properties.  If a problem is found then
     * prints the queue and aborts.  Used for testing.
     */

#endif /*_LIBPRIQUEUE_H_INCLUDED*/

