
/*
 * FILE:  cvcc_innotr.h 
 *
 * Created on Wed Apr 17 10:52:36 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _cvcc_innotr_included
#define _cvcc_innotr_included



/*
 * FILE: cvcc_innotr.drn 
 *
 * This file contains the PDU definitions used for innovative 
 * training exercises for CVCC, involving interactions between
 * CVCC-Send and the CCD workstations.
 *
 *  Copyright (c) 1992.  All rights reserved.
 *  BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 *  History:
 *	06/23/92  Wilbert	created
 *
 * $RCSfile: cvcc_innotr.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <cvcc_instr.h> 
#define cvcc_MaxMsgTextLen      1024 /* Max length of many kinds of text */
#define cvcc_MaxNameLen         32 /* Max length of a student name */
 /* The following are the additional types of messages used during an */
 /* innovative training exercise. */
#define cvcc_OpOrdExtractMsg    0     /* Sends OpOrd text to students */
#define cvcc_TeachingObjMsg     1     /* Sends Teaching Obj. text to students */
#define cvcc_StudentNameMsg     2     /* Sends student name to coordinator */
#define cvcc_StartCollectionMsg 3     /* Notifies CCDs to start collecting  */
#define cvcc_EndCollectionMsg   4     /* Notifies CCDs to stop collecting */
#define cvcc_CoordinatorAlertMsg 5    /* Sends text with alert to student */
#define cvcc_CoordinatorTextMsg 6     /* Sends msg from coord to student */
#define cvcc_StudentReadyMsg    7     /* Named student is ready for exercise */
 /* or feedback */
#define cvcc_SMEReportActionsMsg 8    /* Provides SME's actions for a report */
 /* to the student */
#define cvcc_SMESitRepMsg       9     /* Provides SME's preferred SitRep */
 /* for training exercise. */
typedef uint8    cvcc_TrainingMessageType; 
 /* The following describes a student by the student's name and the computer */
 /* host's name.					-- 32 + 32 = 64 bytes */
typedef struct {
    char student[cvcc_MaxNameLen];   
    char host[cvcc_MaxNameLen];      
} cvcc_StudentName; 
 /* The following describes an alert message which includes direct addressing. */
 /* The direct addressing is "informal".  It is either the host name or the */
 /* student's name.  If the ccd finds a match on either, it assumes the */
 /* message is intended for it.			-- 32 + 1024 = 1056 bytes */
typedef struct {
    char name[cvcc_MaxNameLen];      
    char text[cvcc_MaxMsgTextLen];   
} cvcc_CoordinatorAlert; 
 /* The following is sent as part of a cvcc_SMEReportActionsMsg. There */
 /* should be one message of this type sent for every actual message */
 /* in the training exercise. This message provides the recommended */
 /* actions which should have been taken by the student upon receipt of */
 /* the message.			-- 12 + 2 + 2 + 12*1 + 1024 = 1052 bytes */
typedef struct {
    uint16 numActions;           /* Number of actions on report */

    uint16 _unused_40;           /* padding */
    ivis_MessageAction actions[ivis_MaxMessageActions]; 
} cvcc_ReportActions; 

typedef struct {
    cvcc_ObjID         report;        /* Identifies report */

    cvcc_ReportActions actions;      
    char                  rationale[cvcc_MaxMsgTextLen]; 
} cvcc_SMEReportActions; 
 /* The following is part fo the cvcc_SMESitRepMsg.  This contains the */
 /* SME's actual situation report, as well as his/her rationale for what's */
 /* in the report.				-- 28 + 1024 = 1052 bytes */
typedef struct {
    SituationReport rpt;             
    char               rationale[cvcc_MaxMsgTextLen]; 
} cvcc_SMESitReport; 
 /* CVCC Innovative Training messages */
 /* 				-- 1 + 3 + max(1052,64,1056,1052) = 1060 bytes */
typedef struct {
    cvcc_TrainingMessageType mtype;   /* type of message */

    unsigned      _unused_41 : 24;       /* padding */
    union {
        char                     text[cvcc_MaxMsgTextLen];
         /* 		when (cvcc_TeachingObjMsg)	text */
         /* 			array (cvcc_MaxMsgTextLen) of Character (8), */
        cvcc_StudentName      name;
         /* 		when (cvcc_StartCollectionMsg)	nothing, */
         /* 		when (cvcc_EndCollectionMsg)	nothing, */
        cvcc_CoordinatorAlert alert;
         /* 		when (cvcc_CoordinatorTextMsg)	text */
         /* 			array (cvcc_MaxMsgTextLen) of Character (8), */
         /* 		when (cvcc_StudentReadyMsg)	name	cvcc_StudentName, */
        cvcc_SMEReportActions report;
        cvcc_SMESitReport     sitRep;
    } u;                                
} cvcc_InnovativeTraining; 



#endif
