
/*
 * FILE:  cvcc_instr.h 
 *
 * Created on Wed Apr 17 10:52:37 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _cvcc_instr_included
#define _cvcc_instr_included



/*
 *  Name:  cvcc_instr.drn
 *  Author:  The CVCC gang
 *  Date Created:  December 10, 1991
 *  This file contains definitions for cvcc instrumentation
 *
 *  Copyright (c) 1991.  All rights reserves.
 *  BBN Systems and Technologies Corp., Cambridge, MA 02138
 *
 * $RCSfile: cvcc_instr.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <ivis_com.h> 
     /*  ************************************************************************** */

 /*  CITV instrumentation enumeration types: */

 /*  ************************************************************************** */
#define citvStateOff            0    
#define citvStateCooling        1    
#define citvStateSearchMode     2    
#define citvStateAScanMode      3    
#define citvStateGLOSMode       4    
#define citvStateGPSState       5    
#define citvStateStandby        6    

typedef uint8    CITVState; 

#define citvMag10x              0    
#define citvMag3x               1    

typedef uint8    CITVMagnification; 

#define citvPolarityBlackHot    0    
#define citvPolarityWhiteHot    1    

typedef uint8    CITVPolarity; 

#define noStackButtonPressed    0    
#define stackButton1Pressed     1    
#define stackButton2Pressed     2    
#define stackButton3Pressed     3    
#define stackButton4Pressed     4    

typedef uint8    StackButtonPressed; 
 /*  ************************************************************************** */

 /*  CITV instrumentation PDU: */
 /*  6 + 1 + 1 + 1 + 1/8 + 1/8 + 1/8 + 5/8 + 1 + 1 + 4 + 4 + 4 = 24 bytes */
 /*  ************************************************************************** */
typedef struct {
    VehicleID  vehicle;               /* identity of vehicle */

    CITVState  citvState;            
    CITVMagnification citvMagnification;  /* 3X or 10X */

    CITVPolarity citvPolarity;        /* black or white hot */

    unsigned      designatePushed           : 1;  /* Boolean */
    unsigned      cmdrStackOn               : 1;  /* Boolean */
    unsigned      gunnerStackOn             : 1;  /* Boolean */
    unsigned      _unused_24                : 5;  /* padding */
    StackButtonPressed cmdrStackButtonPushed;  /* 1 to 4, or none */

    StackButtonPressed gunnerStackButtonPushed;  /* 1 to 4, or none */

    float32         autoscanLeftSectorLimit;  /* radians, world coords. */

    float32         autoscanRightSectorLimit;  /* radians, world coords. */

    float32         autoscanRate;          /* radians per second */

} CITVInstrumentationVariant; 
 /*  ************************************************************************** */

 /*  IVIS instrumentation enumeration types: */

 /*  ************************************************************************** */
 /* ivis instrumenation types */
#define ivis_ReportGenerationInstr 0 
#define ivis_MessageInstr       1    
#define ivis_MapInstr           2    
#define ivis_QuantityInstr      3    

typedef uint8 ivis_InstrumentationKind; 
 /* ivis map scroll mode types */
#define ivis_MapJumpMode        0    
#define ivis_MapFollowMode      1    
#define ivis_MapMoveMode        2    

typedef uint8 ivis_MapScrollMode; 
 /* actions taken on ivis messages */
#define ivis_MaxMessageActions  12
#define ivis_RelayMessage       0    
#define ivis_RelayUp            1    
#define ivis_RelayDown          2    
#define ivis_SaveMessage        3    
#define ivis_DeleteMessage      4    
#define ivis_InspectMessage     5    
#define ivis_MessageReceived    6    
#define ivis_MessageRemovedFromReceiveQueue 7
#define ivis_MessageAgedFromReceiveQueue 8
#define ivis_PostMessage        9    
#define ivis_UnpostMessage      10   
#define ivis_MakeRouteActive    11   

typedef uint8 ivis_MessageAction; 
 /* actions taken upon creating a report */
#define ivis_CancelReport       0    
#define ivis_SendReport         1    

typedef uint8 ivis_ReportAction; 
 /* method by which a message is retrieved */
#define ivis_AlreadyRetrieved   0    
#define ivis_RetrieveByQueue    1    
#define ivis_RetrieveByFile     2    
#define ivis_RetrieveByIcon     3    
#define ivis_NotApplicable      4    

typedef uint8 ivis_RetrievalMethod; 
 /* quantity types */
#define ivis_HighPriorityMessagesInQueue 0
#define ivis_LowPriorityMessagesInQueue 1
#define ivis_ThumbDesignatorSelects 2
#define ivis_TouchScreenSelects 3    
#define ivis_MapLocationInputs  4    
#define ivis_LaserRangeFinderLocationInputs 5

typedef uint16 ivis_QuantityType; 

#define ivis_NumberQuantityTypes 8
     /*  ************************************************************************** */

 /*  IVIS instrumentation structure types: */

 /*  ************************************************************************** */
 /* report generation variant */
 /* 12 + 4 + 4 + 2 + 1 + 1 = 24 bytes */
typedef struct {
    cvcc_ObjID  messageID;            /* identity of a report object */

    SP_Time        activeTime;           /* active time to create report */

    SP_Time        elapsedTime;          /* time elapsed since user started report */

    uint16 totalInputs;         
     /* number of inputs required to compose report */

    ivis_ReportAction reportAction;   /* send or cancel */

    cvcc_ReportType reportType;       /* type of report created */

} ivis_ReportGenerationVariant; 
 /* message instrumentation variant */
 /* 12 + 6 + 6 + 1 + 1 + 1 + 1 = 28 bytes */
typedef struct {
    cvcc_ObjID    messageID;          /* identity of a message object */

    cvcc_CallSign originator;         /* identity of the originator  */

    cvcc_CallSign sender;             /* identity of the sender  */

    cvcc_ObjType  otype;              /* type of message (object) */

    cvcc_ReportType rpt;              /* type of rpt (if applicable) */

    ivis_MessageAction actionOnMessage;  /* action performed on message */

    ivis_RetrievalMethod retrievalMethod;  /* how the user selected msg */

} ivis_MessageVariant; 


/*
 * different ivis scroll modes
 */

#define ivis_ScrollFollow       0    
#define ivis_ScrollJump         1    
#define ivis_ScrollMove         2    

typedef uint8 ivis_ScrollMode; 
 /* map usage variant */
 /* 4 + 4 + 1 + 3 = 12 bytes */
typedef struct {
    cvcc_MapScale scale;              /* map scale */

    cvcc_MapFeatures features;        /* map features */

    IVIS_Location loc;                /* center of map */

    ivis_ScrollMode mode;             /* scroll mode */

    unsigned         _unused_25 : 24;    /* padding */
} ivis_MapVariant; 
 /* tuple for ivis quantity variant */
 /* 2 + 2 = 4 bytes */
typedef struct {
    ivis_QuantityType qtype;          /* thing to be quantized */

    uint16 quantity;             /* quantity of that thing */

} ivis_QuantityItem; 
 /* ivis quantity variant */
 /* 8*4 = 32 bytes */
typedef ivis_QuantityItem ivis_QuantityVariant[ivis_NumberQuantityTypes]; 
 /*  ************************************************************************** */

 /*  IVIS instrumentation PDU: */
 /*  1 + 1 + 6 + max(24,28,12,32) = 40 bytes */
 /*  ************************************************************************** */
typedef struct {
    ivis_InstrumentationKind kind;    /* the type of instrumentation */

    uint8 _unused_26;            /* padding */
    VehicleID  veh;                   /* site/host/vehicle */

    union {
        ivis_ReportGenerationVariant rg;
        ivis_MessageVariant          msg;
        ivis_MapVariant              map;
        ivis_QuantityVariant         qty;
    } u;                                
} ivis_InstrumentationVariant; 
 /*  ************************************************************************** */

 /*  BnTOC instrumentation enumeration types: */

 /*  ************************************************************************** */
 /* types of bntoc instrumentation */
#define bntoc_MessageEventInstr 0    
#define bntoc_FolderEventInstr  1    
#define bntoc_MapModuleInstr    2    

typedef uint8 bntoc_InstrumentationKind; 
 /* bntoc filter destinations */
#define bntoc_FiltNone          0x0  
#define bntoc_FiltInFolder      0x1  
#define bntoc_FiltJournal       0x2  
#define bntoc_FiltMapDisplay    0x4  
#define bntoc_FiltS2            0x8  
#define bntoc_FiltS3            0x10 
#define bntoc_FiltSD            0x20 

typedef uint16 bntoc_FilterDestination; 
 /* bntoc message events */
#define bntoc_MsgArrivalKind    0    
#define bntoc_MsgViewingKind    1    
#define bntoc_MsgCopyKind       2    
#define bntoc_MsgLinkKind       3    
#define bntoc_MsgCreateKind     4    
#define bntoc_MsgRelayKind      5    
#define bntoc_MsgAggKind        6    
#define bntoc_MsgFilterKind     7    
#define bntoc_MsgDeleteKind     8    

typedef uint8 bntoc_MessageEventKind; 
 /* folder events */
#define bntoc_FldrCreate        0    
#define bntoc_FldrDelete        1    
#define bntoc_FldrView          2    
#define bntoc_FldrUnview        3    

typedef uint8 bntoc_FolderEventKind; 
 /* map module instrumentation types */
#define bntoc_MapStateKind      0    
#define bntoc_MapOverlayEventKind 1  
#define bntoc_MapOverlayStackKind 2  

typedef uint8 bntoc_MapModuleInstrKind; 
 /* map scroll methods */
#define bntoc_ScrollMethodNotApplicable 0
#define bntoc_MapScrollBars     1    
#define bntoc_MapAutoScroll     2    
#define bntoc_MapDragScroll     3    
#define bntoc_MapHome           4    
#define bntoc_MapGoTo           5    
#define bntoc_MapResize         6    

typedef uint8 bntoc_MapScrollMethod; 
 /* overlay event types */
#define bntoc_OverAdd           0    
#define bntoc_OverRemove        1    
#define bntoc_OverMoveToTop     2    
#define bntoc_OverMoveToBottom  3    
#define bntoc_OverCircleUp      4    
#define bntoc_OverCircleDown    5    
#define bntoc_OverCreate        6    
#define bntoc_OverEdit          7    
#define bntoc_OverSend          8    
#define bntoc_OverDelete        9    
#define bntoc_OverSave          10   
#define bntoc_OverSaveAs        11   
#define bntoc_OverRename        12   
#define bntoc_OverCopy          13   

typedef uint8 bntoc_OverlayEventKind; 
 /*  ************************************************************************** */

 /*  BnTOC instrumentation structure types: */

 /*  ************************************************************************** */
 /* a folder ID - this is a role and the folder's name */
 /* 32 + 2 + 2 = 36 bytes */
#define bntoc_FolderNameLength  32
typedef struct {
    char          name[bntoc_FolderNameLength]; 
    cvcc_Role  role;                  /* the role of the workstation */

    uint16 _unused_27;           /* padding */
} bntoc_FolderID; 
 /* an overlay ID - this is a role and the overlay's name */
 /* 2 + 10 = 12 bytes */
#define bntoc_OverlayNameLength 10
typedef struct {
    cvcc_Role  role;                  /* the role of the workstation */

    char          name[bntoc_OverlayNameLength]; 
} bntoc_OverlayID; 
 /* Following are the variants that go into a message event.  The message */
 /* event itself will have the message (object) id which contains the */
 /* originator (site/host/vehicle), time stamp and serial number.  The */
 /* variants hold other applicable data. */
 /* message arrival  */
 /* 6 + 2 = 8 bytes */
typedef struct {
    cvcc_CallSign sender;             /* call sign of the sender */

    uint16   _unused_28;         /* padding */
} bntoc_MsgArrival; 
 /* message viewing */
 /* 36 bytes */
typedef struct {
    bntoc_FolderID source;           
     /* the folder from which the message was viewed */

} bntoc_MsgViewing; 
 /* message copying */
 /* 36 + 36 = 72 bytes */
typedef struct {
    bntoc_FolderID source;            /* the folder the message came from */

    bntoc_FolderID dest;              /* the folder the message is copied to */

} bntoc_MsgCopy; 
 /* message linking */
 /* 12 bytes */
typedef struct {
    bntoc_OverlayID dest;             /* the overlay the message is linked to */

} bntoc_MsgLink; 
 /* message creation */
 /* 4 bytes */
typedef struct {
    uint32 _unused_29;            /* padding */
} bntoc_MsgCreate; 
 /* message relay */
 /* 36 + 2 + 2 = 40 bytes */
typedef struct {
    bntoc_FolderID source;            /* the folder them message came from */

    bntoc_MessageDestination recipients;  /* message recipients */

    uint16    _unused_30;        /* padding */
} bntoc_MsgRelay; 
 /* message aggregation */
 /* 2 + 2 + 50*12 = 604 bytes */
#define bntoc_MaxMessagesInAggregate           50
typedef struct {
    uint16 n;                    /* number of messages in aggregate */

    uint16 _unused_31;           /* padding */
    cvcc_ObjID  oid[bntoc_MaxMessagesInAggregate]; 
} bntoc_MsgAgg; 
 /* the ids of the messages (objects) in aggregate */
 /* message filter */
 /* 9*2 + 2 = 20 bytes */
typedef struct {
    bntoc_FilterDestination msgFilter[cvcc_NumberReportTypes]; 
    uint16 _unused_32;           /* padding */
} bntoc_MsgFilter; 
 /* message delete */
 /* 36 bytes */
typedef struct {
    bntoc_FolderID source;            /* the folder the msg is deleted from */

} bntoc_MsgDelete; 
 /* message event variant */
 /* 1 + 1 + 6 + 12 + max(8,36,72,36,4,40,604,20,36) = 624 bytes */
typedef struct {
    bntoc_MessageEventKind kind;      /* type of message event */

    cvcc_ObjType  otype;              /* type of message (object) */

    cvcc_CallSign orig;               /* call sign of the originator */

    cvcc_ObjID    oid;                /* the message (object) id */

    union {
        bntoc_MsgArrival  arr;
        bntoc_MsgViewing  view;
        bntoc_MsgCopy     copy;
        bntoc_MsgLink     link;
        bntoc_MsgCreate   create;
        bntoc_MsgRelay    relay;
        bntoc_MsgAgg      agg;
        bntoc_MsgFilter   filter;
        bntoc_MsgDelete   delete;
    } u;                                
} bntoc_MessageEvent; 
 /* folder event variant */
 /* 1 + 3 + 36 = 40 bytes */
typedef struct {
    bntoc_FolderEventKind kind;       /* type of folder event */

    unsigned          _unused_33 : 24;   /* padding */
    bntoc_FolderID fid;               /* folder which event happened */

} bntoc_FolderEvent; 
 /* map state */
 /* 8 + 4 + 4 + 1 + 3 = 20 bytes */
typedef struct {
    IVIS_Location location;           /* location of the center of the map */

    cvcc_MapScale scale;              /* the scale of the map */

    cvcc_MapFeatures features;        /* which features are turned on */

    bntoc_MapScrollMethod scroll;     /* how map was scrolled */

    unsigned         _unused_34 : 24;    /* padding */
} bntoc_MapState; 
 /* overlay creation */
 /* 1/8 + 3-7/8 = 4 bytes */
typedef struct {
    unsigned  coo  : 1;                  /* Boolean */ /* is it a coo overlay? */

    unsigned  _unused_35 : 31;           /* padding */
} bntoc_OverCreateStruct; 
 /* overlay copy */
 /* 12 + 12 = 24 bytes */
typedef struct {
    bntoc_OverlayID source;           /* the "original" */

    bntoc_OverlayID dest;             /* the "copy" */

} bntoc_OverCopyStruct; 
 /* overlay events */
 /* 1 + 3 + 12 + max(4,24) = 40 bytes */
typedef struct {
    bntoc_OverlayEventKind kind;      /* type of overlay event */

    unsigned           _unused_36 : 24;  /* padding */
    bntoc_OverlayID oid;              /* overlay the action is taken on */

    union {
        bntoc_OverCreateStruct create;
        bntoc_OverCopyStruct   copy;
    } u;                                
} bntoc_MapOverlayEvent; 
 /* stack of overlays */
 /* 2 + 2 + 50*12 = 604 */
#define bntoc_MaxOverlaysReported           50
typedef struct {
    uint16     n;                /* number of overlays reported */

    uint16     _unused_37;       /* padding */
    bntoc_OverlayID oid[bntoc_MaxOverlaysReported]; 
} bntoc_MapOverlayStack; 
 /* the overlays in the stack */
 /* map module instrumentation variant */
 /* 1 + 3 + max(20,40,604) = 608 bytes */
typedef struct {
    bntoc_MapModuleInstrKind kind;    /* type of map instrumentation */

    unsigned      _unused_38 : 24;       /* padding */
    union {
        bntoc_MapState        mapState;
        bntoc_MapOverlayEvent overEvent;
        bntoc_MapOverlayStack overStack;
    } u;                                
} bntoc_MapModule; 
 /*  ************************************************************************** */

 /*  BnTOC instrumentation PDU: */
 /*  1 + 1 + 2 + max(40,624,608) = 628 bytes */
 /*  ************************************************************************** */
typedef struct {
    bntoc_InstrumentationKind kind;   /* type of instrumentation */

    uint8 _unused_39;            /* padding */
    cvcc_Role  role;                  /* workstation role */

    union {
        bntoc_MessageEvent msg;
        bntoc_FolderEvent  fldr;
        bntoc_MapModule    map;
    } u;                                
} bntoc_InstrumentationVariant; 



#endif
