
/*
 * FILE:  cvcc_types.h 
 *
 * Created on Wed Apr 17 10:52:39 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _cvcc_types_included
#define _cvcc_types_included



/*
 *  Name:  cvcc_types.drn
 *  Author:  Mitchell Greess, Chris Ross, Beverly Schwartz and Deborah Wilbert
 *  Date Created:  December 10, 1991
 *  This file contains definitions of types useful for the cvcc world.
 *
 *  Copyright (c) 1991.  All rights reserves.
 *  BBN Systems and Technologies Corp., Cambridge, MA 02138
 *
 * $RCSfile: cvcc_types.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <basic.h> 

/*
 *  Each cvcc application is associated with a role within the brigade
 *  organization.  The following is an enumeration of the different
 *  roles in the cvcc world.
 *
 */

#define cvcc_RoleUndefined      0     /* undefined role */
 /* brigade level roles */
#define cvcc_BdeCO              1     /* brigade commander */
#define cvcc_BdeXO              2     /* brigade executive officer */
#define cvcc_BdeFSO             3     /* brigade fire support officer */
#define cvcc_BdeS4              4     /* brigade S4 */
#define cvcc_BdeS3              5     /* brigade S3 */
#define cvcc_BdeS2              6     /* brigade S2 */
#define cvcc_BdeS1              7     /* brigade S1 */
 /* battalion level roles */
#define cvcc_BnCO               8     /* battalion commander */
#define cvcc_BnXO               9     /* battalion executive officer */
#define cvcc_BnFSO              10    /* battalion fire support officer */
#define cvcc_BnS4               11    /* battalion S4 */
#define cvcc_BnS3               12    /* battalion S3 (operations) */
#define cvcc_BnS2               13    /* battalion S2 (intelligence) */
#define cvcc_BnS1               14    /* battalion S1 */
#define cvcc_BnSitDisp          15    /* battalion toc situation display */
 /* company level roles */
#define cvcc_CoCO               16    /* company commander */
#define cvcc_CoXO               17    /* company executive officer */
 /* platoon level roles */
#define cvcc_PltLeader          18    /* platoon leader */
#define cvcc_PltSergeant        19    /* platoon sergeant */
#define cvcc_PltWingman1        20    /* one wingman */
#define cvcc_PltWingman2        21    /* the other wingman */
typedef uint16   cvcc_Role; 


/*
 *  Each cvcc application is associated with a place in the task organization.
 *  In the cvcc world, this is defined as battalion number, company letter,
 *  platoon number and role.  A cvcc application uses this for two things:
 *  1) the application's identity, and 2) definition of who the user reports
 *  to.  The identity is assigned at initialization and does not change
 *  throughout the simulation.  Who a user reports to is dynamic throughout
 *  a simulation; this changes as casualities occur and battle needs change.
 *  						-- (5 * 2) + 2 = 12 bytes
 */

typedef UnitIdentifier cvcc_UnitIdentifier[5]; 

typedef struct {
    cvcc_UnitIdentifier to;           /* position in the organization */

    cvcc_Role           role;         /* user's role */

} cvcc_Tasking; 
 /*  */
 /*  Cvcc users use alphanumeric call signs to identify themselves. */
 /*  PLEASE NOTE!!  The maximum length of the call sign is for both */
 /*  the call sign and the null terminator.  A call sign is really */
 /*  a maximum of FIVE alphanumerics!! */

#define cvcc_MaxCallSignLength  6 /* max. characters in call sign */
typedef char cvcc_CallSign[cvcc_MaxCallSignLength]; 
 /* last character must be a null terminator */

 /*  An object in the cvcc world, is something which gets sent around and */
 /*  saved.  Because it is saved, each object must be uniquely identified. */
 /*  An object is uniquely identified by a vehicle ID (site/host/vehicle), */
 /*  a serial number, and a time stamp.  Both time stamp and serial number */
 /*  are needed because:  the cvcc send utility always starts with the */
 /*  serial number 1.  If send is executed more than once during a simulation, */
 /*  there will be different messages with the same serial number, thus, the */
 /*  time stamp is needed.  The serial number cannot be removed because */
 /*  applications can send out more than one message per second. */
 /* 						-- 6 + 2 + 4 = 12 bytes */
typedef struct {
    VehicleID  veh;                   /* site/host/vehicle of originator */

    uint16 _unused_3;            /* padding */
    uint32 sn;                    /* a unique serial number */

    SP_Time       time;                  /* a time stamp */

} cvcc_ObjID; 


/*
 *  Map scale is used in a number of contexts in the cvcc world.  It must
 *  be a discrete unsigned integer, and it must be large enough to 
 *  accomodate the number 250,000.
 *							-- 4 bytes
 */

typedef uint32 cvcc_MapScale; 


/*
 *  Map features is another concept used throughout the cvcc world.  Each
 *  one must represent a single bit field.
 *							-- 4 bytes
 */

#define cvcc_Grids              0x1  
#define cvcc_Rivers             0x2  
#define cvcc_Roads              0x4  
#define cvcc_Trees              0x8  
#define cvcc_Buildings          0x10 
#define cvcc_Pipelines          0x20 
#define cvcc_Politicals         0x40 
#define cvcc_Railroads          0x80 
#define cvcc_Powerlines         0x100
#define cvcc_Contours           0x200

typedef uint32 cvcc_MapFeatures; 


/*
 * Bntoc can address for networks, or direct addressing within the bntoc.
 * These are bit fields for indicating a destination for a bntoc message.
 */

#define bntoc_RcptNone          0x0  
#define bntoc_RcptBnCommandNet  0x1  
#define bntoc_RcptBnS1          0x2  
#define bntoc_RcptBnS2          0x4  
#define bntoc_RcptBnS3          0x8  
#define bntoc_RcptBnS4          0x10 
#define bntoc_RcptBnFSO         0x20 
#define bntoc_RcptBnXO          0x40 
#define bntoc_RcptBnCO          0x80 
#define bntoc_RcptBnSitDisp     0x100
#define bntoc_RcptAllBnWs       0x1fe
#define bntoc_RcptBdeNet        0x200

typedef uint16 bntoc_MessageDestination; 



#endif
