
/*
 * FILE:  dis203_types.h 
 *
 * Created on Wed Apr 17 10:53:30 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _dis203_types_included
#define _dis203_types_included



/*
 * File:     dis203_types.drn
 * Contents: 
 * System:   BDS-D BIP development.
 * Created:  6-Oct-1993
 * Author:   cchiang/ckuczaj
 *
 * Remarks:
 *
 * History:
 * 6-Oct-93	new file
 *		Contains common types for the DIS protocol as defined in
 *		the "Military Standard (2.0.3) Protocol Data Units 
 *		for Entity Information and Entity Interaction in a 
 *		Distributed Interactive Simulation" document.
 *
 * 13-Jan-1994  The data types that are required to support the Logistics,
 *              Simulation Management, Laser, and Receiver PDUs were
 *              added.
 *    
 *
 * $RCSfile: dis203_types.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 * COPYRIGHT 1993 Loral Advanced Distributed Simulation
 * 50 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3600
 *
 */

#include <stdtypes.h>
#include <dis203_values.h> 

/*
 * Angle Representation
 *
 * Version 2.0.3 Angle Representation (section 5.3.1, page 76)
 * Angles shall be specified as 32-bit float32s expressed in radians
 */

typedef float32            DIS203_ANGLE_TYPE; 


/*
 * The Angular Velocity Representation
 *
 */

typedef float32            DIS203_ANGULAR_VELOCITY_TYPE; 


/*
 * Change Indicator
 *
 */

typedef uint16   DIS203_CHANGE_INDICATOR_TYPE; 


/*
 * ID - Part Attached to
 *
 */

typedef uint16   DIS203_ID_PART_ATTACHED_TO_TYPE; 


/*
 * Parameter Type
 *
 */

typedef uint32    DIS203_PARAMETER_TYPE_TYPE; 


/*
 * Parameter Value
 *
 */

typedef float32            DIS203_PARAMETER_VALUE_TYPE; 
 /* Boolean */

 /* Shall be represented as a single bit representing a true-false */
 /* value.  This bit shall represent an enumeration type of one bit, */
 /* where the value 0 is interpreted as false, and the value 1 as true. */

typedef struct {
    unsigned       value  : 1;          
    unsigned       _unused_1 : 7;        /* padding */
} DIS203_BOOLEAN_RECORD; 
 /*  */
 /* Quantity */

typedef uint16 DIS203_QUANTITY_TYPE; 


/*
 * Rate
 *
 * In units of rounds per second
 *
 */

typedef uint16 DIS203_RATE_TYPE; 


/*
 * Entity Appearance
 *
 */

typedef uint32 DIS203_ENTITY_APPEARANCE_TYPE; 


/*
 * Entity Capabilities
 *
 */

typedef struct {
    unsigned       _unused_2                     : 28;  /* padding */
    unsigned       repair_capability             : 1; 
    unsigned       recovery_capability           : 1; 
    unsigned       fuel_supply_capability        : 1; 
    unsigned       ammunition_supply_capability  : 1; 
} DIS203_ENTITY_CAPABILITIES_RECORD; 


/*
 * Site ID
 *
 */

typedef uint16 DIS203_SITE_ID_TYPE; 
 /*  */
 /* Application ID */

typedef uint16 DIS203_APPLICATION_ID_TYPE; 


/*
 * Entity ID
 *
 */

typedef uint16 DIS203_ENTITY_ID_TYPE; 


/*
 * Group ID
 *
 */

typedef uint16 DIS203_GROUP_ID_TYPE; 
 /*  */
 /* Character Set */

typedef uint8 DIS203_CHARACTER_SET_TYPE; 
 /*  */
 /* Character */

typedef uint8 DIS203_CHARACTER_TYPE; 


/*
 * Entity Kind
 *
 */

typedef uint8 DIS203_KIND_TYPE; 


/*
 * Domain
 *
 */

typedef uint8 DIS203_DOMAIN_TYPE; 


/*
 * Country
 *
 */

typedef uint16 DIS203_COUNTRY_TYPE; 


/*
 * Category
 *
 */

typedef uint8 DIS203_CATEGORY_TYPE; 


/*
 * Subcategory
 *
 */

typedef uint8 DIS203_SUBCATEGORY_TYPE; 


/*
 * Specific
 *
 */

typedef uint8 DIS203_SPECIFIC_TYPE; 


/*
 * Extra
 *
 */

typedef uint8 DIS203_EXTRA_TYPE; 


/*
 * Event ID
 *
 */

typedef uint16 DIS203_EVENT_ID_TYPE; 


/*
 * Exercise ID
 *
 */

typedef uint8 DIS203_EXERCISE_ID_TYPE; 


/*
 * Force ID
 *
 */

typedef uint8 DIS203_FORCE_ID_TYPE; 


/*
 * Protocol Version
 *
 */

typedef uint8 DIS203_PROTOCOL_VERSION_TYPE; 


/*
 * PDU Type
 *
 */

typedef uint8 DIS203_PDU_TYPE_TYPE; 


/*
 * Time Stamp
 *
 */

typedef struct {
    unsigned       value      : 31;     
    unsigned       reference  : 1;      
} DIS203_TIMESTAMP_RECORD; 


/*
 * Time
 *
 */

typedef uint32 DIS203_TIME_TYPE; 


/*
 * Length
 *
 */

typedef uint16 DIS203_LENGTH_TYPE; 


/*
 * Repair
 *
 */

typedef uint16 DIS203_REPAIR_TYPE_TYPE; 


/*
 * Service
 *
 */

typedef uint8 DIS203_SERVICE_TYPE_TYPE; 


/*
 * Supply Quantity
 *
 */

typedef float32 DIS203_SUPPLY_QUANTITY_TYPE; 


/*
 * Vector Component
 *
 */

typedef float32 DIS203_VECTOR_COMPONENT_TYPE; 
 /*  */
 /* World Coordinates Component */

typedef float64 DIS203_WORLD_COORDINATES_COMPONENT_TYPE; 


/*
 * Dead Reckoning Algorithm
 *
 */

typedef uint8 DIS203_DEAD_RECKONING_ALGORITHM_TYPE; 


/*
 * Number of Articulation Parameters
 *
 */

typedef uint8 DIS203_NUMBER_OF_ARTICULATION_PARAMETERS_TYPE; 


/*
 * Range
 *
 */

typedef float32 DIS203_RANGE_TYPE; 


/*
 * Detonation Result
 *
 */

typedef uint8 DIS203_DETONATION_RESULT_TYPE; 


/*
 * Number of Supply Types
 *
 */

typedef uint8 DIS203_NUMBER_OF_SUPPLY_TYPES_TYPE; 


/*
 * Repair Result
 *
 */

typedef uint8 DIS203_REPAIR_RESULT_TYPE; 


/*
 * Mass
 *
 */

typedef float32 DIS203_MASS_TYPE; 


/*
 * Reasons to Stop (4.3.1.10)
 *
 */

typedef uint8 DIS203_REASONS_TO_STOP_TYPE; 


/*
 * Acknowledgement Flags (4.3.1.11)
 *
 */

typedef uint16 DIS203_ACKNOWLEDGEMENT_FLAGS_TYPE; 


/*
 * Response Flags (None defined)
 *
 */

typedef uint16 DIS203_RESPONSE_FLAGS_TYPE; 
 /*  */
 /* Request ID  */

typedef uint32 DIS203_REQUEST_ID_TYPE; 


/*
 * Action ID
 *
 */

typedef uint32 DIS203_ACTION_ID_TYPE; 


/*
 * Request Status
 *
 */

typedef uint32 DIS203_REQUEST_STATUS_TYPE; 


/*
 * Time Interval
 *
 */

typedef uint32 DIS203_TIME_INTERVAL_TYPE; 
 /*   */
 /* Number of Fixed Datum Fields   */

typedef uint32 DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE; 
 /*   */
 /* Number of Variable Datum Fields   */

typedef uint32 DIS203_NUMBER_OF_VARIABLE_DATUM_FIELDS_TYPE; 


/*
 * Datum ID
 *
 */

typedef uint32 DIS203_DATUM_ID_TYPE; 


/*
 * Event Type
 *
 */

typedef uint32 DIS203_EVENT_TYPE_TYPE; 


/*
 * Radio ID
 *
 */

typedef uint16 DIS203_RADIO_ID_TYPE; 


/*
 * Sample rate (for Signal PDU)
 *
 */

typedef uint32 DIS203_SAMPLE_RATE_TYPE; 


/*
 * Signal length (for Signal PDU)
 *
 */

typedef uint16 DIS203_SIGNAL_LENGTH_TYPE; 


/*
 * Number of Samples (for Signal PDU)
 *
 */

typedef uint16 DIS203_SAMPLES_TYPE; 


/*
 * Radio Transmit State
 *
 */

typedef uint8 DIS203_RADIO_TRANSMIT_STATE_TYPE; 


/*
 * Radio Input Source
 *
 */

typedef uint8 DIS203_RADIO_INPUT_SOURCE_TYPE; 


/*
 * Antenna Pattern Type
 *
 */

typedef uint16 DIS203_ANTENNA_PATTERN_TYPE_TYPE; 


/*
 * Antenna Pattern Length
 *
 */

typedef uint16 DIS203_ANTENNA_PATTERN_LENGTH_TYPE; 


/*
 * Frequency
 *
 */

typedef struct {
    uint32 high_order;           
    uint32 low_order;            
} DIS203_TRANSMITTER_FREQUENCY_TYPE; 


/*
 * Frequency Bandwidth
 *
 */

typedef float32 DIS203_FREQUENCY_BANDWIDTH_TYPE; 


/*
 * Power
 *
 */

typedef float32 DIS203_POWER_TYPE; 


/*
 * Crypto System
 *
 */

typedef uint16 DIS203_CRYPTO_SYSTEM_TYPE; 


/*
 * Crypto Key ID
 *
 */

typedef uint16 DIS203_CRYPTO_KEY_ID_TYPE; 


/*
 * Length of Modulation Parameters (for Transmitter PDU)
 *
 */

typedef uint8 DIS203_MODULATION_LENGTH_TYPE; 


/*
 * Receiver State (for Receiver PDU)
 *
 */

typedef uint16 DIS203_RECEIVER_STATE_TYPE; 


/*
 * Transmitter Radio ID (for Receiver PDU)
 *
 */

typedef uint16 TRANSMITTER_RADIO_ID_TYPE; 


/*
 * State Update Indicator (for Emission PDU)
 *
 */

typedef uint8 DIS203_STATE_UPDATE_INDICATOR_TYPE; 


/*
 * Number of Systems (for Emission PDU)
 *
 */

typedef uint8 DIS203_NUMBER_OF_SYSTEMS_TYPE; 


/*
 * System Data Length (for Emission PDU)
 *
 */

typedef uint8 DIS203_SYSTEM_DATA_LENGTH_TYPE; 


/*
 * Number of Beams (for Emission PDU)
 *
 */

typedef uint8 DIS203_NUMBER_OF_BEAMS_TYPE; 


/*
 * Beam Data Length (for Emission PDU)
 *
 */

typedef uint8 DIS203_BEAM_DATA_LENGTH_TYPE; 


/*
 * Beam ID Number (for Emission PDU)
 *
 */

typedef uint8 DIS203_BEAM_ID_NUMBER_TYPE; 


/*
 * Beam Parameter Index (for Emission PDU)
 *
 */

typedef uint16 DIS203_BEAM_PARAMETER_INDEX_TYPE; 


/*
 * Beam Function (for Emission PDU)
 *
 */

typedef uint8 DIS203_BEAM_FUNCTION_TYPE; 


/*
 * Number of Targets (for Emission PDU)
 *
 */

typedef uint8 DIS203_NUMBER_OF_TARGETS_TYPE; 


/*
 * High Density Track Jam (for Emission PDU)
 *
 */

typedef uint8 DIS203_HIGH_DENSITY_TRACK_JAM_TYPE; 


/*
 * Jamming Mode Sequence (for Emission PDU)
 *
 */

typedef uint32 DIS203_JAMMING_MODE_SEQUENCE_TYPE; 


/*
 * Emitter Name (for Emission PDU)
 *
 */

typedef uint16 DIS203_EMITTER_NAME_TYPE; 


/*
 * Emitter Function (for Emission PDU)
 *
 */

typedef uint8 DIS203_EMITTER_FUNCTION_TYPE; 


/*
 * Emitter ID Number (for Emission PDU)
 *
 */

typedef uint8 DIS203_EMITTER_ID_NUMBER_TYPE; 


/*
 * Emission Frequency (for Emission PDU)
 *
 */

typedef float32 DIS203_EMISSION_FREQUENCY_TYPE; 


/*
 * Frequency Range (for Emission PDU)
 *
 */

typedef float32 DIS203_FREQUENCY_RANGE_TYPE; 


/*
 * Effective Radiated Power (for Emission PDU)
 *
 */

typedef float32 DIS203_EFFECTIVE_RADIATED_POWER_TYPE; 


/*
 * Pulse Width (for Emission PDU)
 *
 */

typedef float32 DIS203_PULSE_WIDTH_TYPE; 


/*
 * Beam Sweep Sync (for Emission PDU)
 *
 */

typedef float32 DIS203_BEAM_SWEEP_SYNC_TYPE; 


/*
 * Code Name (for Laser PDU)
 *
 */

typedef uint16 DIS203_CODE_NAME_TYPE; 


/*
 * Laser Code (for Laser PDU)
 *
 */

typedef uint8 DIS203_LASER_CODE_TYPE; 


/*
 * Laser Wave Length
 *
 */

typedef float32 DIS203_LASER_WAVE_LENGTH_TYPE; 
 /* END LOWEST-LEVEL COMPONENTS */
 /* 1st-LEVEL COMPONENTS */

 /* Angular Velocity Vector Record (section 5.3.2, Page 76) */

 /* These rates shall be measured in radians per second about each of the */
 /* entity's own coordinate axes.  The first field shall consist of the */
 /* velocity about the x-axis, the second about the y-axis and the third */
 /* about the z-axis. */

typedef struct {
    DIS203_ANGULAR_VELOCITY_TYPE rate_about_x_axis; 
    DIS203_ANGULAR_VELOCITY_TYPE rate_about_y_axis; 
    DIS203_ANGULAR_VELOCITY_TYPE rate_about_z_axis; 
} DIS203_ANGULAR_VELOCITY_VECTOR_RECORD; 


/*
 * Simulation Address Record
 *
 */

typedef struct {
    DIS203_SITE_ID_TYPE        site_id; 
    DIS203_APPLICATION_ID_TYPE application_id; 
} DIS203_SIMULATION_ADDRESS_RECORD; 


/*
 * Entity Marking Record
 *
 */

typedef struct {
    DIS203_CHARACTER_SET_TYPE Character_set; 
    DIS203_CHARACTER_TYPE     Character_string[11]; 
} DIS203_ENTITY_MARKING_RECORD; 


/*
 * Entity Type Record
 *
 */

typedef struct {
    DIS203_KIND_TYPE        kind;    
    DIS203_DOMAIN_TYPE      domain;  
    DIS203_COUNTRY_TYPE     country; 
    DIS203_CATEGORY_TYPE    category; 
    DIS203_SUBCATEGORY_TYPE subcategory; 
    DIS203_SPECIFIC_TYPE    specific; 
    DIS203_EXTRA_TYPE       extra;   
} DIS203_ENTITY_TYPE_RECORD; 


/*
 * Euler Angles Record
 *
 */

typedef struct {
    DIS203_ANGLE_TYPE psi;           
    DIS203_ANGLE_TYPE theta;         
    DIS203_ANGLE_TYPE phi;           
} DIS203_EULER_ANGLES_RECORD; 
 /*  */
 /* Protocol Data Unit (PDU) Header Record */

typedef struct {
    DIS203_PROTOCOL_VERSION_TYPE protocol_version; 
    DIS203_EXERCISE_ID_TYPE      exercise_id; 
    DIS203_PDU_TYPE_TYPE         pdu_type; 
    uint8                   _unused_3;  /* padding */
    DIS203_TIMESTAMP_RECORD      timestamp; 
    DIS203_LENGTH_TYPE           length; 
    uint16                  _unused_4;  /* padding */
} DIS203_PDU_HEADER_RECORD; 


/*
 * Vector Record
 *
 */

typedef struct {
    DIS203_VECTOR_COMPONENT_TYPE x;  
    DIS203_VECTOR_COMPONENT_TYPE y;  
    DIS203_VECTOR_COMPONENT_TYPE z;  
} DIS203_VECTOR_RECORD; 


/*
 * Entity Coordinate Vector Record
 *
 */

typedef DIS203_VECTOR_RECORD DIS203_ENTITY_COORDINATE_VECTOR_RECORD; 


/*
 * Linear Acceleration Vector Record
 *
 */

typedef DIS203_VECTOR_RECORD DIS203_LINEAR_ACCELERATION_VECTOR_RECORD; 


/*
 * Linear Velocity Vector Record
 *
 */

typedef DIS203_VECTOR_RECORD DIS203_LINEAR_VELOCITY_VECTOR_RECORD; 


/*
 * World Coordinates Record
 *
 */

typedef struct {
    DIS203_WORLD_COORDINATES_COMPONENT_TYPE x; 
    DIS203_WORLD_COORDINATES_COMPONENT_TYPE y; 
    DIS203_WORLD_COORDINATES_COMPONENT_TYPE z; 
} DIS203_WORLD_COORDINATES_RECORD; 


/*
 * Clock Time (5.4.5.1.5)
 *
 */

typedef struct {
    uint32              hour;    
    DIS203_TIMESTAMP_RECORD time_past_the_hour; 
} DIS203_CLOCK_TIME_RECORD; 


/*
 * Fixed Datum Record
 *
 */

typedef struct {
    DIS203_DATUM_ID_TYPE datum_id;   
    uint32           datum_value; 
} DIS203_FIXED_DATUM_RECORD; 


/*
 * Datum Value Record
 *
 */

typedef struct {
    uint32 datum_value_1;        
    uint32 datum_value_2;        
} DIS203_DATUM_VALUE_RECORD; 


/*
 * Variable Datum Record
 *
 */

typedef struct {
    DIS203_DATUM_ID_TYPE      datum_id; 
    uint32                datum_length; 
    DIS203_DATUM_VALUE_RECORD datum_value[1]; 
} DIS203_VARIABLE_DATUM_RECORD; 


/*
 * Modulation Type (5.4.7.2 [17])
 *
 */

typedef struct {
    uint16 spread_spectrum;     
    uint16 mod_major;           
    uint16 mod_detail;          
    uint16 mod_system;          
} DIS203_MODULATION_TYPE_RECORD; 


/*
 * Encoding Scheme (5.4.7.3 [4])
 *
 */

typedef uint16 DIS203_ENCODING_SCHEME; 


/*
 * Emitter System (for Emission PDU)
 *
 */

typedef struct {
    DIS203_EMITTER_NAME_TYPE      emitter_name; 
    DIS203_EMITTER_FUNCTION_TYPE  function; 
    DIS203_EMITTER_ID_NUMBER_TYPE emitter_id_number; 
} DIS203_EMITTER_SYSTEM_RECORD; 


/*
 * Fundamental Parameter Data Record (for Emission PDU)
 *
 */

typedef struct {
    DIS203_EMISSION_FREQUENCY_TYPE       frequency; 
    DIS203_FREQUENCY_RANGE_TYPE          frequency_range; 
    DIS203_EFFECTIVE_RADIATED_POWER_TYPE effective_radiated_power; 
    DIS203_EMISSION_FREQUENCY_TYPE       pulse_repetition_frequency; 
    DIS203_PULSE_WIDTH_TYPE              pulse_width; 
    DIS203_ANGLE_TYPE                    beam_azimuth_center; 
    DIS203_ANGLE_TYPE                    beam_azimuth_sweep; 
    DIS203_ANGLE_TYPE                    beam_elevation_center; 
    DIS203_ANGLE_TYPE                    beam_elevation_sweep; 
    DIS203_BEAM_SWEEP_SYNC_TYPE          beam_sweep_sync; 
} DIS203_FUNDAMENTAL_PARAMETER_DATA_RECORD; 
 /* END 1st-LEVEL COMPONENTS */
 /* 2nd-LEVEL COMPONENTS */

 /* Articulation Parameter Record */

 /* The change indicator shall be set to zero for each exercise and */
 /* sequentially incremented by one for each change in articulation */
 /* parameters.  In the case where all possible values are exhausted, */
 /* the numbers shall be reused beginning at zero. */

 /* The id - part attached to field shall contain the value zero if */
 /* the articulated part is attached directly to the entity. */

typedef struct {
    DIS203_CHANGE_INDICATOR_TYPE    change_indicator; 
    DIS203_ID_PART_ATTACHED_TO_TYPE id_part_attached_to; 

    union {
         /* Parameter Type Record */
         /* if whole, want to access all 32 bits at once for comparisons */
        DIS203_PARAMETER_TYPE_TYPE          value;
         /* if articulated part */
        struct {
            unsigned       part_type         : 1; 
            unsigned       part_type_class   : 26; 
            unsigned       part_type_metric  : 5; 
        } articulation_parameter;
         /* if attached part */
        struct {
            unsigned       part_type          : 1; 
            unsigned       part_type_station  : 31; 
        } attached_part;
    } parameter_type;                   

    union {
         /* Parameter Value Record */
         /* if attached part */
        DIS203_ENTITY_TYPE_RECORD           entity_type;
         /* if articulated part */
        struct {
            DIS203_PARAMETER_VALUE_TYPE value; 
            uint32                  _unused_5;  /* padding */
        } numeric;
    } parameter_value;                  
} DIS203_ARTICULATION_PARAMETER_RECORD; 


/*
 * Supply Quantity Record
 *
 */

typedef struct {
    DIS203_ENTITY_TYPE_RECORD   supply_type; 
    DIS203_SUPPLY_QUANTITY_TYPE supply_quantity; 
} DIS203_SUPPLY_QUANTITY_RECORD; 


/*
 * Burst Descriptor Record
 *
 */

typedef struct {
    DIS203_ENTITY_TYPE_RECORD munition; 
    DIS203_WARHEAD_TYPE       warhead; 
    DIS203_FUZE_TYPE          fuze;  
    DIS203_QUANTITY_TYPE      quantity; 
    DIS203_RATE_TYPE          rate;  
} DIS203_BURST_DESCRIPTOR_RECORD; 


/*
 * Entity Identifier Record
 *
 */

typedef struct {
    DIS203_SIMULATION_ADDRESS_RECORD simulation_address; 
    DIS203_ENTITY_ID_TYPE            entity_id; 
} DIS203_ENTITY_ID_RECORD; 


/*
 * Event Identifier Record
 *
 */

typedef struct {
    DIS203_SIMULATION_ADDRESS_RECORD simulation_address; 
    DIS203_EVENT_ID_TYPE             event_id; 
} DIS203_EVENT_ID_RECORD; 


/*
 * Dead Reckoning Parameters
 *
 */

typedef struct {
    DIS203_DEAD_RECKONING_ALGORITHM_TYPE     dead_reckoning_algorithm; 
    unsigned                                    _unused_6_7               : 24; 
     /* padding */
    uint32                               _unused_6_8;  /* padding */
    uint32                               _unused_6_9;  /* padding */
    uint32                               _unused_6;  /* padding */
    DIS203_LINEAR_ACCELERATION_VECTOR_RECORD entity_linear_accel; 
    DIS203_ANGULAR_VELOCITY_VECTOR_RECORD    entity_angular_velocity; 
} DIS203_DEAD_RECKONING_PARAMETERS_RECORD; 


/*
 * Track Jam Record (for Emission PDU)
 *
 */

typedef struct {
    DIS203_ENTITY_ID_RECORD       entity_id; 
    DIS203_EMITTER_ID_NUMBER_TYPE emitter_id; 
    DIS203_BEAM_ID_NUMBER_TYPE    beam_id; 
} DIS203_TRACK_JAM_RECORD; 
 /* END 2nd-LEVEL COMPONENTS */


#endif
