
/*
 * FILE:  dis_types.h 
 *
 * Created on Wed Apr 17 10:53:22 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _dis_types_included
#define _dis_types_included



/*
 * File:     dis_types.drn
 * Contents: 
 * System:   simnet development.
 * Created:  23-Sep-1992
 * Author:   bschwart
 *
 * Remarks:
 *
 * History:
 *	24-Sep-1992	new file
 *		Contains common types for the DIS protocol as defined in
 *		the "Military Standard (Final Draft) Protocol Data Units 
 *		for Entity Information and Entity Interaction in a 
 *		Distributed Interactive Simulation" document.
 *    
 *
 * $RCSfile: dis_types.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 * COPYRIGHT 1990 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 *
 */

#include <stdtypes.h>
#include <dis_values.h> 

/*
 * Angle Representation (section 5.2.1, page 31)
 * Angles are specified as 32-bit integers expressed as BAMs
 */

typedef uint32    DIS_ANGLE; 


/*
 * Entity Appearance (section 5.3.4.1, letter k, page 48 and appendix D,
 * pages 129-134)
 * Entity Appearance is expressed with 32 bit fields.  Depending on what
 * entity is described is how the bitfields are described.  See the
 * DIS_ENTITY_APPEARANCE_... types in "dis_values.drn"
 */

typedef uint32    DIS_ENTITY_APPEARANCE; 


/*
 * Simulation Address Record (section 5.2.8.1, page 35)
 * The simulation address contains a site number and a host number.
 */

typedef struct {
    uint16 site;                
    uint16 host;                
} DIS_SIMULATION_ADDRESS; 


/*
 * Entity Identifier Record. (section 5.2.8.2, page 35)
 * The entity id is a simulation address and a entity identifier.
 */

typedef struct {
    DIS_SIMULATION_ADDRESS simulator; 
    uint16            entity;   
} DIS_ENTITY_IDENTIFIER; 


/*
 * The Entity Type Record.  (section 5.2.10, page 36)
 * The Entity Type Record specified the entity type, the country associated
 * with that entity, the domain, the specific identification of the
 * entity, and any extra information necessary for describing the entity.
 * domain can be filled with a DIS_ENTITY_DOMAIN or a DIS_MUNITION_DOMAIN,
 * depending on the entity kind.
 */

typedef struct {
    DIS_ENTITY_KIND entity_kind;     
    uint8  domain;              
    DIS_COUNTRY_ID country;          
    uint8  category;            
    uint8  sub_category;        
    uint8  specific;            
    uint8  extra;               
} DIS_ENTITY_TYPE; 
 /*  */
 /* Exercise Identifier. (section 5.2.13, page 39) */
typedef uint8 DIS_EXERCISE_ID; 


/*
 * Protocol Data Unit (PDU) Header.  (section 5.3.3, page 45)
 */

typedef struct {
    DIS_PROTOCOL_VERSION version;    
    DIS_EXERCISE_ID exercise;        
    DIS_PDU_KIND    kind;            
    uint8      _unused_1;        /* padding */
} DIS_PDU_HEADER; 


/*
 * Hierarchy. (section 5.2.14.4, page 40 and appendix e, section 30.1.4,
 * page 136)
 * Service hierarchy consists of eight 8-bit unsigned integers.  Each
 * byte represents a different level, and contains the value for that
 * level.  The set of levels varies dependent on the branch of service.
 * See DIS_HIERARCHY_... types in dis_values.drn.
 */

typedef uint8 DIS_HIERARCHY[8]; 


/*
 * Time Representation (section 5.2.6, page 34)
 * Time is expressed in number of seconds since 00:00:00 GMT January 1st 1970 
 * (unix time) in 32 bits.
 */

typedef uint32 DIS_TIME; 


/*
 * Vector Record. (section 5.2.20, page 42)
 * Vectors are used for entity coordinates (section 5.2.20.1, page 42),
 * linear acceleration (section 5.2.20.2, page 42), and linear velocity
 * (section 5.2.20.3, page 43).  The first component corresponds to 'x',
 * the second to 'y', and the third to 'z'.  For entity coordinates,
 * the vector is expressed in meters.  For linear acceleration, the
 * vector is expressed in meters per second squared.  For linear velocity,
 * the vector is expressed in meters.  For an entity, the x-axis extends
 * in the positive direction out its front; the y-axis extends in the
 * positive direction out the right side; the z-axis extends in the 
 * positive direction downward.
 */

typedef struct {
    float32 x;                            
    float32 y;                            
    float32 z;                            
} DIS_VECTOR; 


/*
 * World Coordinates Record. (section 5.2.21, page 43)
 * A location is expressed in a 3-D cartesian coordinate system.  The shape
 * of the earth is specified using WGS84.  The origin of this coordinate
 * system shall be the centroid of the earth, with the x-axis passing
 * through the Prime Meridian at the equator, the y-axis passing through
 * 90 degrees east int32itude at the equator, and the z-axis passing through
 * the North Pole.
 */

typedef struct {
    float64 x;                           
    float64 y;                           
    float64 z;                           
} DIS_WORLD_COORDINATES; 



#endif
