
/*
 * FILE:  dis_values.h 
 *
 * Created on Wed Apr 17 10:53:20 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _dis_values_included
#define _dis_values_included



/*
 * File:     dis_values.drn
 * Contents: enumerations and values for dis
 * System:   simnet development.
 * Created:  22-Sep-1992
 * Author:   bschwart
 *
 * Remarks:
 *	These are the enumerations defined in "Military Standard Protocol
 *	Data Units for Entity Information and Entity Interaction in a
 *	Distrkbuted Interactive Simulation", issued May 1992
 *
 * History:
 *    
 * $RCSfile: dis_values.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *    
 * COPYRIGHT 1990 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 *
 *
 * Versions of the DIS_ protocol: (section 5.3.3, letter a, page 45)
 */

#include <stdtypes.h>

#define DIS_PROTOCOL_VERSION_JUL92 1 
#define DIS_PROTOCOL_VERSION_NOV93 3 

typedef uint8    DIS_PROTOCOL_VERSION; 


/*
 * The current version: (section 5.3.3, letter a, page 45)
 */

#define DIS_PROTOCOL_VERSION_CURRENT           3

/*
 * The different DIS PDUs: (Appendix F, section 30.1.6.2-3, page 144)
 */

#define DIS_OTHER_PDU_KIND      0     /* defined values (section 30.1.6.2) */
#define DIS_ENTITY_STATE_PDU_KIND 1  
#define DIS_FIRE_PDU_KIND       2    
#define DIS_DETONATION_PDU_KIND 3    
#define DIS_COLLISION_PDU_KIND  4    
#define DIS_SERVICE_REQUEST_PDU_KIND 5
#define DIS_RESUPPLY_OFFER_PDU_KIND 6
#define DIS_RESUPPLY_RECEIVED_PDU_KIND 7
#define DIS_RESUPPLY_CANCEL_PDU_KIND 8
#define DIS_REPAIR_COMPLETE_PDU_KIND 9
#define DIS_REPAIR_RESPONSE_PDU_KIND 10
#define DIS_TRANSMITTER_PDU_KIND 25  
#define DIS_SIGNAL_PDU_KIND     26   
#define DIS_RECEIVER_PDU_KIND   27    /* temporary values (section 30.1.6.3) */
#define DIS_RADAR_PDU_KIND      128  
#define DIS_ACTIVATE_REQUEST_PDU_KIND 129
#define DIS_ACTIVATE_RESPONSE_PDU_KIND 130
#define DIS_DEACTIVATE_REQUEST_PDU_KIND 131
#define DIS_DEACTIVATE_RESPONSE_PDU_KIND 132
#define DIS_EMITTER_PDU_KIND    133  

typedef uint8    DIS_PDU_KIND; 


/*
 * Articulated Parameters (Appendix G, pages 151-164)
 * An articulated parameter is described in two parts - the type of motion
 * of the appendage, and what the appendage is.  The types of motion are
 * defined by constants between 1 and 19, inclusive.  The appendages are 
 * defined by multiples of 20.  An articulation parameter describes one
 * aspect of an appendage.  This is defined by adding the motion type to
 * the appendage type.  There are certain appendages that don't have any
 * motions or positions associated with them; these are included in the
 * articulated parts (appendages) enumeration, but should not have any
 * articulated part types (motions) added to them.
 * Moral of the story:  DIS_ARTICULATED_PARAMETER =
 *		DIS_ARTICULATED_PARAMETER_TYPE + DIS_ARTICULATED_PART
 */

typedef uint32    DIS_ARTICULATED_PARAMETER; 

#define DIS_ARTICULATED_PARAMETER_TYPE_POSITION 1
#define DIS_ARTICULATED_PARAMETER_TYPE_POSITION_RATE 2
#define DIS_ARTICULATED_PARAMETER_TYPE_EXTENSION 3
#define DIS_ARTICULATED_PARAMETER_TYPE_EXTENSION_RATE 4
#define DIS_ARTICULATED_PARAMETER_TYPE_X 5
#define DIS_ARTICULATED_PARAMETER_TYPE_X_RATE 6
#define DIS_ARTICULATED_PARAMETER_TYPE_Y 7
#define DIS_ARTICULATED_PARAMETER_TYPE_Y_RATE 8
#define DIS_ARTICULATED_PARAMETER_TYPE_Z 9
#define DIS_ARTICULATED_PARAMETER_TYPE_Z_RATE 10
#define DIS_ARTICULATED_PARAMETER_TYPE_AZIMUTH 11
#define DIS_ARTICULATED_PARAMETER_TYPE_AZIMUTH_RATE 12
#define DIS_ARTICULATED_PARAMETER_TYPE_ELEVATION 13
#define DIS_ARTICULATED_PARAMETER_TYPE_ELEVATION_RATE 14
#define DIS_ARTICULATED_PARAMETER_TYPE_ROTATION 15
#define DIS_ARTICULATED_PARAMETER_TYPE_ROTATION_RATE 16

typedef uint32    DIS_ARTICULATED_PARAMETER_TYPE; 

#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_STORE_STATIONS 0
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_STORE_STATIONS 20
#define DIS_ARTICULATED_PART_WEAPON_MOUNTS 40
#define DIS_ARTICULATED_PART_SENSOR_MOUNTS 60
#define DIS_ARTICULATED_PART_BULLDOZER_BLADE 80
#define DIS_ARTICULATED_PART_MINE_CLEARING 81
#define DIS_ARTICULATED_PART_ROLLERS 82
#define DIS_ARTICULATED_PART_SWIMMING_GEAR 83
#define DIS_ARTICULATED_PART_LEFT_WING_STATIONS 100
#define DIS_ARTICULATED_PART_RIGHT_WING_STATIONS 120
#define DIS_ARTICULATED_PART_CARGO_STATIONS 140
#define DIS_ARTICULATED_PART_ARMOR_MOUNTS 160
#define DIS_ARTICULATED_PART_RUDDER 1000
#define DIS_ARTICULATED_PART_LEFT_FLAP 1020
#define DIS_ARTICULATED_PART_RIGHT_FLAP 1040
#define DIS_ARTICULATED_PART_LEFT_AILERON 1060
#define DIS_ARTICULATED_PART_RIGHT_AILERON 1080
#define DIS_ARTICULATED_PART_PERISCOPE 2000
#define DIS_ARTICULATED_PART_GENERIC_ANTENNA 2020
#define DIS_ARTICULATED_PART_SNORKEL 2040
#define DIS_ARTICULATED_PART_LANDING_GEAR 3000
#define DIS_ARTICULATED_PART_TAIL_HOOK 3020
#define DIS_ARTICULATED_PART_SPEED_BRAKE 3040
#define DIS_ARTICULATED_PART_LEFT_WEAPON_BAY_DOOR 3060
#define DIS_ARTICULATED_PART_RIGHT_WEAPON_BAY_DOOR 3080
#define DIS_ARTICULATED_PART_TANK_OR_APC_HATCH 3100
#define DIS_ARTICULATED_PART_WINGSWEEP 3120
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_1 4000
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_2 4020
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_3 4040
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_4 4060
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_5 4080
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_6 4100
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_7 4120
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_8 4140
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_9 4160
#define DIS_ARTICULATED_PART_PRIMARY_TURRET_NUMBER_10 4180
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_1 4200
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_2 4220
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_3 4240
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_4 4260
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_5 4280
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_6 4300
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_7 4320
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_8 4340
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_9 4360
#define DIS_ARTICULATED_PART_PRIMARY_GUN_NUMBER_10 4380
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_1 4400
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_2 4420
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_3 4440
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_4 4460
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_5 4480
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_6 4500
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_7 4520
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_8 4540
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_9 4560
#define DIS_ARTICULATED_PART_PRIMARY_LAUNCHER_10 4580
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_1 4600
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_2 4620
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_3 4640
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_4 4660
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_5 4680
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_6 4700
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_7 4720
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_8 4740
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_9 4760
#define DIS_ARTICULATED_PART_PRIMARY_DEFENSE_SYSTEMS_10 4780
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_1 4800
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_2 4820
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_3 4840
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_4 4860
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_5 4880
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_6 4900
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_7 4920
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_8 4940
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_9 4960
#define DIS_ARTICULATED_PART_PRIMARY_RADAR_10 4980
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_1 5000
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_2 5020
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_3 5040
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_4 5060
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_5 5080
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_6 5100
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_7 5120
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_8 5140
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_9 5160
#define DIS_ARTICULATED_PART_SECONDARY_TURRET_NUMBER_10 5180
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_1 5200
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_2 5220
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_3 5240
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_4 5260
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_5 5280
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_6 5300
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_7 5320
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_8 5340
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_9 5360
#define DIS_ARTICULATED_PART_SECONDARY_GUN_NUMBER_10 5380
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_1 5400
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_2 5420
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_3 5440
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_4 5460
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_5 5480
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_6 5500
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_7 5520
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_8 5540
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_9 5560
#define DIS_ARTICULATED_PART_SECONDARY_LAUNCHER_10 5580
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_1 5600
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_2 5620
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_3 5640
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_4 5660
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_5 5680
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_6 5700
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_7 5720
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_8 5740
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_9 5760
#define DIS_ARTICULATED_PART_SECONDARY_DEFENSE_SYSTEMS_10 5780
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_1 5800
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_2 5820
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_3 5840
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_4 5860
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_5 5880
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_6 5900
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_7 5920
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_8 5940
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_9 5960
#define DIS_ARTICULATED_PART_SECONDARY_RADAR_10 5980

typedef uint32    DIS_ARTICULATED_PART; 


/*
 * country to which an entity beint32s (Appendix F, section 30.1.2, 
 * pages 139-143)
 */

#define DIS_COUNTRY_OTHER       0    
#define DIS_COUNTRY_AFGHANISTAN 1    
#define DIS_COUNTRY_ALBANIA     2    
#define DIS_COUNTRY_ALGERIA     3    
#define DIS_COUNTRY_ANDORRA     4    
#define DIS_COUNTRY_ANGOLA      5    
#define DIS_COUNTRY_ANTARCTICA  6    
#define DIS_COUNTRY_ANTIGUA_AND_BARBUDA 7
#define DIS_COUNTRY_ARGENTINA   8    
#define DIS_COUNTRY_AUSTRALIA   9    
#define DIS_COUNTRY_AUSTRIA     10   
#define DIS_COUNTRY_BAHAMAS     11   
#define DIS_COUNTRY_BAHRAIN     12   
#define DIS_COUNTRY_BANGLADESH  13   
#define DIS_COUNTRY_BARBADOS    14   
#define DIS_COUNTRY_BELGIUM     15   
#define DIS_COUNTRY_BELIZE      16   
#define DIS_COUNTRY_BENIN       17   
#define DIS_COUNTRY_BHUTAN      18   
#define DIS_COUNTRY_BOLIVIA     19   
#define DIS_COUNTRY_BOTSWANA    20   
#define DIS_COUNTRY_BRAZIL      21   
#define DIS_COUNTRY_BRUNEI      22   
#define DIS_COUNTRY_BULGARIA    23   
#define DIS_COUNTRY_BURKINA_FASO 24  
#define DIS_COUNTRY_BURMA       25   
#define DIS_COUNTRY_BURUNDI     26   
#define DIS_COUNTRY_CAMBODIA_KAMPUCHEA 27
#define DIS_COUNTRY_CAMEROON    28   
#define DIS_COUNTRY_CANADA      29   
#define DIS_COUNTRY_CAPE_VERDE_ISLANDS 30
#define DIS_COUNTRY_CENTRAL_AFRICAN_REPUBLIC 31
#define DIS_COUNTRY_CHAD        32   
#define DIS_COUNTRY_CHILE       33   
#define DIS_COUNTRY_PEOPLES_REPUBLIC_OF_CHINA 34
#define DIS_COUNTRY_TAIWAN      35   
#define DIS_COUNTRY_COLOMBIA    36   
#define DIS_COUNTRY_COMORO_ISLANDS 37
#define DIS_COUNTRY_CONGO       38   
#define DIS_COUNTRY_COSTA_RICA  39   
#define DIS_COUNTRY_COTE_D_IVOIRE 40 
#define DIS_COUNTRY_CUBA        41   
#define DIS_COUNTRY_CYPRUS      42   
#define DIS_COUNTRY_CZECHOSLOVAKIA 43
#define DIS_COUNTRY_DENMARK     44   
#define DIS_COUNTRY_DJIBOUTI    45   
#define DIS_COUNTRY_DOMINICA    46   
#define DIS_COUNTRY_DOMINICAN_REPUBLIC 47
#define DIS_COUNTRY_ECUADOR     48   
#define DIS_COUNTRY_EGYPT       49   
#define DIS_COUNTRY_EL_SALVADOR 50   
#define DIS_COUNTRY_EQUATORIAL_GUINEA 51
#define DIS_COUNTRY_ETHIOPIA    52   
#define DIS_COUNTRY_FIJI        53   
#define DIS_COUNTRY_FINLAND     54   
#define DIS_COUNTRY_FRANCE      55   
#define DIS_COUNTRY_GABON       56   
#define DIS_COUNTRY_GERMANY     57   
#define DIS_COUNTRY_GHANA       58   
#define DIS_COUNTRY_GREECE      59   
#define DIS_COUNTRY_GRENADA     60   
#define DIS_COUNTRY_GUATEMALA   61   
#define DIS_COUNTRY_GUINEA      62   
#define DIS_COUNTRY_GUINEA_BISSUA 63 
#define DIS_COUNTRY_GUYANA      64   
#define DIS_COUNTRY_HAITI       65   
#define DIS_COUNTRY_HONDURAS    66   
#define DIS_COUNTRY_HUNGARY     67   
#define DIS_COUNTRY_ICELAND     68   
#define DIS_COUNTRY_INDIA       69   
#define DIS_COUNTRY_INDONESIA   70   
#define DIS_COUNTRY_IRAN        71   
#define DIS_COUNTRY_IRAQ        72   
#define DIS_COUNTRY_IRELAND     73   
#define DIS_COUNTRY_ISRAEL      74   
#define DIS_COUNTRY_ITALY       75   
#define DIS_COUNTRY_JAMAICA     76   
#define DIS_COUNTRY_JAPAN       77   
#define DIS_COUNTRY_JORDAN      78   
#define DIS_COUNTRY_KENYA       79   
#define DIS_COUNTRY_KIRIBATI    80   
#define DIS_COUNTRY_KOREA       81   
#define DIS_COUNTRY_DEMOCRATIC_PEOPLES_REPUBLIC_OF_KOREA 82
#define DIS_COUNTRY_REPUBLIC_OF_KOREA 83
#define DIS_COUNTRY_KUWAIT      84   
#define DIS_COUNTRY_LAOS        85   
#define DIS_COUNTRY_LEBANON     86   
#define DIS_COUNTRY_LESOTHO     87   
#define DIS_COUNTRY_LIBERIA     88   
#define DIS_COUNTRY_LIBYA       89   
#define DIS_COUNTRY_LIECHTENSTEIN 90 
#define DIS_COUNTRY_LUXEMBOURG  91   
#define DIS_COUNTRY_MADAGASCAR  92   
#define DIS_COUNTRY_MALAWI      93   
#define DIS_COUNTRY_MALAYSIA    94   
#define DIS_COUNTRY_MALDIVES    95   
#define DIS_COUNTRY_MALI        96   
#define DIS_COUNTRY_MALTA       97   
#define DIS_COUNTRY_MAURITANIA  98   
#define DIS_COUNTRY_MAURITIUS   99   
#define DIS_COUNTRY_MEXICO      100  
#define DIS_COUNTRY_MONACO      101  
#define DIS_COUNTRY_MONGOLIA    102  
#define DIS_COUNTRY_MOROCCO     103  
#define DIS_COUNTRY_MOSAMBIQUE  104  
#define DIS_COUNTRY_NAMIBIA     105  
#define DIS_COUNTRY_NAURU       106  
#define DIS_COUNTRY_NEPAL       107  
#define DIS_COUNTRY_NETHERLANDS 108  
#define DIS_COUNTRY_ARUBA       109  
#define DIS_COUNTRY_NETHERLANDS_ANTILLES 110
#define DIS_COUNTRY_NEW_ZEALAND 111  
#define DIS_COUNTRY_NICARAGUA   112  
#define DIS_COUNTRY_NIGER       113  
#define DIS_COUNTRY_NIGERIA     114  
#define DIS_COUNTRY_NORWAY      115  
#define DIS_COUNTRY_OMAN        116  
#define DIS_COUNTRY_PAKISTAN    117  
#define DIS_COUNTRY_PANAMA      118  
#define DIS_COUNTRY_PAPUA_NEW_GUINEA 119
#define DIS_COUNTRY_PARAGUAY    120  
#define DIS_COUNTRY_PERU        121  
#define DIS_COUNTRY_PHILIPPINES 122  
#define DIS_COUNTRY_POLAND      123  
#define DIS_COUNTRY_PORTUGAL    124  
#define DIS_COUNTRY_QATAR       125  
#define DIS_COUNTRY_ROMANIA     126  
#define DIS_COUNTRY_RWANDA      127  
#define DIS_COUNTRY_ST_CHRISTOPHER_AND_NEVIS 128
#define DIS_COUNTRY_ST_LUCIA    129  
#define DIS_COUNTRY_ST_VINCENT_AND_THE_GRENADINES 130
#define DIS_COUNTRY_SAN_MARINO  131  
#define DIS_COUNTRY_SAO_TOME_AND_PRINCIPE 132
#define DIS_COUNTRY_SAUDI_ARABIA 133 
#define DIS_COUNTRY_SENEGAMBIA  134  
#define DIS_COUNTRY_SENEGAL     135  
#define DIS_COUNTRY_GAMBIA      136  
#define DIS_COUNTRY_SEYCHELLES  137  
#define DIS_COUNTRY_SIERRA_LEONE 138 
#define DIS_COUNTRY_SINGAPORE   139  
#define DIS_COUNTRY_SOLOMON_ISLANDS 140
#define DIS_COUNTRY_SOMALIA     141  
#define DIS_COUNTRY_SOUTH_AFRICA 142 
#define DIS_COUNTRY_BOPHUTHATSWANA 143
#define DIS_COUNTRY_CISKEI      144  
#define DIS_COUNTRY_TRANSKEI    145  
#define DIS_COUNTRY_VENDA       146  
#define DIS_COUNTRY_SPAIN       147  
#define DIS_COUNTRY_SRI_LANKA   148  
#define DIS_COUNTRY_SUDAN       149  
#define DIS_COUNTRY_SURINAME    150  
#define DIS_COUNTRY_SWAZILAND   151  
#define DIS_COUNTRY_SWEDEN      152  
#define DIS_COUNTRY_SWITZERLAND 153  
#define DIS_COUNTRY_SYRIA       154  
#define DIS_COUNTRY_TANZANIA    155  
#define DIS_COUNTRY_THAILAND    156  
#define DIS_COUNTRY_TOGO        157  
#define DIS_COUNTRY_TONGA       158  
#define DIS_COUNTRY_TRINIDAD_AND_TOBAGO 159
#define DIS_COUNTRY_TUNISIA     160  
#define DIS_COUNTRY_TURKEY      161  
#define DIS_COUNTRY_TUVALU      162  
#define DIS_COUNTRY_UGANDA      163  
#define DIS_COUNTRY_UNION_OF_SOVIET_SOCIALIST_REPUBLICS 164
#define DIS_COUNTRY_UNITED_ARAB_EMIRATES 165
#define DIS_COUNTRY_UNITED_KINGDOM 166
#define DIS_COUNTRY_NORTHERN_IRELAND 167
#define DIS_COUNTRY_UNITED_STATES 168
#define DIS_COUNTRY_URUGUAY     169  
#define DIS_COUNTRY_VANUATU     170  
#define DIS_COUNTRY_VATICAN_CITY_STATE 171
#define DIS_COUNTRY_VENEZUELA   172  
#define DIS_COUNTRY_VIETNAM     173  
#define DIS_COUNTRY_WESTERN_SAMOA 174
#define DIS_COUNTRY_YEMEN_ARAB_REPUBLIC 175
#define DIS_COUNTRY_YEMEN_PEOPLES_DEMOCRATIC_REPUBLIC_OF 176
#define DIS_COUNTRY_YUGOSLAVIA  177  
#define DIS_COUNTRY_ZAIRE       178  
#define DIS_COUNTRY_ZAMBIA      179  
#define DIS_COUNTRY_ZIMBABWE    180  
#define DIS_COUNTRY_PALESTINE_LIBERATION_ORGANIZATION 181
#define DIS_COUNTRY_NEUTRAL     200  

typedef uint16   DIS_COUNTRY_ID; 


/*
 * The different dead reckoning algorithms (appendix I, section 30.7,
 * page 178)
 */

#define DIS_DEAD_RECKONING_ALGORITHM_OTHER 0
#define DIS_DEAD_RECKONING_ALGORITHM_STATIC 1
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_F_P_W 2
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_R_P_W 3
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_R_V_W 4
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_F_V_W 5
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_F_P_B 6
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_R_P_B 7
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_R_V_B 8
#define DIS_DEAD_RECKONING_ALGORITHM_DRM_F_V_B 9

typedef uint8    DIS_DEAD_RECKONING_ALGORITHM; 


/*
 * Results of detonation (appendix F, section 30.1.3, page 143)
 */

#define DIS_DETONATION_RESULT_DETONATION 0
#define DIS_DETONATION_RESULT_IMPACT 1
#define DIS_DETONATION_RESULT_NO_IMPACT_OR_DETONATION 2

typedef uint8    DIS_DETONATION_RESULT; 


/*
 * Air domain platform appearances (appendix D, section 30.1.2, pages 130-131)
 */

#define DIS_ENTITY_APPEARANCE_AIR_DESTROYED 0x1
#define DIS_ENTITY_APPEARANCE_AIR_FLAMING 0x2
#define DIS_ENTITY_APPEARANCE_AIR_AFTER_BURNER 0x4
#define DIS_ENTITY_APPEARANCE_AIR_RUNNING_LIGHTS 0x8
#define DIS_ENTITY_APPEARANCE_AIR_NAVIGATION_LIGHTS 0x10
#define DIS_ENTITY_APPEARANCE_AIR_FORMATION_LIGHTS 0x20

typedef uint32    DIS_ENTITY_APPEARANCE_AIR; 


/*
 * Cultural features (appendix D, section 30.5.1, page 134)
 */

#define DIS_ENTITY_APPEARANCE_CULTURAL_SLIGHT_DAMAGE 0x1
#define DIS_ENTITY_APPEARANCE_CULTURAL_MODERATE_DAMAGE 0x2
#define DIS_ENTITY_APPEARANCE_CULTURAL_DESTROYED 0x3
#define DIS_ENTITY_APPEARANCE_CULTURAL_SMOKE_PLUME 0x4
#define DIS_ENTITY_APPEARANCE_CULTURAL_FLAMING 0x10

typedef uint32    DIS_ENTITY_APPEARANCE_CULTURAL; 


/*
 * Environmental appearances (appendix D, section 30.4.1, page 133)
 */

#define DIS_ENTITY_APPEARANCE_ENVIRONMENTAL_HAZY 0x1
#define DIS_ENTITY_APPEARANCE_ENVIRONMENTAL_DENSE 0x2
#define DIS_ENTITY_APPEARANCE_ENVIRONMENTAL_VERY_DENSE 0x3
#define DIS_ENTITY_APPEARANCE_ENVIRONMENTAL_OPAQUE 0x4

typedef uint32    DIS_ENTITY_APPEARANCE_ENVIRONMENTAL; 


/*
 * Land domain platform appearances (appendix D, section 30.1.1, pages 129-130)
 */

#define DIS_ENTITY_APPEARANCE_LAND_DESTROYED 0x1
#define DIS_ENTITY_APPEARANCE_LAND_SMOKE_PLUME 0x2
#define DIS_ENTITY_APPEARANCE_LAND_FLAMING 0x4
#define DIS_ENTITY_APPEARANCE_LAND_SMALL_DUST_CLOUD 0x8
#define DIS_ENTITY_APPEARANCE_LAND_MEDIUM_DUST_CLOUD 0x10
#define DIS_ENTITY_APPEARANCE_LAND_LARGE_DUST_CLOUD 0x18
#define DIS_ENTITY_APPEARANCE_LAND_PAINT_SCHEME_WINTER_COLORS 0x20
#define DIS_ENTITY_APPEARANCE_LAND_PAINT_SCHEME_DESERT_COLORS 0x40
#define DIS_ENTITY_APPEARANCE_LAND_LAUNCHER_RAISED 0x80
#define DIS_ENTITY_APPEARANCE_LAND_ENGINE_SMOKE 0x100
#define DIS_ENTITY_APPEARANCE_LAND_HATCH_CLOSED 0x200
#define DIS_ENTITY_APPEARANCE_LAND_HATCH_OPEN 0x400
#define DIS_ENTITY_APPEARANCE_LAND_HATCH_OPEN_COMMANDER_LOOKING_OUT 0x600

typedef uint32    DIS_ENTITY_APPEARANCE_LAND; 


/*
 * Life form appearances (appendix D, section 30.3.1, page 133)
 */

#define DIS_ENTITY_APPEARANCE_LIFE_UPRIGHT 0x1
#define DIS_ENTITY_APPEARANCE_LIFE_KNEELING 0x2
#define DIS_ENTITY_APPEARANCE_LIFE_PRONE 0x3
#define DIS_ENTITY_APPEARANCE_LIFE_WEAPON_1_DEPLOYED 0x10
#define DIS_ENTITY_APPEARANCE_LIFE_WEAPON_2_DEPLOYED 0x20

typedef uint32    DIS_ENTITY_APPEARANCE_LIFE; 


/*
 * Guided munitions appearances (appendix D, section 30.2.1, page 132)
 */

#define DIS_ENTITY_APPEARANCE_MUNITION_LAUNCH_FLASH 0x1
#define DIS_ENTITY_APPEARANCE_MUNITION_ROCKET_FLAME 0x2

typedef uint32    DIS_ENTITY_APPEARANCE_MUNITION; 


/*
 * Space domain platform appearances (appendix D, section 30.1.5, page 132)
 */

#define DIS_ENTITY_APPEARANCE_SPACE_DESTROYED 0x1

typedef uint32    DIS_ENTITY_APPEARANCE_SPACE; 


/*
 * sub-surface domain platform appearances (appendix D, section 30.1.4, 
 * pages 131-132)
 */

#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_DESTROYED 0x1
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_SMALL_WAKE 0x2
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_MEDIUM_WAKE 0x4
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_LARGE_WAKE 0x6
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_RUNNING_LIGHTS 0x10
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_HATCH_CLOSED 0x20
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_HATCH_OPEN 0x40
#define DIS_ENTITY_APPEARANCE_SUB_SURFACE_HATCH_OPEN_COMMANDER_LOOKING_OUT 0x60

typedef uint32    DIS_ENTITY_APPEARANCE_SUB_SURFACE; 


/*
 * Surface domain platform appearances (appendix D, section 30.1.3, page 131)
 */

#define DIS_ENTITY_APPEARANCE_SURFACE_DESTROYED 0x1
#define DIS_ENTITY_APPEARANCE_SURFACE_FLAMING 0x2
#define DIS_ENTITY_APPEARANCE_SURFACE_SMALL_WAKE 0x4
#define DIS_ENTITY_APPEARANCE_SURFACE_MEDIUM_WAKE 0x8
#define DIS_ENTITY_APPEARANCE_SURFACE_LARGE_WAKE 0xc
#define DIS_ENTITY_APPEARANCE_SURFACE_RUNNING_LIGHTS 0x10

typedef uint32    DIS_ENTITY_APPEARANCE_SURFACE; 


/*
 * The domain which an entity operates (appendix H1, section 30.1.1.1, 
 * page 166)
 */

#define DIS_ENTITY_DOMAIN_OTHER 0    
#define DIS_ENTITY_DOMAIN_LAND  1    
#define DIS_ENTITY_DOMAIN_AIR   2    
#define DIS_ENTITY_DOMAIN_SURFACE 3  
#define DIS_ENTITY_DOMAIN_SUBSURFACE 4
#define DIS_ENTITY_DOMAIN_SPACE 5    

typedef uint8    DIS_ENTITY_DOMAIN; 


/*
 * The kind of entity type record.  (section 5.2.10.1, page 37 and
 * appendix H1, section 30.1, page 165)
 */

#define DIS_ENTITY_KIND_OTHER   0    
#define DIS_ENTITY_KIND_PLATFORM 1   
#define DIS_ENTITY_KIND_MUNITION 2   
#define DIS_ENTITY_KIND_LIFE_FORM 3  
#define DIS_ENTITY_KIND_ENVIRONMENTAL 4
#define DIS_ENTITY_KIND_CULTURAL_FEATURE 5

typedef uint8    DIS_ENTITY_KIND; 


/*
 * Character sets for DIS entity markings (appendix f, section 30.1.4, 
 * page 144)
 */

#define DIS_ENTITY_MARKING_CHAR_SET_UNUSED 0
#define DIS_ENTITY_MARKING_CHAR_SET_ASCII 1

typedef uint8    DIS_ENTITY_MARKING_CHAR_SET; 


/*
 * Force ID (section 5.2.14.1, page 39 and appendix E, section 30.1.1, 
 * page 135)
 * Blue forces are defined as U.S. NATO, SEATO, etc.
 * Red forces are defined as USSR, Warsaw Pact, Medellin Cartel, etc.
 * White forces are neutral.
 */

#define DIS_FORCE_OTHER         0    
#define DIS_FORCE_BLUE          1    
#define DIS_FORCE_RED           2    
#define DIS_FORCE_WHITE         3    

typedef uint8    DIS_FORCE_ID; 


/*
 * Different fuze types (appendix b, section 30.1, page 121)
 */

#define DIS_FUZE_OTHER          0    
#define DIS_FUZE_CONTACT        1000 
#define DIS_FUZE_CONTACT_INSTANT 1100
#define DIS_FUZE_CONTACT_DELAYED 1200
#define DIS_FUZE_TIMED          2000 
#define DIS_FUZE_PROXIMITY      3000 
#define DIS_FUZE_COMMAND        4000 
#define DIS_FUZE_ALTITUDE       5000 
#define DIS_FUZE_DEPTH          6000 
#define DIS_FUZE_ACOUSTIC       7000 

typedef uint16   DIS_FUZE_TYPE; 


/*
 * Air Force Hierarchy offsets in the hierarchy type (appendix E,
 * section 30.1.4.2, page 137)
 */

#define DIS_HIERARCHY_AIR_FORCE_AIR_FORCE 0
#define DIS_HIERARCHY_AIR_FORCE_WING 1
#define DIS_HIERARCHY_AIR_FORCE_SQUADRON 2
#define DIS_HIERARCHY_AIR_FORCE_FLIGHT 3
#define DIS_HIERARCHY_AIR_FORCE_DIVISION 4
#define DIS_HIERARCHY_AIR_FORCE_SECTION 5

typedef uint8    DIS_HIERARCHY_AIR_FORCE; 


/*
 * Army Hierarchy offsets in the hierarchy type (appendix E, 
 * section 30.1.4.1, page 136)
 */

#define DIS_HIERARCHY_ARMY_ARMY 0    
#define DIS_HIERARCHY_ARMY_CORPS 1   
#define DIS_HIERARCHY_ARMY_DIVISION 2
#define DIS_HIERARCHY_ARMY_REGIMENT 3
#define DIS_HIERARCHY_ARMY_BATTALION 4
#define DIS_HIERARCHY_ARMY_COMPANY 5 
#define DIS_HIERARCHY_ARMY_PLATOON 6 
#define DIS_HIERARCHY_ARMY_SQUAD 7   

typedef uint8    DIS_HIERARCHY_ARMY; 


/*
 * Marine Hierarchy offsets in the hierarchy type (appendix E,
 * section 30.1.4.4, page 137)
 */

#define DIS_HIERARCHY_MARINE_CORP_FLEET 0
#define DIS_HIERARCHY_MARINE_CORP_MARINE_EXPEDITIONARY_FORCE 1
#define DIS_HIERARCHY_MARINE_CORP_MARINE_EXPEDITIONARY_BRIGADE 2
#define DIS_HIERARCHY_MARINE_CORP_MARINE_EXPEDITIONARY_UNIT 3

typedef uint8    DIS_HIERARCHY_MARINE_CORP; 


/*
 * Navy Hierarch offsets in the hierarchy type (appendix E, 
 * section 30.1.4.5, page 138)
 */

#define DIS_HIERARCHY_NAVY_FLEET 0   
#define DIS_HIERARCHY_NAVY_FORCE 1   
#define DIS_HIERARCHY_NAVY_GROUP 2   
#define DIS_HIERARCHY_NAVY_UNIT 3    
#define DIS_HIERARCHY_NAVY_ELEMENT 4 

typedef uint8    DIS_HIERARCHY_NAVY; 


/*
 * The domain of the target of a munition (appendix H1, section 30.1.2.1,
 * page 166)
 */

#define DIS_MUNITION_DOMAIN_OTHER 0  
#define DIS_MUNITION_DOMAIN_ANTI_AIR 1
#define DIS_MUNITION_DOMAIN_ANTI_ARMOR 2
#define DIS_MUNITION_DOMAIN_ANTI_GUIDED_MUNITION 3
#define DIS_MUNITION_DOMAIN_ANTI_RADAR 4
#define DIS_MUNITION_DOMAIN_ANTI_SATELLITE 5
#define DIS_MUNITION_DOMAIN_ANTI_SHIP 6
#define DIS_MUNITION_DOMAIN_ANTI_SUBMARINE 7
#define DIS_MUNITION_DOMAIN_BATTLEFIELD_SUPPORT 8
#define DIS_MUNITION_DOMAIN_STRATEGIC 9
#define DIS_MUNITION_DOMAIN_MISCELLANEOUS 10

typedef uint8    DIS_MUNITION_DOMAIN; 


/*
 * Repairs (appendix C, pages 125-128)
 * Repairs are described in two parts - a repair code and the item being
 * repaired.  A repair code is a number between from 0 to 9, inclusive.
 * The items that can be repaired are described in multiples of 10.  A
 * repair is described by adding the type of repair (the repair code) to
 * the item being repaired.  This is done in 16 bits.  In other words,
 * DIS_REPAIR = DIS_REPAIR_CODE + DIS_REPAIR_SYSTEM
 */

typedef uint16   DIS_REPAIR; 

#define DIS_REPAIR_CODE_NO_REPAIRS_PERFORMED 0
#define DIS_REPAIR_CODE_REPAIR_EVERYTHING 1

typedef uint16   DIS_REPAIR_CODE; 

#define DIS_REPAIR_SYSTEM_MOTOR 10   
#define DIS_REPAIR_SYSTEM_STARTER 20 
#define DIS_REPAIR_SYSTEM_ALTERNATOR 30
#define DIS_REPAIR_SYSTEM_GENERATOR 40
#define DIS_REPAIR_SYSTEM_BATTERY 50 
#define DIS_REPAIR_SYSTEM_ENGINE_COOLANT_LEAK 60
#define DIS_REPAIR_SYSTEM_FUEL_FILTER 70
#define DIS_REPAIR_SYSTEM_TRANSMISSION_OIL_LEAK 80
#define DIS_REPAIR_SYSTEM_ENGINE_OIL_LEAK 90
#define DIS_REPAIR_SYSTEM_PUMPS 100  
#define DIS_REPAIR_SYSTEM_PROPULSION_SYSTEM_FILTERS 110
#define DIS_REPAIR_SYSTEM_HULL  1000 
#define DIS_REPAIR_SYSTEM_AIRFRAME 1010
#define DIS_REPAIR_SYSTEM_PROPELLER 1500
#define DIS_REPAIR_SYSTEM_TRANSMISSION 1510
#define DIS_REPAIR_SYSTEM_ENVIRONMENTAL_FILTERS 1520
#define DIS_REPAIR_SYSTEM_BRAKES 1530
#define DIS_REPAIR_SYSTEM_GUN_ELEVATION_DRIVE 2000
#define DIS_REPAIR_SYSTEM_GUN_STABILIZATION_SYSTEM 2010
#define DIS_REPAIR_SYSTEM_GUNNERS_PRIMARY_SIGHT 2020
#define DIS_REPAIR_SYSTEM_GUNNERS_EXTENSION_TO_THE_GPS 2030
#define DIS_REPAIR_SYSTEM_LOADING_MECHANISM 2040
#define DIS_REPAIR_SYSTEM_FUEL_TRANSFER_PUMP 4000
#define DIS_REPAIR_SYSTEM_FUEL_LINES 4010
#define DIS_REPAIR_SYSTEM_GAUGES 4020
#define DIS_REPAIR_SYSTEM_ELECTRONIC_COUNTER_MEASURE_SYSTEMS 4500
#define DIS_REPAIR_SYSTEM_ELECTRONIC_WARFARE_SYSTEMS 4600
#define DIS_REPAIR_SYSTEM_LASER_RANGE_FINDER 4700
#define DIS_REPAIR_SYSTEM_RADIOS 4800
#define DIS_REPAIR_SYSTEM_INTERCOMS 4900
#define DIS_REPAIR_SYSTEM_CODERS 5000
#define DIS_REPAIR_SYSTEM_DECODERS 5100
#define DIS_REPAIR_SYSTEM_LASERS 5200
#define DIS_REPAIR_SYSTEM_COMPUTERS 5300
#define DIS_REPAIR_SYSTEM_EMITTERS 5400
#define DIS_REPAIR_SYSTEM_DETECTION_SYSTEMS 5600
#define DIS_REPAIR_SYSTEM_AIR_SUPPLY 8000
#define DIS_REPAIR_SYSTEM_LIFE_SUPPORT_FILTERS 8010
#define DIS_REPAIR_SYSTEM_LIFE_SUPPORT_WATER_SUPPLY 8020
#define DIS_REPAIR_SYSTEM_REFRIDGERATION_SYSTEM 8030
#define DIS_REPAIR_SYSTEM_CHEMICAL_BIOLOGICAL_RADIOLOGIC_PROTECTION 8040
#define DIS_REPAIR_SYSTEM_WATER_WASH_DOWN_SYSTEMS 8050
#define DIS_REPAIR_SYSTEM_DECONTAMINATION_SYSTEMS 8060
#define DIS_REPAIR_SYSTEM_HYDRAULICS_WATER_SUPPLY 9000
#define DIS_REPAIR_SYSTEM_COOLING_SYSTEM 9010
#define DIS_REPAIR_SYSTEM_WINCHES 9020
#define DIS_REPAIR_SYSTEM_CATAPULTS 9030
#define DIS_REPAIR_SYSTEM_CRANES 9040
#define DIS_REPAIR_SYSTEM_LAUNCHERS 9050
#define DIS_REPAIR_SYSTEM_LIFE_BOATS 10000
#define DIS_REPAIR_SYSTEM_LANDING_CRAFT 10010
#define DIS_REPAIR_SYSTEM_EJECTION_SEATS 10020

typedef uint16   DIS_REPAIR_SYSTEM; 


/*
 * The results of a repair (appendix f, section 30.1.8)
 */

#define DIS_REPAIR_RESULT_OTHER 0    
#define DIS_REPAIR_RESULT_REPAIR_ENDED 1
#define DIS_REPAIR_RESULT_INVALID_REPAIR 2
#define DIS_REPAIR_RESULT_REPAIR_INTERRUPTED 3
#define DIS_REPAIR_RESULT_SERVICE_CANCELLED_BY_THE_SUPPLIER 4

typedef uint8    DIS_REPAIR_RESULT; 


/*
 * The different branches of the service (appendix E, section 30.1.3,
 * page 136)
 */

#define DIS_SERVICE_ID_OTHER    0    
#define DIS_SERVICE_ID_ARMY     1    
#define DIS_SERVICE_ID_AIR_FORCE 2   
#define DIS_SERVICE_ID_COAST_GUARD 3 
#define DIS_SERVICE_ID_MARINES  4    
#define DIS_SERVICE_ID_NAVY     5    

typedef uint8    DIS_SERVICE_ID; 


/*
 * The different types of service an entity can provide (appendix f,
 * section 30.1.9)
 */

#define DIS_SERVICE_TYPE_OTHER  0    
#define DIS_SERVICE_TYPE_RESUPPLY 1  
#define DIS_SERVICE_TYPE_REPAIR 2    

typedef uint8    DIS_SERVICE_TYPE; 


/*
 * Different warhead types (appendix b, section 30.1, page 122)
 */

#define DIS_WARHEAD_OTHER       0    
#define DIS_WARHEAD_HIGH_EXPLOSIVE 1000
#define DIS_WARHEAD_HIGH_EXPLOSIVE_PLASTIC 1100
#define DIS_WARHEAD_HIGH_EXPLOSIVE_INCENDIARY 1200
#define DIS_WARHEAD_HIGH_EXPLOSIVE_FRAGMENTATION 1300
#define DIS_WARHEAD_HIGH_EXPLOSIVE_ANTI_TANK 1400
#define DIS_WARHEAD_HIGH_EXPLOSIVE_BOMBLETS 1500
#define DIS_WARHEAD_HIGH_EXPLOSIVE_SHAPED_CHARGE 1600
#define DIS_WARHEAD_SMOKE       2000 
#define DIS_WARHEAD_ILLUMINATION 3000
#define DIS_WARHEAD_PRACTICE    4000 
#define DIS_WARHEAD_KINETIC     5000 
#define DIS_WARHEAD_UNUSED      6000 
#define DIS_WARHEAD_NUCLEAR     7000 
#define DIS_WARHEAD_CHEMICAL_GENERAL 8000
#define DIS_WARHEAD_CHEMICAL_BLISTER_AGENT 8100
#define DIS_WARHEAD_CHEMICAL_BLOOD_AGENT 8200
#define DIS_WARHEAD_CHEMICAL_NERVE_AGENT 8300
#define DIS_WARHEAD_BIOLOGICAL_GENERAL 9000

typedef uint16   DIS_WARHEAD_TYPE; 


/*
 * Different deactivate reasons (section 6.3.1.3.3 letter c, page 86)
 */

#define DIS_DEACTIVATE_REASON_OTHER 1
#define DIS_DEACTIVATE_REASON_EXERCISE_END 2
#define DIS_DEACTIVATE_REASON_ENTITY_WITHDRAWN 3
#define DIS_DEACTIVATE_REASON_ENTITY_DESTROYED 4

typedef uint8    DIS_DEACTIVATE_REASON; 


/*
 * Different radar modes (section 30.2.7, page 148)
 */

#define DIS_RADAR_MODE_OFF      0    
#define DIS_RADAR_MODE_SEARCH   1    
#define DIS_RADAR_MODE_DOPPLER_HPRF 2
#define DIS_RADAR_MODE_DPPPLER_MPRF 3
#define DIS_RADAR_MODE_DPPPLER_LPRF 4
#define DIS_RADAR_MODE_MONOPULSE 5   
#define DIS_RADAR_MODE_ACQUISITION 6 
#define DIS_RADAR_MODE_TRACKING 7    
#define DIS_RADAR_MODE_TRACK_WHILE_SCAN 8
#define DIS_RADAR_MODE_TERRAIN_FOLLOW 9
#define DIS_RADAR_MODE_DATA_LINK 10  

typedef uint8    DIS_RADAR_MODE; 


/*
 * Different radar system categories (section 30.2.8, page 148)
 */

#define DIS_RADAR_SYSTEM_CATEGORY_RESERVED 0
#define DIS_RADAR_SYSTEM_CATEGORY_AIR_BASED_FIRE_CONTROL 1
#define DIS_RADAR_SYSTEM_CATEGORY_AIR_BASE_SEARCH 2
#define DIS_RADAR_SYSTEM_CATEGORY_GROUND_BASED_FIRE_CONTROL 3
#define DIS_RADAR_SYSTEM_CATEGORY_GROUND_BASED_SEARCH 4
#define DIS_RADAR_SYSTEM_CATEGORY_SEA_BASED_FIRE_CONTROL 5
#define DIS_RADAR_SYSTEM_CATEGORY_SEA_BASED_SEARCH 6

typedef uint8    DIS_RADAR_SYSTEM_CATEGORY; 


/*
 * Different radar system ID's (section 30.2.10, page 149)
 */

#define DIS_RADAR_SYSTEM_ID_RESERVED 0
#define DIS_RADAR_SYSTEM_ID_AN_APG_66 1
#define DIS_RADAR_SYSTEM_ID_AN_APG_68 2
#define DIS_RADAR_SYSTEM_ID_AN_APG_63 3
#define DIS_RADAR_SYSTEM_ID_AN_APG_65 4
#define DIS_RADAR_SYSTEM_ID_AN_APG_70 5
#define DIS_RADAR_SYSTEM_ID_JAYBIRD 6
#define DIS_RADAR_SYSTEM_ID_MIG_31 7 
#define DIS_RADAR_SYSTEM_ID_MIG_29 8 
#define DIS_RADAR_SYSTEM_ID_MIG_27 9 
#define DIS_RADAR_SYSTEM_ID_SU_27 10 
#define DIS_RADAR_SYSTEM_ID_AN_APY_2 11
#define DIS_RADAR_SYSTEM_ID_SUAWACS 12
#define DIS_RADAR_SYSTEM_ID_FOXFIRE 13
#define DIS_RADAR_SYSTEM_ID_HIGHLARK 14
#define DIS_RADAR_SYSTEM_ID_AN_APS_125 15
#define DIS_RADAR_SYSTEM_ID_LN_66_HP 16
#define DIS_RADAR_SYSTEM_ID_AN_APS_116 17
#define DIS_RADAR_SYSTEM_ID_AN_APS_115 18
#define DIS_RADAR_SYSTEM_ID_AN_SPQ_9 19
#define DIS_RADAR_SYSTEM_ID_AN_SPQ_9A 20
#define DIS_RADAR_SYSTEM_ID_AN_SPG_60 21
#define DIS_RADAR_SYSTEM_ID_AN_SPS_49 22
#define DIS_RADAR_SYSTEM_ID_AN_SPS_55 23
#define DIS_RADAR_SYSTEM_ID_AN_SPS_67 24
#define DIS_RADAR_SYSTEM_ID_AN_SPS_10 25
#define DIS_RADAR_SYSTEM_ID_AN_SPY_1 26
#define DIS_RADAR_SYSTEM_ID_AWG_9 27 
#define DIS_RADAR_SYSTEM_ID_APG_71 28

typedef uint16   DIS_RADAR_SYSTEM_ID; 


/*
 * Radio enumerations
 *
 */

#define DIS_RADIO_TRANSMIT_STATE_OFF 0
#define DIS_RADIO_TRANSMIT_STATE_ON_IDLE 1
#define DIS_RADIO_TRANSMIT_STATE_ON_ACTIVE 2

typedef uint8    DIS_RADIO_TRANSMIT_STATE; 


/*
 * Radio antenna pattern types (DIS 2.0 section 5.4.7.1.2, page 146).
 *
 */

#define DIS_RADIO_ANTENNA_PATTERN_OMNI 0
#define DIS_RADIO_ANTENNA_PATTERN_BEAM 1

typedef uint16   DIS_RADIO_ANTENNA_PATTERN_TYPE; 

#define DIS_RADIO_MODULATION_MAJOR_OTHER 0
#define DIS_RADIO_MODULATION_MAJOR_AMPLITUDE 1
#define DIS_RADIO_MODULATION_MAJOR_ANGLE 2
#define DIS_RADIO_MODULATION_MAJOR_AMP_ANG 3
#define DIS_RADIO_MODULATION_MAJOR_COMBO 4
#define DIS_RADIO_MODULATION_MAJOR_PULSE 5

typedef uint16   DIS_RADIO_MODULATION_MAJOR; 

#define DIS_RADIO_MODULATION_DETAIL_OTHER 0
#define DIS_RADIO_MODULATION_DETAIL_FM 1
#define DIS_RADIO_MODULATION_DETAIL_FSK 2
#define DIS_RADIO_MODULATION_DETAIL_PM 3

typedef uint16   DIS_RADIO_MODULATION_DETAIL; 

#define DIS_RADIO_MODULATION_SYSTEM_OTHER 0
#define DIS_RADIO_MODULATION_SYSTEM_GENERIC 1
#define DIS_RADIO_MODULATION_SYSTEM_HQ 2
#define DIS_RADIO_MODULATION_SYSTEM_HQII 3
#define DIS_RADIO_MODULATION_SYSTEM_HQIIA 4
#define DIS_RADIO_MODULATION_SYSTEM_SINCGARS 5

typedef uint16   DIS_RADIO_MODULATION_SYSTEM; 

#define DIS_RADIO_CRYPTO_SYSTEM_OTHER 0
#define DIS_RADIO_CRYPTO_SYSTEM_KY_28 1
#define DIS_RADIO_CRYPTO_SYSTEM_KY_58 2

typedef uint16   DIS_RADIO_CRYPTO_SYSTEM; 

#define DIS_RADIO_ENCODING_SCHEME_MU_LAW 1
#define DIS_RADIO_ENCODING_SCHEME_CVSD 2
#define DIS_RADIO_ENCODING_SCHEME_ADPCM 3
#define DIS_RADIO_ENCODING_SCHEME_PCM 4
#define DIS_RADIO_ENCODING_SCHEME_RAW 16384
#define DIS_RADIO_ENCODING_SCHEME_APPL_S 32768
#define DIS_RADIO_ENCODING_SCHEME_PTR 49152

typedef uint16   DIS_RADIO_ENCODING_SCHEME; 



#endif
