
/*
 * FILE:  ivis_com.h 
 *
 * Created on Wed Apr 17 10:52:48 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _ivis_com_included
#define _ivis_com_included



/*
 *  This file contains definitions related to communication among cvcc
 *  applications..
 *
 *
 *  Copyright (c) 1990.  All rights reserved.
 *  BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 *  History:
 *    12/03/91  Schwartz	protocol overhaul
 *    09/07/90  Schwartz	put Ivis prefix on constants
 *    04/25/90  Pope		translated from previous ivis_comm.h
 *				  and ivis_msg.h
 *
 * $RCSfile: ivis_com.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <mos.h> 
#include <ivis_rpt.h> 

/*
 *  Several combat radio networks are operated within a battalion.  
 *  These networks correspond to platoons, companies and battalions.
 *  Platoon networks correspond to the battalion number, company 
 *  letter and platoon number of the corresponding platoon.  Company
 *  networks are defined by the company's letter and battalion number.
 *  Battalion networks are defined by the battalion's number.  Since
 *  the networks correspond so closely to unit identifiers, and the
 *  largest number of unit identifiers needed is 3 plus one for 
 *  termination, networks can be defined by an array of unit identifiers.
 *
 */

typedef UnitIdentifier cvcc_RadioNetworkID[4]; 
 /* 8 bytes */

 /*  Cvcc applications communicate information using "messages".  The */
 /*  following are enumerations needed by a cvcc message: */

 /*  Messages are labelled, by experimenters, according to their relevance: */
#define cvcc_MsgRelevanceUndefined 0 
#define cvcc_MsgRelevanceLow    1    
#define cvcc_MsgRelevanceMedium 2    
#define cvcc_MsgRelevanceHigh   3    

typedef uint8    cvcc_MsgRelevance; 


/*
 *  Priorities of messages:
 */

#define cvcc_PriorityStatus     0    
#define cvcc_PriorityLow        1    
#define cvcc_PriorityHigh       2    

typedef uint8    cvcc_MsgPriority; 


/*
 *  Cvcc messages can come from different applications.  These are the
 *  different applications that can originate a cvcc message:
 */

#define cvcc_FromIvis           0    
#define cvcc_FromBntoc          1    
#define cvcc_FromSendUtility    2    
#define cvcc_FromSaf            3    

typedef uint8    cvcc_MsgOrigType; 


/*
 *  There are two different types of messages: objects and information.  An
 *  object message is something that should be saved; information messages
 *  contain information pertinent to the simulation, but the message itself
 *  will not be saved.  Objects are reports, routes, and mos symbols.
 *  Information is status of ivis simulators, task organization, bntoc utm
 *  location, and posting to and unposting from the situation display.
 *
 */

#define cvcc_ObjMsg             0     /* an object (something saved) */
#define cvcc_InfoMsg            1     /* information (used and thrown away) */
typedef uint8    cvcc_MsgType; 

#define cvcc_ReportObj          0     /* a report */
#define cvcc_RouteObj           1     /* a route */
#define cvcc_MosObj             2     /* mos symbol */
typedef uint8    cvcc_ObjType; 

#define cvcc_StatusInfo         0     /* status of ivis simulators */
#define cvcc_TaskOrgInfo        1     /* task organization assignment */
#define cvcc_BntocUtmInfo       2     /* where a bntoc is located */
#define cvcc_SitDispInfo        3     /* (un)post to/from situation display */
typedef uint8    cvcc_InfoType; 


/*
 *  cvcc_Report is defined in ivis_rpt.drn; mos_Overlay is defined in mos.drn.
 *  cvcc_Route, used in the navigation module in ivis, is defined here:
 *						-- 6 * 8 = 48 bytes
 */

#define cvcc_MaxWaypointsPerRoute           6
typedef struct {
    IVIS_Location waypoints[cvcc_MaxWaypointsPerRoute]; 
     /* list of waypoints in the route */

} cvcc_Route; 


/*
 *  An object contains the object id, information pertaining to the creation
 *  of the object and its originator, the sender, the object type and its 
 *  contents
 *	-- 12 + 6 + 1 + 1 + 4  + 6 + 1 + 1 + max(516,48,628)  = 660 bytes
 */

typedef struct {
    cvcc_ObjID    objid;              /* object id (defn in cvcc_types.drn) */

    cvcc_CallSign orig;               /* call sign of the msg originator */

    cvcc_MsgOrigType origType;        /* what the originator is */

    cvcc_MsgRelevance relevance;      /* relevance of message */

    SP_Time          origTime;           /* object creation time */

    cvcc_CallSign sender;             /* sender's call sign */

    cvcc_ObjType  otype;              /* type of object */

    uint8    _unused_19;         /* padding */
    union {
        cvcc_Report       rpt;
        cvcc_Route        route;
        mos_OverlayPacket mos;
    } u;                                
} cvcc_Object; 


/*
 *  Periodically, the ivis simulators send out an information message
 *  describing its status.  Status information messages are automatically
 *  gathered and relayed up and down the chain of command.  The following
 *  are the enumerations used in a status message, and definitions of
 *  the structures which describe an ivis vehicle's status
 *
 *  The different status types
 */

#define cvcc_StFireEquipment    0     /* equipment status types */
#define cvcc_StMobilityEquipment 1   
#define cvcc_StCommander        2     /* personnel status types */
#define cvcc_StGunner           3    
#define cvcc_StDriver           4    
#define cvcc_StLoader           5    
#define cvcc_StFuel             6     /* fuel status type */
typedef uint8 cvcc_StatusType; 

#define cvcc_NumberStatusTypes  7 /* number of different status types */
 /*  The different status codes */
#define cvcc_ScUndefined        0    
#define cvcc_ScEquipmentFunctioning 1 /* equipment status codes */
#define cvcc_ScEquipmentNotFunctioning 2
#define cvcc_ScPersonOkay       3     /* personnel status codes */
#define cvcc_ScPersonWounded    4    
#define cvcc_ScPersonKilled     5    

typedef uint8 cvcc_StatusCode; 
 /*  A status tuple				-- 1 + 1 + max(1,2) = 4 bytes */
typedef struct {
    cvcc_StatusType stype;            /* which thing */

    uint8 _unused_20;            /* padding */
    union {
         /* status value */
        cvcc_StatusCode scode;
         /* 		when (cvcc_StMobilityEquipment)	scode	cvcc_StatusCode, */
         /* 		when (cvcc_StCommander)		scode	cvcc_StatusCode, */
         /* 		when (cvcc_StGunner)		scode	cvcc_StatusCode, */
         /* 		when (cvcc_StDriver)		scode	cvcc_StatusCode, */
         /* 		when (cvcc_StLoader)		scode	cvcc_StatusCode, */
        uint16    samount;
    } u;                                
} cvcc_ItemStatus; 
 /*  An ammo tuple				-- 4 + 4 = 8 bytes */
typedef struct {
    ObjectType atype;                 /* ammo type, from obj_type.h */

    uint32 aamount;               /* how much of the ammo type */

} cvcc_AmmoStatus; 

#define cvcc_NumberAmmoTypes    4 /*  Status of a cvcc vehicle	 */
 /*  12 + 10 + 10 + 1/8 + 3-7/8 + 4 + 8 + 4 + 7*4 + 4*8 = 112 bytes */
typedef struct {
    cvcc_Tasking        dutyPosition;  /* the duty position of the vehicle */

    cvcc_UnitIdentifier tasking;      /* the unit the vehicle is reporting to */

    cvcc_UnitIdentifier leadsUnit;    /* unit that is being led */

    unsigned               amUnitLeader  : 1;  /* Boolean */
     /* is this vehicle leading something? */

    unsigned               _unused_21    : 31;  /* padding */
    ObjectType          vehType;      /* the type of vehicle it is */

    IVIS_Location       loc;          /* location of the vehicle */

    SP_Time                timestamp;    /* time this info was issued */

    cvcc_ItemStatus     status[cvcc_NumberStatusTypes]; 
     /* state of affairs in the tank */

    cvcc_AmmoStatus     ammo[cvcc_NumberAmmoTypes]; 
     /* amount of ammo in the tank */

} cvcc_VehicleStatus; 
 /*  Status of a group of vehicles (Arbitrarily, the maximum size of a */
 /*  group has been decided to be 9 vehicles.  This number is derived */
 /*  such that the status packet will be about 1000 bytes */
 /*  					-- 2 + 6 + 9*112 = 1016 bytes */
#define cvcc_StatusGroupSize    9
typedef struct {
    uint16        number;        /* number in group */

    cvcc_CallSign      sender;        /* sender of the status information */

    cvcc_VehicleStatus veh[cvcc_StatusGroupSize]; 
} cvcc_StatusGroup; 


/*
 *  Bntoc can send out task organization messages.  These messages tell
 *  the ivis's who they are supposed to report to.	-- 12 + 12 = 24 bytes
 */

typedef struct {
    cvcc_Tasking toe;                 /* TO&E of unit receiving orders */

    cvcc_Tasking to;                  /* new task organization */

} cvcc_TaskingOrders; 


/*
 *  When the bntocs start up, or change location in the war field, they
 *  need to send each other messages indicating their position.
 *							-- 8 + 2*4 = 16 bytes
 */

typedef struct {
    IVIS_Location  loc;               /* location of the bntoc */

    UnitIdentifier bn[2];             /* battalion number */

} cvcc_BntocUtm; 


/*
 *  A bntoc workstation can post and unpost reports and overlays to
 *  the situation display.		-- 1 + 1 + 6 + max(12,6) = 20 bytes
 */

#define cvcc_SdPost             0     /* post something */
#define cvcc_SdUnpost           1     /* unpost something */
typedef uint8 cvcc_SitDispAction; 

#define cvcc_SdReport           0     /* a report */
#define cvcc_SdOverlay          1     /* an overlay */
typedef uint8 cvcc_SitDispWhat; 

typedef struct {
    cvcc_SitDispAction action;        /* post or unpost */

    cvcc_SitDispWhat what;            /* report or overlay */

    cvcc_CallSign poster;             /* call sign of the (un)poster */

    union {
        cvcc_ObjID        rid;
         /* report id */
         /* overlay name */
        mos_Tag           name;
    } u;                                
} cvcc_SitDisp; 


/*
 *  An information message contains information pertinent to the simulation.
 *  The information is used, but the message it came in is not saved.  This
 *  is the distinction between objects and information.
 *				-- 1 + 3 + max(1016,24,16,20,146) = 1020 bytes
 */

typedef struct {
    cvcc_InfoType itype;              /* type of information message */

    unsigned      _unused_22 : 24;       /* padding */
    union {
        cvcc_StatusGroup   status;
         /* vehicle status message */
        cvcc_TaskingOrders tasking;
         /* task organization orders */
        cvcc_BntocUtm      bntocLoc;
         /* bntoc location */
         /* un/post to situation display */
        cvcc_SitDisp       sitDisp;
    } u;                                
} cvcc_Info; 


/*
 *  A cvcc message is either an object or information.  Messages have 
 *  priorities.			-- 1 + 1 + 2 + max(660,1020) = 1024
 */

typedef struct {
    cvcc_MsgType mtype;               /* type of message */

    cvcc_MsgPriority priority;        /* priority of message */

    uint16 _unused_23;           /* padding */
    union {
        cvcc_Object    obj;
        cvcc_Info      info;
    } u;                                
} cvcc_Msg; 



#endif
