
/*
 * FILE:  ivis_rpt.h 
 *
 * Created on Wed Apr 17 10:52:48 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _ivis_rpt_included
#define _ivis_rpt_included



/*
 * This file defines a machine representation for IVIS reports of various
 * kinds.
 *
 *
 * Copyright (c) 1990.  All rights reserved.
 * BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 * History:
 *   12/09/91	Schwartz	minor report format changes
 *   09/07/90	Schwartz	put Ivis prefix on constants
 *   06/04/90	Schwartz	changed Bool fields in adjust fire rpt to chars
 *   05/16/90 	Kanarick	nbc attack type made IVIS_ObjectType
 *   04/25/90	Pope		translated from previous ivis_report.h
 *   01/02/90	Dubrovsky  	added NBC report sequenceure.
 *   11/04/89	Pope		removed heading field from Shell report
 *   11/11/89	Pope		added Intelligence report
 *   11/17/89	Pope		added FRAGO report
 *
 * $RCSfile: ivis_rpt.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <cvcc_types.h> 
typedef float32            IVIS_Location[2]; 

#define IvisLocationUnknown     1000000 /*  */
 /* Constants identity the various types of IVIS reports. */
#define IvisReportAdjustFire    0     /* adjust fire */
#define IvisReportCallForFire   1     /* call for fire */
#define IvisReportContact       2     /* contact report */
#define IvisReportShell         3     /* shell report */
#define IvisReportSit           4     /* situation report */
#define IvisReportSpot          5     /* spot report */
#define IvisReportIntel         6     /* intelligence report */
#define IvisReportFreeText      7     /* free text report */
#define IvisReportNBC           8     /* Nuc, Bio, Chem report */
typedef uint8    cvcc_ReportType; 

#define cvcc_NumberReportTypes  9 /*  */
 /*  Definitions of values used in various report fields. */


 /* Types of friendly or enemy unit activities: */
#define IvisActivityUnknown     0     /* activity unknown */
#define IvisActivityAirAttack   1     /* air attack */
#define IvisActivityAttack      2     /* attacking */
#define IvisActivityDefend      3     /* defending */
#define IvisActivityDelay       4     /* delaying */
#define IvisActivityFire        5     /* firing */
#define IvisActivityGroundAttack 6    /* ground attack */
#define IvisActivityNoChange    7     /* no change in activity */
#define IvisActivityRecon       8     /* reconnaissance */
#define IvisActivityWithdraw    9     /* withdraw */
typedef uint8    IVIS_ActivityType; 


/*
 * Levels of activity:
 */

#define IvisActivityLevelUnknown 0    /* unknown level of activity */
#define IvisActivityLevelLight  1     /* light enemy activity */
#define IvisActivityLevelMedium 2     /* medium enemy activity */
#define IvisActivityLevelHeavy  3     /* heavy enemy activity */
typedef uint8    IVIS_ActivityLevel; 


/*
 * Types of NBC bursts:
 */

#define IvisNbcBurstUnknown     0     /* unknown type of burst */
#define IvisNbcBurstAir         1     /* air burst */
#define IvisNbcBurstSurface     2     /* surface burst */
typedef uint8    IVIS_BurstType; 


/*
 * Status of a supply quantity:
 */

#define IvisStatusUnknown       0    
#define IvisStatusGreen         1    
#define IvisStatusAmber         2    
#define IvisStatusRed           3    
#define IvisStatusBlack         4    

typedef uint8    IVIS_SupplyStatus; 


/*
 * Types of objects described in IVIS reports:
 */

#define IvisObjectTypeUnknown   0     /* unknown */
#define IvisTargetTypeTank      1     /* tanks */
#define IvisTargetTypeATGM      2     /* anti-tank guided missiles */
#define IvisTargetTypeHelo      3     /* helicopters */
#define IvisTargetTypeFWAir     4     /* fixed-wing aircraft */
#define IvisTargetTypeARTY      5     /* artillery */
#define IvisTargetTypePC        6     /* personnel carriers */
#define IvisTargetTypeTruck     7     /* trucks */
#define IvisTargetTypeTroop     8     /* troops */
 /* Types of friendly units: */
#define IvisFriendTypeUnknown   10    /* unknown friendly */
#define IvisFriendTypeArty      11    /* artillery */
#define IvisFriendTypeC2        12    /* command and control */
#define IvisFriendTypeMech      13    /* mechanized infantry */
#define IvisFriendTypeMortar    14    /* mortar */
#define IvisFriendTypeScout     15    /* scouts */
#define IvisFriendTypeSupport   16    /* support */
#define IvisFriendTypeTank      17    /* tanks */
 /* Types of obstacles: */
#define IvisObstacleTypeUnknown 20    /* unknown obstacle */
#define IvisObstacleTypeAbati   21    /* felled trees */
#define IvisObstacleTypeBlownBridge 22 /* blown bridge */
#define IvisObstacleTypeMinefield 23  /* minefield */
#define IvisObstacleTypeTankDitch 24  /* tank ditch */
 /* Types of NBC attacks: */
#define IvisNbcAttackNuclear    25   
#define IvisNbcAttackBiological 26   
#define IvisNbcAttackChemical   27    /* Observer of NBC attack: */
#define IvisNbcObsLoc           28    /* Artillery shell */
#define IvisArtyShell           29    /* FLOT */
#define IvisFlot                30   

typedef uint8    IVIS_ObjectType; 


/*
 * Constants for indicating unknown numeric values:
 */

#define IvisNumberUnknown       0 /* user has not filled in a number */
#define IvisHeadingUnknown      1000 /* user has not filled in a heading */

 /*  Adjust fire report. */
 /*  1 + 1 + 1 + 1 + 2 + 2 + 8 + 12 = 28 bytes */
typedef struct {
    uint8    relative;           /* 0: abs tgt locn; 1: shift amt */

    uint8    ffe;                /* 0: suppress; 1: fire for effect */

    uint8    eom;                /* 1: end of mission */

    uint8    _unused_4;          /* padding */
    int16            leftRightShift;     /* shift left (-) or right (+) */

    int16            addDropShift;       /* shift add (+) or drop (-) */

    IVIS_Location tgtLocation;        /* absolute target location */

    cvcc_ObjID    cff;                /* cff rpt this rpt is associated with */

} AdjustFireReport; 


/*
 *  Ammunition status report.
 *  This report has gone away
 *
 *  Call for fire report.
 *  8 + 8 + 1 + 7 = 24 bytes
 */

#define cvcc_ConcNoSize         7
typedef struct {
    IVIS_Location targetLoc;          /* location of target */

    IVIS_Location observeLoc;         /* location of observer */

    IVIS_ObjectType what;             /* type of target */

    char             concNo[cvcc_ConcNoSize];  /* concentration number */

    uint32    size;               /* size of sighting */

} CallForFireReport; 
 /*  */
 /*  Contact report. */
 /*  2 * (8 + 2 + 1 + 1) = 24 bytes */
typedef struct {
    IVIS_Location location;           /* enemy location */

    uint16   enemyHdg;           /* enemy heading */

    IVIS_ObjectType what;             /* type of units spotted */

    uint8    _unused_5;          /* padding */
} ContactReportItem; 

#define maxContactReportItems   2
typedef struct {
    ContactReportItem item[maxContactReportItems]; 
} ContactReport; 


/*
 *  Free Text Report
 *  512 bytes
 */

#define maxFreeTextSize         512 /* max length, including null */
typedef struct {
    char text[maxFreeTextSize];       /* text paragraph */

} FreeTextReport; 


/*
 *  Intelligence report.
 *  2 * (8 + 2 + 1 + 1 + 2 + 2) + (1 + 3 + 8 + 8) + 4 = 56 bytes
 */

typedef struct {
    IVIS_Location location;           /* location of unit */

    uint16   number;             /* size of unit */

    IVIS_ObjectType what;             /* type of unit */

    IVIS_ActivityType activity;       /* unit's activity */

    uint16   heading;            /* unit's heading, in degrees */

    uint16   _unused_6;          /* padding */
} IntelReportItem; 

typedef struct {
    IVIS_ObjectType what;             /* type of obstacle */

    unsigned         _unused_7  : 24;    /* padding */
    IVIS_Location location1;          /* one end of obstacle */

    IVIS_Location location2;          /* other end of obstacle */

} ObstacleItem; 

typedef struct {
    IntelReportItem enemy;            /* enemy unit */

    IntelReportItem friend;           /* friendly unit */

    ObstacleItem    obstacle;         /* obstacle */

    SP_Time            time;             /* time of reported status */

} IntelReport; 


/*
 *  Nuclear, biological, or chemical attack report.
 *  8 + 8 + 1 + 1 + 2 + 2 + 2 + 2 + 2 = 28 bytes
 */

typedef struct {
    IVIS_Location obsLocation;        /* location of observer (probably) */

    IVIS_Location atkLocation;        /* location of attack */

    IVIS_BurstType burstType;         /* air, surface, or unknown burst */

    IVIS_ObjectType attackType;       /* nuclear, biological, or chemical */

    uint16   flashToBangTime;    /* flash to bang time in seconds */

    uint16   numberOfShells;     /* number of shells observed */

    uint16   nuclearCraterSize;  /* size in meters, of nuclear crater */

    uint16   nuclearCloudWidth;  /* angle in degrees, of cloud width */

    uint16   nuclearCloudHeight;  /* angle in degrees, of cloud height */

} NBCReport; 


/*
 *  Shell report.
 *  8 + 4 + 2 + 2 = 16 bytes
 */

typedef struct {
    IVIS_Location location;           /* location of shelling */

    SP_Time          time;               /* time of shelling */

    uint16   number;             /* number of shells */

    uint16   _unused_8;          /* padding */
} ShellReport; 


/*
 *  Situation report.
 *  8 + 8 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 4 = 28 bytes
 */

typedef struct {
    IVIS_Location startFLOT;          /* endpoints of FLOT */

    IVIS_Location endFLOT;           
    IVIS_ActivityLevel enActLevel;    /* enemy activity level */

    IVIS_ActivityType enActType;      /* enemy activity type */

    IVIS_ActivityType ownIntent;      /* cmdr's intent */

    uint8    _unused_9;          /* padding */
    uint8    int16Ammo;          /* 1: critical ammunition int16age */

    uint8    int16Equip;         /* 1: critical equipment int16age */

    uint8    int16Fuel;          /* 1: critical fuel int16age */

    uint8    int16Pers;          /* 1: critical personnel int16age */

    SP_Time          time;              
} SituationReport; 


/*
 *  Spot report.
 *  8 + 2*(2 + 2 + 2 + 1 + 1) + 2 + 1 + 1 + 4 = 32 bytes
 */

typedef struct {
    uint16 number;               /* number of units spotted */

    uint16 damaged;              /* number of those damaged */

    uint16 destroyed;            /* number of those destroyed */

    IVIS_ObjectType what;             /* type of units spotted */

    uint8  _unused_10;           /* padding */
} SpotReportItem; 

#define maxSpotReportItems      2
typedef struct {
    IVIS_Location  location;          /* location of the spotted item */

    SpotReportItem item[maxSpotReportItems]; 
    uint16    enHeading;         /* enemy heading */

    IVIS_ActivityType enActType;      /* enemy activity type */

    IVIS_ActivityType ownActType;     /* own activity type */

    SP_Time           time;              /* time at which spotted */

} SpotReport; 
 /*  */
 /* A cvcc_Report data element represents the contents of a single report */
 /* of any type, which is a union of the different reports. */
 /* 1 + 3 + max(28,24,24,512,56,28,16,28,32) = 516 bytes */
typedef struct {
    cvcc_ReportType rtype;           
    unsigned      _unused_11 : 24;       /* padding */
     /* Information specific to individual types of reports: */

    union {
        AdjustFireReport  adjust;
        CallForFireReport cff;
        ContactReport     contact;
        FreeTextReport    freeText;
        IntelReport       intel;
        NBCReport         nbc;
        ShellReport       shell;
        SituationReport   sitRep;
        SpotReport        spot;
    } u;                                
} cvcc_Report; 



#endif
