
/*
 * FILE:  mos.h 
 *
 * Created on Wed Apr 17 10:52:49 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _mos_included
#define _mos_included



/*
 *  File: mos.drn
 *
 *  This file contains definitions for communicating military operational
 *  symbology from BnTOC to IVIS, IVIS to IVIS, and (in the future?) BnTOC
 *  to SAF.  It is wrapped by the IVIS message protocol.  wrapped in the
 *
 *  Copyright (c) 1990, 1991.  All rights reserved.
 *  BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 *  History:
 *    12/10/91  Ross		overhaul
 *    09/12/90  Schwartz	new file
 *
 * $RCSfile: mos.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <cvcc_types.h> 
     /*  mos_WCoord defines an overlay X,Y position in world coordinates, usually */
 /*  associated with a terrain database.  It is used primarily for object */
 /*  locations. */
typedef struct {
    float32 x;                            
    float32 y;                            
} mos_WCoord; 
 /* 8 bytes */
 /*  mos_DCoord defines an X,Y position or offset in device coordinates */
 /*  (usually a screen.)  It is used primarily to specify screen offsets from */
 /*  world coordinates. */
typedef struct {
    int32 x;                             
    int32 y;                             
} mos_DCoord; 
 /* 8 bytes */
 /*  mos_Position is another means of specifying location; it can indicate */
 /*  where one object is placed relative to another, or where a specific part */
 /*  of an object is drawn in relation to the overall object.  This has */
 /*  several potential uses; it is currently used only to place tags on point */
 /*  symbols. */
#define mos_TopLeft             0    
#define mos_TopCenter           1    
#define mos_TopRight            2    
#define mos_Left                3    
#define mos_Center              4    
#define mos_Right               5    
#define mos_BottomLeft          6    
#define mos_BottomCenter        7    
#define mos_BottomRight         8    

typedef uint16 mos_Position; 
 /*  mos_AlignmentType should be self-explanatory. */
#define mos_Friend              0    
#define mos_Foe                 1    

typedef uint8 mos_AlignmentType; 
 /*  mos_PropConfType is used for planning purposes; a given overlay object */
 /*  may be either proposed (e.g., at the IPB stage), or confirmed (e.g., a */
 /*  known enemy platoon location.) */
#define mos_Proposed            0    
#define mos_Confirmed           1    

typedef uint8 mos_PropConfType; 
 /*  mos_String is used to contain a fairly large amount of text (often */
 /*  human-readable) for a variety of purposes, including label objects and */
 /*  overlay remarks. */
#define mos_MaxStringLength     512
typedef struct {
    uint16 length;              
    uint16 _unused_4;            /* padding */
    char           text[mos_MaxStringLength]; 
} mos_String; 
 /* 516 bytes */
 /*  mos_Tag is used to contain a small amount of text, associated with an */
 /*  object as an attribute.  If appropriate for the object, the 'where' */
 /*  field may be used to position the tag for graphical output.  Please */
 /*  note that a tag must have a null terminator.  A tag can really only */
 /*  have 31 alphanumeric characters. */
#define mos_MaxTagLength        32
typedef struct {
    char          tag[mos_MaxTagLength]; 
    mos_Position where;              
    uint16 _unused_5;            /* padding */
} mos_Tag; 
 /* 32 + 2 + 2 = 36 bytes */
 /*  mos_OverlayID is a unique identifier for an overlay, combining the */
 /*  originating machine's sim address and a sequence number unique to this */
 /*  exercise. */

 /*  mos_ObjectID uniquely identifies an object within an overlay. */
typedef struct {
    SimulationAddress machineID;     
    uint32        overlayID;     
} mos_OverlayID; 
 /* 4 + 4 = 8 bytes */
typedef struct {
    mos_OverlayID overlayID;         
    uint32    objectID;          
} mos_ObjectID; 
 /* 8 + 4 = 12 bytes */
 /*  mos_PointSymbol defines an overlay object which occupies a single point */
 /*  on the terrain.  It may be real (e.g., a vehicle, howitzer, building, */
 /*  etc) or conceptual (contact point, target reference, etc.)  For detailed */
 /*  information, and to support visual semantics defined by standards such */
 /*  as FM-101-5-1, tags may be placed around it. */
#define mos_MaxPointTags        8
typedef struct {
    ObjectType  oType;                /* tank, contact point, et al. */

    mos_WCoord  location;            
    UnitType    uType;                /* platoon, company, et al. */

    uint8  _unused_6;            /* padding */
    uint16 nTags;               
    mos_Tag     tags[mos_MaxPointTags]; 
} mos_PointSymbol; 
 /* 4 + 8 + 1 + 1 + 2 + 8*36 = 304 bytes */
 /*  mos_PolyType enumerates the types of linear objects that may appear on */
 /*  an overlay.  At present, there are no "real-world" linear objects */
 /*  analogous to point objects like vehicles; all lines are conceptual. */
#define mos_GenericLine         0    
#define mos_GenericArea         1    
#define mos_DirectionOfAdvance  2    
#define mos_MainDirectionOfAdvance 3 
#define mos_AxisOfAdvance       4    
#define mos_MainAxisOfAdvance   5    
#define mos_BoundaryLine        6    
#define mos_PhaseLine           7    
#define mos_RectangularTarget   8    
#define mos_LinearTarget        9    
#define mos_LinearAbatis        10   
#define mos_UnspecifiedMinefield 11  
#define mos_AntitankMinefield   12   
#define mos_AntipersonnelMinefield 13
#define mos_TankDitch           14   
#define mos_RestrictedFireLine  15   
#define mos_RestrictedFireArea  16   
#define mos_NoFireLine          17   
#define mos_NoFireArea          18   
#define mos_FLOT                19   
#define mos_CFL                 20   

typedef uint16 mos_PolyType; 
 /*  mos_Polyline defines a linear object on an overlay.  thickness is */
 /*  obviously used for visual processing only.  sizes are */
 /*  the width in kilometers of the head and tail of wide polylines such as */
 /*  minefields and axes of advance; for polylines such as minefields, which */
 /*  have a constant width aint32 their length, these two attributes are */
 /*  equal.  unitSize allows a line to be associated with a particular level */
 /*  of command; unitSizeIndex indicates, for graphical representation, at */
 /*  what point index the unit size should be displayed. */
#define mos_MaxPolyCoords       50
typedef struct {
    mos_PolyType ptype;              
    char           thickness;           
    UnitType    unitSize;            
    float32          sizes[3];            
    uint16 unitSizeIndex;       
    uint16 nCoords;             
    unsigned       spline         : 1;   /* Boolean */
    unsigned       _unused_7      : 31;  /* padding */
    mos_WCoord  coords[mos_MaxPolyCoords]; 
} mos_Polyline; 
 /* 2 + 1 + 1 + 4 + 4 + 4 + 2 + 2 + 1/8 + 3-7/8 + 50*8 = 424 bytes */
 /*  mos_TextType enumerates the kinds of single- or multi-line text objects */
 /*  that may appear on an overlay. */

 /*  mos_Text may be used to represent either text placed by itself or a */
 /*  free-float32ing label linked to another object.  It contains both a world */
 /*  coordinate position and an optional device coordinate offset from that */
 /*  position (for added display flexibility when a text object is linked to */
 /*  another object.) */
#define mos_PolyLabelText       0    
#define mos_FreeText            1    
#define mos_RemarkText          2    

typedef uint16 mos_TextType; 

typedef struct {
    mos_TextType ttype;              
    uint16 _unused_8;            /* padding */
    mos_WCoord location;             
    mos_DCoord offset;               
    mos_String text;                 
} mos_Text; 
 /* 2 + 2 + 8 + 8 + 516 = 536 bytes */
 /*  mos_DBObject permits binding to an overlay of an object which does not */
 /*  necessarily have a display representation, but which is present in a */
 /*  CVCC database (e.g., an IVIS message linked to a point symbol.)  A */
 /*  mos_DBObject will always be "linked" to another overlay object; see */
 /*  mos_Object, below. */
typedef struct {
    cvcc_ObjID objID;                
} mos_DBObject; 
 /* 12 bytes */
 /*  Enumerate the above four object types for encapsulation purposes. */
#define mos_OPointSymbol        0    
#define mos_OPolyline           1    
#define mos_OText               2    
#define mos_ODBObject           3    

typedef uint16 mos_ObjectType; 
 /*  mos_OverlayObject encapsulates the above individual object types, and */
 /*  adds the type-independent information. */

 /*  mos_LinkType enumerates the types of links that may be established */
 /*  between one object and another; the comment for each type describes the */
 /*  type of "linked-from" and "linked-to" objects valid for that link type. */

 /*  mos_LinkTuple defines one link.  The vertex field is used when the */
 /*  linked-to object is a mos_Polyline; it defines the point on the polyline */
 /*  to which the linked-from object is bound. */

 /*  Note that the link information is not required to _display_ a received */
 /*  overlay, but is required to _change one.  If only display is required, */
 /*  the receiver of a mos_Object may ignore the link fields, even if nLinks */
 /*  is nonzero. */
#define mos_PolyLink            0     /*  from a mos object to mos_Polyline */
#define mos_UnitParentLink      1     /*  from mos_PointSymbol to mos_PointSymbol */
typedef uint16 mos_LinkType; 

typedef struct {
    mos_LinkType ltype;              
    uint16  vertex;             
    mos_ObjectID other;              
} mos_LinkTuple; 
 /* 2 + 2 + 12 = 16 bytes */
#define mos_MaxLinks            4
typedef struct {
    mos_ObjectType otype;            
    mos_AlignmentType alignment;     
    mos_PropConfType propConf;       
    mos_ObjectID  objID;             
    uint16   nLinks;            
    unsigned         visible    : 1;     /* Boolean */
    unsigned         _unused_9  : 15;    /* padding */
    mos_LinkTuple links[mos_MaxLinks]; 
    union {
        mos_PointSymbol   psym;
        mos_Polyline      poly;
        mos_Text          text;
        mos_DBObject      dbObj;
    } u;                                
} mos_Object; 
 /* 2 + 1 + 1 + 12 + 2 + 2 + 4*16 + max(304,424,536,12) = 620 bytes */
 /*  mos_Overlay defines an overlay, a collection of mos_Objects each of */
 /*  which refer to it by ID.  The receiver is free to do anything it wants */
 /*  with the remarks field. */
typedef struct {
    mos_OverlayID id;                
    mos_Tag       name;              
    mos_String    remarks;           
} mos_Overlay; 
 /* 8 + 36 + 516 = 560 bytes */
 /*  mos_ControlType enumerates the types of control messages which may be */
 /*  sent for overlays.  A mos_CtlBeginOverlay is sent before the objects on */
 /*  the overlay, followed by a mos_CtlAddObject for each object, followed by */
 /*  a mos_CtlEndOverlay.  Reliable sequencing and delivery is assumed.  For */
 /*  display purposes, the bottom-to-top object order is established by the */
 /*  sending order. */

 /*  mos_OverlayPacket, as expected, encapsulates control and object */
 /*  information. */
#define mos_CtlBeginOverlay     0    
#define mos_CtlEndOverlay       1    
#define mos_CtlAddObject        2    

typedef uint16 mos_ControlType; 

typedef struct {
    mos_ControlType ctype;           
    uint16 _unused_10;           /* padding */
    union {
        mos_Overlay    ovl;
         /* 		when (mos_CtlEndOverlay)	ovl	mos_Overlay, */
        mos_Object     obj;
    } u;                                
} mos_OverlayPacket; 
 /* 2 + 2 + max(560,624) = 628 bytes */


#endif
