/*
 * $RCSfile: p_arty.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */

/****************************************************************************
*   File: p_arty.h                                                          *
*                                                                           *
*       Copyright 1994 by Technical Solutions Inc.                          *
*                                                                           *
*               Technical Solutions, Inc.                                   *
*               Highway 478                                                 *
*               Mesilla Park 88047                                          *
*               1-505-524-2154                                              *
*                                                                           *
*       This software was developed by TSI under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Artillery radio message format                            *
*       Created: Wed March 31 10:45:23 MST 1994                             *
*       Author: sst                                                         *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#define artyNotGiven 			0

#define artyCoordinatesTcc 		0
#define artyCoordinatesPolar 		1
#define artyCoordinatesShift   		2

typedef uint8		artyCoordinatesType;

#define artyMessageAdjust 		1
#define artyMessageFireRequest       	2
#define artyMessageSuppress	 	3
#define artyMessageSuppressNow  	4
#define artyMessageRegistration 	5
#define artyMessageIntelligence 	6
#define artyMessageInformation 		7
#define artyMessageFireOrder  		8
#define artyMessageToObserver		9

typedef uint8		artyMessageType;
typedef uint32		artyTime;
typedef uint32		artyVehicleID;

typedef struct {
    artyCoordinatesType      coordinate_system;
    artyMessageType          message_type;
    char		     target_number[6];
    artyVehicleID   	     originator;
    artyVehicleID   	     receiver;
} artyMessageInfo;

typedef struct {
    union {
        struct {
            float64          tcc_x;
            float64          tcc_y;
	} tcc;
        struct {
	    uint16	     OT_direction;
	    int16	     lateral_shift;
 	    int16	     range_shift;
	    int16	     vertical_shift;
 	    uint8	     known_point_num;
	    uint8	     _unused_xx[7];      /* padding */
	} shift;
        struct {
	    uint16	     OT_direction;
	    uint16	     distance;
	    int16	     vertical_shift;
	    uint8	     _unused_zz [2];
	    uint8	     _unused_xx [8];     /* padding */
	} polar;
    } location;
} artyTargetLocation;

typedef struct {
    uint16	 	     height;
    uint16	 	     width;
    uint16	 	     length;
    uint16	 	     radius;
    uint16 		     orientation;
} artyPolygonDescriptor;

#define artyTargetADA		1
#define artyTargetARMOR		2
#define artyTargetARTY		3
#define artyTargetASSY		4
#define artyTargetBLDG		5
#define artyTargetBRIDGE	6
#define artyTargetCEN		7
#define artyTargetEQUIP		8
#define artyTargetMORT		9
#define artyTargetPERS		10
#define artyTargetRKTMSL	11
#define artyTargetSPEC		12
#define artyTargetSUPPLY	13 
#define artyTargetTER		14
#define artyTargetVEH		15
#define artyTargetWPN		16

typedef uint8		artyTargetType;

#define artyTargetSubtypeUNK		0
#define artyTargetSubtypeLT		1
#define artyTargetSubtypeMDM		2
#define artyTargetSubtypeMSL		3
#define artyTargetSubtypePOS		4
#define artyTargetSubtypeAPC		5
#define artyTargetSubtypeHV		6
#define artyTargetSubtypeTRP		7
#define artyTargetSubtypeTRPVEH		8
#define artyTargetSubtypeTRPMEC		9
#define artyTargetSubtypeTRPARM		10
#define artyTargetSubtypeWOOD		11
#define artyTargetSubtypeMASNRY		12
#define artyTargetSubtypeCONC		13
#define artyTargetSubtypeMET		14
#define artyTargetSubtypeSPCL		15
#define artyTargetSubtypeFTPON		16
#define artyTargetSubtypeVEHPON		17
#define artyTargetSubtypeSTEEL		18
#define artyTargetSubtypeSITE		19
#define artyTargetSubtypeRAFT		20
#define artyTargetSubtypeFERRY		21
#define artyTargetSubtypeSMALL		22
#define artyTargetSubtypeBN		23
#define artyTargetSubtypeREGT		24
#define artyTargetSubtypeDIV		25
#define artyTargetSubtypeFWD		26
#define artyTargetSubtypeRADAR		27
#define artyTargetSubtypeEW		28
#define artyTargetSubtypeGDNC		29
#define artyTargetSubtypeLS		30
#define artyTargetSubtypeVH		31
#define artyTargetSubtypeINF		32
#define artyTargetSubtypeOP		33
#define artyTargetSubtypePTL		34
#define artyTargetSubtypeWKPTY		35
#define artyTargetSubtypeAPERS		36
#define artyTargetSubtypeLTMSL		37
#define artyTargetSubtypeMDMMSL		38
#define artyTargetSubtypeHVMSL		39
#define artyTargetSubtypeATANK		40
#define artyTargetSubtypeONCALL		41
#define artyTargetSubtypeILL1		42
#define artyTargetSubtypeILL2		43
#define artyTargetSubtypeILL2DS		44
#define artyTargetSubtypeILL2RG		45
#define artyTargetSubtypeILL4		46
#define artyTargetSubtypeGASNON		47
#define artyTargetSubtypeGASPER		48
#define artyTargetSubtypeLEAF		49
#define artyTargetSubtypeAMMO		50
#define artyTargetSubtypeBRGEQ		51
#define artyTargetSubtypeCLI		52
#define artyTargetSubtypeCLII		53
#define artyTargetSubtypeROAD		54
#define artyTargetSubtypeJCT		55
#define artyTargetSubtypeHILL		56
#define artyTargetSubtypeDEFILE		57
#define artyTargetSubtypeLDGSTR		58
#define artyTargetSubtypeRR		59
#define artyTargetSubtypeLTWHL		60
#define artyTargetSubtypeHVWHL		61
#define artyTargetSubtypeRECON		62
#define artyTargetSubtypeBT		63
#define artyTargetSubtypeACFT		64
#define artyTargetSubtypeHEL		65
#define artyTargetSubtypeLTMG		66
#define artyTargetSubtypeATG		67
#define artyTargetSubtypeHVMG		68
#define artyTargetSubtypeRCLR		69

typedef uint8		artyTargetSubtype;

#define artyTargetProtectionPRAND		1
#define artyTargetProtectionPRONE		2
#define artyTargetProtectionPRUG 		3
#define artyTargetProtectionPROVER		4
#define artyTargetProtectionDUGIN		5
#define artyTargetProtectionCOVER		6

typedef uint8		artyTargetProtection;

typedef struct {
    artyTargetType	     type;
    artyTargetSubtype        subtype;
    uint8		     activity;
    uint16		     elements;
    artyTargetProtection     protection;
    artyPolygonDescriptor    description;
    uint8		     final_protective_fire;
} artyTacfireTargetInfo;

#define artyAtMyCommand  		1
#define artyCancelAtMyCommand   	2
#define artyFire		  	3
#define artyAtMyCommandRepeat		4
#define artyFireAtTime                  5
#define artyCheckFiring                 6
#define artyCeaseLoading                7
#define artyContinuousFire              8
#define artyRepeat                      9
#define artyEndOfMission		10
#define artyFireMission			11
#define artyFireForEffect		12

typedef uint8		     artyFCCommand;

#define artyContIllumination		1
#define artyCoordIllumination		2

typedef uint8		     artyIllumination;

typedef struct {
    uint8		     platoon_left;
    uint8		     firing_interval;
    artyFCCommand	     command;
    uint8		     cant_observe;
    artyTime		     time_on_target;
    artyIllumination	     illumination;
    uint8		     followed_by;
} artyFCMethod;

#define artyPowderWhite			1
#define artyPowderGreen			2
#define artyPowderRed			3

typedef uint8		     artyPowderType;

typedef struct {
    uint32		     warhead;
    uint32		     fuze;
    uint16		     quantity;
    artyPowderType	     powder_type;
    uint8		     powder_amount;
    uint16		     quad_elevation;
    uint16		     deflection_angle;
    uint16		     flight_time;
} artyTechfireInfo;


typedef struct {
    artyMessageInfo 	     message_info;
    uint8    		     fire_element;
    artyTargetLocation	     location;
    artyTacfireTargetInfo    target;
    uint8		     engagement_method;
    artyPolygonDescriptor    distribution_params;
    artyFCMethod	     fc_method;
    artyTechfireInfo	     techfire_info;
    uint32		     _unused_xx   : 32;  /* future expansion */
} artyRadioData;

