/*
 * $RCSfile: p_cac2_ivis.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: p_cac2_ivis.h						    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Network definitions for libivisi			    *
*       Created: October 30, 1994					    *
*       Author: pmonday							    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* This header file is intended to be a readable version of the actual
 * PDU definitions used by LibIvis.  The definitions are not necessarily
 * aligned.
 */

#ifndef _CAC2_IVIS_HEADER_
#define _CAC2_IVIS_HEADER_


#include <stdtypes.h>
#include <p_dis203.h>

/***** Definitions *****/

#define DIS_IVIS_PROTOCOL 1

/* Ivis vehicle style enumeration */
typedef enum
{
   	CAC2_M1A2IVIS = 1,
	CAC2_A2ATDCAC2 = 6
} CAC2_IVIS_VEHICLE_STYLE_TYPE;

/* Software Version */
typedef enum
{
        CAC2_SMI_VERSION_2_0 = 1,
        CAC2_SMI_VERSION_2_1_1,
        CAC2_SMI_VERSION_2_2,
        CAC2_SMI_VERSION_2_3,
        CAC2_SMI_VERSION_2_3a
} CAC2_SOFTWARE_VERSION_TYPE;

/* ivis_message_type */
typedef enum
{
        CAC2_REPORT_ACK = 0,                 /*  NOT A 1553 MESSAGE TYPE  */
	CAC2_CONTACT_REPORT		= 1,
	CAC2_CALL_FOR_FIRE_REPORT,
	CAC2_SPOT_REPORT,
	CAC2_SITUATION_REPORT,
        CAC2_MEDEVAC_AIR_REPORT,
        CAC2_MEDEVAC_GROUND_REPORT,
	CAC2_ENEMY_OVERLAY		= 8,
	CAC2_FIRE_SUPPORT_OVERLAY,
	CAC2_OBSTACLE_OVERLAY,
	CAC2_HIGHER_ECHELONS_OPERATIONS_OVERLAY_1,
	CAC2_ENEMY_OVERLAY_UPDATE,
	CAC2_FIRE_SUPPORT_OVERLAY_UPDATE,
	CAC2_OBSTACLE_OVERLAY_UPDATE,
	CAC2_OPERATIONS_OVERLAY_1_UPDATE,
	CAC2_ID_REPORT,			/* TEU <-> RIU only */
	CAC2_POSITION_REPORT,
	CAC2_HIGHER_ECHELONS_OPERATIONS_OVERLAY_2,
	CAC2_OPERATIONS_OVERLAY_2_UPDATE,
	CAC2_OWN_OPERATIONS_OVERLAY_1,
	CAC2_OWN_OPERATIONS_OVERLAY_2,
	CAC2_ADJUST_FIRE_REPORT,
	CAC2_DMD_FREETEXT_REPORT,
	CAC2_INTELLIGENCE_REPORT,
	CAC2_LOGISTICS_REPORT,
	CAC2_NBC_REPORT,
	CAC2_SHELL_REPORT
} CAC2_MSG_TYPE;

/* These are the different modes of the SINCGARS radio.*/
typedef enum
{
	CAC2_SINGLE_CHANNEL	= 0,
	CAC2_FREQUENCY_HOPPING,
	CAC2_FH_MASTER	= 3
} CAC2_RADIO_MODE;

/* These are the channels on the SINCGARS radios. */
typedef enum
{
	CAC2_MAN		= 0,
	CAC2_CHANNEL_1,
	CAC2_CHANNEL_2,
	CAC2_CHANNEL_3,
	CAC2_CHANNEL_4,
	CAC2_CHANNEL_5,
	CAC2_CHANNEL_6,
	CAC2_CUE		= 7
} CAC2_RADIO_CHANNEL;
 
/* These are the power levels of the SINCGARS radios. */
typedef enum
{
	CAC2_POWER_LOW	= 0,
	CAC2_POWER_AMPLIFIER,
	CAC2_POWER_HIGH,
	CAC2_POWER_MEDIUM	= 4
} CAC2_RADIO_POWER;
 
/* These are the meanings of the squelch command. */
typedef enum
{
	CAC2_SQUELCH_ON	= 0,
	CAC2_SQUELCH_OFF	= 16
} CAC2_RADIO_SQUELCH;
 
/* These are the meanings of the zeroize command. */
typedef enum
{
	CAC2_ZEROIZE_NONE	= 0,
	CAC2_ZEROIZE_ALL
} CAC2_RADIO_ZEROIZE;
 
/* These are the codes for the different COMSEC variables. */
typedef enum
{
	CAC2_VAR5		= 0,
	CAC2_VAR4,
	CAC2_VAR2,
	CAC2_VAR6,
	CAC2_VAR1,
	CAC2_VAR3		= 6
} CAC2_RADIO_COMSEC_VAR;

/* These are the codes for the different COMSEC modes. */
typedef enum
{
	CAC2_PLAIN_TEXT	= 0,
	CAC2_CYPHER_TEXT,
	CAC2_REMOTE_VARIABLE,
	CAC2_MODE_LOW,
	CAC2_MODE_ZERO_ALL	= 6
} CAC2_RADIO_COMSEC_MODE;

/* These are the meanings of the different time delays */
typedef enum
{
	CAC2_TIME_DELAY_ON	= 0,
	CAC2_TIME_DELAY_OFF	= 1
} CAC2_RADIO_TIME_DELAY;
 
/* These are the codes for the different data rates. */
typedef enum
{
	CAC2_RATE_16K	= 0,
	CAC2_RATE_4800,
	CAC2_RATE_2400,
	CAC2_RATE_1200,
	CAC2_RATE_600,
	CAC2_RATE_300,
	CAC2_RATE_150,
	CAC2_RATE_75
} CAC2_RADIO_DATA_RATE;

/* These are the codes for the different communication networks. */
typedef enum
{
	CAC2_PLATOON_NET	= 1,
	CAC2_COMPANY_NET,
	CAC2_BATTALION_NET
} CAC2_RADIO_NETWORK;

/* These are the codes for the radios in the radio select command. */
typedef enum
{
	CAC2_RADIO_A		= 1,
	CAC2_RADIO_B,
	CAC2_BOTH_A_AND_B
} CAC2_RADIO_SELECT;
  
/* These are the codes for the different duty positions. */
typedef enum {
	CAC2_DP_UNDEFINED	= 0,
	CAC2_BN_CMDR		= 1,
	CAC2_S3,
	CAC2_CO_CMDR,
	CAC2_CO_XO,
	CAC2_PLT_LDR,
	CAC2_PLT_SGT,
	CAC2_WM1,
	CAC2_WM2,
	CAC2_TOC,		/* S3 AIR */
	CAC2_CTCP,
	CAC2_BN_XO,
	CAC2_SCT_PLT_LDR,
	CAC2_FSO,
	CAC2_MORT_PLT_LDR,
	CAC2_ENG_CMDR,
	CAC2_ADA_CMDR,
	CAC2_S2,
	CAC2_SPT_PLT_LDR,
	CAC2_MED_PLT_LDR,
	CAC2_CO_1SG,
	CAC2_UMCP,
	CAC2_BMO,
	CAC2_CSM,
	CAC2_FIST	
} CAC2_DUTY_POSITION;
 
typedef enum {
	CAC2_NEW_NODE, 
	CAC2_RESPONSE_TO_NEW_NODE, 
	CAC2_ID_UPDATE, 
	CAC2_EXITING_NET
}  CAC2_ID_REPORT_SUBTYPES;

typedef enum {
	CAC2_OWN_ID = 0,
	CAC2_ID_MESSAGE_UPDATE_ID,
	CAC2_VT_RESPONSE,
	CAC2_CALL_SIGN_UPDATE,
	CAC2_RESPONSE_TO_NEW_NODE_ID
} CAC2_ID_TYPE;

typedef enum {
	CAC2_FT_NONE		= 0,
	CAC2_FT_IMMED_SMOKE,
	CAC2_FT_IMMED_SUPPR,
	CAC2_FT_FIRE_FOR_EFFECT
} CAC2_FIRE_TYPE;

typedef enum {
	CAC2_TT_NONE		= 0,
	CAC2_TT_AIRCRAFT,
	CAC2_TT_ARTILLERY,
	CAC2_TT_APC,
	CAC2_TT_TANK,
	CAC2_TT_OTHER,
	CAC2_TT_AIRCRAFT_FLOGGER,
	CAC2_TT_AIRCRAFT_HAVOC,
	CAC2_TT_AIRCRAFT_HIND,
	CAC2_TT_AIRCRAFT_HIP,
	CAC2_TT_ARTILLERY_SELF_PROP,
	CAC2_TT_ARTILLERY_TOWED,
	CAC2_TT_APC_BMP,
	CAC2_TT_APC_BDRM,
	CAC2_TT_APC_BTR,
	CAC2_TT_TANK_T55,
	CAC2_TT_TANK_T60,
	CAC2_TT_TANK_T72,
	CAC2_TT_TANK_T80,
	CAC2_TT_TANK_NATO,
	CAC2_TT_TANK_FST1,
	CAC2_TT_TANK_FST2,
	CAC2_TT_OTHER_BUNKER,
	CAC2_TT_OTHER_ZSU23,
	CAC2_TT_OTHER_SA9_13,
	CAC2_TT_OTHER_INFANTRY = 26
} CAC2_THREAT_TYPE;

typedef enum {
	CAC2_ACT_NONE 	= 0,
	CAC2_ACT_ATTACKING,
	CAC2_ACT_DEFENDING,
	CAC2_ACT_RECONING,
	CAC2_ACT_WITHDRAWING
} CAC2_ACTIVITY;

typedef enum {
	CAC2_FA_NONE		= 0,
	CAC2_FA_CONTINUE,
	CAC2_FA_OBSERVE
} CAC2_FRIENDLY_ACTION;

typedef enum {
	CAC2_EA_NONE		= 0,
	CAC2_EA_DEFENDING,
	CAC2_EA_ATTACKING,
	CAC2_EA_WITHDRAWING,
	CAC2_EA_SCREENING,
	CAC2_EA_RECONNING
} CAC2_ENEMY_ACTIVITY;

typedef enum {
	CAC2_FUEL_NO_SELECTION	= 0,
	CAC2_FUEL_RED,
	CAC2_FUEL_GREEN,
	CAC2_FUEL_AMBER,
	CAC2_FUEL_BLACK
} CAC2_FUEL_TYPE;

typedef enum {
	CAC2_TACT_NONE 	= 0,
	CAC2_TACT_DEFEND,
	CAC2_TACT_ATTACK,
	CAC2_TACT_WITHDRAW,
	CAC2_TACT_RELIEF,
	CAC2_TACT_MOVE
} CAC2_TACTICAL_TYPE;

typedef enum {
	CAC2_ANTI_TANK_DITCH	= 7,
	CAC2_ANTI_TANK_WEAPON,
	CAC2_APC,
	CAC2_ANTI_PERS_MINE,
	CAC2_ARMOR_BATTALION,
	CAC2_ARMOR_COMPANY,
	CAC2_ARMOR_PLATOON,
	CAC2_ARMOR_TASK_FORCE,
	CAC2_ARMOR_TEAM,
	CAC2_ARMOR_CALV_PLATOON,
	CAC2_ARMOR_CALV_SQUADRON,
	CAC2_ARMOR_CALV_TROOP,
	CAC2_BATLN_BATTLE_POS,
	CAC2_COMPANY_BATTLE_POS,
	CAC2_PLATOON_BATTLE_POS,
	CAC2_BRIDGE,
	CAC2_CONTACT_POINT,
	CAC2_COORDINATING_POINT,
	CAC2_CRITICAL_POINT,
	CAC2_ENGAGEMENT_AREA,
	CAC2_FREE_DRAW,
	CAC2_FRONT_LINE,
	CAC2_HEADQUARTERS,
	CAC2_LANE,
	CAC2_LINEAR_CONCENTRATION,
	CAC2_MECH_INFANTRY_BATLN,
	CAC2_MECH_INFANTRY_COMPANY,
	CAC2_MECH_INFANTRY_PLATOON,
	CAC2_MECH_INFANTRY_TASK_FORCE,
	CAC2_MECH_INFANTRY_TEAM,
	CAC2_MINEFIELD,
	CAC2_OBSERVATION_POINT,
	CAC2_PASSAGE_POINT,
	CAC2_POINT_OBSTACLE,
	CAC2_RELEASE_POINT,
	CAC2_START_POINT,
	CAC2_TANK,
	CAC2_TARGET_REFERENCE_POINT,
	CAC2_UNSPECIFIED_OBSTACLE,
	CAC2_CONTINUATION,
	CAC2_CROSS_HAIR,
	CAC2_FRIENDLY_TANK,
	CAC2_ENEMY_THREAT
}  CAC2_GRAPHIC_TYPE;

typedef enum {
	CAC2_ADD_ICON = 1,
	CAC2_DELETE_ICON = 3
} CAC2_ICON_STATUS;


typedef enum {
        CAC2_INTERNATIONAL_0 = 0,
        CAC2_INTERNATIONAL_10,
        CAC2_CLARK_1866_0,
        CAC2_CLARK_1866_10,
        CAC2_CLARK_1880,
        CAC2_EVEREST_0,
        CAC2_EVEREST_10,
        CAC2_BESSEL_0,
        CAC2_BESSEL_10,
        CAC2_AUSTRALIAN,
        CAC2_WGS_72,
        CAC2_WGS_84
} CAC2_SPHEROID_NAME_TYPE;

typedef enum {
        CAC2_GREEN = 1,
        CAC2_AMBER,
        CAC2_RED,
        CAC2_BLACK
} CAC2_STATUS_TYPE;

typedef enum {
        CAC2_OT_ABATI = 1,
        CAC2_OT_BLOWN_BRIDGE,
        CAC2_OT_MINEFIELD,
        CAC2_OT_TANK_DITCH
} CAC2_OBSTACLE_TYPE;

typedef enum {
        CAC2_NBC_AIR,
        CAC2_NBC_SURFACE
} CAC2_BURST_TYPE;

typedef enum {
        CAC2_NUCLEAR,
        CAC2_BIOLOGICAL,
        CAC2_CHEMICAL
} CAC2_ATTACK_TYPE;

#define CAC2_CALL_SIGN_LENGTH 6
#define CAC2_TANK_ID_LENGTH 6
#define	CAC2_DMD_MSG_SIZE	396

/***** Basic structures *****/

/*   CAC2_GRID_LOC */
typedef struct {
        int16    _unused;			/* 2 */
	unsigned spheroid	: 8;		/* 1 */
	unsigned grid_zone_nbr1	: 8;		/* 1 */
	unsigned grid_zone_nbr2 : 8;		/* 1 */
	unsigned grid_zone_ltr	: 8;		/* 1 */
	unsigned grid_map_col	: 8;		/* 1 */
	unsigned grid_map_row	: 8;		/* 1 */
	uint16   easting;			/* 2 */
	uint16   northing;      		/* 2 */
} CAC2_GRID_LOC;	       			/* 12 */

/* CAC2_PICKUP_DATA_TYPE */
typedef struct {
        unsigned special_equipment : 4;
        unsigned nationality       : 5;
        unsigned contamination     : 4;
        unsigned security          : 3;
} CAC2_PICKUP_DATA_TYPE;

/*   CAC2_RECEIVER_LIST_STRUCT */
typedef struct {
    DIS203_ENTITY_ID_RECORD  receiver_id;	/* 3 x 2 = 6 */
    uint16                  _unused_1;		/* 2 */
} CAC2_RECEIVER_LIST_STRUCT;


#define CAC2_IVIS_MAX_RECEIVERS 30

/****** Header structures *****/

/*   CAC2_MSG_HEADER_STRUCT */
typedef struct {
    uint16           ivis_vehicle_style;	/* 2 */
    uint16           veh_software_version;	/* 2 */
    uint16           unique_report_id;		/* 2 */
    unsigned         ivis_message_type     : 8;	/* 1 */
    unsigned         receiver_count        : 8;	/* 1 */
} CAC2_MSG_HEADER_HDR_STRUCT;

typedef struct {
    CAC2_MSG_HEADER_HDR_STRUCT    header_hdr;		/* 8 */
    CAC2_RECEIVER_LIST_STRUCT     receiver_list[CAC2_IVIS_MAX_RECEIVERS];
} CAC2_MSG_HEADER_STRUCT;					/* 248 */

/*   CAC2_PDU_REPORT_HEADER_STRUCT */  
typedef struct {
	char     call_sign[CAC2_CALL_SIGN_LENGTH];	/* 6 */
	unsigned date		: 8;			/* 1 */
	unsigned time_zone	: 8;	 		/* 1 */
	uint32   time;			 		/* 4 */
} CAC2_PDU_REPORT_HEADER_STRUCT;			/* 12 */


/****** Messages ******/

/*   CAC2_PDU_ID_REPORT_STRUCT */ 
typedef struct {
    char     tank_id[CAC2_TANK_ID_LENGTH];
    unsigned duty_position : 8;
    unsigned platoon       : 8;
    uint8    company;
    uint16   division;
    uint16   brigade;
    uint16   regiment;
    uint16   battalion;
    unsigned radio_network_a  : 8;
    unsigned radio_network_b  : 8;
    unsigned id_report_type   : 8;
} CAC2_PDU_ID_REPORT_STRUCT;


/*   CAC2_POSITION_MSG */
typedef struct {
	unsigned spheroid	: 8;    /*  WORD 12  */
	unsigned grid_zone_nbr1	: 8;    /*  WORD 12  */
	unsigned grid_zone_nbr2	: 8;	/*  WORD 13  */
	unsigned grid_zone_ltr	: 8;	/*  WORD 13  */
	unsigned grid_map_col	: 8;	/*  WORD 14  */
	unsigned grid_map_row	: 8;	/*  WORD 14  */
	uint16   easting;		/*  WORD 15  */
	uint16   northing;		/*  WORD 16  */
} CAC2_POSITION_MSG;

/* CAC2_NEW_MESSAGE */
typedef struct {
	uint32 report_contents[1];
} CAC2_NEW_MESSAGE;

/*  CAC2_CONTACT_MSG */
typedef struct {
	unsigned fire_type	: 8;	/*  WORD 12  */		/* 1 */
	unsigned spare1		: 8;	/*  WORD 12  */		/* 1 */
	int16    _unused;
        int16    location[6];	/* is really CAC2_GRID_LOC */
	unsigned threat_type	: 8;	/*  WORD 18  */
	unsigned threat_size	: 8;	/*  WORD 18  */
} CAC2_CONTACT_MSG;

/*   CAC2_CALL_FOR_FIRE_MSG */
typedef CAC2_CONTACT_MSG CAC2_CALL_FOR_FIRE_MSG;

/*   CAC2_SPOT_MSG */
typedef struct {
	unsigned fire_type	: 8;	/*  WORD 12  */
	unsigned spare1		: 8;	/*  WORD 12  */
	int16    _unused;
        int16    location[6];	/* is really CAC2_GRID_LOC */
	unsigned threat_type_1	: 8;	/*  WORD 18  */
	unsigned threat_size_1	: 8;	/*  WORD 18  */
	unsigned threat_type_2	: 8;	/*  WORD 19  */
	unsigned threat_size_2	: 8;	/*  WORD 19  */
	unsigned threat_type_3	: 8;	/*  WORD 20  */
	unsigned threat_size_3	: 8;	/*  WORD 20  */
	unsigned threat_type_4	: 8;	/*  WORD 21  */
	unsigned threat_size_4	: 8;	/*  WORD 21  */
	unsigned fr_action	: 8;	/*  WORD 22  */
	unsigned activity	: 8;	/*  WORD 22  */
}  CAC2_SPOT_MSG;

/*  CAC2_PDU_SIT_REPORT_STRUCT */
typedef struct { 
	unsigned enemy_act	: 8;				/* 1 */
	unsigned spare1		: 8;				/* 1 */
        int16    friendlies[6][9];   /* is really CAC2_GRID_LOC */
 	unsigned veh_author	: 8;				/* 1 */
	unsigned veh_on_hand	: 8;				/* 1 */
 	uint16   pers_author;				/* 2 */
  	uint16   pers_on_hand;				/* 2 */
 	uint16   sabot;					/* 2 */
 	uint16   heat;					/* 2 */
	uint16   mpat;					/* 2 */
        uint16   staff;					/* 2 */
 	uint16   smk_grenade;				/* 2 */
        uint16   _padding1;				/* 2 */
 	uint32   coax;					/* 4 */
 	uint32   cal50;					/* 4 */
 	unsigned fuel		: 8;				/* 1 */
	unsigned tactical	: 8; 				/* 1 */
} CAC2_PDU_SIT_REPORT_STRUCT;

/*   CAC2_OVERLAY_HEADER_STRUCT */
typedef struct {
    unsigned pdu_count           : 8;
    unsigned pdu_graphic_count   : 8;
    uint16   graphic_count;
} CAC2_OVERLAY_HEADER_STRUCT;

/*   CAC2_PDU_GRAPHIC_STRUCT */
typedef struct {
    unsigned 	graphic_id    : 6;
    unsigned 	status        : 2;
    uint8       points;
    uint16      label_1_count;
    uint16      label_2_count;
    char  	labels[20];
    int16 	graphics[6][9];	/* is really CAC2_GRID_LOC */
} CAC2_PDU_GRAPHIC_STRUCT;

/*   CAC2_OVERLAY_STRUCT */
typedef struct {
    CAC2_OVERLAY_HEADER_STRUCT 	overlay_header;
    CAC2_PDU_GRAPHIC_STRUCT 		graphics[50];
} CAC2_OVERLAY_STRUCT;

/* CAC2_MEDEVAC_AIR_MSG  */
typedef struct {
        CAC2_GRID_LOC 	own_tank_location;
        unsigned 	urgent_casualties    : 8;
        unsigned 	priority_casualties  : 8;
        unsigned 	routine_casualties   : 8;
        uint8           radio_frequency[5];    
   	uint8	 	call_sign[5];
        unsigned 	_spare_1             : 8;
        CAC2_PICKUP_DATA_TYPE pickup_data;
        unsigned 	litter               : 8;
        unsigned 	ambulatory           : 8;      
        unsigned 	pickup_site_markings : 16;
        uint8 	        terrain_desc_1[14];
        uint8 	        terrain_desc_2[14];         
        uint8 	        terrain_desc_3[14];
        uint8 	        terrain_desc_4[14];  
} CAC2_MEDEVAC_AIR_MSG;

/* CAC2_MEDEVAC_GROUND_MSG  */
typedef struct {
        CAC2_GRID_LOC 	own_tank_location;
        unsigned 	urgent_casualties    : 8;
        unsigned 	priority_casualties  : 8;
        unsigned 	routine_casualties   : 8;
        uint8 	        radio_frequency[5];
   	uint8 	        call_sign[5];
        unsigned 	_spare_1             : 8;      
} CAC2_MEDEVAC_GROUND_MSG;

/* CAC2_ADJUST FIRE */
typedef struct {
        CAC2_GRID_LOC 	location;
        int16 		add_drop;
        int16 		left_right;
        int16 		up_down;
        unsigned 	ffe : 1;
        unsigned 	eom : 1;
        unsigned 	_spare_1 : 6;
} CAC2_ADJUST_MSG;

/* CAC2_FREE TEXT */
typedef struct {
	char	text[CAC2_DMD_MSG_SIZE];
} CAC2_DMD_FREETEXT_MSG;

/* CAC2_INTELLIGENCE */
typedef struct {
    unsigned      	enemy_type    : 8;
    unsigned      	enemy_number  : 8;
    unsigned      	enemy_act     : 8;
    CAC2_GRID_LOC      	enemy_location;
    int16         	enemy_heading;
    unsigned      	friend_type    : 8;
    unsigned      	friend_number  : 8;
    unsigned      	friend_act     : 8;
    CAC2_GRID_LOC      	friend_location;
    int16         	friend_heading;
    CAC2_OBSTACLE_TYPE 	obstacle_type;
    CAC2_GRID_LOC      	obstacle_point_1;
    CAC2_GRID_LOC      	obstacle_point_2;
} CAC2_INTELLIGENCE_REPORT_STRUCT;

/* CAC2_LOGISTICS */
typedef struct {
    CAC2_STATUS_TYPE 	personnel_status;
    CAC2_STATUS_TYPE 	equipment_status;
    CAC2_STATUS_TYPE 	ammunition_status;
    CAC2_STATUS_TYPE 	fuel_status;
} CAC2_LOGISTICS_REPORT_STRUCT;

/* CAC2_NBC */
typedef struct {
    	CAC2_GRID_LOC   observer_location;
    	CAC2_GRID_LOC   attacker_location;
    	CAC2_BURST_TYPE	burst_type;
	CAC2_ATTACK_TYPE	attack_type;
	int16		flash_to_bang_time;
	int16		number_of_shells;
	int16		crater_diameter;
	int16		nuclear_cloud_width;
	int16		nuclear_cloud_height;
} CAC2_NBC_REPORT_STRUCT;

/* CAC2_SHELL */
typedef struct {
    int16		shell_number;
    CAC2_GRID_LOC    	shell_location;
    int16       	as_of;
} CAC2_SHELL_REPORT_STRUCT;


/****** Signal PDU data structures ******/

/*  CAC2_PDU_IVIS_STRUCT */
typedef struct
{
    CAC2_MSG_HEADER_STRUCT 			msg_header;	/* 248 */
    CAC2_PDU_REPORT_HEADER_STRUCT 		report_header;	/* 12 */
    union {
	CAC2_POSITION_MSG	      		pos_body;        
	CAC2_PDU_ID_REPORT_STRUCT  		id_body;         
	CAC2_NEW_MESSAGE	      		msg_body;   
	CAC2_CONTACT_MSG           		ct_body;         
	CAC2_CALL_FOR_FIRE_MSG     		cff_body;        
	CAC2_SPOT_MSG	      			sp_body;         
	CAC2_PDU_SIT_REPORT_STRUCT 		sit_body;        
	CAC2_OVERLAY_STRUCT	      		overlay;       
        CAC2_MEDEVAC_AIR_MSG       		meda_body;
        CAC2_MEDEVAC_GROUND_MSG    		medg_body;
	CAC2_ADJUST_MSG            		af_body;
	CAC2_DMD_FREETEXT_MSG      		dmd_body;
	CAC2_INTELLIGENCE_REPORT_STRUCT		intell_body;
	CAC2_NBC_REPORT_STRUCT     		nbc_body;
	CAC2_LOGISTICS_REPORT_STRUCT		log_body;
	CAC2_SHELL_REPORT_STRUCT   		shell_body;
    } report_overlay;
} CAC2_PDU_IVIS_STRUCT;

typedef struct
{
  uint32 			checksum;
  DIS203_ENTITY_ID_RECORD 	vehicle_id;
  int16 			radio_id;
  uint8 			data[1];
} CAC2_IVIS_DATA;

#endif

