
/*
 * FILE:  p_data.h 
 *
 * Created on Fri Apr 26 13:42:12 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_data_included
#define _p_data_included



/*
 * File:     pro_data.drn
 * Contents: The Data Collection Protocol Data Units
 * System:   simnet development.
 * Created:  10-Oct-1989
 * Author:   apope
 *
 * Remarks:
 *
 * History:
 *    12/12/89	Pope		aligned Float(64) to 64-bit boundaries
 *    12/13/89  Pope		fixed some padding errors
 *    01/08/90	Pope		added battleScheme to ExerciseStatusVariant
 *    01/12/90	Kanarick	added Jan90 things
 *
 * $RCSfile: p_data.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 * COPYRIGHT 1990 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 */

#include <stdtypes.h>
#include <basic.h> 
#include <status.h> 

/*
 * Versions of the data collection protocol:
 */

#define dcProtocolVersionAug89  1    
#define dcProtocolVersionJan90  2    
#define dcProtocolVersionJan90Corrected 3

typedef uint8    DataCollectionProtocolVersion; 

#define dcProtocolVersionCurrent 3

/*
 * Various kinds of Data Collection PDUs:
 */

#define exerciseStatusPDUKind   1     /* Exercise Status PDU */
#define simulationStatusPDUKind 2     /* Simulation Status PDU */
#define vehicleStatusPDUKind    3     /* Vehicle Status PDU */
#define statusQueryPDUKind      4     /* Status Query PDU */
#define statusResponsePDUKind   5     /* Status Response PDU */
#define statusChangePDUKind     6     /* Status Change PDU */
#define laserRangePDUKind       7     /* Laser Range PDU */
#define eventFlagPDUKind        8     /* Event Flag PDU */
#define ctasPDUKind             129   /* scatoma/glare PDU for CTAS */
#define hetPDUKind              130   /* Head/Eye Tracker PDU for CTAS */
#define losatStatusPDUKind      131   /* Losat Status PDU */
#define losatAidedScanPDUKind   132   /* Losat Aided Scan PDU */
#define losatAutoTrackPDUKind   133   /* Losat Auto Track PDU */
#define losatAidedCueingPDUKind 134   /* Losat Aided Cueing PDU */
#define losatRangePDUKind       135   /* Losat Range PDU */
#define instrumentationPDUKind  136   /* Instrumentation PDU */
#define missReportPDUKind       137   /* Miss Report PDU */
#define srStatusPDUKind         138   /* Stingray Status PDU */
#define srCMImpactResultKind    139   /* Stingray Impact Result PDU */
typedef uint8    DataCollectionPDUKind; 


/*
 * PDUs for requesting and reporting status.
 *
 *
 * Contents of an Exercise Status PDU:
 */

typedef struct {
    /* Times: */
    SP_Time           realTime;      
    SP_Time           simulatedTime; 

    /* The terrain database chosen for the exercise: */
    TerrainDatabaseID terrain;       

    /* The battle scheme chosen for the exercise: */
    BattleScheme      battleScheme;

    /* Weather conditions: */
    SkyColor          simulatedSkyColor; 
    uint16            _unused_26;   /* padding */
    float32           VLVisibility; /* visibility in visible light in meters */
} ExerciseStatusVariant; 


/*
 * Details of SIMNET MCC system status:
 */

typedef struct {
    /* Whether various optional elements are being simulated: */
    unsigned  optionTOC        : 1;      /* Tactical Operations Center */
    unsigned  optionAdminLog   : 1;      /* Admin/Log Center */
    unsigned  optionBnHQTanks  : 1;      /* Bn HQ tank section */
    unsigned  optionScoutPlt   : 1;      /* scout platoon */
    unsigned  optionFSE        : 1;      /* fire support element */
    unsigned  optionALO        : 1;      /* air liaison officer */
    unsigned  optionCSS        : 1;      /* combat service support */
    unsigned  optionCE         : 1;      /* combat engineering */
    unsigned  _unused_27       : 24;     /* padding */
} SIMNET_MCC_Status; 


/*
 * Contents of a Simulation Status PDU:
 */

typedef struct {
    SimulatorType      simulator;    
    uint8              numberUnits;  
    uint8              _unused_28;    /* padding */

    /* Times: */
    SP_Time            exerciseEntryTime; 
    SP_Time            realTime;     
    SP_Time            simulatedTime; 

    /* The terrain database used by the simulator: */
    TerrainDatabaseID  terrain;      

    /* The user should include vehicle specific information here */

    /* The organizational units simulated: */
    OrganizationalUnit units[1];     
} SimulationStatusVariant; 


/*
 * Contents of a Vehicle Status PDU:
 */

typedef struct {
    VehicleID          vehicleID;    
    uint16             _unused_29;    /* padding */
    OrganizationalUnit unit;         
    VehicleStatus      status;       
    uint32             _unused_30;    /* padding */
} VehicleStatusVariant; 


/*
 * Contents of a Status Query PDU:
 */

#define unitRelationIrrelevant  0    
#define unitSpecified           1    
#define unitIncluded            2    
#define unitIncluding           3    

typedef uint8 UnitRelation; 

typedef struct {
    /* Specifying the type of response sought: */
    DataCollectionPDUKind responseKind;

    /* Specifying the desired respondents: */
    UnitRelation       unitRelation; 
    SimulatorType      simulatorType; 
    VehicleID          vehicleID;    
    uint16             _unused_31;    /* padding */
    OrganizationalUnit unit;         
} StatusQueryVariant; 


/*
 * Excuses for not being able to respond to a Status Query PDU:
 */

#define invalidQueryParameter   1    
#define queryConditionFailed    2    

typedef uint8 StatusResult; 


/*
 * Contents of a Status Response PDU:
 */

typedef struct {
    StatusResult result;             
    unsigned     _unused_32_33 : 24;    /* padding */
    uint32       _unused_32;            /* padding */
} StatusResponseVariant; 


/*
 * PDUs for reporting a change in vehicle status.
 *
 * 
 * What sort of status change has occurred:
 */

#define effectVehicleDestroyed  1    
#define effectVehicleReincarnated 2  
#define effectSubsystemsDamaged 3    
#define effectSubsystemsRepaired 4   

typedef uint8 StatusChangeEffect; 


/*
 * What caused damage to a vehicle:
 */

#define damageCauseOther        0     /* none of those listed below */
#define damageCauseBreakdown    1    
#define damageCauseCollision    2    
#define damageCauseCrewError    3    
#define damageCauseDirectFire   4    
#define damageCauseIndirectFire 5    
#define damageCauseIntervention 6    
#define damageCauseTerrain      7    
#define damageCauseBattlemaster 8    

typedef uint8 DamageCause; 


/*
 * What caused a repair to a vehicle:
 */

#define repairCauseOther        0     /* none of those listed below */
#define repairCauseCrew         1    
#define repairCauseIntervention 2    
#define repairCauseMaintTeam    3    
#define repairCauseBattlemaster 4    

typedef uint8 RepairCause; 


/*
 * Contents of a Status Change PDU:
 */

typedef struct {
    VehicleID           vehicleID;     
    uint8               _unused_34;     /* padding */
    StatusChangeEffect  effect;       
    union {
        struct {
            DamageCause kind;        
            unsigned    _unused_35 : 24;  /* padding */
        } destroyed;
        struct {
            RepairCause kind;        
            unsigned    _unused_36 : 24;  /* padding */
        } reincarnated;
        struct {
            DamageCause kind;        
            unsigned    _unused_37 : 24;  /* padding */
        } damaged;
        struct {
            RepairCause kind;        
            unsigned    _unused_38 : 24;  /* padding */
        } repaired;
    } cause;                            
    EventID             eventID;       
    VehicleID           agentID;       
    VehicleSubsystems   subsystems;    
} StatusChangeVariant; 


/*
 * PDUs for describing various other kinds of events.
 *
 *
 * The result of using a laser rangefinder:
 */

#define laserRangeMalfunction   1    
#define laserRangeMultipleReturns 2  
#define laserRangeNoReturn      3    
#define laserRangeSingleReturn  4    

typedef uint8 LaserRangeResult; 


/*
 * Position of the "first/last return" switch:
 */

#define noReturnSwitch          0     /* no "first/last return" switch */
#define firstReturn             1     /* switch set to "first return" */
#define lastReturn              2     /* switch set to "last return" */
typedef uint8 ReturnSwitch; 


/*
 * Types of laser range finders:
 */

#define gunnerLaser             0     /* gunner's laser */
#define citvLaser               1     /* citv laser */
typedef uint8 LaserRangeFinder; 


/*
 * Contents of a Laser Range PDU:
 */

typedef struct {
    VehicleID               vehicleID; 
    LaserRangeResult        result;  
    ReturnSwitch            returnSwitch; 
    TargetDescriptor        target;  
    SIMNET_WorldCoordinates muzzle;  
    SIMNET_WorldCoordinates location; 
    uint8                   whichLaserRange; 
    unsigned                _unused_39_40    : 24;  /* padding */
    uint32                  _unused_39;  /* padding */
} LaserRangeVariant; 


/*
 * Contents of an Event Flag PDU:
 */

typedef struct {
    int32       number;              
    uint32      sequenceNumber;      
    VehicleID   vehicleID;           
    uint16      textLength;          
    char        text[1];             
} EventFlagVariant; 


/*
 */

/*
 * Temporary PDUs
 */

/*
 *
 * ctas PDU
 *
 * This PDU is sent out after every impact PDU 
 */

typedef struct {
    VehicleID vehicleID;             
    EventID   eventID;               
    EventID   gazeEventID;           
    EventID   impactEventID;         
    uint32    _unused_41;            /* padding */
    float32   glareSize;             
    float32   scatomaSize;           
    float32   scatomaSizeDice;       
    float32   alphaEyeToLaser;       
    float32   distance;              
    float32   range;                 
    float32   alphaSensorToLaser;    
    float32   distanceSensor;        
    float32   rangeSensor;           
    float32   sensorBurnSize;        
    float32   sensorBurnSizeDice;    
    float32   sensorBloomingTime;    
    float32   sensorBloomingTimeDice; 
    float32   missionDefeatDice;     
    int16     missionDefeat;         
    uint16    _unused_42;           /* padding */
} CtasVariant; 


/*
 * Head/Eye Tracker PDU
 *
 */

#define HETicksPerSecond        15
#define EYES_UNKNOWN            0    
#define EYES_UP                 1    
#define EYES_DOWN               2    
#define EYES_CLOSED             3    

typedef uint16 EyeStatus; 

typedef struct {
    VehicleID               vehicleID; 
    EventID                 eventID; 
    EyeStatus               eye_status; 
    uint16                  _unused_43;  /* padding */
    SP_Time                 timestamp; 
    SIMNET_WorldCoordinates eyeGaze; 
    SIMNET_WorldCoordinates sensorGaze; 
    SIMNET_WorldCoordinates location; 
    float32                 rotation[3][3]; 
    int16                   adatsInFOV; 
    uint16                  _unused_44;  /* padding */
} HetVariant; 


/*
 * LOSAT PDUs
 *
 */

typedef struct {
    VehicleID      vehicleID;           
    uint8          mode;                
    unsigned       normal     : 1;       /* Boolean */
    unsigned       cmdr       : 1;       /* Boolean */
    unsigned       flir       : 1;       /* Boolean */
    unsigned       _unused_45 : 5;       /* padding */
} LosatStatusVariant; 

typedef struct {
    VehicleID               vehicleID; 
    uint16                  cueing_on; 
    float32                 topLimit; 
    float32                 bottomLimit; 
    float32                 sensitivity; 
    uint32                  vehicle_count; 
    uint32                  object_count; 
    ObjectID                targetID; 
    uint16                  _unused_46_47;  /* padding */
    uint32                  _unused_46;  /* padding */
    SIMNET_WorldCoordinates location; 
} LosatAidedCueingVariant; 

typedef struct {
    VehicleID   vehicleID;           
    uint16      systemState;         
    uint16      waypointCount;       
    uint16      offsetCount;         
    uint16      boundaryCount;       
    uint16      _unused_48;           /* padding */
    float32     scanRate;            
} LosatAidedScanVariant; 

typedef struct {
    VehicleID   vehicleID;           
    ObjectID    targetID;            
    unsigned    IsVehicle    : 1;     /* Boolean */
    unsigned    reason       : 15;   
    uint16      targetRange;         
} LosatAutoTrackVariant; 

#define laserEstimation         0    
#define manualEstimation        1    

typedef uint16 estimationMethods; 

typedef struct {
    VehicleID   vehicleID;           
    ObjectID    targetID;            
    uint32      targetType;          
    float32     actualRange;         
    float32     estimatedRange;      
    uint16      estimationMethod;    
    uint16      _unused_49;           /* padding */
} LosatRangeVariant; 


/*
 * Stingray (SR) PDU's
 */

#define SR_NO_DAMAGE            0    
#define SR_SIGHT_DEFEAT         1    

typedef uint16 SRDefeatStatus; 

typedef struct {
    EventID         fireEventID;          
    SightID         targetSightID;        
    VehicleID       targetVehicleID;      
    VehicleID       stingrayVehicleID;    
    SP_Time         simulatedTime;        
    SRDefeatStatus  sightDefeat;      
    unsigned        sightOccupied      : 1;  /* Boolean */
    unsigned        _unused_50         : 15;  /* padding */
    float32         range;                
    float32         incidenceAngle;       
    float32         azimuthAngle;         
    float32         missDistance;         
    float32         glareSize;            
    float32         sightDamageSize;      
    float32         defeatTime;            /* time in seconds that sight is */
                                           /* defeated. -1 if permanent */

} SR_CMImpactResultVariant; 

#define SR_OFF                  0    
#define SR_WARMUP               1    
#define SR_STANDBY              2    
#define SR_OPER_STOP            3    
#define SR_OPER_AUTO_SCAN       4    
#define SR_OPER_SEMI_SCAN       5    
#define SR_OPER_MAN_SCAN        6    
#define SR_OPER_RNG_SCAN        7    
#define SR_POWER_DOWN           8    

typedef uint16 SRStateValue; 

typedef struct {
    VehicleID               stingrayVehicleID; 
    SRStateValue            SRState; 
    SIMNET_WorldCoordinates SRLocation; 
    unsigned                laserArmed         : 1;  /* Boolean */
    unsigned                cmEnabled          : 1;  /* cm: Counter-Measure */
    unsigned                _unused_51         : 30; /* padding */
    float32                 leftScanLimit; 
    float32                 rightScanLimit; 
    float32                 innerScanLimit; 
    float32                 outerScanLimit; 
    float32                 leftNFZLimit;  /* NFZ:  No Fire Zone */
    float32                 rightNFZLimit; 
    float32                 innerNFZLimit; 
    float32                 outerNFZLimit; 
    float32                 SRScanRate; 
    float32                 searchWidth; 
    float32                 searchHeight; 
    float32                 SRAzimuth;  /* in radians */
    float32                 SROrientation[3][3]; 
} SRStatusVariant; 


/*
 * Instrumentation PDU
 */

typedef struct {
    VehicleID  vehicle;              
    uint16     _unused_52;           /* padding */
    uint32     control_id;           
    uint32     control_val;          
} InstrumentationVariant; 


/*
 * Miss Report PDU
 */

typedef struct {
    VehicleID attackerID;            
    EventID   eventID;               
    VehicleID intendedTargetID;      
    unsigned  terminatedShort   : 1;  /* Boolean */
    unsigned  _unused_53        : 15;  /* padding */
    float32   x;                     
    float32   z;                     
} MissReportVariant; 


/*
 */

/*
 * Definition of a Data Collection Protocol Data Unit.
 *------------------------------------------------------------------------
 */

typedef struct {
    DataCollectionProtocolVersion version; 
    DataCollectionPDUKind         kind;      
    ExerciseID                    exercise;             
    uint8                         _unused_54_55;         /* padding */
    uint32                        _unused_54;            /* padding */
    union {
        ExerciseStatusVariant     exerciseStatus;
        SimulationStatusVariant   simulationStatus;
        VehicleStatusVariant      vehicleStatus;
        StatusQueryVariant        statusQuery;
        StatusResponseVariant     statusResponse;
        StatusChangeVariant       statusChange;
        LaserRangeVariant         laserRange;
        EventFlagVariant          eventFlag;
        CtasVariant               ctas;
        HetVariant                het;
        LosatStatusVariant        losatStatus;
        LosatAidedCueingVariant   losatAidedCueing;
        LosatAidedScanVariant     losatAidedScan;
        LosatAutoTrackVariant     losatAutoTrack;
        LosatRangeVariant         losatRange;
        InstrumentationVariant    instrumentation;
        MissReportVariant         missReport;
        SRStatusVariant           srStatus;
        SR_CMImpactResultVariant  srCMImpactResult;
    } variant;                          
} DataCollectionPDU; 



#endif
