
/*
 * FILE:  p_desig.h 
 *
 * Created on Fri Apr 26 13:42:12 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_desig_included
#define _p_desig_included



/*
 * File:     p_designator.drn
 * Contents: The remote designation protocol
 * System:   simnet development
 * Created:  5-May-1992
 * Author:   aevans
 * Remarks:
 * History:
 *
 * $RCSfile: p_desig.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 * COPYRIGHT 1991 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 */

#include <stdtypes.h>
#include <basic.h> 

/*
 * Versions of the Designator Protocol
 */

#define DesignatorProtocolVersionMay92 1

typedef uint8    DesignatorProtocolVersion; 


/*
 * The current version
 */

#define designatorProtocolVersionCurrent           1

/*
 * Various kinds of Designator PDUs:
 */

#define designatorServerIdentifyPDUKind 1
#define designatorServerEntityPDUKind 2
#define designatorHandoffPDUKind 3   
#define designatorHandoffAckPDUKind 4
#define designatorCancelHandoffPDUKind 5
#define designatorDesignatePDUKind 6 
#define designatorStopDesignatePDUKind 7
#define designatorAddTargetPDUKind 8 
#define designatorDeleteTargetPDUKind 9
#define designatorAttackPDUKind 10   
#define designatorAttackAckPDUKind 11
#define designatorCancelAttackPDUKind 12

typedef uint8    DesignatorPDUKind; 


/*
 * Datagram.
 *
 * A ServerIdentify is broadcast out by the client when it desires a
 * server to attach to.  The server_address contains a specification which
 * servers should check before sending out a corresponding ServerEntity
 * PDU.  If an address component is 0, it represents a wildcard (eg.
 * server_address.site = 0 mean  a server at any site should respond if
 * the host portion is valid), otherwise it represents the site or host
 * of servers which should respond.
 *
 */

typedef struct {
    SimulationAddress serverAddress; 
} DesignatorServerIdentifyVariant; 


/*
 * Datagram.
 *
 * Sent by an entity offering the missile server service.  
 * The load_average is the load average of the missile server
 * Responding to the broadcast. Server with lightest load
 * is selected by the missile server library. 
 *
 */

typedef struct {
    SimulationAddress serverAddress; 
    uint32            _unused_3;      /* padding */
    float64           loadAverage;   
} DesignatorServerEntityVariant; 

typedef struct {
    uint16  name;                
    uint16  _unused_4;            /* padding */
    uint32  frequency;           
} LaserCode; 


/*
 * Transaction.
 *
 * A handoff is sent by the client when it desires to handoff a missile.
 * When target_mode is set to TARGET_MODE_DEST_UNDEFINED, the missile is 
 * launched in the direction the client is heading.
 *
 * When target_mode is set to TARGET_MODE_DEST_PREDEFINED, the missile
 * is launched so that it will land at the default destination.
 */

#define TARGET_MODE_DEST_UNDEFINED 1 
#define TARGET_MODE_DEST_PREDEFINED 2

typedef uint16 TargetMode; 

typedef struct {
    VehicleID               originator; 
    uint16                  _unused_5;  /* padding */
    ObjectID                designatorID; 
    uint16                  _unused_6;  /* padding */
    VehicleID               handoffID; 
    uint16                  _unused_7;  /* padding */
    ExerciseID              exercise; 
    uint8                    _unused_8;  /* padding */
    EventID                 firingEvent; 
    ObjectType              munitionType; 
    SP_Time                 delay;    /* time to wait until launching */
    LaserCode               laserCode; 
    TargetMode              targetMode; 
    uint16                  flightMode; 
    SIMNET_WorldCoordinates destination;  /* default entered by pilot */
    SIMNET_WorldCoordinates launchOffset;  /* from vehicle origin */
} DesignatorHandoffVariant; 

typedef struct {
    SimulationAddress serverAddress; 
    VehicleID         originator;    
    ObjectID          designatorID;  
    EventID           firingEvent;   
    VehicleID         munitionID;    
} DesignatorHandoffAckVariant; 


/*
 * Transaction.
 *
 * A cancel is sent by the client when it desires to cancel a missile
 * previously handed off.
 *
 */

typedef struct {
    VehicleID originator;            
    EventID   firingEvent;           
} DesignatorCancelHandoffVariant; 


/*
 * Datagram.
 *
 */

typedef struct {
    ObjectID                designatorID; 
    uint16                  _unused_9;  /* padding */
    LaserCode               laserCode; 
    SIMNET_WorldCoordinates source;   /* source of emission. */
    float64                 beamDivergence;  /* in radians. */
    SIMNET_WorldCoordinates location;  /* location and velocity info */
    VelocityVector          velocity;  /* used to RVA laser spot. */
    uint32                  timestamp; 
    EventID                 designationEvent; 
    unsigned                stationary        : 1;  /* Boolean */
    unsigned                _unused_10        : 15;  /* padding */
} DesignatorDesignateVariant; 


/*
 * Datagram.
 *
 */

typedef struct {
    ObjectID  designatorID;          
    uint16    _unused_11;           /* padding */
    LaserCode laserCode;             
    EventID   designationEvent;      
} DesignatorStopDesignateVariant; 

typedef struct {
    VehicleID               targetID; 
    uint16                  classificationStatus; 
    ObjectType              classificationType; 
    uint32                  _unused_12;  /* padding */
    SIMNET_WorldCoordinates position; 
} Target; 


/*
 * Transaction.
 *
 * Sent by the server to indicate targets which are detected.
 *
 */

typedef struct {
    SimulationAddress serverAddress; 
    uint16            _unused_13;     /* padding */
    ObjectID          designatorID;  
    uint32            targetCount;   
    Target            target[1];     
} DesignatorAddTargetVariant; 


/*
 * Transaction.
 *
 * Sent by the server to indicate targets which are no int32er detected.
 *
 */

typedef struct {
    SimulationAddress serverAddress; 
    uint16            _unused_14;     /* padding */
    ObjectID          designatorID;  
    uint32            targetCount;   
    Target            target[1];     
} DesignatorDeleteTargetVariant; 


/*
 * Transaction.
 *
 * Sent by the client to initiate missile attack.
 *
 */

typedef struct {
    VehicleID originator;            
    ObjectID  designatorID;          
    VehicleID missileID;             
    VehicleID targetID;              
} DesignatorAttackVariant; 


/*
 * Transaction.
 *
 * Sent by the server to ack receipt of missile attack.
 *
 */

typedef struct {
    VehicleID originator;            
    ObjectID  designatorID;          
    VehicleID missileID;             
    VehicleID targetID;              
} DesignatorAttackAckVariant; 


/*
 * Transaction.
 *
 * Sent by the server to cancel missile attack.
 *
 */

typedef struct {
    VehicleID originator;            
    ObjectID  designatorID;          
    VehicleID missileID;             
    VehicleID targetID;              
} DesignatorCancelAttackVariant; 

typedef struct {
    DesignatorProtocolVersion version; 
    DesignatorPDUKind         kind;          
    ExerciseID                exercise;             
    uint8                     _unused_15_16;         /* padding */
    uint32                    _unused_15;            /* padding */
    union {
        DesignatorServerIdentifyVariant serverIdentify;
        DesignatorServerEntityVariant   serverEntity;
        DesignatorHandoffVariant        handoff;
        DesignatorHandoffAckVariant     handoffAck;
        DesignatorCancelHandoffVariant  cancelHandoff;
        DesignatorDesignateVariant      designate;
        DesignatorStopDesignateVariant  stopDesignate;
        DesignatorAddTargetVariant      addTarget;
        DesignatorDeleteTargetVariant   deleteTarget;
        DesignatorAttackVariant         attack;
        DesignatorAttackVariant         attackAck;
        DesignatorCancelAttackVariant   cancelattack;
    } variant;                          
} DesignatorPDU; 



#endif
