
/*
 * FILE:  p_dis.h 
 *
 * Created on Wed Apr 17 10:52:46 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_dis_included
#define _p_dis_included



/*
 * File:     p_dis.drn
 * Contents: 
 * System:   simnet development.
 * Created:  8-Jul-1992
 * Author:   jesmith
 *
 * Remarks:
 *
 * History:
 *	22-Sep-1992	overhaul by BIS
 *		Overhaul based on DIS revision released on May 8, 1992
 *		Comments added to describe fields and cross-reference
 *		to DIS document (May 1992 version).  Corrected 
 *		inconsistencies with the May 1992 document.
 *    
 *
 * $RCSfile: p_dis.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 * COPYRIGHT 1990 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 *
 */

#include <stdtypes.h>
#include <dis_values.h> 
#include <dis_types.h> 

/*
 * Please note!  All hard-coded values that are found in the appendices
 * of the DIS document "Military Standard (Final Draft) Protocol Data
 * Units for Entity Information and Entity Interaction in a Distributed
 * Interactive Simulation" are defined in "dis_values.drn".
 *
 * Also note that commonly used types and headers are defined in
 * "dis_types.drn".
 *
 * The Angular Velocity Vector Record.  (section 5.2.2, Page 31)
 * This specifies the rate at which an entity's orientation is changing.
 * Rate is measured in BAMs per millisecond.  The first field
 * represents velocity about the x-axis; the second field 
 * represents velocity about the y-axis; the third field 
 * represents velocity about the z-axis.  Positive direction is
 * counter-clockwise motion about each axis.  Please note!!  This is
 * different than the specification for Euler Angles which are
 * ordered z,y,x.  See section 3.13 (page 4) for description of
 * entity coordinate systems. 
 */

typedef struct {
    DIS_ANGLE x;                     
    DIS_ANGLE y;                     
    DIS_ANGLE z;                     
} DIS_ANGULAR_VELOCITY_VECTOR; 


/*
 * Articulation Parameter Record. (section 5.2.3, page 32)
 * This specificies parameters for movable parts and attached parts of
 * an entity.  It includes a field to indicate if a change has happened,
 * the id of the part this parameter is attached to, parameter types
 * and values.  See Appendix G (pages 151-164) for more information on
 * articulated parts.  change indicates if anything about the parameter
 * has changed.  parm_id indicates the articulated part which this
 * parameter is attached to.  If the articulated part is attached to
 * the entity itself, then this is set to 0.
 */

typedef struct {
    uint16               change; 
    uint16               parm_id; 
    DIS_ARTICULATED_PARAMETER parm_type; 
    uint8                value[8]; 
} DIS_ARTICULATION_PARAMETER; 


/*
 * Burst Descriptor Record. (section 5.2.5, page 33)
 * The burst descriptor record describes the firing of a round or a
 * burst of ammunition.  quantity is the number of rounds fired; rate
 * is rate of firing in rounds per second; munition is the kind of
 * munition.
 */

typedef struct {
    DIS_ENTITY_TYPE munition;        
    DIS_WARHEAD_TYPE warhead;        
    DIS_FUZE_TYPE   fuze;            
    uint16     quantity;        
    uint16     rate;            
} DIS_BURST_DESCRIPTOR; 


/*
 * Dead Reckoning Record. (section 5.3.4.1, letter j, page 48 and 
 * appendix i, pages 175-178)
 * The dead reckoning record provides which algorithm is being used
 * for d.r., linear acceleration and angular velocity.  It has room
 * for other parameters which are not yet specified.
 */

typedef struct {
    DIS_DEAD_RECKONING_ALGORITHM algorithm; 
    unsigned                       _unused_2_3          : 24;  /* padding */
    uint32                  _unused_2_4;  /* padding */
    uint32                  _unused_2_5;  /* padding */
    uint32                  _unused_2;  /* padding */
    DIS_VECTOR                  linear_acceleration; 
    DIS_ANGULAR_VELOCITY_VECTOR angular_velocity; 
} DIS_DEAD_RECKONING_PARAMETERS; 


/*
 * The Entity Capabilities Record. (section 5.2.7, page 34)
 * The entity capabilities record reports on what an entity can and
 * can't do.
 */

typedef struct {
    unsigned       ammunition_supply  : 1; 
    unsigned       fuel_supply        : 1; 
    unsigned       misc_supply        : 1; 
    unsigned       repair             : 1; 
    unsigned       _unused_6          : 28;  /* padding */
} DIS_ENTITY_CAPABILITIES; 


/*
 * The Entity Marking Record. (section 5.2.9, page 36)
 * The Entity Marking Record specified the character set to be used, and
 * 11 characters to be interpreted for display.
 */

#define DIS_MAX_ENTITY_MARKING_LENGTH           11
typedef struct {
    DIS_ENTITY_MARKING_CHAR_SET char_set; 
    uint8  string[DIS_MAX_ENTITY_MARKING_LENGTH]; 
} DIS_ENTITY_MARKING; 


/*
 * Euler Angles Record. (section 3.14, page 5 and section 5.2.11, page 38)
 * A set of three angles used to describe the orientation of an entity
 * as a set of three successive rotations about three different
 * orthogonal axes.  First about z, angle psi; second about y, angle
 * theta; third about x, angle phi.  The angles are expressed in BAMs.
 */

typedef struct {
    DIS_ANGLE psi;                   
    DIS_ANGLE theta;                 
    DIS_ANGLE phi;                   
} DIS_EULER_ANGLE; 


/*
 * Event Identifier Record. (section 5.2.12, page 38)
 * An event is identified by a simlation address and a sequence number.
 */

typedef struct {
    DIS_SIMULATION_ADDRESS simulator; 
    uint16            event;    
} DIS_EVENT_ID; 


/*
 * The Organizational Unit Record. (section 5.2.14, page 39)
 * The organization unit is specified by the force (us, them or neutral),
 * the unit's country, branch of military service, and unit's placement
 * in the military hierarchy.
 */

typedef struct {
    DIS_FORCE_ID  force;             
     /* WARNING: the following should be a DIS_COUNTRY_ID, but due to bad */
         /* byte alignment, it must be split into two bytes. */

    uint8    country1;          
    uint8    country2;          
    DIS_SERVICE_ID service;          
    DIS_HIERARCHY hierarchy;         
} DIS_ORGANIZATIONAL_UNIT; 


/*
 * Supply Quantity Record. (section 5.2.17, page 40)
 * Supply quantity is defined by an entity type and the amount of that
 * entity.
 */

typedef struct {
    DIS_ENTITY_TYPE supply_type;     
    float32              quantity;        
} DIS_SUPPLY_QUANTITY; 


/*
 * Terrain Database Identifier Record. (section 5.2.18, page 41)
 * The database id is an 11 character string and a 1-byte version number.
 */

typedef struct {
    uint8  db_name[11];         
    uint8  version;             
} DIS_TERRAIN_DATABASE_ID; 


/*
 * Entity State PDU. (section 5.3.4.1, pages 47-50)
 */

typedef struct {
    DIS_PDU_HEADER                header; 
    DIS_ENTITY_IDENTIFIER         entity_id; 
    uint8                    _unused_7;  /* padding */
    DIS_FORCE_ID                  force_id; 
    DIS_ENTITY_TYPE               entity_type; 
    DIS_ENTITY_TYPE               alternate_entity_type; 
    DIS_TIME                      time_stamp; 
    DIS_WORLD_COORDINATES         location; 
    DIS_VECTOR                    velocity; 
    DIS_EULER_ANGLE               orientation; 
    DIS_DEAD_RECKONING_PARAMETERS dead_reckoning; 
    DIS_ENTITY_APPEARANCE         appearance; 
    DIS_ENTITY_MARKING            marking; 
    DIS_ENTITY_CAPABILITIES       capabilities; 
    unsigned                         _unused_8              : 24;  /* padding */
    uint8                    articulated_count; 
    DIS_ARTICULATION_PARAMETER    articulated[1]; 
} DIS_ENTITY_STATE_PDU; 


/*
 * Fire PDU. (section 5.3.5.1.1, pages 51-53)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER attacker_id; 
    DIS_ENTITY_IDENTIFIER target_id; 
    DIS_ENTITY_IDENTIFIER munition_id; 
    DIS_EVENT_ID          event_id;  
    DIS_TIME              time_stamp; 
    DIS_WORLD_COORDINATES location;  
    DIS_BURST_DESCRIPTOR  burst;     
    DIS_VECTOR            velocity;  
    float32                    fire_range; 
} DIS_FIRE_PDU; 


/*
 * Detonation PDU. (section 5.3.5.1.2, pages 54-57)
 */

typedef struct {
    DIS_PDU_HEADER             header; 
    DIS_ENTITY_IDENTIFIER      attacker_id; 
    DIS_ENTITY_IDENTIFIER      target_id; 
    DIS_ENTITY_IDENTIFIER      munition_id; 
    DIS_EVENT_ID               event_id; 
    DIS_TIME                   time_stamp; 
    DIS_WORLD_COORDINATES      location; 
    DIS_BURST_DESCRIPTOR       burst; 
    DIS_VECTOR                 velocity; 
    DIS_VECTOR                 location_entity; 
    DIS_DETONATION_RESULT      result; 
    uint8                 articulated_count; 
    uint16                _unused_9;  /* padding */
    DIS_ARTICULATION_PARAMETER articulated[1]; 
} DIS_DETONATION_PDU; 


/*
 * Service request pdu. (section 5.3.5.2.1, pages 58-59)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER receiver_id; 
    DIS_ENTITY_IDENTIFIER supplier_id; 
    DIS_SERVICE_TYPE      service;   
    uint8            num_supply; 
    uint16           _unused_10;  /* padding */
    DIS_SUPPLY_QUANTITY   quantity[1]; 
} DIS_SERVICE_REQUEST_PDU; 


/*
 * Resupply offer pdu. (section 5.3.5.2.2, pages 60-61)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER receiver_id; 
    DIS_ENTITY_IDENTIFIER supplier_id; 
    uint8            num_supply; 
    unsigned                 _unused_11   : 24;  /* padding */
    DIS_SUPPLY_QUANTITY   quantity[1]; 
} DIS_RESUPPLY_OFFER_PDU; 


/*
 * Resupply received pdu. (section 5.3.5.2.3, pages 62-63)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER receiver_id; 
    DIS_ENTITY_IDENTIFIER supplier_id; 
    uint8            num_supply; 
    unsigned                 _unused_12   : 24;  /* padding */
    DIS_SUPPLY_QUANTITY   quantity[1]; 
} DIS_RESUPPLY_RECEIVED_PDU; 


/*
 * Resupply cancelled pdu. (section 5.3.5.2.4, page 64)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER receiver_id; 
    DIS_ENTITY_IDENTIFIER supplier_id; 
} DIS_RESUPPLY_CANCEL_PDU; 


/*
 * Repair complete pdu. (section 5.3.5.2.5, pages 66-67)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER receiver_id; 
    DIS_ENTITY_IDENTIFIER supplier_id; 
    DIS_REPAIR            repair;    
    uint16           _unused_13;  /* padding */
} DIS_REPAIR_COMPLETE_PDU; 


/*
 * Repair response pdu. (section 5.3.5.2.6, pages 68-69)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER receiver_id; 
    DIS_ENTITY_IDENTIFIER supplier_id; 
    DIS_REPAIR_RESULT     result;    
    unsigned                 _unused_14   : 24;  /* padding */
} DIS_REPAIR_RESPONSE_PDU; 


/*
 * Collision pdu. (section 5.3.5.3, pages 70-71)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER issue_id;  
    DIS_ENTITY_IDENTIFIER collide_id; 
    DIS_EVENT_ID          event;     
    uint16           _unused_15;  /* padding */
    DIS_TIME              time_stamp; 
    DIS_VECTOR            velocity;  
    float32                    mass;      
    DIS_VECTOR            location;  
} DIS_COLLISION_PDU; 


/*
 * Deactivate request pdu. (section 6.3.1.3, pages 85-86)
 */

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER entity_id; 
    DIS_DEACTIVATE_REASON deactivate_reason; 
    uint8            _unused_16;  /* padding */
} DIS_DEACTIVATE_REQUEST_PDU; 


/*
 * Radar system (section 6.3.3.3, page 95)
 */

typedef struct {
    DIS_RADAR_SYSTEM_CATEGORY category; 
    uint8  subcategory;         
    DIS_RADAR_SYSTEM_ID id;          
} DIS_RADAR_SYSTEM; 


/*
 * Radar pdu and associated sub types (section 6.3.3, pages 95-99)
 */

typedef struct {
    DIS_ANGLE azimuth_center;        
    DIS_ANGLE azimuth_sweep;         
    DIS_ANGLE elevation_center;      
    DIS_ANGLE elevation_sweep;       
} DIS_RADAR_SWEEP; 

typedef struct {
    DIS_ENTITY_IDENTIFIER target;    
    uint16           _unused_17;  /* padding */
    uint32            radar_data; 
} DIS_ILLUMINED_INFO; 

typedef struct {
    DIS_VECTOR         location;     
    DIS_RADAR_SYSTEM   radar_system; 
    int16                 power;        
    DIS_RADAR_MODE     mode;         
    uint8         illumined;    
    uint32         specific_data0; 
    uint32         specific_data1; 
    DIS_RADAR_SWEEP    sweep;        
    DIS_ILLUMINED_INFO illumined_data[1]; 
} DIS_RADAR_SYSTEM_INFO; 

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_ENTITY_IDENTIFIER entity_id; 
    DIS_EVENT_ID          event;     
    DIS_TIME              time_stamp; 
    unsigned                 _unused_18   : 24;  /* padding */
    uint8            radar_count; 
    DIS_RADAR_SYSTEM_INFO radar[1];  
} DIS_RADAR_PDU; 


/*
 * Radio PDU's
 *
 */

typedef struct {
    unsigned       spread_spectrum[16]; 
    DIS_RADIO_MODULATION_MAJOR mod_major; 
    DIS_RADIO_MODULATION_DETAIL mod_detail; 
    DIS_RADIO_MODULATION_SYSTEM mod_system; 
} DIS_RADIO_MODULATION_TYPE; 

typedef struct {
    DIS_PDU_HEADER            header; 
    DIS_TIME                  time_stamp; 
    uint16               pdu_length; 
    uint16               _unused_19;  /* padding */
    DIS_ENTITY_IDENTIFIER     entity_id; 
    uint16               radio_id; 
    DIS_ENTITY_TYPE           radio_entity_type; 
    DIS_RADIO_TRANSMIT_STATE  transmit_state; 
    uint8                input_source; 
    uint16               _unused_20;  /* padding */
    DIS_WORLD_COORDINATES     antenna_world_loc; 
    DIS_VECTOR                antenna_relative_loc; 
    DIS_RADIO_ANTENNA_PATTERN_TYPE antenna_pattern_type; 
    uint16               antenna_pattern_length; 
    uint32                freq_hack; 
    uint32                frequency; 
    float32                        frequency_bw; 
    float32                        power; 
    DIS_RADIO_MODULATION_TYPE modulation_type; 
    DIS_RADIO_CRYPTO_SYSTEM   crypto_system; 
    uint16               crypto_key_id; 
    uint8                modulation_parm_length; 
    unsigned                     _unused_21              : 24;  /* padding */
} DIS_TRANSMITTER_PDU; 

typedef struct {
    DIS_PDU_HEADER        header;    
    DIS_TIME              time_stamp; 
    uint16           pdu_length; 
    uint16           _unused_22;  /* padding */
    DIS_ENTITY_IDENTIFIER entity_id; 
    uint16           radio_id;  
    DIS_RADIO_ENCODING_SCHEME encoding_scheme; 
    uint16           _unused_23;  /* padding */
    int32                     sample_rate; 
    int16                    length;    
    int16                    samples;   
    uint8            data[1];   
} DIS_SIGNAL_PDU; 



#endif
