
/*
 * FILE:  p_dis203ae.h 
 *
 * Created on Wed Apr 17 10:52:42 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_dis203ae_included
#define _p_dis203ae_included



/*
 * File:     p_dis203ae.drn
 * Contents: 
 * System:   BDS-D BIP development.
 * Created:  6-Oct-93
 * Author:   mgreess
 *
 * Remarks:
 *
 * History:
 * 2-Dec-1993	Data entered in DRN format by mgreess from
 *		IRS DIS-Interface to SAFSim BL 5121 T058 document.
 *    
 *
 * $RCSfile: p_dis203ae.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 * COPYRIGHT 1993 Loral Advanced Distributed Simulation
 * 50 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3600
 *
 */

#include <stdtypes.h>
#include <dis203_types.h> 

/*
 * Please note!  All hard-coded values that are found in the appendices
 * of the DIS document "Military Standard (Final Draft) Protocol Data
 * Units for Entity Information and Entity Interaction in a Distributed
 * Interactive Simulation" are defined in "dis203_values.drn".
 *
 * Also note that commonly used types and headers which are part of the
 * official DIS standard are defined in "dis203_types.drn".
 *
 * Any type definitions or enums which are specific to the AE extensions
 * are defined in this file.
 *
 *
 * AE PDU kinds --
 *
 *
 */

#define DIS203AE_ENVIRONMENT_PDU_KIND           128
#define DIS203AE_DYNAMIC_TERRAIN_PDU_KIND           129

/*
 *
 * AE Enum, Type, and Record definitions for the Environment PDU.
 * (section 3.3.1.1.4, pages 31-32)
 *
 *
 *
 * AE Precipitation Type Enum. (section 3.3.1.1.5, page 30)
 *
 */

#define DIS203AE_PRECIPITATION_TYPE_NONE           0
#define DIS203AE_PRECIPITATION_TYPE_CLOUDS           1
#define DIS203AE_PRECIPITATION_TYPE_FOG           2
#define DIS203AE_PRECIPITATION_TYPE_RAIN           3
#define DIS203AE_PRECIPITATION_TYPE_SNOW           4
#define DIS203AE_PRECIPITATION_TYPE_HAIL           5

/*
 * AE Time. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef float64           DIS203AE_TIME_TYPE; 


/*
 * AE Precipitation Type. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef uint8    DIS203AE_PRECIPITATION_TYPE; 


/*
 * AE Precipitation Level. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef uint8    DIS203AE_PRECIPITATION_LEVEL_TYPE; 


/*
 * AE Temperature. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef char             DIS203AE_TEMPERATURE_TYPE; 


/*
 * AE Humidity. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef uint8    DIS203AE_HUMIDITY_TYPE; 


/*
 * AE Air Pressure. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef float32            DIS203AE_AIR_PRESSURE_TYPE; 


/*
 * AE Precipitation Record. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef struct {
    DIS203AE_PRECIPITATION_TYPE       precip_type; 
    DIS203AE_PRECIPITATION_LEVEL_TYPE precip_level; 
} DIS203AE_PRECIPITATION_RECORD; 


/*
 * AE Weather Record. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef struct {
    DIS203AE_PRECIPITATION_RECORD        precipitation; 
    uint16                          _unused_7;  /* padding */
    DIS203_LENGTH_TYPE                   visual_range; 
    DIS203AE_TEMPERATURE_TYPE            temperature; 
    DIS203AE_HUMIDITY_TYPE               humidity; 
    DIS203AE_AIR_PRESSURE_TYPE           air_pressure; 
    DIS203_LINEAR_VELOCITY_VECTOR_RECORD wind_velocity; 
} DIS203AE_WEATHER_RECORD; 


/*
 * Environment PDU. (section 3.3.1.1.4, pages 31-32)
 *
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD pdu_header; 
    DIS203_ENTITY_ID_RECORD  entity_id; 
    DIS203_EVENT_ID_RECORD   event_id; 
    DIS203AE_TIME_TYPE       time_of_day; 
    DIS203AE_WEATHER_RECORD  weather; 
} DIS203AE_ENVIRONMENT_PDU; 


/*
 *
 * AE Enum, Type, and Record definitions for the Dynamic Terrain PDU.
 * (section 3.3.1.1.3, page 30)
 *
 *
 *
 * AE Dynamic Terrain Object Type Enum. (section 3.3.1.1.5, page 30)
 *
 */

#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE_BUILDING           0
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE_BRIDGE           1
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE_BARRIER           2
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE_CRATER           3
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE_FREE_CRATER           4

/*
 * AE Dynamic Terrain Object State Enum. (section 3.3.1.1.5, page 30)
 *
 */

#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_INVISIBLE           0
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_UNDAMAGED           1
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_MODERATELY_DAMAGED           2
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_HEAVILY_DAMAGED           3
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_DESTROYED           4

/*
 * AE Dynamic Terrain Object Appearance Enum. (section 3.3.1.1.5, page 30)
 *
 */

#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_APPEARANCE_NORMAL           0
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_APPEARANCE_SMOKING           1
#define DIS203AE_DYNAMIC_TERRAIN_OBJECT_APPEARANCE_BURNING           2

/*
 * AE Dynamic Terrain Object ID. (section 3.3.1.1.5, page 30)
 *
 */

typedef uint32 DIS203AE_DYNAMIC_TERRAIN_OBJECT_ID_TYPE; 


/*
 * AE Dynamic Terrain Object Type. (section 3.3.1.1.5, page 30)
 *
 */

typedef uint16 DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE; 


/*
 * AE Dynamic Terrain Object State Type. (section 3.3.1.1.5, page 30)
 *
 */

typedef uint16 DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_TYPE; 


/*
 * AE Dynamic Terrain Object Appearance Type. (section 3.3.1.1.5, page 30)
 *
 */

typedef uint16 DIS203AE_DYNAMIC_TERRAIN_OBJECT_APPEARANCE_TYPE; 


/*
 * Environment PDU. (section 3.3.1.1.5, page 30)
 *
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                        pdu_header; 
    DIS203AE_DYNAMIC_TERRAIN_OBJECT_ID_TYPE         object_id; 
    DIS203_WORLD_COORDINATES_RECORD                 location; 
    DIS203AE_DYNAMIC_TERRAIN_OBJECT_TYPE            object_type; 
    DIS203AE_DYNAMIC_TERRAIN_OBJECT_STATE_TYPE      object_state; 
    DIS203AE_DYNAMIC_TERRAIN_OBJECT_APPEARANCE_TYPE object_appearance; 
} DIS203AE_DYNAMIC_TERRAIN_PDU; 



#endif
