
/*
 * FILE:  p_dis203.h 
 *
 * Created on Wed Apr 17 10:53:32 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_dis203_included
#define _p_dis203_included



/*
 * File:     p_dis203.drn
 * Contents: 
 * System:   BDS-D BIP development.
 * Created:  6-Oct-93
 * Author:   cchiang/ckuczaj
 *
 * Remarks:
 *
 * History:
 * 6-Oct-1993	Data entered in DRN format by cchiang from
 *		C include files created by ckuczaj
 *		Based on DIS revision 2.0.3 released on May 28, 1993
 *		Comments added to describe fields and cross-reference
 *		to DIS document (May 1993 version).
 *
 * 13-Jan-1994  The PDUs not included in the initial implementation 
 *              have been added (Logistics, Sim. Mgmt., Laser, and Receiver).
 *    
 *
 * $RCSfile: p_dis203org.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 * COPYRIGHT 1993 Loral Advanced Distributed Simulation
 * 50 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3600
 *
 */

#include <stdtypes.h>
#include <dis203_types.h> 

/*
 * Please note!  All hard-coded values that are found in the appendices
 * of the DIS document "Military Standard (Final Draft) Protocol Data
 * Units for Entity Information and Entity Interaction in a Distributed
 * Interactive Simulation" are defined in "dis203_values.drn".
 *
 * Also note that commonly used types and headers are defined in
 * "dis203_types.drn".
 *
 *
 * Entity State PDU. (section 5.4.3.1, pages 89-92)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                      pdu_header; 
    DIS203_ENTITY_ID_RECORD                       entity_id; 
    DIS203_FORCE_ID_TYPE                          force_id; 
    DIS203_NUMBER_OF_ARTICULATION_PARAMETERS_TYPE number_of_articulation_parameters; 
    DIS203_ENTITY_TYPE_RECORD                     entity_type; 
    DIS203_ENTITY_TYPE_RECORD                     alternative_entity_type; 
    DIS203_LINEAR_VELOCITY_VECTOR_RECORD          entity_linear_velocity; 
    DIS203_WORLD_COORDINATES_RECORD               entity_location; 
    DIS203_EULER_ANGLES_RECORD                    entity_orientation; 
    DIS203_ENTITY_APPEARANCE_TYPE                 entity_appearance; 
    DIS203_DEAD_RECKONING_PARAMETERS_RECORD       dead_reckoning_parameters; 
    DIS203_ENTITY_MARKING_RECORD                  entity_marking; 
    DIS203_ENTITY_CAPABILITIES_RECORD             capabilities; 
    DIS203_ARTICULATION_PARAMETER_RECORD          articulation_parameters[1]; 
} DIS203_ENTITY_STATE_PDU; 


/*
 * Fire PDU. (section 5.4.4.1.1, pages 93-94)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD             pdu_header; 
    DIS203_ENTITY_ID_RECORD              firing_entity_id; 
    DIS203_ENTITY_ID_RECORD              target_entity_id; 
    DIS203_ENTITY_ID_RECORD              munition_id; 
    DIS203_EVENT_ID_RECORD               event_id; 
    uint32                           _unused_7;  /* padding */
    DIS203_WORLD_COORDINATES_RECORD      location_in_world; 
    DIS203_BURST_DESCRIPTOR_RECORD       burst_descriptor; 
    DIS203_LINEAR_VELOCITY_VECTOR_RECORD velocity; 
    DIS203_RANGE_TYPE                    fire_range; 
} DIS203_FIRE_PDU; 


/*
 * Detonation PDU. (section 5.4.4.1.2, pages 95-98)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                      pdu_header; 
    DIS203_ENTITY_ID_RECORD                       firing_entity_id; 
    DIS203_ENTITY_ID_RECORD                       target_entity_id; 
    DIS203_ENTITY_ID_RECORD                       munition_id; 
    DIS203_EVENT_ID_RECORD                        event_id; 
    DIS203_LINEAR_VELOCITY_VECTOR_RECORD          velocity; 
    DIS203_WORLD_COORDINATES_RECORD               location_in_world; 
    DIS203_BURST_DESCRIPTOR_RECORD                burst_descriptor; 
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD        location_in_entity_coordinates; 
    DIS203_DETONATION_RESULT_TYPE                 detonation_result; 
    DIS203_NUMBER_OF_ARTICULATION_PARAMETERS_TYPE number_of_articulation_parameters; 
    uint16                                   _unused_8;  /* padding */
    DIS203_ARTICULATION_PARAMETER_RECORD          articulation_parameters[1]; 
} DIS203_DETONATION_PDU; 
 /* Logistics  */

 /* Service Request PDU (section 5.4.4.2.1, pages 99-100) */
typedef struct {
    DIS203_PDU_HEADER_RECORD           pdu_header; 
    DIS203_ENTITY_ID_RECORD            requesting_entity_id; 
    DIS203_ENTITY_ID_RECORD            servicing_entity_id; 
    DIS203_SERVICE_TYPE_TYPE           service_type; 
    DIS203_NUMBER_OF_SUPPLY_TYPES_TYPE number_of_supply_types; 
    uint16                        _unused_9;  /* padding */
    DIS203_SUPPLY_QUANTITY_RECORD      supplies[1]; 
} DIS203_SERVICE_REQUEST_PDU; 


/*
 * Resupply Offer PDU (section 5.4.4.2.2, pages 101-102)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD           pdu_header; 
    DIS203_ENTITY_ID_RECORD            receiving_entity_id; 
    DIS203_ENTITY_ID_RECORD            supplying_entity_id; 
    DIS203_NUMBER_OF_SUPPLY_TYPES_TYPE number_of_supply_types; 
    unsigned                              _unused_10              : 24;  /* padding */
    DIS203_SUPPLY_QUANTITY_RECORD      supplies[1]; 
} DIS203_RESUPPLY_OFFER_PDU; 


/*
 * Resupply Received PDU (section 5.4.4.2.3, pages 103-104)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD           pdu_header; 
    DIS203_ENTITY_ID_RECORD            receiving_entity_id; 
    DIS203_ENTITY_ID_RECORD            supplying_entity_id; 
    DIS203_NUMBER_OF_SUPPLY_TYPES_TYPE number_of_supply_types; 
    unsigned                              _unused_11              : 24;  /* padding */
    DIS203_SUPPLY_QUANTITY_RECORD      supplies[1]; 
} DIS203_RESUPPLY_RECEIVED_PDU; 


/*
 * Resupply Cancel PDU (section 5.4.4.2.4, page 105)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD pdu_header; 
    DIS203_ENTITY_ID_RECORD  receiving_entity_id; 
    DIS203_ENTITY_ID_RECORD  supplying_entity_id; 
} DIS203_RESUPPLY_CANCEL_PDU; 


/*
 * Repair Complete PDU (section 5.4.4.2.5, page 106)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD pdu_header; 
    DIS203_ENTITY_ID_RECORD  receiving_entity_id; 
    DIS203_ENTITY_ID_RECORD  repairing_entity_id; 
    DIS203_REPAIR_TYPE_TYPE  repair; 
    uint16              _unused_12;  /* padding */
} DIS203_REPAIR_COMPLETE_PDU; 


/*
 * Repair Response PDU (section 5.4.4.2.6, page 107)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD  pdu_header; 
    DIS203_ENTITY_ID_RECORD   receiving_entity_id; 
    DIS203_ENTITY_ID_RECORD   repairing_entity_id; 
    DIS203_REPAIR_RESULT_TYPE repair_result; 
    unsigned                     _unused_13           : 24;  /* padding */
} DIS203_REPAIR_RESPONSE_PDU; 


/*
 * Collision pdu. (section 5.4.4.3.1, pages 108-109)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD               pdu_header; 
    DIS203_ENTITY_ID_RECORD                issuing_entity_id; 
    DIS203_ENTITY_ID_RECORD                colliding_entity_id; 
    DIS203_EVENT_ID_RECORD                 event_id; 
    uint16                            _unused_14;  /* padding */
    DIS203_LINEAR_VELOCITY_VECTOR_RECORD   velocity; 
    DIS203_MASS_TYPE                       mass; 
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD location; 
} DIS203_COLLISION_PDU; 


/*
 * Simulation Management
 *
 * Create Entity PDU (section 5.4.5.2, page 115)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD  originating_entity_id; 
    uint16              _unused_15;  /* padding */
    DIS203_ENTITY_ID_RECORD  receiving_entity_id; 
    uint16              _unused_16;  /* padding */
} DIS203_CREATE_ENTITY_PDU; 


/*
 * Remove pdu. (section 5.4.5.3, page 116)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD  originating_entity_id; 
    uint16              _unused_17;  /* padding */
    DIS203_ENTITY_ID_RECORD  receiving_entity_id; 
    uint16              _unused_18;  /* padding */
} DIS203_REMOVE_ENTITY_PDU; 


/*
 * Start/Resume pdu. (section 5.4.5.4, pages 117-118)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD  originating_entity_id; 
    uint16              _unused_19;  /* padding */
    DIS203_ENTITY_ID_RECORD  receiving_entity_id; 
    uint16              _unused_20;  /* padding */
    DIS203_CLOCK_TIME_RECORD real_world_time; 
    DIS203_CLOCK_TIME_RECORD simulation_time; 
} DIS203_START_RESUME_PDU; 


/*
 * Stop/Freeze pdu. (section 5.4.5.5, page 119)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD    pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD     originating_entity_id; 
    uint16                 _unused_21;  /* padding */
    DIS203_ENTITY_ID_RECORD     receiving_entity_id; 
    uint16                 _unused_22;  /* padding */
    DIS203_CLOCK_TIME_RECORD    real_world_time; 
    DIS203_REASONS_TO_STOP_TYPE reason; 
    unsigned                       _unused_23             : 24;  /* padding */
} DIS203_STOP_FREEZE_PDU; 


/*
 * Acknowledge pdu. (section 5.4.5.6, page 120)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD          pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD           originating_entity_id; 
    uint16                       _unused_24;  /* padding */
    DIS203_ENTITY_ID_RECORD           receiving_entity_id; 
    uint16                       _unused_25;  /* padding */
    DIS203_ACKNOWLEDGEMENT_FLAGS_TYPE acknowledge_flag; 
    DIS203_RESPONSE_FLAGS_TYPE        response_flag; 
} DIS203_ACKNOWLEDGE_PDU; 


/*
 * Action Request PDU (section 5.4.5.7, pages 121-122)
 * Note : This definition only allows a variable number of fixed datum
 * records.
 *
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                 pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD                  originating_entity_id; 
    uint16                              _unused_26;  /* padding */
    DIS203_ENTITY_ID_RECORD                  receiving_entity_id; 
    uint16                              _unused_27;  /* padding */
    uint32                               _unused_28;  /* padding */
    DIS203_REQUEST_ID_TYPE                   request_id; 
    DIS203_ACTION_ID_TYPE                    action_id; 
    DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE number_of_fixed_datum_fields; 
    DIS203_FIXED_DATUM_RECORD                fixed_datum[1]; 
} DIS203_ACTION_REQUEST_PDU; 


/*
 * Action Response PDU (section 5.4.5.8, pages 123-124)
 * Note : This definition only allows a variable number of fixed datum 
 * records. 
 * 
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                 pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD                  originating_entity_id; 
    uint16                              _unused_29;  /* padding */
    DIS203_ENTITY_ID_RECORD                  receiving_entity_id; 
    uint16                              _unused_30;  /* padding */
    uint32                               _unused_31;  /* padding */
    DIS203_REQUEST_ID_TYPE                   request_id; 
    DIS203_REQUEST_STATUS_TYPE               request_status; 
    DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE number_of_fixed_datum_fields; 
    DIS203_FIXED_DATUM_RECORD                fixed_datum[1]; 
} DIS203_ACTION_RESPONSE_PDU; 


/*
 * Data Query PDU (section 5.4.5.9, pages 125-126)
 * Note : This definition only allows a variable number of fixed datum 
 * records. 
 * Removed 32-bit padding after receiving_entity_id, pmonday 12-94
 * 
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                 pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD                  originating_entity_id; 
    uint16                              _unused_32;  /* padding */
    DIS203_ENTITY_ID_RECORD                  receiving_entity_id; 
    uint16                              _unused_33;  /* padding */
    DIS203_REQUEST_ID_TYPE                   request_id; 
    DIS203_TIME_INTERVAL_TYPE                time_interval; 
    DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE number_of_fixed_datum_fields; 
    DIS203_FIXED_DATUM_RECORD                fixed_datum[1]; 
} DIS203_DATA_QUERY_PDU; 


/*
 * Set Data PDU (section 5.4.5.10, pages 127-128)
 * Note : This definition only allows a variable number of fixed datum
 * records.
 * Removed 32-bit padding after receiving_entity_id, pmonday 12-94
 *
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                 pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will */
         /*       eventually be removed, so we substitute padding */

    DIS203_ENTITY_ID_RECORD                  originating_entity_id; 
    uint16                              _unused_34;  /* padding */
    DIS203_ENTITY_ID_RECORD                  receiving_entity_id; 
    uint16                              _unused_35;  /* padding */
    DIS203_REQUEST_ID_TYPE                   request_id; 
    DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE number_of_fixed_datum_fields; 
    DIS203_FIXED_DATUM_RECORD                fixed_datum[1]; 
} DIS203_SET_DATA_PDU; 
 /*  */
 /* Data PDU (section 5.4.5.11, pages 129-130)  */
 /* Note : This definition only allows a variable number of fixed datum  */
 /* records.  */
 /* Removed 32-bit padding after receiving_entity_id, pmonday 12-94 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                 pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will  */
         /*       eventually be removed, so we substitute padding  */

    DIS203_ENTITY_ID_RECORD                  originating_entity_id; 
    uint16                              _unused_36;  /* padding */
    DIS203_ENTITY_ID_RECORD                  receiving_entity_id; 
    uint16                              _unused_37;  /* padding */
    DIS203_REQUEST_ID_TYPE                   request_id; 
    DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE number_of_fixed_datum_fields; 
    DIS203_FIXED_DATUM_RECORD                fixed_datum[1]; 
} DIS203_DATA_PDU; 
 /*  */
 /* Event Report PDU (section 5.4.5.12, pages 131-132)  */
 /* This definition assumes that number_of_fixed_datum_fields = 0 */
 /* Removed 32-bit padding after receiving_entity_id, pmonday 12-94 */

typedef struct {
    DIS203_PDU_HEADER_RECORD                    pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will  */
         /*       eventually be removed, so we substitute padding  */

    DIS203_ENTITY_ID_RECORD                     originating_entity_id; 
    uint16                                 _unused_38;  /* padding */
    DIS203_ENTITY_ID_RECORD                     receiving_entity_id; 
    uint16                                 _unused_39;  /* padding */
    DIS203_EVENT_TYPE_TYPE                      event_type; 
    DIS203_NUMBER_OF_FIXED_DATUM_FIELDS_TYPE    number_of_fixed_datum_fields; 
    DIS203_NUMBER_OF_VARIABLE_DATUM_FIELDS_TYPE number_of_variable_datum_fields; 
    DIS203_VARIABLE_DATUM_RECORD                variable_datum[1]; 
} DIS203_EVENT_REPORT_PDU; 
 /*   */
 /* Message PDU (section 5.4.5.13, pages 133-134)   */
 /* Note : This definition only allows a variable number of variable datum   */
 /* records.  */
 /* Removed 32-bit padding after receiving_entity_id, pmonday 12-94 */
 /*  */
typedef struct {
    DIS203_PDU_HEADER_RECORD                    pdu_header; 
     /* NOTE: The protocol includes a 'group' in the entity_id's.  This will   */
         /*       eventually be removed, so we substitute padding   */

    DIS203_ENTITY_ID_RECORD                     originating_entity_id; 
    uint16                                 _unused_40;  /* padding */
    DIS203_ENTITY_ID_RECORD                     receiving_entity_id; 
    uint16                                 _unused_41;  /* padding */
    DIS203_NUMBER_OF_VARIABLE_DATUM_FIELDS_TYPE number_of_variable_datum_fields; 
    DIS203_VARIABLE_DATUM_RECORD                variable_datum[1]; 
} DIS203_MESSAGE_PDU; 


/*
 * Distributed Emission Regeneration
 *
 * Emission pdu. (section 5.4.6.2, pages 138-142)
 *
 * NOTE: This definition is good for 1 system with 1 beam.  Anything
 *       more cannot be defined using DRN.
 *
 */

typedef struct {
    DIS203_BEAM_DATA_LENGTH_TYPE             beam_data_length; 
    DIS203_BEAM_ID_NUMBER_TYPE               beam_id_number; 
    DIS203_BEAM_PARAMETER_INDEX_TYPE         beam_parameter_index; 
    DIS203_FUNDAMENTAL_PARAMETER_DATA_RECORD fundamental_parameter_data; 
    DIS203_BEAM_FUNCTION_TYPE                beam_function; 
    DIS203_NUMBER_OF_TARGETS_TYPE            number_of_targets_in_track_jam_field; 
    DIS203_HIGH_DENSITY_TRACK_JAM_TYPE       high_density_track_jam; 
    uint8                               _unused_42;  /* padding */
    DIS203_JAMMING_MODE_SEQUENCE_TYPE        jamming_mode_sequence; 
    DIS203_TRACK_JAM_RECORD                  track_jam[1]; 
} DIS203_EMISSION_BEAM_RECORD; 

typedef struct {
    DIS203_SYSTEM_DATA_LENGTH_TYPE         system_data_length; 
    DIS203_NUMBER_OF_BEAMS_TYPE            number_of_beams; 
    uint16                            _unused_43;  /* padding */
    DIS203_EMITTER_SYSTEM_RECORD           emitter_system; 
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD location_with_respect_to_entity; 
    DIS203_EMISSION_BEAM_RECORD            beams[1]; 
} DIS203_EMISSION_SYSTEM_RECORD; 

typedef struct {
    DIS203_PDU_HEADER_RECORD           pdu_header; 
    DIS203_ENTITY_ID_RECORD            entity_id; 
    DIS203_EVENT_ID_RECORD             event_id; 
    DIS203_STATE_UPDATE_INDICATOR_TYPE state_update_indicator; 
    DIS203_NUMBER_OF_SYSTEMS_TYPE      number_of_systems; 
    uint16                        _unused_44;  /* padding */
    DIS203_EMISSION_SYSTEM_RECORD      systems[1]; 
} DIS203_EMISSION_PDU; 


/*
 * Laser PDU (section 5.4.6.3, pages 143-144) 
 * 
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD        pdu_header; 
    DIS203_ENTITY_ID_RECORD         lasing_entity_id; 
    DIS203_CODE_NAME_TYPE           code_name; 
    DIS203_ENTITY_ID_RECORD         lased_entity_id; 
    uint8                      _unused_45;  /* padding */
    DIS203_LASER_CODE_TYPE          laser_code; 
    DIS203_POWER_TYPE               laser_power; 
    DIS203_LASER_WAVE_LENGTH_TYPE   laser_wave_length; 
    DIS203_VECTOR_RECORD            laser_spot_with_respect_to_entity; 
    DIS203_WORLD_COORDINATES_RECORD laser_spot_location; 
} DIS203_LASER_PDU; 


/*
 * Transmitter pdu. (section 5.4.7.2, pages 152-156)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD               pdu_header; 
    DIS203_ENTITY_ID_RECORD                entity_id; 
    DIS203_RADIO_ID_TYPE                   radio_id; 
    DIS203_ENTITY_TYPE_RECORD              radio_entity_type; 
    DIS203_RADIO_TRANSMIT_STATE_TYPE       transmit_state; 
    DIS203_RADIO_INPUT_SOURCE_TYPE         input_source; 
    uint16                            _unused_46;  /* padding */
    DIS203_WORLD_COORDINATES_RECORD        antenna_location; 
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD relative_antenna_location; 
    DIS203_ANTENNA_PATTERN_TYPE_TYPE       antenna_pattern_type; 
    DIS203_ANTENNA_PATTERN_LENGTH_TYPE     antenna_pattern_length; 
    DIS203_TRANSMITTER_FREQUENCY_TYPE      frequency; 
    DIS203_FREQUENCY_BANDWIDTH_TYPE        transmit_frequency_bandwidth; 
    DIS203_POWER_TYPE                      power; 
    DIS203_MODULATION_TYPE_RECORD          modulation_type; 
    DIS203_CRYPTO_SYSTEM_TYPE              crypto_system; 
    DIS203_CRYPTO_KEY_ID_TYPE              crypto_key_id; 
    DIS203_MODULATION_LENGTH_TYPE          length_of_modulation_parameters; 
    unsigned                                  _unused_47                       : 24; 
     /* padding */                      
     /* NOTE: The protocol includes modulation and antenna parameters */
         /*       at the end of this packet, but does not specify their type. */

} DIS203_TRANSMITTER_PDU; 


/*
 * Signal pdu. (section 5.4.7.3, pages 157-159)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD  pdu_header; 
    DIS203_ENTITY_ID_RECORD   entity_id; 
    DIS203_RADIO_ID_TYPE      radio_id; 
    DIS203_ENCODING_SCHEME    encoding_scheme; 
    uint16               _unused_48;  /* padding */
    DIS203_SAMPLE_RATE_TYPE   sample_rate; 
    DIS203_SIGNAL_LENGTH_TYPE length; 
    DIS203_SAMPLES_TYPE       samples; 
    unsigned                     data[1]; 
} DIS203_SIGNAL_PDU; 


/*
 * Receiver PDU (section 5.4.7.4, pages 160-161)
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD   pdu_header; 
    DIS203_ENTITY_ID_RECORD    entity_id; 
    DIS203_RADIO_ID_TYPE       radio_id; 
    DIS203_RECEIVER_STATE_TYPE receiver_state; 
    uint16                _unused_49;  /* padding */
    DIS203_POWER_TYPE          received_power; 
    DIS203_ENTITY_ID_RECORD    transmitter_entity_id; 
    TRANSMITTER_RADIO_ID_TYPE  transmitter_radio_id; 
} DIS203_RECEIVER_PDU; 



#endif
