
/*
 * FILE:  p_faad.h 
 *
 * Created on Fri Apr 26 13:42:14 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_faad_included
#define _p_faad_included



/*
 * File:     p_faad.drn
 * Contents: The FAADS vehicle protocol
 * System:   simnet development
 * Created:  27-Jun-1990
 * Author:   tshannon
 * Remarks:
 * History:
 *	6/27/90	Shannon		Created from old FAADS protocol to 6.x drn
 *
 * $RCSfile: p_faad.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 * COPYRIGHT 1989 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 */

#include <stdtypes.h>
#include <basic.h> 

/*
 * Versions of the FAADS protocol:
 */

#define FAADSProtocolVersionJun90 1  

typedef uint8    FAADSProtocolVersion; 


/*
 * The current version
 */

#define FAADSProtocolVersionCurrent           1

/*
 * Various kinds of FAADS PDUs:
 */

#define targetEngageROPDUKind   1    
#define targetDisengageROPDUKind 2   
#define targetVisibilityPDUKind 3    
#define targetHandoffEOPDUKind  4    
#define targetFireCommandPDUKind 5   
#define targetStatusPDUKind     6    
#define alertStatusPDUKind      7    
#define fireUnitStatePDUKind    8    
#define fireUnitProbePDUKind    9    
#define radiatingStatePDUKind   10   
#define reloadRequestPDUKind    11   
#define roTargetFireCommandPDUKind 12
#define roAlertStatusPDUKind    13   

typedef uint8    FAADSPDUKind; 


/*
 * Different IFF modes:
 */

#define classifiedFriendByIFF   0    
#define classifiedFriendByUser  1    
#define classifiedFoe           2    
#define classifiedUnknown       3    

typedef uint8    IFFMode; 


/*
 * Following are primarily used by RDDS
 */

#define battalionIDIrrelevant   0
#define unitIDIrrelevant        0

/*
 * PDUs for communication between vehicle and radar simulators.
 */

typedef struct {
    VehicleID   vehicleID;           
    uint16      battalion;           
    uint16      unit;                
} FAADSVehicle; 


/*
 * TargetEngage indicates a unit has started tracking a target.  Sent from
 * EO to radar.
 */

typedef struct {
    FAADSVehicle faadsVeh;           
    VehicleID    targetID;            /* Identity of target vehicle */
} TargetEngageROVariant; 


/*
 * TargetDisengage indicates a unit has ceased tracking/engaging a target.
 * Sent from EO to radar.
 */

typedef struct {
    FAADSVehicle faadsVeh;           
    VehicleID    targetID;            /* Identity of target vehicle */
} TargetDisengageROVariant; 


/*
 * RadiatingState PDU's are sent periodically to let the vehicle
 * know we're radiating (used to disable reloading)
 */

typedef struct {
    FAADSVehicle faadsVeh;           
    uint16       radiatingState;      /* TRUE - we're radiating */
                                      /* FALSE - we're not */
} RadiatingStateVariant; 


/*
 * ReloadRequests are sent from the vehicle to the radar to let the
 * radar know it's not allowed to radiate until reloading is done
 * (which is told by the veh_status PDU)
 */

typedef struct {
    FAADSVehicle faadsVeh;           
} ReloadRequestVariant; 

#define MAX_VISIBLE_TGTS        48 /* maximum number of visible vehicles */
                                   /* that can fit in TargetVisiblityVariant */

/* Target Visibility PDUs are sent from the radar to the EO to give */
/* line-of-sight information on targets within a 12 km radius: */
typedef struct {
    FAADSVehicle faadsVeh;           
    uint16       numVisibleTargets;  
    VehicleID    targetIDs[MAX_VISIBLE_TGTS]; 
} TargetVisibilityVariant; 


/*
 * Target Handoff EO PDUs are sent from the Radar to the EO to indicate
 * which target should be engaged:
 */
typedef struct {
    FAADSVehicle            faadsVeh; 
    uint16                  initialHandoff;  /* TRUE if first time for targ */
    VehicleID               targetID;  /* id of target vehicle */
    VehicleID               sourceID;  /* id of source vehicle */
    SIMNET_WorldCoordinates location;  /* location of target */
    SIMNET_WorldCoordinates sourceLocation;  /* location of detecting veh */
    VelocityVector          velocity;  /* velocity of target */
    IFFMode                 IFF;      /* current IFF of target */
    uint8                   _unused_3;  /* padding */
    uint16                  helo;     /* TRUE if helo, FALSE o.w. */
} TargetHandoffEOVariant; 


/*
 * The following constants are the fire commands of the TargetFireCommand
 */

#define CEASE_ENGAGE_CMD    0 /* End current engagement, no missile effect */
#define CEASE_FIRE_CMD      1 /* Withhold fire, continue to track */
#define HOLD_FIRE_CMD       2 /* Divert launched missile, continue to track */
typedef uint8 FireCommands; 


/*
 * TargetFireCommand PDU's are sent from the Radar to the EO to indicate
 * that tracking or fire status is to be altered.
 */

typedef struct {
    FAADSVehicle faadsVeh;           
    VehicleID    receivingVehID;      /* id of vehicle receiving cmd */
    FireCommands fireCommand;        
    uint8        _unused_4;           /* padding */
} TargetFireCommandVariant; 


/*
 * roTargetFireCommand PDU's are sent from the Radar to the EO to indicate
 * ?something.
 */
typedef struct {
    FAADSVehicle faadsVeh;           
    VehicleID    receivingVehID;      /* id of vehicle receiving cmd */
    FireCommands fireCommand;        
    uint8        _unused_5;           /* padding */
} ROTargetFireCommandVariant; 


/*
 * PDUs used by radar simulators to change the track information. 
 * TargetStatusPDU's are sent to update each fire unit's position, 
 * and appropriate target information according to the fire unit's 
 * netCommoRole. 
 */

#define MAX_TARGETS_TO_TRANSMIT 4 /* max num of targets that may  */
                                  /* be transmitted over the net  */
typedef struct {
    SIMNET_XYCoordinates position;   
    VehicleID            vehicleID;  
    uint16               jamming;    
    IFFMode              IFF;        
    unsigned             _unused_6_7 : 24;  /* padding */
    uint32               _unused_6;   /* padding */
} TargetEntry; 

typedef struct {
    FAADSVehicle faadsVeh;           
    uint16       myCommoMode;         /* commo role of sender */
    uint16       numTargets;          /* # in target list  */
    uint16       _unused_8;           /* padding */
    TargetEntry  targetList[MAX_TARGETS_TO_TRANSMIT]; /* list of targets detected */
    TargetEntry  targetEngaged;       /* target unit is engaging */
} TargetStatusVariant; 

 /* The following constants are the alert status levels:  */
#define RED_ALERT               0     /* Fire when ready, Gridley */
#define YELLOW_ALERT            1     /* Fire only upon command */
#define WHITE_ALERT             2     /* Hold fire */
typedef uint8 AlertLevels; 

 /* AlertStatusPDUs are used to set the current alert status level.  */
 /* Master sends to connected units or Radar may sent to EO in local mode.  */
typedef struct {
    FAADSVehicle faadsVeh;           
    VehicleID    alertVehicle;        /* VehicleID to which status is sent */
                                      /* LOCAL mode, RO->EO sends it own ID */
                                      /* MASTER, uses vehicleIDIrrelevant */
    AlertLevels  alertLevel;         
    uint8        _unused_9;           /* padding */
} AlertStatusVariant; 

 /* ROAlertStatusVariants are used to set the current alert status level.  */
 /* something like AlertSTatus yet different in some way. */
typedef struct {
    FAADSVehicle faadsVeh;           
    VehicleID    alertVehicle;        /* VehicleID to which status is sent */
                                      /* LOCAL mode, RO->EO sends it own ID */
                                      /* MASTER, uses vehicleIDIrrelevant */
    AlertLevels  alertLevel;         
    uint8        _unused_10;          /* padding */
} ROAlertStatusVariant; 


/*
 * FireUnitStatePDU's are used to keep the radar informed of the any
 * state changes which occur as a result of the EO's actions so the radar
 * will know when a fire unit is engaged, ready, etc.
 */

#define fuState_Off             1    
#define fuState_Ready           2    
#define fuState_Search          3    
#define fuState_Tracking        4    
#define fuState_LOS_Reset       5    
#define fuState_MIF             6    
#define fuState_Autonomous      7    

typedef uint8 FireUnitStates; 

typedef struct {
    FAADSVehicle   faadsVeh;           
    VehicleID      targetID;            /* id of engaged vehicle */
    FireUnitStates fuState;           /* Current state of Fire Unit */
    uint8          _unused_11;          /* padding */
} FireUnitStateVariant; 


/*
 * When a Fire Unit sees a FireUnitProbePDU it should broadcast
 * a FireUnitStatePDU.
 */

typedef struct {
    FAADSVehicle faadsVeh;           
} FireUnitProbeVariant; 

typedef struct {
    FAADSProtocolVersion version;    
    FAADSPDUKind         kind;               
    ExerciseID           exercise;             
    uint8                _unused_12_13;         /* padding */
    uint32               _unused_12;            /* padding */
    union {
        TargetEngageROVariant      targetEngageRO;
        TargetDisengageROVariant   targetDisengageRO;
        TargetVisibilityVariant    targetVisibility;
        TargetHandoffEOVariant     targetHandoffEO;
        TargetFireCommandVariant   targetFireCommand;
        TargetStatusVariant        targetStatus;
        AlertStatusVariant         alertStatus;
        FireUnitStateVariant       fireUnitState;
        FireUnitProbeVariant       fireUnitProbe;
        RadiatingStateVariant      radiatingState;
        ReloadRequestVariant       reloadRequest;
        ROTargetFireCommandVariant roTargetFireCommand;
        ROAlertStatusVariant       roAlertStatus;
    } variant;                          
} FAADSPDU; 



#endif
