
/*
 * FILE:  p_gt.h 
 *
 * Created on Fri Apr 26 13:42:14 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_gt_included
#define _p_gt_included

 /* File:     p_gt.drn */
 /* Contents: The GT protocol */
 /* System:   simnet development. */
 /* Created:  16-Aug-1990 */
 /* Author:   kanarick */
 /* Remarks: */
 /* $RCSfile: p_gt.h,v $ */
 /* $Revision: 1.1.1.1 $ */
 /* $State: Exp $ */
 /* COPYRIGHT 1990 BBN Systems and Technologies  */
 /* 10 Moulton Street */
 /* Cambridge, Ma. 02138 */
 /* 617-873-3000 */
#include <stdtypes.h>
#include <basic.h> 
#include <p_assoc.h> 

/*
 * Versions of the simulation protocol:
 */

#define gtProtocolVersionAug90  1    

typedef uint8    GTProtocolVersion; 


/*
 * The current version
 */

#define gtProtocolVersionCurrent 1

/*
 * The different kinds of GT PDUs:
 */

#define gtActivateReqPDUKind    1    
#define gtActivateRspPDUKind    2    
#define gtDeactivateReqPDUKind  3    
#define gtDeactivateRspPDUKind  4    
#define gtErrorPDUKind          5    
#define gtHeartbeatPDUKind      6    
#define gtInitializeMsgPDUKind  7    
#define gtInitializeRspPDUKind  8    
#define gtSimulateMsgPDUKind    9    

typedef uint8    GTPDUKind; 


/*
 * An address for GT's
 */

typedef uint32    GTAddress; 


/*
 * The GTErrorPDU is sent by the GT cig of sim on error condition
 */

typedef struct {
    char errorString1[80];              
} GTErrorVariant; 

typedef uint8 GTDataUnit; 
 /* one byte */

 /*  */
typedef struct {
    uint16      size;                 /* size in bytes */
    uint16      seq_no;               /* sequence number */
    uint16      frame;                /* frame counter */
    unsigned    _unused_5 : 24;       /* padding */
    GTDataUnit  data[1];             
} GTMsgVariant; 

 /* A gt deactivate response */
typedef struct {
    GTAddress controller;            
    uint32    _unused_6_7;           /* padding */
    uint16    _unused_6;            /* padding */
} GTInitMsgRspVariant; 


/*
 * Activate Req for a CIG
 */
typedef struct {
    GTAddress               controller; 
    uint32                  gt_kind; 
    GTAddress               logical; 
    uint32                  _unused_8;  /* padding */
    ExerciseID              exerciseID; 
    ForceID                 force;   
    VehicleID               vehicle; 
    SIMNET_WorldCoordinates view_range; 
} GTActivateReqVariant; 


/*
 * The different possible reasons for deactivating a vehicle
 */

#define gtDeactivateReasonOther 0    
#define gtConnectEnd            1    

typedef uint8 GTDeactivateReason; 


/*
 * The GT Deactivate Req PDU:
 */

typedef struct {
    GTAddress          controller;           
    GTDeactivateReason reason;       
    unsigned           _unused_9   : 24;      /* padding */
} GTDeactivateReqVariant; 


/*
 * The different possible results of trying to activate a gt:
 */

#define gtActivateReqAccepted   0    
#define gtInvalidActivateParameter 1 
#define gtAlreadyActive         2    

typedef uint8 GTActivateResult; 

typedef struct {
    GTActivateResult result;         
    unsigned         _unused_10_11 : 24;    /* padding */
    uint32           _unused_10;            /* padding */
} GTActivateRspVariant; 


/*
 * The different possible results of trying a gt deactivate:
 *
 * Results reported on deactivating a vehicle:
 */

#define gtDeactivateReqAccepted 0    
#define gtInvalidDeactivateParameter 1
#define gtUnexpectedDeactivateReason 2
#define gtVehicleNotActive      3    

typedef uint8 GTDeactivateResult; 


/*
 * A gt deactivate response
 */

typedef struct {
    GTDeactivateResult result;       
    unsigned           _unused_12_13 : 24;    /* padding */
    unsigned           _unused_12 : 24;       /* padding */
} GTDeactivateRspVariant; 


/*
 * A Heartbeat PDU
 */

typedef struct {
    uint16 frame;               
    uint16 seq_no;              
    uint16 errors_ahead;        
    uint16 errors_behind;       
    uint16 errors_seq;          
    uint16 _unused_14_15;        /* padding */
    uint32 _unused_14;           /* padding */
} GTHeartbeatVariant; 


/*
 * The GT PDU:
 */

typedef struct {
    GTProtocolVersion version;       
    GTPDUKind         kind;                 
    uint16            _unused_16_17;        /* padding */
    uint32            _unused_16;            /* padding */
    GTAddress         to;                   
    GTAddress         from;                 
    union {
        GTActivateReqVariant   activateReq;
        GTActivateRspVariant   activateRsp;
        GTDeactivateReqVariant deactivateReq;
        GTDeactivateRspVariant deactivateRsp;
        GTErrorVariant         error;
        GTHeartbeatVariant     heartbeat;
        GTMsgVariant           initializeMsg;
        GTInitMsgRspVariant    initializeRsp;
        GTMsgVariant           simulateMsg;
    } variant;                          
} GTPDU; 



#endif
