
/*
 * FILE:  p_ivis.h 
 *
 * Created on Wed Apr 17 10:52:52 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_ivis_included
#define _p_ivis_included



/*
 *  This file defines the representation of IVIS (and CITV) Protocol Data
 *  Units.
 *
 *  Copyright (c) 1990.  All rights reserved.
 *  BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 *  History:
 *    09/07/90  Schwartz	put Ivis prefix on constants
 *    04/25/90	Pope		updated for compatibility with SIMNET 6.0
 *    10/15/89  Pope    	created
 *
 * $RCSfile: p_ivis.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

#include <stdtypes.h>
#include <cvcc_innotr.h> 
#include <p_radio.h> 

/*
 * Versions of the IVIS protocol:
 */

#define cvcc_ProtocolVersionApr90 1  
#define cvcc_ProtocolVersionDec91 2  

typedef uint8    cvcc_ProtocolVersion; 


/*
 * The current version
 */

#define cvcc_ProtocolVersionCurrent           2

/*
 * Various kinds of CVCC C2 protocol PDUs:
 *
 * NOTE:  New kinds of PDUs should be added to the end of the list so that
 * applications which do not use them do not HAVE to be recompiled.
 */

#define citvOrientationPDUKind  1    
#define citvInstrumentationPDUKind 2 
#define ivis_BroadcastPDUKind   3    
#define ivis_TransmitRequestPDUKind 4
#define ivis_TransmitResponsePDUKind 5
#define ivis_ReceivePDUKind     6    
#define ivis_InstrumentationPDUKind 7
#define bntoc_InstrumentationPDUKind 8
#define cvcc_ExecutionControlPDUKind 9
#define cvcc_InnovativeTrainingPDUKind 10
#define bntoc_PointToPointPDUKind 11 

typedef uint8    cvcc_PDUKind; 


/*
 * Format of a CITV Orientation PDU:
 * 6 + 2 + 4 + 4 = 16 bytes
 */

typedef struct {
    VehicleID vehicle;                /* identity of vehicle */
    uint16    _unused_47;             /* padding */
    Angle     azimuth;                /* azimuth of CITV */
    Angle     elevation;              /* elevation of CITV */
} CITVOrientationVariant; 


/*
 *  PDUs for communicating messages directly among cvcc applications
 *
 *
 * Format of an IVIS Broadcast PDU:
 * 6 + 2 + 8 + 1024 = 1040 bytes
 */

typedef struct {
    VehicleID           vehicle;      /* identity of sender */
    uint16              _unused_48;   /* padding */
    cvcc_RadioNetworkID network;      /* network carrying message */
    cvcc_Msg            message;      /* message */
} ivis_BroadcastVariant; 


/*
 *  PDUs for communication between an IVIS and radio simulator.
 *
 *
 * Maximum number of explicitly specified message recipients
 */

#define ivis_MaxRecipients      2

/*
 * Format of an IVIS Transmit Request PDU:
 * 8 + 8 + 4 + 2*6 + 1 + 1 + 1/8 + 1/8 + 1-3/4 + 1024 = 1060 bytes
 */

typedef struct {
    RadioID             radio;        /* identity of radio transmitter */
    cvcc_RadioNetworkID network;      /* network carrying message */
    uint32              serialNumber;  /* request serial number */
    VehicleID           recipient[ivis_MaxRecipients]; /* failing recipients */
    uint8               priority;     /* priority of communication */
    uint8               numberRecipients;  /* number of specified receivers */
    unsigned            broadcast   : 1;/* whether meant for all receivers */
    unsigned            acknowledge : 1; /* whether to be acked by receivers */
    unsigned            _unused_49  : 14;  /* padding */
    cvcc_Msg            message;      /* message */
} ivis_TransmitRequestVariant; 


/*
 * Format of an IVIS Transmit Response PDU:
 * 8 + 8 + 4 + 1/8 + 7/8 + 1 + 2 + 2*6 = 36 bytes
 */

typedef struct {
    RadioID             radio;        /* identity of radio transmitter */
    cvcc_RadioNetworkID network;      /* network carrying message */
    uint32              serialNumber;  /* request serial number */
    unsigned            transmitted : 1; /* whether message was transmitted */
    unsigned            _unused_50  : 7;  /* padding */
    uint8               numberRecipients;  /* number of failing receivers */
    uint16              _unused_51;   /* padding */
    VehicleID           recipient[ivis_MaxRecipients]; /* failing recipients */
} ivis_TransmitResponseVariant; 


/*
 * Format of an IVIS Receive PDU:
 * 8 + 8 + 1024 = 1040 bytes
 */

typedef struct {
    RadioID             radio;        /* identity of radio receiver */
    cvcc_RadioNetworkID network;      /* network carrying message */
    cvcc_Msg            message;      /* message */

} ivis_ReceiveVariant; 


/*
 *  During the execution of a simulation, certain exercise control commands
 *  are sent.  These commands have meaning in a simulated environment, but
 *  have no meaning in a battle field.  The management commands specific to
 *  the cvcc environment are checkpoint commands, the shutdown command, and
 *  host to duty position and callsign mapping.
 *  (For grins, could you imagine a shutdown command in a real war?)
 */

#define cvcc_CheckpointControl  0    
#define cvcc_Shutdown           1    
#define cvcc_MappingControl     2    

typedef uint8 cvcc_ExecutionControlType; 


/*
 *  Checkpoint messages are for saving and recovering the state of a 
 *  simulation.  There are three checkpoint actions:  save (save the
 *  state of the world), recover (set the environment to a previously
 *  saved state), and delete (delete a checkpoint file).
 *						-- 1 + 15 = 16 bytes
 *  Checkpoint actions
 */

#define cvcc_SaveCheckpoint     0     /* create checkpoint file */
#define cvcc_RecoverCheckpoint  1     /* recover checkpoint file */
#define cvcc_DeleteCheckpoint   2     /* delete checkpoint file */
typedef uint8 cvcc_CheckpointAction; 
 /*  Description of a checkpoint command: */
 /*  Please note that a checkpoint name can have at most 14 characters in */
 /*  it.  The name must be terminated by a null character. */
 /*  1 + 15 = 16 bytes */
#define cvcc_MaxCheckpointName  15 /* max length of ivis checkpoint file */
typedef struct {
    cvcc_CheckpointAction action;     /* checkpoint or recover */
    char          name[cvcc_MaxCheckpointName];  /* name of checkpoint file */
} cvcc_Checkpoint; 


/*
 *  Types of cvcc hosts (needed in the mapping message:		1 byte
 */

#define cvcc_BntocAppl          0    
#define cvcc_IvisAppl           1    
#define cvcc_SafAppl            2    
#define cvcc_SendAppl           3    

typedef uint8 cvcc_ApplType; 

#define cvcc_NumberApplications 4

/*
 *  An ivis vehicle and a bntoc ws can be identified by a number of different
 *  things.  First, there is the name of the host.  Second, there is the
 *  site/host/vehicle of either the host or the simulator.  Third, there is
 *  the TOE of the tank or soldier using the bntoc ws.  Fourth, there is
 *  the callsign of the tank or bntoc ws.  There is a one-to-one mapping
 *  from the host name to call sign in bntoc and ivis, and a one-to-many
 *  mapping from the host name to call sign in SAF.  There is a one-to-many 
 *  mapping from TOE to call sign.  And there is a one-to-many mapping from 
 *  host name to site/host/vehicle.  In the military semantic, the players 
 *  need to know the call sign to TOE mappings.  In the software world, the 
 *  hosts need to know the host to call sign mappings.  The following
 *  mapping information gives the information ivis and bntoc need so that
 *  they can do their thing.  The information here, and the information
 *  in the status message is a melding of three concepts into two different
 *  packet types.  To be cleaner about this information, duty position
 *  to callsign mappings beint32 with tasking information and not here.
 *  The only concept that really beint32s here is host to duty position
 *  or host to callsign. 			-- 12 + 6 + 1 + 33 = 52 bytes 
 */

#define cvcc_MaxHostname        33
typedef struct {
    cvcc_Tasking  dutyPosition;       /* the TO&E of a unit */
    cvcc_CallSign callSign;           /* the call sign of a unit */
    cvcc_ApplType applType;           /* which cvcc application */
    char          hostname[cvcc_MaxHostname]; 
} cvcc_MappingUnit; 
 /* 						-- 2 + 2 + 20*52 = 1044 bytes */
#define cvcc_MaxMappingUnits    20
typedef struct {
    uint16           n;                /* the number of mapping units */
    unsigned         _unused_52 : 32;  /* padding */
    cvcc_MappingUnit map[cvcc_MaxMappingUnits];  /* the mappings */
} cvcc_Mapping; 

 /*  Execution Control command */
 /*  1 + 3 + max(16,1,1044) = 1048 bytes */
typedef struct {
    cvcc_ExecutionControlType etype;  /* type of command */
    unsigned                  _unused_53 : 24;       /* padding */
    union {
        cvcc_Checkpoint check;
        char            nothing;
        cvcc_Mapping    map;
    } u;                                
} cvcc_ExecutionControl; 


/*
 * Format of an BnTOC Point To Point PDU:
 * 6 + 2 + 8 + 1024 = 1040 bytes
 *
 * BnTOC WS's send messages to other WS's in the TOC using this varient.
 * The .recipients is a bit field containing the role of all the BnTOC
 * workstations to which this message is to be sent.  Note, this point to point
 * addressing scheme is intended to be used for sending messages between the
 *  BnTOC Workstations only.  BnTOC do not use the point to point addressing
 *  found in the ivis_TransmissionRequestVarient because BnTOC's do not
 *  necessarily use SINCGARS radios to communicate message among themselves.
 * 
 */

typedef struct {
    bntoc_MessageDestination recipients;  /* the recipients of message */
    uint16                   _unused_54;           /* padding */
    cvcc_Msg                 message;               /* message */
} bntoc_PointToPointVariant; 


/*
 * Definition of IVIS Protocol Data Unit.
 * 1 + 1 + 1 + 1 + max(16,24,1040,1060,36,1040,40,628,1048,1060) = 1064 bytes
 */

typedef struct {
    cvcc_ProtocolVersion version;     /* the version of the protocol */
    cvcc_PDUKind         kind;        /* the type of pdu */
    ExerciseID           exercise;    /* the exercise ID */
    uint8                _unused_55;  /* padding */
    union {
        CITVOrientationVariant       citvOrient;
        CITVInstrumentationVariant   citvInstrument;
        ivis_BroadcastVariant        broadcast;
        ivis_TransmitRequestVariant  transmitRequest;
        ivis_TransmitResponseVariant transmitResponse;
        ivis_ReceiveVariant          receive;
        ivis_InstrumentationVariant  ivisInstr;
        bntoc_InstrumentationVariant bntocInstr;
        cvcc_ExecutionControl        control;
        cvcc_InnovativeTraining      innotr;
        bntoc_PointToPointVariant    bntocPTP;
    } u;                                
} IVIS_PDU; 



#endif
