
/*
 * FILE:  p_logger.h 
 *
 * Created on Wed Apr 17 10:52:54 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_logger_included
#define _p_logger_included

 /* File:     pro_logger.drn */
 /* Contents: The logger/pvd protocol */
 /* System:   simnet development. */
 /* Created:  18-Dec-89 */
 /* Author:   mgreess */
 /* Remarks: */
 /* $RCSfile: p_logger.h,v $ */
 /* $Revision: 1.1.1.1 $ */
 /* $State: Exp $ */
 /* COPYRIGHT 1989 BBN Systems and Technologies Corporation */
 /* 10 Moulton Street */
 /* Cambridge, Ma. 02138 */
 /* 617-873-3000 */
#include <stdtypes.h>
#include <basic.h> 

/*
 * OVERVIEW
 *	This protocol is intended to handle the communication
 *	needs of PVD and Logger when the Logger is being controlled
 *	by from the PVD (i.e. when the Logger is running in 'server'
 *	mode.
 *
 *	Logger/PVD PDUs share the CMC network with the standard simulation 
 *	protocol PDUs.  However, Logger/PVD communications take place on
 *	a unique multi-caste address consisting of the following:
 *
 *		MulticastGroupID	0
 *		AssociationUserProtocol	loggerProtocolNumber
 *
 *	There are three phases of Logger/PVD communications:
 *		1) Connect Phase
 *		2) Command Phase
 *		2) Disconnect Phase
 *	During the first phase, the Logger will respond to all PDU's
 *	arriving on the above Multi-cast group.  Thereafter it responds
 *	only to PDU's which are explicitely directed to it.
 *
 *	In the discussion of the Logger/PVD protocol outlined below,
 *	it is assumed that the Logger and PVD will be using the services
 *	of the Association Protocol for their communications.  All PDU's
 *	are assumed to be sent via the POINT_TO_POINT_DATAGRAM service
 *	unless otherwise stated.
 *
 *
 * MulticastGroupId loggers and pvds use to communicate.
 *
 */

#define loggerAcceptAll         256
#define loggerMaxCommentSz      24
#define loggerMaxErrorStringSz  128
#define loggerMaxExerciseIDs    8
#define loggerMaxFileNameSz     64
#define loggerMaxFileVolumes    2
#define loggerMaxHostNameSz     32
#define loggerMaxProjIDSz       7
#define loggerMulticastGroupID  0

/*
 * PDUKIND loggerAvailRequestPDUKind 
 * OVERVIEW
 *	The loggerAvailRequestPDUKind is sent by the PVD while preparing to
 *	connect to a Logger.  It is sent on the Logger/PVD multi-cast address
 *	using the broadcast Datagram (as opposed to the ptpDatagram) service.
 *
 *	All Logger which are operating in server mode and which are not already
 *	connected to a PVD must respond with a 'loggerAvailReplyPDUKind' PDU.
 *	Both Request and the Reply are sent using the ptpDatagram service.
 *	The PVD must wait a reasonable amount of time for the Loggers
 *	to respond before assuming there are no other Loggers out there.
 *
 * PDUKIND loggerConnectRequestPDUKind
 * OVERVIEW
 *	The PVD sends a 'loggerConnectPDUKind' to the particular Logger
 *	to which it wants to connect using the ptpDatagram service of the
 *	Association Protocol.
 *
 *	The Logger responds with either a 'loggerAckPDUKind' or a 
 *	'loggerNakPDUKind' using the ptpDatagrame service.
 *
 * PDUKIND loggerCommandRequestPDUKind
 * OVERVIEW
 *	The loggerCommandRequestPDUKind is a ptpDatagram PDU sent by the
 *	PVD to the Logger to which it is connected Logger to specify
 *	runtime commands.  The following runtime commands
 *	are currently defined in the PVD/Logger protocol.
 *		START
 *		CONTINUE
 *		SUSPEND
 *		STOP
 *		SEEK_RELATIVE
 *		PLAY_SPEED
 *
 * PDUKIND loggerStatusRequestPDUKind
 * OVERVIEW
 *	The PVD sends a 'loggerStatusRequestPDUKind' to the particular Logger
 *	to which it is connected to gather information about its current
 *	operating status.
 *	The PVD uses the ptpDatagram service of the Association Protocol.
 *
 *	The Logger responds with a 'loggerStatusReplyPDUKind' using the
 *	ptpDatagram service.
 *
 * PDUKIND loggerDisconnectRequestPDUKind
 * OVERVIEW
 *	The PVD sends a 'loggerDisconnectPDUKind' to the particular Logger
 *	to which it is connected to terminate its connection.
 *	The PVD uses the ptpDatagram service of the Association Protocol.
 *
 *	The Logger responds with a 'loggerAckPDUKind' using the
 *	ptpDatagram service.
 *
 *
 *
 * PDUKIND loggerAckPDUKind
 * OVERVIEW
 *	The Logger uses this PDU to reply to successful PVD requests.
 *
 * PDUKIND loggerNakPDUKind
 * OVERVIEW
 *	The Logger uses this PDU to reply to failed PVD requests.
 *
 * PDUKIND loggerAvailReplyPDUKind
 * OVERVIEW
 *	The Logger uses this PDU to reply to PVD Availability Requests.
 *
 * PDUKIND loggerStatusReplyPDUKind
 * OVERVIEW
 *	The Logger uses this PDU to reply to PVD Status Requests.
 *
 *
 *
 * PDUKIND loggerInformationPDUKind
 * OVERVIEW
 *	The Logger uses this PDU to communicate unsolicited information
 *	to the PVD using the ptpDatagram service.
 *
 * PDUKIND loggerClockTickPDUKind
 * OVERVIEW
 *	The Logger uses this PDU to communicate timing information to the PVD
 *	using the ptpDatagram service.
 *
 *
 * Kinds of loggerProtocol PDUs
 *
 *
 * loggerProtocol PDUs which the PVD sends requests to the Logger.
 *
 */

#define loggerAvailRequestPDUKind 1  
#define loggerConnectRequestPDUKind 2
#define loggerCommandRequestPDUKind 3
#define loggerStatusRequestPDUKind 4 
#define loggerDisconnectRequestPDUKind 5
     /* loggerProtocol PDUs with which the Logger replies to PVD requests. */
#define loggerAckPDUKind        6    
#define loggerNakPDUKind        7    
#define loggerAvailReplyPDUKind 8    
#define loggerStatusReplyPDUKind 9   
     /* loggerProtocol PDUs which the Logger sends to the PVD unsolicited info. */
#define loggerInformationPDUKind 10  
#define loggerClockTickPDUKind  11   
#define loggerLastPDUKind       12   

typedef uint8    LoggerPDUKind; 


/*
 * The following defines the set of possible file mediums for Logger files.
 *
 */

#define diskMedium              1    
#define nineTrackTapeMedium     2    
#define cassetteTapeMedium      3    

typedef unsigned         LoggerMedium; 

typedef struct {
    uint8  year;                 /* Offset from 1900 */
    uint8  month;               
    uint8  day;                 
    uint8  hour;                
    uint8  minute;              
    uint8  second;              
    uint8  tzone;                /* Hours WEST of GMT.  -1 if unknown. */
    uint8  dstime;               /* True if daylight savings time. */
} LoggerTime; 

typedef struct {
     /* Recording Only:	Base time of an exercise. */
    LoggerTime  starttime;           
     /* Recording to Disk Only:	Use contiguous disk files. */
    unsigned       contig                : 1;  /* Boolean */
     /* Recording to Disk:	When the end of file is reached, */
     /* 			return to the start and continue recording. */
     /* Play back from Disk:	Follow the loop pointer to the logical start */
     /* 			of file.  When the end of file is reached, */
     /* 			go to the beginning of the file and continue */
     /* 			playing back 'til loop pointer is reached. */
     /* 			(i.e. the loop pointer marks the logical */
     /* 			beginning and end of file. */
     /* Record and Play Back to Tape: */
     /* 			Swapping between available tape drives. */
    unsigned       looping               : 1;  /* Boolean */
     /* True => recording; False => Play back. */
    unsigned       recording             : 1;  /* Boolean */
    unsigned       promiscuous           : 1;  /* Boolean */
    unsigned       racal                 : 1;  /* Boolean */
    unsigned       medium                : 3;  /* LoggerMedium */
     /* Count of the 'flnm' array entries being used. */
    uint8  nvolumes;            
     /* Recording:	exercise ids is an array of 1-byte integers in */
     /* 		the range 1 .. 255.  A value of 'loggerAcceptAll' */
     /* 		indicates that all exercise ids are to be accepted. */
    int16          num_exercises;       
    char           exercises[loggerMaxExerciseIDs]; 
     /* Recording:	A value of 'loggerAcceptAll' indicates that all  */
     /* 		standard or non-standard protocols are accepted. */
     /* 		A value of 0 indicates that none are to be accepted. */
     /* 		For the time being, any value besides  */
     /* 		'loggerAcceptAll' will be interprited as 0. */
    int16          num_stdprotocols;    
    int16          num_nstdprotocols;   
     /* Recording to Contiguous Disk files:	size in bytes  */
     /* Recording to Tape:	1=>600 ft; 2=>1200 ft; 3=>2400 ft reels */
    uint32  size[loggerMaxFileVolumes]; 
     /* Disk:	Names of separate Unix files 'concatenated to  */
     /* 		together to create the logical SIMNET data file.  */
     /* 		A max of loggerMaxFileVolumes files can be involved. */
     /* Tape:	Names of tape drives being used. */
     /* 		Max of loggerMaxFileVolumes drives can be involved. */
    char           flnm[loggerMaxFileVolumes][loggerMaxFileNameSz]; 
     /* Header info */
    char           projID[loggerMaxProjIDSz]; 
    char           comment[loggerMaxCommentSz]; 
     /* Used when the logger is being initialized for Recording. */
     /* Controls whether packets belonging to the Voice Protocol Type */
     /* are recorded. */
    unsigned       rcdVoiceProtocolType  : 1;  /* Boolean */
     /* Used when the logger is being initialized for Recording. */
     /* Controls whether packets belonging to the Simulation Protocol Type */
     /* are recorded. */
    unsigned       rcdSimProtocolType    : 1;  /* Boolean */
     /* Used when the logger is being initialized for Play Back. */
     /* Controls whether packets belonging to the Stealth Protocol of the */
     /* Simulation Protocol Type are recorded. */
    unsigned       playStealthProtocol   : 1;  /* Boolean */
    unsigned       _unused_3             : 5;  /* padding */
} LoggerStartRTC; 

typedef struct {
    int32 relative_curloc;               
    int32 seconds;                       
} LoggerSeekRTC; 

typedef struct {
    int32 factor;                        
} LoggerSpeedRTC; 


/*
 * The following defines the set of possible runtime commands which the
 * PVD can send to the Logger.
 *
 */

#define startRTCType            1    
#define continueRTCType         2    
#define suspendRTCType          3    
#define stopRTCType             4    
#define seekRelativeRTCType     5    
#define playSpeedRTCType        6    

typedef uint8 LoggerRTCType; 

typedef struct {
    LoggerRTCType command;           
    unsigned      _unused_4 : 24;        /* padding */
    union {
        LoggerStartRTC startInfo;
        LoggerSeekRTC  seekInfo;
        LoggerSpeedRTC playInfo;
    } info;                             
} LoggerCommandVariant; 

#define loggerNoActivity        0    
#define loggerRecording         1    
#define loggerPlayback          2    
#define loggerCopying           3    

typedef unsigned  LoggerActivity; 

#define loggerInteractive       1    
#define loggerServer            2    

typedef unsigned   LoggerOperation; 

#define loggerDisconnected      0    
#define loggerConnected         1    
#define loggerSuspended         2    
#define loggerActive            3    

typedef unsigned LoggerState; 

typedef struct {
    unsigned  available  : 1;            /* Boolean */
    unsigned  _unused_5  : 7;            /* padding */
    char      loggerID[loggerMaxHostNameSz]; 
} LoggerAvailReplyVariant; 

typedef struct {
    unsigned          activity    : 2;   /* LoggerActivity */
    unsigned          oper        : 2;   /* LoggerOperation */
    unsigned          state       : 2;   /* LoggerState */
    unsigned          racal       : 1;   /* Boolean */
    unsigned          _unused_6   : 1;   /* padding */
    /* Time factor */
    uint8             speed;            
    uint16            _unused_7;         /* padding */
    /* Start time of the exercise */
    LoggerTime        start;           
    /* Time offset in ms from begin */
    uint32            offset;         
    /* Num of packets handled since start */
    uint32            packets;          
    char              file_id[loggerMaxFileNameSz]; 
    char              project_id[loggerMaxProjIDSz]; 
    char              comment[loggerMaxCommentSz]; 
    uint8             _unused_8;         /* padding */
} LoggerStatusReplyVariant; 

typedef struct {
    unsigned  is_error      : 1;         /* Boolean */
    unsigned  _unused_9     : 7;         /* padding */
    char      error_string[loggerMaxErrorStringSz]; 
} LoggerInformationVariant; 

typedef struct {
    LoggerTime starttime;            
    uint32     offset;               
} LoggerClockTickVariant; 

#define loggerProtocolVersionAug89 1 
#define loggerProtocolVersionJan90 2 

typedef uint8 LoggerProtocolVersion; 

#define loggerProtocolVersionCurrent           2
typedef struct {
    LoggerProtocolVersion version;   
    LoggerPDUKind         kind;          
    uint16                _unused_10;     /* padding */
    SimulationAddress     destination;   
    union {
        LoggerCommandVariant     commandReq;
        LoggerAvailReplyVariant  availReply;
        LoggerStatusReplyVariant statusReply;
        LoggerInformationVariant information;
        LoggerClockTickVariant   clockTick;
    } variant;                          
} LoggerPDU; 



#endif
