
/*
 * FILE:  p_mgmt.h 
 *
 * Created on Fri Apr 26 13:42:15 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_mgmt_included
#define _p_mgmt_included



/*
 * File:     p_mgmt.drn
 * Contents: The management protocol
 * System:   simnet development.
 * Created:  21-Nov-89
 * Author:   kanarick
 * Remarks:
 * History:
 *    28 Nov 89	kanarick	Added errorNumberIrrelevant
 *    10 Jan 90	kanarick	Added ShowEffectVariant and data elements
 *    14 Jan 90	kanarick	Defined mgmtProtocolVersionJan90
 *
 * $RCSfile: p_mgmt.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 * COPYRIGHT 1989 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 */

#include <stdtypes.h>
#include <basic.h> 
#include <p_logger.h> 

/*
 * Versions of the management protocol:
 */

#define mgmtProtocolVersionAug89 1   
#define mgmtProtocolVersionJan90 2   
#define mgmtProtocolVersionJan90Corrected 3

typedef uint8    ManagementProtocolVersion; 

#define mgmtProtocolVersionCurrent           3

/*
 * Various kinds of Management PDUs:
 */

#define equipStatusPDUKind      1     /* Equipment Status PDU */
#define errorReportPDUKind      2     /* Error Report PDU */
#define showEffectPDUKind       3     /* Show Effect PDU */
typedef uint8    ManagementPDUKind; 


/*
 * PDU for reporting equipment status.
 *
 *
 * Status of a component:
 */

#define componentNonexistent    0    
#define componentOperating      1    
#define componentDisabled       2    
#define componentPartiallyOperating 3

typedef uint8    ComponentStatus; 


/*
 * Details of AGPT vehicle simulator status:
 */

typedef struct {
    ComponentStatus hostHealthy;     
    ComponentStatus cabinHealthy;    
    ComponentStatus cigHealthy;      
    ComponentStatus soundHealthy;    
    ComponentStatus apeHealthy;      
} AGPT_Sim_Status; 
 /* Details of the AGPT data logger status */
typedef struct {
    float32          diskUsed;           /* percentage */
    unsigned         hostState   : 2;    /* LoggerActivity */
    unsigned         _unused_11  : 6;    /* padding */
    ComponentStatus hostHealth;      
} AGPT_Data_Logger_Status; 


/*
 * Details of SIMNET vehicle simulator status:
 */

#define max_SIMNET_IDCs         5
typedef struct {
    /* Power supply voltages: */
    float32         voltage12P;            /* +12V level */
    float32         voltage12N;            /* -12V level */
    float32         voltage5;              /* 5V level */
    /* Equipment temperature: */
    float32         temperature;           /* temperature (degrees) */
    /* The existence and status of various simulator components: */
    ComponentStatus hostHealthy;     
    ComponentStatus cigHealthy;      
    ComponentStatus soundHealthy;    
    ComponentStatus idcHealthy[max_SIMNET_IDCs]; 
} SIMNET_Sim_Status; 

#define idcM1DriverHealth       0
#define idcM1TurretHealth       1
#define idcM1AmmoHealth         2
#define idcM2DriverHealth       0
#define idcM2TurretHealth       1

/*
 * Contents of an Equipment Status PDU:
 */

typedef struct {
    SimulatorType simulator;          /* type of equipment reporting */
    uint16        _unused_12;           /* padding */
    union {
        SIMNET_Sim_Status       m1;
        SIMNET_Sim_Status       m2;
        SIMNET_Sim_Status       fred;
        SIMNET_Sim_Status       fwa;
        SIMNET_Sim_Status       faad;
        SIMNET_Sim_Status       stealth;
        SIMNET_Sim_Status       di;
        AGPT_Sim_Status         leo2;
        AGPT_Sim_Status         ulf;
        AGPT_Data_Logger_Status data_logger;
    } status;                           
} EquipStatusVariant; 


/*
 * PDU for reporting unusual conditions.
 *
 *
 * The severity of a report:
 */

#define errorReportInformational 1   
#define errorReportWarning      2    
#define errorReportError        3    
#define errorReportFatal        4    

typedef uint8 ErrorReportSeverity; 


/*
 * Contents of an Error Report PDU:
 */

typedef struct {
    ErrorReportSeverity severity;    
    uint8               _unused_13;           /* padding */
    uint16              number;              
    uint16              textLength;          
    uint16               _unused_14;           /* padding */
    /* to get to longword boundary  */
    char           text[1];             
} ErrorReportVariant; 


/*
 * an irrelevant number
 */

#define errorNumberIrrelevant   0

/*
 * The different types of effects:
 */

#define effectTypeIrrelevant    0    
#define detonateOnGround        1    
#define detonateInAir           2    
#define detonateOnArmor         3    
#define detonateOnWood          4    
#define detonateOnOther         5    
#define detonateOnOurVehicle    6    
#define ourMuzzleFlash          7    
#define otherMuzzleFlash        8    

typedef uint16 EffectType; 


/*
 * The Show Effect PDU
 */

typedef struct {
    ObjectType              projectile; 
    EffectType              effect;  
    uint16                  _unused_15;  /* padding */
    SIMNET_WorldCoordinates effectLocation; 
} ShowEffectVariant; 
 /*  */
 /* Definition of a Management Protocol Data Unit. */

typedef struct {
    ManagementProtocolVersion version; 
    ManagementPDUKind         kind;          
    uint16                    _unused_16_17;        /* padding */
    uint32                    _unused_16;            /* padding */
    union {
        EquipStatusVariant equipStatus;
        ErrorReportVariant errorReport;
        ShowEffectVariant  showEffect;
    } variant;                          
} ManagementPDU; 



#endif
