
/*
 * FILE:  p_mon.h 
 *
 * Created on Wed Apr 17 10:52:55 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_mon_included
#define _p_mon_included



/*
 *  This file defines the representation of Monitor Protocol Data Units.
 *
 * $RCSfile: p_mon.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 *  Copyright (c) 1990.  All rights reserved.
 *  BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 *  History:
 *    07/17/90  Pope    	created from previous pro_mon.h
 *
 *
 * Versions of the Monitor protocol:
 */

#include <stdtypes.h>

#define monitorProtocolVersionJul90 1

typedef uint8    MonitorProtocolVersion; 


/*
 * The current version
 */

#define monitorProtocolVersionCurrent           1

/*
 * Various kinds of Monitor PDUs:
 */

#define monitorQueryPDUKind     1    
#define monitorResponsePDUKind  2    
#define remoteCommandPDUKind    3    
#define shellOpenPDUKind        4    
#define shellStatusPDUKind      5    
#define shellTextPDUKind        6    
#define shellAckPDUKind         7    

typedef uint8    Monitor_PDUKind; 


/*
 *  PDUs that query for the presence of monitors on the network.
 *
 *
 * Format of a Monitor Response PDU:
 */

#define fidoMonitorKind         1    
#define roverMonitorKind        2    

typedef uint8    MonitorKind; 

typedef struct {
    MonitorKind monitorKind;         
    unsigned      _unused_1_2  : 24;     /* padding */
    uint32 _unused_1;             /* padding */
} MonitorResponseVariant; 


/*
 *  PDU for executing a remote command.
 *
 *
 * Format of Remote Command PDU:
 */

typedef struct {
    uint16 length;              
    char   text[1];             
} RemoteCommandVariant; 


/*
 *  PDUs for communicating with a remote shell.
 *
 *
 * Format of Shell Open PDU:
 */

typedef struct {
    uint8     channel;             
    unsigned  _unused_3_4 : 24;     /* padding */
    uint32    _unused_3;            /* padding */
} ShellOpenVariant; 


/*
 * Format of Shell Status PDU:
 */

#define shellOpenSucceeded      1    
#define shellReconnected        2    
#define shellOpenFailed         3    
#define shellTerminated         4    

typedef uint8 ShellStatus; 

typedef struct {
    uint8       channel;             
    ShellStatus status;              
    uint16      _unused_5_6;          /* padding */
    uint32      _unused_5;            /* padding */
} ShellStatusVariant; 


/*
 * Format of Shell Text PDU:
 */

typedef struct {
    uint8      channel;             
    unsigned   _unused_7       : 24;  /* padding */
    uint32     sequenceNumber;       /* of first char included */
    uint16     length;               /* number of characters */
    char       text[1];             
} ShellTextVariant; 


/*
 * Format of Shell Ack PDU:
 */

typedef struct {
    uint8     channel;             
    unsigned  _unused_8       : 24;  /* padding */
    uint32    sequenceNumber;       /* of next char expected */
} ShellAckVariant; 


/*
 * Definition of Monitor Protocol Data Unit.
 *
 */

typedef struct {
    MonitorProtocolVersion version;  
    Monitor_PDUKind kind;            
    uint16 _unused_9_10;         /* padding */
    uint32 _unused_9;             /* padding */
    union {
        MonitorResponseVariant monitorRsp;
        RemoteCommandVariant   remoteCmd;
        ShellOpenVariant       open;
        ShellStatusVariant     status;
        ShellTextVariant       text;
        ShellAckVariant        ack;
    } variant;                          
} Monitor_PDU; 



#endif
