
/*
 * FILE:  p_sensor.h 
 *
 * Created on Fri Apr 26 13:42:15 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_sensor_included
#define _p_sensor_included



/*
 * File:     p_sensor.drn
 * Contents: The sensor protocol
 * System:   simnet development
 * Created:  7-Nov-1991
 * Author:   tshannon
 * Remarks:
 * History:
 *
 * $RCSfile: p_sensor.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 * COPYRIGHT 1991 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 */

#include <stdtypes.h>
#include <basic.h> 

/*
 * Versions of the Sensor Protocol
 */

#define SensorProtocolVersionNov91 1 

typedef uint8    SensorProtocolVersion; 


/*
 * The current version
 */

#define sensorProtocolVersionCurrent           1

/*
 * Various kinds of Sensor PDUs:
 */

#define sensorServerIdentifyPDUKind 1
#define sensorServerEntityPDUKind 2  
#define sensorCreatePDUKind     3    
#define sensorDeletePDUKind     4    
#define sensorSetFOVPDUKind     5    
#define sensorSetViewLocationPDUKind 6
#define sensorSetRangeDetectionPDUKind 7
#define sensorSetTimeDetectionPDUKind 8
#define sensorAddTargetPDUKind  9    
#define sensorDeleteTargetPDUKind 10 
#define sensorStartScanPDUKind  11   
#define sensorStopScanPDUKind   12   
#define sensorServerAlivePDUKind 13  
#define sensorSetWavelengthPDUKind 14

typedef uint8    SensorPDUKind; 


/*
 * A sensor is identified by the vehicle which the sensor is attached to,
 * the simulation address of the host simulating the sensor, and a sensor
 * number on that host.  The combination of sensor_addr and sensor must be
 * unique.
 *
 */

typedef struct {
    VehicleID         veh_id;         /* Address of vehicle sensor is on. */
    SimulationAddress sensor_addr;    /* Address of host simulating sensor. */
    uint16            sensor;        
} sensor_id_t; 


/*
 * On the server side, the server determines FOV in a cone which is somewhat
 * larger than the actual FOV of the sensor.  This is so that network latency
 * will have less of an effect on the accuracy of the data
 * available to the sensor.
 *
 */

typedef float64 sensor_view_vector_t[3]; 


/*
 * A target is either a vehicle, object (terrain features), or smoke
 *
 */

#define TARGET_TYPE_VEHICLE     1    
#define TARGET_TYPE_OBJECT      2    
#define TARGET_TYPE_SMOKE_CLOUD 3    

typedef uint16 target_type_t; 

typedef struct {
    uint16 site;                
    uint16 host;                
    uint16 vehicle;             
    uint16 tgt_type;            
} sensor_target_id_t; 


/*
 *  Each target is uniquely identified by its object id and target type.
 *  In the case of a vehicle the obj_id is the VehicleID of the vehicle.
 *  In the case of a terrain object, it is the patch_row, patch_column, and
 *  index into the list of features for the patch, as returned from the
 *  ctdb library.
 *
 * center_pos is the center of the target in SIMNET_WorldCoordinates.
 * center_offset is the offset of the center_pos from the origin of
 * the target (based on the model being displayed).  This exists so that
 * if the client has a different idea of where the target is it can
 * adjust the center_pos based on its own idea of the target origin.
 *
 *  detection_score represents the "detectability" of an object, typically
 *  from 0.0-1.0.  Exact meaning may vary by application.
 *
 *  smoke_visbility is the visibility after applying a smoke model
 *  to the target.  0.0 = not visible,  1.0 = totally visible
 *
 */

typedef struct {
    SIMNET_WorldCoordinates center_pos; 
    SIMNET_WorldCoordinates center_offset; 
    float32                 visible_area; 
    float32                 horizontal_size; 
    float32                 vertical_size; 
    float32                 detection_score; 
    float32                 range;   
    float32                 smoke_visibility; 
    ObjectID                obj_id;  
    target_type_t           target_type; 
} sensor_target_t; 


/*
 * Datagram.
 *
 * A ServerIdentify is broadcast out by the client when it desires a
 * server to attach to.  The server_address contains a specification which
 * servers should check before sending out a corresponding ServerEntity
 * PDU.  If an address component is 0, it represents a wildcard (eg.
 * server_address.site = 0 mean  a server at any site should respond if
 * the host portion is valid), otherwise it represents the site or host
 * of servers which should respond.
 *
 */

typedef struct {
    SimulationAddress server_address; 
} SensorServerIdentifyVariant; 


/*
 * Datagram.
 *
 * Sent by an entity offering the sensor server service.  num_clients indicates
 * the current number of clients (sensors) the server is currently servicing.
 * The sensor library will pick the least loaded server.
 *
 */

typedef struct {
    SimulationAddress server_address; 
    uint16       num_clients;   
} SensorServerEntityVariant; 


/*
 * Transaction.
 *
 * A create is sent by the client when it desires to create a sensor.
 *
 */

typedef struct {
    sensor_id_t sensor_id;           
    ExerciseID  exercise;            
    uint8       _unused_3;            /* padding */
} SensorCreateVariant; 


/*
 * Transaction.
 *
 * A delete is sent by the client when it desires to delete a previously
 * created sensor.
 *
 */

typedef struct {
    sensor_id_t sensor_id;           
} SensorDeleteVariant; 


/*
 * Transaction.
 *
 * Sent by the client to change or set the field of view of the sensor.
 * vertical and horizontal fovs are half-angles in radians of the width
 * of the fov from a centered view vector.  range is meters.
 *
 */

typedef struct {
    sensor_id_t sensor_id;           
    float32     vertical_fov;        
    float32     horizontal_fov;      
    float32     range;               
} SensorSetFOVVariant; 


/*
 * Transaction.
 *
 * Sent by the client to set the location and orientation of the sensor.
 * The view is the vector of "gaze" of the sensor.
 *
 */

typedef struct {
    sensor_id_t             sensor_id; 
    uint32                 _unused_4;  /* padding */
    SIMNET_WorldCoordinates position; 
    sensor_view_vector_t    view;    
} SensorSetViewLocationVariant; 

#define SENSOR_MAX_RANGE_BUCKETS 5

/*
 * Each range bucket defines a range within which the "mean" and "std_dev"
 * are used as inputs to a random normal distribution to generate detection
 * scores for targets.
 *
 */

typedef struct {
    float32 range_min;                    
    float32 range_max;                    
    float32 mean;                         
    float32 std_dev;                      
} sensor_range_bucket_t; 


/*
 * Sent by the client to set the range bucket parameters.
 *
 */

typedef struct {
    sensor_id_t           sensor_id; 
    sensor_range_bucket_t veh_range_buckets[SENSOR_MAX_RANGE_BUCKETS]; 
    sensor_range_bucket_t obj_range_buckets[SENSOR_MAX_RANGE_BUCKETS]; 
} SensorSetRangeDetectionVariant; 


/*
 * Transaction.
 *
 * Sent by the client to set the frame interval that the sensor's target list
 * should be updated.  frame_rate is the rate at which intervisibility checks
 * are performed.  detect_rate is the rate at which detection_score information
 * in the AddTargetVariant is updated.  detect_time is the amount of time
 * that a target must be detected for before a detection_score is applied.
 *
 */

typedef struct {
    sensor_id_t sensor_id;           
    uint16 frame_rate;          
    uint16 detect_rate;         
    uint16 detect_time;         
} SensorSetTimeDetectionVariant; 


/*
 * Datagram.
 *
 * Sent by the server to indicate targets which are visible to the sensor.
 *
 */

typedef struct {
    SimulationAddress server_addr;   
    sensor_id_t       sensor_id;     
    uint32            _unused_5;      /* padding */
    uint32            target_count;  
    sensor_target_t   target[1];     
} SensorAddTargetVariant; 


/*
 * Datagram.
 *
 * Sent by the server to indicate targets which are no int32er visible to
 * the sensor.
 *
 */

typedef struct {
    SimulationAddress  server_addr;  
    sensor_id_t        sensor_id;    
    uint32             _unused_6;     /* padding */
    uint32             target_count; 
    sensor_target_id_t target[1];    
} SensorDeleteTargetVariant; 


/*
 * Transaction.
 *
 * Sent by the client to indicate that a sensor is scanning, or "on".
 *
 */

typedef struct {
    sensor_id_t sensor_id;           
} SensorStartScanVariant; 


/*
 * Transaction.
 *
 * Sent by the client to indicate that a sensor has stopped scanning, or is 
 * "off".
 * 
 */

typedef struct {
    sensor_id_t sensor_id;           
} SensorStopScanVariant; 


/*
 * Datagram.
 *	Sent by the sensor server to indicate that it is still alive
 *	every X seconds.
 *
 *
 */

typedef struct {
    SimulationAddress server_addr;   
    sensor_id_t       sensor_id;     
} SensorServerAliveVariant; 


/*
 * Datagram
 *	Set the wavelength that the sensor is currently attuned to.  Note
 * 	that this must be a value from libsmkint.h, out of the 
 * 	SMKINT_WAVELENGTH typedef enum.
 *
 */

typedef struct {
    sensor_id_t sensor_id;           
    uint32      wavelength;          
} SensorSetWavelengthVariant; 

typedef struct {
    SensorProtocolVersion version;   
    SensorPDUKind         kind;              
    ExerciseID            exercise;             
    uint8                 _unused_7_8;           /* padding */
    uint32                _unused_7;             /* padding */
    union {
        SensorServerIdentifyVariant    serverIdentify;
        SensorServerEntityVariant      serverEntity;
        SensorCreateVariant            create;
        SensorDeleteVariant            delete;
        SensorSetFOVVariant            setFOV;
        SensorSetViewLocationVariant   setViewLocation;
        SensorSetRangeDetectionVariant setRangeDetection;
        SensorSetTimeDetectionVariant  setTimeDetection;
        SensorAddTargetVariant         addTarget;
        SensorDeleteTargetVariant      deleteTarget;
        SensorStartScanVariant         startScan;
        SensorStopScanVariant          stopScan;
        SensorServerAliveVariant       serverAlive;
        SensorSetWavelengthVariant     setWavelength;
    } variant;                          
} SensorPDU; 



#endif
