
/*
 * FILE:  p_sim.h 
 *
 * Created on Mon Jun 17 19:35:08 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_sim_included
#define _p_sim_included



/*
 * File:     pro_sim.drn
 * Contents: The Simulation protocol
 * System:   simnet development.
 * Created:  10-Oct-1989
 * Author:   apope
 *
 *  History:
 *    10/10/89	Pope		created
 *    7 Dec 89  Kanarick 	changed ImpactVariant
 *    12/12/89	Pope		aligned Float(64) to 64-bit boundaries
 *    12/13/89	Pope		added type fields to ServiceRequestVariant
 *    12/13/89  Pope		fixed some padding errors
 *    01/08/90	Pope		added reason flags to VehicleAppearanceVariant
 *    01/09/90	Pope		added battleScheme to ActivateRequestVariant
 *    01/12/90  Kanarick        added Jan90 version number, removed reason flgs
 *    01/18/90	Kanarick	changed rotation field of VA to RM type
 *    02/27/90  Kanarick	added new fireResult, changed filename
 *    05/18/90	Kanarick	added breached lane PDU
 *
 * $RCSfile: p_sim.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 * COPYRIGHT 1989 BBN Systems and Technologies Corporation
 * 10 Moulton Street
 * Cambridge, Ma. 02138
 * 617-873-3000
 */

#include <stdtypes.h>
#include <basic.h> 
#include <status.h> 
#include <dis203_types.h> 

/*
 * Versions of the simulation protocol:
 */

#define simProtocolVersionAug89 1    
#define simProtocolVersionJan90 2    
#define simProtocolVersionJan90Corrected 3

typedef uint8    SimulationProtocolVersion; 


/*
 * The current version
 */

#define simProtocolVersionCurrent           3

/*
 * Various kinds of Simulation PDUs:
 */

#define activateRequestPDUKind  1     /* Activate Request PDU */
#define activateResponsePDUKind 2     /* Activate Response PDU */
#define deactivateRequestPDUKind 3    /* Deactivate Request PDU */
#define deactivateResponsePDUKind 4   /* Deactivate Response PDU */
#define vehicleAppearancePDUKind 5    /* Vehicle Appearance PDU */
#define radiatePDUKind          6     /* Radiate PDU */
#define firePDUKind             7     /* Fire PDU */
#define impactPDUKind           8     /* Impact PDU */
#define indirectFirePDUKind     9     /* Indirect Fire PDU */
#define collisionPDUKind        10    /* Collision PDU */
#define serviceRequestPDUKind   11    /* Service Request PDU */
#define resupplyOfferPDUKind    12    /* Resupply Offer PDU */
#define resupplyReceivedPDUKind 13    /* Resupply Received PDU */
#define resupplyCancelPDUKind   14    /* Resupply Cancel PDU */
#define repairRequestPDUKind    15    /* Repair Request PDU */
#define repairResponsePDUKind   16    /* Repair Response PDU */
#define markerPDUKind           17    /* Marker PDU */
#define breachedLanePDUKind     18    /* Breached Lane PDU */
#define minefieldPDUKind        19    /* MineField PDU */
#define damageRequestPDUKind    20    /* Damage Request PDU */
#define damageResponsePDUKind   21    /* Damage Response PDU */
#define weatherConditionPDUKind 22    /* Weather Condition PDU */
#define radarPDUKind            30    /* Loral's Radar PDU for CMT */
#define smokePDUKind            31    /* Loral's Smoke PDU for Losat */
#define environmentPDUKind      32    /* Loral's Environmental PDU	 */
#define laserPDUKind            33    /* Loral's Laser PDU for DIS compat. */
#define aggregateAppearancePDUKind 134 /* Loral's ASPDU PDU for DIS compat. */
typedef uint8    SimulationPDUKind; 


/*
 * PDUs for activating and deactivating vehicles.
 *
 *
 * Reasons given for activating a vehicle:
 */

#define activateReasonOther     0     /* none of those listed below */
#define exerciseStart           1     /* initial entry into new exercise */
#define exerciseRestart         2     /* restart of exercise */
#define vehicleReconstitution   3     /* restoration of vehicle status */
#define towingArrival           4     /* towing destination reached */
typedef uint8    ActivateReason; 


/*
 * Contents of an Activate Request PDU:
 */

typedef struct {
    /* Purpose of the activation: */
    ActivateReason          reason;

    /* Identity of the activated vehicle: */ 
    VehicleClass            vehicleClass;
    VehicleID               vehicleID; 
    OrganizationalUnit      unit;    
    VehicleMarking          marking; 
    VehicleGuises           guises;

    /* Information about the simulated world: */
    SP_Time                 simulatedTime; 
    TerrainDatabaseID       terrain; 
    BattleScheme            battleScheme;

    /* Status of the vehicle: */
    unsigned                onSurface          : 1;  /* Boolean */
    unsigned                _unused_32         : 23;  /* padding */
    VehicleStatus           status;  
    SIMNET_WorldCoordinates location;

    /* Depending on vehicle class: */
    union {
         /* A static vehicle: */
        struct {
            Angle  hullAzimuth;       
            uint32 _unused_33;    /* padding */
        } _static;
         /* A simple moving vehicle, without a turret: */
        struct {
            Angle  hullAzimuth;       
            uint32 _unused_34;    /* padding */
        } simple;
         /* A tank: */
        struct {
            Angle  hullAzimuth;       
            Angle  turretAzimuth;     
        } tank;
    } specific;    
                     
    /* More information about the status of the vehicle: */
    VelocityVector          velocity; 
    unsigned                freezeState        : 1;  /* Boolean */
    unsigned                _unused_35         : 31;  /* padding */

    /* More information about the simulated world: */
    /* visible light visibility in meters */
    float32                 VLVisibility; 
    SkyColor                simulatedSkyColor; 
    unsigned                _unused_36         : 24;  /* padding */
} ActivateRequestVariant; 


/*
 * Results reported on activating a vehicle:
 */

#define activateRequestAccepted 0    
#define invalidActivateParameter 1   
#define unexpectedActivateReason 2   
#define invalidVehicleIdentifier 3   
#define terrainDatabaseUnavailable 4 

typedef uint8 ActivateResult; 


/*
 * Contents of an Activate Response PDU:
 */

typedef struct {
    VehicleID      vehicleID;           
    ActivateResult result;           
    uint8          _unused_37;           /* padding */
    uint16         timeLimit;           
    uint16         _unused_38_39;        /* padding */
    uint32         _unused_38;           /* padding */
} ActivateResponseVariant; 


/*
 * Contents of a Breached Lane PDU:
 */

typedef struct {
    SIMNET_XYCoordinates vertices[4];  /* a "lane" */
    ObjectID             breachedLaneID; 
    uint8                numberOfVertices; 
    ForceID              force;       /* the force of the breachers */
    SP_Time              breachTime;  /* when the lane was breached */

} BreachedLaneVariant; 


/*
 * Contents of a Damage Request PDU:
 */

typedef struct {
    VehicleID  receiverID;           
    VehicleID  supplierID;           
    EventID    eventID;              
    DamageType damage;               
} DamageRequestVariant; 


/*
 * Various responses to a Damage Request PDU:
 */

#define damageRequestAccepted   0    
#define invalidDamageType       1    

typedef uint8 DamageResult; 


/*
 * Contents of a Damage Response PDU:
 */

typedef struct {
    VehicleID     recieverID;           
    VehicleID     supplierID;           
    DamageResult  result;             
    unsigned      _unused_40  : 24;      /* padding */
} DamageResponseVariant; 


/*
 * Reasons given for deactivating a vehicle:
 */

#define deactivateReasonOther   0     /* none of those listed below */
#define exerciseEnd             1     /* end of exercise */
#define vehicleWithdrawn        2     /* vehicle withdrawn from exercise */
#define vehicleDestroyed        3     /* vehicle no int32er exists */
#define towingDeparture         4     /* start of towing operation */
#define vehicleHandoff          5     /* vehicle being simulated by another */
                                      /* SimulationAddress */
typedef uint8 DeactivateReason; 


/*
 * Contents of a Deactivate Request PDU:
 */

typedef struct {
    VehicleID        vehicleID;            
    DeactivateReason reason;         
    uint8            _unused_41;            /* padding */
} DeactivateRequestVariant; 


/*
 * Results reported on deactivating a vehicle:
 */

#define deactivateRequestAccepted 0  
#define invalidDeactivateParameter 1 
#define unexpectedDeactivateReason 2 
#define vehicleNotActive        3    

typedef uint8 DeactivateResult; 


/*
 * Contents of a Deactivate Response PDU:
 */

typedef struct {
    VehicleID        vehicleID;            
    DeactivateResult result;         
    uint8            _unused_42;            /* padding */
} DeactivateResponseVariant; 


/*
 * PDUs for reporting the state of simulated vehicles
 *
 *
 * Contents of a Vehicle Appearance PDU:
 */

typedef struct {
    /* Identity of the vehicle: */
    VehicleID               vehicleID; 
    VehicleClass            vehicleClass; 
    ForceID                 force;

    /* Appearance of the vehicle: */
    VehicleGuises           guises;  
    SIMNET_WorldCoordinates location; 
    float32                 rotation[3][3]; 
    uint32                  appearance; 
    VehicleMarking          marking; 
    uint32                  timestamp; 
    VehicleCapabilities     capabilities; 
    uint16                  engineSpeed; 
    unsigned                stationary    : 1;  /* Boolean */
    unsigned                _unused_43    : 15;  /* padding */

    /* Depending on vehicle class: */
    union {
         /* A simple moving vehicle, without a turret: */
        struct {
            VelocityVector velocity; 
            uint32         _unused_44;  /* padding */
        } simple;
         /* A tank: */
        struct {
            VelocityVector velocity; 
            Angle          turretAzimuth; 
            Angle          gunElevation; 
            uint32         _unused_45;  /* padding */
        } tank;
    } specific;                         
} VehicleAppearanceVariant; 

 /* Smoke parameters: */

#define DEAD_RECKONING_ALGORITHM_OTHER           0
#define DEAD_RECKONING_ALGORITHM_STATIC           1
#define DEAD_RECKONING_ALGORITHM_DRM_3D           2
typedef struct {
    float32         sigma_x;              
    float32         sigma_y;              
    float32         sigma_z;              

    /* simulated time is time in milliseconds since detonation */
    uint32 simulated_time;       
} IntervisRecord; 

typedef struct {
    uint8  dead_reckoning_algorithm; 
    unsigned         _unused_46                : 24;  /* padding */
    float32          dsigma_x_dt;         
    float32          dsigma_y_dt;         
    float32          dsigma_z_dt;         
} SmokeDeadReckoningRecord; 

typedef struct {
    IntervisRecord           intervis_factors; 
    SmokeDeadReckoningRecord dead_reckoning; 
} SmokeRecord; 


/*
 * Contents of a Smoke PDU:
 */

typedef struct {
     /* Identity of the vehicle: */
    VehicleID               smokeID; 
    ForceID                 force;   
     /* data_valid is true if the smoke_parameters, downwind_distance, */
         /* centroid_height, velocity and fill_weight are all valid. */

    uint8                   data_valid;  /* Appearance of the smoke cloud: */

    VehicleGuises           guises;  
    SIMNET_WorldCoordinates location; 
    uint32                  appearance; 
    uint32                  timestamp; 
    VelocityVector          velocity; 
    VelocityVector          wind_velocity; 
    float32                 downwind_distance; 
    float32                 centroid_height; 
    float32                 num_sources; 
    float32                 fill_weight; 
    SmokeRecord             smoke_parameters; 
    ObjectType              munition_type; 
} SmokeVariant; 


/*
 * Experimental Environment PDU
 *
 * All constants should match their DIS counterparts
 *
 */

#define EnvStatusOff            0    
#define EnvStatusOn             1    

typedef uint16 EnvStatus; 

#define envWeatherModelSimple   0x1
#define envWeatherModelObserved 0x2
#define envSeaStateModelSimple  0x3
#define envEphemerisModelSimple 0x5
#define envSmokeModelOther      0x10
#define envSmokeModelSimple     0x11
#define envSmokeModelAtlas      0x12
#define envSmokeModelCOMBIC     0x13
#define envFlareModelFLARE      0x14
#define envDustModelVehicular   0x15

typedef uint8 EnvModel; 

#define EnvPaddingKind          0x0 
  /* State Records */
#define EnvStateSimpleIlluminationKind 0x1
#define EnvStateSimpleSoundKind 0x2  
#define EnvStateSimpleObscurantKind 0x3
#define EnvStateSimpleWeatherKind 0x4
#define EnvStateObservedWeatherKind 0x5
#define EnvStateCOMBICPlumeKind 0x100
#define EnvStateCOMBICPuffKind  0x101
#define EnvStateEphemerisKind   0x102
#define EnvStateFlareKind       0x103
#define EnvStateNetSeaKind      0x104
#define EnvStateVehDustKind     0x502
#define EnvStateSimulationDateKind 0x500
 /* Geometry Records */
#define EnvGeometryUniformKind  0x50000
#define EnvGeometryPointKind    0xa0000
#define EnvGeometrySpheroidKind 0xd0000
#define EnvGeometryPolygonKind  0x10000003
#define EnvGeometryExtentsKind  0x10000004
#define EnvGeometrySphericalCoordsKind 0x10000005
#define EnvGeometryGaussianPlumeKind 0x60000000
#define EnvGeometryGaussianPuffKind 0x70000000
#define EnvGeometryConeKind     0x30000000
#define EnvGeometryFlexPlumeKind 0x10000400
 /* Simulation Data */
typedef uint32 EnvRecordKindType; 

typedef struct {
    uint32 _unused_47_48;         /* padding */
    uint32 _unused_47_49;         /* padding */
    uint32 _unused_47_50;         /* padding */
    uint32 _unused_47_51;         /* padding */
    uint32 _unused_47_52;         /* padding */
    uint32 _unused_47_53;         /* padding */
    uint32 _unused_47_54;         /* padding */
    uint32 _unused_47_55;         /* padding */
    uint32 _unused_47_56;         /* padding */
    uint32 _unused_47_57;         /* padding */
    uint32 _unused_47_58;         /* padding */
    uint32 _unused_47_59;         /* padding */
    uint32 _unused_47_60;         /* padding */
    uint32 _unused_47_61;         /* padding */
    uint32 _unused_47_62;         /* padding */
    uint32 _unused_47_63;         /* padding */
    uint32 _unused_47_64;         /* padding */
    uint32 _unused_47_65;         /* padding */
    uint32 _unused_47_66;         /* padding */
    uint32 _unused_47_67;         /* padding */
    uint32 _unused_47_68;         /* padding */
    uint32 _unused_47_69;         /* padding */
    uint32 _unused_47_70;         /* padding */
    uint32 _unused_47_71;         /* padding */
    uint32 _unused_47_72;         /* padding */
    uint32 _unused_47_73;         /* padding */
    uint32 _unused_47_74;         /* padding */
    uint32 _unused_47_75;         /* padding */
    uint32 _unused_47_76;         /* padding */
    uint32 _unused_47_77;         /* padding */
    uint32 _unused_47_78;         /* padding */
    uint32 _unused_47;            /* padding */
} EnvPadding; 

typedef struct {
    SIMNET_WorldCoordinates location; 
} EnvGeometryPoint; 

typedef struct {
    float64              z0;         
    float64              z1;         
    int32                number_of_vertices; 
    uint32               _unused_79_80;  /* padding */
    uint32               _unused_79_81;  /* padding */
    uint32               _unused_79;  /* padding */
    SIMNET_XYCoordinates vertex[1];  
} EnvGeometryPolygon; 

typedef struct {
    SIMNET_WorldCoordinates location; 
    float32                 distance; 
    float32                 azimuth; /* heading/wind direction in radians */
    float32                 sigma_y; 
    float32                 sigma_z; 
    float32                 height;  
    float32                 horz_velocity; 
    float32                 vert_velocity; 
    float32                 d_sigma_y_dt; 
    float32                 d_sigma_z_dt; 
} EnvGeometryGaussianPlume; 

typedef struct {
    SIMNET_WorldCoordinates location; 
    float32                 distance; 
    float32                 azimuth; /* heading/wind direction in radians */
    float32                 sigma_x; 
    float32                 sigma_y; 
    float32                 sigma_z; 
    float32                 height;  
    float32                 horz_velocity; 
    float32                 vert_velocity; 
    float32                 d_sigma_x_dt; 
    float32                 d_sigma_y_dt; 
    float32                 d_sigma_z_dt; 
} EnvGeometryGaussianPuff; 

typedef struct {
    SIMNET_WorldCoordinates origin;  
    SIMNET_WorldCoordinates extent;  
} EnvGeometryExtents; 

typedef struct {
    float32 azimuth;                      
    float32 elevation;                    
    float32 range;                        
} EnvGeometrySphericalCoords; 

typedef struct {
    SIMNET_WorldCoordinates location; 
    float32                 peak_angle; 
    VelocityVector          horz_velocity; 
    float32                 vert_velocity; 
} EnvGeometryCone; 

typedef struct {
    SP_Time     time_since_detonation; 
    ObjectType  source;               
    int32       number_of_sources;    
    int16       geometry_index;       
} EnvStateCOMBIC; 

typedef struct {
    SP_Time     time_since_detonation; 
    ObjectType  source;               
    float32     source_intensity;     
    int32       number_of_sources;    
    int16       geometry_index;       
} EnvStateFlare; 

typedef struct {
    int16    sun_geometry_index;           
    int16    moon_geometry_index;          
    float32  moon_phase;                   
} EnvStateEphemeris; 

typedef struct {
    WavebandType waveband;           
    float32      coeff;                
} EnvExtinctionRecord; 

typedef struct {
    uint32                         simulation_time; 
    SimpleWeatherPrecipitationType precipitation_type;
    /* mm/hr */
    float32                        precipitation_rate; 
    SimpleWeatherExtinctionType    extinction_type;
    /* units tbd */
    float32                        extinction_amount;
    /* deg C */
    float32                        temperature; 
    float32                        dewpoint;
    /* relative humidity percent */
    float32                        relative_humidity; 
    /* barometric pressure in N/M2 (1 N/m2 == 7.501E-4 cm Hg) */
    float32                        barometric_pressure;
    /* illumination level */
    float32                        ambient_light;
    /* wind velocity in m/s */
    VelocityVector                 wind;        
    /* 0.0 is no clouds, 1.0 is complete cloud cover */
    float32                        cloud_cover;
    /* altitude ASL */
    float32                        cloud_top;   
    float32                        cloud_base;  
    /* extinction coefficients(3) and the wavebands they appy to */
    EnvExtinctionRecord            extinctionRecord[3];
    /* where does this weather apply? */
    uint16                         geometry_index; 
    uint16                         _unused_82;   /* padding */
    uint32                         cloud_type;  
    float32                        total_precip_last_hr; 
    uint32                         last_rain_time; 
    uint32                         sub_soil_saturation; 
} EnvStateSimpleWeather; 

#define EnvMaxSiteNameLen       32
typedef struct {
    /* Name of observation site: */
    uint8                          name[EnvMaxSiteNameLen]; 
    /* Time observation was made. Do *NOT* set sim time from this!  */
    uint32                         observation_time; 
    SimpleWeatherPrecipitationType precipitation_type;
    /* mm/hr */
    float32                        precipitation_rate; 
    SimpleWeatherExtinctionType    extinction_type;
    /* units tbd */
    float32                        extinction_amount;
    /* deg C */
    float32                        temperature; 
    float32                        dewpoint;
    /* relative humidity percent */
    float32                        relative_humidity; 
    /* barometric pressure in N/M2 (1 N/m2 == 7.501E-4 cm Hg) */
    float32                        barometric_pressure;
    /* illumination level */
    float32                        ambient_light;
    /* wind velocity in m/s */
    VelocityVector                 wind;        
    /* 0.0 is no clouds, 1.0 is complete cloud cover */
    float32                        cloud_cover;
    /* altitude ASL */
    float32                        cloud_top;   
    float32                        cloud_base;  
    /* extinction coefficients(3) and the wavebands they appy to */
    EnvExtinctionRecord            extinctionRecord[3];
    /* where does this weather apply? */
    uint16                         geometry_index; 
} EnvStateObservedWeather; 

typedef uint32 EnvStateSimulationDate; 

typedef float32 EnvVelocityComponent; 

typedef float32 EnvComponentOffsetType; 

typedef struct {
    EnvComponentOffsetType x_offset; 
    EnvComponentOffsetType y_offset; 
    EnvComponentOffsetType z_offset; 
} EnvLocationOffsetRecord; 

typedef struct {
    uint32                  time_since_creation; 
    VehicleID               attached_vehicle_ID; 
    uint16                  geometry_index; 
    VelocityVector          vehicle_velocity; 
    SIMNET_WorldCoordinates location; 
    EnvLocationOffsetRecord bounding_sphere_center; 
    float32                 bounding_sphere_radius; 
    float32                 plume_length; 
    float32                 sigma_y; 
    float32                 sigma_z; 
    float32                 height;  
    EnvVelocityComponent    horz_velocity; 
    EnvVelocityComponent    vert_velocity; 
    EnvVelocityComponent    d_sigma_y_dt; 
    EnvVelocityComponent    d_sigma_z_dt; 
    ObjectType              vehicle_type; 
} EnvStateVehDust; 

typedef struct {
    EnvLocationOffsetRecord node_location; 
    float32                 node_dsigma_y_dt; 
    float32                 node_dsigme_z_dt; 
    float32                 node_mass; 
    uint16                  node_time_since_creation; 
    uint16                  node_soil_type; 
    float32                 node_d_sigma_info; 
} EnvDustNodeRecord; 

typedef struct {
    EnvDustNodeRecord dust_nodes[4]; 
} EnvGeometryFlexPlume; 

typedef struct {
    EnvRecordKindType record_kind;   
    uint32            _unused_82;            /* padding */
    union {
        EnvPadding                 padding;
        EnvStateCOMBIC             combic_plume;
        EnvStateCOMBIC             combic_puff;
        EnvStateFlare              flare;
        EnvStateVehDust            veh_dust;
        EnvStateSimulationDate     simulation_date;
        EnvStateSimpleWeather      simple_weather;
        EnvStateEphemeris          ephemeris;
        EnvPadding                 uniform;
        EnvGeometryGaussianPlume   gaussian_plume;
        EnvGeometryCone            cone;
        EnvGeometryGaussianPuff    gaussian_puff;
        EnvGeometryPoint           point;
        EnvGeometryPolygon         polygon;
        EnvGeometryExtents         extents;
        EnvGeometrySphericalCoords sphcoords;
        EnvGeometryFlexPlume       flex_plume;
    } variant;                          
} EnvRecord; 


/*
 * Contents of an Environmental PDU:
 */

typedef struct {
    VehicleID  envID;                
    EnvModel   model;                
    EnvStatus  status;               
    uint32     number_of_records;    
    ObjectType envType;              
    uint32     timestamp;            
    EnvRecord  envRecord[1];         
} EnvironmentVariant; 


/*
 * Various modes of a radar:
 */

#define radarModeOther          0     /* none of those listed below */
#define radarModeSearch         1    
#define radarModeAcquisition    2    
#define radarModeTracking       3    

typedef uint8 RadarMode; 


/*
 * Various types of radar duty cycle:
 */

#define radarDutyCycleOther     0     /* none of those listed below */
#define radarDutyCycleContinuous 1   
#define radarDutyCyclePulsed    2    

typedef uint8 RadarDutyCycle; 


/*
 * Charactistics of a radar emitter:
 */

typedef float32 SignalFrequency; /* hertz */
typedef float32 SignalPower; /* watts */

 /* Contents of a Radiate PDU: */
typedef struct {
    VehicleID               vehicleID; 
    RadarMode               mode;    
    RadarDutyCycle          dutyCycle; 
    SIMNET_WorldCoordinates location; 
    SignalFrequency         carrierFrequency; 
    SignalPower             signalPower; 
    float32                 antennaGain; 
    float32                 pulseEnergy; 
    uint8                   numberIllumed; 
    uint8                   numberDetected; 
    VehicleID               targetID[1]; 
} RadiateVariant; 

#define maxRadiateTargets       33

/*
 * PDUs for reporting weapons effects and collisions.
 *
 *
 * Various types of projectile:
 */

#define fireTypeShell           1    
#define fireTypeMissile         2    

typedef uint8 FireType; 


/*
 * Contents of a Fire PDU:
 */

typedef struct {
     /* Common to all shell and missile firings: */
    VehicleID               attackerID; 
    EventID                 eventID; 
    BurstDescriptor         burst;   
    TargetDescriptor        target;  
    VelocityVector          velocity; 
    SIMNET_WorldCoordinates muzzle;  
    VehicleID               projectileID; 
    unsigned                tracerValid   : 1;  /* Boolean */
    unsigned                _unused_83    : 7;  /* padding */
    FireType                fireType; 
     /* Depending on whether a shell or a missile is fired: */

    union {
         /* If a shell is fired: */
        struct {
            float32            range;    
            float32            slewRate; 
            ObjectType         ammoSelected; 
            VelocityVector     gunTipVelocity; 
            float64            sinGunAzim; 
            float64            cosGunAzim; 
            float64            sinGunElev; 
            float64            cosGunElev; 
        } shell;
         /* If a missile is fired: */
        struct {
            uint8       tube;        
            unsigned    _unused_84_85 : 24;  /* padding */
            uint32      _unused_84;   /* padding */
        } missile;
    } specific;                         
} FireVariant; 


/*
 * Various possible results of a firing a projectile:
 */

#define nonImpact               1     /* the projectile did not impact */
#define groundImpact            2     /* the projectile struck terrain */
#define vehicleImpact           3     /* the projectile struck a vehicle */
#define proximateImpact         4     /* the projectile came close */
#define vehicleImpactDirectedDamage 5 /* the projectile hit with specific results */
 /*   (for battle recreation only) */
typedef uint8 FireResult; 


/*
 * Contents of an Impact PDU:
 */

typedef struct {
     /* For any round fired: */
    VehicleID               attackerID; 
    EventID                 eventID; 
    BurstDescriptor         burst;   
    VehicleID               projectileID; 
    FireResult              result;  
    DirectedDamage          damage;   /* Only used for battle reenactment */

    /* For projectiles that impact somewhere: */
    float32                 momentum;        /* newton-seconds */
    float32                 energy;          /* joules */
    float32                 directionality;  /* steradians */
    SIMNET_WorldCoordinates location; 
    float64                 range;   

    /* For shots that strike a particular vehicle: */
    VehicleID               targetID; 
    VehicleComponent        component; 
    VehicleCoordinates      impact;  
    VehicleCoordinates      trajectory; 
} ImpactVariant; 


/*
 * A single indirect fire detonation is described by:
 */

typedef struct {
    SIMNET_WorldCoordinates location; 
    VehicleID               attackerID; 
    EventID                 eventID; 
    uint16                  delay;   
    FireResult              result;  
    uint8                   _unused_86_87;  /* padding */
    uint32                  _unused_86;  /* padding */
} IndirectFireDetonation; 


/*
 * Contents of an Indirect Fire PDU:
 */

typedef struct {
    BurstDescriptor        burst;    
    uint32                 _unused_88;  /* padding */
    IndirectFireDetonation detonations[1]; 
} IndirectFireVariant; 

#define maxIndirectFireDetonations           5

/*
 * Contents of a Collision PDU:
 */

typedef struct {
    VehicleID vehicleID;             
    EventID   eventID;               
    VehicleID targetID;              
    uint16    _unused_89;           /* padding */
} CollisionVariant; 


/*
 * A mine in a field has a density and a type
 */

typedef struct {
    ObjectType mineType;             
    float32    density;              
} MineDescriptor; 

#define maxMinefieldVertices    13
#define maxMineDescriptors      3

/*
 * Contents of a Minefield PDU:
 */

typedef struct {
    MineDescriptor       mines[maxMineDescriptors]; 
    SIMNET_XYCoordinates vertices[maxMinefieldVertices]; 
    uint8                numberOfVertices; 
    uint8                numberOfMineTypes; 
    ObjectID             minefieldID; 
    SP_Time              emplacementTime;  /* When the minefield was placed */
    ForceID              force;      
    unsigned             _unused_90_91      : 24;  /* padding */
    uint32               _unused_90;  /* padding */
} MinefieldVariant; 


/*
 * PDUs for resupplying and repairing vehicles.
 *
 * Contents of a Service Request, Resupply Offer, or Resupply Received PDU:
 */

typedef struct {
    VehicleID        receiverID;     
    VehicleID        supplierID;     
    ObjectType       vehicleType;    
    SimulatorType    simulatorType;  
    uint16           _unused_92_93;   /* padding */
    unsigned         _unused_92       : 24;  /* padding */
    uint8            numberMunitions; 
    MunitionQuantity munitions[1];   
} ResupplyVariant; 

#define maxResupplyMunitions    28

/*
 * Contents of a Resupply Cancel PDU:
 */

typedef struct {
    VehicleID receiverID;            
    VehicleID supplierID;            
    uint32    _unused_94;            /* padding */
} ResupplyCancelVariant; 


/*
 * Contents of a Repair Request PDU:
 */

typedef struct {
    VehicleID  receiverID;           
    VehicleID  supplierID;           
    EventID    eventID;              
    RepairType repair;               
} RepairRequestVariant; 


/*
 * Various responses to a Repair Request PDU:
 */

#define repairRequestAccepted   0    
#define invalidRepairType       1    

typedef uint8 RepairResult; 


/*
 * Contents of a Repair Response PDU:
 */

typedef struct {
    VehicleID    receiverID;           
    VehicleID    supplierID;           
    RepairResult result;             
    unsigned     _unused_95  : 24;      /* padding */
} RepairResponseVariant; 


/*
 * Maximum number of markers per PDU:
 */

#define maxMarkers              32

/*
 * a marker and its status
 */

typedef struct {
    SIMNET_WorldCoordinates location; 
    Angle                   orientation; 
    uint16                  identifier; 
    uint16                  _unused_96;  /* padding */
} MarkerDescriptor; 


/*
 * Every markerPDU is one of a series of PDUs all logically related
 */

typedef struct {
    uint8  number;               /* the number of this one */
    uint8  total;                /* out of this many */
} MarkerVariantNumber; 

typedef struct {
    VehicleGuises       guises;      
    SimulationAddress   simulator;   
    MarkerVariantNumber variantNumber; 
    uint8               numberOfMarkers; 
    ForceID             force;       
    MarkerDescriptor    markers[1];  
} MarkerVariant; 


/*
 * Contents of a Weather Condition PDU
 */

typedef struct {
     /* Unique ID for this piece of weather */
    ObjectID             objectID;   

    /* Priority (higher number overrides lower number in same area) */
    /* 0 is reserved for entire database weather, in which case the */
    /* bounding box is ignored. */
    uint8                priority;

    /* General Conditions */
    WeatherConditions    conditions; 

    /* Precipitation rate, for those General Conditions that are */
    /* associated with precipitation.  This quantity is in m/s, */
    /* but is an accumulation (volume per area per time), not a */
    /* velocity (distance per time). */
    float32              precipitationRate; 

    /* Overcast is between the range of 0.0 and 1.0 and is a percentage */
    /* of cloud cover.  0.0 is no clouds, and 1.0 is complete cloud */
    /* cover. */
    float32              overcast;   

    /* Simulated Time (should be the same for all weather) */
    uint32               secondsSince1970;

    /* Temperature in degrees Celsius */
    float32              temperature;

    /* Relative humidity percent */
    float32              humidity;   

    /* Barometric pressure in N/m2 (1 N/m2 == 7.501E-4 cm Hg) */
    float32              pressure;

    /* Visibility in meters */
    float32              visibility;

    /* Ambient light level in Candles */
    float32              light;

    /* Wind velocity in m/s */
    VelocityVector       wind;       
    uint32               _unused_97;  /* padding */

    /* Area over which this weather applies */
    SIMNET_XYCoordinates SWcorner;   
    SIMNET_XYCoordinates NEcorner;   
} WeatherVariant; 

#define MaxRadarTargets         10
typedef struct {
     /* Identity of the radar emitter: */
    VehicleID       vehicleID;       
    uint8           numberIllumed;    /* Number of illuminated targets */
    uint8           _unused_98;       /* padding */

    /* Radar descriptors  */
    uint32          radarSystem;     
    uint8           radarMode;       
    uint8           _unused_99;       /* padding */

    /* ARG!!!  the next set of fields are 32 bit fields, but */
    /* this packet is not 32 bit alligned.  set up so can define */
    /* packet, and will have to do bcopy into fields. */
    /* each field will be labeled with what it really should */
    /* represent (Thank you, Loral...) */
    uint16          azimuthCenter[2];    /* Float 32 */
    uint16          azimuthWidth[2];     /* Float 32 */
    uint16          elevationCenter[2];  /* Float 32 */
    uint16          elevationWidth[2];   /* Float 32 */
    uint16          power[2];            /* int 32 */
    uint16          timestamp[2];        /* UnsignedInteger 32 */

    /* Illuminated targets info */
    RadarTargetData targets[1];      
} RadarVariant; 


/*
 * Contents of a Laser PDU:
 */

typedef struct {
    DIS203_PDU_HEADER_RECORD        pdu_header; 
    DIS203_ENTITY_ID_RECORD         lasing_entity_id; 
    DIS203_CODE_NAME_TYPE           code_name; 
    DIS203_ENTITY_ID_RECORD         lased_entity_id; 
    uint8                           _unused_100;  /* padding */
    DIS203_LASER_CODE_TYPE          laser_code; 
    DIS203_POWER_TYPE               laser_power; 
    DIS203_LASER_WAVE_LENGTH_TYPE   laser_wave_length; 
    DIS203_VECTOR_RECORD            laser_spot_with_respect_to_entity; 
    DIS203_WORLD_COORDINATES_RECORD laser_spot_location; 
} LaserVariant; 


/*
 */

/*
 * Definition of a Simulation Protocol Data Unit.
 *-------------------------------------------------------------------------
 */

typedef struct {
    SimulationProtocolVersion version; 
    SimulationPDUKind         kind;          
    ExerciseID                exercise;             
    uint8                     _unused_101_102;       /* padding */
    uint32                    _unused_101;           /* padding */
    union {
        ActivateRequestVariant    activateReq;
        ActivateResponseVariant   activateRsp;
        DeactivateRequestVariant  deactivateReq;
        DeactivateResponseVariant deactivateRsp;
        VehicleAppearanceVariant  appearance;
        RadiateVariant            radiate;
        FireVariant               fire;
        ImpactVariant             impact;
        IndirectFireVariant       indirectFire;
        CollisionVariant          collision;
        ResupplyVariant           serviceReq;
        ResupplyVariant           resupplyOffer;
        ResupplyVariant           resupplyReceived;
        ResupplyCancelVariant     resupplyCancel;
        RepairRequestVariant      repairReq;
        RepairResponseVariant     repairRsp;
        MarkerVariant             marker;
        BreachedLaneVariant       breachedLane;
        MinefieldVariant          minefield;
        DamageRequestVariant      damageReq;
        DamageResponseVariant     damageRsp;
        WeatherVariant            weather;
        RadarVariant              radar;
        SmokeVariant              smoke;
        EnvironmentVariant        env;
        LaserVariant              laser;
    } variant;                          
} SimulationPDU; 



#endif
