
/*
 * pro_size.h
 *
 * Created on Sat Nov 11 10:20:20 1989 
 * using vi and my fingers.
 * 
 */

#ifndef _p_size_included
#define _p_size_included

/*  This file defines macros for determining the size of various PDUs. */
/*  As with all macros (potentially) all bets are off if you use args  */
/*  with side effects.                                                 */

/* These size macros have explicit padding on the end so that if allocating
 * an array of PDU's using malloc (n * XXX_SIZE()), each element in the array
 * is properly aligned.
 */

/* $RCSfile: p_size.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

/*  First, the headers.  These macros take a pointer to the            */
/*  appropriate PDU type.                                              */

#define PRO_ASSOC_DATAGRAM_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_ASSOC_REQUEST_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant) + sizeof ((p)->variant.request) - \
	sizeof ((p)->variant.request.data))

#define PRO_ASSOC_RESPONSE_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant) + sizeof ((p)->variant.response) - \
	sizeof ((p)->variant.response.data))

#define PRO_ASSOC_PADDING_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_SIM_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_DATA_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_MGMT_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_P2P_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->data))

#define PRO_STEALTH_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_PO_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_DES_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_SOFTP_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_STNGRY_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

/*  Then the individual PDU types  */


/* **Association Protocol - These macros take a pointer to an   */
/*   AssociationPDU and the length of the data portion.         */

#define PRO_ASSOC_DATAGRAM_SIZE(p, l) \
	(PRO_ASSOC_DATAGRAM_HDR_SIZE (p) + (l))

#define PRO_ASSOC_REQUEST_SIZE(p, l) \
	(PRO_ASSOC_REQUEST_HDR_SIZE (p) + (l))

#define PRO_ASSOC_RESPONSE_SIZE(p, l) \
	(PRO_ASSOC_RESPONSE_HDR_SIZE (p) + (l))

#define PRO_ASSOC_PADDING_SIZE(p, l) \
	(PRO_ASSOC_PADDING_HDR_SIZE (p) + (l))




/* **Simulation Protocol (fixed length) - These macros take a pointer  */
/*   to a SimulationPDU.                                               */

#define PRO_SIM_ACTIVATE_REQ_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ActivateRequestVariant))

#define PRO_SIM_ACTIVATE_RSP_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ActivateResponseVariant))

#define PRO_SIM_DEACTIVATE_REQ_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (DeactivateRequestVariant))

#define PRO_SIM_DEACTIVATE_RSP_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (DeactivateResponseVariant))

#define PRO_SIM_APPEARANCE_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (VehicleAppearanceVariant))

#define PRO_SIM_FIRE_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (FireVariant))

#define PRO_SIM_IMPACT_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ImpactVariant))

#define PRO_SIM_COLLISION_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (CollisionVariant))

#define PRO_SIM_RESUPPLY_CANCEL_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyCancelVariant))

#define PRO_SIM_REPAIR_REQUEST_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RepairRequestVariant))

#define PRO_SIM_REPAIR_RESPONSE_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RepairResponseVariant))

#define PRO_SIM_WEATHER_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (WeatherVariant))


/* **Simulation Protocol (variable length) - These macros take a       */
/*   pointer to a SimulationPDU and the number of variants to be       */
/*   included.                                                         */


#define PRO_SIM_RADIATE_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RadiateVariant) + \
	((v) - 1) * sizeof (VehicleID) + ((v) % 2 ? 0 : sizeof (short)))

/* this packet's internals are only 16 bit alligned, NOT 32.  the tail   */
/* end additions cause it to be 32 bit alligned at the end of the packet */
/* every ODD number of targets */
#define PRO_SIM_RADAR_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RadarVariant) + \
	((v) - 1) * sizeof (RadarTargetData) + (((v) + 1) % 2 ? 0 : sizeof (short)))

#define PRO_SIM_IND_FIRE_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (IndirectFireVariant) + \
	((v) - 1) * sizeof (IndirectFireDetonation))

#define PRO_SIM_SERVICE_REQ_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyVariant) + \
	((v) - 1) * sizeof (MunitionQuantity))

#define PRO_SIM_RESUPPLY_OFFER_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyVariant) + \
	((v) - 1) * sizeof (MunitionQuantity))

#define PRO_SIM_RESUPPLY_RECEIVED_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyVariant) + \
	((v) - 1) * sizeof (MunitionQuantity))

#define PRO_SIM_MARKER_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (MarkerVariant) + \
	((v) - 1) * sizeof (MarkerDescriptor))

/* **Data Collection Protocol (fixed length) - These macros take   */
/*   a pointer to a DataCollectionPDU.                             */

#define PRO_DATA_EXERCISE_STATUS_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (ExerciseStatusVariant))

#define PRO_DATA_VEHICLE_STATUS_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (VehicleStatusVariant))

#define PRO_DATA_STATUS_QUERY_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (StatusQueryVariant))

#define PRO_DATA_STATUS_RESPONSE_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (StatusResponseVariant))

#define PRO_DATA_STATUS_CHANGE_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (StatusChangeVariant))

#define PRO_DATA_LASER_RANGE_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (LaserRangeVariant))

#define PRO_DATA_SR_IMPACT_RESULT_SIZE(p) \
        (PRO_DATA_HDR_SIZE (p) + sizeof (SR_CMImpactResultVariant))



/* **Data Collection Protocol (variable length) - These macros     */
/*   take a pointer to a DataCollectionPDU,  and the number of     */
/*   variants to include.                                          */

#define PRO_DATA_SIMULATION_STATUS_SIZE(p, v) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (SimulationStatusVariant) + \
	((v) - 1) * sizeof (OrganizationalUnit))

/* subtract sizeof long, because long is the largest element       */
/* type in the structure, so that's what kind of padding is being  */
/* done.                                                           */
#define PRO_DATA_EVENT_FLAG_SIZE(p, v) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (EventFlagVariant) - \
	sizeof (long) + (v) * sizeof (char))



/* **Management Protocol (fixed length) - These macros take        */
/*   a pointer to a ManagementPDU.                                 */

#define PRO_MGMT_EQUIP_STATUS_SIZE(p) \
	(PRO_MGMT_HDR_SIZE (p) + sizeof (EquipStatusVariant))

#define PRO_MGMT_SHOW_EFFECT_SIZE(p) \
	(PRO_MGMT_HDR_SIZE (p) + sizeof (ShowEffectVariant))

/* **Management Protocol (variable length) - These macros take     */
/*   a pointer to a ManagementPDU,  and the number of variants     */
/*   to include.                                                   */

/* subtract sizeof short, because short is the largest element     */
/* type in the structure, so that's what kind of padding is being  */
/* done.                                                           */
#define PRO_MGMT_ERROR_REPORT_SIZE(p, v) \
	(PRO_MGMT_HDR_SIZE (p) + sizeof (ErrorReportVariant) - \
	sizeof (short) + (v) * sizeof (char))


/* **Point To Point Protocol - These macros take a pointer to a    */
/*   PointToPointPDU, and the size of the data field.              */

#define PRO_P2P_PDU_SIZE(p, l) \
	(PRO_P2P_HDR_SIZE (p) + (l))


/* **Stealth Protocol (fixed length) - These macros take a pointer */
/*   to a StealthPDU.                                              */

#define PRO_STEALTH_ERROR_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (StealthErrorVariant))

#define PRO_STEALTH_VISIBILITY_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (StealthVisibilityVariant))

#define PRO_STEALTH_TELEPORT_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (TeleportVariant))

#define PRO_STEALTH_ATTACH_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (AttachVariant))

#define PRO_STEALTH_ATTACHED_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (AttachedVariant))

#define PRO_STEALTH_MIMIC_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (MimicVariant))

#define PRO_STEALTH_METAMORPHOSE_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (MetamorphoseVariant))

#define PRO_STEALTH_APPEARANCE_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (StealthAppearanceVariant))

/* **Persistent Object Protocol (fixed length) - These macros take a pointer*/
/*   to a PersistentObjectPDU.                                           */

#define PRO_PO_SIM_PRESENT_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (SimulatorPresentVariant))

#define PRO_PO_NOMINATE_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (NominationVariant))

#define PRO_PO_DESC_OBJ_SIZE(p) \
	(PRO_PO_HDR_SIZE (p) + sizeof (DescribeObjectVariant) - \
	sizeof ((p)->variant.describeObject.variant))

#define PRO_PO_SET_WS_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (SetWorldStateVariant))

/* **Persistent Object Protocol (variable length) - These macros take a */
/*   pointer to a PersistentObjectPDU and the number of variants to be  */
/*   included.                                                          */

#define PRO_PO_OBJ_PRES_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ObjectsPresentVariant) + \
	((v) - 1) * sizeof (ObjectIDSequencePair))

#define PRO_PO_OBJ_REQ_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ObjectRequestVariant) + \
	((v) - 1) * sizeof (ObjectID) + ((v) % 2 ? sizeof (short) : 0))

#define PRO_PO_DELETE_OBJS_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (DeleteObjectsVariant) + \
	((v) - 1) * sizeof (ObjectIDWorldStateIDPair))

#define PRO_PO_ALERT_USER_SIZE(p, v) \
	(PRO_PO_HDR_SIZE (p) + sizeof (AlertUserVariant) + \
	((v) - 2 ) * sizeof (char) + ((v) % 2 ? sizeof (char) : 0))

/* **Persistent Object Protocol (Describe Object).  This is a           */
/*   sub-protocol.  Macros are provided for getting the sizes of each   */
/*   class part.  The application must add the class part to size of    */
/*   PDU returned by the PRO_PO_DESC_OBJ_SIZE macro.                    */

/* Class sizes.  For variants, takes  the number of variants to be      */
/* included.                                                            */

#define PRO_PO_WORLD_STATE_CLASS_SIZE(v) \
	(sizeof (WorldStateClass) + \
	((v) - 1) * sizeof (ObjectID) + ((v) % 2 ? sizeof (short) : 0))

#define PRO_PO_OVERLAY_CLASS_SIZE \
	(sizeof (OverlayClass))

#define PRO_PO_POINT_CLASS_SIZE \
	(sizeof (PointClass))

#define PRO_PO_LINE_CLASS_SIZE(v) \
	(sizeof (LineClass) + \
	((v) - 1) * sizeof (PointDescription))

#define PRO_PO_SECTOR_CLASS_SIZE(v) \
	(sizeof (SectorClass) + \
	((v) - 1) * sizeof (long))

#define PRO_PO_TEXT_CLASS_SIZE(v)					     \
	(sizeof (TextClass) +						     \
	 ((v) - 4) * sizeof (char) + ((v) % 4 ? (4-(v)%4) : 0))

#define PRO_PO_UNIT_CLASS_SIZE \
	(sizeof (UnitClass))

#define PRO_PO_LINK_CLASS_SIZE \
        (sizeof (LinkClass))

#define PRO_PO_CIS_CLASS_SIZE \
	(sizeof (CISClass))

#define PRO_PO_STCON_CLASS_SIZE \
	(sizeof (StealthControllerClass))

#define PRO_PO_HHOUR_CLASS_SIZE \
	(sizeof (HHourClass))

#define PRO_PO_COMMO_CLASS_SIZE \
	(sizeof (CommoClass))

#define PRO_PO_TASK_CLASS_SIZE(v)					     \
        (sizeof (TaskClass) +						     \
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_TASK_STATE_CLASS_SIZE(v)					     \
        (sizeof (TaskStateClass) +					     \
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_TASK_FRAME_CLASS_SIZE(v)				             \
        (sizeof (TaskFrameClass) +					     \
	 ((v) - 1) * sizeof (ObjectID) + ((v) % 2 ? sizeof(short) : 0))

#define PRO_PO_TASK_AUTH_CLASS_SIZE(v)					     \
	(sizeof (TaskAuthorizationClass) +				     \
	 ((v) - 4) * sizeof (char) + ((v) % 4 ? (4-(v)%4) : 0))

#define PRO_PO_PI_CLASS_SIZE(v)						     \
        (sizeof (ParametricInputClass) +				     \
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

#define PRO_PO_PIH_CLASS_SIZE(v)					     \
        (sizeof (ParametricInputHolderClass) +				     \
	 ((v) - 1) * sizeof (TaggedParametricInput) + ((v) % 2 ? 0 : sizeof (short)))

#define PRO_PO_FIRE_PARAM_CLASS_SIZE                                         \
        (sizeof (FireParametersClass))

#define PRO_PO_CHEMICAL_CLASS_SIZE \
        (sizeof (ChemicalClass))

#define PRO_PO_MINEFIELD_CLASS_SIZE(v)					\
  (sizeof (MinefieldClass) - sizeof((v)->variant) +			\
   (((v)->minefieldType == minefieldTypePoint) ?			\
    (sizeof(MinefieldPoints) +						\
     (((v)->variant.point.pointCount-1) * sizeof(PointLocation))) :	\
    (((v)->minefieldType == minefieldTypeArea) ?			\
     (sizeof(MinefieldArea) +						\
      (((v)->variant.area.size-4) * sizeof(char) +			\
       (((v)->variant.area.size+2) % 4 ? 				\
	(4-((v)->variant.area.size+2)%4) : 0))) :			\
     (sizeof(MinefieldRow) +						\
      (((v)->variant.row.size-1) * sizeof(char) +			\
       (((v)->variant.row.size) % 4 ?					\
	(4-((v)->variant.row.size) % 4) : 0))))))

#define PRO_PO_MINEFIELD_PARENT_CLASS_SIZE                              \
       (sizeof (MinefieldParentClass))

#define PRO_PO_SIMULATION_REQUEST_CLASS_SIZE(v)				\
       (sizeof (SimulationRequestClass) +				\
	 ((v) - 8) * sizeof (char) + ((v) % 8 ? (8-(v)%8) : 0))

/* **Remote Designation Protocol (fixed length) - These macros take a  */
/*   pointer to a SimulationPDU.                                       */

#define PRO_DES_SERVER_ENT_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorServerEntityVariant))

#define PRO_DES_HANDOFF_ACK_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorHandoffAckVariant))

#define PRO_DES_ADD_TARGET_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorAddTargetVariant))

#define PRO_DES_DELETE_TARGET_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorDeleteTargetVariant))

#define PRO_DES_DESIGNATE_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorDesignateVariant))

#define PRO_DES_STOP_DESIGNATE_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorStopDesignateVariant))

/* **Stingray Protocol (fixed length) - These macros take a            */
/*   pointer to a StingrayPDU.                                         */

#define PRO_STNGRY_DET_REQ_SIZE(p) \
	(PRO_STNGRY_HDR_SIZE (p) + sizeof (SRDetectionRequestVariant))

#define PRO_STNGRY_DET_RESP_SIZE(p) \
	(PRO_STNGRY_HDR_SIZE (p) + sizeof (SRDetectionResponseVariant))

/* **Radio Protocol (variable length) - These macros take a  */
/*   pointer to a RadioPDU and the number of bytes in the    */
/*   dataLength field                             */

#define PRO_RADIO_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_RADIO_INTERCOM_SIZE(p, v) \
	(PRO_RADIO_HDR_SIZE(p) + sizeof (IntercomVariant) + (v)-sizeof(char))

/*
 * To maintain compatibility 
 * laser pdu is added to simnet
 */
#define PRO_SIM_LASER_SIZE(p) \
        (PRO_SIM_HDR_SIZE (p) + sizeof(LaserVariant))

#endif /* _p_size_included */
