
/*
 * pro_size.h
 *
 * Created on Sat Nov 11 10:20:20 1989 
 * using vi and my fingers.
 * 
 */

#ifndef _p_size_included
#define _p_size_included

/*  This file defines macros for determining the size of various PDUs. */
/*  As with all macros (potentially) all bets are off if you use args  */
/*  with side effects.                                                 */

/* These size macros have explicit padding on the end so that if allocating
 * an array of PDU's using malloc (n * XXX_SIZE()), each element in the array
 * is properly aligned.
 */

/* $RCSfile: p_sizeorg.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 */

/*  First, the headers.  These macros take a pointer to the            */
/*  appropriate PDU type.                                              */

#define PRO_ASSOC_DATAGRAM_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_ASSOC_REQUEST_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant) + sizeof ((p)->variant.request) - \
	sizeof ((p)->variant.request.data))

#define PRO_ASSOC_RESPONSE_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant) + sizeof ((p)->variant.response) - \
	sizeof ((p)->variant.response.data))

#define PRO_ASSOC_PADDING_HDR_SIZE(p) \
	(sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_SIM_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_DATA_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_MGMT_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_P2P_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->data))

#define PRO_STEALTH_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_DES_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_SOFTP_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_STNGRY_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

/*  Then the individual PDU types  */


/* **Association Protocol - These macros take a pointer to an   */
/*   AssociationPDU and the length of the data portion.         */

#define PRO_ASSOC_DATAGRAM_SIZE(p, l) \
	(PRO_ASSOC_DATAGRAM_HDR_SIZE (p) + (l))

#define PRO_ASSOC_REQUEST_SIZE(p, l) \
	(PRO_ASSOC_REQUEST_HDR_SIZE (p) + (l))

#define PRO_ASSOC_RESPONSE_SIZE(p, l) \
	(PRO_ASSOC_RESPONSE_HDR_SIZE (p) + (l))

#define PRO_ASSOC_PADDING_SIZE(p, l) \
	(PRO_ASSOC_PADDING_HDR_SIZE (p) + (l))




/* **Simulation Protocol (fixed length) - These macros take a pointer  */
/*   to a SimulationPDU.                                               */

#define PRO_SIM_ACTIVATE_REQ_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ActivateRequestVariant))

#define PRO_SIM_ACTIVATE_RSP_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ActivateResponseVariant))

#define PRO_SIM_DEACTIVATE_REQ_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (DeactivateRequestVariant))

#define PRO_SIM_DEACTIVATE_RSP_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (DeactivateResponseVariant))

#define PRO_SIM_APPEARANCE_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (VehicleAppearanceVariant))

#define PRO_SIM_FIRE_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (FireVariant))

#define PRO_SIM_IMPACT_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ImpactVariant))

#define PRO_SIM_COLLISION_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (CollisionVariant))

#define PRO_SIM_RESUPPLY_CANCEL_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyCancelVariant))

#define PRO_SIM_REPAIR_REQUEST_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RepairRequestVariant))

#define PRO_SIM_REPAIR_RESPONSE_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RepairResponseVariant))

#define PRO_SIM_WEATHER_SIZE(p) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (WeatherVariant))


/* **Simulation Protocol (variable length) - These macros take a       */
/*   pointer to a SimulationPDU and the number of variants to be       */
/*   included.                                                         */


#define PRO_SIM_RADIATE_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RadiateVariant) + \
	((v) - 1) * sizeof (VehicleID) + ((v) % 2 ? 0 : sizeof (int16)))

/* this packet's internals are only 16 bit alligned, NOT 32.  the tail   */
/* end additions cause it to be 32 bit alligned at the end of the packet */
/* every ODD number of targets */
#define PRO_SIM_RADAR_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (RadarVariant) + \
	((v) - 1) * sizeof (RadarTargetData) + (((v) + 1) % 2 ? 0 : sizeof (int16)))

#define PRO_SIM_IND_FIRE_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (IndirectFireVariant) + \
	((v) - 1) * sizeof (IndirectFireDetonation))

#define PRO_SIM_SERVICE_REQ_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyVariant) + \
	((v) - 1) * sizeof (MunitionQuantity))

#define PRO_SIM_RESUPPLY_OFFER_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyVariant) + \
	((v) - 1) * sizeof (MunitionQuantity))

#define PRO_SIM_RESUPPLY_RECEIVED_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (ResupplyVariant) + \
	((v) - 1) * sizeof (MunitionQuantity))

#define PRO_SIM_MARKER_SIZE(p, v) \
	(PRO_SIM_HDR_SIZE (p) + sizeof (MarkerVariant) + \
	((v) - 1) * sizeof (MarkerDescriptor))

/* **Data Collection Protocol (fixed length) - These macros take   */
/*   a pointer to a DataCollectionPDU.                             */

#define PRO_DATA_EXERCISE_STATUS_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (ExerciseStatusVariant))

#define PRO_DATA_VEHICLE_STATUS_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (VehicleStatusVariant))

#define PRO_DATA_STATUS_QUERY_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (StatusQueryVariant))

#define PRO_DATA_STATUS_RESPONSE_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (StatusResponseVariant))

#define PRO_DATA_STATUS_CHANGE_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (StatusChangeVariant))

#define PRO_DATA_LASER_RANGE_SIZE(p) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (LaserRangeVariant))

#define PRO_DATA_SR_IMPACT_RESULT_SIZE(p) \
        (PRO_DATA_HDR_SIZE (p) + sizeof (SR_CMImpactResultVariant))



/* **Data Collection Protocol (variable length) - These macros     */
/*   take a pointer to a DataCollectionPDU,  and the number of     */
/*   variants to include.                                          */

#define PRO_DATA_SIMULATION_STATUS_SIZE(p, v) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (SimulationStatusVariant) + \
	((v) - 1) * sizeof (OrganizationalUnit))

/* subtract sizeof int32, because int32 is the largest element       */
/* type in the structure, so that's what kind of padding is being  */
/* done.                                                           */
#define PRO_DATA_EVENT_FLAG_SIZE(p, v) \
	(PRO_DATA_HDR_SIZE (p) + sizeof (EventFlagVariant) - \
	sizeof (int32) + (v) * sizeof (char))



/* **Management Protocol (fixed length) - These macros take        */
/*   a pointer to a ManagementPDU.                                 */

#define PRO_MGMT_EQUIP_STATUS_SIZE(p) \
	(PRO_MGMT_HDR_SIZE (p) + sizeof (EquipStatusVariant))

#define PRO_MGMT_SHOW_EFFECT_SIZE(p) \
	(PRO_MGMT_HDR_SIZE (p) + sizeof (ShowEffectVariant))

/* **Management Protocol (variable length) - These macros take     */
/*   a pointer to a ManagementPDU,  and the number of variants     */
/*   to include.                                                   */

/* subtract sizeof int16, because int16 is the largest element     */
/* type in the structure, so that's what kind of padding is being  */
/* done.                                                           */
#define PRO_MGMT_ERROR_REPORT_SIZE(p, v) \
	(PRO_MGMT_HDR_SIZE (p) + sizeof (ErrorReportVariant) - \
	sizeof (int16) + (v) * sizeof (char))


/* **Point To Point Protocol - These macros take a pointer to a    */
/*   PointToPointPDU, and the size of the data field.              */

#define PRO_P2P_PDU_SIZE(p, l) \
	(PRO_P2P_HDR_SIZE (p) + (l))


/* **Stealth Protocol (fixed length) - These macros take a pointer */
/*   to a StealthPDU.                                              */

#define PRO_STEALTH_ERROR_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (StealthErrorVariant))

#define PRO_STEALTH_VISIBILITY_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (StealthVisibilityVariant))

#define PRO_STEALTH_TELEPORT_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (TeleportVariant))

#define PRO_STEALTH_ATTACH_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (AttachVariant))

#define PRO_STEALTH_ATTACHED_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (AttachedVariant))

#define PRO_STEALTH_MIMIC_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (MimicVariant))

#define PRO_STEALTH_METAMORPHOSE_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (MetamorphoseVariant))

#define PRO_STEALTH_APPEARANCE_SIZE(p) \
	(PRO_STEALTH_HDR_SIZE (p) + sizeof (StealthAppearanceVariant))

/* **Persistent Object Protocol moved to po_size.h, found in libpo** */


/* **Remote Designation Protocol (fixed length) - These macros take a  */
/*   pointer to a SimulationPDU.                                       */

#define PRO_DES_SERVER_ENT_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorServerEntityVariant))

#define PRO_DES_HANDOFF_ACK_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorHandoffAckVariant))

#define PRO_DES_ADD_TARGET_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorAddTargetVariant))

#define PRO_DES_DELETE_TARGET_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorDeleteTargetVariant))

#define PRO_DES_DESIGNATE_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorDesignateVariant))

#define PRO_DES_STOP_DESIGNATE_SIZE(p) \
	(PRO_DES_HDR_SIZE (p) + sizeof (DesignatorStopDesignateVariant))

/* **Stingray Protocol (fixed length) - These macros take a            */
/*   pointer to a StingrayPDU.                                         */

#define PRO_STNGRY_DET_REQ_SIZE(p) \
	(PRO_STNGRY_HDR_SIZE (p) + sizeof (SRDetectionRequestVariant))

#define PRO_STNGRY_DET_RESP_SIZE(p) \
	(PRO_STNGRY_HDR_SIZE (p) + sizeof (SRDetectionResponseVariant))

/* **Radio Protocol (variable length) - These macros take a  */
/*   pointer to a RadioPDU and the number of bytes in the    */
/*   dataLength field                             */

#define PRO_RADIO_HDR_SIZE(p) (sizeof (*(p)) - sizeof ((p)->variant))

#define PRO_RADIO_INTERCOM_SIZE(p, v) \
	(PRO_RADIO_HDR_SIZE(p) + sizeof (IntercomVariant) + (v)-sizeof(char))

/*
 * To maintain compatibility 
 * laser pdu is added to simnet
 */
#define PRO_SIM_LASER_SIZE(p) \
        (PRO_SIM_HDR_SIZE (p) + sizeof(LaserVariant))

#endif /* _p_size_included */
