
/*
 * FILE:  p_stlth.h 
 *
 * Created on Fri Apr 26 13:42:19 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_stlth_included
#define _p_stlth_included

 /* File:     pro_stlth.drn */
 /* Contents: The Stealth protocol */
 /* System:   simnet development. */
 /* Created:  20-Jan-1989 */
 /* Author:   wkatz */
 /* Remarks: */
 /* $RCSfile: p_stlth.h,v $ */
 /* $Revision: 1.1.1.1 $ */
 /* $State: Exp $ */
 /* COPYRIGHT 1990 BBN Systems and Technologies Corporation */
 /* 10 Moulton Street */
 /* Cambridge, Ma. 02138 */
 /* 617-873-3000 */
#include <stdtypes.h>
#include <basic.h> 
#include <stlth_dynamics.h>

/*
 * Versions of the Stealth protocol:
 */

#define stealthProtocolVersionJan90 1
#define stealthProtocolVersionJan90Corrected 2

typedef uint8    StealthProtocolVersion; 

#define stealthProtocolVersionCurrent           2  /* The current version */

/* Stealth vehicle type */
#define vehicleStealth          1

/*
 * Kinds of Stealth Protocol PDUs:
 */

#define stealthErrorPDUKind      1    
#define stealthVisibilityPDUKind 2   
#define teleportPDUKind          3    
#define attachPDUKind            4    
#define attachedPDUKind          5    
#define mimicPDUKind             6    
#define metamorphosePDUKind      7    
#define stealthAppearancePDUKind 8   

typedef uint8    StealthPDUKind; 


/*
 * The StealthErrorPDU is sent by the Stealth vehicle in response to any cmd
 * from the PVD.  
 *
 * Contents of a Stealth Error PDU:
 */

typedef struct {
    uint16  result;        /* 0 if command successful, positive integer */
                           /* error number if the command failed.  */
    uint16  _unused_3_4;   /* padding */
    uint32  _unused_3;     /* padding */
    char    string[80];    /* Human readable string describing the error. */
} StealthErrorVariant; 


/*
 * Command sent from the PVD to the Stealth Vehicle.  Allows the Stealth
 * to be dynamically sent to any (or all) exercise ID's.
 */

#define viewModeToggleThermal   1    
#define viewModeTogglePolarity  2    

typedef uint8 stealthViewMode; 

typedef struct {
    ExerciseID      exerciseID;           
    stealthViewMode viewMode;        
    uint16          _unused_5_6;           /* padding */
    unsigned        _unused_5   : 24;      /* padding */
} StealthVisibilityVariant; 


/*
 * The TeleportPDU is sent to the Stealth vehicle by the PVD to cause it
 * to move to a new location and/or assume a new orientation.
 */

typedef struct {
    SIMNET_WorldCoordinates location;  /* New location in meters */
    Angle                   azimuth; 
    uint32                  _unused_7;  /* padding */
} TeleportVariant; 


/*
 * Command Issued by the PVD causing the Stealth to become attached to a
 * particular vehicle.
 *
 */

typedef struct {
    VehicleID  vehicleID;  /* to which the Stealth should become attached  */
    uint16     _unused_8;            /* padding */
} AttachVariant; 


/*
 * This packet allows the PVD to show the stealth's current attachment.  There
 * are two cases when it is sent.
 *
 * Sent by the Stealth whenever it becomes attached to a new vehicle
 * (no matter whether this was caused by a packet sent from the PVD or a 
 * Stealth user's input).  Allows the PVD to show the stealth's current 
 * attachment.
 *
 * Also sent by the Stealth vehicle whenever it ceases to be attached to a
 * vehicle, for whatever reason ( a new attach packet from pvd, entering
 * independent velocity mode, attached vehicle vanishing, etc).  In this
 * case, the 'detached' field is set to 1.
 *
 */

typedef struct {
    VehicleID vehicleID;          /* that to which the Stealth has attached  */
    unsigned  detached   : 1;     /* Boolean */
    unsigned  _unused_9  : 15;    /* padding */
} AttachedVariant; 


/*
 * Command from the PVD to the Stealth to show the "Out the window view"
 * from a particular viewport in a particluar simulator.
 *
 */

typedef struct {
    VehicleID  vehicleID;     /* that Stealth is to mimic */
    uint16     channel;       /* CIG channel of mimic'ed vehicle to display */
} MimicVariant; 


/*
 *  Assume the dynamics of another vehicle.
 */
typedef struct {
    vehiclesDynamics vehicleDynamics;  /* Assume the dynamics of this type */
    unsigned         _unused_10_11 : 24;   /* padding */
    uint32           _unused_10;           /* padding */
} MetamorphoseVariant; 


/*
 * Contents of a Vehicle Appearance PDU:
 */

typedef struct {
     /* Identity of the vehicle: */
    VehicleID               vehicleID; 
    VehicleClass            vehicleClass; 
    ForceID                 force;
    VehicleGuises           guises;    /* Appearance of the vehicle: */  
    SIMNET_WorldCoordinates location; 
    float32                 rotation[3][3]; 
    uint32                  appearance; 
    VehicleMarking          marking; 
    uint32                  timestamp; 
    VehicleCapabilities     capabilities; 
    uint16                  engineSpeed; 
    unsigned                stationary    : 1;  /* Boolean */
    unsigned                _unused_12    : 15;  /* padding */

    /* Depending on vehicle class: */
    union {
	/* A simple moving vehicle, without a turret: */
        struct {
            VelocityVector velocity; 
            uint32         _unused_13;  /* padding */
        } simple;
	/* A tank: */
        struct {
            VelocityVector velocity; 
            Angle          turretAzimuth; 
            Angle          gunElevation; 
            uint32         _unused_14;  /* padding */
        } tank;
    } specific;                         
} StealthAppearanceVariant; 

typedef struct {
    StealthProtocolVersion version;  
    StealthPDUKind kind;             
    uint16 _unused_15_16;        /* padding */
    uint32 _unused_15;            /* padding */
    union {
        StealthErrorVariant      stealthError;
        StealthVisibilityVariant stealthVisibility;
        TeleportVariant          teleport;
        AttachVariant            attach;
        AttachedVariant          attached;
        MimicVariant             mimic;
        MetamorphoseVariant      metamorphose;
        StealthAppearanceVariant stealthAppearance;
    } variant;                          
} StealthPDU; 



#endif
