
/*
 * FILE:  status.h 
 *
 * Created on Wed Apr 17 10:53:02 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _status_included
#define _status_included

 /* File:     status.drn */
 /* Contents: The representation of vehicle status information. */
 /* System:   simnet development. */
 /* Created:  10-Oct-1989 */
 /* Author:   apope */
 /* Remarks: */
 /* $RCSfile: status.h,v $ */
 /* $Revision: 1.1.1.1 $ */
 /* $State: Exp $ */
 /* COPYRIGHT 1989 BBN Systems and Technologies Corporation */
 /* 10 Moulton Street */
 /* Cambridge, Ma. 02138 */
 /* 617-873-3000 */

#include <stdtypes.h>
#include <basic.h> 
#include <stat_m1.h> 
#include <stat_m2.h> 
#include <st_leo2.h> 
#include <st_faad.h> 
#include <st_rwa.h> 
#include <status_const.h>

/*
 * Representation of vehicle subsystem status.
 *
 *
 * Status of airframe subsystems: 
 */

typedef struct {
    /* Airframe components: */
    unsigned  airframeMajor     : 1;     /* Boolean */
    unsigned  leftWing          : 1;     /* Boolean */
    unsigned  rightWing         : 1;     /* Boolean */
    unsigned  _unused_8         : 5;     /* padding */

    /* Control surfaces: */
    unsigned  pitchControl      : 1;     /* Boolean */
    unsigned  rollControl       : 1;     /* Boolean */
    unsigned  yawControl        : 1;     /* Boolean */
    unsigned  flaps             : 1;     /* Boolean */
    unsigned  airBrakes         : 1;     /* Boolean */
    unsigned  _unused_9         : 3;     /* padding */

    /* Landing gear: */
    unsigned  landingGearMajor  : 1;     /* Boolean */
    unsigned  noseWheel         : 1;     /* Boolean */
    unsigned  leftWheel         : 1;     /* Boolean */
    unsigned  rightWheel        : 1;     /* Boolean */
    unsigned  _unused_10        : 12;    /* padding */
} AirframeSubsystems; 


/*
 * Status of ground vehicle chassis subsystems:
 */

typedef struct {
    /* Final drive subsystems: */
    unsigned  finalDriveMajor      : 1;  /* Boolean */
    unsigned  leftTrack            : 1;  /* Boolean */
    unsigned  rightTrack           : 1;  /* Boolean */
    unsigned  leftFrontWheel       : 1;  /* Boolean */
    unsigned  rightFrontWheel      : 1;  /* Boolean */
    unsigned  leftRearWheel        : 1;  /* Boolean */
    unsigned  rightRearWheel       : 1;  /* Boolean */
    unsigned  serviceBrake         : 1;  /* Boolean */
    unsigned  parkingBrake         : 1;  /* Boolean */
    unsigned  _unused_11           : 7;  /* padding */
    
    /* Hull subsystems: */
    unsigned  hullMajor            : 1;  /* Boolean */
    unsigned  _unused_12           : 7;  /* padding */

    /* Driver subsystems: */
    unsigned  driversVisionBlocks  : 1;  /* Boolean */
    unsigned  _unused_13           : 7;  /* padding */
} ChassisSubsystems; 


/*
 * Status of aircraft cockpit subsytems:
 */

typedef struct {
    unsigned  cockpitMajor  : 1;         /* Boolean */
    unsigned  _unused_14    : 31;        /* padding */
} CockpitSubsystems; 


/*
 * Status of electronic subsystems:
 */

typedef struct {
    /* Communication subsystems: */
    unsigned  communicationMajor   : 1;  /* Boolean */
    unsigned  radioAntenna         : 1;  /* Boolean */
    unsigned  intercom             : 1;  /* Boolean */
    unsigned  radio                : 1;  /* Boolean */
    unsigned  _unused_15           : 12;  /* padding */

    /* Sensor subsystems: */
    unsigned  laserRangefinder     : 1;  /* Boolean */
    unsigned  electroOpticalMajor  : 1;  /* Boolean */
    unsigned  eoFLIR               : 1;  /* Boolean */
    unsigned  eoDATV               : 1;  /* Boolean */
    unsigned  eoCtlHandle          : 1;  /* Boolean */
    unsigned  radarMajor           : 1;  /* Boolean */
    unsigned  radarTransceiver     : 1;  /* Boolean */
    unsigned  radarTracking        : 1;  /* Boolean */
    unsigned  radarNetwork         : 1;  /* Boolean */
    unsigned  navigationMajor      : 1;  /* Boolean */
    unsigned  _unused_16           : 6;  /* padding */
} ElectronicSubsystems; 


/*
 * Status of electrical and hydraulic power subsytems:
 */

typedef struct {
    /* Electrical power subsystems: */
    unsigned  priElectrical       : 1;   /* Boolean */
    unsigned  priDistributionBox  : 1;   /* Boolean */
    unsigned  secElectrical       : 1;   /* Boolean */
    unsigned  secDistributionBox  : 1;   /* Boolean */
    unsigned  alternator          : 1;   /* Boolean */
    unsigned  generator           : 1;   /* Boolean */
    unsigned  _unused_17          : 10;  /* padding */

    /* Hydraulic power subsystems: */
    unsigned  priHydraulic        : 1;   /* Boolean */
    unsigned  _unused_18          : 15;  /* padding */
} PowerSubsystems; 

/*
 * Status of ground vehicle locomotion subsystems:
 */
typedef struct {
    /* Engine subsystems: */
    unsigned  engineMajor        : 1;    /* Boolean */
    unsigned  pilotRelay         : 1;    /* Boolean */
    unsigned  starter            : 1;    /* Boolean */
    unsigned  oilFilter          : 1;    /* Boolean */
    unsigned  oilLeak            : 1;    /* Boolean */
    unsigned  airCleaner         : 1;    /* Boolean */
    unsigned  coolantLeak        : 1;    /* Boolean */
    unsigned  fuelFilter         : 1;    /* Boolean */
    unsigned  fuelXferPump       : 1;    /* Boolean */
    unsigned  _unused_19         : 7;    /* padding */

    /* Drive train subsystems: */
    unsigned  transmissionMajor  : 1;    /* Boolean */
    unsigned  transFluidFilter   : 1;    /* Boolean */
    unsigned  transFluidLeak     : 1;    /* Boolean */
    unsigned  leftGearbox        : 1;    /* Boolean */
    unsigned  rightGearbox       : 1;    /* Boolean */
    unsigned  universalJoint     : 1;    /* Boolean */
    unsigned  _unused_20         : 10;   /* padding */
} MotiveSubsystems; 


/*
 * Status of tank turret subsystem:
 */

typedef struct {
    unsigned  turretMajor            : 1;  /* Boolean */
    unsigned  turretTraverse         : 1;  /* Boolean */
    unsigned  stabilization          : 1;  /* Boolean */
    unsigned  gunnersCtlHandle       : 1;  /* Boolean */
    unsigned  cmdrsCtlHandle         : 1;  /* Boolean */
    unsigned  turretPositionInd      : 1;  /* Boolean */
    unsigned  turretSlopeInd         : 1;  /* Boolean */
    unsigned  gunnersPrimarySight    : 1;  /* Boolean */
    unsigned  gunnersSecondarySight  : 1;  /* Boolean */
    unsigned  gpsExtension           : 1;  /* Boolean */
    unsigned  gunnersVisionBlocks    : 1;  /* Boolean */
    unsigned  cmdrsVisionBlocks      : 1;  /* Boolean */
    unsigned  loadersPeriscope       : 1;  /* Boolean */
    unsigned  _unused_21             : 19;  /* padding */
} TurretSubsystems; 


/*
 * Status of weapon subsystems:
 */

typedef struct {
    unsigned  priGunMajor      : 1;      /* Boolean */
    unsigned  priGunMount      : 1;      /* Boolean */
    unsigned  priGunElevation  : 1;      /* Boolean */
    unsigned  priGunTraverse   : 1;      /* Boolean */
    unsigned  priGunMisfire    : 1;      /* Boolean */
    unsigned  secGunMajor      : 1;      /* Boolean */
    unsigned  secGunMount      : 1;      /* Boolean */
    unsigned  secGunElevation  : 1;      /* Boolean */
    unsigned  secGunTraverse   : 1;      /* Boolean */
    unsigned  secGunMisfire    : 1;      /* Boolean */
    unsigned  missileMajor     : 1;      /* Boolean */
    unsigned  launcher1        : 1;      /* Boolean */
    unsigned  launcher2        : 1;      /* Boolean */
    unsigned  _unused_22       : 19;     /* padding */
} WeaponSubsystems; 


/*
 * Collection of subsytems making up a generic air vehicle:
 */

typedef struct {
     /* Common to all vehicles: */
    ElectronicSubsystems electronic[2]; 
    MotiveSubsystems     motive[2];  
    PowerSubsystems      power[2];   
    WeaponSubsystems     weapon[2];   /* Specific to air vehicles: */

    AirframeSubsystems   airframe[2]; 
    CockpitSubsystems    cockpit[2]; 
} AirVehicleSubsystems; 


/*
 * Collection of subsytems making up a generic ground vehicle:
 */

typedef struct {
     /* Common to all vehicles: */
    ElectronicSubsystems electronic[2]; 
    MotiveSubsystems     motive[2];  
    PowerSubsystems      power[2];   
    WeaponSubsystems     weapon[2];   /* Specific to ground vehicles: */

    ChassisSubsystems    chassis[2]; 
    TurretSubsystems     turret[2];  
} GroundVehicleSubsystems; 



 /* Overall status of a vehicle's subsystems: */
typedef struct {
    SubsystemsCategory category;     
    unsigned      operationalSummary    : 1;  /* Boolean */
    unsigned      mobilitySummary       : 1;  /* Boolean */
    unsigned      firepowerSummary      : 1;  /* Boolean */
    unsigned      communicationSummary  : 1;  /* Boolean */
    unsigned      noncriticalSummary    : 1;  /* Boolean */
    unsigned      _unused_23            : 11;  /* padding */
    union {
        AirVehicleSubsystems    air;
        GroundVehicleSubsystems ground;
    } subsystems;                       
} VehicleSubsystems; 


/*
 * Status of a generic vehicle:
 */

typedef struct {
    uint8       enginePower;    
    unsigned            batteryVoltage  : 24; 
    MunitionQuantity stores[maxGenericVehicleStores]; 
} GenericVehicleStatus; 

/*
 * Vehicle-specific status of any vehicle:
 */
typedef struct {
    SpecificStatusCategory category; 
    uint16                 _unused_24;           /* padding */
    union {
        GenericVehicleStatus     generic;
        Generic_RWA_Status       rwa;
        SIMNET_M1_Status         m1;
        SIMNET_M2_Status         m2;
        SIMNET_FAAD_LOS_H_Status faad;
        AGPT_Leo2_Status         leo2;
    } specific;                         
} VehicleSpecificStatus; 


/*
 * Representation of overall vehicle status.
 *
 */

typedef struct {
    ObjectType            vehicleType; 
    float32               odometer;  
    uint8                 age;       
    unsigned              _unused_25   : 24;  /* padding */
    VehicleSubsystems     failures;  
    VehicleSpecificStatus specific;  
} VehicleStatus; 



#endif
