/*
 * $RCSfile: libpvd_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpvd_local.h
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libpvd
*       Created: Mon Sep 28 08:10:21 PDT 1992
*       Author: jesmith
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpvd.h"
#include <libvtab.h>
#include <libclass.h>
#include <libtactmap.h>
#include <libsafgui.h>
#include <libcallback.h>
#include <libsensitive.h>
#include <libcoordinates.h>
#include <libeditor.h>
#include <libpktvalve.h>
#include <libsched.h>
#include <X11/Intrinsic.h>
#include <libentity.h>

/* Pixels of mouse motion before it is considered a drag, and not
 * just a wiggle...
 */
#define DRAG_THRESH 20

/* Number of previous views saved */
#define UNDO_BUFFER_LENGTH 128

/* Short-hand debugging macro (assumes local variable vehicle_id) */
#define DEBUG_PVD(format) \
  CLASS_DEBUG((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id), \
	      pvd_user_data_handle, format)

/* Cached data on parts */
typedef struct pvd_last_part
{    
    READER_UNION    *part_entry;
    uint32          artic_state;
    uint32          artic_state_offset;
} PVD_LAST_PART;

/* Per-vehicle pvd class information */
typedef struct pvd_vars
{
    int32 guiseless;
    int32 highlight;
    READER_UNION *pictures;

    READER_UNION *dfn;       /* picture dfn of last tick */
    
    int32 last_num_artics;    /* number of artics last tick */
    PVD_LAST_PART last_artics[ENT_MAX_PARTS]; /* array of last artics present */
    
    uint32 veh_appearance;   /* the vehicles current appearance */

    struct
    {
#define PVD_CALL_SIGN         0
#define PVD_ALTITUDE          1
#define PVD_SPEED             2
#define PVD_DESIGNATIONS      3

	TACTMAP_OBJECT_DFN picture;
	TACTMAP_OBJECT_DFN global_effect;
	TACTMAP_OBJECT_DFN icon;
	TACTMAP_OBJECT_DFN designations[PVD_DESIGNATIONS];
    } tactmap_dfn;
    struct
    {
	TACTMAP_OBJECT_PTR picture;
	TACTMAP_OBJECT_PTR global_effect;
	TACTMAP_OBJECT_PTR icon;
	TACTMAP_OBJECT_PTR designations[PVD_DESIGNATIONS];
    } tactmap_obj[PVD_MAX_PVDS];
    uint32 last_update;
    /* Used to determine when state changed to inactive so can update */
    int32 last_active;


} PVD_VARS;

/* Stealth editor structure */
struct pvd_stealth_editor
{
    int32   needs_initialization;
    int32   _padding;
    int32   mode;
#define STEALTH_MODE_TETHER  0
#define STEALTH_MODE_MIMIC   1
    int32   attachment;
    float64 location[XYC];
    float64 direction;
    float64 altitude;
};

#define PVD_MAX_ALIGNMENTS  8

typedef struct {
    uint32 sea_state;
    uint32 in_surf_zone;
    uint32 breaker_type;
    float32 wave_height;
    float32 wave_direction;
    float32 sea_salinity;
    float32 surface_temp;
    float64 current_velocity[3];
    float64 breaker_angle;
} Sea_Parms;

typedef struct {
    float64 wind_velocity[3]; 
    float64 cloud_cover;
    float64 precip_rate;
    float64 visual_range;
    uint32  extinct_type;
    uint32  precip_type;
} Weather_Parms;

/* Needed for environmental display window */
typedef struct
{
    Widget env_display_form;
    Widget sea_form;        /* Sea/Surf form */
    Widget skyview_da;      /* skyview drawing area */
    Widget illum_da;        /* illumination level drawing area */
    Widget moonphase_da;        /* moon phase drawing area */
    Widget weather_da;        /* textual weather conditions */
    Widget date_da;         /* textual Date/Time */
    Widget sea_da;         /* textual sea/surf conditions */
    Widget sea_label;      /* label for sea or surf conditions */
    GC   sky_gc;
    GC   black_gc;
    GC   skybackground_gc;
    GC   sun_gc;
    GC   moon_gc;
    GC   horizon_gc;
    GC   outer_horizon_gc;
    GC   direct_gc;
    GC   diffuse_gc;
    GC   other_gc;
    GC   label_gc;
    GC   key_gc;
    float64 view_location[XYZC]; /* Coordinates of point in question in */
                              /* world coordinates */
    SCHED_FNCL_PTR update_function;
    float64 last_solar_illum;
    float64 last_lunar_illum;
    float64 last_sky_illum;
    float64 last_total_illum;
    float64 last_sun_pos;  /* altitude and azimuth added up */
    float64 last_moon_pos; /* altitude and azimuth added up */
    uint32  last_uxtime;    /* last unix time we updated */
    Sea_Parms last_sea_parms;
    Weather_Parms last_wthr_parms;
}PVD_ENV_DISPLAY;

/* GUI structure */
struct pvd_gui
{
    SGUI_PTR             gui;
    /* Store PRIV_PTR as convience, can be retrieved thru SGUI_PTR. */
    PRIV_PTR             priv;
    TACTMAP_PTR          tactmap;
    Widget               map_widget;
    GC                   map_erase_gc;
    SNSTVE_WINDOW_PTR    sensitive;
    CALLBACK_EVENT_PTR   map_event;
/* User data in viewable list:
 *    Commanded vehicle = 1
 *    Vehicle spotted by commanded vehicle = 0
 */
    VTAB_LIST            viewable;
    int32                alignments_displayed[PVD_MAX_ALIGNMENTS];

    /* Flag set to suppress updates */
    uint32 frozen;

    /* Feature editor */
    EDT_EDITOR_PTR editor;

    /* Stealth editor */
    EDT_EDITOR_PTR stealth_editor;

    /* Stealth mode */
    SGUI_MODE_PTR stealth_mode;

    /* Preferred stealth */
    SimulationAddress preferred_stealth;

    /* Flag indicating whether all stealths should be displayed
     * (instead of just the preferred one).
     */
    int32 show_all_stealths;

    /* Flag set when we warn the user that the preferred stealth cannot
     * be found.
     */
    int32 preferred_warned;

    /* Feature & scale stuff */
    struct pvd_default_features
    {
	int32 num_scales_plus_5;
	char *database_name;
	int32 initial_scale;
	int32 highest_tree_line;
	int32 highest_tree;
	struct
	{
	    int32 always_5;
	    int32 scale;
	    int32 minor_contour;
	    int32 major_contour;
	    int32 grid_interval;
	} *scales[1];
    } *default_features;
    Widget scale_toggles[25];

/* We put some of the features in known locations */
#define PVD_TREES         0
#define PVD_CONTOURS      1
#define PVD_GRIDS         2
#define PVD_HYPSO         3
#define PVD_WIRE          4
#define PVD_DRAGONS_TEETH 5
#define PVD_MICLIC_LINE   6
#define PVD_MICLIC_SCORCH 7

    struct
    {
	TACTMAP_OBJECT_CLASS class;
	TACTMAP_OBJECT_DFN dfn;
	TACTMAP_OBJECT_PTR obj;
	Widget toggle;
    } features[TACTMAP_LAST_CLASS];

    /* Menu items for Show As */
    Widget show_as_vehicle_pictures;
    Widget show_as_vehicle_icons;
    Widget show_as_aggregate_platoons;
    Widget show_as_aggregate_companies;
    Widget show_as_aggregate_battalions;
    

    /* Menu items for local force */
    Widget local_force_blue;
    Widget local_force_red;

    SNSTVE_CLASS vehicle_class;
    SNSTVE_CLASS stealth_class;

    /* Current map scale, and the corresponding menu item */
    int32 current_scale_index;
    float64 current_scale;

    /* Flag indicating if we are drawing pictures or icons */
#define PVD_SHOW_PICTURES 	0
#define PVD_SHOW_ICONS    	1
#define PVD_SHOW_PLATOONS       2
#define PVD_SHOW_COMPANIES      3
#define PVD_SHOW_BATTALIONS     4
#define PVD_SHOW_NONE     	5

    int32 show_type[6];

    /* Flag indicating that we were forced to turn off entities,
     * and thus should turn them back on if we get a chance.
     */
    int32 reenable_show_entities;

    /* Flag indicating how do display altitudes */
    int32 altitudes_agl;

    /* Whether contour intervals should be fixed */
    int32 fixed_contours;

    /* Approximate size of vehicle/unit icons (in pixels) */ 
    float32 raw_icon_scale;
    int32 icon_size;
    CALLBACK_EVENT_PTR icon_change_event;

    /* Rate at which we would like to refresh vehicles (ms) */
    uint32 veh_update_rate;

    /* Magnification of vehicles */
    float64 veh_magnification;

    /* Flag which is true when the magnification shouldn't be adjusted */
    int32 veh_mag_dont_adjust;

    /* Raw scale (range:0-1, used to offset markings) */
    float64 raw_scale;

    /* Selected markings (bit mask) */
    uint32 markings;

    /* Fixed contour intervals */
    int32 fixed_minor_contour;
    int32 fixed_major_contour;

    /* GCs used to draw vehicles */
    GC black_gc;
    GC white_gc;
    GC force_gc[5];
    GC hit_gc;
    GC miss_gc;
    GC indf_gc;
    GC mine_gc;
    GC coll_gc;
    GC mf_gc;
    GC opmf_gc;
    GC lane_gc;
    GC miclic_gc;
    GC laser_gc;
    GC shading_gc[19];

    /* Font used for designations */
    XFontStruct *font_struct;

    /* Map modes */
    SGUI_MODE_PTR current;
    SGUI_MODE_PTR zoom;
    SGUI_MODE_PTR pan;
    SGUI_MODE_PTR info;
    SGUI_MODE_PTR env;

    /* Mode state variables */
    int32 dragging;
    XRectangle last_rect;
    GC zoom_gc;
    int32 info_vehicle;

    /* Undo buffer */
    struct
    {
	float64 center_x;
	float64 center_y;
	int32   center_c;
	float64 scale;
    } undo_buffer[UNDO_BUFFER_LENGTH];
    int32 undo_count;

    /* Electro-magnetic band to use to draw environmental effects */
    int32 em_band;

    /* Determines if smoke clouds are drawn filled or in outline mode */
    int32 smoke_display;

    /* Specifies zoom level to stop displaying the minefield tactical symbol */
    int32 mf_sym_cutoff;

#define PVD_SHOW_COLLISION       0
#define PVD_SHOW_FIRE            1
#define PVD_SHOW_IMPACT	         2
#define PVD_SHOW_FOR             3
#define PVD_SHOW_RADIO_TRAFFIC   4
#define PVD_SHOW_LINKS           5
#define PVD_SHOW_LASER_DSG       6

#define PVD_EVENTS               7
    int32 draw_collisions;
    int32 draw_fires;
    int32 draw_impacts;
    int32 draw_for;             /* Field of Regard */
    int32 draw_links;
    int32 draw_radio_traffic;
    int32 draw_laser_designations;

    /* fired when either of the last two change */
    CALLBACK_EVENT_PTR display_settings_changed_event;
    /* fired when the user zooms (but not pans) */
    CALLBACK_EVENT_PTR zoom_event;

    /* if a minefield is on this list, don't display it in too close. */
    VTAB_LIST mf_list;
    /* Needed for environmental display window */
    PVD_ENV_DISPLAY           env_display;

};

/* copied this out of pvd_impact.c so that prototypes could compile */

/* Generic PDU structure used for impact, detonation, indirect fire,
 * and collisions.
 */
#define PVD_MAX_DETONATIONS 20
typedef struct
{
    uint8           result;
    uint8           dis_result;
    int32           firer_id;
    int32           target_id;
    float64         firing_location[XYZ];
    float32         munition_velocity[3];
    BurstDescriptor burst;
    struct pvd_detonation
    {
	uint32  delay;
	float64 location[XYZ];
    } detonations[PVD_MAX_DETONATIONS];
} PVD_IMPACT;

/* PDU for radio signals */
#define PVD_MAX_MSG 1024
typedef struct
{
    int32   hashed_id;
    NETID48 unhashed_id;
    int32   length_bits;
    uint16  encoding_scheme;
    char    data[PVD_MAX_MSG];
} PVD_SIGNAL;

typedef struct
{
    int32   lasing_id;
    int32   lased_id;
    int32   laser_code;
    float64 lasing_location[XYZC];
    float64 lased_location[XYZC];
    uint32  time_last_pdu;
    SCHED_FNCL_PTR erase_fcn;
} PVD_LASER;


/* VERY PRIVATE VARIABLES AND FUNCTIONS!  DO NOT USE OUTSIDE! */

extern int32 PVD_ASPDU;
extern int32 pvd_modstealth;
extern int32 pvd_user_data_handle;
extern void  pvd_init_params(void);
extern void  pvd_zoom_mode(
    SGUI_PTR gui, SGUI_MODE_PTR mode,
    PVD_GUI_PTR pvd_gui, SGUI_MODE_STATE state);
extern void  pvd_pan_mode(
    SGUI_PTR    gui, SGUI_MODE_PTR mode,
    PVD_GUI_PTR pvd_gui, SGUI_MODE_STATE state);
extern void  pvd_info_mode(
    SGUI_PTR    gui, SGUI_MODE_PTR mode,
    PVD_GUI_PTR pvd_gui, SGUI_MODE_STATE state);
extern void  pvd_env_mode(
    SGUI_PTR    gui, SGUI_MODE_PTR mode,
    PVD_GUI_PTR pvd_gui, SGUI_MODE_STATE state);
extern void  pvd_set_scale(PVD_GUI_PTR pvd_gui, float64 scale);
extern void  pvd_set_scale_index(PVD_GUI_PTR pvd_gui, int32 index);
extern int32 pvd_n_pvds;
extern PVD_GUI_PTR pvd_pvds[];
extern int32 pvd_db_key;
extern void  pvd_process_impact(
    PV_PACKET  *packet, int32 format,
    PVD_IMPACT *pdu, PVD_GUI_PTR pvd_gui);
extern void  pvd_process_radio(
    PV_PACKET  *packet, int32 format,
    PVD_SIGNAL *pdu, PVD_GUI_PTR pvd_gui);
extern void pvd_process_laser(
    PV_PACKET  *packet, 
    int32       format,
    PVD_LASER   *pdu,
    PVD_GUI_PTR  pvd_gui);
extern void  pvd_click(
    SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
    SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui);
extern READER_UNION pvd_stealth_picture;
extern READER_UNION pvd_smoke_picture;
extern READER_UNION pvd_parts;
extern void  pvd_stealth_selected(
    SNSTVE_WINDOW_PTR window,
    SNSTVE_ACTION     action,
    SNSTVE_CALL_DATA *call_data,
    PVD_GUI_PTR       pvd_gui);
extern void  pvd_stealth_apply(
    EDT_EDITOR_PTR             editor, int32 transient,
    struct pvd_stealth_editor *old,
    struct pvd_stealth_editor *new,
    PVD_GUI_PTR                pvd_gui);
extern void  pvd_stealth_done(
    EDT_EDITOR_PTR             editor,
    struct pvd_stealth_editor *new,
    PVD_GUI_PTR                pvd_gui, EDT_EXIT_STATUS status);
extern void  pvd_remember_where(PVD_GUI_PTR pvd_gui);
extern void  pvd_init_scrollbar(PVD_GUI_PTR pvd_gui, Widget sb);
extern void  pvd_update_scrollbars(PVD_GUI_PTR pvd_gui);
extern int32 pvd_okay_to_display (
    PVD_GUI_PTR pvd,
    ForceID     force_id,
    int32       vehicle_id,
    int32       commanded_only );

extern void pvd_init_subclass(void);
extern int32 pvd_impact_fmt;
extern int32 pvd_indirect_fmt;
extern int32 pvd_collision_fmt;
extern int32 pvd_fire_fmt;
extern int32 pvd_radio_fmt;
extern int32 pvd_srdetect_fmt;
extern int32 pvd_srimpact_fmt;
extern int32 pvd_laser_fmt;
extern void pvd_init_pdu(void);
extern void pvd_create_env_display(
    Widget           parent,
    PVD_GUI_PTR      pvd_gui,
    Pixel            foreground,
    Font             font,
    Window           root,
    PVD_ENV_DISPLAY *env_display);

extern char *pvd_plow_cat;    /* reader symbol name for a plow */
extern char *pvd_bridge_cat;    /* reader symbol name for a bridge */
extern char *pvd_dogbone_cat;   /* reader symbol name for a dogbone */
extern char *pvd_grizzlyshovel_cat;  /* symbol name for the grizzly shovel */
extern char *pvd_dirtmover_cat;  /* symbol name for the a "dirtmover" */
extern char *pvd_shovel_base;   /* symbol name for the shovel's base */
