#ifndef lint
static char rcsid [] = "$RCSfile: pvd_class.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_class.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Manages class information attached to each vehicle        *
*       Created: Mon Sep 28 08:10:21 PDT 1992
*       Author: jesmith
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <X11/Intrinsic.h>
#include <dis_types.h>
#include <libdisdbobj.h>
#include <libaccess.h>
#include <libaggrmap.h>
#include <libotmatch.h>
#include <libdisdbobj.h>
#include <libentity.h>
#include <libenvent.h>
#include <libmarkent.h>
#include <libphysdb.h>
#include <libreader.h>
#include <libprotocol.h>
#include <libtime.h>
#include <libvecmat.h>
#include <libctdb.h>
#include <libgcs.h>
#include <librfpiedt.h>
#include <libminemgr.h>
#ifndef MODSTEALTH
#include <libstealth.h>
#endif
#include <math.h>
#include <veh_appear.h> 
#include <obj_type.h> 
#include <unit_type.h> 
#include <veh_type.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/
#include <stdstring.h> /*common/include/global*/
#include <ctype.h>


/* Length of buffer for entity description: */
#define PVD_BUF_LEN 256

int32 pvd_user_data_handle;

static void pvd_print(CLASS_USER_DATA_TYPE vars);
#ifdef USE_CLCGF
static void pvd_process_aggregate(int32 vehicle_id);
#endif /* USE_CLCGF */
static void pvd_draw_minefield(Widget widget, Pixmap drawable,
			       ADDRESS draw_arg, XRectangle *boundary);

/* Offsets to various designations (in pixels) */
static struct
{
    int32 h_base, h_mul;
    int32 v_base, v_mul;
} offsets[] = { { -5 , -100 ,  0 ,    0 } ,
		{ -5 , -100 , -5 , -100 } ,
		{  5 ,  100 , -5 , -100 } ,
	    };

static void pvd_add_remote_subclass(int32 vehicle_id, ADDRESS user_data)
{
    /* This is a remote vehicle being added to the system.
     * Unlike locals, parmgr won't be adding subclasses to this object
     * based on parameters.  Intead, this library (libpvd) always
     * knows to add itself to these particular objects.
     */
    pvd_create(vehicle_id, (ADDRESS) NULL, (ADDRESS) NULL, "");
}

static void pvd_remove_remote_subclass(int32 vehicle_id, int32 is_migration,
				       ADDRESS user_data)
{
    /* This is a remote vehicle being deleted from the system.
     * Unlike locals, parmgr won't be deleting subclasses from this object
     * based on parameters.  Intead, this library (libpvd) always
     * knows to delete itself from these particular objects.
     */
    pvd_destroy(vehicle_id, is_migration);
}

void pvd_init_subclass(void)
{
    pvd_user_data_handle =
      class_reserve_user_data(disdbobj_class, "pvd", pvd_print);
    
    /*
     * Register callbacks to create this subclass when remote vehicles
     * are created.
     */
    callback_register_handler(disdbobj_create_event, 
			      pvd_add_remote_subclass, NULL);
    callback_register_handler(disdbobj_destroy_event, 
			      pvd_remove_remote_subclass, NULL);
    
    
    /* Register a callback to tick libpvd when the simulation clock
     * is frozen (for all vehicles)
     */
    callback_register_handler(disdbobj_frozen_tick_event,
			      pvd_tick, (ADDRESS)NULL);
}

/*ARGSUSED*/
static void pvd_print(CLASS_USER_DATA_TYPE vars)
{
    printf(" nothing to print\n");
}

/*ARGSUSED*/
void pvd_create(int32 vehicle_id, ADDRESS params, ADDRESS unit_entry,
		char *name_symbol)
{
    PVD_VARS *pvd;
    int32 i, j;
    VehicleID vid;
    static SNSTVE_CLASS null_class;
    TACTMAP_OBJECT_CLASS tmap_obj_class;

    if (!pvd_n_pvds)
      return;
    
    /* Only attempt to show stuff we want to display */
    if (!(VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),
			   (VTAB_VEHICLE | VTAB_MISSILE | VTAB_STRUCTURE |
			    VTAB_STEALTH | VTAB_ENV | VTAB_MINE_MARKER |
			    VTAB_AGGREGATE | VTAB_MINEFIELD | VTAB_MINE_SET))))
      return;
    
    pvd = (PVD_VARS *)STDALLOC(sizeof(PVD_VARS));
    
    bzero((char *)pvd, sizeof(PVD_VARS));
    
    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
                        pvd_user_data_handle,
                        (CLASS_USER_DATA_TYPE)pvd);
    
    if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_STEALTH))
    {
        /* Initialize the stuff which doesn't change from tick to tick */
        pvd->tactmap_dfn.picture.sensitive.instance = vehicle_id;
        pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
        pvd->tactmap_dfn.picture.picture.highlight = FALSE;
	
        for (i=0;i<pvd_n_pvds;i++)
        {
            pvd->tactmap_dfn.picture.sensitive.class =
              &pvd_pvds[i]->stealth_class;
	    
            pvd->tactmap_obj[i].picture =
              tactmap_create_object(pvd_pvds[i]->tactmap,
                                    TACTMAP_PICTURE,
                                    &pvd->tactmap_dfn.picture);
	    
            vtab_vehicle_simid(vehicle_id, &vid);
            if (pvd_pvds[i]->show_all_stealths ||
                NS_SIM_ADDRESSES_EQUAL(pvd_pvds[i]->preferred_stealth,
                                       vid.simulator))
              tactmap_add_transient_object(pvd_pvds[i]->tactmap,
                                           pvd->tactmap_obj[i].picture);
        }
    }
#if 0
    else if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_ENV))
    {
        pvd->tactmap_dfn.picture.sensitive.instance = vehicle_id;
        pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
        pvd->tactmap_dfn.picture.picture.highlight = FALSE;

	/* Initialize this here in case a subclass does not have */
	/* a picture init function */
	tmap_obj_class = TACTMAP_PICTURE;
        envent_init_picture(vehicle_id,
                           &pvd->tactmap_dfn.picture.picture,
			    &tmap_obj_class);

        
        for (i=0;i<pvd_n_pvds;i++)
        {
            pvd->tactmap_dfn.picture.sensitive.class =
              &pvd_pvds[i]->vehicle_class;
	    
            pvd->tactmap_obj[i].picture =
              tactmap_create_object(pvd_pvds[i]->tactmap,
                                    tmap_obj_class,
         (tmap_obj_class== TACTMAP_PICTURE) ? &pvd->tactmap_dfn.picture : &pvd->tactmap_dfn.global_effect);
	    
            tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					 pvd->tactmap_obj[i].picture);
        }
    }
#endif

    else if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_MINE_MARKER))
    {
	/* For mine markers and breach markers, make a line */
	
        pvd->tactmap_dfn.picture.sensitive.instance = vehicle_id;
        pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
	
	/* markers aren't sensitive, so use a null class */
	bzero((char *)&null_class, sizeof(null_class));
	pvd->tactmap_dfn.picture.sensitive.class = &null_class;
	
	pvd->tactmap_dfn.picture.linear.style = LSplain;
	pvd->tactmap_dfn.picture.linear.dashed = FALSE;
	pvd->tactmap_dfn.picture.linear.thickness = 2;
	pvd->tactmap_dfn.picture.linear.backward_arrow = noArrowHead;
	pvd->tactmap_dfn.picture.linear.forward_arrow = noArrowHead;
	pvd->tactmap_dfn.picture.linear.width = 1;
	
	pvd->tactmap_dfn.picture.linear.n_vertices = 0;
	pvd->tactmap_dfn.picture.linear.vertices = NULL;
	
	/* put the line in the appropriate tactmaps */
        for (i=0;i<pvd_n_pvds;i++)
        {
	    pvd->tactmap_dfn.picture.linear.gc = 
	      pvd_pvds[i]->black_gc;
	    
            pvd->tactmap_obj[i].picture =
              tactmap_create_object(pvd_pvds[i]->tactmap,
                                    TACTMAP_LINEAR,
                                    &pvd->tactmap_dfn.picture);
	    
	    if(pvd_okay_to_display(pvd_pvds[i],
				   markent_get_force_id(vehicle_id),
				   vehicle_id, FALSE))
	      tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					   pvd->tactmap_obj[i].picture);
        }
    }
    else
      pvd->guiseless = 1;
    
    /* Markings aren't sensitive, so initialize them to point to a faked
     * up class.
     */
    bzero((char *)&null_class, sizeof(null_class));
    for (j=0;j<PVD_DESIGNATIONS;j++)
    {
        pvd->tactmap_dfn.designations[j].sensitive.class = &null_class;
        pvd->tactmap_dfn.designations[j].sensitive.instance =
          (vehicle_id << 3) | j;
    }
    
    /*
     * Register callback to tick during the appropriate phase at the
     * appropriate rate
     */
    if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_LOCAL_STEALTH))
      callback_register_handler(disdbobj_get_tick_event(vehicle_id,
							DISDBOBJ_GUI, 250),
				pvd_tick,
				(ADDRESS) NULL);
    else
      callback_register_handler(disdbobj_get_tick_event(vehicle_id,
							DISDBOBJ_GUI, 0),
				pvd_tick,
				(ADDRESS) NULL);
}


void pvd_destroy(int32 vehicle_id, int32 is_migration)
{
    int32 i, j;
    PVD_VARS *pvd = (PVD_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id),
			  pvd_user_data_handle);
    
    if (!pvd) /* Passive error detection */
      return;
    
    for (i=0;i<pvd_n_pvds;i++)
    {
	if (pvd->tactmap_obj[i].picture)
	  tactmap_destroy_object(pvd_pvds[i]->tactmap,
				 pvd->tactmap_obj[i].picture);
	if (pvd->tactmap_obj[i].icon)
	  tactmap_destroy_object(pvd_pvds[i]->tactmap,
				 pvd->tactmap_obj[i].icon);
	
	
	for (j=0;j<PVD_DESIGNATIONS;j++)
	  if (pvd->tactmap_obj[i].designations[j])
	    tactmap_destroy_object(pvd_pvds[i]->tactmap,
				   pvd->tactmap_obj[i].designations[j]);
    }
    
    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			pvd_user_data_handle,
			(CLASS_USER_DATA_TYPE)0);
    
    STDDEALLOC(pvd);
}    

static void get_vitals(PVD_GUI_PTR pvd_gui, int32 vehicle_id, char *buf,
	int32 buf_len)
{
    int32 ndx, heading_degrees;
    char veh_str[8];
    VehicleID pid;
    VehicleID vid;
    VehicleMarking marking;
    float64 pos[XYZC], pitch, roll, speed, heading;
    float32 downwind_distance, extent[3];
    uint32 simtime;
    MunitionQuantity munition;
    
    vtab_vehicle_simid(vehicle_id, &vid);
    
    if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),
			 (VTAB_VEHICLE | VTAB_MISSILE | VTAB_STRUCTURE |
			  VTAB_STEALTH | VTAB_AGGREGATE)))
    {
	ent_get_marking(vehicle_id, &marking);
	ent_set_context(ent_get_cell(vehicle_id));
	ent_get_position(vehicle_id,pos);
	pos[CELL3D]=ent_get_cell(vehicle_id);
	speed = ent_get_speed(vehicle_id);
	ent_get_orientation(vehicle_id, &heading, &pitch, &roll);
	heading_degrees = (int32)RAD_TO_DEG(heading);
	if(heading_degrees < 0)
	  heading_degrees += 360;
	marking.text[maxVehicleMarkingLength-1] = 0;
	
	if(VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id), VTAB_AGGREGATE))
	{
	  ent_get_aggregate_extent(vehicle_id, extent);
	  edt_format(pvd_gui->gui, buf,
		     "Aggregate ID: #D/#D/#D | Marking: #S | Location: #l | Speed: #s | Heading: #D | Width: #D Depth: #D Height: #D",
		     vid.simulator.site, vid.simulator.host,
		     vid.vehicle, marking.text,
		     (int32)pos[CELL3D], pos[X], pos[Y], speed,
		     heading_degrees, (int)extent[0],
		     (int)extent[1], (int)extent[2]);
	}
	else if( rfpiedt_find_unit_id( vid.vehicle, veh_str, &ndx ) != -1 )
	  if(((ent_get_guise(vehicle_id, distinguishedForceID) &
	       objectDomainMask) != objectDomainVehicle) ||
	     ((ent_get_guise(vehicle_id, distinguishedForceID) &
	       vehicleEnvironmentMask) != vehicleEnvironmentAir))
	    edt_format(pvd_gui->gui, buf,
		       "Vehicle ID: #D/#D/#D | Unit ID: #S #D | Marking: #S | Type: #S | Location: #l | Speed: #s",
		       vid.simulator.site, vid.simulator.host,
		       vid.vehicle, veh_str, ndx, marking.text,
		       protocol_find_name(ent_get_guise(vehicle_id, 
							distinguishedForceID)),
		       (int32)pos[CELL3D], pos[X], pos[Y], speed);
	  else
	    edt_format(pvd_gui->gui, buf,
		       "Vehicle ID: #D/#D/#D | Unit ID: #S #D | Marking: #S | Type: #S | Location: #l | Speed: #s | Altitude: #h | Pitch: #D | Roll: #D",
		       vid.simulator.site, vid.simulator.host,
		       vid.vehicle, veh_str, ndx, marking.text,
		       protocol_find_name(ent_get_guise(vehicle_id, 
							distinguishedForceID)),
		       (int32)pos[CELL3D], pos[X], pos[Y], speed, pos[Z],
		       (int)RAD_TO_DEG(pitch), (int)RAD_TO_DEG(roll));
	else
	  if(((ent_get_guise(vehicle_id, distinguishedForceID) &
	       objectDomainMask) != objectDomainVehicle) ||
	     ((ent_get_guise(vehicle_id, distinguishedForceID) &
	       vehicleEnvironmentMask) != vehicleEnvironmentAir))
	    edt_format(pvd_gui->gui, buf,
		       "Vehicle ID: #D/#D/#D | Marking: #S | Type: #S | Location: #l | Speed: #s",
		       vid.simulator.site, vid.simulator.host,
		       vid.vehicle, marking.text,
		       protocol_find_name(ent_get_guise(vehicle_id, 
							distinguishedForceID)),
		       (int32)pos[CELL3D], pos[X], pos[Y], speed);
	  else
	    edt_format(pvd_gui->gui, buf,
		       "Vehicle ID: #D/#D/#D | Marking: #S | Type: #S | Location: #l | Speed: #s | Altitude: #h | Pitch: #D | Roll: #D",
		       vid.simulator.site, vid.simulator.host,
		       vid.vehicle, marking.text,
		       protocol_find_name(ent_get_guise(vehicle_id, 
							distinguishedForceID)),
		       (int32)pos[CELL3D], pos[X], pos[Y], speed, pos[Z],
		       (int)RAD_TO_DEG(pitch), (int)RAD_TO_DEG(roll));
    }
    else if(VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id), VTAB_ENV))
    {
	char *description;

	if (envent_get_description(vehicle_id, buf, buf_len) != ENVENT_SUCCESS)
	  edt_format(pvd_gui->gui, buf,
		     "Environmental Effect ID: #D/#D/#D",
		     vid.simulator.site, vid.simulator.host,
		     vid.vehicle);

    }
    else if(VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id), VTAB_MINEFIELD))
    {
	edt_format(pvd_gui->gui, buf,
		   "Minefield ID: #D/#D/#D",
		   vid.simulator.site, vid.simulator.host,
		   vid.vehicle);
    }


}

/*ARGSUSED*/
void pvd_click(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
	       SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui)
{
    int32 vehicle_id;
    char buf[PVD_BUF_LEN];
    VehicleID vid;
    
    
    if (((call_data->class == &pvd_gui->vehicle_class) ||
	 (call_data->class == &pvd_gui->stealth_class)) &&
	(vehicle_id = (int32)(call_data->user_data)))
      switch (call_data->button)
      {
	case 1:
	  /* Callback here */
	  
	  break;
	case 2:
	  switch(action)
	  {
	    case SNSTVE_PRESS:
	      vtab_vehicle_simid(vehicle_id, &vid);
	      if (vtab_vehicle_type(vehicle_id)==VTAB_REMOTE_STEALTH)
		sprintf(buf, "Stealth ID: %d/%d/%d",
			vid.simulator.site, vid.simulator.host,
			vid.vehicle);
	      else if (vtab_vehicle_type(vehicle_id)==VTAB_LOCAL_STEALTH)
	      {
		  if (pvd_modstealth)
		    sprintf(buf, "Stealth ID: %d/%d/%d",
			    vid.simulator.site, vid.simulator.host,
			    vid.vehicle);
#ifndef MODSTEALTH
		  else
		    sprintf(buf, "Preview #%d",
			    stealth_get_preview_number(vehicle_id));
#endif
	      }
	      else
	      {
		  get_vitals(pvd_gui, vehicle_id, buf, PVD_BUF_LEN);
		  pvd_gui->info_vehicle = vehicle_id;
	      }
	      sgui_push_help(pvd_gui->current, buf);
	      break;
	    case SNSTVE_RELEASE:
	      pvd_gui->info_vehicle = 0;
	      sgui_pop_help(pvd_gui->current, TRUE);
	      break;
	  }
	  break;
      }
}

int32 pvd_okay_to_display(PVD_GUI_PTR pvd, ForceID force_id,
			  int32 vehicle_id, int32 commanded_only)
{
    /* If we are displaying vehicles for that force (irregardless of 
     * commanding)
     *    Then it is okay to display the annotations for this vehicle.
     */
    if (priv_get_local_force(pvd->priv, force_id))
    {
        return TRUE;
    }
    
    /* Now checked viewable list. */
    
    /* If the vehicle is not in the viewable list, it is not commanded
     * or spotted by a commanded vehicle so don't display data for it.
     */
    if (!vtab_in_list(pvd->viewable, vehicle_id))
    {
        return FALSE;
    }
    /* If don't care whether vehicle is commanded
     *    Okay to display data since in viewable list.
     */
    if (!commanded_only)
    {
        return TRUE;
    }
    /* Only want to display if vehicle is commanded, so check user data
     *    Commanded vehicle = 1
     *    Vehicle spotted by commanded vehicle = 0
     */
    if ((VTAB_USER_DATA_TYPE)vtab_get_user_data(pvd->viewable, 
                                                vehicle_id))
    {
        return TRUE;
    }
    return FALSE;
}

/* Initializes the env fields at tick time because some of the envent */
/* info isn't available yet at creation */
static void pvd_init_env(
    PVD_VARS            *pvd,
    int32                vehicle_id,
    TACTMAP_OBJECT_DFN **env_tactmap_dfn)
{
    int32 i;
    TACTMAP_OBJECT_CLASS tmap_obj_class;
    static SNSTVE_CLASS null_class;
    

    /* Initialize this here in case a subclass does not have */
    /* a picture init function */
    tmap_obj_class = TACTMAP_PICTURE;
    envent_init_picture(vehicle_id,
			&pvd->tactmap_dfn.picture.picture,
			&tmap_obj_class);

    switch (tmap_obj_class) {
      case TACTMAP_GLOBAL_EFFECT:
	*env_tactmap_dfn = &pvd->tactmap_dfn.global_effect;
	pvd->tactmap_dfn.global_effect.global_effect.highlight = FALSE;
	break;
      case TACTMAP_PICTURE:
      default:
	*env_tactmap_dfn = &pvd->tactmap_dfn.picture;
	pvd->tactmap_dfn.picture.picture.highlight = FALSE;
	break;
    }

#if 0
    pvd->tactmap_dfn.picture.sensitive.instance = vehicle_id;
    pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
    pvd->tactmap_dfn.picture.picture.highlight = FALSE;
#endif

    (*env_tactmap_dfn)->sensitive.instance = vehicle_id;
    (*env_tactmap_dfn)->sensitive.user_data = (ADDRESS)vehicle_id;

    for (i=0;i<pvd_n_pvds;i++)
    {
#if 0	
	pvd->tactmap_dfn.picture.sensitive.class =
	  &pvd_pvds[i]->vehicle_class;
#endif
	bzero((char *)&null_class, sizeof(null_class));
	(*env_tactmap_dfn)->sensitive.class = &null_class;
	  
	if (tmap_obj_class == TACTMAP_PICTURE)
	  pvd->tactmap_obj[i].picture =
	    tactmap_create_object(pvd_pvds[i]->tactmap,
				  tmap_obj_class,
				  &pvd->tactmap_dfn.picture);
	else
	{
	    bzero((char *)&null_class, sizeof(null_class));
	    pvd->tactmap_dfn.global_effect.sensitive.class = &null_class;
	    pvd->tactmap_obj[i].global_effect =
	      tactmap_create_object(pvd_pvds[i]->tactmap,
				    tmap_obj_class,
				    &pvd->tactmap_dfn.global_effect);
	}
	
	    tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					 (tmap_obj_class== TACTMAP_PICTURE) ? pvd->tactmap_obj[i].picture : pvd->tactmap_obj[i].global_effect
					 /* pvd->tactmap_obj[i].picture */);
    }

}

void pvd_tick(int32 vehicle_id, uint32 interval_since_last_tick,
	      ADDRESS user_data)
{
    PVD_VARS *pvd;
    ObjectType guise;
    VehicleGuises  *guises;
    READER_UNION *new_dfn = NULL, *old_dfn = NULL;
    int32 i, j, id;
    int32 artic_state;
    float64 effectiveness;
    int32 shovel_active = FALSE;
    float64 pos[XYZC], rot[XYZ][XYZ], *dir;
    ForceID force_id;
    uint32 appearance;
    PHYSDB_DATA *pdb;
    float64 speed;
    VehicleMarking marking;
    struct tactmap_picture *pict;
    struct tactmap_bgr *icon;
    uint32 m;
    char alt_buf[20], spd_buf[20];
    int32 is_stealth, is_env, is_mark, is_mf;
    int32 show_picture;
    int32 do_display;
    int32 became_inactive = FALSE;
    float64 locs[MARKENT_MAX_MARKERS_PER_PDU][XY];
    float64 temp_locs[MARKENT_MAX_MARKERS_PER_PDU][XYZ];
    int32 num;
    ENVENT_PICTURE_DATA envent_data;
    static SNSTVE_CLASS null_class;
    char *deploy_name;
    ENT_PART artics[ENT_MAX_PARTS];
    int32 ap;
    int32 current_num_artics = 0;
    PVD_LAST_PART current_artics[ENT_MAX_PARTS];
    int32 artics_changed = 0;
    READER_UNION *part_entry;
    int32 offset;
    READER_UNION *old, *tmp; 
    char *tmp_buf;
    int32 picture_offset = 0;    
    
    char *deploy_FWMP_name;
    char *deploy_AVLB_present_name;

    TACTMAP_OBJECT_DFN *env_tactmap_dfn;
    TACTMAP_OBJECT_CLASS tmap_obj_class;


    /* Someone before me may have just deactivated me prior to this tick
     * event being fired.  Check this before continuing.
     */
    if (!vtab_get_vehicle(vehicle_id))
      return;
#ifdef USE_CLCGF    
    if (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_AGGREGATE))
    {
	if (PVD_ASPDU)
	  pvd_process_aggregate(vehicle_id);
	return;
    }
#endif /* USE_CLCGF */    

    pvd = (PVD_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id),
			  pvd_user_data_handle);
    
    if (!pvd)					/* Passive error detection */
      return;
    
    /* init variables */
    is_env = VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_ENV);
    is_stealth = VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_STEALTH);
    is_mark = VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_MINE_MARKER);
    is_mf = VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_MINEFIELD);
    
    /* If this is an environment entity or a normal entity which is
     * not active
     *    If the entity was active last time 
     *       Need to force an update no matter whether the update time 
     *       has been reached to cause the entity to be removed from 
     *       the pvd since will not execute the rest of the code 
     *       normally for inactive entities.
     *    Else the entity was inactive last time as well and has 
     *    already been removed from the display
     *       Since nothing needs to be updated for an inactive entity,
     *       just return.  Do not need to go through the following for
     *       loop to check each pvd, since each will have been updated
     *       when the entity transitioned to inactive.
     * Else this entity is active (or stealth is considered active)
     *    So need to continue on and see if have reached update time.
     * What this is saying is we only update if an entity switches 
     * from active to inactive or if the entity is active and the
     * update time has been reached.
     */
    if ((is_env && !envent_active(vehicle_id)) ||
        (!is_env && !is_stealth && !is_mark && !is_mf &&
	 !ent_active(vehicle_id)))
    {
        if (pvd->last_active)
        {
            pvd->last_active = FALSE;
            became_inactive  = TRUE;
        }
        else
        {
            pvd->last_active = FALSE;
            return;
        }
    }
    else
    {
        pvd->last_active = TRUE;
    }
    
    /* Before we continue, make sure somebody wants to be updated */
    for (i=0;i<pvd_n_pvds;i++)
      if (!pvd_pvds[i]->frozen &&
	  (pvd->last_update + pvd_pvds[i]->veh_update_rate <
	   time_last_realtime_clock) || became_inactive)
	break;
    if (i == pvd_n_pvds)
      return;
    
    if (is_env)
    {
	/* Initialize this here in case a subclass does not have */
	/* a picture init function */
	tmap_obj_class = TACTMAP_PICTURE;
	envent_init_picture(vehicle_id,
			    &pvd->tactmap_dfn.picture.picture,
			    &tmap_obj_class);
	switch (tmap_obj_class) {
	  case TACTMAP_GLOBAL_EFFECT:
	    env_tactmap_dfn = &pvd->tactmap_dfn.global_effect;
	    pvd->tactmap_dfn.global_effect.global_effect.highlight = FALSE;
	    break;
	  case TACTMAP_PICTURE:
	  default:
	    env_tactmap_dfn = &pvd->tactmap_dfn.picture;
	    pvd->tactmap_dfn.picture.picture.highlight = FALSE;
	    break;
	}

	guise = envent_get_guise(vehicle_id);
    }
    else if (is_mark)
      guise = markent_get_guise(vehicle_id, distinguishedForceID);
    else if (is_mf)
      guise = (ObjectType)minemgr_get_appearance(vehicle_id);
    else
      guise = ent_get_guise(vehicle_id, distinguishedForceID);
    
    /* Stealths and env and marks get all this information right away, but 
     * vehicles wait until it becomes available.
     */
    if (pvd->guiseless)
    {
	/* See if we have a guise yet */
	if (!guise)
	  return;
	
	pvd->guiseless = 0;
	if (is_mf)
	{
	    static SNSTVE_CLASS null_class;
	    
	    /* Minefields are drawn objects */
	    bzero((char *)&null_class, sizeof(null_class));
	    pvd->tactmap_dfn.picture.sensitive.class = &null_class;
	    pvd->tactmap_dfn.picture.sensitive.instance = vehicle_id;
	    pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
	    pvd->tactmap_dfn.picture.drawn.draw = pvd_draw_minefield;
	    pvd->tactmap_dfn.picture.drawn.erase_on_delete_only = TRUE;
	    pvd->tactmap_dfn.picture.drawn.draw_arg = (ADDRESS)vehicle_id;
	    
	    for (i=0;i<pvd_n_pvds;i++)
	    {
		pvd->tactmap_obj[i].picture =
		  tactmap_create_object(pvd_pvds[i]->tactmap, TACTMAP_DRAWN,
					&pvd->tactmap_dfn.picture);
		
		if(pvd_okay_to_display(pvd_pvds[i], 
				       minemgr_get_force_id(vehicle_id),
				       vehicle_id, FALSE))
		  tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					       pvd->tactmap_obj[i].picture);
	    }
	}
	else if (is_env)
	    pvd_init_env(pvd, vehicle_id, &env_tactmap_dfn);
	else
	{	
	    /* Initialize the stuff which doesn't change from tick to tick */
	    pvd->tactmap_dfn.picture.sensitive.instance  = vehicle_id;
	    pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
	    
	    /* Need some initial stuff */
	    pvd->tactmap_dfn.icon.sensitive.instance  = vehicle_id;
	    pvd->tactmap_dfn.icon.sensitive.user_data = (ADDRESS)vehicle_id;
	    
	    pvd->pictures = (READER_UNION *)otm_query(pvd_db_key, guise);
	    
	    /* set the initial appearance to some unlikely value..  this value
	     * will be compared with the real vehicles appearance, and it is
	     * imperitive that these values not be equal initially..
	     */
	    
	    pvd->veh_appearance = (uint32) -1;
	
	    pvd->tactmap_dfn.designations[PVD_CALL_SIGN].text.see_through = 0;
	    pvd->tactmap_dfn.designations[PVD_CALL_SIGN].text.alignment = west;
	    pvd->tactmap_dfn.designations[PVD_ALTITUDE].text.see_through = 0;
	    pvd->tactmap_dfn.designations[PVD_ALTITUDE].text.alignment =
	      southWest;
	    pvd->tactmap_dfn.designations[PVD_SPEED].text.see_through = 0;
	    pvd->tactmap_dfn.designations[PVD_SPEED].text.alignment =
	      southEast;
	    
	    bzero((char*)&null_class, sizeof(null_class));
	
	    for (i=0;i<pvd_n_pvds;i++)
	    {
		if ((VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),
				      VTAB_STRUCTURE)) &&
		    (ent_get_guise(vehicle_id, forceIDIrrelevant) ==
		     structure_Ditch))
		{
		    /* ditches aren't sensitive, so use a null class */
		    pvd->tactmap_dfn.picture.sensitive.class = &null_class;
		}
		else
		{
		    pvd->tactmap_dfn.picture.sensitive.class =
		      &pvd_pvds[i]->vehicle_class;
		    pvd->tactmap_dfn.icon.sensitive.class =
		      &pvd_pvds[i]->vehicle_class;
		    
		    pvd->tactmap_obj[i].icon =
		      tactmap_create_object(pvd_pvds[i]->tactmap,
					TACTMAP_BGR,
					    &pvd->tactmap_dfn.icon);
		}
		
		pvd->tactmap_obj[i].picture =
		  tactmap_create_object(pvd_pvds[i]->tactmap,
				    TACTMAP_PICTURE,
					&pvd->tactmap_dfn.picture);
				
		for (j=0;j<PVD_DESIGNATIONS;j++)
		  pvd->tactmap_obj[i].designations[j] =
		    tactmap_create_object(pvd_pvds[i]->tactmap,
				      TACTMAP_TEXT,
					  pvd->tactmap_dfn.designations+j);
	    }
	}
    }
    
    pict = &pvd->tactmap_dfn.picture.picture;
    icon = &pvd->tactmap_dfn.icon.bgr;
    
    /* ensure that all information from libentity is returned in the
     * current vehicles context.
     */
    ent_set_context(ent_get_cell(vehicle_id));
    pos[CELL3D] = ent_get_cell(vehicle_id);

    if (!is_stealth && !is_env && !is_mark && !is_mf)
    {
        float64 shovel_base_azimuth = 0.0;
	access_get(vehicle_id,
		   ent_rotation, rot,
		   ent_force_id, &force_id,
		   ent_physdb, &pdb,
		   ent_appearance, &appearance,
		   ent_speed, &speed,
		   ent_marking, &marking,
		   ent_position, pos,
		   A_END);	
	    
	/* first, get the vehicle's pictures based upon its appearance 
	 * pvd->pictures is the array for this object type.  start out
	 * by just copying the pointer to the picture...
	 */
	
	switch (appearance &
		(vehDestroyed | vehFlaming |
		 vehMobilityDisabled | vehFirepowerDisabled))
	{
	  case 0:  /* Healthy */
	    new_dfn = pvd->pictures[1].array;
	    picture_offset = 1;
	    break;
	  case vehMobilityDisabled:
	    new_dfn = pvd->pictures[2].array;
	    picture_offset = 2;
	    break;
	  case vehFirepowerDisabled:
	    new_dfn = pvd->pictures[3].array;
	    picture_offset = 3;
	    break;
	  case vehMobilityDisabled | vehFirepowerDisabled:
	    new_dfn = pvd->pictures[4].array;
	    picture_offset = 4;
	    break;
	  default: /* catastrophic */
	    new_dfn = pvd->pictures[5].array;
	    picture_offset = 0;
	    break;
	}


	/* Get all the artic parts, and determine if there is a
	 * match in pvd_parts.rdr for display.
	 */
	bzero(artics, sizeof(ENT_PART)*ENT_MAX_PARTS);
	bzero(current_artics, sizeof(PVD_LAST_PART)*ENT_MAX_PARTS);
	
	ent_get_all_artic(vehicle_id, artics);
	ap = 0;
	while (artics[ap].name)
	{
	    /* Determine whether we have an picture for this part */
	    part_entry = reader_find_tag(artics[ap].name, pvd_parts.array,
					 READER_UNTAGGED, READER_NO_ERRORS);
	    if (part_entry)
	    {	    
		/* We are trying to weed out duplicates by filtering on
		 * the metric.  Note that the dogbone is an exception, 
		 * since it uses the EXTENSION metric for damage and 
		 * doesn't have any other metric.
		 */
		if ((artics[ap].type == ENT_ARTICULATED) &&
		    (artics[ap].u.articulated.metric == ENT_AZIMUTH ||
		     artics[ap].u.articulated.metric == ENT_ELEVATION ||
		     artics[ap].artic_category == pvd_dogbone_cat))
		{
		    /* Determine whether this is a new artic since 
                     * last time 
		     */
		    for (i=0; i<pvd->last_num_artics; i++)
		      if (part_entry == pvd->last_artics[i].part_entry)
			break;
		    
		    /* Save the current artic */
		    current_artics[current_num_artics].part_entry 
		      = part_entry;

		    /* If its a plow or dogbone, get its effectiveness for
		     * damage.
		     */
		    if ((artics[ap].artic_category == pvd_plow_cat) ||
			(artics[ap].artic_category == pvd_dirtmover_cat) ||
			(artics[ap].artic_category == pvd_dogbone_cat))
		    {			
			ent_get_artic_value(vehicle_id, artics[ap].name, 
					    ENT_EXTENSION,
					    &effectiveness);
			artic_state = (1.0 - effectiveness > 0.5);
			
			/* If this is different than last time or this
			 * is the first time for this part, get offset.
			 */
			if ((i>=pvd->last_num_artics) ||
			    (pvd->veh_appearance != appearance) ||
			    (artic_state != pvd->last_artics[i].artic_state))
			{
			    /* its the first time or something has
			     * changed.
			     */
			    current_artics[current_num_artics].artic_state = 
			      artic_state;
			    artics_changed = TRUE;
			    if (picture_offset)
			      current_artics[current_num_artics].
				artic_state_offset = artic_state*4 + 
				  picture_offset;
			    else
			      current_artics[current_num_artics].
				artic_state_offset = 9;
			}
			else
			{
			    /* Update current in case artics array changed.
			     * This is necessary for artics that are
			     * changed, removed, added, or reordered.
			     */
			    current_artics[current_num_artics].artic_state = 
			      pvd->last_artics[i].artic_state;
			    current_artics[current_num_artics].
			      artic_state_offset
				= pvd->last_artics[i].artic_state_offset;
			}			
		    }
		    else if (artics[ap].artic_category == 
			     pvd_grizzlyshovel_cat)
		    {
			shovel_active = 1;
			ent_get_artic_value(vehicle_id, pvd_shovel_base,
					    ENT_AZIMUTH, &shovel_base_azimuth);
			
			/* the shovel is not active if the base is at 
			 * an angle of 45 degrees of less.  There are 
			 * two pictures for an active grizzly shovel.  
			 * One is for 45 <= angle < 135  (shovel_active=1)
			 * and the other is for 135 <= angle 
			 * (shovel_active=2).  This provides a slightly 
			 * better animation than using only two 
			 * pictures.  shovel_active has the following meaning:
			 *   shovel_active == 1  ==> healthy and stowed
			 *   shovel_active == 2  ==> healthy and in position 1
			 *   shovel_active == 3  ==> healthy and in position 2
			 *   shovel_active == -1 ==> damaged and stowed
			 *   shovel_active == -2 ==> damaged and in position 1
			 *   shovel_active == -3 ==> damaged and in position 2
			 *   shovel_active == 0  ==> no shovel attached.
			 */
			shovel_active += 
		            (shovel_base_azimuth >= PI_OVER_TWO * 0.5) +
			    (shovel_base_azimuth >= PI_OVER_TWO * 1.5);

			if (appearance & vehShovelDisabled)
			  shovel_active *= -1;

			/* If this is different than last time or this
			 * is the first time for this part, get offset.
			 */
			if ((i>=pvd->last_num_artics) ||
			    (shovel_active != pvd->last_artics[i].artic_state))
			{
			    current_artics[current_num_artics].artic_state 
			      = shovel_active;
			    artics_changed = TRUE;			
			    if (shovel_active < 0)
			      current_artics[current_num_artics].
				artic_state_offset = 3 - shovel_active;
			    else
			      current_artics[current_num_artics].
				artic_state_offset = shovel_active;
			}	
			else
			{			    
			    /* Update current in case artics array changed */
			    current_artics[current_num_artics].artic_state = 
			      pvd->last_artics[i].artic_state;
			    current_artics[current_num_artics].
			      artic_state_offset = 
			      pvd->last_artics[i].artic_state_offset;
			}
		    }
		    else
		    {
			
			/* Check appropriate appearance bit for damage */
			artic_state =  ((appearance & part_entry[2].integer)
					== part_entry[2].integer);
			
			/* If this is different than last time or this
			 * is the first time for this part, get offset.
			 */
			if ((i>=pvd->last_num_artics) ||
			    (pvd->veh_appearance != appearance) ||
			    (artic_state != pvd->last_artics[i].artic_state))
			{
			    current_artics[current_num_artics].artic_state 
			      = artic_state;
			    artics_changed = TRUE;			    
			    if (picture_offset)
			      current_artics[current_num_artics].
				artic_state_offset = artic_state*4 + 
				  picture_offset;
			    else
			      current_artics[current_num_artics].
				artic_state_offset = 9;
			}
			else
			{			    
			    /* Update current in case artics array changed */
			    current_artics[current_num_artics].
			      artic_state_offset
			      = pvd->last_artics[i].artic_state;
			    current_artics[current_num_artics].
			      artic_state_offset = 
			      pvd->last_artics[i].artic_state_offset;
			}

		    }
		    current_num_artics++;
		}
	    }
	    ap++;
	}
	    
	dir = rot[Y];
	
	for (j=0;j<PVD_DESIGNATIONS;j++)
	{
	    pvd->tactmap_dfn.designations[j].text.x = pos[X];
	    pvd->tactmap_dfn.designations[j].text.y = pos[Y];
	    pvd->tactmap_dfn.designations[j].text.cell = pos[CELL3D];
	}

	pvd->tactmap_dfn.designations[PVD_CALL_SIGN].text.string =
	  (char *)marking.text;

	/* We fill in the alt_buf and spd_buf on a per-gui basis, below */
	pvd->tactmap_dfn.designations[PVD_ALTITUDE].text.string = alt_buf;
	pvd->tactmap_dfn.designations[PVD_SPEED].text.string = spd_buf;
	

	/* the picture used during the last tick is used, unless something
         * has changed that effects it...
         */

        if ((pvd->veh_appearance != appearance) ||
	    (pvd->last_num_artics != current_num_artics) ||
	    artics_changed)
        {

	    /* something has changed.  therefore we must free up the
             * previous picture definition and create a new one..
	     */

	    int32 space_allocated = FALSE;
	    READER_UNION *pictures[ENT_MAX_PARTS];


            /* save a copy of the old dfn pointer for use later.. 
             * note: only save the copy if there was previously 
	     * memory allocated, which will occur when there were
	     * pictures present for artic parts.
             */

	    if (pvd->dfn && pvd->last_num_artics)
                old_dfn = pvd->dfn;

	    /* update the pvd class vars.. */
	    pvd->veh_appearance = appearance;
	    pvd->last_num_artics = current_num_artics;
	    bcopy((char *)&current_artics, (char *)&pvd->last_artics, 
		  sizeof(PVD_LAST_PART)*ENT_MAX_PARTS);

	    /* now, create the picture definition from all the subpart
	     * definitions and the vehicle picture.  Put the vehicle
	     * picture into the first position in an array, followed by
  	     * a pointer to all the part pictures and combine them into
	     * one definition..
	     */
	    pictures[0] = new_dfn;

	    /* Add each icon that is attached */
	    for (i=0; i<current_num_artics; i++)
                pictures[i+1] = current_artics[i].part_entry[3].
                    array[current_artics[i].artic_state_offset].array;

	    if (current_num_artics)
                pvd->dfn = reader_combine_arrays(pictures, 
					         current_num_artics+1);
	    else
		pvd->dfn = pictures[0];
        }
	
	pict->num_csystems = 3;
	
	/* First coordinate system is compass */
	pict->csystems[0].parent_system = 0; /* world */
	pict->csystems[0].angle         = 0.0;
	pict->csystems[0].dimensions[X] = 1.0;
	pict->csystems[0].dimensions[Y] = 1.0;
	pict->csystems[0].origin[X]     = pos[X];
	pict->csystems[0].origin[Y]     = pos[Y];
	pict->csystems[0].cell          = pos[CELL3D];
	
	/* Second coordinate system is hull */
	pict->csystems[1].parent_system = 1; /* world */
	if (dir[X] || dir[Y])
	  pict->csystems[1].angle = -NS_ATAN2(dir[X], dir[Y]);
	else
	  pict->csystems[1].angle = 0.0;
	    
	pict->csystems[1].dimensions[X] = pdb->dimensions[X];
	pict->csystems[1].dimensions[Y] = pdb->dimensions[Y];
	pict->csystems[1].origin[X]     = pos[X];
	pict->csystems[1].origin[Y]     = pos[Y];
	pict->csystems[1].cell          = pos[CELL3D];

	/* Scale the width by the roll to give the impression of 3D.
	 */
	pict->csystems[1].dimensions[X] *=
	  rot[X][X] * rot[X][X] + rot[X][Y] * rot[X][Y];
	
	/* Third coordinate system is turret */
	pict->csystems[2].parent_system = 1; /* hull */
	if (pdb->turrets->count)
	{
	    ent_get_artic_value(vehicle_id, pdb->turrets->turret[0]->name,
				ENT_AZIMUTH, 
				&pict->csystems[2].angle);
	    pict->csystems[2].origin[X] =
	      pdb->turrets->turret[0]->location[X];
	    pict->csystems[2].origin[Y] =
	      pdb->turrets->turret[0]->location[Y];
	    pict->csystems[2].cell = GCS_ILLEGAL_CELL;
	}
	else
	{
	    pict->csystems[2].angle     = 0.0;
	    pict->csystems[2].origin[X] = 0.0;
	    pict->csystems[2].origin[Y] = 0.0;
	    pict->csystems[2].cell      = GCS_ILLEGAL_CELL;
	    
	}

	pict->csystems[2].dimensions[X] = pdb->dimensions[X] * 0.5;

	if (pdb->turrets->count && pdb->turrets->turret[0]->guns->count)
	  pict->csystems[2].dimensions[Y] =
	    pdb->turrets->turret[0]->guns->gun[0]->length;
	else
	  pict->csystems[2].dimensions[Y] = 0.0;
	
	/* Setting up icon data */
	icon->x            = pos[X];
	icon->y            = pos[Y];
	icon->cell         = pos[CELL3D];
	icon->dashed       = FALSE;
	icon->call_sign[0] = '\0';
	icon->object_type  = guise;
	icon->override     = FALSE;
	icon->dfn          = NULL;
	icon->rotation     = -NS_ATAN2(dir[X], dir[Y]);
	
	/* Snap the rotation to be multiples of PI/4 */
	icon->rotation = 
	  floor((icon->rotation + (PI/8.0))/(PI/4.0)) * (PI/4.0);
    }
    else if (is_stealth)		/* STEALTH */
    {
	pict->dfn = pvd_stealth_picture.array;
	pict->num_csystems = 1;
	
#ifndef MODSTEALTH
	if (!pvd_modstealth)
	  /* Note: stealth information is all in a single context
	   * and can not have individual information set (other than 
	   * teleportation which is handled in libstealth.  Therefore, no
	   * reason to set a context for stealth data.
	   */
	  stealth_get_position(vehicle_id, pos);
	else
#endif
	  ent_get_position(vehicle_id,pos);
	
	/* Stealth coordinate system is hull-oriented */
	pict->csystems[0].parent_system = 0; /* world */
	
#ifndef MODSTEALTH
	if (!pvd_modstealth)
	  pict->csystems[0].angle = stealth_get_direction(vehicle_id);
	else
#endif
	{
	    float64 direction[XYZ];
	    ent_get_direction(vehicle_id, direction);

	    if (direction[X] || direction[Y])
              pict->csystems[0].angle = 
                -NS_ATAN2(direction[X], direction[Y]);
	    else
              pict->csystems[0].angle = 0;
	}
	pict->csystems[0].dimensions[X] = 1.0;
	pict->csystems[0].dimensions[Y] = 1.0;
	pict->csystems[0].origin[X] = pos[X];
	pict->csystems[0].origin[Y] = pos[Y];
    }
    else if (is_env)			/* Environmental effect */
    {
	envent_get_position(vehicle_id, pos);

	if (tmap_obj_class == TACTMAP_PICTURE)
	{
	    pict->num_csystems = 1;
	    pict->csystems[0].parent_system = 0; /* world */
	    pict->csystems[0].angle = envent_get_direction(vehicle_id);
	    pict->csystems[0].dimensions[X] = 1.0;
	    pict->csystems[0].dimensions[Y] = 1.0;
	    pict->csystems[0].origin[X] = pos[X];
	    pict->csystems[0].origin[Y] = pos[Y];
	    pict->magnification = 0.0; /* pixels */
	}
	else
	{
	    env_tactmap_dfn->global_effect.num_csystems = 1;
	    env_tactmap_dfn->global_effect.csystems[0].parent_system = 0; /* world */
	    env_tactmap_dfn->global_effect.csystems[0].angle = envent_get_direction(vehicle_id);
	    env_tactmap_dfn->global_effect.csystems[0].dimensions[X] = 1.0;
	    env_tactmap_dfn->global_effect.csystems[0].dimensions[Y] = 1.0;
	    env_tactmap_dfn->global_effect.csystems[0].origin[X] = pos[X];
	    env_tactmap_dfn->global_effect.csystems[0].origin[Y] = pos[Y];
	    env_tactmap_dfn->global_effect.magnification = 0.0; /* pixels */

	}


	for (i = 0; i< pvd_n_pvds; i++)
	{
	    /* The EnvEnt client may override any of the picture settings */
	    envent_data.smoke_display = pvd_pvds[i]->smoke_display;
	    envent_data.em_band = tactmap_get_em_band(pvd_pvds[i]->tactmap);
	    if (tmap_obj_class == TACTMAP_PICTURE)
		envent_get_picture(vehicle_id, (struct tactmap_picture *)pict,
				   &envent_data);
	    else
	      envent_get_picture(vehicle_id, 
		    (struct tactmap_picture *)&env_tactmap_dfn->global_effect, 
				 &envent_data);
	     
	}

    }
    else if (is_mark)  /* Mine marker or breached lane marker */
    {	
	force_id = markent_get_force_id(vehicle_id);
        if (guise == structure_BreachedLaneMarker)
          pvd->tactmap_dfn.picture.linear.style = LSbreachLaneMarker;
        else
          pvd->tactmap_dfn.picture.linear.style = LSminefieldMarker;
	    
        pvd->tactmap_dfn.picture.linear.dashed         = FALSE;
        pvd->tactmap_dfn.picture.linear.thickness      = 2;
        pvd->tactmap_dfn.picture.linear.backward_arrow = noArrowHead;
        pvd->tactmap_dfn.picture.linear.forward_arrow  = noArrowHead;
        pvd->tactmap_dfn.picture.linear.width          = 1;
	
        /* Position the markers */
	num = markent_get_flag_locations(vehicle_id,
					 MARKENT_MAX_MARKERS_PER_PDU,
					 temp_locs);
	
	for (i=0; i<num; i++)
	{
	    locs[i][X] = temp_locs[i][X];
	    locs[i][Y] = temp_locs[i][Y];
	}
	
        pvd->tactmap_dfn.picture.linear.n_vertices = num;
        pvd->tactmap_dfn.picture.linear.vertices = (float64 *)locs;
	
    }
    else if (is_mf)
    {
	force_id = minemgr_get_force_id(vehicle_id);
    }
    
    if (!is_mark)
    {
	pvd->tactmap_dfn.picture.picture.highlight = pvd->highlight;
	pvd->tactmap_dfn.icon.bgr.highlight        = pvd->highlight;
    }
    
    for (i=0;i<pvd_n_pvds;i++)
    {
	if (!pvd_pvds[i]->frozen &&
	    (pvd->last_update + pvd_pvds[i]->veh_update_rate >
	     time_last_realtime_clock) && !became_inactive)
	  continue;

	pvd->last_update = time_last_realtime_clock;
	
        /* Get whether it is okay to display vehicle annotations or not. */
        do_display = 
	  pvd_okay_to_display(pvd_pvds[i], force_id, vehicle_id, FALSE);
	
	if (!is_stealth && !is_env && !is_mark && !is_mf)
	{
	    /* Local vehicles come and go (especially missiles), so check
	     * the active flag to see if the vehicle should even be displayed.
	     * (Note that it costs practically nothing to re-add or re-remove
	     * an object, so we don't bother checking that here.)
	     */
	    if (!ent_active(vehicle_id))
	    {
		tactmap_remove_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].picture);
		tactmap_remove_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].icon);
		for (j=0;j<PVD_DESIGNATIONS;j++)
		  tactmap_remove_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].
					designations[j]);
		continue;
	    }
	    
	    /* Dispatch on whether this PVD wants vehicle pictures,
	     * vehicle icons, or neither.
	     */
	    if (pvd_pvds[i]->show_type[PVD_SHOW_PICTURES])
	      show_picture = TRUE;
	    else if (pvd_pvds[i]->show_type[PVD_SHOW_ICONS])
	      show_picture = FALSE;
	    else
	    {
		/* I'm not drawing these now, so remove them.
		 * Note that it doesn't hurt to remove something not
		 * currently on the display list.
		 */
		tactmap_remove_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].picture);
		tactmap_remove_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].icon);
		for (j=0;j<PVD_DESIGNATIONS;j++)
		  tactmap_remove_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].
					designations[j]);
		continue;
	    }
	    
	    if (show_picture)
	    {
		/* ditch structures aren't sensitive, so use a null class */
		if ((VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),
				      VTAB_STRUCTURE)) &&
		    (guise == structure_Ditch))
		  pict->sensitive.class = &null_class;
		else
		  pict->sensitive.class = &pvd_pvds[i]->vehicle_class;
		pict->magnification = pvd_pvds[i]->veh_magnification;

		/* Keep the ship/sub picture size comparable to tanks, planes, etc. */
                if (((guise &
		      (objectDomainMask    | vehicleEnvironmentMask)) ==
		     (objectDomainVehicle | vehicleEnvironmentWater)) &&
		    
		    ((guise != vehicle_US_RHIB) && 
		     (guise != vehicle_US_EOD)  &&
		     (guise != vehicle_US_MK4)  &&
		     (guise != vehicle_US_MK7)))
		  
		  pict->magnification *= 0.125 ;
		
		/* Keep sweep gear pitcure size smaller than ships.JLK 04/02/96
		 * Magnification was changed in statement above.   JLK 04/02/96
		 * Adjust mildly.   			    JLK 04/02/96 */
		if ((guise & (objectDomainMask|  	 /* JLK 04/02/96 */
			      vehicleEnvironmentMask|vehicleClassMask)) ==
		    (objectDomainVehicle|vehicleEnvironmentWater|
		     vehicleClassMineSweepGear))	
		  pict->magnification *= 0.600;		
		
		if ((pict->magnification < 1.0) || 
		    (VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),
				      VTAB_STRUCTURE)))
		  pict->magnification = 1.0;
		
		pict->num_gcs = 3;
		pict->gcs[0] = pvd_pvds[i]->black_gc;
		pict->gcs[1] = pvd_pvds[i]->white_gc;
		pict->gcs[2] = pvd_pvds[i]->force_gc[force_id];

   	        /* use the newly created definition as the picture */ 
                pict->dfn = pvd->dfn;
		
		tactmap_update_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].picture,
				      &pvd->tactmap_dfn.picture);
	
               /* dealloc the old picture if one exists..  note that we can
                * dealloc this reader union array simply by passing it to
                * STDDEALLOC because we only allocated the top-most array
                * in the reader union (using reader_combine_arrays)..
                */

                if (old_dfn)
                    STDDEALLOC(old_dfn);
	
		/* Now check the viewable list to see if the pictures
		 * should be displayed.
		 */
		if (do_display)
		{
		    tactmap_add_transient_object(pvd_pvds[i]->tactmap,
						 pvd->tactmap_obj[i].picture);
		    tactmap_remove_object(pvd_pvds[i]->tactmap,
					  pvd->tactmap_obj[i].icon);
		    
		}
		else 
		{
		    tactmap_remove_object(pvd_pvds[i]->tactmap,
					  pvd->tactmap_obj[i].icon);
		    
		    tactmap_remove_object(pvd_pvds[i]->tactmap,
					  pvd->tactmap_obj[i].picture);
		}
		
	    }
	    else
	    {
		icon->sensitive.class = &pvd_pvds[i]->vehicle_class;
		icon->size = pvd_pvds[i]->icon_size;
		icon->line_thickness = 3;

		switch (appearance &
			(vehDestroyed | vehFlaming |
			 vehMobilityDisabled | vehFirepowerDisabled))
		{
		  case 0:		/* Healthy */
		    icon->gc = pvd_pvds[i]->force_gc[force_id];
		    break;

		  case vehMobilityDisabled:
		  case vehFirepowerDisabled:
		  case vehMobilityDisabled | vehFirepowerDisabled:
		  default:		/* catastrophic */
		    icon->gc = pvd_pvds[i]->black_gc;
		    break;
		}
		
		tactmap_update_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].icon,
				      &pvd->tactmap_dfn.icon);
		/* Now check the viewable list to see if the icons
		 * should be displayed.
		 */
                if (do_display)
		{
		    tactmap_add_transient_object(pvd_pvds[i]->tactmap,
						 pvd->tactmap_obj[i].icon);
		    tactmap_remove_object(pvd_pvds[i]->tactmap,
					  pvd->tactmap_obj[i].picture);
		    
		    
		}
		else
		{
		    tactmap_remove_object(pvd_pvds[i]->tactmap,
					  pvd->tactmap_obj[i].picture);
		    
		    tactmap_remove_object(pvd_pvds[i]->tactmap,
					  pvd->tactmap_obj[i].icon);
		}
		
	    }
	    
	    
	    /* Suppress all markings on munitions, suppress ATC markings
	     * on non-air vehicles.
	     */
	    m = pvd_pvds[i]->markings;

	    if ((guise & objectDomainMask) == objectDomainMunition)
	      m = 0;
	    else if ((guise & (objectDomainMask|vehicleEnvironmentMask)) !=
		     (objectDomainVehicle | vehicleEnvironmentAir))
	      m &= ~(1 << PVD_ALTITUDE);
	    
	    /* Altitudes are shown in hundreds of feet, unless the
	     * defaults select otherwise
	     */
	    if (pvd_pvds[i]->altitudes_agl)
	    {
		edt_format(pvd_pvds[i]->gui, alt_buf, "#h",
			   ent_get_altitude_agl(vehicle_id));
		/* Remove units display (ft, meters, etc.) */
		strtok(alt_buf, " ");
	    }
	    else
	      sprintf(alt_buf, "%d", (int)(pos[Z] / 30.48 + 0.5 /*ROUND*/));
	    
	    edt_format(pvd_pvds[i]->gui, spd_buf, "#s", speed);

	    /* need to extract the numerical speed from spd_buf, but
	     * ignore the units.  due to mach, the first field is not
	     * always the number (ie, "mach 1.5" instead of "55 kph")
	     * so we need to search for it 
	     */
	    tmp_buf = strtok(spd_buf, " ");
	    if (!isdigit(tmp_buf[0]))
	      tmp_buf = strtok(NULL, " ");
	    strcpy(spd_buf, tmp_buf);
	    
	    for (j=0;j<PVD_DESIGNATIONS;j++)
	    {
                /* Remove if not okay to display.
		 * Note that it doesn't hurt to remove something not
		 * currently on the display list.
                 */
		if (do_display && (m & (1 << j)))
		  tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					       pvd->tactmap_obj[i].
					       designations[j]);
		else
		  tactmap_remove_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].
					designations[j]);
		
		pvd->tactmap_dfn.designations[j].text.gc = 
		  pvd_pvds[i]->black_gc;
		pvd->tactmap_dfn.designations[j].text.font_struct =
		  pvd_pvds[i]->font_struct;
		pvd->tactmap_dfn.designations[j].text.horizontal_offset =
		  offsets[j].h_base + 
		    offsets[j].h_mul * pvd_pvds[i]->raw_scale;
		pvd->tactmap_dfn.designations[j].text.vertical_offset =
		  offsets[j].v_base +
		    offsets[j].v_mul * pvd_pvds[i]->raw_scale;
		
		/* Now check the viewable list to see if the markings
		 * should be displayed.
		 */
                if (do_display)
		  tactmap_update_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].designations[j],
					&pvd->tactmap_dfn.designations[j]);
	    }
	}
	else if (is_stealth)	/* Stealth */
	{
	    VehicleID vid;
	    VehicleMarking stealth_marking;
	    XColor ign, clr;
	    
	    pict->sensitive.class = &pvd_pvds[i]->stealth_class;
	    pict->magnification = 0.0; /*Pixels*/
#ifndef MODSTEALTH
	    if (!pvd_modstealth)
	      force_id = stealth_get_color(vehicle_id);
	    else
#endif
	      force_id = ent_get_force_id(vehicle_id);
	    
	    pict->num_gcs = 3;
	    pict->gcs[0] = pvd_pvds[i]->black_gc;
	    pict->gcs[1] = pvd_pvds[i]->white_gc;
	    pict->gcs[2] = pvd_pvds[i]->force_gc[force_id];
#ifndef MODSTEALTH
	    if (pvd_modstealth) 
	    {
		ent_get_marking(vehicle_id, &stealth_marking);
		if (XAllocNamedColor(XtDisplay(pvd_pvds[i]->map_widget),
				     DefaultColormapOfScreen
				     (XtScreen (pvd_pvds[i]->map_widget)),
				     (char *)stealth_marking.text, &clr, &ign))
		  XSetForeground(XtDisplay(pvd_pvds[i]->map_widget),
				 pict->gcs[2], clr.pixel);
	    }
#endif
	    
	    tactmap_update_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture,
				  &pvd->tactmap_dfn.picture);
	    
	    vtab_vehicle_simid(vehicle_id, &vid);

	    if (pvd_pvds[i]->show_all_stealths ||
		NS_SIM_ADDRESSES_EQUAL(pvd_pvds[i]->preferred_stealth,
				       vid.simulator))
	      tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					   pvd->tactmap_obj[i].picture);
	}
	else if (is_env)
	{
	    int32	gc_count;
	    
	    /* Local env phenomena come and go, so check
	     * the active flag to see if the smoke should even be displayed.
	     * (Note that it costs practically nothing to re-add or re-remove
	     * an object, so we don't bother checking that here.)
	     */
	    if (!envent_active(vehicle_id))
	    {
		if (tmap_obj_class == TACTMAP_PICTURE)
		  
		  tactmap_remove_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].picture);
		else 
		  tactmap_remove_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].global_effect);

		for (j=0;j<PVD_DESIGNATIONS;j++)
		  tactmap_remove_object(pvd_pvds[i]->tactmap,
					pvd->tactmap_obj[i].
					designations[j]);
		continue;
	    }

	    bzero((char*)&null_class, sizeof(null_class));
	    if (tmap_obj_class == TACTMAP_PICTURE)
	    {
		pict->sensitive.class = &null_class;

		pict->num_gcs = 12;
		pict->gcs[0] = pvd_pvds[i]->black_gc;
		pict->gcs[1] = pvd_pvds[i]->white_gc;

		for (gc_count = 0; gc_count < 10; gc_count++)
		  pict->gcs[2+gc_count] = pvd_pvds[i]->shading_gc[gc_count];
		tactmap_update_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].picture,
				      &pvd->tactmap_dfn.picture);
	    
		tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					     pvd->tactmap_obj[i].picture);
	    }
	    else
	    {
		/* This is set up for dust only!!!!! $$$ */
		env_tactmap_dfn->global_effect.sensitive.class = &null_class;

		env_tactmap_dfn->global_effect.num_gcs = 11;
		env_tactmap_dfn->global_effect.gcs[0] = pvd_pvds[i]->black_gc;
		env_tactmap_dfn->global_effect.gcs[1] = pvd_pvds[i]->white_gc;

		for (gc_count = 0; gc_count < 9; gc_count++)
		  env_tactmap_dfn->global_effect.gcs[2+gc_count] = 
		    pvd_pvds[i]->shading_gc[gc_count+10];
	    
		tactmap_update_object(pvd_pvds[i]->tactmap,
				      pvd->tactmap_obj[i].global_effect,
				      &pvd->tactmap_dfn.global_effect);
	    
		tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					     pvd->tactmap_obj[i].global_effect);

	    }
	}
	else if (is_mark)
	{
	    static SNSTVE_CLASS null_class;
	    
	    /* markers aren't sensitive, so use a null class */
	    bzero((char *)&null_class, sizeof(null_class));
	    pvd->tactmap_dfn.picture.sensitive.class = &null_class;
	    
            pvd->tactmap_dfn.picture.linear.gc =
              pvd_pvds[i]->black_gc;
	    
	    tactmap_update_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture,
				  &pvd->tactmap_dfn.picture);

	    if (do_display)
	      tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					   pvd->tactmap_obj[i].picture);
	    else
	      tactmap_remove_object(pvd_pvds[i]->tactmap,
				    pvd->tactmap_obj[i].picture);
	    
	}
	else if (is_mf)
	{
	    /* Minefields are drawn.  The intelligence is in the drawing. */
	    
	    /* If we're too zoomed in, the PO stuff will draw it */
	    if (vtab_in_list(pvd_pvds[i]->mf_list, vehicle_id) &&
		(pvd_pvds[i]->mf_sym_cutoff > pvd_pvds[i]->current_scale))
	      do_display = FALSE;


	    tactmap_update_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture,
				  &pvd->tactmap_dfn.picture);

	    if (do_display)
	      tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					   pvd->tactmap_obj[i].picture);
	    else
	      tactmap_remove_object(pvd_pvds[i]->tactmap,
				    pvd->tactmap_obj[i].picture);
	    
	}
		
	/* If the user is currently getting info about this vehicle,
	 * update it.
	 */
	if (pvd_pvds[i]->info_vehicle == vehicle_id)
	{
	    char buf[PVD_BUF_LEN];
	    
	    sgui_pop_help(pvd_pvds[i]->current, TRUE);
	    get_vitals(pvd_pvds[i], vehicle_id, buf, PVD_BUF_LEN);
	    sgui_push_help(pvd_pvds[i]->current, buf);
	}
    }    
}

void pvd_set_highlight(int32 vehicle_id, int32 value)
{
    PVD_VARS *pvd = (PVD_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id),
			  pvd_user_data_handle);
    
    if (!pvd) /* Passive error detection */
      return;
    
    pvd->highlight = value;
}


static int32 pvd_aggregate_displayable (int32 vehicle_id, int32 i)
{
    ObjectType guise;
    uint32     unit_size, display_size;
    int32      parent;

    
    guise = ent_get_guise(vehicle_id, distinguishedForceID);
    unit_size = guise & unitSizeMask;

    if (pvd_pvds[i]->show_type[PVD_SHOW_PLATOONS])
      display_size = unitSizePlatoon;
    else if (pvd_pvds[i]->show_type[PVD_SHOW_COMPANIES])  
      display_size = unitSizeCompany;    
    else if (pvd_pvds[i]->show_type[PVD_SHOW_BATTALIONS])  
      display_size = unitSizeBattalion;    
    else
      return FALSE;

    if (unit_size == display_size)
      return TRUE;

    if (unit_size >  display_size)
      return FALSE;

    /* 
     *  display only if this unit has no parent or its
     *  parent in not to be displayed.
     */

    parent = aggrmap_get_parent (vehicle_id);
    /* parent may not be set yet: try to find it again. */
    if (!parent)
    {
	aggrmap_find_parent(vehicle_id, &parent);
	
	if (!parent)
	  return TRUE;
    }
	
    if (parent == vehicle_id)
      return TRUE;
    guise = ent_get_guise(parent, distinguishedForceID);
    unit_size = guise & unitSizeMask;
    /* If parent is too high to be displayed, display the unit */
    if (unit_size > display_size)
      return TRUE;
    return FALSE;
}
#ifdef USE_CLCGF
static void pvd_process_aggregate(int32 vehicle_id)
{
    PVD_VARS *pvd;
    ObjectType guise;
    VehicleGuises  *guises;
    READER_UNION *new_dfn = NULL, *old_dfn = NULL;
    int32 i, j, id;
    float64 pos[XYZC], rot[XYZ][XYZ], *dir;
    ForceID force_id;
    uint32 appearance;
    PHYSDB_DATA *pdb;
    float64 speed;
    VehicleMarking marking;
    struct tactmap_picture *pict;
    struct tactmap_bgr *icon;
    uint32 m;
    char alt_buf[20], spd_buf[20];
    int32 show_aggregate;
    int32 do_display;
    int32 became_inactive = FALSE;
    static SNSTVE_CLASS null_class;
    int32 is_aggregate;
	
    /* Someone before me may have just deactivated me prior to this tick
     * event being fired.  Check this before continuing.
     */
    if (!vtab_get_vehicle(vehicle_id))
      return;

    if (!VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),VTAB_AGGREGATE))
      return;

    pvd = (PVD_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id),
			  pvd_user_data_handle);
    
    if (!pvd)					/* Passive error detection */
      return;
    if (!ent_active(vehicle_id))
    {
        if (pvd->last_active)
        {
            pvd->last_active = FALSE;
            became_inactive  = TRUE;
        }
        else
        {
            pvd->last_active = FALSE;
            return;
        }
    }
    else
    {
        pvd->last_active = TRUE;
    }
    
    /* Before we continue, make sure somebody wants to be updated */
    for (i=0;i<pvd_n_pvds;i++)
      if (!pvd_pvds[i]->frozen &&
	  (pvd->last_update + pvd_pvds[i]->veh_update_rate <
	   time_last_realtime_clock) || became_inactive)
	break;
    if (i == pvd_n_pvds)
      return;
    
    guise = ent_get_guise(vehicle_id, distinguishedForceID);
        
    if (pvd->guiseless)
    {
	/* See if we have a guise yet */
	if (!guise)
	  return;
	
	pvd->guiseless = 0;
	
	/* Initialize the stuff which doesn't change from tick to tick */
	pvd->tactmap_dfn.picture.sensitive.instance  = vehicle_id;
	pvd->tactmap_dfn.picture.sensitive.user_data = (ADDRESS)vehicle_id;
	
	/* Need some initial stuff */
	pvd->tactmap_dfn.icon.sensitive.instance  = vehicle_id;
	pvd->tactmap_dfn.icon.sensitive.user_data = (ADDRESS)vehicle_id;
	
	pvd->pictures = (READER_UNION *)otm_query(pvd_db_key, guise);
 
        /* set the initial appearance to some unlikely value..  this value
         * will be compared with the real vehicles appearance, and it is
         * imperitive that these values not be eqaul initially..
         */

        pvd->veh_appearance = (uint32) -1;
	
	pvd->tactmap_dfn.designations[PVD_CALL_SIGN].text.see_through = 0;
	pvd->tactmap_dfn.designations[PVD_CALL_SIGN].text.alignment = west;
	pvd->tactmap_dfn.designations[PVD_ALTITUDE].text.see_through = 0;
	pvd->tactmap_dfn.designations[PVD_ALTITUDE].text.alignment= southWest;
	pvd->tactmap_dfn.designations[PVD_SPEED].text.see_through = 0;
	pvd->tactmap_dfn.designations[PVD_SPEED].text.alignment = southEast;
	
	bzero((char*)&null_class, sizeof(null_class));
	
	for (i=0;i<pvd_n_pvds;i++)
	{

	    pvd->tactmap_dfn.picture.sensitive.class =
	      &pvd_pvds[i]->vehicle_class;
	    pvd->tactmap_dfn.icon.sensitive.class =
	      &pvd_pvds[i]->vehicle_class;
	    
	    pvd->tactmap_obj[i].icon =
	      tactmap_create_object(pvd_pvds[i]->tactmap,
				    TACTMAP_BGR,
				    &pvd->tactmap_dfn.icon);
	    
	    pvd->tactmap_obj[i].picture =
	      tactmap_create_object(pvd_pvds[i]->tactmap,
				    TACTMAP_PICTURE,
				    &pvd->tactmap_dfn.picture);
	    
	    for (j=0;j<PVD_DESIGNATIONS;j++)
	      pvd->tactmap_obj[i].designations[j] =
		tactmap_create_object(pvd_pvds[i]->tactmap,
				      TACTMAP_TEXT,
				      pvd->tactmap_dfn.designations+j);
	}
    }
    
    pict = &pvd->tactmap_dfn.picture.picture;
    icon = &pvd->tactmap_dfn.icon.bgr;

    ent_set_context(ent_get_cell(vehicle_id));
    pos[CELL3D] = ent_get_cell(vehicle_id);

    access_get(vehicle_id,
	       ent_rotation, rot,
	       ent_force_id, &force_id,
	       ent_physdb, &pdb,
	       ent_appearance, &appearance,
	       ent_speed, &speed,
	       ent_marking, &marking,
	       ent_position, pos,
	       A_END);	

    dir = rot[Y];
    
    for (j=0;j<PVD_DESIGNATIONS;j++)
    {
	pvd->tactmap_dfn.designations[j].text.x = pos[X];
	pvd->tactmap_dfn.designations[j].text.y = pos[Y];
	pvd->tactmap_dfn.designations[j].text.cell = pos[CELL3D];
    }

    pvd->tactmap_dfn.designations[PVD_CALL_SIGN].text.string =
      (char *)marking.text;

    /* We fill in the alt_buf and spd_buf on a per-gui basis, below */
    pvd->tactmap_dfn.designations[PVD_ALTITUDE].text.string = alt_buf;
    pvd->tactmap_dfn.designations[PVD_SPEED].text.string = spd_buf;
	
	
    /* Setting up icon data */
    icon->x            = pos[X];
    icon->y            = pos[Y];
    icon->cell         = pos[CELL3D];
    icon->dashed       = FALSE;
    icon->call_sign[0] = '\0';
    icon->object_type  = guise;
    icon->override     = FALSE;
    icon->dfn          = NULL;
    icon->rotation     = -NS_ATAN2(dir[X], dir[Y]);
	
    /* Snap the rotation to be multiples of PI/4 */
    icon->rotation = 
      floor((icon->rotation + (PI/8.0))/(PI/4.0)) * (PI/4.0);

    pvd->tactmap_dfn.picture.picture.highlight = pvd->highlight;
    pvd->tactmap_dfn.icon.bgr.highlight        = pvd->highlight;
    
    for (i=0;i<pvd_n_pvds;i++)
    {
	if (!pvd_pvds[i]->frozen &&
	    (pvd->last_update + pvd_pvds[i]->veh_update_rate >
	     time_last_realtime_clock) && !became_inactive)
	  continue;
    

	pvd->last_update = time_last_realtime_clock;
	
        /* Get whether it is okay to display vehicle annotations or not. */
        do_display = 
	  pvd_okay_to_display(pvd_pvds[i], force_id, vehicle_id, FALSE);
	


	/* Local vehicles come and go (especially missiles), so check
	 * the active flag to see if the vehicle should even be displayed.
	 * (Note that it costs practically nothing to re-add or re-remove
	 * an object, so we don't bother checking that here.)
	 */
	if (!ent_active(vehicle_id))
	{
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture);
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].icon);
	    for (j=0;j<PVD_DESIGNATIONS;j++)
	      tactmap_remove_object(pvd_pvds[i]->tactmap,
				    pvd->tactmap_obj[i].
				    designations[j]);
	    continue;
	}
	    
	/* Dispatch on whether this PVD wants vehicle pictures,
	 * vehicle icons, or neither.
	 */

        if ((ent_get_aggregate_state(vehicle_id) == aggregated)
	    || pvd_aggregate_displayable(vehicle_id, i))
	  show_aggregate = TRUE;
	else
	{
	    /* I'm not drawing these now, so remove them.
	     * Note that it doesn't hurt to remove something not
	     * currently on the display list.
	     */
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture);
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].icon);
	    for (j=0;j<PVD_DESIGNATIONS;j++)
	      tactmap_remove_object(pvd_pvds[i]->tactmap,
				    pvd->tactmap_obj[i].
				    designations[j]);
	    continue;
	}
	    


	icon->sensitive.class = &pvd_pvds[i]->vehicle_class;
	icon->size = pvd_pvds[i]->icon_size;
	icon->line_thickness = 3;
	icon->gc = pvd_pvds[i]->force_gc[force_id];
	 
		
	tactmap_update_object(pvd_pvds[i]->tactmap,
			      pvd->tactmap_obj[i].icon,
			      &pvd->tactmap_dfn.icon);
	/* Now check the viewable list to see if the icons
	 * should be displayed.
	 */
	if (do_display)
	{
	    tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					 pvd->tactmap_obj[i].icon);
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture);
		    
		    
	}	
	else
	{
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].picture);
		    
	    tactmap_remove_object(pvd_pvds[i]->tactmap,
				  pvd->tactmap_obj[i].icon);
	}
		

	    
	    
	/* Suppress all markings on munitions, suppress ATC markings
	 * on non-air vehicles.
	 */
	m = pvd_pvds[i]->markings;
	
	if ((guise & (objectDomainMask|vehicleEnvironmentMask)) !=
	    (objectDomainVehicle | vehicleEnvironmentAir))
	  m &= ~(1 << PVD_ALTITUDE);
	    
	/* Altitudes are shown in hundreds of feet, unless the
	 * defaults select otherwise
	 */
	if (pvd_pvds[i]->altitudes_agl)
	{
	    edt_format(pvd_pvds[i]->gui, alt_buf, "#h",
		       ent_get_altitude_agl(vehicle_id));
	    /* Remove units display (ft, meters, etc.) */
	    strtok(alt_buf, " ");
	}
	else
	  sprintf(alt_buf, "%d", (int)(pos[Z] / 30.48 + 0.5 /*ROUND*/));
	
	edt_format(pvd_pvds[i]->gui, spd_buf, "#s", speed);
	strtok(spd_buf, " ");
	    
	for (j=0;j<PVD_DESIGNATIONS;j++)
	{
	    /* Remove if not okay to display.
	     * Note that it doesn't hurt to remove something not
	     * currently on the display list.
	     */
	    if (do_display && (m & (1 << j)))
	      tactmap_add_transient_object(pvd_pvds[i]->tactmap,
					   pvd->tactmap_obj[i].
					   designations[j]);
	    else	
	      tactmap_remove_object(pvd_pvds[i]->tactmap,
				    pvd->tactmap_obj[i].
				    designations[j]);
		
	    pvd->tactmap_dfn.designations[j].text.gc = 
	      pvd_pvds[i]->black_gc;
	    pvd->tactmap_dfn.designations[j].text.font_struct =
	      pvd_pvds[i]->font_struct;
	    pvd->tactmap_dfn.designations[j].text.horizontal_offset =
	      offsets[j].h_base + 
		offsets[j].h_mul * pvd_pvds[i]->raw_scale;
	    pvd->tactmap_dfn.designations[j].text.vertical_offset =
	      offsets[j].v_base +
		offsets[j].v_mul * pvd_pvds[i]->raw_scale;
	    
	    /* Now check the viewable list to see if the markings
	     * should be displayed.
	     */
	    if (do_display)
	      tactmap_update_object(pvd_pvds[i]->tactmap,
				    pvd->tactmap_obj[i].designations[j],
				    &pvd->tactmap_dfn.designations[j]);
	}


	
	/* If the user is currently getting info about this vehicle,
	 * update it.
	 */
	if (pvd_pvds[i]->info_vehicle == vehicle_id)
	{
	    char buf[PVD_BUF_LEN];
	    
	    sgui_pop_help(pvd_pvds[i]->current, TRUE);
	    get_vitals(pvd_pvds[i], vehicle_id, buf, PVD_BUF_LEN);
	    sgui_push_help(pvd_pvds[i]->current, buf);
	}
    }    
}
#endif /* USE_CLCGF */

/* Draw a minefield */
static void pvd_draw_minefield(Widget widget, Pixmap drawable,
			       ADDRESS draw_arg, XRectangle *boundary)
{
    int32 vehicle_id = (int32)draw_arg;
    PVD_GUI_PTR pg = NULL;
    GC gc;
    ForceID force_id;
    int16 app;
    float64 pos[3];
    float64 rot[3][3];
    float64 extnt[2];
    int32 cell;
    int32 i;
    XPoint points[5];
    XSegment seg[5];
    float64 corners[4][XYZ], circ[XYZ], ocirc[XYZ];
    float64 ear11[XY], ear12[XY], ear21[XY], ear22[XY];
    float64 offset[XYC], factor, earoff[XYC], extra, lane[XY];
    int32 sx,sy,cx,cy;
    float64 size;
    int32 num_circ;

    /* Process for only this particular pvd */
    for (i=0;i<pvd_n_pvds;i++)
    {
	if (pvd_pvds[i]->map_widget == widget)
	{
	    pg = pvd_pvds[i];
	    break;
	}
    }
    if (!pg) return;

    /* Which side is it? */
    force_id = minemgr_get_force_id(vehicle_id);
    
    if (force_id == distinguishedForceID)
      gc = pg->mf_gc;
    else if (force_id == otherForceID)
      gc = pg->opmf_gc;
    else
      gc = pg->black_gc;

    /* What kind is it? */
    app = minemgr_get_appearance(vehicle_id);

    /* Should it be da-shed? */
    if (app & minefieldInactive)
      XSetLineAttributes(XtDisplay(widget), gc, 2, LineOnOffDash,
			 CapButt, JoinMiter);
    else
      XSetLineAttributes(XtDisplay(widget), gc, 2, LineSolid,
			 CapButt, JoinMiter);

    /* Where does it go? */
    minemgr_get_position(vehicle_id, pos);
    minemgr_get_rotation(vehicle_id, rot);
    cell = minemgr_get_cell(vehicle_id);
    minemgr_get_extents(vehicle_id, extnt);

    /* Are these extents bigger than the minimum size--51x23? */
    tactmap_pix_to_offvec(pg->tactmap, 51, -23, offset);
    if ((offset[X] > extnt[X]) && (extnt[X]))
      extnt[X] *= (offset[X]/extnt[X]);
    if ((offset[Y] > extnt[Y]) && (extnt[Y]))
      extnt[Y] *= (offset[Y]/extnt[Y]);
    

    /* Figure out where the 4 corners are on the screen */
    corners[0][X] = corners[1][X] = extnt[X] * 0.5;
    corners[2][X] = corners[3][X] = -extnt[X] * 0.5;
    corners[0][Y] = corners[3][Y] = extnt[Y] * 0.5;
    corners[1][Y] = corners[2][Y] = -extnt[Y] * 0.5;
    corners[0][Z] = corners[1][Z] = corners[2][Z] = corners[3][Z] = 0.0;

    for(i=0;i<4;i++)
    {
	vmat3_vec_mat_mul64(corners[i], rot, corners[i]);
	vmat3_vec_add64(corners[i], pos, corners[i]);
	(void)tactmap_map_to_screen(pg->tactmap, corners[i][X], corners[i][Y],
				    cell, &sx, &sy);
	points[i].x = sx;
	points[i].y = sy;
    }
    points[4].x = points[0].x;
    points[4].y = points[0].y;

    /* Set up clearing rectangle */
    boundary->x = points[0].x;
    boundary->y = points[0].y;
    boundary->width = points[0].x;
    boundary->height = points[0].y;

    for(i=1;i<4;i++)
    {
	if (boundary->x > points[i].x)
	  boundary->x = points[i].x;
	if (boundary->y > points[i].y)
	  boundary->y = points[i].y;
	if (boundary->width < points[i].x)
	  boundary->width = points[i].x;
	if (boundary->height < points[i].y)
	  boundary->height = points[i].y;
    }
    boundary->width -= boundary->x;
    boundary->height -= boundary->y;

    boundary->x--;
    boundary->y--;
    boundary->width+=2;
    boundary->height+=2;

    /* Draw outline */
    XDrawLines(XtDisplay(widget), XtWindow(widget), gc, points, 5, 
	       CoordModeOrigin);

    /* Figure out how many and what size circles to draw */
    
    /* Size is the biggest circle that will fit in world coord */
    /* 9 pixels high when it's min sized */
    size = NS_MAX(NS_MIN(extnt[Y] * 9.0 / 23.0, extnt[X] * 9.0 / 51.0),1);  

    /* Figure out how many circles from this size */
    num_circ = extnt[X] / (size * 14.0/9.0); /* 5 extra pixels between circs */

    /* Figure out spacing between circles */
    extra = (extnt[X] - (size * num_circ)) / (num_circ + 1);
    
    /* Figure out line width for circles and "ears" */
    earoff[X] = size/9.0;
    earoff[Y] = 0;
    earoff[CELL2D] = 0;
    tactmap_offvec_to_pix(pg->tactmap, earoff, &sx, &sy);
    
    XSetLineAttributes(XtDisplay(widget), gc, sx,
		       LineSolid, CapButt, JoinMiter);
    XSetLineAttributes(XtDisplay(widget), pg->lane_gc, sx,
		       LineSolid, CapButt, JoinMiter);
    
    /* Loop through the circles */
    for(i=0;i<num_circ;i++)
    {
	/* Map the location to the screen */
	ocirc[X] = (-extnt[X] * 0.5) + (size + extra) * i + 
	  (size * 0.5 + extra);
	ocirc[Y] = ocirc[Z] = 0;
	vmat3_vec_mat_mul64(ocirc, rot, circ);
	vmat3_vec_add64(circ, pos, circ);
	(void)tactmap_map_to_screen(pg->tactmap, circ[X] - (size * 0.5),
				    circ[Y] + (size * 0.5), cell, &cx, &cy);
	(void)tactmap_map_to_screen(pg->tactmap, circ[X] + (size * 0.5),
				    circ[Y] - (size * 0.5), cell, &sx, &sy);
	
	/* Draw the circle */
	if (app & minefieldInactive)
	  XDrawArc(XtDisplay(widget), XtWindow(widget), gc, cx, cy, 
		   sx-cx, sy-cy, 0, 360*64);
	else
	  XFillArc(XtDisplay(widget), XtWindow(widget), gc, cx, cy, 
		   sx-cx, sy-cy, 0, 360*64);

	/* Draw ears if it's an AP.
	 * Note: this huge gloppy mess is because a mixed AT/AP is supposed to
	 * have ears on every other circle, and because a mixed minefield
	 * might have both appearance bits on or both off.
	 */
	if (((((app & minefieldAP) && (app & minefieldAT)) || 
	      (!(app & minefieldAP) && !(app & minefieldAT))) && !(i%2)) ||
	    (app & minefieldAP))
	{
	    ear11[X] = ocirc[X] + size * 0.225;
	    ear21[X] = ocirc[X] - size * 0.225;
	    ear11[Y] = ear21[Y] = ocirc[Y] + size * 0.390;
	    ear12[X] = ocirc[X] + size * 0.5;
	    ear22[X] = ocirc[X] - size * 0.5;
	    ear12[Y] = ear22[Y] = ocirc[Y] + size *  0.866;

	    vmat3_vec_mat_mul64(ear11, rot, ear11);
	    vmat3_vec_add64(ear11, pos, ear11);
	    (void)tactmap_map_to_screen(pg->tactmap, ear11[X], ear11[Y],
					cell, &sx, &sy);
	    seg[0].x1 = sx;
	    seg[0].y1 = sy;

	    vmat3_vec_mat_mul64(ear12, rot, ear12);
	    vmat3_vec_add64(ear12, pos, ear12);
	    (void)tactmap_map_to_screen(pg->tactmap, ear12[X], ear12[Y],
					cell, &sx, &sy);
	    seg[0].x2 = sx;
	    seg[0].y2 = sy;

	    vmat3_vec_mat_mul64(ear21, rot, ear21);
	    vmat3_vec_add64(ear21, pos, ear21);
	    (void)tactmap_map_to_screen(pg->tactmap, ear21[X], ear21[Y],
					cell, &sx, &sy);
	    seg[1].x1 = sx;
	    seg[1].y1 = sy;

	    vmat3_vec_mat_mul64(ear22, rot, ear22);
	    vmat3_vec_add64(ear22, pos, ear22);
	    (void)tactmap_map_to_screen(pg->tactmap, ear22[X], ear22[Y],
					cell, &sx, &sy);
	    seg[1].x2 = sx;
	    seg[1].y2 = sy;
	    
	    XDrawSegments(XtDisplay(widget), XtWindow(widget), gc, seg, 2);
	}
    }
    
    /* Draw a "lane" symbol through the icon if necessary */
    if (app & minefieldLane)
    {
	lane[X] = 0;
	lane[Y] = -extnt[Y] * 0.5;
	vmat3_vec_mat_mul64(lane, rot, lane);
	vmat3_vec_add64(lane, pos, lane);
	(void)tactmap_map_to_screen(pg->tactmap, lane[X], lane[Y],
				    cell, &sx, &sy);
	seg[0].x2 = sx;
	seg[0].y2 = sy;
	seg[1].x2 = sx;
	seg[1].y2 = sy;
	seg[2].x1 = sx;
	seg[2].y1 = sy;

	lane[X] = -extnt[Y] * 0.2;
	lane[Y] = -extnt[Y] * 0.7;
	vmat3_vec_mat_mul64(lane, rot, lane);
	vmat3_vec_add64(lane, pos, lane);
	(void)tactmap_map_to_screen(pg->tactmap, lane[X], lane[Y],
				    cell, &sx, &sy);
	seg[0].x1 = sx;
	seg[0].y1 = sy;

	lane[X] = extnt[Y] * 0.2;
	lane[Y] = -extnt[Y] * 0.7;
	vmat3_vec_mat_mul64(lane, rot, lane);
	vmat3_vec_add64(lane, pos, lane);
	(void)tactmap_map_to_screen(pg->tactmap, lane[X], lane[Y],
				    cell, &sx, &sy);
	seg[1].x1 = sx;
	seg[1].y1 = sy;

	lane[X] = 0;
	lane[Y] = extnt[Y] * 0.5;
	vmat3_vec_mat_mul64(lane, rot, lane);
	vmat3_vec_add64(lane, pos, lane);
	(void)tactmap_map_to_screen(pg->tactmap, lane[X], lane[Y],
				    cell, &sx, &sy);
	seg[2].x2 = sx;
	seg[2].y2 = sy;
	seg[3].x1 = sx;
	seg[3].y1 = sy;
	seg[4].x1 = sx;
	seg[4].y1 = sy;

	lane[X] = -extnt[Y] * 0.2;
	lane[Y] = extnt[Y] * 0.7;
	vmat3_vec_mat_mul64(lane, rot, lane);
	vmat3_vec_add64(lane, pos, lane);
	(void)tactmap_map_to_screen(pg->tactmap, lane[X], lane[Y],
				    cell, &sx, &sy);
	seg[3].x2 = sx;
	seg[3].y2 = sy;

	lane[X] = extnt[Y] * 0.2;
	lane[Y] = extnt[Y] * 0.7;
	vmat3_vec_mat_mul64(lane, rot, lane);
	vmat3_vec_add64(lane, pos, lane);
	(void)tactmap_map_to_screen(pg->tactmap, lane[X], lane[Y],
				    cell, &sx, &sy);
	seg[4].x2 = sx;
	seg[4].y2 = sy;

	XDrawSegments(XtDisplay(widget), XtWindow(widget), pg->lane_gc, seg,
		      5);
    }


}

#endif  /*USE_MOTIF*/
