#ifndef lint
static char rcsid [] = "$RCSfile: pvd_env.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_env.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle pvd environment mode                       *
*       Created: Mon Apr 11 1994                                            *
*       Author: llovero                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <stdlib.h>
#include <time.h>
#include <libenvironment.h>
#include <math.h>
#include <libvecmat.h>
#include <libgcs.h>
#include <stdext.h> /*common/include/global*/
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <X11/StringDefs.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include "libenvillum.h"
#include "libgcs.h"

static void env_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action, 
			SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui);
static void show_env_display(
    Widget               w,
    PVD_GUI_PTR          pvd_gui,
    XmAnyCallbackStruct *call_data);
static void draw_skyview(
    Widget           da_widget, /* Drawing area */
    PVD_ENV_DISPLAY *env_display);
static void draw_illum_levels(
    Widget           widget,
    PVD_ENV_DISPLAY *env_display);
static void query_environment(PVD_ENV_DISPLAY *env_display);
static int16 get_sun_or_moon_coordinates(
    float64 altitude,
    float64 azimuth,
    uint32  radius,
    int32  *x,
    int32  *y);

static float64 sv_alt = PI/2.0;	/* view center altitude, rads */
static float64 sv_az = PI;	/* view center az, rads */
static float64 sv_fov = PI;	/* field of view, rads */

#define SUN_RADIUS 10
#define HALF_SUN_RADIUS SUN_RADIUS/2
#define MOON_RADIUS 8
#define HALF_MOON_RADIUS MOON_RADIUS/2

#define MOON_OFFSET    20
#define BELOW_HORIZON_OFFSET 20

/*ARGSUSED*/
void 
pvd_env_mode(SGUI_PTR gui, SGUI_MODE_PTR mode, PVD_GUI_PTR pvd_gui,
	     SGUI_MODE_STATE state)
{
    if ((state == SGUI_ACTIVE) ||
	(state == SGUI_RESUMED))
    {
	/* Install the mouse handler */
	callback_register_handler(pvd_gui->map_event,env_gesture,pvd_gui);
	pvd_gui->current = pvd_gui->env;
    }
    else
    {
	/* Remove the mouse handler */
	callback_unregister_handler(pvd_gui->map_event,env_gesture,pvd_gui);
    }
}

/*ARGSUSED*/
static void 
env_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
	    SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui)
{
    float64 map[XYZC];
    int32   tmp_c;
    float64 wind_speed, temperature, humidity, illumination;
    float64 visibility;
    float64 cloud_percent;
    float64 wind_vector[2];
    float64 current_vector[2];
    float64 temp_farht;
    ENV_INTERFACE ei;
    char buf[2048];
    float32 tide_level;
    

    bzero(&ei, sizeof(ENV_INTERFACE));

    /* Ignore button 1, and EXCITED/UNEXCITED callbacks */
    if (call_data->button < 2)
      return;

    switch(action)
    {
      case SNSTVE_MOTION:
#if 0
	/* I don't know why this has been commented out! -rg */
	sgui_pop_help(pvd_gui->env, TRUE);
#endif
	/* Fall through */
      case SNSTVE_PRESS:
	tactmap_screen_to_map(pvd_gui->tactmap,
			      call_data->screen_x, call_data->screen_y, 
			      &map[X], &map[Y], &tmp_c);
	map[CELL3D] = tmp_c;
	map[Z]      = ctdb_lookup_elevation((CTDB *)gcs_get_tdb(tmp_c,FALSE), 
					    map[X], map[Y]);
	VMAT4_VEC_COPY64(map,ei.u.generic_vector_to_value.input_vector);

	/* Get temperature (in celsius) and convert to farenheit: */
	ei.u.temperature.input_vector[X] = map[X];
	ei.u.temperature.input_vector[Y] = map[Y];
	ei.u.temperature.input_vector[Z] = map[Z];
	env_get(ENV_TEMPERATURE, &ei);
	temperature = ei.u.temperature.output_value;
	temp_farht = (9.0 * temperature) / 5.0 + 32.0;

	/* Get visibility (extinction amount in visual band): */
	ei.u.extinction_amount.input_vector[X] = map[X];
	ei.u.extinction_amount.input_vector[Y] = map[Y];
	ei.u.extinction_amount.input_vector[Z] = map[Z];
	env_get(ENV_EXTINCTION_AMOUNT, &ei);
	/* env unit is km. we need meters: */
	visibility = ei.u.extinction_amount.output_value * 1000.0;

	/* Get wind vector (in m/s): */
	ei.u.wind_velocity.input_vector[X] = map[X];
	ei.u.wind_velocity.input_vector[Y] = map[Y];
	ei.u.wind_velocity.input_vector[Z] = map[Z];
	env_get(ENV_WIND_VELOCITY, &ei);
	wind_vector[0] = ei.u.wind_velocity.output_vector[0];
	wind_vector[1] = ei.u.wind_velocity.output_vector[1];

	/* Get cloudiness and convert to a percentage: */
	ei.u.cloud_cover.input_vector[X] = map[X];
	ei.u.cloud_cover.input_vector[Y] = map[Y];
	ei.u.cloud_cover.input_vector[Z] = map[Z];
	env_get(ENV_CLOUD_COVER, &ei);
	cloud_percent = ei.u.temperature.output_value * 100.0;

	/* Get illumination level, totalling direct & diffuse components: */
	ei.u.illumination.vector_info = ENV_BAND_47MM;
	ei.u.illumination.input_vector[X] = map[X];
	ei.u.illumination.input_vector[Y] = map[Y];
	ei.u.illumination.input_vector[Z] = map[Z];
	env_get(ENV_ILLUMINATION, &ei);
	illumination = ei.u.illumination.output_vector[0] +
	  ei.u.illumination.output_vector[1] +
	    ei.u.illumination.output_vector[2];
    
	ei.u.sea_current_velocity.input_vector[X] = map[X];
	ei.u.sea_current_velocity.input_vector[Y] = map[Y];
	ei.u.sea_current_velocity.input_vector[Z] = map[Z];

	
	if(env_get(ENV_SEA_CURRENT_VELOCITY, &ei) != 1)
	  current_vector[0] = ei.u.sea_current_velocity.output_vector[0];
	  current_vector[1] = ei.u.sea_current_velocity.output_vector[1];

	ei.u.sea_tide_level.input_vector[X] = map[X];
	ei.u.sea_tide_level.input_vector[Y] = map[Y];
	ei.u.sea_tide_level.input_vector[Z] = map[Z];

	
	if(env_get(ENV_SEA_TIDE_LEVEL, &ei) != 1)
	  tide_level = ei.u.sea_tide_level.output_value;
		
	edt_format(pvd_gui->gui, buf, 
		   "Coordinates: #l  |  Temp: #D C (#D F)  |  Wind: <#s #s>  | Current: <#s #s> | Tide Level: #d |% Clouds: #D  |  Illum: #D fc | Visibility: #d", 
		   (int32)map[CELL3D], map[X], map[Y],
		   (int32)temperature, 
		   (int32)temp_farht,
		   wind_vector[0],
		   wind_vector[1],
		   current_vector[0],
		   current_vector[1],
		   tide_level,
		   (int32)cloud_percent,
		   (int32)illumination,
		   visibility);

	sgui_push_help(pvd_gui->env, buf);
	
	/* Send a message to the window to redraw the sun/moon, etc. */
	pvd_gui->env_display.view_location[X] = map[X];
	pvd_gui->env_display.view_location[Y] = map[Y];
	pvd_gui->env_display.view_location[Z] = map[Z];
	query_environment(&pvd_gui->env_display);

	break;
      case SNSTVE_RELEASE:
	sgui_pop_help(pvd_gui->env, TRUE);
	
	break;
    }
}


static void done_callback (
    Widget               w,
    PVD_GUI_PTR          pvd_gui,
    XmAnyCallbackStruct *call_data)
{

    /* Cancel update function since display is coming down */
    sched_cancel_fncl(pvd_gui->env_display.update_function);

    XtUnmanageChild(pvd_gui->env_display.env_display_form);
}


/* Calculate the x,y offset of where to draw the sun or moon, */
/* relative to the center of the skyview display. */
/* Return TRUE if should draw, else FALSE. */
static int16 get_sun_or_moon_coordinates(
    float64 altitude,
    float64 azimuth,
    uint32  radius,            /* radius of skyview in pixels */
    int32  *x,
    int32  *y)
{

    float64 range;
    float64 new_y, new_x;

    /* The altitudes are represented by concentric circles */
    /* in the skyview display. The radius of the circle */
    /* represents 90 degrees */
    /* First convert the altitude into a range from the center. */
    /* 0 degrees is sunrise and 90 degrees is straight overhead. */

    if ((altitude < 0) || (altitude > PI))  /* in radians */
    {
	/* sun or moon is below horizon. Check if within  */
	/* twilight ranges (20 degrees). */
	if ((altitude >= - 0.34906585) || (altitude <= -2.792526803))
	{
	    /* Altitude is between 0 and 20 degrees or */ 
	    /* between -160 and -180 degrees. */
	    if (altitude <= -2.792526803)
	      altitude = -PI - altitude;

	    range = (altitude * radius)/PI_OVER_TWO;

	    /* The altitude starts at the horizon.  Instead give offset  */
	    /* from center of circle */
	    range = labs(radius - range);

	}
	else
	  return(FALSE); /* don't draw */
        
    }
    else 
    {
	/* Sun or moon is above the horizon */
	/* Solve for range: PI_OVER_TWO/radius = altitude/range */

	range = (altitude * radius)/PI_OVER_TWO;

	/* The altitude starts at the horizon.  Instead give offset from */
        /* center of circle */
	range = labs(radius - range);

    }

    /* Since azimuth starts at North and goes clockwise, convert */
    /* to appropriate angle */
    azimuth = -(-azimuth - PI_OVER_TWO);

    new_y = sin(azimuth) * range;
    new_x = cos(azimuth) * range;

    *x = (int32) new_x;
    *y = (int32) new_y;

    return(TRUE);
}


/* compute x/y loc of a point at azra/altdec on a circle of radius rad pixels
 *   as viewed from sv_azra/sv_altdec with sv_fov.
 * return 1 if fits on screen, else 0 (but still return x/y).

 * LL: Taken from xephem/skyviewmenu.c
 */
static
int32 sv_loc (
    int32   rad,	/* radius of target display, pixels */
    float64 altdec,	/* angle up from spherical equator, such as alt or dec,
			   rad  */
    float64 azra,	/* angle around spherical pole, such as az or ra,
			   rad  */
    int32  *xp,		/* return X coords within circle */
    int32  *yp)
{
#define	LOCEPS	(1e-5)	/* an angle too small to see on screen, rads */
	static float64 last_sv_altdec = 123.0, last_sa, last_ca;
	float64 a,sa,ca;	/* angle from viewpoint to pole */
	float64 b,sb,cb;	/* angle from object to pole */
	float64 c,sc,cc;	/* difference in polar angles of obj and viewpoint */
	float64 d,sd,cd;	/* angular separation of object and viewpoint */
	float64 r;	/* proportion of d to desired field of view */
	float64 se, ce;	/* angle between (viewpoint,pole) and (viewpoint,obj) */

	a = PI/2 - sv_alt;
	if (sv_alt == last_sv_altdec) {
	    sa = last_sa;
	    ca = last_ca;
	} else {
	    last_sv_altdec = sv_alt;
	    last_sa = sa = sin(a);
	    last_ca = ca = cos(a);
	}

	b = PI/2 - altdec;
	sb = sin(b);
	cb = cos(b);

	/* We're always in altitude/azimuth mode */
	c = azra - sv_az;
	cc = cos(c);
	d = acos(ca*cb + sa*sb*cc);
	if (d < LOCEPS) {
	    *xp = *yp = rad;
	    return (1);
	}

	r = d/(sv_fov/2.0);

	sc = sin(c);
	sd = sin(d);
	se = sc*sb/sd;
	*xp = rad*(1 + r*se) + 0.5;
	if (a < LOCEPS) {
	    /* as viewpoint approaches N pole, e approaches PI - c */
	    ce = -cc;
	} else if (a > PI - LOCEPS) {
	    /* as viewpoint approaches S pole, e approaches c */
	    ce = cc;
	} else {
	    cd = cos(d);
	    ce = (cb - cd*ca)/(sd*sa);
	}
	*yp = rad*(1 - r*ce) + 0.5;

	return (r >= 1.0 ? 0 : 1);
}



/* draw the ecliptic on a circle of radius r, x and y borders xb and yb.
   Taken from XEPHEM/skyviewmenu.c
 */
static void
draw_ecliptic(
    Display *dsp,
    Window   win,
    GC       gc,
    uint32   r,
    uint32   xb,
    uint32   yb,
    float64 *view_location)
{
#define	ECL_TICS	2	/* period of points for ecliptic, pixels */
#define	ECL_CACHE_SZ	100
	XPoint point_cache[ECL_CACHE_SZ];
	float64 elat0, elng0;	/* ecliptic lat and long at center of fov */
	float64 elngmin, elngmax;/* ecliptic long limits */
	float64 ra, dec;
	float64 elng;
	float64 lst;
	int32 ncache;
	ENVILM_EPHEM np;
	float64 ha0;		/* local hour angle */

/*	np = mm_get_now(); */

	/* Fill in all the necessary info for np */
	envilm_geteph(&np, view_location);
	
	envilm_now_lst (&np, &lst);

	/* find ecliptic coords of center of view */
	envilm_aa_hadec (np.n_lat, sv_alt, sv_az, &ha0, &dec);

	ra = ENVILM_HRRAD(lst) - ha0;

	envilm_eq_ecl (np.n_mjd, ra, dec, &elat0, &elng0);

	/* no ecliptic visible if ecliptic latitude at center of view 
	 * is not less than fov/2
	 */
	if (fabs((float64)elat0) >= sv_fov/2.0)
	    return;

	/* worst-case elong limits is center elong += half fov */
	elngmin = elng0 - sv_fov/2.0;
	elngmax = elng0 + sv_fov/2.0;

	/* put a mark at every ECL_TICS pixels */
	ncache = 0;
	for (elng = elngmin; elng <= elngmax; elng += sv_fov/(2.0*r/ECL_TICS)) {
	    int32 x, y;
	    float64 altdec, azra;

	    /* convert longitude along the ecliptic to ra/dec */
	    envilm_ecleq (np.n_mjd, 0.0, elng, &azra, &altdec);

	    /* if in aa mode, we need it in alt/az */
	    /* LL: We're always in altitude/azimuth mode */

	    envilm_eph_hadaa (np.n_lat, ENVILM_HRRAD(lst) - azra, 
			      altdec, &altdec, &azra);
	    envilm_refract (np.n_pressure, np.n_temp, altdec, &altdec);


	    /* if visible, display point */
	    if ((altdec >= 0) && sv_loc (r, altdec, azra, &x, &y)) {
		XPoint *xp = &point_cache[ncache++];
		xp->x = x + xb;
		xp->y = y + yb;
		if (ncache == XtNumber(point_cache)) {
		    XDrawPoints (dsp,win,gc,point_cache,ncache,CoordModeOrigin);
		    ncache = 0;
		}
	    }
	}

	if (ncache > 0)
	    XDrawPoints (dsp, win, gc, point_cache, ncache, CoordModeOrigin);
}


/* return the current circle info about the skyview */
/* drawing area. Taken from XEPHEM */
static void get_drawing_area(
    Widget  drawing_area,
    uint32 *width,
    uint32 *height,
    uint32 *radius,
    uint32 *xborder,
    uint32 *yborder,
    uint32  offset)
{
    Display *dsp = XtDisplay(drawing_area);
    Window win = XtWindow(drawing_area);
    Window root;
    int32 x, y;
    uint32 w, h;
    uint32 border_width, depth;

    XGetGeometry(dsp, win, &root, (int *)&x, (int *)&y, (unsigned int *)width, 
		 (unsigned int *)height,
		 (unsigned int *)&border_width, (unsigned int *)&depth);

    w = *width/2;
    h = *height/2;
    *radius = w > h ? h : w;    
    *radius = *radius - offset;
    *xborder = w - *radius;
    *yborder = h - *radius;

}


static void draw_skyview(
    Widget           da_widget, /* Drawing area */
    PVD_ENV_DISPLAY *env_display)
{
    Arg args[6];
    int16 n;
    Dimension width;
    Dimension height;
    char title[16];
    uint32 w, h, r, xborder, yborder;
    int32 x,y;
    int32 twilight_offset;
    ENV_INTERFACE  ei;


    XClearWindow(XtDisplay(da_widget), XtWindow(da_widget));
    bzero(&ei, sizeof(ENV_INTERFACE));

    get_drawing_area(da_widget, &w, &h, &r, &xborder, &yborder, 
		     (uint32)BELOW_HORIZON_OFFSET);

    /* Draw compass indicators:     N         */
    /*                            E   W       */
    /*                              S         */
    /* A little different from what we're used to! */

    XDrawString(XtDisplay(da_widget), XtWindow(da_widget),
		env_display->outer_horizon_gc,
		xborder + r -5, yborder - 5,
		"N", 1);

    XDrawString(XtDisplay(da_widget), XtWindow(da_widget),
		env_display->outer_horizon_gc,
		xborder + r -5, yborder + (2*r) + 14,
		"S", 1);

    XDrawString(XtDisplay(da_widget), XtWindow(da_widget),
		env_display->outer_horizon_gc,
		xborder - 16, 
		yborder + r + 3,
		"E", 1);

    XDrawString(XtDisplay(da_widget), XtWindow(da_widget),
		env_display->outer_horizon_gc,
		xborder + (2*r) + 6, 
		yborder + r + 3,
		"W", 1);
    
    /* Twilight offsets are 6 degrees from one another */
    /* The radius of the inner circle represents 90 degrees of */
    /* elevation, the very outer represents 108 degrees */
    twilight_offset = r/18;

    /* rebuild the clean circles. Start by drawing the  */
    /* dotted circle representing astronomical twilight */

    XDrawArc(XtDisplay(da_widget), XtWindow(da_widget),
	     env_display->outer_horizon_gc,
	     xborder, yborder, 2*r, 2*r, 0, 360*64);

    /* Now nautical twilight */
    r -= twilight_offset;

    xborder += twilight_offset;
    yborder += twilight_offset;
   
    XDrawArc(XtDisplay(da_widget), XtWindow(da_widget),
	     env_display->outer_horizon_gc,
	     xborder, yborder, 2*r, 2*r, 0, 360*64);

    /* Now civic twilight */
    r -= twilight_offset;

    xborder += twilight_offset;
    yborder += twilight_offset;

    XDrawArc(XtDisplay(da_widget), XtWindow(da_widget),
	     env_display->outer_horizon_gc,
	     xborder, yborder, 2*r, 2*r, 0, 360*64);

    /* Now the horizon itself */
    r -= twilight_offset;

    xborder += twilight_offset;
    yborder += twilight_offset;

    XFillArc(XtDisplay(da_widget), XtWindow(da_widget),
	     env_display->skybackground_gc,
	     xborder, yborder, 2*r, 2*r, 0, 360*64);

    /* Now the ecliptic */
    draw_ecliptic(XtDisplay(da_widget), XtWindow(da_widget),
		  env_display->moon_gc, r, xborder, yborder,
		  env_display->view_location);


    /* Now let's get the sun position */
    /* 0 degrees elevation is sunrise, and 90 degrees is */
    /* straight overhead.  180 or -180 is sunset */
    /* Azimuth runs with 0 degrees being north, and it increases */
    /* as we go clockwise to 360 degrees */

    ei.u.sun_position.input_vector[0] = env_display->view_location[X];
    ei.u.sun_position.input_vector[1] = env_display->view_location[Y];
    ei.u.sun_position.input_vector[2] = env_display->view_location[Z];

    if (env_get(ENV_SUN_POSITION, &ei) != -1)
    {
	/* If the sum of the altitude and azimuth stays the same, the sun */
	/* hasn't moved */
	env_display->last_sun_pos = ei.u.sun_position.output_vector[0] +
	  ei.u.sun_position.output_vector[1];

	/* altitude comes back in first value, then azimuth */
	if (get_sun_or_moon_coordinates(ei.u.sun_position.output_vector[0],
					ei.u.sun_position.output_vector[1],
					r, &x, &y))
	  /* Calculate offset for Motif */
	  XFillArc(XtDisplay(da_widget), XtWindow(da_widget),
		   env_display->sun_gc,
		   xborder+r+x-HALF_SUN_RADIUS, 
		   yborder+r-y-HALF_SUN_RADIUS, SUN_RADIUS, SUN_RADIUS, 
		   0, 360*64);
	
    }

    /* Time for the moon position */
    ei.u.moon_position.input_vector[0] = env_display->view_location[X];
    ei.u.moon_position.input_vector[1] = env_display->view_location[Y];
    ei.u.moon_position.input_vector[2] = env_display->view_location[Z];

    if (env_get(ENV_MOON_POSITION, &ei) != -1)
    {
	/* If the sum of the altitude and azimuth stays the same, the moon */
	/* hasn't moved */
	env_display->last_moon_pos = ei.u.moon_position.output_vector[0] +
	  ei.u.moon_position.output_vector[1];

	/* Only draw the moon when it is above the horizon */
	/* because that's when it contributes illumination */
	if ((ei.u.moon_position.output_vector[0] >=0) && 
	(ei.u.moon_position.output_vector[0] <= PI))
	{
	    
	    /* altitude comes back in first value, then azimuth */
	    if (get_sun_or_moon_coordinates(
					    ei.u.moon_position.output_vector[0],
					    ei.u.moon_position.output_vector[1],
					    r, &x, &y))
	      
	      /* Calculate offset for Motif */
	      XFillArc(XtDisplay(da_widget), XtWindow(da_widget),
		       env_display->moon_gc,
		       xborder+r+x-HALF_MOON_RADIUS, 
		       yborder+r-y-HALF_MOON_RADIUS, MOON_RADIUS, MOON_RADIUS, 
		       0, 360*64);
	}
    }
}

/* Draw the phase of the moon */
/* Taken from XEPHEM/moonmenu.c */
static void draw_moonphase(
    Widget           da_widget, /* Drawing area */
    PVD_ENV_DISPLAY *env_display)
{
    Arg args[6];
    int16 n;
    Dimension width;
    Dimension height;
    char title[16];
    uint32 w, h, r, xborder, yborder;
    int32 x,y;
    int32 twilight_offset;
    ENV_INTERFACE  ei;
    Display *dsp = XtDisplay(da_widget);
    Window win = XtWindow(da_widget);
    Window root;
    float64 el;
    uint32 md;		/* moon circle diameter */
    uint32 border_width, depth;
    int32 la1, la2, ta1, ta2;	/* limb and terminator start and extent */
    int32 wid;			/* distance from meridian to terminator */
    int32 xb, yb;		/* x and y borders, eg, (w-nx)/2 */
#define MRAD           191
#define smallfm_height 480
#define smallfm_width  480
#define	TOPMAR	(smallfm_height/2 - MRAD + 2) /* top margin */
#define	LEFTMAR	(smallfm_width/2 - MRAD) /* left margin */

    bzero(&ei, sizeof(ENV_INTERFACE));
    if (env_get(ENV_MOON_ELONGATION, &ei) != -1)
	el = ei.u.moon_elongation.output_value;

    XGetGeometry(dsp, win, &root, (int *)&x, (int *)&y, (unsigned int *)&w, 
		 (unsigned int *)&h,
		 (unsigned int *)&border_width, (unsigned int *)&depth);

    XClearWindow(dsp, win);

    /* Make background */
    XFillRectangle(dsp, win,
		   env_display->skybackground_gc,
		   0, 0, w, h);

    /* set moon diameter so it is centered within the bounding rectangle
     * of the window.
     * this code sets it to match the borders of the map view and makes
     * the diameter even so it divides evenly.
     */
    md = ((w > h) ? MRAD*h/(MRAD+TOPMAR) : MRAD*w/(MRAD+LEFTMAR)) & (~1);
    xb = (w - md)/2;
    yb = (h - md)/2;

    la1 = el >= 0.0 ? -90*64 : 90*64;
    la2 = 180*64;
    wid = fabs((float64)(md/2*cos(el)))+0.5;
    ta1 = el >= PI/2 || (el <= 0 && el >= -PI/2) ? 90*64 : -90*64;
    ta2 = 180*64;

    /* dsp, win, gc, x, y, w, h, start_ang, ang_extent */
    /* draw the lit hemisphere to the limb */
    XFillArc (dsp, win, env_display->moon_gc, xb, yb, md, md, la1, la2);

    /* draw the portion from the terminator to the meridian */
    if (la1 < 0 && ta1 < 0 || la1 > 0 && ta1 > 0) {
	/* crescent, so draw in background color */
	XFillArc (dsp,win, env_display->skybackground_gc, 
		  md/2-wid+xb, yb, 2*wid, md, ta1, ta2);
    } else {
	/* gibbous, so draw in foreground color */
	XFillArc (dsp,win, env_display->moon_gc, 
		  md/2-wid+xb, yb, 2*wid, md, ta1, ta2);
    }

}

static void handle_moonphase_exposure(
    Widget                       widget,
    PVD_ENV_DISPLAY             *env_display,
    XmDrawingAreaCallbackStruct *cb)
{
    Region region;

    region = XCreateRegion();
    XtAddExposureToRegion(cb->event, region);

    /* Draw moon phase */
    draw_moonphase(widget, env_display);
    XDestroyRegion(region);

}

static void handle_skyview_exposure(
    Widget                       widget,
    PVD_ENV_DISPLAY             *env_display,
    XmDrawingAreaCallbackStruct *cb)
{
    
    Region region;

    region = XCreateRegion();
    XtAddExposureToRegion(cb->event, region);

    /* Draw skyview */
    draw_skyview(widget, env_display);
    XDestroyRegion(region);

}

static uint32 get_time_zone_offset_in_seconds(
    float64 location[])
{
    float64 lat, lon, elev;
    
    if(coord_convert( COORD_GCS, GCS_ILLEGAL_CELL,
		     location[0],location[1],location[2],
		     COORD_LATLON, &lat, &lon, &elev, FALSE))
      return(0);
    else
    {
	if ( lon >= 0.0 )
	  return((int32)((lon+7.5)/15.0) * 3600);
	else
	  return((int32)((lon-7.5)/15.0) * 3600);
    }

    
}


static void draw_date (
    Widget           da_widget, /* Drawing area */
    PVD_ENV_DISPLAY *env_display,
    uint32           cur_time)
{
    Arg args[6];
    char *da_str;
    Display *dsp = XtDisplay(da_widget);
    Window win = XtWindow(da_widget);
    struct tm *tms;

    XClearWindow(dsp, win);

    cur_time += get_time_zone_offset_in_seconds(env_display->view_location);

    tms = gmtime((time_t*)&cur_time);

    da_str = asctime(tms);
    XDrawString(dsp, win,
		env_display->key_gc,
		5,25,
		da_str, strlen(da_str));
}


static uint32 get_env_time(PVD_ENV_DISPLAY *env_display)
{
   ENV_INTERFACE ei;

    bzero(&ei, sizeof(ENV_INTERFACE));

    if (env_get(ENV_SIM_TIME, &ei) != -1)
    {
	return(ei.u.sim_time.uxtime);
    }
   else 
     return (env_display->last_uxtime);

}

static void handle_date_exposure(
    Widget                       widget,
    PVD_ENV_DISPLAY             *env_display,
    XmDrawingAreaCallbackStruct *cb)
{
    
    Region region;

    region = XCreateRegion();
    XtAddExposureToRegion(cb->event, region);

    /* Draw new date */
    env_display->last_uxtime = get_env_time(env_display);
    draw_date(widget, env_display, env_display->last_uxtime);
    XDestroyRegion(region);

}


static struct pvd_dirs 
{
    int32	angle;
    char	*name;
} pvd_dirs[] = 
{
    {0,		"W"},
    {22,	"WSW"},
    {45,	"SW"},
    {67,	"SSW"},
    {90,	"S"},
    {112,	"SSE"},
    {135,	"SE"},
    {157,	"ESE"},
    {180,	"E"},
    {202,	"ENE"},
    {225,	"NE"},
    {247,	"NNE"},
    {270,	"N"},
    {292,	"NNW"},
    {315,	"NW"},
    {337,	"WNW"},
    {360,	"W"}
};

static char *pvd_angle_name(int32 angle)
{
    int32	a;
    int32	d, md = 370, mdi, i;
    
    a = angle % 360;

    for (i = 0; i < 17; i++)
    {
	d = abs(a-pvd_dirs[i].angle);
	if (d <= md)
	{
	    md = d;
	    mdi = i;
	}
    }
    return(pvd_dirs[mdi].name);
}

static char *get_range_string(float64 range)
{

    if (range >= 5.0)
      return("Light");
    else return("Heavy");

}

static char *get_snow_string(uint32 type)
{
    switch (type) {
      case weatherExtinctionBlowingSnow:
	return("Blowing Snow");
      case weatherExtinctionFogSnow:
	return("Fog and Snow");
      case weatherExtinctionSnow:
      default:
	return("Snow");
    }
}


static void draw_weather(
    Widget           da_widget, /* Drawing area */
    PVD_ENV_DISPLAY *env_display)
{
    Arg args[6];
    int16 n;
    Dimension width, height;
    char title[16];
    uint32 w, h, r, xborder, yborder;
    int32 x,y;
    Display *dsp = XtDisplay(da_widget);
    Window win = XtWindow(da_widget);
    Window root;
    char wthr_str[256];
    int32 wind_speed, wind_direction;
    uint32 border_width, depth;
    int32 i = 0;
    Weather_Parms *wthr_parms = &env_display->last_wthr_parms;
#define FONT_HEIGHT 15

    XClearWindow(dsp, win);

    XGetGeometry(dsp, win, &root, (int *)&x, (int *)&y, (unsigned int *)&w, 
		 (unsigned int *)&h,
		 (unsigned int *)&border_width, (unsigned int *)&depth);

    /* Convert to miles per hour */
    wind_speed =
      (int32) rint(sqrt(VMAT3_VEC_MAG_SQ(wthr_parms->wind_velocity)) 
		        / 0.44704);

    wind_direction =
      (int32) RAD_TO_DEG(NS_ATAN2(wthr_parms->wind_velocity[Y],
			          wthr_parms->wind_velocity[X]));
    if (wind_direction < 0)
      wind_direction += 360;
    
    sprintf(wthr_str, "Wind Speed:  %d miles/hour", wind_speed);
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3 , y + (i++*FONT_HEIGHT),
		wthr_str, strlen(wthr_str));

    /* Wind direction is specified as to where wind is */
    /* coming *from*.   */
    sprintf(wthr_str, "Wind From:  %s", pvd_angle_name(wind_direction));
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3 , y+(i++ * FONT_HEIGHT),
		wthr_str, strlen(wthr_str));

    sprintf(wthr_str, "Cloud Cover:  %d%%",
	    (int32) rint(wthr_parms->cloud_cover * 100));
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3 , y+(i++ * FONT_HEIGHT),
		wthr_str, strlen(wthr_str));

    sprintf(wthr_str, "Visibility: %.3f Km\n", 
	    wthr_parms->visual_range/1000);
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3 , y+(i++ * FONT_HEIGHT),
		wthr_str, strlen(wthr_str));

    if (wthr_parms->extinct_type >= weatherExtinctionSnow)
    {
	/* It's snowing. See how much */
	sprintf(wthr_str, "Precipitation: \n");
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3 , y+(i++ * FONT_HEIGHT),
		    wthr_str, strlen(wthr_str));

	sprintf(wthr_str, "     %s %s\n", 
		get_range_string(wthr_parms->visual_range/1000),
		get_snow_string(wthr_parms->extinct_type));
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3 , y+(i++ * FONT_HEIGHT),
		    wthr_str, strlen(wthr_str));
	
    }
    else 
    {
	if (wthr_parms->precip_type == weatherPrecipitationNone)
	{
	    sprintf(wthr_str, "%s", "Precipitation: None");
	    XDrawString(dsp, win,
			env_display->key_gc,
			x+3 , y+(i++ * FONT_HEIGHT),
			wthr_str, strlen(wthr_str));
	}
	else
	{
	    sprintf(wthr_str, "Precipitation: Rain");
	    XDrawString(dsp, win,
			env_display->key_gc,
			x+3 , y+(i++ * FONT_HEIGHT),
			wthr_str, strlen(wthr_str));
	    sprintf(wthr_str, "Precip. Rate: %d mm/hr", 
		    (int32)wthr_parms->precip_rate);
	    XDrawString(dsp, win,
			env_display->key_gc,
			x+3 , y+(i++ * FONT_HEIGHT),
			wthr_str, strlen(wthr_str));
	}
    }
}


/* Note: These must match with seaedit.rdr! */
static char *breaker_type[] = {"None", "Plunging", "Spilling", "Surging"};

static void draw_sea_or_surf(Widget da_widget, PVD_ENV_DISPLAY *env_display)
{
    Arg args[6];
    int16 n;
    Dimension width;
    Dimension height;
    char title[16];
    uint32 w, h, r, xborder, yborder;
    int32 x,y;
    ENV_INTERFACE  ei;
    Display *dsp = XtDisplay(da_widget);
    Window win = XtWindow(da_widget);
    Window root;
    char wthr_str[256];
    uint32 border_width, depth;
    int32 i = 0;
    float32 value;
    int32 value_int;
    char sea_str[256];
    XmString xms;
    float32 wave_height;
    float32 wave_direction;
    
    XClearWindow(dsp, win);

    XGetGeometry(dsp, win, &root, (int *)&x, (int *)&y, (unsigned int *)&w, 
		 (unsigned int *)&h,
		 (unsigned int *)&border_width, (unsigned int *)&depth);

    value = (float32) (sqrt((env_display->last_sea_parms.current_velocity[X] *
			     env_display->last_sea_parms.current_velocity[X] +
			     env_display->last_sea_parms.current_velocity[Y] *
			     env_display->last_sea_parms.current_velocity[Y])));
    sprintf(sea_str, "Current Speed:  %.1f Knots",(value * 1.943844));
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3, y +(i++*FONT_HEIGHT),
		sea_str, strlen(sea_str));
    value_int = 
      (int32)RAD_TO_DEG(NS_ATAN2(env_display->last_sea_parms.current_velocity[Y],
				 env_display->last_sea_parms.current_velocity[X]));
    if (value_int < 0)
      value_int += 360;

    /* Convert to "From direction "  */
    sprintf(sea_str, "Current From:  %s", 
	    pvd_angle_name(value_int));
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3, y +(i++*FONT_HEIGHT),
		sea_str, strlen(sea_str));

    wave_height = (float32) (env_display->last_sea_parms.wave_height);

    wave_direction = (float32) (env_display->last_sea_parms.wave_direction);

    value = (float32)env_display->last_sea_parms.sea_salinity;
    sprintf(sea_str, "Salinity:  %d parts/1000", (int32)value);
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3, y +(i++*FONT_HEIGHT),
		sea_str, strlen(sea_str));

    value = (float32)env_display->last_sea_parms.surface_temp;
    sprintf(sea_str, "Surface Temp:  %d C", (int32)value);
    XDrawString(dsp, win,
		env_display->key_gc,
		x+3, y +(i++*FONT_HEIGHT),
		sea_str, strlen(sea_str));
    if (env_display->last_sea_parms.in_surf_zone)
    {
	/* In the surf zone, get surf parameters */
	xms = XmStringCreate("Surf Conditions", 
			     XmSTRING_DEFAULT_CHARSET);
	n = 0;
	XtSetArg(args[n], XmNlabelString, xms); n++;
	XtSetValues(env_display->sea_label, args, n);
	XmStringFree(xms);

	sprintf(sea_str, "Breaker Height: %.1f feet", wave_height*3.28084);
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));

	/* Convert to clockwise direction with 0 being East */
	value_int = 90 - (int32) RAD_TO_DEG(wave_direction);
	/* Sometimes the editor sends back extremely large values*/
	value_int = value_int % 360;
	if (value_int < 0)
	  value_int += 360;

	sprintf(sea_str, "Breakers From: %s", 
		pvd_angle_name(value_int));
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));

	sprintf(sea_str, "Breaker Angle To Beach:");
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));
	/* Angle is stored counterclockwise, convert to */
	/* clockwise (compass) degrees */
	sprintf(sea_str, "     %d Deg", 
	360-(int32)(RAD_TO_DEG(env_display->last_sea_parms.breaker_angle)));
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));

	sprintf(sea_str, "Breaker Type: %s", 
		breaker_type[env_display->last_sea_parms.breaker_type]);
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));
    }
    else
    {
	/* In the sea (or ground), get sea parameters */
	xms = XmStringCreate("Ocean Conditions", 
			     XmSTRING_DEFAULT_CHARSET);
	n = 0;
	XtSetArg(args[n], XmNlabelString, xms); n++;
	XtSetValues(env_display->sea_label, args, n);
	XmStringFree(xms);
	
	sprintf(sea_str, "Wave Height: %.1f feet", wave_height * 3.28084);
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));

	/* Convert to clockwise direction with 0 being East */	
	value_int = 90-(int32)RAD_TO_DEG(wave_direction);
	/* Sometimes the editor sends back extremely large values*/
	value_int = value_int % 360;

	if (value_int < 0)
	  value_int += 360;
	
	sprintf(sea_str, "Waves From : %s", 
		pvd_angle_name(value_int));
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));
	
	/* Display Sea State */
	sprintf(sea_str, "Sea State: %d", 
		env_display->last_sea_parms.sea_state);
	XDrawString(dsp, win,
		    env_display->key_gc,
		    x+3, y +(i++*FONT_HEIGHT),
		    sea_str, strlen(sea_str));
	
    }
}

/* Returns TRUE is the location is in the sea or surf, */
/* FALSE if on land. */

static int32 get_sea_or_surf_parms(PVD_ENV_DISPLAY *env_display, 
				   Sea_Parms *sea_parms)
{
    ENV_INTERFACE ei;
    static CTDB_ELEV_QUALIFIER qual=(CTDB_INCL_WATER |
				     CTDB_RETURN_MAX |
				     CTDB_INCL_MATERIAL);
    CTDB_ELEV_DATA maxelev;

    /* Determine whether location is on land or on water */

    ctdb_lookup_qual_elevation((CTDB *)gcs_get_tdb(GCS_ILLEGAL_CELL, FALSE),
			       env_display->view_location[X], 
			       env_display->view_location[Y],
			       qual, CTDB_NO_REF_ELEV, &maxelev);
    if (maxelev.class == CTDB_FC_ILLEGAL)
      return FALSE;


    bzero(&ei, sizeof(ENV_INTERFACE));
    /* Get current speed and direction */
    ei.u.sea_current_velocity.input_vector[X] = env_display->view_location[X];
    ei.u.sea_current_velocity.input_vector[Y] = env_display->view_location[Y];
    ei.u.sea_current_velocity.input_vector[Z] = env_display->view_location[Z];
    if (env_get(ENV_SEA_CURRENT_VELOCITY, &ei) != -1)
    {
	sea_parms->current_velocity[X] =
	  ei.u.sea_current_velocity.output_vector[X];
	sea_parms->current_velocity[Y] =
	  ei.u.sea_current_velocity.output_vector[Y];
	sea_parms->current_velocity[Z] =
	  ei.u.sea_current_velocity.output_vector[Z];
    }
    /* Get wave/breaker height and direction */
    ei.u.sea_wave_height.input_vector[X] = env_display->view_location[X];
    ei.u.sea_wave_height.input_vector[Y] = env_display->view_location[Y];
    ei.u.sea_wave_height.input_vector[Z] = env_display->view_location[Z];
    if (env_get(ENV_SEA_WAVE_HEIGHT, &ei) != -1)
    {
	sea_parms->wave_height = (float32) (ei.u.sea_wave_height.output_value);

    }
    else sea_parms->wave_height = 0;

    ei.u.sea_wave_direction.input_vector[X] = env_display->view_location[X];
    ei.u.sea_wave_direction.input_vector[Y] = env_display->view_location[Y];
    ei.u.sea_wave_direction.input_vector[Z] = env_display->view_location[Z];
    if (env_get(ENV_SEA_WAVE_DIRECTION, &ei) != -1)
    {
	sea_parms->wave_direction = 
	  (float32)ei.u.sea_wave_direction.output_value;

    }
    else sea_parms->wave_direction = 0.0;

    /* Get salinity and temperature */
    ei.u.sea_salinity.input_vector[X] =  env_display->view_location[X];
    ei.u.sea_salinity.input_vector[Y] =  env_display->view_location[Y];
    ei.u.sea_salinity.input_vector[Z] =  env_display->view_location[Z];
    if (env_get(ENV_SEA_SALINITY, &ei) != -1)
    {
	sea_parms->sea_salinity = (float32)ei.u.sea_salinity.output_value;
    }
    else
	sea_parms->sea_salinity = 0.0;

    ei.u.sea_surface_temp.input_vector[X] =  env_display->view_location[X];
    ei.u.sea_surface_temp.input_vector[Y] =  env_display->view_location[Y];
    ei.u.sea_surface_temp.input_vector[Z] =  env_display->view_location[Z];
    if (env_get(ENV_SEA_SURFACE_TEMP, &ei) != -1)
    {
	sea_parms->surface_temp = (float32)ei.u.sea_surface_temp.output_value;
    }
    else
      sea_parms->surface_temp = 0.0;

    ei.u.sea_in_surf_zone.input_vector[X] = env_display->view_location[X];
    ei.u.sea_in_surf_zone.input_vector[Y] = env_display->view_location[Y];
    ei.u.sea_in_surf_zone.input_vector[Z] = env_display->view_location[Z];

    if (env_get(ENV_SEA_IN_SURF_ZONE, &ei) != -1)
    {
	sea_parms->in_surf_zone = ei.u.sea_in_surf_zone.output_type;
    }
    else sea_parms->in_surf_zone = 0;

    ei.u.sea_breaker_angle.input_vector[X] = 
      env_display->view_location[X];
    ei.u.sea_breaker_angle.input_vector[Y] = 
      env_display->view_location[Y];
    ei.u.sea_breaker_angle.input_vector[Z] = 
      env_display->view_location[Z];
	    
    if (env_get(ENV_SEA_BREAKER_ANGLE, &ei) != -1)
    {
	sea_parms->breaker_angle = ei.u.sea_breaker_angle.output_value;
    }
    else
      sea_parms->breaker_angle = 0.0;

    ei.u.sea_breaker_type.input_vector[X] = 
      env_display->view_location[X];
    ei.u.sea_breaker_type.input_vector[Y] = 
      env_display->view_location[Y];
    ei.u.sea_breaker_type.input_vector[Z] = 
      env_display->view_location[Z];
    
    if (env_get(ENV_SEA_BREAKER_TYPE, &ei) != -1)
    {
	sea_parms->breaker_type = ei.u.sea_breaker_type.output_type;
    }
    else
      sea_parms->breaker_type = 0;

    ei.u.sea_state.input_vector[X] = 
      env_display->view_location[X];
    ei.u.sea_state.input_vector[Y] = 
      env_display->view_location[Y];
    ei.u.sea_state.input_vector[Z] = 
      env_display->view_location[Z];
    
    if (env_get(ENV_SEA_STATE, &ei) != -1)
    {
	sea_parms->sea_state = ei.u.sea_state.output_type;
    }
    else
      sea_parms->sea_state = 0;

    return(TRUE);
}




static void handle_sea_exposure(
    Widget widget, 
    PVD_ENV_DISPLAY *env_display, 
    XmDrawingAreaCallbackStruct *cb)
{
    
    Region region;

    region = XCreateRegion();
    XtAddExposureToRegion(cb->event, region);

    /* Update sea/surf parameters */
    if (!get_sea_or_surf_parms(env_display, &env_display->last_sea_parms))
    {
	/* Location is not on water, */
	/* Unmanage form widget */
	XtUnmanageChild(env_display->sea_form);
	XDestroyRegion(region);
	return;
    }

    XtManageChild(env_display->sea_form);
    /* Draw Sea/surf conditions */
    draw_sea_or_surf(widget, env_display);
    XDestroyRegion(region);

}


static void draw_illum_levels(
    Widget widget,
    PVD_ENV_DISPLAY *env_display)
{
    int32 n,left_x, right_x, top_y,bottom_y; 
    Dimension width, height;
    ENV_INTERFACE  ei;
    int32  bar_width;
    Arg    args[5];
    float64 direct;
    float64 diffuse;
    float64 total_illum;
    int16   direct_height, diffuse_height;
    float64 illum;
    int16   illum_height;
    float64 running_total;
    int16 tick_separator, i, graph_height;
    char    label_string[16];
#define NUM_TICKS      9
#define TICK_WIDTH     10
#define HALF_TICK_WIDTH     TICK_WIDTH/2
#define RIGHT_OFFSET   10
#define KEY_X_INIT_OFFSET      5
#define KEY_Y_INIT_OFFSET      6
#define KEY_X          40
#define KEY_Y          14
#define KEY_X_OFFSET   10
#define KEY_Y_OFFSET    4
#define LABEL_X        13
#define LABEL_Y         5

    bzero(&ei, sizeof(ENV_INTERFACE));
    XClearWindow(XtDisplay(widget), XtWindow(widget));

    n = 0;
    XtSetArg(args[n], XtNheight, &height); n++;
    XtSetArg(args[n], XtNwidth, &width); n++;
    XtGetValues(widget, args, n);

    /* Draw key */
    left_x = 2;
    top_y = 20;
    bottom_y = height - (2*top_y);
    right_x = width - RIGHT_OFFSET;

    XFillRectangle(XtDisplay(widget), XtWindow(widget),
		   env_display->direct_gc,
		   left_x, bottom_y + KEY_Y_INIT_OFFSET,
		   KEY_X, KEY_Y);

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x + 2,  bottom_y + 17, "Direct", 6);

    XFillRectangle(XtDisplay(widget), XtWindow(widget),
		   env_display->diffuse_gc,
		   left_x, bottom_y + KEY_Y_INIT_OFFSET+ KEY_Y_OFFSET + KEY_Y,
		   KEY_X, KEY_Y);

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x + 2, bottom_y + 37, "Diffuse", 7);

    XFillRectangle(XtDisplay(widget), XtWindow(widget),
		   env_display->other_gc,
		   left_x + KEY_X + KEY_X_OFFSET,
		   bottom_y + KEY_Y_INIT_OFFSET,
		   KEY_X, KEY_Y);

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x + 2 + KEY_X + KEY_X_OFFSET,  
		bottom_y + 17, "Other", 5);

    XFillRectangle(XtDisplay(widget), XtWindow(widget),
		   env_display->sky_gc,
		   left_x+KEY_X + KEY_X_OFFSET, 
		   bottom_y + KEY_Y_INIT_OFFSET+ KEY_Y_OFFSET + KEY_Y,
		   KEY_X, KEY_Y);

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x+2+KEY_X+KEY_X_OFFSET, 
		bottom_y + 37, "Sky", 3);

    XFillRectangle(XtDisplay(widget), XtWindow(widget),
		   env_display->moon_gc,
		   left_x + 2*(KEY_X + KEY_X_OFFSET),
		   bottom_y + KEY_Y_INIT_OFFSET,
		   KEY_X, KEY_Y);

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x + 2 + 2*(KEY_X + KEY_X_OFFSET),  
		bottom_y + 17, "Moon", 4);

    XFillRectangle(XtDisplay(widget), XtWindow(widget),
		   env_display->sun_gc,
		   left_x+ 2*(KEY_X + KEY_X_OFFSET), 
		   bottom_y + KEY_Y_INIT_OFFSET+ KEY_Y_OFFSET + KEY_Y,
		   KEY_X, KEY_Y);

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x+2+ 2*(KEY_X+KEY_X_OFFSET), 
		bottom_y + 37, "Sun", 3);


    /* Draw the graph template */
    left_x = 50;
    bar_width = (right_x-left_x)/12;

    /* Coordinates start at upper left corner */

    XDrawLine(XtDisplay(widget), XtWindow(widget),
	      env_display->black_gc,
	      left_x,top_y, left_x, bottom_y);
    XDrawLine(XtDisplay(widget), XtWindow(widget),
	      env_display->black_gc,
	      left_x, bottom_y, right_x, bottom_y);

    /* Label the Y axis */
    sprintf(label_string, "Log Lumens");

    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->key_gc,
		left_x - 35,
		top_y - 2,
		label_string, strlen(label_string));
    
    /* Draw ticks down the Y axis. */
    tick_separator = (bottom_y-top_y)/10;
    graph_height = bottom_y - top_y - tick_separator;
    for (i=1; i<=NUM_TICKS; i++)
    { 
	XDrawLine(XtDisplay(widget), XtWindow(widget),
		  env_display->black_gc,
		  left_x - HALF_TICK_WIDTH, top_y + i*tick_separator,
		  left_x + HALF_TICK_WIDTH, top_y + i*tick_separator);

	if (!(i%2))
	{
	    sprintf(label_string, "%2d", 6-i);

	    XDrawString(XtDisplay(widget), XtWindow(widget),
			env_display->label_gc,
			left_x - HALF_TICK_WIDTH - LABEL_X,
			top_y + i*tick_separator + LABEL_Y,
			label_string, strlen(label_string));
	}
    }

    /* Bottom tick label */
    sprintf(label_string, "%2d", -4);
    XDrawString(XtDisplay(widget), XtWindow(widget),
		env_display->label_gc,
		left_x - HALF_TICK_WIDTH - LABEL_X,
		top_y + 10*tick_separator + LABEL_Y,
		label_string, strlen(label_string));

    /* Find ambient light level. Our model supports only the visual */
    /* range. Right now illumination is constant over entire terrain */
    /* but we set the input vector for future compatibility */
    ei.u.illumination.vector_info = ENV_BAND_47MM;
    ei.u.illumination.input_vector[0] = env_display->view_location[X];
    ei.u.illumination.input_vector[1] = env_display->view_location[Y];
    ei.u.illumination.input_vector[2] = env_display->view_location[Z];
    if (env_get(ENV_ILLUMINATION, &ei) != -1)
    {
	total_illum = ei.u.illumination.output_vector[0] +
	  ei.u.illumination.output_vector[1];

	env_display->last_total_illum = total_illum;
	/* Convert from footcandles (FC) to lumens, */
	/* take the log, and normalize */
	if ((direct = ei.u.illumination.output_vector[0]) <= 0)
	  direct = 0;
	else if (direct >= 63452.78)
	  direct = 1;
	/* Convert the output to be between 10**-5 and 10**4 lumens. */
	/*  (0.16602675) is the log of FC_TO_Lumens */
	else direct = (log10(direct) + 0.16602675 + 4.0)/9.0;

	/* Convert from footcandles (FC) to lumens, */
	/* and normalize */
	if ((diffuse = ei.u.illumination.output_vector[1]) <= 0)
	  diffuse = 0;
	else if (diffuse >= 63452.78)
	  diffuse = 1;
	/* Convert the output to be between 10**-5 and 10**4 lumens. */
	/*  (0.16602675) is the log of FC_TO_Lumens */
	else 
	  {
	      diffuse = (log10(diffuse) + 0.16602675 + 4.0)/9.0;
	  }
    }
    else 
    {
	direct = 0;
	diffuse = 0;
    }

    /* Now that the values have been normalized, scale a bar chart */
    /* accordingly */
    direct_height = (int16)(graph_height * direct);
    if (direct_height < 0)
      direct_height = 0;

    XFillRectangle(XtDisplay(widget), XtWindow(widget), 
		   env_display->direct_gc,
		   left_x + bar_width, bottom_y-1-direct_height, bar_width, 
		   direct_height);

    diffuse_height = (int16)(graph_height * diffuse);

    if (diffuse_height < 0)
      diffuse_height = 0;

    XFillRectangle(XtDisplay(widget), XtWindow(widget), 
		   env_display->diffuse_gc,
		   left_x + 3*bar_width, 
		   bottom_y-1-diffuse_height,
		   bar_width, diffuse_height);
    running_total = 0;
    /* Now break down the illumination by source */
    if (env_get(ENV_SOLAR_ILLUMINATION, &ei) != -1)
    {
	/* Convert from footcandles (FC) to lumens */
	/* and normalize */
	illum = ei.u.illumination.output_vector[0] + 
	  ei.u.illumination.output_vector[1];

	env_display->last_solar_illum = illum;
	running_total += illum;

	if (illum <= 0)
	  illum = 0;
	else if (illum >= 63452.78)
	  illum = 1;
	else illum = (log10(illum) + 0.16602675 + 4.0)/9.0;
    }
    else
      illum = 0;

    illum_height = (int16)(graph_height * illum);
    if (illum_height < 0)
      illum_height = 0;

    XFillRectangle(XtDisplay(widget), XtWindow(widget), 
		   env_display->sun_gc,
		   left_x + 12*bar_width, bottom_y-1-illum_height, 
		   bar_width, 
		   illum_height);

    if (env_get(ENV_LUNAR_ILLUMINATION, &ei) != -1)
    {
	/* Convert from footcandles (FC) to lumens */
	/* and normalize */
	illum = ei.u.illumination.output_vector[0] + 
	  ei.u.illumination.output_vector[1];
	env_display->last_lunar_illum = illum;
	running_total += illum;
	if (illum <= 0)
	  illum = 0;
	else if (illum >= 63452.78)
	  illum = 1;
	else illum = (log10(illum) + 0.16602675 + 4.0)/9.0;
    }
    else
      illum = 0;

    illum_height = (int16)(graph_height * illum);
    if (illum_height < 0)
      illum_height = 0;
    XFillRectangle(XtDisplay(widget), XtWindow(widget), 
		   env_display->moon_gc,
		   left_x + 10*bar_width, bottom_y-1-illum_height, 
		   bar_width, 
		   illum_height);

    /* Now for stellar */
    if (env_get(ENV_SKY_ILLUMINATION, &ei) != -1)
    {
	/* Convert from footcandles (FC) to lumens */
	/* and normalize */
	illum = ei.u.illumination.output_vector[0] + 
	  ei.u.illumination.output_vector[1];
	env_display->last_sky_illum = illum;
	running_total += illum;
	if (illum <= 0)
	  illum = 0;
	else if (illum >= 63452.78)
	  illum = 1;
	else illum = (log10(illum) + 0.16602675 + 4.0)/9.0;
    }
    else
      illum = 0;

    illum_height = (int16)(graph_height * illum);
    if (illum_height < 0)
      illum_height = 0;
    XFillRectangle(XtDisplay(widget), XtWindow(widget), 
		   env_display->sky_gc,
		   left_x + 8*bar_width, bottom_y-1-illum_height, 
		   bar_width, 
		   illum_height);
	    
    /* Now calculate any other illumination */
    illum = total_illum - running_total;
    if (illum <= 0)
      illum = 0;
    else if (illum >= 63452.78)
      illum = 1;
    else illum = (log10(illum) + 0.16602675 + 4.0)/9.0;

    illum_height = (int16)(graph_height * illum);
    if (illum_height < 0)
      illum_height = 0;
    XFillRectangle(XtDisplay(widget), XtWindow(widget), 
		   env_display->other_gc,
		   left_x + 6*bar_width, bottom_y-1-illum_height, 
		   bar_width, 
		   illum_height);


}


static void handle_illum_level_exposure(
    Widget                       widget,
    PVD_ENV_DISPLAY             *env_display,
    XmDrawingAreaCallbackStruct *cb)
{
    
    Region region;

    region = XCreateRegion();
    XtAddExposureToRegion(cb->event, region);

    /* Draw illumination levels */
    draw_illum_levels(widget, env_display);
    XDestroyRegion(region);

}
    


static void query_skyview(PVD_ENV_DISPLAY *env_display)
{
    ENV_INTERFACE ei;
    int32         drew_skyview = FALSE;

    bzero(&ei, sizeof(ENV_INTERFACE));

    ei.u.sun_position.input_vector[0] = env_display->view_location[X];
    ei.u.sun_position.input_vector[1] = env_display->view_location[Y];
    ei.u.sun_position.input_vector[2] = env_display->view_location[Z];

    if (env_get(ENV_SUN_POSITION, &ei) != -1)
    {
	/* The sum of the altitude and azimuth will have changed if */
	/* either has changed */
	if ((ei.u.sun_position.output_vector[0] +
	    ei.u.sun_position.output_vector[1]) != env_display->last_sun_pos)
	{
	    draw_skyview(env_display->skyview_da, env_display);
	    drew_skyview = TRUE;
	}
    }

    if (!drew_skyview)
    {
	ei.u.moon_position.input_vector[0] = env_display->view_location[X];
	ei.u.moon_position.input_vector[1] = env_display->view_location[Y];
	ei.u.moon_position.input_vector[2] = env_display->view_location[Z];
	if (env_get(ENV_MOON_POSITION, &ei) != -1)
	{
	    /* The sum of the altitude and azimuth will have changed if */
	    /* either has changed */
	    if ((ei.u.moon_position.output_vector[0] +
		 ei.u.moon_position.output_vector[1]) != 
		env_display->last_moon_pos)
	    {
		draw_skyview(env_display->skyview_da, env_display);
	    }
	}
    }

    draw_moonphase(env_display->moonphase_da, env_display);

}

static void query_illum_levels (PVD_ENV_DISPLAY *env_display)
{
    int32         drew_illum = FALSE;
    ENV_INTERFACE ei;

    bzero(&ei, sizeof(ENV_INTERFACE));

    ei.u.illumination.vector_info = ENV_BAND_47MM;
    ei.u.illumination.input_vector[0] = env_display->view_location[X];
    ei.u.illumination.input_vector[1] = env_display->view_location[Y];
    ei.u.illumination.input_vector[2] = env_display->view_location[Z];
    if (env_get(ENV_ILLUMINATION, &ei) != -1)
    {
	/* The sum of the altitude and azimuth will have changed if */
	/* either has changed */
	if ((ei.u.illumination.output_vector[0] +
	    ei.u.illumination.output_vector[1]) != 
	    env_display->last_total_illum)
	{
	    draw_illum_levels(env_display->illum_da, env_display);
	    drew_illum = TRUE;
	}
    }

    if (!drew_illum)
    {
	ei.u.solar_illumination.input_vector[0] = 
	  env_display->view_location[X];
	ei.u.solar_illumination.input_vector[1] = 
	  env_display->view_location[Y];
	ei.u.solar_illumination.input_vector[2] = 
	  env_display->view_location[Z];
	if (env_get(ENV_SOLAR_ILLUMINATION, &ei) != -1)
	{
	    /* The sum of the altitude and azimuth will have changed if */
	    /* either has changed */
	    if ((ei.u.solar_illumination.output_vector[0] +
		 ei.u.solar_illumination.output_vector[1]) != 
		env_display->last_solar_illum)
	    {
		draw_illum_levels(env_display->illum_da, env_display);
		drew_illum = TRUE;
	    }
	}

    }

    if (!drew_illum)
    {
	ei.u.lunar_illumination.input_vector[0] = 
	  env_display->view_location[X];
	ei.u.lunar_illumination.input_vector[1] = 
	  env_display->view_location[Y];
	ei.u.lunar_illumination.input_vector[2] = 
	  env_display->view_location[Z];
	if (env_get(ENV_LUNAR_ILLUMINATION, &ei) != -1)
	{
	    /* The sum of the altitude and azimuth will have changed if */
	    /* either has changed */
	    if ((ei.u.lunar_illumination.output_vector[0] +
		 ei.u.lunar_illumination.output_vector[1]) != 
		env_display->last_lunar_illum)
	    {
		draw_illum_levels(env_display->illum_da, env_display);
		drew_illum = TRUE;
	    }
	}
    }

    if (!drew_illum)
    {
	ei.u.sky_illumination.input_vector[0] = env_display->view_location[X];
	ei.u.sky_illumination.input_vector[1] = env_display->view_location[Y];
	ei.u.sky_illumination.input_vector[2] = env_display->view_location[Z];
	if (env_get(ENV_SKY_ILLUMINATION, &ei) != -1)
	{
	    /* The sum of the altitude and azimuth will have changed if */
	    /* either has changed */
	    if ((ei.u.sky_illumination.output_vector[0] +
		 ei.u.sky_illumination.output_vector[1]) != 
		env_display->last_sky_illum)
	    {
		draw_illum_levels(env_display->illum_da, env_display);
		drew_illum = TRUE;
	    }
	}
    }
}

static void get_weather_parms(
    PVD_ENV_DISPLAY *env_display, 
    Weather_Parms *wthr_parms)
{

    ENV_INTERFACE ei;

    memset(&ei, 0, sizeof(ENV_INTERFACE));

    ei.u.wind_velocity.input_vector[X] = env_display->view_location[X];
    ei.u.wind_velocity.input_vector[Y] = env_display->view_location[Y];
    ei.u.wind_velocity.input_vector[Z] = env_display->view_location[Z];

    if (env_get(ENV_WIND_VELOCITY, &ei) != -1)
    {
	wthr_parms->wind_velocity[X] = 
	  ei.u.wind_velocity.output_vector[X];
	wthr_parms->wind_velocity[Y] = 
	  ei.u.wind_velocity.output_vector[Y];
	wthr_parms->wind_velocity[Z] = 
	  ei.u.wind_velocity.output_vector[Z];
    }

    ei.u.cloud_cover.input_vector[X] = env_display->view_location[X];
    ei.u.cloud_cover.input_vector[Y] = env_display->view_location[Y];
    ei.u.cloud_cover.input_vector[Z] = env_display->view_location[Z];

    wthr_parms->cloud_cover = 0.0;
    if (env_get(ENV_CLOUD_COVER, &ei) != -1)
    {
	wthr_parms->cloud_cover = 
	  ei.u.cloud_cover.output_value;
    }

    ei.u.extinction_type.input_vector[X] = env_display->view_location[X];
    ei.u.extinction_type.input_vector[Y] = env_display->view_location[Y];
    ei.u.extinction_type.input_vector[Z] = env_display->view_location[Z];
    
    if (env_get(ENV_EXTINCTION_TYPE, &ei) != -1)
    {
	wthr_parms->extinct_type = 
	  ei.u.extinction_type.output_type;
    }

    ei.u.visual_range.input_wavelength = ENV_BAND_47MM;
    if (env_get(ENV_VISUAL_RANGE, &ei)  != -1)
    {
	wthr_parms->visual_range = 
	  ei.u.visual_range.output_value;
    }

    ei.u.precipitation_type.input_vector[X] = 
      env_display->view_location[X];
    ei.u.precipitation_type.input_vector[Y] = 
      env_display->view_location[Y];
    ei.u.precipitation_type.input_vector[Z] = 
      env_display->view_location[Z];

    if (env_get(ENV_PRECIPITATION_TYPE, &ei) != -1)
    {
	wthr_parms->precip_type = 
	  ei.u.precipitation_type.output_type;
    }

    ei.u.precipitation_rate.input_vector[X] = env_display->view_location[X];
    ei.u.precipitation_rate.input_vector[Y] = env_display->view_location[Y];
    ei.u.precipitation_rate.input_vector[Z] = env_display->view_location[Z];

    if (env_get(ENV_PRECIPITATION_RATE, &ei) != -1)
    {
	wthr_parms->precip_rate = 
	  ei.u.precipitation_rate.output_value;
    }
}

static void handle_weather_exposure(
    Widget                       widget,
    PVD_ENV_DISPLAY             *env_display,
    XmDrawingAreaCallbackStruct *cb)
{
    
    Region region;

    region = XCreateRegion();
    XtAddExposureToRegion(cb->event, region);

    /* Update Weather parameters */
    get_weather_parms(env_display, &env_display->last_wthr_parms);

    /* Draw Weather conditions */
    draw_weather(widget, env_display);
    XDestroyRegion(region);

}


static void query_weather(PVD_ENV_DISPLAY *env_display)
{
    Weather_Parms wthr_parms;
    
    memset(&wthr_parms, 0, sizeof(wthr_parms));

    /* Get current weather parameters */
    get_weather_parms(env_display, &wthr_parms);

    if (memcmp(&wthr_parms, &env_display->last_wthr_parms, 
	       sizeof(Weather_Parms)))
    {
	/* Something is different.  Update the display */
	memcpy(&env_display->last_wthr_parms, &wthr_parms, 
	       sizeof(Weather_Parms));
	draw_weather(env_display->weather_da, env_display);
    }
}

	
static int16 two_minutes_later(
    uint32  old_time,
    uint32  new_time)
{

    if ((new_time-old_time) >= (2*60))
      return(TRUE);
    else return(FALSE);
}

  
static void query_date(PVD_ENV_DISPLAY *env_display)
{
    uint32 new_time;

    new_time = get_env_time(env_display);

    if (two_minutes_later(env_display->last_uxtime, new_time))
    {
	env_display->last_uxtime = new_time;
	draw_date(env_display->date_da, env_display,
		  env_display->last_uxtime);
    }
}


/* Decide whether sea or surf parameters have changed */
/* enough to warrant redrawing */    
static void query_sea_or_surf(PVD_ENV_DISPLAY *env_display)
{

    Sea_Parms sea_parms;

    memset(&sea_parms, 0, sizeof(sea_parms));

    /* Get current sea parameters */
    /* If location is not on water, don't draw Sea form at all */
    if (!get_sea_or_surf_parms(env_display, &sea_parms))
    {
	/* Location is not on water, unmanage form widget */
	XtUnmanageChild(env_display->sea_form);
	return;
    }

    XtManageChild(env_display->sea_form);
    if (memcmp(&sea_parms, &env_display->last_sea_parms, sizeof(Sea_Parms)))
    {
	/* Something is different.  Update the display. */
	memcpy(&env_display->last_sea_parms, &sea_parms, sizeof(Sea_Parms));
	draw_sea_or_surf(env_display->sea_da, env_display);
    }
    
}



/* Decide whether environmental display parameters */
/* have changed enough to warrant redrawing */
static void query_environment(PVD_ENV_DISPLAY *env_display)
{

    query_skyview(env_display);
    query_illum_levels(env_display);
    query_weather(env_display);
    query_date(env_display);
    query_sea_or_surf(env_display);
    
}


/* Manages Environmental Display Widget and schedules */
/* periodic update function */
static void show_env_display(
    Widget               w,
    PVD_GUI_PTR          pvd_gui,
    XmAnyCallbackStruct *call_data)
{
    /* Pop up window of environmental display */
    XtManageChild(pvd_gui->env_display.env_display_form);
    /* Schedule periodic function to update display if illumination or */
    /* sun or moon positions change. */

    pvd_gui->env_display.update_function = 
      sched_periodic_fncl(query_environment, 5000 /* milliseconds */,
			  5000, 0, A_PTR, &pvd_gui->env_display, A_END);
    
}

void pvd_create_env_display(
    Widget           parent,
    PVD_GUI_PTR      pvd_gui,
    Pixel            foreground,
    Font             font,
    Window           root,
    PVD_ENV_DISPLAY *env_display)
{
    Widget form, done_button, drawing_area, skyview_form, label;
    Widget illum_form, w, weather_form, sea_form;
    Arg   args[10];
    XmString xms;
    int32 n;
    XGCValues gcv;
    Pixmap stipple;
    float64 temp_loc[XYC];
    struct pvd_env_resources
    {
	Pixel black;
	Pixel skybackground;
	Pixel this_sun;
	Pixel moon;
	Pixel sky;
	Pixel other;
	Pixel direct;
	Pixel diffuse;
	Font  label;
	Font  key;
    } resources;
    static XtResource res_list[] = {
	{ "black" , "Black" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, black) ,
	    XmRString , "Black" } ,
	{ "skybackground" , "Skybackground" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, skybackground) ,
	    XmRString , "#00004a" } ,
	{ "sun" , "Sun" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, this_sun) ,
	    XmRString , "Yellow" } ,
	{ "moon" , "Moon" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, moon) ,
	    XmRString , "Gray100" } ,
	{ "sky" , "Sky" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, sky) ,
	    XmRString , "Cornflowerblue" } ,
	{ "other" , "Other" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, other) ,
	    XmRString , "darkorchid2" } ,
	{ "direct" , "Direct" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, direct) ,
	    XmRString , "Grey58" } ,
	{ "diffuse" , "Diffuse" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_env_resources *, diffuse) ,
	    XmRString , "Grey40" } ,
	{ "label" , "Label" , XmRFont , sizeof(Font) ,
	    XtOffset(struct pvd_env_resources *, label) ,
	    XmRString , "Fixed" } ,
	{ "key" , "Key" , XmRFont , sizeof(Font) ,
	    XtOffset(struct pvd_env_resources *, key) ,
	    XmRString , "Fixed" } ,

    };

    /* Set the default query location to be the middle of the playbox */
    wld_get_playbox_center(temp_loc);
    env_display->view_location[X] = temp_loc[X];
    env_display->view_location[Y] = temp_loc[Y];
    env_display->view_location[CELL3D] = temp_loc[CELL2D];
    env_display->view_location[Z] = 
      ctdb_lookup_elevation((CTDB *)gcs_get_tdb(temp_loc[CELL2D], FALSE),
			    temp_loc[X],
			    temp_loc[Y]);

    /* Get the colors, and various other resources used */
    XtGetSubresources(parent, &resources,
		      "env_display", "EnvDisplay",
		      res_list, XtNumber(res_list), NULL, 0);

    /* Add a menu item for showing display */
    w = sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR, "Special",
			 "Show Environmental Display",
			 xmPushButtonWidgetClass);
    XtAddCallback(w, XmNactivateCallback, 
		  (XtCallbackProc)show_env_display,
		  pvd_gui);

    n = 0;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_MODELESS); n++; 
    xms = XmStringCreate("Environmental Display", 
			 XmSTRING_DEFAULT_CHARSET);
    XtSetArg(args[n], XmNdialogTitle, xms); n++;
    XtSetArg(args[n], XmNmarginWidth, 3); n++;
    XtSetArg(args[n], XmNleftOffset, 3); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    form = XmCreateFormDialog(parent, "EnvDisplay", args, n); 

    XmStringFree(xms);
    env_display->env_display_form = form;

    /* Make Skyview form */
    n = 0;

    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 3); n++;
    XtSetArg(args[n], XmNleftOffset, 3); n++;
    XtSetArg(args[n], XmNborderWidth, 2); n++;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_MODELESS); n++; 
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    skyview_form = XtCreateManagedWidget("SkyViewForm",
					 xmFormWidgetClass,
					 form, args, n);

    /* Make Skyview title label */
    xms = XmStringCreate("SkyView", XmSTRING_DEFAULT_CHARSET);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNlabelString, xms); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1)); n++;
    XtSetArg(args[n], XmNforeground, resources.black); n++;
    label = XtCreateManagedWidget("SkyViewLabel", xmLabelWidgetClass,
				  skyview_form, args, n);
    XmStringFree(xms);

    /* Make moon phase drawing area */
    n = 0;
    XtSetArg(args[n], XtNheight, 40); n++;
    XtSetArg(args[n], XtNwidth, 40); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 3); n++;
    XtSetArg(args[n], XmNrightOffset, 3); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1)); n++;

    drawing_area = XtCreateManagedWidget("MoonPhase", 
					 xmDrawingAreaWidgetClass,
					 skyview_form, args, n);

    env_display->moonphase_da = drawing_area;
    XtAddCallback(drawing_area, XmNexposeCallback,
		  (XtCallbackProc)handle_moonphase_exposure, env_display);

    /* Make skyview drawing area */
    n = 0;
    XtSetArg(args[n], XtNheight, 150); n++;
    XtSetArg(args[n], XtNwidth, 150); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, label); n++;
    XtSetArg(args[n], XmNtopOffset, 0); n++; 
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;

    drawing_area = XtCreateManagedWidget("SkyView", 
					 xmDrawingAreaWidgetClass,
					 skyview_form, args, n);

    env_display->skyview_da = drawing_area;
    XtAddCallback(drawing_area, XmNexposeCallback,
		  (XtCallbackProc)handle_skyview_exposure, env_display);
    
    /* Make Close button */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, skyview_form); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNforeground, resources.black); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;

    XtSetArg(args[n], XmNleftOffset, 3); n++;
    
    done_button = XtCreateManagedWidget("Close",
					xmPushButtonWidgetClass,
					form, args, n);
    XtAddCallback (done_button, XmNactivateCallback,
		    (XtCallbackProc)done_callback, pvd_gui);

    /* Make Illumination form */
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, skyview_form);n++;
    XtSetArg(args[n], XmNtopOffset, 3); n++;
    XtSetArg(args[n], XmNleftOffset, 3); n++;
    XtSetArg(args[n], XmNborderWidth, 2); n++;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_MODELESS); n++; 
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    
    illum_form = XtCreateManagedWidget("IlluminationForm",
					 xmFormWidgetClass,
					 form, args, n);

    /* Make title label */
    xms = XmStringCreate("Illumination Levels", 
			 XmSTRING_DEFAULT_CHARSET);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNlabelString, xms); n++;
    XtSetArg(args[n], XmNforeground, resources.black); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    label = XtCreateManagedWidget("IlluminationLabel", 
				  xmLabelWidgetClass,
				  illum_form, args, n);
    XmStringFree(xms);

    n = 0;
    XtSetArg(args[n], XtNheight, 150); n++;
    XtSetArg(args[n], XtNwidth, 150); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, label); n++;
    XtSetArg(args[n], XmNtopOffset, 0); n++; 
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;

    drawing_area = XtCreateManagedWidget("Illumination Levels",
					 xmDrawingAreaWidgetClass,
					 illum_form, 
					 args, n);
    env_display->illum_da = drawing_area;
    XtAddCallback(drawing_area, XmNexposeCallback,
		  (XtCallbackProc)handle_illum_level_exposure, env_display);

    /* Make Weather Condition form */
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, illum_form);n++;
    XtSetArg(args[n], XmNtopOffset, 3); n++;
    XtSetArg(args[n], XmNleftOffset, 3); n++;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_MODELESS); n++; 
    XtSetArg(args[n], XmNborderWidth, 2); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    
    weather_form = XtCreateManagedWidget("WeatherForm",
					 xmFormWidgetClass,
					 form, args, n);

    /* Make title label */
    xms = XmStringCreate("Weather Conditions", 
			 XmSTRING_DEFAULT_CHARSET);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNlabelString, xms); n++;
    XtSetArg(args[n], XmNforeground, resources.black); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    label = XtCreateManagedWidget("WeatherLabel", 
				  xmLabelWidgetClass,
				  weather_form, args, n);
    XmStringFree(xms);

    /* Make Weather drawing area */
    n = 0;
    XtSetArg(args[n], XtNheight, 100); n++;
    XtSetArg(args[n], XtNwidth, 150); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, label); n++;
    XtSetArg(args[n], XmNtopOffset, 0); n++; 
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;

    drawing_area = XtCreateManagedWidget("WeatherConditions",
					 xmDrawingAreaWidgetClass,
					 weather_form, 
					 args, n);

    XtAddCallback(drawing_area, XmNexposeCallback,
		  (XtCallbackProc)handle_weather_exposure, env_display);

    env_display->weather_da = drawing_area;

    /* Make Date/time widget */
    n = 0;
    XtSetArg(args[n], XtNheight, 44); n++;
    XtSetArg(args[n], XtNwidth, 153); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, illum_form); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, weather_form); n++;
    XtSetArg(args[n], XmNtopOffset, 2); n++; 
    XtSetArg(args[n], XmNleftOffset, 3); n++; 
    XtSetArg(args[n], XmNborderWidth, 2); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;

    drawing_area = XtCreateManagedWidget("DateDrawingArea",
			       xmDrawingAreaWidgetClass,
			       form, args, n);
    
    XtAddCallback(drawing_area, XmNexposeCallback,
		  (XtCallbackProc)handle_date_exposure, env_display);

    env_display->date_da = drawing_area;


    /* Make Sea/Surf Conditions form */
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, weather_form);n++;
    XtSetArg(args[n], XmNtopOffset, 3); n++;
    XtSetArg(args[n], XmNleftOffset, 3); n++;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_MODELESS); n++; 
    XtSetArg(args[n], XmNborderWidth, 2); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    
    sea_form = XtCreateManagedWidget("SeaForm",
				     xmFormWidgetClass,
				     form, args, n);

    env_display->sea_form = sea_form;

    /* Make title label */
    xms = XmStringCreate("Ocean Conditions", 
			 XmSTRING_DEFAULT_CHARSET);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNlabelString, xms); n++;
    XtSetArg(args[n], XmNforeground, resources.black); n++;
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;
    env_display->sea_label = XtCreateManagedWidget("SeaLabel", 
				  xmLabelWidgetClass,
				  sea_form, args, n);
    XmStringFree(xms);

    /* Make Sea/Surf conditions drawing area */
    n = 0;
    XtSetArg(args[n], XtNheight, 150); n++;
    XtSetArg(args[n], XtNwidth, 150); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, env_display->sea_label); n++;
    XtSetArg(args[n], XmNtopOffset, 0); n++; 
    XtSetArg(args[n], XmNbackground,tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1) ); n++;

    drawing_area = XtCreateManagedWidget("SeaConditions",
					 xmDrawingAreaWidgetClass,
					 sea_form, 
					 args, n);

    XtAddCallback(drawing_area, XmNexposeCallback,
		  (XtCallbackProc)handle_sea_exposure, env_display);

    env_display->sea_da = drawing_area;

    /* make the black gc */
    gcv.font = font;
    gcv.foreground = resources.black;
    gcv.background = tactmap_get_color(pvd_gui->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1);
    env_display->black_gc = XCreateGC(XtDisplay(form), root,
			       GCFont | GCForeground | GCBackground,
					      &gcv);

    /* make the direct illumination gc */
    
    gcv.foreground = resources.direct;
    env_display->direct_gc = XCreateGC(XtDisplay(form), root,
				       GCForeground|GCBackground, &gcv);

    /* make the diffuse illumination gc */
    gcv.foreground = resources.diffuse;
    env_display->diffuse_gc = XCreateGC(XtDisplay(form), root,
				       GCForeground|GCBackground, &gcv);

    /* make the horizon gc */
    env_display->horizon_gc = XCreateGC(XtDisplay(form), root,
			       GCFont | GCForeground | GCBackground,
					      &gcv);

    /* make the outer horizon gc */
    gcv.foreground = resources.black;
    gcv.line_style = LineOnOffDash;
    env_display->outer_horizon_gc = XCreateGC(XtDisplay(form), root,
			       GCLineStyle | GCForeground | GCBackground,
				&gcv);

    /* make the sun gc */
    gcv.foreground = resources.this_sun;
    env_display->sun_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground,
					      &gcv);

    /* make the skybackground gc */
    gcv.foreground = resources.skybackground;
    gcv.background = resources.moon;
    env_display->skybackground_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground,
					      &gcv);

    /* make the moon gc */
    gcv.foreground = resources.moon;
    gcv.background = resources.skybackground;
    env_display->moon_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground | GCFont,
				     &gcv);

    /* make the sky illum gc */
    gcv.foreground = resources.sky;
    env_display->sky_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground,
				     &gcv);

    /* make the "other" illum gc */
    gcv.foreground = resources.other;
    env_display->other_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground,
				     &gcv);

    /* make the GC to draw labels */
    gcv.foreground = resources.black;
    gcv.font = resources.label;
    env_display->label_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground | GCFont,
				     &gcv);

    /* make the GC to draw labels */
    gcv.foreground = resources.black;
    gcv.font = resources.key;
    env_display->key_gc = XCreateGC(XtDisplay(form), root,
			       GCForeground | GCBackground | GCFont,
				     &gcv);

    /* If we don't do this, we crash if we go into env_gesture */
    /* before bringing up the environmental display because it's */
    /* not managed right away. */
    XtRealizeWidget(form);

}


#endif
/*USE_MOTIF*/
