#ifndef lint
static char rcsid [] = "$RCSfile: pvd_impact.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_impact.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw impacts                                      *
*       Created: Thu Oct  1 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <libpduapi.h>
#include <libentity.h>
#include <libtime.h>
#include <libsched.h>
#include <obj_type.h>
#include <mun_type.h>
#include <p_aspid.h>
#include <stdext.h> /*common/include/global*/
#include <math.h>
#include <libpduapi.h>
#include <libgcs.h> /* AJC */

/* The amount of simulated time, we leave effects on the screen */
#define ERASE_DELAY 2000
#define FIRE_ERASE_DELAY 1000

/* FireResult code for Stingray fire */
#define STINGRAY_DETECT    254
#define STINGRAY_NO_DETECT 253
#define STINGRAY_DAMAGE    252
#define STINGRAY_NO_DAMAGE 251

/* Drawing parameters */
#define CIRCLE_RAD   10
#define BURST_LENGTH 50

static short burst_segs[] = {
    -BURST_LENGTH , -BURST_LENGTH ,  BURST_LENGTH ,  BURST_LENGTH ,
    -BURST_LENGTH , 0             ,  BURST_LENGTH , 0             ,
    -BURST_LENGTH ,  BURST_LENGTH ,  BURST_LENGTH , -BURST_LENGTH ,
    0             ,  BURST_LENGTH , 0             , -BURST_LENGTH };

static short coll_segs[] = {
     0  , -30	,  6  , -12	,  21 , -21	,  12 , -6	,
     30 ,  0	,  12 ,  6	,  21 ,  21	,  6  ,  12	,
     0  ,  30	, -6  ,  12	, -21 ,  21	, -12 ,  6	,
    -30 ,  0	, -12 , -6	, -21 , -21	, -6  , -12	,
     0  , -30 };

static void draw_df(PVD_GUI_PTR pvd_gui, uint32 from, uint32 to,
		    int32 result);
static void draw_laser(
    PVD_GUI_PTR pvd_gui,
    int32 erase,
    int32 sfrom[XY],
    int32 sto[XY]);
static void draw_if(PVD_GUI_PTR pvd_gui, int32 from_x, int32 from_y, 
		    int32 mun);
static void draw_coll(PVD_GUI_PTR pvd_gui, int32 x, int32 y, int32 draw_p);
static void draw_fire(PVD_GUI_PTR pvd_gui, uint32 from, float64 angle,
		      int32 draw_it);

#define IS_MINE(x) (((x) & (objectDomainMask | munitionClassMask)) == \
		    (objectDomainMunition | munitionClassMine))

int32 pvd_impact_fmt;
int32 pvd_indirect_fmt;
int32 pvd_collision_fmt;
int32 pvd_fire_fmt;
int32 pvd_radio_fmt;
int32 pvd_srdetect_fmt;
int32 pvd_srimpact_fmt;
int32 pvd_laser_fmt;

#define LASER_SPOT_TIMEOUT 1000 /* in ms */
#define MAX_LASER_CODE 256
static QUEUE_QUEUE laser_spots[MAX_LASER_CODE];

void pvd_init_pdu(void)
{
    int32 detonation_type =
      pa_define_type("pvd_detonation", sizeof(struct pvd_detonation),
		     "delay", pa_uint32,
		       OFFSETOF(struct pvd_detonation, delay),
		     "location", pa_zup_world_coordinate64,
		       OFFSETOF(struct pvd_detonation, location[X]),
		     A_END);

#define PDU_FIELD(s,t,m) s, t, OFFSETOF(PVD_IMPACT, m)

    pvd_impact_fmt =
      pa_define_input_format("detonation", sizeof(PVD_IMPACT),
		     PDU_FIELD("result", pa_simnet_detonation_result, result),
		     PDU_FIELD("result", pa_dis203_detonation_result, 
			       dis_result),
		     PDU_FIELD("firing_id", pa_hashed_id, firer_id),
		     PDU_FIELD("target_id", pa_hashed_id, target_id),
		     PDU_FIELD("burst", pa_simnet_burst, burst),
		     PDU_FIELD("location",pa_zup_world_coordinate64,
			   detonations[0].location[X]),
			     A_END);

    pvd_indirect_fmt =
      pa_define_input_format("indirect_fire", sizeof(PVD_IMPACT),
			     PDU_FIELD("firing_id", pa_hashed_id, firer_id),
			     PDU_FIELD("burst", pa_simnet_burst, burst),
			     "detonations", pa_array,
			       detonation_type, "burst.quantity",
			       PVD_MAX_DETONATIONS,
			       OFFSETOF(PVD_IMPACT, detonations[0]),
			     A_END);

    pvd_collision_fmt =
      pa_define_input_format("collision", sizeof(PVD_IMPACT),
			     PDU_FIELD("issuing_id", pa_hashed_id, firer_id),
			     PDU_FIELD("colliding_id", pa_hashed_id, target_id)
			     ,
			     A_END);

    pvd_fire_fmt =
      pa_define_input_format("fire", sizeof(PVD_IMPACT),
		     PDU_FIELD("location", pa_zup_world_coordinate64, 
			       firing_location[0]),
		     PDU_FIELD("velocity", pa_zup_velocity32, 
			       munition_velocity[0]),
		     A_END);

    pvd_radio_fmt =
      pa_define_input_format("signal", sizeof(PVD_SIGNAL),
			     "entity_id", pa_hashed_id,
			       OFFSETOF(PVD_SIGNAL, hashed_id),
			     "entity_id", pa_unhashed_id,
			       OFFSETOF(PVD_SIGNAL, unhashed_id),
			     "length", pa_int32,
			       OFFSETOF(PVD_SIGNAL, length_bits),
			     "?.encoding_scheme", pa_uint16,
			       OFFSETOF(PVD_SIGNAL, encoding_scheme),
			     "data", pa_array, pa_uint1, "length",
			       PVD_MAX_MSG*8, OFFSETOF(PVD_SIGNAL, data[0]),
			     A_END);

    pvd_srdetect_fmt =
      pa_define_input_format("detection_response", sizeof(PVD_IMPACT),
		     PDU_FIELD("stingray_id", pa_hashed_id, firer_id),
		     PDU_FIELD("target_id", pa_hashed_id, target_id),
		     PDU_FIELD("result", pa_uint8, result),
			     A_END);

    pvd_srimpact_fmt =

      pa_define_input_format("impact_response", sizeof(PVD_IMPACT),
                     PDU_FIELD("stingray_id", pa_hashed_id, firer_id),
                     PDU_FIELD("target_id", pa_hashed_id, target_id),
                     PDU_FIELD("result", pa_uint8, result),
                             A_END);

    pvd_laser_fmt =
      pa_define_input_format("laser", sizeof(PVD_LASER),
			     "laser_code", pa_int32,
			     OFFSETOF(PVD_LASER, laser_code),
			     "lasing_id", pa_hashed_id,
			     OFFSETOF(PVD_LASER, lasing_id),
			     "lased_id", pa_hashed_id,
			     OFFSETOF(PVD_LASER, lased_id),
			     "laser_spot_location", pa_zup_world_coordinate64,
			     OFFSETOF(PVD_LASER, lased_location),
			     A_END);

    bzero(laser_spots, sizeof(laser_spots));
}


/*ARGSUSED*/
void pvd_process_impact(PV_PACKET *packet, int32 format, PVD_IMPACT *pdu,
			PVD_GUI_PTR pvd_gui)
{
    float64 fire[XYZC]; /* firing position in GCS */
    float64 targ[XYZC]; /* target position in GCS */
    float64 det[XYZC] ; /* detonation position in GCS */
    int32   sfrom[XY] ; /* screen coordinate */
    int32   sto[XY]   ; /* screen coordinates */
    uint32  from, to  ; /* compressed (uint16) screen position (X,Y) */
    int32 i;
    uint32 when;
    int32 vehicle_id;
    ObjectType simnet_munition;
    float64 angle;

    if (format == pvd_impact_fmt)
    {
	int32 treat_as_IF = FALSE;

	if (!pvd_gui->draw_impacts)
	  return;

	/* Translate DIS to simnet */
	if ((pdu->dis_result == DIS203_DETONATION_RESULT_DETONATION) ||
	    (pdu->dis_result == DIS203_DETONATION_RESULT_MICLIC)) 
	  treat_as_IF = TRUE;


	/* All mine detonations, and certain DIS detonations (see above)
	 * are displayed as IF.
	 */
	if (IS_MINE(pdu->burst.projectile) ||
	    treat_as_IF)
	{ 
	    if (!tactmap_map_to_screen(pvd_gui->tactmap,
				       pdu->detonations[0].location[X],
				       pdu->detonations[0].location[Y],
				       (int32)GCS_ILLEGAL_CELL,
				       &sto[X], &sto[Y]))
	      return;

	    /* $$$ Why not just call this now? */
	    sched_simtime_fncl(draw_if, time_last_simulation_clock, 0,
			       A_PTR, pvd_gui,
			       A_INT, sto[X], 
			       A_INT, sto[Y],
			       A_INT, ((pdu->dis_result == 
					DIS203_DETONATION_RESULT_MICLIC) ?
				       -1 :
				       pdu->burst.projectile),
			       A_END);
	    sched_simtime_fncl(draw_if,
			       time_last_simulation_clock+ERASE_DELAY, 0,
			       A_PTR, pvd_gui,
			       A_INT, sto[X], 
			       A_INT, sto[Y], 
			       A_INT, 0,
			       A_END);
	    return;
	}

	if (!pdu->firer_id || !vtab_get_vehicle(pdu->firer_id))
	  return;
	    
	if (pdu->result == nonImpact)
	  return;

	ent_set_context(ent_get_cell(pdu->firer_id));
	ent_get_position(pdu->firer_id, fire);
	fire[CELL3D]=ent_get_cell(pdu->firer_id);

        /*  Ignore Stingray impacts.  Stingray shots will be displayed
         *  based on the Stingray Impact Response PDU.
         */
        if (pdu->burst.projectile == munition_EO_1 ||
            pdu->burst.projectile == munition_EO_2 ||
            pdu->burst.projectile == munition_EO_3 ||
            pdu->burst.projectile == munition_EO_4)
          return;

        /* Note that we are NOT using shortcut AND (&&) because we
         * want both expressions evaluated before returning.
         */
        if (!tactmap_map_to_screen(pvd_gui->tactmap, 
                                   fire[X], fire[Y], (int32)fire[CELL3D],
                                   &sfrom[X], &sfrom[Y]) &
            !tactmap_map_to_screen(pvd_gui->tactmap,
  	                         pdu->detonations[0].location[X],
                                   pdu->detonations[0].location[Y],
                                   (int32)GCS_ILLEGAL_CELL,
                                   &sto[X], &sto[Y]))
           return;

	    
	/* Cobble the locations into a int, so I can do the deferred_fncl
	 * with 4 arguments.
	 */
	from = sfrom[X] << 16 | (sfrom[Y] & 0xffff);
	to = sto[X] << 16 | (sto[Y] & 0xffff);

	draw_df(pvd_gui, from, to, pdu->result);

	sched_simtime_fncl(draw_df,
			   time_last_simulation_clock+ERASE_DELAY, 0,
			   A_PTR, pvd_gui,
			   A_INT, from, A_INT, to, A_INT, 0, A_END);
    }
    else if (format == pvd_srimpact_fmt)
    {
        if (!pdu->firer_id || !vtab_get_vehicle(pdu->firer_id))
          return;

        ent_set_context(ent_get_cell(pdu->firer_id));
        ent_get_position(pdu->firer_id, fire);
        fire[CELL3D]=ent_get_cell(pdu->firer_id);

        ent_set_context(ent_get_cell(pdu->target_id));
        ent_get_position(pdu->target_id, targ);
        targ[CELL3D]=ent_get_cell(pdu->target_id);

        /* Note that we are NOT using shortcut AND (&&) because we
         * want both expressions evaluated before returning.
         */
        if (!tactmap_map_to_screen(pvd_gui->tactmap,
                                   fire[X], fire[Y], (int32)fire[CELL3D],
                                   &sfrom[X], &sfrom[Y]) &
            !tactmap_map_to_screen(pvd_gui->tactmap,
                                   targ[X], targ[Y], (int32)targ[CELL3D],
                                   &sto[X], &sto[Y]))
           return;


        /* Cobble the locations into a int, so I can do the deferred_fncl
         * with 4 arguments.
         */
        from = sfrom[X] << 16 | (sfrom[Y] & 0xffff);
        to = sto[X] << 16 | (sto[Y] & 0xffff);

        if(pdu->result)
          draw_df(pvd_gui, from, to, STINGRAY_DAMAGE);

        sched_simtime_fncl(draw_df,
                           time_last_simulation_clock+ERASE_DELAY, 0,
                           A_PTR, pvd_gui,
                           A_INT, from, A_INT, to, A_INT, 0, A_END);

    }
    else if (format == pvd_collision_fmt)
    {
	if (!pvd_gui->draw_collisions)
	  return;

	/* Get a position to draw it (either vehicle will do) */
	/* $$$ ?? */
	if (!(vehicle_id = pdu->firer_id) && !(vehicle_id = pdu->target_id))
	  return;
	ent_set_context(ent_get_cell(vehicle_id));
	ent_get_position(vehicle_id, fire);
	fire[CELL3D]=ent_get_cell(vehicle_id);

	if (!tactmap_map_to_screen(pvd_gui->tactmap, 
				   fire[X], fire[Y], (int32)fire[CELL3D],
				   &sto[X], &sto[Y]))
	  return;

	draw_coll(pvd_gui, sto[X], sto[Y], (int32)TRUE);

	sched_simtime_fncl(draw_coll,
			   time_last_simulation_clock+ERASE_DELAY, 0,
			   A_PTR, pvd_gui, A_INT, sto[X], A_INT, sto[Y], 
			   A_INT, FALSE,
			   A_END);
    }
    else if (format == pvd_indirect_fmt)
    {
	if (!pvd_gui->draw_impacts)
	  return;

	when = time_last_simulation_clock;

	for (i=0;i<pdu->burst.quantity;i++)
	{
	    when += pdu->detonations[i].delay;

	    if (!tactmap_map_to_screen(pvd_gui->tactmap,
				       pdu->detonations[i].location[X],
				       pdu->detonations[i].location[Y],
				       (int32)GCS_ILLEGAL_CELL,
				       &sfrom[X], &sfrom[Y]))
	      continue;

	    sched_simtime_fncl(draw_if, when, 0,
			       A_PTR, pvd_gui,
			       A_INT, sfrom[X], A_INT, sfrom[Y],
			       A_INT,
			       pdu->burst.projectile,
			       A_END);

	    sched_simtime_fncl(draw_if, when+ERASE_DELAY, 0,
			       A_PTR, pvd_gui,
			       A_INT, sfrom[X], A_INT, sfrom[Y], A_INT, FALSE,
			       A_END);
	}
    }
    else if (format == pvd_fire_fmt)
    {
	if (!pvd_gui->draw_fires)
	  return;

	if (!tactmap_map_to_screen(pvd_gui->tactmap,
				   pdu->firing_location[X],
				   pdu->firing_location[Y],
                                   (int32)GCS_ILLEGAL_CELL,
				   &sfrom[X],&sfrom[Y]))
	  return;

	/* Cobble the locations into a int, so I can do the deferred_fncl
	 * with 4 arguments.
	 */

	from = sfrom[X] << 16 | (sfrom[Y] & 0xffff);
	angle = NS_ATAN2(pdu->munition_velocity[Y], pdu->munition_velocity[X]);

	/* Draw now*/
	draw_fire(pvd_gui, from, angle, TRUE);

	/* Erase later */
	sched_simtime_fncl(draw_fire,
			   time_last_simulation_clock+FIRE_ERASE_DELAY, 0,
			   A_PTR, pvd_gui,
			   A_INT, from,
			   A_DOUBLE, angle,
			   A_INT, FALSE,
			   A_END);
    }
    else if (format == pvd_srdetect_fmt)
    {
	ent_set_context(ent_get_cell(pdu->firer_id));
	ent_get_position(pdu->firer_id, fire);
	fire[CELL3D]=ent_get_cell(pdu->firer_id);
	ent_set_context(ent_get_cell(pdu->target_id));
	ent_get_position(pdu->target_id, targ);
	targ[CELL3D]=ent_get_cell(pdu->target_id);

	/* Note that we are NOT using shortcut AND (&&) because we
	 * want both expressions evaluated before returning.
	 */
	if (!tactmap_map_to_screen(pvd_gui->tactmap, 
				   fire[X], fire[Y], (int32)fire[CELL3D],
				   &sfrom[X], &sfrom[Y]) &
	    !tactmap_map_to_screen(pvd_gui->tactmap,
				   targ[X], targ[Y], (int32)targ[CELL3D],
				   &sto[X], &sto[Y]))
	  return;

	/* Cobble the locations into a int, so I can do the deferred_fncl
	 * with 4 arguments.
	 */
	from = sfrom[X] << 16 | (sfrom[Y] & 0xffff);
	to = sto[X] << 16 | (sto[Y] & 0xffff);

	if(pdu->result)
	  draw_df(pvd_gui, from, to, STINGRAY_DETECT);
	else
	  draw_df(pvd_gui, from, to, STINGRAY_NO_DETECT);

	sched_simtime_fncl(draw_df,
			   time_last_simulation_clock+ERASE_DELAY, 0,
			   A_PTR, pvd_gui,
			   A_INT, from, A_INT, to, A_INT, 0, A_END);
    }
}

void pvd_process_radio(PV_PACKET *packet, int32 format, PVD_SIGNAL *pdu,
		       PVD_GUI_PTR pvd_gui)
{
    static char mbuf[4096]; /* atleast > than radio message + formatting */
    int32 aspid=0;
    char *text;

#define PVD_ENCODING_SCHEME_BIT_MASK 0xc000
#define PVD_ENCODING_SCHEME_APPLICATION_SPECIFIC_DATA 0x8000
    /* Do a check to see if the encoding_scheme that is used is the
     * application specific data.  If not, then currently we want to
     * throw this PDU out.  
     */
    if ((pa_present(format, packet->data, "encoding_scheme") == PA_PRESENT) &&
	((pdu->encoding_scheme & PVD_ENCODING_SCHEME_BIT_MASK) !=
	 PVD_ENCODING_SCHEME_APPLICATION_SPECIFIC_DATA))
      return;
    
    bcopy(pdu->data, (char *)&aspid, sizeof(int32));
    text = pdu->data + 4;

    /* Filter on the force ID of transmitter.  If we aren't displaying 
     * the vehicles for that force (as indicated by the corresponding
     * element of the alignments_displayed array maintained by
     * libpvd) and we are not commanding, then we won't display 
     * their radio traffic either.
     */
    if (!pvd_okay_to_display(pvd_gui, ent_get_force_id(pdu->hashed_id),
			     pdu->hashed_id, TRUE))
      return;

    /* Check the application-specific "user protocol
     * identification number."  This distinguishes between
     * applications in a given exercise.
     */
    /* $$$ This ought to be configurable */
    if ((aspid != ASPID_MODSAF_TEXT) && (aspid != ASPID_MODSAF_ARTILLERY))
      return;

    /* Suppress vehicle-to-vehicle messages */
    if (text[0] == '@')
      return;
    
    edt_format (pvd_gui->gui, mbuf, "#v: #S", &pdu->unhashed_id, text);
    sgui_add_message (pvd_gui->gui, "radio", mbuf);
}


/* Perhaps this laser code should be drawn using tactmap lines
 * rather than straight X calls.
 */

/* When a new laser PDU comes in we need to check whether
 * something with the same code from the same lasing entity
 * already exists.
 */
struct laser_search_cache
{
    PVD_LASER *key;             /* The new PDU */
    PVD_LASER *matching_entry;  /* The matching existing entry,
				   if any */
};
static void search_for_lasing_entry(
    PVD_LASER                 *entry,
    struct laser_search_cache *cache)
{
    if (cache->key->lasing_id == entry->lasing_id)
      cache->matching_entry = entry;
}

/* Cancel the  */
static void update_spot(
    PVD_LASER   *pdu,
    PVD_GUI_PTR  pvd_gui)
{
    int32 sfrom[2], sto[2];

    /* If the laser has timed out or drawing of laser designations
     * has been cancelled, then eliminate this entry.
     * Otherwise draw it.
     */
    if ((pdu->time_last_pdu + LASER_SPOT_TIMEOUT > time_last_simulation_clock) &&
	pvd_gui->draw_laser_designations)
    {
	if (tactmap_map_to_screen(pvd_gui->tactmap, 
				  pdu->lasing_location[X],
				  pdu->lasing_location[Y],
				  (int32) pdu->lasing_location[CELL3D],
				  &sfrom[X], &sfrom[Y]) |
	    tactmap_map_to_screen(pvd_gui->tactmap,
				  pdu->lased_location[X],
				  pdu->lased_location[Y],
				  (int32) pdu->lased_location[CELL3D],
				  &sto[X], &sto[Y]))
	  draw_laser(pvd_gui, FALSE, sfrom, sto);
    }
    else
    {
	/* Erase the line, maybe? */
	if (tactmap_map_to_screen(pvd_gui->tactmap, 
				  pdu->lasing_location[X],
				  pdu->lasing_location[Y],
				  (int32) pdu->lasing_location[CELL3D],
				  &sfrom[X], &sfrom[Y]) |
	    tactmap_map_to_screen(pvd_gui->tactmap,
				  pdu->lased_location[X],
				  pdu->lased_location[Y],
				  (int32) pdu->lased_location[CELL3D],
				  &sto[X], &sto[Y]))
	  draw_laser(pvd_gui, TRUE, sfrom, sto);
	sched_cancel_fncl(pdu->erase_fcn);
	queue_excise(laser_spots[pdu->laser_code], pdu);
	queue_buffer_deallocate(pdu);
    }
}

void pvd_process_laser(
    PV_PACKET  *packet, 
    int32       format,
    PVD_LASER   *pdu,
    PVD_GUI_PTR  pvd_gui)
{
    struct laser_search_cache cache;
    int32 sfrom[2], sto[2];
    PVD_LASER *buf;

    if (!pvd_gui->draw_laser_designations)
      return;

    pdu->time_last_pdu = time_last_simulation_clock;
    ent_get_position(pdu->lasing_id, pdu->lasing_location);
    pdu->lasing_location[CELL3D] = ent_get_cell(pdu->lasing_id);
    if (pdu->lased_id)
      pdu->lased_location[CELL3D] = ent_get_cell(pdu->lased_id);
    else
      pdu->lased_location[CELL3D] = GCS_ILLEGAL_CELL;

    /* See if it already exists in the list */
    cache.key = pdu;
    cache.matching_entry = NULL;
    if (laser_spots[pdu->laser_code])
    {
	queue_traverse_and_apply(laser_spots[pdu->laser_code],
				 (QUEUE_FUNCTION)search_for_lasing_entry,
				 &cache);
    }
    else
      laser_spots[pdu->laser_code] = queue_allocate(0);

    /* Update the list entry, creating one if necessary */
    if (cache.matching_entry == NULL)
    {
	/* Its a new spot */
	buf = (PVD_LASER*) queue_buffer_allocate(sizeof(PVD_LASER));
	bcopy(pdu, buf, sizeof(PVD_LASER));
	buf->erase_fcn =
	  sched_periodic_fncl(update_spot, SCHED_ASAP, LASER_SPOT_TIMEOUT, 0,
			      A_PTR, buf, A_PTR, pvd_gui, A_END);
#if 0
	/* These next few lines belong in the init routine,
	 * the fields are constant for all lines.
	 */
	dfn->linear.style = LSplain;
	dfn->linear.dashed = FALSE;
	dfn->linear.thickness = 1;
	dfn->linear.backward_arrow = FALSE;
	dfn->linear.forward_arrow = FALSE;
	dfn->linear.width = 1;
	dfn->linear.n_vertices = 2;
	dfn->linear_gc = pvd_gui->laser_gc;
	/* What should the class be? These lines should _not_
	 * be sensitive to user input.
	 */
	dfn->sensitive.class = NULL;

	/* These should stay here */
	dfn->linear.verticies[0] = pdu->lasing_location[X];
	dfn->linear.verticies[1] = pdu->lasing_location[Y];
	dfn->linear.verticies[2] = pdu->lased_location[X];
	dfn->linear.verticies[3] = pdu->lased_location[Y];
	/* $$$ What should the instance be if it is not sensitive? */
	/* $$$ When activating this code, up above add a
	 * static int32 uniqid = 1;
	 */
	dfn->sensitive.instance = uniqid++;
	buf->tmap_obj = tactmap_create_object(pvd_gui->tactmap,
				      TACTMAP_LINEAR, dfn);
	tactmap_add_object(tmap, new[j].tactmap_obj);
#endif
					      
	queue_enqueue(laser_spots[pdu->laser_code], buf);
    }
    else
    {
	/* Old spot. */
	/* $$$ Maybe erase old entry & redraw? */
	pdu->erase_fcn = cache.matching_entry->erase_fcn;
	bcopy(pdu, cache.matching_entry, sizeof(PVD_LASER));
    }
}


static void draw_df(PVD_GUI_PTR pvd_gui, uint32 from, uint32 to,
		    int32 result)
{
    int32 from_x, from_y, to_x, to_y;
    GC gc;

    from_x = from >> 16;
    from_y = from & 0xFFFF;
    to_x = to >> 16;
    to_y = to & 0xFFFF;

    if (from_x & 0x8000) from_x |= 0xFFFF0000;
    if (from_y & 0x8000) from_y |= 0xFFFF0000;
    if (to_x & 0x8000) to_x |= 0xFFFF0000;
    if (to_y & 0x8000) to_y |= 0xFFFF0000;

    switch(result)
    {
      case 0:
	gc = pvd_gui->map_erase_gc;
	break;
      case vehicleImpact:
      case proximateImpact:
      case vehicleImpactDirectedDamage:
	gc = pvd_gui->hit_gc;
	break;
      case STINGRAY_DAMAGE:
	gc = pvd_gui->coll_gc;
        break;
/* The NO_DAMAGE case is currently unused. */
      case STINGRAY_NO_DAMAGE:
        gc = pvd_gui->miss_gc;
        break;
      case STINGRAY_DETECT:
	gc = pvd_gui->mine_gc;
	break;
      case STINGRAY_NO_DETECT:
      default:
	gc = pvd_gui->miss_gc;
	break;
    }

    XDrawLine(XtDisplay(pvd_gui->map_widget), XtWindow(pvd_gui->map_widget),
	      gc, from_x, from_y, to_x, to_y);

    if(result != STINGRAY_DETECT && result != STINGRAY_NO_DETECT)
      XFillArc(XtDisplay(pvd_gui->map_widget), XtWindow(pvd_gui->map_widget),
	       gc, to_x-CIRCLE_RAD, to_y-CIRCLE_RAD, 2*CIRCLE_RAD,2*CIRCLE_RAD,
	       0, 360*64);
}

static void draw_fire(PVD_GUI_PTR pvd_gui, uint32 from, float64 angle,
		      int32 draw_it)
{
    int32 from_x, from_y, to_x, to_y;
    GC gc;
    int32 i;
    short segs[12];
    float64 sin_angle;
    float64 cos_angle;

#define FIRE_LENGTH 20
#define FIRE_ANGLE (PI/8.0)

    from_x = from >> 16;
    from_y = from & 0xFFFF;

    if (from_x & 0x8000) from_x |= 0xFFFF0000;
    if (from_y & 0x8000) from_y |= 0xFFFF0000;

    if (draw_it)
      gc = pvd_gui->hit_gc;
    else
      gc = pvd_gui->map_erase_gc;

    /* Remember that screen coordinates have Y going DOWN */

    /* Segment 1 is straight */
    segs[0] = from_x;
    segs[1] = from_y;
    segs[2] = from_x + (FIRE_LENGTH * cos(angle));
    segs[3] = from_y - (FIRE_LENGTH * sin(angle));

    /* Segment 2 is to the left, (facing north) */
    angle -= FIRE_ANGLE;
    segs[4] = from_x;
    segs[5] = from_y;
    segs[6] = from_x + (FIRE_LENGTH * cos(angle));
    segs[7] = from_y - (FIRE_LENGTH * sin(angle));

    /* Segment 3 is to the right, (facing north) */
    angle += FIRE_ANGLE;
    angle += FIRE_ANGLE;
    segs[8] = from_x;
    segs[9] = from_y;
    segs[10] = from_x + (FIRE_LENGTH * cos(angle));
    segs[11] = from_y - (FIRE_LENGTH * sin(angle));
    

    XDrawSegments(XtDisplay(pvd_gui->map_widget),
		  XtWindow(pvd_gui->map_widget),
		  gc,
		  (XSegment *)segs, 3);
}

static void draw_if(PVD_GUI_PTR pvd_gui, int32 from_x, int32 from_y, int32 mun)
{
    GC gc;
    int32 i;
    short segs[XtNumber(burst_segs)];
    ObjectType munition = (ObjectType)mun;

    if (!munition)
      gc = pvd_gui->map_erase_gc;
    else if (IS_MINE(munition))
      gc = pvd_gui->mine_gc;
    else if (mun == -1)
      gc = pvd_gui->miclic_gc;
    else 
      gc = pvd_gui->indf_gc;

    for (i=0;i<XtNumber(burst_segs);i+=2)
    {
	segs[i] = from_x + burst_segs[i];
	segs[i+1] = from_y + burst_segs[i+1];
    }
    XDrawSegments(XtDisplay(pvd_gui->map_widget),
		  XtWindow(pvd_gui->map_widget),
		  gc, (XSegment *)segs, i/4);
}

static void draw_coll(PVD_GUI_PTR pvd_gui, int32 x, int32 y, int32 draw_p)
{
    GC gc;
    int32 i;
    short segs[XtNumber(coll_segs)];

    if (!draw_p)
      gc = pvd_gui->map_erase_gc;
    else
      gc = pvd_gui->coll_gc;
	
    for (i=0;i<XtNumber(coll_segs);i+=2)
    {
	segs[i] = x + coll_segs[i];
	segs[i+1] = y + coll_segs[i+1];
    }
    XFillPolygon(XtDisplay(pvd_gui->map_widget),
		 XtWindow(pvd_gui->map_widget),
		 gc, (XPoint *)segs, i/2, Nonconvex, CoordModeOrigin);
}

static void draw_laser(
    PVD_GUI_PTR pvd_gui,
    int32       erase,
    int32       from[XY],
    int32       to[XY])
{
    GC gc;

    if (erase)
      gc = pvd_gui->map_erase_gc;
    else
      gc = pvd_gui->laser_gc;

    XDrawLine(XtDisplay(pvd_gui->map_widget), XtWindow(pvd_gui->map_widget),
	      gc, from[X], from[Y], to[X], to[Y]);

#if 0
    XDrawSegments(XtDisplay(pvd_gui->map_widget),
		  XtWindow(pvd_gui->map_widget),
		  gc, (XSegment *)segs, 1);
#endif
}


#endif
/*USE_MOTIF*/
