#ifndef lint
static char rcsid [] = "$RCSfile: pvd_info.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_info.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle pvd info mode                              *
*       Created: Tue Sep 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <libctdb.h>
#include <math.h>
#include <libgcs.h>

static void info_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
			 SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui);

/*ARGSUSED*/
void pvd_info_mode(SGUI_PTR gui, SGUI_MODE_PTR mode, PVD_GUI_PTR pvd_gui,
		   SGUI_MODE_STATE state)
{
    if ((state == SGUI_ACTIVE) ||
	(state == SGUI_RESUMED))
    {
	/* Install the mouse handler */
	callback_register_handler(pvd_gui->map_event,info_gesture,pvd_gui);
	pvd_gui->current = pvd_gui->info;
    }
    else
    {
	/* Remove the mouse handler */
	callback_unregister_handler(pvd_gui->map_event,info_gesture,pvd_gui);
    }
}

/*ARGSUSED*/
static void info_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
			 SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui)
{
    float64 map_x, map_y, map_z;
    int32   map_c;
    float64 rot[3][3];
    int32 soil, grade;
    char buf[256];

    /* Ignore button 1, and EXCITED/UNEXCITED callbacks */
    if (call_data->button < 2)
      return;

    switch(action)
    {
      case SNSTVE_MOTION:
	sgui_pop_help(pvd_gui->info, TRUE);
	/* Fall through */
      case SNSTVE_PRESS:
	tactmap_screen_to_map(pvd_gui->tactmap,
				call_data->screen_x, call_data->screen_y,
				&map_x,  &map_y, &map_c);
	ctdb_place_vehicle((CTDB *)gcs_get_tdb(map_c, FALSE), map_x, map_y, 
			   1.0, 1.0, 0.0, 1.0,&map_z, rot, &soil);

	soil = CTDB_SIMNET_SOIL(soil);

	if ((rot[Z][Z] >= 1.0) || (rot[Z][Z] <= -1.0))
	  grade = 0;
	else
	  grade = 90 - RAD_TO_DEG(asin(rot[Z][Z]));

#if 1
	/* $$$ What is altitude? Is it the elevation normal to the cell or
	 * the length of the z vector along gravity
	 */
	edt_format(pvd_gui->gui, buf,
		   "Grade: #a   |   Soil type: #D   |   Center: #l   |   Altitude: #h",
		   DEG_TO_RAD(grade), soil, map_c, map_x, map_y, map_z);
#else
	/* We have decided (for now) that this display is too cluttered
	 * to be useful.  Instead we just show the location in the user's
	 * favorite system.
	 */

	sprintf(buf, "Grade: %d degrees   |   Soil type: %d   |   GCS: ",
		grade, soil);

	coord_convert(COORD_GCS,map_c, map_x, map_y, map_z,
		      COORD_CHARACTER_STRING,hold);
	strcat(buf, hold);

	strcat(buf, "   |   GCC: ");
	coord_convert(COORD_GCS, map_c, map_x, map_y, map_z,
		      COORD_GCC,&gcc_x, &gcc_y, &gcc_z);
	coord_convert(COORD_GCC,gcc_x, gcc_y, gcc_z,
		      COORD_CHARACTER_STRING,hold);
	strcat(buf, hold);

	strcat(buf, "   |  GEO: ");
	coord_convert(COORD_GCS,map_c, map_x, map_y, map_z,
		      COORD_LATLON,&lat, &lon, &ll_z, TRUE);
	coord_convert(COORD_LATLON,lat, lon, ll_z, TRUE,
		      COORD_CHARACTER_STRING,hold);
	strcat(buf, hold);

	strcat(buf, "   |  UTM: ");
	coord_convert(COORD_GCS,map_c, map_x, map_y, map_z,
		      COORD_UTM_GRID,COORD_DEFAULT_ZONE, 5, hold, &utm_z);
	strcat(buf, hold);
#endif
	sgui_push_help(pvd_gui->info, buf);
	break;
      case SNSTVE_RELEASE:
	sgui_pop_help(pvd_gui->info, TRUE);
	break;
    }
}

#endif
/*USE_MOTIF*/
