#ifndef lint
static char rcsid [] = "$RCSfile: pvd_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_init.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libpvd
*       Created: Mon Sep 28 08:10:21 PDT 1992
*       Author: jesmith
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <libenvui.h>
#include <libpduproc.h>
#include <libotmatch.h>
#include <libsafgui.h>
#include <libtactmap.h>
#include <libsensitive.h>
#include <libcallback.h>
#include <libvtab.h>
#include <libprivilege.h>
#include <libreader.h>
#include <libcoordinates.h>
#include <libselect.h>
#include <libworld.h>
#include "pvd_zoom.h"
#include "pvd_pan.h"
#include "pvd_info.h"
#include "pvd_icon.h"
#include "pvd_env.h"
#include <p_aspid.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/
#include <stdstring.h> /*common/include/global*/

#include <Xm/ToggleB.h>

/* This constant is the reciprocal of the scale at which the largest
 * desired vehicle size happens to be magnification == 1.
 */
#define K 0.002


/* The features are represented in the following two structures.
 * The first is used to make the legend.  It drives the creation
 * of features.  The second is the PVD editor which must match
 * the definition in the file pvdedit.rdr.
 */

static struct
{
    /* NOTE: libpvd_local.h says that TREES, CONTOURS, GRIDS, HYPSO, WIRE,
     * DRAGONS_TEETH, MICLIC_LINE, and MICLIC_SCORCH must be the first 
     * features listed.
     */

    TACTMAP_OBJECT_CLASS class;
    char *label;
    int32 sensitive;
} features[] = { { TACTMAP_TREES, "Trees and Canopies", TRUE } ,
		 { TACTMAP_CONTOURS, "Contour lines", FALSE } ,
		 { TACTMAP_GRIDS, "Grid lines", FALSE } ,
		 { TACTMAP_HYPSO, "Hypsometric", FALSE } ,
		 { TACTMAP_WIRE, "Wire", TRUE },
		 { TACTMAP_DRAGONS_TEETH, "Dragon's Teeth", TRUE },
		 { TACTMAP_MICLIC_LINE, "Miclic Line", TRUE },
		 { TACTMAP_MICLIC_SCORCH, "Miclic Scorch", TRUE },
		 { TACTMAP_CELLS, "GCS Cell Boundries", FALSE } ,
		 { TACTMAP_WATER, "Water and Soils", TRUE } ,
		 { TACTMAP_ROADS, "Roads", TRUE } ,
		 { TACTMAP_CONTAMINANT, "Contaminants", TRUE } ,
		 { TACTMAP_BUILDINGS, "Buildings", TRUE } ,
#ifdef AE_FEATURES
		 { TACTMAP_TACT_SIGNS, "TactSigns", TRUE } , 
#endif
		 { TACTMAP_PIPELINES, "Pipelines", TRUE } ,
		 { TACTMAP_POLITICAL, "Political Boundaries", TRUE } ,
		 { TACTMAP_RAILROADS, "Railroads", TRUE } ,
#ifdef AE_FEATURES
		 { TACTMAP_OFFROADS, "OffRoads", TRUE } ,
#endif
		 { TACTMAP_POWERLINES, "Powerlines", TRUE } ,
		 { TACTMAP_TOWNS, "Towns", TRUE } ,
		 { TACTMAP_CRATERS, "Craters", TRUE } ,
		 { TACTMAP_DITCHES, "Ditches", TRUE } ,
		 { TACTMAP_FPOS, "Fighting Positions", TRUE },
		 { TACTMAP_ROCK_DROPS, "Rock Drops", TRUE },
		 { TACTMAP_BERMS, "Berms", TRUE },
		 { TACTMAP_RUBBLE, "Rubble", TRUE },
		 { TACTMAP_SNOW, "Snow", TRUE },
		 { TACTMAP_CLOUDS, "Clouds", FALSE },
		 { TACTMAP_CAMOUFLAGE, "Camouflage", TRUE },
		 { TACTMAP_HYDRO_SURFACE, "Surface Wetness", TRUE },
		 { TACTMAP_HYDRO_SUBSURFACE, "Subsurface Saturation", TRUE }
	     };

struct pvd_editor
{
    float32 vehicle_scale;
    int32   feature_count;
    int32   features[32];
    int32   hypso_kind;
    uint32  update_rate;
    int32   num_markings;
    int8    markings[4];
    uint32  use_these_features;
    float32 icon_scale;
    int32 em_band;
    int32 fixed_minor_contour;
    int32 fixed_major_contour;
    int32 smoke_display;
    int32 mf_sym_cutoff;
    int32   num_events;
    int32   events[8];
};

int32 PVD_ASPDU;
int32 pvd_modstealth;
int32 pvd_db_key = -1;
int32 pvd_n_pvds = 0;
PVD_GUI_PTR pvd_pvds[PVD_MAX_PVDS];
static READER_UNION default_features;
static READER_UNION editor_dfn;
READER_UNION pvd_stealth_picture;
READER_UNION pvd_smoke_picture;
READER_UNION pvd_parts;

char *pvd_plow_cat = NULL, *pvd_bridge_cat = NULL, *pvd_dogbone_cat = NULL;
char *pvd_grizzlyshovel_cat = NULL, *pvd_dirtmover_cat = NULL;
char *pvd_shovel_base = NULL;

static READER_UNION stealth_editor_dfn;

/* private prototypes */
static void make_legend(PVD_GUI_PTR pvd_gui);
static void make_scale(PVD_GUI_PTR pvd_gui);
static void make_show_as(PVD_GUI_PTR pvd_gui);
static void apply_editor(EDT_EDITOR_PTR editor, int32 transient,
			 struct pvd_editor *old, struct pvd_editor *new,
			 PVD_GUI_PTR pvd_gui);
static void done_editor(EDT_EDITOR_PTR editor, struct pvd_editor *new,
			PVD_GUI_PTR pvd_gui, EDT_EXIT_STATUS status);
static void load_editor(EDT_EDITOR_PTR editor, PVD_GUI_PTR pvd_gui);
static void apply_editor_data(PVD_GUI_PTR pvd_gui, struct pvd_editor *data);
static void new_defaults(SGUI_PTR gui, PVD_GUI_PTR pvd_gui);

/*ARGSUSED*/
static void pvd_edit_features(SGUI_PTR gui, SGUI_MODE_PTR mode,
			      PVD_GUI_PTR pvd_gui, SGUI_MODE_STATE state)
{
    edt_state(pvd_gui->editor, mode, state);
    load_editor(pvd_gui->editor, pvd_gui);
}

/*ARGSUSED*/
static void pvd_edit_stealth(SGUI_PTR gui, SGUI_MODE_PTR mode,
			     PVD_GUI_PTR pvd_gui, SGUI_MODE_STATE state)
{
    edt_state(pvd_gui->stealth_editor, mode, state);
}

int32 pvd_init(char *data_path, uint32 flags, int32 modstealth, int32 aspdu)
{
    READER_UNION pictures;
    READER_UNION stealth;
    READER_UNION smoke;
    
    int32 ret0 = reader_read("pvd.rdr", data_path, &pictures,
			     flags & ~READER_TYPING);
    int32 ret1 = reader_read("features.rdr", data_path, &default_features,
			     flags & ~READER_TYPING);
    int32 ret2 = reader_read("pvdedit.rdr", data_path, &editor_dfn,
			     flags & ~READER_TYPING);
    int32 ret3 = reader_read("stealth.rdr", data_path, &stealth,
			     flags & ~READER_TYPING);
    int32 ret4 = reader_read("smoke.rdr", data_path, &smoke,
			     flags & ~READER_TYPING);
    int32 ret5 = reader_read("pvd_parts.rdr", data_path, &pvd_parts,
			     flags & ~READER_TYPING);

    PVD_ASPDU = FALSE;
    if (aspdu)
      PVD_ASPDU = TRUE;

    if (ret0)
      return ret0;
    if (ret1)
      return ret1;
    if (ret2)
      return ret2;
    if (ret3)
      return ret3;
    if (ret4)
      return ret4;
    if (ret5)
      return ret5;

    pvd_db_key = otm_incorporate_db((OTM_DB *)(pictures.array), TRUE, NULL);
    stealth_editor_dfn.array = stealth.array[1].array;
    pvd_stealth_picture.array = stealth.array[2].array;
    pvd_init_subclass();
    
    pvd_modstealth = modstealth;

    pvd_plow_cat = reader_get_symbol("plow");
    pvd_bridge_cat = reader_get_symbol("bridge");
    pvd_dogbone_cat = reader_get_symbol("dogbone");
    pvd_grizzlyshovel_cat = reader_get_symbol("grizzlyshovel");
    pvd_dirtmover_cat = reader_get_symbol("dirtmover");
    pvd_shovel_base = reader_get_symbol("shovel-base");

    return 0;
}

PVD_GUI_PTR pvd_init_gui(SGUI_PTR gui, TACTMAP_PTR tactmap,
			 Widget map_widget, GC map_erase_gc,
			 SNSTVE_WINDOW_PTR sensitive,
			 SELECT_TOOL_PTR select,
			 CALLBACK_EVENT_PTR refresh_event,
			 VTAB_LIST viewable,
			 SimulationAddress *preferred_stealth)
{
    PVD_GUI_PTR result;
    XGCValues gcv;
    char name_sans_version[256];
#ifdef NEVER
    char *tok;
#endif
    int32 version;
    Window root;
    CALLBACK_EVENT_PTR ev;
    int32 i,j;
    char *feat;
    Widget hsb, vsb;
    struct pvd_editor edt_defaults;
    struct pvd_gui_resources
    {
	Pixel zoom;
	Pixel black;
	Pixel white;
	Pixel unkn;
	Pixel dist;
	Pixel othr;
	Pixel obsr;
	Pixel targ;
	Pixel hit;
	Pixel miss;
	Pixel indf;
	Pixel mine;
	Pixel coll;
	Font  designations;
	Pixel grey[10];
	Pixel mf;
	Pixel opmf;
	Pixel lane;
	Pixel brown[9];
	Pixel miclic;
	Pixel laser;
    } resources;
    static XtResource res_list[] = {
	{ "zoom" , "Zoom" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, zoom) ,
	    XmRString , "Yellow" } ,
	{ "black" , "Black" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, black) ,
	    XmRString , "Black" } ,
	{ "white" , "White" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, white) ,
	    XmRString , "White" } ,
	{ "unknown" , "Unknown" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, unkn) ,
	    XmRString , "Gray50" } ,
	{ "distinguished", "Distinguished", XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, dist) ,
	    XmRString , "Blue" } ,
	{ "other" , "Other" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, othr) ,
	    XmRString , "Red" } ,
	{ "observer" , "Observer" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, obsr) ,
	    XmRString , "DeepPink" } ,
	{ "target" , "Target" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, targ) ,
	    XmRString , "Green3" } ,
	{ "hit" , "Hit" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, hit) ,
	    XmRString , "Yellow" } ,
	{ "miss" , "Miss" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, miss) ,
	    XmRString , "Gray50" } ,
	{ "indirectFire" , "indirectFire" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, indf) ,
	    XmRString , "Purple" } ,
	{ "mine" , "Mine" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, mine) ,
	    XmRString , "Orange" } ,
	{ "collision" , "Collision" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, coll) ,
	    XmRString , "Red" } ,
	{ "designations" , "Designations" , XmRFont , sizeof(Font) ,
	    XtOffset(struct pvd_gui_resources *, designations) ,
	    XmRString , "Fixed" } ,
	{ "grey1" , "Grey1" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[0]) ,
	    XmRString , "Grey30" } ,
	{ "grey2" , "Grey2" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[1]) ,
	    XmRString , "Grey33" } ,
	{ "grey3" , "Grey3" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[2]) ,
	    XmRString , "Grey37" } ,
	{ "grey4" , "Grey4" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[3]) ,
	    XmRString , "Grey42" } ,
	{ "grey5" , "Grey5" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[4]) ,
	    XmRString , "Grey48" } ,
	{ "grey6" , "Grey6" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[5]),
	    XmRString, "Grey55"},
	{ "grey7" , "Grey7" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[6]) ,
	    XmRString , "Grey63" } ,
	{ "grey8" , "Grey8" , XmRPixel , sizeof(Pixel) ,
	    XtOffset(struct pvd_gui_resources *, grey[7]),
	    XmRString , "Grey70" } ,
	{ "grey9", "Grey9", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, grey[8]),
	    XmRString, "Grey77" },
	{ "grey10", "Grey10", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, grey[9]),
	    XmRString, "Grey81"},
	{ "minefield", "Minefield", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, mf),
	    XmRString, "DarkGreen"},
	{ "opforminefield", "OpForMinefield", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, opmf),
	    XmRString, "FireBrick"},
	{ "minefieldlane", "MinefieldLane", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, lane),
	    XmRString, "Black"},
	{ "brown0", "Brown0", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[0]),
	    XmRString, "#f7de6f"},
	{ "brown1", "Brown1", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[1]),
	    XmRString, "#ffd6a1"},
	{ "brown2", "Brown2", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[2]),
	    XmRString, "#e8c18f"},
	{ "brown3", "Brown3", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[3]),
	    XmRString, "#be9564"},
	{ "brown4", "Brown4", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[4]),
	    XmRString, "#916d3e"},
	{ "brown5", "Brown5", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[5]),
	    XmRString, "#845224"},
	{ "brown6", "Brown6", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[6]),
	    XmRString, "#6c4028"},
	{ "brown7", "Brown7", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[7]),
	    XmRString, "#59270c"},
	{ "brown8", "Brown8", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, brown[8]),
	    XmRString, "#381400"},
	{ "miclic", "Miclic", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, miclic),
	    XmRString, "Maroon"},
	{ "laser", "Laser", XmRPixel, sizeof(Pixel),
	    XtOffset(struct pvd_gui_resources *, laser),
	    XmRString, "Red"}
    };

    if (pvd_n_pvds == PVD_MAX_PVDS)
    {
	printf("LIBPVD ERROR: Only allowed %d PVDs\n",
	       PVD_MAX_PVDS);
	return 0;
    }

    /* Get the colors, and various other resources used */
    XtGetSubresources(map_widget, &resources,
		      "pvd", "PVD",
		      res_list, XtNumber(res_list), NULL, 0);

    result = (PVD_GUI_PTR)STDALLOC(sizeof(struct pvd_gui));
    bzero((char *)result, sizeof(struct pvd_gui));
	
    result->gui                = gui;
    result->priv               = sgui_get_priv(gui);
    result->tactmap            = tactmap;
    result->map_widget         = map_widget;
    result->map_erase_gc       = map_erase_gc;
    result->sensitive          = sensitive;
    result->map_event          = snstve_miss_callback(sensitive);
    result->viewable           = viewable;
    result->preferred_stealth  = *preferred_stealth;
    result->show_all_stealths  =
      !preferred_stealth->site && !preferred_stealth->host;
    
    /* Initialize the PVD to show all but blue (distinguished) and red 
     * (other) forces, which are set by the menubar.
     */
    priv_set_local_force(result->priv, forceIDIrrelevant, TRUE);
    priv_set_local_force(result->priv, distinguishedForceID, FALSE);
    priv_set_local_force(result->priv, otherForceID, FALSE);
    priv_set_local_force(result->priv, observerForceID, TRUE);
    priv_set_local_force(result->priv, targetForceID, TRUE);

    /* Register callbacks for scrollbars and initialize their position */
    sgui_map_scrollbars(result->gui, &hsb, &vsb);
    pvd_init_scrollbar(result, hsb);
    pvd_init_scrollbar(result, vsb);
    pvd_update_scrollbars(result);

    SNSTVE_INIT_CLASS(result->vehicle_class);

    callback_register_handler(result->vehicle_class.gesture, pvd_click,
			      result);

    result->vehicle_class.sensitive = 1;
    result->vehicle_class.buttons = 1 << 2;

    SNSTVE_INIT_CLASS(result->stealth_class);

    callback_register_handler(result->stealth_class.gesture, pvd_click,
			      result);

    result->stealth_class.sensitive = 1;
    result->stealth_class.buttons = 1 << 2;

    if (select && result->show_all_stealths && !pvd_modstealth)
      select_add_selectable(select, &result->stealth_class, PRIV_OPERATOR,
			    FALSE, pvd_stealth_selected, result);

    result->veh_update_rate     = 500; /* Default to twice per second */
    result->veh_magnification   = 1.0;
    result->veh_mag_dont_adjust = FALSE;
    result->raw_scale           = 0.0;
    result->markings            = 0;
    result->mf_sym_cutoff       = 50000;
    result->draw_collisions     = FALSE;
    result->draw_fires          = FALSE;
    result->draw_impacts        = TRUE;
    result->draw_for            = TRUE;
    result->draw_radio_traffic  = FALSE;
    result->draw_links          = TRUE;
    
    /* Tell libeditor about the vehicle class */
    edt_set_sensitive_class(gui, EDT_VEHICLE, &result->vehicle_class);
    edt_set_sensitive_class(gui, EDT_STEALTH, &result->stealth_class);

    /* Subscribe to incoming PDUs */
    pvd_init_pdu();

    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_impact_fmt,
		      (PDUPROC_CALLBACK)pvd_process_impact, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_indirect_fmt,
		      (PDUPROC_CALLBACK)pvd_process_impact, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_collision_fmt,
		      (PDUPROC_CALLBACK)pvd_process_impact, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_fire_fmt,
		      (PDUPROC_CALLBACK)pvd_process_impact, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_radio_fmt,
		      (PDUPROC_CALLBACK)pvd_process_radio, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_srdetect_fmt,
		      (PDUPROC_CALLBACK)pvd_process_impact, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_srimpact_fmt,
		      (PDUPROC_CALLBACK)pvd_process_impact, result);
    pduproc_catch_PDU(PV_SOURCE_LOCAL | PV_SOURCE_REMOTE, pvd_laser_fmt,
		      (PDUPROC_CALLBACK)pvd_process_laser, result);

    /* Get the default features */
    wld_get_playbox_id(name_sans_version,&version);
    
    if (!(result->default_features = (struct pvd_default_features *)
	  reader_find_tag(reader_get_symbol(name_sans_version),
			  default_features.array,
			  READER_UNTAGGED, READER_NO_ERRORS)))
    {
	result->default_features = (struct pvd_default_features *)
	  reader_find_tag(reader_get_symbol("default"),
			  default_features.array,
			  READER_UNTAGGED, READER_ERRORS_WITH_CONTEXT);
    }

    result->mf_list = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY,0);
    result->zoom_event = callback_define_event(A_INT, A_END);

    make_legend(result);
    make_scale(result);
    make_show_as(result);

    /* Make the feature editor */
    result->editor = edt_create(editor_dfn.array,
				sizeof(struct pvd_editor),
				(EDT_RENDER_FUNCTION)apply_editor, result,
				(EDT_RENDER_FUNCTION)apply_editor, result,
				(EDT_EXIT_FUNCTION)done_editor, result,
				TRUE,
				gui, tactmap, NULL, map_erase_gc,
				sensitive, refresh_event, NULL);

    sgui_add_mode(gui, SGUI_TOOL_MODE,
		  (char *)pvd_features_icon_bits,
		  pvd_features_icon_width, pvd_features_icon_height,
		  (SGUI_MODE_CALLBACK)pvd_edit_features, result,
		  "PVD Editor: Select features for display, use Apply or Done buttons to take effect.",
		  "selecting features");

    /* Color the feature buttons */
    feat = reader_get_symbol("Features");
    for (i=0;i<XtNumber(features);i++)
      edt_color_choice(result->editor, feat,
		       (int32)features[i].class,
		       tactmap_get_color(result->tactmap,
					 features[i].class));

#ifndef MODSTEALTH
    if (!pvd_modstealth)
    {
	/* Make the stealth editor */
	result->stealth_editor =
	  edt_create(stealth_editor_dfn.array,
		     sizeof(struct pvd_stealth_editor),
		     (EDT_RENDER_FUNCTION)pvd_stealth_apply, result,
		     (EDT_RENDER_FUNCTION)pvd_stealth_apply, result,
		     (EDT_EXIT_FUNCTION)pvd_stealth_done, result,
		     TRUE, gui, tactmap, NULL, map_erase_gc,
		     sensitive, refresh_event, NULL);
	result->stealth_mode =
	  sgui_add_mode(gui, SGUI_TOOL_MODE,
			(char *)pvd_stealth_icon_bits,
			pvd_stealth_icon_width, pvd_stealth_icon_height,
			(SGUI_MODE_CALLBACK)pvd_edit_stealth, result,
			"Stealth Control: Control stealth view.",
			"controlling stealth");
    }
#endif

    /* We make the GCs with the root window, since we know it is realized */
    root = RootWindowOfScreen(XtScreen(map_widget));

    /* Make the vehicle GCs */
    gcv.font = resources.designations;
    gcv.foreground = resources.black;
    gcv.background = tactmap_get_color(result->tactmap,
				       (TACTMAP_OBJECT_CLASS)-1);
    result->black_gc = XCreateGC(XtDisplay(map_widget),root,
				 GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.white;
    result->white_gc = XCreateGC(XtDisplay(map_widget),root,
				 GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.unkn;
    result->force_gc[forceIDIrrelevant] =
      XCreateGC(XtDisplay(map_widget),root,
		GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.dist;
    result->force_gc[distinguishedForceID] =
      XCreateGC(XtDisplay(map_widget),root,
		GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.othr;
    result->force_gc[otherForceID] =
      XCreateGC(XtDisplay(map_widget),root,
		GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.obsr;
    result->force_gc[observerForceID] =
      XCreateGC(XtDisplay(map_widget),root,
		GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.targ;
    result->force_gc[targetForceID] =
      XCreateGC(XtDisplay(map_widget),root,
		GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.hit;
    result->hit_gc = XCreateGC(XtDisplay(map_widget),root,
			       GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.miss;
    result->miss_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.indf;
    result->indf_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.mine;
    result->mine_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.coll;
    result->coll_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.mf;
    result->mf_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.opmf;
    result->opmf_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.lane;
    result->lane_gc = XCreateGC(XtDisplay(map_widget),root,
				GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.miclic;
    result->miclic_gc = XCreateGC(XtDisplay(map_widget),root,
				  GCForeground|GCBackground|GCFont,&gcv);
    gcv.foreground = resources.laser;
    result->laser_gc = XCreateGC(XtDisplay(map_widget),root,
				  GCForeground|GCBackground|GCFont,&gcv);
    for (i = 0; i < 10; i++)
    {
	gcv.foreground = resources.grey[i];
	result->shading_gc[i] = XCreateGC(XtDisplay(map_widget),root,
				  GCForeground|GCBackground|GCFont,&gcv);
    }

    /* Continue filling in shading array */
    for (i = 0; i < 9; i++)
    {
	gcv.foreground = resources.brown[i];
	result->shading_gc[i+10] = XCreateGC(XtDisplay(map_widget),root,
				  GCForeground|GCBackground|GCFont,&gcv);
    }

    /* Cache the font metrics, so libtactmap doesn't have to look them
     * up over and over.
     */
    result->font_struct =
      XQueryFont(XtDisplay(map_widget), XGContextFromGC(result->black_gc));

    result->zoom = sgui_add_mode(result->gui, SGUI_MAP_MODE,
				 (char *)pvd_zoom_bits,
				 pvd_zoom_width, pvd_zoom_height,
				 (SGUI_MODE_CALLBACK)pvd_zoom_mode, result,
				 "Zoom: click middle to zoom in around \
point; click right to zoom out around point; click and drag middle to \
set screen area",
				 "changing scale");

    gcv.foreground = resources.zoom;
    result->zoom_gc =
      XCreateGC(XtDisplay(map_widget), root, GCForeground, &gcv);

    result->pan = sgui_add_mode(result->gui, SGUI_MAP_MODE,
				(char *)pvd_pan_bits,
				pvd_pan_width, pvd_pan_height,
				(SGUI_MODE_CALLBACK)pvd_pan_mode, result,
				"Pan: click middle to set map center; click \
and drag right to change viewing area",
				"changing map location");

    result->info = sgui_add_mode(result->gui, SGUI_MAP_MODE,
				 (char *)pvd_info_bits,
				 pvd_info_width, pvd_info_height,
				 (SGUI_MODE_CALLBACK)pvd_info_mode, result,
				 "Information: click middle on vehicle for \
description; click middle on map for coordinates",
				 "getting map information");


    result->env = sgui_add_mode(result->gui, SGUI_MAP_MODE,
				(char *)pvd_env_bits,
				pvd_env_width, pvd_env_height,
				(SGUI_MODE_CALLBACK)pvd_env_mode, result,
				"Information: click middle on map for environmental information",
				 "getting environment information");

    pvd_create_env_display(sgui_shell_widget(gui),
			   result,
			   resources.black,
			   resources.designations,
			   root,
			   &result->env_display);

    sgui_set_mode(result->zoom);

    pvd_pvds[pvd_n_pvds] = result;
    pvd_n_pvds++;

    /* Get the default features from the editor definition */
    if (result->editor)
    {
	edt_query(result->editor, &edt_defaults);
	edt_defaults.use_these_features = 1;
	apply_editor_data(result, &edt_defaults);
    }

    /* Register to be told when the user changes defaults, and get
     * initial values.
     */
    callback_register_handler(edt_defaults_callback, new_defaults, result);
    new_defaults(result->gui, result);

    return result;
}

SNSTVE_CLASS *pvd_vehicle_snstve_class(PVD_GUI_PTR pvd_gui)
{
    return &pvd_gui->vehicle_class;
}

static void pvd_toggle_feature(Widget w, PVD_GUI_PTR pvd_gui,
			       XmToggleButtonCallbackStruct *call_data)
{
    Arg arg;
    int32 which=0;

    XtSetArg(arg, XmNuserData, &which);
    XtGetValues(w, &arg, 1);

    if (call_data->set)
      tactmap_add_object(pvd_gui->tactmap, pvd_gui->features[which].obj);
    else
      tactmap_remove_object(pvd_gui->tactmap, pvd_gui->features[which].obj);

    /* Update the editor */
    if (pvd_gui->editor)
      load_editor(pvd_gui->editor, pvd_gui);
}

void pvd_freeze(PVD_GUI_PTR pvd_gui, uint32 freeze)
{
    pvd_gui->frozen = freeze;
}

static void make_legend(PVD_GUI_PTR pvd_gui)
{
    Arg args[2];
    int32 i;

    for (i=0;i<XtNumber(features);i++)
    {
	/* Note that the definition has already been zero'd out, so all
	 * we need to do is set the non-zero values
	 */
	if (features[i].sensitive)
	{
	    pvd_gui->features[i].dfn.sensitive.class =
	      (SNSTVE_CLASS *)STDALLOC(sizeof(SNSTVE_CLASS));
	    SNSTVE_INIT_CLASS(*pvd_gui->features[i].dfn.sensitive.class);
	}
	
	/* Some feature class need specific option set */
	switch (features[i].class)
	{
	  case TACTMAP_HYPSO:
	    pvd_gui->features[i].dfn.hypso.kind = TACTMAP_COLOR_HYPSO;
	    break;
	  case TACTMAP_TREES:
	    pvd_gui->features[i].dfn.trees.mask = TACTMAP_TREE_CANOPIES;
	    break;
	  case TACTMAP_CONTOURS:
	    pvd_gui->features[i].dfn.contours.minor_spacing = 100.0;
	    pvd_gui->features[i].dfn.contours.major_spacing = 100.0;
	    break;
	  case TACTMAP_GRIDS:
	    pvd_gui->features[i].dfn.grids.interval   = 1000;
	    pvd_gui->features[i].dfn.grids.label_mask = COORD_LABEL_ALL;
	    pvd_gui->features[i].dfn.grids.kind       = TACTMAP_MILGRID;
	    break;
	  case TACTMAP_CELLS:
	    pvd_gui->features[i].dfn.grids.interval   = 0;
	    pvd_gui->features[i].dfn.grids.label_mask = COORD_LABEL_TOP;
	    pvd_gui->features[i].dfn.grids.kind       = TACTMAP_GCS;
	    break;
	}

	pvd_gui->features[i].class = features[i].class;

	pvd_gui->features[i].obj =
	  tactmap_create_object(pvd_gui->tactmap, features[i].class,
				&pvd_gui->features[i].dfn);

	pvd_gui->features[i].toggle =
	  sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
			   "Map Features", features[i].label,
			   xmToggleButtonWidgetClass);

	XtSetArg(args[0], XmNforeground,
		 tactmap_get_color(pvd_gui->tactmap, features[i].class));
	XtSetArg(args[1], XmNuserData, i);
	XtSetValues(pvd_gui->features[i].toggle, args, 2);

	XtAddCallback(pvd_gui->features[i].toggle, XmNvalueChangedCallback,
		      (XtCallbackProc)pvd_toggle_feature, pvd_gui);

	/* Tell libeditor about the road class */
	if (features[i].class == TACTMAP_ROADS)
	  edt_set_sensitive_class(pvd_gui->gui, EDT_ROADS,
				  pvd_gui->features[i].dfn.sensitive.class);
    }	
}

static void pvd_toggle_scale(Widget w, PVD_GUI_PTR pvd_gui,
			     XmToggleButtonCallbackStruct *call_data)
{
    Arg arg;
    int32 which=0;

    XtSetArg(arg, XmNuserData, &which);
    XtGetValues(w, &arg, 1);

    /* Ignore these */
    if (!call_data->set)
      return;

    /* Remember where we were so we can unzoom */
    pvd_remember_where(pvd_gui);

    pvd_set_scale_index(pvd_gui, which);

    /* Update the editor */
    if (pvd_gui->editor)
      load_editor(pvd_gui->editor, pvd_gui);
}

static void make_scale(PVD_GUI_PTR pvd_gui)
{
    int32 i;
    char name[128];
    int32 scale;
    Arg arg;

    for (i=0;i<pvd_gui->default_features->num_scales_plus_5-5;i++)
    {
	scale = pvd_gui->default_features->scales[i]->scale;
	if (scale < 1000)
	  sprintf(name, "1:%d", scale);
	else if (scale < 1000000)
	  sprintf(name, "1:%d,%03d", scale/1000, scale%1000);
	else
	  sprintf(name, "1:%d,%03d,%03d",
		  scale/1000000, (scale/1000)%1000, scale%1000);
	
	pvd_gui->scale_toggles[i] =
	  sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
			   "Map Scale", name, xmToggleButtonWidgetClass);

	XtSetArg(arg, XmNuserData, i);
	XtSetValues(pvd_gui->scale_toggles[i], &arg, 1);

	XtAddCallback(pvd_gui->scale_toggles[i], XmNvalueChangedCallback,
		      (XtCallbackProc)pvd_toggle_scale, pvd_gui);
    }
    XtSetArg(arg, XmNradioBehavior, TRUE);
    XtSetValues(XtParent(pvd_gui->scale_toggles[0]), &arg, 1);

    pvd_set_scale(pvd_gui, (float64)pvd_gui->default_features->initial_scale);
    pvd_gui->veh_magnification = 1.0;
    pvd_gui->veh_mag_dont_adjust = FALSE;
    pvd_gui->raw_scale = 0.0;
}

static void pvd_toggle_show_as(Widget w, PVD_GUI_PTR pvd_gui,
			       XmToggleButtonCallbackStruct *call_data)
{
    Arg arg;
    int32 which=0;

    XtSetArg(arg, XmNuserData, &which);
    XtGetValues(w, &arg, 1);

    pvd_gui->show_type[which] = call_data->set;

    /* The user did this, so don't change it */
    pvd_gui->reenable_show_entities = 0;

    /* Fire the icon change event, but in order to have it fire only once,
     * fire it when a button is SET.  Note that this means that if other
     * button are added to this menu, someone else will have to detect when
     * those other buttons are set and fire this event for those.
     */
    if (call_data->set)
      callback_fire_event(pvd_gui->icon_change_event,
			  pvd_gui);
}

static void make_show_as(PVD_GUI_PTR pvd_gui)
{
    Arg arg;

    pvd_gui->show_as_vehicle_pictures = 
      sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
		       "Show As", "Vehicle Pictures",
		       xmToggleButtonWidgetClass);
    XtSetArg(arg, XmNuserData, PVD_SHOW_PICTURES);
    XtSetValues(pvd_gui->show_as_vehicle_pictures, &arg, 1);
    XtAddCallback(pvd_gui->show_as_vehicle_pictures, XmNvalueChangedCallback,
		  (XtCallbackProc)pvd_toggle_show_as, pvd_gui);

    pvd_gui->show_as_vehicle_icons = 
      sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
		       "Show As", "Vehicle Icons",
		       xmToggleButtonWidgetClass);
    XtSetArg(arg, XmNuserData, PVD_SHOW_ICONS);
    XtSetValues(pvd_gui->show_as_vehicle_icons, &arg, 1);
    XtAddCallback(pvd_gui->show_as_vehicle_icons, XmNvalueChangedCallback,
		  (XtCallbackProc)pvd_toggle_show_as, pvd_gui);

    if (PVD_ASPDU)
    {
	pvd_gui->show_as_aggregate_platoons = 
	  sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
			   "Show As", "Agg Platoons",
			   xmToggleButtonWidgetClass);
	XtSetArg(arg, XmNuserData, PVD_SHOW_PLATOONS);
	XtSetValues(pvd_gui->show_as_aggregate_platoons, &arg, 1);
	XtAddCallback(pvd_gui->show_as_aggregate_platoons, XmNvalueChangedCallback,
		      (XtCallbackProc)pvd_toggle_show_as, pvd_gui);
	
	pvd_gui->show_as_aggregate_companies = 
	  sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
			   "Show As", "Agg Companies",
			   xmToggleButtonWidgetClass);
	XtSetArg(arg, XmNuserData, PVD_SHOW_COMPANIES);
	XtSetValues(pvd_gui->show_as_aggregate_companies, &arg, 1);
	XtAddCallback(pvd_gui->show_as_aggregate_companies, XmNvalueChangedCallback,
		      (XtCallbackProc)pvd_toggle_show_as, pvd_gui);
	
	pvd_gui->show_as_aggregate_battalions = 
	  sgui_add_menubar(pvd_gui->gui, PRIV_OPERATOR,
			   "Show As", "Agg Battalions",
			   xmToggleButtonWidgetClass);
	XtSetArg(arg, XmNuserData, PVD_SHOW_BATTALIONS	);
	XtSetValues(pvd_gui->show_as_aggregate_battalions, &arg, 1);
	XtAddCallback(pvd_gui->show_as_aggregate_battalions, XmNvalueChangedCallback,
		      (XtCallbackProc)pvd_toggle_show_as, pvd_gui);
    }
    

    XtSetArg(arg, XmNradioBehavior, TRUE);
    XtSetValues(XtParent(pvd_gui->show_as_vehicle_icons), &arg, 1);

    pvd_gui->icon_change_event =
      callback_define_event(A_PTR, A_END);

    pvd_gui->display_settings_changed_event =
      callback_define_event(A_PTR, A_END);

    XmToggleButtonSetState(pvd_gui->show_as_vehicle_pictures, TRUE, TRUE);
}

/* External interface which allows the application to force the PVD
 * to accomodate a loss of entity state information by switching to
 * aggregate state display instead.
 */
void pvd_display_entities(PVD_GUI_PTR pvd_gui,
			  int32 do_display)
{
    int32 r;
    Widget pict = pvd_gui->show_as_vehicle_pictures;
    Widget icon = pvd_gui->show_as_vehicle_icons;
    Widget plat = pvd_gui->show_as_aggregate_platoons;

    if (!PVD_ASPDU)
      return;

    if (!do_display)
    {
	/* If we are not allowed to display entities, and we are currently
	 * displaying them, switch to platoon pictures.
	 */
	if (pvd_gui->show_type[PVD_SHOW_PICTURES] ||
	    pvd_gui->show_type[PVD_SHOW_ICONS])
	{
	    /* Note that we should undo this if we get a chance */
	    pvd_gui->reenable_show_entities =
	      ((pvd_gui->show_type[PVD_SHOW_PICTURES] << PVD_SHOW_PICTURES) |
	       (pvd_gui->show_type[PVD_SHOW_ICONS] << PVD_SHOW_ICONS));

	    pvd_gui->show_type[PVD_SHOW_PICTURES] =
	      pvd_gui->show_type[PVD_SHOW_ICONS] =
		FALSE;

	    XmToggleButtonSetState(pict, FALSE, FALSE);
	    XmToggleButtonSetState(icon, FALSE, FALSE);

	    pvd_gui->show_type[PVD_SHOW_PLATOONS] = TRUE;
	    XmToggleButtonSetState(plat, TRUE, FALSE);

	    callback_fire_event(pvd_gui->icon_change_event, pvd_gui);
	}
    }
    else
    {
	/* If we are now allowed to display entities, we were
	 * previously forced to stop displaying them, and the user
	 * has not fiddled with the show-as, then turn them back on
	 * again.
	 */
	if (r = pvd_gui->reenable_show_entities)
	{
	    /* Turn everything off */
	    bzero((ADDRESS)pvd_gui->show_type, sizeof(pvd_gui->show_type));
	    XmToggleButtonSetState(pvd_gui->show_as_aggregate_platoons,
				   FALSE, FALSE);
	    XmToggleButtonSetState(pvd_gui->show_as_aggregate_companies,
				   FALSE, FALSE);
	    XmToggleButtonSetState(pvd_gui->show_as_aggregate_battalions,
				   FALSE, FALSE);

	    /* Turn back on the ones the user had selected before */
	    pvd_gui->show_type[PVD_SHOW_PICTURES] =
	      ((r & (1<<PVD_SHOW_PICTURES)) != 0);
	    XmToggleButtonSetState(pict,
				   pvd_gui->show_type[PVD_SHOW_PICTURES],
				   FALSE);
	    pvd_gui->show_type[PVD_SHOW_ICONS] =
	      ((r & (1<<PVD_SHOW_ICONS)) != 0);
	    XmToggleButtonSetState(icon,
				   pvd_gui->show_type[PVD_SHOW_ICONS],
				   FALSE);
	    pvd_gui->reenable_show_entities = 0;

	    callback_fire_event(pvd_gui->icon_change_event, pvd_gui);
	}
    }       

    XtSetSensitive(pict, do_display);
    XtSetSensitive(icon, do_display);
}

/*
 * Set the alignment of the pvd as buttons are pressed.
 */
static void pvd_set_align(Widget w, PVD_GUI_PTR pvd_gui,
			  XmToggleButtonCallbackStruct *call_data)
{
    Arg   arg;
    int32 alignment=0;

    XtSetArg(arg, XmNuserData, &alignment);
    XtGetValues(w, &arg, 1);
    
    pvd_gui->alignments_displayed[alignment] = call_data->set;
}

static void make_local_force(PVD_GUI_PTR pvd_gui)
{
    Arg arg;

    /* Put up the local workstation alignment menus */

    pvd_gui->local_force_blue =
      sgui_add_menubar(pvd_gui->gui, PRIV_BATTLEMASTER,
		       "Local Force", "Friendly",
		       xmToggleButtonWidgetClass);
    XtSetArg(arg, XmNuserData, distinguishedForceID);
    XtSetValues(pvd_gui->local_force_blue, &arg, 1);
    XtAddCallback(pvd_gui->local_force_blue, XmNvalueChangedCallback, 
		  (XtCallbackProc)pvd_set_align, pvd_gui);

    pvd_gui->local_force_red =
      sgui_add_menubar(pvd_gui->gui, PRIV_BATTLEMASTER,
		       "Local Force", "Enemy",
		       xmToggleButtonWidgetClass);
    XtSetArg(arg, XmNuserData, otherForceID);
    XtSetValues(pvd_gui->local_force_red, &arg, 1);
    XtAddCallback(pvd_gui->local_force_red, XmNvalueChangedCallback,
		  (XtCallbackProc)pvd_set_align, pvd_gui);

    XmToggleButtonSetState(pvd_gui->local_force_blue, TRUE, TRUE);
    XmToggleButtonSetState(pvd_gui->local_force_red, TRUE, TRUE);
}

static void new_defaults(SGUI_PTR gui, PVD_GUI_PTR pvd_gui)
{
    Widget v, h;
    uint8 placement;
    Arg arg;
    EDT_PVD_DEFAULTS dflt;

    if (gui != pvd_gui->gui)
      return;

    edt_pvd_defaults(gui, &dflt);

    pvd_gui->altitudes_agl = dflt.altitudes_agl;
    if (pvd_gui->fixed_contours != dflt.fixed_contour)
    {
	pvd_gui->fixed_contours = dflt.fixed_contour;
	pvd_set_scale(pvd_gui, pvd_gui->current_scale);
    }

    edt_control_widget(pvd_gui->editor,
		       OFFSETOF(struct pvd_editor, fixed_minor_contour),
		       dflt.fixed_contour ? EDT_NORMAL : EDT_UNMANAGED);
    edt_control_widget(pvd_gui->editor,
		       OFFSETOF(struct pvd_editor, fixed_major_contour),
		       dflt.fixed_contour ? EDT_NORMAL : EDT_UNMANAGED);

    sgui_map_scrollbars(pvd_gui->gui, &h, &v);
    if (dflt.scrollbars == EDT_NO_SCROLLBARS)
      XtUnmanageChild(h), XtUnmanageChild(v);
    else
    {
	switch (dflt.scrollbars)
	{
	  case EDT_TOP_LEFT: placement = XmTOP_LEFT; break;
	  case EDT_BOTTOM_LEFT: placement = XmBOTTOM_LEFT; break;
	  case EDT_TOP_RIGHT: placement = XmTOP_RIGHT; break;
	  default:
	  case EDT_BOTTOM_RIGHT: placement = XmBOTTOM_RIGHT;
	}
	XtSetArg(arg, XmNscrollBarPlacement, placement);
	XtSetValues(XtParent(h), &arg, 1);
	XtManageChild(h), XtManageChild(v);
    }
}

static void apply_editor_data(PVD_GUI_PTR pvd_gui, struct pvd_editor *data)
{
    int32 i, j;
    int32 present;
    int32 old_icon_size;

    /* Check the flag value that tells us this is just the editor telling
     * us the defaults, so we know to ignore them.
     */
    if (!data->use_these_features)
      return;

    pvd_gui->raw_icon_scale = data->icon_scale;
    /* Icons will range between 15 and 75 pixels big */
    old_icon_size = pvd_gui->icon_size;
    pvd_gui->icon_size = 15 + (data->icon_scale * 60);

    /* If the icon size changed, fire an event on it */
    if (pvd_gui->icon_size != old_icon_size)
      callback_fire_event(pvd_gui->icon_change_event, pvd_gui);

    pvd_gui->veh_magnification = 
      (K * pvd_gui->current_scale) * data->vehicle_scale;
    pvd_gui->raw_scale = data->vehicle_scale;
    if (data->vehicle_scale == 0.0)
      pvd_gui->veh_mag_dont_adjust = TRUE;
    else
      pvd_gui->veh_mag_dont_adjust = FALSE;

    for (i=0;i<XtNumber(features);i++)
    {
	present = 0;
	for (j=0;j<data->feature_count;j++)
	  if (data->features[j] == (int32)pvd_gui->features[i].class)
	  {
	      present = 1;
	      break;
	  }
	XmToggleButtonSetState(pvd_gui->features[i].toggle,
			       present, FALSE);

	if (present)
	  tactmap_add_object(pvd_gui->tactmap, pvd_gui->features[i].obj);
	else
	  tactmap_remove_object(pvd_gui->tactmap, pvd_gui->features[i].obj);
    }

    pvd_gui->features[PVD_HYPSO].dfn.hypso.kind =
      (TACTMAP_HYPSO_KIND)data->hypso_kind;

    tactmap_set_em_band(pvd_gui->tactmap, data->em_band);
    tactmap_update_object(pvd_gui->tactmap,
			  pvd_gui->features[PVD_HYPSO].obj,
			  &pvd_gui->features[PVD_HYPSO].dfn);

    pvd_gui->fixed_minor_contour = data->fixed_minor_contour;
    pvd_gui->fixed_major_contour = data->fixed_major_contour;

    pvd_set_scale(pvd_gui, pvd_gui->current_scale);

    if (!(pvd_gui->veh_update_rate = data->update_rate * 1000))
      pvd_gui->veh_update_rate = 500;

    pvd_gui->markings = 0;
    for (i=0;i<data->num_markings;i++)
      pvd_gui->markings |= (1 << data->markings[i]);

    pvd_gui->smoke_display = data->smoke_display;

    pvd_gui->mf_sym_cutoff = data->mf_sym_cutoff;

    /* Load up what events to draw on the PVD.  Turn them all off, then
     * turn on only the ones selected.
     */
    pvd_gui->draw_collisions =
      pvd_gui->draw_fires =
	pvd_gui->draw_impacts =
	  pvd_gui->draw_for =
	    pvd_gui->draw_links =
	      pvd_gui->draw_radio_traffic = 
                pvd_gui->draw_laser_designations = FALSE;

    for (i = 0; i < data->num_events; i++)
    {
	if (data->events[i] == PVD_SHOW_COLLISION)
	  pvd_gui->draw_collisions = TRUE;
	else if (data->events[i] == PVD_SHOW_FIRE)
	  pvd_gui->draw_fires = TRUE;
	else if (data->events[i] == PVD_SHOW_IMPACT)
	  pvd_gui->draw_impacts = TRUE;
	else if (data->events[i] == PVD_SHOW_RADIO_TRAFFIC)
	  pvd_gui->draw_radio_traffic = TRUE;
	else if (data->events[i] == PVD_SHOW_FOR)
	  pvd_gui->draw_for = TRUE;
	else if (data->events[i] == PVD_SHOW_LINKS)
	  pvd_gui->draw_links = TRUE;
	else if (data->events[i] == PVD_SHOW_LASER_DSG)
	  pvd_gui->draw_laser_designations = TRUE;
    }

    callback_fire_event(pvd_gui->display_settings_changed_event, pvd_gui);
}

/*ARGSUSED*/
static void apply_editor(EDT_EDITOR_PTR editor, int32 transient,
			 struct pvd_editor *old, struct pvd_editor *new,
			 PVD_GUI_PTR pvd_gui)
{
    apply_editor_data(pvd_gui, new);
}

/*ARGSUSED*/
static void done_editor(EDT_EDITOR_PTR editor, struct pvd_editor *new,
			PVD_GUI_PTR pvd_gui, EDT_EXIT_STATUS status)
{
    /* When the status is ABORT, the editor is kind enough to fill
     * *new with the revert values.
     */
    apply_editor_data(pvd_gui, new);
}

static void load_editor(EDT_EDITOR_PTR editor, PVD_GUI_PTR pvd_gui)
{
    struct pvd_editor cur;
    int32 i, j;

    cur.vehicle_scale = pvd_gui->raw_scale;

    if (cur.vehicle_scale < 0.0)
      cur.vehicle_scale = 0.0;
    else if (cur.vehicle_scale > 1.0)
      cur.vehicle_scale = 1.0;

    cur.icon_scale = pvd_gui->raw_icon_scale;

    for(i=j=0;i<XtNumber(features);i++)
      if (XmToggleButtonGetState(pvd_gui->features[i].toggle))
	cur.features[j] = (int32)pvd_gui->features[i].class, j++;
    cur.feature_count = j;

    cur.hypso_kind = (int32)
      pvd_gui->features[PVD_HYPSO].dfn.hypso.kind;

    cur.update_rate = pvd_gui->veh_update_rate / 1000;

    cur.num_markings = 0;
    for (i=0;i<PVD_DESIGNATIONS;i++)
      if (pvd_gui->markings & (1 << i))
	cur.markings[cur.num_markings++] = i;

    cur.em_band = tactmap_get_em_band(pvd_gui->tactmap);
    cur.smoke_display = pvd_gui->smoke_display;
    cur.mf_sym_cutoff = pvd_gui->mf_sym_cutoff;

    cur.fixed_minor_contour = pvd_gui->fixed_minor_contour;
    cur.fixed_major_contour = pvd_gui->fixed_major_contour;

    cur.smoke_display = pvd_gui->smoke_display;
    cur.num_events = 0;
    if (pvd_gui->draw_collisions)
      cur.events[cur.num_events++] = PVD_SHOW_COLLISION;
    if (pvd_gui->draw_fires)
      cur.events[cur.num_events++] = PVD_SHOW_FIRE;
    if (pvd_gui->draw_radio_traffic)
      cur.events[cur.num_events++] = PVD_SHOW_RADIO_TRAFFIC;
    if (pvd_gui->draw_impacts)
      cur.events[cur.num_events++] = PVD_SHOW_IMPACT;
    if (pvd_gui->draw_for)
      cur.events[cur.num_events++] = PVD_SHOW_FOR;
    if (pvd_gui->draw_links)
      cur.events[cur.num_events++] = PVD_SHOW_LINKS;
    if (pvd_gui->draw_laser_designations)
      cur.events[cur.num_events++] = PVD_SHOW_LASER_DSG;

    cur.use_these_features = 1;

    edt_load(editor, &cur, sizeof(cur));
}

int32 pvd_get_icon_size(PVD_GUI_PTR pvd)
{
    return pvd->icon_size;
}

CALLBACK_EVENT_PTR pvd_get_icon_change_event(PVD_GUI_PTR pvd)
{
    return pvd->icon_change_event;
}

CALLBACK_EVENT_PTR pvd_get_settings_changed_event(PVD_GUI_PTR pvd)
{
    return pvd->display_settings_changed_event;
}

void pvd_get_display_settings(PVD_GUI_PTR pvd, PVD_DISPLAY_SETTINGS *settings)
{
    if (!pvd || !settings)
      return;

    settings->draw_for = pvd->draw_for;
    settings->draw_links = pvd->draw_links;
    settings->mf_sym_cutoff = pvd->mf_sym_cutoff;
}

int32 pvd_get_scale(PVD_GUI_PTR pvd)
{
    return (int32)pvd->current_scale;
}

CALLBACK_EVENT_PTR pvd_get_zoom_event(PVD_GUI_PTR pvd)
{
    return pvd->zoom_event;
}

VTAB_LIST pvd_get_minefield_list(PVD_GUI_PTR pvd)
{
    return pvd->mf_list;
}

TACTMAP_OBJECT_PTR pvd_obj_from_class (PVD_GUI_PTR pvd_gui,
                                       TACTMAP_OBJECT_CLASS class)
{
    int32 i = 0;
    
    for (i=0; i<TACTMAP_LAST_CLASS; i++) 
    {
        if(pvd_gui->features[i].class == class) 
	    return(pvd_gui->features[i].obj);
    }
    return (NULL);
}	
  

#else

int pvd_dummy_archive_variable;

#endif
/*USE_MOTIF*/
