#ifndef lint
static char rcsid [] = "$RCSfile: pvd_pan.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_pan.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle panning                                    *
*       Created: Tue Sep 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"

static void pan_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
			SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui);

/*ARGSUSED*/
void pvd_pan_mode(SGUI_PTR gui, SGUI_MODE_PTR mode, PVD_GUI_PTR pvd_gui,
		  SGUI_MODE_STATE state)
{
    if ((state == SGUI_ACTIVE) ||
	(state == SGUI_RESUMED))
    {
	/* Install the mouse handler */
	callback_register_handler(pvd_gui->map_event,pan_gesture,pvd_gui);
	pvd_gui->current = pvd_gui->pan;
    }
    else
    {
	/* Remove the mouse handler */
	callback_unregister_handler(pvd_gui->map_event,pan_gesture,pvd_gui);
    }
}

/*ARGSUSED*/
static void pan_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
			SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui)
{
    float64 map_x, map_y;
    int32   map_c;
    int32 screen_x, screen_y;
    char buf[128];

    /* Ignore button 1, and EXCITED/UNEXCITED callbacks */
    if (call_data->button < 2)
      return;

    switch(action)
    {
      case SNSTVE_PRESS:
	/* Put us into world view mode for right button */
	if (call_data->button == 3)
	{
	    /* Do this on the press for button 3 */
	    pvd_remember_where(pvd_gui);

	    tactmap_get_center(pvd_gui->tactmap, &map_x, &map_y, &map_c);

	    tactmap_world_view(pvd_gui->tactmap, TRUE);

	    /* Warp the mouse to the map center in world view */
	    if (tactmap_map_to_screen(pvd_gui->tactmap,
				      map_x, map_y, map_c,
				      &screen_x, &screen_y))
	    {
		XWarpPointer(XtDisplay(pvd_gui->map_widget),
			     None, XtWindow(pvd_gui->map_widget),
			     0,0,0,0,screen_x,screen_y);
	    }
	}
	break;
      case SNSTVE_MOTION:
	/* Only do drag stuff on right button */
	if (call_data->button != 3)
	  return;

	/* Set the new center */
	tactmap_screen_to_map(pvd_gui->tactmap,
			      call_data->screen_x, call_data->screen_y,
			      &map_x, &map_y, &map_c);
	tactmap_set_center(pvd_gui->tactmap, (int32)map_x, 
			   (int32)map_y, map_c);
	pvd_update_scrollbars(pvd_gui);

	/* Tell the user the new center on the help line */
	sgui_pop_help(pvd_gui->pan, TRUE);
	edt_format(pvd_gui->gui, buf, "New Center: #l",
		   map_c, map_x, map_y);
	sgui_push_help(pvd_gui->pan, buf);

	break;
      case SNSTVE_RELEASE:
	/* Do this on release for button 2 */
	if (call_data->button == 2)
	  pvd_remember_where(pvd_gui);

	/* Pop the location off the help line */
	if (call_data->button == 3)
	  sgui_pop_help(pvd_gui->pan, TRUE);

	/* Set the new center */
	tactmap_screen_to_map(pvd_gui->tactmap,
			      call_data->screen_x, call_data->screen_y,
			      &map_x, &map_y, &map_c);
	tactmap_set_center(pvd_gui->tactmap, (int32)map_x, 
			   (int32)map_y, map_c);
	pvd_update_scrollbars(pvd_gui);

	/* Make sure we're not in world view mode (we do this regardless
	 * of whether the release was button 2 or 3, in order to protect
	 * ourselves from chords)
	 */
	tactmap_world_view(pvd_gui->tactmap, FALSE);
	break;
    }
}

#endif
/*USE_MOTIF*/
