#ifndef lint
static char rcsid [] = "$RCSfile: pvd_scroll.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_scroll.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Scrolling callback                                        *
*       Created: Wed Mar 10 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <libtactmap.h>
#include <libworld.h>
#include <Xm/ScrollBar.h>

#define SB_MAX 1000
#define SB_SIZE 50

void pvd_update_scrollbars(PVD_GUI_PTR pvd_gui)
{
    Widget h, v;
    int32 value;
    float64 cx, cy;
    Arg arg;
    int32 cc;
    float64 mn[XYZ],mx[XYZ];

    sgui_map_scrollbars(pvd_gui->gui, &h, &v);

    tactmap_get_center(pvd_gui->tactmap, &cx, &cy, &cc);
    wld_get_playbox_extent(cc,mn,mx);
    
    if (cx < mn[X])
      cx = mn[X];
    else if (cx > mx[X])
      cx = mx[X];

    if (cy < mn[Y])
      cy = mn[Y];
    else if (cy > mx[Y])
      cy = mx[Y];

    if ( mx[X] != mn[X])
      value =
	(SB_MAX - SB_SIZE) * (cx - mn[X])/(mx[X] - mn[X]);
    else
      value = 0;

    XtSetArg(arg, XmNvalue, value);
    XtSetValues(h, &arg, 1);

    if (mx[Y] != mn[Y])
      value =
	(SB_MAX - SB_SIZE) * (cy - mn[Y])/(mx[Y] - mn[Y]);
    else
      value = 0;

    XtSetArg(arg, XmNvalue, value);
    XtSetValues(v, &arg, 1);
}

static void pvd_scrollbar(Widget w, PVD_GUI_PTR pvd_gui,
			  XmScrollBarCallbackStruct *call_data)
{
    uint8 which;
    Widget h, v;
    float64 cx, cy, width, height;
    float64 *variable, min_val, max_val, size;
    int32 cc;
    float64 mn[XYZ],mx[XYZ];

    /* Figure out which scrollbar it is */
    sgui_map_scrollbars(pvd_gui->gui, &h, &v);
    if (h == w)
      which = XmHORIZONTAL;
    else if (v = w)
      which = XmVERTICAL;
    else
      return;

    /* When the user is dragging, show the world view */
    tactmap_world_view(pvd_gui->tactmap,(call_data->reason == XmCR_DRAG));
    tactmap_get_center(pvd_gui->tactmap, &cx, &cy, &cc);
    tactmap_get_dimension(pvd_gui->tactmap, &width, &height);
    wld_get_playbox_extent(cc,mn,mx);

    /* Get current map info */
    if (which == XmHORIZONTAL)
    {
	variable = &cx;
	min_val  = mn[X];
	max_val  = mx[X];
	size     = width;
    }
    else
    {
	variable = &cy;
	min_val  = mn[Y];
	max_val  = mx[Y];
	size     = height;
    }

    /* For drag or valueChanged, just go to the new value */
    if ((call_data->reason == XmCR_DRAG) ||
	(call_data->reason == XmCR_VALUE_CHANGED))
    {
	/* Only note the current location on the initial click of a drag */
	if (!pvd_gui->dragging)
	  pvd_remember_where(pvd_gui);

	pvd_gui->dragging = (call_data->reason == XmCR_DRAG);

	*variable = min_val +
	  (max_val - min_val) * (float64) call_data->value / (float64) SB_MAX;

	tactmap_set_center(pvd_gui->tactmap, (int32)cx, (int32)cy, cc);
	return;
    }

    /* Note the current location for the Undo buffer */
    pvd_remember_where(pvd_gui);

    /* Get the new value */
    switch (call_data->reason)
    {
      case XmCR_DECREMENT:	*variable -= size / 2;	break;
      case XmCR_INCREMENT:	*variable += size / 2;	break;
      case XmCR_PAGE_DECREMENT:	*variable -= size;	break;
      case XmCR_PAGE_INCREMENT:	*variable += size;	break;
      case XmCR_TO_TOP:		*variable = min_val;	break;
      case XmCR_TO_BOTTOM:	*variable = max_val;	break;
    }

    /* Clip the value */
    if (*variable < min_val)
      *variable = min_val;

    if (*variable > max_val)
      *variable = max_val;

    tactmap_set_center(pvd_gui->tactmap, (int32)cx, (int32)cy, cc);
    pvd_update_scrollbars(pvd_gui);
}

static Arg sb_args[] = { { XmNminimum , 0 } ,
			 { XmNmaximum , SB_MAX } ,
			 { XmNsliderSize , SB_SIZE } ,
		       };

void pvd_init_scrollbar(PVD_GUI_PTR pvd_gui, Widget sb)
{
    XtAddCallback(sb, XmNdecrementCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNdragCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNincrementCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNpageDecrementCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNpageIncrementCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNtoBottomCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNtoTopCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtAddCallback(sb, XmNvalueChangedCallback, 
                  (XtCallbackProc)pvd_scrollbar, pvd_gui);
    XtSetValues(sb, sb_args, XtNumber(sb_args));
}

#endif /*USE_MOTIF*/
