#ifndef lint
static char rcsid [] = "$RCSfile: pvd_stealth.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_stealth.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle stealth editor                             *
*       Created: Fri Nov 13 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF
#ifndef MODSTEALTH

#include "libpvd_local.h"
#include <libgcs.h>
#include <libstealth.h>
#include <libvtab.h>
#include <math.h>

static int32 find_preferred_stealth(PVD_GUI_PTR pvd);

static void load_for_stealth(int32 id, EDT_EDITOR_PTR editor)
{
    float64 pos[XYZC];
    struct pvd_stealth_editor data;

    stealth_get_position(id, pos);

#if 0
    data.attachment               = stealth_get_attachment(id);
#else
    data.attachment               = 1;
#endif
    data.mode                     = 0;
    data.location[X]              = pos[X];
    data.location[Y]              = pos[Y];
    data.altitude                 = pos[Z] - 
      ctdb_lookup_elevation((CTDB *)gcs_get_tdb((int32)pos[CELL3D], FALSE),
			    pos[X], pos[Y]);
    data.direction                = stealth_get_direction(id);
    data.needs_initialization     = FALSE;

    edt_load(editor,&data, sizeof(struct pvd_stealth_editor));
}

static void apply_editor_data(PVD_GUI_PTR pvd_gui,
			      struct pvd_stealth_editor *new,
			      struct pvd_stealth_editor *old)
{
    float64 pos[XYZC], dir;
    int32 id;

    if (!(id = find_preferred_stealth(pvd_gui)))
    {
	if (!pvd_gui->preferred_warned)
	{
	    if (pvd_gui->show_all_stealths)
	      sgui_error(pvd_gui->gui, "No stealth has been selected");
	    else
	      sgui_error(pvd_gui->gui, "Unable to find preferred stealth");
	    pvd_gui->preferred_warned = 1;
	}

	return;
    }

    if (old->needs_initialization || new->needs_initialization)
    {
	load_for_stealth(id, pvd_gui->stealth_editor);
	return;
    }

    if (new->attachment && new->attachment != old->attachment)
     {
	if (stealth_get_attachment(id) != new->attachment)
	  stealth_attach(id, new->attachment);

	if (new->mode)
	  stealth_set_mode(id, new->mode);
    }
    else if (!new->attachment && old->attachment)
    {
	stealth_get_position(id, pos);
	dir = stealth_get_direction(id);

	stealth_teleport(id, pos, dir);
	}
	else if (new->attachment && new->mode != old->mode)
	{
	if (new->mode)
	  stealth_set_mode(id, new->mode);
	}
	else
	{
		int changed = 0;

		stealth_get_position(id, pos);
		dir = stealth_get_direction(id);
	    
#define COMPARE(amp,field) \
     (bcmp(amp old->field, amp new->field, sizeof(old->field)))

		if (COMPARE(&,location[X]) || COMPARE(&,location[Y]))
		{
			pos[X] = new->location[X];
			pos[Y] = new->location[Y];
			changed = 1;
		}
		else if (COMPARE(&,altitude))
		{
			CTDB *ctdb=gcs_get_tdb((int32)pos[CELL3D], FALSE);
			float64 elev=ctdb_lookup_elevation(ctdb,pos[X],pos[Y]);

		    pos[Z] = new->altitude + elev;
			changed = 1;
		}
		else if (COMPARE(&,direction))
		{
			dir = new->direction;
			changed = 1;
		}

		if (changed)
			stealth_teleport(id, pos, dir);
	}
}

/*ARGSUSED*/
void pvd_stealth_apply(EDT_EDITOR_PTR editor, int32 transient,
		       struct pvd_stealth_editor *old, 
		       struct pvd_stealth_editor *new,
		       PVD_GUI_PTR pvd_gui)
{
#if 0
    if (!transient)
#endif
      apply_editor_data(pvd_gui, new, old);
}

/*ARGSUSED*/
void pvd_stealth_done(EDT_EDITOR_PTR editor, struct pvd_stealth_editor *new,
		      PVD_GUI_PTR pvd_gui, EDT_EXIT_STATUS status)
{
#if 0
    /* When the status is ABORT, the editor is kind enough to fill
     * *new with the revert values.
     */
    apply_editor_data(pvd_gui, new, FALSE, FALSE);
#endif

    pvd_gui->preferred_warned = 0;
}

/*ARGSUSED*/
void pvd_stealth_selected(SNSTVE_WINDOW_PTR window, SNSTVE_ACTION action,
			  SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui)
{
    int32 id;
    VehicleID vid;

    if ((action != SNSTVE_RELEASE) ||
	(call_data->button != 1))
      return;

    /* The user_data attached to the sensitive object is its vehicle ID */
    id = (int32)call_data->user_data;

    /* See if this is what we expect */
    if (!id || !(VTAB_TYPES_MATCH(vtab_vehicle_type(id),VTAB_STEALTH)))
      return;

    vtab_vehicle_simid(id, &vid);
    pvd_gui->preferred_stealth = vid.simulator;

    /* Fire up the editor */
    sgui_set_mode(pvd_gui->stealth_mode);

    /* Load it with initial values */
    load_for_stealth(id, pvd_gui->stealth_editor);
}

/* Finds the preferred stealth's vehicle ID
 */
static int32 find_preferred_stealth(PVD_GUI_PTR pvd)
{
    int32 id;
    VTAB_USER_DATA_TYPE user_data;
    VehicleID vid;

    for (id=vtab_get_first(vtab_vehicle_table, VTAB_ALL, &user_data); id;
	 id=vtab_get_next(vtab_vehicle_table, VTAB_ALL, &user_data))
    {
	vtab_vehicle_simid(id, &vid);

	if (NS_SIM_ADDRESSES_EQUAL(pvd->preferred_stealth, vid))
	  return id;
    }
    return 0;
}
#endif /* MODSTEALTH */
#endif /*USE_MOTIF*/
