#ifndef lint
static char rcsid [] = "$RCSfile: pvd_zoom.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pvd_zoom.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle zoom mode                                  *
*       Created: Tue Sep 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_MOTIF

#include "libpvd_local.h"
#include <Xm/ToggleB.h>

#define SMALLEST_SCALE 250.0  /* Things get ugly any smaller */

static void zoom_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
			 SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui);
static char *scale_string(PVD_GUI_PTR pvd_gui, float64 scale);

/*ARGSUSED*/
void pvd_zoom_mode(SGUI_PTR gui, SGUI_MODE_PTR mode, PVD_GUI_PTR pvd_gui,
		   SGUI_MODE_STATE state)
{
    if ((state == SGUI_ACTIVE) ||
	(state == SGUI_RESUMED))
    {
	/* Install the mouse handler */
	callback_register_handler(pvd_gui->map_event,zoom_gesture,pvd_gui);
	pvd_gui->current = pvd_gui->zoom;
    }
    else
    {
	/* Remove the mouse handler */
	callback_unregister_handler(pvd_gui->map_event,zoom_gesture,pvd_gui);
    }
}

static void draw_signed_rect(Display *display, Drawable d, GC gc,
			     XRectangle *rect)
{
    register int16 x, y, width, height;

    x = rect->x;
    y = rect->y;
    width = rect->width;
    height = rect->height;

    if (width < 0)
      x += width, width = -width;

    if (height < 0)
      y += height, height = -height;

    XDrawRectangle(display, d, gc, x, y, width, height);
}

/*ARGSUSED*/
static void zoom_gesture(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action,
			 SNSTVE_CALL_DATA *call_data, PVD_GUI_PTR pvd_gui)
{
    float64 map_x, map_y;
    int32   map_c;
    float64 scale;

    /* Ignore button 1, and EXCITED/UNEXCITED callbacks */
    if (call_data->button < 2)
      return;

    switch(action)
    {
      case SNSTVE_PRESS:
	pvd_gui->dragging = 0;
	return;
      case SNSTVE_MOTION:
	/* Only do drag stuff on middle button */
	if (call_data->button != 2)
	  return;

	if (pvd_gui->dragging)
	{
	    /* Erase the last box */
	    draw_signed_rect(XtDisplay(pvd_gui->map_widget),
			     XtWindow(pvd_gui->map_widget),
			     pvd_gui->map_erase_gc,
			     &pvd_gui->last_rect);

	    /* Erase the last scale */
	    sgui_pop_help(pvd_gui->zoom, TRUE);
	}
	else
	{
	    /* See if we have moved far enough to consider this a drag */
	    if ((call_data->xoff > DRAG_THRESH) ||
		(call_data->xoff < -DRAG_THRESH) ||
		(call_data->yoff > DRAG_THRESH) ||
		(call_data->yoff < -DRAG_THRESH))
	    {
		pvd_gui->dragging = 1;

		/* Infer the click location */
		pvd_gui->last_rect.x = call_data->screen_x - call_data->xoff;
		pvd_gui->last_rect.y = call_data->screen_y - call_data->yoff;
	    }
	}

	if (pvd_gui->dragging)
	{
	    pvd_gui->last_rect.width = call_data->xoff;
	    pvd_gui->last_rect.height = call_data->yoff;

	    /* Draw the new box */
	    draw_signed_rect(XtDisplay(pvd_gui->map_widget),
			     XtWindow(pvd_gui->map_widget),
			     pvd_gui->zoom_gc,
			     &pvd_gui->last_rect);

	    /* Set the help message to show the new scale */
	    scale = tactmap_get_scale(pvd_gui->tactmap,
				      call_data->xoff,call_data->yoff);
	    if (scale < SMALLEST_SCALE)
	      scale = SMALLEST_SCALE;
	    sgui_push_help(pvd_gui->zoom, scale_string(pvd_gui, scale));
	}
	break;
      case SNSTVE_RELEASE:
	if (pvd_gui->dragging)
	{
	    /* Erase the last box */
	    draw_signed_rect(XtDisplay(pvd_gui->map_widget),
			     XtWindow(pvd_gui->map_widget),
			     pvd_gui->map_erase_gc,
			     &pvd_gui->last_rect);

	    sgui_pop_help(pvd_gui->zoom, TRUE);
	}

	pvd_remember_where(pvd_gui);

	if (call_data->button == 2)
	{
	    if (!pvd_gui->dragging)
	    {
		tactmap_screen_to_map(pvd_gui->tactmap,
					call_data->screen_x,
					call_data->screen_y,
					&map_x,&map_y,&map_c);
		pvd_set_scale_index(pvd_gui, pvd_gui->current_scale_index+1);
	    }
	    else
	    {
		tactmap_screen_to_map(pvd_gui->tactmap,
				      (int32)call_data->screen_x -
				      (call_data->xoff/2),
				      (int32)call_data->screen_y -
				      (call_data->yoff/2),
				      &map_x,&map_y,&map_c);
		scale = tactmap_get_scale(pvd_gui->tactmap,
					  call_data->xoff,call_data->yoff);
		if (scale < SMALLEST_SCALE)
		  scale = SMALLEST_SCALE;
		pvd_set_scale(pvd_gui, (float64)scale);
	    }
	}
	else /* call_data->button == 3 */
	{
	    tactmap_screen_to_map(pvd_gui->tactmap,
				  call_data->screen_x, call_data->screen_y,
				  &map_x,&map_y, &map_c);
	    pvd_set_scale_index(pvd_gui, pvd_gui->current_scale_index-1);
	} 

	tactmap_set_center(pvd_gui->tactmap, (int32)map_x, (int32)map_y, 
			   map_c);
	pvd_update_scrollbars(pvd_gui);
	break;
    }
}

static void check_features(PVD_GUI_PTR pvd_gui, int32 index, int32 scale)
{
    uint32 mask;
    int32 mnr, mjr, grid;

    /* Set the bits for LINES and INDIVIDUAL trees appropriately
     * for this scale.
     */
    mask = pvd_gui->features[PVD_TREES].dfn.trees.mask;

    if (scale > pvd_gui->default_features->highest_tree_line)
      mask &= ~TACTMAP_TREE_LINES;
    else
      mask |= TACTMAP_TREE_LINES;

    if (scale > pvd_gui->default_features->highest_tree)
      mask &= ~TACTMAP_TREE_INDIVIDUAL;
    else
      mask |= TACTMAP_TREE_INDIVIDUAL;

    if (mask != pvd_gui->features[PVD_TREES].dfn.trees.mask)
    {
	pvd_gui->features[PVD_TREES].dfn.trees.mask = mask;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_TREES].obj,
			      &pvd_gui->features[PVD_TREES].dfn);
    }

    /* Use the same rules for wires that we use for treelines.
     */
    mask = pvd_gui->features[PVD_WIRE].dfn.wire.mask;
    if (scale > pvd_gui->default_features->highest_tree_line)
      mask &= ~TACTMAP_WIRE_LINES;
    else
      mask |= TACTMAP_WIRE_LINES;

    if (mask != pvd_gui->features[PVD_WIRE].dfn.wire.mask)
    {
	pvd_gui->features[PVD_WIRE].dfn.wire.mask = mask;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_WIRE].obj,
			      &pvd_gui->features[PVD_WIRE].dfn);
    }

    /* Use the same rules for dragon's teeth that we use for individual trees.
     */
    mask = pvd_gui->features[PVD_DRAGONS_TEETH].dfn.dragons_teeth.mask;
    if (scale > pvd_gui->default_features->highest_tree)
      mask &= ~TACTMAP_DRAGONS_TEETH_IND;
    else
      mask |= TACTMAP_DRAGONS_TEETH_IND;

    if (mask != pvd_gui->features[PVD_DRAGONS_TEETH].dfn.dragons_teeth.mask)
    {
	pvd_gui->features[PVD_DRAGONS_TEETH].dfn.dragons_teeth.mask = mask;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_DRAGONS_TEETH].obj,
			      &pvd_gui->features[PVD_DRAGONS_TEETH].dfn);
    }


    /* Use the same rules for MICLIC LINE that we use for treelines.
     */
    mask = pvd_gui->features[PVD_MICLIC_LINE].dfn.miclic_line.mask;
    if (scale > pvd_gui->default_features->highest_tree_line)
      mask &= ~TACTMAP_MICLIC_LINE_IND;
    else
      mask |= TACTMAP_MICLIC_LINE_IND;

    if (mask != pvd_gui->features[PVD_MICLIC_LINE].dfn.miclic_line.mask)
    {
	pvd_gui->features[PVD_MICLIC_LINE].dfn.miclic_line.mask = mask;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_MICLIC_LINE].obj,
			      &pvd_gui->features[PVD_MICLIC_LINE].dfn);
    }

    /* Use the same rules for MICLIC SCORCH that we use for treelines.
     */
    mask = pvd_gui->features[PVD_MICLIC_SCORCH].dfn.miclic_scorch.mask;
    if (scale > pvd_gui->default_features->highest_tree_line)
      mask &= ~TACTMAP_MICLIC_SCORCH_IND;
    else
      mask |= TACTMAP_MICLIC_SCORCH_IND;

    if (mask != pvd_gui->features[PVD_MICLIC_SCORCH].dfn.miclic_scorch.mask)
    {
	pvd_gui->features[PVD_MICLIC_SCORCH].dfn.miclic_scorch.mask = mask;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_MICLIC_SCORCH].obj,
			      &pvd_gui->features[PVD_MICLIC_SCORCH].dfn);
    }


    /* Set the contour intervals and grid spacing appropriately
     * for this scale.
     */
    if (pvd_gui->fixed_contours)
    {
	mnr = pvd_gui->fixed_minor_contour;
	mjr = mnr * pvd_gui->fixed_major_contour;
    }
    else
    {
	mnr = pvd_gui->default_features->scales[index]->minor_contour;
	mjr = pvd_gui->default_features->scales[index]->major_contour;
    }

    if ((pvd_gui->features[PVD_CONTOURS].dfn.contours.minor_spacing != mnr) ||
	(pvd_gui->features[PVD_CONTOURS].dfn.contours.major_spacing != mjr))
    {
	pvd_gui->features[PVD_CONTOURS].dfn.contours.minor_spacing = mnr;
	pvd_gui->features[PVD_CONTOURS].dfn.contours.major_spacing = mjr;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_CONTOURS].obj,
			      &pvd_gui->features[PVD_CONTOURS].dfn);
    }

    grid = pvd_gui->default_features->scales[index]->grid_interval;

    if (pvd_gui->features[PVD_GRIDS].dfn.grids.interval != grid)
    {
	pvd_gui->features[PVD_GRIDS].dfn.grids.interval = grid;
	tactmap_update_object(pvd_gui->tactmap,
			      pvd_gui->features[PVD_GRIDS].obj,
			      &pvd_gui->features[PVD_GRIDS].dfn);
    }
	
    
}

void pvd_set_scale_index(PVD_GUI_PTR pvd_gui, int32 index)
{
    float64 old_scale = pvd_gui->current_scale;

    if (index < 0)
      index = 0;
    else if (index >= pvd_gui->default_features->num_scales_plus_5-5)
      index = pvd_gui->default_features->num_scales_plus_5-6;

    /* Set the toggles */
    XmToggleButtonSetState(pvd_gui->scale_toggles[pvd_gui->
						  current_scale_index],
			   FALSE, FALSE);
    XmToggleButtonSetState(pvd_gui->scale_toggles[index],
			   TRUE, FALSE);

    pvd_gui->current_scale_index = index;
    pvd_gui->current_scale = pvd_gui->default_features->scales[index]->scale;
    check_features(pvd_gui, index, (int32)pvd_gui->current_scale);

    /* Maintain constant vehicle size */
    if (!pvd_gui->veh_mag_dont_adjust && (old_scale != 0.0))
      pvd_gui->veh_magnification *= pvd_gui->current_scale / old_scale;

    tactmap_set_scale(pvd_gui->tactmap, pvd_gui->current_scale);
    callback_fire_event(pvd_gui->zoom_event, (int32)pvd_gui->current_scale);
}

static char *scale_string(PVD_GUI_PTR pvd_gui, float64 scale)
{
    int32 i;
    int32 int_scale;
    static char buf[128];
    EDT_PVD_DEFAULTS dflt;

    edt_pvd_defaults(pvd_gui->gui, &dflt);

    if (dflt.non_standard_scales)
      int_scale = scale;
    else
    {
	/* Find the nearest scale which is at least as big as the desired
	 * scale.
	 */
	for (i=pvd_gui->default_features->num_scales_plus_5-6;i;i--)
	{
	    if (pvd_gui->default_features->scales[i]->scale >= scale)
	      break;
	}
	int_scale = pvd_gui->default_features->scales[i]->scale;
    }
    
    if (int_scale < 1000)
      sprintf(buf, "1 : %d", int_scale);
    else if (int_scale < 1000000)
      sprintf(buf, "1 : %d,%03d", int_scale/1000, int_scale%1000);
    else
      sprintf(buf, "1 : %d,%03d,%03d",
	      int_scale/1000000, (int_scale/1000)%1000, int_scale%1000);

    return buf;
}

void pvd_set_scale(PVD_GUI_PTR pvd_gui, float64 scale)
{
    int32 i;
    float64 old_scale = pvd_gui->current_scale;
    EDT_PVD_DEFAULTS dflt;

    edt_pvd_defaults(pvd_gui->gui, &dflt);

    /* Find the nearest scale which is at least as big as the desired
     * scale.
     */
    for (i=pvd_gui->default_features->num_scales_plus_5-6;i;i--)
    {
	if (pvd_gui->default_features->scales[i]->scale >= scale)
	  break;
    }

    if (dflt.non_standard_scales &&
	(pvd_gui->default_features->scales[i]->scale != scale))
    {
	XmToggleButtonSetState(pvd_gui->scale_toggles[pvd_gui->
						      current_scale_index],
			       FALSE, FALSE);
	pvd_gui->current_scale_index = i;
	pvd_gui->current_scale = scale;
	check_features(pvd_gui, i, (int32)pvd_gui->current_scale);

	/* Maintain constant vehicle size */
	if (!pvd_gui->veh_mag_dont_adjust && (old_scale != 0.0))
	  pvd_gui->veh_magnification *= pvd_gui->current_scale / old_scale;

	tactmap_set_scale(pvd_gui->tactmap, pvd_gui->current_scale);
	callback_fire_event(pvd_gui->zoom_event,(int32)pvd_gui->current_scale);
    }
    else
    {
	pvd_set_scale_index(pvd_gui, i);
    }
}

void pvd_remember_where(PVD_GUI_PTR pvd_gui)
{
    int32 i;

    /* Make room */
    if (pvd_gui->undo_count == UNDO_BUFFER_LENGTH-1)
    {
	for (i=0;i<UNDO_BUFFER_LENGTH-1;i++)
	  pvd_gui->undo_buffer[i] = pvd_gui->undo_buffer[i+1];
	pvd_gui->undo_count -= 1;
    }

    pvd_gui->undo_buffer[pvd_gui->undo_count].scale =
      pvd_gui->current_scale;
    tactmap_get_center(pvd_gui->tactmap,
		       &pvd_gui->undo_buffer[pvd_gui->undo_count].center_x,
		       &pvd_gui->undo_buffer[pvd_gui->undo_count].center_y,
		       &pvd_gui->undo_buffer[pvd_gui->undo_count].center_c);
    pvd_gui->undo_count += 1;
}

void pvd_unzoom(PVD_GUI_PTR pvd_gui)
{
    if (!pvd_gui->undo_count)
      XBell(XtDisplay(pvd_gui->map_widget), 100);
    else
    {
	pvd_gui->undo_count -= 1;
	pvd_set_scale(pvd_gui,
		      pvd_gui->undo_buffer[pvd_gui->undo_count].scale);

	tactmap_set_center(pvd_gui->tactmap,
			   (int32)pvd_gui->undo_buffer[pvd_gui->undo_count].
			   center_x,
			   (int32)pvd_gui->undo_buffer[pvd_gui->undo_count].
			   center_y,
			   pvd_gui->undo_buffer[pvd_gui->undo_count].
			   center_c);
	pvd_update_scrollbars(pvd_gui);
    }
}


/* This function returns the grid interval for the current scale of */
/* the pvd */
int32 pvd_get_grid_interval(PVD_GUI_PTR pvd_gui)
{    
    return
      (pvd_gui->default_features->scales[pvd_gui->current_scale_index]->grid_interval);  
}
#endif
/*USE_MOTIF*/
