#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libpvd                                   *
*       Created: Tue Sep 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libsafgui.h>
#include <libprivilege.h>
#include <libcoordinates.h>
#include <libcallback.h>
#include <libsensitive.h>
#include <libselect.h>
#include <libtactmap.h>
#include <libpvd.h>
#include <libworld.h>
#include <libgcs.h>
#include <libreader.h>
#include <stdext.h> /*common/include/global*/

#include <Xm/PushB.h>

static Widget toplevel;

static void tick_X();
static void init_environ_resolvers();
static void init_environ_defaults();

/* These should match what is in envinit.rdr!!! */
#define LOCAL 0
#define GLOBAL 1

static int32 get_tz();
static void init_environ_resolvers(void);
static void init_environ_defaults(char *data_path, uint32 flags);

static void quit()
{
    exit(0);
}

main(argc, argv)
    int32 argc;
    char *argv[];
{
    int32 hold_argc = argc;
    SGUI_PTR gui;
    XGCValues gcv;
    GC draw_gc, erase_gc;
    Widget w;
    SNSTVE_WINDOW_PTR sensitive;
    TACTMAP_PTR tactmap;
    SELECT_TOOL_PTR select;
    Window root;
    PVD_GUI_PTR pvd_gui;
    PRIV_PTR priv;
    char *name;
    SimulationAddress dummy;
    XtAppContext context;
    Display *display;
    READER_UNION *dummy_ru;
    CTDB *ctdb;
    COORD_TCC_PTR tcc;

    dummy.site = 0;
    dummy.host = 0;

    /* Initialize libtime */
    time_init(1000);

    /* Initailize libsched */
    sched_init(2, 10, 67);

    /* Initailize libprivilege */
    priv_init();
    priv = priv_create();


    /* Note that use of XtInitialize is now frowned upon, so this does
     * things the way the X11 manual say.
     */

    /* Tell X to look in the current directory for resources */
#ifdef USE_SETENV
    setenv("XUSERFILESEARCHPATH", "./%N", TRUE);
    setenv("XAPPLRESDIR", ".", TRUE);
#else
    putenv("XUSERFILESEARCHPATH=./%N");
    putenv("XAPPLRESDIR=.");
#endif

    XtToolkitInitialize();
    context = XtCreateApplicationContext();
    display = XtOpenDisplay(context, NULL /* Get the display from argv */,
			    argv[0],
			    "ModSAF",
			    NULL, 0, /* We do our own option parsing */
			    (int*)&argc, argv);

/*    toplevel = XtInitialize(argv[0], "../libsafgui/safgui.xrdb",
			    NULL, 0, &hold_argc, argv);
*/
    /* Each application may have multiple independents shells.  There
     * is no "main" shell, merely this unrealized parent which we are
     * about to make.
     */
    toplevel = XtAppCreateShell(name,
			        "ModSAF",
				applicationShellWidgetClass,
				display, NULL, 0);

    if (hold_argc != 2)
    {
	printf("Usage: %s <terrain directory>\n", argv[0]);
	name = NULL;
    }
    else name = argv[1];

    wld_load_playbox(1, 0, argc, argv, "../../data", NULL, NULL, &tcc, &ctdb);

    /* Make the sensitivity GCs */
    root = RootWindowOfScreen(XtScreen(toplevel));
    gcv.foreground = WhitePixelOfScreen(XtScreen(toplevel));
    gcv.line_width = 1;
    draw_gc = XCreateGC(XtDisplay(toplevel), root,
			GCForeground | GCLineWidth, &gcv);
    erase_gc = XCreateGC(XtDisplay(toplevel), root,
			 GCLineWidth, &gcv);
#if 0
    /* Read the terrain databases */
    load_terrain(name, &ctdb, &tcc);
#endif

    if (rdc_init("../../data", READER_OVERRIDE))
      printf("Error initializing librdrconst\n"), exit(0);

    edt_init("../../data", READER_OVERRIDE, FALSE);

    gui = sgui_init(toplevel, "PVD Test", priv, SGUI_FULL_DISPLAY);

    w = sgui_add_menubar(gui, PRIV_OPERATOR,
			 "File", "Quit", xmPushButtonWidgetClass);
    XtAddCallback(w, XmNactivateCallback, (XtCallbackProc)quit, NULL);

    /* Make the sensitive window */
    snstve_init();
    sensitive = snstve_create(sgui_map_widget(gui), draw_gc, erase_gc);

    /* Make the tactical map */
    tactmap_init();
    tactmap = tactmap_create(sgui_map_widget(gui), sensitive, erase_gc, NULL);
    tactmap_set_center(tactmap,
		       (ctdb->min_x+ctdb->max_x)/2, (ctdb->min_y+ctdb->max_y)/2,
		       GCS_ILLEGAL_CELL);

    select_init();
    select = select_create(gui, tactmap, erase_gc, sensitive);

    disdbobj_init();

    /*
     * Ignore reader file if only the constant model is running.
     * Shouldn't turn off -everything-, so at least constant model 
     * should be running.  env_init() will ignore the file if passed
     * a NULL main_options.data.value.  Yes, side effects like this are gross.
     */

    env_init(".", READER_OVERRIDE, ctdb);
/*    envinit_init(".", READER_OVERRIDE,0, 67,
		 tcc, ctdb, NULL, 0, 0, 3);
  */  
/*    eui_init();*/

    /* Initialize all environmental models here.
     * Note that there is an implicit priority
     * scheme that results in the last model set
     * registered having the least priority when
     * conflicts arise.
     *
     * The constant model (econst) should always
     * be last, and must be outside the command-line test.
     */


/*    if (!main_options.envconstant.value)*/
    {
/*	wthr_init(NULL, simulator_LL_SAFSIM); */

/*	envsea_init(ctdb, TRUE); */

	aeenv_init("../../data", READER_OVERRIDE);

	aeflare_init(&ctdb, "../../data", READER_OVERRIDE);
	/* We tried to do it simply by initializing libvisual, but */
	/* it depended on too many things.  Instead, the only */
	/* thing we need is visual.rdr read in, so just do that right here. */
	reader_read("visual.rdr", "../../data", &dummy_ru,
		    READER_OVERRIDE & ~READER_TYPING);

 	envatm_init("../../data", READER_OVERRIDE);

	esimp_init("../../data", READER_OVERRIDE, ctdb);
	
	envilm_init("../../data", READER_OVERRIDE, ctdb); 

	econt_init("../../data", READER_OVERRIDE, ctdb);

	/* smoke intervisibility (combic) */

	smkint_init(&ctdb);

	/* smoke sim  (combic) */

/*
	sms_init(po_db, "../../data", READER_OVERRIDE, 
		 1, &ctdb); 
*/
	/* flare sim */
/*
	 fls_init(po_db, "../../data", READER_OVERRIDE, &ctdb);
*/

	/* set up simple multiply/add/etc resolvers */
	init_environ_resolvers();

    }
    econst_init("../../data", READER_OVERRIDE);

    /* Load default values for environmental queries */
    init_environ_defaults("../../data", READER_OVERRIDE);


    
    printf("Environmental models:\n");
    env_print_registered_models();

    printf("Requestable environmental effects:\n");
    env_print_registered_effects();

    /* Make the PVD */
    pvd_init(".", READER_OVERRIDE,0,0);
    pvd_gui = pvd_init_gui(gui, tactmap,
		 sgui_map_widget(gui), erase_gc,
		 sensitive, select, 0, NULL, &dummy);

    envedit_init_gui("../../data", READER_OVERRIDE, gui, tactmap,
		     tcc, erase_gc, sensitive, 0, NULL, &ctdb);

    printf("The current interval = %d\n",pvd_get_grid_interval(pvd_gui));

    /* Realize the GUI */
    sgui_realize(gui);

    /* Get rid of the message log and editor */
    sgui_message_log(gui, FALSE);
    sgui_editor(gui, FALSE);

    /* Realize the GUI */
    sgui_realize(gui,0);

    /* Post the map tick */
    sched_periodic_fncl(tactmap_tick, 1000, 67, 0,
			A_PTR, tactmap,
			A_INT, 67,
			A_END);

    /* Post the X tick */
    sched_periodic_fncl(tick_X, 0, 10, 0,
			A_PTR, XtWidgetToApplicationContext(toplevel),
			A_END);

    /* Star the scheduler */
    sched_invoke_functions_until(0xFFFFFFFF);
}

static void tick_X(context)
    XtAppContext context;
{
    XEvent event;
    int pend;

    while (pend = XtAppPending(context))
    {
	if (pend & (XtIMXEvent  | XtIMTimer))
	{
	    XtAppProcessEvent(context, pend);
	}
	else
	  break;
    }
}

static void load_terrain(name, ctdb, tccp)
    char          *name;
    CTDB          *ctdb;
    COORD_TCC_PTR *tccp;
{
    char path_name[256];

    if (name)
    {
	if ((name[0] == '/') || (name[0] == '.'))
	  strcpy(path_name, name);
	else
	  sprintf(path_name, "../../terrain/%s/%s.ctdb", name, name);

	ctdb_read(path_name, ctdb, 10*1024*1024);
    }
    else
      ctdb_read(NULL, ctdb, 0);

    coord_init("../../data", 0);
    *tccp = coord_define_tcc(COORD_UTM_NE,
			     ctdb->origin_northing,
			     ctdb->origin_easting,
			     ctdb->origin_zone_number,
			     ctdb->origin_zone_letter,
			     ctdb->datum,
			     ctdb->max_x, ctdb->max_y);
}

static void init_environ_resolvers(void)
{
    int32     md_mult, md_vadd, md_priority, md_stat_print, md_surf_zone;
    
    md_mult = env_register_model("multiply", NULL);
    
    env_register_tag(md_mult,
		     ENV_RAY_VISIBILITY,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER) env_vmm_multiply,
		     NULL);

    md_vadd = env_register_model("vadd", NULL);
    
    env_register_tag(md_vadd,
		     ENV_ILLUMINATION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER) env_vmm_vadd,
		     NULL);
    env_register_tag(md_vadd,
		     ENV_SOLAR_ILLUMINATION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER) env_vmm_vadd,
		     NULL);
    env_register_tag(md_vadd,
		     ENV_LUNAR_ILLUMINATION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER) env_vmm_vadd,
		     NULL);
    env_register_tag(md_vadd,
		     ENV_SKY_ILLUMINATION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER) env_vmm_vadd,
		     NULL);
    
    /*
     * Surf zone resolver.  For sea tags that change in the surf zone.
     *
    ENV_SEA_SURFACE_HEIGHT,
    ENV_SEA_SURFACE_TEMP
    ENV_SEA_WAVE_DIRECTION,
    ENV_SEA_WAVE_PERIOD,
    ENV_SEA_WAVE_HEIGHT,
    ENV_SEA_CURRENT_VELOCITY,
    ENV_SEA_BREAKER_ANGLE,
    ENV_SEA_BREAKER_TYPE,
    ENV_SEA_IN_SURF_ZONE, 
     */

    md_surf_zone = env_register_model("surf_zone", NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_SURFACE_TEMP,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_CURRENT_VELOCITY,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_WAVE_DIRECTION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_WAVE_PERIOD,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_WAVE_HEIGHT,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_BREAKER_ANGLE,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_BREAKER_TYPE,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_IN_SURF_ZONE,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    env_register_tag(md_surf_zone,
		     ENV_SEA_SURFACE_HEIGHT,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_surf_zone,
		     NULL);

    md_priority = env_register_model("priority", NULL);

    env_register_tag(md_priority,
		     ENV_SIM_TIME,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_use_priority,
		     NULL);

    env_register_tag(md_priority,
		     ENV_SUN_POSITION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_use_priority,
		     NULL);

    env_register_tag(md_priority,
		     ENV_MOON_POSITION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_use_priority,
		     NULL);

    env_register_tag(md_priority,
		     ENV_MOON_ELONGATION,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_use_priority,
		     NULL);

    env_register_tag(md_priority,
		     ENV_SKY_OVER_GROUND,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER)env_use_priority,
		     NULL);

    md_stat_print = env_register_model("statistics", NULL);
    env_register_tag(md_stat_print,
		     ENV_INTERNAL_STATS,
		     ENV_RESOLVER_MODEL,
		     (ENV_HANDLER) env_print_internal_stats,
		     NULL);

}

static void init_environ_defaults(char *data_path, uint32 flags)
{
    int32		md;
    ENV_TAG		tag;
    READER_UNION	db;
    READER_UNION	*field;
    TimeDateGroup       tdg;
    ENV_INTERFACE	ei;
    int32               time_zone_offset_in_seconds;
    char                timestring[256];
    char                firsthalf[256];
    char                *secondhalf, *finalstring;

    int32 ret = reader_read("envinit.rdr", data_path, &db,
			    flags & ~READER_TYPING);
    if (ret != 0) 
    {
	printf("error reading envinit.rdr\n");
	return;
    }

    /* Set BATCH mode.  We won't FLUSH here, and will instead require
     * some other chunk of code to do this.  This keeps anything from
     * going out on the network before the user actually requests it.
     */
    env_set(ENV_BATCH, NULL);
    
    for (tag = 0; tag < ENV_NUM_TAGS; tag++)
    {
	bzero(&ei, sizeof(ENV_INTERFACE));
	
	if (field = reader_find_tag(reader_get_symbol(env_tag_name(tag)),
				    db.array, READER_TAGGED,
				    READER_NO_ERRORS))
	{
	    switch(tag)
	    {
	      case ENV_INTERNAL_STATS:
		break;
	      case ENV_BAROMETRIC_PRESSURE:
		/* Barometric pressure is input in atmospheres, 
		 * but ENV_INTERFACE uses pascals.
		 */
		ei.u.barometric_pressure.output_value = 
		  (float64)field[2].real / ENV_ATMS_PER_PASCAL;
		break;
	      case ENV_TEMPERATURE:
	      case ENV_DEWPOINT:
	      case ENV_RELATIVE_HUMIDITY:
	      case ENV_PRECIPITATION_RATE:
	      case ENV_EXTINCTION_AMOUNT:
		ei.u.generic_vector_to_value.output_value = field[2].real;
		break;
	      case ENV_PRECIPITATION_TYPE:
	      case ENV_EXTINCTION_TYPE:
	      case ENV_CLOUD_TYPE:
		ei.u.generic_vector_to_type.output_type = field[2].integer;
		break;
	      case ENV_SUN_POSITION:
	      case ENV_MOON_POSITION:
	      case ENV_WIND_VELOCITY:
		ei.u.generic_vector_to_vector.output_vector[X]
		  = field[2].array[1].real;
		ei.u.generic_vector_to_vector.output_vector[Y]
		  = field[2].array[2].real;
		ei.u.generic_vector_to_vector.output_vector[Z]
		  = field[2].array[3].real;
		ei.u.generic_vector_to_vector.output_max_mag
		  = fsqrt((field[2].array[1].real * field[2].array[1].real)
			  + (field[2].array[2].real * field[2].array[2].real)
			  + (field[2].array[3].real * field[2].array[3].real));
		ei.u.generic_vector_to_vector.output_min_mag
		  = ei.u.generic_vector_to_vector.output_max_mag;
		break;
	      
	      case ENV_ILLUMINATION:
	      case ENV_SOLAR_ILLUMINATION:
	      case ENV_LUNAR_ILLUMINATION:
	      case ENV_SKY_ILLUMINATION:
		ei.u.visual_check_to_vector.output_vector[X] 
		  = field[2].array[1].real;
		ei.u.visual_check_to_vector.output_vector[Y] 
		  = field[2].array[2].real;
		ei.u.visual_check_to_vector.output_vector[Z] 
		  = field[2].array[3].real;
       		break;
		
	      case ENV_RAY_VISIBILITY:
	      case ENV_CONTRAST:
	      case ENV_VISUAL_RANGE:
       	      case ENV_SKY_OVER_GROUND:
	      case ENV_CLOUD_COVER:
	      case ENV_MOON_ELONGATION:
		ei.u.generic_value.output_value = field[2].real;
		ei.u.generic_value.output_max = field[2].real;
		ei.u.generic_value.output_min = field[2].real;
		break;

	      case ENV_CLOUD_CEILING:
	      case ENV_CLOUD_HEIGHT:
		ei.u.generic_vector_to_value.output_value = field[2].real;
		ei.u.generic_vector_to_value.output_max = field[2].real;
		ei.u.generic_vector_to_value.output_min = field[2].real;
		break;

	      case ENV_SEA_SALINITY: case ENV_SEA_TIDE_LEVEL:
	      case ENV_SEA_SURFACE_HEIGHT: case ENV_SEA_SURFACE_TEMP:
	      case ENV_SEA_WAVE_DIRECTION: case ENV_SEA_WAVE_PERIOD:
	      case ENV_SEA_WAVE_HEIGHT: case ENV_SEA_BREAKER_ANGLE:
		ei.u.generic_vector_to_value.output_value = field[2].real;
		break;

	      case ENV_SEA_STATE: case ENV_SEA_NUM_WAVES:
	      case ENV_SEA_NUM_FREQUENCIES: case ENV_SEA_BREAKER_TYPE:
	      case ENV_SEA_IN_SURF_ZONE:
		ei.u.generic_vector_to_type.output_type = field[2].integer;
		break;
	     
	      case ENV_SEA_CURRENT_VELOCITY:
		ei.u.generic_vector_to_vector.output_vector[X]
		  = field[2].array[1].real;
		ei.u.generic_vector_to_vector.output_vector[Y]
		  = field[2].array[2].real;
		ei.u.generic_vector_to_vector.output_vector[Z]
		  = field[2].array[3].real;
		ei.u.generic_vector_to_vector.output_max_mag
		  = fsqrt((field[2].array[1].real * field[2].array[1].real)
			  + (field[2].array[2].real * field[2].array[2].real)
			  + (field[2].array[3].real * field[2].array[3].real));
		ei.u.generic_vector_to_vector.output_min_mag
		  = ei.u.generic_vector_to_vector.output_max_mag;
		break;

              /* extinction coefficient is a little wierd.  its wavelength
               * sensitive, so we accept multiple initializations.
               */
	      case ENV_EXTINCTION_COEFF:
		ei.u.extinction_coeff.input_wavelength
		  = field[2].array[1].integer;
		ei.u.extinction_coeff.output_value
		  = field[2].array[2].real;
		break;
	      case ENV_SIM_TIME:

		strcpy(timestring, field[2].charptr);
		/* If there was no time zone specified, insert the string "UTC" */
		/* The time zone string will be after the time. First look for */
		/* the colon in the time */
		finalstring = timestring;
		secondhalf = strchr(timestring, ':');
		/* Copy the string up to and including the time and the space after*/
		secondhalf+=4; 
		strncpy(firsthalf, timestring, (secondhalf-timestring));
		/* Check if the next characters are alpha characters */
		if (!isalpha((int)secondhalf[0]))
		    {
			/* Time zone letters are not included. Insert "UTC" */
			strcat(firsthalf, "UTC ");
			strcat(firsthalf, secondhalf);
			finalstring = firsthalf;
		    }
		    
		ei.u.sim_time.uxtime = 
		  time_get_full_date_time(finalstring);

		/* If the last field is LOCAL, the time should be taken */
		/* to be local to the database. */
		if (field[3].integer == LOCAL)
		{
		    /* Figure out roughly what timezone this database */
		    /* is in and convert the time to be the same time */
		    /* there. */
		    time_zone_offset_in_seconds = get_tz() * 3600;
		    ei.u.sim_time.uxtime -= time_zone_offset_in_seconds;
		}
		
		time_tdg_from_seconds(ei.u.sim_time.uxtime, &tdg);
		ei.u.sim_time.date = tdg.date;
		ei.u.sim_time.time = tdg.time;
		break;
	    }
	    env_set(tag, &ei);
	}	
    }
}


static int32 get_tz()
{

  float64 lat, lon, elev;

  if(coord_convert(COORD_GCS, GCS_ILLEGAL_CELL,
		   0.0, 0.0, 0.0,
		   COORD_LATLON, &lat, &lon, &elev, FALSE))
  return(0);

  else
    {
      if ( lon >= 0.0 )
	return((int32)((lon+7.5)/15.0));
      else
	return((int32)((lon-7.5)/15.0));
    }
}

