/*
 * $RCSfile: libqueue_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libqueue_local.h                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Libqueue (formerly libcomm) private declarations          *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libqueue.h"
#ifdef MULTIPROCESSOR
#include <liblock.h>
#elif defined(USE_ASYNCIO)
#include <signal.h>
#endif

#ifdef MULTIPROCESSOR
#define QLOCKINIT(qp)	lock_init (&(qp->lock))
#elif defined(USE_ASYNCIO)
#define QLOCKINIT(qp)	qp->old_signal = (int (*)())SIG_IGN
#else
#define QLOCKINIT(qp)	
#endif

#ifdef MULTIPROCESSOR
#define QLOCK(qp)	lock_wait (&(qp->lock))
#elif defined(USE_ASYNCIO)
#define QLOCK(qp)	qp->old_signal = (int (*)()) signal(SIGIO, SIG_IGN)
#else
#define QLOCK(qp)	
#endif

#ifdef MULTIPROCESSOR
#define QUNLOCK(qp)	lock_unlock (&(qp->lock))
#elif defined(USE_ASYNCIO)
#define QUNLOCK(qp)	signal(SIGIO, qp->old_signal)
#else
#define QUNLOCK(qp)	
#endif

/*
 * Simple, fixed-length, no-insertion queue.
 *
 * From this, we build buffer pools.
 */

typedef struct queue_simple
{
#ifdef MULTIPROCESSOR
    LOCK_TYPE lock;
#elif defined(USE_ASYNCIO)
    int (*old_signal)();
#endif
    uint32 queue_length_mask;
    uint32 dequeue_this;
    uint32 enqueue_here;
    struct queue_buffer **queue;
} QUEUE_SIMPLE;

typedef struct queue_standard
{
#ifdef MULTIPROCESSOR
    LOCK_TYPE lock;
#elif defined(USE_ASYNCIO)
    int (*old_signal)();
#endif
    struct queue_element *first;
    struct queue_element *last;
    int32 length;
} QUEUE_STANDARD;

typedef struct queue_buffer
{
    QUEUE_BUFFER_POOL free_pool;
    int32 refcnt;
    int32 length;
    int32 is_a_queue;
} QUEUE_BUFFER;

#ifndef PARANOID
#define char_from_buffer(bp) \
  ((char *)((char *)bp + sizeof(QUEUE_BUFFER)))
#define buffer_from_char(p)  \
  ((QUEUE_BUFFER *)(((char *)(p)) - sizeof(QUEUE_BUFFER)))
#endif

typedef struct queue_element
{
    struct queue_element *prev;
    struct queue_element *next;
    struct queue_buffer *data;
} QUEUE_ELEMENT;

struct queue_queue
{
    int32 is_standard;
    union
    {
	QUEUE_SIMPLE simple;
	QUEUE_STANDARD standard;
    }u;
};

/* Private function prototypes */
#ifdef PARANOID
extern char *char_from_buffer();
extern QUEUE_BUFFER *buffer_from_char();
#endif
extern void          queue_buffer_one_more(void);
extern void          queue_buffer_one_less(void);
extern void          queue_buffer_pool_return(QUEUE_BUFFER *bp);
extern QUEUE_QUEUE   queue_simple_allocate(uint32 size);
extern QUEUE_BUFFER *queue_simple_dequeue(QUEUE_SIMPLE *sqp);
extern QUEUE_BUFFER *queue_simple_stack_pop(QUEUE_SIMPLE *sqp);
extern void          queue_simple_describe(QUEUE_SIMPLE *sqp);
extern int32           queue_simple_enqueue(QUEUE_SIMPLE *sqp,
					    QUEUE_BUFFER *bp);
extern int32           queue_simple_stack_push(QUEUE_SIMPLE *sqp,
					       QUEUE_BUFFER *bp);
extern QUEUE_BUFFER *queue_simple_excise(QUEUE_SIMPLE *sqp, QUEUE_BUFFER *p);
extern void          queue_simple_flush(QUEUE_SIMPLE *sqp);
extern int32           queue_simple_is_full(QUEUE_SIMPLE *sqp);
extern int32           queue_simple_length(QUEUE_SIMPLE *sqp);
extern int32           queue_simple_traverse_and_apply(QUEUE_SIMPLE *sqp,
						       QUEUE_FUNCTION fcn,
						       QUEUE_FUNCTION_ARG arg);
extern QUEUE_BUFFER *queue_simple_peek(QUEUE_SIMPLE *sqp);
extern QUEUE_QUEUE   queue_standard_allocate(void);
extern QUEUE_BUFFER *queue_standard_dequeue(QUEUE_STANDARD *qp);
extern void          queue_standard_describe(QUEUE_STANDARD *qp);
extern void          queue_standard_enqueue(QUEUE_STANDARD *qp,
					    QUEUE_BUFFER *bufp);
extern QUEUE_BUFFER *queue_standard_excise(QUEUE_STANDARD *qp,
					   QUEUE_BUFFER *p);
extern void          queue_standard_flush(QUEUE_STANDARD *qp);
extern int32           queue_standard_length(QUEUE_STANDARD *qp);
extern int32
queue_standard_traverse_and_apply(QUEUE_STANDARD *qp,
				  QUEUE_FUNCTION fcn,
				  QUEUE_FUNCTION_ARG arg);
extern QUEUE_BUFFER *queue_standard_peek(QUEUE_STANDARD *qp);
