#ifndef lint
static char rcsid [] = "$RCSfile: que_buffer.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: que_buffer.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Handles buffer operations                                 *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Original source code by jm.                                *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libqueue_local.h"
#include <stdalloc.h> /*common/include/global*/
#include <stdlib.h>

#ifdef PARANOID

/*
 * For computing pushback.
 */
char *char_from_buffer(QUEUE_BUFFER *bp)
{
    char *p = ((BYTE_POINTER)bp + sizeof(QUEUE_BUFFER));
    return(p);
}

QUEUE_BUFFER *buffer_from_char(char *p)
{
    QUEUE_BUFFER *bp = (QUEUE_BUFFER *)(p - sizeof(QUEUE_BUFFER));
    return(bp);
}
#endif

ADDRESS queue_buffer_allocate (uint32 size)
{
    QUEUE_BUFFER *bp = (QUEUE_BUFFER *)STDALLOC(size + sizeof(QUEUE_BUFFER));

    if (!bp)
    {
#ifdef USE_STDERR
	fprintf (stderr, "LIBQUEUE ERROR: Memory allocation failed\n");
#endif /* USE_STDERR */
	abort ();
    }

    bp->free_pool = (QUEUE_BUFFER_POOL)NULL;
    bp->refcnt = 1;
    bp->length = size;
    bp->is_a_queue = 0;
    queue_buffer_one_more();
    return(char_from_buffer(bp));
}

int32 queue_buffer_deallocate(ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);
    int32 oldrefcnt;

#ifdef MULTIPROCESSOR
    oldrefcnt = atomadd32(&(bp->refcnt), -1);
#else
    oldrefcnt = bp->refcnt;
    (bp->refcnt)--;
#endif

    switch(oldrefcnt)
    {
      case 0:
#ifdef USE_STDERR
	fprintf(stderr, "LIBQUEUE WARNING: Attempt to re-deallocate buffer!\n");
#endif /* USE_STDERR */
	return 0;
      case 1:
	queue_buffer_one_less();
	/*
	 * KLUDGE ALERT.
	 * Really, all buffers should use the same mechanism.
	 */
	if (bp->free_pool)
	{
	    queue_buffer_pool_return(bp);
	}
	else
	{
	    STDDEALLOC(bp);
	}
	return 1;
      default:
	return 0;
    }
}

/* Returns 1 if the buffer was actually free */
void queue_deallocate(QUEUE_QUEUE queue)
{
    QUEUE_BUFFER *bp = buffer_from_char(((char *)queue));
    int32 oldrefcnt;

#ifdef MULTIPROCESSOR
    oldrefcnt = atomadd32(&(bp->refcnt), -1);
#else
    oldrefcnt = bp->refcnt;
    (bp->refcnt)--;
#endif

    switch(oldrefcnt)
    {
      case 0:
#ifdef USE_STDERR
	fprintf(stderr, "LIBQUEUE WARNING: Attempt to re-deallocate queue!\n");
#endif /* USE_STDERR */
	break;
      case 1:
	queue_buffer_one_less();

	/* Flush empty queues before freeing their memory */
	queue_flush(queue);
	STDDEALLOC(bp);
	break;
      default:
	break;
    }
}

int32 queue_buffer_length(ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);
    return(bp->length);
}

void queue_buffer_up_refcnt(ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);
#ifdef MULTIPROCESSOR
    atomadd32(&(bp->refcnt), 1);
#else
    (bp->refcnt)++;
#endif
}

int32 queue_buffer_get_refcnt(ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);
    return bp->refcnt;
}

void queue_buffer_describe(ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);

    printf("BUFFER: 0x%X, p = 0x%X\n", bp, p);
    printf("{\n");
    printf("    refcnt = 0x%X\n", bp->refcnt);
    printf("}\n");
}

static int32 *buffers_allocated = (int32 *)0;

static void buffer_statistics_init(void)
{
    buffers_allocated = (int32 *)STDALLOC(sizeof(int));
    *buffers_allocated = 0;
}

void queue_buffer_one_more(void)
{
    if (!buffers_allocated)
      buffer_statistics_init();

    (*buffers_allocated)++;
}

void queue_buffer_one_less(void)
{
    if (!buffers_allocated)
      buffer_statistics_init();

    (*buffers_allocated)--;
}

void queue_print_statistics(void)
{
    if (!buffers_allocated)
      buffer_statistics_init();

    printf("Buffers in use : %d\n", *buffers_allocated);
}
