#ifndef lint
static char rcsid [] = "$RCSfile: que_bufpool.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: que_bufpool.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Buffer pool operations                                    *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libqueue_local.h"

ADDRESS queue_buffer_allocate_from_pool(QUEUE_BUFFER_POOL bufpool)
{
    QUEUE_BUFFER *bufp = queue_simple_stack_pop(&bufpool->u.simple);

    if (bufp)
    {
	bufp->refcnt = 1;
	queue_buffer_one_more();
	return(char_from_buffer(bufp));
    }
    else
    {
	return (ADDRESS )0;
    }
}

void queue_buffer_pool_return(QUEUE_BUFFER *bp)
{
    (void)queue_simple_stack_push(&bp->free_pool->u.simple, bp);
}

static void buffer_allocate_into_pool(QUEUE_BUFFER_POOL bufpool, uint32 size)
{
    ADDRESS p = queue_buffer_allocate(size);
    QUEUE_BUFFER *bp = buffer_from_char(p);

    /* To undo the effect of the above buffer_allocate()... */
    queue_buffer_one_less();

    bp->free_pool = bufpool;
    queue_buffer_pool_return(bp);
}

QUEUE_BUFFER_POOL queue_buffer_pool_allocate(uint32 size, uint32 number)
{
    QUEUE_QUEUE queue = queue_simple_allocate(number);
    QUEUE_BUFFER_POOL bufpool = queue;
    int32 i;

    for (i=0;i<number;i++)
    {
	buffer_allocate_into_pool(bufpool,size);
    }

    return(bufpool);
}

int32 queue_buffer_pool_deallocate(QUEUE_BUFFER_POOL bufpool)
{
    QUEUE_BUFFER *bufp;

    /* See if the pool is full (no buffers are outstanding).
     */
    if (!queue_simple_is_full(&bufpool->u.simple))
    {
#ifdef USE_STDERR
	fprintf(stderr, "LIBQUEUE WARNING: Could not deallocate buffer pool (outstanding buffers).\n");
#endif /* USE_STDERR */
	return 0;
    }

    /* Free the buffers in the pool */
    while (bufp = queue_simple_stack_pop(&bufpool->u.simple))
    {
	/* Fix it so the buffer won't just go back into the pool
	 * when we deallocate it.
	 */
	bufp->free_pool = 0;

	bufp->refcnt = 1;
	queue_buffer_deallocate(char_from_buffer(bufp));
	queue_buffer_one_more(); /* To compensate for the deallocate */
    }

    /* Now deallocate the pool */
    queue_deallocate((QUEUE_QUEUE)bufpool);

    return 1;
}
