#ifndef lint
static char rcsid [] = "$RCSfile: que_queue.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: que_queue.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Handles queue operations                                  *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Original source code by jm.                                *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libqueue_local.h"

QUEUE_QUEUE queue_allocate(uint32 size)
{
    QUEUE_BUFFER *bp;
    ADDRESS q;

    if (size)
      q = queue_simple_allocate(size);
    else
      q = queue_standard_allocate();

    bp = buffer_from_char(q);
    bp->is_a_queue = 1;
    return q;
}

void queue_flush(QUEUE_QUEUE queue)
{
    if (queue->is_standard)
      queue_standard_flush(&queue->u.standard);
    else
      queue_simple_flush(&queue->u.simple);
}

int32 queue_enqueue(QUEUE_QUEUE qp, ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);

#ifdef MULTIPROCESSOR
    atomadd32(&(bp->refcnt), 1);
#else
    (bp->refcnt)++;
#endif

    if (qp->is_standard)
    {
	queue_standard_enqueue(&qp->u.standard, bp);
	return 1; /* Standard queues always succeed */
    }
    else
    {
	return queue_simple_enqueue(&qp->u.simple, bp);
    }
}

ADDRESS queue_dequeue(QUEUE_QUEUE qp)
{
    QUEUE_BUFFER *bp;

    if (qp->is_standard)
      bp = queue_standard_dequeue(&qp->u.standard);
    else
      bp = queue_simple_dequeue(&qp->u.simple);

    if (bp == NULL)
    {
	return(NULL);
    }
    else
    {
	char *p = char_from_buffer(bp);
	return((ADDRESS)p);
    }
}

ADDRESS queue_peek(QUEUE_QUEUE qp)
{
    QUEUE_BUFFER *bp;

    if (qp->is_standard)
      bp = queue_standard_peek(&qp->u.standard);
    else
      bp = queue_simple_peek(&qp->u.simple);

    if (bp == NULL)
    {
	return(NULL);
    }
    else
    {
	char *p = char_from_buffer(bp);
	return((ADDRESS)p);
    }
}

ADDRESS queue_excise(QUEUE_QUEUE qp, ADDRESS p)
{
    QUEUE_BUFFER *bp = buffer_from_char(p);
    QUEUE_BUFFER *btmp;

    if (qp->is_standard)
      btmp = queue_standard_excise(&qp->u.standard, bp);
    else
      btmp = queue_simple_excise(&qp->u.simple, bp);

    if (btmp == NULL)
    {
#ifdef USE_STDERR
	fprintf(stderr,	"LIBQUEUE WARNING: Buffer not found in queue\n");
#endif /* USE_STDERR */
	return(NULL);
    }
    else
    {
	return char_from_buffer(btmp);
    }
}

int32 queue_length(QUEUE_QUEUE queue)
{
    if (queue->is_standard)
      return queue_standard_length(&queue->u.standard);
    else
      return queue_simple_length(&queue->u.simple);
}

int32 queue_traverse_and_apply(QUEUE_QUEUE queue,
			       QUEUE_FUNCTION fcn,
			       QUEUE_FUNCTION_ARG arg)
{
    if (queue->is_standard)
      return queue_standard_traverse_and_apply(&queue->u.standard, fcn, arg);
    else
      return queue_simple_traverse_and_apply(&queue->u.simple, fcn, arg);
}

void queue_distribute(QUEUE_QUEUE qp, QUEUE_FUNCTION_ARG p)
{
    (void)queue_traverse_and_apply(qp, (QUEUE_FUNCTION)queue_enqueue, p);
}

void queue_describe(QUEUE_QUEUE queue)
{
    if (queue->is_standard)
      queue_standard_describe(&queue->u.standard);
    else
      queue_simple_describe(&queue->u.simple);
}


int32 queue_full(QUEUE_QUEUE queue)
{
   if (!queue->is_standard)
   {
      return (queue_simple_is_full(&queue->u.simple));
   }
   return (0);
}
