#ifndef lint
static char rcsid [] = "$RCSfile: que_simple.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: que_simple.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Simple queue routines                                     *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Original source code by jm.                                *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "libqueue_local.h"
#include <stdalloc.h> /*common/include/global*/

#define MOD_QUEUE_SIZE(sqp,n)	((n) & (sqp)->queue_length_mask)
#define SQDQ(sqp)		MOD_QUEUE_SIZE(sqp,(sqp)->dequeue_this)
#define SQNQ(sqp)		MOD_QUEUE_SIZE(sqp,(sqp)->enqueue_here)
#define SQSPOP(sqp)		MOD_QUEUE_SIZE(sqp,(sqp)->enqueue_here)
#define SQSPUSH(sqp)		MOD_QUEUE_SIZE(sqp,(sqp)->enqueue_here)

/*
 * These macros can only be used while the queue is locked!
 */
#define SQLENGTH(sqp)	(sqp->enqueue_here - sqp->dequeue_this)
#define SQFULL(sqp)	(SQLENGTH(sqp) == (sqp->queue_length_mask + 1))
#define SQEMPTY(sqp)	(SQLENGTH(sqp) == 0)

int32 queue_simple_enqueue(QUEUE_SIMPLE *sqp, QUEUE_BUFFER *bp)
{
    int32 ret;

    QLOCK(sqp);

    if (SQFULL(sqp))
    {
#ifdef MULTIPROCESSOR
	atomadd32(&(bp->refcnt), -1);
#else
	(bp->refcnt)--;
#endif
	ret = 0;
    }
    else
    {
	uint32 index = SQNQ(sqp);

	sqp->queue[index] = bp;
	sqp->enqueue_here++;
	ret = 1;
    }

    QUNLOCK(sqp);
    return(ret);
}

/* almost a copy of queue_simple_enqueue() */
int32 queue_simple_stack_push(QUEUE_SIMPLE *sqp, QUEUE_BUFFER *bp)
{
    int32 ret;

    QLOCK(sqp);

    if (SQFULL(sqp))
    {
#ifdef MULTIPROCESSOR
	atomadd32(&(bp->refcnt), -1);
#else
	(bp->refcnt)--;
#endif
	ret = 0;
    }
    else
    {
	uint32 index = SQSPUSH(sqp);

	sqp->queue[index] = bp;
	sqp->enqueue_here++;
	ret = 1;
    }

    QUNLOCK(sqp);
    return(ret);
}

QUEUE_BUFFER *queue_simple_stack_pop(QUEUE_SIMPLE *sqp)
{
    QUEUE_BUFFER *ret;

    QLOCK(sqp);

    if (SQEMPTY(sqp))
    {
	ret = NULL;
    }
    else
    {
	uint32 index;
	QUEUE_BUFFER *item;

	/* enqueue points at the next empty spot for
	 * enquing, so decrement it before using it to point
	 * at the top buffer that is there.
	 */
	sqp->enqueue_here--;
	index = SQSPOP(sqp);
	item = sqp->queue[index];

	ret = item;
    }

    QUNLOCK(sqp);
    return(ret);
}

QUEUE_BUFFER *queue_simple_dequeue(QUEUE_SIMPLE *sqp)
{
    QUEUE_BUFFER *ret;

    QLOCK(sqp);

    if (SQEMPTY(sqp))
    {
	ret = NULL;
    }
    else
    {
	uint32 index = SQDQ(sqp);
	QUEUE_BUFFER *item = sqp->queue[index];

	sqp->dequeue_this++;
	ret = item;
    }

    QUNLOCK(sqp);
    return(ret);
}

QUEUE_BUFFER *queue_simple_peek(QUEUE_SIMPLE *sqp)
{
    QUEUE_BUFFER *ret;

    if (SQEMPTY(sqp))
    {
	ret = NULL;
    }
    else
    {
	uint32 index = SQDQ(sqp);
	QUEUE_BUFFER *item = sqp->queue[index];

	ret = item;
    }

    return(ret);
}

QUEUE_BUFFER *queue_simple_excise(QUEUE_SIMPLE *sqp, QUEUE_BUFFER *p)
{
    QUEUE_BUFFER *ret = NULL;
    uint32 index, i;

    QLOCK(sqp);

    if (!(SQEMPTY(sqp)))
    {
	for (i = 0; i < SQLENGTH(sqp); i++)
	{
	    index = MOD_QUEUE_SIZE(sqp,(SQDQ(sqp) + i));
	    if (p == sqp->queue[index])
	    {
		uint32 j, k;

		ret = sqp->queue[index];

		for (j=index, k = MOD_QUEUE_SIZE(sqp, (j+1));
		     k != SQNQ(sqp);
		     j=k, k = MOD_QUEUE_SIZE(sqp, (j+1)))
		{
		    sqp->queue[j] = sqp->queue[k];
		}
		sqp->enqueue_here--;
		break;
	    }
	}
    }

    QUNLOCK(sqp);

    return (ret);
}

/*
 * These are safe to use any time.
 */

int32 queue_simple_length(QUEUE_SIMPLE *sqp)
{
    int32 length;

    QLOCK(sqp);
    length = SQLENGTH(sqp);
    QUNLOCK(sqp);
    return(length);
}

int32 queue_simple_is_full(QUEUE_SIMPLE *sqp)
{
    int32 ret;

    QLOCK(sqp);
    ret = SQFULL(sqp);
    QUNLOCK(sqp);
    return(ret);
}

static int32 _queue_simple_is_this_power_of_two(uint32 l)
{
    int32 num_bits_on = 0;
    int32 i;

    for (i=0;i<32;i++)
    {
	num_bits_on += (l>>i)&0x1;
    }

    return (num_bits_on == 1);
}
    
QUEUE_QUEUE queue_simple_allocate(uint32 size)
{
    struct queue_queue *q;
    QUEUE_SIMPLE *sqp;

    /* Queues are buffers */
    q =
      (struct queue_queue *)queue_buffer_allocate(sizeof(struct queue_queue));
    q->is_standard = 0;
    sqp = &(q->u.simple);

    if (_queue_simple_is_this_power_of_two(size))
    {
        QLOCKINIT(sqp);
	sqp->queue_length_mask = size - 1;
	sqp->dequeue_this = 0;
	sqp->enqueue_here = 0;
	sqp->queue =
	  (QUEUE_BUFFER **)STDALLOC(size * sizeof(QUEUE_BUFFER *));
	return(q);
    }
    else
    {
#ifdef USE_STDERR
	fprintf(stderr, "LIBQUEUE ERROR: %d is not a power of two!\n", size);
#endif /* USE_STDERR */
	abort();
	/*NOTREACHED*/
    }
}

void queue_simple_describe(QUEUE_SIMPLE *sqp)
{
    uint32 i;

    QLOCK(sqp);
    
    printf("QUEUE_SIMPLE @ 0x%X, length = %d\n",
	   sqp,
	   SQLENGTH(sqp));
    printf("  queue_length_mask = 0x%X\n", sqp->queue_length_mask);
    printf("  dequeue_this = %d\n", sqp->dequeue_this);
    printf("  enqueue_here = %d\n", sqp->enqueue_here);
    
    for (i=sqp->dequeue_this;i<sqp->enqueue_here;i++)
    {
	uint32 index = MOD_QUEUE_SIZE(sqp,i);

	printf("[%d] = 0x%X\n", index ,sqp->queue[index]);
    }

    QUNLOCK(sqp);
}

void queue_simple_flush(QUEUE_SIMPLE *sqp)
{
    QUEUE_BUFFER *bp;

    while (bp = queue_simple_dequeue(sqp))
    {
	if (bp->is_a_queue)
	  queue_deallocate((QUEUE_QUEUE)char_from_buffer(bp));
	else
	  queue_buffer_deallocate(char_from_buffer(bp));
    }
}

int32 queue_simple_traverse_and_apply(QUEUE_SIMPLE *sqp,
				      QUEUE_FUNCTION fcn,
				      QUEUE_FUNCTION_ARG arg)
{
    uint32 index, i;
    int32  count = 0;

    QLOCK(sqp);

    if (!(SQEMPTY(sqp)))
    {
	for (i = 0; i < SQLENGTH(sqp); i++)
	{
	    index = MOD_QUEUE_SIZE(sqp,(SQDQ(sqp) + i));
	    count += (*fcn)(char_from_buffer(sqp->queue[index]), arg);
	}
    }

    QUNLOCK(sqp);

    return (count);
}
