#ifndef lint
static char rcsid [] = "$RCSfile: que_standard.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: que_standard.c                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Standard (in conrast to simple) queue routines            *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "libqueue_local.h"
#include <stdalloc.h> /*common/include/global*/

#ifdef PARANOID
static int32 _queue_consistency_check(QUEUE_STANDARD *qp);
#endif /*  PARANOID */

QUEUE_QUEUE queue_standard_allocate(void)
{
    struct queue_queue *q;
    QUEUE_STANDARD *qp;

    /* Queues are buffers */
    q =
      (struct queue_queue *)queue_buffer_allocate(sizeof(struct queue_queue));
    q->is_standard = 1;
    qp = &(q->u.standard);

    QLOCKINIT(qp);
    QLOCK(qp);
    qp->first = (QUEUE_ELEMENT *)0;
    qp->last = (QUEUE_ELEMENT *)0;
    qp->length = 0;

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    QUNLOCK (qp);
    return(q);
}

int32 queue_standard_length(QUEUE_STANDARD *qp)
{
    int32 ret = 0;
    
    QLOCK(qp);
    ret = qp->length;
#ifdef PARANOID
    _queue_consistency_check(qp);
#endif
    QUNLOCK(qp);
    return(ret);
}

static QUEUE_ELEMENT *_queue_hanger_hang_this(QUEUE_BUFFER *p)
{
    QUEUE_ELEMENT *h =
      (QUEUE_ELEMENT *)STDALLOC(sizeof(QUEUE_ELEMENT));
    h->prev = (QUEUE_ELEMENT *)0;
    h->next = (QUEUE_ELEMENT *)0;
    h->data = p;
    return(h);
}

static QUEUE_BUFFER *_queue_hanger_unhang(QUEUE_ELEMENT *h)
{
    QUEUE_BUFFER *p = h->data;
    STDDEALLOC(h);
    return(p);
}

void queue_standard_enqueue(QUEUE_STANDARD *qp, QUEUE_BUFFER *bufp)
{
    QUEUE_ELEMENT *qt_ptr;
    QUEUE_ELEMENT *qe_ptr = _queue_hanger_hang_this(bufp);

    QLOCK(qp);

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    if (!qp->first)
    {
        /** The queue is empty **/
        qp->first = qe_ptr;
        qe_ptr->prev = NULL;
    }
    else
    {
        /** Add an element to the end of the (circular) list. **/
        qt_ptr = qp->last;
        qt_ptr->next = qe_ptr;          /** Old last points to new last **/
        qe_ptr->prev = qt_ptr;          /** Last points to next-to-last **/
    }

    qp->last = qe_ptr;
    qe_ptr->next = NULL;
    ++qp->length;

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    QUNLOCK(qp);
}

QUEUE_BUFFER *queue_standard_dequeue(QUEUE_STANDARD *qp)
{
    QUEUE_ELEMENT *qe_ptr;
    QUEUE_BUFFER *ret;

    QLOCK(qp);

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    if (qp->first)
    {
        qe_ptr = qp->first;
        qp->first = qe_ptr->next;
        --qp->length;

	if (qp->first == NULL)
	{
	    qp->last = NULL;
	}

#ifdef DBG
      printf("qe_ptr=%X\n", qe_ptr);
#endif

	ret = _queue_hanger_unhang(qe_ptr);
    }
    else
    {
#ifdef DBG
      printf("qe_ptr=NULL\n");
#endif
	ret = NULL;
    }

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    QUNLOCK(qp);
    return (ret);
}

QUEUE_BUFFER *queue_standard_peek(QUEUE_STANDARD *qp)
{
    QUEUE_BUFFER *ret;

    if (qp->first)
    {
	ret = qp->first->data;
    }
    else
    {
	ret = NULL;
    }

    return (ret);
}

QUEUE_BUFFER *queue_standard_excise(QUEUE_STANDARD *qp, QUEUE_BUFFER *p)
{
    QUEUE_ELEMENT *qn;
    QUEUE_BUFFER *ret = NULL;

    QLOCK(qp);

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    for (qn = qp->first; qn; qn = qn->next)
    {
	if (p == qn->data)
	{
	    if (qn == qp->first)
	      qp->first = qn->next;
	    if (qn == qp->last)
	      qp->last = qn->prev;
	    if (qn->prev)
	      qn->prev->next = qn->next;
	    if (qn->next)
	      qn->next->prev = qn->prev;
	    --qp->length;

	    ret = _queue_hanger_unhang(qn);
	    break;
	}
    }

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    QUNLOCK (qp);

    return (ret);
}

int32 queue_standard_traverse_and_apply(QUEUE_STANDARD *qp,
				        QUEUE_FUNCTION fcn,
				        QUEUE_FUNCTION_ARG arg)
{
    QUEUE_ELEMENT *qn, *next=NULL;
    int32 count = 0;

    QLOCK(qp);

    for (qn = qp->first; qn; qn = next)
    {
	next = qn->next;
	count += (*fcn)(char_from_buffer(qn->data), arg);
    }

    QUNLOCK(qp);

    return (count);
}

/*
 * Assumes the QUEUE is already locked!
 * This seemed easier in recursive form.
 */

static void queue_flush_hangers(QUEUE_ELEMENT *qep)
{
    if (qep == (QUEUE_ELEMENT *)NULL)
    {
	return;
    }
    else
    {
	/*
	 * There are race conditions here.
	 * If you don't REALLY understand them, don't modify this fcn.
	 */

	queue_flush_hangers(qep->next);
	STDDEALLOC(qep);
    }
}

/*
 * Seemed easier to do in two passes.
 * Could make this more efficient by combining with preceding fcn.
 */

void queue_standard_flush(QUEUE_STANDARD *qp)
{
    QUEUE_ELEMENT *qn;

    QLOCK(qp);

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    qn = qp->first;

    /* First, get rid of all the data. */

    for (; qn != 0; qn = qn->next)
    {
	if (qn->data->is_a_queue)
	  queue_deallocate((QUEUE_QUEUE)char_from_buffer(qn->data));
	else
	  queue_buffer_deallocate(char_from_buffer(qn->data));
    }

    /* Next, get rid of all the hangers. */

    queue_flush_hangers(qp->first);

    qp->first = NULL;
    qp->last = NULL;
    qp->length = 0;

#ifdef PARANOID
    _queue_consistency_check(qp);
#endif

    QUNLOCK(qp);
}    

void queue_standard_describe (QUEUE_STANDARD *qp)
{
    QLOCK(qp);

    printf("QUEUE_STANDARD: { first = 0x%X last = 0x%X length = %d }\n",
	   qp->first, qp->last, qp->length);
    fflush (stdout);

    QUNLOCK(qp);
}

#ifdef PARANOID
static int32 _queue_length_forw(QUEUE_ELEMENT *qep)
{
    if (qep == (QUEUE_ELEMENT *)NULL)
    {
	return(0);
    }
    else
    {
	return (1 + _queue_length_forw(qep->next));
    }
}

static int32 _queue_length_back(QUEUE_ELEMENT *qep)
{
    if (qep == (QUEUE_ELEMENT *)NULL)
    {
	return(0);
    }
    else
    {
	return (1 + _queue_length_back(qep->prev));
    }
}

static int32 _queue_consistency_check(QUEUE_STANDARD *qp)
{
    if ((qp->length != _queue_length_forw(qp->first))
	&& (qp->length != _queue_length_back(qp->last)))
    {
#ifdef USE_STDERR
	fprintf(stderr, "QUEUE CONSISTENCY PROBLEM!\n");
#endif /* USE_STDERR */
	abort();
    }
}    
#endif
