#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libqueue                                 *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libqueue.h"
#include <stdext.h>

#define NITS 16

main()
{
    QUEUE_QUEUE q, r;
    int32 *b, *bs[NITS];
    int32 s, i, j;
    QUEUE_BUFFER_POOL pool;
    int32 *first = NULL;

    queue_print_statistics();

    for (s=0;s<=NITS;s+=NITS)
    {
	/* Make a queue */
	q = queue_allocate(s);

	/* Make some buffers and enqueue them */
	for (i=0;i<NITS;i++)
	{
	    b = (int32 *)queue_buffer_allocate(4);
	    *b = i;
	    queue_enqueue(q, b);
	    if (i == 0)
	      first = b;
	    queue_buffer_deallocate(b);
	}
    
	/* Check that peek finds the first item */
	b = (int32 *)queue_peek(q);
	if (b != first)
	  abort();

	/* Dequeue them and see that everything is cool */
	for (i=0;i<NITS;i++)
	{
	    b = (int32 *)queue_dequeue(q);
	    if (!b || (*b != i))
	      abort();
	    queue_buffer_deallocate(b);
	}

	/* Check that peek finds NULL */
	if ((int32 *)queue_peek(q))
	  abort();

	/* Ditch the queue */
	queue_deallocate(q);
    }

    queue_print_statistics();

    /* Make a buffer pool */
    pool = queue_buffer_pool_allocate(4, NITS);

    /* Do the same thing, but this time get the buffers from the pool */
    for (s=0;s<=NITS;s+=NITS)
    {
	/* Make a queue */
	q = queue_allocate(s);

	/* Make some buffers and enqueue them */
	for (i=0;i<NITS;i++)
	{
	    b = (int32 *)queue_buffer_allocate_from_pool(pool);
	    *b = i;
	    queue_enqueue(q, b);
	    if (i == 0)
	      first = b;
	    queue_buffer_deallocate(b);
	}
    
	/* Check that peek finds the first item */
	b = (int32 *)queue_peek(q);
	if (b != first)
	  abort();

	/* Dequeue them and see that everything is cool */
	for (i=0;i<NITS;i++)
	{
	    b = (int32 *)queue_dequeue(q);
	    if (!b || (*b != i))
	      abort();
	    queue_buffer_deallocate(b);
	}

	/* Check that peek finds NULL */
	if (queue_peek(q))
	  abort();

	/* Ditch the queue */
	queue_deallocate(q);
    }

    /* Free the pool */
    if (!queue_buffer_pool_deallocate(pool))
      abort();

    queue_print_statistics();

    for (s=0;s<=NITS;s+=NITS)
    {
	/* Make a queue of queues */
	q = queue_allocate(s);
	for (i=0;i<NITS;i++)
	{
	    r = queue_allocate(s);
	    queue_enqueue(q, r);
	    queue_deallocate(r);
	}

	/* Make some buffers and distribute them */
	for (i=0;i<NITS;i++)
	{
	    b = (int32 *)queue_buffer_allocate(4);
	    *b = i;
	    queue_distribute(q, b);
	    queue_buffer_deallocate(b);
	}
    
	/* Dequeue them and see that everything is cool */
	for (i=0;i<NITS;i++)
	{
	    r = (QUEUE_QUEUE)queue_dequeue(q);
	    for (j=0;j<NITS;j++)
	    {
		b = (int32 *)queue_dequeue(r);
		if (!b || (*b != j))
		  abort();
		queue_buffer_deallocate(b);
	    }
	    queue_deallocate(r);
	}

	/* Ditch the queue */
	queue_deallocate(q);
    }

    queue_print_statistics();

    for (s=0;s<=NITS;s+=NITS)
    {
	/* Make a queue */
	q = queue_allocate(s);

	/* Make some buffers and enqueue them */
	for (i=0;i<NITS;i++)
	{
	    b = (int32 *)queue_buffer_allocate(4);
	    *b = i;
	    bs[i] = b;
	    queue_enqueue(q, b);
	}
    
	/* Excise them in random order */
	for (i=0;i<NITS;i++)
	{
	    do
	      j = random()%NITS;
	    while (!bs[j]);

	    b = (int32 *)queue_excise(q, bs[j]);
	    if (!b || (*b != j))
	      abort();
	    queue_buffer_deallocate(b);

	    queue_buffer_deallocate(bs[j]);
	    bs[j] = 0;
	}

	/* Ditch the queue */
	queue_deallocate(q);
    }

    queue_print_statistics();

    printf("If those all say 0, then everthing works.\n");
}

