%{
#ifndef lint
static char rcsid[]="$RCSfile: rdr_lexer.l,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: rdr_lexer.l                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Lex scanner grammar                                       *
*       Created: 1/29/90                                                    *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "librdr_local.h"
#include "rdr_parser.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdalloc.h>
#include <stdstring.h>
#include <stdext.h>

int32 reader_lex_line_count = 1;
int32 reader_dot_warning_given = 0;


#ifdef USE_READER_READ_FROM_STRING

/* Fancy hack to get lex to read data from a string or from a file based
 * on reader_reading_from_string.  It is a little complicated so it will
 * work with both at&t lex and lbl flex.
 */
#ifndef FLEX_SCANNER

static char *string_to_read_from;
static char *start_of_string;
static int32 reading_from_string=0;


static char lex_do_input(void)
{
    if (!reading_from_string)
    {
	return(input());
    }
    else
    {
	return(*string_to_read_from++);
    }
}

static void lex_do_unput(
    char c)
{
    if (!reading_from_string)
    {
	unput(c);
    }
    else
    {
	*--string_to_read_from = c;
    }
}

#undef input
#undef unput
#define input() lex_do_input()
#define unput(c) lex_do_unput(c)


void reader_string_preparse(
    char *string)
{
    string_to_read_from = (char *)STDALLOC(strlen(string)+1);
    start_of_string = string_to_read_from;
    strcpy(string_to_read_from, string);
    reading_from_string = 1;
}

void reader_string_postparse(void)
{
    STDDEALLOC(start_of_string);
    reading_from_string = 0;
}

#else   /* FLEX_SCANNER */

static YY_BUFFER_STATE prev_state;
static YY_BUFFER_STATE current_state;

void reader_string_preparse(
    char *string)
{
    current_state = yy_scan_string(string);
}

void reader_string_postparse(void)
{
    yy_delete_buffer(current_state);
    if(!YY_CURRENT_BUFFER)
      yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));      
}

#endif   /* FLEX_SCANNER */

#endif  /* USE_READER_READ_FROM_STRING */

%}

delim   [ \t]
newline [\n]
letter  [A-Za-z]
digit   [0-9]
xdigit  [0-9a-fA-F]
odigit  [0-7]
dot     \.
quote   \"
openp   \(
closep  \)
openb   \{
closeb  \}
opensq  \[
closesq \]
semi    \;
mulop   [\*\/\%]
addop   [\+\-]
shiftop ((\<\<)|(\>\>))
eqop    ((\=\=)|(\!\=))
ltop    (\<\=?)
gtop    (\>\=?)
bitaop  \&
bitxop  \^
bitiop  \|
andop   \&\&
orop    \|\|
notop   [\!\~]
questop \?
colonop \:
string  ({quote}[^"]*{quote})
comment {semi}.*$
sign    [\+\-]
id      (({letter}|{digit})({letter}|([\-\_])|{digit})*)
exp     ([Ee]{sign}?{digit}+)
realnum ({sign}?{digit}*{dot}{digit}+{exp}?)
decnum  ({digit}+)
hexnum  (0[xX]{xdigit}+)
octnum  (0[oO]{odigit}+)

%%     

{comment} { }
{string}  {
    yylval.typed.ru.charptr = reader_get_symbol_with_quotes(yytext);
    return SYMBOL;
}
{openp}    {
    yylval.typed.ru.integer = 0;
    return OPEN_PAREN;
}
{closep}  {
    yylval.typed.ru.integer = 0;
    return CLOSE_PAREN;
}
{opensq}  {
    yylval.typed.ru.integer = 0;
    return OPEN_SQUARE;
}
{closesq} {
    yylval.typed.ru.integer = 0;
    return CLOSE_SQUARE;
}
{openb}    {
    yylval.typed.ru.integer = 0;
    return OPEN_BRACE;
}
{closeb}  {
    yylval.typed.ru.integer = 0;
    return CLOSE_BRACE;
}
{realnum} {
    
    yylval.typed.ru.real = atof(yytext);
    return FLOAT;
}
{hexnum}  {
    sscanf(yytext+2,"%x",&(yylval.typed.ru.integer));
    return INTEGER;
}
{octnum}  {
    sscanf(yytext+2,"%o",&(yylval.typed.ru.integer));
    return INTEGER;
}
{decnum}  {
    sscanf(yytext,"%d",&(yylval.typed.ru.integer));
    return INTEGER;
}
{mulop}   {
    yylval.typed.ru.integer = yytext[0];
    return MULOP;
}
{addop}   {
    yylval.typed.ru.integer = yytext[0];
    return ADDOP;
}
{shiftop} {
    yylval.typed.ru.integer = yytext[0];
    return SHIFTOP;
}
{eqop}   {
    yylval.typed.ru.integer = (yytext[0] == '=');
    return EQOP;
}
{ltop}   {
    yylval.typed.ru.integer = (yytext[1] == '=');
    return LTOP;
}
{gtop}   {
    yylval.typed.ru.integer = (yytext[1] == '=');
    return GTOP;
}
{bitaop}  {
    yylval.typed.ru.integer = yytext[0];
    return BITAOP;
}
{bitxop}  {
    yylval.typed.ru.integer = yytext[0];
    return BITXOP;
}
{bitiop}  {
    yylval.typed.ru.integer = yytext[0];
    return BITIOP;
}
{andop}   {
    yylval.typed.ru.integer = yytext[0];
    return ANDOP;
}
{orop}    {
    yylval.typed.ru.integer = yytext[0];
    return OROP;
}
{notop}    {
    yylval.typed.ru.integer = yytext[0];
    return NOTOP;
}
{questop} {
    yylval.typed.ru.integer = yytext[0];
    return QUESTOP;
}
{colonop} {
    yylval.typed.ru.integer = yytext[0];
    return COLONOP;
}
t         {
    yylval.typed.ru.integer = 1;
    return INTEGER;
}
nil       {
    yylval.typed.ru.integer = 0;
    return INTEGER;
}
{id}      {
    yylval.typed.ru.charptr = reader_get_symbol(yytext);
    if (reader_get_symbol_value(yylval.typed.ru.charptr))
      return MACRO_SYMBOL;
    else
      return SYMBOL;
}
{delim}   ;
{newline} {
    reader_lex_line_count++;
}
{dot}     {
    if (!reader_dot_warning_given)
    {
	yyerror("Warning: . encountered (ignored)");
	reader_dot_warning_given = 1;
    }
}
.         {
    printf("Lex error on line %d.  ?? %s ??\n",
	   reader_lex_line_count, yytext);
    exit(1);
}

%%
