%{
#ifndef lint
static char rcsid[]="$RCSfile: rdr_parser.y,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: rdr_parser.y                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Yacc parser grammar                                       *
*       Created: 1/29/90                                                    *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "librdr_local.h"
#include <stdio.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h>
#include <stdstring.h>

static int32 do_typing = 0;
static READER_UNION parsed_data;
static char *file_name;
int32 reader_alert;
static char override_path[4096] = ".";
static int32 n_override_dirs = 1;
static char default_path[4096] = "";
static int32 n_default_dirs = 0;



static READER_STACK *stack_push(
    TYPED_DATA *typed_ptr, 
    READER_STACK *stack_ptr)
{
    READER_STACK * new_stack;

    new_stack = (READER_STACK *)STDALLOC(sizeof(READER_STACK));
    new_stack->typed = *typed_ptr;
    new_stack->next = stack_ptr;
    return new_stack;
}

static READER_STACK *stack_push_array(
    READER_UNION *arr,
    READER_STACK *stack_ptr)
{
    int32 len = arr[0].integer;
    int32 i;

    /*  an array given to us has already been processed with make_array and
     *  is made of READER_UNIONs, but we need to push TYPED_DATAs.
     */

    for (i=1;i<len;i++)
    {
	TYPED_DATA typed;
	typed.ru = arr[i];
	if (do_typing)
	    typed.type = READER_UTYPE(arr,i);
	stack_ptr = stack_push(&typed, stack_ptr);
    }

    return stack_ptr;
}

static void free_stack(
    READER_STACK *stack_ptr)
{
    if (stack_ptr)
    {
	free_stack(stack_ptr->next);
	STDDEALLOC(stack_ptr);
    }
}

static READER_UNION *make_array(
    READER_STACK *stack_ptr, 
    int32 ismac)
{
    int32 len, n_unions;
    READER_STACK * s;
    READER_UNION * result;
    char *types;

    /*  Determine array length.  We need len data unions, and optionally len
     *  type bytes at the beginning rounded up to the nearest multiple of
     *  sizeof READER_UNION.
     */

    for (len=1, s=stack_ptr; s; len++,s=s->next)
	;

    n_unions = len;

    if (do_typing)
	n_unions += len / sizeof(READER_UNION) + 1;
    result = (READER_UNION *)STDALLOC(n_unions * sizeof(READER_UNION));
    result += (n_unions - len);
    result[0].integer = len;

    if (do_typing)
    {
	types = ((char*) result) - 2;
	types[0] = READER_INTEGER;
    }

    /*  copy the data.  top of stack is last array element.
    */

    if (s = stack_ptr)
    {
	READER_UNION *dp = result + len - 1;
	READER_STACK *tmp;
	if (do_typing)
	    types -= len - 1;
	do {
	    *dp = s->typed.ru;
	    if (do_typing){
		*types++ = s->typed.type;
                if ((s->typed.type == READER_ARRAY) && ismac)
                    ((char *)(dp->array))[-1] = ismac;
	    }
            dp--;
	    tmp = s;
	    s = s->next;
	    STDDEALLOC(tmp);
	} while (s);
    }

    return result;
}


%}

%union
{
    TYPED_DATA typed;
    READER_STACK * stackptr;
}

%token <typed.ru.charptr>  SYMBOL
%token <typed.ru.charptr>  MACRO_SYMBOL
%token <typed.ru.integer>  INTEGER
%token <typed.ru.integer>  OPEN_PAREN CLOSE_PAREN
%token <typed.ru.integer>  OPEN_SQUARE CLOSE_SQUARE
%token <typed.ru.integer>  OPEN_BRACE CLOSE_BRACE
%token <typed.ru.real>     FLOAT
%token <typed.ru.integer>  MULOP ADDOP SHIFTOP EQOP LTOP GTOP
%token <typed.ru.integer>  NOTOP BITAOP BITXOP BITIOP ANDOP OROP
%token <typed.ru.integer>  QUESTOP COLONOP

%left OROP
%left ANDOP
%left BITIOP
%left BITXOP
%left BITAOP
%left EQOP
%left LTOP GTOP
%left SHIFTOP
%left ADDOP
%left MULOP
%right NOTOP

%type  <typed.ru.charptr>  symbol_type
%type  <typed.ru.charptr>  macro_symbol_type
%type  <typed.ru.integer>  integer_type
%type  <typed.ru.integer>  math_expr
%type  <typed.ru.real>     float_type
%type  <typed.ru.array>    array_type
%type  <typed.ru.array>    macro_value
%type  <typed.ru.charptr>  macro_defn
%type  <stackptr>          data_list
%type  <typed>             data

%start file

%%
file			: data_list
                          {
			      parsed_data = ($1)->typed.ru;
			      free_stack($1);
			  }
			;
macro_defn              : symbol_type OPEN_BRACE data_list CLOSE_BRACE
                          {
			      READER_UNION *arr;

			      arr = make_array($3, 1);
			      reader_set_symbol_value($1, arr);
			      $$ = $1;
			  }
			;
macro_value             : macro_symbol_type
                          {
			      $$ = reader_get_symbol_value($1);
			  }
			;
macro_symbol_type       : MACRO_SYMBOL
                          { $$ = $1; }
			;
data			: macro_defn
                          {
			      $$.ru.charptr = $1;
			      $$.type = READER_CHARPTR;
			  }
                        | symbol_type
			  {
			      $$.ru.charptr = $1;
			      $$.type = READER_CHARPTR;
			  }
			| integer_type
			  {
			      $$.ru.integer = $1;
			      $$.type = READER_INTEGER;
			  }
			| float_type
			  {
			      $$.ru.real = $1;
			      $$.type = READER_REAL;
			  }
			| array_type
			  {
			      $$.ru.array = $1;
			      $$.type = READER_ARRAY;
			  }
			;
array_type		: OPEN_PAREN data_list CLOSE_PAREN
			  {
			      $$ = make_array($2, 0);
			  }
                        | OPEN_PAREN CLOSE_PAREN
                          {
			      $$ = 0;
			  }
			;
data_list		: data
			  {
			      $$ = stack_push(&($1), (READER_STACK*) NULL);
			  }
			| data_list data
			  {
			      $$ = stack_push(&($2), $1);
			  }
			| macro_value
			  {
			      $$ = stack_push_array($1, (READER_STACK*) NULL);
			  }
			| data_list macro_value
			  {
			      $$ = stack_push_array($2, $1);
			  }
			;
symbol_type		: SYMBOL
			  { $$ = $1; }
			;
integer_type		: OPEN_SQUARE math_expr CLOSE_SQUARE
			  { $$ = $2; }
			| ADDOP integer_type
			  {
			      if ($1 == '+')
				$$ = $2;
			      else
				$$ = - $2;
			  }
			| INTEGER
			  { $$ = $1; }
			;
math_expr		: math_expr QUESTOP math_expr COLONOP math_expr
			  { $$ = $1 ? $3 : $5; }
			| math_expr OROP math_expr
			  { $$ = $1 || $3; }
			| math_expr ANDOP math_expr
			  { $$ = $1 && $3; }
			| math_expr BITIOP math_expr
			  { $$ = $1 | $3; }
			| math_expr BITXOP math_expr
			  { $$ = $1 ^ $3; }
			| math_expr BITAOP math_expr
			  { $$ = $1 & $3; }
			| math_expr EQOP math_expr
			  {
			      if ($2)		/* meaning == */
				$$ = ($1 == $3);
			      else		/* meaning != */
				$$ = ($1 != $3);
			  }
			| math_expr LTOP math_expr
			  {
			      if ($2)		/* meaning <= */
				$$ = ($1 <= $3);
			      else		/* meaning < */
				$$ = ($1 < $3);
			  }
			| math_expr GTOP math_expr
			  {
			      if ($2)		/* meaning >= */
				$$ = ($1 >= $3);
			      else		/* meaning > */
				$$ = ($1 > $3);
			  }
			| math_expr SHIFTOP math_expr
			  {
			      if ($2 == '<')	/* meaning << */
				$$ = $1 << $3;
			      else		/* meaning >> */
				$$ = $1 >> $3;
			  }
			| math_expr ADDOP math_expr
			  {
			      if ($2 == '+')
				$$ = $1 + $3;
			      else
				$$ = $1 - $3;
			  }
			| math_expr MULOP math_expr
			  {
			      if ($2 == '*')
				$$ = $1 * $3;
			      else if ($2 == '/')
				$$ = $1 / $3;
			      else
				$$ = $1 % $3;
			  }
			| NOTOP math_expr %prec NOTOP
			  {
			      if ($1 == '!')
				$$ = !($2);
			      else
				$$ = ~($2);
			  }
			| macro_value
			  {
			      if ($1[0].integer == 2)
				$$ = $1[1].integer;
			      else
				$$ = 1;
			  }
			| integer_type
			  { $$ = $1; }
			;
float_type		: FLOAT
			  { $$ = $1; }
			;
%%

void yyerror(
    char *s)
{
    fprintf(stderr,"%s: %s at line %d\n",
	    file_name, s, reader_lex_line_count);
}

void reader_set_search_paths(
    char *override,
    char *defalt)
{
    if (override)
    {
	strncpy(override_path, override, sizeof(override_path));
	strtok(override_path, ":");
	for(n_override_dirs=1;strtok(NULL, ":");n_override_dirs++)
	  ;
    }
    else
    {
	*override_path = 0;
	n_override_dirs = 0;
    }

    if (defalt)
    {
	strncpy(default_path, defalt, sizeof(default_path));
	strtok(default_path, ":");
	for(n_default_dirs=1;strtok(NULL, ":");n_default_dirs++)
	  ;
    }
    else
    {
	*default_path = 0;
	n_default_dirs = 0;
    }
}

static int32 read_file(
    FILE *f,
    READER_UNION *ru)
{
    int32 errcond;

    yyin = f;
    reader_lex_line_count = 1;
    reader_dot_warning_given = 0;
    errcond = yyparse();
    fclose(f);

    if (errcond)
    {
	ru->integer = 0;
	return READER_READ_ERROR;
    }
    else
    {
	*ru = parsed_data;
	return 0;
    }
}

static int32 search_and_read(
    char *search_path,
    int32 search_len,
    char *fname,
    READER_UNION *ru,
    int32 alert)
{
    char path[256];
    int32 i;
    FILE *f=NULL;

    for (i=0;i<search_len;i++,search_path+=strlen(search_path)+1)
    {
	if (*search_path)
	{
	    strcpy(path, search_path);
	    strcat(path, "/");
	}
	else
	  *path = 0;
	strcat(path, fname);
	if (f = fopen(path,"r"))
	{
	    if (alert)
	      printf("Reading file %s from %s\n", fname, search_path);
	    break;
	}
    }

    if (f)
    {
	file_name = path;
	return read_file(f, ru);
    }
    else
      return READER_FILE_NOT_FOUND;
}

int32 reader_read(
    char *fname,
    char *dirs,
    READER_UNION *ru,
    uint32 flags)
{
    int32 ret;
    int32 n_dirs;
    char dirs_path[4096];


    /* In case something fails */
    ru->integer = 0;

    do_typing = (flags & READER_TYPING) != 0;

    if (flags & READER_OVERRIDE)
    {
	if ((ret = search_and_read(override_path, n_override_dirs,
				   fname, ru, reader_alert)) !=
	    READER_FILE_NOT_FOUND)
	  return ret;
    }

    if (dirs)
    {
	strncpy(dirs_path, dirs, sizeof(dirs_path));
	strtok(dirs_path, ":");
	for(n_dirs=1;strtok(NULL, ":");n_dirs++)
	  ;
    }
    else
    {
	n_dirs=1;
	dirs_path[0]='\0';
    }

    if ((ret = search_and_read(dirs_path, n_dirs, fname, ru, 0)) !=
	READER_FILE_NOT_FOUND)
      return ret;

    if (flags & READER_DEFAULTS)
      ret = search_and_read(default_path, n_default_dirs, fname, ru, 0);

    return ret;
}	  


#ifdef USE_READER_READ_FROM_STRING
int32 reader_read_from_string(
    char *string,
    READER_UNION *ru,
    uint32 flags)
{
    int32 ret;
    int32 errcond;

    /* In case something fails */
    ru->integer = 0;

    do_typing = (flags & READER_TYPING) != 0;

    reader_string_preparse(string);
    reader_lex_line_count = 1;
    reader_dot_warning_given = 0;
    errcond = yyparse();
    reader_string_postparse();

    if (errcond)
    {
	ru->integer = 0;
	return READER_READ_ERROR;
    }
    else
    {
	*ru = parsed_data;
	return 0;
    }
}
#endif /* USE_READER_READ_FROM_STRING */


static void output_string(
    char *s,
    FILE *stream)
{
#define MAXCOL 75
    static int32 column = 0;
    static int32 indent = 0;
    static int32 prespace = 0;
    int32 i;
    int32 len;
    int32 wrap, oparen, cparen;

#ifdef _GT_
    oparen = (index(s, "(")) != -1;
    cparen = (index(s, ")")) != -1;
#else
    oparen = (strchr(s, '(')) != 0;
    cparen = (strchr(s, ')')) != 0;
#endif

    if (oparen)
      column = MAXCOL; /* Force an indent */

    len = strlen(s);

    wrap = column+len >= MAXCOL;

    if (oparen || cparen || wrap)
      prespace = 0;

    if (prespace)
      len += 1; 

    if (wrap)
    {
	if (indent)
	  putc('\n', stream);
	for (i=0;i<indent;i++)
	  putc(' ', stream);

	column = indent;
    }

    if (prespace)
      putc(' ', stream);

    if (oparen)
      indent += 1;
    if (cparen)
      indent -= 1;

    fputs(s, stream);
    column += len;

    prespace = !oparen;

    if (!indent && cparen)
    {
	putc('\n', stream);
	prespace = column = 0;
    }
}

void reader_pretty_print(
    READER_UNION *ru,
    FILE *stream)
{
    char buf[100];
    int32 i, len;

    if (ru == NULL)
	{
        fprintf(stderr,"reader_pretty_print() passed NULL reader union\n");
	return;
	}
	
    len = ru[0].integer;

    output_string("(", stream);

    for (i=1;i<len;i++)
    {
	switch (READER_UTYPE(ru,i))
	{
	  case READER_INTEGER:
	    sprintf(buf, "%d", ru[i].integer);
	    output_string(buf, stream);
	    break;
	  case READER_REAL:
	    /*  Must detect if %g prints without decimal point or
	     *  E-notation, or the number could be read back as an int.
	     */
	    sprintf(buf, "%g", ru[i].real);
#ifdef _GT_
	    if ((index(buf, ".") != -1) && (index(buf, "e") != -1))
#else
	    if (!strchr(buf, '.') && !strchr(buf, 'e'))
#endif
	      strcat(buf, ".0");
	    output_string(buf, stream);
	    break;
	  case READER_CHARPTR:
	    sprintf(buf, "\"%s\"", ru[i].charptr);
	    output_string(buf, stream);
	    break;
	  case READER_ARRAY:
	    if (ru[i].array)
	      reader_pretty_print(ru[i].array, stream);
	    else
	      output_string("()", stream);
	    break;
	  default:
	    output_string("<unknown>", stream);
	    break;
	}
    }

    output_string(")", stream);
}

void reader_free_typed_ru(
    READER_UNION *ru)
{
    int32 i, len = ru[0].integer;

    for ( i = 1; i < len; ++i ){
	switch (READER_UTYPE(ru,i)){
	    case READER_INTEGER:
	        break;
	    case READER_REAL:
	        break;
	    case READER_CHARPTR:
	        break;
	    case READER_ARRAY:
                if (!READER_ISMAC(ru[i].array))
	            reader_free_typed_ru(ru[i].array);
	        break;
	    default:
	        break;
	}
    }
    if (!READER_ISMAC(ru))
        STDDEALLOC(ru - len / sizeof(READER_UNION) - 1);
}



/* combine reader union arrays and return the new reader union */
READER_UNION *reader_combine_arrays(
    READER_UNION *u[],
    int32         count)
{
    int32 length=1, i, last_length=1;
    READER_UNION *new;

    /* count up how many arrays we need to store */
    for (i=0; i<count; i++)
        length += u[i][0].integer - 1;

    /* alloc enough space to store the new array */
    new = (READER_UNION *)STDALLOC(sizeof(READER_UNION) * length);

    /* copy each reader union array into the right spot.. */
    for (i=0; i<count; i++)
    {
	int32 this_length = u[i][0].integer - 1;

        bcopy(&u[i][1], 
	      &new[last_length],
	      sizeof(READER_UNION) * this_length);
	
	last_length += this_length;
    }
    
    /* update the length of the new array.. */
    new[0].integer = length;

    return(new);
}
