#ifndef lint
static char rcsid [] = "$RCSfile: rdr_python.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tags.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Python interface to libreader                             *
*       Created: 6/96                                                       *
*       Author: emalafee                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#ifdef PYTHON_PARSER

#ifdef sgi4
#    define HAVE_STDLIB_H
#endif
#include <Py/Python.h>

#include "librdr_local.h"

extern int py_initialized;

static PyObject *build_reader_array(
    READER_UNION *ru)
{
    PyObject *po;
    int       i;

    po = PyList_New(ru[0].integer-1);

    for ( i = 1; i<ru[0].integer; i++ )
    {
	switch (READER_UTYPE(ru, i))
	{
	    case READER_CHARPTR:
		PyList_SetItem(po, i-1, PyString_FromString(ru[i].charptr));
		break;
	    case READER_INTEGER:
		PyList_SetItem(po, i-1, PyInt_FromLong(ru[i].integer));
		break;
	    case READER_REAL   :
		PyList_SetItem(po, i-1, PyFloat_FromDouble(ru[i].real));
		break;
	    case READER_ARRAY  :
		PyList_SetItem(po, i-1, build_reader_array(ru[i].array));
		break;
	    default            :
		printf("Reader element %d unidentified.\n", i);
		break;
	}
    }

    return(po);
}

static PyObject *py_reader_read(
    PyObject *self,
    PyObject *args)
{
    char         *fname;
    char         *dirs;
    READER_UNION  ru;
    int32         ret;
    uint32        flags;

    if ( !PyArg_ParseTuple(args, "ssd", &fname, &dirs, &flags) )
	return(NULL);

    if (( ret = reader_read(fname, dirs, &ru, flags | READER_TYPING)) != 
	      READER_READ_OK || ru.array[0].integer <= 1 )
    {
	Py_INCREF(Py_None);
	return(Py_None);
    }

    return(build_reader_array(ru.array));
}

static PyObject *py_reader_get_symbol(
    PyObject *self,
    PyObject *args)
{
    char *string;
    char *ret;

    if ( !PyArg_ParseTuple(args, "s", &string))
	return(NULL);

    ret = reader_get_symbol(string);


    return(Py_BuildValue("l", (long)ret));
}

static PyMethodDef ReaderMethods[] =
{
    {	"reader_read",		py_reader_read, 	1	},
    {	"reader_get_symbol",	py_reader_get_symbol,	1	},
    {	NULL,			NULL				}
};

void reader_init_python(void)
{
    (void) Py_InitModule("reader", ReaderMethods);
}

#endif
