#ifndef lint
static char rcsid [] = "$RCSfile: sched_funcall.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_funcall.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/
#include <stdio.h>
#include <stdalloc.h>
#include <libexcept.h>
#include "libsched_local.h"

#define SCHED_FUNCALL_MAX_ARGS 30

static EVENT_OPS funcall_ops = { 
    (EVENT_DESTROY_FCN)sched_funcall_event_destroy,
    (EVENT_DELETE_FCN)sched_event_delete_func,
    (EVENT_INVOKE_FCN)sched_funcall_event_invoke,
    (EVENT_CANCEL_MATCH_FCN)sched_event_cancel_match_func
};

void sched_funcall_event_init(SCHED_FUNCALL_EVENT *event,
			      int32 group,
			      void (*func) (/* ??? */),
			      va_list args)
{
    int32 arg_count;
    int32 arg_type; 
    SCHED_ARG_UNION arg_buf[SCHED_FUNCALL_MAX_ARGS];

    sched_event_init(&event->event_hdr, &funcall_ops, group);

    event->func = func;

    arg_count = 0;
    arg_type = va_arg(args,int);

    while (arg_type != A_END)
    {
	arg_buf[arg_count].tag = arg_type;
	switch(arg_type)
	{
	  case A_INT:
	    arg_buf[arg_count++].arg.int_arg = va_arg(args,int);
	    break;
	  case A_DOUBLE:
	    arg_buf[arg_count++].arg.double_arg = va_arg(args,double);
	    break;
	  default:
	    fprintf(stderr,"Unknown argument type %d.  Goodbye.\n",arg_type);
	    abort();
	}
	arg_type = va_arg(args,int);
    }

    event->arg_count = arg_count;

    if (arg_count)
    {
	event->args = (SCHED_ARG_UNION *)STDALLOC(arg_count * sizeof(SCHED_ARG_UNION));
	bcopy((char *)arg_buf,
	      (char *)event->args,
	      (int)(arg_count*sizeof(SCHED_ARG_UNION)));
    }
    else
      event->args = NULL;

}

SCHED_FUNCALL_EVENT* sched_funcall_event_create(int32 group,
					        void (*func) (/* ??? */),
					        va_list args)
{
    int32 arg_count;
    int32 arg_type;
    SCHED_FUNCALL_EVENT* event;

    event = (SCHED_FUNCALL_EVENT*)STDALLOC(sizeof(SCHED_FUNCALL_EVENT));
    sched_funcall_event_init(event, group, func, args);
    return event;
}

/*VARARGS*/
SCHED_FUNCALL_EVENT* sched_funcall_event_vcreate(int32 group,
						 void (*func) (/* ??? */),
						 ...)
{
    va_list args;

    va_start(args, func);
    return sched_funcall_event_create(group, func, args);
    va_end(args);
}
      


void sched_funcall_event_destroy(SCHED_FUNCALL_EVENT *event)
{
    if (event->args)
      STDDEALLOC(event->args);
}

SCHED_EVENT_INVOKE_RESULT 
sched_funcall_event_invoke(SCHED_FUNCALL_EVENT *event,
			   uint32 qtime,
			   SCHED_QUEUE *queue)
{
    /* The following was copied from the old scheduler. */

    SCHED_ARG_UNION * a = event->args;

    EXCEPT_ASSERT(event->func);

    /* This is a little gross, but I'm hoping the compiler can untangle
       it into some decent code... */

#define C1(t0)          event->func(a[0].arg.t0);
#define C2(t0,t1)       event->func(a[0].arg.t0,a[1].arg.t1);
#define C3(t0,t1,t2)    event->func(a[0].arg.t0,a[1].arg.t1,a[2].arg.t2);
#define C4(t0,t1,t2,t3) event->func(a[0].arg.t0,a[1].arg.t1,a[2].arg.t2,a[3].arg.t3);
#define S(n,x,y) \
	switch (a[n].tag) \
	{ \
	  case A_INT: x ; break ; \
	  case A_DOUBLE: y ; break ; \
	}

    switch(event->arg_count)
    {
      case 0:
	event->func();
	break;
      case 1:
	S(0,
	  { C1(int_arg) },
	  { C1(double_arg) } );
	break;
      case 2:
	S(0,
	  S(1,
	    { C2(int_arg,int_arg) },
	    { C2(int_arg,double_arg) } ),
	  S(1,
	    { C2(double_arg,int_arg) },
	    { C2(double_arg,double_arg) } )
	  );
	break;
      case 3:
	S(0,
	  S(1,
	    S(2,
	      { C3(int_arg,int_arg,int_arg) },
	      { C3(int_arg,int_arg,double_arg) }),
	    S(2,
	      { C3(int_arg,double_arg,int_arg) },
	      { C3(int_arg,double_arg,double_arg) })),
	  S(1,
	    S(2,
	      { C3(double_arg,int_arg,int_arg) },
	      { C3(double_arg,int_arg,double_arg) }),
	    S(2,
	      { C3(double_arg,double_arg,int_arg) },
	      { C3(double_arg,double_arg,double_arg) }))
	  );
	break;
      case 4:
	S(0,
	  S(1,
	    S(2,
	      S(3,
		{ C4(int_arg,int_arg,int_arg,int_arg) },
		{ C4(int_arg,int_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(int_arg,int_arg,double_arg,int_arg) },
		{ C4(int_arg,int_arg,double_arg,double_arg) })),
	    S(2,
	      S(3,
		{ C4(int_arg,double_arg,int_arg,int_arg) },
		{ C4(int_arg,double_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(int_arg,double_arg,double_arg,int_arg) },
		{ C4(int_arg,double_arg,double_arg,double_arg) }))),
	  S(1,
	    S(2,
	      S(3,
		{ C4(double_arg,int_arg,int_arg,int_arg) },
		{ C4(double_arg,int_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(double_arg,int_arg,double_arg,int_arg) },
		{ C4(double_arg,int_arg,double_arg,double_arg) })),
	    S(2,
	      S(3,
		{ C4(double_arg,double_arg,int_arg,int_arg) },
		{ C4(double_arg,double_arg,int_arg,double_arg) }),
	      S(3,
		{ C4(double_arg,double_arg,double_arg,int_arg) },
		{ C4(double_arg,double_arg,double_arg,double_arg) })))
	  );
	break;
      default:
	fprintf(stderr,"Cannot invoke functions with %d args.\n",
		event->arg_count);
	abort();
    }

    return EVENT_DONE;
}


    

