#ifndef lint
static char rcsid [] = "$RCSfile: sched_funcs.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_funcs.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libsched.h"
#include "libsched_local.h"
#include <stdio.h>
#include <libexcept.h>
#include <stdext.h>
#include <stdarg.h>


SCHED_FNCL_PTR sched_deferred_fncl(void (*func) (/* ??? */),
				   uint32 delay,
				   int32 group,
				   ...)
{
    SCHED_FUNCALL_EVENT* event;
    va_list args;
    uint32 current_time;

    va_start(args, group);
    event = sched_funcall_event_create(group, func, args);
    va_end(args);
    if (delay == SCHED_ASAP)
    {
	sched_pq_schedule_asap(&sched_realtime_queue, 
			       SCHED_BASE_EVENT(event));
    }
    else
    {
	current_time = 
	  sched_queue_current_time(SCHED_BASE_QUEUE(&sched_realtime_queue));
	sched_pq_schedule(&sched_realtime_queue,
			  SCHED_BASE_EVENT(event),
			  current_time + delay);
    }
    return SCHED_BASE_EVENT(event);
}

SCHED_FNCL_PTR sched_preemptive_deferred_fncl(void (*func) (/* ??? */),
					      uint32 delay,
					      int32 group,
					      ...)
{
    SCHED_FUNCALL_EVENT* event;
    va_list args;
    uint32 current_time;

    va_start(args, group);
    event = sched_funcall_event_create(group, func, args);
    va_end(args);
    current_time = 
      sched_queue_current_time(SCHED_BASE_QUEUE(&sched_preempt_queue));
    sched_preempt_queue_schedule(&sched_preempt_queue,
				    SCHED_BASE_EVENT(event),
				    current_time + delay);
    return SCHED_BASE_EVENT(event);
}

SCHED_FNCL_PTR sched_simtime_fncl(void (*func) (/* ??? */),
				  uint32 simtime,
				  int32 group,
				  ...)
{
    SCHED_FUNCALL_EVENT* event;
    va_list args;

    va_start(args, group);
    event = sched_funcall_event_create(group, func, args);
    va_end(args);
    sched_pq_schedule(&sched_simtime_queue,
		      SCHED_BASE_EVENT(event),
		      simtime);
    return SCHED_BASE_EVENT(event);
}

SCHED_FNCL_PTR sched_periodic_fncl(void (*func) (/* ??? */),
				   uint32 initial_delay,
				   uint32 period,
				   int32 group,
				   ...)
{
    SCHED_FUNCALL_EVENT* event;
    SCHED_PERIODIC_EVENT* per_event;
    va_list args;
    uint32 current_time;

    va_start(args, group);
    event = sched_funcall_event_create(group, func, args);
    va_end(args);
    per_event = sched_periodic_event_create(period, SCHED_BASE_EVENT(event));
    if (initial_delay == SCHED_ASAP)
    {
	sched_pq_schedule_asap(&sched_realtime_queue,
			       SCHED_BASE_EVENT(per_event));
    }
    else
    {
	current_time = 
	  sched_queue_current_time(SCHED_BASE_QUEUE(&sched_realtime_queue));
	sched_pq_schedule(&sched_realtime_queue,
			  SCHED_BASE_EVENT(per_event),
			  current_time + initial_delay);
    }
    return SCHED_BASE_EVENT(event);
}

SCHED_FNCL_PTR sched_high_priority_fncl(void (*func) (/* ??? */),
				        uint32 initial_delay,
				        uint32 period,
				        int32 group,
				        ...)
{
    SCHED_FUNCALL_EVENT* event;
    SCHED_PERIODIC_EVENT* per_event;
    va_list args;
    uint32 current_time;

    va_start(args, group);
    event = sched_funcall_event_create(group, func, args);
    va_end(args);
    per_event = sched_periodic_event_create(period, SCHED_BASE_EVENT(event));
    if (initial_delay == SCHED_ASAP)
    {
	sched_pq_schedule_asap(&sched_hipri_queue,
			       SCHED_BASE_EVENT(per_event));
    }
    else
    {
	current_time = 
	  sched_queue_current_time(SCHED_BASE_QUEUE(&sched_hipri_queue));
	sched_pq_schedule(&sched_hipri_queue,
			  SCHED_BASE_EVENT(per_event),
			  current_time + initial_delay);
    }
    return SCHED_BASE_EVENT(event);
}

SCHED_FNCL_PTR sched_preemptive_periodic_fncl(void (*func) (/* ??? */),
					      uint32 initial_delay,
					      uint32 period,
					      int32 group,
					      ...)
{
    SCHED_FUNCALL_EVENT* event;
    SCHED_PERIODIC_EVENT* per_event;
    va_list args;
    uint32 current_time;

    va_start(args, group);
    event = sched_funcall_event_create(group, func, args);
    va_end(args);
    per_event = sched_periodic_event_create(period, SCHED_BASE_EVENT(event));
    current_time = 
      sched_queue_current_time(SCHED_BASE_QUEUE(&sched_realtime_queue));
    sched_preempt_queue_schedule(&sched_preempt_queue,
				    SCHED_BASE_EVENT(per_event),
				    current_time + initial_delay);
    return SCHED_BASE_EVENT(event);
}

void sched_cancel_fncl(SCHED_FNCL_PTR event)
{
    sched_queues_cancel_event(event, TRUE);
}


void sched_cancel_fncl_group(int32 group)
{
    sched_queues_cancel_group(group);
}



void sched_change_fncl_period(SCHED_FNCL_PTR event, int32 new_period)
{
    SCHED_PERIODIC_EVENT* periodic_event;
    uint32 current_time;
    SCHED_QUEUE* cancel_queue;
    SCHED_PQ* resched_queue;

    cancel_queue = sched_queues_cancel_event(event, FALSE);
    EXCEPT_ASSERT(cancel_queue != NULL);
    event->cancelled = FALSE;
    periodic_event = sched_periodic_event_create(new_period, event);
    switch(cancel_queue->time_type)
    {
      case SCHED_REAL_TIME:
	resched_queue = &sched_realtime_queue;
	break;
      case SCHED_SIM_TIME:
	resched_queue = &sched_simtime_queue;
	break;
      case SCHED_PREEMPT_TICKS:
	resched_queue = &sched_preempt_queue;
	break;
      case SCHED_HIGH_PRIORITY:
	resched_queue = &sched_hipri_queue;
	break;
      default:
	EXCEPT_ASSERT(FALSE);
	resched_queue = &sched_realtime_queue;
	break;
    }
    current_time = sched_queue_current_time(SCHED_BASE_QUEUE(resched_queue));
    SCHED_QUEUE_SCHEDULE(SCHED_BASE_QUEUE(resched_queue),
			 SCHED_BASE_EVENT(periodic_event),
			 current_time + new_period);
}

