#ifndef lint
static char rcsid [] = "$RCSfile: sched_per.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_per.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libsched_local.h"
#include <stdalloc.h>
#include <stdio.h>
#include <libexcept.h>

static EVENT_OPS event_ops =
{
    (EVENT_DESTROY_FCN)sched_periodic_event_destroy,
    (EVENT_DELETE_FCN)sched_event_delete_func,
    (EVENT_INVOKE_FCN)sched_periodic_event_invoke,
    (EVENT_CANCEL_MATCH_FCN)sched_periodic_event_cancel_match
};

SCHED_PERIODIC_EVENT* sched_periodic_event_create(uint32 period,
						  SCHED_EVENT *a_event)
{
    SCHED_PERIODIC_EVENT* event;

    event = (SCHED_PERIODIC_EVENT*)STDALLOC(sizeof(SCHED_PERIODIC_EVENT));
    sched_event_init(SCHED_BASE_EVENT(event), 
		     &event_ops, 
		     a_event->group);
    event->period = period;
    event->event = a_event;
    event->no_ring = FALSE;
#ifdef DEBUG
    SCHED_BASE_EVENT(event)->name = strdup(a_event->name);
#endif
    return event;
}

void sched_periodic_event_destroy(SCHED_PERIODIC_EVENT *event)
{
    if (event->event)
    {
	SCHED_EVENT_DELETE(event->event);
	event->event = NULL;
    }
}

SCHED_EVENT_INVOKE_RESULT
sched_periodic_event_invoke(SCHED_PERIODIC_EVENT *event,
			    uint32 qtime,
			    SCHED_QUEUE *queue)
{
    SCHED_RING* ring;
    int32 scheduled;
    SCHED_EVENT_INVOKE_RESULT result;
    uint32 current_time;

    result = SCHED_EVENT_INVOKE(event->event, qtime, queue);
    EXCEPT_ASSERT(result != EVENT_RESCHEDULED);

    if (SCHED_BASE_EVENT(event)->cancelled)
    {
	/* If we have been cancelled then don't put
	 * the event on a ring.
	 */
	return EVENT_CANCELLED;
    }

    current_time = sched_queue_current_time(queue);

    if (event->no_ring)
    {
	ring = NULL;
    }
    else
    {
	ring = sched_find_ring(event->period, queue->time_type);
	if (ring == NULL)
	  event->no_ring = TRUE;
    }

    if (ring)
    {
	scheduled = sched_ring_schedule(ring, 
					event->event,
					current_time + event->period);

	if (scheduled)
	{
	    event->event = NULL;
	    /* We return EVENT_DONE, which will result in
	     * us being deleted.
	     */
	    return EVENT_DONE;
	}
#ifdef DEBUG	
	else
	{
	    printf("Couldn't get on ring %d\n", event->period);
	}
#endif
    }

    /* There is no ring of this period.
     * or we couldn't get scheduled on it.
     */
    scheduled = SCHED_QUEUE_SCHEDULE(queue,
				     SCHED_BASE_EVENT(event),
				     current_time + event->period);
    EXCEPT_ASSERT(scheduled);
    return EVENT_RESCHEDULED;
}


int32 sched_periodic_event_cancel_match(SCHED_PERIODIC_EVENT *event,
				        SCHED_EVENT *event_to_match)
{
    return (SCHED_BASE_EVENT(event) == event_to_match ||
	    event->event == event_to_match);
}


