#ifndef lint
static char rcsid[]= "$RCSfile: sched_python.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_python.c
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Tues Jul 30 10:12:23 EDT 1996
*       Author: cws
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef PYTHON_PARSER

#include <Py/Python.h>
#include <stdext.h>
#include "libsched.h"

static void
python_invoke_sched_fncl(PyObject* method, PyObject *arglist)
{
    PyObject *result;

    if ( method )
    {
	result = PyEval_CallObject(method, arglist);
	/*
	 * Don't Py_DECREF the arglist here - it may be used again
	 */
	if ( ! result )
	{
	    printf("There's a problem calling the scheduled method\n");
	    sched_cancel_fncl((SCHED_FNCL_PTR)python_invoke_sched_fncl);
	}
	else
	{
	    Py_DECREF(result);
	}
    }
}

static int32
python_sched_periodic_fncl( PyObject *method,
		        PyObject *arglist,
			uint32    initial_delay,
			uint32    period,
			int32     group
		      )
{

    int32 ret;

    ret = (int32) sched_periodic_fncl(  python_invoke_sched_fncl, 
				initial_delay, period,
				group,
				A_PTR, method,
				A_PTR, arglist,
				A_END );

    return(ret);
}

static int32
python_sched_deferred_fncl( PyObject *method,
			PyObject *arglist,
			uint32    delay_msecs,
			int32     group 
		      )
{
    int32 ret;

    ret = (int32) sched_deferred_fncl(  python_invoke_sched_fncl, 
				delay_msecs, group,
				A_PTR, method,
				A_PTR, arglist,
				A_END );

    return(ret);
}

static int32 
python_sched_simtime_fncl(	PyObject *method,
			PyObject *arglist,
			uint32    simulation_clock,
			int32     group 
		      )
{
    int32 ret;

    ret = (int32) sched_simtime_fncl(  python_invoke_sched_fncl, 
			       simulation_clock, group,
			       A_PTR, method,
			       A_PTR, arglist,
			       A_END );

    return(ret);
}

static void
python_sched_cancel_fncl( int32 sched_id
		    )
{
    sched_cancel_fncl((SCHED_FNCL_PTR) sched_id);
}

static void 
python_sched_cancel_fncl_group(   int32 group
			  )
{
    sched_cancel_fncl_group(group);
}

static void
python_change_fncl_period( int32 sched_id, 
		       int32 new_period_msecs
		     )
{
    sched_change_fncl_period((SCHED_FNCL_PTR)sched_id, new_period_msecs);
}

static PyObject *
python_setup_periodic_fncl(PyObject *self, PyObject *arg)
{
    PyObject *pfncl = NULL;
    PyObject *fargs;
    PyObject *retobj;
    uint32    initial_delay;
    uint32    period;
    int32     group;
    int32     retval;

    if ( arg )
    {
	PyArg_ParseTuple(arg, "OOiii", &pfncl, &fargs, 
				      &initial_delay, &period, &group);
	Py_INCREF(arg);
	Py_INCREF(pfncl);  /*the method or function object*/
	Py_INCREF(fargs);  /*the args for the function (as a tuple)*/

	fargs = Py_BuildValue("(O)", fargs);
	Py_INCREF(fargs);

	retval = python_sched_periodic_fncl(pfncl, fargs, 
					initial_delay, period,  group);
	
	retobj = Py_BuildValue("i", retval);
	return(retobj);
    }
    else
    {
	Py_INCREF(Py_None);
	return(Py_None);
    }
}

static PyObject *
python_setup_deferred_fncl(PyObject *self, PyObject *arg)
{
    PyObject *pfncl = NULL;
    PyObject *fargs;
    PyObject *retobj;
    uint32    delay_msecs;
    int32     group;
    int32     retval;

    if ( arg )
    {
	PyArg_ParseTuple(arg, "OOii", &pfncl, &fargs, 
				      &delay_msecs, &group);
	Py_INCREF(arg);
	Py_INCREF(pfncl);  /*the method or function object*/
	Py_INCREF(fargs);  /*the args for the function (as a tuple)*/

	fargs = Py_BuildValue("(O)", fargs);
	Py_INCREF(fargs);

	retval = python_sched_deferred_fncl(pfncl, fargs, delay_msecs, group);
	
	retobj = Py_BuildValue("i", retval);
	return(retobj);
    }
    else
    {
	Py_INCREF(Py_None);
	return(Py_None);
    }
}

static PyObject *
python_setup_simtime_fncl(PyObject *self, PyObject *arg)
{
    PyObject *pfncl = NULL;
    PyObject *fargs;
    PyObject *retobj;
    uint32    simtime_clock;
    int32     group;
    int32     retval;

    if ( arg )
    {
	PyArg_ParseTuple(arg, "OOii", &pfncl, &fargs, 
				      &simtime_clock, &group);
	Py_INCREF(arg);
	Py_INCREF(pfncl);  /*the method or function object*/
	Py_INCREF(fargs);  /*the args for the function (as a tuple)*/

	fargs = Py_BuildValue("(O)", fargs);
	Py_INCREF(fargs);

	retval = python_sched_deferred_fncl(pfncl, fargs, simtime_clock, group);
	
	retobj = Py_BuildValue("i", retval);
	return(retobj);
    }
    else
    {
	Py_INCREF(Py_None);
	return(Py_None);
    }
}

static PyObject *
python_remove_fncl(PyObject *self, PyObject *arg)
{
    int32 sched_id;

    if ( arg )
    {
	PyArg_ParseTuple(arg, "i", &sched_id);
	Py_INCREF(arg);

	python_sched_cancel_fncl(sched_id);
    }

    Py_INCREF(Py_None);
    return(Py_None);
}

static PyObject *
python_remove_fncl_group(PyObject *self, PyObject *arg)
{
    int32 group;

    if ( arg )
    {
	PyArg_ParseTuple(arg, "i", &group);
	Py_INCREF(arg);

	python_sched_cancel_fncl_group(group);
    }

    Py_INCREF(Py_None);
    return(Py_None);
}

static PyObject *
python_update_fncl_period(PyObject *self, PyObject *arg)
{
    int32 sched_id;
    int32 new_period_msecs;

    if ( arg )
    {
	PyArg_ParseTuple(arg, "ii", &sched_id, &new_period_msecs);
	Py_INCREF(arg);

	python_change_fncl_period(sched_id, new_period_msecs);
    }

    Py_INCREF(Py_None);
    return(Py_None);
}

static PyMethodDef schedMethods[] =
{
    {   "periodic_function",	python_setup_periodic_fncl,	1	},
    {   "deferred_function",	python_setup_deferred_fncl,	1	},
    {   "simtime_function",	python_setup_simtime_fncl,	1	},
    {   "remove_function",	python_remove_fncl,		1	},
    {   "remove_group",		python_remove_fncl_group,	1	},
    {   "update_period",	python_update_fncl_period,	1	},
    {	NULL,			NULL			}
};

void
sched_init_python()
{
    (void) Py_InitModule("sched_if", schedMethods);
    PyRun_SimpleString("import sched");
}


#endif
