#ifndef lint
static char rcsid [] = "$RCSfile: sched_queue.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_queue.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libsched_local.h"
#include <libtime.h>
#include <libexcept.h>
#include <stdio.h>

SCHED_QUEUE* sched_queue_list;

void sched_queue_init(SCHED_QUEUE *queue,
		      QUEUE_OPS *ops,
		      SCHED_TIME_TYPE time_type)
{
    queue->ops = ops;
    queue->time_type = time_type;

    /* Now put the queue on the list. */
    queue->next_queue = sched_queue_list;
    if (queue->next_queue)
      queue->next_queue->prev_queue = queue;
    queue->prev_queue = NULL;
    sched_queue_list = queue;
}

void sched_queue_destroy(SCHED_QUEUE *queue)
{
    /* Take the queue off the list. */
    if (!queue->prev_queue)
    {
	EXCEPT_ASSERT(sched_queue_list == queue);
	sched_queue_list = queue->next_queue;
    }
    else
    {
	queue->prev_queue->next_queue = queue->next_queue;
    }

    if (queue->next_queue)
    {
	queue->next_queue->prev_queue = queue->prev_queue;
    }
}


uint32 sched_queue_current_time(SCHED_QUEUE *queue)
{
    switch(queue->time_type)
    {
      case SCHED_REAL_TIME:
      case SCHED_HIGH_PRIORITY:
	return time_last_realtime_clock;
      case SCHED_SIM_TIME:
	return time_last_simulation_clock;
      case SCHED_PREEMPT_TICKS:
	return time_last_realtime_clock / SCHED_PREEMPT_MSECS_PER_TICK;
      default:
	EXCEPT_ASSERT(FALSE);	/* We shouldn't be here */
	return SCHED_MAX_TIME;
    }
}

uint32 sched_queue_real_time(SCHED_QUEUE *queue, uint32 qtime)
{
    switch(queue->time_type)
    {
      case SCHED_REAL_TIME:
      case SCHED_HIGH_PRIORITY:
	return qtime;
      case SCHED_SIM_TIME:
	return time_simulation_to_realtime(qtime);
      case SCHED_PREEMPT_TICKS:
	return qtime * SCHED_PREEMPT_MSECS_PER_TICK;
      default:
	EXCEPT_ASSERT(FALSE);	/* We shouldn't be here */
	return SCHED_MAX_TIME;
    }
}	



void sched_queues_cancel_group(int32 group)
{
    SCHED_QUEUE* queue;

    for (queue = sched_queue_list; queue; queue = queue->next_queue)
    {
	SCHED_QUEUE_CANCEL_GROUP(queue, group);
    }
}

SCHED_QUEUE* sched_queues_cancel_event(SCHED_EVENT *event, int32 delete_event)
{
    SCHED_QUEUE* queue;

    for (queue = sched_queue_list; queue; queue = queue->next_queue)
    {
	if (SCHED_QUEUE_CANCEL_EVENT(queue, event, delete_event))
	{
	    return queue;
	}
    }
    return NULL;
}

    
