#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libsched                                 *
*       Created: Tue Jun  2 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libsched.h"
#include <libtime.h>
#include <librandom.h>
#include <libexcept.h>
#include <stdext.h> /*common/include/global*/
#include "sched_event.h"

SCHED_FNCL_PTR remotes[800];
SCHED_FNCL_PTR composites[40];
SCHED_FNCL_PTR vehicles[120];
SCHED_FNCL_PTR spuratics[120];
int32 dead_remotes[800];
int32 dead_vehicles[120];
int32 remote_timed_out[800];

static void wait_until(deadline)
    uint32 deadline;
{
    while (time_realtime_clock() < deadline) ;
}

static void remote(id)
    int32 id;
{
    if (remote_timed_out[id])
      abort();

    if (rnd_probability(0.10))
    {
	printf("Timeout %d\n",id);
	sched_cancel_fncl(remotes[id]);
	remote_timed_out[id] = 1;
    }
    else if (rnd_probability(0.05) && !dead_remotes[id])
    {
	printf("Dead %d\n",id);
	dead_remotes[id] = 1;
	sched_change_fncl_period(remotes[id],4000);
    }
}

static void composite()
{
    wait_until(time_realtime_clock()+10);
}

static void spuratic(delay,id,realtime)
    uint32 delay;
    int32 id;
    int32 realtime;
{
    printf("Spuratic %d @ %d\n", id,
	   realtime ? time_last_realtime_clock : time_last_simulation_clock);
    spuratics[id] = 0;
    wait_until(time_realtime_clock()+delay);

    time_set_simulation_rate(rnd_uniform_dist()*2.0, 0);
}

static void vehicle(id)
    int32 id;
{
    uint32 delay, when;

    if (rnd_probability(0.10) && !dead_vehicles[id])
    {
	printf("Killing %d\n",id);
	sched_change_fncl_period(vehicles[id],4000);
	dead_vehicles[id] = 1;
    }
    if (!dead_vehicles[id])
    {
	if (rnd_probability(0.05))
	{
	    delay = rnd_uniform_dist()*2000;
	    printf("Spuratic random realtime event for %d @ %d\n",
		   id, time_last_realtime_clock+delay);

	    sprintf(event_name, 	
		    "Spuratic %d @ %d", id, time_last_realtime_clock+delay);
	    spuratics[id] =
	      sched_deferred_fncl(spuratic,
				  delay, id,
				  A_INT, (uint32)(rnd_uniform_dist()*30),
				  A_INT, id,
				  A_INT, TRUE,
				  A_END);
	}
	if (rnd_probability(0.05))
	{
	    when = (time_last_simulation_clock +
		    (uint32)(rnd_uniform_dist()*2000));
	    printf("Spuratic random simtime event for %d @ %d\n",
		   id, when);
	    sprintf(event_name,
		    "Spuratic simtime %d @ %d", id, time_last_realtime_clock+delay);
	    spuratics[id] =
	      sched_simtime_fncl(spuratic,
				 when, id,
				 A_INT, (uint32)(rnd_uniform_dist()*30),
				 A_INT, id,
				 A_INT, FALSE,
				 A_END);
	}
	if (rnd_probability(0.05) && spuratics[id])
	{
	    printf("Cancelling spuratic for %d\n",id);
	    sched_cancel_fncl(spuratics[id]);
	    spuratics[id] = 0;
	}
	wait_until(time_realtime_clock()+
		   (rnd_probability(0.05) ? 50 : 3));
    }
}

simnet()
{
    static int32 count = 0;

    if (count < 800)
    {
	if (rnd_probability(0.10))
	{
	    printf("New remote %d\n",count);
	    
	    sprintf(event_name, "remote %d", count);
	    remotes[count] =
	      sched_periodic_fncl(remote,SCHED_ASAP,67,count,
				  A_INT,count,A_END);
	    count++;
	}
	else if (rnd_probability(0.10))
	{
	    printf("New static remote %d\n",count);
	    sprintf(event_name, "state remote %d", count);
	    remotes[count] =
	      sched_periodic_fncl(remote,SCHED_ASAP,30000,count,
				  A_INT,count,A_END);
	    count++;
	}
    }
    wait_until(time_realtime_clock()+10);
}

sbx()
{
    static int32 ccount = 0;
    static int32 vcount = 0;

    if (ccount < 40)
    {
	if (rnd_probability(0.10))
	{
	    printf("New platoon %d %d %d %d\n",
		   ccount,vcount,vcount+1,vcount+2);
	    sprintf(event_name, "Composite %d", ccount);
	    composites[ccount] =
	      sched_periodic_fncl(composite,SCHED_ASAP,20000,ccount,A_END);
	    ccount++;
	    sprintf(event_name, "Vehicle %d", vcount);
	    vehicles[vcount] =
	      sched_periodic_fncl(vehicle,SCHED_ASAP,67,vcount,
				  A_INT,vcount,A_END);
	    vcount++;
	    sprintf(event_name, "Vehicle %d", vcount);
	    vehicles[vcount] =
	      sched_periodic_fncl(vehicle,SCHED_ASAP,67,vcount,
				  A_INT,vcount,A_END);
	    vcount++;
	    sprintf(event_name, "Vehicle %d", vcount);
	    vehicles[vcount] =
	      sched_periodic_fncl(vehicle,SCHED_ASAP,67,vcount,
				  A_INT,vcount,A_END);
	    vcount++;
	}
	wait_until(time_realtime_clock()+10);
    }
    else
    {
	printf("Reset\n");
	while (--ccount > 0)
	  sched_cancel_fncl(composites[ccount]);
	while (--vcount > 0)
	  sched_cancel_fncl(vehicles[vcount]);
	bzero(dead_vehicles,sizeof(dead_vehicles));
	bzero(spuratics,sizeof(spuratics));
	wait_until(time_realtime_clock()+1000);
    }
}

static void gasp()
{
    printf("*\n**\n*** GASP! Out of cycles!\n**\n*\n");
}

static void sigh()
{
    printf("  *\n **\n*** Sigh.  All better.\n **\n  *\n");
}

main()
{
    bzero(dead_vehicles,sizeof(dead_vehicles));
    bzero(spuratics,sizeof(spuratics));
    time_init();
    except_init(TRUE, FALSE);
    sched_init(6,67,100,1000,4000,20000,30000);
    sched_perf_monitor_on(30000);
    sched_set_critical_performance(67, 0.90, 500, 15000,
				   gasp, sigh);
    sprintf(event_name, "simnet");
    sched_periodic_fncl(simnet,SCHED_ASAP,67,0,A_END);
    sprintf(event_name, "sbx");
    sched_periodic_fncl(sbx,5000,100,0,A_END);
    sched_invoke_functions_until(time_realtime_clock()+241000);

    printf("\n\nIf the program got here, you can assume everything works.\n");
}

