#ifndef lint
static char rcsid [] = "$RCSfile: time_test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libsched                                 *
*       Created: Tue Jun  2 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libsched.h"
#include <libtime.h>
#include <librandom.h>
#include <libexcept.h>
#include <stdext.h> /*common/include/global*/
#include <stdstring.h>
#include "sched_event.h"

#include <stdio.h>
#include <stdalloc.h>

#include <libparser.h>
#include <libtty.h>


SCHED_FNCL_PTR remotes[800];
SCHED_FNCL_PTR composites[40];
SCHED_FNCL_PTR vehicles[120];
SCHED_FNCL_PTR spuratics[120];
int32 dead_remotes[800];
int32 dead_vehicles[120];
int32 remote_timed_out[800];

/*ARGSUSED*/
static void test_print_time(argc, argv)
    int32 argc;
    int32 argv[];
{
    printf("Sim time: %d\tReal time: %d\n", 
	   time_last_simulation_clock,
	   time_last_realtime_clock);
}

/*ARGSUSED*/
static void set_rate(argc, argv)
    int32 argc;
    int32 argv[];
{
    float32 rate = *(float32 *)argv;
    int32 limit = argv[1];

    if ((rate > 0.999) && (rate < 1.001) || (rate < 0.0))
    {
	printf("Running in real time\n");
	time_set_simulation_rate(1.0, (uint32)0);
    }
    else if (limit)
    {
	printf("Setting simulation to %.2f x real time for %d (simulated) seconds\n",
	       rate, limit),
	time_set_simulation_rate(rate,
				 limit * 1000 + time_last_simulation_clock);
    }
    else
    {
	printf("Setting simulation to %.2f x real time\n",
	       rate, limit),
	time_set_simulation_rate(rate, (uint32)0);
    }
}

/*ARGSUSED*/
static void run_rate(argc, argv)
    int32 argc;
    int32 argv[];
{
    float32 rate = *(float32 *)argv;
    int32 limit = argv[1];

    if ((rate > 0.999) && (rate < 1.001) || (rate < 0.0))
    {
	printf("Running in real time");
	rate = 1.0;
    }
    else 
    {
	printf("Setting simulation to %.2f x real time", rate);
    }

    if (limit)
    {
	printf(" for %d (simulated) seconds\n", limit),
	time_run_simulation_rate(rate,
				 limit * 1000 + time_last_simulation_clock);
    }
    else
    {
	printf("\n");
	time_run_simulation_rate(rate, (uint32)0);
    }
}

/*ARGSUSED*/
static void test_done()
{
    exit(0);
}

DEFINE_TABLE(top_command_table)

    KEYWORD_SELECT("    Test Commands")

	KEYWORD("quit", "-exit")
	    CALL(test_done)
	END_KEYWORD

	KEYWORD("exit", "-quit")
	    CALL(test_done)
	END_KEYWORD

        KEYWORD ("printtime", "- print commands")
            CALL(test_print_time)
        END_KEYWORD

	KEYWORD ("rate", "- modify simulation rate")
	    GETFLOAT("simulation rate (x real time)")
	    GETDECIMAL("duration of modification in seconds (0 for no limit)")
	    CALL (set_rate)
	END_KEYWORD

	KEYWORD ("run", "- run and return to current rate")
	    GETFLOAT("simulation rate (x real time)")
	    GETDECIMAL("duration of modification in seconds (0 for no limit)")
	    CALL (run_rate)
	END_KEYWORD

    END_KEYWORD_SELECT

END_DEFINE_TABLE

extern PARSE_TABLE top_command_table[];

main()
{

    time_init(5000);
    except_init(TRUE, FALSE);
    sched_init(6,67,100,1000,4000,20000,30000);

    tty_parser_init(top_command_table, "test> ");

    sched_periodic_fncl(tty_loop_tick, 67, 67, 0, A_END);
    sched_invoke_functions_until(time_realtime_clock()+241000);
}




