/*
 * $RCSfile: libsnstve_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libsnstve_local.h                                                 *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libsensitive                      *
*       Created: Fri Sep 11 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libsensitive.h"
#include <X11/Intrinsic.h>

/* For simplicity, all sensitive windows are the same size */
#define SNSTVE_TREE_SIZE 2048 /* pixels per side of root node */
#define SNSTVE_LEAF_SIZE 32   /* pixels per side of a leaf node */
#define SNSTVE_NODES     5461 /* 4^0+4^1+4^2+...+4^5 */

typedef struct snstve_node
{
    uint16 alloced;
    uint16 used;
    SNSTVE_OBJECT *objects;
} SNSTVE_NODE;

struct snstve_window
{
    Widget widget;

    GC draw_gc, erase_gc;

    CALLBACK_EVENT_PTR miss_callback;

    /* We do this on the first click, so we know there is a
     * realized window to munch on.
     */
    int32 added_grab;

    int32 object_pressed;
    SNSTVE_OBJECT press_object;
    uint32 press_x, press_y;
    int32 button;
    int32 dragged, drag_dx, drag_dy;

    SNSTVE_OBJECT *hot_object;

    SNSTVE_NODE nodes[SNSTVE_NODES];
};

/* Private prototypes */
extern void snstve_gesture(Widget widget, XtPointer w, XEvent *event,
			   Boolean *unused);
extern void snstve_motion(SNSTVE_WINDOW_PTR window, int x, int y, 
			  SNSTVE_OBJECT *hot_copy);
extern void snstve_leave(Widget widget, XtPointer w, XEvent *event,
			 Boolean *unused);
extern SNSTVE_OBJECT *snstve_find_hot(SNSTVE_WINDOW_PTR window, int x, int y);
extern SNSTVE_OBJECT *snstve_find_press(SNSTVE_WINDOW_PTR window, int x, int y,
					unsigned int button);
extern void snstve_clear_hot(SNSTVE_WINDOW_PTR window);
extern void snstve_get_hot(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *hot_copy);
extern void snstve_redraw_press(SNSTVE_WINDOW_PTR window, GC gc);
