#ifndef lint
static char rcsid [] = "$RCSfile: snstve_mouse.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: snstve_mouse.c                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle mouse callbacks                            *
*       Created: Fri Sep 11 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libsnstve_local.h"
#include <X11/cursorfont.h>
#include <math.h>

static void highlight(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *object, GC gc, int32 xoff, int32 yoff);

/*ARGSUSED*/
void snstve_gesture(Widget widget, XtPointer w, XEvent *event, Boolean *unused)
{
    XEvent next_event;
    SNSTVE_OBJECT *object;
    SNSTVE_CALL_DATA call_data;
    SNSTVE_ACTION action;
    int32 thresh;
    SNSTVE_WINDOW_PTR window = (SNSTVE_WINDOW_PTR)w;

    if (!window->added_grab)
    {
	XGrabButton(XtDisplay(widget), AnyButton, AnyModifier,
		    XtWindow(widget), TRUE,
		    ButtonPressMask | ButtonMotionMask | ButtonReleaseMask,
		    GrabModeAsync, GrabModeAsync, XtWindow(widget),
		    XCreateFontCursor(XtDisplay(widget), XC_crosshair));
	window->added_grab = 1;
    }

    /* Where there are multiple motion events in X's queue, skip all
     * but the latest one.
     */
    if ((event->type == MotionNotify) &&
	(XtAppPending(XtWidgetToApplicationContext(widget)) & XtIMXEvent))
    {
	XPeekEvent(XtDisplay(widget), &next_event);
	if ((next_event.type == MotionNotify) &&
	    (next_event.xmotion.state == event->xmotion.state))
	  return;
    }	

    /* Send non-button motion notifies to snstve_motion */
    if ((event->type == MotionNotify) &&
	!(event->xmotion.state &
	  (Button1Mask|Button2Mask|Button3Mask|Button4Mask|Button5Mask)))
    {
	snstve_motion(window, event->xmotion.x, event->xmotion.y,
		      (SNSTVE_OBJECT *)NULL);
	return;
    }

    if (event->type == ButtonPress)
    {
	/* Protect ourselves from chords.  We set this when a button
	 * is pressed, and zero it out when *that* button is released.
	 * In between, we refuse to process other buttons.
	 */
	if (window->button)
	  return;

	/* If there is a hot object which responds to this button,
	 * then the press *must* go to that object.
	 */
	if (window->hot_object &&
	    (window->hot_object->sensClass->buttons &
	     ((1 << event->xbutton.button) | 1)))
	  object = window->hot_object;
	else
	  object = snstve_find_press(window,
				     event->xbutton.x, event->xbutton.y,
				     event->xbutton.button);

	snstve_clear_hot(window);

	action = SNSTVE_PRESS;
	call_data.button = event->xbutton.button;
	call_data.screen_x = event->xbutton.x;
	if (call_data.screen_x & 0x8000)
	  call_data.screen_x = 0;
	call_data.screen_y = event->xbutton.y;
	if (call_data.screen_y & 0x8000)
	  call_data.screen_y = 0;
	call_data.xoff = call_data.yoff = 0;

	/* Save these for later */
	window->button = call_data.button;
	window->press_x = call_data.screen_x;
	window->press_y = call_data.screen_y;
	window->dragged = window->drag_dx = window->drag_dy = 0;
	if (object)
	{
	    window->press_object = *object;
	    window->object_pressed = 1;
	    highlight(window, object, window->draw_gc, 0, 0);
	}
	else
	  window->object_pressed = 0;
    }
    else if (event->type == MotionNotify)
    {
	action = SNSTVE_MOTION;
	call_data.button = window->button;
	call_data.screen_x = event->xmotion.x;
	if (call_data.screen_x & 0x8000)
	  call_data.screen_x = 0;
	call_data.screen_y = event->xmotion.y;
	if (call_data.screen_y & 0x8000)
	  call_data.screen_y = 0;
	call_data.xoff = event->xmotion.x - window->press_x;
	call_data.yoff = event->xmotion.y - window->press_y;
	if (window->object_pressed)
	{
	    object = &window->press_object;
	    /* Once you start dragging, don't check the thresholds
	     * anymore.
	     */
	    thresh = object->sensClass->drag_thresh;
	    if (window->dragged || (object->sensClass->dragable &&
				    ((call_data.xoff < -thresh) ||
				     (call_data.xoff > thresh) ||
				     (call_data.yoff < -thresh) ||
				     (call_data.yoff > thresh))))
	    {
		highlight(window, object, window->erase_gc,
			  window->drag_dx, window->drag_dy);
		window->drag_dx = call_data.xoff;
		window->drag_dy = call_data.yoff;
		highlight(window, object, window->draw_gc,
			  window->drag_dx, window->drag_dy);
		window->dragged = 1;
	    }	      
	}
	else
	  object = NULL;
    }
    else if (event->type == ButtonRelease)
    {
	/* Protect ourselves from chords */
	if (event->xbutton.button != window->button)
	  return;

	/* Zero this out, indicating new button presses are acceptable */
	window->button = 0;

	action = SNSTVE_RELEASE;
	call_data.button = event->xbutton.button;
	call_data.screen_x = event->xbutton.x;
	if (call_data.screen_x & 0x8000)
	  call_data.screen_x = 0;
	call_data.screen_y = event->xbutton.y;
	if (call_data.screen_y & 0x8000)
	  call_data.screen_y = 0;
	call_data.xoff = event->xbutton.x - window->press_x;
	call_data.yoff = event->xbutton.y - window->press_y;
	if (window->object_pressed)
	{
	    object = &window->press_object;
	    highlight(window, object, window->erase_gc,
		      window->drag_dx, window->drag_dy);

	    /* Dragables return 0 offsets if they weren't dragged past
	     * their thresholds.
	     */
	    if (object->sensClass->dragable && !window->dragged)
	      call_data.xoff = call_data.yoff = 0;
	}
	else
	  object = NULL;
	window->object_pressed = 0;
    }

    if (object)
    {
	call_data.sensClass = object->sensClass;
	call_data.instance = object->instance;
	call_data.user_data = object->user_data;
	if (object->sensClass->gesture)
	  callback_fire_event(object->sensClass->gesture,
			      window, action, &call_data);
    }
    else
    {
	call_data.sensClass = NULL;
	call_data.instance = -1;
	call_data.user_data = NULL;
	if (window->miss_callback)
	  callback_fire_event(window->miss_callback,
			      window, action, &call_data);
    }
}

void snstve_refresh(SNSTVE_WINDOW_PTR window)
{
    snstve_redraw_press(window, window->draw_gc);
    if (window->hot_object)
      highlight(window, window->hot_object, window->draw_gc, 0, 0);
}

void snstve_redraw_press(SNSTVE_WINDOW_PTR window, GC gc)
{
    if (window->object_pressed)
      highlight(window, &window->press_object, gc,
		window->drag_dx, window->drag_dy);
}

void snstve_clear_hot(SNSTVE_WINDOW_PTR window)
{
    SNSTVE_CALL_DATA call_data;

    if (!window->hot_object)
      return;

    highlight(window, window->hot_object, window->erase_gc, 0, 0);
    if (window->hot_object->sensClass->excited)
    {
	call_data.button = 0;
	call_data.screen_x = call_data.screen_y = 0;
	call_data.xoff = call_data.yoff = 0;
	call_data.sensClass = window->hot_object->sensClass;
	call_data.instance = window->hot_object->instance;
	call_data.user_data = window->hot_object->user_data;
	callback_fire_event(window->hot_object->sensClass->excited,
			    window, SNSTVE_UNEXCITED, &call_data);
    }

    window->hot_object = NULL;
}

void snstve_motion(SNSTVE_WINDOW_PTR window, int x, int y, SNSTVE_OBJECT *hot_copy)
{
    SNSTVE_CALL_DATA call_data;
    SNSTVE_OBJECT *object;

    /* Ignore motion events when there is a pressed object */
    if (window->object_pressed)
      return;

    /* Find the object under the pointer */
    object = snstve_find_hot(window, x, y);

    /* If hot_copy is non-NULL, the caller made a copy of the hot object
     * which we should use instead of the value of window->hot_object.
     */
    if (hot_copy)
    {
	/* Erase the old hot object */
	highlight(window, hot_copy, window->erase_gc, 0, 0);

	/* If the same object is currently hot, draw the highlight in the
	 * new position and return.
	 */
	if (object &&
	    (object->sensClass == hot_copy->sensClass) &&
	    (object->instance == hot_copy->instance))
	{
	    /* Draw the highlight in its current position */
	    highlight(window, object, window->draw_gc, 0, 0);

	    window->hot_object = object;
	    return;
	}

	/* The hot object has changed, so notify the caller */
	snstve_clear_hot(window);
    }
    else
    {
	/* See if there has been no change */
	if (object == window->hot_object)
	  return;

	/* Turn off the old hot object, if any */
	snstve_clear_hot(window);
    }

    /* Turn on the new hot object, if any */
    if (object)
    {
	highlight(window, object, window->draw_gc, 0, 0);
	if (object->sensClass->excited)
	{
	    call_data.button = 0;
	    call_data.screen_x = x;
	    if (call_data.screen_x & 0x8000)
	      call_data.screen_x = 0;
	    call_data.screen_y = y;
	    if (call_data.screen_y & 0x8000)
	      call_data.screen_y = 0;
	    call_data.xoff = call_data.yoff = 0;
	    call_data.sensClass = object->sensClass;
	    call_data.instance = object->instance;
	    call_data.user_data = object->user_data;
	    callback_fire_event(object->sensClass->excited,
				window, SNSTVE_EXCITED, &call_data);
	}
    }

    window->hot_object = object;
}

void snstve_get_hot(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *hot_copy)
{
    int root_x, root_y, win_x, win_y;
    unsigned int mask;
    Window root, child;

    /* Get the current pointer position */
    if (XQueryPointer(XtDisplay(window->widget),
		      XtWindow(window->widget),
		      &root, &child,
		      &root_x, &root_y, &win_x, &win_y, &mask) &&
	(child == None))
    {
	snstve_motion(window, win_x, win_y, hot_copy);
    }
}

/*ARGSUSED*/
void snstve_leave(Widget widget, XtPointer w, XEvent *event, Boolean *unused)
{
    SNSTVE_WINDOW_PTR window = (SNSTVE_WINDOW_PTR)w;
    /* Ignore motion events when there is a pressed object */
    if (window->object_pressed)
      return;

    /* Turn off the old hot object, if any */
    snstve_clear_hot(window);
}

static void highlight(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *object, GC gc, int32 xoff, int32 yoff)
{
    float64 mag, dx, dy;
    XPoint points[5];

    switch(object->kind)
    {
      case SNSTVE_RECTANGLE:
	XDrawRectangle(XtDisplay(window->widget),
		       XtWindow(window->widget),
		       gc,
		       xoff + object->attr.rect.x,
		       yoff + object->attr.rect.y,
		       object->attr.rect.width, object->attr.rect.height);
	break;
      case SNSTVE_CIRCLE:
	XDrawArc(XtDisplay(window->widget),
		 XtWindow(window->widget),
		 gc,
		 xoff + object->attr.circle.x - object->attr.circle.radius,
		 yoff + object->attr.circle.y - object->attr.circle.radius,
		 2*object->attr.circle.radius, 2*object->attr.circle.radius,
		 0, 360*64);
	break;
      case SNSTVE_LINE:
	/* Deal with horizontal or vertical lines as a special case */
	if ((object->attr.line.x0 == object->attr.line.x1) ||
	    (object->attr.line.y0 == object->attr.line.y1))
	  XDrawRectangle(XtDisplay(window->widget),
			 XtWindow(window->widget),
			 gc,
			 xoff + object->xlow, yoff + object->ylow,
			 object->xhigh - object->xlow,
			 object->yhigh - object->ylow);
	else
	{
	    dy = (object->attr.line.x1 - object->attr.line.x0);
	    dx = (object->attr.line.y0 - object->attr.line.y1);
	    mag = 1.0 / sqrt(dx * dx + dy * dy);
	    dx *= mag * object->attr.line.width * 0.5;
	    dy *= mag * object->attr.line.width * 0.5;

	    points[0].x = object->attr.line.x0 + dx + xoff;
	    points[0].y = object->attr.line.y0 + dy + yoff;
	    points[1].x = object->attr.line.x1 + dx + xoff;
	    points[1].y = object->attr.line.y1 + dy + yoff;
	    points[2].x = object->attr.line.x1 - dx + xoff;
	    points[2].y = object->attr.line.y1 - dy + yoff;
	    points[3].x = object->attr.line.x0 - dx + xoff;
	    points[3].y = object->attr.line.y0 - dy + yoff;
	    points[4] = points[0];
	    XDrawLines(XtDisplay(window->widget),
		       XtWindow(window->widget),
		       gc,
		       points, 5, CoordModeOrigin);
	}
    }
}

#endif /*USE_X*/
