#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program                                              *
*       Created: Mon Sep 14 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: See the texinfo documentation for details of operation     *
*                                                                           *
****************************************************************************/

#include "libsensitive.h"
#include <librandom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <stdext.h> /*common/include/global*/

#undef TIME

#ifdef TIME
#include <sys/types.h>
#include <sys/timeb.h>
#endif

#define OBJECT_COUNT 1000

SNSTVE_CLASS stationary_rectangle;
SNSTVE_CLASS stationary_circle;
SNSTVE_CLASS stationary_line;
SNSTVE_CLASS dragable_rectangle;
SNSTVE_CLASS dragable_circle;
SNSTVE_CLASS dragable_line;
SNSTVE_CLASS stationary_moving;
SNSTVE_CLASS dragable_moving;
#define N_OBJECT_TYPES        6

#define RECTANGLE(x) \
  ((((x)==&stationary_rectangle) || ((x)==&dragable_rectangle)))
#define CIRCLE(x) \
  ((((x)==&stationary_circle) || ((x)==&dragable_circle)))
#define LINE(x) \
  ((((x)==&stationary_line) || ((x)==&dragable_line)))

#define DRAGABLE(x) \
  ((x==&dragable_rectangle) || (x==&dragable_circle) || (x==&dragable_line))

#define HOT(x) 1

/* Rectangles:left, Circles:right or middle, Lines:any */
#define BUTTONS(x) \
  (RECTANGLE(x) ? 0x2 : (CIRCLE(x) ? 0xC : 0x1))

#define KIND(x) \
  (RECTANGLE(x) ? SNSTVE_RECTANGLE : \
   (CIRCLE(x) ? SNSTVE_CIRCLE : SNSTVE_LINE))

static Widget toplevel;
static Display *display;
static Window window;
static SNSTVE_WINDOW_PTR sensitive;
static Pixmap pixmap = NULL;
static GC draw, erase;
static uint32 width=0, height=0;
static struct
{
    GC black;
    GC white;
    GC red;
    GC blue;
    GC green;
    GC orange;
    GC yellow;
    GC violet;
} colors;
static SNSTVE_OBJECT lines[OBJECT_COUNT];
static SNSTVE_OBJECT circles[OBJECT_COUNT];
static SNSTVE_OBJECT rectangles[OBJECT_COUNT];
static int32 count;
static SNSTVE_OBJECT s_circle, d_circle;
static int32 jump_set=0, jump_dx, jump_dy;

static void make_colors(void)
{
    XColor clr, ign;
    XGCValues gcv;
    Colormap colormap = DefaultColormapOfScreen(XtScreen(toplevel));

    XAllocNamedColor(display, colormap, "black", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.black = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "white", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.white = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "red", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.red = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "blue", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.blue = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "green", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.green = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "orange", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.orange = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "yellow", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.yellow = XCreateGC(display, window, GCForeground, &gcv);
    XAllocNamedColor(display, colormap, "violet", &clr, &ign);
    gcv.foreground = clr.pixel;
    colors.violet = XCreateGC(display, window, GCForeground, &gcv);
}

static void change_size(uint32 w, uint32 h)
{
    if ((w==width) && (h==height))
      return;

    width = w;
    height = h;

    if (pixmap)
      XFreePixmap(display, pixmap);

    pixmap = XCreatePixmap(display, window, width, height,
			   DefaultDepthOfScreen(XtScreen(toplevel)));

    XFillRectangle(display, pixmap, colors.black,
		   0, 0, width, height);

    XSetTile(display, erase, pixmap);
    XSetWindowBackgroundPixmap(display, window, pixmap);
    XClearWindow(display, window);
}    

static void resize(Widget w, int32 ignored, XEvent *event)
{
    if (event->type == ConfigureNotify)
      change_size(event->xconfigure.width, event->xconfigure.height);
}    

static void callback(SNSTVE_WINDOW_PTR sensitive, SNSTVE_ACTION action, SNSTVE_CALL_DATA *call_data)
{
    int32 i;
    int32 old_x, old_y;
    int32 radius;

    switch(action)
    {
      case SNSTVE_EXCITED:
	return "SNSTVE_EXCITED";
      case SNSTVE_UNEXCITED:
	return "SNSTVE_UNEXCITED";
      case SNSTVE_PRESS:
	return "SNSTVE_PRESS";
      case SNSTVE_MOTION:
	return "SNSTVE_MOTION";
      case SNSTVE_RELEASE:
	return "SNSTVE_RELEASE";
    }
    printf(" Object <0x%x %d> Button %d @ %d %d (delta = %d %d)\n",
	   call_data->class, call_data->instance,
	   call_data->button,
	   call_data->screen_x, call_data->screen_y,
	   call_data->xoff, call_data->yoff);

    if ((action == SNSTVE_RELEASE) &&
	DRAGABLE(call_data->class))
    {
	i = call_data->instance;
	if (call_data->class == &dragable_line)
	{
	    XDrawLine(display, pixmap, colors.black,
		      lines[i].attr.line.x0, lines[i].attr.line.y0,
		      lines[i].attr.line.x1, lines[i].attr.line.y1);

	    old_x = lines[i].attr.line.x0;
	    old_y = lines[i].attr.line.y0;
	    lines[i].attr.line.x0 += call_data->xoff;
	    lines[i].attr.line.x1 += call_data->xoff;
	    lines[i].attr.line.y0 += call_data->yoff;
	    lines[i].attr.line.y1 += call_data->yoff;
	    snstve_change_object(sensitive, old_x, old_y, lines+i);

	    XDrawLine(display, pixmap, colors.orange,
		      lines[i].attr.line.x0, lines[i].attr.line.y0,
		      lines[i].attr.line.x1, lines[i].attr.line.y1);
	}
	else if (call_data->class == &dragable_rectangle)
	{
	    XFillRectangle(display, pixmap, colors.black,
			   rectangles[i].attr.rect.x,
			   rectangles[i].attr.rect.y,
			   rectangles[i].attr.rect.width,
			   rectangles[i].attr.rect.height);

	    old_x = rectangles[i].attr.rect.x;
	    old_y = rectangles[i].attr.rect.y;
	    rectangles[i].attr.rect.x += call_data->xoff;
	    rectangles[i].attr.rect.y += call_data->yoff;
	    snstve_change_object(sensitive, old_x, old_y, rectangles+i);

	    XFillRectangle(display, pixmap, colors.red,
			   rectangles[i].attr.rect.x,
			   rectangles[i].attr.rect.y,
			   rectangles[i].attr.rect.width,
			   rectangles[i].attr.rect.height);
	}
	else if (call_data->class == &dragable_circle)
	{
	    radius = circles[i].attr.circle.radius - 2;
	    XFillArc(display, pixmap, colors.black,
		     circles[i].attr.circle.x-radius,
		     circles[i].attr.circle.y-radius,
		     2*radius, 2*radius, 0, 360*64);

	    old_x = circles[i].attr.circle.x;
	    old_y = circles[i].attr.circle.y;
	    circles[i].attr.circle.x += call_data->xoff;
	    circles[i].attr.circle.y += call_data->yoff;
	    snstve_change_object(sensitive, old_x, old_y, circles+i);

	    XFillArc(display, pixmap, colors.green,
		     circles[i].attr.circle.x-radius,
		     circles[i].attr.circle.y-radius,
		     2*radius, 2*radius, 0, 360*64);
	}
	/* Refresh */
	XClearWindow(display, window);
	snstve_refresh(sensitive);
    }
    else if ((action == SNSTVE_RELEASE) &&
	     (call_data->class == &dragable_moving))
    {
	/* This is the dragable moving circle.  Set a flag, so it'll
	 * jump here on its next tick.
	 */
	jump_set = 1;
	jump_dx = call_data->xoff;
	jump_dy = call_data->yoff;
    }
}

#ifdef TIME
static uint32 total, count;
static struct timeb tstart, tend;

#define TIME_INIT total = count = 0
#define TIME_START ftime(&tstart)
#define TIME_END							\
{									\
    ftime(&tend);							\
    total += ((tend.time * 1000 + tend.millitm) -			\
	      (tstart.time * 1000 + tstart.millitm));			\
    count += 1;								\
}
#define TIME_REPORT \
  printf("%6f usec\n", 1000.0 * (float64)total / (float64)count);
#else
#define TIME_INIT
#define TIME_START
#define TIME_END
#define TIME_REPORT
#endif

static void make_object(int32 num)
{
    SNSTVE_OBJECT obj;
    SNSTVE_CLASS *which;

    switch(num)
    {
      case 0: which = &stationary_rectangle; break;
      case 1: which = &stationary_circle; break;
      case 2: which = &stationary_line; break;
      case 3: which = &dragable_rectangle; break;
      case 4: which = &dragable_circle; break;
      case 5: which = &dragable_line; break;
    }

    obj.class = which;
    obj.instance = count;
    obj.kind = KIND(which);
    switch(obj.kind)
    {
      case SNSTVE_RECTANGLE:
	obj.attr.rect.x = rnd_random() % width;
	obj.attr.rect.y = rnd_random() % height;
	obj.attr.rect.width = rnd_random() % 20 + 1;
	obj.attr.rect.height = rnd_random() % 20 + 1;
	XFillRectangle(display, pixmap,
		       obj.class->dragable ? colors.red : colors.blue,
		       obj.attr.rect.x, obj.attr.rect.y,
		       obj.attr.rect.width, obj.attr.rect.height);
	obj.attr.rect.width += 2;
	obj.attr.rect.height += 2;
	break;
      case SNSTVE_CIRCLE:
	obj.attr.circle.x = rnd_random() % width;
	obj.attr.circle.y = rnd_random() % height;
	obj.attr.circle.radius = rnd_random() % 10 + 1;
	XFillArc(display, pixmap,
		 obj.class->dragable ? colors.green : colors.yellow,
		 obj.attr.circle.x-obj.attr.circle.radius,
		 obj.attr.circle.y-obj.attr.circle.radius,
		 2*obj.attr.circle.radius, 2*obj.attr.circle.radius,
		 0, 360*64);
	obj.attr.circle.radius += 2;
	break;
      case SNSTVE_LINE:
	obj.attr.line.x0 = rnd_random() % width;
	obj.attr.line.y0 = rnd_random() % height;
	obj.attr.line.x1 = obj.attr.line.x0 + rnd_random() % 100 - 50;
	obj.attr.line.y1 = obj.attr.line.y0 + rnd_random() % 100 - 50;
	obj.attr.line.width = 10;
	XDrawLine(display, pixmap,
		  obj.class->dragable ? colors.orange : colors.violet,
		  obj.attr.line.x0, obj.attr.line.y0,
		  obj.attr.line.x1, obj.attr.line.y1);
	obj.attr.line.width += 2;
	break;
    }

    TIME_START;
    snstve_add_object(sensitive, &obj);
    TIME_END;

    /* Save the DRAGABLEs, so we can move them around later */
    if (which == &dragable_line)
      lines[count] = obj;
    else if (which == &dragable_rectangle)
      rectangles[count] = obj;
    else if (which == &dragable_circle)
      circles[count] = obj;

    count += 1;
}

static void remove_stuff(void)
{
    snstve_clear_class(sensitive, &stationary_rectangle);
    snstve_set_class(sensitive, &stationary_circle, 0, 0, 0);
}

static void regenerate(void)
{
    int32 i;

    TIME_INIT;

    /* Forget the old objects */
    count = 0;

    /* Clear the pixmap */
    XFillRectangle(display, pixmap, colors.black,
		   0, 0, width, height);

    /* Clear the sensitive window */
    snstve_clear(sensitive);

    /* Generate a bunch of stuff */
    for(i=0;i<OBJECT_COUNT;i++)
      make_object(rnd_random() % N_OBJECT_TYPES);

    /* Refresh */
    XClearWindow(display, window);
    snstve_refresh(sensitive);

    /* Do this again in 30 seconds */
    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel), 30000,
		    regenerate, 0);

    /* Remove stuff in 15 seconds */
    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel), 15000,
		    remove_stuff, 0);

    TIME_REPORT;
}

static void move_circle(SNSTVE_OBJECT *circle)
{
    int32 old_x = -1, old_y = -1;
    static int32 first_time = -2;

    /* Erase the old circle */
    if (first_time)
      first_time += 1;
    else
    {

	old_x = circle->attr.circle.x;
	old_y = circle->attr.circle.y;
	circle->attr.circle.radius -= 2;
	XFillArc(display, window, erase,
		 circle->attr.circle.x-circle->attr.circle.radius,
		 circle->attr.circle.y-circle->attr.circle.radius,
		 2*circle->attr.circle.radius, 2*circle->attr.circle.radius,
		 0, 360*64);
	circle->attr.circle.radius += 2;
    }

#define JUMP 5

    /* Draw the new circle */
    if (jump_set && circle->class->dragable)
    {
	jump_set = 0;
	circle->attr.circle.x += jump_dx;
	circle->attr.circle.y += jump_dy;
    }
    else
    {
	circle->attr.circle.x =
	  (width + (circle->attr.circle.x + rnd_random()%JUMP - JUMP/2)) %
	    width;
	circle->attr.circle.y =
	  (height + (circle->attr.circle.y + rnd_random()%JUMP - JUMP/2)) %
	    height;
    }
    circle->attr.circle.radius += rnd_random()%3 - 1;
    if (circle->attr.circle.radius < 3)
      circle->attr.circle.radius = 3;

    /* Tell libsensitive it moved before we move it, so erasing the
     * highlight does not chop a piece out of the circle
     */
    if (first_time || !snstve_change_object(sensitive, old_x, old_y, circle))
    {
	/* Add the object both the first time, and whenever there is
	 * a "can't find it" error (which should happen whenever
	 * regenerate() is called).
	 */

	snstve_add_object(sensitive, circle);
    }

    /* Move it */
    circle->attr.circle.radius -= 2;
    XFillArc(display, window,
	     circle->class->dragable ? colors.green : colors.yellow,
	     circle->attr.circle.x-circle->attr.circle.radius,
	     circle->attr.circle.y-circle->attr.circle.radius,
	     2*circle->attr.circle.radius, 2*circle->attr.circle.radius,
	     0, 360*64);
    circle->attr.circle.radius += 2;

    /* Do this again soon */
    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel), 100,
		    move_circle, circle);
}

main(int32 argc, char **argv)
{
    int32 hold_argc = argc;
    XWindowAttributes att;
    XSetWindowAttributes set_att;
    XGCValues gcv;
    extern float64 atof(const char *);

    rnd_srandom(0);

#define INIT(which)	\
    SNSTVE_INIT_CLASS(which);	\
    which.sensitive = 1;	\
    which.hot = HOT(&which);	\
    which.dragable = DRAGABLE(&which);	\
    which.drag_thresh = 10;	\
    which.buttons = BUTTONS(&which);

    INIT(stationary_rectangle);
    INIT(stationary_circle);
    INIT(stationary_line);
    INIT(dragable_rectangle);
    INIT(dragable_circle);
    INIT(dragable_line);

    callback_register_handler(stationary_rectangle.gesture, callback, NULL);
    callback_register_handler(stationary_circle.gesture, callback, NULL);
    callback_register_handler(stationary_line.gesture, callback, NULL);
    callback_register_handler(dragable_rectangle.gesture, callback, NULL);
    callback_register_handler(dragable_circle.gesture, callback, NULL);
    callback_register_handler(dragable_line.gesture, callback, NULL);

    toplevel = XtInitialize(argv[0], "Sensitive",
			    NULL, 0, &hold_argc, argv);

    XtRealizeWidget(toplevel);

    display = XtDisplay(toplevel);
    window = XtWindow(toplevel);

    set_att.bit_gravity = ForgetGravity;
    XChangeWindowAttributes(display, window, CWBitGravity, &set_att);

    make_colors();

    draw = colors.white;
    gcv.fill_style = FillTiled;
    erase = XCreateGC(display, window, GCFillStyle, &gcv);
    XSetLineAttributes(display, draw, 2, LineSolid, CapButt, JoinMiter);
    XSetLineAttributes(display, erase, 2, LineSolid, CapButt, JoinMiter);

    snstve_init();
    sensitive = snstve_create(toplevel, draw, erase);

    callback_register_handler(snstve_miss_callback(sensitive), callback,NULL);

    XGetWindowAttributes(display, window, &att);
    change_size(att.width, att.height);

    XtAddEventHandler(toplevel, StructureNotifyMask, FALSE, resize, 0);

    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel), 1000,
		    regenerate, 0);

    SNSTVE_INIT_CLASS(stationary_moving);
    callback_register_handler(stationary_moving.gesture, callback, NULL);
    callback_register_handler(stationary_moving.excited, callback, NULL);

    stationary_moving.sensitive = 1;
    stationary_moving.hot = 1;
    stationary_moving.dragable = 0;
    stationary_moving.drag_thresh = 10;
    stationary_moving.buttons = 0x1; /*Any*/

    dragable_moving = stationary_moving;
    dragable_moving.dragable = 1;

    s_circle.class = &stationary_moving;
    s_circle.instance = 0;
    s_circle.kind = SNSTVE_CIRCLE;
    s_circle.attr.circle.x = att.width/2;
    s_circle.attr.circle.y = att.height/2;
    s_circle.attr.circle.radius = 20;
    d_circle = s_circle;
    d_circle.class = &dragable_moving;
    d_circle.instance = 0;
    d_circle.attr.circle.x += 100;

    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel), 1000,
		    move_circle, &s_circle);
    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel), 1050,
		    move_circle, &d_circle);

    XtMainLoop();
}
