/*
 * $RCSfile: libtactmap.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libtactmap.h                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header for libtactmap                              *
*       Created: Wed Sep 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBTACTMAP_H_INCLUDED
#define _LIBTACTMAP_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdext.h>
#include <libsensitive.h>
#include <libreader.h>
#include <libctdb.h>
#include <libcoordinates.h>
#undef SCALE
#include <p_po.h> /* Needed for LineStyle */
#include "stdtypes.h" /*common/include/global*/
#ifdef USE_X
#undef Time
#undef Boolean
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#endif
#include <libenvironment.h>


#undef AE_FEATURES

/* Classes of tactical map objects.
 */
typedef enum
{
    /* Terrain features (do not instantiate these more than once
     * per tactical map).
     */

#define TACTMAP_FIRST_CLASS TACTMAP_HYPSO

    /* Hypsometric background */
    TACTMAP_HYPSO,

    /* Contaminant */
    TACTMAP_CONTAMINANT,

    /* Water */
    TACTMAP_WATER,

    /* Roads */
    TACTMAP_ROADS,

#ifdef AE_FEATURES
    /* Roads */
    TACTMAP_OFFROADS,
#endif

    /* Trees, tree lines, and tree canopies */
    TACTMAP_TREES,

    /* Buildings, power pilons, other structures */
    TACTMAP_BUILDINGS,

#ifdef AE_FEATURES
    /* Tactical signs */
    TACTMAP_TACT_SIGNS,
#endif

    /* Pipelines */
    TACTMAP_PIPELINES,

    /* Political Boundaries */
    TACTMAP_POLITICAL,

    /* Railroads */
    TACTMAP_RAILROADS,

    /* Powerlines */
    TACTMAP_POWERLINES,

    /* Contour lines */
    TACTMAP_CONTOURS,

    /* Towns */
    TACTMAP_TOWNS,

    /* GCS cell boundary lines */
    TACTMAP_CELLS,

    /* Craters */
    TACTMAP_CRATERS,

    /* Ditches */
    TACTMAP_DITCHES,

    /* Fighting Positions */
    TACTMAP_FPOS,

    /* Wire */
    TACTMAP_WIRE,

    /* Rock Drops */
    TACTMAP_ROCK_DROPS,

    /* Dragon's Teeth */
    TACTMAP_DRAGONS_TEETH,

    /* Berms */
    TACTMAP_BERMS,

    /* Rubble */
    TACTMAP_RUBBLE,

    /* Snow */
    TACTMAP_SNOW,

    /* Clouds */
    TACTMAP_CLOUDS,

    /* Miclic Line */
    TACTMAP_MICLIC_LINE,

    /* Miclic Scorch */
    TACTMAP_MICLIC_SCORCH,

    /* Camouflage */
    TACTMAP_CAMOUFLAGE,

    /* Hydrology Surface */
    TACTMAP_HYDRO_SURFACE,

    /* Hydrology Subsurface */
    TACTMAP_HYDRO_SUBSURFACE,

    /* Grid lines */
    TACTMAP_GRIDS,

#define TACTMAP_MAP_FEATURE(x) ((x) <= TACTMAP_GRIDS)

    /* Multiply instantiable application-defined objects
     */

#define TACTMAP_FIRST_DYNAMIC_CLASS TACTMAP_DRAWN
    /* Objects the application draws itself */
    TACTMAP_DRAWN,

    /* Pixmaps */
    TACTMAP_PIXMAP,

    /* Pictures */
    TACTMAP_PICTURE,


    /* Linear objects (routes, minefields, etc.) */
    TACTMAP_LINEAR,

    /* Text */
    TACTMAP_TEXT,

    /* BGR icon */
    TACTMAP_BGR,

    TACTMAP_LINE_INTERVIS,

    TACTMAP_AREA_INTERVIS,

    TACTMAP_LINE_ENV_EFFECT,
      
    TACTMAP_GLOBAL_EFFECT

#define TACTMAP_LAST_CLASS TACTMAP_GLOBAL_EFFECT
} TACTMAP_OBJECT_CLASS;

typedef enum
{
    TACTMAP_DITHER_HYPSO,
    TACTMAP_COLOR_HYPSO
} TACTMAP_HYPSO_KIND;

typedef enum
{
    TACTMAP_MILGRID,
    TACTMAP_GCS
} TACTMAP_GRID_KIND;

typedef enum
{
    TACTMAP_PICTURE_BLOCK,
    TACTMAP_PICTURE_LINE,
    TACTMAP_PICTURE_DISC,
    TACTMAP_PICTURE_CIRCLE,
    TACTMAP_PICTURE_LABEL
} TACTMAP_PICTURE_INSTRUCTION;

typedef enum
{
    TACTMAP_PICTURE_X      = 'x',
    TACTMAP_PICTURE_Y      = 'y',
    TACTMAP_PICTURE_METERS = 'm'
} TACTMAP_PICTURE_UNIT;

/* Maximum number of vertices in a non-libreader picture definition */
#define TACTMAP_MAX_PAIRS 8

typedef void (*TACTMAP_DRAW_CALLBACK) ( /*
					      Widget widget,
					      Pixmap drawable,
					      ADDRESS draw_arg,
					      XRectangle *boundary*/
					  );

typedef void (*TACTMAP_REFRESH_CALLBACK) (/*
						TACTMAP_PTR tactmap,
						int32 refreshing */
					     );

typedef struct tactmap_sensitive_options
{
    /* See libsensitive for the meaning of these attributes */

    SNSTVE_CLASS *sensClass;
    uint16 instance;		/* Ignored for terrain features */
    ADDRESS user_data;
} TACTMAP_SENSITIVE_OPTIONS;

/* The instance returned by libsensitive for terrain features can
 * be translated to a libctdb index using the following macro:
 */
#define TACTMAP_INSTANCE_TO_INDEX(instance) ((instance) >> 8)

/* The instance returned by libsensitive for user-defined objects
 * can be translated to the user-assigned instance number using the
 * following macro:
 */
#define TACTMAP_INSTANCE_TO_USER_INSTANCE(instance) ((instance) & 0xFFFF)

/* The following macros yield extra information about the instance
 * when the TACTMAP_CLASS is TACTMAP_LINE:
 */
#define TACTMAP_IS_VERTEX(instance) ((instance) & 0x10000)
#define TACTMAP_IS_SEGMENT(instance) (!TACTMAP_IS_VERTEX(instance))
#define TACTMAP_INDEX(instance) ((instance) >> 17)
typedef struct tactmap_block_data
{
    int32                num_xy_pairs; 
    float32              xy_pairs[TACTMAP_MAX_PAIRS];
} TMAP_BLOCK;
typedef struct tactmap_line_data
{
    float32              width;
    int32                num_xy_pairs; 
    float32              xy_pairs[TACTMAP_MAX_PAIRS];
} TMAP_LINE;
typedef struct tactmap_disc_data
{
    float32              center_x;
    float32              center_y;
    float32              radius;
    TACTMAP_PICTURE_UNIT unit;
} TMAP_DISC;
typedef struct tactmap_circle_data
{
    float32              width;
    float32              center_x;
    float32              center_y;
    float32              radius;
    TACTMAP_PICTURE_UNIT unit;
} TMAP_CIRCLE;

typedef struct tactmap_label_data
{
    float32              center_x;
    float32              center_y;
} TMAP_LABEL;

typedef struct tactmap_csystem
{
    /* Index of csystem which is this one's parent (use this
     * one's index to be parented by world).  Note that you
     * *may not* reference a parent with a higher index than
     * your own.
     */
    int32   parent_system;

    float64 angle;     /* 0 down Y axis, increase counter-clockwise */
    float64 dimensions[2]; /* meters */
    int32   cell ;
    float64 origin[2];     /* meters */
} TMAP_CSYSTEMS;

typedef union tactmap_object_dfn
{
    struct tactmap_hypso
    {
	TACTMAP_HYPSO_KIND kind;
    } hypso;
    struct tactmap_water
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } water;
    struct tactmap_contaminant
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } contaminant;
    struct tactmap_roads
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } roads;

#ifdef AE_FEATURES
    struct tactmap_offroads
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } offroads;
#endif
    struct tactmap_trees
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

#define TACTMAP_TREE_CANOPIES   0x1
#define TACTMAP_TREE_LINES      0x2
#define TACTMAP_TREE_INDIVIDUAL 0x4

	uint32 mask;
    } trees;
    struct tactmap_buildings
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } buildings;
#ifdef AE_FEATURES
    struct tactmap_tact_signs
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } tact_signs;
#endif
    struct tactmap_pipelines
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } pipelines;
    struct tactmap_political
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } political;
    struct tactmap_railroads
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } railroads;
    struct tactmap_powerlines
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } powerlines;
    struct tactmap_contours
    {
	/* A contour line is generated every minor_spacing meters in
	 * altitude.  Those which are (roughly) evenly divisible by
	 * major_spacing are emphasized.
	 */
	float64 minor_spacing;
	float64 major_spacing;
    } contours;
    struct tactmap_towns
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } towns;

    struct tactmap_craters
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } craters;

    struct tactmap_ditches
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } ditches;

    struct tactmap_fpos
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } fpos;

    struct tactmap_wire
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

#define TACTMAP_WIRE_LINES            0x1

	uint32 mask;
    } wire;

    struct tactmap_rock_drops
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } rock_drops;

    struct tactmap_dragons_teeth
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

#define TACTMAP_DRAGONS_TEETH_IND     0x1

	uint32 mask;
    } dragons_teeth;

    struct tactmap_berms
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } berms;

    struct tactmap_rubble
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } rubble;

    struct tactmap_snow
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } snow;

    struct tactmap_clouds
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } clouds;

    struct tactmap_miclic_line
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
#define TACTMAP_MICLIC_LINE_IND        0x1
	uint32 mask;
    } miclic_line;

    struct tactmap_miclic_scorch
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
#define TACTMAP_MICLIC_SCORCH_IND       0x1
	uint32 mask;
    } miclic_scorch;

    struct tactmap_camouflage
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } camouflage;

    struct tactmap_hydro_surface
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } hydro_surface;

    struct tactmap_hydro_subsurface
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
    } hydro_subsurface;

    struct tactmap_grids
    {
	/* This structure is being used by both TACTMAP_GRIDS and 
	 * TACTMAP_CELLS
	 */
	int32             interval  ;/* grid interval, generally mult of 10 */
	uint32            label_mask;/* use masks defined in libcoordinates
				      * i.e. COORD_LABEL_LEFT ...*/
	TACTMAP_GRID_KIND kind;
    } grids;

    struct tactmap_linear
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

	/* The line style is one defined by the persistent object
	 * protocol (common/include/protocol/p_po.h).
	 */
	LineStyle style;

	/* Set dashed to TRUE for non-solid line */
	uint16 dashed;

	/* Line thickness (pixels) */
	uint16 thickness;

	/* Arrows */
	uint32 forward_arrow;
	uint32 backward_arrow;

	/* Width of the line (meters) */
	float32 width;

	/* The GC to use for drawing (some attributes, such as line width,
	 * will be modified).
	 */
#ifdef USE_X
	GC gc;
#endif
	/* the cell which contains this set of points */
	int32 cell;
	
	/* n_vertices indicates how many vertices the line has */
	int32 n_vertices;

	/* pointer to x y x y... format list of vertices; the contents
	 * of this array are copied, so it is safe to point this to
	 * a local variable when calling tactmap_create_object().
	 * The size of this array should be (n_vertices*2*sizeof(float64)).
	 */
	float64 *vertices;
    } linear;
    struct tactmap_pixmap
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

	/* The location of the pixmap (cell coordinates) */
	int32   cell;
	float64 x, y;

	/* The size of the pixmap (screen coordinates) */
	uint32 width;
	uint32 height;

	/* The pixmap to copy to the screen (assumes origin 0,0) */
#ifdef USE_X
	Pixmap pixmap;
#endif
    } pixmap;
    struct tactmap_picture
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

#define TACTMAP_MAX_GCS      12
#define TACTMAP_MAX_CSYSTEMS 6

	/* Set dashed to TRUE for non-solid lines */
	uint32 dashed;

	/* Drawing instructions in libreader format (see texinfo docs
	 * for format)
	 */
	READER_UNION *dfn;

	/* If dfn is NULL, then pointer to drawing instructions in
	 * structure form.  If dfn is provided, this is ignored.
	 * (Note: instructions are copied.)
	 */
	int32 num_instructions;
	struct tactmap_picture_instruction
	{
	    int8 interior_gc_index;		/* -1 to omit */
	    int8 border_gc_index;		/* -1 to omit */
	    int8 coordinate_system_index;

	    TACTMAP_PICTURE_INSTRUCTION kind;
	    union
	    {
		/* Note: these are set up to resemble the libreader
		 * format for reuse.  Thus, no float64's allowed.
		 */
		
		TMAP_BLOCK		block;
		TMAP_LINE		line;
		TMAP_DISC		disc;
		TMAP_CIRCLE		circle;
		TMAP_LABEL		label;
	    } data;
	} *instructions;

	/* GCs */
	int32 num_gcs;
#ifdef USE_X
	GC gcs[TACTMAP_MAX_GCS];
#endif

	/* Magnification of the picture */
	float64 magnification;

	/* Optional (pass NULL if not needed) label string */
	char *string;

	/* Flag if the picture is highlighted (non-zero), and if
	 * so, index of the gc to use.
	 */
	int32 highlight;

	/* Coordinate systems */
	int32 num_csystems;
	TMAP_CSYSTEMS	csystems[TACTMAP_MAX_CSYSTEMS];
    } picture;

 struct tactmap_global_effect
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

#define TACTMAP_MAX_GCS      12
#define TACTMAP_MAX_CSYSTEMS 6

	/* Set dashed to TRUE for non-solid lines */
	uint32 dashed;

	/* Drawing instructions in libreader format (see texinfo docs
	 * for format)
	 */
	READER_UNION *dfn;

	/* If dfn is NULL, then pointer to drawing instructions in
	 * structure form.  If dfn is provided, this is ignored.
	 * (Note: instructions are copied.)
	 */
	int32 num_instructions;
	struct tactmap_global_effect_instruction
	{
	    int8 interior_gc_index;		/* -1 to omit */
	    int8 border_gc_index;		/* -1 to omit */
	    int8 coordinate_system_index;

	    TACTMAP_PICTURE_INSTRUCTION kind;
	    union
	    {
		/* Note: these are set up to resemble the libreader
		 * format for reuse.  Thus, no float64's allowed.
		 */
		TMAP_BLOCK		block;
		TMAP_LINE		line;
		TMAP_DISC		disc;
		TMAP_CIRCLE		circle;
		TMAP_LABEL		label;
	    } data;
	} *instructions;

	/* GCs */
	int32 num_gcs;
#ifdef USE_X
	GC gcs[TACTMAP_MAX_GCS];
#endif

	/* Magnification of the picture */
	float64 magnification;

	/* Optional (pass NULL if not needed) label string */
	char *string;

	/* Flag if the picture is highlighted (non-zero), and if
	 * so, index of the gc to use.
	 */
	int32 highlight;

	/* Coordinate systems */
	int32 num_csystems;
	TMAP_CSYSTEMS	csystems[TACTMAP_MAX_CSYSTEMS];
    } global_effect;
    struct tactmap_drawn
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

	/* Function to call to draw */
	TACTMAP_DRAW_CALLBACK draw;
	int32 erase_on_delete_only;
	/* Argument passed to the draw function */
	ADDRESS draw_arg;
    } drawn;
    struct tactmap_text
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;

#ifdef USE_X
	/* The GC to use for drawing. */
	GC gc;

	/* Optional: The XFontStruct for this GC.  Find it using
	 * XQueryFont(display, GContextFromGC(gc))
	 * See texinfo docs for details.
	 */
	XFontStruct *font_struct;
#endif

	/* Location in GCS coordinates */
	int32   cell;
	float64 x, y;

	/* Whether the text shold be see-through */
	uint32 see_through;

	/* Alignment relative to that location */
	TextAlignment alignment;

	/* Pixels away from the location */
	int16 horizontal_offset;
	int16 vertical_offset;

	/* NULL-terminated character string (data is copied) */
	char *string;

    } text;
    struct tactmap_bgr
    {
	TACTMAP_SENSITIVE_OPTIONS sensitive;
	
	/* GCS coordinates */
	int32   cell;
	float64 x, y;

	/* Rotation of icon */
	float64 rotation;

	/* If TRUE, override algorithmic construction of BGR icon
	 * with the following BGR strings.
	 */
	int32 override;

	/* BGR instructions in libreader format (see texinfo docs
	 * for format)
	 */
	READER_UNION *dfn;

	/* object type of this icon */
	ObjectType object_type;

	/* Approximate size (pixels in width and height) of the icon.
	 * LibBGR uses this to calculate an acutal size.
	 */
	int32    size;

	/* Thickness of lines in icon.  Note that the thicker the lines, the
	 * longer libBGR takes to draw something.  An odd thickness
	 * (typically 3) works best.
	 */
	int32    line_thickness;

	/* Whether to draw the icon dashed */
	uint32    dashed;

	/* Flag if the picture is highlighted */
	int32 highlight;

#define TACTMAP_MAX_CALL_SIGN 11
	/* Optional call sign for this unit.  Null terminated. */
	char     call_sign[TACTMAP_MAX_CALL_SIGN];

	/* GC for BGR drawing */
#ifdef USE_X
	GC gc;
#endif
    } bgr;
    struct tactmap_line_intervis
    {
	int32   cell;
	float64 point1[XY];
	float64 point2[XY];
	float64 viewer_agl;
	float64 target_agl;
	float64 target_width;
	float64 target_height;
	int32 n_veh;
	CTDB_VEHICLE_LOCATION *veh;
	int32 ignore1;
	int32 show_percentages;
	int32 grid_interval;
#define TACTMAP_INTERVIS_TERRAIN      0x01
#define TACTMAP_INTERVIS_FOG          0x02
#define TACTMAP_INTERVIS_OBSCURANTS   0x04
#define TACTMAP_INTERVIS_ILLUMINATION 0x08
#define TACTMAP_INTERVIS_SMOKE	      0x10
	uint32 environ_effects;
    } line_intervis;
    struct tactmap_area_intervis
    {
	float64   center_point[XYC];
	float64   radius;
	float64   viewer_agl;
	float64   target_agl;
	float64   target_height;
	float64   target_width;
	uint32    environ_effects;
    } area_intervis;

    struct tactmap_line_env_effect
    {
	int32   cell;
	/* Both points are in "cell's" frame of reference */
	float64 point1[XY];
	float64 point2[XY];
	float64 viewer_agl;
	int32 grid_interval;
	int32 environ_effect;
    } line_env_effect;
	
    /* Generic reference to a sensitive class */
    TACTMAP_SENSITIVE_OPTIONS sensitive;
} TACTMAP_OBJECT_DFN;

typedef struct tactmap *TACTMAP_PTR;
typedef struct tactmap_object *TACTMAP_OBJECT_PTR;

/*
 * Public functions
 */

#ifndef USE_X
#define tactmap_init
#define tactmap_create(a,b,c,d) 0
#define tactmap_set_scale
#define tactmap_set_center
#define tactmap_screen_to_map
#define tactmap_map_to_screen
#define tactmap_pix_to_offvec
#define tactmap_offvec_to_pix
#define tactmap_world_view
#define tactmap_get_scale
#define tactmap_create_object
#define tactmap_delete_object
#define tactmap_add_object
#define tactmap_add_transient_object
#define tactmap_remove_object
#define tactmap_update_object
#define tactmap_tick(a,b) 0
#define tactmap_refresh_non_terrain
#define tactmap_redrawing
#define tactmap_get_dimension(a,b,c) 
#define tactmap_get_center(a,b,c,d) 

#else

/* tactmap_init:
 *
 * Initializes libtactmap, causing it to read its data file
 * (tmap_bgr.rdr) from the specified directory.
 * The flags are as in reader_read.
 * The return value is zero if the read succeeds, or one of the
 * libreader return values: READER_READ_ERROR READER_FILE_NOT_FOUND
 *
 * Note that the libreader function reader_init, the
 * librdrconst function rdc_init, and the libotmatch function otm_init,
 * must be called before this function.
 *
 * Call this before any other libtactmap functions.
 */
extern void 
tactmap_init (void);

/* tactmap_create:
 *
 * Creates a tactical map.  The map starts with no features displayed.
 * The widget specifies the drawing area (libtactmap posts its own
 * resize, expose, etc. handlers).  The sensitive window should be
 * created with a call to snstve_create().  The refresh callback is
 * called at the beginning and end of a map refresh.
 */
extern TACTMAP_PTR            
tactmap_create (
		   Widget                   widget,
		   SNSTVE_WINDOW_PTR        sensitive,
		   GC                       sensitive_erase_gc,
		   TACTMAP_REFRESH_CALLBACK refresh
		   );

/* tactmap_get_color:
 *
 * Returns the pixel value of a color used when drawing objects of
 * the passed class.
 */
extern Pixel
/*tried sensClass AUSTIN 13 Aug 97*/
tactmap_get_color ( 
		      TACTMAP_PTR tactmap,
		      TACTMAP_OBJECT_CLASS sensClass
		      );

/* tactmap_set_scale:
 *
 * Sets the map scale to 1:scale.
 */
extern void 
tactmap_set_scale ( 
		      TACTMAP_PTR tactmap,
		      float64     scale
		      );

/* tactmap_set_center:
 *
 * Sets the map center.
 */
extern void 
tactmap_set_center (
		       TACTMAP_PTR tactmap,
		       int32       map_x,
		       int32       map_y,
		       int32       map_c
		       );

/* tactmap_screen_to_map:
 *
 * Converts a screen location to a map location.
 */
extern void 
tactmap_screen_to_map (
			  TACTMAP_PTR tactmap,
			  int32       screen_x, 
			  int32       screen_y,
			  float64    *map_x, 
			  float64    *map_y,
			  int32      *map_c
			  );

/* tactmap_map_to_screen:
 *
 * Converts a map location to a screen location.  Return value indicates
 * whether the point is visible.
 */
extern int32 
tactmap_map_to_screen (
			  TACTMAP_PTR tactmap,
			  float64     map_x, 
			  float64     map_y,
			  int32       map_c,
			  int32      *screen_x, 
			  int32      *screen_y
			  );

/* tactmap_world_view:
 *
 * Puts the tactical map in a mode where the grid lines for the entire
 * database are shown, and a box representing the viewport is laid on
 * top.
 */
extern void 
tactmap_world_view ( 
		       TACTMAP_PTR tactmap,
		       int32 world_view_on /* Pass TRUE to enable */
		       );

/* tactmap_get_scale:
 *
 * Determines the map scale which would encapsulate a box of the passed
 * size (in pixels).
 */
extern float64 
tactmap_get_scale ( 
		      TACTMAP_PTR tactmap,
		      int32 width,
		      int32 height
		      );

/* tactmap_create_object:
 *
 * Creates an object which can be placed on the tactical map.
 */
extern TACTMAP_OBJECT_PTR 
/*tried sensClass AUSTIN 13 Aug 97*/
tactmap_create_object ( 
			  TACTMAP_PTR tactmap,
			  TACTMAP_OBJECT_CLASS sensClass,
			  TACTMAP_OBJECT_DFN *dfn
			  );

/* tactmap_destroy_object:
 *
 * Destroys an object created with tactmap_create_object.
 */
extern void 
tactmap_destroy_object ( 
			   TACTMAP_PTR        tactmap,
			   TACTMAP_OBJECT_PTR obj  
			   );


/* tactmap_add_object:
 *
 * Adds an object (created with tactmap_create_object) to the map display
 * list.
 */
extern void 
tactmap_add_object ( 
		       TACTMAP_PTR        tactmap,
		       TACTMAP_OBJECT_PTR object
		       );

/* tactmap_remove_object:
 *
 * Removes an object from the map display list.
 */
extern void 
tactmap_remove_object ( 
			  TACTMAP_PTR        tactmap,
			  TACTMAP_OBJECT_PTR object
			  );

/* tactmap_update_object:
 *
 * Changes the attributes of an existing object.
 */
extern void 
tactmap_update_object ( 
			  TACTMAP_PTR        tactmap,
			  TACTMAP_OBJECT_PTR object,
			  TACTMAP_OBJECT_DFN *dfn
			  );

/* tactmap_add_transient_object:
 *
 * Adds an object (created with tactmap_create_object) to the transient
 * display list.  This must not be a terrain feature.
 */
extern void 
tactmap_add_transient_object ( 
				 TACTMAP_PTR        tactmap,
				 TACTMAP_OBJECT_PTR object
				 );

/* tactmap_tick:
 *
 * Updates the map image.  Call this function periodically to redraw
 * the map (if no redraw is needed, the function will just return).
 * max_time_slice specifies the amount of time which is acceptable to
 * spend redrawing this tick.  If drawing operations exceed this time
 * allotment, the refresh will not be completed until the next tick or
 * later.
 */
extern void 
tactmap_tick ( 
		 TACTMAP_PTR tactmap,
		 uint32      max_time_slice
		 );

/* tactmap_refresh:
 *
 * Forces the redrawing of the tactical map.  This may be called after dynamic
 * terrain objects are created, destroyed, or modified.
 */
extern void
tactmap_refresh(TACTMAP_PTR tactmap);

/* tactmap_refresh_non_terrain:
 *
 * Runs through all non-terrain objects and redraws them.  This may
 * be called after editing operations to correct accidental erasures.
 */
extern void 
tactmap_refresh_non_terrain ( TACTMAP_PTR tactmap  );

/* tactmap_redrawing:
 *
 * Returns TRUE if the map is in the process of refreshing itself.
 * Drawing on a map which is in this state will probably disappear on
 * the next tactmap_tick.
 */
extern int32 
tactmap_redrawing (  TACTMAP_PTR tactmap  );

/* tactmap_bundle_changes:
 *
 * Sets the do_refresh flag so that many changes can be made to the
 * tactmap before the tactmap refreshes itself.
 */
extern void 
tactmap_bundle_changes ( TACTMAP_PTR tactmap  );

extern void 
tactmap_set_em_band ( 
			TACTMAP_PTR tactmap,
			ENV_EM_BAND band  
			);

extern ENV_EM_BAND 
tactmap_get_em_band ( TACTMAP_PTR tactmap );

/* File: tmap_state.c
 * Routine: tactmap_pix_to_offvec
 *
 * This routine will convert a screen coordinate delta to a GCS map coordinate
 * delta which consists of an (x,y,cell_id).  This routine should be used 
 * inplace of the previous tactmap_pixel_offset which will be phased out as 
 * applications become GCS aware.  Note: GCS unaware callers will find the 
 * x,y to be identical to the old x,y and the cell field can be ignored
 */
extern void
tactmap_pix_to_offvec (
			  TACTMAP_PTR tactmap,
			  int32       screen_dx,
			  int32       screen_dy,
			  float64     offset_vector[XYC]
			  );
/* File: tmap_state.c
 * Routine: tactmap_offvec_to_pix
 *
 * This routine will convert a GCS map coordinate delta to a screen coordinate 
 * delta.
 */
extern void
tactmap_offvec_to_pix (
			  TACTMAP_PTR tactmap,
			  float64     offset_vector[XYC],
			  int32      *screen_dx,
			  int32      *screen_dy
			  );
/* File: tmap_state.c
 * Routine: tactmap_get_dimension
 *
 * This routine will return the width and height (in meters) of the tactmap.
 */
extern void
tactmap_get_dimension (
			  TACTMAP_PTR tactmap,
			  float64    *width,
			  float64    *height
			  );

/* File: tmap_state.c
 * Routine: tactmap_get_center
 *
 * This routine will return the tactmap center in GCS coordinates (x,y,cell).
 */
extern void
tactmap_get_center (
		       TACTMAP_PTR tactmap,
		       float64    *map_x,
		       float64    *map_y,
		       int32      *map_c
		       );

extern int32 tactmap_obj_display_on (TACTMAP_OBJECT_PTR obj);

#endif /*USE_X*/

#ifdef __cplusplus
}
#endif

#endif /*_LIBTACTMAP_H_INCLUDED*/
