/*
 * $RCSfile: libtmap_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libtmap_local.h                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header for libtactmap                             *
*       Created: Wed Sep 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libtactmap.h"
#include <libcoordinates.h>
#include <libgcs.h>
#include <libctdb.h>
#include <libreader.h>
#include <libbgrdb.h>
#include <libdither.h>
#include <X11/Intrinsic.h>
#ifndef _LIBTMAP_LOCAL_H_INCLUDED
#define _LIBTMAP_LOCAL_H_INCLUDED
#include <libenvui.h>

/* Definitions for intervisibility */
#define PERCENT_SIZE  4
#define N_PERCENT     DITHER_COUNT(PERCENT_SIZE)
#define SAMPLING_PIXELS 10
#define MAX_SAMPLES (2000/SAMPLING_PIXELS)
#define TICK_LENGTH 5

/* Bigger size yields more shades */
#define HYPSO_CMAP_SIZE 32

#define TACTMAP_TREE_WIDTH         3 /*meters*/
#define TACTMAP_MIN_WIDE_TREE_PPM  0.167

/* The minimum scale at which the thinned version of the database will be
 * used for drawing.
 */
#define TACTMAP_THINNED_SCALE    50000

/* The portion of the screen we fill in world-view mode */
#define WORLD_VIEW_FILL 0.9

/* We do this a lot... */
#define DECLARE_REGS						\
    register float64 ppm = tactmap->cur->ppm;		\
    register int32 cc    = tactmap->cur->cc;		\
    register int32 cx    = tactmap->cur->cx;			\
    register int32 cy    = tactmap->cur->cy;			\
    register int32 hsw   = tactmap->cur->hsw;		\
    register int32 hsh   = tactmap->cur->hsh

/* We do this a lot also... */
#define DPYWIN XtDisplay(tactmap->widget), XtWindow(tactmap->widget)

/* This macro will transform a coordinate that is in the incorrect 
 * context (different from tactmap->cur->cc) to the context of
 * tactmap->cur->cc
 */
#define CHANGE_CONTEXT(ocell,ox,oy,ncell)                                    \
	{                                                                    \
	    float64 nx,ny;                                                   \
	    gcs_gcs2e_to_gcscs((int32)(ocell),(float64)(ox),(float64)(oy), \
                                 (int32)(ncell),&nx,&ny);                    \
	    (ocell) = (ncell);                                               \
	    (ox)    = nx;                                                    \
	    (oy)    = ny;                                                    \
	}
/* The user gives us 16 bits of instance data for sensitive objects.
 * We tack another 16 bits on top for lines.  Bit 16 indicates whether
 * the sensitive object is a vertext or a segment; bits 17-31 indicate
 * which vertext or segment.
 */
#define INDEX_SHIFT 17
#define IS_POINT    0x10000
typedef struct tactmap_view_pt
{
    int32   cx;
    int32   cy;
    int32   cc;
    float64 ppm;
    float64 hsw;
    float64 hsh;
} TACTMAP_VIEW_PT;

typedef struct tactmap_bridge_data
{
    float64    loc[XYC] ;
    float64    del[XY]  ;
    float64    width    ;
    int32      mes      ;
} TACTMAP_BRIDGE_DATA;

typedef struct tactmap_bounds
{
    /* Boundaries of require update */
    int32 min_x, max_x;
    int32 min_y, max_y;
} TACTMAP_BOUNDS;


struct tactmap
{
    Widget widget;
    SNSTVE_WINDOW_PTR sensitive;
    TACTMAP_REFRESH_CALLBACK refresh_callback;
    TACTMAP_VIEW_PT wv;/* world view point */
    TACTMAP_VIEW_PT vp;/* view port - standard viewing */
    TACTMAP_VIEW_PT *cur;/* current active viewport */

    CTDB *ctdb;

    /* Flag for alternate display mode */
    int32  world_view_mode;

    int32 *pregenerated_grids;
    XRectangle last_viewport;

    /* Flag indicating a refresh is needed */
    int32 do_refresh;

    /* Flag we set when we start a refresh, and clear when we finish */
    int32 redrawing;

    /* Hourglass cursor */
    Cursor hourglass;

    /* Holds the feature are we due to draw next */
    TACTMAP_OBJECT_CLASS current_class;
    TACTMAP_OBJECT_PTR objects[TACTMAP_LAST_CLASS+1];

    /* Bounds of the playbox in the center_cells frame of reference */
    TACTMAP_BOUNDS playbox;
    
    /* These are in pixels */
    int32 pix_width;
    int32 pix_height;
    int32 depth;

    /* Terrain pixmap */
    Pixmap terrain_pix;

    /* Intervisibility pixmap */
    Pixmap intervis_pix[N_PERCENT];

    /* Clouds pixmaps */
    Pixmap *clouds_pix;

    /* Whole-map pixmap */
    Pixmap background_pix;

    /* Erase GCs */
    GC terrain_gc;
    GC background_gc;

    /* GCs for drawing */
    struct
    {
	GC background;
	GC hypso;
	GC water;
	GC fordable;
	GC no_go;
	GC slow_go;
	GC roads;
        GC contamination;
#ifdef AE_FEATURES
	GC offroads;
#endif
	GC trees;
	GC canopies;
	GC buildings;
        GC contaminant;
#ifdef AE_FEATURES
	GC tact_signs;
#endif
	GC pipelines;
	GC political;
	GC railroads;
	GC powerlines;
	GC contours[2];
	GC towns;
	GC grids;
  	GC compass;
	GC line_intervis;
	GC area_intervis;
	GC bridge_picture;
	GC bridge_symbol;
	GC cells;
	GC craters;
	GC ditches;
	GC fpos;
	GC wire;
	GC rock_drops;
	GC dragons_teeth;
	GC berms;
	GC rubble;
	GC snow;
	GC clouds;
	GC miclic_line;
	GC miclic_scorch;
	GC camouflage;
	GC hydro_surface;
	GC hydro_subsurface;
	GC plug;
    } gc;

    /* Answers to color queries */
    Pixel colors[TACTMAP_LAST_CLASS+1];
    Pixel bg_color;

    /* Colors for various altitudes */
    int32 n_hypso_colors;
    struct
    {
	int32 begin_alt;
	XColor dark_color;
	XColor bright_color;
    } hypso_colors[16];
    Pixel hypso_pixels[HYPSO_CMAP_SIZE];

    /* Environment-specific preferences */
    EUI_PREFS_PTR eui_prefs;
};

struct tactmap_object
{
    /* An object can be on the regular display list, the transient
     * display list, or no list at all.
     */
    enum tactmap_display_mode
    {
	NOT_DISPLAYED,
	PERMANENT,
	TRANSIENT
    } display_mode;

    TACTMAP_OBJECT_CLASS sensClass;
    TACTMAP_OBJECT_DFN dfn;
    void (*draw_fcn) (
#if 0
		      TACTMAP_PTR *tactmap,
		      TACTMAP_OBJECT_PTR object,
		      int32 use_erase_gc,
		      TACTMAP_BOUNDS *bounds,
		      CTDB_SEARCH_SPACE_PTR space,
		      /* -- ignored by non-terrain */
		      int32 clear_area_when_finished,
		      /* -- ignored by terrain features */
		      TACTMAP_OBJECT_DFN *new_definition
		      /* -- ignored by terrain features */
#endif
		      );

    /* We store the boundaries of a drawn, picture, and text objects here */
    XRectangle drawn_bounds;

    /* For BGR objects, we store here the current Pixmap in use (if any),
     * and the data that were used to generate that pixmap.
     */
    struct
    {
	BGRDB_PIXMAP_POOL_ELEMENT *pooled_pixmap;
	int32                      width;
	int32                      height;
	int32                      center_x;/* screen x */
	int32                      center_y;/* screen y */
	ObjectType                 object_type;
	int32                      override;
	READER_UNION              *bgr_strings;
	float64                    rotation;
	int32                      size;
	int32                      line_thickness;
	GC                         gc;
	char                       call_sign[TACTMAP_MAX_CALL_SIGN];
    } bgr_draw_cache;

    /* Other objects of the same class */
    struct tactmap_object *next;
    struct tactmap_object *prev;
};
/*
 * Coordinate system transformation macros
 */

#define METERS(pixel,ppm) \
  ((int32)(0.5 + (pixel) / (ppm)))
#define PIXELS(meter,ppm) \
  ((float64)((meter) * (ppm)))

#define X_P_N(x,cx,hsw,ppm) \
  (PIXELS(- (cx) + (hsw) + (x), ppm))
#define Y_P_N(y,cy,hsh,ppm) \
  (PIXELS((cy) + (hsh) - (y), ppm))
#define X_METER_NUMBER(x,cx,hsw,ppm) \
  ((cx) - (hsw) + METERS(x, ppm))
#define Y_METER_NUMBER(y,cy,hwh,ppm) \
  ((cy) + (hsh) - METERS(y, ppm))

#define X_PIXEL_ON_SCREEN(x,cx,hsw,ppm,buf) \
  ((x) < ((cx)-(hsw)) ? 0 : \
   ((x) > ((cx)+(hsw)-(buf)/(ppm)) ? \
    tactmap_X_pixel_number((int32)((cx)+(hsw)-(buf)/(ppm)),(cx),(hsw),(ppm)) : \
    tactmap_X_pixel_number((int32)(x),(cx),(hsw),(ppm))))

#define Y_PIXEL_ON_SCREEN(y,cy,hsh,ppm,buf) \
  ((y)+(buf)/(ppm) > ((cy)+(hsh)) ? (buf) : \
   ((y) < ((cy)-(hsh)) ? \
    tactmap_Y_pixel_number((int32)(cy)-(hsh),(cy),(hsh),(ppm)) : \
    tactmap_Y_pixel_number((int32)(y),(cy),(hsh),(ppm))))

/* Former macros, now functions to deal with overflow */
extern int16 
tactmap_X_pixel_number (
			   int32   x,
			   int32   cx,
			   int32   hsw,
			   float64 ppm
			   );

extern int16 
tactmap_Y_pixel_number (
			   int32   y,
			   int32   cy,
			   int32   hsh,
			   float64 ppm
			   );
/* Same as above except x/y are float */
extern int16 
tactmap_FX_pixel_number (
			    float64 x,
			    int32   cx,
			    int32   hsw,
			    float64 ppm
			    );
extern int16
tactmap_FY_pixel_number (
			    float64 y,
			    int32   cy,
			    int32   hsh,
			    float64 ppm
			    );

#define TOO_SMALL(data,limit)                   \
  (((data)->x_max - (data)->x_min < (limit)) && \
   ((data)->y_max - (data)->y_min < (limit)))

#define POINT_IN_RECT(px, py, rect) \
    (((px) >= (rect).x) && ((px) <= (rect).x+(int32)(rect).width) && \
     ((py) >= (rect).y) && ((py) <= (rect).y+(int32)(rect).height))

#define LINE_MAY_CROSS_RECT(x0, y0, x1, y1, rect)     \
    (!((NS_MAX((x0),(x1)) < (rect).x) ||                 \
       (NS_MIN((x0),(x1)) > (rect).x+(int32)(rect).width) ||    \
       (NS_MAX((y0),(y1)) < (rect).y) ||                 \
       (NS_MIN((y0),(y1)) > (rect).y+(int32)(rect).height)))

/* Private prototypes */

extern void 
tactmap_draw_pregenerated_grids (    
				    TACTMAP_PTR tactmap,
				    Pixmap      pixmap
				    );
extern void 
tactmap_show_viewport (    
			  TACTMAP_PTR tactmap,
			  Drawable    drawable
			  );

extern void 
tactmap_erase_viewport ( TACTMAP_PTR tactmap );

extern void 
tactmap_draw_class (    
		       TACTMAP_PTR               tactmap,
		       TACTMAP_OBJECT_CLASS      sensClass,
		       enum tactmap_display_mode mode,
		       TACTMAP_BOUNDS           *bounds,
		       CTDB_SEARCH_SPACE_PTR     space 
		       );

extern void                  
tactmap_draw_buildings ( 			 
			   TACTMAP_PTR            tactmap,
			   TACTMAP_OBJECT_PTR     object,
			   int32                  use_erase_gc,
			   TACTMAP_BOUNDS        *bounds,
			   CTDB_SEARCH_SPACE_PTR  space
			   );
extern void 
tactmap_draw_contaminant( 			 
			   TACTMAP_PTR            tactmap,
			   TACTMAP_OBJECT_PTR     object,
			   int32                  use_erase_gc,
			   TACTMAP_BOUNDS        *bounds,
			   CTDB_SEARCH_SPACE_PTR  space
			  );
extern void                  
tactmap_draw_contours ( 	    
			  TACTMAP_PTR             tactmap,
			  TACTMAP_OBJECT_PTR      object,
			  int32                   use_erase_gc,
			  TACTMAP_BOUNDS         *bounds
			  );
extern void                  
tactmap_draw_grids ( 	       
		       TACTMAP_PTR         tactmap,
		       TACTMAP_OBJECT_PTR  object,
		       int32               use_erase_gc,
		       TACTMAP_BOUNDS     *bounds
		       );
extern void                  
tactmap_draw_hypso (         
		       TACTMAP_PTR         tactmap,
		       TACTMAP_OBJECT_PTR  object,
		       int32               use_erase_gc,
		       TACTMAP_BOUNDS     *bounds
		       );
extern void                  
tactmap_draw_linear ( 	      
		     TACTMAP_PTR               tactmap,
		     TACTMAP_OBJECT_PTR        object,
		     int32                     use_erase_gc,
		     TACTMAP_BOUNDS           *bounds,
		     CTDB_SEARCH_SPACE_PTR     space,
		     int32                     clear_area_when_finished,
		     TACTMAP_OBJECT_DFN       *dfn
		     );
extern void                  
tactmap_draw_railroads (
			TACTMAP_PTR            tactmap,
			TACTMAP_OBJECT_PTR     object,
			int32                  use_erase_gc,
			TACTMAP_BOUNDS        *bounds,
			CTDB_SEARCH_SPACE_PTR  space 
			);
extern void                  
tactmap_draw_political ( 	   
			TACTMAP_PTR            tactmap,
			TACTMAP_OBJECT_PTR     object,
			int32                  use_erase_gc,
			TACTMAP_BOUNDS        *bounds,
			CTDB_SEARCH_SPACE_PTR  space
			);
extern void                  
tactmap_draw_pipelines ( 	   
			TACTMAP_PTR            tactmap,
			TACTMAP_OBJECT_PTR     object,
			int32                  use_erase_gc,
			TACTMAP_BOUNDS        *bounds,
			CTDB_SEARCH_SPACE_PTR space
			);
extern void                  
tactmap_draw_powerlines (	   
			 TACTMAP_PTR           tactmap,
			 TACTMAP_OBJECT_PTR    object,
			 int32                 use_erase_gc,
			 TACTMAP_BOUNDS       *bounds,
			 CTDB_SEARCH_SPACE_PTR space
			 );
#ifdef AE_FEATURES
extern void                  
tactmap_draw_offroads ( 	    
		       TACTMAP_PTR             tactmap,
		       TACTMAP_OBJECT_PTR      object,
		       int32                   use_erase_gc,
		       TACTMAP_BOUNDS         *bounds,
		       CTDB_SEARCH_SPACE_PTR   space
		       );
extern void                  
tactmap_draw_tact_signs (
			 TACTMAP_PTR           tactmap,
			 TACTMAP_OBJECT_PTR    object,
			 int32                 use_erase_gc,
			 TACTMAP_BOUNDS       *bounds,
			 CTDB_SEARCH_SPACE_PTR space
			 );
#endif
extern void                  
tactmap_draw_line_effect (	  
			  TACTMAP_PTR          tactmap,
			  TACTMAP_OBJECT_PTR   object,
			  int32                use_erase_gc,
			  TACTMAP_BOUNDS      *bounds,
			  CTDB_SEARCH_SPACE_PTR space,
			  int32                 clear_area_when_finished,
			  TACTMAP_OBJECT_DFN   *dfn
			  );
extern void                      
tactmap_draw_trees (		           
		    TACTMAP_PTR                tactmap,
		    TACTMAP_OBJECT_PTR         object,
		    int32                      use_erase_gc,
		    TACTMAP_BOUNDS            *bounds,
		    CTDB_SEARCH_SPACE_PTR      space
		    );
extern void                      
tactmap_draw_water (             
		    TACTMAP_PTR                tactmap,
		    TACTMAP_OBJECT_PTR         object,
		    int32                      use_erase_gc,
		    TACTMAP_BOUNDS            *bounds,
		    CTDB_SEARCH_SPACE_PTR      space
		    );
extern void                      
tactmap_draw_towns (
		    TACTMAP_PTR                tactmap,
		    TACTMAP_OBJECT_PTR         object,
		    int32                      use_erase_gc,
		    TACTMAP_BOUNDS            *bounds,
		    CTDB_SEARCH_SPACE_PTR      space  
		    );
extern void                      
tactmap_draw_pixmap (	           
		     TACTMAP_PTR               tactmap,
		     TACTMAP_OBJECT_PTR        object,
		     int32                     use_erase_gc,
		     TACTMAP_BOUNDS           *bounds,
		     CTDB_SEARCH_SPACE_PTR     space,
		     int32                     clear_area_when_finished,
		     TACTMAP_OBJECT_DFN       *dfn
		     );
extern void                      
tactmap_draw_picture (	          
		      TACTMAP_PTR              tactmap,
		      TACTMAP_OBJECT_PTR       object,
		      int32                    use_erase_gc,
		      TACTMAP_BOUNDS          *bounds,
		      CTDB_SEARCH_SPACE_PTR    space,
		      int32                    clear_area_when_finished,
		      TACTMAP_OBJECT_DFN      *dfn);
extern void                      
tactmap_draw_global (	          
		      TACTMAP_PTR              tactmap,
		      TACTMAP_OBJECT_PTR       object,
		      int32                    use_erase_gc,
		      TACTMAP_BOUNDS          *bounds,
		      CTDB_SEARCH_SPACE_PTR    space,
		      int32                    clear_area_when_finished,
		      TACTMAP_OBJECT_DFN      *dfn);
extern void                      
tactmap_draw_bgr (	              
		  TACTMAP_PTR                  tactmap,
		  TACTMAP_OBJECT_PTR           object,
		  int32                        use_erase_gc,
		  TACTMAP_BOUNDS              *bounds,
		  CTDB_SEARCH_SPACE_PTR        space,
		  int32                        clear_area_when_finished,
		  TACTMAP_OBJECT_DFN          *dfn
		  );
extern void                      
tactmap_draw_drawn (		           
		    TACTMAP_PTR                tactmap,
		    TACTMAP_OBJECT_PTR         object,
		    int32                      use_erase_gc,
		    TACTMAP_BOUNDS            *bounds,
		    CTDB_SEARCH_SPACE_PTR      space,
		    int32                      clear_area_when_finished,
		    TACTMAP_OBJECT_DFN        *dfn
		    );
extern void                      
tactmap_draw_text (
		   TACTMAP_PTR                 tactmap,
		   TACTMAP_OBJECT_PTR          object,
		   int32                       use_erase_gc,
		   TACTMAP_BOUNDS             *bounds,
		   CTDB_SEARCH_SPACE_PTR       space,
		   int32                       clear_area_when_finished,
		   TACTMAP_OBJECT_DFN         *dfn
		   );
extern void                      
tactmap_draw_area_iv (
		      TACTMAP_PTR              tactmap,
		      TACTMAP_OBJECT_PTR       object,
		      int32                    use_erase_gc,
		      TACTMAP_BOUNDS          *bounds,
		      CTDB_SEARCH_SPACE_PTR    space,
		      int32                    clear_area_when_finished,
		      TACTMAP_OBJECT_DFN      *dfn
		      );
extern void                      
tactmap_draw_line_iv (
		      TACTMAP_PTR              tactmap,
		      TACTMAP_OBJECT_PTR       object,
		      int32                    use_erase_gc,
		      TACTMAP_BOUNDS          *bounds,
		      CTDB_SEARCH_SPACE_PTR    space,
		      int32                    clear_area_when_finished,
		      TACTMAP_OBJECT_DFN      *dfn
		      );
extern void
tactmap_draw_cells (
		       TACTMAP_PTR             tactmap,
		       TACTMAP_OBJECT_PTR      object,
		       int32                   use_erase_gc,
		       TACTMAP_BOUNDS         *bounds
		       );
extern void
tactmap_draw_craters(
		     TACTMAP_PTR tactmap,
		     TACTMAP_OBJECT_PTR object, 
		     int32 use_erase_gc,
		     TACTMAP_BOUNDS *bounds,
		     CTDB_SEARCH_SPACE_PTR space
		     );
extern void
tactmap_draw_ditches(
		     TACTMAP_PTR tactmap,
		     TACTMAP_OBJECT_PTR object, 
		     int32 use_erase_gc,
		     TACTMAP_BOUNDS *bounds,
		     CTDB_SEARCH_SPACE_PTR space
		     );
extern void
tactmap_draw_fpos(
		  TACTMAP_PTR tactmap,
		  TACTMAP_OBJECT_PTR object, 
		  int32 use_erase_gc,
		  TACTMAP_BOUNDS *bounds,
		  CTDB_SEARCH_SPACE_PTR space
		  );
extern void
tactmap_draw_hydro_surface(
			   TACTMAP_PTR tactmap,
			   TACTMAP_OBJECT_PTR object, 
			   int32 use_erase_gc,
			   TACTMAP_BOUNDS *bounds,
			   CTDB_SEARCH_SPACE_PTR space
			   );
extern void
tactmap_draw_hydro_subsurface(
			      TACTMAP_PTR tactmap,
			      TACTMAP_OBJECT_PTR object, 
			      int32 use_erase_gc,
			      TACTMAP_BOUNDS *bounds,
			      CTDB_SEARCH_SPACE_PTR space
			      );

extern void
tactmap_draw_wire(
		  TACTMAP_PTR tactmap,
		  TACTMAP_OBJECT_PTR object, 
		  int32 use_erase_gc,
		  TACTMAP_BOUNDS *bounds,
		  CTDB_SEARCH_SPACE_PTR space
		  );

extern void                  
tactmap_draw_rock_drops( 			 
			TACTMAP_PTR            tactmap,
			TACTMAP_OBJECT_PTR     object,
			int32                  use_erase_gc,
			TACTMAP_BOUNDS        *bounds,
			CTDB_SEARCH_SPACE_PTR  space
			);

extern void
tactmap_draw_dragons_teeth(
			   TACTMAP_PTR tactmap,
			   TACTMAP_OBJECT_PTR object, 
			   int32 use_erase_gc,
			   TACTMAP_BOUNDS *bounds,
			   CTDB_SEARCH_SPACE_PTR space
			   );

extern void
tactmap_draw_berms(
		   TACTMAP_PTR tactmap,
		   TACTMAP_OBJECT_PTR object, 
		   int32 use_erase_gc,
		   TACTMAP_BOUNDS *bounds,
		   CTDB_SEARCH_SPACE_PTR space
		   );

extern void
tactmap_draw_rubble(
		    TACTMAP_PTR tactmap,
		    TACTMAP_OBJECT_PTR object, 
		    int32 use_erase_gc,
		    TACTMAP_BOUNDS *bounds,
		    CTDB_SEARCH_SPACE_PTR space
		    );

extern void
tactmap_draw_snow(
		  TACTMAP_PTR tactmap,
		  TACTMAP_OBJECT_PTR object, 
		  int32 use_erase_gc,
		  TACTMAP_BOUNDS *bounds,
		  CTDB_SEARCH_SPACE_PTR space
		  );

extern void tactmap_init_clouds_pixmaps(TACTMAP_PTR tactmap);

extern void
tactmap_draw_clouds(
		    TACTMAP_PTR tactmap, 
		    TACTMAP_OBJECT_PTR object, 
		    int32 use_erase_gc, 
		    TACTMAP_BOUNDS *bounds, 
		    CTDB_SEARCH_SPACE_PTR space, 
		    int32 clear_area_when_finished, 
		    TACTMAP_OBJECT_DFN *dfn);

extern void
tactmap_draw_miclic_line(
		  TACTMAP_PTR tactmap,
		  TACTMAP_OBJECT_PTR object, 
		  int32 use_erase_gc,
		  TACTMAP_BOUNDS *bounds,
		  CTDB_SEARCH_SPACE_PTR space
		  );

extern void
tactmap_draw_miclic_scorch(
		  TACTMAP_PTR tactmap,
		  TACTMAP_OBJECT_PTR object, 
		  int32 use_erase_gc,
		  TACTMAP_BOUNDS *bounds,
		  CTDB_SEARCH_SPACE_PTR space
		  );

extern void
tactmap_draw_camouflage(
			TACTMAP_PTR tactmap,
			TACTMAP_OBJECT_PTR object, 
			int32 use_erase_gc,
			TACTMAP_BOUNDS *bounds,
			CTDB_SEARCH_SPACE_PTR space
			);

extern void
tactmap_draw_plugs(
			TACTMAP_PTR tactmap,
			TACTMAP_OBJECT_PTR object, 
			int32 use_erase_gc,
			TACTMAP_BOUNDS *bounds,
			CTDB_SEARCH_SPACE_PTR space
			);


extern TACTMAP_PTR 
tactmap_find_sensitive ( SNSTVE_WINDOW_PTR sensitive );

extern void 
tactmap_resize (    
		   Widget      w,
		   TACTMAP_PTR tactmap,
		   XEvent     *event 
		   );

extern void 
tactmap_scale_string (
			 TACTMAP_PTR tactmap,
			 char        scale[] 
			 );

extern int32 
tactmap_off_screen32 (    
			 int32           n_vert,
			 float32        *verts,
			 TACTMAP_BOUNDS *bounds 
			 );

extern int32 
tactmap_off_screen64 (int32           n_vert,
		      float64         verts[][XY],
		      TACTMAP_BOUNDS *bounds 
		      );

extern void 
tactmap_draw_bridge(TACTMAP_PTR          tactmap, 
		    TACTMAP_BRIDGE_DATA *bridges);

		    
#endif /*_LIBTMAP_LOCAL_H_INCLUDED*/
