#ifndef lint
static char rcsid [] = "$RCSfile: tmap_abstract.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_abstract.c                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw abstracts in the ground (i.e. craters,       *
*                 ditches, fighting positions, berms, rubble, camouflage,   *
*                 snow, etc.)                                               *
*       Created: Tue Feb 13 1996                                            *
*       Author: mlongtin                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
/*changed class to sensClass AUSTIN 13 Aug 97*/
#ifdef USE_X

#include "libtmap_local.h"
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/*ARGSUSED*/
static void tactmap_draw_abstract(TACTMAP_PTR tactmap,
				  TACTMAP_OBJECT_PTR object, 
				  int32 use_erase_gc,
				  TACTMAP_BOUNDS *bounds,
				  CTDB_SEARCH_SPACE_PTR space,
				  GC abstract_gc,
				  int32 type_code)
{
    DECLARE_REGS;
    XPoint segs[5000];
    float32 *verts;
    float32 *top, *bottom;
    float32 bottom_diff[2];
    float32 offset1, offset2;
    float32 manhatt;
    int32 l1;
    int32 n_verts;
    int32 num_vrts;
    CTDB_ABSTRACT_DATA *data;
    XGCValues gcv;
    GC gc;
    int32 j;
    int32 xmin,xmax,ymin,ymax;
    SNSTVE_OBJECT snsobj;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;

    gcv.line_width = 0;
    XChangeGC(XtDisplay(tactmap->widget), abstract_gc,
	      GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.background,
	      GCLineWidth, &gcv);

    /* Now draw the abstract. */
    while (n_verts = ctdb_next_abstract(space, type_code, &verts, &data, NULL))
    {
	if (tactmap_off_screen32(n_verts, verts, bounds))
	  continue;

	/* Don't draw a CTDB_ABSTRACT_HYDRO_SURFACE feature if it is
	 * not wet.
	 */
	if(type_code == CTDB_ABSTRACT_HYDRO_SURFACE &&
	   !data->hydro_surface.wetness)
	  continue;

	/* Don't draw a CTDB_ABSTRACT_HYDRO_SUBSURFACE feature if it is
	 * not saturated.
	 */
	if(type_code == CTDB_ABSTRACT_HYDRO_SUBSURFACE &&
	   !data->hydro_subsurface.wetness)
	  continue;

	/* Bundle up the points */
	xmin=ymin = 0x10000;
	xmax=ymax = 0;
	for (j=0;j<n_verts;j++)
	{
	    segs[j].x =
	      tactmap_FX_pixel_number(verts[2*j],cx,hsw,ppm);
	    segs[j].y =
	      tactmap_FY_pixel_number(verts[2*j+1],cy,hsh,ppm);

	    if (segs[j].x < xmin)
	      xmin = segs[j].x;
	    if (segs[j].y < ymin)
	      ymin = segs[j].y;
	    if (segs[j].x > xmax)
	      xmax = segs[j].x;
	    if (segs[j].y > ymax)
	      ymax = segs[j].y;
	}

	if (use_erase_gc)
	  gc = tactmap->gc.background;
	else
	  gc = abstract_gc;

	XFillPolygon(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		     gc, segs, j, Nonconvex, CoordModeOrigin);

	snsobj.kind = SNSTVE_RECTANGLE;
	snsobj.instance = 0;
	snsobj.attr.rect.x = xmin;
	snsobj.attr.rect.y = ymin;
	snsobj.attr.rect.width = xmax-xmin;
	snsobj.attr.rect.height = ymax-ymin;

	snstve_add_object(tactmap->sensitive, &snsobj);

	if ((!(type_code    == CTDB_ABSTRACT_DITCH &&
	       data->ditch.parapet == TRUE)) &&
	    type_code != CTDB_ABSTRACT_FIGHTING_POS)
	  continue;

	/* Draw forward arrows */

	gc             = tactmap->gc.background;

	if (type_code == CTDB_ABSTRACT_FIGHTING_POS)
        {
	    segs[0].x = tactmap_FX_pixel_number(verts[0],cx,hsw,ppm);
            segs[0].y = tactmap_FY_pixel_number(verts[1],cy,hsh,ppm);
	    segs[1].x = tactmap_FX_pixel_number((verts[4] + verts[6]) / 2.0,
				cx,hsw,ppm);
	    segs[1].y = tactmap_FY_pixel_number((verts[5] + verts[7]) / 2.0,
				cy,hsh,ppm);
	    segs[2].x = tactmap_FX_pixel_number(verts[2],cx,hsw,ppm);
	    segs[2].y = tactmap_FY_pixel_number(verts[3],cy,hsh,ppm);

	    if (data->fighting_pos.turret_depth)
	    {
            	segs[3].x = tactmap_FX_pixel_number(
                           (verts[0] + verts[2] + verts[4] + verts[6]) / 4.0,
                           cx,hsw,ppm);
            	segs[3].y = tactmap_FY_pixel_number(
                           (verts[1] + verts[3] + verts[5] + verts[7]) / 4.0,
                           cy,hsh,ppm);
            	segs[4].x = segs[0].x;
            	segs[4].y = segs[0].y;

            	XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix, gc,
                        segs, 5, CoordModeOrigin);
            }
	    else
	      XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix, gc,                        	segs, 3, CoordModeOrigin);	
	}
	else
	{
	    num_vrts = 2 * n_verts - 5;
   	    bottom   = verts + n_verts - 2;
	    top      = bottom + 2;

	    for (l1=n_verts/2-1; l1; l1--, bottom -= 2, top += 2)
	    {
		bottom_diff[X] = bottom[X-2] - bottom[X];
		bottom_diff[Y] = bottom[Y-2] - bottom[Y];
	    	manhatt = 2.0 / (fabs(bottom_diff[X])+fabs(bottom_diff[Y]));

		offset1 = .5 - manhatt;
		offset2 = .5 + manhatt;

	    	segs[--num_vrts].x = tactmap_FX_pixel_number(
			bottom[X] + offset1 * bottom_diff[X],cx,hsw,ppm);
            	segs[num_vrts].y   = tactmap_FY_pixel_number(
			bottom[Y] + offset1 * bottom_diff[Y],cy,hsh,ppm);

	    	segs[--num_vrts].x = tactmap_FX_pixel_number(
			(top[X] + top[2+X]) / 2.0,cx,hsw,ppm);	
	    	segs[num_vrts].y   = tactmap_FY_pixel_number(
			(top[Y] + top[2+Y]) / 2.0,cy,hsh,ppm);

		segs[--num_vrts].x = tactmap_FX_pixel_number(
                        bottom[X] + offset2 * bottom_diff[X],cx,hsw,ppm);
                segs[num_vrts].y   = tactmap_FY_pixel_number(
                        bottom[Y] + offset2 * bottom_diff[Y],cy,hsh,ppm);

		if (l1 > 1)
		{
		    segs[--num_vrts].x = tactmap_FX_pixel_number(
			bottom[X-2],cx,hsw,ppm);
		    segs[num_vrts].y   = tactmap_FY_pixel_number(
			bottom[Y-2],cy,hsh,ppm);
		}
	    }

	    XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix, gc, 
			segs, 2*n_verts - 5, CoordModeOrigin);
	}
    }
}


void tactmap_draw_craters(
			  TACTMAP_PTR tactmap,
			  TACTMAP_OBJECT_PTR object, 
			  int32 use_erase_gc,
			  TACTMAP_BOUNDS *bounds,
			  CTDB_SEARCH_SPACE_PTR space
			  )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.craters, CTDB_ABSTRACT_CRATER);
}


void tactmap_draw_ditches(
			  TACTMAP_PTR tactmap,
			  TACTMAP_OBJECT_PTR object, 
			  int32 use_erase_gc,
			  TACTMAP_BOUNDS *bounds,
			  CTDB_SEARCH_SPACE_PTR space
			  )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.ditches, CTDB_ABSTRACT_DITCH);
}


void tactmap_draw_fpos(
		       TACTMAP_PTR tactmap,
		       TACTMAP_OBJECT_PTR object, 
		       int32 use_erase_gc,
		       TACTMAP_BOUNDS *bounds,
		       CTDB_SEARCH_SPACE_PTR space
		       )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.fpos, CTDB_ABSTRACT_FIGHTING_POS);
}

void tactmap_draw_berms(
			TACTMAP_PTR tactmap,
			TACTMAP_OBJECT_PTR object, 
			int32 use_erase_gc,
			TACTMAP_BOUNDS *bounds,
			CTDB_SEARCH_SPACE_PTR space
			)
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.berms, CTDB_ABSTRACT_BERM);
}

void tactmap_draw_rubble(
			 TACTMAP_PTR tactmap,
			 TACTMAP_OBJECT_PTR object, 
			 int32 use_erase_gc,
			 TACTMAP_BOUNDS *bounds,
			 CTDB_SEARCH_SPACE_PTR space
			 )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.rubble, CTDB_ABSTRACT_RUBBLE);
}

void tactmap_draw_snow(
		       TACTMAP_PTR tactmap,
		       TACTMAP_OBJECT_PTR object, 
		       int32 use_erase_gc,
		       TACTMAP_BOUNDS *bounds,
		       CTDB_SEARCH_SPACE_PTR space
		       )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.snow, CTDB_ABSTRACT_SNOW);
}

void tactmap_draw_camouflage(
			     TACTMAP_PTR tactmap,
			     TACTMAP_OBJECT_PTR object, 
			     int32 use_erase_gc,
			     TACTMAP_BOUNDS *bounds,
			     CTDB_SEARCH_SPACE_PTR space
			     )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.camouflage, CTDB_ABSTRACT_CAMOUFLAGE);
}

void tactmap_draw_hydro_surface(
				TACTMAP_PTR tactmap,
				TACTMAP_OBJECT_PTR object, 
				int32 use_erase_gc,
				TACTMAP_BOUNDS *bounds,
				CTDB_SEARCH_SPACE_PTR space
				)
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.hydro_surface,
			  CTDB_ABSTRACT_HYDRO_SURFACE);
}

void tactmap_draw_hydro_subsurface(
				   TACTMAP_PTR tactmap,
				   TACTMAP_OBJECT_PTR object, 
				   int32 use_erase_gc,
				   TACTMAP_BOUNDS *bounds,
				   CTDB_SEARCH_SPACE_PTR space
				   )
{
    tactmap_draw_abstract(tactmap, object, use_erase_gc, bounds, space,
			  tactmap->gc.hydro_subsurface,
			  CTDB_ABSTRACT_HYDRO_SUBSURFACE);
}

#endif /*USE_X*/
