#ifndef lint
static char rcsid [] = "$RCSfile: tmap_bgr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_bgr.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw bgr icons                                    *
*       Created: Tue Jun 22 1993                                            *
*       Author: ajc                                                         *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
/*changed class to sensClass 13 Aug 97 AUSTIN*/
#ifdef USE_X

#include "libtmap_local.h"
#include <libbgr.h>
#include <obj_type.h>
#include <math.h>
#include <stdalloc.h> /*common/include/global*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdstring.h> /*common/include/global*/

static void 
allocate_cached_pixmap(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
		       int32 width, int32 height)
{
    /* Allocate an element from the pool */
    object->bgr_draw_cache.pooled_pixmap =
      bgrdb_allocate_pixmap(DPYWIN, 
			    DefaultDepthOfScreen(XtScreen(tactmap->widget)),
			    width, height);
}

static void 
deallocate_cached_pixmap(TACTMAP_OBJECT_PTR object)
{
    /* Give the element back to the pool, and zero out our pixmap fields */
    bgrdb_deallocate_pixmap(object->bgr_draw_cache.pooled_pixmap);
    object->bgr_draw_cache.pooled_pixmap = NULL;    

    /* Do not clear out the width and height in the draw cache.  They will
     * be used as an approximation to see if I will ever need to draw 
     * this again. 
     */
}

static void 
compute_bgr_pixmap(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
		   TACTMAP_OBJECT_DFN *dfn)
{
    DECLARE_REGS;
    XRectangle   *bounds = &object->drawn_bounds; /* Pixel Coordinates */

    /* Adjust the bounds to the new location */
    bounds->x      = tactmap_FX_pixel_number(dfn->bgr.x, cx, hsw, ppm);
    bounds->y      = tactmap_FY_pixel_number(dfn->bgr.y, cy, hsh, ppm);

    /* First, see if the object already has a pixmap and all the data
     * is unchanged from the last time that the pixmap was computed.
     */
    if (object->bgr_draw_cache.pooled_pixmap &&
	(object->bgr_draw_cache.object_type    == dfn->bgr.object_type) &&
	(object->bgr_draw_cache.override       == dfn->bgr.override) &&
	(!object->bgr_draw_cache.override ||
	 (object->bgr_draw_cache.bgr_strings   == dfn->bgr.dfn)) &&
	(object->bgr_draw_cache.rotation       == dfn->bgr.rotation) &&
	(object->bgr_draw_cache.size           == dfn->bgr.size) &&
	(object->bgr_draw_cache.line_thickness == dfn->bgr.line_thickness) &&
	(object->bgr_draw_cache.gc             == dfn->bgr.gc) &&
	!strcmp(object->bgr_draw_cache.call_sign, dfn->bgr.call_sign))
      return;
    else
    {
#define MAX_DESC_ELTS	20
	BgrSymbolElt elts[MAX_DESC_ELTS];
	int32        nelts;
	int32        width;
	int32        height;
	int32        cx;
	int32        cy;
	Pixmap       pix;

	/* Calculate the size of the BGR icon */
	bgrdb_calculate_elements_and_extents(elts, &nelts, MAX_DESC_ELTS,
					     &width, &height, &cx, &cy, 
					     XtDisplay(tactmap->widget),
					     dfn->bgr.gc,
					     dfn->bgr.size, 
					     dfn->bgr.line_thickness,
					     dfn->bgr.rotation,
					     dfn->bgr.dashed,
					     dfn->bgr.call_sign,
					     dfn->bgr.object_type,
					     dfn->bgr.dfn);

	/* If I already have a pooled pixmap, give it back, because the
	 * update may need a pixmap of a different size.
	 */

	if (object->bgr_draw_cache.pooled_pixmap)
	  deallocate_cached_pixmap(object);      
	
	/* Allocate a pixmap of the correct size */
	allocate_cached_pixmap(tactmap, object, width, height);

	pix = object->bgr_draw_cache.pooled_pixmap->pixmap;

	/* Clear the pixmap */
	XFillRectangle(XtDisplay(tactmap->widget), pix, 
		       tactmap->gc.background,
		       0, 0, width, height);

	/* Draw the BGR icon on the new allocated pixmap */
	bgrdb_draw_elements_to_pixmap(elts, nelts,
				      cx, cy, 
				      XtDisplay(tactmap->widget), dfn->bgr.gc,
				      pix);
	   
	/* Remember the parameters for all this work I did so that
	 * I may not have to do it again.
	 */
	object->bgr_draw_cache.object_type    = dfn->bgr.object_type;
	object->bgr_draw_cache.override       = dfn->bgr.override;
	object->bgr_draw_cache.bgr_strings    = dfn->bgr.dfn;
	object->bgr_draw_cache.rotation       = dfn->bgr.rotation;
	object->bgr_draw_cache.size           = dfn->bgr.size;
	object->bgr_draw_cache.line_thickness = dfn->bgr.line_thickness;
	object->bgr_draw_cache.gc             = dfn->bgr.gc;
	strcpy(object->bgr_draw_cache.call_sign, dfn->bgr.call_sign);
	object->bgr_draw_cache.width          = width;
	object->bgr_draw_cache.height         = height;
	object->bgr_draw_cache.center_x       = cx;
	object->bgr_draw_cache.center_y       = cy;
    }
}

/*ARGSUSED*/
void 
tactmap_draw_bgr(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		 int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
		 CTDB_SEARCH_SPACE_PTR space, int32 clear_area_when_finished,
		 TACTMAP_OBJECT_DFN *dfn)
{
    DECLARE_REGS;
    int32 x, y;
    int32 change=0, old_x, old_y;
    SNSTVE_OBJECT snsobj;
    int32 should_draw;

    /* Adjust old dfn coordinates to be relative to the 
     * current center cell */
    CHANGE_CONTEXT(object->dfn.bgr.cell, object->dfn.bgr.x,
		   object->dfn.bgr.y,cc);

    /* Adjust new dfn coordinates to be relative to the 
     * current center cell 
     */
    if ( dfn )
    {
	CHANGE_CONTEXT(dfn->bgr.cell, dfn->bgr.x, dfn->bgr.y, cc);
    }
    
    /* Did it draw last time? (It did if it has a pooled pixmap)) */
    if (object->bgr_draw_cache.pooled_pixmap)
    {
	/* Erase the last incarnation */
	x = tactmap_FX_pixel_number(object->dfn.bgr.x, cx, hsw, ppm);
	y = tactmap_FY_pixel_number(object->dfn.bgr.y, cy, hsh, ppm);

	/* Deal with centering */
	x -= object->bgr_draw_cache.center_x;
	y -= object->bgr_draw_cache.center_y;
	

	if (object->display_mode == PERMANENT)
	{
	    XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		      tactmap->background_pix, tactmap->gc.background,
		      x, y,
		      object->bgr_draw_cache.width,
		      object->bgr_draw_cache.height,
		      x, y);

	    if (clear_area_when_finished)
	      XClearArea(DPYWIN,
			 x, y,
			 object->bgr_draw_cache.width,
			 object->bgr_draw_cache.height,
			 0);
	}
	else
	  XCopyArea(XtDisplay(tactmap->widget), tactmap->background_pix,
		    XtWindow(tactmap->widget), tactmap->gc.background,
		    x, y,
		    object->bgr_draw_cache.width,
		    object->bgr_draw_cache.height,
		    x, y);

	old_x = x;
	old_y = y;
	change = 1;
    }

    if (!dfn)
      dfn = &object->dfn;
    
    if (!object->bgr_draw_cache.width  || !object->bgr_draw_cache.height)
      /* 
       * Deal with case of not having computed pixmap extents yet.
       * Compute a pixmap to get the height and width.
       */
      compute_bgr_pixmap(tactmap, object, dfn);

    /*
     * Check whether this object should be drawn 
     */
    if ((dfn->bgr.x + 
	 (object->bgr_draw_cache.width - object->bgr_draw_cache.center_x)/ppm
	 > bounds->min_x) &&
	(dfn->bgr.x - object->bgr_draw_cache.center_x/ppm < bounds->max_x) &&
	(dfn->bgr.y + object->bgr_draw_cache.center_y/ppm > bounds->min_y) &&
	(dfn->bgr.y -
	 (object->bgr_draw_cache.height + object->bgr_draw_cache.center_y)/ppm 
	 < bounds->max_y))
      should_draw = TRUE;
    else
      should_draw = FALSE;

    if (!should_draw)
    {
	if (change)
	{
	    snstve_remove_object(tactmap->sensitive, old_x, old_y,
				 dfn->sensitive.sensClass,
				 dfn->sensitive.instance);
	    deallocate_cached_pixmap(object);
	}
	return;
    }

    compute_bgr_pixmap(tactmap, object, dfn);

    x = tactmap_FX_pixel_number(dfn->bgr.x, cx, hsw, ppm);
    y = tactmap_FY_pixel_number(dfn->bgr.y, cy, hsh, ppm);

    /* Deal with centering */
    x -= object->bgr_draw_cache.center_x;
    y -= object->bgr_draw_cache.center_y;

    if (use_erase_gc)
    {
	if (object->display_mode == PERMANENT)
	  XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		    tactmap->background_pix, tactmap->gc.background,
		    x, y,
		    object->bgr_draw_cache.width,
		    object->bgr_draw_cache.height,
		    x, y);
	else
	  XClearArea(DPYWIN,
		     x, y, 
		     object->bgr_draw_cache.width,
		     object->bgr_draw_cache.height,
		     0);

	snstve_remove_object(tactmap->sensitive, x, y,
			     dfn->sensitive.sensClass,
			     dfn->sensitive.instance);
	deallocate_cached_pixmap(object);
    }
    else
    {
	if (object->display_mode == PERMANENT)
	  XCopyArea(XtDisplay(tactmap->widget),
		    object->bgr_draw_cache.pooled_pixmap->pixmap,
		    tactmap->background_pix, tactmap->gc.background,
		    0, 0,
		    object->bgr_draw_cache.width,
		    object->bgr_draw_cache.height,
		    x, y);
	else
	  XCopyArea(XtDisplay(tactmap->widget),
		    object->bgr_draw_cache.pooled_pixmap->pixmap,
		    XtWindow(tactmap->widget), tactmap->gc.background,
		    0, 0,
		    object->bgr_draw_cache.width,
		    object->bgr_draw_cache.height,
		    x, y);
	snsobj.sensClass            = dfn->sensitive.sensClass;
	snsobj.instance         = dfn->sensitive.instance;
	snsobj.user_data        = dfn->sensitive.user_data;
	snsobj.kind             = SNSTVE_RECTANGLE;
	snsobj.attr.rect.x      = x-2;
	snsobj.attr.rect.y      = y-2;
	snsobj.attr.rect.width  = object->bgr_draw_cache.width  + 4;
	snsobj.attr.rect.height = object->bgr_draw_cache.height + 4;
	
	if (!change ||
	    !snstve_change_object(tactmap->sensitive,
				  old_x, old_y, &snsobj))
	  snstve_add_object(tactmap->sensitive, &snsobj);
    }

    if (clear_area_when_finished &&
	(object->display_mode == PERMANENT))
      XClearArea(DPYWIN, x, y,
		 object->bgr_draw_cache.width,
		 object->bgr_draw_cache.height,
		 0);
}

#endif
/*USE_X*/
