#ifndef lint
static char rcsid [] = "$RCSfile: tmap_buildgs.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_buildings.c                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw buildings                                    *
*       Created: Tue Jul 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <stdext.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>

#define COSINE_30_SQ 0.75		/* Cosine(30 degrees) squared */

/* Returns TRUE iff p1 p2 p3 is roughly a straight line (currently defined
 * as creating an angle of <= 30 degrees)
 */
static int32
pts_in_line(float64 p1[XY],
	    float64 p2[XY],
	    float64 p3[XY])
{
    float64 v1[XY];
    float64 v2[XY];
    float64 dp;
    float64 mag_sq_v1;
    float64 mag_sq_v2;    
    float64 prod;

    vmat2_vec_sub64(p2, p1, v1);
    vmat2_vec_sub64(p3, p2, v2);
    
    dp = fabs(vmat2_dot_prod64(v1, v2));
    
    mag_sq_v1 = vmat2_dot_prod64(v1, v1);
    mag_sq_v2 = vmat2_dot_prod64(v2, v2);
    
    /* If the angle is <= 30 degrees, this will return TRUE--we're not
     * worried about the line being EXACTLY straight, just more-or-less
     * straight.
     */
    return ((dp*dp) >= (mag_sq_v1 * mag_sq_v2 * COSINE_30_SQ));
    
}


static void
tactmap_draw_mes_bridge(TACTMAP_PTR       tactmap,
			CTDB_MES_NODE_KEY mes_key,
			int32             n_vert, 
			float64           verts[][XYZ]
			)
{
    int32 i;
    int32 index = 0;
    int32 start;
    TACTMAP_BRIDGE_DATA bridge;
    float64 vec1[XY], vec2[XY], mag1, mag2;
    
    VMAT2_VEC_ZERO64(bridge.loc);
    
    for (i=0;i<n_vert;i++)
    {
	bridge.loc[X] += verts[i][X];
	bridge.loc[Y] += verts[i][Y];
    }

    /* centroid of the bridge */
    vmat2_scal_vec_mul64(1.0/n_vert,bridge.loc,bridge.loc);
    bridge.loc[CELL2D] = ctdb_get_cell_id(tactmap->ctdb);

    /* Find a corner (this assumes that bridge is rectangular) */
    while((index < n_vert) && (pts_in_line(verts[index % n_vert],
					   verts[(index+1) % n_vert],
					   verts[(index+2) % n_vert])))
    {
	index++;
    }

    
    if (index >= n_vert)
    {
	/* Bridge is basically a straight line.  We don't know how to draw
	 * this, so don't.
	 */
	return;
    }

    
    
    /* Find the next corner */
    start = (index + 1) % n_vert;
    index = start;
    
    while(pts_in_line(verts[index % n_vert],
		      verts[(index+1) % n_vert],
		      verts[(index+2) % n_vert]))
    {
	index = (index + 1) % n_vert;
    }
    
    index = (index + 1) % n_vert;

    /* Index now points to the next corner.  Find the distance between them */
    vmat2_vec_sub64(verts[index], verts[start], vec1);
    mag1 = VMAT2_VEC_MAG_SQ(vec1);

    /* Find a third corner */
    index = start - 2;
    if (index < 0)
    {
	index += n_vert;
    }
    
    while(pts_in_line(verts[index % n_vert],
		      verts[(index+1) % n_vert],
		      verts[(index+2) % n_vert]))
    {
	index = index - 1;
	if (index < 0)
	{
	    index += n_vert;
	}
    }

    index = (index + 1) % n_vert;

    vmat2_vec_sub64(verts[index], verts[start], vec2);
    mag2 = VMAT2_VEC_MAG_SQ(vec2);

    if (mag2 > mag1)
    {
	VMAT2_VEC_COPY64(vec2, bridge.del);
	bridge.width = fsqrt(mag1);
    }
    else
    {
	VMAT2_VEC_COPY64(vec1, bridge.del);
	bridge.width = fsqrt(mag2);
    }
    
    bridge.mes = TRUE;
    tactmap_draw_bridge(tactmap, &bridge);
}

/*ARGSUSED*/
static void 
tactmap_draw_volumes(TACTMAP_PTR           tactmap, 
		     TACTMAP_OBJECT_PTR    object,
		     int32                 use_erase_gc, 
		     TACTMAP_BOUNDS       *bounds,
		     CTDB_SEARCH_SPACE_PTR space, 
		     int32                 type_code)
{
    DECLARE_REGS;
    XPoint pts[CTDB_VERTICES_MAX+1];
    float64 verts[CTDB_VERTICES_MAX][XY];
    int32 n_vert, j;
    CTDB_TERRAIN_SUBCLASS type;
    int32 xmin,xmax,ymin,ymax;
    GC gc;
    SNSTVE_OBJECT snsobj;
    CTDB_MES_NODE_KEY node_key;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;
    snsobj.kind = SNSTVE_RECTANGLE;
    
    while (n_vert = ctdb_next_volume(space, &type, NULL, NULL, &node_key,
				     verts, NULL))
    {
	switch (type)
	{
	  case CTDB_FS_BUILDING:
	    if ( type_code != CTDB_FS_BUILDING )
	      continue;
	    gc = tactmap->gc.buildings;
	    break;
	    
	  case CTDB_FS_MES:
	    if ( type_code == CTDB_FS_BUILDING )
	    {
		CTDB_MES_TYPE     mes_type;

		ctdb_mes_get_mes(tactmap->ctdb, node_key, &mes_type, 
				 NULL, NULL, NULL, NULL);
		if ( mes_type == CTDB_MES_BRIDGE )
		{
		    int32 i;
		    float64 mverts[100][XYZ];
		    
		    ctdb_mes_get_mes(tactmap->ctdb, node_key, &mes_type, 
				     NULL, NULL, &n_vert, mverts);
		    tactmap_draw_mes_bridge(tactmap, node_key, n_vert, mverts);
		    continue;
		}
		gc = tactmap->gc.snow;
		break;
	    }
	    continue;
	  case CTDB_FS_ROCK_DROP:
	    if ( type_code != type )
	      continue;
	    gc = tactmap->gc.rock_drops;
	    break;

	  case CTDB_FS_PLUG:
	    if ( type_code != type )
	      continue;
	    gc = tactmap->gc.plug;
	    break;
	  default:
	    continue;
	}
	
	if (use_erase_gc)
	  gc = tactmap->gc.background;
	
	if (tactmap_off_screen64(n_vert, verts, bounds))
	  continue;

	/* Bundle up the points */
	xmin=ymin = 0x10000;
	xmax=ymax = 0;
	for (j=0;j<n_vert;j++)
	{
	    pts[j].x = tactmap_FX_pixel_number(verts[j][X],cx,hsw,ppm);
	    pts[j].y = tactmap_FY_pixel_number(verts[j][Y],cy,hsh,ppm);
	    
	    if (pts[j].x < xmin)
	      xmin = pts[j].x;
	    if (pts[j].y < ymin)
	      ymin = pts[j].y;
	    if (pts[j].x > xmax)
		  xmax = pts[j].x;
	    if (pts[j].y > ymax)
	      ymax = pts[j].y;
	}
	
	XFillPolygon(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		     gc, pts, j, Nonconvex, CoordModeOrigin);
	
	snsobj.instance = 0;
	snsobj.attr.rect.x = xmin;
	snsobj.attr.rect.y = ymin;
	snsobj.attr.rect.width = xmax-xmin;
	snsobj.attr.rect.height = ymax-ymin;
	    
	snstve_add_object(tactmap->sensitive, &snsobj);
    }
}

void 
tactmap_draw_buildings(TACTMAP_PTR           tactmap, 
		       TACTMAP_OBJECT_PTR    object,
		       int32                 use_erase_gc, 
		       TACTMAP_BOUNDS       *bounds,
		       CTDB_SEARCH_SPACE_PTR space)
{
    tactmap_draw_volumes(tactmap, object, use_erase_gc, bounds, space,
			 CTDB_FS_BUILDING);
}

void 
tactmap_draw_rock_drops(TACTMAP_PTR           tactmap, 
		       TACTMAP_OBJECT_PTR    object,
		       int32                 use_erase_gc, 
		       TACTMAP_BOUNDS       *bounds,
		       CTDB_SEARCH_SPACE_PTR space)
{
    tactmap_draw_volumes(tactmap, object, use_erase_gc, bounds, space,
			 CTDB_FS_ROCK_DROP);
}

void 
tactmap_draw_plugs(TACTMAP_PTR tactmap, 
		   TACTMAP_OBJECT_PTR object,
		   int32 use_erase_gc, 
		   TACTMAP_BOUNDS *bounds,
		   CTDB_SEARCH_SPACE_PTR space)
{
    tactmap_draw_volumes(tactmap, object, use_erase_gc, bounds, space,
			 CTDB_FS_PLUG);
}

#endif /*USE_X*/
