#ifndef lint
static char rcsid [] = "$RCSfile: tmap_clouds.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_cloud.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw clouds                *
*       Created:  May 16, 1996                                          *
*       Author:   Stephen A. Kukolich
       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#ifdef USE_X
#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <libdither.h>
#include <math.h>
#include <p_po.h>
#include <libenvui.h>
#include <libgcs.h>

#ifndef NO_ENV

#include <libenvcloud.h>


/* Constants */

#define CLOUD_PIXMAP_SIZE  4    /* screen pixels wide */
#define N_CLOUD_PIXMAPS    (DITHER_COUNT(CLOUD_PIXMAP_SIZE)/2)
#define CLOUD_PATCH_SIZE   10   /* screen pixels wide */


/* local functions */
static void draw_clouds(TACTMAP_PTR tactmap, Drawable d,
			TACTMAP_BOUNDS *bounds,
			TACTMAP_OBJECT_DFN *dfn, GC gc);

#endif /*not NO_ENV*/

void tactmap_init_clouds_pixmaps(TACTMAP_PTR tactmap)
{

#ifndef NO_ENV

    int i;

    DITHER_MATRIX(mat, CLOUD_PIXMAP_SIZE);
    DITHER_BITMAPS(bits, CLOUD_PIXMAP_SIZE);

    /* Make the generating matrix */
    dither_matrix(CLOUD_PIXMAP_SIZE, mat);

    /* Generate dither bitmaps */
    dither_bitmaps(CLOUD_PIXMAP_SIZE, mat, bits);

    tactmap->clouds_pix = (Pixmap*)STDCALLOC(N_CLOUD_PIXMAPS, sizeof(Pixmap));

    for(i=0;i<N_CLOUD_PIXMAPS;i++)
      tactmap->clouds_pix[i] =
	XCreateBitmapFromData(XtDisplay(tactmap->widget), 
			      RootWindowOfScreen(XtScreen(tactmap->widget)),
			      (char *)bits[i], 
			      CLOUD_PIXMAP_SIZE, 
			      CLOUD_PIXMAP_SIZE);
#else /*NO_ENV*/
    return;
#endif /*not NO_ENV*/

}

void tactmap_draw_clouds(TACTMAP_PTR tactmap, 
			 TACTMAP_OBJECT_PTR object, 
			 int32 use_erase_gc, 
			 TACTMAP_BOUNDS *bounds, 
			 CTDB_SEARCH_SPACE_PTR space, 
			 int32 clear_area_when_finished, 
			 TACTMAP_OBJECT_DFN *dfn)
{

#ifndef NO_ENV

    DECLARE_REGS;
    TACTMAP_BOUNDS   clipped_bounds;
    GC gc;

    if (!ecloud_model_exists_p())  return;

    /* Are we removing this object? */
    gc = use_erase_gc ? tactmap->gc.background : tactmap->gc.clouds;

    if (!dfn)  dfn = &object->dfn;
    
    clipped_bounds.min_x = NS_MAX(bounds->min_x,tactmap->playbox.min_x);
    clipped_bounds.max_x = NS_MIN(bounds->max_x,tactmap->playbox.max_x);
    clipped_bounds.min_y = NS_MAX(bounds->min_y,tactmap->playbox.min_y);
    clipped_bounds.max_y = NS_MIN(bounds->max_y,tactmap->playbox.max_y);

    draw_clouds(tactmap,tactmap->terrain_pix, &clipped_bounds,dfn,gc);

#else /*NO_ENV*/
    return;
#endif /*not NO_ENV*/
}

#ifndef NO_ENV

static void draw_clouds(TACTMAP_PTR tactmap, Drawable d,
			TACTMAP_BOUNDS *bounds,
			TACTMAP_OBJECT_DFN *dfn, GC gc)
{
    DECLARE_REGS;
    int32 min_sx, max_sx, min_sy, max_sy, dsx, dsy, dsxhalf, dsyhalf;
    int32 sx, sy, x, y;
    int32 this_pix;
    XRectangle rect;
    float32  position[3], value;

    min_sx = tactmap_X_pixel_number(bounds->min_x,cx,hsw,ppm);
    max_sx = tactmap_X_pixel_number(bounds->max_x,cx,hsw,ppm);
    max_sy = tactmap_Y_pixel_number(bounds->min_y,cy,hsh,ppm);
    min_sy = tactmap_Y_pixel_number(bounds->max_y,cy,hsh,ppm);
    rect.width  = dsx = CLOUD_PATCH_SIZE;
    rect.height = dsy = CLOUD_PATCH_SIZE;
    dsxhalf = dsx/2;
    dsyhalf = dsy/2;

    position[2] = 0;

    XSetFillStyle(XtDisplay(tactmap->widget), gc, FillStippled);

    for (sx = min_sx; sx < max_sx; sx += dsx)
      for (sy = min_sy; sy < max_sy; sy += dsy)
      {
	  x = X_METER_NUMBER((sx+dsxhalf),cx,hsw,ppm);
	  y = Y_METER_NUMBER((sy+dsyhalf),cy,hsh,ppm);

	  if ((x < cx-hsw) || (x > cx+hsw) || (y < cy-hsh) || (y > cy+hsh))
	    continue;

	  position[0] = x;
	  position[1] = y;
	  value = ecloud_get_2d_value(position);
	  this_pix = floor( value * N_CLOUD_PIXMAPS );
	  if (this_pix < 0)  this_pix = 0;
	  if (this_pix > N_CLOUD_PIXMAPS-1)  this_pix = N_CLOUD_PIXMAPS-1;

	  if (this_pix > 0)
	  {
	      rect.x = sx;
	      rect.y = sy;

	      XSetStipple(XtDisplay(tactmap->widget),gc,
			  tactmap->clouds_pix[this_pix]);
	      XFillRectangles(XtDisplay(tactmap->widget),
			      d, gc,
			      &rect, 1);
	  }
      }

    XSetFillStyle(XtDisplay(tactmap->widget), gc, FillSolid);

}

#endif /*not NO_ENV*/

#endif /*USE_X*/
