#ifndef lint
static char rcsid [] = "$RCSfile: tmap_contours.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_contours.c                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw contours                                     *
*       Created: Tue Jul 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <libctdb.h>
#include <math.h>

static struct
{
    TACTMAP_PTR tactmap;
    Pixmap pixmap;
    float64 major_spacing;
    GC minor_gc;
    GC major_gc;
} dcs;

static void draw_ctdb_segs(int32 nsegs, float64 *segs, float64 *elevs)
{
    register float64 ppm = dcs.tactmap->cur->ppm;
    register int32 cx = dcs.tactmap->cur->cx;
    register int32 cy = dcs.tactmap->cur->cy;
    register int32 hsw = dcs.tactmap->cur->hsw;
    register int32 hsh = dcs.tactmap->cur->hsh;
    int32 i;
    int16 x0,x1,y0,y1,dx,dy;
    int16 pts[2][CTDB_MAX_CONTOUR_PTS];
    int16 lines[2][CTDB_MAX_CONTOUR_PTS];
    int32 p[2], l[2];
    int32 hi;

    p[0] = p[1] = 0;
    l[0] = l[1] = 0;
    for (i=0;i<nsegs;i+=4)
    {
	x0 = tactmap_X_pixel_number((int32)segs[i],cx,hsw,ppm);
	y0 = tactmap_Y_pixel_number((int32)segs[i+1],cy,hsh,ppm);
	x1 = tactmap_X_pixel_number((int32)segs[i+2],cx,hsw,ppm);
	y1 = tactmap_Y_pixel_number((int32)segs[i+3],cy,hsh,ppm);

	dx = x1-x0;
	dy = y1-y0;

	if ((dx == 0) && (dy == 0))
	  continue;

	if (dx < 0) dx = -dx;
	if (dy < 0) dy = -dy;

	hi = (dcs.major_spacing &&
	      (fmod(elevs[i>>2], dcs.major_spacing) < 0.1));

	if ((dx<2) && (dy<2))
	{
	    pts[hi][p[hi]] = x0;
	    pts[hi][p[hi]+1] = y0;
	    pts[hi][p[hi]+2] = x1;
	    pts[hi][p[hi]+3] = y1;
	    p[hi] += 4;
	}
	else
	{
	    lines[hi][l[hi]] = x0;
	    lines[hi][l[hi]+1] = y0;
	    lines[hi][l[hi]+2] = x1;
	    lines[hi][l[hi]+3] = y1;
	    l[hi] += 4;
	}
    }

    if (l[0])
      XDrawSegments(XtDisplay(dcs.tactmap->widget), dcs.pixmap,
		    dcs.minor_gc, (XSegment *)lines[0], l[0]/4);
    if (p[0])
      XDrawPoints(XtDisplay(dcs.tactmap->widget), dcs.pixmap,
		  dcs.minor_gc, (XPoint *)pts[0], p[0]/2,
		  CoordModeOrigin);
    if (l[1])
      XDrawSegments(XtDisplay(dcs.tactmap->widget), dcs.pixmap,
		    dcs.major_gc, (XSegment *)lines[1], l[1]/4);
    if (p[1])
      XDrawPoints(XtDisplay(dcs.tactmap->widget), dcs.pixmap,
		  dcs.major_gc, (XPoint *)pts[1], p[1]/2,
		  CoordModeOrigin);
}

void tactmap_draw_contours(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			   int32 use_erase_gc, TACTMAP_BOUNDS *bounds)
{
    dcs.tactmap = tactmap;
    dcs.pixmap = tactmap->terrain_pix;
    dcs.major_spacing = object->dfn.contours.major_spacing; 
    if (use_erase_gc)
      dcs.minor_gc = dcs.major_gc = tactmap->gc.background;
    else
    {
	dcs.minor_gc = tactmap->gc.contours[0];
	dcs.major_gc = tactmap->gc.contours[1];
    }

    ctdb_contour_segments(tactmap->ctdb, object->dfn.contours.minor_spacing,
			  bounds->min_x, bounds->min_y,
			  bounds->max_x, bounds->max_y,
			  draw_ctdb_segs);
}

#endif /*USE_X*/
