#ifndef lint
static char rcsid [] = "$RCSfile: tmap_dragon.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_dragon.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw dragon's teeth                               *
*       Created: Tue Feb 13 1996                                            *
*       Author: mlongtin                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <libgeometry.h>
#include <math.h>
#include <stdext.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define MIN_ARC_SIZE         2     /* pixels */
#define MAX_ARC_SIZE         10    /* pixels */
#define DESIRED_ARC_SIZE     2     /* meters */
#define FULL_REVOLUTION      23040 /* (360 * 64) */

#define MAX_TEETH_PER_LINE  2000   



static void draw_dt_segment(
    Display *display, 
    Window pixmap, 
    GC gc,
    int32 x0, int32 y0, int32 x1, int32 y1,
    int32 num_teeth,
    XRectangle *clip_box, 
    float64 ppm)
{
    int32 j;
    int32 x, y;
    int32 arc_size;
    float64 dx, dy, sx, sy;

    arc_size = NS_LIMITS(MIN_ARC_SIZE, DESIRED_ARC_SIZE * ppm, MAX_ARC_SIZE);

    if (num_teeth > 1)
    {
        dx = (float64)(x1 - x0) / (float64)(num_teeth-1.0);
        dy = (float64)(y1 - y0) / (float64)(num_teeth-1.0);
     } 
     else
	dy = dy = 0.0;

    /* X wants the x,y pos to be the top corner of the bounding box
     * that contains the arc.. adjust the position from the center to
     * the corner.
     */
    sx = x0 - arc_size*0.5;
    sy = y0 - arc_size*0.5;

    for (j=0;j<num_teeth;j++)
    {
        x = (int32)(sx + dx * (j));
        y = (int32)(sy + dy * (j));

        /* This is not a speed optimization. There seems to
         * be a bug in X which causes it to die if arcs are drawn
         * far off the screen.
         */
        if (!POINT_IN_RECT(x,y,*clip_box))
          continue;

        XFillArc(display, pixmap, gc, x, y,
                 arc_size, arc_size, 0, FULL_REVOLUTION);
    }
}



/*ARGSUSED*/
void tactmap_draw_dragons_teeth(TACTMAP_PTR tactmap,
				TACTMAP_OBJECT_PTR object, 
				int32 use_erase_gc,
				TACTMAP_BOUNDS *bounds,
				CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XSegment segs[5000];
    float32 *verts;
    float64 vertices[CTDB_VERTICES_MAX][XY];
    float64 locations[MAX_TEETH_PER_LINE][XY];  
    CTDB_ABSTRACT_DATA *data;
    XRectangle bounding_box;
    CTDB_LINEAR_DESC feature_info;
    XGCValues gcv;
    GC gc;
    int32 x, y;
    int32 j, k;
    int32 n_verts;
    SNSTVE_OBJECT snsobj;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;

    bounding_box.x      = tactmap_X_pixel_number(bounds->min_x,cx,hsw,ppm);
    bounding_box.y      = tactmap_Y_pixel_number(bounds->max_y,cy,hsh,ppm);
    bounding_box.width  = PIXELS(bounds->max_x-bounds->min_x,ppm);
    bounding_box.height = PIXELS(bounds->max_y-bounds->min_y,ppm);

    gcv.line_width = 0;
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.dragons_teeth,
	      GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.background,
	      GCLineWidth, &gcv);

    /* Draw the aggregate (abstract) representation of the dragon's teeth. */
    k = 0;
    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_DRAGONS_TEETH,
					&verts, &data, NULL))
    {
	if (tactmap_off_screen32(n_verts, verts, bounds))
	  continue;

	/* Bundle up the points */
	for (j=0;j<n_verts;j++)
	{
	    x = tactmap_FX_pixel_number(verts[2*j+X],cx,hsw,ppm);
	    y = tactmap_FY_pixel_number(verts[2*j+Y],cy,hsh,ppm);

	    if (j)
	    {
		snsobj.kind = SNSTVE_LINE;
		snsobj.instance = 0;
		snsobj.attr.line.x0 = segs[k].x1;
		snsobj.attr.line.y0 = segs[k].y1;
		snsobj.attr.line.x1 = x;
		snsobj.attr.line.y1 = y;
		snsobj.attr.line.width = 10;

		snstve_add_object(tactmap->sensitive, &snsobj);

		segs[k].x2 = x;
		segs[k].y2 = y;
		k++;
	    }

	    segs[k].x1 = x;
	    segs[k].y1 = y;
	}

	if (use_erase_gc)
	  gc = tactmap->gc.background;
	else
	  gc = tactmap->gc.dragons_teeth;

	if (k > 0)
	  XDrawSegments(XtDisplay(tactmap->widget), tactmap->terrain_pix,
			gc, segs, k);
    }

    /* Now draw the individual (physical) representations of the dragon's
     * teeth.
     */
    while (n_verts = ctdb_next_linear_ref(space, vertices, NULL,
					  &feature_info))
    {
        int32 num_teeth, i;
        float32 start_offset = 0.0;

	if ((feature_info.type != CTDB_FS_DRAGONS_TEETH) ||
            (!(object->dfn.dragons_teeth.mask & TACTMAP_DRAGONS_TEETH_IND)))
	  continue;

	if (n_verts == 1)
	{
	    x = tactmap_FX_pixel_number(vertices[0][X],cx,hsw,ppm);
	    y = tactmap_FY_pixel_number(vertices[0][Y],cy,hsh,ppm);

	    draw_dt_segment(XtDisplay(tactmap->widget),
			    tactmap->terrain_pix,
			    tactmap->gc.rock_drops,
			    x, y, x, y,
			    1,
			    &bounding_box, ppm);
        }
	else
	{
	    int32 x1, y1;

	    for (j=0; j<n_verts-1; j++)
	    {
		num_teeth = MAX_TEETH_PER_LINE;

		/* locate the teeth on this segment (only 2 verts).. */
		geo_locate_xy_pos_on_segments(
  	            &vertices[j], 2, &start_offset, 
	            feature_info.linear_data.dragons_tooth_info.spacing,
	            &num_teeth, locations);

		/* for efficiency reasons, we are only interested in the
		 * first and last points computed..  this is because
		 * we would have to map all the world positions to X
		 * screen coordinates foreach of the computed positions.
		 */

		x = tactmap_FX_pixel_number(locations[0][X],cx,hsw,ppm);
		y = tactmap_FY_pixel_number(locations[0][Y],cy,hsh,ppm);

		x1 = tactmap_FX_pixel_number(locations[num_teeth-1][X],
		                             cx,hsw,ppm);
		y1 = tactmap_FY_pixel_number(locations[num_teeth-1][Y],
		                             cy,hsh,ppm);

		draw_dt_segment(XtDisplay(tactmap->widget),
				tactmap->terrain_pix,
				tactmap->gc.rock_drops,
				x, y, x1, y1,
				num_teeth,
				&bounding_box, ppm);
	    }
	}
    }

}


#endif /*USE_X*/
