#ifndef lint
static char rcsid [] = "$RCSfile: tmap_enveff.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_enveffect.c                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw environmental effects on lines and area      *
*       Created: Wed Apr 12 1994                                            *
*       Author:  llovero                                                    *
*       Remarks: This code was lifted from tmap_intervis.c                  *
*                                                                           *
****************************************************************************/


#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <libdither.h>
#include <math.h>
#include <p_po.h>

#define PCB(v) (((v)==0.0) ? 0 : (1 + (int)((v) * (N_PERCENT-2))))

/* This routine draws the tick marks on an environment effect line at the */
/* current grid interval to make the line look like a ruler */
static void 
draw_ticked_line(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
		 TACTMAP_OBJECT_DFN *dfn, GC gc, int32 ruler_x0, 
		 int32 ruler_y0, int32 ruler_x1, int32 ruler_y1)
{
    float64 dx, dy, dist, cos_theta, sin_theta;
    float64 tick_pixels, xi;
    int32 x0,y0,x1,y1;
    int32 ruler_tick_interval = 1;
    Drawable d;

    ruler_tick_interval = dfn->line_env_effect.grid_interval;

    tick_pixels = ruler_tick_interval * tactmap->cur->ppm;

    dx = (ruler_x1 - ruler_x0);
    dy = (ruler_y1 - ruler_y0);
    dist = sqrt(dx*dx + dy*dy);

    if (dist < 0.1)
      return;

    cos_theta = dy / dist;
    sin_theta = dx / dist;

    /* Run down the X axis making ticks, then rotating them to
     * the right place
     */
    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);
    for (xi = 0.0; xi <= dist; xi += tick_pixels)
    {
	x0 = xi * sin_theta +  -TICK_LENGTH * cos_theta;
	y0 = xi * (-cos_theta) +  -TICK_LENGTH * sin_theta;
	x1 = xi * sin_theta + TICK_LENGTH * cos_theta;
	y1 = xi * (-cos_theta) + TICK_LENGTH * sin_theta;

	/* Subtract the Y values because the coordinate system is flopped */
	XDrawLine(XtDisplay(tactmap->widget), d, gc, 
		  ruler_x0+x0, ruler_y0-y0, ruler_x0+x1, ruler_y0-y1);
    }
}

/* This routine erases a previously drawn environment effect line */
static void erase_env_effect_line(TACTMAP_PTR tactmap, 
				  TACTMAP_OBJECT_PTR object, 
				  TACTMAP_OBJECT_DFN *dfn, GC gc)
{
    DECLARE_REGS;
    float64 x0,y0,x1,y1;
    int32 sx1,sy1,sx2,sy2;
    Drawable d;

    x0 = dfn->line_env_effect.point1[X];
    y0 = dfn->line_env_effect.point1[Y];
    x1 = dfn->line_env_effect.point2[X];
    y1 = dfn->line_env_effect.point2[Y];

    sx1 = tactmap_FX_pixel_number(x0,cx,hsw,ppm);    
    sy1 = tactmap_FY_pixel_number(y0,cy,hsh,ppm);
    sx2 = tactmap_FX_pixel_number(x1,cx,hsw,ppm);    
    sy2 = tactmap_FY_pixel_number(y1,cy,hsh,ppm);

    XSetLineAttributes(XtDisplay(tactmap->widget), gc, 7,
		       LineSolid, CapButt, JoinMiter);

    draw_ticked_line(tactmap, object, dfn, gc, sx1, sy1, sx2, sy2);


    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);

    XDrawLine(XtDisplay(tactmap->widget),d,gc, sx1,sy1,sx2,sy2);
    XSetLineAttributes(XtDisplay(tactmap->widget), gc, 3,
		       LineSolid, CapButt, JoinMiter);
}

/* This function draws an plain environmental effect line. It is just
   black with tick marks to show the length. */
static void draw_env_effect_line(TACTMAP_PTR tactmap,
				 TACTMAP_OBJECT_PTR object, 
				 TACTMAP_OBJECT_DFN *dfn, GC gc,
				 TACTMAP_BOUNDS *bounds)
{
    DECLARE_REGS;
    int32 x,y;
    float64 sx0, sy0, sx1, sy1;
    float64 x0,y0,x1,y1;
    Drawable d;

    x0 = dfn->line_env_effect.point1[X];
    y0 = dfn->line_env_effect.point1[Y];
    x1 = dfn->line_env_effect.point2[X];
    y1 = dfn->line_env_effect.point2[Y];

    if ((x0 < bounds->min_x) || (y0 < bounds->min_y) ||
	(x0 > bounds->max_x) || (y0 > bounds->max_y) ||
	(x1 < bounds->min_x) || (y1 < bounds->min_y) ||
	(x1 > bounds->max_x) || (y1 > bounds->max_y))
      return;

    sx0 = x = tactmap_FX_pixel_number(x0,cx,hsw,ppm);    
    sy0 = y = tactmap_FY_pixel_number(y0,cy,hsh,ppm);

    sx1 = x = tactmap_FX_pixel_number(x1,cx,hsw,ppm);    
    sy1 = y = tactmap_FY_pixel_number(y1,cy,hsh,ppm);

    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);

    XSetLineAttributes(XtDisplay(tactmap->widget),
		       gc, 3, LineSolid, CapButt, JoinRound);
    XSetFillStyle(XtDisplay(tactmap->widget), gc,FillSolid);

    XDrawLine(XtDisplay(tactmap->widget), d, gc, (int32)sx0, 
	      (int32)sy0, (int32)sx1, (int32)sy1);

    draw_ticked_line(tactmap, object, dfn, gc, (int32)sx0,(int32)sy0,x,y); 
}


void tactmap_draw_line_effect(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
			      int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			      CTDB_SEARCH_SPACE_PTR space, 
			      int32 clear_area_when_finished,
			      TACTMAP_OBJECT_DFN *dfn)
{
    register int32 cc=tactmap->cur->cc;
    GC gc;
    int32 cell;

	/* Erase the last incarnation */
    if ( dfn )
    {
	cell=dfn->line_env_effect.cell;

        CHANGE_CONTEXT(cell,
		       dfn->line_env_effect.point1[X],
		       dfn->line_env_effect.point1[Y],
		       cc);
        CHANGE_CONTEXT(dfn->line_env_effect.cell,
		       dfn->line_env_effect.point2[X],
		       dfn->line_env_effect.point2[Y],
		       cc);

	tactmap_draw_line_effect(tactmap, object, (int32) TRUE, bounds,
				 space, clear_area_when_finished,
				 (TACTMAP_OBJECT_DFN *) NULL);
    }

    cell = object->dfn.line_env_effect.cell;
    CHANGE_CONTEXT(cell,
		   object->dfn.line_env_effect.point1[X],
		   object->dfn.line_env_effect.point1[Y],
		   cc);
    CHANGE_CONTEXT(object->dfn.line_env_effect.cell,
		   object->dfn.line_env_effect.point2[X],
		   object->dfn.line_env_effect.point2[Y],
		   cc);

    if (!dfn)
      dfn = &object->dfn;
    
    /* Are we removing this object? */
    if (use_erase_gc)
    {

	if (object->display_mode == PERMANENT)
	  gc = tactmap->terrain_gc;
	else
	  gc = tactmap->background_gc;

	erase_env_effect_line(tactmap, object, dfn, gc);
    }
    else
    {
	gc = tactmap->gc.line_intervis;
	draw_env_effect_line(tactmap,object,dfn,gc, bounds);
    }
    
    if (clear_area_when_finished && (object->display_mode == PERMANENT))
      XClearWindow(DPYWIN);
}

#endif /*USE_X*/


