#ifndef lint
static char rcsid [] = "$RCSfile: tmap_grids.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_grids.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw grids                                        *
*       Created: Tue Jul 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X
/*#define DEBUG*/
#include <ctype.h>
#include "libtmap_local.h"
#include "tmap_north.h"
#include <libcoordinates.h>
#include <libworld.h>
#include <libgcs.h>
#include <stdalloc.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define MAX_SEGS 20000
#define MAX_LBLS 1000
#define CHARACTER_WIDTH 10 /* Kludge */

static void 
render_grid(    
	    TACTMAP_PTR          tactmap,
	    int32                erase_under_numbers,
	    Pixmap               pixmap,
	    GC                   gc,
	    int32                n_thin_segments,
	    float32              thin_segments[],
	    int32                n_thick_segments,
	    float32              thick_segments[],
	    int32                n_labels,
	    COORD_GRID_LABEL     labels[]
	    );
static void 
render_cell(    
	    TACTMAP_PTR          tactmap,
	    int32                erase_under_numbers,
	    Pixmap               pixmap,
	    GC                   gc,
	    int32                n_segments,
	    float32              segments[],
	    int32                n_labels,
	    COORD_GRID_LABEL     labels[]
	    );
    
static void
render_compass(
	       TACTMAP_PTR tactmap,
	       GC          gc
	       );

void 
tactmap_draw_pregenerated_grids(TACTMAP_PTR tactmap, Pixmap pixmap)
{
    static int32    n_thin,n_thick,n_labels;
    static float32 *thin=NULL;
    static float32 *thick=NULL;
    static COORD_GRID_LABEL *labels=NULL;
    static int32 allocate=1;
    
    if (allocate)
    {
	/* Initial allocation of space */
	thin     = (float32 *)STDALLOC(sizeof(float32)*XY*XY*100);
	n_thin   = 100;
	thick    = (float32 *)STDALLOC(sizeof(float32)*XY*XY*100);
	n_thick  = 100;
	labels   = (COORD_GRID_LABEL *)STDALLOC(sizeof(COORD_GRID_LABEL)*100);
	n_labels = 100;
	allocate = 0;
    }

    while (!wld_get_playbox_utm_grids(&n_thin,thin,
				      &n_thick,thick,
				      &n_labels,labels))
    {
	/* This processing will happen infrequently since the playbox is
	 * not expected to change very often.
	 */
	thin = (float32 *)STDREALLOC(thin,n_thin*XY*XY*sizeof(float32));
	thick= (float32 *)STDREALLOC(thick,n_thick*XY*XY*sizeof(float32));
	labels = (COORD_GRID_LABEL *)
	  STDREALLOC(labels,n_labels*sizeof(COORD_GRID_LABEL));
    }
    
    /* change the view of the world to be the world view */
    tactmap->cur = &tactmap->wv;
    render_grid(tactmap, (int32) TRUE, pixmap, tactmap->gc.grids,
		n_thin,thin,n_thick,thick,n_labels,labels);

    /* change the view back to the viewport view */
    tactmap->cur = &tactmap->vp;
}


static void 
display_scale(TACTMAP_PTR tactmap, GC gc)
{
    XFontStruct *fs;
    XGCValues gcv;
    int32 xoff;
    char buf[64];

    tactmap_scale_string(tactmap, buf);

    XGetGCValues(XtDisplay(tactmap->widget), gc, GCFont, &gcv);

    fs = XQueryFont(XtDisplay(tactmap->widget), gcv.font);

    xoff = fs->max_bounds.width * strlen(buf);

    XDrawImageString(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		     gc, tactmap->pix_width - 25 - xoff,
		     25 + fs->max_bounds.ascent,
		     buf, strlen(buf));

    XFreeFontInfo(NULL, fs, 1);
}



#define MAX_DIM 5
static int32 
tactmap_clip_intersections(float64 *v0, float64 *v1, TACTMAP_BOUNDS *bounds,
			   float64 *clip, int32 dim)
{
    int32   c;
    int32   l[XY],r[XY],b[XY],t[XY];
    int32   out;
    int32   i;
    float64 pts[XY][MAX_DIM], x0,y0,z0,x1,y1,z1, d;

    /* This algorithm finds intersections of the line from v0 to v1 with
     * box specified by min_x, max_x, min_y, max_y.  The code is
     * my own implementation of the Cohen-Sutherland algorithm.
     * Basically, you just keep clipping the line from both ends
     * until it is within the box.  I also will truncate up to MAX_DIM
     * other variables simultaneously (such as a Z coordinate).
     * This is used by ctdb_clip_line for regular line clipping, and
     * within the ctdb compiler for polygon clipping.
     */

    for(i=0;i<dim;i++)
      pts[0][i] = v0[i], pts[1][i] = v1[i];

    while (1)
    {
	for(i=0;i<XY;i++)
	{
	    l[i] = pts[i][X] < bounds->min_x;
	    r[i] = pts[i][X] > bounds->max_x;
	    b[i] = pts[i][Y] < bounds->min_y;
	    t[i] = pts[i][Y] > bounds->max_y;
	}

	if (l[X]&&l[Y] || r[X]&&r[Y] || b[X]&&b[Y] || t[X]&&t[Y])
	  return 0;

	out = 0;
	if (!(l[X] | r[X] | t[X] | b[X]))
	{
	    out = 1;
	    if (!(l[Y] | r[Y] | t[Y] | b[Y]))
	    {
		c = 0;
		if (!((pts[0][X] == v0[X]) && (pts[0][Y] == v0[Y]) ||
		      (pts[0][X] == v1[X]) && (pts[0][Y] == v1[Y])))
		{
		    for(i=0;i<dim;i++)
		      clip[c++] = pts[0][i];
		}
		if (!((pts[1][X] == v0[X]) && (pts[1][Y] == v0[Y]) ||
		      (pts[1][X] == v1[X]) && (pts[1][Y] == v1[Y])))
		{
		    for(i=0;i<dim;i++)
		      clip[c++] = pts[1][i];
		}
		return c / dim;
	    }
	}

	x0 = pts[0][X]; y0 = pts[0][Y];
	x1 = pts[1][X]; y1 = pts[1][Y];

	if (l[out])
	  d = (bounds->min_x - x0) / (x1 - x0);
	else if (r[out])
	  d = (bounds->max_x - x0) / (x1 - x0);
	else if (b[out])
	  d = (bounds->min_y - y0) / (y1 - y0);
	else if (t[out])
	  d = (bounds->max_y - y0) / (y1 - y0);
	else
	  d = 0.0;

	pts[out][X] = x0 + (x1 - x0) * d;
	pts[out][Y] = y0 + (y1 - y0) * d;

	/* Because of rounding error, the above two lines are not
	 * guaranteed to clip as they are supposed to.  Make sure the
	 * points do get clipped.
	 */
	if (l[out])
	  pts[out][X] = bounds->min_x;
	else if (r[out])
	  pts[out][X] = bounds->max_x;
	else if (b[out])
	  pts[out][Y] = bounds->min_y;
	else if (t[out])
	  pts[out][Y] = bounds->max_y;

	for(i=2;i<dim;i++)
	{
	    z0 = pts[X][i];
	    z1 = pts[Y][i];
	    pts[out][i] = z0 + (z1 - z0) * d;
	}
    }
    /* return 1 */
    /*NOTREACHED*/
}


static int32 
tactmap_clip_line(float64 *v0, float64 *v1, TACTMAP_BOUNDS *bounds,
		  float64 *clip, int32 dim)
{
    int32 c=0;
    int32 i;

    /* Copy point 0 if it is OK */
    if ((v0[X] >= bounds->min_x) && (v0[X] <= bounds->max_x) &&
	(v0[Y] >= bounds->min_y) && (v0[Y] <= bounds->max_y))
    {
	for (i=0;i<dim;i++)
	  clip[c++] = v0[i];
    }

    /* Find clipping points */
    c += dim * tactmap_clip_intersections(v0,v1,bounds,clip+c,dim);

    /* Copy point 1 if it is OK */
    if ((v1[X] >= bounds->min_x) && (v1[X] <= bounds->max_x) &&
	(v1[Y] >= bounds->min_y) && (v1[Y] <= bounds->max_y))
    {
	for (i=0;i<dim;i++)
	  clip[c++] = v1[i];
    }

    /* Never return less than two points */
    if (c > dim)
      return c/dim;
    else
      return 0;
}

static int32
generate_cell(TACTMAP_PTR tactmap, TACTMAP_BOUNDS *bounds, int32 max_segments,
	      int32 max_labels, uint32 label_mask, int32 *n_segments,
	      float32 *segments, int32 *n_labels, COORD_GRID_LABEL *labels)
{
    static int32 n_cells=0;
    static int32 *cell_list=NULL;
    int32 nts,i,lbl_cnt;
    float64 z,sw_x,sw_y,se_x,se_y;
    float64 nw_x,nw_y,ne_x,ne_y;
    float64 mn_lat,mx_lat,mn_lng,mx_lng;
    float64 pt[2][XY];
    float64 clip[2][XY];
    TACTMAP_BOUNDS lim;

    lbl_cnt = nts =  *n_labels = *n_segments = 0;
    max_labels--;
    
    lim.min_x = NS_MAX(bounds->min_x,tactmap->playbox.min_x);
    lim.max_x = NS_MIN(bounds->max_x,tactmap->playbox.max_x);
    lim.min_y = NS_MAX(bounds->min_y,tactmap->playbox.min_y);
    lim.max_y = NS_MIN(bounds->max_y,tactmap->playbox.max_y);
    

    while ( !wld_lookup_playbox_cells(&n_cells,cell_list) )
    {
	if ( !cell_list )
	  cell_list = (int32 *)STDALLOC(sizeof(cell_list) * n_cells);
	else 
	  cell_list = (int32 *)STDREALLOC(cell_list,n_cells*sizeof(cell_list));
    }
    
    for (i=0;i<n_cells;i++)
    {
	int32 cell = cell_list[i];
	
	/* $$$ can still make faster, currently computing points twice */
	if ( ((nts>>2) >= max_segments) ||
	    !gcs_lookup_cell_extents(cell,&mn_lat,&mx_lat,
				    &mn_lng,&mx_lng)  || 
	    /* SW CORNER */
	    coord_convert(COORD_LATLON,mn_lat,mn_lng,z=0,0,
			  COORD_GCSCS,tactmap->cur->cc,
			  &sw_x,&sw_y,&z) ||
	    /* NE CORNER */
	    coord_convert(COORD_LATLON,mx_lat,mx_lng,z=0,0,
			  COORD_GCSCS,tactmap->cur->cc,
			  &ne_x,&ne_y,&z) ||
	    /* SE CORNER */
	    coord_convert(COORD_LATLON,mn_lat,mx_lng,z=0,0,
			  COORD_GCSCS,tactmap->cur->cc,
			  &se_x,&se_y,&z) ||
	    /* NW_CORNER */
	    coord_convert(COORD_LATLON,mx_lat,mn_lng,z=0,0,
			  COORD_GCSCS,tactmap->cur->cc,
			  &nw_x,&nw_y,&z))
	{
	    n_segments = 0 ;
	    n_labels   = 0 ;
	    return FALSE   ;
	}
	
	if (( nw_x >= lim.min_x) || ( sw_x >= lim.min_x)) 
	{
	    /* WESTERN CELL BOUNDRY */
	    pt[0][X] = nw_x;
	    pt[0][Y] = nw_y;
	    pt[1][X] = sw_x;
	    pt[1][Y] = sw_y;
	    /*CLIP(nw_x,nw_y,sw_x,sw_y);*/
	    if (tactmap_clip_line(&pt[0][X],&pt[1][X],&lim,&clip[0][0],
				  (int32)XY))
	    {
		segments[nts++] = clip[0][X];
		segments[nts++] = clip[0][Y];
		segments[nts++] = clip[1][X];
		segments[nts++] = clip[1][Y];
		
		labels[lbl_cnt].x = (clip[0][X]+clip[1][X])*0.5 + 	      
		  METERS(20,tactmap->cur->ppm);
		labels[lbl_cnt].y = (clip[0][Y]+clip[1][Y])*0.5 ;
		sprintf(labels[lbl_cnt].txt,"%d",cell);
		if ( lbl_cnt < max_labels ) lbl_cnt++;
	    }
	}
	
	if (( ne_x <= lim.max_x) || ( se_x <= lim.max_x)) 
	{
	    /* EASTERN CELL BOUNDRY */
	    pt[0][X] = ne_x;
	    pt[0][Y] = ne_y;
	    pt[1][X] = se_x;
	    pt[1][Y] = se_y;

	    if (tactmap_clip_line(&pt[0][X], &pt[1][X], &lim, &clip[0][X], 
				  (int32)XY))
	    {
		segments[nts++] = clip[0][X];
		segments[nts++] = clip[0][Y];
		segments[nts++] = clip[1][X];
		segments[nts++] = clip[1][Y];
		
		labels[lbl_cnt].x = (clip[0][X]+clip[1][X])*0.5 - 
		  METERS(70,tactmap->cur->ppm);
		labels[lbl_cnt].y = (clip[0][Y]+clip[1][Y])*0.5;
		sprintf(labels[lbl_cnt].txt,"%d",cell);
		if ( lbl_cnt < max_labels ) lbl_cnt++;
	    }
	}
	
	if (( se_y >= lim.min_y) || ( sw_y >= lim.min_y)) 
	{
	    /* SOUTHERN CELL BOUNDRY */
	    pt[0][X] = sw_x;
	    pt[0][Y] = sw_y;
	    pt[1][X] = se_x;
	    pt[1][Y] = se_y;

	    if (tactmap_clip_line(&pt[0][X], &pt[1][X], &lim, &clip[0][X], 
				  (int32)XY))
	    {
		segments[nts++] = clip[0][X];
		segments[nts++] = clip[0][Y];
		segments[nts++] = clip[1][X];
		segments[nts++] = clip[1][Y];
		
		labels[lbl_cnt].x = (clip[0][X]+clip[1][X])*0.5;
		labels[lbl_cnt].y = (clip[0][Y]+clip[1][Y])*0.5 + 
		  METERS(10,tactmap->cur->ppm);
		sprintf(labels[lbl_cnt].txt,"%d",cell);
		if ( lbl_cnt < max_labels ) lbl_cnt++;
	    }	    
	}
	
	if (( ne_y <= lim.max_y) || ( nw_y <= lim.max_y)) 
	{
	    /* NORTHERN CELL BOUNDRY */
	    pt[0][X] = nw_x;
	    pt[0][Y] = nw_y;
	    pt[1][X] = ne_x;
	    pt[1][Y] = ne_y;

	    if (tactmap_clip_line(&pt[0][X], &pt[1][X], &lim, &clip[0][0],
				  (int32)XY))
	    {
		segments[nts++] = clip[0][X];
		segments[nts++] = clip[0][Y];
		segments[nts++] = clip[1][X];
		segments[nts++] = clip[1][Y];
		
		labels[lbl_cnt].x = (clip[0][X]+clip[1][X])*0.5;
		labels[lbl_cnt].y = (clip[0][Y]+clip[1][Y])*0.5 - 
		  METERS(20,tactmap->cur->ppm);
		sprintf(labels[lbl_cnt].txt,"%d",cell);
		if ( lbl_cnt < max_labels ) lbl_cnt++;
	    }	    
	}
    }	

    *n_segments = nts/4;
    *n_labels   = lbl_cnt;
    return TRUE;
}

static void
render_compass(TACTMAP_PTR tactmap, GC gc)
{
    int16 x,y;
        
    x = tactmap->pix_width - tmap_north_width - 25, y = 50;

    XSetTSOrigin(XtDisplay(tactmap->widget), tactmap->gc.compass, x, y);

    XFillRectangle(XtDisplay(tactmap->widget), tactmap->terrain_pix, gc,
		   x, y, tmap_north_width, tmap_north_height);
}
    

void 
tactmap_draw_grids(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		   int32 use_erase_gc, TACTMAP_BOUNDS *bounds)
{
    int32            digits;
    int32            n_thin_segments;
    float32          thin_segments[MAX_SEGS*4];
    int32            n_thick_segments;
    float32          thick_segments[MAX_SEGS*4];
    int32            n_labels;
    COORD_GRID_LABEL labels[MAX_LBLS];
    TACTMAP_BOUNDS   clipped_bounds;
    GC               draw_gc;

    /* how many digits to print the UTM as */
    if (object->dfn.grids.interval >= 10000)
      digits = 1;
    else if (object->dfn.grids.interval >= 1000)
      digits = 2;
    else if (object->dfn.grids.interval >= 100)
      digits = 3;
    else if (object->dfn.grids.interval >= 10)
      digits = 4;
    else
      digits = 5;

    clipped_bounds       = *bounds;
    clipped_bounds.min_x = NS_MAX(bounds->min_x,tactmap->playbox.min_x);
    clipped_bounds.max_x = NS_MIN(bounds->max_x,tactmap->playbox.max_x);
    clipped_bounds.min_y = NS_MAX(bounds->min_y,tactmap->playbox.min_y);
    clipped_bounds.max_y = NS_MIN(bounds->max_y,tactmap->playbox.max_y);
    
    coord_generate_grid(tactmap->cur->cc,
			clipped_bounds.min_x, clipped_bounds.max_x,
			clipped_bounds.min_y, clipped_bounds.max_y,
			object->dfn.grids.interval, digits,
			object->dfn.grids.label_mask,
			MAX_SEGS, MAX_SEGS, MAX_LBLS,
			&n_thin_segments, thin_segments,
			&n_thick_segments, thick_segments,
			&n_labels, labels);

    draw_gc = use_erase_gc ? tactmap->gc.background : tactmap->gc.grids;
    
    render_grid(tactmap, (int32)TRUE, tactmap->terrain_pix, draw_gc,
		n_thin_segments, thin_segments,
		n_thick_segments, thick_segments,
		n_labels, labels);

    draw_gc = use_erase_gc ? tactmap->gc.background : tactmap->gc.compass;

    render_compass(tactmap,draw_gc);

    display_scale(tactmap, tactmap->gc.grids);
}




void 
tactmap_draw_cells(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		   int32 use_erase_gc, TACTMAP_BOUNDS *bounds)
{
    int32            n_segments;
    float32          segments[MAX_SEGS*4];
    int32            n_labels;
    COORD_GRID_LABEL labels[MAX_LBLS];
    GC               draw_gc;

    /* draw the cell boundaries */
    if ( generate_cell(tactmap,bounds,(int32)MAX_SEGS,(int32)MAX_LBLS,
		       object->dfn.grids.label_mask, &n_segments, segments,
		       &n_labels, labels) )
    {
	draw_gc = use_erase_gc ? tactmap->gc.background : tactmap->gc.cells;
	
	render_cell(tactmap, (int32)TRUE, tactmap->terrain_pix, draw_gc,
		    n_segments, segments, n_labels, labels);
    }

    draw_gc = use_erase_gc ? tactmap->gc.background : tactmap->gc.compass;

    render_compass(tactmap,draw_gc);
    /* don't render the scale */
}

static void 
render_grid(TACTMAP_PTR tactmap, int32 erase_under_numbers, Pixmap pixmap,
	    GC gc, int32 n_thin_segments, float32 *thin_segments, 
	    int32 n_thick_segments, float32 *thick_segments, int32 n_labels,
	    COORD_GRID_LABEL *labels)
{
    DECLARE_REGS;
    int          (*string_fcn)();
    int16          pts[MAX_SEGS][2][XY];
    int32          i,j;
    XGCValues      gcv;
    int32          digits;

    if (erase_under_numbers)
      string_fcn = XDrawImageString;
    else
      string_fcn = XDrawString;
#ifdef DEBUG
    printf("\n\n**** new zoom ***\n");
#endif   
    /* Change the GC to have different line width for thick lines */
    if ( n_thick_segments )
    {
	for (j=0,i=0;j<n_thick_segments;j++)
	{
	    pts[j][0][X] = tactmap_FX_pixel_number(thick_segments[i++],cx,
						   hsw,ppm);
	    pts[j][0][Y] = tactmap_FY_pixel_number(thick_segments[i++],cy,
						   hsh,ppm);
	    pts[j][1][X] = tactmap_FX_pixel_number(thick_segments[i++],cx,
						   hsw,ppm);
	    pts[j][1][Y] = tactmap_FY_pixel_number(thick_segments[i++],cy,
						   hsh,ppm);
#ifdef DEBUG
	    printf("thick seg %d (%f,%f)->(%f,%f)  screen (%d,%d),(%d,%d)\n",
		   j,thick_segments[i-4],thick_segments[i-3],
		   thick_segments[i-2],thick_segments[i-1],
		   pts[j][0][X],pts[j][0][Y],pts[j][1][X],pts[j][1][Y]);
#endif
	    
	}
	
	/* Change the line width for the thick segments */
	gcv.line_width = 2;
	
	XChangeGC(XtDisplay(tactmap->widget), gc, GCLineWidth, &gcv);
	
	XDrawSegments(XtDisplay(tactmap->widget), pixmap, gc, 
		      (XSegment *)pts, n_thick_segments);
    }
    
    if ( n_thin_segments )
    {
	for (j=0,i=0;j<n_thin_segments;j++)
	{
	    pts[j][0][X] = tactmap_FX_pixel_number(thin_segments[i++],cx,
						   hsw,ppm);
	    pts[j][0][Y] = tactmap_FY_pixel_number(thin_segments[i++],cy,
						   hsh,ppm);
	    pts[j][1][X] = tactmap_FX_pixel_number(thin_segments[i++],cx,
						   hsw,ppm);
	    pts[j][1][Y] = tactmap_FY_pixel_number(thin_segments[i++],cy,
						   hsh,ppm);
#ifdef DEBUG
	    printf("thin seg %d (%f,%f)->(%f,%f)  screen (%d,%d),(%d,%d)\n",
		   j,thin_segments[i-4],thin_segments[i-3],
		   thin_segments[i-2],thin_segments[i-1],
		   pts[j][0][X],pts[j][0][Y],pts[j][1][X],pts[j][1][Y]);
#endif    
	}
	
	/* Change the line width for the thin segments */
	gcv.line_width = 0;

	XChangeGC(XtDisplay(tactmap->widget), gc, GCLineWidth, &gcv);
	
	XDrawSegments(XtDisplay(tactmap->widget), pixmap,
		      gc, (XSegment *)pts, n_thin_segments);
    }
    
    for (i=0;i<n_labels;i++)
    {
	digits = strlen(labels[i].txt);
	string_fcn(XtDisplay(tactmap->widget), pixmap, gc,
		   X_PIXEL_ON_SCREEN(labels[i].x,cx,hsw,ppm,
				     digits*CHARACTER_WIDTH),
		   Y_PIXEL_ON_SCREEN(labels[i].y,cy,hsh,ppm,
				     CHARACTER_WIDTH),
		   labels[i].txt, digits);
    }
}

static void 
render_cell(TACTMAP_PTR tactmap, int32 erase_under_numbers, Pixmap pixmap,
	    GC gc, int32 n_segments, float32 *segments, int32 n_labels,
	    COORD_GRID_LABEL *labels)
{
    DECLARE_REGS;
    int          (*string_fcn)();
    int16        pts[MAX_SEGS][2][XY];
    int32        i,j;
    int32        digits;

    if (erase_under_numbers)
      string_fcn = XDrawImageString;
    else
      string_fcn = XDrawString;

    for (j=0,i=0;j<n_segments;j++)
    {
	pts[j][0][X] = tactmap_FX_pixel_number(segments[i++],cx,hsw,ppm);
	pts[j][0][Y] = tactmap_FY_pixel_number(segments[i++],cy,hsh,ppm);
	pts[j][1][X] = tactmap_FX_pixel_number(segments[i++],cx,hsw,ppm);
	pts[j][1][Y] = tactmap_FY_pixel_number(segments[i++],cy,hsh,ppm);
    }

    XDrawSegments(XtDisplay(tactmap->widget), pixmap,
		  gc, (XSegment *)pts, n_segments);

    for (i=0;i<n_labels;i++)
    {
	digits = strlen(labels[i].txt);
	string_fcn(XtDisplay(tactmap->widget), pixmap, gc,
		   X_PIXEL_ON_SCREEN(labels[i].x,cx,hsw,ppm,
				     digits*CHARACTER_WIDTH),
		   Y_PIXEL_ON_SCREEN(labels[i].y,cy,hsh,ppm,
				     CHARACTER_WIDTH),
		   labels[i].txt, digits);
    }
}

#endif /*USE_X*/
