#ifndef lint
static char rcsid [] = "$RCSfile: tmap_hypso.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_hypso.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to do hypsometric tinting                            *
*       Created: Fri Nov  8 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <libctdb.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define MAX_SCREEN_WIDTH 2048
#define MAX_SCREEN_HEIGHT 2048

static void interpolate_color(TACTMAP_PTR tactmap, float64 elev,
			      XColor *color);

/*ARGSUSED*/
void tactmap_draw_hypso(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
			int32 use_erase_gc, TACTMAP_BOUNDS *bounds)
{
    DECLARE_REGS;
    XImage *image;
    char img[MAX_SCREEN_WIDTH * MAX_SCREEN_HEIGHT];
    int32 low_x, low_y, width, height;
    float64 min_z, max_z;
    int32 i;
    XColor local_colors[HYPSO_CMAP_SIZE];
    float64 step;

    low_x = tactmap_X_pixel_number(bounds->min_x,cx,hsw,ppm);
    low_y = tactmap_Y_pixel_number(bounds->max_y,cy,hsh,ppm);
    width = tactmap_X_pixel_number(bounds->max_x,cx,hsw,ppm) - low_x;
    height = tactmap_Y_pixel_number(bounds->min_y,cy,hsh,ppm) - low_y;

    if (object->dfn.hypso.kind == TACTMAP_DITHER_HYPSO)
    {
	image = XCreateImage(XtDisplay(tactmap->widget), NULL,
			     1, XYBitmap, 0, img,
			     width, height, 8, (width+7)/8);

	ctdb_hypso_bitmap(tactmap->ctdb,
			  bounds->min_x, bounds->min_y, width, height,
			  1.0/ppm,
			  tactmap->ctdb->min_z, tactmap->ctdb->max_z,
			  img);
    }
    else
    {
	image = XCreateImage(XtDisplay(tactmap->widget),
			     DefaultVisualOfScreen(XtScreen(tactmap->widget)),
			     tactmap->depth,ZPixmap, 0, img,
			     width, height, 8, width);

	ctdb_hypso_image(tactmap->ctdb,
			 bounds->min_x, bounds->min_y, width, height,
			 1.0/ppm,
			 HYPSO_CMAP_SIZE, 
			 (unsigned long *)((ADDRESS)tactmap->hypso_pixels),
			 BlackPixelOfScreen(XtScreen(tactmap->widget)),
			 img, &min_z, &max_z);

	step = (max_z-min_z)/HYPSO_CMAP_SIZE;
	for(i=0;i<HYPSO_CMAP_SIZE;i++)
	{
	    interpolate_color(tactmap, min_z + i*step, &local_colors[i]);
	    local_colors[i].pixel = tactmap->hypso_pixels[i];
	}
	XStoreColors(XtDisplay(tactmap->widget),
		     DefaultColormapOfScreen(XtScreen(tactmap->widget)),
		     local_colors, HYPSO_CMAP_SIZE);
    }

    XPutImage(XtDisplay(tactmap->widget), tactmap->terrain_pix,
	      tactmap->gc.hypso,
	      image,0,0,low_x,low_y,width,height);

    image->data = NULL;
    XDestroyImage(image);
}

static void interpolate_color(TACTMAP_PTR tactmap, float64 elev, XColor *color)
{
    int32 i;
    float64 delta;

    for (i=tactmap->n_hypso_colors-1;i>=0;i--)
      if (elev > tactmap->hypso_colors[i].begin_alt)
	break;
    if (i==tactmap->n_hypso_colors-1)
      delta = 1.0;
    else if (tactmap->hypso_colors[i].begin_alt > elev)
      delta = 0.0;
    else
      delta = ((elev - tactmap->hypso_colors[i].begin_alt) /
	       (float64)(tactmap->hypso_colors[i+1].begin_alt -
			 tactmap->hypso_colors[i].begin_alt));

    color->red = tactmap->hypso_colors[i].dark_color.red + delta *
      (tactmap->hypso_colors[i].bright_color.red -
       tactmap->hypso_colors[i].dark_color.red);
    color->green = tactmap->hypso_colors[i].dark_color.green + delta *
      (tactmap->hypso_colors[i].bright_color.green -
       tactmap->hypso_colors[i].dark_color.green);
    color->blue = tactmap->hypso_colors[i].dark_color.blue + delta *
      (tactmap->hypso_colors[i].bright_color.blue -
       tactmap->hypso_colors[i].dark_color.blue);

    color->flags = DoRed | DoGreen | DoBlue;
}


#endif /*USE_X*/
