#ifndef lint
static char rcsid [] = "$RCSfile: tmap_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_init.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize/create tactical maps                   *
*       Created: Wed Sep 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include "tmap_trees.h"
#include "tmap_north.h"
#include <libgcs.h>
#include <libworld.h>
#include <X11/bitmaps/cross_weave>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include <stdext.h> /*common/include/global*/
#include <stdalloc.h> /*common/include/global*/

#include <stdstring.h> /*common/include/global*/


static void make_gcs(TACTMAP_PTR tactmap);
static void parse_hypso(TACTMAP_PTR tactmap, char *map_string);
static TACTMAP_PTR tactmaps[1024];
static int32 tactmap_count = 0;

void tactmap_init(void)
{
    /* No-Op */
}

static void init_intervis_pixmaps(TACTMAP_PTR tactmap)
{

    int i;

    DITHER_MATRIX(mat, PERCENT_SIZE);
    DITHER_BITMAPS(bits, PERCENT_SIZE);

    /* Make the generating matrix */
    dither_matrix(PERCENT_SIZE, mat);

    /* Generate dither bitmaps */
    dither_bitmaps(PERCENT_SIZE, mat, bits);

    for(i=0;i<N_PERCENT;i++)
      tactmap->intervis_pix[i] =
	XCreateBitmapFromData(XtDisplay(tactmap->widget), 
			      RootWindowOfScreen(XtScreen(tactmap->widget)),
			      (char *)bits[N_PERCENT-i-1], 
			      PERCENT_SIZE, 
			      PERCENT_SIZE);
    
}
    

TACTMAP_PTR 
tactmap_create(Widget widget, SNSTVE_WINDOW_PTR sensitive, 
	       GC sensitive_erase_gc, TACTMAP_REFRESH_CALLBACK refresh)
{
    float64 pbcenter[XYC];

    TACTMAP_PTR result = (TACTMAP_PTR)STDALLOC(sizeof(struct tactmap));

    bzero((char *)result, sizeof(struct tactmap));

    result->widget           = widget;
    result->sensitive        = sensitive;
    result->refresh_callback = refresh;

    /* Force the setting of the center */
    result->cur         = &result->vp;    
    result->cur->cc     = GCS_ILLEGAL_CELL-1;

    wld_get_playbox_center(pbcenter);

    tactmap_set_center(result,(int32)pbcenter[X],(int32)pbcenter[Y],
		       (int32)pbcenter[CELL2D]);

    result->cur->ppm = 1.0;

    tactmaps[tactmap_count++] = result;

    init_intervis_pixmaps(result);

    tactmap_init_clouds_pixmaps(result);

    result->background_gc = sensitive_erase_gc;
    make_gcs(result);

    /* Create parameters for Environment-specific user preferences */
    result->eui_prefs = eui_create_prefs();

    XtAddEventHandler(widget, StructureNotifyMask, FALSE,
		      (XtEventHandler)tactmap_resize, (XtPointer)result);

    return result;
}

TACTMAP_PTR tactmap_find_sensitive(SNSTVE_WINDOW_PTR sensitive)
{
    int32 i;

    for (i=0;i<tactmap_count;i++)
      if (tactmaps[i]->sensitive == sensitive)
	return tactmaps[i];
    return NULL;
}

Pixel tactmap_get_color(TACTMAP_PTR tactmap, TACTMAP_OBJECT_CLASS sensClass)
{
    if ((sensClass >= TACTMAP_FIRST_CLASS) &&
	(sensClass <= TACTMAP_LAST_CLASS))
      return tactmap->colors[sensClass];
    else
      return tactmap->bg_color;
}

/*ARGSUSED*/
void tactmap_resize(Widget w, TACTMAP_PTR tactmap, XEvent *event)
{
    XWindowAttributes att;
    int32 from_x, from_y, to_x, to_y, width, height;

    if (event->type == ConfigureNotify)
    {
	/* Get the new size */
	XGetWindowAttributes(DPYWIN, &att);

	/* Make sure they are really new */
	if ((att.height == tactmap->pix_height) &&
	    (att.width == tactmap->pix_width))
	  return;

	/* Copy the background to the foreground to reduce visual
	 * unpleasantness related to background pixmap tiling.
	 */
	if (tactmap->background_pix)
	{
	    if (att.width > tactmap->pix_width)
	    {
		from_x = 0;
		to_x = (att.width - tactmap->pix_width)/2;
		width = tactmap->pix_width;
	    }
	    else
	    {
		from_x = (tactmap->pix_width - att.width)/2;
		to_x = 0;
		width = att.width;
	    }
	    if (att.height > tactmap->pix_height)
	    {
		from_y = 0;
		to_y = (1 + att.height - tactmap->pix_height)/2;
		height = tactmap->pix_height;
	    }
	    else
	    {
		from_y = (1 + tactmap->pix_height - att.height)/2;
		to_y = 0;
		height = att.height;
	    }
	    XFillRectangle(DPYWIN, tactmap->gc.background,
			   0, 0, att.width, att.height);
	    XCopyArea(XtDisplay(tactmap->widget), tactmap->background_pix,
		      XtWindow(tactmap->widget), tactmap->background_gc,
		      from_x, from_y, width, height, to_x, to_y);
	}

	/* Store the new values */
	tactmap->pix_height = att.height;
	tactmap->pix_width = att.width;
	tactmap->depth = att.depth;
	tactmap->cur->hsw = att.width/tactmap->cur->ppm/2;
	tactmap->cur->hsh = att.height/tactmap->cur->ppm/2;

	/* Clear the existing Pixmaps, so we know we have to generate
	 * new ones on our next tick.
	 */
	if (tactmap->terrain_pix)
	{
	    XFreePixmap(XtDisplay(tactmap->widget), tactmap->terrain_pix);
	    tactmap->terrain_pix = 0;
	}
	if (tactmap->background_pix)
	{
	    XFreePixmap(XtDisplay(tactmap->widget), tactmap->background_pix);
	    tactmap->background_pix = 0;
	}

	tactmap->do_refresh = 1;
    }
}    

static void make_gcs(TACTMAP_PTR tactmap)
{
    XGCValues gcv;
    Window root;
    struct tactmap_resources
    {
	Pixel background;
	Pixel hypso;
	Pixel water;
	Pixel fordable;
	Pixel no_go;
	Pixel slow_go;
	Pixel road;
#ifdef AE_FEATURES
	Pixel offroad;
#endif
	Pixel tree;
	Pixel building;
        Pixel contaminant;
#ifdef AE_FEATURES
	Pixel tact_sign;
#endif
	Pixel pipeline;
	Pixel political;
	Pixel railroad;
	Pixel powerline;
	Pixel contour;
	Pixel town;
	Pixel grid;
	Pixel cell;
	Pixel crater;
	Pixel ditch;
	Pixel fpos;
	Pixel wire;
	Pixel rock_drop;
	Pixel dragons_tooth;
	Pixel berm;
	Pixel rubble;
	Pixel snow;
        Pixel clouds;
	Pixel miclic_line;
	Pixel miclic_scorch;
	Pixel camouflage;
	Pixel hydro_surface;
	Pixel hydro_subsurface;
	Pixel plug;
	Pixel compass;
	Font town_font;
	Font grid_font;
	String hypso_map;
	Pixel intervis_fg;
	Pixel intervis_bg;
    } resources;
    static XtResource res_list[] = {
	{ "background" , "Background" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, background) ,
	    XtRString , "Gray75" } ,
	{ "hypsoDither" , "HypsoDither" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, hypso) ,
	    XtRString , "Black" } ,
	{ "water" , "Water" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, water) ,
	    XtRString , "steelBlue" } ,
	{ "fordable" , "Fordable" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, fordable) ,
	    XtRString , "SkyBlue" } ,
	{ "noGo" , "NoGo" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, no_go) ,
	    XtRString , "FireBrick" } ,
	{ "slowGo" , "slowGo" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, slow_go) ,
	    XtRString , "Gold3" } ,
	{ "road" , "Road" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, road) ,
	    XtRString , "Red1" } ,
	{ "contaminant" , "Contaminant" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, contaminant) ,
	    XtRString , "PaleVioletRed3" } ,
#ifdef AE_FEATURES
	{ "offroad" , "OffRoad" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, offroad) ,
	    XtRString , "White" } ,
#endif
	{ "tree" , "Tree" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, tree) ,
	    XtRString , "DarkGreen" } ,
	{ "building" , "Building" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, building) ,
	    XtRString , "Gray25" } ,
#ifdef AE_FEATURES
	{ "tact_sign" , "TactSign" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, tact_sign) ,
	    XtRString , "White" } ,
#endif
	{ "pipeline" , "Pipeline" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, pipeline) ,
	    XtRString , "Cyan4" } ,
	{ "political" , "Political" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, political) ,
	    XtRString , "Orange3" } ,
	{ "railroad" , "Railroad" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, railroad) ,
	    XtRString , "Black" } ,
	{ "powerline" , "Powerline" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, powerline) ,
	    XtRString , "Gray25" } ,
	{ "contour" , "Contour" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, contour) ,
	    XtRString , "Sienna4" } ,
	{ "town" , "Town" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, town) ,
	    XtRString , "Black" } ,
	{ "grid" , "Grid" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, grid) ,
	    XtRString , "Black" } ,
	{ "cell" , "Cell" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, cell) ,
	    XtRString , "Yellow" } ,
	{ "crater" , "Crater" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, crater) ,
	    XtRString , "Sienna4" } ,
	{ "ditch" , "Ditch" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, ditch) ,
	    XtRString , "tomato" } ,
	{ "fpos" , "FPos" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, fpos) ,
	    XtRString , "brown" } ,
	{ "wire" , "Wire" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, wire) ,
	    XtRString , "Black" } ,
	{ "rock_drop" , "RockDrop" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, rock_drop) ,
	    XtRString , "Orange3" } ,
	{ "dragons_tooth" , "DragonsTooth" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, dragons_tooth) ,
	    XtRString , "red" } ,
	{ "berm" , "Berm" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, berm) ,
	    XtRString , "Cyan4" } ,
	{ "rubble" , "Rubble" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, rubble) ,
	    XtRString , "black" } ,
	{ "snow" , "Snow" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, snow) ,
	    XtRString , "White" } ,
	{ "clouds" , "Clouds" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, clouds) ,
	    XtRString , "Gray50" } ,
	{ "miclic_line" , "MiclicLine" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, miclic_line) ,
	    XtRString , "White" } ,
	{ "miclic_scorch" , "MiclicScorch" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, miclic_scorch) ,
	    XtRString , "Sienna4" } ,
	{ "camouflage" , "Camouflage" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, camouflage) ,
	    XtRString , "DimGray" } ,
	{ "hydro_surface" , "HydroSurface" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, hydro_surface) ,
	    XtRString , "LightSlateBlue" } ,
	{ "hydro_subsurface" , "HydroSubsurface" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, hydro_subsurface) ,
	    XtRString , "MediumBlue" } ,
	{ "plug" , "Plug" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, plug) ,
	    XtRString , "lavender" } ,
	{ "compass" , "Compass" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, compass) ,
	    XtRString , "#710006" } ,
	{ "townFont" , "TownFont" , XtRFont , sizeof(Font) ,
	    XtOffset(struct tactmap_resources *, town_font) ,
	    XtRString , XtDefaultFont } ,
	{ "gridFont" , "GridFont" , XtRFont , sizeof(Font) ,
	    XtOffset(struct tactmap_resources *, grid_font) ,
	    XtRString , XtDefaultFont } ,
	{ "hypsoMap" , "HypsoMap" , XtRString , sizeof(String) ,
	    XtOffset(struct tactmap_resources *, hypso_map) ,
	    XtRString , "0 #b6c05c #ccd767 200 #ecec84 #f9f98b 500 #d4b068 #efc575 1000 #b87979 #d58c8c 2000 #934548 #b35558 3000 #a99197 #c0a5ab 4000 #bbacc8 #d8c7e7 5000 #cdabb2 #e3bdc5 6000 Gray80 White" } ,
	{ "intervis_fg" , "Intervis_fg" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, intervis_fg) ,
	    XtRString , "Green" } ,
	{ "intervis_bg" , "Intervis_bg" , XtRPixel , sizeof(Pixel) ,
	    XtOffset(struct tactmap_resources *, intervis_bg) ,
	    XtRString , "Black" } ,
    };

    /* Get the colors, and various other resources used */
    XtGetSubresources(tactmap->widget, &resources,
		      "tacticalmap", "TacticalMap",
		      res_list, XtNumber(res_list), NULL, 0);

    bzero((char *)tactmap->colors, sizeof(tactmap->colors));

    /* Get the root window (the tactmap widget may not yet be realized) */
    root = RootWindowOfScreen(XtScreen(tactmap->widget));
#undef DPYWIN
#define DPYWIN XtDisplay(tactmap->widget), root

    /* Make the hourglass cursor */
    tactmap->hourglass = XCreateFontCursor(XtDisplay(tactmap->widget),
					   XC_watch);

    /* Make the GCs */
    gcv.line_width = 0;
    gcv.foreground = resources.background;
    tactmap->bg_color = gcv.background = resources.background;
    tactmap->gc.background =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_HYPSO] = gcv.foreground = resources.hypso;
    tactmap->gc.hypso =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_WATER] = gcv.foreground = resources.water;
    tactmap->gc.water =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    gcv.foreground = resources.fordable;
    tactmap->gc.fordable =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    gcv.foreground = resources.no_go;
    gcv.line_width = 3;
    tactmap->gc.no_go =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    gcv.foreground = resources.slow_go;
    gcv.line_width = 0;
    tactmap->gc.slow_go =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    gcv.foreground = resources.plug;
    tactmap->gc.plug =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_CRATERS] = gcv.foreground = resources.crater;
    tactmap->gc.craters =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_DITCHES] = gcv.foreground = resources.ditch;
    tactmap->gc.ditches =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_FPOS] = gcv.foreground = resources.fpos;
    tactmap->gc.fpos =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_BERMS] = gcv.foreground = resources.berm;
    tactmap->gc.berms =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_SNOW] = gcv.foreground = resources.snow;
    tactmap->gc.snow =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_CLOUDS] = gcv.foreground = resources.clouds;
    tactmap->gc.clouds =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_MICLIC_LINE] = gcv.foreground = 
      resources.miclic_line;
    tactmap->gc.miclic_line =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_MICLIC_SCORCH] = gcv.foreground = 
      resources.miclic_scorch;
    tactmap->gc.miclic_scorch =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_WIRE] = gcv.foreground = resources.wire;
    tactmap->gc.wire =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_ROCK_DROPS]= gcv.foreground = resources.rock_drop;
    tactmap->gc.rock_drops =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_DRAGONS_TEETH]= gcv.foreground =
      resources.dragons_tooth;
    tactmap->gc.dragons_teeth =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_ROADS] = gcv.foreground = resources.road;
    tactmap->gc.roads =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

#ifdef AE_FEATURES
    tactmap->colors[TACTMAP_OFFROADS] = gcv.foreground = resources.offroad;
    tactmap->gc.offroads =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);
#endif

    tactmap->colors[TACTMAP_TREES] = gcv.foreground = resources.tree;
    tactmap->gc.trees =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);
    gcv.stipple =
      XCreateBitmapFromData(DPYWIN, (char *)tactmap_trees_bits,
			    tactmap_trees_width, tactmap_trees_height);
    gcv.fill_style = FillStippled;
    tactmap->gc.canopies =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth |
		GCStipple | GCFillStyle, &gcv);

    tactmap->colors[TACTMAP_RUBBLE]= gcv.foreground = resources.rubble;
    tactmap->gc.rubble =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth |
		GCStipple | GCFillStyle, &gcv);

    tactmap->colors[TACTMAP_CAMOUFLAGE]= gcv.foreground = resources.camouflage;
    tactmap->gc.camouflage =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth |
		GCStipple | GCFillStyle, &gcv);

    tactmap->colors[TACTMAP_HYDRO_SURFACE]= gcv.foreground =
      resources.hydro_surface;
    tactmap->gc.hydro_surface =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth |
		GCStipple | GCFillStyle, &gcv);

    gcv.stipple =
      XCreateBitmapFromData(DPYWIN, (char *)tactmap_hydro_bits,
			    tactmap_trees_width, tactmap_trees_height);
    tactmap->colors[TACTMAP_HYDRO_SUBSURFACE]= gcv.foreground =
      resources.hydro_subsurface;
    tactmap->gc.hydro_subsurface =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth |
		GCStipple | GCFillStyle, &gcv);

    tactmap->colors[TACTMAP_BUILDINGS] = gcv.foreground = resources.building;
    tactmap->gc.buildings =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

#ifdef AE_FEATURES
    tactmap->colors[TACTMAP_TACT_SIGNS] = gcv.foreground = resources.tact_sign;
    tactmap->gc.tact_signs =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);
#endif

    tactmap->colors[TACTMAP_PIPELINES] = gcv.foreground = resources.pipeline;
    tactmap->gc.pipelines =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_CONTAMINANT] = gcv.foreground = resources.contaminant;
    tactmap->gc.contaminant =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_POLITICAL] = gcv.foreground = resources.political;
    tactmap->gc.political =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_RAILROADS] = gcv.foreground = resources.railroad;
    tactmap->gc.railroads =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_POWERLINES]= gcv.foreground = resources.powerline;
    tactmap->gc.powerlines =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_CONTOURS] = gcv.foreground = resources.contour;
    tactmap->gc.contours[0] =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);
    gcv.line_width = 3;
    tactmap->gc.contours[1] =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCLineWidth, &gcv);

    tactmap->colors[TACTMAP_TOWNS] = gcv.foreground = resources.town;
    gcv.font = resources.town_font;
    tactmap->gc.towns =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCFont, &gcv);

    /* Set Up of TACTMAP_GRIDS Color and GC */
    tactmap->colors[TACTMAP_GRIDS] = gcv.foreground = resources.grid;
    gcv.font = resources.grid_font;
    tactmap->gc.grids =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCFont, &gcv);

    /* Set Up of TACTMAP_CELLS Color and GC */
    tactmap->colors[TACTMAP_CELLS] = gcv.foreground = resources.cell;
    gcv.line_width = 4;
    gcv.line_style = LineOnOffDash;
    tactmap->gc.cells =
      XCreateGC(DPYWIN, GCForeground | GCBackground | GCFont | 
		GCLineWidth | GCLineStyle, &gcv);

    gcv.foreground = resources.intervis_fg; 
    gcv.background = resources.intervis_bg;
    gcv.line_width = 4;
    gcv.fill_style = FillOpaqueStippled;
    gcv.font = resources.grid_font;
    tactmap->gc.line_intervis =
      XCreateGC(
		DPYWIN, 
		GCForeground | GCBackground | GCFillStyle | GCLineWidth|GCFont,
		&gcv); 

    gcv.foreground = resources.intervis_bg; 
    gcv.fill_style = FillOpaqueStippled;
    tactmap->gc.area_intervis =
      XCreateGC(DPYWIN, 
		GCForeground | GCFillStyle, &gcv);

    gcv.foreground = resources.building;
    gcv.background = resources.background;
    gcv.fill_style = FillOpaqueStippled;
    gcv.stipple =
      XCreateBitmapFromData(DPYWIN, (char *)cross_weave_bits,
			    cross_weave_width, cross_weave_height);
    tactmap->gc.bridge_picture =
      XCreateGC(DPYWIN,
		GCForeground | GCBackground | GCFillStyle | GCStipple, &gcv);
    tactmap->gc.bridge_symbol = XCreateGC(DPYWIN, GCForeground, &gcv);

    gcv.stipple = XCreateBitmapFromData(DPYWIN, (char *)tmap_north_bits,
					tmap_north_width, tmap_north_height);
    gcv.fill_style = FillStippled;
    gcv.foreground = resources.compass;
    tactmap->gc.compass =
      XCreateGC(DPYWIN, GCForeground | GCFillStyle | GCStipple, &gcv);

    parse_hypso(tactmap, resources.hypso_map);

    gcv.fill_style = FillTiled;
    gcv.font = resources.grid_font;
    tactmap->terrain_gc = XCreateGC(DPYWIN, GCFillStyle | GCFont, &gcv);

    /* The background GC is provide by the caller; set it up so it
     * does what the caller expects.
     */
    XChangeGC(XtDisplay(tactmap->widget), tactmap->background_gc,
	      GCFillStyle, &gcv);
}

static void parse_hypso(TACTMAP_PTR tactmap, char *map_string)
{
    char *s;
    int32 alt;
    unsigned long ignore;
    XColor dark, bright;
    char buf[2048];

    strcpy(buf, map_string);

    tactmap->n_hypso_colors = 0;
    s = strtok(buf, " \t");
    while (s)
    {
	if (!sscanf(s, "%d", &alt) ||
	    !(s = strtok(NULL, " \t")) ||
	    !(XParseColor(XtDisplay(tactmap->widget),
			  DefaultColormapOfScreen(XtScreen(tactmap->widget)),
			  s, &dark)) ||
	    !(s = strtok(NULL, " \t")) ||
	    !(XParseColor(XtDisplay(tactmap->widget),
			  DefaultColormapOfScreen(XtScreen(tactmap->widget)),
			  s, &bright)))
	{
	    printf("LIBTACTMAP WARNING: Error parsing hypsometric map resource\nFormat \"<altitude> <color> <altitude> <color> ...\"\n");
	    break;
	}

	tactmap->hypso_colors[tactmap->n_hypso_colors].begin_alt = alt;
	tactmap->hypso_colors[tactmap->n_hypso_colors].dark_color = dark;
	tactmap->hypso_colors[tactmap->n_hypso_colors].bright_color = bright;
	tactmap->n_hypso_colors += 1;

	s = strtok(NULL, " \t");
    }
    XAllocColorCells(XtDisplay(tactmap->widget),
		     DefaultColormapOfScreen(XtScreen(tactmap->widget)),
		     FALSE, &ignore, 1,
		     tactmap->hypso_pixels, HYPSO_CMAP_SIZE);
}

#else

int tactmap_dummy_archive_variable;

#endif /*USE_X*/
