#ifndef lint
static char rcsid [] = "$RCSfile: tmap_intervis.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_intervis.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw intervisbility lines and area                *
*       Created: Thurs July 8 1992                                          *
*       Author: Carol M. Ladd                                               *
       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#ifdef USE_X
#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <libdither.h>
#include <math.h>
#include <p_po.h>
#include <libenvui.h>
#include <libgcs.h>

static void 
tactmap_draw_intervis_v2v(
			  TACTMAP_PTR tactmap, 
			  TACTMAP_OBJECT_PTR  object,
			  TACTMAP_OBJECT_DFN *dfn, 
			  GC gc
			  );

static ENV_EM_BAND tactmap_em_band = ENV_BAND_47MM;

#define PCB(v) (((v)==0.0) ? 0 : (1 + (int)((v) * (N_PERCENT-2))))

/* This routine draws the tick marks on an intervisibility line at the */
/* current grid interval to make the line look like a ruler */
static void 
draw_ticked_line(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		 TACTMAP_OBJECT_DFN *dfn, GC gc, 
		 int32 ruler_x0, int32 ruler_y0, 
		 int32 ruler_x1, int32 ruler_y1)
{
    float64 dx, dy, dist, cos_theta, sin_theta;
    float64 tick_pixels, xi;
    int32 x0,y0,x1,y1;
    int32 ruler_tick_interval = 1;
    Drawable d;

    ruler_tick_interval = dfn->line_intervis.grid_interval;

    tick_pixels = ruler_tick_interval * tactmap->cur->ppm;

    dx = (ruler_x1 - ruler_x0);
    dy = (ruler_y1 - ruler_y0);
    dist = sqrt(dx*dx + dy*dy);

    if (dist < 0.1)
      return;

    cos_theta = dy / dist;
    sin_theta = dx / dist;

    /* Run down the X axis making ticks, then rotating them to
     * the right place
     */
    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);

    for (xi = 0.0; xi <= dist; xi += tick_pixels)
    {
	x0 = xi * sin_theta +  -TICK_LENGTH * cos_theta;
	y0 = xi * (-cos_theta) +  -TICK_LENGTH * sin_theta;
	x1 = xi * sin_theta + TICK_LENGTH * cos_theta;
	y1 = xi * (-cos_theta) + TICK_LENGTH * sin_theta;

	/* Subtract the Y values because the coordinate system is flopped */
	XDrawLine(XtDisplay(tactmap->widget), d, gc, 
		  ruler_x0+x0, ruler_y0-y0, ruler_x0+x1, ruler_y0-y1);
    }
}

/* This routine erases a previously drawn intervisbility line */
static void 
erase_intervis_line(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		    TACTMAP_OBJECT_DFN *dfn, GC gc)
{
    DECLARE_REGS;
    float64 x0,y0,x1,y1;
    int32 sx1,sy1,sx2,sy2;
    XGCValues gcv;
    char buf[20];
    int32 width, height;
    XFontStruct *fs;
    Drawable d;

    fs = NULL;	/* makes check of whether to free later easy */

    /* Adjust coordinates to be relative to the current center cell.
     */
    gcs_gcs2_to_gcscs(dfn->line_intervis.cell,dfn->line_intervis.point1,
		      cc,dfn->line_intervis.point1);
    gcs_gcs2_to_gcscs(dfn->line_intervis.cell,dfn->line_intervis.point2,
		      cc,dfn->line_intervis.point2);
    dfn->line_intervis.cell=cc;

    x0 = dfn->line_intervis.point1[X];
    y0 = dfn->line_intervis.point1[Y];
    x1 = dfn->line_intervis.point2[X];
    y1 = dfn->line_intervis.point2[Y];

    sx1 = tactmap_FX_pixel_number(x0,cx,hsw,ppm);    
    sy1 = tactmap_FY_pixel_number(y0,cy,hsh,ppm);
    sx2 = tactmap_FX_pixel_number(x1,cx,hsw,ppm);    
    sy2 = tactmap_FY_pixel_number(y1,cy,hsh,ppm);
    /* Make sure we get the percentages also */

    /* Delete the percentages */
    if (dfn->line_intervis.show_percentages)
    {
	int32 strx1,stry1;
	strx1 = sx1 + 9*(sx2-sx1)/10;
	stry1 = sy1 + 9*(sy2-sy1)/10;
	XSetLineAttributes(XtDisplay(tactmap->widget), gc, 20,
			   LineSolid, CapButt, JoinMiter);

	sprintf(buf, "    ");
	XGetGCValues(XtDisplay(tactmap->widget), gc, GCFont, &gcv);

	fs = XQueryFont(XtDisplay(tactmap->widget), gcv.font);
	width = fs->max_bounds.width * 5; /* multiply width by four characters */
	height = fs->max_bounds.ascent + fs->max_bounds.descent; 
	stry1 = stry1 - fs->max_bounds.ascent;
	XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		  tactmap->background_pix, tactmap->gc.background,
		  strx1, stry1, width, height, strx1, stry1);
	XClearArea(XtDisplay(tactmap->widget), XtWindow(tactmap->widget),
		   strx1, stry1, width, height, 0);
    }

    XSetLineAttributes(XtDisplay(tactmap->widget), gc, 7,
		       LineSolid, CapButt, JoinMiter);

    draw_ticked_line(tactmap, object, dfn, gc, sx1,sy1,sx2,sy2);


    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);

    XDrawLine(XtDisplay(tactmap->widget),d,gc, sx1,sy1,sx2,sy2);

    XSetLineAttributes(XtDisplay(tactmap->widget), gc, 3,
		       LineSolid, CapButt, JoinMiter);
    if (fs)
      XFreeFontInfo((char **)NULL, fs, 1);
}

/* This function draw an intervisibility line */
static void 
draw_intervis_line(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
		   TACTMAP_OBJECT_DFN *dfn, GC gc, TACTMAP_BOUNDS *bounds)
{
    DECLARE_REGS;
    float64 ex, ey, z0, z1;
    int32 x,y;
    float64 vis, opacity;
    float64 dx, dy;
    float64 sx0, sy0, sdx, sdy;
    int16 lx,ly,px,py;
    int32 i, n;
    XSegment segs[N_PERCENT][MAX_SAMPLES];
    int32 counts[N_PERCENT];
    int32 this_pix;
    float64 x0,y0,x1,y1;
    Drawable d;
    ENV_INTERFACE	eif;
    int32 md;
    CTDB *ctdb;
    
    /* Adjust coordinates to be relative to the current center cell.
     */
    gcs_gcs2_to_gcscs(dfn->line_intervis.cell,dfn->line_intervis.point1,
		      cc,dfn->line_intervis.point1);
    gcs_gcs2_to_gcscs(dfn->line_intervis.cell,dfn->line_intervis.point2,
		      cc,dfn->line_intervis.point2);
    dfn->line_intervis.cell = cc;

    ctdb = (CTDB *)gcs_get_tdb(cc, FALSE);
    x0 = dfn->line_intervis.point1[X];
    y0 = dfn->line_intervis.point1[Y];
    x1 = dfn->line_intervis.point2[X];
    y1 = dfn->line_intervis.point2[Y];

    if ((x0 < bounds->min_x) ||
	(y0 < bounds->min_y) ||
	(x0 > bounds->max_x) ||
	(y0 > bounds->max_y) ||
	(x1 < bounds->min_x) ||
	(y1 < bounds->min_y) ||
	(x1 > bounds->max_x) ||
	(y1 > bounds->max_y))
      return;

    z0 = ctdb_lookup_elevation(ctdb,x0,y0) + dfn->line_intervis.viewer_agl; 

    x = tactmap_FX_pixel_number(x0,cx,hsw,ppm);    
    y = tactmap_FY_pixel_number(y0,cy,hsh,ppm);
    sx0 = x;
    sy0 = y;

    x = tactmap_FX_pixel_number(x1,cx,hsw,ppm);    
    y = tactmap_FY_pixel_number(y1,cy,hsh,ppm);
    sdx = x - sx0;
    sdy = y - sy0;

    if (sdx >= 0.0)
      n = sdx;
    else
      n = -sdx;

    if ((sdy >= 0.0) && (sdy > n))
      n = sdy;
    else if ((sdy < 0.0) && (-sdy > n))
      n = -sdy;

    n /= SAMPLING_PIXELS;

    dx = (x1 - x0) / n;
    dy = (y1 - y0) / n;
    sdx *= 1.0 / n;
    sdy *= 1.0 / n;

    bzero(counts, sizeof(counts));

    XSetFillStyle(XtDisplay(tactmap->widget),
		  gc,FillOpaqueStippled); 

    lx = sx0;
    ly = sy0;
    for (i=1;i<=n;i++)
    {
	ex = x0 + dx * i;
	ey = y0 + dy * i;

	z1 = ctdb_lookup_elevation(ctdb,ex,ey)+dfn->line_intervis.target_agl; 
	/* Set up tags for libenvironment call from tactmap_dfn */
	
	/* Note that we are just multiplying the visibility results
	 * together, down below.  When we add support for fog, we should
	 * get a resolver from libenvironment to orchestrate the combinations.
	 */

        /* Set up tags for libenvironment call from tactmap_dfn */
   
	if (dfn->line_intervis.environ_effects & TACTMAP_INTERVIS_TERRAIN)
	  vis = ctdb_point_to_point(ctdb,x0, y0, z0,ex, ey, z1,
				  z1+dfn->line_intervis.target_height, 
				  dfn->line_intervis.target_width,
				  NULL, &opacity, 0.05,
				  dfn->line_intervis.n_veh,
				  dfn->line_intervis.veh, 
				  CTDB_DONT_IGNORE, CTDB_DONT_IGNORE);
	else
	{
	    vis = 1.0;
	    opacity = 1.0;
	}
	

	vis *= opacity;
	
/* Code previously used for calculating only smoke intervis effect */
/*	if ((dfn->line_intervis.environ_effects & TACTMAP_INTERVIS_SMOKE))
	{
	    char *modelname;
	    
	    modelname = env_find_model_for_effect(ENV_SMOKE_EFFECT);
	    if (modelname != NULL)
	      md = env_find_model(modelname);

	    eif.u.ray_visibility.input_viewer[X] = x0;
	    eif.u.ray_visibility.input_viewer[Y] = y0;
	    eif.u.ray_visibility.input_viewer[Z] = z0;
	    eif.u.ray_visibility.input_target[X] = ex;
	    eif.u.ray_visibility.input_target[Y] = ey;
	    eif.u.ray_visibility.input_target[Z] = 
	      z1 + (dfn->line_intervis.target_height/2.0);
	    eif.u.ray_visibility.input_wavelength = 
	      eui_get_em_band(tactmap->eui_prefs);
	    if ((md != -1)
		&& (env_get_explicit(ENV_RAY_VISIBILITY, &eif, &md, 1) != -1)) 

	    if(env_get(ENV_RAY_VISIBILITY, &eif) != -1)
	    {
	    printf(" Visibility is %f\n", eif.u.ray_visibility.output_value);
	    vis *= eif.u.ray_visibility.output_value;
	    }
	    
	}
*/
/* All local environmental obscurants are calculated for this ivis */
	if ((dfn->line_intervis.environ_effects & TACTMAP_INTERVIS_OBSCURANTS))
	{

	    eif.u.ray_visibility.input_viewer[X] = x0;
	    eif.u.ray_visibility.input_viewer[Y] = y0;
	    eif.u.ray_visibility.input_viewer[Z] = z0;
	    eif.u.ray_visibility.input_target[X] = ex;
	    eif.u.ray_visibility.input_target[Y] = ey;
	    eif.u.ray_visibility.input_target[Z] = 
	      z1 + (dfn->line_intervis.target_height/2.0);
	    eif.u.ray_visibility.input_wavelength = 
	      eui_get_em_band(tactmap->eui_prefs);

	    if(env_get(ENV_RAY_VISIBILITY, &eif) != -1)
	      vis *= eif.u.ray_visibility.output_value;
	}

	px = sx0 + i*sdx;
	py = sy0 + i*sdy;

	/* For point to point lines, we flop the meaning of shading:
	 * Clear lines mean no vis, Solid lines mean vis.  Hence, we
	 * pass 1-vis to the pix routine.
	 */
	this_pix = PCB(1-vis);

	segs[this_pix][counts[this_pix]].x1 = lx;
	segs[this_pix][counts[this_pix]].y1 = ly;
	segs[this_pix][counts[this_pix]].x2 = px;
	segs[this_pix][counts[this_pix]++].y2 = py;

	lx = px, ly = py;
    }

    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);
    for (i=0;i<N_PERCENT;i++)
    {
	if (counts[i])
	{
	    XSetStipple(XtDisplay(tactmap->widget), gc,
			tactmap->intervis_pix[i]); 
	    XSetLineAttributes(XtDisplay(tactmap->widget),
			       gc, 3, LineSolid, CapButt, JoinRound);
	    XDrawSegments(XtDisplay(tactmap->widget), d, gc, 
			  segs[i], counts[i]);
	}
    }

    XSetFillStyle(XtDisplay(tactmap->widget), gc,FillSolid);

    draw_ticked_line(tactmap, object, dfn, gc, (int32)sx0,(int32)sy0,x,y); 
}


static void 
draw_intervis_area(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		   TACTMAP_OBJECT_DFN *dfn, GC gc)
{
    DECLARE_REGS;
    float64 x0, y0, z0;
    float64 radius;
    int32 sx0, sy0, sx, sy;
    int32 dxy;
    float64 z1, dx,dy;
    int32 x,y;
    float64 vis, opacity;
    int32 i,j,k, n;
    int32 this_pix, last_pix=N_PERCENT-1;
    XRectangle rects[N_PERCENT];
    Drawable d;
    ENV_INTERFACE	eif;
    int32  md;
    CTDB *ctdb;
    
    /* Adjust coordinates to be relative to the current center cell */
    gcs_gcs2_to_gcscs((int32)dfn->area_intervis.center_point[CELL2D],
		      dfn->area_intervis.center_point,
		      cc, dfn->area_intervis.center_point);
    dfn->area_intervis.center_point[CELL2D] = cc;

    ctdb   = (CTDB *)gcs_get_tdb(cc, FALSE);
    x0     = dfn->area_intervis.center_point[X];
    y0     = dfn->area_intervis.center_point[Y];
    z0     = ctdb_lookup_elevation(ctdb,x0,y0)+dfn->area_intervis.viewer_agl;  
    radius = PIXELS(dfn->area_intervis.radius, tactmap->cur->ppm);
    
    /* Clip radius to absolute maximum possible to see */
    dx = NS_MAX((cx+hsw-x0),(x0-(cx-hsw)));
    dy = NS_MAX((cy+hsh-y0),(y0-(cy-hsh)));
    
    if (dx * dx + dy * dy < radius * radius)
      radius = sqrt(dx*dx + dy*dy);

    n =  radius / SAMPLING_PIXELS;

    dxy = SAMPLING_PIXELS;
    
    sx0 = tactmap_FX_pixel_number(x0,cx,hsw,ppm);
    sy0 = tactmap_FY_pixel_number(y0,cy,hsh,ppm);
    
    XSetFillStyle(XtDisplay(tactmap->widget), gc, FillStippled);

    for (k=0;k<N_PERCENT;k++)
    {
	rects[k].width = SAMPLING_PIXELS;
	rects[k].height = 0;
    }

    for (i= -n;i<=n;i++)
    {
	for (j= -n;j<=n;j++)
	  if (i*i + j*j <= n*n)
	  {
	      sx = sx0 + dxy * i;
	      sy = sy0 + dxy * j;
	      
	      x = X_METER_NUMBER(sx,cx,hsw,ppm);
	      y = Y_METER_NUMBER(sy,cy,hsh,ppm);

	      if ((x < cx-hsw) || (x > cx+hsw) || (y < cy-hsh) || (y > cy+hsh))
		continue;

	      z1 = ctdb_lookup_elevation(ctdb,(float64)x,(float64)y) +
		dfn->area_intervis.target_agl;

	      /* Set up call to libenvironment using env_effects from */
	      /* tactmap_dfn. */

	      if (dfn->area_intervis.environ_effects &
		  TACTMAP_INTERVIS_TERRAIN)
		vis = ctdb_point_to_point(ctdb, x0, y0, z0,(float64)x, 
					  (float64)y, z1,
					  z1+dfn->area_intervis.target_height, 
					  dfn->area_intervis.target_width,
					  NULL, &opacity,
					  0.05, 0, 0, 0, 0);   
	 
	      else 
		vis = 1.0;

	      vis *= opacity;
	
	      if ((dfn->area_intervis.environ_effects & 
		   TACTMAP_INTERVIS_OBSCURANTS))
	      {
/* Changed for changing smoke iv button to obscurants button */
/*		  char *modelname;
	    
		  modelname = env_find_model_for_effect(ENV_SMOKE_EFFECT);

		  if (modelname != NULL)
		    md = env_find_model(modelname);
*/

		  eif.u.ray_visibility.input_viewer[X] = x0;
		  eif.u.ray_visibility.input_viewer[Y] = y0;
		  eif.u.ray_visibility.input_viewer[Z] = z0;
		  eif.u.ray_visibility.input_target[X] = x;
		  eif.u.ray_visibility.input_target[Y] = y;
		  eif.u.ray_visibility.input_target[Z] = 
		    z1+(dfn->area_intervis.target_height/2.0);
		  eif.u.ray_visibility.input_wavelength = 
		    eui_get_em_band(tactmap->eui_prefs);
/*		  if ((md != -1)
		      && (env_get_explicit(ENV_RAY_VISIBILITY, &eif, &md, 1)
			  != -1))
*//* Smoke button to obscurants button */
		  if((env_get(ENV_RAY_VISIBILITY, &eif)) != 1)
		    vis *= eif.u.ray_visibility.output_value;
	      }
	      this_pix = PCB(vis);
	      if (this_pix != last_pix)
	      {
		  if ((this_pix!=N_PERCENT-1) && rects[this_pix].height)
		  {
		      XSetStipple(XtDisplay(tactmap->widget),gc,
				  tactmap->intervis_pix[this_pix]);

		      XFillRectangles(XtDisplay(tactmap->widget),
				      (Drawable)
				      ((object->display_mode == PERMANENT) ?
				      tactmap->background_pix :
				      XtWindow(tactmap->widget)), gc,
				      rects+this_pix, 1);
		      rects[this_pix].height = 0;
		  }
		  last_pix = this_pix;
	      }

	      if (!rects[this_pix].height)
	      {
		  rects[this_pix].x = sx-1;
		  rects[this_pix].y = sy-1;
	      }
	      rects[this_pix].height += SAMPLING_PIXELS;
	  }

	for (k=0;k<N_PERCENT-1;k++)
	  if (rects[k].height)
	  {
	      XSetStipple(XtDisplay(tactmap->widget), gc,
			  tactmap->intervis_pix[k]); 

	      XFillRectangles(XtDisplay(tactmap->widget),
			      (Drawable)
			      ((object->display_mode == PERMANENT) ?
			      tactmap->background_pix :
			      XtWindow(tactmap->widget)), gc, 
			      rects+k, 1);
	      rects[k].height = 0;
	  }
    }

    XSetFillStyle(XtDisplay(tactmap->widget), gc, FillSolid);

    /* Draw a little plus sign in the middle */

    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);
    XDrawLine(XtDisplay(tactmap->widget), d, gc, sx0-10, sy0, sx0+10, sy0);
    XDrawLine(XtDisplay(tactmap->widget), d, gc, sx0, sy0-10, sx0, sy0+10);
}

void 
tactmap_draw_area_iv(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		     int32 use_erase_gc, TACTMAP_BOUNDS *bounds, 
		     CTDB_SEARCH_SPACE_PTR space, 
		     int32 clear_area_when_finished, TACTMAP_OBJECT_DFN *dfn)
{
    DECLARE_REGS;
    GC gc;
    int32 x, y;
    int32 pixel_radius_sqr;
    float64 radius;


    /* Are we removing this object? */

    if (use_erase_gc)
    {
	if (object->display_mode == PERMANENT)
	  gc = tactmap->terrain_gc;
	else
	  gc = tactmap->background_gc;
    }
    else
      gc = tactmap->gc.area_intervis;

    radius = PIXELS(object->dfn.area_intervis.radius, ppm);
    pixel_radius_sqr = radius * 2;
    x = tactmap_FX_pixel_number(object->dfn.area_intervis.center_point[X],
				cx,hsw,ppm) - radius;      
    y = tactmap_FY_pixel_number(object->dfn.area_intervis.center_point[Y],
				cy,hsh,ppm) - radius;   
    if (x < 0) 
      x = 0;

    if (y < 0) 
      y = 0;

    /* Are we replacing an area with a new area */
    if (dfn)
    {
	/* Erase the last incarnation */

	/* use the pixel radius squared for the height and width of */
	/* the rectangle */
	if (object->display_mode == PERMANENT)
	{
	    XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		      tactmap->background_pix, tactmap->gc.background,
		      x, y, pixel_radius_sqr,
		      pixel_radius_sqr,
		      x, y);

	    if (clear_area_when_finished)
	      XClearArea(DPYWIN,
			 x, y, pixel_radius_sqr, 
			 pixel_radius_sqr, 0);
	}
	else
	{
	    XCopyArea(XtDisplay(tactmap->widget), tactmap->background_pix,
		      XtWindow(tactmap->widget), tactmap->gc.background,
		      x, y, pixel_radius_sqr, pixel_radius_sqr, x, y);
	}
    }

    if (!dfn)
      dfn = &object->dfn;

    if (use_erase_gc)
    {

	radius = PIXELS(dfn->area_intervis.radius, ppm);
	pixel_radius_sqr =  radius * 2;
	x = tactmap_FX_pixel_number(dfn->area_intervis.center_point[X],
				    cx, hsw,ppm) - radius;
	y = tactmap_FY_pixel_number(dfn->area_intervis.center_point[Y],
				    cy,hsh,ppm) - radius;
	
	if (x < 0) 
	  x = 0;
	
	if (y < 0) 
	  y = 0;

	if (object->display_mode == PERMANENT)
	  XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		    tactmap->background_pix, tactmap->gc.background,
		    x, y, pixel_radius_sqr, pixel_radius_sqr, x, y);
	else
	  XClearArea(DPYWIN,
		     x, y, pixel_radius_sqr, pixel_radius_sqr, 0);
    }
    else 
      draw_intervis_area(tactmap,object,dfn,gc);
    
    if (clear_area_when_finished && (object->display_mode == PERMANENT))
      XClearWindow(DPYWIN);

}


void 
tactmap_draw_line_iv(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		     int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
		     CTDB_SEARCH_SPACE_PTR space,
		     int32 clear_area_when_finished, TACTMAP_OBJECT_DFN *dfn)
{
    GC gc;

    /* Erase the last incarnation */
    if (dfn)
    {
	tactmap_draw_line_iv(tactmap, object, (int32) TRUE, bounds,
			     space, clear_area_when_finished,
			     (TACTMAP_OBJECT_DFN *) NULL);
    }
    
    if (!dfn)
      dfn = &object->dfn;
    
    
    /* Are we removing this object? */
    
    if (use_erase_gc)
    {

	if (object->display_mode == PERMANENT)
	  gc = tactmap->terrain_gc;
	else
	  gc = tactmap->background_gc;

	erase_intervis_line(tactmap, object, dfn, gc);
    }
    else
    {
	gc = tactmap->gc.line_intervis;
	if (dfn->line_intervis.show_percentages)
	  tactmap_draw_intervis_v2v(tactmap, object, dfn, gc);  
	else
	  draw_intervis_line(tactmap,object,dfn,gc, bounds);
    }
    
    if (clear_area_when_finished && (object->display_mode == PERMANENT))
      XClearWindow(DPYWIN);
}

static void 
tactmap_draw_intervis_v2v(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			  TACTMAP_OBJECT_DFN *dfn, GC gc)
{
    DECLARE_REGS;
    float64 vis, opacity;
    int32 sx0, sy0, sx1, sy1;
    float64 x0, y0, z0, x1, y1, z1;
    char buf[20];
    Drawable d;
    ENV_INTERFACE eif;
    CTDB *ctdb;
    int32 orig_cell=dfn->line_intervis.cell;
    
    /* Adjust coordinates to be relative to the tactmap frame of reference
     * Note: This macro will change the cell item to be equal to the new
     * frame of reference, which will preclude the next macro from changing
     * the second point.  Pass a dummy cell_id, with the correct context to 
     * the first macro and let the second macro change the context of both
     * points.
     */
    CHANGE_CONTEXT(orig_cell,
		   dfn->line_intervis.point1[X],
		   dfn->line_intervis.point1[Y],
		   cc);

    CHANGE_CONTEXT(dfn->line_intervis.cell,
		   dfn->line_intervis.point2[X],
		   dfn->line_intervis.point2[Y],
		   cc);

    x0 = dfn->line_intervis.point1[X];
    y0 = dfn->line_intervis.point1[Y];
    x1 = dfn->line_intervis.point2[X];
    y1 = dfn->line_intervis.point2[Y];
    ctdb = (CTDB *)gcs_get_tdb(cc, FALSE);

    z0 = ctdb_lookup_elevation(ctdb, x0, y0) + dfn->line_intervis.viewer_agl;  
    z1 = ctdb_lookup_elevation(ctdb, x1, y1) + dfn->line_intervis.target_agl;

    /* $$$ Set up libenvironment call using env_effects from tactmap_dfn. */
    vis = ctdb_point_to_point(ctdb, x0, y0, z0,x1, y1, z1,
			      z1+dfn->line_intervis.target_height,
			      dfn->line_intervis.target_width,
			      NULL, &opacity,
			      0.05,dfn->line_intervis.n_veh,
			      dfn->line_intervis.veh,
			      dfn->line_intervis.ignore1,
			      CTDB_DONT_IGNORE);

    vis *= opacity;

    /* Skip targets to which there is no visibility */
    if (vis == 0.0)
      return;

    eif.u.ray_visibility.input_viewer[X] = x0;
    eif.u.ray_visibility.input_viewer[Y] = y0;
    eif.u.ray_visibility.input_viewer[Z] = z0;
    eif.u.ray_visibility.input_target[X] = x1;
    eif.u.ray_visibility.input_target[Y] = y1;
    eif.u.ray_visibility.input_target[Z] = 
      z1+(dfn->line_intervis.target_height/2.0);
    eif.u.ray_visibility.input_wavelength = 
      eui_get_em_band(tactmap->eui_prefs);
	      
    if (env_get(ENV_RAY_VISIBILITY, &eif) != -1)
      vis *= eif.u.ray_visibility.output_value;
	
    
    /* Skip targets to which there is no visibility */
    if (vis == 0.0)
      return;
    
    sx0 = tactmap_FX_pixel_number(dfn->line_intervis.point1[X],cx,hsw,ppm);    
    sy0 = tactmap_FY_pixel_number(dfn->line_intervis.point1[Y],cy,hsh,ppm);
    
    sx1 = tactmap_FX_pixel_number(dfn->line_intervis.point2[X],cx,hsw,ppm);    
    sy1 = tactmap_FY_pixel_number(dfn->line_intervis.point2[Y],cy,hsh,ppm);
    
    /* For vehicle to vehicle lines, we flop the meaning of shading:
     * Clear lines mean no vis, Solid lines mean vis.  Hence, we
     * pass 1-vis to the pix routine.
     */
    XSetStipple(XtDisplay(tactmap->widget), gc,
		tactmap->intervis_pix[PCB(1-vis)]);    
    
    XSetFillStyle(XtDisplay(tactmap->widget),gc,FillOpaqueStippled);
    
    XSetLineAttributes(XtDisplay(tactmap->widget),
		       gc, 3, LineSolid, CapButt, JoinRound);
    
    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);

    XDrawLine(XtDisplay(tactmap->widget), d, gc , sx0, sy0, sx1, sy1);

    XSetFillStyle(XtDisplay(tactmap->widget), gc, FillSolid);
    
    sprintf(buf, " %d%%", (int)(100.0*vis + 0.5));

    XDrawImageString(XtDisplay(tactmap->widget), d,
		     gc, sx0 + 9*(sx1-sx0)/10, sy0 + 
		     9*(sy1-sy0)/10, 
		     buf, strlen(buf));
}


void tactmap_set_em_band(TACTMAP_PTR tactmap, ENV_EM_BAND band)
{
    eui_set_em_band(band, tactmap->eui_prefs);
}

ENV_EM_BAND tactmap_get_em_band(TACTMAP_PTR tactmap)
{
    return eui_get_em_band(tactmap->eui_prefs);
}

#endif /*USE_X*/
