#ifndef lint
static char rcsid [] = "$RCSfile: tmap_lines.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_lines.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw lines in world coordinates                   *
*       Created: Fri Dec 14 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include <p_po.h>

#define MAX_LINE_LEN 1000

static void wide_bdry(int16 pts_in[][2], int32 n_pts_in, int32 width,
		      int16 pts_out[][2], int32 *n_pts_out);

#define FRONT_ARC_SIZE         20
#define FRONT_ARC_SPACE         5
#define BERM_ARC_SIZE          20
#define BERM_ARC_SPACE        -15
#define MINEFIELD_ARC_SIZE     10
#define MINEFIELD_ARC_SPACE    20
#define AT_DITCH_TRIANGLE_SIZE 10
#define AT_DITCH_TRIANGLE_SPACE 0
#define WIRE_X_SIZE             7
#define WIRE_X_SPACE           20
#define COMM_SPACE             50
#define COMM_LENGTH            30
#define COMM_HEIGHT             3
#define FORTIFICATION_SIZE     20
#define FORTIFICATION_SPACE     0
#define FORTIFICATION_PART   0.50
#define FORTIFICATION_INV_AMP   2
#define ARROW_SIZE  20
#define ARROW_ANGLE (DEG_TO_RAD(10.0))
#define SINGLE_POINT_ARC_SIZE  15
#define MARKER_ARC_SIZE         5
#define LEFT                 0x01
#define RIGHT                0x02
#define TOP                  0x04
#define BOTTOM               0x08
#define MAX_XY              32000
#define MIN_XY             -32000

static void draw_styled_line(Display *display, Pixmap pixmap, GC gc,
			     int16 x0, int16 y0, int16 x1, int16 y1,
			     LineStyle variant, XRectangle *clip_box)
{
    int32 j, narcs;
    int32 angle;
    float64 dist,dx,dy;
    float64 sint, cost;
    int32 x,y,hx,hy,nx,ny;
    int32 arc_size, arc_size_cost, arc_size_sint;
    int32 arc_space;
    int16 pts[100];

    switch(variant)
    {
      case LSfrontA:
      case LSfrontB:
	arc_size = FRONT_ARC_SIZE;
	arc_space = FRONT_ARC_SPACE;
	break;
      case LSberm:
	arc_size = BERM_ARC_SIZE;
	arc_space = BERM_ARC_SPACE;
	break;
      case LSminefield:
      case LSminefieldAT:
      case LSminefieldAP:
	arc_size = MINEFIELD_ARC_SIZE;
	arc_space = MINEFIELD_ARC_SPACE;
	break;
      case LSATDitchA:
      case LSATDitchB:
	arc_size = AT_DITCH_TRIANGLE_SIZE;
	arc_space = AT_DITCH_TRIANGLE_SPACE;
	break;
      case LSfortification:
	arc_size = FORTIFICATION_SIZE;
	arc_space = FORTIFICATION_SPACE;
	break;
      case LSwire:
	arc_size = WIRE_X_SIZE;
	arc_space = WIRE_X_SPACE;
	break;
      case LScommunication:
	arc_size = 0;
	arc_space = COMM_SPACE;
	break;
    }

    dx = x1 - x0;
    dy = y1 - y0;
    angle = 64 * 180 * NS_ATAN2(-dy,dx) / 3.141592654;

    dist = sqrt(dx*dx + dy*dy);
    sint = dy / dist;
    cost = dx / dist;

    narcs = dist / (arc_size + arc_space);

    if (!narcs)
      return;

    dx /= narcs;
    dy /= narcs;

    arc_size_cost = arc_size * cost;
    arc_size_sint = arc_size * sint;

    for (j=0;j<narcs;j++)
    {
	x = (int32)(x0 + dx * (j) + arc_size_cost);
	y = (int32)(y0 + dy * (j) + arc_size_sint);

	/* This is not a speed optimization. There seems to
	 * be a bug in X which causes it to die if arcs are drawn
	 * far off the screen.
	 */
	if (!POINT_IN_RECT(x,y,*clip_box))
	  continue;

	switch(variant)
	{
	  case LSfrontA:
	  case LSfrontB:
	    XDrawArc(display, pixmap, gc,
		     x, y,
		     arc_size, arc_size, angle,
		     (variant == LSfrontA) ? 180 * 64 : -180 * 64);
	    break;
	  case LSminefield:
	    XDrawArc(display, pixmap, gc, x, y,
		     arc_size, arc_size, angle, 360 * 64);
	    break;
	  case LSminefieldAP:
	    XDrawLine(display, pixmap, gc, x, y-arc_size/2,
		      x+arc_size/2, y+arc_size/2);
	    XDrawLine(display, pixmap, gc, x+arc_size, y-arc_size/2,
		      x+arc_size/2, y+arc_size/2);
	    /* Fall through */
	  case LSminefieldAT:
	    XFillArc(display, pixmap, gc, x, y,
		     arc_size, arc_size, angle, 360 * 64);
	    break;
	  case LSberm:
	    XDrawArc(display, pixmap, gc, x, y,
		     arc_size, arc_size, angle - 90 * 64,
		     (j % 2) ? 180 * 64 : -180 * 64);
	    break;
	  case LSATDitchA:
	    pts[0] = x-dx;
	    pts[1] = y-dy;
	    pts[2] = x-dx/2+arc_size_sint;
	    pts[3] = y-dy/2-arc_size_cost;
	    pts[4] = x;
	    pts[5] = y;
	    XFillPolygon(display, pixmap, gc, (XPoint *)pts, 3,
			 Convex, CoordModeOrigin);
	    break;
	  case LSATDitchB:
	    pts[0] = x-dx;
	    pts[1] = y-dy;
	    pts[2] = x-dx/2-arc_size_sint;
	    pts[3] = y-dy/2+arc_size_cost;
	    pts[4] = x;
	    pts[5] = y;
	    XFillPolygon(display, pixmap, gc, (XPoint *)pts, 3,
			 Convex, CoordModeOrigin);
	    break;
	  case LSfortification:
	    pts[0] = x-dx;
	    pts[1] = y-dy;
	    pts[2] = x-dx+arc_size_sint/FORTIFICATION_INV_AMP;
	    pts[3] = y-dy-arc_size_cost/FORTIFICATION_INV_AMP;
	    pts[4] =
	      x-FORTIFICATION_PART*dx+arc_size_sint/FORTIFICATION_INV_AMP;
	    pts[5] =
	      y-FORTIFICATION_PART*dy-arc_size_cost/FORTIFICATION_INV_AMP;
	    pts[6] = x-FORTIFICATION_PART*dx;
	    pts[7] = y-FORTIFICATION_PART*dy;
	    pts[8] = x;
	    pts[9] = y;
	    XDrawLines(display, pixmap, gc, (XPoint *)pts, 5,
		       CoordModeOrigin);
	    break;
	  case LSwire:
	    hx = x + arc_size_cost;
	    hy = y + arc_size_sint;
	    XDrawLine(display, pixmap, gc, x-arc_size_sint, y+arc_size_cost,
		      hx+arc_size_sint, hy-arc_size_cost);
	    XDrawLine(display, pixmap, gc, hx-arc_size_sint, hy+arc_size_cost,
		      x+arc_size_sint, y-arc_size_cost);
	    break;
	  case LScommunication:
	    hx = x + COMM_LENGTH * cost - COMM_HEIGHT * sint;
	    hy = y + COMM_LENGTH * sint + COMM_HEIGHT * cost;
	    XDrawLine(display, pixmap, gc, x, y, hx, hy);
	    nx = x+dx - COMM_LENGTH * cost + COMM_HEIGHT * sint;
	    ny = y+dy - COMM_LENGTH * sint - COMM_HEIGHT * cost;
	    XDrawLine(display, pixmap, gc, hx, hy, nx, ny);
	    XDrawLine(display, pixmap, gc, nx, ny, 
		      (int32)(x+dx), (int32)(y+dy));
	    break;
	}
    }
}

#define MAX_BOUND_VERTS 1000

static void draw_bounding_box(Display *display, Pixmap pixmap, GC gc, 
			      int16 verts[][2], int32 nverts, 
			      int32 width, XRectangle *clip_box)
{
    int16 pts_out[MAX_BOUND_VERTS][2];
    int16 pts[MAX_BOUND_VERTS*4+4];
    int32 npts_out;
    int16 x0,y0,x1,y1;
    int32 i,j;

    if (!width)
      return;

    if (width < MINEFIELD_ARC_SIZE*2)
      width = MINEFIELD_ARC_SIZE*2;

    wide_bdry(verts, nverts, width/2, pts_out, &npts_out);

    for (i=0,j=0;i<npts_out-1;i++)
    {
	x0 = pts_out[i][0];
	y0 = pts_out[i][1];
	x1 = pts_out[i+1][0];
	y1 = pts_out[i+1][1];

	if (LINE_MAY_CROSS_RECT(x0,y0,x1,y1,*clip_box))
	{
	    pts[j] = x0;
	    pts[j+1] = y0;
	    pts[j+2] = x1;
	    pts[j+3] = y1;
	    j += 4;
	}
    }

    XDrawSegments(display, pixmap, gc, (XSegment *)pts, j/4);
}

static void draw_styled_segments(Display *display, Window pixmap, GC gc,
				 int16 *segs, int32 nsegs, int16 pts[][2],
				 int32 npts, LineStyle variant, 
				 int32 width, XRectangle *clip_box)
{
    int32 i;

    switch(variant)
    {
      case LSATDitchA:
      case LSATDitchB:
      case LSwire:
	XDrawSegments(display, pixmap, gc, (XSegment *)segs, nsegs);
	break;
      case LSminefield:
      case LSminefieldAT:
      case LSminefieldAP:
	draw_bounding_box(display, pixmap, gc, pts, npts, width, clip_box);
	break;
    }

	
    for (i=0;i<nsegs*4;i+=4)
    {
        if (variant == LSminefieldMarker)
	{
	    XDrawLine(display, pixmap, gc, 
		      segs[i], segs[i+1]-MARKER_ARC_SIZE,
		      segs[i]+MARKER_ARC_SIZE, segs[i+1]+MARKER_ARC_SIZE);
	    XDrawLine(display, pixmap, gc, 
		      segs[i]+MARKER_ARC_SIZE*2, segs[i+1]-MARKER_ARC_SIZE,
		      segs[i]+MARKER_ARC_SIZE, segs[i+1]+MARKER_ARC_SIZE);
	    XDrawLine(display, pixmap, gc,
		      segs[i+2], segs[i+3]-MARKER_ARC_SIZE,
		      segs[i+2]+MARKER_ARC_SIZE, segs[i+3]+MARKER_ARC_SIZE);
	    XDrawLine(display, pixmap, gc,
		      segs[i+2]+MARKER_ARC_SIZE*2, segs[i+3]-MARKER_ARC_SIZE,
		      segs[i+2]+MARKER_ARC_SIZE, segs[i+3]+MARKER_ARC_SIZE);
	}
        else if (variant == LSbreachLaneMarker) 
	{
	    XDrawLine(display, pixmap, gc,
		      segs[i], segs[i+1]+MARKER_ARC_SIZE,
		      segs[i]-MARKER_ARC_SIZE, segs[i+1]-MARKER_ARC_SIZE);
	    XDrawLine(display, pixmap, gc,
		      segs[i]+MARKER_ARC_SIZE*2, segs[i+1]+MARKER_ARC_SIZE,
		      segs[i]-MARKER_ARC_SIZE, segs[i+1]-MARKER_ARC_SIZE);
	    XDrawLine(display, pixmap, gc,
		      segs[i+2], segs[i+3]+MARKER_ARC_SIZE,
		      segs[i+2]-MARKER_ARC_SIZE, segs[i+3]-MARKER_ARC_SIZE);
	    XDrawLine(display, pixmap, gc,
		      segs[i+2]+MARKER_ARC_SIZE*2, segs[i+3]+MARKER_ARC_SIZE,
		      segs[i+2]-MARKER_ARC_SIZE, segs[i+3]-MARKER_ARC_SIZE);
	}    
	else 
	{
	    draw_styled_line(display, pixmap, gc,
			     segs[i], segs[i+1], segs[i+2], segs[i+3],
			     variant, clip_box);
	}
	
    }

}

static void arrow(Display *display, Window pixmap, GC gc, 
		  int16 x0, int16 y0, int16 x1, int16 y1)
{
    float64 dx, dy, angle;
    XSegment segs[2];

    dx = x1 - x0;
    dy = y1 - y0;
    angle = NS_ATAN2(dy, dx);

    segs[0].x1 = segs[1].x1 = x1;
    segs[0].y1 = segs[1].y1 = y1;

    segs[0].x2 = segs[0].x1 - ARROW_SIZE*cos(angle+ARROW_ANGLE);
    segs[0].y2 = segs[0].y1 - ARROW_SIZE*sin(angle+ARROW_ANGLE);
    segs[1].x2 = segs[1].x1 - ARROW_SIZE*cos(angle-ARROW_ANGLE);
    segs[1].y2 = segs[1].y1 - ARROW_SIZE*sin(angle-ARROW_ANGLE);

    XDrawSegments(display, pixmap, gc, segs, 2);
}


/*
 *  This macro computes the code for the point
 *  NOTE: this code indicate where the point is
 *  in relation to the (XRectangle) rec
 */
#define   OP_CODE(xx, yy, rec, code)                                \
                 code = 0;                                          \
                 if (yy < rec.y)                     code = BOTTOM; \
                 else if (yy > (rec.y + rec.height)) code = TOP;    \
                 if (xx < rec.x)                     code |= LEFT;  \
                 else if (xx > (rec.x + rec.width))  code |= RIGHT  
                 

/*
 *  do_clip_fails
 * 
 *  Checks to see if the segment is leagle to draw
 *  IF needed it will clip the segment to the bounding box
 *  contained in rec
 *
 *  NOTE:  The cliping algorithm used is the Cohen-Sutherland
 *  Line-Cliping Algorithym 
 */ 
static int32 do_clip_fails(int32 *x0, int32 *y0, int32 *x1, int32 *y1,
			   XRectangle rec)
{
   uint32  code0, code1, temp_code; 
   int32   done = 0, 
           temp_x, temp_y;

  /* IF this segment is completly out of there */ 
   if (!LINE_MAY_CROSS_RECT(*x0, *y0, *x1, *y1, rec))
     return (TRUE);

   /* IF the segment is in an acceptable range */
   if ((*x0 < MAX_XY && *x0 > MIN_XY) &&
       (*x1 < MAX_XY && *x1 > MIN_XY) &&
       (*y0 < MAX_XY && *y0 > MIN_XY) &&
       (*y1 < MAX_XY && *y1 > MIN_XY))
      return (FALSE);
   
   OP_CODE(*x0, *y0, rec, code0);
   OP_CODE(*x1, *y1, rec, code1);
   
   while (!done)
   {
      /* This is a triveal accept */
      if ((code0 == 0) && (code1 == 0))
         done =  1;
 
      /* This is a triveal reject */
      else if ((code0 & code1) != 0)
         done = 2;
    
      /* This segment needs to be cliped a little */
      else
      {
         temp_code = (code0 != 0)? code0:code1;
         if ((temp_code & RIGHT) == RIGHT)
         {
            temp_y = *y0 + (*y1 - *y0) * ((rec.x + rec.width) - *x0)/(*x1 - *x0);
            temp_x = rec.x + rec.width;
         }
         else if ((temp_code & LEFT) == LEFT)
         {
            temp_y = *y0 + (*y1 - *y0) * (rec.x - *x0)/(*x1 - *x0);
            temp_x = rec.x;
         }
         else if ((temp_code & TOP) == TOP)
         {
            temp_x = *x1 + (*x0 - *x1) * ((rec.y + rec.height) - *y1)/(*y0 - *y1);
            temp_y = rec.y + rec.height;
         }
         else if ((temp_code & BOTTOM) == BOTTOM)
         {
            temp_x = *x0 + (*x1 - *x0) * (rec.y - *y0)/(*y1 - *y0);
            temp_y = rec.y;
         }
         if (temp_code == code0)
         {
            *x0 = temp_x;
            *y0 = temp_y;
            OP_CODE(*x0, *y0, rec, code0);
         }
         else
         {
            *x1 = temp_x;
            *y1 = temp_y;
            OP_CODE(*x1, *y1, rec, code1);
         }
      }
   }
   return ((done == 1)? FALSE:TRUE);
}

/*ARGSUSED*/
void tactmap_draw_linear(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
			 int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			 CTDB_SEARCH_SPACE_PTR space,
			 int32 clear_area_when_finished,
			 TACTMAP_OBJECT_DFN *dfn)
                                  
                                 
                                       
                                 
                                
                                                   
                              /* new version of the object */
{
    DECLARE_REGS;
    int32 x, y, x0, y0, old_x, old_y;
    int32 width;
    int32 i, j, k;
    int16 segs[10000];
    int16 verts[5000][XY];
    float64 *pts;
    XRectangle bounding_box;
    GC gc;
    int32 removing_old = use_erase_gc;
    SNSTVE_OBJECT snsobj;
    Drawable d;


    if (dfn)
    {
	/* Erase the last incarnation */
	tactmap_draw_linear(tactmap, object, (int32) TRUE, bounds,
			    space, clear_area_when_finished,
			    (TACTMAP_OBJECT_DFN *) NULL);
    }

    if (!dfn)
      dfn = &object->dfn;

    pts = dfn->linear.vertices;

    if ((dfn->linear.n_vertices < 1) || (!dfn->linear.style))
      return;

    if (use_erase_gc)
    {
	if (object->display_mode == PERMANENT)
	  gc = tactmap->terrain_gc;
	else
	  gc = tactmap->background_gc;
    }
    else
      gc = dfn->linear.gc;
	
    bounding_box.x      = tactmap_X_pixel_number(bounds->min_x,cx,hsw,ppm);
    bounding_box.y      = tactmap_Y_pixel_number(bounds->max_y,cy,hsh,ppm);
    bounding_box.width  = PIXELS(bounds->max_x-bounds->min_x,ppm);
    bounding_box.height = PIXELS(bounds->max_y-bounds->min_y,ppm);

    snsobj.sensClass        = dfn->sensitive.sensClass;
    snsobj.user_data    = dfn->sensitive.user_data;
    
    if (dfn->linear.dashed)
      XSetLineAttributes(XtDisplay(tactmap->widget), gc, dfn->linear.thickness,
			 LineOnOffDash, CapButt, JoinMiter);
    else
      XSetLineAttributes(XtDisplay(tactmap->widget), gc, dfn->linear.thickness,
			 LineSolid, CapButt, JoinMiter);

    width = dfn->linear.width*ppm;

    if ( cc != dfn->linear.cell )
    {
	/* if the context of the points is different than the context
	 * of the tactmap center cell then change the context to match
	 * that of the center cell.
	 */
	int32 i;
	float64 *pt=pts;
	
	for (i=0;i<dfn->linear.n_vertices;i++,pt+=2)
	{
	    int32 cell = dfn->linear.cell; 
	    CHANGE_CONTEXT(cell,pt[X],pt[Y],cc);
	}
	/* update the context of the new points */
	dfn->linear.cell = cc;
    }
    
		   
    old_x = x = tactmap_FX_pixel_number(pts[X],cx,hsw,ppm);
    old_y = y = tactmap_FY_pixel_number(pts[Y],cy,hsh,ppm);

    verts[0][X] = x;
    verts[0][Y] = y;
    k = 1;

    /* Stuff the first sensitive vertex */
    snsobj.instance           = dfn->sensitive.instance | IS_POINT;
    snsobj.kind               = SNSTVE_CIRCLE;
    snsobj.attr.circle.x      = x;
    snsobj.attr.circle.y      = y;
    snsobj.attr.circle.radius = 10;
    
    if (removing_old)
      snstve_remove_object(tactmap->sensitive, x, y,
			   snsobj.sensClass, snsobj.instance);
    else
      snstve_add_object(tactmap->sensitive, &snsobj);

    if (object->display_mode == PERMANENT)
      d = tactmap->background_pix;
    else
      d = XtWindow(tactmap->widget);

    for (i=2,j=0;i<dfn->linear.n_vertices*2;i+=2)
    {
	x0 = old_x;
	y0 = old_y;

	old_x = x = tactmap_FX_pixel_number(pts[i],cx,hsw,ppm);
	old_y = y = tactmap_FY_pixel_number(pts[i+1],cy,hsh,ppm);

	verts[k][X] = x;
	verts[k][Y] = y;
	k += 1;
	
	if ((x==x0) && (y==y0))
	  continue;

        if (do_clip_fails(&x0, &y0, &x, &y, bounding_box))
	  continue;

	/* Stuff the sensitive vertex before the line, so it will be
	 * on top.
	 */
	snsobj.instance =
	  dfn->sensitive.instance | ((i>>1) << INDEX_SHIFT) | IS_POINT;
	snsobj.kind               = SNSTVE_CIRCLE;
	snsobj.attr.circle.x      = x;
	snsobj.attr.circle.y      = y;
	snsobj.attr.circle.radius = 10;

	if (removing_old)
	  snstve_remove_object(tactmap->sensitive, x, y,
			       snsobj.sensClass, snsobj.instance);
	else
	  snstve_add_object(tactmap->sensitive, &snsobj);

	/* Stuff the sensitive segment */
	snsobj.instance =
	  (dfn->sensitive.instance | (((i-2)>>1) << INDEX_SHIFT));
	snsobj.kind                                = SNSTVE_LINE;
	snsobj.attr.line.x0                        = x0;
	snsobj.attr.line.y0                        = y0;
	snsobj.attr.line.x1                        = x;
	snsobj.attr.line.y1                        = y;
	snsobj.attr.line.width                     = width + 10;

	if (removing_old)
	  snstve_remove_object(tactmap->sensitive, x0, y0,
			       snsobj.sensClass,snsobj.instance);
	else
	  snstve_add_object(tactmap->sensitive,&snsobj);

	segs[j]   = x0;
	segs[j+1] = y0;
	segs[j+2] = x;
	segs[j+3] = y;
	j += 4;
    }

    if (j)
    {
	if (dfn->linear.style != LSplain)
	  draw_styled_segments(XtDisplay(tactmap->widget), d,
			       gc, segs, j/4,
			       verts, k, dfn->linear.style, width,
			       &bounding_box);
	else
	  XDrawSegments(XtDisplay(tactmap->widget), d,
			gc, (XSegment *)segs, j/4);

	if (dfn->linear.backward_arrow)
	  arrow(XtDisplay(tactmap->widget), d,
		gc, segs[3], segs[4], segs[0], segs[1]);

	if (dfn->linear.forward_arrow)
	  arrow(XtDisplay(tactmap->widget), d,
		gc, segs[j-4], segs[j-3], segs[j-2], segs[j-1]);
    }
    else
    {
	if (dfn->linear.style == LSminefieldMarker)
	{
	    XDrawLine(XtDisplay(tactmap->widget), d, gc,
		      verts[0][X],
		      verts[0][Y]-MARKER_ARC_SIZE,
		      verts[0][X]+MARKER_ARC_SIZE,
		      verts[0][Y]+MARKER_ARC_SIZE);
	    XDrawLine(XtDisplay(tactmap->widget), d, gc, 
		      verts[0][X]+MARKER_ARC_SIZE*2, 
		      verts[0][Y]-MARKER_ARC_SIZE, 
		      verts[0][X]+MARKER_ARC_SIZE, 
		      verts[0][Y]+MARKER_ARC_SIZE);

	}
	else if (dfn->linear.style == LSbreachLaneMarker)
	{
	    XDrawLine(XtDisplay(tactmap->widget), d, gc,
		      verts[0][X], 
		      verts[0][Y]+MARKER_ARC_SIZE,
		      verts[0][X]-MARKER_ARC_SIZE, 
		      verts[0][Y]-MARKER_ARC_SIZE);
	    XDrawLine(XtDisplay(tactmap->widget), d, gc, 
		      verts[0][X]+MARKER_ARC_SIZE*2, 
		      verts[0][Y]+MARKER_ARC_SIZE, 
		      verts[0][X]-MARKER_ARC_SIZE, 
		      verts[0][Y]-MARKER_ARC_SIZE);

	}
	else
	  XFillArc(XtDisplay(tactmap->widget), d,
		   gc,
		   verts[0][X]-SINGLE_POINT_ARC_SIZE/2,
		   verts[0][Y]-SINGLE_POINT_ARC_SIZE/4,
		   SINGLE_POINT_ARC_SIZE, SINGLE_POINT_ARC_SIZE/2,
		   0, 360*64);
    }

    if (clear_area_when_finished && (object->display_mode == PERMANENT))
      XClearWindow(DPYWIN);
}

static int32 
get_intersection(int16 x0, int16 y0, int16 x1, int16 y1, int16 x2, int16 y2, 
		 int16 x3, int16 y3, int16 *ret)
    /*          |---segment---| |---segment---| */
                          /* segment #1 */
                          /* segment #2 */
                  
{
    float64 a0, b0, c0, d0;
    float64 a1, b1, c1, d1;
    float64 numer, denom;
    float64 t0, t1;

    /* This uses the algorithm from ctdb_segment_intersection_t.
     * see libctdb/ct_poly.c for a description.
     */

    /* The returned miss direction indicates whether the intersection
     * of the lines these segment line on occurs downstream from the
     * first segment (FORWARD) or behind the first segment (BACKWARD).
     * When the intersection is FORWARD, the ends of the two segments
     * are pasted together with an extra segment; when the intersection
     * is BACKWARD, the endpoints are discarded altogether.
     */

#define MISS_FORWARD  1
#define MISS_BACKWARD 2


    a0 = x1 - x0; b0 = x0;
    c0 = y1 - y0; d0 = y0;
    a1 = x3 - x2; b1 = x2;
    c1 = y3 - y2; d1 = y2;

    numer = c0 * (b1 - b0) + a0 * (d0 - d1);
    denom = a0 * c1 - a1 * c0;

    if (denom == 0.0)
      return MISS_BACKWARD;

    t1 = numer / denom;

    /* Are we within the bounds of the second segment? */
    if (t1 < 0.0)
      return MISS_FORWARD;

    if (t1 > 1.0)
      return MISS_BACKWARD;

    /* Derive t0 from t1 using either the x or the y equation.
     */
    if (a0)
      t0 = (b1 - b0 + a1 * t1) / a0;
    else
      t0 = (d1 - d0 + c1 * t1) / c0;

    /* Are we within the bounds of the first segment? */
    if (t0 < 0.0)
      return MISS_BACKWARD;
    if (t0 > 1.0)
      return MISS_FORWARD;

    ret[X] = b1 + a1 * t1;
    ret[Y] = d1 + c1 * t1;
    return 0; /* They do intersect */
}

static void wide_bdry(int16 pts_in[][2], int32 n_pts_in, int32 width, 
		      int16 pts_out[][2], int32 *n_pts_out)
{
    int32 i, j;
    int32 x0, y0, x1, y1;
    float64 dx, dy, mag;
    int32 miss;
    int16 offset[MAX_LINE_LEN][XY];

    if (n_pts_in < 2)
    {
	*n_pts_out = 0;
	return;
    }

    /* Find the offset to move 'width' units away from each segment */
    for (i=0;i<n_pts_in-1;i++)
    {
	x0 = pts_in[i][X]; y0 = pts_in[i][Y];
	x1 = pts_in[i+1][X]; y1 = pts_in[i+1][Y];
	if (x1 == x0)
	{
	    offset[i][Y] = 0.0;
	    if (y1 > y0)
	      offset[i][X] = -width;
	    else
	      offset[i][X] = width;
	}
	else if (y1 == y0)
	{
	    offset[i][X] = 0.0;
	    if (x1 > x0)
	      offset[i][Y] = width;
	    else
	      offset[i][Y] = -width;
	}
	else
	{
	    dy = (x1 - x0);
	    dx = (y0 - y1);
	    mag = 1.0 / sqrt(dx * dx + dy * dy);
	    offset[i][X] = dx * mag * width;
	    offset[i][Y] = dy * mag * width;
	}
    }
    offset[i][X] = offset[i-1][X];
    offset[i][Y] = offset[i-1][Y];

    /* Generate the right side */
    j = 0;
    pts_out[j][X] = pts_in[0][X] + offset[0][X];
    pts_out[j][Y] = pts_in[0][Y] + offset[0][Y];
    j++;
    for (i=1;i<n_pts_in-1;i++)
    {
	if (!(miss = get_intersection(pts_in[i-1][X] + offset[i-1][X],
				      pts_in[i-1][Y] + offset[i-1][Y],
				      pts_in[i][X]   + offset[i-1][X],
				      pts_in[i][Y]   + offset[i-1][Y],
				      pts_in[i][X]   + offset[i][X],
				      pts_in[i][Y]   + offset[i][Y],
				      pts_in[i+1][X] + offset[i][X],
				      pts_in[i+1][Y] + offset[i][Y],
				      pts_out[j])))
	  j++;
	else if (miss == MISS_FORWARD)
	{
	    pts_out[j][X] = pts_in[i][X] + offset[i-1][X];
	    pts_out[j][Y] = pts_in[i][Y] + offset[i-1][Y];
	    j++;
	    pts_out[j][X] = pts_in[i][X] + offset[i][X];
	    pts_out[j][Y] = pts_in[i][Y] + offset[i][Y];
	    j++;
	}
    }
    pts_out[j][X] = pts_in[i][X] + offset[i][X];
    pts_out[j][Y] = pts_in[i][Y] + offset[i][Y];
    j++;

    /* Generate the left side */
    pts_out[j][X] = pts_in[i][X] - offset[i][X];
    pts_out[j][Y] = pts_in[i][Y] - offset[i][Y];
    j++;
    for (--i;i>0;--i)
    {
	if (!(miss = get_intersection(pts_in[i+1][X] - offset[i][X],
				      pts_in[i+1][Y] - offset[i][Y],
				      pts_in[i][X]   - offset[i][X],
				      pts_in[i][Y]   - offset[i][Y],
				      pts_in[i][X]   - offset[i-1][X],
				      pts_in[i][Y]   - offset[i-1][Y],
				      pts_in[i-1][X] - offset[i-1][X],
				      pts_in[i-1][Y] - offset[i-1][Y],
				      pts_out[j])))
	  j++;
	else if (miss == MISS_FORWARD)
	{
	    pts_out[j][X] = pts_in[i][X] - offset[i][X];
	    pts_out[j][Y] = pts_in[i][Y] - offset[i][Y];
	    j++;
	    pts_out[j][X] = pts_in[i][X] - offset[i-1][X];
	    pts_out[j][Y] = pts_in[i][Y] - offset[i-1][Y];
	    j++;
	}
    }
    pts_out[j][X] = pts_in[i][X] - offset[i][X];
    pts_out[j][Y] = pts_in[i][Y] - offset[i][Y];
    j++;
    pts_out[j][X] = pts_out[0][X];
    pts_out[j][Y] = pts_out[0][Y];
    j++;

    *n_pts_out = j;
}

#endif /*USE_X*/
