#ifndef lint
static char rcsid [] = "$RCSfile: tmap_miclic.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_miclic.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw wire                                         *
*       Created: Mon Mar 18 1996                                            *
*       Author: bkenney                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <libvecmat.h>
#include <math.h>

#define MIN_ARC_SIZE         2     /* pixels */
#define MAX_ARC_SIZE         200   /* pixels */
#define DESIRED_ARC_SIZE     14    /* meters */
#define FULL_REVOLUTION      23040 /* (360 * 64) */


void tactmap_draw_miclic_line(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
			      int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			      CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XSegment seg[1];
    float32 *verts;
    float32 first_point[2];
    float32 scalar;
    float32 sec_point[2];
    float32 direction[3];
    int32 n_verts;
    CTDB_ABSTRACT_DATA *data;
    XGCValues gcv;
    int32 j;
    GC gc;
    SNSTVE_OBJECT snsobj;

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.miclic_line;

    gcv.line_width = 0;
    XChangeGC(XtDisplay(tactmap->widget), gc,
	      GCLineWidth, &gcv);
    
    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;
    
    /* Now draw the abstract. */
    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_MICLIC, &verts, 
					&data, NULL))
    {
	if (n_verts == 1)
	  continue;

	if (data->miclic.detonated)
	  continue;
	
	if (object->dfn.miclic_line.mask & TACTMAP_MICLIC_LINE_IND)
	{	    
	    for (j=0;j<n_verts;j++)
	    {
		seg[0].x1 = tactmap_FX_pixel_number(verts[2*j],cx,hsw,ppm);
		seg[0].y1 = tactmap_FY_pixel_number(verts[2*j+1],cy,hsh,ppm);
		
		if (j)
		{
		    snsobj.instance = 0;
		    snsobj.kind = SNSTVE_LINE;
		    snsobj.instance = 0;
		    snsobj.attr.line.x0 = seg[0].x2;
		    snsobj.attr.line.y0 = seg[0].y2;
		    snsobj.attr.line.x1 = seg[0].x1;
		    snsobj.attr.line.y1 = seg[0].y1;
		    snsobj.attr.line.width = 
		        data->miclic.segment_width*ppm + 10;
		    
		    snstve_add_object(tactmap->sensitive, &snsobj);
		  
		    gcv.line_width = data->miclic.segment_width*ppm;
                    XChangeGC(XtDisplay(tactmap->widget), gc,
			      GCLineWidth, &gcv);
 
		    XDrawSegments(XtDisplay(tactmap->widget),
				  tactmap->terrain_pix, gc,
				  seg, 1);
		}
		
		seg[0].x2 = seg[0].x1;
		seg[0].y2 = seg[0].y1;
	    }
	}
    }
}


/* draw the scorch marks created by the miclic detonation... this code
 * duplicated from tmap_wire.c..
 */
static void draw_scorch_segment(Display *display, Window pixmap, GC gc,
				int32 x0, int32 y0, int32 x1, int32 y1,
				float32 width, XRectangle *clip_box, 
				float64 ppm)
{
    int32 j, narcs;
    int32 x,y,hx,hy,nx,ny;
    int32 arc_size_cost, arc_size_sint;
    int32 arc_size, arc_space;
    float64 dist,dx,dy, offset[2];
    float64 sint, cost;
    XGCValues gcv;


    arc_size = width * ppm;

    if(arc_size < MIN_ARC_SIZE)
      arc_size = MIN_ARC_SIZE;

    arc_space = -0.5 * arc_size;

    dx = x1 - x0;
    dy = y1 - y0;
    
    /* Compute the normal of (dx, dy). */
    offset[X] = -dy;
    offset[Y] = dx;
    vmat2_unit64(offset, offset);
    VMAT2_SCAL_VEC_MUL(arc_space, offset, offset);
    
    dist = sqrt(dx*dx + dy*dy);
    sint = dy / dist;
    cost = dx / dist;
    
    narcs = dist / (arc_size + arc_space);
    
    if(!narcs)
      narcs = 1;
    
    dx /= narcs;
    dy /= narcs;
    
    arc_size_cost = arc_size * cost * 0.5;
    arc_size_sint = arc_size * sint * 0.5;
    
    for (j=0;j<narcs;j++)
    {
	x = (int32)(x0 + dx * (j) + arc_size_cost - arc_size * 0.5);
	y = (int32)(y0 + dy * (j) + arc_size_sint - arc_size * 0.5);
	
	/* This is not a speed optimization. There seems to
	 * be a bug in X which causes it to die if arcs are drawn
	 * far off the screen.
	 */
	if (!POINT_IN_RECT(x,y,*clip_box))
	  continue;
	
	XFillArc(display, pixmap, gc, x, y,
		 arc_size, arc_size, 0, FULL_REVOLUTION);
	
    }
}



void tactmap_draw_miclic_scorch(TACTMAP_PTR tactmap, 
				TACTMAP_OBJECT_PTR object, 
				int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
				CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XSegment seg[1];
    float32 *verts;
    int32 n_verts;
    float32 first_point[2];
    float32 sec_point[2];
    float32 direction[3];
    float32 scalar;
    CTDB_ABSTRACT_DATA *data;
    XGCValues gcv;
    XRectangle bounding_box;
    int32 j, k, narcs;
    int32 x, y, hx,hy,nx,ny;
    int32 arc_size_cost, arc_size_sint;
    int32 arc_size, arc_space;
    float64 dist,dx,dy, offset[2];
    float64 sint, cost;
    GC gc;
    SNSTVE_OBJECT snsobj;

    bounding_box.x      = tactmap_X_pixel_number(bounds->min_x,cx,hsw,ppm);
    bounding_box.y      = tactmap_Y_pixel_number(bounds->max_y,cy,hsh,ppm);
    bounding_box.width  = PIXELS(bounds->max_x-bounds->min_x,ppm);
    bounding_box.height = PIXELS(bounds->max_y-bounds->min_y,ppm);
    
    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.miclic_scorch;
    
    gcv.line_style = LineSolid;

    XChangeGC(XtDisplay(tactmap->widget), gc, GCLineStyle, &gcv);

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;
    
    /* Now draw the abstract. */
    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_MICLIC, &verts, 
					&data, NULL))
    {
	if (n_verts == 1)
	  continue;

	if (!data->miclic.detonated)
	  continue;
	
	if (object->dfn.miclic_scorch.mask & TACTMAP_MICLIC_SCORCH_IND)
	{	    
	    for (j=0;j<n_verts;j++)
	    {
		seg[0].x1 = tactmap_FX_pixel_number(verts[2*j],cx,hsw,ppm);
		seg[0].y1 = tactmap_FY_pixel_number(verts[2*j+1],cy,hsh,ppm);
		    
		if (j)
		{
		    snsobj.instance = 0;
		    snsobj.kind = SNSTVE_LINE;
		    snsobj.instance = 0;
		    snsobj.attr.line.x0 = seg[0].x2;
		    snsobj.attr.line.y0 = seg[0].y2;
		    snsobj.attr.line.x1 = seg[0].x1;
		    snsobj.attr.line.y1 = seg[0].y1;
		    snsobj.attr.line.width = 10;
		    
		    snstve_add_object(tactmap->sensitive, &snsobj);
		    
		    draw_scorch_segment(XtDisplay(tactmap->widget),
					tactmap->terrain_pix, gc,
					seg[0].x2, seg[0].y2,
					seg[0].x1, seg[0].y1,
					data->miclic.scorch_width,
					&bounding_box, ppm);
		    
		}		

	    
		seg[0].x2 = seg[0].x1;
		seg[0].y2 = seg[0].y1;
	    }
	}
    }
}


#endif /*USE_X*/
