#ifndef lint
static char rcsid [] = "$RCSfile: tmap_obj.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_obj.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle objects                                    *
*       Created: Wed Sep 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/

static int32 redraw_required(TACTMAP_OBJECT_CLASS sensClass, 
			     TACTMAP_OBJECT_DFN *old, TACTMAP_OBJECT_DFN *new);

/*ARGSUSED*/
TACTMAP_OBJECT_PTR tactmap_create_object(TACTMAP_PTR tactmap,
					 TACTMAP_OBJECT_CLASS sensClass, 
					 TACTMAP_OBJECT_DFN *dfn)
{
    TACTMAP_OBJECT_PTR obj;
    int32 len;

    obj = (TACTMAP_OBJECT_PTR)STDALLOC(sizeof(struct tactmap_object));
    bzero((char *)obj, sizeof(struct tactmap_object));

    switch(sensClass)
    {
      case TACTMAP_HYPSO: obj->draw_fcn = tactmap_draw_hypso; break;
      case TACTMAP_WATER: obj->draw_fcn = tactmap_draw_water; break;
      case TACTMAP_ROADS: obj->draw_fcn = tactmap_draw_water; break;
#ifdef AE_FEATURES
      case TACTMAP_OFFROADS: obj->draw_fcn = tactmap_draw_offroads; break;
      case TACTMAP_TACT_SIGNS: obj->draw_fcn = tactmap_draw_tact_signs; break;
#endif
      case TACTMAP_TREES: obj->draw_fcn = tactmap_draw_trees; break;
      case TACTMAP_CONTAMINANT: 
	obj->draw_fcn = tactmap_draw_contaminant; break;
      case TACTMAP_BUILDINGS:
	obj->draw_fcn = tactmap_draw_buildings; break;
      case TACTMAP_PIPELINES:
	obj->draw_fcn = tactmap_draw_pipelines; break;
      case TACTMAP_POLITICAL:
	obj->draw_fcn = tactmap_draw_political; break;
      case TACTMAP_RAILROADS:
	obj->draw_fcn = tactmap_draw_railroads; break;
      case TACTMAP_POWERLINES:
	obj->draw_fcn = tactmap_draw_powerlines; break;
      case TACTMAP_CONTOURS:
	obj->draw_fcn = tactmap_draw_contours; break;
      case TACTMAP_TOWNS: obj->draw_fcn = tactmap_draw_towns; break;
      case TACTMAP_CELLS: obj->draw_fcn = tactmap_draw_cells; break;
      case TACTMAP_CRATERS: obj->draw_fcn = tactmap_draw_craters; break;
      case TACTMAP_DITCHES: obj->draw_fcn = tactmap_draw_ditches; break;
      case TACTMAP_FPOS: obj->draw_fcn = tactmap_draw_fpos; break;
      case TACTMAP_WIRE: obj->draw_fcn = tactmap_draw_wire; break;
      case TACTMAP_ROCK_DROPS: obj->draw_fcn = tactmap_draw_rock_drops; break;
      case TACTMAP_DRAGONS_TEETH: obj->draw_fcn = tactmap_draw_dragons_teeth;
	break;
      case TACTMAP_BERMS: obj->draw_fcn = tactmap_draw_berms; break;
      case TACTMAP_RUBBLE: obj->draw_fcn = tactmap_draw_rubble; break;
      case TACTMAP_SNOW: obj->draw_fcn = tactmap_draw_snow; break;
      case TACTMAP_CLOUDS: obj->draw_fcn = tactmap_draw_clouds; break;
      case TACTMAP_MICLIC_LINE: obj->draw_fcn = tactmap_draw_miclic_line; 
	break;
      case TACTMAP_MICLIC_SCORCH: obj->draw_fcn = tactmap_draw_miclic_scorch; 
	break;
      case TACTMAP_CAMOUFLAGE: obj->draw_fcn = tactmap_draw_camouflage; break;
      case TACTMAP_HYDRO_SURFACE:
	obj->draw_fcn = tactmap_draw_hydro_surface; break;
      case TACTMAP_HYDRO_SUBSURFACE:
	obj->draw_fcn = tactmap_draw_hydro_subsurface; break;
      case TACTMAP_GRIDS: obj->draw_fcn = tactmap_draw_grids; break;
      case TACTMAP_LINEAR: obj->draw_fcn = tactmap_draw_linear; break;
      case TACTMAP_PIXMAP: obj->draw_fcn = tactmap_draw_pixmap; break;
      case TACTMAP_PICTURE: obj->draw_fcn = tactmap_draw_picture; break;
      case TACTMAP_GLOBAL_EFFECT:obj->draw_fcn = tactmap_draw_global;break;
      case TACTMAP_DRAWN: obj->draw_fcn = tactmap_draw_drawn; break;
      case TACTMAP_TEXT: obj->draw_fcn = tactmap_draw_text; break;
      case TACTMAP_BGR:  obj->draw_fcn = tactmap_draw_bgr; break;
      case TACTMAP_LINE_INTERVIS:  obj->draw_fcn = tactmap_draw_line_iv; break;
      case TACTMAP_AREA_INTERVIS:  obj->draw_fcn = tactmap_draw_area_iv; break;
      case TACTMAP_LINE_ENV_EFFECT:
	obj->draw_fcn = tactmap_draw_line_effect; break;
      default:
	printf("LIBTACTMAP ERROR: Unrecognized object sensClass %d\n", sensClass);
	STDDEALLOC(obj);
	return NULL;
    }

    obj->sensClass = sensClass;
    obj->dfn = *dfn;
    if (sensClass == TACTMAP_LINEAR)
    {
	if (dfn->linear.n_vertices)
	{
	    obj->dfn.linear.vertices =
	      (float64 *)STDALLOC(2*dfn->linear.n_vertices*sizeof(float64));
	    bcopy((char *)dfn->linear.vertices,
		  (char *)obj->dfn.linear.vertices,
		  (long) (2 * dfn->linear.n_vertices * sizeof(float64)));
	}
	else
	  obj->dfn.linear.vertices = NULL;
    }
    else if (sensClass == TACTMAP_TEXT)
    {
	if (dfn->text.string)
	{
	    len = strlen(dfn->text.string) + 1;
	    obj->dfn.text.string = (char *)STDALLOC(len);
	    bcopy((char *)dfn->text.string,
		  (char *)obj->dfn.text.string,
		  len);
	}
	else
	  obj->dfn.text.string = NULL;
    }
    else if (sensClass == TACTMAP_PICTURE)
    {
	if (dfn->picture.string)
	{
	    len = strlen(dfn->picture.string) + 1;
	    obj->dfn.picture.string = (char *)STDALLOC(len);
	    bcopy((char *)dfn->picture.string,
		  (char *)obj->dfn.picture.string,
		  len);
	}
	else
	  obj->dfn.picture.string = NULL;

	if (!dfn->picture.dfn && dfn->picture.instructions)
	{
	    len = dfn->picture.num_instructions *
	      sizeof(struct tactmap_picture_instruction);
	    obj->dfn.picture.instructions =
	      (struct tactmap_picture_instruction *)STDALLOC(len);
	    bcopy((char *)dfn->picture.instructions,
		  (char *)obj->dfn.picture.instructions,
		  len);
	}
	else
	  obj->dfn.picture.instructions = NULL;
    }
    else if (sensClass == TACTMAP_GLOBAL_EFFECT)
    {
	if (dfn->global_effect.string)
	{
	    len = strlen(dfn->global_effect.string) + 1;
	    obj->dfn.global_effect.string = (char *)STDALLOC(len);
	    bcopy((char *)dfn->global_effect.string,
		  (char *)obj->dfn.global_effect.string,
		  len);
	}
	else
	  obj->dfn.global_effect.string = NULL;

	if (!dfn->global_effect.dfn && dfn->global_effect.instructions)
	{
	    len = dfn->global_effect.num_instructions *
	      sizeof(struct tactmap_global_effect_instruction);
	    obj->dfn.global_effect.instructions =
	      (struct tactmap_global_effect_instruction *)STDALLOC(len);
	    bcopy((char *)dfn->global_effect.instructions,
		  (char *)obj->dfn.global_effect.instructions,
		  len);
	}
	else
	  obj->dfn.global_effect.instructions = NULL;
    }

    return obj;
}

void tactmap_destroy_object(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR obj)
{
    tactmap_remove_object(tactmap, obj);

    if (obj->sensClass == TACTMAP_LINEAR)
    {
	if (obj->dfn.linear.n_vertices)
	  STDDEALLOC(obj->dfn.linear.vertices);
    }
    else if (obj->sensClass == TACTMAP_TEXT)
    {
	if (obj->dfn.text.string)
	  STDDEALLOC(obj->dfn.text.string);
    }
    else if (obj->sensClass == TACTMAP_PICTURE)
    {
	if (obj->dfn.picture.string)
	  STDDEALLOC(obj->dfn.picture.string);
	if (obj->dfn.picture.instructions)
	  STDDEALLOC(obj->dfn.picture.instructions);
    }
    else if (obj->sensClass == TACTMAP_GLOBAL_EFFECT)
    {
	if (obj->dfn.global_effect.string)
	  STDDEALLOC(obj->dfn.global_effect.string);
	if (obj->dfn.global_effect.instructions)
	  STDDEALLOC(obj->dfn.global_effect.instructions);
    }

    STDDEALLOC(obj);
}

void tactmap_add_object(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR obj)
{
    TACTMAP_BOUNDS bounds;

    if (!obj)
      return;

    if (obj->display_mode == PERMANENT)
      return;

    if (obj->display_mode == NOT_DISPLAYED)
    {
	obj->prev = NULL;
	obj->next = tactmap->objects[obj->sensClass];

	if (obj->next)
	  obj->next->prev = obj;

	tactmap->objects[obj->sensClass] = obj;
    }

    obj->display_mode = PERMANENT;

    /* When we come out of world_view_mode, we'll redraw automagically */
    if (tactmap->world_view_mode)
      return;

    /* Register a refresh unless the feature is about to be drawn
     * already.
     */
    if (!tactmap->do_refresh &&
	(!tactmap->redrawing || (tactmap->current_class > obj->sensClass)))
    {
	if (TACTMAP_MAP_FEATURE(obj->sensClass))
	{
	    /* When we add a feature, we can shortcut the redraw
	     * process, by skipping the refresh, and starting with
	     * the new feature.
	     */
	    if (tactmap->refresh_callback)
	      tactmap->refresh_callback(tactmap, 1);

	    tactmap->current_class = obj->sensClass;
	    tactmap->redrawing = 1;
	}
	else
	{
	    bounds.min_x = tactmap->cur->cx - tactmap->cur->hsw;
	    bounds.max_x = tactmap->cur->cx + tactmap->cur->hsw;
	    bounds.min_y = tactmap->cur->cy - tactmap->cur->hsh;
	    bounds.max_y = tactmap->cur->cy + tactmap->cur->hsh;

	    /* For non-terrain-features, we just draw the thing right away */
	    obj->draw_fcn(tactmap, obj,
			  FALSE, &bounds, NULL, TRUE, NULL);
	}
    }
}

void tactmap_add_transient_object(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR obj)
{
    TACTMAP_BOUNDS bounds;

    if (!obj)
      return;

    if (TACTMAP_MAP_FEATURE(obj->sensClass))
    {
	fprintf(stderr, "LIBTACTMAP ERROR: Cannot draw terrain features as transients\n");
	return;
    }

    if (obj->display_mode == TRANSIENT)
      return;

    if (obj->display_mode == PERMANENT)
    {
	/* We can't make a permanent object transient without erasing
	 * it first.
	 */
	tactmap_remove_object(tactmap, obj);
    }

    obj->display_mode = TRANSIENT;

    obj->prev = NULL;
    obj->next = tactmap->objects[obj->sensClass];
    if (obj->next)
      obj->next->prev = obj;
    tactmap->objects[obj->sensClass] = obj;

    /* When we come out of world_view_mode, we'll redraw automagically */
    if (tactmap->world_view_mode)
      return;

    bounds.min_x = tactmap->cur->cx - tactmap->cur->hsw;
    bounds.max_x = tactmap->cur->cx + tactmap->cur->hsw;
    bounds.min_y = tactmap->cur->cy - tactmap->cur->hsh;
    bounds.max_y = tactmap->cur->cy + tactmap->cur->hsh;

    /* Draw it unless its about to be drawn (transients are drawn
     * in one swell foop after everything else)
     */
    if (!tactmap->do_refresh && !tactmap->redrawing)
      obj->draw_fcn(tactmap, obj, FALSE, &bounds, NULL, TRUE, NULL);
}

static void refresh_transients(TACTMAP_PTR tactmap, TACTMAP_BOUNDS *bounds)
{
    TACTMAP_OBJECT_CLASS sensClass;
    TACTMAP_OBJECT_PTR obj;

    for (sensClass=TACTMAP_FIRST_DYNAMIC_CLASS;sensClass<=TACTMAP_LAST_CLASS;sensClass++)
      for (obj=tactmap->objects[sensClass];obj;obj=obj->next)
	if (obj->display_mode == TRANSIENT)
	  obj->draw_fcn(tactmap, obj, FALSE, bounds,
			NULL, FALSE, &obj->dfn);
}

void tactmap_remove_object(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR obj)
{
    TACTMAP_BOUNDS bounds;

    if (!obj)
      return;

    if (obj->display_mode == NOT_DISPLAYED)
      return;

    if (obj->next)
      obj->next->prev = obj->prev;
    if (obj->prev)
      obj->prev->next = obj->next;
    if (obj == tactmap->objects[obj->sensClass])
      tactmap->objects[obj->sensClass] = obj->next;

    obj->next = obj->prev = NULL;

    /* When we come out of world_view_mode, we'll redraw automagically */
    if (tactmap->world_view_mode)
      return;

    /* Register a refresh unless the feature is about to be drawn
     * (which will not happen, now).  Note that transients are drawn
     * after everything else.
     */
    if (!tactmap->do_refresh &&
	(!tactmap->redrawing ||
	 ((obj->display_mode == PERMANENT) &&
	  (tactmap->current_class > obj->sensClass))))
    {
	if (TACTMAP_MAP_FEATURE(obj->sensClass))
	  tactmap->do_refresh = 1;
	else
	{
	    bounds.min_x = tactmap->cur->cx - tactmap->cur->hsw;
	    bounds.max_x = tactmap->cur->cx + tactmap->cur->hsw;
	    bounds.min_y = tactmap->cur->cy - tactmap->cur->hsh;
	    bounds.max_y = tactmap->cur->cy + tactmap->cur->hsh;

	    /* For non-terrain-features, we just erase the thing */
	    obj->draw_fcn(tactmap, obj,
			  TRUE, &bounds, NULL, TRUE, NULL);

	    if (obj->display_mode == PERMANENT)
	      refresh_transients(tactmap, &bounds);
	}
    }

    obj->display_mode = NOT_DISPLAYED;
}

void tactmap_update_object(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR obj, 
			   TACTMAP_OBJECT_DFN *dfn)
{
    int32 old_count, old_len;
    float64 *old_verts;
    char *old_string;
    struct tactmap_picture_instruction *old_inst;
    struct tactmap_global_effect_instruction *old_effect;
    int32 len;
    TACTMAP_BOUNDS bounds;

    if (!obj)
      return;

    if (obj->display_mode != NOT_DISPLAYED)
    {
	/* When we come out of world_view_mode, we'll redraw automagically */
	if (!tactmap->world_view_mode)
	{
	    bounds.min_x = tactmap->cur->cx - tactmap->cur->hsw;
	    bounds.max_x = tactmap->cur->cx + tactmap->cur->hsw;
	    bounds.min_y = tactmap->cur->cy - tactmap->cur->hsh;
	    bounds.max_y = tactmap->cur->cy + tactmap->cur->hsh;

	    if (!tactmap->do_refresh &&
		(!tactmap->redrawing ||
		 ((obj->display_mode == PERMANENT) &&
		  (tactmap->current_class > obj->sensClass))))
	    {
		if (TACTMAP_MAP_FEATURE(obj->sensClass))
		{
		    if (redraw_required(obj->sensClass, &obj->dfn, dfn))
		      tactmap->do_refresh = 1;
		}
		else
		{
		    obj->draw_fcn(tactmap, obj,
				  FALSE, &bounds, NULL, TRUE, dfn);

		    if (obj->display_mode == PERMANENT)
		      refresh_transients(tactmap, &bounds);
		}
	    }
	}
    }

    /* Copy in the new definition regardless of whether the object is
     * currently being displayed.
     */

    if (obj->sensClass == TACTMAP_LINEAR)
    {
	old_count = obj->dfn.linear.n_vertices;
	old_verts = obj->dfn.linear.vertices;
    }
    else if (obj->sensClass == TACTMAP_TEXT)
    {
	old_string = obj->dfn.text.string;
	old_count = old_string ? (strlen(old_string) + 1) : 0;
    }
    else if (obj->sensClass == TACTMAP_PICTURE)
    {
	old_string = obj->dfn.picture.string;
	old_count = old_string ? (strlen(old_string) + 1) : 0;
	if (obj->dfn.picture.dfn)
	  old_len = 0, old_inst = NULL;
	else
	{
	    old_len = obj->dfn.picture.num_instructions *
	      sizeof(struct tactmap_picture_instruction);
	    old_inst = obj->dfn.picture.instructions;
	}
    }
    else if (obj->sensClass == TACTMAP_GLOBAL_EFFECT)
    {
	old_string = obj->dfn.global_effect.string;
	old_count = old_string ? (strlen(old_string) + 1) : 0;
	if (obj->dfn.global_effect.dfn)
	  old_len = 0, old_effect = NULL;
	else
	{
	    old_len = obj->dfn.global_effect.num_instructions *
	      sizeof(struct tactmap_global_effect_instruction);
	    old_effect = obj->dfn.global_effect.instructions;
	}
    }

    obj->dfn = *dfn;

    if (obj->sensClass == TACTMAP_LINEAR)
    {
	if (dfn->linear.n_vertices == old_count)
	  obj->dfn.linear.vertices = old_verts;
	else
	{
	    if (old_verts)
	      STDDEALLOC(old_verts);
	    if (dfn->linear.n_vertices)
	      obj->dfn.linear.vertices =
		(float64 *)STDALLOC(2*dfn->linear.n_vertices*sizeof(float64));
	    else
	      obj->dfn.linear.vertices = NULL;
	}
	if (obj->dfn.linear.vertices)
	  bcopy((char *)dfn->linear.vertices,
		(char *)obj->dfn.linear.vertices,
		(long)(2 * dfn->linear.n_vertices * sizeof(float64)));
    }
    else if (obj->sensClass == TACTMAP_TEXT)
    {
	len = dfn->text.string ? (strlen(dfn->text.string) + 1) : 0;

	if (len == old_count)
	  obj->dfn.text.string = old_string;
	else
	{
	    if (old_string)
	      STDDEALLOC(old_string);
	    if (dfn->text.string)
	      obj->dfn.text.string = (char *)STDALLOC(len);
	    else
	      obj->dfn.text.string = NULL;
	}
	if (obj->dfn.text.string)
	  bcopy((char *)dfn->text.string,
		(char *)obj->dfn.text.string,
		len);
    }
    else if (obj->sensClass == TACTMAP_PICTURE)
    {
	len = dfn->picture.string ? (strlen(dfn->picture.string) + 1) : 0;

	if (len == old_count)
	  obj->dfn.picture.string = old_string;
	else
	{
	    if (old_string)
	      STDDEALLOC(old_string);
	    if (dfn->picture.string)
	      obj->dfn.picture.string = (char *)STDALLOC(len);
	    else
	      obj->dfn.picture.string = NULL;
	}
	if (obj->dfn.picture.string)
	  bcopy((char *)dfn->picture.string,
		(char *)obj->dfn.picture.string,
		len);

	if (!dfn->picture.dfn)
	{
	    len = dfn->picture.num_instructions *
	      sizeof(struct tactmap_picture_instruction);

	    if (len == old_len)
	      obj->dfn.picture.instructions = old_inst;
	    else
	    {
		if (old_inst)
		  STDDEALLOC(old_inst);
		if (dfn->picture.instructions)
		  obj->dfn.picture.instructions =
		    (struct tactmap_picture_instruction *)STDALLOC(len);
		else
		  obj->dfn.picture.instructions = NULL;
	    }
	    if (obj->dfn.picture.instructions)
	      bcopy((char *)dfn->picture.instructions,
		    (char *)obj->dfn.picture.instructions,
		    len);
	}
    }
    else if (obj->sensClass == TACTMAP_GLOBAL_EFFECT)
    {
	len = dfn->global_effect.string ? (strlen(dfn->global_effect.string) + 1) : 0;

	if (len == old_count)
	  obj->dfn.global_effect.string = old_string;
	else
	{
	    if (old_string)
	      STDDEALLOC(old_string);
	    if (dfn->global_effect.string)
	      obj->dfn.global_effect.string = (char *)STDALLOC(len);
	    else
	      obj->dfn.global_effect.string = NULL;
	}
	if (obj->dfn.global_effect.string)
	  bcopy((char *)dfn->global_effect.string,
		(char *)obj->dfn.global_effect.string,
		len);

	if (!dfn->global_effect.dfn)
	{
	    len = dfn->global_effect.num_instructions *
	      sizeof(struct tactmap_global_effect_instruction);

	    if (len == old_len)
	      obj->dfn.global_effect.instructions = old_effect;
	    else
	    {
		if (old_effect)
		  STDDEALLOC(old_effect);
		if (dfn->global_effect.instructions)
		  obj->dfn.global_effect.instructions =
		    (struct tactmap_global_effect_instruction *)STDALLOC(len);
		else
		  obj->dfn.global_effect.instructions = NULL;
	    }
	    if (obj->dfn.global_effect.instructions)
	      bcopy((char *)dfn->global_effect.instructions,
		    (char *)obj->dfn.global_effect.instructions,
		    len);
	}
    }

}

void tactmap_refresh(TACTMAP_PTR tactmap)
{
    /* No need to do this in the midst (or just prior to) a refresh */
    if (tactmap->redrawing || tactmap->do_refresh)
      return;

    if(tactmap)
      tactmap->do_refresh = TRUE;
}

void tactmap_refresh_non_terrain(TACTMAP_PTR tactmap)
{
    TACTMAP_OBJECT_CLASS sensClass;
    TACTMAP_BOUNDS bounds;
    TACTMAP_OBJECT_PTR obj;

    /* No need to do this in the midst (or just prior to) a refresh */
    if (tactmap->redrawing || tactmap->do_refresh)
      return;

    if (tactmap->refresh_callback)
      tactmap->refresh_callback(tactmap, 1);
    tactmap->redrawing = 1;

    bounds.min_x = tactmap->cur->cx - tactmap->cur->hsw;
    bounds.max_x = tactmap->cur->cx + tactmap->cur->hsw;
    bounds.min_y = tactmap->cur->cy - tactmap->cur->hsh;
    bounds.max_y = tactmap->cur->cy + tactmap->cur->hsh;

    for (sensClass=TACTMAP_FIRST_DYNAMIC_CLASS;sensClass<=TACTMAP_LAST_CLASS;sensClass++)
      for (obj=tactmap->objects[sensClass];obj;obj=obj->next)
	if (obj->display_mode == PERMANENT)
	  obj->draw_fcn(tactmap, obj, FALSE, &bounds,
			NULL, FALSE, &obj->dfn);

    XClearWindow(DPYWIN);

    refresh_transients(tactmap, &bounds);

    tactmap->redrawing = 0;
    if (tactmap->refresh_callback)
      tactmap->refresh_callback(tactmap, 0);
}

void tactmap_draw_class(TACTMAP_PTR tactmap, TACTMAP_OBJECT_CLASS sensClass, 
			enum tactmap_display_mode mode,
			TACTMAP_BOUNDS *bounds, CTDB_SEARCH_SPACE_PTR space)
{
    TACTMAP_OBJECT_PTR obj;

    for (obj=tactmap->objects[sensClass];obj;obj=obj->next)
      if (obj->display_mode == mode)
	obj->draw_fcn(tactmap, obj, FALSE, bounds, space, FALSE, NULL);
}

static int32 redraw_required(TACTMAP_OBJECT_CLASS sensClass, 
			     TACTMAP_OBJECT_DFN *old, TACTMAP_OBJECT_DFN *new)
{
    switch(sensClass)
    {
      case TACTMAP_HYPSO:
	return old->hypso.kind != new->hypso.kind;

      case TACTMAP_TREES:
	return old->trees.mask != new->trees.mask;

      case TACTMAP_CONTOURS:
	return ((old->contours.minor_spacing!=new->contours.minor_spacing) ||
		(old->contours.major_spacing!=new->contours.major_spacing));

      case TACTMAP_GRIDS:
	return ((old->grids.interval != new->grids.interval) ||
		(old->grids.label_mask != new->grids.label_mask));

      default:
	return 0;
    }
}

int32 tactmap_obj_display_on (TACTMAP_OBJECT_PTR obj)
{

   if (obj == NULL)
      return (FALSE);

   if (obj->display_mode == NOT_DISPLAYED)
       return (FALSE);
   else
       return (TRUE);
}

#endif /*USE_X*/
