#ifndef lint
static char rcsid [] = "$RCSfile: tmap_picture.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_picture.c                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw pictures                                     *
*       Created: Mon Sep 28 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <libreader.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define ROUNDED 0.5

static void draw_picture(TACTMAP_PTR tactmap, Pixmap pixmap, 
			 int32 use_erase_gc, struct tactmap_picture *picture,
			 XRectangle *bounds);

/*ARGSUSED*/
void 
tactmap_draw_picture(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		     int32 use_erase_gc, TACTMAP_BOUNDS *bounds, 
		     CTDB_SEARCH_SPACE_PTR space,
		     int32 clear_area_when_finished, TACTMAP_OBJECT_DFN *dfn)
                                  
                                 
                                       
                                 
                                
                                                   
                              /* new version of the object */
{
    int32 change=FALSE, old_x, old_y;
    SNSTVE_OBJECT snsobj;
    Drawable d;

    if ( dfn )
    {
	/* Erase the last incarnation */
	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    if (object->display_mode == PERMANENT)
	    {
		XCopyArea(XtDisplay(tactmap->widget),
			  tactmap->terrain_pix, tactmap->background_pix,
			  tactmap->gc.background,
			  object->drawn_bounds.x, object->drawn_bounds.y,
			  object->drawn_bounds.width,
			  object->drawn_bounds.height,
			  object->drawn_bounds.x, object->drawn_bounds.y);

		if (clear_area_when_finished)
		  XClearArea(DPYWIN,
			     object->drawn_bounds.x, object->drawn_bounds.y,
			     object->drawn_bounds.width,
			     object->drawn_bounds.height,
			     0);
	    }
	    else
	      if (bcmp((char *)&dfn->picture, (char *)&object->dfn.picture,
		       sizeof(struct tactmap_picture)))
		tactmap_draw_picture(tactmap, object, (int32) TRUE, bounds,
				     space, clear_area_when_finished,
				     (TACTMAP_OBJECT_DFN *) NULL);

	    change = TRUE;
	    old_x = object->drawn_bounds.x;
	    old_y = object->drawn_bounds.y;
	    
	}
    }

    if (!dfn)
      dfn = &object->dfn;
    

    if ((!dfn->picture.dfn && !dfn->picture.instructions) ||
	(dfn->picture.num_csystems < 1) ||
	(dfn->picture.num_gcs < 1))
    {
	if (change)
	  snstve_remove_object(tactmap->sensitive, old_x, old_y,
			       dfn->sensitive.sensClass, dfn->sensitive.instance);
	return;
    }

    if (use_erase_gc)
    {
	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    if (object->display_mode == PERMANENT)
	      XCopyArea(XtDisplay(tactmap->widget),
			tactmap->terrain_pix, tactmap->background_pix,
			tactmap->gc.background,
			object->drawn_bounds.x, object->drawn_bounds.y,
			object->drawn_bounds.width,
			object->drawn_bounds.height,
			object->drawn_bounds.x, object->drawn_bounds.y);
	    else
	    {
		XRectangle dummy_bounds;
		draw_picture(tactmap, XtWindow(tactmap->widget), use_erase_gc,
			     &dfn->picture, &dummy_bounds);
	    }

	}

	snstve_remove_object(tactmap->sensitive,
			     object->drawn_bounds.x,
			     object->drawn_bounds.y,
			     dfn->sensitive.sensClass, dfn->sensitive.instance);
    }
    else
    {
	/* Drawing the object on the tactmap display */
	/* Adjust this coordinate system to be relative to the current
	 * tactmap context.
	 */
	CHANGE_CONTEXT(dfn->picture.csystems[0].cell,
		       dfn->picture.csystems[0].origin[X],
		       dfn->picture.csystems[0].origin[Y],
		       tactmap->cur->cc);

	if ((dfn->picture.csystems[0].origin[X] < bounds->min_x) ||
	    (dfn->picture.csystems[0].origin[Y] < bounds->min_y) ||
	    (dfn->picture.csystems[0].origin[X] > bounds->max_x) ||
	    (dfn->picture.csystems[0].origin[Y] > bounds->max_y))
	  object->drawn_bounds.width = object->drawn_bounds.height = 0;
	else
	{
	    if (object->display_mode == PERMANENT)
	      d = tactmap->background_pix;
	    else
	      d = XtWindow(tactmap->widget);

	    draw_picture(tactmap, d, use_erase_gc,
			 &dfn->picture, &object->drawn_bounds);
	}

	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    snsobj.sensClass            = dfn->sensitive.sensClass;
	    snsobj.instance         = dfn->sensitive.instance;
	    snsobj.user_data        = dfn->sensitive.user_data;
	    snsobj.kind             = SNSTVE_RECTANGLE;
	    snsobj.attr.rect.x      = object->drawn_bounds.x-2;
	    snsobj.attr.rect.y      = object->drawn_bounds.y-2;
	    snsobj.attr.rect.width  = object->drawn_bounds.width+4;
	    snsobj.attr.rect.height = object->drawn_bounds.height+4;

	    if (!change ||
		!snstve_change_object(tactmap->sensitive,
				      old_x, old_y, &snsobj))
	      snstve_add_object(tactmap->sensitive, &snsobj);
	}
	else
	{
	    if (change)
	      snstve_remove_object(tactmap->sensitive, old_x, old_y,
				   dfn->sensitive.sensClass,
				   dfn->sensitive.instance);
	}
    }

    if (clear_area_when_finished &&
	object->drawn_bounds.width && object->drawn_bounds.height &&
	(object->display_mode == PERMANENT))
      XClearArea(DPYWIN, object->drawn_bounds.x, object->drawn_bounds.y,
		 object->drawn_bounds.width, object->drawn_bounds.height, 0);
}

static void 
check_extents(int32 x, int32 y, int32 w, XRectangle *bounds)
{
    int32 old_x = bounds->x;
    int32 old_y = bounds->y;
    int32 old_w = bounds->width;
    int32 old_h = bounds->height;

    w += 1;
    if (x-w < old_x)
    {
	bounds->width = old_w + old_x - (x-w);
	bounds->x = x-w;
    }
    else if (x+w > old_x+old_w)
    {
	bounds->width = x+w - old_x;
    }

    if (y-w < old_y)
    {
	bounds->height = old_h + old_y - (y-w);
	bounds->y = y-w;
    }
    else if (y+w > old_y+old_h)
    {
	bounds->height = y+w - old_y;
    }
}

static void 
draw_circle(Display *display, Pixmap pixmap, XRectangle *bounds, GC interior, 
	    GC border, float64 *reference, float64 *dimensions, float64 scale,
	    int32 cx, int32 cy, uint32 dashed_p, float32 width,
	    float32 center_x, float32 center_y, float32 radius, char rcoord)
{
    int32 x, y, r, w;
    int32 line_style = (dashed_p) ? LineOnOffDash : LineSolid;

    x = cx + (reference[X] * dimensions[X] * center_x -
	      reference[Y] * dimensions[Y] * center_y) + ROUNDED;
    y = cy - (reference[Y] * dimensions[X] * center_x +
	      reference[X] * dimensions[Y] * center_y) + ROUNDED;
    w = width * scale;

    if (w < 1) 
      w = 1;

    if (rcoord == 'x')
      r = radius * dimensions[X] + ROUNDED;
    else if (rcoord == 'y')
      r = radius * dimensions[Y] + ROUNDED;
    else
      r = radius * scale + ROUNDED;

    check_extents(x-r, y-r, (int32) (border?w+2:w), bounds);
    check_extents(x+r, y+r, (int32) (border?w+2:w), bounds);

    if (border)
    {
	XSetLineAttributes(display,border,w+2,line_style,CapButt,JoinBevel);
	XDrawArc(display, pixmap, border, x-r, y-r, 2*r, 2*r, 0, 360*64);
    }
    if (interior)
    {
	XSetLineAttributes(display,interior,w,line_style,CapButt,JoinBevel);
	XDrawArc(display, pixmap, interior, x-r, y-r, 2*r, 2*r, 0, 360*64);
    }
}

static void 
draw_disc(Display *display, Pixmap pixmap, XRectangle *bounds, GC interior,
	  GC border, float64 *reference, float64 *dimensions, float64 scale, 
	  int32 cx, int32 cy, uint32 dashed_p, float32 center_x, 
	  float32 center_y, float32 radius, char rcoord)
{
    int32 x, y, r;
    int32 line_style = (dashed_p) ? LineOnOffDash : LineSolid;

    x = cx + (reference[X] * dimensions[X] * center_x -
	      reference[Y] * dimensions[Y] * center_y) + ROUNDED;
    y = cy - (reference[Y] * dimensions[X] * center_x +
	      reference[X] * dimensions[Y] * center_y) + ROUNDED;
    if (rcoord == 'x')
      r = radius * dimensions[X] + ROUNDED;
    else if (rcoord == 'y')
      r = radius * dimensions[Y] + ROUNDED;
    else
      r = radius * scale + ROUNDED;

    check_extents(x-r, y-r, (int32) (border?3:1), bounds);
    check_extents(x+r, y+r, (int32) (border?3:1), bounds);

    if (border)
    {
	XSetLineAttributes(display,border,1,line_style,CapButt,JoinBevel);
	XDrawArc(display,pixmap,border, x-r-1, y-r-1, 2*r+2, 2*r+2, 0,360*64);
    }
    if (interior)
    {
	XSetLineAttributes(display,interior,1,line_style,CapButt,JoinBevel);
	XFillArc(display, pixmap, interior, x-r, y-r, 2*r, 2*r, 0, 360*64);
    }
}

static void 
draw_line(Display *display, Pixmap pixmap, XRectangle *bounds, GC interior, 
	  GC border, float64 *reference, float64 *dimensions, float64 scale, 
	  int32 cx, int32 cy, uint32 dashed_p, float32 width, int32 n_points,
	  float32 *points)
{
    int32 w, i, j;
    float64 px, py;
    int16 x, y, pts[1000];
    int32 line_style = (dashed_p) ? LineOnOffDash : LineSolid;

    w = width * scale;
    if (w < 2) 
      w = 2;

    for (i=0,j=0;i<n_points;i+=2)
    {
	px = points[i] * dimensions[X];
	py = points[i+1] * dimensions[Y];
	x = cx + (reference[X] * px - reference[Y] * py) + ROUNDED;
	y = cy - (reference[Y] * px + reference[X] * py) + ROUNDED;

	if (!j || (x != pts[j-2]) || (y != pts[j-1]))
	{
	    pts[j++] = x;
	    pts[j++] = y;

	    check_extents((int32)x,(int32)y, (int32)(border?w+2:2), bounds);
	}
    }

    if (j < 4)
      return;

    if (border)
    {
	XSetLineAttributes(display,border,w+2,line_style,CapButt,JoinBevel);
	XDrawLines(display, pixmap, border, (XPoint *)pts, j/2,
		   CoordModeOrigin); 
    }
    if (interior)
    {
	XSetLineAttributes(display,interior,w,line_style,CapButt,JoinBevel);
	XDrawLines(display,pixmap,interior, (XPoint *)pts, j/2,
		   CoordModeOrigin); 
    }
}

/*ARGSUSED*/
static void 
draw_block(Display *display, Pixmap pixmap, XRectangle *bounds, GC interior,
	   GC border, float64 *reference, float64 *dimensions, float64 scale,
	   int32 cx, int32 cy, uint32 dashed_p, int32 n_points, 
	   float32 *points)
{
    int32 i, j;
    float64 px, py;
    int16 x, y, pts[1000];
    int32 line_style = (dashed_p) ? LineOnOffDash : LineSolid;

    for (i=0,j=0;i<n_points;i+=2)
    {
	px = points[i] * dimensions[X];
	py = points[i+1] * dimensions[Y];
	x = cx + (reference[X] * px - reference[Y] * py) + ROUNDED;
	y = cy - (reference[Y] * px + reference[X] * py) + ROUNDED;

	if (!j || (x != pts[j-2]) || (y != pts[j-1]))
	{
	    pts[j++] = x;
	    pts[j++] = y;

	    check_extents((int32)x,(int32)y, (int32)(border?2:1),bounds);
	}
    }

    if (j < 4)
      return;

    if ((pts[j-2] != pts[0]) || (pts[j-1] != pts[1]))
    {
	pts[j++] = pts[0];
	pts[j++] = pts[1];
    }

    if (border)
    {
	XSetLineAttributes(display,border,2,line_style,CapButt,JoinBevel);
	XDrawLines(display, pixmap, border, (XPoint *)pts, j/2,
		   CoordModeOrigin); 
    }
    if (interior)
    {
	XSetLineAttributes(display,interior,1,line_style,CapButt,JoinBevel);
	XFillPolygon(display, pixmap, interior, (XPoint *)pts, j/2,
		     Nonconvex, CoordModeOrigin);
    }
}

/*ARGSUSED*/
static void 
draw_label(Display *display, Pixmap pixmap, XRectangle *bounds, GC interior,
	   GC border, float64 *reference, float64 *dimensions, float64 scale,
	   int32 cx, int32 cy, uint32 dashed_p, float32 origin_x,
	   float32 origin_y, char *label)
{
    int32 x, y, w, h;
    int ign, ascent, descent;
    XFontStruct *xfs;
    XCharStruct result;

    if (!label || !*label)
      return;

    xfs = NULL; /* used to check whether to free or not at end of rtn */

    xfs = XQueryFont(display, XGContextFromGC(interior));

    x = cx + (reference[X] * dimensions[X] * origin_x -
	      reference[Y] * dimensions[Y] * origin_y) + ROUNDED;
    y = cy - (reference[Y] * dimensions[X] * origin_x +
	      reference[X] * dimensions[Y] * origin_y) + ROUNDED;

    XTextExtents(xfs, label, strlen(label),
		 &ign, &ascent, &descent, &result);

    w = result.width;
    h = result.descent + result.ascent;

    x -= w / 2;
    y += h / 2;

    check_extents(x, y, (int32)1, bounds);
    check_extents(x+w, y-h, (int32)1, bounds);

    XDrawString(display,pixmap,interior, x, y,
		label, strlen(label));
    if (xfs)
      XFreeFontInfo((char **)NULL, xfs, 1);
} 

static void draw_picture(TACTMAP_PTR tactmap, Pixmap pixmap,
			 int32 use_erase_gc, struct tactmap_picture *picture,
			 XRectangle *bounds)
{
    DECLARE_REGS;
    int32 num_inst;
    int32 i, j;
    GC interior, border;
    int32 coord;
    char *type;
    int32 center_x, center_y;
    float64 *ref, dim[XY], scale;
    float64 rotations[TACTMAP_MAX_CSYSTEMS][XY];
    float64 offsets[TACTMAP_MAX_CSYSTEMS][XY];
    float64 mag;
    int32 in_pixels;
    float64 x, y;
    int32 p;
    static char * block_sym  = NULL;
    static char * line_sym   = NULL;
    static char * disc_sym   = NULL;
    static char * circle_sym = NULL;
    static char * label_sym  = NULL;

    if (!block_sym)
    {
	block_sym  = reader_get_symbol("block");
	line_sym   = reader_get_symbol("line");
	disc_sym   = reader_get_symbol("disc");
	circle_sym = reader_get_symbol("circle");
	label_sym  = reader_get_symbol("label");
    }

    if (picture->magnification != 0.0)
    {
	in_pixels = 0;
	mag = picture->magnification;
    }
    else
    {
	in_pixels = 1;
	mag = 1.0;
    }

    for (i=0;i<picture->num_csystems;i++)
    {
	/* Is this world-parented? */
	if (picture->csystems[i].parent_system >= i)
	{
	    /* Parent Coordinate Systems must all be parented
	     * to the tactmap coordinate system.
	     */
	    CHANGE_CONTEXT(picture->csystems[i].cell,
			   picture->csystems[i].origin[X],
			   picture->csystems[i].origin[Y],
			   cc);
	    
	    offsets[i][X]   = picture->csystems[i].origin[X];
	    offsets[i][Y]   = picture->csystems[i].origin[Y];
	    rotations[i][X] = cos(picture->csystems[i].angle);
	    rotations[i][Y] = sin(picture->csystems[i].angle);
	}
	else
	{
	    p = picture->csystems[i].parent_system;
	    x = picture->csystems[i].origin[X] * mag;
	    y = picture->csystems[i].origin[Y] * mag;

	    /* Rotate and translate the center to world coordinates */
	    offsets[i][X] =
	      offsets[p][X] + x * rotations[p][X] - y * rotations[p][Y];
	    offsets[i][Y] =
	      offsets[p][Y] + x * rotations[p][Y] + y * rotations[p][X];

	    /* Rotate the rotation to world coordinates */
	    rotations[i][X] =
	      cos(picture->csystems[p].angle + picture->csystems[i].angle);
	    rotations[i][Y] =
	      sin(picture->csystems[p].angle + picture->csystems[i].angle);
	}
    }

    bounds->x = 
      tactmap_FX_pixel_number(picture->csystems[0].origin[X],cx,hsw,ppm);
    bounds->y = 
      tactmap_FY_pixel_number(picture->csystems[0].origin[Y],cy,hsh,ppm);
    bounds->width = bounds->height = 0;

    if (picture->dfn)
      num_inst = picture->dfn[0].integer-1;
    else if (picture->instructions)
      num_inst = picture->num_instructions;
    else /* They have to give us one or the other... */
      return;

    for (i=0;i<num_inst;i++)
    {
	READER_UNION *part;
	struct tactmap_picture_instruction *inst;
	int32 gi, gb;

	if (picture->dfn)
	{
	    part = picture->dfn[i+1].array;
	    gi = part[1].integer;
	    gb = part[2].integer;
	    coord = part[3].integer;
	}
	else
	{
	    inst = picture->instructions+i;
	    gi = inst->interior_gc_index;
	    gb = inst->border_gc_index;
	    coord = inst->coordinate_system_index;
	}

	/* Get the interior GC */
	if ((gi < 0) || (gi >= picture->num_gcs))
	  interior = NULL;
	else
	  interior = picture->gcs[gi];

	/* Get the border GC */
	if ((gb < 0) || (gb >= picture->num_gcs))
	  border = NULL;
	else
	  border = picture->gcs[gb];

	if (use_erase_gc)
	{
	    if (interior)
	      interior = tactmap->background_gc;
	    if (border)
	      border = tactmap->background_gc;
	}

	/* Check the coordinate system */
	if ((coord < 0) || (coord >= picture->num_csystems))
	  coord = 0;

	/* If not first coordinate system, transpose */
	center_x = tactmap_FX_pixel_number(offsets[coord][X],cx,hsw,ppm);
	center_y = tactmap_FY_pixel_number(offsets[coord][Y],cy,hsh,ppm);
	ref      = rotations[coord];

	if (in_pixels)
	  scale = 1.0;
	else
	  scale = ppm * mag;

	dim[X] = picture->csystems[coord].dimensions[X] * scale;
	dim[Y] = picture->csystems[coord].dimensions[Y] * scale;

	if (picture->dfn)
	  for (j=4;j<part[0].integer;j++)
	  {
	      type = part[j].array[1].charptr;
	      if (type == circle_sym)
		draw_circle(XtDisplay(tactmap->widget), pixmap,
			    bounds, interior,border,
			    ref, dim, scale, center_x, center_y,
			    picture->dashed,
			    part[j].array[2].real,
			    part[j].array[3].real,
			    part[j].array[4].real,
			    part[j].array[5].real,
			    part[j].array[6].charptr[0]);
	      else if (type == disc_sym)
		draw_disc(XtDisplay(tactmap->widget), pixmap,
			  bounds, interior, border,
			  ref, dim, scale, center_x, center_y,
			  picture->dashed,
			  part[j].array[2].real,
			  part[j].array[3].real, 
			  part[j].array[4].real,
			  part[j].array[5].charptr[0]);
	      else if (type == line_sym)
		draw_line(XtDisplay(tactmap->widget), pixmap,
			  bounds, interior, border,
			  ref, dim, scale, center_x, center_y,
			  picture->dashed,
			  part[j].array[2].real,
			  part[j].array[0].integer - 3,
			  &(part[j].array[3].real));
	      else if (type == block_sym)
		draw_block(XtDisplay(tactmap->widget), pixmap,
			   bounds, interior, border,
			   ref, dim, scale, center_x, center_y,
			   picture->dashed,
			   part[j].array[0].integer - 2,
			   &(part[j].array[2].real));
	      else if (type == label_sym)
		draw_label(XtDisplay(tactmap->widget), pixmap,
			   bounds, interior, border,
			   ref, dim, scale, center_x, center_y,
			   picture->dashed,
			   part[j].array[2].real,
			   part[j].array[3].real,
			   picture->string);

	  }
	else
	{
	    if (inst->kind == TACTMAP_PICTURE_CIRCLE)
	      draw_circle(XtDisplay(tactmap->widget), pixmap,
			  bounds, interior,border,
			  ref, dim, scale, center_x, center_y,
			  picture->dashed,
			  inst->data.circle.width,
			  inst->data.circle.center_x,
			  inst->data.circle.center_y,
			  inst->data.circle.radius,
			  inst->data.circle.unit);
	    else if (inst->kind == TACTMAP_PICTURE_DISC)
	      draw_disc(XtDisplay(tactmap->widget), pixmap,
			bounds, interior, border,
			ref, dim, scale, center_x, center_y,
			picture->dashed,
			inst->data.disc.center_x,
			inst->data.disc.center_y,
			inst->data.disc.radius,
			inst->data.disc.unit);
	    else if (inst->kind == TACTMAP_PICTURE_LINE)
	    {
		draw_line(XtDisplay(tactmap->widget), pixmap,
			  bounds, interior, border,
			  ref, dim, scale, center_x, center_y,
			  picture->dashed,
			  inst->data.line.width,
			  2*inst->data.line.num_xy_pairs,
			  inst->data.line.xy_pairs);
		
		/* Caller exceeded static array.  No MORE harm to draw
		 * it, but let user know we may be in trouble from
		 * memory stomp.
		 */
		if (2*inst->data.line.num_xy_pairs > TACTMAP_MAX_PAIRS)
		  fprintf (stderr,
			   "TactMap WARNING: %d line pairs > limit of %d\n",
			   inst->data.line.num_xy_pairs, TACTMAP_MAX_PAIRS/2);
	    }
	    else if (inst->kind == TACTMAP_PICTURE_BLOCK)
	    {
		draw_block(XtDisplay(tactmap->widget), pixmap,
			   bounds, interior, border,
			   ref, dim, scale, center_x, center_y,
			   picture->dashed,
			   2*inst->data.block.num_xy_pairs,
			   inst->data.block.xy_pairs);

		/* Caller exceeded static array.  No MORE harm to draw
		 * it, but let user know we may be in trouble from
		 * memory stomp.
		 */
		if (2*inst->data.line.num_xy_pairs > TACTMAP_MAX_PAIRS)
		if (2*inst->data.line.num_xy_pairs > TACTMAP_MAX_PAIRS)
		  fprintf (stderr,
			   "TactMap WARNING: %d block pairs > limit of %d\n",
			   inst->data.line.num_xy_pairs, TACTMAP_MAX_PAIRS/2);
	    }
	    else if (inst->kind == TACTMAP_PICTURE_LABEL)
	      draw_label(XtDisplay(tactmap->widget), pixmap,
			 bounds, interior, border,
			 ref, dim, scale, center_x, center_y,
			 picture->dashed,
			 inst->data.label.center_x,
			 inst->data.label.center_y,
			 picture->string);

	}

    }

    if (picture->highlight)
    {
	XGCValues save_gcv;

	if (use_erase_gc)
	{
	    interior = tactmap->background_gc;
	    XGetGCValues(XtDisplay(tactmap->widget), interior,
			 GCLineWidth|GCLineStyle|GCCapStyle|GCJoinStyle,
			 &save_gcv);
	}
	else
	{
	    if (picture->highlight <= picture->num_gcs)
	      interior = picture->gcs[picture->highlight-1];
	    else
	      interior = picture->gcs[0];
	}

	/* These belong in a resource file */
#define HIGHLIGHT_BORDER 4
#define HIGHLIGHT_WIDTH  2

	XSetLineAttributes(XtDisplay(tactmap->widget), interior,
			   HIGHLIGHT_WIDTH, LineOnOffDash, CapButt, JoinRound);
	  
	XDrawRectangle(XtDisplay(tactmap->widget), pixmap, interior,
		       bounds->x - HIGHLIGHT_BORDER,
		       bounds->y - HIGHLIGHT_BORDER,
		       bounds->width + 2 * HIGHLIGHT_BORDER,
		       bounds->height + 2 * HIGHLIGHT_BORDER);

	bounds->x -= (HIGHLIGHT_BORDER + HIGHLIGHT_WIDTH + 1);
	bounds->y -= (HIGHLIGHT_BORDER + HIGHLIGHT_WIDTH + 1);

	bounds->width += 2 * (HIGHLIGHT_BORDER + HIGHLIGHT_WIDTH + 1);
	bounds->height += 2 * (HIGHLIGHT_BORDER + HIGHLIGHT_WIDTH + 1);

	if (use_erase_gc)
	  XChangeGC(XtDisplay(tactmap->widget), interior,
		    GCLineWidth|GCLineStyle|GCCapStyle|GCJoinStyle,
		    &save_gcv);
    }
}

#endif
/*USE_X*/
