#ifndef lint
static char rcsid [] = "$RCSfile: tmap_pix.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_pix.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw pixmap- and functionally-defined objects     *
*       Created: Thu Sep 17 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/*ARGSUSED*/
void 
tactmap_draw_pixmap(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		    int32 use_erase_gc, TACTMAP_BOUNDS *bounds, 
		    CTDB_SEARCH_SPACE_PTR space, 
		    int32 clear_area_when_finished, TACTMAP_OBJECT_DFN *dfn)
                                  
                                 /* object data container with old data */
                                       
                                 
                                
                                                   
                              /* version of the object */
{
    DECLARE_REGS;
    int32 x, y;
    int32 change=0, old_x, old_y;
    SNSTVE_OBJECT snsobj;
    int32 should_draw;

#define SHOULD_DRAW(dfn)						\
    ((dfn)->pixmap.width && (dfn)->pixmap.height &&			\
     ((dfn)->pixmap.x+(dfn)->pixmap.width/ppm > bounds->min_x) &&	\
     ((dfn)->pixmap.x < bounds->max_x) &&				\
     ((dfn)->pixmap.y > bounds->min_y) &&				\
     ((dfn)->pixmap.y-(dfn)->pixmap.height/ppm < bounds->max_y))

    /* Change the context of the old information if the tactmap center
     * cell has changed since the last rendering.
     */
    CHANGE_CONTEXT(object->dfn.pixmap.cell,object->dfn.pixmap.x,
		   object->dfn.pixmap.y,cc);
    
    if (dfn && SHOULD_DRAW(&object->dfn))
    {
	/* Erase the last incarnation */
	x = tactmap_FX_pixel_number(object->dfn.pixmap.x,cx,hsw,ppm);
	y = tactmap_FY_pixel_number(object->dfn.pixmap.y,cy,hsh,ppm);

	if (object->display_mode == PERMANENT)
	{
	    XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		      tactmap->background_pix, tactmap->gc.background,
		      x, y,
		      object->dfn.pixmap.width, object->dfn.pixmap.height,
		      x, y);

	    if (clear_area_when_finished)
	      XClearArea(DPYWIN,
			 x, y, dfn->pixmap.width, dfn->pixmap.height, 0);
	}
	else
	  XCopyArea(XtDisplay(tactmap->widget), tactmap->background_pix,
		    XtWindow(tactmap->widget), tactmap->gc.background,
		    x, y,
		    object->dfn.pixmap.width, object->dfn.pixmap.height,
		    x, y);

	change = 1;
	old_x = x;
	old_y = y;
    }

    if (!dfn)
      dfn = &object->dfn;

    /* Change the context of the new data to match the context of the tactmap
     * center (tactmap->cur->cc
     */
    CHANGE_CONTEXT(dfn->pixmap.cell, dfn->pixmap.x, dfn->pixmap.y,cc);
    
    should_draw = SHOULD_DRAW(dfn);
    
    if (!should_draw)
    {
	if (change)
	  snstve_remove_object(tactmap->sensitive, old_x, old_y,
			       dfn->sensitive.sensClass,
			       dfn->sensitive.instance);
	return;
    }

    x = tactmap_FX_pixel_number(dfn->pixmap.x,cx,hsw,ppm);
    y = tactmap_FY_pixel_number(dfn->pixmap.y,cy,hsh,ppm);

    if (!dfn->pixmap.pixmap || use_erase_gc)
    {
	/* $$$ What is this test doing for you?  You have precluded 
	 * being here if !should_draw by the test just a few lines up.
	 */
	if (should_draw)
	{
	    if (object->display_mode == PERMANENT)
	      XCopyArea(XtDisplay(tactmap->widget), tactmap->terrain_pix,
			tactmap->background_pix, tactmap->gc.background,
			x, y, dfn->pixmap.width, dfn->pixmap.height, x, y);
	    else
	      XClearArea(DPYWIN,
			 x, y, dfn->pixmap.width, dfn->pixmap.height, 0);
	}

	snstve_remove_object(tactmap->sensitive, x, y,
			     dfn->sensitive.sensClass,
			     dfn->sensitive.instance);
    }
    else
    {
	if (should_draw)
	{
	    if (object->display_mode == PERMANENT)
	      XCopyArea(XtDisplay(tactmap->widget), dfn->pixmap.pixmap,
			tactmap->background_pix, tactmap->gc.background,
			0, 0, dfn->pixmap.width, dfn->pixmap.height, x, y);
	    else
	      XCopyArea(XtDisplay(tactmap->widget), dfn->pixmap.pixmap,
			XtWindow(tactmap->widget), tactmap->gc.background,
			0, 0, dfn->pixmap.width, dfn->pixmap.height, x, y);
		
	    snsobj.sensClass            = dfn->sensitive.sensClass;
	    snsobj.instance         = dfn->sensitive.instance;
	    snsobj.user_data        = dfn->sensitive.user_data;
	    snsobj.kind             = SNSTVE_RECTANGLE;
	    snsobj.attr.rect.x      = x-2;
	    snsobj.attr.rect.y      = y-2;
	    snsobj.attr.rect.width  = dfn->pixmap.width+4;
	    snsobj.attr.rect.height = dfn->pixmap.height+4;
	    
	    if (!change ||
		!snstve_change_object(tactmap->sensitive,
				      old_x, old_y, &snsobj))
	      snstve_add_object(tactmap->sensitive, &snsobj);
	}
	else
	{
	    if (change)
	      snstve_remove_object(tactmap->sensitive, old_x, old_y,
				   dfn->sensitive.sensClass,
				   dfn->sensitive.instance);
	}
    }

    if (clear_area_when_finished &&
	(object->display_mode == PERMANENT))
      XClearArea(DPYWIN, x, y, dfn->pixmap.width, dfn->pixmap.height, 0);
}

/*ARGSUSED*/
void 
tactmap_draw_drawn(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		   int32 use_erase_gc, TACTMAP_BOUNDS *bounds, 
		   CTDB_SEARCH_SPACE_PTR space, int32 clear_area_when_finished,
		   TACTMAP_OBJECT_DFN *dfn)
{
    int32 change=FALSE, old_x, old_y;
    SNSTVE_OBJECT snsobj;

    /* The x and y here are screen coordinates and must not be adjusted
     * based on context.
     */
    if (dfn)
    {
	/* Erase the last incarnation */
	if ((object->drawn_bounds.width && object->drawn_bounds.height) &&
	    (!dfn->drawn.erase_on_delete_only))
	{
	    if (object->display_mode == PERMANENT)
	    {
		XCopyArea(XtDisplay(tactmap->widget),
			  tactmap->terrain_pix, tactmap->background_pix,
			  tactmap->gc.background,
			  object->drawn_bounds.x, object->drawn_bounds.y,
			  object->drawn_bounds.width,
			  object->drawn_bounds.height,
			  object->drawn_bounds.x, object->drawn_bounds.y);

		if (clear_area_when_finished)
		  XClearArea(DPYWIN,
			     object->drawn_bounds.x, object->drawn_bounds.y,
			     object->drawn_bounds.width,
			     object->drawn_bounds.height,
			     0);
	    }
	    else
	      XCopyArea(XtDisplay(tactmap->widget),
			tactmap->background_pix,
			XtWindow(tactmap->widget), tactmap->gc.background,
			object->drawn_bounds.x, object->drawn_bounds.y,
			object->drawn_bounds.width,
			object->drawn_bounds.height,
			object->drawn_bounds.x, object->drawn_bounds.y);

	    change = TRUE;
	    old_x = object->drawn_bounds.x;
	    old_y = object->drawn_bounds.y;
	}
    }

    if (!dfn)
      dfn = &object->dfn;
    
    if (!dfn->drawn.draw)
    {
	if (change)
	  snstve_remove_object(tactmap->sensitive, old_x, old_y,
			       dfn->sensitive.sensClass,
			       dfn->sensitive.instance);
	return;
    }

    if (use_erase_gc)
    {
	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    if (object->display_mode == PERMANENT)
	      XCopyArea(XtDisplay(tactmap->widget),
			tactmap->terrain_pix, tactmap->background_pix,
			tactmap->gc.background,
			object->drawn_bounds.x, object->drawn_bounds.y,
			object->drawn_bounds.width,
			object->drawn_bounds.height,
			object->drawn_bounds.x, object->drawn_bounds.y);
	    else
	      XClearArea(DPYWIN,
			 object->drawn_bounds.x, object->drawn_bounds.y,
			 object->drawn_bounds.width,
			 object->drawn_bounds.height,
			 0);
	}

	snstve_remove_object(tactmap->sensitive,
			     object->drawn_bounds.x,
			     object->drawn_bounds.y,
			     dfn->sensitive.sensClass,
			     dfn->sensitive.instance);
    }
    else
    {
	dfn->drawn.draw(tactmap->widget,
			(object->display_mode == PERMANENT) ?
			tactmap->background_pix : XtWindow(tactmap->widget),
			dfn->drawn.draw_arg,
			&object->drawn_bounds);
	
	if (object->drawn_bounds.x < 0)
	{
	    object->drawn_bounds.width += object->drawn_bounds.x;
	    object->drawn_bounds.x = 0;
	}

	if (object->drawn_bounds.y < 0)
	{
	    object->drawn_bounds.height += object->drawn_bounds.y;
	    object->drawn_bounds.y = 0;
	}
	

	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    snsobj.sensClass            = dfn->sensitive.sensClass;
	    snsobj.instance         = dfn->sensitive.instance;
	    snsobj.user_data        = dfn->sensitive.user_data;
	    snsobj.kind             = SNSTVE_RECTANGLE;
	    snsobj.attr.rect.x      = object->drawn_bounds.x-2;
	    snsobj.attr.rect.y      = object->drawn_bounds.y-2;
	    snsobj.attr.rect.width  = object->drawn_bounds.width+4;
	    snsobj.attr.rect.height = object->drawn_bounds.height+4;

	    if (!change ||
		!snstve_change_object(tactmap->sensitive,
				      old_x, old_y, &snsobj))
	      snstve_add_object(tactmap->sensitive, &snsobj);
	}
	else
	{
	    if (change)
	      snstve_remove_object(tactmap->sensitive, old_x, old_y,
				   dfn->sensitive.sensClass,
				   dfn->sensitive.instance);
	}
    }

    if (clear_area_when_finished &&
	object->drawn_bounds.width && object->drawn_bounds.height &&
	(object->display_mode == PERMANENT))
      XClearArea(DPYWIN, object->drawn_bounds.x, object->drawn_bounds.y,
		 object->drawn_bounds.width, object->drawn_bounds.height, 0);
}

#endif /*USE_X*/
