#ifndef lint
static char rcsid [] = "$RCSfile: tmap_rails.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_rails.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw rails and other simple linear features       *
*       Created: Tue Jul 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/*ARGSUSED*/
static int32 get_simple_linear_segs(TACTMAP_PTR tactmap, 
				    TACTMAP_OBJECT_PTR object,
				    CTDB_SEARCH_SPACE_PTR space, 
				    int32 type_code, XSegment *segs)
{
    DECLARE_REGS;
    float32 *verts;
    int32 n_verts;
    XRectangle bounding_box;
    int32 j, k;
    int16 x, y;
    SNSTVE_OBJECT snsobj;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;
    snsobj.kind = SNSTVE_LINE;
    snsobj.attr.line.width = 10;

    bounding_box.x = 0;
    bounding_box.y = 0;
    bounding_box.width = 2 * hsw * ppm;
    bounding_box.height = 2 * hsh * ppm;

    k = 0;
    while (n_verts = ctdb_next_abstract(space, type_code, &verts, NULL, NULL))
    {
	if (n_verts < 2)
	  continue;

	/* Bundle up the points */
	for (j=0;j<n_verts;j++)
	{
	    x = tactmap_FX_pixel_number(verts[2*j],cx,hsw,ppm);
	    y = tactmap_FY_pixel_number(verts[2*j+1],cy,hsh,ppm);

	    if (j && LINE_MAY_CROSS_RECT(segs[k].x1,segs[k].y1,
					 x, y, bounding_box))
	    {
		snsobj.instance = 0;
		snsobj.attr.line.x0 = segs[k].x1;
		snsobj.attr.line.y0 = segs[k].y1;
		snsobj.attr.line.x1 = x;
		snsobj.attr.line.y1 = y;

		snstve_add_object(tactmap->sensitive, &snsobj);

		segs[k].x2 = x;
		segs[k].y2 = y;
		k++;
	    }

	    segs[k].x1 = x;
	    segs[k].y1 = y;
	}
    }

    return k;
}

/*ARGSUSED*/
void tactmap_draw_railroads(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			    int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			    CTDB_SEARCH_SPACE_PTR space)
{
    XSegment segs[2500];
    int lw;
    int32 nsegs;
    GC gc;

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.railroads;

    lw = PIXELS(5.0, tactmap->cur->ppm);
    if (lw < 1)
      lw = 1;
    if (nsegs =
	get_simple_linear_segs(tactmap, object,
			       space, CTDB_ABSTRACT_RAILROAD, segs)) 
    {
	XSetLineAttributes(XtDisplay(tactmap->widget), gc,
			   lw, LineOnOffDash, CapButt, JoinMiter);
	XDrawSegments(XtDisplay(tactmap->widget),
		      tactmap->terrain_pix, gc, segs, nsegs);

	if (lw > 1)
	{
	    XSetLineAttributes(XtDisplay(tactmap->widget), gc,
			       0, LineSolid, CapButt, JoinMiter);
	    XDrawSegments(XtDisplay(tactmap->widget),
			  tactmap->terrain_pix, gc, segs, nsegs);
	}
    }
}

/*ARGSUSED*/
void tactmap_draw_political(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
			    int32 use_erase_gc, TACTMAP_BOUNDS *bounds, 
			    CTDB_SEARCH_SPACE_PTR space)
{
    XSegment segs[2500];
    int32 nsegs;
    GC gc;

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.political;

    if (nsegs = 
	get_simple_linear_segs(tactmap, object, space,
			       CTDB_ABSTRACT_POLITICAL_BOUNDARY, segs))  
    {
	XSetLineAttributes(XtDisplay(tactmap->widget), gc,
			   4, LineOnOffDash, CapButt, JoinMiter);
	XDrawSegments(XtDisplay(tactmap->widget),
		      tactmap->terrain_pix, gc, segs, nsegs);
    }
}

/*ARGSUSED*/
void tactmap_draw_pipelines(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			    int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			    CTDB_SEARCH_SPACE_PTR space)
{
    XSegment segs[2500];
    int32 nsegs;
    GC gc;

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.pipelines;

    if (nsegs =
	get_simple_linear_segs(tactmap, object, space,
			       CTDB_ABSTRACT_PIPELINE, segs))
    {
	XSetLineAttributes(XtDisplay(tactmap->widget), gc,
			   4, LineOnOffDash, CapButt, JoinMiter);
	XDrawSegments(XtDisplay(tactmap->widget),
		      tactmap->terrain_pix, gc, segs, nsegs);
    }
}

/*ARGSUSED*/
void tactmap_draw_powerlines(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			     int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			     CTDB_SEARCH_SPACE_PTR space)
{
    XSegment segs[2500];
    int32 nsegs;
    GC gc;

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.powerlines;

    if (nsegs =
	get_simple_linear_segs(tactmap, object, space,
			       CTDB_ABSTRACT_POWERLINE, segs))
    {
	XSetLineAttributes(XtDisplay(tactmap->widget), gc,
			   2, LineOnOffDash, CapButt, JoinMiter);
	XDrawSegments(XtDisplay(tactmap->widget),
		      tactmap->terrain_pix, gc, segs, nsegs);
    }
}

#endif /*USE_X*/
