#ifndef lint
static char rcsid [] = "$RCSfile: tmap_state.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_state.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to manipulate the state of the tactical map          *
*       Created: Wed Sep 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <libgcs.h>
#include <libworld.h>
#include <libctdb.h>

static float64 scale_to_ppm(Screen *screen, float64 scale)
{
    float64 pixels, meters;

    pixels = screen->width;
    meters = screen->mwidth * 0.001;

    return (pixels / meters) / scale;
}

static float64 ppm_to_scale(Screen *screen, float64 ppm)
{
    float64 pixels, meters;

    pixels = screen->width;
    meters = screen->mwidth * 0.001;

    return (pixels / meters) / ppm;
}

void tactmap_set_scale(TACTMAP_PTR tactmap, float64 scale)
{
    float64 ppm = scale_to_ppm(XtScreen(tactmap->widget), scale);

    if (ppm != tactmap->cur->ppm)
    {
	tactmap->do_refresh = 1;
	tactmap->cur->ppm = ppm;
	tactmap->cur->hsw =
	  tactmap->pix_width/tactmap->cur->ppm/2;
	tactmap->cur->hsh =
	  tactmap->pix_height/tactmap->cur->ppm/2;
    }
}

void tactmap_scale_string(TACTMAP_PTR tactmap, char *scale)
{
    int32 sval = ppm_to_scale(XtScreen(tactmap->widget),
			      tactmap->cur->ppm);

    if (sval < 1000)
      sprintf(scale, "1:%d", sval);
    else if (sval < 1000000)
      sprintf(scale, "1:%d,%03d", sval/1000, sval%1000);
    else
      sprintf(scale, "1:%d,%03d,%03d",
	      sval/1000000, (sval/1000)%1000, sval%1000);
}

void 
tactmap_set_center(TACTMAP_PTR tactmap, int32 x, int32 y, int32 cell)
{
    float64 gcs_x=x;
    float64 gcs_y=y;
    CTDB *thinned_db;
    
    int32 sval = ppm_to_scale(XtScreen(tactmap->widget),
			      tactmap->cur->ppm);

    /* First check if the new point is still contained within this tdb.
     * If the point is contained within the current tdb then update the 
     * x and y and return.  Otherwise, call the coordinate conversion
     * routine to determine what is the proper cell to be in and update
     * the center dependant information accordingly.
     */
    if ( !ctdb_point_on_database((CTDB *)gcs_get_tdb(cell, FALSE),
				 (float64)x,(float64)y) )
      gcs_gcs2e_to_gcs(cell,gcs_x,gcs_y,&cell,&gcs_x,&gcs_y);
    
    if ( cell  != tactmap->vp.cc )
    {
	float64 sw[XYZ],ne[XYZ];
	
	/* If the new center point cell is not the same as the current
	 * center cell then update cell dependant data otherwise just
	 * update the cell,x,y
	 */
	
	/* Generate database bounds based on new cell */
	wld_get_playbox_extent(cell,sw,ne);
	tactmap->playbox.min_x = (int32)sw[X];
	tactmap->playbox.min_y = (int32)sw[Y];	
	tactmap->playbox.max_x = (int32)ne[X] ;
	tactmap->playbox.max_y = (int32)ne[Y] ;

	/* $$$ setting of ctdb pointer should be removed from tactmap API */
	tactmap->ctdb        = (CTDB *)gcs_get_tdb(cell, FALSE);

	/* Determine whether we are zoomed out far enough to use the thinned
	 * version of the database for redrawing.
	 */
	if((sval >= TACTMAP_THINNED_SCALE))
	   if ((thinned_db = (CTDB *)gcs_get_tdb(cell, TRUE)))
	  tactmap->ctdb = thinned_db;

	tactmap->vp.cc = cell;

	/* set refresh flag incase the coord converted x,y are identical
	 * but the cell has changed
	 */
	tactmap->do_refresh  = TRUE;
    }
    
    if ((gcs_x != tactmap->vp.cx) || (gcs_y != tactmap->vp.cy))
    {
	tactmap->vp.cx    = gcs_x;
	tactmap->vp.cy    = gcs_y;
	tactmap->do_refresh  = TRUE;
    }
}

static void
tactmap_establish_worldview(TACTMAP_PTR tactmap)
{
    register float64 xppm, yppm;
    float64  mn[XYZ],mx[XYZ];
    float64  pbc[XYC];

    wld_get_playbox_center(pbc);
    tactmap->wv.cc = pbc[CELL2D];
    tactmap->wv.cx = pbc[X];
    tactmap->wv.cy = pbc[Y];
    
    wld_get_playbox_extent((int32)pbc[CELL2D],mn,mx);

    xppm = WORLD_VIEW_FILL * tactmap->pix_width / (mx[X] - mn[X]);
    yppm = WORLD_VIEW_FILL * tactmap->pix_height / (mx[Y] - mn[Y]); 

    if (xppm < yppm)
      tactmap->wv.ppm = xppm;
    else
      tactmap->wv.ppm = yppm;
    
    tactmap->wv.hsw = tactmap->cur->hsw * 
      (tactmap->vp.ppm / tactmap->wv.ppm);
    tactmap->wv.hsh = tactmap->cur->hsh * 
      (tactmap->vp.ppm / tactmap->wv.ppm);
}

void 
tactmap_world_view(TACTMAP_PTR tactmap, int32 world_view_on)
{

    XGCValues gcv;

    if (world_view_on != tactmap->world_view_mode)
    {
	tactmap->do_refresh = 1;
	if (tactmap->world_view_mode = world_view_on)
	{
	    XFillRectangle(XtDisplay(tactmap->widget),
			   tactmap->background_pix, tactmap->gc.background,
			   0, 0, tactmap->pix_width, tactmap->pix_height);
	    
	    tactmap_establish_worldview(tactmap);
	    tactmap_draw_pregenerated_grids(tactmap,tactmap->background_pix);

	    gcv.line_style = LineOnOffDash;
	    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.grids,
		      GCLineStyle, &gcv);

	    tactmap_show_viewport(tactmap, tactmap->background_pix);

	    gcv.line_style = LineSolid;
	    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.grids,
		      GCLineStyle, &gcv);

	    XClearWindow(DPYWIN);
	    tactmap->cur = &tactmap->wv;
	}
	else /* restore viewport view */
	  tactmap->cur = &tactmap->vp;
    }
}

void 
tactmap_screen_to_map(TACTMAP_PTR tactmap, int32 sx, int32 sy, 
		      float64 *mx, float64 *my, int32 *mc)
{
    DECLARE_REGS;
    gcs_gcs2e_to_gcs(cc, (float64)X_METER_NUMBER(sx,cx,hsw,ppm),
		     (float64)Y_METER_NUMBER(sy,cy,hsh,ppm),
		     mc,mx,my);
}

int32 
tactmap_map_to_screen(TACTMAP_PTR tactmap, float64 mx, float64 my,
		      int32 mc, int32 *sx, int32 *sy)
{
    DECLARE_REGS;

    /* convert the coordinates of the incoming point into the tactmap
     * context before converting into the screen coordinate system
     */
    gcs_gcs2e_to_gcscs(mc,mx,my,cc,&mx,&my);

    *sx = tactmap_FX_pixel_number(mx,cx,hsw,ppm);
    *sy = tactmap_FY_pixel_number(my,cy,hsh,ppm);

    return ((mx >= (cx - hsw)) && (mx <= (cx + hsw)) &&
	    (my >= (cy - hsh)) && (my <= (cy + hsh)));
}

void 
tactmap_pix_to_offvec(TACTMAP_PTR tactmap, int32 sx, int32 sy, float64 *offset)
{
    offset[X]      = METERS(sx,tactmap->cur->ppm);
    offset[Y]      = -METERS(sy,tactmap->cur->ppm);
    offset[CELL2D] = tactmap->cur->cc; 
}

void 
tactmap_offvec_to_pix(TACTMAP_PTR tactmap, float64 *offset,
		      int32 *sx, int32 *sy)
{
    *sx            = PIXELS(offset[X],tactmap->cur->ppm);
    *sy            = -PIXELS(offset[Y],tactmap->cur->ppm);
}

void 
tactmap_get_dimension(TACTMAP_PTR tactmap, float64 *width, float64 *height)
{
    /* dimensions of the screen in the cell coordinates */
    *width    = tactmap->cur->hsw * 2;
    *height   = tactmap->cur->hsh * 2;
}

void 
tactmap_get_center(TACTMAP_PTR tactmap, float64 *x, float64 *y, int32 *c)
{
    /* GCS coordinates based on the center cell */
    *c = tactmap->cur->cc;
    *x = tactmap->cur->cx;
    *y = tactmap->cur->cy;
}   


float64 tactmap_get_scale(TACTMAP_PTR tactmap, int32 width, int32 height)
{
    DECLARE_REGS;
    float64 ppm_width, ppm_height;
#ifdef lint
    cx = cy; cy = cx;
#endif

    if (width < 0)
      width = -width;
    if (height < 0)
      height = -height;

    ppm_width =
      (double)PIXELS(2*hsw,ppm)/(double)METERS(width,ppm);
    ppm_height =
      (double)PIXELS(2*hsh,ppm)/(double)METERS(height,ppm);

    /* Choose the larger of the two dimensions (smaller ppm) */
    if (ppm_width < ppm_height)
      return ppm_to_scale(XtScreen(tactmap->widget), ppm_width);
    else
      return ppm_to_scale(XtScreen(tactmap->widget), ppm_height);
}

int32 tactmap_redrawing(TACTMAP_PTR tactmap)
{
    return tactmap->redrawing;
}

void tactmap_bundle_changes(TACTMAP_PTR tactmap)
{
  tactmap->do_refresh = 1;

}

int32 tactmap_off_screen64(int32 n_vert, float64 verts[][2],
			   TACTMAP_BOUNDS *bounds)
{
    int32 lt, rt, bt, tp;
    int32 i;

    lt = rt = bt = tp = 1;

    for(i=0;i<n_vert;i++)
    {
	if (verts[i][X] > bounds->min_x) lt = 0;
	if (verts[i][X] < bounds->max_x) rt = 0;
	if (verts[i][Y] > bounds->min_y) bt = 0;
	if (verts[i][Y] < bounds->max_y) tp = 0;
    }
    return (lt || rt || bt || tp);
}

int32 tactmap_off_screen32(int32 n_vert, float32 *verts,
			   TACTMAP_BOUNDS *bounds)
{
    int32 lt, rt, bt, tp;
    int32 i;

    lt = rt = bt = tp = 1;

    n_vert += n_vert;
    for(i=0;i<n_vert;i+=2)
    {
	if (verts[i] > bounds->min_x) lt = 0;
	if (verts[i] < bounds->max_x) rt = 0;
	if (verts[i+1] > bounds->min_y) bt = 0;
	if (verts[i+1] < bounds->max_y) tp = 0;
    }
    return (lt || rt || bt || tp);
}

#define MAX_PIXEL 0x7FFF
int16 tactmap_X_pixel_number(int32 x, int32 cx, int32 hsw, float64 ppm)
{
    float64 result = X_P_N(x,cx,hsw,ppm);
    if (result > MAX_PIXEL) return MAX_PIXEL;
    if (result < -MAX_PIXEL) return -MAX_PIXEL;
    return result;
}

int16 tactmap_Y_pixel_number(int32 y, int32 cy, int32 hsh, float64 ppm)
{
    float64 result = Y_P_N(y,cy,hsh,ppm);
    if (result > MAX_PIXEL) return MAX_PIXEL;
    if (result < -MAX_PIXEL) return -MAX_PIXEL;
    return result;
}

int16 tactmap_FX_pixel_number(float64 x, int32 cx, int32 hsw, float64 ppm)
{
    float64 result = X_P_N(x,cx,hsw,ppm);
    if (result > MAX_PIXEL) return MAX_PIXEL;
    if (result < -MAX_PIXEL) return -MAX_PIXEL;
    return result;
}

int16 tactmap_FY_pixel_number(float64 y, int32 cy, int32 hsh, float64 ppm)
{
    float64 result = Y_P_N(y,cy,hsh,ppm);
    if (result > MAX_PIXEL) return MAX_PIXEL;
    if (result < -MAX_PIXEL) return -MAX_PIXEL;
    return result;
}

#endif /*USE_X*/
