#ifndef lint
static char rcsid [] = "$RCSfile: tmap_text.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_text.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw TEXT class objects                           *
*       Created: Thu Feb 25 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <p_po.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdstring.h> /*common/include/global*/

/*ARGSUSED*/
void 
tactmap_draw_text(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
		  int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
		  CTDB_SEARCH_SPACE_PTR space, int32 clear_area_when_finished,
		  TACTMAP_OBJECT_DFN *dfn)
{
    DECLARE_REGS;
    int32 x, y, width, height;
    int32 change=0, old_x, old_y;
    SNSTVE_OBJECT snsobj;
    int ign, ascent, descent;
    XFontStruct *xfs;
    XCharStruct result;
    int32 i, nlines;
    char buf[10240];
    char *lines[100];
    int32 heights[100];
    Drawable d;


    /* Note: drawn_bounds are in screen pixel coordinates, to transformation
     * required
     */
    if (dfn)
    {
	/* Erase the last incarnation */
	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    if (object->display_mode == PERMANENT)
	    {
		XCopyArea(XtDisplay(tactmap->widget),
			  tactmap->terrain_pix, tactmap->background_pix,
			  tactmap->gc.background,
			  object->drawn_bounds.x, object->drawn_bounds.y,
			  object->drawn_bounds.width,
			  object->drawn_bounds.height,
			  object->drawn_bounds.x, object->drawn_bounds.y);

		if (clear_area_when_finished)
		  XClearArea(DPYWIN,
			     object->drawn_bounds.x, object->drawn_bounds.y,
			     object->drawn_bounds.width,
			     object->drawn_bounds.height,
			     0);
	    }
	    else
	      XCopyArea(XtDisplay(tactmap->widget),
			tactmap->background_pix,
			XtWindow(tactmap->widget), tactmap->gc.background,
			object->drawn_bounds.x, object->drawn_bounds.y,
			object->drawn_bounds.width,
			object->drawn_bounds.height,
			object->drawn_bounds.x, object->drawn_bounds.y);

	    change = 1;
	    old_x = object->drawn_bounds.x;
	    old_y = object->drawn_bounds.y;
	}
    }

    if (!dfn)
      dfn = &object->dfn;


    if (!dfn->text.string || !*dfn->text.string)
    {
	if (change)
	  snstve_remove_object(tactmap->sensitive, old_x, old_y,
			       dfn->sensitive.sensClass,
			       dfn->sensitive.instance);
	return;
    }

    if (use_erase_gc)
    {
	if (object->drawn_bounds.width && object->drawn_bounds.height)
	{
	    if (object->display_mode == PERMANENT)
	      XCopyArea(XtDisplay(tactmap->widget),
			tactmap->terrain_pix, tactmap->background_pix,
			tactmap->gc.background,
			object->drawn_bounds.x, object->drawn_bounds.y,
			object->drawn_bounds.width,
			object->drawn_bounds.height,
			object->drawn_bounds.x, object->drawn_bounds.y);
	    else
	      XClearArea(DPYWIN,
			 object->drawn_bounds.x, object->drawn_bounds.y,
			 object->drawn_bounds.width,
			 object->drawn_bounds.height,
			 0);
	}

	snstve_remove_object(tactmap->sensitive,
			     object->drawn_bounds.x,
			     object->drawn_bounds.y,
			     dfn->sensitive.sensClass,
			     dfn->sensitive.instance);
    }
    else
    {
	/* Get the font from the server */
	if (dfn->text.font_struct)
	  xfs = dfn->text.font_struct;
	else
	  xfs = XQueryFont(XtDisplay(tactmap->widget),
			   XGContextFromGC(dfn->text.gc));

	/* Copy the text into a buffer so we can strtok it */
	strcpy(buf, dfn->text.string);
    
	/* Break the buffer up into lines */
	lines[0] = strtok(buf, "\n");

        if (!lines[0])
          nlines = 0;
        else
        {
	    i = 1;
	    while (lines[i++] = strtok(NULL, "\n"))
	      ;
	    nlines = i-1;
        }

	/* Find the total width & height */
	width = height = 0;

	for (i=0;i<nlines;i++)
	{
	    XTextExtents(xfs,
			 lines[i], strlen(lines[i]),
			 &ign, &ascent, &descent, &result);

	    if (result.width > width)
	      width = result.width;

	    height += (heights[i] = ascent + descent);
	}
	
	CHANGE_CONTEXT(dfn->text.cell,dfn->text.x,dfn->text.y,cc);
	x = tactmap_FX_pixel_number(dfn->text.x,cx,hsw,ppm);
	y = tactmap_FY_pixel_number(dfn->text.y,cy,hsh,ppm);

	x += dfn->text.horizontal_offset;
	y -= dfn->text.vertical_offset;

	switch (dfn->text.alignment)
	{
	  case northWest:	x -= width;	y -= height;	break;
	  case north:		x -= width/2;	y -= height;	break;
	  case northEast:			y -= height;	break;
	  case west:		x -= width;	y -= height/2;	break;
	  case center:		x -= width/2;	y -= height/2;	break;
	  case east:				y -= height/2;	break;
	  case southWest:	x -= width;			break;
	  case south:		x -= width/2;			break;
	  case southEast:					break;
	}
	    
	object->drawn_bounds.x      = x;
	object->drawn_bounds.y      = y;
	object->drawn_bounds.width  = width+2;
	object->drawn_bounds.height = height+1;

	snsobj.sensClass                = dfn->sensitive.sensClass;
	snsobj.instance             = dfn->sensitive.instance;
	snsobj.user_data            = dfn->sensitive.user_data;
	snsobj.kind                 = SNSTVE_RECTANGLE;
	snsobj.attr.rect.x          = object->drawn_bounds.x-2;
	snsobj.attr.rect.y          = object->drawn_bounds.y-2;
	snsobj.attr.rect.width      = object->drawn_bounds.width+4;
	snsobj.attr.rect.height     = object->drawn_bounds.height+4;
	
	if (!change ||
	    !snstve_change_object(tactmap->sensitive,
				  old_x, old_y, &snsobj))
	  snstve_add_object(tactmap->sensitive, &snsobj);

	/* Shift the location so the real upper-left corner of the first
	 * character lands there.
	 */
	x += xfs->max_bounds.lbearing;
	y += ascent;

	/* If we looked up the font, give the memory back now */
	if (!dfn->text.font_struct)
	  XFreeFontInfo(NULL, xfs, 1);

	if (object->display_mode == PERMANENT)
	  d = tactmap->background_pix;
	else
	  d = XtWindow(tactmap->widget);

	for (i=0;i<nlines;i++)
	{
	    if (lines[i])
	    {
		if (dfn->text.see_through)
		  XDrawString(XtDisplay(tactmap->widget), d,
			      dfn->text.gc, x, y,
			      lines[i], strlen(lines[i]));
		else
		  XDrawImageString(XtDisplay(tactmap->widget), d,
				   dfn->text.gc, x, y,
				   lines[i], strlen(lines[i]));
	    }
	    y += heights[i];
	}

    }

    if (clear_area_when_finished &&
	object->drawn_bounds.width && object->drawn_bounds.height &&
	(object->display_mode == PERMANENT))
      XClearArea(DPYWIN, object->drawn_bounds.x, object->drawn_bounds.y,
		 object->drawn_bounds.width, object->drawn_bounds.height, 0);
}

#endif /*USE_X*/
