#ifndef lint
static char rcsid [] = "$RCSfile: tmap_tick.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_tick.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to refresh the map a little at a time                *
*       Created: Wed Sep 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <libtime.h>

static Pixmap allocate_pix(TACTMAP_PTR tactmap);
static void please_wait(TACTMAP_PTR tactmap);

void tactmap_tick(TACTMAP_PTR tactmap, uint32 max_time_slice)
{
    uint32 must_return;
    int32 map_feature, last_was_map_feature;
    CTDB_SEARCH_SPACE_PTR space = NULL;
    TACTMAP_BOUNDS bounds;
    TACTMAP_OBJECT_CLASS sensClass;
    int32 finished = 1;
    uint32 search_flags;

    if (!tactmap->redrawing && !tactmap->do_refresh)
      return;

    /* Deal with world view as a special case */
    if (tactmap->world_view_mode)
    {
	tactmap->redrawing = 0;
	tactmap->do_refresh = 0;

	snstve_clear(tactmap->sensitive);
	tactmap_erase_viewport(tactmap);
	tactmap_show_viewport(tactmap, XtWindow(tactmap->widget));

	return;
    }

    must_return = time_realtime_clock() + max_time_slice;

    /* Make sure we have pixmaps to draw into */
    if (!tactmap->terrain_pix)
    {
	tactmap->terrain_pix = allocate_pix(tactmap);
	XSetTile(XtDisplay(tactmap->widget), tactmap->terrain_gc,
		 tactmap->terrain_pix);
    }
    if (!tactmap->background_pix)
    {
	tactmap->background_pix = allocate_pix(tactmap);
	XSetTile(XtDisplay(tactmap->widget), tactmap->background_gc,
		 tactmap->background_pix);
	XSetWindowBackgroundPixmap(DPYWIN, tactmap->background_pix);
    }

    if (tactmap->do_refresh)
    {
	CTDB *thinned_db;
	float64 pixels, meters;
	int32 sval;
	Screen *screen= XtScreen(tactmap->widget);

	pixels = screen->width;
	meters = screen->mwidth * 0.001;

	if(tactmap->cur->ppm)
	  sval = (pixels / meters) / tactmap->cur->ppm;
	else
	  sval = 1000;

	if (tactmap->refresh_callback)
	  tactmap->refresh_callback(tactmap, 1);
	tactmap->redrawing = 1;
	tactmap->current_class = TACTMAP_FIRST_CLASS;

	snstve_clear(tactmap->sensitive);
	XFillRectangle(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		       tactmap->gc.background,
		       0, 0, tactmap->pix_width, tactmap->pix_height);
	please_wait(tactmap);

	/* Determine whether we are zoomed out far enough to use the thinned
	 * version of the database for redrawing.
	 */
	if((sval >= TACTMAP_THINNED_SCALE) &&
	   (thinned_db = (CTDB *)gcs_get_tdb(tactmap->vp.cc, TRUE)))
	  tactmap->ctdb = thinned_db;
	else
	  tactmap->ctdb = (CTDB *)gcs_get_tdb(tactmap->vp.cc, FALSE);

	tactmap->do_refresh = 0;

	/* Don't start the redraw until the next tick (this gives the
	 * please_wait a chance to come up).
	 */
	return;
    }

    bounds.min_x = tactmap->cur->cx - tactmap->cur->hsw;
    bounds.max_x = tactmap->cur->cx + tactmap->cur->hsw;
    bounds.min_y = tactmap->cur->cy - tactmap->cur->hsh;
    bounds.max_y = tactmap->cur->cy + tactmap->cur->hsh;

    if (tactmap->current_class == TACTMAP_FIRST_CLASS)
      last_was_map_feature = 1;
    else
      last_was_map_feature = 0;

    search_flags = 0;

    /* If individual tree will not be displayed, optimize our search by
     * asking libctdb to not give them to us.
     */
    if (tactmap->objects[TACTMAP_TREES] &&
	!(tactmap->objects[TACTMAP_TREES]->dfn.trees.mask &
	  TACTMAP_TREE_INDIVIDUAL))
      search_flags |= CTDB_NO_INDIVIDUAL_TREES;

    while (tactmap->current_class<=TACTMAP_LAST_CLASS)
    {
	if (!space)
	  space = ctdb_create_search_space(tactmap->ctdb, 
					   bounds.min_x, bounds.min_y,
					   bounds.max_x, bounds.max_y,
					   search_flags);

	ctdb_reset_search_space(space);

	if (tactmap->objects[tactmap->current_class])
	{
	    if (map_feature = TACTMAP_MAP_FEATURE(tactmap->current_class))
	    {
		/* Put up a Please Wait before drawing Hypso */
		if (tactmap->current_class == TACTMAP_HYPSO)
		  please_wait(tactmap);

		tactmap_draw_class(tactmap, tactmap->current_class,
				   PERMANENT, &bounds, space);
	    }
	    else
	    {
		/* Copy the terrain_pix to the background_pix *just
		 * before* we draw the first non-map feature.
		 */
		if (last_was_map_feature)
		  XCopyArea(XtDisplay(tactmap->widget),
			    tactmap->terrain_pix, tactmap->background_pix,
			    tactmap->gc.background,
			    0, 0, tactmap->pix_width, tactmap->pix_height,
			    0, 0);

		tactmap_draw_class(tactmap, tactmap->current_class,
				   PERMANENT, &bounds, space);
	    }
	    last_was_map_feature = map_feature;

	}

	tactmap->current_class++;

	if (time_realtime_clock() >= must_return)
	{
	    finished = 0;
	    break;
	}
    }

    /* Refresh the display */
    if (last_was_map_feature)
      XCopyArea(XtDisplay(tactmap->widget),
		tactmap->terrain_pix, tactmap->background_pix,
		tactmap->gc.background,
		0, 0, tactmap->pix_width, tactmap->pix_height,
		0, 0);

    /* If you omit the 'if', there will be a perodic update of map
     * display instead of the redraw happening all at once.  I've
     * decided it isn't really necessary since we have the
     * "Please Wait" message...
     */
    if (finished)
      XClearWindow(DPYWIN);

    snstve_refresh(tactmap->sensitive);
    if (space)
    {
	ctdb_destroy_search_space(space);
	space = NULL;
    }

    if (finished)
    {
	/* Draw the transients last */
	for (sensClass=TACTMAP_FIRST_DYNAMIC_CLASS;sensClass<=TACTMAP_LAST_CLASS;
	     sensClass++)
	  tactmap_draw_class(tactmap, sensClass, TRANSIENT, &bounds, NULL);

	tactmap->redrawing = 0;
	if (tactmap->refresh_callback)
	  tactmap->refresh_callback(tactmap, 0);
	XUndefineCursor(DPYWIN);
    }
    /* We need this if we do the ClearWindow above
    else
      please_wait(tactmap);
    */
}

static void please_wait(TACTMAP_PTR tactmap)
{
    char *msg = "Redrawing Map.  Please Wait.";
    XFontStruct *fs;
    XGCValues gcv;
    int32 xoff;

    XGetGCValues(XtDisplay(tactmap->widget), tactmap->gc.grids, GCFont, &gcv);

    fs = XQueryFont(XtDisplay(tactmap->widget), gcv.font);

    xoff = fs->max_bounds.width * strlen(msg) / 2;

    XDrawImageString(DPYWIN, tactmap->gc.grids,
		     tactmap->pix_width/2 - xoff, tactmap->pix_height/2,
		     msg, strlen(msg));

    XFreeFontInfo(NULL, fs, 1);

    XDefineCursor(DPYWIN, tactmap->hourglass);

    XFlush(XtDisplay(tactmap->widget));
}

static Pixmap allocate_pix(TACTMAP_PTR tactmap)
{
    XEvent event;

    /* Make sure we know our width & height */
    if (!tactmap->pix_width || !tactmap->pix_height)
    {
	/* Fake a resize event */
	event.type = ConfigureNotify;
	tactmap_resize((Widget)NULL, tactmap, &event);
    }

    return XCreatePixmap(DPYWIN,
			 tactmap->pix_width, tactmap->pix_height,
			 tactmap->depth);
}

#endif /*USE_X*/
