#ifndef lint
static char rcsid [] = "$RCSfile: tmap_trees.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_trees.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw treelines and canopies                       *
*       Created: Tue Jul 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define MAX_SEGS 50000
#define SEND_SEGS 45000

static void draw_canopies(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			  int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			  CTDB_SEARCH_SPACE_PTR space);

void tactmap_draw_trees(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
			int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XSegment segs[MAX_SEGS];
    float64 verts[CTDB_VERTICES_MAX][XY];
    int32 n_verts;
    int32 j, k;
    int32 x, y;
    XGCValues gcv;
    GC gc;
    SNSTVE_OBJECT snsobj;

    if (object->dfn.trees.mask & TACTMAP_TREE_CANOPIES)
      draw_canopies(tactmap, object, use_erase_gc, bounds, space);

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.trees;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;

    if ((gcv.line_width = TACTMAP_TREE_WIDTH * ppm) == 1)
      gcv.line_width = 0;

    if ((gcv.line_width < 2) && (ppm > TACTMAP_MIN_WIDE_TREE_PPM))
      gcv.line_width = 2;

    XChangeGC(XtDisplay(tactmap->widget), gc, GCLineWidth, &gcv);

    k = 0;
    while (n_verts = ctdb_next_tree(space, verts, NULL, NULL, NULL))
    {
	if (k > SEND_SEGS)
	{
	    XDrawSegments(XtDisplay(tactmap->widget), tactmap->terrain_pix,
			  gc, segs, k);
	    k = 0;
	}

	if (n_verts == 1)
	{
	    if (object->dfn.trees.mask & TACTMAP_TREE_INDIVIDUAL)
	    {
		x = tactmap_FX_pixel_number(verts[0][X],cx,hsw,ppm);
		y = tactmap_FY_pixel_number(verts[0][Y],cy,hsh,ppm);

		segs[k  ].x1 = x-3; segs[k  ].y1 = y-3;
		segs[k  ].x2 = x+3; segs[k  ].y2 = y+3;
		segs[k+1].x1 = x-3; segs[k+1].y1 = y+3;
		segs[k+1].x2 = x+3; segs[k+1].y2 = y-3;
		k += 2;

		snsobj.kind = SNSTVE_CIRCLE;
		snsobj.instance = 0;
		snsobj.attr.circle.x = x;
		snsobj.attr.circle.y = y;
		snsobj.attr.circle.radius = 8;

		snstve_add_object(tactmap->sensitive, &snsobj);
	    }
	}
	else
	{
	    if (object->dfn.trees.mask & TACTMAP_TREE_LINES)
	    {
		/* Bundle up the points */
		for (j=0;j<n_verts;j++)
		{
		    x = tactmap_FX_pixel_number(verts[j][X],cx,hsw,ppm);
		    y = tactmap_FY_pixel_number(verts[j][Y],cy,hsh,ppm);

		    if (j)
		    {
			snsobj.kind = SNSTVE_LINE;
			snsobj.instance = 0;
			snsobj.attr.line.x0 = segs[k].x1;
			snsobj.attr.line.y0 = segs[k].y1;
			snsobj.attr.line.x1 = x;
			snsobj.attr.line.y1 = y;
			snsobj.attr.line.width = 10;

			snstve_add_object(tactmap->sensitive, &snsobj);

			segs[k].x2 = x;
			segs[k].y2 = y;
			k++;

		    }

		    segs[k].x1 = x;
		    segs[k].y1 = y;
		}
	    }
	}
    }

    if (k > 0)
      XDrawSegments(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		    gc, segs, k);
}

/*ARGSUSED*/
static void draw_canopies(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			  int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			  CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XPoint segs[5000];
    float32 *verts;
    int32 n_verts;
    int32 j;
    CTDB_ABSTRACT_DATA *data;
    int ls;
    XGCValues gcv;
    GC fill_gc, line_gc;

    if (use_erase_gc)
      fill_gc = line_gc = tactmap->gc.background;
    else
      fill_gc = tactmap->gc.canopies,
      line_gc = tactmap->gc.trees;

    gcv.line_width = 0;
    gcv.line_style = LineSolid;
    XChangeGC(XtDisplay(tactmap->widget), fill_gc, GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), line_gc,
	      GCLineWidth | GCLineStyle, &gcv);

    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_CANOPY,
					&verts, &data, NULL))
    {
	if (n_verts < 3)
	  continue;

	if (tactmap_off_screen32(n_verts, verts, bounds))
	  continue;

	/* Bundle up the points */
	for (j=0;j<n_verts;j++)
	{
	    segs[j].x = tactmap_FX_pixel_number(verts[2*j],cx,hsw,ppm);
	    segs[j].y = tactmap_FY_pixel_number(verts[2*j+1],cy,hsh,ppm);
	}

	XFillPolygon(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		     fill_gc, segs, j, Nonconvex, CoordModeOrigin);
	    
	segs[j++] = segs[0];

	if (data->canopy.impenetrable)
	  ls = LineSolid;
	else
	  ls = LineDoubleDash;

	if (ls != gcv.line_style)
	{
	    gcv.line_style = ls;
	    XChangeGC(XtDisplay(tactmap->widget), line_gc, GCLineStyle, &gcv);
	}

	XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		   line_gc, segs, j, CoordModeOrigin);
    }

    gcv.line_style = LineSolid;
    XChangeGC(XtDisplay(tactmap->widget), line_gc, GCLineStyle, &gcv);
}

#ifdef AE_FEATURES
void tactmap_draw_tact_signs(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object,
			     int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
			     CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XPoint segs[5000];
    float32 *verts;
    int32 n_verts;
    int32 j;
    XGCValues gcv;
    GC line_gc;

    if (use_erase_gc)
      line_gc = tactmap->gc.background;
    else
      line_gc = tactmap->gc.tact_signs;

    gcv.line_width = 0;
    XChangeGC(XtDisplay(tactmap->widget), line_gc, GCLineWidth, &gcv);

    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_TACTICAL_SIGN,
					&verts, &data, NULL))
    {
	if (n_verts < 3)
	  continue;

	if (tactmap_off_screen32(n_verts, verts, bounds))
	  continue;

	/* Bundle up the points */
	for (j=0;j<n_verts;j++)
	{
	    segs[j].x = tactmap_X_pixel_number(verts[2*j],cx,hsw,ppm);
	    segs[j].y = tactmap_Y_pixel_number(verts[2*j+1],cy,hsh,ppm);
	}

	segs[j++] = segs[0];

	XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		   line_gc, segs, j, CoordModeOrigin);
    }
}
#endif

#endif /*USE_X*/
