#ifndef lint
static char rcsid [] = "$RCSfile: tmap_water.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_water.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw rivers and lakes                             *
*       Created: Tue Jul 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include <stdlib.h>
#include "libtmap_local.h"
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

static void 
draw_lakes(TACTMAP_PTR tactmap, 
	   TACTMAP_OBJECT_PTR object, 
	   int32 use_erase_gc, 
	   TACTMAP_BOUNDS *bounds, 
	   CTDB_SEARCH_SPACE_PTR space);

/* Temporary globals (used during calls to tactmap_draw_water */
static int32 n_bridges;
static int32 alloced = 0;
static TACTMAP_BRIDGE_DATA *bridges=NULL;

static void 
maybe_save_bridge(CTDB *ctdb, 
		  int32 node_index)
{
    CTDB_NODE *node=ctdb->nodes[node_index];
    int32              i;
    TACTMAP_BRIDGE_DATA cand;

    if (!node->is_bridge)
      return;

    cand.loc[X]      = node->x;
    cand.loc[Y]      = node->y;
    cand.loc[CELL2D] = ctdb_get_cell_id(ctdb);
    /* See if we already have this bridge saved */
    for (i=0;i<n_bridges;i++)
      if ( gcs_gcs2_pos_equal64(bridges[i].loc, cand.loc) )
	return;

    /* Make room to save this */
    if (n_bridges == alloced)
    {
	alloced += 32;
	
	if (bridges)
	  bridges = (TACTMAP_BRIDGE_DATA *)
	    STDREALLOC(bridges, alloced * sizeof(TACTMAP_BRIDGE_DATA));
	else
	  bridges = (TACTMAP_BRIDGE_DATA *)
	    STDALLOC(alloced * sizeof(TACTMAP_BRIDGE_DATA));
    }

    ctdb_bridge_info_by_index(ctdb, node_index,NULL,&cand.width,
                              &cand.del[X],&cand.del[Y]);
    cand.mes = FALSE;
    /* Save the bridge node and some information about the road */
    bcopy(&cand,&bridges[n_bridges],sizeof(TACTMAP_BRIDGE_DATA));
    n_bridges++;
}

void 
tactmap_draw_bridge(TACTMAP_PTR          tactmap, 
		    TACTMAP_BRIDGE_DATA *bridges)
		    
{
    DECLARE_REGS;
    XPoint pts[4], spts[8];
    XSegment segs[6];
    float64 mag, dx, dy, ddx, ddy, x, y;
    float64 top[XY], bot[XY],halfdel[XY], widthvec[XY], tmp[XY];
    float64 unitvec[XY], rotunitvec[XY];
    static float64 rot45[XY][XY],rot90[XY][XY];
    static int32 init_rotation=TRUE;
    
    if (init_rotation)
    {
	vmat2_primary_rotation64(0.707,0.707,rot45);
	vmat2_primary_rotation64(1.0,0.0,rot90);
	init_rotation = FALSE;
    }
    
    gcs_gcs2_to_gcscs((int32)bridges->loc[CELL2D],
		      bridges->loc,tactmap->cur->cc,
		      bridges->loc);
    bridges->loc[CELL2D] = tactmap->cur->cc;
    
    gcs_gcs2_vec_rotate64(bridges->del,tactmap->cur->cc,bridges->del);
    
    if (!bridges->mes)
    {
	/* make up the bridge dimensions */
	mag = fsqrt(bridges->del[X]* bridges->del[X] +
		    bridges->del[Y] * bridges->del[Y]);
	mag = 0.5 * bridges->width / mag *1.5;
	bridges->del[X] *= mag;
	bridges->del[Y] *= mag;
    }
    
    vmat2_unit64(bridges->del, unitvec);

    vmat2_mat_vec_mul64(rot90, unitvec, widthvec);
    vmat2_scal_vec_mul64(0.5*bridges->width, widthvec, widthvec);
    vmat2_vec_add64(bridges->loc, widthvec, top);
    vmat2_vec_sub64(bridges->loc, widthvec, bot);
    vmat2_scal_vec_mul64(0.5, bridges->del, halfdel);
    
    /* Top part of the bridge */  
    vmat2_vec_sub64(top, halfdel, tmp);
    pts[0].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    pts[0].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);

    vmat2_mat_vec_mul64(rot45, widthvec, rotunitvec);
    vmat2_vec_add64(tmp, rotunitvec, tmp);
    spts[0].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    spts[0].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);
    spts[1]   = pts[0];
    
    vmat2_vec_add64(top, halfdel, tmp);
    pts[1].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    pts[1].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);

    vmat2_mat_vec_mul64(rot90, rotunitvec, rotunitvec);
    vmat2_vec_sub64(tmp, rotunitvec, tmp);
    spts[2]   = pts[1];
    spts[3].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    spts[3].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);

    /* Bottom part of the bridge */
    vmat2_vec_add64(bot, halfdel, tmp);
    pts[2].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    pts[2].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);

    vmat2_mat_vec_mul64(rot90, rotunitvec, rotunitvec);
    vmat2_vec_add64(tmp, rotunitvec, tmp);
    spts[4].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    spts[4].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);
    spts[5]   = pts[2];

    vmat2_vec_sub64(bot, halfdel, tmp);
    pts[3].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    pts[3].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);

    vmat2_mat_vec_mul64(rot90, rotunitvec, rotunitvec);
    vmat2_vec_sub64(tmp, rotunitvec, tmp);
    spts[6]   = pts[3];
    spts[7].x = tactmap_FX_pixel_number(tmp[X],cx,hsw,ppm);
    spts[7].y = tactmap_FY_pixel_number(tmp[Y],cy,hsh,ppm);

#if 0    
    /* Skip tiny bridges */
    if ((abs(pts[0].x - pts[1].x) < 2) &&
	(abs(pts[0].y - pts[1].y) < 2) ||
	(abs(pts[1].x - pts[2].x) < 2) &&
	(abs(pts[1].y - pts[2].y) < 2))
      return;	  
#endif  
  
    /* Blot out the road and river in that box */
    XFillPolygon(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		 tactmap->gc.bridge_picture, pts, 4,
		 Nonconvex, CoordModeOrigin);
    
    XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix,
	       tactmap->gc.bridge_symbol, &spts[0], 4, CoordModeOrigin);
    XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix,
	       tactmap->gc.bridge_symbol, &spts[4], 4, CoordModeOrigin);
}

/* Note: This gets called for both roads and rivers, because CTDB gives
 *       both of those back from the same function.  Since we iterate over
 *       the search space exactly once, the first one called will cause
 *       both to be drawn, and the second one will do nothing.
 */

void 
tactmap_draw_water(TACTMAP_PTR           tactmap, 
		   TACTMAP_OBJECT_PTR    object, 
		   int32                 use_erase_gc, 
		   TACTMAP_BOUNDS       *bounds, 
		   CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    XPoint pts[5000];
    XSegment segs[6];
    float64 verts[CTDB_VERTICES_MAX][XY], widthD;
    float32 *vertices;
    CTDB_ABSTRACT_DATA *data;
    int32 n_verts;
    int32 j, width, soil, ei, esi;
    int32 do_roads, do_rivers, is_a_road;
    XGCValues *current_gcv, ford_gcv, water_gcv, road_gcv;
    GC current_gc, ford_gc, water_gc, road_gc, gc;
    CTDB_NODE *node;
    SNSTVE_OBJECT snsobj;
    CTDB *ctdb;

    n_bridges = 0;

    if (object->sensClass == TACTMAP_WATER)
    {
	draw_lakes(tactmap, object, use_erase_gc, bounds, space);
	tactmap_draw_plugs(tactmap, object, use_erase_gc, bounds, space);
    }

    /* Determine which of roads & rivers are on the list of stuff
     * to draw.
     */
    do_roads = (tactmap->objects[TACTMAP_ROADS] &&
		(tactmap->objects[TACTMAP_ROADS]->display_mode == PERMANENT));
    do_rivers = (tactmap->objects[TACTMAP_WATER] &&
		 (tactmap->objects[TACTMAP_WATER]->display_mode == PERMANENT));

    if (use_erase_gc)
      ford_gc = water_gc = road_gc = tactmap->gc.background,
      do_roads = !do_roads, do_rivers = !do_rivers;
    else
      ford_gc = tactmap->gc.fordable,
      water_gc = tactmap->gc.water,
      road_gc = tactmap->gc.roads;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;
    snsobj.kind = SNSTVE_LINE;

    ford_gcv.line_width = 0;
    water_gcv.line_width = 0;
    road_gcv.line_width = 0;
    XChangeGC(XtDisplay(tactmap->widget), water_gc, GCLineWidth, &ford_gcv);
    XChangeGC(XtDisplay(tactmap->widget), ford_gc, GCLineWidth, &water_gcv);
    XChangeGC(XtDisplay(tactmap->widget), road_gc, GCLineWidth, &road_gcv);
    current_gcv = &road_gcv;

    while (n_verts = ctdb_next_linear_gcs(space, verts, &soil, &widthD, 
					  &ei, &esi, &ctdb))
    {
	if (tactmap_off_screen64(n_verts, verts, bounds))
	  continue;

	/* Which gc */
	is_a_road = FALSE;
	if (CTDB_WATER(soil) && CTDB_SLOW_GO(soil))
	{
	    if (!do_rivers)
	      continue;
	    current_gc = ford_gc, current_gcv = &ford_gcv;
	}
	else if (CTDB_WATER(soil) && CTDB_NO_GO(soil))
	{
	    if (!do_rivers)
	      continue;
	    current_gc = water_gc, current_gcv = &water_gcv;
	}
	else
	{
	    if (!do_roads)
	      continue;
	    current_gc = road_gc, current_gcv = &road_gcv;
	    is_a_road = TRUE;
	}

	/* Use the right sensitive info for roads/rivers */
	if (is_a_road)
	{
	    snsobj.sensClass =
	      tactmap->objects[TACTMAP_ROADS]->dfn.sensitive.sensClass;
	    snsobj.user_data =
	      tactmap->objects[TACTMAP_ROADS]->dfn.sensitive.user_data;
	}
	else
	{
	    snsobj.sensClass =
	      tactmap->objects[TACTMAP_WATER]->dfn.sensitive.sensClass;
	    snsobj.user_data =
	      tactmap->objects[TACTMAP_WATER]->dfn.sensitive.user_data;
	}

	/* If this is a road feature check for bridges.
	 */
	if (is_a_road)
	{
	    /* If this is the 0'th feature in the edge, check the first node.
	     */
	    if (!esi)
	      maybe_save_bridge(ctdb, ctdb->edges[ei]->first_node);

	    /* If this is the n-1'th feature in the edge, check the last node.
	     */
	    if (esi == ctdb->edges[ei]->num_features-1)
	      maybe_save_bridge(ctdb,ctdb->edges[ei]->last_node);
	}

	/* How wide is this segment? */
	width = widthD * ppm;
	if (width == 1) width = 0;
	if (current_gcv->line_width != width)
	{
	    current_gcv->line_width = width;
	    XChangeGC(XtDisplay(tactmap->widget), current_gc,
		      GCLineWidth, current_gcv);
	}

	/* Bundle up the points */
	for (j=0;j<n_verts;j++)
	{
	    pts[j].x = tactmap_FX_pixel_number(verts[j][X],cx,hsw,ppm);
	    pts[j].y = tactmap_FY_pixel_number(verts[j][Y],cy,hsh,ppm);
	}

	XDrawLines(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		   current_gc, pts, j, CoordModeOrigin);

	for (j=0;j<n_verts-1;j++)
	{
	    snsobj.instance = (ei << 8);
	    snsobj.attr.line.x0 = pts[j].x;
	    snsobj.attr.line.y0 = pts[j].y;
	    snsobj.attr.line.x1 = pts[j+1].x;
	    snsobj.attr.line.y1 = pts[j+1].y;
	    snsobj.attr.line.width = width + 10;

	    snstve_add_object(tactmap->sensitive, &snsobj);
	}
    }

    /* Draw the bridges */

    /* First, draw bridges that occur at road/water intersections by looking
     * at the linear topoligy.
     */
    road_gcv.line_width = PIXELS(2, ppm);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.bridge_symbol,
	      GCLineWidth, &road_gcv);

    for (j=0;j<n_bridges;j++)
      tactmap_draw_bridge(tactmap, &bridges[j]);

    /* Next, draw the bridges that occur explicitly as abstract features.
     */
    if(do_roads)
      while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_BRIDGE,
					  &vertices, &data, NULL))
      {
	  if (tactmap_off_screen32(n_verts, vertices, bounds))
	    continue;

	  /* Bundle up the points */
	  for (j=0;j<n_verts;j++)
	  {
	      pts[j].x =
		tactmap_FX_pixel_number(vertices[2*j],cx,hsw,ppm);
	      pts[j].y =
		tactmap_FY_pixel_number(vertices[2*j+1],cy,hsh,ppm);
	  }

	  if (use_erase_gc)
	    gc = tactmap->gc.background;
	  else
	    gc = tactmap->gc.bridge_picture;

	  XFillPolygon(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		       gc, pts, j, Nonconvex, CoordModeOrigin);
      }
}

#define INITIAL_SEGS 5000

/*ARGSUSED*/
static void 
draw_lakes(TACTMAP_PTR           tactmap, 
	   TACTMAP_OBJECT_PTR    object, 
	   int32                 use_erase_gc, 
	   TACTMAP_BOUNDS       *bounds,
	   CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    static XPoint *segs = NULL;
    static int32 num_segs = 0;
    float32 *verts;
    int32 n_verts;
    CTDB_ABSTRACT_DATA *data;
    XGCValues gcv;
    GC gc;
    int32 i, j;
    int32 level, max_level;
    int32 n_saved[16];
    struct
    {
	int32    soil_type;
	int32    n_verts;
	float32 *verts;
    } save_list[16][1024], *saved;

    if (!num_segs)
    {
        /* Initialize the persistent segs array */
        num_segs = INITIAL_SEGS;
        segs = (XPoint *)STDALLOC(INITIAL_SEGS * sizeof(XPoint));
    }

    gcv.line_width = 0;
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.water,
	      GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.fordable,
	      GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.no_go,
	      GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.slow_go,
	      GCLineWidth, &gcv);
    XChangeGC(XtDisplay(tactmap->widget), tactmap->gc.background,
	      GCLineWidth, &gcv);

    /* Iterate through all the soil defrags.  We have to draw them by
     * level number (first 0, then 1, etc.), so we first just pass through
     * and save them up.
     */
    bzero(n_saved, sizeof(n_saved));
    max_level = 0;
    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_SOIL_DEFRAG,
					&verts, &data, NULL))
    {
	if (n_verts < 3)
	  continue;

	level = data->soil_defrag.level;

	if (level > ARRAYLENGTH(n_saved))
	  continue;

	if (n_saved[level] == ARRAYLENGTH(save_list[0]))
	  continue;

	if (level > max_level)
	  max_level = level;

	save_list[level][n_saved[level]].soil_type =
	  ctdb_lookup_soil_from_defrag(tactmap->ctdb,data);
	save_list[level][n_saved[level]].n_verts = n_verts;
	save_list[level][n_saved[level]].verts = verts;
	n_saved[level]++;
    }

    /* Draw the steep areas, too (put them in the top level) */
    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_STEEP_SLOPE,
					&verts, &data, NULL))
    {
	if (n_verts < 3)
	  continue;

	if (n_saved[max_level] == ARRAYLENGTH(save_list[max_level]))
	  continue;

	save_list[max_level][n_saved[max_level]].soil_type = CTDB_NO_GO_FLAG;
	save_list[max_level][n_saved[max_level]].n_verts = n_verts;
	save_list[max_level][n_saved[max_level]].verts = verts;
	n_saved[max_level]++;
    }

#if 0
    /* Draw the contaminated areas, too (put them in level 0) */
    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_CONTAMINANT,
					&verts, &data, NULL))
    {
	if (n_verts < 3)
	  continue;

	if (n_saved[0] == ARRAYLENGTH(save_list[0]))
	  continue;

	save_list[0][n_saved[0]].soil_type = CTDB_NO_GO_FLAG;
	save_list[0][n_saved[0]].n_verts = n_verts;
	save_list[0][n_saved[0]].verts = verts;
	n_saved[0]++;
    }
#endif

    /* Now draw them in order */
    for (level = 0; level <= max_level; level++)
      for (i=0; i < n_saved[level]; i++)
      {
	  saved = &save_list[level][i];

	  if (tactmap_off_screen32(saved->n_verts, saved->verts, bounds))
	    continue;

          if (saved->n_verts > num_segs)
          {
              /* Grow the segs array so that all the points will fit */
              num_segs = saved->n_verts;
              segs = (XPoint *)STDREALLOC(segs, num_segs * sizeof(XPoint));
          }

	  /* Bundle up the points */
	  for (j=0;j<saved->n_verts;j++)
	  {
	      segs[j].x =
		tactmap_FX_pixel_number(saved->verts[(j<<1)],cx,hsw,ppm);
	      segs[j].y =
		tactmap_FY_pixel_number(saved->verts[(j<<1)+1],cy,hsh,ppm);
	  }

	  if (use_erase_gc)
	    gc = tactmap->gc.background;
	  else
	  {
	      int32     soil_type = saved->soil_type;

	      if((CTDB_WATER(soil_type) && CTDB_NO_GO(soil_type)) ||
		 CTDB_SIMNET_SOIL(soil_type) == 14)
		gc = tactmap->gc.water;
	      else if(CTDB_WATER(soil_type) && (CTDB_SLOW_GO(soil_type)))
		gc = tactmap->gc.fordable;
	      else if(CTDB_SLOW_GO(soil_type))
		gc = tactmap->gc.slow_go;
	      else if(CTDB_NO_GO(soil_type))
		gc = tactmap->gc.no_go;
	      else
		gc = tactmap->gc.background;
	  }

	  XFillPolygon(XtDisplay(tactmap->widget), tactmap->terrain_pix,
		       gc, segs, j, Nonconvex, CoordModeOrigin);
      }
}

#endif /*USE_X*/
