#ifndef lint
static char rcsid [] = "$RCSfile: tmap_wire.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tmap_wire.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to draw wire                                         *
*       Created: Wed Feb 14 1996                                            *
*       Author: mlongtin                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libtmap_local.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <libvecmat.h>
#include <math.h>

#define MIN_ARC_SIZE         2     /* pixels */
#define FULL_REVOLUTION  23040 /* (360 * 64) */

static void draw_wire_segment(Display *display, Window pixmap, GC gc,
			      int32 x0, int32 y0, int32 x1, int32 y1,
			      int32 type, float32 width,
			      XRectangle *clip_box, float64 ppm)
{
    int32 j, narcs;
    int32 x,y,hx,hy,nx,ny;
    int32 arc_size_cost, arc_size_sint;
    int32 arc_size, arc_space;
    float64 dist,dx,dy, offset[2];
    float64 sint, cost;
    XGCValues gcv;


    if(type == CTDB_FS_MINEFIELD_FENCE)
      gcv.line_style = LineOnOffDash;
    else
      gcv.line_style = LineSolid;

    XChangeGC(display, gc, GCLineStyle, &gcv);

    switch(type)
    {
      case CTDB_FS_CONCERTINA_WIRE:
      case CTDB_FS_WIRE_ROAD_BLOCK:

	arc_size = width * ppm;
#if 0
	if (type == CTDB_FS_WIRE_ROAD_BLOCK)
	  arc_size /= 2;
#endif
	if(arc_size < MIN_ARC_SIZE)
	  arc_size = MIN_ARC_SIZE;

	arc_space = -0.5 * arc_size;

	dx = x1 - x0;
	dy = y1 - y0;

	/* Compute the normal of (dx, dy). */
	offset[X] = -dy;
	offset[Y] = dx;
	vmat2_unit64(offset, offset);
	VMAT2_SCAL_VEC_MUL(arc_space, offset, offset);

	dist = sqrt(dx*dx + dy*dy);
	sint = dy / dist;
	cost = dx / dist;

	narcs = dist / (arc_size + arc_space);

	if(!narcs)
	  narcs = 1;

	dx /= narcs;
	dy /= narcs;

	arc_size_cost = arc_size * cost * 0.5;
	arc_size_sint = arc_size * sint * 0.5;

	for (j=0;j<narcs;j++)
	{
	    x = (int32)(x0 + dx * (j) + arc_size_cost - arc_size * 0.5);
	    y = (int32)(y0 + dy * (j) + arc_size_sint - arc_size * 0.5);

	    /* This is not a speed optimization. There seems to
	     * be a bug in X which causes it to die if arcs are drawn
	     * far off the screen.
	     */
	    if (!POINT_IN_RECT(x,y,*clip_box))
	      continue;

	    XDrawArc(display, pixmap, gc, x, y,
		     arc_size, arc_size, 0, FULL_REVOLUTION);
#if 0
	    if(type != CTDB_FS_CONCERTINA_WIRE)
	    {
		XDrawArc(display, pixmap, gc, x + (int32)offset[X],
			 y + (int32)offset[Y], arc_size, arc_size, 0,
			 FULL_REVOLUTION);		
		XDrawArc(display, pixmap, gc, x - (int32)offset[X],
			 y - (int32)offset[Y], arc_size, arc_size, 0,
			 FULL_REVOLUTION);
	    }
#endif

	}
	break; 

      case CTDB_FS_FENCE:
      case CTDB_FS_MINEFIELD_FENCE:

	XDrawLine(display, pixmap, gc, x0, y0, x1, y1);
	break;
    }
}

void tactmap_draw_wire(TACTMAP_PTR tactmap, TACTMAP_OBJECT_PTR object, 
		       int32 use_erase_gc, TACTMAP_BOUNDS *bounds,
		       CTDB_SEARCH_SPACE_PTR space)
{
    DECLARE_REGS;
    float64 verts[CTDB_VERTICES_MAX][XY];
    int32 n_verts;
    int32 j;
    int32 x, y, last_x=0, last_y=0;
    XGCValues gcv;
    GC gc;
    SNSTVE_OBJECT snsobj;
    XRectangle bounding_box;
    CTDB_LINEAR_DESC feature_info;

    bounding_box.x      = tactmap_X_pixel_number(bounds->min_x,cx,hsw,ppm);
    bounding_box.y      = tactmap_Y_pixel_number(bounds->max_y,cy,hsh,ppm);
    bounding_box.width  = PIXELS(bounds->max_x-bounds->min_x,ppm);
    bounding_box.height = PIXELS(bounds->max_y-bounds->min_y,ppm);

    if (use_erase_gc)
      gc = tactmap->gc.background;
    else
      gc = tactmap->gc.wire;

    bzero((char *)&snsobj, sizeof(snsobj));
    snsobj.sensClass = object->dfn.sensitive.sensClass;
    snsobj.user_data = object->dfn.sensitive.user_data;

    while (n_verts = ctdb_next_linear_ref(space, verts, NULL, &feature_info))
    {
	if (n_verts == 1)
	  continue;

	if(feature_info.type != CTDB_FS_CONCERTINA_WIRE &&
	   feature_info.type != CTDB_FS_WIRE_ROAD_BLOCK &&
	   feature_info.type != CTDB_FS_FENCE &&
	   feature_info.type != CTDB_FS_MINEFIELD_FENCE)
	  continue;

	if (object->dfn.wire.mask & TACTMAP_WIRE_LINES)
	{
	    for (j=0;j<n_verts;j++)
	    {
		x = tactmap_FX_pixel_number(verts[j][X],cx,hsw,ppm);
		y = tactmap_FY_pixel_number(verts[j][Y],cy,hsh,ppm);

		if (j)
		{
		    snsobj.kind = SNSTVE_LINE;
		    snsobj.instance = 0;
		    snsobj.attr.line.x0 = last_x;
		    snsobj.attr.line.y0 = last_y;
		    snsobj.attr.line.x1 = x;
		    snsobj.attr.line.y1 = y;
		    snsobj.attr.line.width = 10;

		    snstve_add_object(tactmap->sensitive, &snsobj);

		    draw_wire_segment(XtDisplay(tactmap->widget),
				      tactmap->terrain_pix, gc,
				      last_x, last_y, x, y,
				      feature_info.type,
				      feature_info.width,
				      &bounding_box, ppm);
		}

		last_x = x;
		last_y = y;
	    }
	}
    }
}

#endif /*USE_X*/
