#ifndef lint
static char rcsid [] = "$RCSfile: time_clock.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: time_clock.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government         *
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause at               *
*       DFARS 252.227-7013 (OCT 1988).                                      *
*                                                                           *
*       Contents: Simulation and real time clock routines                   *
*                                                                           *
*       Created: 5/9/89                                                     *
*       Author: ajc                                                         *
*       Remarks:                                                            *
*                                                                           *
*       Modified: 7/20/93                                                   *
*       Author: Bryant Collard                                              *
*       Remarks: Moved code associated with real time to time_realtime.c.   *
*                Modified to deal with fixed frame simulations.             *
*                                                                           *
****************************************************************************/

#include "libtime.h"
#include <stdext.h> /*common/include/global*/
#include <stdstring.h>

/* globally accesable value */
uint32 time_last_simulation_clock = 0;
uint32 time_last_simulation_dis_clock = 0;
#ifdef USE_RTI
RTI_FEDERATION_TIME time_last_simulation_rti_clock = 0.0;
#endif

/* locals */
static float64 simulation_clock = 0.0;
static float64 simulation_rate = 1.0;
static float64 realtime_rate = 1.0;
static float64 last_rate = -1.0;
static uint32 rate_limit = 0;
static int simtime_mode = TIME_VARIABLE_FRAME;
static uint32 next_frame_advance_time = 0;
static uint32 frame_time = 67;
static uint32 realtime_at_last_sim_frame = 0;

static void rate_reset();

int32 time_get_current_frame_mode(fixed_frame_time)
    uint32 fixed_frame_time[1];
{
    if(fixed_frame_time)
    {
	if(simtime_mode != TIME_VARIABLE_FRAME)
	  *fixed_frame_time = frame_time;
    }
    return(simtime_mode);
}

void time_set_current_frame_mode(mode, fixed_frame_time)
    int32 mode;
    uint32 fixed_frame_time;
{
    switch(mode)
    {
      default:
      case TIME_VARIABLE_FRAME:
	time_set_variable_frame_mode ();
	break;

      case TIME_REALTIME_SYNC_FIXED_FRAME:
	time_set_realtime_sync_fixed_frame_mode (fixed_frame_time);
	break;
      case TIME_EXTERNAL_SYNC_FIXED_FRAME:
	time_set_external_sync_fixed_frame_mode (fixed_frame_time);
	break;
      case TIME_ASYNC_FIXED_FRAME:
	time_set_async_fixed_frame_mode (fixed_frame_time);
	break;
    }
}

void time_set_variable_frame_mode ()
{
    if (simtime_mode != TIME_VARIABLE_FRAME)
    {
        simulation_clock = (float64) time_last_simulation_clock;
        simtime_mode = TIME_VARIABLE_FRAME;
    }
}


void time_set_realtime_sync_fixed_frame_mode (fixed_frame_time)
uint32 fixed_frame_time;
{
    if (simulation_rate <= 0.0)
    {
        printf ("libtime: Cannot enter TIME_REALTIME_SYNC_FIXED_FRAME mode ");
        printf ("when simulation_rate = 0\n");
        return;
    }

    frame_time = fixed_frame_time;
    next_frame_advance_time = realtime_at_last_sim_frame +
            (uint32) ((float64) frame_time * realtime_rate);
    simtime_mode = TIME_REALTIME_SYNC_FIXED_FRAME;
}

void time_set_external_sync_fixed_frame_mode (fixed_frame_time)
uint32 fixed_frame_time;
{
    frame_time = fixed_frame_time;
    simtime_mode = TIME_EXTERNAL_SYNC_FIXED_FRAME;
}

void time_set_async_fixed_frame_mode (fixed_frame_time)
uint32 fixed_frame_time;
{
    frame_time = fixed_frame_time;
    simtime_mode = TIME_ASYNC_FIXED_FRAME;
}

#ifdef USE_RTI
#include <libntpif.h>
static RTI_FEDERATION_RATE fed_rate = 1.0;
static int32               fed_is_paused = 0;
static RTI_FEDERATION_TIME fed_pause_time = 0.0;
static RTI_WALL_CLOCK_TIME fed_resume_time = 0.0;
static char                fed_pause_label[1024];
static RTI_RESULT rti_new_rate(RTI_FEDERATION_RATE the_rate)
{
    fed_rate = the_rate;

    if (fed_is_paused)
      the_rate = 0.0;

    realtime_rate = 1.0 / the_rate;
    rate_limit = 0;
    simulation_rate = the_rate;
    return RTI_OK;
}
static RTI_RESULT rti_pause_now(RTI_PAUSE_LABEL label)
{
    printf("RTI Pausing: %s\n", label);
    fed_is_paused = 1;
    rti_new_rate(fed_rate);
    rti_pause_achieved(time_last_simulation_rti_clock, label);
    return RTI_OK;
}
static RTI_RESULT rti_resume_now()
{
    printf("RTI Resuming\n");
    fed_is_paused = 0;
    rti_new_rate(fed_rate);
    rti_resume_achieved();
    return RTI_OK;
}
static RTI_RESULT rti_time_advance_grant(RTI_FEDERATION_TIME the_time)
{
    if ((fed_pause_time != 0.0) && (the_time >= fed_pause_time))
    {
	rti_pause_now(fed_pause_label);
	fed_pause_time = 0.0;
    }

    time_last_simulation_clock = the_time * 1000.0;
    time_last_simulation_rti_clock = the_time;
    time_last_simulation_dis_clock =
      TIME_MS_TO_DIS(time_last_simulation_clock);
    return RTI_OK;
}
static void rti_check_resume()
{
    if (fed_resume_time != 0.0)
    {
	NTP_TIMEVAL tv;
	ntp_get_timeval(&tv);
	if (fed_resume_time <= (tv.sec + tv.usec * 0.000001))
	{
	    rti_resume_now();
	    fed_resume_time = 0.0;
	}
    }
}
static RTI_RESULT rti_pause
  (RTI_FEDERATION_TIME the_time, RTI_PAUSE_LABEL label)
{
    strcpy(fed_pause_label, label);
    fed_pause_time = the_time;
    return RTI_OK;
}
static RTI_RESULT rti_resume(RTI_WALL_CLOCK_TIME the_time)
{
    fed_resume_time = the_time;
    return RTI_OK;
}

void time_init_from_rti(void)
{
    RTI_FEDERATION_TIME the_time;
    RTI_FEDERATION_RATE the_rate;
    static int32 initted = 0;

    if (!rti_request_federation_time(&the_time))
      rti_time_advance_grant(the_time);
    if (!rti_request_federation_rate(&the_rate))
      rti_new_rate(the_rate);

    if (!initted)
    {
	RTI_SET_CALLBACK(TIME_ADVANCE_GRANT, rti_time_advance_grant);
	RTI_SET_CALLBACK(NEW_FEDERATION_TIME, rti_time_advance_grant);
	RTI_SET_CALLBACK(NEW_FEDERATION_RATE, rti_new_rate);
	RTI_SET_CALLBACK(PAUSE, rti_pause);
	RTI_SET_CALLBACK(PAUSE_NOW, rti_pause_now);
	RTI_SET_CALLBACK(RESUME, rti_resume);
	RTI_SET_CALLBACK(RESUME_NOW, rti_resume_now);
    }
    initted = 1;
}
#endif

/* Call before the fork on MP */
void time_init(max_forward_step)
uint32 max_forward_step;
{
    static int32 clocks_initted = 0;

    if (clocks_initted)
      return;

    clocks_initted = 1;
    time_reset (max_forward_step);
}

void time_reset (max_forward_step)
uint32 max_forward_step;
{
    time_realtime_init (max_forward_step);
    time_simtime_init ();
}

void time_simtime_init ()
{
    time_last_simulation_clock = 0;
    time_last_simulation_dis_clock = 0;
#ifdef USE_RTI
    time_last_simulation_rti_clock = 0.0;
#endif
    simulation_clock = 0.0;
    realtime_at_last_sim_frame = time_realtime_clock ();
    next_frame_advance_time = (uint32) ((float64) frame_time * realtime_rate) +
            realtime_at_last_sim_frame;
}

int time_check_simulation_clock_advance ()
{
    return ((simtime_mode != TIME_REALTIME_SYNC_FIXED_FRAME) ||
            (time_realtime_clock () >= next_frame_advance_time));
}

static void rate_reset(now, deadline)
    float64 now;
    uint32 deadline;
{
    /* If we are supposed to stop the acceleration at a given time,
     * see if that time has come.
     */
    if (deadline && (now >= deadline))
    { 
	/* Restore simulation rate indefinitely*/
	deadline = 0;

	if(last_rate >= 0.0) 
	  /* last_rate will only be >= 0.0 if run_rate was called */
	{
	    printf("Simulation rate being reset to %.1f\n", last_rate);
	    time_set_simulation_rate(last_rate, 0);
	    printf("Realtime rate has been reset to %.1f\n", realtime_rate);
	    printf("Simulation rate has been reset to %.1f\n", simulation_rate);
	    last_rate = -1.0;
	}
	else 
	{
	    time_set_simulation_rate(1.0, 0);
	}
    }
}

uint32 time_advance_simulation_clock()
{
    uint32 current_time;
    float64 delta_time;

    switch (simtime_mode)
    {
      case TIME_VARIABLE_FRAME:
        current_time = time_realtime_clock ();
        frame_time = current_time - realtime_at_last_sim_frame;
        realtime_at_last_sim_frame = current_time;

        /* Update the simulation clock */
        if (simulation_rate == 1.0)
            simulation_clock += (float64) frame_time;
        else
        {
            delta_time = (float64) frame_time * simulation_rate;
            simulation_clock += delta_time;
            frame_time = (uint32) delta_time;
        }
	rate_reset(simulation_clock, rate_limit);

        /* Cache the value */
#ifdef USE_RTI
	if (rti_in_use)
	{
	    /* When using the RTI, we cannot just advance time.
	     * Instead we have to ask that time be advanced and
	     * hope that it is granted right away.
	     */
	    rti_check_resume();
	    rti_time_advance_request(simulation_clock * 0.001);
	}
	else
#endif
        time_last_simulation_clock = simulation_clock;
        break;
      case TIME_REALTIME_SYNC_FIXED_FRAME:
        time_last_simulation_clock += frame_time;
        realtime_at_last_sim_frame = time_realtime_clock ();
        while (realtime_at_last_sim_frame >= next_frame_advance_time)
            next_frame_advance_time +=
                    (uint32) ((float64) frame_time * realtime_rate);
	rate_reset(simulation_clock, rate_limit);
        break;
      case TIME_EXTERNAL_SYNC_FIXED_FRAME:
      case TIME_ASYNC_FIXED_FRAME:
        realtime_at_last_sim_frame = time_realtime_clock ();
        time_last_simulation_clock += frame_time;
        break;
      default:
        break;
    }

    time_last_simulation_dis_clock =
            TIME_MS_TO_DIS(time_last_simulation_clock);
#ifdef USE_RTI
    time_last_simulation_rti_clock =
      time_last_simulation_clock * 0.001;
#endif
    return(time_last_simulation_clock);
}

void time_set_simulation_rate(rate, limit)
    float64 rate;
    uint32  limit;
{
    if (rate < 0.0)
    {
	printf("Simulated time may not go backward\n");
	return;
    }
    else if (rate <= 0.0)
    {
        if (simtime_mode == TIME_REALTIME_SYNC_FIXED_FRAME)
        {
            printf ("libtime: Cannot set simulation rate to 0 in \n");
            printf ("TIME_REALTIME_SYNC_FIXED_FRAME mode.\n");
            printf ("Set the frame time to zero to stop simulation time.\n");
            return;
        }
    }

#ifdef USE_RTI
    if (rti_in_use)
    {
	/* We cannot change the rate, we can only request that it change */
	if (rate == 0.0)
	{
	    char buf[256];
	    gethostname(buf, sizeof(buf));
	    rti_fex_request_pause_now(buf);
	}
	else
	{
	    rti_fex_request_resume_now();
	    rti_fex_change_federation_rate(rate);
	}
	return;
    }
#endif

    if (rate != 0.0)
      realtime_rate = 1.0 / rate;
    rate_limit = limit;
    simulation_rate = rate;
}

void time_run_simulation_rate(rate, limit)
    float64 rate;
    uint32  limit;
{
    last_rate = simulation_rate;
    time_set_simulation_rate(rate, limit);
}

/* Returns the simulation rate. */
float64 time_get_simulation_rate()
{
    return simulation_rate;
}


int time_set_frame_time (fixed_frame_time)
uint32 fixed_frame_time;
{
    if (simtime_mode == TIME_VARIABLE_FRAME)
      return (FALSE);

    frame_time = fixed_frame_time;
    return (TRUE);
}

uint32 time_simulation_to_realtime(simtime)
    uint32 simtime;
{
    int32 delta;

    if (simtime_mode == TIME_ASYNC_FIXED_FRAME)
      return (0xffffffff);

    /* How long from now is this time? */
    delta = simtime - time_last_simulation_clock;

    /* How long is that in real time? */
    if (simulation_rate == 0.0)
      return 0xffffffff;
    if (simulation_rate == 1.0)
      return realtime_at_last_sim_frame + delta;
    return realtime_at_last_sim_frame + ((double)delta * realtime_rate);
}

uint32 time_realtime_to_simulation(realtime)
    uint32 realtime;
{
    int32 delta;

    /* How long from now is this time? */
    delta = realtime - realtime_at_last_sim_frame;

    if ((simtime_mode == TIME_ASYNC_FIXED_FRAME) ||
	(simulation_rate == 0.0) ||
	(simulation_rate == 1.0))
    {
	return (time_last_simulation_clock + delta);
    }

    /* How long is that in simulation time? */
    return time_last_simulation_clock + ((double)delta / realtime_rate);
}

uint32 time_simulation_time ()
{
    return (time_last_simulation_clock);
}

uint32 time_frame_time ()
{
    return (frame_time);
}


uint32 time_realtime_last_sim_frame ()
{
    return ( realtime_at_last_sim_frame ) ;
}
