#ifndef lint
static char rcsid [] = "$RCSfile: time_tdg.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: time_tdg.c                                                        *
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Time Date Group functions                                 *
*       Created: Fri Oct 12 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libtime.h"
#include <basic.h>
#include <sys/types.h>
#include <time.h>
#include <stdstring.h> /*common/include/global*/

static int32 century = 2000; /* adjust this at turn of century */

/* This isn't always prototyped in <time.h> */
extern time_t time();

#ifndef __isleap
/* Nonzero if YEAR is a leap year (every 4 years,
   except every 100th isn't, and every 400th is).  */
#define	__isleap(year)	\
  ((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))
#endif

static int32 dd_per_mm[] = { 31,28,31,30,31,30,31,31,30,31,30,31 };
/* Short utility function for time_mktime */
#include <stdio.h>
#include <ctype.h>

/* Print the values in a `struct tm'. */
void time_print_tms (tms)
    struct tm *tms;
{
    printf ("%02d/%02d/%02d %02d:%02d:%02d wday:%02d yday:%03d DST:%d ",
	    tms->tm_mon,
	    tms->tm_mday,
	    tms->tm_year,
	    tms->tm_hour,
	    tms->tm_min,
	    tms->tm_sec,
	    tms->tm_wday,
	    tms->tm_yday,
	    tms->tm_isdst);
}


/* Short utility function for time_mktime */
static void normalize(incr, min, max, unit)
    int32 *incr, 
           min, 
           max, 
          *unit;
{
    while (*incr < min) 
    { 
	(*unit)--; 
	*incr = max - ((min - 1) - *incr); 
    } 
    while (*incr > max) 
    { 
	(*unit)++; 
	*incr = min + (*incr - (max + 1)); 
    }
    
}

time_t time_mktime(tms_orig)
    struct tm *tms_orig;
{
    struct tm tms_mem; 
    struct tm *tms;
    int32 i;
    struct tm *tms_guess;
    time_t seconds_guess = 0;
    time_t distance = 0;
    static int32 times_through_search;
    uint32 tms_secs, tms_guess_secs;
    time_t tmp_distance = 0;
    int32 diff_flag = 0;
    int32 year;
    
    tms = &tms_mem;
    
    bcopy((char *)tms_orig, (char *)tms, sizeof(*tms));
    
    year = tms->tm_year + century - 100;
    if(tms->tm_isdst > 0)
    {
	/* we assume these values are in the alternate time zone, and
	 * that it is off by -1 hour of the main time zone. */
	tms->tm_sec += 3600;
	tms->tm_isdst = -1;
    }

    normalize (&(tms->tm_sec), 0, 59, &(tms->tm_min)); 
    /* seconds after the minute - [0, 59] */ 

    normalize (&(tms->tm_min), 0, 59, &(tms->tm_hour));
    /* minutes after the hour - [0, 59] */
 
    normalize (&(tms->tm_hour), 0, 23, &(tms->tm_mday));
    /* hour since midnight - [0, 23] */
    
    /* Do the month first, so day range can be found. */
    normalize (&(tms->tm_mon), 0, 11, &(tms->tm_year));
    /* 0-11    month */
    
    if(__isleap(tms->tm_year))
      dd_per_mm[1] = 29;
    else 
      dd_per_mm[1] = 28;

    normalize (&(tms->tm_mday), 1, dd_per_mm[tms->tm_mon], &(tms->tm_mon));
    /* 1-31    day of month */


    /* Do the month again, because the day may have pushed it out of range. */
    normalize (&(tms->tm_mon), 0, 11, &(tms->tm_year));
    /* 0-11    month */

    /* Do the day again, because the month may have changed the range. */
    normalize (&(tms->tm_mday), 1, dd_per_mm[tms->tm_mon], &(tms->tm_mon));
    /* 1-31    day of month */

    /*
     ** set wday and yday: 
     */

    tms->tm_yday = 0;
    for(i = 0; i < tms->tm_mon; i++)
      tms->tm_yday += dd_per_mm[i];
    tms->tm_yday += (tms->tm_mday - 1);

    /* 
     ** first set day that Jan 1 starts on for this year: 
     */
    /* normal gregorian calendar - one extra day per four years */
    tms->tm_wday = 4 + year + (year + 3)/4;
    /* julian calendar: regular gregorian less three days per 400 */
    if(year > 1800) 
    {
	tms->tm_wday -= (year - 1701)/100;
	tms->tm_wday += (year - 1601)/400;
    }
    /* great calendar changeover instant: */
    if(year > 1752)
      tms->tm_wday += 3;
    tms->tm_wday %=7;
    /* OK, wday now has the day that Jan 1 starts on for this year: */
    tms->tm_wday = (tms->tm_yday + tms->tm_wday) % 7;
       

   /* Direct calculation of the number of seconds is difficult
    * because you need to know the number of days in every year and
    * month.  Instead), we make a good guess assuming there are 365
    * days in every year, and 30 days in every month; then we can
    * just do a simple binary search to quickly get the exact
    * number. 
    *
    * (this is very similair to the algorithm used for
    * time_tdg_from_date_time) 
    */

    times_through_search = 0;
    do
    {
	seconds_guess += distance;
	
	times_through_search++;     

	tms_guess = localtime (&seconds_guess);
      
	/* Check the guess: 
	 */

	diff_flag = 0;
	tms_secs = tms_guess_secs = 0;

	/* The diff_flag tells us not only if tms_secs <
	 * tms_guess_secs, but also if any of the components making up
	 * those numbers are different (imagine tms has a leap year,
	 * and tms_guess doesn't take this into account, then tms_secs
	 * and tms_guess_secs could be the same, but tms_guess's date
	 * is off by a day.
	 */
#define set_diff_flag(unit, secs_per_unit)           \
  tms_secs += tms->unit * secs_per_unit;              \
  tms_guess_secs += tms_guess->unit * secs_per_unit;             \
  if (!diff_flag)                                    \
    {                                                \
      if (tms->unit < tms_guess->unit)          \
	diff_flag = -1;                              \
      else if (tms->unit > tms_guess->unit)     \
	diff_flag = 1;                               \
     }

        /* Do the biggest unit first so the diff_flag gets set
	 * properly: */
        set_diff_flag (tm_year, 31536000); /* assume 365 days/year */
	set_diff_flag (tm_mon, 2592000);   /* assume 30 days/month */
	set_diff_flag (tm_mday, 86400);
	set_diff_flag (tm_hour, 3600);
	set_diff_flag (tm_min, 60);
	set_diff_flag (tm_sec, 1);
  
#undef set_diff_flag
  
	/* We should also make sure that the sign of TMP_DISTANCE is
	   correct -- if DIFF_FLAG is positive, the distance should be
	   positive and vice versa. 
	   */
  
	tmp_distance = (tms_secs > tms_guess_secs) ? 
	  (tms_secs - tms_guess_secs) : (tms_guess_secs - tms_secs);
	if (diff_flag < 0)
	  tmp_distance = -tmp_distance;

	if (times_through_search > 20) 
	  /* Arbitrary # of calls, but makes sure we
	     never hang if there's a problem with
	     this algorithm.  */
	  return -1;

	if ((tmp_distance == 0) && diff_flag)
	  /* Add a day (86400secs/day) */
	  tmp_distance = 86400 * diff_flag; 

	distance = tmp_distance;
	
    } while (distance != 0);

    /* localtime has set the daylight savings time flag for us: */
    tms->tm_isdst = tms_guess->tm_isdst;

    printf("guess from localtime:  ");
    time_print_tms(tms_guess);
    printf("\n");
    printf("from mktime:  ");
    time_print_tms(tms);
    printf("\n\n");

    if(tms->tm_isdst > 0)
      /* We're in the "alternate time zone"; assume this is 1 hour off
       * from the main time zone */
      seconds_guess -= 3600;

    bcopy(tms, tms_orig, sizeof(*tms));

    /* At this point, you know you've guess the right number of
     * seconds, so just return your guess. */
    return seconds_guess;

}

uint32 time_parse_date(s)
    char *s; 
{
    int32 val, yy,mm,dd;

    if (!s || (strlen(s) != 6) || !sscanf(s,"%d",&val))
      return 0;

    yy = val / 10000;
    if (yy < 0 || yy > 99)
      return 0;

    if(__isleap(yy + century - 100))
      dd_per_mm[2] = 29;
    else
      dd_per_mm[2] = 28;

    mm = (val - (10000 * yy)) / 100;
    if (mm < 1 || mm > 12)
      return 0;
    dd = (val - (10000 * yy) - (100 * mm));
    if (dd < 1 || dd > dd_per_mm[mm-1])
      return 0;
    return val;
}

/*
 * Returns -1 if string is in bad format.
 */
uint32 time_parse_time(s)
    char *s;
{
    int32 val, hh,mm;

    if (!s || (strlen(s) != 4) || !sscanf(s,"%d",&val))
      return -1;
    hh = val/100;
    if (hh < 0 || hh > 23 )
      return -1;
    mm = val - (100 * hh);
    if (mm < 0 || mm > 59)
      return -1;
    return val;
}

void time_tdg_from_seconds(seconds, tdg_ptr)
    uint32 seconds;
    TimeDateGroup *tdg_ptr;
{
    struct tm *lt;
    time_t sec;

    sec = seconds;
    lt = localtime(&sec);
    tdg_ptr->secondsSince1970 = sec;

    tdg_ptr->date = lt->tm_year*10000 + (lt->tm_mon + 1)*100 + lt->tm_mday;
    tdg_ptr->time = lt->tm_hour*100 + lt->tm_min;
}

#define SECONDS_PER_MINUTE    (60)
#define SECONDS_PER_HOUR      (60*SECONDS_PER_MINUTE)
#define SECONDS_PER_DAY       (24*SECONDS_PER_HOUR)
#define MAX_SECONDS_PER_MONTH (31*SECONDS_PER_DAY)
#define MAX_SECONDS_PER_YEAR  (366*SECONDS_PER_DAY)
#define SECOND_TEST_SEED      ((2010 - 1970) * MAX_SECONDS_PER_YEAR)

/* The algorithm here is pretty simple:
 * Guess a value for seconds.
 * Adjust the value to guarantee the year is <= the objective.
 * Search forward to the correct year by halves.
 * Adjust the value to guarantee the month is <= the objective.
 * Search forward to the correct month by halves.
 * etc.
 * The order of this is roughly constant.  A cheaper solution is doable
 * but this only took a couple minutes to write.
 */
void time_tdg_from_date_time(date, time, tdg_ptr)
    uint32 date;
    uint32 time;
    TimeDateGroup *tdg_ptr;
{
    struct tm *lt;
    int32 minute, hour, day, month, year;
    int32 sec_test;

    minute = time%100;
    hour = time/100;
    day = date%100;
    month = (date%10000)/100 - 1;
    year = date/10000;

    sec_test = SECOND_TEST_SEED;
    for(;;)
    {
	lt = localtime(&sec_test);

	if (lt->tm_year > year)
	{ 
            sec_test -= (lt->tm_year - year) * MAX_SECONDS_PER_YEAR; 
            continue; 
        }
	else if (lt->tm_year < year)
	{ 
            sec_test += MAX_SECONDS_PER_YEAR / 2; 
            continue; 
        }
	if (lt->tm_mon > month)
	{ 
            sec_test -= (lt->tm_mon - month) * MAX_SECONDS_PER_MONTH; 
            continue;
        }
	else if (lt->tm_mon < month)
	{
	    sec_test += (month - lt->tm_mon) * MAX_SECONDS_PER_MONTH / 2;
	    continue;
	}
	if (lt->tm_mday > day)
	{ 
            sec_test -= (lt->tm_mday - day) * SECONDS_PER_DAY; 
            continue; 
        }
	else if (lt->tm_mday < day)
	{
	    sec_test += (day - lt->tm_mday) * SECONDS_PER_DAY / 2;
	    continue;
	}

	if (lt->tm_hour > hour)
	{ 
            sec_test -= (lt->tm_hour - hour) * SECONDS_PER_HOUR; 
            continue; 
        }
	else if (lt->tm_hour < hour)
	{
	    sec_test += (hour - lt->tm_hour) * SECONDS_PER_HOUR / 2;
	    continue;
	}

	if (lt->tm_min > minute)
	{ 
            sec_test -= (lt->tm_min - minute) * SECONDS_PER_MINUTE; 
            continue; 
        }
	else if (lt->tm_min < minute)
	{
	    sec_test += (minute - lt->tm_min) * SECONDS_PER_MINUTE / 2;
	    continue;
	}

	/* At this point, all the variables match up.  Adjust the seconds
	 * back to the beginning of the minute, and we're done.
	 */
	sec_test -= lt->tm_sec;
	break;
    }

    tdg_ptr->date = date;
    tdg_ptr->time = time;
    tdg_ptr->secondsSince1970 = sec_test;
}
    
void time_tdg_from_unix(tdg_ptr)
    TimeDateGroup *tdg_ptr;
{
    uint32 seconds = time(0);

    time_tdg_from_seconds(seconds, tdg_ptr);
}

