#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Rotation test jig                                         *
*       Created: Thu May 18 1995                                            *
*       Author: mcalef                                                      *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libvecmat.h>
#include <libcoordinates.h>
#include <math.h>
#include <librandom.h>
#include <libtime.h>
#include <stdio.h>

#define pi 3.14159
#define twopi 6.28318 

static void general_angle_to_rotation();
static void general_rotation_to_angle();

main()
{
   float64 mtrx_64[3][3];  	/* A 64 bit rotation matrix */
   float32 mtrx_32[3][3];       /* A 32 bit rotation matrix */
   float32 psi_i;		/* Initial euler angles */
   float32 theta_i; 		                      
   float32 phi_i;		                      
   float32 psi_f;		/* Final euler angles */                   
   float32 theta_f; 		 		      
   float32 phi_f;		 
   int i;			
   int j;
   int fs;			/* Function selection */ 

   COORD_TCC_PTR tcc;		/* Topocentric coordinate system */
   float64 gcc_tcc_rot[3][3];	/* Rotation matrix from geocentric coords to topocentric coords */
   float64 tcc_gcc_rot[3][3];	/* Rotation matric from topocentric coords to geocentric coords */
   float32 gcc_tcc_rot32[3][3];
   float32 tcc_gcc_rot32[3][3];

   /* Initialize utility libraries */
   reader_init(0);
   vtab_init(1, 1, 1);
   rdc_init("../../data", 0);
   coord_init("../../data", 0);
   time_init(10);

   /* Make a tcc */
   tcc = coord_define_tcc(COORD_GCC,
			   4155000.0, 545000.0, 16, 'S', 3, 50, 50);

   /* Create tcc_gcc_rot matrix */
   coord_tcc_gcc_rotation(tcc, tcc_gcc_rot);

   /* Use tcc_gcc_rot matrix to get gcc_tcc_rot matrix */
   vmat3_transpose64(tcc_gcc_rot, gcc_tcc_rot);

   for (i=0;i<3;i++)
     for (j=0;j<3;j++)
     {
	gcc_tcc_rot32[i][j] = (float32) gcc_tcc_rot[i][j];
	tcc_gcc_rot32[i][j] = (float32) tcc_gcc_rot[i][j];
     }
   
   printf("What combinations of functions to test?\n");   
   printf("   1) vmat3_gcc_angles32_to_rotation64 - vmat3_rotation64_to_gcc_angles32\n");
   printf("   2) vmat3_gcc_angles32_to_rotation32 - vmat3_rotation32_to_gcc_angles32\n");
   printf("   3) vmat3_zdown_angles32_to_rotation64 - vmat3_rotation64_to_zdown_angles32\n");
   printf("   4) vmat3_zdown_angles32_to_rotation32 - vmat3_rotation32_to_zdown_angles32\n");
   printf(">");
   scanf("%d",&fs);

   printf("Number of angle combinations to test? ");
   scanf("%d",&i);

   rnd_srandom(time_realtime_clock());

   while (i>0)
   {   
       if (i != 1)
       {
	   psi_i = (twopi * rnd_uniform_dist()) - pi;
	   theta_i = (pi * rnd_uniform_dist()) - (pi / 2.0);
	   phi_i = (twopi * rnd_uniform_dist()) - pi;
       }
       else
       {
	   psi_i = 0.0;
	   theta_i = 0.0;
	   phi_i = 0.0;
       }
       
       printf("Initial angles: psi: %.3f theta :%.3f phi: %.3f\n ",
	      psi_i,theta_i,phi_i);

       general_angle_to_rotation(fs, psi_i, theta_i, phi_i, mtrx_32, mtrx_64, gcc_tcc_rot,gcc_tcc_rot32); 
       
       switch (fs)
       {
	 case 1:
	 case 3: 
	   VMAT3_PRINT_MAT(Resulting 64 bit matrix :, mtrx_64);
	   break;
	 case 2:
	 case 4: 
	   VMAT3_PRINT_MAT(Resulting 32 bit matrix :, mtrx_32);
	   break;
       }
       
       general_rotation_to_angle(fs, &psi_f, &theta_f, &phi_f, mtrx_32, mtrx_64, tcc_gcc_rot, tcc_gcc_rot32);

       printf("Final angles: psi: %.3f theta :%.3f phi: %.3f\n\n\n",
	      psi_f,theta_f,phi_f);

       if ((fabs(psi_i-psi_f) > 0.001) ||
	   (fabs(theta_i-theta_f) > 0.001) ||
	   (fabs(phi_i-phi_f) > 0.001))
       {
	   fprintf(stderr, "Problem with angles:\n");
	   fprintf(stderr, "Initial angles: psi: %.3f theta :%.3f phi: %.3f\n",
		  psi_i,theta_i,phi_i);
	   fprintf(stderr, "Final angles: psi: %.3f theta :%.3f phi: %.3f\n",
		  psi_f,theta_f,phi_f);
       }

       i--;
   }
   
}					

static void general_angle_to_rotation(fs, psi, theta, phi, mtrx32, mtrx64, gcc_tcc, gcc_tcc32)
    int fs;
    float32 psi;
    float32 theta;
    float32 phi;
    float32 mtrx32[3][3];
    float64 mtrx64[3][3];
    float64 gcc_tcc[3][3];
    float32 gcc_tcc32[3][3];
{
    switch (fs)
    {
      case 1:
	vmat3_gcc_angles32_to_rotation64(psi, theta, phi, mtrx64, gcc_tcc);
	break;
      case 2:
	vmat3_gcc_angles32_to_rotation32(psi, theta, phi, mtrx32, gcc_tcc32);
	break;
      case 3:
	vmat3_zdown_angles32_to_rotation64(psi, theta, phi, mtrx64);
	break;
      case 4:
	vmat3_zdown_angles32_to_rotation32(psi, theta, phi, mtrx32);
	break;
    }	    
} 

static void general_rotation_to_angle(fs, psi, theta, phi, mtrx32, mtrx64, tcc_gcc, tcc_gcc32)
    int fs;
    float32 *psi;
    float32 *theta;
    float32 *phi;
    float32 mtrx32[3][3];
    float64 mtrx64[3][3];
    float64 tcc_gcc[3][3];
    float32 tcc_gcc32[3][3];
{    
    switch (fs)
    {
      case 1:
	vmat3_rotation64_to_gcc_angles32(mtrx64, psi, theta, phi, tcc_gcc);
	break;
      case 2:
	vmat3_rotation32_to_gcc_angles32(mtrx32, psi, theta, phi, tcc_gcc32);
	break;
      case 3:
	vmat3_rotation64_to_zdown_angles32(mtrx64, psi, theta, phi);
	break;
      case 4:
	vmat3_rotation32_to_zdown_angles32(mtrx32, psi, theta, phi);
	break;
    }
}

