/****************************************************************************
*   File: vmat_add.c                                                          *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to add/subtract vectors                              *
*       Created: Wed Jul 15 1992                                            *
*       Author: jesmith                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libvecmat.h"
#include <stdext.h> /*common/include/global*/
#include <math.h>

#define TWO_DIMENSIONAL(op, wx, wy, vx, vy, r)				\
{									\
    r[X] = wx op vx;							\
    r[Y] = wy op vy;							\
}

#define THREE_DIMENSIONAL(op, wx, wy, wz, vx, vy, vz, r)		\
{									\
    r[X] = wx op vx;							\
    r[Y] = wy op vy;							\
    r[Z] = wz op vz;							\
}

void vmat2_vec_add32(
    float32 w[2],
    float32 v[2],
    float32 r[2])
TWO_DIMENSIONAL(+, w[X], w[Y], v[X], v[Y], r)

void vmat3_vec_add32(
    float32 w[3],
    float32 v[3],
    float32 r[3])
THREE_DIMENSIONAL(+, w[X], w[Y], w[Z], v[X], v[Y], v[Z], r)

void vmat2_vec_add64(
    float64 w[2],
    float64 v[2],
    float64 r[2])
TWO_DIMENSIONAL(+, w[X], w[Y], v[X], v[Y], r)

void vmat3_vec_add64(
    float64 w[3],
    float64 v[3],
    float64 r[3])
THREE_DIMENSIONAL(+, w[X], w[Y], w[Z], v[X], v[Y], v[Z], r)

void vmat2e_vec_add32(
    float32 wx,
    float32 wy,
    float32 vx,
    float32 vy,
    float32 r[2])
TWO_DIMENSIONAL(+, wx, wy, vx, vy, r)

void vmat3e_vec_add32(
    float32 wx,
    float32 wy,
    float32 wz,
    float32 vx,
    float32 vy,
    float32 vz,
    float32 r[3])
THREE_DIMENSIONAL(+, wx, wy, wz, vx, vy, vz, r)

void vmat2e_vec_add64(
    float64 wx,
    float64 wy,
    float64 vx,
    float64 vy,
    float64 r[2])
TWO_DIMENSIONAL(+, wx, wy, vx, vy, r)

void vmat3e_vec_add64(
    float64 wx,
    float64 wy,
    float64 wz,
    float64 vx,
    float64 vy,
    float64 vz,
    float64 r[3])
THREE_DIMENSIONAL(+, wx, wy, wz, vx, vy, vz, r)

void vmat2_vec_sub32(
    float32 w[2],
    float32 v[2],
    float32 r[2])
TWO_DIMENSIONAL(-, w[X], w[Y], v[X], v[Y], r)

void vmat3_vec_sub32(
    float32 w[3],
    float32 v[3],
    float32 r[3])
THREE_DIMENSIONAL(-, w[X], w[Y], w[Z], v[X], v[Y], v[Z], r)

void vmat2_vec_sub64(
    float64 w[2],
    float64 v[2],
    float64 r[2])
TWO_DIMENSIONAL(-, w[X], w[Y], v[X], v[Y], r)

void vmat3_vec_sub64(
    float64 w[3],
    float64 v[3],
    float64 r[3])
THREE_DIMENSIONAL(-, w[X], w[Y], w[Z], v[X], v[Y], v[Z], r)

void vmat2e_vec_sub32(
    float32 wx,
    float32 wy,
    float32 vx,
    float32 vy,
    float32 r[2])
TWO_DIMENSIONAL(-, wx, wy, vx, vy, r)

void vmat3e_vec_sub32(
    float32 wx,
    float32 wy,
    float32 wz,
    float32 vx,
    float32 vy,
    float32 vz,
    float32 r[3])
THREE_DIMENSIONAL(-, wx, wy, wz, vx, vy, vz, r)

void vmat2e_vec_sub64(
    float64 wx,
    float64 wy,
    float64 vx,
    float64 vy,
    float64 r[2])
TWO_DIMENSIONAL(-, wx, wy, vx, vy, r)

void vmat3e_vec_sub64(
    float64 wx,
    float64 wy,
    float64 wz,
    float64 vx,
    float64 vy,
    float64 vz,
    float64 r[3])
THREE_DIMENSIONAL(-, wx, wy, wz, vx, vy, vz, r)
