/****************************************************************************
*   File: vmat_trans.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to do matrix transform operations                    *
*       Created: Wed Jul 15 1992                                            *
*       Author: jesmith                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libvecmat.h"
#include "vmat_trans.h"
#include <stdext.h> /*common/include/global*/
#include <stdio.h>
#include <math.h>

void vmat2_transpose32(
    float32 m[2][2],
    float32 r[2][2])
VMAT2_TRANSPOSE(float32, m, r)

void vmat3_transpose32(
    float32 m[3][3],
    float32 r[3][3])
VMAT3_TRANSPOSE(float32, m, r)

void vmat2_transpose64(
    float64 m[2][2],
    float64 r[2][2])
VMAT2_TRANSPOSE(float64, m, r)

void vmat3_transpose64(
    float64 m[3][3],
    float64 r[3][3])
VMAT3_TRANSPOSE(float64, m, r)

void vmat3_adjugate32(
    float32 m[3][3],
    float32 r[3][3])
{
    float32 s[3][3];

    if (m == r)
    {
	VMAT3_ADJUGATE(m, r)
    }
    else
    {
	VMAT3_ADJUGATE(m, s)
	bcopy((char *)s, (char *)r, sizeof(s));
    }
}

void vmat3_adjugate64(
    float64 m[3][3],
    float64 r[3][3])
{
    float64 s[3][3];

    if (m == r)
    {
	VMAT3_ADJUGATE(m, r)
    }
    else
    {
	VMAT3_ADJUGATE(m, s)
	bcopy((char *)s, (char *)r, sizeof(s));
    }
}

void vmat2_inverse32(
    float32 m[2][2],
    float32 r[2][2])
VMAT2_INVERSE(float32, m, r)

void vmat2_inverse64(
    float64 m[2][2],
    float64 r[2][2])
VMAT2_INVERSE(float64, m, r)

void vmat3_inverse32(
    float32 m[3][3],
    float32 r[3][3])
{
    float32 s[3][3], scale;

    scale = vmat3_determinant32(m);
    if (scale == 0.0)
    {
	printf("LIBVECMAT ERROR: Matrix uninvertable\n");
	if (m != r)
	  bcopy((char *)m, (char *)r, sizeof(m));
    }

    vmat3_adjugate32(m,s);
    vmat3_transpose32(s,s);
    vmat3_scal_mat_mul32(1.0/scale,s,r);
}

void vmat3_inverse64(
    float64 m[3][3],
    float64 r[3][3])
{
    float64 s[3][3], scale;

    scale = vmat3_determinant64(m);
    if (scale == 0.0)
    {
	printf("LIBVECMAT ERROR: Matrix uninvertable\n");
	if (m != r)
	  bcopy((char *)m, (char *)r, sizeof(m));
    }

    vmat3_adjugate64(m,s);
    vmat3_transpose64(s,s);
    vmat3_scal_mat_mul64(1.0/scale,s,r);
}

float32 vmat2_determinant32(
    float32 m[2][2])
{   return VMAT2_DETERMINANT(m);  }

float32 vmat3_determinant32(
    float32 m[3][3])
{   return VMAT3_DETERMINANT(m);  }

float64 vmat2_determinant64(
    float64 m[2][2])
{   return VMAT2_DETERMINANT(m);  }

float64 vmat3_determinant64(
    float64 m[3][3])
{   return VMAT3_DETERMINANT(m);  }




