/*
 * $RCSfile: vmat_trans.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: vmat_trans.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for vmat_matmat.c                     *
*       Created: Wed May 10 1995                                            *
*       Author:  bkenney                                                    *
*       Remarks:  These macros were lifted from vmat_trans.c so that        *
*                 other vmat files could easily reference them.             *
*                                                                           *
****************************************************************************/

#include <stdext.h> /*common/include/global*/
#include <math.h>

#define SWAP(a,b) hold=a;a=b;b=hold

#define VMAT2_TRANSPOSE(type, m, r)					\
{									\
    type hold;								\
									\
    if (m == r)								\
    {									\
	SWAP(r[X][Y],r[Y][X]);						\
    }									\
    else								\
    {									\
	r[X][X] = m[X][X]; r[X][Y] = m[Y][X];				\
	r[Y][X] = m[X][Y]; r[Y][Y] = m[Y][Y];				\
    }									\
}


#define VMAT3_TRANSPOSE(type, m, r)					\
{									\
    type hold;								\
									\
    if (m == r)								\
    {									\
	SWAP(r[X][Y],r[Y][X]);						\
	SWAP(r[X][Z],r[Z][X]);						\
	SWAP(r[Y][Z],r[Z][Y]);						\
    }									\
    else								\
    {									\
	r[X][X] = m[X][X]; r[X][Y] = m[Y][X]; r[X][Z] = m[Z][X];	\
	r[Y][X] = m[X][Y]; r[Y][Y] = m[Y][Y]; r[Y][Z] = m[Z][Y];	\
	r[Z][X] = m[X][Z]; r[Z][Y] = m[Y][Z]; r[Z][Z] = m[Z][Z];	\
    }									\
}



#define VMAT3_ADJUGATE(m, r)						\
{									\
    r[X][X] =  m[Y][Y] * m[Z][Z] - m[Z][Y] * m[Y][Z];			\
    r[X][Y] =  m[Z][X] * m[Y][Z] - m[Y][X] * m[Z][Z];			\
    r[X][Z] =  m[Y][X] * m[Z][Y] - m[Z][X] * m[Y][Y];			\
									\
    r[Y][X] =  m[Z][Y] * m[X][Z] - m[X][Y] * m[Z][Z];			\
    r[Y][Y] =  m[X][X] * m[Z][Z] - m[Z][X] * m[X][Z];			\
    r[Y][Z] =  m[Z][X] * m[X][Y] - m[X][X] * m[Z][Y];			\
									\
    r[Z][X] =  m[X][Y] * m[Y][Z] - m[Y][Y] * m[X][Z];			\
    r[Z][Y] =  m[Y][X] * m[X][Z] - m[X][X] * m[Y][Z];			\
    r[Z][Z] =  m[X][X] * m[Y][Y] - m[Y][X] * m[X][Y];			\
}


#define VMAT2_INVERSE(type, m, r)					     \
{									     \
    type scale, rxx, rxy, ryx, ryy;					     \
									     \
    scale = m[X][X] * m[Y][Y] - m[X][Y] * m[Y][X];			     \
    if (scale == 0.0)							     \
    {									     \
	printf("LIBVECMAT ERROR: Matrix uninvertable\n");		     \
	if (m != r)							     \
	  bcopy((char *)m, (char *)r, sizeof(m));			     \
    }									     \
    else								     \
    {									     \
	scale = 1.0 / scale;						     \
									     \
	rxx =  scale * m[Y][Y];						     \
	rxy = -scale * m[X][Y];						     \
	ryx = -scale * m[Y][X];						     \
	ryy =  scale * m[X][X];						     \
	r[X][X] = rxx;							     \
	r[X][Y] = rxy;							     \
	r[Y][X] = ryx;							     \
	r[Y][Y] = ryy;							     \
    }									     \
}






