/****************************************************************************
*   File: vmat_unit.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to make a unit vector                                *
*       Created: Wed Jul 15 1992                                            *
*       Author: jesmith                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libvecmat.h"
#include <stdext.h> /*common/include/global*/
#include <math.h>

#define TWO_DIMENSIONAL(type,vx,vy,r)					\
{									\
    type mag;								\
									\
    mag = vx*vx + vy*vy;						\
    if (mag == 0.0)							\
      r[X] = 1.0, r[Y] = 0.0;						\
    else								\
    {									\
	mag = 1.0/sqrt(mag);						\
	r[X] = vx * mag;						\
	r[Y] = vy * mag;						\
    }									\
}

#define THREE_DIMENSIONAL(type,vx,vy,vz,r)				\
{									\
    type mag;								\
									\
    mag = vx*vx + vy*vy + vz*vz;					\
    if (mag == 0.0)							\
      r[X] = 1.0, r[Y] = 0.0, r[Z] = 0.0;				\
    else								\
    {									\
	mag = 1.0/sqrt(mag);						\
	r[X] = vx * mag;						\
	r[Y] = vy * mag;						\
	r[Z] = vz * mag;						\
    }									\
}

void vmat2_unit32(
    float32 v[2],
    float32 r[2])
TWO_DIMENSIONAL(float32, v[X], v[Y], r)

void vmat3_unit32(
    float32 v[3],
    float32 r[3])
THREE_DIMENSIONAL(float32, v[X], v[Y], v[Z], r)

void vmat2_unit64(
    float64 v[2],
    float64 r[2])
TWO_DIMENSIONAL(float64, v[X], v[Y], r)

void vmat3_unit64(
    float64 v[3],
    float64 r[3])
THREE_DIMENSIONAL(float64, v[X], v[Y], v[Z], r)

void vmat2e_unit32(
    float32 vx,
    float32 vy,
    float32 r[2])
TWO_DIMENSIONAL(float32, vx, vy, r)

void vmat3e_unit32(
    float32 vx,
    float32 vy,
    float32 vz,
    float32 r[3])
THREE_DIMENSIONAL(float32, vx, vy, vz, r)

void vmat2e_unit64(
    float64 vx,
    float64 vy,
    float64 r[2])
TWO_DIMENSIONAL(float64, vx, vy, r)

void vmat3e_unit64(
    float64 vx,
    float64 vy,
    float64 vz,
    float64 r[3])
THREE_DIMENSIONAL(float64, vx, vy, vz, r)


