/*
 * $RCSfile: libvtab.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libvtab.h                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Libvtab public function/data definitions                  *
*       Created: Wed Apr  8 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBVTAB_H_INCLUDED
#define _LIBVTAB_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <libbasic.h>
#include <basic.h>
#undef Time
#undef Boolean

/* VTAB_USER_DATA_TYPE
 *
 * This is the type of data stored by libvtab.  You will probably cast
 * to or from the actual type to this type when dealing with libvtab.
 */
typedef uint32 VTAB_USER_DATA_TYPE;

/* Types of objects which *can* be stored in vehicle tables.
 * Note that traditionally, SAF has not stored remote flown out ammo.
 */
#define VTAB_LOCAL_VEHICLE      0x4001
#define VTAB_LOCAL_MISSILE      0x4002
#define VTAB_LOCAL_STRUCTURE    0x4004
#define VTAB_LOCAL_DESIGNATOR   0x4008
#define VTAB_LOCAL_STEALTH      0x4010
#define VTAB_LOCAL_ENV          0x4020
#define VTAB_LOCAL_AEFLARE      0x4040 /* temporarily for AE */
#define VTAB_LOCAL_MINE_MARKER  0x4080 
#define VTAB_LOCAL_BREACH_LANE  0x4100
#define VTAB_LOCAL_AGGREGATE    0x4200 /* Added for SAIC linkage work */
#define VTAB_LOCAL_PROXY        0x4400 /* Added for SAIC linkage work */
#define VTAB_LOCAL_RADIO        0x4800 /* Added for IVIS */
#define VTAB_LOCAL_MINEFIELD    0x5000
#define VTAB_LOCAL_MINE_SET     0x6000 
#define VTAB_REMOTE_VEHICLE     0x8001
#define VTAB_REMOTE_MISSILE     0x8002
#define VTAB_REMOTE_STRUCTURE   0x8004
#define VTAB_REMOTE_DESIGNATOR  0x8008
#define VTAB_REMOTE_STEALTH     0x8010
#define VTAB_REMOTE_ENV         0x8020
#define VTAB_REMOTE_AEFLARE     0x8040 /* temporarily for AE */
#define VTAB_REMOTE_MINE_MARKER 0x8080 
#define VTAB_REMOTE_BREACH_LANE 0x8100
#define VTAB_REMOTE_AGGREGATE   0x8200 /* Added for SAIC linkage work */
#define VTAB_REMOTE_PROXY       0x8400 /* Added for SAIC linkage work */
#define VTAB_REMOTE_RADIO       0x8800 /* Added for IVIS */
#define VTAB_REMOTE_MINEFIELD   0x9000
#define VTAB_REMOTE_MINE_SET    0xA000

/* Masks of typical combinations */
#define VTAB_ALL                0xFFFF
#define VTAB_LOCAL              0x7FFF
#define VTAB_REMOTE             0xBFFF
#define VTAB_VEHICLE            0xC001
#define VTAB_MISSILE            0xC002
#define VTAB_STRUCTURE          0xC004
#define VTAB_DESIGNATOR         0xC008
#define VTAB_STEALTH            0xC010
#define VTAB_ENV                0xC020
#define VTAB_AEFLARE            0xC040 /* temporarily for AE */
#define VTAB_MINE_MARKER        0xC080
#define VTAB_BREACH_LANE        0xC100
#define VTAB_AGGREGATE          0xC200
#define VTAB_PROXY              0xC400
#define VTAB_RADIO              0xC800
#define VTAB_MINEFIELD          0xD000
#define VTAB_MINE_SET           0xE000                                

/* Since routines outside of libvtab do VTab mask operations,
 * VAB_TYPES_MATCH has to be made public.  Likewise, so do VTAB_WHERE
 * and VTAB_WHAT, though they should not be used directly
 */
#define VTAB_WHERE 0xC000              /* do not use directly */
#define VTAB_WHAT  0x3FFF              /* do not use directly */
#define VTAB_TYPES_MATCH(value, mask) \
  ((((value) & (mask) & VTAB_WHERE) != 0) && \
   (((value) & (mask) & VTAB_WHAT) != 0))

/* Types of data structures available for making vehicle lists.
 */

/* Doubly linked lists are nice for relatively short lists which
 * will have lots of additions/deletions and not much search.
 */
#define VTAB_STRUCTURE_DOUBLY_LINKED      1

/* Packed binary trees are good when there is a lot of search, and
 * the members change fairly frequently.
 */
#define VTAB_STRUCTURE_BINARY_TREE        2

/* Order arrays are good when there will be a lot of search, and
 * almost no additions or deletions.  These take about a third
 * of the space needed by binary trees, and can be searched
 * faster, but adding and deleting items takes forever.
 */
#define VTAB_STRUCTURE_ORDERED_ARRAY      3

/* The primary data type managed by libvtab...
 */
typedef struct vtab_list *VTAB_LIST;

/* *THE* vehicle table */
extern VTAB_LIST vtab_vehicle_table;

/* The type of function passed to vtab_for_vehicles_do.
 */
typedef void (*VTAB_QUERY_FUNCTION)(
    int32 id,
    VTAB_USER_DATA_TYPE user_data,
    VTAB_USER_DATA_TYPE query_user_data);

/* The type of function passed to vtab_set_removal_callback.
 */
typedef void (*VTAB_REMOVAL_FUNCTION)(
    VTAB_LIST           list,
    int32               id,
    VTAB_USER_DATA_TYPE user_data,
    ADDRESS             argument);

/* vtab_init:
 *
 * Initializes libvtab.  Call this before any other libvtab function.
 * Libvtab can support as many tables as the application wants (such as
 * to support multiple simultaneous exercises), but I imagine most will
 * just want one.
 */
extern void vtab_init(
    int32  num_tables,
    uint16 site,
    uint16 host);

/* vtab_select_table:
 *
 * Selects which vehicle table is currently active.  There is no need
 * to call this if using only one vehicle table.  Note that selecting
 * the table sets up the context in which all libvtab functions occur.
 * For example, each vehicle list which is created with vtab_create_list
 * holds vehicles which exist in only in the context of the table
 * which was selected when the list was created.
 */
extern void vtab_select_table(
    int32 which_table);

/* vtab_vehicle_number:
 *
 * Given a SIMNET vehicleID (site/host/vehicle) this function returns
 * the integer used internally to represent the vehicle.  Note that
 * this number is unique to this vehicle only if vtab_new_vehicle is
 * called before the next call to this function.  A return value
 * of 0 indicates that the vehicle_id is invalid, and the vehicle
 * should be discarded.
 */
extern int32 vtab_vehicle_number(
    VehicleID *vehicle_id);

/* vtab_new_vehicle:
 *
 * Call this function to create a new vehicle in the global vehicle
 * table.  Get the id with a call to vtab_vehicle_number.  The type
 * is one of VTAB_LOCAL_VEHICLE, VTAB_REMOTE_MISSILE, etc.
 */
extern void vtab_new_vehicle(
    int32               id,
    VehicleID          *vehicle_id,
    uint16              type,
    VTAB_USER_DATA_TYPE user_data);

/* vtab_create_simnet_id:
 *
 * Generates a new SIMNET vehicleID.  The sequence to create a new
 * local vehicle is:	vtab_create_simnet_id(&simnet_id);
 *			id = vtab_vehicle_number(&simnet_id);
 *			vtab_new_vehicle(id, &simnet_id, VTAB_LOCAL_VEHICLE);
 */
extern void vtab_create_simnet_id(
    VehicleID *vehicle_id);

/* vtab_remove_vehicle:
 *
 * Call this function to remove a vehicle from all vehicle tables
 * (including the global one).
 */
extern void vtab_remove_vehicle(
    int32 id);

/* vtab_set_vehicle:
 *
 * This sets the user_data slot for a vehicle in the global table.
 * As you can see, this is actually just a shorthand.
 */
#define vtab_set_vehicle(id, x) \
  vtab_set_user_data(vtab_vehicle_table, id, x)

/* vtab_get_vehicle:
 *
 * This gets the user_data slot for a vehicle in the global table.
 * This is equivalent to vtab_get_user_data(vtab_vehicle_table, id),
 * but is optimized for the structure of the global table.
 */
extern VTAB_USER_DATA_TYPE vtab_get_vehicle(
    int32 id);
#define vtab_get_vehicle(id) \
  (*((*((VTAB_USER_DATA_TYPE ***)vtab_vehicle_table))[id]))

/* vtab_vehicle_type:
 *
 * Returns the type of the passed vehicle (LOCAL_VEHICLE, etc.).
 * The return value 0 indicates the specified vehicle does not exist.
 */
extern uint16 vtab_vehicle_type(
    int32 id);

/* vtab_vehicle_simid:
 *
 * Returns the SIMNET id of the passed vehicle (LOCAL_VEHICLE, etc.).
 */
extern void vtab_vehicle_simid(
    int32 id,
    VehicleID *simnet_id);

/* vtab_create_list:
 *
 * Creates a list of vehicles.  The data structure used for the list
 * is independent of functionality (you can change the data_structure
 * argument withouth changing anything else).  The only data sizes
 * respected are 0, 8, 16, and 32 bits; other values are rounded up.
 */
extern VTAB_LIST vtab_create_list(
    int32 data_structure,
    int32 bits);

/* vtab_clear_list:
 *
 * Empties a list of vehicles.  It is not necessary to call this when
 * the list is first created.
 */
extern void vtab_clear_list(
    VTAB_LIST list);

/* vtab_free_list:
 *
 * Frees the memory used by a list of vehicles.  Do not refer to the
 * list ever again after calling this function.
 */
extern void vtab_free_list(
    VTAB_LIST list);

/* vtab_add_to_list:
 *
 * Adds a vehicle to the passed list, and sets its associated user
 * data.  Note that although an int32 of user data is passed, only the
 * low order 0, 8, 16, or 32 bits will be stored, depending upon
 * the definition of the table.
 */
extern void vtab_add_to_list(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);

/* vtab_remove_from_list:
 *
 * Removes the vehicle from the list (if the vehicle is not there,
 * nothing will happen).
 */
extern void vtab_remove_from_list(
    VTAB_LIST list,
    int32 id);

/* vtab_set_removal_callback:
 *
 * Requests that the passed function be triggered when vehicles are
 * removed from this list.  This is the only way to detect removals
 * resulting from vehicles being deleted from the global table.
 */
extern void vtab_set_removal_callback(
    VTAB_LIST list,
    VTAB_REMOVAL_FUNCTION callback,
    ADDRESS argument);

/* vtab_set_user_data:
 *
 * Sets the user data slot of a vehicle stored in the passed vehicle
 * list.  Note that although an int32 argument is passed, only the low
 * order 0, 8, 16, or 32 bits will be stored, depending upon the
 * definition of the table.
 */
extern void vtab_set_user_data(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);

/* vtab_get_user_data:
 *
 * Gets the user data slot of a vehicle stored in the passed vehicle
 * list.  Note that although an int32 is returned, only the low order
 * 0, 8, 16, or 32 bits will be used, depending upon the definition
 * of the table.  The value 0 will be returned if the vehicle does
 * not appear in the list.
 */
extern VTAB_USER_DATA_TYPE vtab_get_user_data(
    VTAB_LIST list,
    int32 id);

/* vtab_in_list:
 *
 * Returns 0 if the passed id does not appear in the list, and 1
 * if the id does appear.  This is especially useful for tables
 * with user data size 0.
 */
extern int32 vtab_in_list(
    VTAB_LIST list,
    int32 id);

/* vtab_count_vehicles:
 *
 * Counts the vehicles in the list whos type matches the passed mask.
 */
extern int32 vtab_count_vehicles(
    VTAB_LIST list,
    uint16 type_mask);

/* These next two functions allow the construction of a loop:

       for (id=vtab_get_first(list, VTAB_ALL, &user_data); id;
	    id=vtab_get_next(list, VTAB_ALL, &user_data))
*/

/* vtab_get_first:
 *
 * Returns the first id and user data in the passed list which matches
 * the type mask.  Remembers what was returned for successive calls to
 * vtab_get_next.  A return value of 0 means the list is empty.
 */
extern int32 vtab_get_first(
    VTAB_LIST list,
    uint16 type_mask,
    VTAB_USER_DATA_TYPE *user_data);

/* vtab_get_next:
 *
 * Returns the next id and user data after the ones just returned by
 * either vtab_get_first or vtab_get_next which matches the type
 * mask.  A return value of 0 means no id is available.
 */
extern int32 vtab_get_next(
    VTAB_LIST list,
    uint16 type_mask,
    VTAB_USER_DATA_TYPE *user_data);

/* vtab_print_list:
 *
 * Prints a vehicle list.  Handy for debugging.
 */
extern void vtab_print_list(
    VTAB_LIST list);

/* Set-like operations.
 * This is sort of RISCy, in that if you want to do (z = x op y),
 * you have to do: (z), (z = z U x), (x = x op y).
 * Note that a unary compliment operator would lead to rediculously
 * long lists,so none is provided.
 */

/* vtab_list_union: a = a U b
 */
extern void vtab_list_union(
    VTAB_LIST a, 
    VTAB_LIST b);

/* vtab_list_intersect: a = a /\ b
 */
extern void vtab_list_intersect(
    VTAB_LIST a,
    VTAB_LIST b);

/* vtab_list_intersect_not: a = a /\ ~b
 */
extern void vtab_list_intersect_not(
    VTAB_LIST a,
    VTAB_LIST b);

/* vtab_list_equal: a == b
 */
extern int32 vtab_lists_equal(
    VTAB_LIST a,
    VTAB_LIST b,
    int32 check_user_data);

#ifdef USE_RTI

/**
 ** RTI-related functionality extensions...
 **/

#include <librti.h>

/* vtab_id_to_rti:
 *
 * Lookup the 48-bit ID associated with the passed hashed vehicle_id,
 * and derive the RTI_OBJECT_ID from it, by using the low-order 32
 * bits. Note that RTI_OBJECT_ID is passed by reference, because we
 * expect it may be expanded beyond 32 bits in a future API revision.
 * If it expands beyond 48 bits, libvtab will have to be extended to
 * more expliclty support this kind of ID.
 */
extern void vtab_id_to_rti(
    int32 vehicle_id,
    RTI_OBJECT_ID *rti_id);

/* vtab_id_from_rti:
 *
 * Translate the passed RTI_OBJECT_ID into a corresponding 48-bit ID
 * by placing a 1 in the high-order-short, and the RTI_OBJECT_ID split
 * into the two lower-order shorts.  Then translate this to an
 * internal hashed ID using the normal methods.  Note that
 * RTI_OBJECT_ID is passed by reference, because we expect it may be
 * expanded beyond 32 bits in a future API revision.  If it expands
 * beyond 48 bits, libvtab will have to be extended to more expliclty
 * support this kind of ID.
 */
extern int32 vtab_id_from_rti(
    RTI_OBJECT_ID *rti_id);

/* vtab_set_rti_class:
 *
 * Registers a mapping between the passed vtab_type and a
 * corresponding RTI class identifier.  Libvtab will use this
 * information in object management functions.  All type registrations
 * should be performed very early in the initialization sequence, to
 * ensure these mappings are known when needed.
 */
extern void vtab_set_rti_class(
    uint32 vtab_type,
    RTI_OBJECT_CLASS_IMAGE the_class);

/* vtab_get_rti_class:
 *
 * Returns the RTI class ID of the passed vehicle, based on its
 * VTAB_TYPE.
 */
extern RTI_OBJECT_CLASS vtab_get_rti_class(
    int32 vehicle_id);

#endif

#endif /*_LIBVTAB_H_INCLUDED*/
