#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Massive test program for libvtab                          *
*       Created: Fri Apr 10 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libvtab.h"
#include <basic.h>

#define NVEH 1000
#define NITS 1000

main()
{
    VTAB_LIST list[12];
    int32 i, j;
    int32 mycount, cksum;
    int32 ids[NVEH];
    VTAB_USER_DATA_TYPE ign;
    VehicleID simnet_id;
    int32 id;

    /* Initialize libvtab */
    vtab_init(1, 1, 1);

    /* Make twelve lists */
    list[ 0] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED, 0);
    list[ 1] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,   0);
    list[ 2] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY, 0);

    list[ 3] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED, 8);
    list[ 4] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,   8);
    list[ 5] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY, 8);

    list[ 6] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED,16);
    list[ 7] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,  16);
    list[ 8] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY,16);

    list[ 9] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED,32);
    list[10] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,  32);
    list[11] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY,32);

    /* Make a bunch of local vehicles */
    for (i=0;i<NVEH;i++)
    {
	vtab_create_simnet_id(&simnet_id);
	if (ids[i] = vtab_vehicle_number(&simnet_id))
	{
	    vtab_new_vehicle(ids[i], &simnet_id, VTAB_LOCAL_VEHICLE);
	    vtab_set_vehicle(ids[i], i);
	    if (vtab_get_vehicle(ids[i]) != i)
	      abort();
	}
    }

    /* Count them */
    if (vtab_count_vehicles(vtab_vehicle_table, VTAB_ALL) != NVEH)
      abort();

    /* Make a bunch of remotes */
    for (i=2;i<10;i++)
      for (j=1;j<256;j++)
      {
 	  simnet_id.simulator.site = i;
	  simnet_id.simulator.host = j;
	  simnet_id.vehicle = 1;
	  if (id = vtab_vehicle_number(&simnet_id))
	    vtab_new_vehicle(id, &simnet_id, VTAB_REMOTE_VEHICLE);
	  else
	    break;
      }

    /* Remove a remote, then make sure the hashing still finds other
     * vehicles with the same key.
     */
    vtab_remove_vehicle(201);
    simnet_id.simulator.site = 3;
    simnet_id.simulator.host = 200;
    simnet_id.vehicle = 1;
    if (vtab_vehicle_number(&simnet_id) == 201)
      abort();

    /* Now make a new vehicle and see if it reuses the available id */
    simnet_id.simulator.site = 20;
    simnet_id.simulator.host = 200;
    simnet_id.vehicle = 1;
    if (vtab_vehicle_number(&simnet_id) != 201)
      abort();

    /* Randomly add and delete vehicles from the lists */
    for (i=0;i<NITS;i++)
    {
	if (random()&0x1)
	{
	    int32 l = random()%12;
	    int32 id = ids[random()%NVEH];
	    int32 ud;

	    if (l < 3) ud = 1;
	    else if (l < 6) ud = random()%0x100;
	    else if (l < 9) ud = random()%0x10000;
	    else ud = random();

	    vtab_add_to_list(list[l], id, ud);
	    if (!vtab_in_list(list[l], id))
	      abort();
	    if (vtab_get_user_data(list[l], id) != ud)
	      abort();
	    vtab_set_user_data(list[l], id, 1);
	    if (vtab_get_user_data(list[l], id) != 1)
	      abort();
	}
	else
	{
	    int32 l = random()%12;
	    int32 id = ids[random()%NVEH];

	    vtab_remove_from_list(list[l], id);
	    if (vtab_in_list(list[l], id))
	      abort();
	    if (vtab_get_user_data(list[l], id))
	      abort();
	}
    }

    /* Delete all the vehicles */
    for (i=0;i<NVEH;i++)
    {
	vtab_remove_vehicle(ids[i]);
	if (vtab_get_vehicle(ids[i]))
	  abort();
    }

    /* Make sure they are gone from all the lists */
    for (i=0;i<12;i++)
      if (vtab_count_vehicles(list[i], VTAB_LOCAL_VEHICLE))
	abort();

    /* Delete the lists */
    for (i=0;i<12;i++)
      vtab_free_list(list[i]);

    /* Make twelve new lists */
    list[ 0] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED, 0);
    list[ 1] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,   0);
    list[ 2] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY, 0);

    list[ 3] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED, 8);
    list[ 4] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,   8);
    list[ 5] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY, 8);

    list[ 6] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED,16);
    list[ 7] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,  16);
    list[ 8] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY,16);

    list[ 9] = vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED,32);
    list[10] = vtab_create_list(VTAB_STRUCTURE_BINARY_TREE,  32);
    list[11] = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY,32);

    /* Make a bunch of local vehicles */
    for (i=0;i<NVEH;i++)
    {
	VehicleID simnet_id;
	int32 id;

	vtab_create_simnet_id(&simnet_id);
	ids[i] = vtab_vehicle_number(&simnet_id);
	vtab_new_vehicle(ids[i], &simnet_id, VTAB_LOCAL_VEHICLE);
	vtab_set_vehicle(ids[i], i);
	if (vtab_get_vehicle(ids[i]) != i)
	  abort();
    }

    /* Do the same thing to every list, and make sure they remain the
     * same.
     */
    cksum = 0;
    for (i=0;i<NITS;i++)
    {
	int32 id = ids[random()%NVEH];

	if (random()&0x1)
	{
	    if (!vtab_in_list(list[0], id))
	      cksum += id;
	    for (j=0;j<12;j++)
	      vtab_add_to_list(list[j], id, 0);
	}
	else
	{
	    if (vtab_in_list(list[0], id))
	      cksum -= id;
	    for (j=0;j<12;j++)
	      vtab_remove_from_list(list[j], id);
	}
    }

    /* Randomly delete vehicles from each list to ensure deletion
     * during traversal works correctly.
     */
    for (j=0;j<12;j++)
      id = vtab_get_first(list[j], VTAB_ALL, &ign);
    while (id)
    {
	if (random() & 0x1)
	{
	    for (j=0;j<12;j++)
	      vtab_remove_from_list(list[j], id);
	    cksum -= id;
	}

	for (j=0;j<12;j++)
	  id = vtab_get_next(list[j], VTAB_ALL, &ign);
    }

    mycount = 0;
    for (i=0;i<NVEH;i++)
    {
	int32 in = vtab_in_list(list[0], ids[i]);
	if (in)
	  mycount++;
	for (j=1;j<12;j++)
	  if (vtab_in_list(list[j], ids[i]) != in)
	    abort();
    }    

    /* Check if we agree about how many should be in each */
    for (j=0;j<12;j++)
      if (vtab_count_vehicles(list[j], VTAB_LOCAL_VEHICLE) != mycount)
	abort();

    /* Check the checksums */
    for (j=0;j<12;j++)
    {
	int32 dblchek=0;

	for (i=vtab_get_first(list[j], VTAB_ALL, &ign); i;
	     i=vtab_get_next(list[j], VTAB_ALL, &ign))
	  dblchek += i;

	if (dblchek != cksum)
	  abort();
    }

    /* Make sure each list has the same entries in the same order. */
    i = vtab_get_first(list[0], VTAB_ALL, &ign);
    for (j=1;j<12;j++)
      if (i != vtab_get_first(list[j], VTAB_ALL, &ign))
	abort();
    while (i)
    {
	i = vtab_get_next(list[0], VTAB_ALL, &ign);
	for (j=1;j<12;j++)
	  if (i != vtab_get_next(list[j], VTAB_ALL, &ign))
	    abort();
    }

    /* Test the set operations */
    for (i=VTAB_STRUCTURE_DOUBLY_LINKED;i<=VTAB_STRUCTURE_BINARY_TREE;i++)
    {
	VTAB_LIST t;

	t = vtab_create_list(i, 32);
	vtab_list_union(t, list[11]);
	if (!vtab_lists_equal(t, list[11], 1))
	  abort();
	vtab_list_intersect(t, list[11]);
	if (!vtab_lists_equal(t, list[11], 1))
	  abort();
	vtab_list_intersect_not(t, list[11]);
	if (vtab_count_vehicles(t, VTAB_ALL))
	  abort();
	vtab_free_list(t);
    }

    printf("Yea Verily, everthing works.\n");
}

