#ifndef lint
static char rcsid [] = "$RCSfile: vt_dbll.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_dbll.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Doubly Linked List structure                              *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libvtab_local.h"

#define GNFLAG ((VTAB_DOUBLY_LINKED_ENTRY *)-1)

static void dl_init(
    VTAB_DOUBLY_LINKED *dl);
static void dl_free(
    VTAB_LIST list);
static void dl_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static void dl_remove(
    VTAB_LIST list,
    int32 id);
static void dl_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static VTAB_USER_DATA_TYPE dl_get(
    VTAB_LIST list,
    int32 id);
static int32 dl_in(
    VTAB_LIST list,
    int32 id);
static int32 dl_count(
    VTAB_LIST list,
    uint16 mask);
static int32 dl_get_first(
    VTAB_LIST list,
    uint16 mask, 
    VTAB_USER_DATA_TYPE *user_data);
static int32 dl_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);

void vtab_init_doubly_linked(
    VTAB_LIST list,
    int32 bits)
{
    list->structure = VTAB_STRUCTURE_DOUBLY_LINKED;
    list->ud_size = bits;
    list->clear_fcn = dl_free;
    list->free_fcn = dl_free;
    list->add_fcn = dl_add;
    list->remove_fcn = dl_remove;
    list->set_fcn = dl_set;
    list->get_fcn = dl_get;
    list->in_fcn = dl_in;
    list->count_fcn = dl_count;
    list->getf_fcn = dl_get_first;
    list->getn_fcn = dl_get_next;

    dl_init(&list->data.dl);

    return;
}

static void dl_init(
    VTAB_DOUBLY_LINKED *dl)
{
    dl->head = NULL;
    dl->last_got = GNFLAG;
}

static void dl_free(
    VTAB_LIST list)
{
    VTAB_DOUBLY_LINKED_ENTRY *e, *n;

    for(e=list->data.dl.head;e;e=n)
    {
	n = e->next;
	STDDEALLOC(e);
    }
    list->data.dl.head = NULL;
}

static void dl_add(
    VTAB_LIST list, 
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    int32 cmp;
    VTAB_DOUBLY_LINKED_ENTRY *i, *next, *prev, *new;

    /* Figure out where to put the new id */
    prev = NULL;
    for (i=dl->head;i;prev=i,i=i->next)
    {
	if (!(cmp = vtab_packed_cmp(&i->data, id, list->ud_size)))
	{
	    vtab_packed_add(&i->data, id, user_data, list->ud_size);
	    return;
	}
	if (cmp > 0)
	  break;
    }
    next = i;

    /* Make a new node */
    new =
      (VTAB_DOUBLY_LINKED_ENTRY *)STDALLOC(sizeof(VTAB_DOUBLY_LINKED_ENTRY));
    bzero(new, sizeof(VTAB_DOUBLY_LINKED_ENTRY));

    /* Right now, 'next' is the one after the new one and 'prev' is the
     * one before it.
     */
    new->next = next;
    if (new->next)
      new->next->prev = new;
    new->prev = prev;
    if (new->prev)
      new->prev->next = new;
    else
      dl->head = new;

    vtab_packed_add(&new->data, id, user_data, list->ud_size);
}

static VTAB_DOUBLY_LINKED_ENTRY *dl_find(
    VTAB_DOUBLY_LINKED_ENTRY *head,
    int32 id,
    int32 ud_size)
{
    int32 cmp;

    for(;head;head=head->next)
    {
	if (!(cmp = vtab_packed_cmp(&head->data, id, ud_size)))
	  return head;
	if (cmp > 0)
	  break;
    }
    return NULL;
}

static void dl_remove(
    VTAB_LIST list,
    int32 id)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    VTAB_DOUBLY_LINKED_ENTRY *found;

    if (found = dl_find(dl->head, id, list->ud_size))
    {
	if (list->removal_callback)
	  list->removal_callback(list, id,
				 vtab_packed_get(&found->data, id,
						 list->ud_size),
				 list->removal_callback_arg);

	vtab_packed_remove(&found->data, id, list->ud_size);

	/* Collect this node if empty */
	if (!found->data.used)
	{
	    /* Protect ourselves from deletion mid-traversal */
	    if (found == dl->last_got)
	    {
		dl->last_got = dl->last_got->next;
		dl->last_which = -1;
	    }

	    if (found->next)
	      found->next->prev = found->prev;
	    if (found->prev)
	      found->prev->next = found->next;
	    else
	      dl->head = found->next;
	    STDDEALLOC(found);
	}
    }
}

static void dl_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    VTAB_DOUBLY_LINKED_ENTRY *found;

    if (found = dl_find(dl->head, id, list->ud_size))
      vtab_packed_set(&found->data, id, user_data, list->ud_size);
}
    
static VTAB_USER_DATA_TYPE dl_get(
    VTAB_LIST list,
    int32 id)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    VTAB_DOUBLY_LINKED_ENTRY *found;

    if (found = dl_find(dl->head, id, list->ud_size))
      return vtab_packed_get(&found->data, id, list->ud_size);
    else
      return 0;
}
    
static int32 dl_in(
    VTAB_LIST list,
    int32 id)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    VTAB_DOUBLY_LINKED_ENTRY *found;

    if (found = dl_find(dl->head, id, list->ud_size))
      return vtab_packed_in(&found->data, id, list->ud_size);
    else
      return 0;
}
    
static int32 dl_count(
    VTAB_LIST list,
    uint16 mask)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    VTAB_DOUBLY_LINKED_ENTRY *i;
    int32 count;

    for (i=dl->head,count=0;i;i=i->next)
      count += vtab_packed_count(&i->data, mask, list->ud_size);

    return count;
}

static int32 dl_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;

    dl->last_which = -1;
    dl->last_got = dl->head;
    return dl_get_next(list, mask, user_data);
}

static int32 dl_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_DOUBLY_LINKED *dl = &list->data.dl;
    VTAB_DOUBLY_LINKED_ENTRY *i;
    int32 which, id;

    if (dl->last_got == GNFLAG)
    {
	fprintf(stderr, "LIBVTAB WARNING: Called get_next without first calling get_first!\n");
	return 0;
    }

    /* This can happen if the last vehicle in the list is deleted
     * just after being traversed.
     */
    if (!dl->last_got)
    {
	dl->last_got = GNFLAG;
	return 0;
    }

    which = dl->last_which;
    for (i=dl->last_got;i;i=i->next)
    {
	while (id = vtab_packed_get_next(&i->data, user_data, &which,
					 list->ud_size))
	{
	    if (VTAB_MATCH(id, mask))
	    {
		dl->last_which = which;
		dl->last_got = i;
		return id;
	    }
	}
	which = -1;
    }
    dl->last_got = GNFLAG;
    return 0;
}
