#ifndef lint
static char rcsid [] = "$RCSfile: vt_rti.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_rti.c                                                          *
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: RTI-related functionality                                 *
*       Created: Mon Apr 15 1996                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_RTI

#include "libvtab_local.h"

/* Lookup the 48-bit ID associated with the passed hashed vehicle_id,
 * and derive the RTI_OBJECT_ID from it.
 */
void vtab_id_to_rti(
    int32 vehicle_id,
    RTI_OBJECT_ID *rti_id)
{
    VehicleID vid;

    vtab_vehicle_simid(vehicle_id, &vid);
    *rti_id = (vid.simulator.host << 16) | vid.vehicle;
}

/* Translate the passed RTI_OBJECT_ID into a corresponding 48-bit ID.
 */
int32 vtab_id_from_rti(
    RTI_OBJECT_ID *rti_id)
{
    VehicleID vid;

    vid.simulator.site = 1;
    vid.simulator.host = ((uint32)*rti_id) >> 16;
    vid.vehicle = (uint32)*rti_id & 0xffff;
    return vtab_vehicle_number(&vid);
}

/* Table of RTI classes (initialzed through vtab_set_rti_class calls)
 */
#define ILLEGAL_INDEX 15
static RTI_OBJECT_CLASS table[16];

/* Find the index corresponding to a given vtab mask value */
static uint32 type_index(
    uint32 mask)
{
    int32 i;
    mask &= VTAB_WHAT;
    for (i=0;i<ILLEGAL_INDEX;i++,mask=(mask>>1))
      if (mask & 0x1)
	break;
    return i;
}

/* Registers a mapping between the passed vtab_type and a
 * corresponding RTI class identifier.
 */
void vtab_set_rti_class(
    uint32 vtab_type,
    RTI_OBJECT_CLASS_IMAGE the_class)
{
    int32 idx = type_index(vtab_type);
    RTI_OBJECT_CLASS lookup;
    if (!rti_in_use)
      return;
    if (idx == ILLEGAL_INDEX)
      return;
    if (!rti_get_object_class(the_class, &lookup))
      table[idx] = lookup;
}

/* Returns the RTI class ID of the passed vehicle, based on its
 * VTAB_TYPE.
 */
RTI_OBJECT_CLASS vtab_get_rti_class(
    int32 vehicle_id)
{
    int32 idx = type_index(vtab_vehicle_type(vehicle_id));
    if (!rti_in_use)
      return;
    if (idx == ILLEGAL_INDEX)
      return;
    return table[idx];
}

#endif /*USE_RTI*/
