/*
 * $RCSfile: libwld_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libwld_local.h
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC/SAIC under U. S. Government     *
*       contracts, and may be reproduced by or for the U. S. Government     *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013                                               *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libworld                           *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libworld.h"
#include <libctdb.h>
#include <basic.h>

#ifndef _LIBWLD_LOCAL_H_INCLUDED
#define _LIBWLD_LOCAL_H_INCLUDED

#define WLD_MAX_SEARCH_PATH_LEN 132
#define WLD_NOWHERE "Nowhere"
#define WLD_LAT 0
#define WLD_LNG 1

/* command line options structure */
extern struct wld_options_struct
{
    CMD_STRING_OPTION  terrain;
    CMD_STRING_OPTION  tdbpath;
    CMD_STRING_OPTION  gcscell;
    CMD_INTEGER_OPTION tdbmem;
} wld_options;


typedef struct wld_grids
{
    int32             valid;
    int32             n_thin;
    int32             s_thin;
    float32          *thin;
    int32             n_thick;
    int32             s_thick;    
    float32          *thick;
    int32             n_label;
    int32             s_label;
    COORD_GRID_LABEL *label;
} WLD_GRIDS;

typedef struct wld_playbox
{
    CTDB_TDB_TYPE kind;

    struct wld_gcs_extent
    {
	int32   valid;
	float64 cntr[XYZC];
	float64 sw_crnr[XYZC];
	float64 ne_crnr[XYZC];
    } gcs;

    struct wld_ll_extent
    {
	int32   valid;
	float64 cntr[3]   ;/* (WLD_LAT,WLD_LNG,Z) */
	float64 sw_crnr[3];/* (WLD_LAT,WLD_LNG,Z) */
	float64 ne_crnr[3];/* (WLD_LAT,WLD_LNG,Z) */
    } ll;
   
    TerrainDatabaseID pbname;
    int32             n_cells;
    int32            *cells;
    WLD_GRIDS         utm_grid;
    WLD_GRIDS         gcs_grid;
} WLD_PLAYBOX;


/* File    : wld_set.c
 * Routine : wld_set_extent
 *
 * This routine will set the extents of the playbox in the WLD_PLAYBOX 
 * structure.
 */
extern void
wld_set_extent_and_center (
			      WLD_PLAYBOX *pb,   
			      float64 mn_lat,
			      float64 mn_lng,
			      float64 mx_lat,
			      float64 mx_lng
			      );

/* File    : wld_playbx.c
 * Routine : wld_generate_utm_grids
 *
 * This routine will generate the utm grid lines for the complete playbox. 
 * The grid points are in the playbox center frame of reference and are 
 * updated when the playbox changes in size.
 */
extern void
wld_generate_utm_grids ( WLD_PLAYBOX *pb );

/* File    : wld_lookup.c
 * Routine : wld_get_playbox_extent_ll
 *
 * This routine will the playbox extent's southwest corner in south_lat and 
 * west_lng and the northeast corner in north_lat, east_lng.
 */
extern void
wld_get_playbox_extent_ll (
			   WLD_PLAYBOX *pb,
			   float64 *south_lat,
			   float64 *west_lng,
			   float64 *north_lat,
			   float64 *east_lng
			   );

/* File    : wld_init.c
 * Routine : wld_terrain_cmdline_help
 *
 * This is the help routine for the terrain command line option
 */
extern int32
wld_terrain_cmdline_help ( char *dbname );

/* Private-to-this-library globals */
extern uint32       wld_gcs_aware;
extern WLD_PLAYBOX  wld_pb;
extern char         wld_db_dir[WLD_MAX_SEARCH_PATH_LEN];



#endif /*_LIBWLD_LOCAL_H_INCLUDED*/

