/*
 * $RCSfile: libworld.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libworld.h
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC/SAIC under U. S. Government     *
*       contracts, and may be reproduced by or for the U. S. Government     *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013                                               *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libworld                           *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBWORLD_H_INCLUDED
#define _LIBWORLD_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdext.h>
#include <stdtypes.h> /*common/include/global*/
#include <libcoordinates.h>
#include <libcmdline.h>
#include <libctdb.h>


/* File    : wld_init.c
 * Routine : wld_init
 *
 * This routine will initialize library global data
 */
extern void
wld_init (void);

/* File    : wld_init.c
 * Routine : wld_get_tdbpath
 *
 * Returns a pointer to the terrain database path specified on the command line
 */
extern char *
wld_get_tdbpath (void);


/* File    : wld_init.c
 * Routine : wld_get_terrain
 *
 * Returns a pointer to the terrain database specified on the command line
 */
extern char *
wld_get_terrain (void);

/* File    : wld_init.c
 * Routine : wld_lookup_playbox_cells
 *
 * This routine will return the list of cells that are active for the
 * simulation playbox.  The item n_cell must be set on input to the 
 * size of the allocated array cell.  If the number of cells in the 
 * playbox exceeds the list size the function will return 0 (FALSE) and 
 * n_cells will be set to the number of cells in the playbox.  This number
 * should be used to reallocate a cell list and the function will need to be
 * called to populate the list.
 */
extern int32
wld_lookup_playbox_cells (
			  int32 *n_cell,
			  int32 cell[]
			  );

/* File    : wld_lookup.c
 * Routine : wld_get_playbox_center
 *
 * This routine will return the center point of the active simulation 
 * playbox in GCS coordinates.
 */
extern void
wld_get_playbox_center ( float64 cntr[XYC] );

/* File    : wld_lookup.c
 * Routine : wld_get_playbox_extent
 *
 * This routine will return the min and max points of the active simulation 
 * playbox in GCS coordinates.  Where the min and max are defined as the 
 * southwest(lower left) and northeast(upper right) corners, respectively,
 * of the bounding box that contains the playbox.
 */
extern void
wld_get_playbox_extent (
			   int32   destination_cell,
			   float64 sw_corner[XYZ],
			   float64 ne_corner[XYZ]
			   );
/* File    : wld_playbx.c
 * Routine : wld_expand_playbox
 * 
 * This routine will check to see if cell is already in the playbox.
 * If it isn't, it will add cell to the playbox, and allocate
 * all necessary gcs structures for it.
 */
extern void
wld_expand_playbox ( int32 cell );

/* File    : wld_playbx.c
 * Routine : wld_get_playbox_utm_grids
 * 
 * This routine will get the pregenerated utm grids for the active playbox.
 * The number of thin, thick and labels must be set to the dimension of the
 * incoming arrays, respectively, and are then set to the number of segments
 * or labels that are stored in the passed in arrays.  If the size of an 
 * allocated array is to small then the function will return false and the 
 * output n_thin, n_thick, n_labels will contain the need array dimensions.
 */ 
extern int32
wld_get_playbox_utm_grids (
			      int32            *n_thin,/* number of segments 
							*allocated for */
			      float32          thin[],/* segment contains 
						       * XY pair */
			      int32           *n_thick,/* number of segments 
							* allocated for */
			      float32          thick[],/* segment contains 
							* XY pair */
			      int32           *n_labels,    
			      COORD_GRID_LABEL labels[]
			      );

/* File    : wld_playbx.c
 * Routine : wld_point_in_playbox
 *
 * This routine will return TRUE if the GCS point is contained within the 
 * active playbox.
 */
extern int32
wld_point_in_playbox (
			 int32   cell,
			 float64 x,
			 float64 y,
			 float64 z
			 );
/* File    : wld_playbx.c
 * Routine : wld_get_playbox_id
 *
 * This routine will return the playbox name and version, this originally 
 * comes from the initial playbox loading (ctdb headers).
 */
extern void
wld_get_playbox_id (
		       char   name[],
		       int32 *version
		       );

/* File    : wld_init.c
 * Routine : wld_terrain_aggregate_options
 *
 * This routine will parse the terrain options from the command line
 */
extern void 
wld_tdb_aggregate_options(
			  CMD_OPTION **options,
			  uint32      *sizeof_options,
			  CMD_TYPES ***map
			  );

/* File    : wld_playbx.c
 * Routine : wld_load_playbox
 *
 * This routine will load a terrain playbox by initializing the proper 
 * required libraries and creating a playbox.  This routine should be 
 * used in place of the load_terrain routine that was called from various 
 * test routines.
 */
extern void 
wld_load_playbox(
		 int32          sim,
		 int32          gui,
		 int32          argc, 
		 char          *argv[],
		 char          *data, 
		 char          *playbox_name, 
		 int32         *playbox_version,
		 COORD_TCC_PTR *tccp,            /* SIMNET only */
		 CTDB          **ctdb            /* SIMNET only */
		 );
#ifdef __cplusplus
}
#endif

#endif /*_LIBWORLD_H_INCLUDED*/
