#ifndef lint
static char rcsid [] = "$RCSfile: wld_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: wld_init.c                                                        *
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC/SAIC under U. S. Government     *
*       contracts, and may be reproduced by or for the U. S. Government     *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013                                               *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libworld                               *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libwld_local.h"
#include <stdstring.h>
#include <stdalloc.h>

WLD_PLAYBOX  wld_pb;
char         wld_db_dir[WLD_MAX_SEARCH_PATH_LEN];
static int32 library_init=FALSE;


/* Options for libcmdline */
#define DEFAULT_TERRAIN "knox-0311"
#define DEFAULT_TDBPATH "../../terrain"
#define DEFAULT_GCSCELL "none"

struct wld_options_struct wld_options = {
    {
        "Terrain Database" , "Terrain name or geographic extent, use -terrain help for more details",
        NULL ,
        CMD_STRING ,    "terrain" ,    "nowhere" , DEFAULT_TERRAIN ,
    } ,
    {
        "TDB Directory" , "Specifies directory where terrain database files are stored" ,
        NULL ,
        CMD_STRING ,    "tdbpath" , NULL , DEFAULT_TDBPATH ,
    } ,
    {
        "GCS Cell" , "Specifies which type of GCS cell database that is being saught (values: none, single, multi)",
        NULL ,
        CMD_STRING ,    "gcscell" , NULL , DEFAULT_GCSCELL
    } ,
    {
        "TDB Memory" , "Specifies the amount of memory (in Mb) reserved for terrain database services (0 defines unlimited memory).  Recommend a minimum of 5Mb when simulating and 20Mb when not simulating.",
        NULL ,
        CMD_INTEGER ,   "tdbmem" , "tdbmem_all" , 5 ,
    } ,
};

int32
wld_terrain_cmdline_help(char *dbname)
{
    char simnet[256],single[256],multi[256];
    
    if ( strcmp(dbname,"help") != 0 )
      return 0;
    
    sprintf(single,"<dbname>.%s%d%s",CTDB_SINGLE_CELL_ID,CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
    sprintf(multi,"<dbname>.%s/<dbname>.%s%d%s",CTDB_MULTI_CELL_DIR,CTDB_MULTI_CELL_ID,
	    CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
    sprintf(simnet,"<dbname>.%s%d%s",CTDB_SIMNET_ID,CTDB_FILE_FORMAT,CTDB_ENDIAN_ID);
    printf("Since historically the terrain command line option hasn't specified\n");   
    printf("the database extension, it is required that the software\n"); 
    printf("determine the type of database to load.  The table below specifies\n"); 
    printf("the loading procedure for the legal terrain configurations.  It is\n"); 
    printf("important to note that the \"%d\" in the filename extension denotes\n",
	   CTDB_FILE_FORMAT); 
    printf("format %d and will always reflect the ctdb file format identifier.\n\n",
	   CTDB_FILE_FORMAT);  
    printf("--------------------------------------------------------------------\n");
    printf("|                |                |                                 \n");
    printf("|  -gcscell      |  -terrain      | Resultant Processing \n");
    printf("|  (dflt=%s)   |(dflt=%s)|\n",DEFAULT_GCSCELL,DEFAULT_TERRAIN);
    printf("====================================================================\n");
    printf("|     multi      | lat/lng extent | Determine which cells are      \n");
    printf("|                | (NOTE 1)       | needed to fill the extent.  Search\n");
    printf("|                |                | for single cell tdbs (%s) to\n",
	   single);
    printf("|                |                | fill in the extent with terrain.\n");
    printf("|                |                |                                  \n");
    printf("|                |                | Failure Mode:                  \n");
    printf("|                |                | Change the Terrain Command Line\n");
    printf("|                |                | to the default terrain database\n");
    printf("|                |                | (i.e. %s) and follow \n",DEFAULT_TERRAIN);
    printf("|                |                | none/%s processing\n",DEFAULT_TERRAIN);
    printf("--------------------------------------------------------------------\n");
    printf("|     single     | lat/lng extent | Not available                  \n");
    printf("|                | (NOTE 1)       |                                  \n");
    printf("|                |                | Failure Mode:                  \n");
    printf("|                |                | Search for default SIMNET database\n");
    printf("|                |                | (i.e. %s).  If not found\n",DEFAULT_TERRAIN);
    printf("|                |                | then initialize as none/nowhere. \n");
    printf("--------------------------------------------------------------------\n");
    printf("|     none       | lat/lng extent | Not available                  \n");
    printf("|                |                |                                  \n");
    printf("|                |                | Failure Mode:                  \n");
    printf("|                |                | Search for default SIMNET database\n");
    printf("|                |                | (i.e. %s).  If not found \n",DEFAULT_TERRAIN);
    printf("|                |                | then initialize as none/nowhere. 	\n");
    printf("--------------------------------------------------------------------\n");
    printf("|     none       | database name  | Search for %s in the\n",
	   simnet);
    printf("|                | <dbname>       | terrain directory.  Load database.\n");
    printf("|                |                |                                  \n");
    printf("|                |                | Failure Mode:                  \n");
    printf("|                |                | If database not found then initial-\n");
    printf("|                |                | ize as none/nowhere.    \n");
    printf("|                |                | \n");
    printf("--------------------------------------------------------------------\n");
    printf("|     single     | database name  | Search for %s in the\n",
	   single);
    printf("|                | <dbname>       | terrain directory.  Load database.\n");
    printf("|                |                |                                  \n");
    printf("|                |                | Failure Mode:                  \n");
    printf("|                |                | If database not found then process\n");
    printf("|                |                | as none/<dbname> processing\n");
    printf("--------------------------------------------------------------------\n");
    printf("|     multi      | database name  | Search for <dbname>.dir in the \n");
    printf("|                | <dbname>       | terrain directory.  If found,    \n");
    printf("|                |                | load the database's meta data  \n");
    printf("|                |                | (%s) which \n",multi);
    printf("|                |                | contains extent information.    \n");
    printf("|                |                | Determine which cells are   \n");
    printf("|                |                | needed to fill the extent.  Search\n");
    printf("|                |                | for single cell tdbs (%s) to\n",
	   single);
    printf("|                |                | fill in the extent with terrain.\n");
    printf("|                |                | Load cells.\n");
    printf("|                |                |                                  \n");
    printf("|                |                | Failure Mode:                  \n");
    printf("|                |                | If the meta data is not found \n");
    printf("|                |                | transition into single/<dbname> \n");
    printf("|                |                | processing.\n");
    printf("--------------------------------------------------------------------\n");
    printf("|    multi       | <noterrain>    | Illegal Combination    \n");
    printf("|                |                |                                  \n");
    printf("|    single      |                | Failure Mode:                  \n");
    printf("|                |                | Process as none/nowhere           \n");
    printf("--------------------------------------------------------------------\n");
    printf("|    none        | <noterrain>    | Process as none/nowhere \n");
    printf("--------------------------------------------------------------------\n\n");
    printf("NOTE 1: format of the latitude/longitude extents looks like\n");
    printf("37.53E85.64Sx38.00E86.49S where the first set (delimited by the\n");
    printf("\"x\") of lat/long define the minimum/lower-left and the second set defines the \n");
    printf("maximum/upper-right.  Each Latitude and Longitude is represented in degrees and \n");
    printf("with the letter \"E\" or \"W\" separating the longitude from the latitude.\n");
    return 1;
}



/* Process our options using libcmdline */
void 
wld_tdb_aggregate_options(CMD_OPTION **options,
			  uint32 *sizeof_options,
			  CMD_TYPES ***map)
{
    cmd_aggregate(options, sizeof_options,
		  (CMD_OPTION *)&wld_options, /* options array */
		  sizeof(wld_options), /* size of options array */
		  map);
}

void
wld_init(void)
{
    /* --- Initialize LIBRARY GLOBAL items --- */
    if ( library_init )
       return;

    /* Initialize the world database structure */
    bzero(&wld_pb,sizeof(wld_pb));
	  
    if ( strlen(wld_options.tdbpath.value) > WLD_MAX_SEARCH_PATH_LEN )
    {
	printf("LIBWORLD: Maximum directory path length exceeded\n\tSearching in the default path (../../terrain)");
	/* Assign search path to the default path */
	strcpy(wld_options.tdbpath.value,"../../terrain");
    }
    
    if ( wld_options.terrain.value == NULL )
    {
	wld_options.terrain.value = STDALLOC(256*sizeof(char));
	strcpy(wld_options.terrain.value,WLD_NOWHERE);
    }
    
    strcpy(wld_db_dir,wld_options.tdbpath.value);
    library_init=TRUE;
}

char *
wld_get_tdbpath(void)
{
    return(wld_options.tdbpath.value);
}

char *
wld_get_terrain(void)
{
    return(wld_options.terrain.value);
}

