#ifndef lint
static char rcsid [] = "$RCSfile: wld_lookup.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: wld_lookup.c
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC/SAIC under U. S. Government     *
*       contracts, and may be reproduced by or for the U. S. Government     *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013 (OCT 1988).                                   *
*                                                                           *
*       Contents: Code to access libworld stored data                       *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libwld_local.h"
#include <libgcs.h>
#include <stdstring.h>

int32
wld_get_playbox_utm_grids(int32 *n_thin,
			  float32 *thin,
			  int32 *n_thick,
			  float32 *thick,
			  int32 *n_labels,
			  COORD_GRID_LABEL *labels)
                            /* number of segments allocated for */
                            /* segment contains XY pair */
                             /* number of segments allocated for */
                             /* segment contains XY pair */
                                  
                              
{
    int32 state=0;

    if ( !wld_pb.utm_grid.valid )
      wld_generate_utm_grids(&wld_pb);
    
    if ( *n_thin >= wld_pb.utm_grid.n_thin )
    {
	bcopy(wld_pb.utm_grid.thin,thin,wld_pb.utm_grid.s_thin);
	if ( *n_thick >= wld_pb.utm_grid.n_thick )
	{
	    bcopy(wld_pb.utm_grid.thick,thick,wld_pb.utm_grid.s_thick);
	    if ( *n_labels >=  wld_pb.utm_grid.n_label )
	    {
		bcopy( wld_pb.utm_grid.label,labels,
		      wld_pb.utm_grid.s_label);
		state = 1;
	    }
	}
    }
    
    *n_thin   = wld_pb.utm_grid.n_thin;
    *n_thick  = wld_pb.utm_grid.n_thick;
    *n_labels = wld_pb.utm_grid.n_label;
    return state;
}

int32
wld_lookup_playbox_cells(int32 *n_cell, int32 *cell)
{
    /* Check if the list passed is long enough.  If not, return false and 
     * set n_cells to the size that should be allocated
     */
    if ( *n_cell < wld_pb.n_cells)
    {
	*n_cell = wld_pb.n_cells;
	return 0;
    }

    *n_cell = wld_pb.n_cells;
    bcopy(wld_pb.cells,cell,sizeof(int32)*wld_pb.n_cells);
    return 1;
}

#define CONVERT_LL_TO_GCS(ll,gcs)                                       \
	{int32 cell;                                                    \
	 coord_convert(COORD_LATLON,(ll)[WLD_LAT],(ll)[WLD_LNG],(ll)[Z],\
		       0,COORD_GCS,&cell,&(gcs)[X],&(gcs)[Y],&(gcs)[Z]); \
         (gcs)[CELL3D]=cell; }
 
void
wld_get_playbox_center(float64 *cntr)
{
    /* NOTE: GCS conversions only legal after the database has been read
     * and the rotation and offset matrix have been initialized.
     */
    if ( !wld_pb.gcs.valid )
    {
	if ( wld_pb.ll.valid )
	{
	    CONVERT_LL_TO_GCS(wld_pb.ll.cntr,  wld_pb.gcs.cntr);
	    CONVERT_LL_TO_GCS(wld_pb.ll.sw_crnr, wld_pb.gcs.sw_crnr);
	    CONVERT_LL_TO_GCS(wld_pb.ll.ne_crnr, wld_pb.gcs.ne_crnr);
	    wld_pb.gcs.valid = 1;
	}
	else
	{
	    VMAT3_VEC_ZERO64(cntr);
	    return;
	    /*printf("Playbox Extents undefined - center\n");*/
	}
    }
   
    GCS_CAST_VEC_3D_TO_2D(wld_pb.gcs.cntr,cntr);
}

void
wld_get_playbox_extent(int32 cell, float64 *sw_crnr, float64 *ne_crnr)
{
    if ( !wld_pb.gcs.valid )
    {
	int32 cell;
	
	if ( wld_pb.ll.valid )
	{
	    CONVERT_LL_TO_GCS(wld_pb.ll.cntr,wld_pb.gcs.cntr);
	    CONVERT_LL_TO_GCS(wld_pb.ll.sw_crnr,wld_pb.gcs.sw_crnr);
	    CONVERT_LL_TO_GCS(wld_pb.ll.ne_crnr,wld_pb.gcs.ne_crnr);
	    wld_pb.gcs.valid = 1;
	}
	else
	{
	    VMAT3_VEC_ZERO64(sw_crnr);
	    VMAT3_VEC_ZERO64(ne_crnr);
	    return;
	    /*printf("Playbox Extents undefined - max and min \n");*/
	}
	
    }
    else
    {
	gcs_gcs3_to_gcscs((int32)wld_pb.gcs.sw_crnr[CELL3D],
			   wld_pb.gcs.sw_crnr,cell,sw_crnr);
	gcs_gcs3_to_gcscs((int32)wld_pb.gcs.ne_crnr[CELL3D],
			   wld_pb.gcs.ne_crnr,cell,ne_crnr);
    }
}


void
wld_get_playbox_extent_ll(WLD_PLAYBOX *pb,
			  float64 *south_lat,
			  float64 *west_lng,
			  float64 *north_lat,
			  float64 *east_lng)
{
    if ( wld_pb.ll.valid )
    {
	*south_lat = pb->ll.sw_crnr[WLD_LAT];
	*west_lng  = pb->ll.sw_crnr[WLD_LNG];
	*north_lat = pb->ll.ne_crnr[WLD_LAT];
	*east_lng  = pb->ll.ne_crnr[WLD_LNG];
    }
    else
      *south_lat = *west_lng = *north_lat = *east_lng = 0;
}


void
wld_get_playbox_id(char *name, int32 *version)
{
    if ( strcmp(wld_pb.pbname.terrainName,"") != 0 )
      strncpy(name,wld_pb.pbname.terrainName,maxTerrainNameLength);
    else
      strcpy(name,WLD_NOWHERE);
    
    name[maxTerrainNameLength-1] = '\0';
    *version = wld_pb.pbname.terrainVersion;
}
    
