#ifndef lint
static char rcsid [] = "$RCSfile: wld_playbx.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: wld_playbx.c
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC/SAIC under U. S. Government     *
*       contracts, and may be reproduced by or for the U. S. Government     *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013 (OCT 1988).                                   *
*                                                                           *
*       Contents: Code to initialize libworld                               *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include "libwld_local.h"
#include <libreader.h>
#include <libctdb.h>
#include <libgcs.h>
#ifndef WLD_NO_PBT
#include <libpbtab.h>
#endif
#include <stdalloc.h>
#include <stdstring.h>

extern uint32       wld_gcs_aware;
extern WLD_PLAYBOX  wld_pb;
extern char         wld_db_dir[WLD_MAX_SEARCH_PATH_LEN];

static int32 
wld_process_latlong_string(char *dbname, WLD_PLAYBOX *pb)
{
    char letter[4];
    float32 smx_lng,smx_lat,smn_lat,smn_lng;
    float64 mnlat,mnlng,mxlat,mxlng;
    
    if ( sscanf(dbname,"%f%c%f%cx%f%c%f%c",&smn_lat,&letter[0],
		&smn_lng,&letter[1],&smx_lat,&letter[2],&smx_lng,
		&letter[3]) != 8 )
      return 0;
    
    
    /* Using the Letter after the float change the value accordingly */
    if ( letter[0] == 'S' || letter[0] == 's' )
      mnlat = -smn_lat ;
    else
      mnlat = smn_lat;
    
    if ( letter[2] == 'S' || letter[2] == 's' )
      mxlat= -smx_lat;
    else
      mxlat= smx_lat;
    
    if ( letter[1] == 'W' || letter[1] == 'w' )
      mnlng = -smn_lng;
    else
      mnlng = smn_lng;
    
    if ( letter[3] == 'W' || letter[3] == 'w' )
      mxlng = -smx_lng;
    else
      mxlng = smx_lng;
    
    wld_set_extent_and_center(pb,mnlat,mnlng,mxlat,mxlng);
    return 1;    
}

static void
wld_set_playbox_name(WLD_PLAYBOX *pb, CTDB *ctdb)
{  
    char  name[132];
    int32 version;
    
    if ( ctdb )
    {
	ctdb_get_tdb_identifier(ctdb,name,&version);
	strncpy(pb->pbname.terrainName,name,maxTerrainNameLength);
	pb->pbname.terrainName[maxTerrainNameLength-1] = '\0';
	pb->pbname.terrainVersion = version;
    }
    else
    {
	/* If no ctdb header read then this must have been entered with a 
	 * lat/long extent defined.
	 */
	strncpy(pb->pbname.terrainName,"lat/lng extent\0",
		maxTerrainNameLength);
	pb->pbname.terrainVersion = 0;
    }      
}

static void
wld_dbname_to_extent(char *dbname, WLD_PLAYBOX *pb)
{
    char *filename;
    CTDB ctdb;
        
    gcs_set_cell_awareness(GCS_MULTI_CELL);
    /* Extract the latitude/longitude information from a string in the 
     * proper format.
     */
    if ( pb->kind == CTDB_MODE_MULTI_CELL )
      if ( wld_process_latlong_string(dbname,pb) )
      {
	  wld_set_playbox_name(pb,(CTDB *)NULL);
	  return;
      }   

    /* At this point the -terrain command line text string must be 
     * specifying the name of a known database.  
     */
    
    if ( pb->kind == CTDB_MODE_MULTI_CELL )
    {
	/* Look for the multi-cell version of the database */
	filename = 
	  ctdb_generate_tdb_filename(CTDB_MODE_MULTI_HDR,wld_db_dir,
				     dbname,CTDB_FILE_FORMAT,GCS_ILLEGAL_CELL,
				     FALSE);
	
	if ( ctdb_read_hdr(filename,&ctdb) )
	{
	    float64 mnlat,mnlng,mxlat,mxlng;
	    
	    ctdb_get_geographic_extent(&ctdb,&mnlat,&mnlng,&mxlat,&mxlng);
	    wld_set_extent_and_center(pb,mnlat,mnlng,mxlat,mxlng);
	    wld_set_playbox_name(pb,&ctdb);
	    return;    
	}
	
	/* failed to find a multi cell database, look for a single cell
	 * database.
	 */
	printf("Can't Find GCS MultiCell %s - Search for GCS SingleCell version\n",
	       dbname);
	pb->kind = CTDB_MODE_SINGLE_CELL;	
    }
    
    gcs_set_cell_awareness(GCS_SINGLE_CELL);

    /* look for GCS single-cell version of the database */
    if ( pb->kind == CTDB_MODE_SINGLE_CELL )
    {
	filename = 
	  ctdb_generate_tdb_filename(CTDB_MODE_SINGLE_CELL,wld_db_dir,
				     dbname,CTDB_FILE_FORMAT,GCS_ILLEGAL_CELL,
				     FALSE);
	
	if ( ctdb_read_hdr(filename,&ctdb) )
	{
	    float64 mnlat,mnlng,mxlat,mxlng;
	    
	    ctdb_get_geographic_extent(&ctdb,&mnlat,&mnlng,&mxlat,&mxlng);
	    wld_set_extent_and_center(pb,mnlat,mnlng,mxlat,mxlng);
	    wld_set_playbox_name(pb,&ctdb);
	    return;
	}

	printf("Can't Find GCS SingleCell %s - Search for SIMNET version\n",
	       dbname);
	pb->kind = CTDB_MODE_SIMNET;	
    }
    
    /* Try to load a SIMNET database */
    filename = 
      ctdb_generate_tdb_filename(CTDB_MODE_SIMNET,wld_db_dir, dbname,
				 CTDB_FILE_FORMAT,GCS_ILLEGAL_CELL,
				 FALSE);
    
    if ( ctdb_read_hdr(filename,&ctdb) )
    {
	float64 mnlat,mnlng,mxlat,mxlng;
	
	ctdb_get_geographic_extent(&ctdb,&mnlat,&mnlng,&mxlat,&mxlng);
	wld_set_extent_and_center(pb,mnlat,mnlng,mxlat,mxlng);
	wld_set_playbox_name(pb,&ctdb);
	return;
    }
    
    /* Well at this point we haven't found any database that matches the 
     * string provided.  So have libctdb generate a default database with
     * no features and continue processing.
     */ 
    filename = 
      ctdb_generate_tdb_filename(CTDB_MODE_SIMNET, wld_db_dir, WLD_NOWHERE,
				 CTDB_FILE_FORMAT, GCS_ILLEGAL_CELL, FALSE);

    if ( ctdb_read_hdr(filename,&ctdb) )
      wld_set_playbox_name(pb,&ctdb);
}

/* If the user is not GCS aware then the database is a non GCS 
 * format 5 database which will be used.  Therefore search the 
 * terrain directory for a valid database name.
 */
static CTDB_TDB_TYPE
wld_parse_tdb_type(char *dbtype)
{
    if ( strcmp(dbtype,"none") == 0 )
      return CTDB_MODE_SIMNET;
    else if ( strcmp(dbtype,"single") == 0 )
      return CTDB_MODE_SINGLE_CELL;
    else
      return CTDB_MODE_MULTI_CELL; 
}

#define MAX_SEGS 20000
#define MAX_LBLS 1000
#define XYPAIR   4

void 
wld_generate_utm_grids(WLD_PLAYBOX *pb)
{
    int32             n_thin_segments;
    float32           thin_segments[MAX_SEGS*XYPAIR];
    int32             n_thick_segments;
    float32           thick_segments[MAX_SEGS*XYPAIR];
    int32             n_labels;
    COORD_GRID_LABEL  labels[MAX_LBLS];
    int32             thin_bytes, thick_bytes, label_bytes;
    float64           pbcenter[XYC],pbmin[XYZ],pbmax[XYZ];
    
    wld_get_playbox_center(pbcenter);
    wld_get_playbox_extent((int32)pbcenter[CELL2D],pbmin,pbmax);
    
    coord_generate_grid((int32)pbcenter[CELL2D],(int32)pbmin[X],
			(int32)pbmax[X],
			(int32)pbmin[Y],(int32)pbmax[Y], 
			10000, 5, COORD_LABEL_ALL,
			MAX_SEGS, MAX_SEGS, MAX_LBLS,
			&n_thin_segments, thin_segments,
			&n_thick_segments, thick_segments,
			&n_labels, labels);

    thin_bytes  = n_thin_segments * sizeof(float32) * XYPAIR;
    thick_bytes = n_thick_segments * sizeof(float32) * XYPAIR;
    label_bytes = n_labels * sizeof(COORD_GRID_LABEL);

    if (pb->utm_grid.s_thin < thin_bytes)
    {
	pb->utm_grid.thin = (float32 *)STDREALLOC(pb->utm_grid.thin,
						  thin_bytes);
	pb->utm_grid.s_thin = thin_bytes;
    }

    bcopy(thin_segments,pb->utm_grid.thin,thin_bytes);
    pb->utm_grid.n_thin = n_thin_segments;

    if (pb->utm_grid.s_thick < thick_bytes)
    {
	pb->utm_grid.thick = (float32 *)STDREALLOC(pb->utm_grid.thick,
						   thick_bytes);
	pb->utm_grid.s_thick = thick_bytes;
    }

    pb->utm_grid.n_thick = n_thick_segments;
    bcopy(thick_segments,pb->utm_grid.thick,thick_bytes);

    if (pb->utm_grid.s_label < label_bytes)
    {
	pb->utm_grid.label = (COORD_GRID_LABEL *)
	  STDREALLOC(pb->utm_grid.label,label_bytes);
	pb->utm_grid.s_label = label_bytes;
    }

    pb->utm_grid.n_label = n_labels;
    bcopy(labels,pb->utm_grid.label,label_bytes);
    pb->utm_grid.valid = TRUE;
}


static int32
wld_mem_per_cell(int32 sim_type, int32 n_cells)
{
    return ( !sim_type ? 
	    ((wld_options.tdbmem.value*1024*1024*4)/n_cells):
	    (wld_options.tdbmem.value*1024*1024/n_cells));    
}

static void
wld_create_playbox(int32 simulate, int32 gui)
{
    int32   cell_mem_lim,i;
    float64 mnx,mny,mxx,mxy;
    float64 origin_x, origin_y;
    GCS_FLAT_EARTH flat_earth;
        
    if ( wld_terrain_cmdline_help(wld_options.terrain.value) )
      exit(0);

    wld_pb.kind = wld_parse_tdb_type(wld_options.gcscell.value);
	
    wld_dbname_to_extent(wld_options.terrain.value,&wld_pb);
    
    /* Get the list of cells which make up this playbox extent.  
     * If the list length is not correct the routine will return 
     * the proper length and  a new list will be allocated to 
     * store all of the cells.
     */
    if ( wld_pb.kind == CTDB_MODE_MULTI_CELL )
    {
	while (!gcs_extent_to_cell_list(wld_pb.ll.sw_crnr[WLD_LAT], 
					wld_pb.ll.sw_crnr[WLD_LNG],
					wld_pb.ll.ne_crnr[WLD_LAT], 
					wld_pb.ll.ne_crnr[WLD_LNG],
					&wld_pb.n_cells,wld_pb.cells))
	{
	    if (wld_pb.cells) 
	      wld_pb.cells = (int32 *)STDREALLOC(wld_pb.cells,
						 sizeof(int32)*wld_pb.n_cells);
	    else
	      wld_pb.cells = (int32 *)STDALLOC(sizeof(int32) * wld_pb.n_cells);
	}
    }
    else
    {
	wld_pb.n_cells  = 1;
	wld_pb.cells    = (int32 *)STDALLOC(sizeof(int32));
	wld_pb.cells[0] = GCS_ILLEGAL_CELL;
    }   

    cell_mem_lim = wld_mem_per_cell(simulate,wld_pb.n_cells);
    
    bzero(&flat_earth, sizeof(GCS_FLAT_EARTH));
    
    for (i=0;i<wld_pb.n_cells;i++)
    {
	CTDB *ctdb;
	char *filename;
	int32 cell=wld_pb.cells[i];
	
	ctdb     = (CTDB *)STDALLOC(sizeof(CTDB));

	filename = 
	  ctdb_generate_tdb_filename(wld_pb.kind, wld_db_dir,
				     wld_options.terrain.value,
				     CTDB_FILE_FORMAT,cell, FALSE);
		
	ctdb_set_tiling_mode((wld_pb.kind==CTDB_MODE_MULTI_CELL));
	
	/* read the ctdb cell in */
	ctdb_read(filename,ctdb,cell_mem_lim);

	ctdb_print_description(ctdb);

	/* Use the cell id in the ctdb header as the proper one */
	if ( wld_pb.kind == CTDB_MODE_SIMNET )
	{
	    float64         minx,miny,maxx,maxy;
	    GCS_FLAT_EARTH  flat_earth;
	    CTDB_UTM_DATA   utm;
	    float64         gcc[XYZ];
	    float64         tcc_rot[XYZ][XYZ],rotation[XYZ][XYZ];
	    
	    ctdb_get_utm_data(ctdb, &utm);
	    ctdb_get_extent(ctdb,&minx,&miny,&maxx,&maxy);
	    
	    /* store the conversion function pointers */
	    flat_earth.tccf_to_gcc = 
	      (GCS_TCC_TO_GCC_FUNCTION)coord_tccf_to_gcc;
	    flat_earth.gcc_to_tccf = 
	      (GCS_GCC_TO_TCC_FUNCTION)coord_gcc_to_tccf;

	    /* generate the proper coordinate transformation data */
	    flat_earth.tcc_ptr = 
	      coord_define_tcc((utm.source?COORD_UTM_NE:COORD_GCC),
			       utm.northing,
			       utm.easting,
			       utm.zone_number,
			       utm.zone_letter,
			       utm.datum,
			       (int32)(maxx-minx),
			       (int32)(maxy-miny));

	    if ( utm.source )
	      /* Only flat earth UTM requires special processing.
	       * Register function pointers to libcoordinate routines
	       * that must be called inplace of standard GCS functions.
	       */
	      gcs_init_cell( GCS_ILLEGAL_CELL, &flat_earth);
	    else
	      gcs_init_cell( GCS_ILLEGAL_CELL, NULL);

	    /* get the rotation matrix from the tcc pointer and 
	     * transform it into the format stored in libGCS.
	     */
	    coord_get_rotation((COORD_TCC_PTR)flat_earth.tcc_ptr,tcc_rot);
	    /* Note: tcc stored rotation matrix
	     *
	     *       -sin_lon           cos_lon             0.0 
	     *       -sin_lat*cos_lon  -sin_lat*sin_lon     cos_lat
	     *        cos_lat*cos_lon   cos_lat*sin_lon     sin_lat
	     */
	    vmat3_transpose64(tcc_rot,rotation);

	    /* get the GCC center */
	    coord_get_gcc_origin((COORD_TCC_PTR)flat_earth.tcc_ptr,gcc);
	    
	    gcs_set_coord_data(GCS_ILLEGAL_CELL,rotation,gcc);
	    cell=GCS_ILLEGAL_CELL;
	}
	else
	  gcs_init_cell( ctdb_get_cell_id(ctdb), NULL );
	
	ctdb_get_origin(ctdb, &origin_x, &origin_y);
	gcs_offset_origin(cell, origin_x, origin_y);
	
	/* register the mapping from cell to ctdb */
	gcs_set_tdb(cell, (ADDRESS)ctdb, FALSE);
#ifndef WLD_NO_PBT
	/* alloc space for each cell pbtable and perform the proper
	 * registration functions.
	 */
	gcs_set_pbt_info(cell, (ADDRESS) pbt_alloc_info());

	pbt_select_table(cell);

	ctdb_get_extent(ctdb, &mnx, &mny, &mxx, &mxy);
	    
	pbt_set_size((int32)mnx, (int32)mny, (int32)mxx, (int32)mxy);
#endif
    }

    /* Now, look for and read in any thinned versions of the databases.
     */
    if(gui)
      for (i=0;i<wld_pb.n_cells;i++)
      {
	  CTDB *ctdb;
	  char *filename;
	  int32 cell=wld_pb.cells[i];

	  ctdb     = (CTDB *)STDALLOC(sizeof(CTDB));

	  filename = 
	    ctdb_generate_tdb_filename(wld_pb.kind, wld_db_dir,
				       wld_options.terrain.value,
				       CTDB_FILE_FORMAT,cell, TRUE);

	  ctdb_set_tiling_mode((wld_pb.kind==CTDB_MODE_MULTI_CELL));

	  /* read the ctdb cell in */
	  if(!ctdb_read(filename,ctdb,0))
	    /* register the mapping from cell to ctdb */
	    gcs_set_tdb(cell, (ADDRESS)ctdb, TRUE);
	  else
	  {
	      printf("%s not found, using full version.\n", filename);
	      STDDEALLOC(ctdb);
	  }
      }

    wld_generate_utm_grids(&wld_pb);
}

void
wld_expand_playbox(int32 cell)
{
    int32 i;
    CTDB *ctdb;
    float64 slat,wlng,nlat,elng;
    float64 cur_slat,cur_wlng,cur_nlat,cur_elng;
    int32 new_extent=FALSE;
    
    if ((wld_pb.kind != CTDB_MODE_MULTI_CELL) || (cell == GCS_ILLEGAL_CELL))
      return;

    /* Search the list of active cells, if found then processing is complete 
     * and return.
     */
    for (i=0; i < wld_pb.n_cells; i++)
      if (wld_pb.cells[i] == cell)
	return;

    /* Add this cell to the playbox */
    wld_pb.n_cells++;
    wld_pb.cells = 
      (int32 *)STDREALLOC(wld_pb.cells, sizeof(int32) * wld_pb.n_cells);
    wld_pb.cells[wld_pb.n_cells-1] = cell;
    
    /* Initialize the cell in GCS */
    gcs_init_cell(cell, NULL);
    
    /* In the future, we might want it to load in the CTDB */
#ifdef LOAD_CTDB
    ctdb = (CTDB *)STDALLOC(sizeof(CTDB)); 
#else
    ctdb = (CTDB *)NULL;
#endif   
    /* register the mapping from cell to ctdb */
    gcs_set_tdb(cell, (ADDRESS)ctdb, FALSE);
    
#ifdef LOAD_CTDB
    float64 origin_x,origin_y;
    
    ctdb_get_origin(ctdb, &origin_x, &origin_y); 
    gcs_offset_origin(cell, origin_x, origin_y);
#else
    gcs_offset_origin(cell, 0.0, 0.0);
#endif
    
#ifndef WLD_NO_PBT
    /* alloc space for each cell pbtable and perform the proper
     * registration functions.
     */
    gcs_set_pbt_info(cell, (ADDRESS)pbt_alloc_info());
    pbt_select_table(cell);
    
    /* Since we can't get cell size information from the CTDB,
       just use nominal value */
    pbt_set_size((int32) 0, (int32) 0, (int32) GCS_MAX_CELL_SIDE_METERS,
		 (int32) GCS_MAX_CELL_SIDE_METERS);
#endif
    if ( ctdb )
      ctdb_get_geographic_extent(ctdb,&slat,&wlng,&nlat,&elng);
    else
      gcs_lookup_cell_extents(wld_pb.cells[i],&slat,&wlng,&nlat,&elng);
    
    wld_get_playbox_extent_ll(&wld_pb,&cur_slat,&cur_wlng,&cur_nlat,&cur_elng);
    
    /* Check if the playbox should be expanded */
    if ( cur_slat > slat ) 
      cur_slat = slat, new_extent=TRUE;
    
    if ( cur_nlat < nlat ) 
      cur_nlat = nlat, new_extent=TRUE;
    
    /* check for straddling of IDL for both cell and playbox */
    if ( wlng > elng )
      wlng -= 360.0;/* cell straddles IDL */

    if ( cur_wlng > cur_elng )
      cur_wlng -= 360.0;/* playbox straddles IDL */ 
    
    if ( wlng < cur_wlng )
      cur_wlng = wlng, new_extent=TRUE ;
    
    if ( elng > cur_elng )
      cur_elng =  elng, new_extent=TRUE ;
	
    if ( new_extent )
      wld_set_extent_and_center(&wld_pb, cur_slat, cur_wlng, 
				cur_nlat, cur_elng);
}

int32 
wld_point_in_playbox(int32 c, float64 x, float64 y, float64 z)
{
    float64 sw[XYZ],ne[XYZ];
    
    wld_get_playbox_extent(c,sw,ne);    
    return ((x>=sw[X]) && (x<=ne[X]) && (y>=sw[Y]) && (y<=ne[Y]));
}

void 
wld_load_playbox(int32 sim, int32 gui, int32 argc, char *argv[], 
		 char *data, char *playbox_name, 
		 int32 *playbox_version, COORD_TCC_PTR *tccp, CTDB **ctdb)
{
    float64         minx,miny,maxx,maxy,cntr[XYC];
    CTDB_UTM_DATA   utm;
    
    /* initialize coordinate system */
    coord_init(data, READER_OVERRIDE);

    /* tell libcoordinates to treat all COORD_TCC conversions like a 
     * COORD_GCS conversion.
     */
    coord_set_tcc_awareness(0);
 
    /* global coordinate system initialization */
    gcs_init(coord_get_datum_info);

    /* command line arguments to parse */
    if ( argc && argv )
    {
	CMD_OPTION *options;
	uint32      sizeof_options = 0;
	CMD_TYPES **map;
	int32 leftover_argc;
	char **leftover_argv;

	wld_tdb_aggregate_options(&options, &sizeof_options, &map);
	cmd_process_aggregate_options(argc, argv, /* args to main() */
		      &leftover_argc, &leftover_argv, /* unrecognized */
		      options, sizeof_options, map,
		      TRUE /* Verbose mode */
		      );
	cmd_gripe(leftover_argc, leftover_argv);
    }
      
    /* initialize the playbox */
    wld_init();

    /* If you are running with GCS on then allocate more memory since
     * there are more databases.  The multiplication 
     * by 4 is to have the same default modsaf behavior (default if tdbmem
     * is not specified is 5Mb).
     */
    wld_create_playbox(sim, gui);
   
    /* Get the database/playbox name */
    if ( playbox_name && playbox_version )
      wld_get_playbox_id( playbox_name,  playbox_version );

    if ( ctdb )
    {
	wld_get_playbox_center(cntr);
	*ctdb = (CTDB *)gcs_get_tdb((int32)cntr[CELL2D], FALSE);

	if ( tccp )
	{
	    ctdb_get_extent(*ctdb,&minx,&miny,&maxx,&maxy);
	    ctdb_get_utm_data(*ctdb, &utm);
	    *tccp = coord_define_tcc((utm.source?COORD_UTM_NE:COORD_GCC),
				     utm.northing,
				     utm.easting,
				     utm.zone_number,
				     utm.zone_letter,
				     utm.datum,
				     (int32)(maxx-minx), 
				     (int32)(maxy-miny));
	}    
    }
}
