#ifndef lint
static char rcsid [] = "$RCSfile: wld_set.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: wld_set.c
*                                                                           *
*       Copyright 1995 by SAIC & TASC                                       *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Drive                *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7600                                                        *
*                                                                           *
*	This software was developed by TASC/SAIC under U. S. Government     *
*       contracts, and may be reproduced by or for the U. S. Government     *
*       pursuant to the copyright license under the clause                  *
*       at DFARS 252.227-7013 (OCT 1988).                                   *
*                                                                           *
*       Contents: Code to set items in the WLD_PLAYBOX data structure       *
*       Created: Tue Sep  5 16:10:49 EDT 1995                               *
*       Author: buettner                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libwld_local.h"
/* Routine : wld_set_center
 *
 * This routine will set the center of the playbox in the WLD_PLAYBOX structure
 */
static void
wld_set_center(
	       WLD_PLAYBOX *pb,   
	       float64      mn_lat,
	       float64      mn_lng,
	       float64      mx_lat,
	       float64      mx_lng
	       );


void
wld_set_extent_and_center(WLD_PLAYBOX *pb,
			  float64 mn_lat,
			  float64 mn_lng,
			  float64 mx_lat,
			  float64 mx_lng)
{
    /* Set the maximum and minimum latitude and longitude points */
    wld_set_center(pb,mn_lat,mn_lng,mx_lat,mx_lng);
    
    pb->ll.sw_crnr[WLD_LAT] = mn_lat;
    pb->ll.sw_crnr[WLD_LNG] = mn_lng;
    pb->ll.ne_crnr[WLD_LAT] = mx_lat;
    pb->ll.ne_crnr[WLD_LNG] = mx_lng;

    wld_set_center(pb,mn_lat,mn_lng,mx_lat,mx_lng);
    
    /* Invalidate the GCS transformed version after setting the new 
     * latitude and longitude.  This will force the recalculation of
     * the GCS version when it is requested.
     */
    pb->ll.valid       = TRUE;
    pb->gcs.valid      = FALSE;

    /* Invalidate the pregenerated grids so that when they are requested 
     * they will be regenerated first.
     */
    pb->utm_grid.valid = FALSE;
}

static void
wld_set_center(WLD_PLAYBOX *pb,
	       float64 mn_lat,
	       float64 mn_lng,
	       float64 mx_lat,
	       float64 mx_lng)
{
    
    pb->ll.cntr[WLD_LAT] = ( mx_lat + mn_lat )*0.5;
    
    if ( mn_lng > mx_lng )
    {
	/* The longitude extent straddles the international dateline */
	float64 cntr;
	
	mx_lng += 360.0;
	cntr = (mx_lng-mn_lng)*0.5;

	if ( (mn_lng+cntr) > 180.0 )
	  pb->ll.cntr[WLD_LNG] = mx_lng - cntr - 360.0 ;
    }    
    else
      pb->ll.cntr[WLD_LNG] = ( mx_lng + mn_lng ) * 0.5 ;

    /* Invalidate the GCS transformed version of the maximum */
    pb->gcs.valid = 0;
}
