#include "Globals.h"
#include "CmsControl.h"
#include "NetMgr.h"

#include <stdio.h>

#include <sys/time.h>
#include <math.h>
#include <new.h>

#include <stddef.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/syssgi.h>
#include <unistd.h>

char Datapath[256] = "../../data";
char DatapathConst[256];
char Const[256] = "/constants";
char Setupdir[256] = "./setup";
argv_t av;
int ac;

//   global constants, first simple types then objects in order of construction
EBoolean         gbGui  = true;
EBoolean         gbHost = true;

cNetMgr*         gpNetMgr = NULL;
cCmsMgr          gCmsMgr;

const char*   kpCmsVersion = "3.0b2.2";
char*         gpRunName = "CMS test run";
const double kUsecToSimTime = 0.001;
const double kSecToSimTime = 1000.0;
const double kSimTimeToDis = 2147.4836470 / 3.6;   // ~ 596.5


long   gSimStartTime = 0;      // in seconds, not TTimeSim

#ifdef sgi
volatile iotimer_t *gpTimerAddr = NULL;
double gCycleVal = 0.0;
long   gStartTimeMilli = 0;
unsigned int gSimStartCycle = 0;
static unsigned int max_diff = 0;

#endif

#ifdef sgi
void   SetSimStartTime() {

   unsigned int     tm1, tm2, tm3, cycleval;

   //  Most of the below code can be found by doing a man on syssgi
   //  Don't really understand it all, but it works.
   int              fd, poffmask;
   __psunsigned_t   phys_addr, raddr;
   timeval t;
   int cnt, counter_size;

   poffmask = getpagesize() - 1;
   phys_addr = syssgi(SGI_QUERY_CYCLECNTR, &cycleval);

   // This multiplier converts cycles to milliseconds.
   // cycleval is given in number of picoseconds per counter cycle.
   gCycleVal = ((double)cycleval)/1000000000.00;
   max_diff = (unsigned int)(2000.0/gCycleVal);

   raddr = phys_addr & ~poffmask;
   fd = open("/dev/mmem", O_RDONLY);
   gpTimerAddr = (volatile iotimer_t *)mmap(0, poffmask, PROT_READ,
					     MAP_PRIVATE, fd, (__psint_t)raddr);

   gpTimerAddr = (iotimer_t *)((__psunsigned_t)gpTimerAddr + (phys_addr & poffmask));

   // Now determine how many bits are used for the counter
   tm1 = *gpTimerAddr;
   tm2 = *gpTimerAddr;
   tm3 = *gpTimerAddr;

   if ((tm1 == tm2) || (tm2 == tm3)) {
      fprintf (stdout, "Using 64 bit cycle counter\n");
      gpTimerAddr += 1;
      counter_size = 64;
      gSimStartCycle = *gpTimerAddr;
   }
   else {
      for (cnt=0;cnt<3000000;cnt++)
	 if (*gpTimerAddr & 0xff000000)
	    {
	       fprintf(stderr, "Using 32 bit cycle counter.\n");
	       counter_size = 32;
	       gSimStartCycle = *gpTimerAddr;
	       break;
	    }
      if (counter_size == 0) {
	 fprintf(stderr,
		 "Only 24 bit cycle counter available.  Using gettimeofday.\n");
	 gpTimerAddr = NULL;
	 gSimStartCycle = 0;
      }
   }


   // Set simulation start time.
   // Used for periodic checking that cycle counter is accurate.
   // i.e.  - we haven't missed any rollovers.
   gettimeofday(&t);
   gStartTimeMilli = (t.tv_usec * kUsecToSimTime );
   gSimStartTime = t.tv_sec;
}

TTimeSim   GetCurSimTime() {
      //   returns TTimeSim since start of sim
      //
      // looks at hardware cycle counter and determines sim time,
      // in milliseconds.
   static TTimeSim LastRolloverTime = 0;
   static unsigned int lastcycle = 0;
   unsigned int cycle;
   timeval t;
   //printf ("max_diff = %u, gCycleVal = %lf\n", max_diff, gCycleVal);

   // Check if we can use the cycle counter
   // I would like to not to have to do this every time.
   if (gpTimerAddr != NULL) {
      
      // Get the current cycle count
      cycle = *gpTimerAddr;
 
      // Make sure cycle counter has not rolled over or gone unchecked for 
      // more than 2 seconds.  The latter case should only occur if other processes 
      // are hogging the CPU or are suspending CMS, such as a debugger.  In these cases
      // we play the odds that the current cycle count does not land close to the lastcycle 
      // count too often.
      if ( lastcycle < cycle && cycle - lastcycle < max_diff ) { 

         // reset lastcycle
	 lastcycle = cycle;

	 // Determine time.
	 return LastRolloverTime + (TTimeSim)((double)(cycle - gSimStartCycle)*gCycleVal);
      }
      else { // The counter has rolled over, or gone unchecked for too long.  Use gettimeofday.

	 gettimeofday(&t);
	 // Reset the time of the last rollover
	 // gLastRolloverTime = (TTimeSim)( (MAXNUMCYCLES - ((double)gSimStartCycle))*gCycleVal);
	 LastRolloverTime = ((t.tv_sec - gSimStartTime) * kSecToSimTime + 
			      (TTimeSim)(((double)t.tv_usec - (double)gStartTimeMilli) * (double)kUsecToSimTime ));
         // reset lastcycle
	 lastcycle = cycle;

	 // Cycle when Rollover time was set
	 gSimStartCycle = cycle;

	 // Now return current time.
	 return LastRolloverTime;
      }
   }
   else {
      gettimeofday(&t);
      return ( (t.tv_sec - gSimStartTime) * kSecToSimTime +  
	       (TTimeSim)(((double)t.tv_usec - (double)gStartTimeMilli) * (double)kUsecToSimTime ));
   }   
}

#else

void   SetSimStartTime() {
   timeval t;
   gettimeofday(&t);
   gSimStartTime = t.tv_sec;
}

TTimeSim   GetCurSimTime() {
      //   returns TTimeSim since start of sim
      //
      // system call 
      //   int   gettimeofday(timeval* t)
      // returns 0 for no error, and sets seconds and microseconds in 
      //   t.tv_sec and t.tv_usec (each are longs)
   timeval t;
   gettimeofday(&t);
   
   return ( (t.tv_sec - gSimStartTime) * kSecToSimTime + t.tv_usec * kUsecToSimTime );
}

#endif /*sgi*/

char*     cppstrdup( const char* pStr)
{
   char* pNew = new char[strlen(pStr) + 1];
   strcpy(pNew, pStr);
   return pNew;
}

//void   NewFailureHandler()
//{
 //  Assert( false, kerrMemory, "NewFailureHandler" );
//}

//   main
void   Assert(int test, EerrorNumber errNum, char* fnName, int sysErr )
{
}

void main (int argc, char* argv[])
{
	ac=argc;
	av=argv;
	strcpy(DatapathConst, Datapath);
	strcat(DatapathConst, Const); 

 //  set_new_handler( &NewFailureHandler );
   
   gCmsMgr.Initialize( &argc, argv);
   gCmsMgr.Run();
   gCmsMgr.Shutdown();
}


int   yywrap()
{
   return 1;
}
