
#define MULTIADDR 224.002.140.183

#include "NetMgr.h"
#include "CmsControl.h"
#include <stdio.h>
#include <net/raw.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <unistd.h>
#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <iostream.h>
#include <stdlib.h>

int      gPduLen  = 0;

cPort::cPort() {}
cPort::~cPort() {}

int    cUdpPort::OpenPort(int udpPort)
{
   int on = 1;
   u_char off = (u_char) 0;
   int cc;
   char* pMultireq;

   sport = rport = udpPort;
   hostid = gethostid();
   
      // SET UP THE RECEIVING SOCKET
   cc = 1024 * ebufSize;
   r  = socket(AF_INET, SOCK_DGRAM, 0);
   
   setsockopt(r, SOL_SOCKET, SO_REUSEADDR, (const char*)&on, sizeof(on));
   setsockopt(r, SOL_SOCKET, SO_REUSEPORT, (const char*)&on, sizeof(on));   
   Assert ( !setsockopt(r, SOL_SOCKET, SO_BROADCAST, &on, sizeof (on)), kerrNoRecvBcast, "OpenPort");
   Assert ( !setsockopt(r, SOL_SOCKET, SO_RCVBUF, &cc, sizeof(cc)), kerrNoBuff, "OpenPort");
 
   rin.sin_family      = AF_INET;
   rin.sin_addr.s_addr = htonl(INADDR_ANY);
   rin.sin_port        = htons(udpPort);
 
   //Bind the recieving socket
   Assert ( !bind(r, (sockaddr *)&rin, sizeof(sockaddr_in)), kerrBindSocket, "OpenPort", errno);
 
   Assert ( !ioctl(r, FIONBIO, &on), kerrNoBlock, "OpenPort", errno);
   
   //Create the sending socket
   Assert ( (s = socket(AF_INET, SOCK_DGRAM, 0)) != -1, kerrCreateSendSocket, "OpenPort", errno);

   //Set the socket for broadcasting
   Assert ( !setsockopt(s, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)), kerrNoSendBcast, "OpenPort", errno);

   //  --- no need to bind sending socket ---
 

   //Check if the user specified multicast
   if ((pMultireq = (char *)gCmsMgr.fStartupParam.fMulticastAddr) != NULL)
   {
      fprintf(stdout, "Using multicast address %s\n", pMultireq);

      //Check we have a valid multicast address
      Assert(IN_MULTICAST(inet_addr (pMultireq)), kerrInvalidMcast, "OpenPort", errno);
      //Get the TTL from command line input
      sMttl = (u_char)gCmsMgr.fStartupParam.fTtl;
      fprintf(stdout, "Multicast TTL = %d\n", sMttl);

      //Now set multicast parameters for RECIEVING SOCKET
      //Set the multicast address.
      sMreq.imr_multiaddr.s_addr = inet_addr (pMultireq);
      sMreq.imr_interface.s_addr = INADDR_ANY;
      
      //Now join the multicast group.  We don't need to worry about TTL for receiving.
      Assert(!setsockopt(r, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char*)&sMreq, sizeof(sMreq)), 
	     kerrNoSubscribeMcast, "OpenPort (recv socket)", errno);
   
      // Disable Loopback
      Assert(!setsockopt(r, IPPROTO_IP, IP_MULTICAST_LOOP, (char *) &off, sizeof (off)), 
	     kerrNoDisableLoop, "OpenPort (recv socket)", errno);


      //Now set up Multicast parameters for SENDING SOCKET
      //Set the multicast destination address
      sin.sin_family      = AF_INET;
      sin.sin_addr.s_addr = inet_addr (pMultireq);
      sin.sin_port        = htons(udpPort);

      //Set the TTL.  This is where it matters 
      Assert(!setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL, (char *) &sMttl, sizeof(sMttl)),
	     kerrNoTtlMcast, "OpenPort (send socket)", errno);

      //This setsockopt doesn't work.
      //if (setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF,(char *)sMreq.imr_interface.s_addr, sizeof(struct in_addr)) == -1)
      // perror("Error setting  IP_MULTICAST_IF in OpenPort for sending socket");

      //Assert(!setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&sMreq, sizeof(sMreq)),
      //     kerrNoDisableLoop, "OpenPort (send socket)", errno);
	     

   }
   else  // if (fMulticastAddr != NULL)
   {
      // Set the broadcast destination address
      sin.sin_family      = AF_INET;
      sin.sin_addr.s_addr = htonl(INADDR_BROADCAST);
      sin.sin_port        = htons(udpPort);
   }
 
   return 0;

}

int    cUdpPort::ClosePort()
{
   close(s);
   close(r);
   return 0;
}

ENetReturn    cUdpPort::GetPdu(void *buffer, int buflen)
{
   rlen = 0;
   rinlen = sizeof(sockaddr_in);
   errno = 0;
      rlen = recvfrom(r, buffer, buflen, 0, (sockaddr *)&rin, &rinlen);
      switch (errno) {
         case 0:   
                  errcode = enetMsgRecv;   //success, msg read
            break;
   
         case EWOULDBLOCK:   // nothing to read
               errcode = enetNoData;
            break;
            
         default:
               errcode = enetErr;
            break;
      }
   
   if ( (errcode != enetNoData) && (rlen < 0) ) errcode = enetErr;
   
   Assert( errcode != enetErr, kerrRecvError, "cUdpPort::GetPdu", errno);

   return errcode;
}

int    cUdpPort::PutPdu(void *buffer, int buflen)
{
   int pdutype = ((DIS203_PDU_HEADER_RECORD*) buffer)->pdu_type;
  cout << "PutPdu - pdutype "<< pdutype << endl;

   errno = errcode = enetMsgRecv;
 
   slen = sendto(s, buffer , buflen, 0, (sockaddr *)&sin, sizeof(sockaddr_in));
   
   Assert ( slen > 0, kerrNoPduSend, "cUdpPort::PutPdu");
 
   return slen;
}

cUdpPort::cUdpPort(int udpPort)
{
   OpenPort(udpPort);
   
}

cUdpPort::~cUdpPort()
{
   ClosePort();
}

EBoolean         cNetMgr::RecvPdu()
{
         // check against last quit time, to warn about delays
   
   ENetReturn   getResultCode;      
   EBoolean     recv = false;
   
   TTimeSim   start = GetCurSimTime();
   EBoolean bBreak;
   
   do  {
      bBreak = false;
      if(fpDISPort->GetPdu(fpBuffer, ePduBufSize)!=enetNoData)
      { int pdutype = ((DIS203_PDU_HEADER_RECORD*) fpBuffer)->pdu_type;
           cout << "dis - hla PutPdu - pdutype "<< pdutype << endl;

          if(fpHLAPort) fpHLAPort->PutPdu(fpBuffer, ePduBufSize);
      }
      if(fpHLAPort->GetPdu(fpBuffer, ePduBufSize)!=enetNoData)
      { 
           int pdutype = ((DIS203_PDU_HEADER_RECORD*) fpBuffer)->pdu_type;
           cout << "hla - dis PutPdu - pdutype "<< pdutype << endl;
          if(fpDISPort) fpDISPort->PutPdu(fpBuffer, ePduBufSize);
      }
     
      if ( GetCurSimTime() - start >= 50 ) {
         bBreak = true;
      }
   } while ( (getResultCode == enetMsgRecv) && (bBreak == false) );
   
   return recv;
}

void    cNetMgr::NetShutdown()
{
}

cNetMgr::cNetMgr( TStartupParam & startup )
{

   fExerciseId = startup.fExerciseId;
   fpBuffer = new long[ ePduBufSize/sizeof(long) + 1 ];
   fpDISPort = new cUdpPort(startup.fDisPort);
   fpHLAPort = new cHLAPort();
}


cNetMgr::~cNetMgr()
{
   if (fpDISPort) delete fpDISPort;
   if (fpHLAPort) delete fpHLAPort;
}

int    cHLAPort::OpenPort()
{  try
      {
         rtiAmb.createFederationExecution( FederationName, FederationName );
      }
      catch ( RTI::FederationExecutionAlreadyExists& e )
      {
         cerr << "FED_HW: Note: Federation execution already exists." << &e
<< endl;
      }
      catch ( RTI::RTIinternalError& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
      }
      catch ( RTI::Exception& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
      }
    RTI::Boolean Joined    = RTI::RTI_FALSE;
    int          numTries  = 0;

     while( !Joined && (numTries++ < 20) )
      {
         try
         {
            federateId = rtiAmb.joinFederationExecution( 
									(char* const) federateName,
                                                  FederationName,
                                                  &fedAmb);
            Joined = RTI::RTI_TRUE;
         }
         catch (RTI::FederateAlreadyExecutionMember& e)
         {
            cerr << "FED_HW: ERROR: " << federateName 
                 << " already exists in the Federation Execution "
                 << FederationName << "." << endl;
            cerr << &e << endl;
            return -1;
         }
         catch (RTI::FederationExecutionDoesNotExist& e)
         {
            cerr << "FED_HW: ERROR: " << FederationName 
                 << " Federation Execution "                 
                 << "does not exists."<< endl;
#ifdef _WIN32
            Sleep(2000);
#else
            sleep (2); // sleep 2 seconds ::Sleep(2000) on Windows
#endif   
         }
         catch ( RTI::Exception& e )
         {
            cerr << "FED_HW: ERROR:" << &e << endl;
         }
      } // end of while

   return 0;

}
int    cHLAPort::ClosePort()
{
   try
      {
         rtiAmb.resignFederationExecution(
                   RTI::DELETE_OBJECTS_AND_RELEASE_ATTRIBUTES );
      }
      catch ( RTI::Exception& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
         return -1;
      }

   try
      {
         rtiAmb.destroyFederationExecution( FederationName );
      }
      catch ( RTI::Exception& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
         return -1;
      }

   return 0;
}

ENetReturn    cHLAPort::GetPdu(void *buffer, int buflen)
{ 
  char * cbuffer = (char *) buffer;
  if(fedAmb.q.length()>0) {
     charbuffer msg  = fedAmb.q.deQ();
     if(msg.data!=NULL) {
       for(int i=0;i<msg.len;i++) cbuffer[i] = msg.data[i];
       buflen = msg.len;
       return enetMsgRecv;
     }
  }
  return enetNoData; 
}

int    cHLAPort::PutPdu(void *buffer, int buflen)
{
  int ret = 0;
  RTI::ParameterHandleValuePairSet* pParams = NULL;
  pParams = RTI::ParameterSetFactory::create( 1 );

  pParams->add(disParamTypeId,
              (char*) buffer,
              ((buflen+1)*sizeof(char)));
  try
  {
    rtiAmb.sendInteraction(disInterClassTypeId, *pParams, NULL);
  }
  catch ( RTI::Exception& e )
  {
     cerr << "FED_HW: Error:" << &e << endl;
     return -1;
  }
  delete pParams;
  
  /*RTIfedTime requestTime(1);
  requestTime += grantTime;
  try {
     rtiAmb.timeAdvanceRequest(requestTime);
  }
  catch(RTI::Exception& e)
  { cerr << "FED_HW: Error:" << &e << endl;
  }
  */
  return ret;
}
cHLAPort::cHLAPort()
{  
   unsigned int temp  = (unsigned int)(kSimTimeToDis * GetCurSimTime ()) << 1;
   sprintf(federateName,"%s_%d","DISHLAPub",temp);
   timeAdvGrant = RTI::RTI_FALSE;
   OpenPort();
   PublishAndSubscribe();
   decideTiming();
}
cHLAPort::~cHLAPort()
{
   ClosePort();
}

int cHLAPort::PublishAndSubscribe()
{ int ret = 0;
  try {
    disInterClassTypeId = rtiAmb.getInteractionClassHandle(disInterClassName);
    fedAmb.SetdisInterClassTypeId(disInterClassTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    disParamTypeId = rtiAmb.getParameterHandle(disParamName,
                                               disInterClassTypeId);
    fedAmb.SetdisParamTypeId(disParamTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.subscribeInteractionClass(disInterClassTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.publishInteractionClass(disInterClassTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }
  return ret;
}

int cHLAPort::decideTiming()
{ int ret = 0;
  EBoolean doTick = false;

  try {
    rtiAmb.disableTimeConstrained();
    doTick = true;
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.disableTimeRegulation();
    doTick = true;
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.enableAsynchronousDelivery();
    doTick = true;
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  return ret;
}
