
#ifndef _UNetMgr_
#define _UNetMgr_

#include "HwFederateAmbassador.h"
#include "Globals.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

enum ENetReturn { enetMsgRecv = 0, enetMsgSent, enetNoData, enetErr, enetSelfMsg };
enum EbufKb { ebufSize = 512 };

extern  int      gPduLen;

class cPort {
      // base class for either UDP or Assoc net port

public:
   virtual   ENetReturn      GetPdu(void *buffer, int buflen) = 0;
   virtual   int             PutPdu(void *buffer, int buflen) = 0;

   cPort();
   virtual   ~cPort();
};


class cUdpPort : public   cPort {

protected:
   ulong         hostid;		   // ip addr of host
   int           sport;          // send port number
   int           s;   		   	// send socket
   sockaddr_in   sin;            // send sockaddr_in
   int           slen;			   // # bytes sent
   struct ip_mreq  sMreq;                   // multicast address
   u_char        sMttl;                     // multicast time-to-live
   int           rport;			   // recv port number
   int           r;			      // recv socket
   sockaddr_in   rin;	         // recv sockaddr_in
   int           rinlen;		   // length of addr recv'd from
   int           rlen;			   // # bytes recvd
   ENetReturn     errcode;		   // err code set during opens/reads/or writes
   
   int            OpenPort(int udpPort);
   int            ClosePort();
   
public:
   virtual ENetReturn    GetPdu(void *buffer, int buflen);
   virtual int           PutPdu(void *buffer, int buflen);

              cUdpPort(int udpPort);
   virtual   ~cUdpPort();
};

class cNetMgr {

enum EBufSize { ePduBufSize = 1536 };

      cPort*            fpDISPort;
      cPort*            fpHLAPort;
      void*             fpBuffer;
      int               fPduLength;
      
      int               fExerciseId;

public:
      EBoolean          RecvPdu();      // true if one or more processed
      void              SendPdu(void * pBuf, int length);
            
      void              NetShutdown();
   
                        cNetMgr( TStartupParam & startup );
                       ~cNetMgr();
};


class cHLAPort : public   cPort { 

protected:
   RTI::RTIambassador rtiAmb;
   HwFederateAmbassador fedAmb;
   RTI::FederateHandle federateId;

   char* const FederationName = "HelloWorld";
   char* const disInterClassName = "Communication";
   char* const disParamName = "Message ";
   char federateName[200];

   RTI::Boolean        timeAdvGrant;
   RTIfedTime          grantTime;

 
   short disInterClassTypeId; 
   int disParamTypeId;

   ENetReturn     errcode;

   int            OpenPort();
   int            ClosePort();
   
public:
   virtual ENetReturn    GetPdu(void *buffer, int buflen);
   virtual int           PutPdu(void *buffer, int buflen);
   cHLAPort();
   virtual   ~cHLAPort();
   int PublishAndSubscribe();
   int decideTiming();

};

#endif
