
#include <iostream.h>

template <class T> class item {

  public:
    item(const T &t, item<T> *&link);
    ~item();
    T& value();
	void value(T t);
    item<T>*& prev();   
    item<T>*& next();   
    item<T> operator = (const item<T> & rhs);
  private:
    item<T> *_prev;
    item<T> *_next;
    T _t;
};

template <class T> 
inline item<T>::~item() { }

template <class T> 
inline T& item<T>::value() { return _t; }

template <class T> 
inline void item<T>::value(T t) {  _t = t; }

template <class T> 
inline item<T>*& item<T>::prev() { return _prev; }

template <class T> 
inline item<T>*& item<T>::next() { return _next; }

template <class T> 
item<T> item<T>::operator = (const item<T> & rhs)
{ _prev = rhs._prev;
  _next = rhs._next;
  _t = rhs._t;
  return *this;
}

template <class T> class mtlist {

  public:
    mtlist();
    ~mtlist();
    void add(const T &t);
    bool remove(T t);
	void clean();
	bool advanceCurrent();
	bool getCurrent(T& t);
	bool replaceCurrent(T& t);
	void rewindCurrent();
	void removeCurrent();
    int length();

  private:
    item<T> *head;
	item<T> *tail;
	item<T> *current;
    int len;
};

template <class T> inline int mtlist<T>::length() { return len; }

template <class T> item<T>::item(const T &t, item<T> *&link) {

  _t = t;         // store value
  _prev = link;   // this is at head of list so there is no prev
  _next = NULL;   // link to old first item
}

template <class T> 
mtlist<T>::mtlist() : head(NULL), tail(NULL), 
                       current(NULL), len(0) { 

}

//----< queue destructor >-------------------------------------
//
//  walk down list and delete each item
//
template <class T> mtlist<T>::~mtlist() {

  while(head != NULL) {
    item<T> *temp = head;
    head = head->next();
    delete temp;
  }
  head = tail = current = NULL;
}

template <class T> void mtlist<T>::clean() {

  while(head != NULL) {
    item<T> *temp = head;
    head = head->next();
    delete temp;
  }
  head = tail = current = NULL;
}

//----< add element to queue >-------------------------

template <class T> void mtlist<T>::add(const T &t) {

  tail = new item<T>(t,tail);     
  if(head!=NULL)
     tail->prev()->next() = tail;
  else head = tail; 
  len++;
}

template <class T> bool mtlist<T>::remove(T t) {
  bool found = false; 
  if(head == NULL) {              // empty queue so return safe
    return found;
  }
  
 item<T> *temp = head;         
  while(temp!=NULL&&temp->value()!=t)
  { temp = temp->next();
    
  }
  if(temp!=NULL&&temp->value()==t) 
  { len--;
     item<T> *tmp = temp->prev();
     if(tmp!=NULL)
		tmp->next() = temp->next();
	 else {
		head = temp->next();
		if(head!=NULL) 
			head->prev()=NULL;
	 }
	tmp = temp->next();
	if(tmp!=NULL) 
		tmp->prev() = temp->prev();
	else {
        tail = temp->prev();
		if(tail!=NULL) 
			tail->next() = NULL;
	 }
	delete temp;
	found = true;
  }
  
  return found;
}

template <class T> bool mtlist<T>::advanceCurrent() {
	bool found = false;
	if(current!=NULL) {
		current = current->next();
		if(current!=NULL) found = true;
	}
    return found;
}

template <class T> bool mtlist<T>::getCurrent(T& t) {
    bool found = false;
	if(current!=NULL) {
		t = current->value();
		found = true;
	}
	return found;
}

template <class T> bool mtlist<T>::replaceCurrent(T& t) {
    bool found = false;
	if(current!=NULL) {
		current->value(t);
		found = true;
	}
	return found;
}

template <class T> void mtlist<T>::removeCurrent() {
	if(current!=NULL) { 
	  item<T> * temp = current; 
	  current = current->next();  
	  len--;
      item<T> *tmp = temp->prev();
      if(tmp!=NULL)
		tmp->next() = temp->next();
	  else {
		head = temp->next();
		if(head!=NULL) 
			head->prev()=NULL;
	  }
	  tmp = temp->next();
	  if(tmp!=NULL) 
		tmp->prev() = temp->prev();
	  else {
        tail = temp->prev();
		if(tail!=NULL) 
			tail->next() = NULL;
	 }
	delete temp;
	}
}

template <class T> void mtlist<T>::rewindCurrent() {
	current = head;
}

