/*
$RCSfile: AnalData.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:26 $
$Log: AnalData.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:26  jchappel
Comprehensive Mine Simulator

// Revision 2.1  1998/12/08  20:13:06  jchappel
// Made adjustments for target id
//
// Revision 2.0  1998/01/26  20:52:09  ksnively
// Baseline cms before AP replacement mine additions.
//
// Revision 2.0  1998/01/26  20:52:09  ksnively
// Baseline cms before AP replacement mine additions.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.4  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.2  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.8  1996/08/09  14:33:27  rcarp
// changed to support los check by terrain manager, and its use by offroute mines
//
// Revision 1.7  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.6  1996/06/19  19:05:45  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.5  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.3  1996/02/16  19:32:41  rcarp
// revisions for 1.1b7 version
//
// Revision 1.2  1996/02/05  14:32:43  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: AnalData.cc
   
      Copyright 1996 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   AnalData.cc
   
      Analysis data output implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "AnalData.h"
#include "Minefield.h"
#include "AtWideArea.h"
#include "Vulnerability.h"

#include <math.h>

char*   cInteraction::fsResults[eNumResultTypes] = {
   "TrackDropout",
   "NotEngaged",
   "MissedDetection",
   "AttackedTarget",
   "SelfDefenseAttack",
   "HitTarget",
   "MissedTarget",
   "TargetSkippedCounting",
   "MineFailure",
   "TargetGone",
   "MineNeutralized",
   "MineDetonated",
   "MineDisplaced",
   "MineRollback",
   "AttackConditions",
   "TargetTooSlow",
   "TargetTooFast",
   "TargetOutOfRange",
   "TargetTooClose",
   "NoIntervisToTarget",
   "PrimaryTarget",
   "TargetSkippedWindow"
};

void cInteraction::AddIgnoreTime (TTimeSim time)
{
   int i = 0;

   while (fIgnoreTimes[i] != 0 && i < 10)  i++;

   if (i < 10)
      fIgnoreTimes[i] = time;
}

void cInteraction::AddDetTime (TTimeSim time)
{
   int i = 0;

   while (fDetTimes[i] != 0 && i < 10)  i++;

   if (i < 10)
      fDetTimes[i] = time;
}

void   cInteraction::SetResult(EInteractionResult eResult)
{
   feResult = eResult;
}

void    cInteraction::UpdateInteraction(cLinkVehGeom* pLinkVeh)
{
   if ( pLinkVeh->GetRange() < fClosestRange) {
      fClosestRange  = pLinkVeh->GetRange();
      fClosestPt     = pLinkVeh->GetLocation();
      fVehLocAtClosestPt = pLinkVeh->GetVehicle()->GetTccLoc();
      fClosestTime   = GetCurSimTime();
   }
   if ( fabs(pLinkVeh->GetLocation().y) < fProjClosestOffset )
      fProjClosestOffset = fabs( pLinkVeh->GetLocation().y );
}

void    cInteraction::WriteInterSpecifics()
{
}

void    cInteraction::WriteTimeSumm()
{
   int i;
      *gpInteract << "Start and end time: " 
		  << fStartTime * kSimTimeToSec << " " 
		  << fEndTime  * kSimTimeToSec
		  << endl;

      if (fIgnoreTimes[0] != 0) {
	 i = 0;
	 *gpInteract << "Target ignore times: ";
	 while (i < 10 && fIgnoreTimes[i] != 0) {
	    *gpInteract << fIgnoreTimes[i] * kSimTimeToSec << " ";
	    i++;
	 }
	 *gpInteract << endl;
      }
	    
      if (fDetTimes[0] != 0) {
	 i = 0;
	 *gpInteract << "Target detonate times: ";
	 while (i < 10 && fDetTimes[i] != 0) {
	    *gpInteract << fDetTimes[i] * kSimTimeToSec << " ";
	    i++;
	 }
	 *gpInteract << endl;
      }
}

void    cInteraction::WriteClosestSumm()
{
   *gpInteract << "Closest approach time, proj/actual distance, map loc, down and cross range: " 
               << fClosestTime  * kSimTimeToSec << " " 
               << fProjClosestOffset << " " 
               << fClosestRange << " " 
               << fVehLocAtClosestPt.x << " "
               << fVehLocAtClosestPt.y << " "
               << fVehLocAtClosestPt.z << " "
               << fClosestPt.x << " "
               << fClosestPt.y
               << endl;
}

void    cInteraction::WriteInterResult()
{
   *gpInteract << "result: " 
               << fsResults[feResult]
               << endl;
}

void   cInteraction::WriteInteraction(ofstream* pOut)
{
   if ( fEndTime == 0 )
      fEndTime = GetCurSimTime();
   if ( fClosestTime == 0 )
      fClosestTime = GetCurSimTime();
      
   *pOut << "{ Interaction" << endl;
   WriteTimeSumm();
   fpAttEnt->WriteEntSumm(pOut);
   *pOut << endl;
   fpTgtVeh->WriteEntSumm(pOut);
   *pOut << endl;
   WriteClosestSumm();
   WriteInterResult();
   WriteInterSpecifics();
   *pOut << "}\n" << endl;
}

cInteraction::cInteraction(cEntity* pAttacker, cLinkVehGeom* pLinkVeh)
{
   fpAttEnt = pAttacker;
   fpTgtVeh = pLinkVeh->GetVehicle();
   fStartTime = GetCurSimTime();

   for (int i=0; i < 10; i++) {
      fDetTimes[i] = fIgnoreTimes[i] = 0;
   }

   fEndTime = 0;
   fClosestTime = 0;
   fProjClosestOffset = fClosestRange = fClosestPt.x = fClosestPt.y = FLT_MAX;
   feResult = eTracking;
}

cInteraction::~cInteraction()
{
}

void   cWamInteraction::SetResult(EInteractionResult eResult)
{
   cInteraction::SetResult(eResult);
   
   if ( eResult == eAttackedTarget )
      fFireAz = ((cAtWideArea*)fpAttEnt)->GetFireAz();
}

void    cWamInteraction::WriteInterSpecifics()
{
   if ( fFireAz < 99.9 ) {
      *gpInteract << "firing azimuth: "
                  << fFireAz << endl;
   }

   *gpInteract << "target type, range, vel, az errors: ";
   ((cAtWideArea*)fpAttEnt)->WriteTgtType(fTgtType, gpInteract);
   *gpInteract << " "
               << fRangeError << " "
               << fVelError << " "
               << fAzimuthError 
               << endl;
}

cWamInteraction::cWamInteraction(cMine* pMine, cLinkVehGeom* pLinkVeh) :
   cInteraction(pMine, pLinkVeh)
{
   fFireAz        = 99.9;
   fRangeError    = ((cLinkVehErr*)pLinkVeh)->GetRangeErr();
   fVelError      = ((cLinkVehErr*)pLinkVeh)->GetVelErr();
   fAzimuthError  = ((cLinkVehErr*)pLinkVeh)->GetAzErr();
   fTgtType       = ((cLinkVehErr*)pLinkVeh)->GetTgtType();
}

cWamInteraction::~cWamInteraction()
{
}

void   cDetonVuln::WriteDetonResult( ofstream* pFile, cDetonation* /*pDet*/, 
                                      float range, int result )
{
   *pFile << "  range: " 
          << range
          << " result: "
          << TVulnProb::GetVulnResultText(result)
          << " ";
//          << endl;

   fpMine->cMine::WriteEntSumm(pFile);
   *pFile << endl;
}
