/*
$RCSfile: AnalData.h,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:45 $
$Log: AnalData.h,v $
Revision 1.1.1.1  1999/02/16 14:54:45  jchappel
Comprehensive Mine Simulator

 * Revision 2.1  1998/12/08  20:13:26  jchappel
 * made adjustments for target Id
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.3  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.2  1997/10/02  17:47:47  austin
 * first working port to modsaf3.0
 *
Revision 1.2  1997/10/02 17:47:47  austin
first working port to modsaf3.0

 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.6  1996/08/09  14:33:27  rcarp
 * changed to support los check by terrain manager, and its use by offroute mines
 *
 * Revision 1.5  1996/07/03  22:55:15  rcarp
 * enhancements and bug fixes for version 1.5b9
 *
 * Revision 1.4  1996/06/19  19:05:45  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.3  1996/02/16  19:32:41  rcarp
 * revisions for 1.1b7 version
 *
 * Revision 1.3  1996/02/16  19:32:41  rcarp
 * revisions for 1.1b7 version
 *
 * Revision 1.2  1996/02/05  14:32:43  rcarp
 * Updates and bug fixes of release version 1.1b5
 *
$State: Exp $
*/
 
/*******************************************

   file: AnalData.h
   
      Copyright 1996 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   AnalData.h
   
      The header file for analysis data routines
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#ifndef _UAnalData_
#define _UAnalData_

#include "Globals.h"
//#include "Minefield.h"
//#include "Vehicle.h"

enum EInteractionResult {
   eTracking,      // to be determined
   eNoAttack,      // attack conditions not satisfied
   eMissedDetect,      // detection error, e.g. for smart mine
   eAttackedTarget,      // for mines that fire sublets
   eAttackedSelfDefense,      // for mines that fire sublets
   eHitTarget,
   eMissedTarget,
   eDecTgtCounter,      // decremented counter
   eDetonationFailed,
   eTargetDropout,      // e.g, deleted for some reason
   eMineNeutralized,
   eMineDetonated,      // by external attack
   eMineDisplaced,
   eMineRollback,
         // mostly for side attack mines
   eAttackGeomSat,
   eSlowTarget,
   eFastTarget,
   eTooFarWay,
   eTooClose,
   eNoIntervis,

   ePrimaryTarget,
   eMissedWindow,  // target skipped fire window
   
   eNumResultTypes
};

class cLinkVehGeom;

class cInteraction {
 protected:
   static  char*      fsResults[eNumResultTypes];
   
   virtual void       WriteInterSpecifics();
   
           void       WriteTimeSumm();
           void       WriteClosestSumm();
           void       WriteInterResult();

   cEntity*           fpAttEnt;
   cEntity*           fpTgtVeh;
   TTimeSim           fStartTime;
   TTimeSim           fDetTimes[10];  //Used for determining delay with command detonated mines
   TTimeSim           fIgnoreTimes[10];  //Used for determining delay with command detonated mines
   TTimeSim           fEndTime;
   TTimeSim           fClosestTime;
   float              fClosestRange;
   TEntCoord          fClosestPt;
   TLocation          fVehLocAtClosestPt;   // map coordinates
   float              fProjClosestOffset;
   EInteractionResult feResult;

 public:
   void               AddIgnoreTime(TTimeSim);  //Temporary measure.  Used for Cmd Det Mines
   void               AddDetTime(TTimeSim);

   EInteractionResult GetResult()   { return feResult; }
   
   virtual void       UpdateInteraction(cLinkVehGeom*);
   virtual void       WriteInteraction(ofstream*);
   
   virtual void       SetResult(EInteractionResult eResult);
   
              cInteraction(cEntity* pAttacker, cLinkVehGeom* pLinkVeh);
   virtual   ~cInteraction();
};

class cMine;

class cWamInteraction : public cInteraction {
 protected:
      // errors from link
   float         fRangeError;      // multiplier of true range
   float         fVelError;        // multiplier of true velocity
   float         fAzimuthError;    // added to true az
   int           fTgtType;         // might be wrong
      // firing parameter
   float         fFireAz;
   
   virtual void       WriteInterSpecifics();

 public:
   
   virtual void       SetResult(EInteractionResult eResult);
   
              cWamInteraction(cMine*, cLinkVehGeom*);
   virtual   ~cWamInteraction();
};

#endif
