/*
$RCSfile: ApMine.cc,v $
$Revision: 1.7 $
$Author: ksnively $
$Date: 1999/06/04 17:31:53 $
$Log: ApMine.cc,v $
Revision 1.7  1999/06/04 17:31:53  ksnively
Revision 4.2.  Compilies under multiple OS.  Single detonation for Cmd Det mines.

Revision 1.6  1999/05/18 19:33:13  ksnively
Ported CMS to compile under IRIX 6.5
Altered make files to allow cms to be compiled under various OS simulataneously
Removed the Startup GUI

Revision 1.5  1999/05/10 20:18:19  ksnively
Added a command line option for enabling Communication Capability.

Revision 1.4  1999/04/29 19:34:00  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.3  1999/04/26 16:58:40  ksnively
*** empty log message ***

Revision 1.2  1999/03/26 21:17:24  ksnively

Added radio communication capability for Command Detonated Mines.  Included sending
and receiving transmitter and signal PDUs.  All signals have the option of being
processed by the Communication Effects Federate.

Revision 1.1.1.1  1999/02/16 14:54:27  jchappel
Comprehensive Mine Simulator

// Revision 2.5  1998/12/08  19:47:48  jchappel
// Modified detonation PDU to include Target ID.
//
// Revision 2.4  1998/12/02  16:37:58  jchappel
// added gui detonator
//
// Revision 2.3  1998/11/25  18:37:51  ksnively
// *** empty log message ***
//
// Revision 2.2  1998/02/18  20:10:01  ksnively
// Ap mine replacement bug fixes.
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.6  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.5  1996/06/19  19:05:45  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.4  1996/05/08  00:46:50  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.4  1996/05/08  00:46:50  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.3  1996/02/16  21:41:55  rcarp
// revisions for getting mine wide area and directional range
//
// Revision 1.2  1996/02/05  14:31:34  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: ApMine.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   ApMine.cc
   
      The AP mine implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "ApMine.h"
#include "AtMine.h"
#include "AtWideArea.h"
#include "Display.h"
#include "NetMgr.h"
#include "Message.h"
#include "EntityDB.h"
#include "CmsControl.h"
#include "MapGeometry.h"
#include "AnalData.h"
#include "Communication.h"

#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>


// ************************************
//   messages
// ************************************

// ************************************
//   components
// ************************************

const TKeywordEntry TFuzeParamApCmdDet::fKeyApCdFuzeTable[eapcdfuzeSizeTable] = {
   "fuzeType",      ekeyQuotedString,
   "fireOffset",    ekey2DPoint,
   "tgtRange",      ekeyDouble,
   "tgtCrossRange", ekeyDouble,
   "rangeErr",      ekeyDouble,
   "azErr",         ekeyDouble
};

const TKeywordEntry TFuzeParamApTripCmdDet::fKeyApTripCdFuzeTable[eaptcdfuzeSizeTable] = {
   "fuzeType",      ekeyQuotedString,
   "numWires",      ekeyInt,
   "multiShot",     ekeyChar,      // if t or T, has shot for each wire
   "shotRange",     ekeyFloat,    // detonation range for each shot
   "wireLength",    ekeyFloat,
   "tripProb",      ekeyFloat,
   "armDuration",   ekeyFloat,   // time waiting for fire command
   "msgDelay",      ekey2DPoint,   // radio delay interval in notifying controller
   "cmdDelay",      ekey2DPoint,   // radio delay interval in getting fire command
   "tgtRange",      ekeyFloat,   // keeping track of tgts in range, may not be wire length
   "directional",   ekeyChar,      // if t or T, is directional and can be pointed
   "halfAng",       ekeyFloat    // half of trip wire open angle, in degrees on input
};

cTripWire::cTripWire(float prob, float az, float len,
                     TDetonationData mineDetData, cTripWire* pPrevWire) :
                     fDetonationData(mineDetData)
{
   fbTripped = false;
   fbFired   = false;
   fTimeTripped   = 0.0;
   fTimeout   = 0.0;
   fTripProb = prob;
   fAzimuth   = az;
   fLength   = len;
   if (pPrevWire)
      pPrevWire->fpNextWire = this;
   fpNextWire = NULL;
}

EBoolean   cTripWire::UpdateTripWire(T2Dpolar loc1, T2Dpolar loc2, EBoolean bCross, TTimeSim timeout)
   // returns true if tripped this update
{
   if ( !fbTripped ) {
      if ( bCross == true ) {   // tgt crosses neg x axis, azimuth jumps
         if ( ( (fAzimuth <= loc1.theta) && (fAzimuth <= loc2.theta) ) ||
              ( (fAzimuth > loc2.theta) && (fAzimuth > loc1.theta) ) )  {
            fTimeTripped = GetCurSimTime();
            fTimeout = timeout;
            fbTripped = true;
            return true;
         }
      }
      else {
         if ( ( (fAzimuth <= loc1.theta) && (fAzimuth > loc2.theta) ) ||
              ( (fAzimuth <= loc2.theta) && (fAzimuth > loc1.theta) ) )  {
            fTimeTripped = GetCurSimTime();
            fTimeout = timeout;
            fbTripped = true;
            return true;
         }
      }
   }
   return false;
}

cTripWire::~cTripWire()
{
}

TLinkDetData::~TLinkDetData()
{
}

EBoolean     TFuzeParamApCmdDet::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeyApCdFuzeTable, eapcdfuzeSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case eapcdfuzekwFuzeTypeName: {
                     // read name, look up index
   					fTypeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( (fTypeNameIndex != efuzetypeCommandFire) &&
                      (fTypeNameIndex != efuzetypeCommandLoc) &&
                      (fTypeNameIndex != efuzetypeCommandFireDir) ) {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable fuze type for command detonated mine"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case eapcdfuzekwFireOffset:
                  fMaxRange =  ((T2Dsingle*)returnVal)->x;
                  fCrossRange =  ((T2Dsingle*)returnVal)->y;
                  fFireOffset.x = fMaxRange;
                  fFireOffset.y = fCrossRange;
                  fTrackRange = fMaxRange;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eapcdfuzekwRange:
                  fMaxRange = *(double*)returnVal;
                  fFireOffset.x = fMaxRange;
                  fTrackRange = fMaxRange;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eapcdfuzekwCrossRange:
                  fCrossRange = *(double*)returnVal;
                  fFireOffset.y = fCrossRange;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eapcdfuzekwRangeErrMax:
                  fRangeErrMax = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eapcdfuzekwAzErrMax:
                  fAzimuthErrMax = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

               
				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TFuzeParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TFuzeParamApCmdDet::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

cControl*     TFuzeParamApCmdDet::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
      case efuzetypeCommandFire:
      case efuzetypeCommandFireDir:
            pNewControl = new cCmdFireControl(pMine, fFireOffset, fuzeOpt);
         break;
   
      case efuzetypeCommandLoc:
            pNewControl = new cCmdFireControl(pMine, fFireOffset, fuzeOpt);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TFuzeParamApCmdDet::CreateControl");
   }
   
   return pNewControl;
}

EBoolean   TFuzeParamApCmdDet::IsDirectional(TFlatEntCoord* pMinMax)
{
   if ( fTypeNameIndex == efuzetypeCommandFireDir) {
      if ( pMinMax ) {
         pMinMax->x = 0.0;
         pMinMax->y = fMaxRange;
      }
      
      return true;
   }
   
   return false;
}

EBoolean   TFuzeParamApCmdDet::IsWideArea(float* pRange)
{
   if ( fTypeNameIndex == efuzetypeCommandLoc) {
      if ( pRange )
         *pRange = fFireOffset.x;
      
      return true;
   }
   
   return false;
}

TFuzeParamApCmdDet::TFuzeParamApCmdDet(int type) :
   TFuzeParam(type)
{
   fFireOffset.x = fMaxRange    =  100.0;   // meters
   fFireOffset.y = fCrossRange  =  0.0;

   fTrackRange       = fMaxRange;

   fRangeErrMax       = 0.2;
   fAzimuthErrMax     = 0.1;
   
   fDetonResult      = edetDetonation;

//   fFireOffset.x = 200.0;
//   fFireOffset.y = 0.0;
//   fTrackRange       = 200.0;
}

TFuzeParam*   TFuzeParamApCmdDet::CloneFuzeParam()   
{
   TFuzeParam* pClone = new TFuzeParamApCmdDet(*this);
   pClone->fpOptSelector = fpOptSelector->CloneFuzeOptionSelector(pClone);
   
   return pClone;
}

TFuzeParamApCmdDet::~TFuzeParamApCmdDet()
{
}

/*
int    cLifeTracker::GetTrackType(const TEntityType& type)
   // tracks both land platforms and land life forms
{ 
   if ( ( (type.kind == ekindLifeform) || (type.kind == ekindPlatform) ) &&
          (type.domain == eplatdomLand)   )
      return cMine::etrkYes;
   else
      return cMine::etrkNot;
}

cLifeTracker::~cLifeTracker()
{ }
*/

//***************************
// Trip wire addtitions
// Dick Carpenter (IDA)
// Bill Davis (ARDEC)
//**************************

EBoolean     TFuzeParamApTripCmdDet::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeyApTripCdFuzeTable, eaptcdfuzeSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case eaptcdfuzekwFuzeTypeName: {
                     // read name, look up index
   					fTypeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( (fTypeNameIndex != efuzetypeTripCommandFire)  ) {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable fuze type for command detonated mine"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case eaptcdfuzekwNumWires:
                  fNumWires = *(int*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwMultiShot:
                  if ( (*(char*)returnVal == 't') ||
                       (*(char*)returnVal == 'T')  )
                     fbMultiShot = true;
                  else
                     fbMultiShot = false;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwShotRange:
                  fShotRange =  *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwWireLength:
                  fWireLength = *(float*)returnVal;
                  fTrackRange = fWireLength + 5.0;
                  fFireOffset.x     = fWireLength;
                  fTgtRange = fWireLength;
                   delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwTripProb:
                  fTripProb = *(float*)returnVal;
                   delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwArmDuration:
                  fArmDuration = *(float*)returnVal * kSecToSimTime;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwMsgDelay:
                  fMsgDelay.x = kSecToSimTime * ((T2Dsingle*)returnVal)->x;
                  fMsgDelay.y = kSecToSimTime * ((T2Dsingle*)returnVal)->y;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwCmdDelay:
                  fCmdDelay.x = kSecToSimTime * ((T2Dsingle*)returnVal)->x;
                  fCmdDelay.y = kSecToSimTime * ((T2Dsingle*)returnVal)->y;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwTgtRange:
                  fTgtRange = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwDirectional:
                  if ( (*(char*)returnVal == 't') ||
                       (*(char*)returnVal == 'T')  )
                     fbDirectional = true;
                  else
                     fbDirectional = false;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eaptcdfuzekwHalfAng:
                  fHalfAngle = *(float*)returnVal  * kDegToRad;
                  delete returnVal;
                  returnVal = NULL;
 					break;

               
				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TFuzeParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TFuzeParamApTripCmdDet::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

cControl*     TFuzeParamApTripCmdDet::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = new cTripCmdFireControl(pMine, fFireOffset, fuzeOpt, this);
      
   return pNewControl;
}

EBoolean   TFuzeParamApTripCmdDet::IsDirectional(TFlatEntCoord* pMinMax)
{
   
   return fbDirectional;
}

EBoolean   TFuzeParamApTripCmdDet::IsWideArea(float* pRange)
{
   *pRange = fFireOffset.x;
      
   if ( fFireOffset.x > 0.0  )      
      return true;
   
   return false;
}

TFuzeParamApTripCmdDet::TFuzeParamApTripCmdDet(int type) :
   TFuzeParam(type)
{
   fNumWires         = 1;
   fbMultiShot       = false;
   fShotRange        = 3.0;
   fWireLength       = 5.0;
   fFireOffset.x     = fWireLength;
   fFireOffset.y     =  0.0;  
   fTrackRange       = fWireLength + 5.0;
   fTripProb         = 1.0;
   fArmDuration      = 1000.0;
   fMsgDelay.x =  fMsgDelay.y    = 0.0;
   fCmdDelay.x =  fCmdDelay.y    = 0.0;
   fTgtRange         = fWireLength;
   fbDirectional     = false;
   fHalfAngle        = 90.0 * kDegToRad;

   
   fDetonResult      = edetDetonation;

}

TFuzeParam*   TFuzeParamApTripCmdDet::CloneFuzeParam()   
{
   TFuzeParam* pClone = new TFuzeParamApTripCmdDet(*this);
   pClone->fpOptSelector = fpOptSelector->CloneFuzeOptionSelector(pClone);
   
   return pClone;
}

TFuzeParamApTripCmdDet::~TFuzeParamApTripCmdDet()
{
}

cLinkVehGeom*   cTripWireTracker::CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev)
{
   cLinkVehCross* pNewLink = new cLinkVehCross((cLinkVehCross*)pOldTrkLink, pPrev);

   return pNewLink;
}

cLinkVehGeom*   cTripWireTracker::CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev)
{
   cLinkVehCross* pNewLink = new cLinkVehCross(pProxLink, pPrev, fpRefEnt, false);


   return pNewLink;
}

cTripWireTracker::~cTripWireTracker()
{}


// ************************************
//   mines 
// ************************************


void cCmdFireControl::WriteInteraction (int event)
{
   char* result;
   if (gfInteract) {
      fReactTime = GetCurSimTime();
      if (event == cApCommandDet::eopIgnore)
	 result = "Ignore";
      else if (event == cApCommandDet::eopDetonate)
	 result = "Detonate";
      else 
	 result = "Unknown";

      *gpInteract << "{ Command Interaction" << endl;
      *gpInteract << "Trip and React Time: " << fTripTime * kSimTimeToSec << " " 
		  << fReactTime * kSimTimeToSec << endl;
      fpMine->cMine::WriteEntSumm(gpInteract);
      *gpInteract << "\nResult: " << result << " }" << endl << endl;
   }
}

void cCmdFireControl::CbDetonate (Widget w, XtPointer pObj, 
				      XmAnyCallbackStruct* )
{
  cSimMine* fp = ((cCmdFireControl*)pObj)->fpMine;
  ((cCmdFireControl*)pObj)->fbTgtPending = false;
  fp->CmdDet();
  XtDestroyWidget (w);
}

void cCmdFireControl::CbIgnore (Widget w, XtPointer pObj, 
				      XmAnyCallbackStruct* )
{
  /*  gpGuiWin->GetCurMinefield()->fbRepos = false;
  new cReadMinefieldMsg (gpGuiWin->GetCurMinefield(), gMfieldMgr.fpInFile);*/
  cCmdFireControl* pControl = (cCmdFireControl*)pObj;
  pControl->fbTgtPending = false;
  ((cApCommandDet*)(pControl->fpMine))->feOpEvent = cApCommandDet::eopIgnore;
  pControl->WriteInteraction(((cApCommandDet*)(pControl->fpMine))->feOpEvent);
  XtDestroyWidget (w);
}

void      cCmdFireControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* /*pEmissList*/, 
                                          TTimeSim*  /* pWaitTime */)
      // just sets state on detection/ no detection
{
   EMineInternalState state = fpMine->GetInternalState();
   char         buf[199];
   cEntity*     pController;
 
   
   if ( state >= emisOn ) {
      EMineInternalState newstate = emisOn;
      
         // go through each tgt in track list.  Determine tracking state
      for (T2link* pLn  = pVehList->GetFirstLink();
                   pLn != pVehList->GetNULLLink();
                   pLn  = pVehList->GetNextLink() ) {
        if ( ((cLinkVehGeom*)pLn)->GetHorzRange() <  fFireOffset.x  ) {
            newstate = emisTracking;
	    break;
	}
      }

      // Go through each target and determine if any have tripped the mine.  If so, notify operator.
      for (        pLn  = pVehList->GetFirstLink();
                   pLn != pVehList->GetNULLLink();
                   pLn  = pVehList->GetNextLink() ) {
	 if (((cLinkVehGeom*)pLn)->GetHorzRange() <  fFireOffset.x  && InFootprint ((cLinkVehGeom*)pLn)) {
	    // Is this the first time this target has tripped this mine and the mine has no peding target
	    if ( fpMine->IsCmdDet() && !fbTgtPending && (((cLinkVehGeom*)pLn)->GetDetectionResult() == edrNotEngaged)) {
	       fbTgtPending = true;  //We have a pending target
	       fTripTime = GetCurSimTime();
	       //Set up pop up GUI
	       //sprintf (buf, "Do you want to detonate the mine\nat coordinates: X Y\nor do you want ignore?");
		     
	       /*
		 gpGuiWin->PopupMessage (gpGuiWin->GetMainWin(), 
				       cGuiWindow::eDialogQuestion,
				       "Ignore",
				       "Detonate",
				       (XtCallbackProc)CbIgnore,
				       (XtCallbackProc)CbDetonate,( XtPointer)this,
				       NULL, buf);
	    */
	       pController =((cSimMine*)fpMine)->GetController(); 
	       if (((cSimMine*)fpMine)->GetC2Interface()) {
		  fpMine->GetC2Interface()->SendMessage(cMineC2Interface::emtActivated, 
							&pController, 1);
		  cout << "Sending Mine Detonation Message for " << fpMine->GetEntityID() << endl;
	       }
	       else {
		  cout << "ERROR: "<< fpMine->GetEntityID() << " is unable to find C2 Interface!" << endl;
	       }
		     
	    }
	    // Mark this guy as already considered
	    ((cLinkVehGeom*)pLn)->SetDetectionResult(edrInFootprint);
	 }
	 else
	    // This guy has gone outside footprint.  Mark as not engaged
	    ((cLinkVehGeom*)pLn)->SetDetectionResult(edrNotEngaged);
      }

      if ( state != newstate )
          fpMine->StateChange( newstate );

   }
   
   if ( gfInteract ){
      fpMine->UpdateInteractions();
 
   }
}

cCmdFireControl::~cCmdFireControl()
{
}

void      cTripCmdFireControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* /*pEmissList*/, 
                                          TTimeSim* pWaitTime )
      // sends msg to GUI/minefield controller on detection
{
   EMineInternalState state = fpMine->GetInternalState();
   
   if ( state >= emisOn ) {
      EMineInternalState newstate = emisOn;
      T2link* pLn  = pVehList->GetFirstLink();
      if ( pLn != pVehList->GetNULLLink() ) {
         newstate = emisTracking;
         *pWaitTime = fpMine->GetFastUpdateRate();
         TFuzeParamApTripCmdDet* pFP = (TFuzeParamApTripCmdDet*)(fpMine->GetFuzeParams());
         TTimeSim fireTimeout = GetCurSimTime() + pFP->fArmDuration;
         
         // go through each tgt in track list
         for ( pLn  = pVehList->GetFirstLink();
               pLn != pVehList->GetNULLLink();
               pLn  = pVehList->GetNextLink() ) {
//            cout << "target rng " << ((cLinkVehCross*)pLn)->GetHorzRange() ;
            if ( ((cLinkVehCross*)pLn)->GetHorzRange() <  fFireOffset.x  ) {
               // tgt is within wire radius, pass loc to trip wires
               T2Dpolar loc1 = ((cLinkVehCross*)pLn)->GetLocPolar();
               T2Dpolar loc2 = ((cLinkVehCross*)pLn)->GetLastLocPolar();
               
               // tell wires whether tgt crossed the neg x axis
               EBoolean bCross = ( fabs(loc1.theta - loc2.theta) > kPi) ? true: false;
//               cout << " in rng, azs " << loc1.theta << " " << loc2.theta 
//                    << " cross? " << bCross;
               for ( cTripWire* pW = fpTripWireList;
                                pW != NULL;
                                pW = pW->fpNextWire )
                  if ( pW->UpdateTripWire(loc1, loc2, bCross, fireTimeout) ) {
                     fFireTimeout = fireTimeout;
                     ((cApTripCommandDet*)fpMine)->TripWireDetect(pW, fFireTimeout);
                  }
            }
         }
      }
   
      if ( state != newstate )
          fpMine->StateChange( newstate );
   }
   
   if ( gfInteract )
      fpMine->UpdateInteractions();
}

void    cTripCmdFireControl::ProcessFireMsg(EMfCmdMode eCurMode)
   // the command normally comes from the operator, but in autofire it is 
   // internally generated.
   // The mode is that of the minefield. It is allowed for the minefield
   // to have multiiple mine types, so we should check that the mode is allowed
   // for this mine type (specifically, that a mode designed to accomodate
   // multifire mines does not affect the behavior of single fire mines).
{
   TTimeSim curTime = GetCurSimTime();
   if ( curTime <= fFireTimeout ) {   // we will fire
      if ( fbMultiShot == true ) {
         for ( cTripWire* pW = fpTripWireList;
                          pW != NULL;
                          pW = pW->fpNextWire )
            if ( (pW->IsFired() == false) &&
                 ( ( curTime <= pW->GetTimeout() ) ||
                   ( eCurMode == eCmdModeCmdFireAll ) ||
                   ( eCurMode == eCmdModeTripFireAll ) ) ) {
               ((cApTripCommandDet*)fpMine)->Detonate( pW );
               
               pW->SetFired();
            }
      }
      else {   // single shot, not multishot
         ((cApTripCommandDet*)fpMine)->Detonate( NULL );
      }
      fFireTimeout = curTime;
   }
}

cTripCmdFireControl::cTripCmdFireControl(cSimMine* pMine, TFlatEntCoord fireOffset, 
         const TFuzeOpt& fuzeOpt, TFuzeParamApTripCmdDet* pFuzeParam) : 
          cMineControl(pMine, fireOffset, fuzeOpt) 
{
   double kTwoPi = 2.0 * kPi;
      // directional mines open in mine direction, others have randon wire directions
   float az, azStep;
   if ( fpMine->IsDirectional() ) {
      az = fpMine->GetTccOrient().psi;
      az += pFuzeParam->fHalfAngle;
      if ( az > kPi )
         az -= kTwoPi;
      azStep = (kTwoPi - 2.0 * pFuzeParam->fHalfAngle) / (pFuzeParam->fNumWires - 1);
   }
   else {
      az =  kTwoPi * ( UniformRand() - 0.5 );   // between -pi and pi
      azStep = kTwoPi / pFuzeParam->fNumWires;
   }
   
   fFireOffset.x = pFuzeParam->fWireLength;
   cTripWire* pWire = fpTripWireList = new cTripWire(pFuzeParam->fTripProb, 
                                    az, fFireOffset.x, pMine->GetDetonationData(), NULL);
   for ( int i = 1; i < pFuzeParam->fNumWires; i++ ) {
      az += azStep;
      if ( az > kPi )
         az -= kTwoPi;
      new cTripWire(pFuzeParam->fTripProb, az, fFireOffset.x, pMine->GetDetonationData(), pWire);
      pWire = pWire->fpNextWire;
   }
   fbMultiShot = pFuzeParam->fbMultiShot;
   fFireTimeout = 0.0;
}

cTripCmdFireControl::~cTripCmdFireControl()
{
}

cApMine::cApMine(cMineComp* pParent, float radius) :
      cSimMine(pParent, radius)
{
}

cApMine::cApMine( cPlanMine* pPlan ) :
      cSimMine(pPlan)
{
}

cApMine::~cApMine()
{ }

/*
cApConv::cApConv(cMineComp* pParent, float radius) :
      cApMine(pParent, radius)
{
}
*/

cApConv::cApConv( cPlanMine* pPlan ) :
      cApMine( pPlan )
{
}

cApConv::~cApConv()
{ }

/*
cApConvBding::cApConvBding(cMineComp* pParent, float radius) :
      cApMine(pParent, radius)
{
}
*/

cApConvBding::cApConvBding( cPlanMine* pPlan ) :
      cApMine( pPlan )
{
}

cApConvBding::~cApConvBding()
{ }

cApScatterable::cApScatterable(cMineComp* pParent, float radius) :
      cApMine(pParent, radius)
{
}

cApScatterable::cApScatterable( cPlanMine* pPlan ) :
      cApMine( pPlan )
{
}

cApScatterable::~cApScatterable()
{ }

cApScatBding::cApScatBding(cMineComp* pParent, float radius) :
      cApMine(pParent, radius)
{
}

cApScatBding::cApScatBding( cPlanMine* pPlan ) :
      cApMine( pPlan )
{
}

cApScatBding::~cApScatBding()
{ }

EBoolean   cCmdFireControl::InFootprint(cLinkVehGeom* pTgtLink)
{
   //char         buf[199];
 
   if ( pTgtLink->GetHorzRange() <= fFireOffset.x ) {
      if ( (fAzimuth > 0.0) &&
        ( fabs(pTgtLink->GetAzimuth()) > fAzimuth ) ) {
//         cout << "az check for footprint" << endl;
         return false;
      }
      else
         return true;
   }
   else
      return false;
}

cCmdFireControl::cCmdFireControl(cSimMine* pMine, TFlatEntCoord fireOffset, const TFuzeOpt& fuzeOpt) : 
          cMineControl(pMine, fireOffset, fuzeOpt) 
{ 
   fAzimuth = atan2( fireOffset.y/2.0, fireOffset.x ); 
   fbTgtPending = false;
}

void     cApCommandDet::CmdDet()
     // fires mine
{
   if ( (fCmdType != ecmdFire) &&
        (fCmdType != ecmdFireDir) )
      return;
   // cLinkVehGeom* pLn = NULL;    
         // detonate at the mine
    if ( feIntState >= emisOff ) {
       SetDetonationResults((cLinkVehGeom *)NULL);   // set target in det PDU  fDetonationData.fTargetEnt)

       new cMineDetonationMsg(fDetonationData);
       gCmsMgr.GetNextEventId(&fDetonationData.fEventId);
       fFireData.fEventId = fDetonationData.fEventId;

       new cFireProjectileMsg(fFireData);   // NULL for no proj entity state

       feOpEvent = eopDetonate;
       ((cCmdFireControl*)fpControl)->WriteInteraction(feOpEvent);
         //** affects entities in range
       CheckTgts( (fCmdType == ecmdFireDir) ? true : false , (fCmdType == ecmdFireDir) ? true : false );
       //CheckTgts( true  );
   }
    if ( gfInteract )
       UpdateInteractions();
}
/*          quickfix to affect entities in range, detonate at each one
      if (fCmdType == ecmdFireDir) {
         TFlatEntCoord& fireOffset = fpControl->GetFireOffset();
         float az = atan2( fireOffset.y/2.0, fireOffset.x );
//         cout << "range, az " << fireOffset.x << " " << az << endl;
         for (cLinkVehGeom* pLn  = (cLinkVehGeom*)(fTrackVeh.GetFirstLink());
                            pLn != (cLinkVehGeom*)(fTrackVeh.GetNULLLink());
                            pLn  = (cLinkVehGeom*)(fTrackVeh.GetNextLink()) ) {
            if ( ( pLn->GetHorzRange() < fireOffset.x )  &&
                 ( fabs(pLn->GetAzimuth()) <= az )  ) {
             cout << "veh, rng and az " << pLn->GetHorzRange() << " " << pLn->GetAzimuth() << endl;
               TDetonationData* pDetData = new TDetonationData(this);
            
               pDetData->fTargetEnt =  pLn->GetVehicle()->GetEntityID();
   //            pDetData->fDetLocation =  pLn->GetVehicle()->GetGeoLoc();
   //            pDetData->fFireRange = pLn->GetHorzRange();
               pLn->GetLocTgtRef( & pDetData->fLocTargetRef );
               pDetData->feDetResult = edetEntityProximate;
   
               ((cMineComp*)fpMfComponent)->IncrementVehHit();
      
               new cMineDetonationMsg(*pDetData);
      
               TLinkDetData* pDetLink = new TLinkDetData(pDetData, &fHitTgtDetList);
            }
         }
      }
   }
}
*/


void    cApCommandDet::CheckTgts(EBoolean eUseVehLoc, EBoolean eSendDetPdu)
{
   for (cLinkVehGeom* pLn  = (cLinkVehGeom*)(fTrackVeh.GetFirstLink());
                      pLn != (cLinkVehGeom*)(fTrackVeh.GetNULLLink());
                      pLn  = (cLinkVehGeom*)(fTrackVeh.GetNextLink()) ) {
      if ( ((cCmdFireControl*)fpControl)->InFootprint(pLn)  ) {
         TDetonationData* pDetData = new TDetonationData(this);
	 TFireData* pFireData = new TFireData(this);
     
	 gCmsMgr.GetNextEventId(&(pDetData->fEventId));
	 pFireData->fEventId = pDetData->fEventId;

         pDetData->fTargetEnt =  pFireData->fTargetEnt = pLn->GetVehicle()->GetEntityID();

	 if (eUseVehLoc) {
	    pDetData->fDetLocation =  pLn->GetVehicle()->GetGeoLoc();
	    pDetData->fLocTargetRef = 0;
	 }
	 else 
	    pLn->GetLocTgtRef( & pDetData->fLocTargetRef );


	 pFireData->fFireRange = pLn->GetHorzRange();
	 // Make sure target assesses direct fire damage
         //pDetData->fLocTargetRef = TEntCoord();
//               pDetData->feDetResult = edetEntityProximate;
         pDetData->feDetResult = GetFuzeParams()->fDetonResult;
         pLn->SetInteractResult(eHitTarget);

         ((cMineComp*)fpMfComponent)->IncrementVehHit();

         TLinkDetData* pDetLink = new TLinkDetData(pDetData, pLn->GetVehicle(), &fHitTgtDetList);

	 /*
         if ( eSendDetPdu ) {
	    new cFireProjectileMsg(*pFireData);   // NULL for no proj entity state
            new cMineDetonationMsg(*pDetData);
            new cCheckVehDamageMsg(this, fsDamageCheckDelay, pLn->GetVehicle());
         }*/

	 if (eSendDetPdu) {
	    new cFireProjectileMsg(*pFireData);   // NULL for no proj entity state
	    new cMineDetonationMsg(*pDetData);
	    new cCheckVehDamageMsg(this, fsDamageCheckDelay, pLn->GetVehicle());
	 }
      }
   }
}


void     cApCommandDet::CmdDet(TLocation tccPt)
     // fires mine
{
   if ( fCmdType != ecmdLoc )
      return;
      
   if ( feIntState >= emisOff ) {
      SetDetonationResults((cLinkVehGeom *)NULL);   // no target in det PDU
            // reset det location to impact point
      gpTerrainMgr->MapToSimLoc(tccPt, &fDetonationData.fDetLocation);

//      fFireData.fFireRange = fpControl->GetFireOffset().x;
      fFireData.fFireRange = fTccLocation.Distance(tccPt);

      //SetChangeIndicator();
      gCmsMgr.GetNextEventId(&fDetonationData.fEventId);
      fFireData.fEventId = fDetonationData.fEventId;

      new cFireProjectileMsg(fFireData);
      new cMineDetonationMsg(fDetonationData);
      
      CheckTgts(false, false);
   
      if ( gfInteract )
         UpdateInteractions();
   }
}

void    cApCommandDet::SelectObj()
{
   if ( ! fbSelected ) {
      fbSelected = true;
      if ( fpDispObj) {
         ((cMineDisp*)fpDispObj)->SetFireActiveMine();
         fpDispObj->RedrawSelected ();
      }
   }
}

void    cApCommandDet::DeSelectObj()
{
   if ( fbSelected ) {
      fbSelected = false;
      if ( fpDispObj) {
         if ( fCmdType == ecmdLoc )
            ((cMineDisp*)fpDispObj)->ClearFireActive();
         fpDispObj->RedrawSelected ();
      }
   }
}


void    cApCommandDet::CheckVehDamage(cVehicle* pVeh)
{
   if ( pVeh ) {
      if ( ! pVeh->IsMobile() ) {
         fKilledTgts++;
         ((cMineComp*)fpMfComponent)->IncrementVehKill();
      }
   }
}

void   cApCommandDet::UpdateInteractions()
{
         // for output, update each link
   if ( gfInteract ) {
      cInteraction* pInter;
      TTimeSim now = GetCurSimTime();

      for (cLinkVehGeom* pLn  = (cLinkVehGeom*)(fTrackVeh.GetFirstLink());
                         pLn != (cLinkVehGeom*)(fTrackVeh.GetNULLLink());
                         pLn  = (cLinkVehGeom*)(fTrackVeh.GetNextLink()) ) {
         pInter = pLn->GetInteraction();
         if ( pInter ) {      // there is an interaction record
	    if (feOpEvent == eopIgnore)
	       pInter->AddIgnoreTime(now);
	    else if (feOpEvent == eopDetonate)
	       pInter->AddDetTime(now);
	       
            pInter->UpdateInteraction(pLn);
            if ( feIntState < emisLowestUpdate) {
                  // detonated, set result if not already done
               if ( pInter->GetResult() == eTracking) {
                  pInter->SetResult(eMineDetonated);
                  if ( pLn->GetVehicle() == fpTarget ) {
                     if (feIntState == emisDetonatedHit)
                        pInter->SetResult(eHitTarget);
                     else if (feIntState == emisDetonatedMissed)
                        pInter->SetResult(eMissedTarget);
                     else if (feIntState == emisDetonated)
                        pInter->SetResult(eAttackedTarget);
                  }
               }
            }
            else {      // not detonated - keep tracking or drop
               if ( ! ((cCmdFireControl*)fpControl)->InFootprint(pLn)  ) {
                  if ( pInter->GetResult() == eTracking) {
                     if ( feIntState == emisFailed )
                        pInter->SetResult(eMineNeutralized);
                     else
                        pInter->SetResult(eNoAttack);
                  }
                  pLn->DeleteInteraction();
	       }
            }
         }
         else {            // no interaction record for this link, may need one
            if ( feIntState >= emisLowestUpdate) {
               if ( ((cCmdFireControl*)fpControl)->InFootprint(pLn)  ) {
                  pInter = CreateNewInteraction(pLn);
                  pLn->SetInteraction( pInter );
                  pInter->UpdateInteraction(pLn);
                  pInter->SetResult(eTracking);
		  if (feOpEvent == eopIgnore)
		     pInter->AddIgnoreTime(now);
		  else if (feOpEvent == eopDetonate)
		     pInter->AddDetTime(now);

	       }
            }
         }
      }
   }
   feOpEvent = eopNone;
}

cApCommandDet::cApCommandDet( cPlanMine* pPlan ) :
      cApMine( pPlan )
{
   switch (fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)->fTypeNameIndex) {
      case TFuzeParam::efuzetypeCommandFire:
            fCmdType = ecmdFire;
         break;
   
      case TFuzeParam::efuzetypeCommandLoc:
            fCmdType = ecmdLoc;
         break;
   
      case TFuzeParam::efuzetypeCommandFireDir:
            fCmdType = ecmdFireDir;
         break;
   }
   feOpEvent = eopNone;

   if (GetCompParent()) {
      if (gCmsMgr.fStartupParam.fbController) {
	 fpRadio = new cRadio ((cEntity*)this);
	 fpC2Interface = new cMineC2Interface ((cEntity*)this);
	 fpController = gVehMgr.FindVehicle(GetCompParent()->GetControllerId(), &gVehMgr.GetVehList());
	 
	 if (fpController == NULL) {
	    cout << "Error: cApCommandDet cannot find controller " << GetCompParent()->GetControllerId() << endl;
	    
	    if (fpRadio) {
	       delete fpRadio;
	       fpRadio = NULL;
	    }
	    if (fpC2Interface) {
	       delete fpC2Interface;
	       fpC2Interface = NULL;
	    }
	 }
      }
   }
}

cApCommandDet::~cApCommandDet()
{ 
   if (fpRadio)
      delete fpRadio;
   if (fpC2Interface)
      delete fpC2Interface;
}
void     cApTripCommandDet::CmdFire(EMfCmdMode eMfCmdMode)
     // fire button was pressed, or in autofire the cmd is internally generated
{
   if ( eMfCmdMode == eCmdModeAutoFire )
      fAutoFireTimeout = GetCurSimTime() + ((TFuzeParamApTripCmdDet*)GetFuzeParams())->fArmDuration;
   ((cTripCmdFireControl*)fpControl)->ProcessFireMsg(eMfCmdMode);
}

void     cApTripCommandDet::Detonate( cTripWire* pWire )
     // controller tells us to detonate
{
   SetDetonationResults( (cLinkVehGeom*)NULL );   // includes setting state

   if ( fCmdType == ecmdSingleShot ) {
      new cMineDetonationMsg( GetDetonationData() );
   }
   else {
         // fill the tripwire's detonation data, using the mine data just set above
      pWire->GetDetonationData() = GetDetonationData();
      
         // figure the explostion location, from az and range from mine's map location
      TFuzeParamApTripCmdDet* pFP = (TFuzeParamApTripCmdDet*) GetFuzeParams();
               
      float range = pFP->fShotRange;
      float az = pWire->GetAz();
      
      TLocation locLoc( range * cos(az), range * sin(az), 2.0);
      
      cout<< "directional fire: az= " << az
          << " range " << range
          << " locLoc " << locLoc.x << " " << locLoc.y << " " << locLoc.z
          << endl;

      locLoc += fTccLocation;
      gpTerrainMgr->MapToSimLoc( locLoc, &(pWire->GetDetonationData().fDetLocation));
      
      cout<< "tcc and geoloc "
          << locLoc.x << " " << locLoc.y << " " << locLoc.z << " "
          << pWire->GetDetonationData().fDetLocation.x << " " 
          << pWire->GetDetonationData().fDetLocation.y << " " 
          << pWire->GetDetonationData().fDetLocation.z
          << endl;
      
      new cMineDetonationMsg( pWire->GetDetonationData() );
   }
}

void    cApTripCommandDet::TripWireDetect( cTripWire* pW, TTimeSim timeout)
    // the reaction to a wire tripping depends on the minefield's commmand mode
{
   TFuzeParamApTripCmdDet* pFP = (TFuzeParamApTripCmdDet*)GetFuzeParams();
   TTimeSim delayOut = UniformRand() * (pFP->fMsgDelay.y - pFP->fMsgDelay.x) + pFP->fMsgDelay.x;
   TTimeSim delayIn = UniformRand() * (pFP->fCmdDelay.y - pFP->fCmdDelay.x) + pFP->fCmdDelay.x;
   EMfCmdMode em = fpMinefield->GetCmdMode();

   switch ( em ) {
      case eCmdModeFire:   //normal, send msg to operator
      case eCmdModeCmdFireAll:   // will fire all shots, on operator cmd
            fpMinefield->CmdMineDetect(delayOut, delayIn, timeout);
            
         break;
   
      case eCmdModeTripped: // we have been directed to fire without operator
      case eCmdModeTripFireAll:   // fire all shots ourselves when 1 wire is tripped
            CmdFire(em);
         break;
   
      case eCmdModeAutoFire:   // after first fire cmd, fire ourselves on subsequent trips
            if ( GetCurSimTime() <= fAutoFireTimeout )
               CmdFire(em);
            else
               fpMinefield->CmdMineDetect(delayOut, delayIn, timeout);
         break;
   }
}

cApTripCommandDet::cApTripCommandDet( cPlanMine* pPlan ) :
      cApMine( pPlan )
{
   fAutoFireTimeout = 0.0;
   if ( ((TFuzeParamApTripCmdDet*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fbMultiShot ) {
      fCmdType = ecmdMultiShot;
      fRoundsLeft = ((TFuzeParamApTripCmdDet*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fNumWires;
      fpMinefield->SetMultishotCmdDet(true);
   }
   else
      fCmdType = ecmdSingleShot;
}

cApTripCommandDet::~cApTripCommandDet()
{ }
