/*
$RCSfile: CoordList.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:30 $
$Log: CoordList.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:30  jchappel
Comprehensive Mine Simulator

// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.9  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.8  1996/05/08  00:47:57  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.8  1996/05/08  00:47:57  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.7  1996/04/04  16:25:34  mkappel
// Improve performance with large mine sets.
//
// Revision 1.6  1996/03/28  19:16:02  mkappel
// Create 'On' label for mine toggles.
//
// Revision 1.5  1996/03/23  14:26:12  rcarp
// improvements to test mode and selection
//
// Revision 1.4  1996/03/13  00:25:53  mkappel
// Elim toggles in popup.
//
// Revision 1.3  1996/03/13  00:12:33  mkappel
// Create scrolling set of toggles for mines.
//
// Revision 1.2  1996/02/16  20:33:22  rcarp
// revisions associated with 1.1b7
//
$State: Exp $
*/
 
/*******************************************

   file: CoordList.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   CoordList.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "CoordEntry.h"
#include "Minefield.h"
#include "MapGeometry.h"

#include <Xm/BulletinB.h>
#include <Xm/DialogS.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/LabelG.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>


cCoordList::cCoordList (Widget wParent, ECoordListType type)
{
   Widget        wLabel;
   Dimension     height1, height2;

   fType = type;

   fwCoordForm =
     XtVaCreateManagedWidget ("form",
			      xmFormWidgetClass, wParent,
			      XmNfractionBase, 20,
			      NULL);

   fwNumEntries = 
     XtVaCreateManagedWidget (type == ecltMine ? "0 Mines" : "0 Vertices", 
			      xmLabelGadgetClass, fwCoordForm,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 1,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 19,
			      NULL);

   wLabel = 
     XtVaCreateManagedWidget ("X", 
			      xmLabelGadgetClass, fwCoordForm,
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, fwNumEntries,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 5,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 9,
			      XmNalignment, XmALIGNMENT_CENTER,
			      NULL);

   wLabel = 
     XtVaCreateManagedWidget ("Y", 
			      xmLabelGadgetClass, fwCoordForm,
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, fwNumEntries,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 15,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 19,
			      XmNalignment, XmALIGNMENT_CENTER,
			      NULL);

   fwCoordList = 
     XmCreateScrolledList (fwCoordForm, "list", NULL, 0);
   XtVaSetValues (fwCoordList, 
		  XmNvisibleItemCount, 8,
		  XmNselectionPolicy, XmBROWSE_SELECT,
		  NULL);
   XtVaSetValues (XtParent(fwCoordList),
		  XmNtopAttachment, XmATTACH_WIDGET,
		  XmNtopWidget, wLabel,
		  XmNrightAttachment, XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNwidth, 160,
		  NULL);

   if (type == ecltMine) {
     XtVaSetValues (fwCoordList, XmNvisibleItemCount, 3, NULL);
     XtVaGetValues (fwCoordList, XmNheight, &height1, NULL);
     XtVaSetValues (fwCoordList, XmNvisibleItemCount, 4, NULL);
     XtManageChild (fwCoordList);
     XtVaGetValues (fwCoordList, XmNheight, &height2, NULL);
     fListLineSize = height2 - height1;

     XtVaGetValues (XtParent(fwCoordList), XmNverticalScrollBar, &fwScrollbar, NULL);
     XtAddCallback(fwScrollbar, XmNdecrementCallback    , (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNdragCallback         , (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNincrementCallback    , (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNpageDecrementCallback, (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNpageIncrementCallback, (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNtoBottomCallback     , (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNtoTopCallback        , (XtCallbackProc)CbScroll, this);
     XtAddCallback(fwScrollbar, XmNvalueChangedCallback , (XtCallbackProc)CbScroll, this);

     wLabel = 
       XtVaCreateManagedWidget ("On", 
				xmLabelGadgetClass, fwCoordForm,
				XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			 XmNtopWidget, wLabel,
			 XmNleftAttachment, XmATTACH_FORM,
			 NULL);

     fwToggle[0] = 
       XtVaCreateWidget ("",
			 xmToggleButtonWidgetClass, fwCoordForm,
			 XmNheight, fListLineSize,
			 XmNtopAttachment, XmATTACH_WIDGET,
			 XmNtopWidget, wLabel,
			 XmNleftAttachment, XmATTACH_FORM,
			 NULL);
     XtVaSetValues (XtParent(fwCoordList),
		    XmNleftAttachment, XmATTACH_WIDGET,
		    XmNleftWidget, fwToggle[0],
		    NULL);
     fMaxToggles = 1;
     XtAddEventHandler (fwCoordForm, StructureNotifyMask, FALSE,
			(XtEventHandler)CbResize, (XtPointer)this);
   }
   else
     XtVaSetValues (XtParent(fwCoordList),
		    XmNleftAttachment, XmATTACH_FORM,
		    NULL);

   XtAddCallback (fwCoordList, XmNbrowseSelectionCallback,
		  (XtCallbackProc)CbSelect, this);
   XtManageChild (fwCoordList);
}

cCoordList::~cCoordList ()
{
}

void cCoordList::UpdateNumEntries ()
{
   int         nentries;
   char        buf[20];
   XmString    str1;

   if (fType == ecltPerimVertex || fType == ecltRowVertex) {
     if (fType == ecltPerimVertex)
       nentries = ((cPerimeter*)(gpGuiWin->GetCurComponent()))->GetNumPoints();
     else
       nentries = ((cRowComp*)(gpGuiWin->GetCurComponent()))->GetNumPoints();

     if (nentries == 1)
       str1 = XmStringCreateSimple ("1 Vertex");
     else {
       sprintf (buf, "%d Vertices", nentries);
       str1 = XmStringCreateSimple (buf);
     }
   }
   else {
     nentries = ((cMineComp*)(gpGuiWin->GetCurComponent()))->GetNumMines();
     if (nentries == 1)
       str1 = XmStringCreateSimple ("1 Mine");
     else {
       sprintf (buf, "%d Mines", nentries);
       str1 = XmStringCreateSimple (buf);
     }

     XEvent event;
     event.type = ConfigureNotify;
     CbResize (NULL, this, &event);
   }

   XtVaSetValues (fwNumEntries,
		  XmNlabelString, str1,
		  NULL);
   XmStringFree (str1);
}

void cCoordList::AppendCoordEntry ()
   /* add a blank line for new entry */
{
   XmString  xmstr;

   xmstr = XmStringCreateSimple ("");
   XmListAddItemUnselected (fwCoordList, xmstr, 0);
   XmStringFree (xmstr);
   XtVaSetValues (XtParent(fwCoordList),
		  XmNwidth, 160,
		  NULL);
}

void cCoordList::AppendCoordEntry (TLocation tccPt)
{
   char      buf[25];
   XmString  xmstr;
   int       pos = 0;

   /* place before blank entry if present */

   xmstr = XmStringCreateSimple ("");
   pos = XmListItemPos (fwCoordList, xmstr);
   XmStringFree (xmstr);
   if (pos > 0) XmListDeselectPos (fwCoordList, pos);

   sprintf (buf, "%9.1lf%9.1lf", tccPt.x, tccPt.y);
   xmstr = XmStringCreate (buf, "list_charset");
   XmListAddItem (fwCoordList, xmstr, pos);
   XmStringFree (xmstr);

   UpdateNumEntries();
   XtVaSetValues (XtParent(fwCoordList),
		  XmNwidth, 160,
		  NULL);
}

void cCoordList::AppendCoordEntries (TLocation* tccPts, int nEntries)
{
   XmString  xmstr;
   XmString* xmstr_arr = new XmString [ nEntries ];
   char      buf[25];
   int       pos, i;

   XtVaSetValues (XtParent(fwCoordList),
		  XmNwidth, 160,
		  NULL);

   /* place before blank entry if present */

   xmstr = XmStringCreateSimple ("");
   pos = XmListItemPos (fwCoordList, xmstr);
   XmStringFree (xmstr);
   if (pos > 0) XmListDeselectPos (fwCoordList, pos);

   for (i = 0; i < nEntries; i++) {
     sprintf (buf, "%9.1lf%9.1lf", tccPts[i].x, tccPts[i].y);
     xmstr_arr[i] = XmStringCreate (buf, "list_charset");
   }     
   XmListAddItems (fwCoordList, xmstr_arr, nEntries, pos);

   for (i = 0; i < nEntries; i++)
     XmStringFree (xmstr_arr[i]);
   delete[] xmstr_arr;

   UpdateNumEntries();
   XtVaSetValues (XtParent(fwCoordList),
		  XmNwidth, 160,
		  NULL);
}

void cCoordList::DeleteEntries ()
{
   XmListDeleteAllItems (fwCoordList);
}

void     cCoordList::DeleteEntry (int index)
{
   XmListDeletePos (fwCoordList, index);

   UpdateNumEntries();
}
void cCoordList::ReplaceEntry (int index, TLocation tccPt)
{
   char      buf[20];
   XmString  xmstr[1];

   sprintf (buf, "%9.1lf%9.1lf", tccPt.x, tccPt.y);
   xmstr[0] = XmStringCreate (buf, "list_charset");

   XmListReplaceItemsPos (fwCoordList, xmstr, 1, index);
   XmStringFree (xmstr[0]);
}

void cCoordList::CbSelect (Widget w, XtPointer pObj,
			   XmListCallbackStruct *cbs)
{
   cCoordList  *pCoordList = (cCoordList*)pObj;

   if (pCoordList->fType == ecltMine) {
     cMineComp* pComp = (cMineComp*)(gpGuiWin->GetCurComponent());
     cMine*     pMine = pComp->GetMineNum (cbs->item_position);
     new cCoordEntryWindow (w, pMine);
     gpGuiWin->SetSelection(false, pComp->GetMinefield(), pComp, pMine);
   }
   else if (pCoordList->fType == ecltPerimVertex)
     new cCoordEntryWindow (w, (cPerimeter*)(gpGuiWin->GetCurComponent()), cbs->item_position);
   else if (pCoordList->fType == ecltRowVertex)
     new cCoordEntryWindow (w, (cRowComp*)(gpGuiWin->GetCurComponent()), cbs->item_position);
}

void cCoordList::CbScroll (Widget /*sbar*/, XtPointer pObj, 
				      XmScrollBarCallbackStruct */*cbs*/)
{
  cCoordList* pWin = (cCoordList*)pObj;

  pWin->SetMineToggles ();
}

void cCoordList::CbResize (Widget /*w*/, XtPointer pObj, XEvent* event)
{
   cCoordList* pWin = (cCoordList*)pObj;
   int         i, nitems;

   if (event->type == ConfigureNotify) {
     XtVaGetValues (pWin->fwCoordList, XmNvisibleItemCount, &nitems, NULL);
     
     if (nitems > pWin->fMaxToggles) {
       for (i = pWin->fMaxToggles; i < nitems; i++) {
	 pWin->fwToggle[i] = 
	   XtVaCreateWidget ("",
			     xmToggleButtonWidgetClass, pWin->fwCoordForm,
			     XmNheight, pWin->fListLineSize,
			     XmNleftAttachment, XmATTACH_FORM,
			     NULL);
	 if (i == 0)
	   XtVaSetValues (pWin->fwToggle[i],
			  XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			  XmNtopWidget, pWin->fwCoordList,
			  NULL);
	 else
	   XtVaSetValues (pWin->fwToggle[i], 
			  XmNtopAttachment, XmATTACH_WIDGET,
			  XmNtopWidget, pWin->fwToggle[i-1],
			  NULL);
	 XtAddCallback (pWin->fwToggle[i], XmNvalueChangedCallback,
			(XtCallbackProc)CbOnToggle, (XtPointer)i);
       }
     }
     else {
       for (i = nitems; i < pWin->fMaxToggles; i++)
	 XtDestroyWidget (pWin->fwToggle[i]);
     }
     
     pWin->fMaxToggles = nitems;
     pWin->SetMineToggles ();
  }
}

void cCoordList::SetMineToggles ()
{
   cMineComp*  pMineComp = (cMineComp*)(gpGuiWin->GetCurComponent());
   int         scroll_val, i = -1, j, nmines;

   if (pMineComp == NULL) return;
     
   XtVaGetValues (fwScrollbar, XmNvalue, &scroll_val, NULL);
   nmines = pMineComp->GetNumMines();

   for (cMine* pMine = pMineComp->GetMineNum (scroll_val + 1);
	++i < fMaxToggles && i < nmines && pMine != NULL;
	pMine = pMine->GetNextMine() ) {
   
     XtManageChild (fwToggle[i]);
     XtVaSetValues (fwToggle[i], XmNset, pMine->GetEditOnState(), NULL);
   }
   for (j = i; j < fMaxToggles; j++)
     XtUnmanageChild (fwToggle[j]);
}

void cCoordList::CbOnToggle (Widget w, XtPointer pObj,
			     XmToggleButtonCallbackStruct */*cbs*/)
{
   cMineCompEditorWindow* pWin = (cMineCompEditorWindow*)gpGuiWin->GetCurEditor();
   cCoordList*            pList = (cCoordList*)pWin->GetMineCoordWidget();
   cMineComp*             pMineComp = (cMineComp*)(gpGuiWin->GetCurComponent());
   cMine*                 pMine;
   int                    scroll_val, itog = (int)pObj;

   XtVaGetValues (pList->fwScrollbar, XmNvalue, &scroll_val, NULL);
   pMine = pMineComp->GetMineNum (scroll_val + itog + 1);
   pMine->SetMineOn ((EBoolean)XmToggleButtonGetState (w));
}

void cCoordEntryWindow::CbOK (Widget /*w*/, XtPointer  pObj, 
			      XmPushButtonCallbackStruct */*cbs*/)
{
   cCoordEntryWindow* pWin = (cCoordEntryWindow*)pObj;
   TLocation          tccPt;
   float              psi;

   sscanf ((char*)XmTextGetString (pWin->fwXBox), "%lf", &tccPt.x);
   sscanf ((char*)XmTextGetString (pWin->fwYBox), "%lf", &tccPt.y);
   
   tccPt.z = gpTerrainMgr->GetMapElev(tccPt.x, tccPt.y);

   if (pWin->fType == ecltMine) {
     cMine*     pMine = pWin->fpMine;

     if (pMine) {
       if (! pWin->fpMine->MoveMine (tccPt))
	 return;
     }
     else { /* only indiv comp can add mines */
       if (! ((cIndivComp*)gpGuiWin->GetCurComponent())->AddTccMine (tccPt))
         return;
       pWin->fpMine = ((cIndivComp*)gpGuiWin->GetCurComponent())->GetLastMine();
     }

     if (pWin->fpMine->IsDirectional()) {
       sscanf ((char*)XmTextGetString (pWin->fwOrientBox), 
	       "%f", &psi);
       pWin->fpMine->SetTccOrient (psi, 0., 0.);
     }
   }

   else if (pWin->fType == ecltPerimVertex) {
     cPerimeter*  pPerim = pWin->fpPerim;
     TPerimPoint* pPt = pPerim->GetPointNum (pWin->fVertex);

     if (pPt){
       if (! pPerim->MovePoint (pWin->fVertex, tccPt))
	 return;
     }
     else {
       if (! pPerim->AddTccPoint (tccPt))
	 return;
       cCoordList*  pVertexCoord = 
	 gpGuiWin->GetCurEditor()->GetPerimVertexCoordWidget();
       if (pVertexCoord) pVertexCoord->AppendCoordEntry (tccPt);
     }
   }
     
   else if (pWin->fType == ecltRowVertex) {
     cRowComp*    pRowComp = pWin->fpRowComp;
     TPerimPoint* pPt = pRowComp->GetPointNum (pWin->fVertex);

     if (pPt){
       if (! pRowComp->MovePoint (pWin->fVertex, tccPt))
	 return;
     }
     else {
       if (! pRowComp->AddTccPoint (tccPt))
	 return;
       cCoordList*  pVertexCoord = 
	 gpGuiWin->GetCurEditor()->GetRowVertexCoordWidget();
       if (pVertexCoord) pVertexCoord->AppendCoordEntry (tccPt);
     }
   }
     
   delete (cCoordEntryWindow*)pObj;
}

void cCoordEntryWindow::CbDelete (Widget /*w*/, XtPointer  pObj, 
				  XmPushButtonCallbackStruct */*cbs*/)
{
   cCoordEntryWindow* pWin = (cCoordEntryWindow*)pObj;

   if (pWin->fType == ecltMine)
     gMfieldMgr.DeleteMine (pWin->fpMine);

   else if (pWin->fType == ecltPerimVertex) {
     pWin->fpPerim->DeletePoint (pWin->fVertex);
   }
     
   else if (pWin->fType == ecltRowVertex) {
     pWin->fpRowComp->DeletePoint (pWin->fVertex);
   }
     
   delete (cCoordEntryWindow*)pObj;
}

void cCoordEntryWindow::CbCancel (Widget /*w*/, XtPointer pObj, 
				   XmPushButtonCallbackStruct */*cbs*/)
{
  delete (cCoordEntryWindow*)pObj;
}

Widget cCoordEntryWindow::CreateMainWin (Widget parent)
{
   Widget        wBB, wPane, wWorkArea, wActionArea;

   fwMainWin = 
     XtVaCreatePopupShell ("dialog",
			   xmDialogShellWidgetClass, parent,
			   XmNtitle, "Coordinate Entry", 
			   NULL);
   wBB = 
     XtVaCreateWidget ("dialog",
		       xmBulletinBoardWidgetClass, fwMainWin,
		       XmNdialogStyle, XmDIALOG_PRIMARY_APPLICATION_MODAL,
		       NULL);
   wPane = 
     XtVaCreateWidget ("pane", 
		       xmPanedWindowWidgetClass, wBB,
		       NULL);

   wWorkArea = CreateWorkArea (wPane);
   wActionArea = CreateActionArea (wPane);
      
   XtManageChild( wPane );
   XtManageChild( wBB );
   XtPopup( fwMainWin, XtGrabNone );

   return fwMainWin;
}

Widget cCoordEntryWindow::CreateWorkArea (Widget wParent)
{
   Widget     wWorkArea, wLabel;
   TLocation  tccPt;
   char       orient_text[22], xtext[22], ytext[22];

   wWorkArea = 
     XtVaCreateWidget ("work_area", 
		       xmFormWidgetClass, wParent,
		       NULL);

   if (fType == ecltMine) {
     if (fpMine)
       tccPt = fpMine->GetTccLoc();
     else
       tccPt.x = tccPt.y = 0.;
   }

   else if (fType == ecltPerimVertex) {
     TPerimPoint* pPt = fpPerim->GetPointNum(fVertex);
     if (pPt)
       tccPt = pPt->GetTccPoint();
     else
       tccPt.x = tccPt.y = 0.;
   }

   else if (fType == ecltRowVertex) {
     TPerimPoint* pPt = fpRowComp->GetPointNum(fVertex);
     if (pPt)
       tccPt = pPt->GetTccPoint();
     else
       tccPt.x = tccPt.y = 0.;
   }

   sprintf (xtext, "%.1lf", tccPt.x);
   sprintf (ytext, "%.1lf", tccPt.y);

   fwXBox = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, wWorkArea, 
			      XmNvalue, xtext,
			      XmNcolumns, 8,
			      XmNleftAttachment, XmATTACH_FORM,
			      XmNbottomAttachment, XmATTACH_FORM,
			      NULL);

   fwYBox = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, wWorkArea, 
			      XmNvalue, ytext,
			      XmNcolumns, 8,
			      XmNbottomAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, fwXBox,
			      NULL);
  
   wLabel = 
     XtVaCreateManagedWidget ("X", 
			      xmLabelGadgetClass, wWorkArea,
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwXBox,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, fwXBox,
			      XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNrightWidget, fwXBox,
			      XmNalignment, XmALIGNMENT_CENTER,
			      NULL);
   wLabel = 
     XtVaCreateManagedWidget ("Y", 
			      xmLabelGadgetClass, wWorkArea,
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwYBox,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, fwYBox,
			      XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNrightWidget, fwYBox,
			      XmNalignment, XmALIGNMENT_CENTER,
			      NULL);

   if (fType == ecltMine && (!fpMine || fpMine->IsDirectional())) {
     if (fpMine)
       sprintf (orient_text, "%6.4f", 
	      fpMine->GetTccOrient().psi);
     else
       strcpy (orient_text, "0.0");

     fwOrientBox = 
       XtVaCreateManagedWidget ("textbox", 
				xmTextFieldWidgetClass, wWorkArea, 
				XmNvalue, orient_text,
				XmNcolumns, 8,
				XmNbottomAttachment, XmATTACH_FORM,
				XmNleftAttachment, XmATTACH_WIDGET,
				XmNleftWidget, fwYBox,
				NULL);
     wLabel = 
       XtVaCreateManagedWidget ("+ to North)", 
				xmLabelGadgetClass, wWorkArea,
				XmNbottomAttachment, XmATTACH_WIDGET,
				XmNbottomWidget, fwOrientBox,
				XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
				XmNleftWidget, fwOrientBox,
				XmNalignment, XmALIGNMENT_BEGINNING,
				NULL);
     wLabel = 
       XtVaCreateManagedWidget ("(rads, 0 = East,", 
				xmLabelGadgetClass, wWorkArea,
				XmNbottomAttachment, XmATTACH_WIDGET,
				XmNbottomWidget, wLabel,
				XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
				XmNleftWidget, fwOrientBox,
				XmNalignment, XmALIGNMENT_BEGINNING,
				NULL);
     wLabel = 
       XtVaCreateManagedWidget ("Orientation", 
				xmLabelGadgetClass, wWorkArea,
				XmNbottomAttachment, XmATTACH_WIDGET,
				XmNbottomWidget, wLabel,
				XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
				XmNleftWidget, fwOrientBox,
//				XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET,
//				XmNrightWidget, fwOrientBox,
				XmNalignment, XmALIGNMENT_CENTER,
				NULL);
   }

   XtManageChild( wWorkArea );
   return wWorkArea;
}

Widget cCoordEntryWindow::CreateActionArea (Widget wParent)
{
   Widget  wActionArea, wButton;

   wActionArea = 
     XtVaCreateWidget ("action_area", 
		       xmFormWidgetClass, wParent,
		       XmNfractionBase, 13,
		       NULL);
   wButton = 
     XtVaCreateManagedWidget ("OK", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 1,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 4,
			      NULL);
   XtAddCallback( wButton, XmNactivateCallback, 
		 (XtCallbackProc)CbOK, (XtPointer)this);

   if ((fType == ecltMine        && fpMine) ||
       (fType == ecltPerimVertex && fpPerim->GetPointNum (fVertex)) ||
       (fType == ecltRowVertex   && fpRowComp->GetPointNum (fVertex))) {

     wButton = 
       XtVaCreateManagedWidget ("Delete", 
				xmPushButtonWidgetClass, wActionArea,
				XmNtopAttachment, XmATTACH_FORM,
				XmNleftAttachment, XmATTACH_POSITION,
				XmNleftPosition, 5,
				XmNrightAttachment, XmATTACH_POSITION,
				XmNrightPosition, 8,
				NULL);
     XtAddCallback( wButton, XmNactivateCallback, 
		   (XtCallbackProc)CbDelete, (XtPointer)this);
   }

   wButton = 
     XtVaCreateManagedWidget ("Cancel", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 9,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 12,
			      NULL);
   XtAddCallback( wButton, XmNactivateCallback, 
		 (XtCallbackProc)CbCancel, (XtPointer)this);

   XtManageChild( wActionArea );
   return wActionArea;
}

cCoordEntryWindow::cCoordEntryWindow (Widget parent, cMine* pMine)
{
   fType = ecltMine;
   fpMine = pMine;

   CreateMainWin (parent);
}

cCoordEntryWindow::cCoordEntryWindow (Widget parent, cPerimeter* pPerim, 
				      int ivert)
{
   fType = ecltPerimVertex;
   fpPerim = pPerim;
   fVertex = ivert;

   CreateMainWin (parent);
}

cCoordEntryWindow::cCoordEntryWindow (Widget parent, cRowComp* pRowComp, 
				      int ivert)
{
   fType = ecltRowVertex;
   fpRowComp = pRowComp;
   fVertex = ivert;

   CreateMainWin (parent);
}

cCoordEntryWindow::~cCoordEntryWindow()
{
}
