/*
$RCSfile: EntityDB.h,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/04/29 19:34:09 $
$Log: EntityDB.h,v $
Revision 1.2  1999/04/29 19:34:09  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:49  jchappel
Comprehensive Mine Simulator

 * Revision 2.6  1998/12/02  16:41:39  jchappel
 * added gui detonator
 *
 * Revision 2.5  1998/07/27  17:50:42  ksnively
 * Added single detection trip wire sensor
 *
 * Revision 2.4  1998/07/23  14:15:15  jchappel
 * Made changes for DIS switch.
 *
 * Revision 2.3  1998/04/20  15:10:36  ksnively
 * *** empty log message ***
 *
 * Revision 2.2  1998/02/26  21:26:56  ksnively
 * Created Mf Camera object to control MAK Stealth view.
 * Created vulnerability tables specific to command detonated mines
 * and the camera object.
 *
 * Revision 2.1  1998/02/18  17:17:24  ksnively
 * Incorporated AP mine replacement code
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.6  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.5  1997/05/07  12:33:45  ksnively
 * *** empty log message ***
 *
Revision 1.5  1997/05/07 12:33:45  ksnively
*** empty log message ***

 * Revision 1.4  1997/03/20  19:16:16  ksnively
 * Added emission tracking, updated to CMS2.0B6.
 *
 * Revision 1.3  1997/02/12  17:04:49  ksnively
 * *** empty log message ***
 *
 * Revision 1.2  1997/01/02  19:48:54  nsipes
 * *** empty log message ***
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.15  1996/08/21  21:58:20  rcarp
 * CP and QRP file output, other updates
 *
 * Revision 1.14  1996/07/03  22:55:15  rcarp
 * enhancements and bug fixes for version 1.5b9
 *
 * Revision 1.13  1996/06/22  13:04:11  rcarp
 * deton vuln bug fixes for 1.5b4
 *
 * Revision 1.12  1996/06/21  17:04:20  rcarp
 * upgrades for release 1.5b3, mostly mine vuln
 *
 * Revision 1.11  1996/06/19  19:05:45  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.10  1996/05/23  18:42:07  mkappel
 * Use DbMine list accessors.
 *
 * Revision 1.9  1996/05/22  15:58:40  rcarp
 * fix for db mine list
 *
 * Revision 1.8  1996/05/21  22:15:40  rcarp
 * text and lists from mine db, and activating mine state save cmd
 *
 * Revision 1.7  1996/05/16  00:01:58  rcarp
 * miscellaneous bugs
 *
 * Revision 1.6  1996/05/15  17:17:31  rcarp
 * refinements and fixes for 1.5 delivery
 *
 * Revision 1.5  1996/05/14  16:15:08  rcarp
 * refinements and bug fixes for 1.5 delivery
 *
$State: Exp $
*/
 
/*******************************************

   file: EntityDB.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/**********************************************

   EntityDB.h
   
      Class implementation for a entity types
      database manager
   
   Created by Richard Vuduc
   Modification history:
      June 1995: created   
*********************************************/

#ifndef _MineDB_
#define _MineDB_

#include <string.h>
#include "Globals.h"

/*
 * cEntityDBMgr: Entity Database Manager
 *   Class for maintaining a hierarchial entity tree,
 * to provide data for various purposes
 */

class cEntityDBMgr
{
protected:
   cDbEntry *         fpRoot;      // entity search tree
   int                fRegCount;   // # of entities registered so far

   virtual  EBoolean IsMatch(const TEntityType&, const TEntityType&,
                            int level) const;
   virtual  int      MaxLevel();

   void      ReplaceNode( cDbEntry* oldnode,
                                    cDbEntry* newnode );

//   void      DeleteSubTree( cDbEntry * pEnt);
   EBoolean  Insert( cDbEntry *newnode, cDbEntry *curnode,
                               int level, int endlevel );
   cDbEntry* RecFindName(const char* pName, cDbEntry* curnode) const;
   EBoolean  RecMatchAll( const TEntityType& enttype,
   						cDbEntry ***retbuf, int *count, cDbEntry *curnode,
   						int level, EBoolean bGoSib = true ) const;
   EBoolean  RecMatchAllLevel( const TEntityType& enttype,
   						cDbEntry ***retbuf, int *count, cDbEntry *curnode,
   						int level, int matchLevel, EBoolean bGoSib = true ) const;
   cDbEntry* RecMatchClose( const TEntityType& type,
                                 cDbEntry* curnode, int level );
   EBoolean  RecGetCountryList( cDbEntry* pParent,
                     int **retbuf, int *count, EBoolean bGoSib = true ) const;

public:
               cEntityDBMgr( void );
   virtual    ~cEntityDBMgr( void );

   EBoolean  Register( cDbEntry *newtype, int endlevel );

   cDbEntry* FindByName(const char* pName);
   cDbEntry* MatchClosestOne(const TEntityType& type);
   EBoolean  MatchAll( const TEntityType& enttype,
                     cDbEntry ***retbuf, int *count ) const;
   EBoolean  MatchAllLevel( int matchLevel, const TEntityType& enttype,
                     cDbEntry ***retbuf, int *count ) const;
   cDbEntry* MatchExact( const TEntityType& type );

      // for getting descendents of particular parent, we allow additional
      // filtering primarily to filter leaf mine list by country
   EBoolean  GetNextLevel( cDbEntry* pParent, const TEntityType& filterType,
                     cDbEntry ***retbuf, int *count ) const;
   EBoolean  GetDescendentLeaves( cDbEntry* pParent, const TEntityType& filterType,
                     cDbEntry ***retbuf, int *count ) const;

   EBoolean  GetCountryList( cDbEntry* pParent,
                     int **retbuf, int *count ) const;

   cDbEntry* GetRoot() { return fpRoot; }
   int       GetRegCount( void ) const { return fRegCount; }
};

class cMineDBMgr : public cEntityDBMgr {
      // Maintains the mine database.
      //  4 or 5 level hierarchy:
      // 1 - fixed munition, ent type fields kind and category
      // 2 - domain:  antiarmor, etc
      // 3 - subcategory - conventional, etc
      // 4 - specific and country, 
      // 5 - entity: extra
 public:
   enum { emdblevMine = 1, emdblevDomain, emdblevSubcat, 
          emdblevSpecific, emdblevExtra, emdblevEntity = emdblevExtra };

   virtual  EBoolean IsMatch( const TEntityType&, const TEntityType&,
                              int level ) const;
   virtual  int      MaxLevel();

               cMineDBMgr();
   virtual    ~cMineDBMgr();
};

class cVehDBMgr : public cEntityDBMgr {
      // Maintains the vehicle database.
      //  4 or 5 level hierarchy:
      // 1 - platform ent type field kind = platform
      // 2 - domain:  land, etc
      // 3 - category - tank, small wheeled, bomber, etc
      // 4 - subcategory and country  ( may specify exact entity type,
      //                                in which case specific will still
      //                                have wild cards at level 4
      //                                and 0 at level 5)
      // 5 - specific, entity level
      // ( 6 - extra is never used )
 public:
   enum { evdblevPlatform = 1, evdblevDomain, evdblevCat, 
          evdblevSubcat, evdblevSpecific, evdblevEntity = evdblevSpecific };

   virtual  EBoolean IsMatch( const TEntityType&, const TEntityType&,
                              int level ) const;
   virtual  int      MaxLevel();
   
   void    InitializeVehDb();

               cVehDBMgr();
   virtual    ~cVehDBMgr();
};

class cProjDBMgr : public cEntityDBMgr {
      // Maintains the projectile database.
      //  3-4 level hierarchy:
      // 1 - kind and domain: munition, antiarmor
      // 2- category: 2 branches, guided and ballistic
      // 3 - subcategory and country - things like missile type, may be end level
      // 4 - specific - used for some projectiles 
 public:
   enum { epjdblevKindDom = 1, epjdblevCat, epjdblevSubcat, 
          epjdblevSpecific, epjdblevEntity = epjdblevSpecific };

   virtual  EBoolean IsMatch( const TEntityType&, const TEntityType&,
                              int level ) const;
   virtual  int      MaxLevel();
   
               cProjDBMgr();
   virtual    ~cProjDBMgr();
};

      // this is the standard parameter package, 
      // static values kept in mine database,
      // for which cMine will read values
class cProjectile;
struct TMineParam : public cDbStruct {
   TEntityType         fDetonationType;
   TEntityType         fDetectionType;     /// 204 here ???
   EWarhead            fWarhead;
   TOrient             fTccOrient;      // just the default, at plan mine creation
   int                 fNumRounds;
// unsigned int        fPaintScheme;    Had to change for PDU
   unsigned char       fPaintScheme;
   float               fMeanTimeThermal;  // minutes between updates, should change with real model

	enum {
		empkwDetonType,
		empkwDetectType,
		empkwDetonType204,
		empkwDetectType204,
		empkwDetonTypeList,
		empkwDetectTypeList,
		empkwDetonTypeList204,
		empkwDetectTypeList204,
		empkwWarhead,
		empkwOrientation,
		empkwTimeTherm,
		empkwNumRounds,   // for multi-shot mines, e.g., popup Claymore
		eSizeMineParamTable
	};
   
	static const TKeywordEntry fKeyMineParamTable[eSizeMineParamTable];

         // associating projectile type numbers with text names and db entries
   enum   {
      eprojtypeNone,
      eprojtypeWamSublet,

      eNumProjTypes
   };
   static const char* fpProjTextType[ eNumProjTypes ];
  
 public:
   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);
   virtual void         WriteMineOpt(ofstream* file);
   
   virtual TMineParam*   CloneMineParam()   { return new TMineParam(*this); }
   virtual cProjectile*  CreateProjectile()  { return NULL; }
   
            TMineParam(TEntityType& type);
            TMineParam();
   virtual ~TMineParam();
};

struct TProjParam : public cDbStruct {
   TEntityType         fEntityStateType;
   TEntityType         fDetonationType;
   TEntityType         fGroundImpactType;
   EWarhead            fWarhead;
   T2Dsingle           fLaunchVel;   // horz and z velocities
//   float               fLookAngle;      // from straight down
   float               fImpactVelocity;
//   float               fScanDetectProb;
//   float               fHitProb;        // given detection

   enum {
      eprojkwEntStateTypeList,
      eprojkwDetonTypeList,
      eprojkwImpactTypeList,
      eprojkwWarhead,
      eprojkwLaunchVel,
//      eprojkwLookAng,
      eprojkwImpactVel,
//      eprojkwScanDetProb,
//      eprojkwHitProb,
      
      eSizeProjParamTable
   };
   
   static const TKeywordEntry fKeyProjParamTable[eSizeProjParamTable];
   
 public:
   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);
   void                 WriteProjOpt(ofstream* file);
   
   TProjParam*   CloneProjParam()   { return new TProjParam(*this); }
   
            TProjParam(TEntityType& type);
            TProjParam();
   virtual ~TProjParam();
};

enum EYesNoOpt { eNo, eYes, eBoth };
class TFuzeOptionSelector;

struct TFuzeOpt {
   float               fSelfDestTimeHrs;  // += cur time when sim start
   int                 fTgtCount;         // wait for the nth target
   EBoolean            fbAntihandling;
   EBoolean            fbBlastHard;

   void                SetSelfDestTime(float sdtime) { fSelfDestTimeHrs = sdtime; }
   void                SetTgtCount(int count) { fTgtCount = count; }
   void                SetAntihandling(EBoolean bAh) { fbAntihandling = bAh; }
   void                SetBlastHard(EBoolean bBh) { fbBlastHard = bBh; }

   float               GetSelfDestTime() const { return fSelfDestTimeHrs; }
   int                 GetTgtCount() const { return fTgtCount; }
   EBoolean            GetAntihandling() const { return fbAntihandling; }
   EBoolean            GetBlastHard() const { return fbBlastHard; }
   
   TFuzeOpt(float sdtime = FLT_MAX, int tgtcnt = 0, EBoolean bAH = false, EBoolean bBH = false)
           { fSelfDestTimeHrs = sdtime;
             fTgtCount = tgtcnt;
             fbAntihandling = bAH;
             fbBlastHard = bBH; }
};

class cSimMine;
class cPlanMine;
class TDbOptSet;
class cTracker;
class cSensor;
class cControl;
class cMechanVuln;
class cDetonVuln;
class cEmissVuln;
class cMineComp;
class TDetVulnTable;
class TMechVulnTable;

struct TVulnMechParam : public cDbStruct{

      // the table names are in a keyword list, indexed by enum list
   enum {
      evmtblTwmPPres, evmtblTwmPTilt, evmtblTwmPMag, evmtblTwmPAp, evmtblTwmPCmdAp, evmtblTwmPCam, evmtblTwmPOther,
      evmtblFwmPPres, evmtblFwmPTilt, evmtblFwmPMag, evmtblFwmPAp, evmtblFwmPCmdAp, evmtblFwmPCam, evmtblFwmPOther,
      evmtblTwmRPres, evmtblTwmRTilt, evmtblTwmRMag, evmtblTwmRAp, evmtblTwmRCmdAp, evmtblTwmRCam, evmtblTwmROther,
      evmtblFwmRPres, evmtblFwmRTilt, evmtblFwmRMag, evmtblFwmRAp, evmtblFwmRCmdAp, evmtblFwmRCam, evmtblFwmROther,
      evmtblDbTilt, evmtblDbMag, evmtblDbCmdAp, evmtblDbCam, evmtblDbOther,
      evmtblMagCmMag, evmtblMagCmOther,
      
      evmtblSizeTable
   };
   
   static const char* fKeyMechVulnTables[evmtblSizeTable];
//   static const TKeywordEntry fKeyVulnMechTables[evmtblSizeTable];
   
 public:
   TMechVulnTable*   fpTwmPTable;
   TMechVulnTable*   fpFwmPTable;
   TMechVulnTable*   fpTwmRTable;
   TMechVulnTable*   fpFwmRTable;
   TMechVulnTable*   fpTwmrDbTable;
   TMechVulnTable*   fpMagCmTable;
   
   static const char*   GetNameByIndex( int index );    // key table lookup
 
   virtual EBoolean  ParseDbStructItem(cInFileObj& infile);
   
   TVulnMechParam*   CloneVulnMechParam()   { return new TVulnMechParam(*this); }
   
   cMechanVuln*      CreateVulnMech( cSimMine* pMine );

   void InitVulnMechParam(int fuzeType);
   
   TVulnMechParam();
   TVulnMechParam(int fuzeType);
   virtual ~TVulnMechParam() {}
};

struct cDetVulnList : public T2linkHead {

   TDetVulnTable*   GetVulnDetTable( int detType );         
   float            GetMaxRange(int detType, float depth);

   cDetonVuln*      CreateVulnDet( cSimMine* pMine, float extraRange = 0.0 );

   void        AddVulnDetEntry(int detType, TDetVulnTable* pDvTable);
   void        InitDetVulnList(int fuzeType, EBoolean bBlastHard = false);

   cDetVulnList(int fuzeType, EBoolean bBlastHard = false);
   cDetVulnList(cDetVulnList* pOriginal = NULL);
};

struct TFuzeParam : public cDbStruct {

   int                 fTypeNameIndex; // internal fuze type, e.g., WAM, pressure
   EFuze               fFuze;          // used in detonation PDU
   TFlatEntCoord       fFireOffset;    // x and y params, used differently by each fuze type
   float               fProbDud;       // startup in failed state
   float               fProbDeton;     // given  good encounter
   float               fArmTimeHrs;       // += cur time when sim start
   float               fProbEmplaceDet;       // self-destruct on emplacement
   float               fTimeEmplaceDet;       // after this delay
   EDetonResult        fDetonResult;
   
   float               fTrackRange;    // not intended for user
   
   TFuzeOptionSelector*   fpOptSelector;
   
   TVulnMechParam      fVulnMechParam;
   cDetVulnList        fVulnDetList;
   cDetVulnList        fBhVulnDetList;
//   TVulnDetParam       fVulnDetParam;
//   TVulnDetParam       fBhVulnDetParam;      // set to use for blast hardened fuzes

	enum {
		efuzekwDeleteType,      // used for deleting inherited fuze option
      
		efuzekwFuzeTypeName,
		efuzekwFuzeTypeEnum,
		efuzekwFireOffset,
		efuzekwProbDud,
		efuzekwProbDeton,
		efuzekwArmDelay,
		efuzekwProbEmplaceDet,
		efuzekwTimeEmplaceDet,
		efuzekwDetonResult,
		efuzekwTrackRange,
      
		efuzekwSelfDestDelays,
		efuzekwTgtCountOpt,
		efuzekwAntiHandlingOpt,
		efuzekwSelfBlastHardOpt,

		efuzeSizeTable
	};
 
   static const TKeywordEntry fKeyFuzeTable[efuzeSizeTable];

         // associating internal type numbers with text names and DIS enum types
         //   NOTE: the name/index determines functionality, fFuze above is used
         //   in detonation PDU and can be arbitrarily set in input file
   enum   {
      efuzetypeOther,
      efuzetypePressure,
      efuzetypeTiltrod,
      efuzetypeMagnetic,
      efuzetypeAcousticTracker,
      efuzetypeAcousticTrigger,
      
      efuzetypeTopSublet,
      
      efuzetypeApPressure,
      efuzetypeApTrip,

      efuzetypeCommandFire,
      efuzetypeCommandLoc,
      efuzetypeCommandFireDir,
      efuzetypeTripCommandFire,

      esenstypeCommandCamDir,
      esenstypeAutoCmdCamDir,
      
      esenstypePerDet,      // yes/no trigger
      esenstypeVehDet,
      esenstypePerVehDet,
      esenstypeTripWire,    // One shot detection
      esenstypeDirDet,
      esenstypePerTrack,
      esenstypeVehTrack,
      esenstypePerVehTrack,
      esenstypeDirPVTrack,
      esenstypeDirTrack,

      econtypeAuto,
      econtypeRelay,

      eNumFuzeTypes
   };
   static const TKeyValuePair fpFuzeTextType[ eNumFuzeTypes ];
      
 public:
   void			InitVulnTables(int type);
   static  int            GetTypeIndex(char* pTypeName);
   int 			GetType(){ return fTypeNameIndex;};
   
   TVulnMechParam&      GetVulnMechParam()  { return fVulnMechParam; }
   cDetVulnList&       GetVulnDetList(const TFuzeOpt& fuzeOpt)   
                           { if (fuzeOpt.GetBlastHard())
                                return fBhVulnDetList;
                             else
                                return fVulnDetList; }
 
   virtual EBoolean       ParseDbStructItem(cInFileObj& infile);
           EBoolean       ValidParam();      // not marked for deletion
   virtual void           WriteFuzeOpt(ofstream* file);
   
   const char*            GetFuzeName() const { return fpFuzeTextType[fTypeNameIndex].key; }
   TFuzeOptionSelector*   GetOptSelector() { return fpOptSelector; }
   
   virtual EBoolean   IsDirectional(TFlatEntCoord*);
   virtual EBoolean   IsWideArea(float*);
   virtual EBoolean   IsCmdDet();
   virtual EBoolean   IsClickFire();
   virtual EBoolean   IsController();
   
   virtual   TFuzeParam*   CloneFuzeParam();
   
             cTracker*     CreateTracker( cEntity* pMine, const TFuzeOpt& fuzeOpt, float radius = 0.0 );
             cSensor*      CreateSensor( cEntity* pMine, const TFuzeOpt& fuzeOpt, float radius = 0.0 );
//	     cEmissVuln*   CreateEmissVuln( cEntity* pMine);
    virtual   cControl*     CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt );
  
	       
            TFuzeParam(int type = efuzetypeOther);    // defaults to type "Other"
   virtual ~TFuzeParam();
};

struct TFuzeOptionSelector {

   TFuzeParam*   fpFuzeSet;
   
   int           fNumDestTimeOpt;
   float*        fpDestTimeOpt;   // unit: hours, list of options
   EBoolean      fbTgtCountOpt;
   EYesNoOpt     fAntihandlingOpt;
   EYesNoOpt     fBlasthardOpt;
   
 public:
   void        WriteFuzeOptSel(ofstream* file);
   
   void        GetDestTimeOpt( int* num, float** list);   // may be zero, NULL
   EBoolean    GetTgtCountOpt()     { return fbTgtCountOpt; }
   EYesNoOpt   GetAntihandlingOpt() { return fAntihandlingOpt; }   // returns whether option allowed, cur setting
   EYesNoOpt   GetBlasthardOpt()    { return fBlasthardOpt; }   // returns whether option allowed, cur setting

   void        SetSelfDestOpts( TItemList* pList );
   void        SetTgtCntOpt( EYesNoOpt );
   void        SetAntihanOpt( EYesNoOpt );
   void        SetBlastHardOpt( EYesNoOpt );
   
   TFuzeOptionSelector*   CloneFuzeOptionSelector(TFuzeParam* pFuzeSet);

            TFuzeOptionSelector(TFuzeParam*);
           ~TFuzeOptionSelector();
};

class TParamList {
   enum { eBlockSize = 5 };
   int              fNumBlocks;
   
   cDbStruct**      fppParamList;
   int              fNumParam;
   
 public:
   int              GetNumParam() { return fNumParam; }
   
    
   EBoolean         ReplaceParam(cDbStruct* pOldParam, cDbStruct* pParam);
   
    
   void             AddParam(cDbStruct* pParam);
   cDbStruct*       GetParam(int);
   
   EBoolean         RemoveParam(int);
   void             SetList(cDbStruct* pParam);
   
   TParamList(cDbStruct* pParam = NULL);
   ~TParamList();
};

class cDbMine : public cDbEntry {
 protected:
 
   enum { eminekwEntityType, eminekwEntityTypeList,
          eminekwMine, eminekwFuze, eminekwSensor, 
          eminekwNewMem, emineSizeTable };
   static const TKeywordEntry fKeyMineTable[emineSizeTable];
   
   TParamList      fMineOpt;
   TParamList      fFuzeOpt;
  
 public:
   EBoolean   ParseDbEntry(cInFileObj& infile);
   EBoolean   ValidType(TEntityType& type);
   int        GetDbLevel();
   void       WriteMineParam(ofstream* file);

   void                AddMineParam(TMineParam* pParam)          { fMineOpt.AddParam(pParam); }
   void                AddFuzeParam(TFuzeParam* pParam)          { fFuzeOpt.AddParam(pParam); }
 
   int                 GetNumFuzeParam()        { return fFuzeOpt.GetNumParam(); }
   TMineParam*         GetMineParam(int opt)    { return (TMineParam*)(fMineOpt.GetParam(opt)); }
   TFuzeParam*         GetFuzeParam(int opt)    { return (TFuzeParam*)(fFuzeOpt.GetParam(opt)); }
   int                 GetFuzeOpt(int fuzeTypeIndex);
   
      // usage for following list functions
      // If this entry is at or below requested list level, returns
      //   parent's list along with this entry's index
      // If this entry is above requested level, returns list of entries below
      //   self and -1 for index
      // Indices start at 0, are array index
   EBoolean     GetDomainList( cDbMine* **retbuf, int *count, int* thisindex = NULL );
   EBoolean     GetSubcatList( cDbMine* **retbuf, int *count, int* thisindex = NULL );
   EBoolean     GetSpecificList( cDbMine* **retbuf, int *count,
                            unsigned short country, int* thisindex = NULL );
   EBoolean     GetCountryList( int **retbuf, int *count, int* thisindex = NULL );
   
   char*          GetDomainName();
   char*          GetDomainAltName();
   char*          GetSubcatName();
   char*          GetSubcatAltName();
   const char*          GetCountryName();

   EBoolean     IsDirectional(TDbOptSet& opt, TFlatEntCoord* pMinMax = NULL);  // default false
   EBoolean     IsWideArea(TDbOptSet& opt, float* pRange = NULL);              // default false
   EBoolean     IsCmdDet(TDbOptSet& opt);              // default false
   EBoolean     IsController(TDbOptSet& opt);                //default false
   EBoolean     IsClickFire(TDbOptSet& opt);              // default false
   
   cDbMine*     CloneDbMine();      // returns unlinked copy, of correct specific type
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams) = 0;  // blank copy, does not fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan) = 0;       // bare mine, of appropriate type

   cProjectile*           CreateProjectile(int mineOpt);
   
             cSimMine*    CreateSimMine(cPlanMine* pPlan, cMineComp* pComp);   // for simulating, in component
   virtual   cSimMine*    CreateVirtualMine(cMineComp* pComp, float radius);   // components sensor reference copy
   
             cDbMine();         // with NULLs for params, used when cloning the params
             cDbMine(int fuzeType, int projType);   // user types
             cDbMine(int fuzeType, int projType, TEntityType& type, int level, 
                 char* name = NULL, char* altname = NULL);
   virtual  ~cDbMine();
};

// designed as template, SGI and I could not come to terms on how to do templates
/*
template <class CMine>  class cDbMineSpec : public cDbMine {

 public:
   
   virtual   cDbMine*     CreateDbMine();  // blank copy, doesn't fill fields
   virtual   cSimMine*    CreateMine();       // bare mine, of appropriate type
   virtual   cSimMine*    CreateSimMine(cPlanMine* pPlan, cMineComp* pComp);   // for simulating, in component
   virtual   cSimMine*    CreateVirtualMine(cMineComp* pComp, float radius);   // component's sensor reference copy

             cDbMineSpec(int fuzeType);
             cDbMineSpec(int fuzeType, TEntityType& type, int level, char* pName = NULL);
   virtual  ~cDbMineSpec();
};
*/

class cDbMine_cMine : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, does not fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cMine();
             cDbMine_cMine(int fuzeType);
             cDbMine_cMine(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cMine();
};

class cDbMine_cAtMine : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cAtMine();
             cDbMine_cAtMine(int fuzeType);
             cDbMine_cAtMine(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cAtMine();
};

class cDbMine_cApMine : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApMine();
             cDbMine_cApMine(int fuzeType);
             cDbMine_cApMine(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApMine();
};

class cDbMine_cAtConv : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cAtConv();
             cDbMine_cAtConv(int fuzeType);
             cDbMine_cAtConv(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cAtConv();
};

class cDbMine_cAtScatterable : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cAtScatterable();
             cDbMine_cAtScatterable(int fuzeType);
             cDbMine_cAtScatterable(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cAtScatterable();
};

class cDbMine_cAtOffRoute : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cAtOffRoute();
             cDbMine_cAtOffRoute(int fuzeType);
             cDbMine_cAtOffRoute(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cAtOffRoute();
};

class cDbMine_cAtWideArea : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cAtWideArea();
             cDbMine_cAtWideArea(int fuzeType);
             cDbMine_cAtWideArea(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cAtWideArea();
};

class cDbMine_cApConv : public cDbMine
{

 public:
   
   virtual cDbMine*  CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApConv();
             cDbMine_cApConv(int fuzeType);
             cDbMine_cApConv(int fuzeType, TEntityType& type, int level, 
            char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApConv();
};

class cDbMine_cApScatterable : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApScatterable();
             cDbMine_cApScatterable(int fuzeType);
             cDbMine_cApScatterable(int fuzeType, TEntityType& type, 
            int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApScatterable();
};

class cDbMine_cApConvBding : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApConvBding();
             cDbMine_cApConvBding(int fuzeType);
             cDbMine_cApConvBding(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApConvBding();
};

class cDbMine_cApScatBding : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApScatBding();
             cDbMine_cApScatBding(int fuzeType);
             cDbMine_cApScatBding(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApScatBding();
};

/*
class cDbMine_cApWideArea : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine();       // bare mine, of appropriate type

             cDbMine_cApWideArea();
             cDbMine_cApWideArea(int adenType);
             cDbMine_cApWideArea(int adenType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApWideArea();
};
*/

class cDbMine_cApCommandDet : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApCommandDet();
             cDbMine_cApCommandDet(int adenType);
             cDbMine_cApCommandDet(int adenType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApCommandDet();
};

class cDbMine_cApTripCommandDet : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cApTripCommandDet();
             cDbMine_cApTripCommandDet(int adenType);
             cDbMine_cApTripCommandDet(int adenType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cApTripCommandDet();
};

class cDbMine_cFalseMine : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual   cSimMine*    CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cFalseMine();
             cDbMine_cFalseMine(int fuzeType);
             cDbMine_cFalseMine(int fuzeType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cFalseMine();
};

class cDbProj : public cDbEntry {
 protected:
 
	enum { eprojkwEntityType, eprojkwEntityTypeList,
          eprojkwProj, eprojkwFuze, 
          eprojSizeTable };
   static const TKeywordEntry fKeyProjTable[eprojSizeTable];
   
   TFuzeParam*              fpFuzeParam;
   TProjParam*              fpProjParam;
  
 public:
   EBoolean   ParseDbEntry(cInFileObj& infile);
   EBoolean   ValidType(TEntityType& type);
   int        GetDbLevel();
   void       WriteProjParam(ofstream* file);
 
   TProjParam*            GetProjParam()    { return fpProjParam; }
   TFuzeParam*  			  GetFuzeParam()    { return fpFuzeParam; }
   
   cDbProj*     CloneDbProj();      // returns unlinked copy
   
   virtual   cDbProj*        CreateDbProj(EBoolean bWithParams) = 0;  // blank copy, doesn't fill fields
   virtual   cProjectile*    CreateProj() = 0;
   
             cDbProj();         // with NULLs for params, used when cloning the params
             cDbProj(int fuzeType);   // user types
             cDbProj(int fuzeType, TEntityType& type, int level, char* pName = NULL);
   virtual  ~cDbProj();
};

class cDbProj_cProj : public cDbProj
   //
{

 public:
   
   virtual   cDbProj*        CreateDbProj(EBoolean bWithParams);  // blank copy, doesn't fill fields
   virtual   cProjectile*    CreateProj();

             cDbProj_cProj();
             cDbProj_cProj(int fuzeType);
             cDbProj_cProj(int fuzeType, TEntityType& type, int level, char* pName = NULL);
   virtual  ~cDbProj_cProj();
};

class cDbProj_Guided : public cDbProj
   //
{

 public:
   
   virtual   cDbProj*        CreateDbProj(EBoolean bWithParams);  // blank copy, doesn't fill fields
   virtual   cProjectile*    CreateProj();

             cDbProj_Guided();
             cDbProj_Guided(int fuzeType);
             cDbProj_Guided(int fuzeType, TEntityType& type, int level, char* pName = NULL);
   virtual  ~cDbProj_Guided();
};

class cDbProj_Ballistic : public cDbProj
   //
{

 public:
   
   virtual   cDbProj*        CreateDbProj(EBoolean bWithParams);  // blank copy, doesn't fill fields
   virtual   cProjectile*    CreateProj();

             cDbProj_Ballistic();
             cDbProj_Ballistic(int fuzeType);
             cDbProj_Ballistic(int fuzeType, TEntityType& type, int level, char* pName = NULL);
   virtual  ~cDbProj_Ballistic();
};

/*
class cDbProj_cWamSublet : public cDbProj
{

 public:
   
   virtual   cDbProj*        CreateDbProj(EBoolean bWithParams);  // blank copy, doesn't fill fields
   virtual   cProjectile*    CreateProj();

             cDbProj_cWamSublet();
             cDbProj_cWamSublet(TEntityType& type, int level, char* pName = NULL);
   virtual  ~cDbProj_cWamSublet();
};
*/

#endif

/* eof */

