/*
$RCSfile: FileObj.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:34 $
$Log: FileObj.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:34  jchappel
Comprehensive Mine Simulator

// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.4  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.2  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.11  1996/08/21  21:58:20  rcarp
// CP and QRP file output, other updates
//
// Revision 1.10  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.9  1996/06/19  19:05:45  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.8  1996/05/21  22:15:40  rcarp
// text and lists from mine db, and activating mine state save cmd
//
// Revision 1.7  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.2  1996/02/05  14:32:43  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: FileObj.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/******************************************

   FileObj.cc
   
      The file parser implementations
         
   Created by Richard Vuduc
   Modification history:
      June 1995: created   
      
*********************************************/

#include <iostream.h>
#include <strstream.h>
#include <ctype.h>
#include <fstream.h>
#include <string.h>

#include "FileObj.h"
#include "Vulnerability.h"

   // file types
char*   gftypeUndefined  = "UndefinedCMSfile";
      // both input and output
char*   gftypeMinefields = "MinefieldStore.2";
      // input
char*   gftypeTestFile  = "InternalTestFile";
char*   gftypeParam      = "PerfParams.2";
char*   gftypeNetParam      = "NetParams";
char*   gfmechVulnParam      = "CountermineVuln";
char*   gfdetVulnParam      = "DetonationVuln";
      // output
char*   gftypeInteract   = "MineInteractions";
char*   gftypeMineState  = "MineStateSumm";
char*   gftypeMineInputs  = "MineInputSumm";
char*   gftypeVehicleInputs  = "VehicleInputSumm";
char*   gftypePduData      = "MineDetectPduSumm";

   // file objects
ofstream*   gpInteract            = NULL;
ofstream*   gpMineStateSumm       = NULL;
ofstream*   gpMineInputParams     = NULL;
ofstream*   gpVehicleInputParams  = NULL;
ofstream*   gpPduData             = NULL;
   // file names
char*   gpInteractFileName      = "Interactions";
char*   gpMineStateFileName     = "MineStates";
char*   gpMineInputFileName     = "MineInputs";
char*   gpVehicleInputFileName  = "VehicleInputs";
char*   gpPduDataFileName       = "PduSummary";


//static char *gpDelims = " \t\n";
static const char *gpSeparators = "<>{}[](),;=\"";

/******************* Base file object class implementation ****************/
cFileObjBase::cFileObjBase( void )
{
   ClearStatus();
}

void cFileObjBase::ClearStatus( void )
{
	fLevel = 1;
	fLineCount = 1;
	fErrorCode = eFileNoErr;
}

/******************* Input object class implementation ****************/
cInFileObj::cInFileObj( const char *filename, const char *fileType )
{
	fpToken = new char[ kMaxTokenLength ];
   fbLevelOneSkip = false;
	Reset();

	if( filename != NULL )
		Open( filename, fileType );
}

void cInFileObj::Reset( void )
{
   ClearStatus();
	fpToken[0] = NULL;
}

EBoolean cInFileObj::Open( const char *filename, const char *fileType )
{
	fpInfile.open( filename );

	if( fpInfile.bad() || fpInfile.fail() )
	{
		fErrorCode = eCantOpen;
		return false;
	}
   
	// verify that this is the correct file type
	TKeywordEntry keyTable[1] = {
		"fileType", ekeyString
	};
	int /*retVal,*/ retIndex;
   EFileReturn  retCode = eFileNoErr;
	char *retBuf;

	retCode = ReadItem( keyTable, 1, &retIndex, (void **)&retBuf );
	switch( retCode )
	{
		case eFileNoErr:
			if( fileType != NULL )
				if( strcmp(fileType, retBuf) != 0 ) {
	         	fErrorCode = eWrongFileType;
               Close();
            }

			// else, accept the file
			break;

		default:
	         fErrorCode = retCode;
	}

   if ( fErrorCode == eFileNoErr)
	   return true;
   else
      return false;
}

void cInFileObj::Close( void )
{
	fpInfile.close();
}


cInFileObj::~cInFileObj()
{
//	if( fpToken )
//		delete[] fpToken;
}


/*
 * ParseItem( keyTable, size, idxFound, bufFound )
 *   Assumes a token has been read in and extracted, and is available
 * via GetToken().  Matches it to an entry in 'keyTable' (which has 'size'
 * elements), and parses the statement appropriately.  Returns the index
 * in 'keyTable' of the matched item via 'idxFound', and any return data
 * in 'bufFound.'

 * Returns:
 *   eFileNoErr - successful read
 *   eEOF  - end-of-file, no items read
 *   eNextLevel - A new sub-level structure encountered
 *   others - syntax or other critical error; parsing cannot proceed
 */

EFileReturn cInFileObj::ParseItem( const TKeywordEntry keyTable[], int size,
	int *idxFound, void ** bufFound )
{
	// parse assignment statement or structure

	//--------process this keyword----------
	int i = FindToken( keyTable, size, GetToken() );
	if( i < 0 )   // couldn't find the token
		return fErrorCode = eBadKeyword;

	*idxFound = i;
   switch( keyTable[i].type )
   {
      case ekeyLev1Struct:
         fErrorCode = ReadLev1Struct( bufFound );
         break;

      // add special handling of other types here, e.g., macros, etc.

      default:     // assumed to be an assignment statement
         fErrorCode = ReadAssign( keyTable[i].type, bufFound );
         break;
   }

   return fErrorCode;
}

/*
 * NextItem()
 *   Extracts the next token, checks for end of level, and then returns.
 *
 * Return values:
 *    eEOF       - end of file
 *    eLevelDone - end of level
 *    other      - file error
 */
EFileReturn cInFileObj::NextItem( void )
{
	// get the next token from the input stream
	if( NextToken() == false )
		return GetErrorCode();

	// check for end of this level
	EFileReturn retval = CheckEndLevelToken();
	if( retval != eFileNoErr )   // if end of level encountered, or error
		return fErrorCode = retval;

	return eFileNoErr;
}

/*
 * ReadItem( keyTable, size, idxFound, bufFound )
 *   Reads an 'item' from the input file.  Extracts the next token from the
 * input, and calls ParseItem.
 *
 * Returns:
 *   eFileNoErr - successful read
 *   eEOF  - end-of-file, no items read
 *   eNextLevel - A new sub-level structure encountered
 *   eLevelDone - End of last sub-level
 *   others - syntax or other critical error; parsing cannot proceed
 */
EFileReturn cInFileObj::ReadItem( const TKeywordEntry keyTable[], int size,
	int *idxFound, void **bufFound )
{
	EFileReturn returnCode = NextItem();
	if( returnCode != eFileNoErr )
		return returnCode;
	return ParseItem( keyTable, size, idxFound, bufFound );
}

/*
 * ReadLev1Struct - Read in a structure from the input file of the form
 *
 *    keyword "struct_name" < ... >
 *
 * Upon encountering '<', sets object at next logical parsing level.
 * Returns pointer to a buffer allocated for "struct_name" through
 * bufFound.  Returns an error code, or eFileNoErr on no error.
 *
 * Function assumes the keyword token has already been extracted.
 */
EFileReturn cInFileObj::ReadLev1Struct( void **bufFound )
{
	// extract the name
	if( NextToken() == false )
		return eLev1MissingStructName;

	// implicitly return name of minefield
   // remove quotes if they're there
   if( fpToken[0] == '\"' )
   {
      fpToken[ strlen(fpToken) - 1] = NULL;
      *bufFound = cppstrdup( &fpToken[1] );
   }
   else *bufFound = cppstrdup( fpToken );

	// make sure an open '<' bracket appears
	if( NextToken() )
	{
		if( TokenMatch("<") )
		{
			fLevel++;
			return eNextLevel;
		}
		return eLev1NoOpenBracket;
	}
   return GetErrorCode();
}

/*
 * ReadLev2Struct()
 *    Tries to parse a statement of the form:
 *        keyword = [ ... ] ;
 * Assumes 'keyword =' has already been processed, and that the
 * next token (a possible '[') is ready for processing (i.e., checks for
 * '[' and that's it).  If proper syntax exists, advances the state of
 * the class to the next level.  Returns an error code.
 */
EFileReturn cInFileObj::ReadLev2Struct( void )
{
	// new level 2 or 3 sub level
	if( TokenMatch("[") )
	{
		if ( fLevel == 1 ) {   // skipping the first level block
         fLevel = 3;
         fbLevelOneSkip = true;
      }
      else {
         fLevel++;
         fbLevelOneSkip = false;
      }
		return eNextLevel;
	}
	else return eLev2NoOpenBracket;
}

/*
 * ReadAssign( type, bufFound )
 *    Tries to parse a statement of the form:
 *
 *         keyword = value ;
 *
 * Creates a buffer to store "value" as appropriate, and returns it through
 * bufFound.  "type" indicates the type of value to parse.
 * Assumes "keyword" has already been read.
 */
EFileReturn cInFileObj::ReadAssign( EKeywordType type, void **bufFound )
{
	if( NextToken() == false )
		return GetErrorCode();

	// must be an assignment statement
	if( TokenMatch("=") == false )
		return eExpectedAssign;

	// extract the value
	if( NextToken() == false )
		return eExpectedValue;

	istrstream tok( GetToken() );      // for parsing out info
   EFileReturn retval = eFileNoErr;         // error return value
	switch( type )
	{
         // Now check actual input for quoted or unquoted
		case ekeyString:
		case ekeyQuotedString:
			if( fpToken[0] != '\"' )  { // no quotation marks
	   		*bufFound = cppstrdup( GetToken() );
         }

			// NextToken is guaranteed to return a token with
			// open AND closing quotes

         else {
            fpToken[strlen(fpToken)-1] = NULL;    // replace last quote mark
            *bufFound = cppstrdup( GetToken() + 1 ); // skip first quote
         }
         break;
/*
		case ekeyString:
			*bufFound = cppstrdup( GetToken() );
			break;
*/
		case ekeyFloat:
			*bufFound = new float;
			tok >> *((float *)*bufFound);
			break;

		case ekeyDouble:
			*bufFound = new double;
			tok >> *((double *)*bufFound);
			break;

		case ekeyInt:
			*bufFound = new int;
			tok >> *((int *)*bufFound);
			break;

		case ekeyChar:
			*bufFound = new char;
			tok >> *((char *)*bufFound);
			break;

		case ekeyLong:
			*bufFound = new long;
			tok >> *((long *)*bufFound);
			break;

		case ekeyUnsignedInt:
			*bufFound = new unsigned;
			tok >> *((unsigned int *)*bufFound);
			break;

		case ekeyUnsignedLong:
			*bufFound = new unsigned long;
			tok >> *((unsigned long *)*bufFound);
			break;

		case ekey3DPoint:
			*bufFound = new T3Ddouble;
			retval = ReadPoint( (T3Ddouble *)(*bufFound) );
			break;

		case ekey2DPoint:
			*bufFound = new T2Dsingle;
			retval = ReadPoint( (T2Dsingle *)(*bufFound) );
			break;

		case ekeyVulnPoint:
			*bufFound = new TVulnProb;
			retval = ReadPoint( (TVulnProb *)(*bufFound) );
			break;

      // the following may require substructure parsing
      case ekeyIntList:
      case ekeyFloatList:
		case ekey3DPointList:
		case ekey2DPointList:
		case ekeyVulnPointList:
         retval = ReadList( type, bufFound );
			break;

		case ekeyLev2Struct:
         return ReadLev2Struct();

		default:
         return eUnknownKeyValue;
	}

   // allow error return codes to propagate up
   if( retval != eFileNoErr )
      return retval;

	// parse out the ';' from the assignment
	if( NextToken() == false )
		return eUnexpectedEOF;

	if( TokenMatch(";") == false )
		return eMissingSemicolon;

	return eFileNoErr;
}


/*
 * ReadList( type, bufFound )
 *    Read a list of items in the form:
 *
 *       { item1 [, item2, ..., itemN] }
 *
 * Stores the allocated list as a TItemList structure and returns it via
 * bufFound.  The type of items to read and store is indicated by type.
 */
EFileReturn cInFileObj::ReadList( EKeywordType type, void **bufFound )
{
   // make sure we have open '{'
   if( TokenMatch("{") == false )
      return eListNoOpenBrace;

   EBoolean done = false;
   EFileReturn retval;
   int N = 0;                 // number of items found so far
   TLink *head = new TLink;   // keep track of our new items
   TLink *current = head;     // current node at which to add item

   // loop to read the items
   while( !done )
   {
      retval = ReadListItem( type, current );
      switch( retval )
      {
         case eFileNoErr:         // no errors; get ready for next
            current->next = new TLink;
            current = current->next;
            break;

         case eEOF:
            done = true;     // break out of loop
            break;

         default:
            return retval;   // fatal error
      }
      N++;  // an item successfully added
   }

   // make sure there's a close '}'; should already be in token
   if( TokenMatch("}") == false )
      return eListNoCloseBrace;

   // create an array for these items
   TItemList *list = new TItemList;

   list->count = N;
   CreateList( type, head, list );  // frees head's links as it goes
   *bufFound = list;
   return eFileNoErr;
}

/*
 * ReadListItem( type, current )
 *    Tries to read an item appropriate for the list given by 'type'; stores
 * data in the link node 'current'.  Returns any file errors, or eFileNoErr if
 * the item successfully read and another item is waiting, or eEOF if the
 * item was read and end-of-file || end-of-list encountered.
 */
EFileReturn cInFileObj::ReadListItem( EKeywordType type, TLink *current )
{
   EFileReturn retval;

   T3Ddouble *p3d;
   T2Dsingle *p2s;
   TVulnProb *pvp;
   int       *pI;
   float     *pF;

   switch( type )
   {
      case ekey3DPointList:
         p3d = new T3Ddouble;
         NextToken();  // i don't like this, but oh well
         retval = ReadPoint( p3d );
         if( retval != eFileNoErr )
            return retval;
         current->value = p3d;
         break;

      case ekey2DPointList:
         p2s = new T2Dsingle;
         NextToken();  // i don't like this, but oh well
         retval = ReadPoint( p2s );
         if( retval != eFileNoErr )
            return retval;
         current->value = p2s;
         break;

      case ekeyVulnPointList:
         pvp = new TVulnProb;
         NextToken();  // i don't like this, but oh well
         retval = ReadPoint( pvp );
         if( retval != eFileNoErr )
            return retval;
         current->value = pvp;
         break;

      case ekeyIntList:
         pI = new int;
         fpInfile >> (*pI);
         current->value = pI;
         break;

      case ekeyFloatList:
         pF = new float;
         fpInfile >> (*pF);
         current->value = pF;
         break;

      default:
         return eUnknownKeyValue;
   }

   current->next = NULL;

   // check for next point
   if( NextToken() )
      if( TokenMatch(",") )
         return eFileNoErr;

   return eEOF;
}

/*
 * CreateList( type, head, list )
 *    Creates an array to store in 'list' that consists of the items in
 * the linked list, 'head'.  Assumes that list->count is valid.
 */
void cInFileObj::CreateList( EKeywordType type, TLink *head, TItemList *list )
{
   // allocate memory for array
   switch( type )
   {
      case ekey3DPointList:
         list->ptrItems = (void *) (new T3Ddouble[list->count]);
         break;

      case ekey2DPointList:
         list->ptrItems = (void *) (new T2Dsingle[list->count]);
         break;

      case ekeyIntList:
         list->ptrItems = (void *) (new int[list->count]);         
         break;

      case ekeyFloatList:
         list->ptrItems = (void *) (new float[list->count]);         
         break;

      case ekeyVulnPointList:
         list->ptrItems = (void *) (new TVulnProb[list->count]);         
         break;
   }

   // copy stuff to array
   TLink *cur = head, *next;
   int i = 0;
   while( cur )
   {
      next = cur->next;  // save pointer to next

      // copy from linked list to array
      switch( type )
      {
         case ekey3DPointList:
            ((T3Ddouble *)list->ptrItems)[i++] = *((T3Ddouble *)cur->value);
            break;

         case ekey2DPointList:
            ((T2Dsingle *)list->ptrItems)[i++] = *((T2Dsingle *)cur->value);
            break;

         case ekeyIntList:
            ((int *)list->ptrItems)[i++] = *((int *)cur->value);
            break;

         case ekeyFloatList:
            ((float *)list->ptrItems)[i++] = *((float *)cur->value);
            break;

         case ekeyVulnPointList:
            ((TVulnProb *)list->ptrItems)[i++] = *((TVulnProb *)cur->value);
            break;

         default:
            break;
       }

      // free linked-list node
      delete cur->value;
      delete cur;
      cur = next;
   }
}

EFileReturn cInFileObj::ReadCoord( double *c, EBoolean checkComma )
{
	// assumes point needs to be read in (i.e., not already in token buf)
	if( NextToken() )
	{
		istrstream numstr( GetToken() );

		numstr >> *c;
		if( checkComma )
		{
			if( NextToken() )
				if( TokenMatch(",") )
					return eFileNoErr;
				else return eMissingComma;
		}
		return eFileNoErr;
	}

	return eUnexpectedEOF;
}

EFileReturn cInFileObj::ReadCoord( float *c, EBoolean checkComma )
{
	double tmp;
	EFileReturn retval = ReadCoord( &tmp, checkComma );

	if( retval == eFileNoErr )
		*c = (float)tmp;

	return retval;
}

EFileReturn cInFileObj::ReadPoint( T3Ddouble *ptbuf )
{
	EFileReturn retval;

	// assumes first "(" already stored in token buffer
	if( TokenMatch("(") == false )
		return ePointNoOpenParen;

	// the following ensures that point is specified in the form
	//    (x, y, z)

	retval = ReadCoord( &(ptbuf->x) );
	if( retval != eFileNoErr )
		return retval;
	retval = ReadCoord( &(ptbuf->y) );
	if( retval != eFileNoErr )
		return retval;
	retval = ReadCoord( &(ptbuf->z), false );
	if( retval != eFileNoErr )
		return retval;

	if( NextToken() == false )
		return eUnexpectedEOF;
	if( TokenMatch(")") == false )
		return ePointNoCloseParen;

	return eFileNoErr;
}

EFileReturn cInFileObj::ReadPoint( T2Dsingle *ptbuf )
{
	EFileReturn retval;

	// assumes first "(" already stored in token buffer
	if( TokenMatch("(") == false )
		return ePointNoOpenParen;

	retval = ReadCoord( &(ptbuf->x) );
	if( retval != eFileNoErr )
		return retval;
	retval = ReadCoord( &(ptbuf->y), false );
	if( retval != eFileNoErr )
		return retval;

	if( NextToken() == false )
		return eUnexpectedEOF;
	if( TokenMatch(")") == false )
		return ePointNoCloseParen;

	return eFileNoErr;
}

EFileReturn cInFileObj::ReadPoint( TVulnProb *ptbuf )
{
	EFileReturn retval;

	// assumes first "(" already stored in token buffer
	if( TokenMatch("(") == false )
		return ePointNoOpenParen;

	retval = ReadCoord( &(ptbuf->fDetProb) );
	if( retval != eFileNoErr )
		return retval;
      
	retval = ReadCoord( &(ptbuf->fNeutProb) );
   ptbuf->fNeutProb += ptbuf->fDetProb;
	if( retval != eFileNoErr )
		return retval;
      
	retval = ReadCoord( &(ptbuf->fDisplaceProb) );
   ptbuf->fDisplaceProb += ptbuf->fNeutProb;
	if( retval != eFileNoErr )
		return retval;
      
	retval = ReadCoord( &(ptbuf->fRollbackProb), false );
   ptbuf->fRollbackProb += ptbuf->fDisplaceProb;
	if( retval != eFileNoErr )
		return retval;

	if( NextToken() == false )
		return eUnexpectedEOF;
	if( TokenMatch(")") == false )
		return ePointNoCloseParen;

	return eFileNoErr;
}

/*
 * NextToken:  Extracts the next token from the input stream.
 *   Stores the NULL-terminated token in the class's internal buffer,
 * fpToken.  Includes surrounding quotes on quoted tokens; stops adding
 * characters to the token upon encountering whitespace or separators.
 */
EBoolean cInFileObj::NextToken( void )
{
	char inchar;
	int i;
	EBoolean done;    // true if we've found a separator token
	EBoolean quoted;  // true if we're parsing a quoted string

	SkipBlanks();

	i = 0;            // fpToken[i] is location for next token character
	done = false;
	quoted = false;
	do {

		if( !fpInfile.get(inchar) )  // break on eof
			break;

		switch( inchar )
		{
			case '\n':
				fLineCount++;

				if( quoted )
				{
					fErrorCode = eMissingCloseQuote;
					return false;
				}
				done = true;
				break;

			case '\"':                 // open or close quoted string
            if( i != 0 )
               done = true;

            if( i == 0 || quoted )  // include quotes at beg/end of token
            {
               quoted = (quoted ? false : true);
               fpToken[i++] = inchar;
            }
            else fpInfile.putback( inchar ); // delimiter; return to input

				break;

			case ' ':
			case '\t':
				if( !quoted )
				{
					done = true;
					break;
				}

				// deliberate fall through

			default:
				// #comment this complicated if sometime
				if( strchr(gpSeparators, inchar) && !quoted )
				{
					done = true;
					if( i >= 1 )
					{
						fpInfile.putback( inchar );
						break;
					}
				}
				// add this character to the current token being formed
				fpToken[i++] = inchar;
				break;
		}
	} while( !done && (i < kMaxTokenLength) );

	// on token-size overflow, return error
	if( i == kMaxTokenLength )
	{
		fErrorCode = eTokenTooLong;
		return false;
	}
	else if( i == 0 )   // no more tokens; eof
	{
		fErrorCode = eEOF;
		return false;
	}

	// null-terminate the current token
	fpToken[i] = NULL;

	return true;
}

/*
 * SkipBlanks - Skip over possible whitespace region; maintain line counter
 *    Assumes no characters have been read in yet; reads in and throws away
 * whitespace.  On encountering '\n', updates line counter.
 */
EBoolean cInFileObj::SkipBlanks( void )
{
	char inchar, nextchar;

	while( fpInfile.get(inchar) )   // until eof
	{
		switch( inchar )
		{
			// whitespace
			case '\n':
				fLineCount++;
			case ' ':
			case '\t':
				break;

			// comments
			case '/':
				nextchar = fpInfile.peek();

				if( nextchar == '/' || nextchar == '*' ) // if this is a comment
					if( !SkipComments() )  // error while processing comment
						return false;
					else break;

					// otherwise, this '/' is a delimiter;
					// do deliberate fall-through

			default:
				fpInfile.putback(inchar);
				return true;
		}
	}

	return true;
}

/*
 * SkipComments - Skips over possible comment blocks.
 *   Assumes the last character read from the file stream was a '/';
 * if the next character is a '*' or '/', processes the comment
 * appropriately.  Updates line counter when necessary
 */
EBoolean cInFileObj::SkipComments( void )
{
	char nextchar;

	fpInfile.get( nextchar );
	switch( nextchar )
	{
		case '/':		// process c++ style comment
			// ignore the rest of this line
         do {
            if( !fpInfile.get(nextchar) )     // on eof, exit function
               return true;
         } while( nextchar != '\n' );
         fLineCount++;
			break;

		case '*':      // process c style comment
			do {
				do {
					if( !fpInfile.get(nextchar) )
						// unexpected eof
						return false;

					if( nextchar == '\n' )
						fLineCount++;

				} while( nextchar != '*' );

				fpInfile.get( nextchar );
			} while( nextchar != '/' );
			break;

		default:
			fpInfile.putback( nextchar );
			break;
	}

	return true;
}

/*
 * CheckEndLevelToken - Inspects the current token to see if it is
 *   an "end-structure delimiter" (i.e., close bracket) appropriate
 *   to the current level.
 */
EFileReturn cInFileObj::CheckEndLevelToken( void )
{
	switch( fLevel )
	{
		case 2:
			if( TokenMatch(">") )
			{
				fLevel--;
				return eLevelDone;
			}
			break;

		case 3:
			if( TokenMatch("]") == false )
				break;

			// check for semi-colon
			if( NextToken() == false )
				return GetErrorCode();
			if( TokenMatch(";") == false )
				return eMissingSemicolon;

         if ( fbLevelOneSkip ) {
            fLevel = 1;
            fbLevelOneSkip = false;
         }
         else
			   fLevel--;
			return eLevelDone;

		case 4:
			if( TokenMatch("]") == false )
				break;

			// check for semi-colon
			if( NextToken() == false )
				return GetErrorCode();
			if( TokenMatch(";") == false )
				return eMissingSemicolon;

			fLevel--;
			return eLevelDone;
	}

	return eFileNoErr;
}

/*
 * FindToken( keyTable, size, searchString )
 *    Return the index of the key in keyTable that matches searchString,
 * or -1 if no such match exists.
 */
int cInFileObj::FindToken( const TKeywordEntry keyTable[], int size,
	const char *searchString ) const
{
	int i;

	for( i = 0; i < size; i++ )
		if( keyTable[i] == searchString )
			return i;

	return -1;
}

/******************* Output object class implementation ****************/

cOutFileObj::cOutFileObj( const char *filename, const char *fileType )
{
   Reset();

   if( filename )
      Open( filename, fileType );
}

cOutFileObj::~cOutFileObj( void )
{
   Close();
}

void cOutFileObj::Reset( void )
{
   ClearStatus();
}

EBoolean cOutFileObj::Open( const char *filename, const char *fileType )
{
	fpOutfile.open( filename );

	if( fpOutfile.bad() || fpOutfile.fail() )
	{
		fErrorCode = eCantOpen;
		return false;
	}

	// write indicator
	if( fileType == NULL )
		fileType = gftypeUndefined;

	WriteItem( "fileType", ekeyString, (void *)fileType );

	return true;
}

void cOutFileObj::Close( void )
{
   fpOutfile.close();
}

EFileReturn cOutFileObj::WriteItem( const char *key, EKeywordType type,
   void *buf )
{
   int i;
   EFileReturn retval = eFileNoErr;

   // nice indentation
   for( i = 0; i < (GetLevel()-1); i++ )
      fpOutfile << '\t';

   fpOutfile << key; // keyword first
   switch( type )
   {
      case ekeyLev1Struct:    // output: key "name" <
         fpOutfile << " \"" << (char *)buf << '\"' << endl;
         fpOutfile << "<" << endl;
         fLevel++;
         break;

		case ekeyQuotedString:  // output: key = "string" ;
         fpOutfile << " = \"" << (char *)buf << "\";" << endl;
         break;

		case ekeyString:
         fpOutfile << " = " << (char *)buf << ";" << endl;
			break;

		case ekeyFloat:
         fpOutfile << " = " << *((float *)buf) << ";" << endl;
			break;

		case ekeyDouble:
         fpOutfile << " = " << *((double *)buf) << ";" << endl;
			break;

		case ekeyInt:
         fpOutfile << " = " << *((int *)buf) << ";" << endl;
			break;

		case ekeyChar:
         fpOutfile << " = " << *((char *)buf) << ";" << endl;
			break;

		case ekeyLong:
         fpOutfile << " = " << *((long *)buf) << ";" << endl;
			break;

		case ekeyUnsignedInt:
         fpOutfile << " = " << *((unsigned *)buf) << ";" << endl;
			break;

		case ekeyUnsignedLong:
         fpOutfile << " = " << *((long *)buf) << ";" << endl;
			break;

		case ekey3DPoint:
         fpOutfile << " = ";
         WritePoint( (T3Ddouble *)buf );
         fpOutfile << ";" << endl;
			break;

		case ekey2DPoint:
         fpOutfile << " = ";
         WritePoint( (T2Dsingle *)buf );
         fpOutfile << ";" << endl;
			break;

      // the following may require substructure parsing
      case ekeyIntList:
		case ekey3DPointList:
		case ekey2DPointList:
         fpOutfile << " = " << endl;
         retval = WriteList( type, buf );
			fpOutfile << ";" << endl;
			break;

		case ekeyLev2Struct:
         fpOutfile << " = [" << endl;
         fLevel++;
         break;

		default:
         return eUnknownKeyValue;
   }

   return retval;
}

EFileReturn cOutFileObj::EndLevel( void )
{
   int i;
   // nice indentation
   for( i = 0; i < (GetLevel()-2); i++ )
      fpOutfile << '\t';

   switch( fLevel )
   {
      case 2:
         fpOutfile << ">" << endl << endl;
         break;
         
      case 3:
         fpOutfile << "];" << endl;
         break;
         
      case 4:
         fpOutfile << "];" << endl;
         break;

      default:
         return eBadLevel;
   }
   fLevel--;
   return eFileNoErr;
}

EFileReturn cOutFileObj::WriteComment( const char *s )
{
   if( strchr(s, '\n') )
      fpOutfile << "/* " << s << " */" << endl;
   else fpOutfile << "// " << s << endl;
   return eFileNoErr;
}

EFileReturn cOutFileObj::WriteList( EKeywordType type, void *buf )
{
   int i, j = 1, k;
   TItemList *list = (TItemList *)buf;

	for( i = 0; i < GetLevel(); i++ )
		fpOutfile << '\t';
   fpOutfile << "{ ";

   for( i = 0; i < list->count; i++, j++ )
   {
      switch( type )
      {
         case ekeyIntList:
            fpOutfile << *( ((int *)list->ptrItems)+i );
            break;
   		case ekey3DPointList:
            WritePoint( (T3Ddouble *)list->ptrItems + i );
            break;
	   	case ekey2DPointList:
            WritePoint( (T2Dsingle *)list->ptrItems + i );
            break;

         default:
            return eUnknownKeyValue;
      }

      if( i < (list->count-1) ) // not last item
      {
         fpOutfile << ", ";
         if( (j % 5) == 0 )
			{
				fpOutfile << '\n';
            for( k = 0; k <= GetLevel(); k++ )
               fpOutfile << '\t';
			}
      }
   }

	fpOutfile << endl;
	for( i = 0; i < GetLevel(); i++ )
		fpOutfile << '\t';
   fpOutfile << "}";
   return eFileNoErr;
}

EFileReturn cOutFileObj::WritePoint( const T3Ddouble *p )
{
   fpOutfile << "(" << p->x << ", " << p->y << ", " << p->z << ")";
   return eFileNoErr;
}

EFileReturn cOutFileObj::WritePoint( const T2Dsingle *p )
{
   fpOutfile << "(" << p->x << ", " << p->y << ")";
   return eFileNoErr;
}

/* eof */

