/*
$RCSfile: FileObj.h,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:50 $
$Log: FileObj.h,v $
Revision 1.1.1.1  1999/02/16 14:54:50  jchappel
Comprehensive Mine Simulator

 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.2  1998/01/09  18:29:24  ksnively
 * Integrated latest SAIC release cms3.0b2
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.3  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.2  1997/10/02  17:47:47  austin
 * first working port to modsaf3.0
 *
Revision 1.2  1997/10/02 17:47:47  austin
first working port to modsaf3.0

 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.6  1996/07/03  22:55:15  rcarp
 * enhancements and bug fixes for version 1.5b9
 *
 * Revision 1.5  1996/06/19  19:08:07  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.4  1996/05/14  16:15:08  rcarp
 * refinements and bug fixes for 1.5 delivery
 *
$State: Exp $
*/
 
/*******************************************

   file: FileObj.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   FileObj.h
   
		Class implementation for tokenized
		file input/output
   
   Created by
   Modification history:
      June 1995: created   
*********************************************/

#ifndef _FileObj_
#define _FileObj_

#include <fstream.h>
#include <string.h>
#include "Globals.h"


/* list of possible file I/O result codes */
enum EFileReturn {
         // normal result codes
	eFileNoErr, eNextLevel, eLevelDone, eEOF, eListEnd,
         // error codes
   eSafeErrors,
	eCantOpen, eWrongFileType,
   eTokenTooLong, eBadKeyword, eUnexpectedEOF,
	eLev1MissingStructName, eLev1NoOpenBracket,
	eLev2NoOpenBracket, eExpectedAssign, eExpectedValue,
	eMissingCloseQuote, eMissingSemicolon, eMissingComma,
	ePointNoOpenParen, ePointNoCloseParen, eBadValue, eUnknownKeyValue,
   eListNoOpenBrace, eListNoCloseBrace,

   /* output file errors */
   eWriteError, eBadLevel
};

// operator for short-cut comparison of a keyword and a given string
int inline operator==( const TKeywordEntry& entry,
	const char *key )
{
	return (strcmp(key, entry.key) == 0);
}

/* structure for storing list items (lists of points, lists of ints, etc.) */
struct TItemList
{
	int            count;          // number of items in list
	void           *ptrItems;      // actual list of items
};

/* simple linked list */
struct TLink
{
   void          *value;         // pointer to a buffer
   TLink         *next;          // next item in list

   TLink( void *val = NULL, TLink *n = NULL ) {
      next = n; value = val;
   }
};

const int kMaxTokenLength = 256;     // maximum token length


/*
 * cFileObjBase
 *    Contains almost nothing, really.
 */
class cFileObjBase
{
protected:
	int         fLineCount;         // current line # in file
	int         fLevel;             // current level in file
	EFileReturn fErrorCode;         // error code

public:
   cFileObjBase( void );

	virtual void Reset( void ) = 0;
	virtual EBoolean Open( const char *, const char * ) = 0;
	virtual void Close( void ) = 0;

	EBoolean IsEOF( void ) const {
		return (fErrorCode == eEOF || fErrorCode == eUnexpectedEOF) ?
         true : false;
	}
   
	int GetLevel( void ) const { return fLevel; }
	int GetLineCount( void ) const { return fLineCount; }
	EFileReturn GetErrorCode( void ) const { return fErrorCode; }

   void ClearStatus( void );
};

/*
 * cOutFileObj
 */
class cOutFileObj : public cFileObjBase
{
protected:
   ofstream fpOutfile;

   EFileReturn WriteList( EKeywordType type, void *buf );
   EFileReturn WritePoint( const T3Ddouble * );
   EFileReturn WritePoint( const T2Dsingle * );

public:
   cOutFileObj( const char * = NULL, const char * = NULL );
   ~cOutFileObj( void );

	void Reset( void );
	EBoolean Open( const char *, const char * = NULL );
	void Close( void );

   EFileReturn WriteItem( const char *key, EKeywordType type, void *buf );
   EFileReturn EndLevel( void );
   EFileReturn WriteComment( const char *s );
};

/*
 * cInFileObj
 */
//class TVulnProb;
struct TVulnProb;

class cInFileObj : public cFileObjBase
{
protected:
	ifstream   fpInfile;
	char *     fpToken;
   EBoolean   fbLevelOneSkip;

public:
	cInFileObj( const char * = NULL, const char * = NULL );
	~cInFileObj();

	void       Reset( void );
	EBoolean   Open( const char *, const char * = NULL );
	void       Close( void );

	int        FindToken( const TKeywordEntry[], int, const char * ) const;
	char *     GetToken( void ) { return fpToken; }
   EBoolean   TokenMatch( const char *s ) {
      return (strcmp(GetToken(), s)==0) ? true : false;
   }
	EBoolean    NextToken( void );

	EFileReturn ReadItem( const TKeywordEntry [], int, int *, void ** );
	EFileReturn ParseItem( const TKeywordEntry [], int, int *, void ** );
	EFileReturn NextItem( void );

protected:
	EFileReturn ReadLev1Struct( void ** );
   EFileReturn ReadLev2Struct( void );
	EFileReturn ReadAssign( EKeywordType, void ** );
   EFileReturn ReadList( EKeywordType, void ** );
   EFileReturn ReadListItem( EKeywordType, TLink * );

   void        CreateList( EKeywordType, TLink *, TItemList * );

	EFileReturn ReadPoint( T3Ddouble * );
	EFileReturn ReadPoint( T2Dsingle * );
	EFileReturn ReadPoint( TVulnProb * );
	EFileReturn ReadCoord( double *, EBoolean = true );
	EFileReturn ReadCoord( float *, EBoolean = true );

	EFileReturn CheckEndLevelToken( void );
	EBoolean    SkipBlanks( void );
	EBoolean    SkipComments( void );
};

#endif

/* eof */

