/*
$RCSfile: MapControls.h,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:51 $
$Log: MapControls.h,v $
Revision 1.1.1.1  1999/02/16 14:54:51  jchappel
Comprehensive Mine Simulator

 * Revision 2.2  1998/10/27  19:31:15  jchappel
 * modifications for versioning
 *
 * Revision 2.1  1998/02/18  17:17:24  ksnively
 * Incorporated AP mine replacement code
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.3  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.2  1997/10/02  17:47:47  austin
 * first working port to modsaf3.0
 *
Revision 1.2  1997/10/02 17:47:47  austin
first working port to modsaf3.0

 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.2  1996/07/03  22:55:15  rcarp
 * enhancements and bug fixes for version 1.5b9
 *
 * Revision 1.1  1996/01/23  16:48:05  rcarp
 * Initial revision
 *
 * Revision 1.1  1996/01/23  16:48:05  rcarp
 * Initial revision
 *
$State: Exp $
*/
 
/*******************************************

   file: MapControls.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   MapControls.h
   
      The header file defines the map controls interface classes.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#ifndef _MapControls_
#define _MapControls_

#include "Display.h"
#include "libworld.h"

#define   NUM_SCALES    13
#define   NUM_FEATURES  12

struct scale_data
{
    int                   value;
    char                 *label;
    Widget                wOption;
};

struct feature_data
{
    TACTMAP_OBJECT_CLASS  tmap_class;
    char                 *label;
    cMapDisp*             pMapDisp;
    Widget                wToggle;
};

struct pvd_default_features
{
   int32 num_scales_plus_5;
   char *database_name;
   int32 initial_scale;
   int32 highest_tree_line;
   int32 highest_tree;
   struct
   {
      int32 always_5;
      int32 scale;
      int32 minor_contour;
      int32 major_contour;
      int32 grid_interval;
   } *scales[1];
};

class cMapControlsWindow : public cShellWindow {

friend cGuiWindow;

private:
   static struct scale_data    fScale[];
   static struct feature_data  fFeature[];
   static EBoolean             fMToggleSet[];
   static EBoolean             fVToggleSet[];

   static struct pvd_default_features* fDefaultFeatures;

   Widget     fwScaleMenu;
   Widget     fwXCenter;
   Widget     fwYCenter;
   Widget     fwMinefieldToggle[3];
   Widget     fwVehicleToggle[3];

   Widget     CreateWorkArea (Widget wParent);
   Widget     CreateActionArea (Widget wParent);

   static void  CbOK     (Widget, XtPointer, XmPushButtonCallbackStruct*);
   static void  CbCancel (Widget, XtPointer, XmPushButtonCallbackStruct*);
   static void  Mine_size(Widget, XtPointer, XtPointer);
   static void  Veh_size (Widget, XtPointer, XtPointer);

   static void  SetDefaultFeatures();
   static void  CheckFeatures(int32 index, int32 scale);

   static int   GetFeatureIndex(TACTMAP_OBJECT_CLASS tmClass);

public:
   
   static   EBoolean   IsMineDisplayed( EForce force );
   static   EBoolean   IsVehDisplayed( EForce force );

   void         InitControls ();

                cMapControlsWindow (Widget parent);
   virtual     ~cMapControlsWindow ();
};

#endif
