/*
$RCSfile: MapScroll.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:37 $
$Log: MapScroll.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:37  jchappel
Comprehensive Mine Simulator

// Revision 2.1  1998/06/03  17:18:04  jchappel
// ..
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.3  1996/03/23  14:22:31  rcarp
// improvements to test mode and selection
//
// Revision 1.3  1996/03/23  14:22:31  rcarp
// improvements to test mode and selection
//
// Revision 1.2  1996/02/23  18:40:08  mkappel
// Use cGuiWindow::SetMapCenter.
//
// Initial revision
//
$State: Exp $
*/

#include "Display.h"
#include "MapGeometry.h"
#include <Xm/ScrollBar.h>

#define SB_MAX 1000
#define SB_SIZE 50

void cGuiWindow::UpdateScrollbars ()
{
    int32     value, cc;
    float64   cx, cy /*, width, height*/;
    Arg       arg;
    CTDB     *pCtdb = gpTerrainMgr->GetCtdbAddr();

/*replaced tactmap_locale with tactmap_get_center to match modsaf3.0 libtactmap.h AUSTIN 4 Aug 97*/
/*tactmap_get_dimension (gpTactmap, &width, &height);*/
     tactmap_get_center (gpTactmap, &cx, &cy, &cc);
/*tactmap_locale (gpTactmap, &cx, &cy, &width, &height);*/

    if (cx < pCtdb->min_x)
      cx = pCtdb->min_x;
    else if (cx > pCtdb->max_x)
      cx = pCtdb->max_x;

    if (cy < pCtdb->min_y)
      cy = pCtdb->min_y;
    else if (cy > pCtdb->max_y)
      cy = pCtdb->max_y;

    value =
      (SB_MAX - SB_SIZE) * (cx - pCtdb->min_x)/(pCtdb->max_x - pCtdb->min_x);
    XtSetArg (arg, XmNvalue, value);
    XtSetValues (wHScrollbar, &arg, 1);

    value =
      (SB_MAX - SB_SIZE) * (cy - pCtdb->min_y)/(pCtdb->max_y - pCtdb->min_y);
    XtSetArg (arg, XmNvalue, value);
    XtSetValues (wVScrollbar, &arg, 1);
}

void cGuiWindow::CbScroll (Widget w, XtPointer pObj,
			   XmScrollBarCallbackStruct *call_data)
{
    cGuiWindow* pWindow = (cGuiWindow*)pObj;
    uint8       which;
    float64     cx, cy, width, height;
    float64    *variable, min_val, max_val, size;
    CTDB       *pCtdb = gpTerrainMgr->GetCtdbAddr();
	int32 cc;

    /* Figure out which scrollbar it is */
    if (pWindow->wHScrollbar == w)
      which = XmHORIZONTAL;
    else if (pWindow->wVScrollbar == w)
      which = XmVERTICAL;
    else
      return;
    //printf("begin debug\n");
  /*net two lines put in for debugging*/
       	tactmap_get_center (gpTactmap, &cx, &cy, &cc);
	//printf("tacmap before worldview from scroll %d %d %d\n",(int32)cx,(int32)cy,(int32)cc);
 
    /* When the user is dragging, show the world view */

    tactmap_world_view (gpTactmap, (call_data->reason == XmCR_DRAG));

    /* Get current map info */
/*replaced tactmap_locale with tactmap_get_dimension and tactmap_get_center to match modsaf3.0 libtactmap.h AUSTIN 4 Aug 97*/
     tactmap_get_dimension (gpTactmap, &width, &height);
     tactmap_get_center (gpTactmap, &cx, &cy, &cc);
     //printf("tacmap after world view %d %d %d\n",(int32)cx,(int32)cy,(int32)cc);
/*
     tactmap_locale (gpTactmap, &cx, &cy, &width, &height);
*/
    if (which == XmHORIZONTAL)
    {
	variable = &cx;
	min_val = pCtdb->min_x;
	max_val = pCtdb->max_x;
	size = width;
    }
    else
    {
	variable = &cy;
	min_val = pCtdb->min_y;
	max_val = pCtdb->max_y;
	size = height;
    }

    /* For drag or valueChanged, just go to the new value */
    if ((call_data->reason == XmCR_DRAG) ||
	(call_data->reason == XmCR_VALUE_CHANGED))
    {

	*variable = min_val +
	  (max_val - min_val) * (float64) call_data->value / (float64) SB_MAX;
        /*modified to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
	//printf("tactmap setting center after world view to %d %d %d\n",(int32)cx,(int32)cy,(int32)cc);
	tactmap_set_center (gpTactmap, (int32)cx, (int32)cy, (int32)cc);
	return;
    }

    /* Get the new value */
    switch (call_data->reason)
    {
      case XmCR_DECREMENT:	*variable -= size / 2;	break;
      case XmCR_INCREMENT:	*variable += size / 2;	break;
      case XmCR_PAGE_DECREMENT:	*variable -= size;	break;
      case XmCR_PAGE_INCREMENT:	*variable += size;	break;
      case XmCR_TO_TOP:		*variable = min_val;	break;
      case XmCR_TO_BOTTOM:	*variable = max_val;	break;
    }

    /* Clip the value */
    if (*variable < min_val)
      *variable = min_val;
    if (*variable > max_val)
      *variable = max_val;

    pWindow->SetMapCenter ((int32)cx, (int32)cy, cc);
}

static Arg sb_args[] = { { XmNminimum , 0 } ,
			 { XmNmaximum , SB_MAX } ,
			 { XmNsliderSize , SB_SIZE } ,
		       };

void cGuiWindow::InitScrollbar (Widget sb)
{
    XtAddCallback(sb, XmNdecrementCallback    , (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNdragCallback         , (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNincrementCallback    , (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNpageDecrementCallback, (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNpageIncrementCallback, (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNtoBottomCallback     , (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNtoTopCallback        , (XtCallbackProc)CbScroll, this);
    XtAddCallback(sb, XmNvalueChangedCallback , (XtCallbackProc)CbScroll, this);

    XtSetValues(sb, sb_args, XtNumber(sb_args));
}
