/*
$RCSfile: MineStatusCB.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:39 $
$Log: MineStatusCB.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:39  jchappel
Comprehensive Mine Simulator

// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.8  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.7  1996/06/22  13:04:11  rcarp
// deton vuln bug fixes for 1.5b4
//
// Revision 1.6  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.5  1996/05/11  21:35:41  mkappel
// Add mine age.
//
// Revision 1.4  1996/05/08  00:46:50  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.3  1996/03/23  14:25:01  rcarp
// improvements to test mode and selection
//
// Revision 1.2  1996/03/07  21:29:06  mkappel
// Use IsTargetKilled.
//
// Revision 1.1  1996/03/06  19:45:05  mkappel
// New Mine Status Window
//
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: MineStatusCB.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   MineStatusCreate.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include <Xm/List.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>


void cMineStatusWindow::ShowEdWindow ()
{
   char        buf[225];
   XmString    xmstr;
   TLocation   tccPt;

   fpMineComp = (cMineComp*)(gpGuiWin->GetCurComponent());
   Assert( fpMineComp != NULL, kerrValidityCheck, "cMineStatusWindow::ShowEdWindow");

   XtVaSetValues (fwMinefieldName,
		  XmNvalue, gpGuiWin->GetCurMinefield()->GetMfName(),
		  NULL);

      // component type name
   switch (fpMineComp->GetCompType()) {

   case eCompIndiv:
     xmstr = XmStringCreateSimple("Individual");
     break;

   case eCompGrid:
     xmstr = XmStringCreateSimple("Grid");
     break;

   case eCompRow:
     xmstr = XmStringCreateSimple("Row");
     break;
   }

   XtVaSetValues (fwCompType,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   xmstr = XmStringCreateSimple ((char*)fpMineComp->GetEntType().GetMunitionDomainText());
   XtVaSetValues (fwDomain,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);
  
   if (fpMineComp->GetDbEntity())
     xmstr = XmStringCreateSimple ((char*)fpMineComp->GetDbEntity()->GetEntryEntityName());
   else
     xmstr = XmStringCreateSimple ("");
   XtVaSetValues (fwMineType,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   sprintf (buf, "%.2f", fpMineComp->GetCompMineDepth() * kMeterToInch);
   xmstr = XmStringCreateSimple (buf);
   XtVaSetValues (fwMineDepth,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   sprintf (buf, "%.2f", fpMineComp->GetCompHoursInPlace());
   xmstr = XmStringCreateSimple (buf);
   XtVaSetValues (fwMineAge,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   XmListDeleteAllItems (fwMineList);

   for (cSimMine* pMine = (cSimMine*)(fpMineComp->GetFirstMine());
	pMine != NULL; pMine = (cSimMine*)(pMine->GetNextMine()) ) {

     pMine->SetMinePvdSelect( true );

     tccPt = pMine->GetTccLoc();
     TEntityID targ_id = pMine->GetTargetID ();
     if (targ_id.entityNum != 0) {
       TEntityType targ_type = pMine->GetTargetType();
       sprintf (buf, "%10.1lf%10.1lf   %-12s(%4hu,%4hu,%6hu)   %-12s%-12s", 
		tccPt.x, tccPt.y,
		pMine->GetInternalStateText (pMine->GetInternalState()),
		targ_id.site, targ_id.application, targ_id.entityNum,
		targ_type.GetKindText (),
		pMine->IsTargetKilled()?"  X":"");
     }
     else
       sprintf (buf, "%10.1lf%10.1lf   %-12s", tccPt.x, tccPt.y,
		pMine->GetInternalStateText (pMine->GetInternalState()));

     xmstr = XmStringCreate (buf, "list_charset");
     XmListAddItem (fwMineList, xmstr, 0);
     XmStringFree (xmstr);
   }
   if (gbGui) tactmap_refresh_non_terrain (gpTactmap);

   XtManageChild (fwMainWin);
}

void    cMineStatusWindow::ClearEdWindow()
{
   cEditorWindow::ClearEdWindow();

/*   
   tactmap_bundle_changes (gpTactmap);
   for (cSimMine* pMine = (cSimMine*)(fpMineComp->GetFirstMine());
                  pMine != NULL; 
                  pMine = (cSimMine*)(pMine->GetNextMine()) ) {

      pMine->SetMinePvdSelect( false );
   }
*/
   
   if (gbGui) tactmap_refresh_non_terrain (gpTactmap);
   fpMineComp = NULL;
}

void    cMineStatusWindow::SetButtonState ()
{
   cMfieldObj* pCurMine = gpGuiWin->GetCurMfObject();
   if ( pCurMine ) {
      fpMineComp->GetObjIndex(pCurMine);
      XmListSelectPos (fwMineList, fpMineComp->GetObjIndex(pCurMine), FALSE);
   }
}

void cMineStatusWindow::CbScroll (Widget /*sbar*/, XtPointer pObj, 
				  XmScrollBarCallbackStruct */*cbs*/)
{
  cMineStatusWindow* pWin = (cMineStatusWindow*)pObj;

  pWin->SetMineToggles ();
}

void cMineStatusWindow::CbResize (Widget /*w*/, XtPointer pObj, XEvent* event)
{
   cMineStatusWindow* pWin = (cMineStatusWindow*)pObj;
   int                i, nitems;

   if (event->type == ConfigureNotify) {
     XtVaGetValues (fwMineList, XmNvisibleItemCount, &nitems, NULL);
     
     if (nitems > fMaxToggles) {
       for (i = fMaxToggles; i < nitems; i++) {
	 fwToggle[i] = 
	   XtVaCreateWidget ("",
			     xmToggleButtonWidgetClass, fwStatusArea,
			     XmNheight, fListLineSize,
			     XmNrightAttachment, XmATTACH_WIDGET,
			     XmNrightWidget, fwMineList,
			     NULL);
	 if (i == 0)
	   XtVaSetValues (fwToggle[i], XmNtopAttachment, XmATTACH_FORM, NULL);
	 else
	   XtVaSetValues (fwToggle[i], 
			  XmNtopAttachment, XmATTACH_WIDGET,
			  XmNtopWidget, fwToggle[i-1],
			  NULL);
	 XtAddCallback (fwToggle[i], XmNvalueChangedCallback,
			(XtCallbackProc)CbOnToggle, (XtPointer)i);
       }
     }
     else {
       for (i = nitems; i < fMaxToggles; i++)
	 XtDestroyWidget (fwToggle[i]);
     }
     
     fMaxToggles = nitems;
     pWin->SetMineToggles ();
  }
}

void cMineStatusWindow::SetMineToggles ()
{
   cMineComp*  pMineComp = (cMineComp*)(gpGuiWin->GetCurComponent());
   int         scroll_val, i = -1, j;

   if (pMineComp == NULL) return;
     
   XtVaGetValues (fwScrollbar, XmNvalue, &scroll_val, NULL);

   for (cMine* pMine = pMineComp->GetMineNum (scroll_val + 1);
	++i < fMaxToggles &&pMine != NULL;
	pMine = pMine->GetNextMine() ) {
   
     XtManageChild (fwToggle[i]);
     XtVaSetValues (fwToggle[i], XmNset, pMine->GetMineToggleState(), NULL);
     //XtVaSetValues (fwToggle[i], XmNset, pMine->GetEditOnState(), NULL);
   }
   for (j = i; j < fMaxToggles; j++)
     XtUnmanageChild (fwToggle[j]);
}

void cMineStatusWindow::CbOnToggle (Widget w, XtPointer pObj,
				    XmToggleButtonCallbackStruct */*cbs*/)
{
   cMine*      pMine;
   int         scroll_val, itog = (int)pObj;

   XtVaGetValues (fwScrollbar, XmNvalue, &scroll_val, NULL);
   pMine = fpMineComp->GetMineNum (scroll_val + itog + 1);
   
   if ( pMine->IsCmdDet() )
      pMine->CmdDet();
   else
      pMine->SetMineOn ((EBoolean)XmToggleButtonGetState (w));
      
   ((cMineStatusWindow*)(gpGuiWin->GetCurEditor()))->SetMineToggles();
}

void cMineStatusWindow::CbDone (Widget /*w*/, XtPointer /*pObj*/, 
				XmPushButtonCallbackStruct */*cbs*/)
{
   tactmap_bundle_changes (gpTactmap);
   for (cSimMine* pMine = (cSimMine*)(fpMineComp->GetFirstMine());
                  pMine != NULL; 
                  pMine = (cSimMine*)(pMine->GetNextMine()) ) {

      pMine->SetMinePvdSelect( false );
   }
   gpGuiWin->SetSelection( false, fpMineComp->GetMinefield(), fpMineComp );
   gpGuiWin->ShowEditor(cGuiWindow::eEditCompStatus);
}

void cMineStatusWindow::CbCancel (Widget /*w*/, XtPointer /*pObj*/, 
				  XmPushButtonCallbackStruct */*cbs*/)
{
   tactmap_bundle_changes (gpTactmap);
   for (cSimMine* pMine = (cSimMine*)(fpMineComp->GetFirstMine());
                  pMine != NULL; 
                  pMine = (cSimMine*)(pMine->GetNextMine()) ) {

      pMine->SetMinePvdSelect( false );
   }
   gpGuiWin->SetSelection( false, fpMineComp->GetMinefield(), fpMineComp );
  gpGuiWin->ShowEditor(cGuiWindow::eEditCompStatus);
}

void cMineStatusWindow::CbMineItem (Widget /*w*/, XtPointer /*pObj*/,
				    XmListCallbackStruct *cbs)
{
         //** do we need to worry about scrolling of list?
   cMine*   pMine = fpMineComp->GetMineNum (cbs->item_position);
   cMinefield* pMf = fpMineComp->GetMinefield();
   
   gpGuiWin->SetSelection(false, pMf, fpMineComp, pMine);
}
