/*
$RCSfile: NetMgr.h,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/06/03 19:45:00 $
$Log: NetMgr.h,v $
Revision 1.2  1999/06/03 19:45:00  ksnively
Updated CMS Makefile to compile CMS and create DEPENDS file for multiple platforms allowing a DEPENDS file to exist for each type of platform CMS is compiled on concurrently.

Updated supporting make-cms file to support platform dependent DEPENDS file.

Small corrections to NetMgr files.

Revision 1.1.1.1  1999/02/16 14:54:52  jchappel
Comprehensive Mine Simulator

 * Revision 2.1  1998/10/27  19:32:36  jchappel
 * modifications for versioning
 *
 * Revision 2.1  1998/10/27  19:32:36  jchappel
 * modifications for versioning
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.2  1998/01/09  18:29:24  ksnively
 * Integrated latest SAIC release cms3.0b2
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.4  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.3  1997/10/02  17:47:47  austin
 * first working port to modsaf3.0
 *
Revision 1.3  1997/10/02 17:47:47  austin
first working port to modsaf3.0

 * Revision 1.2  1996/11/06  15:13:32  nsipes
 * *** empty log message ***
 *
 * Revision 1.2  1996/11/06  15:13:32  nsipes
 * *** empty log message ***
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.5  1996/09/06  18:57:57  rcarp
 * CP bug fixes
 *
 * Revision 1.4  1996/08/21  21:58:20  rcarp
 * CP and QRP file output, other updates
 *
 * Revision 1.3  1996/08/07  14:19:30  rcarp
 * initial revisions for CP protocol
 *
 * Revision 1.2  1996/05/08  00:45:10  rcarp
 * changes for 1.5, incl user mines, mine params, param i/o, selection
 *
 * Revision 1.2  1996/05/08  00:45:10  rcarp
 * changes for 1.5, incl user mines, mine params, param i/o, selection
 *
$State: Exp $
*/
 
/*******************************************

   file: NetMgr.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   NetMgr.h
   
      The network manager header file defines the cNetMan class, which
      is the simulation's API to the network.
      
      Specific PDUs and the network manager's lower level net
      access routines are defined in ????????????
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#ifndef _UNetMgr_
#define _UNetMgr_

#include "Globals.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

//REGION NPAC - UPDATE (HLA)!
#include "HwFederateAmbassador.h"
//END OF REGION 

enum ENetReturn { enetMsgRecv = 0, enetMsgSent, enetNoData, enetErr, enetSelfMsg };
enum EbufKb { ebufSize = 512 };

extern  int      gPduLen;

class cMine;
class cVehicle;
class cSimVehicle;
class cWamSublet;
class cProjectile;

class cPort {
      // base class for either UDP or Assoc net port

public:
   virtual   ENetReturn      GetPdu(void *buffer, int buflen) = 0;
   virtual   int             PutPdu(void *buffer, int buflen) = 0;

   cPort();
   virtual   ~cPort();
};

class cNonetPort : public cPort {
      // does not send or receive PDUs

public:
   virtual   ENetReturn      GetPdu(void *buffer, int buflen);
   virtual   int             PutPdu(void *buffer, int buflen);

   cNonetPort();
   virtual   ~cNonetPort();
};

   // udp interface
class cUdpPort : public   cPort {

protected:
   ulong         hostid;		   // ip addr of host
   int           sport;          // send port number
   int           s;   		   	// send socket
   sockaddr_in   sin;            // send sockaddr_in
   int           slen;			   // # bytes sent
   struct ip_mreq  sMreq;                   // multicast address
   u_char        sMttl;                     // multicast time-to-live
   int           rport;			   // recv port number
   int           r;			      // recv socket
   sockaddr_in   rin;	         // recv sockaddr_in
   int           rinlen;		   // length of addr recv''d from
   int           rlen;			   // # bytes recvd
   ENetReturn     errcode;		   // err code set during opens/reads/or writes
   
   int            OpenPort(int udpPort, int ttl = 0, char* pMultireq = NULL);
   int            ClosePort();
   
public:
   virtual ENetReturn    GetPdu(void *buffer, int buflen);
   virtual int           PutPdu(void *buffer, int buflen);

              cUdpPort(int udpPort, int ttl = 0, char* pMultireq = NULL);
   virtual   ~cUdpPort();
};

class cAssocPort : public   cPort {

public:
   virtual   ENetReturn     GetPdu(void *buffer, int buflen);
   virtual   int            PutPdu(void *buffer, int buflen);

   cAssocPort();
   virtual   ~cAssocPort();
};

class cPduTranslator {

protected:
      void *      fpOutBuf;
      int         fLength;
      char	  fProtocolFamily; 

public:
      virtual   void        ReceivePdu(void* buf, TTimeSim timeDelta = 0) = 0;
      
                void        SetLength( int len ) { fLength = len; }
	    	void	    SetProtocolFamily(char pf) {fProtocolFamily = pf; }

      cPduTranslator( int bufSize = 1500 );
};

class cDisPduTranslator;
//class cDisPdu204;  /*takeout?*/
class cCmsTranslator;
class cNetMgr {

enum EBufSize { ePduBufSize = 1536 };

      cPort*            fpSimPort;
      cPort*            fpCmsPort;
      cDisPduTranslator*   fpPduTranslator;
//    cDisPdu204*       fpPduTranslator;  /*takeout?*/
      cCmsTranslator*   fpCmsTranslator;
      void*             fpBuffer;
      int               fPduLength;
      
      int               fExerciseId;

public:
      EBoolean          RecvPdu();      // true if one or more processed
      void              SendPdu(void * pBuf, int length);
            
      cDisPduTranslator*       GetPduTrans() { return fpPduTranslator; }      
      int               GetExerciseId();
      
      EBoolean          UsesCp();
      EBoolean          IncludeThermal();
      EBoolean          IncludeBurial();
      EBoolean          IncludeOrient();
      EBoolean          UsesQrp();
      
      void              NetShutdown();
   
                        cNetMgr( TStartupParam & startup );
                       ~cNetMgr();
};


// REGION NPAC - UPDATE (HLA)!

class cHLAPort : public   cPort {

protected:
   RTI::RTIambassador rtiAmb;
   HwFederateAmbassador fedAmb;
   RTI::FederateHandle federateId;

   char* const FederationName = "HelloWorld";
   char* const disInterClassName = "Communication";
   char* const disParamName = "Message ";
   char federateName[200];

   RTI::Boolean        timeAdvGrant;
   RTIfedTime          grantTime;


   short disInterClassTypeId;
   int disParamTypeId;

   ENetReturn     errcode;

   int            OpenPort();
   int            ClosePort();
  
public:
   virtual ENetReturn    GetPdu(void *buffer, int buflen);
   virtual int           PutPdu(void *buffer, int buflen);
   cHLAPort();
   virtual   ~cHLAPort();
   int PublishAndSubscribe();
   int decideTiming();

};

//END OF REGION 

#endif
