/*
$RCSfile: RowEditorCB.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:43 $
$Log: RowEditorCB.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:43  jchappel
Comprehensive Mine Simulator

// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.4  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.7  1996/07/13  19:04:20  rcarp
// version 1.5b13, fixes bug dragging with two buttons and displays damamged vehicles
//
// Revision 1.6  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.5  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.4  1996/04/04  16:25:34  mkappel
// Improve performance with large mine sets.
//
// Revision 1.3  1996/03/23  14:23:28  rcarp
// improvements to test mode and selection
//
// Revision 1.2  1996/02/21  19:45:02  rcarp
// revisions for map scroll, mine ranges
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: RowEditorCB.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   RowEditorCB.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include "EntityDB.h"
#include "Message.h"

#include <Xm/List.h>
#include <Xm/Text.h>
#include <bstring.h>
#include <math.h>
#include <stdext.h>

SNSTVE_CLASS  cRowDisp::fSensNull;
SNSTVE_CLASS  cRowDisp::fSensRow;
SNSTVE_CLASS  cRowDisp::fSensRowSelect;
GC            cRowDisp::fRowGC[1];
int           cRowDisp::fSensRowInstanceCounter;

void cRowEditorWindow::ShowEdWindow ()
{
   char       str[20];
   cRowComp*  pRowComp = (cRowComp*)gpGuiWin->GetCurComponent();

   cMineCompEditorWindow::ShowEdWindow();

   sprintf (str, "%.2f", pRowComp->GetSpacing());
   XtVaSetValues (fwSpacingText,
		  XmNvalue, str,
		  NULL);
      
   sprintf (str, "%.2f", pRowComp->GetOffset());
   XtVaSetValues (fwOffsetText,
		  XmNvalue, str,
		  NULL);
      
   // also show entries for vertices of row
   fpVertexCoord->DeleteEntries();
   fpVertexCoord->AppendCoordEntry ();

   for (TPerimPoint* pPt = pRowComp->GetFirstPoint();
	pPt; pPt = (TPerimPoint*)pRowComp->GetNextPoint())

     fpVertexCoord->AppendCoordEntry (pPt->GetTccPoint());

   fpVertexCoord->UpdateNumEntries();

   // make a copy in case cancel
/*
   if (((cMineComp*)gpCurComponent)->GetNumMines() > 0)
     fpCompCopy = new cRowComp ((cRowComp*)gpCurComponent, 
				gpCurComponent->GetMinefield(),
				false);
   else
     fpCompCopy = NULL;
*/
}

void  cRowEditorWindow::MapClick (TLocation& loc)
{
  if (((cRowComp*)(gpGuiWin->GetCurComponent()))->AddTccPoint (loc))
    fpVertexCoord->AppendCoordEntry (loc);
}

void cRowEditorWindow::CbSpacing (Widget /*w*/, XtPointer pObj, 
				  XmAnyCallbackStruct */*cbs*/)
{
  cRowEditorWindow*    pWindow = (cRowEditorWindow*)pObj;
  float                spacing, offset;
  //int                  nmines;
  char                 msg[200];
  cRowComp*            pRowComp = (cRowComp*) (gpGuiWin->GetCurComponent());

  sscanf ((char*)XmTextGetString (pWindow->fwSpacingText), "%f", &spacing);

  if (spacing > 0.) {
    fNmines = pRowComp->RowMineCount (spacing);
    if (fNmines > kMaxMinesInField) {
      sprintf (msg, "A row spacing of %.2f meters might\ngenerate more than %d mines.\nPlease increase the Spacing to\nreduce the number of mines.", 
	       spacing, kMaxMinesInField);

      gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL, msg);
    }
    else if (fNmines > 100) {
      sprintf (msg, "A row spacing of %.2f meters\nwill generate about %d mines.  Are you sure\nthat you want to create that many?", 
	       spacing, fNmines);

      gpGuiWin->DialogMessage (cGuiWindow::eDialogWarning, 
			       (XtCallbackProc)CbSpacingOK, msg);
    }
    else {
      sscanf ((char*)XmTextGetString (pWindow->fwOffsetText), "%f", &offset);
      pRowComp->SetSpacing (spacing, offset, fNmines);
      ((cMineCompEditorWindow*)pWindow)->SetButtonState();
    }
  }
}

void cRowEditorWindow::CbSpacingOK (Widget w, XtPointer  pObj,
				     XmAnyCallbackStruct */*cbs*/)
{
  cRowEditorWindow*   pWindow = (cRowEditorWindow*)pObj;
  cRowComp*           pRowComp = (cRowComp*)(gpGuiWin->GetCurComponent());
  float               spacing, offset;

  sscanf ((char*)XmTextGetString (pWindow->fwSpacingText), "%f", &spacing);
  sscanf ((char*)XmTextGetString (pWindow->fwOffsetText), "%f", &offset);
  pRowComp->SetSpacing (spacing, offset, fNmines);
  ((cMineCompEditorWindow*)pWindow)->SetButtonState();
  XtDestroyWidget (w);
}

void cRowDisp::CbMove (SNSTVE_WINDOW_PTR /*sensitive*/, 
			 SNSTVE_ACTION action,
			 SNSTVE_CALL_DATA *call_data, 
			 char */*name*/)
{
  float64         dx, dy, xmap, ymap, arccos, a2, b2, c2, off_vec[XYC];
  static float64  xorig, yorig;
  int32           cmap, v;
  static int32    corig;
  TLocation       tccPt;
  T2dMove         move;
  cRowDisp*       pRowDisp = (cRowDisp *)call_data->user_data; 
  cRowComp*       pRowComp = pRowDisp->fpRow;

  if (call_data->button == 2) {
    /* since libsensitive is not returning correct offset for Motion */
    if (action == SNSTVE_PRESS)
/*modified parameters to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
       tactmap_screen_to_map (gpTactmap,
			     call_data->screen_x,
			     call_data->screen_y,
			     &xorig, &yorig, &corig);

    else if (action == SNSTVE_MOTION || action == SNSTVE_RELEASE) {
      TLocation    tccCenter = pRowComp->GetTccLoc();

/*modified parameters to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
       tactmap_screen_to_map (gpTactmap,
			     call_data->screen_x,
			     call_data->screen_y,
			     &xmap, &ymap, &cmap);

      a2 = (xmap - tccCenter.x) * (xmap - tccCenter.x) +
	   (ymap - tccCenter.y) * (ymap - tccCenter.y);
      b2 = (xorig - tccCenter.x) * (xorig - tccCenter.x) +
	   (yorig - tccCenter.y) * (yorig - tccCenter.y);
      c2 = (xorig - xmap) * (xorig - xmap) +
	   (yorig - ymap) * (yorig - ymap);

      arccos = acos(-.5 * (c2 - a2 - b2) / (sqrt(a2) * sqrt(b2)));

      /* point right of line implies counterclockwise rotation */
      if (xmap * (tccCenter.y - yorig) + ymap * (xorig - tccCenter.x) +
	  (tccCenter.x * yorig - tccCenter.y * xorig) < 0)
	arccos = -arccos;

      move.SetMove (0., 0., (float)arccos);
      
      if (action == SNSTVE_MOTION)
	pRowDisp->RotateObj (move);
      else /* RELEASE */
	if (! pRowComp->MoveCompOnly (move)) {
	  /* restore original position */
	  move.SetMove (0., 0., 0.);
	  pRowDisp->RotateObj (move);
	}
    }
  }

  else if (action == SNSTVE_RELEASE) {

    /* move */

    if (call_data->button == 1) {
/*replaced tactmap_pixel_offset with tactmap_pix_to_offvec to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
      tactmap_pix_to_offvec (gpTactmap,
			    call_data->xoff,
			    call_data->yoff,
			    off_vec);
	dx=off_vec[0];
	dy=off_vec[1];			

/*
      tactmap_pixel_offset (gpTactmap,
			    call_data->xoff,
			    call_data->yoff,
			    &dx, &dy);
*/
      /* vertex - just move vertex */
      if (TACTMAP_IS_VERTEX (call_data->instance)) {
	v = TACTMAP_INDEX (call_data->instance);
	tccPt.x = pRowDisp->fDfn.linear.vertices[v*2] + dx;
	tccPt.y = pRowDisp->fDfn.linear.vertices[v*2+1] + dy;
	pRowComp->MovePoint ((int)v + 1, tccPt);
      }

      /* edge - move whole row */
      else {
	tccPt.x = dx;
	tccPt.y = dy;
	pRowComp->MoveRow (tccPt);
      }
    }

    /* delete */

    else if (call_data->button == 3) {
      if (TACTMAP_IS_VERTEX (call_data->instance)) {
	v = TACTMAP_INDEX (call_data->instance);
	pRowComp->DeletePoint ((int)v + 1);
      }
    }
  }
}

void cRowDisp::CbClickSelect (SNSTVE_WINDOW_PTR /*sensitive*/, 
			 SNSTVE_ACTION /*action*/,
			 SNSTVE_CALL_DATA *call_data, 
			 char */*name*/)
{
  cRowDisp*       pRowDisp = (cRowDisp *)call_data->user_data; 
  cRowComp*       pRowComp = pRowDisp->fpRow;

  cMinefield*     pMf = pRowComp->GetMinefield();
  
  gpGuiWin->SetSelection(false, pMf, pRowComp);
  
}

void cRowDisp::RotateObj (T2dMove move)
{
  TLocation       tccPt, tccCenter = fpRow->GetTccLoc();
  int             iPt = 0;

  for (TPerimPoint* pPt = fpRow->GetFirstPoint();
       pPt; pPt = (TPerimPoint*)fpRow->GetNextPoint()) {

    move.fRotMatrix.Mult (pPt->GetTccPoint() - tccCenter, &tccPt);
    fDfn.linear.vertices[iPt*2]   = tccPt.x + tccCenter.x;
    fDfn.linear.vertices[iPt*2+1] = tccPt.y + tccCenter.y;
    iPt++;
  }
  tactmap_update_object (gpTactmap, fpObj, &fDfn);
}

void cRowDisp::MoveVertex (int iPt, TLocation tccPt)
{
   //int32        nvert = fDfn.linear.n_vertices;
   cCoordList*  pVertexCoord = 
     gpGuiWin->GetCurEditor()->GetRowVertexCoordWidget();

   if (pVertexCoord)
     pVertexCoord->ReplaceEntry (iPt - 1, tccPt);

   fDfn.linear.vertices[iPt*2] = tccPt.x;
   fDfn.linear.vertices[iPt*2+1] = tccPt.y;

   tactmap_update_object (gpTactmap, fpObj, &fDfn);
}

cRowDisp::cRowDisp (cRowComp *pRow)
{
   fpRow = pRow;
   ++fSensRowInstanceCounter;

   fDfn.sensitive.sensClass = &fSensNull;
   fDfn.sensitive.instance = fSensRowInstanceCounter;
   fDfn.sensitive.user_data = (char*)this;

   fDfn.linear.style = LSplain;
   fDfn.linear.dashed = 0;
   fDfn.linear.width = 0;
   fDfn.linear.thickness = 1;
   fDfn.linear.gc = fRowGC[0];
   fDfn.linear.n_vertices = 0;
   fDfn.linear.vertices = new float64 [ 20 ];
   /*added AUSTIN 13 Aug 97*/
   fDfn.linear.cell = (int32)0;

   fpObj = tactmap_create_object (gpTactmap, TACTMAP_LINEAR, &fDfn);
   tactmap_add_object (gpTactmap, fpObj);
}

void cRowEditorWindow::CbOffset (Widget /*w*/, XtPointer pObj, 
				 XmAnyCallbackStruct */*cbs*/)
{
  cRowEditorWindow*    pWindow = (cRowEditorWindow*)pObj;
  cRowComp*            pRowComp = (cRowComp*)(gpGuiWin->GetCurComponent());
  float                spacing, offset;
//  char                 msg[200];

  sscanf ((char*)XmTextGetString (pWindow->fwSpacingText), "%f", &spacing);

  if (spacing > 0.) {
    sscanf ((char*)XmTextGetString (pWindow->fwOffsetText), "%f", &offset);
    pRowComp->SetSpacing (spacing, offset);
  }
}

void cRowDisp::AddVertex (int nvert, TLocation tccPt)
{
   if (nvert % 10 == 0)
     fDfn.linear.vertices = 
       (float64 *) realloc (fDfn.linear.vertices,
			    2 * (nvert + 10) * sizeof (float64));

   fDfn.linear.vertices[(nvert-1)*2] = tccPt.x;
   fDfn.linear.vertices[(nvert-1)*2+1] = tccPt.y;
   fDfn.linear.n_vertices = nvert;
   tactmap_update_object (gpTactmap, fpObj, &fDfn);
}

void cRowDisp::DeleteVertex (int iPt)
{
   int          v, nvert = (int)fDfn.linear.n_vertices;
   TLocation    tccPt;
   cCoordList*  pVertexCoord = 
     gpGuiWin->GetCurEditor()->GetRowVertexCoordWidget();

   if (pVertexCoord) {
     tccPt.x = fDfn.linear.vertices[iPt*2];
     tccPt.y = fDfn.linear.vertices[iPt*2+1];
     pVertexCoord->DeleteEntry (iPt + 1);
   }

   for (v = iPt; v < nvert; v++) {
     fDfn.linear.vertices[v*2] = fDfn.linear.vertices[(v+1)*2];
     fDfn.linear.vertices[v*2+1] = fDfn.linear.vertices[(v+1)*2+1];
   }
   fDfn.linear.n_vertices--;
   tactmap_update_object (gpTactmap, fpObj, &fDfn);
   gpGuiWin->SetEditorButtonState ();
}

void cRowDisp::SetSensitive (EBoolean sens)
{
  if (sens) {
    fDfn.sensitive.sensClass = &fSensRow;
    tactmap_update_object (gpTactmap, fpObj, &fDfn);
  }
  else if (fDfn.sensitive.sensClass == &fSensRow) {
    fDfn.sensitive.sensClass = &fSensNull;
    tactmap_update_object (gpTactmap, fpObj, &fDfn);
  }
}

void cRowDisp::SetRowSelectSens (EBoolean sens)
{
  if (sens) {
    fDfn.sensitive.sensClass = &fSensRowSelect;
    tactmap_update_object (gpTactmap, fpObj, &fDfn);
  }
  else if ( fDfn.sensitive.sensClass == &fSensRowSelect ) {
    fDfn.sensitive.sensClass = &fSensNull;
    tactmap_update_object (gpTactmap, fpObj, &fDfn);
  }
  
}

cRowDisp::~cRowDisp ()
{
}

void cRowDisp::InitGCs (Widget wShell)
{
   Window            root;
   XGCValues         gcv;

   struct row_disp_resources
     {
       Pixel row;
       Font  designations;
     } resources;

   static XtResource res_list[] = {
     { "row" , "Row" , XmRPixel , sizeof(Pixel) ,
       XtOffset(struct row_disp_resources *, row) ,
       XmRString , "White" } ,
     { "designations" , "Designations" , XmRFont , sizeof(Font) ,
       XtOffset(struct row_disp_resources *, designations) ,
       XmRString , "Fixed" } ,
   };

   root = RootWindowOfScreen(XtScreen(wShell));

   /* Get the colors, and various other resources used */
   XtGetSubresources (wShell, &resources, "row", "ROW",
		      res_list, XtNumber(res_list), NULL, 0);

   gcv.font = resources.designations;
   gcv.background = 
     tactmap_get_color(gpTactmap,(TACTMAP_OBJECT_CLASS)-1);

   gcv.foreground = resources.row;
   fRowGC [0] = 
     XCreateGC(gpDisplay,root,GCForeground|GCBackground|GCFont,&gcv);

   SNSTVE_INIT_CLASS(fSensNull);
   SNSTVE_INIT_CLASS(fSensRow);
   callback_register_handler(fSensRow.gesture, 
			     (CALLBACK_HANDLER)CbMove,
			     "row");
   fSensRow.buttons = 0x7; /* Buttons 1, 2, and 3 */
   fSensRow.sensitive = fSensRow.hot = fSensRow.dragable = 1;

   SNSTVE_INIT_CLASS(fSensRowSelect);
   callback_register_handler(fSensRowSelect.gesture, 
			     (CALLBACK_HANDLER)CbClickSelect,
			     "rowSel");
   fSensRowSelect.buttons = 0x2; /* 1 << 1, Button 1 */
   fSensRowSelect.sensitive = fSensRowSelect.hot = 1;

   fSensRowInstanceCounter = 0;
}
