/*
$RCSfile: SaSensor.h,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:53 $
$Log: SaSensor.h,v $
Revision 1.1.1.1  1999/02/16 14:54:53  jchappel
Comprehensive Mine Simulator

 * Revision 2.4  1998/12/02  16:46:40  jchappel
 * added directional sensor
 *
 * Revision 2.3  1998/07/23  14:00:29  jchappel
 * *** empty log message ***
 *
 * Revision 2.2  1998/02/18  17:17:24  ksnively
 * Incorporated AP mine replacement code
 *
 * Revision 2.2  1998/02/18  17:17:24  ksnively
 * Incorporated AP mine replacement code
 *
 * Revision 2.1  1998/02/18  17:11:56  ksnively
 * *** empty log message ***
 *
 * Revision 2.1  1998/01/28  14:04:11  ksnively
 * New file for Ap mine replacement.
 *
 * Revision 1.2  1997/07/18  13:47:07  rcarp
 * 2.1b4, mostly changes to fire APLR and hide mine states
 *
 * Revision 1.1  1997/05/12  22:10:05  rcarp
 * Area Denial mods, version 2.1b1
 *
$State: Exp $
*/
 
/*******************************************

   file: SaSensor.h
   
      Copyright 1996 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   SaSensor.h
   
      The sensor and gateway header file
      
      
   
   Created by Richard Carpenter
      
*********************************************/

#ifndef _USaSensor_
#define _USaSensor_

#include "Globals.h"
#include "Minefield.h"
#include "Vehicle.h"
#include "EntityDB.h"
#include "Display.h"
#include "ApMine.h"
#include "AtMine.h"
#include "AtWideArea.h"

// ************************************
//   basic types
// ************************************

// ************************************
//   component classes
// ************************************
class cDispVehicle : public cVehicle
{
   
 public:
 
      void       UpdateDisplay();

                 cDispVehicle( const cVehicle& origVeh );
      virtual   ~cDispVehicle();
};

class cLinkVehTrkDisp : public cLinkVehErr
      // tracking with a track display capability,
      // creates a new cVehicle that it updates using the original
      // cVehicle
{

 protected:
    cDispVehicle*       fpDispVehicle;
   
  
 public:
      void              DisplayVeh( EBoolean disp );
      

      virtual void      UpdateTrkRec();

   cDispVehicle*        GetDispVehicle() const  { return fpDispVehicle; }

                 cLinkVehTrkDisp(cLinkVehTrkDisp* pTrkVeh, cLinkVeh* pPrev);
                 cLinkVehTrkDisp(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev, cEntity* pRefEnt, 
      EBoolean refInvert, int tgtType, float rangeErr, float velErr, float azErr);
      virtual   ~cLinkVehTrkDisp();
};

class cSensTrkTracker : public cLandTracker {

 protected:

 public:

   virtual cLinkVehGeom*   CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev);
   virtual cLinkVehGeom*   CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev);

            cSensTrkTracker(cEntity* pRefEnt, float range, 
                        EBoolean useVelCoord = true, 
                        EBoolean vehTrack = true, 
                        EBoolean lifeTrack = false) : 
                     cLandTracker(pRefEnt, range, useVelCoord, vehTrack, lifeTrack) {}
   virtual ~cSensTrkTracker();
};

struct TSensParamDumb : public TFuzeParam {

      // variable size arrays of ranges and detect probs
   int           fNumBands;
   float*        fpRange;
   float*        fpDetectProb;
  
   float         fCrossRange; 
   float         fMaxRange;  


	enum {
		edmbsenskwFuzeTypeName,      // intercept type name, donot allow unuseable types
      edmbsenskwRange,
      edmbsenskwProb,

		edmbsensSizeTable
	};
 
   static const TKeywordEntry fKeyDmbSensTable[edmbsensSizeTable];
   
   EBoolean      DetectsTgt( float range );
 
   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);

//   virtual   cTracker*     CreateTracker( cEntity* pMine, const TFuzeOpt& fuzeOpt, float radius = 0.0 );
   virtual   cControl*     CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt );

   virtual EBoolean   IsDirectional(TFlatEntCoord* pMinMax);
   virtual EBoolean   IsDirectional();
   
   virtual   TFuzeParam*   CloneFuzeParam();
   
            TSensParamDumb(int sensType = TFuzeParam::esenstypePerDet);
   virtual ~TSensParamDumb();
};

struct TSensParamTrack : public TSensParamDumb {

      // variable size arrays of ranges and detect probs
   float        fRangeErr;
   float        fAzError;

   float         fCrossRange; 
   float         fMaxRange;  
	enum {
		etrksenskwFuzeTypeName,      // intercept type name, do not allow unuseable types
                etrksenskwRangeErr,
                etrksenskwAzErr,

		etrksensSizeTable
	};
 
   static const TKeywordEntry fKeyTrkSensTable[etrksensSizeTable];
   
   EBoolean      TracksTgt( float range );

   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);

//   virtual   cTracker*     CreateTracker( cEntity* pMine, const TFuzeOpt& fuzeOpt, float radius = 0.0 );
   virtual   cControl*     CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt );
   virtual EBoolean   IsDirectional(TFlatEntCoord* pMinMax);
   virtual EBoolean   IsDirectional();

   virtual   TFuzeParam*   CloneFuzeParam();
   
            TSensParamTrack(int sensType = TFuzeParam::esenstypePerTrack);
   virtual ~TSensParamTrack();
};


class cDbMine_cSaSensor : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual cSimMine*     CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cSaSensor();
             cDbMine_cSaSensor(int sensType);
             cDbMine_cSaSensor(int sensType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cSaSensor();
};

class cDbMine_cSensDetector : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual cSimMine*     CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cSensDetector();
             cDbMine_cSensDetector(int sensType);
             cDbMine_cSensDetector(int sensType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cSensDetector();
};

class cDbMine_cSensTracker : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual cSimMine*     CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cSensTracker();
             cDbMine_cSensTracker(int sensType);
             cDbMine_cSensTracker(int sensType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cSensTracker();
};

class cDbMine_cSaController : public cDbMine
{

 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual cSimMine*     CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

             cDbMine_cSaController();
             cDbMine_cSaController(int contType);
             cDbMine_cSaController(int contType, TEntityType& type, 
         int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cSaController();
};

class cDetSensorControl : public cMineControl {
protected:
   cSimMine*         fpSensor;
   TSensParamDumb*   fpSensParam;
   float      fAzimuth;   // fire angle, from centerline   
public:
   virtual   void      UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList, TTimeSim* pWaitTime);
           EBoolean   InFootprint(cLinkVehGeom* pTgtLink);
             
   virtual   void      CalcGeom(const cLinkVehGeom* pLink, TAttackGeom* pGeom);

   virtual      EBoolean   NeedsPrUpdate( EMineInternalState minestate);
   virtual      EBoolean   NeedsUpdate( EMineInternalState minestate);

             cDetSensorControl(cSimMine* pSensor, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt, 
                                   TSensParamDumb* pSensParam );
   virtual  ~cDetSensorControl();
};

class cTrkSensorControl : public cMineControl {
protected:
   cSimMine*         fpSensor;
   TSensParamTrack*  fpSensParam;
   
public:
   virtual   void      UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList, TTimeSim* pWaitTime);
             
   virtual   void      CalcGeom(const cLinkVehGeom* pLink, TAttackGeom* pGeom);

   virtual      EBoolean   NeedsPrUpdate( EMineInternalState minestate);
   virtual      EBoolean   NeedsUpdate( EMineInternalState minestate);

             cTrkSensorControl(cSimMine* pSensor, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt, 
                                   TSensParamTrack* pSensParam );
   virtual  ~cTrkSensorControl();
};


// ************************************
//   mine classes
// ************************************
class cSaSensor : public cSimMine {

 friend class cDbMine_cSaSensor;

 private:
   static  TEntityType   fsSaSensorEntityType;
   static  char*         fsSaSensorName;
   static  char*         fsSaSensorAltName;
   static  int           fsDefaultSensType;
   
   static   cDbMine_cSaSensor          fsDbSaSensor;      // the instantiation in mine database
   
 protected:

public:
           // constructor for virtual mine, for component sensor
//   cSaSensor(cMineComp* pParent, float radius):
//      cSimMine(pParent, radius);

   cSaSensor( cPlanMine* pPlan = NULL );
   virtual   EMineInternalState    GetDisplayState();

   virtual   ~cSaSensor();
};

class cSensDetector : public cSaSensor {

 friend class cDbMine_cSensDetector;

 private:
   static  TEntityType   fsSensDetectorEntityType;
   static  char*         fsSensDetectorName;
   static  char*         fsSensDetectorAltName;
   static  int           fsDefaultSensDetectorType;
   
   static   cDbMine_cSensDetector      fsDbSensDetector;      // the instantiation in mine database
   
 protected:

public:
           // constructor for virtual mine, for component sensor
//   cSensDetector(cMineComp* pParent, float radius):
//      cSaSensor(pParent, radius);

   cSensDetector( cPlanMine* pPlan = NULL );

   virtual   ~cSensDetector();
};

class cSensTracker : public cSaSensor {

 friend class cDbMine_cSensTracker;

 private:
   static  TEntityType   fsSensTrackerEntityType;
   static  char*         fsSensTrackerName;
   static  char*         fsSensTrackerAltName;
   static  int           fsDefaultSensTrackerType;
   
   static   cDbMine_cSensTracker          fsDbSensTracker;      // the instantiation in mine database
   
 protected:

public:
   float      GetRangeErr()
      { return ((TSensParamTrack*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fRangeErr; }
   float      GetAzErr()
      { return ((TSensParamTrack*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fAzError; }

   cSensTracker( cPlanMine* pPlan = NULL );

   virtual   ~cSensTracker();
};

class cSaController : public cSimMine {

 friend class cDbMine_cSaController;

 private:
   static  TEntityType   fsSaControllerEntityType;
   static  char*         fsSaControllerName;
   static  char*         fsSaControllerAltName;
   static  int           fsDefaultContlType;
   
   static   cDbMine_cSaController          fsDbSaController;      // the instantiation in mine database
   
 protected:

public:

   virtual   void          UpdateMine();

   cSaController(  cPlanMine* pPlan = NULL );

   virtual   ~cSaController();
};


#endif
