/*
$RCSfile: TestMode.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:44 $
$Log: TestMode.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:44  jchappel
Comprehensive Mine Simulator

// Revision 2.2  1998/04/20  15:10:36  ksnively
// *** empty log message ***
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.9  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.8  1996/06/21  17:04:20  rcarp
// upgrades for release 1.5b3, mostly mine vuln
//
// Revision 1.7  1996/06/19  19:08:07  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.6  1996/05/14  20:40:15  rcarp
// bug fix
//
// Revision 1.5  1996/05/15  14:56:45  mkappel
// Change button label.
//
// Revision 1.3  1996/05/11  15:15:24  rcarp
// test mode and mine selector work
//
// Revision 1.1  1996/03/16  18:36:36  mkappel
// Interface for Test Mode files.
//
// Revision 1.1  1996/03/16  18:36:36  mkappel
// Interface for Test Mode files.
//
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: TestMode.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

#include "CmsControl.h"
#include "TestMode.h"
#include "MapGeometry.h"

#include <math.h>

#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>

const TKeywordEntry TTurnParam::fTurnParamTable[eturnSizeTable] = {
	"turnTimeRel",   ekeyFloat,      // in seconds
	"turnVelocity",  ekey2DPoint
};

const TKeywordEntry cTestVehicle::fTestVehTable[etestvehSizeTable] = {
	"startTimeRel",  ekeyFloat,      // in seconds
	"endTimeRel",    ekeyFloat,
	"startLocation", ekey2DPoint,
	"startVelocity", ekey2DPoint,
	"killProb",      ekeyFloat,
	"turnParam",     ekeyLev2Struct,
	"partParam",     ekeyLev2Struct
};

const TKeywordEntry TVehPart::fKeyVehPartTable[evpkwSizeTable] = {
	"partType",        ekeyInt,
	"depth",           ekeyFloat,
	"attachPoint",     ekeyInt,
	"attached",        ekeyChar,
	"valueType",       ekeyUnsignedInt,
	"value",           ekeyFloat,
	"entityTypeList",  ekeyIntList
};

void   cTestVehTurnMsg::SendMsg()
{
   fpTestVeh->NextTurn();
}

cTestVehTurnMsg::cTestVehTurnMsg(cTestVehicle* pTestVeh, 
      TTimeSim time, EBoolean bAbsTime) :
   cMsgLowRateUpdt(time, bAbsTime)
{
   fpTestVeh = pTestVeh;
}

cTestVehTurnMsg::~cTestVehTurnMsg()
{
}

cTestModeWindow::cTestModeWindow (Widget wParent)
{
   Widget        wPane, wWorkArea, wActionArea, wButton;
   int           i, nfiles;
   XmString      xmstr;

   fwMainWin = 
     XtCreatePopupShell	("Test Mode Input", transientShellWidgetClass,
			wParent, NULL, 0);
   wPane = 
     XtVaCreateWidget ("pane", 
		       xmPanedWindowWidgetClass, fwMainWin,
		       NULL);

   wWorkArea = 
     XtVaCreateWidget ("work_area", 
		       xmFormWidgetClass, wPane, 
		       NULL);
   fwList = 
     XmCreateScrolledList (wWorkArea, "list", NULL, 0);
   XtVaSetValues (fwList, 
		  XmNvisibleItemCount, 10,
		  XmNselectionPolicy, XmMULTIPLE_SELECT,
		  NULL);
   nfiles = gCmsMgr.GetNumTestFiles();
   for (i = 0; i < nfiles; i++) {
      xmstr = XmStringCreateSimple (gCmsMgr.GetTestFilenames(i));
      XmListAddItemUnselected (fwList, xmstr, 0);
      XmStringFree (xmstr);
    }

   XtAddCallback (fwList, XmNmultipleSelectionCallback,
		  (XtCallbackProc)CbSelect, this);
   XtManageChild (fwList);
   XtManageChild( wWorkArea );

   wActionArea = 
     XtVaCreateWidget ("action_area", 
		       xmFormWidgetClass, wPane,
		       XmNfractionBase, 13,
		       NULL);
   fwStartButton = 
     XtVaCreateManagedWidget ("Start", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 1,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 4,
			      XmNsensitive, FALSE,
			      NULL);
   XtAddCallback( fwStartButton, XmNactivateCallback, 
		 (XtCallbackProc)CbStart, (XtPointer)this);
   wButton = 
     XtVaCreateManagedWidget ("New File", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 5,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 8,
			      NULL);
   XtAddCallback( wButton, XmNactivateCallback, 
		 (XtCallbackProc)CbNewFile, (XtPointer)this);
   wButton = 
     XtVaCreateManagedWidget ("Cancel", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 9,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 12,
			      NULL);
   XtAddCallback( wButton, XmNactivateCallback, 
		 (XtCallbackProc)CbCancel, (XtPointer)this);

   {
     Dimension  h;
     XtVaGetValues (wButton, XmNheight, &h, NULL);
     XtVaSetValues (wActionArea, XmNpaneMaximum, h, XmNpaneMinimum, h, NULL);
   }
   XtManageChild( wActionArea );
      
   XtManageChild( wPane );
}

cTestModeWindow::~cTestModeWindow ()
{
}

void cTestModeWindow::CbSelect (Widget w, XtPointer pObj, 
				XmListCallbackStruct */*cbs*/)
{
   cTestModeWindow* pWin = (cTestModeWindow*)pObj;
   int              nsel;

   XtVaGetValues (w, XmNselectedItemCount, &nsel, NULL);
   XtSetSensitive (pWin->fwStartButton, (nsel > 0));
}

void cTestModeWindow::CbStart (Widget /*w*/, XtPointer pObj, 
			       XmPushButtonCallbackStruct */*cbs*/)
{
      // get the selected files
   cTestModeWindow* pWin = (cTestModeWindow*)pObj;
   int              nsel;
   XmString*        pXmstr;      // array of compound strings
   char*            pFileName;

   XtVaGetValues (pWin->fwList, XmNselectedItemCount, &nsel, 
                                XmNselectedItems, &pXmstr, NULL);

   for ( int i = 0; i < nsel; i++ ) {
      XmStringGetLtoR ( pXmstr[i], XmFONTLIST_DEFAULT_TAG, &pFileName );
      gVehMgr.ReadTestFile(pFileName);
      XtFree( pFileName );
   }
   XtPopdown (((cTestModeWindow*)pObj)->GetMainWin());
}

void cTestModeWindow::CbNewFile (Widget /*w*/, XtPointer pObj, 
				 XmPushButtonCallbackStruct */*cbs*/)
{
   Widget          file_sb;
   XmString        xmstr;

   file_sb = XmCreateFileSelectionDialog (fwTopLevelShell,
					  "file_sb", NULL, 0);
   XtVaSetValues (file_sb, 
		  XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL, 
		  XmNtitle, "Test Mode Input Files",
		  NULL);
   xmstr = XmStringCreateSimple ("Add");
   XtVaSetValues (XmFileSelectionBoxGetChild (file_sb,
					      XmDIALOG_OK_BUTTON),
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);
   xmstr = XmStringCreateSimple ("Done");
   XtVaSetValues (XmFileSelectionBoxGetChild (file_sb,
					      XmDIALOG_CANCEL_BUTTON),
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);
   XtAddCallback (file_sb, XmNcancelCallback, 
		  (XtCallbackProc)XtDestroyWidget, NULL);
   XtAddCallback (file_sb, XmNokCallback, 
		  (XtCallbackProc)CbFileOK, pObj);
   XtManageChild (file_sb);
}

void cTestModeWindow::CbFileOK (Widget /*fs*/, XtPointer pObj, 
				XmFileSelectionBoxCallbackStruct *cbs)
{
   cTestModeWindow* pWin = (cTestModeWindow*)pObj;
   char     *filename;
   XmString  xmstr;

   XmStringGetLtoR (cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);
   xmstr = XmStringCreateSimple (filename);
   XmListAddItem (pWin->fwList, xmstr, 0);
   XmListSelectPos (pWin->fwList, 0, TRUE);
   XmListSetBottomPos (pWin->fwList, 0);
   XmStringFree (xmstr);
}

void cTestModeWindow::CbCancel (Widget /*w*/, XtPointer pObj, 
				XmPushButtonCallbackStruct */*cbs*/)
{
   XtPopdown (((cTestModeWindow*)pObj)->GetMainWin());
}


EBoolean  TVehPart::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fKeyVehPartTable, evpkwSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
/*
		case eNextLevel:
			switch( returnIndex )
			{
				case evpkwEntityTypeStruct: {
               break;

				default: // ??
                  success =  false;
					break;
			}
			break;
*/

		case eFileNoErr:
			switch( returnIndex )
			{
				case evpkwPartType:
                  fParamType = (EPartParam)(*(int*)returnVal);
                  delete returnVal;
                  returnVal = NULL;
					break;

				case evpkwDepth:
                  fDepth = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
					break;

				case evpkwAttachPt:
                  fAttachPoint = *(int*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
					break;

				case evpkwAttached:
                  fAttachPoint = ((*(char*)returnVal) == 't') ? true:false;
                  delete returnVal;
                  returnVal = NULL;
					break;

				case evpkwValueType:
                  fValueType = *(unsigned int*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
					break;

				case evpkwValue:
                  fValue = *(float*)returnVal;
                  delete returnVal;
					break;

				case evpkwEntityTypeList: {
                  TItemList* pList = (TItemList *)returnVal;
                  fEntType.ParseList(pList);
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
                  }
					break;

				default: // what?
                  success =  false;
					break;
			}
			break;

		case eBadKeyword:
            success =  false;
         break;

		default:
			   Assert(false, kerrReadParamfile, "TVehPart::ParseDbStructItem");
			   success = false;
   }
   return success;
}

EBoolean      TTurnParam::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fTurnParamTable, eturnSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
		case eFileNoErr:
			switch( returnIndex )
			{
				case eturnkwTime:
                  fTurnTime = (*(float*)returnVal) * kSecToSimTime;
                  delete returnVal;
                  returnVal = NULL;
					break;

				case eturnkwVelocity:
                  fNewVelocity.x = ((T2Dsingle*)returnVal)->x;
                  fNewVelocity.y = ((T2Dsingle*)returnVal)->y;
                  fNewVelocity.z = 0.0;
                  delete returnVal;
                  returnVal = NULL;
					break;
               
				default: // what?
                  success =  false;
					break;
			}
			break;

		case eBadKeyword:
            success =  false;
         break;

		default:
			   Assert(false, kerrReadParamfile, "TTurnParam::ParseDbStructItem");
			   success = false;
   }
   return success;
}
   
TTurnParam::TTurnParam(TTimeSim turnTime) 
{ 
   fTurnTime = turnTime;
}

TTurnParam::~TTurnParam()
{
}

EBoolean  cTestVehicle::ParseTestVehEntry(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fTestVehTable, etestvehSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
   case eNextLevel: {
      switch( returnIndex )
	 {
	 case etestvehkwTurnParam: {
	    TTurnParam* pNewTurn = new TTurnParam;
	    if ( pNewTurn->ReadDbStruct(infile) )
	       fTurnParam.AddParam(pNewTurn);
	    else
	       success = false;
	 }
	 break;
	 
	 case etestvehkwPartParam: {
	    if ( fpVehPartList[fNumParts++].ReadDbStruct(infile) )
	       fbHasCmPart = true;
	    else
	       success = false;
	 }
	 break;
	 
	 default: // ??
	    success =  false;
	    break;
	 }
   }
   break;
   
   
   case eFileNoErr:
      switch( returnIndex )
	 {
	 case etestvehkwStartTime:
	    fStartTime = (*(float*)returnVal) * kSecToSimTime;
	    delete returnVal;
	    returnVal = NULL;
	    break;
	    
	 case etestvehkwEndTime:
	    fEndTime = (*(float*)returnVal) * kSecToSimTime;
	    delete returnVal;
	    returnVal = NULL;
	    break;
	    
	 case etestvehkwStartLoc: {
	    TLocation tccPt;
	    tccPt.x = ((T2Dsingle*)returnVal)->x;
	    tccPt.y = ((T2Dsingle*)returnVal)->y;
	    tccPt.z = gpTerrainMgr->GetMapElev(tccPt.x, tccPt.y);
	    SetCoordTccRef(tccPt, fTccOrientation.psi, fTccOrientation.theta, fTccOrientation.phi);
	 }
	 delete returnVal;
	 returnVal = NULL;
	 break;
	 
	 case etestvehkwStartVel: {
	    TVelocity vel;
	    vel.x = ((T2Dsingle*)returnVal)->x;
	    vel.y = ((T2Dsingle*)returnVal)->y;
	    vel.z = 0.0;
	    float psi = atan2( vel.y, vel.x );
	    SetCoordTccRef(fTccLocation, psi, 0.0, kPi);
	    fEntVelocity.x = sqrt(vel.x*vel.x + vel.y*vel.y);
	    fEntVelocity.y = fEntVelocity.z = 0.0;
	    cEntity::LocalToGeoVel(fEntVelocity, &fGeoVelocity);
	    fVelocity = fGeoVelocity;
	 }
	 delete returnVal;
	 returnVal = NULL;
	 break;
	 
	 case etestvehkwProbKill:
	    fProbKill = *(float*)returnVal;
	    delete returnVal;
	    returnVal = NULL;
	    break;
	    
	 default: // what?
	    success =  false;
	    break;
	 }
      break;
      
   case eBadKeyword:
      success =  false;
      break;
      
   default:
      Assert(false, kerrReadParamfile, "cDbEntry::ParseDbEntry");
      success = false;
   }
   return success;
}

EBoolean  cTestVehicle::ReadTestVehParams(cInFileObj& infile)
   // reads each parameter into buffer, passes it to ParseTestVehEntry to 
   // interpret it.
{
	EBoolean done = false;
	EBoolean success = true;
	do {
            // NextItem() stores an item for us to look at
		EFileReturn returnCode = infile.NextItem();
		switch( returnCode ) {
			case eLevelDone:      // done with  first level, entire param set
			   	done = true;
				break;

			case eFileNoErr:
				if( !ParseTestVehEntry(infile) )
					success = false;
				break;

			default: // error getting next token
			   	Assert(false, kerrReadParamfile, "cTestVehicle::ReadTestVehParams");
				   success =  false;
		}
	} while( !done );
   
   if ( success ) {
      new cTestVehTurnMsg(this, fStartTime, false);
      fBaseTime = GetCurSimTime();
      fEndTime += fBaseTime;
      fTimeout = fEndTime;
   }
   

	return success;	
}

void      cTestVehicle::NextTurn()
{
   if ( fNextTurnParam == -1 ) {   // just getting started
      new cVehStateMsg( this, true );
      fTime = GetCurSimTime();
   }
   else {
      TVelocity vel = ((TTurnParam*)(fTurnParam.GetParam(fNextTurnParam)))->fNewVelocity;
   
      float psi = atan2( vel.y, vel.x );
      SetCoordTccRef(fTccLocation, psi, 0.0, kPi);
      fEntVelocity.x = sqrt(vel.x*vel.x + vel.y*vel.y);
      fEntVelocity.y = fEntVelocity.z = 0.0;
      cEntity::LocalToGeoVel(fEntVelocity, &fGeoVelocity);
      fVelocity = fGeoVelocity;
      cout << "turn num, time " << fNextTurnParam << " " << GetCurSimTime() << endl;
   }
   fNextTurnParam++;
   if ( fNextTurnParam < fTurnParam.GetNumParam() )
      new cTestVehTurnMsg(this, ((TTurnParam*)(fTurnParam.GetParam(fNextTurnParam)))->fTurnTime + fBaseTime, true);
}

void      cTestVehicle::UpdateVehicle()
{
   fTccLocation.z = gpTerrainMgr->GetMapElev(fTccLocation.x, fTccLocation.y);
   SetCoordTccRef(fTccLocation, fTccOrientation.psi, fTccOrientation.theta, fTccOrientation.phi);
   
   cVehicle::UpdateVehicle();
}

cTestVehicle::cTestVehicle( cDbVeh* pDbVeh)
{
   fDrAlg       = eDrFpw;
   fForce       = eEnemy;
   
   fpDbVeh = pDbVeh;
   fCmsType = fAltType = pDbVeh->GetEntryEntityType();
   fNextTurnParam = -1;   // haven't started simulating yet
   fBaseTime = 0;
   
   fpVehPartList = new TVehPart[10];
}

cTestVehicle::~cTestVehicle()
{
}
