/*
$RCSfile: AtMine.h,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/05/10 20:18:20 $
$Log: AtMine.h,v $
Revision 1.2  1999/05/10 20:18:20  ksnively
Added a command line option for enabling Communication Capability.

Revision 1.1.1.1  1999/02/16 14:54:46  jchappel
Comprehensive Mine Simulator

 * Revision 2.1  1998/02/18  17:17:24  ksnively
 * Incorporated AP mine replacement code
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.4  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.3  1997/10/02  17:47:47  austin
 * first working port to modsaf3.0
 *
Revision 1.3  1997/10/02 17:47:47  austin
first working port to modsaf3.0

 * Revision 1.2  1996/11/06  15:13:32  nsipes
 * *** empty log message ***
 *
 * Revision 1.2  1996/11/06  15:13:32  nsipes
 * *** empty log message ***
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.9  1996/06/19  19:05:45  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.9  1996/06/19  19:05:45  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.8  1996/05/21  22:12:12  rcarp
 * text and lists from mine db, and activating mine state save cmd
 *
 * Revision 1.7  1996/05/15  17:17:31  rcarp
 * refinements and fixes for 1.5 delivery
 *
 * Revision 1.6  1996/05/08  00:46:50  rcarp
 * changes for 1.5, incl user mines, mine params, param i/o, selection
 *
 * Revision 1.6  1996/05/08  00:46:50  rcarp
 * changes for 1.5, incl user mines, mine params, param i/o, selection
 *
 * Revision 1.5  1996/02/16  21:41:55  rcarp
 * revisions for getting mine wide area and directional range
 *
 * Revision 1.4  1996/02/16  19:32:41  rcarp
 * revisions for 1.1b7 version
 *
 * Revision 1.3  1996/02/05  14:32:43  rcarp
 * Updates and bug fixes of release version 1.1b5
 *
$State: Exp $
*/
 
/*******************************************

   file: AtMine.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   AtMine.h
   
      The AT mine header file defines AT mine classes,
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#ifndef _UAtMine_
#define _UAtMine_

#include "Globals.h"
#include "Minefield.h"
#include "Vehicle.h"
#include "EntityDB.h"

// ************************************
//   basic types
// ************************************

// ************************************
//   component classes
// ************************************

class cLandTracker : public cTracker {
   // tracks all land platforms (vehicles).
   //  Note that even buried pressure mines need to track all vehicles,
   //   they may carry proximity countermeasures
protected:
   EBoolean       fLifeTracker;   // tracks life forms
   EBoolean       fVehTracker;   // tracks platforms

public:
   virtual int    GetTrackType(const TEntityType& type);

            cLandTracker(cEntity* pRefEnt, float range, EBoolean useVelCoord = true,
			 EBoolean tracksVeh = true, EBoolean tracksPers = false); 
   virtual ~cLandTracker();
};

class cTrWidthControl : public cMineControl {

protected:
public:
   virtual EBoolean  FireGeometry(const cLinkVehGeom* pLink, TAttackGeom* pGeom);
   virtual EBoolean  FireGeometry(const cMineLinkEmiss* pLink, TAttackGeom* pGeom);         // target should be engaged

            cTrWidthControl(cSimMine* pMine, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt ) : 
                        cMineControl(pMine, offset, fuzeOpt) {}
   virtual ~cTrWidthControl();
};

class cFullWidthControl : public cMineControl {

protected:
   
public:
            cFullWidthControl(cSimMine* pMine, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt ) : 
                     cMineControl(pMine, offset, fuzeOpt) {}
   virtual ~cFullWidthControl();
};


// ************************************
//   mine classes
// ************************************
class cAtMine : public cSimMine {

 friend class cDbMine_cAtMine;

 private:
   static  TEntityType   fsAtMineEntityType;
   static  char*         fsAtMineName;
   static  char*         fsAtMineAltName;
   static int           fsDefaultFuzeType;   // for initializing database

   static   cDbMine_cAtMine          fsDbAtMine;      // the instantiation in mine database

 protected:
         // param input functions

public:

           // constructor for virtual mine, for component sensor
    cAtMine(cMineComp* pParent, float radius);

           // constructor for creation from gui, planning mine
    cAtMine( cPlanMine* pPlan = NULL );

           // constructor for reading from file, only used when sim host without gui
//    cAtMine(cMineComp* pParent, cDbMine* pDbMine, TMinePlaceParam& emplace, TDbOptSet opt);

   virtual   ~cAtMine();
};

class cAtConv : public cAtMine {

 friend class cDbMine_cAtConv;

 private:
   static  TEntityType   fsAtConvMineEntityType;
   static  char*         fsAtConvMineName;
   static  char*         fsAtConvMineAltName;
   static int           fsDefaultFuzeType;   // for initializing database

   static   cDbMine_cAtConv    fsDbAtConv;      // the instantiation in mine database
   
 protected:

public:
           // constructor for virtual mine, for component sensor
    cAtConv(cMineComp* pParent, float radius);

           // constructor for creation from gui, planning mine
    cAtConv( cPlanMine* pPlan = NULL );

           // constructor for reading from file, only used when sim host without gui
//    cAtConv(cMineComp* pParent, cDbMine* pDbMine, TMinePlaceParam& emplace, TDbOptSet opt);
   virtual   ~cAtConv();
};

class cAtScatterable : public cAtMine {

 friend class cDbMine_cAtScatterable;

 private:
   static  TEntityType   fsAtScatMineEntityType;
   static  char*         fsAtScatMineName;
   static  char*         fsAtScatMineAltName;
   static int           fsDefaultFuzeType;   // for initializing database

   static   cDbMine_cAtScatterable   fsDbAtScatterable;      // the instantiation in mine database
   
 protected:
         // param input functions

public:
           // constructor for virtual mine, for component sensor
    cAtScatterable(cMineComp* pParent, float radius);

           // constructor for creation from gui, planning mine
    cAtScatterable( cPlanMine* pPlan = NULL );

           // constructor for reading from file, only used when sim host without gui
//    cAtScatterable(cMineComp* pParent, cDbMine* pDbMine, TMinePlaceParam& emplace, TDbOptSet opt);

   virtual   ~cAtScatterable();
};
// ************************************
//   messages
// ************************************


#endif
