/*
$RCSfile: AtWideArea.h,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/04/29 19:34:03 $
$Log: AtWideArea.h,v $
Revision 1.2  1999/04/29 19:34:03  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:46  jchappel
Comprehensive Mine Simulator

 * Revision 2.4  1998/11/25  18:39:13  ksnively
 * Update for APLA Exercise
 *
 * Revision 2.3  1998/02/26  21:26:56  ksnively
 * Created Mf Camera object to control MAK Stealth view.
 * Created vulnerability tables specific to command detonated mines
 * and the camera object.
 *
 * Revision 2.2  1998/02/18  20:10:33  ksnively
 * Ap mine replacement bug fixes
 *
 * Revision 2.1  1998/02/18  17:17:24  ksnively
 * Incorporated AP mine replacement code
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.4  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.3  1997/10/02  17:47:47  austin
 * first working port to modsaf3.0
 *
Revision 1.3  1997/10/02 17:47:47  austin
first working port to modsaf3.0

 * Revision 1.2  1997/03/20  19:16:16  ksnively
 * Added emission tracking, updated to CMS2.0B6.
 *
 * Revision 1.2  1997/03/20  19:16:16  ksnively
 * Added emission tracking, updated to CMS2.0B6.
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.11  1996/08/09  14:33:27  rcarp
 * changed to support los check by terrain manager, and its use by offroute mines
 *
 * Revision 1.10  1996/06/22  13:04:11  rcarp
 * deton vuln bug fixes for 1.5b4
 *
 * Revision 1.9  1996/06/19  19:08:07  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.8  1996/05/21  22:12:12  rcarp
 * text and lists from mine db, and activating mine state save cmd
 *
 * Revision 1.7  1996/05/15  17:18:44  rcarp
 * refinements and fixes for 1.5 delivery
 *
 * Revision 1.6  1996/05/08  00:47:57  rcarp
 * changes for 1.5, incl user mines, mine params, param i/o, selection
 *
 * Revision 1.6  1996/05/08  00:47:57  rcarp
 * changes for 1.5, incl user mines, mine params, param i/o, selection
 *
 * Revision 1.5  1996/02/16  21:41:55  rcarp
 * revisions for getting mine wide area and directional range
 *
 * Revision 1.4  1996/02/16  19:32:41  rcarp
 * revisions for 1.1b7 version
 *
 * Revision 1.3  1996/02/05  14:32:43  rcarp
 * Updates and bug fixes of release version 1.1b5
 *
$State: Exp $
*/
 
/*******************************************

   file: AtWideArea.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   AtWideArea.h
   
      The AT wide area mine header file defines wide area AT mine classes,
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/


#ifndef _UAtWideAreaMine_
#define _UAtWideAreaMine_

#include "Globals.h"
#include "Minefield.h"
#include "AtMine.h"
#include "Vehicle.h"
#include "EntityDB.h"

// ************************************
//   basic types
// ************************************

enum EProjectileIntState {  eprojisNone = 0,
                            eprojisFailed,
                            eprojisHitGround,
                            eprojisFiredMissed,   // those that fire, eg WAM sublet
                            eprojisFiredHit,
                            eprojisFlying,
                               eprojisLowestUpdate  = eprojisFlying,
                               eprojisLowestTracker = eprojisFlying,
                               eprojisLowestControl = eprojisFlying,
                               
                            eprojisNumStates
};

struct TFuzeParamOffRoute : public TFuzeParam {

   float         fMinRange;
   float         fMaxRange;
   float         fMinTgtSpeed;
   float         fMaxTgtSpeed;
   float         fDetectProb;
   float         fHitProb;
   float         fSpeed;   // both launch and impact

	enum {
		eorfuzekwFuzeTypeName,      // intercept type name, donot allow unuseable types
      eorfuzekwFireOffset,         // intercept tgtoffset, same as range gate here
      eorfuzekwRangeGate,
		eorfuzekwSpeedGate,
		eorfuzekwDetectProb,
		eorfuzekwHitProb,
		eorfuzekwMissileSpeed,

		eorfuzeSizeTable
	};
 
   static const TKeywordEntry fKeyOrFuzeTable[eorfuzeSizeTable];

   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);

   virtual   cControl*     CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt );

   virtual EBoolean   IsDirectional(TFlatEntCoord* pMinMax);
   
   virtual   TFuzeParam*   CloneFuzeParam();
   
            TFuzeParamOffRoute();
   virtual ~TFuzeParamOffRoute();
};

class cProjectile;
struct TFuzeParamWamSublet : public TFuzeParam {
   float             fConeAngle;
   float             fScanDetProb;
   float             fHitProb;

	enum {
		esubfuzekwFuzeTypeName,      // intercept type name, donot allow unuseable types
      esubfuzekwConeAngle,
      esubfuzekwScanDetProb,
      esubfuzekwHitProb,

		esubfuzeSizeTable
	};
 
   static const TKeywordEntry fKeySubletFuzeTable[esubfuzeSizeTable];

   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);
   virtual void           WriteFuzeOpt(ofstream* file);

   virtual cControl*    CreateControl(cProjectile* pProj, const TFuzeOpt& fuzeOpt );
   
   virtual   TFuzeParam*   CloneFuzeParam();
   
            TFuzeParamWamSublet();
   virtual ~TFuzeParamWamSublet();
};

struct TMineParamWam : public TMineParam {
   
   cDbProj*            fpDbProj;      // type of projectile to launch

	enum {
		empwamkwProjType,
      
		eSizeMineParamWamTable
	};
   
	static const TKeywordEntry fKeyMineParamWamTable[eSizeMineParamWamTable];
   
 public:
   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);
   virtual void         WriteMineOpt(ofstream* file);
   
   virtual TMineParam*   CloneMineParam()   { return new TMineParamWam(*this); }
   virtual cProjectile*  CreateProjectile()  { return fpDbProj->CreateProj(); }
   
            TMineParamWam(TEntityType& type, int projType = eprojtypeWamSublet);
            TMineParamWam(int projType = eprojtypeWamSublet);
   virtual ~TMineParamWam();
};

struct TFuzeParamWideArea : public TFuzeParam {
   float               fSelfDefenseRange;
   float               fRangeErrMax;
   float               fVelocityErrMax;
   float               fAzimuthErrMax;
   float               fTypeErrProb;
   float               fTypeErrMult;
   float               fFireWindow;

	enum {
		ewafuzekwFuzeTypeName,      // intercept type name, do not allow unuseable types
      ewafuzekwSelfDefRange,
		ewafuzekwRangeErrMax,
		ewafuzekwVelErrMax,
		ewafuzekwAzErrMax,
		ewafuzekwProbTypeErr,
		ewafuzekwTypeErrMult,
		ewafuzekwFireWindow,

		ewafuzeSizeTable
	};
 
   static const TKeywordEntry fKeyWaFuzeTable[ewafuzeSizeTable];

   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);

   virtual EBoolean   IsWideArea(float* pRange);

   virtual cControl*    CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt );
   
   virtual   TFuzeParam*   CloneFuzeParam();
   
            TFuzeParamWideArea();
   virtual ~TFuzeParamWideArea();
};

class cLinkVehErr : public cLinkVehGeom
      // tracking with errors
{

 protected:
   
      // errors
   float         fRangeError;      // multiplier of true range
   float         fVelError;        // multiplier of true velocity
   float         fAzimuthError;    // added to true az
   int           fTgtType;         // might be wrong
   
  
 public:
      virtual void   UpdateTrkRec();
      
              float  GetRangeErr()   { return fRangeError; }
              float  GetVelErr()   { return fVelError; }
              float  GetAzErr()   { return fAzimuthError; }
              int    GetTgtType()   { return fTgtType; }
	      void   SetTgtType(int newTgtType)  { fTgtType = newTgtType; }

                 cLinkVehErr(cLinkVehErr* pTrkVeh, cLinkVeh* pPrev);
                 cLinkVehErr(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev, cEntity* pRefEnt, 
      EBoolean refInvert, int tgtType, float rangeErr, float velErr, float azErr);
      virtual   ~cLinkVehErr();
};

class cLinkVehSublet : public cLinkVehGeom
    // used by sublets that compare current and last elevation angles
{

 protected:
   float         fLastElevation;

 public:
   float         GetLastElevation()  const  { return fLastElevation; } 
      
                 cLinkVehSublet(cLinkVehSublet* pTrkVeh, cLinkVeh* pPrev);
                 cLinkVehSublet(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev, 
                                cEntity* pRefEnt, EBoolean refTgtVel = false);
      virtual   ~cLinkVehSublet();

};

class cLinkVehCross : public cLinkVehGeom
    // used by off-route mines to determine if target crosses x-axis
{

 protected:
   float         fLastY;   // saves last y coordinate value
   float         fLastAz;   // saves last azimuth
   float         fLastRange;

 public:
   float         GetLastY()  const  { return fLastY; } 
   float         GetLastAz()  const  { return fLastAz; } 
   
   T2Dpolar      GetLastLocPolar()  
                     { T2Dpolar pLoc(fLastRange, fLastAz); return pLoc; }
      
                 cLinkVehCross(cLinkVehCross* pTrkVeh, cLinkVeh* pPrev);
                 cLinkVehCross(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev, 
                                cEntity* pRefEnt, EBoolean refTgtVel = false);
      virtual   ~cLinkVehCross();

};

// ************************************
//   component classes
// ************************************

class cWideAreaTracker : public cLandTracker {
   // tracks all land platforms (vehicles), adds errors and vehicle categories
protected:

public:
   virtual   int            GetTrackType(const TEntityType& type);

   virtual cLinkVehGeom*   CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev);
   virtual cLinkVehGeom*   CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev);

            cWideAreaTracker(cEntity* pRefEnt, float range, EBoolean useVelCoord = true) : 
                     cLandTracker(pRefEnt, range, useVelCoord) {}
   virtual ~cWideAreaTracker();
};

class cSubletTracker : public cLandTracker {

 public:
   virtual   int           GetTrackType(const TEntityType& type);

   virtual cLinkVehGeom*   CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev);
   virtual cLinkVehGeom*   CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev);

            cSubletTracker(cEntity* pRefEnt, float range, EBoolean useVelCoord = false) : 
                     cLandTracker(pRefEnt, range, useVelCoord) {}
   virtual ~cSubletTracker();
};

class cOffRouteTracker : public cLandTracker {

 public:
   virtual   int           GetTrackType(const TEntityType& type);

   virtual cLinkVehGeom*   CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev);
   virtual cLinkVehGeom*   CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev);

            cOffRouteTracker(cEntity* pRefEnt, float range, EBoolean useVelCoord = false) : 
                     cLandTracker(pRefEnt, range, useVelCoord) {}
   virtual ~cOffRouteTracker();
};

class cWamControl : public cMineControl {

protected:
   float         fSelfDefenseRange;
   
public:
   virtual void    UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList,  TTimeSim* pWaitTime);

             EBoolean  TgtGeometry( cLinkVehGeom* pLink,  const cMineLinkEmiss* pEmLink);
             void  SetupForFire(cLinkVehErr* pLink, cMineLinkEmiss* pEmLink,  TTimeSim* pWait);
             void  SelfDefenseFire(cLinkVehErr* pLink,  cMineLinkEmiss* pEmLink = NULL);

            cWamControl(cSimMine* pMine, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt ) : 
                             cMineControl(pMine, offset, fuzeOpt) { }
   virtual ~cWamControl();
};

class cProjectile;
class cWamSubletControl : public cControl {

protected:
   cProjectile*      fpProjectile;
   float             fConeAngle;
   float             fScanDetProb;
   float             fHitProb;
   
public:
   virtual   void      UpdateControl(cVehListHead*, cEmissListHead*, TTimeSim*);

            cWamSubletControl(cProjectile* pProj, float coneAng,
                              float detProb, float hitProb)
                { fpProjectile = pProj;   fConeAngle = coneAng;
                  fScanDetProb = detProb; fHitProb = hitProb; }
   virtual ~cWamSubletControl();
};

class cOffRouteControl : public cMineControl {

protected:
   float      fDetectProb;
   float      fHitProb;
   float      fMinTgtSpeed;
   float      fMaxTgtSpeed;
   
      // these should be added to mine or fuze params, allow user input
   static   double   fsHeight;
   static   double   fsMinVis;
   
   
             EBoolean   AttackGeomFilter(cLinkVehCross*);
             EBoolean   AttackGeomFilter(cMineLinkEmiss*);
   
public:
   virtual void    UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList, TTimeSim* pWaitTime);

            cOffRouteControl(cSimMine* pMine, TFlatEntCoord fireOffset, const TFuzeOpt& fuzeOpt, 
                            float probDet, float probHit, float minSpeed, float maxSpeed) : 
                             cMineControl(pMine, fireOffset, fuzeOpt) 
                             { fDetectProb = probDet; fHitProb = probHit;
                               fMinTgtSpeed = minSpeed; fMaxTgtSpeed = maxSpeed;}
   virtual ~cOffRouteControl();
};

// ************************************
//   mine classes
// ************************************

class cProjectile : public cSimVehicle {

 public:
      // all munition types use 0 for not tracked, other values for tracked
   enum { etrkNot, etrkYes };

 protected:
   static TEntityType   fsProjEntityType;
   static char*         fsProjName;
   static int           fsDefaultFuzeType;   // for initializing database
   
   static   cDbProj_cProj     fsDbProj;      // the instantiation in database

   static   TTimeSim     fsDamageCheckDelay;
   static   char*        fsStateText[eprojisNumStates];
   
   cDbProj*              fpDbProj;      // points to the proj db object

   cTracker*             fpTracker;
   cControl*             fpControl;

   cSimMine*             fpFiringMine;
   TEntityID             fEventID;      // event that launched it
   EProjectileIntState   feIntState;
   TTimeSim              fLaunchTime;
   TLocation             fLaunchTccLoc;
   TLocation             fLaunchGeoLoc;
   TEntCoord             fLocRelLaunch;
 
      // info filled in during simulation, mostly at detonation
   cVehicle*             fpTarget;
   int                   fKilledTgts;
   EBoolean              fbKilledTgt;
   TEntityType           fTargetType;
   TEntityID             fTargetID;
   TLocation             fDetonLoc;      // may not be mine location
   TEntCoord             fLocTargetRef;
   EDetonResult          feDetonResult;
   float                 fFireRange;
 
 public:
   void                 SetTracker(cTracker* pTracker) { fpTracker = pTracker; }
   void                 SetControl(cControl* pControl) { fpControl = pControl; }
 
   TEntityID            GetEventID()   { return fEventID; }
   EProjectileIntState  GetIntState()  { return feIntState; }
   
   void                 SetIntState(EProjectileIntState state) 
                                       { feIntState = state; }
   
   virtual   EWarhead    GetWarhead();
   virtual   EFuze       GetFuze();
   virtual   float       GetHitProb();

   cVehicle*             GetTarget() { return fpTarget; }
   virtual   TEntityType GetDetonationType();
   TEntityType           GetTargetType() { return fTargetType; }
   TEntityID             GetTargetID() { return fTargetID; }
   TEntCoord             GetDetVehCoord() { return fLocTargetRef; }
   TLocation             GetDetLoc() { return fDetonLoc; }
   EDetonResult          GetDetResult() { return feDetonResult; }
   float                 GetFireRange() { return fFireRange; }
   
   static char*          GetInternalStateText(EProjectileIntState state) { return fsStateText[state]; }

   void                  SetDetonationResults(cLinkVehGeom* pLink);
//   void                  SetTarget(cVehicle* veh) {fpTarget = veh;}
//   void                  SetTargetType(const TEntityType& type) {fTargetType = type;}
//   void                  SetTargetID(const TEntityID& id) {fTargetID = id;}
//   void                  SetTargetRefLoc(const TEntCoord& loc) {fLocTargetRef = loc;}
//   void                  SetDetonLoc(const TLocation& loc) {fDetonLoc = loc;}
   void                  SetDetonResult(EDetonResult result) {feDetonResult = result;}
//   void                  SetFireRange(float range) { fFireRange = range; }

                  cProjectile(cDbProj* pDbProj);
   virtual       ~cProjectile();
};

class cAtWideArea;
class cWamSublet : public cProjectile {

 private:
   static TEntityType   fsWamSubEntityType;
   static char*         fsWamSubletName;
   static int           fsDefaultFuzeType;   // for initializing database
   
   static   cDbProj_Ballistic     fsDbWamSublet;      // the instantiation in database
   
 protected:
   static const   TTimeSim   fsUpdateRate;
   
      // two veh lists, master from mine and tracked each update
   cVehListHead          fMasterVeh;
   cVehListHead          fTrackVeh;
   cEmissListHead        fEmissList;
//   cGenericWam*          fpMine;

 public:
   virtual   char*         GetEntName();
   virtual   EWarhead      GetWarhead();
   virtual   EFuze         GetFuze();
   virtual   TEntityType   GetDetonationType();
   virtual   float         GetHitProb();
   
   const     T2Dsingle&  GetLaunchVel() { return (fpDbProj->GetProjParam())->fLaunchVel; }
 
   virtual   void        UpdateVehicle();
   
   virtual   void        UpdateLocation();
         
   virtual cInteraction*  CreateNewInteraction(cLinkVehGeom*);
   virtual void           UpdateInteractions();
   virtual   void        WriteEntSumm(ofstream*);
   
             void        InitWamSublet(cAtWideArea* pMine, float tccAimAz);
 
               cWamSublet(cDbProj* pDbProj);
   virtual    ~cWamSublet();
};

class cAtOffRoute : public cAtMine {

 friend class cDbMine_cAtOffRoute;

 public:
   enum { etrkNot, etrkYes, etrkNumTrkTypes };

 private:
   static  TEntityType   fsAtOffrouteMineEntityType;
   static  char*         fsAtOffrouteMineName;
   static  char*         fsAtOffrouteMineAltName;
   static int           fsDefaultFuzeType;   // for initializing database

   static  cDbMine_cAtOffRoute          fsDbAtOffRoute;      // the instantiation in mine database
   
 protected:

 public:
   virtual  void   FireProjectile(const cLinkVehGeom* pLink);
   virtual  void   SetDetonationResults(cLinkVehGeom* pLink, TVehPart* pPart = NULL, EBoolean all = false);
   virtual  void   SetDetonationResults(cMineLinkEmiss* pEmLink, EBoolean all = false);
   
   float      GetHitProb();
   float      GetLaunchSpeed();
   float      GetMaxProjRange();
   float      GetMinProjRange();


           // constructor for virtual mine, for component sensor
    cAtOffRoute(cMineComp* pParent, float radius);

           // constructor for creation from gui, planning mine
    cAtOffRoute( cPlanMine* pPlan = NULL );

           // constructor for reading from file, only used when sim host without gui
//    cAtOffRoute(cMineComp* pParent, cDbMine* pDbMine, TMinePlaceParam& emplace, TDbOptSet opt);

   virtual   ~cAtOffRoute();
};

class cAtWideArea : public cAtMine {

 friend class cDbMine_cAtWideArea;

 public:
   enum { etrkNot, etrkORSMC,  etrkTank, etrkOtherHeavyTracked, etrkHeavyWheel, 
          etrkLightTracked, etrkLightWheel, etrkNumTrkTypes };

 private:
   static  TEntityType   fsAtWideareaMineEntityType;
   static  char*         fsAtWideareaMineName;
   static  char*         fsAtWideareaMineAltName;
   static  int           fsDefaultFuzeType;   // for initializing database

   static   cDbMine_cAtWideArea      fsDbAtWideArea;      // the instantiation in mine database
   
 protected:
         
   float            fFireAzimuth;

public:

          float        GetSelfDefenseRange() 
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fSelfDefenseRange; }
          float        GetTypeErrProb()
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fTypeErrProb; }
          float        GetTypeErrMult()
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fTypeErrMult; }
          float        GetRangeErrMax() 
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fRangeErrMax; }
          float        GetVelErrMax()
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fVelocityErrMax; }
          float        GetAzErrMax() 
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fAzimuthErrMax; }
          float        GetFireWindow()
      { return ((TFuzeParamWideArea*)(fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fFireWindow; }
   const  T2Dsingle&   GetSubletLaunchVel()  
      { return ((TMineParamWam*)(fpDbMine->GetMineParam(fOptSet.fMineOpt)))->fpDbProj->GetProjParam()->fLaunchVel; }
         
                 float          GetFireAz()  { return fFireAzimuth; }

   virtual    void   SetDetonationResults(const cLinkVehGeom* pLink,const cMineLinkEmiss* pEmLink,  
                     cProjectile* pProj, float fireAz, EBoolean all = false);
   virtual    void   SetProjectileResults(int); 
              void   WriteTgtType(int type, ofstream* pOut);
              
   cProjectile*      CreateProjectile() { return fpDbMine->CreateProjectile(fOptSet.fMineOpt); }

   virtual cInteraction* CreateNewInteraction(cLinkVehGeom*);
          
           // constructor for virtual mine, for component sensor
    cAtWideArea(cMineComp* pParent, float radius);

           // constructor for creation from gui, planning mine
    cAtWideArea( cPlanMine* pPlan = NULL );

   virtual   ~cAtWideArea();
};

// ************************************
//   messages
// ************************************

class cFireProjectileMsg : public cMsgPrEvent {

protected:
   TFireData&                 fFireData;
   
public:
      void        SendMsg();

               cFireProjectileMsg(TFireData& fireData):
		  fFireData(fireData)
                    { }
   virtual    ~cFireProjectileMsg(); 
};

class cUpdateProjectileMsg : public cMsgPrUpdt {

protected:
   cProjectile*     fpProjectile;
   
public:
      void        SendMsg();

               cUpdateProjectileMsg(TTimeSim wait, cProjectile* pProjectile) :
                  cMsgPrUpdt(wait)      { fpProjectile = pProjectile; }
   virtual    ~cUpdateProjectileMsg(); 
};

class cProjFireMunitionMsg : public cMsgPrEvent {

protected:
   cProjectile*   fpProjectile;
   
public:
      void        SendMsg();

               cProjFireMunitionMsg(cProjectile* pProjectile)
                  { fpProjectile = pProjectile; }
   virtual    ~cProjFireMunitionMsg(); 
};

class cProjDetonationMsg : public cMsgPrEvent {

protected:
   cProjectile*   fpProjectile;
   float          fBlastSize;
   TTimeSim       fDelay;
   
public:
      void        SendMsg();

               cProjDetonationMsg(cProjectile* pProjectile, 
                           float blastSize = 5.0, TTimeSim delay = 3000 )
                  { fpProjectile = pProjectile; 
                    fBlastSize   = blastSize;
                    fDelay       = delay;   }
   virtual    ~cProjDetonationMsg(); 
};

#endif
