/*
$RCSfile: CompStatusCB.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:29 $
$Log: CompStatusCB.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:29  jchappel
Comprehensive Mine Simulator

// Revision 2.2  1998/10/27  19:34:43  jchappel
// modifications for versioning
//
// Revision 2.2  1998/10/27  19:34:43  jchappel
// modifications for versioning
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.4  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/05/20  15:35:11  ksnively
// Redraw display when turning on or off a component.
//
// Revision 1.2  1997/05/20  15:35:11  ksnively
// Redraw display when turning on or off a component.
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.7  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.6  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.4  1996/03/23  14:26:12  rcarp
// improvements to test mode and selection
//
// Revision 1.3  1996/03/13  01:09:55  mkappel
// Move CompStatus button.
//
// Revision 1.2  1996/03/06  19:45:05  mkappel
// New Mine Status Window
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: CompStatusCB.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   CompStatusCB.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include "Minefield.h"
#include <Xm/List.h>
#include <Xm/RowColumn.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>

void cComponentStatusWindow::CbScroll (Widget sbar, XtPointer /*wForm*/, 
				       XmScrollBarCallbackStruct *cbs)
{
  cMinefield*    pCurMinefield = gpGuiWin->GetCurMinefield();
  Assert( pCurMinefield != NULL, kerrValidityCheck, "cComponentStatusWindow::CbScroll");
  cMfComponent*  pComp = (cMfComponent*)pCurMinefield->GetPerimeter();
  int            count = 0;
  Widget         wMainWin, wPrev;

  if (pComp == NULL) return;
  pComp = pComp->GetNext(); /* skip perimeter */

  while (pComp != NULL) {
    wMainWin = pComp->GetCompStatus()->GetMainWin();
    if (count < cbs->value)
      XtUnmanageChild (wMainWin);
    else {
      XtManageChild (wMainWin);
      if (count == cbs->value)
	XtVaSetValues (wMainWin,
		       XmNtopAttachment, XmATTACH_FORM,
		       NULL);
      else
	XtVaSetValues (wMainWin,
		       XmNtopAttachment, XmATTACH_WIDGET,
		       XmNtopWidget, wPrev,
		       NULL);
      wPrev = wMainWin;
    }
    
    count++;
    pComp = pComp->GetNext();
  }

  XtVaSetValues (sbar, XmNmaximum, (count > 1) ? count : 2, NULL);
}

void  cCompStatusLine::CbSelect(Widget /*w*/, XtPointer pObj, 
				XmPushButtonCallbackStruct */*cbs*/)
{
   cMfComponent*  pComp = ((cCompStatusLine*)pObj)->fpComp;

   gpGuiWin->SetSelection( false, pComp->GetMinefield(), pComp );
}

void cComponentStatusWindow::CbStatus (Widget /*w*/, XtPointer  /*pObj*/,
					XmPushButtonCallbackStruct */*cbs*/)
{
   if (gpGuiWin->GetCurComponent()) {
      tactmap_bundle_changes (gpTactmap);
      cMfComponent*  pComp = (cMfComponent*)fpMinefield->GetPerimeter();
      if (pComp) {
         pComp = pComp->GetNext(); /* skip perimeter */
   
         while (pComp != NULL) {
            pComp->SetCompPvdSelect( true );
            pComp = pComp->GetNext();
         }
      }
      tactmap_refresh_non_terrain (gpTactmap);
      gpGuiWin->ShowEditor(cGuiWindow::eEditMineStatus);

   }
   /*
   if (gpGuiWin->GetCurComponent())
     gpGuiWin->ShowEditor(cGuiWindow::eEditMineStatus);
     */
}

void cComponentStatusWindow::CbDone (Widget /*w*/, XtPointer  /*pObj*/,
					XmPushButtonCallbackStruct */*cbs*/)
{
   tactmap_bundle_changes (gpTactmap);
   cMfComponent*  pComp = (cMfComponent*)fpMinefield->GetPerimeter();
   if (pComp) {
      pComp = pComp->GetNext(); /* skip perimeter */

      while (pComp != NULL) {
         pComp->SetCompPvdSelect( true );
         pComp = pComp->GetNext();
      }
   }

   gpGuiWin->SetSelection( false, fpMinefield );
   tactmap_refresh_non_terrain (gpTactmap);
   fpMinefield = NULL;

   gpGuiWin->ShowEditor(cGuiWindow::eEditMfStatus);
}

void cComponentStatusWindow::CbCancel (Widget /*w*/, XtPointer  /*pObj*/,
					XmPushButtonCallbackStruct */*cbs*/)
{
  cMfComponent*  pComp = (cMfComponent*)fpMinefield->GetPerimeter();
  if (pComp) {
     pComp = pComp->GetNext(); /* skip perimeter */

     while (pComp != NULL) {
        pComp->SetCompPvdSelect( true );
        pComp = pComp->GetNext();
     }
  }

  gpGuiWin->ShowEditor(cGuiWindow::eEditMfStatus);
}

/*  this code is not necessary
void cComponentStatusWindow::SetCmdCompSens( cMinefield* pMf, EBoolean sens )
{

  // set the sensitivity of command detonated mines

  cMfComponent*  pComp = (cMfComponent*)pMf->GetPerimeter();

  if (pComp) {
  pComp = pComp->GetNext(); // skip perimeter 
 
    while (pComp != NULL) {
      pComp->SetCompPvdSelect( sens );
      pComp = pComp->GetNext();
    }
  }

  if (gbGui) tactmap_refresh_non_terrain (gpTactmap);

}
*/
void cComponentStatusWindow::ShowEdWindow()
{
  fpMinefield = gpGuiWin->GetCurMinefield();
  Assert( fpMinefield != NULL, kerrValidityCheck, "cComponentStatusWindow::ShowEdWindow");
  cMfComponent*  pComp = (cMfComponent*)fpMinefield->GetPerimeter();
  
  fpMinefield->GetCompStatus()->ShowInfo();

  if (pComp) {
    pComp = pComp->GetNext(); /* skip perimeter */

    while (pComp != NULL) {
      pComp->GetCompStatus()->ShowInfo();
      pComp->SetCompPvdSelect( true );
      pComp = pComp->GetNext();
    }
  }
  if (gbGui) tactmap_refresh_non_terrain (gpTactmap);

  XtManageChild (fwMainWin);
}

void    cComponentStatusWindow::ClearEdWindow()
{
   cEditorWindow::ClearEdWindow();
   
//   cMfComponent*  pComp = (cMfComponent*)fpMinefield->GetPerimeter();
//   if (pComp) {
//      pComp = pComp->GetNext(); /* skip perimeter */

//      while (pComp != NULL) {
//         pComp->SetCompPvdSelect( false );
//         pComp = pComp->GetNext();
//      }
//   }
   fpMinefield = NULL;
   if (gbGui) tactmap_refresh_non_terrain (gpTactmap);
}

void cCompStatusLine::CbSimToggle (Widget w, XtPointer pObj, 
				  XmToggleButtonCallbackStruct */*cbs*/)
{
  cCompStatusLine *pCompStatus = (cCompStatusLine*)pObj;

  if (XmToggleButtonGetState (w)) {
    ((cMineComp*)pCompStatus->fpComp)->SetSimState 
      ((EBoolean)XmToggleButtonGetState (w));
    XtSetSensitive (w, FALSE);  /* once simulating, always simulating */
  }
}

void cCompStatusLine::CbOnToggle (Widget w, XtPointer pObj, 
				 XmToggleButtonCallbackStruct */*cbs*/)
{
  cCompStatusLine *pCompStatus = (cCompStatusLine*)pObj;
  cMinefield* pCompParent = pCompStatus->fpComp->GetMinefield();
  EBoolean show = pCompParent->GetShowMines();

  pCompParent->ShowMines(false);
  ((cMineComp*)pCompStatus->fpComp)->SetOnState((EBoolean)XmToggleButtonGetState (w));
  pCompParent->ShowMines(show);
}

void cCompStatusLine::ShowInfo ()
{
  char        str[20];
  XmString    xmstr;
  cMineComp*  pComp = (cMineComp*)fpComp;

  XtVaSetValues (fwSimToggle,
		 XmNset, pComp->GetSimState(),
		 XmNsensitive, ! pComp->GetSimState(),
		 NULL);

  XtVaSetValues (fwOnToggle,
		 XmNset, pComp->GetOnState(),
		 NULL);

  sprintf (str, "%d", pComp->GetNumMines());
  xmstr = XmStringCreateSimple (str);
  XtVaSetValues (fwTotalMines,
		 XmNlabelString, xmstr,
		 NULL);

   sprintf (str, "%d", pComp->GetLiveMines());
   xmstr = XmStringCreateSimple (str);
   XtVaSetValues (fwLiveMines,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree(xmstr);

   sprintf (str, "%d", pComp->GetFiredMines());
   xmstr = XmStringCreateSimple (str);
   XtVaSetValues (fwFiredMines,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree(xmstr);

   sprintf (str, "%d", pComp->GetOtherMines());
   xmstr = XmStringCreateSimple (str);
   XtVaSetValues (fwOtherMines,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree(xmstr);

   sprintf (str, "%d", pComp->GetVehHits());
   xmstr = XmStringCreateSimple (str);
   XtVaSetValues (fwVehHits,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree(xmstr);

   sprintf (str, "%d", pComp->GetVehKills());
   xmstr = XmStringCreateSimple (str);
   XtVaSetValues (fwVehKills,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree(xmstr);

  if (pComp->GetDbEntity())
     xmstr = XmStringCreateSimple ((char*)pComp->GetEntType().
           GetMunitionDomainText());
  else
    xmstr = XmStringCreateSimple ("");
  XtVaSetValues (fwDomain,
		 XmNlabelString, xmstr,
		 NULL);
  XmStringFree (xmstr);
  
  if (pComp->GetDbEntity())
    xmstr = XmStringCreateSimple ((char*)pComp->GetDbEntity()->GetEntryEntityName());
  else
    xmstr = XmStringCreateSimple ("");
  XtVaSetValues (fwMineType,
		 XmNlabelString, xmstr,
		 NULL);
  XmStringFree (xmstr);
}

void   cCompStatusLine::SwitchHilite()
{
   Pixel   backColor, foreColor;
   
   XtVaGetValues( fwTypeText,
                  XmNbackground, &backColor,
                  XmNforeground, &foreColor, 
                  NULL);
                  
   XtVaSetValues( fwTypeText,
                  XmNbackground, foreColor,
                  XmNforeground, backColor, 
                  NULL);
}

void cCompStatusArea::ShowInfo ()
{
   XmString  xmstr;

   cMinefield*    pCurMinefield = gpGuiWin->GetCurMinefield();
   Assert( pCurMinefield != NULL, kerrValidityCheck, "cCompStatusArea::ShowInfo");

   XtVaSetValues (fwMinefieldName,
		  XmNvalue, pCurMinefield->GetMfName(),
		  NULL);

   switch (pCurMinefield->GetMfType()) {

   case cMinefield::eDeliberate:
     xmstr = XmStringCreateSimple ("Deliberate");
     break;
   case cMinefield::eHasty:
     xmstr = XmStringCreateSimple ("Hasty");
     break;
   case cMinefield::eTactical:
     xmstr = XmStringCreateSimple ("Tactical");
     break;
   }

   XtVaSetValues (fwTypeLabel,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   switch (pCurMinefield->GetForce()) {

   case eFriendly:
     xmstr = XmStringCreateSimple ("Friendly");
     break;
   case eEnemy:
     xmstr = XmStringCreateSimple ("Enemy");
     break;
   default:
     xmstr = XmStringCreateSimple ("Other");
     break;
   }

   XtVaSetValues (fwForceLabel,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   if (pCurMinefield->GetOnState())
     xmstr = XmStringCreateSimple ("On");
   else
     xmstr = XmStringCreateSimple ("Off");

   XtVaSetValues (fwStateLabel,
		  XmNlabelString, xmstr,
		  NULL);
   XmStringFree (xmstr);

   XtManageChild (fwMainWin);
}

