/*
$RCSfile: Component.cc,v $
$Revision: 1.3 $
$Author: ksnively $
$Date: 1999/05/18 19:33:16 $
$Log: Component.cc,v $
Revision 1.3  1999/05/18 19:33:16  ksnively
Ported CMS to compile under IRIX 6.5
Altered make files to allow cms to be compiled under various OS simulataneously
Removed the Startup GUI

Revision 1.2  1999/04/29 19:34:05  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:30  jchappel
Comprehensive Mine Simulator

// Revision 2.3  1998/11/25  18:37:51  ksnively
// Update for APLA Exercise
//
// Revision 2.2  1998/04/21  20:37:34  ksnively
// Corrected PointInside algorithm
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  18:41:50  ksnively
// eliminated warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.5  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.4  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.4  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/05/20  15:35:11  ksnively
// Redraw display when turning on or off a component.
//
// Revision 1.3  1997/05/20  15:35:11  ksnively
// Redraw display when turning on or off a component.
//
// Revision 1.2  1997/03/20  19:14:41  ksnively
// added emission tracking, updated to CMS2.0B6
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.23  1996/09/06  18:57:57  rcarp
// CP bug fixes
//
// Revision 1.22  1996/08/21  21:58:20  rcarp
// CP and QRP file output, other updates
//
// Revision 1.21  1996/08/07  14:19:30  rcarp
// initial revisions for CP protocol
//
// Revision 1.20  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.19  1996/06/22  13:04:11  rcarp
// deton vuln bug fixes for 1.5b4
//
// Revision 1.18  1996/06/21  17:04:20  rcarp
// upgrades for release 1.5b3, mostly mine vuln
//
// Revision 1.17  1996/06/19  19:05:45  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.16  1996/05/21  22:15:40  rcarp
// text and lists from mine db, and activating mine state save cmd
//
// Revision 1.15  1996/05/16  00:01:58  rcarp
// miscellaneous bugs
//
// Revision 1.14  1996/05/15  19:28:58  mkappel
// Restore changes.
//
// Revision 1.13  1996/05/15  18:10:10  rcarp
// grid and row emplacement allocation bug
//
// Revision 1.12  1996/05/15  00:11:09  rcarp
// mostly wam and offroute fixes, also deleting selection
//
// Revision 1.11  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.9  1996/05/11  20:29:26  rcarp
// bug fixes
//
// Revision 1.8  1996/05/08  00:49:35  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.7  1996/04/04  16:25:34  mkappel
// Improve performance with large mine sets.
//
// Revision 1.6  1996/03/28  22:34:41  mkappel
// Add Mine Display toggle to component editors.
//
// Revision 1.5  1996/03/23  14:22:31  rcarp
// improvements to test mode and selection
//
// Revision 1.4  1996/03/06  19:45:05  mkappel
// New Mine Status Window
//
// Revision 1.3  1996/02/16  19:32:41  rcarp
// revisions for 1.1b7 version
//
// Revision 1.2  1996/02/05  14:33:28  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: Component.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   Component.cc
   
      The minefield implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "Minefield.h"
#include "CmsControl.h"
#include "AtMine.h"
#include "FileObj.h"
#include "MapGeometry.h"
#include "Message.h"
#include "NetMgr.h"
#include "McmPdu.h"
#include "Display.h"

#include <iostream.h>
#include <stdlib.h>
#include <math.h>

const TKeywordEntry cMineComp::fKeyMCompTable[eSizeMCompTable] =
{
	"type",         ekeyIntList,
	"typestring",   ekeyQuotedString,
	"depth",        ekeyFloat,
	"inPlaceHours", ekeyFloat,
	"fuzeType",     ekeyString,
	"selfDestTime", ekeyFloat,
	"tgtCount",     ekeyInt,
	"antiHandling", ekeyChar,
	"blastHard",    ekeyChar,
   "locations",    ekey2DPointList,     // in minefield coords
   "3dlocations",  ekey3DPointList,     // in minefield coords
   "orientations", ekey2DPointList      // map coord sys ref
};

const TKeywordEntry cIndivComp::fKeyIndivTable[ 1 /* eSizeGridTable */] =
{
	"filler",      ekeyFloat
};

const TKeywordEntry cGridComp::fKeyGridTable[eSizeGridTable] =
{
	"spacing",     ekeyFloat
};

const TKeywordEntry cRowComp::fKeyRowTable[eSizeRowTable] =
{
	"spacing",     ekeyFloat,
	"offset",      ekeyFloat,
	"points",      ekey2DPointList      // in minefield coords
};

// ************************************
//   messages
// ************************************

void    cUpdateComponentMsg::SendMsg()
{
   fpComponent->UpdateComponent();
}

cUpdateComponentMsg::~cUpdateComponentMsg()
{} 

void    cReadGridMsg::SendMsg()
{
   fpGrid->ReadGridMines(fpMineFile);
}

cReadGridMsg::cReadGridMsg(cGridComp* pGrid, cInFileObj* pMineFile)
{
   fpGrid = pGrid;
   fpMineFile = pMineFile;
}
 
cReadGridMsg::~cReadGridMsg()
{} 

void    cReadIndivMsg::SendMsg()
{
   fpIndiv->ReadIndivMines(fpMineFile);
}

cReadIndivMsg::cReadIndivMsg(cIndivComp* pIndiv, cInFileObj* pMineFile)
{
   fpIndiv = pIndiv;
   fpMineFile = pMineFile;
}
 
cReadIndivMsg::~cReadIndivMsg()
{} 

void    cReadRowMsg::SendMsg()
{
   fpRow->ReadRowMines(fpMineFile);
}

cReadRowMsg::cReadRowMsg(cRowComp* pRow, cInFileObj* pMineFile)
{
   fpRow = pRow;
   fpMineFile = pMineFile;
}
 
cReadRowMsg::~cReadRowMsg()
{} 

void   cSaveCompMsg::SendMsg()
{
   fpMinefield->SaveNextComponent(fpOutfile);
}

cSaveCompMsg::~cSaveCompMsg()
{
}

// ************************************
//   component
// ************************************
int     cMfComponent::GetObjIndex(cMfieldObj* )
{
   return 0;
}

EBoolean    cMfComponent::IsMineComp()
{
   return false;
}

EBoolean    cMfComponent::CompleteComp()
{
   return false;
}

void    cMfComponent::UpdateStatusInfo(EBoolean bSentCompStatMsg)
{
   if (fpCompEntry)
      fpCompEntry->ShowInfo();
   if (fpCompStatus)
      fpCompStatus->ShowInfo();
      
   if (fpParentMf)
      fpParentMf->SetMineSumm();

   if (bSentCompStatMsg) fpUpdateCompStatMsg = NULL;
}

void    cMfComponent::ReplaceInList (cMfComponent* pComp)
{
  // this component is last on minefield list
  Assert ( fpParentMf->fpCompTail == this, kerrTypeMatch, 
	  "cMfComponent::ReplaceInList");

  fpParentMf->fpCompTail = fpPrev;
  fpPrev->fpNext = NULL;

  fpNext = pComp->fpNext;
  if (pComp->fpNext)
    pComp->fpNext->fpPrev = this;

  fpPrev = pComp->fpPrev;
  if (pComp->fpPrev)
    pComp->fpPrev->fpNext = this;

  if (fpParentMf->fpCompHead == pComp) fpParentMf->fpCompHead = this;
  if (fpParentMf->fpCompTail == pComp) fpParentMf->fpCompTail = this;

  pComp->fpNext = NULL;
  pComp->fpPrev = NULL;
  delete pComp;
}

void    cMfComponent::OnOffComponent (EBoolean )   
{
}

void    cMfComponent::SimulateComponent (EBoolean )
{
}

void    cMfComponent::UpdateComponent()
{
}

void    cMfComponent::SaveComponent(cOutFileObj* pOutfile)
{
   pOutfile->WriteComment("cannot save generic component");
}

void    cMfComponent::StartEditingComp()
{
   feIsEditing = true; 
}

void    cMfComponent::StopEditingComp()
{
   feIsEditing = false;
}

void     cMfComponent::MakeBackup()
{
   fpBackup = NULL;
}

void     cMfComponent::RestoreBackup()
{
}

void     cMfComponent::DeleteBackup()
{
   if ( fpBackup ) {
      delete fpBackup;
      fpBackup = NULL;
   }
}

void     cMfComponent::ResetTccCoord()
{
   fRefSet = eTccRef;
}

cMfComponent*    cMfComponent::DuplicateComp(T2dMove&, cMinefield*, EBoolean)
{
   return NULL;
}

EBoolean  cMfComponent::MoveCompOnly(T2dMove&)
   // to be overridden
{
   return false;
}

EBoolean  cMfComponent::MoveComp (T2dMove&)
   // minefield itself moves
{
   return false;
}

void   cMfComponent::InDetComp(cDetonation* )
{
}

void     cMfComponent::SelectComp()
{
   if ( ! feSelected ) {
      fpCompEntry->SwitchHilite();
      feSelected = true;
   }
}

void     cMfComponent::DeSelectComp()
{
   if ( feSelected ) {
      fpCompEntry->SwitchHilite();
      feSelected = false;
   }
}

void    cMfComponent::SetCompPvdSelect( EBoolean /*on*/ )
{
}

EBoolean    cMfComponent::AcceptsCommands()
{
   return false;
}

void    cMfComponent::CmdFire(EMfCmdMode eMineCmdMode)
{
}


void    cMfComponent::WriteCompTypeText(ofstream* pOut)
{
   switch (fCompType) {
      case eCompPerim:
            *pOut << "perimeter";
         break;
   
      case eCompIndiv:
            *pOut << "individual";
         break;
   
      case eCompGrid:
            *pOut << "grid";
         break;
   
      case eCompRow:
            *pOut << "row";
         break;

   }
}

void    cMfComponent::WriteCompSumm(ofstream* pOut)
{
   *pOut << " Component type: ";
   WriteCompTypeText(pOut);
   *pOut << endl;
}

void   cMfComponent::InitMfComponent(cMinefield* pParentMf,
				     EBoolean /*isBackup*/)
    // if parent is NULL, this is backup for individual component
    //  A backup for an entire minefield will have the parent set,
    //  and need linking with that backup parent
{
   feSelected    = false;
   feIsEditing   = false;
   fpCompEntry   = NULL;
   fpCompStatus  = NULL;
   fpBackup      = NULL;
   
   fpParentMf = pParentMf;
   fpNext     = NULL;
   fpPrev     = NULL;
   
   fpUpdateCompStatMsg = NULL;

   if ( pParentMf ) {
      fpParentMf = pParentMf;
      ++(fpParentMf->fNumComponents);
   
      fpNext = NULL;
      fpPrev = fpParentMf->fpCompTail;
   
      if (fpPrev)      // we're not the first component
        fpPrev->fpNext = this;
   
      fpParentMf->fpCompTail = this;
      if ( !fpParentMf->fpCompHead) fpParentMf->fpCompHead = this;
   
      // fpCompEntry and fpCompStatus left to lowest levels
   }
}

void   cMfComponent::SaveCompState()
{   
}

cMfComponent::cMfComponent(cMfComponent* pOriginal, cMinefield* pParentMf,
			   EBoolean isBackup)
      // used for duplicating and backing up
{
   fCompType     = pOriginal->fCompType;
   fRefSet       = pOriginal->fRefSet;

   InitMfComponent(pParentMf, isBackup);
}

cMfComponent::cMfComponent(cMinefield* pParentMf)
   // the child class sets the editor display widget
{
   fCompType     = eCompNull;      // set by realized component
   fRefSet       = eNullRef;

   InitMfComponent(pParentMf, false);
}

cMfComponent::cMfComponent()
{
   fCompType     = eCompNull;      // set by realized component
   fRefSet       = eNullRef;

   feSelected    = false;
   feIsEditing   = false;
   fpCompEntry   = NULL;
   fpCompStatus  = NULL;
   fpBackup      = NULL;
   
   fpParentMf = NULL;
   fpNext     = NULL;
   fpPrev     = NULL;
   fpUpdateCompStatMsg = NULL;
}   


cMfComponent::~cMfComponent()
{
   if ( feSelected && gbGui)
      gpGuiWin->DeleteSelection(fpParentMf, this, NULL);

   if ( fpBackup ) {
      delete fpBackup;
      fpBackup = NULL;
   }
   
   if (fpParentMf) {
      if (fpParentMf->fpCompHead == this) fpParentMf->fpCompHead = fpNext;
      if (fpParentMf->fpCompTail == this) fpParentMf->fpCompTail = fpPrev;
      --(fpParentMf->fNumComponents);
   }

   if (fpPrev) fpPrev->fpNext = fpNext;
   if (fpNext) fpNext->fpPrev = fpPrev;
   
   if (gbGui) {
      if (fpCompEntry) delete fpCompEntry;
      if (fpCompStatus) delete fpCompStatus;
   }
   if (fpUpdateCompStatMsg) fpUpdateCompStatMsg->GetMsgList()->DeleteMsg(fpUpdateCompStatMsg);
}

void    cMfComponent::RecreateVisuals()
{
}

void    cPerimeter::RecreateVisuals()
{
   // recreate perimeter display object
   fpPerimDisp = new cPerimDisp (this);
   fNumPoints  = 0;

   for ( TPerimPoint* pPt = GetFirstPoint(); pPt; pPt = GetNextPoint() )
        fpPerimDisp->AddVertex (++fNumPoints, pPt->GetTccPoint());
   fpPerimDisp->CloseLoop (fNumPoints + 1);

   fpCompEntry = new cPerimSummLine
     (this, fpParentMf->GetCompList()->GetMainWin());
}

void    cMineComp::RecreateVisuals()
{
   // recreate display objects for all mines in component
   for ( cMine* pMine = fpMineListHead; pMine; pMine = pMine->GetNextMine() )
     pMine->CompleteVisuals();

   fpCompEntry = new cMineCompSummLine
     (this, fpParentMf->GetCompList()->GetMainWin());
   fpCompStatus = new cCompStatusLine 
     (this, fpParentMf->GetCompStatus()->GetMainWin());
}

void    cRowComp::RecreateVisuals()
{
   cMineComp::RecreateVisuals();
   // recreate row display object
   fpRowDisp = new cRowDisp (this);
   fNumPoints  = 0;
   
   for (TPerimPoint* pPt = GetFirstPoint(); pPt; 
                     pPt = GetNextPoint() )
     fpRowDisp->AddVertex (++fNumPoints, pPt->GetTccPoint());
}

void cMineComp::CbOk (Widget w, XtPointer , 
				      XmAnyCallbackStruct* )
{
  XtDestroyWidget (w);
}

TMineParam*    cMineComp::GetCompMineParams()
{
   if ( fpDbMine )
      return fpDbMine->GetMineParam( fDbOpt.fMineOpt );
   else
      return NULL;
}

TFuzeParam*    cMineComp::GetCompFuzeParams()
{
   if ( fpDbMine )
      return fpDbMine->GetFuzeParam( fDbOpt.fFuzeOpt );
   else
      return NULL;
}

EBoolean    cMineComp::IsMineDirectional()
{
   if ( fpDbMine )
      return ( fpDbMine->IsDirectional(fDbOpt) );
   else
      return false;
}

char*   cMineComp::GetEntName()
{
   if ( fpDbMine )
      return fpDbMine->GetEntryEntityName();
   else
      return NULL;
}

const TEntityType&   cMineComp::GetEntType()
{
   static TEntityType type(0);

   if ( fpDbMine )
      return fpDbMine->GetEntryEntityType();
   else
      return type;
}

const TEntityType&   cMineComp::GetDetectEntType()
{
   static TEntityType type(0);

   if ( GetCompMineParams() )
      return GetCompMineParams()->fDetectionType;
   else
      return type;
}

EBoolean    cMineComp::IsMineComp()
{
   return true;
}

void     cMineComp::SetOnState (EBoolean on)
{
  fEditOn = on;
  
     // if we're simulating, the simulation object will take care of each
     // mine, otherwise we must set each mine toggle here
  if (fRealSim)
     new cOnOffComponentMsg(this, on);
  else {
   for ( cMine* pM = GetFirstMine(); pM; pM = pM->GetNextMine() )
      pM->SetMineOn(on);
  }
}

void     cMineComp::SetSimState (EBoolean sim)
{
  fEditSim = sim;
  
  if (fRealSim)
     new cSimulateComponentMsg(this, sim);
}

void     cMineComp::ShowMines (EBoolean show)
{
  fShowMines = show;
  
   if ( GetNumMines() > 0 ) {
     if ( gbGui )
       tactmap_bundle_changes (gpTactmap);

     for ( cMine* pM = GetFirstMine(); pM; pM = pM->GetNextMine() )
       pM->ShowObj (show);
   }
}

void     cMineComp::IncrementLive()
{
   fMineSumm.fTotalMines++;
   fMineSumm.fLiveMines++;
   
//   UpdateStatusInfo();
}

void     cMineComp::DecrementMine(EMineExternalState oldState)
{
   fMineSumm.fTotalMines--;
   switch (oldState) {
      case emsAlive:
            fMineSumm.fLiveMines--;
         break;
         
      case emsFired:
            fMineSumm.fFiredMines--;
         break;
         
      case emsOther:
            fMineSumm.fOtherDeadMines--;
         break;
         
   }
   
   //UpdateStatusInfo();
   if ( !fpUpdateCompStatMsg ) {
      fpUpdateCompStatMsg = new cUpdateMfCompStatusMsg(this);
   }
}

void     cMineComp::IncrementFired(EBoolean decLive)
{
   fMineSumm.fFiredMines++;
   if (decLive)
      fMineSumm.fLiveMines--;
  if ( !fpUpdateCompStatMsg ) {
      fpUpdateCompStatMsg = new cUpdateMfCompStatusMsg(this);
   }
}
   
void     cMineComp::IncrementOtherDead()
{
   fMineSumm.fLiveMines--;
   fMineSumm.fOtherDeadMines++;
   
   //UpdateStatusInfo();
   if ( !fpUpdateCompStatMsg ) {
      fpUpdateCompStatMsg = new cUpdateMfCompStatusMsg(this);
   }
}
void     cMineComp::IncrementVehHit()
{
   fMineSumm.fVehiclesHit++;
   
   //UpdateStatusInfo();
  if ( !fpUpdateCompStatMsg ) {
     fpUpdateCompStatMsg = new cUpdateMfCompStatusMsg(this);
  }
}

void     cMineComp::IncrementVehKill()
{
   fMineSumm.fVehiclesKilled++;
   
   //UpdateStatusInfo();
   if ( !fpUpdateCompStatMsg ) {
      fpUpdateCompStatMsg = new cUpdateMfCompStatusMsg(this);
   }
}


void     cMineComp::SetCompMineType(cDbMine* pDbMine, EBoolean bDraw)  
{
   Assert ( !fRealSim, kerrValidityCheck, "cMineComp::SetCompMineType");
   
   fpDbMine = pDbMine;
	fDbOpt.fMineOpt = fDbOpt.fFuzeOpt = fDbOpt.fVulnDetOpt = fDbOpt.fVulnMechOpt = 0;
      // set each mine type
   for ( cPlanMine* pMine = (cPlanMine*)fpMineListHead; 
         pMine; 
         pMine = (cPlanMine*)(pMine->GetNextMine()) ) {
      pMine->SetMineType(pDbMine);
   }
   if (bDraw) RedrawComp ();
}

void     cMineComp::SetCompHoursInPlace(float  hours)  
{
   Assert ( !fRealSim, kerrValidityCheck, "cMineComp::SetCompHoursInPlace");
   fHoursInPlace = hours;
   
      // set each mine depth
   for ( cPlanMine* pMine = (cPlanMine*)fpMineListHead; 
         pMine; 
         pMine = (cPlanMine*)(pMine->GetNextMine()) ) {
      pMine->SetMineHoursInPlace(hours);
   }
}

void     cMineComp::SetCompMineDepth(float  depth)  
{
   Assert ( !fRealSim, kerrValidityCheck, "cMineComp::SetCompMineDepth");
   fDepth = depth;
   
      // set each mine depth
   for ( cPlanMine* pMine = (cPlanMine*)fpMineListHead; 
         pMine; 
         pMine = (cPlanMine*)(pMine->GetNextMine()) ) {
      pMine->SetMineDepth(fDepth);
   }
}

void     cMineComp::SelectComp()
{
   if ( ! feSelected ) {
      fpCompEntry->SwitchHilite();
      fpCompStatus->SwitchHilite();
      feSelected = true;
      RedrawComp();
   }
}

void     cMineComp::DeSelectComp()
{
   if ( feSelected ) {
      fpCompEntry->SwitchHilite();
      fpCompStatus->SwitchHilite();
      feSelected = false;
      RedrawComp();
   }
}

void    cMineComp::SetCompPvdSelect( EBoolean on )
{
   if ( gbGui && GetNumMines() > 10 )
      tactmap_bundle_changes (gpTactmap);
   for ( cMine* pMine = fpMineListHead; 
         pMine; 
         pMine = pMine->GetNextMine() ) {
      pMine->SetCompPvdSelect(on);
   }
}

EBoolean    cMineComp::AcceptsCommands()
{
   if ( fpDbMine )
      return ( fpDbMine->IsCmdDet(fDbOpt) );
   else
      return false;
}

void    cMineComp::CmdFire( EMfCmdMode eMineCmdMode )
{
   if ( fpDbMine->IsCmdDet(fDbOpt) ) 
      for ( cMine* pMine = fpMineListHead; 
            pMine; 
            pMine = pMine->GetNextMine() ) 
         pMine->CmdFire(eMineCmdMode);
}

void     cMineComp::RedrawComp ()
{
  for ( cMine* pM = GetFirstMine(); pM; pM = pM->GetNextMine() )
    if (pM->GetDispObj())
    pM->GetDispObj()->RedrawSelected();
}

EBoolean    cMineComp::CompleteComp()
      // computes center location, radius, sets coords
{
   cMine*        pMine;
   TEntCoord     loc;
   
   double xMin = DBL_MAX;
   double xMax = - DBL_MAX;
   double yMin = DBL_MAX;
   double yMax = - DBL_MAX;
   double zMin = DBL_MAX;
   double zMax = - DBL_MAX;
   
   if (fMineSumm.fTotalMines == 0) return false;
   
   for (pMine = GetFirstMine(); pMine; pMine = pMine->GetNextMine() ) {
      if ( ! pMine->CompleteMfObj() ) return false;
      loc = pMine->GetMfLoc();
      if ( loc.x < xMin ) xMin = loc.x;
      if ( loc.x > xMax ) xMax = loc.x;
      if ( loc.y < yMin ) yMin = loc.y;
      if ( loc.y > yMax ) yMax = loc.y;
      if ( loc.z < zMin ) zMin = loc.z;
      if ( loc.z > zMax ) zMax = loc.z;
   }
   
   fMfLocation.x = (xMin + xMax) / 2.0;
   fMfLocation.y = (yMin + yMax) / 2.0;
   fMfLocation.z = (zMin + zMax) / 2.0;
   fpParentMf->MfToMapLoc(fMfLocation, &fTccLocation);
   fpParentMf->LocalToGeoLoc(fMfLocation, &fGeoLocation);
   fGeoOrient = fpParentMf->GetGeoOrient();
   fGeoRotation = fpParentMf->GetGeoRot();
   fRefSet = eRefComplete;
      
      // could get more accurate radius
   fRadius = sqrt( (xMax-xMin)*(xMax-xMin) + (yMax-yMin)*(yMax-yMin) + (zMax-zMin)*(zMax-zMin) ) / 2.0;
   
   return true;
}

void    cMineComp::SimulateComponent(EBoolean /*sim*/)
{
   //   int                   num;
   cSimMine*             pMine;
   T2linkMineSetList*    plinkCpMineSet = NULL;
   
      // get the component update on stack before its objects
   fRealSim = true;
   
   if ( fEditSim) {
      //NPAC fpUpdateCompMsg = new cUpdateComponentMsg(this);
   
         // write info to mine state file
      if ( gfMineStateSumm ) {
         *gpMineStateSumm << " { CompSimStart\n"
                          << " Simulation start time: "
                          << GetCurSimTime() * kSimTimeToSec
                          << endl;
         WriteCompSumm(gpMineStateSumm);
      }
   
      if ( gpNetMgr->UsesCp() )
         plinkCpMineSet = new T2linkMineSetList( fpDbMine->GetMineParam(fDbOpt.fMineOpt)->fDetectionType, 
                                              this, fMineSetList.GetLastLink() );
      
      if (gbGui) tactmap_bundle_changes (gpTactmap);
      //NPAC
      int i, j, numMines=GetNumMines(); //NPAC
      cSimMineList = new cSimMine*[numMines]; //NPAC
      cMine* pM; //NPAC

      /*NPAC 
         for ( cMine* pM = GetFirstMine(); pM; pM = pM->GetNextMine() ) {*/

       for(i=0;i<numMines;i++) { //NPAC
         pM = GetMineNum(i+1); //NPAC
         pMine = fpDbMine->CreateSimMine((cPlanMine*)pM, this);
         ReplaceMine(pM, pMine);
         cSimMineList[i]=pMine;  //NPAC 
         cSimMineList[i]->SimulateMine();  //NPAC 

         //NPAC pMine->SimulateMine();
         delete pM;
         pM = pMine;
         
         if (plinkCpMineSet) {
            if ( plinkCpMineSet->GetNumMines() >=  cCpManager::GetMinesPerPdu() ) {
               plinkCpMineSet->MineChange(NULL);   // initiates first message
               plinkCpMineSet = new T2linkMineSetList( fpDbMine->GetMineParam(fDbOpt.fMineOpt)->fDetectionType, 
                                              this, fMineSetList.GetLastLink() );
            }
            //NPAC 

            cSimMineList[i]->SetMineSetList(plinkCpMineSet);

            //pMine->SetMineSetList(plinkCpMineSet);
         }
      }
      if (plinkCpMineSet)
          plinkCpMineSet->MineChange(NULL);   // initiates first message

      if ( gfMineStateSumm ) *gpMineStateSumm << " }\n" << endl;
      
         // now the virtual mine, not actually simulated
      fpVirtualMine = fpDbMine->CreateVirtualMine(this, fRadius);
      
      fbSimulating = true;
   }
}

void     cMineComp::ReplaceMine(cMine* pOld, cMine* pNew)
{
   if ( pOld->GetDispObj() ) {
      pNew->TakeDispObj( pOld->GetDispObj() );
      ( (cMineDisp*)(pNew->GetDispObj()) )->SetMine(pNew);
   }

         // do the mine links
   pNew->ReplaceEnt(pOld);

         // the component links
   if ( fpMineListHead == pOld) fpMineListHead = pNew;
   if ( fpMineListTail == pOld) fpMineListTail = pNew;
   if ( fpCurMine == pOld) fpCurMine = pNew;

         // clear out pointers, so pOld can be deleted
   pOld->fpMinefield = NULL;
   pOld->fpMfComponent = NULL;
   pOld->fpDispObj = NULL;
}

void    cMineComp::OnOffComponent(EBoolean on)
   // called from msg, to really turn the object on when simulating
{
   fEditOn = on;

   for ( cMine* pM = GetFirstMine(); pM; pM = pM->GetNextMine() )
      ((cSimMine*)pM)->OnOffMine(on);
}

void    cMineComp::UpdateComponent()
{
   double range;
   cLinkVehLoc* pProxVeh;
   //cLinkEmiss* pProxEmiss;
   
   cVehListHead&  vehMasterList = gVehMgr.GetVehList();
   cEmissListHead&  emissMasterList = gEmissionMgr.GetEmissList();
   
   fProxVehList.DeleteList();
   fProxEmissList.DeleteList();
   
   for ( cVehicle* veh = vehMasterList.GetFirstVeh(); veh; veh = vehMasterList.GetNextVeh() ) {
      if ( veh->IsMobile() ) {
         range = fGeoLocation.Distance( veh->GetGeoLoc() );
         if ( fpVirtualMine->InMineRange(veh, range) ) {
            pProxVeh = new cLinkVehLoc(veh, (cLinkVeh*) (fProxVehList.GetLastLink()), 
                                 fpParentMf, false);
         }
      }
   }
   
   for ( cEmission* emiss = emissMasterList.GetFirstEmiss(); emiss; emiss = emissMasterList.GetNextEmiss() ) {
      if ( emiss->GetVehicle()->IsMobile() ) {
         range = (double) fTccLocation.HorzDistance( emiss->GetTccLoc() );
         if ( fpVirtualMine->InMineRange(emiss, range) ) {
            /*pProxEmiss =*/ new cMineLinkEmiss(emiss, (cLinkEmiss*)(fProxEmissList.GetLastLink()), 
	    			fpParentMf, false);   // adds at end
         }
      }
   }
   
   //NPAC 
   int i,numMines=GetNumMines();

   for(i=0;i<numMines;i++)
   { cSimMineList[i]->UpdateMine(); }
   //NPAC fpUpdateCompMsg = new cUpdateComponentMsg(this);
}

void    cMineComp::StartEditingComp()
{
   feIsEditing = true;
   MakeBackup();

         // create entries for location text
   if ( GetNumMines() > 0 ) {
      if ( gbGui )
      tactmap_bundle_changes (gpTactmap);

      for (cMine* pMine = GetFirstMine(); pMine; pMine = pMine->GetNextMine() )
          ((cMineDisp*)pMine->GetDispObj())->SetSensitive (true);
   }

   if (fCompType == eCompRow && ((cRowComp*)this)->GetRowDisp())
      ((cRowComp*)this)->GetRowDisp()->SetSensitive (true);
}

void    cMineComp::StopEditingComp()
{
   feIsEditing = false;
   if (fpBackup) {
      delete fpBackup;
      fpBackup = NULL;
   }
   
   if ( gbGui && GetNumMines() > 0 )
      tactmap_bundle_changes (gpTactmap);

   ResetTccCoord();
   if (fpParentMf->IsComplete())
      CompleteComp();
   
   if (gbGui) tactmap_bundle_changes (gpTactmap);
   for (cMine* pMine = GetFirstMine(); pMine; pMine = pMine->GetNextMine())
     ((cMineDisp*)pMine->GetDispObj())->SetSensitive (false);

   if (fCompType == eCompRow)
     ((cRowComp*)this)->GetRowDisp()->SetSensitive (false);

   fpCompEntry->ShowInfo();
   fpParentMf->SetMineSumm();
   gpGuiWin->SetSelection( false, fpParentMf, this );
//   SetMineSelection (NULL);
}

void     cMineComp::ResetTccCoord()
{
   cMine* pMine;
   
   fRefSet = eTccRef;
   for (pMine=GetFirstMine(); pMine; pMine = pMine->GetNextMine() )
      pMine->ResetTccCoord();
}

EBoolean  cMineComp::MoveCompOnly(T2dMove& move)
   // moves component, but not minefield
{
   EBoolean success = true;
   
   if ( gbGui )
     tactmap_bundle_changes (gpTactmap);

      // make sure we have a component location
   if ( fRefSet != eRefComplete )
      success = CompleteComp();
   
   for (cMine* pM = GetFirstMine(); pM && success; pM = pM->GetNextMine() )
      success = pM->MoveMineWithComp(fMfLocation, move);
   
   if ( success)
      success = CompleteComp();
   
   if ( ! success && gbGui)
      gpGuiWin->DialogMessage (cGuiWindow::eDialogInfo, 
			   NULL,
			   "Unable to move component.  Please delete.");

   return success;
}

EBoolean  cMineComp::MoveComp(T2dMove& move)
   // minefield is moving
{
   EBoolean success = true;
   
      // make sure we have a component location
   if ( fRefSet != eRefComplete )
      success = CompleteComp();

   if ( success ) {
      fMfLocation = move.fRotMatrix * fMfLocation;
      fRefSet = eMfRef;
      
      for (cMine* pM = GetFirstMine(); pM && success; pM = pM->GetNextMine() ) 
         success = pM->MoveMineWithMf(move);
   }
   
   return success;
}

void   cMineComp::InDetComp(cDetonation* pDetonation)
{
   float detRange  = fGeoLocation.Distance(pDetonation->fGeoLoc);
   float vulnRange = fpVirtualMine->GetDetVulnRange(pDetonation->fDetonType);
   
   if (detRange <= vulnRange) {
      if ( gfInteract ) {
         *gpInteract << "{ Detonation near component" << endl;
         *gpInteract << "time: " << GetCurSimTime() * kSimTimeToSec << endl;
         pDetonation->WriteDetInfo(gpInteract);
         
         *gpInteract << "mines affected:" << endl;
      }
      TEntCoord detMfLoc;
      /* NPAC 
      for (cMine* pObj = GetFirstMine(); pObj; pObj = pObj->GetNextMine() )
         if (pObj->IsSimulating())
            ((cSimMine*)pObj)->InDetMine(pDetonation);
      */
       cMine* pObj;int i, numMines=GetNumMines();
       for(i=0;i<numMines;i++)
        { pObj = cSimMineList[i];
          if (pObj->IsSimulating())
             ((cSimMine*)pObj)->InDetMine(pDetonation);
        }

      if ( gfInteract ) {
         *gpInteract << "}\n" << endl;
      }
   }
}

void        cMineComp::SaveComponent(cOutFileObj* pOutfile)
      // specific component type will have written header
      // We don't write level close, so descendents can add trailer
{
   cMine*    pMine;
   TItemList ptlist;
   char      bool = 'f';   // 't' or 'f'

         // mine type in component
   pOutfile->WriteComment("type of mine in this component");
   int   type[7];
   const TEntityType& entType = fpDbMine->GetEntryEntityType();
   type[0] = entType.kind;
   type[1] = entType.domain;
   type[2] = entType.country;
   type[3] = entType.category;
   type[4] = entType.subcategory;
   type[5] = entType.specific;
   type[6] = entType.extra;
   TItemList typelist;
   typelist.count = 7;
   typelist.ptrItems = (void *)type;
   pOutfile->WriteItem( fKeyMCompTable[emcompkwType].key,
                       fKeyMCompTable[emcompkwType].type,
                       (void*)(&typelist) );
   
         //depth
   float depth = fDepth * kMeterToInch;
   pOutfile->WriteItem( fKeyMCompTable[emcompkwDepth].key,
                       fKeyMCompTable[emcompkwDepth].type,
                       (void*)(&depth) );
   
         //age
   pOutfile->WriteItem( fKeyMCompTable[emcompkwAge].key,
                       fKeyMCompTable[emcompkwAge].type,
                       (void*)(&fHoursInPlace) );
   
         //fuze type
   pOutfile->WriteItem( fKeyMCompTable[emcompkwFuzeType].key,
                       fKeyMCompTable[emcompkwFuzeType].type,
                       (void*)(fpDbMine->GetFuzeParam(fDbOpt.fFuzeOpt)->GetFuzeName()) );
   
         //self dest time
   pOutfile->WriteItem( fKeyMCompTable[emcompkwSelfDestTime].key,
                       fKeyMCompTable[emcompkwSelfDestTime].type,
                       (void*)(&(fFuzeOpt.fSelfDestTimeHrs)) );
   
         //tgt counter
   pOutfile->WriteItem( fKeyMCompTable[emcompkwTgtCount].key,
                       fKeyMCompTable[emcompkwTgtCount].type,
                       (void*)(&(fFuzeOpt.fTgtCount)) );
   
         //antihandle
   bool = (fFuzeOpt.fbAntihandling == true) ? 't' : 'f';
   pOutfile->WriteItem( fKeyMCompTable[emcompkwAntiHandle].key,
                       fKeyMCompTable[emcompkwAntiHandle].type,
                       (void*)(&bool) );
   
         //blast hard
   bool = (fFuzeOpt.fbBlastHard == true) ? 't' : 'f';
   pOutfile->WriteItem( fKeyMCompTable[emcompkwBlastHard].key,
                       fKeyMCompTable[emcompkwBlastHard].type,
                       (void*)(&bool) );

         // mine locations
   pOutfile->WriteComment("saving mine locations in 2 dim minefield coords");
   TFlatEntCoord* points = new TFlatEntCoord[fMineSumm.fTotalMines];
   int i = 0; 
   for (       pMine = GetFirstMine();
               pMine != NULL;
               pMine = pMine->GetNextMine())   {
      points[i].x = pMine->GetMfLoc().x;
      points[i].y = pMine->GetMfLoc().y;
      ++i;
   }
   ptlist.count = fMineSumm.fTotalMines;
   ptlist.ptrItems = (void *)points;
   pOutfile->WriteItem( fKeyMCompTable[emcompkwLocList].key,
                        fKeyMCompTable[emcompkwLocList].type,
                        (void*)(&ptlist) );
   delete[] points;
   points = NULL;

         // mine orientations
   pOutfile->WriteComment("saving mine psi and theta orientations, map ref");
   T2Dsingle* orients = new T2Dsingle[fMineSumm.fTotalMines];
   i = 0; 
   for (       pMine = GetFirstMine();
               pMine != NULL;
               pMine = pMine->GetNextMine())   {
      orients[i].x = pMine->GetTccOrient().psi;
      orients[i].y = pMine->GetTccOrient().theta;
      ++i;
   }
   ptlist.count = fMineSumm.fTotalMines;
   ptlist.ptrItems = (void *)orients;
   pOutfile->WriteItem( fKeyMCompTable[emcompkwOrientList].key,
                        fKeyMCompTable[emcompkwOrientList].type,
                        (void*)(&ptlist) );
   delete[] orients;
   orients = NULL;
}

void   cMineComp::DeleteMine (cMine* pM)
{
  // first decrement the count, so that gui can get new count
  DecrementMine( pM->GetExternalState() );
      
  // delete entry, while index and links are still valid
  if (gbGui )
    gpGuiWin->GetCurEditor()->DeleteObjEntry(this, GetObjIndex(pM));

  // fix indices
  if (fpMineListHead == pM) 
    fpMineListHead = pM->GetNextMine();
  if (fpMineListTail == pM) 
    fpMineListTail = (cMine*)(pM->GetPrevEnt());

  delete pM;
  RedrawComp();
}

void   cMineComp::DeleteMines()
{
  if (gbGui && fMineSumm.fTotalMines > 1)
    tactmap_bundle_changes (gpTactmap);

   /* delete coord list entries in a batch to improve performance */
   if ( gbGui && 
        (gpGuiWin->GetCurComponent() == this) &&
        gpGuiWin->GetCurEditor()->GetMineCoordWidget() )
     gpGuiWin->GetCurEditor()->GetMineCoordWidget()->DeleteEntries();

   while (fpMineListHead) {
      cMine* pM = fpMineListHead;
      fpMineListHead = pM->GetNextMine();
      pM->fpMfComponent = NULL;
      delete pM;
    }
   fpMineListTail = NULL;
   fMineSumm.fTotalMines = fMineSumm.fLiveMines = 
     fMineSumm.fFiredMines = fMineSumm.fOtherDeadMines = 0;
   //UpdateStatusInfo();
   UpdateStatusInfo();
}

void   cMineComp::ParseMineItem(cInFileObj* pMineFile)
{
   EFileReturn   returnCode;
   int           returnIndex;
   void*         returnVal;
   TItemList *   list;
   cPlanMine*    pPlanMine;
   int           i;
   TEntityType   type;
   cDbMine*      pDbMine;
   int           fuzeTypeIndex = 0;
      
   returnCode = pMineFile->ParseItem(fKeyMCompTable, eSizeMCompTable,
		&returnIndex, &returnVal);
   
   switch (returnCode) {
      case eFileNoErr:
         switch (returnIndex) {
         
            case emcompkwType:      // list of field values
                  list = (TItemList *)returnVal;
                  type.ParseList(list);
                  
                  pDbMine  = (cDbMine*)(gMineDbMgr.MatchExact(type));
                  
                  if ( pDbMine ) {
                     //fpDbMine = pDbMine;   // need to set the descriptor
		     SetCompMineType(pDbMine, false);
		     fDbOpt.fFuzeOpt = fpDbMine->GetFuzeOpt(fuzeTypeIndex);
                     /*
		       for ( pPlanMine = (cPlanMine*)GetFirstMine(); 
		       pPlanMine; 
		       pPlanMine = (cPlanMine*)(pPlanMine->GetNextMine()) )
		       pPlanMine->SetMineType(fpDbMine);
		       */
		  }
                  else if (gbGui) {
                     char msg[] = "Could not find mine type, not creating mines";
                     gpGuiWin->PopupMessage (gpGuiWin->GetMainWin(), cGuiWindow::eDialogWarning,
                        "", "", NULL, NULL, NULL, msg);
                  }

                  delete[] (int *)(list->ptrItems);
                  delete list;
                  list = NULL;
                  returnVal = NULL;
               break;
      
            case emcompkwTypeString:      // stored as string, need TEntityType
                  pDbMine  = (cDbMine*)(gMineDbMgr.FindByName((char*)returnVal));
                  
                  if ( pDbMine ) {
                     //fpDbMine = pDbMine;   // need to set the descriptor
		     SetCompMineType(pDbMine, false);
                     fDbOpt.fFuzeOpt = fpDbMine->GetFuzeOpt(fuzeTypeIndex);
		     /*
		       for ( pPlanMine = (cPlanMine*)GetFirstMine(); 
                           pPlanMine; 
                           pPlanMine = (cPlanMine*)(pPlanMine->GetNextMine()) )
                        pPlanMine->SetMineType(fpDbMine);
			*/
                  }
                  else if (gbGui) {
                     char msg[] = "Could not find mine type, not creating mines";
                     gpGuiWin->PopupMessage (gpGuiWin->GetMainWin(), cGuiWindow::eDialogWarning,
                        "", "", NULL, NULL, NULL, msg);
                  }

                  delete[] returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwDepth:
                  fDepth = *(float*)returnVal * kInchToMeter;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwAge:
                  fHoursInPlace = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwFuzeType: {  // need to find by name
                  fuzeTypeIndex = TFuzeParam::GetTypeIndex((char*)returnVal);
                  if (fpDbMine)
                     fDbOpt.fFuzeOpt = fpDbMine->GetFuzeOpt(fuzeTypeIndex);
                  delete[] returnVal;
                  returnVal = NULL;
               }
               break;
      
            case emcompkwSelfDestTime:
                  fFuzeOpt.fSelfDestTimeHrs = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwTgtCount:
                  fFuzeOpt.fTgtCount = *(int*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwAntiHandle:
                  fFuzeOpt.fbAntihandling = 
                     ( *(char*)returnVal == 't') ? true : false;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwBlastHard:
                  fFuzeOpt.fbBlastHard = 
                     ( *(char*)returnVal == 't') ? true : false;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case emcompkwLocList:      // minefield coords
                  fRefSet = eMfRef;
                  list = (TItemList *)returnVal;
                  while (fpMineListHead) 
                     delete fpMineListHead;
                     
                  Assert( (fMineSumm.fTotalMines == 0) && (fpMineListHead == NULL) && (fpMineListTail == NULL), 
                           kerrList, "cMineComp::ParseMineItem" );
                  for (i = 0; i < list->count; i++) {
                     pPlanMine = new cPlanMine(this, (TFlatEntCoord *)(list->ptrItems) + i);
                     pPlanMine->SetMineType(fpDbMine);
                  }
                  delete[] (TFlatEntCoord *)(list->ptrItems);
                  delete list;
                  list = NULL;
                  returnVal = NULL;
               break;
      
            case emcompkw3dLocList:    {  // minefield coords, including elevation
                  TEntCoord entLoc;
                  TLocation* pLoc;
                  fRefSet = eMfRef3d;
                  list = (TItemList *)returnVal;
                  while (fpMineListHead) 
                     delete fpMineListHead;
                     
                  Assert( (fMineSumm.fTotalMines == 0) && (fpMineListHead == NULL) && (fpMineListTail == NULL), 
                           kerrList, "cMineComp::ParseMineItem:2" );
                  for (i = 0; i < list->count; i++) {
                     pLoc = (TLocation *)(list->ptrItems) + i;
                     entLoc.x = pLoc->x;
                     entLoc.y = pLoc->y;
                     entLoc.z = pLoc->z;
                     pPlanMine = new cPlanMine(this, &entLoc);
                     pPlanMine->SetMineType(fpDbMine);
                  }
                  delete[] (TLocation *)(list->ptrItems);
                  delete list;
                  list = NULL;
                  returnVal = NULL;
               break;
               }
               
            case emcompkwOrientList:      // minefield coords
                  list = (TItemList *)returnVal;
                     
                  Assert( (GetNumMines() == list->count) , 
                           kerrList, "cMineComp::ParseMineItem.3" );
                  for (i = 0, pPlanMine = (cPlanMine*)GetFirstMine(); 
                       i < list->count; 
                       i++, pPlanMine = (cPlanMine*)(pPlanMine->GetNextMine()) ) {
                     pPlanMine->SetTccOrient( ((T2Dsingle *)(list->ptrItems) + i)->x,
                                              ((T2Dsingle *)(list->ptrItems) + i)->y,
                                              0.0);
                  }
                  delete[] (T2Dsingle *)(list->ptrItems);
                  delete list;
                  list = NULL;
                  returnVal = NULL;
               break;

            }

         break;

      case eLevelDone:   // done reading this component
            new cReadMinefieldMsg(fpParentMf, pMineFile);
               // check that we have enough info
            if ( (fMineSumm.fTotalMines == 0) || (fpDbMine == NULL) ) 
               delete this;
            else
               fpParentMf->SetMineSumm();
         break;

      default: // error returns.  Give the user a message
            Assert(false, kerrReadMinefield, "cMineComp::ParseMineItem:4");
            delete pMineFile;
         break;
   }
}

cMine*  cMineComp::GetMineNum (int iPt) 
{
  cMine* pMine = fpMineListHead;
  for (int i = 1; i < iPt; i++) {
    if (pMine == NULL) break;
    pMine = pMine->GetNextMine();
  }
  return pMine;
}

int     cMineComp::GetObjIndex(cMfieldObj* pObj)
{
  cMine* pMine = fpMineListHead;
  for (int i = 1; pMine; i++, pMine = pMine->GetNextMine()) {
    if ( (cMfieldObj*)pMine == pObj) 
       return i;
  }
  return 0;
}

void   cMineComp::CopyMineCompParams(cMineComp* pOriginal)
{
   fEditOn         = pOriginal->fEditOn;
   fEditSim        = pOriginal->fEditSim;
   fDepth          = pOriginal->fDepth;
   fHoursInPlace   = pOriginal->fHoursInPlace;
   
   fpDbMine        = pOriginal->fpDbMine;
   fDbOpt          = pOriginal->fDbOpt;

   fGeoLocation    = pOriginal->fGeoLocation;
   fGeoOrient      = pOriginal->fGeoOrient;
   fGeoRotation    = pOriginal->fGeoRotation;
   fTccLocation    = pOriginal->fTccLocation;
   fTccOrientation = pOriginal->fTccOrientation;
   fTccRotation    = pOriginal->fTccRotation;

   fMfLocation     = pOriginal->fMfLocation;
   fRadius         = pOriginal->fRadius;
   fShowMines      = pOriginal->fShowMines;
};

cMineComp::cMineComp(cMineComp* pOriginal, cMinefield* pParent, 
		     EBoolean isBackup) : 
         cMfComponent(pOriginal, pParent, isBackup)
      // used for duplicating components
{
   fRealSim        = false;      // never start in simulated state

   fpVirtualMine   = NULL;
   fpCurMine       = NULL;
   fpUpdateCompMsg = NULL;
   fpMineListHead  = NULL;
   fpMineListTail  = NULL;
//   fpDbMine      = NULL;
   
   CopyMineCompParams( pOriginal );
   
         // fMineSumm incremented when adding mines
   cPlanMine* pNew;
   for (cMine* pM = pOriginal->GetFirstMine();
               pM; pM = pM->GetNextMine())
      pNew = new cPlanMine(this, pM, isBackup );
}


cMineComp::cMineComp(cMinefield* pParentMf) : 
   cMfComponent(pParentMf)
{   
   fRadius          = 0.0;
   
   fpMineListHead = NULL;
   fpMineListTail = NULL;
   fpVirtualMine  = NULL;
   fpCurMine      = NULL;
   fpUpdateCompMsg = NULL;
   fpDbMine       = NULL;
   
   fEditOn           = true;
   fEditSim          = true;
   fRealSim          = false;
   fDepth            = 0.0;
   fHoursInPlace     = 0.0;
   fShowMines        = true;
      // fCmsType, locations, fProxVehList have constructors
}

void   cMineComp::SaveCompState()
{   
      // write info to mine state file
   if ( fbSimulating && gfMineStateSumm ) {
      *gpMineStateSumm << "{ Save Comp State Command\n"
                       << "at time: "
                       << GetCurSimTime() * kSimTimeToSec
                       << endl;
      WriteCompSumm(gpMineStateSumm);
      
      for ( cSimMine* pMine = (cSimMine*)fpMineListHead;
                      pMine;
                      pMine = (cSimMine*) (pMine->GetNextMine()) ) {
         pMine->SaveMineState();
      }
   }
}

void     cMineComp::WriteCompSumm(ofstream* pOut)
{
   *pOut << "  contains mine type: "
         << fpDbMine->GetEntryEntityName()
         << endl;
   *pOut << "  mine depth (inches): "
         << fDepth * kMeterToInch
         << endl;
   *pOut << "  fuze type: "
         << fpDbMine->GetFuzeParam(fDbOpt.fFuzeOpt)->GetFuzeName();
          
   if ( fFuzeOpt.GetAntihandling() )
      *pOut << " antihandling";
   if ( fFuzeOpt.GetBlastHard() )
      *pOut << " blast hardened";

   *pOut << endl;
   
   *pOut << "  totals (tot/live/fired/dead/vehhit/vehkill): "
         << fMineSumm.fTotalMines << " "
         << fMineSumm.fLiveMines << " "
         << fMineSumm.fFiredMines << " "
         << fMineSumm.fOtherDeadMines << " "
         << fMineSumm.fVehiclesHit << " "
         << fMineSumm.fVehiclesKilled << endl;
}

cMineComp::~cMineComp()
{
      // prox vehicle list destructor called automatically
   
   
         // write info to mine state file
      if ( fbSimulating && gfMineStateSumm ) {
         *gpMineStateSumm << " { CompDelete\n"
                          << " Simulation end time: "
                          << GetCurSimTime() * kSimTimeToSec
                          << endl;
         WriteCompSumm(gpMineStateSumm);
      }
   
      // delete all objects
   DeleteMines();
   
   T2linkMineSetList* pMineList = (T2linkMineSetList*) (fMineSetList.GetLastLink());
   while ( pMineList != fMineSetList.GetNULLLink() ) {
      delete pMineList;
      pMineList = (T2linkMineSetList*) (fMineSetList.GetLastLink());
   }

   if ( fbSimulating && gfMineStateSumm ) 
      *gpMineStateSumm << " }\n" << endl;
   
   if (fpVirtualMine) {
      delete fpVirtualMine;
      fpVirtualMine = NULL;
   }
   if (fpUpdateCompMsg) (fpUpdateCompMsg->GetMsgList())->DeleteMsg(fpUpdateCompMsg);
}

EBoolean   cMineComp::AddTccMine (TLocation& tccPt)
{
   fRefSet = eNullRef;
   cMine* pMine = new cPlanMine (this, tccPt);
   pMine->CreateDispObj();
   CompleteComp();

   if (gbGui && gpGuiWin->GetCurComponent() == this) {
     if (pMine->GetDispObj())
       ((cMineDisp*)pMine->GetDispObj())->SetSensitive (true);
     gpGuiWin->NewMine (this, tccPt);

     gpGuiWin->SetSelection (false, fpParentMf, this, pMine);
//     RedrawComp();
   }
   return true;
}

EBoolean   cMineComp::AddTccMines (TLocation* tccPts, int nMines)
{
   cMine* pMine;
   int    i;

   for (i = 0; i < nMines; i++) {
     pMine = new cPlanMine (this, tccPts[i]);
     pMine->CreateDispObj();

     if (gbGui && gpGuiWin->GetCurComponent() == this) {
       if (pMine->GetDispObj())
	 ((cMineDisp*)pMine->GetDispObj())->SetSensitive (true);
     }
   }

   if (gbGui && gpGuiWin->GetCurComponent() == this)
     gpGuiWin->NewMines (this, tccPts, nMines);

   return true;
}

EBoolean   cIndivComp::AddTccMine (TLocation& tccPt)
{
  if (GetMinefield()->GetPerimeter()->PointInside (tccPt)) {
    cMineComp::AddTccMine (tccPt);
    return true;
  }
  else {
    if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
					"Invalid point.  Mine must be emplaced\ninside minefield perimeter.");

    return false;
  }
}

void   cIndivComp::ReadIndivMines(cInFileObj* pMineFile)
{
   EFileReturn   returnCode;
   int           returnIndex;
   void*         returnVal;
      
   returnCode = pMineFile->ReadItem(fKeyIndivTable, eSizeIndivTable, &returnIndex, &returnVal);
   
   switch (returnCode) {
         
      case eFileNoErr:
         switch (returnIndex) {

            }

            new cReadIndivMsg(this, pMineFile);
         break;

		case eBadKeyword:
            cMineComp::ParseMineItem(pMineFile);
            new cReadIndivMsg(this, pMineFile);
         break;

      case eLevelDone:   // done reading this component
            new cReadMinefieldMsg(fpParentMf, pMineFile);
               // check that we have enough info
            if ( (fMineSumm.fTotalMines == 0) || (fpDbMine == NULL) ) 
               delete this;
            else
               fpParentMf->SetMineSumm();
         break;

      default: // error returns.  Give the user a message
            Assert(false, kerrReadMinefield, "cIndivComp::ReadIndivMines");
            delete pMineFile;
         break;
   }
}

void        cIndivComp::SaveComponent(cOutFileObj* pOutfile)
{
   pOutfile->WriteComment("individual mine placement component");

         // individual level start
   pOutfile->WriteItem( cMinefield::fKeyMfldTable[cMinefield::emfkwIndiv].key,
                       cMinefield::fKeyMfldTable[cMinefield::emfkwIndiv].type,
                       NULL );

   cMineComp::SaveComponent(pOutfile);

   pOutfile->EndLevel();
}

cMfComponent*    cIndivComp::DuplicateComp(T2dMove& move, cMinefield* pParent,
					   EBoolean isBackup)
{
   cIndivComp* pC = new cIndivComp(this, pParent, isBackup);

   if ( ! isBackup) {
      pC->MoveCompOnly(move);
      pC->SetCompMineType(pC->GetDbEntity());
      if (pC->fpCompEntry)
	 pC->fpCompEntry->ShowInfo();
      if (pC->fpCompStatus)
	 pC->fpCompStatus->ShowInfo();
   }
   return pC;
}

void     cIndivComp::MakeBackup()
{
   if ( fpBackup )
      delete fpBackup;
      
   fpBackup = NULL;
   
   if ( GetNumMines() > 0 ) 
      fpBackup = new cIndivComp(this, NULL, true);
}

void     cIndivComp::RestoreBackup()
{
   cIndivComp* pBackup = (cIndivComp*) fpBackup;
   
   if ( pBackup ) {
      CopyMineCompParams(pBackup);
         // delete existing mines, but don't update displays yet
      if ( gbGui )
	tactmap_bundle_changes (gpTactmap);
      while (fpMineListHead)
         delete fpMineListHead;

      cPlanMine* pNew;
      for (cMine* pM = pBackup->GetFirstMine();
                  pM; pM = pM->GetNextMine())
         pNew = new cPlanMine(this, pM, false );

      UpdateStatusInfo();
   }
}

void   cIndivComp::InitIndivComp(cMinefield* pParentMf, EBoolean isBackup)
{
   fCompType = eCompIndiv;
   fpBackup = NULL;

   if ( ( ! isBackup ) && gbGui) {
      fpCompEntry = new cMineCompSummLine
            (this, pParentMf->GetCompList()->GetMainWin());
      fpCompStatus = new cCompStatusLine 
            (this, pParentMf->GetCompStatus()->GetMainWin());
   }
   else {
     fpCompEntry = NULL;
     fpCompStatus = NULL;
  }
}

cIndivComp::cIndivComp(cIndivComp* pOriginal, cMinefield* pParentMf,
		       EBoolean isBackup):   // duplication or backup
   cMineComp(pOriginal, pParentMf, isBackup)
{
   InitIndivComp(pParentMf, isBackup);
}

cIndivComp::cIndivComp(cMinefield* pParentMf) : 
   cMineComp(pParentMf)
{
   InitIndivComp(pParentMf);
}

cIndivComp::~cIndivComp()
{
}

void   cGridComp::ReadGridMines(cInFileObj* pMineFile)
{
   EFileReturn   returnCode;
   int           returnIndex;
   void*         returnVal;
      
   returnCode = pMineFile->ReadItem(fKeyGridTable, eSizeGridTable, &returnIndex, &returnVal);
   
   switch (returnCode) {
      case eFileNoErr:
         switch (returnIndex) {
      
            case egridkwSpacing:
                  fSpacing = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
               break;
            }
      
            new cReadGridMsg(this, pMineFile);
         break;

		case eBadKeyword:
            cMineComp::ParseMineItem(pMineFile);
            new cReadGridMsg(this, pMineFile);
         break;

      case eLevelDone:   // done reading this component
            new cReadMinefieldMsg(fpParentMf, pMineFile);
               // check that we have enough info
            if ( (fMineSumm.fTotalMines == 0) || (fpDbMine == NULL) ) 
               delete this;
            else
               fpParentMf->SetMineSumm();
         break;

      default: // error returns.  Give the user a message
            Assert(false, kerrReadMinefield, "cGridComp::ReadGridMines");
            delete pMineFile;
         break;
   }
}

void        cGridComp::SaveComponent(cOutFileObj* pOutfile)
{
   pOutfile->WriteComment("grid mine placement component");

         // grid level start
   pOutfile->WriteItem( cMinefield::fKeyMfldTable[cMinefield::emfkwGrid].key,
                       cMinefield::fKeyMfldTable[cMinefield::emfkwGrid].type,
                       NULL );
         //spacing
   pOutfile->WriteItem( fKeyGridTable[egridkwSpacing].key,
                       fKeyGridTable[egridkwSpacing].type,
                       (void*)(&fSpacing) );

   cMineComp::SaveComponent(pOutfile);

   pOutfile->EndLevel();
}

void        cRowComp::SaveComponent(cOutFileObj* pOutfile)
{
   TFlatEntCoord* points;
   TItemList ptlist;
   int i = 0; 

   pOutfile->WriteComment("row mine placement component");

         // row level start
   pOutfile->WriteItem( cMinefield::fKeyMfldTable[cMinefield::emfkwRow].key,
                       cMinefield::fKeyMfldTable[cMinefield::emfkwRow].type,
                       NULL );
         //spacing
   pOutfile->WriteItem( fKeyRowTable[erowkwSpacing].key,
                       fKeyRowTable[erowkwSpacing].type,
                       (void*)(&fSpacing) );

         //offset
   pOutfile->WriteItem( fKeyRowTable[erowkwOffset].key,
                       fKeyRowTable[erowkwOffset].type,
                       (void*)(&fOffset) );

         // points, in mf coordinates
   pOutfile->WriteComment("saving row points in 2 dim minefield coords");
   points = new TFlatEntCoord[fNumPoints];

   i = 0; 
   for (T2link* pPerimPt = fPointList.GetFirstLink();
                pPerimPt != fPointList.GetNULLLink();
                pPerimPt = fPointList.GetNextLink()) {
      points[i].x   = ((TPerimPoint*) pPerimPt)->GetMfPoint().x;
      points[i++].y = ((TPerimPoint*) pPerimPt)->GetMfPoint().y;
   }
   
   ptlist.count = fNumPoints;
   ptlist.ptrItems = (void *)points;

   pOutfile->WriteItem( fKeyRowTable[erowkwMfRowCoords].key,
                        fKeyRowTable[erowkwMfRowCoords].type,
                        (void*)(&ptlist) );
   delete[] points;

   cMineComp::SaveComponent(pOutfile);

   pOutfile->EndLevel();
}

void     cGridComp::MakeBackup()
{
   if ( fpBackup )
      delete fpBackup;
      
   fpBackup = NULL;
   
   if ( GetNumMines() > 0 ) 
      fpBackup = new cGridComp(this, NULL, true);
}

void     cGridComp::RestoreBackup()
{
   cGridComp* pBackup = (cGridComp*) fpBackup;
   
   if ( pBackup ) {
      fSpacing = pBackup->fSpacing;
      CopyMineCompParams(pBackup);
         // delete existing mines, but don't update displays yet
      if ( gbGui )
	tactmap_bundle_changes (gpTactmap);
      while (fpMineListHead)
         delete fpMineListHead;

      cPlanMine* pNew;
      for (cMine* pM = pBackup->GetFirstMine();
                  pM; pM = pM->GetNextMine())
         pNew = new cPlanMine(this, pM, false );

      UpdateStatusInfo();
   }
}

cMfComponent*    cGridComp::DuplicateComp(T2dMove& move, cMinefield* pParent,
					  EBoolean isBackup)
      // duplication, or a backup when an entire minefield is backed up
      //  MakeBackup is called when only the component is backed up
{
   cGridComp* pC = new cGridComp(this, pParent, isBackup);

   if ( ! isBackup) {
     pC->MoveCompOnly(move);
     pC->SetCompMineType(pC->GetDbEntity());
     if (pC->fpCompEntry)
       pC->fpCompEntry->ShowInfo();
     if (pC->fpCompStatus)
       pC->fpCompStatus->ShowInfo();
   }
   return pC;
}

void cGridComp::EmplaceGrid (int nEstMines)
{
  TPerimPoint        *pPt;
  cPerimeter*         pPerim = fpParentMf->GetPerimeter();
  TLocation           tccPt, *pTccPts;
  float64             xmin =  DBL_MAX, ymin =  DBL_MAX, 
                      xmax = - DBL_MAX, ymax = - DBL_MAX,
                      x, y;
  float               rand_max = pow(2., 15.) - 1.;
  int                 nMines = 0;
  int                 arrayDim = 1.1 * (nEstMines + 5);	// allow some slop

  if (! pPerim) return;

  gpGuiWin->SetSelection( false, fpParentMf, this );
  if (fMineSumm.fTotalMines == 0) {
    if (gbGui && fSpacing > 0.)
      tactmap_bundle_changes (gpTactmap);
  }
  else
    DeleteMines();

  if (fSpacing <= 0.) return;

  for (pPt = pPerim->GetFirstPoint(); pPt; pPt = pPerim->GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    if (tccPt.x < xmin) xmin = (int)tccPt.x;
    if (tccPt.x > xmax) xmax = (int)tccPt.x;
    if (tccPt.y < ymin) ymin = (int)tccPt.y;
    if (tccPt.y > ymax) ymax = (int)tccPt.y;
  }

  if (xmax <= xmin || ymax <= ymin) return;
  
  pTccPts = new TLocation[arrayDim];

  for (x = xmin; x < xmax; x += fSpacing) {
    for (y = ymin; y < ymax; y += fSpacing) {
      pTccPts[nMines].x = x + .5 * fSpacing * (float)rand() / rand_max;
      pTccPts[nMines].y = y + .5 * fSpacing * (float)rand() / rand_max;
      if (pPerim->PointInside (pTccPts[nMines]))
         if (++nMines >= arrayDim ) break;
    }
    if ( nMines >= arrayDim ) break;
  }

  AddTccMines (pTccPts, nMines);
  
  delete[] pTccPts;
}

void cGridComp::SetSpacing (float spacing, int nEstMines)
{
  if (spacing <= 0.) return;
  fSpacing = spacing;
  EmplaceGrid (nEstMines);
}

void   cGridComp::InitGridComp(cMinefield* pParentMf, EBoolean isBackup)
{
   fCompType = eCompGrid;
   fSpacing = 0.;
   fpBackup = NULL;

   if ( ( ! isBackup ) && gbGui) {
      fpCompEntry = new cMineCompSummLine
            (this, pParentMf->GetCompList()->GetMainWin());
      fpCompStatus = new cCompStatusLine 
            (this, pParentMf->GetCompStatus()->GetMainWin());
   }
   else {
     fpCompEntry = NULL;
     fpCompStatus = NULL;
  }
}

cGridComp::cGridComp(cGridComp* pOriginal, cMinefield* pParentMf,
		     EBoolean isBackup):   // duplication or backup
   cMineComp(pOriginal, pParentMf, isBackup)
{
   InitGridComp(pParentMf, isBackup);

   fSpacing = pOriginal->fSpacing;
}

cGridComp::cGridComp(cMinefield* pParentMf) : cMineComp(pParentMf)
{
   InitGridComp(pParentMf, false);
}

cGridComp::~cGridComp()
{
}

void   cRowComp::ReadRowMines(cInFileObj* pMineFile)
{
   EFileReturn   returnCode;
   int           returnIndex;
   void*         returnVal;
   TItemList *   list;
   int           i;
   TEntityType   type;
      
   returnCode = pMineFile->ReadItem(fKeyRowTable, eSizeRowTable, &returnIndex, &returnVal);
   
   fRefSet    = eMfRef;

   switch (returnCode) {
      case eFileNoErr:
         switch (returnIndex) {
      
            case erowkwSpacing:
                  fSpacing = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case erowkwOffset:
                  fOffset = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
               break;
      
            case erowkwMfRowCoords:
//                  fRefSet    = eMfRef;
                  list = (TItemList *)returnVal;
                  fNumPoints = list->count;
               
                  fPointList.DeleteList();
                     
                  for (i = 0; i < fNumPoints; i++)
                     new TPerimPoint( ((TFlatEntCoord *)(list->ptrItems))[i], fPointList.GetLastLink());
                        // don't yet have tcc coords, to do map display
                  delete[] list->ptrItems;
                  delete list;
                  list = NULL;
                  returnVal = NULL;
               break;
               
            }
            new cReadRowMsg(this, pMineFile);
         break;

		case eBadKeyword:
            cMineComp::ParseMineItem(pMineFile);
            new cReadRowMsg(this, pMineFile);
         break;

      case eLevelDone:   // done reading this component
            new cReadMinefieldMsg(fpParentMf, pMineFile);
               // check that we have enough info
            if ( (fMineSumm.fTotalMines == 0) || (fpDbMine == NULL) ) 
               delete this;
            else
               fpParentMf->SetMineSumm();
         break;

      default: // error returns.  Give the user a message
            Assert(false, kerrReadMinefield, "cRowComp::ReadRowMines:4");
            delete pMineFile;
         break;
   }
}

cMfComponent*    cRowComp::DuplicateComp(T2dMove& move, cMinefield* pParent,
					 EBoolean isBackup)
{
   cRowComp* pC = new cRowComp(this, pParent, isBackup);

   if ( ! isBackup ) {
      pC->MoveCompOnly(move);
      pC->SetCompMineType(pC->GetDbEntity());
      if (pC->fpCompEntry)
	 pC->fpCompEntry->ShowInfo();
      if (pC->fpCompStatus)
	 pC->fpCompStatus->ShowInfo();
   }
   return pC;
}

void     cRowComp::MakeBackup()
{
   if ( fpBackup )
      delete fpBackup;
      
   fpBackup = NULL;
   
   if ( GetNumMines() > 0 ) 
      fpBackup = new cRowComp(this, NULL, true);
}

void     cRowComp::RestoreBackup()
{
   cRowComp* pBackup = (cRowComp*) fpBackup;
   
   if ( pBackup ) {
      CopyMineCompParams(pBackup);
      fSpacing = pBackup->fSpacing;
      fOffset  = pBackup->fOffset;
         // delete existing mines and rows, but don't update displays yet
      if ( gbGui )
	tactmap_bundle_changes (gpTactmap);
      if ( fpRowDisp ) 
         for ( int i = fNumPoints; i > 0; i-- )
            fpRowDisp->DeleteVertex(i);
      fPointList.DeleteList();   // deletes each point
      fNumPoints = 0;
      while (fpMineListHead)
         delete fpMineListHead;

      
      for (TPerimPoint* pPt = pBackup->GetFirstPoint(); pPt; 
                           pPt = pBackup->GetNextPoint() ) {
         new TPerimPoint (pPt->GetTccPoint(), fPointList.GetLastLink());
         ++fNumPoints;
         if ( fpRowDisp )
            fpRowDisp->AddVertex (fNumPoints, pPt->GetTccPoint());
         
      }

      cPlanMine* pNew;
      for (cMine* pM = pBackup->GetFirstMine();
                  pM; pM = pM->GetNextMine())
         pNew = new cPlanMine(this, pM, false );

      UpdateStatusInfo();
   }
}

int cRowComp::RowMineCount (float spacing)
{
  TPerimPoint        *pPt;
  TLocation           tccPt, lastPt, newPt;
  float64             xdelta, ydelta, dist;
  int                 count = 0;

  if (fNumPoints < 2) return 0;

  pPt = GetFirstPoint();
  lastPt = pPt->GetTccPoint();

  for (pPt = GetNextPoint(); pPt; pPt = GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    xdelta = tccPt.x - lastPt.x;
    ydelta = tccPt.y - lastPt.y;
    dist = sqrt (xdelta * xdelta + ydelta * ydelta);
    count += (int)(dist / spacing) + 1;

    lastPt = tccPt;
  }
  return count;
}

void cRowComp::SetSpacing (float spacing, float offset, int nEstMines)
{
  TPerimPoint        *pPt;
  TLocation           tccPt, lastPt, newPt, *pTccPts;
  float64             xdelta, ydelta, xstep, ystep, dist, xoff, yoff;
  int                 i, nsteps, nMines = 0;

  fSpacing = spacing;
  fOffset = offset;

  gpGuiWin->SetSelection( false, fpParentMf, this );

  if (fMineSumm.fTotalMines == 0) {
    if (gbGui && spacing > 0.)
      tactmap_bundle_changes (gpTactmap);
  }
  else
    DeleteMines();

  if (spacing <= 0.) return;
  if (fNumPoints < 2) return;

  fRefSet = eTccRef;
  if ( nEstMines <= 0)
     nEstMines = RowMineCount(spacing);
  int   arrayDim = 1.1 * (nEstMines + 5);	// allow some slop
  pTccPts = new TLocation[arrayDim];

  // space mines on row with offset

  pPt = GetFirstPoint();
  lastPt = pPt->GetTccPoint();

  for (pPt = GetNextPoint(); pPt; pPt = GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    xdelta = tccPt.x - lastPt.x;
    ydelta = tccPt.y - lastPt.y;
    dist = sqrt (xdelta * xdelta + ydelta * ydelta);
    xstep = spacing * xdelta / dist;
    ystep = spacing * ydelta / dist;
    nsteps = (int)(dist / spacing);

    xoff = -ystep * offset / spacing;
    yoff =  xstep * offset / spacing;
    newPt = lastPt;

    for (i = 0; i <= nsteps; i++) {
      pTccPts[nMines].x = newPt.x + ((i % 2) ? 1 : -1) * xoff;
      pTccPts[nMines].y = newPt.y + ((i % 2) ? 1 : -1) * yoff;
      if (++nMines >= arrayDim) break;
      newPt.x += xstep;
      newPt.y += ystep;
    }
    lastPt = pPt->GetTccPoint();
    if (nMines >= arrayDim) break;
  }

  AddTccMines (pTccPts, nMines);
  
  delete[] pTccPts;
}

TPerimPoint*  cRowComp::GetFirstPoint ()  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetFirstLink();

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

TPerimPoint*  cRowComp::GetNextPoint ()  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetNextLink();

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

TPerimPoint*  cRowComp::GetLastPoint ()  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetLastLink();

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

TPerimPoint*  cRowComp::GetPointNum (int iPt)  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetLinkNum (iPt);

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

EBoolean  cRowComp::CompleteComp()
{
   TPerimPoint* pPt;
   int   i;
   
      // the row coords and mines in it completed as other components
   ERefSet oldRefSet = fRefSet;
   cMineComp::CompleteComp();
   
      // work on the points in the row, similar to perimeter
   switch (oldRefSet) {
      case eTccRef:
            for ( pPt = (TPerimPoint*) (fPointList.GetFirstLink()); 
                  pPt != (TPerimPoint*) &fPointList; 
                  pPt = (TPerimPoint*) (fPointList.GetNextLink()) ) 
               fpParentMf->MapToMfLoc( pPt->GetTccPoint(), &(pPt->GetMfPoint()) );
         break;
         
      case eMfRef:
            for ( pPt = (TPerimPoint*) (fPointList.GetFirstLink()), i = 1; 
                  pPt != (TPerimPoint*) &fPointList; 
                  pPt = (TPerimPoint*) (fPointList.GetNextLink()), i++ ) {
                     // have 2D mf ref, need both tcc and mf elevation
               TFlatEntCoord loc2D(pPt->GetMfPoint().x, pPt->GetMfPoint().y);
               fpParentMf->MfToMapLoc( loc2D, &(pPt->GetTccPoint()) );
               fpParentMf->MapToMfLoc(pPt->GetTccPoint(), &(pPt->GetMfPoint()) );
               
               if ( fpRowDisp ) {
                  fpRowDisp->AddVertex (i, pPt->GetTccPoint());
               }
            }
          break;
        
      case eRefComplete:
         break;
      
      default:
            return false;
         // break;
   }
   
   fRefSet = eRefComplete;
   return true;
}

void    cRowComp::ComputeCompLoc()
   // called after SetSpacing, used to provide center location before
   //   row is completed.
   // If no mines have been created, uses row point locations
{
   cMine*        pTempMine;
   TPerimPoint*  pPt;
   TLocation     loc;

   double xMin = DBL_MAX;
   double xMax = - DBL_MAX;
   double yMin = DBL_MAX;
   double yMax = - DBL_MAX;
   double zMin = DBL_MAX;
   double zMax = - DBL_MAX;
   
   Assert(fRefSet == eTccRef, kerrValidityCheck, "cRowComp::ComputeCompLoc");

   if ( GetNumMines() > 0 ) {
      for (pTempMine=GetFirstMine(); pTempMine; pTempMine=pTempMine->GetNextMine() ) {
         loc = pTempMine->GetTccLoc();
         if ( loc.x < xMin ) xMin = loc.x;
         if ( loc.x > xMax ) xMax = loc.x;
         if ( loc.y < yMin ) yMin = loc.y;
         if ( loc.y > yMax ) yMax = loc.y;
         if ( loc.z < zMin ) zMin = loc.z;
         if ( loc.z > zMax ) zMax = loc.z;
      }
   }
   else {
      for ( pPt = (TPerimPoint*) (fPointList.GetFirstLink()); 
            pPt != (TPerimPoint*) &fPointList; 
            pPt = (TPerimPoint*) (fPointList.GetNextLink()) ) {
         loc = pPt->GetTccPoint();
         if ( loc.x < xMin ) xMin = loc.x;
         if ( loc.x > xMax ) xMax = loc.x;
         if ( loc.y < yMin ) yMin = loc.y;
         if ( loc.y > yMax ) yMax = loc.y;
         if ( loc.z < zMin ) zMin = loc.z;
         if ( loc.z > zMax ) zMax = loc.z;
      }
   }

   fTccLocation.x = (xMin + xMax) / 2.0;
   fTccLocation.y = (yMin + yMax) / 2.0;
   fTccLocation.z = (zMin + zMax) / 2.0;
}

void   cRowComp::InitRowComp(cMinefield* pParentMf, EBoolean isBackup)
{
   fCompType = eCompRow;
   fpBackup = NULL;
   fNumPoints = 0;
   fRefSet    = eTccRef;      // default, when created by gui

   if ( ( ! isBackup ) && gbGui) {
     fpCompEntry = new cMineCompSummLine
       (this, pParentMf->GetCompList()->GetMainWin());
     fpCompStatus = new cCompStatusLine 
       (this, pParentMf->GetCompStatus()->GetMainWin());
     fpRowDisp = new cRowDisp (this);
   }
   else {
     fpCompEntry = NULL;
     fpCompStatus = NULL;
     fpRowDisp = NULL;
   }
}

cRowComp::cRowComp(cRowComp* pOriginal, cMinefield* pParentMf,
		   EBoolean isBackup):   // duplication or backup
   cMineComp(pOriginal, pParentMf, isBackup)
{
   InitRowComp(pParentMf, isBackup);

   fSpacing = pOriginal->fSpacing;
   fOffset  = pOriginal->fOffset;

   TPerimPoint* pNewPt;
   for (TPerimPoint* pPt = pOriginal->GetFirstPoint();
                     pPt; 
	                  pPt = pOriginal->GetNextPoint() ) {
      pNewPt = new TPerimPoint (pPt->GetTccPoint(), fPointList.GetLastLink());
//      pNewPt->SetMfPoint( pPt->GetMfPoint() );
      ++fNumPoints;
      if ( fpRowDisp )
         fpRowDisp->AddVertex (fNumPoints, pNewPt->GetTccPoint());
   }
}

cRowComp::cRowComp(cMinefield* pParentMf) : cMineComp(pParentMf)
{
   fSpacing = fOffset = 0.;
   InitRowComp(pParentMf, false);
}

cRowComp::~cRowComp()
{
   fPointList.DeleteList();
   if (fpRowDisp)  delete fpRowDisp;
}

EBoolean    cRowComp::DeletePoint (int iPt)
{
  fPointList.DeletePoint (iPt);
  fNumPoints--;
  fpRowDisp->DeleteVertex (iPt - 1);
  SetSpacing (fSpacing, fOffset);
  return true;
}

EBoolean    cRowComp::MoveCompOnly (T2dMove& move)
   // moves component around its own center, but not minefield
   //  We move each vertex point and mine
{
  TPerimPoint*  pPt;
  TLocation     tccPt;
  TEntCoord     mfLoc;
  int           iPt = 0;

  fRefSet = eTccRef;

  for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
    move.fRotMatrix.Mult(pPt->GetTccPoint() - fTccLocation, &tccPt);
    tccPt.x = tccPt.x + fTccLocation.x + move.fDelX;
    tccPt.y = tccPt.y + fTccLocation.y + move.fDelY;
    tccPt.z =  gpTerrainMgr->GetMapElev(tccPt.x, tccPt.y);
    if (! GetMinefield()->GetPerimeter()->PointInside (tccPt)) {
      if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
					  "Invalid move.  Row must be emplaced\ninside minefield perimeter.");
      return false;
    }
  }

  for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
    move.fRotMatrix.Mult(pPt->GetTccPoint() - fTccLocation, &tccPt);
    tccPt.x = tccPt.x + fTccLocation.x + move.fDelX;
    tccPt.y = tccPt.y + fTccLocation.y + move.fDelY;
    tccPt.z =  gpTerrainMgr->GetMapElev(tccPt.x, tccPt.y);
    pPt->SetTccPoint(tccPt);
    fpParentMf->MapToMfLoc(tccPt, &mfLoc);
    pPt->SetMfPoint(mfLoc); 

    fpRowDisp->MoveVertex (iPt++, tccPt);
  }
//  SetSpacing (fSpacing, fOffset);
//  CompleteComp();
//  return true;

   return cMineComp::MoveCompOnly(move);   // moves mines, completes
}

EBoolean  cRowComp::MoveComp(T2dMove& move)
   // minefield is moving.  Rotate in mf coords, let tcc and geo
   // get set on completion of move.
{
   TPerimPoint*  pPt;
   TEntCoord     mfLoc;
   TLocation     tccPt;
   EBoolean success = true;
   
      // make sure we have a component location
   if ( fRefSet != eRefComplete )
      success = CompleteComp();

   if ( success ) {
      fMfLocation = move.fRotMatrix * fMfLocation;
      fRefSet = eMfRef;
   
      for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
         move.fRotMatrix.Mult(pPt->GetMfPoint(), &mfLoc);
         pPt->SetMfPoint(mfLoc);
      }
      
      for (cMine* pM = GetFirstMine(); pM && success; pM = pM->GetNextMine() ) 
         success = pM->MoveMineWithMf(move);
   }
   
   return success;
}

EBoolean    cRowComp::MoveRow (TLocation offset)
{
  TPerimPoint*  pPt;
  TLocation     tccPt;
  int           iPt = 0;

  for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    tccPt.x += offset.x;
    tccPt.y += offset.y;
    if (! GetMinefield()->GetPerimeter()->PointInside (tccPt)) {
      if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
					  "Invalid move.  Row must be emplaced\ninside minefield perimeter.");
      return false;
    }
  }

  for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    tccPt.x += offset.x;
    tccPt.y += offset.y;
    pPt->SetTccPoint(tccPt); 

    fpRowDisp->MoveVertex (iPt++, tccPt);
  }
  SetSpacing (fSpacing, fOffset);
  return true;
}

EBoolean    cRowComp::MovePoint (int iPt, TLocation& tccPt)
{
  TPerimPoint*  pPt;

  if (GetMinefield()->GetPerimeter()->PointInside (tccPt)) {
    pPt = (TPerimPoint*) fPointList.GetLinkNum(iPt);
    pPt->SetTccPoint(tccPt);
    fpRowDisp->MoveVertex (iPt - 1, tccPt);
    SetSpacing (fSpacing, fOffset);
    return true;
  }
  else {
    if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
					"Invalid move.  Row must be emplaced\ninside minefield perimeter.");
    return false;
  }
}

EBoolean    cRowComp::AddTccPoint (TLocation& tccPt)
{
   /* is new point valid ? */
  if (GetMinefield()->GetPerimeter()->PointInside (tccPt)) {
    tccPt.z = gpTerrainMgr->GetMapElev( tccPt.x, tccPt.y);
    new TPerimPoint (tccPt, fPointList.GetLastLink());
   
    ++fNumPoints;
    if ( fpRowDisp )
      fpRowDisp->AddVertex (fNumPoints, tccPt);

    fRefSet = eTccRef;
    SetSpacing (fSpacing, fOffset, 0);
    ComputeCompLoc();
    if ( gbGui )
      gpGuiWin->SetEditorButtonState ();
    return true;
  }
  else {
    if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
					"Invalid point.  Row must be emplaced\ninside minefield perimeter.");

    return false;
  }
}

void    cRowComp::SetCompPvdSelect( EBoolean on )
{
   cMineComp::SetCompPvdSelect( on );
   fpRowDisp->SetRowSelectSens( on );
}

