/**********************************************************************
 * File: DisPdu203Exp.h  Contains type, class and enumeration definitions for 
 *                       experimental PDUs.
 *
 * Types defined: cpduEnvironmental
 *
 * Functions defined:
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1997
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */

#ifndef _experimentalPdus__
#define _experimentalPdus__
 
#include "dis203_types.h"
#include "dis203exp_values.h"
#include "dis203exp_types.h"

#include "dis204_types.h"
#include "dis204exp_values.h"
#include "dis204exp_types.h"

struct TEnvironmentRecord {
    uint32                                  fRecordKind;
    uint16                                  fRecordLength;
    uint8                                   fRecordIndex;
    uint8                                   _unused_padding;

    union {
        DIS203_ENV_PADDING_VARIANT                 padding;
        DIS203_ENV_STATE_COMBIC_VARIANT            combic_plume;
        DIS203_ENV_STATE_COMBIC_VARIANT            combic_puff;
        DIS203_ENV_STATE_FLARE_VARIANT             flare;
        DIS203_ENV_STATE_EPHEMERIS_VARIANT         ephemeris;
        DIS203_ENV_STATE_SIMPLE_WEATHER_VARIANT    simple_weather;
        DIS203_ENV_STATE_OBSERVED_WEATHER_VARIANT  observed_weather;
        DIS203_ENV_STATE_SIMULATION_DATE_VARIANT   simulation_date;
        DIS203_ENV_GEOMETRY_UNIFORM_VARIANT        uniform;
        DIS203_ENV_GEOMETRY_POINT_VARIANT          point;
        DIS203_ENV_GEOMETRY_SPHEROID_VARIANT       spheroid;
        DIS203_ENV_GEOMETRY_GAUSSIAN_PLUME_VARIANT gaussian_plume;
        DIS203_ENV_GEOMETRY_CONE_VARIANT           cone;
        DIS203_ENV_GEOMETRY_GAUSSIAN_PUFF_VARIANT  gaussian_puff;
        DIS203_ENV_GEOMETRY_POLYGON_VARIANT        polygon;
        DIS203_ENV_GEOMETRY_EXTENTS_VARIANT        extents;
        DIS203_ENV_STATE_VEH_DUST_VARIANT          vehicle_dust;
        DIS203_ENV_GEOMETRY_FLEX_PLUME_VARIANT     flex_plume;
    } fRecordVariant;
};

struct cpduEnvironmental {
    DIS203_PDU_HEADER_RECORD                fHeader;
    DIS203_ENTITY_ID_RECORD                 fEnvId;
    DIS203_ENTITY_TYPE_RECORD               fEnvType;

    uint8                                   fEnvModel;
    uint8                                   fEnvStatus;
    uint16                                  fNumberRecords;
    uint16                                  _unused_padding;

    TEnvironmentRecord                      fpEnvRecord[1];
};


// Create a class which encapsulates TEnvironmentRecord and can 
// process(translate) each of the records.
class cTransEnvRecord : private TEnvironmentRecord
{
public:
    // Declare functions to process (translate) environment records.
    void ProcessSimpleWeatherRecord();
};


/************************************************************************
 * MAK Stealth View Control PDU prototype.
 ***********************************************************************/


struct TAbsoluteParams
{
    uint32	            	        ignoreFlags;
    uint32                              padding;
    DIS203_WORLD_COORDINATES_RECORD	world;
    DIS203_EULER_ANGLES_RECORD	        orientation;
};

struct TTetherParams
{
    uint32			         ignoreFlags;
    DIS203_VECTOR_RECORD                  offset;
    DIS203_EULER_ANGLES_RECORD	         orientation;
    DIS203_ENTITY_ID_RECORD	         attachId;
};

struct TCompassParams
{
    uint32       		ignoreFlags;
    float32    		        radius;
    float32    		        bodyHeading;
    float32    		        pitch;
    float32    		        headingOffset;
    float32    		        pitchOffset;
    float32	    	        roll;
    DIS203_ENTITY_ID_RECORD	attachId;	
};

struct TMimicParams
{
    uint32			                ignoreFlags;
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD	offset;
    DIS203_EULER_ANGLES_RECORD		        bodyOrientation;
    DIS203_ENTITY_ID_RECORD	                attachId;
};

struct TOrbitParams
{
    uint32 			ignoreFlags;
    float32    		        radius;
    float32    		        azimuthRate;
    float32       		heading;
    float32       		pitch;
    float32       		headingOffset;
    float32      		pitchOffset;
    float32	       	        roll;
    DIS203_ENTITY_ID_RECORD	attachId;
};

struct TTrackParams
{
    uint32    	                       ignoreFlags;
    uint32                             padding;
    DIS203_WORLD_COORDINATES_RECORD    world;
    DIS203_ENTITY_ID_RECORD	       trackId;
};

struct TTetherTrackParams
{
    uint32			             ignoreFlags;
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD   offset;
    DIS203_ENTITY_ID_RECORD	             attachId;
    DIS203_ENTITY_ID_RECORD	             trackId;
};

struct TMimicTrackParams
{
    uint32			             ignoreFlags;
    DIS203_ENTITY_COORDINATE_VECTOR_RECORD   offset;
    DIS203_ENTITY_ID_RECORD	             attachId;
    DIS203_ENTITY_ID_RECORD	             trackId;
};

#define MAX_EIG_ENTITIES	100

struct TEntitiesInGroup
{
    uint32			numToRemove;	/* -1 means remove all */
    uint32			numToAdd;
    DIS203_ENTITY_ID_RECORD	groupId[MAX_EIG_ENTITIES];
};

struct TGroupParams
{
    uint32 		ignoreFlags;
    float32       	radialOffset;
    float32       	heading;
    float32       	pitch;
    float32      	headingOffset;
    float32       	pitchOffset;
    uint32		groupCmd;
    TEntitiesInGroup 	groupIds;
};

struct TModelessParams
{
    uint32 			ignoreFlags;
    DIS203_ENTITY_ID_RECORD 	primaryId;
    DIS203_ENTITY_ID_RECORD 	secondaryId;
    TEntitiesInGroup 		groupIds;
};
									
union UNetViewControlParams
{
    TAbsoluteParams 		absolute;
    TTetherParams		tether;
    TCompassParams		compass;
    TMimicParams		mimic;
    TOrbitParams		orbit;
    TTrackParams		track;
    TTetherTrackParams		tetherTrack;
    TMimicTrackParams		mimicTrack;
    TGroupParams		group;
    TModelessParams		modeless;
};
typedef UNetViewControlParams UViewControlParams;


struct TViewControlPdu
{
    DIS203_PDU_HEADER_RECORD       header;
    DIS203_ENTITY_ID_RECORD        viewerId;       /* The ID of the stealth */
    DIS203_ENTITY_ID_RECORD        senderId;
    uint16	       	           unused0;
    uint16                         command;        /* attach, teleport, etc. */
    uint32       	           unused;
    UNetViewControlParams	   params;	       
};

#endif
