/*
$RCSfile: Distance.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:32 $
$Log: Distance.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:32  jchappel
Comprehensive Mine Simulator

// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.9  1996/08/26  14:20:30  mkappel
// Improve handling of sensitivity.
//
// Revision 1.8  1996/08/20  20:11:35  mkappel
// Add intervisibility to Distance Tool.
//
// Revision 1.7  1996/07/13  19:04:20  rcarp
// version 1.5b13, fixes bug dragging with two buttons and displays damamged vehicles
//
// Revision 1.6  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.5  1996/06/19  19:05:45  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.4  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.3  1996/04/04  17:19:20  mkappel
// Fix creation of Text Dfn.
//
// Revision 1.3  1996/04/04  17:19:20  mkappel
// Fix creation of Text Dfn.
//
// Revision 1.2  1996/04/02  19:58:14  mkappel
// Add sensitivity at end of drag.
//
// Revision 1.1  1996/03/30  19:35:34  mkappel
// Add Distance tool.
//
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: Distance.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/


#include "Distance.h"
#include <bstring.h>
#include <math.h>
#include <stdext.h>

SNSTVE_CLASS cDistanceDisp::fSensNull;
SNSTVE_CLASS cDistanceDisp::fSensDist;
GC           cDistanceDisp::fDistanceGC;
XFontStruct* cDistanceDisp::fFontInfo;
int          cDistanceDisp::fDispSensIndex = 0;


void cDistanceDisp::InitGCs (Widget wShell)
{
   XGCValues         gcv;

   struct dist_disp_resources
     {
       Pixel color;
       int   line_width;
       Font  designations;
     } resources;

   static XtResource res_list[] = {
     { "color" , "Color" , XmRPixel , sizeof(Pixel) ,
       XtOffset(struct dist_disp_resources *, color) ,
       XmRString , "Yellow" } ,
     { "line_width" , "Line_width" , XmRInt , sizeof(int) ,
       XtOffset(struct dist_disp_resources *, line_width) ,
       XmRString , "2" } ,
     { "designations" , "Designations" , XmRFont , sizeof(Font) ,
       XtOffset(struct dist_disp_resources *, designations) ,
       XmRString , "Fixed" } ,
   };

   /* Get the colors, and various other resources used */
   XtGetSubresources (wShell, &resources, "dist", "DIST",
		      res_list, XtNumber(res_list), NULL, 0);

   gcv.foreground = resources.color;
   gcv.background = tactmap_get_color (gpTactmap,(TACTMAP_OBJECT_CLASS)-1);
   gcv.line_width = resources.line_width;
   gcv.font = resources.designations;
   fDistanceGC = XCreateGC (gpDisplay, gMapWindow, 
			    GCForeground|GCBackground|GCLineWidth|GCFont, &gcv);
   fFontInfo = XQueryFont (gpDisplay, gcv.font);
   SNSTVE_INIT_CLASS(fSensNull);
   SNSTVE_INIT_CLASS(fSensDist);
   callback_register_handler(fSensDist.gesture, 
			     (CALLBACK_HANDLER)CbDelete,
			     "distance");
   fSensDist.buttons = 0x8; /* binary 1000, Button 3 */
   fSensDist.sensitive = fSensDist.hot = 1;
}

cDistanceDisp::cDistanceDisp (SNSTVE_CALL_DATA *call_data)
{
  int32 dummyc;
  bzero(&fDfn, sizeof(TACTMAP_OBJECT_DFN));
  fDfn.sensitive.sensClass = &fSensNull;

  fDfn.line_intervis.viewer_agl = 2.5;
  fDfn.line_intervis.target_agl = 5.0;
  fDfn.line_intervis.target_width = 5.0;
  fDfn.line_intervis.target_height = 5.0;
/*removed to match ModSAF3.0 libtactmap.h AUSTIN 4 Aug 97*/
/*
  fDfn.line_intervis.tree_opacity = 0.95;
*/
  fDfn.line_intervis.n_veh = 0;
  fDfn.line_intervis.show_percentages = 0;
  fDfn.line_intervis.grid_interval = gpGuiWin->GetMapScale() / 50.;
  fDfn.line_intervis.environ_effects = TACTMAP_INTERVIS_TERRAIN;

  /* Infer the click location */
  /*modified to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
tactmap_screen_to_map (gpTactmap,
         (int32)(call_data->screen_x - call_data->xoff),
         (int32)(call_data->screen_y - call_data->yoff),
         &fXStart, &fYStart, &dummyc);
  fDfn.line_intervis.point1[0] = fXStart;
  fDfn.line_intervis.point1[1] = fYStart;

  fpObj = tactmap_create_object (gpTactmap, TACTMAP_LINE_INTERVIS, &fDfn);
  tactmap_add_transient_object (gpTactmap, fpObj);

  bzero(&fTextDfn, sizeof(TACTMAP_OBJECT_DFN));
  fTextDfn.sensitive.sensClass = &fSensNull;

  fTextDfn.text.gc = fDistanceGC;
  fTextDfn.text.font_struct = fFontInfo;

  fpTextObj = tactmap_create_object (gpTactmap, TACTMAP_TEXT, &fTextDfn);
  tactmap_add_transient_object (gpTactmap, fpTextObj);
}

void cDistanceDisp::UpdateDist (SNSTVE_CALL_DATA *call_data)
{
  float64            x, y, xtext, ytext;
  int32		     c;
  char               txt[15];
  int                dist, ascent, descent, dir;
  XCharStruct        overall;

  /*modified to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
  tactmap_screen_to_map (gpTactmap,
			 (int32)call_data->screen_x, 
			 (int32)call_data->screen_y,
			 &x, &y, &c);
  fDfn.line_intervis.point2[0] = x;
  fDfn.line_intervis.point2[1] = y;
  tactmap_update_object (gpTactmap, fpObj, &fDfn);

  dist = (int)fsqrt ((float)((fXStart - x) * (fXStart - x) + 
			     (fYStart - y) * (fYStart - y)));
  sprintf (txt, "%d", dist);
  fTextDfn.text.string = cppstrdup (txt);
  XTextExtents (fFontInfo, fTextDfn.text.string, strlen(fTextDfn.text.string), 
		&dir, &ascent, &descent, &overall);
  /*modified to match modsaf3.0 libtactmap.h AUSTIN 5 Aug 97*/
  tactmap_screen_to_map (gpTactmap,
			 (int32)(call_data->screen_x - overall.width),
			 (int32)(call_data->screen_y - ascent - descent),
			 &xtext, &ytext, &c);
  if (x < fXStart) fTextDfn.text.x = xtext;
  else             fTextDfn.text.x = x;
  if (y > fYStart) fTextDfn.text.y = ytext;
  else             fTextDfn.text.y = y;

  tactmap_update_object (gpTactmap, fpTextObj, &fTextDfn);
}

void cDistanceDisp::CompleteDist ()
{
  fTextDfn.sensitive.sensClass = &fSensDist;
  fTextDfn.sensitive.instance = ++fDispSensIndex;
  fTextDfn.sensitive.user_data = (char *)this;
  tactmap_update_object (gpTactmap, fpTextObj, &fTextDfn);
}

void cDistanceDisp::CbDelete (SNSTVE_WINDOW_PTR /* sensitive */, 
			      SNSTVE_ACTION action,
			      SNSTVE_CALL_DATA *call_data, 
			      char * /* name */ )
{
  if (call_data->button == 3 && action == SNSTVE_RELEASE) 
    delete (cDistanceDisp*)call_data->user_data;
}

cDistanceDisp::~cDistanceDisp ()
{
  tactmap_remove_object (gpTactmap, fpObj);
  tactmap_destroy_object (gpTactmap, fpObj);
  delete[] fTextDfn.text.string;
  fTextDfn.text.string = NULL;
  tactmap_remove_object (gpTactmap, fpTextObj);
  tactmap_destroy_object (gpTactmap, fpTextObj);
}


