/**********************************************************************
 * File: Emission.cc
 *
 * Types defined: cLinkEmiss, cMineLinkEmiss, cEmissListHead, cEmission
 *                cEmissionMgr
 *
 * Functions defined: 
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1996
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */
#ifndef _EMISSION_H_
#define _EMISSION_H_

#include "EntityEnums.h"
#include "Globals.h"
// #include "Area.h"
/*
#include "p_dis203.h"
*/
#include "dis203_values.h"
#include "dis204_values.h"
#include "Vehicle.h"

const int MAX_EMISSIONS = 100;

/*
class ELLIPSE_DATA
{
public:
        float a;
        float b;
	float MAX_AB;
        float a2;
        float b2;
	float a2b2;
        TOrient orientation;
	float COS_PSI;
	float SIN_PSI;
        float x_offset;
        float y_offset;
	float z_offset;

	ELLIPSE_DATA() {};
	~ELLIPSE_DATA(){};
};
*/

struct TEmissionSystem 
{
	// part of SP_DIS203_EMISSION_SYSTEM_RECORD
	unsigned char	fSystemDataLen;
	unsigned char	fNumBeams;
	
	// part of SP_DIS203_EMITTER_SYSTEM_RECORD
	unsigned short	fEmitterName;
	unsigned char	fFunction;
	unsigned char	fEmitterID;
	
	TEntCoord	fRelEmitterLoc;
	

	// part of SP_DIS203_EMISSION_BEAM_RECORD
	unsigned char	fbeam_data_length;
	unsigned char	fbeam_id_number;
	unsigned short	fbeam_parameter_index;
	
	// part of SP_DIS203_FUNDAMENTAL_PARAMETER_DATA_RECORD
	float		ffrequency;
	float		ffrequency_range;
	float		fERP;
	float		fPRF;
	float		fpulse_width;
	float		fbeam_azimuth_center;
	float		fbeam_azimuth_sweep;
	float		fbeam_elevation_center;
	float		fbeam_elevation_sweep;
	float		fbeam_sweep_sync;

	// part of SP_DIS203_EMISSION_BEAM_RECORD
	unsigned char	fbeam_function;
	unsigned char	fnumber_of_targets_in_track_jam_field;
	unsigned char	fhigh_density_track_jam;
	unsigned long	fjamming_mode_sequence;
	
	
	// SP_DIS203_SIMULATION_ADDRESS_RECORD
	unsigned short	fsite_id;
	unsigned short	fapplication_id;
	
	// part of SP_DIS203_ENTITY_ID_RECORD
	unsigned short	fentity_id;

	// part of SP_DIS203_TRACK_JAM_RECORD
	unsigned char	femitter_id;
	unsigned char	fbeam_id;	

};


class cEmission 
{
friend class cDisPduTranslator;

public:
	cEmission();
	cEmission(int numsys, uint8 emitter_name);
	//cEmission(TEntCoord emitter_location, TOrient emitter_orient, float erp, SP_DIS203_EMISSION_PDU* pdu);
	~cEmission();
	
	unsigned short GetEmitterName() const { return fpSystem->fEmitterName; }
	
//	void cEmission::GetAffectedArea(TEntCoord pLocation[]);
	TEntCoord GetRelLoc()	{ return fpSystem->fRelEmitterLoc; }
	TLocation GetTccLoc();
	TLocation GetDrTccLoc(TTimeSim fTime);
	TOrient GetTccOrient();
	float GetERP() { return fpSystem->fERP; };

/*
	int IsInAffectedAreaEllipse(TEntCoord pt);
	int IsInAffectedAreaEllipse(TLocation spt);
	int IsInAffectedArea(T3Dsingle* pt);
	int IsInAffectedArea(T3Ddouble* pt);
	EBoolean ProcessEmission(float HorzRange);
	void    Rotate ( float psi, float theta, float phi, T3Dsingle* pt );

	float	GetRadius() const { return radius; }
*/

	const TVehDimensions&	GetEmissDim();

	cVehicle*	GetVehicle() const { return fpVeh; }
	
	void		ProcEmissionMsg(cEmission* pEmiss);

protected:
	cDbVeh* 	 fDbEmiss;
        TEmissionSystem* fpSystem;
	unsigned char	 fNumSystems;

        cVehicle*	fpVeh;
//	cArea*		fpArea;
	TEntCoord	fEmitterLoc;
	TOrient		fEmitterOrient;	
//	TEntCoord	fAffectedArea[4];

/*
	ELLIPSE_DATA   fEllipseData;
	TLocation Rotate (TLocation t);
 	void ComputeAffectedArea(TEntCoord ctr, TOrient orientation);
	void ComputeAffectedAreaEllipse(TEntCoord ctr, float fa, float fb, float sin_psi, float cos_psi);

*/

	unsigned char	state_update_indicator;
	
};


class cEmissionMsg : public cMsgPrEvent
{
	cEmission* fpEmission;
public:
        cEmissionMsg() {};
        cEmissionMsg(cEmission* pEmission)  { fpEmission = pEmission; };
        virtual ~cEmissionMsg() { /*if(fpEmission) delete fpEmission; fpEmission = NULL;*/};
	void SendMsg();

};

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/*
       //TEntityID target_entity_id;
        //SP_DIS203_EVENT_ID_RECORD             event_id;
        //SP_DIS203_STATE_UPDATE_INDICATOR_TYPE state_update_indicator; //unsigned char
        //SP_DIS203_NUMBER_OF_SYSTEMS_TYPE number_of_systems; //unsigned char
        //SP_DIS203_EMISSION_SYSTEM_RECORD systems[1];
                //SP_DIS203_SYSTEM_DATA_LENGTH_TYPE         system_data_length; //unsigned char
                //SP_DIS203_NUMBER_OF_BEAMS_TYPE            number_of_beams; //unsigned char
                //SP_DIS203_EMITTER_SYSTEM_RECORD           emitter_system;
                        //SP_DIS203_EMITTER_NAME_TYPE      emitter_name; //unsigned short
                        //SP_DIS203_EMITTER_FUNCTION_TYPE  function;  //unsigned char
                        //SP_DIS203_EMITTER_ID_NUMBER_TYPE emitter_id_number; //unsigned char
                //SP_DIS203_ENTITY_COORDINATE_VECTOR_RECORD location_with_respect_to_entity;
                        //SP_DIS203_VECTOR_COMPONENT_TYPE x;
                        //SP_DIS203_VECTOR_COMPONENT_TYPE y;
                        //SP_DIS203_VECTOR_COMPONENT_TYPE z;
                //SP_DIS203_EMISSION_BEAM_RECORD            beams[1];
                        //SP_DIS203_BEAM_DATA_LENGTH_TYPE             beam_data_length; // unsigned char
                        //SP_DIS203_BEAM_ID_NUMBER_TYPE               beam_id_number; //unsigned char
                        //SP_DIS203_BEAM_PARAMETER_INDEX_TYPE         beam_parameter_index; //unsigned short
                        //SP_DIS203_FUNDAMENTAL_PARAMETER_DATA_RECORD fundamental_parameter_data;
                                //SP_DIS203_EMISSION_FREQUENCY_TYPE       frequency; //float
                                //SP_DIS203_FREQUENCY_RANGE_TYPE          frequency_range; //float
                                //SP_DIS203_EFFECTIVE_RADIATED_POWER_TYPE effective_radiated_power; //float
                                //SP_DIS203_EMISSION_FREQUENCY_TYPE       pulse_repetition_frequency; //float
                                //SP_DIS203_PULSE_WIDTH_TYPE              pulse_width; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_azimuth_center; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_azimuth_sweep; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_elevation_center; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_elevation_sweep; //float
                                //SP_DIS203_BEAM_SWEEP_SYNC_TYPE          beam_sweep_sync; // float

                        //SP_DIS203_BEAM_FUNCTION_TYPE                beam_function; //unsigned char
                        //SP_DIS203_NUMBER_OF_TARGETS_TYPE            number_of_targets_in_track_jam_field; //unsigned char
                        //SP_DIS203_HIGH_DENSITY_TRACK_JAM_TYPE       high_density_track_jam; //unsigned char
                        //unsigned char                               _unused_42;  
                        //SP_DIS203_JAMMING_MODE_SEQUENCE_TYPE        jamming_mode_sequence; //unsigned long
                        //SP_DIS203_TRACK_JAM_RECORD                  track_jam[1];
                                //SP_DIS203_ENTITY_ID_RECORD       entity_id;
                                        //SP_DIS203_SIMULATION_ADDRESS_RECORD simulation_address;
                                                //SP_DIS203_SITE_ID_TYPE        site_id; //unsigned short
                                                //SP_DIS203_APPLICATION_ID_TYPE application_id; //unsigned short
                                        //SP_DIS203_ENTITY_ID_TYPE            entity_id; //unsigned short
                                //SP_DIS203_EMITTER_ID_NUMBER_TYPE emitter_id; //unsigned char
                                //SP_DIS203_BEAM_ID_NUMBER_TYPE    beam_id; //unsigned char





*/

class cEmissionMgr;
class cLinkEmiss;

class cEmissListHead : public  T2linkHead{

   friend class cEmissionMgr;
  
protected:

public:
      
      cEmission*   GetFirstEmiss();
      cEmission*   GetNextEmiss();
            
      EBoolean        IsInList(const cEmission* pEm, cLinkEmiss** ppEmissNode); //matches veh and returns node ptr
      
      void	ResetList();
      
                 cEmissListHead(cEmissListHead* pOldHead);  // takes over old list
                 cEmissListHead();
      virtual   ~cEmissListHead();
};

class cLinkEmiss : public T2link 
{
	friend class cEmissionMgr;
	friend class cEmissListHead;
	
protected:
	cEmission*	fpEmission;
//	float		fHorzRange;

public:
	cEmission*	GetEmission() const { return fpEmission; }
	unsigned short	GetEmitterType() const    { return fpEmission->GetEmitterName(); }
//	float 	        GetHorzRange() const { return fHorzRange; }


	cLinkEmiss(cEmission* pEm,  cLinkEmiss* pPrev);
	cLinkEmiss(cEmission* pEm,  cLinkEmiss* pPrev,  float hrange);
//	cLinkEmiss(cEmission* pEmission) { pNextEmission = NULL; }
	// data
};

//class cLinkEmissCross : 


class cMineLinkEmiss : public cLinkEmiss 
{
	friend class cEmissionMgr;
	friend class cEmissListHead;
	
protected:
	float	       fHorzRange;
	float	       fRange;
	cEntity*       fpRefEnt;      // entity providing ref coord system
	EBoolean       fbTgtVelRef; // use vehicle vel to get az rot angle
	float          fAzRot;       // from map to veh system, only if fbTgtVelRef
	TEntCoord      fLocation;
	TVelocity      fVelocity;
	TTimeSim       fTime;         // time of setting geometry
	EPartResult    fMagCmResult;
	
	float	       fLastY;	

	EBoolean       fbApproaching;
	

public:
	float 		    GetHorzRange() const { return fHorzRange; }
	void		    SetLocation();
   const TEntCoord&	    GetLocation() const   {return fLocation;}
         void		    GetLocTgtRef(TEntCoord* pLocTgtRef) const;
   const TVelocity&	    GetVelocity() const   {return fVelocity;}
         float		    GetAzRot() const      {return fAzRot;}
	 float		    GetRange() const	  { return fRange; }
	 float		    GetLastY() const      { return fLastY; }
	 void		    SetMagCmResult(EPartResult res)  { fMagCmResult=res; }
	 EPartResult	    GetMagCmResult()  { return fMagCmResult; }
	 
	 void		    UpdateSsrRec();
         
         void		    SnapZ();      // fixes elevation for detonation
	
         cMineLinkEmiss(cMineLinkEmiss* pPrEmiss, cLinkEmiss* pPrev, cEntity* pRefEnt = NULL, EBoolean refTgtVel = false);   // copy of other node pPrEmiss
         cMineLinkEmiss(cEmission* pEmiss, cLinkEmiss* pPrev, cEntity* pRefEnt,
                             EBoolean refTgtVel = false);
	 
//	cMineLinkEmiss(cEmission* pEmission) : cLinkEmiss( pEmission){ }
};


class cEmissionMgr {
      // cEmissionMgr manages the Emission lists, finds Emissions to
      // update from PDUs, etc.
      // **It has one instantiated object, no virtual functions
 
 protected:
	enum { eEmissmgrkwEmissType, eEmissmgrSizeTable };
   static const TKeywordEntry fEmissMgrTable[eEmissmgrSizeTable];
   
   cEmissListHead    fEmissList;         // fpNext is the head, fpPrev the tail
   
public:
   
   cEmission*   FindEmission(const TEntityID & vID, const unsigned short feN, cEmissListHead* pListHead);
   
   cEmissListHead&   GetEmissList()   {return fEmissList;}

  void		ForwardEmission(cEmission* fpEmiss);
  void		RemoveEmiss ( cEmission* pEmiss );      // remove Emission from list
/*
  void        ReadTestFile(char* pFileName);
   
   EBoolean    ReadTestEmiss(cInFileObj& infile, char* EmissTypeName);
*/
   void        EmissShutdown();
   
               cEmissionMgr();
              ~cEmissionMgr();

};




#endif
