 
/*******************************************

   file: FieldControl.cc
   
      Copyright 1998 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   FieldControl.cc
   
      The operator-controlled APLA implementation.
      
      
   
   Created by Richard Carpenter 11/98
      
*********************************************/

#include "Display.h"
#include "Minefield.h"
#include "Message.h"

#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>

cMinefield*         cFieldControlWindow::fpMinefield        = NULL;


void cFieldControlWindow::CbDone (Widget /*w*/, XtPointer  /*pObj*/,
					XmPushButtonCallbackStruct */*cbs*/)
{
   gpGuiWin->SetSelection( false, fpMinefield );
//   tactmap_refresh_non_terrain (gpTactmap);
   fpMinefield = NULL;

   gpGuiWin->ShowEditor(cGuiWindow::eEditMfStatus);
}

void cFieldControlWindow::CbFire (Widget w, XtPointer  pObj,
					XmPushButtonCallbackStruct * /*cbs*/)
{
   XtVaSetValues( w, XmNsensitive, FALSE, 
                  XmNbackground, ((cFieldControlWindow*)pObj)->fForeColor,
                  NULL);
   
   new cFireMunitionMsg(fpMinefield->GetMsgDelay(), fpMinefield);
}

void cFieldControlWindow::CbModeCheck (Widget /*w*/, XtPointer  mode,
					XmPushButtonCallbackStruct */*cbs*/)
{
   fpMinefield->SetCmdMode( (EMfCmdMode)((int)mode));
}

Widget cFieldControlWindow::CreateActionArea (Widget wParent)
   // The action area has controls along the top, with the Done button below
{
   Widget     wActionArea, wBox;

   wActionArea = 
     XtVaCreateWidget ("action_area", 
		       xmFormWidgetClass, wParent,
		       XmNfractionBase, 100,
		       XmNtopAttachment, XmATTACH_WIDGET,
		       XmNtopWidget, fwMinefieldName,
		       XmNbottomAttachment, XmATTACH_FORM,
		       XmNleftAttachment, XmATTACH_FORM,
		       XmNrightAttachment, XmATTACH_FORM,
		       NULL);
             
// the mode radio buttons on the left
   wBox = 
     XtVaCreateManagedWidget ("modes", xmRowColumnWidgetClass, wActionArea,
                              XmNpacking, XmPACK_COLUMN,
                              XmNnumColumns, 1,
                              XmNradioBehavior, TRUE,
			                     XmNtopAttachment, XmATTACH_FORM,
			                     XmNleftAttachment, XmATTACH_FORM,
                              NULL);

   fwFireOnCmd = 
     XtVaCreateManagedWidget ("Fire On Command",
			      xmToggleButtonWidgetClass, wBox,
			      XmNset, True,
			      NULL);
   XtAddCallback (fwFireOnCmd, XmNvalueChangedCallback, 
		  (XtCallbackProc)CbModeCheck, (XtPointer)eCmdModeFire);

   fwFireWhenTripped = 
     XtVaCreateManagedWidget ("Fire When Tripped",
			      xmToggleButtonWidgetClass, wBox,
			      NULL);
   XtAddCallback (fwFireWhenTripped, XmNvalueChangedCallback, 
		  (XtCallbackProc)CbModeCheck, (XtPointer)eCmdModeTripped);

   fwAutoFire = 
     XtVaCreateManagedWidget ("Autofire After First Fire",
			      xmToggleButtonWidgetClass, wBox,
			      XmNsensitive, FALSE,
			      NULL);
   XtAddCallback (fwAutoFire, XmNvalueChangedCallback, 
		  (XtCallbackProc)CbModeCheck, (XtPointer)eCmdModeAutoFire);

   fwCmdFireAll = 
     XtVaCreateManagedWidget ("All Shots Fire on Command",
			      xmToggleButtonWidgetClass, wBox,
			      XmNsensitive, FALSE,
			      NULL);
   XtAddCallback (fwCmdFireAll, XmNvalueChangedCallback, 
		  (XtCallbackProc)CbModeCheck, (XtPointer)eCmdModeCmdFireAll);

   fwTripFireAll = 
     XtVaCreateManagedWidget ("All Shots Fire When Tripped",
			      xmToggleButtonWidgetClass, wBox,
			      XmNsensitive, FALSE,
			      NULL);
   XtAddCallback (fwTripFireAll, XmNvalueChangedCallback, 
		  (XtCallbackProc)CbModeCheck, (XtPointer)eCmdModeTripFireAll);


// the fire button in the middle
   fwFireButton = 
     XtVaCreateManagedWidget ("Fire", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, wBox,
			      XmNtopOffset, 20,
			      XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNbottomWidget, wBox,
			      XmNbottomOffset, 20,
//			      XmNheight, 25,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, wBox,
			      XmNleftOffset, 20,
//			      XmNrightAttachment, XmATTACH_NONE,
			      XmNwidth, 100,
			      XmNsensitive, FALSE,
			      NULL);
   XtAddCallback (fwFireButton, XmNactivateCallback, 
		  (XtCallbackProc)CbFire, (XtPointer) this);
   
      // start with button dulled, in forecolor
   XtVaGetValues( fwFireButton,
                  XmNbackground, &fBackColor,
                  XmNforeground, &fForeColor, 
                  NULL);
   XtVaSetValues( fwFireButton,
                  XmNbackground, fForeColor,
                  NULL);


// the mine list on the right

// the Done button on the bottom
   fwDoneButton = 
     XtVaCreateManagedWidget ("Done", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_NONE,
			      XmNbottomAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 25,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 40,
			      NULL);
   XtAddCallback (fwDoneButton, XmNactivateCallback, 
		  (XtCallbackProc)CbDone, this);

   XtManageChild( wActionArea );
   return wActionArea;
}

void cFieldControlWindow::ShowEdWindow ()
{
   fpMinefield = gpGuiWin->GetCurMinefield();

   XtVaSetValues (fwMinefieldName,
		  XmNvalue, fpMinefield->GetMfName(),
		  NULL);
        
   if ( fpMinefield->IsMultishotCmdDet() ) {
      XtVaSetValues( fwAutoFire, XmNsensitive, TRUE, NULL);
      XtVaSetValues( fwCmdFireAll, XmNsensitive, TRUE, NULL);
      XtVaSetValues( fwTripFireAll, XmNsensitive, TRUE, NULL);
   }
   else {
      XtVaSetValues( fwAutoFire, XmNsensitive, FALSE, NULL);
      XtVaSetValues( fwCmdFireAll, XmNsensitive, FALSE, NULL);
      XtVaSetValues( fwTripFireAll, XmNsensitive, FALSE, NULL);
   }

//*** set the mode radio buttons
   switch (fpMinefield->GetCmdMode()) {
      case eCmdModeFire:
            XtVaSetValues( fwFireOnCmd, XmNset, TRUE, NULL );
         break;
   
      case eCmdModeTripped:
            XtVaSetValues( fwFireWhenTripped, XmNset, TRUE, NULL );
         break;
   
      case eCmdModeAutoFire:
            XtVaSetValues( fwAutoFire, XmNset, TRUE, NULL );
         break;
   
      case eCmdModeCmdFireAll:
            XtVaSetValues( fwCmdFireAll, XmNset, TRUE, NULL );
         break;
   
      case eCmdModeTripFireAll:
            XtVaSetValues( fwTripFireAll, XmNset, TRUE, NULL );
         break;
   }

      // it's possible that the minefield is in a firing mode
   if ( fpMinefield->GetFireButtonState() == true )
      XtVaSetValues( fwFireButton, XmNsensitive, TRUE, 
                     XmNbackground, fBackColor,
                     NULL);
   else
      XtVaSetValues( fwFireButton, XmNsensitive, FALSE, 
                     XmNbackground, fForeColor,
                     NULL);
         
/*
   fpMineCoord->DeleteEntries();
   
   nMines = pMineComp->GetNumMines();
   if (nMines > 0) {
     tccPts = new TLocation [ nMines ];
     i = 0;
     for (cMine* pMine = pMineComp->GetFirstMine();
	  pMine != NULL; pMine = pMine->GetNextMine() )
       tccPts[i++] = pMine->GetTccLoc();

     fpMineCoord->AppendCoordEntries (tccPts, nMines);
     delete[] tccPts;
   }
   fpMineCoord->UpdateNumEntries();
*/

   XtManageChild (fwMainWin);
}

void   cFieldControlWindow::FireButtonMsg(cMinefield* pMinefield, EBoolean on)
   // called by GUI window, from the minefield, when the button should be
   // turned on or off (sensitive to operator or not) as a result of a target
   // detection or a timeout
{
   if ( pMinefield == fpMinefield) {
      int   sens;
      XtVaGetValues( fwFireButton, XmNsensitive, &sens, NULL);
      
      cout << " Fire button " << on << endl;
      if ( sens && (on == false) ) {
         XtVaSetValues( fwFireButton, XmNsensitive, FALSE, 
                        XmNbackground, fForeColor,
                        NULL);
      }
      else if ( !sens && (on == true) ) {
         XtVaSetValues( fwFireButton, XmNsensitive, TRUE, 
                        XmNbackground, fBackColor,
                        NULL);
      }
   }
}

cFieldControlWindow::cFieldControlWindow (Widget wParent)
   // the constructor creates the window and its widgets,
   // initializing with input values
   // We use the general XmForm widget, with the field name on the top,
   // the controls (fire button, mode settings) below, and the Done on the bottom
   // The APLA munition list is to the right
{
   Widget     wLabel;

// The name of the field
   fwMainWin = 
     XtVaCreateWidget ("comp_form", 
		       xmFormWidgetClass, wParent, 
		       XmNfractionBase, 100,
		       NULL);

   wLabel = 
     XtVaCreateManagedWidget ("Minefield Name", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 22,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
   fwMinefieldName = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, fwMainWin, 
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, wLabel,
			      XmNleftOffset, 5,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      XmNcolumns, 20,
			      XmNeditable, FALSE,
			      NULL);

// The action area is all the rest
   fwActionArea = CreateActionArea (fwMainWin);
   
}

cFieldControlWindow::~cFieldControlWindow ()
{
}
