/*
$RCSfile: MapCB.cc,v $
$Revision: 1.4 $
$Author: ksnively $
$Date: 1999/05/18 19:33:20 $
$Log: MapCB.cc,v $
Revision 1.4  1999/05/18 19:33:20  ksnively
Ported CMS to compile under IRIX 6.5
Altered make files to allow cms to be compiled under various OS simulataneously
Removed the Startup GUI

Revision 1.3  1999/04/29 19:34:13  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.2  1999/04/22 22:20:01  ksnively
Sreamlined radio communication routines.  Allowed to send to unlimited number of receivers.

Revision 1.1.1.1  1999/02/16 14:54:36  jchappel
Comprehensive Mine Simulator

// Revision 2.2  1998/11/25  18:37:51  ksnively
// Update for APLA Exercise
//
// Revision 2.1  1998/06/03  17:18:04  jchappel
// ..
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.19  1996/08/20  13:44:41  mkappel
// Add Detonation Display.
//
// Revision 1.18  1996/08/08  13:19:27  mkappel
// Update minefield name label when zooming.
//
// Revision 1.17  1996/07/13  19:04:20  rcarp
// version 1.5b13, fixes bug dragging with two buttons and displays damamged vehicles
//
// Revision 1.16  1996/05/15  00:11:09  rcarp
// mostly wam and offroute fixes, also deleting selection
//
// Revision 1.15  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.11  1996/04/04  16:25:34  mkappel
// Improve performance with large mine sets.
//
// Revision 1.10  1996/04/02  19:58:14  mkappel
// Add sensitivity at end of drag.
//
// Revision 1.9  1996/03/30  19:35:34  mkappel
// Add Distance tool.
//
// Revision 1.8  1996/03/23  14:24:14  rcarp
// improvements to test mode and selection
//
// Revision 1.7  1996/03/16  18:36:36  mkappel
// Interface for Test Mode files.
//
// Revision 1.6  1996/03/15  16:03:02  mkappel
// Allow repositioning of minefield when loading (standard minefields).
//
// Revision 1.5  1996/03/07  14:22:46  mkappel
// Add Output Statistics to main menu.
//
// Revision 1.4  1996/02/23  21:16:33  mkappel
// Use cShellWindow::PopupMessage.
//
// Revision 1.3  1996/02/23  18:36:44  mkappel
// Add Site/Application to Start Up Screen.
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: MapCB.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   MapCB.cc
   
      The display interface implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "CmsControl.h"
#include "Distance.h"
#include "MapControls.h"
#include "Minefield.h"
#include "TestMode.h"
#include "Communication.h"

#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>

#include <stdext.h>


/* Indicates what to draw vis-a-vis trees (CANOPIES, LINES, INDIVIDUAL) */
#define TREE_MASK TACTMAP_TREE_CANOPIES | TACTMAP_TREE_LINES ;


void     cWindow::InitMotif( int* pArgc, char* argv[] )
{
   //Set the xenvironment variable to point to the cms resource file
   //File will then be loaded by XtAppInitialize
   putenv("XENVIRONMENT=cms.xrdb");

   fwTopLevelShell = XtAppInitialize( &fApp, "CMSGUI", NULL, 0, pArgc, argv, NULL, NULL, 0);
}

void    cMapDisp::print_gesture(SNSTVE_WINDOW_PTR /*sensitive*/, 
				SNSTVE_ACTION action,
                                SNSTVE_CALL_DATA *call_data, char *name)
{
    if (action == SNSTVE_PRESS)
      cout<<"Press on "<<name<<" "<<
         (!strcmp(name, "vehicle") ?
        TACTMAP_INSTANCE_TO_USER_INSTANCE(call_data->instance) :
        TACTMAP_INSTANCE_TO_INDEX(call_data->instance)) << endl;
}

cMapDisp::cMapDisp (TACTMAP_OBJECT_CLASS tmClass)
{
    switch (tmClass) {
       case TACTMAP_HYPSO:
          fDfn.hypso.kind = TACTMAP_COLOR_HYPSO;
	  break;

       case TACTMAP_CONTOURS:
          fDfn.contours.minor_spacing = 10.0;
          fDfn.contours.major_spacing = 100.0;
	  break;

       case TACTMAP_GRIDS:
	  fDfn.grids.interval = 1000;
	  fDfn.grids.label_mask = COORD_LABEL_ALL;
	  break;
    
       default:
             SNSTVE_INIT_CLASS(fSensClass);
             fDfn.sensitive.sensClass = &fSensClass;
             
             switch (tmClass) {
                case TACTMAP_WATER:
                      callback_register_handler(fSensClass.gesture, 
                               (CALLBACK_HANDLER)print_gesture,
                              "water");
                   break;
             
                case TACTMAP_ROADS:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "road");
                   break;
             
                case TACTMAP_BUILDINGS:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "building");
                   break;
             
                case TACTMAP_PIPELINES:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "pipeline");
                   break;
             
                case TACTMAP_POLITICAL:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "political");
                   break;
             
                case TACTMAP_RAILROADS:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "railroad");
                   break;
             
                case TACTMAP_POWERLINES:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "powerline");
                   break;
             
                case TACTMAP_TOWNS:
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "town");
                   break;
             
                case TACTMAP_TREES:
		      fDfn.trees.mask = TREE_MASK;
                      callback_register_handler(fSensClass.gesture,  
                               (CALLBACK_HANDLER)print_gesture,
                              "tree");
                   break;
                   
                default:
                      return;
             }
    }
    fpObj = tactmap_create_object(gpTactmap, tmClass, &fDfn);
    tactmap_add_object(gpTactmap, fpObj);

}

cMapDisp::~cMapDisp()
{
}

void   cEditorWindow::DeleteObjEntry(cMfComponent* /*pComp*/, int /*index*/)
{
}

void   cEditorWindow::ShowEdWindow()
{
   SetButtonState ();
   XtManageChild(fwMainWin);
}

void   cEditorWindow::ClearEdWindow()
{
   XtUnmanageChild(fwMainWin);
}

void   cEditorWindow::MapClick (TLocation& /*loc*/)
{
}

void   cEditorWindow::NewMine (TLocation& /*loc*/)
{
}

void   cEditorWindow::NewMines (TLocation* /*tccPts*/, int /*nMines*/)
{
}

void cEditorWindow::SetButtonState ()
{
}

void cEditorWindow::SetWidgetState ()
{
}

cCoordList*  cEditorWindow::GetMineCoordWidget ()
{ return NULL; }

cCoordList*  cEditorWindow::GetPerimVertexCoordWidget ()
{ return NULL; }

cCoordList*  cEditorWindow::GetRowVertexCoordWidget ()
{ return NULL; }

cEntityIdTextEditor*  cEditorWindow::GetControllerIdWidget ()
{ return NULL; }

cEditorWindow::cEditorWindow()
{
}

cEditorWindow::~cEditorWindow()
{
}

void    cGuiWindow::CbOutput (Widget /*w*/, XtPointer  /*pObj*/,
			      XmPushButtonCallbackStruct */*cbs*/)
{
  XmString  xmstr;
  char     *str;

  XtVaGetValues (fwOutputButton, XmNlabelString, &xmstr, NULL);
  XmStringGetLtoR (xmstr, XmSTRING_DEFAULT_CHARSET, &str);

  if ( ! gfInteract ) {
    /* start saving */
    xmstr = XmStringCreateSimple ("Stop Saving Interactions");
    gCmsMgr.SetInteractionOutput(true);
  }
  else {
    /* stop saving */
    xmstr = XmStringCreateSimple ("Start Saving Interactions");
    gCmsMgr.SetInteractionOutput(false);
  }

  XtVaSetValues (fwOutputButton, XmNlabelString, xmstr, NULL);
  XmStringFree (xmstr);
}

void    cGuiWindow::CbMineStateDisp (Widget /*w*/, XtPointer  /*pObj*/,
			      XmPushButtonCallbackStruct */*cbs*/)
{
  XmString  xmstr;
  char     *str;

  XtVaGetValues (fwMineDispButton, XmNlabelString, &xmstr, NULL);
  XmStringGetLtoR (xmstr, XmSTRING_DEFAULT_CHARSET, &str);

  if ( !gfMineStateDisp ) {
    /* show mine state */
    xmstr = XmStringCreateSimple ("Hide Mine State Display");
    gfMineStateDisp = 1;
  }
  else {
    /* hide mine state */
    xmstr = XmStringCreateSimple ("Show Mine State Display");
    gfMineStateDisp = 0;
  }

  XtVaSetValues (fwMineDispButton, XmNlabelString, xmstr, NULL);
  XmStringFree (xmstr);
}

void    cGuiWindow::CbCommEffects (Widget /*w*/, XtPointer  /*pObj*/,
			      XmPushButtonCallbackStruct */*cbs*/)
{
  XmString  xmstr;
  char     *str;

  XtVaGetValues (fwCommEffectsButton, XmNlabelString, &xmstr, NULL);
  XmStringGetLtoR (xmstr, XmSTRING_DEFAULT_CHARSET, &str);

  cout << "cGuiWindow::CbCommEffects" << endl;
  if ( gpRadioNetMgr ) {
     cout << "Found Radio Net Manager" << endl;
     if (!(gpRadioNetMgr->GetUseCEM())) {

	cout << "\t\tSetting UseCEM to TRUE" << endl;
	xmstr = XmStringCreateSimple ("Ignore Communication Effects");
	gpRadioNetMgr->SetUseCEM(true);
     }
     else {
	
	cout << "\t\tSetting UseCEM to FALSE" << endl;
	xmstr = XmStringCreateSimple ("Use Communication Effects");
	gpRadioNetMgr->SetUseCEM(false);
     }
  }

  XtVaSetValues (fwCommEffectsButton, XmNlabelString, xmstr, NULL);
  XmStringFree (xmstr);
}

void    cGuiWindow::CbQuit (Widget /*w*/, XtPointer  /*pObj*/,
			    XmPushButtonCallbackStruct */*cbs*/)
{
  gpGuiWin->DialogMessage (cGuiWindow::eDialogWarning, 
			   (XtCallbackProc)CbQuitOK,
			   "Are you sure that you want to quit?");
}

void    cGuiWindow::CbQuitOK (Widget w, XtPointer  /*pObj*/,
			      XmPushButtonCallbackStruct */*cbs*/)
{
   gCmsMgr.SetState( estQuit );   
   XtDestroyWidget (w);
}

void cGuiWindow::CbMap (Widget /*w*/, XtPointer /*pObj*/,
			XmPushButtonCallbackStruct */*cbs*/)
{
  XtPopup (cGuiWindow::fpMapControlsWin->GetMainWin(), XtGrabNone);
  cGuiWindow::fpMapControlsWin->InitControls();
}

void cGuiWindow::CbTestMode (Widget /*w*/, XtPointer /*pObj*/,
			     XmPushButtonCallbackStruct */*cbs*/)
{
  XtPopup (cGuiWindow::fpTestModeWin->GetMainWin(), XtGrabNone);
}

void cGuiWindow::CbDisableUser (Widget, XtPointer,
         XmPushButtonCallbackStruct*)
{
   gpGuiWin->DisableUser();
}

void cGuiWindow::RedrawAll ()
{
  for (cMinefield* pMf = gMfieldMgr.GetField(1); pMf; 
       pMf = pMf->GetNextMf()) {
    pMf->GetPerimeter()->GetPerimDisp()->UpdateLabelPos();

    for (cMfComponent* pC = pMf->GetFirstComp(); pC; 
	 pC = pC->GetNext())
      if (pC->IsMineComp())
	for (cMine* pMine = ((cMineComp*)pC)->GetFirstMine();
	     pMine; pMine = pMine->GetNextMine())
	  pMine->CompleteVisuals();
  }
}

void cGuiWindow::SetMapScaleIndex (int Scale)
{
  if (Scale < 0) Scale = 0;
  else if (Scale >= NUM_SCALES) Scale = NUM_SCALES - 1;
  
  if ( fScaleIndex != Scale ) {
     fScaleIndex = Scale;
     cMapControlsWindow::CheckFeatures(fScaleIndex, cMapControlsWindow::fScale [fScaleIndex].value);
     tactmap_set_scale (gpTactmap, cMapControlsWindow::fScale [fScaleIndex].value);
     RedrawAll ();
     fpMapControlsWin->InitControls();
  }
}

void cGuiWindow::SetMapScale (float64 Scale)
{
  int      i;

  for (i = 0; i < NUM_SCALES; i++) {
    if (Scale == cMapControlsWindow::fScale[i].value) {
      SetMapScaleIndex (i);
      break;
    }
    else if (Scale > cMapControlsWindow::fScale[i].value && i > 0) {
      SetMapScaleIndex (i - 1);
      break;
    }
  }
  if (i == NUM_SCALES) 
    SetMapScaleIndex (NUM_SCALES - 1);
}

float64 cGuiWindow::GetMapScale ()
{
  return cMapControlsWindow::fScale [fScaleIndex].value;
}

void cGuiWindow::SetMapCenter (int32 x, int32 y, int32 /*c*/)
{
float64 xf,yf;
int32 ci=0;

/*added parameter to upgrade to modsaf3.0 libtactmap AUSTIN 4 Aug 97*/
//printf("SetMapCenter1 will set to x: %d y: %d: %d\n",x,y);

  tactmap_get_center(gpTactmap,&xf,&yf,&ci);
  //printf("SetMapCenter2 current was to x: %f y: %f\n",xf,yf);

  tactmap_set_center (gpTactmap, x, y, ci);
  //printf("done with SetMapCenter2\n");

  fpMapControlsWin->InitControls();
  //printf("done with InitControls\n");

  UpdateScrollbars ();
  //printf("done with UpdateScrollbars\n");

}

static void draw_signed_rect (Display *display, Drawable d, 
			      GC gc, XRectangle *rect)
{
    register int16 x, y, width, height;

    x = rect->x;
    y = rect->y;
    width = rect->width;
    height = rect->height;

    if (width < 0)
      x += width, width = -width;

    if (height < 0)
      y += height, height = -height;

    XDrawRectangle(display, d, gc, x, y, width, height);
}

#define DRAG_THRESH 20

void    cGuiWindow::CbClick(SNSTVE_WINDOW_PTR /*window*/, 
                            SNSTVE_ACTION action, 
                            SNSTVE_CALL_DATA *call_data)
{
   TLocation          tccPt;
   float64            x, y, scale;
   static EBoolean        dragging = false;
   static XRectangle      last_rect;
   static cDistanceDisp*  pDistDisp = NULL;
	int32		c;

   if (call_data->button == 1) {
      if (action == SNSTVE_RELEASE) {
/*added parameter to upgrade to modsaf3.0 libtactmap AUSTIN 4 Aug 97*/
         tactmap_screen_to_map (gpTactmap,
            (int32)call_data->screen_x, 
            (int32)call_data->screen_y,
            &tccPt.x, &tccPt.y, &c);
         
         fpCurrentEditor->MapClick(tccPt);
       }
    }

    else if (call_data->button == 2) {

      switch(action) {

      case SNSTVE_PRESS:
         dragging = false;
         break;

      case SNSTVE_MOTION:

         if (dragging)
           /* Erase the last box */
           draw_signed_rect (gpDisplay, gMapWindow,
                   gpGuiWin->fgcSensErase,
                   &last_rect);
         else {
           /* See if we have moved far enough to consider this a drag */
           if ((call_data->xoff > DRAG_THRESH) ||
               (call_data->xoff < -DRAG_THRESH) ||
               (call_data->yoff > DRAG_THRESH) ||
               (call_data->yoff < -DRAG_THRESH)) {
      
              dragging = true;

             /* Infer the click location */
             last_rect.x = call_data->screen_x - call_data->xoff;
             last_rect.y = call_data->screen_y - call_data->yoff;
           }
         }

         if (dragging) {
           last_rect.width = call_data->xoff;
           last_rect.height = call_data->yoff;
      
           /* Draw the new box */
           draw_signed_rect (gpDisplay, gMapWindow,
                   gpGuiWin->fgcSensDraw,
                   &last_rect);
         }
         break;

      case SNSTVE_RELEASE:
         if (dragging) {
            dragging = false;
            /* Erase the last box */
            draw_signed_rect (gpDisplay, gMapWindow,
                    gpGuiWin->fgcSensErase,
                    &last_rect);
       
/*added parameter to upgrade to modsaf3.0 libtactmap AUSTIN 4 Aug 97*/
            tactmap_screen_to_map (gpTactmap,
                    (int32)(call_data->screen_x - call_data->xoff/2),
                    (int32)(call_data->screen_y - call_data->yoff/2),
                    &x, &y, &c);
            scale = tactmap_get_scale (gpTactmap,
                        (int32)call_data->xoff,
                        (int32)call_data->yoff);
            gpGuiWin->SetMapScale (scale);
         }
         else {
/*added parameter to upgrade to modsaf3.0 libtactmap AUSTIN 4 Aug 97*/
            tactmap_screen_to_map (gpTactmap,
                  (int32)call_data->screen_x, 
                  (int32)call_data->screen_y,
                  &x, &y, &c);
            gpGuiWin->SetMapScaleIndex (gpGuiWin->fScaleIndex + 1);
         } 
      
         gpGuiWin->SetMapCenter ((int32)x, (int32)y, (int32)c);
         break;
      }
   }

   else if (call_data->button == 3) {

      switch(action) {

      case SNSTVE_PRESS:
         dragging = false;
         break;

      case SNSTVE_MOTION:

         if ( ! dragging) {
           /* See if we have moved far enough to consider this a drag */
           if ((call_data->xoff > DRAG_THRESH) ||
               (call_data->xoff < -DRAG_THRESH) ||
               (call_data->yoff > DRAG_THRESH) ||
               (call_data->yoff < -DRAG_THRESH)) {
      
             dragging = true;
             pDistDisp = new cDistanceDisp (call_data);
           }
         }
         if (dragging)
           pDistDisp->UpdateDist (call_data);
      
         break;

      case SNSTVE_RELEASE:
         if (dragging) {
           pDistDisp->CompleteDist ();
           dragging = false;
         }
         else {
/*added parameter to upgrade to modsaf3.0 libtactmap AUSTIN 4 Aug 97*/
           tactmap_screen_to_map (gpTactmap,
                   (int32)call_data->screen_x, 
                   (int32)call_data->screen_y,
                   &x, &y, &c);
               
           gpGuiWin->SetMapScaleIndex (gpGuiWin->fScaleIndex - 1);
           gpGuiWin->SetMapCenter ((int32)x, (int32)y, (int32)c);
         }
         break;
      }
    }
}

void     cGuiWindow::EditMinefield(cMinefield* pField)
{
   SetSelection( false, pField );
   ShowEditor(eEditMinefield);
}

void     cGuiWindow::ComponentStatus(cMinefield* pField)
{
   SetSelection( false, pField );
   ShowEditor(eEditCompStatus);
}

void     cGuiWindow::FieldControl(cMinefield* pField)
{
   SetSelection( false, pField );
   ShowEditor(eEditMfCommand);
}

void    cGuiWindow::ProcessFireButMsg(cMinefield* pMinefield, EBoolean on)
{
   if ( GetCurEditor() == fpFieldControlWin )
      fpFieldControlWin->FireButtonMsg(pMinefield, on);
}

void     cGuiWindow::EditComponent(cMfComponent* pComp)
{
   SetSelection( false, pComp->GetMinefield(), pComp );
   
   switch (pComp->GetCompType()) {
      case eCompPerim:
            ShowEditor(eEditPerim);
         break;

      case eCompIndiv:
            ShowEditor(eEditIndiv);
         break;

      case eCompGrid:
            ShowEditor(eEditGrid);
         break;

      case eCompRow:
            ShowEditor(eEditRow);
         break;
   }
}

cMinefield*          cGuiWindow::GetCurMinefield()
   // searches selection list for minefields, 
   // returns NULL if multiple minefields are found
{
   cMinefield* pCurMf = NULL;
   
   for ( TSelection* pSel  = (TSelection*) (fSelectionList.GetFirstLink());
                     pSel != (TSelection*) (fSelectionList.GetNULLLink());
                     pSel  = (TSelection*) (fSelectionList.GetNextLink()) ) {
      if ( (pCurMf != pSel->GetMinefield()) && (pSel->GetMinefield() != NULL) ) {
         if (pCurMf != NULL )   // found a second minefield
            return NULL;
         else
            pCurMf = pSel->GetMinefield();
      }
   }
   
   return pCurMf;
}

cMfComponent*        cGuiWindow::GetCurComponent()
   // searches selection list for components, 
   // returns NULL if multiple components are found
{
   cMfComponent* pCurComp = NULL;
   
   for ( TSelection* pSel  = (TSelection*) (fSelectionList.GetFirstLink());
                     pSel != (TSelection*) (fSelectionList.GetNULLLink());
                     pSel  = (TSelection*) (fSelectionList.GetNextLink()) ) {
      if ( (pCurComp != pSel->GetComponent()) && (pSel->GetComponent() != NULL) ) {
         if (pCurComp != NULL )   // found a second component
            return NULL;
         else
            pCurComp = pSel->GetComponent();
      }
   }
   
   return pCurComp;
}

cMfieldObj*          cGuiWindow::GetCurMfObject()
   // searches selection list for objects, 
   // returns NULL if multiple objects are found
{
   cMfieldObj* pCurObj = NULL;
   
   for ( TSelection* pSel  = (TSelection*) (fSelectionList.GetFirstLink());
                     pSel != (TSelection*) (fSelectionList.GetNULLLink());
                     pSel  = (TSelection*) (fSelectionList.GetNextLink()) ) {
      if ( (pCurObj != pSel->GetMfieldObj()) && (pSel->GetMfieldObj() != NULL) ) {
         if (pCurObj != NULL )   // found a second object
            return NULL;
         else
            pCurObj = pSel->GetMfieldObj();
      }
   }
   
   return pCurObj;
}

void     cGuiWindow::SetSelection(EBoolean multiSelect, cMinefield* pMf,
                                  cMfComponent* pComp, cMfieldObj* pObj )
{
   TSelection* pSel;
   TSelection* pNextSel;

   if ( multiSelect ) {
      for ( pSel  = (TSelection*)fSelectionList.GetFirstLink();
            pSel != (TSelection*)fSelectionList.GetNULLLink();
            pSel  = (TSelection*)fSelectionList.GetNextLink() ) {
         if ( pSel->MatchSel(pMf, pComp, pObj) ) {
               // already selected, remove selection
            delete pSel;
               // changing selections, we always need to check button states
            SetEditorButtonState();
            return;
         }
      }
         // not already selected, add to selection list
      if ( pMf)
         new TSelection( (TSelection*)fSelectionList.GetLastLink(),
                      pMf, pComp, pObj);
   }
   
   else {            // single selection, clear old selections first
      pSel  = (TSelection*)fSelectionList.GetFirstLink();
      pNextSel = (TSelection*)fSelectionList.GetNextLink();
      while ( pSel != (TSelection*)fSelectionList.GetNULLLink() ) {
         if ( ! (pSel->IsParentSel(pMf, pComp, pObj)) )
            delete pSel;
         pSel = pNextSel;
         pNextSel = (TSelection*)fSelectionList.GetNextLink();
      }
      
      if ( pMf)
         new TSelection( (TSelection*)fSelectionList.GetLastLink(),
                      pMf, pComp, pObj);
   }

      // changing selections, we always need to check button states
   SetEditorButtonState();
}

void		cGuiWindow::DeleteSelection(cMinefield* pMf,
             cMfComponent* pComp, cMfieldObj* pObj)
{
   TSelection* pSel;
   
   for ( pSel  = (TSelection*)fSelectionList.GetFirstLink();
         pSel != (TSelection*)fSelectionList.GetNULLLink();
         pSel  = (TSelection*)fSelectionList.GetNextLink() ) {
      if ( pSel->MatchSel(pMf, pComp, pObj) ||
         ( pSel->IsChildSel(pMf, pComp, pObj) )   ) {
         delete pSel;
      }
   }
}

void     cGuiWindow::UpdateTactmap()
{
   tactmap_tick (gpTactmap, 67);
}

void     cGuiWindow::NewMine (cMineComp* /*pComp*/, TLocation& tccPt)
{
  fpCurrentEditor->NewMine (tccPt);
  fpCurrentEditor->SetButtonState ();
}

void     cGuiWindow::NewMines (cMineComp* /*pComp*/, TLocation* tccPts, int nMines)
{
  fpCurrentEditor->NewMines (tccPts, nMines);
  fpCurrentEditor->SetButtonState ();
}

void cGuiWindow::DialogMessage (EDialog eDialogType, 
				XtCallbackProc cb_proc,
				char *msg)
{
  if ( ! fsUserEnabled )
     EnableUser();

  PopupMessage (fwMainWin, eDialogType, "", "", cb_proc, NULL, fpCurrentEditor, msg);
}

void cShellWindow::PopupMessage (Widget wParent,
				 EDialog eDialogType,
				 char* ok_str,
				 char* cancel_str,
				 XtCallbackProc ok_cb_proc,
				 XtCallbackProc cancel_cb_proc,
				 XtPointer pObj,
				 char *msg)
{
  Widget          dialog;
  XmString        xmstr;

  switch (eDialogType) {

  case cGuiWindow::eDialogError:
    dialog = 
      XmCreateErrorDialog (wParent, "error", NULL, 0);
    break;

  case cGuiWindow::eDialogInfo:
    dialog = 
      XmCreateInformationDialog (wParent, "info", NULL, 0);
    break;

  case cGuiWindow::eDialogWarning:
    dialog = 
      XmCreateWarningDialog (wParent, "warning", NULL, 0);
    break;

  case cGuiWindow::eDialogQuestion:
    dialog = 
      XmCreateQuestionDialog (wParent, "question", NULL, 0);
    break;
  }

  xmstr = XmStringCreateLtoR (msg, XmSTRING_DEFAULT_CHARSET);
  XtVaSetValues (dialog,
		 XmNmessageString, xmstr,
		 XmNdefaultButtonType, XmDIALOG_OK_BUTTON,
		 XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
		 NULL);
  XmStringFree (xmstr);

  if (strlen (ok_str) > 0) {
    xmstr = XmStringCreateSimple (ok_str);
    XtVaSetValues (dialog, XmNokLabelString, xmstr, NULL);
    XmStringFree (xmstr);
  }
  if (strlen (cancel_str) > 0) {
    xmstr = XmStringCreateSimple (cancel_str);
    XtVaSetValues (dialog, XmNcancelLabelString, xmstr, NULL);
    XmStringFree (xmstr);
  }

  if (ok_cb_proc) {
    XtAddCallback (dialog, XmNokCallback, ok_cb_proc, pObj);
    if (cancel_cb_proc)
      XtAddCallback (dialog, XmNcancelCallback, cancel_cb_proc, pObj);
  }
  else
    XtUnmanageChild (XmMessageBoxGetChild (dialog, 
					   XmDIALOG_CANCEL_BUTTON));
  XtUnmanageChild (XmMessageBoxGetChild (dialog, 
					 XmDIALOG_HELP_BUTTON));

  XtManageChild (dialog);
  XtPopup (XtParent(dialog), XtGrabNone);
  XtSetSensitive (dialog, TRUE);
}
