/**********************************************************************
 * File: MfCamera.h
 *
 * Types defined: 
 *
 * Functions defined: 
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1998
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */

#include "MfCamera.h"
#include "AnalData.h"
#include "CmsControl.h"
#include "MapGeometry.h"
#include <math.h>

const TKeywordEntry cDbMine_cMfCamera::fKeyCameraTable[ecameraSizeTable] = {
   "height", ekeyFloat
};

const TKeywordEntry TSensParamCam::fKeyCamSensTable[ecamSizeTable] = {
   "sensType",        ekeyString,
   "feedDelay", ekeyFloat
};

// ************************************
//   stealth view types for camera view
// ************************************

void cStealthEntity::SetControl(cMfCamera* pNewControlCam, TLocation* pLoc, TOrient* pOrient)
{
   //Cuts off vehicle updates
   if (fpUpdateVehMsg){
      (fpUpdateVehMsg->GetMsgList())->DeleteMsg(fpUpdateVehMsg);
      fpUpdateVehMsg = NULL;
   }
   
   /*
   if (fpSendViewControlMsg){
      (fpSendViewControlMsg->GetMsgList())->DeleteMsg(fpSendViewControlMsg);
      fpSendViewControlMsg = NULL;
   }
   */

   
   //Set pointer to entity controllong the view
   fpControlCam = pNewControlCam;

   if (fpControlCam) {
      //The entity exists.  Reset our coordinates and update.
      SetCoordTccRef (*pLoc, pOrient->psi, pOrient->theta, pOrient->phi);
      //Restarts vehicle updates
      feInternalState = estisActive;
      fpSendViewControlMsg = new cSendViewControlMsg((cVehicle*) this, toStealthId);
      UpdateVehicle();
   }
   else {
      //The entity does not exist.  Set view pointing to nothing in the sky.
      //Mimic an "off" state
      fTccLocation.z += 1000;
      SetCoordTccRef (fTccLocation, fTccOrientation.psi, 0, 0);
      fpSendVehStateMsg = new cSendVehStateMsg ((cVehicle*) this);
      fpSendViewControlMsg = new cSendViewControlMsg((cVehicle*) this, toStealthId);
      feInternalState = estisNone;
   }
}

void   cStealthEntity::UpdateVehicle()
{
   fpUpdateVehMsg = NULL;
   TTimeSim   now = GetCurSimTime();


   if (feInternalState >= estisActive ) {
      if (now >= fNextVehStateTime) {
	 //update the network of our position
	 fpSendVehStateMsg = new cSendVehStateMsg ((cVehicle*) this);

	 //commented this out to allow people to pan and tilt from Mak Stealth
	 //fpSendViewControlMsg = new cSendViewControlMsg((cVehicle*) this, toStealthId);
	 fNextVehStateTime = now + STEALTHPDURATE;
      }
      
      fpUpdateVehMsg = new cUpdateVehicleMsg(this);
   }
   if ( fpDisplay)
         fpDisplay->UpdateVehicleDisp();

}

void cStealthEntity::InitStealth()
{
   feInternalState = estisActive;
   //fpSendVehStateMsg = new cSendVehStateMsg ((cVehicle*) this);
   fNextVehStateTime = GetCurSimTime() + STEALTHPDURATE;

   fpSendViewControlMsg = new cSendViewControlMsg((cVehicle*) this, toStealthId);
   
   SimulateVehicle();
}

cStealthEntity::cStealthEntity(cMfCamera* pControlCam, TLocation* pLoc, TOrient* pOrient) :
   cSimVehicle()
{
   TEntityType stealthType(0, 0, 0, 0, 0, 0, 0);
   fCmsType = fAltType = stealthType;

   // Send to the stealth ID that was specified on the command line.
   if (gCmsMgr.fStartupParam.fstealthID.site != (unsigned short) -1) {
      toStealthId = gCmsMgr.fStartupParam.fstealthID;
   }
   else
      toStealthId = fId;

   fpUpdateVehMsg = NULL;
   fpSendViewControlMsg = NULL;
   fpSendVehStateMsg = NULL;

   SetControl(pControlCam, pLoc, pOrient);
}

cStealthEntity::~cStealthEntity()
{
   if (fpUpdateVehMsg){
      (fpUpdateVehMsg->GetMsgList())->DeleteMsg(fpUpdateVehMsg);
      fpUpdateVehMsg = NULL;
   }

   /*
   if (fpSendViewControlMsg){
      (fpSendViewControlMsg->GetMsgList())->DeleteMsg(fpSendViewControlMsg);
      fpSendViewControlMsg = NULL;
   }
*/
}

// ************************************
//   component classes
// ************************************


cStealthEntity* cMfCamera::fpViewEntity = NULL;
TEntityType   cMfCamera::fsCameraEntityType(ekindMunition,
                                             efunnydomainMfCamera,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cMfCamera::fsCameraName = "Camera";
char*            cMfCamera::fsCameraAltName = "Cam";
int     cMfCamera::fsDefaultCameraType =  TFuzeParam::esenstypeCommandCamDir;
cDbMine_cMfCamera      cMfCamera::fsDbCamera( cMfCamera::fsDefaultCameraType,
                                                cMfCamera::fsCameraEntityType,
                                                cMineDBMgr::emdblevDomain, 
                                                cMfCamera::fsCameraName,
                                                cMfCamera::fsCameraAltName); 



// ************************************
//   components
// ************************************

EBoolean     TSensParamCam::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   
   returnCode = infile.ParseItem(fKeyCamSensTable, ecamSizeTable,
		&returnIndex, &returnVal);


   //fprintf (stdout, "\n\n\nEntering ParseDbStructItem \n\n\n");
   switch( returnCode ) {

      case eFileNoErr:
	 switch( returnIndex )
	    {
	       case ecamkwFuzeTypeName: {
		  // read name, look up index
		  fTypeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
						   (char *)returnVal );
		  
		  // if we get a type we cannot use, mark us as not valid
		  if( (fTypeNameIndex != esenstypeCommandCamDir) &&
		      (fTypeNameIndex != esenstypeAutoCmdCamDir)) 
		     {
			fTypeNameIndex = efuzetypeOther;
			cout<< "ignoring unuseable sens type for standalone sensor"<<endl;
		     }
                  
                  delete[] returnVal;
                  returnVal = NULL;
		  break;
               }
	       case ecamkwFeedDelay: {
		  fFeedDelay = (TTimeSim)(1000. * (*((float*)returnVal))); // Convert seconds to milliseconds
		  //fprintf (stdout, "Setting Delay = %u\n", fFeedDelay);

		  delete returnVal;
		  returnVal = NULL;
		  break;
	       }
	       
	       
	       default: // what?
		  break;
	    }
	 break;
	 
      case eBadKeyword:
	 cout<< "passing keyword up to parent fuze class"<<endl;
         return TSensParamDumb::ParseDbStructItem(infile);
	 
      default:
	 Assert(false, kerrReadParamfile, "TSensParamCam::ParseDbStructItem");
	 return false;
	 // error, error...
   }
   return true;
}

cControl*     TSensParamCam::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
      case esenstypeCommandCamDir:
      case esenstypeAutoCmdCamDir:
            pNewControl = new cMfCameraControl(pMine, fFireOffset, fuzeOpt, this);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TFuzeParamApOffRoute::CreateControl");
   }
   
   return pNewControl;
}

EBoolean   TSensParamCam::IsDirectional(TFlatEntCoord* /*pMinMax*/)
{
   if ( fTypeNameIndex == esenstypeCommandCamDir || fTypeNameIndex == esenstypeAutoCmdCamDir  ) {
     /*
      if ( pMinMax ) {
         pMinMax->x = 0.0;
         pMinMax->y = fMaxRange;
      }
      */      
      return true;
   }
   
   return false;
}

TFuzeParam* TSensParamCam::CloneFuzeParam()
{
   TFuzeParam* pClone = new TSensParamCam(fTypeNameIndex);
   pClone->fpOptSelector = fpOptSelector->CloneFuzeOptionSelector(pClone);
   
   if ( fpRange ) {
      fpRange = new float[fNumBands];
      fpDetectProb = new float[fNumBands];
      for ( int i = 0; i < fNumBands; i++ ) {
         fpRange[i] = ((TSensParamCam*)pClone)->fpRange[i];
         fpDetectProb[i] = ((TSensParamCam*)pClone)->fpDetectProb[i];
      }
   }
   ((TSensParamCam*)pClone)->fFeedDelay = fFeedDelay;
   
   return pClone;
}

/*
EBoolean TSensParamCam::ParseDbStructItem(cInFileObj& infile) 
{
}
*/


TSensParamCam::TSensParamCam(int fuze):
   TSensParamDumb(fuze)
{
   fMaxRange = 0.0;
   fFeedDelay = 0.0;
}

TSensParamCam::~TSensParamCam()
{
}

EBoolean cDbMine_cMfCamera::ParseDbEntry(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   EBoolean success = true;

   returnCode = infile.ParseItem(fKeyCameraTable, ecameraSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

               case eFileNoErr:
			switch( returnIndex )
			{
			   case ecamerakwHeight: {
			      fCamHeight = *((float*)returnVal);
			      //fprintf(stdout, "Setting Cam height = %f\n", fCamHeight);
			   }
			   break;

			   default:
			      success =  false;
			}
			break;

              case eBadKeyword:   // parent might recognize it
		 success =  cDbMine::ParseDbEntry(infile);
		 break;

              default:
		 success = false;
   }
   delete returnVal;

   return success;
}


cDbMine* cDbMine_cMfCamera::CreateDbMine(EBoolean bWithParams)
{
   cDbMine_cMfCamera* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cMfCamera( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cMfCamera();
   return pNewDbMine;
}

cSimMine* cDbMine_cMfCamera::CreateMine(cPlanMine* pPlan)
{
   //Assert(false, kerrValidityCheck, "cDbMine_cMfCamera::CreateMine");
   return (cSimMine*) new cMfCamera(pPlan);
}

cDbMine_cMfCamera::cDbMine_cMfCamera() :
   cDbMine()
{
   fCamHeight = 1.0;
   //fprintf(stdout, "Setting Cam height = %f\n", fCamHeight);
}

cDbMine_cMfCamera::cDbMine_cMfCamera(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
   fCamHeight = 1.0;
   //fprintf(stdout, "Setting Cam height = %f\n", fCamHeight);
}

cDbMine_cMfCamera::cDbMine_cMfCamera(int fuzeType, TEntityType& type, 
				     int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
   fCamHeight = 1.0;
   //fprintf(stdout, "Setting Cam height = %f\n", fCamHeight);
}

cDbMine_cMfCamera::~cDbMine_cMfCamera()
{
   fCamHeight = 1.0;
   //fprintf(stdout, "Setting Cam height = %f\n", fCamHeight);
}

void cMfCameraControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* /*pEmissList*/, TTimeSim* /*pWaitTime*/)
{
   TAttackGeom   attGeom;
   ECamInternalState state = (ECamInternalState)(fpMine->GetInternalState());
   float range = FLT_MAX;
   cLinkVehGeom* pLink = NULL;
   EBoolean detect = false;
   TTimeSim curTime = GetCurSimTime();

   static TTimeSim timeOff = 0;
   


   if ( state >= ecisOn ) {
      ECamInternalState newstate = state;

      // Check if we were viewing but the view was stolen.
      if (state > ecisOn && !(fpCamera->IsViewing()) ) {
	 newstate = ecisOn;
	 fbDelaying = false;
      }

      // if the Camera is doing a delay feed and is still the viewing camera,
      // check if it is feed time.  If so, activate view.
      if (fbDelaying && newstate > ecisOn && fFeedTime <= curTime) {
	 fpCamera->CamOn(false); 
	 fbDelaying = false;
      }

         // find closest vehicle
      for (T2link* pLn  = pVehList->GetFirstLink();
                   pLn != pVehList->GetNULLLink();
                   pLn  = pVehList->GetNextLink() ) {
         if ( ((cLinkVehGeom*)pLn)->GetHorzRange() < range ) {
	    pLink = (cLinkVehGeom*)pLn;
            range = pLink->GetHorzRange();
	 }
	 //check if camera is auto and detects target for each vehicle in range.  If so, detect.
	 if ( !detect && fpCamParam->fTypeNameIndex == TFuzeParam::esenstypeAutoCmdCamDir && 
	      fpCamParam->DetectsTgt( ((cLinkVehGeom*)pLn)->GetHorzRange() )) 
	    detect = true;       
	
      }
      
      if (fpCamParam->fTypeNameIndex == TFuzeParam::esenstypeAutoCmdCamDir) {
	 switch (newstate) {
	 case ecisOn:
	    if (detect && fpCamera->ViewAvail()) {
	       fpCamera->CamOn();
	       newstate = ecisAutoViewing;
	    }
	    break;
	 case ecisAutoShutdown:
	    if (detect) {
	       newstate = ecisAutoViewing;
	    }
	    else if (curTime >= timeOff) {
	       fpCamera->CamOff();
	       newstate = ecisOn;
	    }
	    break;
	 case ecisAutoViewing:
	    if (!detect) {
	       timeOff = curTime + 5000;
	       newstate = ecisAutoShutdown;
	    }
	    break;
	 }
      }
	    

      // found a close vehicle
      if (pLink)  {
	 //check if vehicle has run over camera.  If so, nuetralize camera
	 CalcGeom( pLink, &attGeom);
	 if ( attGeom.fbAttackNow ) {
	    /*
	    if (fpCamera->GetViewEntity()->InControl(fpCamera))
	       fpCamera->GetViewEntity()->SetControl();
	       */
	    pLink->SetInteractResult(eMineNeutralized);
	    //newstate = ecisFailed;
	 }
      }
      //Set new mine state
      if ( state != newstate )
          fpMine->StateChange( (EMineInternalState)newstate );
   }
}

/*
void cMfCameraControl::CalcGeom(const cLinkVehGeom* pLink, TAttackGeom* pGeom)
{
}
*/

EBoolean cMfCameraControl::NeedsPrUpdate( EMineInternalState minestate)
{
   return (minestate > emisOn &&
	   fFeedTime - GetCurSimTime() < fpMine->GetUpdateRate()) ? true : false;
}


EBoolean cMfCameraControl::NeedsUpdate( EMineInternalState minestate)
{
   return minestate >= emisLowestUpdate ? true : false;
}

cMfCameraControl::cMfCameraControl(cSimMine* pCamera, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt, 
                                   TSensParamCam* pCamParam ) :
            cMineControl(pCamera, offset, fuzeOpt)
{
   fpCamera = (cMfCamera*)pCamera;
   fpCamParam = pCamParam;
}

cMfCameraControl::~cMfCameraControl()
{
}


// ************************************
//   mines 
// ************************************


void     cMfCamera::CmdDet()
     // activates camera, repositions stealth entity
{
   if (GetInternalState() >= ecisOn) {
      //cout << "Camera activated." << endl;
      if (IsViewing()) {
	 // We already have the camera on, turn it off
	 CamOff();
	 StateChange ((EMineInternalState)ecisOn);
      }
      else {
	 //Turn on the camera
	 CamOn();
	 StateChange ((EMineInternalState)ecisViewing);
      }
   }
}

void     cMfCamera::CmdDet(TLocation /*tccPt*/)
     // activates camera, repositions stealth entity
{
   CmdDet();
}

void    cMfCamera::SelectObj()
{
   if ( ! fbSelected ) {
      fbSelected = true;
      if ( fpDispObj) {
         ((cMineDisp*)fpDispObj)->SetFireActiveMine();
         fpDispObj->RedrawSelected ();
      }
   }
}

void cMfCamera::DeSelectObj()
{
   if ( fbSelected ) {
      fbSelected = false;
      if ( fpDispObj) {
         fpDispObj->RedrawSelected ();
      }
   }
}


void   cMfCamera::StateChange(EMineInternalState state)
{
   EMineExternalState oldExtState = GetExternalState();
   feIntState = state;
   EMineExternalState newExtState = GetExternalState();
   
   if ( fpDispObj ) fpDispObj->DrawObj();
   
   if(state == emisMoving)
      fpMinefield->IncrementSeq();  

   if(state <= ecisOff)
      //Camera is not active.  Release the view.
      if (IsViewing())
	 CamOff();
   
   if ( oldExtState != newExtState ) {
      switch ( newExtState ) {
/*     FireMine() now takes care of fired case
         case emsFired:
               ((cMineComp*)fpMfComponent)->IncrementFired();
               if ( (fDetonationData.feDetResult == edetEntityImpact) ||
                    (fDetonationData.feDetResult == edetEntityProximate)   ) {
                  ((cMineComp*)fpMfComponent)->IncrementVehHit();
               } 
            break;
*/            
         case emsOther:
               ((cMineComp*)fpMfComponent)->IncrementOtherDead();
            break;
     }
     fpMinefield->IncrementSeq(); 
  }
      
}

EMineInternalState cMfCamera::GetDisplayState()
{
   return GetInternalState();
}

void cMfCamera::CamOn(EBoolean bDelay)
{
   TLocation viewLoc = fViewLocation;

   fprintf(stdout, "Camera has a delay of %f seconds\n", fFeedDelay);
   // If there is a feed delay, take control of camera but set an empty view
   if (bDelay && fFeedDelay != 0.0) {
      fprintf(stdout, "Setting camera for a delay of %f seconds\n", fFeedDelay);
      viewLoc.z += 2000.00;
      ((cMfCameraControl*)fpControl)->SetFeedTime (GetCurSimTime() + fFeedDelay);
   }

   //Get control of the view if it exists.  Otherwise create the view.
   if (fpViewEntity) {
      fpViewEntity->SetControl(this, &viewLoc, &fViewOrient);
   }
   else {
      fpViewEntity = new cStealthEntity(this, &viewLoc, &fViewOrient);
   }
}

void cMfCamera::CamOff()
{
   //Release the camera view.
   fpViewEntity->SetControl();
}

void cMfCamera::PanCam(EBoolean clkwise)
{
   float incr = 2.0;
   if (!clkwise)
      incr *= -1.0;

   fTccOrientation.psi += incr;
   SetTccOrient(fTccOrientation.psi, fTccOrientation.theta, 0.0);
   fViewOrient = GetTccOrient();

   CamOn();
}

void cMfCamera::TiltCam(EBoolean up)
{
   float incr = 2.0;
   if (!up)
      incr *= -1.0;

   fTccOrientation.theta += incr;
   SetTccOrient(fTccOrientation.psi, fTccOrientation.theta, 0.0);
   fViewOrient = GetTccOrient();

   CamOn();
}

EBoolean cMfCamera::IsViewing()
{
   if (fpViewEntity)
      return fpViewEntity->InControl(this);
   else
      return false;
}

EBoolean cMfCamera::ViewAvail()
{
   if (fpViewEntity)
      return fpViewEntity->InControl((cMfCamera*)NULL);
   else
      return true;
}


cMfCamera::cMfCamera( cPlanMine* pPlan ) :
      cSimMine( pPlan )
{
   switch (fsDbCamera.GetFuzeParam(fOptSet.fFuzeOpt)->fTypeNameIndex) {
      case TFuzeParam::esenstypeCommandCamDir:
            fCmdType = ecmdCamDir;
         break;
      case TFuzeParam::esenstypeAutoCmdCamDir:
            fCmdType = ecmdAutoCam;
         break;
      default:
	 cout << "cMfCamera: Unrecognized fuze type!" << endl;      
   }
   
   //When this camera is created, phi orientation angle should be 0
   //Look at a point 1 meter directly in front of camera to get 
   //a theta angle of view which follows the ground.

   double ref_elev = gpTerrainMgr->GetMapElev(fTccLocation.x + sin(fTccOrientation.psi),
					      fTccLocation.y + cos(fTccOrientation.psi));

   fTccOrientation.theta = atan(ref_elev - fTccLocation.z);
								   
   SetTccOrient(fTccOrientation.psi, fTccOrientation.theta, 0.0);

   fViewLocation = GetTccLoc();
   fViewLocation.z += ((cDbMine_cMfCamera*)fpDbMine)->GetHeight();          //Camera View "X" meters off the ground
   fViewOrient = GetTccOrient();

   fFeedDelay = ((TSensParamCam*)(((cDbMine_cMfCamera*)fpDbMine)->GetFuzeParam(fOptSet.fFuzeOpt)))->fFeedDelay;
   
   //fprintf(stdout, "Height = %f\nDelay = %u\n", ((cDbMine_cMfCamera*)fpDbMine)->GetHeight(), fFeedDelay);

}

cMfCamera::~cMfCamera()
{
   if (IsViewing()) {
      CamOff();
   }
}
