/**********************************************************************
 * File: MfCamera.h
 *
 * Types defined: 
 *
 * Functions defined: 
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1998
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */

#ifndef _UCaControl_
#define _UCaControl_

#define STEALTHPDURATE 5000

#include "Globals.h"
#include "Minefield.h"
#include "Vehicle.h"
#include "EntityDB.h"
#include "Display.h"
#include "ApMine.h"
#include "AtMine.h"
#include "AtWideArea.h"
#include "SaSensor.h"

enum   ECamInternalState {
                           ecisNone      = emisNone,
                           ecisFailed    = emisFailed,
                                 ecisLowestUpdate    = esisFailed,
                                 ecisLowestTracker   = esisFailed,
                           ecisOff       = emisOff,
                           ecisOn        = emisOn,
			   ecisAutoShutdown = emisTracking,
			   ecisAutoViewing = emisActivating,
                           ecisViewing  = emisCloseTracking
                           
                           };
// ************************************
//   stealth view types for camera view
// ************************************
class cMfCamera;

class cStealthEntity : public cSimVehicle
{
private:
   TEntityID             toStealthId;
   cMfCamera             *fpControlCam;

   enum EStealthState { estisNone=0, estisActive, estisMoving, estisNumStates };

   EStealthState feInternalState;
   TTimeSim      fNextVehStateTime;

   cMessage*     fpSendVehStateMsg;
   cMessage*     fpSendViewControlMsg;
   

public:
   virtual void           SetControl(cMfCamera* pNewControlCam = NULL, TLocation* pLoc = NULL, TOrient* pOrient = NULL);
   virtual EBoolean       InControl(cMfCamera* pMfCam) { return (pMfCam == fpControlCam) ? true : false; }
   virtual void           UpdateVehicle();
   virtual void           InitStealth();

            cStealthEntity (cMfCamera* pNewControlCam = NULL, TLocation* pLoc = NULL, TOrient* pOrient = NULL);
   virtual ~cStealthEntity();
};

// ************************************
//   component classes
// ************************************

struct TSensParamCam : public TSensParamDumb {

   float   fMaxRange;

   TTimeSim   fFeedTime;  // Time when video feed is ready
   TTimeSim   fFeedDelay;  // How long it takes to get video feed in milliseconds.
   

      // variable size arrays of ranges and detect probs

	enum {
		ecamkwFuzeTypeName,      // intercept type name, donot allow unuseable types
                ecamkwFeedDelay,         // Get feed delay from reader file in seconds
		ecamSizeTable
	};
 
   static const TKeywordEntry fKeyCamSensTable[ecamSizeTable];
   
   //EBoolean      TracksTgt( float range );

   virtual EBoolean     ParseDbStructItem(cInFileObj& infile);

//   virtual   cTracker*     CreateTracker( cEntity* pMine, const TFuzeOpt& fuzeOpt, float radius = 0.0 );
   virtual   cControl*     CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt );
   
   virtual   TFuzeParam*   CloneFuzeParam();
   
   virtual EBoolean   IsDirectional(TFlatEntCoord* pMinMax);

            TSensParamCam(int fuze = esenstypeAutoCmdCamDir);
   virtual ~TSensParamCam();
};

class cDbMine_cMfCamera : public cDbMine {
 protected:
   enum { ecamerakwHeight,  ecameraSizeTable};

   static const TKeywordEntry fKeyCameraTable[ecameraSizeTable];

   float fCamHeight;
   
 public:
   
   virtual   cDbMine*     CreateDbMine(EBoolean bWithParams);  // blank copy, doesnot fill fields
   virtual cSimMine*     CreateMine(cPlanMine* pPlan);       // bare mine, of appropriate type

   virtual EBoolean  ParseDbEntry(cInFileObj& infile);

   float     GetHeight() { return fCamHeight; }

             cDbMine_cMfCamera();
             cDbMine_cMfCamera(int sensType);
             cDbMine_cMfCamera(int sensType, TEntityType& type, int level, char* pName = NULL, char* altname = NULL);
   virtual  ~cDbMine_cMfCamera();
};

class cMfCameraControl : public cMineControl {
protected:
   cMfCamera*         fpCamera;
   TSensParamCam*  fpCamParam;

   EBoolean        fbDelaying;
   TTimeSim        fFeedTime;

public:
   virtual   void      UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList, TTimeSim* pWaitTime);
             
   //virtual   void      CalcGeom(const cLinkVehGeom* pLink, TAttackGeom* pGeom);
   void                    SetFeedTime(TTimeSim time) { fbDelaying = true; fFeedTime = time; }

   virtual      EBoolean   NeedsPrUpdate( EMineInternalState minestate);
   virtual      EBoolean   NeedsUpdate( EMineInternalState minestate);

             cMfCameraControl(cSimMine* pSensor, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt, 
                                   TSensParamCam* pCamParam );
   virtual  ~cMfCameraControl();
};



// ************************************
//   mine classes
// ************************************

class cMfCamera : public cSimMine {

 friend class cDbMine_cMfCamera;

 public:
   enum ECmdType { ecmdCamDir, ecmdAutoCam };

 private:
   static  TEntityType   fsCameraEntityType;
   static  char*         fsCameraName;
   static  char*         fsCameraAltName;
   static  int           fsDefaultCameraType;
   
   static   cDbMine_cMfCamera      fsDbCamera;      // the instantiation in mine database

   //The camera view entity
   //Only one view for all cameras.
   static  cStealthEntity*  fpViewEntity;

   TLocation  fViewLocation;
   TOrient    fViewOrient;

   TTimeSim   fFeedDelay;
   
 protected:
   ECmdType           fCmdType;

 public:
   cStealthEntity*      GetViewEntity() { return fpViewEntity; }
   EBoolean             IsViewing();
   EBoolean             ViewAvail();

   virtual      void          StateChange(EMineInternalState state);
   virtual EMineInternalState GetDisplayState();
   virtual      void    CmdDet();
   virtual      void    CmdDet(TLocation tccPt);
   virtual      void    CamOn(EBoolean bDelay = true);  // By default, there is a video feed delay
   virtual      void    CamOff();

   virtual      void    PanCam(EBoolean clkwise);
   virtual      void    TiltCam(EBoolean up);

   virtual      void    SelectObj();      // sets for cmd det
   virtual      void    DeSelectObj();

   cMfCamera( cPlanMine* pPlan = NULL );

   virtual   ~cMfCamera();
};
#endif
