/*
$RCSfile: Minefield.cc,v $
$Revision: 1.3.2.1 $
$Author: ksnively $
$Date: 1999/06/17 14:28:00 $
$Log: Minefield.cc,v $
Revision 1.3.2.1  1999/06/17 14:28:00  ksnively
Corrected bug in version 4.2 with using -mfs option with the -nogui option.
Declared version 4.2.1.

Revision 1.3  1999/05/18 19:33:22  ksnively
Ported CMS to compile under IRIX 6.5
Altered make files to allow cms to be compiled under various OS simulataneously
Removed the Startup GUI

Revision 1.2  1999/04/29 19:34:18  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:40  jchappel
Comprehensive Mine Simulator

// Revision 2.3  1998/11/25  18:37:51  ksnively
// Update for APLA Exercise
//
// Revision 2.2  1998/04/20  15:10:36  ksnively
// *** empty log message ***
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1998/01/09  18:21:47  ksnively
// Integrated latest SAIC release cms3.0b2.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.6  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.5  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.5  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.4  1997/05/20  15:35:11  ksnively
// Redraw screen when turning on or off minefield
//
// Revision 1.3  1997/03/20  19:14:41  ksnively
// added emission tracking, updated to CMS2.0B6
//
// Revision 1.2  1997/01/21  15:08:14  nsipes
// *** empty log message ***
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.20  1996/10/19  18:06:12  rcarp
// minor editing
//
// Revision 1.19  1996/09/06  18:57:57  rcarp
// CP bug fixes
//
// Revision 1.18  1996/08/21  21:58:20  rcarp
// CP and QRP file output, other updates
//
// Revision 1.17  1996/08/15  19:15:44  mkappel
// Change perimeter color when minefield state changes.
//
// Revision 1.16  1996/08/07  14:19:30  rcarp
// initial revisions for CP protocol
//
// Revision 1.15  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.14  1996/05/21  22:15:40  rcarp
// text and lists from mine db, and activating mine state save cmd
//
// Revision 1.13  1996/05/15  18:30:00  mkappel
// Bundle tactmap changes.
//
// Revision 1.11  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.9  1996/04/04  22:16:03  mkappel
// Label minefield on PVD.
//
// Revision 1.8  1996/04/04  16:25:34  mkappel
// Improve performance with large mine sets.
//
// Revision 1.7  1996/03/28  22:34:41  mkappel
// Add Mine Display toggle to component editors.
//
// Revision 1.6  1996/03/23  18:09:52  mkappel
// Fix repositioning on load of minefield.
//
// Revision 1.5  1996/03/23  14:26:12  rcarp
// improvements to test mode and selection
//
// Revision 1.4  1996/03/15  16:03:02  mkappel
// Allow repositioning of minefield when loading (standard minefields).
//
// Revision 1.3  1996/02/05  14:33:28  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: Minefield.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   Minefield.cc
   
      The minefield implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "Minefield.h"
#include "CmsControl.h"
#include "AtMine.h"
#include "FileObj.h"
#include "MapGeometry.h"
#include "Message.h"
#include "McmPdu.h"
#include "EntityDB.h"
#include "NetMgr.h"
#include "DisPdu203.h"
#include "Display.h"

#include <strings.h>
#include <bstring.h>
#include <iostream.h>
#include <stdlib.h>
#include <math.h>

#include <Xm/Xm.h>
#include <Xm/RowColumn.h>


// ************************************
//   static minefield enumeration values
// ************************************
const  TTimeSim    cMinefield::fsMfStateMsgRate = 5000;   // 5 seconds

TEntityType      cMinefield::fsMinefieldEntType( 5, 1, 225, 21, 0, 0, 0 );

int                cMinefield::fNewNameNum     = 0;   

// ************************************
//   file parameters
// ************************************
const TKeywordEntry cMfieldMgr::fKeyMmgrTable[eSizeMmgrTable] = {
	"minefield",   ekeyLev1Struct
};

const char*   cMinefield::fpMfTypeText[eNumMfTypes] = {"Deliberate", "Hasty", "Tactical"};

const TKeywordEntry cMinefield::fKeyMfldTable[eSizeMfldTable] =
{
	"type",          ekeyString,             // hasty, etc.
	"force",         ekeyString,            // friendly, etc.
	"maplocation",   ekey3DPoint,
	"geolocation",   ekey3DPoint,
	"orientation",   ekey3DPoint,
	"mfperimeter",   ekey2DPointList,
	"mapperimeter",  ekey3DPointList,
	"2dmapperimeter",  ekey2DPointList,
	"individual",    ekeyLev2Struct,
	"grid",          ekeyLev2Struct,
	"row",           ekeyLev2Struct
};

// float cTracker::fsSensRange = 100.0;

// ************************************
//   miscellany
// ************************************

TMineSumm&   TMineSumm::operator+=(const TMineSumm& mSumm)
{
   fTotalMines     += mSumm.fTotalMines;
   fLiveMines      += mSumm.fLiveMines;
   fFiredMines     += mSumm.fFiredMines;
   fOtherDeadMines += mSumm.fOtherDeadMines;
   fVehiclesHit    += mSumm.fVehiclesHit;
   fVehiclesKilled += mSumm.fVehiclesKilled;
   
   return *this;
}

TMineSumm&   TMineSumm::operator-=(const TMineSumm& mSumm)
{
   fTotalMines     -= mSumm.fTotalMines;
   fLiveMines      -= mSumm.fLiveMines;
   fFiredMines     -= mSumm.fFiredMines;
   fOtherDeadMines -= mSumm.fOtherDeadMines;
   fVehiclesHit    -= mSumm.fVehiclesHit;
   fVehiclesKilled -= mSumm.fVehiclesKilled;
   
   return *this;
}

void    TMineSumm::WriteTotals(ofstream* pOut)
{
   *pOut << " mine totals: "
         << fTotalMines << " "
         << fLiveMines << " "
         << fFiredMines << " "
         << fOtherDeadMines << " "
         << fVehiclesHit << " "
         << fVehiclesKilled;
}

TMfObjTypeDesc*   cObjTypeListHead::GetObjDesc( const TEntityType& type)
{
   for (TMfObjTypeDesc* pObjdesc = (TMfObjTypeDesc*)fpNext; 
              (T2link*) pObjdesc != (T2link*) this; 
                        pObjdesc = (TMfObjTypeDesc*)(pObjdesc->fpNext)) {
      if ( pObjdesc->fType == type ) {
         return pObjdesc;
      }
   }
   return NULL;
}

TMfObjTypeDesc::TMfObjTypeDesc(const TEntityType& type, int numTot, T2link* pPrev) :
         T2link(pPrev), fMineSumm( numTot )
{
   fType = type;
}

TMfObjTypeDesc::TMfObjTypeDesc(const TEntityType& type, const TMineSumm& msumm, T2link* pPrev) :
         T2link(pPrev)
{
   fType = type;
   fMineSumm = msumm;
}

TMfObjTypeDesc::~TMfObjTypeDesc()
{
}

cObjTypeListHead::~cObjTypeListHead()
{
}
// ************************************
//   messages
// ************************************

void    cSendMfStateMsg::SendMsg()
{
   gpNetMgr->GetPduTrans()->SendMinefieldState(fpMinefield);
   fpMinefield->ScheduleNextMfState();
}

cSendMfStateMsg::cSendMfStateMsg(cMinefield* pMinefield, TTimeSim wait) :
      cMsgLowRateUpdt(wait)
{
// cout<<"creating Send Minefield State message"<<endl;
   fpMinefield = pMinefield;
} 

cSendMfStateMsg::~cSendMfStateMsg()
{} 

void    cCreateMinefieldMsg::SendMsg()
   // message from gui to minefield manager
   //  Sending to separate host awaits starting simulating the minefield
{
   cMinefield*   pNewfield = gMfieldMgr.CreateMinefield(NULL, esimGuiInit);
   gpGuiWin->SetSelection( false, pNewfield);
   pNewfield->StartEditingMf();
   
   new cEditMinefieldMsg();
}

cCreateMinefieldMsg::~cCreateMinefieldMsg()
{
   //** at least for now, minefield keeps the string
} 

void    cReadMineFileMsg::SendMsg()
{
   gMfieldMgr.ReadMinefile(fpMineFile, fbSim);
}

cReadMineFileMsg::~cReadMineFileMsg()
{} 

void    cReadMinefieldMsg::SendMsg()
{
   fpField->ReadMinefield(fpMineFile);
}

cReadMinefieldMsg::cReadMinefieldMsg(cMinefield* pField, cInFileObj* pMineFile)
{
   fpField = pField;
   fpMineFile = pMineFile;
}
 
cReadMinefieldMsg::~cReadMinefieldMsg()
{} 

void    cCreateCompMsg::SendMsg()
{
   cMfComponent* pComp = fpMfield->CreateComp(feCompType);
   
   if (feSetCurrent) {
      new cEditCompMsg(pComp);
      pComp->StartEditingComp();

				//**test
      if (gbGui) gpGuiWin->SetSelection( false, fpMfield, pComp ); // CMS2.0B6
   }
}
 
cCreateCompMsg::~cCreateCompMsg()
{} 

void    cSimulateMinefieldMsg::SendMsg()
{
   fpField->SimulateMinefield();
   if (gbGui)
      gpGuiWin->SetEditorButtonState();
}

cSimulateMinefieldMsg::~cSimulateMinefieldMsg()
{} 

void    cOnOffMinefieldMsg::SendMsg()
{
   fpField->OnOffMinefield(fOn);
}

cOnOffMinefieldMsg::~cOnOffMinefieldMsg()
{} 

char*   cMinefield::GetEntName()
{
   return fName;
}

const TEntityType&   cMinefield::GetEntType()
{
   return fCmsType;
}

EBoolean         cMinefield::GetOnState()
   // If at least one component has same on state, we return that state..
   //  If all components have other state, we switch
{
   for ( cMfComponent* pComp = GetFirstComp();
                       pComp;
                       pComp = pComp->GetNext() ) {
      if ( pComp->IsMineComp() ) {
         if ( ((cMineComp*)pComp)->GetOnState() == fOn )
            return fOn;
      }
   }
      // all mine containing components differ
   if ( fOn ) fOn = false;
   else       fOn = true;
   
   return fOn;
}




EBoolean         cMinefield::GetShowMines()
   // If at least one component has same on state, we return that state..
   //  If all components have other state, we switch
{
   for ( cMfComponent* pComp = GetFirstComp();
                       pComp;
                       pComp = pComp->GetNext() ) {
      if ( pComp->IsMineComp() ) {
         if ( ((cMineComp*)pComp)->GetShowMines() == fShowMines )
            return fShowMines;
      }
   }
      // all mine containing components differ
   if ( fShowMines ) fShowMines = false;
   else              fShowMines = true;
   
   return fShowMines;
}

void    cMinefield::RecreateVisuals ()
{
  fpMfSumm = new cDispMfSummary(this);
  fpCompList = gpGuiWin->CreateCompListArea();
  fpCompStatus = gpGuiWin->CreateCompStatusArea();

  for ( cMfComponent* pC = GetFirstComp(); pC; pC = pC->GetNext() )
    pC->RecreateVisuals();
}

void    cMinefield::ReplaceInList (cMinefield* pMf)
{
  // this minefield is last on minefield manager list
  Assert ( gMfieldMgr.fpLastMf == this, kerrTypeMatch, 
	  "cMinefield::ReplaceInList");

  gMfieldMgr.fpLastMf = fpPrev;
  fpPrev->fpNext = NULL;

  fpNext = pMf->fpNext;
  if (pMf->fpNext)
    pMf->fpNext->fpPrev = this;

  fpPrev = pMf->fpPrev;
  if (pMf->fpPrev)
    pMf->fpPrev->fpNext = this;

  if (gMfieldMgr.fpFirstMf == pMf) gMfieldMgr.fpFirstMf = this;
  if (gMfieldMgr.fpLastMf == pMf) gMfieldMgr.fpLastMf = this;

  pMf->fpNext = NULL;
  pMf->fpPrev = NULL;
  delete pMf;
}

void    cMinefield::UpdateStatusInfo()
{
   if (fpMfSumm)
      fpMfSumm->ShowInfo();
}

void    cMinefield::ScheduleNextMfState()
{
   fpSendMfStateMsg = new cSendMfStateMsg(this, fsMfStateMsgRate);
   
}

//NPAC
void   cMinefield::Simulate() {
 if(fbSim){
 //Simulate the minefield if told to do so from the command line
   SetSimStart( esimstSimulating );
 // newly added to make mines sensitive after simulation
   cMfComponent* pComp = GetPerimeter();

 // more new stuff
    if (pComp && gbGui) {
          pComp = pComp->GetNext();
          while (pComp != NULL) {
              pComp->SetCompPvdSelect( true ); // not sure about this one
              pComp = pComp->GetNext();
          }
    }
    if (fpMfSumm) fpMfSumm->ShowInfo();
 }
}

void   cMinefield::ReadMinefield(cInFileObj* pMineFile)
{
   EFileReturn   returnCode;
   int           returnIndex;
   void*         returnVal;
   int           i;
      
   fShowMines = false;
   returnCode = pMineFile->ReadItem(fKeyMfldTable, eSizeMfldTable, &returnIndex, &returnVal);
   
   switch (returnCode) {
      case eFileNoErr:
            switch (returnIndex) {
               case emfkwForce:
                     for (i = 0; i < eNumForceTypes; i++)
                        if ( !strcmp(pForceText[i], (char*)returnVal ) ) {
                           fForce = (EForce)i;
                           goto forceDone;
                        }
                     // pMineFile->ReportError();
                  forceDone: delete[] returnVal;
                  returnVal = NULL;
                  break;
         
               case emfkwType:
                     for (i = 0; i < eNumMfTypes; i++)
                        if ( !strcmp(fpMfTypeText[i], (char*)returnVal ) ) {
                           fCmsType.specific = i;
                           goto typeDone;
                        }
                     // pMineFile->ReportError();
                  typeDone: delete[] returnVal;
                  returnVal = NULL;
                  break;
         
               case emfkwMapLoc:
		     if (! fbRepos) {
		       SetCoordTccRef(*((TLocation*)returnVal), fTccOrientation.psi, 
				      fTccOrientation.theta, fTccOrientation.phi);
//                     gpTerrainMgr->MapToSimLoc(*((TLocation*)returnVal), &fGeoLocation);
		     }
                     delete returnVal;
                     returnVal = NULL;
                  break;
         
               case emfkwGeoLoc:
		     if (! fbRepos) {
		       SetCoordGeoRef(*((TLocation*)returnVal), fGeoOrient);
//                     fGeoLocation = *((TLocation*)returnVal);
		     }
                     delete returnVal;
                     returnVal = NULL;
                  break;
         
               case emfkwOrient:
                     fGeoOrient.psi   = ((T3Ddouble*)returnVal)->x;
                     fGeoOrient.theta = ((T3Ddouble*)returnVal)->y;
                     fGeoOrient.phi   = ((T3Ddouble*)returnVal)->z;
                     SetCoordGeoRef(fGeoLocation, fGeoOrient);
                     delete returnVal;
                     returnVal = NULL;
                  break;
         
               case emfkwMfPerim:
                     GetPerimeter()->ReadMfPerim( (TItemList *)returnVal );
                     returnVal = NULL;
                  break;
         
               case emfkwMapPerim:
                     GetPerimeter()->ReadMapPerim( (TItemList *)returnVal );
                     returnVal = NULL;
                  break;
         
               case emfkw2dMapPerim:
                     GetPerimeter()->Read2dMapPerim( (TItemList *)returnVal );
                     returnVal = NULL;
                  break;
            }
            new cReadMinefieldMsg(this, pMineFile);
         break;

      case eNextLevel:   // start on a component
            switch (returnIndex) {      
               case emfkwIndiv: {
                     cIndivComp* pIndiv = new cIndivComp( this );
										 pIndiv->ShowMines(false);
                     new cReadIndivMsg(pIndiv, pMineFile);
                  }
                  break;
         
               case emfkwGrid: {
                     cGridComp* pGrid = new cGridComp( this );
										 pGrid->ShowMines(false);
                     new cReadGridMsg(pGrid, pMineFile);
                  }
                  break;
   
               case emfkwRow: {
                     cRowComp* pRow = new cRowComp( this );
		     pRow->ShowMines(false);
                     new cReadRowMsg(pRow, pMineFile);
                  }
                  break;
            }
         break;

      case eLevelDone:   // done reading this minefield
	    feSimState = esimFileRead;
	    //** should give user msg if field is incomplete
	    if ( ! DoCompleteMf() ) delete this;

	    if (fbRepos && gbGui) {
	      gpGuiWin->SetSelection( false, this, GetPerimeter() );
	      new cEditCompMsg(gpGuiWin->GetCurComponent());
	    }
	    else {
	      new cReadMineFileMsg(pMineFile, fbSim);
	      SetPvdSelect( true);

	      /*NPAC if (fbSim) {
		 //Simulate the minefield if told to do so from the command line 
		 fprintf(stdout, "Setting simulation of minefield\n");
		 SetSimStart( esimstSimulating );
		 // newly added to make mines sensitive after simulation
		 cMfComponent* pComp = GetPerimeter(); 
	       
		 // more new stuff
		 if (pComp && gbGui) {
		    pComp = pComp->GetNext();
		    while (pComp != NULL) {
		       pComp->SetCompPvdSelect( true ); // not sure about this one
		       pComp = pComp->GetNext();
		    }
		 }
		 if (fpMfSumm)
		    fpMfSumm->ShowInfo();
	      }
              */
	      break;
	    }
	 break;
	    
      default: // error returns.  Give the user a message
            Assert(false, kerrReadMinefield, "cMfieldMgr::ReadMinefield");
            delete pMineFile;
         break;
   }
}

void    cMinefield::SaveMinefield(cOutFileObj* pMineFile)
   // writes minefield level data, and starts messages to save components
{
      // manager has already written name
   pMineFile->WriteItem(fKeyMfldTable[emfkwType].key,
                        fKeyMfldTable[emfkwType].type,
                        (void*) (fpMfTypeText[fMfType]) );
   
   pMineFile->WriteItem(fKeyMfldTable[emfkwForce].key,
                        fKeyMfldTable[emfkwForce].type,
                        (void*) (pForceText[fForce]) );
   
   pMineFile->WriteItem(fKeyMfldTable[emfkwMapLoc].key,
                        fKeyMfldTable[emfkwMapLoc].type,
                        (void*) (&fTccLocation) );

   fpNextSaveComp = fpCompHead;
   new cSaveCompMsg(this, pMineFile);
}

void    cMinefield::SaveNextComponent(cOutFileObj* pMineFile)
{
   if (fpNextSaveComp) {
      fpNextSaveComp->SaveComponent(pMineFile);
      fpNextSaveComp = fpNextSaveComp->GetNext();
      new cSaveCompMsg(this, pMineFile);
   }
   else {   // tell the manager to save the next minefield on list
      new cSaveMinefieldMsg();
   }
}

void    cMinefield::CmdFire()    // called from GUI, directly
{
   for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() )
      pC->CmdFire(feMineCmdMode);
}

void    cMinefield::CmdMineDetect(TTimeSim delayOut, /* time interval, trip msg */
                                  TTimeSim delayIn, /* time interval, fire msg */
                                  TTimeSim timeout /* clock time */ )
   // a command-detonated mine within the field has been triggered.
   // The minefield sends a msg to itself to turn on the operator's
   // fire button, after the commmunications delay
{
   if ( timeout >= fFireBitOffTime ) {
      fFireBitOffTime = timeout;
      fFireMsgDelay = delayIn;
   }
      
   TTimeSim stopDelay = timeout - GetCurSimTime();
   if ( stopDelay > delayOut ) {
      new cFireButtonActivateMsg( delayOut, true, this);
      new cFireButtonActivateMsg( stopDelay, false, this);
   }
}

void     cMinefield::ProcessFireButMsg(EBoolean on)
   // note that more than one mine may have tripped and caused sending
   // the cFireButtonActivateMsg, so that we need to ensure that this is
   // the message that should be heeded. We do this by checking against the
   // fFireBitOffTime.
{
   if ( (on == true) && 
        (fbFireButtonOn == false) &&
        (GetCurSimTime() < fFireBitOffTime) ) {  // turn on
      gpGuiWin->ProcessFireButMsg(this, true);
      fbFireButtonOn = true;
   }
   else if ( (on == false) && 
             (fbFireButtonOn == true) &&
             (GetCurSimTime() >= fFireBitOffTime) ) { // turn off
      gpGuiWin->ProcessFireButMsg(this, false);
      fbFireButtonOn = false;
   }
}

EBoolean    cMinefield::DoCompleteMf()
{
   if ( ! DoMfCoords() ) return false;
   
   if (fNumComponents == 0) return false;
   if ( gbGui )
     tactmap_bundle_changes (gpTactmap);
   for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() ) {
      if ( ! pC->CompleteComp() ) return false;
   }
   
   feSimState = esimComplete;
   return true;
}

void    cMinefield::ResetCoords()
{
   fGeoLocation.x = fGeoLocation.y = fGeoLocation.z = 0.0;

   for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() )
      pC->ResetTccCoord();
   
   DoCompleteMf ();
}

EBoolean    cMinefield::HasMfCoords()
{
   if ( fGeoLocation.x != 0.0 || fGeoLocation.y != 0.0 || fGeoLocation.z != 0.0 )
      return true;
   else
      return false;
}

void     cMinefield::SetMfName( char* name )
{
  fName = cppstrdup(name); 
  if (GetPerimeter()->GetPerimDisp())
    GetPerimeter()->GetPerimDisp()->UpdateLabelName (name);
}

void     cMinefield::SetMfElevation( TEntCoord* mfLoc )
{
   mfLoc->z = gpTerrainMgr->GetMapElev(mfLoc->x + fTccLocation.x, mfLoc->y + fTccLocation.y) - fTccLocation.z;
}

void     cMinefield::MapToMfLoc(const TLocation& mapLoc, TEntCoord* pMfLoc)
{
   pMfLoc->x = mapLoc.x - fTccLocation.x;
   pMfLoc->y = mapLoc.y - fTccLocation.y;
   pMfLoc->z = mapLoc.z - fTccLocation.z;
}

void     cMinefield::MapToMfLoc(const TLocation& mapLoc, TFlatEntCoord* pMfLoc)
{
   pMfLoc->x = mapLoc.x - fTccLocation.x;
   pMfLoc->y = mapLoc.y - fTccLocation.y;
}

void     cMinefield::MfToMapLoc(const TEntCoord& mfLoc, TLocation* pMapLoc)
{
   pMapLoc->x = mfLoc.x + fTccLocation.x;
   pMapLoc->y = mfLoc.y + fTccLocation.y;
   pMapLoc->z = mfLoc.z + fTccLocation.z;
}

void     cMinefield::MfToMapLoc(const TFlatEntCoord& mfLoc, TLocation* pMapLoc)
{
   pMapLoc->x = mfLoc.x + fTccLocation.x;
   pMapLoc->y = mfLoc.y + fTccLocation.y;
   pMapLoc->z = gpTerrainMgr->GetMapElev(pMapLoc->x, pMapLoc->y);
}

void     cMinefield::SetMineSumm()
{
   fMineTypeList.DeleteList();
   fMineSumm.Clear();
   fNumMineTypes = fNumDetectable = 0;
   
   for ( cMfComponent* pComp = fpCompHead;
                       pComp != NULL;
                       pComp = pComp->GetNext() ) {
      if (pComp->IsMineComp()) {
         const TMineSumm&   mineSumm  = ((cMineComp*)pComp)->GetMineSumm();
         const TEntityType& type       = ((cMineComp*)pComp)->GetDetectEntType();
         
         fMineSumm += mineSumm;
         
//    cout<<" component added "<< fMineSumm.fTotalMines<<endl;
    
        TMfObjTypeDesc* pDesc = fMineTypeList.GetObjDesc( type );
         if (pDesc) {      // already have some of this type
            pDesc->fMineSumm += mineSumm;
         }
         else {            // create new type entry
            new TMfObjTypeDesc(type, mineSumm, fMineTypeList.GetLastLink());
            fNumMineTypes += 1;
        }
         fNumDetectable += mineSumm.GetDetectable();
      }
   }
   
   if (fpMfSumm)
      fpMfSumm->ShowInfo();
}

EBoolean    cMinefield::DoMfCoords()
{
   EBoolean good = true;
   
      // geo location either set already, or from perimeter tcc
   if ( fGeoLocation.x == 0.0 && fGeoLocation.y == 0.0 && fGeoLocation.z == 0.0 ) {
      fGeoOrient.psi = fGeoOrient.theta = fGeoOrient.phi = 0.0;
      good = GetPerimeter()->GetMfGeoLoc(&fGeoLocation);
   }

   gpTerrainMgr->GetMapRotationGeoPt( &fGeoRotation, fGeoLocation);
   fGeoRotation.SetOrientation( &fGeoOrient );
   gpTerrainMgr->SimToMapLoc( fGeoLocation, &fTccLocation );
   
   if ( good )
      return true;
   else
      return false;
}

short cMinefield::GetMinefieldType()
{
	if(fMinefieldType.ATbit)
	{
		if(fMinefieldType.APbit)
			return 0;
		else
			return 2;
	}

	else
		return 1;
}

	
void cMinefield::SetMinefieldType()
{
	bzero(&fMinefieldType, sizeof(fMinefieldType));
	// check each component to see if it has matching type
	for (cMfComponent* pC = GetFirstComp(); pC; pC = pC->GetNext())
	{
		if ( pC->IsMineComp()  )
		{
			short domain = ((cMineComp*)pC)->GetFirstMine()->GetDomain(); 

			if(domain == emundomAntipersonnel)
				fMinefieldType.APbit = 1;	
			else if(domain == emundomAntitank)
				fMinefieldType.ATbit = 1;	
			else
				printf("MinefieldType other than AT or AP\n");
		}
	}
}


void    cMinefield::SimulateMinefield()
   //** fill in summary data, if not done as components are defined
{
      //** need to check return, notify user if cannot complete
   if (feSimState != esimComplete) 
      if ( !DoCompleteMf() ) return;
      
   gCmsMgr.GetNextEntityId(this, &fId);
   
      // write info to mine state file
   if ( gfMineStateSumm ) {
      *gpMineStateSumm << "{ MinefieldSimStart\n"
                       << "Simulation start time: "
                       << GetCurSimTime() * kSimTimeToSec
                       << endl;
      WriteMinefieldSumm(gpMineStateSumm);
   }

   // determine minefield type (at,ap,mixed)
   SetMinefieldType();
   fbMultiShotCmdDet = false;   // may get overwritten by mine
      
      // simulate the components
   for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() )
      pC->SimulateComponent(true);

   if ( gfMineStateSumm ) *gpMineStateSumm << "}\n" << endl;
         
   feSimState = esimSimulated;
   fbSimulating = true;
   if ( AcceptsCommands() )
      feMineCmdMode       = eCmdModeFire;
   fbFireButtonOn = false;
   fFireBitOffTime = 0.0;
   fFireMsgDelay = 0.0;
   
//   feSimStart used by gui, set on toggle button
   OnOffMinefield (fOn);
   
   fpSendMfStateMsg = new cSendMfStateMsg(this, fsMfStateMsgRate);
   
   MinefieldChange();   // first update of change number
   
   
         //** special for testing
/*
   cPerimeter*   pPerimeter = GetPerimeter();
   TEntCoord* pPerim = new TEntCoord[ pPerimeter->GetNumPoints() ];
   for ( int i = 0; i < pPerimeter->GetNumPoints(); i++) 
      pPerim[i] = pPerimeter->GetPointNum(i+1)->GetMfPoint();
     
     
   TEntityType type(2, kMatchAllChar, kMatchAllShort, kMatchAllChar, kMatchAllChar, kMatchAllChar, kMatchAllChar);
   new cMinefieldRequestMsg(this, GetEntityID(), 1, 
               pPerimeter->GetNumPoints(), 0, type, pPerim);
   
*/
}

void  cMinefield::OnOffMinefield(EBoolean on)
   // called when we're already simulating
{
   fOn = on;
   
   for (cMfComponent* pComp = fpCompHead; pComp; pComp = pComp->GetNext() ) 
      pComp->OnOffComponent(on);
}

void  cMinefield::ShowMinefield (EBoolean show)
{
  GetPerimeter()->ShowPerim (show);

  for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() ) {
    if (pC->GetCompType() == eCompIndiv || pC->GetCompType() == eCompGrid) {

      /* if first mine is shown and we're trying to show 
	 (or vice versa) then exit to save time */
      cMine* pMine = ((cMineComp*)pC)->GetFirstMine();
      if (pMine && pMine->GetDispObj()->IsVisible() == show) return;

      while (pMine != NULL) {
	((cMfieldObj*)pMine)->ShowObj (show);
	pMine = pMine->GetNextMine();
      }
    }
  }
}

void cMinefield::ProcessMfRequest(TEntityID requestor, int reqNum, unsigned short filter,
                      TEntityType objType, int numPts, TEntCoord* pPerim,
                        short numReqSensors, TEntityType* pReqSensorTypes)
{
	if(pReqSensorTypes == NULL)
		ProcessMfRequest1(requestor,reqNum,filter,objType,numPts,pPerim);
	else	
		ProcessMfRequest3(requestor,reqNum,filter,objType,numPts,pPerim,numReqSensors,pReqSensorTypes);
}

void   cMinefield::ProcessMfRequest1(TEntityID requestor, int reqNum, unsigned short /*filter*/,
                      TEntityType objType, int numPts, TEntCoord* pPerim)
{
	T2linkHead*       pTypeListHead = new T2linkHead;
	T2linkMineList*   pMineListLink;
   
	// check each component to see if it has matching type
	for (cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext()) 
	{
		if ( pC->IsMineComp() )
		{
			const TEntityType& type = ((cMineComp*)pC)->GetDetectEntType();
			if ( objType.MatchType(type) ) 
			{
				// type matches, either create new mine list or add to existing
				pMineListLink = NULL;
				for ( T2link* pLink  = pTypeListHead->GetFirstLink();
					pLink != pTypeListHead->GetNULLLink();
					pLink  = pTypeListHead->GetNextLink() ) 
				{
					if ( ((T2linkMineList*)pLink)->GetType() == type ) 
					{
						// cout<<"processing request, have type match, adding to old list "<<endl;
						pMineListLink = (T2linkMineList*)pLink;
						break;
					}
				}
				if ( !pMineListLink)
				{
					// cout<<"processing request, have type match, creating new list "<<endl;
					pMineListLink = new T2linkMineList(type, pTypeListHead->GetLastLink() );
				}
               
				// go through mines in component, adding those in perimeter
				for (cMine* pM = ((cMineComp*)pC)->GetFirstMine(); 
					pM; 
					pM = pM->GetNextMine() ) 
				{
					if (  (pM->IsDetectable()) &&
						(PtInPolygon( pM->GetMfLoc(), numPts, pPerim )) ) 
					{
						pMineListLink->AddMine(pM);
						// cout<<"mine in polygon, adding"<<endl;
					}
				}
			}
		}
	}
   
	//   cout<<"created total response, sending msg back to network mgr"<<endl;
	new cMinefieldResponseMsg(this, pTypeListHead, requestor, reqNum);

	delete[] pPerim;
}

void   cMinefield::ProcessMfRequest3(TEntityID requestor, int reqNum, unsigned short /*filter*/,
                      TEntityType objType, int numPts, TEntCoord* pPerim,
			short numReqSensors, TEntityType* pReqSensorTypes)
{
	T2linkHead*       pTypeListHead = new T2linkHead;
	T2linkMineList*   pMineListLink;
  
	// check each component to see if it has matching type
	for (cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext()) 
	{
		if ( pC->IsMineComp() ) 
		{
			const TEntityType& type = ((cMineComp*)pC)->GetDetectEntType();
			if ( objType.MatchType(type) ) 
			{
				// type matches, either create new mine list or add to existing
				pMineListLink = NULL;
				for ( T2link* pLink  = pTypeListHead->GetFirstLink();
					pLink != pTypeListHead->GetNULLLink();
					pLink  = pTypeListHead->GetNextLink() ) 
				{
					if ( ((T2linkMineList*)pLink)->GetType() == type ) 
					{
						// cout<<"processing request, have type match, adding to old list "<<endl;
						pMineListLink = (T2linkMineList*)pLink;
						break;
					}
				}
				if ( !pMineListLink) 
				{
					// cout<<"processing request, have type match, creating new list "<<endl;
					pMineListLink = new T2linkMineList(type, pTypeListHead->GetLastLink() );
				}
              
				// go through mines in component, adding those in perimeter
				for (cMine* pM = ((cMineComp*)pC)->GetFirstMine();
					pM;
					pM = pM->GetNextMine() ) 
				{
					if (  (pM->IsDetectable()) &&
						(PtInPolygon( pM->GetMfLoc(), numPts, pPerim )) ) 
					{
						pMineListLink->AddMine(pM);
						// cout<<"mine in polygon, adding"<<endl;
					}
				}
			}
		}
	}


	//   cout<<"created total response, sending msg back to network mgr"<<endl;
	new cMinefieldResponseMsg(this, pTypeListHead, requestor, reqNum, numReqSensors,
			pReqSensorTypes);
	delete[] pPerim;
}

void  cMinefield::SetMfOnState (EBoolean on)
{
  fOn = on;
  
     // this call came from the gui, so we should check feSimStart
     // rather than feSimState to see if the user has started us
     // simulating
  
  if ( feSimStart == esimstSimulating )
     new cOnOffMinefieldMsg(this, on);      // turns all components
  else
     for ( cMfComponent* pComp = GetFirstComp();
                         pComp;
                         pComp = pComp->GetNext() ) {
        if ( pComp->IsMineComp() )
           ((cMineComp*)pComp)->SetOnState(on);
     }
}

void  cMinefield::ShowMines (EBoolean show)
{
  fShowMines = show;
  
  for ( cMfComponent* pComp = GetFirstComp();
                      pComp;
                      pComp = pComp->GetNext() ) {
    if ( pComp->IsMineComp() )
      ((cMineComp*)pComp)->ShowMines (show);
  }
}

void  cMinefield::RedrawMines ()
{
  for ( cMfComponent* pComp = GetFirstComp();
                      pComp;
                      pComp = pComp->GetNext() ) {
     if ( pComp->IsMineComp() ) {
	for ( cMine* pM = ((cMineComp*)pComp)->GetFirstMine(); pM; pM = pM->GetNextMine() ) {
	   if ( pM->GetDispObj() ) pM->GetDispObj()->DrawObj();
	}
     }
  }
}

cMfComponent*   cMinefield::CreateComp(ECompType eCompType)
{
   cMfComponent* pComp = NULL;
   
   switch (eCompType) {
      case eCompPerim:
            pComp = new cPerimeter(this);
         break;
      
      case eCompIndiv:
            pComp = new cIndivComp(this);
         break;
      
      case eCompGrid:
            pComp = new cGridComp(this);
         break;

      case eCompRow:
            pComp = new cRowComp(this);
         break;
   }
   
   return pComp;
}

void    cMinefield::DeleteComp(cMfComponent* pComp)
{
  delete pComp;
}

void     cMinefield::SetSimStart(ESimStart simStart, TTimeSim time)
{
   feSimStart = simStart;
   fSimTime = time;
   
   SimulateMinefield();
/* NPAC 
   if (simStart == esimstSimulating)
      new cSimulateMinefieldMsg(this);
*/
}

cMinefield*    cMinefield::DuplicateMf(T2dMove& move)
{
   cMinefield* pNewMf = new cMinefield(this, false);
   pNewMf->MoveMinefield(move);
   
   DoCompleteMf();
   pNewMf->SetMineSumm();  // Set the mine summary for new minefield
   
   return pNewMf;
}

void    cMinefield::StartEditingMf()
{
   feIsEditing = true;
//   fpCompList->ShowInfo();
   
   if (fpMfBackup)
      delete fpMfBackup;
      
   fpMfBackup = NULL;
   
   if ( (GetNumComps() > 1) || (GetPerimeter()->GetNumPoints() > 0) ) 
      fpMfBackup = new cMinefield(this, true);
}

void    cMinefield::RestoreMfBackup()
{
   CopyParams(fpMfBackup);
   while (fpCompHead)    // delete all components
      delete fpCompHead;
   
   T2dMove move(0.0, 0.0, 0.0);
   for ( cMfComponent* pC = fpMfBackup->GetFirstComp();
                       pC;
                       pC = pC->GetNext() ) {
      pC->DuplicateComp(move, this, false);
   }
   
   if (fpMfSumm)
      fpMfSumm->ShowInfo();
}

void    cMinefield::StopEditingMf()
{
   feIsEditing = false;
   if (gbGui) gpGuiWin->SetSelection( false, this);   // removes selection of components
 
   if (fpMfBackup) {
      delete fpMfBackup;
      fpMfBackup = NULL;
   }
}

//NPAC
void   cMinefield::UpdateMfield()
{
   for (cMfComponent* pComp = fpCompHead; pComp; pComp = pComp->GetNext() )
      if ( ( pComp->IsMineComp() ) &&
           ( ((cMineComp*)pComp)->IsSimulating() ) )
         pComp->UpdateComponent();
}


void   cMinefield::InDetMfield(cDetonation* pDetonation)
{
   for (cMfComponent* pComp = fpCompHead; pComp; pComp = pComp->GetNext() )
      if ( ( pComp->IsMineComp() ) &&
           ( ((cMineComp*)pComp)->IsSimulating() ) )
         pComp->InDetComp(pDetonation);
}

void     cMinefield::SetPvdSelect( EBoolean on)
{
      // the perimeter is the sensitive display
   cPerimeter* pPerim = GetPerimeter();
   if ( pPerim ) {
      pPerim->SetPerimSelect(on);
   }
}

void     cMinefield::SelectMfield()
{
      // minefield level highlighting, which includes perimeter PVD
      //  but not perimeter component listing in editor
   if ( ! feSelected ) {
      fpMfSumm->SwitchHilite();
      feSelected = true;
      GetPerimeter()->GetPerimDisp()->RedrawSelected();
   }
}

void     cMinefield::DeSelectMfield()
{ 
   if ( feSelected ) {
      fpMfSumm->SwitchHilite();
      feSelected = false;
      GetPerimeter()->GetPerimDisp()->RedrawSelected();
      
      XtUnmanageChild(fpCompList->GetMainWin());
      XtUnmanageChild(fpCompStatus->GetMainWin());
   }
}

EBoolean      cMinefield::AcceptsCommands()
{
   for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() )
      if ( pC->AcceptsCommands() )
         return true;
         
   return false;
}

void cMinefield::MoveMinefield (T2dMove& move)
   //  Instead of the minefield moving its own coord system, we move
   // each component and then re-set the minefield's system from the
   // moved perimeter.  This allow us to keep the minefield system pointing
   // east/north/up.
{
   EBoolean success = true;
   
   if ( ! HasMfCoords() )
      success = DoCompleteMf();
      
   if ( success ) {
      for ( cMfComponent* pC = fpCompHead; pC && success; pC = pC->GetNext() )
         success = pC->MoveComp (move);
         
      if ( success ) {
         fGeoLocation.x = fGeoLocation.y = fGeoLocation.z = 0.0;
         success = DoCompleteMf();
      }
   }
   
   if ( ! success && gbGui)
      gpGuiWin->DialogMessage (cGuiWindow::eDialogInfo, 
			   NULL,
			   "Failure in trying to move minefield. Please delete.");
}

void    cMinefield::MakeMineCompBackups()
   // not the perimeter, this is called from perimeter
{
   cMfComponent* pC = fpCompHead;
   for ( pC = pC->GetNext(); pC; pC = pC->GetNext() )
      pC->MakeBackup();
}

void    cMinefield::RestoreMineCompBackups()
   // not the perimeter, this is called from perimeter
{
   cMfComponent* pC = fpCompHead;
   for ( pC = pC->GetNext(); pC; pC = pC->GetNext() )
      pC->RestoreBackup();
}

void    cMinefield::DeleteMineCompBackups()
   // not the perimeter, this is called from perimeter
{
   cMfComponent* pC = fpCompHead;
   for ( pC = pC->GetNext(); pC; pC = pC->GetNext() )
      pC->DeleteBackup();
}

void cMinefield::ResetComponents ()
{
  cPerimeter*   pPerim = GetPerimeter();
  cMine*        pMine, *pNextMine;
  TPerimPoint*  pPt;
  char          msg[110];
  int           icount = 0, rcount = 0, iPt;

  ResetAllGrids();

  for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() ) {

    /* check if all individual component mines are within new perimeter */

    if (pC->GetCompType() == eCompIndiv) {

      pMine = ((cMineComp*)pC)->GetFirstMine();
      while (pMine) {
	pNextMine = pMine->GetNextMine();

	if (! pPerim->PointInside (pMine->GetTccLoc())) {
	  ((cMineComp*)pC)->DeleteMine (pMine);
	  icount++;
	}
	pMine = pNextMine;
      }
    }

    /* check if all row component vertices are within new perimeter */

    else if (pC->GetCompType() == eCompRow) {
      iPt = 1;
      while (1) {
	pPt = ((cRowComp*)pC)->GetPointNum (iPt);
	if (pPt == NULL) break;
	if (! pPerim->PointInside (pPt->GetTccPoint())) {
	  ((cRowComp*)pC)->DeletePoint (iPt);
	  rcount++;
	}
	else
	  iPt++;
      }
    }
  }

  if (icount + rcount > 0 && gbGui) {
    sprintf (msg, 
	     "%d individual mine(s) and %d row vertices\nwere lost because they fell outside\nnewly drawn perimeter.",
	     icount, rcount);
    gpGuiWin->DialogMessage (cGuiWindow::eDialogInfo, NULL, msg);
  }

  ResetCoords();
}

void cMinefield::ResetAllGrids ()
{
  for ( cMfComponent* pC = fpCompHead; pC; pC = pC->GetNext() ) {
    if (pC->GetCompType() == eCompGrid)
      ((cGridComp*)pC)->EmplaceGrid ( GridMineCount( ((cGridComp*)pC)->GetSpacing () ) );
  }
}

int  cMinefield::GridMineCount (float spacing)
{
  TPerimPoint        *pPt;
  cPerimeter*         pPerim = GetPerimeter();
  TLocation           tccPt;
  float64             xmin =  DBL_MAX, ymin =  DBL_MAX, 
                      xmax = -DBL_MAX, ymax = -DBL_MAX;
  int                 count = 0;

  if (! pPerim) return 0;

  for (pPt = pPerim->GetFirstPoint(); pPt; pPt = pPerim->GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    if (tccPt.x < xmin) xmin = (int)tccPt.x;
    if (tccPt.x > xmax) xmax = (int)tccPt.x;
    if (tccPt.y < ymin) ymin = (int)tccPt.y;
    if (tccPt.y > ymax) ymax = (int)tccPt.y;
  }

  if (xmax <= xmin || ymax <= ymin) return 0;

  for (tccPt.x = xmin; tccPt.x < xmax; tccPt.x += spacing) {
    for (tccPt.y = ymin; tccPt.y < ymax; tccPt.y += spacing) {
      if (pPerim->PointInside (tccPt))
	if (++count > kMaxMinesInField) return count;
    }
  }
  return count;
}

void    cMinefield::WriteMinefieldSumm(ofstream* pOut)
{
   *pOut << "\""
         << GetMfName()
         << "\"";
   *pOut << " id: "
         << GetEntityID().entityNum;
   *pOut << " force: "
         << (int)fForce;
   *pOut << " mapLocation: "
         << fTccLocation.x << " "
         << fTccLocation.y << " "
         << fTccLocation.z;
   int   utmZone = 0;
   double utmNorth = 0.0, utmEast = 0.0, utmZ = 0.0;
   if ( ! gpTerrainMgr->GeoToUtmLoc(fGeoLocation, &utmZone, &utmNorth, 
                                    &utmEast, &utmZ) )
      *pOut << " utmLocation: "
            << utmZone  << " "
            << utmNorth << " "
            << utmEast  << " "
            << utmZ     << " ";
   else
      *pOut << " utm coordinates not available" 
            << " ";
            
   fMineSumm.WriteTotals(pOut);
   *pOut << endl;
}

void   cMinefield::InitMinefield(char* pName, cMinefield* pOldMf, EBoolean isBackup)
{
   gMfieldMgr.ZfListCount++;
   if ( isBackup) {
      fpNext = NULL;
      fpPrev = NULL;
   }
   else {
      fpNext = NULL;
      fpPrev = gMfieldMgr.fpLastMf;
      if (fpPrev)  fpPrev->fpNext = this;
      gMfieldMgr.fpLastMf = this;
      if ( !gMfieldMgr.fpFirstMf) gMfieldMgr.fpFirstMf = this;
   }

   fNumMineTypes       = 0;
   fNumDetectable      = 0;
   fNumFixtureTypes    = 0;
   fNumFixtures        = 0;
   fpSendMfStateMsg    = NULL;
   fpCpMfMsg           = NULL;

   fNumComponents      = 0;
   fpCompHead          = NULL;
   fpCompTail          = NULL;
   fpMfBackup          = NULL;
   fpNextSaveComp      = NULL;
   feSelected          = false;
   feIsEditing         = false;
   fChangeNum          = 0;
   fCurMfStateSeq.u    = 0;
   fLastMfStateSeq.u   = 0;
   fbSim               = false;
   
 
   if (pOldMf)
      CopyParams(pOldMf);
   else {
      if ( pName )
         fName = pName;
      else {
         fName            = new char [kNameLength];
         sprintf (fName, "Minefield %d", ++fNewNameNum);
      }
      fForce              = eFriendly;
      fMfType             = eDeliberate;
      feSimStart          = esimstCommand;
      feSimState          = esimGuiInit;
      fOn                 = true;
      fSimTime            = 0.0;
      fEmplacementTime    = GetCurSimTime();
      fShowMines          = true;
      fShowLabel	  = true;
      feMineCmdMode       = eCmdModeNone;
   }
        
   if ( (!isBackup) && gbGui) {
      fpMfSumm = new cDispMfSummary(this);
      fpCompList = gpGuiWin->CreateCompListArea();
      fpCompStatus = gpGuiWin->CreateCompStatusArea();
   }
   else {
      fpMfSumm = NULL;
      fpCompList = NULL;
      fpCompStatus = NULL;
   }

}

void   cMinefield::CopyParams( cMinefield* pOldMf)
{
   fName   = new char [kNameLength];
   strcpy( fName, pOldMf->fName);
   fForce              = pOldMf->fForce;
   fMfType             = pOldMf->fMfType;
   feSimStart          = esimstCommand;
   feSimState          = esimGuiInit;  // cannot init in simulated state
   fOn                 = pOldMf->fOn;
   fSimTime            = pOldMf->fSimTime;
   fEmplacementTime    = pOldMf->fEmplacementTime;
   fShowMines          = pOldMf->fShowMines;
   fMinefieldType      = pOldMf->fMinefieldType;
   feMineCmdMode       = pOldMf->feMineCmdMode;
}

void   cMinefield::SaveMfState()
{   
      // write info to mine state file
   if ( fbSimulating && gfMineStateSumm ) {
      *gpMineStateSumm << "{ Save State Command\n"
                       << "at time: "
                       << GetCurSimTime() * kSimTimeToSec
                       << endl;
      WriteMinefieldSumm(gpMineStateSumm);
      
      for ( cMfComponent* pComp = fpCompHead; pComp; pComp = pComp->GetNext() ) {
         pComp->SaveCompState();
      }
      *gpMineStateSumm << "}\n" << endl;
   }
}

void   cMinefield::MinefieldChange()
{
   fChangeNum++;
   
   IncrementSeq();
   
   if ( fpCpMfMsg ) {   // already a msg in queue
      if ( fpCpMfMsg->GetSendTime() > GetCurSimTime() ) {
                  // its a future msg, reschedule for now
         fpCpMfMsg->GetMsgList()->DeleteMsg( fpCpMfMsg );
         fpCpMfMsg = new cSendCpMfMsg( this, 0 );
      }
   }
   
   else
      fpCpMfMsg = new cSendCpMfMsg( this, 0 );
}

cMinefield::cMinefield( cMinefield* pOldMf, EBoolean isBackup )
      // for duplicating and backing up
{
   fCmsType = fsMinefieldEntType;
   fAltType = fsMinefieldEntType;
   InitMinefield(NULL, pOldMf, isBackup);

   T2dMove move(0.0, 0.0, 0.0);
   for ( cMfComponent* pC = pOldMf->fpCompHead; pC; pC = pC->GetNext() ) {
     pC->DuplicateComp(move, this, isBackup);
   }
}

cMinefield::cMinefield( char* pName, ESimState startSimState)
   // id, loc, orient have constuctors init to zeros
{   
   fCmsType = fsMinefieldEntType;
   fAltType = fsMinefieldEntType;
   InitMinefield(pName);
   
   feSimState          = startSimState;
  
//   new cCreateCompMsg(this, eCompPerim, false); 

   CreateComp(eCompPerim);
}

cMinefield::~cMinefield()
{
    gMfieldMgr.ZfListCount--;
   if (fpNext) fpNext->fpPrev = fpPrev;
   if (fpPrev) fpPrev->fpNext = fpNext;
   if (gMfieldMgr.fpFirstMf == this) gMfieldMgr.fpFirstMf = fpNext;
   if (gMfieldMgr.fpLastMf == this) gMfieldMgr.fpLastMf = fpPrev;
   
   
      // write info to mine state file
   if ( fbSimulating && gfMineStateSumm ) {
      *gpMineStateSumm << "{ MinefieldDelete\n"
                       << "Simulation end time: "
                       << GetCurSimTime() * kSimTimeToSec
                       << endl;
      WriteMinefieldSumm(gpMineStateSumm);
   }
   
   while (fpCompHead) delete fpCompHead;
   
   if ( fbSimulating && gfMineStateSumm ) *gpMineStateSumm << "}\n" << endl;

   if (gbGui) {
      if (fpMfSumm) {
         delete fpMfSumm;
         fpMfSumm = NULL;
      }
      if (fpCompList) {
         delete fpCompList;
         fpCompList = NULL;
      }
      if (fpCompStatus) {
         delete fpCompStatus;
         fpCompStatus = NULL;
      }
   }
   if (fpSendMfStateMsg) (fpSendMfStateMsg->GetMsgList())->DeleteMsg(fpSendMfStateMsg);
   if (fpCpMfMsg) (fpCpMfMsg->GetMsgList())->DeleteMsg(fpCpMfMsg);
   if (fName) {
      delete[] fName;
      fName = NULL;
   }
   fMineTypeList.DeleteList();
   fFixtureTypeList.DeleteList();

   if (gbGui) tactmap_refresh_non_terrain (gpTactmap);
}

cMinefield*   cMfieldMgr::CreateMinefield(char* pName, ESimState startSimState)
{
   cMinefield* pNewField = new cMinefield(pName, startSimState);
   
   return pNewField;
}

EBoolean cMfieldMgr::SaveMinefile (char* filename, int count,
				 cMinefield** ppMfList, char** returnMsg)
    // if problem setting up file, returns false with message
{
   *returnMsg = new char[100];
   
   Assert ( feMfMgrState == emgrstIdle, kerrValidityCheck, "cMfieldMgr::SaveMinefile");
   
   if (count <= 0) {
      *returnMsg = "No minefields selected to save";
      return false;
   }
   if (fpMineFile) {
      *returnMsg = "Please wait until last save is completed";
      return false;
   }
   
   fpMineFile = new cOutFileObj;
   if ( !fpMineFile->Open(filename, gftypeMinefields) ) {
      delete fpMineFile;
      fpMineFile = NULL;
      sprintf (*returnMsg, "Cannot open %s for output", filename);
      return false;
   }   
   
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstSaving;
   fppMfList = ppMfList;
   fListCount = count;
   fListIndex = 0;
   
   new cSaveMinefieldMsg();
   delete[] *returnMsg;
   return true;
}

void   cMfieldMgr::SaveNextMinefield()
{
   if (fListIndex >= fListCount)   {
      fpMineFile->EndLevel();
      delete fpMineFile;
      delete [] fppMfList;
      fppMfList = NULL;
      fpMineFile = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) gpGuiWin->EnableUser();
   }
   else {
      if (fListIndex > 0)
         fpMineFile->EndLevel();
      cMinefield*   pMf = fppMfList[fListIndex++];
      fpMineFile->WriteItem( fKeyMmgrTable[emmgrkwMinefield].key, 
                             fKeyMmgrTable[emmgrkwMinefield].type,
                             (void*) (pMf->GetMfName()) );
      pMf->SaveMinefield(fpMineFile);
         // minefield creates next save message when it is done
   }
}

void cMfieldMgr::LoadMinefields ()
{
   // here is where the minefields from the command line are loaded
   char /*filename, */ msg[199];
   char *cwd = NULL;
   char *tmp = NULL;
   char *mfin = NULL;
   cInFileObj* pMineFile;
   int i, sbuff;

   if ( gCmsMgr.fStartupParam.fNumMineFiles != 0 ){

      sbuff = 100;

      //Need to get directory for minefield input into cwd.
      //We need only do this once.
      if (gCmsMgr.fStartupParam.fDirMfInput[0] != '/') {
	 //The input directory is relative
	 while((tmp = getcwd(NULL, sbuff)) == NULL){
	    sbuff = sbuff + 50;
	 }
         cwd = new char[strlen(tmp) + strlen(gCmsMgr.fStartupParam.fDirMfInput) + 3];
         strcpy (cwd, tmp);
	 free (tmp);
	 strcat (cwd, "/" );
	 strcat (cwd, gCmsMgr.fStartupParam.fDirMfInput );
	 strcat (cwd, "/" );
      }
      else { // if ( fDirMfInput[0] != '/')
	 //The input directory is an absolute path
	 cwd = new char[strlen(gCmsMgr.fStartupParam.fDirMfInput) + 2];
	 strcpy (cwd, gCmsMgr.fStartupParam.fDirMfInput);
	 if (cwd[strlen(cwd)-1] != '/')
	    strcat(cwd, "/");
      }// end else

      //cout << endl << endl << "dir: " << cwd << endl<< endl;
      //Open and read each input file
      for ( i = 0 ; i < gCmsMgr.fStartupParam.fNumMineFiles ; i++ ){
	 // Build Minefile name(s)
	 if (gCmsMgr.fStartupParam.fMineFilenames[i].ffilename[0] != '.' 
	     && gCmsMgr.fStartupParam.fMineFilenames[i].ffilename[0] != '/') {
	    //Minefield file is in the minefield input directory
	    mfin = new char[strlen(cwd) + strlen (gCmsMgr.fStartupParam.fMineFilenames[i].ffilename) + 1];
	    strcpy (mfin, cwd);
	    strcat (mfin, gCmsMgr.fStartupParam.fMineFilenames[i].ffilename);
	 }
	 else {
	    //Minefile has been specified from current directory
	    mfin = new char[strlen (gCmsMgr.fStartupParam.fMineFilenames[i].ffilename) + 1];
	    strcpy (mfin, gCmsMgr.fStartupParam.fMineFilenames[i].ffilename);
	 }
	 //cout << endl << endl << "mf: " << mfin << endl<< endl;

         pMineFile = new cInFileObj;
         if ( !pMineFile->Open(mfin, gftypeMinefields) ) {
            sprintf (msg, "Cannot open %s for input", cwd);
         }
	 else {
	    //We have a valid file, read in the minefield.
	    new cReadMineFileMsg (pMineFile, gCmsMgr.fStartupParam.fMineFilenames[i].fesim ); 
	 }
	 delete [] mfin;
      } // end for
      delete [] cwd;
 
   } // end if (NumMineFiles != 0)

}

void cMfieldMgr::DuplicateMinefields (int count, cMinefield** ppMfList,
                                      double dx, double dy, float daz)
{
   fMove.SetMove(dx, dy, daz);
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstDuplicating;
   fppMfList = ppMfList;
   fListCount = count;
   fListIndex = 0;
   
   new cDuplicateMinefieldMsg();
}

void   cMfieldMgr::DuplicateNextMinefield ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppMfList;
      fppMfList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) gpGuiWin->EnableUser();
   }
   else {
      cMinefield*   pMf = fppMfList[fListIndex++];
      pMf->DuplicateMf(fMove);
      new cDuplicateMinefieldMsg();
   }
}

void   cMfieldMgr::DeleteNextMinefield ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppMfList;
      fppMfList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) gpGuiWin->EnableUser();
   }
   else {
      cMinefield*   pMf = fppMfList[fListIndex++];
      delete pMf;
      new cDeleteMinefieldMsg();
   }
}

void cMfieldMgr::DeleteMinefields (int count, cMinefield** ppMfList)
{
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstDeleting;
   fppMfList = ppMfList;
   fListCount = count;
   fListIndex = 0;
   
   new cDeleteMinefieldMsg();
}

void cMfieldMgr::SaveMfState (int count, cMinefield** ppMfList)
{
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstSavingState;
   fppMfList = ppMfList;
   fListCount = count;
   fListIndex = 0;
   
   new cSaveMfStateMsg();
}

void   cMfieldMgr::SaveNextMfState ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppMfList;
      fppMfList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) gpGuiWin->EnableUser();
   }
   else {
      cMinefield*   pMf = fppMfList[fListIndex++];
      pMf->SaveMfState();
      new cSaveMfStateMsg();
   }
}

void cMfieldMgr::MoveMinefields (int count, cMinefield** ppMfList, 
                                 double dx, double dy, float daz)
{
   fMove.SetMove(dx, dy, daz);
   
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstMoving;
   fppMfList = ppMfList;
   fListCount = count;
   fListIndex = 0;
   
   new cMoveMinefieldMsg();
}

void   cMfieldMgr::MoveNextMinefield ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppMfList;
      fppMfList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) gpGuiWin->EnableUser();
   }
   else {
      cMinefield*   pMf = fppMfList[fListIndex++];
      pMf->MoveMinefield (fMove);
      new cMoveMinefieldMsg();
   }
}

void   cMfieldMgr::DeleteNextComponent ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppCompList;
      fppCompList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) {
         gpGuiWin->EnableUser();
         if ( gpGuiWin->GetCurMinefield() )
            gpGuiWin->GetCurMinefield()->GetCompList()->ShowInfo();
      }
   }
   else {
      cMfComponent*   pComp = (fppCompList[fListIndex++]);
      if ( ! (pComp->GetCompType() == eCompPerim) )
	    delete pComp;
            //** should give a message, cannot delete perim
      new cDeleteComponentMsg();
   }
}

void cMfieldMgr::DeleteComponents (int count, cMfComponent** ppCompList)
{
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstDeleting;
   fppCompList = ppCompList;
   fListCount = count;
   fListIndex = 0;
   
   new cDeleteComponentMsg();
}

void cMfieldMgr::DuplicateComponents (int count, cMfComponent** ppCompList,
                                      double dx, double dy, float daz)
{
   if ( count > 0 ) {
      fMove.SetMove(dx, dy, daz);
      if (gbGui) gpGuiWin->DisableUser();
      feMfMgrState = emgrstDuplicating;
      fppCompList = ppCompList;
      fListCount = count;
      fListIndex = 0;
   
      if (gbGui) gpGuiWin->SetSelection(false, fppCompList[0]->GetMinefield());
      
      new cDuplicateComponentMsg ();
   }
}

void    cMfieldMgr::DuplicateNextComponent ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppCompList;
      fppCompList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) {
         gpGuiWin->EnableUser();
         if ( gpGuiWin->GetCurMinefield() )
            gpGuiWin->GetCurMinefield()->GetCompList()->ShowInfo();
      }
   }
   else {
      cMfComponent*   pC = fppCompList[fListIndex++];
      cMfComponent*   pNew = pC->DuplicateComp(fMove, pC->GetMinefield());
      if (gbGui) gpGuiWin->SetSelection(true, pNew->GetMinefield(), pNew);
      new cDuplicateComponentMsg();
   }
}

void cMfieldMgr::DeleteMine (cMine* pMine)
      // convenience for individual mine
{
  cMine** ppMineList = new cMine*[1];
  ppMineList[0] = pMine;
  DeleteMines(1, ppMineList);
}

void   cMfieldMgr::DeleteNextMine ()
{
   if (fListIndex >= fListCount)   {
      delete [] fppMineList;
      fppMineList = NULL;
      fListCount = 0;
      fListIndex = 0;
      feMfMgrState = emgrstIdle;
      if (gbGui) gpGuiWin->EnableUser();
   }
   else {
      cMine*   pMine = fppMineList[fListIndex++];
//      cMfComponent* pComp = pMine->GetCompParent();
      delete pMine;
//      if (pComp) 
//         ((cMineComp*)pComp)->RedrawComp();
      new cDeleteMineMsg();
   }
}

void cMfieldMgr::DeleteMines (int count, cMine** ppMineList)
{
   if (gbGui) gpGuiWin->DisableUser();
   feMfMgrState = emgrstDeleting;
   fppMineList = ppMineList;
   fListCount = count;
   fListIndex = 0;
   
   new cDeleteMineMsg();
}

/*
void cMfieldMgr::SetMineType (cMineComp* pMineComp, TEntityType ent_type)
{

}

void cMfieldMgr::SetMineDepth (cMineComp* pMineComp, float depth)
{

}

void cMfieldMgr::SetMineSpacing (cMineComp* pMineComp, float spacing)
{
  ((cGridComp*)pMineComp)->SetSpacing (spacing);
}
*/

void cMfieldMgr::CbStoredCoordinates (Widget w, XtPointer , 
				      XmAnyCallbackStruct* )
{
  gpGuiWin->GetCurMinefield()->fbRepos = false;
  new cReadMinefieldMsg (gpGuiWin->GetCurMinefield(), gMfieldMgr.fpInFile);
  XtDestroyWidget (w);
}

void cMfieldMgr::CbRepositionNow (Widget w, XtPointer , 
				  XmAnyCallbackStruct* )
{
  TLocation  tccPt;
  float64    xcenter, ycenter, width, height;
  int32 c;
  
  cMinefield*  pCurMinefield = gpGuiWin->GetCurMinefield();

  pCurMinefield->fbRepos = true;\
/*replaced tactmap_locale with tactmap_get_dimension and tactmap_get_center
  to match modsaf3.0 libtactmap.h AUSTIN 4 Aug 97
*/
tactmap_get_dimension(gpTactmap, &width, &height);
tactmap_get_center(gpTactmap, &xcenter, &ycenter, &c);
/*
  tactmap_locale (gpTactmap, &xcenter, &ycenter, &width, &height);
*/
  tccPt.x = xcenter;
  tccPt.y = ycenter;
  tccPt.z = gpTerrainMgr->GetMapElev (tccPt.x, tccPt.y);
  pCurMinefield->SetCoordTccRef(tccPt, 
				 pCurMinefield->fTccOrientation.psi, 
				 pCurMinefield->fTccOrientation.theta, 
				 pCurMinefield->fTccOrientation.phi);
  new cReadMinefieldMsg (pCurMinefield, gMfieldMgr.fpInFile);
  XtDestroyWidget (w);
}

void   cMfieldMgr::ReadMinefile( cInFileObj* pFile, EBoolean sim)
{
   EFileReturn   returnCode;
   int           returnIndex;
   char*         pReturnVal, buf[199];
      
   feMfMgrState = emgrstReading;
   fpInFile = pFile;
   returnCode = pFile->ReadItem(fKeyMmgrTable, eSizeMmgrTable, &returnIndex, (void**)(&pReturnVal));

   switch (returnCode) {
      case eNextLevel:
            if ( emmgrkwMinefield == returnIndex ) {
	       cMinefield* pField = gMfieldMgr.CreateMinefield( (char*)pReturnVal, esimFileInit );
	       if (gbGui) gpGuiWin->SetSelection(false, pField);

// here it checks for command line loading of minefields  
// if it finds any, the modify coordinates gui doesn't come up 

               if ( gCmsMgr.fStartupParam.fNumMineFiles != 0 ){
		  /*
                  gpGuiWin->GetCurMinefield()->fbRepos = false;
		  gpGuiWin->GetCurMinefield()->fbSim = sim;
                  new cReadMinefieldMsg (gpGuiWin->GetCurMinefield(), gMfieldMgr.fpInFile);
		  */
                  //gCmsMgr.fStartupParam.fNumMineFiles--;
		  pField->fbRepos = false;
		  pField->fbSim = sim;
		  new cReadMinefieldMsg (pField, gMfieldMgr.fpInFile); 
               }
               else if (gbGui) {

	       sprintf (buf, "Do you want to use the stored coordinates\nfor minefield '%s'\nor do you want to reposition it now?",
			pReturnVal);
	      gpGuiWin->PopupMessage (gpGuiWin->GetMainWin(), 
				      cGuiWindow::eDialogQuestion,
				      "Stored Coordinates",
				      "Reposition Now",
				      (XtCallbackProc)CbStoredCoordinates,
				      (XtCallbackProc)CbRepositionNow,
				      NULL, buf);
	       }
            }
         break;

      case eEOF:
	    feMfMgrState = emgrstIdle;
            delete pFile;
            pFile = NULL;
	    if (gCmsMgr.fStartupParam.fNumMineFiles != 0)
	       gCmsMgr.fStartupParam.fNumMineFiles--;
            cout << "Reading is DONE!"<<endl;
            new cSimulateMsg(); //NPAC
         break;

      default: // error returns.  Give the user a message
            Assert(false, kerrReadMinefield, "cMfieldMgr::ReadMinefile");
	    feMfMgrState = emgrstIdle;
            delete pFile;
            pFile = NULL;
	    if (gCmsMgr.fStartupParam.fNumMineFiles != 0)
	       gCmsMgr.fStartupParam.fNumMineFiles--;
         break;
   }
}

EBoolean   cMfieldMgr::ResumeReadMinefile(EBoolean sim)
{
  if (feMfMgrState == emgrstReading) {
    if (gbGui) gpGuiWin->GetCurMinefield()->fbRepos = false;
    new cReadMineFileMsg(fpInFile, sim);
    return true;
  }
  else
    return false;
}

cMinefield*   cMfieldMgr::GetField(int index)
{
   int i;
   cMinefield*   pField;

   if ( index < 1 )
      return NULL;
      
   for ( i = 0, pField = fpFirstMf; (i < index - 1) && (pField != NULL); i++) {
      pField = pField->GetNextMf();
   }
   
   return pField;
}

cMinefield*   cMfieldMgr::GetFieldByEntId(TEntityID& id)
{
   cMinefield*   pField;
      
   for ( pField = fpFirstMf; pField != NULL; pField = pField->GetNextMf()) {
      if ( pField->GetEntityID() == id )
         return pField;
   }
   
   return NULL;
}

//NPAC
void   cMfieldMgr::Simulate() {
  cMinefield* pField = fpFirstMf;
  if(pField!=NULL) {
    int i,n=ZfListCount+1,id=0;
    double start_t,stop_t;
    start_t = clock();
    #pragma parallel local(pField,i) byvalue(n)
    {
       #pragma pfor iterate(i=1;n;1)
       for(i=1;i<n;i++)
       { 
         pField = GetField(i);
         pField->Simulate();
        }
       #pragma critical
         id++;
    }
    stop_t = clock();
    cout << "Starting time = " << (stop_t-start_t) /CLOCKS_PER_SEC
         << " secs over "<< id << " number of threads" << endl;
    new cUpdateMsg();
    cout << "Start Simulation " << endl;
  }
}

//NPAC
void   cMfieldMgr::UpdateGeneral()
{ int proc_count[64],proc_simcount[64],proc_totalmines[64];
  cMinefield* pField = fpFirstMf;
  if(pField!=NULL) {
    double start_t, stop_t;

    cout <<"\n -> Total number of minefields = "<<ZfListCount<<endl;
    int i,n=ZfListCount+1;int id=0; int count,simcount,totalmines;
    start_t = clock();
    #pragma parallel shared(id, proc_count, proc_simcount)
    #pragma local(pField, i,count,simcount,totalmines) byvalue(n)
    { count =0; simcount=0; totalmines =0;
      #pragma pfor iterate(i=1;n;1)
       for(i=1;i<n;i++)
       { count ++;
         pField = GetField(i);
          if ( pField->IsSimulating()) {
             simcount++;
             pField->UpdateMfield();
             totalmines = totalmines+pField->GetTotalMines();
          }
        }
        #pragma critical
        { proc_count[id] = count;
          proc_simcount[id] = simcount;
          proc_totalmines[id] = totalmines;
          id++;
        }
     }
     stop_t = clock();
     for(i=0;i<id;i++)
     { cout <<"ID -> "<<i<<" / #minefields = "<<proc_count[i]<<" / Simulated Minefileds = "<<proc_simcount[i]<<" / Total mines = "<<proc_totalmines[i]<<endl;
     }

     double diff = stop_t-start_t;
     cout << "NPAC :  cMfieldMgr::UpdateGeneral Time = " << diff/CLOCKS_PER_SEC 
          << " secs over " << id << " number of threads"<< endl;
     if(monitorPort)
        monitorPort->sendTo((float)diff,id); //NPAC monitor update!!!

   }
   new cUpdateMsg();
}

void   cMfieldMgr::InDetGeneral(cDetonation* pDetonation)
{
/*   for (cMinefield* pField = fpFirstMf; pField; pField = pField->GetNextMf() )
      if ( pField->IsSimulating() )
         pField->InDetMfield(pDetonation);
*/
//NPAC
      double start_t, stop_t;
  start_t = clock();
   cMinefield* pField = fpFirstMf; int i,n=ZfListCount+1,id=0;
   #pragma parallel local(pField, i) byvalue(n)
   {
     #pragma pfor iterate(i=1;n;1)
      for(i=0;i<n;i++)
      { pField = GetField(i);
        if ( pField->IsSimulating() )
         pField->InDetMfield(pDetonation);
      }
      #pragma critical
            id++;
   }
   stop_t = clock();
   double diff = stop_t-start_t;
   cout << "NPAC  cMfieldMgr::InDetGeneral Time = " <<diff/CLOCKS_PER_SEC
        << " secs over " << id << " number of threads" <<endl;
   if(monitorPort)
        monitorPort->sendTo((float)diff,id); //NPAC monitor update!!!
}

void    cMfieldMgr::MfShutdown()
{
   while ( fpFirstMf )
      delete fpFirstMf;
}

cMfieldMgr::cMfieldMgr() : fpFirstMf(NULL),  fpLastMf(NULL)
{
   fppMfList      = NULL;
   fListCount   = 0;
   ZfListCount   = 0;
   fListIndex   = 0;
   fpMineFile     = NULL;
   feMfMgrState = emgrstIdle;
}

cMfieldMgr::~cMfieldMgr()
{
   cMinefield* pNextField =  NULL;
   cMinefield* pCurField =  fpFirstMf;
   
   while ( pCurField ) {
      pNextField = pCurField->GetNextMf();
      delete pCurField;
      pCurField = pNextField;
   }
}

