/*
$RCSfile: RowEditorCreate.cc,v $
$Revision: 1.3 $
$Author: ksnively $
$Date: 1999/05/10 20:18:26 $
$Log: RowEditorCreate.cc,v $
Revision 1.3  1999/05/10 20:18:26  ksnively
Added a command line option for enabling Communication Capability.

Revision 1.2  1999/03/26 21:17:43  ksnively

Added radio communication capability for Command Detonated Mines.  Included sending
and receiving transmitter and signal PDUs.  All signals have the option of being
processed by the Communication Effects Federate.

Revision 1.1.1.1  1999/02/16 14:54:43  jchappel
Comprehensive Mine Simulator

// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.7  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.6  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.5  1996/05/11  21:35:41  mkappel
// Add mine age.
//
// Revision 1.4  1996/05/11  16:49:17  mkappel
// New mine type selector and fuze options.
//
// Revision 1.3  1996/03/28  22:34:41  mkappel
// Add Mine Display toggle to component editors.
//
// Revision 1.2  1996/03/13  00:12:33  mkappel
// Create scrolling set of toggles for mines.
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: RowEditorCreate.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   RowEditorCreate.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include "MineType.h"
#include "EntityIdEdit.h"

#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/TextF.h>

int    cRowEditorWindow::fNmines = 0;	// estimated number of mines, passed to CbSpacingOK

cCoordList*    cRowEditorWindow::GetRowVertexCoordWidget ()
{
   return fpVertexCoord;
}

cRowEditorWindow::cRowEditorWindow (Widget wParent, EBoolean bController)
   : cMineCompEditorWindow (wParent, bController)
{
   Widget            wLabel, wCoordForm;
   cHelpTextWidget*  pHelp;
    char             *help_text;

   wLabel = 
     XtVaCreateManagedWidget ("Row Emplacement Editor", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 0,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 20,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);


   XtVaSetValues (fpMineCoord->GetCoordForm(),
		  XmNwidth, 100,
		  NULL);

   fpVertexCoord = new cCoordList (fwMainWin, ecltRowVertex);
   wCoordForm = fpVertexCoord->GetCoordForm();
   XtVaSetValues (wCoordForm,
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNtopOffset, 5,
		  XmNbottomAttachment, XmATTACH_WIDGET,
		  XmNbottomWidget, fwActionArea,
		  XmNbottomOffset, 10,
		  XmNrightAttachment, XmATTACH_WIDGET,
		  XmNrightWidget, fpMineCoord->GetCoordForm(),
		  XmNrightOffset, 5,
		  XmNwidth, 100,
		  NULL);


   help_text = cppstrdup ("Add Vertex: left click on map or select first blank\nEmplace Mines: set Spacing and Offset and hit <Enter>.\nDelete Vertex/Mine: right click or use Delete button.\nMove Vertex/Mine: left click-and-drag or use pop-up.\nRotate Row: middle click-and-drag."); 

   pHelp = new cHelpTextWidget (fwMainWin, help_text);
   XtVaSetValues (pHelp->GetHelpForm(),
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_WIDGET,
		  XmNleftWidget, fpMineType->GetTypeForm(),
		  XmNleftOffset, 5,
		  XmNrightAttachment, XmATTACH_WIDGET,
		  XmNrightWidget, wCoordForm,
		  XmNrightOffset, 5,
		  NULL);
   delete[] help_text;
   help_text = NULL;

   if (fpControllerId) {
      XtVaSetValues (fpControllerId->GetIdForm(),
		     XmNtopAttachment, XmATTACH_WIDGET,
		     XmNtopWidget, pHelp->GetHelpForm(),
		     NULL);
   }

   XtVaSetValues (fwMineDispToggle,
		  XmNrightAttachment, XmATTACH_WIDGET,
		  XmNrightWidget, wCoordForm,
		  XmNrightOffset, 10,
		  NULL);

   fwSpacingText = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, fwMainWin, 
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwDepthText,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, fwDepthText,
			      XmNcolumns, 7,
			      NULL);
   XtAddCallback (fwSpacingText, XmNactivateCallback, 
		  (XtCallbackProc)CbSpacing, this);

   wLabel = 
     XtVaCreateManagedWidget ("Mine Spacing (m)", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, fwSpacingText,
			      XmNrightAttachment, XmATTACH_WIDGET,
			      XmNrightWidget, fwSpacingText, 
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);

   fwOffsetText = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, fwMainWin, 
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwSpacingText,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, fwDepthText,
			      XmNcolumns, 7,
			      NULL);
   XtAddCallback (fwOffsetText, XmNactivateCallback, 
		  (XtCallbackProc)CbOffset, this);

   wLabel = 
     XtVaCreateManagedWidget ("Row Offset (m)", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, fwOffsetText,
			      XmNrightAttachment, XmATTACH_WIDGET,
			      XmNrightWidget, fwOffsetText, 
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);
}

cRowEditorWindow::~cRowEditorWindow ()
{
}
