/*
$RCSfile: SaSensor.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:43 $
$Log: SaSensor.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:43  jchappel
Comprehensive Mine Simulator

// Revision 2.5  1998/12/02  16:46:56  jchappel
// added directional sensor
//
// Revision 2.4  1998/11/24  15:11:11  ksnively
// Updated for APLA Exercise
//
// Revision 2.3  1998/04/20  15:10:36  ksnively
// *** empty log message ***
//
// Revision 2.2  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.1  1998/02/18  17:11:56  ksnively
// *** empty log message ***
//
// Revision 1.1  1998/02/17  13:34:42  ksnively
// Initial revision
//
// Revision 2.1  1998/01/28  14:04:11  ksnively
// *** empty log message ***
//
// Revision 1.3  1997/07/18  13:47:07  rcarp
// 2.1b4, mostly changes to fire APLR and hide mine states
//
// Revision 1.2  1997/06/05  15:56:22  rcarp
// updates for b3
//
// Revision 1.1  1997/05/12  22:10:05  rcarp
// Area Denial mods, version 2.1b1
//
$State: Exp $
*/
 
/*******************************************

   file: SaSensor.cc
   
      Copyright 1996 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   SaSensor.cc
   
      The sensor and gateway implementation.
      
      
   
   Created by Richard Carpenter
      
*********************************************/

#include "SaSensor.h"
#include "Message.h"
#include "CmsControl.h"
#include "MapControls.h"
#include <math.h>
      // standalone sensors
TEntityType   cSaSensor::fsSaSensorEntityType(ekindMunition,
                                             efunnydomainSaSensor,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cSaSensor::fsSaSensorName = "Standalone Sensor";
char*            cSaSensor::fsSaSensorAltName = "Sens";
int     cSaSensor::fsDefaultSensType =  TFuzeParam:: esenstypePerDet;
cDbMine_cSaSensor      cSaSensor::fsDbSaSensor( cSaSensor::fsDefaultSensType,
                                                cSaSensor::fsSaSensorEntityType,
                                                cMineDBMgr::emdblevDomain, 
                                                cSaSensor::fsSaSensorName,
                                                cSaSensor::fsSaSensorAltName); 

TEntityType   cSensDetector::fsSensDetectorEntityType(ekindMunition,
                                             efunnydomainSaSensor,
                                kMatchAllShort,
                                             emuncatFixed,
                                             esensscatDetecting,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cSensDetector::fsSensDetectorName = "Detecting Sensor";
char*            cSensDetector::fsSensDetectorAltName = "DSen";
int     cSensDetector::fsDefaultSensDetectorType =  TFuzeParam:: esenstypePerDet;
cDbMine_cSensDetector      cSensDetector::fsDbSensDetector( cSensDetector::fsDefaultSensDetectorType,
                                                cSensDetector::fsSensDetectorEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cSensDetector::fsSensDetectorName,
                                                cSensDetector::fsSensDetectorAltName); 

TEntityType   cSensTracker::fsSensTrackerEntityType(ekindMunition,
                                             efunnydomainSaSensor,
                                kMatchAllShort,
                                             emuncatFixed,
                                             esensscatTracking,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cSensTracker::fsSensTrackerName = "Tracking Sensor";
char*            cSensTracker::fsSensTrackerAltName = "TSen";
int     cSensTracker::fsDefaultSensTrackerType =  TFuzeParam:: esenstypePerTrack;
cDbMine_cSensTracker      cSensTracker::fsDbSensTracker( cSensTracker::fsDefaultSensTrackerType,
                                                cSensTracker::fsSensTrackerEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cSensTracker::fsSensTrackerName,
                                                cSensTracker::fsSensTrackerAltName); 

      // standalone controllers
TEntityType   cSaController::fsSaControllerEntityType(ekindMunition,
                                             efunnydomainSaController,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cSaController::fsSaControllerName = "Standalone Control";
char*            cSaController::fsSaControllerAltName = "Cntrl";
int     cSaController::fsDefaultContlType =  TFuzeParam:: econtypeAuto;
cDbMine_cSaController      cSaController::fsDbSaController( cSaController::fsDefaultContlType,
                                                cSaController::fsSaControllerEntityType,
                                                cMineDBMgr::emdblevDomain, 
                                                cSaController::fsSaControllerName,
                                                cSaController::fsSaControllerAltName); 


const TKeywordEntry TSensParamDumb::fKeyDmbSensTable[edmbsensSizeTable] = {
   "sensType",        ekeyString,
   "sensRangeList",   ekeyFloatList,
   "detProbList",     ekeyFloatList
};

const TKeywordEntry TSensParamTrack::fKeyTrkSensTable[etrksensSizeTable] = {
   "sensType",        ekeyString,
   "rangeError",      ekeyFloat,
   "azimuthError",    ekeyFloat
};

void        cDispVehicle::UpdateDisplay()
{
   if ( fpDisplay)
      fpDisplay->UpdateVehicleDisp();
}

cDispVehicle::cDispVehicle( const cVehicle& origVeh )
{
      // copy/set cEntity needed fields
   fGeoLocation    =    origVeh.GetGeoLoc();
   fGeoOrient      =    origVeh.GetGeoOrient();
   fGeoRotation    =    origVeh.GetGeoRot();
   fTccLocation    =    origVeh.GetTccLoc();
   fTccOrientation =    origVeh.GetTccOrient();
   fTccRotation    =    origVeh.GetTccRot();

      // copy/set cVehicle needed fields
   fpDbVeh    =    origVeh.fpDbVeh;
   fTime    =    origVeh.fTime;
   fForce = eOther;
   fCmsType    =    origVeh.fCmsType;
   fAltType    =    origVeh.fAltType;
   fVelocity    =    origVeh.fVelocity;
   fGeoVelocity    =    origVeh.fGeoVelocity;
   fEntVelocity    =    origVeh.fEntVelocity;
   fTurretAzimuth    =    origVeh.fTurretAzimuth;
   fAppearance    =    eAprStandard;
   
   fpDisplay = new cVehDisp(this, false );
   
   //cout << "creating new disp veh, and its disp" << endl;

//   fpVehPartList = NULL;   // may have to copy the array,
                           // but then will have to update
}

cDispVehicle::~cDispVehicle()
{
   if (fpDisplay)
      delete fpDisplay;
}

void        cLinkVehTrkDisp::DisplayVeh( EBoolean disp )
{
   fpDispVehicle->ShowVehicle(disp);
   if (disp)
      fpDispVehicle->UpdateDisplay();
}



void   cLinkVehTrkDisp::UpdateTrkRec()
{
      // first update to error values
   cLinkVehErr::UpdateTrkRec();
   
   TLocation tccLoc;
   TOrient orient;
   
      // set the tcc values, with the errors
   if ( fbTgtVelRef ) {
      orient.SetOrient(fAzRot, 0.0, kPi);
      TRotation rot(orient);
      tccLoc = ( rot | fLocation ) + fpRefEnt->GetTccLoc();
   }
   
   else {
      orient = fpVehicle->GetTccOrient();
      fpRefEnt->LocalToTccLoc(fLocation, &tccLoc);
   }
   
//   fpDispVehicle->SetCoordTccRef(fpVehicle->GetTccLoc(), fpVehicle->GetTccOrient().psi, 
//               fpVehicle->GetTccOrient().theta, fpVehicle->GetTccOrient().phi);
   fpDispVehicle->SetCoordTccRef(tccLoc, orient.psi, orient.theta, orient.phi);
}

cLinkVehTrkDisp::cLinkVehTrkDisp(cLinkVehTrkDisp* pTrkVeh, cLinkVeh* pPrev) :
   cLinkVehErr(pTrkVeh, pPrev)
{
   //cout << "copying track" << endl;
   fpDispVehicle   = pTrkVeh->fpDispVehicle;
   pTrkVeh->fpDispVehicle   = NULL;
}

cLinkVehTrkDisp::cLinkVehTrkDisp(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev, cEntity* pRefEnt, 
      EBoolean refInvert, int tgtType, float rangeErr, float velErr, float azErr) :
    cLinkVehErr(pProxVeh, pPrev, pRefEnt, refInvert, tgtType, rangeErr, velErr, azErr)
{
   //cout << "creating new track"<<endl;
   fpDispVehicle   = new cDispVehicle( *(pProxVeh->GetVehicle()) );
   fpDispVehicle->UpdateDisplay();
}

cLinkVehTrkDisp::~cLinkVehTrkDisp()
{
   if (fpDispVehicle)
      delete fpDispVehicle;
}

EBoolean     TSensParamDumb::DetectsTgt( float range )
{
   for ( int i = 0; i < fNumBands; i++ ) {
      if ( range <= fpRange[i] ) {
         if ( UniformRand() <= fpDetectProb[i] ) 
            return true;
         else
            return false;
      }
   }
   return false;
}
 
EBoolean     TSensParamDumb::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TItemList *   list;
   int           i;
   
   returnCode = infile.ParseItem(fKeyDmbSensTable, edmbsensSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case edmbsenskwFuzeTypeName: {
                     // read name, look up index
   					fTypeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( (fTypeNameIndex != esenstypePerDet) &&
                      (fTypeNameIndex != esenstypeVehDet) &&
                      (fTypeNameIndex != esenstypePerVehDet) &&
                      (fTypeNameIndex != esenstypeTripWire) &&
                      (fTypeNameIndex != esenstypeDirDet) &&
		      (fTypeNameIndex != esenstypeCommandCamDir) &&
		      (fTypeNameIndex != esenstypeAutoCmdCamDir) &&
		      (fTypeNameIndex != esenstypeDirPVTrack)  ) 
                  {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable sens type for standalone sensor"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case edmbsenskwRange:
                  list = (TItemList *)returnVal;
                  if ( fpRange ) delete fpRange;
                  
                        // if list sizes are unequal,
                        // num bands is the smaller list size
                  if ( (fNumBands > list->count) ||
                       (fNumBands == 0) )
                     fNumBands = list->count;
                  fpRange = new float[list->count];
                     
                  for (i = 0; i < list->count; i++) {
                     fpRange[i] = *((float *)(list->ptrItems) + i);
                  }
                  
                  fTrackRange = fpRange[list->count - 1] + 30.0;
                  
                  delete[] (float *)(list->ptrItems);
                  delete list;
                  list = NULL;
                  returnVal = NULL;
 					break;

				case edmbsenskwProb:
                  list = (TItemList *)returnVal;
                  if ( fpDetectProb ) delete fpDetectProb;
                  
                        // if list sizes are unequal,
                        // num bands is the smaller list size
                  if ( (fNumBands > list->count) ||
                       (fNumBands == 0) )
                     fNumBands = list->count;
                  
                  fpDetectProb = new float[list->count];
                     
                  for (i = 0; i < list->count; i++) {
                     fpDetectProb[i] = *((float *)(list->ptrItems) + i);
                  }
                  delete[] (float *)(list->ptrItems);
                  delete list;
                  list = NULL;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
     cout<< "passing keyword up to parent fuze class"<<endl;
         return TFuzeParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TSensParamDumb::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

cControl*     TSensParamDumb::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
   
      case esenstypePerDet:
      case esenstypeVehDet:
      case esenstypePerVehDet:
      case esenstypeTripWire:
      case esenstypeDirDet:
            pNewControl = new cDetSensorControl(pMine, fFireOffset, fuzeOpt, this);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TSensParamDumb::CreateControl");
   }
   
   return pNewControl;
}

TFuzeParam*   TSensParamDumb::CloneFuzeParam()   
{
   TFuzeParam* pClone = new TSensParamDumb(*this);
   pClone->fpOptSelector = fpOptSelector->CloneFuzeOptionSelector(pClone);
   
   if ( fpRange ) {
      fpRange = new float[fNumBands];
      fpDetectProb = new float[fNumBands];
      for ( int i = 0; i < fNumBands; i++ ) {
         fpRange[i] = ((TSensParamDumb*)pClone)->fpRange[i];
         fpDetectProb[i] = ((TSensParamDumb*)pClone)->fpDetectProb[i];
      }
   }
   
   return pClone;
}

EBoolean   TSensParamDumb::IsDirectional(TFlatEntCoord* pMinMax)
{
   if ( fTypeNameIndex == esenstypeDirDet) {
      if ( pMinMax ) {
         pMinMax->x = 0.0;
         pMinMax->y = fMaxRange;
      }
      
      return true;
   }
   
   return false;
}

EBoolean   TSensParamDumb::IsDirectional()
{
   if ( fTypeNameIndex == esenstypeDirDet) {
      return true;
   }
   
   return false;
}

TSensParamDumb::TSensParamDumb(int sensType) :
   TFuzeParam(sensType)
{
   fNumBands     =  0;
   fpRange       = NULL;
   fpDetectProb  = NULL;
   
   fDetonResult      = edetNone;

   
   fFireOffset.x = fMaxRange     =  100.0;
   fFireOffset.y = fCrossRange  =  0.0;
   fTrackRange   = fMaxRange;
   
}

TSensParamDumb::~TSensParamDumb()
{
}

EBoolean     TSensParamTrack::TracksTgt( float range )
{
   for ( int i = 0; i < fNumBands; i++ ) {
      if ( range <= fpRange[i] ) {
         if ( UniformRand() <= fpDetectProb[i] ) {
            return true;
         }
         else
            return false;
      }
   }
   return false;
}

 
EBoolean     TSensParamTrack::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeyTrkSensTable, etrksensSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case etrksenskwFuzeTypeName: {
                     // read name, look up index
   					fTypeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( (fTypeNameIndex != esenstypePerTrack) &&
                      (fTypeNameIndex != esenstypeVehTrack) &&
                      (fTypeNameIndex != esenstypePerVehTrack) ) 
                  {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable fuze type for wide area mine"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case etrksenskwRangeErr:
                  fRangeErr = * (float *)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case etrksenskwAzErr:
                  fAzError = * (float *)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TSensParamDumb::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TSensParamTrack::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

cControl*     TSensParamTrack::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
   
      case esenstypePerTrack:
      case esenstypeVehTrack:
      case esenstypePerVehTrack:
      case esenstypeDirPVTrack:
            pNewControl = new cTrkSensorControl(pMine, fFireOffset, fuzeOpt, this);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TSensParamTrack::CreateControl");
   }
   
   return pNewControl;
}

TFuzeParam*   TSensParamTrack::CloneFuzeParam()   
{
   TFuzeParam* pClone = new TSensParamTrack(*this);
   pClone->fpOptSelector = fpOptSelector->CloneFuzeOptionSelector(pClone);
   
   if ( fpRange ) {
      fpRange = new float[fNumBands];
      fpDetectProb = new float[fNumBands];
      for ( int i = 0; i < fNumBands; i++ ) {
         fpRange[i] = ((TSensParamDumb*)pClone)->fpRange[i];
         fpDetectProb[i] = ((TSensParamDumb*)pClone)->fpDetectProb[i];
      }
   }
   
   return pClone;
}
EBoolean   TSensParamTrack::IsDirectional(TFlatEntCoord* pMinMax)
{
   if ( fTypeNameIndex == esenstypeDirPVTrack) {
      if ( pMinMax ) {
         pMinMax->x = 0.0;
         pMinMax->y = fMaxRange;
      }
      
      return true;
   }
   
   return false;
}

EBoolean   TSensParamTrack::IsDirectional()
{
   if ( fTypeNameIndex == esenstypeDirPVTrack) {
      return true;
   }
   
   return false;
}

TSensParamTrack::TSensParamTrack(int sensType) :
   TSensParamDumb(sensType)
{
   fRangeErr     =  0.2;
   fAzError      =  0.1;
   
   fFireOffset.x = fMaxRange     =  100.0;
   fFireOffset.y = fCrossRange  =  0.0;
   fTrackRange   = fMaxRange;

}

TSensParamTrack::~TSensParamTrack()
{
}

cDbMine*  cDbMine_cSaSensor::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cSaSensor* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cSaSensor( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cSaSensor();
   return pNewDbMine;
}

cSimMine*     cDbMine_cSaSensor::CreateMine(cPlanMine* pPlan)
{
   Assert(false, kerrValidityCheck, "cDbMine_cSaSensor::CreateMine");
   return (cSimMine*) new cSaSensor(pPlan);
}

cDbMine_cSaSensor::cDbMine_cSaSensor() :
   cDbMine()
{
}

cDbMine_cSaSensor::cDbMine_cSaSensor(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cSaSensor::cDbMine_cSaSensor(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cSaSensor::~cDbMine_cSaSensor()
{
}

cDbMine*  cDbMine_cSensDetector::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cSensDetector* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cSensDetector( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cSensDetector();
   return pNewDbMine;
}

cSimMine*     cDbMine_cSensDetector::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cSensDetector(pPlan);
}

cDbMine_cSensDetector::cDbMine_cSensDetector() :
   cDbMine()
{
}

cDbMine_cSensDetector::cDbMine_cSensDetector(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cSensDetector::cDbMine_cSensDetector(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cSensDetector::~cDbMine_cSensDetector()
{
}

cDbMine*  cDbMine_cSensTracker::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cSensTracker* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cSensTracker( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cSensTracker();
   return pNewDbMine;
}

cSimMine*     cDbMine_cSensTracker::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cSensTracker(pPlan);
}

cDbMine_cSensTracker::cDbMine_cSensTracker() :
   cDbMine()
{
}

cDbMine_cSensTracker::cDbMine_cSensTracker(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cSensTracker::cDbMine_cSensTracker(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cSensTracker::~cDbMine_cSensTracker()
{
}

cDbMine*  cDbMine_cSaController::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cSaController* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cSaController( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cSaController();
   return pNewDbMine;
}

cSimMine*     cDbMine_cSaController::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cSaController(pPlan);
}

cDbMine_cSaController::cDbMine_cSaController() :
   cDbMine()
{
}

cDbMine_cSaController::cDbMine_cSaController(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cSaController::cDbMine_cSaController(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cSaController::~cDbMine_cSaController()
{
}

cLinkVehGeom*   cSensTrkTracker::CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev)
{
   cSensTracker* pSens = (cSensTracker*)fpRefEnt;

   float rangeErr = 1.0 + ( (2.0 * UniformRand()) - 1.0) * pSens->GetRangeErr();
   float azErr = ( (2.0 * UniformRand()) - 1.0) * pSens->GetAzErr();


   cLinkVehTrkDisp* pNewLink = new cLinkVehTrkDisp(pProxLink, pPrev, pSens, fbUseVehCoord, 
                           cMine::etrkYes, rangeErr, 0.0, azErr);

   return pNewLink;
}

cLinkVehGeom*   cSensTrkTracker::CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev)
{
   cLinkVehTrkDisp* pNewLink = new cLinkVehTrkDisp( (cLinkVehTrkDisp*)pOldTrkLink, pPrev);

   return pNewLink;
}

cSensTrkTracker::~cSensTrkTracker()
{ }


void      cDetSensorControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* /*pEmissList*/, TTimeSim* /*pWaitTime*/)
      // just sets state on detection/ no detection
{
   ESensInternalState state = (ESensInternalState) fpMine->GetInternalState();

   if (state > esisOn && fpSensParam->fTypeNameIndex == TFuzeParam::esenstypeTripWire) {
      // We have a tripwire fuzed sensor.
      // The trip wire has been activated and cannot be activated again.
      state = esisFailed;
      fpMine->StateChange((EMineInternalState)state);
   }
   
   if ( state >= esisOn ) {
      ESensInternalState newstate = esisOn;
      
         // go through each tgt in track list
      for (T2link* pLn  = pVehList->GetFirstLink();
                   pLn != pVehList->GetNULLLink();
                   pLn  = pVehList->GetNextLink() ) {
         if ( fpSensParam->DetectsTgt( ((cLinkVehGeom*)pLn)->GetHorzRange() ) ) {
            if(/* (fpSensParam->IsDirectional() ) && */
                (InFootprint((cLinkVehGeom*)pLn) ) ){
               newstate = esisDetecting;
               break;
	    }
         } 
      }
   
      if ( state != newstate )
          fpMine->StateChange( (EMineInternalState)newstate );
   }
   
//   if ( gfInteract )
//      fpMine->UpdateInteractions();
}
 
EBoolean  cDetSensorControl::InFootprint(cLinkVehGeom* pTgtLink)
{
   if ( pTgtLink->GetHorzRange() <= fFireOffset.x ){
      if ( (fAzimuth > 0.0) && ( fabs(pTgtLink->GetAzimuth()) > fAzimuth) ) {
         return false;
      }
      else
	 return true;
   }
   else
      return false;
}
           
void      cDetSensorControl::CalcGeom(const cLinkVehGeom* /*pLink*/, TAttackGeom* /*pGeom*/)
{
}

EBoolean   cDetSensorControl::NeedsPrUpdate( EMineInternalState /*minestate*/)
{ 
   return false;
}


EBoolean   cDetSensorControl::NeedsUpdate( EMineInternalState minestate)
{ 
   return minestate >= esisLowestUpdate ? true : false;
}

cDetSensorControl::cDetSensorControl(cSimMine* pSensor, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt, 
                                   TSensParamDumb* pSensParam ) :
            cMineControl(pSensor, offset, fuzeOpt)
{
   fpSensParam = pSensParam;
}
                             
cDetSensorControl::~cDetSensorControl()
{
}

void      cTrkSensorControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* /*pEmissList*/, TTimeSim* /*pWaitTime*/)
      // when coming into range, start track at first detect
      // On leaving, go until first dropout
{
   ESensInternalState state = (ESensInternalState) fpMine->GetInternalState();
   
   if ( state >= esisOn ) {
      ESensInternalState newstate = esisOn;
      
         // go through each tgt in track list
      for (T2link* pLn  = pVehList->GetFirstLink();
                   pLn != pVehList->GetNULLLink();
                   pLn  = pVehList->GetNextLink() ) {
         if ( fpSensParam->TracksTgt( ((cLinkVehGeom*)pLn)->GetHorzRange() ) ) {
            newstate = esisTracking;
            ((cLinkVehTrkDisp*)pLn)->DisplayVeh(true);
         }
         else {
            ((cLinkVehTrkDisp*)pLn)->DisplayVeh(false);
         }
      }
   
      if ( state != newstate )
          fpMine->StateChange( (EMineInternalState)newstate );
   }
   
//   if ( gfInteract )
 //     fpMine->UpdateInteractions();
}
             
void      cTrkSensorControl::CalcGeom(const cLinkVehGeom* /*pLink*/, TAttackGeom* /*pGeom*/)
{
}

EBoolean   cTrkSensorControl::NeedsPrUpdate( EMineInternalState /*minestate*/)
{ 
   return false;
}


EBoolean   cTrkSensorControl::NeedsUpdate( EMineInternalState minestate)
{ 
   return minestate >= esisLowestUpdate ? true : false;
}

cTrkSensorControl::cTrkSensorControl(cSimMine* pSensor, TFlatEntCoord offset, const TFuzeOpt& fuzeOpt, 
                                   TSensParamTrack* pSensParam ) :
            cMineControl(pSensor, offset, fuzeOpt)
{
   fpSensParam = pSensParam;
}
                             
cTrkSensorControl::~cTrkSensorControl()
{
}

EMineInternalState    cSaSensor::GetDisplayState()
{
   return GetInternalState();
}

cSaSensor::cSaSensor( cPlanMine* pPlan ) :
   cSimMine( pPlan)
{
}

cSaSensor::~cSaSensor()
{ }

cSensDetector::cSensDetector( cPlanMine* pPlan ) :
   cSaSensor( pPlan)
{
}

cSensDetector::~cSensDetector()
{ }

cSensTracker::cSensTracker( cPlanMine* pPlan ) :
   cSaSensor(pPlan)
{
}

cSensTracker::~cSensTracker()
{ }

void      cSaController::UpdateMine()
{
   if ( feIntState >= emisLowestUpdate)
      fpUpdateMsg = new cUpdateMineMsg(this);
   else {
      delete fpDetonVuln;
      fpDetonVuln = NULL;
   }
}

cSaController::cSaController(  cPlanMine* pPlan  ) :
   cSimMine( pPlan )
{
}

cSaController::~cSaController()
{ }
