/*
$RCSfile: AtMine.cc,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/05/10 20:18:20 $
$Log: AtMine.cc,v $
Revision 1.2  1999/05/10 20:18:20  ksnively
Added a command line option for enabling Communication Capability.

Revision 1.1.1.1  1999/02/16 14:54:27  jchappel
Comprehensive Mine Simulator

// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.7  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.6  1996/06/19  19:05:45  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.5  1996/05/08  00:46:50  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.5  1996/05/08  00:46:50  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.4  1996/02/16  21:41:55  rcarp
// revisions for getting mine wide area and directional range
//
// Revision 1.3  1996/02/05  14:31:34  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: AtMine.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   AtMine.cc
   
      The AT mine implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "AtMine.h"
#include "EntityDB.h"
#include "AnalData.h"

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// ************************************
//   messages
// ************************************

// ************************************
//   components
// ************************************

int    cLandTracker::GetTrackType(const TEntityType& type)
{
   if ( (type.kind == ekindPlatform) && 
        (type.domain == eplatdomLand) &&
        fVehTracker )
      return cMine::etrkYes;
      
   else if( (type.kind == ekindLifeform) && 
            (type.domain == eplatdomLand) &&
            fLifeTracker )
      return cMine::etrkYes;
   
   else
      return cMine::etrkNot;
}

cLandTracker::cLandTracker(cEntity* pRefEnt, float range, EBoolean useVelCoord,
             EBoolean tracksVeh, EBoolean tracksPers) : 
                        cTracker(pRefEnt, range, useVelCoord) 
{
   fLifeTracker = tracksPers;
   fVehTracker = tracksVeh;
}

cLandTracker::~cLandTracker()
{ }

EBoolean  cTrWidthControl::FireGeometry(const cLinkVehGeom* pLink, TAttackGeom* pGeom)
{
   if ( (pGeom->fTgtOffset.x == 0.0) && 
        (pGeom->fTgtOffset.y == 0.0)   ) {
      
      const  TVehDimensions& vehDim = pLink->GetVehicle()->GetDimensions();
      const  TEntCoord&      vehLoc = pLink->GetLocation();
      
      if ( ( fabs(vehLoc.y) <= (vehDim.trackYmax + fFireOffset.y) ) &&
           ( fabs(vehLoc.y) >= (vehDim.trackYmin - fFireOffset.y) ) )
         return true;
   }
   return false;
}

EBoolean  cTrWidthControl::FireGeometry(const cMineLinkEmiss* /* pLink*/, TAttackGeom* pGeom)
{
   if ( (pGeom->fTgtOffset.x == 0.0) && 
        (pGeom->fTgtOffset.y == 0.0)   ) {
      return true;
   }
   return false;
}

cTrWidthControl::~cTrWidthControl()
{ }

cFullWidthControl::~cFullWidthControl()
{ }

// ************************************
//   mines 
// ************************************
cAtMine::cAtMine(cMineComp* pParent, float radius) :
      cSimMine(pParent, radius)
{
}

cAtMine::cAtMine( cPlanMine* pPlan ) :
      cSimMine(  pPlan )
{
}

cAtMine::~cAtMine()
{ }

cAtConv::cAtConv(cMineComp* pParent, float radius) :
      cAtMine(pParent, radius)
{
}

cAtConv::cAtConv( cPlanMine* pPlan ) :
      cAtMine( pPlan )
{
}

cAtConv::~cAtConv()
{ }

cAtScatterable::cAtScatterable(cMineComp* pParent, float radius) :
      cAtMine(pParent, radius)
{
}

cAtScatterable::cAtScatterable( cPlanMine* pPlan ) :
      cAtMine( pPlan )
{
}

cAtScatterable::~cAtScatterable()
{ }
