/*
$RCSfile: AtWideArea.cc,v $
$Revision: 1.3 $
$Author: ksnively $
$Date: 1999/05/18 19:33:13 $
$Log: AtWideArea.cc,v $
Revision 1.3  1999/05/18 19:33:13  ksnively
Ported CMS to compile under IRIX 6.5
Altered make files to allow cms to be compiled under various OS simulataneously
Removed the Startup GUI

Revision 1.2  1999/04/29 19:34:02  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:28  jchappel
Comprehensive Mine Simulator

// Revision 2.3  1998/05/08  15:42:04  jchappel
// almost made correction to ORSMC damage.
//
// Revision 2.2  1998/02/18  20:10:01  ksnively
// Ap mine replacement bug fixes.
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1998/01/09  18:21:47  ksnively
// Integrated latest SAIC release cms3.0b2.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.5  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.4  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.4  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.2  1997/03/20  19:14:41  ksnively
// added emission tracking, updated to CMS2.0B6
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.14  1996/08/21  21:58:20  rcarp
// CP and QRP file output, other updates
//
// Revision 1.13  1996/08/09  14:33:27  rcarp
// changed to support los check by terrain manager, and its use by offroute mines
//
// Revision 1.12  1996/08/07  14:19:30  rcarp
// initial revisions for CP protocol
//
// Revision 1.11  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.10  1996/06/19  19:08:07  rcarp
// revisions for mine vulnerability, version 1.5b2
//
// Revision 1.9  1996/05/15  17:18:44  rcarp
// refinements and fixes for 1.5 delivery
//
// Revision 1.8  1996/05/15  00:11:09  rcarp
// mostly wam and offroute fixes, also deleting selection
//
// Revision 1.7  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.6  1996/05/11  20:29:26  rcarp
// bug fixes
//
// Revision 1.5  1996/05/08  00:47:57  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.4  1996/02/16  21:41:55  rcarp
// revisions for getting mine wide area and directional range
//
// Revision 1.2  1996/02/05  14:32:43  rcarp
// Updates and bug fixes of release version 1.1b5
//
$State: Exp $
*/
 
/*******************************************

   file: AtWideArea.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   AtWideArea.cc
   
      The wide area AT mine implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "AtWideArea.h"
#include "NetMgr.h"
#include "CmsControl.h"
#include "MapGeometry.h"
#include "Message.h"
#include "CmsPdu.h"
#include "DisPdu203.h"
#include "AnalData.h"
#include "Vehicle.h"
#include "Display.h"

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// ************************************
//   static variables
// ************************************

char*   cProjectile::fsStateText[eprojisNumStates] = {
   "None",
   "Failed",
   "HitGround",
   "MissedTarget",
   "HitTarget",
   "Flying"
};

const TKeywordEntry TFuzeParamOffRoute::fKeyOrFuzeTable[eorfuzeSizeTable] = {
   "fuzeType",     ekeyQuotedString,
   "fireOffset",   ekey2DPoint,
   "tgtRangeGate", ekey2DPoint,
   "tgtSpeedGate", ekey2DPoint,
   "detectProb",   ekeyDouble,
   "hitProb",      ekeyDouble,
   "missileSpeed", ekeyDouble
};

const TKeywordEntry TFuzeParamWamSublet::fKeySubletFuzeTable[esubfuzeSizeTable] = {
   "fuzeType",          ekeyQuotedString,
   "coneAngle",         ekeyDouble,
   "probScanDetect",    ekeyDouble,
   "probHitTgt",        ekeyDouble
};

const TKeywordEntry TFuzeParamWideArea::fKeyWaFuzeTable[ewafuzeSizeTable] = {
   "fuzeType",          ekeyQuotedString,
   "selfDefenseRange",  ekeyDouble,
   "rangeErrMax",       ekeyDouble,
   "velErrMax",         ekeyDouble,
   "azimuthErrMax",     ekeyDouble,
   "probTypeErr",       ekeyDouble,
   "typeErrMultiplier", ekeyDouble,
   "fireWindow",        ekeyDouble
};

const TKeywordEntry TMineParamWam::fKeyMineParamWamTable[eSizeMineParamWamTable] = {
   "subletType",        ekeyQuotedString
};

TTimeSim   cProjectile::fsDamageCheckDelay = 500;      // milliiseconds

const   TTimeSim   cWamSublet::fsUpdateRate  = 100;

double   cOffRouteControl::fsHeight = 1.0;
double   cOffRouteControl::fsMinVis = 0.1;


// ************************************
//   messages
// ************************************
void   cFireProjectileMsg::SendMsg()
{
   gpNetMgr->GetPduTrans()->SendMineFire(fFireData);
}

cFireProjectileMsg::~cFireProjectileMsg()
{ } 

void   cUpdateProjectileMsg::SendMsg()
{
   fpProjectile->UpdateVehicle();
}

cUpdateProjectileMsg::~cUpdateProjectileMsg()
{ } 

void   cProjFireMunitionMsg::SendMsg()
{
   gpNetMgr->GetPduTrans()->SendProjFire(fpProjectile);
}

cProjFireMunitionMsg::~cProjFireMunitionMsg()
{ } 

void   cProjDetonationMsg::SendMsg()
{
   gpNetMgr->GetPduTrans()->SendProjDeton(fpProjectile);
   
   if ( gbGui && (fpProjectile->GetDetResult() == edetEntityImpact) ) {
      TLocation detLoc;
      gpTerrainMgr->SimToMapLoc( fpProjectile->GetDetLoc(), &detLoc);
      new cMineExplosion( detLoc, fpProjectile->GetTccLoc(), 
                          fBlastSize, fDelay );
   }
}

cProjDetonationMsg::~cProjDetonationMsg()
{ } 

// ************************************
//   database
// ************************************
EBoolean     TFuzeParamOffRoute::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeyOrFuzeTable, eorfuzeSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case eorfuzekwFuzeTypeName: {
                     // read name, look up index
   					int typeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( typeNameIndex != efuzetypeAcousticTrigger ) {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable fuze type for off route mine"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case eorfuzekwRangeGate:
				case eorfuzekwFireOffset:
                  fMinRange =  ((T2Dsingle*)returnVal)->x;
                  fMaxRange =  ((T2Dsingle*)returnVal)->y;
                  fFireOffset.x = fMinRange;
                  fFireOffset.y = fMaxRange;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eorfuzekwSpeedGate:
                  fMinTgtSpeed =  ((T2Dsingle*)returnVal)->x;
                  fMaxTgtSpeed =  ((T2Dsingle*)returnVal)->y;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eorfuzekwDetectProb:
				   //fProbDud = *(double*)returnVal;
		  fDetectProb = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eorfuzekwHitProb:
				   //fProbDeton = *(double*)returnVal;
		  fHitProb = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eorfuzekwMissileSpeed:
                  fArmTimeHrs = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;
               
				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TFuzeParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TFuzeParamOffRoute::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

cControl*     TFuzeParamOffRoute::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
      case efuzetypeAcousticTrigger:
            pNewControl = new cOffRouteControl(pMine, fFireOffset, fuzeOpt, fDetectProb,
                              fHitProb, fMinTgtSpeed, fMaxTgtSpeed);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TFuzeParamOffRoute::CreateControl");
   }
   
   return pNewControl;
}

EBoolean   TFuzeParamOffRoute::IsDirectional(TFlatEntCoord* pMinMax)
{
   if ( pMinMax ) {
      pMinMax->x = fMinRange;
      pMinMax->y = fMaxRange;
   }
   
   return true;
}

TFuzeParamOffRoute::TFuzeParamOffRoute() :
   TFuzeParam(efuzetypeAcousticTrigger)
{
   fFireOffset.x = fMinRange      =  20.0;   // meters
   fFireOffset.y = fMaxRange      = 100.0;
   
   fDetonResult      = edetEntityImpact;

   fMinTgtSpeed   = 0.833;   // m/s
   fMaxTgtSpeed   = 13.89;
   fDetectProb    = 0.95;
   fHitProb       = 0.8;
   fSpeed         = 100.0;   // m/sec, both launch and impact

   fTrackRange       = 200.0;
}

TFuzeParamOffRoute::~TFuzeParamOffRoute()
{
}

EBoolean     TFuzeParamWideArea::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeyWaFuzeTable, ewafuzeSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case ewafuzekwFuzeTypeName: {
                     // read name, look up index
   					int typeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( typeNameIndex != efuzetypeAcousticTracker ) {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable fuze type for wide area mine"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case ewafuzekwSelfDefRange:
                  fSelfDefenseRange = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case ewafuzekwRangeErrMax:
                  fRangeErrMax = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case ewafuzekwVelErrMax:
                  fVelocityErrMax = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case ewafuzekwAzErrMax:
                  fAzimuthErrMax = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case ewafuzekwProbTypeErr:
                  fTypeErrProb = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case ewafuzekwTypeErrMult:
                  fTypeErrMult = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case ewafuzekwFireWindow:
                  fFireWindow = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TFuzeParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TFuzeParamWideArea::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

cControl*     TFuzeParamWideArea::CreateControl(cSimMine* pMine, const TFuzeOpt& fuzeOpt )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
   
      case efuzetypeAcousticTracker:
            pNewControl = new cWamControl(pMine, fFireOffset, fuzeOpt);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TFuzeParamWideArea::CreateControl");
   }
   
   return pNewControl;
}

EBoolean   TFuzeParamWideArea::IsWideArea(float* pRange)
{
   if ( pRange )
      *pRange = fFireOffset.y;
   
   return true;
}

TFuzeParamWideArea::TFuzeParamWideArea() :
   TFuzeParam(efuzetypeAcousticTracker)
{
   fSelfDefenseRange  =  5.0;
   fRangeErrMax       = 0.2;
   fVelocityErrMax    = 0.2;
   fAzimuthErrMax     = 0.1;
   fTypeErrProb       = 0.05;
   fTypeErrMult       = 0.5;
   fFireWindow        = 0.1;
   
   fFireOffset.x = 0.0;
   fFireOffset.y = 100.0;

   fTrackRange       = 200.0;
   
   fDetonResult      = edetNone;
}

TFuzeParamWideArea::~TFuzeParamWideArea()
{
}

EBoolean     TMineParamWam::ParseDbStructItem(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses the one statement.
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
//   TItemList *    pList;

   returnCode = infile.ParseItem(fKeyMineParamWamTable, eSizeMineParamWamTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
  				case empwamkwProjType: {
                     // read name, look up db entry
                  fpDbProj = (cDbProj*)(gProjDbMgr.FindByName((char *)returnVal));
                                    
                  delete[] returnVal;
                  returnVal = NULL;
                  if ( ! fpDbProj )
                     return false;
 					break;
               }

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TMineParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TMineParamWam::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

void    TMineParamWam::WriteMineOpt(ofstream* file)
{
   TMineParam::WriteMineOpt(file);
   if ( fpDbProj )
      *file << "projectile " << fpDbProj->GetEntryEntityName() << endl;
   else
      *file << "projectile type undefined" << endl;
}

TMineParamWam::TMineParamWam(TEntityType& type, int projType) :
   TMineParam(type)
{
   fpDbProj   = (cDbProj*)(gProjDbMgr.FindByName(fpProjTextType[projType]));
}

TMineParamWam::TMineParamWam(int projType) :
   TMineParam()
{
   fpDbProj   = (cDbProj*) ( gProjDbMgr.FindByName(fpProjTextType[projType]) );
}

TMineParamWam::~TMineParamWam()
{
}

EBoolean     TFuzeParamWamSublet::ParseDbStructItem(cInFileObj& infile)
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeySubletFuzeTable, esubfuzeSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case esubfuzekwFuzeTypeName: {
                     // read name, look up index
   					int typeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
                      
                      // if we get a type we cannot use, mark us as not valid
   					if( typeNameIndex != efuzetypeTopSublet ) {
                     fTypeNameIndex = efuzetypeOther;
                     cout<< "ignoring unuseable fuze type for top attack sublet"<<endl;
                  }
                  
                  delete[] returnVal;
                  returnVal = NULL;
 					break;
               }

				case esubfuzekwConeAngle:
                  fConeAngle = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case esubfuzekwScanDetProb:
                  fScanDetProb = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case esubfuzekwHitProb:
                  fHitProb = *(double*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;
               
				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return TFuzeParam::ParseDbStructItem(infile);

		default:
			Assert(false, kerrReadParamfile, "TFuzeParamWamSublet::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

void    TFuzeParamWamSublet::WriteFuzeOpt(ofstream* file)
{
   *file << "cone angle " << fConeAngle << endl;
   *file << "scan detect prob " << fScanDetProb << endl;
   *file << "hit prob " << fHitProb << endl;
}

cControl*     TFuzeParamWamSublet::CreateControl(cProjectile* pProj, const TFuzeOpt& /*fuzeOpt*/ )
{
   cControl* pNewControl = NULL;
   
   switch (fTypeNameIndex) {
   
      case efuzetypeTopSublet:
            pNewControl = new cWamSubletControl(pProj, fConeAngle,
                                             fScanDetProb, fHitProb);
         break;
         
      default:
         Assert( false, kerrValidityCheck, "TFuzeParamWamSublet::CreateControl");
   }
   
   return pNewControl;
}

TFuzeParamWamSublet::TFuzeParamWamSublet() :
   TFuzeParam(efuzetypeTopSublet)
{
   fDetonResult      = edetEntityImpact;

   fConeAngle  =  55.0 * kDegToRad;
   fScanDetProb     = 0.95;
   fHitProb    = 0.7;

   fTrackRange       = 200.0;
}

TFuzeParamWamSublet::~TFuzeParamWamSublet()
{
}

// ************************************
//   components
// ************************************

void   cLinkVehErr::UpdateTrkRec()
{
      // first update to true values
   cLinkVehGeom::UpdateTrkRec();
   
      // apply errors
   fRange     *= fRangeError;
   fHorzRange *= fRangeError;
   fVelocity  *= fVelError;
   
   fAzimuth   += fAzimuthError;
   if (fAzimuth > kPi)
      fAzimuth -= 2.0 * kPi;
   else if (fAzimuth <= -kPi)
      fAzimuth += 2.0 * kPi;
      
   fLocation.x = fHorzRange * cos(fAzimuth);
   fLocation.y = fHorzRange * sin(fAzimuth);
   fLocation.z = fRange     * sin(fElevation);

      // fbApproaching was set earlier in SetLocation(),
      // should re-set now to reflect new location and velocity
   if ( fbTgtVelRef ) {
      if ( IsVerySmall(fVelocity.x) ) fbApproaching = false;
      else {
         if ( (fLocation.x / fVelocity.x) < 0.0 ) fbApproaching = true;
         else                                     fbApproaching = false;
      }
   }
   
   else {
      if ( IsVerySmall(fVelocity.y) ) fbApproaching = false;
      else {
         if ( (fLocation.y / fVelocity.y) < 0.0 ) fbApproaching = true;
         else                                     fbApproaching = false;
   
      }
   }
}

cLinkVehErr::cLinkVehErr(cLinkVehErr* pTrkVeh, cLinkVeh* pPrev) :
      cLinkVehGeom( (cLinkVehGeom*)pTrkVeh, pPrev)
         // copies from an existing error track link
{
   fRangeError   = pTrkVeh->fRangeError;
   fAzimuthError = pTrkVeh->fAzimuthError;
   fVelError     = pTrkVeh->fVelError;
   fTgtType      = pTrkVeh->fTgtType;
}

cLinkVehErr::cLinkVehErr(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev, cEntity* pRefEnt, 
      EBoolean refInvert, int tgtType, float rangeErr, float velErr, float azErr) :
      cLinkVehGeom(pProxVeh, pPrev, pRefEnt, refInvert)
         // copies from the original proximity link
{
   fTgtType      = tgtType;
   fRangeError   = rangeErr;
   fVelError     = velErr;
   fAzimuthError = azErr;
}

cLinkVehErr::~cLinkVehErr()
{
}

cLinkVehSublet::cLinkVehSublet(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev,
                              cEntity* pRefEnt, EBoolean refTgtVel) :
      cLinkVehGeom(pProxVeh, pPrev, pRefEnt, refTgtVel)
         // copies from the original proximity link
{
   fLastElevation      = 0.0;
}

cLinkVehSublet::cLinkVehSublet(cLinkVehSublet* pTrkVeh, cLinkVeh* pPrev) :
      cLinkVehGeom( (cLinkVehGeom*)pTrkVeh, pPrev)
         // copies from an existing sublet track link
{
   fLastElevation = pTrkVeh->fElevation;
}

cLinkVehSublet::~cLinkVehSublet()
{
}

cLinkVehCross::cLinkVehCross(cLinkVehLoc* pProxVeh, cLinkVeh* pPrev,
                              cEntity* pRefEnt, EBoolean refTgtVel) :
      cLinkVehGeom(pProxVeh, pPrev, pRefEnt, refTgtVel)
         // copies from the original proximity link
{
   fLastY      = fLocation.y;
   fLastAz     = fAzimuth;
   fLastRange  = fRange;
}

cLinkVehCross::cLinkVehCross(cLinkVehCross* pTrkVeh, cLinkVeh* pPrev) :
      cLinkVehGeom( (cLinkVehGeom*)pTrkVeh, pPrev)
         // copies from an existing crossing track link
{
   fLastY = pTrkVeh->fLocation.y;
   fLastAz = pTrkVeh->fAzimuth;
   fLastRange  = pTrkVeh->fRange;
}

cLinkVehCross::~cLinkVehCross()
{
}

EWarhead      cProjectile::GetWarhead()
{ 
   return ewhOther;
}

EFuze         cProjectile::GetFuze()
{
   return efzOther;
}

float         cProjectile::GetHitProb()
{
   return 1.0;
}

TEntityType         cProjectile::GetDetonationType()
{ 
   return fCmsType;
}

cProjectile::cProjectile(cDbProj* pDbProj)
{ 
   fpDbProj      = pDbProj;
   fpFiringMine  = NULL;
   
   fpTracker = NULL;
   fpControl = NULL;

   gCmsMgr.GetNextEventId(&fEventID);
   feIntState   = eprojisFlying;
   fLaunchTime  = GetCurSimTime();
         // specific projectile type constructors must initialize
         // all geometry, eg fLaunchTccLoc as well as entity data
         
   fpTarget = NULL;
   fbKilledTgt = false;
   fKilledTgts = 0;
   feDetonResult = edetNone;
   fFireRange    = 0.0;
}

cProjectile::~cProjectile()
{ }

void   cProjectile::SetDetonationResults(cLinkVehGeom* pLink)
{
   cVehicle* pVeh = pLink->GetVehicle();
   TEntCoord tgtRef( -1.0, 0.0, 0.0);
   
   if ( UniformRand() < GetHitProb() ) {
      fpTarget = pVeh;
      fTargetType = pVeh->GetEntType();
      fTargetID = pVeh->GetEntityID();

      fLocTargetRef = tgtRef;
      feDetonResult = edetEntityImpact;

      SetIntState(eprojisFiredHit);
      pLink->SetInteractResult(eHitTarget);
      ((cAtWideArea*)fpFiringMine)->SetProjectileResults(eprojisFiredHit);
      new cCheckVehDamageMsg(fpFiringMine, fsDamageCheckDelay, fpTarget);
   }
   
   else {
      feDetonResult = edetGroundImpact;
      tgtRef.x -= 4.0;

      SetIntState(eprojisFiredMissed);
      pLink->SetInteractResult(eMissedTarget);
      ((cAtWideArea*)fpFiringMine)->SetProjectileResults(eprojisFiredMissed);
   }
   
   TLocation impact = pVeh->GetDrGeoLoc( GetCurSimTime() );
   TEntCoord offset = pVeh->GetGeoRot() * tgtRef;
   impact += offset;
   fDetonLoc = impact;
      
   fFireRange = pLink->GetRange();

   if(fpDisplay)  {
       delete fpDisplay;
       fpDisplay = NULL;
   }
}



void    cAtOffRoute::FireProjectile(const cLinkVehGeom*)
{}

//
//   cGenericWam
//
int    cWideAreaTracker::GetTrackType(const TEntityType& type)
{
   if ( (type.kind == ekindPlatform) && (type.domain == eplatdomLand) ) {
      switch (type.category) {
         case eplatlandTank:
               return cAtWideArea::etrkTank;
            // break;
            
         case eplatlandArmorFightVeh:
         case eplatlandSelfPropArty:
         case eplatlandLgTrkUtil:
               return cAtWideArea::etrkOtherHeavyTracked;
            // break;
            
         case eplatlandLgWheelUtil:
         case eplatlandArmorUtilVeh:
         case eplatlandTowedArty:
               return cAtWideArea::etrkHeavyWheel;
            // break;
            
         case eplatlandSmTrkUtil:
               return cAtWideArea::etrkLightTracked;
            // break;
            
         case eplatlandSmWheelUtil:
               if ( type.MatchType(kEntTypeORSMC) )
                  return cAtWideArea::etrkTank;
               else
                  return cAtWideArea::etrkLightWheel;
         
         case eplatlandMortar:
               return cAtWideArea::etrkLightWheel;
            // break;
            
         default:
               return cAtWideArea::etrkNot;
      }
   }
   
   else
      return cAtWideArea::etrkNot;
}

cLinkVehGeom*   cWideAreaTracker::CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev)
{
   int tgtType = GetTrackType(pProxLink->GetVehicle()->GetEntType());
   cAtWideArea* pMine = (cAtWideArea*)fpRefEnt;

      // it there is a vehicle type error, range and velocity errors multiply
   float   typeMult = 1.0;
   if ( UniformRand() < pMine->GetTypeErrProb() ) {
      switch (tgtType) {
         case cAtWideArea::etrkTank: 
            tgtType = cAtWideArea::etrkOtherHeavyTracked;
            break;
         case cAtWideArea::etrkOtherHeavyTracked: 
            tgtType = cAtWideArea::etrkTank;
            break;
         case cAtWideArea::etrkHeavyWheel: 
            tgtType = cAtWideArea::etrkLightWheel;
            break;
         case cAtWideArea::etrkLightTracked: 
            tgtType = cAtWideArea::etrkOtherHeavyTracked;
            break;
         case cAtWideArea::etrkLightWheel: 
            tgtType = cAtWideArea::etrkHeavyWheel;
            break;
      }
      if ( UniformRand() < 0.5 )
         typeMult = pMine->GetTypeErrMult();
      else
         typeMult = 1.0 / pMine->GetTypeErrMult();
   }

   float rangeErr   = typeMult * (1.0 + ( UniformRand() * pMine->GetRangeErrMax() ) );
   float velErr     = typeMult * (1.0 + ( UniformRand() * pMine->GetVelErrMax() ) );
   if (UniformRand() < 0.5) {
      rangeErr = 1.0 / rangeErr;
      velErr = 1.0 / velErr;
   }
   float azErr = ( (2.0 * UniformRand()) - 1.0) * pMine->GetAzErrMax();


   cLinkVehErr* pNewLink = new cLinkVehErr(pProxLink, pPrev, pMine, fbUseVehCoord, 
                           tgtType, rangeErr, velErr, azErr);

   return pNewLink;
}

cLinkVehGeom*   cWideAreaTracker::CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev)
{
   cLinkVehErr* pNewLink = new cLinkVehErr( (cLinkVehErr*)pOldTrkLink, pPrev);

   return pNewLink;
}

cWideAreaTracker::~cWideAreaTracker()
{ }

void    cWamControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList, TTimeSim* pWaitTime)
{
   cMineLinkEmiss* pEmLink = NULL;	   
   cLinkVehErr* pLink = NULL;
//   float range = FLT_MAX;
   EMineInternalState newState = fpMine->GetInternalState();
   
      //** should report data on state emisInhibit
   if ( newState >= emisOn ) {
      newState = emisOn;
      *pWaitTime = fpMine->GetUpdateRate();
      
         // go through entire list looking for something to fire at
         //   However, tanks, tracked vehicles, and heavies have priority
      for (T2link* pLn  = pVehList->GetFirstLink();
                   pLn != pVehList->GetNULLLink();
                   pLn  = pVehList->GetNextLink() ) {
		   
      	 for (cMineLinkEmiss* pEmLn  = (cMineLinkEmiss*) pEmissList->GetFirstLink();
                      	      pEmLn != (cMineLinkEmiss*) pEmissList->GetNULLLink();
                              pEmLn  = (cMineLinkEmiss*) pEmissList->GetNextLink() ) {
		if(pEmLn->GetEmission()->GetVehicle()->GetEntityID() == ((cLinkVehGeom*)pLn)->GetVehicle()->GetEntityID() &&
		   (pEmLn->GetEmission()->GetEmitterName() == DIS203_EMITTER_SYSTEM_ORSMC ||
		    pEmLn->GetEmission()->GetEmitterName() == 0))
		{ 
		   pEmLink = pEmLn;
		   break;
		}
	 }	
	 
	 if(pEmLink) {
            if ( pEmLn->GetHorzRange() < ((cAtWideArea*)fpMine)->GetSelfDefenseRange()) {
               pLink = (cLinkVehErr*)pLn;     // self defense, don't look any further
	       pLink->SetTgtType(cAtWideArea::etrkORSMC);
               SelfDefenseFire(pLink, pEmLink);
               return;
            }
	 }
         if ( ((cLinkVehGeom*)pLn)->GetHorzRange() < ((cAtWideArea*)fpMine)->GetSelfDefenseRange()) {
            pLink = (cLinkVehErr*)pLn;     // self defense, don't look any further
            SelfDefenseFire(pLink);
            return;
         }
         
         if (TgtGeometry((cLinkVehErr*)pLn, pEmLink)) {   // this is possible tgt
            if (pLink) {  // already have tentative tgt
               if ( ( ((cLinkVehErr*)pLn)->GetTgtType() < pLink->GetTgtType() ) ||
                    ( ( ((cLinkVehErr*)pLn)->GetTgtType() == pLink->GetTgtType()) &&
                      ( ((cLinkVehErr*)pLn)->GetHorzRange() < pLink->GetHorzRange()) ) )
                  pLink = (cLinkVehErr*)pLn;
            }
            else  // put as tentative tgt, if it is heavy enough
               if ( ((cLinkVehErr*)pLn)->GetTgtType() <= cAtWideArea::etrkHeavyWheel )
                  pLink = (cLinkVehErr*)pLn;
         }
      }
      
         // look at geometry for fire or close track
      if (pLink) {
         pLink->SetInteractResult(ePrimaryTarget);
         SetupForFire(pLink, pEmLink, pWaitTime);
      }
   }
      
   if ( gfInteract )
      fpMine->UpdateInteractions();
}

void  cWamControl::SelfDefenseFire(cLinkVehErr* pLink, cMineLinkEmiss* pEmLink)
{
      // need tcc ref fire azimuth.  Link has az rotation of coord system, as
      // well as az within this system
   float fireAz = pLink->GetAzimuth() + pLink->GetAzRot();
   if (fireAz > kPi) 
      fireAz -= 2.0 * kPi;
   else if (fireAz <= - kPi) 
      fireAz += 2.0 * kPi;

   cWamSublet* pSublet = (cWamSublet*) ( ((cAtWideArea*)fpMine)->CreateProjectile() );
   pSublet->InitWamSublet((cAtWideArea*)fpMine, fireAz);
   ((cAtWideArea*)fpMine)->SetDetonationResults(pLink, pEmLink, pSublet, fireAz);
   new cFireProjectileMsg(fpMine->GetFireData());   // fire PDU
   //fpMine->StateChange(emisDetonated);
   pLink->SetInteractResult(eAttackedSelfDefense);
/*
   cout<<" WAM control, self defense firing, tcc az = "
      <<fireAz<<endl;
*/
}

void  cWamControl::SetupForFire(cLinkVehErr* pLink, cMineLinkEmiss* pEmLink, TTimeSim* pWait)
{
   TEntCoord loc;
   if (pEmLink && pLink->GetTgtType() == cAtWideArea::etrkORSMC)
   {
   	loc = pEmLink->GetLocation();
   }	 
   else 
   	loc = pLink->GetLocation();
   
   const  TVelocity& vel = pLink->GetVelocity();
   float  tof = fabs(loc.y) / ((cAtWideArea*)fpMine)->GetSubletLaunchVel().x;   // seconds, not sim time
   float  cpaDelay = -loc.x / vel.x;
   float  fireDelay = cpaDelay - tof;
   float  fireTimeWindow = ((cAtWideArea*)fpMine)->GetFireWindow();

//   cout << "wam fire delay " << fireDelay;
   
   if ( fireDelay < - 2.0 * fireTimeWindow ) {
      cout<<" WAM control, lost fire opportunity"<<endl;
      pLink->SetInteractResult(eMissedWindow);
      return;      // lost firing opportunity
   } 
   else if ( fireDelay < fireTimeWindow ) {   
            // time to fire

         //   aim pt is along local y axis, assume flat launch
         //** get local terrain slope
      float fireAz;
      if ( loc.y >= 0.0 ) {
         fireAz = pLink->GetAzRot() - kHalfPi;
         if (fireAz <= - kPi) fireAz += 2.0 * kPi;
      }
      else {
         fireAz = pLink->GetAzRot() + kHalfPi;
         if (fireAz >  kPi) fireAz -= 2.0 * kPi;
      }
         
      cWamSublet* pSublet = (cWamSublet*) ( ((cAtWideArea*)fpMine)->CreateProjectile() );
      pSublet->InitWamSublet((cAtWideArea*)fpMine, fireAz);
      ((cAtWideArea*)fpMine)->SetDetonationResults(pLink, pEmLink, pSublet, fireAz);
      
      new cFireProjectileMsg(fpMine->GetFireData());   // fire PDU
      //fpMine->StateChange(emisDetonated);
      pLink->SetInteractResult(eAttackedTarget);
      cout<<" WAM control, firing "<< pLink->GetTgtType() << endl;
   }
   
   else if ( fireDelay < ( 2.0 * kSimTimeToSec * fpMine->GetUpdateRate() ) ) {
            // close to fire
      if ( fireDelay > 0.1 )
         *pWait = kSecToSimTime * 0.1;
      else
         *pWait = kSecToSimTime * fireDelay;
      
//      cout << " close tracking, wait sim time " << *pWait<<endl;
      if ( fpMine->GetInternalState() != emisCloseTracking)
         fpMine->StateChange(emisCloseTracking);
   }
   
   else if ( fireDelay < ( 5.0 * kSimTimeToSec * fpMine->GetUpdateRate() ) ) {
            // approaching
      *pWait = fpMine->GetUpdateRate() / 4;
      
//      cout << " near tracking, wait sim time " << *pWait<<endl;
      if ( fpMine->GetInternalState() != emisCloseTracking)
         fpMine->StateChange(emisCloseTracking);
   }
   
   else {            // need to wait
      *pWait = fpMine->GetUpdateRate();
//      cout << " normal tracking, wait sim time " << *pWait<<endl;
      if ( fpMine->GetInternalState() != emisOn)
         fpMine->StateChange(emisOn);
   }
}

EBoolean  cWamControl::TgtGeometry( cLinkVehGeom* pLink, const cMineLinkEmiss* pEmLink)
   // tgt within cross range, and not yet to intercept pt
{
   TEntCoord loc;
   if (pEmLink) 
   {
      loc = pEmLink->GetLocation();
      if ( ( fabs(loc.y) <= fFireOffset.y) && (loc.x < fFireOffset.x) ) {
         if ( pLink->GetInteractResult() == eTracking )
            pLink->SetInteractResult(eAttackGeomSat);
	 ((cLinkVehErr*)pLink)->SetTgtType(cAtWideArea::etrkORSMC);
         return true;
      }
   }  
   loc = pLink->GetLocation();
   
   if ( ( fabs(loc.y) <= fFireOffset.y) && (loc.x < fFireOffset.x) ) {
      if ( pLink->GetInteractResult() == eTracking )
         pLink->SetInteractResult(eAttackGeomSat);
      return true;
   }
   else
      return false;
}

cWamControl::~cWamControl()
{ }

// ************************************
//   cAtWideArea 
// ************************************

void   cAtWideArea::SetDetonationResults(const cLinkVehGeom* pLink, const cMineLinkEmiss*  pEmLink,
                     cProjectile* pProj, float fireAz, EBoolean all)
   // parameters set for fire, not for detonation
{
   fFireAzimuth = fireAz;
   
   fpTarget    = pLink->GetVehicle();
   fTargetType = fpTarget->GetEntType();

   fFireData.fEventId = fDetonationData.fEventId;
   fFireData.fMunitionEnt        = pProj->GetEntityID();
   fFireData.fFiringEnt = fId;
   fFireData.fTargetEnt = fpTarget->GetEntityID();
   fFireData.fLocation =  fGeoLocation;
   fFireData.fFireTccLocation =  fTccLocation;
   fFireData.fGeoLaunchVel  = pProj->GetVehicleGeoVel();
   fFireData.fDetonationType =  GetDetonationType();
   fFireData.fWarhead =  GetWarhead();
   fFireData.fFuze =  GetFuze();
   fFireData.fFireRange = pLink->GetRange();
   
   //feDetonResult = edetDetonation;
   if (pEmLink && ((cLinkVehErr*)pLink)->GetTgtType() == etrkORSMC)
   	fFireData.fFireRange  = pEmLink->GetRange();
   else fFireData.fFireRange  = pLink->GetRange();
   
   FireMine(emisDetonated, all);

   SetChangeIndicator();
}

void   cAtWideArea::SetProjectileResults(int state) 
   // gets feedback from projectile, updates component's count
{
   if ( (EProjectileIntState)state == eprojisFiredHit ) {
      feIntState = emisDetonatedHit;
      ((cMineComp*)fpMfComponent)->IncrementVehHit();
   }
   else
      feIntState = emisDetonatedMissed;
}

void       cAtWideArea::WriteTgtType(int type, ofstream* pOut)
{
   switch (type) {
      case etrkNot:
            *pOut << "noTarget";
         break;
   
      case etrkTank:
            *pOut << "Tank";
         break;
   
      case etrkOtherHeavyTracked:
            *pOut << "HeavyTrack";
         break;

      case etrkHeavyWheel:
            *pOut << "HeavyWheel";
         break;
   
      case etrkLightTracked:
            *pOut << "LightTrack";
         break;
   
      case etrkLightWheel:
            *pOut << "LightWheel";
         break;
   }
}

cInteraction*   cAtWideArea::CreateNewInteraction(cLinkVehGeom* pLinkVeh)
{
   return new cWamInteraction(this, pLinkVeh);
}

cAtWideArea::cAtWideArea(cMineComp* pParent, float radius) :
      cAtMine(pParent, radius)
{
}

cAtWideArea::cAtWideArea( cPlanMine* pPlan ) :
      cAtMine( pPlan )
{
}

cAtWideArea::~cAtWideArea()
{ }

/*
void   cGenericWam::InitGenericWam()
{
   SetEntityType(fsAtGenericWamEntType);

   fpTracker   = new cWideAreaTracker(this, fsMineParams.fTrackRange, true);
   fpMechanVuln  = new cMechanVuln;
   fpControl   = new cWamControl(this, fsMineParams.fFireOffset);
   fpDetonVuln = new cDetonVuln(this,  fsMineParams.fESMBcumDetProb, fsMineParams.fESMBcumNeutProb, 
                                fsMineParams.fIndetRange);

   if ( UniformRand() > fsMineParams.fProbDud ) 
      feIntState = emisOff;
   else
      feIntState = emisFailed;
}

cGenericWam::cGenericWam( cMineDBMgr& mineDbMgr)
{
   SetEntityType(fsAtGenericWamEntType);

   fsMineParams.fDetonationType = fsAtGenericWamEntType;
   fsMineParams.fTrackRange = 200.0;
   fsMineParams.fFireOffset = TFlatEntCoord( 0.0, 100.0 );
   fsMineParams.fAltFireOffset = TFlatEntCoord( 0.0, 100.0 );
   fsMineParams.fWarhead = ewhMine;
   fsMineParams.fFuze = efzAcoustic;
   fsMineParams.fProbDud = 0.0;
   fsMineParams.fProbDeton = 1.0;
   fsMineParams.fArmDelay = 0.0;
   fsMineParams.fSelfDestDelay = INT_MAX;
   fsMineParams.fESMBcumDetProb = 0.8;
   fsMineParams.fESMBcumNeutProb = 0.9;

   fsMineParams.fIndetRange[edvinNo]          = -1.0;
   fsMineParams.fIndetRange[edvinSmallBlast]  =  2.5;
   fsMineParams.fIndetRange[edvinLargeBlast]  = 10.0;
   fsMineParams.fIndetRange[edvinESMB]        = 75.0;
   fsMineParams.fIndetRange[edvinMiclic]      = 75.0;
   
   mineDbMgr.Register(this, cMineDBMgr::emdblevSpecific);
}

cGenericWam::cGenericWam(cMineComp* pParent, double extraRange) : 
                  cAtWideArea(pParent)
      // constructor for virtual mine, never simulated
{ 
   SetEntityType(fsAtGenericWamEntType);

   fpTracker   = new cWideAreaTracker(this, fsMineParams.fTrackRange + extraRange);
   fpDetonVuln = new cDetonVuln(this,  fsMineParams.fESMBcumDetProb, fsMineParams.fESMBcumNeutProb, 
                                fsMineParams.fIndetRange, extraRange);
}

cGenericWam::cGenericWam(cMine* pFormerObj, unsigned char ) : 
                  cAtWideArea(pFormerObj)
{ 
   InitGenericWam();
}

cGenericWam::cGenericWam( cMineComp* pParent, TEntCoord* pLoc) : 
                        cAtWideArea(pParent, pLoc)
{ 
   InitGenericWam();
}

cGenericWam::~cGenericWam()
{ }
*/

int    cSubletTracker::GetTrackType(const TEntityType& type)
   // sees everything but ORSMC
{
   if ( (type.kind == ekindPlatform) && (type.domain == eplatdomLand)
         && ( ! type.MatchType(kEntTypeORSMC) ) )
      return cProjectile::etrkYes;
   else
      return cProjectile::etrkNot;
}

cLinkVehGeom*   cSubletTracker::CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev)
{
   cLinkVehSublet* pNewLink = new cLinkVehSublet(pProxLink, pPrev, fpRefEnt, false);

   return pNewLink;
}

cLinkVehGeom*   cSubletTracker::CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev)
{
   cLinkVehSublet* pNewLink = new cLinkVehSublet((cLinkVehSublet*)pOldTrkLink, pPrev);

   return pNewLink;
}

cSubletTracker::~cSubletTracker()
{}

void   cWamSubletControl::UpdateControl(cVehListHead* pVehList, cEmissListHead*, TTimeSim*)
   // go through tgt veh list, see if any have passed through cone angle
{
   for (cLinkVehSublet* pLn  = (cLinkVehSublet*) (pVehList->GetFirstLink());
                        pLn != (cLinkVehSublet*) (pVehList->GetNULLLink());
                        pLn  = (cLinkVehSublet*) (pVehList->GetNextLink()) ) {
      if ( ( (pLn->GetElevation() <= fConeAngle) && (pLn->GetLastElevation() > fConeAngle) ) ||
           ( (pLn->GetElevation() >= fConeAngle) && (pLn->GetLastElevation() < fConeAngle) ) ) {
         if ( UniformRand() <= fScanDetProb ) {
//            cout<<"sublet target detected, firing"<<endl;
            fpProjectile->SetDetonationResults(pLn);
            new cProjFireMunitionMsg(fpProjectile);
            new cProjDetonationMsg(fpProjectile);
		
            if ( gfInteract )
               fpProjectile->UpdateInteractions();
            return;
         }
         
         else {         // scan error, miss detection
            pLn->SetInteractResult(eMissedDetect);
         }
      }
   }

   if ( gfInteract )
      fpProjectile->UpdateInteractions();
      
      // didn't fire this time
   new cSendVehStateMsg( fpProjectile );
}

cWamSubletControl::~cWamSubletControl()
{}

char*     cWamSublet::GetEntName()
{
   if ( fpDbProj )
      return fpDbProj->GetEntryEntityName();
   else
      return NULL;
}

void   cWamSublet::UpdateVehicle()
{
   fpUpdateVehMsg = NULL;
   TTimeSim waitTime = 0.0;

   if ( feIntState >= eprojisLowestUpdate ) {
      UpdateLocation();
      
      if ( fTccLocation.z <= gpTerrainMgr->GetMapElev(fTccLocation.x, fTccLocation.y) ) {
         feIntState = eprojisHitGround;
         fAppearance  = eAprDeact;
         new cProjDetonationMsg(this);   // ends flight
         if (fpDisplay) {
		delete fpDisplay;
		fpDisplay = NULL;
         }
      }
      
      else {
         if ( feIntState >= eprojisLowestTracker ) // prox track list is relative to launch loc
            fpTracker->UpdateTracker( fMasterVeh, &fTrackVeh, fLocRelLaunch);
   
         if ( feIntState >= eprojisLowestControl ) 
            fpControl->UpdateControl(&fTrackVeh, (cEmissListHead*)NULL, &waitTime);
      }
      
      if ( feIntState >= eprojisLowestUpdate)
         fpUpdateVehMsg = new cUpdateProjectileMsg(fsUpdateRate, this);
      else {
         fTrackVeh.DeleteList();
         delete fpTracker;
         fpTracker = NULL;
         delete fpControl;
         fpControl = NULL;
      }
   }
   if ( fpDisplay)
         fpDisplay->UpdateVehicleDisp();

}

void   cWamSublet::UpdateLocation()
   // our rotations/orientations never change, so we don't go back to SetCoordRef calls
   //  Update tcc and geo locations right here
{
   const     T2Dsingle&  launchVel = GetLaunchVel();
   fTime = GetCurSimTime();
   double time = kSimTimeToSec * (fTime - fLaunchTime);
   fLocRelLaunch.x = launchVel.x * time;
   fLocRelLaunch.y = 0.0;
         // entity z is down, launch vel is up, grav is down
   fLocRelLaunch.z = -launchVel.y * time + (kHalfGravAccel * time * time);
   
   fEntVelocity.x = launchVel.x;
   fEntVelocity.y = 0.0;
   fEntVelocity.z = -launchVel.y + kHalfGravAccel * time;
   fGeoRotation.MultTrans( fEntVelocity, &fGeoVelocity);
   if ( fDrAlg < eDrLowestBody )
      fVelocity = fGeoVelocity;
   else
      fVelocity = fEntVelocity;
   
      //   operator | is mult by transpose
   fGeoLocation = (fGeoRotation | fLocRelLaunch) + fLaunchGeoLoc;
   fTccLocation = (fTccRotation | fLocRelLaunch) + fLaunchTccLoc;
}

EWarhead      cWamSublet::GetWarhead()
{
   return fpDbProj->GetProjParam()->fWarhead;
}

EFuze         cWamSublet::GetFuze()
{
   return fpDbProj->GetFuzeParam()->fFuze;
}

TEntityType         cWamSublet::GetDetonationType()
{
   return fpDbProj->GetProjParam()->fDetonationType;
}

float         cWamSublet::GetHitProb()
{
   return ( (TFuzeParamWamSublet*) (fpDbProj->GetFuzeParam()) )->fHitProb;
}
         
cInteraction*  cWamSublet::CreateNewInteraction(cLinkVehGeom* pLinkVeh)
{
   return new cInteraction(this, pLinkVeh);
}

void    cWamSublet::WriteEntSumm(ofstream* pOut)
{
   *pOut << GetEntName();
   *pOut << " id: "
         << GetEntityID().site        << " "
         << GetEntityID().application << " "
         << GetEntityID().entityNum;
   *pOut << " state: "
         << GetInternalStateText(feIntState);
   *pOut << " mapLocation: "
         << fTccLocation.x << " "
         << fTccLocation.y << " "
         << fTccLocation.z;
   int   utmZone = 0;
   double utmNorth = 0.0, utmEast = 0.0, utmZ = 0.0;
   
   if ( ! gpTerrainMgr->GeoToUtmLoc(fGeoLocation, 
               &utmZone, &utmNorth, &utmEast, &utmZ) ) {
      *pOut << " utmLocation: "
            << utmZone  << " "
            << utmNorth << " "
            << utmEast  << " "
            << utmZ;
   }
   else {
      *pOut << " utm coords not available";
   }
   
   if (fpTarget) {
      *pOut << endl
            << "Attacked target: ";
      fpTarget->WriteEntSumm(pOut);
      *pOut << " target killed: "
            << (fbKilledTgt == true ? "true" : "false");
   }
   
   *pOut << endl
         << "[\nFiring mine: ";
   fpFiringMine->WriteEntSumm(pOut);
   *pOut << endl
         <<"]";
}

void    cWamSublet::UpdateInteractions()
{
         // for output, update each link
   if ( gfInteract ) {
      cInteraction* pInter;
      for (cLinkVehGeom* pLn  = (cLinkVehGeom*)(fTrackVeh.GetFirstLink());
                         pLn != (cLinkVehGeom*)(fTrackVeh.GetNULLLink());
                         pLn  = (cLinkVehGeom*)(fTrackVeh.GetNextLink()) ) {
         pInter = pLn->GetInteraction();
         if ( pInter ) {      // there is an interaction record
            pInter->UpdateInteraction(pLn);
            if ( pInter->GetResult() == eTracking) {
               switch (feIntState) {
                  case eprojisNone:
                  case eprojisFailed:
                  case eprojisHitGround:
                        pInter->SetResult(eNoAttack);
                     break;
               
                  case eprojisFiredMissed:
                  case eprojisFiredHit:
                        if ( pLn->GetVehicle() == fpTarget ) {
                           if (feIntState == eprojisFiredHit)
                              pInter->SetResult(eHitTarget);
                           else
                              pInter->SetResult(eMissedTarget);
                        }

                     break;
               
               
                  case eprojisFlying:
                        if ( ! pLn->IsApproaching() ) {
                           pInter->SetResult(eNoAttack);
                           pLn->DeleteInteraction();
                        }
                     break;
               }
            }
         }

         else {            // no interaction record for this link, may need one
            if ( pLn->GetVehicle() == fpTarget ) {
               pInter = CreateNewInteraction(pLn);
               pLn->SetInteraction( pInter );
               pInter->UpdateInteraction(pLn);
               if (feIntState == eprojisFiredHit)
                  pInter->SetResult(eHitTarget);
               else if (feIntState == eprojisFiredMissed)
                  pInter->SetResult(eMissedTarget);
            }
            else {      // no detonation, create interaction record for approaching veh
               if ( feIntState >= eprojisLowestUpdate) {
                  if ( pLn->IsApproaching() ) {
                     pInter = CreateNewInteraction(pLn);
                     pLn->SetInteraction( pInter );
                     pInter->UpdateInteraction(pLn);
                     pInter->SetResult(eTracking);
                  }
               }
            }
         }
      }
   }
}

void   cWamSublet::InitWamSublet(cAtWideArea* pMine, float tccAimAz)
{
   fForce       = pMine->GetMinefield()->GetForce();
   fpFiringMine = pMine;
   
      // entities have down pointing z, rotate around forward/x axis
   SetCoordTccRef (pMine->GetTccLoc(), tccAimAz, 0.0, kPi);
   fLaunchTccLoc = pMine->GetTccLoc();
   fLaunchGeoLoc = pMine->GetGeoLoc();

   fAppearance  = eAprStandard;
      //** check dead reckoning alg
      // accel and ang accel init to zeros
   fDrAlg       = eDrFpb;
   
      // need prox track list in own coordinates
   cLinkVeh* pLink;
   cLinkVehLoc* pNewLink;
   cVehListHead*   pList = pMine->TakeVehList();
   for (pLink  = (cLinkVeh*)(pList->GetFirstLink());
        pLink != (cLinkVeh*)(pList->GetNULLLink());
        pLink  = (cLinkVeh*)(pList->GetNextLink()) ) {
      pNewLink = new cLinkVehLoc(pLink->GetVehicle(), 
                (cLinkVeh*)(fMasterVeh.GetLastLink()), this, false);
   }
   TFuzeOpt fuzeOpt;      // not used for wam sublet
   fpTracker = fpDbProj->GetFuzeParam()->CreateTracker(this, fuzeOpt, false);
   fpControl = ((TFuzeParamWamSublet*)(fpDbProj->GetFuzeParam()))->CreateControl(this, fuzeOpt);
   fCmsType = pMine->GetDetonationType();
   fAltType=fCmsType;

   SimulateVehicle();

    new cUpdateProjectileMsg(fsUpdateRate, this);
}

cWamSublet::cWamSublet(cDbProj* pDbProj) :
   cProjectile(pDbProj)
{
}

cWamSublet::~cWamSublet()
{ }

// ************************************
//   cGenOffRoute 
// ************************************
EBoolean   cOffRouteControl::AttackGeomFilter(cLinkVehCross* pLn)
{
      // range gate, crossing detect, speed gate
   float horzSpeed = sqrt( (pLn->GetVelocity().x * pLn->GetVelocity().x)
                         + (pLn->GetVelocity().y * pLn->GetVelocity().y) );
   //cout << "Veh: HorzSpeed = " << horzSpeed << endl;
   //cout << "Veh: Last Y = " << pLn->GetLastY() << ", Cur Y = " << pLn->GetLocation().y << endl;

   if ( ( (pLn->GetLocation().y <= 0.0) && (pLn->GetLastY() > 0.0) ) ||
        ( (pLn->GetLocation().y >= 0.0) && (pLn->GetLastY() < 0.0) ) ) {
            // this is crossing time
      if ( pLn->GetLocation().x < fFireOffset.x ) {
         pLn->SetInteractResult(eTooClose);
         return false;
      }
      if ( pLn->GetLocation().x > fFireOffset.y ) {
         pLn->SetInteractResult(eTooFarWay);
         return false;
      }
      if ( horzSpeed < fMinTgtSpeed ) {
         pLn->SetInteractResult(eSlowTarget);
         return false;
      }
      if ( horzSpeed > fMaxTgtSpeed ) {
         pLn->SetInteractResult(eFastTarget);
         return false;
      }
      
      double vehheight = pLn->GetVehicle()->GetDimensions().height;
      if ( vehheight < 1.0 ) vehheight = 1.0;
      double vis = gpTerrainMgr->GetIntervis( fpMine->GetTccLoc(), fsHeight,
                                              pLn->GetVehicle()->GetTccLoc(), vehheight,
                                              fsMinVis, 1.5 );
      //TLocation tmp = pLn->GetVehicle()->GetTccLoc();
      //printf("Vehicle visibility = %lf, TccLoc = (%lf, %lf, %lf)\n", vis, tmp.x, tmp.y, tmp.z);
      if ( vis < fsMinVis ) {
         pLn->SetInteractResult(eNoIntervis);
         return false;
      }
      
      pLn->SetInteractResult(eAttackGeomSat);
      return true;
   }

   else
      return false;
}

EBoolean   cOffRouteControl::AttackGeomFilter(cMineLinkEmiss* pEmLn)
{
      // range gate, crossing detect, speed gate
   float horzSpeed = sqrt( (pEmLn->GetVelocity().x * pEmLn->GetVelocity().x)
                         + (pEmLn->GetVelocity().y * pEmLn->GetVelocity().y) );
   //cout << "HorzSpeed = " << horzSpeed << endl;
   //cout << "Last Y = " << pEmLn->GetLastY() << ", Cur Y = " << pEmLn->GetLocation().y << endl;
 
   if ( ( (pEmLn->GetLocation().y <= 0.0) && (pEmLn->GetLastY() > 0.0) ) ||
        ( (pEmLn->GetLocation().y >= 0.0) && (pEmLn->GetLastY() < 0.0) ) ) {
            // this is crossing time
      if ( pEmLn->GetLocation().x < fFireOffset.x ) {
         //pEmLn->SetInteractResult(eTooClose);
	 //cout << "Failed Fire, Location x = " << pEmLn->GetLocation().x << endl;
         return false;
      }
      if ( pEmLn->GetLocation().x > fFireOffset.y ) {
         //pLn->SetInteractResult(eTooFarWay);
	 //cout << "Failed Fire, Location x = " << pEmLn->GetLocation().x << endl;
         return false;
      }
      if ( horzSpeed < fMinTgtSpeed ) {
         //pLn->SetInteractResult(eSlowTarget);
	 //cout << "Failed Fire, Min speed = " << fMinTgtSpeed << endl;
         return false;
      }
      if ( horzSpeed > fMaxTgtSpeed ) {
         //pLn->SetInteractResult(eFastTarget);
	 //cout << "Failed Fire, Max speed = " << fMaxTgtSpeed << endl;
         return false;
      }
      
      double emissheight = pEmLn->GetEmission()->GetEmissDim().height;
      if ( emissheight < 1.0 ) emissheight = 1.0;
      //TLocation tmp = pEmLn->GetEmission()->GetTccLoc();
      double vis = gpTerrainMgr->GetIntervis( fpMine->GetTccLoc(), fsHeight,
                                              pEmLn->GetEmission()->GetTccLoc(), 
					      emissheight,
                                              fsMinVis, 1.5 );
      //printf("Emission visibility = %lf, TccLoc = (%lf, %lf, %lf)\n", vis, tmp.x, tmp.y, tmp.z);

      if ( vis < fsMinVis ) {
         //pLn->SetInteractResult(eNoIntervis);
	 //cout << "Failed Fire, visibility = " << vis << endl;
	 //TLocation tmpv = pEmLn->GetEmission()->GetVehicle()->GetTccLoc();
	 //printf (" TccLoc Vehicle = (%lf, %lf, %lf)\n", tmpv.x, tmpv.y, tmpv.z);	 
         return false;
      }
      
      //pLn->SetInteractResult(eAttackGeomSat);
      return true;
   }

   else
      return false;
}

void    cOffRouteControl::UpdateControl(cVehListHead* pVehList, cEmissListHead* pEmissList, TTimeSim* /*pWaitTime*/)
{
   EMineInternalState oldstate = fpMine->GetInternalState();
   EBoolean mine_fired = false;

   if ( oldstate >= emisOn ) {

   EMineInternalState newstate = emisOn;

      for (cMineLinkEmiss* pEmLn  = (cMineLinkEmiss*) (pEmissList->GetFirstLink());
                           pEmLn != (cMineLinkEmiss*) (pEmissList->GetNULLLink());
                           pEmLn  = (cMineLinkEmiss*) (pEmissList->GetNextLink()) ) {
        unsigned short emit_name = ((cMineLinkEmiss*) pEmLn)->GetEmission()->GetEmitterName();
        if (emit_name == DIS203_EMITTER_SYSTEM_ORSMC || emit_name == 0)
        {
         newstate = emisTracking;
         if ( AttackGeomFilter(pEmLn) ) {
            if ( UniformRand() <= fDetectProb ) {
               fpMine->SetDetonationResults(pEmLn);
               fpMine->SetChangeIndicator();
               new cFireProjectileMsg(fpMine->GetFireData());   // NULL for no proj entity state
               new cMineDetonationMsg(fpMine->GetDetonationData());
               //cout << "Shooting at Emission." << endl;
	       mine_fired = true;
               break;
            }
//          else
//               pLn->SetInteractResult(eMissedDetect);
         }
        }
      }
      if (!mine_fired) {
         for (cLinkVehCross* pLn  = (cLinkVehCross*) (pVehList->GetFirstLink());
                             pLn != (cLinkVehCross*) (pVehList->GetNULLLink());
                             pLn  = (cLinkVehCross*) (pVehList->GetNextLink()) ) {
            newstate = emisTracking;
            if ( AttackGeomFilter(pLn) ) {
               if ( UniformRand() <= fDetectProb ) {
                  fpMine->SetDetonationResults(pLn);
                  fpMine->SetChangeIndicator();
                  new cFireProjectileMsg(fpMine->GetFireData());   // NULL for no proj entity state
                  new cMineDetonationMsg(fpMine->GetDetonationData());
		  mine_fired = true;
	          break;
               }
               else
                  pLn->SetInteractResult(eMissedDetect);
            }
	 }
      }
      if ( (oldstate != newstate) &&
           (fpMine->GetInternalState() >= emisOn) )
         fpMine->StateChange(newstate);
   }
   if ( gfInteract )
         fpMine->UpdateInteractions();
}

cOffRouteControl::~cOffRouteControl()
{ }

float   cAtOffRoute::GetLaunchSpeed()
{
   return ((TFuzeParamOffRoute*)
      (fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fSpeed; 
}

float   cAtOffRoute::GetHitProb()
{
   return ((TFuzeParamOffRoute*)
      (fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fHitProb;
}

float      cAtOffRoute::GetMaxProjRange()
{
   return ((TFuzeParamOffRoute*)
      (fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fMaxRange;
}

float      cAtOffRoute::GetMinProjRange()
{
   return ((TFuzeParamOffRoute*)
         (fpDbMine->GetFuzeParam(fOptSet.fFuzeOpt)))->fMinRange;
}

void   cAtOffRoute::SetDetonationResults(cLinkVehGeom* pLink, TVehPart* /*pPart*/, EBoolean all)
{
   fpTarget = pLink->GetVehicle();
   TEntCoord tgtRef( -1.0, 0.0, 0.0);
   gCmsMgr.GetNextEventId(&fDetonationData.fEventId);

   fTargetType = fpTarget->GetEntType();

   fFireData.fEventId = fDetonationData.fEventId;
   fFireData.fFiringEnt = fId;
   fFireData.fTargetEnt = fpTarget->GetEntityID();
   fFireData.fLocation =  fGeoLocation;
   fFireData.fFireTccLocation =  fTccLocation;
   fFireData.fDetonationType =  GetDetonationType();
   fFireData.fWarhead =  GetWarhead();
   fFireData.fFuze =  GetFuze();
   fFireData.fFireRange = pLink->GetRange();

   fDetonationData.fFiringEnt = fId;
   fDetonationData.fTargetEnt = fpTarget->GetEntityID();
   fDetonationData.fDetonationType =  GetDetonationType();
   fDetonationData.fWarhead =  GetWarhead();
   fDetonationData.fFuze =  GetFuze();
   fDetonationData.fFireTccLocation =  fTccLocation;
   
   if ( ( UniformRand() < GetHitProb() )  ) {

      fDetonationData.feDetResult = GetFuzeParams()->fDetonResult;

      
      FireMine(emisDetonatedHit, all);
      pLink->SetInteractResult(eHitTarget);
      new cCheckVehDamageMsg(this, fsDamageCheckDelay, fpTarget);
   }
   
   else {
      fDetonationData.feDetResult = edetGroundImpact;
      tgtRef.x -= 4.0;

      FireMine(emisDetonatedMissed, all);
      pLink->SetInteractResult(eMissedTarget);
   }

   fDetonationData.fLocTargetRef = tgtRef;
   
   TVelocity vel(GetLaunchSpeed(), 0.0, 0.0);
   LocalToGeoVel( vel, &fFireData.fGeoLaunchVel);
   TLocation impact = fpTarget->GetDrGeoLoc( GetCurSimTime() );
   TEntCoord offset = fpTarget->GetGeoRot() * tgtRef;
   impact += offset;
   fDetonationData.fDetLocation = impact;
      
   SetChangeIndicator();
}

void   cAtOffRoute::SetDetonationResults(cMineLinkEmiss* pEmLink, EBoolean all)
{
   TVehPart* pArtPart;
   
   fpTarget = pEmLink->GetEmission()->GetVehicle();
   TEntCoord tgtRef = pEmLink->GetEmission()->GetRelLoc();
   gCmsMgr.GetNextEventId(&fDetonationData.fEventId);

   fTargetType = fpTarget->GetEntType();

   fFireData.fEventId = fDetonationData.fEventId;
   fFireData.fFiringEnt = fId;
   fFireData.fTargetEnt = fpTarget->GetEntityID();
   fFireData.fLocation =  fGeoLocation;
   fFireData.fFireTccLocation =  fTccLocation;
   fFireData.fDetonationType =  GetDetonationType();
   fFireData.fWarhead =  GetWarhead();
   fFireData.fFuze =  GetFuze();
   fFireData.fFireRange = pEmLink->GetRange();

   fDetonationData.fFiringEnt = fId;
   fDetonationData.fTargetEnt = fpTarget->GetEntityID();
   fDetonationData.fDetonationType =  GetDetonationType();
   fDetonationData.fWarhead =  GetWarhead();
   fDetonationData.fFuze =  GetFuze();
   fDetonationData.fFireTccLocation =  fTccLocation;
   
   if ( ( UniformRand() < GetHitProb() ) ) {

      fDetonationData.feDetResult = edetEntityProximate;
      //fpPartHit = new TVehPart();
      //Need to add code here to look up ORSMC part on vehicle
      //and target this part in the detonation PDU.

      pArtPart = fpTarget->GetPartByType(ePrtORSMC);
      fDetonationData.fTgtPart.fValueType = 0xFFFFFFE0;

      if (pArtPart == NULL)
         fDetonationData.fTgtPart.SetPartType(ePrtORSMC);
      else{
         fDetonationData.fTgtPart = *pArtPart;

	 /*         fDetonationData.fTgtPart.fParamType  = pArtPart->GetPartType();
         fDetonationData.fTgtPart.fDepth  = pArtPart->GetDepth();
         fDetonationData.fTgtPart.fChange  = pArtPart->GetPartChange();
         fDetonationData.fTgtPart.fAttachPoint = pArtPart->GetPartAttachPt();
         fDetonationData.fTgtPart.fAttached = pArtPart->IsAttached();
         fDetonationData.fTgtPart.fValueType = pArtPart->GetPartValueType();
         fDetonationData.fTgtPart.fValue = pArtPart->GetPartValue();
         fDetonationData.fTgtPart.fEntType = pArtPart->GetEntType(); */
      }

      FireMine(emisDetonated, all);

//      pLink->SetInteractResult(eHitTarget);
//      new cCheckVehDamageMsg(this, fsDamageCheckDelay);
   }
   
   else {

      fDetonationData.feDetResult = edetGroundImpact;
      tgtRef.x += 4.0;

      FireMine(emisDetonatedMissed, all);
//      pLink->SetInteractResult(eMissedTarget);
   }

   fDetonationData.fLocTargetRef = tgtRef;
//   fLocTargetRef.y = tgtRef.x;
//   fLocTargetRef.z = tgtRef.z;
   
   TVelocity vel(GetLaunchSpeed(), 0.0, 0.0);
   LocalToGeoVel( vel, &fFireData.fGeoLaunchVel);
   TLocation impact = fpTarget->GetDrGeoLoc( GetCurSimTime() );
   TEntCoord offset = fpTarget->GetGeoRot() * tgtRef;
   impact += offset;
   fDetonationData.fDetLocation = impact;
 
   SetChangeIndicator();
}

cAtOffRoute::cAtOffRoute(cMineComp* pParent, float radius) :
      cAtMine(pParent, radius)
{
}

cAtOffRoute::cAtOffRoute( cPlanMine* pPlan ) :
      cAtMine( pPlan )
{
}

cAtOffRoute::~cAtOffRoute()
{ }

int    cOffRouteTracker::GetTrackType(const TEntityType& type)
   // sees all land vehicles
{
   if ( (type.kind == ekindPlatform) && (type.domain == eplatdomLand) )
      return cProjectile::etrkYes;
   else
      return cProjectile::etrkNot;
}

cLinkVehGeom*   cOffRouteTracker::CreateNewTrack(cLinkVehLoc* pProxLink, cLinkVeh* pPrev)
{
   cLinkVehCross* pNewLink = new cLinkVehCross(pProxLink, pPrev, fpRefEnt, false);


   return pNewLink;
}

cLinkVehGeom*   cOffRouteTracker::CreateCopyTrack(cLinkVehGeom* pOldTrkLink, cLinkVeh* pPrev)
{
   cLinkVehCross* pNewLink = new cLinkVehCross((cLinkVehCross*)pOldTrkLink, pPrev);

   return pNewLink;
}

cOffRouteTracker::~cOffRouteTracker()
{}
