/**********************************************************************
 * File: Emission.cc
 *
 * Types defined: cLinkEmiss::*, cMineLinkEmiss::*, cEmissListHead::*, cEmission::*
 *
 * Functions defined: 
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1997
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */
#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <stdio.h>
#include "libvecmat.h"
#include "Emission.h"
//#include "Vehicle.h"
#include "Minefield.h"
//#include "Sensor.h"
#include "MapGeometry.h"



cLinkEmiss::cLinkEmiss(cEmission* pEm, cLinkEmiss* pPrev) :
   T2link(pPrev)
{   
   fpEmission = pEm;
}

cLinkEmiss::cLinkEmiss(cEmission* pEm, cLinkEmiss* pPrev, float /*hrange*/) :
   T2link(pPrev)
{   
   fpEmission = pEm;
}

void cMineLinkEmiss::SnapZ()
{
}

cMineLinkEmiss::cMineLinkEmiss(cMineLinkEmiss* pPrEmiss, cLinkEmiss* pPrev, cEntity* pRefEnt,
      EBoolean refTgtVel) : cLinkEmiss(pPrEmiss->fpEmission, pPrev)
     //copies data from another link
     //  if pRefEnt is given, new data are computed for new ref system.
     //   if pRefEnt is null, complete copy from old ref system is made.
{
   fLastY = pPrEmiss->fLocation.y;
   fMagCmResult = pPrEmiss->GetMagCmResult();
   if (pRefEnt) {
      fpRefEnt       = pRefEnt;
      fbTgtVelRef    = refTgtVel;
      //UpdateSsrRec();
   }
   else {
      fpRefEnt      = pPrEmiss->fpRefEnt;
      fbTgtVelRef   = pPrEmiss->fbTgtVelRef;
      fAzRot        = pPrEmiss->fAzRot;
      fLocation     = pPrEmiss->fLocation;
      fVelocity     = pPrEmiss->fVelocity;
      fTime         = pPrEmiss->fTime;
   }
   UpdateSsrRec();
}

cMineLinkEmiss::cMineLinkEmiss(cEmission* pEmiss, cLinkEmiss* pPrev, cEntity* pRefEnt,
      EBoolean refTgtVel) : cLinkEmiss(pEmiss, pPrev)
         // gets data from the original master emission link
{
   fMagCmResult   = eprNotEngaged;
   fpRefEnt       = pRefEnt;
   fbTgtVelRef    = refTgtVel;
   UpdateSsrRec();
   fLastY	  = fLocation.y;
}

void     cMineLinkEmiss::SetLocation()
{
   if ( (fpRefEnt != NULL) && (fpEmission != NULL) ) {
      fTime  = GetCurSimTime();
      
      if ( fbTgtVelRef ) {
         TRotation mapRot, emissTccRot;
         gpTerrainMgr->GetMapRotationGeoPt( &mapRot, fpRefEnt->GetDrGeoLoc(fTime) );
         TVelocity mapVel = mapRot * fpEmission->GetVehicle()->GetVehicleGeoVel();
				 
            // if emiss is stopped, use its orientation, otherwise its velocity for direction
         if ( (mapVel.x > 0.1) || (mapVel.y > 0.1) ) {
            fAzRot = atan2(mapVel.y, mapVel.x);
            emissTccRot.SetRotation( fAzRot, 0.0, kPi );   // z points down in veh coords
         }
         else {
            fAzRot = fpEmission->GetTccOrient().psi;
            emissTccRot = fpEmission->GetVehicle()->GetTccRot();
         }
         fVelocity = emissTccRot * mapVel;
         fLocation = emissTccRot * (fpEmission->GetDrTccLoc(fTime) - fpRefEnt->GetDrTccLoc(fTime));

         if ( IsVerySmall(fVelocity.x) ) fbApproaching = false;
         else {
            if ( (fLocation.x / fVelocity.x) < 0.0 ) fbApproaching = true;
            else                                     fbApproaching = false;
         }
      }
      
      else {
         fpRefEnt->TccToLocalLoc(fpEmission->GetDrTccLoc(fTime), &fLocation);
         fpRefEnt->GeoToLocalVel(fpEmission->GetVehicle()->GetVehicleGeoVel(), &fVelocity);
         fAzRot = 0.0;
         
         if ( IsVerySmall(fVelocity.y) ) fbApproaching = false;
         else {
            if ( (fLocation.y / fVelocity.y) < 0.0 ) fbApproaching = true;
            else                                     fbApproaching = false;
      
         }
      }
   }
}

void    cMineLinkEmiss::GetLocTgtRef(TEntCoord* pLocTgtRef) const
{
   if ( fbTgtVelRef ) {
      (* pLocTgtRef).x = -fLocation.x;
      (* pLocTgtRef).y = -fLocation.y;
      (* pLocTgtRef).z = -fLocation.z;
   }
   else {
         // rotate to global and then to tgt but without translation
      TEntCoord geoRef;
      fpRefEnt->GetGeoRot().MultTrans( ((T3Dsingle)fLocation) * float(-1.0), &geoRef);
      fpEmission->GetVehicle()->GetGeoRot().Mult(geoRef, pLocTgtRef);
   }
}

void    cMineLinkEmiss::UpdateSsrRec()
{
   SetLocation();    // get up to date emission info
   
   double dHSq = fLocation.x*fLocation.x + fLocation.y*fLocation.y;
   double dSq = dHSq + fLocation.z*fLocation.z;
   
   fHorzRange = (float) sqrt(dHSq);
   fRange     = (float) sqrt(dSq);
   //fAzimuth   = atan2(fLocation.y, fLocation.x);
   //fElevation = atan2(fLocation.z, fHorzRange);
}


cEmissListHead::cEmissListHead(cEmissListHead* pOldHead) : T2linkHead(pOldHead)
   // this list head takes over the list from the old head
{
}

cEmissListHead::cEmissListHead() : T2linkHead()
{
}


/*
int cEmission::IsInAffectedArea(T3Dsingle* pt)
{
//	return fpArea->IsPtInArea(pt);
}

int cEmission::IsInAffectedArea(T3Ddouble* pt)
{
	return fpArea->IsPtInArea(pt);
}

*/

cEmission*   cEmissListHead::GetFirstEmiss()
{ 
   fpCurrent = fpNext;
   if (fpCurrent == this)
      return NULL;
   else
      return ((cLinkEmiss*)fpCurrent)->fpEmission; 
}

cEmission*   cEmissListHead::GetNextEmiss()
{ 
   fpCurrent = fpCurrent->fpNext; 
   
   if (fpCurrent == this)
      return NULL;

   return ((cLinkEmiss*)fpCurrent)->fpEmission; 
}

EBoolean    cEmissListHead::IsInList(const cEmission* pEmiss, cLinkEmiss** ppEmissNode)
{
   for (cLinkEmiss* pVN = (cLinkEmiss*)fpNext; (T2link*)pVN != this; pVN = (cLinkEmiss*)(pVN->fpNext)) {
      if ( pVN->fpEmission == pEmiss ) {
         *ppEmissNode = pVN;
         return true;
      }
   }
   *ppEmissNode = NULL;
   return false;
}

void  cEmissListHead::ResetList()
{
 	fpNext = this;
	fpPrev = this;
}

cEmissListHead::~cEmissListHead()
{
   while (fpNext != this)
      delete fpNext;
}

const TVehDimensions& cEmission::GetEmissDim()
{
    if (fDbEmiss)
	return fDbEmiss->GetVehDim();
    else
        return TVehDimensions();
}


cEmission::cEmission(int numsys,  uint8 emitter_name)
{
	TEntityType type;
	type.kind = ekindPlatform;
	type.domain = eplatdomLand;
	type.country = ecntryOther;
	type.category = eplatlandOther;
	type.subcategory = emitter_name;
	fpSystem = new TEmissionSystem[numsys];
	fNumSystems = numsys;
//	fpArea = NULL;
	fDbEmiss = (cDbVeh*)(gVehDbMgr.MatchExact(type));
//	fprintf(stderr, "Length = %f, Width = %f\n", GetEmissDim().length, GetEmissDim().width);
}



/*
cEmission::cEmission(TEntCoord emitter_location, TOrient emitter_orient, float erp, SP_DIS203_EMISSION_PDU* pdu) 
{
        fEmitterLoc.x = emitter_location.x;
	fEmitterLoc.y = emitter_location.y;
	fEmitterLoc.z = emitter_location.z;
	fEmitterOrient = emitter_orient;
	fERP = erp;

	int offset = 10;
	float sin_psi = sin(emitter_orient.psi);
	float cos_psi = cos(emitter_orient.psi);
	TEntCoord ctr(emitter_location.x + (offset*cos_psi),emitter_location.y + (offset*sin_psi),0);


//	emitter_name		= pdu->systems->emitter_system.emitter_name;
	function		= pdu->systems->emitter_system.function;
//	emitter_id_number	= pdu->emitter_id_number;

	if (emitter_name == SP_DIS203_EMITTER_SYSTEM_ORSMC);
	
	else if (emitter_name == SP_DIS203_EMITTER_SYSTEM_MMCM || emitter_name == SP_DIS203_EMITTER_SYSTEM_MMCM)
		fpArea = new cRectangle (&ctr, &emitter_orient, 50, 50);



	fTargetEntity = something;

	state_update_indicator	= (SP_DIS203_EMISSION_PDU*)buf->state_update_indicator;
	number_of_systems 	= (SP_DIS203_EMISSION_PDU*)buf->number_of_systems;
	
	// part of SP_DIS203_EMISSION_SYSTEM_RECORD
	system_data_length	= (SP_DIS203_EMISSION_PDU*)buf->system_data_length;
	number_of_beams		= (SP_DIS203_EMISSION_PDU*)buf->number_of_beams;
	
	// part of SP_DIS203_EMITTER_SYSTEM_RECORD
	emitter_name		= (SP_DIS203_EMISSION_PDU*)buf->emitter_name;
	function		= (SP_DIS203_EMISSION_PDU*)buf->function;
	emitter_id_number	= (SP_DIS203_EMISSION_PDU*)buf->emitter_id_number;
	
	// part of SP_DIS203_EMISSION_BEAM_RECORD
	beam_data_length	= (SP_DIS203_EMISSION_PDU*)buf->beam_data_length;
	beam_id_number		= (SP_DIS203_EMISSION_PDU*)buf->beam_id_number;
	beam_parameter_index	= (SP_DIS203_EMISSION_PDU*)buf->beam_parameter_index;
	
	// part of SP_DIS203_FUNDAMENTAL_PARAMETER_DATA_RECORD
	frequency		= (SP_DIS203_EMISSION_PDU*)buf->frequency;
	frequency_range         = (SP_DIS203_EMISSION_PDU*)buf->frequency_range;
	pulse_repetition_frequency	= (SP_DIS203_EMISSION_PDU*)buf->pulse_repetition_frequency;
	pulse_width             = (SP_DIS203_EMISSION_PDU*)buf->pulse_width;
	beam_azimuth_center	= (SP_DIS203_EMISSION_PDU*)buf->beam_azimuth_center;
	beam_azimuth_sweep      = (SP_DIS203_EMISSION_PDU*)buf->beam_azimuth_sweep;
	beam_elevation_center   = (SP_DIS203_EMISSION_PDU*)buf->beam_elevation_center;
	beam_elevation_sweep    = (SP_DIS203_EMISSION_PDU*)buf->beam_elevation_sweep;
	beam_sweep_sync         = (SP_DIS203_EMISSION_PDU*)buf->beam_sweep_sync;

	// part of SP_DIS203_EMISSION_BEAM_RECORD
	beam_function		= (SP_DIS203_EMISSION_PDU*)buf->beam_function;
	number_of_targets_in_track_jam_field	= (SP_DIS203_EMISSION_PDU*)buf->number_of_targets_in_track_jam_field;
	high_density_track_jam	= (SP_DIS203_EMISSION_PDU*)buf->high_density_track_jam;
	jamming_mode_sequence	= (SP_DIS203_EMISSION_PDU*)buf->jamming_mode_sequence;
	
	
	// SP_DIS203_SIMULATION_ADDRESS_RECORD
	site_id			= (SP_DIS203_EMISSION_PDU*)buf->site_id;
	application_id		= (SP_DIS203_EMISSION_PDU*)buf->application_id;

	// part of SP_DIS203_ENTITY_ID_RECORD
	entity_id		= (SP_DIS203_EMISSION_PDU*)buf->entity_id;

	// part of SP_DIS203_TRACK_JAM_RECORD
	emitter_id		= (SP_DIS203_EMISSION_PDU*)buf->emitter_id;
	beam_id			= (SP_DIS203_EMISSION_PDU*)buf->beam_id;	

}
*/

cEmission::~cEmission()
{
	if (fpSystem) {
		delete fpSystem;
		fpSystem = NULL;
	}
/*
	if (fpArea) {
		delete fpArea;
		fpArea = NULL;
	}
*/

}




/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/*
       //TEntityID target_entity_id;
        //SP_DIS203_EVENT_ID_RECORD             event_id;
        //SP_DIS203_STATE_UPDATE_INDICATOR_TYPE state_update_indicator; //unsigned char
        //SP_DIS203_NUMBER_OF_SYSTEMS_TYPE number_of_systems; //unsigned char
        //SP_DIS203_EMISSION_SYSTEM_RECORD systems[1];
                //SP_DIS203_SYSTEM_DATA_LENGTH_TYPE         system_data_length; //unsigned char
                //SP_DIS203_NUMBER_OF_BEAMS_TYPE            number_of_beams; //unsigned char
                //SP_DIS203_EMITTER_SYSTEM_RECORD           emitter_system;
                        //SP_DIS203_EMITTER_NAME_TYPE      emitter_name; //unsigned short
                        //SP_DIS203_EMITTER_FUNCTION_TYPE  function;  //unsigned char
                        //SP_DIS203_EMITTER_ID_NUMBER_TYPE emitter_id_number; //unsigned char
                //SP_DIS203_ENTITY_COORDINATE_VECTOR_RECORD location_with_respect_to_entity;
                        //SP_DIS203_VECTOR_COMPONENT_TYPE x;
                        //SP_DIS203_VECTOR_COMPONENT_TYPE y;
                        //SP_DIS203_VECTOR_COMPONENT_TYPE z;
                //SP_DIS203_EMISSION_BEAM_RECORD            beams[1];
                        //SP_DIS203_BEAM_DATA_LENGTH_TYPE             beam_data_length; // unsigned char
                        //SP_DIS203_BEAM_ID_NUMBER_TYPE               beam_id_number; //unsigned char
                        //SP_DIS203_BEAM_PARAMETER_INDEX_TYPE         beam_parameter_index; //unsigned short
                        //SP_DIS203_FUNDAMENTAL_PARAMETER_DATA_RECORD fundamental_parameter_data;
                                //SP_DIS203_EMISSION_FREQUENCY_TYPE       frequency; //float
                                //SP_DIS203_FREQUENCY_RANGE_TYPE          frequency_range; //float
                                //SP_DIS203_EFFECTIVE_RADIATED_POWER_TYPE effective_radiated_power; //float
                                //SP_DIS203_EMISSION_FREQUENCY_TYPE       pulse_repetition_frequency; //float
                                //SP_DIS203_PULSE_WIDTH_TYPE              pulse_width; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_azimuth_center; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_azimuth_sweep; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_elevation_center; //float
                                //SP_DIS203_ANGLE_TYPE                    beam_elevation_sweep; //float
                                //SP_DIS203_BEAM_SWEEP_SYNC_TYPE          beam_sweep_sync; // float

                        //SP_DIS203_BEAM_FUNCTION_TYPE                beam_function; //unsigned char
                        //SP_DIS203_NUMBER_OF_TARGETS_TYPE            number_of_targets_in_track_jam_field; //unsigned char
                        //SP_DIS203_HIGH_DENSITY_TRACK_JAM_TYPE       high_density_track_jam; //unsigned char
                        //unsigned char                               _unused_42;  
                        //SP_DIS203_JAMMING_MODE_SEQUENCE_TYPE        jamming_mode_sequence; //unsigned long
                        //SP_DIS203_TRACK_JAM_RECORD                  track_jam[1];
                                //SP_DIS203_ENTITY_ID_RECORD       entity_id;
                                        //SP_DIS203_SIMULATION_ADDRESS_RECORD simulation_address;
                                                //SP_DIS203_SITE_ID_TYPE        site_id; //unsigned short
                                                //SP_DIS203_APPLICATION_ID_TYPE application_id; //unsigned short
                                        //SP_DIS203_ENTITY_ID_TYPE            entity_id; //unsigned short
                                //SP_DIS203_EMITTER_ID_NUMBER_TYPE emitter_id; //unsigned char
                                //SP_DIS203_BEAM_ID_NUMBER_TYPE    beam_id; //unsigned char






*/
/*
EBoolean cEmission::ProcessEmission(float HorzRange)
{
    if (HorzRange < radius)
       return true;
    else
       return false;  
}
*/
void cEmission::ProcEmissionMsg(cEmission* pEmiss)
{
//    radius = pEmiss->radius;
    fpVeh = pEmiss->fpVeh;
    
    if(fpSystem)  {
	delete fpSystem;
    }
    fNumSystems = pEmiss->fNumSystems;
    fpSystem = new TEmissionSystem[fNumSystems];
    
    fpSystem->fEmitterName =pEmiss->fpSystem->fEmitterName; 
    fpSystem->fFunction =pEmiss->fpSystem->fFunction;
    fpSystem->fEmitterID = pEmiss->fpSystem->fEmitterID;
    fpSystem->fRelEmitterLoc =pEmiss->fpSystem->fRelEmitterLoc;
    fpSystem->fERP = pEmiss->fpSystem->fERP;
/*
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    fpSystem->
    
*/    
    delete pEmiss;
    
}


TLocation  cEmission::GetDrTccLoc(TTimeSim fTime)
{
    TLocation emitter_loc;
    TLocation  VehLoc = fpVeh->GetDrTccLoc(fTime);
        
    TEntCoord  rel_emitter_loc = GetRelLoc();
//    float b = sqrt( (rel_emitter_loc.x*rel_emitter_loc.x) + (rel_emitter_loc.y*rel_emitter_loc.y));
//    float angle = GetTccOrient().psi - atan(rel_emitter_loc.y/rel_emitter_loc.x);

    float COS_PSI = fcos(GetTccOrient().psi);
    float SIN_PSI = fsin(GetTccOrient().psi);
    
//    emitter_loc.x = VehLoc.x + (b * cos(angle));
//    emitter_loc.y = VehLoc.y + (b * sin(angle));
//    emitter_loc.z = VehLoc.z;

    emitter_loc.x = VehLoc.x + (COS_PSI * rel_emitter_loc.x + SIN_PSI * rel_emitter_loc.y);
    emitter_loc.y = VehLoc.y + (SIN_PSI * rel_emitter_loc.x - COS_PSI * rel_emitter_loc.y);
    emitter_loc.z = VehLoc.z;

    return emitter_loc;
}

TLocation  cEmission::GetTccLoc()
{
    TLocation emitter_loc;
    TLocation  VehLoc = fpVeh->GetTccLoc();
        
    TEntCoord  rel_emitter_loc = GetRelLoc();
//    float b = sqrt( (rel_emitter_loc.x*rel_emitter_loc.x) + (rel_emitter_loc.y*rel_emitter_loc.y));
//    float angle = GetTccOrient().psi - atan(rel_emitter_loc.y/rel_emitter_loc.x);

    float COS_PSI = fcos(GetTccOrient().psi);
    float SIN_PSI = fsin(GetTccOrient().psi);
    
//    emitter_loc.x = VehLoc.x + (b * cos(angle));
//    emitter_loc.y = VehLoc.y + (b * sin(angle));
//    emitter_loc.z = VehLoc.z;

    emitter_loc.x = VehLoc.x + (COS_PSI * rel_emitter_loc.x + SIN_PSI * rel_emitter_loc.y);
    emitter_loc.y = VehLoc.y + (SIN_PSI * rel_emitter_loc.x - COS_PSI * rel_emitter_loc.y);
    emitter_loc.z = VehLoc.z;

    return emitter_loc;
}

TOrient  cEmission::GetTccOrient()
{
    return fpVeh->GetTccOrient();
}

void cEmissionMsg::SendMsg()
{
	gEmissionMgr.ForwardEmission(fpEmission);
}

/*
void    cEmission::Rotate ( float psi, float theta, float phi, T3Dsingle* pt )
{
   double sinpsi   = sin(psi);
   double cospsi   = cos(psi);
   double sintheta = sin(theta);
   double costheta = cos(theta);
   double sinphi   = sin(phi);
   double cosphi   = cos(phi);

   T3Ddouble x,y,z;

   x.x   =    costheta * cospsi;
   x.y   =    costheta * sinpsi;
   x.z   =   -sintheta;
  
   y.x   =   -cosphi * sinpsi    + sinphi * sintheta * cospsi;
   y.y   =    cosphi * cospsi    + sinphi * sintheta * sinpsi;
   y.z   =    sinphi * costheta;
  
   z.x   =    sinphi * sinpsi    + cosphi * sintheta * cospsi;
   z.y   =   -sinphi * cospsi    + cosphi * sintheta * sinpsi;
   z.z   =    cosphi * costheta;

   pt->x = x.x*pt->x + x.y*pt->y + x.z*pt->z;
   pt->y = y.x*pt->x + y.y*pt->y + y.z*pt->z;
   pt->z = z.x*pt->x + z.y*pt->y + z.z*pt->z;
}
*/


cEmission*   cEmissionMgr::FindEmission(const TEntityID & vID, const unsigned short feN, cEmissListHead* pListHead)
{
   cEmission* pEmiss;
   
   cEmissListHead* pHead = pListHead;
   if ( pHead == NULL ) pHead = &fEmissList;
   
   for ( pEmiss = pHead->GetFirstEmiss(); 
         pEmiss; 
         pEmiss = pHead->GetNextEmiss() )
      if (pEmiss->GetVehicle()->GetEntityID() == vID && pEmiss->GetEmitterName() == feN)
         return pEmiss;

   return NULL;
}

void   cEmissionMgr::EmissShutdown()
{
   fEmissList.DeleteList();
}

void cEmissionMgr::RemoveEmiss ( cEmission* pEmiss )      // remove Emission from list
{
   cLinkEmiss* pEmissLink;
   for ( pEmissLink  = (cLinkEmiss*) (fEmissList.GetFirstLink()); 
         pEmissLink != (cLinkEmiss*) (fEmissList.GetNULLLink()); 
         pEmissLink  = (cLinkEmiss*) (fEmissList.GetNextLink()) )
      if (pEmissLink->GetEmission() == pEmiss) {
         delete pEmissLink;
         break;
      }
}
 
 
void cEmissionMgr::ForwardEmission(cEmission* pEmiss)
{
   cEmission*   pExistingEm = FindEmission(pEmiss->GetVehicle()->GetEntityID(), pEmiss->GetEmitterName(), &fEmissList);
   
   if ( pExistingEm == NULL ) {
      if (pEmiss->GetERP() == 0.0) {
	  delete pEmiss;
      }

      else {
	new cLinkEmiss(pEmiss, (cLinkEmiss*)(fEmissList.GetLastLink()));   // adds at end
	//cout << "Adding new Emission." << endl;
      }
   }
   else if (pEmiss->GetERP() == 0.0) {
      RemoveEmiss(pExistingEm);
      cout << "Removing Emission." << endl;
   }
   else { // should check state change number here
      pExistingEm->ProcEmissionMsg(pEmiss);
      //cout << "Updating Emission." << endl;
   }
 

}  

cEmissionMgr::cEmissionMgr() :
      fEmissList()
{
}

cEmissionMgr::~cEmissionMgr()
{
   // should delete every Emission
}
