/*
$RCSfile: EntityDB.cc,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/04/29 19:34:07 $
$Log: EntityDB.cc,v $
Revision 1.2  1999/04/29 19:34:07  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:33  jchappel
Comprehensive Mine Simulator

// Revision 2.9  1998/11/25  18:37:51  ksnively
// Update for APLA Exercise
//
// Revision 2.8  1998/10/27  19:30:25  jchappel
//  modifications for versioning.
//
// Revision 2.7  1998/07/27  19:47:41  jchappel
// Made changes for DIS switch to 204.
//
// Revision 2.6  1998/07/27  17:50:42  ksnively
// Added single detection trip wire sensor
//
// Revision 2.5  1998/07/23  15:17:57  jchappel
// Made DIS switch changes
//
// Revision 2.4  1998/06/04  18:21:56  jchappel
// made DIS changes to parser
//
// Revision 2.3  1998/04/20  15:10:36  ksnively
// *** empty log message ***
//
// Revision 2.2  1998/02/26  21:24:37  ksnively
// Created Mf Camera object to control MAK Stealth view.
// Created vulnerability tables specific to command detonated mines
// and the camera object.
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.7  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.6  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.6  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.5  1997/05/20  15:35:11  ksnively
// Redraw screen when turning on or off minefield
//
// Revision 1.5  1997/05/20  15:35:11  ksnively
// Redraw screen when turning on or off minefield
//
// Revision 1.4  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.3  1997/03/20  19:14:41  ksnively
// added emission tracking, updated to CMS2.0B6
//
// Revision 1.2  1997/01/02  19:48:54  nsipes
// *** empty log message ***
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.4  1996/08/21  21:58:20  rcarp
// CP and QRP file output, other updates
//
// Revision 1.3  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.2  1996/06/22  13:04:11  rcarp
// deton vuln bug fixes for 1.5b4
//
// Revision 1.10  1996/05/23  18:42:07  mkappel
// Use DbMine list accessors.
//
// Revision 1.9  1996/05/22  15:58:40  rcarp
// fix for db mine list
//
// Revision 1.8  1996/05/21  22:15:40  rcarp
// text and lists from mine db, and activating mine state save cmd
//
// Revision 1.7  1996/05/16  00:01:58  rcarp
// miscellaneous bugs
//
// Revision 1.6  1996/05/15  17:18:44  rcarp
// refinements and fixes for 1.5 delivery
//
// Revision 1.5  1996/05/15  00:11:09  rcarp
// mostly wam and offroute fixes, also deleting selection
//
// Revision 1.4  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.1  1996/05/13  19:27:13  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: EntityDB.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/******************************************

   EntityDB.cc
   
      The entity type database manager
      implementation.
         
   Created by Richard Vuduc
   Modification history:
      June 1995: created   
      
*********************************************/

#include <iostream.h>
#include <stdio.h>
#include "Vehicle.h"
#include "Minefield.h"
#include "AtMine.h"
#include "ApMine.h"
#include "AtWideArea.h"
#include "FalseMine.h"
#include "EntityDB.h"
#include "SaSensor.h"
#include "MfCamera.h"
#include "CmsControl.h"

/**** entity database manager object class implementation ****/

const TKeywordEntry cDbEntry::fKeyEntryTable[eentrySizeTable] = {
   "entityTypeStruct",     ekeyLev2Struct,
   "entityType",     ekeyIntList,
   "entityName",     ekeyQuotedString
};

const TKeywordEntry cDbMine::fKeyMineTable[emineSizeTable] = {
   "entityTypeStruct",     ekeyLev2Struct,
   "entityType",     ekeyIntList,
   "mineParam",     ekeyLev2Struct,
   "fuzeParam",     ekeyLev2Struct,
   "sensorParam",     ekeyLev2Struct,
   "addMine",    ekeyLev2Struct
};

const TKeywordEntry TMineParam::fKeyMineParamTable[eSizeMineParamTable] = {
   "detonationTypeStruct",     ekeyLev2Struct,
   "detectionTypeStruct",      ekeyLev2Struct,
   "detonationTypeStruct204",  ekeyLev2Struct,
   "detectionTypeStruct204",   ekeyLev2Struct,
   "detonationType",     ekeyIntList,
   "detectionType",      ekeyIntList,
   "detonationType204",  ekeyIntList,
   "detectionType204",   ekeyIntList,
   "warheadType",        ekeyInt,
   "orientation",        ekey3DPoint,
   "meanThermChange",    ekeyFloat,
   "numRounds",          ekeyInt
};

const TKeywordEntry TProjParam::fKeyProjParamTable[eSizeProjParamTable] = {
   "entityType",         ekeyIntList,
   "detonationType",     ekeyIntList,
   "impactType",         ekeyIntList,
   "warheadType",        ekeyInt,
   "launchVel",          ekey2DPoint,
   "impactSpeed",        ekeyFloat
};

const char*   TMineParam::fpProjTextType[eNumProjTypes] = {
   "noProjectile",
   "WAM sublet"
};

const TKeywordEntry TFuzeParam::fKeyFuzeTable[efuzeSizeTable] = {
   "deleteFuzeType", ekeyString,
   "fuzeType",       ekeyString,
   "fuzeEnum",       ekeyUnsignedInt,
   "fireOffset",     ekey2DPoint,
   "dudProb",        ekeyFloat,
   "detonProb",      ekeyFloat,
   "armingDelay",    ekeyFloat,
   "emplaceDetProb", ekeyFloat,
   "emplaceDetTime", ekeyFloat,
   "detonResult",    ekeyUnsignedInt,
   "trackRange",     ekeyFloat,
   
   "selfDestTimes",     ekeyFloatList,
   "targetCounterOpt",  ekeyString,         // no, yes, choice
   "antihandlingOpt",   ekeyString,
   "blasthardOpt",      ekeyString
};

const TKeywordEntry cDbVeh::fKeyVehTable[evehSizeTable] = {
   "sizeParam",     ekeyLev2Struct
};

const TKeywordEntry TVehDimensions::fKeyVehSizeParamTable[eSizeVehSizeParamTable] = {
   "length",        ekeyFloat,
   "width",         ekeyFloat,
   "height",        ekeyFloat,
   "trackMin",      ekeyFloat,
   "trackMax",      ekeyFloat
};

const TKeywordEntry cDbProj::fKeyProjTable[eprojSizeTable] = {
   "entityTypeStruct",    ekeyLev2Struct,
   "entityType",          ekeyIntList,
   "projectileParam",     ekeyLev2Struct,
   "fuzeParam",           ekeyLev2Struct
};
      // instantiate the databases

   // we put the initial vehicle type hierarchy in InitializeVehDb()
   // since we do not  have to use static class data
cVehDBMgr       gVehDbMgr;
   // The projectile db constructor initializes the top level, to be ready
   // for the cWamSublet static entry
cProjDBMgr       gProjDbMgr;

      // projectiles are antiarmor munitions
TEntityType   cProjectile::fsProjEntityType( ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*   cProjectile::fsProjName = "Projectile";
int     cProjectile::fsDefaultFuzeType = TFuzeParam::efuzetypeTopSublet;
cDbProj_cProj     cProjectile::fsDbProj(    cProjectile::fsDefaultFuzeType,
                                       cProjectile::fsProjEntityType,
                                       cProjDBMgr::epjdblevKindDom, 
                                       cProjectile::fsProjName);

      // guided projectiles not implemented
/*
TEntityType   cProjectile::fsProjEntityType( ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*   cProjectile::fsProjName = "Projectile";
int     cProjectile::fsDefaultFuzeType = TFuzeParam::efuzetypeOther;
cDbProj_cProj     cProjectile::fsDbProj(    cProjectile::fsDefaultFuzeType,
                                       cProjectile::fsProjEntityType,
                                       cProjDBMgr::epjdblevCat, 
                                       cProjectile::fsProjName);
*/

      // wam sublets are ballistic
TEntityType   cWamSublet::fsWamSubEntityType( ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                             emuncatBallistic,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*   cWamSublet::fsWamSubletName = "WAM sublet";
int     cWamSublet::fsDefaultFuzeType = TFuzeParam::efuzetypeTopSublet;
cDbProj_Ballistic     cWamSublet::fsDbWamSublet(    cWamSublet::fsDefaultFuzeType,
                                       cWamSublet::fsWamSubEntityType,
                                       cProjDBMgr::epjdblevCat, 
                                       cWamSublet::fsWamSubletName);

      //      We put all the basic mine database objects here
      //      to ensure correct construction order
cMineDBMgr      gMineDbMgr;


      // mines are fixed munitions
TEntityType   cMine::fsMineEntityType(       ekindMunition,
                                kMatchAllChar,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*   cMine::fsMineName = "Mine";
int     cMine::fsDefaultFuzeType = TFuzeParam::efuzetypeOther;
cDbMine_cMine     cMine::fsDbMine(    cMine::fsDefaultFuzeType,
                                       cMine::fsMineEntityType,
                                       cMineDBMgr::emdblevMine, 
                                       cMine::fsMineName);

      // at mines are antiarmor fixed munitions
TEntityType   cAtMine::fsAtMineEntityType(   ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*         cAtMine::fsAtMineName = "Antiarmor";
char*         cAtMine::fsAtMineAltName = "AT";
int     cAtMine::fsDefaultFuzeType = TFuzeParam::efuzetypeOther;
cDbMine_cAtMine   cAtMine::fsDbAtMine(cAtMine::fsDefaultFuzeType,
                                       cAtMine::fsAtMineEntityType,
                                       cMineDBMgr::emdblevDomain, 
                                       cAtMine::fsAtMineName,
                                       cAtMine::fsAtMineAltName);
      // at conv
TEntityType   cAtConv::fsAtConvMineEntityType(ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eatscatConv,
                                kMatchAllChar,
                                kMatchAllChar);
char*         cAtConv::fsAtConvMineName = "AT Conventional";
char*         cAtConv::fsAtConvMineAltName = "Conventional";
int     cAtConv::fsDefaultFuzeType = TFuzeParam::efuzetypePressure;
cDbMine_cAtConv      cAtConv::fsDbAtConv(    cAtConv::fsDefaultFuzeType,
                                              cAtConv::fsAtConvMineEntityType,
                                              cMineDBMgr::emdblevSubcat, 
                                              cAtConv::fsAtConvMineName,
                                              cAtConv::fsAtConvMineAltName);
      // at scatterable
TEntityType   cAtScatterable::fsAtScatMineEntityType(ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eatscatScatter,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cAtScatterable::fsAtScatMineName = "AT Scatterable";
char*            cAtScatterable::fsAtScatMineAltName = "Scatterable";
int     cAtScatterable::fsDefaultFuzeType = TFuzeParam::efuzetypeMagnetic;
cDbMine_cAtScatterable     cAtScatterable::fsDbAtScatterable(cAtScatterable::fsDefaultFuzeType,
                                                cAtScatterable::fsAtScatMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cAtScatterable::fsAtScatMineName,
                                                cAtScatterable::fsAtScatMineAltName);
      // at off route
TEntityType   cAtOffRoute::fsAtOffrouteMineEntityType(ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eatscatOffroute,
                                kMatchAllChar,
                                kMatchAllChar);
char*         cAtOffRoute::fsAtOffrouteMineName = "AT Offroute";
char*         cAtOffRoute::fsAtOffrouteMineAltName = "Offroute";
int     cAtOffRoute::fsDefaultFuzeType = TFuzeParam::efuzetypeAcousticTrigger;
cDbMine_cAtOffRoute      cAtOffRoute::fsDbAtOffRoute(cAtOffRoute::fsDefaultFuzeType,
                                                cAtOffRoute::fsAtOffrouteMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cAtOffRoute::fsAtOffrouteMineName,
                                                cAtOffRoute::fsAtOffrouteMineAltName);
      // at wide area
TEntityType   cAtWideArea::fsAtWideareaMineEntityType(ekindMunition,
                                             emundomAntitank,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eatscatWidearea,
                                kMatchAllChar,
                                kMatchAllChar);
char*         cAtWideArea::fsAtWideareaMineName = "AT Wide Area";
char*         cAtWideArea::fsAtWideareaMineAltName = "Wide Area";
int     cAtWideArea::fsDefaultFuzeType = TFuzeParam::efuzetypeAcousticTracker;
cDbMine_cAtWideArea      cAtWideArea::fsDbAtWideArea(cAtWideArea::fsDefaultFuzeType,
                                                cAtWideArea::fsAtWideareaMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cAtWideArea::fsAtWideareaMineName,
                                                cAtWideArea::fsAtWideareaMineAltName);

      // ap mines are antipersonnel fixed munitions
TEntityType   cApMine::fsApMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*         cApMine::fsApMineName = "Antipersonnel";
char*         cApMine::fsApMineAltName = "AP";
int     cApMine::fsDefaultFuzeType = TFuzeParam::efuzetypeOther;
cDbMine_cApMine      cApMine::fsDbApMine(  cApMine::fsDefaultFuzeType,
                                                cApMine::fsApMineEntityType,
                                                cMineDBMgr::emdblevDomain, 
                                                cApMine::fsApMineName,
                                                cApMine::fsApMineAltName);
      // ap conventional
TEntityType   cApConv::fsApConvMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatConv,
                                kMatchAllChar,
                                kMatchAllChar);
char*         cApConv::fsApConvMineName = "AP Conventional";
char*         cApConv::fsApConvMineAltName = "Conventional";
int     cApConv::fsDefaultFuzeType = TFuzeParam::efuzetypeApPressure;
cDbMine_cApConv      cApConv::fsDbApConv(  cApConv::fsDefaultFuzeType,
                                                cApConv::fsApConvMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApConv::fsApConvMineName,
                                                cApConv::fsApConvMineAltName); 
      // ap scatterable
TEntityType   cApScatterable::fsApScatMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatScatter,
                                kMatchAllChar,
                                kMatchAllChar);
char*    cApScatterable::fsApScatMineName = "AP Scatterable";
char*    cApScatterable::fsApScatMineAltName = "Scatterable";
int     cApScatterable::fsDefaultFuzeType = TFuzeParam::efuzetypeApTrip;
cDbMine_cApScatterable         cApScatterable::fsDbApScatterable( cApScatterable::fsDefaultFuzeType,
                                                cApScatterable::fsApScatMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApScatterable::fsApScatMineName,
                                                cApScatterable::fsApScatMineAltName); 
      // ap conventional bounding
TEntityType   cApConvBding::fsApConvBdgMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatConvBounding,
                                kMatchAllChar,
                                kMatchAllChar);
char*    cApConvBding::fsApConvBdgMineName = "AP Conv Bounding";
char*    cApConvBding::fsApConvBdgMineAltName = "Conv Bounding";
int     cApConvBding::fsDefaultFuzeType = TFuzeParam::efuzetypeApPressure;
cDbMine_cApConvBding         cApConvBding::fsDbApConvBding( cApConvBding::fsDefaultFuzeType,
                                                cApConvBding::fsApConvBdgMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApConvBding::fsApConvBdgMineName,
                                                cApConvBding::fsApConvBdgMineAltName); 
      // ap scatterable bounding
TEntityType   cApScatBding::fsApScatBdgMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatScatBounding,
                                kMatchAllChar,
                                kMatchAllChar);
char*    cApScatBding::fsApScatBdgMineName = "AP Scat Bounding";
char*    cApScatBding::fsApScatBdgMineAltName = "Scat Bounding";
int     cApScatBding::fsDefaultFuzeType = TFuzeParam::efuzetypeApTrip;
cDbMine_cApScatBding         cApScatBding::fsDbApScatBding( cApScatBding::fsDefaultFuzeType,
                                                cApScatBding::fsApScatBdgMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApScatBding::fsApScatBdgMineName,
                                                cApScatBding::fsApScatBdgMineAltName); 
/*
      // ap wide area
TEntityType   cApWideArea::fsApWideAreaMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatScatWideArea,
                                kMatchAllChar,
                                kMatchAllChar);
char*    cApWideArea::fsApWideAreaMineName = "AP Wide Area";
char*    cApWideArea::fsApWideAreaMineAltName = "ApWAM";
int     cApWideArea::fsWideAreaDefaultFuzeType = TFuzeParam::efuzetypeCommandLoc;
cDbMine_cApWideArea         cApWideArea::fsDbApWideArea( cApWideArea::fsWideAreaDefaultFuzeType,
                                                cApWideArea::fsApWideAreaMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApWideArea::fsApWideAreaMineName,
                                                cApWideArea::fsApWideAreaMineAltName); 
*/

      // ap commanded
TEntityType   cApCommandDet::fsApCommandDetMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatScatCmdDet,
                                kMatchAllChar,
                                kMatchAllChar);
char*    cApCommandDet::fsApCommandDetMineName = "AP Cmd Det";
char*    cApCommandDet::fsApCommandDetMineAltName = "ApCD";
int     cApCommandDet::fsApCommandDetDefaultFuzeType = TFuzeParam::efuzetypeCommandFire;
cDbMine_cApCommandDet         cApCommandDet::fsDbApCommandDet( cApCommandDet::fsApCommandDetDefaultFuzeType,
                                                cApCommandDet::fsApCommandDetMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApCommandDet::fsApCommandDetMineName,
                                                cApCommandDet::fsApCommandDetMineAltName); 

      // ap tripwire and commanded
TEntityType   cApTripCommandDet::fsApTripCommandDetMineEntityType(ekindMunition,
                                             emundomAntipersonnel,
                                kMatchAllShort,
                                             emuncatFixed,
                                             eapscatScatTripCmdDet,
                                kMatchAllChar,
                                kMatchAllChar);
char*    cApTripCommandDet::fsApTripCommandDetMineName = "AP Tripwire/Cmd Det";
char*    cApTripCommandDet::fsApTripCommandDetMineAltName = "ApTripCD";
int     cApTripCommandDet::fsApTripCommandDetDefaultFuzeType = TFuzeParam::efuzetypeTripCommandFire;
cDbMine_cApTripCommandDet         cApTripCommandDet::fsDbApTripCommandDet( cApTripCommandDet::fsApTripCommandDetDefaultFuzeType,
                                                cApTripCommandDet::fsApTripCommandDetMineEntityType,
                                                cMineDBMgr::emdblevSubcat, 
                                                cApTripCommandDet::fsApTripCommandDetMineName,
                                                cApTripCommandDet::fsApTripCommandDetMineAltName); 


      // false alarms
TEntityType   cFalseMine::fsFalseMineEntityType(ekindMunition,
                                             efunnydomainFalseAlarm,
                                kMatchAllShort,
                                             emuncatFixed,
                                kMatchAllChar,
                                kMatchAllChar,
                                kMatchAllChar);
char*            cFalseMine::fsFalseMineName = "False Alarm";
char*            cFalseMine::fsFalseMineAltName = "FA";
int     cFalseMine::fsDefaultFuzeType = TFuzeParam::efuzetypeOther;
cDbMine_cFalseMine      cFalseMine::fsDbFalseMine( cFalseMine::fsDefaultFuzeType,
                                                cFalseMine::fsFalseMineEntityType,
                                                cMineDBMgr::emdblevDomain, 
                                                cFalseMine::fsFalseMineName,
                                                cFalseMine::fsFalseMineAltName); 


cEntityDBMgr::cEntityDBMgr( void )
{
   fpRoot = NULL;
   fRegCount = 0;
}

cEntityDBMgr::~cEntityDBMgr( void )
{
}

EBoolean cEntityDBMgr::Register( cDbEntry *newtype, int endlevel )
/*
 * Register( newtype )
 *   Adds an entity descriptor 'newtype' into the database/search tree.
 * Increments the current database register count if node sucessfully added.
 */
{
   EBoolean retval;
   
   if (fpRoot) {
      retval = Insert( newtype, fpRoot, 1, endlevel );
   }
   else {
      fpRoot = newtype;
      fRegCount = 1;
      retval = true;
   }
   
   return retval;
}

cDbEntry* cEntityDBMgr::FindByName(const char* pName)
{
   if( ( ! fpRoot ) || ( ! pName ) )
      return NULL;

   return RecFindName( pName, fpRoot );
}

/*
 * MatchAll( attribs, retbuf, count )
 *   Searches the registration tree for entities matching the attributes
 * given by 'attribs.'  Returns a list of pointers to entity descriptors
 * in 'retbuf.'  on return, 'count' contains the number of items in
 * 'retbuf.'
 *
 * Returns false if no items were found, true otherwise.
 */
EBoolean cEntityDBMgr::MatchAll( const TEntityType& enttype,
   cDbEntry ***retbuf, int *count ) const
{
      // allocate array of descriptor pointers
   *retbuf = new cDbEntry *[fRegCount];
   *count = 0;

   if( !fpRoot )
      return false;

   return RecMatchAll( enttype, retbuf, count, fpRoot, 1 );
}

EBoolean cEntityDBMgr::MatchAllLevel( int matchLevel, const TEntityType& enttype,
   cDbEntry ***retbuf, int *count ) const
      // finds all entries at the specified level, that match the entity type
      // defined to that level, with wildcards allowed as matches
{
      // allocate array of descriptor pointers
   *retbuf = new cDbEntry *[fRegCount];
   *count = 0;

   if( !fpRoot )
      return false;

   return RecMatchAllLevel( enttype, retbuf, count, fpRoot, 1,  matchLevel);
}

EBoolean cEntityDBMgr::GetNextLevel( cDbEntry* pParent, const TEntityType& filterType,
                     cDbEntry ***retbuf, int *count ) const
      // finds all entries at the next level from parent
      // that match the entity type
{
   if( !pParent )
      return false;
      
      // allocate array of descriptor pointers
   *retbuf = new cDbEntry *[fRegCount];
   *count = 0;

   int level = pParent->GetDbLevel();
   return RecMatchAllLevel( filterType, retbuf, count, pParent, level,  level+1, false);
}

EBoolean cEntityDBMgr::GetDescendentLeaves( cDbEntry* pParent, const TEntityType& filterType,
                     cDbEntry ***retbuf, int *count ) const
      // finds all leaf nodes descending from parent
      // that match the filter type
{
   if( !pParent )
      return false;

      // allocate array of descriptor pointers
   *retbuf = new cDbEntry *[fRegCount];
   *count = 0;
      
   int level = pParent->GetDbLevel();
   return RecMatchAll( filterType, retbuf, count, pParent, level, false);
}

EBoolean  cEntityDBMgr::GetCountryList( cDbEntry* pParent,
                     int **retbuf, int *count ) const
{
   if( !pParent )
      return false;

   *retbuf = new int [enumCountry];
   *count = 0;
   return RecGetCountryList( pParent, retbuf, count, false );
}

EBoolean  cEntityDBMgr::RecGetCountryList( cDbEntry* pParent,
                     int **retbuf, int *count, EBoolean bGoSib ) const
{
   unsigned short country = pParent->GetEntryEntityType().country;
   
   if ( country != kMatchAllShort ) {
      EBoolean match = false;
      int index = 0;   // insert before this one
      for ( int i = 0; i < *count; i++ ) {
         if ( (int)country == (*retbuf)[i] ) {  // already have this country
            match = true;
            break;
         }
         else if ( (int)country < (*retbuf)[i] ) {
            index = i;
            break;
         }
      }
      if ( ! match ) {
         Assert( *count < enumCountry, 
                  kerrValidityCheck, "cEntityDBMgr::RecGetCountryList");
         for ( int i = *count - 1; i >= index; i-- )
            (*retbuf)[i+1] = (*retbuf)[i];
         (*retbuf)[index] = country;
         ++ (*count);
      }
   }
   
   if ( pParent->fpChildEntry != NULL )
      RecGetCountryList(pParent->fpChildEntry, retbuf, count);
   
   if ( bGoSib && (pParent->fpNextEntry != NULL) )
      RecGetCountryList(pParent->fpNextEntry, retbuf, count);
      
   return true;
}

/*
 * MatchClosestOne( const TEntityType& type, int endlevel )
 *   Finds and returns one entity descriptor which represents the
 * closest match to "type."  Search proceeds until "endlevel."
 * The closest match will be at or just above 'endlevel.'
 *
 * Returns NULL if nothing could be found.
 */

cDbEntry* cEntityDBMgr::MatchClosestOne(const TEntityType& type)
{
   if( !fpRoot )
      return NULL;

   return RecMatchClose( type, fpRoot, 1 );
}

/*
 * Insert( newnode, curnode, level, endlevel )
 *   Tries to place 'newnode' into the tree at level 'endlevel' beginning at
 * 'curnode.'  'level' indicates the level of 'curnode' in the tree.
 */
EBoolean cEntityDBMgr::Insert( cDbEntry *newnode, cDbEntry *curnode,
                               int level, int endlevel )
{

   cDbEntry *nextnode;
   int nextlevel = level;

   if( ! IsMatch(curnode->GetEntryEntityType(), newnode->GetEntryEntityType(), level) )
   {
         // if not a match, go to sibling or else this should be sibling
      if( curnode->GetNextEntry() != NULL )
         nextnode = curnode->GetNextEntry();
      else {
         Assert ( level == endlevel, kerrValidityCheck, "cEntityDBMgr::Insert/1");
         curnode->fpNextEntry = newnode;
         newnode->fpPrevEntry = curnode;
         fRegCount++;
         return true;
      }
   }
   else
   {
         // types match.  Go to next level, if at last level this 
         // is a replacement
      if( level == endlevel )
      {
         ReplaceNode( curnode, newnode );   // no fRegCount increment
         return true;
      }

      nextlevel = nextlevel + 1;
      if( curnode->fpChildEntry != NULL )
         nextnode = curnode->fpChildEntry;
      else {
            // first at new level, must be only one level below
            // current depth
         Assert( nextlevel == endlevel, kerrValidityCheck, "cEntityDBMgr::Insert/2");
         curnode->fpChildEntry = newnode;
         newnode->fpPrevEntry = curnode;
         fRegCount++;
         return true;
      }
   }

   return Insert( newnode, nextnode, nextlevel, endlevel );
}

cDbEntry* cEntityDBMgr::RecFindName(const char* pName, cDbEntry* curnode) const
{
   cDbEntry* retval;

   if( ( curnode->GetEntryEntityName() ) &&
       ( ! strcmp(pName, curnode->GetEntryEntityName() ) ) )
      retval = curnode;
      
   else {
      if( curnode->fpChildEntry != NULL ) {
            // try to go deeper
         retval = RecFindName( pName, curnode->fpChildEntry );
         if( retval != NULL )
            return retval;
      }
      if( curnode->fpNextEntry )
         retval = RecFindName( pName, curnode->fpNextEntry );
      else 
         retval = NULL;
   }

   return retval;
}

/*
 * RecMatchAll( attribs, retbuf, count, curnode, level )
 *   Recursive match routine.  Returns in retbuf a list of pointers
 * to all nodes matching 'attribs' beginning in search tree at 'curnode.'
 * 'level' indicates the level (elevXXX) of 'curnode' in the search tree.
 *
 * assumes curnode != null
 */

EBoolean cEntityDBMgr::RecMatchAll( const TEntityType& enttype,
   cDbEntry ***retbuf, int *count, cDbEntry *curnode,
   int level, EBoolean bGoSib ) const
{
   // return values; true if a node was successfully added
   // at curnode, below curnode, or next to curnode, respectively
   EBoolean thisRetval, depthRetval, breadthRetval;

   // base case: we've reached actually entities in the tree
   if ( ! curnode->IsAbstract() )
   {
      if ( IsMatch(enttype, curnode->GetEntryEntityType(), level) ) {
         (*retbuf)[(*count)] = curnode;
         ++(*count);
         thisRetval = true;   // found at least one
      }
   }
   else // otherwise, must try to descend
   {
      thisRetval = false;

      // if we match at this level, try to search below
      if( IsMatch(enttype, curnode->GetEntryEntityType(), level) && 
                 (curnode->fpChildEntry != NULL) ) {
         depthRetval = RecMatchAll( enttype, retbuf, count,
            curnode->fpChildEntry, level+1 );
      }
      else 
         depthRetval = false;
   }

   // try siblings
   if( bGoSib && (curnode->fpNextEntry) )
   {
      breadthRetval = 
         RecMatchAll( enttype, retbuf, count, curnode->fpNextEntry, level );
   }
   else 
      breadthRetval = false;

   EBoolean retval = (thisRetval || breadthRetval || depthRetval) ? true : false;
   return retval;
}

EBoolean cEntityDBMgr::RecMatchAllLevel( const TEntityType& enttype,
   cDbEntry ***retbuf, int *count, cDbEntry *curnode,
   int level, int matchLevel, EBoolean bGoSib ) const
{
   // return values; true if a node was successfully added
   EBoolean thisRetval, depthRetval, breadthRetval;

   // if we have reached the level, check and add if match
   if ( level >=  matchLevel)
   {
      if ( IsMatch(enttype, curnode->GetEntryEntityType(), level) ) {
         (*retbuf)[(*count)] = curnode;
         ++(*count);
         thisRetval = true;   // found at least one
      }
   }
   else // otherwise, must try to descend
   {
      thisRetval = false;

      // if we match at this level, try to search below
      if( IsMatch(enttype, curnode->GetEntryEntityType(), level) && 
                 (curnode->fpChildEntry != NULL) ) {
         depthRetval = RecMatchAllLevel( enttype, retbuf, count,
            curnode->fpChildEntry, level+1, matchLevel );
      }
      else 
         depthRetval = false;
   }

   // try siblings
   if( bGoSib && (curnode->fpNextEntry) )
   {
      breadthRetval = 
         RecMatchAllLevel( enttype, retbuf, count, curnode->fpNextEntry,
               level, matchLevel );
   }
   else 
      breadthRetval = false;

   EBoolean retval = (thisRetval || breadthRetval || depthRetval) ? true : false;
   return retval;
}

// assumes curnode != NULL

cDbEntry* cEntityDBMgr::RecMatchClose( const TEntityType& type,
                                 cDbEntry* curnode, int level )
{
   cDbEntry* retval;

   if( IsMatch(type, curnode->GetEntryEntityType(), level) ) {
      if( curnode->fpChildEntry != NULL ) {
            // try to go deeper
         retval = RecMatchClose( type, curnode->fpChildEntry, level+1 );
         if( retval == NULL )
            retval = curnode;
      }
      else 
         retval = curnode;  // we're the (closest) match!
   }
   else {
      if( curnode->fpNextEntry )
         retval = RecMatchClose( type, curnode->fpNextEntry, level );
      else 
         retval = NULL;
   }

   return retval;
}

cDbEntry* cEntityDBMgr::MatchExact( const TEntityType& type )
{
   cDbEntry* pEntry = RecMatchClose( type, fpRoot, 1 );
   
   if ( type == pEntry->GetEntryEntityType() )
      return pEntry;
   else
      return NULL;
}

/*
 * IsMatch( t1, t2, level, endlevel )
 *   If t1 is equivalent to t2 (not necessarily exactly equal, may include
 * wild-card type values like kMatchAllChar, etc.), returns true.  Compares
 * relevant fields only up to 'level'.
 */
EBoolean cEntityDBMgr::IsMatch(
   const TEntityType&, const TEntityType&, int) const
{
   return false;
}

int      cEntityDBMgr::MaxLevel()
{
   return 1;
}

void     cEntityDBMgr::ReplaceNode( cDbEntry* oldnode,
                                    cDbEntry* newnode )
{
   newnode->fpChildEntry = oldnode->fpChildEntry;
   newnode->fpNextEntry = oldnode->fpNextEntry;
   newnode->fpPrevEntry = oldnode->fpPrevEntry;

   Assert ( oldnode->fpPrevEntry != NULL, kerrValidityCheck, "cEntityDBMgr::ReplaceNode/1");

   if( oldnode->fpPrevEntry )
   {
      if( oldnode->fpPrevEntry->fpChildEntry == oldnode )
         oldnode->fpPrevEntry->fpChildEntry = newnode;
      else if( oldnode->fpPrevEntry->fpNextEntry == oldnode )
         oldnode->fpPrevEntry->fpNextEntry = newnode;
      else {
         Assert ( false, kerrValidityCheck, "cEntityDBMgr::ReplaceNode/2");
      }
   }

   if( oldnode->fpNextEntry )
   {
      Assert ( oldnode->fpNextEntry->fpPrevEntry == oldnode, kerrValidityCheck, "cEntityDBMgr::ReplaceNode/3");
      oldnode->fpNextEntry->fpPrevEntry = newnode;
   }

   if( oldnode->fpChildEntry )
   {
      Assert ( oldnode->fpChildEntry->fpPrevEntry == oldnode, kerrValidityCheck, "cEntityDBMgr::ReplaceNode/44");
      oldnode->fpChildEntry->fpPrevEntry = newnode;
   }

   oldnode->fpChildEntry = NULL;
   oldnode->fpNextEntry = NULL;
   oldnode->fpPrevEntry = NULL;
   delete oldnode;
}

EBoolean cMineDBMgr::IsMatch(
   const TEntityType& t1, const TEntityType& t2, int level) const
{
      // matches entity type fields, according to mine hierarchy levels,
      // up to a specific level.
   switch( level )
   {               // NOTE fall through strategy to match all fields
                   // up to the needed level
//      case emdblevExtra: 
      case emdblevEntity:
            if ( ! t1.MatchExtra(t2) ) return false;
      
      case emdblevSpecific:   // includes matching both spec and country 
            if ( ! ( (t1.MatchSpecific(t2)) &&  (t1.MatchCountry(t2))) ) return false;
 
      case emdblevSubcat:
            if ( ! t1.MatchSubcategory(t2) ) return false;
 
      case emdblevDomain:
            if ( ! t1.MatchDomain(t2) ) return false;

      case emdblevMine:   // includes matching both kind and category 
            if ( ! ( (t1.MatchKind(t2)) &&  (t1.MatchCategory(t2))) ) return false;

         break;

      default:
            return false; 
   }

   return true;
}

int      cMineDBMgr::MaxLevel()
{
   return emdblevEntity;
}

cMineDBMgr::cMineDBMgr()
{
}

cMineDBMgr::~cMineDBMgr()
{
}

EBoolean cVehDBMgr::IsMatch(
   const TEntityType& t1, const TEntityType& t2, int level) const
{
      // matches entity type fields, according to vehicle hierarchy levels,
      // up to specific level.
   switch( level )
   {               // NOTE fall through strategy to match all fields
                   // up to the needed level
//      case evdblevSpecific: 
      case evdblevEntity:
            if ( ! t1.MatchSpecific(t2) ) return false;
      
      case evdblevSubcat:   // includes matching both spec and country 
            if ( ! ( (t1.MatchSubcategory(t2)) &&  (t1.MatchCountry(t2))) ) return false;
 
      case evdblevCat:
            if ( ! t1.MatchCategory(t2) ) return false;
 
      case evdblevDomain:
            if ( ! t1.MatchDomain(t2) ) return false;

      case evdblevPlatform:   // includes matching both kind and category 
            if ( ! t1.MatchKind(t2)  ) return false;

         break;

      default:
            return false; 
   }

   return true;
}

int      cVehDBMgr::MaxLevel()
{
   return evdblevEntity;
}

void    cVehDBMgr::InitializeVehDb()
{
   TEntityType type( ekindPlatform,
                        kMatchAllChar,
                        kMatchAllShort,
                        kMatchAllChar,
                        kMatchAllChar,
                        kMatchAllChar,
                        kMatchAllChar  );
   TVehDimensions tankDim(7.32, 3.56, 2.33, 1.35, 1.75);
   
   new cDbVeh(type, evdblevPlatform, "Platform", &tankDim);
   
   type.domain = eplatdomLand;
	new cDbVeh(type, evdblevDomain, "Land_Platform", &tankDim);
   
   type.domain = eplatdomAir;
   TVehDimensions airDim(10.0, 2.0, 2.0, 0.0, 0.0);
	new cDbVeh(type, evdblevDomain, "Air_Platform", &airDim);
   
   type.domain = eplatdomAir;
   type.category = eplatairUAVRecon;
   TVehDimensions uavDim(10.0, 2.0, 2.0, 0.0, 0.0);
	new cDbVeh(type, evdblevCat, "UAV_Recon", &uavDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandTank;
	new cDbVeh(type, evdblevCat, "Tank", &tankDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandArmorFightVeh;
   TVehDimensions afvDim(5.0, 2.6, 2.0, 1.0, 1.3);
	new cDbVeh(type, evdblevCat, "AFV", &afvDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandArmorUtilVeh;
   TVehDimensions arutilDim(5.0, 2.6, 2.0, 1.0, 1.3);
	new cDbVeh(type, evdblevCat, "ArmUtilVeh", &arutilDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandSelfPropArty;
   TVehDimensions spartyDim(5.0, 2.6, 2.0, 1.0, 1.3);
	new cDbVeh(type, evdblevCat, "SpArty", &spartyDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandTowedArty;
   TVehDimensions towartyDim(5.0, 2.6, 2.0, 1.0, 1.3);
	new cDbVeh(type, evdblevCat, "TowedArty", &towartyDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandSmWheelUtil;
   TVehDimensions swheelDim(4.0, 2.0, 2.0, 0.8, 1.0);
	new cDbVeh(type, evdblevCat, "SmallWheelUtil", &swheelDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandLgWheelUtil;
   TVehDimensions lgwheelDim(5.0, 2.6, 2.0, 1.0, 1.3);
	new cDbVeh(type, evdblevCat, "LargeWheelUtil", &lgwheelDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandSmTrkUtil;
   TVehDimensions strackDim(4.0, 2.0, 2.0, 0.8, 1.0);
	new cDbVeh(type, evdblevCat, "SmallTrackedUtil", &strackDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandLgTrkUtil;
   TVehDimensions lgtrackDim(5.0, 2.6, 2.0, 1.0, 1.3);
	new cDbVeh(type, evdblevCat, "LargeTrackedUtil", &lgtrackDim);
   
   type.domain = eplatdomLand;
   type.category = eplatlandMortar;
   TVehDimensions mortarDim(4.0, 2.0, 2.0, 0.8, 1.0);
	new cDbVeh(type, evdblevCat, "Mortar", &mortarDim);

   type.domain = eplatdomLand;
   type.category = eplatlandOther;
   TVehDimensions emissDim(0.0, 20.0, 10.0, 0.0, 20.0);
	new cDbVeh(type, evdblevCat, "Emission", &emissDim);


   type.kind = ekindLifeform;
   type.domain = kMatchAllChar;
   type.category = kMatchAllChar;
   TVehDimensions persDim(0.3, 0.5, 2.0, 0.4, 0.5);
   
   new cDbVeh(type, evdblevPlatform, "LifeForm", &persDim);
   
   type.domain = eplatdomLand;
	new cDbVeh(type, evdblevDomain, "Land_LifeForm", &persDim);
   
   type.category = elifelandOther;
	new cDbVeh(type, evdblevCat, "OtherLife", &persDim);
   
   type.category = elifelandDI;
	new cDbVeh(type, evdblevCat, "DILife", &persDim);

}

cVehDBMgr::cVehDBMgr()
{
    InitializeVehDb();
}

cVehDBMgr::~cVehDBMgr()
{
}

EBoolean cProjDBMgr::IsMatch(
   const TEntityType& t1, const TEntityType& t2, int level) const
{
      // matches entity type fields, according to mine hierarchy levels,
      // up to a specific level.
   switch( level )
   {               // NOTE fall through strategy to match all fields
                   // up to the needed level
      
      case epjdblevSpecific:
            if ( ! t1.MatchSpecific(t2) ) return false;
 
      case epjdblevSubcat:   // includes matching both subcat and country 
            if ( ! ( t1.MatchSubcategory(t2)&&  (t1.MatchCountry(t2)) ) ) return false;
 
      case epjdblevCat:
            if ( ! t1.MatchCategory(t2) ) return false;

      case epjdblevKindDom:   // includes matching both kind and category 
            if ( ! ( (t1.MatchKind(t2)) &&  (t1.MatchDomain(t2))) ) return false;

         break;

      default:
            return false; 
   }

   return true;
}

int      cProjDBMgr::MaxLevel()
{
   return epjdblevEntity;
}

cProjDBMgr::cProjDBMgr()
{
}

cProjDBMgr::~cProjDBMgr()
{
}

EBoolean         TParamList::ReplaceParam(cDbStruct* pOldParam, cDbStruct* pParam)
{
   for ( int i = 0; i < fNumParam; i++) {
      if ( fppParamList[i] == pOldParam ) {
         delete pOldParam;
         fppParamList[i] = pParam;
         return true;
      }
   }
   
   return false;
}

void         TParamList::SetList(cDbStruct* pParam)
{
   for ( int i = 0; i < fNumParam; i++)
      delete fppParamList[i];

   fNumParam = 1;
   fppParamList[0] = pParam;
}

void         TParamList::AddParam(cDbStruct* pParam)
{
   if ( pParam ) {
      if ( fNumParam >= (fNumBlocks * eBlockSize) ) {  // need to add a block
         cDbStruct** ppNewList = new cDbStruct*[++fNumBlocks * eBlockSize];
         for ( int i = 0; i < fNumParam; i++)
            ppNewList[i] = fppParamList[i];
         delete[] fppParamList;
         fppParamList = ppNewList;
      }
      
      fppParamList[fNumParam++] = pParam;
   }
}

cDbStruct*         TParamList::GetParam(int opt)
{
   if ( (opt < fNumParam) && (opt >= 0) )
      return fppParamList[opt];
   else
      return NULL;
}

EBoolean         TParamList::RemoveParam(int opt)
{
   if ( (opt < 0) || (opt >= fNumParam) )
      return false;
   
   delete fppParamList[opt];
   
   for ( int i = opt; i < fNumParam-1; i++)
      fppParamList[i] = fppParamList[i+1];
      
   --fNumParam;
   return true;
}

TParamList::TParamList(cDbStruct* pParam)
{
   fNumBlocks = 1;
   fppParamList = new cDbStruct*[fNumBlocks * eBlockSize];
   
   fppParamList[0] = pParam;
   if ( pParam )
      fNumParam = 1;
   else
      fNumParam = 0;
}

TParamList::~TParamList()
{
   for ( int i = 0; i < fNumParam; i++)
      delete fppParamList[i];
   delete[] fppParamList;
}

EBoolean   cDbStruct::ParseDbStructItem(cInFileObj&)
{
   return false;
}

EBoolean     cDbStruct::ReadDbStruct(cInFileObj& infile)
   // reads each parameter into buffer, passes it to parser to 
   // interpret it.
{
	EBoolean done = false;
	EBoolean success = true;
	do {
            // NextItem() stores an item for us to look at
		EFileReturn returnCode = infile.NextItem();
		switch( returnCode )
		{
			case eLevelDone:      // done with second level struct
//				cout << "level done!" << endl;
				done = true;
				break;

			case eFileNoErr:
			        if (!(ParseDbStructItem(infile)))
				    success = false;  //must be successful at all levels
				break;

			default: // error getting next token
				Assert(false, kerrReadParamfile, "cDbStruct::ReadDbStruct");
				return false;
		}
	} while( !done );

	return success;	
}

cDbStruct::~cDbStruct()
{ }

EBoolean     TMineParam::ParseDbStructItem(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses the one statement.
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TItemList *    pList;

   returnCode = infile.ParseItem(fKeyMineParamTable, eSizeMineParamTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
		case eNextLevel:

		  switch( returnIndex )
			{
		  case empkwDetonType:
		    //if( gCmsMgr.fStartupParam.feProtocol == /*EProtocol:: */eDis203 )
                        fDetonationType.ReadEntityStruct( infile );
                     break;

		  case empkwDetectType:
		    //if( gCmsMgr.fStartupParam.feProtocol == /*EProtocol:: */eDis203 )
                        fDetectionType.ReadEntityStruct( infile );
                     break;

 		  case empkwDetonType204:
		     if( gCmsMgr.fStartupParam.feProtocol == /*EProtocol:: */eDis204 )
                        fDetonationType.ReadEntityStruct( infile );
                     break;

		  case empkwDetectType204:
		     if( gCmsMgr.fStartupParam.feProtocol == /*EProtocol:: */eDis204 )
                        fDetectionType.ReadEntityStruct( infile );
                     break;


		  default: // ??
                     return false;
                     //	break;
			}
		  
         break;

		case eFileNoErr:
		  if( gCmsMgr.fStartupParam.feProtocol == /* EProtocol*/eDis203 ){


			switch( returnIndex )
			{
				case empkwDetonTypeList:
                  pList = (TItemList *)returnVal;
                  fDetonationType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				case empkwDetectTypeList:
                  pList = (TItemList *)returnVal;
                  fDetectionType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;
				default: // what?
					break;
			}
		  } // end if
		  else {

			switch( returnIndex )
			{
				case empkwDetonTypeList204:
                  pList = (TItemList *)returnVal;
                  fDetonationType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				case empkwDetectTypeList204:
                  pList = (TItemList *)returnVal;
                  fDetectionType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
		  } // end else

			switch( returnIndex )
			{
				case empkwWarhead:
                  fWarhead = (EWarhead) (*(int*)returnVal);
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case empkwNumRounds:
                  fNumRounds = *(int*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case empkwOrientation:
                  fTccOrient.psi   = ((T3Ddouble*)returnVal)->x;
                  fTccOrient.theta = ((T3Ddouble*)returnVal)->y;
                  fTccOrient.phi   = ((T3Ddouble*)returnVal)->z;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case empkwTimeTherm:
                  fMeanTimeThermal = (*(float*)returnVal);
                  delete returnVal;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return false;

		default:
			Assert(false, kerrReadParamfile, "TMineParam::ParseDbStructItem");
			return false;
			// error, error...
   }
   //   cout << fDetectionType.kind << " " << fDetectionType.domain << " " << fDetectionType.country << " " << fDetectionType.category << " " << fDetectionType.subcategory << " " << fDetectionType.specific << " " << fDetectionType.extra << endl;

   return true;
}

void    TMineParam::WriteMineOpt(ofstream* file)
{
   *file << "detonation " ;
   fDetonationType.WriteEntityType(file);
   *file << "detection " ;
   fDetectionType.WriteEntityType(file);
   *file << "warhead " << (int)fWarhead << endl;
   *file << "num rounds " << fNumRounds << endl;
   *file << "orientation ps-t-ph " << fTccOrient.psi << " "
                                   << fTccOrient.theta << " "
                                   << fTccOrient.phi << endl;
}

TMineParam::TMineParam(TEntityType& type) :
   fTccOrient(0.0, kHalfPi, 0.0)
{
   fDetonationType = type;
   fDetectionType  = type;
   fWarhead = ewhMine;
   fNumRounds = 1;
   fPaintScheme = 0;
   fMeanTimeThermal = 15.0;   //minutes
}

TMineParam::TMineParam() :
   fTccOrient(0.0, kHalfPi, 0.0)
{
      // entity types default to all zeros
   fWarhead = ewhMine;
   fPaintScheme = 0;
   fMeanTimeThermal = 15.0;   //minutes
}

TMineParam::~TMineParam()
{
}

int    TFuzeParam::GetTypeIndex(char* pTypeName)
{
    return KeyIndexLookup(fpFuzeTextType, eNumFuzeTypes, pTypeName);
}

EBoolean     TFuzeParam::ParseDbStructItem(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses the one statement.
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TItemList *    pList;
   int i;

   returnCode = infile.ParseItem(fKeyFuzeTable, efuzeSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case efuzekwDeleteType:
   					i = KeyIndexLookup( gpYesNoText, kYesNoPairs,
   						 (char *)returnVal );
   					if( i < 0 ) return false;
                  
                  if ( (EYesNoOpt) i == eYes )
                     fFuze = (EFuze)(-1);      // cannot be overwritten by input enum value
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwFuzeTypeName:
                     // read name, look up index
   					fTypeNameIndex = KeyIndexLookup( fpFuzeTextType, eNumFuzeTypes,
   						 (char *)returnVal );
   					if( fTypeNameIndex < 0 ) {
                     delete[] returnVal;
                     returnVal = NULL;
                     return false;
                  }
                     
                  fFuze = (EFuze) (fpFuzeTextType[fTypeNameIndex].value);
               
                  delete[] returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwFuzeTypeEnum:
                  if ( fFuze != (EFuze)(-1) )   // don't overwrite a deletion
                     fFuze =  (EFuze) ( *(int*)returnVal );
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwFireOffset:
                  fFireOffset =  *(TFlatEntCoord*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwProbDud:
                  fProbDud = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwProbDeton:
                  fProbDeton = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwArmDelay:
                  fArmTimeHrs = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwProbEmplaceDet:
                  fProbEmplaceDet = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwTimeEmplaceDet:
                  fTimeEmplaceDet = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwDetonResult:
                  fDetonResult =  (EDetonResult) ( *(int*)returnVal );
                  delete returnVal;
		  cout << "setting deton result: " << fDetonResult << endl;
                  returnVal = NULL;
 					break;

				case efuzekwSelfDestDelays:
                  pList = (TItemList *)returnVal;
                  fpOptSelector->SetSelfDestOpts(pList);
                  delete[] (float *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				case efuzekwTgtCountOpt:
   					i = KeyIndexLookup( gpYesNoText, kYesNoPairs,
   						 (char *)returnVal );
   					if( i < 0 ) return false;
                  
                  fpOptSelector->SetTgtCntOpt( (EYesNoOpt) i);
                  delete[] returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwAntiHandlingOpt:
   					i = KeyIndexLookup( gpYesNoText, kYesNoPairs,
   						 (char *)returnVal );
   					if( i < 0 ) return false;
                  
                  fpOptSelector->SetAntihanOpt( (EYesNoOpt) i);
                  delete[] returnVal;
                  returnVal = NULL;
 					break;

				case efuzekwSelfBlastHardOpt:
   					i = KeyIndexLookup( gpYesNoText, kYesNoPairs,
   						 (char *)returnVal );
   					if( i < 0 ) return false;
                  
                  fpOptSelector->SetBlastHardOpt( (EYesNoOpt) i);
                  delete[] returnVal;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return false;

		default:
			Assert(false, kerrReadParamfile, "TMineParam::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

EBoolean TFuzeParam::ValidParam()
{
   if ( fFuze > efzOther )
      return true;
   else
      return false;
}

void    TFuzeParam::WriteFuzeOpt(ofstream* file)
{
   *file << "fuze type index " << fTypeNameIndex << endl;
   *file << "fuze type enumerator " << (int)fFuze << endl;
   *file << "fire offset x y " << fFireOffset.x << " "
                                   << fFireOffset.y << endl;
   *file << "dud probability " << fProbDud << endl;
   *file << "detonation prob " << fProbDeton << endl;
   *file << "arm time (hrs) " << fArmTimeHrs << endl;
   *file << "detonating on emplacement prob " << fProbEmplaceDet << endl;
   *file << "detonating on emplace delay " << fTimeEmplaceDet << endl;
   
   fpOptSelector->WriteFuzeOptSel(file);
}

EBoolean   TFuzeParam::IsDirectional(TFlatEntCoord*)
{
   return false;
}

EBoolean   TFuzeParam::IsWideArea(float*)
{
   return false;
}

EBoolean   TFuzeParam::IsCmdDet()
{
   switch (fTypeNameIndex) {
      case efuzetypeCommandFire:
      case efuzetypeCommandLoc:
      case efuzetypeCommandFireDir:
      case efuzetypeTripCommandFire:
      case esenstypeCommandCamDir:
      case esenstypeAutoCmdCamDir:
            return true;
//         break;
   
   }
   
   return false;
}

EBoolean   TFuzeParam::IsClickFire()
{
   switch (fTypeNameIndex) {
      case efuzetypeCommandFire:
      case efuzetypeCommandFireDir:
      case esenstypeCommandCamDir:
      case esenstypeAutoCmdCamDir:
            return true;
//         break;
   
   }
   
   return false;
}

EBoolean   TFuzeParam::IsController()
{
   return false;
}

void    TFuzeOptionSelector::WriteFuzeOptSel(ofstream* file)
{
   if ( fNumDestTimeOpt > 0 ) {
      *file << "self destruct time options";
      for ( int i = 0; i < fNumDestTimeOpt; i++ )
         *file << " " << fpDestTimeOpt[i];
      *file << endl;
   }
   else
      *file << "no self destruct option" << endl;
      
   if ( fbTgtCountOpt )
      *file << "target counting option" << endl;
   else
      *file << "no target counting option" << endl;
      
   switch (fAntihandlingOpt) {
      case eNo:
            *file << "no antihandling" << endl;
         break;
   
      case eYes:
            *file << "antihandling" << endl;
         break;
   
      case eBoth:
            *file << "antihandling choice" << endl;
         break;
   }
      
   switch (fBlasthardOpt) {
      case eNo:
            *file << "no blast hardening" << endl;
         break;
   
      case eYes:
            *file << "blast hardening" << endl;
         break;
   
      case eBoth:
            *file << "blast hardening choice" << endl;
         break;
   }

}

EBoolean     TProjParam::ParseDbStructItem(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses the one statement.
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TItemList *    pList;

   returnCode = infile.ParseItem(fKeyProjParamTable, eSizeProjParamTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case eprojkwEntStateTypeList:
                  pList = (TItemList *)returnVal;
                  fEntityStateType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				case eprojkwDetonTypeList:
                  pList = (TItemList *)returnVal;
                  fDetonationType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				case eprojkwImpactTypeList:
                  pList = (TItemList *)returnVal;
                  fGroundImpactType.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
 					break;

				case eprojkwWarhead:
                  fWarhead = (EWarhead) (*(int*)returnVal);
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case eprojkwLaunchVel:
                  fLaunchVel.x   = ((T2Dsingle*)returnVal)->x;
                  fLaunchVel.y = ((T2Dsingle*)returnVal)->y;
                  delete returnVal;
                  returnVal = NULL;
 					break;

/*
				case eprojkwLookAng:
                  fLookAngle = *(float*)returnVal;
                  delete returnVal;
 					break;
*/

				case eprojkwImpactVel:
                  fImpactVelocity = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

/*
				case eprojkwScanDetProb:
                  fScanDetectProb = *(float*)returnVal;
                  delete returnVal;
 					break;

				case eprojkwHitProb:
                  fHitProb = *(float*)returnVal;
                  delete returnVal;
 					break;
*/

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return false;

		default:
			Assert(false, kerrReadParamfile, "TProjParam::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

void    TProjParam::WriteProjOpt(ofstream* file)
{
   *file << "entity " ;
   fEntityStateType.WriteEntityType(file);
   *file << "detonation " ;
   fDetonationType.WriteEntityType(file);
   *file << "ground impact " ;
   fGroundImpactType.WriteEntityType(file);
   *file << "warhead " << (int)fWarhead << endl;
   *file << "launch vel, x y " << fLaunchVel.x << " "
                                   << fLaunchVel.y << endl;
//   *file << "look angle " << fLookAngle << endl;
   *file << "impact speed " << fImpactVelocity << endl;
//   *file << "scan detect prob " << fScanDetectProb << endl;
//   *file << "hit prob " << fHitProb << endl;
}

TProjParam::TProjParam(TEntityType& type) :
   fLaunchVel(28.84, 22.84)
{
   fEntityStateType = type;
   fDetonationType = type;
   fGroundImpactType  = type;
   fWarhead = ewhHEAntitank;
//   fLookAngle = 55.0 * kDegToRad;
   fImpactVelocity = 675.0;
//   fScanDetectProb = 0.9;
//   fHitProb = 0.7;
}

TProjParam::TProjParam() :
   fLaunchVel(28.84, 22.84)
{
      // entity types default to all zeros
   fWarhead = ewhHEAntitank;
//   fLookAngle = 55.0 * kDegToRad;
   fImpactVelocity = 675.0;
//   fScanDetectProb = 0.9;
//   fHitProb = 0.7;
}

TProjParam::~TProjParam()
{
}

EBoolean     TVehDimensions::ParseDbStructItem(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses the one statement.
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;

   returnCode = infile.ParseItem(fKeyVehSizeParamTable, eSizeVehSizeParamTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {

		case eFileNoErr:
			switch( returnIndex )
			{
				case evehkwLength:
                  length = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case evehkwWidth:
                  width = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case evehkwHeight:
                  height = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case evehkwTrackMin:
                  trackYmin = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				case evehkwTrackMax:
                  trackYmax = *(float*)returnVal;
                  delete returnVal;
                  returnVal = NULL;
 					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:
         return false;

		default:
			Assert(false, kerrReadParamfile, "TMineParam::ParseDbStructItem");
			return false;
			// error, error...
   }
   return true;
}

void    TVehDimensions::WriteDimensions(ofstream* file)
{
   *file << "len, wid, ht " << length << " "
                            << width << " "
                            << height << endl;
   *file << "track offset from centerline, min and max " 
                            << trackYmin << " "
                            << trackYmax << endl;
                            
}

cDbEntry*       cDbEntry::GetParentEntry() const
{
   if (fpPrevEntry) {
      if (fpPrevEntry->fpChildEntry == this)
         return fpPrevEntry;
      else
         return fpPrevEntry->GetParentEntry();
   }
   return NULL;
}

EBoolean     cDbEntry::ReadParamSet(cInFileObj& infile)
   // reads each parameter into buffer, passes it to ParseDbEntry to 
   // interpret it.
{
	EBoolean done = false;
	EBoolean success = true;
	do {
            // NextItem() stores an item for us to look at
		EFileReturn returnCode = infile.NextItem();
		switch( returnCode )
		{
			case eLevelDone:      // done with  first level, entire param set
//				cout << "level done!" << endl;
				done = true;
				break;

			case eFileNoErr:
				if( !ParseDbEntry(infile) )
				{
//					cout << "error parsing parameter" << endl;
					success = false;
				}
				break;

			default: // error getting next token
				Assert(false, kerrReadParamfile, "cDbEntry::ReadParamSet");
				return false;
		}
	} while( !done );

	return success;	
}

EBoolean     cDbEntry::ParseDbEntry(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses the one statement.
   //     This level gets tokens not recognized by descendent classes.
{
   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TEntityType    type;
   TItemList *    pList;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fKeyEntryTable, eentrySizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
		case eNextLevel:
			switch( returnIndex )
			{
				case eentrykwEntityType:
               type = fEntityType;
               type.extra = 0;
					if( !type.ReadEntityStruct(infile) )
						return false;
               if ( ValidType(type) )
                  fEntityType = type;
               else
                  success = false;
               break;

				default: // ??
               success =  false;
					break;
			}
			break;


		case eFileNoErr:
			switch( returnIndex )
			{
				case eentrykwName:
                  if (fpName) delete[] fpName;
                  fpName = cppstrdup((char*)returnVal);
   					delete[] returnVal;
                  returnVal = NULL;
					break;

				case eentrykwEntityTypeList:
                  pList = (TItemList *)returnVal;
                  type.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  returnVal = NULL;
                  pList = NULL;
                  
                  if ( ValidType(type) )
                     fEntityType = type;
                  else
                     success = false;
					break;

				default: // what?
                  success =  false;
					break;
			}
			break;

		case eBadKeyword:
            success =  false;
         break;

		default:
			   Assert(false, kerrReadParamfile, "cDbEntry::ParseDbEntry");
			success = false;
			// error, error...
   }
   //   cout << fEntityType.kind << " " << fEntityType.domain << " " << fEntityType.country << " " << fEntityType.category << " " << fEntityType.subcategory << " " << fEntityType.specific << " " << fEntityType.extra << endl;

   return success;
}

EBoolean    cDbEntry::ValidType(TEntityType& type)
      // the default probably never gets called
{
   if ( (type.kind           != kMatchAllChar) &&
        (type.domain         != kMatchAllChar) && 
        (type.country        != kMatchAllShort) && 
        (type.category       != kMatchAllChar) && 
	/*    (type.category204    != kMatchAllChar) && */
        (type.subcategory    != kMatchAllChar) && 
	/*        (type.subcategory204 != kMatchAllChar) &&  */
        (type.specific       != kMatchAllChar) && 
	/*        (type.specific204    != kMatchAllChar) &&  */
        (type.extra          != kMatchAllChar)  ) {
	/*        (type.extra204       != kMatchAllChar)  ) { */
      return true;
   }
   else
      return false;
}

int  cDbEntry::GetDbLevel()
{
   return 0;
}

void    cDbEntry::WriteDbEntry(ofstream* file)
{
   *file << "Database entry for "<< fpName << endl;
   fEntityType.WriteEntityType(file);
}

void  cDbEntry::InitFuzeParamTables(TFuzeParam* pFuzeParam)
{
   int fuzeType = pFuzeParam->fTypeNameIndex;
   
   switch (fuzeType) {
      case TFuzeParam::efuzetypeAcousticTrigger:
      case TFuzeParam::efuzetypeAcousticTracker:
      case TFuzeParam::efuzetypeTopSublet:
	 break;
      default:
	    pFuzeParam->InitVulnTables(fuzeType);
	 break;
   }
       
}

TFuzeParam*  cDbEntry::CreateNewFuzeParam(int fuzeType)
{
   TFuzeParam* pNewFuzeParam;
   switch (fuzeType) {
      case TFuzeParam::efuzetypeAcousticTrigger:
            pNewFuzeParam = new TFuzeParamOffRoute;
         break;
   
      case TFuzeParam::efuzetypeAcousticTracker:
            pNewFuzeParam = new TFuzeParamWideArea;
         break;
   
      case TFuzeParam::efuzetypeTopSublet:
            pNewFuzeParam = new TFuzeParamWamSublet;
         break;

      case TFuzeParam::esenstypeCommandCamDir:
      case TFuzeParam::esenstypeAutoCmdCamDir:
            pNewFuzeParam = new TSensParamCam(-1);
         break;

      case TFuzeParam::efuzetypeCommandFire:
      case TFuzeParam::efuzetypeCommandLoc:
      case TFuzeParam::efuzetypeCommandFireDir:
            pNewFuzeParam = new TFuzeParamApCmdDet(-1);
         break;

      case TFuzeParam::efuzetypeTripCommandFire:
            pNewFuzeParam = new TFuzeParamApTripCmdDet(-1);
         break;

/*   
      case TFuzeParam::efuzetypeCommandLoc:
            pNewFuzeParam = new TFuzeParamApWideArea;
         break;
*/   
      case TFuzeParam::esenstypePerDet:
      case TFuzeParam::esenstypeVehDet:
      case TFuzeParam::esenstypePerVehDet:
      case TFuzeParam::esenstypeTripWire:
            pNewFuzeParam = new TSensParamDumb(-1);
         break;
   
      case TFuzeParam::esenstypePerTrack:
      case TFuzeParam::esenstypeVehTrack:
      case TFuzeParam::esenstypePerVehTrack:
            pNewFuzeParam = new TSensParamTrack(-1);
         break;
   
      default:
            pNewFuzeParam = new TFuzeParam(-1);
         break;
   }
   return pNewFuzeParam;
}

TFuzeParam*  cDbEntry::CreateFuzeParam(int fuzeType)
{
   TFuzeParam* pNewFuzeParam;
   switch (fuzeType) {
      case TFuzeParam::efuzetypeAcousticTrigger:
            pNewFuzeParam = new TFuzeParamOffRoute;
         break;
   
      case TFuzeParam::efuzetypeAcousticTracker:
            pNewFuzeParam = new TFuzeParamWideArea;
         break;
   
      case TFuzeParam::efuzetypeTopSublet:
            pNewFuzeParam = new TFuzeParamWamSublet;
         break;
      case TFuzeParam::esenstypeCommandCamDir:
      case TFuzeParam::esenstypeAutoCmdCamDir:
            pNewFuzeParam = new TSensParamCam(fuzeType);
         break;
      case TFuzeParam::efuzetypeCommandFire:
      case TFuzeParam::efuzetypeCommandLoc:
      case TFuzeParam::efuzetypeCommandFireDir:
            pNewFuzeParam = new TFuzeParamApCmdDet;
         break;

      case TFuzeParam::efuzetypeTripCommandFire:
            pNewFuzeParam = new TFuzeParamApTripCmdDet;
         break;

/*   
      case TFuzeParam::efuzetypeCommandLoc:
            pNewFuzeParam = new TFuzeParamApWideArea;
         break;
*/   
      case TFuzeParam::esenstypePerDet:
      case TFuzeParam::esenstypeVehDet:
      case TFuzeParam::esenstypePerVehDet:
      case TFuzeParam::esenstypeTripWire:
            pNewFuzeParam = new TSensParamDumb(fuzeType);
         break;
   
      case TFuzeParam::esenstypePerTrack:
      case TFuzeParam::esenstypeVehTrack:
      case TFuzeParam::esenstypePerVehTrack:
            pNewFuzeParam = new TSensParamTrack(fuzeType);
         break;
   
      default:
            pNewFuzeParam = new TFuzeParam(fuzeType);
         break;
   }
   return pNewFuzeParam;
}

TMineParam*  cDbEntry::CreateMineParam(int projType)
{
   TMineParam* pNewMineParam;
   switch (projType) {
      case TMineParam::eprojtypeNone:
            pNewMineParam = new TMineParam;
         break;
   
      case TMineParam::eprojtypeWamSublet:
            pNewMineParam = new TMineParamWam(projType);
         break;
   
      default:
            pNewMineParam = new TMineParam;
         break;
   }
   return pNewMineParam;
}

EBoolean     cDbEntry::IsAbstract()
{
   if ( (fEntityType.extra == kMatchAllChar) ||
        (fpChildEntry) )
      return true;
   else
      return false;
}

cDbEntry::cDbEntry()
{
   fpName = fpAltName = NULL;
   fpPrevEntry = fpNextEntry = fpChildEntry = NULL;
}

cDbEntry::cDbEntry(cEntityDBMgr* pDb, TEntityType& type, int level, 
                  char* name, char* altname)
{
   fpPrevEntry = fpNextEntry = fpChildEntry = NULL;
   if (name)
      fpName = cppstrdup(name);
   else
      fpName = NULL;
   if (altname)
      fpAltName = cppstrdup(altname);
   else
      fpAltName = fpName;
   fEntityType = type;
   
   pDb->Register(this, level);
}

cDbEntry::~cDbEntry()
{
   if (fpChildEntry) {
      delete fpChildEntry;
      fpChildEntry = NULL;
   }
/*
   if (fpName)
      delete[] fpName;
   if (fpAltName)
      delete[] fpAltName;
*/
   if (fpPrevEntry) {
      if ( fpPrevEntry->fpNextEntry == this)
         fpPrevEntry->fpNextEntry = fpNextEntry;
      if ( fpPrevEntry->fpChildEntry == this)
         fpPrevEntry->fpChildEntry = fpNextEntry;
   }
   if (fpNextEntry)
      fpNextEntry->fpPrevEntry = fpPrevEntry;
}

EBoolean     cDbMine::ParseDbEntry(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses it by passing to relevant struct.
   //  Unrecognized token is passed to parent.
{

   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TEntityType    type;
   TItemList *    pList;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fKeyMineTable, emineSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
		case eNextLevel:
			switch( returnIndex )
			{
				case eminekwEntityType:
               type = fEntityType;
               type.extra = 0;
					if( !type.ReadEntityStruct(infile) )
                  success = false;
               else if ( ValidType(type) ) {
                  fEntityType = type;
                  ((TMineParam*)(fMineOpt.GetParam(0)))->fDetonationType = fEntityType;
                  ((TMineParam*)(fMineOpt.GetParam(0)))->fDetectionType = fEntityType;
               }
               else {
                  success = false;
               }
               break;

	 
				case eminekwMine:
               if ( ! (((TMineParam*)(fMineOpt.GetParam(0)))->ReadDbStruct(infile)) )
                  success = false;
               break;

				case eminekwFuze:
				case eminekwSensor: // we treat sensors as fuzes internally
	      {
                  TFuzeParam* pParam = CreateNewFuzeParam(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);   // create correct type
                  TFuzeParam* pOldParam = NULL;
                  if  ( pParam->ReadDbStruct(infile) ) {
		    InitFuzeParamTables(pParam);
                    for ( int i = 0; i < fFuzeOpt.GetNumParam(); i++ ) {
                        if ( pParam->fTypeNameIndex == ((TFuzeParam*)(fFuzeOpt.GetParam(i)))->fTypeNameIndex ) {
                           pOldParam = (TFuzeParam*)(fFuzeOpt.GetParam(i));
                           break;
                        }
                     }
		     if ( pOldParam ) {
                       if ( pParam->fFuze != (EFuze(-1)) ) {
                           cout<<"replacing param set"<<endl;
                           fFuzeOpt.ReplaceParam(pOldParam, pParam);
                        }
                        else {
                           cout<<"removing param set"<<endl;
                           fFuzeOpt.RemoveParam(i);
                        }
		     }
                     else {
                        cout<<"adding param set"<<endl;
                        fFuzeOpt.AddParam(pParam);
                     }
                  }
                  else
                     success = false;
	       }
               break;

				default: // ??
               success = false;
					break;
			}
			break;


		case eFileNoErr:
			switch( returnIndex )
			{
				case eminekwEntityTypeList:   // intercept, to set mine params
                  pList = (TItemList *)returnVal;
                  type.ParseList( pList );
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
                  
                  if ( ValidType(type) ) {
                     fEntityType = type;
                     ((TMineParam*)(fMineOpt.GetParam(0)))->fDetonationType = fEntityType;
                     ((TMineParam*)(fMineOpt.GetParam(0)))->fDetectionType = fEntityType;
                  }
					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:   // parent might recognize it
            success =  cDbEntry::ParseDbEntry(infile);
         break;

		default:
			   Assert(false, kerrReadParamfile, "cDbMine::ParseDbEntry");
			success = false;
			// error, error...
   }
   return success;
}

EBoolean    cDbMine::ValidType(TEntityType& type)
      // For new (input) mine type, child of this cDbMine
      // country can change, specific can change, extra not wild
{
   if ( (type.kind        == fEntityType.kind) &&
        (type.domain      == fEntityType.domain) && 
        (type.country     != kMatchAllShort) && 
        (type.category    == fEntityType.category) && 
        (type.specific    != kMatchAllChar) && 
        (type.extra       != kMatchAllChar)  ) {

      if ( fEntityType.subcategory == kMatchAllChar ) {   // false alarms
         if ( type.subcategory != kMatchAllChar )
            return true;
      }
      else {   // mines
         if (type.subcategory == fEntityType.subcategory)
            return true;
      }

      return false;
   }
   else
      return false;
}

int  cDbMine::GetDbLevel()
{
   if ( (fEntityType.kind != ekindMunition) ||
        (fEntityType.category != emuncatFixed) )
      return 0;
      
   if ( (fEntityType.domain == kMatchAllChar) || 
        (fEntityType.domain == 0) )
      return cMineDBMgr::emdblevMine;
      
   if ( (fEntityType.subcategory == kMatchAllChar) ||
        (fEntityType.subcategory == 0) )
      return cMineDBMgr::emdblevDomain;
      
   if ( (fEntityType.specific == kMatchAllChar) ||
        (fEntityType.specific == 0) )
      return cMineDBMgr::emdblevSubcat;
      
   if ( (fEntityType.extra == kMatchAllChar) ||
        (fEntityType.extra == 0) )
      return cMineDBMgr::emdblevSpecific;
      
   return cMineDBMgr::emdblevExtra;
}

void  cDbMine::WriteMineParam(ofstream* file)
{
   *file << endl;
   WriteDbEntry(file);
   
   ((TMineParam*)(fMineOpt.GetParam(0)))->WriteMineOpt(file);
   
   for ( int i = 0; i < fFuzeOpt.GetNumParam(); i++ )
      ((TFuzeParam*)(fFuzeOpt.GetParam(i)))->WriteFuzeOpt(file);
      
//   fVulnDetOpt.WriteVulnDetOpt(file);
//   fVulnMechOpt.WriteVulnMechOpt(file);
}

int    cDbMine::GetFuzeOpt(int fuzeTypeIndex)
{
   for ( int i = 0; i < fFuzeOpt.GetNumParam(); i++) {
      if ( ((TFuzeParam*)(fFuzeOpt.GetParam(i)))->fTypeNameIndex ==  fuzeTypeIndex)
         return i;
   }
   return 0;
}

EBoolean     cDbMine::GetDomainList( cDbMine* **retbuf, int *count, int* thisindex )
{
   if ( GetDbLevel() >= cMineDBMgr::emdblevDomain) {
      ((cDbMine*)GetParentEntry())->GetDomainList(retbuf, count, thisindex);
      if (thisindex) {
         for ( int i = 0; i < *count; i++) {
            if ( (*retbuf)[i]->GetEntryEntityType().domain == fEntityType.domain  ) {
               *thisindex = i;
               break;
            }
         }
      }
   }
   else {
      Assert(GetDbLevel() >= cMineDBMgr::emdblevMine, kerrValidityCheck, "cDbMine::GetDomainList");
      gMineDbMgr.GetNextLevel(this, fEntityType, (cDbEntry***)retbuf, count);
      if ( thisindex )
         *thisindex = -1;
   }
   return true;
}

EBoolean     cDbMine::GetSubcatList( cDbMine* **retbuf, int *count, int* thisindex )
{
   if ( GetDbLevel() >= cMineDBMgr::emdblevSubcat) {
      ((cDbMine*)GetParentEntry())->GetSubcatList(retbuf, count, thisindex);
      if (thisindex) {
         for ( int i = 0; i < *count; i++) {
            if ( (*retbuf)[i]->GetEntryEntityType().subcategory == fEntityType.subcategory  ) {
               *thisindex = i;
               break;
            }
         }
      }
   }
   else if (GetDbLevel() == cMineDBMgr::emdblevDomain) {
      if (fEntityType.domain == efunnydomainFalseAlarm) {      // for false alarms, subcat is entity level
         *retbuf = new cDbMine*[1];
         *count = 0;
         *thisindex = -1;
      }
      else {
      gMineDbMgr.GetNextLevel(this, fEntityType, (cDbEntry***)retbuf, count);
      if ( thisindex )
         *thisindex = -1;
      }
   }
   else {   //** SHOULD WE ALLOW THIS CALL?????
      gMineDbMgr.MatchAllLevel(cMineDBMgr::emdblevSubcat, fEntityType, 
               (cDbEntry***)retbuf, count);
      if ( thisindex )
         *thisindex = -1;
   }
   return true;
}

EBoolean     cDbMine::GetSpecificList( cDbMine* **retbuf, int *count,
             unsigned short country, int* thisindex )
{
   TEntityType filtertype;
   
   if ( IsAbstract() ) {
      filtertype = fEntityType;
      if ( country < enumCountry )
         filtertype.country = country;
      gMineDbMgr.GetDescendentLeaves( this, filtertype, (cDbEntry***)retbuf, count);
      if ( thisindex )
         *thisindex = -1;
	 }
	 else {
      filtertype = GetParentEntry()->GetEntryEntityType();
      if ( country < enumCountry )
         filtertype.country = country;
      gMineDbMgr.GetDescendentLeaves( GetParentEntry(), filtertype, (cDbEntry***)retbuf, count);
      if (thisindex) {
         for ( int i = 0; i < *count; i++) {
            if ( (*retbuf)[i] == this  ) {
               *thisindex = i;
               break;
            }
         }
      }
	 }
   return true;
}

EBoolean     cDbMine::GetCountryList( int **retbuf, int *count, int* thisindex )
{
   if ( GetDbLevel() >= cMineDBMgr::emdblevSpecific) {
      ((cDbMine*)GetParentEntry())->GetCountryList(retbuf, count, thisindex);
      if (thisindex) {
         for ( int i = 0; i < *count; i++) {
            if ( (*retbuf)[i] == fEntityType.country  ) {
               *thisindex = i;
               break;
            }
         }
      }
   }
   else {
      gMineDbMgr.GetCountryList(this, retbuf, count);
      if ( thisindex )
         *thisindex = -1;
   }
   return true;
}

   
char*    cDbMine::GetDomainName()
{
   if ( GetDbLevel() > cMineDBMgr::emdblevDomain ) 
      return ((cDbMine*)GetParentEntry())->GetDomainName();
   else if (GetDbLevel() == cMineDBMgr::emdblevDomain)
      return fpName;

   return NULL;
}

char*    cDbMine::GetDomainAltName()
{
   if ( GetDbLevel() > cMineDBMgr::emdblevDomain ) 
      return ((cDbMine*)GetParentEntry())->GetDomainAltName();
   else if (GetDbLevel() == cMineDBMgr::emdblevDomain)
      return fpAltName;

   return NULL;
}

char*    cDbMine::GetSubcatName()
{
   if ( GetDbLevel() > cMineDBMgr::emdblevSubcat ) 
      return ((cDbMine*)GetParentEntry())->GetSubcatName();
   else if (GetDbLevel() == cMineDBMgr::emdblevSubcat)
      return fpName;

   return NULL;
}

char*    cDbMine::GetSubcatAltName()
{
   if ( GetDbLevel() > cMineDBMgr::emdblevSubcat ) 
      return ((cDbMine*)GetParentEntry())->GetSubcatAltName();
   else if (GetDbLevel() == cMineDBMgr::emdblevSubcat)
      return fpAltName;

   return NULL;
}

const char*    cDbMine::GetCountryName()
{
   return fEntityType.GetCountryText();
}

EBoolean  cDbMine::IsDirectional(TDbOptSet& opt, TFlatEntCoord* pMinMax)
{
   return ((TFuzeParam*)(fFuzeOpt.GetParam(opt.fFuzeOpt)))->IsDirectional(pMinMax);
}

EBoolean  cDbMine::IsWideArea(TDbOptSet& opt, float* pRange)
{
   return ((TFuzeParam*)(fFuzeOpt.GetParam(opt.fFuzeOpt)))->IsWideArea(pRange);
}

EBoolean  cDbMine::IsCmdDet(TDbOptSet& opt)
{
   return ((TFuzeParam*)(fFuzeOpt.GetParam(opt.fFuzeOpt)))->IsCmdDet();
}

EBoolean cDbMine::IsController(TDbOptSet& opt)
{
   return ((TFuzeParam*)(fFuzeOpt.GetParam(opt.fFuzeOpt)))->IsController();
}

EBoolean  cDbMine::IsClickFire(TDbOptSet& opt)
{
   return ((TFuzeParam*)(fFuzeOpt.GetParam(opt.fFuzeOpt)))->IsClickFire();
}

cProjectile*    cDbMine::CreateProjectile(int mineOpt)
{
   return ( (TMineParamWam*) fMineOpt.GetParam(mineOpt) )->CreateProjectile();
}

cDbMine*  cDbMine::CloneDbMine()
{
   int i;
   TMineParam* pMineParam;
   TFuzeParam* pFuzeParam;
   
   cDbMine* pNewDbMine = CreateDbMine(false);   // returns one of correct type
   
      // copy the cDbEntry members - only the entity type gets copied
   pNewDbMine->fEntityType = fEntityType;
   
      // copy the cDbMine members
   for ( i = 0; i < fMineOpt.GetNumParam(); i++ ) {
      pMineParam = ((TMineParam*)(fMineOpt.GetParam(i)))->CloneMineParam();
      pNewDbMine->AddMineParam(pMineParam);
   }
   for ( i = 0; i < fFuzeOpt.GetNumParam(); i++ ) {
      pFuzeParam = ((TFuzeParam*)(fFuzeOpt.GetParam(i)))->CloneFuzeParam();
      pNewDbMine->AddFuzeParam(pFuzeParam);
   }

   return pNewDbMine;
}

cDbMine::cDbMine()
{ 
}

cDbMine::cDbMine(int fuzeType, int projType)
{ 
   TMineParam* pMineOpt     = CreateMineParam(projType);
   TFuzeParam* pFuzeOpt     = CreateFuzeParam(fuzeType);
   
   fMineOpt.AddParam(pMineOpt);
   fFuzeOpt.AddParam(pFuzeOpt);
}

cDbMine::cDbMine(int fuzeType, int projType, TEntityType& type, int level, 
                 char* name, char* altname) :
   cDbEntry(&gMineDbMgr, type, level, name, altname)
{
   TMineParam* pMineOpt     = CreateMineParam(projType);
   TFuzeParam* pFuzeOpt     = CreateFuzeParam(fuzeType);
   
   fMineOpt.AddParam(pMineOpt);
   fFuzeOpt.AddParam(pFuzeOpt);
}

cDbMine::~cDbMine()
{
}

cSimMine*     cDbMine::CreateSimMine(cPlanMine* pPlan, cMineComp* pComp)
   // does not link mine, just creates the new one
{
   const TDbOptSet&   partOpt = pComp->GetDbOpt();
   const TFuzeOpt&    fuzeOpt = pComp->GetFuzeOpt();
   
   cSimMine*   pMine   = CreateMine(pPlan);
   //pMine->CopyEntityParams(pPlan);
   //pMine->CopyMfObjParams(pPlan);
   //pMine->CopyMineParams(pPlan);
   
   TFuzeParam* pFuzeParam = GetFuzeParam(partOpt.fFuzeOpt);
   pMine->SetTracker( pFuzeParam->CreateTracker(pMine, fuzeOpt) );
   pMine->SetSensor( pFuzeParam->CreateSensor(pMine,fuzeOpt) );
   pMine->SetControl( (cMineControl*)(pFuzeParam->CreateControl(pMine, fuzeOpt)) );
   pMine->SetDetVuln( pFuzeParam->GetVulnDetList(fuzeOpt).CreateVulnDet( pMine ) );
   pMine->SetMechVuln( pFuzeParam->GetVulnMechParam().CreateVulnMech( pMine ) );
//   pMine->SetEmissVuln( pFuzeParam->CreateEmissVuln(pMine)); 

   EFuze fuze = pMine->GetFuze();
   EFuze planfuze = pPlan->GetFuze();
   return pMine;
}

cSimMine*     cDbMine::CreateVirtualMine(cMineComp* pComp, float radius)
{
   const TDbOptSet&   partOpt = pComp->GetDbOpt();
   const TFuzeOpt&    fuzeOpt = pComp->GetFuzeOpt();
   cSimMine*   pMine      = CreateMine(NULL); 

   TFuzeParam* pFuzeParam = GetFuzeParam(partOpt.fFuzeOpt);

   pMine->SetTracker( pFuzeParam->CreateTracker(pMine, fuzeOpt, radius) );
   pMine->SetSensor ( pFuzeParam->CreateSensor(pMine, fuzeOpt, radius) );
   pMine->SetDetVuln( pFuzeParam->GetVulnDetList(fuzeOpt).CreateVulnDet( pMine, radius ) );
   
   return pMine;
}

cSimMine*     cDbMine_cMine::CreateMine(cPlanMine* pPlan)
{
   Assert(false, kerrValidityCheck, "cDbMine_cMine::CreateMine");
   return (cSimMine*) new cMine(pPlan);
}

cDbMine*  cDbMine_cMine::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cMine* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cMine(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cMine();
   return pNewDbMine;
}

cDbMine_cMine::cDbMine_cMine() :
   cDbMine()
{
}

cDbMine_cMine::cDbMine_cMine(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cMine::cDbMine_cMine(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cMine::~cDbMine_cMine()
{
}

cSimMine*     cDbMine_cApMine::CreateMine(cPlanMine* pPlan)
{
   Assert(false, kerrValidityCheck, "cDbMine_cApMine::CreateMine");
   return (cSimMine*) new cApMine(pPlan);
}

cDbMine*  cDbMine_cApMine::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApMine* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApMine(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cApMine();
   return pNewDbMine;
}

cDbMine_cApMine::cDbMine_cApMine() :
   cDbMine()
{
}

cDbMine_cApMine::cDbMine_cApMine(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cApMine::cDbMine_cApMine(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApMine::~cDbMine_cApMine()
{
}

cSimMine*     cDbMine_cAtMine::CreateMine(cPlanMine* pPlan)
{
   Assert(false, kerrValidityCheck, "cDbMine_cAtMine::CreateMine");
   return (cSimMine*) new cAtMine(pPlan);
}

cDbMine*  cDbMine_cAtMine::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cAtMine* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cAtMine(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cAtMine();
   return pNewDbMine;
}

cDbMine_cAtMine::cDbMine_cAtMine() :
   cDbMine()
{
}

cDbMine_cAtMine::cDbMine_cAtMine(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cAtMine::cDbMine_cAtMine(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cAtMine::~cDbMine_cAtMine()
{
}

cSimMine*     cDbMine_cAtConv::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cAtConv(pPlan);
}

cDbMine*  cDbMine_cAtConv::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cAtConv* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cAtConv(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cAtConv();
   return pNewDbMine;
}

cDbMine_cAtConv::cDbMine_cAtConv() :
   cDbMine()
{
}

cDbMine_cAtConv::cDbMine_cAtConv(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cAtConv::cDbMine_cAtConv(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cAtConv::~cDbMine_cAtConv()
{
}

cSimMine*     cDbMine_cAtScatterable::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cAtScatterable(pPlan);
}

cDbMine*  cDbMine_cAtScatterable::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cAtScatterable* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cAtScatterable(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cAtScatterable();
   return pNewDbMine;
}

cDbMine_cAtScatterable::cDbMine_cAtScatterable() :
   cDbMine()
{
}

cDbMine_cAtScatterable::cDbMine_cAtScatterable(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cAtScatterable::cDbMine_cAtScatterable(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cAtScatterable::~cDbMine_cAtScatterable()
{
}

cSimMine*     cDbMine_cAtOffRoute::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cAtOffRoute(pPlan);
}

cDbMine*  cDbMine_cAtOffRoute::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cAtOffRoute* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cAtOffRoute(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cAtOffRoute();
   return pNewDbMine;
}

cDbMine_cAtOffRoute::cDbMine_cAtOffRoute() :
   cDbMine()
{
}

cDbMine_cAtOffRoute::cDbMine_cAtOffRoute(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
   GetMineParam(0)->fTccOrient.SetOrient( 0.0, 0.0, 0.0 );
}

cDbMine_cAtOffRoute::cDbMine_cAtOffRoute(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
   GetMineParam(0)->fTccOrient.SetOrient( 0.0, 0.0, 0.0 );
}

cDbMine_cAtOffRoute::~cDbMine_cAtOffRoute()
{
}

cSimMine*     cDbMine_cAtWideArea::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cAtWideArea(pPlan);
}

cDbMine*  cDbMine_cAtWideArea::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cAtWideArea* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cAtWideArea(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cAtWideArea();
   return pNewDbMine;
}

cDbMine_cAtWideArea::cDbMine_cAtWideArea() :
   cDbMine()
{
}

cDbMine_cAtWideArea::cDbMine_cAtWideArea(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeWamSublet  )
{
}

cDbMine_cAtWideArea::cDbMine_cAtWideArea(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeWamSublet, type, level, pName, altname )
{
}

cDbMine_cAtWideArea::~cDbMine_cAtWideArea()
{
}

cSimMine*     cDbMine_cApConv::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cApConv(pPlan);
}

cDbMine*  cDbMine_cApConv::CreateDbMine(EBoolean bWithParams)
{
   cDbMine_cApConv* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApConv(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cApConv();
   return pNewDbMine;
}

cDbMine_cApConv::cDbMine_cApConv() :
   cDbMine()
{
}

cDbMine_cApConv::cDbMine_cApConv(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone )
{
}

cDbMine_cApConv::cDbMine_cApConv(int fuzeType, TEntityType& type, int level, 
            char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApConv::~cDbMine_cApConv()
{
}

cSimMine*     cDbMine_cApScatterable::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cApScatterable(pPlan);
}

cDbMine*  cDbMine_cApScatterable::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApScatterable* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApScatterable(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cApScatterable();
   return pNewDbMine;
}

cDbMine_cApScatterable::cDbMine_cApScatterable() :
   cDbMine()
{
}

cDbMine_cApScatterable::cDbMine_cApScatterable(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cApScatterable::cDbMine_cApScatterable(int fuzeType, TEntityType& type, 
            int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApScatterable::~cDbMine_cApScatterable()
{
}

cSimMine*     cDbMine_cApConvBding::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cApConvBding(pPlan);
}

cDbMine*  cDbMine_cApConvBding::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApConvBding* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApConvBding(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cApConvBding();
   return pNewDbMine;
}

cDbMine_cApConvBding::cDbMine_cApConvBding() :
   cDbMine()
{
}

cDbMine_cApConvBding::cDbMine_cApConvBding(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cApConvBding::cDbMine_cApConvBding(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApConvBding::~cDbMine_cApConvBding()
{
}

cSimMine*     cDbMine_cApScatBding::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cApScatBding(pPlan);
}

cDbMine*  cDbMine_cApScatBding::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApScatBding* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApScatBding(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cApScatBding();
   return pNewDbMine;
}

cDbMine_cApScatBding::cDbMine_cApScatBding() :
   cDbMine()
{
}

cDbMine_cApScatBding::cDbMine_cApScatBding(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cApScatBding::cDbMine_cApScatBding(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApScatBding::~cDbMine_cApScatBding()
{
}

/*
cDbMine*  cDbMine_cApWideArea::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApWideArea* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApWideArea( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cApWideArea();
   return pNewDbMine;
}

cSimMine*     cDbMine_cApWideArea::CreateMine()
{
   return (cSimMine*) new cApWideArea;
}

cDbMine_cApWideArea::cDbMine_cApWideArea() :
   cDbMine()
{
}

cDbMine_cApWideArea::cDbMine_cApWideArea(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cApWideArea::cDbMine_cApWideArea(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApWideArea::~cDbMine_cApWideArea()
{
}
*/

cDbMine*  cDbMine_cApCommandDet::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApCommandDet* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApCommandDet( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cApCommandDet();
   return pNewDbMine;
}

cSimMine*     cDbMine_cApCommandDet::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cApCommandDet(pPlan);
}

cDbMine_cApCommandDet::cDbMine_cApCommandDet() :
   cDbMine()
{
}

cDbMine_cApCommandDet::cDbMine_cApCommandDet(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cApCommandDet::cDbMine_cApCommandDet(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cApCommandDet::~cDbMine_cApCommandDet()
{
}

cDbMine*  cDbMine_cApTripCommandDet::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cApTripCommandDet* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cApTripCommandDet( ((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex );
   else
      pNewDbMine = new cDbMine_cApTripCommandDet();
   return pNewDbMine;
}

cSimMine*     cDbMine_cApTripCommandDet::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cApTripCommandDet(pPlan);
}

cDbMine_cApTripCommandDet::cDbMine_cApTripCommandDet() :
   cDbMine()
{
}

cDbMine_cApTripCommandDet::cDbMine_cApTripCommandDet(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
   GetMineParam(0)->fTccOrient.SetOrient( 0.0, 0.0, 0.0 );
}

cDbMine_cApTripCommandDet::cDbMine_cApTripCommandDet(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
   GetMineParam(0)->fTccOrient.SetOrient( 0.0, 0.0, 0.0 );
}

cDbMine_cApTripCommandDet::~cDbMine_cApTripCommandDet()
{
}

cSimMine*     cDbMine_cFalseMine::CreateMine(cPlanMine* pPlan)
{
   return (cSimMine*) new cFalseMine(pPlan);
}

cDbMine*  cDbMine_cFalseMine::CreateDbMine(EBoolean bWithParams)
{
    cDbMine_cFalseMine* pNewDbMine;
   if ( bWithParams )
      pNewDbMine = new cDbMine_cFalseMine(((TFuzeParam*)(fFuzeOpt.GetParam(0)))->fTypeNameIndex);
   else
      pNewDbMine = new cDbMine_cFalseMine();
   return pNewDbMine;
}

cDbMine_cFalseMine::cDbMine_cFalseMine() :
   cDbMine()
{
}

cDbMine_cFalseMine::cDbMine_cFalseMine(int fuzeType) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone  )
{
}

cDbMine_cFalseMine::cDbMine_cFalseMine(int fuzeType, TEntityType& type, 
         int level, char* pName, char* altname) :
   cDbMine( fuzeType, TMineParam::eprojtypeNone, type, level, pName, altname )
{
}

cDbMine_cFalseMine::~cDbMine_cFalseMine()
{
}

EBoolean     cDbVeh::ParseDbEntry(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses it by passing to relevant struct.
   //  Unrecognized token is passed to parent.
{

   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fKeyVehTable, evehSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
		case eNextLevel:
			switch( returnIndex )
			{
				case evehkwSize:
                  success = fVehDim.ReadDbStruct(infile);
               break;

				default: // ??
                  success = false;
					break;
			}
			break;


		case eFileNoErr:
			switch( returnIndex )
			{

				default: // what?
                  success = false;
					break;
			}
			break;

		case eBadKeyword:   // parent might recognize it
            success = cDbEntry::ParseDbEntry(infile);
          break;

		default:
		   	Assert(false, kerrReadParamfile, "cDbVeh::ParseDbEntry");
			success = false;
			// error, error...
   }
   return success;
}

int  cDbVeh::GetDbLevel()
{
   if ( (fEntityType.kind != ekindPlatform) &&
        (fEntityType.kind != ekindLifeform)   )
      return 0;
      
   if ( fEntityType.domain == kMatchAllChar )
      return cVehDBMgr::evdblevPlatform;
      
   if ( fEntityType.category == kMatchAllChar )
      return cVehDBMgr::evdblevDomain;
      
   if ( fEntityType.subcategory == kMatchAllChar )
      return cVehDBMgr::evdblevCat;
      
   if ( (fEntityType.specific == kMatchAllChar) || 
        (fEntityType.specific == 0) )
      return cVehDBMgr::evdblevSubcat;
      
   return cVehDBMgr::evdblevSpecific;
}

void  cDbVeh::WriteVehicleParam(ofstream* file)
{
   *file << endl;
   WriteDbEntry(file);
      
   fVehDim.WriteDimensions(file);
}

EBoolean    cDbVeh::ValidType(TEntityType& type)
      // For new vehicles, kind, domain, and cat must
      // match this parent type
{
   if ( (type.kind        == fEntityType.kind) &&
        (type.domain      == fEntityType.domain) && 
        (type.country     != kMatchAllShort) && 
        (type.category    == fEntityType.category) && 
        (type.subcategory != kMatchAllChar)  &&
        (type.specific    != kMatchAllChar)  &&
        (type.extra       != kMatchAllChar) ) {
      return true;
   }
   else
      return false;
}

cDbVeh*  cDbVeh::CloneDbVeh()
{
   cDbVeh* pNewDbVeh = new cDbVeh;
   pNewDbVeh->fVehDim = fVehDim;
   
      // copy the cDbEntry members - only the entity type gets copied
   pNewDbVeh->fEntityType = fEntityType;

   return pNewDbVeh;
}

cDbVeh::cDbVeh() :
   cDbEntry()
{
   // veh dim default initialize to zeros
}

cDbVeh::cDbVeh(TEntityType& type, int level,
        char* name, TVehDimensions* pVehDim) :
   cDbEntry(&gVehDbMgr, type, level, name)
{
   if ( pVehDim )
      fVehDim = *pVehDim;
}

cDbVeh::~cDbVeh()
{
}

EBoolean     cDbProj::ParseDbEntry(cInFileObj& infile)
   // looks at a token waiting to be read in the parameter file.
   //  If it recognizes the token, parses it by passing to relevant struct.
   //  Unrecognized token is passed to parent.
{

   EFileReturn    returnCode;
   int            returnIndex;
   void*          returnVal = NULL;
   TEntityType    type;
   TItemList *    pList;
   EBoolean       success = true;

   returnCode = infile.ParseItem(fKeyProjTable, eprojSizeTable,
		&returnIndex, &returnVal);

   switch( returnCode ) {
		case eNextLevel:
			switch( returnIndex ) {
				case eprojkwEntityType:
                  type = fEntityType;
                  type.specific = 0;
                  type.extra = 0;
   					if ( (type.ReadEntityStruct(infile)) && (ValidType(type)) ) {
                     fEntityType = type;
                     fpProjParam->fEntityStateType  = fEntityType;
                     fpProjParam->fDetonationType   = fEntityType;
                     fpProjParam->fGroundImpactType = fEntityType;
                  }
   					else
                     success = false;
               break;

				case eprojkwProj:
                  if ( ! fpProjParam->ReadDbStruct(infile) )
                     success = false;
              break;

				case eprojkwFuze:
                  if ( ! fpFuzeParam->ReadDbStruct(infile) )
                     success = false;
               break;

				default: // ??
                  success = false;
					break;
            }
			break;


		case eFileNoErr:
			switch( returnIndex )
			{
				case eprojkwEntityTypeList:
                  pList = (TItemList *)returnVal;
                  type.ParseList( pList );
                  
                  if ( type.ParseList(pList) && ValidType(type) ) {
                     fEntityType = type;
                     fpProjParam->fEntityStateType  = fEntityType;
                     fpProjParam->fDetonationType   = fEntityType;
                     fpProjParam->fGroundImpactType = fEntityType;
                  }
                  else
                     success = false;
                 
                  delete[] (int *)(pList->ptrItems);
                  delete pList;
                  pList = NULL;
                  returnVal = NULL;
					break;

				default: // what?
					break;
			}
			break;

		case eBadKeyword:   // parent might recognize it
            success =  cDbEntry::ParseDbEntry(infile);
         break;

		default:
			   Assert(false, kerrReadParamfile, "cDbProj::ParseDbEntry");
			success = false;
			// error, error...
   }
   return success;
}

EBoolean    cDbProj::ValidType(TEntityType& type)
      // For new (input) proj type,
      // subcat and country must be specified
{
   if ( (type.kind        == fEntityType.kind) &&
        (type.domain      == fEntityType.domain) && 
        (type.country     != kMatchAllShort) && 
        (type.category    == fEntityType.category) && 
        (type.subcategory != kMatchAllChar) && 
        (type.specific    != kMatchAllChar) && 
        (type.extra       != kMatchAllChar)  ) {
      return true;
   }
   else
      return false;
}

int  cDbProj::GetDbLevel()
{
   if ( (fEntityType.kind != ekindMunition) ||
        (fEntityType.domain != emundomAntitank) )
      return 0;
      
   if ( fEntityType.category == kMatchAllChar )
      return cProjDBMgr::epjdblevKindDom;
      
   if ( fEntityType.subcategory == kMatchAllChar )
      return cProjDBMgr::epjdblevCat;
      
   if ( fEntityType.specific == kMatchAllChar )
      return cProjDBMgr::epjdblevSubcat;
      
   return cProjDBMgr::epjdblevEntity;
}

void  cDbProj::WriteProjParam(ofstream* file)
{
   *file << endl;
   WriteDbEntry(file);
   
   fpProjParam->WriteProjOpt(file);
   fpFuzeParam->WriteFuzeOpt(file);
}

cDbProj*  cDbProj::CloneDbProj()
{
   
   cDbProj* pNewDbProj = CreateDbProj(false);   // returns one of correct type
   
      // copy the cDbEntry members - only the entity type gets copied
   pNewDbProj->fEntityType = fEntityType;
   
      // copy the cDbProj members
   pNewDbProj->fpProjParam = fpProjParam->CloneProjParam();

   pNewDbProj->fpFuzeParam = fpFuzeParam->CloneFuzeParam();

   return pNewDbProj;
}

cDbProj::cDbProj()
{ 
}

cDbProj::cDbProj(int fuzeType)
{ 
   Assert ( fuzeType == TFuzeParam::efuzetypeTopSublet, kerrValidityCheck, "cDbProj::cDbProj");
   fpFuzeParam = new TFuzeParamWamSublet;
   fpProjParam = new TProjParam;
}

cDbProj::cDbProj(int fuzeType, TEntityType& type, int level, char* name) :
   cDbEntry(&gProjDbMgr, type, level, name)
{
   Assert ( fuzeType == TFuzeParam::efuzetypeTopSublet, kerrValidityCheck, "cDbProj::cDbProj");
   fpFuzeParam = new TFuzeParamWamSublet;
   fpProjParam = new TProjParam;
}

cDbProj::~cDbProj()
{
   delete fpFuzeParam;
   delete fpProjParam;
}

cDbProj*        cDbProj_cProj::CreateDbProj(EBoolean bWithParams)
{
   cDbProj_cProj* pNewDbProj;
   if ( bWithParams )
      pNewDbProj = new cDbProj_cProj(fpFuzeParam->fTypeNameIndex);
   else
      pNewDbProj = new cDbProj_cProj();
   
   return pNewDbProj;
}

cProjectile*    cDbProj_cProj::CreateProj()
{
   Assert(false, kerrValidityCheck, "cDbProj_cProj::CreateProj");
   return new cProjectile(this);
}

cDbProj_cProj::cDbProj_cProj()
{ 
}

cDbProj_cProj::cDbProj_cProj(int fuzeType) :
   cDbProj(fuzeType)
{ 
}

cDbProj_cProj::cDbProj_cProj(int fuzeType, TEntityType& type, int level, char* name) :
   cDbProj(fuzeType, type, level, name)
{
}

cDbProj_cProj::~cDbProj_cProj()
{
}

cDbProj*        cDbProj_Guided::CreateDbProj(EBoolean bWithParams)
{
   cDbProj_Guided* pNewDbProj;
   if ( bWithParams )
      pNewDbProj = new cDbProj_Guided(fpFuzeParam->fTypeNameIndex);
   else
      pNewDbProj = new cDbProj_Guided();
   
   return pNewDbProj;
}

cProjectile*    cDbProj_Guided::CreateProj()
{
   Assert(false, kerrValidityCheck, "cDbProj_Guided::CreateProj");
   return new cProjectile(this);
}

cDbProj_Guided::cDbProj_Guided()
{ 
}

cDbProj_Guided::cDbProj_Guided(int fuzeType) :
   cDbProj(fuzeType)
{ 
}

cDbProj_Guided::cDbProj_Guided(int fuzeType, TEntityType& type, int level, char* name) :
   cDbProj(fuzeType, type, level, name)
{
}

cDbProj_Guided::~cDbProj_Guided()
{
}

cDbProj*        cDbProj_Ballistic::CreateDbProj(EBoolean bWithParams)
{
   cDbProj_Ballistic* pNewDbProj;
   if ( bWithParams )
      pNewDbProj = new cDbProj_Ballistic(fpFuzeParam->fTypeNameIndex);
   else
      pNewDbProj = new cDbProj_Ballistic();
   
   return pNewDbProj;
}

cProjectile*    cDbProj_Ballistic::CreateProj()
{
   cWamSublet*  pProj = new cWamSublet(this);

      // TFuzeOpt not used for projectile fuzes
   TFuzeOpt fuzeOpt;
   pProj->SetTracker( fpFuzeParam->CreateTracker(pProj, fuzeOpt) );
   pProj->SetControl( ( (TFuzeParamWamSublet*) fpFuzeParam)->CreateControl(pProj, fuzeOpt) );
//   pProj->SetDetVuln( GetVulnDetList(partOpt.fVulnDetOpt)->CreateVulnDet( pMine, fuzeOpt ) );
//   pProj->SetMechVuln( GetVulnMechParam(partOpt.fVulnMechOpt)->CreateVulnMech( pMine, fuzeOpt ) );

   return pProj;
}

cDbProj_Ballistic::cDbProj_Ballistic()
{ 
}

cDbProj_Ballistic::cDbProj_Ballistic(int fuzeType) :
   cDbProj(fuzeType)
{ 
}

cDbProj_Ballistic::cDbProj_Ballistic(int fuzeType, TEntityType& type, int level, char* name) :
   cDbProj(fuzeType, type, level, name)
{
}

cDbProj_Ballistic::~cDbProj_Ballistic()
{
}

/* eof */
