/*
$RCSfile: Globals.h,v $
$Revision: 1.4 $
$Author: jchappel $
$Date: 1999/05/28 13:49:03 $
$Log: Globals.h,v $
Revision 1.4  1999/05/28 13:49:03  jchappel


Coms code that has been combined with cms code for one distribution.  Makefile
distinguishes which executable to create.

Revision 1.3  1999/05/18 19:33:18  ksnively
Ported CMS to compile under IRIX 6.5
Altered make files to allow cms to be compiled under various OS simulataneously
Removed the Startup GUI

Revision 1.2  1999/03/26 21:17:35  ksnively

Added radio communication capability for Command Detonated Mines.  Included sending
and receiving transmitter and signal PDUs.  All signals have the option of being
processed by the Communication Effects Federate.

Revision 1.1.1.1  1999/02/16 14:54:50  jchappel
Comprehensive Mine Simulator

 * Revision 2.2  1998/10/27  19:34:43  jchappel
 * modifications for versioning
 *
 * Revision 2.2  1998/10/27  19:34:43  jchappel
 * modifications for versioning
 *
 * Revision 2.1  1998/03/31  14:49:21  ksnively
 * *** empty log message ***
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.5  1998/01/09  18:29:24  ksnively
 * Integrated latest SAIC release cms3.0b2
 *
 * Revision 1.4  1998/01/08  13:05:54  ksnively
 * Altered time routine to read clock cycles, faster then using gettimeofday.
 *
 * Revision 1.3  1997/12/22  21:04:56  ksnively
 * Added interupt for updating Sim Time.
 *
 * Revision 1.2  1997/12/17  14:40:21  ksnively
 *  Debugged Minefield duplication routine.
 * Added input file "netparams" for reading PDU type enumerations.
 * Got rid of warning messages during compilation.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.5  1997/10/27  17:08:57  trang
 * BUILD1
 *
 * Revision 1.4  1997/10/27  17:05:38  trang
 * BUILD1
 *
 * Revision 1.4  1997/10/27  17:05:38  trang
 * BUILD1
 *
Revision 1.3  1997/03/20 19:16:16  ksnively
Added emission tracking, updated to CMS2.0B6.

 * Revision 1.2  1997/01/02  19:48:54  nsipes
 * *** empty log message ***
 *
 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.10  1996/08/21  21:58:20  rcarp
 * CP and QRP file output, other updates
 *
 * Revision 1.9  1996/07/03  22:55:15  rcarp
 * enhancements and bug fixes for version 1.5b9
 *
 * Revision 1.8  1996/06/19  19:05:45  rcarp
 * revisions for mine vulnerability, version 1.5b2
 *
 * Revision 1.7  1996/05/14  16:15:08  rcarp
 * refinements and bug fixes for 1.5 delivery
 *
 * Revision 1.4  1996/03/23  14:22:31  rcarp
 * improvements to test mode and selection
 *
 * Revision 1.3  1996/02/21  19:45:02  rcarp
 * revisions for map scroll, mine ranges
 *
 * Revision 1.2  1996/02/16  19:33:39  rcarp
 * revisions for 1.1b7 version
 *
$State: Exp $
*/
 
/*******************************************

   file: Globals.h
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   Globals.h
   
      Global variables and functions declared here, most are 
      defined in main.cc
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/


#ifndef _UGlobals_
#define _UGlobals_


#include <limits.h>
#include <float.h>

#include "EntityEnums.h"
#include "GlobalTypes.h"

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <libtactmap.h>

// global constants
extern char Datapath[256]; 
extern char DatapathConst[256];
extern char Const[256];
extern char Setupdir[256];

      //+++++++++error numbers
enum EerrorNumber   {
   kerrNone = 0,
   kerrDebug,
   kerrNoBuff,
   kerrNoPduSend,
   kerrRecvError,
   kerrReadTerrain,
   kerrReadMinefield,
	kerrReadParamfile,
   
         //errors after this lead to shutdown
   kerrLowestFatal = 500,
   kerrMemory = kerrLowestFatal,
   kerrValidityCheck,
   kerrNoRecvBcast,
   kerrBindSocket,
   kerrNoBlock,
   kerrCreateSendSocket,
   kerrNoSendBcast,
   kerrInvalidMcast,
   kerrNoSubscribeMcast,
   kerrNoTtlMcast,
   kerrNoDisableLoop,
   kerrCoordRead,
   kerrProtocolFamily,
   kerrList,
   kerrTypeMatch
   
};

//Valid Component Types
enum ECompType { eCompNull, eCompPerim, eCompIndiv, eCompGrid, eCompRow };


//   Global constants, declared extern here.  Most defined in main.cc

extern   const char*  kpCmsVersion;
extern   const double kSimTimeToSec;
extern   const double kSecToSimTime;
extern   const double kSimTimeToDis;
extern   const double kDisToSimTime;
extern   const double kSimTimeToUsec;
extern   const double kUsecToSimTime;
extern   int   kMaxMinesInField;

extern   const double kPi;
extern   const double kHalfPi;
extern   const double kDegToRad;
extern   const double kRadToDeg;
extern   const double kHalfGravAccel;
extern   const double kInchToMeter;
extern   const double kMeterToInch;

//   Global variables, declared extern here.  Most defined in main.cc

//   message lists
extern   cPrEventList     gPrEventList;
extern   cPrUpdtList      gPrUpdtList;
extern   cEventList       gEventList;
extern   cUpdtList        gUpdtList;
extern   cDispUpdtList    gDispUpdtList;
extern   cLowPrEventList  gLowPrEventList;
extern   cLowRateUpdtList gLowRateUpdtList;

   // file types, defined in FileObj
extern   char*   gftypeUndefined;
extern   char*   gftypeMinefields;
extern   char*   gftypeParam;
extern   char*   gftypeNetParam;
extern   char*   gfmechVulnParam;
extern   char*   gfdetVulnParam;
extern   char*   gftypeInteract;
extern   char*   gftypeMineState;
extern   char*   gftypeMineInputs;
extern   char*   gftypeVehicleInputs;
extern   char*   gftypeTestFile;
extern   char*   gftypePduData;

extern   const int kYesNoPairs;
extern   const TKeyValuePair gpYesNoText[];

//   miscellaneous
class cMinefield;
class cMfComponent;
class cMfieldObj;
class cPerimeter;
#ifdef COMS
class cCOMSComponent;
#endif

extern   EBoolean         gbGui;
extern   EBoolean         gbHost;
extern   TACTMAP_PTR      gpTactmap;
extern   Display*         gpDisplay;
extern   Window           gMapWindow;
extern   long             gSimStartTime;      // in seconds, not TTimeSim

#ifdef sgi
typedef  unsigned int     iotimer_t;
extern   unsigned int     gSimStartCycle;        // in cycles
extern   long             gStartTimeMilli;
extern volatile iotimer_t       *gpTimerAddr;
extern   double           gCycleVal;
extern   unsigned int     max_diff;
#define  MAXNUMCYCLES     4294967295.0           // 2^32 - 1
#endif

//   managers etc.
class   cEntityDBMgr;
class   cMineDBMgr;
class   cVehDBMgr;
class   cProjDBMgr;
class   cMechThreatList;
class   cDetThreatList;
class   cMsgMgr;
class   cNetMgr;
class   cCmsMgr;
class   cVehicleMgr;
class   cEmissionMgr;
class   cMfieldMgr;
class   cTerrain;
class   cGuiWindow;
class   cDisPduType;
class   cRadioNetworkMgr;
#ifdef COMS
class   cCOMSMgr;
extern   cCOMSMgr         gCOMSMgr;
#endif
extern   cMineDBMgr       gMineDbMgr;
extern   cVehDBMgr        gVehDbMgr;
extern   cProjDBMgr       gProjDbMgr;
extern   cMechThreatList  gMechThreatList;
extern   cDetThreatList   gDetThreatList;
extern   cMsgMgr          gMsgMgr;
extern   cNetMgr*         gpNetMgr;
extern   cCmsMgr          gCmsMgr;
extern   cEmissionMgr     gEmissionMgr;
extern   cVehicleMgr      gVehMgr;
extern   cMfieldMgr       gMfieldMgr;
extern   cTerrain*        gpTerrainMgr;
extern   cGuiWindow *     gpGuiWin;
extern   cDisPduType      gPduType;
extern   cRadioNetworkMgr* gpRadioNetMgr;


      //    Dis version and related pointers
class   cDisPdu;
extern   cDisPdu*         gpDisVersion;       // pointer to dis version

      // output control flags, file object pointers, and file names
extern int         gFileProt;      // protection mode for created output files
extern int         gfInteract;
extern ofstream*   gpInteract;
extern char*       gpInteractFileName;

extern int         gfMineStateDisp;
extern int         gfMineStateSumm;
extern ofstream*   gpMineStateSumm;
extern char*       gpMineStateFileName;

extern int         gfMineInputParams;
extern ofstream*   gpMineInputParams;
extern char*       gpMineInputFileName;

extern int         gfVehicleInputParams;
extern ofstream*   gpVehicleInputParams;
extern char*       gpVehicleInputFileName;

extern int         gfPduData;
extern ofstream*   gpPduData;
extern char*       gpPduDataFileName;

extern int      gfTestRun;
extern int      gfVehicle;
extern int      gfMinefield;
extern int      gfMine;
extern int      gfFixObj;
extern int      gfSensor;
extern int      gfDamage;
extern int      gfControl;
extern int      gfAttack;
extern int      gfNetif;
extern int      gfDetectProt;

extern char*    gpRunName;

//   Global functions.  Most defined in main.cc

extern   float            UniformRand();         // uniform on (0,1)
extern   void             SetRand(unsigned int seed);
extern   TTimeSim         GetCurSimTime();
extern   void             SetSimStartTime();
extern   char*            cppstrdup( const char* pStr);

extern   void   NewFailureHandler();         //called if new cannot get memory it needs
extern   void   Assert(int, EerrorNumber, char* fnName = 0, int sysErr = 0);    //condition to test, err num on failure
extern   void   Shutdown(EerrorNumber);       // global, can be called regardless of CMS state

extern "C" int yywrap();

extern   TLocation     EntToSimLoc( const TEntCoord& entLoc, const TLocation& entOrigin, const TRotation& entRot );
extern   TEntCoord     SimToEntLoc( const TLocation& simLoc, const TLocation& entOrigin, const TRotation& entRot );
extern   EBoolean      IsVerySmall( const double x );

   // geometry helpers
extern   int   CntrClk(double pt0x, double pt0y, double pt1x, double pt1y, double pt2x, double pt2y);
extern   int   CntrClk(const T3Dsingle& pt0, const T3Dsingle& pt1, const T3Dsingle& pt2);

extern   EBoolean   SegIntersect(const T3Dsingle& seg1pt1, const T3Dsingle& seg1pt2, const T3Dsingle& seg2pt1, const T3Dsingle& seg2pt2);
extern   EBoolean   SegIntersect(const T3Ddouble& seg1pt1, const T3Ddouble& seg1pt2, const T3Ddouble& seg2pt1, const T3Ddouble& seg2pt2);
extern   EBoolean   PtInPolygon(const T3Dsingle& pt, int numPolyPts, const T3Dsingle* pPolyPts);
extern   EBoolean   PtInPolygon(const T3Ddouble& pt, int numPolyPts, const T3Ddouble* pPolyPts);

#endif
