/*
$RCSfile: IndivEditor.cc,v $
$Revision: 1.3 $
$Author: ksnively $
$Date: 1999/05/10 20:18:24 $
$Log: IndivEditor.cc,v $
Revision 1.3  1999/05/10 20:18:24  ksnively
Added a command line option for enabling Communication Capability.

Revision 1.2  1999/03/26 21:17:36  ksnively

Added radio communication capability for Command Detonated Mines.  Included sending
and receiving transmitter and signal PDUs.  All signals have the option of being
processed by the Communication Effects Federate.

Revision 1.1.1.1  1999/02/16 14:54:35  jchappel
Comprehensive Mine Simulator

// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.8  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.7  1996/05/11  21:35:41  mkappel
// Add mine age.
//
// Revision 1.7  1996/05/11  21:35:41  mkappel
// Add mine age.
//
// Revision 1.6  1996/05/11  16:49:17  mkappel
// New mine type selector and fuze options.
//
// Revision 1.5  1996/05/08  00:48:49  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.4  1996/03/28  22:34:41  mkappel
// Add Mine Display toggle to component editors.
//
// Revision 1.3  1996/03/23  14:26:12  rcarp
// improvements to test mode and selection
//
// Revision 1.2  1996/03/13  00:12:33  mkappel
// Create scrolling set of toggles for mines.
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: IndivEditor.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   IndivEditor.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include "MineType.h"
#include "EntityIdEdit.h"

#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include <stdio.h>


cIndivEditorWindow::cIndivEditorWindow (Widget wParent, EBoolean bController)
   : cMineCompEditorWindow (wParent, bController)
{
   Widget            wLabel, wCoordForm = fpMineCoord->GetCoordForm();
   cHelpTextWidget*  pHelp;
   char             *help_text;
   
   wLabel = 
     XtVaCreateManagedWidget ("Individual Emplacement Editor", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 0,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 20,
			      XmNalignment, XmALIGNMENT_BEGINNING,
			      NULL);

   help_text = cppstrdup ("Emplace Mine: left click on map or select first blank entry at right.\nDelete Mine: right click on map or use Delete button in entry pop-up.\nMove Mine: left click-and-drag or use entry pop-up.\nRe-orient Mine: middle click-and-drag or use entry pop-up."); 

   pHelp = new cHelpTextWidget (fwMainWin, help_text);
   XtVaSetValues (pHelp->GetHelpForm(),
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_WIDGET,
		  XmNleftWidget, fpMineType->GetTypeForm(),
		  XmNleftOffset, 5,
		  XmNrightAttachment, XmATTACH_WIDGET,
		  XmNrightWidget, wCoordForm,
		  XmNrightOffset, 5,
		  NULL);
   delete[] help_text;

   if (fpControllerId) {
      XtVaSetValues (fpControllerId->GetIdForm(),
		     XmNtopAttachment, XmATTACH_WIDGET,
		     XmNtopWidget, pHelp->GetHelpForm(),
		     NULL);
   }

   XtVaSetValues (fwMineDispToggle,
		  XmNrightAttachment, XmATTACH_WIDGET,
		  XmNrightWidget, wCoordForm,
		  XmNrightOffset, 10,
		  NULL);
}

cIndivEditorWindow::~cIndivEditorWindow ()
{
}

void cIndivEditorWindow::ShowEdWindow ()
{
   cMineCompEditorWindow::ShowEdWindow();
   fpMineCoord->AppendCoordEntry ();
}

void  cIndivEditorWindow::MapClick (TLocation& loc)
{
  ((cIndivComp*)(gpGuiWin->GetCurComponent()))->AddTccMine (loc);
}

cHelpTextWidget::cHelpTextWidget (Widget wParent, char *help_text)
{
   fwHelpForm = 
     XtVaCreateWidget ("help",
		       xmTextWidgetClass, wParent, 
		       XmNeditMode      , XmMULTI_LINE_EDIT,
//		       XmNwordWrap      , TRUE,
		       XmNresizeHeight  , TRUE,
		       XmNresizeWidth   , TRUE,
		       XmNeditable      , FALSE,
		       XmNvalue         , help_text,
		       NULL);

   XtManageChild (fwHelpForm);
}

cHelpTextWidget::~cHelpTextWidget()
{
}
