/*
$RCSfile: MapControlsCB.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:36 $
$Log: MapControlsCB.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:36  jchappel
Comprehensive Mine Simulator

// Revision 2.2  1998/02/26  21:24:37  ksnively
// Created Mf Camera object to control MAK Stealth view.
// Created vulnerability tables specific to command detonated mines
// and the camera object.
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:51:30  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:48:55  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.3  1996/02/23  18:36:44  mkappel
// Add Site/Application to Start Up Screen.
//
// Revision 1.3  1996/02/23  18:36:44  mkappel
// Add Site/Application to Start Up Screen.
//
// Revision 1.2  1996/02/16  20:33:22  rcarp
// revisions associated with 1.1b7
//
$State: Exp $
*/
 
/*******************************************

   file: MapControlsCB.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   MapControlsCB.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "MapControls.h"
#include "Minefield.h"
#include "Vehicle.h"
#include <Xm/Text.h>
#include <Xm/ToggleB.h>

EBoolean   cMapControlsWindow::IsMineDisplayed( EForce force )
{
   if ( force == eOther )
      return fMToggleSet[eNeutral - 1];
   else
      return fMToggleSet[force - 1];
}

EBoolean   cMapControlsWindow::IsVehDisplayed( EForce force )
{
   if ( force == eOther )
      return fVToggleSet[eNeutral - 1];
   else
      return fVToggleSet[force - 1];
}

void cMapControlsWindow::InitControls ()
{
   int           i;
   char          str[22];
/*removed width and height AUSTIN 7 Aug 97*/
   float64       xcenter, ycenter;
   int32	ccenter;

   if (! XtIsRealized (fwMainWin)) return;

   XtVaSetValues (fwScaleMenu,
		  XmNmenuHistory, fScale[gpGuiWin->GetMapScaleIndex()].wOption,
		  NULL);
/*replaced tactmap_locale with tactmap_get_dimension and tactmap_get_center
  to match ModSAF3.0 libtactmap.h AUSTIN 4 Aug 97
  Don't think we need the get dimension
*/
/*
  tactmap_get_dimension (gpTactmap, &width, &height);
*/
   tactmap_get_center (gpTactmap, &xcenter, &ycenter, &ccenter);
   //printf("tactmap_get_center in MapControlsCB gets xy: %d %d\n",(int32)xcenter,(int32)ycenter);

/*
   tactmap_locale (gpTactmap, &xcenter, &ycenter, &width, &height);
*/
   sprintf (str, "%lf", xcenter);
   XtVaSetValues (fwXCenter,
		  XmNvalue, str,
		  NULL);
   sprintf (str, "%lf", ycenter);
   XtVaSetValues (fwYCenter,
		  XmNvalue, str,
		  NULL);

   for (i = 0; i < NUM_FEATURES; i++)
     XtVaSetValues (fFeature[i].wToggle,
		    XmNset, fFeature[i].pMapDisp->IsVisible(),
		    NULL);
}

void cMapControlsWindow::CbOK (Widget /*w*/, XtPointer  pObj, 
			       XmPushButtonCallbackStruct */*cbs*/)
{
   cMapControlsWindow* pWin = (cMapControlsWindow*)pObj;
   Widget              widget;
   XmString            xmstr;
   char               *str;
   int                 i;
   float64             xcenter, ycenter;
   cMinefield*         pMf;
   cVehicle*           pVeh;
   int                 forceInt;

   /* features */
   //printf("In MapControlsCb with %d features\n",NUM_FEATURES);
   for (i = 0; i < NUM_FEATURES; i++)
       pWin->fFeature[i].pMapDisp->SetVisible 
	 ((EBoolean)XmToggleButtonGetState (pWin->fFeature[i].wToggle));

   /* minefield toggles */

   for (i = 0; i < 3; i++)
     pWin->fMToggleSet[i] =
       (EBoolean)XmToggleButtonGetState (pWin->fwMinefieldToggle[i]);

   pMf = gMfieldMgr.GetField(1);
   while (pMf != NULL) {
     forceInt =  (int)pMf->GetForce() - 1;
     if ( forceInt < 0 ) forceInt = 2;
     pMf->ShowMinefield (pWin->fMToggleSet[forceInt]);
     pMf = pMf->GetNextMf();
   }

   /* vehicle toggles */

   for (i = 0; i < 3; i++)
     pWin->fVToggleSet[i] =
       (EBoolean)XmToggleButtonGetState (pWin->fwVehicleToggle[i]);

   pVeh = gVehMgr.GetVehList().GetFirstVeh();
   while (pVeh != NULL) {
     forceInt =  (int)pVeh->GetVehicleForce() - 1;
     if ( forceInt < 0 ) forceInt = 2;
     pVeh->ShowVehicle (pWin->fVToggleSet[forceInt]);
     pVeh = gVehMgr.GetVehList().GetNextVeh();
   }
     
   /* center */

   sscanf ((char*)XmTextGetString (pWin->fwXCenter), "%lf", &xcenter);
   sscanf ((char*)XmTextGetString (pWin->fwYCenter), "%lf", &ycenter);

   /* scale */

   XtVaGetValues (pWin->fwScaleMenu,
		  XmNmenuHistory, &widget,
		  NULL);
   XtVaGetValues (widget, 
		  XmNlabelString, &xmstr,
		  NULL);
   XmStringGetLtoR (xmstr, XmSTRING_DEFAULT_CHARSET, &str);

   for (i = 0; i < NUM_SCALES; i++) {
     if (strcmp (str, pWin->fScale[i].label) == 0) {
       gpGuiWin->SetMapScaleIndex (i);
       break;
     }
   }

   gpGuiWin->SetMapCenter ((int32)xcenter, (int32)ycenter, (int32)0);
   XtPopdown (((cMapControlsWindow*)pObj)->GetMainWin());
}


void cMapControlsWindow::CbCancel (Widget /*w*/, XtPointer pObj, 
				   XmPushButtonCallbackStruct */*cbs*/)
{
   XtPopdown (((cMapControlsWindow*)pObj)->GetMainWin());
}

