/*
$RCSfile: MapControlsCreate.cc,v $
$Revision: 1.1.1.1 $
$Author: jchappel $
$Date: 1999/02/16 14:54:36 $
$Log: MapControlsCreate.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:36  jchappel
Comprehensive Mine Simulator

// Revision 2.2  1998/04/20  15:10:36  ksnively
// *** empty log message ***
//
// Revision 2.1  1998/03/16  19:10:09  jchappel
// Added icon size sliders
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.3  1996/05/14  16:15:08  rcarp
// refinements and bug fixes for 1.5 delivery
//
// Revision 1.2  1996/03/16  18:36:36  mkappel
// Interface for Test Mode files.
//
// Revision 1.2  1996/03/16  18:36:36  mkappel
// Interface for Test Mode files.
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: MapControlsCreate.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   MapControlsCreate.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "MapControls.h"

#include <Xm/ArrowBG.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/LabelG.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>
#include <stdio.h>

struct scale_data cMapControlsWindow::fScale[] = {
  {1500000, "1:1,500,000", NULL } ,
  {1000000, "1:1,000,000", NULL } ,
  {500000,  "1:500,000",   NULL } ,
  {250000,  "1:250,000",   NULL } ,
  {100000,  "1:100,000",   NULL } ,
  {50000,   "1:50,000",    NULL } ,
  {25000,   "1:25,000",    NULL } ,
  {10000,   "1:10,000",    NULL } ,
  {5000,    "1:5,000",     NULL } ,
  {2500,    "1:2,500",     NULL } ,
  {1000,    "1:1,000",     NULL } ,
  {500,     "1:500",       NULL } ,
  {250,     "1:250",       NULL }
};

struct feature_data cMapControlsWindow::fFeature[] = {
  { TACTMAP_TREES     , "Trees and Canopies"  , NULL, NULL } ,
  { TACTMAP_CONTOURS  , "Contour lines"       , NULL, NULL } ,
  { TACTMAP_GRIDS     , "Grid lines"          , NULL, NULL } ,
  { TACTMAP_HYPSO     , "Hypsometric"         , NULL, NULL } ,
  { TACTMAP_WATER     , "Water and Soils"     , NULL, NULL } ,
  { TACTMAP_ROADS     , "Roads"               , NULL, NULL } ,
  { TACTMAP_BUILDINGS , "Buildings"           , NULL, NULL } ,
  { TACTMAP_PIPELINES , "Pipelines"           , NULL, NULL } ,
  { TACTMAP_POLITICAL , "Political Boundaries", NULL, NULL } ,
  { TACTMAP_RAILROADS , "Railroads"           , NULL, NULL } ,
  { TACTMAP_POWERLINES, "Powerlines"          , NULL, NULL } ,
  { TACTMAP_TOWNS     , "Towns"               , NULL, NULL } ,
};

EBoolean cMapControlsWindow::fMToggleSet[] = {true, true, true};
EBoolean cMapControlsWindow::fVToggleSet[] = {true, true, true};

struct pvd_default_features* cMapControlsWindow::fDefaultFeatures = NULL;

int cMapControlsWindow::GetFeatureIndex(TACTMAP_OBJECT_CLASS tmClass)
{
   for (int i = 0; i < NUM_FEATURES; i++) {
      if (fFeature[i].tmap_class == tmClass)
	 return i;
   }
   return -1;
}


void cMapControlsWindow::SetDefaultFeatures()
{
   char name_sans_version[256];
   int32 version;

   static READER_UNION default_features;

   int32 ret1 = reader_read("features.rdr", Datapath, &default_features,
			    0x0 & ~READER_TYPING);


   /* Get the default features */
   wld_get_playbox_id(name_sans_version,&version);
    
   if (!(fDefaultFeatures = (struct pvd_default_features *)
	  reader_find_tag(reader_get_symbol(name_sans_version),
			  default_features.array,
			  READER_UNTAGGED, READER_NO_ERRORS)))
    {
	fDefaultFeatures = (struct pvd_default_features *)
	  reader_find_tag(reader_get_symbol("default"),
			  default_features.array,
			  READER_UNTAGGED, READER_ERRORS_WITH_CONTEXT);
    }
}

void cMapControlsWindow::CheckFeatures(int32 index, int32 scale)
{
    uint32 mask;
    int32 mnr, mjr, grid;

    int tree_ind = GetFeatureIndex(TACTMAP_TREES);
    int contour_ind = GetFeatureIndex(TACTMAP_CONTOURS);
    int grid_ind = GetFeatureIndex(TACTMAP_GRIDS);

    //idiot check
    if (fDefaultFeatures == NULL)
       return;

    /* Set the bits for LINES and INDIVIDUAL trees appropriately
     * for this scale.
     */
    mask = fFeature[tree_ind].pMapDisp->fDfn.trees.mask;

    if (scale > fDefaultFeatures->highest_tree_line)
      mask &= ~TACTMAP_TREE_LINES;
    else
      mask |= TACTMAP_TREE_LINES;

    if (scale > fDefaultFeatures->highest_tree)
      mask &= ~TACTMAP_TREE_INDIVIDUAL;
    else
      mask |= TACTMAP_TREE_INDIVIDUAL;

    if (mask != fFeature[tree_ind].pMapDisp->fDfn.trees.mask)
    {
	fFeature[tree_ind].pMapDisp->fDfn.trees.mask = mask;
	tactmap_update_object(gpTactmap,
			      fFeature[tree_ind].pMapDisp->fpObj,
			      &fFeature[tree_ind].pMapDisp->fDfn);
    }



    /* Set the contour intervals and grid spacing appropriately
     * for this scale.
     */
    mnr = fDefaultFeatures->scales[index]->minor_contour;
    mjr = fDefaultFeatures->scales[index]->major_contour;

    if ((fFeature[contour_ind].pMapDisp->fDfn.contours.minor_spacing != mnr) ||
	(fFeature[contour_ind].pMapDisp->fDfn.contours.major_spacing != mjr))
    {
	fFeature[contour_ind].pMapDisp->fDfn.contours.minor_spacing = mnr;
	fFeature[contour_ind].pMapDisp->fDfn.contours.major_spacing = mjr;
	tactmap_update_object(gpTactmap,
			      fFeature[contour_ind].pMapDisp->fpObj,
			      &fFeature[contour_ind].pMapDisp->fDfn);
    }

    grid = fDefaultFeatures->scales[index]->grid_interval;

    if (fFeature[grid_ind].pMapDisp->fDfn.grids.interval != grid)
    {
	fFeature[grid_ind].pMapDisp->fDfn.grids.interval = grid;
	tactmap_update_object(gpTactmap,
			      fFeature[grid_ind].pMapDisp->fpObj,
			      &fFeature[grid_ind].pMapDisp->fDfn);
    }
	
    
}

Widget cMapControlsWindow::CreateWorkArea (Widget wParent)
{
   Widget        wWorkArea, wMenu, wFrame, wLabel, wRowcol, wMinescale, wVehscale;
   XmString      xmstr;
   Arg           args[2];
   int           i;

   wWorkArea = 
     XtVaCreateWidget ("work_area", 
		       xmFormWidgetClass, wParent, 
		       NULL);

   wMenu = XmCreatePulldownMenu (wWorkArea, "_pulldown", NULL, 0);
   xmstr = XmStringCreateSimple ("Map Scale");
   XtSetArg (args[0], XmNsubMenuId, wMenu);
   XtSetArg (args[1], XmNlabelString, xmstr);
   fwScaleMenu = 
     XmCreateOptionMenu (wWorkArea, "menu", args, 2);

   for (i = 0; i < NUM_SCALES; i++) {
     fScale[i].wOption =
       XtVaCreateManagedWidget (fScale[i].label,
				xmPushButtonGadgetClass, wMenu, 
				NULL);
   }
   XtManageChild (fwScaleMenu);
   XmStringFree (xmstr);

   XtVaSetValues (fwScaleMenu,
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_FORM,
		  NULL);

   wLabel = 
     XtVaCreateManagedWidget ("Center Point", 
			      xmLabelGadgetClass, wWorkArea, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, fwScaleMenu,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, fwScaleMenu,
			      XmNtopOffset, 10,
			      NULL);

   fwXCenter = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, wWorkArea, 
			      XmNcolumns, 7,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, wLabel,
			      XmNtopOffset, -5,
			      NULL);

   fwYCenter = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, wWorkArea, 
			      XmNcolumns, 7,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, fwXCenter,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, fwXCenter,
			      NULL);

   wLabel = 
     XtVaCreateManagedWidget ("Map Features:", 
			      xmLabelGadgetClass, wWorkArea, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, wLabel,
			      XmNtopOffset, 10,
			      NULL);
   wFrame = 
     XtVaCreateManagedWidget ("frame", 
			      xmFrameWidgetClass, wWorkArea,
			      XmNshadowThickness, 4, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_FORM,
			      XmNleftWidget, wLabel,
			      NULL);
   wRowcol = 
     XtCreateWidget ("rowcol", xmRowColumnWidgetClass, 
		     wFrame, NULL, 0);
   for (i = 0; i < NUM_FEATURES; i++)
     fFeature[i].wToggle = 
       XtVaCreateManagedWidget (fFeature[i].label,
				xmToggleButtonWidgetClass, wRowcol,
				NULL);
   XtManageChild (wRowcol);

   wLabel = 
     XtVaCreateManagedWidget ("Show Mineflds:", 
			      xmLabelGadgetClass, wWorkArea, 
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, wFrame,
			      XmNleftOffset, 10,
			      NULL);
   wFrame = 
     XtVaCreateManagedWidget ("frame", 
			      xmFrameWidgetClass, wWorkArea,
			      XmNshadowThickness, 4, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, wLabel,
			      NULL);
   wRowcol = 
     XtCreateManagedWidget ("rowcol", xmRowColumnWidgetClass, 
			    wFrame, NULL, 0);
   fwMinefieldToggle[0] = 
     XtVaCreateManagedWidget ("Friendly",
			      xmToggleButtonWidgetClass, wRowcol,
			      XmNset, fMToggleSet[0],
			      NULL);
   fwMinefieldToggle[1] = 
     XtVaCreateManagedWidget ("Enemy",
			      xmToggleButtonWidgetClass, wRowcol,
			      XmNset, fMToggleSet[1],
			      NULL);
   fwMinefieldToggle[2] = 
     XtVaCreateManagedWidget ("Other",
			      xmToggleButtonWidgetClass, wRowcol,
			      XmNset, fMToggleSet[2],
			      NULL);
   XtManageChild (wRowcol);
   
   wLabel = 
     XtVaCreateManagedWidget ("Show Vehicles:", 
			      xmLabelGadgetClass, wWorkArea, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wFrame,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, wFrame,
			      XmNtopOffset, 10,
			      NULL);
   wFrame = 
     XtVaCreateManagedWidget ("frame", 
			      xmFrameWidgetClass, wWorkArea,
			      XmNshadowThickness, 4, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, wLabel,
			      NULL);
   wRowcol = 
     XtCreateManagedWidget ("rowcol", xmRowColumnWidgetClass, 
			    wFrame, NULL, 0);
   fwVehicleToggle[0] = 
     XtVaCreateManagedWidget ("Friendly   ",
			      xmToggleButtonWidgetClass, wRowcol,
			      XmNset, fVToggleSet[0],
			      NULL);
   fwVehicleToggle[1] = 
     XtVaCreateManagedWidget ("Enemy",
			      xmToggleButtonWidgetClass, wRowcol,
			      XmNset, fVToggleSet[1],
			      NULL);
   fwVehicleToggle[2] = 
     XtVaCreateManagedWidget ("Other",
			      xmToggleButtonWidgetClass, wRowcol,
			      XmNset, fVToggleSet[2],
			      NULL);
   XtManageChild (wRowcol);

   wLabel = 
     XtVaCreateManagedWidget ("Icon Size:", 
			      xmLabelGadgetClass, wWorkArea, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wFrame,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, wFrame,
			      XmNtopOffset, 10,
			      NULL);
   wFrame = 
     XtVaCreateManagedWidget ("frame", 
			      xmFrameWidgetClass, wWorkArea,
			      XmNshadowThickness, 4, 
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, wLabel,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, wLabel,
			      NULL);

  
   wRowcol = 
     XtCreateManagedWidget ("rowcol", xmRowColumnWidgetClass, 
			    wFrame, NULL, 0);
  
   wMinescale = XtVaCreateManagedWidget("mines", xmScaleWidgetClass, wRowcol,
				    XtVaTypedArg, XmNtitleString, XmRString, "Mines", 6,
				    XmNmaximum,     20,
				    XmNminimum,     1,
				    XmNvalue,       10,
				    XmNshowValue,   True,
				    XmNorientation, XmHORIZONTAL,
				    NULL);
   
   XtAddCallback( wMinescale, XmNvalueChangedCallback, Mine_size, NULL);

   wVehscale = XtVaCreateManagedWidget("veh", xmScaleWidgetClass, wRowcol,
				    XtVaTypedArg, XmNtitleString, XmRString, "Vehicles", 9,
				    XmNmaximum,     20,
				    XmNminimum,     1,
				    XmNvalue,       10,
				    XmNshowValue,   True,
				    XmNorientation, XmHORIZONTAL,
				    NULL);

   XtAddCallback( wVehscale, XmNvalueChangedCallback, Veh_size, NULL);

   XtManageChild (wRowcol);
 
   XtManageChild( wWorkArea );
   return wWorkArea;
}

Widget cMapControlsWindow::CreateActionArea (Widget wParent)
{
   Widget  wActionArea, wButton;

   wActionArea = 
     XtVaCreateWidget ("action_area", 
		       xmFormWidgetClass, wParent,
		       XmNfractionBase, 9,
		       NULL);
   wButton = 
     XtVaCreateManagedWidget ("OK", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 1,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 4,
			      NULL);
   XtAddCallback( wButton, XmNactivateCallback, 
		 (XtCallbackProc)CbOK, (XtPointer)this);
   wButton = 
     XtVaCreateManagedWidget ("Cancel", 
			      xmPushButtonWidgetClass, wActionArea,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_POSITION,
			      XmNleftPosition, 5,
			      XmNrightAttachment, XmATTACH_POSITION,
			      XmNrightPosition, 8,
			      NULL);
   XtAddCallback( wButton, XmNactivateCallback, 
		 (XtCallbackProc)CbCancel, (XtPointer)this);

   {
     Dimension  h;
     XtVaGetValues (wButton, XmNheight, &h, NULL);
     XtVaSetValues (wActionArea, XmNpaneMaximum, h, XmNpaneMinimum, h, NULL);
   }
   XtManageChild( wActionArea );
   return wActionArea;
}

cMapControlsWindow::cMapControlsWindow(Widget parent)
   // the constructor creates the window and its widgets,
   // initializing with input values
{
   Widget        wPane, wWorkArea, wActionArea;

   fDefaultFeatures = NULL;

   fwMainWin = 
     XtCreatePopupShell( "Map Controls", transientShellWidgetClass,
			parent, NULL, 0);
   wPane = 
     XtVaCreateWidget ("pane", 
		       xmPanedWindowWidgetClass, fwMainWin,
		       NULL);

   wWorkArea = CreateWorkArea (wPane);
   wActionArea = CreateActionArea (wPane);
      
   XtManageChild( wPane );
}

cMapControlsWindow::~cMapControlsWindow()
{
}

void cMapControlsWindow::Mine_size(Widget /*scale*/, XtPointer /*client_data*/, XtPointer call_data)
{

   XmScaleCallbackStruct *cbs = (XmScaleCallbackStruct *) call_data;
   cMineDisp* Mag = (cMineDisp *)0;

   Mag->SetMineMag( (cbs->value * .00008) );

   if ( gbGui )
      tactmap_bundle_changes (gpTactmap);

   
   for ( cMinefield* pMf = gMfieldMgr.GetField(1); pMf; pMf = pMf->GetNextMf()) {
      pMf->RedrawMines();
   }
}

void cMapControlsWindow::Veh_size(Widget /*scale*/, XtPointer /*client_data*/, XtPointer call_data)
{

   XmScaleCallbackStruct *cbs = (XmScaleCallbackStruct *) call_data;
   cVehDisp* Mag = (cVehDisp *)0;

   Mag->SetVehMag( (cbs->value * .00008) );

}
