/*
$RCSfile: MineCompEditor.cc,v $
$Revision: 1.5 $
$Author: ksnively $
$Date: 1999/05/10 20:18:25 $
$Log: MineCompEditor.cc,v $
Revision 1.5  1999/05/10 20:18:25  ksnively
Added a command line option for enabling Communication Capability.

Revision 1.4  1999/04/29 19:34:18  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.3  1999/04/22 22:20:04  ksnively
Sreamlined radio communication routines.  Allowed to send to unlimited number of receivers.

Revision 1.2  1999/03/26 21:17:41  ksnively

Added radio communication capability for Command Detonated Mines.  Included sending
and receiving transmitter and signal PDUs.  All signals have the option of being
processed by the Communication Effects Federate.

Revision 1.1.1.1  1999/02/16 14:54:39  jchappel
Comprehensive Mine Simulator

// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.2  1997/12/17  14:38:03  ksnively
// Debugged Minefield duplication routine.
// Added input file "netparams" for reading PDU type enumerations.
// Got rid of warning messages during compilation.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.4  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.3  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/05/20  15:35:11  ksnively
// Redraw screen when turning on or off minefield
//
// Revision 1.2  1997/05/20  15:35:11  ksnively
// Redraw screen when turning on or off minefield
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.12  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.11  1996/06/22  13:04:11  rcarp
// deton vuln bug fixes for 1.5b4
//
// Revision 1.10  1996/05/23  18:42:07  mkappel
// Use DbMine list accessors.
//
// Revision 1.9  1996/05/15  15:01:14  mkappel
// Set fuze option menus.
//
// Revision 1.8  1996/05/11  21:36:32  mkappel
// Add mine age.
//
// Revision 1.7  1996/05/11  16:49:17  mkappel
// New mine type selector and fuze options.
//
// Revision 1.6  1996/05/08  00:48:49  rcarp
// changes for 1.5, incl user mines, mine params, param i/o, selection
//
// Revision 1.5  1996/04/04  16:25:34  mkappel
// Improve performance with large mine sets.
//
// Revision 1.4  1996/03/28  22:34:41  mkappel
// Add Mine Display toggle to component editors.
//
// Revision 1.3  1996/03/23  14:25:01  rcarp
// improvements to test mode and selection
//
// Revision 1.2  1996/03/13  00:12:33  mkappel
// Create scrolling set of toggles for mines.
//
// Revision 1.1  1996/01/23  16:52:30  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: MineCompEditor.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   MineCompEditor.cc
   
      The display interface implementation.
      
      
   
   Created by Michael Kappel
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include "Minefield.h"
#include "MineType.h"
#include "EntityIdEdit.h"

#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>


cMineCompEditorWindow::cMineCompEditorWindow (Widget wParent, EBoolean bController)
   : cCompEditorWindow (wParent)
{
   Widget        wLabel;

   fpMineType = new cMineTypeSelector (fwMainWin);
   XtVaSetValues (fpMineType->GetTypeForm(),
		  XmNtopAttachment, XmATTACH_WIDGET,
		  XmNtopWidget, fwMinefieldName,
		  XmNtopOffset, 5,
		  XmNleftAttachment, XmATTACH_FORM,
		  NULL);

   fpMineCoord = new cCoordList (fwMainWin, ecltMine);
   XtVaSetValues (fpMineCoord->GetCoordForm(),
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNtopOffset, 5,
		  XmNbottomAttachment, XmATTACH_WIDGET,
		  XmNbottomWidget, fwActionArea,
		  XmNbottomOffset, 10,
		  XmNrightAttachment, XmATTACH_FORM,
		  NULL);

   fwMineDispToggle = 
     XtVaCreateManagedWidget ("Display Mines",
			      xmToggleButtonWidgetClass, fwMainWin,
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwActionArea,
			      XmNbottomOffset, 10,
			      NULL);
   XtAddCallback (fwMineDispToggle, XmNvalueChangedCallback,
		  (XtCallbackProc)CbOnToggle, NULL);

   fwAgeText = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, fwMainWin, 
			      XmNcolumns, 7,
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwMineDispToggle,
			      XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNrightWidget, fwMineDispToggle,
			      NULL);
   wLabel = 
     XtVaCreateManagedWidget ("Age (hrs.)", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, fwAgeText,
			      XmNrightAttachment, XmATTACH_WIDGET,
			      XmNrightWidget, fwAgeText,
			      NULL);

   fwDepthText = 
     XtVaCreateManagedWidget ("textbox", 
			      xmTextFieldWidgetClass, fwMainWin, 
			      XmNcolumns, 7,
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, fwAgeText,
			      XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNleftWidget, fwAgeText,
			      NULL);
   wLabel = 
     XtVaCreateManagedWidget ("Mine Depth (in.)", 
			      xmLabelGadgetClass, fwMainWin,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
			      XmNtopWidget, fwDepthText,
			      XmNrightAttachment, XmATTACH_WIDGET,
			      XmNrightWidget, fwDepthText,
			      NULL);

   if (bController) {
      fpControllerId = new  cEntityIdTextEditor (fwMainWin);
      XtVaSetValues (fpControllerId->GetIdForm(),
		     XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
		     XmNbottomWidget, fpMineType->GetTypeForm(),
		     XmNbottomOffset, 10,
		     XmNleftAttachment, XmATTACH_WIDGET,
		     XmNleftWidget, fpMineType->GetTypeForm(),
		     XmNrightAttachment, XmATTACH_WIDGET,
		     XmNrightWidget, wLabel,
		     XmNrightOffset, 10,
		     XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
		     XmNtopWidget, fwDepthText,
		     //XmNheight, 150,
		     //XmNwidth, 200,
		     NULL);
   }
   else {
      fpControllerId = NULL;
   }

   XtAddCallback( fwDoneButton, XmNactivateCallback, 
		 (XtCallbackProc)CbDone, this);
   XtAddCallback( fwCancelButton, XmNactivateCallback, 
		 (XtCallbackProc)CbCancel, this);
}

cMineCompEditorWindow::~cMineCompEditorWindow ()
{
}

void cMineCompEditorWindow::ShowEdWindow ()
{
   cMineComp*     pMineComp = (cMineComp*)gpGuiWin->GetCurComponent();
   cMinefield*    pCurMinefield = gpGuiWin->GetCurMinefield();
   char           str[20];
   TLocation*     tccPts;
   int            i, nMines;

   Assert( pMineComp != NULL, kerrValidityCheck, 
	  "cMineCompEditorWindow::ShowEdWindow");

   XtVaSetValues (fwMinefieldName,
		  XmNvalue, pCurMinefield->GetMfName(),
		  NULL);

   fpMineType->SetMineTypeMenus (pMineComp);

   XtVaSetValues (fwMineDispToggle,
		  XmNset, pMineComp->GetShowMines(),
		  NULL);

   sprintf (str, "%.2f", pMineComp->GetCompMineDepth() * kMeterToInch);
   XtVaSetValues (fwDepthText,
		  XmNvalue, str,
		  NULL);

   sprintf (str, "%.2f", pMineComp->GetCompHoursInPlace());
   XtVaSetValues (fwAgeText,
		  XmNvalue, str,
		  NULL);

   fpMineCoord->DeleteEntries();

   nMines = pMineComp->GetNumMines();
   if (nMines > 0) {
//     tccPts = (TLocation*) malloc (nMines * sizeof(TLocation));
     tccPts = new TLocation [ nMines ];
     i = 0;
     for (cMine* pMine = pMineComp->GetFirstMine();
	  pMine != NULL; pMine = pMine->GetNextMine() )
       tccPts[i++] = pMine->GetTccLoc();

     fpMineCoord->AppendCoordEntries (tccPts, nMines);
     delete[] tccPts;
   }
   fpMineCoord->UpdateNumEntries();

   SetButtonState();
   SetWidgetState();
   XtManageChild (fwMainWin);
}

void cMineCompEditorWindow::SetButtonState ()
{
   int    ntypes = fpMineType->GetNumEntries();
   int    nmines = 0;
   
   cMineComp* pCurComp = (cMineComp*)(gpGuiWin->GetCurComponent());
   if ( pCurComp )
      nmines = pCurComp->GetNumMines();

   XtSetSensitive (fwDoneButton, (nmines > 0 && ntypes > 0));
}

void cMineCompEditorWindow::SetWidgetState ()
{
   //Set the sensitivity for the controller ID editor based on mine type
   if (fpMineType->GetDbMine()->GetFuzeParam(fpMineType->GetFuze())->IsCmdDet()) {
      //cout << "Setting Controller Id Sensitivity to TRUE" << endl;
      if (GetControllerIdWidget())
	 GetControllerIdWidget()->SetSensitive(true);
   }
   else {
      //cout << "Setting Controller Id Sensitivity to FALSE" << endl;
      if (GetControllerIdWidget())
	 GetControllerIdWidget()->SetSensitive(false);
   }
}

void cMineCompEditorWindow::CbOnToggle (Widget w, XtPointer /*pObj*/, 
					XmToggleButtonCallbackStruct */*cbs*/)
{
   cMineComp*     pMineComp = (cMineComp*)gpGuiWin->GetCurComponent();

   pMineComp->ShowMines ((EBoolean)XmToggleButtonGetState (w));
}

void cMineCompEditorWindow::CbDone (Widget /*w*/, XtPointer pObj, 
				    XmPushButtonCallbackStruct */*cbs*/)
{
  cMineCompEditorWindow*   pWin      = (cMineCompEditorWindow*)pObj;
  cMineComp*               pMineComp = (cMineComp*)gpGuiWin->GetCurComponent();
  float                    depth, age;
  TDbOptSet                OptSet;
  
  sscanf ((char*)XmTextGetString (pWin->fwDepthText), "%f", &depth);
  pMineComp->SetCompMineDepth (depth * kInchToMeter);
  sscanf ((char*)XmTextGetString (pWin->fwAgeText), "%f", &age);
  pMineComp->SetCompHoursInPlace (age);

  //Get the controller Id
  if (pWin->fpControllerId)
     pWin->fpControllerId->GetInputId();

  OptSet.fFuzeOpt = pWin->fpMineType->GetFuze();
  pMineComp->SetDbOpt (OptSet);
  pMineComp->SetFuzeOpt (pWin->fpMineType->GetFuzeOptions());

  pMineComp->StopEditingComp();
//  delete pWin->fpCompCopy;
  gpGuiWin->ShowEditor(cGuiWindow::eEditMinefield);
}

void cMineCompEditorWindow::CbCancel (Widget /*w*/, XtPointer /*pObj*/, 
				      XmPushButtonCallbackStruct */*cbs*/)
{
   cMfComponent* pCurComponent = gpGuiWin->GetCurComponent();
   
   if ( pCurComponent->HasCompBackup() ) {
      gpGuiWin->SetSelection( false, pCurComponent->GetMinefield(), pCurComponent );
      pCurComponent->RestoreBackup();
      pCurComponent->StopEditingComp();
   }
   
   else {
     gpGuiWin->SetSelection ( false, gpGuiWin->GetCurMinefield(), (cMfComponent*)NULL);
     cMfComponent** ppCompList = new (cMfComponent*);
     ppCompList[0] =  pCurComponent;
     gMfieldMgr.DeleteComponents (1, ppCompList);
   }

  gpGuiWin->ShowEditor(cGuiWindow::eEditMinefield);
}

cCoordList*    cMineCompEditorWindow::GetMineCoordWidget () 
{
   return fpMineCoord;
}

cEntityIdTextEditor*    cMineCompEditorWindow::GetControllerIdWidget () 
{
   return fpControllerId;
}

void  cMineCompEditorWindow::NewMine (TLocation& loc)
{
   fpMineCoord->AppendCoordEntry (loc);
}

void  cMineCompEditorWindow::NewMines (TLocation* tccPts, int nMines)
{
   fpMineCoord->AppendCoordEntries (tccPts, nMines);
}

void   cMineCompEditorWindow::DeleteObjEntry(cMfComponent* pComp, int index)
{
//   Assert ( pComp == gpCurComponent, kerrValidityCheck, "cMineCompEditorWindow::DeleteObjEntry");

   if (pComp == gpGuiWin->GetCurComponent() ) {
     fpMineCoord->DeleteEntry(index);
     SetButtonState ();
   }
}
