/*
$RCSfile: Perimeter.cc,v $
$Revision: 1.2 $
$Author: ksnively $
$Date: 1999/04/29 19:34:22 $
$Log: Perimeter.cc,v $
Revision 1.2  1999/04/29 19:34:22  ksnively
Incorporated ARDEC addtions:  Tripwire AP mines and a Control Window.

Revision 1.1.1.1  1999/02/16 14:54:42  jchappel
Comprehensive Mine Simulator

// Revision 2.2  1998/11/25  18:37:51  ksnively
// Update for APLA Exercise
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/10/02  17:53:37  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.2  1997/10/02  17:50:22  austin
// first working port to modsaf3.0
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.5  1996/09/06  18:57:57  rcarp
// CP bug fixes
//
// Revision 1.4  1996/08/15  19:15:44  mkappel
// Change perimeter color when minefield state changes.
//
// Revision 1.3  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.2  1996/03/23  14:22:31  rcarp
// improvements to test mode and selection
//
// Revision 1.2  1996/03/23  14:22:31  rcarp
// improvements to test mode and selection
//
$State: Exp $
*/
 
/*******************************************

   file: Perimeter.cc
   
      Copyright 1995 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   Perimeter.cc
   
      The minefield implementation.
      
      
   
   Created by Richard Carpenter
   Modification history:
      created   
      
*********************************************/

#include "Display.h"
#include "MapGeometry.h"
#include "Message.h"
#include "Minefield.h"

/* the next two functions are from Sedgewick 'Algorithms in C' */
/* line intersect helper function looks at slopes */

static int ccw (TLocation p0, TLocation p1, TLocation p2)
{
   float64 dx1 = p1.x - p0.x;
   float64 dy1 = p1.y - p0.y;
   float64 dx2 = p2.x - p0.x;
   float64 dy2 = p2.y - p0.y;

   if((dx1*dy2) > (dy1*dx2)) return  1;
   if((dx1*dy2) < (dy1*dx2)) return -1;

   if((dx1*dx2 < 0) || (dy1*dy2 < 0)) return -1;

   if((dx1*dx1+dy1*dy1) < (dx2*dx2+dy2*dy2)) return  1;

   return 0;
}

/* do lines extended from these two segments intersect ? */
int intersect (TLocation p11, TLocation p12, TLocation p21, TLocation p22)
{
   return ((ccw (p11, p12, p21) * ccw (p11, p12, p22)) < 0)
     &&   ((ccw (p21, p22, p11) * ccw (p21, p22, p12)) < 0);
}

void       TPerimPoint::SetTccPoint(const TLocation& loc)
{
   fTccPoint = loc;
   gpTerrainMgr->MapToSimLoc(loc, &fGeoPoint);
}

TPerimPoint::TPerimPoint(const TFlatEntCoord& mfPt, T2link* pPrev) :
   T2link(pPrev)
{
   fMfPoint.x = mfPt.x;
   fMfPoint.y = mfPt.y;
   fMfPoint.z = 0.0;         // haven't translated to fcc coords yet
//   fMfPoint.z = gpTerrainMgr->GetMapElev(fMfPoint.x, fMfPoint.y);
}

TPerimPoint::TPerimPoint(const TLocation& tccPt, T2link* pPrev) :
   T2link(pPrev)
{
   fTccPoint = tccPt;
   
   gpTerrainMgr->MapToSimLoc(tccPt, &fGeoPoint);
}

TPerimPoint::~TPerimPoint()
{
}

EBoolean    cPerimeter::AddTccPoint (TLocation& newPt)
{
   /* is new point valid ? */
  if (! ValidPerim (newPt))
    return false;

  newPt.z = gpTerrainMgr->GetMapElev( newPt.x, newPt.y);
   
  new TPerimPoint (newPt, fPointList.GetLastLink());
  ++fNumPoints;
  fpParentMf->DoMfCoords();
  fpParentMf->ResetAllGrids();
   
  if ( fpPerimDisp )
    fpPerimDisp->AddVertex (fNumPoints, newPt);
  if ( gpGuiWin )
    gpGuiWin->SetEditorButtonState ();

  return true;
}

EBoolean    cPerimeter::DeletePoint (int iPt)
{
  fPointList.DeletePoint (iPt);
  fNumPoints--;
  fpPerimDisp->DeleteVertex (iPt - 1);
  fpParentMf->ResetComponents();
  return true;
}

EBoolean    cPerimeter::MoveComp (T2dMove& move)
   // moves component, called from cMinefield::MoveMinefield.  
   // Since this is perimeter, all moves move the entire minefield
   //   NOTE: we move with tcc ref, which allows the minefield to set
   // its new coordinate system based on the new perimeter.
{
  TPerimPoint*  pPt;
  TLocation     tccPt;
  TLocation     mfCenter = fpParentMf->GetTccLoc(); // still old loc
  TEntCoord     mfLoc;
  int           iPt = 0;

  fRefSet = eTccRef;
  
  if ( ! move.IsNoMove() ) {
     for ( pPt = GetFirstPoint(); pPt; pPt = GetNextPoint() ) {
       move.fRotMatrix.Mult(pPt->GetTccPoint() - mfCenter, &tccPt);
       tccPt = tccPt + mfCenter;
       tccPt.x += move.fDelX;
       tccPt.y += move.fDelY;
       tccPt.z =  gpTerrainMgr->GetMapElev(tccPt.x, tccPt.y);
       pPt->SetTccPoint(tccPt);
       
       if (fpPerimDisp) 
          fpPerimDisp->MoveVertex (iPt++, tccPt);
     }
  }

  return true;
}

/*
EBoolean    cPerimeter::MovePerim (TLocation offset)
{
  TPerimPoint*  pPt;
  TLocation     tccPt;
  int           iPt = 0;

  for ( pPt = GetFirstPoint(); pPt; pPt = GetNextPoint() ) {
    tccPt = pPt->GetTccPoint();
    tccPt.x += offset.x;
    tccPt.y += offset.y;
    tccPt.z =  gpTerrainMgr->GetMapElev(tccPt.x, tccPt.y);
    pPt->SetTccPoint(tccPt); 

    fpPerimDisp->MoveVertex (iPt++, tccPt);
  }

  fpParentMf->ResetCoords();

  return true;
}
*/

EBoolean    cPerimeter::MovePoint (int iPt, TLocation newPt)
{
      //** need to set mf loc also, if we are complete.
      //** what if we are mf based?
  TPerimPoint*  curPt, *prevPt, *nextPt;
  int           i;

   /* is new location valid ? */
  if (fNumPoints > 3) {

    /* first check edge between iPt and iPt + 1 */
    if (iPt == fNumPoints)
      nextPt = GetFirstPoint();
    else
      nextPt = (TPerimPoint*) (fPointList.GetLinkNum(iPt + 1));

    prevPt = GetLastPoint();

    i = 0;
    for ( curPt = GetFirstPoint(); curPt; curPt = GetNextPoint() ) {
      i++;
      if (i != iPt && i != (iPt % fNumPoints) + 1  && 
	  i != ((iPt + 1) % fNumPoints) + 1) {

	if (intersect (prevPt->GetTccPoint(), curPt->GetTccPoint(),
		       nextPt->GetTccPoint(), newPt)) {
	  if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
				   "Invalid move.  Minefield\nperimeter cannot cross itself.");
	  return false;
	}
      }
      prevPt = curPt;
    }

    /* next check edge between iPt - 1 and iPt */

    if (iPt == 1)
      nextPt = GetLastPoint();
    else
      nextPt = (TPerimPoint*) (fPointList.GetLinkNum(iPt - 1));

    prevPt = GetLastPoint();

    i = 0;
    for ( curPt = GetFirstPoint(); curPt; curPt = GetNextPoint() ) {
      i++;
      if (i != iPt && i != (iPt % fNumPoints) + 1  && 
	  (i % fNumPoints) != iPt - 1) {

	if (intersect (prevPt->GetTccPoint(), curPt->GetTccPoint(), 
		       nextPt->GetTccPoint(), newPt)) {
	  if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
				   "Invalid move.  Minefield\nperimeter cannot cross itself.");
	  return false;
	}
      }

      prevPt = curPt;
    }
  }

  newPt.z =  gpTerrainMgr->GetMapElev(newPt.x, newPt.y);
  fpPerimDisp->MoveVertex (iPt - 1, newPt);
  nextPt = (TPerimPoint*) fPointList.GetLinkNum(iPt);
  nextPt->SetTccPoint(newPt); 

  fpParentMf->ResetComponents();
  return true;
}

void    cPerimeter::ReadMfPerim( TItemList* list )
{
   fRefSet    = eMfRef;
   fNumPoints = list->count;

   fPointList.DeleteList();
   
   for (int i = 0; i < fNumPoints; i++) 
      new TPerimPoint( ((TFlatEntCoord *)(list->ptrItems))[i], fPointList.GetLastLink());
   
   delete[] list->ptrItems;
   delete list;
}

void    cPerimeter::ReadMapPerim( TItemList* list )
{
   fRefSet    = eTccRef;
   fNumPoints = 0;

   fPointList.DeleteList();
   
   for (int i = 0; i < list->count; i++) {
      AddTccPoint( ((TLocation *)(list->ptrItems))[i]);
   }
   
   delete[] list->ptrItems;
   delete list;
}

void     cPerimeter::Read2dMapPerim( TItemList* list )
{
   fRefSet    = eTccRef;
   fNumPoints = 0;
   TLocation   loc;
   
   fPointList.DeleteList();
   
   for (int i = 0; i < list->count; i++) {
      loc.x = ((T2Dsingle *)(list->ptrItems))[i] .x;
      loc.y = ((T2Dsingle *)(list->ptrItems))[i] .y;
      loc.z = gpTerrainMgr->GetMapElev(loc.x, loc.y);
      
      AddTccPoint( loc );
   }
   
   delete[] list->ptrItems;
   delete list;
}

void    cPerimeter::ComputePerim()
{

}

void    cPerimeter::ShowPerim (EBoolean show)
{
  if (fpPerimDisp)
    fpPerimDisp->SetVisible (show);
}

void    cPerimeter::SetPerimSelect (EBoolean on)
{
  if (fpPerimDisp)
    fpPerimDisp->SetSelectSens (on);
}

EBoolean    cPerimeter::CompleteComp()
{
   TPerimPoint* pPt;
   int   i;
   
   switch (fRefSet) {
      case eTccRef:
            for ( pPt = (TPerimPoint*) (fPointList.GetFirstLink()); 
                  pPt != (TPerimPoint*) &fPointList; 
                  pPt = (TPerimPoint*) (fPointList.GetNextLink()) ) 
               fpParentMf->MapToMfLoc( pPt->GetTccPoint(), &(pPt->GetMfPoint()) );
         break;
         
      case eMfRef:
            for ( pPt = (TPerimPoint*) (fPointList.GetFirstLink()), i = 1; 
                  pPt != (TPerimPoint*) &fPointList; 
                  pPt = (TPerimPoint*) (fPointList.GetNextLink()) ) {
                     // may have 2D mf ref
               TFlatEntCoord loc2D(pPt->GetMfPoint().x, pPt->GetMfPoint().y);
               TLocation tccLoc;
               fpParentMf->MfToMapLoc( loc2D, &tccLoc );
               pPt->SetTccPoint(tccLoc);
               fpParentMf->MapToMfLoc(tccLoc, &(pPt->GetMfPoint()) );
               
               if ( fpPerimDisp ) {
                  fpPerimDisp->AddVertex (i++, pPt->GetTccPoint());
               }
            }
	    if ( fpPerimDisp ) fpPerimDisp->CloseLoop (i);
         break;

      case eRefComplete:
         break;
         
      default:
            return false;
         // break;
   }
   
   fRefSet = eRefComplete;
   if (fNumPoints < 3)
      return false;
   else
      return true;
}

void        cPerimeter::SaveComponent(cOutFileObj* pOutfile)
{
   pOutfile->WriteComment("saving perimeter in 2 dim minefield coords");
   
   TFlatEntCoord* points = new TFlatEntCoord[fNumPoints];
   int i = 0; 
   
   for (T2link* pPerimPt = fPointList.GetFirstLink();
                pPerimPt != fPointList.GetNULLLink();
                pPerimPt = fPointList.GetNextLink()) {
      points[i].x   = ((TPerimPoint*) pPerimPt)->GetMfPoint().x;
      points[i++].y = ((TPerimPoint*) pPerimPt)->GetMfPoint().y;
   }
   
   TItemList ptlist;
   ptlist.count = fNumPoints;
   ptlist.ptrItems = (void *)points;

   pOutfile->WriteItem( cMinefield::fKeyMfldTable[cMinefield::emfkwMfPerim].key,
                       cMinefield::fKeyMfldTable[cMinefield::emfkwMfPerim].type,
                       (void*)(&ptlist) );
   delete[] points;
}

void    cPerimeter::OnOffComponent (EBoolean on)   
{
  if (fpPerimDisp)
    fpPerimDisp->SetStateColor (on);
}

EBoolean    cPerimeter::GetMfGeoLoc(TLocation* pLoc)
{
   if ( fNumPoints > 0 ) {
      double xMin = DBL_MAX;
      double xMax = - DBL_MAX;
      double yMin = DBL_MAX;
      double yMax = - DBL_MAX;
      double zMin = DBL_MAX;
      double zMax = - DBL_MAX;
      
      for ( TPerimPoint* pPt = (TPerimPoint*) (fPointList.GetFirstLink()); 
                         pPt != (TPerimPoint*) &fPointList; 
                         pPt = (TPerimPoint*) (fPointList.GetNextLink()) ) {
         TLocation tccPt = pPt->GetTccPoint();
         if (tccPt.x < xMin) xMin = tccPt.x;
         if (tccPt.x > xMax) xMax = tccPt.x;
         if (tccPt.y < yMin) yMin = tccPt.y;
         if (tccPt.y > yMax) yMax = tccPt.y;
         if (tccPt.z < zMin) zMin = tccPt.z;
         if (tccPt.z > zMax) zMax = tccPt.z;
      }
      
      TLocation mapLoc;
      mapLoc.x = (xMin + xMax) / 2.0;
      mapLoc.y = (yMin + yMax) / 2.0;
      mapLoc.z = (zMin + zMax) / 2.0;
      
      gpTerrainMgr->MapToSimLoc(mapLoc, pLoc);
   }
   else {
      pLoc->x = 0.0;
      pLoc->y = 0.0;
      pLoc->z = 0.0;
   }
   if ( fNumPoints >= 3 )
      return true;
   else
      return false;
}

EBoolean    cPerimeter::ValidPerim (TLocation& newPt)
{
  TPerimPoint  *curPt, *prevPt, *lastPt;

  if (fNumPoints > 2) {
    lastPt = GetLastPoint();
    prevPt = GetFirstPoint();

    for ( curPt = GetNextPoint(); curPt != lastPt; curPt = GetNextPoint() ) {
      if (intersect (prevPt->GetTccPoint(), curPt->GetTccPoint(), 
		     lastPt->GetTccPoint(), newPt)) {
         if (gbGui) gpGuiWin->DialogMessage (cGuiWindow::eDialogError, NULL,
				 "Invalid vertex.  Minefield\nperimeter cannot cross itself.");
         return false;
      }
      prevPt = curPt;
    }
  }
  return true;
}

void   cPerimeter::StartEditingComp()
{
   feIsEditing = true;
   MakeBackup();   // causes all comps to backup
}

void    cPerimeter::StopEditingComp()
   // called when leaving perim editor window, set/reset mf coord system
{
   fpParentMf->ResetCoords();
   fpCompEntry->ShowInfo();

   feIsEditing = false;

   if (fpBackup) {
      delete fpBackup;
      fpBackup = NULL;
      fpParentMf->DeleteMineCompBackups();
   }

   if (fpPerimDisp) {
     fpPerimDisp->SetSensitive (false);
     fpPerimDisp->CloseLoop (fNumPoints + 1);
   }
   if (gbGui) gpGuiWin->SetSelection( false, fpParentMf, this);
}

TPerimPoint*  cPerimeter::GetFirstPoint ()  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetFirstLink();

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

TPerimPoint*  cPerimeter::GetNextPoint ()  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetNextLink();

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

TPerimPoint*  cPerimeter::GetLastPoint ()  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetLastLink();

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

TPerimPoint*  cPerimeter::GetPointNum (int iPt)  
{
  TPerimPoint* pPoint = (TPerimPoint*)fPointList.GetLinkNum (iPt);

  if (pPoint == fPointList.GetNULLLink())
    return NULL;
  else
    return pPoint;
}

EBoolean  cPerimeter::PointInside (TLocation newPt)
{
  TPerimPoint        *pPt, *lastPt;
  TLocation           tccPt, prevPt, infPt;
  int                 count = 0;

  // Need to have TCC coordinates set
  if (fRefSet == eMfRef)
     CompleteComp();

  infPt = newPt;
  infPt.x += 999999.;
  infPt.y += 999999.;
  lastPt = GetLastPoint();
  prevPt = lastPt->GetTccPoint();

  // Choose an infPt so no Perimeter vertex lies on the line 
  // from newPt to infPt.  Otherwise the algorithm is invalid.
  for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
     if (pPt->GetTccPoint().y > newPt.y ) {
	tccPt = pPt->GetTccPoint();
	if (tccPt.y < infPt.y)
	   infPt.y = tccPt.y;
     }
  }

  for (pPt = GetFirstPoint(); pPt; pPt = GetNextPoint()) {
    tccPt = pPt->GetTccPoint();
    //if (! intersect (tccPt, tccPt, newPt, infPt))
      if (intersect (prevPt, tccPt, newPt, infPt)) count++;

    prevPt = tccPt;
  }

  if (count & 1)
    return true;
  else
    return false;    
}

cMfComponent*    cPerimeter::DuplicateComp(T2dMove& move, cMinefield* pParent,
					   EBoolean isBackup)
{
   cPerimeter* pC = new cPerimeter(this, pParent, isBackup);

   if ( ! isBackup) {
     if ( ! move.IsNoMove() )
        pC->MoveComp(move);
     if (fpCompEntry)
       fpCompEntry->ShowInfo();
     if (fpCompStatus)
       fpCompStatus->ShowInfo();
   }
   return pC;
}

void     cPerimeter::MakeBackup()
{
   if ( fpBackup )
      delete fpBackup;
   fpBackup = new cPerimeter(this, NULL, true);
   fpParentMf->MakeMineCompBackups();
}

void     cPerimeter::RestoreBackup()
{
   cPerimeter* pBackup = (cPerimeter*) fpBackup;
   
   if ( pBackup ) {
         // delete existing mines and rows, but don't update displays yet
      if ( fpPerimDisp )
         for ( int i = fNumPoints; i > 0; i-- )
            fpPerimDisp->DeleteVertex(i);
//         fpPerimDisp->DeleteAll();
      fPointList.DeleteList();   // deletes each point
      fNumPoints = 0;
      
      for (TPerimPoint* pPt = pBackup->GetFirstPoint(); pPt; 
                           pPt = pBackup->GetNextPoint() ) {
//         AddTccPoint(pPt->GetTccPoint());
         new TPerimPoint (pPt->GetTccPoint(), fPointList.GetLastLink());
         ++fNumPoints;
         if ( fpPerimDisp )
           fpPerimDisp->AddVertex (fNumPoints, pPt->GetTccPoint());
      }
      fpParentMf->DoMfCoords();
      fpParentMf->ResetAllGrids();

      fpParentMf->RestoreMineCompBackups();
      
      UpdateStatusInfo();
   }
}

cPerimeter::cPerimeter(cPerimeter* pOriginal, cMinefield* pParent,
		       EBoolean isBackup) : 
      cMfComponent(pOriginal, pParent, isBackup)
    // used when duplicating or backing up minefields
{
   fNumPoints  = 0;
   fCompType   = pOriginal->fCompType;
   fRefSet     = pOriginal->fRefSet;
   
   if ( ( ! isBackup) && gbGui) {
      fpPerimDisp = new cPerimDisp (this);
      fpCompEntry = new cPerimSummLine
         (this, pParent->GetCompList()->GetMainWin());
   }
   else  {
      fpPerimDisp = NULL;
      fpCompEntry = NULL;
   }
   
   for ( TPerimPoint* pPt = pOriginal->GetFirstPoint();
         pPt;
         pPt = pOriginal->GetNextPoint() ) {
   
      new TPerimPoint (pPt->GetTccPoint(), fPointList.GetLastLink());
      ++fNumPoints;
      if ( fpPerimDisp )
        fpPerimDisp->AddVertex (fNumPoints, pPt->GetTccPoint());
   }
   
   if ( fpPerimDisp ) {
     fpPerimDisp->SetSensitive (false);
     fpPerimDisp->CloseLoop (fNumPoints + 1);
   }
   
   if (pParent) {
      pParent->DoMfCoords();
      pParent->ResetAllGrids();
   }
}

cPerimeter::cPerimeter(cMinefield* pParent) : cMfComponent(pParent)
{
   fNumPoints    = 0;
   fCompType = eCompPerim;
   fRefSet    = eTccRef;      // default, when created by gui
   
   if (gbGui) {
      fpPerimDisp = new cPerimDisp (this);
      fpCompEntry = new cPerimSummLine
         (this, pParent->GetCompList()->GetMainWin());
   }
   else  {
      fpPerimDisp = NULL;
      fpCompEntry = NULL;
   }
}

cPerimeter::~cPerimeter()
{
   if (gbGui) {
      if (fpPerimDisp) {
         delete fpPerimDisp;
         fpPerimDisp = NULL;
      }
   }
   fPointList.DeleteList();
   
}

