/**
*
* @author H. Timucin Ozdemir
* @date   September 14, 1998 ( Last Updated September 15, 1995)
*
*/

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.net.*;
import java.io.*;

import org.omg.CORBA.*;

import HeartBeat.*;

public class 
HeartBeatServant extends HeartBeat._HeartBeatListenerImplBase 
                 implements HBLDebug
{
    /** */
    public static final int SLAVE     = 0;
    public static final int CENTRAL   = SLAVE+1;
    public static final int G_CENTRAL = CENTRAL+1;

    /** */
    String hostName;

    /** */
    int registers = 0;

    /** */
    Hashtable maps;

    /** */
    Hashtable simaps;

    /** */
    ServerInfoList siList;

    /** */
    HeartBeat.HeartBeatListener central;
    String                      central_name=null;
    org.omg.CORBA.StringHolder  cnHolder=new org.omg.CORBA.StringHolder("");
    org.omg.CORBA.StringHolder  nHolder=new org.omg.CORBA.StringHolder("");

    int   role = SLAVE;
    String  myId;
    int     myNum;
    Ticker  tick;

    RegisterBroadcaster rgBroadcast;
    StrikeOutUpdater    strikeBroadcast;

    HeartBeatDaemon     master=null;

    long  timeLimit   = 5000;
    int   strikeLimit = 3;

    /**
    */
    public HeartBeatServant() 
    {
        this(null);

    }// end of CONSTUCTOR

    /**
    */
    public HeartBeatServant(HeartBeat.HeartBeatListener _central) 
    {
       maps   = new Hashtable();
       simaps = new Hashtable();
       hostName = getMachineName();
       siList = new ServerInfoList();
       //head = tail = null;
       central = _central;
       org.omg.CORBA.ORB.init().connect(this);
       if( central!=null ) 
         try
         {
            myNum = central.joinGroup((HeartBeat.HeartBeatListener)this,nHolder,cnHolder);
            central_name = cnHolder.value;
            myId         = nHolder.value;
            registers    = myNum;
            debug(" My Master is "+central_name);

         } catch(org.omg.CORBA.SystemException se)
           {  central=null; role=CENTRAL; myId = hostName; central_name=null; } 
           catch(Exception se)
           {  central=null; role=CENTRAL; myId = hostName; central_name=null; } 

       if( central==null ) 
       {
           myId = hostName+"_central";
           role=CENTRAL;
           myNum=0;
           registers++;

       }// end of if

       if(central!=null)  register(central_name,0,central);

       strikeBroadcast = new StrikeOutUpdater(this);
       rgBroadcast = new RegisterBroadcaster(); 
       tick = new Ticker(this,3000);
       tick.start();

       debug(" HeartBeatServant is started! ");

       switch(role)
       {
         case CENTRAL   : debug(" I am the CENTRAL! "); 
                          publishMyAddress("hbl.ior",
                             org.omg.CORBA.ORB.init().object_to_string(this));
                          break;

         case SLAVE     : debug(" I am the SLAVE! "); break;

         case G_CENTRAL : debug(" I am the Group CENTRAL! "); break;

              default   : debug(" I am WEIRD! "); break;

       }// end of switch()

    }// end of CONSTUCTOR

    /**
    * <ol>
    *  <p><li> define a name for this object.
    *  <p><li> register this object to the records.
    *  <p><li> if HBL is a CENTRAL role, then broadcast this 
    *          registration to all the group members.
    *  <p><li> return a unique id to this new comer.
    * <ol>
    */
    public int 
    joinGroup( HeartBeat.HeartBeatListener sa, org.omg.CORBA.StringHolder sname,
               org.omg.CORBA.StringHolder ret_name) 
    {
       int retVal = -1;

       switch(role)
       {
          case SLAVE     : // Forward to Central
                           retVal = central.joinGroup(sa,sname,ret_name);
                           break;

          case CENTRAL   : // Accept Registering
                           String ret = hostName+registers;
                           registers++;
                           register( ret, registers-1, sa );
                           broadcast( ret, registers-1, sa );
                           //rgBroadcast.broadcast( this, ret, sa , maps );
                           ret_name.value = myId;
                           sname.value = ret;
                           retVal = registers-1;
                           break;

          case G_CENTRAL : // Send lower level
                           /*Enumeration keys = lower_maps.keys();
                           if( keys.hasMoreElements() )
                           {
                             HeartBeat.HeartBeatListener low = 
                               (HeartBeat.HeartBeatListener) lower_maps.get(keys.nextElement());
                             retVal = low.joinGroup(sa,sname,ret_name);


                           }// end of if
                           */
                             
                           break;

       }// end of switch

       return retVal;

    }// end of joinGroup()

    /**
    */
    public void 
    leave(String HBname) 
    {
       delete(HBname);

    }// end of leave()

    /**
    *
    * Processes the alive message coming from server <b>HBname</b>.
    *
    * <ol>
    *  <p><li> get record about this server.
    *  <p><li> if it knows this server, then update its last check-in
    *          time and resets its strike-out paramater. 
    *  <p><li> if it does not know this server, then ? nothing
    * </ol>
    *
    */
    public void 
    alive(String HBname) 
    {
       ServerInfo sv = (ServerInfo) simaps.get( HBname );
       debug(" ALIVE Message is received for "+HBname);

       if( sv!=null )
       {

          debug(HBname+" is alive! <<<<< ");
          siList.removeFromList(sv);
          sv.setTime( System.currentTimeMillis() );
          sv.reset();
          siList.addFromTail(sv);

       }// end of if

    }// end of alive()

    /**
    * A server is asked with this message to check whether it is runnning
    * or not. If it is gonna be strike-out from some server's records, it
    * will first asks the object to make sure that it really is out of 
    * the loop.
    */
    public boolean 
    areYouAlive() 
    {
       boolean ret=true;
       return ret;

    }// end of areYouAlive()

    /**
    * 
    * <ol> 
    *  <p><li> gets the record about this new central server. If it does
    *          not find anything, then it defines one. 
    *  <p><li> updates its central server information.
    * </ol> 
    * 
    */
    public void 
    replace( String prevCntName, String newCntName, 
             HeartBeat.HeartBeatListener currCentralServerAddress) 
    {
       HeartBeat.HeartBeatListener hb = 
           (HeartBeat.HeartBeatListener) maps.get(newCntName);

       if(hb==null) register(newCntName,0,currCentralServerAddress);

       central_name = new String(newCntName);
       central = currCentralServerAddress;

    }// end of replace()

    /**
    */
    public void 
    updateGroup( HeartBeat.HeartBeatListener centralServerAddress, 
                 String[] group) 
    {
    }// end of updateGroup()

    /**
    * A server registry message. A central object is supposed to send this
    * message to group members to anounce the presence of new server.
    *     
    * <ol>
    *  <p><li> Check whether it knows this server
    *   <ol>
    *    <p><li> If it does know this server,
    *     <ol>
    *      <p><li> define a record for this new server.
    *      <p><li> adds this record to the list.
    *      <p><li> updates its counter according to the given server's id
    *              information so that it can track down the next number 
    *              it should be assigned to the new guy.
    *     </ol>
    *   </ol>
    * </ol>
    */
    public void 
    register( String HBname, int id, HeartBeat.HeartBeatListener hbListener)
    {
       maps.put(HBname,hbListener);
       ServerInfo sv = new ServerInfo( HBname, id, System.currentTimeMillis());
       simaps.put(HBname,sv);
       siList.addFromTail(sv);

       // In case this will be a CENTRAL
       if( registers<=id ) registers=id+1;

    }// end of register()

    /** Utilities */

    /**
    * A timer of this server sends this message to HBL.
    * <ol>
    *  <p><li> If it has a HBD to report to , then it reports the recent
    *          states of servers it knows of.
    * </ol>
    */
    public void 
    ticked() 
    {  
       debug(" HeartBeatServant ticked! ");

       // Process Strike-outs 
       strikeBroadcast.update(siList.enumerateList().elements(),strikeLimit,timeLimit);

       if(master!=null) master.report(produceReport());
       else debug(" master is null! ");

    }// end of ticked()

    /**
    * Prepares a report about each server including itself.
    */
    final Enumeration produceReport() 
    {  
       ServerInfo me  = new ServerInfo(getName(),myNum,System.currentTimeMillis());
       Vector a = siList.enumerateList();
       a.addElement( me );

       return a.elements();

    }// end of ticked()

    /**
    */
    public HeartBeatDaemon getMaster() { return master; }

    /**
    */
    public void setMaster(HeartBeatDaemon _master) { master=_master; }

    /**
    */
    public String getName() { return myId; }

    /**
    */
    public Enumeration getBudies() { return maps.elements(); }

    /**
    */
    public HeartBeat.HeartBeatListener 
    getHBL( String HBLname ) 
    { 
       return (HeartBeat.HeartBeatListener) maps.get(HBLname);

    }// end of getHBL()

    /**
    */
    String getMachineName()
    {
       String ret = null;
       InetAddress addr;

       try
       {
         addr = InetAddress.getLocalHost();

         ret = addr.getHostName();
         if (ret == null || ret.indexOf('.') == -1)
         {
            /****************************************************/
            /** If host does not appear to be fully qualified,  */
            /** then return just the inet address instead.      */
            /****************************************************/
            ret = addr.getHostAddress();

         }// end of if

       } catch (UnknownHostException e) { ret=null; }

       return ret;

    }// end of getMachineName()

    /**
    * prints the address information to the given file. It is useful when
    * the HBL becomes CENTRAL so that it can publish its address 
    * information to the file.
    */
    final void publishMyAddress(String fileName, String myAddr)
    {
       try
       {
          PrintWriter pw = new PrintWriter( new FileWriter(new File(fileName)));
          pw.println(myAddr);
          pw.close();

       } catch( java.io.IOException io) {}

    }// end of publishMyAddress()

    /**
    *
    *  Deletes the server object from group.
    *
    * <ol>
    *  <p><li> It obtains the record about this server. Of it does not
    *          know it, then that's OK(?).
    *  <p><li> It removes the record from list.
    *  <p><li> If this serer was a CENTRAL, then it needs to figure out
    *          which server will take over.
    *          If HBL has the minimum id, then it should take over
    *     <ol>
    *      <p><li> changes its role to CENTRAL
    *      <p><li> resets its central server related information
    *      <p><li> publishes its address and anounce that it is taking over.
    *     </ol>
    *          If HBL does not have the minimum id, then it should update
    *          its central server information according to its registry
    *          information. After all, it will get notified by the central
    *          server.
    * </ol>
    *
    */
    void delete( String HBname )
    {
       debug(HBname+" will be removed!"); 
       maps.remove( HBname );
       ServerInfo sv = (ServerInfo) simaps.get( HBname );
       if( sv!=null )
       {
          siList.removeFromList(sv);
          simaps.remove( HBname );

          // Anounce that I took over CENTRAL Role if
          // That is the case.
          if( HBname.equals(central_name) )
          {
              sv = siList.findMinimum();
              if( sv==null || sv.getId() == myNum )
              {
                  // Promote to CENTRAL Server Role
                  role=CENTRAL;

                  central = null;
                  broadcastReplace(siList.enumerateList().elements(),central_name,myId);
                  central_name = null;
                  debug("I took over removed!"); 
                  publishMyAddress( "hbl.ior",
                                    org.omg.CORBA.ORB.init().object_to_string(this));
 
              }
              else
                {  // I could not get the CENTRAL Role, so let's
                   // set the record 
                   central_name = sv.getName();
                   central = (HeartBeat.HeartBeatListener) maps.get(central_name);
                }

          }// end of if

       }// end of if

    }// end of delete()

    /**
    * 
    * <ol> 
    * </ol> 
    * 
    */
    void broadcastReplace(Enumeration eles, String prevName, String myName)
    {
       HeartBeat.HeartBeatListener rl;
       while( eles.hasMoreElements() )
       {
           rl = (HeartBeat.HeartBeatListener) eles.nextElement();
           try { rl.replace(prevName,myName,this); } catch(Exception e) {}

       }// end of while
       
    }// end of broadcastReplace()

    /**
    * 
    * <ol> 
    * </ol> 
    * 
    */
    void broadcastLeave(Enumeration eles, String myName)
    {
       HeartBeat.HeartBeatListener rl;
       while( eles.hasMoreElements() )
       {
           rl = (HeartBeat.HeartBeatListener) eles.nextElement();
           try { rl.leave(myName); } catch(Exception e) {}

       }// end of while
       
    }// end of broadcastLeave()

    /**
    * 
    * <ol> 
    *  <p><li> Stops timer 
    *  <p><li> If it has a central server to inform, then it does and sends
    *          a leave message to it.
    * </ol> 
    * 
    */
    public void 
    stop() 
    { 
        debug(" in HeartBeatServant::stop() ");
        tick.destroy();

        switch( role )
        {
           case SLAVE : 
                        Vector bList = new Vector();
                        Enumeration eles = maps.elements();

                        if( central!=null ) bList.addElement(central);
                        else 
                          while( eles.hasMoreElements() )
                                 bList.addElement(eles.nextElement());
                        broadcastLeave(bList.elements(),myId);
                        break;

           case CENTRAL : 
                        broadcastLeave(maps.elements(),myId);
                        break;

           case G_CENTRAL : 
                        //broadcastLeave(lower_maps.elements(),myId);
                        //broadcastLeave(maps.elements(),myGrId);
                        break;
                        
        }// end of switch

    }// end of stop()

    /**
    *  Broadcasts the registration of a new server.
    *  <ol>
    *   <p><li> Broadcast all others this registration
    *   <p><li> Broadcast all others to the new comer
    *  </ol>
    */
    void broadcast( String hName, int nm, HeartBeat.HeartBeatListener sa)
    {
       ServerInfo sv;
       HeartBeat.HeartBeatListener hb;
       Enumeration e = maps.keys();
       String srvKey;

       // Broadcast all others this registration
       while( e.hasMoreElements() )
       {
          srvKey = (String) e.nextElement();

          if( srvKey!=null && !srvKey.equals(hName) )
          {
             hb = (HeartBeat.HeartBeatListener) maps.get(srvKey);
             if( hb!=null )
               try
               {  hb.register(hName,nm, sa);
               } catch(  org.omg.CORBA.SystemException se)
                 {  // delete(srvKey);
                 } 

          }// end of if
          
       }// end of while

       e = maps.keys();

       // Broadcast all others to the new comer
       // try
       // {
       while( e.hasMoreElements() )
       {
          srvKey = (String) e.nextElement();

          if( srvKey!=null && !srvKey.equals(hName) )
          {
             hb = (HeartBeat.HeartBeatListener) maps.get(srvKey);
             sv = (ServerInfo) simaps.get(srvKey);
             if( hb!=null )
               try
               {  sa.register(srvKey,sv.getId(), hb);
               } catch(  org.omg.CORBA.SystemException se)
                 {  // delete(hName);
                 } 

          }// end of if
          
       }// end of while
       // } catch(  org.omg.CORNBA.SystemException se)
       //   {  // delete(hName); } 

    }// end of

    /**
    */
    public void
    debug(String mess)
    {
       if(HBL_DEBUG) System.out.println(mess);

    }// end of debug


}// end of HeartBeatServant()

