/**
*
* Thread-Safe list.
*
* @author H. Timucin Ozdemir
* @date   September 15, 1998 ( Last Updated September 17, 1995)
*
*/

import java.util.Vector;
import java.util.Enumeration;

public class ServerInfoList implements HBLDebug
{
    /** */
    ServerInfo  head;
    ServerInfo  tail;

    /**
    */
    public ServerInfoList() 
    {
        head = tail = null;

    }// end of CONSTUCTOR

    /**
    *
    *  Walks backwards from end of the list and inserts the object
    *  into the appropriate place. Note that list will be kept in 
    *  non-decreasing order accroding to their last check in time.
    *
    */
    public synchronized void addFromTail( ServerInfo sv )
    {
       ServerInfo current = tail;

       // Go backwards as much as possible
       while( current!=null && current.getTime()>sv.getTime() )
              current = current.getPrevious();

       // If the current is not null, then it did not hit the head.
       if( current!=null ) 
       { 
           current.insert(sv);
           if( current.getNext()==null ) tail = current;

       } else head = current;
 
       if( head==null ) head = sv;

       // If we added something to the head, we need to update the tail
       // incase it is the first element.
       if( head.getNext()==null ) tail = head;
       else head.getNext().setPrevious(head);

    }// end of addFromTail()

    /**
    *  Case analysis. It can be written more compact form.
    */
    public synchronized void removeFromList( ServerInfo sv )
    {

       ServerInfo prev = sv.getPrevious();
       ServerInfo next = sv.getNext();
 
       if( prev!=null && next!=null)
       {
           prev.setNext(next);
           next.setPrevious(prev);

       }// end of if

       if( prev!=null && next==null)
       {
           prev.setNext(next);
           tail = prev;

       }// end of if

       if( prev==null && next!=null)
       {
           head = next;
           next.setPrevious(prev);

       }// end of if

       if( prev==null && next==null)
       {
           head = next;
           tail = next;

       }// end of if

    }// end of addFromTail()

    /**
    * retuns all the records in the list.
    */
    public synchronized Vector enumerateList( )
    {
       Vector a = new Vector();
       ServerInfo current = head;
       while( current!=null )
       {
          a.addElement(current);
          current = current.getNext();

       }// end of while

       return a;

    }// end of enumerateList()

    /**
    * returns the minimum server info according to id.
    */
    public synchronized ServerInfo findMinimum( )
    {
       ServerInfo current = head;
       ServerInfo ret     = head;
       while( current!=null )
       {
          if(ret.getId()>current.getId()) ret=current;
          current = current.getNext();

       }// end of while

       return ret;

    }// end of findMinimum()

    /**
    */
    public void
    debug(String mess)
    {
       if(LST_DEBUG) System.out.println(mess);

    }// end of debug


}// end of ServerInfoList()

