
import java.io.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import owrti.*;
import org.omg.CORBA.*;
import npacrti.*;
import RTIcap.*;


/**
*
* This servlet forks a HelloWorld federate object to produce an
* interactions for the OWRTI demos with DirectX donuts.
*
* @author H. Timucin Ozdemir
* @date   August 9, 1998 ( Last Update August 11, 1998 )
*
*/
public
class RHello extends GenericServlet 
{

    String    IorFile  = null;
    String    nameOftheCountry   = new String("us");
    String    initialPopulation  = new String("100");
    String    numberOfIterations = new String("10");
    String    sleepTime          = new String("20");

    public void 
    init(ServletConfig initialParameters) throws ServletException
    {
       IorFile = initialParameters.getInitParameter("IOR_FILE");

       System.out.println(" IorFile : "+IorFile);
  
    }// end of init()

    /**
    * Reads the file 
    */
    public void
    service(ServletRequest req, ServletResponse res) 
            throws ServletException,IOException
    {
       String command = req.getParameter("cmd");

       System.out.println(" command : "+command);

       if(command!=null && command.equals("runHello") ) 
            service_runHello(req,res);
       else service_unknown(res,command);
       
    }// end of init() 

    /**
    * Unknown command received on URL. Return a message with "unknown command"
    * in it.
    * @param res: a ServletResponse, an object that contains the response of the
    *             servlet.
    * @param command: String containing the command with which the servlet was
    *                 invoked
    */
    void service_unknown(ServletResponse res, String command)
    {
       res.setContentType("text/html");
       try
       {
          OutputStream os = res.getOutputStream();
          PrintStream out = new PrintStream(os);
          out.println("Unknown command: " + command);
          out.flush();

       }catch(java.io.IOException e){System.out.println(e);}

    }// end of service_unknown()

   /**
   */
   public  void 
   service_runHello(ServletRequest req, ServletResponse res)
   { 
     Date   today = new Date(System.currentTimeMillis());
     String args[] = new String[4];
     args[0] = new String(nameOftheCountry+today);
     args[0] = args[0].replace(' ','_');
     args[1] = new String(initialPopulation);
     args[2] = new String(numberOfIterations);
     args[3] = new String(sleepTime);

     String var = req.getParameter("cname");
     if(var!=null) args[0] = new String(var);

     var = req.getParameter("ipop");
     if(var!=null) args[1] = new String(var);

     var = req.getParameter("inum");
     if(var!=null) args[2] = new String(var);

     var = req.getParameter("stime");
     if(var!=null) args[3] = new String(var);
        
     try
     {
      
        // Create and initialize the ORB
        ORB orb = org.omg.CORBA.ORB.init(args, null);
      
        /*  ----- This Part is required if we are using Naming Service ---
        // Get the root naming context
        org.omg.CORBA.Object objRef = orb.resolve_initial_references("NameService");
        NamingContext ncRef = NamingContextHelper.narrow(objRef);
      
        // Resolve the object reference in naming
        NameComponent nc = new NameComponent("RTIKernel", " ");
        NameComponent path[] = {nc};
        RTIKernelInterface rtiKernel = RTIKernelInterfaceHelper.narrow(ncRef.resolve(path));
        */

        try 
        {
           System.out.println(" IorFile : "+IorFile);
           BufferedReader input = new BufferedReader(new FileReader(IorFile));
           //String strIOR = input.readLine();  // get IOR address of RTIKernelInterface
           String strIOR = owrti.util.KernelLocator.getRTIKernelIOR(IorFile);
           if( strIOR != null )
           {
      
              RTIKernelInterface rtiKernel = RTIKernelInterfaceHelper.narrow(orb.string_to_object(strIOR));

              // Get the Version of the Kernel
              String version = rtiKernel.getVersion();
              System.out.println(" Version is : "+version);      

              if(args.length>2)
              {
                 RTIambassador rtiAmb = rtiKernel.createRTIAmbassador();
                 owrtihello.HelloWorld hello = 
                         new owrtihello.HelloWorld(rtiAmb,args[0],
                                        Double.valueOf(args[1]).doubleValue(),
                                        Integer.parseInt(args[2]),
                                        Double.valueOf(args[3]).doubleValue(), orb);

                 /*
                  Return ack.
                 */
                 res.setContentType("text/html");
                 try
                 {
                    OutputStream os = res.getOutputStream();
                    PrintStream out = new PrintStream(os);
                    out.println(" Interaction Generator has been activated.<br>");
                    out.println(" It will produce an event in each <b>"+args[3]+"</b> msec.<br>");
                    out.println("...ENJOY....");
                    out.flush();

                 } catch(java.io.IOException e){System.out.println(e);}
                
                 hello.start();

              }// end of if

           } else System.out.println(" IOR file is empty| \n "+
                        "I can not find the address of RTIKernelInterface ");

        } catch(java.io.FileNotFoundException fnf) {}
          catch(java.io.IOException ioe) {}
          catch(NumberFormatException ioe) {}
          
     } catch(Exception e) 
       {
         System.out.println("ERROR : " + e);
         e.printStackTrace(System.out);
       }  

     
    }// end of service_runHello()

    /**
    */
    public String 
    getServletInfo() 
    {
	   return "RHello servlet : produces interactions for OWRTI demo" ;
    }// end of getServletInfo()

}// end of RHello

