
/*
 * @(#)SnoopServlet.java        1.20 97/11/17
 * 
 * Copyright (c) 1995-1997 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.0
 */

import java.io.*;
import java.util.*;
//import java.security.cert.*;

import javax.servlet.*;
import javax.servlet.http.*;


/**
 * Snoop servlet. This servlet simply echos back the request line and
 * headers that were sent by the client, plus any HTTPS information
 * which is accessible.
 *
 * @version     1.20
 * @author      Various
 */
public
class SnoopServlet extends HttpServlet {

    public void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
    {
        ServletOutputStream     out;

        res.setContentType("text/html");
        out = res.getOutputStream();

        out.println("<html>");
        out.println("<head><title>Snoop Servlet</title></head>");
        out.println("<body>");

        out.println("<h1>Requested URL:</h1>");
        out.println("<pre>");
        out.println (HttpUtils.getRequestURL (req).toString ());
        out.println("</pre>");

        Enumeration enum = getServletConfig().getInitParameterNames();
        if (enum != null) {
            boolean first = true;
            while (enum.hasMoreElements()) {
                if (first) {
                    out.println("<h1>Init Parameters</h1>");
                    out.println("<pre>");
                    first = false;
                }
                String param = (String) enum.nextElement();
                out.println(" "+param+": "+getInitParameter(param));
            }
            out.println("</pre>");
        }

        out.println("<h1>Request information:</h1>");
        out.println("<pre>");

        out.println( "Request method "+req.getMethod());
        out.println("Request URI "+req.getRequestURI());
        out.println("Request protocol "+req.getProtocol());
        out.println("Servlet path "+req.getServletPath());
        out.println("Path info "+req.getPathInfo());
        out.println("Path translated "+req.getPathTranslated());
        out.println("Query string "+req.getQueryString());
        out.println("Content length "+req.getContentLength());
        out.println("Content type "+req.getContentType());
        out.println("Server name "+req.getServerName());
        out.println("Server port "+req.getServerPort());
        out.println("Remote user "+req.getRemoteUser());
        out.println("Remote address "+req.getRemoteAddr());
        out.println("Remote host "+req.getRemoteHost());
        out.println("Authorization scheme "+req.getAuthType());

        out.println("</pre>");
        
        Enumeration e = req.getHeaderNames();
        if (e.hasMoreElements()) {
            out.println("<h1>Request headers:</h1>");
            out.println("<pre>");
            while (e.hasMoreElements()) {
                String name = (String)e.nextElement();
                out.println(" " + name + ": " + req.getHeader(name));
            }
            out.println("</pre>");
        }

        e = req.getParameterNames();
        if (e.hasMoreElements()) {
            out.println("<h1>Servlet parameters (Single Value style):</h1>");
            out.println("<pre>");
            while (e.hasMoreElements()) {
                String name = (String)e.nextElement();
                out.println(" " + name + " = " + req.getParameter(name));
            }
            out.println("</pre>");
        }

        e = req.getParameterNames();
        if (e.hasMoreElements()) {
            out.println("<h1>Servlet parameters (Multiple Value style):</h1>");
            out.println("<pre>");
            while (e.hasMoreElements()) {
                String name = (String)e.nextElement();
                String vals[] = (String []) req.getParameterValues(name);
                if (vals != null) {
                    out.print("<b> " + name + " = </b>"); 
                    out.println(vals[0]);
                    for (int i = 1; i<vals.length; i++)
                        out.println("           " + vals[i]);
                }
                out.println("<p>");
            }
            out.println("</pre>");
        }

        String  cipherSuite = (String)
            req.getAttribute ("javax.net.ssl.cipher_suite");

        /*
        if (cipherSuite != null) {
            X509Certificate     certChain [] = (X509Certificate [])
                req.getAttribute ("javax.net.ssl.peer_certificates");
                
            out.println ("<h1>HTTPS Information:</h1>");
            out.println("<pre>");

            out.println ("Cipher Suite:  " + cipherSuite);

            if (certChain != null) {
                for (int i = 0; i < certChain.length; i++) {
                    out.println ("client cert chain [" + i + "] = "
                        + certChain [i].toString ());
                }
            }


            out.println("</pre>");
        }
        */
        
        out.println("</body></html>");
    }

    private void print (PrintWriter out, String name, String value)
    {
        out.print(" " + name + ": ");
        out.println(value == null ? "&lt;none&gt;" : value);
    }

    private void print (PrintWriter out, String name, int value)
    {
        out.print(" " + name + ": ");
        if (value == -1) {
            out.println("&lt;none&gt;");
        } else {
            out.println(value);
        }
    }
}
