/**
* @date   September 14, 1998 ( Last Updated September 15, 1998)
*
*/

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.*;
import java.net.*;

import org.omg.CORBA.*;

import HeartBeat.*;
import HeartBeat.HeartBeatListenerPackage.*;

/**
* Functionality:
*  <p>
*  <ol>
*   <p><li> Start/Stop HeartBeatListener Object
*   <p><li> Register/Remove ReportListener Objects
*  </ol>
*  <p>
*
* @author H. Timucin Ozdemir
* @date   September 14, 1998 ( Last Updated September 15, 1998)
*
*/
public class 
HeartBeatDaemon extends HeartBeat._ReportAdminImplBase implements HBLDebug
{
    
    /** */
    public static final int DEAD    = 0;
    public static final int BEATING = 1;

    /** */
    public static final String  DEFAULT_FILE_NAME = new String("hbl.srv");

    /** */
    HeartBeatServant  hbl;

    /** */
    String            cFileName = null;
    String            logFile   = new String("log.hbl");

    /** */
    int               state = DEAD;

    /** */
    ReportBroadcaster Rbroadcaster ;

    /** */
    Hashtable         listeners ;
    int               registered=0; 

    /**
    */
    public HeartBeatDaemon() 
    {
       Rbroadcaster = new ReportBroadcaster(this);
       listeners    = new Hashtable();
       debug(" HeartBeatDaemon() defined ");
       try
       { PrintWriter out = new PrintWriter(new FileWriter(logFile,false));
         out.println(" start");
         out.close();
       } catch(java.io.IOException e) {}

       start();

    }// end of CONSTUCTOR

    //// HeartBeat.HeartBeatDaemon Interface

    /**
    * Starts the HBL.
    * 
    * <ol>
    *  <p><li> If it is not running already,
    *  <ol>
    *   <p><li> looks at /servlet/naming objects on given servers to find
    *            another running HBL object.
    *   <p><li> If it could not find anybody, then tries to read
    *           <b>HBL.ior</b> file to obtain this info.
    *   <p><li> defines a HBL object, and sets its master to this object.
    *  </ol>
    * </ol>
    * 
    */
    public void 
    start()
    {
       debug(" In start() ");
       if( state == DEAD )
       {
          HeartBeat.HeartBeatListener cHBL=null;

          try
          {
            String hblIOR = getHBLIOR(cFileName);
            if( hblIOR==null ) 
              try { hblIOR = getHBLIOR_FF("hbl.ior"); }
              catch( java.io.IOException ioe) {}
 
            if( hblIOR!=null )
              cHBL = HeartBeat.HeartBeatListenerHelper.narrow(org.omg.CORBA.ORB.init().string_to_object(hblIOR));
              
          } catch(java.io.FileNotFoundException fnf) 
            { System.out.println(" ERROR: "+cFileName+" not found!"); 
              cHBL=null;
            }
            catch(java.lang.Exception fnf) { cHBL=null; }
              
          hbl = new HeartBeatServant(cHBL);
          hbl.setMaster(this);

          state = BEATING;

       }// end of if

    }// end of start()

    /**
    * Stops the HBL.
    * <ol>
    *  <p><li> If the HBL is running, sends a stop message to it.
    * </ol>
    */
    public void 
    stop()
    {
       debug(" in HeartBeatDaemon::stop() ");
       if( state == BEATING )
       {
          hbl.stop();
          state = DEAD;
 
       }// end of if

       if( state == BEATING )
           debug(" in HeartBeatDaemon::stop() ALIVE!!! ");
       else if( state == DEAD )
           debug(" in HeartBeatDaemon::stop() EXTERMINATED!!! ");
       else 
           debug(" in HeartBeatDaemon::stop() WEIRD!!! ");

    }// end of stop()

    //// HeartBeat.ReportAdmin Interface

    /**
    * Registers ReportListener object.
    */
    public void 
    register(HeartBeat.ReportListener rl)
    {
       int ret = 0;
       Integer ll = (Integer) listeners.get(rl);
       if(ll==null) 
       {
          synchronized(listeners) 
          { listeners.put(rl,new Integer(registered++)); }
          log(registered+" registered ");
       }
       else ret=1;

       //return ret;

    }// end of register()

    /**
    * Removes ReportListener object.
    */
    public void 
    remove(HeartBeat.ReportListener rl)
    {
       int ret = 0;
       Integer ll = (Integer) listeners.get(rl);
       if(ll!=null)
       {
         synchronized(listeners) { listeners.remove(rl); }
         log(rl+" is removed "+listeners.size());
       }
       else { 
              ret = 1;
              debug(" ********* HeartBeatDaemon::remover() Element Does NOT EXISTS!");
            }

       //return ret;

    }// end of remove()

    /** UTILITIES **/

    /**
    */
    final String getHBLIOR(String cnfFile) throws java.io.FileNotFoundException
    {
      String fileName = new String(HeartBeatDaemon.DEFAULT_FILE_NAME);
      if(cnfFile!=null) fileName = cnfFile;
      return IORProvider.getIOR(fileName,"IDL:HeartBeat/HeartBeatListener:1.0");

    }// end of getHBLIOR()

    /**
    */
    final String getHBLIOR_FF(String cnfFile) 
                 throws java.io.FileNotFoundException, java.io.IOException
    {
      return IORProvider.getIORFromFile(cnfFile);

    }// end of getHBLIOR_FF()

    /**
    * It reports ServerInfo objects in Enumeration
    * HBL will invoke this method after each tick. It produces neccessary
    * data structures  and publishes this report to each ReportListener.
    * Publishes report if it has some listeners of this report.
    */
    public void report(Enumeration e)
    {
       if( e.hasMoreElements() ) publishReport(e);

    }// end of report()

    /**
    */
    public void publishReport(Enumeration e)
    {
       debug(" ********* in HeartBeatDaemon::report() ");
       Vector a = new Vector();
       ServerInfo sv;

       while( e.hasMoreElements() )
       {
          sv = (ServerInfo) e.nextElement();
          if( sv!=null )
          {
             a.addElement(new HeartBeat.ReportListenerPackage.ServerItem("",0,sv.getName(),(int)sv.getTime(),sv.getStrikeOut()));
          }// end of if

       }// end of while

       HeartBeat.ReportListenerPackage.ServerItem r[] = new HeartBeat.ReportListenerPackage.ServerItem[a.size()];
       a.copyInto(r);

       debug(" in HeartBeatDaemon::report() ");
       if( listeners.size()>0 ) 
       { 
            log(" in HeartBeatDaemon::report() for "+listeners.size());
            Rbroadcaster.broadcast(listeners,r);

       } else debug(" There is no listener HeartBeatDaemon::report() ");

    }// end of report()

    /**
    */
    public void
    debug(String mess)
    {
       if(HBD_DEBUG) System.out.println(mess);

    }// end of debug

    /**
    */
    public final void
    log(String mess)
    {
       try
       {  PrintWriter out = new PrintWriter(new FileWriter(logFile,true));
          out.println(mess);
          out.close();
       } catch(java.io.IOException e) {}

    }// end of debug

}// end of HeartBeatDaemon

