/**
* Functionality:
*  <p>
*  <ol>
*   <p><li> Gets IOR from /servlet/naming
*   <p><li> Gets IOR from file
*  </ol>
*  <p>
*
* @author H. Timucin Ozdemir
* @date   September 14, 1998 ( Last Updated September 15, 1995)
*
*/

import java.util.Enumeration;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.*;
import java.net.*;

import org.omg.CORBA.*;

public class 
IORProvider
{
    /** */
    public static final String  DEFAULT_FILE_NAME = new String("naming.srv");

    /** */
    String            cFileName = null;

    /**
    * Default configuration file will be <b>naming.srv</b>.
    */
    public IORProvider() { this(DEFAULT_FILE_NAME); }

    /**
    * Default configuration file will be <b>_cFileName</b>.
    */
    public IORProvider(String _cFileName) 
    { 
      if(_cFileName!=null) cFileName = _cFileName;
      else cFileName = DEFAULT_FILE_NAME;

    }// end of CONSTRUCTOR

    public final String 
    getConfigFileName() { return cFileName; } 

    /**
    * Returns the IOR address of the interface. 
    */
    public static final String 
    getIOR(String IDL) throws java.io.FileNotFoundException
    {
       return IORProvider.getIOR(DEFAULT_FILE_NAME,IDL);

    }// end of getIOR()

    /**
    * Returns the IOR address of the <b>interface</b> by using entries in
    * <b>cnfFile</b> file. 
    */
    public static final String 
    getIOR(String cnfFile, String IDL) 
           throws java.io.FileNotFoundException
    {
      if( !IDL.startsWith("IDL:") ) return null;

      String  ret = null;
      String  hostName;
      String  portNumber;
      int     port;
      boolean notFound = true;
      String fileName = new String(IORProvider.DEFAULT_FILE_NAME);
      if(cnfFile!=null) fileName = cnfFile;

      try
      {
        BufferedReader input = new BufferedReader(new FileReader(fileName));
        String line = input.readLine();

        while( ret==null && line!=null )
        {
           if( !line.startsWith("#") || !line.startsWith("%") ||
               !line.startsWith("/") || !line.startsWith("$") ) 
           {
               StringTokenizer sToken = new StringTokenizer(line);
               if(sToken.countTokens()>1)
               {
                 hostName   = (String) sToken.nextElement(); 
                 portNumber = (String) sToken.nextElement(); 

                 try
                 {
                   port = Integer.parseInt(portNumber);
                   
                   ret = askIOR(hostName,port,IDL);

                 } catch(java.lang.NumberFormatException nfe) {}

               }// end of if

           }// end of if

           line = input.readLine();

        }// end of while

      } catch(java.io.IOException fnf) { }
      
      return ret; 

   }// end of getHBLIOR()

   /**
   */
   static final 
   String askIOR(String host, int port, String interfaceName)
   {

      String retIOR=null;

      try
      {
         String query = new String("http://"+host+":"+port+"/servlet/naming?");
         query += "cmd=getIOR&";
         query += ("IDL="+interfaceName);
         System.out.println(" query :"+query);

         URL namingURL = new URL(query);

         DataInputStream dis = new DataInputStream( namingURL.openStream());

         retIOR = dis.readLine();

         if(retIOR!=null && !retIOR.startsWith("IOR:")) retIOR = null;

         dis.close();

      } catch(java.net.MalformedURLException mfue) {}
        catch(java.io.IOException ioe) {}

      return retIOR;

    }// end of askIOR()

    /**
    * Returns the IOR value from file <b>cFile</b>. 
    * This file should contain externalized IOR definition of the object.
    */
    public static final String 
    getIORFromFile(String cFile) 
           throws java.io.FileNotFoundException, java.io.IOException
    {
       String ret = null;
       BufferedReader input = new BufferedReader(new FileReader(cFile));
       String strIOR = input.readLine();  

       if( strIOR!=null && strIOR.startsWith("IOR:") ) ret = strIOR;

       return ret;

    }// end of getIORFromFile()

}// end of IORProvider

