/**
* A simple test for remote invocations.
*
* <p>
* It reads IOR of the test object from file IOR in the current directory 
* unless user provides another file name for this.
*
* <p>
* Usage : java test1Client [<IOR_file_name>]
*
* @author H. Timucin Ozdemir
* @date  July 6, 1998
*
*/

import org.omg.CosNaming.*;  
import org.omg.CORBA.*;     

import java.net.*;
import java.io.*;
import java.util.*;

import jworbTest3Package.*;

public class 
test3Client
{

    /**
    */
    public static void 
    main(String args[]) 
    {
        try 
        {
           int i;
           String fileName = new String("IOR");
           if(args.length>0) fileName = args[0];
           BufferedReader input = new BufferedReader(new FileReader(fileName));
           String strIOR = input.readLine();  // get IOR address of RTIKernelInt erface
           if( strIOR!=null && strIOR.startsWith("IOR:") )
           {
 
              Properties prop = new Properties();
              prop.put("org.omg.CORBA.ORBClass","worb.protocols.iiop.ORBImpl");
              prop.put("org.omg.CORBA.ORBSingletonClass","worb.protocols.iiop.ORBImpl");
              System.out.println(" I read IOR --> "+strIOR);

              // Create and initialize the ORB
              ORB orb = ORB.init(args, prop);

              System.out.println(" IOR --> "+strIOR);
              org.omg.CORBA.Object objRef = orb.string_to_object(strIOR);
      
              jworbTest3 tst3Ref = jworbTest3Helper.narrow(objRef);

              tst3Ref.ping();
              System.out.println(" (Ping) I pinged ");

              char val1 = 'A';
              System.out.println(" (char) I said "+val1);
              System.out.println(" (char) I received "+tst3Ref.echoChar(val1));


              byte val2 = (byte) 0x000000f0;
              System.out.println(" (byte) I said "+(int)val2);
              System.out.println(" (byte) I received "+(int)tst3Ref.echoOctet(val2));

              boolean val3 = true;
              System.out.println(" (boolean) I said "+val3);
              System.out.println(" (boolean) I received "+tst3Ref.echoBoolean(val3));

              int val4 = 11;
              System.out.println(" (Long) I said "+val4);
              System.out.println(" (Long) I received "+tst3Ref.echoLong(val4));

              short val5 = 11;
              System.out.println(" (short) I said "+val5);
              System.out.println(" (short) I received "+tst3Ref.echoShort(val5));

              float val6 = 3.142f;
              System.out.println(" (Float) I said "+val6);
              System.out.println(" (Float) I received "+tst3Ref.echoFloat(val6));

              double val7 = 3545463.142;
              System.out.println(" (Double) I said "+val7);
              System.out.println(" (Double) I received "+tst3Ref.echoDouble(val7));

              String val8 = "Hello";
              System.out.println(" (string) I said "+val8);
              System.out.println(" (string) I received "+tst3Ref.echoString(val8));

              try
              {
                System.out.println(" (Exception) I received ");
                tst3Ref.getException();

              } catch(jworbTest3Package.ObjectNotKnown onk)
                {
                   System.out.println(" jworbTest3Package.ObjectNotKnow Exception is received!");
                }

              int  val9[] = new int[2];
              val9[0] = 10; val9[1]=1111;
              System.out.print(" (echoLongArray) I said ");
              for(i=0;i<2;i++) System.out.print(val9[i]+" , ");
              System.out.println();

              int  ret[] = tst3Ref.echoLongArray(val9);
              System.out.println(" (echoLongArray) I received ");
              for(i=0;i<2;i++) System.out.print(ret[i]+" , ");
              System.out.println();

              jworbTest3Package.long_struct val10 = new jworbTest3Package.long_struct(12322);
              System.out.println(" (echoLong_struct) I said "+val10.number);
              System.out.println(" (echoLong_struct) I received "+tst3Ref.echoLong_struct(val10).number);

              jworbTest3Package.long_struct  val11[] = new jworbTest3Package.long_struct[2];
              val11[0] = new jworbTest3Package.long_struct(222); 
              val11[1] = new jworbTest3Package.long_struct(323); 
              System.out.print(" (echoLong_structArr) I said ");
              for(i=0;i<2;i++) System.out.print(val11[i].number+" , ");
              System.out.println();

              jworbTest3Package.long_struct  ret2[] = tst3Ref.echoLong_structArray(val11);
              System.out.println(" (echoLong_structArr) I received ");
              for(i=0;i<2;i++) System.out.print(ret2[i].number+" , ");
              System.out.println();

              int cId = 1;
              short[] myAttributes = new short[2];
              myAttributes[0] = (short) 33;
              myAttributes[1] = (short) 12;
              System.out.println(" (subscribeObjectClassAttributes) Before ");
              System.out.println(" (subscribeObjectClassAttributes) cId "+cId);
              System.out.println(" (subscribeObjectClassAttributes) myAttributes[0] "+myAttributes[0]);
              System.out.println(" (subscribeObjectClassAttributes) myAttributes[1] "+myAttributes[1]);
              
              tst3Ref.subscribeObjectClassAttributes( (short)cId, myAttributes , jworbTest3Package._Boolean.RTI_TRUE);
              System.out.println(" (subscribeObjectClassAttributes) After ");

           }
           else 
            {
               System.err.println(fileName+" does not have poper IOR address ");
               System.exit(1);
            }

        } catch (Exception e) 
          {
             System.out.println("### Exception !!! ### \n");
             e.printStackTrace();
          }

    }// end of main()

}// end of Class test1Client 
