/**
*
* This is a basic test program for JWORB. 
*
* org.omg.* classes come with Java 1.2 beta2.
*
* @author H. Timucin Ozdemir
* @date   July 6, 1998
*/

// the naming service.
import org.omg.CosNaming.*;

// The package containing special exceptions thrown by the name service.
import org.omg.CosNaming.NamingContextPackage.*;

// All CORBA applications need these classes.
import org.omg.CORBA.*;

import java.util.*;
import java.io.*;

public class testDynArray {


  /** */
  public static void printStringDynArray(DynArray darr, String label) {
    try
    {  darr.seek(0);
       boolean stay=true; int i=0;
       while(stay) {
         DynAny a = darr.current_component();
         if(a!=null) System.out.println((label!=null?label:"a")+"["+i+"]: "+a.get_string());
         stay=darr.next();  i++;
       }// end of for
    } catch(org.omg.CORBA.DynAnyPackage.TypeMismatch in) 
      { System.out.println("It does not contain *string* "); }
  }// end of for

  /** */
  public static void main(String args[]) {
    try{
    
      int i;
      Properties prop = new Properties();
      
      // to force our IIOP implementation.
      prop.put("org.omg.CORBA.ORBClass","worb.protocols.iiop.ORBImpl");
      prop.put("org.omg.CORBA.ORBSingletonClass","worb.protocols.iiop.ORBImpl");

      // Create and initialize the ORB
      ORB orb = ORB.init(args, prop);
      
      // Create a typecode
      String[] enumValues = new String[3];
      enumValues[0]=new String("RED");  
      enumValues[1]=new String("BLUE");  
      enumValues[2]=new String("YELLOW");
      TypeCode eleTC= orb.get_primitive_tc(TCKind.tk_string);
      TypeCode colTC= orb.create_sequence_tc(3,eleTC);
      DynArray darr = orb.create_dyn_array(colTC); 

      System.out.println("\n Test: darr.insert_string ");

      for(i=0;i<3;i++) darr.insert_string(enumValues[i]);

      System.out.println("\n Test: darr.get_string() ");
      try { System.out.println("ERROR: "+darr.get_string()+" should not be produced!"); }
      catch(org.omg.CORBA.DynAnyPackage.TypeMismatch in) 
      { System.out.println("CORRECT: It is after the end of the array "); }

      System.out.println("\n Test: darr.rewind() and darr.get_ulong() ");
      darr.rewind();
      try { System.out.println("ERROR: "+darr.get_ulong()+" should not be produced!"); }
      catch(org.omg.CORBA.DynAnyPackage.TypeMismatch in) 
      { System.out.println("CORRECT: It only contains *string* values"); }

      System.out.println("\n Test: darr.rewind() and access each element with get_string() ");
      darr.rewind();
      boolean stay=true; i=0;
      while(stay) {
        DynAny a = darr.current_component();
        if(a!=null) System.out.println("darr["+i+"]: "+a.get_string());
        stay=darr.next();  i++;
      }// end of for

      System.out.println("\n Test: darr.seek(0) and access each element with ");
      System.out.println("       current_component() and next()");
      testDynArray.printStringDynArray(darr,"darr");

      System.out.println("\n Test: clone1=darr.copy() - print clone1 ");
      DynArray clone1 = (DynArray)darr.copy();
      testDynArray.printStringDynArray(clone1,"clone1");

      System.out.println("\n Test: darr.set_elements() - add 3 elements ");
      Any[] vals = new Any[3];
      for(i=0;i<3;i++) { 
        vals[i]=orb.create_any();
        vals[i].type(orb.get_primitive_tc(TCKind.tk_string));
        vals[i].insert_string("A"+i);
      }// end of for

      darr.set_elements(vals);
      testDynArray.printStringDynArray(darr,"darr");

      System.out.println("\n Test: vals=clone1.get_elements() - print vals ");
      vals = clone1.get_elements();
      if(vals!=null)
        for(i=0;i<vals.length;i++)  
          System.out.println("val["+i+"]: "+vals[i].extract_string());

      System.out.println("\n Test: seqA=clone1.to_any(), darr2.from_any(seqA) - print darr2 ");
      Any seqA = clone1.to_any();
      DynArray darr2 = orb.create_dyn_array(colTC); 
      darr2.from_any(seqA);
      testDynArray.printStringDynArray(darr2,"darr2");
     
      System.out.println("\n Test: darr.assign(darr2) - print darr ");
      darr.assign(darr2);
      testDynArray.printStringDynArray(darr,"darr");

      System.out.println("**** DynArray Interface TEST IS OVER **** ");

    } catch(Exception e) 
      { System.err.println("ERROR: " + e); e.printStackTrace(System.out); }  

  }// end of main

}// end of testDynArray


