/**
*
* This is a basic test program for JWORB. 
*
* org.omg.* classes come with Java 1.2 beta2.
*
* @author H. Timucin Ozdemir
* @date   July 6, 1998
*/

// the naming service.
import org.omg.CosNaming.*;

// The package containing special exceptions thrown by the name service.
import org.omg.CosNaming.NamingContextPackage.*;

// All CORBA applications need these classes.
import org.omg.CORBA.*;

import java.util.*;
import java.io.*;

public class testDynEnum {

  /** */
  public static void printMyIOR(String fileName, String IOR) {
     try
     {  PrintWriter pw = new PrintWriter( new FileWriter(new File(fileName)));
        pw.println(IOR);
        pw.close();
     } catch( java.io.IOException io) {}
  }// end of printMyIOR()

  /** */
  public static void main(String args[]) {
    try{
    
      int i;
      Properties prop = new Properties();
      
      // to force our IIOP implementation.
      prop.put("org.omg.CORBA.ORBClass","worb.protocols.iiop.ORBImpl");
      prop.put("org.omg.CORBA.ORBSingletonClass","worb.protocols.iiop.ORBImpl");

      // Create and initialize the ORB
      ORB orb = ORB.init(args, prop);
      
      // Create a typecode
      String[] enumValues = new String[3];
      enumValues[0]=new String("RED");  
      enumValues[1]=new String("BLUE");  
      enumValues[2]=new String("YELLOW");  
      TypeCode colTC= orb.create_enum_tc("IDL:Colors:0.1","myColors",enumValues);
      DynEnum denum = orb.create_dyn_enum(colTC); 

      System.out.println("\n Test: denum.value_as_string() and value_as_ulong()");
      for(i=0;i<enumValues.length;i++) {
        denum.value_as_string(enumValues[i]);
        System.out.println(" set to "+enumValues[i]+" equals to "+
                             denum.value_as_ulong());
      }

      System.out.println("\n Test: denum.value_as_ulong() and value_as_string()");
      for(i=0;i<enumValues.length;i++) {
        denum.value_as_ulong(i);
        System.out.println(" set to "+i+" equals to "+ denum.value_as_string());
      }

      System.out.println("\n Test:  denum2 = (DynEnum)denum.copy()  ");
      DynEnum denum2 = (DynEnum)denum.copy();
      System.out.println(" denum is set to "+denum.value_as_ulong()+
                         " denum2 is set to "+ denum2.value_as_string());

      System.out.println("\n Test: denum2.to_any() ");
      Any val = denum2.to_any();
      if(val.type().kind()==TCKind.tk_enum) 
           System.out.println("Any contains: " + val.extract_ulong()); 
      else System.out.println("Any does not contain tk_enum! " );

      System.out.println("\n Test: from_any() insert 0");
      Any val2 = orb.create_any();
      val2.type(denum2.type());
      val2.insert_ulong(0);
      denum2.from_any(val2); 
      System.out.println(" denum2 is set to "+ denum2.value_as_string());
         

      System.out.println("\n Test: assign() ");
      System.out.println(" denum is set to "+denum.value_as_string()+
                         " denum2 is set to "+ denum2.value_as_string());
      denum.assign(denum2);
      System.out.println(" denum is set to "+denum.value_as_string()+
                         " denum2 is set to "+ denum2.value_as_string());

      try { System.out.println(denum.get_ulong()+" should not be printed! "); }
      catch(org.omg.CORBA.DynAnyPackage.TypeMismatch iv )
      { System.out.println("CORRECT: you can not invoke get_ulong()! " ); ; }  

      try { System.out.println(denum.get_boolean()+" should not be printed! "); }
      catch(org.omg.CORBA.DynAnyPackage.TypeMismatch iv )
      { System.out.println("CORRECT: you can not invoke get_boolean()! " ); ; }  

      System.out.println("**** DynEnum Interface TEST IS OVER **** ");

    } catch(Exception e) 
      { System.err.println("ERROR: " + e); e.printStackTrace(System.out); }  

  }// end of main

}// end of test1


