/**
*
* This is a basic test program for JWORB. 
*
* org.omg.* classes come with Java 1.2 beta2.
*
* @author H. Timucin Ozdemir
* @date   July 6, 1998
*/

// the naming service.
import org.omg.CosNaming.*;

// The package containing special exceptions thrown by the name service.
import org.omg.CosNaming.NamingContextPackage.*;

// All CORBA applications need these classes.
import org.omg.CORBA.*;

import java.util.*;
import java.io.*;

public class testDynSeq {

  /** */
  public static void printMyIOR(String fileName, String IOR) {
     try
     {  PrintWriter pw = new PrintWriter( new FileWriter(new File(fileName)));
        pw.println(IOR);
        pw.close();
     } catch( java.io.IOException io) {}
  }// end of printMyIOR()

  /** */
  public static void print(DynSequence seq) {
    try {for(int i=0;i<seq.length();i++) 
       System.out.println("seq["+i+"]: "+seq.get_ulong());
    } catch(org.omg.CORBA.DynAnyPackage.TypeMismatch tm) {}
  }// end of print()

  /** */
  public static void main(String args[]) {
    try{
    
      int i;
      Properties prop = new Properties();
      
      // to force our IIOP implementation.
      prop.put("org.omg.CORBA.ORBClass","worb.protocols.iiop.ORBImpl");
      prop.put("org.omg.CORBA.ORBSingletonClass","worb.protocols.iiop.ORBImpl");

      // Create and initialize the ORB
      ORB orb = ORB.init(args, prop);
      
      // Create a typecode
      String[] enumValues = new String[3];
      enumValues[0]=new String("RED");  
      enumValues[1]=new String("BLUE");  
      enumValues[2]=new String("YELLOW");
      TypeCode eleTC= orb.get_primitive_tc(TCKind.tk_ulong);
      TypeCode colTC= orb.create_sequence_tc(3,eleTC);
      DynSequence dseq = orb.create_dyn_sequence(colTC); 

      System.out.println("\n Test: dseq.insert_ulong ");

      for(i=0;i<dseq.length();i++) dseq.insert_ulong(i*5);

      System.out.print("\n Test: dseq.get_ulong() ");
      try { System.out.println("ERROR: "+dseq.get_ulong()+" should not be produced!"); }
      catch(org.omg.CORBA.DynAnyPackage.TypeMismatch in) 
      { System.out.println("CORRECT: It is after the end of the sequence "); }

      System.out.println("\n Test: dseq.rewind() and access each element with get_ulong() ");
      dseq.rewind();
      for(i=0;i<dseq.length();i++) System.out.println("seq["+i+"]: "+dseq.get_ulong());

      System.out.println("\n Test: dseq.seek(0) and access each element with ");
      System.out.println("       current_component() and next()");
      dseq.seek(0);
      for(i=0;i<dseq.length();i++) {
        DynAny a = dseq.current_component();
        System.out.println("seq["+i+"]: "+a.get_ulong());
        dseq.next();  
      }// end of for

      System.out.println("\n Test: dseq.set_elements() - add 4 elements after");
      System.out.println("       changing the size with length() ");
      Any[] vals = new Any[4];
      for(i=0;i<4;i++) { 
        vals[i]=orb.create_any();
        vals[i].type(orb.get_primitive_tc(TCKind.tk_ulong));
        vals[i].insert_ulong(3*i);
      }// end of for

      dseq.length(4);
      dseq.set_elements(vals);
      for(i=0;i<dseq.length();i++) 
          System.out.println("seq["+i+"]: "+dseq.get_ulong());
      
      System.out.println("\n Test: clone1=dseq.copy() - print clone1 ");
      DynSequence clone1 = (DynSequence)dseq.copy();
      for(i=0;i<clone1.length();i++) 
          System.out.println("seq["+i+"]: "+clone1.get_ulong());

      System.out.println("\n Test: vals=clone1.get_elements() - print vals ");
      vals = clone1.get_elements();
      for(i=0;i<clone1.length();i++)  
        System.out.println("val["+i+"]: "+vals[i].extract_ulong());

      System.out.println("\n Test: seqA=clone1.to_any(), dseq2.from_any(seqA) - print dseq2 ");
      Any seqA = clone1.to_any();
      TypeCode colTC2= orb.create_sequence_tc(4,eleTC);
      DynSequence dseq2 = orb.create_dyn_sequence(colTC2); 
      dseq2.from_any(seqA);
      for(i=0;i<dseq2.length();i++) 
          System.out.println("dseq2["+i+"]: "+dseq2.get_ulong());
     
      System.out.println("\n Test: dseq.assign(dseq2) - print dseq ");
      dseq.assign(dseq2);
      for(i=0;i<dseq.length();i++) 
          System.out.println("dseq["+i+"]: "+dseq.get_ulong());

      System.out.println("**** DynSequence Interface TEST IS OVER **** ");

    } catch(Exception e) 
      { System.err.println("ERROR: " + e); e.printStackTrace(System.out); }  

  }// end of main

}// end of testDynSeq


