/**
*
* This is a basic test program for JWORB. 
*
* org.omg.* classes come with Java 1.2 beta2.
*
* @author H. Timucin Ozdemir
* @date   July 6, 1998
*/

// the naming service.
import org.omg.CosNaming.*;

// The package containing special exceptions thrown by the name service.
import org.omg.CosNaming.NamingContextPackage.*;

// All CORBA applications need these classes.
import org.omg.CORBA.*;

import java.util.*;
import java.io.*;

public class testDynStruct {

  /** */
  public static void printMyIOR(String fileName, String IOR) {
     try
     {  PrintWriter pw = new PrintWriter( new FileWriter(new File(fileName)));
        pw.println(IOR);
        pw.close();
     } catch( java.io.IOException io) {}
  }// end of printMyIOR()

  /** */
  public static void printDynStruct(DynStruct dstruct, String label) {
    try{
      dstruct.rewind();
      DynAny a = dstruct.current_component();
      System.out.println("===> "+label);
      System.out.println("testStruct1.s : "+a.get_string());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.l : "+a.get_long());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.f : "+a.get_float());
    } catch(org.omg.CORBA.DynAnyPackage.TypeMismatch tm) {}
  }// end of print()

  /** */
  public static void main(String args[]) {
    try{
    
      int i;
      Properties prop = new Properties();
      
      // to force our IIOP implementation.
      prop.put("org.omg.CORBA.ORBClass","worb.protocols.iiop.ORBImpl");
      prop.put("org.omg.CORBA.ORBSingletonClass","worb.protocols.iiop.ORBImpl");

      // Create and initialize the ORB
      ORB orb = ORB.init(args, prop);
      
      // Create a typecode
      StructMember[] tc_members = new StructMember[3];
      tc_members[0]=new StructMember("s",orb.get_primitive_tc(TCKind.tk_string),null);
      tc_members[1]=new StructMember("l",orb.get_primitive_tc(TCKind.tk_long),null);
      tc_members[2]=new StructMember("f",orb.get_primitive_tc(TCKind.tk_float),null);
      String[] enumValues = new String[3];
      enumValues[0]=new String("RED");  
      enumValues[1]=new String("BLUE");  
      enumValues[2]=new String("YELLOW");
      TypeCode eleTC= orb.get_primitive_tc(TCKind.tk_ulong);
      TypeCode colTC= orb.create_struct_tc("IDL:testStruct1:1.0","testStruct1",tc_members);
      DynStruct dstruct = orb.create_dyn_struct(colTC); 

      System.out.println("dstruct.insert_string(\"hello\")");
      dstruct.insert_string("hello");
      System.out.println("dstruct.insert_long(111)");
      dstruct.insert_long(111);
      System.out.println("dstruct.insert_float((float)222.222)");
      dstruct.insert_float((float)222.222);
      System.out.println("\n insert_XXXX is over ");

      dstruct.seek(0);
      DynAny a = dstruct.current_component();
      System.out.println("testStruct1.s : "+a.get_string());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.l : "+a.get_long());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.f : "+a.get_float());

      System.out.println("\n Test: a=dstruct.current_component(), a.insert_XXX ");
      dstruct.rewind();
      a = dstruct.current_component();
      System.out.println("a.insert_string(\"hello2\")");
      a.insert_string("hello2");
      if(dstruct.next()) {
        a = dstruct.current_component();
        System.out.println("a.insert_long(222)");
        a.insert_long(222);
      }// nd of if
      if(dstruct.next()) {
        a = dstruct.current_component();
        System.out.println("a.insert_float((float)333.333)");
        a.insert_float((float)333.333);
      }// end of if

      dstruct.rewind();
      a = dstruct.current_component();
      System.out.println("testStruct1.s : "+a.get_string());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.l : "+a.get_long());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.f : "+a.get_float());

      System.out.println("\n Test: dstruct.seek(1), a=dstruct.current_component(), a.insert_long(4444) ");
      dstruct.seek(1);
      a = dstruct.current_component();
      a.insert_long(4444);

      dstruct.rewind();
      a = dstruct.current_component();
      System.out.println("testStruct1.s : "+a.get_string());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.l : "+a.get_long());
      dstruct.next();  a = dstruct.current_component();
      System.out.println("testStruct1.f : "+a.get_float());

      System.out.println("\n Test: nvp=dstruct.get_members() ");
      NameValuePair[] nvp = dstruct.get_members();
      System.out.println("nvp[0].id    : "+nvp[0].id);  
      System.out.println("nvp[0].value : "+nvp[0].value.extract_string());  
      System.out.println("nvp[1].id    : "+nvp[1].id);  
      System.out.println("nvp[1].value : "+nvp[1].value.extract_long());  
      System.out.println("nvp[2].id    : "+nvp[2].id);  
      System.out.println("nvp[2].value : "+nvp[2].value.extract_float());  

      Any a1=orb.create_any();
      a1.type(orb.get_primitive_tc(TCKind.tk_string));
      a1.insert_string("*hello*");
      nvp[0] = new NameValuePair("s",a1);
      Any a2=orb.create_any();
      a2.type(orb.get_primitive_tc(TCKind.tk_long));
      a2.insert_long(321);
      nvp[1] = new NameValuePair("l",a2);
      Any a3=orb.create_any();
      a3.type(orb.get_primitive_tc(TCKind.tk_float));
      a3.insert_float((float)1234.4321);
      nvp[2] = new NameValuePair("f",a3);
     
      System.out.println("\n Test: dstruct.set_members(\"*hello*\",321,1234.4321) ");
      dstruct.set_members(nvp);
      testDynStruct.printDynStruct(dstruct,"dstruct");

      System.out.println("\n Test: dstruct2 = (DynStruct)dstruct.copy() - print dstruct2 ");
      DynStruct dstruct2 = (DynStruct)dstruct.copy();
      testDynStruct.printDynStruct(dstruct2,"dstruct2");

      System.out.println("\n Test: print dstruct2 ");
      dstruct2.rewind();
      a = dstruct2.current_component();
      a.insert_string("%hello%");
      testDynStruct.printDynStruct(dstruct2,"dstruct2");

      System.out.println("\n Test: dstruct3.assign(dstruct2) - print dstruct3 ");
      DynStruct dstruct3 = orb.create_dyn_struct(colTC);
      dstruct3.assign(dstruct2) ;
      testDynStruct.printDynStruct(dstruct3,"dstruct3");

      System.out.println("\n Test: aa=dstruct3.to_any(),dstruct.from_any(aa) - print dstruct ");
      Any aa = dstruct3.to_any();
      testDynStruct.printDynStruct(dstruct,"dstruct");
      dstruct.from_any(aa);
      testDynStruct.printDynStruct(dstruct,"dstruct");
  
      System.out.println("**** DynStruct Interface TEST IS OVER **** ");

    } catch(Exception e) 
      { System.err.println("ERROR: " + e); e.printStackTrace(System.out); }  

  }// end of main

}// end of testDynStruct


