//
// Tests the ORB for several basic data types.
//
// Usage: test1_clt 
//
//        Gets the IOR address of the server object from IOR file in the 
//        current directory.
//
// author H. Timucin Ozdemir
// date   July 6, 1998
//

#include <fstream.h>
#include <iostream.h>
#include <time.h>
#include "anyidl.hh"

/**
* Test each method on the interface
*/
void
test(CORBA::Object_ptr obj)
{
  AnyTest::aTest_var e = AnyTest::aTest::_narrow(obj);

  if (CORBA::is_nil(e)) {
    cerr << "jworbTest1: cannot invoke on a nil object reference.\n" << endl;
    return;
  }

  cout << " Object is OK! " << endl;

  /*
  // First ping
  cerr << " Before ping " << endl;
  e->ping();
  cerr << " Object pinged! " << endl;

  // echo char 
  CORBA::Char val1 = 'A';
  cerr << "(echoChar) I said,\"" << val1 << "\"."
       << " The Object said,\"" << e->echoChar(val1) <<"\"" << endl;

  // echo octet 
  CORBA::Octet val2 = 11 ;
  cerr << "(echoOctet) I said,\"" << (int)val2 << "\"."
       << " The Object said,\"" << (int)e->echoOctet(val2) <<"\"" << endl;

  // echo boolean 
  CORBA::Boolean val3 = 1 ;
  cerr << "(echoBoolean) I said,\"" << (int)val3 << "\"."
       << " The Object said,\"" << (int)e->echoBoolean(val3) <<"\"" << endl;

  // echo long value
  CORBA::Long dest;
  CORBA::Long val=1;

  cerr << " Before echoLong( " << val << " ) " << endl;
  dest = e->echoLong(val);

  cerr << "I said,\"" << val << "\"."
       << " The Object said,\"" << dest <<"\"" << endl;

  // echo Short value
  CORBA::Short val5  = 12344;
  CORBA::Short dest5 = e->echoShort(val5);

  cerr << "(echoShort) I said,\"" << val5 << "\"."
       << " The Object said,\""   <<   dest5 <<"\"" << endl;

  // echo Float value
  CORBA::Float val6  = 12344.3434;
  CORBA::Float dest6 = e->echoFloat(val6);

  cerr << "(echoFloat) I said,\"" << val6 << "\"."
       << " The Object said,\""   << dest6 <<"\"" << endl;

  // echo Double value
  CORBA::Double val7  = 3312344.3434;
  CORBA::Double dest7 = e->echoDouble(val7);

  cerr << "(echoDouble) I said,\"" << val7 << "\"."
       << " The Object said,\""    << dest7 <<"\"" << endl;

  // echo String value
  CORBA::String_var val8  = (const char*)"Hello There!!!";
  CORBA::String_var dest8 = e->echoString(val8);

  cerr << "(echoString) I said,\"" << val8 << "\"."
       << " The Object said,\""    << dest8 <<"\"" << endl;

  // echo structure value
  jworbTest1::long_struct valStruct = {123};
  jworbTest1::long_struct retStruct ;

  cerr << " Before echoLong_struct( " << valStruct.number << " ) " << endl;
  retStruct = e->echoLong_struct(valStruct);
  cerr << "I said,\"" << valStruct.number << "\"."
       << " The Object said,\"" << retStruct.number <<"\"" << endl;

  // Run echoLongArray
  int i, size = 10;
  jworbTest1::longArray *destIntArr;
  jworbTest1::longArray valIntArr;
  valIntArr.length(size);
  for(i=0;i<size;i++) valIntArr[i] = i;

  cerr << " Before echoLongArray( " ;
  for(i=0;i<size;i++) cerr << valIntArr[i] <<" , " ;
  cerr << " ) " << endl;

  destIntArr = e->echoLongArray(valIntArr);

  cerr << " The object said  ( " ;
  for(i=0;i<destIntArr->length();i++) cerr << (*destIntArr)[i] <<" , " ;
  cerr << " ) " << endl;

  // Run echoLong_struct
  jworbTest1::long_structArray *destStrArr;
  jworbTest1::long_structArray valStrArr;
  valStrArr.length(size);
  for(i=0;i<size;i++) 
  {
     jworbTest1::long_struct a ;
     a.number = i;
     valStrArr[i] = a;
  }

  cerr << " Before echoLong_structArray( " ;
  for(i=0;i<valStrArr.length();i++) cerr << valStrArr[i].number <<" , " ;
  cerr << " ) " << endl;

  destStrArr = e->echoLong_structArray(valStrArr);

  cerr << " The Object said ( " ;
  for(i=0;i<destStrArr->length();i++) cerr << (*destStrArr)[i].number <<" , " ;
  cerr << " ) " << endl;

  // Exception

  try
  {
    cout << " Before exception test! " << endl;
    e->getException();

  } catch(jworbTest1::ObjectNotKnown& ex )    
    {
       cout << " ObjectNotKnown is received! " << endl;
    }
  */

}// end of test()

int
main (int argc, char **argv) 
{
  CORBA::ORB_ptr orb = CORBA::ORB_init(argc,argv,"omniORB2");
  CORBA::BOA_ptr boa = orb->BOA_init(argc,argv,"omniORB2_BOA");

  ifstream in("IOR");
  char *buf = new char[400];
  in.getline(buf,400);
  cout << ">" << buf<< "<" << endl;
  in.close();

  try 
  {

    CORBA::Object_var obj = orb->string_to_object(buf);
    test(obj);
  } catch(CORBA::COMM_FAILURE& ex) {
    cerr << "Caught system exception COMM_FAILURE, unable to contact the "
         << "object." << endl;
  } catch(omniORB::fatalException& ex) {
    cerr << "Caught omniORB2 fatalException. This indicates a bug is caught "
         << "within omniORB2.\nPlease send a bug report.\n"
         << "The exception was thrown in file: " << ex.file() << "\n"
         << "                            line: " << ex.line() << "\n"
         << "The error message is: " << ex.errmsg() << endl;
  } catch(...) 
    { cerr << "Caught a system exception." << endl; }

  return 0;

}// end of main
