// test2_clt.cc - tests exception, receiving an object reference from
//                 server and registering a callback object to the server.
//
// Usage: test2_clt 
//
//                This is the client. The object reference is read from
//                IOR file in the current directory.
//
//
// author H. Timucin Ozdemir
// date   July 6, 1998
//
#include <fstream.h>
#include <iostream.h>
#include <time.h>
#include "jworbTest2.hh"
#include "callback_i.h"

/**
* Test each method on the interface
*/
void
test(CORBA::Object_ptr obj)
{
  jworbTest2_var e = jworbTest2::_narrow(obj);

  if (CORBA::is_nil(e)) {
    cerr << "hello: cannot invoke on a nil object reference.\n" << endl;
    return;
  }

  cout << " Object is OK! " << endl;

  cerr << " Before getEchoObject( " ;

  echo1_ptr echo_obj =  e->getEchoObject();

  try
  { 
     echo_obj =  e->getEchoObject();
     CORBA::String_var src = (const char*) "Hello!"; // String literals are not
     CORBA::String_var ans;
     cerr << " ******* I got the Object **** " << endl;

     ans = echo_obj->echo(src);

     cerr << "I said,\"" << src << "\"."
          << " The Object said,\"" << ans <<"\"" << endl;

  } catch( jworbTest2::ObjectNotKnown &e )
    {
        cout << " Object NotKnown " << endl;
    }

  cerr << " I defined the object " << endl;
  callback_i *myCallBack = new callback_i();

  int argv = 0;
  char **args = NULL;
  CORBA::ORB_ptr orb = CORBA::ORB_init(argv,args,"omniORB2");
  CORBA::BOA_ptr boa = orb->BOA_init(argv,args,"omniORB2_BOA");
  myCallBack->_obj_is_ready(boa);
  boa->impl_is_ready(0,1);

  cerr << " myCallBack is bound to BOA " << endl;
  CORBA::String_var p = orb->object_to_string(myCallBack->_this());
  cerr << "'" << (char*)p << "'" << endl;

  e->registerCallback(myCallBack); 
  cerr << " myCallBack is registered " << endl;

  char myMessageBuffer[180];
  int stay = 1;

  while(stay)
  {
     cout << " Message ";
     cin >> myMessageBuffer;
     if(!strcmp(myMessageBuffer,"exit")) stay = 0;
     else e->broadcast(myMessageBuffer);

  }// end of while

  if (!CORBA::is_nil(e))
     try
     { 
        cerr << "Let's try exception " << endl;

        echo_obj =  e->getEchoObjectWithException();

     } catch( jworbTest2::ObjectNotKnown &e )
       {
        cout << " Object NotKnown " << endl;
       }

}// end of test()

int
main (int argc, char **argv) 
{
  CORBA::ORB_ptr orb = CORBA::ORB_init(argc,argv,"omniORB2");
  CORBA::BOA_ptr boa = orb->BOA_init(argc,argv,"omniORB2_BOA");

  ifstream in("IOR");
  char *buf = new char[400];
  in.getline(buf,400);
  cout << ">" << buf<< "<" << endl;
  in.close();

  try 
  {

    CORBA::Object_var obj = orb->string_to_object(buf);
    test(obj);
  } catch(CORBA::COMM_FAILURE& ex) {
    cerr << "Caught system exception COMM_FAILURE, unable to contact the "
         << "object." << endl;
  } catch(omniORB::fatalException& ex) {
    cerr << "Caught omniORB2 fatalException. This indicates a bug is caught "
         << "within omniORB2.\nPlease send a bug report.\n"
         << "The exception was thrown in file: " << ex.file() << "\n"
         << "                            line: " << ex.line() << "\n"
         << "The error message is: " << ex.errmsg() << endl;
  } catch(...) 
    { cerr << "Caught a system exception." << endl; }

  return 0;

}// end of main()
