#ifndef _GlobalTypes_
#define _GlobalTypes_

#ifndef NULL
#define NULL 0
#endif
#if defined(_WIN32) || defined(SGI6_5)
typedef bool EBoolean;
#else
enum EBoolean {
   false, true
};
#endif

// simple dictionary entry/association; helper structure for
// parsing input files
struct TKeyValuePair
{
#ifndef _WIN32
	const char* key;
#else 
	char * key;
#endif
	int value;
};

enum EKeywordType {
	ekeyString, 
   ekeyFloat, 
   ekeyDouble, 
   ekeyInt, 
   ekeyChar, 
   ekeyLong,
	ekeyUnsignedInt, 
   ekeyUnsignedLong, 
   ekey3DPoint,          //  T3Ddouble, written and read in parentheses
   ekey2DPoint,          // T2Dsingle, written and read in parentheses
   ekey3DPointList,      // array of T3Ddouble, r/w in brackets and parens
   ekey2DPointList,      // array of T2Dsingle, r/w in brackets and parens
   ekeyIntList, 
   ekeyFloatList,
	ekeyLev1Struct, 
   ekeyLev2Struct,
	ekeyQuotedString,
   ekeyVulnPoint,        // 4 floats, enclosed in  parens
   ekeyVulnPointList     // array of vuln pts, in  brackets
};

struct TKeywordEntry
{
	char *         key;
	EKeywordType   type;
};
/*
int KeyIndexLookup( const TKeyValuePair *, int, const char * );
int StringIndexLookup( const char** strArray, int size, const char* str );
*/

//   Time units.  DIS keeps time as high 31 bits for 1 hour, lsb to indicate absolute time
typedef   unsigned long   TTimeMs;    // SIMNET protocols need milliseconds
typedef   float           TTimeSec;
typedef   unsigned long   TTimeDis;   // second lsb = 1.676.. microseconds
typedef   unsigned long   TTimeSim;   // internal simulator time, ms for now


//   id for other simulated entities
struct TEntityID {
   unsigned short   site;
   unsigned short   application;      // when using SIMNET, this is host
   unsigned short   entityNum;
   
   int   operator==(const TEntityID& eId) const { return (site == eId.site) && 
                                                   (application == eId.application) &&
                                                   (entityNum == eId.entityNum); }
   
   TEntityID( unsigned short s = 0, unsigned short a = 0, unsigned short e = 0)
         { site = s; application = a; entityNum = e;}
};

//  Entity types, in DIS format
const unsigned char  kMatchAllChar  = 0xff;
const unsigned short kMatchAllShort = 0xffff;
class cInFileObj;
class ofstream;
class TItemList;

struct TEntityType   {
   /*unsigned char      kind;    
   unsigned char      domain;  
   unsigned short     country; 
   unsigned char      category; 
   unsigned char      subcategory; 
   unsigned char      specific; 
   unsigned char      extra;  

      // keywords in files, defining input fields
   enum   {
		eentkwKindInt,eentkwDomainInt,
		eentkwCountryInt, eentkwCategoryInt,
		eentkwSubcategoryInt, eentkwSpecificInt, 
      eentkwExtraInt, eentkwKind, 
      eentkwDomain, eentkwCountry, 
      eentkwCategory, eentkwSubcategory, 

		eSizeEntityTable
   };
   static const TKeywordEntry fKeyEntityTable[eSizeEntityTable];
   
   enum  { enumMineSubcatPairs = 9 };
//   static const TKeyValuePair fsCountryText[eNumCountryPairs];
   
   static const char*   fsKindText[enumKind];
   static const char*   fsMunDomText[enumMunDom];   // short names, for tables
   static const char*   fsCountryText[enumCountry];
   static const char*   fsMunCategoryText[enumMunCat];
   static const TKeyValuePair fsMineSubcatText[enumMineSubcatPairs];
    
       // accessors for static text values
   static   const char*       GetKindText(int knd);
   static   const char*       GetMunitionDomainText(int dom);
   static   const char*       GetCountryText(int cntry);
        // accessors for text values for this TEntityType object
            const char*       GetKindText() const;
            const char*       GetMunitionDomainText() const;
            const char*       GetCountryText() const;
  
      // getting values from input files
   EBoolean   ParseList( TItemList* pList);
   EBoolean   ParseEntityType(cInFileObj& infile);
   EBoolean   ReadEntityStruct(cInFileObj& infile);
   void       WriteEntityType(ofstream* file);

   void     SetAllWilds();

   int   MatchKind(const TEntityType& t) const { return (kind == t.kind) || 
                        (kind == kMatchAllChar) || (t.kind == kMatchAllChar); }
   int   MatchDomain(const TEntityType& t) const { return (domain == t.domain) || 
                        (domain == kMatchAllChar) || (t.domain == kMatchAllChar); }
   int   MatchCountry(const TEntityType& t) const { return (country == t.country) || 
                        (country == kMatchAllShort) || (t.country == kMatchAllChar); }
   int   MatchCategory(const TEntityType& t) const { return (category == t.category) || 
                        (category == kMatchAllChar) || (t.category == kMatchAllChar); }
   int   MatchSubcategory(const TEntityType& t) const { return (subcategory == t.subcategory) || 
                        (subcategory == kMatchAllChar) || (t.subcategory == kMatchAllChar); }
   int   MatchSpecific(const TEntityType& t) const { return (specific == t.specific) || 
                        (specific == kMatchAllChar) || (t.specific == kMatchAllChar); }
   int   MatchExtra(const TEntityType& t) const { return (extra == t.extra) || 
                        (extra == kMatchAllChar) || (t.extra == kMatchAllChar); }
   int   MatchType(const TEntityType& t) const {return MatchKind(t)
                                                    && MatchDomain(t)
                                                    && MatchCountry(t)
                                                    && MatchCategory(t)
                                                    && MatchSubcategory(t)
                                                    && MatchSpecific(t)
                                                    && MatchExtra(t); }

   int   operator==(const TEntityType& eTp) const { return (kind == eTp.kind) && 
                                                    (domain      == eTp.domain) &&
                                                    (country     == eTp.country) &&
                                                    (category    == eTp.category) &&
                                                    (subcategory == eTp.subcategory) &&
                                                    (specific    == eTp.specific) &&
                                                    (extra       == eTp.extra); }
    
   TEntityType(unsigned char k = 0, unsigned char d = 0, unsigned short cn = 0, unsigned char ct = 0,
               unsigned char sc = 0, unsigned char sp = 0, unsigned char e = 0)
               {kind = k; domain = d; country = cn; category = ct; subcategory = sc; specific = sp; extra = e;}
*/
};

   // database entry base class, all db entries must descend from this
   // in general using template arguments and params to support specific
   // classes
/*
class TVehDimensions;
class TFuzeParam;
class TMineParam;

class cDbEntry {

 friend class cEntityDBMgr;
 
	enum { eentrykwEntityType, eentrykwEntityTypeList, eentrykwName, eentrySizeTable };
   static const TKeywordEntry fKeyEntryTable[eentrySizeTable];

 protected:
   TEntityType   fEntityType;
   char*         fpName;
   char*         fpAltName;   // generally used for shorter names

      // for convenience we keep three links, left/parent, right, and child
   cDbEntry*         fpPrevEntry;       // left sibling or parent
   cDbEntry*         fpNextEntry;       // right sibling or NULL
   cDbEntry*         fpChildEntry;      // first child (would be left in normal binary tree)
   
 public:
         cDbEntry*       GetPrevEntry()  const     {return fpPrevEntry;}
         cDbEntry*       GetNextEntry()  const     {return fpNextEntry;}
         cDbEntry*       GetChildEntry() const     {return fpChildEntry;}
         cDbEntry*       GetParentEntry() const;

   virtual  EBoolean     ReadParamSet(cInFileObj& infile);
   virtual  EBoolean     ParseDbEntry(cInFileObj& infile);
   virtual  EBoolean     ValidType(TEntityType& type);
   virtual  int          GetDbLevel();
   void                  WriteDbEntry(ofstream* file);
   void                  InitFuzeParamTables(TFuzeParam* pFuzeParam);  // blank copy of appropriate type
   TFuzeParam*           CreateNewFuzeParam(int fuzeType);  // blank copy of appropriate type
   TFuzeParam*           CreateFuzeParam(int fuzeType);  // blank copy of appropriate type
   TMineParam*           CreateMineParam(int projType);  // blank copy of appropriate type

            EBoolean     IsAbstract();   // not instantiable, not a terminal node

            void         SetEntryEntityType(TEntityType& type) { fEntityType = type; }
   
   TEntityType&   GetEntryEntityType() { return fEntityType; }
   char*          GetEntryEntityName() { return fpName; }
   char*          GetEntryAltName()    { return fpAltName; }
 
              cDbEntry();         // new input definitions, data not yet available
              cDbEntry(cEntityDBMgr* pDb, TEntityType& type, int level, 
                  char* name = NULL, char* altname = NULL);
//              cDbEntry(TEntityType& type, char* name, TVehDimensions size);
   virtual   ~cDbEntry();
};

class cDbStruct   {

 public:
             EBoolean   ReadDbStruct(cInFileObj& infile);
   virtual   EBoolean   ParseDbStructItem(cInFileObj&);
   
   virtual   ~cDbStruct();
};

//  startup parameters
*/
enum EState          { estNull, estInitialize, estStartup, estRun, estQuit };

enum EProtocol       { eDis203, eDis204, eNonet, eDis213, eDisPP, eHLA };

enum EQrpProtocol    { eQrpNo, eQrp1, eQrp2, eQrp3 };
enum ECpProtocol     { eCpNo, eCp1HeartBeat, eCp1, eCp1Mcast, eCp2HeartBeat, eCp2, eCp2Mcast };

// struct and definitions for command line mfield loading

struct TLoadMf {
    char*     ffilename;
    EBoolean  fesim;       // true if simulating, false otherwise 
};

struct TStartupParam {
   EState     fStartResult;
   EProtocol  feProtocol;
   EQrpProtocol   feQrpProtocol;
   ECpProtocol    feCpProtocol;
   EBoolean       fbOrientation;
   EBoolean       fbBurialDepth;
   EBoolean       fbTemperature;
   char*      fSetupFileName;
   char*      fTerrainName;
   char*      fDirMineParams;
   char*      fDirMfInput;
   char*      fDirMfOutput;
   char*      fDirStats;
   char*      fExerciseName;
   char*      fTestFilenames[50];
   char*      fOutputFilePrefix;
   int        fNumTestFiles;
   int        fDisPort;         // DIS protocol, including M/CM
   int        fCmsPort;         // CMS internal, for separate gui/host
   char*      fMulticastAddr;    //Multicast address
   int        fTtl;             //Multicast time-to-live (must also specify fMulticastAddr)
   int        fSite;
   int        fApplication;
   int        fExerciseId;
   TEntityID  fstealthID;
   EBoolean   fbGui;
   EBoolean   fbHost;
   EBoolean   fbRandomSeed;
   EBoolean   fbOperatorPriv;
   int        fRnSeed;
   int        fNumMineFiles;
   TLoadMf    fMineFilenames[50];  // filename + simulate boolean
   
   TStartupParam();      // provides defaults
};

/*
//   Force
enum   EForce {   eOther = 0, eFriendly, eEnemy, eNeutral, eNumForceTypes };
extern const char*   pForceText[eNumForceTypes];

//   doubly linked list
struct T2link {
 friend struct T2linkHead;
 friend class cVehListHead;
 friend class cObjTypeListHead;
 friend class cEmissListHead;

 protected:
   T2link*   fpNext;
   T2link*   fpPrev;
   
 public:
             T2link(T2link* pPrev);
   virtual  ~T2link();
};

struct T2linkHead : public T2link {

 protected:
   T2link*   fpCurrent;
   
 public:
   T2link*        GetFirstLink();
   T2link*        GetNextLink();
   T2link*        GetLastLink();
   T2link*        GetNULLLink()      { return this; }
   T2link*        GetLinkNum(int iLink);
	
   EBoolean	  IsEmpty() { return fpNext == this ? true : false; }

   virtual void   DeleteList();
   void           DeletePoint (int iPt);
   
                  T2linkHead(T2linkHead* pOldHead);   // takes the list
                  T2linkHead() : T2link( this ) { fpCurrent = this; }      // point to self
   virtual       ~T2linkHead();               // doesn't' delete list, call DeleteList first
};

class cMinefield;
class cMfComponent;
class cMfieldObj;
struct TSelection : public T2link {
      // selections form a list.  The first link should be the
      // minefield only, then the mf + component, then mf + comp + mfobj
      //  Multiselection at any level precludes going to lower level
      
   cMinefield*   fpMinefield;
   cMfComponent* fpComponent;
   cMfieldObj*   fpMfieldObj;
   
   cMinefield*   GetMinefield() { return fpMinefield; }
   cMfComponent* GetComponent() { return fpComponent; }
   cMfieldObj*   GetMfieldObj() { return fpMfieldObj; }
   
   EBoolean      MatchSel(cMinefield* pMf = NULL, cMfComponent* pCmp = NULL, 
                             cMfieldObj* pObj = NULL);  // is *this the same selection
   EBoolean      IsParentSel(cMinefield* pMf = NULL, cMfComponent* pCmp = NULL, 
                             cMfieldObj* pObj = NULL);  // is *this TSelection the parent
   EBoolean      IsChildSel(cMinefield* pMf = NULL, cMfComponent* pCmp = NULL, 
                             cMfieldObj* pObj = NULL);  // is *this TSelection the child
   
   TSelection(TSelection* pPrev, cMinefield* pMf = NULL, 
              cMfComponent* pCmp = NULL, cMfieldObj* pObj = NULL);
   virtual ~TSelection();
};

//   Rectangular coordinates

struct T3Dsingle {
   float   x;
   float   y;
   float   z;

   float       Distance( const T3Dsingle& pt) const;
   float       HorzDistance( const T3Dsingle& pt) const;
   T3Dsingle   unit() const;
   
   T3Dsingle    operator*(float);
   T3Dsingle    operator*(double);
   T3Dsingle    operator*(TTimeSim);
   T3Dsingle    operator-(const T3Dsingle&);
   T3Dsingle    operator+(const T3Dsingle&);
   T3Dsingle&   operator+=(const T3Dsingle&);
   T3Dsingle&   operator-=(const T3Dsingle&);
   T3Dsingle&   operator*=(float);

   T3Dsingle( const T3Dsingle& a )
            { x = a.x; y = a.y; z = a.z;}
   T3Dsingle(float a = 0.0, float b = 0.0, float c = 0.0)
            { x = a; y = b; z = c;}

};

struct   TEntCoord : public T3Dsingle
{
   TEntCoord(const T3Dsingle& s);
   TEntCoord(float a = 0.0, float b = 0.0, float c = 0.0) :
            T3Dsingle(a,b,c) {}
};
struct T2Dint {
   int   x;
   int   y;
   
   T2Dint(int a = 0, int b = 0) {x = a; y = b;}
};

struct   TMfieldObjCoord : public T3Dsingle
{
   float   depth;
   
   TMfieldObjCoord(float a = 0.0, float b = 0.0, float c = 0.0, float d = 0.0) :
            T3Dsingle(a,b,c) { depth = d; }
};

struct   TOrient
{
   float   psi;
   float   theta;
   float   phi;
   
   void    SetOrient( float a, float b, float c)
            { psi = a; theta = b; phi = c;}
   
   TOrient(float a = 0.0, float b = 0.0, float c = 0.0)
            { psi = a; theta = b; phi = c;}
};

struct   TAngVel : public T3Dsingle
{
   TAngVel(float a = 0.0, float b = 0.0, float c = 0.0) :
            T3Dsingle(a,b,c) {}
};

struct T3Ddouble {
   double   x;
   double   y;
   double   z;
   
   double       Distance( const T3Ddouble& pt) const;
   double       HorzDistance( const T3Ddouble& pt) const;
   
   T3Ddouble    operator*(double);
   T3Ddouble    operator-(const T3Ddouble&) const;
   T3Ddouble    operator+(const T3Ddouble&) const;
   T3Ddouble    operator+(const T3Dsingle&) const;
   T3Ddouble&   operator+=(const T3Dsingle&);
   T3Ddouble&   operator+=(const T3Ddouble&);
   T3Ddouble(double a = 0.0, double b = 0.0, double c = 0.0)
            { x = a; y = b; z = c; }
};

struct   TVelocity : public T3Dsingle
{
   TVelocity( const T3Dsingle& a ) : T3Dsingle(a) {}
   TVelocity(float a = 0.0, float b = 0.0, float c = 0.0) :
            T3Dsingle(a,b,c) {}
};

struct   TLocation : public T3Ddouble
{
   TLocation(const T3Ddouble& d);
   TLocation(double a = 0.0, double b = 0.0, double c = 0.0) :
            T3Ddouble(a,b,c) {}
};

struct T2Dsingle {
   float   x;
   float   y;
   
   T2Dsingle(float a = 0.0, float b = 0.0) {x = a; y = b;}
};

struct   TFlatEntCoord : public T2Dsingle
{
   TFlatEntCoord(float a = 0.0, float b = 0.0) :
            T2Dsingle(a,b) {}
};

struct   TAccel : public T3Dsingle
{
   TAccel(float a = 0.0, float b = 0.0, float c = 0.0) :
            T3Dsingle(a,b,c) {}
};

struct   TRotation {
   T3Ddouble   x;
   T3Ddouble   y;
   T3Ddouble   z;

   TRotation   operator*(const TRotation&) const;
   T3Dsingle   operator*(const T3Ddouble&) const;
   T3Dsingle   operator*(const T3Dsingle&) const;
   T3Ddouble   operator|(const T3Dsingle&) const;      // mult by transpose
   
   void        Mult( const T3Dsingle& in, T3Dsingle* out) const;
   void        Mult( const T3Ddouble& in, T3Dsingle* out) const;
   void        Mult( const T3Ddouble& in, T3Ddouble* out) const;
   void        MultTrans( const T3Dsingle& in, T3Dsingle* out) const;
   void        MultTrans( const T3Dsingle& in, T3Ddouble* out) const;
   void        MultTrans( const TRotation& in, TRotation* out) const;
   void        MultOtherTrans( const TRotation& other, TRotation* out) const;

   void        SetRotation ( float psi, float theta, float phi );
   void        SetRotation ( const TOrient& );
   void        SetOrientation( TOrient* orient );
   
   TRotation ();
   TRotation( float psi, float theta = 0.0, float phi = 0.0 );
   TRotation ( const TOrient& );
};

struct T2dMove {
      // used to move points, not change coord systems,
      // note rotation matrix initialization
   double    fDelX;
   double    fDelY;
   float     fDelAzimuth;
   TRotation fRotMatrix;
   
   void   SetMove(double dx, double dy = 0.0, double daz = 0.0);
   
   EBoolean   IsNoMove() { return ( ( fDelX == 0.0 ) &&
                                    ( fDelY == 0.0 ) &&
                                    ( fDelAzimuth == 0.0 )  ) ? true: false; }
   
   T2dMove( double dx = 0.0, double dy = 0.0, double daz = 0.0 ) :
      fRotMatrix( - daz, 0.0, 0.0) {fDelX = dx; fDelY = dy; fDelAzimuth = daz; }
};

struct TMunitionMapEntry {
   TEntityType   fMunitionType;
   int           fDetonType;
};

struct TWarheadMapEntry {
   int           fWarheadType;
   int           fDetonType;
};

struct TDetonNameBlock : public T2link {
   enum   { eDetNameBlockSize = 250 };
   
   int      fNumEntries;
   int      fBaseIndex;
   
   char*    fpName[eDetNameBlockSize];

   EBoolean     AddDetonName( const char* pName, int* pIndex );
   EBoolean     GetDetIndexName( const char* pName, int* pIndex );

   TDetonNameBlock(T2link* pPrev, EBoolean first = false);
   virtual ~TDetonNameBlock();
};

struct TMunitionMapBlock : public T2link {
   enum   { eMunMapBlockSize = 80 };
   
   int      fNumEntries;

   TMunitionMapEntry   fMunMap[eMunMapBlockSize];

   EBoolean     AddMunEntry(const TEntityType& entType, int index);
   EBoolean     GetDetIndexMun( const TEntityType& entType, int* pIndex );

   TMunitionMapBlock(T2link* pPrev) : 
         T2link(pPrev) { fNumEntries = 0; }
   virtual ~TMunitionMapBlock();
};

struct TWarheadMapBlock : public T2link {
   enum   { eWhMapBlockSize = 125 };
   
   int      fNumEntries;

   TWarheadMapEntry   fWhMap[eWhMapBlockSize];

   EBoolean     AddWhEntry(int warhead, int index);
   EBoolean     GetDetIndexWh( int warhead, int* pIndex );

   TWarheadMapBlock(T2link* pPrev) : 
         T2link(pPrev) { fNumEntries = 0; }
   virtual ~TWarheadMapBlock();
};

struct cDetonNameListHead : public T2linkHead {

   int      GetDetIndexName(const char* pName);
   
   int      AddDetonName( char* pName );   // returns index

   cDetonNameListHead();
   virtual ~cDetonNameListHead();
};

struct cMunMapListHead : public T2linkHead {

   int      GetDetIndexMun(const TEntityType& entType);
   
   void     AddMunMapEntry(const TEntityType& entType, int index);
   
   cMunMapListHead();
   virtual ~cMunMapListHead();
};

struct cWarheadMapListHead : public T2linkHead {

   int      GetDetIndexWh(int warhead);
   
   void     AddWhMapEntry(int warhead, int index);

   cWarheadMapListHead();
   virtual ~cWarheadMapListHead();
};

class cDetonationDisp;

struct cDetonation {

   enum   {
      edetOtherDet,
      edetEsmb,
      edetMiclic,
      edetLgBlast,
      edetMedBlast,
      edetSmBlast,
      edetLgFrag,
      edetSmFrag,
      
      edetNumTypes
   };
 protected:
      // names for constructing default det types
   static   char*      fDetTypeNames[edetNumTypes];
   enum { emundetkwName, emundetkwMunTypeList, emundetkwMunTypeStruct,
          eSizeMunParamTable};
   static   TKeywordEntry   fKeyMunParamTable[eSizeMunParamTable];
   enum { ewhdetkwName, ewhdetkwWarhead,
          eSizeWhParamTable};
   static   TKeywordEntry   fKeyWhParamTable[eSizeWhParamTable];

   static   cMunMapListHead       fMunitionMapList;
   static   cWarheadMapListHead   fWarheadMapList;
   static   cDetonNameListHead    fDetonNameList;
            cDetonationDisp*      fpDisplay;
   
 public:
      // reflection of detonation pdu, in cms types
   TEntityID      fFiringEntity;
   TEntityID      fTargetEntity;
   TEntityID      fMunitionEntity;
   TVelocity      fVelocity;
   TEntityType    fMunitionType;
   EWarhead       feWarhead;
   EFuze          feFuze;
   
   TLocation      fGeoLoc;
   TEntCoord      fEntityLoc;      // in target coord system, if targeted
   EDetonResult   feResult;
   
      // cms interpretation
   int            fDetonType;
   TLocation      fTccLoc;
   TRotation      fTccRot;	      // used for ESMB and MICLIC, where vel gives direction
   TOrient        fTccOrient;	      // used for ESMB and MICLIC, where vel gives direction
   
            void  FinishDetInit();   // fill the cms interpretation from pdu stuff
            
   int            GetDetonType() { return fDetonType; }
//   char*          GetDetonTypeName() { return fDetonNameList(fDetonType); }
   TLocation&     GetTccLoc()    { return fTccLoc; }
   TRotation&     GetTccRot()    { return fTccRot; }
   TOrient&       GetTccOrient() { return fTccOrient; }
   TEntCoord&     GetEntityLoc() { return fEntityLoc; }
   EBoolean       IsAreaType();      // esmb or miclic
   void           WriteDetInfo(ofstream* pFile);
   
   static   void  InitDetonMap();
   static   void  ReadParamMunDetType( cInFileObj& infile );
   static   void  ReadParamWhDetType( cInFileObj& infile );
   
   cDetonation();
   virtual ~cDetonation();
};

   // cEntity is an abstract root class for objects that have entity ids.
   //  The primary purpose is to keep track of internally simulated objects
   //  and quickly finding one addressed in incoming PDUs
struct TKeywordEntry;
class cInFileObj;
class cInteraction;
class cLinkVehGeom;
class ofstream;
class cEntity {
 
 private:
   static       TEntityType   fsEntEntityType;
   static       char*         fsEntName;

 protected:
   TEntityID        fId;

   TLocation        fGeoLocation;
   TOrient          fGeoOrient;
   TRotation        fGeoRotation;

   TLocation        fTccLocation;
   TOrient          fTccOrientation;
   TRotation        fTccRotation;
   
      // links are used to support linear list in mf component, 
   cEntity*         fpPrevEnt;       // left
   cEntity*         fpNextEnt;       // right
   
   EBoolean         fbSimulating;
   
 public:
   const TEntityID&     GetEntityID()  const    {return fId; }
   virtual  const TEntityType&   GetEntType() = 0;
   const TLocation&     GetGeoLoc()  const      {return fGeoLocation;}
   const TOrient&       GetGeoOrient()  const   {return fGeoOrient;}
   const TRotation&     GetGeoRot()  const      {return fGeoRotation;}
   const TLocation&     GetTccLoc()  const      {return fTccLocation;}
   const TOrient&       GetTccOrient()  const   {return fTccOrientation;}
   const TRotation&     GetTccRot()  const      {return fTccRotation;}
   
         cEntity*       GetPrevEnt()  const     {return fpPrevEnt;}
         cEntity*       GetNextEnt()  const     {return fpNextEnt;}
         void           ReplaceEnt( cEntity* pOld );
         
   virtual cInteraction*  CreateNewInteraction(cLinkVehGeom*);
   virtual void           UpdateInteractions();
   virtual   void        WriteEntSumm(ofstream*);
         
                 EBoolean       IsSimulating()   { return fbSimulating; }
   virtual       char*          GetEntName() = 0;

   virtual TLocation    GetDrGeoLoc(TTimeSim time);
   virtual TLocation    GetDrTccLoc(TTimeSim time);

   virtual void   SetCoordTccRef(TLocation tccLoc, float psi, float theta, float phi);
   virtual void   SetCoordGeoRef(TLocation geoLoc, float psi, float theta, float phi);
   virtual void   SetCoordGeoRef(TLocation geoLoc, TOrient orient);
   virtual void   SetTccOrient(float psi, float theta, float phi);      // existing location
   virtual void   SetTccCoordFromGeo();
   virtual void   SetGeoCoordFromTcc();
                  // allows specific types to initialize their orientation on creation
   void           CopyEntityParams(cEntity* pOriginal);
   
   void                 GeoToLocalLoc(const T3Ddouble& geoLoc, T3Dsingle* pLocLoc);
   void                 TccToLocalLoc(const T3Ddouble& tccLoc, T3Dsingle* pLocLoc);
   void                 LocalToGeoLoc(const T3Dsingle& locLoc, T3Ddouble* pGeoLoc);
      // vehicles with moving ref systems override velocity conversions
   virtual   void       GeoToLocalVel(const T3Dsingle& geoVel, T3Dsingle* pLocVel);
   virtual   void       LocalToGeoVel(const T3Dsingle& locVel, T3Dsingle* pGeoVel);
   void                 LocalToTccLoc(const T3Dsingle& locLoc, T3Ddouble* pTccLoc);
   
 
   virtual void    InDetTargeted(cDetonation* pDet);
    
                 cEntity(cEntity* pOrig = NULL);
   virtual      ~cEntity();
};

//   Messages have a root class for each message list, that puts itself on the list
class cMessage {

   friend class cMessageList;
   
   protected:
      TTimeSim           fSendTime;
      cMessage*          fpNextMsg;
      cMessage*          fpPrevMsg;
   
   public:
      virtual   cMessageList*  GetMsgList() = 0;
      virtual   void           SendMsg() = 0;
      TTimeSim                 GetSendTime() const;
      
                         cMessage(TTimeSim time, EBoolean bAbsTime = false);
      virtual           ~cMessage();           // does nothing, caller must fix links
};

class cMsgPrEvent: public cMessage {

   friend class cPrEventList;
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgPrEvent();
      virtual           ~cMsgPrEvent();
};

class cMsgPrUpdt: public cMessage {

   friend class cPrUpdtList;
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgPrUpdt(TTimeSim waitTime);
      virtual           ~cMsgPrUpdt();
};

class cMsgEvent: public cMessage {

   friend class cEventList;
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgEvent();
      virtual           ~cMsgEvent();
};

class cMsgUpdt: public cMessage {

   friend class cUpdtList;
   
   private:
      static   TTimeSim   fUpdtWaitTime;      // between updates
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgUpdt();
      virtual           ~cMsgUpdt();
};

class cMsgDispUpdt: public cMessage {

   friend class cDispUpdtList;
   
   private:
      static   TTimeSim   fDispUpdtWaitTime;      // between updates
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgDispUpdt();
      virtual           ~cMsgDispUpdt();
};

class cMsgLowPrEvent: public cMessage {

   friend class cLowPrEventList;
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgLowPrEvent();
      virtual           ~cMsgLowPrEvent();
};

class cMsgLowRateUpdt: public cMessage {

   friend class cLowRateUpdtList;
   
   public:
      virtual   cMessageList*  GetMsgList();
      
                         cMsgLowRateUpdt(TTimeSim time, EBoolean bAbsTime = false);
      virtual           ~cMsgLowRateUpdt();
};
*/

class cDisPduType {
public:
   unsigned char pdu[256];

   cDisPduType(void);


private:

enum { epdukwNone, epdukwEntityState, epdukwFire, epdukwDetonation, epdukwCollision, 
       epdukwServiceRequest, epdukwResupplyOffer, epdukwResupplyReceived, epdukwResupplyCancel,
       epdukwRepairComplete, epdukwRepairResponse, epdukwCreateEntity, epdukwRemoveEntity,
       epdukwStartResume, epdukwStopFreeze, epdukwAcknowledge, epdukwActionRequest, 
       epdukwActionResponse, epdukwDataQuery, epdukwSetData, epdukwData, epdukwEventReport, 
       epdukwMessage, epdukwEmission, epdukwLaser, epdukwTransmitter, epdukwSignal, epdukwReceiver, 
       epdukwCpSequencer, epdukwCpNACK, epdukwExpEnvironment, epdukwCpMinefield, epdukwCpMines, 
       epdukwMinefieldRequest, epdukwMinefieldResponse, epdukwMinefieldState, epdukwMinefieldNACK,
       epdukwViewControl, epdukwSizeTable };

static const TKeywordEntry fKeyPduType[epdukwSizeTable];

};

#endif
