
#define MULTIADDR 224.002.140.183

#include "NetMgr.h"
#include "DemoMgr.h"
#include "DisPdu203.h"

#ifdef _WIN32
#include <winsock.h>
#include <sys/types.h>
#else
#include <net/raw.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <unistd.h>
#include <arpa/inet.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <iostream.h>
#include <stdlib.h>

int      gPduLen  = 0;
static int      numberof =0;

cPort::cPort() {}
cPort::~cPort() {}


int    cUdpPort::OpenPort(int udpPort)
{
   int on = 1;
   u_char off = (u_char) 0;
   int cc;
   char* pMultireq;
   int status;
   sport = rport = udpPort;

#ifdef _WIN32      // ZEYNEP
   
   WSADATA Data;
   status=WSAStartup(MAKEWORD(1,1),&Data);
   if(status!=0)
     cerr << "ERROR : WSAStartup "<< endl;
   char hostname[100];
   gethostname(hostname,sizeof(hostname));
   //cout << "HOSTNAME "<<hostname<<endl;
   HOSTENT *hoststr = gethostbyname(hostname);
   hostid = (unsigned long)hoststr->h_addr;
   //cout << "HOST IP "<<hostid<<endl;

#else
   hostid = gethostid();
#endif   
      // SET UP THE RECEIVING SOCKET
   cc = 1024 * ebufSize;
   r  = socket(AF_INET, SOCK_DGRAM, 0);

#ifdef _WIN32 // ZEYNEP

   status=setsockopt(r, SOL_SOCKET, SO_BROADCAST, (const char*)&on, sizeof (on));
   cout << status <<endl;
   status=setsockopt(r, SOL_SOCKET, SO_RCVBUF, (const char*)&cc, sizeof(cc));
   cout << status << endl;
#else
   Assert ( !setsockopt(r, SOL_SOCKET, SO_BROADCAST, &on, sizeof (on)), kerrNoRecvBcast, "OpenPort");
   Assert ( !setsockopt(r, SOL_SOCKET, SO_RCVBUF, &cc, sizeof(cc)), kerrNoBuff, "OpenPort");
#endif
   
   //memset(&rin,0,sizeof(rin));
   rin.sin_family      = AF_INET;
   rin.sin_addr.s_addr = htonl(INADDR_ANY);
   rin.sin_port        = htons(udpPort);
			    
   //Bind the recieving socket
#ifndef _WIN32   
   Assert ( !bind(r, (sockaddr *)&rin, sizeof(sockaddr_in)), kerrBindSocket, "OpenPort", errno);
#else
   status=bind(r, (sockaddr *)&rin, sizeof(sockaddr_in));
   cout << status <<endl;
#endif                            

#ifdef _WIN32   // ZEYNEP
  status=ioctlsocket(r, FIONBIO, (unsigned long *)&on);
  cout << status << endl;

#else 
   Assert ( !ioctl(r, FIONBIO, &on), kerrNoBlock, "OpenPort", errno);
#endif
   

   //Check if the user specified multicast
   if ((pMultireq = (char *)gDemoMgr.fStartupParam.fMulticastAddr) != NULL)
   {
      //Check we have a valid multicast address
#ifdef _WIN32   // ZEYNEP
#else
      Assert(IN_MULTICAST(inet_addr (pMultireq)),kerrInvalidMcast,"OpenPort ",errno);
#endif
      //Get the TTL from command line input
      sMttl = (u_char) gDemoMgr.fStartupParam.fTtl;
      
      sMreq.imr_multiaddr.s_addr = inet_addr (pMultireq);
      sMreq.imr_interface.s_addr = INADDR_ANY;
      
      #ifndef _WIN32
      Assert(!setsockopt(r, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char*)&sMreq, sizeof(sMreq)), kerrNoSubscribeMcast, "OpenPort (recv socket)", errno);
      
      // Disable Loopback
      Assert(!setsockopt(r, IPPROTO_IP, IP_MULTICAST_LOOP, (char *) &off, sizeof (off)), kerrNoDisableLoop, "OpenPort (recv socket)", errno);
      #else
	status=setsockopt(r, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char*)&sMreq, sizeof(sMreq));
      cout <<"STATUS "<<status<<endl;
	// Disable Loopback
	//status=setsockopt(r, IPPROTO_IP, IP_MULTICAST_LOOP, (char *) &off, sizeof (off));
      cout <<"STATUS " << status <<endl;
      #endif   
   }
   return 0;

}

int    cUdpPort::ClosePort()
{
#ifdef _WIN32    // ZEYNEP
//   closesocket(s);
   closesocket(r);
#else
//   close(s);
   close(r);
#endif
   return 0;
}

ENetReturn    cUdpPort::GetPdu(void *buffer, int buflen)
{
   rlen = 0;
   rinlen = sizeof(sockaddr_in);
   errno = 0;
   errcode = enetSelfMsg;
   int myErrno=0;
   
   while ( errcode == enetSelfMsg ) {

#ifndef _WIN32  // ZEYNEP
      rlen = recvfrom(r, buffer, buflen, 0, (sockaddr *)&rin, &rinlen); 
      myErrno = errno;
#else  
      rlen = recvfrom(r, (char *)buffer, buflen, 0, (sockaddr *)&rin, &rinlen);
      if(rlen==SOCKET_ERROR)
      {  myErrno = WSAGetLastError();
//         cout << "myErrno "<< myErrno<<endl;
      }
#endif
     
      switch (myErrno) {
	 case 0:     
	       if (hostid == rin.sin_addr.s_addr) {
		  errcode = enetSelfMsg;
		  cout<<" from self"<<endl;
		  cout << "HOSTID "<<hostid<<" adr "<<rin.sin_addr.s_addr<<endl;
	       }
	       else {
		  errcode = enetMsgRecv;   //success, msg read
		  numberof++;
		  //cout<<" from elsewhere "<< numberof<<endl;
		  //cout << "HOSTID "<<hostid<<" adr "<<rin.sin_addr.s_addr<<endl;
	       }
	       break;
	 #ifndef _WIN32                  
	 case EWOULDBLOCK:   // nothing to read
	 #else
	 case WSAEWOULDBLOCK:
	 #endif;
	       errcode = enetNoData;
  //             cout << "NO DATA"<<endl;
	       break;
			   
	 default:
	       errcode = enetErr;
	       break;
      }
   }
   
   if ( (errcode != enetNoData) && (rlen < 0) ) {
		errcode = enetErr;
   }
#ifndef _WIN32   
   Assert( errcode != enetErr, kerrRecvError, "cUdpPort::GetPdu", errno);
#endif
   return errcode;
}

cUdpPort::cUdpPort(int udpPort)
{
   OpenPort(udpPort);
}

cUdpPort::~cUdpPort()
{
   ClosePort();
}


cPduTranslator::cPduTranslator( int bufSize )
{
   fpOutBuf = new long[ bufSize/sizeof(long) + 1 ];
   fLength = 0;
   fProtocolFamily = 0;
}


EBoolean         cNetMgr::RecvPdu()
{
	 // check against last quit time, to warn about delays
   
   ENetReturn   getResultCode;      
   EBoolean     recv = false;
   
   TTimeSim   start = GetCurSimTime();
   EBoolean bBreak;
   
   do  {
      bBreak = false;

      getResultCode = fpSimPort->GetPdu(fpBuffer, ePduBufSize);
      if (getResultCode == enetMsgRecv) {
	 //if ( gfNetif ) cout << "enetMsgRecv " << endl;
	 fpPduTranslator->ReceivePdu(fpBuffer);
	 recv = true;
      }
      else if ( fpCmsPort && (getResultCode == enetNoData) ) {
	 if ( gfNetif ) cout << "fpCmsPort&enetNoData " << endl;
	 getResultCode = fpCmsPort->GetPdu(fpBuffer, ePduBufSize);
	 if (getResultCode == enetMsgRecv) {
	   if ( gfNetif ) cout << "fpCmsPort&enetNoData- enetMsgRecv " << endl;
	    //fpCmsTranslator->ReceivePdu(fpBuffer);
	    recv = true;
	 }
      }
     
      if ( GetCurSimTime() - start >= 50 ) {
	 //cout << "breaking net to check sim" << endl;
	 bBreak = true;
      }
   } while ( (getResultCode == enetMsgRecv) && (bBreak == false) );
   
   
   return recv;
}


int    cNetMgr::GetExerciseId()
{ 
   return fExerciseId;
}
/*

EBoolean     cNetMgr::UsesCp() 
{
   return fpPduTranslator->UsesCp();
}

EBoolean     cNetMgr::IncludeThermal()  
{ 
   return fpPduTranslator->IncludeThermal(); 
}

EBoolean     cNetMgr::IncludeBurial()  
{ 
   return fpPduTranslator->IncludeBurial(); 
}

EBoolean     cNetMgr::IncludeOrient()  
{ 
   return fpPduTranslator->IncludeOrient();
}

EBoolean    cNetMgr::UsesQrp()
{
   return fpPduTranslator->UsesQrp();
}
*/
void    cNetMgr::NetShutdown()
{
   /*if ( fpPduTranslator )
      fpPduTranslator->TranslatorShutdown();
   */
}

cNetMgr::cNetMgr( TStartupParam & startup )
{
   
   //unsigned int   CpBitFlags;      // determines optional fields used
// CpBitFlags = startup.fbOrientation << ? +   
//              startup.fbBurialDepth << ? + 
//              startup.fbTemperature << ?;

   fExerciseId = startup.fExerciseId;
   
   fpBuffer = new long[ ePduBufSize/sizeof(long) + 1 ];
   
   switch (startup.feProtocol) {
   
      // REGION 2. ZEYNEP - UPDATE (HLA)!
      case eHLA:
            fpPduTranslator = new cDisPduTranslator(startup.feQrpProtocol,
                                                    startup.feCpProtocol,
                                                    startup.fbOrientation,
                                                    startup.fbBurialDepth,
                                                    startup.fbTemperature
);

            fpSimPort = new cHLAPort();
         break;
     // END OF REGION 2.

      case eDis203:
	    fpPduTranslator = new cDisPduTranslator(startup.feQrpProtocol, 
						    startup.feCpProtocol,
						    startup.fbOrientation,
						    startup.fbBurialDepth,
						    startup.fbTemperature );
	    
	    fpSimPort = new cUdpPort(startup.fDisPort);
	 break;
   
      default:
	    cerr << "unknown simulation protocol"<<endl;
	    Assert ( 0, kerrProtocolFamily, "cNetMgr contructor");
	 break;
   }
  /* 
      // if this is half duplex, need CMS port
   if ( ! (startup.fbGui && startup.fbHost) ) {
      fpCmsPort = new cUdpPort(startup.fCmsPort);
      //fpCmsTranslator = new cCmsTranslator();
   }
   else {
      fpCmsPort = NULL;
      //fpCmsTranslator = NULL;
   }
  */ 
}


cNetMgr::~cNetMgr()
{
   if (fpSimPort) delete fpSimPort;
   if (fpCmsPort) delete fpCmsPort;
   //if (fpPduTranslator) delete fpPduTranslator;
   //if (fpCmsTranslator) delete fpCmsTranslator;
}

// REGION 3. ZEYNEP - UPDATE (HLA)!
int    cHLAPort::OpenPort()
{  try
      {
         cout << "FED_HW: CREATING FEDERATION EXECUTION" << endl;
         rtiAmb.createFederationExecution( FederationName, FederationName
);
         cout << "FED_HW: SUCCESSFUL CREATE FEDERATION EXECUTION" << endl;
      }
      catch ( RTI::FederationExecutionAlreadyExists& e )
      {
         cerr << "FED_HW: Note: Federation execution already exists." << &e
<< endl;
      }
      catch ( RTI::RTIinternalError& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
      }
      catch ( RTI::Exception& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
      }
    RTI::Boolean Joined    = RTI::RTI_FALSE;
    int          numTries  = 0;

     while( !Joined && (numTries++ < 20) )
      {
         try
         {
            cout << "FED_HW: JOINING FEDERATION EXECUTION: "  << endl;
            federateId = rtiAmb.joinFederationExecution(
                                    (char* const) federateName,
                                                  FederationName,
                                                  &fedAmb);
            Joined = RTI::RTI_TRUE;
         }
         catch (RTI::FederateAlreadyExecutionMember& e)
         {
            cerr << "FED_HW: ERROR: " << federateName
                 << " already exists in the Federation Execution "
                 << FederationName << "." << endl;
            cerr << &e << endl;
            return -1;
         }
         catch (RTI::FederationExecutionDoesNotExist& e)
         {
            cerr << "FED_HW: ERROR: " << FederationName
                 << " Federation Execution "
                 << "does not exists."<< endl;
#ifdef _WIN32
            Sleep(2000);
#else
            sleep (2); // sleep 2 seconds ::Sleep(2000) on Windows
#endif
         }
         catch ( RTI::Exception& e )
         {
            cerr << "FED_HW: ERROR:" << &e << endl;
         }
      } // end of while

      cout << "FED_HW: JOINED SUCCESSFULLY: "
           << ": Federate Handle = " << federateId << endl;

   return 0;
}
int    cHLAPort::ClosePort()
{
   try
      {
         cout << "FED_HW: RESIGN FEDERATION EXECUTION CALLED" << endl;
         rtiAmb.resignFederationExecution(
                   RTI::DELETE_OBJECTS_AND_RELEASE_ATTRIBUTES );
         cout << "FED_HW: SUCCESSFUL RESIGN FEDERATION EXECUTION CALLED" <<
endl;
      }
      catch ( RTI::Exception& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
         return -1;
      }

   try
      {
         cout << "FED_HW: DESTROY FEDERATION EXECUTION CALLED" << endl;
         rtiAmb.destroyFederationExecution( FederationName );
         cout << "FED_HW: SUCCESSFUL DESTROY FEDERATION EXECUTION CALLED"
<< endl;
      }
      catch ( RTI::Exception& e )
      {
         cerr << "FED_HW: ERROR:" << &e << endl;
         return -1;
      }

   return 0;
}

ENetReturn    cHLAPort::GetPdu(void *buffer, int buflen)
{ 
   rtiAmb.tick();
/*
#ifdef WIN32
            Sleep(1000);
#else
            sleep (1); // sleep 2 seconds ::Sleep(2000) on Windows
#endif
*/
   return enetNoData;
}

int    cHLAPort::PutPdu(void *buffer, int buflen)
{ cout << "int    cHLAPort::PutPdu(void *buffer, int buflen)"<< endl;
  int ret = 0;
  RTI::ParameterHandleValuePairSet* pParams = NULL;
  pParams = RTI::ParameterSetFactory::create( 1 );

  pParams->add(disParamTypeId,
              (char*) buffer,
              ((buflen+1)*sizeof(char)));
  try
  {
    rtiAmb.sendInteraction(disInterClassTypeId, *pParams, NULL);
  }
  catch ( RTI::Exception& e )
  {
     cerr << "FED_HW: Error:" << &e << endl;
     return -1;
}
  delete pParams;
  RTIfedTime requestTime(1);
  requestTime += grantTime;
  try {
     rtiAmb.timeAdvanceRequest(requestTime);
  }
  catch(RTI::Exception& e)
  { cerr << "FED_HW: Error:" << &e << endl;
  }
  cout << "end of int    cHLAPort::PutPdu(void *buffer, int buflen)"<<
endl;
  return ret;
}
cHLAPort::cHLAPort()
{ 
   unsigned int temp  = (unsigned int)(kSimTimeToDis * GetCurSimTime ()) <<
1;
   
     strcpy(FederationName, "HelloWorld");
     strcpy(disInterClassName,"Communication");
     strcpy(disParamName,"Message ");
   sprintf(federateName,"%s_%d","DISHLAPub",temp);
   cout << "FEDERATENAME : "<< federateName << " !!!!!!" << endl;
   timeAdvGrant = RTI::RTI_FALSE;
   OpenPort();
   PublishAndSubscribe();
   decideTiming();
}
cHLAPort::~cHLAPort()
{
   ClosePort();
}

int cHLAPort::PublishAndSubscribe()
{ int ret = 0;
  try {
    disInterClassTypeId = rtiAmb.getInteractionClassHandle(disInterClassName);
    fedAmb.SetdisInterClassTypeId(disInterClassTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    disParamTypeId = rtiAmb.getParameterHandle(disParamName,
                                               disInterClassTypeId);
    fedAmb.SetdisParamTypeId(disParamTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.publishInteractionClass(disInterClassTypeId);
}
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }
    try {
    rtiAmb.subscribeInteractionClass(disInterClassTypeId);
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  return ret;
}

int cHLAPort::decideTiming()
{ int ret = 0;
  EBoolean doTick = false;

  try {
    rtiAmb.disableTimeConstrained();
    doTick = true;
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.disableTimeRegulation();
    doTick = true;
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  try {
    rtiAmb.enableAsynchronousDelivery();
    doTick = true;
  }
  catch ( RTI::Exception& e )
  {
    cerr << "FED_HW: ERROR:" << &e << endl;
    ret = -1;
  }

  return ret;
}
//END OF REGION 3.

