#ifndef _UNetMgr_
#define _UNetMgr_

#include "Globals.h"

#ifdef _WIN32
#include <winsock.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif

//REGION 6. ZEYNEP - UPDATE (HLA)!
#include "HwFederateAmbassador.h"
//END OF REGION 6.

enum ENetReturn { enetMsgRecv = 0, enetMsgSent, enetNoData, enetErr, enetSelfMsg };
enum EbufKb { ebufSize = 512 };

extern  int      gPduLen;

class cPort {
      // base class for either UDP or Assoc net port

public:
   virtual   ENetReturn      GetPdu(void *buffer, int buflen) = 0;

   cPort();
   virtual   ~cPort();
};

   // udp interface
class cUdpPort : public   cPort {

protected:
#ifdef _WIN32
   ULONG         hostid;                   // ip addr of host
#else
   ulong         hostid;                   // ip addr of host
#endif
   int           sport;          // send port number
   int           s;                     // send socket
   sockaddr_in   sin;            // send sockaddr_in
   int           slen;                     // # bytes sent
   struct ip_mreq  sMreq;                   // multicast address
   u_char        sMttl;                     // multicast time-to-live
   int           rport;                    // recv port number
   int           r;                           // recv socket
   sockaddr_in   rin;            // recv sockaddr_in
   int           rinlen;                   // length of addr recv'd from
   int           rlen;                     // # bytes recvd
   ENetReturn     errcode;                 // err code set during opens/reads/or writes
   
   int            OpenPort(int udpPort);
   int            ClosePort();
   
public:
   virtual ENetReturn    GetPdu(void *buffer, int buflen);

	      cUdpPort(int udpPort);
   virtual   ~cUdpPort();
};


class cPduTranslator {

protected:
      void *      fpOutBuf;
      int         fLength;
      char        fProtocolFamily; 

public:
      virtual   void        ReceivePdu(void* buf, TTimeSim timeDelta = 0) = 0;
      
		void        SetLength( int len ) { fLength = len; }
		void        SetProtocolFamily(char pf) {fProtocolFamily = pf; }

      cPduTranslator( int bufSize = 1500 );
};

class cDisPduTranslator;
/*
class cCmsTranslator;
*/

class cNetMgr {

enum EBufSize { ePduBufSize = 1536 };

      cPort*            fpSimPort;
      cPort*            fpCmsPort;
      cDisPduTranslator*   fpPduTranslator;
      //cCmsTranslator*   fpCmsTranslator;
      void*             fpBuffer;
      int               fPduLength;
      
      int               fExerciseId;

public:
      EBoolean          RecvPdu();      // true if one or more processed
	    
      //cDisPduTranslator*   GetPduTrans() { return fpPduTranslator; }      
      int               GetExerciseId();
     /* 
      EBoolean          UsesCp();
      EBoolean          IncludeThermal();
      EBoolean          IncludeBurial();
      EBoolean          IncludeOrient();
      EBoolean          UsesQrp();
     */ 
      void              NetShutdown();
   
			cNetMgr( TStartupParam & startup );
		       ~cNetMgr();
};

// REGION 1. ZEYNEP - UPDATE (HLA)!

class cHLAPort : public   cPort {

protected:
   RTI::RTIambassador rtiAmb;
   HwFederateAmbassador fedAmb;
   RTI::FederateHandle federateId;

//   char* const FederationName = "HelloWorld";
//   char* const disInterClassName = "Communication";
//   char* const disParamName = "Message ";
 
     char FederationName[20];
     char disInterClassName[20];
     char disParamName[20];
 
   char federateName[200];

   RTI::Boolean        timeAdvGrant;
   RTIfedTime          grantTime;


   short disInterClassTypeId;
   int disParamTypeId;

   ENetReturn     errcode;

   int            OpenPort();
   int            ClosePort();
  
public:
   virtual ENetReturn    GetPdu(void *buffer, int buflen);
   virtual int           PutPdu(void *buffer, int buflen);
   cHLAPort();
   virtual   ~cHLAPort();
   int PublishAndSubscribe();
   int decideTiming();

};

//END OF REGION 1.

#endif
