/*
 * $RCSfile$ $Revision$ $State$
 */
/****************************************************************************
*   File: stdtypes.h                                                        *
*                                                                           *
*       Copyright 1990 by BBN Systems and Technology Corporation            *
*                                                                           *
*               BBN Systems and Technology Corporation                      *
*               10 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed under U.S. Government                   *
*       contract MDA903-86-C-0309.  Use, duplication and                    *
*       distribution of this software is subject to the                     *
*       provisions of DFARS 52.227-7013                                     *
*                                                                           *
*       Contents: Common "machine independent" type definitions             *
*       Created: Fri May 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks: Use #ifdefs to customize for new hardware.                 *
*                                                                           *
****************************************************************************/

#ifndef _STDTYPES_H_INCLUDED
#define _STDTYPES_H_INCLUDED

/* Machine dependent native (useful for mixed K&R and ANSI parameters) 
   "and fastest" integer to use.
 */
typedef int FAST_INT;               /* here for backward compatibility */
typedef int NATIVE_INT;
typedef unsigned int NATIVE_UINT;

/* Machine dependent "fastest" floating point to use.
 */
typedef double FAST_REAL;

/* Generic address.  void *, caddr_t, and other such definitions are
 * not fully portable.  It is best to use void * if the compiler supports
 * it.
 */
#if defined(__STDC__) || (defined(mips) && !defined(sony)) || defined(__alpha) || defined(sun5)
typedef void * ADDRESS;
#else
typedef char * ADDRESS;
#endif

/* Generic numeric types.
 * U means unsigned, the trailing digit is bit length.
 */

#if defined(__STDC__) || defined(sgi) || defined(AIXV3)
typedef signed char    int8;
#else
typedef char           int8;
#endif

typedef short          int16;

#ifdef __alpha
typedef int            int32;
#else
typedef long           int32;
#endif

typedef unsigned char  uint8;

typedef unsigned short uint16;

#ifdef __alpha
typedef unsigned       uint32;
#else
typedef unsigned long  uint32;
#endif

typedef float          float32;

typedef double         float64;

/* The following are vector definitions */
typedef float32 vec3d32[3];
typedef float64 vec3d64[3];

typedef float32 vec2d32[2];
typedef float64 vec2d64[2];

typedef float32 rmatrix32[3][3];
typedef float64 rmatrix64[3][3];

typedef float32 rtmatrix32[4][3];
typedef float64 rtmatrix64[4][3];

typedef struct {
    vec3d64     vec;   /* Placed before single-precision to assure alignment */
    rmatrix32   rot;
} mtx32vec64;

/* Thirty-two and fourty-eight bit identifiers, as used in network protocols
 * H == High-order short
 * M == Medium-order short
 * L == Low-order short
 */
typedef struct {uint16 h; uint16 l;} NETID32, *NETID32_PTR;
typedef struct {uint16 h; uint16 m; uint16 l;} NETID48, *NETID48_PTR;

/* We do the following to get a type for argv that will work on all machines.
   This is really required because the DEC Alpha cannot make the argv pointer
   be 32 bits wide. */
#ifdef __alpha
#pragma pointer_size save
#pragma pointer_size long
typedef char **argv_t;
#pragma pointer_size restore
#else
typedef char **argv_t;
#endif

#endif
