import RTIcap.*;
import owrti.util.CDRCoderEncoder;

/**
*
* Simulated Chatter object. This object knows the state of Chatter Object
* and it can produce the necessary state update messages for required
* attributes to the RTI so that other interested federates can receive 
* these updates.
*
* @date May 11, 1998 ( Updated May 11, 1998 )
* @author H. Timucin Ozdemir
*
*/
public class SimChatBudy extends ChatBudy {
 
   double grRate = 0.5;
   RTIcap.RTIambassador rtiAmb;
   ChatGUI  boss;
   boolean  sendNameUpdateFlag = true;
   boolean  sendLastSayUpdateFlag = true;
   boolean  nameChanged = true;
   boolean  lastSayChanged = true;

   /**
   */
   public 
   SimChatBudy( RTIcap.RTIambassador _rtiAmb, ChatGUI _father, int _ObjectID) 
   { 
      rtiAmb = _rtiAmb;
      setObjectID(_ObjectID);
      boss = _father;
      resetNameChanged(); 
      resetLastSayChanged();

   }// end of CONSTRUCTOR

   /**
   */
   public 
   SimChatBudy( RTIcap.RTIambassador _rtiAmb, ChatGUI _father, String _name, 
               String _lastSay , int _ObjectID)
   {
      rtiAmb = _rtiAmb;
      setName(_name);
      setLastSay(_lastSay);
      setObjectID(_ObjectID);
      boss = _father;
      setNameChanged(); 
      setLastSayChanged();

   }// end of CONSTRUCTOR

   /**
   *
   * After each time adavancement in the Main simulation loop,
   * this method will be invoked. So that object can update its state,
   * and broadcast its updates if it is necessary.
   *
   */
   public final void
   UpdateTime( double time )
   {
      double diff = time - currentTime;

      if( diff > 0 )
      {
         setCurrentTime(time);
       
         /*******************************/
         /** Set the population changed */
         /*******************************/
         setLastSayChanged();
         UpdateLastSay(getLastSay());

         UpdateName( getName() );

      }// end of if

   }// end of Update()

   /**
   */
   public final void
   UpdateName( String cntName )
   {

      setName(cntName);

      System.out.println(" rtiAmb: "+(rtiAmb != null?" not ":"")+"null");
      System.out.println(" sendNameUpdateFlag: "+sendNameUpdateFlag);
      System.out.println(" isNameChanged: "+isNameChanged());

      /********************************************************************/
      /** If we are allowed to send this update and name is changed, then */
      /** we will send this update message to the RTI.                    */
      /********************************************************************/
      if( rtiAmb != null && sendNameUpdateFlag && isNameChanged() )
      {
           
          RTIcap.AttributeHandleValuePair[] eles = new RTIcap.AttributeHandleValuePair[1];
              
          eles[0] = new RTIcap.AttributeHandleValuePair(boss.getChatterObjAttrNameTypeId(), myName.getBytes()); 

          try
          {   
              rtiAmb.updateAttributeValues(ObjectID,eles,"Name");

              /******************************************************/
              /** Since we sent this update, we no longer required  */
              /** to send it again. Unless we are asked to do so    */
              /** ( Look at provideAttributesUpdates )              */ 
              /******************************************************/
              resetNameChanged();

          } catch (RTIcap.ObjectNotKnown onk) {}
            catch (RTIcap.AttributeNotDefined onk) {}
            catch (RTIcap.AttributeNotOwned onk) {}
            catch (RTIcap.FederateNotExecutionMember onk) {}
            catch (RTIcap.SaveInProgress onk) {}
            catch (RTIcap.RestoreInProgress onk) {}
            catch (RTIcap.RTIinternalError onk) {}

      }// end of if

   }// end of UpdateName()

   /**
   */
   public final void
   UpdateLastSay( String _lastSay )
   {
      if( _lastSay!=null && _lastSay.trim().length()>0 && 
          !_lastSay.trim().equals(lastSay) )
      { lastSay = _lastSay ; setLastSayChanged(); }

      if(isNameChanged()) UpdateName(getName());

      /********************************************************************/
      /** If we are allowed to send this update and population is changed,*/ 
      /** then we will send this update message to the RTI.               */
      /********************************************************************/
      if( rtiAmb != null && sendLastSayUpdateFlag && isLastSayChanged() )
      {
           
          RTIcap.AttributeHandleValuePair[] eles = new RTIcap.AttributeHandleValuePair[1];
              
          eles[0] = new RTIcap.AttributeHandleValuePair(
                        boss.getChatterObjAttrLastSayTypeId(), _lastSay.getBytes() ); 
          try
          {   
            rtiAmb.updateAttributeValues(ObjectID,eles,"Say");

            resetLastSayChanged();

          } catch (RTIcap.ObjectNotKnown onk) {}
            catch (RTIcap.AttributeNotDefined onk) {}
            catch (RTIcap.AttributeNotOwned onk) {}
            catch (RTIcap.FederateNotExecutionMember onk) {}
            catch (RTIcap.SaveInProgress onk) {}
            catch (RTIcap.RestoreInProgress onk) {}
            catch (RTIcap.RTIinternalError onk) {}

      }// end of if

   }// end of Update()

   /**
   *
   * <b>turnUpdates[On/Off]ForObjectInstance</b> in HwFederateAmbassador uses 
   * this method to tell this object to start sending attribute updates.
   *
   */
   public final void
   UpdateControls( boolean status, short[] theAttributes )
   {
      for(int i=0;i<theAttributes.length;i++)
         if( theAttributes[i] == boss.getChatterObjAttrLastSayTypeId() )
                sendLastSayUpdateFlag = status;
         else if( theAttributes[i] == boss.getChatterObjAttrNameTypeId() )
                sendNameUpdateFlag = status;

   }// end of UpdateControls()

   public final void setNameChanged()   { nameChanged = true; }
   public final void resetNameChanged() { nameChanged = false; }
   public final boolean isNameChanged() { return nameChanged ; }

   public final void setLastSayChanged()   { lastSayChanged = true; }
   public final void resetLastSayChanged() { lastSayChanged = false; }
   public final boolean isLastSayChanged() { return lastSayChanged ; }

}// end of SimChatBudy

