import RTIcap.*;
import owrti.util.CDRCoderEncoder;


public class
SimNode extends Node
{
 
   double grRate = 0.5;
   RTIcap.RTIambassador rtiAmb;
   Server   boss;
   boolean  sendNameUpdateFlag = true;
   boolean  sendPopuUpdateFlag = true;
   boolean  nameChanged = true;
   boolean  popuChanged = true;
   short interactionId = -1;
   int role;

   public static final int SLAVE     = 0;
   public static final int CENTRAL   = SLAVE+1;
   /**
   */
   public 
   SimNode( RTIcap.RTIambassador _rtiAmb, Server _father, int _ObjectID) 
   { 
      rtiAmb = _rtiAmb;
      setObjectID(_ObjectID);
      boss = _father;
      resetNameChanged(); resetPopuChanged();

   }// end of CONSTRUCTOR

   /**
   */
   public 
   SimNode( RTIcap.RTIambassador _rtiAmb, Server _father, String _name,double _initialID, int _ObjectID,short centralCommTypeId,int _role)
   {
      rtiAmb = _rtiAmb;
      setName(_name);
      setId(_initialID);
      setObjectID(_ObjectID);
	  interactionId = centralCommTypeId;
      boss = _father;
      setNameChanged(); 
	  setPopuChanged();
	  role = _role;

   }// end of CONSTRUCTOR
    


   /**
   *
   * After each time adavancement in the Main simulation loop,
   * this method will be invoked. So that object can update its state,
   * and broadcast its updates if it is necessary.
   *
   */
   public final void
   UpdateTime( double time )
   {
      double diff = time - currentTime;

     // if( diff > 0 )
     // {
         setCurrentTime(time);
       
         /*******************************/
         /** Set the IDchanged */
         /*******************************/
         setPopuChanged();

         UpdateName( getName() );
		 UpdateId(myID);//+diff*grRate);

        

    

   }// end of Update()

   /**
   */
   public final void
   UpdateName( String cntName )
   {

      setName(cntName);
	  short a =5;

      /********************************************************************/
      /** If we are allowed to send this update and name is changed, then */
      /** we will send this update message to the RTI.                    */
      /********************************************************************/
      if( rtiAmb != null && sendNameUpdateFlag && isNameChanged() )
      {
           
          RTIcap.AttributeHandleValuePair[] eles = new RTIcap.AttributeHandleValuePair[1];
                
              
          eles[0] = new RTIcap.AttributeHandleValuePair(boss.getNameAttributeId(), 
                                                     myCountryName.getBytes()); 

		  
          try
          {   
              rtiAmb.updateAttributeValues(ObjectID,eles,"name");
			 

              /******************************************************/
              /** Since we sent this update, we no longer required  */
              /** to send it again. Unless we are asked to do so    */
              /** ( Look at provideAttributesUpdates )              */ 
              /******************************************************/
            //  resetNameChanged();

          } catch (RTIcap.ObjectNotKnown onk) {}
            catch (RTIcap.AttributeNotDefined onk) {}
            catch (RTIcap.AttributeNotOwned onk) {}
            catch (RTIcap.FederateNotExecutionMember onk) {}
            catch (RTIcap.SaveInProgress onk) {}
            catch (RTIcap.RestoreInProgress onk) {}
            catch (RTIcap.RTIinternalError onk) {} 
			



      }// end of if

   }// end of Update()

   /**
   */
   public final void
   UpdateId( double newPop)
   {
      myID= newPop ;
	  short a =5;


      /********************************************************************/
      /** If we are allowed to send this update and IDis changed,*/ 
      /** then we will send this update message to the RTI.               */
      /********************************************************************/
      if( rtiAmb != null && sendPopuUpdateFlag && isPopuChanged() )
      {
           
          RTIcap.AttributeHandleValuePair[] eles = new RTIcap.AttributeHandleValuePair[1];


              
          byte[] values = CDRCoderEncoder.toByte(myID);
          eles[0] = new RTIcap.AttributeHandleValuePair(
                        boss.getPopuAttributeId(),
                        values ); 
           try
          {   
            rtiAmb.updateAttributeValues(ObjectID,eles,"Id");

            resetPopuChanged();

          } catch (RTIcap.ObjectNotKnown onk) {}
            catch (RTIcap.AttributeNotDefined onk) {}
            catch (RTIcap.AttributeNotOwned onk) {}
            catch (RTIcap.FederateNotExecutionMember onk) {}
            catch (RTIcap.SaveInProgress onk) {}
            catch (RTIcap.RestoreInProgress onk) {}
            catch (RTIcap.RTIinternalError onk) {}

	if(role==CENTRAL)
	{
		  RTIcap.ParameterHandleValuePair[] eles1 = new RTIcap.ParameterHandleValuePair[1];
                
          eles1[0] = new RTIcap.ParameterHandleValuePair(boss.getCentralMsgTypeId(),values); 
			System.out.println("Central Sending Alive ");
			try{
			rtiAmb.sendInteraction((short)interactionId,eles1,"Central");
			}catch(RTIcap.InteractionClassNotDefined onk){}
			catch(RTIcap.InteractionClassNotPublished onk){}
			catch(RTIcap.InteractionParameterNotDefined onk){}
			catch(RTIcap.FederateNotExecutionMember onk){}
			catch (RTIcap.SaveInProgress onk) {}
            catch (RTIcap.RestoreInProgress onk) {}
            catch (RTIcap.RTIinternalError onk) {} 
	}

      }// end of if

   }// end of Update()

   /**
   *
   * <b>turnUpdates[On/Off]ForObjectInstance</b> in HwFederateAmbassador uses 
   * this method to tell this object to start sending attribute updates.
   *
   */
   public final void
   UpdateControls( boolean status, short[] theAttributes )
   {
      for(int i=0;i<theAttributes.length;i++)
         if( theAttributes[i] == boss.getPopuAttributeId() )
                sendPopuUpdateFlag = status;
         else if( theAttributes[i] == boss.getNameAttributeId() )
                sendNameUpdateFlag = status;

   }// end of UpdateControls()

   public final void setNameChanged()   { nameChanged = true; }
   public final void resetNameChanged() { nameChanged = false; }
   public final boolean isNameChanged() { return nameChanged ; }

   public final void setPopuChanged()   { popuChanged = true; }
   public final void resetPopuChanged() { popuChanged = false; }
   public final boolean isPopuChanged() { return popuChanged ; }

   public void setRole(int _role)
   {
	   role =_role;
   }

}// end of SimNode

