package edu;
import java.util.*;
import edu.rti.*;
import edu.dmi.*;
import edu.util.ErrorReport;

class RequestBroadcaster extends Thread { 
   Vector  objVect;
   RTIcap.RTIambassador rtiAmb;
   Object  objVectLock = new Object();

   RequestBroadcaster(RTIcap.RTIambassador rtiAmb) { setPriority(Thread.NORM_PRIORITY-1); this.rtiAmb=rtiAmb; start(); }
   
   public synchronized void run() {
     objVect = new Vector();
     try
     { while(true) {
         while(objVect.size()==0) wait();
         int objId = -1;
         try{ sleep(100); } catch(InterruptedException ie) {}
         synchronized(objVectLock) { Integer a = (Integer) objVect.elementAt(0); objVect.removeElementAt(0); objId=a.intValue(); }; 
         if(objId>-1 && rtiAmb!=null) 
           try
           { rtiAmb.requestObjectAllAttributeValueUpdate(objId);
           } catch( RTIcap.ObjectNotKnown sip) {}
             catch( RTIcap.SaveInProgress sip) {}
             catch( RTIcap.RestoreInProgress rip) {}
             catch( RTIcap.FederateNotExecutionMember icnd) {}
             catch( RTIcap.RTIinternalError ie) {} 
       }// end of while()
     } catch(InterruptedException ex) {}
   }// end of run()

   public synchronized void addRequest(int objId)
   { synchronized(objVectLock) { objVect.addElement(new Integer(objId)); } notify(); }

}// end of RequestBroadcaster

/**
* @author Hasan Timucin Ozdemir
* @date   March 16, 1999 (Updated on March 16, 1999)
*/
public class EntityManager extends UpdateBroadcaster implements UpdateListener, RTIObjectsContainer, ElementsContainer { 
  RTIcap.RTIambassador rtiAmb=null;
  RequestBroadcaster rb;

  /**
  * @param RTIcap.RTIambassador
  */
  public EntityManager(RTIcap.RTIambassador rtiAmb) { 
     this.rtiAmb=rtiAmb; rb = new RequestBroadcaster(rtiAmb);
  }

  Hashtable objIDToElement   = new Hashtable();
  Hashtable objIDToClassId   = new Hashtable();
  Object    guard1  = new Object();
  int       localId = -2;

  /**
  * @return RTIObject
  */
  public RTIObject getObjectInstance(int objId) 
  { return (RTIObject) objIDToElement.get(new Integer(objId));}

  /**
  * @param Object identifier.
  */
  public void removeObjectInstance(int objId) { 
    synchronized(guard1) {objIDToElement.remove(new Integer(objId)); }
    fireUpdate(new UpdateEvent(this)); 
  }// end of removeObjectInstance()

  /**
  * @param Object identifier.
  * @param Class identifier.
  */
  public void   addObjectInstance(int objId, short classId) { 
    ObjFactory oFac = getObjectFactoryForRtiObjectClass(RTILookUpTable.getObjectClassName(classId));
    try 
    { if( oFac!=null) { Object o = oFac.getNewInstance();
       if( o instanceof RTIObject) { 
         ((RTIObject)o).setInstanceRTIid(objId);
         ((RTIObject)o).setObjectClassRTIid(classId); 
         addElement((RTIObject)o);
         if(o instanceof Element) ((Element)o).setUpdateListener(this);
         if(rb!=null) rb.addRequest(objId);
       }// end of if
      }// end of if
    } catch(ClassNotFoundException e1) {
       ErrorReport.displayErrorMessage(e1.getMessage()); 
      } catch(ObjectClassNameUnknown e2) {
         ErrorReport.displayErrorMessage(e2.getMessage());  }
  }// end of addObjectInstance()

  /**
  * @param RTIObject.
  */
  public void   addElement(RTIObject ele) { 
    try
    { if(ele.getInstanceRTIid()==-1) assignRTIid(ele);
      if(getObjectInstance(ele.getInstanceRTIid())==null) { 
        synchronized (guard1) {objIDToElement.put(new Integer(ele.getInstanceRTIid()),ele); }
        fireUpdate(new UpdateEvent(ele));
      }// end  of if 
    } catch(Exception e) { e.printStackTrace(); }
  }// end of addElement()

  public void   removeElement(RTIObject ele) 
  { if(getObjectInstance(ele.getInstanceRTIid())!=null) removeObjectInstance(ele.getInstanceRTIid()); }
  
  public Enumeration   getElements() { return objIDToElement.elements(); }

  /**
  * @param Element.
  */
  public void   addElement(Element ele) {
    if(ele!=null ) { 
       ele.setUpdateListener(this); 
       try
       { if( ele instanceof RTIObject ) { 
           addElement((RTIObject)ele); ((RTIObject)ele).setPublish(); }
       } catch( Exception e) { 
          //e.printStackTrace();  
          synchronized (guard1) 
          { objIDToElement.put(new Integer(localId--),ele); }
          fireUpdate(new UpdateEvent(ele));
         }// end of catch
    }// end of if
  }// end of addElement()

  public void   removeElement(Element ele) 
  { if(ele!=null && ele instanceof RTIObject ) removeObjectInstance(((RTIObject)ele).getInstanceRTIid()); }
  
  public void updateOccurred(UpdateEvent evt) { fireUpdate(evt.getSource()); }

  Hashtable objNameToFactory = new Hashtable();
  Object    objNameToFactoryLock = new Object();

  /**
  * @param rtiClassName.
  * @param ObjFactory 
  */
  public void registerObjectFactory(String rtiClassName, ObjFactory oFactory) {  
   synchronized(objNameToFactoryLock) { objNameToFactory.put(rtiClassName,oFactory); }
  }// end of registerObjectFactory()

  /**
  * @param rtiClassName.
  * @return ObjFactory 
  */
  public ObjFactory getObjectFactoryForRtiObjectClass(String rtiClassName)
                    throws ObjectClassNameUnknown {
    short a = RTILookUpTable.getObjectClassRTIid(rtiClassName);
    return (ObjFactory)objNameToFactory.get(rtiClassName);
  }// end of registerObjectFactory()

  final void assignRTIid(RTIObject ele) throws Exception {
    if( rtiAmb != null ) {
      short objID = ele.getObjectClassRTIid();
      if( objID!=-1 )
      try
      { int objInstanceId = rtiAmb.registerObjectInstance(objID);
        ele.setInstanceRTIid(objInstanceId);
      } catch( RTIcap.ObjectClassNotDefined sip) {}
        catch( RTIcap.ObjectClassNotPublished sip) {}
        catch( RTIcap.SaveInProgress sip) {}
        catch( RTIcap.RestoreInProgress rip) {}
        catch( RTIcap.FederateNotExecutionMember icnd) {}
        catch( RTIcap.RTIinternalError ie) {} 
    }// end of if 
  }// end of assignRTIid

  final void askRTIObjectAttributes(int objID) {
    if( rtiAmb != null ) {
      if( objID!=-1 )
      try
      { rtiAmb.requestObjectAllAttributeValueUpdate(objID);
      } catch( RTIcap.ObjectNotKnown sip) {}
        catch( RTIcap.SaveInProgress sip) {}
        catch( RTIcap.RestoreInProgress rip) {}
        catch( RTIcap.FederateNotExecutionMember icnd) {}
        catch( RTIcap.RTIinternalError ie) {} 
    }// end of if 
  }// end of askRTIObjectAttributes

}// end of EntityManager

