/*
$RCSfile: Vulnerability.h,v $
$Revision
$Author: jchappel $
$Date $
$Log: Vulnerability.h,v $
Revision 1.1.1.1  1999/02/16 14:54:54  jchappel
Comprehensive Mine Simulator

 * Revision 2.1  1998/06/03  17:18:04  jchappel
 * ..
 *
 * Revision 2.0  1998/01/26  21:04:20  ksnively
 * Baseline CMS before additions for AP mine replacements.
 *
 * Revision 1.1  1997/12/16  13:59:57  ksnively
 * Initial revision
 *
 * Revision 1.3  1997/10/27  17:07:30  trang
 * BUILD1
 *
 * Revision 1.2  1997/05/07  12:33:45  ksnively
 * *** empty log message ***
 *
Revision 1.2  1997/05/07 12:33:45  ksnively
*** empty log message ***

 * Revision 1.1  1996/10/23  13:54:40  nsipes
 * Initial revision
 *
 * Revision 1.4  1996/07/03  22:55:15  rcarp
 * enhancements and bug fixes for version 1.5b9
 *
 * Revision 1.3  1996/06/22  13:04:11  rcarp
 * deton vuln bug fixes for 1.5b4
 *
 * Revision 1.2  1996/06/21  17:04:20  rcarp
 * upgrades for release 1.5b3, mostly mine vuln
 *
 * Revision 1.1  1996/06/19  19:05:45  rcarp
 * Initial revision
 *
$State: Exp $
*/
 
/*******************************************

   file: Vulnerability.h
   
      Copyright 1996 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   Vulnerability.h
   
      MIne vulnerability to explosives and countermeasures
   
*********************************************/


#ifndef _UVulnerability_
#define _UVulnerability_

#include "Globals.h"

enum EVulnResult {
   evrNoVuln,
   evrNone,      // but was in range, random no result
   evrDet,
   evrNeut,
   evrDisplace,
   evrRollback
};

struct TVulnProb {
      // stored as cumulative probabilities in given order,
      // but initialized as independent probabilities
   float   fDetProb;
   float   fNeutProb;
   float   fDisplaceProb;
   float   fRollbackProb;      // given a displacement
      
   static  const char*   GetVulnResultText(int result);
   
   TVulnProb( float detPr = 0.0, float neutPr = 0.0, 
              float displacePr = 0.0, float rollbackPr = 0.0 );
};

struct TVulnTable : public T2link {

    enum { etbltypeMech,  etbltypeDet };

  enum { evkwDepths, evkwRanges, evkwProbs, evkwLinLength, evkwOffset, 
          evkwSizeTable };
   static const TKeywordEntry fKeyVulnParam[evkwSizeTable];
   
   char*       fpName;      // e.g., esmbPresFuze, lgblastPresBhFuze
   int         fIndex;      
   float       fLinearXlength;   // for ESMB and MICLIC, length of line
   int         fNumDepth;
   int         fNumRange;
   float*      fpDepths;
   float*      fpRanges;
   T2Dsingle      fOffset;
   TVulnProb*  fpVulnTab;   // 2D depth and range, depths vary fastest
   
//   float            GetLinearLength()   { return fLinearXlength; }
   
   const char*      GetTableName()   { return fpName; }

   EBoolean         ReadVulnTable( cInFileObj& infile, int tblType );
   int		    GetTableIndex()   { return fIndex; }
   
   TVulnTable(T2link* pPrev);
   TVulnTable( T2link* pPrev, char* pName );
//   TVulnTable(T2link* pPrev, int index, int numDepth, int numRange,
//        TVulnProb* pTable, const float* pDepths, const float* pRanges, float length = 0.0);

   virtual  ~TVulnTable();
};

class cThreatList : public T2linkHead {
 
 public:
//   void           ConstructBasicList();
//   void           ReadParamVulnTable( cInFileObj& infile, char* pName );
//   void           ReadParamNewVulnTable( cInFileObj& infile, char* pName );
   
   TVulnTable*   FindTableByName(char* pName);
   virtual TVulnTable*   FindTableByIndex(int index);

                  cThreatList();
   virtual       ~cThreatList();
};


struct TMechVulnTable : public TVulnTable {
      // these tables are linked into a global cMechThreatList list,
      // with individual mines/cMechanVuln parts keeping a TVulnMechParam
      // which has pointers into the list
//   char*       fpName;      // e.g., twmpPresFuze, twmrMagFuze
//   int         fIndex;      // for TVulnMechParam default table types
//   int         fNumDepth;
//   float*      fpDepths;
//   TVulnProb*  fpVulnTab;   // array indexed by depth
//   static const char* fKeyMechVulnTables[evmtblSizeTable];
//   T2Dsingle   fOffset;     // from critical part, e.g. front of track
   
      // careful of coord systems.  relLoc is from mine, but in veh oriented system.
      // which means relLoc.x is negative when mine is on veh leading edge
   EBoolean         IsVehicleEncounter(const TEntCoord& relLoc, const TVehDimensions& vehDim);   // with entire vehicle
   EBoolean         IsTrackEncounter(const TEntCoord& relLoc, const TVehDimensions& vehDim);   // with veh track
   EBoolean         IsBellyEncounter(const TEntCoord& relLoc, const TVehDimensions& vehDim);   // used for dogbone
   EVulnResult      GetRandomResult(float depth,  float effectiveness = 1.00);
   
//   const char*      GetTableName()   { return fpName; }
//         int        GetTableIndex()   { return fIndex; }
   const T2Dsingle& GetTableOffset()  { return fOffset; }
   
   TMechVulnTable(T2link* pPrev, T2Dsingle& offset, char* pName, 
                  int numDepth, TVulnProb* pTable, const float* pDepths);
   TMechVulnTable(T2link* pPrev, T2Dsingle& offset, int index, 
                  int numDepth, TVulnProb* pTable, const float* pDepths);
   virtual  ~TMechVulnTable();
};

enum {
   enumdepthTwmP   = 3,
   enumdepthFwmP   = 3,
   enumdepthTwmR   = 3,
   enumdepthFwmR   = 3,
   enumdepthDb     = 3,
   enumdepthMagCm  = 3
};

class cMechThreatList : public cThreatList {
 
 public:
   void           ConstructBasicList();
   void           ReadParamVulnTable( cInFileObj& infile, char* pName );
   void           ReadParamNewVulnTable( cInFileObj& infile, char* pName );
   
//   TMechVulnTable*   FindTableByName(char* pName);
//   TMechVulnTable*   FindTableByIndex(int index);

                  cMechThreatList();
   virtual       ~cMechThreatList();
};

struct TDetVulnTable : public TVulnTable {

      // default table names are in a character array
   enum {
      evdtblLgBlastPres, evdtblLgBlastBhPres, evdtblLgBlastTilt, evdtblLgBlastAp,  
      evdtblLgBlastCmdAp,  evdtblLgBlastCam, evdtblLgBlastOther,
      evdtblMedBlastPres, evdtblMedBlastBhPres, evdtblMedBlastTilt, evdtblMedBlastAp, 
      evdtblMedBlastCmdAp, evdtblMedBlastCam, evdtblMedBlastOther,
      evdtblSmBlastPres, evdtblSmBlastBhPres, evdtblSmBlastTilt, evdtblSmBlastAp, 
      evdtblSmBlastCmdAp, evdtblSmBlastCam, evdtblSmBlastOther,
      evdtblMiclicPres, evdtblMiclicTilt, evdtblMiclicBhPres, evdtblMiclicAp, 
      evdtblMiclicCmdAp, evdtblMiclicCam, evdtblMiclicOther,
      evdtblEsmbPres, evdtblEsmbBhPres, evdtblEsmbTilt, evdtblEsmbAp, 
      evdtblEsmbCmdAp, evdtblEsmbCam, evdtblEsmbOther,
      evdtblLgFragPres, evdtblLgFragBhPres, evdtblLgFragTilt, evdtblLgFragAp, 
      evdtblLgFragCmdAp, evdtblLgFragCam, evdtblLgFragOther,
      evdtblSmFragPres, evdtblSmFragBhPres, evdtblSmFragTilt, evdtblSmFragAp, 
      evdtblSmFragCmdAp, evdtblSmFragCam, evdtblSmFragOther,
      evdtblOtherPres, evdtblOtherBhPres, evdtblOtherTilt, evdtblOtherAp, 
      evdtblOtherCmdAp, evdtblOtherCam, evdtblOtherOther,

      evdtblSizeTable
   };
   static const char* fVulnDetDefNames[evdtblSizeTable];
      // keywords for reading parameters
   enum { evdkwDepths, evdkwRanges, evdkwProbs, evdkwLinLength,
          evdkwSizeTable };
//   static const TKeywordEntry fKeyVulnDetParam[evdkwSizeTable];
   
//   char*       fpName;      // e.g., esmbPresFuze, lgblastPresBhFuze
//   int         fIndex;      
//   float       fLinearXlength;   // for ESMB and MICLIC, length of line
//   int         fNumDepth;
//   int         fNumRange;
//   float*      fpDepths;
//   float*      fpRanges;
//   TVulnProb*  fpVulnTab;   // 2D depth and range, depths vary fastest
   
   EVulnResult      GetRandomResult(float depth, float range);
   float            GetLinearLength()   { return fLinearXlength; }
   
//   const char*      GetTableName()   { return fpName; }

   EBoolean         ReadDetVulnTable( cInFileObj& infile );
   
   TDetVulnTable( T2link* pPrev, char* pName );
   TDetVulnTable(T2link* pPrev, int index, int numDepth, int numRange,
        TVulnProb* pTable, const float* pDepths, const float* pRanges, float length = 0.0);

   virtual  ~TDetVulnTable();
};

struct TDetVulnPair {
   int              fDetType;
   TDetVulnTable*   fDetVulnTable;
};

struct TDetVulnBlock : public T2link {
   enum { eNumDetVulnPairs = 100 };
   
   int              fNumEntries;
   TDetVulnPair     fDetVulnPair[eNumDetVulnPairs];
   
   TDetVulnTable*   GetVulnDetTable( int detType );
   EBoolean         AddVulnDetEntry(int detType, TDetVulnTable* pDvTable);
      
   TDetVulnBlock(T2link* pPrev, TDetVulnBlock* pdvOrig);
   TDetVulnBlock(T2link* pPrev) : 
         T2link(pPrev) { fNumEntries = 0; }

};

enum {
   enumdepthEsmb   = 3,
   enumdepthMiclic   = 3,
   enumdepthLgBlast   = 3,
   enumdepthMedBlast   = 3,
   enumdepthSmBlast   = 3,
   enumdepthLgFrag   = 3,
   enumdepthSmFrag   = 3,
   enumdepthOtherWh = 3
};

enum {
   enumrangeEsmb   = 1,
   enumrangeMiclic   = 4,
   enumrangeLgBlast   = 4,
   enumrangeMedBlast   = 4,
   enumrangeSmBlast   = 4,
   enumrangeLgFrag   = 4,
   enumrangeSmFrag   = 4,
   enumrangeOtherWh = 4
};

class cDetThreatList : public cThreatList {
 
 public:
   void           ConstructBasicList();
   void           ReadParamVulnTable( cInFileObj& infile, char* pName );
   void           ReadParamNewVulnTable( cInFileObj& infile, char* pName );
   
//   TDetVulnTable*   FindTableByName(char* pName);
   TVulnTable*   FindTableByIndex(int index);

                  cDetThreatList();
   virtual       ~cDetThreatList();
};

#endif
