/*
$RCSfile: Vulnerability.cc,v $
$Revision
$Author: jchappel $
$Date $
$Log: Vulnerability.cc,v $
Revision 1.1.1.1  1999/02/16 14:54:45  jchappel
Comprehensive Mine Simulator

// Revision 2.5  1998/12/02  16:45:33  jchappel
// added directional sensor
//
// Revision 2.4  1998/11/25  18:37:51  ksnively
// Update for APLA Exercise
//
// Revision 2.3  1998/04/20  15:10:36  ksnively
// *** empty log message ***
//
// Revision 2.2  1998/02/26  21:24:37  ksnively
// Created Mf Camera object to control MAK Stealth view.
// Created vulnerability tables specific to command detonated mines
// and the camera object.
//
// Revision 2.1  1998/02/18  17:16:10  ksnively
// Incorporated AP mine replacement code.
//
// Revision 2.0  1998/01/26  21:01:35  ksnively
// Baseline CMS before additions for AP mine replacements.
//
// Revision 1.1  1997/12/16  13:59:57  ksnively
// Initial revision
//
// Revision 1.3  1997/11/14  21:48:42  ksnively
// Added multicast command line options.  Changed memory allocation calls.
//
// Revision 1.2  1997/05/07  12:32:33  ksnively
// *** empty log message ***
//
// Revision 1.1  1996/10/23  13:54:40  nsipes
// Initial revision
//
// Revision 1.5  1996/07/13  19:04:20  rcarp
// version 1.5b13, fixes bug dragging with two buttons and displays damamged vehicles
//
// Revision 1.4  1996/07/03  22:55:15  rcarp
// enhancements and bug fixes for version 1.5b9
//
// Revision 1.3  1996/06/22  13:04:11  rcarp
// deton vuln bug fixes for 1.5b4
//
// Revision 1.2  1996/06/21  17:04:20  rcarp
// upgrades for release 1.5b3, mostly mine vuln
//
// Revision 1.1  1996/06/19  19:05:45  rcarp
// Initial revision
//
$State: Exp $
*/
 
/*******************************************

   file: Vulnerability.cc
   
      Copyright 1996 by the Institute for Defense Analyses
      
         Institute for Defense Analyses
         1801 N. Beauregard Street
         Alexandria, Virginia 22311-1772
         703-845-2000
         
      This software was developed by IDA under U. S. Government
      contract.  For distribution information contact:
      
         U. S. Army Night Vision Directorate
         AMSEL RD NV OD CD
         10221 Burbeck Road, suite 430
         Ft. Belvoir VA 22060-5806
      
*********************************************/

/*******************************************

   Vulnerability.cc
   
      MIne vulnerability to explosives and countermeasures
   
*********************************************/


#include "Vulnerability.h"
#include "EntityDB.h"
#include "Minefield.h"
#include "FalseMine.h"

#include <math.h>

const char* TVulnMechParam::fKeyMechVulnTables[evmtblSizeTable] = {
   "twmpPresFuze" , //ekeyVulnPointList,
   "twmpTiltFuze", //ekeyVulnPointList,
   "twmpMagFuze", //ekeyVulnPointList,
   "twmpApFuze", //ekeyVulnPointList,
   "twmpCmdApFuze", //ekeyVulnPointList,
   "twmpCamera", //ekeyVulnPointList,
   "twmpOtherFuze", //ekeyVulnPointList,
   
   "fwmpPresFuze", //ekeyVulnPointList,
   "fwmpTiltFuze", //ekeyVulnPointList,
   "fwmpMagFuze", //ekeyVulnPointList,
   "fwmpApFuze", //ekeyVulnPointList,
   "fwmpCmdApFuze", //ekeyVulnPointList,
   "fwmpCamera", //ekeyVulnPointList,
   "fwmpOtherFuze", //ekeyVulnPointList,
   
   "twmrPresFuze", //ekeyVulnPointList,
   "twmrTiltFuze", //ekeyVulnPointList,
   "twmrMagFuze", //ekeyVulnPointList,
   "twmrApFuze", //ekeyVulnPointList,
   "twmrCmdApFuze", //ekeyVulnPointList,
   "twmrCamera", //ekeyVulnPointList,
   "twmrOtherFuze", //ekeyVulnPointList,
   
   "fwmrPresFuze", //ekeyVulnPointList,
   "fwmrTiltFuze", //ekeyVulnPointList,
   "fwmrMagFuze", //ekeyVulnPointList,
   "fwmrApFuze", //ekeyVulnPointList,
   "fwmrCmdApFuze", //ekeyVulnPointList,
   "fwmrCamera", //ekeyVulnPointList,
   "fwmrOtherFuze", //ekeyVulnPointList,

   "dogboneTiltFuze", //ekeyVulnPointList,
   "dogboneMagFuze", //ekeyVulnPointList,
   "dogboneCmdApFuze", //ekeyVulnPointList,
   "dogboneCamera", //ekeyVulnPointList,
   "dogboneOtherFuze", //ekeyVulnPointList,

   "magcmMagFuze", //ekeyVulnPointList,
   "magcmOtherFuze" //, ekeyVulnPointList
};

const char* TDetVulnTable::fVulnDetDefNames[evdtblSizeTable] = {
    "lgblastPresFuze",
    "lgblastBHPresFuze",
    "lgblastTiltFuze", 
    "lgblastApFuze",
    "lgblastCmdApFuze",
    "lgblastCamera",
    "lgblastOtherFuze",
    
    "medblastPresFuze",
    "medblastBHPresFuze",
    "medblastTiltFuze",
    "medblastApFuze",
    "medblastCmdApFuze",
    "medblastCamera",
    "medblastOtherFuze",
    
    "smblastPresFuze",
    "smblastBHPresFuze",
    "smblastTiltFuze",
    "smblastApFuze",
    "smblastCmdApFuze",
    "smblastCamera",
    "smblastOtherFuze",

    "miclicPresFuze", 
    "miclicTiltFuze",
    "miclicBHPresFuze",
    "miclicApFuze",
    "miclicCmdApFuze",
    "miclicCamera",
    "miclicOtherFuze",
    
    "esmbPresFuze", 
    "esmbBHPresFuze", 
    "esmbTiltFuze",
    "esmbApFuze", 
    "esmbCmdApFuze",
    "esmbCamera",
    "esmbOtherFuze",
    
    "lgfragPresFuze",
    "lgfragBHPresFuze", 
    "lgfragTiltFuze", 
    "lgfragApFuze",
    "lgfragCmdApFuze",
    "lgfragCamera",
    "lgfragOtherFuze",
    
    "smfragPresFuze",
    "smfragBHPresFuze",
    "smfragTiltFuze",
    "smfragApFuze", 
    "smfragCmdApFuze",
    "smfragCamera",
    "smfragOtherFuze",
    
    "otherPresFuze",
    "otherBHPresFuze",
    "otherTiltFuze",
    "otherApFuze",
    "otherCmdApFuze",
    "otherCamera",
    "otherOtherFuze"
    
};
const TKeywordEntry TVulnTable::fKeyVulnParam[evkwSizeTable] = {
   "depths", ekeyFloatList,
   "ranges", ekeyFloatList,
   "tables", ekeyVulnPointList,
   "length", ekeyFloat,
   "offset", ekeyFloatList
};
//***********************************
//**      Initializers for vulnerability tables
//********   Track width mine plow
//    default depths const, used to initialize depths for particular tables
const float   twmpDepths[enumdepthTwmP] = {
   -3.0 / 39.37,      // convert inches to meters
   0.0,
   3.0 / 39.37
};
T2Dsingle   twmpPresFuzeOffset(1.0, 0.0);
TVulnProb   ptwmpPresFuzeTable[enumdepthTwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   twmpTiltFuzeOffset(1.0, 0.0);
TVulnProb   ptwmpTiltFuzeTable[enumdepthTwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   twmpMagFuzeOffset(1.0, 0.0);
TVulnProb   ptwmpMagFuzeTable[enumdepthTwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   twmpApFuzeOffset(1.0, 0.0);
TVulnProb   ptwmpApFuzeTable[enumdepthTwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   twmpCmdApFuzeOffset(1.0, 0.0);
TVulnProb   ptwmpCmdApFuzeTable[enumdepthTwmP] = {
   TVulnProb(0.0, 0.1, 0.85, 0.05),
   TVulnProb(0.0, 0.05, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   twmpCameraOffset(1.0, 0.0);
TVulnProb   ptwmpCameraTable[enumdepthTwmP] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0)
};
T2Dsingle   twmpOtherFuzeOffset(1.0, 0.0);
TVulnProb   ptwmpOtherFuzeTable[enumdepthTwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};

//********   Full width mine plow
const float   fwmpDepths[enumdepthFwmP] = {
   -3.0 / 39.37,      // convert inches to meters
   0.0,
   3.0 / 39.37
};
T2Dsingle   fwmpPresFuzeOffset(1.0, 0.0);
TVulnProb   pfwmpPresFuzeTable[enumdepthFwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmpTiltFuzeOffset(1.0, 0.0);
TVulnProb   pfwmpTiltFuzeTable[enumdepthFwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmpMagFuzeOffset(1.0, 0.0);
TVulnProb   pfwmpMagFuzeTable[enumdepthFwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmpApFuzeOffset(1.0, 0.0);
TVulnProb   pfwmpApFuzeTable[enumdepthFwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmpCmdApFuzeOffset(1.0, 0.0);
TVulnProb   pfwmpCmdApFuzeTable[enumdepthFwmP] = {
   TVulnProb(0.0, 0.1, 0.85, 0.05),
   TVulnProb(0.0, 0.05, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmpCameraOffset(1.0, 0.0);
TVulnProb   pfwmpCameraTable[enumdepthFwmP] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0)
};
T2Dsingle   fwmpOtherFuzeOffset(1.0, 0.0);
TVulnProb   pfwmpOtherFuzeTable[enumdepthFwmP] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};

//********   Track width mine roller
const float   twmrDepths[enumdepthTwmR] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
T2Dsingle   twmrPresFuzeOffset(1.0, 0.0);
TVulnProb   ptwmrPresFuzeTable[enumdepthTwmR] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(0.5, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};
T2Dsingle   twmrTiltFuzeOffset(1.0, 0.0);
TVulnProb   ptwmrTiltFuzeTable[enumdepthTwmR] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0),
};
T2Dsingle   twmrMagFuzeOffset(1.0, 0.0);
TVulnProb   ptwmrMagFuzeTable[enumdepthTwmR] = {
   TVulnProb(0.5, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};
T2Dsingle   twmrApFuzeOffset(1.0, 0.0);
TVulnProb   ptwmrApFuzeTable[enumdepthTwmR] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0)
};
T2Dsingle   twmrCmdApFuzeOffset(1.0, 0.0);
TVulnProb   ptwmrCmdApFuzeTable[enumdepthTwmR] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0)
};
T2Dsingle   twmrCameraOffset(1.0, 0.0);
TVulnProb   ptwmrCameraTable[enumdepthTwmR] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0)
};
T2Dsingle   twmrOtherFuzeOffset(1.0, 0.0);
TVulnProb   ptwmrOtherFuzeTable[enumdepthTwmR] = {
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};

//********   Full width mine rake
const float   fwmrDepths[enumdepthFwmR] = {
   -3.0 / 39.37,      // convert inches to meters
   0.0,
   3.0 / 39.37
};
T2Dsingle   fwmrPresFuzeOffset(1.0, 0.0);
TVulnProb   pfwmrPresFuzeTable[enumdepthFwmR] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmrTiltFuzeOffset(1.0, 0.0);
TVulnProb   pfwmrTiltFuzeTable[enumdepthFwmR] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmrMagFuzeOffset(1.0, 0.0);
TVulnProb   pfwmrMagFuzeTable[enumdepthFwmR] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmrApFuzeOffset(1.0, 0.0);
TVulnProb   pfwmrApFuzeTable[enumdepthFwmR] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmrCmdApFuzeOffset(1.0, 0.0);
TVulnProb   pfwmrCmdApFuzeTable[enumdepthFwmR] = {
   TVulnProb(0.0, 0.1, 0.85, 0.05),
   TVulnProb(0.0, 0.05, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};
T2Dsingle   fwmrCameraOffset(1.0, 0.0);
TVulnProb   pfwmrCameraTable[enumdepthFwmR] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0)
};
T2Dsingle   fwmrOtherFuzeOffset(1.0, 0.0);
TVulnProb   pfwmrOtherFuzeTable[enumdepthFwmR] = {
   TVulnProb(0.1, 0.0, 0.85, 0.05),
   TVulnProb(0.05, 0.0, 0.5, 0.05),
   TVulnProb(0.0, 0.0, 0.0, 0.05)
};

//********   dogbone
const float   dbDepths[enumdepthDb] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
T2Dsingle   dbTiltFuzeOffset(1.0, 0.0);
TVulnProb   pdbTiltFuzeTable[enumdepthDb] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0)
};
T2Dsingle   dbMagFuzeOffset(1.0, 0.0);
TVulnProb   pdbMagFuzeTable[enumdepthDb] = {
   TVulnProb(0.5, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};
T2Dsingle   dbCmdApFuzeOffset(1.0, 0.0);
TVulnProb   pdbCmdApFuzeTable[enumdepthDb] = {
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};
T2Dsingle   dbCameraOffset(1.0, 0.0);
TVulnProb   pdbCameraTable[enumdepthDb] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0)
};
T2Dsingle   dbOtherFuzeOffset(1.0, 0.0);
TVulnProb   pdbOtherFuzeTable[enumdepthDb] = {
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};

//********   magnetic cm
const float   magcmDepths[enumdepthMagCm] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
T2Dsingle   magcmMagFuzeOffset(1.0, 0.0);
TVulnProb   pmagcmMagFuzeTable[enumdepthMagCm] = {
   TVulnProb(0.5, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};
T2Dsingle   magcmOtherFuzeOffset(1.0, 0.0);
TVulnProb   pmagcmOtherFuzeTable[enumdepthMagCm] = {
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//***********************************
//********  ESMB
const float   kEsmbHalfLength = 72.5;
const float   esmbDepths[enumdepthEsmb] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   esmbRanges[enumrangeEsmb] = {
   2.5
};
TVulnProb   pesmbPresFuzeTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.8, 0.1, 0.0, 0.0), TVulnProb(0.5, 0.05, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pesmbPresBhFuzeTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.8, 0.1, 0.0, 0.0), TVulnProb(0.5, 0.05, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pesmbTiltFuzeTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.8, 0.1, 0.0, 0.0), TVulnProb(0.5, 0.05, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pesmbApFuzeTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.8, 0.1, 0.0, 0.0), TVulnProb(0.5, 0.05, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pesmbCmdApFuzeTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.0, 0.9, 0.0, 0.0), TVulnProb(0.0, 0.55, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pesmbCameraTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.00, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pesmbOtherFuzeTable[enumdepthEsmb * enumrangeEsmb] = {
   TVulnProb(0.8, 0.1, 0.0, 0.0), TVulnProb(0.5, 0.05, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  MICLIC
const float   kMiclicHalfLength = 5.0;
const float   miclicDepths[enumdepthMiclic] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   miclicRanges[enumrangeMiclic] = {
   0.375, 0.75, 4.0, 7.0
};
TVulnProb   pmiclicPresFuzeTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.93, 0.0, 0.0, 0.0), TVulnProb(0.45, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.91, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0)
};
TVulnProb   pmiclicPresBhFuzeTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmiclicTiltFuzeTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmiclicApFuzeTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmiclicCmdApFuzeTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.93, 0.0, 0.0), TVulnProb(0.0, 0.45, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.91, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0)
};
TVulnProb   pmiclicCameraTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   pmiclicOtherFuzeTable[enumdepthMiclic * enumrangeMiclic] = {
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  Large blast
const float   lgblastDepths[enumdepthLgBlast] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   lgblastRanges[enumrangeLgBlast] = {
   2.0, 5.0, 10.0, 20.0
};
TVulnProb   plgblastPresFuzeTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgblastPresBhFuzeTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgblastTiltFuzeTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgblastApFuzeTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgblastCmdApFuzeTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0),
   TVulnProb(0.0, 0.8, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.3, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0),
   TVulnProb(0.0, 0.2, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgblastCameraTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   plgblastOtherFuzeTable[enumdepthLgBlast * enumrangeLgBlast] = {
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.15, 0.0, 0.0, 0.0), TVulnProb(0.05, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  Medium blast
const float   medblastDepths[enumdepthMedBlast] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   medblastRanges[enumrangeMedBlast] = {
   1.5, 3.0, 7.0, 12.0
};
TVulnProb   pmedblastPresFuzeTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmedblastPresBhFuzeTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmedblastTiltFuzeTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmedblastApFuzeTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmedblastCmdApFuzeTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0),
   TVulnProb(0.0, 0.8, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.3, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0),
   TVulnProb(0.0, 0.2, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   pmedblastCameraTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   pmedblastOtherFuzeTable[enumdepthMedBlast * enumrangeMedBlast] = {
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.15, 0.0, 0.0, 0.0), TVulnProb(0.05, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  small blast
const float   smblastDepths[enumdepthSmBlast] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   smblastRanges[enumrangeSmBlast] = {
   1.0, 2.0, 5.0, 10.0
};
TVulnProb   psmblastPresFuzeTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmblastPresBhFuzeTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.05, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmblastTiltFuzeTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.95, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.8, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0)
};
TVulnProb   psmblastApFuzeTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmblastCmdApFuzeTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0),
   TVulnProb(0.0, 0.8, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.3, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0),
   TVulnProb(0.0, 0.2, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmblastCameraTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   psmblastOtherFuzeTable[enumdepthSmBlast * enumrangeSmBlast] = {
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.15, 0.0, 0.0, 0.0), TVulnProb(0.05, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  Large frag
const float   lgfragDepths[enumdepthLgFrag] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   lgfragRanges[enumrangeLgFrag] = {
   2.0, 5.0, 10.0, 20.0
};
TVulnProb   plgfragPresFuzeTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgfragPresBhFuzeTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgfragTiltFuzeTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgfragApFuzeTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgfragCmdApFuzeTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0),
   TVulnProb(0.0, 0.8, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.3, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0),
   TVulnProb(0.0, 0.2, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   plgfragCameraTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   plgfragOtherFuzeTable[enumdepthLgFrag * enumrangeLgFrag] = {
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.15, 0.0, 0.0, 0.0), TVulnProb(0.05, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  small frag
const float   smfragDepths[enumdepthSmFrag] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   smfragRanges[enumrangeLgBlast] = {
   1.0, 3.0, 8.0, 12.0
};
TVulnProb   psmfragPresFuzeTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmfragPresBhFuzeTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmfragTiltFuzeTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmfragApFuzeTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmfragCmdApFuzeTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0),
   TVulnProb(0.0, 0.8, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.3, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0),
   TVulnProb(0.0, 0.2, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   psmfragCameraTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   psmfragOtherFuzeTable[enumdepthSmFrag * enumrangeSmFrag] = {
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.15, 0.0, 0.0, 0.0), TVulnProb(0.05, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
//********  Other Warhead
const float   otherwhDepths[enumdepthOtherWh] = {
   0.0,
   3.0 / 39.37,
   9.0 / 39.37
};
const float   otherwhRanges[enumrangeOtherWh] = {
   2.0, 5.0, 10.0, 20.0
};
TVulnProb   potherwhPresFuzeTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(0.95, 0.0, 0.0, 0.0), TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0),
   TVulnProb(0.8, 0.0, 0.0, 0.0), TVulnProb(0.5, 0.0, 0.0, 0.0), TVulnProb(0.2, 0.0, 0.0, 0.0),
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   potherwhPresBhFuzeTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.2, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   potherwhTiltFuzeTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   potherwhApFuzeTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0),
   TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0), TVulnProb(1.0, 0.0, 0.0, 0.0)
};
TVulnProb   potherwhCmdApFuzeTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(0.0, 0.95, 0.0, 0.0), TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.4, 0.0, 0.0),
   TVulnProb(0.0, 0.8, 0.0, 0.0), TVulnProb(0.0, 0.5, 0.0, 0.0), TVulnProb(0.0, 0.2, 0.0, 0.0),
   TVulnProb(0.0, 0.6, 0.0, 0.0), TVulnProb(0.0, 0.3, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0),
   TVulnProb(0.0, 0.2, 0.0, 0.0), TVulnProb(0.0, 0.1, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};
TVulnProb   potherwhCameraTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0),
   TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0), TVulnProb(0.0, 1.0, 0.0, 0.0)
};
TVulnProb   potherwhOtherFuzeTable[enumdepthOtherWh * enumrangeOtherWh] = {
   TVulnProb(0.6, 0.0, 0.0, 0.0), TVulnProb(0.4, 0.0, 0.0, 0.0), TVulnProb(0.1, 0.0, 0.0, 0.0),
   TVulnProb(0.3, 0.0, 0.0, 0.0), TVulnProb(0.15, 0.0, 0.0, 0.0), TVulnProb(0.05, 0.0, 0.0, 0.0),
   TVulnProb(0.1, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0),
   TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0), TVulnProb(0.0, 0.0, 0.0, 0.0)
};

cMechThreatList  gMechThreatList;
cDetThreatList   gDetThreatList;

//***********************************
const char*   TVulnProb::GetVulnResultText(int result)
{
   switch (result) {
      case evrNoVuln:
            return "Not in range";
//         break;
   
      case evrNone:
            return "No damage";
//         break;
   
      case evrDet:
            return "Detonated";
//         break;
   
      case evrNeut:
            return "Neutralized";
//         break;
   
      case evrDisplace:
            return "Displaced";
//         break;
   
      case evrRollback:
            return "Displaced/Rollback";
//         break;
   }
   return "Vuln result error" ;
   
}

TVulnProb::TVulnProb( float detPr, float neutPr, 
              float displacePr, float rollbackPr )
{
   fDetProb = detPr;
   fNeutProb = neutPr + detPr;
   fDisplaceProb = displacePr + fNeutProb; 
   fRollbackProb = rollbackPr + fDisplaceProb; 
}

TVulnTable::TVulnTable(T2link* pPrev) :
    T2link(pPrev)
{
   fpName = NULL;
   fLinearXlength = 0.0;
   fNumDepth = 0;
   fNumRange = 0;
   fpRanges = NULL;
   fpDepths = NULL;
   fpVulnTab = NULL;
}

TVulnTable::TVulnTable(T2link* pPrev, char* pName) :
    T2link(pPrev)
{
   if ( pName ) fpName = cppstrdup(pName);
   fLinearXlength = 0.0;
   fNumDepth = 0;
   fNumRange = 0;
   fpRanges = NULL;
   fpDepths = NULL;
   fpVulnTab = NULL;
}

TVulnTable::~TVulnTable()
{
   if (fpRanges) delete[] fpRanges;
   if (fpDepths) delete[] fpDepths;
   if (fpVulnTab) delete[] fpVulnTab;
   if (fpName) delete[] fpName;
}

EBoolean   TMechVulnTable::IsVehicleEncounter(const TEntCoord& relLoc, 
                     const TVehDimensions& vehDim)
{
   if ( ( fabs(relLoc.x) < vehDim.length/2.0 + fOffset.x ) && 
        ( fabs(relLoc.y) < vehDim.width/2.0 + fOffset.y ) )
      return true;
   else
      return false;
}

EBoolean   TMechVulnTable::IsTrackEncounter(const TEntCoord& relLoc, 
                     const TVehDimensions& vehDim)
{
   float absrelY = fabs(relLoc.y);
   
   if ( ( fabs(relLoc.x) < vehDim.length/2.0 + fOffset.x ) && 
        ( absrelY < vehDim.trackYmax + fOffset.y ) &&
        ( absrelY > vehDim.trackYmin - fOffset.y ) )
      return true;
      
   else
      return false;
}

EBoolean   TMechVulnTable::IsBellyEncounter(const TEntCoord& relLoc, 
                     const TVehDimensions& vehDim)
{
   if ( ( fabs(relLoc.x) < vehDim.length/2.0 + fOffset.x ) && 
        ( fabs(relLoc.y) < vehDim.trackYmin + fOffset.y ) )
      return true;
   else
      return false;
}

EVulnResult      TMechVulnTable::GetRandomResult(float depth, float effectiveness)
{
   int i = 0;
   for ( i = 0; i < fNumDepth; i++ )
      if ( depth <= fpDepths[i] )
         break;
         
   if ( i >= fNumDepth )
      return evrNone;
      
   float ran = UniformRand();
   //cout << "TMechVulnTable::GetRandomResult rand = " << ran << endl;
   //cout << "Det = " << (fpVulnTab[i].fDetProb) << "Nuet = " << (fpVulnTab[i].fNeutProb);
   //cout << "Plow = " << (fpVulnTab[i].fDisplaceProb) << "Rback = " << (fpVulnTab[i].fRollbackProb) << endl;
   if ( ran < (fpVulnTab[i].fDetProb)*effectiveness )
      return evrDet;
   else if ( ran < (fpVulnTab[i].fNeutProb)*effectiveness )
      return evrNeut;
   else if ( ran < (fpVulnTab[i].fDisplaceProb)*effectiveness )
      return evrDisplace;
   else if ( ran < (fpVulnTab[i].fRollbackProb)*effectiveness )
      return evrRollback;
      
   return evrNone;
}

TMechVulnTable::TMechVulnTable(T2link* pPrev, T2Dsingle& offset, int index, 
                  int numDepth, TVulnProb* pTable, const float* pDepths) : 
   TVulnTable(pPrev)
{
   fNumDepth = numDepth;
   fOffset = offset;
   fpVulnTab = pTable;
   fIndex   = index;
   
   fpName = cppstrdup(TVulnMechParam::GetNameByIndex(index));
      
   if (fNumDepth > 0) {
      fpDepths = new float[fNumDepth];
      fpVulnTab = new TVulnProb[fNumDepth];
      for ( int i = 0; i < fNumDepth; i++ ) {
         fpDepths[i] = pDepths[i];
         fpVulnTab[i] = pTable[i];
      }
   }
   else {
      fpDepths = NULL;
      fpVulnTab = NULL;
   }
}

TMechVulnTable::~TMechVulnTable()
{
   if (fpDepths != NULL)
      delete [] fpDepths;
   if (fpVulnTab != NULL)
      delete [] fpVulnTab;
}

cThreatList::cThreatList()
{
    
}

cThreatList::~cThreatList()
{
    
}

/*
TMechVulnTable*   cMechThreatList::FindTableByName(char* pName)
{
   for ( TMechVulnTable* pTable = (TMechVulnTable*)GetFirstLink();
                         pTable != (TMechVulnTable*)GetNULLLink();
                         pTable = (TMechVulnTable*)GetNextLink() ) {
	if ( ! strcmp(pTable->GetTableName(), pName) )
         return pTable;
   }
   return NULL;
}


TMechVulnTable*   cMechThreatList::FindTableByIndex(int index)
{
   for ( TMechVulnTable* pTable = (TMechVulnTable*)GetFirstLink();
                         pTable != (TMechVulnTable*)GetNULLLink();
                         pTable = (TMechVulnTable*)GetNextLink() ) {
      if ( pTable->GetTableIndex() == index )
         return pTable;
   }
   return NULL;
}
*/

void   cMechThreatList::ConstructBasicList()
   // constructs vuln tables for basic cm and fuze types
{
         // track width plows
   new TMechVulnTable(this, twmpPresFuzeOffset, TVulnMechParam::evmtblTwmPPres, 
                     enumdepthTwmP, ptwmpPresFuzeTable, twmpDepths);
   new TMechVulnTable(this, twmpTiltFuzeOffset, TVulnMechParam::evmtblTwmPTilt, 
                     enumdepthTwmP, ptwmpTiltFuzeTable, twmpDepths);
   new TMechVulnTable(this, twmpMagFuzeOffset, TVulnMechParam::evmtblTwmPMag, 
                     enumdepthTwmP, ptwmpMagFuzeTable, twmpDepths);
   new TMechVulnTable(this, twmpApFuzeOffset, TVulnMechParam::evmtblTwmPAp, 
                     enumdepthTwmP, ptwmpApFuzeTable, twmpDepths);
   new TMechVulnTable(this, twmpCmdApFuzeOffset, TVulnMechParam::evmtblTwmPCmdAp, 
                     enumdepthTwmP, ptwmpCmdApFuzeTable, twmpDepths);
   new TMechVulnTable(this, twmpCameraOffset, TVulnMechParam::evmtblTwmPCam, 
                     enumdepthTwmP, ptwmpCameraTable, twmpDepths);
   new TMechVulnTable(this, twmpOtherFuzeOffset, TVulnMechParam::evmtblTwmPOther, 
                     enumdepthTwmP, ptwmpOtherFuzeTable, twmpDepths);
   
         // full width plows
   new TMechVulnTable(this, fwmpPresFuzeOffset, TVulnMechParam::evmtblFwmPPres, 
                     enumdepthFwmP, pfwmpPresFuzeTable, fwmpDepths);
   new TMechVulnTable(this, fwmpTiltFuzeOffset, TVulnMechParam::evmtblFwmPTilt, 
                     enumdepthFwmP, pfwmpTiltFuzeTable, fwmpDepths);
   new TMechVulnTable(this, fwmpMagFuzeOffset, TVulnMechParam::evmtblFwmPMag, 
                     enumdepthFwmP, pfwmpMagFuzeTable, fwmpDepths);
   new TMechVulnTable(this, fwmpApFuzeOffset, TVulnMechParam::evmtblFwmPAp, 
                     enumdepthFwmP, pfwmpApFuzeTable, fwmpDepths);
   new TMechVulnTable(this, fwmpCmdApFuzeOffset, TVulnMechParam::evmtblFwmPCmdAp, 
                     enumdepthFwmP, pfwmpCmdApFuzeTable, fwmpDepths);
   new TMechVulnTable(this, fwmpCameraOffset, TVulnMechParam::evmtblFwmPCam, 
                     enumdepthFwmP, pfwmpCameraTable, fwmpDepths);
   new TMechVulnTable(this, fwmpOtherFuzeOffset, TVulnMechParam::evmtblFwmPOther, 
                     enumdepthFwmP, pfwmpOtherFuzeTable, fwmpDepths);
   
         // track width rollers
   new TMechVulnTable(this, twmrPresFuzeOffset, TVulnMechParam::evmtblTwmRPres, 
                     enumdepthTwmR, ptwmrPresFuzeTable, twmrDepths);
   new TMechVulnTable(this, twmrTiltFuzeOffset, TVulnMechParam::evmtblTwmRTilt, 
                     enumdepthTwmR, ptwmrTiltFuzeTable, twmrDepths);
   new TMechVulnTable(this, twmrMagFuzeOffset, TVulnMechParam::evmtblTwmRMag, 
                     enumdepthTwmR, ptwmrMagFuzeTable, twmrDepths);
   new TMechVulnTable(this, twmrApFuzeOffset, TVulnMechParam::evmtblTwmRAp, 
                     enumdepthTwmR, ptwmrApFuzeTable, twmrDepths);
   new TMechVulnTable(this, twmrCmdApFuzeOffset, TVulnMechParam::evmtblTwmRCmdAp, 
                     enumdepthTwmR, ptwmrCmdApFuzeTable, twmrDepths);
   new TMechVulnTable(this, twmrCameraOffset, TVulnMechParam::evmtblTwmRCam, 
                     enumdepthTwmR, ptwmrCameraTable, twmrDepths);
   new TMechVulnTable(this, twmrOtherFuzeOffset, TVulnMechParam::evmtblTwmROther, 
                     enumdepthTwmR, ptwmrOtherFuzeTable, twmrDepths);
   
         // full width rakes
   new TMechVulnTable(this, fwmrPresFuzeOffset, TVulnMechParam::evmtblFwmRPres, 
                     enumdepthFwmR, pfwmrPresFuzeTable, fwmrDepths);
   new TMechVulnTable(this, fwmrTiltFuzeOffset, TVulnMechParam::evmtblFwmRTilt, 
                     enumdepthFwmR, pfwmrTiltFuzeTable, fwmrDepths);
   new TMechVulnTable(this, fwmrMagFuzeOffset, TVulnMechParam::evmtblFwmRMag, 
                     enumdepthFwmR, pfwmpMagFuzeTable, fwmrDepths);
   new TMechVulnTable(this, fwmrApFuzeOffset, TVulnMechParam::evmtblFwmRAp, 
                     enumdepthFwmR, pfwmrApFuzeTable, fwmrDepths);
   new TMechVulnTable(this, fwmrCmdApFuzeOffset, TVulnMechParam::evmtblFwmRCmdAp, 
                     enumdepthFwmR, pfwmrCmdApFuzeTable, fwmrDepths);
   new TMechVulnTable(this, fwmrCameraOffset, TVulnMechParam::evmtblFwmRCam, 
                     enumdepthFwmR, pfwmrCameraTable, fwmrDepths);
   new TMechVulnTable(this, fwmrOtherFuzeOffset, TVulnMechParam::evmtblFwmROther, 
                     enumdepthFwmR, pfwmrOtherFuzeTable, fwmrDepths);
   
         // dogbone assemblies
   new TMechVulnTable(this, dbTiltFuzeOffset, TVulnMechParam::evmtblDbTilt, 
                     enumdepthDb, pdbTiltFuzeTable, dbDepths);
   new TMechVulnTable(this, dbMagFuzeOffset, TVulnMechParam::evmtblDbMag, 
                     enumdepthDb, pdbMagFuzeTable, dbDepths);
   new TMechVulnTable(this, dbCmdApFuzeOffset, TVulnMechParam::evmtblDbCmdAp, 
                     enumdepthDb, pdbCmdApFuzeTable, dbDepths);
   new TMechVulnTable(this, dbCameraOffset, TVulnMechParam::evmtblDbCam, 
                     enumdepthDb, pdbCameraTable, dbDepths);
   new TMechVulnTable(this, dbOtherFuzeOffset, TVulnMechParam::evmtblDbOther, 
                     enumdepthDb, pdbOtherFuzeTable, dbDepths);
   
         // magnetic countermeasures
   new TMechVulnTable(this, magcmMagFuzeOffset, TVulnMechParam::evmtblMagCmMag, 
                     enumdepthMagCm, pmagcmMagFuzeTable, magcmDepths);
   new TMechVulnTable(this, magcmOtherFuzeOffset, TVulnMechParam::evmtblMagCmOther, 
                     enumdepthMagCm, pmagcmOtherFuzeTable, magcmDepths);


}

cMechThreatList::cMechThreatList()
{
	ConstructBasicList();
}

cMechThreatList::~cMechThreatList()
{
}

const char*   TVulnMechParam::GetNameByIndex( int index )
{
   if ( (index < evmtblSizeTable) && (index >= 0) )
      return fKeyMechVulnTables[index] /*.key*/;
   else
      return NULL;
}

EBoolean     TVulnMechParam::ParseDbStructItem(cInFileObj& /* infile */)
{
   return true;
}

cMechanVuln*      TVulnMechParam::CreateVulnMech( cSimMine* pMine )
{
   return new cMechanVuln( this, pMine );
}

TVulnMechParam::TVulnMechParam()
{
    fpTwmPTable = NULL;
    fpFwmPTable = NULL;
    fpTwmRTable = NULL;
    fpFwmRTable = NULL;
    fpTwmrDbTable = NULL;
    fpMagCmTable = NULL;
}

TVulnMechParam::TVulnMechParam(int fuzeType)
{
    InitVulnMechParam(fuzeType);
}

void TVulnMechParam::InitVulnMechParam(int fuzeType)
{
   //cout << "Fuze type = " << fuzeType << endl;
   switch (fuzeType) {
      case TFuzeParam::efuzetypePressure:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPPres);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPPres);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmRPres);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmRPres);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbOther);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
   
      case TFuzeParam::efuzetypeTiltrod:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPTilt);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPTilt);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmRTilt);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmRTilt);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbTilt);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
   
      case TFuzeParam::efuzetypeMagnetic:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPMag);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPMag);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmRMag);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmRMag);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbMag);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmMag);
         break;
   
      case TFuzeParam::efuzetypeAcousticTracker:
      case TFuzeParam::efuzetypeAcousticTrigger:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPOther);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPOther);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmROther);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmROther);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbOther);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
   
      case TFuzeParam::efuzetypeApPressure:
      case TFuzeParam::efuzetypeApTrip:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPAp);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPAp);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmRAp);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmRAp);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbOther);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
         
      case TFuzeParam::efuzetypeCommandFire:
      case TFuzeParam::efuzetypeCommandLoc:
      case TFuzeParam::efuzetypeCommandFireDir:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPCmdAp);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPCmdAp);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmRCmdAp);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmRCmdAp);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbCmdAp);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
   
      case TFuzeParam::esenstypeCommandCamDir:
      case TFuzeParam::esenstypeAutoCmdCamDir:
      case TFuzeParam::esenstypePerDet:
      case TFuzeParam::esenstypeVehDet:
      case TFuzeParam::esenstypePerVehDet:
      case TFuzeParam::esenstypeDirDet:
      case TFuzeParam::esenstypeTripWire:
      case TFuzeParam::esenstypePerTrack:
      case TFuzeParam::esenstypeVehTrack:
      case TFuzeParam::esenstypePerVehTrack:
      case TFuzeParam::esenstypeDirPVTrack:
      case TFuzeParam::esenstypeDirTrack:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPCam);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPCam);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmRCam);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmRCam);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbCam);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
         
      default:
            fpTwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmPOther);
            fpFwmPTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmPOther);
            fpTwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblTwmROther);
            fpFwmRTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblFwmROther);
            fpTwmrDbTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblDbOther);
            fpMagCmTable = (TMechVulnTable*) gMechThreatList.FindTableByIndex(evmtblMagCmOther);
         break;
   }
}

EVulnResult      TDetVulnTable::GetRandomResult(float depth, float range)
{
   int i = 0;
   int j = 0;
   
   for ( i = 0; i < fNumDepth; i++ )
      if ( depth <= fpDepths[i] )
         break;
   if ( i >= fNumDepth )
      return evrNoVuln;
   
   for ( j = 0; j < fNumRange; j++ )
      if ( range <= fpRanges[j] )
         break;
   if ( j >= fNumRange )
      return evrNoVuln;
      
   float ran = UniformRand();
   if ( ran < fpVulnTab[ j*fNumDepth + i ].fDetProb )
      return evrDet;
   else if ( ran < fpVulnTab[ j*fNumDepth + i ].fNeutProb )
      return evrNeut;
   else if ( ran < fpVulnTab[ j*fNumDepth + i ].fDisplaceProb )
      return evrDisplace;
   else if ( ran < fpVulnTab[ j*fNumDepth + i ].fRollbackProb )
      return evrRollback;
      
   return evrNone;
}

EBoolean    TVulnTable::ReadVulnTable( cInFileObj& infile, int tblType )
{
   EFileReturn  returnCode;
   int          returnIndex;
   void*        returnVal = NULL;
   EBoolean     done = false;
   TItemList*   pList;
   EBoolean     success = true;
   
      // temporaries, in case inconsistent values are read
   int          numRange = 0, numDepth, numEntry, i, numOffset = 0;
   float        linLength = fLinearXlength;
   float*       ranges = NULL;
   float*       depths;
   TVulnProb*   probs;
   float*	offset=NULL;

   while( !done )
   {
      returnCode = infile.ReadItem( fKeyVulnParam, evkwSizeTable,
         &returnIndex, &returnVal );

      switch (returnCode)
      {
         case eFileNoErr:
            switch( returnIndex )
            {
               case evkwDepths:
                  pList = (TItemList *)returnVal;
                  numDepth = pList->count;
                  depths = new float[numDepth];
                  for ( i = 0; i < numDepth; i++ )
                     depths[i] = *( (float*)(pList->ptrItems) + i);

                  delete[] (float *)(pList->ptrItems);
                  delete pList;
									pList = NULL;
									returnVal = NULL;
                  break;

               case evkwRanges:
                  pList = (TItemList *)returnVal;
                  numRange = pList->count;
                  ranges = new float[numRange];
                  for ( i = 0; i < numRange; i++ )
                     ranges[i] = *( (float*)(pList->ptrItems) + i);

                  delete[] (float *)(pList->ptrItems);
                  delete pList;
									pList = NULL;
									returnVal = NULL;
                  break;

               case evkwProbs:
                  pList = (TItemList *)returnVal;
                  numEntry = pList->count;
                  probs = new TVulnProb[numEntry];
                  for ( i = 0; i < numEntry; i++ )
                     probs[i] = *( (TVulnProb*)(pList->ptrItems) + i);

                  delete[] (TVulnProb *)(pList->ptrItems);
                  delete pList;
									pList = NULL;
									returnVal = NULL;
                  break;

               case evkwLinLength:
                     linLength = *(float*)returnVal;
                     delete returnVal;
										 returnVal = NULL;
                  break;

               case evkwOffset:
                  pList = (TItemList *)returnVal;
                  numOffset = pList->count;
                  offset = new float[numOffset];
                  for ( i = 0; i < numOffset; i++ )
                     offset[i] = *( (float*)(pList->ptrItems) + i);

                  delete[] (float *)(pList->ptrItems);
                  delete pList;
									pList = NULL;
									returnVal = NULL;
                  break;

               default:
                  // should not go here
                  break;
            }
            break;

			case eLevelDone:      // finished reading, check consistency
	    if ( (numEntry == numDepth * numRange && tblType == etbltypeDet) ||
			 (numEntry == numDepth  && tblType == etbltypeMech) ) {
               fNumDepth = numDepth;
               fNumRange = numRange;
               fLinearXlength = linLength;
               if ( fpDepths ) delete[] fpDepths;
               if ( fpRanges ) delete[] fpRanges;
               if ( fpVulnTab ) delete[] fpVulnTab;
               fpDepths = depths;
               fpRanges = ranges;
               fpVulnTab = probs;
	       if (offset) {
		  fOffset.x = *offset;
		  fOffset.y = *(offset+1);
	       }
            }
            else
               success = false;
            
            done = true;
				break;

         default:  // error return
            Assert(false, kerrReadParamfile, "TVulnTable::ReadVulnTable");
            success = false;
            break;
      }
   }
   delete[] offset;
   return success;
}

TDetVulnTable::TDetVulnTable( T2link* pPrev, char* pName ) :
   TVulnTable (pPrev)
{
   if ( pName ) fpName = cppstrdup(pName);
   fLinearXlength = 0.0;
   fNumDepth = 0;
   fNumRange = 0;
   fpRanges = NULL;
   fpDepths = NULL;
   fpVulnTab = NULL;
}

TDetVulnTable::TDetVulnTable(T2link* pPrev, int index, int numDepth, int numRange,
        TVulnProb* pTable, const float* pDepths, const float* pRanges, float length) : 
   TVulnTable (pPrev)
{
   int i;
   
   fLinearXlength = length;
   fNumDepth = numDepth;
   fNumRange = numRange;

   fpName = (char*)fVulnDetDefNames[index];
      
   if (fNumDepth > 0) {
      fpDepths = new float[fNumDepth];
      for ( i = 0; i < fNumDepth; i++ )
         fpDepths[i] = pDepths[i];
   }
   else
      fpDepths = NULL;
      
   if (fNumRange > 0) {
      fpRanges = new float[fNumRange];
      for ( i = 0; i < fNumRange; i++ )
         fpRanges[i] = pRanges[i];
   }
   else
      fpRanges = NULL;
			
   if ( (fNumDepth > 0) && (fNumRange > 0) ) {
      fpVulnTab = new TVulnProb[fNumDepth * fNumRange];
      for ( i = 0; i < fNumDepth * fNumRange; i++ )
         fpVulnTab[i] = pTable[i];
   }
   else
      fpVulnTab = NULL;
}

TDetVulnTable::~TDetVulnTable()
{
   if (fpRanges != NULL) 
      delete [] fpRanges;

   if (fpVulnTab != NULL)
      delete [] fpVulnTab;

   if (fpDepths != NULL)
      delete [] fpDepths;
}


TVulnTable*   cThreatList::FindTableByName(char* pName)
{
   for ( TVulnTable*     pTable = (TVulnTable*)GetFirstLink();
                         pTable != (TVulnTable*)GetNULLLink();
                         pTable = (TVulnTable*)GetNextLink() ) {
      if ( ! strcmp(pTable->GetTableName(), pName) )
         return pTable;
   }
   return NULL;
}


TVulnTable*   cThreatList::FindTableByIndex(int index)
{
   for ( TVulnTable*     pTable = (TVulnTable*)GetFirstLink();
                         pTable != (TVulnTable*)GetNULLLink();
                         pTable = (TVulnTable*)GetNextLink() ) {
      if ( pTable->GetTableIndex() == index )
         return pTable;
   }
   return NULL;
}

TVulnTable*   cDetThreatList::FindTableByIndex(int index)
{
   int i;
   TVulnTable* pTable;
   
   for ( i = 0,          pTable = (TVulnTable*)GetFirstLink();
         (i < index) && (pTable != (TVulnTable*)GetNULLLink());
         i++,            pTable = (TVulnTable*)GetNextLink() ) {
   }
   if ( i == index )
      return pTable;
   else
      return NULL;
}


void   cDetThreatList::ConstructBasicList()
   // constructs vuln tables for basic threat munitions and fuze types
{
         // large blast
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastPres, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastPresFuzeTable, lgblastDepths, lgblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastBhPres, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastPresBhFuzeTable, lgblastDepths, lgblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastTilt, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastTiltFuzeTable, lgblastDepths, lgblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastAp, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastApFuzeTable, lgblastDepths, lgblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastCmdAp, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastCmdApFuzeTable, lgblastDepths, lgblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastCam, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastCameraTable, lgblastDepths, lgblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgBlastOther, enumdepthLgBlast, enumrangeLgBlast,
                     plgblastOtherFuzeTable, lgblastDepths, lgblastRanges);

         // med blast
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastPres, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastPresFuzeTable, medblastDepths, medblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastBhPres, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastPresBhFuzeTable, medblastDepths, medblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastTilt, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastTiltFuzeTable, medblastDepths, medblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastAp, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastApFuzeTable, medblastDepths, medblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastCmdAp, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastCmdApFuzeTable, medblastDepths, medblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastCam, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastCameraTable, medblastDepths, medblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMedBlastOther, enumdepthMedBlast, enumrangeMedBlast,
                     pmedblastOtherFuzeTable, medblastDepths, medblastRanges);

         // small blast
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastPres, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastPresFuzeTable, smblastDepths, smblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastBhPres, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastPresBhFuzeTable, smblastDepths, smblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastTilt, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastTiltFuzeTable, smblastDepths, smblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastAp, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastApFuzeTable, smblastDepths, smblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastCmdAp, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastCmdApFuzeTable, smblastDepths, smblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastCam, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastCameraTable, smblastDepths, smblastRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmBlastOther, enumdepthSmBlast, enumrangeSmBlast,
                     psmblastOtherFuzeTable, smblastDepths, smblastRanges);

         // miclic
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicPres, enumdepthMiclic, enumrangeMiclic,
                     pmiclicPresFuzeTable, miclicDepths, miclicRanges, kMiclicHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicTilt, enumdepthMiclic, enumrangeMiclic,
                     pmiclicTiltFuzeTable, miclicDepths, miclicRanges, kMiclicHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicBhPres, enumdepthMiclic, enumrangeMiclic,
                     pmiclicPresBhFuzeTable, miclicDepths, miclicRanges, kMiclicHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicAp, enumdepthMiclic, enumrangeMiclic,
                     pmiclicApFuzeTable, miclicDepths, miclicRanges, kMiclicHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicCmdAp, enumdepthMiclic, enumrangeMiclic,
                     pmiclicCmdApFuzeTable, miclicDepths, miclicRanges, kMiclicHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicCam, enumdepthMiclic, enumrangeMiclic,
                     pmiclicCameraTable, miclicDepths, miclicRanges, kMiclicHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblMiclicOther, enumdepthMiclic, enumrangeMiclic,
                     pmiclicOtherFuzeTable, miclicDepths, miclicRanges, kMiclicHalfLength);

         // esmb
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbPres, enumdepthEsmb, enumrangeEsmb,
                     pesmbPresFuzeTable, esmbDepths, esmbRanges, kEsmbHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbBhPres, enumdepthEsmb, enumrangeEsmb,
                     pesmbPresBhFuzeTable, esmbDepths, esmbRanges, kEsmbHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbTilt, enumdepthEsmb, enumrangeEsmb,
                     pesmbTiltFuzeTable, esmbDepths, esmbRanges, kEsmbHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbAp, enumdepthEsmb, enumrangeEsmb,
                     pesmbApFuzeTable, esmbDepths, esmbRanges, kEsmbHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbCmdAp, enumdepthEsmb, enumrangeEsmb,
                     pesmbCmdApFuzeTable, esmbDepths, esmbRanges, kEsmbHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbCam, enumdepthEsmb, enumrangeEsmb,
                     pesmbCameraTable, esmbDepths, esmbRanges, kEsmbHalfLength);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblEsmbOther, enumdepthEsmb, enumrangeEsmb,
                     pesmbOtherFuzeTable, esmbDepths, esmbRanges, kEsmbHalfLength);


         // large frag
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragPres, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragPresFuzeTable, lgfragDepths, lgfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragBhPres, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragPresBhFuzeTable, lgfragDepths, lgfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragTilt, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragTiltFuzeTable, lgfragDepths, lgfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragAp, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragApFuzeTable, lgfragDepths, lgfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragCmdAp, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragCmdApFuzeTable, lgfragDepths, lgfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragCam, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragCameraTable, lgfragDepths, lgfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblLgFragOther, enumdepthLgFrag, enumrangeLgFrag,
                     plgfragOtherFuzeTable, lgfragDepths, lgfragRanges);

         // small frag
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragPres, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragPresFuzeTable, smfragDepths, smfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragBhPres, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragPresBhFuzeTable, smfragDepths, smfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragTilt, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragTiltFuzeTable, smfragDepths, smfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragAp, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragApFuzeTable, smfragDepths, smfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragCmdAp, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragCmdApFuzeTable, smfragDepths, smfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragCam, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragCameraTable, smfragDepths, smfragRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblSmFragOther, enumdepthSmFrag, enumrangeSmFrag,
                     psmfragOtherFuzeTable, smfragDepths, smfragRanges);

         // other warheads
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherPres, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhPresFuzeTable, otherwhDepths, otherwhRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherBhPres, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhPresBhFuzeTable, otherwhDepths, otherwhRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherTilt, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhTiltFuzeTable, otherwhDepths, otherwhRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherAp, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhApFuzeTable, otherwhDepths, otherwhRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherCmdAp, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhCmdApFuzeTable, otherwhDepths, otherwhRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherCam, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhCameraTable, otherwhDepths, otherwhRanges);
   new TDetVulnTable(fpPrev, TDetVulnTable::evdtblOtherOther, enumdepthOtherWh, enumrangeOtherWh,
                     potherwhOtherFuzeTable, otherwhDepths, otherwhRanges);


}

void  cMechThreatList::ReadParamVulnTable( cInFileObj& infile, char* pName )
{
   TMechVulnTable*  pdvt = (TMechVulnTable*) FindTableByName(pName);
   
   if ( pdvt ) {
      cout << "Reading parameters for countermine vuln table '" << pName << "'" << endl;
      if ( ! pdvt->ReadVulnTable(infile, TVulnTable::etbltypeMech) )
         cout << "Error in reading vuln table for '" << pName << "'" << endl;
         
   }
   else {
      cout << "Warning:  Encountered unknown vuln table, '" << pName << "'" << endl;
   }
}

void  cMechThreatList::ReadParamNewVulnTable( cInFileObj& infile, char* pName )
{
   TVulnTable*  pdvt = new TVulnTable( fpPrev, pName );
   
   cout << "Creating new countermine vuln table '" << pName << "'" << endl;
   if ( ! pdvt->ReadVulnTable(infile, TVulnTable::etbltypeMech) ) {
      cout << "Error in reading vuln table for '" << pName << "'" << endl;
      delete pdvt;
   }
}

void  cDetThreatList::ReadParamVulnTable( cInFileObj& infile, char* pName )
{
   TDetVulnTable*  pdvt = (TDetVulnTable*) FindTableByName(pName);
   
   if ( pdvt ) {
      cout << "Reading parameters for detonation vuln table '" << pName << "'" << endl;
      if ( ! pdvt->ReadVulnTable(infile, TVulnTable::etbltypeDet) )
         cout << "Error in reading vuln table for '" << pName << "'" << endl;
         
   }
   else {
      cout << "Warning:  Encountered unknown vuln table, '" << pName << "'" << endl;
   }
}

void  cDetThreatList::ReadParamNewVulnTable( cInFileObj& infile, char* pName )
{
   TDetVulnTable*  pdvt = new TDetVulnTable( fpPrev, pName );
   
   cout << "Creating new detonation vuln table '" << pName << "'" << endl;
   if ( ! pdvt->ReadVulnTable(infile, TVulnTable::etbltypeDet) ) {
      cout << "Error in reading vuln table for '" << pName << "'" << endl;
      delete pdvt;
   }
}

cDetThreatList::cDetThreatList()
{
   ConstructBasicList();
}

cDetThreatList::~cDetThreatList()
{
}

cDetonVuln*      cDetVulnList::CreateVulnDet( cSimMine* pMine, float extraRange )
{
   cDetonVuln* pNewDetVuln = NULL;

/*   
   if ( pMine->IsFalseAlarm() )
      pNewDetVuln = new cFalseDetectDetonVuln(this, pMine, extraRange);
   else
*/

   pNewDetVuln = new cDetonVuln(this, pMine, extraRange);
   
   return pNewDetVuln;
}

TDetVulnTable*     cDetVulnList::GetVulnDetTable( int detType )         
{
   TDetVulnTable* pdvTable = NULL;
   for ( TDetVulnBlock* pdvBlock = (TDetVulnBlock*)GetFirstLink();
                        pdvBlock != GetNULLLink();
                        pdvBlock = (TDetVulnBlock*)GetNextLink() ) {
     if ( pdvTable = pdvBlock->GetVulnDetTable(detType) )
        break;
   }
   
   return pdvTable;
}

float    cDetVulnList::GetMaxRange(int detType, float  /*depth*/ )
{
   TDetVulnTable* pTable = GetVulnDetTable(detType);
   if ( pTable )
      return pTable->fpRanges[pTable->fNumRange - 1];
   else
      return 0.0;
}

void        cDetVulnList::AddVulnDetEntry(int detType, TDetVulnTable* pDvTable)
{
   if ( ! ( (TDetVulnBlock*)fpPrev )->AddVulnDetEntry(detType, pDvTable) ) {
      TDetVulnBlock* pNewBlock = new TDetVulnBlock(fpPrev);
      pNewBlock->AddVulnDetEntry(detType, pDvTable);
   }
}

cDetVulnList::cDetVulnList(cDetVulnList* pOriginal)
{
   if ( pOriginal ) {
      for ( TDetVulnBlock* pdvB = (TDetVulnBlock*)pOriginal->GetFirstLink();
                           pdvB = (TDetVulnBlock*)pOriginal->GetNULLLink();
                           pdvB = (TDetVulnBlock*)pOriginal->GetNextLink() ) {
         new TDetVulnBlock(fpPrev, pdvB);
      }
   }
}

TDetVulnTable*   TDetVulnBlock::GetVulnDetTable( int detType )
{
   for ( int i = 0; i < fNumEntries; i++ ) {
      if ( fDetVulnPair[i].fDetType == detType )
         return fDetVulnPair[i].fDetVulnTable;
   }
   return NULL;
}

EBoolean         TDetVulnBlock::AddVulnDetEntry(int detType, TDetVulnTable* pDvTable)
{
   if ( fNumEntries < eNumDetVulnPairs ) {
      fDetVulnPair[fNumEntries].fDetType = detType;
      fDetVulnPair[fNumEntries].fDetVulnTable = pDvTable;
      fNumEntries++;
      return true;
   }
   else
      return false;
}

TDetVulnBlock::TDetVulnBlock(T2link* pPrev, TDetVulnBlock* pdvOrig) :
   T2link(pPrev)
{
   fNumEntries = pdvOrig->fNumEntries;
   for ( int i = 0; i < fNumEntries; i++ )
      fDetVulnPair[i] = pdvOrig->fDetVulnPair[i];
}


cDetVulnList::cDetVulnList(int fuzeType, EBoolean bBlastHard)
{
    InitDetVulnList(fuzeType, bBlastHard);
}

void cDetVulnList::InitDetVulnList(int fuzeType, EBoolean bBlastHard)
{
   //cout << "Fuzetype in cDetVulnList is " << fuzeType << endl;
   switch (fuzeType) {
      case TFuzeParam::efuzetypePressure:
            if (bBlastHard) {
               AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbBhPres) );
               AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicBhPres) );
               AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastBhPres) );
               AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastBhPres) );
               AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastBhPres) );
               AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragBhPres) );
               AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragBhPres) );
               AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherBhPres) );
            }
            else {
               AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbPres) );
               AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicPres) );
               AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastPres) );
               AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastPres) );
               AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastPres) );
               AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragPres) );
               AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragPres) );
               AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherPres) );
            }

         break;
   
      case TFuzeParam::efuzetypeTiltrod:
            AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbTilt) );
            AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicTilt) );
            AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastTilt) );
            AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastTilt) );
            AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastTilt) );
            AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragTilt) );
            AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragTilt) );
            AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherTilt) );
         break;
   
      case TFuzeParam::efuzetypeApPressure:
      case TFuzeParam::efuzetypeApTrip:
            AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbAp) );
            AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicAp) );
            AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastAp) );
            AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastAp) );
            AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastAp) );
            AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragAp) );
            AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragAp) );
            AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherAp) );
         break;
   
      case TFuzeParam::efuzetypeCommandFire:
      case TFuzeParam::efuzetypeCommandLoc:
      case TFuzeParam::efuzetypeCommandFireDir:
            AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbCmdAp) );
            AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicCmdAp) );
            AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastCmdAp) );
            AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastCmdAp) );
            AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastCmdAp) );
            AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragCmdAp) );
            AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragCmdAp) );
            AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherCmdAp) );
         break;
   
      case TFuzeParam::esenstypeCommandCamDir:
      case TFuzeParam::esenstypeAutoCmdCamDir:
      case TFuzeParam::esenstypePerDet:
      case TFuzeParam::esenstypeVehDet:
      case TFuzeParam::esenstypePerVehDet:
      case TFuzeParam::esenstypeTripWire:
      case TFuzeParam::esenstypeDirDet:
      case TFuzeParam::esenstypePerTrack:
      case TFuzeParam::esenstypeVehTrack:
      case TFuzeParam::esenstypePerVehTrack:
      case TFuzeParam::esenstypeDirPVTrack:
      case TFuzeParam::esenstypeDirTrack:
            AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbCam) );
            AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicCam) );
            AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastCam) );
            AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastCam) );
            AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastCam) );
            AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragCam) );
            AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragCam) );
            AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherCam) );
         break;
   
      default:
            AddVulnDetEntry( cDetonation::edetEsmb, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblEsmbOther) );
            AddVulnDetEntry( cDetonation::edetMiclic, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMiclicOther) );
            AddVulnDetEntry( cDetonation::edetLgBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgBlastOther) );
            AddVulnDetEntry( cDetonation::edetMedBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblMedBlastOther) );
            AddVulnDetEntry( cDetonation::edetSmBlast, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmBlastOther) );
            AddVulnDetEntry( cDetonation::edetLgFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblLgFragOther) );
            AddVulnDetEntry( cDetonation::edetSmFrag, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblSmFragOther) );
            AddVulnDetEntry( cDetonation::edetOtherDet, (TDetVulnTable*) gDetThreatList.FindTableByIndex(TDetVulnTable::evdtblOtherOther) );
         break;
   }
}
